
require "util/weightedlist"
require "content/character_aspects/character_traits"
require "battle/battleutil"
require "content/opinion_events"
local chargen = require "charactergen"
local CastMember = require "castmember"
local memtokens = require "memtokens"

require "property_bag"
require "eventsystem"
require "agentutil"
require "brain"
-----------------------------------------------------------------
-- An instanced 'character' in the game


local Agent = class("Agent")
CastMember.AddInterface( Agent )
Agent.can_reload = true

function Agent:init( content_id, skin_table )
    self.tags = table.empty
    self.aspects = table.empty
    self.money = 0
    self.brain = Brain(self)

    local def = Content.GetCharacterDef( content_id )
    if not def then
        assert(nil, loc.format("Trying to spawn in invalid character def {1}", content_id))
    end

    assert( def, content_id )
    assert_warning( not def.is_template, "Shouldn't create Agent from definition that 'is_template': %s", content_id )

    if skin_table == nil and def.default_skin then
        skin_table = Content.GetCharacterSkin( def.default_skin )
    end

    self:InitializeAgent( def, skin_table )
end

function Agent:__index( k )
    local v = Agent[ k ]
    if v == nil then
        local aspects = rawget( self, "aspects" )
        v = aspects and aspects[ k ]
    end
    if v == nil then
        local def = rawget( self, "def" )
        v = def and def[k]
    end
    return v
end

function Agent:GetBrain()
    return self.brain
end

function Agent:GetMoney()
    return self.money
end

function Agent:SetMoney(v)
    self.money = math.round(v)
end

function Agent:DeltaMoney(delta)
    
    local old_money = self.money
    self.money = math.max(0, math.round( self.money + delta))

    if self:IsPlayer() then
        local actual_delta = self.money - old_money
        if actual_delta > 0 then
            self.money_gained = (self.money_gained or 0) + actual_delta
        end
        if actual_delta ~= 0 then
            TheGame:BroadcastEvent( "player_money_change", old_money, self.money )
        end
    end
end


function Agent:GainStaticAspects( def )
    -- Add any aspects defined by this character def.
    local aspects = def.aspects
    if aspects then
        for aspect_id, data in pairs( aspects ) do
            -- NOTE: GainAspect silently returns if the aspect already exists.
            if type(data) == "table" then
                self:GainAspect( aspect_id, data )
            elseif data then
                self:GainAspect( aspect_id )
            end
        end
    end
end

function Agent:RemoveStaticAspects( def )
    -- Remove any aspects defined by this character def.
    local aspects = def.aspects
    if aspects then
        for aspect_id, data in pairs( aspects ) do
            self:RemoveAspect( aspect_id )
        end
    end
end

function Agent:ReinitializeAgent( def )
    assert( is_instance( def, CharacterDef ))

    local idx = self.def:RemoveProperties( self.agent_def )
    self.def:PushProperties( def, idx )

    self.agent_def = def

    self:GainStaticAspects( def )
    self.loc_table = nil

    self:UpdateMaxHealth()
    self:UpdateMoraleRange()
end

function Agent:InitializeAgent( def, skin_table )
    assert( is_instance( def, CharacterDef ))
    
    assert( skin_table == nil or is_instance( skin_table, CharacterSkin ))
    assert( self.def == nil )

    self.def = PropertyBag()
    self.def:PushProperties( def )
    self.agent_def = def

    if skin_table then
        self.def:PushProperties( skin_table )
        self.skin_def = skin_table
    end

    -- No skin: run character gen.
    for i = #def.on_inits, 1, -1 do
        def.on_inits[ i ]( self )
    end

    assert( self.species, def.id )
    if self.species ~= SPECIES.MECH then
        assert( self.gender, def.id )
    end
    assert( self.faction_id, def.id )

    self:GainStaticAspects( def )

    self.loc_table = nil

    self:UpdateMaxHealth()
    self:UpdateMoraleRange()
end

function Agent:IsUnique()
    return self.def.unique == true or self:GetSkinID() ~= nil
end

function Agent:GetUniqueID()
    return self:GetSkinID() or self:GetContentID()
end

function Agent:GetSkinID()
    return self.skin_def and self.skin_def:GetContentID()
end

function Agent:GetSkinOverrides()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end

    local skin = {
        content_id = self:GetContentID(),
        name = self:GetName(),
        gender = self.gender,
        species = self.species,
        build = self.build,
        head = self.head,
        skin_colour = self.skin_colour,
        hair_colour = self.hair_colour,
    }
    return skin
end

function Agent:SetBuildOverride(id)
    self.build = id
    self:BroadcastEvent( "build_changed", self )
end

function Agent:GetContentID()
    return self.agent_def.id
end

function Agent:GetID()
    return self.guid
end

function Agent:GenerateRandomName()
    local props = SPECIES_PROPS[ self.species ]
    local name, nameid = Content.GenerateCharacterName( props and props.name_pool )
    self.nameid = nameid
    self:GenerateLocTable()
end

function Agent:GenerateLocTable()
    -- fullname: Title, nickname, and name.
    -- Usage: Anywhere, but mostly UI (titles are generally capitalized)
    --      Merchant Bob
    --      Mercenary 'Shadowstalker' Carl
    --      Admiralty Guard Dave
    --      Alpha Crayote

    if self.loc_table == nil then
        self.loc_table = setmetatable( {}, GENDER_META[ self.gender or GENDER.UNDISCLOSED ] )
    end

    local language = TheGame:GetLanguage()
    local name
	if self.custom_names and self.custom_names[ language ] then
        name = self.custom_names[ language ]

    elseif self.nameid then
        local props = SPECIES_PROPS[ self.species ]
        local declensions
        name, declensions = Content.GetCharacterName( props.name_pool, self.nameid )
        for case, name in pairs( declensions or table.empty ) do
            self.loc_table[ case ] = name
        end

    elseif self.skin_def and self.skin_def:GetLocalizedName( self ) then
        name = self.skin_def:GetLocalizedName( self )
        for case, name in pairs( self.skin_def:GetLocalizedNameDeclensions() ) do
            self.loc_table[ case ] = name
            self.loc_table[ "name_" .. case ] = name
        end

    else
        name = self.agent_def:GetLocalizedName( self )
    end
    
    for case, name in pairs( self.agent_def:GetLocalizedNameDeclensions() ) do
        if self.loc_table[ case ] == nil then
            self.loc_table[ case ] = name
            self.loc_table[ "name_" .. case ] = name
        end
    end

    local nickname = self.agent_def:GetLocalizedNickName( self ) or (self.skin_def and self.skin_def:GetLocalizedNickName( self ))

    local fullname
    if nickname and name then
        local pre, nick = nickname:match( "^([*]?)(.+)$" )
        if #pre > 0 then
            name = loc.format( "{1} '{2}'", name, nick )
        else
            name = loc.format( "'{1}' {2}", nick, name )
        end
    end

    local title = self.agent_def:GetLocalizedTitle( self )
    if self.trained_support_negotiation then
        if self:IsInPlayerParty() then
            -- Pets in the player party are either 'trained' or 'untrained'
            if self.support_negotiation == self.trained_support_negotiation then
                title = loc.format( LOC "UI.TRAINED_PET", title )
            else
                title = loc.format( LOC "UI.UNTRAINED_PET", title )
            end

        elseif self:GetParty() == nil or not self:GetParty():GetLeader():IsSentient() and not self:HasTag("NOT_WILD") then
            -- Pets without a sentient leader are "Wild"
            title = loc.format( LOC "UI.WILD_PET", title )
        end
    end

    if title and name then
        fullname = loc.format(LOC"UI.AGENT_TITLE_FORMAT_STR", name, title)
    elseif title then
        fullname = title
    else
        fullname = name
    end
    

    -- .the_desc, The_Desc, a_desc, A_desc: Noun descriptive with title or species.
    -- May be overriden by the character def.
    -- Usage: Sentences and dialog, hence the case variants.
    --      the Mysterious Stranger
    --      The Human Admiralty guard
    --      a Human Admiralty guard
    --      a Crayote
    --      an Alpha Crayote
    local desc = self.agent_def:GetLocalizedDesc( self )
    local species = self.species or SPECIES.HUMAN
    local species_str = loc.format( LOC("SPECIES." .. species), self )
    if desc == nil then
        if species ~= SPECIES.MECH and species ~= SPECIES.BEAST and title then
            desc = loc.format( LOC"AGENT.SPECIES_FORMAT_STR", species_str, title )
        elseif title then
            desc = title
        else
            desc = species_str
        end
    end

    local the_desc = loc.format( LOC"AGENT.THE_FORMAT_STR", desc, self )
    local a_desc = loc.format( LOC"AGENT.A_FORMAT_STR", desc, self )

    -- .identity: faction name and title combined, used for various UI elements.
    local identity
    local faction_name = self:GetFaction() and self:GetFaction():GetName()
    if faction_name and title then
        identity = loc.format( LOC"AGENT.IDENTITY_FORMAT_STR", faction_name, title )
    elseif faction_name then
        identity = faction_name
    else
        identity = title
    end

    -- .who, .Who:  Relation-specific description; the name or description with article 'the'.
    -- Usage: Sentences
    --      Bob
    --      the Shroke Merchant 
    --      the Alpha Crayote
    local who
    if name then
        who = name
    else
        who = the_desc
    end

    local bio = (self.skin_def and self.skin_def:GetLocalizedBio( self )) or self.agent_def:GetLocalizedBio( self )

    local t = self.loc_table
    t.name = name
    t.nick = nickname and nickname:match( "^[*]?(.+)$" ) or name
    t.fullname = fullname
    t.title = title
    t.who = who
    t.Who = loc.cap( t.who or "" )
    t.faction = self:GetFaction() and self:GetFaction():GetName() or "???"
    t.identity = identity
    t.desc = desc
    t.the_desc = the_desc
    t.a_desc = a_desc
    t.species = species_str
    t.bio = bio
    t.The_desc = loc.cap( t.the_desc )
    t.A_desc = loc.cap( t.a_desc )
    t.Species = loc.cap( t.species )

    self:BroadcastEvent( "loc_changed", self )
end

function Agent:LocTable()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table
end

function Agent:GetLoreUnlocks()
    do return table.empty end
    
    local unlocks = shallowcopy( self.lore_unlocks ) or {}
    if self.loved_bio then
        unlocks.loved_bio = self.loved_bio
    end
    if self.hated_bio then
        unlocks.hated_bio = self.hated_bio
    end
    if self.killed_bio then
        unlocks.killed_bio = self.killed_bio
    end
    if self.drank_bio then
        unlocks.drank_bio = self.drank_bio
    end

    return unlocks
end

function Agent:CanTalk()
    return self:IsSentient() or self.can_talk
end

function Agent:GetSpecies()
    return self.species
end

function Agent:HasAlias( alias )
    if alias == nil then
        return self.alias ~= nil
    else
        return self.alias == alias
    end
end

function Agent:GetAlias()
    return self.alias
end

function Agent:GetBio()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table.bio
end

-- Noun-based description
-- "Shroke Mercenary"
-- "Tall Merchant"
function Agent:GetDesc()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table.desc
end

-- "Smith"
function Agent:GetName()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end

    return self.loc_table.name or self:GetShortTitle()
end

function Agent:SetCustomName( name )
    if self.custom_names == nil then
        self.custom_names = {}
    end
    local language = TheGame:GetLanguage()
    self.custom_names[ language ] = name
    self:GenerateLocTable()
end

-- Similar to loc_table.who, but not meant to be used in a sentence (no article and capitalized)
-- "Captain Mishra"
-- "Tall Kra'deshi Merchant"
function Agent:GetIdentity()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table.identity
end

-- "Captain Mishra", "Bandit Lord Felgrimm 'The Legless'""
function Agent:GetFullName()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table.fullname
end

-- "Guard", "Merchant", "Shopkeeper"
function Agent:GetShortTitle()
    if self.loc_table == nil then
        self:GenerateLocTable()
    end
    return self.loc_table.title
end

function Agent:GetRenown()
    return self.def.renown
end

function Agent:ModifyCombatStrength( modifier )
    self.combat_strength_modifier = (self.combat_strength_modifier or 0) + modifier
end

function Agent:GetCombatStrength()
    if self:IsPlayer() then
        return TheGame:GetGameState():GetCurrentBaseDifficulty()
    end

    return (self.def.combat_strength or 1) + (self.combat_strength_modifier or 0)
end

function Agent:ListenForEvent( ... )
    if self.events == nil then
        self.events = EventSystem()
    end
    self.events:ListenForEvent( ... )
end


function Agent:ListenForAny(listener, fn)
    if self.events == nil then
        self.events = EventSystem()
    end
    self.events:ListenForAny( listener, fn )
end

function Agent:BroadcastEvent( event, ... )
    if self.events then
        self.events:BroadcastEvent( event, ... )
    end
end

function Agent:RemoveListener( listener )
    if self.events then
        self.events:RemoveListener( listener )
        if not self.events:HasListeners() then
            self.events = nil
        end
    end
end



function Agent:__tostring()
    if self:IsRetired() then
        return string.format("[%s:%s**]", self.guid or rawstring(self), tostring((self:GetName()) or self.id) )   
    elseif self:InLimbo() then
        return string.format("[%s:%s*]", self.guid or rawstring(self), tostring((self:GetName()) or self.id) )
    else
        return string.format("[%s:%s]", self.guid or rawstring(self), tostring((self:GetName()) or self.id) )
    end
end

function Agent:Load()
    local has_skin = self.def[2] ~= nil -- ccc: The second property class is ALWAYS a skin, if it exists.
    self.def:ResolveIDs()
    self.agent_def = self.def:LookupPropertiesByClass( CharacterDef )
    self.skin_def = self.def:LookupPropertiesByClass( CharacterSkin )

    if has_skin and self.skin_def == nil then
        -- No skin: run character gen.
        for i = #self.agent_def.on_inits, 1, -1 do
            self.agent_def.on_inits[ i ]( self )
        end
        print( self, " missing skin: regenerated properties" )
    end

    if self.aspects ~= self.def.aspects then
        self.sorted_aspects = {}
        for k,v in pairs(self.aspects) do
            if v.id ~= k then
                -- Old Aspect: probably deleted, or renamed.
                print( "MISSING ASPECT:", self, k )
                self.aspects[k] = nil
            else
                table.insert( self.sorted_aspects, v )
            end
        end
        table.sort( self.sorted_aspects, Aspect.SortCompare )
        for i, aspect in ipairs( self.sorted_aspects ) do
            if aspect.OnAgentReady then
                aspect:OnAgentReady( self )
            end
        end
    else
        self.aspects = table.empty
    end

    if rawget( self, "tags" ) == nil then
        self.tags = table.empty
    end
end

function Agent:PostLoad( game_state )
    if self.seed == nil then
        self.seed = game_state:SeededRandom( 1, 2^32-1 )
    end

    for i, aspect in self:Aspects() do
        if aspect.PostLoad then
            aspect:PostLoad( game_state )
        end
    end

    if self.support_negotiation then
        if Content.GetNegotiationCard( self.support_negotiation ) == nil then
            self.support_negotiation = nil
        end
    end

    if self.trained_support_negotiation then
        if Content.GetNegotiationCard( self.support_negotiation ) == nil then
            self.trained_support_negotiation = nil
        end
    end

    if self.brain then
        self.brain:PostLoad()
    end

end


function Agent:GetSummaryData()
    return {
        content_id = self:GetContentID(),
        skin_id = self:GetSkinID(),
        guid = self.guid,
        seed = self.seed,
        nameid = self.nameid,
        species = self.species,
        gender = self.gender,

        hair_colour = self.hair_colour,
        skin_colour = self.skin_colour,
        hair_colour = self.hair_colour,
        build = self.build,
        head = self.head,
    }
end


function Agent:Save()
    local data =
    {
        def = self.def,
        guid = self.guid,
        seed = self.seed,
        money = self.money,
        money_gained = self.money_gained,

        location = self.location,
        location_role = self.location_role,

        nameid = self.nameid,
        custom_names = self.custom_names,
        agent_buckets = self.agent_buckets,
        species = rawget( self, "species" ),
        gender = rawget( self, "gender" ),

        hair_colour = rawget( self, "hair_colour" ),
        voice_actor = rawget( self, "voice_actor" ),
        skin_colour = rawget( self, "skin_colour" ),
        hair_colour = rawget( self, "hair_colour" ),
        build = rawget( self, "build" ),
        head = rawget( self, "head" ),

        removed = self.removed,
        death_data = self.death_data,
        alias = self.alias,

        brain = self.brain,
        plax_slot_id = self.plax_slot_id,
        plax_name = self.plax_name,

        combat_strength_modifier = self.combat_strength_modifier,
        support_negotiation = self.support_negotiation,
        in_hiding = self.in_hiding,
        death_item_override = self.death_item_override,
        death_money_override = self.death_money_override,
    }

    if self.aspects and self.aspects ~= table.empty then
        data.aspects = self.aspects
    end

    if self.quest_membership and next(self.quest_membership) then
        data.quest_membership = self.quest_membership
    end
    if self.memory and next(self.memory) then
        data.memory = self.memory
    end
    if self.memory_negotiation and next(self.memory_negotiation) then
        data.memory_negotiation = self.memory_negotiation
    end

    if next(self.tags) then
        data.tags = self.tags
    end

    return data
end

-- One-time init when added to game state
function Agent:PopulateGameState( game_state )
    self.guid = game_state:GenerateID()
    self.seed = game_state:SeededRandom( 1, 2^32-1 )

    if TheGame:GetLocalSettings().DEBUG then
        self.spawn_traceback = debug.traceback()
    end

    -- Until this Agent is added officially to game state, it really shouldn't be in ANY buckets.
    assert( self.agent_buckets == nil )

    if self.faction_id then
        self:AddToBucket( self.faction_id )
    end
end

function Agent:GetSeededRandom( property_id )
    -- BOTH seeds must be unique if we expect the streams to be sufficiently different
    -- across Agent instances.
    -- This means we must perturb the property_id by something static and unique to this agent.
    local property_hash = engine.inst:HashString( property_id .. tostring(self.guid))
    return engine.Random( property_hash, self.seed or 0 )
end


-- Calculates a "virtual" property about this agent, where fn() is the property
-- generator function. The property is "virtual" because there is no concrete field for its value, it is deterministic and const for this Agent's lifetime.
-- eg: this yields a constant & deterministic "FEAR" value for agent in the range [1, 10]
--
-- agent:CalculateProperty( "FEAR", function( agent )
--      return math.random( 1, 10 )
-- end )

function Agent:CalculateProperty( property_id, fn )
    local rng = self:GetSeededRandom( property_id )
    push_random( function( n, m )
        return rng:Random( n, m )
    end )

    local ok, result = xpcall( fn, generic_error, self )
    if not ok then
        LOGWARN( result )
    end

    pop_random( rng )
    return result
end


function Agent:HasActiveQuestMembership()
    -- has membership in an "active" quest
    if not self:HasQuestMembership() then
        return false
    end

    for _, quest in pairs(self.quest_membership) do
        if quest:IsActive() then
            return true
        end
    end
end

function Agent:IsCastInQuest(test_quest)
    return self.quest_membership and table.arrayfind( self.quest_membership, test_quest)
end

function Agent:HasInactiveQuestMembership()
    -- has membership in an "active" quest
    if not self:HasQuestMembership() then
        return false
    end

    for _, quest in pairs(self.quest_membership) do
        if not quest:IsActive() then
            return true
        end
    end
end

function Agent:GetActiveQuests()
    local active_quests = {}

    if self:HasQuestMembership() then
    
        for _, quest in pairs(self.quest_membership) do
            if quest:IsActive() then
                table.insert( active_quests, quest )
            end
        end

    end
    
    return active_quests
end

function Agent:IsQuestProvider()
    if self.quest_membership then
        for _, quest in pairs(self.quest_membership) do
            if quest:IsActive() and quest:GetProvider() == self then
                return true, quest
            end
        end
    end
end

function Agent:HasQuestMembership()
    return self.quest_membership and (next(self.quest_membership) ~= nil)
end

function Agent:GetSocialConnections()
    return self.social_connections
end

function Agent:GetDeathData()
    return self.death_data
end

function Agent:GetHomeLocation()
    if self.brain then
        return self.brain:GetHome()
    end
end

function Agent:GetLocalizedString( key )
    return self.def:GetLocalizedString( key )
end

function Agent:GetLocation()
    return self.location
end

function Agent:IsRetired()
    return self.removed == true
end

function Agent:InLimbo()
    return self.location == nil
end

function Agent:AddToBucket( bucket_id )
    assert( self.guid ~= nil )
    
    if self.agent_buckets == nil then
        self.agent_buckets = {}
    end
    if table.arrayfind( self.agent_buckets, bucket_id ) ~= nil then
        assert(nil, "ALREADY IN THAT BUCKET " ..  tostring(bucket_id))
    end

    assert( table.arrayfind( self.agent_buckets, bucket_id ) == nil )
    table.insert( self.agent_buckets, bucket_id )
    TheGame:GetGameState():AddAgentToBucket( self, bucket_id )
end

function Agent:RemoveFromBucket( bucket_id )
    if self.agent_buckets then
        local i = table.arrayfind( self.agent_buckets, bucket_id )
        assert( i, bucket_id )
        table.remove( self.agent_buckets, i )
        TheGame:GetGameState():RemoveAgentFromBucket( self, bucket_id )
    end
end

--this is super-specific player party stuff that should be moved to caravan
function Agent:Recruit( member_type )
    assert( not self:IsInPlayerParty(), tostring(self) .. " is already a party member" )
    print( "Agent:Recruit", self, member_type )

    if self:GetParty() then
        self:GetParty():RemoveMember(self)
    end

    local gs = TheGame:GetGameState()

    gs:GetCaravan():AddMember( self, member_type or PARTY_MEMBER_TYPE.CREW )
    self:GenerateLocTable()

    if member_type == PARTY_MEMBER_TYPE.CAPTIVE then
        self:OpinionEvent(OPINION.TOOK_CAPTIVE)
    elseif member_type == PARTY_MEMBER_TYPE.HIRED then
        TheGame:AddGameplayStat( "hired_help", 1 )        
    end

    AgentUtil.CancelInactiveSideQuests(self)
    return self
end

--this is super-specific player party stuff that should be moved to caravan
function Agent:Dismiss()
    --print (debugstack())
    assert( not self:IsPlayer() )
    
    if not self:IsInPlayerParty() then
        return
    end

    print( "Agent:Dismiss()", self )

    local gs = TheGame:GetGameState()

    gs:GetCaravan():RemoveMember( self )

    -- New squad? In locations or elsewhere, no need to do anything.
    if self.location and self.location:CheckFlag( LOCFLAGS.MAP ) then
        self:GetBrain():MoveToHome()
    end
    self:BroadcastEvent( "dismissed", self )

    return self
    -- self.brain:Update()
end

local function MoveToLocationInternal( self, location )

    if self.location ~= location then
        --print( "MoveToLocationInternal", self, self.location, location, debugstack() )
 
        local old_loc = self.location       
        self.location = location

        if old_loc then
            local guid = old_loc:GetGUID()
            self.location_role = nil
            self.plax_slot_id = nil
            self.plax_name = nil

            old_loc:RemoveAgent( self )
        end

        if location then
            assert( not self.removed, tostring(self) )
        
            location:AddAgent( self )
            self.moveto_traceback = debug.traceback()
        else
            assert_warning( not self:IsPlayer() )

            self.limbo_traceback = debug.traceback()

            --There are no parties in limbo
            if self.party_member and not self:IsInPlayerParty() then
                self.party_member:GetParty():RemoveMember(self)
            end
        end

        self:BroadcastEvent( "agent_location_changed", self, old_loc, location )
    end
end

function Agent:MoveToLimbo()
    MoveToLocationInternal( self, nil )

    self.location_role = nil
    self.plax_slot_id = nil
    self.plax_name = nil
    return self
end

function Agent:MoveToLocation( location )
    assert( location )
    MoveToLocationInternal( self, location )
    return self
end

function Agent:MoveToAgent( agent )
    if agent.location then
        self:MoveToLocation( agent.location )
    else
        self:MoveToLimbo()
    end
    return self
end

function Agent:Retire()
    assert( not self.removed or error( "Retired at:" .. tostring(self.retire_traceback)))
    assert( not self:IsPlayer() )

    print( "Agent:Retire()", self )

    if self:IsInPlayerParty() then
        self:Dismiss()
    end

    if self.party_member then
        self.party_member:GetParty():RemoveMember(self)
    end

    self:MoveToLimbo()
    assert_warning( self:InLimbo() )

    local gs = TheGame:GetGameState()
    if self.guid then
        gs:RemoveAgent( self )
    end


    if TheGame:GetLocalSettings().DEBUG then
        self.retire_traceback = debug.traceback()
    end

    -- Clear a whackload of properties so they don't end up save-loaded.
    for aspect_id, aspect in pairs( self.aspects ) do
        if not aspect.keep_on_retire then
            self:RemoveAspect( aspect_id, true )
        end
    end
    if self.sorted_aspects and #self.sorted_aspects == 0 then
        self.aspects = table.empty
        self.sorted_aspects = nil
    end

    if self.agent_buckets then
        while #self.agent_buckets > 0 do
            gs:RemoveAgentFromBucket( self, table.remove( self.agent_buckets ))
        end
        self.agent_buckets = nil
    end


    self.brain:Clear()

    -- ccc: want to clear these too, but Encounter frequently queries agents (CreateScenario) after Retire().
    -- need to shore up those holes first.
    self.brain = nil
    self.quest_membership = nil
    self.memory = nil
    self.memory_negotiation = nil
    --self.voice_actor = nil --keep this because we might want to show them in a memorial
    self.sim = nil

    -- IsRetired() now officially returns true.
    self.removed = true

    self:BroadcastEvent( "agent_retired", self )
    return self
end

function Agent:Kill()
    self.death_data = {
        time = Now(),
    }

    self:AddTag( "dead" )

    if self.social_connections then
        local rel = self.social_connections:GetRelationship(TheGame:GetGameState():GetPlayerAgent()) 
        if rel == RELATIONSHIP.LOVED or rel == RELATIONSHIP.HATED then
            TheGame:GetGameState():GetPlayerAgent().graft_owner:RemoveSocialGraft(self)
        end
    end

    self:Retire()

    self:BroadcastEvent( "agent_killed", self )
    return self
end

function Agent:GetRoleAtLocation()
    return self.location_role
end

function Agent:SetLocationRole( role )
    if self.location_role ~= role then
        self.location_role = role
        if self.location then
            self.location:BroadcastEvent( "agent_role_changed", self.location, self, role )
        end
    end
end

function Agent:GetPlaxSlotID( plaxname )
    return self.plax_slot_id
end

function Agent:SetPlaxSlotID( slot_id )
    self.plax_slot_id = slot_id
end

function Agent:HasMemoryOfNegotiation( token, duration )
    if self.memory_negotiation == nil then
        return false
    else
        return memtokens.HasMemory( self.memory_negotiation, token, duration )
    end
end

function Agent:RememberNegotiation( token, duration )
    if self.memory_negotiation == nil then
        self.memory_negotiation = {}
    end

    return memtokens.Remember( self.memory_negotiation, token, duration )
end

function Agent:ForgetNegotiation( token )
    if token then
        return memtokens.Forget( self.memory_negotiation, token)
    
    elseif self.memory_negotiation then
        self.memory_negotiation = nil
    end
end

function Agent:HasMemory(token, duration)
    return memtokens.HasMemory( self.memory, token, duration)
end

function Agent:HasMemoryFromToday(token)
    local phase = TheGame:GetGameState():GetDayPhase()
    local duration = (phase == DAY_PHASE.DAY) and 1 or 2
    return memtokens.HasMemory( self.memory, token, duration)
end

function Agent:Remember(token, val)
    if self.memory == nil then
        self.memory = {}
    end
    return memtokens.Remember( self.memory, token, val )
end

function Agent:Forget(token)
    return memtokens.Forget( self.memory, token )
end

local function IterateAspects( state, i )
    local aspect
    while true do
        i, aspect = next( state.aspects, i )
        if not aspect then
            return
        else
            local ok = true
            for i, tag in ipairs( state ) do
                if not aspect[ tag ] then
                    ok = false
                    break
                end
            end
            if ok then
                return i, aspect
            end
        end
    end
end

-- Iterates over self.aspects, returning <idx, aspect> pairs, where aspect has non-nil values for all keys in the parameter list.
function Agent:Aspects( ... )
    return IterateAspects, { aspects = self.sorted_aspects or table.empty, ... }
end

function Agent:GetAspects()
    return self.aspects
end

function Agent:GetAspect( name )
    return self.aspects[ name ]
end

function Agent:FindAspect( fn )
    for i, aspect in ipairs( self.sorted_aspects ) do
        if fn( aspect ) then
            return aspect
        end
    end
end

function Agent:HasAspect( name )
    return self.aspects[ name ] ~= nil
end

local function SortAspects( x, y )
    return x.id < y.id -- *shrug*
end

function Agent:AddAspect(name, ...)
    return self:GainAspect(name, ...)
end

function Agent:AddAspectStacks( name, stacks )
    stacks = stacks or 1
    
    local aspect = self.aspects[ name ]
    if aspect then
        aspect:DeltaStacks(stacks)
        self:BroadcastEvent( "aspects_changed", self, true, aspect )
        return aspect

    else
        local aspect = self:AddAspect(name)
        if stacks > 1 then
            aspect:DeltaStacks(stacks-1)
        end
        return aspect
    end
end

function Agent:GainAspect( name, ... )
    local aspect = self.aspects[ name ]
    if aspect then
        local current, max = aspect:GetStacks()
        if max == 1 then
            return aspect
        else
            aspect:DeltaStacks(1)
            self:BroadcastEvent( "aspects_changed", self, true, aspect )
            return aspect
        end
    end

    assert( Content.GetAspect( name ), name )
    aspect = Content.GetAspect( name )( self, ... )

    if self.aspects == table.empty then
        self.aspects, self.sorted_aspects = {}, {}
    end
    self.aspects[ name ] = aspect
    table.binsert( self.sorted_aspects, aspect, Aspect.SortCompare )

    if aspect.OnGainAspect then
        aspect:OnGainAspect( self )
    end
    
    if aspect.OnAgentReady then
        aspect:OnAgentReady( self )
    end

    self:BroadcastEvent( "aspects_changed", self, true, aspect )
    
    return aspect
end

function Agent:IsInConversation()
    return TheGame:GetGameState():GetCaravan():GetCurrentEncounter() and TheGame:GetGameState():GetCaravan():GetCurrentEncounter():IsAgentPresent(self)
end

function Agent:RemoveAspect( name )
    assert( type(name) == "string" )
    local aspect = self.aspects[ name ]
    if aspect then
        self.aspects[ name ] = nil
        table.arrayremove( self.sorted_aspects, aspect )

        if aspect.OnLoseAspect then
            aspect:OnLoseAspect( self )
        end

        self:BroadcastEvent( "aspects_changed", self, false, aspect )
    end
end


function Agent:OnTimePass()
    if self.sorted_aspects then
        --this is done in a work table in case the agent adds or removes an aspect from an update
        local wt = ObtainWorkTable()
        table.arrayadd( wt, self.sorted_aspects )
        for k,aspect in ipairs(wt) do
            if aspect.OnTimePass then
                aspect:OnTimePass()
            end
        end
        ReleaseWorkTable(wt)
    end
end

--when the player rests, people reset timeouts, regenerate, etc.
function Agent:OnPlayerRest()

    self:ForgetNegotiation()

    if self.sorted_aspects then
        --this is done in a work table in case the agent adds or removes an aspect from an update
        local wt = ObtainWorkTable()
        table.arrayadd( wt, self.sorted_aspects )
        for k,aspect in ipairs(wt) do
            if aspect.OnPlayerRest then
                aspect:OnPlayerRest()
            end
        end
        ReleaseWorkTable(wt)
    end

end

function Agent:UpdateAgent()
    local now = Now()
    if now == self.last_update then
        return false
    end

    self.last_update = now

    if self.sorted_aspects then
        --this is done in a work table in case the agent adds or removes an aspect from an update
        local wt = ObtainWorkTable()
        table.arrayadd( wt, self.sorted_aspects )
        for k,aspect in ipairs(wt) do
            if aspect.UpdateAspect then
                aspect:UpdateAspect()
            end
        end
        ReleaseWorkTable(wt)
    end

    if self.removed then
        -- Aspects may retire us in which case there is nothing left to do.
        return
    end


    if not self:IsInPlayerParty() then
        if self:GetAspect("health") then
            self:GetAspect("health"):SetPercent(1)
        end

        -- Any skinned agents are real, any randos are not.
        if not self:IsUnique() and
            self:InLimbo() and
            not self:HasTag("NO_AUTO_CULL") and
            not self:HasQuestMembership() and
            not self:GetBrain():GetWorkPosition() then
            -- Cull randos who have no other important duties.
            print( "## CULLING:", self )
            self:Retire()
        end
    end

    return true
end


function Agent:UpdateMaxHealth()
    local new_max = self.fight_data.MAX_HEALTH

    if not self:HasTag("PC") then
        local scale

        if self.boss then
            scale = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_HEALTH )
        else
            scale = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_HEALTH ) or 1.0
        end
        new_max = math.round( new_max * scale )
    else
        local advancement_mod = GetAdvancementModifier( ADVANCEMENT_OPTION.LOWER_MAX_HP )
        if advancement_mod then
            new_max = math.round(new_max * advancement_mod)
        end
    end

    local health = self:GainAspect("health")
    local p = health:GetPercent()
    local current, max, min = health:Get()

    health:SetBaseMax(new_max)
    health:SetPercent(math.max( p, 1/new_max ))
end

function Agent:UpdateMoraleRange()
    local value
    if self.fight_data.MAX_MORALE then
        value = self.fight_data.MAX_MORALE
    end


    if (value or 0) > 0 then
        self:GainAspect("morale", value )
    end
end

function Agent:GetRallyReaction()
    return self.def.rally_reaction or "RALLIED"
end

function Agent:GetHealth()
    if self:GetAspect("health") then
        return self:GetAspect("health"):Get()
    end
end

function Agent:GetMorale()
    local morale_aspect = self:GetAspect("morale")
    if morale_aspect then
        return morale_aspect.value or morale_aspect.cap
    end
end

function Agent:GetFactionID()
    return self.faction_id
end

function Agent:GetFaction( game_state )
    if self.faction_id then
        game_state = game_state or TheGame:GetGameState()
        if game_state then
            return game_state:GetFaction( self.faction_id )
        end
    end
end

------------------------------------------------------------------

function Agent:IsPC()
    return self.faction_id == PLAYER_FACTION
end

function Agent:IsPlayer()
    return TheGame:GetGameState() and self == TheGame:GetGameState():GetPlayerAgent()
end

function Agent:IsInPlayerParty()
    return TheGame:GetGameState() and ( self:GetParty() == TheGame:GetGameState():GetCaravan():GetParty() )
end

function Agent:IsLeader()
    return self.aspects.party_member ~= nil and self.aspects.party_member:IsLeader()
end

function Agent:IsPet()
    return self:IsCrewMember() and not self:IsSentient()
end

function Agent:IsCrewMember()
    return self.aspects.party_member ~= nil and self.aspects.party_member:IsCrew()
end

function Agent:IsHiredMember()
    return self.aspects.party_member ~= nil and self.aspects.party_member:IsHired()
end

function Agent:IsEscortMember()
    return self.aspects.party_member ~= nil and self.aspects.party_member:IsEscort()
end

function Agent:IsCaptiveMember()
    return self.aspects.party_member ~= nil and self.aspects.party_member:IsCaptive()
end

local function IteratePartyMembers( agent, i )
    if agent:GetParty() then
        if i == nil then 
            return 1, agent.party_member:GetParty().members[1]
        else
            local agent = agent.party_member:GetParty().members[1+i]
            if agent then
                return i + 1, agent
            end
        end
    else
        if i == nil then
            return 1, agent
        end
    end
end

function Agent:Party()
    return IteratePartyMembers, self
end

function Agent:GetParty()
    return self.party_member and self.party_member:GetParty()
end

function Agent:IsBoss()
    return self.def.boss
end


function Agent:ClearParty()
    local party = self:GetParty()
    if party then
        party:Disband()
    end
end

function Agent:AddTag( tag )
    if self.tags == table.empty then
        self.tags = {}
    end
    table.insert_unique(self.tags, tag)
    return self
end

function Agent:RemoveTag( tag )
    if self.tags ~= table.empty then
        table.arrayremove(self.tags, tag)
    end
end

function Agent:HasTag( tag )
    if self.tags and table.arraycontains(self.tags, tag) then
        return true
    end
    
    for k,v in self.def:PropertyStack() do
        
        if v.tags and table.arraycontains( v.tags, tag ) then
            return true
        end
    end
	return false
end

function Agent:IsDead()
    return self.death_data ~= nil
end

function Agent:IsAlive()
    return not self:IsDead()
end

function Agent:LookupQuip( tags )
    if self.quest_membership then
        for _,quest in pairs(self.quest_membership) do
            local quip = quest:LookupQuip(tags)
            if quip then
                return quip
            end
        end
    end
    
    local quip = Content.LookupQuip(tags)
    return quip

end

function Agent:FillOutQuipTags(tags)
    table.arrayadd( tags, self.tags )
    if self.def.tags then
        table.arrayadd( tags, self.def.tags )
    end

    
    if TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent() then
        if TheGame:GetGameState():GetPlayerAgent().player_quip_tag then
            table.insert( tags, TheGame:GetGameState():GetPlayerAgent().player_quip_tag)
        end
    end

    
    table.insert_unique(tags, self:GetContentID():lower())

    if self:IsUnique() then
        table.insert( tags, self:GetName() )
    end

    local faction_id = self:GetFactionID()
    if faction_id then
        table.insert_unique(tags, faction_id:lower())
    end

    if self.alias then
        table.insert_unique( tags, self.alias:lower() )
    end
    table.insert_unique( tags, self.species:lower() )

    if not self:IsPlayer() then
        local rel = self:GetRelationship()
        -- DO NOT add hated/loved quips: this forces us to write too many categories.
        if rel < RELATIONSHIP.NEUTRAL then
            table.insert_unique( tags, "disliked" )
        elseif rel > RELATIONSHIP.NEUTRAL then
            table.insert_unique( tags, "liked" )
        else
            table.insert_unique( tags, "neutral" )
        end

        if self:IsInPlayerParty() then
            table.insert_unique( tags, "in_party" )
        elseif self:KnowsPlayer() then
            table.insert_unique( tags, "met" )
        end
    end

    if self:IsDead() then
        table.insert_unique(tags, "dead")
    end


    if not TheGame:GetGameState() or TheGame:GetGameState():GetDayPhase() == DAY_PHASE.DAY then
        table.insert_unique( tags, "day" )
    else
        table.insert_unique( tags, "night" )
    end

    if self.location then
        for tag, v in self.location:Tags() do 
           table.insert_unique( tags, tag )
        end
        if self.location:GetContent().tags then
            for i, tag in ipairs( self.location:GetContent().tags ) do
                table.insert_unique( tags, tag )
            end
        end
        if self.location:GetAgentByRole("guard") == self then
            table.insert_unique(tags, "guard")
        end
    end

    if TheGame:GetGameState() then
        local why_event = self.social_connections and self.social_connections:GetRelationshipReason(TheGame:GetGameState():GetPlayerAgent())
        if why_event then
            table.insert_unique(tags, why_event.id)
        end

    end

    for k,v in pairs(self.aspects) do
        if v:IsQuipTag() then
            table.insert_unique(tags, v.id)
        end
    end
end

local BASE_BUILDS = 
{
    [GENDER.FEMALE] = "medium_female",
    [GENDER.MALE] = "medium_male",
}

function Agent:GetBaseBuild()
    local base_build
    local gender = self.gender

    -- Character def specification.
    if self.build then
        base_build = self.build
    elseif self.def.base_builds then
        base_build = self.def.base_builds[ gender ]
    end

    -- ...Fallback to species.
    if base_build == nil then
        local props = SPECIES_PROPS[ self.species ]
        base_build = props and props.base_build
    end

    -- ..Fallback to faction specification
    if base_build == nil then
        local faction = self:GetFaction()
        if faction and faction:GetContent().base_builds then
            base_build = faction:GetContent().base_builds[ gender ]
        end
    end

    -- ..Fallback to gender builds.
    if base_build == nil then
        base_build = BASE_BUILDS[ gender ]
    end
    return base_build
end

function Agent:GetBuild()
    local build_id = self:GetBaseBuild()

    local build =  Content.GetCharacterBuild(build_id)
    if not build then
        assert(false, string.format("COULD NOT FIND BUILD: %s", build_id))
    end

    return build
end

function Agent:IsCurrentlyImportant()
    if TheGame:GetGameState() then
        for i, party_member in self:Party() do
            if TheGame:GetGameState():GetTracker():IsTracked( party_member ) then
                return true
            end
        end
    end

    return false
end


function Agent:JoinParty( boss )
    if self:GetParty() and self:GetParty() == boss:GetParty() then
        return
    end

    if self:GetParty() then
        self:GetParty():RemoveMember(self)
    end

    if not boss:GetParty() then
        local party = Party()
        party:AddMember(boss, PARTY_MEMBER_TYPE.LEADER)
    end


    boss:GetParty():AddMember(self)
    return self
end

function Agent:IsInSameParty( agent )
    return self:GetParty() and self:GetParty() == agent:GetParty()
end

function Agent:IsWithAgent( agent )
    return self.location == agent.location
end

function Agent:IsSentient()
    
    if self.is_hologram then
        return false
    end

    if SPECIES_PROPS[ self.species ] and SPECIES_PROPS[ self.species ].sentient then
        return true
    else
        return false
    end
end

-----------------BEHAVIOUR MANAGEMENT

function Agent:IsEnforcing()
    if not self:IsInPlayerParty() then
        return self:HasTag("enforcer") and not self:HasAspect( "intimidated" )
    end
end


function Agent:KnowsPlayer()
    local player = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    return self.social_connections and player and self.social_connections:HasOpinionEvents( player )
end

function Agent:KnowsAgent( agent )
    return self.social_connections and self.social_connections:GetRelationship( agent ) ~= RELATIONSHIP.NEUTRAL
end

function Agent:GetRelationship( agent )
    agent = agent or (TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent())
    if agent and self.social_connections then
        return self.social_connections:GetRelationship( agent )
    end

    return RELATIONSHIP.NEUTRAL
end

function Agent:IsLiked( agent )
    return self:GetRelationship( agent ) >= RELATIONSHIP.LIKED
end

function Agent:IsDisliked( agent )
    return self:GetRelationship( agent ) <= RELATIONSHIP.DISLIKED
end

function Agent:IsLoved( agent )
    return self:GetRelationship( agent ) >= RELATIONSHIP.LOVED
end

function Agent:IsHated( agent )
    return self:GetRelationship( agent ) <= RELATIONSHIP.HATED
end

-----------------------------------------------------------------------------------------------------------------------------

function Agent:GetFactionRelationship( faction_id )
    return TheGame:GetGameState():GetFactions():GetFactionRelationship( self.faction_id, faction_id )
end

function Agent:OpinionEvent( opinion_event_data, instance_data, from_agent )
    if self.social_connections then
        assert_warning( not is_instance( instance_data, Agent ), "Wrong parameters to OpinionEvent" )
        from_agent = from_agent or TheGame:GetGameState():GetPlayerAgent()
        local event = self.social_connections:ProcessOpinionEvent(opinion_event_data, instance_data, from_agent)
        return event
    end
end

function Agent:GetWorkBoss()
    local work_owner = self:GetBrain() and self:GetBrain():GetWorkPosition() and self:GetBrain():GetWorkPosition().location
    if is_instance(work_owner, Location) then
        if not self:GetBrain():GetWorkPosition():GetData().no_boss then
            
            local prop = work_owner:GetProprietor()
            if prop and prop ~= self then
                return prop
            end
        end
    end
end

function Agent:ClearStatusEffects()
    local aspects_to_remove = ObtainWorkTable()
    
    for k,aspect in pairs(self.aspects) do
        if aspect:IsStatusEffect() then
            table.insert( aspects_to_remove, aspect )
        end
    end

    for k, aspect in ipairs(aspects_to_remove) do
        self:RemoveAspect(aspect.id)
    end

    ReleaseWorkTable(aspects_to_remove)
end

function Agent:GetSocialGraft(rel)
    if rel == RELATIONSHIP.LOVED then
        if self.def.loved_graft and self.def.loved_graft ~= "" then
            return self.def.loved_graft
        end
        
    elseif rel == RELATIONSHIP.HATED then
        if self.def.hated_graft and self.def.hated_graft ~= "" then
            return self.def.hated_graft
        end
    end
end

function Agent:CollectQuestMarks(all_marks, in_location)
    all_marks = all_marks or {}
    local marks = ObtainWorkTable()


    for k, quest in TheGame:GetGameState():ActiveQuests() do
        table.clear(marks)
        quest:CollectMarkedCast(marks, in_location)
        for obj, objectives in pairs(marks) do
            if obj == self then
                table.insert(all_marks, {agent = obj, quest = quest, objectives = objectives})
                break
            end
        end
    end

    ReleaseWorkTable(marks)
    return all_marks
end

function Agent:LeaveHiding()
    self.in_hiding = false
end

function Agent:SendToHiding()
    self.in_hiding = true
    self:MoveToLimbo()
end

function Agent:GetDeathMoney()
    if TheGame:GetGameState() and TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY then
        return 0
    end

    return self.death_money_override or self.def.death_money or 0
end

function Agent:GetDeathItem()
    return self.death_item_override or self.def.death_item
end

function Agent:GetCombatVOEvent(vo)
    if self.voice_actor then
        self.vo_combat_events = self.vo_combat_events or {}
        if not self.vo_combat_events[vo] then
            --local actor = "admiraltyHighClassFemale01" -- self.voice_actor
            self.vo_combat_events[vo] = "event:/vo/combat/" .. self.voice_actor .. "/" .. vo
        end
        return self.vo_combat_events[vo]
    end
end


function Agent:TrainPet()
    if not self:IsPet() then
        print("ATTEMPTED TO TRAIN NON-PET", self)
        return
    elseif self.support_negotiation == "trained_pet" then
        print("ATTEMPTED TO TRAIN ALREADY TRAINED PET", self, self.support_negotiation)
        return
    end
    self.health:AddStatModifier( "TRAINED_PET", 10 )
    self.support_negotiation = self.trained_support_negotiation
    self:GenerateLocTable()
end

function Agent:IsTrained()
    return self:IsPet() and self.support_negotiation == "trained_pet"
end