local ANIM_META = {}
local ANIM_CACHE = {} -- Caching engine.asset.Anim.
local ANIM_INFO_CACHE = {} -- Caching state info per anim (indices, durations)
local FX_TRACKS = {} -- Array of fx tracks.

function CacheAnim(anim)
    local asset = ANIM_CACHE[ anim ]
    if asset == nil then
        asset = engine.asset.Anim( anim )
        ANIM_CACHE[ anim ] = asset
        --print ("CACHING ANIM ", anim)
    end
end

-- Should just be for debugging.
function GetAnimCache()
    return ANIM_CACHE, ANIM_INFO_CACHE
end

function GetAnim(anim, state)

    local asset = ANIM_CACHE[ anim ]
    if asset == nil then
        asset = engine.asset.Anim( anim )
        --ANIM_CACHE[ anim ] = asset
    end

    local asset_info
    if state then
        asset_info = ANIM_INFO_CACHE[ anim ]
        if asset_info == nil then
            asset_info = {}
            for k = 1, asset:GetNumAnims() do
                local name = asset:GetAnimName(k-1)
                local dur = asset:GetAnimDuration(k-1)
                asset_info[ name ] = { idx = k-1, duration=dur }
            end
            --ANIM_INFO_CACHE[ anim ] = asset_info
        end
    end

    return asset, asset_info and asset_info[ state ]
end

function ForgetAnim(anim)
    ANIM_CACHE[anim] = nil
    ANIM_INFO_CACHE[anim] = nil
end

function LoadAnimMeta()
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/anim/*.lua')) do
        local name = filename:match("^(.+)[.]lua$")
        local filepath = string.format( "content/anim/%s", name )
        package.loaded[ filepath ] = nil
        if name == GetFXTrackFilename() then
            FX_TRACKS = require( filepath )
        else
            ANIM_META[ name ] = require( filepath )
        end
    end
end

function SaveAnimMeta( filename )
    local anims
    if filename == GetFXTrackFilename() then
        anims = FX_TRACKS
    else
        anims = ANIM_META[ filename ]
    end
    if anims == nil then
        LOGWARN( "Missing anim meta information for filename '%s'", filename )
        return
    end

    local SERPENT_OPT = { comment = false, nocode = true, sortkeys = true,  indent = "\t" }

    local outs = 
    {
        "-- Autogenerated: DO NOT EDIT\n",
        "local sounds = require \"sound_event_names\"\n",
        "local vo = require \"voice_event_names\"\n\n",

        serpent.dump( anims, SERPENT_OPT )
    }
    outs = table.concat( outs )

    return outs
end

function GetAnimMetaFilename( animfile )
    return string.match( animfile:GetFileName(), "anim/(.+)[.]zip$" )
end

function GetFilenameFromAnimMeta( t )
    for anim_file, anims in pairs( ANIM_META ) do
        for anim, meta in pairs( anims ) do
            if meta == t then
                return anim_file
            end
        end
    end
end

function GetAnimMeta(filename, anim)
    local anim_table = ANIM_META[ filename ]
    if anim_table == nil then
        anim_table = {}
        print( "Created anim meta table for ", filename, anim )
        -- print( debug.traceback() )
        ANIM_META[ filename ] = anim_table
    end
    if anim == nil then
        return anim_table, filename
    end

    local anim_meta = anim_table[ anim ]

    if anim_meta and anim_meta.filename then
        assert( filename ~= anim_meta.filename )
        return GetAnimMeta( anim_meta.filename, anim ), anim_meta.filename

    elseif anim_meta then
        return anim_meta, filename

    elseif anim_table._parent then
        -- Inherit data from another anim meta table.
        return GetAnimMeta( anim_table._parent, anim ), anim_table._parent
    end
    return nil, filename
end

function GetFXTrackFilename()
    return "generic_fx"
end

function GetFXTrack( name )
    if name then
        return FX_TRACKS and FX_TRACKS[ name ]
    else
        return FX_TRACKS
    end
end

function GetAnimMetaTable()
    return ANIM_META
end

-- These anims are preloaded whenever you begin a game, and stay resident, to avoid having to load
-- whenever going into battle.
local PRELOAD_COMBAT_ANIMS = 
{
    "anim/card_expend_mask_fx.zip",
    "anim/card_expend_mask_fx.zip",
    "anim/card_electric_fx.zip",
    "anim/card_sticky_fx.zip",
    "anim/UI_combat_character_intent_fx.zip",
    "anim/UI_negotiation_round_fx.zip",
    "anim/card_flourish_fx.zip",
    "anim/UI_healthbar_hits_fx.zip",
}

local function ProcessAnimsInTracks( anim_file, anims, fn )
    for anim, meta in pairs( anims ) do
        if meta.fx_tracks then
            for i, v in ipairs( meta.fx_tracks ) do
                if v.fx and v.fx ~= "" then
                    fn(v.fx)
                end
                if v.buildfile and v.buildfile ~= "" then
                    fn(v.buildfile)
                end
            end
        end
    end
end

function PreloadFXTracks()
    for i, anim_file in ipairs(PRELOAD_COMBAT_ANIMS) do
        CacheAnim(anim_file)
    end

    ProcessAnimsInTracks( GetFXTrackFilename(), FX_TRACKS, CacheAnim )

    for anim_file, anims in pairs( ANIM_META ) do
        ProcessAnimsInTracks( anim_file, anims, CacheAnim )
    end
end

function UnloadFXTracks()
    for anim_file, anims in pairs( ANIM_META ) do
        ProcessAnimsInTracks( anim_file, anims, ForgetAnim )
    end

    ProcessAnimsInTracks( GetFXTrackFilename(), FX_TRACKS, ForgetAnim )

    for i, anim_file in ipairs(PRELOAD_COMBAT_ANIMS) do
        ForgetAnim(anim_file)
    end
end


LoadAnimMeta()
