
--this is so that you don't need to nil-check requested sound instances when the sound system is disabled. 
local DummySoundInstance = class("DummySoundInstance")
function DummySoundInstance:init() end
function DummySoundInstance:Start() end
function DummySoundInstance:Stop() end
function DummySoundInstance:SetPaused() end
function DummySoundInstance:Release() end
function DummySoundInstance:IsValid() return false end
function DummySoundInstance:SetPos() end
function DummySoundInstance:SetParameterValue() end
function DummySoundInstance:IsVirtual() return false end
function DummySoundInstance:GetPlaybackState() return engine.sound.STOPPED end
function DummySoundInstance:SetVolume() end
function DummySoundInstance:SetUserData() end
function DummySoundInstance:GetUserData() return 0 end

local DummySoundDesc = class("DummySoundDesc")
function DummySoundDesc:CreateInstance() end
function DummySoundDesc:PlayOneShot() end
function DummySoundDesc:IsOneShot() return true end



local AudioSystem = class("AudioSystem")

function AudioSystem:init(enabled)
    self.enabled = enabled
    self.sound_banks = {}
    self.current_sounds = {}
    self.recent_sounds = {}
    self.speech_event_lookup = {}
    self.phonemes_per_actor = {}
    self.recent_start_idx = 1
    self.played_stats = {}
    self.phoneme_lookup = {}
    self.last_played_time = {}
    self.ambience_events = {}
    self.music_events = {}
    self.default_bank = nil

    self.default_audio_device = engine.inst:GetSetting("audio", "default_output_device", true)

    if not self.default_audio_device then
        local preferred_audio_device = engine.inst:GetSetting("audio", "output_device", "")
        if preferred_audio_device and preferred_audio_device ~= "" then
            local devices = engine.sound and engine.sound:GetOutputDevices()
            for k, v in pairs(devices or table.empty) do
                if v.name == preferred_audio_device then
                    self.audio_device_name = v.name
                    engine.sound:SetOutputDevice(v.id)
                    break
                end
            end
        end
    end
end

function AudioSystem:IsEnabled()
    return self.enabled
end

function AudioSystem:Reset()
    for k,v in pairs(self.current_sounds) do
        if k:IsValid() then
            k:Stop(true)    -- stop right now
        end
    end

    self.current_sounds = {}
    self.recent_sounds = {}

    local force_mono = engine.inst:GetSetting("audio", "force_mono", false)
    if force_mono then
        self:SetForceMono(true)
    end
    
end


function AudioSystem:GetNumSpeechSamples(actor, emotion, is_question)
    if self.speech_event_lookup[actor] and self.speech_event_lookup[actor][emotion] then
        local t = is_question and self.speech_event_lookup[actor][emotion].question or self.speech_event_lookup[actor][emotion].statement
        return #t
    end
    return 0
end

function AudioSystem:GetSpeechEvent(actor, emotion, is_question, idx)    
    if self.speech_event_lookup[actor] and self.speech_event_lookup[actor][emotion] then
        local t = is_question and self.speech_event_lookup[actor][emotion].question or self.speech_event_lookup[actor][emotion].statement
        return t[idx and clamp(idx, 1, #t) or math.random(#t)]
    end
end



local SPECIES_VALS = 
{
    [SPECIES.HUMAN] = 1,
    [SPECIES.SHROKE] = 2,
    [SPECIES.KRADESHI] = 3,
    [SPECIES.JARACKLE] = 4,
    [SPECIES.PHICKET] = 5,
}

function AudioSystem:PlaySpeechEvent(actor, emotion, is_question, species)

    if not self.speech_event_lookup[actor] then
        print ("INVALID VOICE ACTOR", actor)
        return
    end

    if not self.speech_event_lookup[actor][emotion] then
        print ("INVALID VOICE EMOTION FOR ACTOR", actor, emotion)
        emotion = "neutralDefault"
        if not self.speech_event_lookup[actor][emotion] then
            return
        end
    end

    if self.speech_event_lookup[actor] and self.speech_event_lookup[actor][emotion] then
        local t = is_question and self.speech_event_lookup[actor][emotion].question or self.speech_event_lookup[actor][emotion].statement
        
        if t and #t > 0 then
            local sorted = ObtainWorkTable()
            for k,v in ipairs(t) do
                table.insert(sorted, v)
            end
            
            table.shuffle(sorted)
            
            table.stable_sort(sorted, function(a, b) 
                local a_played = self.played_stats[a.event_name] or 0
                local b_played = self.played_stats[b.event_name] or 0
                if a_played == b_played then
                    local a_play_time = self.last_played_time[a.event_name] or 0
                    local b_play_time = self.last_played_time[a.event_name] or 0
                    return a_play_time < b_play_time
                else
                    return a_played < b_played
                end
            end)

            local info = sorted[1]
            ReleaseWorkTable(sorted)
            self.played_stats[info.event_name] = self.played_stats[info.event_name] and self.played_stats[info.event_name] + 1 or 1
            self.last_played_time[info.event_name] = GetWallTime()
            
            local inst = self:CreateEventInstance(info.event_name, "sound/vo.bank") 
            local species_val = species and SPECIES_VALS[species]
            if species_val then
                inst:SetParameterValue("species", species_val)
            end
            
            AUDIO:LogSound( info.event_name )
            if inst then 
                inst:Start()
                return info, inst
            end
        end
    end
end


function AudioSystem:GetVOMetaForEvent(event_name)
    for actor, actor_dat in pairs(self.speech_event_lookup) do
        for voice, voice_data in pairs(actor_dat) do
            for vtype, vtype_data in pairs(voice_data) do
                for _, dat in ipairs(vtype_data) do
                    if dat.event_name == event_name then
                        return dat
                    end
                end     
            end
        end
    end
end


function AudioSystem:GetForceMono()
    return self.force_mono == true
end

function AudioSystem:SetForceMono(force_mono)
    print ("FORCE MONO", self.force_mono, force_mono)
    if self.force_mono ~= force_mono then
        
        if not force_mono and self.force_mono_event then
            self.force_mono_event:Stop()
            self.force_mono_event:Release()
            self.force_mono_event = nil
            print ("STOPPING EVENT")
        end

        if force_mono then
            self.force_mono_event = self:CreateEventInstance("event:/mix_events/MonoMix")
            self.force_mono_event:Start()
            print ("PLAYING EVENT", self.force_mono_event)
        end
        self.force_mono = force_mono
    end
end

function AudioSystem:RegenVoiceData(actor_id, event_name)
    

    local WAVE_FILE_FINDER = {
        chemist = "civilianlowclassfemale01",
    }

    local info = self:GetVOMetaForEvent(event_name)
    if info then
        
        pcall(function() 

            local meta_filename =  "vo/".. actor_id ..".lua"
            local data = {}
            pcall(function() data = dofile(meta_filename) end)
            local file_name = "vo_"..info.id
            local old_file_name = file_name

            if WAVE_FILE_FINDER[actor_id] then
                file_name = file_name:gsub(actor_id, WAVE_FILE_FINDER[actor_id])
            end

            local wav_file_name = loc.format( "../contentsrc/fmod/planetoids/Assets/vo/non_proc/{1}.wav", file_name)
            local cmd = {"..\\tools\\rhubarb-lip-sync-1.8.0-win32\\rhubarb.exe", wav_file_name, "--consoleLevel error"}
            local exec = table.concat(cmd, " ") 
            print (exec)
            
            local f = io.popen(exec)
            local new_phonemes = {}
            for line in f:lines() do
                local _,_,time, char = string.find(line, "(.*)\t(.*)")
                if time and char then
                    table.insert( new_phonemes, tonumber(time))
                    table.insert( new_phonemes, char)
                end
            end
            f:close()
            
            data[old_file_name] = new_phonemes
            engine.data:SaveFile( meta_filename, serpent.dump(data, {sparse=false})) 
            self.phoneme_lookup[info.id] = new_phonemes

        end)
        
    end
end

function AudioSystem:ProcessAmbienceBank(bank_info)
    if not bank_info or not self.enabled then return end
    for event_name, event_def in pairs(bank_info.events) do
        local ambience_name = event_name:match("sfx/ambience/(.*)")
        if ambience_name then
            table.insert(self.ambience_events, {txt = ambience_name, event_id = event_name, event_def = event_def})
        end
    end

    table.sort(self.ambience_events, function(a,b) return a.txt < b.txt end)
end

function AudioSystem:ProcessMusicBank(bank_info)
    if not bank_info or not self.enabled then return end

    for event_name, event_def in pairs(bank_info.events) do
        local music_name = event_name:match("music/(.*)")
        if music_name then
            table.insert(self.music_events, {txt = music_name, event_id = event_name, event_def = event_def})
        end
    end

    table.sort(self.music_events, function(a,b) return a.txt < b.txt end)
end


function AudioSystem:MatchEvents(pattern, bank_info)
    bank_info = bank_info or self.default_bank
    local ret = {}
    if not bank_info or not self.enabled then return end

    for event_name, event_def in pairs(bank_info.events) do
        local music_name = event_name:match(pattern)
        if music_name then
            table.insert(ret, {txt = music_name, event_id = event_name, event_def = event_def})
        end
    end

    table.sort(self.music_events, function(a,b) return a.txt < b.txt end)
    return ret
end


function AudioSystem:GetPhonemes(id)
    return self.phoneme_lookup[id]
end


    
--[[
    local fname = 'scripts/content/vo_data.lua'
    print("Saving vo data to " .. fname)

    local save_str = serpent.block(event_lookup, {comment=false})

    save_str = loc.format("--THIS FILE IS AUTOGENERATED BY THE EDITOR. DO NOT CHANGE MANUALLY\nreturn {1}, {2}", save_str)

    engine.data:SaveFile( fname, save_str, 
        function(ok) 
            if ok then 
                print("VO " .. fname .. " saved successfully.")
            else
                print("ERROR: Can't save VO " .. fname)
            end
            if onsavedfn then onsavedfn(ok) end
        end)    
end
--]]

function AudioSystem:LoadVOMetadata()
    
    engine.inst:ProfilerPush("AudioSystem:LoadVOMetadata")
    local CONVO_DIR = 'vo'
    self.phonemes_per_actor = {}
    
    local filepath = require "util/filepath"
    
    local t0 = engine.inst:QueryPreciseTime()
    for k, filepath in ipairs( filepath.list_files( CONVO_DIR, "*.lua")) do
        local name = filepath:match( "vo/(.+)[.]lua$" )
        if name then
            self.phonemes_per_actor[name] = dofile(filepath)
        end
    end

    self.speech_event_lookup = require "content/vo_data"
    
    for actor, actor_dat in pairs(self.speech_event_lookup) do
        for voice, voice_data in pairs(actor_dat) do
            for vtype, vtype_data in pairs(voice_data) do
                for _, dat in ipairs(vtype_data) do
                    self.phoneme_lookup[dat.id] = self.phonemes_per_actor[actor] and self.phonemes_per_actor[actor]["vo_"..dat.id]
                end     
            end
        end
    end

    engine.inst:ProfilerPop()
    local t1 = engine.inst:QueryPreciseTime()
    print( "Load phonemes time:", t1-t0 )
end


function AudioSystem:ProcessSpeechBank(bank_info)
    if not bank_info or not self.enabled then return end

    engine.inst:ProfilerPush("AudioSystem:ProcessSpeechBank")
    local t1 = engine.inst:QueryPreciseTime()

    for event_name, event_def in pairs(bank_info.events) do

        local _, _, actor = string.find(event_name, "dialogue/([%w_]*)/")
        local _, _, emo = string.find(event_name, "/conversation/(%w+)_")
        
        local _, _, num = string.find(event_name, "(%d+)$")
        local is_question = string.find(event_name, "_question_") ~= nil
        
        if actor and emo and num then
            self.speech_event_lookup[actor] = self.speech_event_lookup[actor] or {}
            self.speech_event_lookup[actor][emo] = self.speech_event_lookup[actor][emo] or {}
            
            self.speech_event_lookup[actor][emo].statement = self.speech_event_lookup[actor][emo].statement or {}
            self.speech_event_lookup[actor][emo].question = self.speech_event_lookup[actor][emo].question or {}
            
            --local phoneme_lookup = loc.format("vo_{1}_{2}_{3}_{4}", actor, emo, is_question and "question" or "statement", num)
            local id = loc.format("{1}_{2}_{3}_{4}", actor, emo, is_question and "question" or "statement", num)
            local file_name = "vo_"..id
            
            --print (actor, phoneme_lookup, all_phonemes[actor] and all_phonemes[actor][phoneme_lookup])
            local info = {
                --actor_name = actor, 
                id = id, 
                event_name = event_name, 
            }


            self.phoneme_lookup[id] = self.phonemes_per_actor[actor] and self.phonemes_per_actor[actor][file_name]

            if is_question then
                table.insert(self.speech_event_lookup[actor][emo].question, info )
                --info.idx = #self.speech_event_lookup[actor][emo].question
            else
                table.insert(self.speech_event_lookup[actor][emo].statement, info )
                --info.idx = #self.speech_event_lookup[actor][emo].statement
            end
        end
    end
    local t2 = engine.inst:QueryPreciseTime()
    print( "event iteration:", t2-t1 )

    
    local fname = 'scripts/content/vo_data.lua'
    print("Saving vo data to " .. fname)


    local save_str = serpent.block(self.speech_event_lookup, {compact = true, comment=false})

    save_str = loc.format("--THIS FILE IS AUTOGENERATED BY THE EDITOR. DO NOT CHANGE MANUALLY\nreturn {1}", save_str)

    engine.data:SaveFile( fname, save_str, 
        function(ok) 
            if ok then 
                print("VO " .. fname .. " saved successfully.")
            else
                print("ERROR: Can't save VO " .. fname)
            end
            --if onsavedfn then onsavedfn(ok) end
        end)    
    --]]
    engine.inst:ProfilerPop()
end


function AudioSystem:SaveSpeechMeta()
    self:FinalizeVoiceLoad()

    local bank_info = self.sound_banks["sound/vo.bank"]
    local events = {}
    for event_name, event_def in pairs(bank_info.events) do

        local _, _, actor = string.find(event_name, "dialogue/([%w_]*)/")
        local _, _, emo = string.find(event_name, "/conversation/(%w+)_")
        
        local _, _, num = string.find(event_name, "(%d+)$")
        local is_question = string.find(event_name, "_question_") ~= nil
        
        if actor and emo and num then
            events[actor] = events[actor] or {}
            events[actor][emo] = events[actor][emo] or {}
            
            events[actor][emo].statement = events[actor][emo].statement or {}
            events[actor][emo].question = events[actor][emo].question or {}
            
            local id = loc.format("{1}_{2}_{3}_{4}", actor, emo, is_question and "question" or "statement", num)
            local file_name = "vo_"..id
            local info = {
                id = id, 
                event_name = event_name, 
            }
            if is_question then
                table.insert(events[actor][emo].question, info )
            else
                table.insert(events[actor][emo].statement, info )
            end
        end
    end

    local fname = 'scripts/content/vo_data.lua'
    print("Saving vo data to " .. fname)
    local save_str = serpent.block(events, {compact = true, comment=false})
    save_str = loc.format("--THIS FILE IS AUTOGENERATED BY THE EDITOR. DO NOT CHANGE MANUALLY\nreturn {1}", save_str)
    engine.data:SaveFile( fname, save_str, 
        function(ok) 
            if ok then 
                print("VO " .. fname .. " saved successfully.")
            else
                print("ERROR: Can't save VO " .. fname)
            end
        end)    
    engine.inst:ProfilerPop()

    self.speech_event_lookup = events
end


function AudioSystem:DoInitialLoad()
    if not self.enabled then
        return
    end

    self:LoadVOMetadata()
    AUDIO:LoadBank("sound/Main Bank.strings.bank", false)

    -- Do Main Bank with asynchronous loading.
    self.ma_bank_handle = AUDIO:LoadBank("sound/Main Bank.bank", true)

    -- Do Voice Over with asynchronous loading.
    self.vo_bank_handle = AUDIO:LoadBank("sound/vo.bank", true)
end


function AudioSystem:FinalizeMainLoad()
    if not (self.enabled and self.ma_bank_handle) then
        return
    end
    print("FINALIZE LOAD OF Main Bank")
    local master_bank = AUDIO:FinalizeBankLoad(self.ma_bank_handle)
    if master_bank then
        self.sound_banks["sound/Main Bank.bank"] = master_bank
        local master_bank_info = self.sound_banks["sound/Main Bank.bank"]
        AUDIO:ProcessAmbienceBank(master_bank_info)
        AUDIO:ProcessMusicBank(master_bank_info)
        AUDIO:SetDefaultBank(master_bank_info.id)
    end
    self.ma_bank_handle = nil
end 


function AudioSystem:FinalizeVoiceLoad()
    if not (self.enabled and self.vo_bank_handle) then
        return
    end
    print("FINALIZE LOAD OF SPEECH BANK")
    local speech_bank = AUDIO:FinalizeBankLoad(self.vo_bank_handle)
    if speech_bank then
        self.sound_banks["sound/vo.bank"] = speech_bank
        self.speech_bank = speech_bank
    end
    self.vo_bank_handle = nil
end 


function AudioSystem:LoadBank(bnk, async)
    if not self.enabled then
        return
    end
    if self.sound_banks[bnk] then
        --already loaded
    else        
        --do the loading
        local bank = engine.sound:LoadBank( bnk, async)
		if not bank then
		    print("failed to load", bnk)
		end
        if bank and not async then
            local bank_info = {id = bnk, bank = bank, events = bank:GetEvents(), buses = bank:GetBusList(), vcas = bank:GetVCAList() }
            self.sound_banks[bnk] = bank_info
        end
        return bank
    end
end

function AudioSystem:FinalizeBankLoad(bnk)
    if not self.enabled then
        return
    end

    local bank = engine.sound:FinalizeBankLoad(bnk)
    if not bank then
        assert(nil, "COULD NOT LOAD SOUND BANK (is there another copy of the game running?)")
    end
    local bank_info = {id = bnk, bank = bank, events = bank:GetEvents(), buses = bank:GetBusList(), vcas = bank:GetVCAList() }
    self.sound_banks[bnk] = bank_info
    return bank_info
end


function AudioSystem:SetDefaultBank(id)
    self.default_bank = self.sound_banks[id]
    self.default_bank_id = id
end

function AudioSystem:StopSpeech()
    if not self.enabled then return end
end

--"sound/vo.bank"
function AudioSystem:PlayParamEventBank( str, bank, ... )
    local n = select("#", ...)
    if n <= 0 then
        self:PlayEvent(str, bank)
    else
        local inst, desc = self:CreateEventInstance( str, bank )
        local params
        if inst then
            assert(desc:IsOneShot(), "Attempted to play looping sound using :PlayParamEvent(), use :CreateInstance() instead!")
            for n=1,select('#',...), 2 do
                params = params or {}
                local name = select(n,...)
                local val = select(n+1,...)
                if name and val then
                    inst:SetParameterValue(name, val)
                    params[name] = val
                end
            end
            inst:Start()
            inst:Release()
        end
        AUDIO:LogSound( str, nil, params )
    end

end

--this is for playing a simple one shot event WITH parameters.
function AudioSystem:PlayParamEvent( str, ... )
    self:PlayParamEventBank(str, self.default_bank_id, ... )
end

function AudioSystem:FindEvent(str)
    if not self.enabled then
        return
    end
    if (str == "" or not str) then
        return
    end

    for bank_name, bank in pairs( self.sound_banks ) do
        if bank.events[ str ] then
            return bank_name, bank
        end
    end
end


--this is for playing a simple one shot event with no parameters. Its is set up in a way so as to not create much garbage
function AudioSystem:PlayEvent(str, bank_name, x, y, z)
    if not self.enabled then return end
    if (str == "" or not str) then return end

    local bank = self.default_bank
    if bank_name then
        bank = self.sound_banks[bank_name]
        if not bank then
            print ("SOUND BANK DOES NOT EXIST:", bank_name)
            return
        end
    end

    if bank then
        local desc = bank.events[str]
        if not desc then
            print ("MISSING SOUND EVENT:", str)
            return
        end
        desc:PlayOneShot(x,y,z)
        self:LogSound(str, {x=x,y=y,z=z})
    end 

end

function AudioSystem:IsOneShot(str, bank_name)
    local bank = self.default_bank
    if bank_name then
        bank = self.sound_banks[bank_name]
        if not bank then
            print ("SOUND BANK DOES NOT EXIST:", bank_name)
            return
        end
    end

    if bank then
        local desc = bank.events[str]
        if not desc then
            print ("MISSING SOUND EVENT:", str)
            return
        end
        return desc:IsOneShot()
    end 
end

function AudioSystem:GetBus(bus_name, bank_name)

    local bank = self.default_bank
    if bank_name then
        bank = self.sound_banks[bank_name]
        if not bank then
            print ("SOUND BANK DOES NOT EXIST:", bank_name)
            return
        end
    end

    if bank and bank.buses[bus_name] then
        return bank.buses[bus_name]
    end

    print ("SOUND BUS DOES NOT EXIST:", bus_name)
end


function AudioSystem:GetVCA(vca_name, bank_name)

    local bank = self.default_bank
    if bank_name then
        bank = self.sound_banks[bank_name]
        if not bank then
            print ("SOUND BANK DOES NOT EXIST:", bank_name)
            return
        end
    end

    if bank and bank.vcas[vca_name] then
        return bank.vcas[vca_name]
    end

    print ("SOUND VCA DOES NOT EXIST:", vca_name)
end


local MAX_HISTORY = 64
function AudioSystem:LogSound(name, guid, params)
    if not self.enabled then return end

    local rec = {name=name, guid=guid, time = GetWallTime(), params = params}

    if TheGame:GetLocalSettings().SHOW_AUDIO_CALLSTACK then
        rec["callstack"] = debugstack()
    end

    if #self.recent_sounds < MAX_HISTORY then
        table.insert(self.recent_sounds, rec)
    else

        self.recent_sounds[self.recent_start_idx] = rec
        self.recent_start_idx = self.recent_start_idx + 1
        if self.recent_start_idx > MAX_HISTORY then
            self.recent_start_idx = 1
        end
    end
end

--for looping sounds or sounds that you need to set parameters...
function AudioSystem:CreateEventInstance(str, bank_name)
    local inst, desc
    if self.enabled then 
        local bank = self.default_bank
        if bank_name then
            bank = self.sound_banks[bank_name]
            if not bank then
                print ("SOUND BANK DOES NOT EXIST:", bank_name)
        		-- print (debug.traceback())
            end
        end

        if bank then
            desc = bank.events[str]
            if not desc then
                print ("MISSING SOUND EVENT: ", str)
		print (debug.traceback())
            else
                inst = desc:CreateInstance()
            end
        else
            print ("NO SOUND BANKS LOADED")
        end
    end

    if inst and desc then
        self.current_sounds[inst] = str
        return inst, desc
    else
        return DummySoundInstance(), DummySoundDesc()
    end
end

function AudioSystem:Update(dt)
    local num = 0
    for k,v in pairs(self.current_sounds) do
        if not k:IsValid() then
            self.current_sounds[k] = nil
        else
            num = num + 1
        end
    end

    self:CheckForDefaultAudioChange()
end

function AudioSystem:GetAudioDeviceOverrideName()
    if not self.default_audio_device then
        return self.audio_device_name
    else
        return 
    end
end

function AudioSystem:SetDefaultOutputDevice()
    local audio_devices = engine.sound and engine.sound:GetOutputDevices() or {}
    local device = audio_devices[1]
    if device then
        engine.sound:SetOutputDevice(device.id)
        self.audio_device_name = device.name
        self.default_audio_device = true
        engine.inst:SetSetting("audio", "default_output_device", true)
        engine.inst:SetSetting("audio", "output_device", "")
    end
end

function AudioSystem:SetOutputDeviceByID(id)
    local audio_devices = engine.sound and engine.sound:GetOutputDevices() or {}
    for k,v in ipairs(audio_devices) do
        if v.id == id then
            engine.sound:SetOutputDevice(v.id)
            self.audio_device_name = v.name
            self.default_audio_device = false
            engine.inst:SetSetting("audio", "output_device", v.name)
            engine.inst:SetSetting("audio", "default_output_device", false)
            break
        end
    end
end

function AudioSystem:CheckForDefaultAudioChange()
    if self.default_audio_device and engine.sound and engine.sound:AreDevicesDirty() then
        engine.sound:ClearDevicesDirty()
        local audio_devices = engine.sound and engine.sound:GetOutputDevices()
        local device = audio_devices[1]
        if self.audio_device_name ~= device.name then
            engine.sound:SetOutputDevice(device.id)
            self.audio_device_name = device.name
        end
    end
end


function AudioSystem:DumpDebugInfo()

    print("----------------------AUDIO INFO---------------------------")


    print( "LIVE SOUNDS")
    do
        local snds = {}
        local tots = {}
        for inst,name in pairs(self.current_sounds) do
            snds[name] = snds[name] or {}
            if inst:GetUserData() > 0 then
                tots[name] = (tots[name] or 0) + 1
            end
            table.insert(snds[name], inst)
        end
        
        for i, name, v in sorted_pairs(snds) do
            print(  string.format("%s (%d/%d)", name, (tots[name] or 0), #snds[name]) )
        end
    end

    local ambience = TheGame.ambience
    local music = TheGame.music

    print( loc.format( "MUSIC: {1}", music and music.music_event_name or "NONE"))
    print( loc.format( "AMBIENCE: {1}", ambience and ambience.ambience_event_id or "NONE"))

    print( "RECENT SOUND EVENTS:")
    local now = GetWallTime()
    local idx = self.recent_start_idx
    for k = 1, math.min(#self.recent_sounds, 10) do
        idx = idx - 1 
        if idx == 0 then
            idx =  #self.recent_sounds
        end

        local snd = self.recent_sounds[idx]
        if snd then
            local time_ago = now - snd.time
            print( string.format( "\t%s (%.2f ago)", snd.name, time_ago) )
            if snd.params then
                for k,v in pairs(snd.params) do
                    print( loc.format("\t{1}:{2}", k, v))
                end
            end
        end            
    end
    print( "-----------------------------------------------------------")
end



local PET_SOUNDS = {
    AFFIRMATIVE_YAP = {
        VROC = "event:/sfx/enemy/vroc/conversation/affirmative_yap",
        FLEAD = "event:/sfx/enemy/flead/conversation/affirmative_yap",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/affirmative_yap",
    },
    ANGRY_GROWL = {
        VROC = "event:/sfx/enemy/vroc/conversation/angry_growl",
        FLEAD = "event:/sfx/enemy/flead/conversation/angry_growl",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/angry_growl",
    },
    EXCITED_BORK = {
        VROC = "event:/sfx/enemy/vroc/conversation/excited_bork",
        FLEAD = "event:/sfx/enemy/flead/conversation/excited_bork",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/excited_bork",
    },
    PANT = {
        VROC = "event:/sfx/enemy/vroc/conversation/pant",
        FLEAD = "event:/sfx/enemy/flead/conversation/pant",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/pant",
    },
    SCARED_YELP = {
        VROC = "event:/sfx/enemy/vroc/conversation/scared_yelp",
        FLEAD = "event:/sfx/enemy/flead/conversation/scared_yelp",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/scared_yelp",
    },
    THREATEN = {
        VROC = "event:/sfx/enemy/vroc/conversation/threaten",
        FLEAD = "event:/sfx/enemy/flead/conversation/threaten",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/threaten",
    },
    WHIMPER = {
        VROC = "event:/sfx/enemy/vroc/conversation/whimper",
        FLEAD = "event:/sfx/enemy/flead/conversation/whimper",
        CRAYOTE = "event:/sfx/enemy/crayote/conversation/whimper",
    }
}

function AudioSystem:PlayPetSounds(pet, event_id)
    if PET_SOUNDS[event_id][pet:GetContentID()] and PET_SOUNDS[event_id][pet:GetContentID()] then
        self:PlayEvent(PET_SOUNDS[event_id][pet:GetContentID()] )
    end
end

