local battle_defs = require "battle/battle_defs"

local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

--------------------------------------------------------------------------
-- A single hit upon a target within a BattleAttack.
-- Each hit has a target, damage, and various hit results independent of the others.

local BattleHit = class( "Battle.Hit" )
BattleHit.can_reload = true

function BattleHit:SetHitResult( key, value )
    if value == nil then
        value = true
    end

    self[ key ] = value
end

function BattleHit:CheckHitResult( key )
    return self[ key ] == true
end

--------------------------------------------------------------------------
-- Wraps a Battle.Card for when it attacks a primary target.

local BattleAttack = class( "Battle.Attack" )
BattleAttack.can_reload = true

function BattleAttack:init( resolve_card )
    self.card = resolve_card
    self.name = resolve_card:GetName()
    self.icon = resolve_card.icon
    self.flags = resolve_card.flags
    self.engine = resolve_card.engine

    if self.engine then
        if resolve_card:GetRandSeed() then
            self.rand = engine.Random( self.engine.battle_seed, resolve_card:GetRandSeed() )
        else
            self.rand = engine.Random( self.engine.battle_seed, self.engine.resolve_count )
        end
        self.damage_accumulators = {}
        self.hit_accumulator = CardEngine.ScalarAccumulator( self.engine, BATTLE_EVENT.CALC_HIT_COUNT )
    end

    self.attacker = resolve_card:GetOwner()
    self.owner = self.attacker
    
    self.anim = resolve_card.anim
    self.hit_tags = shallowcopy( resolve_card.hit_tags ) -- Potentially modified dynamically.

    self.target = resolve_card:GetTarget()

    self.total_damage = 0
end

function BattleAttack:GenerateHits( hit_count )
    local card = self.card
    self.hits = {}

    local target_count = card.target_count or 1
    if hit_count == nil then
        hit_count = self.hit_accumulator and self.hit_accumulator:CalculateValue( card.hit_count or 1, card, self.target ) or card.hit_count or 1
    end

    -- Collect a set of targets for each target_count.
    for i = 1, target_count do
        local targets = card.secondary_targets and card.secondary_targets[i] or table.empty
        for j, target in ipairs( targets ) do

            -- Shouldn't hit_tags be per-target?
            if target:HasHitTag("small") then
                table.insert_unique( self.hit_tags, "small" )
            end
            if target:HasHitTag("large") then
                table.insert_unique( self.hit_tags, "large" )
            end
                        
            -- Hit each target once for each hit_count.
            for k = 1, hit_count do
                local hit = self:AddHit( target, (i-1) * hit_count + k, i )
                hit.targets = targets
                self:RollHitDamage( hit )
                if k == hit_count then
                    hit.features = shallowcopy( card.features )
                end
            end
        end
    end
end

function BattleAttack:ReassignTarget( target, new_target )
    if self.target == target then
        self.target = new_target
    end

    for i, hit in ipairs( self.hits ) do
        if hit.target == target then
            hit.target = new_target
            self:RollHitDamage( hit )
        end
    end
end

function BattleAttack:AddHit( target, hit_index, target_index )
    if hit_index == nil then
        hit_index = 1
        for i, hit in ipairs( self.hits ) do
            if hit.target == target then
                hit_index = hit_index + 1
            end
        end
    end

    local hit = setmetatable(
    {
        hit_index = hit_index,
        target_index = target_index,
        attacker = self.attacker,
        target = target,
        card = self.card,
        attack = self,
        piercing = self.card:IsFlagged( battle_defs.CARD_FLAGS.PIERCING )
    }, BattleHit )

    table.insert( self.hits, hit )
    return hit
end

function BattleAttack:RollHitDamage( hit )
    local target = hit.target

    if not self.engine then
        -- Editor only branch: For DoHitResult, having damage changes the presentation.
        hit.damage = math.random( 5, 20 )

    elseif self.card:IsDamageCard() then
        if self.card.no_modifiers then
            hit.min_damage, hit.max_damage = self.card.min_damage, self.card.max_damage
        else
            if self.damage_accumulators[ hit ] == nil then
                self.damage_accumulators[ hit ] = Battle.DamageAccumulator( self.engine )
            end
            hit.min_damage, hit.max_damage = self.damage_accumulators[ hit ]:PreviewDamage( self.card, target, hit )
        end

        if self.rand then
            hit.damage = self.rand:Random( hit.min_damage, hit.max_damage )
        else
            hit.damage = math.random( hit.min_damage, hit.max_damage )
        end
    end

    return hit
end

function BattleAttack:IsDamageModifier( source )
    for hit, acc in pairs( self.damage_accumulators ) do
        if acc:IsSource( source ) then
            return true
        end
    end
    return false
end

function BattleAttack:Hits()
    return ipairs( self.hits )
end

function BattleAttack:FindNextHit( hit )
    if hit == nil then
        return self.hits[ 1 ]
    else
        local idx = table.arrayfind( self.hits, hit )
        if idx then
            return self.hits[ idx + 1 ]
        end
    end
end

function BattleAttack:IsLastHitForTarget( hit )
    local i = table.arrayfind( self.hits, hit )
    if i then
        for j = i + 1, #self.hits do
            if self.hits[j].target == hit.target then
                return false
            end
        end
        return true
    end

    return false
end

function BattleAttack:GetHitCount()
    local hit_index
    for i, hit in ipairs( self.hits ) do
        if hit_index == nil or hit.hit_index > hit_index then
            hit_index = hit.hit_index
        end
    end
    return hit_index or 0
end

function BattleAttack:IsTarget( fighter, target_index )
    for i, hit in ipairs( self.hits ) do
        if hit.target == fighter and (target_index == nil or hit.target_index == target_index) then
            return true
        end
    end
    return false
end

function BattleAttack:CountUniqueTargets()
    local t = ObtainWorkTable()
    for i, hit in ipairs( self.hits ) do
        table.insert_unique( t, hit.target )
    end
    local count = #t
    ReleaseWorkTable( t )
    return count
end

function BattleAttack:IsApplyingFeature( fighter, feature_name )
    for i, hit in ipairs( self.hits ) do
        if hit.target == fighter and hit.features and hit.features[ feature_name ] then
            return true
        end
    end
    return false
end

-- Applies the condition to each distinct target at most once (or only to target, if non-nil)
function BattleAttack:AddCondition( condition_id, stacks, source, target )
    local targets = {}
    for i, hit in ipairs( self.hits ) do
        if hit.evaded and Content.GetBattleCondition( condition_id ).ctype == CTYPE.DEBUFF then
            -- DEBUFFS are evaded.  This is case-by-case for fully defended attacks.
        elseif not table.contains( targets, hit.target ) and (target == nil or target == hit.target) then
            hit.target:AddCondition( condition_id, stacks, source )
            table.insert( targets, hit.target )
        end
    end
end

-- is 'key' true for ALL hit results for the given fighter? (at least one hit result required)
function BattleAttack:CheckHitResult( fighter, key )
    local found
    for i, hit in ipairs( self.hits ) do
        if hit.target == fighter then
            if hit[ key ] ~= true then
                return false
            end
            found = true
        end
    end

    return found
end

-- is 'key' true for ALL hits?
function BattleAttack:CheckHitResults( key )
    local ok = true
    for i, hit in ipairs( self.hits ) do
        if hit[ key ]  ~= true then
            ok = false
            break
        end
    end
    return ok
end

function BattleAttack:GetTotalDamage()
    return self.total_damage
end

function BattleAttack:RunAttack()
    local card = self.card
    local rand = self.rand
    local last_hit_index

    local any_small, any_large = false
    local i = 1
    while i <= #self.hits do
        local hit = self.hits[i]
        if hit.target.team and not table.contains( hit.target.team.fighters, hit.target ) then
            -- No longer active.  IF target_mod is RANDOMX and the attacker is the player, then retarget, otherwise discard the hit entirely.
            local new_target
            if self.attacker:IsPlayer() and
                (card.target_mod == TARGET_MOD.RANDOM1 or card.target_mod == TARGET_MOD.RANDOM2 or card.target_mod == TARGET_MOD.RANDOM3) then
                --
                local targets = self.engine:CollectAllTargets( self.card )
                for i = #targets, 1, -1 do
                    -- Exclude anyone that is already targetted for this target_index.
                    if self:IsTarget( targets[i], hit.target_index ) then
                        table.remove( targets, i )
                    end
                end
                new_target = table.arraypick( targets )
            end
            if new_target then
                hit.target = new_target
                self:RollHitDamage( hit )
            else
                -- No new target, remove this hit entirely. (presentation will expect it to do something otherwise)
                table.remove( self.hits, i )
            end
        end

        if self.hits[i] == hit then
            if hit.damage then
                local damage_dealt = hit.target:ApplyDamage( hit.damage, card:GetOwner(), hit, hit.piercing )
                if self.attacker:IsPlayer() then
                    local def = Content.FindAchievement( "fist_o_bricks" )
                    if def and damage_dealt >= def.count then
                        TheGame:UnlockAchievement( def.id )
                    end
                end

                self.total_damage = self.total_damage + damage_dealt
                hit.damage_dealt = damage_dealt

                if card.threshold then
                    local threshold = self.engine:CalculateThreshold( card )
                    if hit.damage >= threshold then
                        card:ThresholdEffect( self.engine, self, hit )
                        self.engine:BroadcastEvent( BATTLE_EVENT.THRESHOLD_TRIGGERED, self.engine, card, hit.target )
                        table.insert( self.hit_tags, "threshold" )
                    end
                end
                if self.attacker:IsPlayer() and damage_dealt == hit.max_damage and hit.min_damage ~= hit.max_damage then
                    table.insert( self.hit_tags, "critical" )
                end
            end
        end

        -- Even if this hit was removed/invalidated, we still should attempt to apply appropriate features since some
        -- may effect the attacker (eg, Defend, Combo, ...)
        if hit.features then
            for feature_name, stacks in pairs( hit.features ) do
                local feature = Content.GetBattleCardFeature( feature_name )
                if feature and feature.apply then
                    feature:apply( self.engine, self, stacks, hit.target )
                end
            end
        end

        if self.hits[i] == hit then
            self.engine:BroadcastEvent( BATTLE_EVENT.ON_HIT, self.engine, self, hit )
            i = i + 1
        end
    end


    if self:CheckHitResults( "evaded" ) then
        table.insert( self.hit_tags, "evaded" )
    end

    if self:CheckHitResults( "defended" ) then
        table.insert( self.hit_tags, "defended" )
    end

end


function BattleAttack:__tostring()
    return string.format( "Attack.%s", self.card.id )
end
