require "eventsystem"

local METTLESOME_CHANCE = {0, .05, .1, .15, 0.2}
local METTLESOME_AMT = { 3, 4, 5, 6, 9 }

local STATUS = MakeEnum{
    "INACTIVE", -- Engine hasn't been started yet.
    "ACTIVE", -- Currently executing.
    "PLAY_CARD", -- Player is choosing a card from hand to play.
    "CHOOSE_CARDS", -- Player is choosing card(s), such as what to expend/etc.
    "SYNC", -- Waiting for UI to sync.
}

local BATTLE_EVENT = MakeEnum
{
    "BEGIN_TURN",
    "END_TURN",

    "BEGIN_PLAYER_TURN",
    "SHUFFLE_DISCARDS",
    "END_PLAYER_TURN",

    "BEGIN_ENEMY_TURN",
    "END_ENEMY_TURN",

    "ACTIVATED_FIGHTER",
    "DEACTIVATED_FIGHTER",
    "FIGHTER_KILLED",
    "ADDED_FIGHTER",
    "SETTLE_FIGHTERS",
    "FIGHTER_REMAP_ANIMS",
    "FIGHTER_RALLY",
    "FIGHTER_SURRENDER",
    "FIGHTER_PREVIEW",
    
    "START_RESOLVE",
    "END_RESOLVE",
    "PRE_RESOLVE",      -- Before resolving an action
    "ON_HIT",           -- Used to sync hit presentation to anim.
    "ON_RIPOSTE",       -- Attack was riposted
    "ON_DODGE",
    "POST_RESOLVE",
    "BEGIN_COMBAT",
    "END_COMBAT",

    "DELTA_ACTION_COUNT",
    "DELTA_STAT",
    "HEAL_HEALTH",
    "DELTA_MONEY",

    "DRAW_CARD",

    "CARD_DEALT",
    "CARD_PINNED",
    "CARDS_STOLEN",
    "CARD_MOVED",
    "CARD_CHANGED",
    "CARD_DISCARDED",
    "CARD_EXPENDED",
    "CARD_TRIGGERED",
    "CARD_TRIGGERED_PRE",
    "CARD_TRIGGERED_POST",
    "CARD_IMPROVISED",
    "CARD_UPGRADED",

    "DECK_TOO_LARGE",

    "STATUS_CHANGED",   -- Fighter status changed.
    "CONDITION_ADDED",
    "CONDITION_REMOVED",
    "CONDITION_BLOCKED",
    "REFRESH_CONDITIONS",
    "GRAFT_TRIGGERED",
    "REFRESH_GRAFTS",
    "DAMAGE_APPLIED",
    "THRESHOLD_TRIGGERED",

    "CALC_DAMAGE",
    "CALC_MORALE",
    "CALC_THRESHOLD",
    "CALC_ACTION_COST", -- Event for calculating the action cost of cards
    "CALC_MODIFY_STACKS", -- Event for calculating stacks of conditions when applied.
    "CALC_TURN_DRAW",  -- Event for calculating the number of cards drawn on turn
    "CALC_CARD_DRAW", -- For calculating draw count after the initial turn draw.
    "CALC_COMBO_LOSS", -- Event for calculating combo lost when receiving damage
    "CALC_CARD_FLAGS", -- Event for calculating flags on a Battle.Card
    "CALC_TARGET_MOD", -- Event for calculating a card's target_mod.
    "CALC_HIT_COUNT", -- Event for calculating how a card's hit_count.

    "UPDATE_HOME_POSITIONS",
    "EXECUTES_ADDED",
    "CHECKED_FOR_SURRENDER",

    "QUIP",
    "SPEECH",
    "SOUND",
    "DELAY",
    "PLAY_ANIM",
    "PLAY_FX",
    "PLAY_FLOATER",
    "CUSTOM",
    "CUSTOM_BEGIN",
    "CUSTOM_END",
    "FIGHTER_REPLACED",
    "FLOURISH_CHANGED",
}


local CARD_FLAGS, CARD_FLAG_BITSTRINGS = MakeBitField
{
    -- Cannot be played.
    "UNPLAYABLE",
    -- Automatically played at end of turn.
    "AUTOPLAY",
    -- "one-shot": Moved into the trash deck once played.
    "EXPEND",
    -- Destroys this card, permanently (eg. from your meta deck)
    "CONSUME",
    -- "multi-shot": Stays in hand when resolved.
    "MULTISHOT",
    -- This card's patience cost is variable/calculated dynamically.
    -- Shows "X" in place of the patience cost.
    "VARIABLE_COST",
    -- Costs 0 until played.
    "FREEBIE",
    -- Extend card draw by 1 when drawn.
    "REPLENISH",
    -- This card is an upgraded version
    "UPGRADED",
    -- This card is inserted into your hand at the start of combat if you have advantage.
    "AMBUSH",
    -- This card requires and consumes COMBO points on the card owner.
    "COMBO_FINISHER",
    -- This card damage ignores DEFEND
    "PIERCING",
    -- This is an NPC card.
    "NPC",
    -- This card applies positive effects to the target.
    "BUFF",
    -- This card applies negative effects to the target.
    "DEBUFF",
    -- This card heals and has a different intent icon
    "HEAL",
    -- Costs health to remove from deck
    "PARASITE",
    -- This card is not removed from your hand at the end of your turn
    "STICKY",
    -- This card doesn't upgrade but triggers an event when ready to upgrade
    "HATCH",
    -- This card contains other cards and is sticky
    "TOOLBOX",
    -- This card is expended at the end of the turn if not used.
    "BURNOUT",
    -- The damage of this card can't be affect
    "RESTRAINED",
    "GALVANIZED",
    "SLIMY",
    -- This card is Destroyed on sleep
    "SLEEP_IT_OFF",
    "DUD",
    "FLOURISH",

    -- The following flags are card type tags for applying various card rules.
    "MELEE", "RANGED", "SKILL", "STATUS", "ITEM", "SPECIAL", "SCORE",

    "ONE_OF_A_KIND",
}

local CARD_FLAG_STRINGS =
{
    [CARD_FLAGS.UNPLAYABLE] = "Unplayable",
    [CARD_FLAGS.EXPEND] = "Expend",
    [CARD_FLAGS.MULTISHOT] = "Multishot",
    [CARD_FLAGS.AMBUSH] = "Ambush",
    [CARD_FLAGS.CONSUME] = "Destroy",
    [CARD_FLAGS.PIERCING] = "Piercing",
    [CARD_FLAGS.PARASITE] = "Parasite",
    [CARD_FLAGS.FLOURISH] = "Flourish",
    -- [CARD_FLAGS.AUTOPLAY] = "Autoplay",
    [CARD_FLAGS.STICKY] = "Sticky",
    [CARD_FLAGS.TOOLBOX] = "Toolbox",
    [CARD_FLAGS.BURNOUT] = "Burnout",
    [CARD_FLAGS.RESTRAINED] = "Restrained",
    [CARD_FLAGS.SCORE] = "Score",
    [CARD_FLAGS.GALVANIZED] = "Galvanized",
    [CARD_FLAGS.SLIMY] = "Slimy",
    [CARD_FLAGS.SLEEP_IT_OFF] = "Sleep It Off",
    [CARD_FLAGS.COMBO_FINISHER] = "Finisher",
    [CARD_FLAGS.DUD] = "Dud",
}

local CARD_FLAG_DETAILS =
{
    [CARD_FLAGS.UNPLAYABLE] = "This card cannot be played.",
    [CARD_FLAGS.EXPEND] = "Remove this card for the rest of this battle.",
    [CARD_FLAGS.MULTISHOT] = "This card can be played multiple times per turn.",
    [CARD_FLAGS.AMBUSH] = "Automatically placed in your hand at the start of battle.",
    [CARD_FLAGS.CONSUME] = "Remove this card permanently.",
    -- [CARD_FLAGS.AUTOPLAY] = "This card cannot be played, and resolves automatically at the end of your turn",
    [CARD_FLAGS.PIERCING] = "Ignores damage blocked by {DEFEND}.",
    [CARD_FLAGS.ONE_OF_A_KIND] = "Your deck can only contain 1 copy of this card.",
    [CARD_FLAGS.FLOURISH] = "This is a special attack that will {EXPEND} and {BURNOUT}.",
    [CARD_FLAGS.PARASITE] = "Removing this card from your deck lowers your maximum health by 10.",
    [CARD_FLAGS.STICKY] = "This card is not discarded at the end of your turn.",
    [CARD_FLAGS.TOOLBOX] = "This card contains other cards that can be improvised. This card stays in your hand at the end of your turn, is placed in your hand at the start of every turn, and is <b>Destroyed</> once empty.",
    [CARD_FLAGS.BURNOUT] = "If this card is in your hand at the end of your turn, {EXPEND} it.",
    [CARD_FLAGS.RESTRAINED] = "The damage of this card cannot be increased.",
    [CARD_FLAGS.GALVANIZED] = "Playing this card triggers {voltaic}. <b>Galvanized</> is removed when this card goes into your discard pile.",
    [CARD_FLAGS.SLIMY] = "This card is not discarded at the end of your turn. If this card is played <b>Slimy</> is removed and the owner takes 3 damage.",
    [CARD_FLAGS.SLEEP_IT_OFF] = "This card is removed upon sleep.",
    [CARD_FLAGS.COMBO_FINISHER] = "Finishers spend some or all of your <b>Combo</b> points to enact powerful effects",
    [CARD_FLAGS.DUD] = "Created when trying to improvise a card type you do not have."
}

COMBAT_STAT = MakeEnum
{
    "HEALTH",
    "MORALE",
}

--------------------------------------------------------------------------------
-- Modifiers follow.

local CONDITIONS = 
{
    KRILLED_UP =
    {
        name = "Krilled Up",
        desc = "Lose {1} {POWER} every turn. Lose {2} health when {POWER} reaches zero.",
        desc_fn = function(self, fmt_str)
            return loc.format( fmt_str, self.decay_rate * (self.stacks or 1), self.health_loss * (self.stacks or 1) )
        end,

        decay_rate = 2,
        health_loss = 10,
        ctype = CTYPE.DEBUFF,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    local power_stacks = self.owner:GetConditionStacks("POWER")

                    self.owner:RemoveCondition( "POWER", math.min(power_stacks, self.decay_rate * (self.stacks or 1) ))
                    
                    if not self.owner:HasCondition("POWER") then
                        self.owner:DeltaHealth( -self.health_loss * (self.stacks or 1))
                        self.battle:BroadcastEvent( BATTLE_EVENT.SOUND, SoundEvents.battle_impact_sharp_leather )
                        self.owner:CheckForSurrender( true )
                        self.owner:RemoveCondition(self.id)
                    end
                    
                end
            end,
        }
    },

    SURRENDER = {
        name = "Panic",
        desc = "You can choose to win battle if all enemies <b>Panic</b>. {1} is unable to act for {2} {2*turn|turns}.",
        feature_tooltip = "You can choose to win battle if all enemies <b>Panic</b>. A fighter with <b>Panic</b> cannot act, and reduces their <b>Panic</b> by 1 at the end of their turn.",

        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.owner and self.owner:GetName(), self.stacks or 1 )
        end,

        hud_fx = {"surrender"},
        apply_sound = SoundEvents.battle_morale_broken,
        default_stacks = SURRENDER_DURATION,

		apply_sound = SoundEvents.battle_status_panic,
        
        event_handlers =
        {
            [ BATTLE_EVENT.END_ENEMY_TURN ] = function( self )
                if self.stacks == 1 then
                    if self.owner:IsLastStanding() then
                        self.delayed = true
                    else
                        self.owner:Rally()
                    end
                else
                    self.owner:RemoveCondition(self.id, 1)
                end
            end,

            [ BATTLE_EVENT.DEACTIVATED_FIGHTER ] = function ( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition(self.id, self.stacks)
                end
            end
        },
    },

    RALLIED = {
        name = "Rallied",
        desc = "Rallied. Deals 25% more damage.",
        ctype = CTYPE.BUFF,
		
		apply_sound = SoundEvents.battle_status_rallied,

        OnApply = function( self, battle )
            local HEAL_PERCENT = 0.1
            local MIN_HEAL = 5

            local min, max = self.owner:GetStatBounds( COMBAT_STAT.HEALTH )
            local current_health = self.owner:GetHealth()

            self.owner:DeltaHealth( math.max( MIN_HEAL, max * HEAL_PERCENT ) )
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner then
                    dmgt:ModifyDamage( math.round( dmgt.min_damage * 1.25 ),
                                       math.round( dmgt.max_damage * 1.25 ),
                                       self )
                end
            end,

            [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                if condition:GetID() == "SURRENDER" and fighter == self.owner then
                    self.owner:RemoveCondition(self.id)
                end
            end,
        }
    },

    RUNNING = {
        name = "Running Away",
        desc = "Next turn, this fighter will flee from battle.",

        max_stacks = 1,

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner and fighter:IsAlive() then
                    if not self.battle:CheckGameOver() and fighter:GetEnemyTeam():GetTeamStatus() ~= FIGHT_STATUS.SURRENDER then
                        self.owner:Flee()
                    end
                end
            end,
        }
    },

    DAUNTLESS =
    {
        name = "Dauntless",
        desc = "Will not {SURRENDER} while they have allies that have not panicked.",
        ctype = CTYPE.BUFF,

        OnApply = function( self, battle )
            self.had_morale = self.owner:HasMorale()
            if self.had_morale then
                self.owner:EnableMorale( self.owner:GetTeam():NumActiveFighters() <= 1 )
            end
        end,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",

        event_handlers =
        {
            [ BATTLE_EVENT.DEACTIVATED_FIGHTER ] = function ( self, fighter )
                if not CheckBits(self.battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER) then
                    if self.had_morale then
                        self.owner:EnableMorale( self:CountActive() <= 1 )
                    end
                end
            end,

            [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                if not CheckBits(self.battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER) then
                    if self.had_morale then
                        self.owner:EnableMorale( self:CountActive() <= 1 )
                    end
                end
            end,

            [ BATTLE_EVENT.FIGHTER_SURRENDER ] = function( self, fighter )
                if not CheckBits(self.battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER) then
                    if self.had_morale then
                        self.owner:EnableMorale( self:CountActive() <= 1 )
                    end
                end
            end,
        },

        CountActive = function( self )
            local count = 0
            for i, fighter in self.owner:GetTeam():Fighters() do
                if fighter:IsActive() then
                    count = count + 1
                end
            end
            return count
        end,
    },

    COMBO = {
        name = "Combo",
        feature_desc = "Gain {1} {COMBO}.",
        desc = "Combo points are accumulated until they are spent by a <b>Finisher</b>. Lose half of your <b>Combo</b> whenever you take damage.",

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Combo",

        event_handlers =
        {
            [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                if fighter == self.owner then
                    local combo_loss = math.ceil(self.stacks * 0.5)
                    combo_loss = self.battle:CalculateComboLoss( combo_loss, source, fighter )
                    self.owner:RemoveCondition( self.id, combo_loss )
                end
            end,
        },
    },

    COMBO_CARRY = {
        name = "Combo Carry",
        desc = "Gain <#HILITE>{1}</> {COMBO} next turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Combo",

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                self.owner:AddCondition("COMBO", self.stacks, self)
                self.owner:RemoveCondition(self.id)
            end
        },
    },

    BLEED =
    {
        name = "Bleed",
        feature_desc = "Apply {1} {BLEED}.",
        desc = "At the beginning of their turn, take <#HILITE>{1} damage</>, then halve <b>Bleed</b> count.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        ctype = CTYPE.DEBUFF,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Bleed",

        target_type = TARGET_TYPE.ENEMY,

        CanApply = function( self, source )
            return not self.owner:HasCondition("METALLIC")
        end,

        CalculatePreviewDamage = function( self )
            return self.stacks
        end,
        
        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    local scan = self.battle:GetPlayerFighter() and self.battle:GetPlayerFighter():GetCondition("arterial_scanner")
                    local decay = scan and scan:GetDecayRate() or 0.5

                    --jcheng: cleaved makes bleed never decay
                    decay = self.owner:GetConditionStacks("cleaved") == 0 and decay or 0

                    local attacker = self:GetRecentApplicant()
                    if attacker and not is_instance( attacker, Fighter ) then
                        attacker = attacker.owner
                    end

                    self.owner:ApplyDamage( self.stacks, attacker, nil, true, {"bleed"} )
                    self.battle:BroadcastEvent( BATTLE_EVENT.SOUND, SoundEvents.battle_impact_sharp_leather )
                    self.battle:BroadcastEvent( BATTLE_EVENT.SOUND, "event:/sfx/battle/status/debuff/bleeding_damage_layer" )

                    self.owner:CheckForSurrender()

                    local decay_stacks = math.max( 0, math.round(self.owner:GetConditionStacks("BLEED") * decay))
                    self.owner:RemoveCondition( self.id, decay_stacks )
                end
            end,
        }
    },
    
    WANTED_DEAD = 
    {
        name = "Wanted Dead",
        desc = "You've been told to kill this target.",
        max_stacks = 1,
        event_handlers =
        {
            [ BATTLE_EVENT.FIGHTER_SURRENDER ] = function( self, fighter )
                if fighter == self.owner then
                    self.highlighted = true
                    self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
                end
            end,

            [ BATTLE_EVENT.FIGHTER_RALLY ] = function( self, fighter )
                if fighter == self.owner then
                    self.highlighted = false
                    self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
                end
            end
        },
    },
    
    WANTED_ALIVE = 
    {
        name = "Wanted Alive",
        desc = "You've been told to keep this target alive.",
        max_stacks = 1,
        event_handlers =
        {
            [ BATTLE_EVENT.FIGHTER_SURRENDER ] = function( self, fighter )
                if fighter == self.owner then
                    self.highlighted = true
                    self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
                end
            end,

            [ BATTLE_EVENT.FIGHTER_RALLY ] = function( self, fighter )
                if fighter == self.owner then
                    self.highlighted = false
                    self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
                end
            end
        },
    },

    IMPAIR =
    {
        name = "Impair",
        desc = "Attack damage by this target is reduced by {1}%.\nAt the beginning of their turn, reduce <b>Impair</b> by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, math.round((1 - self.damage_mult) * 100))
        end,
        feature_desc = "Apply {1} {IMPAIR}.",
		
        ctype = CTYPE.DEBUFF,
		
        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Impair",
        apply_fx = {"impair"},
        fx_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Impair_FX",
        fx_sound_delay = 0.85,
        target_type = TARGET_TYPE.ENEMY,

        damage_mult = 0.666,

        event_priorities =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner then
                    dmgt:ModifyDamage( math.round( dmgt.min_damage * self.damage_mult ),
                                       math.round( dmgt.max_damage * self.damage_mult ),
                                       self )
                end
            end,

            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition( self.id, 1 )
                end
            end,
        }
    },

    STUN =
    {
        name = "Stun",
        desc = "Can't act this turn. Remove Stun at the end of the turn. This condition does not stack.",
        feature_desc = "Apply {STUN}.",
        hud_fx = {"stunned"},
        idle_state_override = IDLE_STATE.STUNNED,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Debuff_Physical",

        max_stacks = 1,
        ctype = CTYPE.DEBUFF,

        target_type = TARGET_TYPE.ENEMY,

        CanApply = function( self, source )
            return not self.owner:IsPlayer()
        end,

        CanPlayCard = function ( self, battle, target )
            return false, CARD_PLAY_REASONS.STUNNED
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition( self.id, 1 )
                    self.owner:CheckForSurrender()
                end
            end,

            [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                if self.owner == fighter and condition.id == "STUN" then
                    self.owner:PrepareTurn()
                end
            end
        }
    },

    STAGGER =
    {
        name = "Stagger",
        desc = "When <b>Stagger</b> reaches {1}, clear it and apply {STUN 1}.\nDecrease by 1 at the end of your turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stun_stacks or 5 )
        end,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Debuff_Physical",

        target_type = TARGET_TYPE.ENEMY,
        stun_stacks = 5,

        OnSetStacks = function( self, delta, old_stacks, new_stacks )
            if new_stacks >= self.stun_stacks then
                self.owner:AddCondition( "STUN", 1, self )
                self:SetStacks( new_stacks - self.stun_stacks )
            end
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition( self.id, 1 )
                end
            end,
        }
    },

    DEFECT =
    {
        name = "Defect",
        feature = "Apply {1} {DEFECT}.",
        desc = "The next attack on this fighter does double damage and removes 1 <b>Defect</b>.",

        ctype = CTYPE.DEBUFF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Wound",
        fx_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Wound_FX",
        fx_sound_delay = .3,
        apply_fx = { "wound"},

        target_type = TARGET_TYPE.ENEMY,

        event_priorities = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },
        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt, hit )
                if target == self.owner then
                    local found = false
                    local attack_count = 0
                    for i, fighter in self.battle:ActiveFighters() do
                        if fighter.prepared_attacks and not found then
                            for i, attack in ipairs( fighter.prepared_attacks ) do
                                if attack:IsTarget( target ) then
                                    if attack.card == card then
                                        -- Ok, found this specific attack.
                                        found = true
                                        break
                                    elseif attack.card:IsDamageCard() then
                                        attack_count = attack_count + 1
                                    end
                                end
                            end
                        end
                    end

                    if not found or attack_count < self.stacks then
                        dmgt:AddDamage( dmgt.min_damage, dmgt.max_damage, self )
                    end
                end
            end,

            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if attack:IsTarget( self.owner ) and attack.card:IsAttackCard() then
                    if attack:IsDamageModifier( self ) and not hit.evaded then
                        self.owner:RemoveCondition( self.id, 1 )
                    end
                end
            end,
        }
    },

    WOUND =
    {
        name = "Wound",
        feature_desc = "Apply {1} {WOUND}.",
        desc = "Attacks inflict <#HILITE>{1}</> bonus damage on this target.\nAt the beginning of their turn, reduce <b>Wound</b> by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        ctype = CTYPE.DEBUFF,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Wound",
        fx_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Wound_FX",
        fx_sound_delay = .3,
        apply_fx = { "wound"},

        target_type = TARGET_TYPE.ENEMY,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if target == self.owner then
                    dmgt:ModifyDamage( dmgt.min_damage + self.stacks, dmgt.max_damage + self.stacks, self )
                end
            end,

            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition( self.id, 1 )
                end
            end,
        },

        CanApply = function( self, source )
            return not self.owner:HasCondition("METALLIC")
        end,
    },

    MENDING = 
    {
        name = "Mending",
        feature_desc = "Apply {1} {MENDING}.",
        feature_tooltip = "At the end of turn, {HEAL} an amount equal to <b>Mending</b> and reduce <b>Mending</b> by 1.",

        desc = "At the end of their turn, <#HILITE>{HEAL} {1}</> then reduce <b>Mending</b> by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Heal",

        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

        CalculatePreviewDamage = function( self )
            return -self.stacks or 0
        end,
        
        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:HealHealth( self.stacks, self )
                    self.owner:RemoveCondition( self.id, 1 )
                end
            end,
        }
    },

    DEFEND =
    {
        name = "Defense",
        feature_desc = "Gain {1} {DEFEND}.",
        feature_desc_fn = function( self, fmt_str, num, engine, card )
            return loc.format( fmt_str, card:CalculateDefendText( num ))
        end,
        desc = "Prevents attack damage for this turn.",
        min_stacks = 0,
        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend_Shield",
        hidden = true,

        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

        priority = 2, -- After evasion, before temp health

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    local retain_defend = self.owner:GetConditionStacks( "RETAIN_DEFEND" )
                    self.owner:RemoveCondition( self.id, math.max( 0, self.stacks - retain_defend ), self)
                end
            end,
        },

        OnSetStacks = function( self, delta, old_stacks, new_stacks )
            local def = Content.FindAchievement( "impervious" )
            if def and old_stacks < def.count and new_stacks >= def.count then
                TheGame:UnlockAchievement( def.id )
            end
        end,

        OnPreDamage = function( self, damage, attacker, battle, source, piercing )
            if piercing then
                return damage
            end

            local defended = math.min( damage, self.stacks )
            if defended > 0 then
                self.owner:AppendLog( BATTLE_LOG.GENERAL, "{1}'s DEFEND reduces damage by {2} (from {3})", self.owner, defended, damage )
                self.owner:RemoveCondition( self.id, defended, source )
            end

            -- If fully mitigated, mark as 'defended'.
            if defended >= damage and is_instance( source, Battle.Hit ) and not source:CheckHitResult( "evaded" ) then
                if not self.owner:HasCondition("STUN") and not source.attack:IsApplyingFeature( self.owner, "STUN" ) then
                   source:SetHitResult( "defended" )
               end
            end

            return damage - defended
        end,
    },

    TARGETED = 
    {
        name = "Targeted",
        feature_desc = "Gain {1} {TARGETED}.",
        desc = "All non-player characters will prioritize this target.",

        --apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",

        target_type = TARGET_TYPE.SELF,
        ctype = CTYPE.DEBUFF,


        CanBeTargetted = function( self, card, fighter )
            -- Only applies to attacks by NPCs...
            if card:IsAttackCard() and not card.owner:IsPlayer() and not card:IsFlagged( CARD_FLAGS.SPECIAL ) then
                -- ... who are on the other team, targetting the enemy team (eg our team)
                if fighter:GetTeam() ~= card:GetOwner():GetTeam() and fighter:GetTeam() == self.owner:GetTeam() then
                    -- ... if the target is not us.
                    if fighter ~= self.owner and self.owner:IsAlive() and not fighter:WillProtectOthers() then
                        return false, "Another enemy has Targeted"
                    end
                end
            end
            return true
        end,


        OnApply = function( self, battle )
            for k,v in pairs( self.owner.team.fighters ) do
                if v ~= self.owner then
                    v:RemoveCondition("TARGETED")
                end
            end
        end,
    },

    EVASION =
    {
        name = "Evasion",
        feature_desc = "Gain {1} {EVASION}.",
        desc = "Dodge next hit. Remove 1 <b>Evasion</b> when you dodge a hit and at the start of your turn.",
        desc_fn  = function( self, fmt_str )
            return loc.format( fmt_str, self.chance )
        end,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",

        target_type = TARGET_TYPE.SELF,

        chance = .5,

        priority = 1, -- before defend

        PreviewFn = function( self, damage_preview, targets )
            if targets[self.owner] ~= nil then
               local count = math.min(self.stacks, #targets[self.owner])
               for i=1, count do
                    if targets[self.owner][i].card ~= nil and is_instance(Battle.Card, targets[self.owner][i].card) then
                       targets[self.owner][i].min_damage = 0
                       targets[self.owner][i].max_damage = 0
                    else
                        count = count + 1
                    end
               end
            end
        end,


        OnPreDamage = function( self, damage, attacker, battle, source )
            if source == nil then
                return damage
            end

            if is_instance( source, Battle.Hit ) and source.card and source.card:IsAttackCard() and source.attack.owner ~= self.owner then
                self.owner:RemoveCondition("EVASION", 1)

                if not self.owner:HasCondition("STUN") then
                    source:SetHitResult( "evaded" )
                    battle:BroadcastEvent( BATTLE_EVENT.ON_DODGE, self.owner, source, attacker )
                    return 0
                end
            end

            return damage
        end,

        event_handlers = 
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition("EVASION", 1)
                end
            end
        },
    },


    RETAIN_DEFEND =
    {
        name = "Retain Defend",
        desc = "Retain 5 {DEFEND} at the end of your turn.",
        target_type = TARGET_TYPE.SELF,
    },

    EXPOSED =
    {
        name = "Exposed",
        desc = "When receiving {DEFEND}, reduce the amount by 50%.\nAt the end of your turn, reduce <b>Exposed</b> by 1.",
        feature_desc = "Apply {1} {EXPOSED}.",

        ctype = CTYPE.DEBUFF,

         apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                if condition_id == "DEFEND" and fighter == self.owner then
                    if acc.value > 0 then
                        acc:AddValue( -math.floor( acc.value / 2 ), self )
                    end
                end
            end,

            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition( self.id, 1 )
                end
            end,
        }
    },
    
    ARMOURED =
    {
        name = "Armored",
        desc = "Gain <#HILITE>{1}</> {DEFEND} at the end of every turn.\n",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.stacks or 1)
        end,

        hide_gained = true,
        
        OnApply = function( self, battle )
            self.owner:AddCondition("DEFEND", self.stacks or 1, self)
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if self.owner == fighter and self.owner:IsActive() then
                    self.owner:AddCondition( "DEFEND", self.stacks or 1, self )
                end
            end,
        }
    },        
    --same as mech, BUT mech does some other roboty stuff, too. So we dupe it.
    METALLIC =
    {
        name = "Metallic",
        desc = "Immune to <b>Bleed</b> and <b>Wound</b>.\n{PIERCING} attacks do double damage.",
        hide_gained = true,

        event_priorities =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt, hit )
                if (card:IsFlagged( CARD_FLAGS.PIERCING ) or (hit and hit.piercing)) and target == self.owner then
                    dmgt:AddDamage( dmgt.min_damage, dmgt.max_damage, self )
                end
            end,
        }
    },    

    PROTECT =
    {
        name = "Protect",
        feature_desc = "Gain {1} {PROTECT}.",
        desc = "Opponents must target enemies that have <b>Protect</>. Remove one stack when damage is received.",

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",
        target_type = TARGET_TYPE.SELF,

        loc_strings =
        {
            PROTECTED = "Protected by ally",
        },

        CanBeTargetted = function( self, card, fighter )
            if card:IsAttackCard() and not self.owner:HasCondition("STUN") then
                if fighter:GetTeam() ~= card:GetOwner():GetTeam() and card.owner:GetTeam() ~= self.owner:GetTeam() then
                    if self.owner:WillProtectOthers() and fighter ~= self.owner and not fighter:WillProtectOthers() then
                        return false, self:GetLocalizedString( "PROTECTED" )
                    end
                end
            end
            return true
        end,

        OnPostDamage = function( self, damage, attacker, battle, source )
            if damage > 0 then
                self.owner:RemoveCondition( self.id, 1 )
            end
        end,

        event_handlers =
        {
            [BATTLE_EVENT.STATUS_CHANGED] = function( self )
                if not self.owner:IsActive() then
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },

    OPPORTUNISTIC =
    {
        name = "Opportunistic",
        desc = "Gain <#HILITE>{POWER {1}}</> after an attack damages a target.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack",

        event_handlers =
        {
            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if not hit:CheckHitResult( "defended" ) and not hit:CheckHitResult( "evaded" ) and hit.card:IsDamageCard() and hit.attacker == self.owner then
                    self.owner:AddCondition( "POWER", self.stacks, self )
                end
            end,
        }
    },

    RIPOSTE =
    {
        name = "Counter",
        feature_desc = "Gain {1} {RIPOSTE}.",
        desc = "Deal <#HILITE>{1}</> damage back to the enemy when hit. Remove all <b>Counter</b> at the beginning of your turn.",
		desc_fn = function( self, fmt_str, battle )
            return loc.format( fmt_str, self.stacks or 1 )
        end,
		
        hud_fx = {"counter"},
		apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend_Counter",

        target_type = TARGET_TYPE.SELF,

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition( self.id )
                end
            end,

            [ BATTLE_EVENT.ON_HIT] = function( self, battle, attack, hit )
                local resolve_card = attack.card
                if self.owner:IsActive() and not self.owner:HasCondition( "STUN" ) and not self.owner:HasCondition("SURRENDER") then
                    if resolve_card.owner and resolve_card.owner ~= self.owner and not resolve_card.owner.call_in and resolve_card:IsAttackCard() and hit.damage and (hit.damage > 0 or (attack.card.max_damage and attack.card.max_damage > 0)) and not attack.is_counter then
                        --
                        if hit.target == self.owner then
                            self.counter_attack_counter = (self.counter_attack_counter or 0) + 1
                            attack.riposted = true
                        end
                    end
                end
            end,

            [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, resolve_card )
                if self.counter_attack_counter and self.counter_attack_counter > 0 then
                    self.owner:CheckForSurrender()
                    if self.owner:IsActive() and not self.owner:HasCondition( "STUN" ) and not self.owner:HasCondition("SURRENDER") and resolve_card.owner:IsActive() then
                        --
                        local card = Battle.Card( "counter_attack", self.owner )
                        card.engine = battle -- HAX
                        card.min_damage = self.stacks
                        card.max_damage = card.min_damage 

                        local fight_data = self.owner.agent.fight_data
                        if fight_data.riposte_tags then
                            card.hit_tags = self.owner.riposte_tags
                        end
                        if fight_data.ranged_riposte then
                            card.flags = CARD_FLAGS.SPECIAL | CARD_FLAGS.RANGED
                        end

                        card:AssignTarget( resolve_card.owner, { { resolve_card.owner } } )

                        for i=1, self.counter_attack_counter do
                            if resolve_card.owner:IsAlive() then
                                local counter_attack = Battle.Attack( card )
                                counter_attack:GenerateHits()
                                counter_attack.is_counter = true
                                battle:BroadcastEvent( BATTLE_EVENT.ON_RIPOSTE, counter_attack, i == 1, resolve_card, i == self.counter_attack_counter )
                                counter_attack:RunAttack()
                            end
                        end

                        resolve_card.owner:CheckForSurrender()
                    end
                end
                self.counter_attack_counter = 0
            end,

            [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                if fighter == self.owner and condition:GetID() == "STUN" then
                    self.owner:RemoveCondition(self.id)
                end
            end
        }
    },

    HEMOPHAGE =
    {
        name = "Hemophage",
		desc = "Whenever you apply {BLEED}, <#HILITE>{HEAL} {1}</>.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Heal",

        event_handlers =
        {
            [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                if condition:GetID() == "BLEED" then
                    if source and source.owner == self.owner then
                        self.owner:HealHealth( self.stacks, self )
                    end
                end
            end,
        },
    },

    POWER =
    {
        name = "Power",
        feature_desc = "Gain {1} {POWER}.",
		desc = "Attack damage is increased by <#HILITE>{1}</>.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,
		
		apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Power",

        fx_sound = "event:/sfx/battle/status/system/Status_Buff_Attack_Power_FX",
        fx_sound_delay = .4,
        apply_fx = {"power"},

        target_type = TARGET_TYPE.SELF,
        max_stacks = 99,
        min_stacks = -99,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner and card:IsAttackCard() then
                    dmgt:ModifyDamage( dmgt.min_damage + self.stacks, dmgt.max_damage + self.stacks, self )
                end
            end,
        }
    },

    POWER_LOSS =
    {
        name = "Temporary Power",
        desc = "Remove <#HILITE>{1} {POWER}</> at the end of {2}'s turn.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks, self:GetOwnerName() )
        end,
        
        target_type = TARGET_TYPE.SELF,
        ctype = CTYPE.DEBUFF,
        max_stacks = 99,

        OnApply = function( self, battle )
            self.apply_turn = battle:GetTurns()
        end,

        event_priorities =
        {
            [ BATTLE_EVENT.END_TURN  ] = 10,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    if self.owner.player_controlled then
                        fighter:RemoveCondition( "POWER", self.stacks )
                        fighter:RemoveCondition( self.id )
                    else
                        if self.apply_turn < self.battle:GetTurns() then
                            fighter:RemoveCondition( "POWER", self.stacks )
                            fighter:RemoveCondition( self.id )
                        end
                    end
                end
            end,
        }
    },


    EXISTENTIAL_CRISIS = 
    {
        name = "Existential Crisis",
        desc = "At the end of each turn, increase this fighter's {SURRENDER} meter by <#HILITE>{1}</>.",

        desc_fn = function(self, fmt_str, battle)
            return loc.format(fmt_str, self.stacks or 1)
        end,
        
        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    fighter:ApplyMoraleDamage(self.stacks, self)
                    fighter:CheckForSurrender()
                end
            end,
        },

    },
    FATIGUED =
    {
        name = "Fatigued",
        desc = "Cards no longer gain XP for this battle",
        ctype = CTYPE.DEBUFF,

        apply_sound = "event:/sfx/battle/status/system/Status_Debuff_Fatigue",
    },

    SHATTER =
    {
        name = "Shatter",
        desc = "Deal 2x damage to enemies with {DEFEND}.",

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack",

        event_priorities =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner and target and target:GetConditionStacks("DEFEND") > 0 then
                    dmgt:ModifyDamage( dmgt.min_damage * 2, dmgt.max_damage * 2, self )
                end
            end,
        },
    },

    COMBAT_DRAW = 
    {
        name = "Card Draw",
		desc = "Draw <#HILITE>{1}</> extra {1*card|cards} per turn.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",

        event_handlers = 
        {
            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                battle:DrawCards( self.stacks )
            end
        },
    },

    EXERT =
    {
        name = "Exert",
        desc =  "Next turn, gain <#HILITE>{1}</> fewer {1*action|actions}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,

        loc_strings =
        {
            EXERT_ONE = "{EXERT}",
        },

        feature_desc = "{1} {EXERT}",
        feature_desc_fn = function( self, fmt_str, num, engine, card )
            if num == 1 then
                return self:GetLocalizedString( "EXERT_ONE" )
            else
                return loc.format( fmt_str, num )
            end
        end,

        target_type = TARGET_TYPE.SELF,

        ctype = CTYPE.DEBUFF,

        apply_sound = "event:/sfx/battle/status/system/Status_Debuff_Other",

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    if self.owner.player_controlled then
                        self.battle:ModifyExertedCount( self.stacks )
                    end
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },

    NEXT_TURN_DEFENSE =
    {
        name = "Next-Turn Defense",
        desc =  "Next turn, gain <#HILITE>{1}</> {DEFEND}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:AddCondition("DEFEND", self.stacks)
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },

    NEXT_TURN_ACTION =
    {
        name = "Next-Turn Action",
        desc =  "Next turn, gain <#HILITE>{1}</> additional {1*action|actions}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    if self.owner.player_controlled then
                        self.battle:ModifyActionCount( self.stacks )
                    end
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },
    
    BONUS_ACTION =
    {
        name = "Bonus Action",
        desc =  "This turn, gain <#HILITE>{1}</> additional {1*action|actions}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,
        target_type = TARGET_TYPE.SELF,
        
        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",
        OnApply = function(self, battle)
            battle:ModifyActionCount( self.stacks )
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },

    NEXT_TURN_DRAW =
    {
        name = "Next-Turn Draw",
        desc =  "Next turn, draw <#HILITE>{1}</> additional {1*card|cards}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,
        feature_desc = "Draw {1} extra {1*card|cards} next turn.",

        target_type = TARGET_TYPE.SELF,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Other",

        event_handlers =
        {
            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    if self.owner.player_controlled then
                        self.battle:DrawCards( self.stacks )
                    end
                    self.owner:RemoveCondition( self.id )
                end
            end,
        }
    },

    VENDETTA = 
    {
        name = "Vendetta",
        desc =  "{1} hates {2}, and will always attack them.",
        loc_strings =
        {
            ALT_DESC = "Will always attack the player in battle.",
        },

        desc_fn = function( self, fmt_str, battle )
            if battle then
                return loc.format(fmt_str, self.owner:GetName(), self.battle:GetPlayerFighter().agent:GetName() )
            else
                return self.feature_def:GetLocalizedString( "ALT_DESC" )
            end
        end,

        max_stacks = 1,
        target_type = TARGET_TYPE.SELF,

        ShouldAutoAttach = function( self, battle, fighter )
            if fighter:GetTeamID() == TEAM.RED and not fighter.call_in and fighter.agent:GetRelationship() == RELATIONSHIP.HATED then
                return 1
            end
        end,

        CanBeTargetted = function( self, card, fighter )
            if card:IsAttackCard() and card.owner == self.owner and not fighter:IsPlayer() and not card:IsFlagged( CARD_FLAGS.SPECIAL ) then
                if not fighter:HasCondition( "TARGETED" ) and not fighter:HasCondition( "PROTECT" ) and card.target_mod == TARGET_MOD.SINGLE then
                    return false, "Only targets player"
                end
            end
            return true
        end,
    },

    KRADESHI_BOON = 
    {
        name = "Kra'deshi Regeneration",
        desc = "At the end of {1}'s turn, apply <#HILITE>{HEAL {2}}</>.\nThis is a species boon.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.owner.agent:GetName(), self.stacks )
        end,

        ShouldAutoAttach = function( self, battle, fighter )
            if not fighter.agent:IsPlayer() and not fighter.call_in and fighter.agent:GetSpecies() == SPECIES.KRADESHI then
                local stacks = 2
                if fighter:GetScale() > 1 then
                    stacks = math.floor( stacks * GetAdvancementModifier( ADVANCEMENT_OPTION.UPGRADED_SPECIES_BOONS ))
                end
                return stacks
            end
        end,

        apply_sound = "event:/sfx/battle/status/system/Status_Heal",

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    fighter:HealHealth(self.stacks, self)
                end
            end,
        }
    },

    JARACKLE_BOON = 
    {
        name = "Jarackle's Instinct",
        desc = "Gain <#HILITE>{POWER {1}}</> at the start of battle.\nThis is a species boon.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,
        
        power_amt = {1, 1, 1, 2, 2},

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Attack",

        ShouldAutoAttach = function( self, battle, fighter )
            if not fighter.agent:IsPlayer() and not fighter.call_in and fighter.agent:GetSpecies() == SPECIES.JARACKLE then
                local stacks = 1
                if fighter:GetScale() > 1 then
                    stacks = math.floor( stacks * GetAdvancementModifier( ADVANCEMENT_OPTION.UPGRADED_SPECIES_BOONS ))
                end
                return stacks
            end
        end,

        OnApply = function(self, battle)
            self.owner:AddCondition("POWER", self.stacks, self)
        end,
    },

    SHROKE_BOON = 
    {
        name = "Shroke Skin",
        desc = "The first time {1} takes damage each turn, they gain <#HILITE>{DEFEND {2}}</>.\nThis is a species boon.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.owner.agent:GetName(), self.stacks )
        end,

        ShouldAutoAttach = function( self, battle, fighter )
            if not fighter.agent:IsPlayer() and not fighter.call_in and fighter.agent:GetSpecies() == SPECIES.SHROKE then
                local stacks = 2
                if fighter:GetScale() > 1 then
                    stacks = math.floor( stacks * GetAdvancementModifier( ADVANCEMENT_OPTION.UPGRADED_SPECIES_BOONS ))
                end
                return stacks
            end
        end,

        apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",

        event_handlers =
        {
            [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                if fighter == self.owner and stat == COMBAT_STAT.HEALTH and not self.has_gained_defense then
                    if delta < 0 then
                        self.has_gained_defense = true
                        self.owner:AddCondition("DEFEND", self.stacks, self)
                    end
                end
            end,

            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.has_gained_defense = false
                end
            end
        }
    },

    SPARK_VISOR_MARK =
    {
        name = "Spark Visor",
        desc = "Attacking this target grants 8 {DEFEND} and reduces this by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetOwnerName() )
        end,

        ctype = CTYPE.DEBUFF,

        event_handlers = 
        {
            [ BATTLE_EVENT.PRE_RESOLVE ] = function( self, battle, attack )
                for i, hit in attack:Hits() do
                    if hit.target == self.owner and self.stacks > 0 then
                        attack.attacker:AddCondition("DEFEND", 8, self)
                        self.owner:RemoveCondition("SPARK_VISOR_MARK", 1)
                    end
                end                    
            end,
        },
    },

    METTLESOME = 
    {
        name = "Mettlesome",
        desc =  "Killing {1} grants {2} extra {METTLE} at the end of the fight.",

        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.owner:GetName(), self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,

        ShouldAutoAttach = function( self, battle, fighter )
            if (TheGame:GetGameState() and TheGame:GetGameState():CanDoMettle()) and
                        not battle.mettle_applied and 
                        fighter:GetTeamID() == TEAM.RED and 
                        battle:GetPlayerFighter() then
                    local player = battle:GetPlayerFighter():GetAgent()
                    local chance = METTLESOME_CHANCE[math.min( #METTLESOME_CHANCE, fighter and fighter.scale or 1)]
                    if math.random() < chance then
                        battle.mettle_applied = true
                        local amt = METTLESOME_AMT[math.min( #METTLESOME_AMT, fighter and fighter.scale or 1)]
                        return amt
                    end
                end
        end,


        event_handlers = 
        {
            [ BATTLE_EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if fighter == self.owner and status == FIGHT_STATUS.DEAD then
                    local player = self.battle:GetPlayerFighter():GetAgent()
                    self.battle:GetPlayerFighter():AddCondition("METTLE_BANK", self.stacks)
                end
            end,
        },
    },

    METTLE_BANK = 
    {
        name = "Bonus Mettle",
        desc =  "You will get {1} bonus {METTLE} if you win the fight.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,
        target_type = TARGET_TYPE.SELF,
    },

    BATTERED = 
    {
        name = "Battered",
        desc = "{1} draws 1 fewer cards at the start of {1}'s turn. At the end of {1}'s turn, remove 1 <b>Battered</>.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        ctype = CTYPE.DEBUFF,

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition(self.id, 1, self)
                end
            end,

            [ BATTLE_EVENT.CALC_TURN_DRAW ] = function( self, acc )
                acc:AddValue( -1, self )
            end,
        },
    },
}

for id, def in pairs( CONDITIONS ) do
    Content.AddBattleCondition( id, def )
end

-------------------------------------------------------------
-- Generate features, a superset of CARD_FLAGS and CONDITIONS.

local CARD_FEATURES = {}

for i, name in ipairs( CARD_FLAG_BITSTRINGS ) do
    local bit = 1 << (i-1)
    local details = CARD_FLAG_DETAILS[ bit ]
    if details then
		local data =
		{
            name = CARD_FLAG_STRINGS[ bit ],
            desc = details,
        }

		local def = BattleFeatureDef(name, data)

		Content.AddBattleCardFeature(def.id, def)
        CARD_FEATURES[ bit ] = def
        CARD_FEATURES[ name ] = def
    end
end

local FEATURES = 
{
    HATCH_X =
    {
        name = "<p img='icons/ic_hatch.tex' scale=1.0> Hatch",
        desc = "When this card's effect is triggered the specified number of times, it is destroyed and a special event occurs. The hatch counter is not changed while fatigued.",
    },

	IMPROVISE =
    {
        name = "Improvise",
        desc = "Choose 1 of 3 cards to put into your hand.",
    },

    IMPROVISE_PLUS =
    {
        name = "Improvise+",
        desc = "Choose 1 of 5 cards to put into your hand.",
    },

    WARPING = 
    {
        name = "Warping",
        desc = "At the start of your turn, this card changes its effect.",
    },

    PREPARED = 
    {
        name = "Prepared",
        desc = "This card has an additional effect when it is the left-most card in your hand.",
    },

    PREPARE = 
    {
        name = "Prepare",
        desc = "Move a card to the left-most of your hand.",
    },

	HEAL =
	{
		name = "Heal",
		desc = "Restores health up to max.",
	},

    ABILITY =
    {
        name = "Ability",
        desc = "Gain a condition for the rest of this battle",
    },

    THRESHOLD =
    {
        name = "Threshold",
        desc = "Activates an added effect if damage is at or above the Threshold number.",
    },

    OUTNUMBERED =
    {
        name = "Outnumbered",
        desc = "This card has an added effect when you are outnumbered.",
    },

    REPLENISH =
    {
        name = "Replenish",
        desc = "When this card is drawn, draw another card.",
    },


    BLEED =
    {
        name = "Bleed",
        feature_desc = "Apply {1} {BLEED}.",
        desc = "At the beginning of their turn, take damage equal to the count of <b>Bleed</b>, then halve <b>Bleed</b> count.",
        apply = function ( self, battle, attack, stacks, target )
            if not attack:CheckHitResult( target, "evaded" ) then
                if target.player_controlled then
                    local cards = {}
                    for i=1,stacks do
                        local incepted_card = Battle.Card( "status_bleed", target)
                        table.insert(cards, incepted_card )                                  
                    end
                    battle:DealCards( cards, battle:GetDiscardDeck() )
                else
                    target:AddCondition("BLEED", stacks, attack.card)
                end
            end
        end
    },

    TEMP_POWER =
    {
        name = "Temporary Power",
        desc = "{POWER} that is removed at the end of your turn."
    },

    KEEPSAKE =
    {
        name = "Keepsake",
        desc = "Gain additional points if this is still in your deck at the end of the daily challenge."
    },

    METTLE =
    {
        name = "Mettle",
        desc = "Accumulate Mettle to purchase bonuses that permanently upgrade your character."
    }
}

for id, data in pairs( FEATURES ) do
	local def = BattleFeatureDef(id, data)
	Content.AddBattleCardFeature(id, def)
	CARD_FEATURES[id] = def
end

return
{
    IS_BATTLE = true,

    STATUS = STATUS,
    BATTLE_FLAGS = BATTLE_FLAGS,
    BATTLE_EVENT = BATTLE_EVENT,
    EVENT = BATTLE_EVENT,

    CARD_FLAGS = CARD_FLAGS,
    CARD_FLAG_BITSTRINGS = CARD_FLAG_BITSTRINGS,
    CARD_FLAG_DETAILS = CARD_FLAG_DETAILS,
    CARD_FLAG_STRINGS = CARD_FLAG_STRINGS,

    GetFeature = Content.GetBattleCardFeature,
    CONDITIONS = Content.GetAllBattleConditions(),
}
