require "eventsystem"
require "cards/card_engine"
require "battle/battle_log"
require "battle/team"
require "battle/fighter"
require "battle/damage_accumulator"
require "battle/attack"

local battle_defs = require "battle/battle_defs"

local STATUS = battle_defs.STATUS
local BATTLE_FLAGS = battle_defs.BATTLE_FLAGS
local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

local START_FATIGUE_TURN = 6
local INFINITE_THRESHOLD = 15 -- If more cards are played, it's probably an infinite
--------------------------------------------------------------------------------
-- A single card.

local Card = class( "Battle.Card", CardEngine.Card )
Card.can_reload = true
Card.CARD_FLAGS = CARD_FLAGS

function Card:init( id, owner, userdata )
    local data = Content.GetBattleCard( id )
    if data == nil then
        LOGWARN( "Unknown card: %s. Converting to drunk", id )
        data = Content.GetBattleCard( "drunk" )
        id = "drunk"
    end

    for k,v in pairs(data) do
        self[k] = v
    end
    self.features = deepcopy(data.features)

    self.def = data
    self.owner = owner
    self.userdata = userdata or {}

    self:RollUpgrades()

    if self:GetMaxXP() ~= nil and self.userdata.xp == nil then
        self.userdata.xp = 0
    end

    -- Charges are stored in user-data as they are persistent.
    if (self.userdata.max_charges or self.max_charges) and self.userdata.charges == nil then
        self.userdata.charges = self.userdata.max_charges or self.max_charges
    end

    if data.OnInit then
        data.OnInit( self )
    end
end

function Card:CanTrash()
    if CheckAnyBits( self.flags, CARD_FLAGS.ITEM ) then
        if is_instance( self.owner, Agent ) then
            return self.owner.battler and self.owner.battler:HasCard( self )
        end
    end
    return false
end

function Card:IsAutoplay()
    return self:IsFlagged( CARD_FLAGS.AUTOPLAY )
end

function Card:CanGainXP()
    if is_instance( self.owner, Fighter ) then
        return not self.owner:HasCondition("FATIGUED")
    end

    return true
end

function Card:Load()
    self.def = Content.GetBattleCard( self.id )
    if self.def then
        for k,v in pairs(self.def) do 
            self[k] = v
        end
    end
end


function Card:OnAdded( battle )
    self.engine = battle
    if self.event_handlers then
        for event_name, fn in pairs( self.event_handlers ) do
            local priority = self.event_priorities and self.event_priorities[ event_name ]
            battle:ListenForEvent( event_name, self, self.OnEvent, priority )
        end
    end
    self:UpdateDamageAndBuffScale()
    if self.on_added then
        self:on_added( battle )
    end

    -- If this card doesn't exist in the permanent deck, flag it as temporary.
    if is_instance( self.owner, Fighter ) and self.owner.agent.battler and not self.owner.agent.battler:FindCardByID( self.id, self.userdata ) then
        self:MakeTemporary()
    end
end

function Card:UpdateDamageAndBuffScale()
    if self.generate_damage_fn then
        self:generate_damage_fn()
    end

    if self.generate_buff_fn then
        self:generate_buff_fn()
    end
end


function Card:OnRemoved( battle )
    if self.event_handlers then
        battle:RemoveListener( self )
    end
end

function Card:IsUpgraded()
    return CheckBits( self.flags, CARD_FLAGS.UPGRADED ) 
end

function Card:IsHatched()
    return CheckBits( self.flags, CARD_FLAGS.HATCH )
end

function Card:IsFriendly()
    return self.target_type == TARGET_TYPE.FRIENDLY or
        self.target_type == TARGET_TYPE.FRIENDLY_OR_SELF or
        self.target_type == TARGET_TYPE.SELF
end

function Card:IsFlagged( flags )
    local my_flags = self.engine and self.engine:CalculateCardFlags( self ) or self.flags
    return CheckBits( my_flags, flags )
end

function Card:IsFlaggedAny( flags )
    local my_flags = self.engine and self.engine:CalculateCardFlags( self ) or self.flags
    return CheckAnyBits( my_flags, flags )
end

function Card:NotifyChanged( hint )
    self.deck.engine:BroadcastEvent( BATTLE_EVENT.CARD_CHANGED, self, self:Clone(), hint )
end

function Card:GetOwner( owner )
    return self.owner
end

function Card:Consume( minigame )
    local agent
    if is_instance( self.owner, Agent ) then
        agent = self.owner
    elseif is_instance( self.owner, Fighter ) then
        agent = self.owner.agent
    else
       print( self, "NO OWNER TO CONSUME:", self.owner )
    end

    if agent and agent.battler then
        agent.battler:RemoveCardByID( self.id, self.userdata )
    end
end   

function Card:AssignTarget( target, secondary_targets )
    assert( target == nil or is_instance( target, Fighter ))
    self.target = target
    self.secondary_targets = secondary_targets or table.empty
end

function Card:GetTarget()
    return self.target
end

function Card:SetRandSeed( seed )
    self.rand_seed = seed
end

function Card:GetRandSeed()
    return self.rand_seed
end

function Card:CreateRNG()
    if self.rand_seed then
        return engine.Random( self.engine.battle_seed, self.rand_seed )
    end
end

function Card:NotifyTriggered()
    if self.engine then
        self.engine:BroadcastEvent( BATTLE_EVENT.CARD_TRIGGERED, self, self.deck, self.deck and self.deck:FindCard( self ))
    end
end
function Card:NotifyTriggeredPre()
    if self.engine then
        self.engine:BroadcastEvent( BATTLE_EVENT.CARD_TRIGGERED_PRE, self )
    end
end
function Card:NotifyTriggeredPost()
    if self.engine then
        self.engine:BroadcastEvent( BATTLE_EVENT.CARD_TRIGGERED_POST, self, self.deck, self.deck and self.deck:FindCard( self ))
    end
end

function Card:HasFeature( feature )
    return self.features and self.features[ feature ] ~= nil
end

function Card:AddFeature( feature, value )
    assert( value == nil or value > 0 )
    if self.features == nil then
        self.features = {}
    end
    self.features[ feature ] = (self.features[ feature ] or 0) + value
end

function Card:RemoveFeature( feature, value )
    if self.features[ feature ] then
        if value == nil or value >= self.features[ feature ] then
            self.features[ feature ] = nil
        else
            self.features[ feature ] = self.features[ feature ] - value
        end
    end
end

function Card:CalculateDefendText( defend, target )
    if self.engine then
        local modified_defend = self.engine:CalculateModifiedStacks( "DEFEND", defend, target or self.target, self )
        --modified_defend = math.max(modified_defend, 0)
        if modified_defend < defend then
            return string.format( "<#PENALTY_CARD_TEXT>%d</>", modified_defend )
        elseif modified_defend > defend then
            return string.format( "<#BONUS_CARD_TEXT>%d</>", modified_defend )       
        end
    end

    return tostring(defend)
end

function Card:CalculateThresholdText( card )
    if self.engine and card.threshold then
        local modified_threshold = self.engine:CalculateThreshold( card )
        if modified_threshold > card.threshold then
            return string.format( "<#PENALTY_CARD_TEXT>%d</>", modified_threshold )
        elseif modified_threshold < card.threshold then
            return string.format( "<#BONUS_CARD_TEXT>%d</>", modified_threshold )       
        end
    end

    return tostring(card.threshold)
end

function Card:RenderDebugUI( ui, panel )
    ui:Text( "Flags: ".. StringizeBitField( self.flags or 0, battle_defs.CARD_FLAG_BITSTRINGS ))
end

function Card:GetDesc( battle )
    return CardEngine.GenerateCardDesc( self, nil, battle, battle_defs )
end

function Card:IsAttackCard()
    return CheckAnyBits( self.flags, CARD_FLAGS.MELEE | CARD_FLAGS.RANGED )
end

function Card:IsItemCard()
    return self:IsFlagged( CARD_FLAGS.ITEM )
end

function Card:IsDamageCard()
    return self.min_damage ~= nil and self.max_damage ~= nil
end

function Card:IsMeleeAttack()
    return CheckAnyBits( self.flags, CARD_FLAGS.MELEE )
end

function Card:IsRangedAttack()
    return CheckAnyBits( self.flags, CARD_FLAGS.RANGED )
end

function Card:LookupCardDef( content_id )
    return Content.GetBattleCard( content_id )
end

function Card:__tostring()
    return string.format( "Card[%s]", tostring(self.id))
end

----------------------------------------------------------------------------
-- A single deck of cards.  Could be a draw pile, a discard pile, etc..

local Deck = class("Battle.Deck", CardEngine.Deck)
Deck.can_reload = true

function Deck:init( battle, deck_type, is_event_handler )
    Deck._base.init( self, deck_type, is_event_handler )
    self.battle = battle
    self.engine = battle
end

function Deck:RenderDebugUI( ui, panel, battle )
    if ui:Button( "Insert Card" ) then
        ui:OpenPopup( "INSERT_CARD" )
    end

    if ui:BeginPopup( "INSERT_CARD" ) then
        self.filter_txt = ui:InputText( "Filter", self.filter_txt or "" ) or self.filter_txt
        if ui:Checkbox( "Show All", self.show_all ) then
            self.show_all = not self.show_all
        end

        self.owner_idx = ui:Combo( "Owner", self.owner_idx or 1, battle.player_team.fighters ) or self.owner_idx

        ui:Separator()
        ui:BeginChild( "CARDS", 400, 400 )

        local owner = battle.player_team.fighters[ self.owner_idx ]
        if owner then
            for i, carddef in ipairs( Content.GetAllBattleCards() ) do
                if self.show_all or AgentUtil.CanUseCard( owner.agent, carddef ) then
                    local id = carddef.id
                    if self.filter_txt == nil or id:find( self.filter_txt ) then
                        if ui:Selectable( id, nil,  ui.DontClosePopups ) then
                            local card = Card( id, owner )
                            card.show_dealt = false
                            print( "Debug dealing:", card, " to ", owner )
                            battle:DealCard( card, self )
                        end
                    end
                end
            end
        end
        ui:EndChild()
        ui:EndPopup()
    end

    for i, card in ipairs( self.cards ) do
        ui:PushID( "CARD"..tostring(i) )
        ui:Indent( 20 )
        panel:AppendTable( ui, card )
        ui:Unindent( 20 )
        ui:PopID()
    end
end

--------------------------------------------------------------------------
-- Battle card engine state.

local BattleEngine = class( "BattleEngine", CardEngine )
BattleEngine.can_reload = true

BattleEngine.MAX_ACTIONS = 99
BattleEngine.MAX_HAND_SIZE = 10

BattleEngine.STATUS = STATUS
BattleEngine.EVENT = BATTLE_EVENT
BattleEngine.CARD_FLAGS = CARD_FLAGS

function BattleEngine:init( scenario )
    CardEngine.init( self )
    assert( is_instance( scenario, Scenario ))

    self.draw_deck = Deck( self, DECK_TYPE.DRAW )
    self.resolve_deck = Deck( self, DECK_TYPE.RESOLVE, true )
    self.discard_deck = Deck( self, DECK_TYPE.DISCARDS )
    self.hand_deck = Deck( self, DECK_TYPE.IN_HAND, true )
    self.hand_deck:SetMaxSize( self.MAX_HAND_SIZE )
    self.trash_deck = Deck( self, DECK_TYPE.TRASH )

    self.scenario = scenario

    self.battle_log = BattleLog( self )
    self.damage_accumulator = Battle.DamageAccumulator( self )
    self.cost_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_ACTION_COST )
    self.combo_loss_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_COMBO_LOSS )
    self.stacks_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_MODIFY_STACKS )
    self.threshold_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_THRESHOLD )
    self.draw_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_TURN_DRAW )
    self.card_flags_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_CARD_FLAGS )
    self.target_mod_accumulator = CardEngine.ScalarAccumulator( self, BATTLE_EVENT.CALC_TARGET_MOD )

    self.turns = 0
    self.action_count = 0
    self.exerted_count = 0
    self.resolve_count = 0
    self.status = STATUS.INACTIVE
    self.battle_seed = os.time()

    self.flags = self.scenario.flags or BATTLE_FLAGS.NONE

    self.player_team = Team( TEAM.BLUE, self, self.scenario.blue_team_size )
    self.enemy_team = Team( TEAM.RED, self, self.scenario.red_team_size )

    self.teams = {
        [ TEAM.BLUE ] = self.player_team,
        [ TEAM.RED ] = self.enemy_team,
    }

    for team_id, agent_list in pairs(self.scenario.teams) do
        for k,agent in pairs(agent_list) do
            local scale = self.scenario:GetAgentScale(agent)
            local pc = self.scenario:IsPlayerControlled(agent)
            local fighter = Fighter.CreateFromAgent( agent, scale, pc )
            self.teams[team_id]:AddFighter( fighter )
        end
    end

    self.active_fighters = {}
    self.resolve_queue = {}

    self.hand_size = 5
    self.cards_played = {} -- this turn
    self.all_cards_played = {} -- since battle began
    self.cards_drawn = 0
    self.actions_per_turn = 3
    self.fatigue_turn = START_FATIGUE_TURN
    
    self.coro = coroutine.create( self.Play )
end

function BattleEngine:GetDeck( deck_type )
    if deck_type == DECK_TYPE.DRAW then
        return self.draw_deck
    elseif deck_type == DECK_TYPE.IN_HAND then
        return self.hand_deck
    elseif deck_type == DECK_TYPE.RESOLVE then
        return self.resolve_deck
    elseif deck_type == DECK_TYPE.DISCARDS then
        return self.discard_deck
    elseif deck_type == DECK_TYPE.TRASH then
        return self.trash_deck
    end
end

function BattleEngine:AddBattleModule( mod )
    if self.mods == nil then
        self.mods = {}
    end
    table.insert( self.mods, mod )
    if mod.OnAddedToBattle then
        mod:OnAddedToBattle( self )
    end
end

function BattleEngine:RemoveBattleModule( mod )
    local idx = table.find( self.mods, mod )
    if idx then
        table.remove( self.mods, idx )
        if mod.OnRemovedFromBattle then
            mod:OnRemovedFromBattle( self )
        end
    end
end

function BattleEngine:Fighters( team_id )
    return self.teams[ team_id ]:Fighters()
end

function BattleEngine:ActiveFighters()
    return ipairs( self.active_fighters )
end

function BattleEngine:GetPlayerFighter()
    local primary = self:GetPlayerTeam():Primary()
    if primary and primary:IsPlayer() then
        return primary
    end
end

function BattleEngine:GetActiveFighters()
    return self.active_fighters
end

function BattleEngine:AllFighters()
    local fighters = {}
    for i, team in pairs( self.teams ) do
        for i, fighter in team:AllFighters() do
            table.insert( fighters, fighter )
        end
    end
    return pairs( fighters )
end

function BattleEngine:GetScenario()
    return self.scenario
end

function BattleEngine:GetTeam( team_id )
    return self.teams[ team_id ]
end

function BattleEngine:GetCurrentTeam()
    return self.current_team
end

function BattleEngine:IsPlayerTurn()
    return self.player_turn == true
end

function BattleEngine:GetPlayerTeam()
    return self.player_team
end

function BattleEngine:GetEnemyTeam()
    return self.enemy_team
end

function BattleEngine:GetOtherTeam( team_id )
    for k, team in pairs( self.teams ) do
        if k ~= team_id then
            return k
        end
    end
end

function BattleEngine:AppendLog( log_type, ... )
    self.battle_log:AppendLog( log_type, ... )
end

function BattleEngine:PrintLog( log_type, ... )
    self.battle_log:PrintLog( log_type, ... )
end

function BattleEngine:YieldPlay( status, ... )
    self:AppendLog( BATTLE_LOG.YIELD, status )
    return BattleEngine._base.YieldPlay( self, status, ... )
end

function BattleEngine:GetTurns()
    return self.turns
end

function BattleEngine:GetFlags()
    return self.flags
end

function BattleEngine:Win()
    assert_warning( self.result == nil )
    self.result = BATTLE_RESULT.WON
    print( "Battle Won!" )
end

function BattleEngine:Lose()
    assert_warning( self.result == nil )
    self.result = BATTLE_RESULT.LOST
    print( "Battle Lost!" )
end

function BattleEngine:Runaway()
    assert_warning( self.result == nil )
    self.result = BATTLE_RESULT.RUNAWAY
    print( "Battle was Fled!" )
end

function BattleEngine:GetBattleResult()
    return self.result
end

function BattleEngine:GetLootMoney()
    return self.loot_money or 0
end

function BattleEngine:AddLootMoney(amt)
    self.loot_money = (self.loot_money or 0) + (amt or 0)
end

function BattleEngine:AddLootItem( card_id )
    if self.loot_items == nil then
        self.loot_items = {}
    end
    table.insert( self.loot_items, card_id )
end

function BattleEngine:GetLootItems()
    return self.loot_items or table.empty
end

function BattleEngine:GetHandDeck()
    return self.hand_deck
end

function BattleEngine:GetDrawDeck()
    return self.draw_deck
end

function BattleEngine:GetDiscardDeck()
    return self.discard_deck
end

function BattleEngine:GetTrashDeck()
    return self.trash_deck
end

function BattleEngine:InsertCard( card, deck_type )
    local player = self:GetPlayerFighter()
    if player then
        if type(card) == "string" then
            card = Battle.Card( card, player )
        end
        local deck = self:GetDeck( deck_type or DECK_TYPE.IN_HAND )
        if not deck:InsertCard( card ) then
            if deck:GetDeckType() == DECK_TYPE.IN_HAND then
                -- Insert into discards instead.
                card:TransferCard( self.discard_deck )
            else
                -- Just trash it, though this can't ever happen atm.
                card:TransferCard( self.trash_deck )
            end
        end

        return card
    end
end

function BattleEngine:GetCardsOwnedBy( fighter )
    assert( fighter )
    if fighter:GetTeam() == self.player_team then
        local all_cards = table.merge( self.hand_deck.cards, self.discard_deck.cards, self.draw_deck.cards )
        for i = #all_cards, 1, -1 do
            local card = all_cards[i]
            if card:GetOwner() ~= fighter then
                table.remove( all_cards, i )
            end
        end
        return all_cards

    else
        return table.empty
    end
end

function BattleEngine:GetFighterFromSource( source )
    if is_instance( source, Fighter ) then
        return source

    elseif is_instance( source, CombatCondition ) then
        return source.owner

    elseif is_instance( source, Battle.Card ) then
        return source.owner

    elseif is_instance( source, Battle.Hit ) then
        return source.attack.attacker

    elseif is_instance( source, Battle.Attack ) then
        return source.attacker
    end
end

function BattleEngine:IsPlayerSource( source )
    local fighter = self:GetFighterFromSource( source )
    return fighter and fighter:IsPlayer()
end

function BattleEngine:IsAllySource( fighter, source )
    local other = self:GetFighterFromSource( source )
    return other and fighter:GetTeam() == other:GetTeam()
end

function BattleEngine:IsEnemySource( fighter, source )
    local other = self:GetFighterFromSource( source )
    return other and fighter:GetTeam() ~= other:GetTeam()
end

function BattleEngine:IsResolving( card )
    return self.resolve_deck:HasCard( card )
end

function BattleEngine:IsAmbushing( team )
    return self.turns == 0 and self.scenario:GetAdvantageTeam() == team:GetID()
end

function BattleEngine:GetActionCount()
    return self.action_count, self.actions_per_turn
end

function BattleEngine:GetActionsPerTurn()
    return self.actions_per_turn
end

function BattleEngine:CountCardsDrawn()
    return self.cards_drawn
end

function BattleEngine:GetAllCardsPlayed()
    return self.all_cards_played
end

function BattleEngine:GetCardsPlayed()
    return self.cards_played
end

function BattleEngine:CountCardsPlayed( fn )
    if fn then
        return table.count_if( self.cards_played, fn )
    else
        return #self.cards_played
    end
end

function BattleEngine:CountCardsDiscarded()
    return self.cards_discarded or 0
end

function BattleEngine:GetExertedCount()
    return self.exerted_count
end

function BattleEngine:ModifyActionCount( delta, turn_start )
    local action_count = clamp( self.action_count + delta, 0, self.MAX_ACTIONS )
    if action_count ~= self.action_count then
        delta = action_count - self.action_count
        self.action_count = action_count
        self:BroadcastEvent( BATTLE_EVENT.DELTA_ACTION_COUNT, self.action_count, self.actions_per_turn, delta, turn_start == true )
    end
end

function BattleEngine:ModifyExertedCount( delta )
    self.exerted_count = self.exerted_count + delta

    if delta < 0 then 
        --jcheng: do not let you go above your max actions
        delta = math.max( self.action_count - self.actions_per_turn, delta )
    end

    self:ModifyActionCount(-delta)
end

function BattleEngine:CollectRandomTargets( target_fighters, fighters, n, card )
    if #fighters == 0 then
        return
    end

    -- Convert to a weighted table.
    fighters = self:WeightTargets( card, fighters )

    -- Don't select any fighters which are already targetted.
    for i, fighter in ipairs( target_fighters ) do
        fighters[ fighter ] = nil
    end

    -- Track the surrendered (last resort targets) and invalid ones.
    local surrendered = {}
    for fighter, wt in pairs( fighters ) do
        if fighter:HasCondition("SURRENDER") then
            table.insert(surrendered, fighter)
            fighters[ fighter ] = nil

        elseif card and not fighter:CanBeTargetted( card ) then
            fighters[ fighter ] = nil
        end
    end

    for i = 1, math.min( n, table.count( fighters ) + #surrendered ) do
        local fighter = weightedpick( fighters, nil )
        if fighter ~= nil then
            fighters[ fighter ] = nil
            table.insert(target_fighters, fighter)

        elseif #surrendered > 0 then
            local fighter = table.remove( surrendered, math.random( 1, #surrendered ))
            table.insert(target_fighters, fighter)
        else
            break
        end        
    end
end

function BattleEngine:CollectTargets(card, primary_target)
    local attacker = card:GetOwner()

    --Pick a primary target from the list of all valid ones
    if primary_target == nil then
        local t = self:CollectAllTargets( card, nil, 1 )
        primary_target = t[1]
        if primary_target == nil then
            return table.empty
        end
    end

    local target_mod = self.target_mod_accumulator:CalculateValue( card.target_mod, card, primary_target )
    local target_fighters = {}

    if target_mod == TARGET_MOD.SINGLE then
        if primary_target:IsAlive() then
            table.insert( target_fighters, primary_target )
        end

    elseif target_mod == TARGET_MOD.RANDOM1 then
        self:CollectRandomTargets( target_fighters, primary_target.team.fighters, 1, card )

    elseif target_mod == TARGET_MOD.RANDOM2 then
        self:CollectRandomTargets( target_fighters, primary_target.team.fighters, 2, card )

    elseif target_mod == TARGET_MOD.RANDOM3 then
        self:CollectRandomTargets( target_fighters, primary_target.team.fighters, 3, card )

    elseif target_mod == TARGET_MOD.TEAM then
        -- target everyone on the same team as primary target
        for _, fighter in ipairs(primary_target.team.fighters) do
            table.insert(target_fighters, fighter)
        end

    elseif target_mod == TARGET_MOD.ALL then
        -- target everyone on battlefield
        for _, fighter in ipairs(attacker.team.fighters) do
            table.insert(target_fighters, fighter)
        end

        for _, fighter in ipairs(attacker.team:GetEnemyTeam().fighters) do
            table.insert(target_fighters, fighter)
        end

    elseif target_mod == TARGET_MOD.CUSTOM then
        card:target_fn( self, primary_target, target_fighters )
    end

    return target_fighters
end


function BattleEngine:CollectAllTargets( card, t, max_count )
    max_count = max_count or math.huge
    t = t or {}

    for i, fighter in self:ActiveFighters() do
        if card.owner:CanTarget( card, fighter ) then
            table.insert( t, fighter )
            if #t >= max_count then
                return t
            end
        end
    end

    return t
end

function BattleEngine:GetNumTargeters( target )
    local num_targeters = 0
    for i, fighter in self:ActiveFighters() do
        num_targeters = num_targeters + (fighter:IsTargetting( target ) and 1 or 0)
    end
    
    return num_targeters
end

function BattleEngine:WeightTargets( card, t )
    local weighted_targets = {}

    local min_targetting = math.huge
    for i, fighter in ipairs( t ) do
        min_targetting = math.min( min_targetting, self:GetNumTargeters(fighter) )
    end
    -- We only start removing target options for targets with 2 or more attackers, if possible.
    min_targetting = math.max( 1, min_targetting )

    for i, fighter in ipairs( t ) do
        
        if card and card.EvaluateTargetWeight then
            weighted_targets[ fighter ] = card:EvaluateTargetWeight( fighter, t )
        else
            weighted_targets[ fighter ] = fighter:GetTargetWeight()
        end
        
        if card and card:IsAttackCard() and not card.owner:IsPlayer() and weighted_targets[fighter] then
            --
            local num_targeters = self:GetNumTargeters(fighter)
            local hp = fighter:GetHealthPercent()
            local p = .5 + .5*hp

            if num_targeters <= min_targetting then
                weighted_targets[ fighter ] = weighted_targets[ fighter ]*p * (1 - num_targeters/(min_targetting + 1))
            else
                weighted_targets[ fighter ] = nil
            end
        end
    end
    return weighted_targets
end

function BattleEngine:ValidateTargets( card, attack )
    -- Validate the primary target.
    local target = card:GetTarget()
    if target and not card.owner:CanTarget( card, target ) then
        print( "\tRetargetting:", card.owner, card, "from", target, card.owner:CanTarget( card, target ))

        card:AssignTarget( nil )
        self:AssignPrimaryTarget( card )

        if attack then
            attack:ReassignTarget( target, card:GetTarget() )
        end
    end

    -- Validate secondary targets for non-AOEs
    if card.target_mod == TARGET_MOD.SINGLE then
        for i, targets in ipairs( card.secondary_targets or table.empty ) do
            for j, target in ipairs( targets ) do
                if not card.owner:CanTarget( card, target ) then
                    print( "\tRetargetting secondary:", card, "from", target, card.owner:CanTarget( card, target ))
                    targets[j] = self:CollectPrimaryTarget( card )
                    if attack then
                        attack:ReassignTarget( target, targets[j] )
                    end
                end
            end
        end
    end
end

function BattleEngine:CollectPrimaryTarget( card, target )

    if target and not card.owner:CanTarget( card, target ) then
        target = nil
    end

    -- print( "ASSIGN TARGET:", card, target, tostr(t), tostr(weighted_targets))

    if target == nil then
        local t = {}
        self:CollectAllTargets( card, t )

        local weighted_targets = self:WeightTargets( card, t )
        target = weightedpick( weighted_targets )
    end

    return target
end

function BattleEngine:AssignPrimaryTarget( card, target )
    target = self:CollectPrimaryTarget( card, target or card:GetTarget() )

    local secondary_targets = card.secondary_targets
    if target ~= card.target then
        -- Generate secondary targets from the primary target.
        secondary_targets = {}
        if target then
            secondary_targets = {}
            for i = 1, card.target_count or 1 do
                secondary_targets[i] = self:CollectTargets( card, target )
            end
        end

        -- print( "ASSIGNIGN secondary_targets to", card, target, tostr(secondary_targets, 2))
    end

    card:AssignTarget( target, secondary_targets )

    return target ~= nil
end

function BattleEngine:CanPlayAny( ignore_trivials )
    for i, card in self.hand_deck:Cards() do
        if self:CanPlayCard( card ) then
            if ignore_trivials and (card.id == "empty_bottle" or card:IsFlagged( CARD_FLAGS.TOOLBOX )) then
                -- skip these cards for the UI warning, they are trivial plays.
            else
                return true
            end
        end
    end
    return false
end

function BattleEngine:CanPlayCard( card, target )
    local owner = card:GetOwner()

    if owner.player_controlled then
        if self.hand_deck:HasCard( card ) then
            if not card:IsFlaggedAny( CARD_FLAGS.VARIABLE_COST | CARD_FLAGS.FREEBIE ) then
                local cost = self:CalculateActionCost( card, target )
                if cost > self.action_count then
                    return false, CARD_PLAY_REASONS.NO_ACTIONS
                end
            end
        end

        if not card:IsFlaggedAny( CARD_FLAGS.SPECIAL | CARD_FLAGS.FLOURISH) and card.deck ~= self.hand_deck then
            return false --, CARD_PLAY_REASONS.NOT_IN_HAND
        end

    else
        if card.deck == nil or card.deck:GetDeckType() ~= DECK_TYPE.AI then
            return false --, CARD_PLAY_REASONS.NOT_IN_HAND
        end
    end

    if card:IsFlaggedAny( CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.AUTOPLAY ) then
        return false, CARD_PLAY_REASONS.UNPLAYABLE
    end

    -- if owner:GetCondition( "STUN" ) then
    --     return false, CARD_PLAY_REASONS.STUNNED
    -- end

    local ok, reason = true, nil

    local initial_conditions = shallowcopy( owner:GetConditions() )
    local conditions = owner:SortConditionTriggers( "CanPlayCard", initial_conditions )
    for i, condition in ipairs( conditions ) do
        if condition.CanPlayCard and ok then
            ok, reason = condition:CanPlayCard( self, target, card )
        end
    end

    if not ok then
        return ok, reason
    end
    
    local charges = card:GetCharges()
    if charges and charges <= 0 then
        return false, CARD_PLAY_REASONS.NO_CHARGE
    end

    if not card:GetOwner():IsActive() then
        return false
    end

    -- If there is a provided target, validate versus it.
    if target then
        local ok, reason = owner:CanTarget( card, target )
        if not ok then
            return false, reason or CARD_PLAY_REASONS.INVALID_TARGET
        end
    
    -- Otherwise, this card is playable if it is playable against SOME active target.
    else
        if card.CanPlayCard then
            local ok, reason = card:CanPlayCard( self )
            if not ok then
                return false, reason
            end
        end

        local t = ObtainWorkTable()
        self:CollectAllTargets( card, t, 1 )
        local target_count = #t
        ReleaseWorkTable( t )

        if target_count == 0 then
            return false, CARD_PLAY_REASONS.NO_VALID_TARGETS
        end
    end

    return true
end

function BattleEngine:CalculateTurnDraw()
    local hand_size, details = self.draw_accumulator:PreviewValue( self.hand_size )
    hand_size = math.max( 1, hand_size )
    return hand_size, details
end

function BattleEngine:ModifyTurnDraw( delta )
    self.hand_size = clamp( self.hand_size + delta, 0, 10 )
end

function BattleEngine:OnActivateFighter( fighter )
    table.insert( self.active_fighters, fighter )

    self:BroadcastEvent( BATTLE_EVENT.ACTIVATED_FIGHTER, fighter )
end

function BattleEngine:OnDeactivateFighter( fighter )
    table.arrayremove( self.active_fighters, fighter )

    local cards = {}
    for deck_type, _ in pairs( DECK_TYPE ) do
        local deck = self:GetDeck( deck_type )
        if deck then
            for i = #deck.cards, 1, -1 do
                local card = deck.cards[i]
                if card:GetOwner() == fighter then
                    deck:RemoveCard( card )
                    table.insert( cards, card )
                end
            end
        end
    end

    for i = #self.resolve_queue, 1, -1 do
        local card = self.resolve_queue[i].card
        if card.owner == fighter then
            table.remove( self.resolve_queue, i)
        end
    end

    self:BroadcastEvent( BATTLE_EVENT.DEACTIVATED_FIGHTER, fighter, cards )
end

function BattleEngine:Play()
    -- move the sound listener so it's actually in the right spot for battles
    engine.sound:SetListenerPosUpFwd(0, 1, 5, 0, 1, 0,  0, 0, -1) 

    self.player_team:ActivateNewFighters()
    self.enemy_team:ActivateNewFighters()
    self.draw_deck:Shuffle()

    self:BroadcastEvent( BATTLE_EVENT.BEGIN_COMBAT )

    if self.scenario.on_start_battle then
        self.scenario.on_start_battle(self)
    end


    -- Sync so that viz setup can query battle state.
    self:YieldPlay( STATUS.SYNC )

    self:RunResolvePhase()

    while not self:CheckGameOver() do

        self.turns = self.turns + 1
        self:ModifyActionCount( self.actions_per_turn - self.action_count, true )

        if self.scenario.on_turn_start then
            self.scenario.on_turn_start(self)
        end


        if self.turns == 1 and self.scenario:GetAdvantageTeam() == TEAM.RED then
            -- 1st turn advantage hack: skip player phase
        else
            
            if self.turns == 1 and self.scenario.time_penalty then
                self:ModifyActionCount( -self.scenario.time_penalty )
            end

            self:RunBeginPlayerTurn()

            while not self:CheckGameOver() do

                -- Wait for user choice...
                local card = self:YieldPlay( STATUS.PLAY_CARD )
                assert( card == nil or is_instance( card, Card ))

                if not card then
                    if not self.added_executes then
                        self:AppendLog( BATTLE_LOG.GENERAL, "End Turn! ({1} time left)", self.action_count )
                        break
                    end

                else
                    self:PayCosts( card )
                    self:PlayCard( card )
                end

                -- Resolve cards played.
                self:RunResolvePhase()
            end

            -- Discard hand, etc.
            self:RunEndPlayerTurn()
            self.exerted_count = 0

        end

        if not self:CheckGameOver() then
            self:RunEnemyPhase()
        end
    end

    self:BroadcastEvent( BATTLE_EVENT.END_COMBAT )
    self:AppendLog( BATTLE_LOG.GENERAL, "Completed with result: {1}", self.result )
end


function BattleEngine:GetFighterForAgent(agent)
    return self.player_team:GetFighterForAgent(agent) or self.enemy_team:GetFighterForAgent(agent)
end

function BattleEngine:RePrepareTurn()
    for i, enemy in self.enemy_team:Fighters() do
        enemy:ClearPreparedCards()
        enemy:PrepareTurn()
    end

    for i, ally in self.player_team:Fighters() do
        ally:ClearPreparedCards()
        ally:PrepareTurn()
    end
end
function BattleEngine:CheckFatigue()
    local player = self:GetPlayerFighter()
    if player then
        local extra_turns = math.min(self.scenario:GetStartingTeamSize(TEAM.RED)-1, 3)
        if self.scenario:IsBossFight() then
            extra_turns = extra_turns + 3
        end

        if (self.turns >= self:GetFatiguedTurn() + extra_turns) and not player:HasCondition("FATIGUED") then
            player:AddCondition("FATIGUED")
        end
    end
end

function BattleEngine:RunBeginPlayerTurn()

    if self.scenario.before_player_turn_start then
        self.scenario.before_player_turn_start(self)
    end

    self.player_turn = true
    self.cards_discarded = 0
    self.current_team = self.player_team
    self.player_team:ActivateNewFighters()

    self:CheckFatigue()

    local fighters = shallowcopy( self.player_team:GetFighters())
    for i, fighter in ipairs( fighters ) do
        fighter:OnBeginTurn()
        fighter:PrepareTurn()
    end

    for i, enemy in self.enemy_team:Fighters() do
        enemy:PrepareTurn()
    end

    if self:GetEnemyTeam():IsSurrendering() then
        self:AddExecutes()
    else

        if self.turns == 1 then
            if self.scenario:GetAdvantageTeam() == self.player_team:GetID() then
                -- Gain one action.
                self:ModifyActionCount( 1 )
            end

            -- Ambush cards go directly into hand on turn 1 only.
            local function IsAmbush( card )
                return card:IsFlagged( CARD_FLAGS.AMBUSH )
            end
            self.draw_deck:TransferCardsIf( self.hand_deck, IsAmbush )
        end

        -- Toolbox are always in hand.
        local function IsToolBox( card )
            return card:IsFlagged( CARD_FLAGS.TOOLBOX )
        end
        self.draw_deck:TransferCardsIf( self.hand_deck, IsToolBox )
        self.discard_deck:TransferCardsIf( self.hand_deck, IsToolBox )

        local hand_size = self:CalculateTurnDraw()
        self:DrawCards( hand_size, true )
    end


    self:BroadcastEvent( BATTLE_EVENT.BEGIN_PLAYER_TURN, self )

    if self.scenario.on_player_turn_start then
        self.scenario.on_player_turn_start(self)
    end
end

function BattleEngine:RunEnemyPhase()

    self.current_team = self.enemy_team

    self:BroadcastEvent( BATTLE_EVENT.BEGIN_ENEMY_TURN )

    local fighters = shallowcopy( self.enemy_team:GetFighters())
    for i, fighter in ipairs( fighters ) do
        if fighter:IsAlive() then
            fighter:OnBeginTurn()
        end
    end

    local fighters = shallowcopy( self.enemy_team:GetFighters())
    for i, fighter in ipairs( fighters ) do
        if fighter:IsAlive() then
            fighter:OnEndTurn()
            self:YieldPlay( STATUS.SYNC )
        end

        if self:CheckGameOver() then
            break
        end
    end

    self.current_team = nil
    
    self.enemy_team:ActivateNewFighters()
    

    if self.scenario.on_enemy_turn_end then
        self.scenario.on_enemy_turn_end(self)
    end


    self:BroadcastEvent( BATTLE_EVENT.END_ENEMY_TURN )
    self:YieldPlay( STATUS.SYNC )
end

function BattleEngine:RunResolvePhase()
    self.resolving = true

    while #self.resolve_queue > 0 do
        local t = table.remove( self.resolve_queue, 1 )
        local resolve_card = t.card
        if t.target == nil or t.target ~= resolve_card.target then
           self:AssignPrimaryTarget( resolve_card, t.target )
       else
           self:ValidateTargets( resolve_card )
       end

        print( "Resolving: ", resolve_card, resolve_card:GetTarget() )
        resolve_card:TransferCard( self.resolve_deck )

        self:BroadcastEvent(BATTLE_EVENT.START_RESOLVE, self, resolve_card)
        if resolve_card:IsFlagged( CARD_FLAGS.TOOLBOX) then
            if resolve_card.userdata.stored_cards and #resolve_card.userdata.stored_cards > 0 then
                self:SearchToolbox(resolve_card.userdata.stored_cards, resolve_card)
            else
                -- failsafe, this really shouldn't happen
                if resolve_card.refreshing then
                    self:ExpendCard(resolve_card)
                else
                    resolve_card:TransferCard( self.trash_deck )
                    resolve_card:Consume()
                end
            end
        else
            do
                local attack = t.attack
                if attack == nil then
                    attack = Battle.Attack( resolve_card )
                    attack:GenerateHits()
                end

                self:AppendLog( BATTLE_LOG.GENERAL, "{1} resolving {2} vs {3}", resolve_card:GetOwner(), resolve_card, attack.target )

                if resolve_card.OnPreResolve then
                    resolve_card:OnPreResolve( self, attack )
                end
                self:BroadcastEvent( BATTLE_EVENT.PRE_RESOLVE, self, attack )

                if not attack.attacker:IsActive() then
                    -- LOGWARN( " is not active after PRE_RESOLVE...%s", tostring(attack) )
                else
                    attack:RunAttack()
                end

                if resolve_card.OnPostResolve then
                    resolve_card:OnPostResolve( self, attack )
                end
                self:BroadcastEvent( BATTLE_EVENT.POST_RESOLVE, self, attack )

                -- Clean up the resolve card and figure out where it should go
                --------------------------------------------------------------

                resolve_card:AddXP(1)

                if resolve_card:IsFlagged( CARD_FLAGS.HATCH ) and resolve_card:UpgradeReady() then
                    if resolve_card.hatch_fn then 
                        if not resolve_card.hatched then
                            resolve_card:hatch_fn(self)
                            resolve_card.hatched = true
                        else
                            resolve_card:TransferCard( self.trash_deck )
                            resolve_card:Consume()
                        end

                    elseif resolve_card.userdata.hatch_id then
                        self:HatchCard( resolve_card, resolve_card.userdata.hatch_id )
                    end
                end

                resolve_card:AssignTarget( nil )
                
                resolve_card:ConsumeCharge()

                if resolve_card:IsSpent() then
                    resolve_card:SetFlags( CARD_FLAGS.CONSUME )
                end
                resolve_card.flags = ClearBits( resolve_card.flags, CARD_FLAGS.FREEBIE )

                if resolve_card.owner and resolve_card.owner:IsPlayer() and not resolve_card:IsFlagged( CARD_FLAGS.SPECIAL ) then
                    StatUtil.PlayedCard(resolve_card.id)
                    table.insert( self.cards_played, resolve_card )
                    table.insert( self.all_cards_played, resolve_card )
                    if #self.cards_played >= INFINITE_THRESHOLD and not self.reported_infinite then
                        self.reported_infinite = true
                        local player_state = TheGame:GetGameState():GetPlayerState()
                        if player_state then
                            local json_t = {
                                PLAYER_DATA = player_state,
                            }
                            SendMetricsData("BATTLE_INFINITE_DETECTED", json_t )
                        end
                    end
                end

                self.resolve_count = self.resolve_count + 1
            end

            -- Where does this resolved card go now?
            if resolve_card:IsFlagged( CARD_FLAGS.SPECIAL ) then
                -- Special cards are implementation-only cards: they're so special they are removed entirely.
                resolve_card:RemoveCard()

            elseif resolve_card:IsFlagged(CARD_FLAGS.CONSUME) then
                --
                resolve_card:TransferCard( self.trash_deck )
                resolve_card:Consume()

            -- Expended: go directly to the trash.
            elseif resolve_card:IsFlagged(CARD_FLAGS.EXPEND) or resolve_card:IsFlagged(CARD_FLAGS.FLOURISH) then
                self:ExpendCard( resolve_card )

            elseif resolve_card.deck == self.resolve_deck then
                -- If the card is no longer in the resolve deck, then something moved it
                -- somewhere else and we should not attempt any post-resolve transfer.
                if not resolve_card.owner:IsPlayer() then
                    local deck = resolve_card.owner:GetDiscardDeck()
                    resolve_card:TransferCard( deck )

                elseif resolve_card:IsFlagged( CARD_FLAGS.MULTISHOT ) then
                    -- Multi-shots get re-inserted into your hand.
                    resolve_card:TransferCard( self.hand_deck )
                else
                    -- By default, most things go into the discards after being resolved.
                    resolve_card:TransferCard( self.discard_deck )
                end
            end

            self:BroadcastEvent(BATTLE_EVENT.END_RESOLVE, self, resolve_card)
            self:CheckAllSurrenders()

            -- Potential changes on the player turn could require re-preparing new valid actions
            for i, enemy in self.enemy_team:Fighters() do
                if enemy ~= resolve_card.owner and enemy.prepared_attacks then
                    enemy:PrepareTurn()
                end
            end

            for i, ally in self.player_team:Fighters() do
                if ally ~= resolve_card.owner and ally.prepared_attacks then
                    ally:PrepareTurn()
                end
            end
        end

        if self.result ~= nil then
            table.clear( self.resolve_queue )
            break
        end
    end

    self.resolving = false
end

function BattleEngine:HatchCard( card, card_id )
    card:TransferCard( self.trash_deck )
    card:Consume()

    local new_card = Battle.Card( card_id, card.owner:GetAgent() )
    new_card.owner.battler:AddCard(new_card)

    local battle_card = new_card:Clone()
    battle_card.owner = card.owner
    self:DealCard(battle_card, self:GetHandDeck())
end

function BattleEngine:CheckAllSurrenders()
    local all_fighters = shallowcopy(self.active_fighters)
    for i,fighter in ipairs(all_fighters) do
        if fighter ~= self:GetPlayerFighter() then
            fighter:CheckForSurrender()
        end
    end
    self.player_team:ActivateNewFighters(true)
    self.enemy_team:ActivateNewFighters(true)
    self:BroadcastEvent( BATTLE_EVENT.CHECKED_FOR_SURRENDER )
end

function BattleEngine:AddExecutes()
    if not self.added_executes then
        self.hand_deck:TransferCards( self.trash_deck )
        self.draw_deck:TransferCards( self.trash_deck )
        self.discard_deck:TransferCards( self.trash_deck )
        self.resolve_deck:TransferCards( self.trash_deck )

        for i=1, #self.enemy_team.fighters do
            if self.enemy_team.fighters[i]:IsAlive() then
                local card = Battle.Card( "execute", self.player_team:Primary() )
                card.show_dealt = false
                self:DealCard( card, self:GetHandDeck() )
            end
        end
        self.added_executes = true
        table.clear(self.resolve_queue)

        self:BroadcastEvent( BATTLE_EVENT.EXECUTES_ADDED )
    end
end

function BattleEngine:HasExecutes()
    return self.added_executes
end

function BattleEngine:DealCard( card, deck )
    if self.added_executes then
        print( "Executes added: not dealing ", card )
        return
    end
    return self._base.DealCard( self, card, deck )
end

function BattleEngine:Delay( delay )
    self:BroadcastEvent( BATTLE_EVENT.DELAY, delay )
end

function BattleEngine:RunEndPlayerTurn()

    if not self:CheckGameOver() then
        -- Resolve all AUTOPLAY cards remaining in hand.
        self:PlayCards( self.hand_deck, Card.IsAutoplay )
    end

    -- This happens before discards so that cards can trigger off END_PLAYER_TURN.
    self:BroadcastEvent( BATTLE_EVENT.END_PLAYER_TURN, self )

    local NeedsDiscarding = function( card )
        return not card:IsFlagged( CARD_FLAGS.STICKY ) and not card:IsFlagged( CARD_FLAGS.TOOLBOX ) and not card:IsFlagged( CARD_FLAGS.SLIMY )
    end

    local wt = ObtainWorkTable()
    table.arrayadd( wt, self.hand_deck.cards )
    for k,card in ipairs(wt) do
        if card:IsFlagged(CARD_FLAGS.BURNOUT) or card:IsFlagged(CARD_FLAGS.FLOURISH) then
            self:ExpendCard(card)
        end
    end
    ReleaseWorkTable(wt)

    -- Lastly, discard all cards in hand.
    if not self.no_discard then
        self.hand_deck:TransferCardsIf( self.discard_deck, NeedsDiscarding)
    end

    self.player_turn = false

    if not self:CheckGameOver() then
        for i, fighter in self.player_team:Fighters() do
            fighter:OnEndTurn()
            self:YieldPlay( STATUS.SYNC )
        end
    end

    table.clear( self.cards_played )
    self.cards_drawn = 0
    
    self:YieldPlay( STATUS.SYNC )
end

function BattleEngine:PayCosts( card )
    if card.owner:IsPlayer() then
        local cost = self:CalculateActionCost( card, card:GetTarget() )
        self:ModifyActionCount( -cost )

        if card:IsFlagged( CARD_FLAGS.FLOURISH ) then
            local flourish_graft = self:GetPlayerFighter().agent.graft_owner:GetGraft("flourish_tracker")
            if flourish_graft then
                flourish_graft:GetDef().SetBattleFlourish( flourish_graft, self, 0 )
            end
        end
    end
end

function BattleEngine:PlayCard( card, target, attack )
    if self.result ~= nil then
        return
    end
    if card:IsFlagged(CARD_FLAGS.UNPLAYABLE) then
        return
    end
    local t = { card = card, target = target or card:GetTarget(), attack = attack }
    table.insert( self.resolve_queue, t )

    if not self.resolving then
        self:RunResolvePhase()
    end
end

function BattleEngine:CheckGameOver()
    
    if self.result == nil then
        local enemy_status = self.enemy_team:GetTeamStatus()
        if enemy_status == FIGHT_STATUS.FLED or enemy_status == FIGHT_STATUS.DEAD then
            -- NOTE: Surrender is a manual option in FightScreen.
            self:Win()

        else
            local player_status = self.player_team:GetTeamStatus()
            if player_status == FIGHT_STATUS.FLED then
                self:Runaway()
            elseif player_status ~= FIGHT_STATUS.ACTIVE then
                self:Lose()
            else
                local turns = self.scenario:GetSurvivalTurns()
                if turns and self.turns > turns then
                    self:Win()
                end 
            end
        end
    end

    return self.result ~= nil
end

function BattleEngine:CalculateActionCost( card, target )
    if card:IsFlagged( CARD_FLAGS.FREEBIE ) then
        return 0
    else
        local cost, details = self.cost_accumulator:PreviewValue( card.cost, card, target )
        cost = math.max( 0, cost )
        return cost, details
    end
end

function BattleEngine:CalculateComboLoss( combo, source, target )
    local new_combo, details = self.combo_loss_accumulator:PreviewValue( combo, source, target )
    new_combo = math.max( 0, new_combo )
    return new_combo, details
end

function BattleEngine:CalculateCardFlags( card )
    local flags = self.card_flags_accumulator:PreviewValue( card.flags, card )
    return flags
end

function BattleEngine:CalculateTargetMod( card )
    return self.target_mod_accumulator:CalculateValue( card.target_mod, card, card:GetTarget() )
end

function BattleEngine:CalculateModifiedStacks( condition_id, stacks, fighter, source )
    local condition_def = Content.GetBattleCondition( condition_id )
    local stacks, details = self.stacks_accumulator:PreviewValue( stacks, condition_id, fighter, source )
    return stacks, details
end


function BattleEngine:PreviewDamage( card, target, hit )
    return self.damage_accumulator:PreviewDamage( card, target, hit )
end

function  BattleEngine:CalculateThreshold( card )
    local threshold, details = self.threshold_accumulator:CalculateValue( card.threshold, card )
    return threshold, details
end

function BattleEngine:ChooseCardsForHand( cards, count, event_name, replacement_id, can_skip, txt, source )
    assert( cards )

    count = count or 1
    local min_count = can_skip and 0 or count

    if #cards < count and replacement_id then
        local bonus = count - #cards
        for i=1, bonus do
            table.insert(cards, Battle.Card(replacement_id, self:GetPlayerFighter()))            
        end
    end

    if #cards > count then
        txt = txt or loc.format( LOC"CARD_ENGINE.CHOOSE_CARDS", count )
        AUDIO:PlayEvent("event:/sfx/battle/cards/neg/improvised_card_presented")
        cards = self:ChooseCardsFromTable( cards, min_count, count, nil, txt, source )
    end

    for i, card in ipairs( cards ) do
        if card.deck == nil then
            -- Because you chose it, seeing the CARD_DEALT viz is annoying.
            card.show_dealt = false
            self:DealCard( card, self:GetHandDeck() )
        else
            card.deck:TransferCard( card, self:GetHandDeck() )
        end
        if card.deck ~= self:GetHandDeck() then
            if card.deck then
                card:RemoveCard()
            end
            card:TransferCard( self:GetDiscardDeck() )
        else
            AUDIO:PlayEvent("event:/sfx/battle/cards/neg/improvised_goes_to_deck")
            if event_name then
                self:BroadcastEvent( event_name, card, self )
            end
        end
    end
    return cards
end

function BattleEngine:ImproviseCards( cards, count, replacement_id, can_skip, txt, source )
    return self:ChooseCardsForHand( cards, count, BATTLE_EVENT.CARD_IMPROVISED, replacement_id, can_skip, txt, source )
end


function BattleEngine:SearchToolbox( cards, toolbox_card )
    assert( cards )
    local options = shallowcopy(cards)
    table.insert(options, Battle.Card("cancel_toolbox", toolbox_card.owner))
    local chosen_card = self:ChooseCardsFromTable( options, 1, 1, nil, loc.format( LOC"CARD_ENGINE.CHOOSE_CARDS", 1 ) )[1]
    if chosen_card == nil then
        return -- This can happen if the battle is already over by the time this runs.
    end

    if chosen_card.id ~= "cancel_toolbox" then
        local cloned_card = chosen_card:Duplicate()
        cloned_card.show_dealt = false
        cloned_card.owner = toolbox_card.owner
        if chosen_card:GetCharges() then
            chosen_card:ConsumeCharge()
            if chosen_card:IsSpent() then
                table.arrayremove(toolbox_card.userdata.stored_cards, chosen_card)
            end
            cloned_card:SetFlags(CARD_FLAGS.EXPEND) -- The base card has multiple charges so this clone should only function as a single charge.
        else
            table.arrayremove(toolbox_card.userdata.stored_cards, chosen_card)
        end
        if #toolbox_card.userdata.stored_cards == 0 then
            if toolbox_card.refreshing then
                self:ExpendCard(toolbox_card)
            else
                toolbox_card:TransferCard( self.trash_deck )
                toolbox_card:Consume()
            end
        else
            toolbox_card:TransferCard( self.discard_deck )
        end
        self:DealCard( cloned_card, self:GetHandDeck() )

    else
        toolbox_card:TransferCard( self.hand_deck )
    end

    -- FIXME: hack identified
    -- This card isn't/shouldn't be actually in the engine, but this causes it to animate to the Trash.
    chosen_card:TransferCard(self:GetTrashDeck())
    chosen_card:RemoveCard()
end

function BattleEngine:GetFatiguedTurn()
    return self.fatigue_turn
end

function BattleEngine:SetFatiguedTurn( val )
    self.fatigue_turn = val
end

function BattleEngine:RenderDebugUI( ui, panel )
    ui:Text( string.format( "Turn: %d", self.turns ))
    ui:SameLine( nil, 20 )
    ui:TextColored( 0x00CCCCFF, self.status )
    if ui:IsItemHovered() then
        ui:SetTooltip( tostr(self.wait_params) )
    end

    local action_count = ui:SliderInt( "Actions", self.action_count, 0, 9 )
    if action_count then
        self:ModifyActionCount( action_count - self.action_count )
    end

    if ui:TreeNode( "Decks" ) then
        self.debug_deck = self.debug_deck or table.arrayfind( DECK_TYPE_ARRAY, DECK_TYPE.IN_HAND )
        local new_deck = ui:Combo( "Deck", self.debug_deck, DECK_TYPE_ARRAY )
        if new_deck and new_deck ~= self.debug_deck then
            self.debug_deck = new_deck
        end

        local deck = self:GetDeck( DECK_TYPE_ARRAY[ self.debug_deck ])
        if deck then
            deck:RenderDebugUI( ui, panel, self )
        end
        ui:TreePop()
    end

    if ui:TreeNode( "Log" ) then
        self.battle_log:RenderDebug( ui, panel )
        ui:TreePop()
    end

    ui:Spacing()

    panel:AppendTable( ui, self.scenario, "Scenario" )

    if self.coro then
        ui:SameLine( nil, 10 )
        if ui:Button( "Battle Coroutine" ) then
            TheGame:GetDebug():CreatePanel( DebugTraceback( self.coro ))
        end
    end
end

