--------------------------------------------------------------
-- Logs battle events for debugging and UI presentation.

local BattleLog = class("BattleLog")

BATTLE_LOG, BATTLE_LOG_STRINGS = MakeBitField{
    "GENERAL", -- Log text for generally relevant battle events.
    "AI",   -- AI-decisions.
    "PHASE", -- Phases.
    "STAT", -- Fighter stat changes
    "YIELD", -- BattleEngine coroutine yielding
}

function BattleLog:init( battle )
    assert( is_instance( battle, BattleEngine ))
    self.battle = battle
    self.items = {}
    if TheGame:GetDebug() then
        self.enabled_types = BATTLE_LOG.ANY
    else
        self.enabled_types = BATTLE_LOG.GENERAL
    end
    self.filter_types = BATTLE_LOG.ANY
    self.start_time = engine.inst:QueryPreciseTime()
end

function BattleLog:AppendLog( log_type, ... )
    if not CheckBits( self.enabled_types, log_type ) then
        return
    end

    local item = { t = engine.inst:QueryPreciseTime(), log_type = log_type }
    if log_type == BATTLE_LOG.GENERAL then
        item.text = loc.format( ... )

    elseif log_type == BATTLE_LOG.AI then
        item.text = loc.format( ... )

    elseif log_type == BATTLE_LOG.PHASE then
        local fighter = ...
        item.text = loc.format( "{1} turn {2}: {3}", fighter, self.battle:GetTurns(), select( 2, ... ))

    elseif log_type == BATTLE_LOG.STAT then
        local fighter, stat, delta = ...
        item.text = loc.format( "{1} {2} changed by {3} (currently {4})", fighter, stat, delta, fighter:GetStat( stat ) )

    elseif log_type == BATTLE_LOG.YIELD then
        local status = ...
        item.text = loc.format( "{1}: {2}", status, item.t - self.start_time )
    end

    table.insert( self.items, item )
    return item
end

function BattleLog:PrintLog( log_type, ... )
    local item = self:AppendLog( log_type, ... )
    print( item and item.text )
end

function BattleLog:Dump()
    print( "\nBATTLE LOG\n##########")
    for i, item in ipairs( self.items ) do
        if CheckBits( self.filter_types, item.log_type ) then
            local s1 = string.format( "[%s] %s", StringizeBitField( item.log_type, BATTLE_LOG_STRINGS ), tostring( item.text ))
            print( s1 )
        end
    end
end

function BattleLog:RenderDebug( ui, panel )
    if ui:Button( "Clear" ) then
        table.clear( self.items )
    end
    ui:SameLine( nil, 10 )
    if ui:Button( "Print-Dump" ) then
        self:Dump()
    end

    ui:Separator()
    ui:Columns(2)
    ui:SetColumnOffset(1, 120)
    ui:Text( "Log Type" )
    ui:NextColumn()

    ui:Text( "Data")
    ui:NextColumn()

    ui:Separator()
    for i, item in ipairs( self.items ) do
        if CheckBits( self.filter_types, item.log_type ) then
            ui:TextColored( 0xCC00FFFF, StringizeBitField( item.log_type, BATTLE_LOG_STRINGS ))
            ui:NextColumn()

            if item.text then
                ui:Text( tostring( item.text ))
            end
            ui:NextColumn()
        end
    end

    ui:Columns(1)

    if ui:Button( "Separator" ) then
        self:AppendLog( BATTLE_LOG.GENERAL, "---------------------"  )
    end
end
