------------------------------------------------------------------------
-- Helper functions for battle engine.

local battle_defs = require "battle/battle_defs"

class("BattleUtil")

local function ResolveBattleForAgent( battle, fighter, agent, winning_team, repercussions )
    local status = fighter:GetStatus()
    local loot, murders = repercussions.loot, repercussions.murders
    local self_defence = CheckBits( battle:GetScenario():GetFlags(), battle_defs.BATTLE_FLAGS.SELF_DEFENCE )
    local team = fighter:GetTeam()

    if winning_team == team:GetID() then
        -- Winning team!

    else
        -- Losing team!
        if status ~= FIGHT_STATUS.DEAD then
            -- Enemy NPCS who lost become 'intimidated'
            if not agent:IsRetired() and not agent:IsInPlayerParty() then
                if team ~= battle:GetPlayerTeam() and fighter.has_been_activated then
                    agent:GainAspect("intimidated")
                end
            end
        end

        local stat_id = string.format( "defeated_%s", agent:GetContentID() )
        TheGame:AddGameplayStat( stat_id, 1 )

        if CheckBits(battle:GetFlags(), BATTLE_FLAGS.BOSS_FIGHT) then
            if agent.boss then
                TheGame:GetGameProfile():SetBossAdvancement( agent:GetContentID(), TheGame:GetGameState():GetAdvancementLevel())
            end
        end
    end

    -- Everyone has persistent health loss.
    if not agent:IsRetired() then
        if agent:IsPlayer() then
            agent.health:Set(fighter:GetStat(COMBAT_STAT.HEALTH))
        else
            agent.health:SetPercent(fighter:GetStatPercent( COMBAT_STAT.HEALTH ))
        end
    end

    if agent.fight_data.OnResolveBattle then
        agent.fight_data.OnResolveBattle( fighter, repercussions )
    end

    if fighter:GetTeam() ~= battle:GetPlayerTeam() and battle:GetBattleResult() == BATTLE_RESULT.WON then
        if (status == FIGHT_STATUS.DEAD or CheckBits( battle:GetFlags(), BATTLE_FLAGS.REWARD_DEATH_LOOT )) and not CheckBits( battle:GetFlags(), BATTLE_FLAGS.NO_DEATH_LOOT) and not fighter.no_death_loot then
            AgentUtil.CollectDeathLoot(agent, loot)
        end
    end    


    if status == FIGHT_STATUS.DEAD then
        if not CheckBits( battle:GetFlags(), BATTLE_FLAGS.NO_DEATH ) and not agent:IsPlayer() then
            
            if not CheckBits( battle:GetFlags(), BATTLE_FLAGS.SELF_DEFENCE ) then
                local location = battle:GetScenario():GetLocation()
                if location and location:GetProprietor() and not CheckBits( location:GetFlags(), LOCFLAGS.HOME ) and not AgentUtil.HasPlotArmour(location:GetProprietor()) then
                    if battle:GetScenario():GetTeamForAgent(location:GetProprietor()) ~= battle:GetPlayerTeam() then
                        location:GetProprietor():OpinionEvent(OPINION.DID_VIOLENCE_IN_SHOP)
                    end
                end
            end

            if agent:IsInPlayerParty() and winning_team ~= team:GetID() then
                agent:Dismiss( true )
            end

            
            local killer = fighter:GetKiller()
            if killer and killer:IsPlayer() and not CheckBits( battle:GetFlags(), BATTLE_FLAGS.ISOLATED ) then
                if agent:IsSentient() then
                    TheGame:GetGameProfile():SetKilledAgent( agent:GetUniqueID() )
                    table.insert(murders, agent)
                end
            end

            if not agent:IsRetired() then
                agent:Kill()
                
            end
        end

    elseif CheckBits( battle:GetScenario():GetFlags(), battle_defs.BATTLE_FLAGS.SEND_TO_LIMBO ) then
        if not agent:IsInPlayerParty() then
            agent:MoveToLimbo()
        end
    end

    if status == FIGHT_STATUS.FLED and not agent:IsSentient() and not agent:IsRetired() then
        if agent:IsInPlayerParty() and agent:IsPet() and agent:GetSpecies() == SPECIES.BEAST then
            QuestUtil.RunAwayPet(agent)
        else
            agent:MoveToLimbo()
            agent:Retire()
        end
    end

    if agent:IsPlayer() then
        for i,graft in ipairs(agent.graft_owner:GetGrafts(GRAFT_TYPE.COMBAT)) do
            if graft:HasUpgrades() and not graft:GetDef().manual_xp and not graft:UpgradeReady() then
                graft:AddXP(1)
            end
        end
    end
end


function BattleUtil.ResolveBattle( battle, primary_enemy )
    
    local repercussions = {
        murders = {},        
        forced_cards = {},
        loot = {    
            money = 0,
            items = {},
            grafts = {},
        },
    }

    local scenario = battle:GetScenario()
    scenario:ApplyRepercussions( repercussions )

    local team = battle:GetTeam(scenario:GetEnemyTeam())
   
    --apply post-fight results 
    local winning_team
    if battle:GetBattleResult() == BATTLE_RESULT.WON then
        winning_team = battle:GetPlayerTeam():GetID()
        table.arrayadd( repercussions.loot.items, battle:GetLootItems() )
        repercussions.loot.money = repercussions.loot.money + battle:GetLootMoney()

        if battle:GetTurns() <= 1 then
            TheGame:GetGameState():AddScore("one_turn_battle_win", 1)
        elseif battle:GetTurns() <= 4 then
            TheGame:GetGameState():AddScore("quick_battle_win", 1)
        end

        local cards_played, item_cards_played, attack_cards_played = 0, 0, 0
        for i, card in ipairs(battle:GetAllCardsPlayed()) do
            if card:IsFlagged( battle_defs.CARD_FLAGS.ITEM ) then
                item_cards_played = item_cards_played + 1
            end
            if card:IsAttackCard() then
                attack_cards_played = attack_cards_played + 1
            end
            if not card:IsFlagged( battle_defs.CARD_FLAGS.SPECIAL ) then
                cards_played = cards_played + 1
            end
        end
        if item_cards_played > 0 and item_cards_played == cards_played then
            TheGame:UnlockAchievement( "machinist" )
        end
        if CheckBits(battle:GetFlags(), BATTLE_FLAGS.BOSS_FIGHT) then
            if attack_cards_played == 0 then
                TheGame:UnlockAchievement( "counterplay" )
            end
        end
    else
        winning_team = battle:GetEnemyTeam():GetID()
    end

    for i, fighter in battle:AllFighters() do
        if fighter.agent then
            ResolveBattleForAgent( battle, fighter, fighter.agent, winning_team, repercussions )
        end
    end

    TheGame:BroadcastEvent( "resolve_battle", battle, primary_enemy, repercussions )
    
    -- Handle murders.

    if next (repercussions.murders) then
        local revengers = {}
        for k, agent in pairs(repercussions.murders) do
            local revenger
            if math.random() < 0.5 then
                revenger = AgentUtil.CommitMurder( agent, revengers )
            end

            if revenger then
                revengers[revenger] = true
            else
                local cards = NegotiationCardCollection(function(def) return def.murder_card == true end):Pick(1)
                for k, v in ipairs(cards) do
                    table.insert( repercussions.forced_cards, v.id )
                end
            end
        end
    end

    local self_defence = CheckBits( scenario:GetFlags(), BATTLE_FLAGS.SELF_DEFENCE )

    if not self_defence then
        -- if primary_enemy and not primary_enemy:IsDead() then
        --     primary_enemy:OpinionEvent(OPINION.ATTACKED)
        -- end

        if scenario:IsAggressive() then
            table.insert( repercussions.forced_cards, "aggressive" )
        end

        local location = battle:GetScenario():GetLocation()
        if location then
            LocationUtil.Kerfuffle(location)
            location:OnDoViolence(primary_enemy)
        end
    end


    return repercussions
    
end


function BattleUtil.ResolveBattleAutomatically( battle, battle_policy )
    local team_blue = battle:GetTeam(TEAM.BLUE)
    for k, v in team_blue:AllFighters() do
        if battle_policy ~= "SPARE" and math.random() < .25 then
            v:SetStatus(FIGHT_STATUS.DEAD)
        end
    end

    local can_surrender = not CheckBits(battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER)
    local team_red = battle:GetTeam(TEAM.RED)
    for k, v in team_red:AllFighters() do
        local death_chance = can_surrender and .5 or 1
        if battle_policy == "KILL" then
            death_chance = 1
        elseif battle_policy == "SPARE" and can_surrender then
            death_chance = 0
        end

        if math.random() < death_chance then
            v:SetStatus(FIGHT_STATUS.DEAD)
        end
    end

    battle:Win()

    return BATTLE_RESULT.WON
end

