---------------------------------------------------------
-- Class handling combat conditions.

local CombatCondition = class("CombatCondition")

function CombatCondition:init(id, owner, stacks)
    local def = Content.GetBattleCondition( id )
    assert( def, id )
    self.feature_def = Content.GetBattleCardFeature( id )
    assert( self.feature_def, id )

	self.id = id
    self.def = def

    for k,v in pairs(def) do
        self[k] = v
    end

    self.owner = owner
    self.stacks = stacks
    assert( self.max_stacks )
end

function CombatCondition:GetID()
    return self.id
end

function CombatCondition:GetBattle()
    return self.owner and self.owner.battle
end

function CombatCondition:GetOwner()
    return self.owner
end

function CombatCondition:GetOwnerName()
    return self.owner and self.owner:GetName() or LOC "UI.CARDS.OWNER"
end

function CombatCondition:GetIdleOverride()
    return self.idle_override_anim
end

function CombatCondition:AddApplicant( source )
    if self.applied_by == nil then
        self.applied_by = {}
    end

    -- If this applicant already exists, remove and re-add at the end.
    -- Applicants are sorted in order of oldest -> most recent.
    table.arrayremove( self.applied_by, source )
    table.insert( self.applied_by, source )
end

function CombatCondition:RemoveApplicant( source )
    if self.applied_by then
        table.arrayremove( self.applied_by, source )
    end
end


function CombatCondition:GetApplicants()
    return self.applied_by or table.empty
end

function CombatCondition:GetRecentApplicant()
    if self.applied_by then
        return self.applied_by[ #self.applied_by ]
    end
end

function CombatCondition:GetRecentFighterApplicant()
    local applicant = self:GetRecentApplicant()
    if applicant and not is_instance( applicant, Fighter ) then
        applicant = applicant.owner
    end
    return applicant
end

function CombatCondition:GetStacks()
    return self.stacks, self.max_stacks
end

function CombatCondition:ApplyStacks(num)
    local old_stacks = self.stacks
    self:SetStacks( self.stacks + num )
    return self.stacks ~= old_stacks
end

function CombatCondition:SetStacks(num)
    local old_stacks = self.stacks

    self.stacks = clamp(num, self.min_stacks or 0, self.max_stacks)

    if self.OnSetStacks then
        self:OnSetStacks( num, old_stacks, self.stacks )
    end
end

function CombatCondition:GetIcon()
    return self.icon
end

function CombatCondition:GetName()
	local key = self.feature_def:GetLocNameKey()
	local name = LOC(key)
    return name
end

function CombatCondition:ApplyCondition( battle )
    if self.event_handlers then
        local battle_defs = require "battle/battle_defs"
        for event_name, fn in pairs( self.event_handlers ) do
            local priority = self.event_priorities and self.event_priorities[ event_name ]
            battle:ListenForEvent( event_name, self, self.OnBattleEvent, priority )
        end
    end

    self.battle = battle

    local init_stacks = self.stacks
    self.stacks = 0 -- To ensure that the delta within SetStacks starts from 0.

    self:SetStacks( init_stacks )

    if self.OnApply then
        self:OnApply( battle )
    end
end

function CombatCondition:OnBattleEvent( eventname, ... )
    if self.event_handlers and self.event_handlers[ eventname ] then
        self.event_handlers[ eventname ]( self, ... )
    end
end

function CombatCondition:UnapplyCondition( battle )
    if self.OnUnapply then
        self:OnUnapply( battle )
    end

    if self.event_handlers then
        battle:RemoveListener( self )
    end
end

function CombatCondition:CalculateDefendText( defend, target )
    if self.battle then
        local modified_defend = self.battle:CalculateModifiedStacks( "DEFEND", defend, target or self.target, self )
        --modified_defend = math.max(modified_defend, 0)
        if modified_defend < defend then
            return string.format( "<#PENALTY_CARD_TEXT>%d</>", modified_defend )
        elseif modified_defend > defend then
            return string.format( "<#BONUS_CARD_TEXT>%d</>", modified_defend )       
        end
    end

    return tostring(defend)
end

function CombatCondition:GetLocalizedString( id )
    return self.feature_def:GetLocalizedString( id )
end

function CombatCondition:GetDesc( battle, widget, features )
	local battle_defs = require "battle/battle_defs"

	local key = self.feature_def:GetLocDescKey()
	local desc = LOC(key)

    if self.desc_fn then
        desc = self.desc_fn( self, desc, battle or self.battle )
    end

    if desc then
        desc = CardEngine.ExtractKeywords( desc, widget, battle or self.battle, battle_defs, features )
    end
    return desc
end

function CombatCondition:__tostring()
    return string.format( "%s (x%d)", self.id, self.stacks )
end
