local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

---------------------------------------------------------------------
-- Used for calculating and tracking preview damage

local DamagePreview = class( "Battle.DamagePreview" )
DamagePreview.can_reload = true

function DamagePreview:init( battle )
    self.battle = battle

    -- Map of [Fighter] -> array of { card, min_damage, max_damage }
    self.targets = {}
end

function DamagePreview:Clear()
    table.clear( self.targets )
end

function DamagePreview:Targets()
    return pairs( self.targets )
end

function DamagePreview:GetPreviews( target )
    return self.targets[ target ] or table.empty
end

function DamagePreview:CountDamageHits( target )
    local count = 0
    if self.targets[ target ] then
        for i, v in ipairs( self.targets[ target ] ) do
            if v.damage or v.min_damage then
                count = count + 1
            end
        end
    end
    return count
end

function DamagePreview:IsPreviewingCard( target, card )
    local previews = self.targets[ target ]
    if previews then
        for i, preview in ipairs( previews ) do
            if preview.card == card then
                return true
            end
        end
    end

    return false
end

function DamagePreview:PreviewAttack( attack )
    -- Convert attack data into previews.
    local card = attack.card
    for i, hit in attack:Hits() do
        local previews = self.targets[ hit.target ]
        if previews == nil then
            previews = {}
            self.targets[ hit.target ] = previews
        end

        if hit.damage then
            if card.owner:IsPlayer() then
                -- Don't preview the rolled damage for the Player.
                -- TODO: this number doesn't even use a seeded random atm, so it changes whenever a new card is chosen.
                local preview = { card = card, min_damage = hit.min_damage, max_damage = hit.max_damage, hit_index = i, owner = card.owner, piercing = hit.piercing }
                table.insert( previews, preview )
            else
                local preview = { card = card, min_damage = hit.damage, max_damage = hit.damage, hit_index = i, owner = card.owner, piercing = hit.piercing }
                table.insert( previews, preview )
            end

        else
            local preview = { card = card, owner = card.owner }
            table.insert( previews, preview )
        end
    end
end

function DamagePreview:PreviewCard( card )
    local target_mod = self.battle:CalculateTargetMod( card )
    if card.owner:IsPlayer() and (target_mod == TARGET_MOD.RANDOM1 or target_mod == TARGET_MOD.RANDOM2 or target_mod == TARGET_MOD.RANDOM3) then
        -- Random targets shouldn't be previewed
        return
    end

    if card.PreviewFn then
        card:PreviewFn( self )
    else
        local attack = Battle.Attack( card )
        attack:GenerateHits()
        self:PreviewAttack( attack )
    end
end

function DamagePreview:PreviewCondition( condition )
    if condition.PreviewFn then
        condition:PreviewFn( self, self.targets )
    end
end

function DamagePreview:RemovePreview( card )
	for target, previews in pairs( self.targets ) do
		for i = #previews, 1, -1 do
			local preview = previews[i]
			if preview.card == card then
				table.remove( previews, i )
			end
		end
        if #previews == 0 then
            self.targets[ target ] = nil
        end
	end
end

function DamagePreview:AddPreview( preview )
    for target, previews in pairs( preview.targets ) do
        if self.targets[ target ] == nil then
            self.targets[ target ] = shallowcopy( previews )

        else
            for i, preview in ipairs( previews ) do
                assert( not self:IsPreviewingCard( target, preview.card ))
            end

            for i, preview in ipairs( previews ) do
                table.insert( self.targets[ target ], preview )
            end
        end
    end
end

function DamagePreview:RenderDebugUI( ui, panel )
    for target, previews in pairs( self.targets ) do
        panel:AppendTable( ui, target )
        ui:Indent( 20 )

        for i, preview in ipairs( previews ) do
            panel:AppendTable( ui, preview.card )
            
            if preview.min_damage then
                ui:SameLine( nil, 20 )
                ui:TextColored( 0xFF0000FF, string.format( "%d-%d", preview.min_damage, preview.max_damage ))
            end
        end

        ui:Unindent( 20 )
    end
end
