require "battle/fighter_behaviour"
local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

--------------------------------------------------------------


local Fighter = class("Fighter")
Fighter.can_reload = true


function Fighter.CreateFromAgent( agent, scale, player_controlled )
    scale = scale or 1
    
    assert( is_instance( agent, Agent ))
    assert( agent.fight_data )
    
    local fighter = Fighter( agent )
    local health, max_health = agent:GetHealth()

    fighter:CreateStat( COMBAT_STAT.HEALTH, health, 0, max_health )
    fighter:EnableMorale( true )

    fighter.rally_modifier = agent:GetRallyReaction()
    
    fighter.player_controlled = player_controlled
    fighter.can_surrender = agent:IsSentient()

    fighter.ranged_riposte = agent.fight_data.ranged_riposte
    fighter.riposte_tags = agent.fight_data.riposte_tags

    fighter.formation = agent.fight_data.formation or 0
    fighter.scale = scale
    fighter.fight_data = agent.fight_data
    
    if agent.fight_data.behaviour then
        fighter.behaviour = FighterBehaviour( fighter, agent.fight_data.behaviour )
    end

    if agent.battler and (fighter.behaviour == nil or player_controlled) then
        fighter.cards = {}
        for i, card in ipairs( agent.battler:GetCards() ) do
            local card = card:Clone()
            card.show_dealt = false
            card.owner = fighter
            table.insert( fighter.cards, card )
        end
    end

    if TheGame:GetGameState() and TheGame:GetGameState().available_quests then
        for k, quest in pairs(TheGame:GetGameState().available_quests) do
            quest:ProcessFighter(fighter)
        end
    end

    
    return fighter
end

function Fighter:init( agent )
    assert( is_instance( agent, Agent ))
    self.agent = agent

    self.conditions = {}
 
    self.stat_value = {} --The non-modified value of a stat.
    self.stat_bounds = {} --The minimum and maximum values of a stat, EVEN WITH MODIFIERS.

    self.status = FIGHT_STATUS.ACTIVE

    self.last_turn = -1
end

function Fighter:GetScale()
    return self.scale
end

function Fighter:SetAnimScale( anim_scale )
    self.anim_scale = anim_scale
end

function Fighter:GetAnimScale()
    if self.anim_scale then
        return self.anim_scale
    else
        return self.agent.fight_data.battle_scale or 1.0
    end
end


function Fighter:GetName()
    return self.agent:GetFullName()
end

function Fighter:GetHitTags()
    return self.agent.fight_data.hit_tags or table.empty
end

function Fighter:HasHitTag( tag )
    if self.agent:HasTag( tag ) then
        return true
    end
    if table.contains( self:GetHitTags(), tag ) then
        return true
    end
    return false
end

function Fighter:GetDiscardDeck()
    if self.behaviour then
        return self.behaviour:GetDeck()
    else
        return self.battle:GetDiscardDeck()
    end
end

function Fighter:GetHealthHidden()
    return self.health_hidden
end

function Fighter:SetHealthHidden(viz)
    self.health_hidden = viz
end


function Fighter:GetBattle()
    return self.battle
end


function Fighter:GetTeam()
    return self.team
end

function Fighter:GetEnemyTeam()
    return self.team:GetEnemyTeam()
end

function Fighter:GetTeamID()
    return self.team and self.team.id
end

function Fighter:GetStartTurn()
    return self.start_turn or 0
end

function Fighter:GetAgent()
    return self.agent
end

function Fighter:GetTargetWeight()
    return self.agent.fight_data.target_weight or 1
end

function Fighter:AppendLog( log_type, ... )
    if self.battle then
        self.battle:AppendLog( log_type, ... )
    end
end

function Fighter:PrintLog( log_type, ... )
    if self.battle then
        self.battle:PrintLog( log_type, ... )
    end
end

function Fighter:BroadcastEvent( event_name, ... )
    if self.battle then
        self.battle:BroadcastEvent( event_name, self, ... )
    end
end

function Fighter:SetStatus(status)
    
    if self.status ~= status then
        self.status = status
        self:PrintLog( BATTLE_LOG.GENERAL, "{1} status is now {2}!", self, status )
        --print (debugstack())
        self:BroadcastEvent( BATTLE_EVENT.STATUS_CHANGED, status )

        if self.team:GetID() == TEAM.RED and self.team:IsSurrendering() and self.battle:GetCurrentTeam() ~= self.team then
            self.battle:AddExecutes()
        end
    end
end

function Fighter:GetStatus()
    return self.status
end

function Fighter:IsDead()
    return self.status == FIGHT_STATUS.DEAD
end

function Fighter:IsAlive()
    return self:GetStatus() == FIGHT_STATUS.ACTIVE or self:GetStatus() == FIGHT_STATUS.SURRENDER
end

function Fighter:IsActive()
    return self:GetStatus() == FIGHT_STATUS.ACTIVE
end

function Fighter:GetKiller()
    return self.killer
end

function Fighter:CanMove( rank )
    for id, condition in pairs(self.conditions) do
        if condition:PreventsMove() then
            return false, "Can't move"
        end
    end

    return true
end

function Fighter:CanSurrender()
    return self.can_surrender
end

function Fighter:CreateStat(stat, current, min, max)
    assert( IsEnum( stat, COMBAT_STAT ))
    assert( self.stat_value[ stat ] == nil )
    assert( current )
    assert( min and max and min <= max )

    self.stat_bounds[ stat ] = {min = min, max = max}
    self.stat_value[ stat ] = current
end

function Fighter:DestroyStat(stat)
    self.stat_bounds[ stat ] = nil
    self.stat_value[ stat ] = nil
end

function Fighter:DeltaMaxStat(stat, delta_max)
    if self.stat_value[stat] == nil then
        return
    end

    local old_max = self.stat_bounds[ stat ].max
    local new_max = math.max( self.stat_bounds[ stat ].min, self.stat_bounds[ stat ].max + delta_max )

    self.stat_bounds[ stat ].max = new_max
    local delta = math.min( 0, new_max - self.stat_value[ stat ] )
    self.stat_value[ stat ] = math.min( self.stat_value[ stat ], new_max )

    self:BroadcastEvent( BATTLE_EVENT.DELTA_STAT, stat, delta, self:GetStat( stat ), 0,
        new_max, new_max - old_max )

    if stat == COMBAT_STAT.HEALTH then --or stat == COMBAT_STAT.MORALE then
        self:CheckForDeath()
    end
end

function Fighter:DeltaStat(stat, delta, mitigated)
    if self.stat_value[stat] == nil then
        return
    end

    self:AppendLog( BATTLE_LOG.STAT, self, stat, delta )

    self:SetStat( stat, self.stat_value[stat] + delta )

    self:BroadcastEvent( BATTLE_EVENT.DELTA_STAT, stat, delta, self:GetStat( stat ), mitigated or 0 )

    if stat == COMBAT_STAT.HEALTH then --or stat == COMBAT_STAT.MORALE then
        self:CheckForDeath()
    end
end

function Fighter:SetStat(stat, value)
    local bounds = self.stat_bounds[ stat ]
    if bounds == nil then
        return
    end
    self.stat_value[ stat ] = clamp( value, bounds.min, bounds.max )
end

function Fighter:GetStatBounds(stat)
    return self.stat_bounds[ stat ].min, self.stat_bounds[ stat ].max
end

function Fighter:GetRawStat( stat )
    return self.stat_value[ stat ]
end

function Fighter:GetStat( stat )
    return self.stat_value[stat]
end

function Fighter:GetStatPercent(stat)
    local min, max = self.stat_bounds[ stat ].min, self.stat_bounds[ stat ].max
    if max > 0 then
        return self.stat_value[ stat ] / max
    else
        return 1.0
    end
end

function Fighter:GetHealth()
    return self:GetStat( COMBAT_STAT.HEALTH )
end

function Fighter:GetMaxHealth()
    return self.stat_bounds[ COMBAT_STAT.HEALTH ].max
end

function Fighter:DeltaMaxHealth( delta )
    self:DeltaMaxStat( COMBAT_STAT.HEALTH, delta )
end

function Fighter:GetHealthPercent()
    return self:GetStatPercent( COMBAT_STAT.HEALTH )
end

function Fighter:SetMaxHealth( value )
    self.stat_bounds[ COMBAT_STAT.HEALTH ].max = value
end

function Fighter:SetHealthPercent(p)
    local min, max = self.stat_bounds[ COMBAT_STAT.HEALTH ].min, self.stat_bounds[ COMBAT_STAT.HEALTH ].max
    self:SetStat(COMBAT_STAT.HEALTH, math.max(1, math.round( min + (min+max)*p)))
end

function Fighter:DeltaHealth( delta, attacker, mitigated )
    self.last_attacker = attacker
    delta = math.floor( delta ) -- Health is always integral mang.
    self:DeltaStat( COMBAT_STAT.HEALTH, delta, mitigated )
end

function Fighter:HealHealth( delta, source )
    if delta > 0 then
        self:DeltaStat( COMBAT_STAT.HEALTH, delta )
        self:BroadcastEvent( BATTLE_EVENT.HEAL_HEALTH, delta, source )
    end
end

function Fighter:AlleviateHealth( delta, source )
    delta = math.min(delta, self:GetConditionStacks("RECOVERABLE_HEALTH"))
    if delta > 0 then
        self:DeltaStat( COMBAT_STAT.HEALTH, delta )
        self:BroadcastEvent( BATTLE_EVENT.HEAL_HEALTH, delta, source )
        self:RemoveCondition("RECOVERABLE_HEALTH", delta, self)
    end
end

function Fighter:HasMorale()
    return self:GetMorale() ~= nil
end

function Fighter:EnableMorale( enabled )
    if enabled and self:GetMorale() == nil then
        local morale_percent = self.agent:GetMorale()
        if morale_percent then
            morale_percent = morale_percent * self:GetHealthPercent()
            self:CreateStat(COMBAT_STAT.MORALE, 0, 0, 1)
            self:DeltaStat(COMBAT_STAT.MORALE, morale_percent )
        end

    elseif not enabled and self:GetMorale() then
        self:DestroyStat( COMBAT_STAT.MORALE )
    end
end

function Fighter:GetMorale()
    return self:GetStat( COMBAT_STAT.MORALE )
end

function Fighter:GetSurrenderHealth()
    local health_max = self:GetMaxHealth()
    local morale = self:GetMorale() or 0
    return math.ceil( health_max * morale )
end

function Fighter:ApplyMoraleDamage( delta, source )
    local min, max = self:GetStatBounds( COMBAT_STAT.HEALTH )
    self:DeltaMorale( delta / max )
end

function Fighter:DeltaMorale( delta )
    if self.status == FIGHT_STATUS.ACTIVE or delta > 0 then
        self:DeltaStat( COMBAT_STAT.MORALE, delta )

        -- Don't immediately check for surrender. Like health deltas, wait until a
        -- CheckForSurrender is called at the end of card resolve so that fighters
        -- don't flee until AFTER the attack is over.
        -- self:CheckForSurrender()
    end    
end

function Fighter:CheckForDeath()
    local status = self:GetStatus()
    
    if status ~= FIGHT_STATUS.DEAD and status ~= FIGHT_STATUS.FLED then
        if self:GetStat( COMBAT_STAT.HEALTH ) <= 0 then
            self:Kill()
        end
    end

    return self:GetStatus() == FIGHT_STATUS.DEAD
end

function Fighter:CheckForSurrender()
    local status = self:GetStatus()
    if not self.player_controlled and self:HasMorale() then
        if status == FIGHT_STATUS.ACTIVE and not self:HasCondition("STUN") and self:GetHealth() <= self:GetSurrenderHealth() then
            if self.can_surrender then
                self:Surrender()
            else
                self:Flee()
            end
        end
    end
end

function Fighter:Rally()
    if self.battle:HasExecutes() then
        return
    end

    local MAX_MORALE_PERCENT = 0.25

    assert( self.status == FIGHT_STATUS.SURRENDER )
    self:RemoveCondition( "SURRENDER" )
    self:AddCondition(self.rally_modifier)
    self:SetStatus(FIGHT_STATUS.ACTIVE)

    self:DeltaStat(COMBAT_STAT.MORALE, self:GetStatPercent(COMBAT_STAT.HEALTH) * MAX_MORALE_PERCENT)

    for i, fighter in self.team:Fighters() do
        if fighter ~= self then
            fighter:OnAllyRally( self )
        end
    end
    self:RunBehaviour( "OnRally" )

    
    self.battle:BroadcastEvent( BATTLE_EVENT.FIGHTER_RALLY, self)
end

function Fighter:Surrender()
    assert( self.status == FIGHT_STATUS.ACTIVE )


    if self:HasMorale() then
        self:AddCondition( "SURRENDER", SURRENDER_DURATION )
    end

    self:SetStatus(FIGHT_STATUS.SURRENDER)
    local allies = shallowcopy(self.team:GetFighters())
    local enemies = shallowcopy(self.team:GetEnemyTeam():GetFighters())
    
    for i, fighter in ipairs(allies) do
        if fighter ~= self then
            fighter:OnAllySurrender( self )
        end
    end

    for i, fighter in ipairs(enemies) do
        fighter:OnEnemySurrender( self )
    end

    local morale = self:GetMorale()
    if morale then
        self:DeltaStat( COMBAT_STAT.MORALE, -morale )
    end

    self.battle:BroadcastEvent( BATTLE_EVENT.FIGHTER_SURRENDER, self)
    self.team:LookForReplacement(self)
end

function Fighter:Kill( killer )
    
    assert( self:IsAlive() )
    self.killer = killer or self.last_attacker
    local battle = self.battle
    
    self:SetStatus(FIGHT_STATUS.DEAD)
    self.team:RemoveFromField( self )

    local allies = shallowcopy(self.team:GetFighters())
    local enemies = shallowcopy(self.team:GetEnemyTeam():GetFighters())
    for i, fighter in ipairs(allies) do
        if fighter ~= self and fighter:IsAlive() then
            fighter:OnAllyKill( self )
        end
    end

    for i, fighter in ipairs(enemies) do
        if fighter:IsAlive() then
            fighter:OnEnemyKill( self )
        end
    end
    battle:BroadcastEvent( BATTLE_EVENT.FIGHTER_KILLED, self)
end

function Fighter:Flee()
    assert( self:IsAlive() )
    self:SetStatus(FIGHT_STATUS.FLED)
    self.team:RemoveFromField( self )
    local allies = shallowcopy(self.team:GetFighters())
    local enemies = shallowcopy(self.team:GetEnemyTeam():GetFighters())

    for i, fighter in ipairs(allies) do
        if fighter ~= self then
            fighter:OnAllyFlee( self )
        end
    end

    for i, fighter in ipairs(enemies) do
        fighter:OnEnemyFlee( self )
    end
end

function Fighter:Depart()
    self:SetStatus(FIGHT_STATUS.DEPARTED)

    -- This may be used for "called-in" fighters, who aren't legitimately on the Team.
    if table.contains( self.team:GetFighters(), self ) then
        self.team:RemoveFromField( self )
    else
        self:OnDeactivateFighter()
    end
end

function Fighter:WillProtectOthers( )
    return self:HasCondition("PROTECT")
end

function Fighter:CanBeTargetted( card )
    assert( card )
    if not self.team then return false end
    for i, fighter in self.team:Fighters() do
        for id, condition in pairs( fighter.conditions ) do
            if condition.CanBeTargetted then
                local ok, reason = condition:CanBeTargetted( card, self )
                assert_warning( ok ~= nil, "You really should return true or false explicitly" )
                if not ok then
                    return false, reason
                end
            end
        end
    end
    for i, fighter in self:GetEnemyTeam():Fighters() do
        for id, condition in pairs( fighter.conditions ) do
            if condition.CanBeTargetted then
                local ok, reason = condition:CanBeTargetted( card, self )
                assert_warning( ok ~= nil, "You really should return true or false explicitly" )
                if not ok then
                    return false, reason
                end
            end
        end
    end

    return true
end

function Fighter:IsTargetting( target )
    if self.prepared_attacks then
        for i, attack in ipairs( self.prepared_attacks ) do
            if attack:IsTarget( target ) then
                return true
            end
        end
    end
    return false
end

function Fighter:CanTarget(card, target_fighter)

    -- Validate primary target first
    if target_fighter:GetStatus() == FIGHT_STATUS.DEAD then
        return false, CARD_PLAY_REASONS.TARGET_DEAD

    elseif target_fighter:GetStatus() == FIGHT_STATUS.FLED then
        return false, CARD_PLAY_REASONS.TARGET_FLED

    elseif target_fighter:GetStatus() == FIGHT_STATUS.SURRENDER and
            not self.player_controlled and
            card:IsAttackCard() and
            not target_fighter:HasCondition( "TARGETED" ) then
        return false, CARD_PLAY_REASONS.TARGET_SURRENDERED
    end

    if card.target_mod == TARGET_MOD.ALL then
        -- Anyone is valid.
    elseif card.target_type == TARGET_TYPE.SELF then
        if card.target_mod == TARGET_MOD.TEAM and self.team ~= target_fighter.team then
            return false, CARD_PLAY_REASONS.TARGET_ALLY_OR_SELF
        elseif card.target_mod ~= TARGET_MOD.TEAM and self ~= target_fighter then
            return false, CARD_PLAY_REASONS.TARGET_SELF
        end
    elseif card.target_type == TARGET_TYPE.ANY_NOT_SELF then
        if target_fighter == self then
            return false, CARD_PLAY_REASONS.TARGET_ANY_NOT_SELF
        end
    elseif card.target_type == TARGET_TYPE.FRIENDLY_OR_SELF then
        if self.team ~= target_fighter.team then
            return false, CARD_PLAY_REASONS.TARGET_ALLY_OR_SELF
        end
    elseif card.target_type == TARGET_TYPE.FRIENDLY then
        if self.team ~= target_fighter.team or target_fighter == self then
            return false, CARD_PLAY_REASONS.TARGET_ALLY
        end
    elseif card.target_type == TARGET_TYPE.ENEMY then
        if self.team == target_fighter.team then
            return false, CARD_PLAY_REASONS.TARGET_ENEMY
       end
    end

    local ok, reason = target_fighter:CanBeTargetted( card )
    if not ok then
        return false, reason
    end

    if card.CanPlayCard then
        local ok, reason = card:CanPlayCard( self.battle, target_fighter )
        if not ok then
            return false, reason
        end
    end

    local targets = self.battle:CollectTargets( card, target_fighter )
    if #targets == 0 then
        return false, "NO VALID TARGETS"
    end

    return true
end

function Fighter:RemoveBuff( source )
    local conditions = ObtainWorkTable()

    for j, condition in pairs( self.conditions ) do
        if condition.ctype == CTYPE.BUFF and not condition.hidden then
            table.insert( conditions, condition )
        end
    end

    local condition = table.arraypick( conditions )
    if condition then
        self:RemoveCondition( condition.id, nil, source )
    end

    ReleaseWorkTable( conditions )

    return condition ~= nil
end

function Fighter:RemoveDebuff( source )
    local conditions = ObtainWorkTable()

    for j, condition in pairs( self.conditions ) do
        if condition.ctype == CTYPE.DEBUFF and not condition.hidden then
            table.insert( conditions, condition )
        end
    end

    local condition = table.arraypick( conditions )
    if condition then
        self:RemoveCondition( condition.id, condition.stacks, source )
    end

    ReleaseWorkTable( conditions )

    return condition ~= nil
end

function Fighter:RemoveAllDebuffs( source )
    local debuffs = ObtainWorkTable()
    for i, condition in pairs( self.conditions ) do
        if condition.ctype == CTYPE.DEBUFF then
            table.insert( debuffs, condition )
        end
    end

    for i, condition in ipairs(debuffs) do
        self:RemoveCondition( condition.id, condition.stacks, source )
    end
    local count = #debuffs
    ReleaseWorkTable( debuffs )

    return count
end

function Fighter:GetAllDebuffs()
    local debuffs = ObtainWorkTable()
    for i, condition in pairs( self.conditions ) do
        if condition.ctype == CTYPE.DEBUFF then
            table.insert( debuffs, condition )
        end
    end

    ReleaseWorkTable( debuffs )

    return debuffs
end

function Fighter:AddCondition( id, stacks, source )
    self:DeltaCondition( id, stacks or 1, source )
    return self.conditions[ id ]
end

function Fighter:RemoveCondition( id, stacks, source )
    if stacks == nil then
        local current_condition = self.conditions[ id ]
        if current_condition then
            self:DeltaCondition( id, -current_condition.stacks, source )
        end
    else
        self:DeltaCondition( id, -stacks, source )
    end
end

function Fighter:DeltaCondition( id, stacks, source )
    assert( id )
    assert( stacks )
    assert_warning( source == nil or is_instance( source, Battle.Card ) or is_instance( source, Battle.CombatCondition ))


    if self.battle == nil then
        -- Must queue inactive fighter conditions so they get applied when joining battle.
        if self.queued_conditions == nil then
            self.queued_conditions = {}
        end
        table.insert( self.queued_conditions, { id = id, stacks = stacks, source = source } )
        return
    end

    stacks = self.battle:CalculateModifiedStacks( id, stacks, self, source )
    if stacks == 0 then
        return
    end

    local current_condition = self.conditions[ id ]
    local old_stacks = current_condition and current_condition:GetStacks() or 0

    if current_condition then
        current_condition:ApplyStacks( stacks )
       
        if current_condition:GetStacks() == 0 then
            current_condition:UnapplyCondition( self.battle )
            self.conditions[ id ] = nil
        end

    elseif stacks > 0 then
        current_condition = CombatCondition( id, self, stacks )
        if current_condition.CanApply and not current_condition:CanApply( self, source ) then
            self:BroadcastEvent( BATTLE_EVENT.CONDITION_BLOCKED, current_condition)
            return
        end

        self.conditions[ id ] = current_condition
        current_condition:ApplyCondition( self.battle )

 
    elseif stacks < 0 then
        -- Applying negative stacks can, strangely, CREATE a condition if it permits negative stacks.
        local def = battle_defs.CONDITIONS[ id ]
        if def.min_stacks and def.min_stacks < 0 then
            current_condition = CombatCondition( id, self, stacks )

            self.conditions[ id ] = current_condition
            current_condition:ApplyCondition( self.battle )
        end
    end

    if current_condition then
        local delta = current_condition:GetStacks() - old_stacks
        if delta > 0 then
            self:AppendLog( BATTLE_LOG.GENERAL, "{1} added {2} stacks of {3} (now {4})", self, delta, id, current_condition.stacks )
            self:BroadcastEvent( BATTLE_EVENT.CONDITION_ADDED, current_condition, delta, source )
            if source then
                current_condition:AddApplicant( source )
            end

        elseif delta < 0 then
            self:AppendLog( BATTLE_LOG.GENERAL, "{1} removed {2} stacks of {3} (now {4})", self, -delta, id, current_condition.stacks )
            self:BroadcastEvent( BATTLE_EVENT.CONDITION_REMOVED, current_condition, -delta, source )
        end
    end

end

function Fighter:GetConditions()
    return self.conditions
end

function Fighter:GetCondition( id )
    return self.conditions[ id ]
end

function Fighter:GetConditionStacks( id )
    local condition = self.conditions[ id ]
    return condition and condition.stacks or 0
end

function Fighter:PickRandomCondition( fn )
    local picks = ObtainWorkTable()
    for id, condition in pairs( self.conditions ) do
        if fn( condition ) then
            table.insert( picks, condition )
        end
    end
    local condition = table.arraypick( picks )
    ReleaseWorkTable( picks )
    return condition
end

function Fighter:HasCondition(condition)
    return self.conditions[condition] ~= nil
end

function Fighter:ConsumeCombo( count )
    count = count or self:GetConditionStacks( "COMBO" )
    self:RemoveCondition( "COMBO", count)
    return count
end

-- NOTE: source may be nil, a Card, a Graft.Instance, or a Battle.Hit.
function Fighter:ApplyDamage( damage, attacker, source, piercing, hit_tags )

    -- Feh. This is so that any newly applied conditions within DeltaHealth don't get
    -- an OnPostDamage handler for this application.
    local initial_conditions = shallowcopy( self.conditions )
    local delta = damage

    local conditions = self:SortConditionTriggers( "OnPreDamage", initial_conditions )
    for i, condition in ipairs( conditions ) do
        delta = condition:OnPreDamage( delta, attacker, self.battle, source, piercing ) or delta
    end

    if delta > 0 then
        self:BroadcastEvent( BATTLE_EVENT.DAMAGE_APPLIED, damage, delta, source, hit_tags )
    end
    local source_fighter = source and self.battle and self.battle:GetFighterFromSource( source )
    if source_fighter and source_fighter:IsPlayer() then
        TheGame:AddGameplayStat( "damage_dealt", delta )
    end

    self:DeltaHealth( -delta, attacker, damage - delta )
    
    local conditions = self:SortConditionTriggers( "OnPostDamage", initial_conditions )
    for i, condition in ipairs( conditions ) do
        if self.conditions[ condition.id ] then
            condition:OnPostDamage( delta, attacker, self.battle, source )
        end
    end
    if self.fight_data.can_alleviate and self.battle and self.battle:GetCurrentTeam() == self:GetTeam() then
        self:AddCondition("RECOVERABLE_HEALTH", delta, self)
    end
    return delta
end

function Fighter:SelfDamage( damage, source, piercing )
    self:ApplyDamage( damage, self, source, piercing, { "no_reaction" } )
end

function Fighter:OnActivateFighter( battle )
    assert( self.battle == nil )

    self.battle = battle

    self.morale_accumulator = CardEngine.ScalarAccumulator( battle, BATTLE_EVENT.CALC_MORALE )

    battle:OnActivateFighter( self )

    if self.cards then
        battle:DealCards( self.cards )
        self.cards = nil
    end

    if self.queued_conditions then
        while #self.queued_conditions > 0 do
            local t = table.remove( self.queued_conditions )
            self:AddCondition( t.id, t.stacks, t.source )
        end
        self.queued_conditions = nil
    end

    for i, id, condition_def in sorted_pairs(battle_defs.CONDITIONS) do
        if condition_def.ShouldAutoAttach then
            local stacks = condition_def:ShouldAutoAttach(battle, self)
            if (stacks or 0) > 0 then
                self:AddCondition(condition_def.id, stacks)
            end
        end
    end

    if self.agent:GetAspect("graft_owner") then
        self.agent:GetAspect("graft_owner"):OnActivateFighter(self)
    end    

    if self.behaviour and not self.call_in then
        self.behaviour:OnActivateFighter( battle )
        self:PrepareTurn()
    end

    self:CheckForDeath()
    if self:GetTeam() == battle:GetEnemyTeam() and self:GetMorale() then
        -- If this enemy has morale and it's a no_surrender fight, turn morale off
        self:EnableMorale( not CheckBits(battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER))
    end

    --meet the backing agent (useful for agents that spawn as parts of tag-teams)
    if TheGame:GetGameProfile() then
        TheGame:GetGameProfile():SetMetCharacter( self.agent:GetUniqueID() )    
    end
end

function Fighter:OnDeactivateFighter()
    assert( self.battle ~= nil )

    self.prepared_attacks = nil

    if self.behaviour then
        self.behaviour:OnDeactivateFighter( self.battle )
    end

    -- Conditions may do things when removed, like add other conditions, so this ensures all possible conditions are cleared.
    while next(self.conditions) ~= nil do
        local found = false
        for id, condition in pairs(self.conditions) do
            if not condition.persist_on_death then
                self:RemoveCondition( id )
                found = true
                break
            end
        end
        if not found then
            break
        end
    end

    self.battle:OnDeactivateFighter( self )

    self.battle = nil
end

function Fighter:NotifyPreviewChanged()
    self:BroadcastEvent( BATTLE_EVENT.FIGHTER_PREVIEW )
end

local function SortConditionByPriority( c1, c2 )
    -- TODO: not going to second guess how complicated this priority mechanism is going to be.
    -- Right now, it's optional and per condition, not per function.
    return (c1.priority or math.huge) < (c2.priority or math.huge)
end

function Fighter:SortConditionTriggers( fn, t )
    local conditions = {}
    for id, condition in pairs( t or self.conditions ) do
        if type(condition[ fn ]) == "function" then
            table.insert( conditions, condition )
        end
    end

    table.sort( conditions, SortConditionByPriority )

    return conditions
end

function Fighter:ClearPreparedCards()
    self.prepared_attacks = nil
    self:NotifyPreviewChanged()
end

function Fighter:PrepareTurn()
    if self:IsActive() and self.battle:GetTurns() > self.last_turn and (self.battle:GetTurns() > self.start_turn or self:GetTeamID() == TEAM.BLUE) then
        local attacks = self.prepared_attacks
        local invalid_attacks = {}
        local notify = false
        if attacks and #attacks > 0 then
            for k, attack in pairs(attacks) do
                if not attack.card:GetTarget() or not self.battle:CanPlayCard( attack.card, attack.card:GetTarget() ) then
                    table.insert( invalid_attacks, attack )

                elseif self.battle:IsPlayerTurn() then
                    -- Recalculate the attack.
                    attacks[k] = Battle.Attack( attack.card )
                    attacks[k]:GenerateHits()
                    notify = true
                else
                    -- This attack is locked in, so no attack is regenerated.
                end
            end
            for i=1,#invalid_attacks do
                table.arrayremove( attacks, invalid_attacks[i])
            end
        end

        -- Prepare a card
        if attacks == nil or (attacks and #attacks == 0) then
            self.prepared_attacks = {}
            local cards
            if self.battle:IsPlayerTurn() or #invalid_attacks == 0 then
                cards = self:RunBehaviour()
            end
            if not cards or #cards == 0 then
                if self.behaviour then
                    cards = self.behaviour:ChooseDefaultCard()
                end
            end

            for i, card in ipairs( cards or table.empty ) do
                local attack = Battle.Attack( card )
                table.insert( self.prepared_attacks, attack )
                attack:GenerateHits() -- AFTER adding to prepared_attacks.
            end

            notify = true
        end

        if notify then
            self:NotifyPreviewChanged()
        end

    elseif self.prepared_attacks then
        self:ClearPreparedCards()
    end

    return self.prepared_attacks
end

function Fighter:OnBeginTurn()
    self:AppendLog( BATTLE_LOG.PHASE, self, "OnBeginTurn" )

    self:BroadcastEvent( BATTLE_EVENT.BEGIN_TURN )
    -- self.recoverable_health = 0
end

function Fighter:OnEndTurn()
    self:AppendLog( BATTLE_LOG.PHASE, self, "OnEndTurn" )

    -- Only enemy AI have any self-volition
    self:PlayBehaviour()

    self:BroadcastEvent( BATTLE_EVENT.END_TURN )
end

function Fighter:PlayBehaviour()
    local attacks = self:PrepareTurn()
    self.prepared_attacks = nil
    self.last_turn = self.battle:GetTurns()

    if attacks then
        for k, attack in pairs(attacks) do
            assert( attack.target, tostring(attack) )
            self.battle:PlayCard( attack.card, nil, attack )
            if self.battle == nil then
                break
            end
        end 
    end
end

function Fighter:OnAllyFlee( ally )
    self:RunBehaviour( "OnAllyFlee", ally )
end

function Fighter:OnAllySurrender( ally )
    self:RunBehaviour( "OnAllySurrender", ally )
end

function Fighter:OnAllyKill( ally )
    self:RunBehaviour( "OnAllyKill", ally )
end

function Fighter:OnAllyRally( ally )
    self:RunBehaviour( "OnAllyRally", ally )
end


function Fighter:IsLastStanding()
    local team = self:GetTeam()
    if #team.fighters == 1 then
        return false
    end
    for i,fighter in ipairs(team.fighters) do
        if fighter:GetStatus() == FIGHT_STATUS.ACTIVE and fighter ~= self then
            return false
        end
    end
    return true
end

function Fighter:IsOnlyActiveFighterOnTeam()
    if self:GetStatus() ~= FIGHT_STATUS.ACTIVE then
        return false
    end

    local team = self:GetTeam()
    for i,fighter in ipairs(team.fighters) do
        if fighter:GetStatus() == FIGHT_STATUS.ACTIVE and fighter ~= self then
            return false
        end
    end
    return true
end

function Fighter:OnEnemyFlee( enemy )
    -- May need to re-target due to surrendered target.
    self:PrepareTurn()
end
 
function Fighter:OnEnemySurrender( enemy )
    -- May need to re-target due to surrendered target.
    self:PrepareTurn()
end

function Fighter:OnEnemyKill( enemy )
    -- May need to re-target due to surrendered target.
    self:PrepareTurn()
end

function Fighter:GetIdleState()
    local priority = -10
    local idle_state
    if self.status == FIGHT_STATUS.SURRENDER then
        idle_state = IDLE_STATE.SURRENDER
    elseif self.status == FIGHT_STATUS.DEAD then
        idle_state = IDLE_STATE.DEAD
    else
        idle_state = IDLE_STATE.NEUTRAL
    end

    for id,v in pairs(self.conditions) do
        if v.idle_state_override then
            if priority < (v.priority or 0) then
                idle_state = v.idle_state_override
            end
        end
    end
    return idle_state
end

function Fighter:HasBehaviour()
    return self.behaviour ~= nil
end

function Fighter:RunBehaviour( pattern_name, ... )
    if self.behaviour then
        return self.behaviour:RunBehaviour( self.battle, pattern_name, ... )
    end
end

function Fighter:Delay( delay )
    self.battle:Delay( delay )
end

function Fighter:SaySpeech( delay, txt )
    self:BroadcastEvent( BATTLE_EVENT.SPEECH, delay, txt )
end

function Fighter:SayQuip( quip, ... )
    self:BroadcastEvent( BATTLE_EVENT.QUIP, quip, ... )
end

function Fighter:PlayAnim( anim, wait, idle )
    self:BroadcastEvent( BATTLE_EVENT.PLAY_ANIM, anim, wait, idle )
end

function Fighter:Present( fn )
    self:BroadcastEvent( BATTLE_EVENT.CUSTOM, fn )
end


function Fighter:IsPlayer()
    return self.player_controlled
end

function Fighter:__tostring()
    return string.format( "[%s]", tostring(self.agent))
end

