require "cards/card_picker"
local battle_defs = require "battle/battle_defs"

---------------------------------------------------------------
-- Basic fighter: uses a weighted pick to choose a random card.

function FighterBehaviour_Basic( attacks )
    local behaviour_def =
    {
        OnActivate = function( self )
            self.cards = self:MakePicker()
            for id, wt in pairs( attacks ) do
                self.cards:AddID( id, wt )
            end
            self:SetPattern( self.Pattern )
        end,
        Pattern = function( self )
            self.cards:ChooseCard()
        end,
    }

    return behaviour_def
end

---------------------------------------------------------------------------
-- FighterBehaviour class. An instance is created for AI controlled fighters.

local FighterBehaviour = class("FighterBehaviour")
FighterBehaviour.can_reload = true

FighterBehaviour.Basic = FighterBehaviour_Basic

function FighterBehaviour:init(fighter, data)
    self.fighter = fighter

    self.prepared_cards = {}

    for k, v in pairs(data) do
        assert( rawget( self, k ) == nil, k )
        self[k] = v
    end
end

function FighterBehaviour:GetDeck()
    return self.hand
end

function FighterBehaviour:MakePicker()
    local picker = CardEngine.CardPicker( self, self.engine )
    return picker
end

function FighterBehaviour:ChooseDefaultCard()
    table.clear( self.prepared_cards )

    if self.default_card == nil then
        self.default_card = self:AddCard( "ai_defend_lo" )
    end

    self:ChooseCard( self.default_card )

    return self.prepared_cards
end

function FighterBehaviour:ChooseCard( card )
    card:SetRandSeed( math.random( 1, 2^32 ) )
    card:UpdateDamageAndBuffScale()
    
    local ok, reason = self.engine:CanPlayCard( card )
    if not ok then
        return false
    end

    if not self.engine:AssignPrimaryTarget( card ) then
        return false
    end

    -- if #self.prepared_cards > 0 then
    --     LOGWARN( "%s already has prepared '%s', now preparing '%s'",
    --         self.fighter, tostr(self.prepared_cards), card )
    --     table.clear( self.prepared_cards )
    -- end

    table.insert( self.prepared_cards, card )
    return true
end

function FighterBehaviour:ChooseCards( cards )
    for i, card in ipairs( cards ) do
        self:ChooseCard( card )
    end
end

function FighterBehaviour:SetPattern( pattern )
    if type(pattern) == "function" then
        self.current_pattern = pattern
    else
        self.current_pattern = self[ pattern ]
    end
    if type(self.current_pattern) ~= "function" then
        LOGWARN( "Unknown pattern for %s: '%s'", self.fighter, pattern )
    end
end

function FighterBehaviour:AddCard( attack_id )
    local card = Battle.Card( attack_id, self.fighter )
    card.show_dealt = false
    self.engine:DealCard( card, self.hand )
    return card
end

function FighterBehaviour:AddCards( t )
    local cards = {}
    for i, card_id in ipairs( card_ids ) do
        local card = self:AddCard( card_id )
        table.insert( cards, card )
    end
    return cards
end

function FighterBehaviour:OnActivateFighter( battle )
    self.engine = battle
    self.battle = battle

    self.hand = Battle.Deck( battle, DECK_TYPE.AI, true )

    if self.OnActivate then
        self:OnActivate( self.fighter )
    end

    if self.event_handlers then
        for event_name, fn in pairs( self.event_handlers ) do
            battle:ListenForEvent( event_name, self, self.OnBattleEvent )
        end
    end
end

function FighterBehaviour:OnBattleEvent( eventname, ... )
    if self.event_handlers and self.event_handlers[ eventname ] then
        self.event_handlers[ eventname ]( self, ... )
    end
end

function FighterBehaviour:OnDeactivateFighter( battle )
    if self.battle == nil then
        -- Never activated. Can happen if the fighter is deactivated -during- activation.
        return
    end

    if self.OnDeactivate then
        self:OnDeactivate( self.fighter )
    end

    while not self.hand:IsEmpty() do
        local card = self.hand:Peek()
        self.hand:RemoveCard( card )
    end

    battle:RemoveListener( self )
end

function FighterBehaviour:RunBehaviour( battle, pattern_name, ... )
    local turns = self.engine:GetTurns()
    local pattern_fn
    if pattern_name == nil then
        pattern_fn = self.current_pattern
        
        -- This really is jsut for diagnostics.
        for k, v in pairs( self ) do
            if v == self.current_pattern and k ~= "current_pattern" then
                pattern_name = k
                break
            end
        end
        table.clear( self.prepared_cards )

    else
        pattern_fn = self[ pattern_name ]
    end

    if pattern_fn then
        battle:AppendLog( BATTLE_LOG.AI, "{1} using pattern {2}", self.fighter, pattern_name )
        local ok, result = xpcall( pattern_fn, generic_error, self, turns, ... )

        if not ok then
            LOGWARN( result )
        end
    end

    return self.prepared_cards 
end

function FighterBehaviour:Delay( delay )
    self.fighter:Delay( delay )
end

function FighterBehaviour:SaySpeech( delay, txt )
    self.fighter:SaySpeech( delay, txt )
end

function FighterBehaviour:SayQuip( quip )
    self.fighter:SayQuip( quip )
end

function FighterBehaviour:PlayAnim( anim )
    self.fighter:PlayAnim( anim )
end

function FighterBehaviour:Present( fn )
    self.fighter:Present( fn )
end
