require "battle/battle_engine"
local battle_defs = require "battle/battle_defs"

------------------------------------------------------------------------
-- A scenario is used to describe the configuration for a battle scene.
-- It is used as an input parameter to BattleSim, and describes things
-- like the units involved, their location, the map layout, the environment, etc.
--
-- The Scenario class is fully mutable and dynamic, but is usually first
-- initialized from static content by the scenario_id constructor parameter.

class("Scenario")

function Scenario:init( location, params, quest )
    self.teams = {
        [ TEAM.BLUE ] = {},
        [ TEAM.RED ] = {}
    }
    self.reason = {}
    self.reason_txt = {}

    self.bystanders = {}
    self.on_start_battle = params and params.on_start_battle
    self.on_player_turn_start = params and params.on_player_turn_start
    self.before_player_turn_start = params and params.before_player_turn_start
    self.on_enemy_turn_end = params and params.on_enemy_turn_end
    
    self.reason_fn = params and params.reason_fn
    self.fight_reason_txt = params and params.reason_txt
    self.location = location
    self.params = params or {}
    self.flags = (params and params.flags) or battle_defs.BATTLE_FLAGS.NONE
    self.quest = quest
    self.primary = {}
    self.player_controled = {}
    self.blue_team_size = params and params.blue_team_size or 4
    self.red_team_size = params and params.red_team_size or 4

    if params and params.advantage then
        self:SetAdvantage( params.advantage )
    end

    if params and params.survival_turns then
        -- Convert to team-based turns (one turn per phase)
        self.survival_turns = params.survival_turns
        self.flags = SetBits( self.flags, battle_defs.BATTLE_FLAGS.SURVIVAL )
    end

    if TheGame:GetGameState() then
        if not TheGame:GetGameState():GetDayPhase() == DAY_PHASE.DAY then
            self.flags = SetBits( self.flags, battle_defs.BATTLE_FLAGS.NIGHT )
        end
    end

    if location == nil then
        self.location = Location( "OUTDOORS" )
    end

    self.enemy_scale = params and params.enemy_scale or 1
    self.ally_scale = params and params.ally_scale or 1
end

--this is the difficulty of the "The world" that this battle takes place in. It scales your allies
function Scenario:GetAllyScale()
    return self.ally_scale
end

--this is the difficulty of the the foes that you face.
function Scenario:GetEnemyScale()
    return self.enemy_scale
end

function Scenario:IsAggressive()
    return false
    -- local primary_enemy = self:GetPrimary(TEAM.RED)
    -- if primary_enemy and primary_enemy:IsSentient() and not CheckBits( self.flags, BATTLE_FLAGS.ISOLATED ) and not CheckBits( self.flags, BATTLE_FLAGS.SELF_DEFENCE ) then
    --     return true
    -- end
end

function Scenario:ApplyRepercussions( repercussions )
    if self.params.loot then
        repercussions.loot.money = repercussions.loot.money + (self.params.loot.money or 0)
        table.arrayadd( repercussions.loot.items, self.params.loot.items or table.empty )
        table.arrayadd( repercussions.loot.grafts, self.params.loot.grafts or table.empty )
    end
end

function Scenario:GetAgentScale(agent)
    
    if agent:IsBoss() then
        return 1
    end

    if self:IsPlayerControlled(agent) then
        return 1
    end

    if self:GetTeamForAgent(agent) == TEAM.BLUE then
        return self:GetAllyScale()
    else
        return self:GetEnemyScale()
    end
end

function Scenario:GetSceneScaling()
    -- Not to be confusd with difficulty scaling, this is a viz scaling factor.
    if self.location then
        return self.location:GetContent().scene_scale or 1.0
    end

    return 1.0
end

function Scenario:GetPrimary(team_id)
    return self.primary[team_id]
end

function Scenario:AddTimePenalty()
    self.time_penalty = self.time_penalty and self.time_penalty + 1 or 1
end

function Scenario:AllowsBackup()
    return (self.flags & battle_defs.BATTLE_FLAGS.NO_BACKUP) == 0
end

function Scenario:IsNight()
    return CheckBits( self.flags, battle_defs.BATTLE_FLAGS.NIGHT )
end

function Scenario:IsBossFight()
    return CheckBits( self.flags, battle_defs.BATTLE_FLAGS.BOSS_FIGHT )
end

function Scenario:IsSelfDefence()
    return CheckAnyBits( self.flags, BATTLE_FLAGS.SELF_DEFENCE | BATTLE_FLAGS.SURVIVAL )
end

function Scenario:CanFlee()
    return not CheckAnyBits( self.flags, BATTLE_FLAGS.NO_FLEE | BATTLE_FLAGS.SURVIVAL )
end

function Scenario:GetSurvivalTurns()
    if CheckBits( self.flags, BATTLE_FLAGS.SURVIVAL ) then
        return self.survival_turns or 6
    end
end

function Scenario:GetRewardDifficulty()
    return self.params.reward_difficulty
end

function Scenario:GetFlags()
    return self.flags
end

function Scenario:SetFlags( flags )
    self.flags = SetBits( self.flags, flags )
end

function Scenario:GetLocation()
    return self.location
end

function Scenario:GetTeam( team_id )
    return self.teams[ team_id ]
end

function Scenario:GetPlayerTeam()
    return TEAM.BLUE
end

function Scenario:GetEnemyTeam()
    return TEAM.RED
end


function Scenario:IsPlayerControlled(agent)
    return self.player_controled[agent] ~= nil
end


function Scenario:AddBystander(agent, reason, txt)
    table.insert(self.bystanders, 1, agent)
    self.reason[agent] = reason
    self.reason_txt[agent] = reason
end

function Scenario:GetFightReason(agent)
    return self.reason[agent], self.reason_txt[agent]
end

function Scenario:SetPlayerControlled(agent)
    self.player_controled[agent] = true
end


function Scenario:AddAgentToTeam( agent, team_id, reason, txt )
    if agent:IsRetired() then
        return
    end

    assert( IsEnum( team_id, TEAM ))
    assert( not agent:IsPlayer() or team_id == TEAM.BLUE )
    local team = self.teams[ team_id ]
    table.insert(team, agent)
    
    self.reason[agent] = reason
    self.reason_txt[agent] = txt

    if self.primary[team_id] == nil then
        self.primary[team_id] = agent
    end
end


function Scenario:GetStartingTeamSize(team_id)
    return self.teams[team_id] and #self.teams[team_id] or 0
end



function Scenario:GetTeamForAgent( agent )
    for k,v in pairs(self.teams) do
        if table.arrayfind(v, agent) then
            return k
        end
    end
end

function Scenario:SetAdvantage( team_id )
    self.advantage = team_id
end

function Scenario:LoseAdvantage( )
    self.advantage = nil
end

function Scenario:GetAdvantageTeam( )
    return self.advantage
end

function Scenario:GetStringDump()
    local t= {}
    table.insert(t, "---------FIGHT---------")
    table.insert(t,  loc.format( "Quest: {1}", self.quest ))
    table.insert(t, loc.format("Location: {1}", tostring(self.location)))
    table.insert(t, loc.format("Flags: {1}", StringizeBitField( self.flags, BATTLE_FLAGS_STRINGS, ", " )))

    table.insert(t, "====TEAM BLUE====")

    if self.advantage == TEAM.BLUE then
        table.insert(t, "Has Advantage!")
    end

    for k,v in ipairs(self.teams[TEAM.BLUE]) do
        table.insert(t, loc.format("\t{4}{1}: {2} ({3}) - {5}", v:GetContentID(), v:GetName(), self.reason_txt and self.reason_txt[v] or "", self:GetPrimary(TEAM.BLUE) == v and "* " or "  ", self:GetAgentScale(v)))

        if v.graft_owner then
            local grafts = {}
            for i, v in ipairs( v.graft_owner:GetAllGrafts() ) do
                if v:GetType() == GRAFT_TYPE.COMBAT or v:GetType() == GRAFT_TYPE.SOCIAL then
                    table.insert( grafts, tostring( v ))
                end
            end
            if #grafts > 0 then
                table.insert(t, loc.format( "\tGrafts: {1#listing}", grafts ))
            end
        end
    end

    table.insert(t, "====TEAM RED====")

    if self.advantage == TEAM.RED then
        table.insert(t, "Has Advantage!")
    end

    for k,v in ipairs(self.teams[TEAM.RED]) do
        table.insert(t, loc.format("\t{4}{1}: {2} ({3}) - {5}", v:GetContentID(), v:GetName(), self.reason_txt and self.reason_txt[v] or "", self:GetPrimary(TEAM.RED) == v and "* " or "  ", self:GetAgentScale(v)))
    end

    table.insert(t, "-----------------------")

    return table.concat(t, "\n")
end

