require "battle/combatcondition"
require "battle/fighter"
local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

--------------------------------------------------------------
local Team = class("Team")
Team.can_reload = true

function Team:init(id, battle, max_size)
    self.id = id
    self.battle = battle
    self.reserve_fighters = {}
    self.fighters = {}
    self.inactive_fighters = {}
    self.max_fighters = max_size or 4
end

function Team:CalculateStrength()
    local total_points = 0
    local max_points = 0
    for k, fighter in pairs(self.fighters) do
        local t = fighter:GetHealthPercent()
        total_points = total_points + fighter.agent:GetCombatStrength()*t
        max_points = max_points + fighter.agent:GetCombatStrength()
    end

    for k, fighter in pairs(self.reserve_fighters) do
        
        local t = fighter:GetHealthPercent()
        total_points = total_points + fighter.agent:GetCombatStrength()*t
        max_points = max_points + fighter.agent:GetCombatStrength()
    end

    
    return total_points, max_points
end

function Team:GetMaxFighters()
    return self.max_fighters
end

function Team:SetMaxFighters( max_fighters )
    self.max_fighters = max_fighters
end

function Team:SetCustomFormation( formation )
    -- formation is a map of [#fighters] -> array of positions
    -- OR a function that returns an array of positions.
    self.custom_formation = formation
end

function Team:GetCustomFormation()
    if type(self.custom_formation) == "function" then
        return self.custom_formation( self )
    elseif type(self.custom_formation) == "table" then
        return self.custom_formation[ #self.fighters ]
    end
end

function Team:GetEnemyTeam()
    return self.battle:GetTeam( self.battle:GetOtherTeam( self.id ) )
end

function Team:GetID()
    return self.id
end

function Team:GetFighters()
    return self.fighters
end

function Team:GetReserveFighters()
    return self.reserve_fighters
end

function Team:IsFull()
    return self:NumActiveFighters() >= self:GetMaxFighters()
end

function Team:Fighters()
    return ipairs(self.fighters)
end

function Team:AllFighters()
    return ipairs( self:GetAllFighters() )
end

function Team:GetAllFighters()
    local fighters = table.merge(self.fighters, self.inactive_fighters, self.reserve_fighters)
    return fighters
end

function Team:GetFighterForAgent(agent)
    for k,v in pairs(self.fighters) do
        if v:GetAgent() == agent then
            return v
        end
    end
    for k,v in pairs(self.inactive_fighters) do
        if v:GetAgent() == agent then
            return v
        end
    end
    for k,v in pairs(self.reserve_fighters) do
        if v:GetAgent() == agent then
            return v
        end
    end
end

function Team:NumActiveFighters()
    return #self.fighters
end

function Team:NumFighters()
    return #self.fighters + #self.inactive_fighters + #self.reserve_fighters
end

function Team:GetTeamStatus()
    -- If the player is dead/fled, that's it.
    for i, fighter in ipairs( self.inactive_fighters ) do
        if fighter.agent:IsPlayer() then
            if fighter:GetStatus() == FIGHT_STATUS.DEAD then
                return FIGHT_STATUS.DEAD
            elseif fighter:GetStatus() == FIGHT_STATUS.FLED then
                return FIGHT_STATUS.FLED
            end
        end
    end

    -- If there are pending reserves, then the team is alive.
    if #self.reserve_fighters > 0 then
        return FIGHT_STATUS.ACTIVE
    end

    -- If ANYONE is alive and kicking, then the team is alive.
    local surrender_count = 0
    for i, fighter in ipairs( self.fighters ) do
        if fighter:GetStatus() == FIGHT_STATUS.ACTIVE then
            return FIGHT_STATUS.ACTIVE
        elseif fighter:GetStatus() == FIGHT_STATUS.SURRENDER then
            surrender_count = surrender_count + 1
        end
    end

    -- If only surrendered fighters remain....
    if surrender_count > 0 then
        return FIGHT_STATUS.SURRENDER
    end

    -- If any inactive fighters fled, then the team has fled.
    for i, fighter in ipairs( self.inactive_fighters ) do
        if fighter:GetStatus() == FIGHT_STATUS.FLED then
            return FIGHT_STATUS.FLED
        end
    end

    -- Otherwise, we must be dead.
    return FIGHT_STATUS.DEAD
end

function Team:Primary()
    -- The primary is the first fighter added -- either the instigator or victim of battle.
    return self.primary_fighter
end

function Team:GetRandomTeammate( fighter )
    local t = ObtainWorkTable()
    table.arrayadd( t, self.fighters )
    if fighter then
        table.arrayremove( t, fighter )
    end
    local teammate = #t > 0 and t[ math.random( #t )]
    ReleaseWorkTable( t )
    return teammate
end

function Team:RunAway()
    self.run_away = true

    for i, fighter in self:Fighters() do
        fighter:Flee()
    end
end

function Team:HasRunAway()
    return self.run_away
end

-- PRE-battle function ONLY
function Team:RemoveFighter(fighter)
    assert( fighter ~= self.primary_fighter )
    assert( self.battle == nil )
    table.arrayremove(self.reserve_fighters, fighter)
    table.arrayremove(self.fighters, fighter)
    table.arrayremove(self.inactive_fighters, fighter)
    fighter.team = nil
end

function Team:AddFighter(fighter, start_turn)
    assert( is_instance( fighter, Fighter ))
    assert( fighter.team == nil )    
    fighter.team = self
    fighter.starting_team = fighter.starting_team or self

    local current_turn = self.battle and self.battle:GetTurns() or 0
    fighter.start_turn = current_turn + (start_turn or 0)

    if self.primary_fighter == nil or fighter:IsPlayer() then
        self.primary_fighter = fighter
    end
    
    local function SortByStartTurn( a, b )
        -- Sort in descending order of start_turn, with primary_fighter last.
        if a.start_turn == b.start_turn then
            return self.primary_fighter == b
        else
            return a.start_turn > b.start_turn
        end
    end

    table.insert( self.reserve_fighters, fighter )
    table.sort(self.reserve_fighters, SortByStartTurn )

    if self.battle then
        self.battle:BroadcastEvent( BATTLE_EVENT.ADDED_FIGHTER, fighter )
    end

    return fighter
end

function Team:IsSurrendering()
    return self:GetTeamStatus() == FIGHT_STATUS.SURRENDER
end

local function TryAddFighter( self, fighter, add_delay, replacing )

    local replace_fighter
    if #self.fighters >= self:GetMaxFighters() then
        return false
    end

    table.arrayremove(self.reserve_fighters, fighter)
    table.insert(self.fighters, fighter)

    fighter.has_been_activated = true
    fighter.start_turn = self.battle and self.battle:GetTurns()
    if add_delay and self.battle then
        self.battle:BroadcastEvent( BATTLE_EVENT.DELAY, .1 + math.random()*.25 )
    elseif replacing and self.battle then
        self.battle:BroadcastEvent( BATTLE_EVENT.DELAY, 1.3 )
    end

    fighter:OnActivateFighter( self.battle )

    return true
end

function Team:ActivateNewFighters( replacing )
    local function SortByHealth( a, b )
        
        if a.agent:IsPlayer() ~= b.agent:IsPlayer() then
            return b.agent:IsPlayer()
        end

        -- Sort in ascending order of health.
        if a.start_turn <= b.start_turn then
            if self.primary_fighter == b then
                return false
            elseif a.agent:IsInPlayerParty() ~= b.agent:IsInPlayerParty() then
                -- Prioritize against crew members, because that's technically in the player's favor.
                return a.agent:IsInPlayerParty()
            else
                return a:GetHealthPercent() < b:GetHealthPercent()
            end
        else
            return false
        end
    end
    table.sort(self.reserve_fighters, SortByHealth )

    local new_fighters = 0
    while true do
        local fighter = self.reserve_fighters[ #self.reserve_fighters ]
        if fighter == nil or fighter:GetStartTurn() > self.battle:GetTurns() then
            break
        end
        if TryAddFighter( self, fighter, new_fighters > 0, replacing ) then
            new_fighters = new_fighters + 1
        else
            break
        end
    end

    if new_fighters > 0 then
        self.battle:BroadcastEvent( BATTLE_EVENT.SETTLE_FIGHTERS )
    end
end

function Team:RemoveFromField( fighter )

    assert( table.arrayfind( self.fighters, fighter ) ~= nil, tostring(fighter) )
    table.arrayremove( self.fighters, fighter )

    assert( not table.contains( self.inactive_fighters, fighter ))
    table.insert( self.inactive_fighters, fighter )

    fighter:OnDeactivateFighter()
end

function Team:LookForReplacement( fighter )
    if #self.reserve_fighters > 0 and fighter ~= self:Primary() then
        self.battle:BroadcastEvent( BATTLE_EVENT.FIGHTER_REPLACED, fighter)
        fighter:Flee()
        self:ActivateNewFighters()
    end
end