require "collection"
local negotiation_defs = require "negotiation/negotiation_defs"
local battle_defs = require "battle/battle_defs"

----------------------------------------------------------------------------
-- Card filters

local function IsRewardable( card_def )
    return card_def.rarity ~= CARD_RARITY.UNIQUE and
            card_def.rarity ~= CARD_RARITY.BOSS and
            card_def.rarity ~= CARD_RARITY.BASIC
end

local function BasicFilter( card_def, owner )
    owner = owner or (TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent())

    return 
        card_def.rarity ~= CARD_RARITY.UNIQUE and
        card_def.rarity ~= CARD_RARITY.BOSS and
        card_def.rarity ~= CARD_RARITY.BASIC and
        (owner == nil or AgentUtil.CanUseCard( owner, card_def ))
end

local function NotFlag(card_def, flag)
    return not CheckBits( card_def.flags, flag) 
end

local function HasFlag(card_def, flag)
    return CheckBits( card_def.flags, flag ) 
end

local function IsNativeToRegion( card_def )
    if card_def.local_item then
        local act = TheGame:GetGameState() and TheGame:GetGameState():GetCurrentAct()
        if act then
            return card_def.local_item & (act.item_tag_mask or 0) ~= 0
        end
    end
    return true
end


local function IsUnlocked( card_def )
    local game_type = TheGame:GetGameState() and TheGame:GetGameState():GetGameType()
    if game_type == GAME_TYPE.DAILY then
        return true
    end
    
    return not Content.IsUnlock( card_def.id ) or TheGame:GetGameProfile():HasUnlock( card_def.id, TheGame:GetGameState() )
end

local function UniqueDuplicate( card_def, flag, cardholder )
    return not CheckBits( card_def.flags, flag) or not cardholder:HasCardID(card_def.id)
end

local function IsSeries( card_def, series )
    return card_def.series == series
end

----------------------------------------------------------------------------

local BattleCardCollection = class( "BattleCardCollection", Collection )

function BattleCardCollection:init(fn, ...)
    BattleCardCollection._base.init(self)
    self:AddItems(Content.GetAllBattleCards(), fn, ...)
end

function BattleCardCollection:IsUnlocked()
    self:Filter(IsUnlocked)
    return self
end

function BattleCardCollection:GenerateRandom(num, owner)
    owner = owner or TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    return table.map(self:Pick(num), function(def) return Battle.Card(def.id, owner) end)
end

function BattleCardCollection.AllUsableCards( fn, owner )
    local player = owner or TheGame:GetGameState():GetPlayerAgent()
    local collection = BattleCardCollection( function( card_def )
                            return 
                                AgentUtil.CanUseCard( player, card_def )
                            end )
                            :Filter( IsUnlocked )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.STATUS )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.SPECIAL )
                            :Filter( fn )

    return collection

end

function BattleCardCollection.AllRewardableCards( fn, owner )
    local player = owner or TheGame:GetGameState():GetPlayerAgent()
    local collection = BattleCardCollection( BasicFilter, owner )
                            :Filter( IsUnlocked )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.ITEM )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.UPGRADED )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.PARASITE )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.FLOURISH )
                            :Filter( UniqueDuplicate, battle_defs.CARD_FLAGS.ONE_OF_A_KIND, player:GetAspect("battler"))
                            :Filter( fn )

    return collection
end

function BattleCardCollection.RewardableCardsFromSeries( fn, owner, series )
    owner = owner or TheGame:GetGameState():GetPlayerAgent()
    local collection = BattleCardCollection( IsRewardable )
                            :Filter( IsSeries, series )
                            :Filter( IsUnlocked )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.ITEM )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.UPGRADED )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.PARASITE )
                            :Filter( NotFlag, battle_defs.CARD_FLAGS.FLOURISH )                            
                            :Filter( UniqueDuplicate, battle_defs.CARD_FLAGS.ONE_OF_A_KIND, owner:GetAspect("battler"))
                            :Filter( fn )

    return collection

end

function BattleCardCollection.AllItems( fn, owner )
    return BattleCardCollection( BasicFilter, owner )
                            :Filter( IsUnlocked )
                            :Filter( HasFlag, battle_defs.CARD_FLAGS.ITEM )
                            :Filter( fn )
end


function BattleCardCollection.AllLocalItems( fn, owner )
    return BattleCardCollection( BasicFilter, owner )
                            :Filter( IsNativeToRegion ) 
                            :Filter( IsUnlocked )
                            :Filter( HasFlag, battle_defs.CARD_FLAGS.ITEM )
                            :Filter( fn )
end
----------------------------------------------------------------------------
--

local NegotiationCardCollection = class( "NegotiationCardCollection", Collection )

function NegotiationCardCollection:init(fn, ...)
    NegotiationCardCollection._base.init(self)
    self:AddItems(Content.GetAllNegotiationCards(), fn, ...)
end

function NegotiationCardCollection:IsUnlocked()
    self:Filter(IsUnlocked)
    return self
end

function NegotiationCardCollection:GenerateRandom(num, owner)
    owner = owner or TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    return table.map(self:Pick(num), function(def) return Negotiation.Card(def.id, owner) end)
end

function NegotiationCardCollection.AllUsableCards( fn, owner )
    local player = owner or TheGame:GetGameState():GetPlayerAgent()
    local collection = NegotiationCardCollection( function( card_def )
                                return 
                                    AgentUtil.CanUseCard( player, card_def )
                                end )
                            :Filter( IsUnlocked )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.STATUS )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.BYSTANDER )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.SPECIAL)
                            :Filter( fn )

    return collection

end


function NegotiationCardCollection.AllRewardableCards(fn, owner)
    local player = owner or TheGame:GetGameState():GetPlayerAgent()
    return NegotiationCardCollection( BasicFilter, owner )
                    :Filter( IsUnlocked )
                    :Filter( NotFlag, negotiation_defs.CARD_FLAGS.ITEM )
                    :Filter( NotFlag, negotiation_defs.CARD_FLAGS.UPGRADED )
                    :Filter( NotFlag, negotiation_defs.CARD_FLAGS.PARASITE )
                    :Filter( NotFlag, negotiation_defs.CARD_FLAGS.FLOURISH )
                    :Filter( UniqueDuplicate, negotiation_defs.CARD_FLAGS.ONE_OF_A_KIND, player:GetAspect("negotiator"))
                    :Filter( fn )
end

function NegotiationCardCollection.RewardableCardsFromSeries( fn, owner, series )
    owner = owner or TheGame:GetGameState():GetPlayerAgent()
    local collection = NegotiationCardCollection( IsRewardable )
                            :Filter( IsSeries, series )
                            :Filter( IsUnlocked )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.ITEM )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.UPGRADED )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.PARASITE )
                            :Filter( NotFlag, negotiation_defs.CARD_FLAGS.FLOURISH )
                            :Filter( UniqueDuplicate, negotiation_defs.CARD_FLAGS.ONE_OF_A_KIND, owner:GetAspect("negotiator"))
                            :Filter( fn )

    return collection

end

function NegotiationCardCollection.AllItems(fn, owner)
    return NegotiationCardCollection( BasicFilter, owner )
                    :Filter( IsUnlocked )
                    :Filter( HasFlag, negotiation_defs.CARD_FLAGS.ITEM )
                    :Filter( fn )
end

function NegotiationCardCollection.AllLocalItems(fn, owner)
    return NegotiationCardCollection( BasicFilter, owner )
                    :Filter( IsNativeToRegion ) 
                    :Filter( IsUnlocked )
                    :Filter( HasFlag, negotiation_defs.CARD_FLAGS.ITEM )
                    :Filter( fn )
end
----------------------------------------------------------------------------
--

function GetUniqueFactionCards(player, faction_id, num_battle, num_negotiation, ret)
    local rng = TheGame:GetGameState():GetRNG( "CARDS" )
    local ret = ret or {}
    for k,v in ipairs( NegotiationCardCollection(function(card) return card.faction == faction_id and card.rarity == CARD_RARITY.UNIQUE and card.special_unlock ~= true end):SeededPick(num_negotiation, rng) ) do
        table.insert(ret, Negotiation.Card(v.id, player))
    end

    for k,v in ipairs( BattleCardCollection(function(card) return card.faction == faction_id and card.rarity == CARD_RARITY.UNIQUE and card.special_unlock ~= true end):SeededPick(num_battle, rng) ) do
        table.insert(ret, Battle.Card(v.id, player))
    end

    return ret
end
