require "eventsystem"

DECK_TYPE = MakeEnum
{
    "DRAW", -- You draw cards into your hand from here.
    "IN_HAND", -- Cards currently in hand.
    "RESOLVE", -- Cards currently being resolved (typically one or zero)
    "DISCARDS", -- Discarded cards.
    "AI", -- A deck used by the AI (FighterBehaviour)
    "TRASH", -- Trashed cards (no longer in play)
}


DECK_TYPE_ARRAY = MakeEnumArray( DECK_TYPE )

ALL_DECKS = DECK_TYPE_ARRAY

--------------------------------------------------------------------------------
-- Common code for card engines.

local CardEngine = class( "CardEngine")
CardEngine.can_reload = true

require "cards/scalar_accumulator"

local function AppendFeature( features, id )
    for i, v in ipairs( features ) do
        if v.id == id then
            return v
        end
    end

    local t = { id = id }
    table.insert( features, t )
    return t
end


local function FormatFeature( feature, stacks, features, engine, defs )
    local posted_feature = feature.posted_feature and defs.GetFeature( feature.posted_feature )
    if posted_feature then
        -- This is the descriptive used specifically when extracting this feature as a keyword.
        assert( posted_feature ~= feature )
        FormatFeature( posted_feature, stacks, features, engine, defs )

        return feature:GetLocalizedName()

    else
        local t = AppendFeature( features, feature.id )
        if t.name then
            return t.name
        else
            -- Otherwise, default to the feature's usual name and desc.
            t.name = feature:GetLocalizedName()

            local desc = feature:GetLocalizedDesc()
            if feature.desc_fn then
                desc = feature:desc_fn( desc, stacks, engine, features )
            end
            t.desc = CardEngine.ExtractKeywords( desc or "-", nil, engine, defs, features )
        end

        return t.name
    end
end


-- str: the localized string from which to extract keywords like {POWER} or {EXPEND}.
-- widget: (optional) Widget to post extracted features to.
-- engine: (optional) reference to an instance of the CardEngine. Used for more precise description replacements.
-- defs: either battle_defs or negotiation_defs; used to lookup features.
-- features: (optional) array of extracted features so far, used to detect loops and self-references

function CardEngine.ExtractKeywords( str, widget, engine, defs, features )
	if str == nil or #str == 0 then
		return nil
	end

    if features == nil then
        features = {}
    end

    local j, k, match_str, num, bar, postfix
    repeat
        -- Matches: {prefix.keyword num|postfix}
        -- prefix, num and postfix are all optional decorators.
        j, k = str:find( "%b{}", j )
        if j then
            local _, _, match_str, num, bar, postfix = str:sub( j + 1, k - 1 ):find( "%s*([_.%w]+)%s*([^|]*)%s*([|]?)(.*)" )
            local prefix, keyword
            if match_str then
                prefix, keyword = match_str:match( "([^.]+)[.](.+)")
                if keyword == nil then
                    keyword = match_str
                end
            end

            if prefix == "battle" then
                defs = require "battle/battle_defs"
                prefix = nil
            elseif prefix == "negotiation" then
                defs = require "negotiation/negotiation_defs"
                prefix = nil
            end

            local name, desc
            local feature = defs and defs.GetFeature( keyword )
            if feature and (prefix == nil or prefix == "feature") then
                local stacks = tonumber(num)
                if stacks == nil then
                    stacks = tonumber(SanitizeString(num))
                end
                name, desc = FormatFeature( feature, stacks, features, engine, defs )

                if stacks then
                    name = loc.format( LOC "UI.CARDS.FEATURE_NAME_STACKS", name, num )
                else
                    name = loc.format( LOC "UI.CARDS.FEATURE_NAME_NOSTACKS", name )
                end

            elseif (prefix == nil or prefix == "card") then
                local carddef = Content.GetBattleCard( keyword ) or Content.GetNegotiationCard( keyword )
                if carddef then
                    local t = AppendFeature( features, keyword )
                    t.card = carddef
                    name = loc.format( "<b>{1}</b>", carddef:GetLocalizedName() )
                end

            elseif (prefix == nil or prefix == "graft") then
                local graftdef = Content.GetGraft(keyword)
                if graftdef then
                    local t = AppendFeature( features, keyword )
                    t.graft = GraftInstance( keyword )
                    name = loc.format( "<b>{1}</b>", graftdef:GetLocalizedName() )
                end
            end

            if bar and #bar > 0 then
                if postfix and #postfix > 0 then
                    name = loc.format( "<b>{1}</b>", postfix )
                else
                    name = ""
                end
            end

            if name then
                -- Substitute the given localized name.
                str = str:sub( 1, j - 1 ) .. name .. str:sub( k + 1 )
                j = j + name:len()

            else
                -- Couldn't find any card or feature, so leave the substring alone.
                j = k + 1
            end
        end
    until not j

    if widget then
        for i, t in ipairs( features ) do
            if t.card then
                widget:PostCard( t.id, t.card, engine )
            elseif t.graft then
                widget:PostGraft( t.id, t.graft )
            else
                widget:PostFeature( t.id, t.name, t.desc )
            end
        end
    end

    return str
end


function CardEngine.GenerateCardDesc( card, widget, engine, defs, hilight_upgrades )
	local desc = ObtainWorkTable()
    local card_content
    if is_instance( card, BattleCardDef ) or is_instance( card, NegotiationCardDef ) then
        card_content = card
    else
        card_content = card.def
    end
    assert( card_content, card._classname )

	local loc_desc = card_content:GetLocalizedDesc()
    if loc_desc and not hilight_upgrades then
        loc_desc = SanitizeString( loc_desc, "<#UPGRADE>" )
        loc_desc = SanitizeString( loc_desc, "<#DOWNGRADE>" )
    end
    
    if card_content.desc_fn then
        loc_desc = card_content.desc_fn( card, loc_desc, widget, engine )
    end
    if loc_desc then
       table.insert( desc, loc_desc )
    end

    if not card.manual_desc then
        -- Add keyword flags.
        local keywords = ObtainWorkTable()
        for i, bitstr in ipairs( defs.CARD_FLAG_BITSTRINGS ) do
            local bit = 1 << (i-1)
            if card:IsFlagged( bit ) and defs.GetFeature( bitstr ) then
                if hilight_upgrades and card_content.flag_difs and CheckBits(card_content.flag_difs, bit) then
                    local keyword = string.format( "<#UPGRADE>{%s}</>", bitstr ) 
                    table.insert( keywords, keyword )
                else
                    local keyword = string.format( "{%s}", bitstr ) 
                    table.insert( keywords, keyword )
                end
            end
        end

        -- Add feature descriptions
        if card.features then
            for feature_id, num in pairs( card.features ) do
                local feature = defs.GetFeature( feature_id )
                local feature_desc = feature:GetFeatureDesc()
    			if feature_desc then 
    				if feature.feature_desc_fn then
    					local formatted_data = feature.feature_desc_fn( feature, feature_desc, num, engine, card )
    					table.insert( desc, formatted_data )
    				else
    					local formatted_data = loc.format( feature_desc, num )
    					table.insert( desc, formatted_data )
    				end
    			else
    				local keyword = string.format( "{%s %d}", feature_id, num )
    				table.insert( keywords, keyword )
    			end
            end
        end
        if #keywords > 0 then
    		local formatted_data = loc.format( "{1#comma_listing}", keywords )
            table.insert( desc, formatted_data )
        end
        ReleaseWorkTable( keywords )
    end

    local charges = card.GetCharges and card:GetCharges()
    if charges then
        table.insert( desc, loc.format( LOC "UI.CARDS.CONSUME_DESC", charges ) )
    end

    if card:IsFlagged( card.CARD_FLAGS.HATCH ) then
        local xp, max_xp = card:GetXP()
        if xp then
            local hatch_count = math.max( 0, max_xp - xp )
            table.insert( desc, loc.format( LOC "UI.CARDS.HATCH_DESC", hatch_count ))
        end
    end
    
    local desc_str
    if #desc > 0 then
        desc_str = table.concat( desc, "\n" )
        desc_str = CardEngine.ExtractKeywords( desc_str, widget, engine, defs )
    end

    ReleaseWorkTable( desc )
    return desc_str
end

---------------------------------------------------------------------
-- Card engine framework

function CardEngine:init()
    self.events = EventSystem()
    self.card_draw_accumulator = CardEngine.ScalarAccumulator( self, self.EVENT.CALC_CARD_DRAW )
end

function CardEngine:BroadcastEvent( eventname, ... )
    self.events:BroadcastEvent( eventname, ... )

    while self.post_handlers and #self.post_handlers > 0 do
        local t = table.remove( self.post_handlers )
        call_delegate( t )
    end
end

function CardEngine:PushPostHandler( fn, ... )
    -- FIXME: this is a major event-priority issue hack.
    self.post_handlers = self.post_handlers or {}
    table.insert( self.post_handlers, { fn, ... })
end

function CardEngine:RemoveListener( listener )
    self.events:RemoveListener( listener )
end

function CardEngine:StopListening( event_name, listener )
    self.events:StopListening( event_name, listener )
end

function CardEngine:ListenForAny( listener, fn, priority )
    self.events:ListenForAny( listener, fn, priority )
end

function CardEngine:ListenForEvent( event_name, listener, fn, priority )
    self.events:ListenForEvent( event_name, listener, fn, priority )
end

function CardEngine:YieldPlay( status, ... )
    assert( IsEnum( status, self.STATUS ))
    return coroutine.yield( status, ... )
end

function CardEngine:ResumePlay( ... )
    if self.coro == nil then
        return self.status, table.empty
    end

    self.status = self.STATUS.ACTIVE

    local result = { coroutine.resume( self.coro, ... ) }
    local ok, err = result[1], result[2]
    if not ok then
        error( "ERR:" ..tostring(err).."\n"..debug.traceback( self.coro ) )
    end

    local status = coroutine.status( self.coro )
    if status ~= "suspended" then
        self.coro = nil
        self.wait_params = nil
        return self.status, table.empty

    else
        table.shiftdown( result ) -- remove the 'coroutine.resume' result

        self.status = table.shiftdown( result ) -- remove the STATUS.
        assert( IsEnum( self.status, self.STATUS ), self.status )

        self.wait_params = result
        return self.status, result
    end
end

function CardEngine:IsDone()
    return self.coro == nil
end

function CardEngine:GetStatus()
    return self.status
end

function CardEngine:PlayCards( deck, pred )
    local cards = ObtainWorkTable()
    while true do
        local autoplay_card
        for i, card in self.hand_deck:Cards() do
            if pred( card ) and not table.contains( cards, card ) then
                autoplay_card = card
                break
            end
        end
        if autoplay_card then
            table.insert( cards, autoplay_card )
            self:PlayCard( autoplay_card )
        else
            break
        end
    end
    ReleaseWorkTable( cards )
end

function CardEngine:DealCard( card, deck )
    if deck == nil then
        deck = self:GetDeck( card.starting_deck or DECK_TYPE.DRAW )
    end

    -- Dealing a card means it's never been attached yet in any way.
    assert( card.deck == nil, tostring(card) )

    if deck:GetDeckType() == DECK_TYPE.IN_HAND then
        if not deck:InsertCard( card ) then
            card:TransferCard( self:GetDeck( DECK_TYPE.DISCARDS ))
        end
        return true

    else
        return deck:InsertCard( card, math.random( 1, deck:CountCards() + 1 ) )
    end
end

function CardEngine:DealCards( dealt_cards, deck )
    for i, card in ipairs( dealt_cards ) do
        self:DealCard( card, deck )
    end
end

function CardEngine:FindDeckWithCard( card )
    for deck_type, _ in pairs( DECK_TYPE ) do
        local deck = self:GetDeck( deck_type )
        if deck:HasCard( card ) then
            return deck
        end
    end
end

function CardEngine:CalculateCardDraw( n )
    local draw_count, details = self.card_draw_accumulator:CalculateValue( n )
    draw_count = math.max( 0, draw_count )
    return draw_count, details
end

function CardEngine:DrawCards( n, start_of_turn )
    local draw_cards = {}
    n = self:CalculateCardDraw( n or 1 )

    local i = 1
    while i <= n do
        if self.draw_deck:CountCards() == 0 then
            -- Ran outta draw?  Shuffle in from the discards.
            self:ShuffleDiscardToDraw()
        end


        local card = self.draw_deck:Peek()
        if card == nil then
            break
        end

        if not start_of_turn then
            self.cards_drawn = (self.cards_drawn or 0) + 1
        end

        if not self.hand_deck:InsertCard( card, self.hand_deck:CountCards() + 1 ) then
            -- Insert into discards instead.
            card:TransferCard( self.discard_deck )
            i = i + 1
        else
            table.insert( draw_cards, card )
            self:BroadcastEvent( self.EVENT.DRAW_CARD, self, card, start_of_turn )

            if not card:IsFlagged( self.CARD_FLAGS.REPLENISH ) then
                i = i + 1
            end
        end
    end
    
    return draw_cards
end

function CardEngine:ShuffleDiscardToDraw()
    if self.discard_deck:CountCards() > 0 then
        self:BroadcastEvent( self.EVENT.SHUFFLE_DISCARDS, self.discard_deck:CountCards() )
        self.discard_deck:TransferCards( self.draw_deck )
        self.draw_deck:Shuffle()
    end
end

function CardEngine:ChooseCard( pred, txt, source )
    return self:ChooseCards( 1, 1, pred, txt, source )[1]
end

function CardEngine:ChooseCards( min_count, max_count, pred, txt, source )
    return self:ChooseCardsFromTable( self.hand_deck, min_count, max_count, pred, txt, source )
end

function CardEngine:ChooseCardsFromDeck( min_count, max_count, deck_type, pred, txt, source )
    local deck = self:GetDeck( deck_type )
    return self:ChooseCardsFromTable( deck, min_count, max_count, pred, txt, source )
end

local CHOOSE_STRINGS =
{
    "CARD_ENGINE.CHOOSE_CARDS", -- Choose exactly this many
    "CARD_ENGINE.CHOOSE_MIN_CARDS", -- Choose at least this many
    "CARD_ENGINE.CHOOSE_MAX_CARDS", -- Choose up to this many
    "CARD_ENGINE.CHOOSE_SOME_CARDS", -- Choose between N and M
}

function CardEngine:ChooseCardsFromTable( t, min_count, max_count, pred, txt, source )
    if self:CheckGameOver() then
        return table.empty
    end

    if min_count == nil and max_count == nil then
        min_count, max_count = 0, math.huge -- Any number.
    elseif min_count == nil then
        min_count = max_count
    elseif max_count == nil then
        max_count = min_count
    end
    assert( min_count <= max_count )

    -- t is either a Deck or a plain table of cards.
    local cards
    if is_instance( t, CardEngine.Deck ) then
        cards = t:GetCards()
    else
        cards = t
    end

    -- Only select from valid cards.
    for i = #cards, 1, -1 do
        if pred and not pred( cards[i] ) then
            table.remove( cards, i )
        end
    end

    max_count = math.min( max_count, #cards )
    if max_count <= 0 then
        return table.empty -- Nothing valid for choosing.
    end
    min_count = math.min( min_count, max_count )

    -- Auto-choose if there isn't actually any choice involved.
    if #cards <= min_count then
        return cards -- should this be a copy?

    else
        if txt == nil then
            txt = CHOOSE_STRINGS
        end
        if type(txt) == "table" then
            if min_count == max_count then
                txt = loc.format( LOC( txt[1] ), max_count )
            elseif min_count > 0 and max_count >= #cards then
                txt = loc.format( LOC( txt[2] ), min_count )
            elseif min_count == 0 and max_count <= #cards then
                txt = loc.format( LOC( txt[3] ), max_count )
            else
                txt = loc.format( LOC( txt[4] ), min_count, max_count )
            end
        end

        return self:YieldPlay( self.STATUS.CHOOSE_CARDS, t, min_count, max_count, txt, pred, source ) or table.empty
    end
end

local DISCARD_STRINGS =
{
    "CARD_ENGINE.CHOOSE_DISCARDS", -- Choose exactly this many
    "CARD_ENGINE.CHOOSE_MIN_DISCARDS", -- Choose at least this many
    "CARD_ENGINE.CHOOSE_MAX_DISCARDS", -- Choose up to this many
    "CARD_ENGINE.CHOOSE_SOME_DISCARDS", -- Choose between N and M
}

function CardEngine:DiscardCards( min_count, max_count, source )
    local cards = self:ChooseCardsFromTable( self.hand_deck, min_count, max_count, nil, DISCARD_STRINGS, source )
    for i, card in ipairs( cards ) do
        self:DiscardCard( card )
    end

    return cards
end

function CardEngine:DiscardCard( card )
    if card == nil then
        card = self.hand_deck:PeekRandom()
    end

    if card and self.hand_deck:HasCard( card ) then
        print( "DiscardCard()", card )
        self.cards_discarded = (self.cards_discarded or 0) + 1
        self:BroadcastEvent( self.EVENT.CARD_DISCARDED, card, self )
        self.hand_deck:TransferCard( card, self.discard_deck )
    end
end

function CardEngine:ExpendCard( card )
    if card.deck and card.deck ~= self.trash_deck then
        card:TransferCard( self.trash_deck )
        self:BroadcastEvent( self.EVENT.CARD_EXPENDED, self, card )
    end
end

local EXPEND_STRINGS =
{
    "CARD_ENGINE.CHOOSE_EXPEND", -- Choose exactly this many
    "CARD_ENGINE.CHOOSE_MIN_EXPEND", -- Choose at least this many
    "CARD_ENGINE.CHOOSE_MAX_EXPEND", -- Choose up to this many
    "CARD_ENGINE.CHOOSE_SOME_EXPEND", -- Choose between N and M
}

function CardEngine:ExpendCards( min_count, max_count, source )
    local cards = self:ChooseCardsFromTable( self.hand_deck, min_count, max_count, nil, EXPEND_STRINGS, source )
    for i, card in ipairs( cards ) do
        self:ExpendCard( card )
    end

    return cards
end

---------------------------------------------------------------------
-- Base class for any Card.

local Card = class( "CardEngine.Card" )

function Card:GetContent()
    return self.def
end

function Card:GetName()
    return self.def:GetLocalizedName()
end

function Card:ListenForGlobalEvents()
    local handlers = self.def.global_event_handlers
    if handlers then
        for event_name, fn in pairs( handlers ) do
            local priority = self.event_priorities and self.event_priorities[ event_name ]
            TheGame:GetEvents():ListenForEvent( event_name, self, self.OnGlobalEvent, priority )
        end
    end
end

function Card:RemoveGlobalListener()
    local handlers = self.def.global_event_handlers
    if handlers then
        TheGame:GetEvents():RemoveListener( self )
    end
end

function Card:OnGlobalEvent( event_name, ... )
    local handler = self.def.global_event_handlers[ event_name ]
    if handler then
        handler( self, ... )
    end
end

function Card:OnEvent( eventname, ... )
    if self.deck == nil then
        return
    end
    local handled = self.deck:IsEventHandler()
    if not handled and self.deck_handlers then
        handled = table.arrayfind( self.deck_handlers, self.deck:GetDeckType() )
    end
    if handled and self.event_handlers and self.event_handlers[ eventname ] then
        self.event_handlers[ eventname ]( self, ... )
    end
end

-- Creates a literal copy of the Card, used for snapshots etc. To create a copy that is
-- playable independent from the original, use Duplicate()
function Card:Clone()
    return setmetatable( shallowcopy( self ), self._class )
end

-- Creates an independent copy of the card
-- It is not attached to any Deck until inserted.
function Card:Duplicate()
    local copy = setmetatable( shallowcopy( self ), self._class )
    copy.deck = nil
    copy.engine = nil
    copy.target = nil
    copy.show_dealt = nil

    copy.userdata = shallowcopy( copy.userdata )
    -- We don't want the clone to bother with xp gain, but other userdata is important (eg. charges)
    -- copy.userdata.upgrade_ids = nil
    copy.userdata.xp = nil

    if copy.on_duplicated then
        copy:on_duplicated()
    end

    copy:MakeTemporary()

    return copy
end

function Card:IsTemporary()
    return self.temporary == true
end

function Card:MakeTemporary( temporary )
    if temporary == nil then
        self.temporary = true
    else
        self.temporary = temporary
    end
end

function Card:TransferCard( deck, i )
    if self.deck then
        self.deck:TransferCard( self, deck, i )
    else
        deck:InsertCard( self, i )
    end
end

function Card:RemoveCard()
    if self.deck then
        self.deck:RemoveCard( self)
    end
end

function Card:SetFlags( flags )
    self.flags = SetBits( self.flags, flags )
end

function Card:ClearFlags( flags )
    self.flags = ClearBits( self.flags, flags )
end

function Card:IsFlagged( flags )
    return CheckBits( self.flags, flags )
end

function Card:IsFlaggedAny( flags )
    return CheckAnyBits( self.flags, flags )
end

function Card:GetUserData()
    return self.userdata
end

function Card:GetMaxXP()
    return self.userdata.max_xp or self.def.max_xp
end

function Card:HasUpgrades()
    if self:IsHatched() then
        return true

    elseif not self:IsTemporary() then
        return self.userdata.upgrade_ids and #self.userdata.upgrade_ids > 0
    end
    return false
end

function Card:UpgradeReady()
    if self.userdata.xp and self:HasUpgrades() then
        return (self.userdata.xp or 0) >= self:GetMaxXP()
    else
        return false
    end
end

function Card:UpgradeCard( upgrade_id )
    if upgrade_id == nil then
        upgrade_id = table.arraypick( self.userdata.upgrade_ids or table.empty )
    end
    local upgrade_def = upgrade_id and self:LookupCardDef( upgrade_id )
    if upgrade_def then
        local snapshot
        if self.engine then
            snapshot = self:Clone()
        end
        
        local modified_flags = self.flags ~ self.def.flags

        -- Clear existing properties.
        for k,v in pairs(self.def) do
            self[k] = nil
        end

        -- Assign new upgraded ones.
        for k,v in pairs(upgrade_def) do
            self[k] = v
        end
        self.features = deepcopy(upgrade_def.features)
        self.flags = SetBits( self.flags, modified_flags )

        self.def = upgrade_def

        -- Its best if userdata is preserved but not a copy. (so that transient upgrades
        -- during card playing don't affect the original card)
        self.userdata = shallowcopy( self.userdata )
        if (self.def.max_xp or 0) <= 0 then
            self.userdata.xp = nil
            self.userdata.upgrade_ids = nil
        else
            self.userdata.xp = 0
            self:RollUpgrades()
        end
        -- Upgrade may add or remove charges, e.g. mutator mutator
        if self.userdata.charges and not self.def.max_charges then
            self.userdata.charges = nil
            self.userdata.max_charges = nil
        elseif self.def.max_charges then
            self.userdata.charges = self.def.max_charges
        end

        if self.engine then
            local engine = self.engine
            self:OnRemoved(engine)
            self:OnAdded(engine)
            self.engine:BroadcastEvent( self.engine.EVENT.CARD_UPGRADED, self, snapshot )
            -- It's pretty weird to play in the future.
            self.engine:YieldPlay( self.engine.STATUS.SYNC )
        else
            TheGame:BroadcastEvent( "card_upgraded", self )
        end
    end
end

function Card:GetXP()
    return self.userdata.xp, self:GetMaxXP()
end

function Card:GetFlavourText()
    local card_content
    local card_instance
    if is_instance( self, BattleCardDef ) or is_instance( self, NegotiationCardDef ) then
        card_content = self
    else
        card_content = self.def
        card_instance = self
    end

    if card_content.flavour_fn then
        return card_content.flavour_fn(card_instance, card_content and card_content:GetLocalizedFlavour() or "" )
    end
    return card_content and card_content:GetLocalizedFlavour()
end

-- Adds XP to a card without a notification (to be used in negotiations/battles)
function Card:AddXP(xp, ignore_fatigue )
    if xp and self.userdata.xp ~= nil and (ignore_fatigue or self:CanGainXP()) then
        local was_ready = self:UpgradeReady()
        self.userdata.xp = math.min( self.userdata.xp + xp, self:GetMaxXP())
        if self.owner and self.owner:IsPlayer() and was_ready ~= self:UpgradeReady() then
            TheGame:BroadcastEvent( "deck_upgrade_status_change")
        end

    end
end

function Card:AddRandomXP( range_start, range_end )
    local max_xp = self:GetMaxXP()
    if max_xp then
        local xp = math.floor( math.randomGauss( range_start, range_end ) * max_xp )
        self:AddXP( xp )
    end
end

function Card:RemoveXP(xp)
    if xp and self.userdata.xp ~= nil and self:CanGainXP() then
        local was_ready = self:UpgradeReady()
        self.userdata.xp = math.max( (self.userdata.xp or 0) - xp, 0)
        if self.owner and self.owner:IsPlayer() and was_ready ~= self:UpgradeReady() then
            TheGame:BroadcastEvent( "deck_upgrade_status_change")
        end
    end
end

function Card:ClearXP()
    -- Prevents this card from showing/gaining XP, without actually preventing it from upgrading if it can.
    self.userdata.xp = nil
end

function Card:GetTitle() -- REMOVE ME! Why do I exist!
    return self:GetName()
end

function Card:GetCharges()
    return self.userdata.charges, self.userdata.max_charges or self.max_charges
end

function Card:ConsumeCharge()
    if self.userdata.charges then
        self.userdata.charges = math.max( 0, self.userdata.charges - 1 )
    end
end

function Card:IsSpent()
    local charges, max_charges = self:GetCharges()
    return charges == 0 and max_charges ~= nil
end

function Card:IsPartiallySpent()
    local charges, max_charges = self:GetCharges()
    return charges and max_charges and charges < max_charges
end

function Card:RestoreCharges( count )
    local new_charges
    if count and self.userdata.charges then
        new_charges = math.min( self.userdata.charges + count, self.userdata.max_charges or self.max_charges )
    elseif count == nil then
        new_charges = self.userdata.max_charges or self.max_charges
    end

    if new_charges ~= self.userdata.charges then
        self.userdata.charges = new_charges
        if self.deck then
            self:NotifyChanged( "restore_charges" )
        end
    end
end

function Card:CanGainXP()
    return true
end

function Card:GetSummaryData()
    local data = { id = self.id, userdata = {} }
    for k, v in pairs(self.userdata) do
        if type(v) ~= "table" or not v._class then
            data.userdata[k] = v
        end
    end
    return data
end

function Card:Save()
    return { id = self.id, userdata = self.userdata, owner = self.owner }
end

function Card:PostLoad( game_state )
    -- ccc: START save-game backwards compatability (merge wild_picks table and upgrade_ids)
    do
        if self.userdata and self.userdata.wild_picks then
            self.userdata.upgrade_ids = self.userdata.wild_picks
            self.userdata.wild_picks = nil
            print( "Card:PostLoad", self, " converting wild_picks", tostr(self.userdata.upgrade_ids))

        elseif self.def.upgrade_ids and not self.userdata.upgrade_ids then
            self:RollUpgrades()
            print( "Card:PostLoad", self, " creating upgrade_ids", tostr(self.userdata.upgrade_ids))       
        end
    end
    -- ccc: END save-game backwards compatability

    if self.userdata.charges and not (self.userdata.max_charges or self.max_charges) then
        -- Max charges were removed from this item.
        self.userdata.charges = nil
    end

    if self.userdata and self.userdata.upgrade_ids then
        for i, upgrade_id in ipairs( self.userdata.upgrade_ids ) do
            if self:LookupCardDef( upgrade_id ) == nil then
                print( "Card:PostLoad", self, " sanitizing upgrade_id: ", upgrade_id )
                self:RollUpgrades()
                break
            end
        end
    end

    if self.userdata and self.userdata.hatch_id then
        -- flags aren't save-loaded, so if a card has dynamically assigned hatch data, we need to add the HATCH flag.
        self:SetFlags( self.CARD_FLAGS.HATCH )
    end

    if self:GetMaxXP() == nil then
        self.userdata.xp = nil
    end
end

function Card:RollUpgrades()
    if self.def.upgrade_ids then       
        local upgrade_ids = {}
        local picks = shallowcopy( self.def.upgrade_ids )
        local rng = TheGame:GetGameState() and TheGame:GetGameState():GetRNG("CARD_UPGRADES")
        for i = 1, 2 do
            if #picks > 0 then
                if rng then
                    table.insert(upgrade_ids, table.remove(picks, rng:Random(#picks)))
                else
                    table.insert(upgrade_ids, table.remove(picks, math.random(#picks)))
                end
            end
        end
        self.userdata.upgrade_ids = upgrade_ids

    else
        self.userdata.upgrade_ids = nil
    end

    TheGame:BroadcastEvent( "roll_upgrades", self )
end

---------------------------------------------------------------------
-- Container for a set of cards.

local Deck = class( "CardEngine.Deck" )
Deck.can_reload = true

function Deck:init( deck_type, is_event_handler )
    self.deck_type = deck_type
    self.cards  = {}
    self.is_event_handler = is_event_handler
end

function Deck:IsEventHandler()
   return self.is_event_handler
end

function Deck:GetDeckType()
    return self.deck_type
end

function Deck:SetMaxSize( max_size )
    self.max_size = max_size
end

function Deck:GetMaxSize()
    return self.max_size or math.huge
end

function Deck:GetCards()
    return shallowcopy( self.cards )
end

function Deck:Cards()
    return ipairs( self.cards )
end

function Deck:Peek( offset )
    offset = offset or 1
    return self.cards[ #self.cards - (offset - 1) ]
end

function Deck:PickCardIf( fn )
    local t = ObtainWorkTable()
    for i, card in ipairs( self.cards ) do
        if fn == nil or fn( card ) then
            table.insert( t, card )
        end
    end
    local card = table.arraypick( t )
    ReleaseWorkTable( t )
    return card
end

function Deck:PickCardsIf( fn, count )
    local t = ObtainWorkTable()
    for i, card in ipairs( self.cards ) do
        if fn == nil or fn( card ) then
            table.insert( t, card )
        end
    end
    local cards = table.multipick( t, count )
    ReleaseWorkTable( t )
    return cards
end

function Deck:PeekRandom( fn )
    if fn then
        local tmp = ObtainWorkTable()
        table.arrayadd( tmp, self.cards )
        local card
        while card == nil and #tmp > 0 do
            card = table.remove( tmp, math.random( #tmp ) )
            if not fn( card ) then
                card = nil
            end
        end
        ReleaseWorkTable( tmp )
        return card
    else
        return table.arraypick( self.cards )
    end
end

function Deck:IsEmpty()
    return #self.cards == 0
end

function Deck:CountCards( card_id )
    local count = 0
    for i, card in ipairs( self.cards ) do
        if card_id == nil or card.id == card_id then
            count = count + 1
        end
    end
    return count
end

function Deck:CountCardsIf( pred )
    local count = 0
    for i, card in ipairs( self.cards ) do
        if pred == nil or pred( card ) then
            count = count + 1
        end
    end
    return count
end

function Deck:RemoveCard( card )
    local i = table.arrayfind( self.cards, card )
    assert( i, card )
    table.remove( self.cards, i )
    card.deck = nil

    self.engine:BroadcastEvent( self.engine.EVENT.CARD_MOVED, card, self, i)

    card:OnRemoved( self.engine )
end

function Deck:InsertCard( card, i )
    if #self.cards >= self:GetMaxSize() and card.deck ~= self then
        -- Full! insertion fails.
        self.engine:BroadcastEvent( self.engine.EVENT.DECK_TOO_LARGE, card, self )
        return false
    
    else
        local old_deck = card.deck
        local old_idx = old_deck and table.arrayfind( old_deck.cards, card )

        if old_deck == self and (i == nil or old_idx == i or old_idx + 1 == i) then
            -- these are no-ops: moving the card from the same deck (this one) to the same index it already occupies.
            -- this is considered success, though no CARD_MOVED, etc. should occur.
            return true
        end

        local new_idx = i or (#self.cards + 1)
        if old_deck == self and old_idx < new_idx then
            new_idx = new_idx - 1
        end

        if old_idx == nil then
            card:OnAdded( self.engine )
            self.engine:BroadcastEvent( self.engine.EVENT.CARD_DEALT, card, self )

        else
            table.remove( old_deck.cards, old_idx )
        end

        card.deck = self
        table.insert( self.cards, new_idx, card )

        self.engine:BroadcastEvent( self.engine.EVENT.CARD_MOVED, card, old_deck, old_idx, self, new_idx )

        return true
    end
end

function Deck:TransferCard( card, deck, i )
    if card.deck == self then
        return deck:InsertCard( card, i )
    end
end

function Deck:TransferCards( deck )
    self:TransferCardsIf( deck, nil )
end

function Deck:TransferCardsIf( deck, fn )
    assert( is_instance( deck, Deck ))
    assert( deck ~= self )

    -- This multi-pass process is necessary to be as robust as possible to deck
    -- modifications during iteration  since card transferring can trigger behaviour.
    -- Specifically:
    -- (1) resilient to card removal from this deck
    -- (2) new cards that are added should get a chance to be transferred
    -- (3) any individual card is transferred at MOST once, to prevent infinite looping.

    local cards = ObtainWorkTable()
    local done
    repeat
        done = true
        for i = 1, #self.cards do
            local card = self.cards[i]
            if not cards[ card ] and (fn == nil or fn( card )) then
                cards[ card ] = true
                self:TransferCard( card, deck )
                done = false
                break
            end
        end
    until done

    ReleaseWorkTable( cards )
end

function Deck:HasCard( card )
    return table.contains( self.cards, card )
end

function Deck:FindCard( card )
    return table.arrayfind( self.cards, card )
end


function Deck:FindCardByID( card_id )
    for i = 1, #self.cards do
        if self.cards[i].id == card_id then
            return self.cards[i]
        end
    end
end

function Deck:FindCardByIndex( i )
    return self.cards[ i ]
end

function Deck:Shuffle()
    table.shuffle( self.cards )
end

function Deck:__tostring()
    return string.format( "%s.%s (%d cards)", self._classname, self.deck_type, #self.cards )
end
