---------------------------------------------------------------------------
-- Handles card picking

local CardPicker = class( "CardEngine.CardPicker" )
CardPicker.can_reload = true

function CardPicker:init( behaviour, engine )
    self.behaviour = behaviour
    self.engine = engine

    self.weighted_options = {} -- May of card -> weight
    self.max_uses = {} -- Map of card -> max_uses
    self.picks = {}
end

function CardPicker:Clear()
    table.clear( self.weighted_options )
    table.clear( self.picks )
    self.current_options = nil
end

function CardPicker:FindCard( v )
    for card, wt in pairs( self.weighted_options ) do
        if card == v or card.id == v then
            return card
        end
    end
end

function CardPicker:AddCard( card, wt, max_use )
    assert( is_instance( card, CardEngine.Card ))
    assert( type(wt) == "number" )

    self.weighted_options[ card ] = wt
    self.max_uses[ card ] = max_use
    self.current_options = nil

    return self
end

function CardPicker:AddArgument( modifier_id, wt, max_use )
    local card = self.behaviour:AddArgument( modifier_id )
    self:AddCard( card, wt, max_use )
    return self, card

end

function CardPicker:AddID( card_id, wt, max_use )
    -- assert( Content.GetNegotiationCard( card_id ), tostring(card_id) )
    assert( type(wt) == "number" )

    local card = self.behaviour:AddCard( card_id )
    self:AddCard( card, wt, max_use )

    return self, card
end

function CardPicker:RemoveCard( card )
    self.weighted_options[ card ] = nil
    self.max_uses[ card ] = nil
    if self.current_options then
        self.current_options[ card ] = nil
    end
end

function CardPicker:AddWeight( card, wt )
    assert( type(wt) == "number" )

    if type(card) == "string" then
        card = self:FindCard( card )
    end

    if card then
        self.weighted_options[ card ] = (self.weighted_options[ card ] or 0) + wt
        self.current_options = nil
    end

    return self
end

function CardPicker:ChooseCard( id )
    local card = id and self:FindCard( id )
    if card and self.behaviour:ChooseCard( card ) then
        return 1
    else
        return self:ChooseCards( 1 )
    end
end

function CardPicker:ChooseCards( count )
    table.clear( self.picks )

    local max_tries = table.count( self.weighted_options )
    for i = 1, max_tries do

        if #self.picks >= count then
            break
        end

        if self.current_options == nil or next(self.current_options) == nil then
            self.current_options = shallowcopy( self.weighted_options )
        end

        -- weighted-pick from current_options to find the next valid card.
        local tmp = ObtainWorkTable()
        shallowcopy( self.current_options, tmp )
        local card = weightedpick( tmp )
        while card do
            if not table.contains( self.picks, card ) and self.behaviour:ChooseCard( card ) then
                table.insert( self.picks, card )

                self.current_options[ card ] = self.current_options[ card ] - 1
                if self.current_options[ card ] == 0 then
                    self.current_options[ card ] = nil
                end
                if self.max_uses[ card ] then
                    self.max_uses[ card ] = self.max_uses[ card ] - 1
                    if self.max_uses[ card ] <= 0 then
                        self:RemoveCard( card )
                    end
                end

                break
            else
                tmp[ card ] = nil
                -- print( "INVALID PICK:", card )
                card = weightedpick( tmp )
            end
        end
        if card == nil then
            -- No valid picks! try again with a reset current_options
            -- print( "NO VALID PICKS! RESETTING" )
            self.current_options = nil
        end
        ReleaseWorkTable( tmp )
    end

    return #self.picks
end
