--------------------------

local function ValidateFighterData( id, fighter_data )
    assert( type(fighter_data) == "table", id )
    assert( type(fighter_data.MAX_HEALTH) == "number", id )
    fighter_data.MAX_HEALTH = math.round(fighter_data.MAX_HEALTH)
    local battle_defs = require "battle/battle_defs"


    if fighter_data.conditions then
        for condition_id, t in pairs( fighter_data.conditions ) do
            Content.AddBattleCondition( condition_id, t )
        end
        fighter_data.conditions = nil
    end

    -- These provide combat cards for this agent if they fight for the player.
    if fighter_data.attacks then
        if fighter_data.cards == nil then
            fighter_data.cards = {}
        end

        for k, attack in pairs( fighter_data.attacks ) do
            assert( type(k) == "string", id )
            attack.cost = attack.cost or 1
            attack.flags = (attack.flags or 0) | battle_defs.CARD_FLAGS.NPC
            attack.rarity = CARD_RARITY.BASIC
            attack.series = CARD_SERIES.NPC
            
            Content.AddBattleCard( k, attack )

            table.insert_unique( fighter_data.cards, k )
        end
        fighter_data.attacks = nil
    end
end

--------------------------


local CharacterDef = class("CharacterDef" )

function CharacterDef:init(id, data)
	self.id = id
	for k, v in pairs(data) do
		self[k] = v
	end

    assert( self.custom_name == nil, id ) -- custom_name is run-time, user-edited.

    if self.unique and not self.alias then
        self.alias = id
    end

    self.on_inits = { data.on_init }

    if self.loved_graft and self.loved_graft ~= "" then
        local graft = Content.GetGraft( self.loved_graft )
        assert( graft, self.loved_graft )
        assert( graft.is_good == nil or graft.is_good == true, graft.id )
        graft.is_good = true
    end
    if self.hated_graft and self.hated_graft ~= "" then
        local graft = Content.GetGraft( self.hated_graft )
        assert( graft, self.hated_graft )
        assert( graft.is_good == nil or graft.is_good == false, graft.id )
        graft.is_good = false
    end

    if self.negotiation_data then
        local negotiation_defs = require "negotiation/negotiation_defs"

        if self.negotiation_data.modifiers then
            for i, id, def in sorted_pairs( self.negotiation_data.modifiers ) do
                Content.AddNegotiationModifier( id, def )
            end
        end

        if self.negotiation_data.defs then
            for id, carddef in pairs(self.negotiation_data.defs) do
                carddef.flags = (carddef.flags or 0) | negotiation_defs.CARD_FLAGS.OPPONENT
                carddef.cost = 0
                carddef.rarity = CARD_RARITY.UNIQUE
                carddef.series = carddef.series or CARD_SERIES.NPC
                Content.AddNegotiationCard( id, carddef )
            end
        end
    end

	if self.fight_data then
		ValidateFighterData(id, self.fight_data )
	end

    local loc_name = self.name and Content.LookupString( self:GetLocNameKey() )
    if loc_name then
        local loc_translator = require "util/loc_translator"
        self.name, self.name_declensions = loc_translator.parseNameDeclensions( loc_name )
    end
end

function CharacterDef:InheritBaseDef()
    if self.base_def and self.base_ref == nil then
        local def = Content.GetCharacterDef( self.base_def )
        assert( def, self.base_def )

        def:InheritBaseDef()

        for k, v in pairs( def ) do
            if k == "is_template" or k == "base_ref" or k == "child_refs" then
                -- Template status not inherited.

            elseif k == "promotion_def" then
                -- promoted status not inherited

            elseif self[k] == nil then
                self[k] = v
            
            elseif k == "tags" then
                table.arrayadd( self[k], v ) -- Merge tags by default.

            elseif k == "aspects" then
                for aspect_name, aspect_data in pairs(v) do
                    if self[k][aspect_name] == nil then
                        self[k][aspect_name] = aspect_data
                    end
                end
            end
        end
        for i, fn in ipairs( def.on_inits ) do
            table.insert( self.on_inits, fn )
        end

        self.base_ref = def

        if def.child_refs == nil then
            def.child_refs = {}
        end
        table.insert( def.child_refs, self )
    end

    if self.fight_data and self.fight_data.cards then
        for i = #self.fight_data.cards, 1, -1 do
            local card_id = self.fight_data.cards[ i ]
            if not Content.GetBattleCard( card_id ) then
                LOGWARN( "Missing card_id '%s'", card_id )
                table.remove( self.fight_data.cards, i )
            end
        end
    end
    if self.death_item then
        local card_id = self.death_item
        if not Content.GetBattleCard( card_id ) and not Content.GetNegotiationCard( card_id ) then
            LOGWARN( "Missing death_item card '%s'", card_id )
            self.death_item = nil
        end
    end

    if self.support_negotiation == "untrained_pet" then
        local suffix = self.id:gsub("_UPGRADED", ""):lower() 
        local modifier_id = string.format( "untrained_pet_plus_%s", suffix )
        local modifier =
        {
            name = "Untrained Pet",
            icon = string.format( "negotiation/modifiers/untrained_%s.tex", suffix ),
            sound = string.format("event:/sfx/battle/cards/neg/create_argument/unruly_%s", suffix ),
        }
        Content.AddNegotiationModifier( modifier_id, modifier )
    end
    if self.trained_support_negotiation == "trained_pet" then
        local suffix = self.id:gsub("_UPGRADED", ""):lower() 
        local modifier_id = string.format( "trained_pet_plus_%s", suffix )
        local modifier =
        {
            name = "Trained Pet",
            icon = string.format( "negotiation/modifiers/trained_%s.tex", suffix ),
            sound = string.format("event:/sfx/battle/cards/neg/create_argument/unruly_%s", suffix ),
        }
        Content.AddNegotiationModifier( modifier_id, modifier )
    end

    local is_sentient = self.species == nil or SPECIES_PROPS[ self.species ].sentient
    if not self.is_template and not self.unique and self.faction_id ~= PLAYER_FACTION and self.faction_id ~= MONSTER_FACTION and is_sentient then
        if self.loved_graft == nil then
            printf( "Missing loved_graft for '%s'", self.id )
        end
        if self.hated_graft == nil then
            printf( "Missing hated_graft for '%s'", self.id)
        end
    end
end


function CharacterDef:SetModID( mod_id )
    self.mod_id = mod_id
end

function CharacterDef:GetModID()
    return self.mod_id
end

function CharacterDef:GetContentID()
    return self.id
end

function CharacterDef:GetLocPrefix()
	return "CHARACTER." .. string.upper(self.id)
end

function CharacterDef:GetLocNameKey()
	return self:GetLocPrefix() .. ".NAME"
end

function CharacterDef:GetLocTitleKey()
	return self:GetLocPrefix() .. ".TITLE"
end

function CharacterDef:GetLocDescKey()
    return self:GetLocPrefix() .. ".DESC"
end

function CharacterDef:GetLocBioKey()
	return self:GetLocPrefix() .. ".BIO"
end

function CharacterDef:GetLocNickNameKey()
    return self:GetLocPrefix() .. ".NICKNAME"
end

function CharacterDef:GetLocalizedName( agent )
    if self.name then
        return loc.format( self.name, agent )
    end
end

function CharacterDef:GetLocalizedNameWithDeclension( agent, case )
    local result = nil
    if self.name_declensions and self.name_declensions[ case ] then
        result = loc.format( self.name_declensions[ case ], agent )
    else
        result = self:GetLocalizedName( agent )
    end

    return result
end

function CharacterDef:GetLocalizedNameDeclensions()
    return self.name_declensions or table.empty
end

function CharacterDef:GetLocalizedDesc( agent )
    local result = nil
    if self.desc then
        result = loc.format( LOC(self:GetLocDescKey()), agent )
    end
    return result
end

function CharacterDef:GetLocalizedTitle( agent )
	local result = nil
    if self.title_fn then
        result = self.title_fn()
	elseif self.title then
        result = loc.format( LOC(self:GetLocTitleKey()), agent )
	end
	return result
end

function CharacterDef:GetLocalizedBio( agent )
	local result = nil
	if self.bio then
		result = loc.format( LOC(self:GetLocBioKey()), agent )
	end
	return result
end

function CharacterDef:GetLocalizedNickName( agent )
    local result = nil
    if self.nickname then
        result = loc.format( LOC(self:GetLocNickNameKey()), agent )
    end
    return result
end

function CharacterDef:GetLocalizedString( key )
    return LOC(self:GetLocPrefix() .. "." .. key:upper())
end

function CharacterDef:HarvestString(t, key, str)
    if str == nil then
        str = self[ key ]
    end
    if type(str) == "string" then
        key = self:GetLocPrefix() .. "." .. key:upper()
        assert_warning( t[ key ] == nil, "Duplicate loc key: %s", key )
        t[ key ] = str
    else
        LOGWARN( "Missing key '%s'", key )
    end
end

function CharacterDef:HarvestStrings(t)
	if self.name then
        -- NOTE: english has no declension for proper names, so we do not bother to harvest name_declensions.
		t[self:GetLocNameKey()] = self.name
	end
	if self.title then
		t[self:GetLocTitleKey()] = self.title
	end
	if self.bio then
		t[self:GetLocBioKey()] = self.bio
	end
    if self.nickname then
        t[self:GetLocNickNameKey()] = self.nickname
    end
    if self.loc_strings then
        for key, str in pairs( self.loc_strings ) do
            self:HarvestString( t, key, str )
        end
    end
end




