local lyaml = require "lyaml"
local lpeg = require "lpeg"

-- Shared lpeg patterns for both re-use and less redundant allocations.
------------------------------------------------------------------------

local open = lpeg.P("<>") + lpeg.P( "<" * (1 - lpeg.S("/")) * (1 - lpeg.S(">"))^0 * ">" )
local close = lpeg.P( "</>" )
local lpeg_locale = lpeg.locale()
local lpeg_patterns = 
{
	-- lpeg is not inherently unicode friendly, so alnum always fails for multi-byte codepoints, regardless what the current locale is.
	-- not_alnum = lpeg.P( 1 - (lpeg.locale().alnum + lpeg.S("_"))),
	not_alnum = lpeg.P( lpeg_locale.punct + lpeg_locale.space + lpeg_locale.cntrl ),
	b1 = lpeg.P{ "{" * ((1 - lpeg.S"{}") + lpeg.V(1))^0 * "}" },
	b2 = lpeg.P{ open * (lpeg.V(1) + (1 - close))^0 * close },
}

-------------------------------------------------------------------------------

local CodexEntry = class( "CodexEntry" )

function CodexEntry:GetLocalizedTitle()
	local prefix = string.format( "CODEX.%s.TITLE", self.name:upper() )
	return LOC( prefix )
end

function CodexEntry:GetLocalizedBody()
	local prefix = string.format( "CODEX.%s.BODY", self.name:upper() )
	return LOC( prefix )
end

-------------------------------------------------------------------------------
-- A glorified string lookup table for lore and other informational content.

local Codex = class( "Codex" )

function Codex:init()
	self.db = {} -- Mapping of name -> entry
	self.patterns = {} -- Flat array of all lpeg pattern-matching substitutions
end

function Codex:SetModID( mod_id )
	self.mod_id = mod_id
end

function Codex:GetModID()
	return self.mod_id
end

function Codex:AddFilename( filename )
	local id = filename:match( "([%w_]+)[.]yaml$" )

    local text = engine.inst:LoadTextFile( filename )
    assert( text, filename )
    
    local ok, yaml = pcall( lyaml.load, text )
    if not ok then
    	LOGWARN( "Failed parsing: %s\n%s", filename, tostring(yaml))
    end

	if ok then
		self:AddTable( yaml )
	end
end

-- Add a table of entries.
function Codex:AddTable( entries )

    for name, entry in pairs( entries ) do
		assert( type(name) == "string", "Entry names must be strings" )
		assert( self.db[ name ] == nil, "Duplicate entry: '"..name.."'" )
		
		assert( type(entry.title) == "string", "Missing title")
		assert( type(entry.body) == "string", "Missing body")

		-- Add the lookup by name.
		setmetatable( entry, CodexEntry )
		self.db[ name ] = entry
		entry.name = name

		-- Load localized trigger string if it exists.
		local prefix = string.format( "CODEX.%s.TRIGGER", entry.name:upper() )
		local str = TheGame:LookupStr( prefix )
		local triggers
		if str then
			triggers = str:split("|")
		elseif type(entry.trigger) == "string" then
			triggers = entry.trigger:split("|")
		elseif type(entry.trigger) == "table" then
			triggers = entry.trigger
		end
		for i, trigger in ipairs(triggers) do
			local prefix = string.format( "CODEX.%s.TRIGGER%d", entry.name:upper(), i )
			local trigger_name
			if TheGame:LookupStr( prefix ) then
				trigger_name = LOC( prefix )
			else
				trigger_name = trigger
			end
			self:AddPatternSubstitution( trigger_name, entry.name )
		end
	end
end

local function PatternCompare( p1, p2 )
	return p1.sort_order > p2.sort_order
end

function Codex:AddPatternSubstitution( p, name )
	local not_alnum, b1, b2 = lpeg_patterns.not_alnum, lpeg_patterns.b1, lpeg_patterns.b2
	local substr = string.format( "<!%s>%%1</>", name )
	local cp = (lpeg.C(lpeg.P(p)) / substr) -- The capture (trigger) string
	local w = cp * (not_alnum + lpeg.P(-1)) -- Capture string followed only by punctuation, or the end.
	local w_internal = lpeg.B(not_alnum) * w -- ...preceded by punctuation.
	local seq = b1 + b2 + w_internal + 1
	local pattern = lpeg.Cs( w^0 * seq^0 )

	table.binsert( self.patterns, { pattern = pattern, sort_order = #p }, PatternCompare )
end

-- Substitution formatting for codex entries.
-- WARNING: expensive!  Scales linearly with the number of codex entries.
function Codex:SubstituteFormatting( txt )
    for i, v in ipairs( self.patterns ) do
		txt = v.pattern:match( txt ) or txt
	end
	return txt
end

-- Lookup a specific entry.
function Codex:LookupEntry( name )
	return self.db[ name ]
end

function Codex:HarvestStrings( t )
	for key, entry in pairs( self.db ) do
		local prefix = string.format( "CODEX.%s", key:upper() )

		if type(entry.title) == "string" then
			-- Title: the title of this codex entry, optional.
			t[ prefix .. ".TITLE" ] = entry.title
		end
		
		if type(entry.body) == "string" then
			-- Body: The body of this codex entry, optional.
			t[ prefix .. ".BODY" ] = entry.body
		end

		-- Trigger words: the pattern-matched text options.
		if type(entry.trigger) == "string" then
			t[ prefix .. ".TRIGGER" ] = entry.trigger
		else
			for i, v in ipairs( entry.trigger ) do
				local nprefix = string.format( "%s.TRIGGER%d", prefix, i )
				t[ nprefix ] = v
			end
		end
	end
end
