require "component"
require "util/heap"
require "components/fxtrack"

local AnimController = class("AnimController", Component)

AnimController.Priority = 3

function AnimController:init( rotation )
    AnimController._base.init(self)
    self.anims = {}
    self.scale = 1
    self.rotation = rotation
    self.flip = false
    self.animnode = engine.rendering.SceneGraphNode()
    self.animnode:SetRotation(self.rotation or -FIGHT_CAM_ANGLE, 0, 0)
    self.animmodel = engine.model.AnimModel()
    self.animnode:SetModel(self.animmodel)
    self.animmodel:SetLayer(RENDER_LAYER_WORLD)
    
    if TheGame:GetLocalSettings().DEBUG_ANIMS then
        --self:SetRenderBB(true)
        self.anim_history = {}
    end
    self.anim_speeds = {}
    self.subanims = {}
end

function AnimController:SetOutline(size, colour)
    self.animmodel:SetOutlineSize(size)    
    local r,g,b,a = HexColour(colour)
    if not self.outline_colour_constant then
        self.outline_colour_constant = engine.rendering.ShaderConstantValue("OUTLINE_COLOUR", r, g, b, a)
        self.animmodel:SetShaderConstant(self.outline_colour_constant)
    else
        self.outline_colour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end
end

function AnimController:HideOutline()
    self.animmodel:SetOutlineSize(0)    
end

function AnimController:GetSymbolWorldPos(sym, x, y, z)
    return self.animmodel:SymbolToWorld(sym, x or 0, y or 0, z or 0)
end

function AnimController:GetSymbolPos(sym, x, y, z)
    return self.animmodel:SymbolToLocal(sym, x or 0, y or 0, z or 0)
end

function AnimController:SetScale(scale)
    self.scale = scale
    -- self.animnode:SetScale(scale*(self.flip and -1 or 1),scale,scale)
    self.entity:SetLocalScale( scale*(self.flip and -1 or 1), scale, scale )
end

function AnimController:SetXFlip(flip)
    self.flip = flip
    local scale = self.scale or 1.0
    -- self.animnode:SetScale(self.scale*(self.flip and -1 or 1),self.scale,self.scale)
    self.entity:SetLocalScale( scale*(self.flip and -1 or 1), scale, scale )
end

function AnimController:GetAnimFlip()
    local anim_flip = false
    local anim_info = self.anims[ self.animName ]
    local flip_frames = anim_info and anim_info.meta and anim_info.meta.flip_frames
    if flip_frames then
        local frame = self.animmodel:GetFrame() + 1
        for i, flip_frame in ipairs( flip_frames ) do
            if flip_frame.frame > frame then
                break
            else
                anim_flip = flip_frame.is_flipped
            end
        end
    end
    return anim_flip
end

function AnimController:GetXFlip()
    return self.flip == true
end

function AnimController:IsFacingRight()
    local anim_flip = self:GetAnimFlip()
    return (self.flip and anim_flip) or (not self.flip and not anim_flip)
end    

function AnimController:OnAddedToEntity()  
    self.entity.cmp.Transform:AddChildNode(self.animnode)
    self.animmodel:SetGUID(self.entity:GetGUID())

    self.animmodel:SetEffect( global_shaders.UI_ANIM )

    self.animmodel:SetOutlineEffect(global_shaders.ANIM_OUTLINE)
    end

function AnimController:Reset()
    self.animmodel:Reset()
    table.clear(self.anims)
end

function AnimController:Clear()
    if self.task then
        self.task:Cancel()
        self.task = nil
    end
    self.animmodel:Clear()

    if self.anim_track then
        self.anim_track:KillLinkedFX()
        self.anim_track = nil
    end

    return self
end

function AnimController:OnDespawn()
    self:Clear()
    self.animmodel = nil
end

function AnimController:GetAnimInfo(anim)
    return self.anims[anim or self.animName]
end

function AnimController:GetAnimName()
    return self.animName
end

function AnimController:ClearAnims()
    table.clear( self.anims )
end

function AnimController:RegenerateAnimInfo( name )
    local anim_info = self.anims[ name ]
    local key = GetAnimMetaFilename( anim_info.anim )
    anim_info.meta = GetAnimMeta( key, name )
end


function AnimController:AddAnims(anim)
    local key = GetAnimMetaFilename( anim )
    for k = 1, anim:GetNumAnims() do
        local name = anim:GetAnimName(k-1)
        local meta = GetAnimMeta( key, name )
        local anim_info = self.anims[ name ]
        if anim_info == nil then
            anim_info = {}
            self.anims[ name ] = anim_info
        end

        anim_info.anim = anim
        anim_info.idx = k - 1
        anim_info.meta = meta
    end
    return self
end

function AnimController:SetTime(t)
    local time = (self.walltime and GetWallTime()) or self.entity.sim:GetTime()

    return self.animmodel:SetStartTime(time + t)
end

function AnimController:GetDuration()
    return self.animmodel:GetDuration()
end

function AnimController:SetBuild(build)
    self.build = build
    self.animmodel:SetBuild(build)
    return self
end

function AnimController:SetWallTime(val)
    self.walltime = val
    self.animmodel:SetWallTime(val)
end

function AnimController:SetPhase(v)
    self.phase = clamp(v,0,1)
    if self.start_time then
        local dur = self.animmodel:GetDuration()
        self.animmodel:SetStartTime(self.start_time - self.phase*dur)
    end
    return self
end

function AnimController:SetAnimSpeed(anim, v)
    self.anim_speeds[ anim ] = v
    if self.animName == anim then
        self.animmodel:SetPlaySpeed(v)
        if self.task then
            self.task:Cancel()

            local speed = self:GetPlaySpeed()
            self.animmodel:GetBoundingBox() -- Force update buffers.
            local delay = (self:GetDuration() - self:GetFrame() * FRAMES) / speed
            if delay > 0 then
                self.task = self.entity:DoTaskInTime("anim", delay, self.OnAnimDone, self)
            end
        end
        if self.anim_track then
            self.anim_track:Kill()
        end
    end
    return self
end

function AnimController:GetPlaySpeed(anim)
    anim = anim or self.animName
    return self.anim_speeds[ anim ] or self.speed or 1.0
end

function AnimController:SetSpeed(v)
    self.speed = v
    self.animmodel:SetPlaySpeed(v)
    return self
end

function AnimController:HasAnim(anim)
    return self.anims[anim] ~= nil
end

function AnimController:StopAnim()
    self.looping = false
    if self.anim_queue then
        table.clear( self.anim_queue )
    end

    if not self:IsAnimDone() then
        self:OnAnimDone()
    end
end

function AnimController:StartAnim(anim, looping, speed, time_offset)
    local animdat = self.anims[anim]
    
    if animdat then
        local offset = time_offset or 0
        self.speed = speed or 1
        assert(self.speed >= 0, "can't play anims backwards")
        self.animdone = false
        self.looping = looping == true
        self.animName = anim
        self.animmodel:SetLooping( self.looping )
        local time = (self.walltime and GetWallTime()) or self.entity.sim:GetTime()
        self.animmodel:SetStartTime(time + offset)
        self.start_time= time + offset
        
        self.animmodel:SetAnim(self.anims[anim].anim, self.anims[anim].idx)
        local speed = self:GetPlaySpeed( anim )
        self.animmodel:SetPlaySpeed(speed)

        local delay = math.max(0, (self.animmodel:GetDuration() - (time_offset or 0)) / speed)
        if self.task then
            self.task:Cancel()
        end
        if delay > 0 then
            self.task = self.entity:DoTaskInTime("anim", delay, self.OnAnimDone, self)
        else
            self:OnAnimDone()
        end

        if self.anim_track then
            self.anim_track:KillLinkedFX()
        end

        if animdat.meta and (animdat.meta.fx_tracks ~= nil or animdat.meta.sound_frames ~= nil) then
            if self.anim_track == nil then
                -- FXTrack associated with the anim meta data of the current animation.
                local fx_track = FXTrack( anim, animdat.meta, self )
                fx_track:SetTrackTags( self.fx_track_tags )
                fx_track:AdvanceTrack( 0.0 )
                self.anim_track = fx_track

            else
                self.anim_track:SetTrackTags( self.fx_track_tags )
                self.anim_track:Reset( anim, animdat.meta )
                self.anim_track:AdvanceTrack( 0.0 )
            end

        else
            if self.anim_track then
                self.anim_track:KillLinkedFX()
                self.anim_track = nil
            end
        end

        if self.missinganim then
            
            self.missinganim = nil
            if self.defaulttint then 
            	self:SetTintMult(self.defaulttint.r, self.defaulttint.g, self.defaulttint.b, self.defaulttint.a)
            else
            	self:SetTintMult(1,1,1,1)
            end
        end

        --this can be really really slow!
        if self.anim_history then
            table.insert( self.anim_history, 1, debug.traceback() )
            table.insert( self.anim_history, 1, os.clock() )
            table.insert( self.anim_history, 1, anim )
            while #self.anim_history > 3 * 10 do
                table.remove( self.anim_history )
            end
        end
    else
        -- print( "MISSING ANIM:", anim, debug.traceback() )
        --LOGWARN("Missing animation state: %s", anim)
        if self.anim_history then
            table.insert( self.anim_history, 1, debug.traceback() )
            table.insert( self.anim_history, 1, os.clock() )
            table.insert( self.anim_history, 1, string.format( "MISSING: %s", anim ))
        end
        self.missinganim = anim
        self.animdone = true
        self:SetTintMult(1,0,0,1)
    end

end

function AnimController:IsAnimDone()
    return self.animdone
end

function AnimController:IsAnimQueued()
    return self.anim_queue and #self.anim_queue > 0
end

function AnimController:Push(anim, loop)
    if self.looping or self.animdone then
        self:Play(anim, loop, 1, 0) --just play it directly if we are at the end of an anim or looping
    else
        self.anim_queue = self.anim_queue or {}
        local next_anim
        if loop then
            next_anim = { anim = anim, loop = loop }
        else
            next_anim = anim
        end
        table.insert(self.anim_queue, next_anim)
    end
    return self
end

function AnimController:Play(anim, looping, speed, time_offset)
    -- Abort the entire queue if there is one.
	if self.anim_queue and #self.anim_queue > 0 then
	    table.clear(self.anim_queue)
    end

    -- Playing the same anim that is already looping is a nop.
    -- But playing the same non-looping anim should start it anew.
    if anim ~= self.animName or not self.looping or (self.looping and not looping) then
        return self:StartAnim(anim, looping, speed, time_offset)
    end
    return self
end

function AnimController:ListenForAnim( fn, listener )
    self.entity:ListenForLocalEvent( "anim_done", fn, listener )
    self.listener_count = (self.listener_count or 0) + 1
end

function AnimController:StopAnimListening( listener )
    self.entity:StopListeningToLocalEvent( "anim_done", listener )
    self.listener_count = self.listener_count - 1
    assert( self.listener_count >= 0 )
end

function AnimController:OnAnimDone()

    if self.task then
        self.task:Cancel()
        self.task = nil
    end

    self.animdone = true

    if (self.listener_count or 0) > 0 then
        self.entity:PushLocalEvent("anim_done", self.animName, self.token)
    end

    if self.anim_queue and #self.anim_queue > 0 then
        local next_anim = table.remove( self.anim_queue, 1 )
        local loop = false

        if type(next_anim) == "table" then
            loop = next_anim.loop
            next_anim = next_anim.anim
        end

        self:StartAnim(next_anim, loop, 1, 0)

    elseif self.looping then

        if self.anim_track then
            self.anim_track:Reset()
        end

        local speed = self:GetPlaySpeed( self.animName )
        local delay = self.animmodel:GetDuration() / speed
        if delay > 0 then
            self.task = self.entity:DoTaskInTime("anim", delay, self.OnAnimDone, self)
        end
    end
end

function AnimController:SetLayer(layer)
    self.layer = layer
    self.animmodel:SetLayer(layer)
end

function AnimController:SetFXTrackTags(tag_list)
    self.fx_track_tags = tag_list
end

function AnimController:SetFXTrackFn( fn )
    self.fx_track_fn = fn
end

function AnimController:OnRender( dt )
    if self.anim_track then
        self.anim_track:UpdateTransforms()
        self.anim_track:AdvanceTrack( self.animmodel:GetFrame() )
        if not self.anim_track:IsValid() and not self.looping then
            self.anim_track = nil
        end
    end
end

function AnimController:IsOnScreen()
    return self.animmodel:IsOnScreen(self.entity.sim:GetCameraInfo())
end

function AnimController:RayTest(x,y, ignore_hashes)
    local x1, y1, z1 = self.animnode:ScreenToLocal( self.sim:GetCameraInfo(), x, y )
    if ignore_hashes then
        
        local hashes = ObtainWorkTable()
        local hit = self.animmodel:TestPoint(x1, y1, hashes)
        if hit then
            
            for k,v in ipairs(hashes) do
                
                if not table.arrayfind(ignore_hashes, v) then
                    --use this when you need to figure out the crazy exporter-derived names for glowing subsymbols
                    --[[local strs = {}
                    for k,v in ipairs(hashes) do
                        table.insert(strs, engine.inst:ReverseHashString(v))
                    end
                    print (table.concat(strs, " "))
                    --]]
                    ReleaseWorkTable(hashes)
                    return true
                end
            end
            
            ReleaseWorkTable(hashes)
            return false
        else
            ReleaseWorkTable(hashes)
            return false
        end
        return hit
    else
        return self.animmodel:TestPoint(x1, y1)
    end

    -- if self.animmodel:IsOnScreen(self.entity.sim:GetCameraInfo()) then
    --     return self.animmodel:RayTest(self.entity.sim:GetCameraInfo(), x, y)
    -- end
end

function AnimController:GetBoundingBox()
    local x, y, z = self.entity:GetLocalPosition()
    local x1, y1, x2, y2 = self.animmodel:GetBoundingBox()
    if x1 == nil then
        return 0, 0, 0, 0, 0, 0
    else
        local z1, z2 = z, z
        x1, y1, z1 = self.animnode:LocalToWorld( x1, y1, z1 )
        x2, y2, z2 = self.animnode:LocalToWorld( x2, y2, z2 )
        return x1, y1, z1, x2, y2, z2
    end
end

function AnimController:SetShaderConstant(shader_constant)
    self.animmodel:SetShaderConstant(shader_constant)
    self.constants = self.constants or {}
    self.constants[shader_constant:GetName()] = shader_constant    
end

function AnimController:ClearShaderConstant(name)
    self.animmodel:ClearShaderConstant(name)
end

function AnimController:SetSymbolShaderConstant(name, shader_constant)
    self.animmodel:SetSymbolShaderConstant(name, shader_constant)
    self.symbol_constants = self.symbol_constants or {}
    self.symbol_constants[name] = shader_constant    
end

function AnimController:ClearSymbolShaderConstant(name)
    self.animmodel:ClearSymbolShaderConstant(name)
    if self.symbol_constants then
        self.symbol_constants[name] = nil
    end
end

function AnimController:SetBlendMode(mode)
    self.animmodel:SetBlendMode(mode)
end

function AnimController:SetDepthState(depthtest, depthwrite)
    self.animmodel:SetDepthState(depthtest, depthwrite)
end

function AnimController:GetTimeInAnim()
    return self.entity.sim:GetTime() - (self.start_time or 0)
end

function AnimController:GetFrame()
    return self.animmodel:GetFrame()
end

function AnimController:SetBloom(b)
    self.animmodel:SetBloom(b)
end

function AnimController:SetDefaultTintMult(r, g, b, a)
    self.defaulttint = { r = r, g = g, b = b, a = a}
    self:SetTintMult( r, g, b, a )
end

function AnimController:TintColour(r, g, b, a)
    self:SetTintMult(r, g, b, a )
end

function AnimController:SetTintMult(r, g, b, a)
    if self.missinganim then
        r, g, b, a = 1, 0, 0, 1
    end

    if not self.tint_mult_constant then
        self.tint_mult_constant = engine.rendering.ShaderConstantValue("TINT_MULT", r, g, b, a)
        self.animmodel:SetShaderConstant(self.tint_mult_constant)
    else
        self.tint_mult_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end
end

function AnimController:SetTintAdd(r, g, b, a)
    if not self.tint_add_constant then
        self.tint_add_constant = engine.rendering.ShaderConstantValue("TINT_ADD", r, g, b, a)
        self.animmodel:SetShaderConstant(self.tint_add_constant)
    else
        self.tint_add_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end
end

function AnimController:HideSymbol(name)
    self.animmodel:SetSymbolVisible(name, false)
end

function AnimController:ShowSymbol(name)
    self.animmodel:SetSymbolVisible(name, true)
end

function AnimController:GetAnimFrameLength()
    return self.animmodel:GetAnimFrameLength()
end

function AnimController:HasRenderBB()
    return self.animmodel:HasTag(engine.rendering.SceneGraphModel.TAG_DRAW_BOUNDING_BOX)
end

function AnimController:SetRenderBB(do_it)
    if do_it then
        self.animmodel:SetTag(engine.rendering.SceneGraphModel.TAG_DRAW_BOUNDING_BOX)
        if TheGame.local_settings.DEBUG_ANIM_QUADS then
            self.animmodel:SetTag(engine.rendering.SceneGraphModel.TAG_DRAW_ANIM_QUADS)
        end
    else
        self.animmodel:ClearTag(engine.rendering.SceneGraphModel.TAG_DRAW_BOUNDING_BOX)
        if TheGame.local_settings.DEBUG_ANIM_QUADS then
            self.animmodel:ClearTag(engine.rendering.SceneGraphModel.TAG_DRAW_ANIM_QUADS)
        end
    end
end

function AnimController:SetSymbolTint(symbol, r,g,b,a)
    self.animmodel:SetSymbolTint(symbol, r, g, b, a)
end

function AnimController:AddOverrideFile(anim_file_name, symbol)
    self.animmodel:AddOverrideFile(anim_file_name, symbol)
end

function AnimController:RemoveOverrideFile(anim_file_name)
    self.animmodel:RemoveOverrideFile(anim_file_name)
end

function AnimController:SetLayerTint(layer, r,g,b,a, slot)
    self.animmodel:SetLayerTint(layer, r, g, b, a, slot)
end

function AnimController:CoroWaitForDoneAnims()
    while (self.anim_queue and #self.anim_queue > 0) or not self.animdone do
        if self.missinganim then
            return
        end 
        coroutine.yield()
    end
end

function AnimController:CoroWaitForAnim()
    local anim_name = self.animName
    while self.animName == anim_name and not self.animdone do
        coroutine.yield()
    end
end


--- THis stuff is common(ish) with the anim widget. Should probably centralize it, eventually

function AnimController:PushSubAnim(symbol, anim, loop_at_end)
    
    if not self.ondonesubtask and not self.ondonesubtask[symbol] then
        self:PlaySubAnim(symbol, anim, loop_at_end, 0)
        return
    end

    self.subqueue = self.subqueue or {}
    self.subqueue[symbol] = self.subqueue[symbol] or {}
    self.loop_at_end_of_subqueue = self.loop_at_end_of_subqueue or self.loop_at_end_of_subqueue
    self.loop_at_end_of_subqueue[symbol] = loop_at_end
    table.insert(self.subqueue[symbol], anim)
end

function AnimController:ClearSubAnim(symbol)
    if self.subqueue and self.subqueue[symbol] then
        table.clear(self.subqueue[symbol])
    end

    if self.ondonesubtask and self.ondonesubtask[symbol] then
        self.ondonesubtask[symbol]:Cancel()
        self.ondonesubtask[symbol] = nil
    end

    self.animmodel:ClearSymbolAnim(symbol)
end


function AnimController:PlaySubAnim(symbol, anim, looping, timeoffset, hold)
    --print (symbol, anim, looping, timeoffset) 
    --print (self.subanims, self.subanims[symbol], self.subanims[symbol] and self.subanims[symbol][anim])
    if self.subanims and self.subanims[symbol] and self.subanims[symbol][anim] then
        
        local animdat = self.subanims[symbol][anim] 
        self.ondonesubtask = self.ondonesubtask or {}

        if self.ondonesubtask[symbol] then
            self.ondonesubtask[symbol]:Cancel()
            self.ondonesubtask[symbol] = nil
        end

        if self.subqueue and self.subqueue[symbol] then
            table.clear(self.subqueue[symbol])
        end
        
        if animdat then
            
            if self.animmodel then
                --print ("PLAY SUBANIM", symbol, anim)
                self.animmodel:PlaySymbolAnim(symbol, animdat.anim, animdat.idx, looping or false, timeoffset or 0)
            end

            if not looping and not hold then
                self.ondonesubtask[symbol] = self.entity:DoTaskInTime("SUBDONE", animdat.duration, AnimController.OnSubAnimDone, self, animdat, symbol)
            end

        else
            print("Missing UI Sub anim " .. anim)
        end
    else
        print ("NO SUBANIMS BY THAT NAME" )
    end
    return self
end


function AnimController:AddSymbolAnim(symbol, anim_name, anim_asset, idx)
    self.subanims = self.subanims or {}
    self.subanims[symbol] = self.subanims[symbol] or {}
    self.subanims[symbol][anim_name] = {anim=anim_asset, idx = idx, name=anim_name, duration=anim_asset:GetAnimDuration(idx)}
end

function AnimController:OnSubAnimDone(animdat, symbol)
    if self.removed then
        return
    end
    self.ondonesubtask[symbol] = nil
    if self.subqueue and self.subqueue[symbol] and #self.subqueue[symbol] > 0 then
        local anim = table.remove( self.subqueue[symbol], 1)
        self:PlaySubAnim(anim, self.loop_at_end_of_subqueue and self.loop_at_end_of_subqueue[symbol] and #self.subqueue[symbol] == 0)
    else
        self:ClearSubAnim(symbol)
    end
end

function AnimController:RenderDebugAnimInfo( ui, panel )
    if not self.animmodel then
        return
    end
    local animname, animfile = self.animmodel:GetAnimName()
    ui:Text( string.format( "Anim: %s (%s)\n", animname, animfile ))
    local info = self.anims[ animname ]
    if info and info.meta then
        ui:SameLine( nil, 10 )
        panel:AppendTable( ui, info.meta, "Anim meta-data" )
    end

    ui:Text( string.format( "Looping: %s\n", tostring(self.looping) ))
    ui:Text( string.format( "Sim Time: %g\n", self.sim:GetTime() ))
    ui:Text( string.format( "Frame: %d / %d\n", self.animmodel:GetFrame()+1, self.animmodel:GetAnimFrameLength() ))
    ui:Text( string.format( "Build: %s\n", self.build and self.build:GetFileName() or "nil"))
    if self.anim_queue and #self.anim_queue > 0 then
        ui:Text( string.format( "Queue: %s\n", tostr(self.anim_queue)))
    end
    ui:Text( string.format( "Done: %s\n", tostr(self.animdone)))
    if ui:Checkbox("flip", self.flip) then
        self:SetXFlip( not self.flip )
    end

    if self.missinganim then
        ui:TextColored( 0xFF0000FF, string.format( "Missing Anim: '%s'", self.missinganim ))
    end
end

function AnimController:RenderDebugUI( ui, panel )
    self:RenderDebugAnimInfo( ui, panel )

    if self.anim_history then
        ui:Text( "History:" )
        local now = os.clock()
        for i = 1, #self.anim_history, 3 do
            ui:PushID( tostring(i) )
            local anim, st, trace = self.anim_history[i], self.anim_history[i+1], self.anim_history[i+2]
            local show = ui:TreeNode( anim )
            ui:SameLine( nil, 10 )
            ui:Text( string.format( "(-%.1f dt)", now - st ))
            if show then
                ui:TextColored( 0x00CCCCFF, trace )
                ui:TreePop()
            end
            ui:PopID()
            now = st
        end
        
        if ui:Button("Clear History") then
            table.clear(self.anim_history)
        end

    else
        if ui:Button("Create History") then
            self.anim_history = {}
        end
    end

    ui:Separator()
    
    panel:AppendKeyValues( ui, self )
end


function AnimController:SetSelected(enabled)
    self.animmodel:SetSelected(enabled)
end

function AnimController:SetSelectedLayerSpeeds(layer1Speed, layer2Speed, layer3Speed)
    self.animmodel:SetSelectedLayerSpeeds(layer1Speed, layer2Speed, layer3Speed)
end

function AnimController:SetSelectedLayerScale(scale)
    self.animmodel:SetSelectedLayerScale(scale)
end

function AnimController:SetSelectedLayerAlphas(layer1Alpha, layer2Alpha, layer3Alpha)
    self.animmodel:SetSelectedLayerAlphas(layer1Alpha, layer2Alpha, layer3Alpha)
end

function AnimController:SetSelectedFadeDuration(fadeDuration)
    self.animmodel:SetSelectedFadeDuration(fadeDuration)
end
function AnimController:SetSelectedColour(colour)
    local r,g,b,a = HexColour(colour)
    self.animmodel:SetSelectedColor(r or 1, g or 1, b or 1, a or 1)
end





function AnimController:SetGlowThresholdAndAmount( threshold, amount)
    self.animmodel:SetGlowThresholdAndAmount(threshold, amount);
end

function AnimController:SetGlowColour(colour)
    local r,g,b,a = HexColour(colour)
    self.animmodel:SetGlowColour(r, g, b, a);
end


function AnimController:SetXRayOutline(enabled, fillColour, lineColour, thresholdMin, thresholdMax)
    if enabled then
        self.animmodel:SetEffect(global_shaders.UI_ANIM_XRAY)
        self.animmodel:SetSelectedEffect(global_shaders.UI_ANIM_XRAY_SELECTED)

        local minthreshold = math.min( thresholdMin or 0.05, thresholdMax or 0.1)
        local maxthreshold = math.max( thresholdMin or 0.05, thresholdMax or 0.1)

        if not self.xray_threshold_constant then
            self.xray_threshold_constant = engine.rendering.ShaderConstantValue("XRAY_THRESHOLD", minthreshold, maxthreshold)
            self.animmodel:SetShaderConstant(self.xray_threshold_constant)
        else
            self.xray_threshold_constant:SetValue( minthreshold, maxthreshold )
        end

        local r,g,b,a = HexColour(fillColour)
        if not self.xray_fillcolour_constant then
            self.xray_fillcolour_constant = engine.rendering.ShaderConstantValue("XRAY_FILL_COLOUR", r, g, b, a)
            self.animmodel:SetShaderConstant(self.xray_fillcolour_constant)
        else
            self.xray_fillcolour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
        end

        r,g,b,a = HexColour(lineColour)
        if not self.xray_linecolour_constant then
            self.xray_linecolour_constant = engine.rendering.ShaderConstantValue("XRAY_LINE_COLOUR", r, g, b, a)
            self.animmodel:SetShaderConstant(self.xray_linecolour_constant)
        else
            self.xray_linecolour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
        end
    else
        self.animmodel:SetEffect(global_shaders.UI_ANIM)
    end

end


-- local parameters =
-- {
--     scale = -1.5, 
--     colour = 0x9900ffff,
--     alpha = {0.7, 0.3, 2.0}, -- Alpha for each layer of the FX
--     speed = {0.05, 0.07, 0.5},
--     glitchSpeed = 0.07, 
--     glitchScale = 6.0,
--     layersTexture = engine.asset.Texture("rgba/character_hover.tex"),
--     glitchTexture = engine.asset.Texture("glitch/glitch.tex"),
-- }
function AnimController:SetHologram(enabled, parameters)
    self.animmodel:SetHologramEnabled(enabled)
    if enabled then
        local model = self.animmodel

        model:SetHologramGlitchSpeedAndScale(parameters.glitchSpeed, parameters.glitchScale)
        model:SetHologramGlitchTexture(parameters.glitchTexture)

        model:SetHologramLayerSpeeds(parameters.speed[1], parameters.speed[2], parameters.speed[3])
        model:SetHologramLayerScale(parameters.scale)
        model:SetHologramLayerAlphas(parameters.alpha[1], parameters.alpha[2], parameters.alpha[3])
        model:SetHologramLayerTexture(parameters.layersTexture)

        local r,g,b,a = HexColour(parameters.colour)
        model:SetHologramLayerColor(r, g, b, a)


        model:SetHologramHueSatBriCon(parameters.colormatrixHue[1], parameters.colormatrixHue[2], parameters.colormatrixHue[3], parameters.colormatrixSaturation, parameters.colormatrixBrightness, parameters.colormatrixContrast)


        -- only create she shaderconstants if an hsv table is supplied. Once an hsvtable is supplied, keep setting the values every frame after that.
        local hsvtable = parameters.hsvtable

        if hsvtable then
            if not self.colour_correct_constants then
                self.colour_correct_constants = {}
                self.colour_correct_constants.shadows = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST", 0,1,1)
                self.colour_correct_constants.midtones = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST", 0,1,1)
                self.colour_correct_constants.highlights = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST", 0,1,1)
                self.colour_correct_constants.shadows_col = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST_COL", 1,1,1)
                self.colour_correct_constants.midtones_col = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST_COL", 1,1,1)
                self.colour_correct_constants.highlights_col = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST_COL", 1,1,1)

                model:SetShaderConstant(self.colour_correct_constants.shadows)
                model:SetShaderConstant(self.colour_correct_constants.midtones)
                model:SetShaderConstant(self.colour_correct_constants.highlights)
                model:SetShaderConstant(self.colour_correct_constants.shadows_col)
                model:SetShaderConstant(self.colour_correct_constants.midtones_col)
                model:SetShaderConstant(self.colour_correct_constants.highlights_col)
            end
        end

        if self.colour_correct_constants then
            if hsvtable then
                self.colour_correct_constants.shadows:SetValue(table.unpack(hsvtable.shadows.hsv))
                self.colour_correct_constants.midtones:SetValue(table.unpack(hsvtable.midtones.hsv))
                self.colour_correct_constants.highlights:SetValue(table.unpack(hsvtable.highlights.hsv))
                self.colour_correct_constants.shadows_col:SetValue(table.unpack(hsvtable.shadows.colour))
                self.colour_correct_constants.midtones_col:SetValue(table.unpack(hsvtable.midtones.colour))
                self.colour_correct_constants.highlights_col:SetValue(table.unpack(hsvtable.highlights.colour))
            else
                self.colour_correct_constants.shadows:SetValue(0,1,1)
                self.colour_correct_constants.midtones:SetValue(0,1,1)
                self.colour_correct_constants.highlights:SetValue(0,1,1)
                self.colour_correct_constants.shadows_col:SetValue(1,1,1)
                self.colour_correct_constants.midtones_col:SetValue(1,1,1)
                self.colour_correct_constants.highlights_col:SetValue(1,1,1)
            end
        end

    end
end
