require "component"
local paperdoll = require "paperdoll_util"
local sounds = require "sound_event_names"
local battle_defs = require "battle/battle_defs"

local RUN_DIST = 10
local RUN_SPEED = 40
local DODGE_KNOCKBACK = 2.5

-- This will override all hit reactions to the specified anim.-
-- Assign to 'nil' to disable
local HIT_OVERRIDE = "hit_mid"

-- Scaling factor applied to all AnimFighters.
-- Individual scaling can be assigned through the character def 'scale_adjust' property.
local FIGHTER_SCALE = 1.23

local FIGHT_STATUS_WIDGET_OFFSET = -0.9 -- Default offset.  Customize via fight_data.
local FIGHT_STATUS_WIDGET_HEAD_YOFFSET = 5.0 -- Default head yoffset.
local FIGHT_STATUS_WIDGET_HEAD_XOFFSET = 1.0 -- Default head xoffset.

-- Default attack distance and depth offset for melee attacks.
local DEFAULT_ATTACK_DIST = 5.0
local DEFAULT_ATTACK_DEPTH = 0.05

-- Anim outline tuning
local OUTLINE_WIDTH = 3
local SELECTED_OUTLINE_WIDTH = 8

local DEFAULT_SHADOW_SCALE = 7.5

-- run_forward_pst_turn kicks in at this amount of time prior to the end of movement.
local PST_TURN_DURATION = (3*FRAMES)

-----------------------------------------------------------------

local IDLE_ANIMS =
{
    [ IDLE_STATE.NEUTRAL ] = { priority = 1, loop = "idle" },
    [ IDLE_STATE.AIMING ] = { priority = 2, loop = "aim_loop", },
    [ IDLE_STATE.SURRENDER ] = { priority = 3, loop = "surrender", sfx = {colour = 0x00ffffff} },
    [ IDLE_STATE.STUNNED ] = { priority = 4, loop = "stunned", sfx = {colour = 0xffff00ff} },
    [ IDLE_STATE.DEAD ] = { priority = 5, anim = "dead" },
}

local ANIM_TRANSITIONS =
{
    hit_mid =
    {
        idle = "hit_mid_pst_idle",
        stunned = "hit_mid_pst_stunned",
        surrender = "hit_mid_pst_surrender",
    },
    hit_low =
    {
        idle = "hit_low_pst_idle",
        stunned = "hit_low_pst_stunned",
        surrender = "hit_low_pst_surrender",
    },
    hit_back =
    {
        idle = "hit_back_pst_idle",
        stunned = "hit_back_pst_stunned",
        surrender = "hit_back_pst_surrender",
    },
    hit_up =
    {
        idle = "hit_up_pst_idle",
        stunned = "hit_up_pst_stunned",
        surrender = "hit_up_pst_surrender",
    },
    idle =
    {
        stunned = "stunned_pre",
        surrender = "surrender_pre"
    },
    stunned =
    {
        idle = "stunned_pst",
        surrender = { "stunned_pst", "surrender_pre" },
    },
    surrender =
    {
        idle = { "surrender_pst", "taunt" },
        stunned = { "surrender_pst", "surrender_pre" },
    },
    defend =
    {
        idle = "defend_pst",
        surrender = "defend_pst"
    },
}

local function MakeVOLine( actor, line )
    local actor = actor or "sal/vo_sal"
    actor = actor:match("^(.+)/.+$")
    local line = loc.format("event:/vo/combat/{1}/{2}", actor, line)
    return line
end

----------------------------------------------------------------------


local AnimFighter = class("AnimFighter", Component)
AnimFighter.can_reload = true
AnimFighter.DEFAULT_ATTACK_DIST = DEFAULT_ATTACK_DIST
AnimFighter.DEFAULT_ATTACK_DEPTH = DEFAULT_ATTACK_DEPTH

function AnimFighter:init()
    AnimFighter._base.init(self)
    self.fake_scale = 2

    self.mouth_queue = {}
    self.select_fx = {}
    self.idle_states = {} -- sorted array of applied idle states
    self.talk_lines = {}

    self.outline_width = OUTLINE_WIDTH
    self.selected_outline_width = SELECTED_OUTLINE_WIDTH

    self.fx_tracks = {}

    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function AnimFighter:GetFakeScale()
    return self.fake_scale
end

function AnimFighter:OnScreenModeChange( sm )
    -- This outline is also set on fightscreen.lua:OnScreenModeChange()
    -- It has to be set here too, since when the fightscreen is initialized, there aren't animfighters for it to change the outlines on
    if sm == SCREEN_MODE.MONITOR then
        self:SetOutlineWidth( 3 )
        self:SetSelectedOutlineWidth( 8 )
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        self:SetOutlineWidth( 7 )
        self:SetSelectedOutlineWidth( 16 )
    end
end

function AnimFighter:SetOutlineWidth( outline_width )
    self.outline_width = outline_width or OUTLINE_WIDTH
    return self
end

function AnimFighter:SetSelectedOutlineWidth( selected_outline_width )
    self.selected_outline_width = selected_outline_width or SELECTED_OUTLINE_WIDTH
    return self
end

function AnimFighter:HideTargetFX()
    if self.select_fx then
        for k,v in pairs(self.select_fx) do
            self.entity.sim.fx_mgr:KillFX(v)
        end
        table.clear(self.select_fx)
    end
end

function AnimFighter:IsEditor()
    return TheGame:FE():FindScreen( "Screen.AttackTester" ) ~= nil
end

function AnimFighter:PlayTargetFX()
    if #self.select_fx == 0 then
        local cmp = self.entity.cmp.FXContainer

        local scale = 6.0
        local ground_fx = cmp:PlayFXOnGround("anim/selected_footprint_fx.zip", "loop", 0, 0, scale, scale, 0, false, true )
        ground_fx.model:SetWallTime(true)
        table.insert(self.select_fx, ground_fx)
 
        local scale = self.fighter.team.id == TEAM.BLUE and 4 or -4
        local back_fx = cmp:PlayBillboardFX("anim/selected_fx.zip", "ground_selected_behind", 0, 0, -.05, scale, scale, scale, 0, false, true )
        back_fx.model:SetWallTime(true)
        table.insert(self.select_fx, back_fx)

        local clr = self.fighter.player_controlled and UICOLOURS.FRIENDLY_COLOUR or UICOLOURS.ENEMY_COLOUR
        local r,g,b = HexColour(clr)

        for k,v in pairs(self.select_fx) do
            v.model:SetShaderConstant(engine.rendering.ShaderConstantValue("TINT_MULT", r,g,b, 1))
            v.model:SetBloom(.5)
        end
    end
end

function AnimFighter:HideSelectFX()
    if self.holo_active then 
        self.entity.cmp.AnimController:HideOutline()
    else
        self:SetAnimOutline(self.outline_width, 0x000000ff)
    end
end

function AnimFighter:PlaySelectFX()
    local clr = (self.team == TEAM.BLUE and UICOLOURS.FRIENDLY_COLOUR or UICOLOURS.ENEMY_COLOUR)
    self:SetAnimOutline(self.selected_outline_width, clr)
end

function AnimFighter:OnAddedToEntity()
    self:StartTickUpdating()
    
    self.home_x, self.home_z = 0, 0

    self:SetAnimOutline(self.outline_width, 0x000000ff)
end

function AnimFighter:OnDespawn()
    self.sim.mgr:RemoveFighter( self.fighter )
    self:KillUpdatingFX()
end

function AnimFighter:SetFighter(fighter, battle)
    self.fighter = fighter
    self.battle = battle
    self.fight_data = fighter.agent.fight_data or table.empty
    self.current_anim_mapping = self.fight_data.anim_mapping
    self.status = FIGHT_STATUS.ACTIVE --fighter:GetStatus() (this is future peeking)
    self.team = fighter:GetTeamID()
    self.height = fighter.agent.fight_data.height or 0
    self.anim_transitions = shallowcopy(ANIM_TRANSITIONS)
    if self.fight_data.anim_transitions then
        table.add(self.anim_transitions, self.fight_data.anim_transitions)
    end

    local shadow_scale = self.fight_data.shadow_scale or DEFAULT_SHADOW_SCALE
    if shadow_scale > 0 then
        self.dropshadow = self.entity.sim:SpawnPrefab("dropshadow")
        self.entity:AddChild(self.dropshadow)
        self.dropshadow.cmp.Transform:SetPosRot(0,.01 - self.height,0)
        self.dropshadow.cmp.Transform:SetLocalScale( shadow_scale, shadow_scale, shadow_scale )

        if self.fight_data.shadow_alpha then
            self.dropshadow.cmp.PlaneRenderer:TintColour( 0, 0, 0, self.fight_data.shadow_alpha )
        end
    end

    self:ShowShield(false)

    if self.fight_data.fx then
        for name, fx_def in pairs( self.fight_data.fx ) do
            if fx_def.auto then
                self:PlayUpdatingFX( name )
            end
        end
    end

    paperdoll.SetCharacterData(self.entity.cmp.AnimController, self.fighter.agent, true, FIGHTER_SCALE)

end

function  AnimFighter:SetAnimMapping( new_mapping )
    if new_mapping then
        self.current_anim_mapping = new_mapping
    else
        self.current_anim_mapping = self.fight_data.anim_mapping
    end

    local current_anim = self.entity.cmp.AnimController:GetAnimName()
    local mapped_anim = self.current_anim_mapping and self.current_anim_mapping[ current_anim ] or current_anim
    if current_anim ~= mapped_anim then
        self:PlayAnim( current_anim, self.entity.cmp.AnimController.looping )
    end
end

function AnimFighter:AnimShield()

    local GetSineVal = function(mod, abs, time, timemod)
        local time = (time + (timemod or 0)) * (mod or 1)
        local val = math.sin(PI * time)
        if abs then
            return math.abs(val)
        else
            return val
        end
    end

    local total_time = 0

    local min_scale = 8.75
    local max_scale = 9

    local min_rot = 175
    local max_rot = 185

    while self.show_shield do
        total_time = total_time + FRAMES
        local scale_t = GetSineVal( 0.10, true, total_time )
        local rot_t = GetSineVal( 0.05, true, total_time )
        local scale = easing.linear( scale_t, min_scale, max_scale - min_scale, 1.0 )
        local rot = easing.linear( rot_t, min_rot, max_rot - min_rot, 1.0 )

        self.groundprotect.cmp.Transform:SetLocalScale( scale, scale, scale )
        self.groundprotect.cmp.Transform:SetLocalRotation( 0, rot*DEGREES, 0 )
        coroutine.yield()
    end
end

function AnimFighter:ShowShield( bool )
    if bool then
        if not self.show_shield then
            if self.dropshadow then
                self.dropshadow:Hide()
            end

            if self.groundprotect == nil then
                self.groundprotect = self.entity.sim:SpawnPrefab("groundprotect")
                self.entity:AddChild(self.groundprotect)
                self.groundprotect.cmp.Transform:SetPosRot(0,.01 - self.height,0, 0, 180*DEGREES )
                self.groundprotect.cmp.Transform:SetLocalScale( DEFAULT_SHADOW_SCALE, DEFAULT_SHADOW_SCALE, -DEFAULT_SHADOW_SCALE )
            end

            self.groundprotect:Show()
            self.show_shield = true
            self:StartCoroutine(AnimFighter.AnimShield, self)
        end

    else
        if self.show_shield then
            self.show_shield = false
            self.groundprotect:Hide()
            if self.dropshadow then
                self.dropshadow:Show()
            end
        end
    end
end

function AnimFighter:GetTeam()
    return self.team
end

function AnimFighter:StopMoving()

    self.destx, self.destz = nil, nil
    self.interp_t = nil
    self.next_x = nil
    self.last_x = nil
    self.next_y = nil
    self.last_y = nil
    self.next_z = nil
    self.last_z = nil
    self.movement_timer = nil
    self.move_ease_fn = nil
    self.force_run = nil
    self.arriving = false
    self.run_speed = nil
    self.run_speed_throttle = nil

    self.entity.cmp.AnimController:SetXFlip(self.team == TEAM.RED)
end

function AnimFighter:FaceFighter( fighter )
    local ent = self.sim.mgr:GetAnimFighterByFighter( fighter )
    if ent then
        local x0 = self.entity:GetLocalPosition()
        local tx = ent:GetLocalPosition()
        self.entity.cmp.AnimController:SetXFlip( x0 > tx )
    end
end

function AnimFighter:GetPresentation( anim )
    if self.current_anim_mapping then
        anim = self.current_anim_mapping[ anim ] or anim
    end
    return anim, self:GetMoveAnim( anim )
end

function AnimFighter:GetIdleAnim(idle_state)
    return IDLE_ANIMS[ idle_state ]
end

function AnimFighter:GetMoveAnim(anim)
    local anim_info = self.entity.cmp.AnimController:GetAnimInfo( anim )
    return anim_info and anim_info.meta or table.empty
end

function AnimFighter:IsBusy()
    if self.coro and not self.coro:IsDone() and not self.coro:IsRunning() then
        return true, self.coro
    end
    return false
end

function AnimFighter:CanInterrupt()
    return self.can_interrupt == true
end

function AnimFighter:GetDestPos()
    if self.destx then
        return self.destx, self.destz
    end

    local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
    return x, z
end

function AnimFighter:StartMoving( x, z, move_type, fx_tags, force_run )
    local mapx, _, mapz = self.entity.cmp.Transform:GetLocalPosition()
    local dx, dz = x - mapx, z - mapz
    local dist = (math.abs(dx) > 0 or math.abs(dz) > 0) and math.sqrt( dx*dx + dz*dz ) or 0

    if self.fight_data.StartMoving then
        self.fight_data.StartMoving( self, x, z, move_type, fx_tags, force_run )
        return
    end

    if self.fight_data.stationary then
        self:StopMoving()
        return
    end
    
    if dist <= 0 then
        self:StopMoving()
        return
    end

    self.arriving = false
    self.run_speed = nil
    self.force_run = force_run
    self.destx = x
    self.destz = z

    local fwd = false
    local cmp = self.entity.cmp.AnimController
    local facing_right = cmp:IsFacingRight()
    if facing_right then
        fwd = dx >= 0
    else
        fwd = dx <= 0
    end

    local vis_crunch = .5
    local vis_dist = math.sqrt(dx*dx + (dz*vis_crunch*dz*vis_crunch))
   
    self.move_start_x = mapx
    self.move_start_z = mapz
    self.move_dx = dx
    self.move_dz = dz

    cmp:SetFXTrackTags( fx_tags )

    if not move_type and (vis_dist >= RUN_DIST or force_run) then
        local adj_amt = .25
        local angle = math.atan2(dz, dx)
        local pct_throttle = (1-adj_amt) + adj_amt*math.abs(math.sin(angle))
        local speed = RUN_SPEED * pct_throttle * (self.run_speed_throttle or 1.0)

        self.run_speed = speed

        if fwd then
            self.start_move_time = 0 * FRAMES
            if cmp:GetAnimFlip() then
                cmp:SetXFlip( not cmp:GetXFlip() )
            end
            self:PlayAnim("run_forward_pre")    
        else
            self.start_move_time = 0 * FRAMES
            cmp:SetXFlip(facing_right)
            self:PlayAnim("run_forward_pre_turn")
        end
        self:PushAnim("run_forward", true )

        self.end_move_time = (dist/speed) + self.start_move_time

    else
        if move_type == nil then
            move_type = fwd and "step_forward" or "step_back"
        end

        local move_anim = self:PlayAnim( move_type )

        local start_frame = move_anim.start_move_frame or 0
        local end_frame = move_anim.end_move_frame or start_frame + 10
        self.start_move_time = start_frame * FRAMES
        self.end_move_time = end_frame * FRAMES
    end

    self.movement_timer = 0
end

function AnimFighter:UpdateMoving( dt )
    self.movement_timer = self.movement_timer + dt

    if self.movement_timer < self.start_move_time then
        -- wait until start_move_time to start moving.
    elseif self.movement_timer >= self.start_move_time and self.movement_timer <= self.end_move_time then
        -- do the movement between these times
        local time_moving = self.movement_timer - self.start_move_time
        local total_movement_time = self.end_move_time - self.start_move_time
        local ease_fn = self.move_ease_fn or easing.outQuad
        local p = clamp(ease_fn(time_moving, 0, 1, total_movement_time), 0, 1)
        self:SetNextPos( self.move_start_x + self.move_dx * p, self.height, self.move_start_z + self.move_dz * p)
    
        if self.run_speed and not self.arriving and (total_movement_time - time_moving) <= PST_TURN_DURATION then
            self.arriving = true
            local do_flip_pst = (self.entity.cmp.AnimController.flip and self.team ~= TEAM.RED) or (not self.entity.cmp.AnimController.flip and self.team ~= TEAM.BLUE)
            if do_flip_pst then
                self.entity.cmp.AnimController:SetXFlip(self.team == TEAM.RED)
                self:PlayAnim("run_forward_pst_turn")
            elseif not self.force_run then
                self:PlayAnim("run_forward_pst")
            end
        end

    else
        self.entity.cmp.Transform:SetPosRot(self.destx, self.height, self.destz)

        local current_anim = self.entity.cmp.AnimController:GetAnimName()
        if (current_anim ~= "run_forward_pst" and
            current_anim ~= "run_forward_pst_turn" and
            current_anim ~= "step_back" and
            current_anim ~= "step_forward") or self.entity.cmp.AnimController:IsAnimDone() then
            -- Done moving anims.
            self:StopMoving()
        end
    end 
end

function AnimFighter:SetNextPos(x, y, z)
    self.last_x, self.last_y, self.last_z = self.entity.cmp.Transform:GetLocalPosition()
    self.next_x = x
    self.next_y = y
    self.next_z = z
    self.interp_t = 0
end

function AnimFighter:IsSettled()
    if self:IsMoving() then
        return false, "Moving"
    end

    local facing_right = not self.entity.cmp.AnimController.flip
    if (self.team == TEAM.BLUE and not facing_right) or (self.team == TEAM.RED and facing_right) then
        return false, "Flipped"
    end

    if self:IsBusy() then
        return false, "Busy"
    end

    -- if not self:IsIdle() then
    --     return false, "Not Idle"
    -- end

    if not self:InHomePosition() then
        return false, "Not Home"
    end
    
    return true
end

function AnimFighter:GetHomePosition()
    return self.home_x, self.home_z
end

function AnimFighter:SetHomePosition( x, z )
    if self.destx == self.home_x and self.destz == self.home_z then
        self:StartMoving( x, z )
    end
    self.home_x, self.home_z = x, z
end

function AnimFighter:InHomePosition()
    if self.fight_data.stationary then
        return true
    end
    
    local cx, cy, cz = self.entity.cmp.Transform:GetLocalPosition()
    local hx, hz = self:GetHomePosition()
    local dist = distsq(hx, hz, cx, cz)
    return dist <= 0.15
end

function AnimFighter:GetPerspectiveScale( z )
    return (1+ clamp(z / FIGHT_FIELD_DEPTH, -1, 1)*FIGHT_PERSPECTIVE_SCALE)
end

function AnimFighter:OnRender( dt )
    for i, fx_track in ipairs( self.fx_tracks ) do
        fx_track:UpdateTransforms( dt )
    end
end

function AnimFighter:UpdateFX( dt )
    local i = 1
    while i <= #self.fx_tracks do
        local track = self.fx_tracks[i]
        track:UpdateElapsedTime( dt )

        if not track:IsValid() then
            table.remove( self.fx_tracks, i )
        else
            i = i + 1
        end
    end
end

function AnimFighter:OnTickUpdate(dt)
    local scale = FIGHTER_SCALE * paperdoll.GetCharacterScale( self.fighter.agent )
    scale = scale * (self.fighter:GetAnimScale())
    if self.battle then
        scale = scale * self.battle:GetScenario():GetSceneScaling()
    end

    local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
    scale = scale * self:GetPerspectiveScale( z )
    self.entity.cmp.AnimController:SetScale(scale)
    self.fake_scale = scale
    
    self.interp_t = self.interp_t or 0
    self.interp_t = self.interp_t + dt
    if self.next_x and self.last_x then
        local t = clamp( self.interp_t / SIM_TICK_TIME, 0, 1)
        local x = self.last_x + (self.next_x - self.last_x)*t
        local y = self.last_y + (self.next_y - self.last_y)*t
        local z = self.last_z + (self.next_z - self.last_z)*t
        self.entity.cmp.Transform:SetPosRot(x,y,z)
    end
    
    if self.talk_time then
        self.talk_time = self.talk_time - dt
        if self.talk_time <= 0 then
            self:SayNextLine()
        end
    end

    if self.flash then
        
        self.flash.t = self.flash.t + dt
        local r,g,b,a = HexColour(self.flash.colour)
        local intensity = 0
        if self.flash.t < self.flash.time_in then
            intensity = easing.outQuad(self.flash.t, 0, self.flash.max_intensity, self.flash.time_in)
        elseif self.flash.t < self.flash.time_in+self.flash.time_out  then
            intensity = easing.outQuad(self.flash.t - self.flash.time_in, self.flash.max_intensity, -self.flash.max_intensity, self.flash.time_out)
        else
            intensity = 0
            if self.flash.periodic then
                self.flash.t = self.flash.t -(self.flash.time_in+self.flash.time_out)
            else
                self.flash = nil
            end
        end

        self.entity.cmp.AnimController:SetTintAdd(r*intensity,g*intensity,b*intensity,0)

    end

    if #self.mouth_queue > 0 then
        local time = GetWallTime()
        if self.mouth_queue[#self.mouth_queue].time < time then
            local shp = self.mouth_queue[#self.mouth_queue].shape
            if shp then
                local anim = shp[self.mood == MOOD.HAPPY and MOOD.HAPPY or MOOD.NEUTRAL]
                if anim then
                    self:PlayMouthAnim(anim, false, true, #self.mouth_queue ~= 1 )
                else
                    self.entity.cmp.AnimController:ClearSubAnim("mouth_swap")
                end
            end
            table.remove(self.mouth_queue)
        end
    end

    if self.fight_data.shadow_symbol then
        local lx, ly, lz = self.entity.cmp.AnimController.animmodel:SymbolToLocal( self.fight_data.shadow_symbol, 0, 0, 0 )
        if self.dropshadow then
            self.dropshadow:SetLocalPosition( lx, -self.height, lz )
        end
    end

    if self:IsMoving() then
        if self.fight_data.UpdateMoving then
            self.fight_data.UpdateMoving( self, dt )
        else
            self:UpdateMoving(dt)
        end
    end

    self:UpdateFX( dt )

    self:UpdateCoroutines( dt )
end

function AnimFighter:SetPos(x, z)
    self:StopMoving()
    self.entity.cmp.Transform:SetPosRot(x,self.height,z)

    local scale = FIGHTER_SCALE * paperdoll.GetCharacterScale( self.fighter.agent )
    scale = scale * (self.fighter.agent.fight_data.battle_scale or 1.0) 

    local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
    scale = scale * self:GetPerspectiveScale( z )
    self.entity.cmp.AnimController:SetScale(scale)
    self.fake_scale = scale
end

function AnimFighter:IsMoving()
    if self.destx then
        return true
    end
    return false
end

function AnimFighter:DoAttack( attack, screen )

    if self.post_attack then
        -- Aborting a post_attack.
        if self.post_attack.card.anims or (self.post_attack.card.post_anim and self.post_attack.card.anim ~= attack.card.anim) then
            -- If there are anim overrides, then we MUST wait for Post to finish, otherwise things look bizarre.
            self:WaitForSequences()
        else
            self.can_interrupt = true
            self:AbortSequence()
        end

    elseif self:IsBusy() then
        -- We're doing something else?
        if self.can_interrupt then
            self:AbortSequence()
        else
            self:WaitForSequences()
        end
        if self.entity.removed then
            return
        end
    end

    local anim, presentation = self:GetPresentation( attack.anim )
    if not self.entity.cmp.AnimController:GetAnimInfo( anim ) then
        -- Potentially missing anim, find a sensible default.
        if CheckBits( attack.flags, battle_defs.CARD_FLAGS.MELEE ) then
            anim, presentation = self:GetPresentation( "punch" )
        elseif CheckBits( attack.flags, battle_defs.CARD_FLAGS.RANGED ) then
            anim, presentation = self:GetPresentation( "shoot" )
        else
            anim, presentation = self:GetPresentation( "taunt" )
        end
    end

    local hit_tags = attack.hit_tags
    if presentation.hit_tags then
        for i, tag in ipairs( presentation.hit_tags ) do
            table.insert_unique( hit_tags, tag )
        end
    end

    -- Move into position.
    if attack.card.anim_delay then
        self:CoroDelay(attack.card.anim_delay)
    end
    if not self:DoAttackApproach( attack, presentation ) then
        return
    end

    self:DeferConditions()

    if attack.name then
        self:ShowAttackFloater( attack.name, 0.8, UICOLOURS.HEALTH )
    end
    if attack.card.quip then
        self:SayQuip( attack.card.quip )
    end
    if attack.card.sound then
        AUDIO:PlayEvent( attack.card.sound )
    end

    self.entity.cmp.AnimController:SetFXTrackTags( hit_tags )

    -- if I am post attacking, probably want to SKIP the post_anim if possible, AND skip the pre_anim, but still need to unmap..
    if attack.card.pre_anim and not self.post_attack then
        local pre_anim, pres = self:GetPresentation( attack.card.pre_anim )
        self:PlayAnim( pre_anim )
    end

    if anim then
        if attack.card.anims then
            for k,v in pairs( attack.card.anims ) do
                self.entity.cmp.AnimController:AddOverrideFile(v) 
            end
        end
        if attack.card.pre_anim then
            self:WaitAnim()
        end
        self:PlayAnim( anim )
    end

    -- Wait for hit frames.
    local hit_frames = presentation[ attack.hit_frames ] or presentation.hit_frames or table.empty
    local t = 0
    local hit_count = attack:GetHitCount()
    local original_hit_tags = hit_tags

    for hit = 1, hit_count do

        if attack.card.PrePresAnim then
            attack.card:PrePresAnim( self )
        end

        if attack.card.pre_anim and hit > 1 then
            -- This attack has a pre/post: we need to anim for each additional hit.
            self:WaitAnim()
            self:PlayAnim( anim )

        elseif attack.card.hit_anim then
            if hit > 1 then
                self:WaitAnim()
            end
            self:PlayAnim( anim )
        end

        local hit_frame = hit_frames[ math.min( #hit_frames, hit ) ]
        local hit_time = hit_frame and hit_frame.frame*FRAMES or t
        local delay = hit_time -  t
        local dodge_pre

        if hit == 1 and table.arrayfind( hit_tags, "evaded" ) ~= nil then
            dodge_pre = math.min( delay, 0.3 )
            delay = delay - dodge_pre
        end

        if delay > 0 then
            self:CoroDelay( delay )
        end

        if hit_frame then
            if hit_frame.must_have_tags then
                for i, tag in ipairs( hit_frame.must_have_tags) do
                    table.insert_unique(hit_tags, tag)
                end
            end
            if hit_frame.cant_have_tags then
                for i, tag in ipairs( hit_frame.cant_have_tags) do
                    table.arrayremove(hit_tags, tag)
                end
            end
        end

        if attack.card:IsAttackCard() and hit == 1 and not dodge_pre and screen then
            screen:ShowHit( self.fighter, hit_tags )
        end
        
        for i, t in attack:Hits() do
            if t.hit_index == hit then
                local target = t.target
                -- Target presentation: target anim, hit FX, ...
                local ent = self.sim.mgr:GetAnimFighterByFighter( target )
                if ent then
                    local cmp = ent.cmp.AnimFighter
                    cmp:DoSequence( AnimFighter.DoHitResult, cmp, attack, t, presentation, hit_tags, hit_frames, screen )
                end
            end
        end

        if attack.card.pre_anim and hit_count > 1 then
            -- Since we are replaying the anim per hit, the hit_time just resets.
            self:WaitAnim()
            t = 0
        elseif attack.card.hit_anim then
            self:WaitAnim()
            t = 0
        else
            t = hit_time
        end
    end

    -- PostAttack is finished in a coroutine, so that the main thread doesn't need to
    -- wait for stupid things like the Fighter returning home, etc.
    self.post_attack = nil

    if attack.card and attack.card.synchronous then
        self:DoPostAttack( attack )
    else
        self:DoSequence( self.DoPostAttack, self, attack )
    end
end

function AnimFighter:DoPostAttack( attack )
    self.post_attack = attack

    self:WaitAnim()

    -- Unmap anim overrides.
    if attack.card.anims then
        for k,v in pairs( attack.card.anims ) do
            self.entity.cmp.AnimController:RemoveOverrideFile(v) 
        end
    end

    if attack.card.post_anim then
        self:WaitAnim()
        self:PlayAnim( attack.card.post_anim )
        self:WaitAnim()
        self:PlayIdleAnim()
        self:ShowDeferredConditions()
    end

    if attack.card.PostPresAnim then
        attack.card:PostPresAnim( self )
    end

    if attack.riposted then
        -- If we're flipped due to attack anim, remain flipped until we return home.
        if self.entity.cmp.AnimController:GetAnimFlip() then
            self.entity.cmp.AnimController:SetXFlip( not self.entity.cmp.AnimController:GetXFlip() )
        end

    elseif not attack.card.post_anim then
        self:ResetFighter( easing.outSine )
    end

    self.post_attack = nil
end


function AnimFighter:DoAttackApproach( attack, presentation )
    local target_ent 
    for i, hit in attack:Hits() do
        target_ent = self.sim.mgr:GetAnimFighterByFighter( hit.target )
        if target_ent then
            break
        end
    end

    if target_ent == nil then
        -- This can happen if something during PRE_RESOLVE killed our target. The
        -- attack is already underway at that point, so we must validate here.
        return false
    end

    if CheckBits( attack.flags, battle_defs.CARD_FLAGS.MELEE ) then

        local x0, y0, z0 = self.entity:GetLocalPosition()

        local tx, ty, tz = target_ent.cmp.Transform:GetLocalPosition()
        local target_fighter = target_ent.cmp.AnimFighter.fighter

        -- This modifiers the target position directly.
        if target_fighter.fight_data.approach_distance then
            if tx > x0 then
                tx = tx - target_fighter.fight_data.approach_distance
            else
                tx = tx + target_fighter.fight_data.approach_distance
            end
        end

        local prev_mapping = self.current_anim_mapping
        if attack.card.anim_mapping then
            self:SetAnimMapping( attack.card.anim_mapping )
        end
        local force_run = attack.card.force_run == nil or attack.card.force_run ~= false
        if presentation.waypoints == nil then
            local dz = DEFAULT_ATTACK_DEPTH
            local dx = DEFAULT_ATTACK_DIST * self:GetPerspectiveScale( tz + dz ) * (x0 > tx and 1 or -1)
            self:StartMoving( tx + dx, tz + dz, nil, nil, force_run )
            self:CoroWaitForArrive() 
            self:FaceFighter( target_fighter )

        else
            for i, t in ipairs( presentation.waypoints or table.empty ) do
                local waypoint_dz = t[2]
                local waypoint_dx = t[1] * self:GetPerspectiveScale( tz + waypoint_dz ) * (x0 > tx and 1 or -1)
                self:StartMoving( tx + waypoint_dx, tz + waypoint_dz, nil, nil, force_run )
                self:CoroWaitForArrive()
                self.entity.cmp.AnimController:SetXFlip( x0 > tx )
            end
        end

        if attack.card.anim_mapping then
            self:SetAnimMapping( prev_mapping )
        end
 
    elseif CheckBits( attack.flags, battle_defs.CARD_FLAGS.RANGED ) then
        if attack.card.lined_up then
            local x0, y0, z0 = self.entity:GetLocalPosition()
            local tx, ty, tz = target_ent.cmp.Transform:GetLocalPosition()
            local target_fighter = target_ent.cmp.AnimFighter.fighter
            local force_run = attack.card.force_run == nil or attack.card.force_run ~= false
            self:StartMoving( x0, tz, nil, nil, force_run )
            self:CoroWaitForArrive() 
            self:FaceFighter( target_fighter )
        else
            self:ReturnHomeCoro()
        end
    end

    return true
end

function AnimFighter:DoHitResult( attack, hit, presentation, hit_tags, hit_frames, screen )
    local wx, wy, wz = self:GetStatusWidgetPosition()
    local x, y = self.sim:GetCamera():WorldToScreen( wx, wy, wz )
    
    local width = TheGame:FE():GetScreenDims()
    local screen_pos = math.abs(x)/width
    local pan_pos = easing.linear( screen_pos, -1, 2, 1 )


    local total_damage = 0
    if hit.deltas then
        for i, ev in ipairs( hit.deltas ) do
            -- Because these are extracted from the event stream, these events haven't had
            -- a chance to be propogated through the mod API.
            Content.InvokeModAPI( "OnBattleEvent", screen, table.unpack( ev ))

            local fighter, stat, delta, value, mitigated, new_cap, delta_cap = table.unpack( ev, 2 )
            local ent = self.sim.mgr:GetAnimFighterByFighter( fighter )
            if ent then
                if stat == COMBAT_STAT.HEALTH and delta < 0 then
                    total_damage = total_damage + math.abs(delta)
                end
                if ent.cmp.AnimFighter.status_widget then
                    ent.cmp.AnimFighter.status_widget:ShowDeltaStat( stat, delta, value, mitigated, new_cap, delta_cap, attack )
                end
            end
        end
    end

    if type(hit.dodge) == "table" then
        local fighter, source, attacker = table.unpack( hit.dodge, 2 )
        local ent = self.sim.mgr:GetAnimFighterByFighter( fighter )
        if ent and ent.cmp.AnimFighter.status_widget then
            ent.cmp.AnimFighter.status_widget:ShowTextFloater( LOC "UI.FIGHT_SCREEN.EVADED", 24 )
        end
    end

    if hit.conditions then
        for i, ev in ipairs( hit.conditions ) do
            -- Because these are extracted from the event stream, these events haven't had
            -- a chance to be propogated through the mod API.
            Content.InvokeModAPI( "OnBattleEvent", screen, table.unpack( ev ))

            local event_name, fighter, condition, stacks, source = table.unpack( ev )
            local ent = self.sim.mgr:GetAnimFighterByFighter( fighter )
            if ent then
                if event_name == battle_defs.EVENT.CONDITION_ADDED then
                    ent.cmp.AnimFighter:AddCondition( condition, stacks, source )
                elseif event_name == battle_defs.EVENT.CONDITION_REMOVED then
                    ent.cmp.AnimFighter:RemoveCondition( condition, stacks, source )
                end
            end
        end
    end

    if hit.editor_damage then
        total_damage = hit.editor_damage
    end

    if (attack.card:IsAttackCard() or total_damage > 0) and (hit_tags == nil or not table.arrayfind( hit_tags, "evaded" )) then
        self:Flash(0xFFB4A3ff, 0, .2, 1)
    end

    local damage_type = attack.card.damage_type
    if damage_type == nil and attack.attacker.fight_data.hit_sounds then
        damage_type = attack.attacker.fight_data.hit_sounds[ attack.card.anim ]
    end

    if attack.card:IsAttackCard() or total_damage > 0 or damage_type ~= nil then
        if damage_type == nil then
            damage_type = 1
        end

        local intensity = -total_damage
        if intensity == 0 then
            intensity = -10 -- This is an arbitrary intensity for sound purposes.
        end

        -- print(attack.card, total_damage, hit.editor_damage, damage_type)
        if type(damage_type) == "number" then
            if self.team == TEAM.BLUE then
                AUDIO:PlayParamEvent(SoundEvents.battle_health_change_friendly, "intensity", intensity, "damage_type", damage_type - 1 )
            else
                AUDIO:PlayParamEvent(SoundEvents.battle_health_change_enemy, "intensity", intensity, "damage_type", damage_type - 1 )
            end
        end
    end

    self.entity.cmp.AnimController:SetFXTrackTags( hit_tags )

    local new_status = attack.status_changed and attack.status_changed[ self.fighter ]
    if new_status == FIGHT_STATUS.DEAD and attack:IsLastHitForTarget( hit ) then
        -- Death is special: skip the hit reaction and go directly to dead.
        self:OnStatusChanged( new_status )

    elseif attack.card:IsAttackCard() or total_damage > 0 then
        local move_type, move_dist
        if hit.defended then
            move_type = "defend"
            AUDIO:PlayParamEvent(SoundEvents.battle_attack_defended, "position", pan_pos)
            -- self.entity.cmp.SoundEmitter:PlayEvent( sounds.battle_event_dodge )

        elseif hit.evaded then
            if hit.hit_index == 1 then
                if self.entity.cmp.AnimController:HasAnim( "dodge" ) then
                    move_type = "dodge"
                else
                    move_type = "step_back"
                    move_dist = DODGE_KNOCKBACK
                end
            end
        else
            move_type = HIT_OVERRIDE or hit_frames[hit].type or "hit_mid"
        end

        self.can_interrupt = true

        if move_type then
            local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
            local hit_dist = move_dist or hit_frames[hit.hit_index] and hit_frames[hit.hit_index].dist
            if hit_dist and not self.fighter.knockback_immune then
                local destx, destz = x, z
                destx = destx + hit_dist*(self.team == TEAM.BLUE and -1 or 1)
                self:StartMoving(destx, destz, move_type, hit_tags)
                self:CoroWaitForArrive()
            else
                self:PlayAnim( move_type )
                self:WaitAnim()
            end
        end

        if new_status and attack:IsLastHitForTarget( hit ) then
            self:OnStatusChanged( new_status )

        else
            -- In case a previous hit result was a knockback and interrupted, we must
            -- be sure to return home.
            if not hit.evaded or attack:IsLastHitForTarget( hit ) then
                self:PlayIdleAnim()
                self:ShowDeferredConditions()
                -- self:ReturnHomeCoro()
            end
        end

    elseif new_status and attack:IsLastHitForTarget( hit ) then
         -- buffer time between attack & resultant status change on target
         -- (eg rallying cry)
        self:CoroDelay( 30 * FRAMES )

        self:OnStatusChanged( new_status )

    elseif attack.card.target_anim then
        self.can_interrupt = true

        if attack.card.target_anim then
            self:PlayAnim( attack.card.target_anim )
        end
        -- if not self:IsIdle() then
            self:WaitAnim()
        -- end

        self:PlayIdleAnim()
        self:ShowDeferredConditions()

    else
        self:ShowDeferredConditions()
    end

    self.entity.cmp.AnimController:SetFXTrackTags( nil )
end

function AnimFighter:ShowAttackFloater( attack_name, duration, colour )
    if self.status_widget then self.status_widget:ShowAttackFloater( attack_name, duration, colour ) end
end

function AnimFighter:ResetFighter( move_ease, delay )
    if delay then self:CoroDelay(delay) end
    self:ReturnHomeCoro(move_ease)
    self:PlayIdleAnim()
    self:ShowDeferredConditions()
end

function AnimFighter:ReturnHomeCoro( move_ease )
    if not self:InHomePosition() then
        local x, z = self:GetHomePosition()
        self.move_ease_fn = move_ease
        self:StartMoving(x, z)
        self:CoroWaitForArrive()
    else
        self:StopMoving()
    end
end

function AnimFighter:Flash(colour, time_in, time_out, max_intensity, periodic)
    self.flash = self.flash or {}
    self.flash.colour = colour or 0xffffffff
    self.flash.max_intensity = max_intensity or 1
    self.flash.t = 0
    self.flash.periodic = periodic
    self.flash.time_in = time_in or .2
    self.flash.time_out = time_out or .5
end

function AnimFighter:IsIdle()
    local idle_state = self.idle_states[1] or IDLE_STATE.NEUTRAL
    local idle_anims = self:GetIdleAnim( idle_state )

    -- if you're playing any of the anims from this anim set, you're idle.
    local current_anim = self.entity.cmp.AnimController:GetAnimName()
    for state, anim in pairs(idle_anims) do
        if self.current_anim_mapping then
            anim = self.current_anim_mapping[ anim ] or anim
        end
        if current_anim == anim then
            return true
        end
    end
    return false
end

function AnimFighter:CelebrateCoro()
    self.status_widget:Hide()
    self:ResetFighter()

    self:CoroDelay(.1 + math.random()*.2)

    self:PlayAnim( "taunt" )

    local x, y = self.sim:GetCamera():WorldToScreen( self:GetStatusWidgetPosition() )
    local width = TheGame:FE():GetScreenDims()
    local screen_pos = math.abs(x)/width
    local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
    AUDIO:PlayParamEvent("event:/sfx/battle/gen/win_burst", "position", pan_pos)

    local sx, sy, sz = 5, 3, 3
    local fx = self.entity.cmp.FXContainer:PlayFX( "anim/map_new_location_fx.zip", "pre", 0.5, 0, 0, sx, sy, sz )

    fx.model:SetBloom( 0.25 )
    local r, g, b, a = HexColour( 0xFFFFFF66 )
    local tint = engine.rendering.ShaderConstantValue("TINT_MULT", r, g, b, a)
    fx.model:SetShaderConstant(tint)

    self:WaitAnim()
    self:PlayIdleAnim()
end

function AnimFighter:Celebrate()
    self:AbortSequence()
    self.coro = self:StartCoroutine(AnimFighter.CelebrateCoro, self)
end


function AnimFighter:HumiliateCoro()
    self.status_widget:Hide()
    self:ResetFighter()    

    -- TODO: remove this, player doesn't have surrender atm
    if not self.fighter:IsPlayer() then
        self:AddIdleState( IDLE_STATE.SURRENDER )
        if self:IsBusy() then
            self:PlayIdleAnim()
        end
    end
end

function AnimFighter:Humiliate()
    self:AbortSequence()
    self.coro = self:StartCoroutine(AnimFighter.HumiliateCoro, self)
end

function AnimFighter:DoKill( no_despawn )

    self:KillUpdatingFX()

    self:StopMoving()
    self:PlayAnim( "death" )

    local teammate = self.fighter:GetTeam():GetRandomTeammate( self.fighter )
    local teammate_ent = self.sim.mgr:GetAnimFighterByFighter( teammate )
    if teammate_ent then
        teammate_ent.cmp.AnimFighter:SayQuip( self.fight_data.ally_killed_quip or "battle_ally_killed" )
    end

    local death_event
    if self.fighter.player_controlled then
        self.entity.cmp.SoundEmitter:PlayEvent( SoundEvents.battle_player_death )
    else
        self.entity.cmp.SoundEmitter:PlayEvent( SoundEvents.battle_death )
    end
    
    self.entity.cmp.AnimController:HideOutline()
    self:SetAnimOutline(self.outline_width, 0x000000ff)

    if self.status_widget then
        self.status_widget:AlphaTo( 0, 0.5 )
    end

    self:CoroDelay( self.fight_data and self.fight_data.death_fade_delay or 0.6 ) -- PRE_DELAY
    local t = 0 
    local total_t = self.fight_data and self.fight_data.death_fade_time or 0.5 -- FADE_TIME
    while t < total_t do
        local dt = coroutine.yield()
        t = t + dt
        local alpha = 1 - (t/total_t)
        self.entity.cmp.AnimController:SetTintMult(1,1,1, alpha)
        if self.dropshadow then
            self.dropshadow.cmp.PlaneRenderer:TintColour(0,0,0, .5*alpha)
        end
    end

    self:WaitAnim()

    -- self.sim.mgr:CalculateHomePositions( self.fighter:GetTeam(), true )

    if not no_despawn and not (self.fight_data and self.fight_data.no_despawn) then
        self.entity:Despawn()
    else
        self:AddIdleState( IDLE_STATE.DEAD )
    end
end

function AnimFighter:WaitAnim()
    if not self:IsIdle() then
        -- print( "WAIT", self.fighter, self.entity.cmp.AnimController:GetAnimName(), os.clock() )
        -- print( debug.traceback() )
        self.entity.cmp.AnimController:CoroWaitForDoneAnims()
        -- print( "WAIT DONE", self.fighter, self.entity.cmp.AnimController:GetAnimName(), os.clock() )
    end
end

function AnimFighter:PlayAnim( anim, loop )
    local base_anim = anim
    if self.current_anim_mapping then
        anim = self.current_anim_mapping[ anim ] or anim
    end

    local current_anim = self.entity.cmp.AnimController:GetAnimName()

    -- Determine if transition anims need to be played first.
    local transition_anims = self.anim_transitions[ current_anim ] and self.anim_transitions[ current_anim ][ anim ] 
    if not transition_anims and self.current_anim_mapping then
        for k,v in pairs(self.current_anim_mapping) do
            
            if v == current_anim then
                transition_anims = self.current_anim_mapping[ self.anim_transitions[ k ] and self.anim_transitions[ k ][ base_anim ] ]
                break
            end
        end
        if not transition_anims then
            transition_anims = self.current_anim_mapping[ self.anim_transitions[ current_anim ] and self.anim_transitions[ current_anim ][ base_anim ] ]
        end
    end

    

    if type(transition_anims) == "string" then
        if self.current_anim_mapping then
            transition_anims = self.current_anim_mapping[ transition_anims ] or transition_anims
        end
        self.entity.cmp.AnimController:Play( transition_anims )
    
    elseif transition_anims then
        for i, transition_anim in ipairs( transition_anims ) do
            if self.current_anim_mapping then
                transition_anim = self.current_anim_mapping[ transition_anim ] or transition_anim
            end
            if i == 1 then
                self.entity.cmp.AnimController:Play( transition_anim )
            else
                self.entity.cmp.AnimController:Push( transition_anim )
            end
        end
    end

    local anim_dat = self:GetMoveAnim( anim )
    if anim_dat.anim then
        anim = anim_dat.anim
    end

    if transition_anims then
        self.entity.cmp.AnimController:Push( anim, loop )
    else
        self.entity.cmp.AnimController:Play( anim, loop )
    end

    return anim_dat
end

local function SortIdles( a, b )
    return IDLE_ANIMS[ a ].priority > IDLE_ANIMS[ b ].priority
end

function AnimFighter:AddIdleState( idle_state )
    local old_idle = self.idle_states[1] or IDLE_STATE.NEUTRAL
    table.insert( self.idle_states, idle_state )
    table.sort( self.idle_states, SortIdles )

    local new_idle = self.idle_states[1] or IDLE_STATE.NEUTRAL
    if old_idle ~= new_idle and not self:IsBusy() then
        self:PlayIdleAnim()
    end

    self:RefreshIdleSFX()
end

function AnimFighter:RefreshIdleSFX( )
    local idle = self.idle_states[1] or IDLE_STATE.NEUTRAL
    

    if IDLE_ANIMS[idle].sfx then
        self:SetSpecialEffect(true, IDLE_ANIMS[idle].sfx.scale, IDLE_ANIMS[idle].sfx.colour)
    else
        self:SetSpecialEffect()
    end

end

function AnimFighter:RemoveIdleState( idle_state )
    local old_idle = self.idle_states[1] or IDLE_STATE.NEUTRAL
    table.arrayremove( self.idle_states, idle_state )
    table.sort( self.idle_states, SortIdles )

    local new_idle = self.idle_states[1] or IDLE_STATE.NEUTRAL
    if old_idle ~= new_idle and not self:IsBusy() then
        self:PlayIdleAnim()
    end

    self:RefreshIdleSFX()
end

function AnimFighter:PlayIdleAnim()
    local idle_state = self.idle_states[1] or IDLE_STATE.NEUTRAL
    local animdata = self:GetIdleAnim( idle_state )
    if animdata.loop then
        self:PlayAnim( animdata.loop, true )
    elseif animdata.anim then
        self:PlayAnim( animdata.anim )
    end
end

function AnimFighter:PlayFlourish()
    self:WaitForSequences()
    if self.entity.removed then
        return
    end
    self:PlayAnim("flourish")
    self:Flash( 0xFC4F4FFF, 0.2, 0.2, 0.5, true)
    self:WaitAnim()
    if self.flash then
        self.flash.periodic = false
    end
end

function AnimFighter:QueueIdleAnim()
    self:DoSequence( function()
            self.can_interrupt = true
            self:WaitAnim()
            self:PlayIdleAnim()
            self.can_interrupt = nil
        end )
end

function AnimFighter:PushAnim( anim, loop )
    if self.current_anim_mapping then
        anim = self.current_anim_mapping[ anim ] or anim
    end

    if anim and self.entity.cmp.AnimController:GetAnimInfo( anim ) then
        self.entity.cmp.AnimController:Push( anim, loop )
    end
end

function  AnimFighter:SetAnimOutline(width, tint)
   self.entity.cmp.AnimController:SetOutline(width, tint)
end

function AnimFighter:OnHit( attack, hit_count )
    self.hit_count = hit_count
end

function AnimFighter:OnStatusChanged( status )
    if self.status ~= status then
        self.status = status

        if status == FIGHT_STATUS.SURRENDER then
            self:DoSurrender()

        elseif status == FIGHT_STATUS.DEAD then
            self:DoKill()

        elseif status == FIGHT_STATUS.FLED then
            self:DoFlee()

        elseif status == FIGHT_STATUS.ACTIVE then
            self:DoRally()

        elseif status == FIGHT_STATUS.DEPARTED then
            self:DoRun()
        end
    end
end

function AnimFighter:SetStatusWidget(w)
    self.status_widget = w 
end

function AnimFighter:ShowDeltaStat( stat, delta, value, mitigated, new_cap, delta_cap )
    if self.status_widget then
        self.status_widget:ShowDeltaStat( stat, delta, value, mitigated, new_cap, delta_cap )
    end

    if stat == COMBAT_STAT.HEALTH and delta > 0 then
        self:PlayOneShotFX("fighter_heal")
    end
end

function AnimFighter:GetStatusWidgetPosition()
    local feet_wx, feet_wy, feet_wz = self.entity:GetLocalPosition()
    local fight_data = self.fight_data
    local dy = fight_data.status_widget_dy or FIGHT_STATUS_WIDGET_OFFSET
    local dx = fight_data.status_widget_dx or 0
    if self.entity.cmp.AnimController.flip then
        dx = - dx
    end

    local head_x, head_y, head_z
    if self.fight_data.track_symbol then
        head_x, head_y, head_z = self.entity.cmp.AnimController.animmodel:SymbolToWorld(self.fight_data.track_symbol, 0, 0, 0)
    else
        local head_dx = fight_data.status_widget_head_dx or FIGHT_STATUS_WIDGET_HEAD_XOFFSET
        if self.entity.cmp.AnimController.flip then
            head_dx = -head_dx
        end
        head_x = feet_wx + head_dx
        head_y = feet_wy + self.fake_scale * (fight_data.status_widget_head_dy or FIGHT_STATUS_WIDGET_HEAD_YOFFSET)
        head_z = feet_wz
    end
    return feet_wx + dx, feet_wy + dy, feet_wz, head_x, head_y, head_z
end

function AnimFighter:DoFlee()
    if self.fight_data.stationary then
        self:PlayAnim( "surrender" )
        self:WaitAnim()
        self.entity:Despawn()
    else
        
        local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
        local destx = (FIGHT_MAX_DIST+12)*(self.team == TEAM.RED and 1 or -1)

        self:PlayIdleAnim()
        self:SayQuip( "battle_run" )

        if self.fighter:IsPlayer() then
            self:CoroDelay( 30 * FRAMES )
        else
            self:CoroDelay( 10 * FRAMES )
        end

        self:PlayAnim( "surrender_pre" )
        self:WaitAnim()

        self.run_speed_throttle = 0.5
        self:StartMoving(destx, z)
        self:CoroWaitForArrive() -- wait for run away to finish

        self.entity:Despawn()
    end

end

function AnimFighter:DoRally()
    self:RemoveIdleState( IDLE_STATE.SURRENDER )
    if self:IsBusy() then
        self:PlayIdleAnim()
        self:WaitAnim()
    end
    if self.status_widget then
        self.status_widget:ShowMorale()
    end

    self:SayQuip( "battle_rally" )
    self:PlayAnim( "surrender_pst" )
    self:WaitAnim()
    self:PlayAnim( "taunt" )
    self:WaitAnim()
    self:PlayIdleAnim()
end

function AnimFighter:DoRun()
    if self.fight_data.stationary then
        self:PlayAnim( "surrender" )
        self:WaitAnim()
        self.entity:Despawn()
    else
        if self:IsBusy() then
            self:WaitAnim()
        end
        local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
        local destx = (FIGHT_MAX_DIST+12)*(self.team == TEAM.RED and 1 or -1)
        self.run_speed_throttle = 0.5
        self:StartMoving(destx, z)
        self:CoroWaitForArrive() -- wait for run away to finish

        self.entity:Despawn()
    end
end

function AnimFighter:JoinBattle( status_widget, screen )

    self:DeferConditions()

    self:SetStatusWidget( status_widget )

    if self.fight_data.OnJoinBattle then
        self.fight_data.OnJoinBattle( self.fighter, self, screen )
    else
        local x, z = self:GetHomePosition()
        if self.team == TEAM.BLUE then
            x = x - FIGHT_MAX_DIST
        else
            x = x + FIGHT_MAX_DIST
        end
    
        -- Start offscreen and charge in.
        self:SetPos( x, z )
    end

    if self.battle:GetTurns() == 0 and not self.fighter.call_in then
        -- Slight random delay on initial turn to offset approaches: this has the effect of desynching the idle anims.
        if self.team == TEAM.RED then
            self:CoroDelay( math.random() * 0.2 + 0.2 )
        else
            self:CoroDelay( math.random() * 0.2 )
        end
    end
    AUDIO:PlayEvent("event:/sfx/battle/gen/character_whoosh_in")
    if self.fighter.call_in then
        self:ReturnHomeCoro()
    else
        self:ResetFighter()
    end
    
    if self.team == TEAM.RED and math.random() < 0.5 then
        self:SayQuip( "battle_join")
    end

    if self.team == TEAM.BLUE and not self.fighter.agent:IsPlayer() and math.random() < 0.5 and not self.fighter.call_in then
        self:SayQuip( "battle_join_ally")
    end

    if self.status_widget and not self.fighter.call_in then
        self.status_widget:AlphaTo( 1.0, 0.5 ):Refresh()
    end
end

function AnimFighter:DoSurrender()
    -- print ("DO SURRENDER" )
    if self.status_widget then
        self.status_widget:HideMorale()
    end
    self:ResetFighter()
    self:AddIdleState( IDLE_STATE.SURRENDER )
    if self:IsBusy() then
        self:PlayIdleAnim()
    end

    self:SayQuip( "battle_surrender" )
end

function AnimFighter:AbortSequence()
    if self.coro then
        if not self.coro:IsDone() then
            if not self.can_interrupt then
                print( debug.traceback() )
                print( "ABORTED coroutine:", self.fighter )
                print( debug.traceback( self.coro.c ) )
            end
            self:StopCoroutine( self.coro )
        end

        self.coro = nil
        self.can_interrupt = nil
    end
end

function AnimFighter:WaitForSequences()
    while self.coro and not self.coro:IsDone() do
        coroutine.yield()
    end
end

function AnimFighter:DoSequence( fn, ... )
    if self.can_interrupt then
        self:AbortSequence()
    else
        self:WaitForSequences()
    end

    if not self.entity.removed then
        -- Manually pass in a coroutine so we can assign to self.coro BEFORE resuming
        local c = coroutine.create( fn )
        self.coro = self:StartCoroutine( c )
        self.coro:Resume( ... )

        return self.coro
    end
end

function AnimFighter:FormatString( text )
    local function Formatter( word, fn, field )
        local locobj
        if word:lower() == "player" then
            local fighter = self.battle:GetPlayerFighter()
            locobj = fighter and fighter.agent
        end

        if locobj then
            if fn == "." and is_instance( locobj, Agent ) then
                return loc.format( string.format( "{1%s%s}", fn, field ), locobj:LocTable() )
            elseif fn then
                return loc.format( string.format( "{1%s%s}", fn, field ), locobj )
            elseif is_instance( locobj, Agent ) then
                return loc.format( "{1#agent}", locobj )
            elseif is_instance( locobj, Location ) then
                return loc.format( "{1#location}", locobj )
            end
        end
    end

    return loc.custom_format( text, Formatter )
end

function AnimFighter:GetTalkText()
    local txt = self.talk_line and self.talk_line.txt
    if txt then
        return self:FormatString( txt )
    end
end

function AnimFighter:SayQuip( quip, ... )
    if not self.fighter.agent:IsSentient() or self.fighter.mute then
        return
    end

    local tags = { quip, ...}
    self.fighter.agent:FillOutQuipTags( tags )

    local txt = self.fighter.agent:LookupQuip( tags )
    self:SaySomething( txt )
end

function AnimFighter:SaySomething(script)
    if script == nil then
        return
    end
    table.clear(self.talk_lines)
    
    local lines = script:split("\n")
    local last_perf
    for i, line in ipairs(lines) do
        local find_performance = string.match(line, "^#([^%s]+)")
        if find_performance then
            last_perf = find_performance
        else
            table.insert(self.talk_lines, { txt=line})
            last_perf = nil
        end
    end
    self:SayNextLine()    

end

function AnimFighter:SayNextLine()

    local line = table.remove(self.talk_lines, 1)
    if line then
        self.talk_line = line
        self.talk_time = 2.5
    else
        self.talk_line = nil
        self.talk_time = nil
    end

end

function AnimFighter:CoroWaitForArrive()
    while self:IsMoving() do
        coroutine.yield()
    end
end

function AnimFighter:QueueLipsynch(sounds)
    --print (serpent.block(sounds))
    table.clear(self.mouth_queue)
    local time = GetWallTime()
    for k = 1, #sounds/2, 2 do
        table.insert( self.mouth_queue, {shape=sounds[k] or MOUTH_SHAPE.IDLE, time=sounds[k+1] + time})
        --self.mouth_queue[#sounds-(k-1)] = {shape=v.mouth_anim or MOUTH_SHAPE.IDLE, time=v.start_time + time}
    end

end

function AnimFighter:ClearMouthAnim()
    table.clear(self.mouth_queue)
    self.entity.cmp.AnimController:ClearSubAnim("mouth_swap")
end

function AnimFighter:PlayMouthAnim(anim, loop, preserve_queue, hold)
    if not preserve_queue then
        table.clear(self.mouth_queue)
    end
    self.entity.cmp.AnimController:PlaySubAnim("mouth_swap", anim, loop, 0, hold)
    return self
end

function AnimFighter:PlayUpdatingFX( fx_name )
    if self.fight_data.hud_fx_remapping and self.fight_data.hud_fx_remapping[ fx_name ] then
        fx_name = self.fight_data.hud_fx_remapping[ fx_name ]
    end

    local hit_tags = self.fighter:GetHitTags()

    for i, fx_track in ipairs( self.fx_tracks ) do
        if fx_track.id == fx_name then
            return
        end
    end

    local meta = GetFXTrack( fx_name )
    if meta and (meta.fx_tracks or meta.sound_frames) then
        local fx_track = FXTrack( fx_name, meta, self.entity.cmp.AnimController )
        fx_track:SetTrackTags( hit_tags )
        fx_track:AdvanceTrack( 0.0 )
        table.insert( self.fx_tracks, fx_track )
        return fx_track
    end
end

function AnimFighter:KillUpdatingFX( fx_name )
    if fx_name then
        if self.fight_data.hud_fx_remapping and self.fight_data.hud_fx_remapping[ fx_name ] then
            fx_name = self.fight_data.hud_fx_remapping[ fx_name ]
        end
    end

    for i, fx_track in ipairs( self.fx_tracks ) do
        if fx_name == nil or fx_track.id == fx_name then
            fx_track:Kill() -- Will get removed next update.
        end
    end
end

function AnimFighter:PlayOneShotFX( fx_name, hit_tags )
    if hit_tags then
        table.arrayadd( hit_tags, self.fighter:GetHitTags() )
    else
        hit_tags = self.fighter:GetHitTags()
    end

    local meta = GetFXTrack( fx_name )
    if meta and (meta.fx_tracks or meta.sound_frames) then
        local fx_track = FXTrack( fx_name, meta, self.entity.cmp.AnimController )
        fx_track:SetTrackTags( hit_tags )
        fx_track:AdvanceTrack( 0.0 )
        table.insert( self.fx_tracks, fx_track )
        return fx_track
    end
end

function AnimFighter:BlockedCondition( )
    if self.status_widget then
        self.status_widget:ShowConditionBlockedFloater( )
    end
end

function AnimFighter:ShowDeferredConditions()
    local conditions = self.deferred_conditions
    if conditions then
        self.deferred_conditions = nil

        for i, t in ipairs( conditions ) do
            if t.added then
                self:AddCondition( table.unpack( t ))
            end
        end
        for i, t in ipairs( conditions ) do
            if t.removed then
                self:RemoveCondition( table.unpack( t ))
            end
        end
    end
end

function AnimFighter:DeferConditions()
    if self.deferred_conditions == nil then
        self.deferred_conditions = {}
    end
end

function AnimFighter:AddCondition( condition, stacks, source )
    if self.deferred_conditions and condition.id ~= "DEFEND" and not condition.no_defer then
        table.insert( self.deferred_conditions, { condition, stacks, source, added = true })
        return
    end

    if self.status_widget then
        if condition.id == "DEFEND" then
            self.status_widget:AnimDefend( stacks, source ~= nil)
            if condition.apply_sound and not self.no_condition_sound then
                local wx, wy, wz = self:GetStatusWidgetPosition()
                local x, y = self.sim:GetCamera():WorldToScreen( wx, wy, wz )
                local width = TheGame:FE():GetScreenDims()
                local screen_pos = math.abs(x)/width
                local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
                AUDIO:PlayParamEvent(condition.apply_sound, "position", pan_pos)
            end
        else
            if condition.id == "PROTECT" then
                self.status_widget:AnimProtect( self.fighter:HasCondition("PROTECT") )
                self:ShowShield( self.fighter:HasCondition("PROTECT") )
            end
            if not condition.hide_gained and not self.hide_condition_gained then
                self.status_widget:ShowConditionFloater( condition, stacks )
            else
                if condition.apply_sound and not self.no_condition_sound then
                    local wx, wy, wz = self:GetStatusWidgetPosition()
                    local x, y = self.sim:GetCamera():WorldToScreen( wx, wy, wz )
                    local width = TheGame:FE():GetScreenDims()
                    local screen_pos = math.abs(x)/width
                    local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
                    AUDIO:PlayParamEvent(condition.apply_sound, "position", pan_pos)
                end
            end
            self.status_widget:ShowCondition( condition, stacks )
        end
    end

    if condition.hud_fx then
        for i, fx_name in ipairs(condition.hud_fx) do
            self:PlayUpdatingFX( fx_name )
        end
    end

    if condition.apply_fx and not condition.hide_gained and not self.hide_condition_gained then
        for i, fx_name in ipairs(condition.apply_fx) do
            self:PlayOneShotFX( fx_name )
            if condition.fx_sound and not self.no_condition_sound then
                local ent = self.entity
                ent:DoTaskInTime("fx_sound", condition.fx_sound_delay or 0, function()
                    local x, y = self.sim:GetCamera():WorldToScreen( self:GetStatusWidgetPosition() )
                    local width = TheGame:FE():GetScreenDims()
                    local screen_pos = math.abs(x)/width
                    local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
                    AUDIO:PlayParamEvent(condition.fx_sound, "position", pan_pos)
                end)   
            end
        end
    end

    if condition.idle_state_override then
        self:AddIdleState( condition.idle_state_override )
        -- self:CoroDelay( 1.0 )
    end

    if condition.play_anim then
        self:DoSequence( function()
            self.can_interrupt = true
            self:PlayAnim( condition.play_anim, false )
            self:WaitAnim()
            self:PlayIdleAnim()
            self.can_interrupt = nil
        end )
    end

    if condition.anim_tint then
        self.entity.cmp.AnimController:SetTintMult( table.unpack( condition.anim_tint ))
    end

    

    if condition.quip then
        self:SayQuip( condition.quip )
    end

    if condition.id == "PROTECT" then
        self.sim.mgr:CalculateHomePositions( self.fighter:GetTeam(), true )
        self:ResetFighter()
    end

    return condition
end

function AnimFighter:RemoveCondition( condition, stacks, source )
    if self.deferred_conditions and condition.id ~= "DEFEND" and not condition.no_defer then
        table.insert( self.deferred_conditions, { condition, stacks, source, removed = true })
        return
    end

    if self.status_widget then
        -- No removal floaters.
        if condition.id == "DEFEND" then
            self.status_widget:AnimDefend( -stacks, source ~= nil )
        else

            if condition.id == "PROTECT" then
                self.status_widget:AnimProtect( self.fighter:HasCondition("PROTECT") )
                self:ShowShield( self.fighter:HasCondition("PROTECT") )
            end

            self.status_widget:ShowCondition( condition, -stacks )
        end
    end


    if condition.hud_fx and condition.stacks <= 0 then
        for i, fx_name in ipairs(condition.hud_fx) do
            self:KillUpdatingFX( fx_name )
        end
    end

    if condition.unapply_fx then
        for i, fx_name in ipairs(condition.unapply_fx) do
            self:PlayOneShotFX( fx_name )
            if condition.unapply_fx_sound and not self.no_condition_sound then
                local ent = self.entity
                ent:DoTaskInTime("fx_sound", condition.fx_sound_delay or 0, function()
                    local x, y = self.sim:GetCamera():WorldToScreen( self:GetStatusWidgetPosition() )
                    local width = TheGame:FE():GetScreenDims()
                    local screen_pos = math.abs(x)/width
                    local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
                    AUDIO:PlayParamEvent(condition.unapply_fx_sound, "position", pan_pos)
                end)   
            end
        end
    end
    
    if condition.idle_state_override then
        self:RemoveIdleState( condition.idle_state_override )
    end

    if condition.anim_tint then
        self.entity.cmp.AnimController:SetTintMult( 1, 1, 1, 1 )
    end

    if condition.unapply_sound then
        AUDIO:PlayEvent( condition.unapply_sound )
    end
end

function AnimFighter:MoveUpdatingFX( fx_value )
    local x,y,z = self.entity.cmp.Transform:GetLocalPosition()
    local def = fx_value.def
    local new_x, new_y, new_z = def.x_offset or 0, def.y_offset or 0, def.z_depth or 0
    local rotz
    local x_scale = self.team == TEAM.BLUE and 1 or -1

    if def.symbol then
        if def.rotation then
            local x0, y0 = self.entity.cmp.AnimController.animmodel:SymbolToLocal( def.symbol, 0, 0, 0 )
            local x1, y1 = self.entity.cmp.AnimController.animmodel:SymbolToLocal( def.symbol, new_x, new_y, 0 )
            local angle = math.atan2( y1 - y0, x1 - x0 )
            angle = angle>=0 and angle or 2*math.pi+angle
            rotz = (angle + (def.rotation or 0)) * x_scale
            new_x, new_y = x1, y1
        else
            local x0, y0 = self.entity.cmp.AnimController.animmodel:SymbolToLocal( def.symbol, 0, 0, 0 )
            new_x, new_y = new_x + x0, new_y + y0
        end
    end

    new_x = x + (new_x * self.fake_scale) * x_scale

    if def.ground then
        fx_value.fx.node:SetTranslation( new_x, -self.height, z + new_z )
    else
        new_z = z + new_z + (new_y * self.fake_scale) * math.sin(-FIGHT_CAM_ANGLE)
        new_y = y + (new_y * self.fake_scale) * math.cos(-FIGHT_CAM_ANGLE)
        fx_value.fx.node:SetPosRot( new_x, new_y, new_z,-FIGHT_CAM_ANGLE, 0, rotz )
    end

    -- Storing x pos for audio events on the FX
    fx_value.fx.x_pos = new_x
    local scale = self.fake_scale * (def.scale or 1.0)
    fx_value.fx.node:SetScale( scale * x_scale, scale, scale )
end

function AnimFighter:RenderDebugUI( ui, panel )
    panel:AppendTable( ui, self.fighter )
    ui:SameLine( nil, 10 )
    panel:AppendTable( ui, self.entity.cmp.AnimController, "AnimController" )

    ui:SameLine( nil, 10 )
    
    ui:PushStyleColor( ui.Style_Button, 0x775555FF )
    if ui:Button( "Flash" ) then
        ui:OpenPopup( "FLASH_CXT" )
    end
    ui:PopStyleColor()

    if ui:BeginPopup( "FLASH_CXT" ) then
        self.d_col = ui:ColorSelector( "flash_colour", self.d_col or 0xffffffff ) 
        self.d_time_in = ui:SliderFloat("time_in", self.d_time_in or .2, 0, 1)
        self.d_time_out =  ui:SliderFloat("time_out", self.d_time_out or .5, 0, 1)
        self.d_intensity = ui:SliderFloat("intensity", self.d_intensity or .5, 0, 1)
        if ui:Button("FLASH") then
            self:Flash(self.d_col, self.d_time_in, self.d_time_out, self.d_intensity)
        end
        ui:EndPopup()
    end

    if self:IsMoving() then
        ui:Separator()

        ui:TextColored( 0x00FFFFFF, string.format( "Move Distance: %2.2f, Duration: %.2f",
            distance(0, 0, self.move_dx, self.move_dz),
            self.end_move_time - self.start_move_time ))

        if self.run_speed then
            ui:TextColored( 0x00FFFFFF, string.format( "Run Speed: %2.2f", self.run_speed ))
        end
    end

    if not self.entity.removed then
        local x, y, z = self.entity:GetLocalPosition()
        ui:TextColored( 0xFFCCCCFF, string.format( "Position: <%2.2f, %2.2f>", x, z ))
    end

    ui:Separator()

    self.entity.cmp.AnimController:RenderDebugAnimInfo( ui, panel )

    ui:Columns( 2 )
    for i, fx_track in ipairs( self.fx_tracks ) do
        panel:AppendTable( ui, fx_track, fx_track.id )
        ui:NextColumn()

        ui:Text( string.format( "%d active fx", #fx_track.active_fx ))
        ui:NextColumn()
    end
    ui:Columns( 1 )

    ui:Separator()

    panel:AppendKeyValues( ui, self )
end


function AnimFighter:DoDebugRender()
    
    if self.destx then
        local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
    
        self.sim:GetScene():DrawDebugLine(x,y,z, self.destx, 0, self.destz, 1, 0, 1)
    end

end

function AnimFighter:SetSpecialEffect(enabled, scale, colour, alpha, speed)
    colour = colour or 0xff00ffff
    scale = scale or -1.5
    alpha = alpha or {0.1, 0.1, 0.7}
    speed = speed or {0.025, 0.04, 0.4}
    self.entity.cmp.AnimController:SetSelectedLayerSpeeds(speed[1], speed[2], speed[3])
    self.entity.cmp.AnimController:SetSelectedLayerScale(scale)
    self.entity.cmp.AnimController:SetSelectedLayerAlphas(alpha[1], alpha[2], alpha[3])
    self.entity.cmp.AnimController:SetSelectedFadeDuration(0.2)
    self.entity.cmp.AnimController:SetSelectedColour(colour)
    self.entity.cmp.AnimController:SetSelected(enabled == true)   
end

function AnimFighter:SetXRayOutline(enabled, fillColour, lineColour, thresholdMin, thresholdMax)
    self.entity.cmp.AnimController:SetXRayOutline(enabled, fillColour, lineColour, thresholdMin, thresholdMax)
end



-- local parameters =
-- {
--     scale = -1.5, 
--     colour = 0x9900ffff,
--     alpha = {0.7, 0.3, 2.0}, -- Alpha for each layer of the FX
--     speed = {0.05, 0.07, 0.5},
--     glitchSpeed = 0.07, 
--     glitchScale = 6.0,
--     layersTexture = engine.asset.Texture("rgba/character_hover.tex"),
--     glitchTexture = engine.asset.Texture("glitch/glitch.tex"),
-- }
function AnimFighter:SetHologramEffect(enabled, parameters)
    self.holo_active = enabled

    if enabled then
        self.entity.cmp.AnimController:HideOutline()
    else
        self:SetAnimOutline(self.outline_width, 0x000000ff)
    end
    self.entity.cmp.AnimController:SetHologram(enabled, parameters)
end
