require "component"
require "util/camshake"

local Camera = class("Camera", Component)

Camera.UpdateOrderSortKey = 10000
Camera.Priority = -999

local ORTHO_UNIT = 96  --screen pixels per meter
local ORTHO_DELTA = 0.25
local ORTHO_MAX = 3
local ORTHO_MIN = -2

PERSP_MIN = 10
PERSP_MAX = 100

local PAN_RATE = 2

local GAIN_HIGH = .1
local GAIN_LOW = .025

function Camera:init()
    Camera._base.init(self)
    self.caminfo = engine.rendering.CameraInfo()
    self:SetRenderSize(engine.inst:GetScreenSize())
    self:SetFOV(45*DEGREES)
    self:SetPerspective()
    self:SetNearViewPlane(.01)
    self:SetFarViewPlane(500)

    self.gain = GAIN_LOW
    self.zoom_gain = GAIN_LOW
    self.angle = 0*DEGREES
    self.direction = 0*DEGREES
    self.orthoscale = 0
    self.current_dist = 0
    self.target_dist = 20
    self.target_x, self.target_y, self.target_z = 0, 0, 0
    self.current_x, self.current_y, self.current_z = 0, 0, 0
end

function Camera:ResetToDefaults( defaults )
    if defaults then
        self.defaults = defaults
    end
    assert( self.defaults, "Camera not assigned defaults" )

    if self.defaults.angle then
        self:SetAngle( self.defaults.angle )
    end
    if self.defaults.fov then
        self:SetFOV( self.defaults.fov )
    end
    if self.defaults.distance then
        self:SetDistance( self.defaults.distance )
    end
    if self.defaults.direction then
        self:SetDirection( self.defaults.direction )
    end
    if self.defaults.farview then
        self:SetFarViewPlane( self.defaults.farview )
    end
    if self.defaults.gain then
        self.gain = self.defaults.gain
    end

    if self.defaults.zoom_gain then
        self.zoom_gain = self.defaults.zoom_gain
    end
    if self.defaults.basex or self.defaults.basey or self.defaults.basez then
        self:SetBasePos( self.defaults.basex or 0, self.defaults.basey or 0, self.defaults.basez or 0 )
    end
    if self.defaults.orthoscale then
        self:SetOrtho()
        self:SetOrthoScale( self.defaults.orthoscale )
    end
end

function Camera:GetEyePosition()
    return self.caminfo:GetPosition()
end

function Camera:SetRenderSize(w,h)
    self.caminfo:SetScreenSize(w,h)
end

function Camera:WorldToScreen(x,y,z)
    return self.caminfo:WorldToScreen(x,y,z)
end

function Camera:ScreenToViewPlane(x,y)
    return self.caminfo:ScreenToViewPlane(x,y)
end



function Camera:SetNearViewPlane(dist)
    self.caminfo:SetMinDist(dist)
end

function Camera:SetFarViewPlane(dist)
    self.caminfo:SetMaxDist(dist)
end

function Camera:SetPerspective()
    self.ortho = false
    self.caminfo:SetPerspective()
end

function Camera:SetOrtho()
    self.ortho = true
    self.caminfo:SetOrtho()
end

function Camera:SetOrthoProjection( left, right, top, bottom )
    self.caminfo:SetOrthoProjection( left, right, top, bottom )
end


function Camera:SetOrthoScale( ortho_scale )
    self.ortho_scale = ortho_scale
    self.caminfo:SetOrthoScale( self.ortho_scale )
end

function Camera:GetOrthoScale()
    return self.ortho_scale
end

function Camera:SetFOV(rads)
    self.caminfo:SetFOV(rads/DEGREES)
end


function Camera:SetFollowEntity(ent, x, y, z)
    self.follow_ent = ent
    self.target_offset_x, self.target_offset_y, self.target_offset_z = x or 0, y or 0, z or 0
end


function Camera:GetCurrentBasePos()
    return self.current_x, self.current_y, self.current_z
end


function Camera:Shake(mag, speed, time )
    if TheGame:CamShakeEnabled() then
        self.shaker = CamShake( mag, speed, time )
    end
end

function Camera:SetBasePos(x,y,z)
    self:StopMotion()
    self.current_x = x
    self.current_y = y
    self.current_z = z
    self:UpdatePosition()

end

function Camera:SetDistance(dist)
    self:StopMotion()
    self.current_dist = dist
    self:UpdatePosition()
end

function Camera:SetDirection(angle)
    self:StopMotion()
    self.direction = angle
end

function Camera:SetAngle(angle)
    self:StopMotion()
    self.angle = angle
end


function Camera:GetDistance()
    return self.current_dist
end


function Camera:GetTargetDistance()
    return self.target_dist
end

function Camera:StopMotion()
    self.target_x = nil
    self.target_y = nil
    self.target_z = nil
    self.target_dist = nil
end

function Camera:WarpToTargets()
    
    self.current_x = self.target_x or self.current_x
    self.current_y = self.target_y or self.current_y
    self.current_z = self.target_z or self.current_z
    self.current_dist = self.target_dist or self.current_dist
    
    self:StopMotion()
    self:UpdatePosition()
end

function Camera:GetDirection()
    return self.direction
end


function Camera:GetAngle()
    return self.angle
end

function Camera:GetFOV()
    return self.caminfo:GetFOV()
end

function Camera:MoveTo(x,y,z,d, gain)
    self.target_x = x
    self.target_y = y
    self.target_z = z
    self.target_dist = d
    self.target_gain = gain or self.gain
end

function Camera:OnUpdate(wall_dt)

    if self.follow_ent then
        if self.follow_ent.removed then
            self.follow_ent = nil
        else
            self.current_x, self.current_y, self.current_z = self.follow_ent.cmp.Transform:LocalToWorld(self.target_offset_x or 0, self.target_offset_y or 0 , self.target_offset_z or 0 )
        end
    end

    if self.shaker then
        if self.shaker:IsDone() then
            self.shakex, self.shakey = nil, nil
            self.shaker = nil
        else
            self.shakex, self.shakey = self.shaker:UpdateShake( wall_dt )
        end
    end

    if self.target_x then
        local cx, cy, cz 
        if self.camx then
            cx, cy, cz = self.camx, self.camy, self.camz
        else
            cx, cy, cz = self.entity:GetLocalPosition(0,0,0)
        end
        local tx = self.target_x + self.target_dist*math.cos(self.direction)*math.cos(self.angle)
        local ty = self.target_y + self.target_dist*math.sin(self.angle)
        local tz = self.target_z + self.target_dist*math.cos(self.angle)*math.sin(self.direction)

        
        local dx, dy, dz = tx - cx, ty - cy, tz - cz
        
        local SMALL = .01
        if math.abs(dx) < SMALL and math.abs(dy) < SMALL and math.abs(dz) < SMALL then
            self:WarpToTargets()
        else
            local c = math.min( 1, self.target_gain * (wall_dt * 60))
            local x, y, z = cx + dx*c, cy + dy*c, cz + dz*c
            self:UpdatePosition(x,y,z)
        end
        
    else
        self:UpdatePosition()
    end
end

function Camera:UpdatePosition(x, y, z)

    x = x or (self.current_x + self.current_dist*math.cos(self.direction)*math.cos(self.angle))
    y = y or (self.current_y + self.current_dist*math.sin(self.angle))
    z = z or (self.current_z + self.current_dist*math.cos(self.angle)*math.sin(self.direction))
    self.camx, self.camy, self.camz = x, y, z

    x = x + (self.shakex or 0)
    y = y + (self.shakey or 0)
    
    
    self.entity:SetLocalRotation(-self.angle, -self.direction+90*DEGREES, 0)
    self.entity:SetLocalPosition(x,y,z)
    self.caminfo:Update(self.entity.transform_provider.scenegraphnode)
end
