require "component"
local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

local TRAVEL_TIME = 0.4

local DroneController = class("DroneController", Component)

local drone_actions = 
{
    ["self_destruct"] = function( self, target, manager, battle )
        battle:BroadcastEvent( BATTLE_EVENT.CUSTOM, nil, function( fight_screen, ent, fighter )
            self.entity.cmp.AnimController:CoroDelay(0.35)
            self.moving = true
            self.idle = false
            self.entity.cmp.AnimController:Play("run_forward", true)
            local target_ent = self.entity.cmp.AnimController.sim.mgr:GetAnimFighterByFighter(target)
            local fighter_x, fighter_y, fighter_z = target_ent.cmp.Transform:GetLocalPosition()
            self.destx = fighter_x
            self.desty = fighter_y + 0.5
            self.destz = fighter_z + 1
            self.can_interrupt = true
            local x, y, z = self.entity.cmp.Transform:GetLocalPosition()
            local startx, starty, startz = x, y, z
            local distX = self.destx - x
            local distY = self.desty - y
            local distZ = self.destz - z
            local t = 0
            while t < TRAVEL_TIME do
                local vx = startx + (distX * (t/TRAVEL_TIME))
                local vy = starty + (distY * (t/TRAVEL_TIME))
                local vz = startz + (distZ * (t/TRAVEL_TIME))
                self.entity.cmp.Transform:SetPosRot(vx,vy,vz)
                x, y, z = self.entity.cmp.Transform:GetLocalPosition()
                local dt = coroutine.yield()
                t = t + dt
            end
            self.entity.cmp.AnimController:Play("explosion")
            self.entity.cmp.AnimController:CoroWaitForAnim()
            manager:RemoveDrone(self.entity)
            manager:Reorganize()
            self.can_interrupt = nil
            self:Kill(true)
        end, self.entity )
    end,

    ["expand"] = function( self, target, manager, battle )
        battle:BroadcastEvent( BATTLE_EVENT.CUSTOM, nil, function( fight_screen, ent, fighter )
            self.entity.cmp.AnimController:Play("transform")
            self.entity.cmp.AnimController:Push("idle2", true)
            self.transformed = true
        end, self.entity )
    end,

    ["blast"] = function( self, target, manager, battle )
        battle:BroadcastEvent( BATTLE_EVENT.CUSTOM, nil, function( fight_screen, ent, fighter )
            self.entity.cmp.AnimController:CoroDelay(0.4)
            self.entity.cmp.AnimController:Play("blast")
            self.entity.cmp.AnimController:CoroWaitForAnim()
            self.entity.cmp.AnimController:Play("transform2")
            self.entity.cmp.AnimController:Push("idle", true)
            self.transformed = false
        end, self.entity )
    end,

    ["kill"] = function( self, target, manager, battle )
        battle:BroadcastEvent( BATTLE_EVENT.CUSTOM, nil, function( fight_screen, ent, fighter )
            self:Kill()
        end, self.entity )
    end,
}

function DroneController:init(entity_to_follow, team, fighter)
	self.fighterToFollow = entity_to_follow
    self.fighter = fighter
    self.drone_y = self.drone_y or 4
    self.drone_x = self.drone_x or -0.5
    self.follow = self.follow or true
    self.team = team
end

local fake_perspective_scale_max_amount = .15
function DroneController:OnTickUpdate(dt)
    if self.destx and not self.moving then
    	local dx, dy, dz = self.entity.cmp.Transform:GetLocalPosition()
        local xDist = self.destx - dx
        local yDist = self.desty - dy
        local zDist = self.destz - dz
        local dist = math.sqrt( (xDist ^ 2) + (yDist ^ 2) + (zDist ^ 2))

        if dist > 0 then
            local velocity = self.velocity or 0.05
            
            local vx = dx + (xDist * velocity)
            local vy = dy + (yDist * velocity)
            local vz = dz + (zDist * velocity)
            self.entity.cmp.Transform:SetPosRot(vx,vy,vz)
        end
    end
    if self.idle and self.follow then
        self:SetHomePos()
    end
end

function DroneController:OnDespawn()
	self:StopTickUpdating()
end

function DroneController:ShowDrone()
    if not self.drone_visible then 
        self.drone_visible = true
    	self:StartTickUpdating()

    	self:SetHomePos()
        self.entity.cmp.AnimController:SetXFlip(self.team.id == TEAM.RED)
    	self.entity.cmp.AnimController:StartAnim("idle", true)
        local delay = math.random(0, 10) * 0.3
        self.entity.cmp.AnimController.animmodel:SetStartTime(delay)
        self.entity.cmp.AnimController.anim_track:AdvanceTrack(delay)
    	self.entity:Show()
        self.idle = true
    end
end

function DroneController:SetHomePos()
    if self.fighter and self.fighter:IsAlive() then
        local fighter_scale = self.fighterToFollow.cmp.AnimFighter:GetFakeScale()
        local fighter_x, fighter_y, fighter_z = self.fighterToFollow.cmp.Transform:GetLocalPosition()
        
        local x_flip = self.fighterToFollow.cmp.AnimFighter.fighter.team and self.fighterToFollow.cmp.AnimFighter.fighter.team.id == TEAM.BLUE and -1 or 1
        self.destx = fighter_x + (self.drone_x*x_flip)
        self.desty = (self.drone_y - fighter_y)
        self.destz = fighter_z + 1

        self.entity.cmp.AnimController:SetScale(fighter_scale)
    end
end

function DroneController:Kill( ignore_anim )
    if self.entity.removed then
        return
    end
    
    self.idle = false
    self.follow = false
    if not ignore_anim then
        self.entity.cmp.AnimController:Play( self.transformed and "death2" or "death" )
        self.entity.cmp.AnimController:CoroWaitForDoneAnims()
    end
    self.entity:Despawn()
end

function DroneController:PlayAction( action_id, target, manager, battle )
    if drone_actions[action_id] then
        drone_actions[action_id]( self, target, manager, battle )
    end
end

function DroneController:SetOffset( x, y )
    self.drone_x = x
    self.drone_y = y
end

function DroneController:SetFollow( follow )
    self.follow = follow
end

-- Functionality from AnimFighter
function DroneController:AbortSequence()
    if self.coro then
        if not self.coro:IsDone() then
            self:StopCoroutine( self.coro )
        end

        self.coro = nil
        self.can_interrupt = nil
    end
end

function DroneController:WaitForSequences()
    while self.coro and not self.coro:IsDone() do
        coroutine.yield()
    end
end

function DroneController:DoSequence( fn, ... )
    if self.can_interrupt then
        self:AbortSequence()
    else
        self:WaitForSequences()
    end

    if not self.entity.removed then
        local c = coroutine.create( fn )
        self.coro = self:StartCoroutine( c )
        self.coro:Resume( ... )

        return self.coro
    end
end