require "component"

local FloatText = class("FloatText", Component)

local ICON_SIZE = 1.0, 1.0

local DEFAULT_KEYFRAMES =
{
    -- time, value, ease, ...
    dy = { 0, 0, easing.linear, 1.5, 0.33 },
    alpha = { 4.0, 1.0, easing.linear, 5.0, 0.0 }
}

function FloatText:init()
    FloatText._base.init(self)  
    self.model = engine.rendering.TextNode()
    self.model:SetFont( "default" )
    self.model:SetFontSize( 1.0 )
    self.node = engine.rendering.SceneGraphNode(self.model) 
    self.node:Hide()
    self.node:BillboardToCamera( true )
end

function FloatText:AddIcon( texture, size )
    self.icon_models = self.icon_models or {}
    self.icon_nodes = self.icon_nodes or {}

    local model = engine.model.PlaneModel()
    model:SetSize( size, size )
    model:SetUVOffset(.5, .5)
    model:Rebuild()
    model:SetDepthState(DEPTH_TEST.LESSEQUAL, false, 0)
    model:SetBlendMode(BLEND_MODE.BLENDED)

    -- self.icon_model = model

    local node = engine.rendering.SceneGraphNode(model) 
    node:BillboardToCamera( true, -math.pi/2, 0, 0 )
    self.entity.cmp.Transform:AddChildNode(node)

    model:SetTexture( 0, texture )
    node:Show()

    table.insert(self.icon_models, model)
    table.insert(self.icon_nodes, node)
end

function FloatText:GetModel()
    return self.model
end

function FloatText:OnEntityReady()
    self.entity.cmp.Transform:AddChildNode(self.node)
end

function FloatText:SetKeyFrames( keyframes )
    self.keyframes = keyframes
end

function FloatText:AddKeyFrames( prop, keyframes )
    assert( #keyframes % 3 == 2 )
    if self.keyframes == nil then
        self.keyframes = {}
    end
    self.keyframes[ prop ] = keyframes
end

function FloatText:SetText( txt )
    SetFormattedText( self.model, txt )
end

function FloatText:StartFloater( txt, icons, x, y, z, icon_size )
    self.node:Show()
    self.dt = 0
    self.x0, self.y0, self.z0 = x or 0, y or 0, z or 0
    self.entity:SetLocalPosition( self.x0, self.y0, self.z0 )

    if txt then
        self:SetText( txt )
    else
        self.node:Hide()
    end

    if icons and #icons > 0 then
        icon_size = icon_size or ICON_SIZE
        for i, icon in ipairs(icons) do
            self:AddIcon( icon, icon_size )
        end

        local x1, y1, x2, y2 = self.model:GetRenderBounds()

        if txt then
            for i, node in ipairs(self.icon_nodes) do
                node:SetTranslation( x1 - (icon_size * i) * 0.6, 0, -0.05 * (i - 1) )
            end
        else
            for i, node in ipairs(self.icon_nodes) do
                node:SetTranslation(-icon_size * (i - 1) * 0.6, 0, -0.05 * (i - 1) )
            end
        end
    elseif self.icon_nodes then
        for i, node in ipairs(self.icon_nodes) do
            node:Hide()
        end
    end

    self:OnUpdate( 0 ) -- Apply values at t == 0.

    if self.entity:IsReady() then
        self:StartWallUpdating()
    end
end

function FloatText:ApplyValue( key, value )
    if key == "dx" then
        self.dx = value
    elseif key == "dy" then
        self.dy = value
    elseif key == "dz" then
        self.dz = value
    elseif key == "size" then
        self.model:SetFontSize( value )
    elseif key == "scale" then
        self.entity:SetLocalScale( value, value, value )
    elseif key == "bloom" then
        self.model:SetBloom( value )
        if self.icon_models then
            for i, model in ipairs(self.icon_models) do
                model:SetBloom( value )
            end
        end
    elseif key == "alpha" then
        if not self.tint_mult_constant then
            self.tint_mult_constant = engine.rendering.ShaderConstantValue("TINT_MULT", 1, 1, 1, 1)
            self.model:SetShaderConstant(self.tint_mult_constant)
            if self.icon_models then
                for i, model in ipairs(self.icon_models) do
                    model:SetShaderConstant(self.tint_mult_constant)
                end
            end
        end
        self.tint_mult_constant:SetValue(1,1,1, value)
    end
end

function FloatText:OnUpdate(dt)
    self.dt = self.dt + dt
    local dt = self.dt

    self.dx, self.dy, self.dz = nil, nil, nil
    local done = true
    for k, v in pairs( self.keyframes or DEFAULT_KEYFRAMES ) do
        for i = 1, #v, 3 do
            local t0, v0, ease_fn, t1, v1 = table.unpack( v, i, i + 5 )
            if dt >= t0 then
                if t1 == nil then
                    self:ApplyValue( k, v0 )
                    break
                elseif dt <= t1 then
                    local t = clamp( (dt - t0) / (t1 - t0), 0, 1.0 )
                    local v = ease_fn( t, v0, v1 - v0, 1.0 )
                    self:ApplyValue( k, v )
                    done = false
                    break
                end
            else
                done = false
                break
            end
        end
    end

    if self.dx or self.dy or self.dz then
        self.entity:SetLocalPosition( self.x0 + (self.dx or 0), self.y0 + (self.dy or 0), self.z0 + (self.dz or 0))
    end

    if done then
        if self.auto_despawn then
            self.entity:Despawn()
        else
            self.node:Hide()
            self:StopWallUpdating()
        end
    end
end

function FloatText:SetAutoDespawn( auto_despawn )
    self.auto_despawn = auto_despawn
end

function FloatText:GetRenderBounds()
    return self.model:GetRenderBounds()
end