-- Scale and flip from AnimController is dumb.
-- can fx_sound be moved into the timeline?
-- fix COMBAT_DRONE stun override placement
-- "small" version of wound fx.
-- Support for Delete and Duplicate fx tracks in tool.



local FXTrack = class( "FXTrack" )

function FXTrack:init( id, track_data, cmp )
    assert( is_instance( cmp, AnimController ))

    self.cmp = cmp
	self.animmodel = cmp.animmodel
    self.animnode = cmp.animnode
    self.entity = cmp.entity
    self.fx_mgr = cmp.sim.fx_mgr

    self.active_fx = {}
    self.track_data_tags = table.empty

    self.pointers = {}

    self:Reset( id, track_data )
end

function FXTrack:SetTrackTags( tag_list )
    self.track_data_tags = tag_list
end

-- Finds track frames with a folder name followed by a digit, implying a variant.
-- Fills out the 'variants' table with the variant name and list of variants discovered:
--      eg. variants = { Lumin = { "Lumin1", "Lumin2", "Lumin4", }, Acid = { ... } }

local function ParseVariants( fx_tracks, variants )
    for i, v in ipairs( fx_tracks ) do
        local variant_root = v.folder_name and v.folder_name:match( "^(.+)%d+$" )
        if variant_root then
            if variants == nil then
                variants = {}
            end
            if variants[ variant_root ] == nil then
                variants[ variant_root ] = {}
            end
            table.insert_unique( variants[ variant_root ], v.folder_name )
        end
    end

    return variants
end

function FXTrack:Reset( id, track_data )  

    table.clear( self.pointers )
    self.elapsed_frames = nil
    self.elapsed_time = nil
    self.variant_roll = math.random()

    if track_data then
        self.id = id
        self.track_data = track_data
        assert_warning( id ~= nil, "missing id" )
    end

    if self.track_data.fx_tracks then
        self:AddTrackPointer( self.track_data.fx_tracks )
    end

    if self.track_data.sound_frames then
        self:AddTrackPointer( self.track_data.sound_frames )
    end

    if self.track_data.shake_frames then
        self:AddTrackPointer( self.track_data.shake_frames )
    end
end

function FXTrack:IsValid()
    if #self.active_fx > 0 then
        return true
    end

    if #self.pointers > 0 then
        return true
    end

    return false
end

function FXTrack:IsCurrentVariant( fxdata )
    local variant_root = fxdata.folder_name and fxdata.folder_name:match( "^(.+)%d+$" )
    if variant_root and self.variants then
        local variant_idx = table.find( self.variants[ variant_root ], fxdata.folder_name )
        local max_variants = #self.variants[ variant_root ]
        local random_idx = math.ceil( self.variant_roll * max_variants )
        -- print( "ROLLED VARIANT:", random_idx, max_variants, " for variant", variant_idx, fxdata.folder_name, rawstring(fxdata) )
        if variant_idx ~= random_idx then
            return false
        end
    end

    return true
end

function FXTrack.MatchFxTrackTags( fxdata, tags )

    local tags_ok = true

    if tags_ok and fxdata.must_have_tags then
        for k,tag in ipairs(fxdata.must_have_tags) do
            if tags == nil or not table.arrayfind(tags, tag) then
                tags_ok = false
                break
            end
        end
    end
    
    if tags_ok and tags then
        if fxdata.cant_have_tags then
            for k,tag in ipairs(fxdata.cant_have_tags) do
                if table.arrayfind(tags, tag) then
                    tags_ok = false
                    break
                end
            end
        end
    end

    return tags_ok
end

function FXTrack:Kill()
    for i = #self.active_fx, 1, -1 do
        self.fx_mgr:KillFX( self.active_fx[ i ] )
    end
    assert_warning( #self.active_fx == 0 ) -- will trigger on_dispose which will remove this entry
    table.clear( self.pointers )
end

function FXTrack:KillLinkedFX()
    for i = #self.active_fx, 1, -1 do
        local fx = self.active_fx[i]
        if fx.fxdata.anim_linked then
            self.fx_mgr:KillFX( fx )
        end
    end
end

function FXTrack:AddTrackPointer( track_data )
    local idx = 1
    while idx <= #track_data and self.elapsed_frames and track_data[ idx ].frame <= self.elapsed_frames do
        idx = idx + 1
    end

    -- Discover variants from this track.
    self.variants = ParseVariants( track_data, self.variants or {} )

    local pointer = {
        track_data = track_data,
        idx = idx,
        offsetx = 0, offsety = 0, offsetz = 0
    }
    table.insert( self.pointers, pointer )

    return pointer
end


function FXTrack:AdvanceTrackPointer( pointer, frame )
    while pointer.idx <= #pointer.track_data and pointer.track_data[ pointer.idx ].frame <= frame do

        local fxdata = pointer.track_data[ pointer.idx ]
        pointer.idx = pointer.idx + 1

        if self.cmp.fx_track_fn then
            if self.cmp.fx_track_fn( self, fxdata, pointer, self.track_data_tags ) then
                break
            end

        elseif fxdata.vis_status == VIS_STATUS.HIDE_IN_GAME then
            -- Hidden!

        elseif FXTrack.MatchFxTrackTags( fxdata, self.track_data_tags ) then
            if not self:IsCurrentVariant( fxdata ) then
                -- Not current variant
            elseif self:TriggerFXTrack( fxdata, pointer ) then
                break
            end
        end
    end
end

function FXTrack:AdvanceTrack( frame )
    local i = 1
    while i <= #self.pointers do
        local pointer = self.pointers[i]
        self:AdvanceTrackPointer( pointer, frame )

        if pointer.idx > #pointer.track_data and self.pointers[i] == pointer then
            table.remove( self.pointers, i )
        else
            i = i + 1
        end
    end
end

function FXTrack:UpdateTransform( fx )
    local x,y,z = self.entity.cmp.Transform:GetLocalPosition()
    local fxdata = fx.fxdata
    local xscale = self.cmp.flip and -1 or 1
    local scale = self.cmp.scale or 1.0
    local new_x, new_y, new_z
    local rotz = (fxdata.rot or 0) + (fx.rot or 0)

    if fxdata.symbol then

        if fxdata.model_rotation then
            local x0, y0 = self.animmodel:SymbolToLocal( fxdata.symbol, 0, 0, 0 )
            local x1, y1 = self.animmodel:SymbolToLocal( fxdata.symbol, 10, 10, 0 )
            local angle = math.atan2( y1 - y0, x1 - x0 )
            angle = angle>=0 and angle or 2*math.pi+angle
            rotz = (angle + rotz)  * xscale
            local dx, dy = rotate2d( fxdata.x or 0, fxdata.y or 0, angle )
            new_x, new_y, new_z = dx + x0, dy + y0, fxdata.z or 0
        else
            local x0, y0 = self.animmodel:SymbolToLocal( fxdata.symbol, 0, 0, 0 )
            new_x, new_y, new_z = (fxdata.x or 0) + x0 + fx.offsetx, (fxdata.y or 0) + y0 + fx.offsety, (fxdata.z or 0) + fx.offsetz
            rotz = rotz * xscale
        end

        new_x = x + (new_x * scale) * xscale

        if fxdata.ground then
            new_z = z + new_z
            fx.node:SetTranslation( new_x, 0, new_z )

        else
            new_z = z + new_z + (new_y * scale) * math.sin(-FIGHT_CAM_ANGLE)
            new_y = y + (new_y * scale) * math.cos(-FIGHT_CAM_ANGLE)
            fx.node:SetPosRot( new_x, new_y, new_z, -FIGHT_CAM_ANGLE, 0, rotz )
        end

    elseif fxdata.ground then
        new_x, new_y, new_z = (fxdata.x or 0) + fx.offsetx, (fxdata.y or 0) + fx.offsety, (fxdata.z or 0) + fx.offsetz

        new_x = x + (new_x * scale) * xscale
        new_z = z + (new_z * scale)
        rotz = rotz * xscale

        fx.node:SetPosRot(new_x, 0, new_z, -90*DEGREES, 0, rotz)

    else
        -- Just follow our root transform.
        new_x, new_y, new_z = (fxdata.x or 0) + fx.offsetx, (fxdata.y or 0) + fx.offsety, (fxdata.z or 0) + fx.offsetz

        new_x = x + (new_x * scale) * xscale
        new_z = z + new_z + (new_y * scale) * math.sin(-FIGHT_CAM_ANGLE)
        new_y = y + (new_y * scale) * math.cos(-FIGHT_CAM_ANGLE)
        rotz = rotz * xscale

        fx.node:SetPosRot(new_x, new_y, new_z, -FIGHT_CAM_ANGLE, 0, rotz)
    end

    local sx = fxdata.scalex or fxdata.scale or 1
    local sy = fxdata.scaley or fxdata.scale or 1
    if fxdata.model_scale or fxdata.symbol then
        sx, sy = sx * scale, sy * scale
    end
    if fx.scale then
        sx = sx * fx.scale
        sy = sy * fx.scale
    end
    fx.node:SetScale( sx * xscale, sy, 1.0 )
end

function FXTrack:UpdateTransforms()
    for i, fx in ipairs(self.active_fx) do
        if fx.fxdata.follow then
            self:UpdateTransform( fx )
        end
    end
end

-- NOTE: This elapsed-time update is only called for tracks which are not bound to an animation
-- (in which case the animation frame determines the advancement on the timeline)
function FXTrack:UpdateElapsedTime( dt )
    self.elapsed_time = (self.elapsed_time or 0) + dt
    self.elapsed_frames = math.ceil( (self.elapsed_time or 0) / FRAMES )

    self:AdvanceTrack( self.elapsed_frames )
end

function FXTrack:TriggerFX( fxdata, pointer )
    local tx, ty, tz = (fxdata.x or 0) + pointer.offsetx, (fxdata.y or 0) + pointer.offsety, (fxdata.z or 0) + pointer.offsetz
    local sx = (fxdata.scalex or fxdata.scale or 1)
    local flip = (fxdata.flip == true)
    local sy = (fxdata.scaley or fxdata.scale or 1)
    local r = (fxdata.rot or 0)

    if fxdata.model_scale then
        local scale = self.cmp.scale or 1.0
        tx = tx * scale
        ty = ty * scale
        tz = tz * scale
        sx = sx * scale
        sy = sy * scale
    end
    if pointer.scale then
        sx = sx * pointer.scale
        sy = sy * pointer.scale
    end
    if pointer.rot then
        r = r + pointer.rot
    end

    local speed = self.cmp:GetPlaySpeed()
    local x, y, z = self.animnode:LocalToWorld( tx, ty, tz )
    local fx
    if fxdata.ground then
        fx = self.fx_mgr:PlayFXOnGround(fxdata.fx, fxdata.state or "idle", x,z, sx, sy, r, flip, fxdata.loop, speed )
    else
        fx = self.fx_mgr:PlayBillboardFX(fxdata.fx, fxdata.state or "idle", x,y,z, sx, sy, 1, r, flip, fxdata.loop, speed )
    end
    
    if not fx then
        return
    end

    if fx.model and fxdata.buildfile then
        fx.model:AddOverrideFile(fxdata.buildfile)
    end

    if fxdata.bloom then
        fx.model:SetBloom(fxdata.bloom)
    end

    if fxdata.blendmode then
        fx.model:SetBlendMode(fxdata.blendmode)

        if fxdata.blendmode == BLEND_MODE.MULTIPLY then
            fx.model:SetEffect( global_shaders.FX_ANIM_PREMULT )
        end
    end

    local r, g, b, a = 1, 1, 1, 1
    if fxdata.alpha or fxdata.colour then
        if fxdata.colour then
            r, g, b, a = HexColour(fxdata.colour, fxdata.alpha)
        end

        local tint = engine.rendering.ShaderConstantValue("TINT_MULT", r, g, b, a)
        fx.model:SetShaderConstant(tint)
    end

    if fxdata.erode then
        fx.model:SetEffect( global_shaders.FX_ANIM_ERODE )
    end

    fx.fxdata = fxdata

    fx.offsetx = pointer.offsetx
    fx.offsety = pointer.offsety
    fx.offsetz = pointer.offsetz
    fx.scale = pointer.scale
    fx.rot = pointer.rot

    fx.on_dispose = function( fx )
        for i, v in ipairs( self.active_fx ) do
            if v == fx then
                table.remove( self.active_fx, i )
            end
        end
    end

    self:UpdateTransform( fx )

    table.insert( self.active_fx, fx )
end

function FXTrack:TriggerShake( fxdata )
    for i = #TheGame:FE().screenstack, 1, -1 do
        local screen = TheGame:FE().screenstack[i]
        if screen.ScreenShake then
            screen:ScreenShake(fxdata.shake)
            break
        end
    end
end

function FXTrack:TriggerSound( fxdata )
    local event_name = SoundEvents[ fxdata.sound ]
    if not event_name then
        print ("MISSING SOUND EVENT", fxdata.sound)
    elseif event_name then
        if self.entity.cmp.AnimFighter then
            local wx, wy, wz = self.entity.cmp.AnimFighter:GetStatusWidgetPosition()
            local x, y = self.entity.cmp.AnimFighter.sim:GetCamera():WorldToScreen( wx, wy, wz )
            local width = TheGame:FE():GetScreenDims()
            local screen_pos = math.abs(x)/width
            local pan_pos = easing.linear( screen_pos, -1, 2, 1 )

            AUDIO:PlayParamEvent(event_name, "position", pan_pos)
        else
            local wx, wy, wz = self.entity.cmp.Transform:GetLocalPosition()
            local x, y = self.entity.cmp.AnimController.sim:GetCamera():WorldToScreen( wx, wy, wz )
            local width = TheGame:FE():GetScreenDims()
            local screen_pos = math.abs(x)/width
            local pan_pos = easing.linear( screen_pos, -1, 2, 1 )

            AUDIO:PlayParamEvent(event_name, "position", pan_pos)
        end
    end
end

function FXTrack:TriggerVO( fxdata )
    local agent = self.entity.cmp.AnimFighter.fighter.agent
    local event_name = agent and agent:GetCombatVOEvent(fxdata.vo)
    if event_name then
        local wx, wy, wz = self.entity.cmp.AnimFighter:GetStatusWidgetPosition()
        local x, y = self.entity.cmp.AnimFighter.sim:GetCamera():WorldToScreen( wx, wy, wz )
        local width = TheGame:FE():GetScreenDims()
        local screen_pos = math.abs(x)/width
        local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
        AUDIO:PlayParamEventBank(event_name, "sound/vo.bank", "position", pan_pos)
    end
end

function FXTrack:TriggerFXTrack( fxdata, pointer )
    if fxdata.fx and #fxdata.fx > 0 then
        if not self.entity.sim:IsHidingFX() then
            self:TriggerFX( fxdata, pointer )
        end

    elseif fxdata.vo then
        self:TriggerVO( fxdata )
    elseif fxdata.shake then
        self:TriggerShake( fxdata )
    elseif fxdata.sound then
        self:TriggerSound( fxdata )

    elseif fxdata.command == "loop" then
        self:Reset()
        return true -- bit of a hack, break out of our track processing.

    elseif fxdata.command == "merge" then
        local meta = fxdata.filename and GetAnimMeta( fxdata.filename, fxdata.anim )
        if meta then
            if meta.fx_tracks and fxdata.no_fx ~= true then
                local pointer = self:AddTrackPointer( meta.fx_tracks )
                pointer.offsetx = fxdata.offsetx or 0
                pointer.offsety = fxdata.offsety or 0
                pointer.offsetz = fxdata.offsetz or 0
                pointer.scale = fxdata.scale
                pointer.rot = fxdata.rot
            end
            if meta.sound_frames and fxdata.no_sound ~= true then
                self:AddTrackPointer( meta.sound_frames, fxdata )
            end
            if meta.shake_frames then
                self:AddTrackPointer( meta.shake_frames, fxdata )
            end
        else
            print( "FxTrack merge failed:", self.id, fxdata.filename, fxdata.anim )
        end
    end
end