local battle_defs = require "battle/battle_defs"
local EVENT = battle_defs.EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local DEBUFF_DAMAGE_MULT = 0.5

local NPC_MELEE = table.extend( NPC_ATTACK )
{
    flags = CARD_FLAGS.MELEE
}

local NPC_RANGED = table.extend( NPC_ATTACK )
{
    flags = CARD_FLAGS.RANGED
}

-----------------------------------------------------------------------
--

local conditions =
{
    power_barrier = 
    {
        name = "Power Barrier",
        desc = "Deal <#HILITE>{1}</> bonus damage next turn. Increase by {2} when receiving attack damage. Effect is broken at {3} bonus damage",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks, self.bonus_increment, self.bonus_threshold)
        end,

        bonus_increment = 5,
        bonus_threshold = 30,

        event_handlers = 
        {
            [ EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                if self.owner == fighter then
                    self.owner:AddCondition("power_barrier", self.bonus_increment)
                    if self.stacks >= self.bonus_threshold then
                        self.owner:RemoveCondition("power_barrier")
                    end
                end
            end,

            [ EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner then
                    if card:IsAttackCard() then
                        dmgt:AddDamage( self.stacks, self.stacks, self )
                    end
                end
            end,

            [ EVENT.POST_RESOLVE ] = function( self, battle, attack )
                if self.owner == attack.attacker and attack ~= self.skip_attack then
                    self.owner:RemoveCondition("power_barrier")
                end
            end
        },
    },

    assassin_mark_tracker = 
    {
        hidden = true,

        CalculateBonusDamage = function( self )
            local bonus_damage = 0
            for i, card in self.battle:GetHandDeck():Cards() do
                if card.id == "assassins_mark" or card.id == "assassins_mark_plus" then
                    bonus_damage = bonus_damage + (card.bonus_damage or 0)
                end
            end
            return bonus_damage
        end,

        event_handlers =
        {

            [ EVENT.END_ENEMY_TURN ] = function( self, battle )
                self.locked_in = false
            end,

            [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                if card.id == "assassins_mark" or card.id == "assassins_mark_plus" then
                    self.owner:PrepareTurn()
                end
            end,

            [ EVENT.END_PLAYER_TURN ] = function( self, battle )
                self.locked_in = true
                self.bonus_damage = self:CalculateBonusDamage()
            end,

            [ EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner then
                    local bonus_damage = self.locked_in and self.bonus_damage or self:CalculateBonusDamage()                   
                    dmgt:AddDamage( bonus_damage, bonus_damage, self )
                end
            end
        },
    },

    grout_burr = 
    {
        name = "Grout Bog Burr",
        desc = "Action on the surface has disturbed the Bog, and caused a burr to emerge. This burr will return to the bog in {1} turns.",

        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks)
        end,

        event_priorities =
        {
            -- Allow for all other events to happen first in case we die before we get to leave
            [ EVENT.END_TURN  ] = 999,
        },

        event_handlers = 
        {
            [ EVENT.END_ENEMY_TURN ] = function( self, fighter )
                self:CheckShutdown()
            end,
            
            [ EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition( self.id, 1 )
                    if self.stacks == 0 then
                        self.owner:Depart()
                    end
                end
                
                if self.stacks > 0 then
                    self:CheckShutdown()
                end
            end,

            -- [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
            --     if fighter:GetTeam() and fighter:GetTeam() == self.owner:GetTeam() then
            --         if status == FIGHT_STATUS.FLED or status == FIGHT_STATUS.SURRENDER or status == FIGHT_STATUS.DEAD then
            --             self:CheckShutdown()
            --         end
            --     end
            -- end,

            [ EVENT.CHECKED_FOR_SURRENDER ] = function( self, battle )
                self:CheckShutdown()
            end,

            [ EVENT.END_RESOLVE ] = function( self, battle, card )
                self:CheckShutdown()
            end,
        },

        CountActiveSentients = function( self )
            local count = 0
            for i, fighter in self.owner:GetTeam():Fighters() do
                if fighter:IsActive() and not fighter:HasCondition("no_alone_drone") and not fighter:HasCondition("grout_burr") then
                    count = count + 1
                end
            end
            return count
        end,

        CheckShutdown = function( self )
            if self:CountActiveSentients() == 0 then
                if self.owner:IsActive() then
                    self.owner:Flee()
                end
            elseif self.owner:GetStatus() == FIGHT_STATUS.SURRENDER then
                self.owner:Rally()
            end
        end,        

    },

    no_alone_drone = 
    {
        icon = "battle/conditions/quitter.tex",
        name = "Quitter",
        desc = "Will stop fighting if only drones are left on this team.",
        loc_strings =
        {
            SHUTTING_DOWN = "SHUTTING DOWN",
            REACTIVATING = "REACTIVATING",
        },

        event_handlers = 
        {
            [ EVENT.FIGHTER_RALLY ] = function( self, fighter )
                if fighter:GetTeam() and fighter:GetTeam() == self.owner:GetTeam() then
                    self:CheckShutdown()
                end
            end,

            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if fighter:GetTeam() and fighter:GetTeam() == self.owner:GetTeam() then
                    if status == FIGHT_STATUS.FLED or status == FIGHT_STATUS.SURRENDER or status == FIGHT_STATUS.DEAD then
                        self:CheckShutdown()
                    end
                end
            end,

            [ EVENT.CHECKED_FOR_SURRENDER ] = function( self, battle )
                self:CheckShutdown()
            end,
        },

        CountActiveSentients = function( self )
            local count = 0
            for i, fighter in self.owner:GetTeam():Fighters() do
                if fighter:IsActive() and not fighter:HasCondition("no_alone_drone") and not fighter:HasCondition("grout_burr") then
                    count = count + 1
                end
            end
            return count
        end,

        CheckShutdown = function( self )
            if self:CountActiveSentients() == 0 then
                if self.owner:IsActive() then
                    self.owner:Delay( 1.0 )
                    self.owner:SaySpeech( 1.0, self:GetLocalizedString( "SHUTTING_DOWN" ))
                    self.owner:Surrender()
                    self.owner:AddCondition("drone_shutdown")
                end
            
            elseif self.owner:GetStatus() == FIGHT_STATUS.SURRENDER then
                self.owner:Delay( 1.0 )
                self.owner:SaySpeech( 1.0, self:GetLocalizedString( "REACTIVATING" ))
                self.owner:Rally()
                self.owner:RemoveCondition("drone_shutdown")
            end
        end,        

    },
    
    drone_shutdown = 
    {
        name = "Shutdown Sequence",
        desc = "Will not act while their allies are surrendering.",
    },

    regeneration_barrier =
    {
        name = "Regeneration Barrier",
        desc = "Fully defending attacks will {HEAL} {1} for {2} health.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.owner.agent:GetName(), math.floor(self.heal_amt * self.buff_strength))
        end,

        heal_amt = 3, 
        buff_strength = 1,

        event_handlers = 
        {
            [ EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if hit.target == self.owner and hit.defended then
                    self.owner:HealHealth(math.floor(self.heal_amt * self.buff_strength), self)
                end
            end,

            [ EVENT.BEGIN_TURN ] = function( self, fighter )
                if self.owner == fighter then
                    self.owner:RemoveCondition("regeneration_barrier")
                end
            end
        },
    },

    unstable_drone = 
    {
        name = "Unstable Drone",
        desc = "Receiving damage reduces this by 1. Drone is destroyed at 0 stacks, dealing 5 damage to {1}.",
        desc_fn = function( self, fmt_str )
            local drone_owner = self.owner and self.owner.drone_owner
            return loc.format( fmt_str, drone_owner and drone_owner.agent:GetName())
        end,
        hide_gained = true,

        event_handlers = 
        {
            [ EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                if self.owner == fighter then
                    if self.stacks == 1 then
                        self.used = true
                        self.owner:RemoveCondition(self.id, 1)
                        self.owner:Kill()

                        self.battle:BroadcastEvent( EVENT.DELAY, .29)
                        local drone_owner = self.owner and self.owner.drone_owner
                        if drone_owner then
                            drone_owner:ApplyDamage(5, nil, nil, nil, {"spark"})
                        end
                    else
                        self.owner:RemoveCondition(self.id, 1)
                    end
                end
            end,

            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if not self.used and fighter == self.owner and status == FIGHT_STATUS.DEAD then
                    self.battle:BroadcastEvent( EVENT.DELAY, .29)
                    local drone_owner = self.owner and self.owner.drone_owner
                    if drone_owner then
                        drone_owner:ApplyDamage(5, nil, nil, nil, {"spark"})
                    end
                end
            end,
        },
    },

    spark_aura =
    {
        name = "Spark Aura",
        desc = "Allied mechs deal bonus damage",
        hidden = true,
        mech_boost = 3,

        event_handlers =
        {
            [ EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner:GetTeam() == self.owner:GetTeam() and self.owner.agent:GetSpecies() == SPECIES.MECH then
                    dmgt:ModifyDamage(self.mech_boost, self.mech_boost, self)
                end
            end
        }
    },

    --Merchant
    pickpocket =
    {
        name = "Pickpocket",
        desc = "{1} steals {2} shills if their attacks are not defended.\nAll money is recovered if {1} is killed.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.shill_amount)
        end,

        ctype = CTYPE.INNATE,
        shill_amount = 10,

        OnUnapply = function( self )
            if self.shills_stolen then
                self.battle:GetPlayerTeam():Primary().agent:DeltaMoney( self.shills_stolen )

                self.battle:GetPlayerTeam():Primary():BroadcastEvent( EVENT.DELTA_MONEY, self.shills_stolen )
            end
        end,

        event_handlers = 
        {
            [ EVENT.ON_HIT ] = function( self, battle, attack )
                if attack.attacker == self.owner then
                    local all_defended = attack:CheckHitResults( "defended" ) or attack:CheckHitResults( "evaded" )
                    if not all_defended and attack.card:IsDamageCard() then
                        if attack.target.agent and attack.target.agent:IsPlayer() then
                            local money = math.min( self.shill_amount, attack.target.agent:GetMoney())
                            if money > 0 then
                                attack.target.agent:DeltaMoney( -money )
                                self.owner.agent:DeltaMoney( money )

                                attack.target:BroadcastEvent( EVENT.DELTA_MONEY, -money )
                                self.shills_stolen = (self.shills_stolen or 0) + money
                            end
                        end
                    end
                end
            end,
        },
    },

    --Chemist
    bombard = 
    {
        name = "Bombard",
        desc = "{1}'s attacks insert a random potion into your draw pile.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        potions = {"stimulating_brew", "noxious_vial", "muddler", "bonded_elixir"},
        ctype = CTYPE.INNATE,

        event_handlers = 
        {
            [ EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if attack.attacker == self.owner and hit.target:IsPlayer() and self.owner:GetTeam() ~= battle:GetPlayerTeam() and not attack.is_counter and not hit.evaded then
                    --
                    local incepted_card = Battle.Card( self.potions[math.random(#self.potions)], hit.target )
                    incepted_card.show_dealt = true
                    battle:DealCards( {incepted_card}, battle:GetDrawDeck() )
                end
            end,
        },
    },
}


local attacks =
{
    ai_knife_slice = table.extend(NPC_ATTACK)
    {
        name = "Slice",
        anim = "slice",
        flags = CARD_FLAGS.MELEE,
    },

    ai_shoot = table.extend(NPC_ATTACK)
    {
        name = "Shoot",
        anim = "blast",
        pre_anim = "blast_pre",
        post_anim = "blast_pst",

        flags = CARD_FLAGS.RANGED,

        OnPostResolve = function( self, battle, attack)
            self.owner:RemoveCondition( "ammo", 1 )
        end,
    },

    ai_defend_lo = table.extend(NPC_BUFF)
    {
        name = "Defend",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
        target_type = TARGET_TYPE.SELF,

        features =
        {
            DEFEND = 3,
        }
    },

    ai_defend_med = table.extend(NPC_BUFF)
    {
        name = "Defend",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
        target_type = TARGET_TYPE.SELF,

        features =
        {
            DEFEND = 5,
        }
    },

    ai_defend_hi = table.extend(NPC_BUFF)
    {
        name = "Defend",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
        target_type = TARGET_TYPE.SELF,

        features =
        {
            DEFEND = 8,
        }
    },

    ai_defend_other = table.extend(NPC_BUFF)
    {
        name = "Guard",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
        target_type = TARGET_TYPE.FRIENDLY,

        features =
        {
            DEFEND = 5,
        }
    },

    ai_power_boost = table.extend(NPC_BUFF)
    {
        name = "Power Boost",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
        target_type = TARGET_TYPE.SELF,

        features =
        {
            POWER = 1,
        }
    },

    oshnu_acid = table.extend(NPC_BUFF)
    {
        name = "Oshnu Acid",
        anim = "throw",
        anims = { "anim/weapon_grenade_bile.zip"},

        flags = CARD_FLAGS.RANGED | CARD_FLAGS.DEBUFF,

        features =
        {
            WOUND = 2,
        },
    },

    volatile_substance = table.extend(NPC_RANGED)
    {
        name = "Volatile Substance",
        anim = "throw",
        anims = { "anim/weapon_grenade_bile.zip"},

        target_mod = TARGET_MOD.RANDOM2,
        flags = CARD_FLAGS.RANGED,

        min_damage = 2,
        max_damage = 4,
    },

    vapours = table.extend(NPC_BUFF)
    {
        name = "Vapours",
        anim = "taunt",

        target_mod = TARGET_MOD.RANDOM2,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF | CARD_FLAGS.HEAL,

        heal_amts = { 3, 3, 4, 4, 5 },

        OnPostResolve = function( self, battle, attack )
            for i, hit in attack:Hits() do
                hit.target:HealHealth( self:ScaleBuff( self.heal_amts ) * (1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )), self)
            end
        end,
    },

    assassins_mark = 
    {
        name = "Assassin's Mark",
        desc = "{1}'s attacks deal {2} extra damage while this is in your hand.",
        flavour = "'Hanbi has his eyes on you.'",
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self.source and self.source:GetName() or loc.cap( LOC "UI.CARDS.OWNER" ), self.bonus_damage)
        end,
        max_xp = 0,

        bonus_damage = 3,

        target_type = TARGET_TYPE.SELF,

        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.STATUS,
    },

    assassins_mark_plus = 
    {
        bonus_damage = 4,
        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.STATUS,
    },

    status_lumin_burn = 
    {
        name = "Lumin Burn",
        flavour = "'The Blue solution! May it burn the brightest throughout the dark!'",
        desc = "Take {1} damage.",
        alt_desc = "If this card is in your hand at the end of the turn, divide it into 2.",
        
        desc_fn = function(self, fmt_str)
            if (self.userdata.count or 0) > 1 then
                return loc.format(fmt_str, self.userdata.count ) .. "\n" .. (self.def or self):GetLocalizedString("alt_desc")
            else
            return loc.format(fmt_str, self.userdata and self.userdata.count or 1)
            end
        end,

        OnInit = function( self )
            self.userdata.count = 8
        end,

        target_type = TARGET_TYPE.SELF,

        flags =  CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,

        OnPostResolve = function ( self, battle, attack )
            self.owner:ApplyDamage( self.userdata.count or 1, nil, nil, nil, {"lumin"} )
        end,

        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, battle )
                local new_amt = math.floor(self.userdata.count / 2)
                if new_amt > 0 then
                self:NotifyTriggered()
                self.engine:BroadcastEvent( EVENT.DELAY, 0.25 )
                battle:ExpendCard(self)
                    local cards = {}
                    for k = 1, 2 do
                        local incepted_card = Battle.Card( "status_lumin_burn", self.owner) 
                        incepted_card.userdata.count = new_amt
                        table.insert(cards, incepted_card )                                  
                    end
                    battle:DealCards( cards, battle:GetDiscardDeck() )

                end
            end
        }

    },

    whip = table.extend(NPC_MELEE)
    {
        name = "Whip",
        anim = "whip",
    },


    gnaw = table.extend(NPC_ATTACK)
    {
        name = "Gnaw",
        anim = "bite",
        flags = CARD_FLAGS.MELEE,

        features =
        {
            BLEED = 1,
        }
    },

    double_whammy = table.extend( NPC_MELEE )
    {
        name = "Double Whammy",
        anim = "crush",

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

        features =
        {
            WOUND = 2,
        },

        hit_count = 2,
        damage_mult = 0.5,
    },

    bandit_captain_thump = table.extend( NPC_MELEE )
    {
        name = "Thump",
        anim = "thump",

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

        damage_mult = 0.6,
        OnPostResolve = function( self, battle, attack )
            if not attack:CheckHitResult( attack.target, "evaded" ) and not attack:CheckHitResult( attack.target, "defended" ) then
                if attack.target:IsPlayer() then
                    local cards =
                    {
                        Battle.Card( "status_injury", attack.target )
                    }
                    battle:DealCards( cards, battle:GetDrawDeck() )

                else
                    attack.target:AddCondition( "STUN", 1 )
                end
            end
        end,
    },

    summon_goon =
    {
        name = "Summon Goon",
        anim = "taunt",
        flags = CARD_FLAGS.SKILL,

        OnPostResolve = function( self, battle, attack )
            local content_id = "BANDIT_GOON"
            local agent = Agent( content_id )
            local new_fighter = Fighter.CreateFromAgent( agent, self.owner:GetScale() )
            self.owner:GetTeam():AddFighter( new_fighter )
            self.owner:GetTeam():ActivateNewFighters( battle )
        end,
    },

    --normal oshnus
    oshnu_bite = table.extend(NPC_MELEE)
    {
        name = "Oshnu Bite",
        anim = "bite",
    },

    -- Chemist potions
    stimulating_brew = 
    {
        name = "Stimulating Brew",
        icon = "battle/bombard_stimulating_brew.tex",
        desc = "{HEAL} 10 health.",
        flavour = "'That definitely hit the spot, I could go all night!'",
        cost = 3,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

        OnPostResolve = function( self, battle, attack )
            attack.target:HealHealth(10, self)
        end
    },

    noxious_vial =
    {
        name = "Noxious Vial",
        anim = "throw",
        icon = "battle/bombard_noxious_vial.tex",
        desc = "When drawn, take 5 damage.",
        flavour = "'This one isn’t for drinking.'",
        cost = 2,
        min_damage = 5,
        max_damage = 10,

        flags = CARD_FLAGS.RANGED | CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS,

        event_handlers = 
        {
            [ EVENT.DRAW_CARD ] = function( self, battle, card, start_of_turn )
                if card == self then
                    self.owner:ApplyDamage(5, self.owner, card, false)
                end
            end,
        },
    },

    muddler = 
    {
        name = "Muddler",
        anim = "drink",
        anims = { "anim/grog_beer_glass2.zip"},
        icon = "battle/bombard_muddler.tex",
        desc = "While in hand, all enemy targets are random.",
        flavour = "'You’ll definitely be feeling that in the morning.'",
        cost = 2,

        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS,

        target_type = TARGET_TYPE.SELF,

        event_handlers =
        {
            [ EVENT.CALC_TARGET_MOD ] = function( self, acc, card, target )
                if card.owner == self.owner and card:IsAttackCard() and acc.value ~= TARGET_MOD.TEAM and acc.value ~= TARGET_MOD.ALL then
                    acc:ModifyValue( TARGET_MOD.RANDOM1, self )
                end
            end,
        },
    },

    bonded_elixir = 
    {
        name = "Bonded Elixir",
        icon = "battle/bombard_bonded_elixir.tex",
        desc = "While in hand, apply 1 {WOUND} to yourself for every card played.",
        flavour = "'Strong stuff!'",
        cost = 2,

        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS,
        target_type = TARGET_TYPE.SELF,
        event_handlers =
        {
            [ EVENT.START_RESOLVE ] = function( self, battle, card )
                if card.owner == self.owner and card ~= self then
                    self.owner:AddCondition("WOUND")
                end
            end,
        },
    },
}

------------------------------------------------------------------------
--

for condition_id, t in pairs( conditions ) do
    Content.AddBattleCondition( condition_id, t )
end

for i, id, data in sorted_pairs(attacks) do
    data.cost = data.cost or 1
    data.rarity = CARD_RARITY.UNIQUE
    if not CheckBits( data.flags or 0, CARD_FLAGS.STATUS ) then
        data.flags = SetBits( data.flags or 0, CARD_FLAGS.NPC )
    end
    data.series = CARD_SERIES.NPC

    Content.AddBattleCard(id, data)
end


