local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

local attacks = {

    little_score = 
    {
        name = "Little Score",
        rarity = CARD_RARITY.RARE,
        desc = "Gain {1#points_total}.",
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self.scoring.value)
        end,
        
        scoring = {
            name = "Played Little Score",
            value = 30,
        },

        anim = "taunt",
        flavour = "'If you're going to win, at least win with style!'",
        max_charges = 5,
        cost = 1,
        flags = CARD_FLAGS.ITEM | CARD_FLAGS.REPLENISH,
        target_type = TARGET_TYPE.SELF,
        item_tags = ITEM_TAGS.COMBAT,
        
        OnPostResolve = function( self, battle, attack)
            TheGame:GetGameState():AddScore(self.scoring.id, 1)
        end
    },
    
    big_score = 
    {
        name = "Big Score",
        rarity = CARD_RARITY.RARE,
        anim = "taunt",
        desc = "Gain {1#points_total} if hatched.",
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self.scoring.value)
        end,

        scoring = {
            name = "Hatched Big Score",
            value = 300,
        },

        flavour = "'Now you are just showing off!'",
        max_xp = 4,
        cost = 2,
        flags = CARD_FLAGS.ITEM | CARD_FLAGS.REPLENISH | CARD_FLAGS.HATCH,
        target_type = TARGET_TYPE.SELF,
        item_tags = ITEM_TAGS.COMBAT,
        hatch = true,
        hatch_fn = function( self, battle )
            self:TransferCard( battle.trash_deck )
            self:Consume()
            TheGame:GetGameState():AddScore(self.scoring.id, 1)
        end,
    },

    -- Curse cards
    twisted_figurine =
    {
        name = "Twisted Figurine",
        desc = "At the end of your turn, take 2 damage for each attack card played this turn.{2}\n{KEEPSAKE} {1#points_total}",
        loc_strings =
        {
            CARDS_PLAYED = "\n({1} attack {1*card|cards} played)",
        },
        desc_fn = function( self, fmt_str )
            if self.engine then
                local str = loc.format( self.def:GetLocalizedString( "CARDS_PLAYED" ), self:CountCardsPlayed() )
                return loc.format(fmt_str, self.userdata and self.userdata.points or self.base_points, str)
            else
                return loc.format(fmt_str, self.userdata and self.userdata.points or self.base_points, "")
            end
        end,
        flavour = "'Hedge witches craft these to inflict malady on their targets. If you come into one meant for you it’s best to not lose it before you can dispose of it properly.'",

        cost = 0,
        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS | CARD_FLAGS.ITEM,
        target_type = TARGET_TYPE.SELF,

        base_points = 150,

        CountCardsPlayed = function( self )
            if not self.engine then
                return 0
            end
            return self.engine:CountCardsPlayed( function(card) return card:IsAttackCard() end )
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                local cards_played = self:CountCardsPlayed()
                if cards_played > 0 then
                    if self.deck:GetDeckType() == DECK_TYPE.IN_HAND and not battle:CheckGameOver() then
                        self:NotifyTriggered()
                        self.owner:ApplyDamage(cards_played * 2, self.owner, self)
                    end
                end
            end,
        },
    },

    barbed_doll =
    {
        name = "Barbed Doll",
        desc = "While this is in your hand, all your attacks deal half damage.\n{KEEPSAKE} {1#points_total}",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.userdata and self.userdata.points or self.base_points)
        end,
        flavour = "'Carried by Jarackle combat trainees that skip curfew. Usually carried inside the boots.'",

        cost = 0,
        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS | CARD_FLAGS.ITEM,
        target_type = TARGET_TYPE.SELF,
        base_points = 120,

        event_priorities =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner then
                    dmgt:ModifyDamage( math.round( dmgt.min_damage * 0.5 ),
                                       math.round( dmgt.max_damage * 0.5 ),
                                       self )
                end
            end,
        }
    },

    shrokian_failure_idol =
    {
        name = "Shrokian Failure Idol",
        desc = "At the end of your turn, {EXPEND} all cards in your hand and this card.\n{KEEPSAKE} {1#points_total}",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.userdata and self.userdata.points or self.base_points)
        end,
        flavour = "'Gifted to those who have taken an unacceptable loss and carried with shame.'",

        cost = 0,
        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS | CARD_FLAGS.ITEM,
        target_type = TARGET_TYPE.SELF,
        base_points = 100,

        event_handlers =
        {
            [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                local options = {}
                for i,card in battle:GetHandDeck():Cards() do
                    if card ~= self then
                        table.insert(options, card)
                    end
                end
                for i,card in ipairs(options) do
                    battle:ExpendCard(card)
                end
                battle:ExpendCard(self)
            end
        },
    },

    blasted_eye =
    {
        name = "Blasted Eye",
        desc = "At the end of your turn, gain 3 {IMPAIR}.\n{KEEPSAKE} {1#points_total}",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.userdata and self.userdata.points or self.base_points)
        end,
        flavour = "'Pulled from a disgraced Kra’deshi. Some say if the eye doesn’t grow back that justice was served and the stigma is lifted.'",

        cost = 0,
        flags = CARD_FLAGS.CONSUME | CARD_FLAGS.STATUS | CARD_FLAGS.ITEM,
        target_type = TARGET_TYPE.SELF,
        base_points = 80,

        event_handlers =
        {
            [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                self.owner:AddCondition(self.id, 3) -- So that IMPAIR gets added after IMPAIR decay occurs.
            end
        },
        condition =
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:AddCondition("IMPAIR", self.stacks, self)
                        self.owner:RemoveCondition( self.id )
                    end
                end
            },
        }

    },

    legacy_blade =
    {
        name = "Legacy Blade",
        anim =  "stab",
        flavour = "'Names are etched across a blade that flickers and hums with an ancient energy.'",
        desc = "Whenever you kill an enemy with this card, gain {1#points}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.scoring.value)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        min_damage = 3,
        max_damage = 5,

        scoring = {
            name = "Legacy Blade Kill",
            value = 100,
        },

        OnPostResolve = function( self, battle, attack )
            if self.target and not self.target:IsActive() then
                TheGame:GetGameState():AddScore(self.scoring.id)
            end
        end,
    },

    bring_it_on =
    {
        name = "Bring It On!",
        flavour = "’Take your shot loser!’",
        desc = "Remove all your {DEFEND}, gain {1#points} per {DEFEND} removed.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.scoring.value)
        end,

        cost = 1,

        scoring = {
            name = "Bring It On!",
            value = 5,
        },

        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local count = self.owner:GetConditionStacks("DEFEND")
            if count then
                self.owner:RemoveCondition("DEFEND", count, self)
                TheGame:GetGameState():AddScore(self.scoring.id, count)
            end
        end
    },

    boots_blaster =
    {
        name = "Boot's Blaster",
        anim =  "stab",
        flavour = "'Tough as Hesh but hard to work with.'",
        desc = "Gain {1#points} per damage dealt by this card.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.scoring.value )
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
        min_damage = 2,
        max_damage = 2,

        scoring = {
            name = "Boot's Blaster",
            value = 10,
        },

        event_handlers = 
        {
            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if hit.card == self then
                    TheGame:GetGameState():AddScore(self.scoring.id, hit.damage)
                end
            end
        },
    },
}

for id, data in pairs(attacks) do
    data.cost = data.cost or 0
    data.rarity = data.rarity or CARD_RARITY.UNIQUE
    data.series = CARD_SERIES.DAILY
    data.flags = (data.flags or 0) | CARD_FLAGS.SCORE 
    if data.base_points then
        
        data.global_event_handlers =
        {
            [ "card_added" ] = function( self, card )
                if card == self then
                    card.userdata.points = card.userdata.points or self.base_points
                end
            end,

            [ "GAME_OVER" ] = function( self, gamestate )
                gamestate:AddScore(data.scoring.id, 1, self.userdata.points)
            end
        }
        
        data.scoring = {
            name = "Kept " .. data.name,
            no_combine = true,  --don't combine elements with the same userdata into count instances 
            GetValue = function(def, instance)
                return instance.userdata or 0
            end,
        }

    end

    Content.AddBattleCard(id, data)
end

