local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT


local ITEMS =
{
    emergency_shield_generator =
    {
        name = "Emergency Shield Generator",
        desc = "If you take lethal damage while this is in your deck, block all damage until the start of your turn and destroy this card.",
        cost = 0,
        flavour = "'Saved me more than once, although I did pick it up off of a dead body.'",
        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.UNPLAYABLE,

        deck_handlers = { DECK_TYPE.DRAW, DECK_TYPE.DISCARDS, DECK_TYPE.IN_HAND },

        series = "ROOK",

        OnAdded = function( self, battle )
            Battle.Card.OnAdded( self, battle )
            self.owner:AddCondition("emergency_shield_generator", 1, self)
        end,

        condition =
        {
            desc = "Taking lethal damage will trigger your Emergency Shield Generator, blocking all damage until the start of your turn.",
            loc_strings =
            {
                FLOATER = "Emergency Shield!",
            },
            OnPreDamage = function( self, damage, attacker, battle, source, piercing )
                if self.activated then
                    if is_instance( source, Battle.Hit ) then 
                        source:SetHitResult( "defended" )
                    end
                    self.owner:BroadcastEvent( BATTLE_EVENT.PLAY_FLOATER, self:GetLocalizedString("FLOATER"), 0.8, UICOLOURS.STATUS_HEAL )
                    return 0

                elseif damage >= self.owner:GetHealth() then
                    self.activated = true
                    local applicant = self:GetRecentApplicant()
                    if is_instance( applicant, Battle.Card ) and applicant.deck ~= nil then
                        self:RemoveApplicant( applicant )
                        applicant:TransferCard( battle.trash_deck )
                        applicant:Consume( battle )
                    end
                    if is_instance( source, Battle.Hit ) then 
                        source:SetHitResult( "defended" )
                    end
                    self.owner:BroadcastEvent( BATTLE_EVENT.PLAY_FLOATER, self:GetLocalizedString("FLOATER"), 0.8, UICOLOURS.STATUS_HEAL )

                    return 0
                end
                return damage
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    if self.activated then
                        self.owner:RemoveCondition(self.id, 1, self)
                        self.activated = false
                    end
                end
            },
        },
    },

    shock_grenade =
    {
        name = "Shock Grenade",
        desc = "Hits all enemies. {STUN} enemies with {METALLIC}.",
        anim = "throw",
        flavour = "'Rattles up those buckets really nice, even folks wearing too much metal have a problem with these guys.'",
        anims = { "anim/weapon_grenade_coil.zip"},
        hit_tags = {"spark"},
        max_charges = 1,
        cost = 1,
        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.PIERCING | CARD_FLAGS.RANGED,
        target_mod = TARGET_MOD.TEAM,

        min_damage = 5,
        max_damage = 8,

        OnPostResolve = function( self, battle, attack )
            for i, hit in attack:Hits() do
                if hit.target:HasCondition("METALLIC") then
                    hit.target:AddCondition("STUN", 1, self)
                end
            end
        end,
    },

    bog_brew =
    {
        name = "Bog Brew",
        anim = "drink",
        flavour = "'That really is quite refreshing.'",
        desc = "{HEAL {1}}.",
        anims = { "anim/grog_beer_glass2.zip"},
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self.heal_amount )
        end,

        cost = 1,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
        flags = CARD_FLAGS.REPLENISH,
        rarity = CARD_RARITY.COMMON,

        max_charges = 2,
        heal_amount = 8,
        
        OnPostResolve = function( self, battle, attack)
            self.target:HealHealth( self.heal_amount, self )
        end,
    },

    fuel_canister =
    {
        name = "Fuel Canister",
        flavour = "'There's always time for a Barbeque.'",
        anim = "throw",
        desc = "Apply {1} {BURN}. Hits all enemies.",
        anims = { "anim/weapon_grenade_fuel_canister.zip"},
        hit_tags = { "spark" },
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.burn_amt)
        end,

        max_charges = 1,
        cost = 1,

        target_mod = TARGET_MOD.TEAM,
        flags = CARD_FLAGS.REPLENISH | CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 2,
        max_damage = 4,
        burn_amt = 5,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("BURN", self.burn_amt, self)
        end
    },

    jump_cables = 
    {
        name = "Jump Cables",
        desc = "Gain {1} {CHARGE}.",
        flavour = "'It's an older trick but it gets the job done.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.charge_gain)
        end,

        series = "ROOK",

        max_charges = 2,
        cost = 1,

        flags = CARD_FLAGS.AMBUSH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.SELF,

        charge_gain = 4,

        OnPostResolve = function( self, battle, attack)
            local tracker = self.owner:GetCondition("lumin_tracker")
            if tracker then tracker:AddLuminCharges(self.charge_gain, self) end
        end,
    },

    skimmer = 
    {
        name = "Skimmer",
        anim = "mark",
        flavour = "'Taken right from the targeting matrix of old shock-cannons. Those things were too slow to use anyways.'",
        desc = "Apply {1} {MARK} to a random target.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.mark_amt)
        end,

        max_charges = 3,
        cost = 1,
        series = "ROOK",

        target_mod = TARGET_MOD.RANDOM1,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.COMMON,

        mark_amt = 3,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("MARK", self.mark_amt, self)
        end
    },
}


for i, id, data in sorted_pairs( ITEMS ) do
    data.series = data.series or CARD_SERIES.GENERAL
    data.local_item = LOCAL_ITEMS.BOG
    data.item_tags = (data.item_tags or 0) | ITEM_TAGS.COMBAT
    data.flags = (data.flags or 0) | CARD_FLAGS.ITEM 
    assert(data.rarity, loc.format("Rarity missing on {1}", data.name))
    
    Content.AddBattleCard( id, data )
end

