local fun = require "util/fun"

local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT


local function CreateDrawnHatcher(fn)
    local ret = function( self, battle, card, start_of_turn )
                 if card == self then
                    self.engine:PushPostHandler( function()
                        self:NotifyTriggeredPre()
                        --self.owner:ApplyDamage(1, self.owner, self)
                        if self.engine then
                            self.engine:BroadcastEvent( BATTLE_EVENT.DELAY, 0.5 )
                            self:AddXP(1)

                            -- NOTE: fn() might do anything, including kiling our owner.
                            fn(self)

                            if card.deck ~= nil and card:UpgradeReady() then
                                if card.hatch_fn then 
                                    if not card.hatched then
                                        card:hatch_fn(self.engine)
                                        card.hatched = true
                                    else
                                        card:TransferCard( self.engine.trash_deck )
                                        card:Consume()
                                    end
                                end
                            end
                        end
                        self:NotifyTriggeredPost()
                    end )
                end
            end
    return ret  
end


local PARASITES =
{
    twig = 
    {
        name = "Twig",
        desc = "When this card is drawn, take 1 damage.",
        cost = 0,
        flavour = "'Stop picking at it, it's only going to get worse.'",
        max_xp = 10,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.UNPLAYABLE,
        rarity = CARD_RARITY.BASIC,
        event_handlers = {
           [ BATTLE_EVENT.DRAW_CARD ] = 
                CreateDrawnHatcher(function(self) 
                    self.owner:ApplyDamage(1, self.owner, self)
                end)
        }
    },

    -- TIER 1
    neurotic_haze =
    {
        name = "Neurotic Haze",
        anim = "taunt",
        flavour = "'He pulled down the rags around his mouth to speak, the air filling with an awful electricity.'",
        desc = "Discard {1} random cards from your hand and gain {2} actions.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.discard_count, self.action_gain)
        end,

        cost = 1,

        flags = CARD_FLAGS.HATCH,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.SELF,

        discard_count = 2,
        action_gain = 2,

        OnPostResolve = function( self, battle, attack )
            local cards_to_discard = table.multipick(battle:GetHandDeck().cards, self.discard_count)
            for i,card in ipairs(cards_to_discard) do
                battle:DiscardCard(card)
            end
            battle:ModifyActionCount(self.action_gain)
        end
    },

    tendrils = 
    {
        name = "Tendrils",
        anim = "taunt",
        flavour = "'I looked on in horror as new appendages sloughed out from my body, far too many and far too easily. It was as if they'd always been.'",
        desc = "While this is in your hand, gain {1} {POWER}. Take {2} damage whenever you play a card.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.power_amt, self.damage_amt)
        end,

        cost = 0,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.SELF,

        power_amt = 2,
        damage_amt = 2,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                if card == self then
                    if target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND and not (source_deck and source_deck:GetDeckType() == DECK_TYPE.IN_HAND) then
                        self.owner:AddCondition( "POWER", self.power_amt )

                    elseif source_deck and source_deck:GetDeckType() == DECK_TYPE.IN_HAND and not (target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND) then
                        self.owner:RemoveCondition( "POWER", self.power_amt )
                    end
                end
            end,

            [ BATTLE_EVENT.PRE_RESOLVE ] = function ( self, minigame, attack, targets )
                if attack.card ~= self and self.deck and self.deck:GetDeckType() == DECK_TYPE.IN_HAND then
                    self.attack_played = attack
                end
            end,

            [ BATTLE_EVENT.POST_RESOLVE ] = function ( self, battle, attack )
                if attack == self.attack_played and attack.attacker == self.owner and self.deck:GetDeckType() == DECK_TYPE.IN_HAND then
                    self.attack_played = nil
                    self.owner:ApplyDamage(self.damage_amt, self.owner, self)
                end
            end
        }
    },

    bog_brain = 
    {
        name = "Bog Brain",
        anim = "taunt",
        flavour = "'The headaches worsened, gouging out old memories. But in their stead, were visions of greater importance.'",
        desc = "When this card is drawn, gain {1} {IMPAIR} and draw {2} cards.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.impair_amt, self.draw_count)
        end,

        cost = 1,

        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.SELF,

        impair_amt = 2,
        draw_count = 2,

        event_handlers = 
        {
            [ BATTLE_EVENT.DRAW_CARD ] = function( self, battle, card, start_of_turn )
                if card == self then
                    self:NotifyTriggeredPre()
                    self.engine:BroadcastEvent( BATTLE_EVENT.DELAY, 0.5 )
                    self.owner:AddCondition("IMPAIR", self.impair_amt)
                    self.engine:DrawCards(self.draw_count)
                    self:NotifyTriggeredPost()

                end
            end,
        },
    },

    -- TIER 2

    branch = 
    {
        name = "Branch",
        anim = "taunt",
        flavour = "'My veins were hot fire and erupted from my arm in a viciousness so brutal that it had to have been a terrible dream.'",
        desc = "Gain {1} {TEMP_POWER}. Take {2} damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.power_amt, self.damage_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.HATCH,
        rarity = CARD_RARITY.UNCOMMON,
        target_type = TARGET_TYPE.SELF,

        power_amt = 3,
        damage_amt = 8,

        CanPlayCard = function( self, battle, targets )
            return self.owner:GetHealth() + self.owner:GetConditionStacks("DEFEND") > self.damage_amt
        end,

        OnPostResolve = function( self, battle, attack )
            self.owner:ApplyDamage(self.damage_amt, self.owner, self)
            self.owner:AddCondition("POWER", self.power_amt, self)
            self.owner:AddCondition("POWER_LOSS", self.power_amt, self)
        end,
    },

    fever = 
    {
        name = "Fever",
        anim = "taunt",
        flavour = "'Feed a fever and starve a cold? I can never remember which it is...'",
        desc = "Remove all debuffs on yourself, take {1} damage per debuff removed.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage_amt)
        end,
        cost = 0,

        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,
        target_type = TARGET_TYPE.SELF,

        damage_amt = 3,

        CanPlayCard = function( self, battle, targets )
            return self.owner:GetHealth()  + self.owner:GetConditionStacks("DEFEND") > #self.owner:GetAllDebuffs() * self.damage_amt
        end,

        OnPostResolve = function( self, battle, attack )
            local count = self.owner:RemoveAllDebuffs( self )
            self.owner:ApplyDamage(count * self.damage_amt, self.owner, self)
        end
    },

    delirium = 
    {
        name = "Delirium",
        anim = "taunt",
        flavour = "'I thought that I had woken up that day, free from the bog and the eyes that followed incessantly.'",
        desc = "Draw {1} cards. Play a random one for free and {EXPEND} any others.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.card_draw)
        end,

        cost = 1,

        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,
        target_type = TARGET_TYPE.SELF,

        card_draw = 2,

        OnPostResolve = function( self, battle, attack )
            local cards = battle:DrawCards(self.card_draw)
            local chosen_card = table.arraypick(cards)    

            for i, card in ipairs( cards ) do
                battle:Delay(0.4)
                if card == chosen_card then
                    battle:PushPostHandler( function()
                        card:SetFlags(CARD_FLAGS.FREEBIE)
                        card:NotifyTriggeredPre()
                        battle:Delay(0.3)
                        battle:PlayCard(card)
                        card:NotifyTriggeredPost()
                    end )
                    
                else
                    battle:PushPostHandler( function()
                        card:NotifyTriggeredPre()
                        battle:ExpendCard(card)
                        card:NotifyTriggeredPost()
                    end )    
                end
            end
        end
    },

    -- TIER 3
    funky_fungi =
    {
        name = "Funky Fungi",
        anim = "tipsy",
        flavour = "'It's like biting into a lumin battery!'",
        desc = "Upgrade all cards in your hand for the rest of this fight. Take {1} damage for each card upgraded.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage_amt)
        end,

        cost = 1,

        max_xp = 5,

        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,
        target_type = TARGET_TYPE.SELF,

        damage_amt = 2,

        OnPostResolve = function( self, battle, attack )
            local cards = battle:GetHandDeck().cards
            local count = 0
            for i,card in ipairs(cards) do
                if card:HasUpgrades() then
                    card:UpgradeCard()
                    count = count + 1
                end
            end
            self.owner:ApplyDamage( count * self.damage_amt, self.owner, self )
        end,
    },

    bough = 
    {
        name = "Bough",
        anim = "taunt",
        flavour = "'And as my wretched arm constricted around my enemies and blinked, I knew that I had become the Bough.'",
        desc = "Gain {1} {POWER}. Take {2} damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.power_amt, self.damage_amt)
        end,
        
        cost = 1,
        flags = CARD_FLAGS.HATCH,
        rarity = CARD_RARITY.RARE,
        target_type = TARGET_TYPE.SELF,

        damage_amt = 10,
        power_amt = 3,

        CanPlayCard = function( self, battle, targets )
            return self.owner:GetHealth()  + self.owner:GetConditionStacks("DEFEND") > self.damage_amt
        end,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("POWER", self.power_amt, self)
            self.owner:ApplyDamage(self.damage_amt, self.owner, self)
        end
    },

    butcher_of_the_bog =
    {
        name = "Butcher of the Bog",
        anim = "taunt",
        flavour = "'A lumbering shape shambled towards the group and then shifted into the mire. We thought we would be spared until we awoke to Veep in pieces.'",
        desc = "Gain {1} {WOUND}. Your next attack this turn deals double damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.wound_amt)
        end,

        cost = 1,
        flags = CARD_FLAGS.HATCH,
        rarity = CARD_RARITY.RARE,
        target_type = TARGET_TYPE.SELF,
        wound_amt = 4,


        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("WOUND", self.wound_amt, self)
            self.owner:AddCondition(self.id)
        end,

        condition = 
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition("butcher_of_the_bog")
                end,

                [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                    if card.owner == self.owner then
                        dmgt:AddDamage(dmgt.min_damage, dmgt.max_damage, self)
                    end
                end,

                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.owner == self.owner and attack.card:IsAttackCard() then
                        self.owner:RemoveCondition("butcher_of_the_bog", 1, self)
                    end
                end
            }
        },
    },    

    beast_of_the_bog = 
    {
        name = "Beast of the Bog",
        flavour = "'Such a howl! A scream that shimmered through fangs and bone!'",
        desc = "Gain {1} {TEMP_POWER}. You can only play {2} more cards this turn.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.power_amt, self.card_count)
        end,
        
        cost = 1,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,
        target_type = TARGET_TYPE.SELF,

        
        power_amt = 4,
        card_count = 2,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("POWER", self.power_amt, self)
            self.owner:AddCondition("POWER_LOSS", self.power_amt, self)
            local con = self.owner:AddCondition(self.id)
            if con then 
                con.card_count = self.card_count + 1
                con.power_amt = self.power_amt
            end

        end,

        condition = 
        {
            cards_played = 0,
            card_count = 2,
            hidden = true,
            CanPlayCard = function( self, battle, target )
                if self.cards_played >= self.card_count and not battle:GetEnemyTeam():IsSurrendering() then
                    return false, Content.GetBattleCard( "beast_of_the_bog" ):GetLocalizedName()
                else
                    return true, nil
                end
            end,
            event_handlers = 
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id)
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function (self, battle, card )
                    if card.owner == self.owner then
                        self.cards_played = self.cards_played + 1
                    end
                end
            },
        },
    },

    -- TIER 4

    bog_symbiosis =
    {
        name = "Bog Symbiosis",
        desc = "Kill enemy Grout Burr and gain 3 maximum health.",
        flavour = "'It was beneath the abhorrent layers of cocoon that I realized, that as long as I remained, I would remain whole.'",

        cost = 1,

        flags = CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        CanPlayCard = function( self, battle, target )
            return target == nil or target:HasCondition("grout_burr")
        end,

        OnPostResolve = function( self, battle, attack )
            if attack.target and attack.target:HasCondition("grout_burr") then
                attack.target:Kill()
                self.owner:DeltaMaxHealth(3)
                
                local health = self.owner.agent:GetAspect("health")
                if health then
                    health:AddStatModifier("BOG_SYMBIOSIS", 3)
                end
            end
        end
    }, 
}

for i, id, data in sorted_pairs( PARASITES ) do
    data.series = "ROOK"
    if data.rarity ~= CARD_RARITY.UNIQUE then
        data.max_xp = data.max_xp or 4
    end

    data.flags = (data.flags or 0) | CARD_FLAGS.PARASITE 
    assert(data.rarity, loc.format("Rarity missing on {1}", data.name))
    
    Content.AddBattleCard( id, data )
end



