local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS

local BATTLE_EVENT = ExtendEnum( battle_defs.EVENT,
{
    "DRINK",
    "BOTTLES_SPENT",
})
---------------------------------------------------------------------------------

-- Shared Drink function
local function Drink( source, battle, count )
    local cards = { }
    for i=1, count or 1 do
        local card = Battle.Card( "empty_bottle", source.owner )
        card.auto_deal = true
        table.insert(cards, card)
    end
    battle:DealCards( cards, battle:GetDiscardDeck() )
    battle:BroadcastEvent( BATTLE_EVENT.DRINK, battle, count or 1, cards )
end

local function GetBottleCount( self )
    self.bottle_tracker = self.bottle_tracker or self.owner:GetCondition("BOTTLE_TRACKER")
    return self.bottle_tracker and self.bottle_tracker:GetBottleCount() or 0
end

local function SpendBottles( self, count )
    self.bottle_tracker = self.bottle_tracker or self.owner:GetCondition("BOTTLE_TRACKER")
    return self.bottle_tracker and self.bottle_tracker:SpendBottles( count, self ) or 0
end

local function CanPlaySelfDamage( self, self_damage, hits )
    local total_damage = 0

    -- This is a huge hack and fails to apply generally, but there's no way to simulate OnPreDamage and the like atm.
    for i = 1,  (hits or 1) do
        local damage = self_damage
        if self.owner:HasCondition("power_through") then
            damage = math.min( 1, damage )
        end
        total_damage = total_damage + damage
    end
    total_damage = math.max( 0, total_damage - self.owner:GetConditionStacks("DEFEND"))

    if self.owner:GetHealth() > total_damage then
        return true
    else
        return false, CARD_PLAY_REASONS.NOT_ENOUGH_HEALTH
    end
end

local attacks =
{
    empty_bottle =
    {
        name = "Empty Bottle",
        anim = "bottle_flourish",
        desc = "Draw a card.",
        flavour = "'Heads up!'",

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            battle:DrawCards(1)
        end
    },

    smiths_arsenal =
    {
        name = "Smith's Arsenal",
        anim = "taunt",
        desc = "Insert {hammer_throw} or {bottle_smash} into your hand.",
        flavour = "'Never go anywhere without a basher and a brew.'",

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        cost = 1,

        sound = "event:/sfx/battle/status/special/sals_daggers",

        OnPostResolve = function( self, battle, attack)
            local cards = {
                Battle.Card( "hammer_throw", self.owner ),
                Battle.Card( "bottle_smash", self.owner ),
            }
            battle:ImproviseCards( cards, nil, nil, nil, nil, self )
        end,
    },

    hammer_throw =
    {
        name = "Hammer Throw",
        anim = "hammer_throw",
        desc = "Apply {1} {TRAUMA}.",
        flavour = "'Comes right back!'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.trauma_amt)
        end,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.RANGED | CARD_FLAGS.EXPEND,

        min_damage = 2,
        max_damage = 3,
        trauma_amt = 1,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("TRAUMA", self.trauma_amt, self)
        end
    },

    bottle_smash =
    {
        name = "Bottle Smash",
        anim = "bottle_throw",
        desc = "{EXPEND} one {empty_bottle} card from your deck to play this.\nApply {1} {WOUND}.",
        flavour = "'Well I’m not cleaning this up.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.wound_amount)
        end,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.RANGED | CARD_FLAGS.EXPEND,

        min_damage = 2,
        max_damage = 3,
        wound_amount = 1,

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) > 0, CARD_PLAY_REASONS.NO_BOTTLES

        end,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self, 1)
            attack:AddCondition("WOUND", self.wound_amount, self)
        end,
    },

    smiths_arsenal_plus =
    {
        name = "Smith's Hammer",
        anim = "taunt",
        desc = "Insert {maul} or {one_one_one} into your hand.",
        icon = "battle/smiths_hammer.tex",

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        cost = 1,

        sound = "event:/sfx/battle/status/special/sals_daggers",

        OnPostResolve = function( self, battle, attack)
            local cards = {
                Battle.Card( "maul", self.owner ),
                Battle.Card( "one_one_one", self.owner ),
            }
            battle:ImproviseCards( cards, nil, nil, nil, nil, self )
        end,
    },

    maul =
    {
        name = "Maul",
        anim = "bash",
        desc = "If the target has no {TRAUMA}, apply {1} {TRAUMA}.",
        flavour = "'That’s surely going to put someone in the hospital.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.trauma_amt)
        end,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,

        min_damage = 3,
        max_damage = 3,
        trauma_amt = 2,

        OnPostResolve = function( self, battle, attack )
            for i,hit in ipairs(attack.hits) do
                if hit.target and not hit.target:HasCondition("TRAUMA") and not hit.evaded then
                    hit.target:AddCondition("TRAUMA", self.trauma_amt, self)
                end
            end
        end
    },

    one_one_one =
    {
        name = "One One One",
        anim = "gut_shot",
        desc = "Attack three times.",
        flavour = "'It’s all a setup.'",

        hit_anim = true,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND | CARD_FLAGS.RESTRAINED,

        min_damage = 1,
        max_damage = 1,
        hit_count = 3,
    },

    smiths_arsenal_plus2 =
    {
        name = "Smith's Bottle",
        anim = "taunt",
        desc = "Insert {fracture} or {slam} into your hand.",
        icon = "battle/smiths_bottle.tex",

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        cost = 1,

        sound = "event:/sfx/battle/status/special/sals_daggers",

        OnPostResolve = function( self, battle, attack)
            local cards = {
                Battle.Card( "fracture", self.owner ),
                Battle.Card( "slam", self.owner ),
            }
            battle:ImproviseCards( cards, nil, nil, nil, nil, self )
        end,
    },

    slam =
    {
        name = "Slam",
        anim = "drink",
        desc = "{DRINK} 2 and gain {1} {ADRENALINE}.",
        flavour = "'Take it all down in one go!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_amt)
        end,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        adrenaline_amt = 2,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, 2)
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
        end
    },

    fracture =
    {
        name = "Fracture",
        anim = "bottle_stab",
        desc = "{EXPEND} one {empty_bottle} card from your deck to play this.\nApply {WOUND {1}}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.wound_amount)
        end,

        flavour = "'I’m just getting warmed up here.'",

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,

        min_damage = 5,
        max_damage = 5,
        wound_amount = 1,

        CanPlayCard = function( self, battle, target )
            if GetBottleCount(self) > 0 then
                return true
            else
                return false, CARD_PLAY_REASONS.NO_BOTTLES
            end
        end,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles( self, 1 )
            attack:AddCondition("WOUND", self.wound_amount, self)
        end,
    },

    bash = 
    {
        name = "Bash",
        anim = "smash",
        flavour = "'I call it my speciality.'",
        wild = true,

        cost = 1,

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.MELEE,

        min_damage = 2,
        max_damage = 4,
    },

    bash_plus2a =
    {
        name = "Rooted Bash",
        min_damage = 4,
    },

    bash_plus2b =
    {
        name = "Biting Bash",
        desc = "<#UPGRADE>Take 2 damage.</>",

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, 2 )
        end,

        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( 2, self )
        end
    },

    bash_plus2c =
    {
        name = "Bash of Clarity",

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.CONSUME,  

        min_damage = 8,
        max_damage = 8,
    },

    bash_plus2d =
    {
        name = "Traumatic Bash",
        desc = "<#UPGRADE>Apply 2 {TRAUMA}.</>",

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("TRAUMA", 2, self)
        end
    },

    bash_plus2e =
    {
        name = "Rapid Bash",
        desc = "<#UPGRADE>Attack twice.</>",
        hit_anim = true,
        hit_count = 2,
        min_damage = 3,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.RESTRAINED,
    },

    bash_plus2f =
    {
        name = "Quenched Bash",
        desc = "<#UPGRADE>Deals 4 bonus damage if you drank this turn.</>",

        PreReq = function( self, battle, target )
            return self.drank
        end,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and self.drank then
                    dmgt:AddDamage(4, 4, self)
                end 
            end,

            [ BATTLE_EVENT.DRINK ] = function( self, battle, count, cards )
                self.drank = true
            end,

            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                self.drank = false
            end
        },
    },

    bash_plus2g =
    {
        name = "Lucid Bash",
        min_damage = 6,
        max_damage = 8,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
    },

    bash_plus2h =
    {
        name = "Rolling Bash",
        desc = "<#UPGRADE>Deal 1 bonus damage per card played this turn.</>",

        min_damage = 1,
        max_damage = 2,
        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    dmgt:AddDamage(self.engine:CountCardsPlayed(), self.engine:CountCardsPlayed(), self)
                end 
            end,
        },
    },

    bash_plus2i =
    {
        name = "Six Pack Bash",
        desc = "<#UPGRADE>Deals 6 bonus damage if you have 6 or more {empty_bottle} cards in your deck.</>",

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= 6
        end,
        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    if GetBottleCount(self) >= 6 then
                        dmgt:AddDamage(6, 6, self)
                    end
                end 
            end,
        },
    },

    toughen_up =
    {
        name = "Toughen Up",
        anim = "taunt",
        desc = "Apply {1} {DEFEND}.",
        flavour = "'I didn’t feel a thing!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:CalculateDefendText( self.defend_amt ))
        end,

        cost = 1,

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
        wild = true,

        defend_amt = 3,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition( "DEFEND", self.defend_amt, self )
        end
    },

    toughen_up_plus2a = 
    {
        name = "Easing Toughen Up",
        desc = "<#UPGRADE>{ALLEVIATE} 2,</> apply {1} {DEFEND}.",
        OnPostResolve = function( self, battle, attack )
            attack:AddCondition( "DEFEND", self.defend_amt, self )
            self.owner:AlleviateHealth(2, self)
        end
    },

    toughen_up_plus2b =
    {
        name = "Boosted Toughen Up",
        desc = "Apply <#UPGRADE>{1} {DEFEND}</>.",
        defend_amt = 5,
    },

    toughen_up_plus2c =
    {
        name = "Quenched Toughen Up",
        desc = "Apply {1} {DEFEND}.\n<#UPGRADE>Apply 3 additional {DEFEND} if you drank this turn.</>",

        PreReq = function( self, battle, target )
            return self.drank
        end,
        
        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.DRINK ] = function( self, battle, count, cards )
                self.drank = true
            end,

            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                self.drank = false
            end
        },

        OnPostResolve = function( self, battle, attack )
            local bonus = self.drank and 3 or 0
            attack:AddCondition( "DEFEND", self.defend_amt + bonus, self )
        end
    },

    toughen_up_plus2d =
    {
        name = "Rolling Toughen Up",
        desc = "Apply <#DOWNGRADE>{1} {DEFEND}</>. <#UPGRADE>Increase by {2} for each card played this turn.\n{3}</>",
        desc_fn = function(self, fmt_str)
            if self.engine then
                local str = loc.format( LOC"CARD_ENGINE.CARDS_PLAYED", self.engine:CountCardsPlayed() )
                return loc.format(fmt_str, self:CalculateDefendText( self.defend_amt + (self.defend_bonus * self.engine:CountCardsPlayed()) ), self.defend_bonus, str )
            else
                return loc.format(fmt_str, self:CalculateDefendText( self.defend_amt ), self.defend_bonus, "" )
            end
        end,

        flags = CARD_FLAGS.SKILL,
        defend_amt = 2,
        defend_bonus = 1,
        OnPostResolve = function( self, battle, attack)
            attack:AddCondition( "DEFEND", self.defend_amt + (self.defend_bonus * self.engine:CountCardsPlayed()), self )
        end,
    },

    toughen_up_plus2e =
    {
        name = "Excited Toughen Up",
        desc = "Apply {1} {DEFEND}.\n<#UPGRADE>Gain 2 {ADRENALINE}.</>",
        OnPostResolve = function( self, battle, attack )
            attack:AddCondition( "DEFEND", self.defend_amt, self )
            self.owner:AddCondition("ADRENALINE", 2, self)
        end
    },

    toughen_up_plus2f =
    {
        name = "Toughen Up of Clarity",
        desc = "Apply <#UPGRADE>{1} {DEFEND}</>.",
        defend_amt = 10,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.CONSUME,
    },

    toughen_up_plus2g =
    {
        name = "Lucid Toughen Up",
        desc = "Apply <#UPGRADE>{1} {DEFEND}</>.",
        defend_amt = 7,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
    },

    dropkick =
    {
        name = "Dropkick",
        anim = "dropkick",
        desc = "Take {1} damage.",
        flavour = "'Both feet first!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.self_damage)
        end,

        cost = 1,

        wild = true,

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.MELEE,

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,

        min_damage = 6,
        max_damage = 6,
        self_damage = 3,

        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
        end
    },

    dropkick_plus2a =
    {
        name = "Boosted Dropkick",
        min_damage = 8,
        max_damage = 8,
    },

    dropkick_plus2b =
    {
        name = "Biting Dropkick",
        desc = "Take <#UPGRADE>{1}</> damage <#DOWNGRADE>twice.</>",

        self_damage = 2,
        self_damage_hits = 2,

        OnPostResolve = function( self, battle, attack )
            for i=1, self.self_damage_hits do
                self.owner:SelfDamage( self.self_damage, self )
            end
        end
    },

    dropkick_plus2c =
    {
        name = "Quenched Dropkick",
        desc = "Take {1} damage.\n<#UPGRADE>Deals 4 bonus damage if you drank this turn.</>",

        PreReq = function( self, battle, target )
            return self.drank
        end,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and self.drank then
                    dmgt:AddDamage(4, 4, self)
                end 
            end,

            [ BATTLE_EVENT.DRINK ] = function( self, battle, count, cards )
                self.drank = true
            end,

            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                self.drank = false
            end
        },
    },

    dropkick_plus2d =
    {
        name = "Dropkick of Clarity",
        desc = "Take {1} damage.",
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.CONSUME,
        min_damage = 16,
        max_damage = 16,
    },

    dropkick_plus2e =
    {
        name = "Heavy Dropkick",
        desc = "Take {1} damage.\n<#UPGRADE>Apply 2 {TRAUMA}.</>",

        OnPostResolve = function( self, battle, attack)
            self.owner:SelfDamage( self.self_damage, self )
            attack:AddCondition("TRAUMA", 2, self)
        end,
    },

    dropkick_plus2f =
    {
        name = "Lucid Dropkick",
        desc = "Take {1} damage.",
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
        min_damage = 11,
        max_damage = 11,
    },

    dropkick_plus2g =
    {
        name = "Flying Dropkick",
        desc = "Take {1} damage.\n<#UPGRADE>Gain 1 {EVASION}</>.",
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
        OnPreResolve = function( self, battle, attack )
            self.owner:AddCondition("EVASION", 1, self)
        end,
        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
        end
    },

    smiths_flask = 
    {
        name = "Smith's Flask",
        anim = "drink",
        desc = "{DRINK}.\n{IMPROVISE} a card from a pool of special cards.",
        flavour = "'I stole this from a guy, so now it’s mine.'",

        cost = 1,

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        pool_size = 3,
        pool_cards = {"improvise_green_juice", "improvise_murder_bay_blaster", "improvise_hair_of_the_vroc", "improvise_big_brain", "improvise_stack", "improvise_chug", "improvise_kradeshi_collider", "improvise_shrokian_pepper_shot", "improvise_rumble"},

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, 1)
            local cards = {
                Battle.Card( "improvise_green_juice", self.owner ),
                Battle.Card( "improvise_murder_bay_blaster", self.owner ),
                Battle.Card( "improvise_hair_of_the_vroc", self.owner ),
                Battle.Card( "improvise_big_brain", self.owner ),
                Battle.Card( "improvise_stack", self.owner ),
                Battle.Card( "improvise_kradeshi_collider", self.owner),
                Battle.Card( "improvise_shrokian_pepper_shot", self.owner),
                Battle.Card( "improvise_rumble", self.owner),
            }
            if self.generation == nil or self.generation < 2 then
                table.insert(cards, Battle.Card( self.pool_size == 5 and "improvise_chug_upgraded2" or "improvise_chug", self.owner ))
            end

            cards = table.multipick( cards, self.pool_size )
            local chosen = battle:ImproviseCards( cards, 1, nil, nil, nil, self )
            chosen[1].generation = self.generation or 0
        end,
    },

    smiths_flask_plus =
    {
        name = "Smith's Boosted Flask",
        desc = "{DRINK}.\n{IMPROVISE} a card from a pool of <#UPGRADE>upgraded</> special cards.",
        pool_cards = {"improvise_green_juice_upgraded", "improvise_murder_bay_blaster_upgraded", "improvise_hair_of_the_vroc_upgraded", "improvise_big_brain_upgraded", "improvise_stack_upgraded", "improvise_chug_upgraded", "improvise_kradeshi_collider_upgraded", "improvise_shrokian_pepper_shot_upgraded", "improvise_rumble_upgraded"},
        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, 1)
            local cards = {
                Battle.Card( "improvise_green_juice_upgraded", self.owner ),
                Battle.Card( "improvise_murder_bay_blaster_upgraded", self.owner ),
                Battle.Card( "improvise_hair_of_the_vroc_upgraded", self.owner ),
                Battle.Card( "improvise_big_brain_upgraded", self.owner ),
                Battle.Card( "improvise_stack_upgraded", self.owner ),
                Battle.Card( "improvise_kradeshi_collider_upgraded", self.owner),
                Battle.Card( "improvise_shrokian_pepper_shot_upgraded", self.owner),
                Battle.Card( "improvise_rumble_upgraded", self.owner),
            }
            if self.generation == nil or self.generation < 2 then
                table.insert(cards, Battle.Card( "improvise_chug_upgraded", self.owner ))
            end

            cards = table.multipick( cards, self.pool_size )
            local chosen = battle:ImproviseCards( cards, 1, nil, nil, nil, self )
            chosen[1].generation = self.generation or 0
        end,
    },

    smiths_flask_plus2 = 
    {
        name = "Smith's Wide Flask",
        desc = "{DRINK}.\n<#UPGRADE>{IMPROVISE_PLUS}</> a card from a pool of special cards.",
        pool_size = 5,
    },

    improvise_kradeshi_collider =
    {
        name = "Kra'deshi Collider",
        anim = "drink",
        desc = "{DRINK}.\nTake 2 damage.",
        flavour = "'Pretty sure they use this to start fires.'",

        cost = 0,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        self_damage = 2,


        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,
        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
            Drink(self, battle)
        end
    },

    improvise_kradeshi_collider_upgraded =
    {
        name = "Boosted Kra'deshi Collider",
        anim = "drink",
        desc = "{DRINK}.\nTake 2 damage twice.",

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        self_damage_hits = 2,

        OnPostResolve = function( self, battle, attack )
            for i=1, self.self_damage_hits do
                self.owner:SelfDamage( self.self_damage, self )
            end
            Drink(self, battle)
        end
    },


    improvise_green_juice = 
    {
        name = "Green Juice",
        anim = "drink",
        desc = "{DRINK}.\n{ALLEVIATE} 4 health.",
        flavour = "'So refreshing!'",
        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        max_xp = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AlleviateHealth(4)
            Drink(self, battle)
        end,
    },

    improvise_green_juice_upgraded =
    {
        name = "Boosted Green Juice",
        anim = "drink",
        desc = "{DRINK}.\n{ALLEVIATE} 6 health.",
        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AlleviateHealth(6)
            Drink(self, battle)
        end,
    },

    improvise_murder_bay_blaster =
    {
        name = "Murder Bay Blaster",
        anim = "drink_fancy",
        desc = "{DRINK}.\nGain 1 {POWER}.",
        flavour = "'Don’t swallow the battery, unless you want a real rough couple of days.'",

        cost = 1,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("POWER", 1, self)
            Drink(self, battle)
        end,
    },

    improvise_murder_bay_blaster_upgraded =
    {
        name = "Pale Murder Bay Blaster",
        anim = "drink_fancy",
        desc = "{DRINK}.\nGain 1 {POWER}.",

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("POWER", 1, self)
            Drink(self, battle)
        end,
    },

    improvise_hair_of_the_vroc =
    {
        name = "Hair of the Vroc",
        anim = "drink_fancy",
        desc = "{DRINK}.\nGain {1} {ADRENALINE}.",
        flavour = "'Oh yeah, that’s the stuff.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_amt)
        end,

        cost = 0,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        adrenaline_amt = 2,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
            Drink(self, battle)
        end,
    },

    improvise_hair_of_the_vroc_upgraded =
    {
        name = "Boosted Hair of the Vroc",
        anim = "drink_fancy",
        desc = "{DRINK}.\nGain {1} {ADRENALINE}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_amt)
        end,

        cost = 0,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        adrenaline_amt = 3,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
            Drink(self, battle)
        end,    
    },

    improvise_big_brain = 
    {
        name = "Big Brain",
        anim = "drink",
        desc = "{DRINK}.\nDraw 2 cards.",
        flavour = "'A burst of creativity, but it’s mainly from the garnish.'",

        cost = 0,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            battle:DrawCards(2)
            Drink(self, battle)
        end,
    },

    improvise_big_brain_upgraded = 
    {
        name = "Visionary Big Brain",
        anim = "drink",
        desc = "{DRINK}.\nDraw 3 cards.",

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            battle:DrawCards(3)
            Drink(self, battle)
        end,
    },

    -- improvise_last_drop =
    -- {
    --     name = "Last Drop",
    --     anim = "bottle_flourish",
    --     desc = "Draw an {empty_bottle} card from anywhere in your deck.",

    --     cost = 0,
    --     max_xp = 0,

    --     rarity = CARD_RARITY.UNIQUE,
    --     flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
    --     target_type = TARGET_TYPE.SELF,

    --     OnPostResolve = function( self, battle, attack )
    --         local found_bottle = false
    --         for i,card in battle:GetDrawDeck():Cards() do
    --             if card.id == "empty_bottle" then
    --                 card:TransferCard(battle:GetHandDeck())
    --                 found_bottle = true
    --                 break
    --             end
    --         end
    --         if not found_bottle then
    --             for i,card in battle:GetDiscardDeck():Cards() do
    --                 if card.id == "empty_bottle" then
    --                     card:TransferCard(battle:GetHandDeck())
    --                     found_bottle = true
    --                     break
    --                 end
    --             end
    --         end
    --     end,
    -- },

    -- improvise_last_drop_upgraded =
    -- {
    --     name = "Boosted Last Drop",
    --     anim = "bottle_flourish",
    --     desc = "Draw up to 2 {empty_bottle} cards from anywhere in your deck.",

    --     cost = 0,

    --     rarity = CARD_RARITY.UNIQUE,
    --     flags = CARD_FLAGS.EXPEND | CARD_FLAGS.SKILL,
    --     target_type = TARGET_TYPE.SELF,

    --     OnPostResolve = function( self, battle, attack )
    --         local bottles_found = 0
    --         for i,card in battle:GetDrawDeck():Cards() do
    --             if card.id == "empty_bottle" then
    --                 card:TransferCard(battle:GetHandDeck())
    --                 bottles_found = bottles_found + 1
    --                 if bottles_found >= 2 then
    --                     break
    --                 end
    --             end
    --         end
    --         if bottles_found < 2 then
    --             for i,card in battle:GetDiscardDeck():Cards() do
    --                 if card.id == "empty_bottle" then
    --                     card:TransferCard(battle:GetHandDeck())
    --                     bottles_found =  bottles_found + 1
    --                     if bottles_found >= 2 then
    --                         break
    --                     end
    --                 end
    --             end
    --         end
    --     end,
    -- },

    improvise_stack =
    {
        name = "Stack",
        anim = "bottle_throw",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck: Deal {2} damage for each bottle.",
        flavour = "'Hold these.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_count, self.damage_amt)
        end,

        cost = 0,
        max_xp = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        min_damage = 0,
        max_damage = 0,

        bottle_count = 2,
        damage_amt = 3,

        OnPostResolve = function( self, battle, targets )
            SpendBottles(self, self.bottle_count)
        end,

        event_handlers = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    local count = math.min(self.bottle_count, GetBottleCount(self))
                    count = count * self.damage_amt
                    dmgt:AddDamage( count, count, self)
                end
            end
        },
    },

    improvise_stack_upgraded =
    {
        name = "Boosted Stack",
        anim = "bottle_throw",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck: Deal {2} damage for each bottle.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_count, self.damage_amt)
        end,

        cost = 0,

        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        min_damage = 0,
        max_damage = 0,

        bottle_count = 2,
        damage_amt = 4,

        OnPostResolve = function( self, battle, targets )
            SpendBottles(self, self.bottle_count)
        end,

        event_handlers = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    local count = math.min(self.bottle_count, GetBottleCount(self))
                    count = count * self.damage_amt
                    dmgt:AddDamage( count, count, self)
                end
            end
        },
    },

    improvise_chug = 
    {
        name = "Chug",
        anim = "drink",
        desc = "Play {smiths_flask} again.",
        flavour = "'Sir, that’s one per table not customer.'",

        cost = 0,
        max_xp = 0,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local card = Battle.Card("smiths_flask", self.owner)
            card.show_dealt = false
            card.generation = (self.generation or 0) + 1
            card:SetFlags(CARD_FLAGS.EXPEND)
            battle:DealCard(card, battle:GetHandDeck())
            battle:PlayCard(card)
        end,
    },

    improvise_chug_upgraded = 
    {
        name = "Boosted Chug",
        anim = "drink",
        desc = "Play {smiths_flask_plus} again.",

        cost = 0,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local card = Battle.Card("smiths_flask_plus", self.owner)
            card.show_dealt = false
            card.generation = (self.generation or 0) + 1
            card:SetFlags(CARD_FLAGS.EXPEND)
            battle:DealCard(card, battle:GetHandDeck())
            battle:PlayCard(card)
        end,
    },

    improvise_chug_upgraded2 = 
    {
        name = "Wide Chug",
        anim = "drink",
        desc = "Play {smiths_flask_plus2} again.",

        cost = 0,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local card = Battle.Card("smiths_flask_plus2", self.owner)
            card.show_dealt = false
            card.generation = (self.generation or 0) + 1
            card:SetFlags(CARD_FLAGS.EXPEND)
            battle:DealCard(card, battle:GetHandDeck())
            battle:PlayCard(card)
        end,
    },

    improvise_shrokian_pepper_shot =
    {
        name = "Shrokian Pepper Shot",
        anim = "bottle_throw",
        desc = "{EXPEND} 1 {empty_bottle} card from your deck: Apply 1 {IMPAIR}.",
        flavour = "'You spit it right back in the eyes, pretty effective if you ask me.'",

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        max_xp = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, 1) > 0 then
                attack:AddCondition("IMPAIR", 1, self)
            end
        end
    },

    improvise_shrokian_pepper_shot_upgraded =
    {
        name = "Boosted Shrokian Pepper Shot",
        anim = "bottle_throw",
        desc = "{EXPEND} 1 {empty_bottle} card from your deck: Apply 2 {IMPAIR}.",

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        max_xp = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, 1) > 0 then
                attack:AddCondition("IMPAIR", 2, self)
            end
        end
    },

    improvise_rumble =
    {
        name = "Rumble",
        anim = "bottle_throw",
        desc = "{EXPEND} 1 {empty_bottle} card from your deck: Apply 2 {TRAUMA}.",
        flavour = "'Let’s get this party started!'",

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        max_xp = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, 1) > 0 then
                attack:AddCondition("TRAUMA", 2, self)
            end
        end
    },

    improvise_rumble_upgraded =
    {
        name = "Boosted Rumble",
        anim = "bottle_throw",
        desc = "{EXPEND} 1 {empty_bottle} card from your deck: Apply 3 {TRAUMA}.",

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        max_xp = 0,
        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, 1) > 0 then
                attack:AddCondition("TRAUMA", 3, self)
            end
        end
    },

    -- kradeshi_strike =
    -- {
    --     name = "Kra'deshi Strike",
    --     anim = "punch",
    --     desc = "{FOCUSED}: Heal health equal to damage dealt by this card.",

    --     cost = 1,

    --     rarity = CARD_RARITY.COMMON,
    --     flags = CARD_FLAGS.MELEE,

    --     min_damage = 3,
    --     max_damage = 6,

    --     PreReq = function( self, battle )
    --         return battle:GetHandDeck():CountCards() <= 1
    --     end,

    --     event_handlers = 
    --     {
    --         [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
    --             if source and source.card == self and self.engine:GetHandDeck():CountCards() <= 1 and delta > 0 then
    --                 self.owner:HealHealth(damage)
    --             end
    --         end,
    --     },
    -- },

    -- backhand = 
    -- {
    --     name = "Backhand",
    --     anim = "punch",
    --     desc = "Draw a card.\n{FOCUSED}: this card costs 0.",

    --     cost = 1,

    --     rarity = CARD_RARITY.COMMON,
    --     flags = CARD_FLAGS.MELEE,

    --     min_damage = 5,
    --     max_damage = 5,

    --     PreReq = function( self, battle )
    --         return battle:GetHandDeck():CountCards() <= 1
    --     end,

    --     event_priorities =
    --     {
    --         [ BATTLE_EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_SETTOR,
    --     },

    --     event_handlers = 
    --     {
    --         [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
    --             if card == self and self.engine:GetHandDeck():CountCards() <= 1 then
    --                 cost_acc:ModifyValue(0)
    --             end
    --         end,
    --     },

    --     OnPostResolve = function( self, battle, attack )
    --         battle:DrawCards(1)
    --     end
    -- },

    -- swift_kick = 
    -- {
    --     name = "Swift Kick",
    --     anim = "punch",
    --     desc = "{OPENER}: This card deals {1} bonus damage.",
    --     desc_fn = function( self, fmt_str )
    --         return loc.format(fmt_str, self.bonus_damage)
    --     end,

    --     cost = 1,

    --     rarity = CARD_RARITY.COMMON,
    --     flags = CARD_FLAGS.MELEE,

    --     min_damage = 4,
    --     max_damage = 6,
    --     bonus_damage = 4,

    --     PreReq = function( self, battle )
    --         return battle:CountCardsPlayed() == 0
    --     end,

    --     event_handlers = 
    --     {
    --         [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
    --             if card == self and self.engine:CountCardsPlayed() == 0 then
    --                 dmgt:AddDamage( self.bonus_damage, self.bonus_damage, self )
    --             end
    --         end,
    --     },
    -- },

    freebie = 
    {
        name = "Freebie",
        anim = "gut_shot",
        desc = "{EXPEND} 2 {empty_bottle} cards from your deck: This card costs 0.",
        flavour = "'Too quick for you huh?'",

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.MELEE,

        min_damage = 4,
        max_damage = 4,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= 2
        end,

        OnPostResolve = function( self, battle, attack )
            if GetBottleCount(self) >= 2 then
                SpendBottles(self, 2)
            end
        end,

        event_priorities =
        {
            [ BATTLE_EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_SETTOR,
        },

        event_handlers = 
        {
            [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                if card == self then
                    if GetBottleCount(self) >= 2 then
                        cost_acc:ModifyValue(0, self)
                    end
                end
            end,
        },
    },

    freebie_plus =
    {
        name = "Tall Freebie",
        max_damage = 8,
    },

    freebie_plus2 =
    {
        name = "Rooted Freebie",
        min_damage = 6,
        max_damage = 6,
    },

    forced_adrenaline =
    {
        name = "Push It",
        anim = "taunt",
        desc = "Take {1} damage. Draw {2} cards.",
        flavour = "'Now wait a minute y'all.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.self_damage, self.draw_count)
        end,

        cost = 0,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        draw_count = 2,
        self_damage = 5,

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,

        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
            battle:DrawCards(self.draw_count)
        end
    },

    forced_adrenaline_plus = 
    {
        name = "Visionary Push It",
        desc = "Take {1} damage. Draw <#UPGRADE>{2}</> cards.",
        draw_count = 3,
    },

    forced_adrenaline_plus2 =
    {
        name = "Split Push It",
        desc = "Take <#UPGRADE>{1}</> damage <#DOWNGRADE>twice</>. Draw {2} cards.",

        self_damage = 3,
        self_damage_hits = 2,

        OnPostResolve = function( self, battle, attack )
            for i=1, self.self_damage_hits do
                self.owner:SelfDamage( self.self_damage, self )
            end
            battle:DrawCards(self.draw_count)
        end
    },

    smash =
    {
        name = "Smash",
        anim = "super_smash",
        flavour = "'This one gets a little technical.'",
        desc = "Hits all enemies.\nTake {1} damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.self_damage)
        end,

        cost = 2,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.MELEE,

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,

        min_damage = 6,
        max_damage = 8,
        self_damage = 4,
        target_mod = TARGET_MOD.TEAM,

        OnPostResolve = function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
        end
    },

    smash_plus =
    {
        name = "Softened Smash",
        desc = "Hits all enemies.\nTake <#UPGRADE>{1}</> damage.",
        self_damage = 2,
    },

    smash_plus2 =
    {
        name = "Boosted Smash",
        min_damage = 8,
        max_damage = 10,
    },

    raw_power =
    {
        name = "Raw Power",
        anim = "taunt",
        desc = "Reduce the cost of all cards in your hand to 0 until the end of your turn or until played. Take {1} damage per card.",
        flavour = "'I feel great!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage)
        end,

        cost = 1,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        damage = 3,

        CanPlayCard = function( self, battle, target )
            local damage = 0
            for i,card in battle:GetHandDeck():Cards() do
                if card ~= self and battle:CalculateActionCost(card) > 0 then
                    damage = damage + self.damage
                end
            end
            return CanPlaySelfDamage( self, damage )
        end,

        OnPostResolve = function( self, battle, attack )
            local con = self.owner:AddCondition("raw_power", 1, self)
            if con then
                con.chosen_cards = con.chosen_cards or {}
                local damage = 0
                for i,card in battle:GetHandDeck():Cards() do
                    if card ~= self and battle:CalculateActionCost(card) > 0 then
                        damage = damage + self.damage
                        table.insert(con.chosen_cards, card)
                    end
                end
                self.owner:SelfDamage( damage, self )
            end
        end,

        condition = 
        {
            hidden = true,
            event_priorities =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST  ] = EVENT_PRIORITY_SETTOR,
            },
            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.chosen_cards and table.contains(self.chosen_cards, card) then
                        cost_acc:ModifyValue( 0, self )
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if table.contains(self.chosen_cards, card) then
                        table.arrayremoveall(self.chosen_cards, card)
                    end
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    raw_power_plus =
    {
        name = "Pale Raw Power",
        cost = 0,
    },

    raw_power_plus2 =
    {
        name = "Enduring Raw Power",
        flags = CARD_FLAGS.SKILL,
    },

    masochist =
    {
        name = "Masochist",
        desc = "{ABILITY}: Whenever you take damage on your turn, gain 1 {ADRENALINE}.",
        flavour = "'This is kind of your thing isn’t it?'",
        on_play_screen_shake = true,

        cost = 2,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        count = 1,

        OnPostResolve = function ( self, battle, attack )
            self.owner:AddCondition("masochist", self.count, self)
        end,

        
        condition = 
        {
            desc = "Whenever you take damage on your turn, gain <#HILITE>{1} {ADRENALINE}</>.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,

            OnApply = function( self, battle )
                self.active = true
            end,
            event_priorities =
            {
                -- before: chafe
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN  ] = 10,
            },


            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.active = true
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.active = false
                end,

                [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                    if self.active and fighter == self.owner then
                        self.owner:AddCondition("ADRENALINE", self.stacks, self)
                    end
                end,
            },
        },
    },

    masochist_plus =
    {
        name = "Pale Masochist",
        cost = 1,
    },

    masochist_plus2 =
    {
        name = "Boosted Masochist",
        desc = "{ABILITY}: Whenever you take damage on your turn, gain <#UPGRADE>2 </>{ADRENALINE}.",
        count = 2,
    },

    snellecks_finest =
    {
        name = "Snellick's Finest",
        anim = "drink_fancy",
        desc = "{DRINK}.\nGain 5 {smith_regeneration}.",
        flavour = "'Hesh, even Snellick’s worst isn’t half bad.'",

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        regen_amt = 5,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle)
            local healing = self.owner:GetCondition("smith_regeneration")
            if healing then
                healing:IncreaseMax(self.regen_amt, self)
            end
        end
    },

    snellecks_finest_plus =
    {
        name = "Enduring Snellick's Finest",
        desc = "{DRINK}.\nGain <#DOWNGRADE>3 {smith_regeneration}</>.",
        flags = CARD_FLAGS.SKILL,
        regen_amt = 3,
    },

    snellecks_finest_plus2 =
    {
        name = "Boosted Snellick's Finest",
        desc = "{DRINK}.\nGain <#UPGRADE>8 {smith_regeneration}</>.",
        regen_amt = 8,
    },

    the_pinto_pour =
    {
        name = "The Pinto Pour",
        anim = "drink",
        desc = "{DRINK}.\nGain 2 {POWER}.",
        flavour = "'Only the keenest eyes appreciate the absolute craftsmanship of the bottle.'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        power_amt = 2,
        drink_count = 1,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("POWER", self.power_amt, self)
            Drink(self, battle, self.drink_count)
        end
    },

    the_pinto_pour_plus =
    {
        name = "The Fortified Pinto Pour",
        desc = "{DRINK}.\nGain <#UPGRADE>3</> {POWER}.",
        power_amt = 3,
    },

    the_pinto_pour_plus2 =
    {
        name = "The Pale Pinto Pour",
        cost = 0,
    },

    fire_breather = 
    {
        name = "Fire Breather",
        anim = "drink",
        desc = "{DRINK} and gain {1} {ADRENALINE}.",
        flavour = "'Fermented by the more cruel Shrokian brewmasters. You’re going to want a chaser.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_amt)
        end,

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.SKILL,
        target_type = TARGET_TYPE.SELF,

        drink_count = 1,
        adrenaline_amt = 2,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_count)
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
        end,
    },

    fire_breather_plus =
    {
        name = "Gulping Fire Breather",
        desc = "{DRINK} <#UPGRADE>2</> and gain {1} {ADRENALINE}.",
        drink_count = 2,
    },

    fire_breather_plus2 =
    {
        name = "Boosted Fire Breather",
        desc = "{DRINK} and gain <#UPGRADE>{1}</> {ADRENALINE}.",
        adrenaline_amt = 3,
    },

    mouthwash = 
    {
        name = "Mouthwash",
        anim = "drink",
        desc = "{DRINK}.\nRemove a random debuff. Draw a card.",
        flavour = "'Now with flavor!'",

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        draw_count = 1,
        drink_count = 1,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_count)
            self.owner:RemoveDebuff( self ) 
            battle:DrawCards(self.draw_count)
        end
    },

    mouthwash_plus =
    {
        name = "Visionary Mouthwash",
        desc = "{DRINK}.\nRemove a random debuff. Draw <#UPGRADE>2 cards</>.",
        draw_count = 2,
    },

    mouthwash_plus2 =
    {
        name = "Boosted Mouthwash",
        desc = "{DRINK} <#UPGRADE>2</>.\nRemove a random debuff. Draw a card.",
        drink_count = 2,
    },

    scream = 
    {
        name = "Scream",
        anim = "taunt",
        desc = "Gain 3 {ADRENALINE}. Take 3 damage every time you play a card until the end of this turn.",
        flavour = "'Hesh! That’s Heshing loud!'",

        cost = 0,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,

        target_type = TARGET_TYPE.SELF,

        self_damage = 3,
        adrenaline_amt = 3,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
            local con = self.owner:AddCondition("scream", self.self_damage, self)
            con.source_card = self
        end,

        condition = 
        {
            ctype = CTYPE.DEBUFF,
            desc = "Take <#HILITE>{1}</> damage every time you play a card until the end of this turn.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,
            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack and attack.card ~= self.source_card and attack.attacker == self.owner then
                        self.owner:SelfDamage( self.stacks, self )
                    end
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    scream_plus =
    {
        name = "Softened Scream",
        desc = "Gain 3 {ADRENALINE}. Take <#UPGRADE>2</> damage every time you play a card until the end of this turn.",
        self_damage = 2,
    },

    scream_plus2 =
    {
        name = "Boosted Scream",
        desc = "Gain <#UPGRADE>4</> {ADRENALINE}. Take 3 damage every time you play a card until the end of this turn.",
        adrenaline_amt = 4,
    },

    clean_house =
    {
        name = "Clean House",
        anim = "taunt",
        desc = "Discard up to 3 cards and draw that many. Take {1} damage per card discarded.",
        flavour = "'Oof, that’s been here a while.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.discard_dmg)
        end,

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,

        target_type = TARGET_TYPE.SELF,
        discard_dmg = 2,

        OnPostResolve = function( self, battle, attack )
            local cards = battle:DiscardCards(0, 3, self)
            if #cards > 0 then
                for i=1,#cards do
                    self.owner:SelfDamage( self.discard_dmg, self )
                end
                battle:DrawCards(#cards)
            end
        end
    },

    clean_house_plus =
    {
        name = "Pale Clean House",
        cost = 0,
    },

    clean_house_plus2 = 
    {
        name = "Wide Clean House",
        desc = "Discard <#UPGRADE>any number of</> cards and draw as many. Take {1} damage per card discarded.",
        OnPostResolve = function( self, battle, attack )
            local cards = battle:DiscardCards(nil, nil, self)
            if #cards > 0 then
                for i=1,#cards do
                    self.owner:SelfDamage( self.discard_dmg, self )
                end
                battle:DrawCards(#cards)
            end
        end
    },

    inspirational_blow = 
    {
        name = "Inspirational Blow",
        desc = "{THRESHOLD} {1}: Draw {2} cards.",
        flavour = "'Ok I have to give it to him, that was well done.'",
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self:CalculateThresholdText(self), self.draw_count)
        end,
        anim = "back_hamm",

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON, 

        cost = 1,
        min_damage = 3,
        max_damage = 5,

        threshold = 5,
        draw_count = 2,

        ThresholdEffect = function( self, battle, attack )
           battle:DrawCards(self.draw_count)
        end,
    },

    inspirational_blow_plus = 
    {
        name = "Boosted Inspirational Blow",
        min_damage = 4,
        max_damage = 6,
    },

    inspirational_blow_plus2 =
    {
        name = "Visionary Blow",
        desc = "{THRESHOLD} {1}: Draw <#UPGRADE>{2}</> cards.",
        draw_count = 3,
    },

    nerve_strike =
    {
        name = "Nerve Strike",
        desc = "{THRESHOLD} {1}: Apply {IMPAIR {2}}.",
        flavour = "'Ahh! Right on the funny bone!'",
        desc_fn = function(self, fmt_str)
            return loc.format(fmt_str, self:CalculateThresholdText(self), self.impair_amount)
        end,
        anim = "punch",

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON, 

        cost = 1,
        min_damage = 2,
        max_damage = 6,

        threshold = 6,
        impair_amount = 2,

        ThresholdEffect = function( self, battle, attack )
           attack:AddCondition("IMPAIR", self.impair_amount)
        end,
    },

    nerve_strike_plus =
    {
        name = "Boosted Nerve Strike",
        min_damage = 3,
        max_damage = 7,
    },

    nerve_strike_plus2 =
    {
        name = "Focused Nerve Strike",
        desc = "{THRESHOLD} {1}: Apply <#UPGRADE>{IMPAIR {2}}</>.",
        impair_amount = 4,
    },

    -- quickened = 
    -- {
    --     name = "Quickened",
    --     desc = "Choose a card from your hand, then discard the rest of your hand. Gain {1} actions next turn.",
    --     desc_fn = function(self, fmt_str)
    --         return loc.format(fmt_str, self.action_count)
    --     end,
    --     anim = "taunt",

    --     cost = 2,
    --     action_count = 3,

    --     rarity = CARD_RARITY.RARE,
    --     flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND, 

    --     target_type = TARGET_TYPE.SELF,

    --     OnPostResolve = function ( self, battle, attack )
    --         local keep_card = battle:ChooseCards( 1, 1, nil, "Choose card to keep.")

    --         local tbl = ObtainWorkTable()
    --         for i, card in battle:GetHandDeck():Cards() do
    --             if card ~= keep_card[1] then
    --                 table.insert(tbl, card)
    --             end
    --         end

    --         for i, card in ipairs(tbl) do
    --             battle:DiscardCard(card)
    --         end

    --         ReleaseWorkTable(tbl)

    --         local con = self.owner:AddCondition("quickened")
    --         con.action_count = self.action_count
    --     end,

    --     condition = 
    --     {
    --         hidden = true,

    --         event_handlers = 
    --         {
    --             [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
    --                 battle:ModifyActionCount( self.action_count )
    --                 self.owner:RemoveCondition(self.id)
    --             end
    --         },
    --     },
    -- },

    -- quickened_plus =
    -- {
    --     name = "Pale Quickened",
    --     cost = 1,
    -- },

    -- quickened_plus2 =
    -- {
    --     name = "Boosted Quicken",
    --     desc = "Choose a card from your hand, then discard the rest of your hand. Gain <#UPGRADE>{1}</> actions next turn.",
    --     action_count = 4,
    -- },

    down_the_hatch =
    {
        name = "Down the Hatch",
        anim = "drink2",
        desc = "{DRINK} 2, draw {1} {1*card|cards}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.draw_count )
        end,
        flavour = "'Chug, chug, chug!'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,
        drink_count = 2,
        draw_count = 1,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_count)
            battle:DrawCards(self.draw_count)
        end
    },

    down_the_hatch_plus =
    {
        name = "Boosted Down the Hatch",
        desc = "<#UPGRADE>{DRINK} 4</>, draw {1} {1*card|cards}.",
        drink_count = 4,
    },

    down_the_hatch_plus2 =
    {
        name = "Visionary Down the Hatch",
        desc = "{DRINK} 2, draw <#UPGRADE>{1} {1*card|cards}</>.",
        draw_count = 2,
    },

    cheers =
    {
        name = "Cheers",
        anim = "cheers_smash",
        flavour = "'To absent friends!'",
        desc = "{DRINK}.",

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        min_damage = 3,
        max_damage = 6,

        drink_count = 1,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_count)
        end
    },

    cheers_plus =
    {
        name = "Boosted Cheers",
        desc = "{DRINK} <#UPGRADE>2</>.",
        drink_count = 2,
    },

    cheers_plus2 =
    {
        name = "Rooted Cheers",
        min_damage = 6,
    },

    swig =
    {
        name = "Swig",
        anim = "drink2",
        desc = "{DRINK} {1}. Gain one of the following: {2} {smith_regeneration}, {ADRENALINE {3}}, {DEFEND {4}} or {POWER {5}}.",
        flavour = "’This is great! What do you call it?’",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.drink_amount, self.moxie_amt, self.adrenaline_amt, self:CalculateDefendText(self.defend_amt), self.power_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,

        drink_amount = 2,
        moxie_amt = 2,
        adrenaline_amt = 2,
        defend_amt = 4,
        power_amt = 1,

        options = {
            ["MOXIE"] = 2,
            ["ADRENALINE"] = 2,
            ["DEFEND"] = 4,
            ["POWER"] = 1
        },

        count = 1,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_amount)
            for i=1,self.count do
                local con, stacks = table.tablepick(self.options)
                if con == "MOXIE" then
                    local healing = self.owner:GetCondition("smith_regeneration")
                    if healing then healing:IncreaseMax(stacks, self) end
                else
                    self.owner:AddCondition(con, stacks, self)
                end
            end
        end
    },

    swig_plus =
    {
        name = "Mirrored Swig",
        desc = "{DRINK} {1}. Gain one of the following: {2} {smith_regeneration}, {ADRENALINE {3}}, {DEFEND {4}} or {POWER {5}}, <#UPGRADE>twice</>.",
        count = 2,
    },

    swig_plus2 = 
    {
        name = "Boosted Swig",
        desc = "{DRINK} {1}. Gain one of the following: <#UPGRADE>{2} {smith_regeneration}, {ADRENALINE {3}}, {DEFEND {4}} or {POWER {5}}</>.",
        moxie_amt = 4,
        adrenaline_amt = 4,
        defend_amt = 8,
        power_amt = 2,
        options = {
            ["MOXIE"] = 4,
            ["ADRENALINE"] = 4,
            ["DEFEND"] = 8,
            ["POWER"] = 2,
        },
    },

    dumpster_dive =
    {
        name = "Dumpster Dive",
        anim = "bottle_flourish2",
        desc = "Move up to {1} {empty_bottle} cards from your draw pile to your discard pile.\nDraw {2} {2*card|cards}.",
        flavour = "'Now what do we have here?'",
        desc_fn = function ( self, fmt_str )
            return loc.format(fmt_str, self.bottle_count, self.draw_count)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        bottle_count = 3,
        draw_count = 2,

        OnPostResolve = function( self, battle, attack )
            local cards_to_move = {}
            for i,card in ipairs(battle:GetDrawDeck().cards) do
                if card.id == "empty_bottle" then
                    table.insert(cards_to_move, card)
                    if #cards_to_move >= self.bottle_count then
                        break
                    end
                end
            end
            for i,card in ipairs(cards_to_move) do
                card:TransferCard(battle:GetDiscardDeck())
            end
            battle:DrawCards(self.draw_count)
        end,
    },

    dumpster_dive_plus =
    {
        name = "Visionary Dumpster Dive",
        desc = "Move up to {1} {empty_bottle} cards from your draw pile to your discard pile. Draw <#UPGRADE>{2}</> {2*card|cards}.",
        draw_count = 3,
    },

    dumpster_dive_plus2 =
    {
        name = "Boosted Dumpster Dive",
        desc = "Move up to <#UPGRADE>{1}</> {empty_bottle} cards from your draw pile to your discard pile. Draw {2} {2*card|cards}.",
        bottle_count = 4,
    },

    reduce =
    {
        name = "Reduce",
        anim = "bottle_flourish",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck, gain {2} {DEFEND} per bottle.",
        flavour = "'Want less, waste less!'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cap, self:CalculateDefendText( self.defend_amt ))
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        bottle_cap = 2,
        defend_amt = 3,

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) >= 1, CARD_PLAY_REASONS.NO_BOTTLES
        end,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= 1
        end,

        OnPostResolve = function( self, battle, targets )
            local spent = SpendBottles(self, self.bottle_cap)
            if spent > 0 then
                for i=1, spent do
                    self.owner:AddCondition("DEFEND", self.defend_amt, self)
                end
            end
        end,
    },

    reduce_plus =
    {
        name = "Stone Reduce",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck, gain <#UPGRADE>{2}</> {DEFEND} per bottle.",
        defend_amt = 4,
    },

    reduce_plus2 =
    {
        name = "Wide Reduce",
        desc = "{EXPEND} up to <#UPGRADE>{1}</> {empty_bottle} cards from your deck, gain {2} {DEFEND} per bottle.",
        bottle_cap = 4,
    },

    reuse = 
    {
        name = "Reuse",
        anim = "bottle_flourish2",
        desc = "Discard {empty_bottle} cards from your hand and draw that many cards.",
        flavour = "'Still good!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_count)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        bonus = 0,

        OnPostResolve = function( self, battle, attack )
            local cards_to_move = {}
            for i,card in ipairs(battle:GetHandDeck().cards) do
                if card.id == "empty_bottle" then
                    table.insert(cards_to_move, card)
                end
            end
            if #cards_to_move > 0 then
                for i,card in ipairs(cards_to_move) do
                    battle:DiscardCard(card)
                end
                battle:DrawCards(#cards_to_move + self.bonus)
            end
        end
    },

    reuse_plus =
    {
        name = "Pale Reuse",
        cost = 0,
    },

    reuse_plus2 = 
    {
        name = "Boosted Reuse",
        desc = "Discard {empty_bottle} cards from your hand and draw that many cards <#UPGRADE>+ 1</>.",
        bonus = 1,
    },

    recycle = 
    {
        name = "Recycle",
        anim = "bottle_flourish",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck, {HEAL} {2} health per bottle.",
        flavour = "'It’s all about shrinking that carbon footprint.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cap, self.heal_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        heal_amt = 2,
        bottle_cap = 2,

        OnPostResolve = function( self, battle, attack )
            local spent = SpendBottles(self, self.bottle_cap)
            if spent > 0 then
                self.owner:HealHealth(self.heal_amt * spent, self)
            end
        end
    },

    recycle_plus =
    {
        name = "Boosted Recycle",
        desc = "{EXPEND} up to <#UPGRADE>{1}</> {empty_bottle} cards from your deck, {HEAL} {2} health per bottle.",
        bottle_cap = 4,
    },

    recycle_plus2 =
    {
        name = "Healing Recycle",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck, {HEAL} <#UPGRADE>{2}</> health per bottle.",
        heal_amt = 3,
    },

    gulper = 
    {
        name = "Gulper",
        anim = "drink2",
        desc = "Draw {1} {1*card|cards}. {EXPEND} all {empty_bottle} cards in your hand. {DRINK} 2 per card expended.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.draw_count )
        end,
        flavour = "'Whoah, slow down there.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        bonus_bottles = 0,
        draw_count = 2,

        OnPostResolve = function( self, battle, attack)
            battle:DrawCards(self.draw_count)
            local bottles = {}
            for i,card in ipairs(battle:GetHandDeck().cards) do
                if card.id == "empty_bottle" then
                    table.insert(bottles, card)
                end
            end
            if #bottles + self.bonus_bottles > 0 then
                Drink(self, battle, (#bottles * 2) + self.bonus_bottles)
                for i, card in ipairs(bottles) do
                    battle:ExpendCard(card)
                end
            end
        end
    },

    gulper_plus = 
    {
        name = "Visionary Gulper",      
        desc = "Draw <#UPGRADE>{1} {1*card|cards}</>. {EXPEND} all {empty_bottle} cards in your hand. {DRINK} 2 per card expended.",
        draw_count = 3,
    },

    gulper_plus2 =
    {
        name = "Boosted Gulper",
        desc = "Draw {1} {1*card|cards}. {EXPEND} all {empty_bottle} cards in your hand. {DRINK} 2 per card expended <#UPGRADE>+ 2</>.",
        bonus_bottles = 2,
    },

    right_in_the_face =
    {
        name = "Right in the Face",
        anim = "bottle_throw_quick",
        hit_anim = true,
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck. Attack a random enemy per card expended.",
        flavour = "'Wait, what are you doing with tha-'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cap)
        end,

        cost = 1,

        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.UNCOMMON,

        target_mod = TARGET_MOD.RANDOM1,

        min_damage = 4,
        max_damage = 4,

        bottle_cap = 3,

        CanPlayCard = function( self, battle, target )
            if self:UpdateTargetCount() == 0 then
                return false, CARD_PLAY_REASONS.NO_BOTTLES
            end
            return true
        end,

        PreReq = function( self, battle, target )
            return self:UpdateTargetCount() > 0
        end,

        UpdateTargetCount = function( self )
            self.target_count = math.min(GetBottleCount(self), self.bottle_cap)
            return self.target_count
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self, self.target_count)
            self.target_count = 0
        end,
    },

    right_in_the_face_plus =
    {
        name = "Boosted Right in the Face",
        min_damage = 6,
        max_damage = 6,
    },

    right_in_the_face_plus2 =
    {
        name = "Wide Right in the Face",
        desc = "{EXPEND} up to <#UPGRADE>{1}</> {empty_bottle} cards from your deck. Attack a random enemy per card expended.",
        bottle_cap = 4,
    },

    batter_up =
    {
        name = "Batter Up",
        anim = "bottle_throw_quick",
        hit_anim = true,
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck. Apply {2} {IMPAIR} and {3} {BLEED} to a random enemy per {empty_bottle} expended.",
        flavour = "'Things are about to get messy.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cap, self.impair_amt, self.bleed_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.UNCOMMON,

        target_mod = TARGET_MOD.RANDOM1,

        impair_amt = 1,
        bleed_amt = 2,
        bottle_cap = 3,
        min_damage = 1,
        max_damage = 2,

        CanPlayCard = function( self, battle, target )
            self.target_count = math.min(GetBottleCount(self), self.bottle_cap)
            if self.target_count == 0 then
                return false, CARD_PLAY_REASONS.NO_BOTTLES
            end
            return true
        end,

        PreReq = function( self, battle, target )
            return (self.target_count or 0) > 0
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self, self.target_count)
            self.target_count = 0
            for i,hit in ipairs(attack.hits) do
                if not hit.evaded then
                    hit.target:AddCondition("IMPAIR", self.impair_amt, self)
                    hit.target:AddCondition("BLEED", self.bleed_amt, self)
                end
            end
        end,

    },

    batter_up_plus =
    {
        name = "Boosted Batter Up",
        desc = "{EXPEND} up to <#UPGRADE>{1}</> {empty_bottle} cards from your deck. Apply {2} {IMPAIR} and {3} {BLEED} to a random enemy per {empty_bottle} expended.",
        bottle_cap = 4,
    },

    batter_up_plus2 = 
    {
        name = "Stifling Batter Up",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck. Apply <#UPGRADE>{2}</> {IMPAIR} and {3} {BLEED} to a random enemy per {empty_bottle} expended.",
        impair_amt = 2,
    },

    bonkers = 
    {
        name = "Bonkers",
        anim = "back_hamm",
        flavour = "'WHAAA-HA-HA-HAAAAA!'",
        hit_anim = true,
        desc = "{EXPEND} {1} {empty_bottle} {1*card|cards} from your deck: attack twice.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cost)
        end,

        cost = 2,

        UpdateTargetCount = function( self )
            if GetBottleCount(self) >= self.bottle_cost then
                self.target_count = 2
            else
                self.target_count = 1
            end
            return self.target_count
        end,

        PreReq = function( self, battle, target )
            return self:UpdateTargetCount() >= 2
        end,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        bottle_cost = 2,

        min_damage = 7,
        max_damage = 9,

        OnPostResolve = function( self, battle, attack )
            self:UpdateTargetCount()
            if GetBottleCount(self) >= self.bottle_cost then
                SpendBottles(self, self.bottle_cost)
            end
        end,
    },

    bonkers_plus =
    {
        name = "Broken Bonkers",
        desc = "{EXPEND} <#UPGRADE>{1}</> {empty_bottle} {1*card|cards} from your deck: attack twice.",
        bottle_cost = 1,
    },

    bonkers_plus2 =
    {
        name = "Boosted Bonkers",
        min_damage = 10,
        max_damage = 10,
    },

    long_night =
    {
        name = "Long Night",
        anim = "cheers_smash",
        desc = "{EXPEND} {1} {empty_bottle} card from your deck: {DRINK} {2}.",
        flavour = "'Why are those birds so loud?'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cost, self.drink_amount)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) > 0
        end,

        bottle_cost = 1,
        drink_amount = 2,

        min_damage = 4,
        max_damage = 6,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, self.bottle_cost) > 0 then
                Drink(self, battle, self.drink_amount)
            end
        end
    },

    long_night_plus =
    {
        name = "Boosted Long Night",
        min_damage = 5,
        max_damage = 8,
    },

    long_night_plus2 =
    {
        name = "Gulping Long Night",
        desc = "{EXPEND} {1} {empty_bottle} card from your deck: {DRINK} <#UPGRADE>{2}</>.",
        drink_amount = 3,
    },

    shatter = 
    {
        name = "Shatter",
        anim = "bottle_stab",
        desc = "{EXPEND} {1} {empty_bottle} cards from your deck: apply {2} {WOUND}.",
        flavour = "'Things are about to go real bad for you pal.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cost, self.wound_amount)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= self.bottle_cost
        end,

        bottle_cost = 2,
        wound_amount = 2,

        min_damage = 3,
        max_damage = 5,

        OnPostResolve = function( self, battle, attack )
            if GetBottleCount(self) >= self.bottle_cost then
                attack:AddCondition("WOUND", self.wound_amount, self)
                SpendBottles(self, self.bottle_cost)
            end
        end
    },

    shatter_plus = 
    {
        name = "Broken Shatter",
        desc = "{EXPEND} <#UPGRADE>{1}</> {empty_bottle} card from your deck: apply {2} {WOUND}.",
        bottle_cost = 1,
    },

    shatter_plus2 =
    {
        name = "Boosted Shatter",
        min_damage = 5,
        max_damage = 7,
    },

    endless_supply =
    {
        name = "Endless Supply",
        anim = "taunt",
        desc = "{EXPEND} up to 4 cards from your hand. Add that many {empty_bottle} cards to your discard pile.",
        flavour = "'Can I have that when you are finished?'",

        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local cards = battle:ExpendCards(0, 4, self)
            if #cards > 0 then
                for i=1,#cards do
                    local card = Battle.Card( "empty_bottle", self.owner )
                    battle:DealCard( card, battle:GetDiscardDeck() )
                end
            end
        end,
    },

    endless_supply_plus =
    {
        name = "Wide Supply",
        desc = "{EXPEND} <#UPGRADE>any number of</> cards from your hand. Add that many {empty_bottle} cards to your discard pile.",
        OnPostResolve = function( self, battle, attack )
            local cards = battle:ExpendCards(nil, nil, self)
            if #cards > 0 then
                for i=1,#cards do
                    local card = Battle.Card( "empty_bottle", self.owner )
                    battle:DealCard( card, battle:GetDiscardDeck() )
                end
            end
        end,
    },

    endless_supply_plus2 =
    {
        name = "Visionary Supply",
        desc = "<#UPGRADE>Draw 3 cards.</>\n{EXPEND} up to 4 cards from your hand. Add that many {empty_bottle} cards to your discard pile.",
        OnPostResolve = function( self, battle, attack )
            battle:DrawCards(3)
            local cards = battle:ExpendCards(0, 4, self)
            if #cards > 0 then
                for i=1,#cards do
                    local card = Battle.Card( "empty_bottle", self.owner )
                    battle:DealCard( card, battle:GetDiscardDeck() )
                end
            end
        end,
    },

    -- hoarder = 
    -- {
    --     name = "Hoarder",
    --     anim = "taunt",
    --     desc = "{ABILITY}: At the end of your turn, keep up to {1} {empty_bottle} in your hand instead of discarding it.",
    --     desc_fn = function( self, fmt_str )
    --         return loc.format(fmt_str, self.keep_count)
    --     end,

    --     cost = 1,

    --     flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
    --     rarity = CARD_RARITY.UNCOMMON,

    --     target_type = TARGET_TYPE.SELF,

    --     keep_count = 2,

    --     OnPostResolve = function( self, battle, attack )
    --         self.owner:AddCondition("hoarder", self.keep_count, self)
    --     end,

    --     condition = 
    --     {
    --         desc = "At the end of your turn, keep up to {1} {empty_bottle} {1*card|cards} in your hand instead of discarding it.",
    --         desc_fn = function( self, fmt_str )
    --             return loc.format(fmt_str, self.stacks)
    --         end,
    --         event_handlers =
    --         {
    --             [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
    --                 local count = 0
    --                 for i,card in battle:GetHandDeck():Cards() do
    --                     if card.id == "empty_bottle" then
    --                         self.stickied_cards = self.stickied_cards or {}
    --                         table.insert(self.stickied_cards, card)
    --                         card:SetFlags(CARD_FLAGS.STICKY)
    --                         count = count + 1
    --                         if count >= self.stacks then
    --                             break
    --                         end
    --                     end
    --                 end
    --             end,

    --             [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
    --                 if self.stickied_cards and #self.stickied_cards > 0 then
    --                     for i,card in ipairs(self.stickied_cards) do
    --                         card:ClearFlags(CARD_FLAGS.STICKY)
    --                     end
    --                     self.stickied_cards = {}
    --                 end
    --             end
    --         },
    --     },
    -- },

    -- hoarder_plus =
    -- {
    --     name = "Initial Hoarder",
    --     desc = "{ABILITY}: At the end of your turn, keep up to {1} {empty_bottle} in your hand instead of discarding it.",
    --     flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.AMBUSH,
    -- },

    -- hoarder_plus2 =
    -- {
    --     name = "Boosted Hoarder",
    --     desc = "{ABILITY}: At the end of your turn, keep up to <#UPGRADE>{1}</> {empty_bottle} in your hand instead of discarding it.",
    --     keep_count = 3,
    -- },

    packrat = 
    {
        name = "Packrat",
        anim = "taunt",
        desc = "{ABILITY}: At the start of each turn, move {1} {empty_bottle} {1*card|cards} from your draw pile to your discard.",
        on_play_screen_shake = true,
        flavour = "'Wow! This is a nice box!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_count)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        bottle_count = 1,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("packrat", self.bottle_count, self)
        end,

        condition = 
        {
            desc = "At the start of each turn, move <#HILITE>{1}</> {empty_bottle} {1*card|cards} from your draw pile to your discard.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle)
                    local bottles = ObtainWorkTable()
                    for i,card in battle:GetDrawDeck():Cards() do
                        if card.id == "empty_bottle" then
                            table.insert(bottles, card)
                            if #bottles >= self.stacks then
                                break
                            end
                        end
                    end
                    if #bottles > 0 then
                        for i,card in ipairs(bottles) do
                            card:TransferCard(battle:GetDiscardDeck())
                        end
                    end
                end
            },
        },
    },

    packrat_plus =
    {
        name = "Initial Packrat",
        desc = "{ABILITY}: At the start of each turn, move {1} {empty_bottle} {1*card|cards} from your draw pile to your discard.",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.AMBUSH,
    },

    packrat_plus2 =
    {
        name = "Boosted Packrat",
        desc = "{ABILITY}: At the start of each turn, move <#UPGRADE>{1}</> {empty_bottle} {1*card|cards} from your draw pile to your discard.",
        bottle_count= 2,
    },

    bottomless_vial =
    {
        name = "Bottomless Vial",
        anim = "drink",
        desc = "{DRINK}.\nGain {1} {smith_regeneration} and return this card to your hand and give it {BURNOUT}.",
        flavour = "'Bartender approved!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.moxie_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        moxie_amt = 3,
        drink_amount = 1,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_amount)
            local healing = self.owner:GetCondition("smith_regeneration")
            if healing then
                healing:IncreaseMax(self.moxie_amt, self)
            end
        end,

        deck_handlers = { DECK_TYPE.DISCARDS },

        event_handlers =
        {
            [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS and source_deck and source_deck:GetDeckType() == DECK_TYPE.RESOLVE then
                    -- Return!
                    self:SetFlags(CARD_FLAGS.BURNOUT)
                    self:TransferCard( self.engine:GetHandDeck() )
                end
            end
        }
    },

    bottomless_vial_plus =
    {
        name = "Gulping Bottomless Vial",
        desc = "<#UPGRADE>{DRINK} 3</>.\nGain {1} {smith_regeneration} and return this card to your hand and give it {BURNOUT}.",
        drink_amount = 3,
    },

    bottomless_vial_plus2 =
    {
        name = "Healing Bottomless Vial",
        desc = "{DRINK}.\nGain <#UPGRADE>{1} {smith_regeneration}</> and return this card to your hand and give it {BURNOUT}.",
        moxie_amt = 4,
    },

    entire_supply = 
    {
        name = "Entire Supply",
        anim = "bottle_throw",
        desc = "Hits all enemies. {EXPEND} all {empty_bottle} cards in your deck. Deal {1} bonus damage per card expended.",
        flavour = "'Now you’re finished!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bonus_damage)
        end,

        cost = 2,

        flags = CARD_FLAGS.RANGED | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        bonus_damage = 2,
        min_damage = 2,
        max_damage = 2,
        target_mod = TARGET_MOD.TEAM,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self)
        end,

        event_handlers = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    local bottles = GetBottleCount(self)
                    if bottles > 0 then
                        dmgt:AddDamage(self.bonus_damage * bottles, self.bonus_damage * bottles, self)
                    end
                end
            end
        },
    },

    entire_supply_plus =
    {
        name = "Pale Entire Supply",
        cost = 1,
    },

    entire_supply_plus2 =
    {
        name = "Boosted Entire Supply",
        min_damage = 4,
        max_damage = 4,
    },

    kradeshi_cup =
    {
        name = "Kra'deshi Cup",
        anim = "cheers_smash",
        desc = "{DRINK}.\n{CHAIN}{kradeshi_cup_II|}",
        flavour = "'More of a rite of passage than a drink.'",

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 4,
        max_damage = 4,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, 1)
            local card = Battle.Card("kradeshi_cup_II", self.owner)
            card.base_card = self
            card.auto_deal = true
            battle:DealCard(card, battle:GetDrawDeck())
            self:TransferCard( battle.trash_deck )
        end,
    },

        kradeshi_cup_II =
        {
            name = "Kra'deshi Cup II",
            icon = "battle/kradeshi_cup.tex",
            anim = "cheers_smash",
            desc = "{DRINK} 2.\n{CHAIN}{kradeshi_cup_III|}",
            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 4,
            max_damage = 4,

            OnPostResolve = function( self, battle, attack )
                Drink(self, battle, 2)
                local card = Battle.Card("kradeshi_cup_III", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        kradeshi_cup_III =
        {
            name = "Kra'deshi Cup III",
            icon = "battle/kradeshi_cup.tex",
            anim = "cheers_smash",
            desc = "{DRINK} 3.\n{CHAIN}{kradeshi_cup_IV|}",
            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 4,
            max_damage = 4,

            OnPostResolve = function( self, battle, attack )
                Drink(self, battle, 3)
                local card = Battle.Card("kradeshi_cup_IV", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        kradeshi_cup_IV = 
        {
            name = "Kra'deshi Cup IV",
            icon = "battle/kradeshi_cup.tex",
            anim = "cheers_smash",
            desc = "{DRINK} 4.\nDraw 2 cards.\n{CHAIN}",
            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 4,
            max_damage = 4,

            OnPostResolve = function( self, battle, attack )
                Drink(self, battle, 4)
                battle:DrawCards(2)
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        if source_deck and source_deck:GetDeckType() == DECK_TYPE.RESOLVE then
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            self:TransferCard(self.engine:GetDrawDeck(), idx)
                        else
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                            self:TransferCard( self.engine.trash_deck )
                        end
                    end
                end
            }
        },

    back_stretch = 
    {
        name = "Back Stretch",
        anim = "stretch",
        desc = "{ALLEVIATE} {1}.",
        flavour = "'Limber up!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.alleviate_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,

        alleviate_amt = 5,

        OnPostResolve = function ( self, battle, attack )
            self.owner:AlleviateHealth(self.alleviate_amt, self)
        end,
    },

    back_stretch_plus =
    {
        name = "Stony Back Stretch",
        desc = "{ALLEVIATE} {1},<#UPGRADE> gain {2} {DEFEND}</>.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.alleviate_amt, self:CalculateDefendText( self.defend_amt ))
        end,

        defend_amt = 3,

        OnPostResolve = function ( self, battle, attack )
            self.owner:AlleviateHealth(self.alleviate_amt, self)
            self.owner:AddCondition("DEFEND", self.defend_amt, self)
        end,
    },

    back_stretch_plus2 =
    {
        name = "Boosted Back Stretch",
        desc = "{ALLEVIATE} <#UPGRADE>{1}</>.",
        alleviate_amt = 8,
    },

    chest_pound =
    {
        name = "Chest Pound",
        anim = "chest_pound",
        desc = "Take {1} damage, gain {2} {POWER}.",
        flavour = "'I’ll take on any one of you losers here!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.self_damage, self.power_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,

        self_damage = 5,
        power_amt = 1,

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,

        OnPostResolve= function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
            self.owner:AddCondition("POWER", self.power_amt, self)
        end
    },

    chest_pound_plus =
    {
        name = "Biting Chest Pound",
        desc = "Take <#UPGRADE>{1}</> damage <#DOWNGRADE>twice</>, gain {2} {POWER}.",
        self_damage = 2,
        OnPostResolve= function( self, battle, attack )
            for i=1,2 do
                self.owner:SelfDamage( self.self_damage, self )
            end
            self.owner:AddCondition("POWER", self.power_amt, self)
        end
    },

    chest_pound_plus2 =
    {
        name = "Boosted Chest Pound",
        desc = "Take {1} damage, gain <#UPGRADE>{2} {POWER}</>.",
        power_amt = 2,
    },

    final_blow = 
    {
        name = "Hammer Down",
        anim = "super_smash",
        desc = "{POWER} and {ADRENALINE} are twice as effective on this card.",
        flavour = "'This is going to be a big one.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage_amt)
        end,

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        damage_mult = 1,

        min_damage = 7,
        max_damage = 12,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and (self.owner:HasCondition("POWER") or self.owner:HasCondition("ADRENALINE")) then
                    local stacks = (self.owner:GetConditionStacks("POWER") + self.owner:GetConditionStacks("ADRENALINE")) * self.damage_mult
                    dmgt:AddDamage(stacks, stacks, self)
                end
            end
        },
    },

    final_blow_plus =
    {
        name = "Boosted Hammer Down",
        desc = "{POWER} and {ADRENALINE} are <#UPGRADE>three times</> as effective on this card.",
        damage_mult = 2,
    },

    final_blow_plus2 =
    {
        name = "Pale Hammer Down",
        cost = 1,
    },

    headbutt =
    {
        name = "Headbutt",
        anim = "headbutt",
        desc = "Take {1} damage. Apply {2} {IMPAIR}.",
        flavour = "'I wasn’t using it.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.self_damage, self.impair_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        min_damage = 4,
        max_damage = 6,

        self_damage = 4,
        impair_amt = 2,

        CanPlayCard = function( self, battle, target )
            return CanPlaySelfDamage( self, self.self_damage, self.self_damage_hits )
        end,

        OnPostResolve= function( self, battle, attack )
            self.owner:SelfDamage( self.self_damage, self )
            attack:AddCondition("IMPAIR", self.impair_amt, self)
        end,
    },

    headbutt_plus =
    {
        name = "Softened Headbutt",
        desc = "Take <#UPGRADE>{1}</> damage. Apply {2} {IMPAIR}.",
        self_damage = 3,
    },

    headbutt_plus2 =
    {
        name = "Stifling Headbutt",
        desc = "Take {1} damage. Apply <#UPGRADE>{2}</> {IMPAIR}.",
        impair_amt = 3,
    },

    power_through = 
    {
        name = "Power Through",
        anim = "taunt",
        desc = "For the rest of your turn, any damage you receive is reduced to 1.",
        flavour = "'Pain is just weakness leaving the body, or something.'",
        
        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("power_through", 1, self)
        end,

        condition = 
        {
            priority = 1, -- Before Defend

            OnPreDamage = function( self, damage, attacker, battle, source, piercing )
                if battle:GetCurrentTeam() == self.owner:GetTeam() then
                    return math.min(damage, 1)
                else
                    return damage
                end
            end,

            event_priorities = {
                [ BATTLE_EVENT.END_TURN ] = 999,
            },

            event_handlers = 
            {
                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:RemoveCondition(self.id, 1)
                    end
                end
            },
        },
    },

    power_through_plus =
    {
        name = "Pale Power Through",
        cost = 0,
    },

    power_through_plus2 =
    {
        name = "Enduring Power Through",
        flags = CARD_FLAGS.SKILL,
    },

    hammer_swing = 
    {
        name = "Hammer Swing",
        anim = "heavy_smash",
        desc = "Apply {1} {TRAUMA}.\n{CHAIN}{hammer_swing_II|}",
        flavour = "'If you can keep up the momentum there’s no stopping you.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.trauma_amt)
        end,

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 2,
        max_damage = 2,
        trauma_amt = 2,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("TRAUMA", self.trauma_amt, self)
            local card = Battle.Card("hammer_swing_II", self.owner)
            card.base_card = self
            card.auto_deal = true
            battle:DealCard(card, battle:GetDrawDeck())
            self:TransferCard( battle.trash_deck )
        end,
    },
        hammer_swing_II =
        {
            name = "Hammer Swing II",
            icon = "battle/hammer_swing.tex",
            anim = "heavy_smash",
            desc = "Apply {1} {TRAUMA}.\n{CHAIN}{hammer_swing_III|}",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.trauma_amt)
            end,

            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 2,
            max_damage = 2,
            trauma_amt = 2,

            OnPostResolve = function( self, battle, attack )
                attack:AddCondition("TRAUMA", self.trauma_amt, self)
                local card = Battle.Card("hammer_swing_III", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        hammer_swing_III =
        {
            name = "Hammer Swing III",
            icon = "battle/hammer_swing.tex",
            anim = "heavy_smash",
            desc = "Apply {1} {TRAUMA}.\n{CHAIN}{hammer_swing_IV|}",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.trauma_amt)
            end,

            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 4,
            max_damage = 4,
            trauma_amt = 2,

            OnPostResolve = function( self, battle, attack )
                attack:AddCondition("TRAUMA", self.trauma_amt, self)
                local card = Battle.Card("hammer_swing_IV", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        hammer_swing_IV =
        {
            name = "Hammer Swing IV",
            icon = "battle/hammer_swing.tex",
            anim = "heavy_smash",
            hit_anim = true,
            desc = "Attack twice.\n{CHAIN}",

            hide_in_cardex = true,

            cost = 1,

            flags = CARD_FLAGS.MELEE,
            rarity = CARD_RARITY.UNIQUE,

            min_damage = 8,
            max_damage = 8,
            hit_count = 2,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        if source_deck and source_deck:GetDeckType() == DECK_TYPE.RESOLVE then
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            self:TransferCard(self.engine:GetDrawDeck(), idx)
                        else
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                            self:TransferCard( self.engine.trash_deck )
                        end
                    end
                end
            }
        },

    obliterate =
    {
        name = "Obliterate",
        anim = "smash",
        flavour = "'Bonk!'",
        desc = "Apply 1 {STUN}.",

        cost = 3,

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,

        min_damage = 10,
        max_damage = 12,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("STUN", 1, self)
        end,
    },

    obliterate_plus =
    {
        name = "Pale Obliterate",
        cost = 2,
    },

    obliterate_plus2 =
    {
        name = "Boosted Obliterate",
        min_damage = 16,
        max_damage = 18,
    },

    bottle_hurl =
    {
        name = "Bottle Hurl",
        anim = "bottle_throw",
        desc = "{EXPEND} {1} {empty_bottle} cards from your deck to play this.",
        flavour = "'This one is on me.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cost)
        end,

        cost = 0,
        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.COMMON,

        min_damage = 5,
        max_damage = 5,

        bottle_cost = 2,

        CanPlayCard = function( self, battle, target )
            return GetBottleCount(self) >= self.bottle_cost, CARD_PLAY_REASONS.NO_BOTTLES
        end,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= self.bottle_cost
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self, self.bottle_cost)
        end,
    },

    bottle_hurl_plus = 
    {
        name = "Boosted Bottle Hurl",
        min_damage = 8,
        max_damage = 8,
    },

    bottle_hurl_plus2 =
    {
        name = "Pale Bottle Hurl",
        desc = "{EXPEND} <#UPGRADE>{1}</> {empty_bottle} card from your deck to play this.",
        bottle_cost = 1,
    },

    splinter =
    {
        name = "Splinter",
        anim = "bottle_throw",
        desc = "Attack a random enemy.\n{EXPEND} 1 {empty_bottle} card from your deck: Apply {1} {IMPAIR}.",
        flavour = "'Gah! Almost got it in my eyes!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.impair_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.COMMON,

        target_mod = TARGET_MOD.RANDOM1,

        min_damage = 5,
        max_damage = 7,
        impair_amt = 2,
        bottle_cost = 1,

        PreReq = function( self, battle, target )
            return GetBottleCount(self) >= self.bottle_cost
        end,

        OnPostResolve = function( self, battle, attack )
            if SpendBottles(self, self.bottle_cost) > 0 then
                attack:AddCondition("IMPAIR", self.impair_amt)
            end
        end,
    },

    splinter_plus =
    {
        name = "Boosted Splinter",
        min_damage = 8, 
        max_damage = 8,
    },

    splinter_plus2 = 
    {
        name = "Aimed Splinter",
        desc = "{EXPEND} 1 {empty_bottle} card from your deck: Apply {1} {IMPAIR}.",
        target_mod = TARGET_MOD.SINGLE,
    },

    slug_smash =
    {
        name = "Slug Smash",
        anim = "back_hamm",
        desc = "This card deals bonus damage equal to your {smith_regeneration}.",
        flavour = "'Where is he getting all this strength from?'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage_gain)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 0,
        max_damage = 0,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    local moxie = self.owner:GetCondition("smith_regeneration")
                    if moxie then
                        dmgt:AddDamage(moxie.max_heal, moxie.max_heal, self)
                    end
                end
            end,
        },
    },

    slug_smash_plus =
    {
        name = "Boosted Slug Smash",
        min_damage = 3,
        max_damage = 3,
    },

    slug_smash_plus2 =
    {
        name = "Soothing Slug Smash",
        desc = "This card deals bonus damage equal to your {smith_regeneration}.\n<#UPGRADE>{ALLEVIATE} 5 health.</>",
        OnPostResolve = function( self, battle, attack )
            self.owner:AlleviateHealth(5, self)
        end
    },

    bio_strike =
    {
        name = "Bio-strike",
        anim = "punch",
        desc = "Increase your {smith_regeneration} by the amount of damage dealt by this card.",
        flavour = "'Yeah it’s wired right into your bio-metrics. It’s not something that you are supposed to be doing. There’s legality involved.'",

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.RARE,

        min_damage = 4,
        max_damage = 6,

        event_handlers = 
        {
            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if hit.card == self then
                    local healing = self.owner:GetCondition("smith_regeneration")
                    if healing then healing:IncreaseMax(hit.damage or 0, self) end
                end
            end
        },
    },

    bio_strike_plus =
    {
        name = "Rooted Bio-Strike",
        min_damage = 6,
    },

    bio_strike_plus2 =
    {
        name = "Tall Bio-Strike",
        max_damage = 10,
    },

    meditation =
    {
        name = "Meditation",
        anim = "taunt",
        desc = "Increase your {smith_regeneration} by {1} this turn.\n{ALLEVIATE} {2} health.\n{CHAIN}{meditation_II|}",
        flavour = "'Clear your mind, no food or drink allowed on the premises sir.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.regen_amt, self.alleviate_amt )
        end,

        cost = 2,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        regen_amt = 1,
        alleviate_amt = 2,

        OnPostResolve = function( self, battle, attack )
            local healing = self.owner:GetCondition("smith_regeneration")
            if healing then
                healing:IncreaseMax(self.regen_amt, self)
            end
            self.owner:AlleviateHealth(self.alleviate_amt, self)
            local card = Battle.Card("meditation_II", self.owner)
            card.base_card = self
            card.auto_deal = true
            battle:DealCard(card, battle:GetDrawDeck())
            self:TransferCard( battle.trash_deck )
        end,
    },

        meditation_II =
        {
            name = "Meditation II",
            icon = "battle/meditation.tex",
            anim = "taunt",
            hide_in_cardex = true,
            desc = "Increase your {smith_regeneration} by {1} this turn.\n{ALLEVIATE} {2} health.\n{CHAIN}{meditation_III|}",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.regen_amt, self.alleviate_amt )
            end,

            cost = 1,

            flags = CARD_FLAGS.SKILL,
            rarity = CARD_RARITY.UNIQUE,

            target_type = TARGET_TYPE.SELF,

            regen_amt = 2,
            alleviate_amt = 3,

            OnPostResolve = function( self, battle, attack )
                local healing = self.owner:GetCondition("smith_regeneration")
                if healing then
                    healing:IncreaseMax(self.regen_amt, self)
                end
                self.owner:AlleviateHealth(self.alleviate_amt, self)
                local card = Battle.Card("meditation_III", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        meditation_III =
        {
            name = "Meditation III",
            icon = "battle/meditation.tex",
            anim = "taunt",
            hide_in_cardex = true,
            desc = "Increase your {smith_regeneration} by {1} this turn.\n{ALLEVIATE} {2} health.\n{CHAIN}{meditation_IV|}",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.regen_amt, self.alleviate_amt )
            end,

            cost = 1,

            flags = CARD_FLAGS.SKILL,
            rarity = CARD_RARITY.UNIQUE,

            target_type = TARGET_TYPE.SELF,

            regen_amt = 3,
            alleviate_amt = 4,

            OnPostResolve = function( self, battle, attack )
                local healing = self.owner:GetCondition("smith_regeneration")
                if healing then
                    healing:IncreaseMax(self.regen_amt, self)
                end
                self.owner:AlleviateHealth(self.alleviate_amt, self)
                local card = Battle.Card("meditation_IV", self.owner)
                card.base_card = self.base_card
                card.auto_deal = true
                battle:DealCard(card, battle:GetDrawDeck())
                self:TransferCard( battle.trash_deck )
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                        if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                        self:TransferCard( self.engine.trash_deck )
                    end
                end
            }
        },

        meditation_IV =
        {
            name = "Meditation IV",
            icon = "battle/meditation.tex",
            anim = "taunt",
            hide_in_cardex = true,
            desc = "Increase your {smith_regeneration} by {1} this turn.\nTrigger your {smith_regeneration}.\n{CHAIN}",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.regen_amt )
            end,

            cost = 1,

            flags = CARD_FLAGS.SKILL,
            rarity = CARD_RARITY.UNIQUE,

            target_type = TARGET_TYPE.SELF,

            regen_amt = 3,

            OnPostResolve = function( self, battle, attack )
                local healing = self.owner:GetCondition("smith_regeneration")
                if healing then
                    healing:IncreaseMax(self.regen_amt, self)
                    healing:ActivateHealing(self, true)
                end
            end,

            deck_handlers = { DECK_TYPE.DISCARDS },

            event_handlers =
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card == self and target_deck and target_deck:GetDeckType() == DECK_TYPE.DISCARDS then
                        if source_deck and source_deck:GetDeckType() == DECK_TYPE.RESOLVE then
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            self:TransferCard(self.engine:GetDrawDeck(), idx)
                        else
                            local idx = self.engine:GetDrawDeck():CountCards() > 1 and math.random( self.engine:GetDrawDeck():CountCards() ) or 1
                            if self.base_card then self.base_card:TransferCard(self.engine:GetDrawDeck(), idx) end
                            self:TransferCard( self.engine.trash_deck )
                        end
                    end
                end
            }
        },  

    bloodflow =
    {
        name = "Blood Flow",
        anim = "taunt",
        desc = "Double your {smith_regeneration}.",
        flavour = "'Feels… funny.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.RARE,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local healing = self.owner:GetCondition("smith_regeneration")
            if healing then
                if self.bonus then
                    healing:IncreaseMax(self.bonus, self)
                end
                healing:IncreaseMax(healing.max_heal, self)
            end
        end,

        features =
        {
            EXERT = 1,
        },
    },

    bloodflow_plus =
    {
        name = "Boosted Blood Flow",
        desc = "<#UPGRADE>Gain 3 {smith_regeneration}, then</> double your {smith_regeneration}.",
        bonus = 3,
    },

    bloodflow_plus2 = 
    {
        name = "Pure Blood Flow",
        features = {},
    },

    juicer =
    {
        name = "Juicer",
        anim = "taunt",
        desc = "{ALLEVIATE} health equal to your {smith_regeneration}.",   
        flavour = "'Are you sure that’s legal?'",

        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,

        target_type = TARGET_TYPE.SELF,

        PreReq = function( self, battle )
            return self.owner:GetConditionStacks("RECOVERABLE_HEALTH") > 0
        end,

        OnPostResolve = function ( self, battle, attack )
            local healing = self.owner:GetCondition("smith_regeneration")
            if healing and healing.max_heal > 0 then
                self.owner:AlleviateHealth(healing.max_heal, self)
            end
        end
    },

    juicer_plus =
    {
        name = "Enduring Juicer",
        flags = CARD_FLAGS.SKILL,
    },

    juicer_plus2 = 
    {
        name = "Pale Juicer",
        cost = 0,
    },

    nap =
    {
        name = "Nap",
        anim = "stretch",
        desc = "Apply {1} {DEFEND}.\nDraw {2} cards next turn.",
        flavour = "'How can you sleep at a time like this?'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:CalculateDefendText(self.defend_amt), self.draw_count)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

        defend_amt = 4,
        draw_count = 2,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("DEFEND", self.defend_amt, self)
            self.owner:AddCondition("nap", self.draw_count, self)
        end,

        condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:DrawCards(self.stacks)
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    nap_plus =
    {
        name = "Stone Nap",
        desc = "Apply <#UPGRADE>{1}</> {DEFEND}.\nDraw {2} cards next turn.",
        defend_amt = 6,
    },

    nap_plus2 =
    {
        name = "Visionary Nap",
        desc = "Apply {1} {DEFEND}.\nDraw <#UPGRADE>{2}</> cards next turn.",
        draw_count = 3,
    },

    stoke =
    {
        name = "Stoke",
        anim = "taunt",
        desc = "Gain {1} {ADRENALINE} at the start of your next turn, you can only play 1 attack that turn.",
        flavour = "'Oh yeah, I’m getting excited for this one.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        adrenaline_amt = 5,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("stoke", self.adrenaline_amt, self)
        end,

        condition = 
        {
            hidden = true,
            loc_strings =
            {
                CANT_PLAY = "Stoke: Can only play 1 attack this turn.",
            },

            CanPlayCard = function( self, battle, target, card )
                if card and card:IsAttackCard() and card.id ~= "execute" then
                    if (self.attacks_played or 0) >= 1 then
                        return false, self.feature_def:GetLocalizedString("CANT_PLAY")
                    end
                end
                return true
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.active = true
                    self.owner:AddCondition("ADRENALINE", self.stacks, self)
                    self.adrenaline_added = (self.adrenaline_added or 0) + self.stacks
                    self.attacks_played = 0
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    if self.adrenaline_added and self.adrenaline_added > 0 then
                        self.owner:RemoveCondition(self.id, self.adrenaline_added, self)
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if self.active and card:IsAttackCard() then
                        self.attacks_played = (self.attacks_played or 0) + 1
                    end
                end
            },
        },
    },

    stoke_plus =
    {
        name = "Boosted Stoke",
        desc = "Gain <#UPGRADE>{1}</> {ADRENALINE} at the start of your next turn, you can only play 1 attack that turn.",
        adrenaline_amt = 7,
    },

    stoke_plus2 =
    {
        name = "Pale Stoke",
        cost = 0,
    },

    shelve =
    {
        name = "Shelve",
        anim = "bottle_flourish",
        desc = "Choose a card in your hand, place it on top of your draw pile.",
        flavour = "'That’s a great idea!'",

        cost = 0,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,
        draw_count = 0,

        OnPostResolve = function( self, battle, attack )
            if self.draw_count > 0 then
                battle:DrawCards(self.draw_count)
            end
            local card = battle:ChooseCard()
            if card then
                card:TransferCard(battle:GetDrawDeck(), battle:GetDrawDeck():CountCards() + 1)
            end
        end
    },

    shelve_plus =
    {
        name = "Sticky Shelve",
        desc = "Choose a card in your hand, place it on top of your draw pile.",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.STICKY,
    },

    shelve_plus2 = 
    {
        name = "Visionary Shelve",
        desc = "<#UPGRADE>Draw a card.</>\nChoose a card in your hand, place it on top of your draw pile.",
        draw_count = 1,
    },

    kiss_the_fists = 
    {
        name = "Kiss the Fists",
        anim = "gut_shot",
        hit_anim = true,
        desc = "Attack with this card twice.",
        flavour = "'For luck.'",

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        min_damage = 2,
        max_damage = 3,

        target_count = 2,
    },

    kiss_the_fists_plus =
    {
        name = "Mirrored Kiss the Fists",
        desc = "Attack with this card <#UPGRADE>three times</>.",
        target_count = 3,
    },

    kiss_the_fists_plus2 =
    {
        name = "Boosted Kiss the Fists",
        min_damage = 4,
        max_damage = 6,
    },

    hole_up =
    {
        name = "Hole Up",
        anim = "taunt",
        desc = "Gain {1} {DEFEND} at the start of your next turn.",
        flavour = "'Back in the hole you mutts!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.defend_amt, self.immediate_defend or 0)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,

        target_type = TARGET_TYPE.SELF,
        defend_amt = 6,

        OnPostResolve = function( self, battle, attack )
            if self.immediate_defend then
                attack:AddCondition("DEFEND", self.immediate_defend, self)
            end
            self.owner:AddCondition("NEXT_TURN_DEFENSE", self.defend_amt, self)
        end,
    },

    hole_up_plus =
    {
        name = "Stone Hole Up",
        desc = "Gain <#UPGRADE>{1}</> {DEFEND} at the start of your next turn.",
        defend_amt = 9,
    },

    hole_up_plus2 =
    {
        name = "Twisted Hole Up",
        desc = "Apply <#UPGRADE>{2} {DEFEND} this turn and </>gain {1} {DEFEND} at the start of your next turn.",
        immediate_defend = 3,
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
    },

    intensify =
    {
        name = "Intensify",
        anim = "taunt",
        desc = "Choose a card in your hand, it deals double damage until played.",
        flavour = "'Don’t tell anyone where you got those, I’m in enough trouble as it is.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local card = battle:ChooseCard()
            if card then
                local con = self.owner:AddCondition("intensify", 1, self)
                if con then
                    con.chosen_cards = con.chosen_cards or {}
                    table.insert(con.chosen_cards, card)
                end
            end
        end,

        condition =
        {
            hidden = true,
            event_priorities =
            {
                [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
            },

            event_handlers =
            {
                [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                    if self.chosen_cards and table.contains(self.chosen_cards, card) then
                        local count = table.count_if(self.chosen_cards, function( v ) return v == card end)
                        for i=1,count do
                            dmgt:AddDamage( dmgt.min_damage, dmgt.max_damage, self )
                        end
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if self.chosen_cards and table.contains(self.chosen_cards, card) then
                        table.arrayremoveall(self.chosen_cards, card)
                        if #self.chosen_cards == 0 then
                            self.owner:RemoveCondition(self.id)
                        end
                    end
                end,
            },
        },
    },

    intensify_plus =
    {
        name = "Pale Intensify",
        cost = 0,
    },

    intensify_plus2 =
    {
        name = "Enduring Intensify",
        flags = CARD_FLAGS.SKILL,
    },

    conquer =
    {
        name = "Conquer",
        anim = "super_smash",
        desc = "Hits all enemies.\nThis card deals bonus damage equal to half of your {DEFEND}.",
        flavour = "'We should've made our move earlier, we’re doomed.'",

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        target_mod = TARGET_MOD.TEAM,

        min_damage = 4,
        max_damage = 4,

        mult = 0.5,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self then
                    local bonus = math.floor(self.owner:GetConditionStacks("DEFEND") * self.mult)
                    dmgt:AddDamage( bonus, bonus, self )
                end
            end,
        },
    },

    conquer_plus =
    {
        name = "Focused Conquer",
        desc = "This card deals bonus damage <#UPGRADE>equal to your {DEFEND}</>.",
        target_mod = TARGET_MOD.SINGLE,
        mult = 1,
    },

    conquer_plus2 =
    {
        name = "Boosted Conquer",
        min_damage = 7,
        max_damage = 7,
    },

    body_blow = 
    {
        name = "Body Blow",
        anim = "smash",
        desc = "If your target has {TRAUMA} or {TRAUMATIZED}, this deals double damage.",
        flavour = "'He’ll be feeling that in the morning.'",

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 2,
        max_damage = 4,

        event_priorities = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and target and (target:HasCondition("TRAUMA") or target:HasCondition("TRAUMATIZED")) then
                    dmgt:AddDamage(dmgt.min_damage, dmgt.max_damage, self)
                end
            end,
        },
    },

    body_blow_plus =
    {
        name = "Rooted Body Blow",
        min_damage = 4,
    },

    body_blow_plus2 =
    {
        name = "Tall Body Blow",
        max_damage = 6,
    },

    the_sledge =
    {
        name = "The Sledge",
        anim = "heavy_smash",
        desc = "If your target has {TRAUMA}, apply {1} {IMPAIR}.",
        flavour = "'All over this land.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.impair_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 3,
        max_damage = 5,

        impair_amt = 2,

        event_priorities =
        {
            [ BATTLE_EVENT.ON_HIT ] = 9, -- before trauma
        },

        event_handlers =
        {
            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if not hit.evaded and attack.card == self and hit.target:HasCondition("TRAUMA") then
                    hit.target:AddCondition("IMPAIR", self.impair_amt, self)
                end
            end,
        },
    },

    the_sledge_plus =
    {
        name = "The Stifling Sledge",
        desc = "If your target has {TRAUMA}, apply <#UPGRADE>{1}</> {IMPAIR}.",
        impair_amt = 3,
    },

    the_sledge_plus2 =
    {
        name = "The Rooted Sledge",
        min_damage = 5,
    },

    hurricane =
    {
        name = "Hurricane",
        anim = "bottle_throw_quick",
        hit_anim = true,
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck: Attack a random enemy and apply {2} {TRAUMA} per card expended.",
        flavour = "'I’m not going near that thing!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bottle_cap, self.trauma_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.UNCOMMON,
        target_mod = TARGET_MOD.RANDOM1,

        min_damage = 3,
        max_damage = 4,

        bottle_cap = 3,
        trauma_amt = 1,

        CanPlayCard = function( self, battle, target )
            if self:UpdateTargetCount() == 0 then
                return false, CARD_PLAY_REASONS.NO_BOTTLES
            end
            return true
        end,

        PreReq = function( self, battle, target )
            return self:UpdateTargetCount() > 0
        end,

        UpdateTargetCount = function( self )
            self.target_count = math.min(GetBottleCount(self), self.bottle_cap)
            return self.target_count
        end,

        OnPostResolve = function( self, battle, attack )
            SpendBottles(self, self.target_count)
            self.target_count = 0
            for i,hit in ipairs(attack.hits) do
                if not hit.evaded then
                    hit.target:AddCondition("TRAUMA", self.trauma_amt, self)
                end
            end
        end,
    },

    hurricane_plus =
    {
        name = "Boosted Hurricane",
        desc = "{EXPEND} up to <#UPGRADE>{1}</> {empty_bottle} cards from your deck: Attack a random enemy and apply {2} {TRAUMA} per card expended.",
        bottle_cap = 4,
    },

    hurricane_plus2 =
    {
        name = "Traumatic Hurricane",
        desc = "{EXPEND} up to {1} {empty_bottle} cards from your deck: Attack a random enemy and apply <#UPGRADE>{2}</> {TRAUMA} per card expended.",
        trauma_amt = 2,
    },

    shrewd =
    {
        name = "Shrewd",
        anim = "taunt",
        desc = "Gain {DEFEND} equal to the damage of your target's next attack.",
        flavour = "'Nice try.'",

        cost = 2,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        bonus = 0,

        OnPostResolve = function( self, battle, attack )
            local defend_amt = 0
            for i,hit in ipairs(attack.hits) do
                if hit.target and hit.target.prepared_attacks then
                    for i,prep_attack in ipairs(hit.target.prepared_attacks) do
                        for i,prep_hit in ipairs(prep_attack.hits) do
                            if prep_hit.damage then
                                defend_amt = defend_amt + prep_hit.damage
                            end
                        end
                    end
                end
            end
            if defend_amt + self.bonus > 0 then
                self.owner:AddCondition("DEFEND", defend_amt + self.bonus, self)
            end
            if self.counter_amt then
                self.owner:AddCondition("RIPOSTE", self.counter_amt, self)
            end
        end,
    },

    shrewd_plus =
    {
        name = "Boosted Shrewd",
        desc = "Gain {DEFEND} equal to the damage of your target's next attack <#UPGRADE>+ {1}</>.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bonus)
        end,
        bonus = 4,
    },

    shrewd_plus2 =
    {
        name = "Thorny Shrewd",
        desc = "Gain {DEFEND} equal to the damage of your target's next attack.\n<#UPGRADE>Gain {RIPOSTE {1}}.</>",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.counter_amt )
        end,
        counter_amt = 5,
    },

    frenzy =
    {
        name = "Frenzy",
        anim = "smash",
        desc = "While this card is in your hand, it gains {1} damage until played whenever you take damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.damage_delta)
        end,
        flavour = "'Come on! Is that all you wimps got!?'",

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 4,
        max_damage = 6,

        damage_delta = 1,
        reset = true,

        OnPostResolve = function( self, battle, attack )
            if self.reset then
                self.bonus_damage = 0
            end
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and self.bonus_damage then
                    dmgt:AddDamage( self.bonus_damage, self.bonus_damage, self)
                end
            end,

            [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                if fighter == self.owner and stat == COMBAT_STAT.HEALTH and delta < 0 and self.deck == self.engine:GetHandDeck() then
                    self.bonus_damage = (self.bonus_damage or 0) + self.damage_delta
                end
            end,
        },
    },

    frenzy_plus =
    {
        name = "Boosted Frenzy",
        desc = "While this card is in your hand, it gains <#UPGRADE>{1}</> damage until played whenever you take damage.",
        damage_delta = 2,
    },

    frenzy_plus2 =
    {
        name = "Lasting Frenzy",
        desc = "While this card is in your hand, it gains {1} damage until <#UPGRADE>the end of this battle</> whenever you take damage.",
        reset = false,
        min_damage = 2,
        max_damage = 4,
    },

    the_pain_train =
    {
        name = "The Pain Train",
        anim = "super_smash",
        desc = "Whenever you take damage, reduce the cost of this card by {1} until played.",
        flavour = "'All aboard!'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.cost_reduction)
        end,
        flavour = "'All aboard.'",

        cost = 10,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.RARE,

        min_damage = 8,
        max_damage = 14,

        cost_reduction = 1,

        OnPostResolve = function( self, battle, attack )
            self.discount = 0
        end,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                if card == self and self.discount then
                    cost_acc:AddValue( -self.discount, self)
                end
            end,

            [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                if fighter == self.owner and stat == COMBAT_STAT.HEALTH and delta < 0 then
                    self.discount = (self.discount or 0) + self.cost_reduction
                end
            end,
        },
    },

    the_pain_train_plus =
    {
        name = "The Rooted Pain Train",
        min_damage = 15,
        max_damage = 15,
    },

    the_pain_train_plus2 =
    {
        name = "The Pale Pain Train",
        desc = "Whenever you take damage, reduce the cost of this card by <#UPGRADE>{1}</> until played.",
        cost_reduction = 2,
    },

    desperation =
    {
        name = "Desperation",
        anim = "smash",
        desc = "While you have less than 50% health, this card costs 1 less.",
        flavour = "'desperate times...'",

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        min_damage = 10,
        max_damage = 12,
        health_pct = 0.5,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                if card == self and self.owner:GetHealthPercent() < self.health_pct then
                    cost_acc:AddValue( -1, self)
                end
            end,
        },
    },

    desperation_plus =
    {
        name  = "Boosted Desperation",
        min_damage = 12,
        max_damage = 14,
    },

    desperation_plus2 =
    {
        name = "Pale Desperation",
        desc = "While you have less than <#UPGRADE>65%</> health, this card costs 1 less.",
        health_pct = 0.65,
    },

    power_trip =
    {
        name = "Power Trip",
        anim = "taunt",
        desc = "Gain {1} {ADRENALINE}. If you had no {ADRENALINE}, gain an additional {2} {ADRENALINE}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.base_gain, self.max_gain)
        end,
        flavour = "'It ain’t bragging if it’s true.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        base_gain = 1,
        max_gain = 2,

        OnPostResolve = function( self, battle, attack )
            if not self.owner:HasCondition("ADRENALINE") then
                self.owner:AddCondition("ADRENALINE", self.max_gain, self)
            end
            self.owner:AddCondition("ADRENALINE", self.base_gain, self)
        end
    },

    power_trip_plus =
    {
        name = "Pale Power Trip",
        cost = 0,
    },

    power_trip_plus2 =
    {
        name = "Boosted Power Trip",
        desc = "Gain <#UPGRADE>{1}</> {ADRENALINE}. If you had no {ADRENALINE}, gain an additional {2} {ADRENALINE}.",
        base_gain = 2,
    },

    mettle =
    {
        name = "Fortitude",
        anim = "taunt",
        desc = "Convert up to {1} {ADRENALINE} into {POWER}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_cap)
        end,
        flavour = "'I’m just getting started.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.RARE,

        target_type = TARGET_TYPE.SELF,

        loc_strings = 
        {
            NO_ADRENALINE = "Requires at least 1 Adrenaline"
        },

        CanPlayCard = function( self, battle, target )
            return self.owner:HasCondition("ADRENALINE"), self.def:GetLocalizedString("NO_ADRENALINE")
        end,

        adrenaline_cap = 2,

        OnPostResolve = function( self, battle, attack )
            local count = math.min(self.adrenaline_cap, self.owner:GetConditionStacks("ADRENALINE"))
            if count > 0 then
                self.owner:RemoveCondition("ADRENALINE", count, self)
                self.owner:AddCondition("POWER", count, self)
                if self.defend_gain then
                    self.owner:AddCondition("DEFEND", count * 2, self)
                end
            end
        end
    },

    mettle_plus =
    {
        name = "Boosted Fortitude",
        desc = "Convert up to <#UPGRADE>{1}</> {ADRENALINE} into {POWER}.",
        adrenaline_cap = 3,
    },

    mettle_plus2 =
    {
        name = "Stone Fortitude",
        desc = "Convert up to {1} {ADRENALINE} into {POWER}<#UPGRADE> and twice that much {DEFEND}</>.",
        defend_gain = true,
    },

    pivot =
    {
        name = "Pivot",
        anim = "taunt",
        desc = "Convert up to {1} {ADRENALINE} into twice that much {DEFEND}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.adrenaline_cap)
        end,
        flavour = "'I’ve never seen anything keep fighting for that long.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        CanPlayCard = function( self, battle, target )
            return self.owner:HasCondition("ADRENALINE")
        end,

        adrenaline_cap = 4,

        OnPostResolve = function( self, battle, attack )
            local count = math.min(self.adrenaline_cap, self.owner:GetConditionStacks("ADRENALINE"))
            if count > 0 then
                self.owner:RemoveCondition("ADRENALINE", count, self)
                if self.alleviate then
                    self.owner:AlleviateHealth(count, self)
                else
                    self.owner:AddCondition("DEFEND", count * 2, self)
                end
            end
        end
    },

    pivot_plus =
    {
        name = "Boosted Pivot",
        desc = "Convert up to <#UPGRADE>{1}</> {ADRENALINE} into twice that much {DEFEND}.",
        adrenaline_cap = 6,
    },

    pivot_plus2 =
    {
        name = "Soothing Pivot",
        desc = "Remove up to <#UPGRADE>{1}</> {ADRENALINE} <#UPGRADE>and {ALLEVIATE} that much health.</>",
        adrenaline_cap = 6,
        alleviate = true,
    },

    ruthless =
    {
        name = "Ruthless",
        anim = "knee",
        desc = "If your target has {TRAUMA}, deal {1} bonus damage.",
        flavour = "'That’s some of the most awful stuff I’ve ever seen done to somebody.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.bonus_damage)
        end,
        flavour = "'Uhh… Pretty sure I saw that guy give up 6 punches ago.'",

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        min_damage = 3,
        max_damage = 5,

        bonus_damage = 3,

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card == self and target and target:HasCondition("TRAUMA") then
                    dmgt:AddDamage( self.bonus_damage, self.bonus_damage, self)
                end
            end
        },
    },

    ruthless_plus =
    {
        name = "Boosted Ruthless",
        desc = "If your target has {TRAUMA}, deal <#UPGRADE>{1}</> bonus damage.",
        bonus_damage = 5,
    },

    ruthless_plus2 =
    {
        name = "Rooted Ruthless",
        min_damage = 6,
        max_damage = 6
    },

    rib_cracker =
    {
        name = "Rib Cracker",
        anim = "punch",
        desc = "If your target has {TRAUMA}, apply {1} {WOUND}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.wound_amount)
        end,
        flavour = "'Thanks for helping me break in my new boots!'",

        hit_anim = true,

        cost = 1,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.COMMON,

        min_damage = 3,
        max_damage = 5,

        wound_amount = 1,

        PreReq = function( self, battle, target )
            return target and target:HasCondition("TRAUMA")
        end,

        event_priorities =
        {
            [ BATTLE_EVENT.ON_HIT ] = 9, -- before trauma
        },

        event_handlers =
        {
            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                if not hit.evaded and attack.card == self and hit.target:HasCondition("TRAUMA") then
                    hit.target:AddCondition("WOUND", self.wound_amount, self)
                end
            end,
        },
    },

    rib_cracker_plus =
    {
        name = "Wide Rib Cracker",
        desc = "<#UPGRADE>Hits all enemies.</>\nIf your target has {TRAUMA}, apply {1} {WOUND}.",
        target_mod = TARGET_MOD.TEAM,
        min_damage = 2,
        max_damage = 4,
    },

    rib_cracker_plus2 =
    {
        name = "Boosted Rib Cracker",
        desc = "If your target has {TRAUMA}, apply <#UPGRADE>{1}</> {WOUND}.",
        wound_amount = 2,
        min_damage = 4,
        max_damage = 6,
    },

    knuckle_down =
    {
        name = "Knuckle Down",
        anim = "gut_shot",
        desc = "Attack once per {TRAUMA} on your target.",
        flavour = "'This will keep him busy.'",

        cost = 2,

        min_damage = 2,
        max_damage = 2,

        flags = CARD_FLAGS.MELEE | CARD_FLAGS.RESTRAINED,
        rarity = CARD_RARITY.UNCOMMON,

        OnPostResolve = function( self, battle, attack )
            self.hit_count = 1
        end,

        PreviewFn = function( self, damage_preview )
            local attack = Battle.Attack( self )

            if self.target and self.target:HasCondition("TRAUMA") then
                attack:GenerateHits( self.target:GetConditionStacks("TRAUMA") )
            else
                attack:GenerateHits( )
            end

            damage_preview:PreviewAttack( attack )
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.START_RESOLVE ] = function( self, battle, card )
                if card == self then
                    if card:GetTarget() and card:GetTarget():HasCondition("TRAUMA") then
                        self.hit_count = card:GetTarget():GetConditionStacks("TRAUMA")
                        self.hit_anim = true
                    else
                        self.hit_count = 1
                        self.hit_anim = false
                    end
                end
            end
        },
    },

    knuckle_down_plus = 
    {
        name = "Boosted Knuckle Down",
        min_damage = 3,
        max_damage = 3,
    },

    knuckle_down_plus2 = 
    {
        name = "Pale Knuckle Down",
        cost = 1,
    },

    rampart =
    {
        name = "Rampart",
        anim = "super_smash",
        desc = "Hits all enemies.\nGain {1} {DEFEND}.",
        desc_fn = function( self, fmt_str)
            return loc.format(fmt_str, self:CalculateDefendText(self.defend_amt, self.owner))
        end,
        flavour = "'I’ll take you all on! Bring it!'",

        cost = 2,

        flags = CARD_FLAGS.MELEE,
        rarity = CARD_RARITY.UNCOMMON,

        target_mod = TARGET_MOD.TEAM,

        min_damage = 3,
        max_damage = 5,

        defend_amt = 6,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("DEFEND", self.defend_amt, self)
        end
    },

    rampart_plus =
    {
        name = "Stone Rampart",
        desc = "Hits all enemies.\nGain <#UPGRADE>{1}</> {DEFEND}.",
        defend_amt = 10,
    },

    rampart_plus2 =
    {
        name = "Boosted Rampart",
        min_damage = 5,
        max_damage = 7,
    },

    thickness =
    {
        name = "Thickness",
        anim = "taunt",
        desc = "Double your {DEFEND}. At the start of your next turn, lose no {DEFEND}.",
        flavour = "'I unloaded my whole magazine into them, and the psycho just smiled! Who does that!?'",

        cost = 2,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.RARE,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            if self.defend_amt then
                self.owner:AddCondition("DEFEND", self.defend_amt, self)
            end
            if self.moxie_amt then
                local healing = self.owner:GetCondition("smith_regeneration")
                if healing then healing:IncreaseMax(self.moxie_amt, self) end
            end
            local count = self.owner:GetConditionStacks("DEFEND")
            if count and count > 0 then
                self.owner:AddCondition("DEFEND", count, self)
            end
            self.owner:AddCondition("thickness", 1, self)
        end,

        condition =
        {
            desc = "At the start of your next turn, lose no {DEFEND} and remove <b>Thickness</>.",
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id, self.stacks, self)
                end,

                [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                    if fighter == self.owner and condition_id == "DEFEND" and source == self.owner:GetCondition("DEFEND") then
                        if acc.value < 0 then
                            acc:ModifyValue(0, self)
                        end
                    end
                end,
            },
        },
    },

    thickness_plus =
    {
        name = "Stone Thickness",
        desc = "<#UPGRADE>Gain {1} {DEFEND}, then</> double your {DEFEND}. At the start of your next turn, lose no {DEFEND}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:CalculateDefendText(self.defend_amt))
        end,
        defend_amt = 5,
    },

    thickness_plus2=
    {
        name = "Regenerating Thickness",
        desc = "<#UPGRADE>Gain {1} {smith_regeneration}.</> Double your {DEFEND}. At the start of your next turn, lose no {DEFEND}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.moxie_amt)
        end,
        moxie_amt = 3,
    },

    the_rust_bucket = 
    {
        name = "The Rust Bucket",
        anim = "drink",
        desc = "{DRINK} {1} and gain {2} {DEFEND}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.drink_amt, self:CalculateDefendText(self.defend_amt))
        end,
        flavour = "’Tastes as old as an autodog and garnished with a ball bearing.’",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        drink_amt = 2,
        defend_amt = 5,

        OnPostResolve = function( self, battle, attack )
            Drink(self, battle, self.drink_amt)
            self.owner:AddCondition("DEFEND", self.defend_amt, self)
        end
    },

    the_rust_bucket_plus =
    {
        name = "Stone Rust Bucket",
        desc = "{DRINK} {1} and gain <#UPGRADE>{2} {DEFEND}</>.",
        defend_amt = 8,
    },

    the_rust_bucket_plus2 =
    {
        name = "Gulping Rust Bucket",
        desc = "<#UPGRADE>{DRINK} {1}</> and gain {2} {DEFEND}.",
        drink_amt = 3,
    },

    tough_and_angry =
    {
        name = "Tough and Angry",
        anim = "taunt",
        desc = "Apply {1} {DEFEND}. Increase by 1 for each {POWER} and {ADRENALINE}.",
        desc_fn = function( self, fmt_str )
            if is_instance( self.owner, Fighter ) then
                local bonus = self.owner:GetConditionStacks("POWER")
                bonus = bonus + self.owner:GetConditionStacks("ADRENALINE")
                bonus = bonus * (self.boost or 1)
                return loc.format(fmt_str, self:CalculateDefendText(self.defend_amt + bonus))
            else
                return loc.format(fmt_str, self.defend_amt)
            end
        end,
        flavour = "’Look, I’m not going in there. The guy looked crazy.’",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.RARE,

        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

        defend_amt = 2,
        boost = 1,

        OnPostResolve = function( self, battle, attack )
            local bonus = self.owner:GetConditionStacks("POWER")
            bonus = bonus + self.owner:GetConditionStacks("ADRENALINE")
                bonus = bonus * (self.boost or 1)
            attack:AddCondition("DEFEND", self.defend_amt + bonus, self)
        end,
    },

    tough_and_angry_plus =
    {
        name = "Stone Tough and Angry",
        desc = "Apply <#UPGRADE>{1} {DEFEND}</>. Increase by 1 for each {POWER} and {ADRENALINE}.",
        defend_amt = 6,
    },

    tough_and_angry_plus2 =
    {
        name = "Boosted Tough and Angry",
        desc = "Apply {1} {DEFEND}. Increase by <#UPGRADE>2</> for each {POWER} and {ADRENALINE}.",
        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
        boost = 2,
    },

    tracer =
    {
        name = "Tracer",
        anim = "taunt",
        desc = "Apply {1} {tracer}.",
        flavour = "'I’ve got your number now pal.'",

        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.tracer_amt)
        end,

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        tracer_amt = 2,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("tracer", self.tracer_amt, self)
        end,

        condition =
        {
            name = "Tracer",
            desc = "Whenever {2} is hit by an attack, you gain {1} {DEFEND}. At the start of {2}'s turn, remove 1 stack of <b>Tracer</>.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks, self:GetOwnerName())
            end,
            OnApply = function( self, battle )
                self.player = battle:GetPlayerFighter()
            end,

            ctype = CTYPE.DEBUFF,

            event_handlers =
            {
                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if not hit.evaded and hit.target == self.owner and attack.card and attack.card:IsAttackCard() and self.player then
                        self.player:AddCondition("DEFEND", self.stacks, self)
                    end
                end,

                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                end
            },
        },
    },

    tracer_plus =
    {
        name = "Boosted Tracer",
        desc = "Apply <#UPGRADE>{1} {tracer}</>.",
        tracer_amt = 4,
    },

    tracer_plus2 = 
    {
        name = "Stone Tracer",
        desc = "<#UPGRADE>Gain {2} {DEFEND}.</> Apply {1} {tracer}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.tracer_amt, self:CalculateDefendText(self.defend_amt))
        end,
        defend_amt = 5,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("DEFEND", self.defend_amt, self)
            attack:AddCondition("tracer", self.tracer_amt, self)
        end,
    },

    cavalcade =
    {
        name = "Cavalcade",
        desc = "Gain {DEFEND} equal to the damage dealt by your next attack.",
        flavour = "’You just hit them so hard that they all leave you alone, it’s science.’",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,
        cavalcade_amt = 1,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            if self.bonus then
                self.owner:AddCondition("DEFEND", self.bonus, self)
            end
            self.owner:AddCondition("cavalcade", self.cavalcade_amt, self)
        end,

        condition = 
        {
            desc = "Gain {DEFEND} equal to the damage dealt by your next <#HILITE>{1}</> {1*attack|attacks}.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,
            ctype = CTYPE.BUFF,
            event_handlers =
            {
                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if self.chosen_card and attack.card == self.chosen_card and attack.attacker == self.owner and (hit.damage or 0) > 0 then
                        self.owner:AddCondition("DEFEND", hit.damage, self)
                    end
                end,

                [ BATTLE_EVENT.START_RESOLVE ] = function( self, battle, card )
                    if card.owner == self.owner and card:IsAttackCard() then
                        self.chosen_card = card
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if self.chosen_card and self.chosen_card == card then
                        self.chosen_card = nil
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                end,
            },
        },
    },

    cavalcade_plus =
    {
        name = "Lasting Cavalcade",
        desc = "Gain {DEFEND} equal to the damage dealt by your next <#UPGRADE>two attacks</>.",
        cavalcade_amt = 2,
    },

    cavalcade_plus2 =
    {
        name = "Boosted Cavalcade",
        desc = "<#UPGRADE>Gain {DEFEND {1}}</>. Gain {DEFEND} equal to the damage dealt by your next attack.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateDefendText(self.bonus))
        end,
        bonus = 6,
    },

    rubble =
    {
        name = "Rubble",
        desc = "Discard up to {1} cards, gain {DEFEND {2}} per card discarded.",
        flavour = "'Why yes, I’ll gladly show you what’s inside the briefcase.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.discard_amt, self:CalculateDefendText(self.defend_amt))
        end,
        flavour = "’That loser shows up with three briefcases and I have NO idea what he’s going to do next.’",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.COMMON,
        target_type = TARGET_TYPE.SELF,

        discard_amt = 2,
        defend_amt = 3,

        OnPostResolve = function( self, battle, attack )
            if self.draw_count then
                battle:DrawCards(self.draw_count)
            end
            local cards = battle:DiscardCards(0, self.discard_amt, self)
            if #cards > 0 then
                for i,card in ipairs(cards) do
                    self.owner:AddCondition("DEFEND", self.defend_amt, self)
                end
            end
        end,
    },

    rubble_plus =
    {
        name = "Visionary Rubble",
        desc = "<#UPGRADE>Draw {2} cards,</> discard up to {1} cards, gain {DEFEND {3}} per card discarded.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.discard_amt, self.draw_count, self:CalculateDefendText(self.defend_amt))
        end,
        draw_count = 2,
    },

    rubble_plus2 =
    {
        name = "Stone Rubble",
        desc = "Discard up to {1} cards, gain <#UPGRADE>{DEFEND {2}}</> per card discarded.",
        defend_amt = 4,
    },

    callus =
    {
        name = "Callus",
        anim = "taunt",
        desc = "For the rest of your turn, gain {STORED_DEFENSE} equal to the damage dealt to yourself.",
        flavour = "'Kra’deshi biology can be very surprising when combined with drastic and illegal experimentation.'",

        cost = 1,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNCOMMON,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            if self.draw_count then
                battle:DrawCards(self.draw_count)
            end
            self.owner:AddCondition("callus", 1, self)
        end,

        condition =
        {
            desc = "For the rest of your turn, gain {STORED_DEFENSE} equal to{1* | {1} times} the damage dealt to yourself.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,
            ctype = CTYPE.BUFF,
            event_handlers =
            {
                [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                    if fighter == self.owner and delta > 0 and self.battle:GetFighterFromSource(source) == self.owner then
                        self.owner:AddCondition("STORED_DEFENSE", delta * self.stacks, self)
                    end
                end,

                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    callus_plus =
    {
        name = "Pale Callus",
        cost = 0,
    },

    callus_plus2 =
    {
        name = "Visionary Callus",
        desc = "<#UPGRADE>Draw 2 cards.</> For the rest of your turn, gain {STORED_DEFENSE} equal to the damage dealt to yourself.",
        draw_count = 2,
    },

    gumption =
    {
        name = "Gumption",
        desc = "Alleviate all of your {RECOVERABLE_HEALTH}.",
        flavour = "'Nothing beats a hard day’s work.'",

        cost = 1,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,

        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            local stacks = self.owner:GetConditionStacks("RECOVERABLE_HEALTH")
            if stacks > 0 then
                self.owner:AlleviateHealth(stacks, self)
            end
        end
    },

    gumption_plus =
    {
        name = "Pale Gumption",
        cost = 0,
    },

    gumption_plus2 =
    {
        name = "Enduring Gumption",
        flags = CARD_FLAGS.SKILL,
        cost = 2,
    },

    certainty =
    {
        name = "Certainty",
        anim = "taunt",
        desc = "{ABILITY}: Whenever you gain {DEFEND}, gain {ADRENALINE {1}}.",
        flavour = "'I’ve got this one in the bag.'",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.adrenaline_amt )
        end,

        cost = 2,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        adrenaline_amt = 2,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("certainty", self.adrenaline_amt, self)
        end,

        condition =
        {
            desc = "Whenever you gain {DEFEND}, gain {ADRENALINE {1}}.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.stacks )
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                    if fighter == self.owner and condition.id == "DEFEND" then
                        self.owner:AddCondition("ADRENALINE", self.stacks, self)
                    end
                end
            }
        },
    },

    certainty_plus =
    {
        name = "Pale Certainty",
        cost = 1,
    },

    certainty_plus2 = 
    {
        name = "Initial Certainty",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.AMBUSH,
    },

    protective_procedure =
    {
        name = "Protective Procedure",
        anim = "taunt",
        desc = "{ABILITY}: At the end of your turn, gain {DEFEND {1}}.",
        flavour = "'Wow I look better than ever! I can’t even see the scars!'",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.defend_amt )
        end,

        cost = 2,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        defend_amt = 3,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("protective_procedure", self.defend_amt, self)
        end,

        condition =
        {
            desc = "At the end of your turn, gain {DEFEND {1}}.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self:CalculateDefendText(self.stacks, self.owner) )
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:AddCondition("DEFEND", self.stacks, self)
                    end
                end
            }
        },
    },

    protective_procedure_plus =
    {
        name = "Pale Protective Procedure",
        cost = 1,
    },

    protective_procedure_plus2 =
    {
        name = "Initial Protective Procedure",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.AMBUSH,
    },

    soothing_concoctions = 
    {
        name = "Soothing Concoctions",
        desc = "{ABILITY}: Whenever you {DRINK}, {HEAL} {1} health.",
        flavour = "'Let me show you how to drink.'",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.heal_amt )
        end,

        cost = 3,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        heal_amt = 1,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("soothing_concoctions", self.heal_amt, self)
        end,

        condition =
        {
            desc = "Whenever you {DRINK}, {HEAL} {1} health.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self:CalculateDefendText(self.stacks) )
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.DRINK ] = function( self, battle, count, cards )
                    self.owner:HealHealth( self.stacks, self )
                end
            }
        },
    },

    soothing_concoctions_plus =
    {
        name = "Pale Soothing Concoctions",
        cost = 2,
    },

    soothing_concoctions_plus2 =
    {
        name = "Initial Soothing Concoctions",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.AMBUSH,
    },

    deal =
    {
        name = "Deal",
        anim = "bash",
        desc = "{THRESHOLD} {1}: Gain {ADRENALINE {2}}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self), self.adrenaline_amt)
        end,
        flavour = "'Hurry it up partner, I haven’t got all day.'",

        cost = 1,

        rarity = CARD_RARITY.COMMON,
        flags = CARD_FLAGS.MELEE,

        threshold = 3,
        adrenaline_amt = 2,

        min_damage = 2,
        max_damage = 4,

        ThresholdEffect = function( self, battle, attack )
            self.owner:AddCondition("ADRENALINE", self.adrenaline_amt, self)
        end
    },

    deal_plus =
    {
        name = "Tall Deal",
        max_damage = 6,
    },

    deal_plus2 =
    {
        name = "Enhanced Deal",
        desc = "{THRESHOLD} {1}: Gain <#UPGRADE>{ADRENALINE {2}}</>.",
        adrenaline_amt = 3,
    },

    high_stakes =
    {
        name = "High Stakes",
        anim = "taunt",
        desc = "{ABILITY}: Whenever you trigger a card's {THRESHOLD}, gain {DEFEND {1}}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.defend_amt)
        end,
        flavour = "'Looks like you’re at the wrong table, friend.'",

        cost = 2,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
        target_type = TARGET_TYPE.SELF,

        defend_amt = 2,
        OnPostResolve = function( self, battle, attack )
            attack:AddCondition("high_stakes", self.defend_amt, self)
        end,

        condition =
        {
            desc = "Whenever you trigger a card's {THRESHOLD}, gain {DEFEND {1}}.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self:CalculateDefendText( self.stacks ))
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.THRESHOLD_TRIGGERED ] = function( self, battle, card, target )
                    if card.owner == self.owner then
                        self.owner:AddCondition("DEFEND", self.stacks, self)
                    end
                end
            }
        },
    },

    high_stakes_plus =
    {
        name = "Pale High Stakes",
        cost = 1,
    },

    high_stakes_plus2 =
    {
        name = "Stone Stakes",
        desc = "{ABILITY}: Whenever you trigger a card's {THRESHOLD}, gain <#UPGRADE>{DEFEND {1}}.</>",
        defend_amt = 3,
    },

    the_gamble =
    {
        name = "The Gamble",
        anim = "bash",
        desc = "{THRESHOLD} {1}: Gain {2} actions.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self), self.action_count)
        end,
        flavour = "'This better work.'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.MELEE | CARD_FLAGS.RESTRAINED,

        min_damage = 1,
        max_damage = 4,

        threshold = 4,
        action_count = 3,

        ThresholdEffect = function( self, battle, attack )
            battle:ModifyActionCount(self.action_count)
            if self.draw_count then
                battle:DrawCards(self.draw_count)
            end
        end
    },

    the_gamble_plus =
    {
        name = "Tall Gamble",
        max_damage = 5,
    },

    the_gamble_plus2 =
    {
        name = "Visionary Gamble",
        desc = "{THRESHOLD} {1}: Gain {2} actions <#UPGRADE>and draw a card.</>",
        draw_count = 1,
    },

    hold_em =
    {
        name = "Hold 'em",
        anim = "bash",
        desc = "{THRESHOLD} {1}: Increase your {smith_regeneration} by {2}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self), self.moxie_amt)
        end,
        flavour = "'Know when to walk away.'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.MELEE,

        threshold = 4,
        moxie_amt = 4,

        min_damage = 2,
        max_damage = 4,

        ThresholdEffect = function( self, battle, attack )
            local moxie = self.owner:GetCondition("smith_regeneration")
            if moxie then 
                moxie:IncreaseMax(self.moxie_amt, self)
            end
        end
    },

    hold_em_plus =
    {
        name = "Tall Hold 'em",
        max_damage = 6,
    },

    hold_em_plus2 =
    {
        name = "Enhanced Hold 'em",
        desc = "{THRESHOLD} {1}: Increase your {smith_regeneration} by <#UPGRADE>{2}</>.",
        moxie_amt = 6,
    },

    fold_em =
    {
        name = "Fold 'em",
        anim = "smash",
        desc = "{THRESHOLD} {1}: Apply {2} {WOUND}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self), self.wound_amt)
        end,
        flavour = "'Know when to run.'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.MELEE,

        threshold = 6,
        wound_amt = 2,

        min_damage = 1,
        max_damage = 6,

        ThresholdEffect = function( self, battle, attack, hit )
            attack:AddCondition("WOUND", self.wound_amt, self, hit.target )
        end
    },

    fold_em_plus =
    {
        name = "Boosted Fold 'em",
        min_damage = 3,
        max_damage = 6,
    },

    fold_em_plus2 =
    {
        name = "Enhanced Fold 'em",
        desc = "{THRESHOLD} {1}: Apply <#UPGRADE>{2} {WOUND}</>.",
        wound_amt = 3,
    },

    raise =
    {
        name = "Raise",
        anim = "punch",
        desc = "{THRESHOLD} {1}: Reduce the cost of a random card in your hand by 1 until played.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self) )
        end,
        flavour = "'I’ll raise you this Heshing table!'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.MELEE,

        threshold = 5,

        min_damage = 2,
        max_damage = 6,
        card_count = 1,

        ThresholdEffect = function( self, battle, attack )
            local con = self.owner:AddCondition("raise", 1, self)
            if con then
                local cards = {}
                for i, card in battle:GetHandDeck():Cards() do
                    if battle:CalculateActionCost( card ) > 0 and not card:IsFlagged( CARD_FLAGS.VARIABLE_COST ) then
                        table.insert( cards, card )
                    end
                end
                if con.chosen_cards == nil then
                    con.chosen_cards = {}
                end
                table.arrayadd( con.chosen_cards, table.multipick( cards, self.card_count, false ))
            end
        end,

        condition = 
        {
            hidden = true,
            event_priorities =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST  ] = EVENT_PRIORITY_SETTOR,
            },
            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.chosen_cards and table.contains(self.chosen_cards, card) then
                        cost_acc:AddValue( -1, self )
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if table.contains(self.chosen_cards, card) then
                        table.arrayremoveall(self.chosen_cards, card)
                        if #self.chosen_cards == 0 then
                            self.owner:RemoveCondition(self.id, self.stacks, self)
                        end
                    end
                end,
            },
        },
    },

    raise_plus = 
    {
        name = "Boosted Raise",
        min_damage = 3,
        max_damage = 7,
    },

    raise_plus2 =
    {
        name = "Wide Raise",
        desc = "{THRESHOLD} {1}: Reduce the cost of <#UPGRADE>two random cards</> in your hand by 1 until played.",
        card_count = 2,
        min_damage = 1,
        max_damage = 5,
    },

    jackpot =
    {
        name = "Jackpot",
        anim = "back_hamm",
        desc = "{THRESHOLD} {1}: Apply 1 {STUN}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self) )
        end,
        flavour = "'DING DING DING!'",

        cost = 2,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.MELEE,

        threshold = 10,
        min_damage = 2,
        max_damage = 5,

        ThresholdEffect = function( self, battle, attack )
            attack:AddCondition("STUN", 1, self)
        end
    },

    jackpot_plus =
    {
        name = "Boosted Jackpot",
        min_damage = 4,
        max_damage = 6,
    },

    jackpot_plus2 =
    {
        name = "Pale Jackpot",
        cost = 1,
    },

    mean_streak =
    {
        name = "Mean Streak",
        anim = "bash",
        desc = "{THRESHOLD} {1}: Reduce all <b>Thresholds</> by 1 for the rest of this turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:CalculateThresholdText(self))
        end,
        flavour = "'I’ve made a living out of reading people’s faces.'",

        cost = 1,

        rarity = CARD_RARITY.RARE,
        flags = CARD_FLAGS.MELEE,

        threshold = 5,
        min_damage = 1,
        max_damage = 5,

        ThresholdEffect = function( self, battle, attack )
            self.owner:AddCondition("mean_streak", 1, self)
        end,

        condition =
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.CALC_THRESHOLD ] = function( self, threshold, card )
                    if card.owner == self.owner then
                        threshold:AddValue(-self.stacks, self)
                    end
                end,

                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:RemoveCondition(self.id, nil, self)
                    end
                end
            },
        },
    },

    mean_streak_plus =
    {
        name = "Tall Mean Streak",
        max_damage = 7,
    },

    mean_streak_plus2 =
    {
        name = "Boosted Mean Streak",
        min_damage = 3,
    },

    duck =
    {
        name = "Duck!",
        desc = "Gain 1 {EVASION}.",
        flavour = "'Where?'",

        cost = 1,

        rarity = CARD_RARITY.UNCOMMON,
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND | CARD_FLAGS.BURNOUT,
        target_type = TARGET_TYPE.SELF,

        OnPostResolve = function( self, battle, attack )
            self.owner:AddCondition("EVASION", 1, self)
        end
    },

    duck_plus =
    {
        name = "Stable Duck!",
        flags = CARD_FLAGS.SKILL | CARD_FLAGS.EXPEND,
    },

    duck_plus2 =
    {
        name = "Pale Duck!",
        cost = 0,
    },
}

for i, id, data in sorted_pairs(attacks) do
    data.series = "SMITH"

    Content.AddBattleCard( id, data )
end


local CONDITIONS =
{
    smith_regeneration =
    {
        name = "Moxie",
        desc = "Whenever you take damage on your turn, increase your <b>Moxie</> by a random amount between 1 and {4}. At the end of your turn, {HEAL} health equal to your <b>Moxie</> and then halve your <b>Moxie</> down to a minimum of {3}.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.min_heal, self.max_heal, self.base_max, self.passive_gain )
        end,
        hidden = true,

        apply_sound = "event:/sfx/battle/status/system/Status_Heal",

        min_heal = 0,
        max_heal = 1,

        base_max = 1,
        passive_gain = 2,

        GetRange = function( self )
            return loc.format("{1}-{2}", self.min_heal, self.max_heal )
        end,

        IncreaseMax = function( self, delta, source )
            self.max_heal = self.max_heal + delta
            self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
        end,

        ActivateHealing = function( self, source, ignore_reset )
            local heal = self.max_heal
            
            if heal > 0 and not self.owner:HasCondition("CORROSIVE_BILE") then
                self.owner.battle:BroadcastEvent( BATTLE_EVENT.CUSTOM, self.owner, function( fight_screen, ent, fighter )
                    local anim_fighter = ent.cmp.AnimFighter
                    local anim_controller = ent.cmp.AnimController
                    self.owner:HealHealth(heal, self)
                    if anim_fighter:IsIdle() then
                        anim_fighter:PushAnim(heal < 6 and "moxie_low" or "moxie_high")
                        anim_controller:CoroWaitForAnim()
                        anim_fighter:PlayIdleAnim()
                    end
                end )
            end
            if not ignore_reset then
                self:ResetMaximum(self)
            end
        end,

        ResetMaximum = function( self, source )
            self.max_heal = math.max(math.floor(self.max_heal * .5), self.base_max)
            self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                self:ActivateHealing(self)
            end,

            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                if not self.initialized then
                    self.battle:BroadcastEvent( BATTLE_EVENT.REFRESH_CONDITIONS, self.owner, self )
                    self.initialized = true
                end
            end,

            [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                if self.battle:GetCurrentTeam() == self.owner:GetTeam() and fighter == self.owner and stat == COMBAT_STAT.HEALTH and delta < 0 then
                    self:IncreaseMax(math.random(1, self.passive_gain), self)
                end
            end,
        },

        fighter_info_widget = function( fighter, condition, width )
            return Widget.MoxieInfoWidget( fighter, condition, width )
        end
    },

    RECOVERABLE_HEALTH = 
    {
        name = "Recoverable Health",
        desc = "{1} can {ALLEVIATE} this much health. This is increased by the amount of damage taken on your turn.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, loc.cap( self:GetOwnerName() ))
        end,

        hide_gained = true,

        event_priorities =
        {
            [ BATTLE_EVENT.END_TURN ] = 999,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition(self.id)
                end
            end,

            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition(self.id)
                end
            end,
        },
    },

    TRAUMA = 
    {
        name = "Trauma",
        desc = "Whenever {1} is attacked, increase <b>Trauma</> by 1. At 10 stacks, {1} gains {TRAUMATIZED} and loses <b>Trauma</>.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetOwnerName() )
        end,

        ctype = CTYPE.DEBUFF,
        max_stacks = 10,

        OnSetStacks = function( self )
            if self.stacks >= 10 then
                self.owner:RemoveCondition(self.id)
                local con = self.owner:AddCondition("TRAUMATIZED", 1, self)
                if con and self.most_recent then
                    con.ignore = self.most_recent
                end
            end
        end,

        event_priorities =
        {
            [ BATTLE_EVENT.ON_HIT ] = 11,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.ON_HIT] = function( self, battle, attack, hit )
                if hit.target == self.owner and not hit.evaded and attack.card and attack.card:IsAttackCard() then
                    self.most_recent = attack.card
                    self.owner:AddCondition(self.id, 1, self)
                end
            end,
        },
    }, 

    TRAUMATIZED = 
    {
        name = "Traumatized",
        desc = "The next attack against {1} deals double damage and reduces this by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetOwnerName() )
        end,

        ctype = CTYPE.DEBUFF, 

        event_priorities = 
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = EVENT_PRIORITY_MULTIPLIER,
        },
        event_handlers =
        {
            [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                if attack:IsTarget( self.owner ) and not attack.card:IsFlagged(CARD_FLAGS.RESTRAINED) and (self.ignore == nil or attack.card ~= self.ignore) and attack.card:IsAttackCard() then
                    if attack:IsDamageModifier( self ) and not attack:CheckHitResult( self.owner, "evaded" ) then
                        self.owner:RemoveCondition( self.id, 1 )
                    end
                end
            end,

            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt, hit )
                if target == self.owner then
                    local found = false
                    local attack_count = 0
                    for i, fighter in self.battle:ActiveFighters() do
                        if fighter.prepared_attacks and not found then
                            for i, attack in ipairs( fighter.prepared_attacks ) do
                                if attack:IsTarget( target ) then
                                    if attack.card == card then
                                        -- Ok, found this specific attack.
                                        found = true
                                        break
                                    elseif attack.card:IsDamageCard() then
                                        attack_count = attack_count + 1
                                    end
                                end
                            end
                        end
                    end

                    if not found or attack_count < self.stacks then
                        dmgt:AddDamage( dmgt.min_damage, dmgt.max_damage, self )
                    end
                end
            end,
        },
    },

    ADRENALINE =
    {
        name = "Adrenaline",
        desc = "Attack damage is increased by <#HILITE>{1}</>. At the end of your turn, remove all <b>Adrenaline</>.",
        desc_fn = function( self, fmt_str, battle )
            return loc.format(fmt_str, self.stacks )
        end,

        target_type = TARGET_TYPE.SELF,
        max_stacks = 99,
        min_stacks = -99,

        event_priorities =
        {
            [ BATTLE_EVENT.END_TURN  ] = 10,
        },

        event_handlers =
        {
            [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                if card.owner == self.owner and card:IsAttackCard() then
                    dmgt:ModifyDamage( dmgt.min_damage + self.stacks, dmgt.max_damage + self.stacks, self )
                end
            end,

            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:RemoveCondition(self.id)
                end
            end
        }
    },

    BOTTLE_TRACKER =
    {
        hidden = true,

        OnApply = function( self, battle )
            self.bottles = {}
            local cards = battle:GetCardsOwnedBy( self.owner )
            for i, card in ipairs( cards ) do
                if card.id == "empty_bottle" then
                    table.insert( self.bottles, card )
                end
            end
        end,

        GetBottleCount = function( self )
            return #self.bottles
        end,

        SpendBottles = function( self, count, source )
            local bottles = {}
            local all_cards = table.merge( self.battle:GetDrawDeck().cards, self.battle:GetDiscardDeck().cards, self.battle:GetHandDeck().cards )
            for i,card in ipairs(all_cards) do
                if card.id == "empty_bottle" then
                    table.insert(bottles, card)
                    if count and #bottles >= count then
                        break
                    end
                end
            end
            local total_count = #bottles
            for i,card in ipairs(bottles) do
                self.battle:ExpendCard(card)
                table.arrayremove(self.bottles, card)
            end

            self.battle:BroadcastEvent( BATTLE_EVENT.BOTTLES_SPENT, total_count, source)
            return total_count
        end,

        event_handlers =
        {
            [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                if (target_deck == nil or target_deck:GetDeckType() == DECK_TYPE.RESOLVE) and card.id == "empty_bottle" and table.contains(self.bottles, card) then
                    table.arrayremove(self.bottles, card)
                end
                if source_deck == nil and card.id == "empty_bottle" then
                    table.insert(self.bottles, card)
                end
            end,
        }
    },

    STORED_DEFENSE = 
    {
        name = "Stored Defense",
        desc = "At the end of your turn, gain {DEFEND} equal to <b>Stored Defense</> and remove all <b>Stored Defense</>.",

        event_handlers =
        {
            [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    self.owner:AddCondition("DEFEND", self.stacks, self)
                    self.owner:RemoveCondition(self.id, self.stacks, self)
                end
            end
        },
    },
}

for id, def in pairs( CONDITIONS ) do
    Content.AddBattleCondition( id, def )

end


local FEATURES =
{
    FOCUSED =
    {
        name = "Focused",
        desc = "This card has an additional effect if it is the only card in your hand.",
    },

    OPENER = 
    {
        name = "Opener",
        desc = "This card has an additional effect if it's the first card played this turn.",
    },

    DRINK =
    {
        name = "Drink",
        desc = "Add an {empty_bottle} card to your discard pile.",
    },

    BOTTLE =
    {
        name = "Bottle",
        desc = "This is a Bottle type card."
    },

    ALLEVIATE =
    {
        name = "Alleviate",
        desc = "Recover health up to the amount you have lost this turn.",
    },

    CHAIN =
    {
        name = "Chain",
        desc = "Transform this card and add it to your draw pile. If this goes into your discard pile, this card reverts to its basic form.",
    },
}

for id, data in pairs( FEATURES ) do
    local def = BattleFeatureDef(id, data)
    Content.AddBattleCardFeature(id, def)
end