require "content/notification_events"
require "aspect"
require "cards/card_set"
local battle_defs = require "battle/battle_defs"
local negotiation_def = require "negotiation/negotiation_defs"


local CardHolder = class( "Aspect.CardHolder", Aspect )

function CardHolder:init( agent )
    assert( self.CARD_CLASS )
    CardHolder._base.init( self, agent )
    self.cards = CardSet( self.CARD_CLASS, agent )
    self.paid_removals = 0
end

function CardHolder:AddDeck( deck )
    for i, card_id in deck:Cards() do
        local card = self:LearnCard( card_id )

        -- So that these cards don't get shown as added at End of Day.
        card.userdata.time_added = nil
    end
end

function CardHolder:AddCard( card, is_draft )
    self.cards:AddCard( card )
    assert_warning( card.owner == self.agent, "card '%s' is not owned by agent", tostring(card) )
    card.owner = self.agent

    if self.agent:IsPlayer() then
        card.userdata.time_added = Now()

        StatUtil.AddedCard(card.id)
        card:ListenForGlobalEvents()
        TheGame:BroadcastEvent( "card_added", card, is_draft )
    end
end

function CardHolder:LearnCard( card_id, userdata )
    local card = self.CARD_CLASS( card_id, self.agent, userdata )
    self:AddCard( card )
    return card
end

function CardHolder:GetRemovePrice()
    return REMOVE_CARD_PRICE + (REMOVE_CARD_RAMP * self.paid_removals)
end

function CardHolder:IncPaidRemovals()
    self.paid_removals = self.paid_removals + 1
end


function CardHolder:ClearCards()
    for i, card in ipairs( self.cards.cards ) do
        card:RemoveGlobalListener()
    end
    self.cards:ClearCards()
    return self
end

function CardHolder:GetNumCardsReadyForUpgrade()
    local num = 0
    for k,card in ipairs(self.cards.cards) do
        if card:UpgradeReady() and not card:IsHatched() then
            num = num + 1
        end
    end
    return num
end

function CardHolder:GetCardsReadyForUpgrade(t)
    t = t or {}
    local num = 0
    for k,card in ipairs(self.cards.cards) do
        if card:UpgradeReady() and not card:IsHatched() then
            table.insert(t, card)
        end
    end
    return t
end


function CardHolder:UpgradeCard( card, new_id )
    assert_warning( table.contains( self.cards.cards, card ), tostring(card))

    card:UpgradeCard( new_id )

    print( string.format( "UpgradeCard( %s, %s )", tostring(card), tostring(new_id) ))

    -- don't show these on end of day screen.
    --card.userdata.time_added = Now()
    
    if self.agent:IsPlayer() then
        TheGame:GetGameState():AddScore("CARDS_UPGRADED", 1)
        TheGame:BroadcastEvent( "deck_upgrade_status_change")
        TheGame:AddGameplayStat( "cards_upgraded", 1 )

        if card.base_def then
            StatUtil.UpgradedCard(card.base_def.id)
        end
        StatUtil.AddedCard(card.id)
    end

    return card
end

function CardHolder:HasCard( card )
    for k, v in ipairs(self.cards.cards) do
        if v == card then
            return true
        end
    end
    return false
end

function CardHolder:HasCardID( id )
    for k, v in ipairs(self.cards.cards) do
        if v.id == id then
            return true
        end
    end
    return false
end

function CardHolder:FindCardByID( id, userdata )
    for k, card in ipairs(self.cards.cards) do
        if card.id == id and (userdata == nil or userdata == card.userdata) then
            return card
        end
    end
end
function CardHolder:FindCard( fn )
    for k,card in ipairs(self.cards.cards) do
        if fn(card) then
            return card
        end
    end
end

function CardHolder:RemoveCard( card )
    
    print( string.format( "RemoveCard( %s )", tostring(card)))

    self.cards:RemoveCard( card )

    card:RemoveGlobalListener()

    if self.agent:IsInPlayerParty() then
        if card:UpgradeReady() then
            TheGame:BroadcastEvent( "deck_upgrade_status_change")
        end
        TheGame:BroadcastEvent( "card_removed", card )
    end

end

function CardHolder:RemoveCardByID( card_id, userdata )
    local card = self:FindCardByID( card_id, userdata )
    if card then
        self:RemoveCard( card )
    end
end

function CardHolder:SetCards(card_ids)
    self.cards:ClearCards()
    for k, v in ipairs(card_ids) do
        local card = self:LearnCard( v )
        card.userdata.time_added = nil
    end
end

function CardHolder:SetCardCount( card_id, count )
    local current = self:GetCardCount( card_id )
    if current < count then
        for i = 1, count - current do
            self:LearnCard( card_id )
        end

    elseif count < current then
        for i = 1, current - count do
            self:RemoveCardByID( card_id, nil )
        end
    end
end

function CardHolder:GetCardCount( card_id )
    if card_id == nil then
        return #self.cards:GetCards()
    else
        local count = 0
        for i, card in ipairs( self.cards:GetCards() ) do
            if card.id == card_id then
                count = count + 1
            end
        end
        return count
    end
end

function CardHolder:GetCards()
    return self.cards:GetCards()
end

function CardHolder:IsFullyUpgraded()
    for i, card in ipairs( self.cards:GetCards() ) do
        if card:HasUpgrades() and not card:IsHatched() and card:GetXP() ~= nil then
            return false
        end
    end
    return true
end

function CardHolder:PostLoad( game_state )
    -- Delete any cards that no longer exist.
    self.cards:ClearInvalidCards()

    -- Sanitize existing cards.
    for i, card in ipairs( self.cards:GetCards() ) do
        card:PostLoad( game_state )
        card:ListenForGlobalEvents()
    end
end

---------------------------------------------------------------------
-- Can hold Battle.Cards

local Battler = class( "Aspect.Battler", Aspect.CardHolder )
Battler.CARD_CLASS = Battle.Card
Battler.DATA_NAME = "fight_data"

function Battler:RemoveCard( card )
    Battler._base.RemoveCard(self, card)
    if CheckBits( card.flags, battle_defs.CARD_FLAGS.PARASITE ) and not card.userdata.removed_safely then
        if self.agent.health then
            self.agent.health:AddStatModifier( "PARASITE_REMOVAL", -PARASITE_HEALTH_LOSS )
        end
    end
end
Content.AddAspect( "battler", Battler )

---------------------------------------------------------------------
-- Can hold Negotiation.Cards

local Negotiator = class( "Aspect.Negotiator", Aspect.CardHolder )
Negotiator.CARD_CLASS = Negotiation.Card
Negotiator.DATA_NAME = "negotiation_data"

function Negotiator:RemoveCard( card )
    Negotiator._base.RemoveCard(self, card)
    if CheckBits( card.flags, negotiation_def.CARD_FLAGS.PARASITE ) and not card.userdata.removed_safely then
        if self.agent.health then
            self.agent.health:AddStatModifier( "PARASITE_REMOVAL", -PARASITE_HEALTH_LOSS )
        end
    end
end

Content.AddAspect( "negotiator", Negotiator )
