require "aspect"
require "eventsystem"
local memtokens = require "memtokens"

local GraftInstance = class("GraftInstance")
function GraftInstance:init( id, userdata, owner )
    local data = Content.GetGraft( id )
    if data == nil then
        LOGWARN( "Unknown Graft: %s. Converting to robustify", id )
        data = Content.GetGraft( "robustify" )
        id = "robustify"
    end

    self.id = id
    self.userdata = userdata or {}
    self.owner = owner or TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    local def = self:GetDef()
    self.userdata.upgrade_ids = def.upgrade_ids
    if not self.userdata.max_xp then
        self.userdata.max_xp = def.max_xp
        TheGame:BroadcastEvent( "roll_graft_upgrades", self )
    end
end

function GraftInstance:IsHidden()
    return self:GetDef().hidden
end

function GraftInstance:GetSummaryData()
    local data = { id = self.id, userdata = {} }
    for k, v in pairs(self.userdata) do
        if type(v) ~= "table" or not v._class then
            data.userdata[k] = v
        end
    end
    return data
end

function GraftInstance:IsRemovable()
    if self:GetDef().unremovable then
        return false
    end
    
    if self.userdata.unremovable then
        return false
    end

    return true
end

function GraftInstance:SetUnremovable(val)
    self.userdata.unremovable = val
end


function GraftInstance:Save()
    return self
end

function GraftInstance:GetRarity()
    return self:GetDef().rarity
end

function GraftInstance:GetDef()
    return Content.GetGraft(self.id)
end

function GraftInstance:OnUpdated()
    if self:GetDef().OnUpdated then
        self:GetDef().OnUpdated(self)
    end

    if self.owner then
        self.owner:BroadcastEvent("graft_changed", self)
    end
end

function GraftInstance:GetImage()
    local img = self:GetDef().img
    if type(img) == "function" then
        return img(self)
    end
    return img
end

function GraftInstance:GetType()
    return self:GetDef().type
end

function GraftInstance:GetName()
    local graftdef = self:GetDef()
    assert( graftdef, tostring(self.id) )

    local name    
    if graftdef.name then
        name = graftdef:GetLocalizedName()
    end

    return name or tostring(self.id)
end

function GraftInstance:GetDesc( widget )
    local graftdef = self:GetDef()
    local desc = self.id

    if graftdef.desc then
		desc = graftdef:GetLocalizedDesc()
        if graftdef.desc_fn then
			desc = graftdef.desc_fn(self, desc)
        end
    end

    desc = CardEngine.ExtractKeywords( desc, widget, nil, graftdef.card_defs )

    return desc
end

function GraftInstance:GetXPDesc()
    return self:GetDef().xp_desc or self:GetType() == GRAFT_TYPE.COMBAT and LOC"UI.BATTLE_GRAFT_XP" or LOC"UI.NEGOTIATION_GRAFT_XP"
end

function GraftInstance:GetDisplayNumber()
    local graftdef = self:GetDef()
    if graftdef.display_number then
        return graftdef.display_number(self)
    end
end

function GraftInstance:GetProgressText()
    local graftdef = self:GetDef()   
    if graftdef.GetProgressText then
        return graftdef.GetProgressText(self)
    end
end

function GraftInstance:IncrementCounter()
    self.userdata.counter = (self.userdata.counter or 0) + 1
end

function GraftInstance:ResetCounter()
    self.userdata.counter = 0
end

function GraftInstance:GetCounter()
    return self.userdata.counter or 0
end

function GraftInstance:HasMemory(token, duration)
    return memtokens.HasMemory( self.userdata, token, duration)
end

function GraftInstance:Remember(token, val)
    return memtokens.Remember( self.userdata, token, val )
end

function GraftInstance:Forget(token)
    return memtokens.Forget( self.userdata, token )
end

function GraftInstance:IsNegotiationGraft()
    local def = self:GetDef()
    if def.type ~= GRAFT_TYPE.MUTATOR then
        if def.OnStartNegotiation then
            return true
        end
        local modifier_id = def.negotiation_modifier
        if modifier_id == nil and def.base_id then
            modifier_id = Content.GetGraft( def.base_id ).negotiation_modifier
        end
        if modifier_id then
            return true
        end
    end

    return false
end

function GraftInstance:IsBattleGraft()
    local def = self:GetDef()
    if def.type ~= GRAFT_TYPE.MUTATOR then
        if def.OnActivateFighter then
            return true
        end
        local condition_id = def.battle_condition
        if condition_id == nil and def.base_id then
            condition_id = Content.GetGraft( def.base_id ).battle_condition
        end
        if condition_id then
            return true
        end
    end
    return false
end

function GraftInstance:OnStartNegotiation(minigame)
    if self:GetDef().OnStartNegotiation then
        self:GetDef().OnStartNegotiation(self, minigame)
    else
        local modifier_id = self:GetDef().negotiation_modifier
        if modifier_id == nil and self:GetDef().base_id then
            modifier_id = Content.GetGraft( self:GetDef().base_id ).negotiation_modifier
        end
        if modifier_id then
            self:AddNegotiationModifier( minigame )
        end
    end
end

function GraftInstance:AddBattleCondition(fighter)
    local stacks = self:GetDef().num_stacks or 1
    local condition_id = self:GetDef().battle_condition or Content.GetGraft(self:GetDef().base_id).battle_condition
    local condition = fighter:AddCondition( condition_id, stacks)
    if condition then
        condition.graft = self
    end
    return condition
end

function GraftInstance:AddNegotiationModifier(minigame)
    local stacks = self:GetDef().num_stacks or 1
    local modifier_id = self:GetDef().negotiation_modifier or Content.GetGraft(self:GetDef().base_id).negotiation_modifier
    local modifier = minigame:GetNegotiator( self.owner ):AddModifier( modifier_id, stacks )
    if modifier then
        modifier.graft = self
    end
    return modifier
end


function GraftInstance:OnActivateFighter(fighter)
    if self:GetDef().OnActivateFighter then
        self:GetDef().OnActivateFighter( self,  fighter )
    else
        local condition_id = self:GetDef().battle_condition
        if condition_id == nil and self:GetDef().base_id then
            condition_id = Content.GetGraft( self:GetDef().base_id ).battle_condition
        end
        if condition_id then
            self:AddBattleCondition( fighter )
        end
    end
end

function GraftInstance:AddXP(xp)
    if xp and self:HasUpgrades() then
        self.userdata.xp = math.min( (self.userdata.xp or 0) + xp, (self.userdata.max_xp or self:GetDef().max_xp))
    end
end

function GraftInstance:GetXP()
    return (self.userdata.xp or 0), (self.userdata.max_xp or self:GetDef().max_xp)
end

function GraftInstance:UpgradeReady()
    if self.userdata.xp and self:HasUpgrades() then
        return (self.userdata.xp or 0) >= (self.userdata.max_xp or self:GetDef().max_xp)
    else
        return false
    end
end

function GraftInstance:HasUpgrades()
    return (self.userdata.max_xp or self:GetDef().max_xp or 0) > 0 and self.userdata.upgrade_ids and #self.userdata.upgrade_ids > 0
end

function GraftInstance:Upgrade()
    self.id = self.userdata.upgrade_ids[1]
    assert( self.id )

    self.userdata =
    {
        time_added = self.userdata.time_added,
        upgrade_ids = self:GetDef().upgrade_ids
    }

    if self:GetDef().OnUpgraded then
        self:GetDef().OnUpgraded( self, self.owner )
    end

    self:OnUpdated()
    
    if TheGame:GetGameState() then
        TheGame:GetGameState():LogNotification( NOTIFY.GRAFT_UPGRADED, self.owner, self)
    end
end

function GraftInstance:SetListeners()
    local def = self:GetDef()
    local handlers = def.event_handlers
    if handlers then
        for event_name, fn in pairs( handlers ) do
            local priority = def.event_priorities and def.event_priorities[ event_name ]
            TheGame:GetEvents():ListenForEvent( event_name, self, self.OnGlobalEvent, priority )
        end
    end
end

function GraftInstance:RemoveListeners()
    local handlers = self:GetDef().event_handlers
    if handlers then
        for event_name, fn in pairs( handlers ) do
            TheGame:GetEvents():StopListening( event_name, self )
        end
    end
end

function GraftInstance:OnGlobalEvent( event_name, ... )
    local def = self:GetDef()

    if def.event_handlers ~= nil and def.event_handlers[ event_name ] ~= nil then
        def.event_handlers[ event_name ]( self, ... )
    end
end


function GraftInstance:GetFlavourText()
    local graftdef = self:GetDef()
    local flavour = ""
    
    if graftdef.flavour then
        flavour = graftdef:GetLocalizedFlavour()
    end

    return flavour
end

function GraftInstance:__tostring()
    return string.format( "[%s]", self.id )
end

-------------------------------------------------------------------------

local GraftOwner = class( "Aspect.GraftOwner", Aspect )

function GraftOwner:init( agent )
    GraftOwner._base.init( self, agent )
    self.grafts = {}
    self.max_grafts = shallowcopy( self.agent.def.max_grafts )
    
    local advancement_mod_combat = GetAdvancementModifier( ADVANCEMENT_OPTION.FEWER_COMBAT_GRAFT_SLOTS )
    if advancement_mod_combat then
        self.max_grafts[GRAFT_TYPE.COMBAT] = self.max_grafts[GRAFT_TYPE.COMBAT] - advancement_mod_combat
    end
    local advancement_mod_negotiation = GetAdvancementModifier( ADVANCEMENT_OPTION.FEWER_NEGOTIATION_GRAFT_SLOTS )
    if advancement_mod_negotiation then
        self.max_grafts[GRAFT_TYPE.NEGOTIATION] = self.max_grafts[GRAFT_TYPE.NEGOTIATION] - advancement_mod_negotiation
    end

    if agent.starting_grafts and TheGame:GetGameState() and (agent == TheGame:GetGameState():GetPlayerAgent() or TheGame:GetGameState():GetPlayerAgent() == nil ) then
        for i,graft_id in ipairs(agent.starting_grafts) do
            self:AddGraft(GraftInstance(graft_id))
        end
    end
end

function GraftOwner:ClearGrafts()
    for i = #self.grafts, 1, -1 do
        self:RemoveGraft( self.grafts[ #self.grafts ] )
    end
end

function GraftOwner:CanInstallGraft(graft_id)
    local def = Content.GetGraft(graft_id)
    if def == nil then
        return false, "no_def"
    end

    local max_grafts = self:GetMaxGrafts( def.type )
    if max_grafts then
        local num = 0
        for k,v in pairs(self.grafts) do
            if v:GetType() == def.type then
                num = num + 1  
            end
         end
         if num >= max_grafts then
            return false, "too many"
        end
    end

    return true
end

function GraftOwner:IsFullyUpgraded()
    for k,v in ipairs(self.grafts) do
        if v:HasUpgrades() then
            return false
        end
    end
    return true
end

function GraftOwner:GetMaxGrafts(graft_type)
    return self.max_grafts[graft_type] or 0
end

function GraftOwner:IncreaseMaxGrafts(graft_type, count)
    count = count or 1
    self.max_grafts[graft_type] = self.max_grafts[graft_type] + count

    if self.agent:IsPlayer() then
        self.agent:BroadcastEvent("graft_slot_added")
        if TheGame:GetGameState() then
            TheGame:BroadcastEvent("graft_slot_added")
            TheGame:GetGameState():LogNotification( NOTIFY.GRAFT_SLOT_UPGRADED, graft_type )
        end
    end
end


function GraftOwner:GetGrafts(graft_type, t)
    t = t or {}
    for k,v in ipairs(self.grafts) do
        if graft_type == nil or v:GetType() == graft_type then
            table.insert(t, v)
        end
    end

    return t
end

function GraftOwner:GetAllGrafts()
    return self.grafts
end

function GraftOwner:GetGraft(id)
    for k,v in ipairs(self.grafts) do
        if v.id == id then
            return v
        end
    end
    return nil
end

-- Also checks for upgraded grafts
function GraftOwner:HasGraft( id )
    for k,v in ipairs(self.grafts) do
        if v.id == id then
            return true -- Exact match.
        else
            local def = v:GetDef()
            if def.base_id == id then
                -- We have an upgraded version of id.
                return true
            elseif def.upgrade_ids then
                for i, upgrade_id in ipairs( def.upgrade_ids ) do
                    if upgrade_id == id then
                        -- We have a base version of the upgraded graft 'id'.
                        return true
                    end
                end
            end
        end
    end
    return false
end

function GraftOwner:FindGraft( fn )
    for k,v in ipairs(self.grafts) do
        if fn(v) then
            return v
        end
    end
    return nil
end

function GraftOwner:CountGraftsByType(graft_type)
    local count = 0
    for k,v in ipairs(self.grafts) do
        if graft_type == nil or v:GetType() == graft_type then
            count = count + 1
        end
    end

    return count
end
function GraftOwner:CountGraftsByID( id )
    local count = 0
    for k,v in ipairs(self.grafts) do
        if v.id == id then
            count = count + 1
       end
    end
    return count
end

function GraftOwner:GetTotalGraftsCount()
    local installed_grafts = 0
    installed_grafts = installed_grafts + #self:GetGrafts( GRAFT_TYPE.COMBAT )
    installed_grafts = installed_grafts + #self:GetGrafts( GRAFT_TYPE.NEGOTIATION )
    
    local max_grafts = 0
    max_grafts = max_grafts + self:GetMaxGrafts( GRAFT_TYPE.COMBAT )
    max_grafts = max_grafts + self:GetMaxGrafts( GRAFT_TYPE.NEGOTIATION )

    return installed_grafts, max_grafts
end

function GraftOwner:ReplaceGraft(old_graft, new_graft)
    local idx = table.arrayfind( self.grafts, old_graft )
    assert( idx, tostring(old_graft)) -- old_graft doesn't exist, something's up.

    self:RemoveGraft( old_graft )
    self:AddGraft( new_graft, idx )
end

function GraftOwner:RemoveGraft(graft)
    assert( is_instance( graft, GraftInstance ))
    for k,v in ipairs(self.grafts) do
        if v == graft then
            print( "RemoveGraft:", self.agent, graft )    
            table.remove( self.grafts, k )
            if graft:GetDef().OnRemoved then
                graft:GetDef().OnRemoved(graft, self.agent)
            end
            graft:RemoveListeners()
            self.agent:BroadcastEvent("graft_removed", graft)
            if TheGame:GetGameState() then
                TheGame:GetGameState():LogNotification( NOTIFY.GRAFT_REMOVED, self.agent, graft)
            end
            return
        end
    end

end

function GraftOwner:AddGraft(graft, idx)
    print( "InstallGraft:", self.agent, graft )    
    graft.owner = self.agent
    table.insert( self.grafts, idx or (#self.grafts + 1), graft)
    
    graft:SetListeners()
    graft.userdata.time_added = Now()
    self.agent:BroadcastEvent("graft_added", graft)
    TheGame:BroadcastEvent("graft_added", graft)
    if graft:GetType() == GRAFT_TYPE.COMBAT or graft:GetType() == GRAFT_TYPE.NEGOTIATION then
        StatUtil.InstallGraft( graft.id )
        TheGame:AddGameplayStat( "grafts_installed", 1 )
    end

    if TheGame:GetGameState() and not graft:GetDef().suppressed then
        TheGame:GetGameState():LogNotification( NOTIFY.GRAFT_ADDED, self.agent, graft)
    end

    TheGame:GetGameProfile():SetInstalledGraft( graft.id )

    if graft:GetDef().OnAdded then
        graft:GetDef().OnAdded(graft, self.agent)
    end

    return graft
end

function GraftOwner:GetGraftsReadyForUpgrade()
    local ret = {}
    for k,graft in ipairs(self:GetGrafts()) do
        if graft:UpgradeReady() then
            table.insert(ret, graft)
        end
    end
    return ret
end

function GraftOwner:GetBattleGraftsReadyForUpgrade()
    local ret = {}
    for k, graft in ipairs(self:GetGrafts()) do
        if graft:UpgradeReady() and graft:GetType() == GRAFT_TYPE.COMBAT then
            table.insert(ret, graft)
        end
    end
    return ret
end

function GraftOwner:GetNegotiationGraftsReadyForUpgrade()
    local ret = {}
    for k, graft in ipairs(self:GetGrafts()) do
        if graft:UpgradeReady() and graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            table.insert(ret, graft)
        end
    end
    return ret
end

function GraftOwner:OnActivateFighter(fighter)
    for k,graft in ipairs(self.grafts) do
        graft:OnActivateFighter(fighter)
    end
end

function GraftOwner:OnStartNegotiation(minigame)
    for k,graft in ipairs(self.grafts) do
        graft:OnStartNegotiation(minigame)
    end
end

function GraftOwner:PostLoad( game_state )
    self:VerifyGrafts(game_state)
    self:VerifySocialGrafts(game_state)
    self:VerifyRequiredGrafts(game_state)
    for i,graft in ipairs(self.grafts) do
        graft:SetListeners()
    end
end

function GraftOwner:GetSocialGraft(agent)
    for k,v in ipairs(self.grafts) do
        if v:GetType() == GRAFT_TYPE.SOCIAL and v.userdata and v.userdata.agents and table.arrayfind(v.userdata.agents, agent) then
            return v
        end
    end
end

function GraftOwner:AddSocialGraft(agent, rel)
    local graft_id = agent:GetSocialGraft(rel)
    if graft_id then
        return self:AddGraft( GraftInstance(graft_id, {agents = {agent}}, self.agent) )
    end
end

function GraftOwner:RemoveSocialGraft(agent)
    for k,v in ipairs(self.grafts) do
        if v:GetType() == GRAFT_TYPE.SOCIAL and v.userdata and v.userdata.agents and table.arrayfind(v.userdata.agents, agent) then
            table.arrayremove(v.userdata.agents, agent)
            if #v.userdata.agents == 0 then
                self:RemoveGraft(v)
            else
                v:OnUpdated()
            end
            return
        end
    end
end

function GraftOwner:VerifySocialGrafts(game_state)
    local player = game_state.player_agent
    
    --Replace existing grafts that are wrong
    for k,v in ipairs(self.grafts) do
        if v:GetType() == GRAFT_TYPE.SOCIAL then 
            if v.userdata and v.userdata.agents then
                local agent = v.userdata.agents[1]
                local rel = agent:GetRelationship(player)
                if v.id ~= agent:GetSocialGraft(rel) then
                    print( "SWAPPING SOCIAL GRAFT:", v.id, agent:GetSocialGraft(rel))
                    self:RemoveGraft(v)
                    self:AddSocialGraft(agent, rel)
                end
            end            
        end
    end
end

function GraftOwner:VerifyGrafts(game_state)
    local player = game_state.player_agent
    
    --Remove grafts that no longer exist
    for i = #self.grafts, 1, -1 do
        local graft = self.grafts[i]
        if graft:GetDef() == nil then
            print( "REMOVING MISSING GRAFT:", graft.id )
            table.remove( self.grafts, i )
        end
    end

end

function GraftOwner:VerifyRequiredGrafts(game_state)
    local player = game_state.player_agent
    if player.def.required_grafts then
        for i,id in ipairs(player.def.required_grafts) do
            if not self:HasGraft(id) then
                self:AddGraft(GraftInstance( id, {}, player ) )
            end
        end
    end
end

function GraftOwner:SetGrafts(graft_ids, ignore_starting)
    
    for k,graft in ipairs(self.grafts) do
        if graft:GetDef().OnRemoved then
            graft:GetDef().OnRemoved(graft, self.agent)
        end
        graft:RemoveListeners()
    end
    
    table.clear(self.grafts)
    
    local num_battle, num_negotiation = 0, 0
    for k, v in ipairs(graft_ids) do
        local graft = GraftInstance(v, {}, self.agent)
        
        if graft:GetType() == GRAFT_TYPE.COMBAT then
            num_battle = num_battle + 1
        elseif graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            num_negotiation = num_negotiation + 1
        end
        
        if graft:GetType() ~= GRAFT_TYPE.METTLE_UPGRADE then
            self.max_grafts[GRAFT_TYPE.COMBAT] = math.max(self.max_grafts[GRAFT_TYPE.COMBAT], num_battle)
            self.max_grafts[GRAFT_TYPE.NEGOTIATION] = math.max( self.max_grafts[GRAFT_TYPE.NEGOTIATION], num_negotiation)


            self:AddGraft( graft )
        end
    end
    if not ignore_starting then
        if self.agent.starting_grafts and TheGame:GetGameState() and (self.agent == TheGame:GetGameState():GetPlayerAgent() or TheGame:GetGameState():GetPlayerAgent() == nil ) then
            for i,graft_id in ipairs(self.agent.starting_grafts) do
                if not self:HasGraft(graft_id) then
                    self:AddGraft(GraftInstance(graft_id))
                end
            end
        end
    end
end

Content.AddAspect( "graft_owner", GraftOwner )