require "aspect"
------------------------------------------------------------------------
-- SocialConnections

local SocialConnections = class("Aspects.SocialConnections", Aspect)

SocialConnections.name = "SocialConnections"
SocialConnections.keep_on_retire = true

function SocialConnections:init( agent )
    SocialConnections._base.init(self, agent)
    self.relationships = {}
    self.history = {}
end       

function SocialConnections:GetNumberOfConnections(rel)
    local num = 0
    for agent, arel in pairs(self.relationships) do
        if not agent:IsPlayer() and not agent:IsDead() then
            if not rel or rel == arel then
                num = num + 1
            end
        end
    end
    return num
end

function SocialConnections:GetConnectionsOfLevel(rel)
    local agents = {}

    for agent, arel in pairs(self.relationships) do
        if not agent:IsPlayer() then
            if not rel or rel == arel then
                table.insert( agents, agent )
            end
        end
    end

    return agents
end

function SocialConnections:CountConnections( fn )
    local count = 0
    for agent, rel in pairs( self.relationships ) do
        if fn( agent, rel ) then
            count = count + 1
        end
    end
    
    return count
end


function SocialConnections:UpdateAspect()
  
end

function SocialConnections:GetDesc(gamestate, char )
    return "Has connections to the world"
end

function SocialConnections:GetRelationship(other_agent)
    if other_agent then
        if self.relationships[other_agent] then
            return self.relationships[other_agent]
        end
    end

    return RELATIONSHIP.NEUTRAL
end

local REL_STRING_LOOKUP = 
{
    HATED = RELATIONSHIP.HATED,
    LOVED = RELATIONSHIP.LOVED,
    LIKED = RELATIONSHIP.LIKED,
    DISLIKED = RELATIONSHIP.DISLIKED,
    NEUTRAL = RELATIONSHIP.NEUTRAL,
}

function SocialConnections:SetRelationship(rel, other_agent)
    if type(rel) == "string" then
        rel = REL_STRING_LOOKUP[rel]
        assert(rel, "BAD RELATIONSHIP STRING")
    end


    local old_rel = self.relationships[other_agent] or RELATIONSHIP.NEUTRAL

    self.relationships[other_agent] = rel
    if other_agent.social_connections then
        other_agent.social_connections.relationships[self.agent] = rel
    end


    if old_rel ~= rel then
        if other_agent:IsPlayer() then

            self.agent:BroadcastEvent("relationship_changed", self, self:GetRelationship(other_agent))
            if old_rel == RELATIONSHIP.LOVED or old_rel == RELATIONSHIP.HATED then
                TheGame:GetGameState():GetPlayerAgent().graft_owner:RemoveSocialGraft(self.agent)
            end
            if rel == RELATIONSHIP.LOVED or rel == RELATIONSHIP.HATED then
                TheGame:GetGameState():GetPlayerAgent().graft_owner:AddSocialGraft(self.agent, rel)
            end
            
            TheGame:GetEvents():BroadcastEvent( "agent_relationship_changed", self.agent, old_rel, rel )
            if rel == RELATIONSHIP.LOVED then
                TheGame:GetGameProfile():SetLovedByAgent( self.agent:GetUniqueID() )
                TheGame:GetGameProfile():SetUnlockedBoon( self.agent:GetUniqueID() )
            elseif rel == RELATIONSHIP.HATED then
                TheGame:GetGameProfile():SetHatedByAgent( self.agent:GetUniqueID() )
                TheGame:GetGameProfile():SetUnlockedBane( self.agent:GetUniqueID() )
            end
        end
    end
end


function SocialConnections:GetRandomRelation()
    local candidates = ObtainWorkTable()
    for other_agent, reldat in self:Relationships() do
        candidates[other_agent] = 1
    end
    local target = weightedpick(candidates)
    ReleaseWorkTable(candidates)    
    return target
end

function SocialConnections:Relationships()
    return pairs( self.relationships )
end

function SocialConnections:IsFriend(other)
    return self.relationships[other] and self.relationships[other] > RELATIONSHIP.NEUTRAL 
end

function SocialConnections:IsEnemy(other)
    return self.relationships[other] and self.relationships[other] < RELATIONSHIP.NEUTRAL 
end

function SocialConnections:GetWorstEnemy(filter_fn, ...)
    local worst, worst_op
    for agent,rel in pairs(self.relationships) do
        if not agent:IsRetired() then
            if not filter_fn or filter_fn(agent, ...) then
                if rel <  RELATIONSHIP.NEUTRAL then
                    if not worst_op or rel < worst_op then
                        worst, worst_op = agent, rel
                    end
                end
            end
        end
    end
    return worst, worst_op
end

function SocialConnections:GetRandomConnection(filter)
    local candidates = ObtainWorkTable()

    for agent, rel in pairs(self.relationships) do
        if filter == nil or filter(agent, rel) then
            table.insert(candidates, agent)
        end
    end
    local ret = #candidates > 0 and table.arraypick(candidates)
    ReleaseWorkTable(candidates)
    return ret
end

function SocialConnections:GetConnections(filter)
    local candidates = {}

    for agent, rel in pairs(self.relationships) do
        if filter == nil or filter(agent, rel) then
            table.insert(candidates, agent)
        end
    end
    return candidates
end

function SocialConnections:GetBestFriend(filter_fn, ...)
    local best, best_op
    for agent,rel in pairs(self.relationships) do
        if not agent:IsRetired() then
            if not filter_fn or filter_fn(agent, ...) then
                if rel >  RELATIONSHIP.NEUTRAL then
                    if not best_op or rel > best_op then
                        best, best_op = agent, rel
                    end
                end
            end
        end
    end
    return best, best_op
end

function SocialConnections:HasEnemies()
    for k,rel in pairs(self.relationships) do
        if rel < RELATIONSHIP.NEUTRAL then
            return true
        end
    end
end

function SocialConnections:HasFriends()
    for k,rel in pairs(self.relationships) do
        if rel > RELATIONSHIP.NEUTRAL then
            return true
        end
    end
end

function SocialConnections:HasOpinionEvents( agent )
    return self.history[ agent ] ~= nil
end

function SocialConnections:FindOpinionEvent( agent, event, min_age, max_age, fn )
    -- Most recent first, so search backwards.
    local now = Now()
    local events = self.history[ agent ]
    if events then
        for i = #events, 1, -1 do
            local ev = events[i]
            if max_age and (now - ev.time) > max_age then
                return
            end

            if min_age and (now - ev.time) < min_age then
                return
            end

            if ev.id == event.id then
                if fn and not fn(event) then
                    return 
                end
                return ev
            end
        end
    end
end

function SocialConnections:ProcessOpinionEvent(opinion_event_data, instance_data, from_agent)
    local opinion_event = 
    {
        data = instance_data and shallowcopy( instance_data ) or nil,
        time = Now(),
        where = self.agent:GetLocation(),
        id = opinion_event_data.id,
        agent = self.agent,
        from_agent = from_agent,
        guid = TheGame:GetGameState():GenerateID(),
    }

    if self.history[from_agent] == nil then
        self.history[from_agent] = {}
    end
    table.insert(self.history[from_agent], opinion_event)
    
    if from_agent.social_connections.history[self.agent] == nil then
        from_agent.social_connections.history[self.agent] = {}
    end
    table.insert(from_agent.social_connections.history[self.agent], opinion_event)


    local dat
    if type(opinion_event_data.delta) == "function" then
        dat = opinion_event_data.delta(self.agent, instance_data, from_agent)
    else
        dat = opinion_event_data.delta
    end
    if dat == nil then
        return
    end

    local old_rel = self:GetRelationship(from_agent)
    local new_rel = old_rel

    if dat.relationship_delta[old_rel] then
        if type(dat.relationship_delta[old_rel]) == "function" then
            new_rel = dat.relationship_delta[old_rel](self.agent, instance_data, from_agent)
        else
            new_rel = dat.relationship_delta[old_rel]
        end
    
        new_rel = new_rel or old_rel


        if new_rel ~= old_rel then
            if from_agent:IsPlayer() and not self.agent:IsDead() then
                
                --maybe we should just suppress these from popping up? 
                local skip = TheGame:GetGameState():GetCaravan():GetCurrentEncounter() and (new_rel == RELATIONSHIP.LOVED or new_rel == RELATIONSHIP.HATED)
                if not skip then
                    TheGame:GetGameState():LogNotification( NOTIFY.RELATIONSHIP_CHANGED, self.agent, old_rel, new_rel )            
                end

            end

            
            opinion_event.changed_relationship = old_rel
            self:SetRelationship(new_rel, from_agent)

            if from_agent:IsPlayer() then
                if new_rel == RELATIONSHIP.LOVED then
                    TheGame:AddGameplayStat( "loved", 1 )

                    local def = Content.FindAchievement( "beloved" )
                    if def then
                        local count = from_agent.social_connections:CountConnections( function( agent, rel ) return rel == RELATIONSHIP.LOVED end )
                        if count >= def.count then
                            TheGame:UnlockAchievement( def.id )
                        end
                    end

                elseif new_rel == RELATIONSHIP.HATED then
                    TheGame:AddGameplayStat( "hated", 1 )

                    local def = Content.FindAchievement( "archenemy" )
                    if def then
                        local count = from_agent.social_connections:CountConnections( function( agent, rel ) return rel == RELATIONSHIP.HATED end )
                        if count >= def.count then
                            TheGame:UnlockAchievement( def.id )
                        end
                    end
                end
            end
       end
    end

    if dat.post_apply then
        dat.post_apply(self.agent)
    end

    return opinion_event
end

function SocialConnections:GetRelationshipReason(who)
    if self.history[who] then
        for k = #self.history[who], 1, -1 do
            
            if self.history[who][k].changed_relationship then
                return self.history[who][k]
            end
        end
    end
end

function SocialConnections:GetRelationshipHistory(who)
    local txt = {}
    if self.history[who] then
        for k = 1, #self.history[who] do
            local op = self.history[who][k]
            if op.changed_relationship then
                if #txt == 0 then
                    table.insert( txt, loc.format( LOC"UI.AGENT_RELATIONSHIP_HEADER_TT", GetRelationshipModifierString( self:GetRelationship( who ), self.agent )))
                end
                table.insert( txt, "•" .. FormatOpinionEvent( op, who ))
            end
        end
    end
    return table.concat( txt, "\n" )
end

function SocialConnections:OnAgentReady( )
    for agent,rel in pairs(self.relationships) do
        if self.history[agent] ~= nil then
            self:SanitizeOpinionEvents(agent)
        end
    end
end

function SocialConnections:SanitizeOpinionEvents( agent )
    local events = self.history[ agent ]
    for i = #events, 1, -1 do
        if VerifyOpinionEvent(events[i]) == false then
            table.remove(self.history[agent], i)
        end
    end
end

function SocialConnections:RenderDebugUI( ui, panel )
    ui:Columns( 2 )
    ui:SetColumnOffset(1, 120)

    for agent, relationship in pairs (self.relationships) do
        panel:AppendTable( ui, agent )
        ui:NextColumn()

        ui:Text( RELATIONSHIP_ARRAY[ relationship ] )
        ui:NextColumn()
    end

    ui:Separator()

    for agent, history in pairs( self.history ) do
        panel:AppendTable( ui, agent )
        ui:NextColumn()

        for i, ev in ipairs( history ) do
            panel:AppendTable( ui, ev, FormatOpinionEvent( ev ))
        end
        ui:NextColumn()
    end
    ui:Columns( 1 )
end

return SocialConnections

