require "content/characters/base_types"
local battle_defs = require "battle/battle_defs"
local negotiation_defs = require "negotiation/negotiation_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local ADMIRALTY_NEGOTIATION =
{
	PAPERS = 0,

	MAX_INTERROGATE = { 2, 3, 3, 4, 4 },

	OnInit = function( self, difficulty )
		self.frisk = self:AddCard( "frisk" )
		self.frisk.money = self.frisk.money * difficulty
		self.frisk.stacks = difficulty

		self.attacks = self:MakePicker()
		self.attacks:AddArgument( "INTERROGATE", 1 )

		if self.PAPERS > 0 then
			self.attacks:AddID( "paper_pusher", self.PAPERS )
		end

		if difficulty <= 2 then
			self:SetPattern( self.BasicCycle )
		else
			self:SetPattern( self.Cycle )
		end

		self.max_args = self.MAX_INTERROGATE[ difficulty ] or 4
		self.negotiator:AddModifier("POWER_ABUSE")
	end,

	BasicCycle = function( self, turns )
		local args = self.negotiator:GetModifierInstances( "interrogate" )

		-- Interrogate, every 3 turns up to max_args.
		if (turns - 1) % 3 == 0 and args < self.max_args then
			self.attacks:ChooseCards( 1 )
			self:ChooseGrowingNumbers( 1, 0 )

		elseif turns % 3 == 0 then
			self:ChooseGrowingNumbers( 2, 0 )

		else
			self:ChooseGrowingNumbers( 1, 1 )
		end

		if (turns + 1) % 3 == 0 then
			if self.frisk and math.random() < 0.5 then
				self:ChooseCard( self.frisk )
				self.frisk = nil
			else
				self:ChooseComposure( 1, 1 + self.difficulty, 1 + self.difficulty )
			end
		end
	end,

	Cycle = function( self, turns )
		local args = self.negotiator:GetModifierInstances( "interrogate" )

		-- Interrogate, every 3 turns up to max_args.
		if (turns - 1) % 3 == 0 and args < self.max_args then
			if (turns - 1) % 6 == 0 then
				-- Double up every 6 turns.
				self.attacks:ChooseCards( 2 )
			else
				self.attacks:ChooseCards( 1 )
			end
			self:ChooseGrowingNumbers( 1, 0 )

		elseif turns % 3 == 0 then
			self:ChooseGrowingNumbers( 2, 0, 0.75 )

		else
			self:ChooseGrowingNumbers( 1, 1 )
		end

		if (turns + 1) % 3 == 0 then
			if self.frisk and math.random() < 0.5 then
				self:ChooseCard( self.frisk )
				self.frisk = nil
			else
				self:ChooseComposure( 1, 1 + self.difficulty, 1 + self.difficulty )
			end
		end
	end,
}

Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_BASE",
	{
		base_def = "NPC_BASE",
		is_template = true,

		text_colour = 0x3f61d0FF,
		colour =
		{
			flag = 0x363639ff,
		},

		faction_id = "ADMIRALTY",

		voice_actors = {
		    [GENDER.MALE] = { "admiraltyLowClassMale01" },
    		[GENDER.FEMALE] = { "admiraltyLowClassFemale01" },
		},

	    death_money = DEATH_MONEY_MEDIUM,
	    death_item = "admiralty_intel", -- Really should never happen but is here as a falback

		social_boons =
		{
			[ RELATIONSHIP.LIKED ] = { GIVE_ITEM = 1, SOCIALIZE = 1},
			[ RELATIONSHIP.LOVED ] = { GIVE_ITEM = 1, UPGRADE_BATTLE_CARD = 2},
		},

		support_negotiation = "admiralty_support",
		combat_backup = "ADMIRALTY_PATROL_BACKUP",

		tags = {"guard"},

		renown = 1,
	})
)

--------------------------------------------

Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_PATROL_LEADER",
	{
		designer_note = "Protects allies, and buffs them. Won't surrender if any allies are still around.",
		base_def = "ADMIRALTY_BASE",

		title = "Patrol Leader",
		tags = { "patrolleader", "enforcer"},
		renown = 3,
		combat_strength = 4,
		gen_count = 5,

		voice_actors = {
		    [GENDER.MALE] = { "admiraltyHighClassMale01" },
    		[GENDER.FEMALE] = { "admiraltyHighClassFemale01" },
		},

		loved_graft = "sparring",
		hated_graft = "intimidated",

	    negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_patrol_leader.tex"),

    	death_money = DEATH_MONEY_MEDIUM,
    	death_item = "spear_head",

		fight_data = 
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.HIGH,

	        status_widget_head_dx = 0.0,
	        status_widget_head_dy = 5.0,

	        attacks = 
	        {
			    admiralty_thrust = table.extend(NPC_ATTACK)
			    {
			        name = "Mighty Thrust",
			        anim = "mightythrust",
			        flags = CARD_FLAGS.MELEE,
			        damage_mult = 1.25,
			    },

			    npc_cyclone = table.extend(NPC_ATTACK)
			    {
			        name = "Cyclone",
			        anim = "hamstring",
			        flags = CARD_FLAGS.MELEE,
			        target_mod = TARGET_MOD.TEAM,
			        damage_mult = 0.5,
			    },

			    prepared_stance = table.extend(NPC_BUFF)
			    {
			        name = "Prepared Stance",
			        anim = "taunt",
			        target_type = TARGET_TYPE.SELF,
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,

			        defend = { 3, 6, 9, 12, 15 },
			        riposte = { 0, 1, 2, 3, 4 },

			        OnPreResolve = function(self, battle, attack)
			            self.owner:AddCondition( "PROTECT", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
			            self.owner:AddCondition( "DEFEND", self:ScaleBuff( self.defend ))
			            self.owner:AddCondition( "RIPOSTE", self:ScaleBuff( self.riposte ))
			        end,
			    },

			    patrol_leader_to_arms = 
			    {
			        name = "To Arms",
			        anim = "taunt",
			        target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
			        target_mod = TARGET_MOD.TEAM,
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,

			        OnPostResolve = function(self, battle, attack)
			        	local ally_count = #self.owner:GetTeam():GetFighters()
			        	local power = 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
			        	for i, fighter in ipairs(self.owner:GetTeam():GetFighters()) do
			        		if fighter.agent:GetFactionID() == "ADMIRALTY" then
				        		fighter:AddCondition("POWER", power)
				        	end
			        	end
			    	end,
			    },

	    	},

            anim_mapping =
            {
                riposte = "hamstring",
            },

			cards = { "admiralty_thrust", "npc_cyclone", "ai_defend_med", "prepared_stance" },

			behaviour =
			{
				OnActivate = function( self )
					self.attacks = self:MakePicker()
						:AddID( "admiralty_thrust", 1 )
						:AddID( "npc_cyclone", 1 )

					self.buff = self:MakePicker()
						:AddID( "prepared_stance", 1 )
						:AddID( "patrol_leader_to_arms", 1 )

					self.fighter:AddCondition( "DAUNTLESS", 1 )

					self:SetPattern( self.Buff )
				end,

				Buff = function( self )
					self.buff:ChooseCard()
					self:SetPattern( self.Attack )
				end,

				Attack = function( 	self )
					local enemy_count = #self.fighter:GetEnemyTeam():GetFighters()
					local ally_count = #self.fighter:GetTeam():GetFighters()
					self.attacks:ChooseCard()

					local p = ally_count > 1 and 0.75 or 0.5
					if math.random() < p then
						self:SetPattern( self.Buff )
					end
				end
			},
		},

		base_builds = {
			[ GENDER.MALE ] = "guard_corporal_male",
			[ GENDER.FEMALE ] = "guard_corporal_female"
		},
		
		anims = {"anim/weapon_spear_guard_lumin.zip"},
		combat_anims = { "anim/med_combat_spear_admiralty_patrol_leader.zip"},

		idle_anims = 
		{
			guarding = "idle_neutral_guarding"
		},

		negotiation_data = 
		{
			behaviour = table.extend( ADMIRALTY_NEGOTIATION )
			{
			}
		},
	})
)

------------------------------------------------------
Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_CLERK",
	{
		designer_note = "Gives you injury cards and surrenders if anyone else is surrendered.",
		base_def = "ADMIRALTY_BASE",

		title = "Clerk",
		tags = {"clerk"},
		renown = 3,
		combat_strength = 1.25,
		death_item = "letter_opener",
 		negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_clerk.tex"),

		voice_actors = {
		    [GENDER.MALE] = { "admiraltyHighClassMale01" },
    		[GENDER.FEMALE] = { "admiraltyHighClassFemale01" },
		},

		loved_graft = "authorization",
		hated_graft = "red_tape",

		fight_data =
		{
			MAX_HEALTH = MAX_HEALTH_LOOKUP.VERY_LOW,
			MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,

			attacks = 
			{
			    paper_cut = table.extend(NPC_ATTACK)
			    {
			        name = "Paper Cut",
			        anim = "slice",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        damage_mult = 0.75,


			        event_handlers =
			        {
			            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
			                if attack.card == self and attack.attacker == self.owner and not hit.evaded and not hit.defended then
								local num_injuries = math.floor( self.owner:GetScale() - 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
								num_injuries = clamp( num_injuries, 1, 3 )

			                    if attack.target:IsPlayer() then

			                        local cards = {}
			                        for i=1, num_injuries do
			                        	table.insert(cards, Battle.Card( "status_injury", attack.target ))
			                        end

			                        battle:DealCards( cards )
			                    else
			                        attack.target:AddCondition("IMPAIR", num_injuries)
			                    end
			                end
			            end,
			        }
				},

			    tend = table.extend(NPC_BUFF)
			    {
			        name = "Tend",
			        anim = "taunt",
			        quip = "admiralty_clerk_tend",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF | CARD_FLAGS.HEAL,
			        target_type = TARGET_TYPE.SELF,

			        min_triage = 4,
			        max_triage = 10,

			        OnInit = function( self )
			        	local bonus = math.floor( self.owner:GetScale() * (1.0 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )))
			        	self.min_triage = self.min_triage + bonus
			        	self.max_triage = self.max_triage + bonus
			        end,

			        CanPlayCard = function(self, battle, target )
			            if target then
			                local health = target:GetHealth()
			                local min_health, max_health = target:GetStatBounds( COMBAT_STAT.HEALTH )
			                if max_health - health < self.min_triage then
			                    return false
			                end
			            end

			            return true
			        end,

			        OnPostResolve = function(self, battle, attack)
			            local delta = math.random( self.min_triage, self.max_triage )
			            attack.target:HealHealth( delta, self )
			        end,
			    },
			},

			cards = { "paper_cut", "tend" },

			conditions =
			{
				skittish =
				{
					name = "Skittish",
			        desc = "Will {SURRENDER} whenever an ally {SURRENDER|Panics} or dies.",
			        ctype = CTYPE.DEBUFF,

			        event_handlers =
			        {
			        	[ BATTLE_EVENT.FIGHTER_SURRENDER ] = function( self, fighter )
			        		if fighter ~= self.owner and self.owner:IsActive() and fighter:GetTeamID() == self.owner:GetTeamID() and self.owner:HasMorale() then
			        			self.owner:Surrender()
			        		end
			        	end,

			        	[BATTLE_EVENT.STATUS_CHANGED] = function( self, fighter, status )
			                if status == FIGHT_STATUS.FLED and fighter ~= self.owner and self.owner:IsActive() and fighter:GetTeamID() == self.owner:GetTeamID() and self.owner:HasMorale() then
			                    self.owner:Surrender()
			                end
			            end,

			        	[ BATTLE_EVENT.FIGHTER_KILLED ] = function( self, fighter )
			        		if fighter ~= self.owner and self.owner:IsActive() and fighter:GetTeamID() == self.owner:GetTeamID() and self.owner:HasMorale() then
			        			self.owner:Surrender()
			        		end
			        	end,
				    }
			    },
			},

			behaviour = 
			{
				OnActivate = function( self )
					self.attacks = self:MakePicker()
						:AddID( "paper_cut", 2, 4 )
						:AddID( "ai_knife_slice", 3 )

					self.tend = self:AddCard( "tend" )
					self:SetPattern( self.Attack )
					self.tend_count = 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )

					if not CheckBits(self.battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER) then
						self.fighter:AddCondition( "skittish" )
					end
				end,

				Attack = function( self )

					self.attacks:ChooseCard()

					if self.tend_count > 0 and math.random() < 0.6 and self.fighter:GetHealthPercent() < 0.6 then 
						self:SetPattern( self.Tend )
					end
				end,

				Tend = function( self )
					self:ChooseCard( self.tend )
					self.tend_count = self.tend_count - 1
					self:SetPattern( self.Attack )
				end,
			},
		},

		base_builds = {
			[ GENDER.MALE ] = "male_guard_civ",
			[ GENDER.FEMALE ] = "guard_female_civ"
		},

		idle_anims = 
		{
			guarding = "idle_neutral_guarding_ranged"
		},

		negotiation_data = 
		{
			behaviour = table.extend( ADMIRALTY_NEGOTIATION )
			{
				PAPERS = 3,
			}
		},

		anims = {"anim/weapon_knife_guard_lumin.zip"},
		combat_anims = { "anim/med_combat_knife_admiralty_clerk.zip" },
	})
)

------------------------------------------------------

local GOON_FIGHT_BEHAVIOUR = 
{
	elite = false,
	OnActivate = function( self )
		self.attacks = self:MakePicker()
			:AddID( "heavy_swing", 1 )
			:AddID( "disembowel", 1 )

		self.elite_attacks = self:MakePicker()
			:AddID( "heavy_swing", 2 )
			:AddID( "lumin_swing", 1 )

		self.buffs = self:MakePicker()
			:AddID( "ai_defend_lo", 1 )

		if math.random() < 0.5 then
			self:SetPattern( self.Attack )
		else
			self:SetPattern( self.Buff )
		end
		local cond = self.fighter:AddCondition( "formation_tracker", 1 )
		if cond then
			cond:ApplyFormation()
		end
	end,

	Attack = function( self )
		if self.elite then
			self.elite_attacks:ChooseCard()
		else
			self.attacks:ChooseCard()
		end
		self:SetPattern( self.Buff )
	end,

	Buff = function( self )
		self.buffs:ChooseCard()
		self:SetPattern( self.Attack )
	end,
}

Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_GOON",
	{
		designer_note = "Buffs allies with defend if you don't attack them",
		base_def = "ADMIRALTY_BASE",

		title = "Goon",
		tags = {"goon", "enforcer"},
		renown = 1,
		combat_strength = 2,
		promotion_def = "ADMIRALTY_GOON_PROMOTED",
    	negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_goon.tex"),

		loved_graft = "piece_of_the_pie",
		hated_graft = "hunted",

		death_item = "admiralty_intel",

		voice_actors = {
		    [GENDER.MALE] = { "admiraltyLowClassMale01" },
    		[GENDER.FEMALE] = { "admiraltyLowClassFemale01" },
		},
		
		base_builds = {
			[ GENDER.MALE ] = "guard_male",
			[ GENDER.FEMALE ] = "guard_female"
		},

		anims = {"anim/weapon_sword_common.zip"},
		combat_anims = { "anim/med_combat_sword_admiralty_goon.zip" },


		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.LOW,

			conditions = 
			{
			    formation_tracker =
			    {
			        name = "Formation",

			        hidden = true,
			        defend_amt = {3,5,8,10},

			        GetDefendAmt = function( self )
			            local i = math.min(#self.defend_amt, self.owner:GetScale())
			            return self.defend_amt[i] + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
			        end,

			        ApplyFormation = function( self )
	                    self.owner:AddCondition("formation", self:GetDefendAmt())
			        end,

			        event_handlers = 
			        {
			            [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
			                if self.owner == fighter then
			                    self.owner:RemoveCondition("formation")
			                end
			            end,

			            [ BATTLE_EVENT.END_TURN ] = function( self, fighter)
			            	if fighter == self.owner and not self.owner:HasCondition("formation") and self.owner:IsActive() then
			            		self:ApplyFormation()
			                end
			            end,
			        },
			    },

			    formation =
			    {
			        name = "Formation",
			        desc = "At the end of {1}'s turn, apply <#HILITE>{DEFEND {2}}</> to all allies.\n\n<b>Formation</b> is removed when health is lost and regained at the end of {1}'s turn.",
			        desc_fn = function( self, fmt_str )
			            return loc.format( fmt_str, self:GetOwnerName(), self.stacks)
			        end,

        			apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",
			        ctype = CTYPE.INNATE,

			        event_handlers = 
			        {
			            [ BATTLE_EVENT.END_TURN ] = function ( self, fighter )
			                if fighter == self.owner and fighter:IsActive() then
			                    local card = Battle.Card("formation_defend", self.owner)
			                    card.defend_amt = self.stacks
                            	self.battle:PlayCard( card, self.owner )
			                end
			            end,
			        },
			    },

			    lumin_burnt = 
			    {
			    	name = "Lumin Burnt",
			    	desc = "Whenever an attack removes all {1}'s {DEFEND}, take <#HILITE>{2} damage</>. Reduce <b>Lumin Burnt</> by 1 at the end of your turn.",
			    	desc_fn = function( self, fmt_str )
			    		return loc.format( fmt_str, self:GetOwnerName(), self.stacks)
			    	end,

			    	ctype = CTYPE.DEBUFF,
			    	event_handlers =
			    	{
			    		[ BATTLE_EVENT.END_TURN ] = function( self, fighter )
			    			if fighter == self.owner then
			    				self.owner:RemoveCondition(self.id, 1, self)
			    			end
			    		end,

			    		[ BATTLE_EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
	                        if fighter == self.owner and condition:GetID() == "DEFEND" and condition:GetStacks() == 0 and source and source.owner ~= self.owner then
	                            self.owner:ApplyDamage(self.stacks, source.owner, self, false, {"lumin"})
	                        end
	                    end,
			    	},
			    },

			},

			attacks = 
			{
			    disembowel = table.extend(NPC_ATTACK)
			    {
			        name = "Disembowel",
			        anim = "disembowel",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,
			        damage_mult = 0.5,

			        features =
			        {
			            EXPOSED = 2,
			        }
			    },

			    heavy_swing = table.extend(NPC_ATTACK)
			    {
			        name = "Heavy Swing",
			        anim = "swing",
			        flags = CARD_FLAGS.MELEE,
			    },

			    lumin_swing = table.extend(NPC_ATTACK)
			    {
			    	name = "Lumin Swing",
			    	anim = "elite_attack",
			    	flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			    	burnt_amt = {2, 3, 4, 4},

			    	OnPostResolve = function( self, battle, attack )
			    		local amt = self.burnt_amt[math.min(#self.burnt_amt, self.owner:GetScale())]
			    		attack:AddCondition("lumin_burnt", amt + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ), self)
			    	end
			    },

			    formation_defend = 
			    {
			        name = "Formation Defend",
			        anim = "taunt_rally",
			        flags = CARD_FLAGS.SKILL,
			        target_type = TARGET_TYPE.SELF,

			        OnPostResolve = function(self, battle)
			            assert(self.defend_amt)
			            for i, fighter in self.owner:GetTeam():Fighters() do
			            	if fighter ~= self.owner then
				                fighter:AddCondition("DEFEND", self.defend_amt)
				            end
			            end
			        end,
			    },
			},

			behaviour = table.extend(GOON_FIGHT_BEHAVIOUR)
			{},
		},

		negotiation_data =
		{
			behaviour = table.extend( ADMIRALTY_NEGOTIATION )
			{
			}
		}
	})
)

Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_GOON_PROMOTED",
	{
		base_def = "ADMIRALTY_GOON",

		title = "Promoted Goon",
		combat_strength = 3,

		death_item = "admiralty_intel_plus",

		anims = {"anim/weapon_sword_guard_promoted.zip"},
		combat_anims = { "anim/med_combat_sword_guard_promoted.zip" },

		

		base_builds = {
			[ GENDER.MALE ] = "male_guard_promoted_build",
			[ GENDER.FEMALE ] = "female_guard_promoted_build",
		},

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM,
			cards = { "heavy_swing", "lumin_swing", "formation_defend", "ai_defend_lo" },

			behaviour = table.extend(GOON_FIGHT_BEHAVIOUR)
			{
				elite = true,
			},
		},
	})
)


Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_GUARD",
	{
		designer_note = "Gets stronger if they hit you.",
		base_def = "ADMIRALTY_BASE",

		title = "Guard",
		tags = {"goon", "enforcer"},
		renown = 2,
		combat_strength = 3,

		loved_graft = "commander",
		hated_graft = "lowered_rank",

		death_item = "admiralty_requisition",
		
		voice_actors = {
		    [GENDER.MALE] = { "admiraltyLowClassMale01" },
    		[GENDER.FEMALE] = { "admiraltyLowClassFemale01" },
		},

		base_builds = {
			[ GENDER.MALE ] = "male_guard_02",
			[ GENDER.FEMALE ] = "female_guard_02"
		},

		anims = { "anim/weapon_hammer_guard.zip"},
		combat_anims = { "anim/med_combat_hammer_admiralty_guard.zip"},
	    negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_guard.tex"),

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM,

	        status_widget_head_dx = 0.0,
	        status_widget_head_dy = 4.6,

	        attacks = 
	        {
			    hammer_smash = table.extend(NPC_ATTACK)
			    {
			        name = "Hammer Smash",
			        anim = "smash",
			        flags = CARD_FLAGS.MELEE,
			        damage_type = DAMAGE_TYPE.DULL,
			    },
	        },	

			cards = { "hammer_smash", "ai_defend_lo" },

			behaviour =
			{
				OnActivate = function( self, fighter )
					self.attack = self:AddCard( "hammer_smash" )
					self.buffs = self:MakePicker()
						:AddID( "ai_defend_lo", 1 )

					local stacks = 1 + math.floor( fighter:GetScale() / 2 ) + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
					fighter:AddCondition( "OPPORTUNISTIC", stacks )

					if math.random() < 0.5 then
						self:SetPattern( self.Attack )
					else
						self:SetPattern( self.Buff )
					end
				end,

				Attack = function( self, fighter, battle )
					self:ChooseCard( self.attack )
					self:SetPattern( self.Buff )
				end,

				Buff = function( self, fighter, battle )
					self.buffs:ChooseCard()
					self:SetPattern( self.Attack )
				end
			},
		},

		negotiation_data =
		{
			behaviour = table.extend( ADMIRALTY_NEGOTIATION )
			{
			}
		}
	})
)

Content.AddCharacterDef
(
	CharacterDef("ADMIRALTY_INVESTIGATOR",
	{
		base_def = "ADMIRALTY_BASE",

		title = "Investigator",
		renown = 4,
		combat_strength = 3,

		loved_graft = "informant",
		hated_graft = "watched",

		death_item = "classified_file",
		
		-- voice_actors = {
		--     [GENDER.MALE] = { "admiraltyLowClassMale01" },
  --   		[GENDER.FEMALE] = { "admiraltyLowClassFemale01" },
		-- },

		base_builds = {
			[ GENDER.MALE ] = "male_guard_investigator_build",
			[ GENDER.FEMALE ] = "female_guard_investigator_build"
		},

		anims = { "anim/weapon_halberd_guard_investigator.zip"},
		combat_anims = { "anim/med_combat_halberd_guard_investigator.zip"},
	    -- negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_guard.tex"),

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM,
			death_fade_delay = 0.8,

	        attacks = 
	        {
			    investigator_slice = table.extend(NPC_ATTACK)
			    {
			        name = "Slice and Dice",
			        anim = "attack",
			        flags = CARD_FLAGS.MELEE,
			        hit_count = 2,
			        damage_mult = 0.5,
			    },

			    investigator_lunge = table.extend(NPC_ATTACK)
			    {
			        name = "Lunge",
			        anim = "attack2",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        OnPostResolve = function( self, battle, attack )
			        	for i,hit in ipairs(attack.hits) do
			        		if hit.target:IsPlayer() then
			        			hit.target:AddCondition("BATTERED", 2 + 1 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ), self)
			        		else
			        			hit.target:AddCondition("IMPAIR", 2 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ), self)
			        		end
			        	end
			        end
			    },

			    investigator_taunt = table.extend(NPC_BUFF)
			    {
			    	name = "Taunt",
			    	anim = "taunt",
			    	flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
			    	target_type = TARGET_TYPE.SELF,

			    	OnPostResolve = function( self, battle, attack )
			    		attack:AddCondition("DEFEND", 8 + 4 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) , self)
			    		attack:AddCondition("POWER", 2 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) , self)
			    		attack:AddCondition("POWER_LOSS", 2 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) , self)
			    	end
			    },
	        },

			behaviour =
			{
				OnActivate = function( self, fighter )
					self.attacks = self:MakePicker()
						:AddID("investigator_slice", 2)
						:AddID("investigator_lunge", 1)
					self.buff = self:AddCard("investigator_taunt")

					self.switch_turn = 1
					if math.random() < 0.5 then
						self:SetPattern( self.Attack )
					else
						self:SetPattern( self.Buff )
					end
				end,

				Attack = function( self, turns )
					self.attacks:ChooseCard()
					if turns - self.switch_turn > 2 then
						
						self:SetPattern( self.Buff )
					end
				end,

				Buff = function( self, turns )
					self:ChooseCard( self.buff )
					self.switch_turn = turns
					self:SetPattern( self.Attack )
				end
			},
		},

		negotiation_data =
		{
			behaviour = 
			{
				MAX_SEARCH = { 1, 2, 2, 3, 3 },

				OnInit = function( self, difficulty )
					self.frisk = self:AddCard( "frisk" )
					self.frisk.money = self.frisk.money * difficulty
					self.frisk.stacks = difficulty

					self.attacks = self:MakePicker()
					self.attacks:AddArgument( "THOROUGH_SEARCH", 1 )

					if difficulty <= 2 then
						self:SetPattern( self.BasicCycle )
					else
						self:SetPattern( self.Cycle )
					end

					self.max_args = self.MAX_SEARCH[ difficulty ] or 2
					self.negotiator:AddModifier("INVESTIGATOR")
				end,

				BasicCycle = function( self, turns )
					local args = self.negotiator:GetModifierInstances( "interrogate" )

					-- Search, every 3 turns up to max_args.
					if (turns + 1) % 3 == 0 and args < self.max_args then
						self.attacks:ChooseCards( 1 )
						self:ChooseGrowingNumbers( 2, 0, 0.75, 3 )

					elseif turns % 3 == 0 then
						self:ChooseGrowingNumbers( 3, 0 )

					else
						self:ChooseGrowingNumbers( 2, 0, 0.75, 3 )
					end
				end,

				Cycle = function( self, turns )
					local args = self.negotiator:GetModifierInstances( "interrogate" )

					-- Search, every 3 turns up to max_args.
					if (turns + 1) % 3 == 0 and args < self.max_args then
						if (turns - 1) % 6 == 0 then
							-- Double up every 6 turns.
							self.attacks:ChooseCards( 2 )
						else
							self.attacks:ChooseCards( 1 )
						end
						self:ChooseGrowingNumbers( 2, 0, 0.75, 3 )

					elseif turns % 3 == 0 then
						self:ChooseGrowingNumbers( 3, 0, 0.75 )

					else
						self:ChooseGrowingNumbers( 2, 0, 0.75, 3 )
					end
				end,
			}
		}
	})
)

------------------------------------------------------
