local battle_defs = require "battle/battle_defs"
local negotiation_defs = require "negotiation/negotiation_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local BANDIT_NEGOTIATION =
{
	KINGPIN = 0,

	OnInit = function( self )
		self.escalation = self:AddArgument( "ESCALATION" )
		self.kingpin = self:AddArgument( "KINGPIN" )
		self.greedy = self:AddArgument( "GREEDY" )

		if self.difficulty <= 2 then
			self:SetPattern( self.BasicCycle )
		else
			self:SetPattern( self.Cycle )
		end
		
		self.negotiator:AddModifier("SHORT_FUSE")
	end,
 	
	BasicCycle = function( self, turns )
		if turns == 2 then
			self:ChooseCard( self.escalation )
			self:ChooseGrowingNumbers( 1, 1 )

		elseif (turns-1) % 3 == 0 then
			-- Double attack.
			self:ChooseGrowingNumbers( 2, 1 )

		else
			-- Single attack.
			self:ChooseGrowingNumbers( 1, 1 )
		end
	end,

	Cycle = function( self, turns )
		-- Starting turn 3, "Buff" every 3 turns.
		if turns % 3 == 0 then
			self:ChooseCard( self.greedy )
		end

		-- Double attack every 2 rounds; Single attack otherwise.
		if self.difficulty >= 4 and turns % 2 == 0 then
			self:ChooseGrowingNumbers( 3, -1 )
		elseif turns % 2 == 0 then
			self:ChooseGrowingNumbers( 2, 1 )
		else
			self:ChooseGrowingNumbers( 1, 3 )
		end

		-- Escalate every 5 turns if it doesn't exist.
		-- NOTE: Added last so it doesn't immediately apply to current attacks
		if (turns - 1) % 5 == 0 and not self.negotiator:FindModifier( "ESCALATION" ) then
			self:ChooseCard( self.escalation )
		end

		-- Kingpin every 5 turns, starting turn 2, if doesn't exist.
		if self.KINGPIN > 0 then
			if (turns + 3) % 5 == 0 and not self.negotiator:FindModifier( "KINGPIN" ) then
				self:ChooseCard( self.kingpin )
			end
		end

	end,
}

-------------------------------------------------------------
-- Basic bandit template

Content.AddCharacterDef
(
	CharacterDef("BANDIT_BASE",
	{
		base_def = "NPC_BASE",
		is_template = true,

		faction_id = "BANDITS",
		text_colour = 0xc44a2aFF,

		voice_actors = {
			[GENDER.MALE] = { "spreeLowClassMale01" },
    		[GENDER.FEMALE] = { "spreeLowClassFemale01" },
		},

		loved_graft = "surprise_attack",
		hated_graft = "shakey",

		tags = { "bandit", "goon" },

		renown = 1,

		support_negotiation = "spree_support",
		combat_backup = "BANDIT_PATROL_BACKUP",

		social_boons =
		{
			[ RELATIONSHIP.LIKED ] = { GIVE_ITEM = 1, SOCIALIZE = 1},
			[ RELATIONSHIP.LIKED ] = { GIVE_ITEM = 1, SOCIALIZE = 1, UPGRADE_BATTLE_CARD = 1, HIRE = 1 },
		},

		base_builds = {
			[ GENDER.MALE ] = "bandit_male",
			[ GENDER.FEMALE ] = "bandit_female"
		},

    	death_money = DEATH_MONEY_LOW,
    	death_item = "vial_of_slurry",
	})
)


-------------------------------------------------------------
-- Bandit jobs.

local GOON_FIGHT_BEHAVIOUR = 
{
	elite = false,
	OnActivate = function( self )
			self.fighter:AddCondition( "last_stand", 3 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) )
			self.attacks = self:MakePicker()
				:AddID( "multihit", 1 )
				:AddID( "slug", 1 )
			self.elite_attacks = self:MakePicker()
				:AddID( "slug", 1 )
				:AddID( "cheese_grater", 1 )
			self.revup = self:AddCard( "revup" )

			self:SetPattern( math.random() < 0.5 and self.Attack or self.Buff )
		end,

		Attack = function( self )
			if self.elite then
				self.elite_attacks:ChooseCards( 1 )
			else
				self.attacks:ChooseCards( 1 )
			end
			self:SetPattern( self.Buff )
		end,

		Buff = function( self )
			if not self:ChooseCard( self.revup ) then
				if self.elite then
					self.elite_attacks:ChooseCards( 1 )
				else
					self.attacks:ChooseCards( 1 )
				end
			end
			self:SetPattern( self.Attack )
		end,
}

Content.AddCharacterDef
(
	CharacterDef("BANDIT_GOON",
	{
		designer_note = "Deals more damage when low on health, and bleeds the enemy.",

		base_def = "BANDIT_BASE",
		title = "Thug",
		tags = { "enforcer" },
		promotion_def = "BANDIT_GOON_PROMOTED",

		loved_graft = "sucker_punch",
		hated_graft = "marked",

		base_builds = {
			[ GENDER.MALE ] = "hunter_male",
			[ GENDER.FEMALE ] = "hunter_female"
		},    

		voice_actors = {
			[GENDER.MALE] = { "spreeLowClassMale01" },
    		[GENDER.FEMALE] = { "spreeLowClassFemale01" },
    	},

		renown = 1,
		combat_strength = 2,

		anims = {"anim/weapon_knuckles_bandit_common.zip"},
		combat_anims = { "anim/med_combat_knuckles_bandit_goon.zip" },

		fight_data = 
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.LOW+5,

			anim_mapping =
			{
				riposte = "cut"
			},

			attacks = 
			{
			    revup = 
			    {
			        name = "Rev Up",
			        anim = "taunt",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
			        target_type = TARGET_TYPE.SELF,

			        condition = 
			        {
			            name = "Revved Up",
			            desc = "Attacks apply 1 {BLEED}, then reduce Revved Up by 1.",

			            event_handlers =
			            {
			                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
			                    if attack.attacker == self.owner and attack.card:IsAttackCard() and not hit.evaded then
			                    	if hit.target:IsPlayer() then
		                                local cards = {}
		                                local incepted_card = Battle.Card( "status_bleed", hit.target)
		                                table.insert(cards, incepted_card )                                  
		                                battle:DealCards( cards, battle:GetDiscardDeck() )
		                            else
		                                hit.target:AddCondition("BLEED", 1, attack)
		                            end
		                            self.owner:RemoveCondition( self.id, 1 )
			                    end
			                end
			            },
			        },
			        CanPlayCard = function(self)
			            return self.owner:GetConditionStacks(self.id) == 0
			        end,

			        OnPostResolve = function(self, battle, attack)
			            self.owner:AddCondition( self.id, 4 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) )
			        end,
			    },

			    multihit = table.extend(NPC_ATTACK)
			    {
			        name = "Fisticuffs",
			        anim = "thousand_cuts",
			        flags = CARD_FLAGS.MELEE,

			        base_damage = 2,
			        hit_count = 3,
			    },

			    slug = table.extend(NPC_ATTACK)
			    {
			        name = "Slug",
			        anim = "cut",
			        flags = CARD_FLAGS.MELEE,
			    },

			    cheese_grater = table.extend(NPC_ATTACK)
			    {
			        name = "Cheese Grater",
			        anim = "promoted_attack",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        shredded_count = { 0, 1, 1, 1 },
			        damage_mult = 0.8,
			        hit_count = 2,

			        OnPostResolve = function( self, battle, attack )
			        	for i,hit in ipairs(attack.hits) do
			        		if hit.target:IsPlayer() and not hit.evaded and not hit.defended then
			        			local incepted_cards = {}
			        			for i=1, self.shredded_count[math.min(#self.shredded_count, self.owner:GetScale())] do
			        				local card = Battle.Card( "shredded", hit.target )
				                    card.show_dealt = true
			        				table.insert(incepted_cards, card)
			        			end
			                    battle:DealCards( incepted_cards, battle:GetDiscardDeck() )
			        		end
			        	end
			        end
			    },
			},

			cards = { "multihit", "slug", "revup" },

			conditions = 
			{
			    last_stand =
			    {
			        name = "Last Stand",
			        desc = "Gains <#HILITE>{POWER {1}}</> when below {2} health.",
			        desc_fn = function(self, fmt_str)
			        	local max = self.owner:GetMaxHealth()
			        	return loc.format( fmt_str, self.stacks, math.ceil(max*self.rage_percent) )
			        end,

			        rage_percent = 0.6,
			        ctype = CTYPE.INNATE,

			        event_handlers =
			        {
			            [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
			                if fighter == self.owner and stat == COMBAT_STAT.HEALTH then
			                	if fighter:GetHealthPercent() < self.rage_percent then
			                		fighter:AddCondition( "POWER", self.stacks )
			                		fighter:RemoveCondition( self.id )
			                	end
			                end
		                end,
			        },  
			    },
			},

			behaviour = GOON_FIGHT_BEHAVIOUR,
		},
		

		idle_anims = 
		{
			guarding = "idle_neutral_guarding_ranged"
		},

		negotiation_data =
		{
			behaviour = table.extend( BANDIT_NEGOTIATION )
			{
			}
		},
		death_item = "makeshift_dagger",
	})
)

Content.AddCharacterDef
(
	CharacterDef("BANDIT_GOON_PROMOTED",
	{
		base_def = "BANDIT_GOON",
		title = "Promoted Thug",

		renown = 1,
		combat_strength = 2,

		base_builds = {
			[ GENDER.MALE ] = "male_bandit_promoted_build",
			[ GENDER.FEMALE ] = "female_bandit_promoted_build"
		},  

		anims = {"anim/weapon_knuckles_promoted_bandit.zip"},
		combat_anims = { "anim/med_combat_knuckles_promoted_bandit.zip" },

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.LOW + 10,
			cards = { "slug", "revup", "cheese_grater" },

			behaviour = table.extend(GOON_FIGHT_BEHAVIOUR)
			{
				elite = true,
			},
		},
		death_item = "makeshift_dagger_plus",
	})
)

Content.AddCharacterDef
(
	CharacterDef("BANDIT_GOON2",
	{
		designer_note = "Pins your cards down, making you pay actions to get them back.",
		base_def = "BANDIT_BASE",
		title = "Thief",
		tags = { "enforcer" },

		base_builds = {
			[ GENDER.MALE ] = "bandit_male",
			[ GENDER.FEMALE ] = "bandit_female"
		},

		voice_actors = {
			[GENDER.MALE] = { "spreeLowClassMale01" },
    		[GENDER.FEMALE] = { "spreeLowClassFemale01" },
    	},

		renown = 1,
		combat_strength = 3,

		loved_graft = "surprise_attack",
		hated_graft = "outwitted",
		
		anims = {"anim/weapon_coilgun_bandit_common.zip"},
		combat_anims = { "anim/med_combat_coilgun_bandit_goon2.zip" },

		fight_data = 
		{
			ignore_anim_symbols = {"gun"},

			formation = FIGHTER_FORMATION.BACK_X,
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,

			anim_mapping =
			{
				riposte = "pierce"
			},
			ranged_riposte = true,
			
			attacks = 
			{
			    ai_bandit_goon2_attack1 = table.extend(NPC_ATTACK)
			    {
			        name = "Pin down",
			        anim = "pierce",
			        flags = CARD_FLAGS.RANGED | CARD_FLAGS.DEBUFF,

			        damage_mult = 0.66,

			    	EvaluateTargetWeight = function( self, fighter, targets )
			    		return fighter:IsPlayer() and 2 or 1
			    	end,

			        OnPostResolve = function(self, battle, attack)
                        for i,hit in ipairs(attack.hits) do
                        	if not hit.evaded then
	                        	if hit.target:IsPlayer() and hit.target:IsActive() then
		                            local incepted_card = Battle.Card( "status_pinned", hit.target )
		                            incepted_card:DoPinning( battle, hit.target, 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) )
		                        else
		                        	hit.target:AddCondition("IMPAIR", 2)
		                        end
		                    end
                        end
			       	end,
			    },

			    ai_bandit_goon2_attack2 = table.extend(NPC_ATTACK)
			    {
			        name = "Impale",
			        anim = "open wounds",
			        flags = CARD_FLAGS.RANGED,

			        damage_mult = 1.4,
			    },

			    bandit_aim = table.extend(NPC_BUFF)
			    {
			        name = "Take Aim",
			        anim = "taunt",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
			        target_type = TARGET_TYPE.SELF,

			        OnPostResolve = function( self, battle, attack )
				        attack:AddCondition( "POWER", 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
			        end,
			    },

			},

			cards = { "ai_bandit_goon2_attack1", "ai_bandit_goon2_attack2" },

			behaviour = 
			{
				OnActivate = function( self )
					self.attacks = self:MakePicker()
						:AddID( "ai_bandit_goon2_attack2", 1 )
						
					self.pin = self:AddCard("ai_bandit_goon2_attack1")

					self.buffs = self:MakePicker()
						:AddCard( self.pin, 2 )
						:AddID( "bandit_aim", 1 )

					self:SetPattern( self.Start )
				end,

				Start = function( self )
					self:ChooseCard( self.pin )
					self:SetPattern( self.Attack )
				end,

				Attack = function( self )
					self.attacks:ChooseCard()
					self:SetPattern( self.Buff )
				end,

				Buff = function( self )
					self.buffs:ChooseCard()
					self:SetPattern( self.Attack )
				end,
			},
		},

		idle_anims = 
		{
			guarding = "idle_neutral_guarding_ranged"
		},

		negotiation_data =
		{
			behaviour = table.extend( BANDIT_NEGOTIATION )
			{
			}
		},
		death_item = "makeshift_dagger",
	})
)

Content.AddCharacterDef
(
	CharacterDef("BANDIT_RAIDER",
	{
		designer_note = "They cleave you, making bleeds last forever.",
		base_def = "BANDIT_BASE",
		title = "Raider",
		renown = 2,
		combat_strength = 4,
		tags = { "enforcer" },

		loved_graft = "regimented",
		hated_graft = "brutality",

    	death_money = DEATH_MONEY_HIGH,

		base_builds = {
			[ GENDER.MALE ] = "male_bandit_03",
			[ GENDER.FEMALE ] = "female_bandit_03"
		},

		voice_actors = {
			[GENDER.MALE] = { "spreeLowClassMale01" },
    		[GENDER.FEMALE] = { "spreeLowClassFemale01" },
    	},

		combat_anims = { "anim/med_combat_sickle_bandit_raider.zip"},
		anims = {"anim/weapon_sickle_bandit.zip"},
		
		fight_data = 
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_HIGH,

			OnResolveBattle = function( fighter, repercussions )
				if fighter:GetStatus() == FIGHT_STATUS.DEAD or fighter:GetStatus() == FIGHT_STATUS.SURRENDER then
					repercussions.loot.stolen_money = (repercussions.loot.stolen_money or 0) + (fighter.money_stolen or 0)
				end
			end,

			anim_mapping =
            {
                riposte = "crush"
            },

  			conditions = 
  			{
			    cleaved_player = 
			    {
			        name = "Cleaved",
			        desc = "{BLEED} cards no longer {EXPEND}.",

			        ctype = CTYPE.DEBUFF,
			    },

			    cleaved = 
			    {
			        name = "Cleaved",
			        desc = "{BLEED} no longer decays.",
			        ctype = CTYPE.DEBUFF,
			    },

  			},

  			attacks = 
  			{
			    spinning_scythe = table.extend(NPC_ATTACK)
			    {
			        name = "Raid",
			        anim = "thump",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        base_damage = 2,
			        hit_count = 3,

			        features = 
			        {
			        	BLEED = 3
			        }
			    },

			    raid = table.extend(NPC_ATTACK)
			    {
			        name = "Raid",
			        anim = "crush",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        damage_mult = 1,
			        money_stolen = 10,
			        quip = "bandit_raid",
			        loc_strings =
			        {
			        	SHILLS_STOLEN =  "{1#money} stolen!"
				    },

			        OnInit = function( self )
			        	self.money_stolen = 10 + self.owner:GetScale() * 5
			        end,

			        OnPostResolve = function( self, battle, attack )
			            for i, hit in attack:Hits() do
			                if not hit.evaded and hit.target:IsPlayer() and TheGame:GetGameState():GetCaravan():GetMoney() > 0 then
			                	local money_stolen = math.min( TheGame:GetGameState():GetCaravan():GetMoney(), self.money_stolen )
			                	TheGame:GetGameState():GetCaravan():PayMoney( money_stolen, self.owner.agent )
			                    self.owner:BroadcastEvent( BATTLE_EVENT.PLAY_FX, "steal_money" )
			                    local txt = loc.format( self.def:GetLocalizedString( "SHILLS_STOLEN" ), money_stolen )
			                    hit.target:BroadcastEvent( BATTLE_EVENT.PLAY_FLOATER, txt, 0.8, UICOLOURS.MONEY )
			                    self.owner.money_stolen = (self.owner.money_stolen or 0) + money_stolen
			                end
			            end

			        end,
			    },

			    cleave = table.extend(NPC_ATTACK)
			    {
			        name = "Cleave",
			        anim = "crush",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

			        CanPlayCard = function(self, battle, target)
			        	if target then
				        	return target:GetConditionStacks("cleaved") == 0 and target:GetConditionStacks("cleaved_player") == 0
						else
				           	return true
				        end
			        end,

			        OnPostResolve = function( self, battle, attack)
						if not attack:CheckHitResults( "evaded" ) then
				        	if attack.target.player_controlled then
				        		attack.target:AddCondition("cleaved_player", 1)
				        	else
				        		attack.target:AddCondition("cleaved", 1)
				        	end
				        end
			        end,
			    },

			    blood_thirst = table.extend(NPC_BUFF)
			    {
			        name = "Blood Thirst",
			        anim = "taunt",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
			        target_type = TARGET_TYPE.SELF,

			        OnPostResolve = function(self, battle, attack)
			            for i, fighter in self.owner:GetTeam():Fighters() do
			                fighter:AddCondition("POWER", 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
			            end
			        end,
			    },
  			},

			behaviour = 
			{
				OnActivate = function( self )
					self.cleave = self:AddCard( "cleave" )
					self.blood_thirst = self:AddCard( "blood_thirst" )
					self.intensify_count = math.random( 0, 2 )

					self.attacks = self:MakePicker()
						:AddCard( self.cleave, 2 )
						:AddID( "raid", 2 )
						:AddID( "spinning_scythe", 2 )

					self:SetPattern( self.Start )
				end,

				Start = function( self )				
					self:ChooseCard( self.cleave )
					self:SetPattern( self.Attack )
				end,

				Attack = function( self )
					self.intensify_count = self.intensify_count + 1
					if self.intensify_count >= 4 then
						self:ChooseCard( self.blood_thirst )
						self.intensify_count = 0
					else
						self.attacks:ChooseCard()
					end
				end,
			},
		},
		
		idle_anims = 
		{
			guarding = "idle_neutral_guarding_ranged"
		},

		negotiation_data =
		{
			behaviour = BANDIT_NEGOTIATION
		},
		death_item = "heavy_cleaver",
	})
)

Content.AddCharacterDef
(
	CharacterDef("BANDIT_CAPTAIN",
	{
		base_def = "BANDIT_BASE",
		title = "Captain",

		renown = 4,
		combat_strength = 5,

		loved_graft = "savagery",
		hated_graft = "terrorized",
		death_item = "grisly_trophy",

		base_builds = {
			[ GENDER.MALE ] = "male_jarackle_bandit_02",
			[ GENDER.FEMALE ] = "female_jarackle_bandit_02"
		},

		voice_actors = {
			[GENDER.MALE] = { "spreeLowClassMale01" },
    		[GENDER.FEMALE] = { "spreeLowClassFemale01" },
    	},

		combat_anims = { "anim/med_combat_club_bandit_captain.zip"},
		anims = {"anim/weapon_club_bandit_captain.zip"},

		fight_data = 
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.HIGH,

			status_widget_head_dx = 0.0,
			status_widget_head_dy = 4.6,

			anim_mapping =
			{
				riposte = "thump",
			},

			cards = { "double_whammy", "bandit_captain_thump" },
			
			conditions =
			{
			    vicious =
			    {
			        name = "Vicious",
			        desc = "Gain <#HILITE>{1} {RIPOSTE}</> when attacked.",
			        desc_fn = function( self, fmt_str )
			        	return loc.format( fmt_str, self.stacks )
		        	end,
			        
			        event_priorities =
			        {
			        	[ BATTLE_EVENT.END_RESOLVE ] = 99, -- AFTER riposte.
				    },
			        event_handlers =
			        {
			            [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
			                if hit.target == self.owner and attack.card:IsAttackCard() then
			                	self.hits = (self.hits or 0) + 1
				            end
			            end,

			            [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, resolve_card )
			            	if self.hits then
			            		self.owner:AddCondition( "RIPOSTE", self.stacks * self.hits )
			            		self.hits = nil
			            	end
			            end,
				    }
			    },
			},

			behaviour = {
				--DIALOG_RALLY = [[Backup!]],

				OnActivate = function( self )
					self.fighter:AddCondition( "vicious", 2 + 1 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
					self.attack = self:MakePicker()
						:AddID( "double_whammy", 2 )
						:AddID( "bandit_captain_thump", 1)
						:AddID( "ai_defend_lo", 1 )
					--self.summon_goon = self:AddCard( "summon_goon" )
					self:SetPattern( self.Attack )
				end,


				Attack = function( self, fighter, battle )
					return self.attack:ChooseCard()
				end,
			},
		},

		idle_anims = 
		{
			guarding = "idle_neutral_guarding"
		},

		negotiation_data =
		{
			behaviour = table.extend( BANDIT_NEGOTIATION )
			{
				KINGPIN = 1,
			}
		},
	})
)

