local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS
local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

----------------------------------------------------
-- Member of the Cult of Hesh.

Content.AddCharacterDef
(
	CharacterDef("HESH_BASE",
	{
		base_def = "NPC_BASE",
		is_template = true,

		tags = {"heshian"},
		renown = 1,

		faction_id = "CULT_OF_HESH",
		support_negotiation = "hesh_support",

		death_item = "heshian_amulet",
	})
)

local HESH_NEGOTIATION =
{
	OnInit = function( self, difficulty )
		self.tithe = self:AddArgument( "TITHE" )
		self.tithe.stacks = 2

		self.prayer = self:AddArgument( "prayer_of_hesh" )
		self.wrath = self:AddArgument( "wrath_of_hesh" )

		self.negotiator:AddModifier( "ZEAL" )

		self:SetPattern( self.Cycle )
	end,

	Cycle = function( self, turns )
		if turns == 1 then
			self:ChooseCard( self.wrath )
			self:ChooseCard( self.prayer )
		else
			if turns < 12 and turns % 3 == 0 then
				self:ChooseCard( self.tithe )

			elseif turns < 12 and turns % 4 == 0 then
				self:ChooseGrowingNumbers( 1, 1 )

			elseif turns % 3 == 0 or turns % 4 == 0 then
				self:ChooseGrowingNumbers( 1, 1, 2 )

			else
				local wraths = self.negotiator:GetModifierInstances( "wrath_of_hesh" )
				local prayers = self.negotiator:GetModifierInstances( "prayer_of_hesh" ) + self.negotiator:GetModifierInstances( "prayer_of_hesh_plus" )
				if wraths + prayers >= 2 + self.difficulty then
					self:ChooseGrowingNumbers( 1, 1 )

				elseif wraths >= prayers then
					self:ChooseCard( self.wrath )

				else
					self:ChooseCard( self.prayer )
				end
			end
		end
	end,
}

----------------------------------------------------
-- Work positions.


local PRIEST_FIGHT_BEHAVIOUR =
{
	elite = false,
	OnActivate = function( self )

		self.attacks = self:MakePicker()
			:AddID( "poke", 1 )
			:AddID( "denounce", 1 )
			:AddID( "annhiliation", 4 )

		self.elite_attacks = self:MakePicker()
			:AddID( "priest_slash", 1)
			:AddID( "priest_uppercut", 2)

		self.elite_taunts = self:MakePicker()
			:AddID("annhiliation", 1)
			:AddID( "ai_defend_med", 1)

		self:SetPattern( self.elite and self.EliteCycle or self.Cycle )

		self.fighter:AddCondition( "favorite_of_hesh" )
	end,

	Cycle = function( self )
		self.attacks:ChooseCard()
	end,

	EliteCycle = function( self )
		if self.battle:GetTurns() % 3 == 0 then
			self.elite_taunts:ChooseCard()
		else
			self.elite_attacks:ChooseCard()
		end
	end
}


Content.AddCharacterDef
(
	CharacterDef("PRIEST",
	{
		designer_note = "Curses you, making you expend cards. If you kill them, they demoralize your allies.",
		base_def = "HESH_BASE",

		title = "Priest",
		renown = 4,
		combat_strength = 2.5,

		loved_graft = "bio_feedback",
		hated_graft = "bad_faith",

		promotion_def = "PRIEST_PROMOTED",

		base_builds = {
			[ GENDER.MALE ] = "priest_male",
			[ GENDER.FEMALE ] = "priest_female"
		},

		social_boons =
		{
			[ RELATIONSHIP.LIKED ] = { HEALING = 2, SOCIALIZE = 1, UPGRADE_NEGOTIATION_CARD = 1 },
		},

		death_item = "heshian_amulet",

		fight_data = 
		{

			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.LOW,

	        status_widget_head_dx = 0,
			status_widget_head_dy = 5.0,

			cards = { "poke", "annhiliation" },

			behaviour = table.extend(PRIEST_FIGHT_BEHAVIOUR)
			{
			},

			conditions = {
	 			favorite_of_hesh = {
	 				name = "Favored by Hesh",
	 				desc = "All {SURRENDER} meters increase by <#HILITE>{1}</> upon this fighter's death.\nThis amount increases by {2} every turn.",
	 				desc_fn = function( self, fmt_str )
	 					return loc.format(fmt_str, self.stacks, self.inc)
	 				end,
	                inc = 2,
	                event_handlers =
	                {
	                    [BATTLE_EVENT.END_TURN] = function(self, fighter)
	                    	if self.owner == fighter then
	                    		fighter:AddCondition("favorite_of_hesh", self.inc)
	                    	end
	                    end,

						[ BATTLE_EVENT.FIGHTER_KILLED ] = function ( self, fighter )
			                if self.owner == fighter then
			                	for _, other_fighter in pairs( fighter.battle:GetTeam(TEAM.RED).fighters ) do
			                		if other_fighter.agent:IsSentient() and other_fighter ~= fighter then
			                			other_fighter:ApplyMoraleDamage(self.stacks)
			                		end
			                	end

			                	for _, other_fighter in pairs( fighter.battle:GetTeam(TEAM.BLUE).fighters ) do
			                		if other_fighter.agent:IsSentient() and other_fighter ~= fighter then
			                			other_fighter:ApplyMoraleDamage(self.stacks)
			                		end
			                	end
			                end
						end,
	                }	 				
	 			},


	 			annihilation =
	            {
	                name = "Annihilation",
	                desc = "The next <#HILITE>{1*card| {1} cards}</> played will {EXPEND}.", 
	                desc_fn = function( self, fmt_str )
	                	return loc.format( fmt_str, self.stacks )
	                end,
	                hud_fx = {"annihilation"},

	                ctype = CTYPE.DEBUFF,

	                event_handlers =
	                {
			            [BATTLE_EVENT.POST_RESOLVE] = function( self, battle, attack )
			                if attack.card.owner == self.owner then
			                	attack.card:SetFlags( CARD_FLAGS.EXPEND )
		                		self.owner:RemoveCondition( self.id, 1 )
			                end
						end,
	                }
	            },				
			},
			attacks = 
			{
				poke = table.extend(NPC_ATTACK)
			    {
			        name = "Poke",
			        anim = "poke",
			        flags = CARD_FLAGS.MELEE,
			    },
			
			    annhiliation = table.extend(NPC_BUFF)
			    {
			        name = "Promise of Annihilation",
			        anim = "taunt",

			        quip = "hesh_annhiliation",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.DEBUFF,
			        target_type = TARGET_TYPE.ENEMY,

                	CanPlayCard = function(self, battle, target)
                		if target then
		                    return (target.player_controlled and not target:HasCondition("annihilation"))
		                else
		                	return true
		                end
                	end,

			        OnPostResolve = function( self, battle, attack)
		            	attack.target:AddCondition("annihilation", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
			        end,
			    },

			    denounce = table.extend(NPC_BUFF)
			    {
			        name = "Doomsayer",
			        anim = "taunt",
					morale_damage = 8,
			        quip = "hesh_denounce",
			        flags = CARD_FLAGS.SKILL | CARD_FLAGS.DEBUFF,
			        target_type = TARGET_TYPE.ENEMY,

                	CanPlayCard = function(self, battle, target)
                		if target then
		                    return (target:HasMorale() and target.agent:IsSentient())
		                else
		                	return true
		                end
                	end,

			        OnPostResolve = function( self, battle, attack)
						attack.target:ApplyMoraleDamage(self.morale_damage)
			        end,
			    },

			},

		},
		

		anims = {"anim/weapon_staff_priest.zip"},
		combat_anims = {"anim/med_combat_staff_priest.zip"},

		negotiation_data = 
		{
			behaviour = table.extend( HESH_NEGOTIATION )
			{
			}
		},    
	})
)

Content.AddCharacterDef
(
	CharacterDef("PRIEST_PROMOTED",
	{
		base_def = "PRIEST",
		title = "Promoted Priest",

		death_item = "heshian_amulet_plus",

		 base_builds = {
		 	[ GENDER.MALE ] = "male_priest_promoted_build",
		 	[ GENDER.FEMALE ] = "female_priest_promoted_build"
		 },

		fight_data = 
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,
			cards = {  },

			attacks = 
			{
				priest_slash = table.extend(NPC_ATTACK)
				{
					name = "Slash",
					anim = "attack1",

					flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

					features = 
					{
						WOUND = 1,
					},
				},

				priest_uppercut = table.extend(NPC_ATTACK)
				{
					name = "Slice and Dice",
					anim = "attack2",

					flags = CARD_FLAGS.MELEE,
					target_count = 2,
					damage_mult = 0.7,
				},
			},

			behaviour = table.extend(PRIEST_FIGHT_BEHAVIOUR)
			{
				elite = true,
			},
		},

		anims = {"anim/weapon_scythe_priest_promoted.zip"},
		combat_anims = { "anim/med_combat_scythe_promoted_priest.zip" },
	})
)


Content.AddCharacterDef
(
	CharacterDef("LUMINARI",
	{
		designer_note = "Gains defend every turn, is immune to wound / bleed, and gives you a replicating curse card.",
		base_def = "HESH_BASE",

		title = "Luminari",
		renown = 3,
		combat_strength = 5,
		death_item = "heshian_mask",

		base_builds = {
			[ GENDER.MALE ] = "luminari_male",
			[ GENDER.FEMALE ] = "luminari_female"
		},

		loved_graft = "bell_of_hesh",
		hated_graft = "curse_of_hesh",

		social_boons =
		{
			[ RELATIONSHIP.LIKED ] = { UPGRADE_BATTLE_CARD = 1 },
			[ RELATIONSHIP.LOVED ] = { HIRE = 1 },
		},

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.VERY_HIGH,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.HIGH,
			
			formation = FIGHTER_FORMATION.BACK_Z,

			cards = { "ai_luminari_slice", "ai_luminari_swing" },

			attacks = {
			    ai_luminari_slice = table.extend( NPC_ATTACK )
			    {
			        name = "Lumin Slice",
			        anim = "slice",
			        flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,
			        damage_mult = .5,

			        OnPostResolve = function( self, battle, attack)
	                    for i,hit in ipairs(attack.hits) do
	                    	if not hit.evaded then
	                    		if hit.target.player_controlled then
		                    		local incepted_card = Battle.Card( "status_lumin_burn", battle:GetPlayerFighter())
		                    		battle:DealCards( {incepted_card}, battle:GetDrawDeck() )
		                    	else
		                    		hit.target:AddCondition("WOUND", 2, self)
		                    	end
	                    	end
	                    end
	                    
			        end,
			    },
			    ai_luminari_swing = table.extend( NPC_ATTACK )
			    {
			        name = "Hesh's Scythe",
			        anim = "swing",
			        flags = CARD_FLAGS.MELEE,
			    },
			},
			behaviour = {
				OnActivate = function( self, fighter )
					
					self.fighter:AddCondition( "METALLIC" )
					self.fighter:AddCondition( "ARMOURED", 4 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
					
					self.attacks = self:MakePicker()
						:AddID( "ai_luminari_swing", 2 )
						:AddID( "ai_defend_med", 1)
						
					
					self.insert_attacks = self:MakePicker()
						:AddID( "ai_luminari_slice", 1 )

					self:SetPattern( self.Attack )
					self.insert_counter = 0
				end,

				Attack = function( self )
					if self.insert_counter <= 0 then
						self.insert_attacks:ChooseCard()
						self.insert_counter = math.random(2,3)
					else
						self.insert_counter = self.insert_counter - 1
						self.attacks:ChooseCard()
					end
				end,
			},

			battle_scale = 1.1,
		},

		anims = {"anim/weapon_scythe_luminari.zip"},
		combat_anims = {"anim/med_combat_scythe_luminari.zip"},
	
		tags = {  "luminari", "enforcer" },

		negotiation_data = 
		{
			behaviour = table.extend( HESH_NEGOTIATION )
			{
			}
		},    

	})
)

Content.AddBattleCard( "lumin_kunai", 
{
    name = "Lumin Kunai",
    anim = "taunt",
    desc = "At the end of your turn, take 2 damage per attack card played this turn.\n{1}",
    flavour = "'Ahh! Get this thing outta me!'",
    desc_fn = function( self, fmt_str )
        if self.engine then
            local str = loc.format( LOC"CARD_ENGINE.CARDS_PLAYED", self:CountCardsPlayed())
            return loc.format( fmt_str, str )
        else
            return loc.format( fmt_str, "" )
        end
    end,

    series = CARD_SERIES.GENERAL,
    rarity = CARD_RARITY.UNIQUE,
    cost = 0,
    target_type = TARGET_TYPE.SELF,

    flags = CARD_FLAGS.STATUS | CARD_FLAGS.BURNOUT | CARD_FLAGS.UNPLAYABLE,

    CountCardsPlayed = function( self )
        if not self.engine then
            return 0
        end
        return self.engine:CountCardsPlayed( function(card) return card:IsAttackCard() end )
    end,

  	event_handlers =
    {
    	[ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
    		local cards_played = self:CountCardsPlayed()
    		if self.deck == battle:GetHandDeck() and cards_played > 0 then
    			self.owner:ApplyDamage(cards_played * 2, self.owner, self)
    		end
    	end
    },
})

Content.AddCharacterDef
(
	CharacterDef("ZEALOT",
	{
		base_def = "HESH_BASE",

		title = "Zealot",
		renown = 3,
		combat_strength = 3,
		death_item = "lumin_darts",

		base_builds = {
			[ GENDER.MALE ] = "male_hesh_zealot",
			[ GENDER.FEMALE ] = "female_hesh_zealot"
		},

		loved_graft = "grace_of_hesh",
		hated_graft = "heshian_obscenity",

		-- social_boons =
		-- {
		-- 	[ RELATIONSHIP.LIKED ] = { UPGRADE_BATTLE_CARD = 1 },
		-- 	[ RELATIONSHIP.LOVED ] = { HIRE = 1 },
		-- },

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_HIGH,

			attacks = {
				zealot_slice = table.extend(NPC_MELEE)
				{
					name = "Slicing Strike",
					anim = "melee_attack",

					damage_mult = 1.5,

					flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

					features =
					{
						WOUND = 2
					},
				},

				zealot_throw = table.extend(NPC_RANGED)
				{
					name = "Fan of daggers",
					anim = "throw_knives",
					flags = CARD_FLAGS.RANGED,
					target_count = 2,
					damage_mult = 0.6,

					OnPostResolve = function( self, battle, attack )
						local player_hits = 0
						for i,hit in ipairs(attack.hits) do
							if hit.target:IsPlayer() and not hit.evaded then
								player_hits = player_hits + 1
							end
						end
						if player_hits > 0 then
							local cards = {}
							for i=1,player_hits do
								table.insert(cards, Battle.Card("lumin_kunai", battle:GetPlayerFighter()))
							end
							battle:DealCards(cards, battle:GetDrawDeck())
						end
					end
				},

				zealot_taunt = table.extend(NPC_BUFF)
				{
					name = "Taunt",
					anim = "taunt",
					target_type = TARGET_TYPE.SELF,
					flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,

					features =
					{
						DEFEND = 6,
						RIPOSTE = 3,
					}
				},
			},
			behaviour = {
				OnActivate = function( self, fighter )
					self.slice = self:AddCard("zealot_slice")
					self.throw = self:AddCard("zealot_throw")
					self.taunt = self:AddCard("zealot_taunt")

					self:SetPattern( self.Attack )
				end,

				Attack = function( self )
					local turns = self.battle:GetTurns()
					if turns > 0 and turns % 3 == 0 then
						return self:ChooseCard(self.taunt)
					end
					if turns % 2 == 0 then
						self:ChooseCard(self.slice)
					else
						self:ChooseCard(self.throw)
					end
				end,
			},

			anim_mapping =
			{
				riposte = "throw_knives",
			},
			ranged_riposte = true,

			battle_scale = 1,
		},

		anims = {"anim/weapon_throwing_knives_hesh_zealot.zip"},
		combat_anims = {"anim/med_combat_throving_knives_hesh_zealot.zip"},

		negotiation_data = 
		{
			behaviour = table.extend( HESH_NEGOTIATION )
			{
			}
		},    

	})
)





Content.AddCharacterDef
(
	CharacterDef("LUMINITIATE_PROMOTED",
	{
		
		base_def = "LUMINITIATE",
		title = "Promoted Luminitiate",
		renown = 1,
		combat_strength = 2,
		
		death_item = "luminitiate_coin",

		base_builds = {
			[ GENDER.MALE ] = "male_hesh_luminiciate_promoted_build",
			[ GENDER.FEMALE ] = "female_hesh_luminiciate_promoted_build"
		},		

		anims = {"anim/weapon_cane_hesh_luminiciate_promoted.zip"},
		combat_anims = {"anim/med_combat_cane_hesh_luminiciate_promoted.zip"},

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,

			anim_mapping =
			{
				riposte = "attack2",
			},
			ranged_riposte = true,

			conditions =
			{
				
			},

			attacks = {
				luminitiate_cane_slam = table.extend(NPC_MELEE)
				{
					name = "Slam",
					anim = "attack1",

					flags = CARD_FLAGS.MELEE,
					damage_mult = 1.3,
				},
				luminitiate_cane_shoot = table.extend(NPC_RANGED)
				{
					name = "Shoot",
					anim = "attack2",

					flags = CARD_FLAGS.RANGED | CARD_FLAGS.PIERCING,
					damage_mult = 0.4,
				},
				luminiate_cane_taunt = table.extend(NPC_BUFF)
				{
					name = "Taunt",
					anim = "taunt",
					flags = CARD_FLAGS.BUFF,
					target_type = TARGET_TYPE.SELF,

					OnPostResolve = function( self, battle, attack )
						self.owner:AddCondition("DEFEND", 6, self)
						self.owner:AddCondition("RIPOSTE", 3, self)
					end
				},
			},
			behaviour = {
				believer_stacks = { 2, 3, 4, 4 },
				OnActivate = function( self, fighter )
					self.attacks = self:MakePicker()
						:AddID("luminitiate_cane_slam", 2)
						:AddID("luminitiate_cane_shoot", 2)
						:AddID("luminiate_cane_taunt", 1)

					self:SetPattern( self.Attack )
					self.fighter:AddCondition("true_believer", self.believer_stacks[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ), #self.believer_stacks)])
				end,

				Attack = function( self )
					self.attacks:ChooseCard()
				end,
			},

 
		},
	})
)

Content.AddCharacterDef
(
	CharacterDef("LUMINITIATE",
	{
		
		base_def = "PRIEST",
		title = "Luminitiate",
		renown = 1,
		combat_strength = 1.5,

		loved_graft = "throng",
		hated_graft = "affiliation",
		death_item = "luminitiate_coin",

		promotion_def = "LUMINITIATE_PROMOTED",

		base_builds = {
			[ GENDER.MALE ] = "male_hesh_luminiciate_build",
			[ GENDER.FEMALE ] = "female_hesh_luminiciate_build"
		},

		social_boons =
		{
			[ RELATIONSHIP.LIKED ] = { HEALING = 2, SOCIALIZE = 1, UPGRADE_NEGOTIATION_CARD = 1 },
		},
		

		anims = {"anim/weapon_knife_hesh_luminiciate.zip"},
		combat_anims = {"anim/med_combat_knife_hesh_luminiciate.zip"},

		negotiation_data = 
	    {
	        modifiers =
	        {
	            DEVOTED =
	            {
	            	name = "Devoted",
	            	desc = "Whenever a {FAITH_IN_HESH} argument is destroyed, this argument restores {1} resolve.",
	            	desc_fn = function( self, fmt_str )
	            		return loc.format( fmt_str, self.restore_amt or 2)
	            	end,

	            	modifier_type = MODIFIER_TYPE.CORE,

	            	restore_table = { 2, 2, 3, 3, 4 },

	            	OnInit = function( self, source )
	            		local difficulty = self.engine and self.engine:GetDifficulty() or 1
	            		self.restore_amt = self.restore_table[difficulty]
	            	end,

	            	event_handlers =
	            	{
	            		[ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
		                    if source and modifier.negotiator == self.negotiator and modifier.id == "FAITH_IN_HESH" then
		                        self:RestoreResolve(self.restore_amt or 2, self)
		                    end
		                end
	            	},
	            },

	            FAITH_IN_HESH = 
	            {
	            	name = "Faith in Hesh",
			        damage_tt = "UI.NEGOTIATION_PANEL.ARGUMENT_RANDOM2_DAMAGE_TT", 

			        max_stacks = 1,
			        hide_stacks = true,
			        modifier_type = MODIFIER_TYPE.ARGUMENT,
			        target_enemy = TARGET_ANY_RESOLVE,
			        target_mod = TARGET_MOD.RANDOM2,

			        sound = "event:/sfx/battle/cards/neg/create_argument/wrath_of_hesh",

			        OnInit = function( self )
			            self:SetResolve( 2, MODIFIER_SCALING.LOW )
			        end,

			        min_persuasion = 2,
			        max_persuasion = 2,

			        OnBeginTurn = function( self, minigame )
			            self:ApplyPersuasion()
			        end,
	            },

	            LEVY =
	            {
	            	name = "Levy",
	            	desc = "If this is destroyed, {1} loses {4} shills and {2}'s core argument takes {3} damage. Can only be targeted by cards.",
	            	desc_fn = function( self, fmt_str )
	            		return loc.format( fmt_str, self:GetOpponentName(), self:GetOwnerName(), self.damage_amt, self.shill_amt)
	            	end,

	            	modifier_type = MODIFIER_TYPE.BOUNTY,

	            	max_resolve = 1,

	            	CanPlayCard = function( self, card, engine, target )
			            if target == self and (card.negotiator ~= self.negotiator or card:IsAttack()) then
			                return is_instance(card, Negotiation.Card)
			            end
			            return true
			        end,

	            	damage_amt = 5,
	            	damage_table = { 5, 7, 10, 14, 19 },
	            	shill_amt = 15,
	            	shill_table = { 15, 20, 25, 30, 35 },

	            	OnInit = function( self, source )
			            local difficulty = self.engine and self.engine:GetDifficulty() or 1
			            self.damage_amt = self.damage_table[difficulty]
			            self.shill_amt = self.shill_table[difficulty]
			        end,

			        OnBounty = function( self, card )
			        	self.negotiator:AttackResolve(self.damage_amt, self)
			        	self.engine:ModifyMoney(-self.shill_amt, self)
			        end
	            },
	        },

	        behaviour = 
	        {
	            OnInit = function( self, difficulty )
	                self.faith = self:AddArgument("FAITH_IN_HESH")
	                self.levy = self:AddArgument("LEVY")

	                self.negotiator:AddModifier( "DEVOTED" )

	                self:SetPattern( self.Cycle )
	            end,

	            Cycle = function( self, turns )
                    self:ChooseGrowingNumbers( 1, 1, 1, 2 )
	                if (turns - 1) % 2 == 0 then
	                    self:ChooseComposure( 2, 1, 3 )
	                end
	                if turns % 3 == 0 and self.negotiator:GetModifierInstances("LEVY") < 2 then
	                	self:ChooseCard(self.levy)
	                end
	                if turns % 2 == 0 then
	                    self:ChooseCard(self.faith)
	                end
	            end,
	        }
	    },

		fight_data =
		{
			MAX_MORALE = MAX_MORALE_LOOKUP.LOW,
			MAX_HEALTH = MAX_HEALTH_LOOKUP.LOW,

			conditions =
			{
				true_believer =
				{
					name = "True Believer",
					desc = "Whenever {1} deals unmitigated damage, they gain <#HILITE>{2}</> {POWER}. Whenever their attacks are fully defended or evaded, {1} loses all {POWER} gained from <b>True Believer</>.",
					desc_fn = function( self, fmt_str )
						return loc.format(fmt_str, self:GetOwnerName(), self.stacks)
					end,

					event_handlers =
					{
						[ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
			                if attack.attacker == self.owner and attack.card:IsDamageCard() then
				                if not hit:CheckHitResult( "defended" ) and not hit:CheckHitResult( "evaded" )  then
				                    self.owner:AddCondition( "POWER", self.stacks, self )
				                    self.power_added = (self.power_added or 0) + self.stacks
				                else
				                	if self.power_added and self.power_added > 0 then
				                		self.owner:RemoveCondition("POWER", self.power_added, self)
				                		self.power_added = 0
				                	end
				                end
				            end
			            end,
					},
				},
			},

			attacks = {
				luminitiate_slash = table.extend(NPC_MELEE)
				{
					name = "Slash",
					anim = "slash",

					damage_mult = 1.2,

					flags = CARD_FLAGS.MELEE,
				},

				luminitiate_double_stab = table.extend(NPC_MELEE)
				{
					name = "Double Stab",
					anim = "double_stab",
					flags = CARD_FLAGS.MELEE | CARD_FLAGS.PIERCING,
					target_count = 2,
					base_damage = 1,
				},
			},
			behaviour = {
				believer_stacks = { 1, 2, 3, 3 },
				OnActivate = function( self, fighter )
					self.attacks = self:MakePicker()
						:AddID("luminitiate_slash", 2)
						:AddID("luminitiate_double_stab", 1)
						:AddID("ai_defend_med", 1)

					self:SetPattern( self.Attack )
					self.fighter:AddCondition("true_believer", self.believer_stacks[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ), #self.believer_stacks)])
				end,

				Attack = function( self )
					self.attacks:ChooseCard()
				end,
			},

 
		}, 
	})
)