local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS
local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local def = CharacterDef("EDEN",
{
    unique = true,
    base_def = "ADMIRALTY_PATROL_LEADER",
    alias = "EDEN",
    species = SPECIES.HUMAN,
    gender = GENDER.FEMALE,
    renown = 4,
    combat_strength = 3,
    name = "Eden",
    tags = {"plot_armour", "no_patron"},    
    
    loved_graft = "in_on_it",
    hated_graft = "undermined",
    death_item = "arachni_blade",
    faction_id = "ADMIRALTY",
    title = "Regional Commander",
    theme_music = "event:/music/story/smith/story_smith_eden",
    -- voice_actor = "",
    
    build = "female_eden_yllonwa_build",
    head = "head_female_eden_yllonwa_promoted",

    anims = { "anim/weapon_gun_edan.zip", },
    combat_anims = { "anim/med_combat_edan.zip" },

    negotiation_data = 
    {
        modifiers =
        {
            WIDE_INFLUENCE =
            {
                name = "Wide Influence",
                desc = "At the end of {1}'s turn, incept either a {straw_man} or {PLANTED_EVIDENCE}, chosen randomly, if neither exists.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOwnerName() )
                end,

                modifier_type = MODIFIER_TYPE.CORE,

                event_handlers =
                {
                    [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                        if negotiator == self.negotiator then
                            local has_inception = self.anti_negotiator:HasModifier("straw_man") or self.anti_negotiator:HasModifier("PLANTED_EVIDENCE") or false
                            if not has_inception then
                                if math.random() < 0.5 then
                                    self.anti_negotiator:InceptModifier("PLANTED_EVIDENCE", 5, self)
                                else
                                    self.anti_negotiator:InceptModifier("straw_man", 1, self)
                                end
                            end
                        end
                    end,
                }
            },
        },

        behaviour = 
        {
            OnInit = function( self, difficulty )
                local difficulty = self.engine and self.engine:GetDifficulty() or 1
                self.all_business = self:AddArgument( "ALL_BUSINESS" )
                self.interrogate = self:AddArgument( "INTERROGATE" )

                self.negotiator:AddModifier( "WIDE_INFLUENCE" )

                self:SetPattern( self.Cycle )
            end,

            Cycle = function( self, turns )
                if turns % 3 == 0 then
                    self:ChooseGrowingNumbers( 3, 1, 0.5 )
                else
                    self:ChooseGrowingNumbers( 1, 1 )
                end
                if (turns - 1) % 3 == 0 then
                    self:ChooseComposure( 1, 2, 5 )
                end
                if turns % 2 == 0 then
                    self:ChooseCard(self.interrogate)
                end
                if turns % 6 == 0 and not self.negotiator:HasModifier( "ALL_BUSINESS" ) then
                    self:ChooseCard( self.all_business )
                end
            end,
        }
    },

    fight_data = 
    {
        MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
        MAX_HEALTH = 85,

        anim_mapping =
        {
            run_forward_pre = "run_forward_entry",
            run_forward = "run_forward_entry",
            run_forward_pst = "run_forward_entry_pst",
        },

        post_entry_anim_mapping = {},

        anim_transitions =
        {
            run_forward_entry_pst = 
            {
                idle = "idle_pre",
            },
        },

        conditions =
        {
            energized =
            {
                name = "Energized",
                desc = "{1}'s next attack applies {2} {lumin_burnt} if it causes health loss.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOwnerName(), self.lumin_burn_amt + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                end,

                lumin_burn_amt = 3,

                event_handlers =
                {
                    [ BATTLE_EVENT.ON_HIT] = function( self, battle, attack, hit )
                        if attack.attacker == self.owner and attack.card:IsAttackCard() and not hit.evaded and not hit.defended then
                            hit.target:AddCondition("lumin_burnt", self.lumin_burn_amt + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH))
                        end
                    end,

                    [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, resolve_card )
                        if resolve_card.owner == self.owner and resolve_card:IsAttackCard() and resolve_card ~= self.ignore_card then
                            self.owner:RemoveCondition("energized", self.stacks, self)
                        end
                    end,
                },
            },
        },

        attacks = 
        {
            eden_shoot = table.extend(NPC_ATTACK)
            {
                name = "Shoot",
                anim = "shoot",
                flags = CARD_FLAGS.RANGED,

                CanPlayCard = function ( self, battle, target )
                    return self.owner:HasCondition("energized")
                end,

                damage_mult = 1.2,
            },

            eden_attack = table.extend(NPC_ATTACK)
            {
                name = "Attack",
                anim = "attack1",
                flags = CARD_FLAGS.MELEE,

                CanPlayCard = function ( self, battle, target )
                    return not self.owner:HasCondition("energized")
                end,

                damage_mult = 0.7,
            },

            eden_taunt = table.extend(NPC_BUFF)
            {
                name = "Gather",
                anim = "taunt",

                flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
                target_type = TARGET_TYPE.SELF,
                CanPlayCard = function ( self, battle, target )
                    return not self.owner:HasCondition("energized")
                end,

                OnPostResolve = function( self, battle, attack )
                    self.owner:AddCondition("DEFEND", 5, self)
                    local con = self.owner:AddCondition("energized", 1, self)
                    con.ignore_card = attack.card
                end,
            },

            eden_anim_switch = 
            {
                OnPostResolve = function( self, battle, attack )
                    self.owner:BroadcastEvent( BATTLE_EVENT.FIGHTER_REMAP_ANIMS, self.owner.agent.fight_data.post_entry_anim_mapping )
                end
            },
        
        },

        behaviour =
        {
            OnActivate = function( self, fighter )
                self.attacks = self:MakePicker()
                    :AddID("eden_shoot", 1)
                    :AddID("eden_attack", 3)
                    :AddID("eden_taunt", 1)

                self:SetPattern(self.Cycle)
                self.anim_swap = self:AddCard("eden_anim_switch")
                self:Present( function( screen, ent )
                    fighter.battle:PlayCard(self.anim_swap)
                end )
            end,


            Cycle = function( self )
                self.attacks:ChooseCard()
            end
        },
    },
})

Content.AddCharacterDef( def )
    