local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS
local EVENT = battle_defs.BATTLE_EVENT
---------------------------------------------------

Content.AddBattleCard("flead_larvae_card",
{
    name = "Tiny Stinger",
    anim = "throw",
    flavour = "'Small, but deadly.'",
    desc = "Remove all {DEFEND} on target.",

    flags = CARD_FLAGS.EXPEND | CARD_FLAGS.RANGED | CARD_FLAGS.ITEM | CARD_FLAGS.STICKY,
    series = CARD_SERIES.NPC,
    rarity = CARD_RARITY.UNIQUE,

    min_damage = 4,
    max_damage = 6,
    cost = 0,

    features =
    {
        IMPAIR = 1,
    },

    OnPreResolve = function( self, battle, attack )
        for i, hit in attack:Hits() do
            hit.target:RemoveCondition("DEFEND", nil, self)
        end
    end,
})


Content.AddCharacterDef
(
    CharacterDef("FLEAD_LARVA",
    {
        base_def = "MONSTER",
        title = "Flead Larva",
        build = "flead_larvae",
        combat_strength = .25,

        fight_data =
        {
            MAX_HEALTH = 16,

            status_widget_dx = -0.0,
            status_widget_dy = -0.4,
            status_widget_head_dx = 0.0,
            status_widget_head_dy = 0.5,

            OnJoinBattle = function( fighter, anim_fighter )
                local parent_ent = anim_fighter.sim.mgr:GetAnimFighterByFighter( fighter.parent )
                if parent_ent then
                    anim_fighter.dropshadow:Hide()

                    local x, y, z = parent_ent.cmp.AnimController:GetSymbolWorldPos( "head_swap", 0, 0, 0 )
                    local x0, y0, z0 = parent_ent:GetLocalPosition()
                    anim_fighter:SetPos( x - 1.2, z0 - 0.01 )

                    anim_fighter.entity.cmp.AnimController:Play("spawn")
                    anim_fighter:WaitAnim()
                end
            end,

            conditions =
            {

                run_away_flead_larva = {
                    name = "Quitter",
                    desc = "Will run away if the Flead Queen is killed.",
                    icon = "battle/conditions/quitter.tex",
                    event_handlers = 
                    {
                        [ EVENT.FIGHTER_KILLED ] = function( self, fighter )
                            if fighter:GetTeam() and fighter:GetTeam() == self.owner:GetTeam() and fighter.agent:GetContentID() == "FLEAD_QUEEN" then
                                self.owner:Flee()
                            end
                        end,
                    },
                },

                tiny_stinger =
                {
                    name = "Tiny Stinger",
                    desc = "When this larva dies, insert {flead_larvae_card} into your hand.",
                    hide_gained = true,

                    event_handlers =
                    {
                        [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                            if status == FIGHT_STATUS.DEAD and fighter == self.owner then
                                self.battle:InsertCard( "flead_larvae_card" )
                            end
                        end,
                    }
                }
            },

            attacks =
            {
                tiny_fangs = table.extend(NPC_ATTACK)
                {
                    name = "Tiny Fangs",
                    anim = "attack1",

                    damage_mult = 0.5,
                    flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

                    features =
                    {
                        WOUND = 2,
                    }
                }
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.moves = self:MakePicker()
                        :AddID( "tiny_fangs", 1 )
                        :AddID( "ai_defend_lo", 1 )

                    self.fighter:AddCondition( "tiny_stinger" )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.moves:ChooseCard( 1 )
                end,

            }
        },
    })
)

local QUEEN_FORMATION =
{
    { 7.0, 0.0, FIGHTER_FORMATION.FRONT_X },  -- centre: always queen

    { 3.0, -3, nil }, -- top-left
    { 1.0, 1, nil }, -- left
    { 4.0, 5, nil }, -- bottom-left

    { 11.0, -3, nil }, -- top-right
    { 13.0, 3, nil }, -- right
    { 10.0, 7, nil }, -- bottom-right
}

Content.AddCharacterDef
(
    CharacterDef("FLEAD_QUEEN",
    {
        base_def = "MONSTER",
        title = "Brooding Matron",
        name = "Flead Queen",
        build = "flead_queen",
        boss = true,
        gender = GENDER.FEMALE,
        battle_preview_anim = "anim/boss_flead_slide.zip",
        battle_preview_offset = { x = -90, y = 90 },
        battle_preview_glow = { colour = 0x8BCAD7FF, bloom = 0.15, threshold = 0.02 },
        battle_preview_audio = "event:/ui/prebattle_overlay/prebattle_overlay_whooshin_boss_fleadqueen",
        icon = engine.asset.Texture( "icons/b_flead_queen.tex" ),
        death_item = "giant_stinger",
        compendium_offset = { -200, -60 },

        fight_data =
        {
            battle_scale = 1.25,
            MAX_HEALTH = 140,
            formation = FIGHTER_FORMATION.FRONT_X,

            status_widget_head_dx = 0.25,
            status_widget_head_dy = 5.5,
            shadow_scale = 10.5,

            attacks =
            {
                flead_queen_sting = table.extend(NPC_ATTACK)
                {
                    name = "Sting",
                    anim = "sting",
                    flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,
                    base_damage = {9, 9, 11, 13 },
                    wound_amt = { 2, 2, 3, 3},
                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition("WOUND", self.wound_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                    end
                },

                spawn_larvae =
                {
                    name = "Spawn Larvae",
                    anim = "larvae_spew",
                    flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,

                    target_type = TARGET_TYPE.SELF,

                    CanPlayCard = function( self, battle, target )
                        return self.owner:GetTeam():NumActiveFighters() < self.owner:GetTeam():GetMaxFighters(), CARD_PLAY_REASONS.TEAM_FULL
                    end,

                    OnPostResolve = function( self, battle, attack )
                        local to_summon = math.min( 3, self.owner:GetTeam():GetMaxFighters() - self.owner:GetTeam():NumActiveFighters() )
                        for k =1 , to_summon do
                            local agent = Agent( "FLEAD_LARVA" )
                            local new_fighter = Fighter.CreateFromAgent( agent, self.owner:GetScale() )
                            new_fighter.parent = self.owner
                            new_fighter:AddCondition("run_away_flead_larva")                            
                            self.owner:GetTeam():AddFighter( new_fighter )
                        end
                        self.owner:GetTeam():ActivateNewFighters( battle )
                    end,
                },

                harden_carapace = table.extend(NPC_BUFF)
                {
                    name = "Spiked Carapace",
                    anim = "taunt",
                    flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,

                    CanPlayCard = function( self, battle, target )
                        return self.owner:GetTeam():NumActiveFighters() > 1
                    end,

                    features =
                    {
                        DEFEND = 8,
                        RIPOSTE = 8,
                    }
                },

                spine_storm = table.extend( NPC_ATTACK )
                {
                    name = "Spine Storm",
                    anim = "spine_storm",
                    flags = CARD_FLAGS.RANGED,
                    preview_icon = "UI/fightscreen_aoe_damagepreview.tex",

                    hit_count = 3,
                    target_type = TARGET_TYPE.ENEMY,
                    target_mod = TARGET_MOD.TEAM,

                    base_damage = {4, 4, 6, 7 },
                }
            },

            behaviour =
            {
                CUSTOM_FIGHT_FORMATIONS =
                {
                    [1] = QUEEN_FORMATION,
                    [2] = QUEEN_FORMATION,
                    [3] = QUEEN_FORMATION,
                    [4] = QUEEN_FORMATION,
                    [5] = QUEEN_FORMATION,
                    [6] = QUEEN_FORMATION,
                    [7] = QUEEN_FORMATION,
                },

                OnActivate = function( self )
                    self.fighter:GetTeam():SetMaxFighters( 7 )
                    self.fighter:GetTeam():SetCustomFormation( self.CUSTOM_FIGHT_FORMATIONS )

                    self.attacks = self:MakePicker()
                        :AddID( "flead_queen_sting", 1 )
                        :AddID( "spawn_larvae", 1 )
                        :AddID( "harden_carapace", 1 )

                    self.storm = self:AddCard( "spine_storm" )

                    local chitinous_amt = {8, 10, 16, 18}
                    self.fighter:AddCondition( "chitinous", chitinous_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #chitinous_amt)] )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    if self.battle:GetTurns() % 5 == 1 then
                        self:ChooseCard( self.storm )
                    else
                        self.attacks:ChooseCard()
                    end
                end,
            },

            anim_mapping =
            {
                riposte = "sting",
            },
        },
    })
)
