local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS


local build = CreateCreatureBuild({
    dialog_build = "anim/med_grawkit_build.zip",
    dialog_anim = "anim/med_dial_grawkit.zip",
    combat_anims = { "anim/med_combat_grawkit.zip" },
    portrait_anim = "anim/med_portrait_grawkit.zip",
})

Content.AddCharacterBuild( "grawkit", build )

Content.AddCharacterDef
(
    CharacterDef("GRAWKIT",
    {
        base_def = "MONSTER",

        title = "Grawkit",
        build = "grawkit",

        combat_strength = 4,
        death_item = "carapace_fragment",

        fight_data = 
        {
            MAX_HEALTH = MAX_HEALTH_LOOKUP.HIGH * 1.5,
            MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,
            ranged_riposte = true,
            shadow_scale = 12.0,

            anim_mapping =
            {
                riposte = "charge",
            },

            attacks = {
                grawkit_charge = table.extend(NPC_MELEE)
                {
                    name = "Charge",
                    anim = "charge",

                    damage_mult = 1.25,

                    features =
                    {
                        STUN = 1,
                    },
                },

                stand_up =
                {
                    flags = CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,

                    OnPostResolve = function( self, battle, attack )
                        self.owner.behaviour:TransitionStanding()
                    end,
                },

                thorny_carapace =
                {
                    name = "Thorny Carapace",
                    anim = "taunt",

                    target_type = TARGET_TYPE.SELF,
                    flags = CARD_FLAGS.BUFF,

                    OnPostResolve = function( self, battle, attack )
                        local condition = self.owner:AddCondition( "RIPOSTE", self.owner:GetScale() + 1 )
                        self.owner:AddCondition( "POWER", self.owner:GetScale() - 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end,
                },

                belly_flop = table.extend(NPC_MELEE)
                {
                    name = "Belly Flop",
                    anim = "flop",
                    damage_variance = 0, -- Variance for big hits bad.
                    base_damage = { 14, 20, 24, 28, 32 },
                    damage_mult = 1,

                    preview_icon = "UI/fightscreen_aoe_damagepreview.tex",
                    target_mod = TARGET_MOD.TEAM,
                    flags = CARD_FLAGS.RANGED,

                    OnPostResolve = function( self, battle, attack )
                        self.owner.behaviour:TransitionGround()
                    end
                },
            },

            hit_tags = { "small" },

            behaviour =
            {
                OnActivate = function( self )
                    self.belly_flop = self:AddCard( "belly_flop" )
                    self.stand_up = self:AddCard( "stand_up" )
                    self.carapace = 3 + self.fighter:GetScale() + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )

                    self:TransitionGround()
                end,

                TransitionGround = function( self )
                    self.moves = self:MakePicker()
                        :AddID( "grawkit_charge", 2, 2 )
                        :AddID( "ai_defend_hi", 1, 1 )

                    self.fighter:RemoveCondition( "EXPOSED_BELLY" )
                    self.fighter:AddCondition( "SLATED_CARAPACE", math.max( 1, self.carapace ))
                    self.fighter:AddCondition( "TEMPER", 10 )
                    self:SetPattern( self.Ground )
                end,

                Ground = function( self )
                    if self.fighter:HasCondition( "TEMPER" ) then
                        if self.moves:ChooseCard( 1 ) == 0 then
                            self.moves:AddID( "grawkit_charge", 2, 2 )
                            self.moves:AddID( "ai_defend_hi", 1, 1 )
                            self:ChooseCard( self:AddCard( "thorny_carapace" ) )
                        end
                    else
                        self:ChooseCard( self.stand_up )
                    end
                end,

                TransitionStanding = function( self )
                    self.fighter:AddCondition( "EXPOSED_BELLY" )
                    self.fighter:RemoveCondition( "SLATED_CARAPACE" )
                    self:SetPattern( self.Standing )
                end,

                Standing = function( self )
                    self:ChooseCard( self.belly_flop )
                end,
            },

            conditions = {
                THORNY_CARAPACE = 
                {
                    name = "Thorny Carapace",
                    desc = "Adds 6 {RIPOSTE} at the end of each turn.",
                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            fighter:AddCondition( "RIPOSTE", 6 )
                        end,
                    }
                },            
                SLATED_CARAPACE = 
                {
                    name = "Slated Carapace",
                    desc = "Ignores any non-{PIERCING} damage less than or equal to <#HILITE>{1}</>. Remove 1 stack when attacked.",
                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, self.stacks or 1 )
                    end,

                    OnPreDamage = function( self, damage, attacker, battle, source, piercing )
                        if source and attacker ~= self.owner then
                            if damage <= self.stacks and not piercing then
                                damage = 0
                            end
                            self.owner:RemoveCondition( "SLATED_CARAPACE", 1 )
                        end

                        return damage
                    end,

                    event_priorities =
                    {
                        [ BATTLE_EVENT.CALC_DAMAGE ] = 999,
                    },

                    event_handlers =
                    {
                        [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt, hit )
                            if target == self.owner and dmgt.max_damage <= self.stacks and not card:IsFlagged( CARD_FLAGS.PIERCING ) and (hit == nil or not hit.piercing) then
                                dmgt:ModifyDamage(0, 0, self)
                            end
                        end
                    },
                },

                TEMPER =
                {
                    name = "Temper",
                    desc = "Remove 1 stack whenever you play an attack card targeting the grawkit. When removed, the grawkit will prepare a dangerous attack.",

                    event_handlers =
                    {
                        [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                            if attack:IsTarget( self.owner ) and attack.card:IsAttackCard() and attack.attacker:IsPlayer() then
                                self.owner:RemoveCondition( self.id, 1 )
                                if not self.owner:HasCondition( self.id ) then
                                    self.owner:ClearPreparedCards()
                                    self.owner:PrepareTurn()
                                end
                            end
                        end
                    }
                },

                EXPOSED_BELLY =
                {
                    name = "Exposed Belly",
                    desc = "All attacks do double damage.",

                    anim_mapping =
                    {
                        idle = "idle_belly",
                        hit_mid = "hit_mid_belly",
                        hit_mid_pst_idle = "hit_mid_belly_pst",
                        hit_mid_pst_stunned = "hit_mid_belly_pst_stunned",
                    },

                    OnApply = function( self, battle )
                        self.owner:PlayAnim( "idle_belly_pre", true )
                        self.owner:BroadcastEvent( battle_defs.BATTLE_EVENT.FIGHTER_REMAP_ANIMS, self.anim_mapping )
                        self.owner.knockback_immune = true
                    end,

                    OnUnapply = function( self, battle )
                        -- no pst: returns to idle directly from belly_flop.
                        -- self.owner:PlayAnim( "idle_belly_pst", true )
                        self.owner:BroadcastEvent( battle_defs.BATTLE_EVENT.FIGHTER_REMAP_ANIMS )
                        self.owner.knockback_immune = false
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                            if target == self.owner then
                                dmgt:AddDamage( dmgt.min_damage, dmgt.max_damage )
                            end
                        end
                    }
                }
            },
        },
    })
)
