local fun = require "util/fun"
local battle_defs = require "battle/battle_defs"
local EVENT = battle_defs.EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local CONDITIONS =
{
    ENCASING =
    {
        name = "Encasing",
        desc = "An object is encased inside! If left safely until the end of battle, it could be retrieved.",

        event_handlers =
        {
            [ EVENT.END_COMBAT] = function( self )
                self.battle:AddLootItem( "fuel_canister" )
            end,
        }
    },

    POWER_POLLEN =
    {
        name = "Power Pollen",
        desc = "When destroyed, all fighters gain <#HILITE>{1} {TEMP_POWER}</>.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,

        event_handlers = 
        {
            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if fighter == self.owner and status == FIGHT_STATUS.DEAD then
                    for i, f in self.battle:AllFighters() do
                        if f:IsActive() then
                            f:AddCondition( "POWER", self.stacks )
                            f:AddCondition( "POWER_LOSS", self.stacks )
                        end
                    end
                end
                --self.battle:YieldPlay( self.battle.STATUS.SYNC )
            end,
        },
    },

    ENERGIZING_VAPORS =
    {
        name = "Energizing Vapors",
        desc = "When destroyed, the attacker gains <#HILITE>{MENDING {1}}</>.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,

        event_handlers =
        {
            [ EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source, hit_tags )
                self.last_attacker = self.battle:GetFighterFromSource( source )
            end,

            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if fighter == self.owner and status == FIGHT_STATUS.DEAD and self.last_attacker then
                    self.last_attacker:AddCondition( "MENDING", self.stacks )
                end
            end,

            [ EVENT.POST_RESOLVE ] = function( self, battle, attack )
                self.last_attacker = nil
            end
        }
    },

    EXPLOSIVE =
    {
        name = "Explosive",
        desc = "When destroyed, deal <#HILITE>{1} damage</> to ALL fighters on same team.\nGains stacks every turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,
        max_stacks = 99,
        event_handlers =
        {
            [ EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                self.owner:DeltaCondition("EXPLOSIVE", 2+ math.random(3))
            end,
            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                if fighter == self.owner and status == FIGHT_STATUS.DEAD then
                    
                    local valid_targets = {}
                    for i,fighter in self.owner:GetTeam():Fighters() do
                        if fighter ~= self.owner then
                            table.insert(valid_targets, fighter)
                        end
                    end
                    if #valid_targets > 0 then
                        fighter.battle:BroadcastEvent( EVENT.DELAY, .4)
                        for k,v in ipairs(valid_targets) do
                            v:ApplyDamage(self.stacks, self.owner)
                        end
                    end
                end
            end,
        }       
    },

    REFLEX_BARB =
    {
        name = "Reflex Barb",
        desc = "Gain {RIPOSTE 5} at the end of each turn.",

        event_handlers =
        {
            [ EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    fighter:AddCondition( "RIPOSTE", 5 )
                end
            end,
        }
    },
    
    HIGH_VALUE_TARGET =
    {
        name = "High Value Target",
        desc = "Grants  {card.{1}} and {2#money} after the fight if killed.",
        desc_fn = function(self, fmt_str)
            local str = loc.format( fmt_str, self.item, self.money )
            return str
        end,
        OnApply = function( self, battle )
            
            self.item = fun{"jammed_blaster", "flead_egg", "lock_box", "deep_bog_battle"}
                            :randomPick()
            
            local difficulty = TheGame:GetGameState() and TheGame:GetGameState():GetCurrentBaseDifficulty() or 1
            self.money = math.round(math.randomGauss( 50 * difficulty, 50 * (difficulty+1)))

        end,
        event_handlers =
        {
            [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
                
                if fighter == self.owner and status == FIGHT_STATUS.DEAD then
                    self.battle:AddLootItem( self.item )
                    self.battle:AddLootMoney( self.money)
                end
            end,
        }


    },

    GESTATING =
    {
        name = "Gestating",
        desc = "{1#percent} chance to spawn a creature at the end of the turn. Increase this chance and the strength of the creature at the end of each turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetSpawnChance() )
        end,

        GetSpawnChance = function( self )
            return math.min( 1.0, 0.1 + self.stacks * 0.1 )
        end,

        SpawnCreature = function( self )
            local agent
            if self.stacks <= 3 then
                agent = Agent( "FLEAD_LARVA" )
            elseif self.stacks <= 6 then
                agent = Agent( "LUMINTHIAN" )
            else
                agent = Agent( "GROUT_KNUCKLE" )
            end

            local new_fighter = Fighter.CreateFromAgent( agent, self.owner:GetScale() )
            self.owner:GetTeam():AddFighter( new_fighter )
            self.owner:GetTeam():ActivateNewFighters( self.battle )

            self.owner:Kill()
        end,

        event_handlers =
        {
            [ EVENT.END_TURN ] = function( self, fighter )
                if fighter == self.owner then
                    local chance = self:GetSpawnChance()
                    if math.random() < chance then
                        self:SpawnCreature()
                    else
                        fighter:AddCondition( "GESTATING", 1 ) 
                    end
                end
            end,
        }
    }
}

for id, def in pairs( CONDITIONS ) do
    Content.AddBattleCondition( id, def )
end


local function GroutJoin( fighter, anim_fighter )
    local x, z = anim_fighter:GetHomePosition()
    anim_fighter.entity.cmp.AnimController:SetXFlip(fighter:GetTeamID() == TEAM.RED)
    anim_fighter.entity:SetLocalPosition( x, 0, z )
    anim_fighter:PlayAnim( "emerge" )
    anim_fighter:WaitAnim()
end

local function GroutMoving( anim_fighter, x, z, move_type, fx_tags, force_run )
    -- To satisfy AnimFighter:IsMoving()
    anim_fighter.destx = x
    anim_fighter.destz = z

    anim_fighter:PlayAnim( "surrender" )
    anim_fighter:WaitAnim()
    anim_fighter.entity.cmp.Transform:SetPosRot(x,anim_fighter.height,z)
    anim_fighter:PlayAnim( "emerge" )
    anim_fighter:WaitAnim()

    anim_fighter:StopMoving()
end

local function GroutUpdateMoving( anim_fighter, dt )
end

--this one is a turret
Content.AddCharacterDef
(
    CharacterDef("GROUT_KNUCKLE",
    {
        base_def = "MONSTER",
        
        title = "Grout Knuckle",
        build = "grout_knuckles",
        combat_strength = 1,
        is_grout = true,

        fight_data =
        {
            MAX_HEALTH = 16,

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0

            OnJoinBattle = GroutJoin,

            StartMoving = GroutMoving,

            UpdateMoving = GroutUpdateMoving,

            ally_killed_quip = "battle_grout_killed",

            ranged_riposte = true,
            anim_mapping =
            {
                riposte = "attack1",
            },

            attacks =
            {
                puncture = table.extend(NPC_ATTACK)
                {
                    name = "Puncture",
                    anim = "attack1",
                    flags = CARD_FLAGS.RANGED,
                },

                fossilize =
                {
                    name = "Fossilize",
                    anim = "pose",
                    flags = CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                    OnPostResolve = function( self, battle, attack )
                        for i, hit in attack:Hits() do
                            hit.target:AddCondition( "DEFEND", 4 )
                        end
                        self.owner:AddCondition("RIPOSTE", 8)                        
                    end,
                }

            },

            behaviour =
            {
                OnActivate = function( self )
                    self.buffs = self:MakePicker()
                        :AddID( "puncture", 1 )
                        :AddID( "fossilize", 1 )

                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("grout_burr", 2+math.random(4))
                    self.fighter:AddCondition("RIPOSTE", 8)
                end,

                Cycle = function( self )
                    self.buffs:ChooseCard()
                end,
            },
        },
    })
)

--this one is a mine
Content.AddCharacterDef
(
    CharacterDef("GROUT_SPARK_MINE",
    {
        base_def = "MONSTER",
        is_grout = true,

        title = "Spark Mine",
        build = "grout_spark_mine",
        combat_strength = 1,

        fight_data =
        {
            MAX_HEALTH = 15,

            death_fade_delay = 0.8,

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            ally_killed_quip = "battle_grout_killed",
            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0

            OnJoinBattle = GroutJoin,

            StartMoving = GroutMoving,

            UpdateMoving = GroutUpdateMoving,

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition("EXPLOSIVE", math.ceil(self.fighter:GetHealth()/2) )
                    self.fighter:AddCondition("grout_burr", 2+math.random(4))
                end,
            },
        },
    })
)


Content.AddCharacterDef
(
    CharacterDef("GROUT_SPARK_MINE_BOSS",
    {
        base_def = "MONSTER",

        title = "Spark Mine",
        build = "grout_spark_mine",
        combat_strength = 1,
        is_grout = true,

        fight_data =
        {
            MAX_HEALTH = 22,

            death_fade_delay = 0.8,

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0

            OnJoinBattle = GroutJoin,

            StartMoving = GroutMoving,

            UpdateMoving = GroutUpdateMoving,

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition("EXPLOSIVE", 13 + math.random(4) )
                    self.fighter:AddCondition("grout_burr", 3+math.random(3))
                end,
            },
        },
    })
)

--this one gives loot if you kill it
Content.AddCharacterDef
(
    CharacterDef("GROUT_LOOT_CLUSTER",
    {
        base_def = "MONSTER",
        is_grout = true,
        grout_team_restrict = TEAM.RED,

        title = "Grout Loot Cluster",
        build = "grout_loot",
        combat_strength = 1,

        fight_data =
        {
            MAX_HEALTH = 14,

            death_fade_delay = 0.8,
            ally_killed_quip = "battle_grout_killed",

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0
            
            OnJoinBattle = GroutJoin,

            StartMoving = GroutMoving,

            UpdateMoving = GroutUpdateMoving,

            anim_mapping =
            {
                stunned = "idle",
            },
            
            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition("HIGH_VALUE_TARGET" )
                    self.fighter:AddCondition("grout_burr", 1 + math.random(2))
                end,
            },
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("GROUT_BOG_CYST",
    {
        base_def = "MONSTER",
        is_grout = true,
        title = "Bog Cyst",
        build = "grout_egg_batch",
        combat_strength = 1,

        fight_data =
        {
            MAX_HEALTH = 8,

            death_fade_delay = 1.5,
            ally_killed_quip = "battle_grout_killed",

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0

            OnJoinBattle = GroutJoin,

            StartMoving = GroutMoving,

            UpdateMoving = GroutUpdateMoving,

            behaviour =
            {
                OnActivate = function( self )
                    if math.random() < .5 then
                        self.fighter:AddCondition( "ENERGIZING_VAPORS", 2+math.random(5) )
                    else
                        self.fighter:AddCondition( "POWER_POLLEN", 2+math.random(2) )
                    end
                    self.fighter:AddCondition("grout_burr", 2+math.random(2))
                end,
            },
        },
    })
)
