local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local def = CharacterDef("HESH_BOSS",
{
    unique = true,

    base_def = "NPC_BASE",
    alias = "HESH_BOSS",
    species = SPECIES.HUMAN,
    gender = GENDER.FEMALE,
    renown = 3,
    combat_strength = 2,
    boss = true,
    death_item = "lumin_generator",
    title = "Rogue Zealot",

    loved_graft = "heshian_armature",
    hated_graft = "gremlins",

    battle_preview_anim = "anim/boss_hesh_slide.zip",
    battle_preview_offset = { x = 185, y = 10 },
    battle_preview_glow = { colour = 0x10ffffFF, bloom = 0.3, threshold = 0.02 },
    battle_preview_audio = "event:/ui/prebattle_overlay/prebattle_overlay_whooshin_boss_heshboss",
    icon = engine.asset.Texture( "icons/b_hesh_boss.tex"),
    faction_id = "CULT_OF_HESH",
    voice_actor = "jakeLowClassFemale01",
    
    combat_anims = {"anim/med_combat_female_hesh_boss.zip"},
    build = "female_hesh_boss_build",
    head = "head_female_hesh_boss",

    negotiation_data = 
    {
        modifiers =
        {
            VENGEFUL =
            {
                name = "Vengeful",
                desc = "Whenever you create a new argument, {1} creates 1 {COUNTER_POINT}.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self:GetOwnerName())
                end,

                modifier_type = MODIFIER_TYPE.CORE,

                event_handlers =
                {
                    [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                        if modifier.modifier_type ~= MODIFIER_TYPE.CORE and not modifier.hidden and modifier.negotiator ~= self.negotiator then
                            self.negotiator:CreateModifier("COUNTER_POINT", 1, self)
                        end
                    end
                },
            },

            COUNTER_POINT =
            {
                name = "Counter Point",
                modifier_type = MODIFIER_TYPE.ARGUMENT,
                target_enemy = TARGET_ANY_RESOLVE,

                max_resolve = 2,
                min_persuasion = 2,
                max_persuasion = 3,

                OnBeginTurn = function( self, minigame )
                    self:ApplyPersuasion()
                end,
            },

            MAGNETIC_PULSE =
            {
                name = "Magnetic Pulse",
                desc = "Whenever you {GAMBLE}, this argument deals {1} damage to a random argument.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self.damage_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage_amt)])
                end,

                modifier_type = MODIFIER_TYPE.ARGUMENT,
                max_resolve = 10,
                target_enemy = TARGET_ANY_RESOLVE,

                damage_amt = { 2, 2, 2, 3 },

                event_handlers =
                {
                    [ EVENT.GAMBLE ] = function( self, result, source )
                        local targets = self.engine:CollectAlliedTargets( self.anti_negotiator, nil, nil, self )
                        local target = table.arraypick(targets)
                        if target then
                            target:AttackResolve( self.damage_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage_amt)], self )
                        end
                    end,
                },
            },

            CYBERNETIC_HEARING =
            {
                name = "Cybernetic Hearing",
                desc = "Whenever you play a Diplomacy card, this argument deals {1} damage to a random argument.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self.damage_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage_amt)])
                end,

                modifier_type = MODIFIER_TYPE.ARGUMENT,
                max_resolve = 10,

                damage_amt = { 2, 2, 2, 3 },
                target_enemy = TARGET_ANY_RESOLVE,

                event_handlers =
                {
                    [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                        if card.negotiator == minigame:GetPlayerNegotiator() and card:IsFlagged( negotiation_defs.CARD_FLAGS.DIPLOMACY) then
                            local damage = self.damage_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage_amt)]
                            minigame:ApplyPersuasion( self, nil, damage, damage )
                        end
                    end,
                },
            },
        },
        behaviour = 
        {
            OnInit = function( self, difficulty )
                if CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.WORDSMITH) then
                    self.hearing = self:AddArgument("CYBERNETIC_HEARING")
                    self.magnet = self:AddArgument("MAGNETIC_PULSE")

                    if self.engine:GetPlayerNegotiator().agent:GetContentID() == "ROOK" then
                        self:SetPattern(self.RookBossCycle)
                    else
                        self:SetPattern(self.BossCycle)
                    end
                    self.negotiator:AddModifier("VENGEFUL")
                else
                    self.moment_of_weakness = self:AddCard( "moment_of_weakness" )
                    self.negotiator:AddModifier( "UNCERTAINTY" )
                    self:SetPattern( self.BasicCycle )
                end
            end,

            BasicCycle = function( self, turns )
                if not self.player_negotiator:HasModifier("moment_of_weakness") then
                    self:ChooseCard( self.moment_of_weakness )
                    self:ChooseGrowingNumbers( 2 )
                else
                    self:ChooseGrowingNumbers( 2, 2 )
                    if math.random() < 0.5 then
                        self:ChooseComposure( 1, 2, 2 )
                    end
                end
            end,

            RookBossCycle = function( self, turns )
                self:ChooseGrowingNumbers( 2, 2 )

                if turns % 2 == 0 then
                    self:ChooseComposure( 1, 2, 2 )
                end
                if turns % 3 == 0 and not self.negotiator:HasModifier("MAGNETIC_PULSE") then
                    self:ChooseCard(self.magnet)
                end
            end,

            BossCycle = function( self, turns )
                self:ChooseGrowingNumbers( 2, 2 )

                if turns % 2 == 0 then
                    self:ChooseComposure( 1, 2, 2 )
                end
                if turns % 3 == 0 and not self.negotiator:HasModifier("CYBERNETIC_HEARING") then
                    self:ChooseCard(self.hearing)
                end
            end,
        }
    },
    fight_data = 
    {
        MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
        MAX_HEALTH = 100,
        battle_scale = 1.1,
        death_fade_delay = 1,
        shadow_symbol = "hips",
        shadow_scale = 8.5,

        conditions = 
        { 
            charged_attacks = 
            {
                name = "Charged Attacks",
                desc = "When {1} causes you to lose health, gain {2} {CHARGE}.",
                desc_fn = function( self, fmt_str )
                    local charge = self.charge_amt[ math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.charge_amt) ]
                    return loc.format(fmt_str, self:GetOwnerName(), charge )
                end,

                charge_amt = { 2, 2, 3, 3 },

                event_handlers = 
                {
                    [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                        if attack.card:IsAttackCard() and attack.attacker == self.owner then
                            if hit.target:IsPlayer() then
                                if not hit.evaded and not hit.defended then
                                    local tracker = hit.target:GetCondition("lumin_tracker")
                                    if tracker then 
                                        local charge = self.charge_amt[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.charge_amt)]
                                        tracker:AddLuminCharges(charge, self) 
                                    end
                                end
                            end
                        end
                    end,
                },
            },

            malfunctioning_charge_cells =
            {
                name = "Malfunctioning Charge Cells",
                desc = "At the end of your turn, take {1} damage per {CHARGE}. At the end of your turn, remove one <b>Malfunctioning Charge Cells</>.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self.damage[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage)])
                end,

                ctype = CTYPE.DEBUFF,

                damage = { 2, 2, 2, 2 },

                target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

                CalculatePreviewDamage = function( self )
                    local tracker  = self.owner:GetCondition("lumin_tracker")
                    if tracker then
                        return self.damage[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage)] * tracker:GetCharges()
                    else
                        return 0
                    end
                end,

                PreviewFn = function( self, damage_preview, targets )
                    local damage = self:CalculatePreviewDamage()
                    if targets[self.owner] == nil then
                       targets[self.owner] = {} 
                    end
                    table.insert(targets[self.owner], { condition = self, min_damage = damage, max_damage = damage, owner = self.owner })
                end,

                event_handlers = 
                {
                    [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                        local tracker = self.owner:GetCondition("lumin_tracker")
                        if self.applier:IsActive() and tracker and tracker:GetCharges() > 0 then
                            self.owner:ApplyDamage(tracker:GetCharges() * self.damage[math.min(GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1, #self.damage)], nil, nil, nil, {"lumin"})
                        end
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                },
            },
        },

        attacks = 
        {
            hesh_boss_sabotage = table.extend(NPC_ATTACK)
            {
                name = "Sabotage",
                anim = "throw",

                flags = CARD_FLAGS.RANGED | CARD_FLAGS.DEBUFF,
                base_damage = { 2, 2, 3, 4 },
                sabotage_stacks = { 2, 2, 3, 4},

                CanPlayCard = function(self, battle, target)
                    return target == nil or target:IsPlayer()
                end,

                OnPostResolve = function( self, battle, attack )
                    for i, hit in attack:Hits() do
                        if not hit.evaded then
                            local con = hit.target:AddCondition("malfunctioning_charge_cells", self.sabotage_stacks[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                            if con then 
                                con.applier = self.owner
                            end
                        end
                    end
                end,
            },
            hesh_boss_attack1 = table.extend(NPC_ATTACK)
            {
            	name = "Slashing Claws",
            	anim = "attack1",
                base_damage = { 6, 7, 8, 10 },

                target_mod = TARGET_MOD.TEAM,
            	flags = CARD_FLAGS.RANGED,
            },

            hesh_boss_attack2 = table.extend(NPC_ATTACK)
            {
            	name = "Double Whip",
            	anim = "attack2",
                base_damage = { 3, 3, 4, 5 },
                
                target_mod = TARGET_MOD.TEAM,
            	flags = CARD_FLAGS.RANGED,
                hit_count = 2,
            },

            hesh_boss_taunt =
            {
                name = "Taunt",
                anim = "taunt",

                flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
                target_type = TARGET_TYPE.SELF,

                defend_amt = { 7, 8, 9, 11},
                evasion_amt = { 1, 1, 1, 2},

                OnPostResolve = function( self, battle, attack )
                    self.owner:AddCondition("DEFEND", self.defend_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                    self.owner:AddCondition("EVASION", self.evasion_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                end
            },
        },

        behaviour =
        {
            OnActivate = function( self, fighter )
                self.sabotage = self:AddCard("hesh_boss_sabotage")
                self.attacks = self:MakePicker()
                    :AddID( "hesh_boss_attack1", 2 )
                    :AddID( "hesh_boss_attack2", 2 )
                    :AddID( "hesh_boss_taunt", 1 )
                self:SetPattern(self.Cycle)
                local fighting_rook = self.battle:GetScenario():GetPrimary(TEAM.BLUE):GetContentID() == "ROOK"
                if fighting_rook then
                    self.fighter:AddCondition("charged_attacks")
                end
            end,

            Cycle = function( self )
                local fighting_rook = self.battle:GetScenario():GetPrimary(TEAM.BLUE):GetContentID() == "ROOK"
                if fighting_rook and self.battle:GetTurns() % 4 == 0 then
                    self:ChooseCard(self.sabotage)
                else
                    self.attacks:ChooseCard()
                end
            end
        },
    },
})

Content.AddCharacterDef( def )
    