local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local FLAGS = battle_defs.CARD_FLAGS

local negotiation_defs = require "negotiation/negotiation_defs"
local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT

-----------------------------------------------------------------------------------------------

local MONSTER_DEF = CharacterDef("MONSTER",
{
    base_def = "BASE",
    is_template = true,

    species = SPECIES.BEAST,

    death_money = 0,
    death_item = nil,

    renown = 1,
    combat_strength = 1, 
    plax_flags = PLAX_FLAGS.ALLOW_MONSTERS, -- Only fits in slots flagged "ALLOW_MONSTERS"

    faction_id = MONSTER_FACTION,
    tags = { "beast", "no_rob", "enforcer" },

    nohead = true,
    on_init = function( agent )
        local gender = agent.gender
        if gender == nil then
            gender = math.random() > .5 and GENDER.MALE or GENDER.FEMALE
            agent.gender = gender
        end

        local chargen = require "charactergen" 
        chargen.ApplyBuild( agent )
    end,

    fight_data =
    {
        MAX_HEALTH = MAX_HEALTH_LOOKUP.CREATURE,
        MAX_MORALE = MAX_MORALE_LOOKUP.LOW,

        ally_killed_quip = "battle_pet_killed",
        sound_events = "CRAYOTE",
        status_widget_dx = 1.0,
        status_widget_dy = 0.0,
    }
})

Content.AddCharacterDef(MONSTER_DEF)

local DRONE_DEF = CharacterDef("DRONE",
{
    base_def = "BASE",
    is_template = true,

    species = SPECIES.MECH,
    
    nohead = true,
    gender = GENDER.UNDISCLOSED,

    fight_data = 
    {
        shadow_scale = 5.0,
        height = 2.0,
        status_widget_dy = 0.0,
        ally_killed_quip = "battle_pet_killed",

        MAX_HEALTH = 24,
    },
})
Content.AddCharacterDef(DRONE_DEF)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("CRAYOTE_SPITTER",
    {
        designer_note = "Spits at you, making your cards cost 1 more to play until it's played.",
        base_def = "MONSTER",
        
        title = "Spitting Yote",
        build = "spitter_dire_dog",
        combat_strength = 2,
        death_money = 0,
        death_item = "razorspine",
        voice_event = "event:/vo/compendium/speak_yote",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = MAX_HEALTH_LOOKUP.CREATURE,
            formation = FIGHTER_FORMATION.BACK_X | FIGHTER_FORMATION.FRONT_Z,

            conditions = 
            {
                acidic_slime = 
                {
                    name = "Acidic Slime",
                    desc = "Every turn, <#HILITE>{1}</> random {1*card|cards} in your hand {1*costs|cost} 1 more until played. At the end of your turn, remove 1 <b>Acidic Slime</>.",
                    desc_fn = function(self, fmt_str)
                        return loc.format(fmt_str, self.stacks)
                    end,

                    ctype = CTYPE.DEBUFF,

                    OnApply = function( self, battle )
                        self.modify_cards = {}
                    end,

                    event_priorities =
                    {
                        [ BATTLE_EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_ADDITIVE,
                    },

                    event_handlers =
                    {
                        [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                            for i, card in ipairs(table.multipick( self.battle:GetHandDeck().cards, self.stacks, false )) do
                                table.insert_unique( self.modify_cards, card )
                            end
                        end,

                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            if self.owner == fighter then
                                fighter:RemoveCondition(self.id, 1)
                            end
                        end,

                        [ BATTLE_EVENT.START_RESOLVE ] = function( self, battle, resolve_card )
                            local i = table.find(self.modify_cards, resolve_card)
                            if i then
                                table.remove(self.modify_cards, i)
                            end
                        end,

                        [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                            if self.owner == card.owner then
                                if table.find(self.modify_cards, card) then
                                    cost_acc:AddValue( 1, self )
                                end
                            end
                        end,

                    },
                },

            },

            attacks = 
            {
                yote_spit2 = table.extend(NPC_ATTACK)
                {
                    name = "Acidic Splash",
                    anim = "spit",
                    flags = FLAGS.RANGED | FLAGS.DEBUFF,

                    OnPostResolve = function( self, battle, attack )
                        if not attack:CheckHitResults( "evaded" ) then
                            if attack.target == battle:GetPlayerFighter() then
                                attack.target:AddCondition("acidic_slime", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                            else
                                attack.target:AddCondition("IMPAIR", 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                            end
                        end
                    end,

                    damage_mult = 0.2,
                },

                yote_spit = table.extend(NPC_ATTACK)
                {
                    name = "Spit",
                    anim = "spit",
                    flags = FLAGS.RANGED,
                },

                yote_slaver = table.extend(NPC_BUFF)
                {
                    name = "Slaver",
                    anim = "taunt",
                    flags = FLAGS.SKILL | FLAGS.BUFF | FLAGS.HEAL,
                    target_type = TARGET_TYPE.SELF,

                    heal_amt = 3,

                    OnInit = function( self )
                        self.heal_amt = self.heal_amt * (1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end,

                    CanPlayCard = function( self, battle, target )
                        return self.owner:GetHealthPercent() < 1.0
                    end,

                    OnPostResolve = function(self, battle, attack)
                        for i, hit in attack:Hits() do
                            hit.target:HealHealth( self.heal_amt, self )
                        end
                    end,
                }
            },

            anim_mapping =
            {
                riposte = "spit",
            },

            ranged_riposte = true,

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,
            
            hit_tags = { "small" },

            behaviour =
            {
                OnActivate = function( self )
                    self.slaver = self:AddCard( "yote_slaver" )
                    self.spit = self:AddCard( "yote_spit" )
                    self.acid = self:AddCard( "yote_spit2" )
                    self.cycle_idx = math.random( 1, 3 )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.cycle_idx = (self.cycle_idx % 3) + 1

                    if self.cycle_idx <= 1 and self:ChooseCard( self.slaver ) then
                        return
                    end
                    if self.cycle_idx <= 2 and self:ChooseCard( self.spit ) then
                        return
                    end
                    if self.cycle_idx <= 3 and self:ChooseCard( self.acid ) then
                        return
                    end
                end,
            }
        },
    })
)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("CRAYOTE",
    {
        designer_note = "Nothing special, they just power up over time and bleed you once in a while.",
        base_def = "MONSTER",

        title = "Yote",
        build = "dire_dog",
        
        death_money = 0,
        death_item= "carapace",

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",
        voice_event = "event:/vo/compendium/speak_yote",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {

            status_widget_dx = 0,
            status_widget_dy = -0.6,
            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            hit_tags = { "small" },
            formation = FIGHTER_FORMATION.FRONT_Z,
            cards = { "yote_gnaw", "bristle", "yote_bite" },

            anim_mapping = 
            {
                riposte = "savage_bite"
            },

            attacks = 
            {
                bristle = table.extend(NPC_BUFF)
                {
                    name = "Bristle",
                    anim = "taunt",
                    target_type = TARGET_TYPE.SELF,
                    flags = FLAGS.SKILL | FLAGS.BUFF,

                    buff = { 2, 2, 3, 3, 3 },

                    OnPostResolve = function( self, battle, attack )
                        local buff = self:ScaleBuff( self.buff ) + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
                        if not self.used_power then
                            self.owner:AddCondition( "POWER", buff )
                            self.used_power = true
                        else
                            self.owner:AddCondition( "DEFEND", buff )
                        end
                    end,
                },

                yote_bite = table.extend(NPC_ATTACK)
                {
                    name = "Savage Bite",
                    anim = "savage_bite",
                    flags = FLAGS.MELEE,
                    hit_count = 2,
                    damage_mult = 0.5,
                },

                yote_gnaw = table.extend(NPC_ATTACK)
                {
                    name = "Gnaw",
                    anim = "tear_flesh",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    target_type = TARGET_TYPE.ENEMY,

                    features =
                    {
                        BLEED = 2,
                    },

                    damage_mult = 0.8,
                },

                yote_call_in = table.extend(NPC_ATTACK)
                {
                    name = "Gnaw",
                    anim = "tear_flesh",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF | FLAGS.SPECIAL,
                    combo_starter = true,
                    synchronous = true,
                    combo_followup = true,
                    features =
                    {
                        BLEED = 3,
                    },
                    damage_mult = 1.5,
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "yote_gnaw", 1 )
                        :AddID( "yote_bite", 1 )
                        :AddID( "bristle", 1 )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("CRAYOTE_UPGRADED",
    {
        base_def = "CRAYOTE",

        title = "Yote",
        build = "upgraded_dire_dog",
        
        death_money = 0,
        death_item = "carapace",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            
            MAX_HEALTH = 40,

            status_widget_dx = 0,
            status_widget_dy = -0.6,
            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,

            hit_tags = { "small" },
            formation = FIGHTER_FORMATION.FRONT_Z,
            cards = { "yote_gnaw_upgraded", "bristle_upgraded", "yote_bite_upgraded" },

            anim_mapping = 
            {
                riposte = "savage_bite"
            },

            attacks = 
            {
                bristle_upgraded = table.extend(NPC_BUFF)
                {
                    name = "Bristle",
                    anim = "taunt",
                    target_type = TARGET_TYPE.SELF,
                    flags = FLAGS.SKILL | FLAGS.BUFF,

                    buff = { 2, 2, 3, 3, 3 },

                    OnPostResolve = function( self, battle, attack )
                        local buff = self:ScaleBuff( self.buff ) + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
                        if not self.used_power then
                            self.owner:AddCondition( "POWER", buff )
                            self.used_power = true
                        else
                            self.owner:AddCondition( "DEFEND", buff )
                        end
                    end,
                },

                yote_bite_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Savage Bite",
                    anim = "savage_bite",
                    flags = FLAGS.MELEE,
                    hit_count = 2,
                    damage_mult = 0.5,
                },

                yote_gnaw_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Gnaw",
                    anim = "tear_flesh",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    target_type = TARGET_TYPE.ENEMY,

                    features =
                    {
                        BLEED = 2,
                    },

                    damage_mult = 0.8,
                },

                defend_allies = table.extend(NPC_BUFF)
                {
                    name = "Pack Protection",
                    anim = "defense_taunt",
                    flags = FLAGS.SKILL | FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                    target_mod = TARGET_MOD.TEAM,

                    buff_strength = 4,

                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition("DEFEND", self.buff_strength, self)
                    end
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "yote_gnaw_upgraded", 1 )
                        :AddID( "yote_bite_upgraded", 1 )
                        :AddID( "bristle_upgraded", 1 )
                        :AddID( "defend_allies", 1)
                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("ARMOURED", 4, self)
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("HORNED_CRAYOTE",
    {
        base_def = "MONSTER",

        title = "Horned Yote",
        build = "alpha_dire_dog",
        
        death_money = 0,
        death_item= "yote_horn",
        

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",
        combat_strength = 2.5,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            battle_scale = 1.25,
            hit_tags = { "small" },
            cards = {"bristle" },

            anim_mapping = 
            {
                riposte = "savage_bite"
            },

            attacks = 
            {
                yote_headbutt = table.extend(NPC_ATTACK)
                {
                    name = "Headbutt",
                    anim = "attack_alpha",

                    flags = FLAGS.MELEE,
                    damage_mult = 0.6,
                    target_count = 2,
                },

                yote_scratch = table.extend(NPC_ATTACK)
                {
                    name = "Scratch",
                    anim = "claw",

                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    damage_mult = 0.8,

                    features =
                    {
                        EXPOSED = 2,
                    },
                },

                horned_yote_call_in = table.extend(NPC_ATTACK)
                {
                    name = "Headbutt",
                    anim = "attack_alpha",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF | FLAGS.SPECIAL,
                    combo_starter = true,
                    synchronous = true,
                    combo_followup = true,
                    features =
                    {
                        WOUND = 2,
                    },
                    damage_mult = 1.25,
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "yote_headbutt", 2 )
                        :AddID( "yote_scratch", 1 )
                        :AddID( "bristle", 1 )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("HORNED_CRAYOTE_UPGRADED",
    {
        base_def = "MONSTER",

        title = "Horned Yote",
        build = "upgraded_alpha_dire_dog",
        
        death_money = 0,
        death_item= "yote_horn",

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            battle_scale = 1.25,
            hit_tags = { "small" },
            cards = {"bristle_upgraded" },

            anim_mapping = 
            {
                riposte = "savage_bite"
            },

            attacks = 
            {
                yote_headbutt_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Headbutt",
                    anim = "attack_alpha",

                    flags = FLAGS.MELEE,
                    damage_mult = 1.35
                },

                yote_scratch_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Scratch",
                    anim = "claw",

                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    damage_mult = 0.8,

                    features =
                    {
                        EXPOSED = 1,
                        WOUND = 1,
                    },
                }
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "yote_headbutt_upgraded", 2 )
                        :AddID( "yote_scratch_upgraded", 1 )
                        :AddID( "bristle_upgraded", 1 )
                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("ARMOURED", 4, self)
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("VROC",
    {
        designer_note = "Has a single evade, and wounds you. Has potential to do a fair amount of damage by continual wounds, but is paired with a goon that defends it.",
        base_def = "MONSTER",

        title = "Vroc",
        build = "vroc",
        scale_adjust = 0.9,
        tags = {'guard', 'vroc'},
        formation = FIGHTER_FORMATION.FRONT_Z,
        voice_event = "event:/vo/compendium/speak_vroc",

        combat_anims = { "anim/med_combat_vroc.zip" },
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_admiralty_vroc.tex"),

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            battle_scale = 1.15,
            formation = FIGHTER_FORMATION.FRONT_Z,
            status_widget_head_dx = 0.5,
            status_widget_head_dy = 2.0,
            status_widget_dy = -1.2,

            cards = { "vroc_bite" },
            hit_tags = { "small" },

            conditions = 
            {
                nimble = 
                {
                    name = "Nimble",
                    desc = "At the start of each enemy turn, gain 1 {EVASION} and reduce <b>Nimble</b> by 1.",

                    ctype = CTYPE.INNATE,
                    
                    event_handlers = 
                    {
                        [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                            if self.owner:GetTeamID() == TEAM.RED then
                                self.owner:AddCondition("EVASION")
                                self.owner:RemoveCondition( self.id, 1 )
                            end
                        end,
                        [ BATTLE_EVENT.BEGIN_ENEMY_TURN ] = function( self, battle )
                            if self.owner:GetTeamID() == TEAM.BLUE then
                                self.owner:AddCondition("EVASION")
                                self.owner:RemoveCondition( self.id, 1 )
                            end
                        end
                    },
                },
            },

            attacks = 
            {
                vroc_charge = table.extend(NPC_ATTACK)
                {
                    name = "Charge",
                    anim = "attack1",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,

                    damage_mult = 0.5,

                    OnPostResolve = function(self, battle, attack)
                        attack:AddCondition("WOUND", 2, self)
                    end,
                },

                vroc_bite = table.extend(NPC_ATTACK)
                {
                    name = "Savage Bite",
                    anim = "attack3",
                    flags = FLAGS.MELEE,

                    hit_count = 2,
                    base_damage = 2,
                },

                vroc_taunt = table.extend(NPC_BUFF)
                {
                    name = "Howl",
                    anim = "taunt",

                    flags = FLAGS.SKILL | FLAGS.BUFF,

                    target_type = TARGET_TYPE.SELF,
                    features =
                    {
                        DEFEND = 3,
                    }
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "vroc_bite", 2 )
                        :AddID( "vroc_charge", 1 )
                        :AddID( "vroc_taunt", 1)
                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("nimble", 1 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
        death_item = "vroc_plume",
    })
)


Content.AddCharacterDef
(
    CharacterDef("VROC_UPGRADED",
    {
        base_def = "VROC",
        
        title = "Vroc",
        build = "upgraded_vroc",
        scale_adjust = 0.9,
        tags = {'guard', 'vroc'},
        formation = FIGHTER_FORMATION.FRONT_Z,

        combat_anims = { "anim/med_combat_vroc.zip" },

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = 40,
            battle_scale = 1.15,
            formation = FIGHTER_FORMATION.FRONT_Z,
            status_widget_head_dx = 0.5,
            status_widget_head_dy = 2.0,
            status_widget_dy = -1.2,

            cards = { "vroc_bite_upgraded" },
            hit_tags = { "small" },

            attacks = 
            {
                vroc_charge_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Charge",
                    anim = "attack1",
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,

                    damage_mult = 0.5,

                    CanPlayCard = function(self, battle, target)
                        return target == nil or target and not target:HasCondition("METALLIC")
                    end,

                    OnPostResolve = function(self, battle, attack)
                        attack.target:AddCondition("WOUND", 2)
                    end,
                },

                vroc_bite_upgraded = table.extend(NPC_ATTACK)
                {
                    name = "Savage Bite",
                    anim = "attack3",
                    flags = FLAGS.MELEE,

                    hit_count = 2,
                    base_damage = 3,
                },

                vroc_howl = table.extend(NPC_BUFF)
                {
                    name = "Rallying Howl",
                    anim = "taunt",
                    flags = FLAGS.BUFF | FLAGS.SKILL,
                    target_type = TARGET_TYPE.FRIENDLY_OR_SELF,
                    CanPlayCard = function(self, battle, target)
                        return target == nil or target and target:IsPlayer()
                    end,

                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition("vroc_howl", 1, self)
                        self.owner:AddCondition("DEFEND", 3, self)
                    end,

                    condition = 
                    {
                        name = "Rallying Howl",
                        desc = "At the start of your turn, gain 1 action and draw 1 card. At the end of your turn, remove this.",

                        event_handlers =
                        {
                            [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                                battle:ModifyActionCount(1)
                                battle:DrawCards(1)
                            end,

                            [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                                self.owner:RemoveCondition(self.id)
                            end
                        },
                    },
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "vroc_bite_upgraded", 1 )
                        :AddID( "vroc_charge_upgraded", 2 )
                        :AddID( "vroc_howl", 1 )
                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("nimble", 1 + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    self.fighter:AddCondition("METALLIC")
                    self.fighter:AddCondition("ARMOURED", 4, self)
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
        death_item = "vroc_plume",
    })
)


------------------------

Content.AddCharacterDef
(
    CharacterDef("COMBAT_DRONE",
    {
        base_def = "DRONE",

        title = "Combat Drone",
        build = "combat_drone",
        combat_strength = 1,
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_spark_baron_drone.tex"),

        fight_data = table.extend( DRONE_DEF.fight_data )
        {
            formation = FIGHTER_FORMATION.FRONT_Z,
            track_symbol = "head",
            shadow_symbol = "eye",
            shadow_scale = 0.0, -- Doesn't add a shadow prefab if 0
            shadow_alpha = 0.0,

            status_widget_head_dx = 0,
            status_widget_head_dy = 2.6,
            status_widget_dx = 0,
            status_widget_dy = -1.3,

            anim_mapping =
            {
                run_forward_pst = "idle",
                run_forward_pre = "idle",
                run_forward = "idle",
            },

            fx =
            {
                drone_thrust = { auto = true },
                combat_drone_thrust_dust = { auto = true },

                -- ccc: FIXME: I broke the drone stunn override, which basically sets it to a different symbol.
                -- stunned = { anim="anim/stunned_fx.zip", state="back", walltime=true, loop= true, symbol="back", z_depth=-0.5, bloom = 0.5, alpha = 0.7, x_offset = 0.5, y_offset = -0.3 },
                -- stunned_frt = { anim="anim/stunned_fx.zip", state="front", walltime=true, loop= true, symbol="back", z_depth= 0.5, bloom = 0.5, alpha = 1.0, x_offset = 0.5, y_offset = -0.3 },
            },

            cards = { "drone_discharge", "drone_charge" },
            behaviour =
            {
                OnActivate = function( self )
                    if not CheckBits(self.engine:GetFlags(), BATTLE_FLAGS.SURVIVAL) then
                        self.fighter:AddCondition("no_alone_drone")
                    end
                    self.fighter:AddCondition( "METALLIC" )
                    self.attacks = self:MakePicker()
                        :AddID( "drone_discharge", 3 )
                        :AddID( "drone_charge", 3 )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)

---------------------------------------------------

Content.AddCharacterDef
(
    CharacterDef("FLEAD",
    {
        designer_note = "Gains defense every turn and debuffs you.",
        base_def = "MONSTER",
        title = "Flead",
        build = "flead",
        voice_event = "event:/vo/compendium/speak_flead",

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = 16,
            MAX_MORALE = MAX_MORALE_LOOKUP.VERY_LOW,
            formation = FIGHTER_FORMATION.FRONT_Z,

            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,
            shadow_symbol = "hips",

            anim_mapping =
            {
                riposte = "ball"
            },

            cards = { "flead_bite", "flead_sting", "flead_ball" },

            fx =
            {
                flead_dust = { auto = true },
            },

            attacks = 
            {
                flead_bite = table.extend(NPC_ATTACK)
                {
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    name = "Flead Bite",
                    anim = "bite",

                    damage_mult = 0.33,

                    features =
                    {
                        WOUND = 2,
                    },
                },

                flead_sting = table.extend(NPC_ATTACK)
                {
                    flags = FLAGS.MELEE | FLAGS.DEBUFF,
                    name = "Flead Sting",
                    anim = "sting",

                    damage_mult = 0.5,

                    features =
                    {
                        IMPAIR = 1,
                    },
                },

                flead_ball = table.extend(NPC_ATTACK)
                {
                    flags = FLAGS.MELEE,
                    name = "Flead Ball",
                    anim = "ball",

                    damage_mult = 0.5,
                },
            },

            hit_tags = { "small" },

            conditions =
            {
                chitinous =
                {
                    name = "Chitinous",
                    desc = "Gain <#HILITE>{DEFEND {1}}</> at the end of their turn. Whenever all {DEFEND} is broken, remove 1 <b>Chitinous</b>.",   
                    desc_fn = function(self, fmt_str)
                        return loc.format(fmt_str, self.stacks)
                    end,

                    apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",     
                    ctype = CTYPE.INNATE,

                    OnApply = function(self, battle)
                        self.owner:AddCondition( "DEFEND", self.stacks )
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            if self.owner == fighter then
                                self.owner:AddCondition( "DEFEND", self.stacks )
                            end
                        end,

                        [ BATTLE_EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
                            if fighter == self.owner and condition:GetID() == "DEFEND" and condition:GetStacks() == 0 and not (source and source.owner == self.owner) then
                                self.owner:RemoveCondition( self.id, 1 )
                            end
                        end,
                    }
                }
            },

            behaviour =
            {
                OnActivate = function( self )

                    self.attacks = self:MakePicker()
                        :AddID( "flead_bite", 1 )
                        :AddID( "flead_sting", 1 )
                        :AddID( "flead_ball", 1 )
                    self:SetPattern( self.Cycle )

                    local CHITINOUS_STACKS = {3, 4, 5}
                    local idx = math.min(self.fighter:GetScale(), #CHITINOUS_STACKS)
                    local stacks = CHITINOUS_STACKS[ idx ]
                    stacks = stacks + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )

                    self.fighter:AddCondition( "chitinous", stacks )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
        death_item = "mandible",
    })
)

Content.AddCharacterDef
(
    CharacterDef("FLEAD_UPGRADED",
    {
        base_def = "FLEAD",
        title = "Flead",
        build = "flead_armoured",

        combat_strength = 1,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = 30,
            MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,
            status_widget_head_dx = 1.0,
            status_widget_head_dy = 2.5,
            shadow_symbol = "chest",
            
            fx =
            {
                flead_thrust = { auto = true },
                flead_dust = { auto = true, },
            },

            cards = { "flead_upgraded_headbutt", "flead_upgraded_shoot", "flead_upgraded_taunt" },

            attacks = 
            {
                flead_upgraded_headbutt = table.extend(NPC_ATTACK)
                {
                    flags = FLAGS.MELEE,
                    name = "Headbutt",
                    anim = "attack2",

                    damage_mult = 1,
                },

                flead_upgraded_shoot = table.extend(NPC_ATTACK)
                {
                    flags = FLAGS.RANGED | FLAGS.DEBUFF,
                    name = "Back Cannon",
                    anim = "attack1",

                    damage_mult = 1.2,

                    features =
                    {
                        WOUND = 2,
                    },
                },

                flead_upgraded_taunt = table.extend(NPC_BUFF)
                {
                    flags = FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                    name = "Defensive Maneuver",
                    anim = "taunt2",

                    features =
                    {
                        DEFEND = 6
                    },
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.attacks = self:MakePicker()
                        :AddID( "flead_upgraded_headbutt", 1 )
                        :AddID( "flead_upgraded_shoot", 1 )
                        :AddID( "flead_upgraded_taunt", 1 )

                    local ARMOURED_STACKS = {4, 5, 6}
                    local idx = math.min(self.fighter:GetScale(), #ARMOURED_STACKS)
                    local stacks = ARMOURED_STACKS[ idx ]
                    stacks = stacks + 2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )

                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("ARMOURED", stacks, self)
                    self.fighter:AddCondition("METALLIC")
                end,

                Cycle = function( self )
                    return self.attacks:ChooseCard()
                end,
            },
        },
    })
)

---------------------------------------------------


Content.AddCharacterDef
(
    CharacterDef("LUMINTHIAN",
    {
        designer_note = "Big monster that gains power whenever you attack it.",
        base_def = "MONSTER",
        title = "Krill",
        build = "luminthian",
        combat_strength = 3,
        death_money = 0,
        death_item = "krill_ichor",

        fight_data =
        {
            battle_scale = 1.5,
            cards = { "luminthian_spit", "luminthian_slash" },
            
            MAX_HEALTH = MAX_HEALTH_LOOKUP.HIGH,
            MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,

            conditions = 
            {
                scratching_post =
                {
                    name = "Anger",
                    desc = "Gain <#HILITE>{1}</> {POWER} whenever attacked.\nPeriodically purges all status effects.",
                    desc_fn = function( self, fmt_str )
                         return loc.format( fmt_str, self.stacks )
                    end,

                    ctype = CTYPE.INNATE,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit  )
                            if attack.card:IsAttackCard() and hit.target == self.owner then
                                self.owner:AddCondition("POWER", self.stacks)
                            end
                        end,
                    }
                },

            },


            attacks = 
            {
                luminthian_spit = table.extend(NPC_ATTACK)
                {
                    name = "Lumin Spittle",
                    pre_anim = "spit_pre",
                    anim = "spit",
                    post_anim = "spit_pst",
                    flags = FLAGS.RANGED,
                    target_mod = TARGET_MOD.RANDOM1,

                    target_count = 2,
                    base_damage = 3,
                },

                luminthian_slash = table.extend(NPC_ATTACK)
                {
                    name = "Rending Claws",
                    anim = "slash",
                    flags = FLAGS.MELEE,
                },

                luminthian_seethe =
                {
                    name = "Seethe",
                    anim = "taunt",
                    flags = FLAGS.SKILL | FLAGS.BUFF | FLAGS.HEAL,
                    target_type = TARGET_TYPE.SELF,

                    CanPlayCard = function(self, battle)
                        return self.owner:GetConditionStacks("POWER") > 0
                    end,

                    OnPostResolve = function( self )
                        local count = table.count( self.owner:GetConditions() ) - 1
                        for condition_id, condition in pairs( self.owner:GetConditions() ) do
                            if condition_id ~= "scratching_post" then
                                self.owner:RemoveCondition( condition_id )
                            end
                        end

                        if self.owner:GetScale() > 1 then
                            local health = math.floor( count * (self.owner:GetScale() - 1 ))
                            self.owner:HealHealth( health, self )
                        end
                    end,
                }
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition( "scratching_post", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ) )
                    
                    local spit_bonus = self.fighter:GetScale() >= 4 and 1 or 0
                    local slash_bonus = self.fighter:GetScale() >= 2 and 1 or 0
                    self.attacks = self:MakePicker()
                        :AddID( "luminthian_spit", 1 + spit_bonus )
                        :AddID( "luminthian_slash", 1 + slash_bonus )
                        :AddID( "luminthian_seethe", 1 )

                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            },
        },
    })
)

---------------------------------------------------


Content.AddCharacterDef
(
    CharacterDef("SHROOGLET",
    {
        designer_note = "See Shroog",
        base_def = "MONSTER",

        title = "Shrooglet",
        build = "shrooglet",
        combat_strength = .5,
        
        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",
        voice_event = "event:/vo/compendium/speak_shrooglet",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            hit_tags = { "small" },
            formation = FIGHTER_FORMATION.FRONT_X | FIGHTER_FORMATION.FRONT_Z,

            status_widget_head_dx = 0.5,
            status_widget_head_dy = 2.5,

            MAX_HEALTH = 8,
            MAX_MORALE = MAX_MORALE_LOOKUP.IMMUNE,

            cards = { "shrooglet_feed", "shrooglet_head_butt", "shrooglet_rev_up"},

            conditions = 
            {
                
                run_away_shrooglet = {
                    name = "Quitter",
                    desc = "Will run away if the Shroog is killed.",
                    icon = "battle/conditions/quitter.tex",
                    event_handlers = 
                    {
                        [ BATTLE_EVENT.FIGHTER_KILLED ] = function( self, fighter )
                            if fighter:GetTeam() and fighter:GetTeam() == self.owner:GetTeam() and fighter.agent:GetContentID() == "SHROOG" then
                                self.owner:Flee()
                            end
                        end,
                    },
                },

                hungry_shrooglet =
                {
                    name = "Hungry",
                    desc = "{1}'s Feed attack heals it for {2} health.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self:GetOwnerName(), self.drain_amt or self.base_drain_amt)
                    end,
                    OnApply = function( self)
                        self.drain_amt = math.floor(self.base_drain_amt * (1.0 + 0.5 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )))
                    end,
                    DoHeal = function( self )
                        self.owner:HealHealth(self.drain_amt or self.base_drain_amt, self)
                    end,

                    base_drain_amt = 6,
                },
            },


            attacks = 
            {
                shrooglet_feed = table.extend(NPC_ATTACK)
                {
                    name = "Feed",
                    anim = "feed",
                    flags = FLAGS.MELEE,
                    buff_strength = 1,
                    OnPostResolve = function( self, battle, attack )
                        if not attack:CheckHitResult( attack.target, "evaded" ) then
                            local con = self.owner:GetCondition( "hungry_shrooglet" )
                            if con then
                                con:DoHeal()
                            end
                        end
                    end,

                    base_damage = {3, 6, 10},                   
                },


                shrooglet_head_butt = table.extend(NPC_ATTACK)
                {
                    name = "Head Butt",
                    anim = "head_butt",
                    flags = FLAGS.MELEE,

                    base_damage = {3, 6, 10},
                },

                shrooglet_rev_up = table.extend(NPC_BUFF)
                {
                    name = "Rev Up",
                    anim = "taunt",
                    target_type = TARGET_TYPE.SELF,
                    flags = FLAGS.SKILL | FLAGS.BUFF,

                    features = 
                    {
                        POWER = 2,
                        DEFEND = 3,
                    }

                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self:SetPattern( self.Cycle )
                    
                    self.attacks = self:MakePicker()
                        :AddID( "shrooglet_head_butt", 1 )
                        :AddID( "shrooglet_rev_up", 1 )
                    
                    self.wounded_attacks = self:MakePicker()
                        :AddID( "shrooglet_feed", 1 )

                    self.fighter:AddCondition("hungry_shrooglet")
                end,

                Cycle = function( self )
                    if self.fighter:GetHealthPercent() < 1 then
                        self.wounded_attacks:ChooseCard()
                    else
                        self.attacks:ChooseCard()
                    end
                end,
            },
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("SHROOGLET_UPGRADED",
    {
        base_def = "SHROOGLET",

        title = "Shrooglet",
        build = "upgraded_shrooglet",

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            hit_tags = { "small" },
            formation = FIGHTER_FORMATION.FRONT_X | FIGHTER_FORMATION.FRONT_Z,

            status_widget_head_dx = 0.5,
            status_widget_head_dy = 2.5,

            MAX_HEALTH = 20,
            MAX_MORALE = MAX_MORALE_LOOKUP.IMMUNE,
            ally_killed_quip = "battle_pet_killed",

            cards = { "shrooglet_feed", "shrooglet_head_butt", "shrooglet_rev_up"},

            conditions = 
            {
                
                shrooglet_frenzy =
                {
                    name = "Shrooglet Frenzy",
                    desc = "Gains <#HILITE>{POWER {1}}</> whenever {3} attacks. Lose {POWER {2}} whenever {3} takes damage.",
                    desc_fn = function(self, fmt_str)
                        return loc.format(fmt_str, self.stacks, self.power_loss, self:GetOwnerName() )
                    end,
                    ctype = CTYPE.BUFF,
                    power_loss = 2,

                    OnPostDamage = function( self, delta, attacker, battle, source )
                        if delta > 0 then
                            self.owner:RemoveCondition("POWER", math.min(self.owner:GetConditionStacks("POWER"), self.power_loss), self )
                        end
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                            if attack.attacker == self.owner then
                                if attack.card:IsDamageCard() then
                                    self.owner:AddCondition( "POWER", self.stacks, self )
                                end
                            end
                        end,
                    }
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self:SetPattern( self.Cycle )
                    
                    self.attacks = self:MakePicker()
                        :AddID( "shrooglet_head_butt", 1 )
                        :AddID( "shrooglet_rev_up", 1 )
                    
                    self.wounded_attacks = self:MakePicker()
                        :AddID( "shrooglet_feed", 2 )
                        :AddID( "shrooglet_head_butt", 1)

                    self.fighter:AddCondition("shrooglet_frenzy", 1, self)
                    self.fighter:AddCondition("ARMOURED", 4, self)
                    self.fighter:AddCondition("hungry_shrooglet")
                end,

                Cycle = function( self )
                    if self.fighter:GetHealthPercent() < 1 then
                        self.wounded_attacks:ChooseCard()
                    else
                        self.attacks:ChooseCard()
                    end
                end,
            },
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("OOLO_SNAIL",
    {
        designer_note = "See Oolo",
        base_def = "MONSTER",

        title = "Battle Oshnu",
        name = "Bertha",
        build = "combat_snail",
        faction_id = "ADMIRALTY",
        hide_in_compendium = true,

        combat_strength = 1,
        boss = true,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            status_widget_dy = -1.2,
            battle_scale = 1.25,
            MAX_HEALTH = 30,
            MAX_MORALE = MAX_MORALE_LOOKUP.IMMUNE,


            formation = FIGHTER_FORMATION.BACK_Z | FIGHTER_FORMATION.FRONT_X,

            cards = { "oolo_oshnu_bite", "oolo_oshnu_blast", "oolo_oshnu_shell", "oolo_oshnu_unshell" },

            conditions = 
            {
                oolo_oshnu_defenses =
                {
                    name = "Oshnu Shell",
                    desc = "The oshnu cannot be killed but will hide in its shell and lose all {POWER} when brought to 1 health.",

                    disabled = false,

                    CanBeTargetted = function(self, card, fighter)
                        if self.disabled and card:IsAttackCard() and fighter == self.owner and self.owner:IsActive() then
                            return false, self.name
                        end
                        
                        return true
                    end,

                    event_handlers = 
                    {
                        [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                            if fighter == self.owner and stat == COMBAT_STAT.HEALTH then
                                if value == 1 and not self.disabled then
                                    self.disabled = true


                                    local card = Battle.Card("oolo_oshnu_shell", self.owner)
                                    self.battle:PlayCard( card )

                                    fighter:ClearPreparedCards()
                                    
                                    self.owner.behaviour:SetPattern( self.owner.behaviour.Unshell )

                                    if self.battle:GetCurrentTeam() == fighter:GetTeam() then
                                        fighter.behaviour.unshell_turn = self.battle:GetTurns() + 1
                                    else
                                        fighter.behaviour.unshell_turn = self.battle:GetTurns() + 1
                                    end
                                end
                            end
                        end,
                    },
                },

                oolo_oshnu_devour = 
                {
                    name = "Muscular Glands",
                    desc = "Gains {1} {POWER} at the end of its turn.",
                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, self.power_gain )
                    end,

                    power_gain = 1,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter)
                            if fighter == self.owner and self.owner:IsActive() then
                                self.owner:AddCondition("POWER", self.power_gain)
                            end
                        end,
                    },
                },

            },
            attacks = 
            {
                oolo_oshnu_unshell =
                {
                    name = "Unshell",
                    anim = "shell_pst",

                    target_type = TARGET_TYPE.SELF,

                    PostPresAnim = function( self, anim_fighter )
                        anim_fighter:SetAnimMapping()
                    end,
                    flags =  FLAGS.HEAL,

                    OnPostResolve = function(self, battle)
                        self.owner:HealHealth( self.owner:GetMaxHealth(), self )
                        self.owner:RemoveAllDebuffs( self )
                        self.owner:RemoveCondition("POWER")
                        if self.owner:HasCondition("oolo_oshnu_defenses") then
                            self.owner:GetCondition("oolo_oshnu_defenses").disabled = false
                        end
                    end,
                },

                oolo_oshnu_bite = table.extend(NPC_ATTACK)
                {
                    name = "Devour",
                    anim = "bite",
                    flags = FLAGS.MELEE,

                    base_damage = { 6, 7, 9, 10 },
                    heal_amt = 12,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:HealHealth( self.heal_amt, self )
                    end
                },

                oolo_oshnu_blast_pre = table.extend(NPC_BUFF)
                {
                    name = "Blaster Charge",
                    anim = "taunt",
                    target_type = TARGET_TYPE.SELF,

                    features = 
                    {
                        DEFEND = 10,
                        RIPOSTE = 5,
                    },
                },

                oolo_oshnu_blast = table.extend(NPC_ATTACK)
                {
                    name = "Oshnu Blast",
                    anim = "blast",
                    flags = FLAGS.RANGED,

                    target_mod = TARGET_MOD.TEAM,

                    base_damage = { 2, 3, 4, 6 },
                    hit_count = 3,
                },

                oolo_oshnu_shell =
                {
                    name = "Shell up",
                    anim = "shell",
                    target_type = TARGET_TYPE.SELF,

                    PostPresAnim = function( self, anim_fighter )
                        anim_fighter:SetAnimMapping(self.owner.agent.fight_data.anim_mapping_shell)
                    end,

                    OnPostResolve = function( self, battle, attack )
                    end
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter.stat_bounds[ COMBAT_STAT.HEALTH ].min = 1
                    self.bite = self:AddCard("oolo_oshnu_bite")
                    self.blast = self:AddCard("oolo_oshnu_blast")
                    self.blast_pre = self:AddCard("oolo_oshnu_blast_pre")
                    self:SetPattern( self.Cycle )

                    self.unshell = self:AddCard( "oolo_oshnu_unshell" )

                    self.fighter:AddCondition( "oolo_oshnu_defenses" )
                    self.fighter:AddCondition("oolo_oshnu_devour")
                end,

                Unshell = function(self)
                    if self.battle:GetTurns() >= (self.unshell_turn or 0) then
                        if self:ChooseCard( self.unshell ) then
                            self:SetPattern( self.Cycle )
                        end
                    end
                end,

                Cycle = function( self )
                    if math.random() < 0.5 then
                        self:ChooseCard(self.blast)
                    else
                        self:ChooseCard(self.bite)
                    end
                end,

                OnAllySurrender = function( self )
                    if self.fighter:GetStatus() == FIGHT_STATUS.ACTIVE then
                        self.fighter:Surrender()
                    end
                end,

                OnAllyKill = function( self )
                    if self.fighter:GetStatus() == FIGHT_STATUS.ACTIVE then
                        self.fighter:Surrender()
                    end
                end,

                OnAllyRally = function( self, turns, ally )
                    if self.fighter:GetStatus() == FIGHT_STATUS.SURRENDER then
                        self.fighter:Rally()
                    end
                end,
            },

            anim_mapping_shell =
            {
                idle = "shell_idle",
                hit_mid = "shell_idle",
                surrender = "shell_idle",
                stunned = "shell_idle",
            },
        },
    })
)


Content.AddCharacterDef
(
    CharacterDef("MERCHANT_SNAIL",
    {
        base_def = "MONSTER",

        title = "Pack Oshnu",
        build = "merchant_snail",

        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",
        voice_event = "event:/vo/compendium/speak_racing_snail",

        combat_strength = 1,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            battle_scale = 1.25,
            MAX_HEALTH = 40,
            MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
            ally_killed_quip = "battle_pet_killed",

            status_widget_dx = -1.0,
            status_widget_dy = -0.5,
            status_widget_head_dx = 2.0,
            status_widget_head_dy = 5.25,

            formation = FIGHTER_FORMATION.BACK_Z | FIGHTER_FORMATION.BACK_X,
            conditions =
            {
                thick_shell =
                {
                    name = "Thick Shell",
                    desc = "Gain <#HILITE>{DEFEND {1}}</> at the end of their turn.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self.stacks or 1)
                    end,
                    apply_sound = "event:/sfx/battle/status/system/Status_Buff_Defend",     
                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            if self.owner == fighter then
                                self.owner:AddCondition( "DEFEND", self.stacks, self )
                            end
                        end,
                    }
                }
            },

            cards = { "oshnu_bite" },

            attacks = 
            {
                oshnu_slime_attack = table.extend(NPC_ATTACK)
                {
                    name = "Slobber",
                    anim = "bite",

                    flags = FLAGS.MELEE | FLAGS.DEBUFF,

                    OnPostResolve = function( self, battle, attack )
                        for i,hit in ipairs(attack.hits) do
                            if not hit.evaded then
                                if hit.target:IsPlayer() then
                                    hit.target:AddCondition("IMPAIR", 2, self)
                                else
                                    hit.target:AddCondition("gluey_antennae", 1, self)
                                end
                            end
                        end
                    end
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition( "thick_shell", 5 )
                    self.fighter:AddCondition( "DEFEND", 5 )

                    self.attacks = self:MakePicker()
                        :AddID("oshnu_bite", 3)
                        :AddID("oshnu_slime_attack", 1)

                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    return self.attacks:ChooseCard()
                end,
            },
        },

        negotiation_data =
        {
            modifiers =
            {
                SHELL_DEFENSE =
                {
                    name = "Shell Defense",
                    desc = "Every 3rd turn, {1} retracts into its shell, shielding its core argument.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self:GetOwnerName())
                    end,
                    modifier_type = MODIFIER_TYPE.CORE,

                    loc_strings =
                    {
                        SHELL_DEFENSE_DESC = "{SHIELDED}. This argument is shielded until the start of the oshnu's turn!",
                    },

                    event_handlers =
                    {
                        [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                            if negotiator == self.negotiator and minigame:GetTurns() % 3 == 0 then
                                self:SetShieldStatus( true, self.def:GetLocalizedString("SHELL_DEFENSE_DESC") )
                            end
                        end,

                        [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                            if negotiator == self.negotiator and self:GetShieldStatus() then
                                self:SetShieldStatus( false )
                            end
                        end
                    },
                },

                PUPPY_SNAIL_EYES =
                {
                    name = "Puppy Snail Eyes",
                    desc = "All Hostility cards deal {1} less damage. Whenever you play a Hostility card, this deals {2} damage to a random argument.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self.damage_red, self.damage_amt)
                    end,

                    modifier_type = MODIFIER_TYPE.ARGUMENT,

                    max_resolve = 8,
                    target_enemy = TARGET_ANY_RESOLVE,

                    damage_red = 1,
                    damage_amt = 2,

                    event_handlers =
                    {
                        [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion, minigame, target )
                            if source and is_instance(source, Negotiation.Card) and source:IsFlagged(CARD_FLAGS.HOSTILE) then
                                persuasion:AddPersuasion( -self.damage_red, -self.damage_red, self )
                            end
                        end,

                        [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                            if card:IsFlagged(CARD_FLAGS.HOSTILE) then
                                minigame:ApplyPersuasion( self, nil, self.damage_amt, self.damage_amt )
                            end
                        end
                    },
                },
            },
            behaviour =
            {
                OnInit = function( self, difficulty )
                    self.slime = self:AddCard("oshnu_slime")
                    self.eyes = self:AddArgument("PUPPY_SNAIL_EYES")
                    self:SetPattern( self.Cycle )

                    self.negotiator:AddModifier("SHELL_DEFENSE")
                end,

                Cycle = function( self, turns )
                    local core = self.negotiator:FindCoreArgument()
                    if core and not core:GetShieldStatus() then
                        self:ChooseGrowingNumbers( 2, 1, 1 )
                        self:ChooseComposure( 2, 1, 2 )
                    end
                    if (turns - 1) % 3 == 0 then
                        self:ChooseCard(self.slime)
                    end
                    if turns % 2 == 0 and not self.negotiator:HasModifier("PUPPY_SNAIL_EYES") then
                        self:ChooseCard(self.eyes)
                    end
                end, 
            }
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("RACING_SNAIL",
    {
        base_def = "MERCHANT_SNAIL",
        title = "Racing Oshnu",
        build = "racing_snail",
        rel_y_offset = -300,
        rel_x_offset = 100,
        loved_graft = "gluey_acquaintance",
        support_negotiation = "untrained_pet",
        trained_support_negotiation = "trained_pet",
    })
)

Content.AddCharacterDef
(
    CharacterDef("HESHBLADDER",
    {
        base_def = "MONSTER",

        title = "Heshbladder",
        build = "heshbladder",

        death_item = "bladder_gland",

        combat_strength = .5,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = 24,
            conditions =
            {
                GASEOUS_HERD = 
                {
                    name = "Gaseous Herd",
                    desc = "At the end of {1}'s turn, it gains 1 {INFLATED} per friendly Heshbladder alive.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self:GetOwnerName())
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            if fighter == self.owner then
                                local count = 0
                                for i,fighter in self.owner:GetTeam():Fighters() do
                                    if fighter ~= self.owner and fighter:GetAgent():GetContentID() == "HESHBLADDER" then
                                        count = count + 1
                                    end
                                end
                                if count > 0 then
                                    self.owner:AddCondition("INFLATED", count, self)
                                end
                            end
                        end
                    },
                },

                INFLATED =
                {
                    name = "Inflated",
                    desc = "{1} is storing fumes in order to apply {FETOR} to an enemy.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, loc.cap( self:GetOwnerName()))
                    end,
                },

                FETOR = 
                {
                    name = "Fetor",
                    desc = "At the end of {1}'s turn, take <#HILITE>{2} damage</> and remove all <b>Fetor</>.",
                    desc_fn = function( self, fmt_str )
                        return loc.format(fmt_str, self:GetOwnerName(), self.stacks or 1)
                    end,

                    ctype = CTYPE.DEBUFF,

                    CalculatePreviewDamage = function( self )
                        return self.stacks
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                            if fighter == self.owner then
                                self.owner:ApplyDamage( self.stacks, nil, self )
                                self.owner:RemoveCondition(self.id, self.stacks, self)
                            end
                        end
                    },
                },
            },

			fx =
            {
                flead_dust = { auto = true },
            },
            
            attacks =
            {
                bladder_spit = table.extend(NPC_ATTACK)
                {
                    name = "Spit",
                    anim = "spit",

                    hit_tags = {"attack_regular", "snug"},

                    flags = FLAGS.MELEE | FLAGS.DEBUFF,

                    OnPostResolve = function( self, battle, attack )
                        local stacks = self.owner:GetConditionStacks("INFLATED")
                        self.owner:RemoveCondition("INFLATED", stacks, self)
                        attack:AddCondition("FETOR", stacks, self)
                    end,
                    damage_mult = 0.3,
                },

                bladder_grab = table.extend(NPC_ATTACK)
                {
                    name = "Face Pinch",
                    anim = "grab",

                    hit_tags = {"attack_regular", "snug"},

                    flags = FLAGS.MELEE,
                    damage_mult = 0.5,
                },

                bladder_puff_up = table.extend(NPC_BUFF)
                {
                    name = "Puff Up",
                    anim = "taunt",
                    hit_tags = {"attack_regular"},

                    flags = FLAGS.SKILL | FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:AddCondition("DEFEND", 4, self)
                        self.owner:AddCondition("INFLATED", 4, self)
                    end,
                },
            },

            hit_tags = { "small" },

            behaviour =
            {
                OnActivate = function( self )
                    self.charge_up = self:MakePicker()
                        :AddID("bladder_grab", 1)
                        :AddID("bladder_puff_up", 2)

                    self.spit = self:AddCard("bladder_spit")

                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("GASEOUS_HERD")
                end,

                Cycle = function( self )
                    if self.fighter:GetConditionStacks("INFLATED") >= 8 and math.random() < 0.6 then
                        self:ChooseCard(self.spit)
                    else
                        self.charge_up:ChooseCard()
                    end
                end,
            },
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("HESHBLADDER_VIX",
    {
        base_def = "HESHBLADDER",

        title = "Mutant Heshbladder",
        build = "heshbladder_vix",

        combat_strength = .5,

        fight_data = table.extend( MONSTER_DEF.fight_data )
        {
            MAX_HEALTH = 16,
            MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
            battle_scale = 1.1,
            conditions =
            {
            },

            fx =
            {
                flead_dust = { auto = true },
            },
            
            cards = {"bladder_grab", "bladder_puff_up", "bladder_spit"},

            hit_tags = { "small" },

            attacks =
            {
                vix_bladder_spit = table.extend(NPC_ATTACK)
                {
                    name = "Spit",
                    anim = "spit",

                    hit_tags = {"attack_alt", "acid"},

                    flags = FLAGS.MELEE | FLAGS.DEBUFF,

                    OnPostResolve = function( self, battle, attack )
                        local stacks = self.owner:GetConditionStacks("INFLATED")
                        self.owner:RemoveCondition("INFLATED", stacks, self)
                        attack:AddCondition("FETOR", stacks, self)
                    end,
                    damage_mult = 0.3,
                },

                vix_bladder_grab = table.extend(NPC_ATTACK)
                {
                    name = "Face Pinch",
                    anim = "grab",
                    hit_tags = {"attack_alt", "acid"},

                    flags = FLAGS.MELEE,
                    damage_mult = 0.5,
                },

                vix_bladder_puff_up = table.extend(NPC_BUFF)
                {
                    name = "Puff Up",
                    anim = "taunt",
                    hit_tags = {"attack_alt"},

                    flags = FLAGS.SKILL | FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:AddCondition("DEFEND", 6, self)
                        self.owner:AddCondition("INFLATED", 6, self)
                    end,
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.charge_up = self:MakePicker()
                        :AddID("vix_bladder_grab", 1)
                        :AddID("vix_bladder_puff_up", 2)

                    self.spit = self:AddCard("vix_bladder_spit")

                    self:SetPattern( self.Cycle )
                    self.fighter:AddCondition("GASEOUS_HERD")
                end,

                Cycle = function( self )
                    if self.fighter:GetConditionStacks("INFLATED") >= 6 then
                        self:ChooseCard(self.spit)
                    else
                        self.charge_up:ChooseCard()
                    end
                end,
            },
        },
    })
)

Content.AddCharacterDef
(
    CharacterDef("LUMIN_TURRET",
    {
        base_def = "DRONE",
        anims = { "anim/med_combat_drone_tei.zip", },
        combat_anims = { "anim/med_combat_drone_tei.zip" },
        build = "lumin_turret",
        faction_id = "CULT_OF_HESH",
        title = "Luminious Tripod",

        fight_data = table.extend( DRONE_DEF.fight_data )
        {
            height = 0,
            status_widget_head_dx = 0,
            status_widget_head_dy = 2.5,

            anim_mapping =
            {
                taunt = "taunt2",
            },
            attacks = 
            {
                lumin_turret_melee = table.extend(NPC_ATTACK)
                {
                    name = "Smack",
                    anim = "attack1",
                    flags = FLAGS.MELEE,
                    target_count = 2,
                },

                lumin_turret_shoot = table.extend(NPC_ATTACK)
                {
                    name = "Blast",
                    anim = "shoot",
                    flags = FLAGS.RANGED,
                },

                lumin_turret_taunt = 
                {
                    name = "Taunt",
                    anim = "taunt2",
                    flags = FLAGS.SKILL | FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                },
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.fighter:AddCondition("no_alone_drone")
                    self.fighter:AddCondition( "METALLIC" )
                    self.attacks = self:MakePicker()
                        :AddID( "lumin_turret_melee", 1 )
                        :AddID( "lumin_turret_shoot", 2 )
                        :AddID( "lumin_turret_taunt", 1 )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.attacks:ChooseCard()
                end,
            }
        },
    })
)