local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

-- Weakness cards
Content.AddBattleCard( "weakness_slow",
{
    name = "Weakness: Slow",
    desc = "Add 2 {weakness_quick_jab} cards to your hand.",
    anim = "taunt",
    flavour = "'Saw that coming from a mile away.'",

    argument_icon = "negotiation/modifiers/flekfis_weakness_slow.tex",

    cost = 1,

    rarity = CARD_RARITY.UNIQUE,
    flags = CARD_FLAGS.SKILL,
    series = CARD_SERIES.NPC,

    target_type = TARGET_TYPE.SELF,

    OnPostResolve = function( self, battle, attack )
        local cards = {}
        for i=1,2 do
            table.insert(cards, Battle.Card("weakness_quick_jab", self.owner))
        end
        battle:DealCards(cards, battle:GetHandDeck())
    end
})

Content.AddBattleCard( "weakness_quick_jab",
{
    name = "Quick Jab",
    anim = "whip",
    flavour = "'It's like he's underwater!'",

    cost = 0,

    rarity = CARD_RARITY.UNIQUE,
    flags = CARD_FLAGS.MELEE | CARD_FLAGS.EXPEND,
    series = CARD_SERIES.NPC,

    min_damage = 3,
    max_damage = 3,
})

Content.AddBattleCard( "weakness_inflammable", 
{
    name = "Weakness: Inflammable",
    anim = "crackle",
    flavour = "'All that fur and mechanical grease, he was a walking fire hazard to begin with.'",
    desc = "Apply {1} {BURN} and {2} {SCORCHED}.",
    desc_fn = function( self, fmt_str )
        return loc.format(fmt_str, self.burn_amt, self.scorched_amt)
    end,

    argument_icon = "negotiation/modifiers/flekfis_weakness_inflammable.tex",

    cost = 1,

    rarity = CARD_RARITY.UNIQUE,
    flags = CARD_FLAGS.MELEE,
    series = CARD_SERIES.NPC,

    min_damage = 4,
    max_damage = 6,

    burn_amt = 6,
    scorched_amt = 2,

    OnPostResolve = function( self, battle, attack )
        attack:AddCondition("BURN", self.burn_amt, self)
        attack:AddCondition("SCORCHED", self.scorched_amt, self)
    end
})

Content.AddBattleCard( "weakness_old_injury",
{
    name = "Weakness: Old Injury",
    anim = "kick",
    flavour = "'He has been favoring his left side this whole time.'",
    desc = "Apply {1} {IMPAIR}.",
    desc_fn = function( self, fmt_str )
        return loc.format(fmt_str, self.impair_amt)
    end,

    argument_icon = "negotiation/modifiers/flekfis_weakness_old_injury.tex",

    cost = 1,

    rarity = CARD_RARITY.UNIQUE,
    flags = CARD_FLAGS.MELEE,
    series = CARD_SERIES.NPC,

    min_damage = 5,
    max_damage = 8,

    impair_amt = 4,

    OnPostResolve = function( self, battle, attack )
        attack:AddCondition("IMPAIR", self.impair_amt, self)
    end,
})

-- Junk card
Content.AddBattleCard( "flekfis_junk", 
{
    name = "Assorted Junk",
    anim = "taunt",
    flavour = "'That initial explosion sure did hit hard, wonder what I can do with all these...'",
    rarity = CARD_RARITY.UNIQUE,
    series = CARD_SERIES.NPC,

    cost = 0,
    target_type = TARGET_TYPE.SELF,

    flags = CARD_FLAGS.ITEM | CARD_FLAGS.CONSUME | CARD_FLAGS.REPLENISH,
})

local def = CharacterDef("RISE_SECOND",
{
    unique = true,
    base_def = "RISE_BASE",
    alias = "RISE_SECOND",
    title = "Lieutenant",
    renown = 3,
    combat_strength = 3,

    name = "Flekfis",
    build = "male_rise_flekfis_build",
    head = "head_male_rise_flekfis",
    gender = "MALE",
    species = "JARACKLE",
    voice_actor = "flekfis",
    loved_graft = "gear_head",
    hated_graft = "wrenched",
    death_item = "force_spanner",

    theme_music = "event:/music/story/rook/story_rook_fleckfis",

    combat_anims = {"anim/med_combat_flekfis.zip"},
    anims = {"anim/weapon_wrench_flekfis.zip"},

    boss = true,
    battle_preview_anim = "anim/boss_flekfis_slide.zip",
    battle_preview_offset = { x = 300, y = -35 },
    battle_preview_glow = { colour = 0xF08C49FF, bloom = 0.2, threshold = 0.02 },
    battle_preview_audio = "event:/ui/prebattle_overlay/prebattle_overlay_whooshin_boss_flekfis",
    icon = engine.asset.Texture( "icons/b_rise_second.tex"),

    weakness_cards = {"weakness_inflammable", "weakness_slow", "weakness_old_injury"},

    negotiation_data =
    {
        modifiers =
        {
            SLUGGISH =
            {
                name = "Sluggish",
                desc = "Whenever you play a card, reduce the damage of {1}'s intents by 1. Whenever {1}'s attacks are fully mitigated, his core argument takes {2} damage.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOwnerName(), self.self_damage )
                end,

                modifier_type = MODIFIER_TYPE.CORE,
                self_damage = 3,

                event_priorities =
                {
                    [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
                },

                event_handlers =
                {
                    [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion, minigame, target )
                        if source.negotiator == self.negotiator and target and target.negotiator ~= self.negotiator and self.counter and self.counter > 0 then
                            persuasion:AddPersuasion( -self.counter, -self.counter, self )
                        end
                    end,

                    [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                        if negotiator == self.negotiator then
                            self.counter = 0
                            self.active = false
                        end
                    end,

                    [ EVENT.BEGIN_ENEMY_TURN ] = function( self )
                        self.active = true
                    end,

                    [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                        if self.active and source.negotiator == self.negotiator and target.negotiator ~= self.negotiator and damage <= defended then
                            if self.negotiator:FindCoreArgument():GetShieldStatus() then
                                local targets = self.engine:CollectAlliedTargets(self.negotiator:FindCoreArgument())
                                table.arrayremove( targets, self.negotiator:FindCoreArgument() )
                                local self_target = table.arraypick( targets )
                                if self_target then
                                    self.engine:ApplyPersuasion( target, self_target, self.self_damage, self.self_damage )
                                end
                            else
                                self.engine:ApplyPersuasion( target, self.negotiator, self.self_damage, self.self_damage )
                            end
                        end
                    end,

                    [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                        if card.negotiator ~= self.negotiator then
                            self.counter = (self.counter or 0) + 1
                            minigame:BroadcastEvent( EVENT.CUSTOM, function( panel, ... )
                                    panel.intents:RefreshIntents()
                                end )
                        end
                    end
                }
            },

            COLD_SHOULDER =
            {
                name = "Cold Shoulder",
                desc = "At the start of your turn, increase the cost of a random card in your hand by 1 until the end of your turn.",

                modifier_type = MODIFIER_TYPE.ARGUMENT,
                max_resolve = 10,

                event_handlers =
                {
                    [ EVENT.HAND_DRAWN ] = function( self, minigame )
                        self.chosen_card = table.arraypick(minigame:GetHandDeck().cards)
                    end,

                    [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                        if self.chosen_card and card == self.chosen_card then
                            cost_acc:AddValue(1, self)
                        end
                    end
                },
            },
        },

        behaviour =
        {
            OnInit = function( self )
                if CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.NO_CORE_RESOLVE) then
                    self.negotiator:AddModifier("FLEKFIS_WEAKNESS")
                    self:SetPattern( self.BossCycle )
                elseif CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.WORDSMITH) then
                    self.cold_shoulder = self:AddArgument("COLD_SHOULDER")

                    self.negotiator:AddModifier("SLUGGISH")
                    self:SetPattern( self.SilverCycle )
                else
                    self.brainwash = self:AddArgument( "BRAINWASH" )
                    self.bounty = self:AddArgument( "SYMPATHIZER" )
                    self.negotiator:AddModifier("CALL_TO_RISE")
                    self:SetPattern( self.Cycle )
                end
                
            end,

            -- Generic Rise behaviour
            Cycle = function( self, turn )
                self:ChooseGrowingNumbers( 1, 1 )

                if (turn-1) % 3 == 0 and self.negotiator:GetModifierStacks( "brainwash" ) < 1 then
                    self:ChooseCard( self.brainwash )

                elseif (turn-1) % 2 == 0 and self.negotiator:GetModifierStacks( "SYMPATHIZER" ) == 0 then
                    self:ChooseCard( self.bounty )

                else
                    self:ChooseComposure( 1, self.difficulty, self.difficulty )
                end
            end,

            -- Boss behaviour
            BossCycle = function( self, turn )
                self:ChooseComposure( 2, 4, 6 )
                self:ChooseGrowingNumbers( 3, 2)
            end,
            

            SilverCycle = function( self, turns )
                if turns % 3 == 0 then
                    for i=1,6 do
                        self:ChooseFlatNumbers( 1, 2 )
                    end
                else
                    self:ChooseGrowingNumbers( 1, 1)
                end
                if (turns - 1) % 3 == 0 and not self.negotiator:HasModifier("COLD_SHOULDER") then
                    self:ChooseCard(self.cold_shoulder)
                end
                self:ChooseComposure( 2, 4, 6 )
            end
        },
    },

    fight_data = 
    {
        MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
        MAX_HEALTH = 220,
        death_fade_delay = 1.5,

        conditions = 
        {
            flekfis_wrenched = 
            {
                name = "Wrenched",
                desc = "At the start of your turn, increase the cost of a card in your hand by 1. At the end of your turn, remove 1 <b>Wrenched</>.",

                ctype = CTYPE.DEBUFF,

                event_handlers = 
                {
                    [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                        local card = battle:GetHandDeck():PeekRandom()
                        if card then
                            self.affected_card = card
                            card:NotifyTriggered()
                        end
                    end,

                    [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                        if self.affected_card and card == self.affected_card then
                            cost_acc:AddValue(1, self)
                        end
                    end,

                    [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                },
            },

            flekfis_momentum =
            {
                name = "Momentum",
                desc = "{1} is strong but slow. His damage is reduced by 1 for every card you play. At the end of {1}'s turn, reset this to 1.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self.owner:GetName(), self.stacks)
                end,

                event_handlers =
                {
                    [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                        if card.owner == self.owner then
                            dmgt:AddDamage(-self.stacks, -self.stacks, self)
                        end
                    end,

                    [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                        if attack.attacker and attack.attacker == battle:GetPlayerFighter() then
                            self.owner:AddCondition(self.id, 1, self)
                        end
                    end,

                    [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                        if fighter == self.owner then
                            self.owner:RemoveCondition(self.id, self.stacks-1, self)
                        end
                    end
                },
            },

            flekfis_thick_fur =
            {
                name = "Thick Fur",
                desc = "At the end of {1}'s turn, {1} gains {2} {DEFEND}.",
                desc_fn = function(self, fmt_str)
                    return loc.format(fmt_str, self.owner:GetName(), self.defend_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1])
                end,

                defend_amt = { 10, 11, 12, 14},

                event_handlers = 
                {
                    [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                        if fighter == self.owner then
                            self.owner:AddCondition("DEFEND", self.defend_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                        end
                    end
                },
            },
        },
        attacks = 
        {

            flekfis_light_attack = table.extend(NPC_ATTACK)
            {
                name = "Light Attack",
                anim = "light_attack",

                flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

                base_damage = { 9, 10, 11, 12 },
                OnPostResolve = function( self, battle, attack )
                    for i,hit in attack:Hits() do
                        if hit.target:IsPlayer() and not hit.evaded then
                            hit.target:AddCondition("flekfis_wrenched", 1, self)
                        end
                    end
                end
            },

            flekfis_med_attack = table.extend(NPC_ATTACK)
            {
                name = "Wrench",
                anim = "med_attack",

                flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

                base_damage = { 14, 15, 16, 17},

                OnPostResolve = function( self, battle, attack )
                    for i,hit in attack:Hits() do
                        if hit.target:IsPlayer() and not hit.evaded then
                            hit.target:AddCondition("flekfis_wrenched", 1, self)
                        end
                    end
                end
            },

            flekfis_whirlwind = table.extend(NPC_ATTACK)
            {
                name = "Whirlwind",
                anim = "whirlwind",

                flags = CARD_FLAGS.MELEE,
                target_mod = TARGET_MOD.TEAM,

                base_damage = { 4, 5, 6, 7},

                hit_count = 7,
            },

            flekfis_taunt = table.extend(NPC_BUFF)
            {
                name = "Defend",
                anim = "taunt2",

                flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
                target_type = TARGET_TYPE.SELF,

                defend_amt = { 6, 6, 8, 10},

                OnPostResolve = function( self, battle, attack )
                    self.owner:AddCondition("DEFEND", self.defend_amt[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                end
            },

            flekfis_trap = table.extend(NPC_RANGED)
            {
                name = "Activate Trap",
                anim = "taunt1",

                flags = CARD_FLAGS.RANGED | CARD_FLAGS.DEBUFF,
                target_mod = TARGET_MOD.TEAM,

                base_damage = { 10, 10, 11, 12},

                junk_count = { 6, 6, 5, 4},

                OnPostResolve = function( self, battle, attack )
                    self.owner.behaviour.has_trapped = true
                    for i,hit in attack:Hits() do
                        if hit.target:IsPlayer() then
                            local cards = {}
                            for i=1,self.junk_count[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1] do
                                table.insert(cards, Battle.Card("flekfis_junk", hit.target))
                            end
                            battle:DealCards(cards, battle:GetDiscardDeck())
                        end
                    end
                end
            },

            flekfis_ambush = table.extend(NPC_MELEE)
            {
                name = "First Strike",
                anim = "med_attack",

                CanPlayCard = function( self, battle, target )
                    return target ==  nil or target:IsPlayer()
                end,

                base_damage = {0, 0, 0, 0},

                event_handlers = 
                {
                    [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                        if card == self then
                            local damage = math.ceil(math.max(target:GetMaxHealth() * (target:GetHealthPercent() - .65), 8))
                            dmgt:AddDamage(damage, damage, self)
                        end
                    end
                },
            },
        },


        behaviour =
        {
            OnActivate = function( self, fighter )
                self.attacks = self:MakePicker()
                    :AddID( "flekfis_light_attack", 2)
                    :AddID( "flekfis_med_attack", 2)
                    :AddID( "flekfis_taunt", 1)

                self.whirlwind = self:AddCard( "flekfis_whirlwind")

                self.trap = self:AddCard("flekfis_trap")

                self.ambush_attack = self:AddCard("flekfis_ambush")

                self.fighter:AddCondition("flekfis_momentum", 1, self)
                self.fighter:AddCondition("flekfis_thick_fur", 1, self)
                self:SetPattern( self.Cycle )
            end,

            Cycle = function( self )
                if self.engine:IsAmbushing(self.fighter:GetTeam()) then
                    self:ChooseCard(self.ambush_attack)
                else
                    if not self.has_trapped then
                        self:ChooseCard(self.trap)
                    else
                        if self.battle:GetTurns() % 4 == 0 then
                            self:ChooseCard(self.whirlwind)
                        else
                            self.attacks:ChooseCard()
                        end
                    end
                end
            end,

        },
        anim_mapping =
        {
            taunt = "taunt2"
        },
    },
})

Content.AddCharacterDef( def )