local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

Content.AddNegotiationCard("war_story", 
{
    name = "War Story",
    flavour = "'That’s not a scar, I’ll show you a scar!'",
    desc = "{GAMBLE}.\n{HEADS}: Gain {INFLUENCE 5}.\n{SNAILS}: Gain {DOMINANCE 4}.",
    loc_strings =
    {
        ALT_DESC = "Randomly gain\n{INFLUENCE 5} or {DOMINANCE 4}.",
    },
    desc_fn = function( self, fmt_str )
        local content_id
        if self.owner then
            content_id = self.owner:GetContentID()
        elseif TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent() then
            content_id = TheGame:GetGameState():GetPlayerAgent():GetContentID()
        end

        if content_id == nil or content_id == "ROOK" then
            return fmt_str
        else
            return self.def:GetLocalizedString( "ALT_DESC" )
        end
    end,

    cost = 1,
    flags = negotiation_defs.CARD_FLAGS.MANIPULATE | negotiation_defs.CARD_FLAGS.EXPEND,

    rarity = CARD_RARITY.UNIQUE,
    series = CARD_SERIES.GENERAL,

    OnPostResolve = function( self, minigame )
        local coin = self.negotiator:FindModifier("LUCKY_COIN")
        local result = coin and coin:Gamble(self) or math.random() <.5 and "HEADS"
        if result == "HEADS" then
            self.negotiator:AddModifier("INFLUENCE", 5, self)
        else
            self.negotiator:AddModifier("DOMINANCE", 4, self)
        end
    end,
})

local def = CharacterDef("RENTORIAN_BOSS",
{
    unique = true,

    base_def = "NPC_BASE",
    species = SPECIES.SHROKE,
    gender = GENDER.MALE,
    renown = 3,
    combat_strength = 2,
    boss = true,
    battle_preview_anim = "anim/boss_rentorian_slide.zip",
    battle_preview_offset = { x = -330, y = 0, scale = 0.85 },
    battle_preview_glow = { colour = 0xfdcd90FF, bloom = 0.10, threshold = 0.02 },
    battle_preview_audio = "event:/ui/prebattle_overlay/prebattle_overlay_whooshin_boss_rentorian",
    icon = engine.asset.Texture( "icons/b_rentorian_boss.tex"),
    death_item = "war_story",

    faction_id = "RENTORIAN",

    title = "Hunter",
    voice_actor = "rentorian_boss",

    loved_graft = "rentorian_battle_armor",
    hated_graft = "informed_enemies",
    
    combat_anims = {"anim/med_combat_rentorian.zip"},
    anims = {"anim/weapon_double_bladed_staff_rentorian.zip"},
    
    build = "male_rentorian",
    head = "head_male_rentorian",

    support_negotiation = "formality",

    negotiation_data = 
    {
        modifiers =
        {

        },
        behaviour = 
        {
            OnInit = function( self, difficulty )
                self.moment_of_weakness = self:AddCard( "moment_of_weakness" )
                self.negotiator:AddModifier( "UNCERTAINTY" )
                self:SetPattern( self.BasicCycle )
            end,

            BasicCycle = function( self, turns )
                if not self.player_negotiator:HasModifier("moment_of_weakness") then
                    self:ChooseCard( self.moment_of_weakness )
                    self:ChooseGrowingNumbers( 2 )
                else
                    self:ChooseGrowingNumbers( 2, 2 )
                    if math.random() < .5 then
                        self:ChooseComposure( 1, 2, 2 )
                    end
                end
            end,
        }
    },
    fight_data = 
    {
        MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
        MAX_HEALTH = 90,
        battle_scale = 1.1,

        conditions = 
        { 
            overzealous = 
            {
                name = "Overzealous",
                desc = "When {1}'s attacks cause you to lose health, gain {WOUND}. {1} also takes 20% more damage.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, loc.cap( self:GetOwnerName() ))
                end,

                event_handlers = 
                {
                    [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                        if attack.attacker == self.owner and not hit.defended and not hit.evaded then
                            hit.target:AddCondition("WOUND", 1, self)
                        end
                    end,

                    [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                        if target and target == self.owner then
                            dmgt:ModifyDamage( math.round( dmgt.min_damage * 1.2 ),
                                               math.round( dmgt.max_damage * 1.2 ),
                                               self )
                        end
                    end,
                },
            },

            cautious =
            {
                name = "Cautious",
                desc = "Gains {1} {DEFEND} and {2} {POWER} at the end of every turn. Loses {2} {POWER} whenever all of his {DEFEND} is broken.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self.defend_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self.power_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1])
                end,

                defend_amount = {6, 6, 6, 6},
                power_amount = {3, 3, 3, 4},

                event_handlers = 
                {
                    [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                        if fighter == self.owner then
                            self.owner:AddCondition("POWER", self.power_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                            self.owner:AddCondition("DEFEND", self.defend_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                        end
                    end,

                    [ BATTLE_EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
                        if fighter == self.owner and condition:GetID() == "DEFEND" and condition:GetStacks() == 0 and source and source.owner ~= self.owner then
                            local count = math.min(self.owner:GetConditionStacks("POWER"), self.power_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1])
                            self.owner:RemoveCondition("POWER", count, self)
                        end
                    end,
                },
            },

            stance_counter = 
            {
                hidden = true,
                event_handlers =
                {
                    [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                        if self.owner == fighter then
                            self.owner:RemoveCondition(self.id, 1)
                        end
                    end
                },
            },
        },

        attacks = 
        {
            -- Dual Blade attacks
            dual_blade_flurry = table.extend(NPC_MELEE)
            {
                anim = "attack2_dual",
                flags = CARD_FLAGS.MELEE,

                base_damage = { 4, 4, 5, 6},
                hit_count = 2,
            },

            dual_blade_slice = table.extend(NPC_MELEE)
            {
                anim = "attack_dual",
                flags = CARD_FLAGS.MELEE,

                base_damage = { 7, 8, 10, 12 },
            },

            dual_blade_buff = table.extend(NPC_BUFF)
            {
                anim = "taunt2",
                flags = CARD_FLAGS.BUFF | CARD_FLAGS.SKILL,
                target_type = TARGET_TYPE.SELF,

                power_amount = { 2, 2, 3, 4},
                defend_amount = { 5, 6, 8, 8},

                OnPostResolve = function( self, battle, attack )
                    self.owner:AddCondition("POWER", self.power_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                    self.owner:AddCondition("DEFEND", self.defend_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                end
            },

            halberd_switch = table.extend(NPC_BUFF)
            {
                anim = "transition2",
                flags = CARD_FLAGS.BUFF | CARD_FLAGS.SKILL,
                target_type = TARGET_TYPE.SELF,

                PostPresAnim = function( self, anim_fighter )
                    anim_fighter:SetAnimMapping(nil)
                end,

                defend_amount = {6, 8, 10, 12},

                OnPostResolve = function( self, battle, attack )
                    self.owner.behaviour:SwapStances()
                    self.owner:AddCondition("cautious", 1, self)
                    self.owner:RemoveCondition("overzealous", 1, self)
                    self.owner:AddCondition("DEFEND", self.defend_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                end
            },
            -- Halberd attacks
            halberd_swipe = table.extend(NPC_MELEE)
            {
                anim = "attack1",
                flags = CARD_FLAGS.MELEE,

                base_damage = { 4, 5, 6, 8 },
                target_mod = TARGET_MOD.TEAM,
            },

            halberd_stab = table.extend(NPC_MELEE)
            {
                anim = "attack2",
                flags = CARD_FLAGS.MELEE,

                base_damage = { 5, 7, 9, 10 },
            },

            halberd_buff = table.extend(NPC_BUFF)
            {
                anim = "taunt",
                flags = CARD_FLAGS.BUFF | CARD_FLAGS.SKILL,
                target_type = TARGET_TYPE.SELF,

                defend_amount = {4, 6, 8, 10},

                OnPostResolve = function( self, battle, attack )
                    self.owner:AddCondition("DEFEND", self.defend_amount[GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_BOSS_DIFFICULTY ) or 1], self)
                end
            },

            dual_blade_switch = table.extend(NPC_BUFF)
            {
                anim = "transition1",
                flags = CARD_FLAGS.BUFF | CARD_FLAGS.SKILL,
                target_type = TARGET_TYPE.SELF,

                PostPresAnim = function( self, anim_fighter )
                    anim_fighter:SetAnimMapping(self.owner.agent.fight_data.anim_mapping_dual)
                end,

                OnPostResolve = function( self, battle, attack )
                    self.owner.behaviour:SwapStances()
                    self.owner:AddCondition("overzealous", 1, self)
                    self.owner:RemoveCondition("cautious", 1, self)
                    self.owner:AddCondition("RIPOSTE", 4, self)
                end
            },
        },


        behaviour =
        {
            OnActivate = function( self, fighter )
                self.dual_blade_attacks = self:MakePicker()
                    :AddID("dual_blade_flurry", 2)
                    :AddID("dual_blade_slice", 3)
                    :AddID("dual_blade_buff", 1)
                self.switch_to_halberd = self:AddCard("halberd_switch")

                self.halberd_attacks = self:MakePicker()
                    :AddID("halberd_swipe", 2)
                    :AddID("halberd_stab", 3)
                    :AddID("halberd_buff", 1)
                self.switch_to_dual = self:AddCard("dual_blade_switch")

                self.fighter:AddCondition("stance_counter", 3)
                self.fighter:AddCondition("cautious", 1, self)
                self:SetPattern(self.Halberd)
            end,

            DualBlade = function( self )
            	if self.fighter:HasCondition("stance_counter") then
                    self.dual_blade_attacks:ChooseCards(1)
                else
                    self:ChooseCard(self.switch_to_halberd)
                end
            end,

            Halberd = function( self )
                if self.fighter:HasCondition("stance_counter") then
                    self.halberd_attacks:ChooseCards(1)
                else
                    self:ChooseCard(self.switch_to_dual)
                end
            end,

            SwapStances = function( self )
                self.fighter:AddCondition("stance_counter", 4, self)
                if self.current_pattern == self.DualBlade then
                    self:SetPattern(self.Halberd)
                else
                    self:SetPattern(self.DualBlade)
                end
            end
        },

        anim_mapping_dual =
        {
            idle = "idle2",
            defend = "defend2",
            death = "death2",
            hit_mid = "hit_mid2",
            hit_mid_pst_idle = "hit_mid2_pst_idle",
            hit_mid_pst_stunned = "hit_mid2_pst_stunned",
            hit_mid_pst_surrender = "hit_mid2_pst_surrender",
            run_forward = "run_forward2",
            run_forward_pre = "run_forward2_pre",
            run_forward_pre_turn = "run_forward2_pre_turn",
            run_forward_pst = "run_forward2_pst",
            run_forward_pst_turn = "run_forward2_pst_turn",
            step_back = "step_back2",
            step_forward = "step_forward2",
            stunned = "stunned2",
            stunned_pre = "stunned2_pre",
            stunned_pst = "stunned2_pst",
            surrender_pre = "surrender2_pre",
            surrender_pst = "surrender2_pst",
            surrender = "surrender2",
            riposte = "attack_dual",
        },
    },
})

Content.AddCharacterDef( def )
    