local battle_defs = require "battle/battle_defs"
local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

--------------------------------------------------
-- The Rise

Content.AddCharacterDef
(
    CharacterDef("RISE_BASE",
    {
        designer_note = "** UNUSED **",
        base_def = "NPC_BASE",
        is_template = true,
        combat_backup = "RISE_BACKUP",

        title = "Rebel",

        faction_id = "RISE",
        tags = { "rise" },

        support_negotiation = "rabble_rouser",

        base_builds = {
            [ GENDER.MALE ] = "rise_male",
            [ GENDER.FEMALE ] = "rise_female"
        },

        negotiation_data =
        {
            modifiers =
            {
                CALL_TO_RISE =
                {
                    name = "Call to Rise",
                    desc = "At the start of each turn, a random card in hand costs 1 extra Action this negotiation.",
                    modifier_type = MODIFIER_TYPE.CORE,

                    OnApply = function( self )
                        self.card_costs = {}
                    end,

                    RemoveCard = function( self, card )
                        while table.arrayremove( self.card_costs, card ) do end
                    end,

                    event_priorities =
                    {
                        [ BATTLE_EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_ADDITIVE,
                    },

                    CanChoose = function( card )
                        return card:IsPlayable() and not card:IsFlagged(battle_defs.CARD_FLAGS.VARIABLE_COST)
                    end,

                    event_handlers = 
                    {
                        [ EVENT.HAND_DRAWN ] = function( self, minigame )
                            local card = minigame:GetHandDeck():PeekRandom( self.CanChoose )
                            if card then
                                table.insert( self.card_costs, card )
                                card:NotifyTriggeredPre()
                                self:NotifyTriggered()
                                card:NotifyTriggeredPost()
                            end
                        end,

                        [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                            if table.contains(self.card_costs, card) then
                                local count = 0
                                for i, affected_card in ipairs(self.card_costs) do
                                    if affected_card == card then
                                        count = count + 1
                                    end
                                end
                                cost_acc:AddValue( count, self )
                            end
                        end,
                    },
                },

                SYMPATHIZER =
                {
                    name = "Sympathizer",
                    desc = "Gain <#HILITE>{1}</> actions.\nDraw a card for each friend you have in the Rise. ({2} friends).",

                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, (self.stacks or 1) * self.action_bonus, self:CountRiseFriends() )
                    end,

                    sound = "event:/sfx/battle/cards/neg/create_argument/scruple",

                    modifier_type = MODIFIER_TYPE.BOUNTY,
                    action_bonus = 2,

                    CountRiseFriends = function( self )
                        local function IsRiseFriend( agent, rel )
                            return agent:GetFactionID() == "RISE" and rel >= RELATIONSHIP.LIKED
                        end
                        return self.engine:GetPlayer().social_connections:CountConnections( IsRiseFriend )
                    end,

                    OnInit = function( self, source )
                        self:SetResolve( 2, MODIFIER_SCALING.LOW )
                    end,

                    OnBounty = function( self )
                        self.engine:ModifyActionCount( self.action_bonus * self.stacks, self )
                        local draw = self:CountRiseFriends()
                        if draw > 0 then
                            self.engine:DrawCards( draw )
                        end
                    end,
                },

                BRAINWASH =
                {
                    name = "Brainwash",
                    -- sound = "event:/sfx/battle/cards/neg/create_argument/interrogate",

                    desc = "Insert {1} {card.propaganda} card into {2}'s draw pile at the beginning of {3}'s turn.",

                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, self.stacks, self.anti_negotiator:GetName(), self.negotiator.agent:GetName() )
                    end,

                    OnInit = function( self )
                        self:SetResolve( 1, MODIFIER_SCALING.LOW )
                    end,

                    hide_stacks = true,
                    max_stacks = 1,
                    modifier_type = MODIFIER_TYPE.ARGUMENT,

                    OnBeginTurn = function ( self, minigame )
                        local card = Negotiation.Card( "propaganda", self.anti_negotiator.agent )
                        minigame:InceptCard( card, self )
                    end
                },

                BRAINWASH_plus =
                {
                    desc = "Insert {1} {card.propaganda} card into {2}'s draw pile at the beginning of {3}'s turn.\n\nWhen this argument is destroyed, {3} restores 5 resolve for each {card.propaganda} in your hand.",
                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, self.stacks, self.anti_negotiator:GetName(), self.negotiator.agent:GetName(), self:CountPropaganda() )
                    end,

                    CountPropaganda = function (self )
                        local count = self.engine:GetHandDeck():CountCards( "propaganda" )
                        return count
                    end,

                    OnBounty = function( self )
                        local count = self:CountPropaganda()
                        if count > 0 then
                            self.negotiator:RestoreResolve( 5 * count, self )
                        end
                    end,
                }
            },

            behaviour =
            {
                OnInit = function( self, difficulty )
                    self:SetPattern( self.Cycle )

                    self.brainwash = self:AddArgument( "BRAINWASH" )
                    self.bounty = self:AddArgument( "SYMPATHIZER" )
                    self.negotiator:AddModifier("CALL_TO_RISE")
                end,

                Cycle = function( self, turns )
                    self:ChooseGrowingNumbers( 1, 1 )

                    if (turns-1) % 3 == 0 and self.negotiator:GetModifierStacks( "brainwash" ) < 1 then
                        self:ChooseCard( self.brainwash )

                    elseif (turns-1) % 2 == 0 and self.negotiator:GetModifierStacks( "SYMPATHIZER" ) == 0 then
                        self:ChooseCard( self.bounty )

                    else
                        self:ChooseComposure( 1, self.difficulty, self.difficulty )
                    end
                end,
            }
        }
    })
)

--------------------------


--rise footsoldier
Content.AddCharacterDef
(
    CharacterDef("RISE_REBEL",
    {
        base_def = "RISE_BASE",

        title = "Rebel",
        renown = 2,
        combat_strength = 1.5,

        faction_id = "RISE",
        tags = { "rise" },
        promotion_def = "RISE_REBEL_PROMOTED",

        base_builds = {
            [ GENDER.MALE ] = "rise_male_rebel",
            [ GENDER.FEMALE ] = "rise_female_rebel"
        },

        anims = {"anim/weapon_axe_rise_rebel.zip"},
        combat_anims = { "anim/med_combat_axe_rise_rebel.zip" },
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_rise_rebel.tex"),

        loved_graft = "morale_boost",
        hated_graft = "civil_unrest",
        death_item = "guerrilla_tactic",

        fight_data =
        {
            ignore_anim_symbols = {"axe"},
            MAX_MORALE =  0.5,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,

            attacks =
            {
                rebel_slash = table.extend(NPC_ATTACK)
                {
                    name = "Frenzied Attack",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.MELEE,
                },
                rebel_slash_call_in = table.extend(NPC_ATTACK)
                {
                    name = "Frenzied Attack",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.MELEE | battle_defs.CARD_FLAGS.SPECIAL,

                    damage_mult = 1.1,

                    features =
                    {
                        IMPAIR = 2,
                    },

                    OnPostResolve = function( self, battle, attack )
                        self.owner:Depart()
                    end
                },
                rallying_blow = table.extend(NPC_ATTACK)
                {
                    name = "Rallying Blow",
                    anim = "swing",
                    flags = battle_defs.CARD_FLAGS.MELEE,

                    OnPostResolve = function( self, battle, attack )
                        if attack:GetTotalDamage() > 0 then
                            for i, fighter in self.owner:GetTeam():Fighters() do
                                if fighter:HasMorale() then
                                    local max_delta = math.max( 0, fighter:GetMorale() - 0.1 )
                                    local morale = math.min( max_delta, 0.5 * attack:GetTotalDamage() / fighter:GetMaxHealth())
                                    morale = morale * (1.0 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                                    if morale > 0 then
                                        fighter:DeltaMorale( -morale )
                                    end
                                end
                            end
                        end
                    end
                },
                rise_frenzy = table.extend(NPC_BUFF)
                {
                    name = "Frenzy",
                    anim = "taunt",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.BUFF | battle_defs.CARD_FLAGS.HEAL,
                    target_type = TARGET_TYPE.SELF,


                    OnPostResolve = function( self, battle, attack )
                        local health = math.ceil( self.owner:GetMaxHealth() * 0.25 )
                        self.owner:HealHealth( health, self )

                        attack:AddCondition( "POWER", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end,
                },

                rebel_smoke_bomb = table.extend(NPC_BUFF)
                {
                    name = "Smoke Bomb",
                    anim = "throw",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                    target_mod = TARGET_MOD.TEAM,

                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition("EVASION", 2, self)
                    end
                },
            },
            behaviour =
            {
                OnActivate = function( self )
                    self.basic = self:MakePicker()
                        :AddID( "rallying_blow", 1 )
                        :AddID( "ai_defend_lo", 1 )

                    self.frenzy_cycle = self:MakePicker()
                        :AddID( "rebel_slash", 2 )

                    self.frenzy = self:AddCard( "rise_frenzy" )

                    self:SetPattern( self.Basic )
                end,

                Basic = function( self )
                    if self.fighter:GetMorale() and self.fighter:GetMorale() <= 0 then
                        if self.frenzy then
                            self:ChooseCard( self.frenzy )
                            self.frenzy = nil
                        else
                            self.frenzy_cycle:ChooseCard( 1 )
                        end
                    else
                        self.basic:ChooseCard( 1 )
                    end
                end,
            }
        },

    })
)

Content.AddCharacterDef
(
    CharacterDef("RISE_REBEL_PROMOTED",
    {
        base_def = "RISE_REBEL",

        title = "Promoted Rebel",
        renown = 2,

        faction_id = "RISE",
        tags = { "rise" },

        base_builds = {
        [ GENDER.MALE ] = "male_rise_promoted_build",
        [ GENDER.FEMALE ] = "female_rise_promoted_build"
        },

        anims = {"anim/weapon_grenade_promoted_rise.zip"},
        combat_anims = { "anim/med_combat_unarmed_promoted_rise.zip" },

        death_item = "guerrilla_tactic_plus",

        fight_data =
        {
            MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,

            attacks =
            {
                rise_rebel_throw = table.extend(NPC_ATTACK)
                {
                    name = "Throw",
                    anim = "throw",
                    flags = battle_defs.CARD_FLAGS.RANGED,
                    hit_tags = { "burn" },
                },

                rise_rebel_double_throw = table.extend(NPC_ATTACK)
                {
                    name = "Double Throw",
                    anim = "throw2",
                    flags = battle_defs.CARD_FLAGS.RANGED,
                    hit_tags = { "burn" },
                    target_count = 2,
                    damage_mult = .6,
                },
            },
            cards = {"rise_rebel_throw", "rise_rebel_double_throw", "ai_defend_med"},
            behaviour =
            {
                OnActivate = function( self )
                    self.basic = self:MakePicker()
                        :AddID( "rise_rebel_throw", 2 )
                        :AddID( "rise_rebel_double_throw", 1 )
                        :AddID( "ai_defend_med", 1 )

                    self:SetPattern( self.Basic )
                end,

                Basic = function( self )
                    self.basic:ChooseCard( 1 )
                end,
            }
        },

    })
)

--rise clerk-ish
Content.AddCharacterDef
(
    CharacterDef("RISE_PAMPHLETEER",
    {
        base_def = "RISE_BASE",

        title = "Pamphleteer",
        renown = 3,
        combat_strength = 2,

        faction_id = "RISE",
        tags = { "rise" },

        death_item = "rise_manifesto",
        loved_graft = "propaganda_machine",
        hated_graft = "slander",
        
        base_builds = {
            [ GENDER.MALE ] = "male_rise_pamphleteer",
            [ GENDER.FEMALE ] = "female_rise_pamphleteer",--"female_rise_pamphleteer" not done yet
        },

        anims = table.empty,
        combat_anims = { "anim/med_combat_unarmed_rise_pamphleteer.zip" },
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_rise_pamphleteer.tex"),

        fight_data =
        {
            MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,

            attacks =
            {
                can_opener = table.extend(NPC_ATTACK)
                {
                    name = "Can Opener",
                    anim = "stun",
                    flags = battle_defs.CARD_FLAGS.DEBUFF | battle_defs.CARD_FLAGS.MELEE,

                    damage_mult = 0.5,

                    features =
                    {
                        DEFECT = 1,
                    }
                },

                can_opener_call_in =table.extend(NPC_ATTACK)
                {
                    name = "Can Opener",
                    anim = "stun",
                    flags = battle_defs.CARD_FLAGS.DEBUFF | battle_defs.CARD_FLAGS.MELEE | battle_defs.CARD_FLAGS.SPECIAL,

                    damage_mult = 0.5,

                    features =
                    {
                        DEFECT = 1,
                    },
                    OnPostResolve = function( self, battle, attack )
                        self.owner:Depart()
                    end
                },
                feeble_swing = table.extend(NPC_ATTACK)
                {
                    name = "Swing",
                    anim = "punch",
                    damage_mult = 1,

                    flags = battle_defs.CARD_FLAGS.MELEE,
                },
                incite = table.extend(NPC_BUFF)
                {
                    name = "Incite",
                    desc = "Apply 3 {POWER} to an ally with less than 50% health.",
                    anim = "taunt",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.FRIENDLY_OR_SELF,

                    CanPlayCard = function( self, battle, target )
                        if target then
                            return target:GetHealthPercent() < 0.5
                        else
                            return true
                        end
                    end,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:SayQuip( "battle_rally" )
                        attack:AddCondition( "POWER", 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end
                }
            },
            behaviour =
            {
                OnActivate = function( self )
                    self.moves = self:MakePicker()
                        :AddID( "incite", 1 )
                        :AddID( "feeble_swing", 2 )
                        :AddID( "can_opener", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                        :AddID( "ai_defend_med", 1 )

                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.moves:ChooseCard( 1 )
                end,
            }
        },

    })
)


--rise automech
Content.AddCharacterDef
(
    CharacterDef("RISE_AUTOMECH",
    {
        base_def = "RISE_BASE",

        title = "Cobblebot",
        renown = 1,
        combat_strength = 3,
        can_talk = true,
        species = SPECIES.MECH,
        head = "head_automech_rise_01",

        faction_id = "RISE",
        tags = { "rise" },

        base_builds = {
            [ GENDER.UNDISCLOSED ] = "automech_rise_01",
        },
        voice_actor = "robotBroken01",
        
        death_item = "overloaded_core",

        aspects = {
            social_connections = false,
        },

        anims = {"anim/weapon_club_automech_rise.zip"},
        combat_anims = { "anim/med_combat_automech_rise.zip" },
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_rise_cobblebot.tex"),

        loc_strings =
        {
            ACQUIRING_TARGET = "Acquiring target.",
            TARGET_LOCKED = "Target locked.",
        },

        fight_data =
        {
            ignore_anim_symbols = {"club", "club_glow", "club_layer_1"},
            MAX_MORALE = MAX_MORALE_LOOKUP.IMMUNE,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_LOW,
            death_fade_delay = 1,
            
            attacks =
            {
                auto_targetting =
                {
                    name = "Auto-Targetting",
                    anim = "taunt",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.DEBUFF,
                    target_type = TARGET_TYPE.ENEMY,

                    OnPostResolve = function( self, battle, attack )
                        -- Remove existing reticles.
                        for i, fighter in self.owner:GetEnemyTeam():Fighters() do
                            fighter:RemoveCondition( "RETICLE" )
                        end
                        local stacks = math.floor( 2 + self.owner:GetScale()/2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                        attack.target:AddCondition( "RETICLE", stacks, self )
                        self.owner:SaySpeech( 0.5, self.owner.agent:GetLocalizedString( "TARGET_LOCKED" ))
                    end,
                },

                cobblebot_scan =
                {
                    name = "Scan Target",
                    anim = "scan",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.DEBUFF,
                    target_type = TARGET_TYPE.ENEMY,

                    CanPlayCard = function( self, battle, target )
                        if target then
                            if not target:HasCondition( "RETICLE" ) then
                                return false
                            end
                            if target:HasCondition( "DEFECT") then
                                return false
                            end
                        end
                        return true
                    end,

                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition( "DEFECT", 1 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end,

                    features =
                    {
                        DEFEND = 4,
                    }
                },

                mech_cross = table.extend(NPC_ATTACK)
                {
                    name = "Cross",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.MELEE,

                    CanPlayCard = function( self, battle, target )
                        if target and not target:HasCondition( "RETICLE" ) then
                            return false
                        end
                        return true
                    end
                },
                mech_cross_call_in = table.extend(NPC_ATTACK)
                {
                    name = "Cross",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.MELEE | battle_defs.CARD_FLAGS.SPECIAL,

                    damage_mult = 1.75,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:Depart()
                    end
                },
                mech_taunt = table.extend(NPC_BUFF)
                {
                    name = "Defend",
                    anim = "taunt",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.BUFF,
                    target_type = TARGET_TYPE.SELF,
                    CanPlayCard = function( self, battle, target )
                        if self.owner:GetTeam():NumActiveFighters() > 1 then
                            for i,enemy in self.owner:GetEnemyTeam():Fighters() do
                                if enemy:HasCondition("RETICLE") then
                                    return true
                                end
                            end
                        end

                        return false
                    end,
                    features = 
                    {
                        DEFEND = 5,
                    },
                },
            },

            conditions =
            {
                RETICLE =
                {
                    name = "Reticle",
                    desc = "{1} will attack exclusively this fighter.\n\n1 <b>Reticle</b> is removed when this fighter is the target of any Maneuver card.",
                    ctype = CTYPE.DEBUFF,
                    desc_fn = function( self, fmt_str )
                        local applicant = self:GetRecentFighterApplicant()
                        return loc.format( fmt_str, applicant and applicant:GetName(), self.stacks or 1 )   
                    end,

                    event_handlers =
                    {
                        [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                            if attack:IsTarget( self.owner ) and attack.card:IsFlagged( battle_defs.CARD_FLAGS.SKILL ) and
                                attack.attacker:GetTeam() == self.owner:GetTeam() then
                                --
                                self.owner:RemoveCondition( "RETICLE", 1 )
                            end
                        end,
                    }
                }
            },

            behaviour =
            {
                OnActivate = function( self )
                    self.retarget = self:AddCard( "auto_targetting" )
                    self.moves = self:MakePicker()
                        :AddID( "cobblebot_scan", 1 )
                        :AddID( "mech_cross", 2 )
                        :AddID( "mech_taunt", 1 )

                    self.fighter:AddCondition( "METALLIC" )
                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    if self.moves:ChooseCard( 1 ) == 0 then
                        self:ChooseCard( self.retarget )
                        self.fighter:SaySpeech( 0, self.fighter.agent:GetLocalizedString( "ACQUIRING_TARGET" ) )
                    end
                end,

            }
        },

    })
)



--rise elite
Content.AddCharacterDef
(
    CharacterDef("RISE_RADICAL",
    {
        base_def = "RISE_BASE",

        title = "Radical",
        renown = 3,
        combat_strength = 4,

        loved_graft = "war_machine",
        hated_graft = "depleted",

        faction_id = "RISE",
        tags = { "rise" },

        base_builds = {
            [ GENDER.MALE ] = "rise_male",
            [ GENDER.FEMALE ] = "rise_female"
        },

        anims = {"anim/weapon_bazooka_rise_radical.zip"},
        combat_anims = { "anim/med_combat_bazooka_rise_radical.zip" },
        negotiation_ally_image = engine.asset.Texture( "negotiation/modifiers/recruit_rise_radical.tex"),

        death_item = "rpg",

        fight_data =
        {
            ignore_anim_symbols = {"bazooka"},
            MAX_MORALE = MAX_MORALE_LOOKUP.VERY_HIGH,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM_HIGH,

            attacks =
            {
                fire_in_the_hole = table.extend(NPC_ATTACK)
                {
                    name = "Fire in the Hole",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.RANGED,

                    target_mod = TARGET_MOD.TEAM,
                },
                fire_in_the_hole_call_in = table.extend(NPC_ATTACK)
                {
                    name = "Fire in the Hole",
                    anim = "attack1",
                    flags = battle_defs.CARD_FLAGS.RANGED | battle_defs.CARD_FLAGS.SPECIAL,

                    target_mod = TARGET_MOD.TEAM,

                    OnPostResolve = function( self, battle, attack )
                        self.owner:Depart()
                    end
                },
                disrupt = table.extend(NPC_ATTACK)
                {
                    name = "Disrupt",
                    desc = "Your hand size is reduced by <#HILITE>{1}</>. ({2} max cards).\nAt the beginning of your turn, reduce <b>Disrupt</b> by 1.",
                    desc_fn = function( self, fmt_str )
                        return loc.format( fmt_str, self.stacks, (self.battle and self.battle.MAX_HAND_SIZE or 10) - self.stacks )
                    end,
                    anim = "swing",
                    flags = battle_defs.CARD_FLAGS.MELEE | battle_defs.CARD_FLAGS.DEBUFF,

                    CanPlayCard = function( self, battle, target )
                        if target then
                            if not target:IsPlayer() then
                                return false
                            end
                            if target:GetConditionStacks( "disrupt" ) >= 5 then
                                return false
                            end
                        end
                        return true
                    end,

                    OnPostResolve = function( self, battle, attack )
                        attack:AddCondition( "disrupt", 2 + GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                    end,

                    condition =
                    {
                        ctype = CTYPE.DEBUFF,

                        OnSetStacks = function( self, num, old_stacks, stacks )
                            local deck = self.battle:GetHandDeck()
                            deck:SetMaxSize( self.battle.MAX_HAND_SIZE - num )
                        end,

                        event_handlers =
                        {
                            [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                                if self.owner == fighter then
                                    self.owner:RemoveCondition( self.id, 1 )
                                end
                            end,
                        }
                    }
                },
                demoralize =
                {
                    name = "Demoralize",
                    anim = "taunt",
                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.DEBUFF,
                    target_mod = TARGET_MOD.TEAM,

                    OnPostResolve = function( self, battle, attack )
                        local morale = 0.2 + 0.05 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )
                        for i, hit in attack:Hits() do
                            if hit.target:HasMorale() then
                                hit.target:DeltaMorale( morale )
                            end
                        end
                    end,
                }
            },
            behaviour =
            {
                OnActivate = function( self )
                    self.moves = self:MakePicker()
                        :AddID( "fire_in_the_hole", 1 )
                        :AddID( "disrupt", 1 )
                        :AddID( "demoralize", 1, 1 )
                        :AddID( "ai_power_boost", 1, 2 )

                    self:SetPattern( self.Cycle )
                end,

                Cycle = function( self )
                    self.moves:ChooseCard( 1 )
                end,
            }
        },

    })
)

--rise footsoldier
Content.AddCharacterDef
(
    CharacterDef("RISE_VALET",
    {
        base_def = "RISE_BASE",

        title = "Valet",
        renown = 2,
        combat_strength = 2,

        faction_id = "RISE",
        tags = { "rise" },

        base_builds = {
            [ GENDER.MALE ] = "male_rise_valet_build",
            [ GENDER.FEMALE ] = "female_rise_valet_build"
        },

        anims = {"anim/weapon_valet_rise_valet.zip"},
        combat_anims = { "anim/med_combat_rise_valet.zip" },

        loved_graft = "convenience",
        hated_graft = "inconvenience",
        death_item = "stolen_jewelry",

        fight_data =
        {
            MAX_MORALE = MAX_MORALE_LOOKUP.HIGH,
            MAX_HEALTH = MAX_HEALTH_LOOKUP.MEDIUM,

            attacks =
            {
                valet_double = table.extend(NPC_ATTACK)
                {
                    name = "Double Strike",
                    anim = "double_swipe",

                    flags = battle_defs.CARD_FLAGS.MELEE,

                    hit_count = 2,
                    damage_mult = 0.6,
                },

                valet_spin = table.extend(NPC_ATTACK)
                {
                    name = "Flying Spin",
                    anim = "spin_attack",

                    flags = battle_defs.CARD_FLAGS.MELEE,
                },

                valet_taunt = table.extend(NPC_BUFF)
                {
                    name = "Bow",
                    anim = "taunt",

                    flags = battle_defs.CARD_FLAGS.SKILL | battle_defs.CARD_FLAGS.BUFF | battle_defs.CARD_FLAGS.HEAL,
                    target_type = TARGET_TYPE.SELF,

                    CanPlayCard = function( self, battle, target )
                        return self.owner:GetHealthPercent() < 1
                    end,

                    OnPostResolve = function( self, battle, attack )
                        local power_amt = 2 + (2 * GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH ))
                        attack:AddCondition( "POWER", power_amt, self)
                        attack:AddCondition( "POWER_LOSS", power_amt, self)
                        local heal_pct =  math.random(6, 12) * 0.01
                        self.owner:HealHealth( math.floor(self.owner:GetMaxHealth() * heal_pct), self )
                    end
                },
            },
            behaviour =
            {
                OnActivate = function( self )
                    self.double_hit = self:AddCard("valet_double")

                    self.attacks = self:MakePicker()
                        :AddID("valet_spin", 2)
                        :AddID("valet_taunt", 1)

                    self:SetPattern( self.Basic )
                end,

                Basic = function( self )
                    if self.fighter:HasCondition("POWER_LOSS") then
                        self:ChooseCard(self.double_hit)
                    else
                        self.attacks:ChooseCard()
                    end
                end,
            }
        },

    })
)