local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT
local CARD_FLAGS = battle_defs.CARD_FLAGS
local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local def = CharacterDef("SEEMLI",
{
    unique = true,
    base_def = "RISE_RADICAL",
    alias = "SEEMLI",
    species = SPECIES.JARACKLE,
    gender = GENDER.FEMALE,
    renown = 2,
    combat_strength = 3,
    name = "Seemli",
    tags = {"plot_armour", "no_patron"},
    
    loved_graft = "covert_supplier",
    hated_graft = "disorder",
    death_item = "seemlis_dagger",
    faction_id = "RISE",
    title = "Firebrand",
    theme_music = "event:/music/story/smith/story_smith_seemli",
    -- voice_actor = "",
    
    anims = { "anim/weapon_knife_seemli.zip", },
    combat_anims = { "anim/med_combat_knife_seemli.zip" },
    head = "head_female_jarackle_15",
    build = "female_seemli_build",

    negotiation_data = 
    {
        modifiers =
        {
            SLY =
            {
                name = "Sly",
                desc = "Whenever {1}'s attacks cause resolve loss, {1} gains 1 {RISE_HANDOUT}.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOwnerName() )
                end,

                modifier_type = MODIFIER_TYPE.CORE,

                event_handlers =
                {
                    [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                        if source.negotiator == self.negotiator and target.negotiator ~= self.negotiator and damage > defended then
                            self:NotifyTriggered()
                            self.negotiator:AddModifier("RISE_HANDOUT", 1, self)
                        end
                    end,
                }
            },

            RISE_HANDOUT =
            {
                name = "Rise Handout",
                desc = "At the start of {1}'s turn, <#HILITE>{2} {2*card|cards}</> in {1}'s hand {2*costs|cost} 1 additional action.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOpponentName(), self.stacks or 1)
                end,
                modifier_type = MODIFIER_TYPE.ARGUMENT,
                max_resolve = 4,

                event_handlers =
                {
                    [ EVENT.HAND_DRAWN ] = function( self, minigame )
                        self.affected_cards = {}
                        local options = shallowcopy(minigame:GetHandDeck().cards)
                        for i=1,self.stacks do
                            local chosen = table.arraypick(options)
                            if chosen then
                                table.arrayremove(options, chosen)
                                table.insert(self.affected_cards, chosen)
                            end
                        end
                    end,

                    [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                        if self.affected_cards and table.contains(self.affected_cards, card) then
                            cost_acc:AddValue( 1, self )
                        end
                    end,
                }
            },

            DECEPTION =
            {
                name = "Deception",
                desc = "Your next {1} {1*card|cards} {1*has|have} random targets. Remove 1 <b>Deception</> whenever you play a card.",
                desc_fn = function( self, fmt_str )
                    return loc.format(fmt_str, self.stacks or 1)
                end,
                modifier_type = MODIFIER_TYPE.INCEPTION,

                event_handlers =
                {
                    [ EVENT.END_RESOLVE ] = function( self, minigame, card )
                        if card.negotiator == self.negotiator then
                            self.negotiator:RemoveModifier(self.id, 1, self)
                        end
                    end,

                    [ EVENT.CALC_TARGET_MOD ] = function( self, acc, source, target )
                        if source.negotiator == self.negotiator and is_instance( source, Negotiation.Card ) and acc.value ~= TARGET_MOD.TEAM and acc.value ~= TARGET_MOD.ALL then
                            acc:ModifyValue( TARGET_MOD.RANDOM1, self )
                        end
                    end,
                },
            },
        },

        behaviour = 
        {
            OnInit = function( self, difficulty )
                local difficulty = self.engine and self.engine:GetDifficulty() or 1
                self.deception = self:AddCard("deception_inception")
                self.deception.stacks = math.min( 3, difficulty)

                self.negotiator:AddModifier( "SLY" )

                self:SetPattern( self.Cycle )
            end,

            Cycle = function( self, turns )
                if turns % 3 == 0 then
                    self:ChooseGrowingNumbers( 2, 2 )
                else
                    self:ChooseGrowingNumbers( 1, 1 )
                end
                if (turns - 1) % 2 == 0 then
                    self:ChooseComposure( 1, 2, 5 )
                end
                if (turns + 1) % 3 == 0 then
                    self:ChooseCard(self.deception)
                end
            end,
        }
    },

    fight_data = 
    {
        MAX_MORALE = MAX_MORALE_LOOKUP.MEDIUM,
        MAX_HEALTH = 65,
        death_fade_delay = 1,

        conditions =
        {
            shadow_blades =
            {
                name = "Shadow Blades",
                desc = "{1} gains {2} {EVASION} whenever {3.hisher} attacks cause health loss.",
                desc_fn = function( self, fmt_str )
                    return loc.format( fmt_str, self:GetOwnerName(), self.stack, self.owner)
                end,

                event_handlers =
                {
                    [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                        if not hit:CheckHitResult( "defended" ) and not hit:CheckHitResult( "evaded" ) and hit.card:IsDamageCard() and hit.attacker == self.owner then
                            self.owner:AddCondition( "EVASION", self.stacks, self )
                        end
                    end,
                }
            },
        },

        attacks = 
        {
            seemli_stab = table.extend(NPC_ATTACK)
            {
                name = "Stab",
                anim = "stab",
                flags = CARD_FLAGS.MELEE | CARD_FLAGS.DEBUFF,

                OnPostResolve = function( self, battle, attack )
                    attack:AddCondition("WOUND", 2, self)
                end
            },

            seemli_spin = table.extend(NPC_ATTACK)
            {
                name = "Spin",
                anim = "spin_attack",
                flags = CARD_FLAGS.MELEE,
            },

            seemli_taunt = table.extend(NPC_BUFF)
            {
                name = "Taunt",
                anim = "taunt",

                flags = CARD_FLAGS.SKILL | CARD_FLAGS.BUFF,
                target_type = TARGET_TYPE.SELF,

                OnPostResolve = function( self, battle, attack )
                    attack:AddCondition("DEFEND", 6, self)
                    attack:AddCondition("POWER", 2 + (2 *GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )), self)
                    attack:AddCondition("POWER_LOSS", 2 + (2 *GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ABILITY_STRENGTH )), self)
                end
            },        
        },

        behaviour =
        {
            OnActivate = function( self, fighter )
                self.attacks = self:MakePicker()
                    :AddID("seemli_stab", 2)
                    :AddID("seemli_spin", 3)
                    :AddID("seemli_taunt", 1)

                self:SetPattern(self.Cycle)
            end,


            Cycle = function( self )
                self.attacks:ChooseCard()
            end
        },
    },
})

Content.AddCharacterDef( def )
    