require "constants"
require "quest_def"
require "location_def"
require "character_def"
require "localization_classes"
require "quip_database"
require "codex"
require "workposition"
require "deck"
require "namegen"
require "stringtable"
require "playerbackground"
require "content/combat_parties"
require "content/card_shop_defs"


local filepath = require "util/filepath"
local lyaml = require "lyaml"

----------------------------------------------------------
-- Master table of all preloaded moddable game content. --
----------------------------------------------------------

local CURRENT_MOD_ID = nil -- Keeps track of which mod is currently adding content; nil if base game.
local new_quest_act_filter

local CONTENT =
{
    LOCATIONS = {},
    PLAX_DATA = {},
    WORLD_REGIONS = {},
    CHARACTER_NAMES = {},
    CHARACTERHEADS = {},
    CHARACTERBUILDS = {},
    CHARACTERDEFS = {},
    CHARACTER_ASPECTS = {},
    FACTIONS = {},
    BATTLE_CARDS = {},
    GRAFTS = {},
    QUIPS = {},
    CODICES = {},
    MODIFIERS = {},
    NEGOTIATION_CARDS = {},
    NEGOTIATION_FEATURES = {},
	BATTLE_FEATURES = {},
    BATTLE_CONDITIONS = {},
    OPINION_EVENTS = {},
    UNLOCKS = {},
    PARTICLE_PARAMS = {},
    BATTLE_DECKS = {},
    NEGOTIATION_DECKS = {},
    UNLOCK_TRACKS = {},
    GRIFTS = {},
    ACHIEVEMENTS = {},
    OUTFITS = {},
    LOCALIZATIONS = {},
    ENGLISH_STRINGS = {},
    STRING_TABLES = {},
    STRINGS = {},
    RUMBLES = {},
    SCORE_COMPONENTS = {},

    ACT_DATA = {},
    SORTED_PLAYER_DATA = {},
    PLAYER_DATA = {},

    MODS = {},
    CONVOSG = {},
    CONVOSG_BY_TAG = {},
    QUESTDEFS = {},
    CHARACTER_SKINS = {},
    SLIDE_SHOWS = {},
}

-- Processed content, has duplicate references so it is not in the CONTENT table.
local BATTLE_CARDS_ARRAY, NEGOTIATION_CARDS_ARRAY, GRAFTS_ARRAY
local SORTED_PLAX_DATA

---------------------------------------------------------------------------------

local function InvokeModAPIForMod( mod, fn, ... )
    if not mod.finalized then
        return
    end
    if type(fn) ~= "function" then
        return
    end

    CURRENT_MOD_ID = mod.id
    local ok, result = xpcall( fn, generic_error, mod, ... )
    if not ok then
        LOGWARN( "Failed calling mod API '%s.%s': \n %s", mod.id, table.find( mod, fn ), tostring(result))
        result = nil
    end
    CURRENT_MOD_ID = nil

    return result
end

local function InvokeModAPI( fn_name, ... )
    for i, mod in ipairs( CONTENT.MODS ) do
        InvokeModAPIForMod( mod, mod[ fn_name ], ... )
    end
end

local function FindMod( id )
    for i, mod in ipairs( CONTENT.MODS ) do
        if mod.id == id then
            return mod
        end
    end
end

local function GetModSetting( mod_table, key )
    if type(mod_table) == "string" then
        local id = mod_table
        mod_table = FindMod( id )
        if mod_table == nil then
            print( "GetModSetting could not find mod", id )
            return
        end
    end

    local profile = TheGame:GetGameProfile()
    local mod_settings = profile:GetModSettings()[ mod_table.id ]
    local value
    if mod_settings then
        value = mod_settings[ key ]
    end
    if value == nil then
        -- See if this mod has an option with this key with a default_value.
        for i, option in ipairs( mod_table.mod_options or table.empty ) do
            if option.key == key then
                value = option.default_value
                break
            end
        end
    end
    return value
end

local function SetModSetting( mod_table, key, value )
    if type(mod_table) == "string" then
        local id = mod_table
        mod_table = FindMod( id )
        if mod_table == nil then
            print( "GetModSetting could not find mod", id )
            return
        end
    end

    local profile = TheGame:GetGameProfile()
    local mod_settings = profile:GetModSettings()
    if mod_settings[ mod_table.id ] == nil then
        mod_settings[ mod_table.id ] = {}
    end

    mod_settings[ mod_table.id ][ key ] = value

    profile:ForceSaveSettings()
end

local function LoadWorkshopID( mod )
    print( string.format( "LoadWorkshopID( '%s' )", mod.installPath ))
    local filename = string.format( "%s/steam_workshop.txt", mod.installPath )
    local f = io.open( filename, "r" )
    local workshop_id
    if f then
        workshop_id = f:read()
        if type(workshop_id) == "string" then
            workshop_id = math.floor( tonumber(workshop_id))
        end
        f:close()
    end
    return workshop_id
end

local function SaveWorkshopID( mod, workshop_id )
    if mod.workshop_id == workshop_id then
        return
    end

    mod.workshop_id = workshop_id

    local filename = string.format( "%s/steam_workshop.txt", mod.installPath )
    local fl = io.open( filename, "w+" )
    if fl then
        fl:write( tostring(workshop_id) )
        fl:close()
    else
        print( "Failed to write workshop ID:", filename )
    end
end

local function IsModEnabled( mod_table, version )
    if mod_table == nil then
        return false
    end
    if mod_table.status == MOD_STATUS.BUILTIN then
        -- Built-ins cannot be disabled.
        return true
    elseif mod_table.status == MOD_STATUS.LOCAL then
        -- Local mods are enabled by default.
        if GetModSetting( mod_table, MOD_SETTING.ENABLED ) == false then
            return false
        end
        if mod_table.version and version and not IsCompatibleVersion( mod_table.version, version ) then
            return false
        end

        return true
    else
        -- Any other mods are disabled by default.
        if GetModSetting( mod_table, MOD_SETTING.ENABLED ) ~= true then
            return false
        end
        if mod_table.version and version and not IsCompatibleVersion( mod_table.version, version ) then
            return false
        end
        return true
    end
end

local function SetModEnabled( mod_table, enabled )
    if mod_table.status == MOD_STATUS.BUILTIN then
        return -- builtin mods cannot have their enabled status changed.
    end

    if not enabled and IsModEnabled( mod_table ) then
        SetModSetting( mod_table, MOD_SETTING.ENABLED, false )

    elseif enabled and not IsModEnabled( mod_table ) then
        SetModSetting( mod_table, MOD_SETTING.ENABLED, true )
    end

    if TheGame:GetEvents() then
        TheGame:BroadcastEvent( "mod_changed", mod_table )
    end
end

local function HarvestModStrings( mod_table, strings )
    if mod_table.mod_options then
        for i, option in ipairs( mod_table.mod_options ) do
            local prefix = string.format( "%s.%s", mod_table.alias or mod_table.id, option.key:upper() )
            local title_key = string.format( "%s.TITLE", prefix )
            strings[ title_key ] = option.title

            if option.values and option.values.desc then
                local desc_key = string.format( "%s.DESC", prefix )
                strings[ desc_key ] = option.values.desc

            elseif option.values and #option.values > 0 then
                for i, v in ipairs( option.values ) do
                    if v.desc then
                        local name_key = string.format( "%s.NAME%d", prefix, i )
                        strings[ name_key ] = v.name

                        local desc_key = string.format( "%s.DESC%d", prefix, i )
                        strings[ desc_key ] = v.desc
                    end
                end
            end
        end
    end
end

local function LocalizeModOptions( mod_table )
    for k, option in ipairs( mod_table.mod_options or table.empty ) do
        local prefix = string.format( "%s.%s", mod_table.alias or mod_table.id, option.key:upper() )

        local title_key = string.format( "%s.TITLE", prefix )
        option.title = TheGame:LookupStr( title_key ) or option.title
        if option.values and option.values.desc then
            local desc_key = string.format( "%s.DESC", prefix )
            option.values.desc = TheGame:LookupStr( desc_key ) or option.values.desc

        elseif option.values and #option.values > 0 then
            for i, v in ipairs( option.values ) do
                if v.desc then
                    local name_key = string.format( "%s.NAME%d", prefix, i )
                    v.name = TheGame:LookupStr( name_key ) or v.name

                    local desc_key = string.format( "%s.DESC%d", prefix, i )
                    v.desc = TheGame:LookupStr( desc_key ) or v.desc
                end
            end
        end
    end
end

local function GetModID( mod_table )
    return mod_table.id
end

-- Runs installPath/modinit.lua which should return a table of values.
local function InstallMod( mod )

    if not mod.installPath then
        -- BUILTIN or subscribed workshop mods which are not installed will have no installPath.
        return
    end

    local mod_env =
    {
        __index = function( t, k )
            local v = rawget( _G, k )
            return v
        end,

        mod = mod,
        engine = { asset = engine.asset, model = engine.model },
        debug = { traceback = debug.traceback },
        io = table.empty,

        MountModData = function( alias )
            engine.inst:Mount( mod.installPath, alias )
        end,
    }
    setmetatable( mod_env, mod_env )

    print( string.format( "InstallMod( '%s' )", tostring(mod.installPath )))

    local init_filename = string.format( "%s/modinit.lua", mod.installPath )
    local f, err = loadfile( init_filename, nil, mod_env )
    if type(f) ~= "function" then
        error( tostring(err) )
    end

    local result = f()
    assert(type(result) == "table", "modinit must return a table" )

    -- ccc: frankly I'm not sure this is a good idea or not?
    -- if result.id ~= nil then
    --     assert( type(result.id) == "string", "unique mod id must be a string" )
    --     mod.id = result.id
    -- end

    -- REMOVE ME: Content should not be loaded inline in modinit
    CURRENT_MOD_ID = mod.id -- Loaded content will be registered to this mod.

    if type(result.version) == "string" then
        local major, minor, patch = ParseVersion( result.version )
        assert( major and minor and patch, "version field should be a string of the form MAJOR.MINOR.PATCH" )
    end

    if result.alias == nil then
        result.alias = mod.id
    end

    assert( result.HarvestStrings == nil, "cannot override HarvestStrings" )
    result.HarvestStrings = HarvestModStrings
    result.GetModID = GetModID

    if result.mod_options then
        for k, option in ipairs( result.mod_options ) do
            assert( type(option.title) == "string", "string expected for option.title" )
            assert( option.slider or option.spinner or option.button, "option.spinner, option.slider, or option.button expected" )
            assert( type(option.key) == "string", "string expected for option.key" )
            if option.slider or option.spinner then
                assert( option.default_value ~= nil, "non-nil option.default_value expected" )
                assert( type(option.values) == "table", "option.values table expected" )
            elseif option.button then
                assert( type(option.on_click) == "function", "option.on_click function expected" )
            end
        end
    end

    -- success
    for k, v in pairs( result ) do
        if mod[ k ] == nil then
            mod[ k ] = v
        end
    end

    if mod.description_file then
        if mod.alias then
            engine.inst:Mount( mod.installPath, mod.alias )
        end

        local fl = io.open( mod.description_file, "r" )
        if fl then
            mod.description = fl:read( "a" )
            fl:close()
        else
            print( "Could not find", mod.description_file )
        end

        if mod.alias then
            engine.inst:Unmount( mod.alias )
        end
    end
    -- title always overrides the 'workshop title' if possible.
    mod.title = result.title or mod.title
end

local function RegisterMod( mod_table )
    assert( not table.contains( CONTENT.MODS, mod_table ))
    assert( mod_table.id, "No unique 'id' specified for mod:"..tostr(mod_table) )
    assert( not FindMod( mod_table.id ), "Already exists: "..tostring(mod_table.id))

    if mod_table.status == nil then
        mod_table.status = MOD_STATUS.BUILTIN
        mod_table.finalized = true

    elseif mod_table.status ~= MOD_STATUS.WORKSHOP then
        mod_table.workshop_id = LoadWorkshopID( mod_table )
    end

    local ok, err = xpcall( InstallMod, generic_error, mod_table )
    if not ok then
        print( string.format( "\"%s\" failed to execute modinit.lua.", tostring(mod_table.installPath) ))
        print( tostr(err))
        if mod_table.status == MOD_STATUS.LOCAL or mod_table.status == MOD_STATUS.DEV then
            engine.inst:ShowWarning( string.format("\"%s\" failed to execute modinit.lua:\n%s", tostring(mod_table.installPath) , tostr(err) ))
        else
            SetModEnabled( mod_table, false )
        end

        mod_table.error = err
    end

    CURRENT_MOD_ID = nil

    if FindMod( mod_table.id ) then
        LOGWARN( "Mod '%s' loaded at '%s' already exists from '%s'", mod_table.id, mod_table.installPath, FindMod( mod_table.id ).installPath )
    else
        table.insert( CONTENT.MODS, mod_table )
    end
end

local function UnregisterMod( mod_table )
    table.arrayremove( CONTENT.MODS, mod_table )
end

local function AllMods()
    return ipairs( CONTENT.MODS )
end

local function GetInstalledMods()
    local mods = {}
    for i, mod in AllMods() do
        if mod.status ~= MOD_STATUS.BUILTIN and mod.finalized then
            table.insert( mods, mod.id )
        end
    end
    return mods
end

local function GetOutfit(outfit_id)
    return CONTENT.OUTFITS[outfit_id]
end

local function GetOutfitsForCharacter(character_id, t)
    t = t or {}
    for k,v in pairs(CONTENT.OUTFITS) do
        if v.character == character_id then
            table.insert(t, v)
        end
    end

    table.sort(t, function(a,b) return a.sort_key < b.sort_key end)
    return t
end

local function AddRumble(rumble)
    CONTENT.RUMBLES[rumble:GetID()] = rumble
    return rumble
end

local function GetRumble(id)
    return CONTENT.RUMBLES[id]
end


local function AddOutfit(outfit)
    local num = 0
    for k,v in pairs(CONTENT.OUTFITS) do
        num = num + 1 
    end
    outfit.sort_key = num
    outfit:SetModID( CURRENT_MOD_ID )

    CONTENT.OUTFITS[outfit.id] = outfit
end

local function ReloadSlideShows()
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/slides/*.lua')) do
        local root_name = filename:match( "^(.+)[.]lua$" )
        local filename = "content/slides/" .. root_name
        package[ filename ] = nil
        local slideshow = require(filename)
        Content.AddSlideShow( root_name, slideshow )
    end
end

local function AddSlideShow( id, data )
    CONTENT.SLIDE_SHOWS[ id ] = SlideShowDef( id, data )
end

local function GetSlideShow( id )
    return CONTENT.SLIDE_SHOWS[ id ]
end

local function GetSlideShows()
    return CONTENT.SLIDE_SHOWS
end

local function HarvestStrings(strings, filter_fn)
    for k, v in pairs(CONTENT) do
        if v.HarvestStrings then
            if filter_fn == nil or filter_fn == false or filter_fn( v ) then
                v:HarvestStrings(strings)
            end

        elseif type(v) == "table" then
            for sk, sv in pairs(v) do
                if type(sv) == "table" and sv.HarvestStrings then
                    if filter_fn == nil or filter_fn( sv ) then
                        sv:HarvestStrings(strings)
                    end
                end
            end
        end
    end
end

local function AddLocalization(settings)
    CONTENT.LOCALIZATIONS[settings.id] = settings
end

local function AddPOFileToLocalization(id, po_file)
    if CONTENT.LOCALIZATIONS[id] then
        if not CONTENT.LOCALIZATIONS[id].po_filenames then
            CONTENT.LOCALIZATIONS[id].po_filenames = {}
        end

        if not po_file:match("^(%w+)[:]" ) then
            po_file = string.format( "localizations/%s.po", po_file)
        end

        assert( engine.inst:FileExists( po_file ), "Missing specified .po file: "..tostring(po_file) )

        table.insert(CONTENT.LOCALIZATIONS[id].po_filenames, po_file)
    end
end

local function GetLocalizations()
    return CONTENT.LOCALIZATIONS
end

local function LookupString( id )
    return CONTENT.STRINGS[ id ] or CONTENT.ENGLISH_STRINGS[ id ] 
end

local function AddStringTable( id, t )
    local def = StringTable( id, t )
    def:SetModID( CURRENT_MOD_ID )

    CONTENT.STRING_TABLES[ id ] = def
end

local function AddDeck(deck)
    assert(deck.id)
    local t = is_instance( deck, NegotiationDeck ) and CONTENT.NEGOTIATION_DECKS or CONTENT.BATTLE_DECKS

    local num = 0
    for k,v in pairs(t) do
        num = num + 1 
    end
    t[deck.id] = deck
    deck.sort_key = num

    deck:SetModID( CURRENT_MOD_ID )
end

local function GetDeck(deck)
    return CONTENT.BATTLE_DECKS[deck] or CONTENT.NEGOTIATION_DECKS[deck]
end

local function GetBattleDecks(owner_id, t)
    t = t or {}
    for k,v in pairs(CONTENT.BATTLE_DECKS) do
        if v.owner == owner_id then
            table.insert(t, v)
        end
    end
    table.sort(t, function(a,b) return a.sort_key < b.sort_key end)
    return t
end

local function GetNegotiationDecks(owner_id, t)
    t = t or {}
    for k,v in pairs(CONTENT.NEGOTIATION_DECKS) do
        if v.owner == owner_id or owner_id == nil then
            table.insert(t, v)
        end
    end
    table.sort(t, function(a,b) return a.sort_key < b.sort_key end)
    return t
end

local function UpgradeModifier( id, def, basic_id )
    local basic_def = CONTENT.MODIFIERS[ basic_id ]
    assert( basic_def, basic_id )  

    for k, v in pairs(basic_def) do
        if def[k] == nil then
            def[k] = deepcopy(v)
        end
    end

    if def.name == nil then
        def.name = loc.format( "{1}+", basic_def.name )
    elseif def.name == basic_def.name then
        def.name = nil
    end
    
    def.base_id = basic_id
    def.base_def = basic_def
end

local function ValidateModifier( id, def )
    def.id = id

    -- Handle attack upgrade inheritance.
    local basic_id = def.base_id or id:match( "(.*)_plus.*$" ) or id:match( "(.*)_PLUS*$")
    if basic_id then
        UpgradeModifier( id, def, basic_id )
    end

    if def.modifier_type == nil then
        if def.max_resolve then
            def.modifier_type = MODIFIER_TYPE.ARGUMENT
        elseif def.hidden then
            def.modifier_type = MODIFIER_TYPE.HIDDEN
        else
            error( "Missing modifier_type: " ..tostring(id) )
        end
    end
    assert( def.modifier_type ~= nil, id )
    assert( def.modifier_type == MODIFIER_TYPE.HIDDEN or def.modifier_type == MODIFIER_TYPE.PERMANENT or TARGET_FLAG[ def.modifier_type ] ~= nil, id )

    -- Generate conventional icon paths for non-upgraded cards. 
    if def.modifier_type ~= MODIFIER_TYPE.HIDDEN then
        if type(def.icon) ~= "userdata" then
            local path
            if def.icon then
                path = def.icon
            else
                path = string.format( "negotiation/modifiers/%s.tex", id:lower() )
            end
            def.icon_path = path

            def.icon = engine.asset.Texture( path, true )
            if def.icon:IsDefault() then
                def.icon = global_images.default
            end
        end
    end
    if def.modifier_type == MODIFIER_TYPE.CORE and def.hide_stacks == nil then
        def.hide_stacks = true
    end

    def.max_stacks = def.max_stacks or 99
end



local function AddQuips( db )
    assert( is_instance( db, QuipDatabase ))
    db:SetModID( CURRENT_MOD_ID )
    table.insert( CONTENT.QUIPS, db )
end

local function AddCodex( codex )
    assert( is_instance( codex, Codex ))
    codex:SetModID( CURRENT_MOD_ID )
    table.insert( CONTENT.CODICES, codex )
end

local function AddQuestDef( id, qdef )
    
    if not qdef:IsDisabled() then
        assert(not CONTENT.QUESTDEFS[ id ], "QuestDef already loaded: " .. id )
        CONTENT.QUESTDEFS[ id ] = qdef
        
        if qdef.act_filter == nil then
            qdef.act_filter = new_quest_act_filter
        end

        qdef:SetModID( CURRENT_MOD_ID )
    end
end

local function LoadConvoLua( filename )
    package.loaded[ filename ] = nil
    local ok, result = xpcall( require, generic_error, filename )
    if not ok then
        error( result )
    end
    return ok, result
end

local function GetLocationsThatUsePlax(plax_id, t)
    local t = t or {}
    table.clear(t)
    for loc_id,locdef in pairs(CONTENT.LOCATIONS) do
        if locdef.plax == plax_id then
            table.insert(t, loc_id)
        end
    end
    return t
end

local function LoadAllPlax()
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/plax/*.lua')) do
        local id = filename:match("(.+)%..+")
        filename = string.format( "content/plax/%s", id )
        Content.AddPlaxData( id, filename )
    end
end

local function AllPlax()
    return pairs( CONTENT.PLAX_DATA )
end

local function GetAllPlaxSorted()
    if SORTED_PLAX_DATA == nil then
        SORTED_PLAX_DATA = {}
        for id, v in pairs( CONTENT.PLAX_DATA ) do
            table.insert( SORTED_PLAX_DATA, id )
        end
        table.sort( SORTED_PLAX_DATA )
    end

    return SORTED_PLAX_DATA
end

local function AddPlaxData( plax_id, t )
    assert( type(t) == "string", tostring(t))
    CONTENT.PLAX_DATA[ plax_id ] = t
    SORTED_PLAX_DATA = nil
end

local function RemovePlaxData( plax_id )
    CONTENT.PLAX_DATA[ plax_id ] = nil
    if SORTED_PLAX_DATA then
        table.arrayremove( SORTED_PLAX_DATA, plax_id )
    end
end

local function GetPlaxPath( plax_id )
    return CONTENT.PLAX_DATA[ plax_id ]
end

local function GetPlaxData( plax_id, reload )
    local filename = CONTENT.PLAX_DATA[ plax_id ]
    if filename then
        if reload then
            package.loaded[ filename ] = nil
        end
        local t = require( filename )
        return t
    end
end

local function GetAllQuests()
    return CONTENT.QUESTDEFS
end

local function GetQuestDef(id)
    return CONTENT.QUESTDEFS[id]
end

local function ReloadConvoStateGraphs()
    package.loaded[ "content/convo_loc_common" ] = nil
    require "content/convo_loc_common"

    package.loaded[ "encounter/base_hub" ] = nil
    require "encounter/base_hub"

    package.loaded[ "encounter/convostategraph" ] = nil
    require "encounter/convostategraph"

    package.loaded[ "encounter/convo_option" ] = nil
    require "encounter/convo_option"

    package.loaded[ "encounter/stategraphutil" ] = nil
    require "encounter/stategraphutil"

    package.loaded[ "encounter/convoutil" ] = nil
    require "encounter/convoutil"
end

local function ReloadCharacters()
    engine.inst:ProfilerPush("characters")
    table.clear( CONTENT.CHARACTERDEFS )
    table.clear( CONTENT.CHARACTERBUILDS )
    table.clear( CONTENT.CHARACTER_SKINS )

    package.loaded[ "content/character_builds/character_builds" ] = nil
    require( "content/character_builds/character_builds" )

    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/characters/*.lua')) do
        local name = filename:match( "^(.+)[.]lua$" )
        local filename = "content/characters/"..name
        package.loaded[ filename ] = nil
    end

    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/characters/*.lua')) do
        local name = filename:match( "^(.+)[.]lua$" )
        local filename = "content/characters/"..name
        require( filename )
    end

    for id, def in pairs( CONTENT.CHARACTERDEFS ) do
        def:InheritBaseDef()
    end

    engine.inst:ProfilerPop()
end

local function ReloadBattleContent()

    package.loaded[ "battle/battle_defs" ] = nil
    require "battle/battle_defs"

    package.loaded[ "battle/battle_engine" ] = nil
    require "battle/battle_engine"
    
    package.loaded[ "content/fighter_data" ] = nil
    require "content/fighter_data"

    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/attacks/*.lua')) do
        local name = filename:match("^(.+)[.]lua$" )
        local file = "content/attacks/"..name
        package.loaded[ file ] = nil
        require ( file )
    end
end

local function ReloadNegotiationContent()

    package.loaded[ "negotiation/negotiation_defs" ] = nil
    local negotiation_defs = require "negotiation/negotiation_defs"

    package.loaded[ "negotiation/negotiation_engine" ] = nil
    require "negotiation/negotiation_engine"

    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/negotiation/*.lua')) do
        local name = filename:match("^(.+)[.]lua$" )
        local file = "content/negotiation/"..name
        package.loaded[ file ] = nil
        require ( file )
    end
end

local function ReloadGraftContent()
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/grafts/*.lua')) do
        local name = filename:match("^(.+)[.]lua$" )
        local file = "content/grafts/"..name
        package.loaded[ file ] = nil
        require ( file )
    end
end

local function ReloadConversation()
    table.clear( CONTENT.QUESTDEFS )
    table.clear( CONTENT.CONVOSG )
    table.clear( CONTENT.CONVOSG_BY_TAG )

    table.clear( CONTENT.BATTLE_CARDS )
    BATTLE_CARDS_ARRAY = nil

    table.clear( CONTENT.BATTLE_CONDITIONS )
    table.clear( CONTENT.BATTLE_FEATURES )

    table.clear( CONTENT.NEGOTIATION_CARDS )
    NEGOTIATION_CARDS_ARRAY = nil

    table.clear( CONTENT.MODIFIERS )

    table.clear( CONTENT.GRAFTS )
    GRAFTS_ARRAY = nil

    -- Cards, grafts.
    ReloadBattleContent()
    ReloadNegotiationContent()
    ReloadGraftContent()

    -- This content may add card data inline.
    ReloadCharacters()

    package.loaded[ "content/notification_events" ]= nil
    require "content/notification_events"

    local ok, err = pcall( ReloadConvoStateGraphs )
    assert_warning( ok, err )

    local db = QuipDatabase( "QUIPS" )
    for k, filename in pairs(engine.inst:ListFiles( 'quips/*.yaml')) do
        db:AddFilename( "quips/"..filename )
    end 
    AddQuips( db )

    engine.inst:ProfilerPush("quests")


    for k, filepath in ipairs( filepath.list_files( "scripts/content/quests/", "*.lua", true )) do
        local name = filepath:match( "scripts/(.+)[.]lua$" )
        
        if filepath:find( "/SAL/" ) then
            new_quest_act_filter = "SAL"
        elseif filepath:find( "/ROOK/" ) then
            new_quest_act_filter = "ROOK"
        elseif string.upper(filepath):find( "/SMITH/" ) then
            new_quest_act_filter = "SMITH"
        else
            new_quest_act_filter = nil
        end

        
        if filepath:find( "/deprecated/" ) then
        else
            if name then
                package.loaded[ name ] = nil
                require( name )
                assert( rawget( _G, "QDEF" ) == nil or error( string.format( "Stop declaring global QDEFS %s", name )))
            end
        end
    end
    
    new_quest_act_filter = nil

    engine.inst:ProfilerPop()

    local CONVO_DIR = 'scripts/content/conversations'
    for k, filepath in ipairs( filepath.list_files( CONVO_DIR, "*.lua", true )) do
        local name = filepath:match( "scripts/(.+)[.]lua$" )
        if name then
            LoadConvoLua( name )
        end
    end
end

local function GetLocationContent(loc)
    return CONTENT.LOCATIONS[loc]
end

local function GetLocationContents()
    return CONTENT.LOCATIONS
end

local function GetFaction( id )
    return CONTENT.FACTIONS[id]
end


local function AddFaction( faction )
    assert( faction.id )
    assert( CONTENT.FACTIONS[ faction.id ] == nil )

    if faction.titles then
        faction.sorted_titles = {}
        for k,v in pairs(faction.titles) do
            v.id = k
            table.insert(faction.sorted_titles, v)
        end
        table.sort(faction.sorted_titles, function(a,b) return (a.rank or 0) < (b.rank or 0) end )
    end

    local def = FactionDef(faction.id, faction)
    def:SetModID( CURRENT_MOD_ID )

    CONTENT.FACTIONS[faction.id] = def
end

local function AllFactions()
    return pairs( CONTENT.FACTIONS )
end

local function AddWorldRegion( t )
    assert( t.name, "Missing world region name" )
    CONTENT.WORLD_REGIONS[ t.id ] = t

    assert( t.population, t.id )
    assert( t.locations, t.id )
end

local function GetWorldRegion( id )
    return CONTENT.WORLD_REGIONS[ id ]
end

local function AllWorldRegions()
    return pairs( CONTENT.WORLD_REGIONS )
end


local function UpgradeBattleCard( id, carddef, basic_id )
    local battle_defs = require "battle/battle_defs"
    local CARD_FLAGS = battle_defs.CARD_FLAGS

    local basic_def = CONTENT.BATTLE_CARDS[ basic_id ]
    assert( basic_def, basic_id )
    
    carddef.name = carddef.name or loc.format( "{1}+", basic_def.name )

    for k, v in pairs(basic_def) do
        if carddef[k] == nil and k~= "condition" and k ~= "upgrade_ids" then
            carddef[k] = deepcopy(v)
        end
    end

    if not basic_def.max_xp and not (basic_def.upgrade_ids and #basic_def.upgrade_ids > 0) then
        local default_xp = DEFAULT_CARD_XP - math.max( 0, 2 * basic_def.cost - 1 )
        if CheckAnyBits( carddef.flags or 0, CARD_FLAGS.EXPEND | CARD_FLAGS.COMBO_FINISHER ) then
            default_xp = default_xp - 2
        end
        if basic_def.mod_xp then
            default_xp = default_xp + basic_def.mod_xp
        end

        basic_def.max_xp = math.max( 3, default_xp )

    elseif basic_def.max_xp == 0 then
        basic_def.max_xp = nil
    end
    carddef.flag_difs = ToggleBits(carddef.flags, basic_def.flags)
    carddef.flags = SetBits( carddef.flags or 0, CARD_FLAGS.UPGRADED )
    carddef.base_id = basic_id
    carddef.base_def = basic_def
    carddef.max_xp = nil

    basic_def.upgrade_ids = basic_def.upgrade_ids or {}
    table.insert(basic_def.upgrade_ids, id)
end

local function LoadCardIcon( carddef, prefix )
    if type(carddef.icon) ~= "userdata" then
        -- Generate conventional icon paths for non-upgraded cards. 
        local path
        if carddef.icon then
            path = carddef.icon
        else
            path = string.format( "%s/%s.tex", prefix, carddef.id:lower() )
        end
        
        carddef.path_str = path
        carddef.icon = engine.asset.Texture( path, true )
        if carddef.icon:IsDefault() then
            carddef.icon = global_images.default
        end
    end
    if type(carddef.icon) ~= "userdata" then
        error( "Invalid icon: " .. carddef.id .. " : " .. tostring(carddef.path_str) )
    end
end

local function HatchBattleParasite( self, battle )
    self:TransferCard( battle.trash_deck )
    self.userdata.removed_safely = true
    self:Consume()
    
    local parasites = {}
    local battle_defs = require "battle/battle_defs"
    for i, def in ipairs( Content.GetAllBattleCards() ) do
        if CheckBits( def.flags, battle_defs.CARD_FLAGS.PARASITE ) then
            parasites[ def.id ] = def
        end
    end

    local fun = require "util/fun"
    local new_card_id = fun(parasites)
                            :filter(function(v) return v.id ~= self.id and CARD_RARITY_ORDINALS[v.rarity] == CARD_RARITY_ORDINALS[self.rarity]+1 end)
                            :keys()
                            :shuffle()
                            :first()
                            
    if new_card_id then
        local card = self.owner.agent.battler:LearnCard(new_card_id)
        local battle_card = card:Clone()
        battle_card.owner = self.owner
        battle:DealCard( battle_card, battle:GetHandDeck( ) )
    end
end

local function HatchNegotiationParasite( self, minigame )
    self:TransferCard( minigame.trash_deck )
    self.userdata.removed_safely = true
    self:Consume()
    
    local parasites = {}
    local negotiation_defs = require "negotiation/negotiation_defs"
    for i, def in ipairs( Content.GetAllNegotiationCards() ) do
        if CheckBits( def.flags, negotiation_defs.CARD_FLAGS.PARASITE ) then
            parasites[ def.id ] = def
        end
    end

    local fun = require "util/fun"
    local new_card_id = fun(parasites)
                            :filter(function(v) return v.id ~= self.id and CARD_RARITY_ORDINALS[v.rarity] == CARD_RARITY_ORDINALS[self.rarity]+1 end)
                            :keys()
                            :shuffle()
                            :first()

    if new_card_id then
        local card = self.owner.negotiator:LearnCard(new_card_id)
        local cloned_card = card:Clone()
        cloned_card.owner = self.owner
        minigame:DealCard( cloned_card, minigame:GetHandDeck( ) )
    end
end

local function ValidateBattleCard( id, carddef )
    --print(loc.format("Add Attack: {1}", id))
    local battle_defs = require "battle/battle_defs"
    local CARD_FLAGS = battle_defs.CARD_FLAGS

    assert( carddef.id == nil, carddef.id )
    carddef.id = id

    -- Handle attack upgrade inheritance.
    local basic_id = carddef.base_id or id:match( "(.*)_plus.*$" ) or id:match( "(.*)_upgraded[%w]*$")
    if basic_id then
        UpgradeBattleCard( id, carddef, basic_id )
    end

    carddef.flags = carddef.flags or CARD_FLAGS.NONE

    if not CheckAnyBits( carddef.flags, CARD_FLAGS.NPC | CARD_FLAGS.SPECIAL ) or carddef.icon ~= nil then
        LoadCardIcon( carddef, "battle" )
    end

    if carddef.preview_icon then
        carddef.preview_icon = engine.asset.Texture( carddef.preview_icon, true )
        if carddef.preview_icon:IsDefault() then
            carddef.preview_icon = global_images.default
        end
    end

    --assert_warning( carddef.name ~= nil, "Missing attack name: %s", id )

    if carddef.desc then
        local str = SetFormattedText( nil, carddef.desc )
        if not str:find( "[!.}]$") and str:len() > 0 then
            error( "Missing closing period: ".. id .. ": " ..str )
        end
    end

    assert( carddef.tags == nil )
    assert( carddef.series ~= nil, id )

    -- Validate features.
    if carddef.features then
        for name, v in pairs( carddef.features ) do
            assert( CONTENT.BATTLE_FEATURES[ name ] or (carddef.condition and carddef.id == name), name )
        end
    end

    -- Default values for targetting specifiers.    
    assert( carddef.target == nil, id ) -- use target_type, this will CRASH (Battle.Card uses target)
    carddef.target_type = carddef.target_type or TARGET_TYPE.ENEMY
    carddef.target_mod = carddef.target_mod or TARGET_MOD.SINGLE
    assert( carddef.target_mod ~= TARGET_MOD.CUSTOM or carddef.target_fn ~= nil, "Missing custom target_fn" )

    if carddef.min_damage and carddef.max_damage then
        assert( carddef.min_damage <= carddef.max_damage )
    end

    assert( carddef.damage_type == nil or type(carddef.damage_type) == "number")

    if CheckBits( carddef.flags, CARD_FLAGS.UNPLAYABLE ) then
        carddef.cost = 0
    else
        assert( carddef.cost, "no cost: "..id )
    end

    -- Verify presentation and rarity.
    -- STATUS or SPECIAL cards do not require presentation or rarity.
    if not CheckAnyBits( carddef.flags, CARD_FLAGS.STATUS | CARD_FLAGS.SPECIAL ) then
        assert( carddef.presentation == nil, id )

        if not CheckBits( carddef.flags, CARD_FLAGS.UPGRADED ) then
            assert( IsEnum( carddef.rarity, CARD_RARITY ), loc.format("{1} unknown rarity: {2}", id, tostring(carddef.rarity) ) )
        end
    end

    if CheckBits( carddef.flags, CARD_FLAGS.RANGED ) then
        if carddef.hit_tags == nil then
            carddef.hit_tags = {}
        end
        table.insert_unique( carddef.hit_tags, "ranged" )

    elseif carddef.hit_tags == nil then
        carddef.hit_tags = table.empty
    end

    if CheckBits( carddef.flags, CARD_FLAGS.PARASITE | CARD_FLAGS.HATCH ) and carddef.hatch_fn == nil then
        carddef.hatch_fn = HatchBattleParasite
    end

    if carddef.condition then
        if not carddef.condition.hidden then
            carddef.condition.name = carddef.condition.name or carddef.name
            if carddef.condition.desc == nil then
                carddef.condition.desc = carddef.desc
                carddef.condition.desc_fn = carddef.condition.desc_fn or carddef.desc_fn
            end
        end
        carddef.condition.icon = carddef.condition.icon or carddef.icon

        Content.AddBattleCondition( id, carddef.condition )
    end

    if carddef.scoring then
        carddef.scoring.id = loc.toupper("BATTLE_CARD_SCORE_" .. id)
        Content.AddScoreComponent(ScoreComponentDef(carddef.scoring))
    end


    if carddef.remove_on_rest then
        if carddef.global_event_handlers == nil then
            carddef.global_event_handlers = {}
        end

        assert( carddef.global_event_handlers[ "do_sleep" ] == nil )
        carddef.global_event_handlers[ "do_sleep" ] = function( self, player, sleep_data )
            table.insert( sleep_data.cards_removed, self )
        end
    end
end

local function AddNegotiationCardFeature(id, def)
    assert( is_instance( def, NegotiationFeatureDef ))
    def:SetModID( CURRENT_MOD_ID )
    CONTENT.NEGOTIATION_FEATURES[id] = def
end

local function GetNegotiationCardFeatures()
   return CONTENT.NEGOTIATION_FEATURES
end

local function GetNegotiationCardFeature(id)
	return CONTENT.NEGOTIATION_FEATURES[id]
end

local function AddBattleCondition(id, t)
    assert( CONTENT.BATTLE_CONDITIONS[ id ] == nil, id )

    local def = CombatConditionDef(id, t)

    if not def.hidden then
        local path
        if type(def.icon) == "string" then
            path = def.icon
        else
            path = string.format( "battle/conditions/%s.tex", id:lower() )
        end
        def.icon_path = path
        def.icon = engine.asset.Texture( path, true ) 
        if def.icon:IsDefault() then
            def.icon = global_images.default
        end
    end

    def.max_stacks = def.max_stacks or 99
    def.stacks = 1
    def.ctype = def.ctype or CTYPE.BUFF
    assert( not def.hidden or def.ctype == CTYPE.BUFF )

    def:SetModID( CURRENT_MOD_ID )

    CONTENT.BATTLE_CONDITIONS[ id ] = def

    local feature_def = {
        name = def.name,
        desc = def.desc,
        loc_strings = def.loc_strings,
        feature_desc = def.feature_desc,
        feature_desc_fn = def.feature_desc_fn,
        feature_tooltip = def.feature_tooltip,

        desc_fn = function( self, fmt_str, stacks, engine, features )
            local result = fmt_str
            if def.feature_tooltip then
                result = self:GetFeatureTooltip()
            elseif def.desc_fn then
                local condition = CombatCondition( id, nil, def.default_stacks or stacks or 1 )
                result = condition:GetDesc( engine, nil, features )
            end
            return  result
        end,
        apply = function( self, battle, attack, stacks, target )
            local applied = false
            if target:GetTeam() == attack.attacker:GetTeam() then
                -- Friendly conditions apply to friendly targets.
                if def.target_type == TARGET_TYPE.SELF or def.target_type == TARGET_TYPE.FRIENDLY_OR_SELF then
                    target:AddCondition( id, stacks, attack.card )
                    applied = true
                end
            else
                if def.target_type ~= TARGET_TYPE.SELF and def.target_type ~= TARGET_TYPE.FRIENDLY_OR_SELF and
                    not attack:CheckHitResult( target, "evaded" ) then
                    target:AddCondition( id, stacks, attack.card )
                    applied = true
                end
            end

            if not applied and
                (def.target_type == TARGET_TYPE.SELF or def.target_type == TARGET_TYPE.FRIENDLY_OR_SELF) then
                local owner = attack.card:GetOwner()
                owner:AddCondition( id, stacks, attack.card )
            end
        end,
    }

    local fdef = BattleFeatureDef(id, feature_def)
    Content.AddBattleCardFeature(id, fdef)
end

local function GetBattleCondition(id)
    return CONTENT.BATTLE_CONDITIONS[id]
end

local function GetAllBattleConditions()
    return CONTENT.BATTLE_CONDITIONS
end

local function AddBattleCardFeature(id, feature)
    assert( is_instance( feature, BattleFeatureDef ))
    feature:SetModID( CURRENT_MOD_ID )
    CONTENT.BATTLE_FEATURES[id] = feature
end

local function GetBattleCardFeature(id)
	return CONTENT.BATTLE_FEATURES[id]
end

local function GetAllBattleFeatures()
    return CONTENT.BATTLE_FEATURES
end

local function AddNegotiationModifier(id, modifier)
    ValidateModifier(id, modifier)
    local def = ModifierDef(id, modifier)
    def:SetModID( CURRENT_MOD_ID )
    CONTENT.MODIFIERS[id] = def

    local feature = NegotiationFeatureDef( id,    
    {
        name = def.name,
		desc = def.desc,
        feature_desc = def.feature_desc,
        feature_tooltip = def.feature_tooltip,

        desc_fn = function( self, fmt_str, stacks, engine, features )
            if self.feature_tooltip then
                return self:GetFeatureTooltip()
            else
                -- This makes me sad.
                local mod = Negotiation.Modifier( id, engine, stacks or 1 )
                return mod:GetFeatureDesc( engine, nil, features )
            end
        end,

        apply = function( self, engine, card, targets, stacks )
            local negotiation_defs = require "negotiation/negotiation_defs"
            local modifier
            if def.modifier_type == MODIFIER_TYPE.INCEPTION then
                if card:IsFlagged( negotiation_defs.CARD_FLAGS.INCEPTED ) then
                    modifier = card.negotiator:AddModifier( id, stacks, card )
                else
                    modifier = card.anti_negotiator:AddModifier( id, stacks, card )
                end
            else
                if def.max_stacks == 1 and card.negotiator:HasModifier( id ) then
                    modifier = card.negotiator:CreateModifier( id, stacks, card )
                else
                    modifier = card.negotiator:AddModifier( id, stacks, card )
                end
            end
            return modifier
        end
    })
    AddNegotiationCardFeature(id, feature)
end

local function GetNegotiationModifier(id)
    return CONTENT.MODIFIERS[id]
end

local function GetAllNegotiationModifiers()
    return CONTENT.MODIFIERS
end

local function UpgradeNegotiationCard( id, carddef, basic_id )
    local negotiation_defs = require "negotiation/negotiation_defs"
    local CARD_FLAGS = negotiation_defs.CARD_FLAGS

    local basic_def = CONTENT.NEGOTIATION_CARDS[ basic_id ]
    assert( basic_def, basic_id )
    
    carddef.name = carddef.name or loc.format( "{1}+", basic_def.name )

    for k, v in pairs(basic_def) do
        if carddef[k] == nil then
            if k ~= "modifier" and k ~= "upgrade_ids" then
                carddef[k] = deepcopy(v)
            end
        end
    end

    if not basic_def.max_xp and not (basic_def.upgrade_ids and #basic_def.upgrade_ids > 0) then
        local default_xp = DEFAULT_CARD_XP - math.max( 0, 2 * basic_def.cost - 1 )
        if CheckBits( carddef.flags or 0, CARD_FLAGS.EXPEND ) then
            default_xp = default_xp - 2
        end
        if basic_def.mod_xp then
            default_xp = default_xp + basic_def.mod_xp
        end

        basic_def.max_xp = math.max( 3, default_xp )

    elseif basic_def.max_xp == 0 then
        basic_def.max_xp = nil
    end
    carddef.flag_difs = ToggleBits(carddef.flags, basic_def.flags)
    carddef.max_xp = nil
    carddef.flags = SetBits( carddef.flags, CARD_FLAGS.UPGRADED )
    carddef.base_id = basic_id
    carddef.base_def = basic_def
    
    basic_def.upgrade_ids = basic_def.upgrade_ids or {}
    table.insert(basic_def.upgrade_ids, id)
end

local function ValidateNegotiationCard( id, carddef )
    local negotiation_defs = require "negotiation/negotiation_defs"
    local CARD_FLAGS = negotiation_defs.CARD_FLAGS

    carddef.id = id
    carddef.is_negotiation = true

    -- Handle upgrade magic.
    local basic_id = carddef.base_id or id:match( "(.*)_plus.*$" ) or id:match( "(.*)_upgraded[%w]*$")
    if basic_id then
        -- This is an upgraded card.  Do magic.659
        UpgradeNegotiationCard( id, carddef, basic_id )
    end

    carddef.flags = carddef.flags or CARD_FLAGS.NONE

    if not CheckAnyBits( carddef.flags, CARD_FLAGS.OPPONENT | CARD_FLAGS.SPECIAL | CARD_FLAGS.BYSTANDER ) or carddef.icon ~= nil then
        LoadCardIcon( carddef, "negotiation" )
    end

    assert( carddef.tags == nil )
    assert( carddef.series ~= nil, id )

    if carddef.desc then
        if carddef.desc == "" then
            -- This would be used to specifically remove the descripton from an upgraded card.
            carddef.desc = nil
        else
            local str = SetFormattedText( nil, carddef.desc )
            if not str:find( "[!.}]$") then
                error( "Missing closing period: ".. id .. ": " ..str )
            end
        end
    end

    if carddef.modifier then
        carddef.modifier.name = carddef.modifier.name or carddef.name
        if carddef.modifier.desc == nil then
            carddef.modifier.desc = carddef.desc
        end
        if carddef.modifier.desc_fn == nil then
            carddef.modifier.desc_fn = carddef.desc_fn
        end
        if carddef.modifier.alt_desc == nil then
            carddef.modifier.alt_desc = carddef.alt_desc
        end

        AddNegotiationModifier(carddef.modifier.id or id, carddef.modifier)
    end


    if carddef.min_persuasion and carddef.max_persuasion then
        assert( carddef.max_persuasion >= carddef.min_persuasion, id )
    else
        assert( carddef.min_persuasion == nil and carddef.max_persuasion == nil, id )
    end

    if not CheckBits( carddef.flags, CARD_FLAGS.OPPONENT ) then
        assert( CheckAnyBits( CARD_TYPE_FLAGS, carddef.flags ), "Missing type flag: "..id )
    end

    if not CheckAnyBits( carddef.flags, CARD_FLAGS.OPPONENT | CARD_FLAGS.STATUS ) then
        assert( IsEnum( carddef.rarity, CARD_RARITY ), "Missing rarity: " ..id )
    end

    -- Check target type
    if carddef.min_persuasion then
        if carddef.target_enemy == nil then
            carddef.target_enemy = TARGET_ANY_RESOLVE
        end
    end

    carddef.target_mod = carddef.target_mod or TARGET_MOD.SINGLE
    assert( carddef.target_mod ~= TARGET_MOD.CUSTOM or carddef.target_fn ~= nil, "Missing custom target_fn" )
    assert( carddef.target_type == nil, id )

    if carddef.quips then
        local lines = carddef.quips
        carddef.quips = QuipDatabase( carddef:GetLocPrefix() )
        carddef.quips:ParseTable( lines, "negotiation", {loc.format("{1}_card", id)} )
    end

    if carddef.quip == nil then
        if CheckBits( carddef.flags, CARD_FLAGS.DIPLOMACY ) then
            carddef.quip = "diplomacy"
        elseif CheckBits( carddef.flags, CARD_FLAGS.HOSTILE ) then
            carddef.quip = "hostile"
        elseif CheckBits( carddef.flags, CARD_FLAGS.MANIPULATE ) then
            carddef.quip = "manipulate"
        end
    end

    if CheckAnyBits( carddef.flags, CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.BYSTANDER ) then
        carddef.cost = 0
    else
        assert( carddef.cost, "no cost: "..id )
    end

    if CheckBits( carddef.flags, CARD_FLAGS.PARASITE | CARD_FLAGS.HATCH ) and carddef.hatch_fn == nil then
        carddef.hatch_fn = HatchNegotiationParasite
    end

    if carddef.remove_on_rest then
        if carddef.global_event_handlers == nil then
            carddef.global_event_handlers = {}
        end

        assert( carddef.global_event_handlers[ "do_sleep" ] == nil )
        carddef.global_event_handlers[ "do_sleep" ] = function( self, player, sleep_data )
            table.insert( sleep_data.cards_removed, self )
        end
    end
end

local function AddNegotiationCard( id, carddef )
    assert_warning( CONTENT.NEGOTIATION_CARDS[ id ] == nil, id )

	local def = NegotiationCardDef(id, carddef)
    ValidateNegotiationCard( id, def )
    def:SetModID( CURRENT_MOD_ID )

    -- you konw the drill.
    -- def:AssignMod( current_mod )

    CONTENT.NEGOTIATION_CARDS[ id ] = def
    NEGOTIATION_CARDS_ARRAY = nil
end

local function GetAspectModID( self ) return self.mod_id end

local function AddAspect( id, aspect_class )
    assert( is_class( aspect_class, Aspect ))
    aspect_class.id = id
    aspect_class.mod_id = CURRENT_MOD_ID

    aspect_class.GetModID = GetAspectModID

    if aspect_class.texture ~= nil then
        if type (aspect_class.texture ) == "string" then
            aspect_class.texture = engine.asset.Texture( aspect_class.texture, true )
        end
    end

    assert( CONTENT.CHARACTER_ASPECTS[ id ] == nil, id )
    CONTENT.CHARACTER_ASPECTS[ id ] = aspect_class
end

local function GetAspect( name )
    return CONTENT.CHARACTER_ASPECTS[ name ]
end

local function SortedAspects()
    return sorted_pairs( CONTENT.CHARACTER_ASPECTS )
end


local function AddOpinionEvent( t )
    assert( t.id )
    t = setmetatable( t, OpinionEventDef )
    CONTENT.OPINION_EVENTS[ t.id ] = t 
    return t
end

local function GetCharacterBuild(id)
    return CONTENT.CHARACTERBUILDS[id]
end

local function GetCharacterHead(id)
    return CONTENT.CHARACTERHEADS[id]
end

local function GetCharacterHeads()
    return CONTENT.CHARACTERHEADS
end

local function AddCharacterHead(id, head)
    
    assert(CONTENT.CHARACTERHEADS[id] == nil, "Trying to add duplicate character head")    
    CONTENT.CHARACTERHEADS[id] = head

    if head.species == SPECIES.HUMAN then
        if not head.skin_tone then
            assert( nil, loc.format("MISSING SKIN TONE DEF FOR HUMAN BUILD {1}", id ))
        end
    else
        if head.skin_tone then
            assert( nil, loc.format("EXTRANEOUS SKIN TONE DEF FOR NONHUMAN BUILD {1}", id ))
        end
    end

    if not engine.inst:FileExists( head.anim ) then
        error( "Missing head anim: " ..tostring(id))
    end
end

local function AddCharacterHeads(heads)
    for k,v in pairs(heads) do
        AddCharacterHead(k,v)
    end
end

local function AddCharacterBuild(id, comm)
    assert(CONTENT.CHARACTERBUILDS[id] == nil, "Trying to add duplicate character builds")
    comm.id = id
    CONTENT.CHARACTERBUILDS[id] = comm
end

local function AddCharacterBuilds(t)
    for k,v in pairs(t) do AddCharacterBuild(k, v) end
end

local function GetAllCharacterBuilds()
    return CONTENT.CHARACTERBUILDS
end

local function AddNamePool( name_pool, filename )
    CONTENT.CHARACTER_NAMES[ name_pool ] = NameList( name_pool, filename )
end

local function ExtendNamePool( name_pool, filename )
    CONTENT.CHARACTER_NAMES[ name_pool ]:ReadFile( filename )
end

local function GenerateCharacterName( name_pool )
    local pool = CONTENT.CHARACTER_NAMES[ name_pool or "NAMES" ]
    if pool then
        return pool:GenerateName()
    end
    return ""
end

local function GetCharacterName( name_pool, nameid )
    local pool = CONTENT.CHARACTER_NAMES[ name_pool or "NAMES" ]
    if pool then
        return pool:GetName( nameid )
    end
    return ""
end

local function AddCharacterDef(t)
    assert( is_instance( t, CharacterDef ))
    assert( type(t.id) == "string" )
    assert( CONTENT.CHARACTERDEFS[t.id] == nil, t.id )
    
    if t.fight_data and t.fight_data.ignore_anim_symbols then
        for k,v in pairs(t.fight_data.ignore_anim_symbols) do
            t.fight_data.ignore_anim_symbols[k] = engine.inst:HashString(v)
        end
    end
    t:SetModID( CURRENT_MOD_ID )

    CONTENT.CHARACTERDEFS[t.id] = t
end

local function GetCharacterDef(id)
    return CONTENT.CHARACTERDEFS[id]
end

local function GetAllCharacterDefs()
    return CONTENT.CHARACTERDEFS
end

local function AddCharacterSkin( content_id, skin )
    if CONTENT.CHARACTER_SKINS[ content_id ] == nil then
        CONTENT.CHARACTER_SKINS[ content_id ] = CharacterSkinsList( content_id )
    end

    if not is_instance( skin, CharacterSkin ) then
        skin = CharacterSkin( content_id, skin )
    end

    skin:SetModID( CURRENT_MOD_ID )

    table.insert( CONTENT.CHARACTER_SKINS[ content_id ], skin )
end

local function AddCharacterSkins( t )
    for content_id, skins in pairs( t ) do
        for i, skin in ipairs( skins ) do
            AddCharacterSkin( content_id, skin )
        end
    end
end

local function GetCharacterSkin( uuid )
    for content_id, skins in pairs( CONTENT.CHARACTER_SKINS ) do
        for i, skin in ipairs( skins ) do
            if skin.uuid == uuid then
                return skin
            end
        end
    end
end

local function GetCharacterSkinByAlias( alias )
    if alias then
        for content_id, skins in pairs( CONTENT.CHARACTER_SKINS ) do
            for i, skin in ipairs( skins ) do
                if skin.alias == alias then
                    return content_id, skin
                end
            end
        end
    end
end

local function GetCharacterForSkin( skin )
    for content_id, skins in pairs( CONTENT.CHARACTER_SKINS ) do
        if table.contains( skins, skin ) then
            return content_id, skins
        end
    end
end

local function GetAllCharacterSkins( content_id )
    if content_id then
        return CONTENT.CHARACTER_SKINS[ content_id ]
    else
        return CONTENT.CHARACTER_SKINS
    end
end

local function CreateCharacterSkin( content_id )
    -- FIXME: this is a pretty roundabout way of generating a skin.
    -- Generate a rando, and extract skin properties.
    local agent = Agent( content_id )
    local skin = agent:GetSkinOverrides()
    if CONTENT.CHARACTER_SKINS[ content_id ] == nil then
        CONTENT.CHARACTER_SKINS[ content_id ] = {}
    end

    skin.uuid = engine.inst:NewGUID()

    skin = CharacterSkin( content_id, skin )
    skin.name = UNNAMED_SKIN
    table.insert( CONTENT.CHARACTER_SKINS[ content_id ], skin )
    return #CONTENT.CHARACTER_SKINS[ content_id ], skin
end

local function SaveAllCharacterSkins()
    -- TODO: this will go in the correct MOD directory.
    local filename = "scripts/content/characters/character_skins.lua"
    local st = os.clock()
    local SERPENT_OPT = { comment = false, nocode = true, sortkeys = true }
    local outs = 
    {
        "-- Autogenerated by content.lua: DO NOT EDIT\n",
        "Content.AddCharacterSkins",
        serpent.block( CONTENT.CHARACTER_SKINS, SERPENT_OPT )
    }

    engine.data:SaveFile( filename, table.concat( outs ), function(ok) 
        print( string.format( "Write file %s (%.1f ms) : '%s'", ok and "OK" or "FAILED", (os.clock() - st) * 1000.0, filename))
    end )   
end

local function GetAllParticleParams()
    return CONTENT.PARTICLE_PARAMS
end

local function GetParticleParams(id)
    return CONTENT.PARTICLE_PARAMS[id]
end

local function ReloadParticles()
    table.clear(CONTENT.PARTICLE_PARAMS)
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/particles/*.lua')) do
        local name = filename:match( "^(.+)[.]lua$" )
        local filename = "content/particles/"..name
        package.loaded[ filename ] = nil
        CONTENT.PARTICLE_PARAMS[name] = require( filename )
    end
end


local function AddLocationContent( location )
    -- assert( CONTENT.LOCATIONS[ location.id ] == nil, "Duplicate location_id detected:"..location.id )
    
    if location.faction_id ~= nil and Content.GetFaction( location.faction_id ) == nil then
        assert (nil, "Invalid Faction: " .. tostring(location.faction_id))
    end

    if location.codex_entry then
        if Content.LookupCodexEntry( location.codex_entry ) == nil then
            LOGWARN( "Location '%s' references missing codex entry '%s'", location.id, tostring(location.codex_entry ))
        end
    end
    
    if location.sublocations then
        location.sublocations = deepcopy( location.sublocations )
        for home_id, v in pairs( location.sublocations ) do
            if type(v) == "table" then
                v.id = string.format( "%s.%s", location.id, home_id )
                AddLocationContent( v )
                location.sublocations[ home_id ] = v.id
            end
        end
    end

    if location.work then
        for work_id, params in pairs(location.work) do
            assert( params.id == nil or params.id == work_id )
            params.id = work_id
        end
    end

    assert( location.name, location.id )

	local def = LocationDef(location.id, location)
    CONTENT.LOCATIONS[def.id] = def

    return def
end

local function LookupQuip( tags )
    local entries
    for i, db in ipairs( CONTENT.QUIPS ) do
        entries = db:LookupEntries( tags, entries )
    end

    if #entries > 0 then
        return CONTENT.QUIPS[1]:LookupQuipInEntries( entries )
    end
end

local function LookupCodexEntry( name )
    for i, codex in ipairs( CONTENT.CODICES ) do
        local entry = codex:LookupEntry( name )
        if entry then
            return entry
        end
    end
end

local function SubstituteCodexFormatting( txt )
    for i, codex in ipairs( CONTENT.CODICES ) do
        txt = codex:SubstituteFormatting( txt )
    end
    return txt
end

local function GetBattleCard( id )
    return CONTENT.BATTLE_CARDS[id]
end

local function GetGraft( id )
    return CONTENT.GRAFTS[id]
end

local function GetAllGrafts()
    if GRAFTS_ARRAY == nil then
        GRAFTS_ARRAY = copyvalues( CONTENT.GRAFTS )
        table.sort( GRAFTS_ARRAY, function( a, b ) return a.id < b.id end )
    end

    return GRAFTS_ARRAY
end

local function UpgradeGraft( id, graft_def, basic_id )
    local battle_defs = require "battle/battle_defs"
    local CARD_FLAGS = battle_defs.CARD_FLAGS
    local basic_def = CONTENT.GRAFTS[ basic_id ]
    assert( basic_def, basic_id )
    
    graft_def.name = graft_def.name or loc.format( "{1}+", basic_def.name )

    for k, v in pairs(basic_def) do
        if graft_def[k] == nil and k~= "battle_condition" and k~= "negotiation_modifier" and k ~= "upgrade_ids" and k ~= "max_xp" then
            graft_def[k] = deepcopy(v)
        end
    end

    if not basic_def.max_xp then
        basic_def.max_xp = GRAFT_MAX_XP[basic_def.rarity] or 10
        if basic_def.mod_xp then
            basic_def.max_xp = basic_def.max_xp + basic_def.mod_xp
        end
    end
    
    graft_def.flags = CARD_FLAGS.UPGRADED
    graft_def.rarity = basic_def.rarity
    graft_def.base_id = basic_id
    graft_def.base_def = basic_def

    basic_def.upgrade_ids = basic_def.upgrade_ids or {}
    table.insert(basic_def.upgrade_ids, id)
end

local function AddGraft( id, graft_def )
    assert(id, "NO ID FOR GRAFT")

    graft_def.id = id

    -- Generate icon textures
    if type(graft_def.img) ~= "userdata" then
        local path
        if type(graft_def.img) == "string" then
            path = graft_def.img
        elseif graft_def.type == GRAFT_TYPE.MUTATOR then
            path = string.format( "icons/mutators/%s.tex", id:lower() )
        elseif graft_def.type == GRAFT_TYPE.PERK then
            path = string.format( "icons/perks/perk_%s.tex", id:lower() )
        else
            path = string.format( "icons/items/graft_%s.tex", graft_def.icon_override or id:lower() )
        end
        graft_def.img = engine.asset.Texture( path, true )
        graft_def.img_path = path
        if graft_def.img:IsDefault() then
            graft_def.img = global_images.default
        end
    end
    
    local basic_id = graft_def.base_id or id:match( "(.*)_plus.*$" )
    if basic_id then
        UpgradeGraft( id, graft_def, basic_id )
    end

    assert(graft_def.rarity, id)
    graft_def.series = graft_def.series or CARD_SERIES.GENERAL

    if graft_def.battle_condition then
        local battle_defs = require "battle/battle_defs"
        
        local id = graft_def.battle_condition.id or graft_def.id
        
        if not graft_def.battle_condition.desc then
            if graft_def.desc then
                if type(graft_def.desc) == "function" then
                    graft_def.battle_condition.desc = function(condition, battle) return condition.graft:GetDesc() end
                else 
                    graft_def.battle_condition.desc = graft_def.desc
                end
            end
        end
        
        graft_def.battle_condition.apply_sound = graft_def.battle_condition.apply_sound or "event:/sfx/battle/status/system/Status_Graft_PowerUp"
        graft_def.battle_condition.desc = graft_def.battle_condition.desc or "MISSING GRAFT CONDITION DESCRIPTION"
        graft_def.battle_condition.name = graft_def.battle_condition.name or graft_def.name
        graft_def.battle_condition.icon = graft_def.battle_condition.icon or graft_def.img
        
        Content.AddBattleCondition( id, graft_def.battle_condition )
        
        graft_def.battle_condition = id
    end

    if graft_def.negotiation_card then
        AddNegotiationCard( id, graft_def.negotiation_card )
    end

    if graft_def.negotiation_modifier then
        
        local negotiation_defs = require "negotiation/negotiation_defs"
        
        local id = graft_def.negotiation_modifier.id or graft_def.id
        
        if not graft_def.negotiation_modifier.desc then
            if graft_def.desc then
                if type(graft_def.desc) == "function" then
                    graft_def.negotiation_modifier.desc = function(condition, minigame) return condition.graft:GetDesc() end
                else 
                    graft_def.negotiation_modifier.desc = graft_def.desc
                end
            end
        end
        
        graft_def.negotiation_modifier.desc = graft_def.negotiation_modifier.desc or "MISSING GRAFT CONDITION DESCRIPTION"
        graft_def.negotiation_modifier.name = graft_def.negotiation_modifier.name or graft_def.name
        graft_def.negotiation_modifier.icon = graft_def.negotiation_modifier.icon or graft_def.img

        AddNegotiationModifier(id, graft_def.negotiation_modifier)
        graft_def.negotiation_modifier = id
    end    

    local def = GraftDef(id, graft_def)
    def:SetModID( CURRENT_MOD_ID )

    CONTENT.GRAFTS[graft_def.id] = def
    GRAFTS_ARRAY = nil
end

local function AddSocialGraft( id, graft_def )
    local battle_defs = require "battle/battle_defs"
    local negotiation_defs = require "negotiation/negotiation_defs"
    
    graft_def.type = GRAFT_TYPE.SOCIAL
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE

    if graft_def.negotiation_modifier and graft_def.negotiation_modifier.modifier_type == nil then
        graft_def.negotiation_modifier.modifier_type = MODIFIER_TYPE.HIDDEN
    end
    if graft_def.negotiation_modifier then
        graft_def.card_defs = negotiation_defs
    else
        graft_def.card_defs = battle_defs    
    end

    if graft_def.battle_condition then
        graft_def.battle_condition.apply_sound = graft_def.battle_condition.apply_sound or "event:/sfx/battle/status/system/Status_SocialBoon_PowerUp"
    end

    AddGraft( id, graft_def )
end

local function AddMutatorGraft( id, graft_def )
    graft_def.type = GRAFT_TYPE.MUTATOR
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE
    graft_def.series = graft_def.series or CARD_SERIES.GENERAL

    AddGraft( id, graft_def )
end

local function AddStoryGraft( id, graft_def )
    graft_def.type = GRAFT_TYPE.STORY
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE
    AddGraft( id, graft_def )
end

local function AddFeatGraft( id, graft_def )
    graft_def.type = GRAFT_TYPE.FEAT
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE
    graft_def.series = graft_def.series or CARD_SERIES.GENERAL
    AddGraft( id, graft_def )
end

local function AddPerkGraft( id, graft_def )
    graft_def.type = GRAFT_TYPE.PERK
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE
    graft_def.series = graft_def.series or CARD_SERIES.GENERAL
    AddGraft( id, graft_def )
end

local function AddMettleGraft( id, graft_def )
    graft_def.type = GRAFT_TYPE.METTLE_UPGRADE
    graft_def.rarity = graft_def.rarity or CARD_RARITY.UNIQUE
    graft_def.series = graft_def.series or CARD_SERIES.GENERAL
    AddGraft( id, graft_def )
end

local function GetMettleGrafts( character_id )
    local all_grafts = GetAllGrafts()
    local ret = {}
    for i,v in ipairs(all_grafts) do
        if v.type == GRAFT_TYPE.METTLE_UPGRADE and (character_id == nil or (v.series == CARD_SERIES.GENERAL or v.series == character_id)) then
            table.insert(ret, v)
        end
    end
    return ret
end

local function AddBattleCard( id, carddef )
    assert_warning( CONTENT.BATTLE_CARDS[id] == nil, id )

    ValidateBattleCard( id, carddef )
	local def = BattleCardDef(id, carddef)
    def:SetModID( CURRENT_MOD_ID )

    CONTENT.BATTLE_CARDS[id] = def
    BATTLE_CARDS_ARRAY = nil
end

local function GetAllBattleCards()
    if BATTLE_CARDS_ARRAY == nil then
        BATTLE_CARDS_ARRAY = copyvalues( CONTENT.BATTLE_CARDS )
        table.sort( BATTLE_CARDS_ARRAY, function( a, b ) return a.id < b.id end )
    end

    return BATTLE_CARDS_ARRAY
end
 
local function GetAllNegotiationCards()
    if NEGOTIATION_CARDS_ARRAY == nil then
        NEGOTIATION_CARDS_ARRAY = copyvalues( CONTENT.NEGOTIATION_CARDS )
        table.sort( NEGOTIATION_CARDS_ARRAY, function( a, b ) return a.id < b.id end )
    end
    return NEGOTIATION_CARDS_ARRAY
end

local function GetNegotiationCard( card_id )
    return CONTENT.NEGOTIATION_CARDS[ card_id ]
end

local function AddConvoStateGraph(sg, tag)
    if CheckBits( sg.flags, CONVO_FLAGS.DISABLED ) then
        return
    end
    sg:SetModID( CURRENT_MOD_ID )

    if tag == nil then
        assert( CONTENT.CONVOSG[ sg.id ] == nil, sg.id )
        CONTENT.CONVOSG[sg.id] = sg
    else
        local t = CONTENT.CONVOSG_BY_TAG[ tag ]
        if t == nil then
            t = {}
            CONTENT.CONVOSG_BY_TAG[ tag ] = t
        end
        assert( table.arrayfind( t, sg ) == nil, sg.id )
        table.insert( t, sg )
    end
end

local function GetConvoStateGraph(id)
    return CONTENT.CONVOSG[id]
end

local function IterateConvoStateGraphs( tag )
    if tag then
        return ipairs( CONTENT.CONVOSG_BY_TAG[ tag ] or table.empty )
    else
        return pairs( CONTENT.CONVOSG )
    end
end

local function PreloadCharacterAnims()
    for k,v in pairs(CONTENT.CHARACTERBUILDS) do
        if v.file then
            GetAnim(v.file)
        end
    end
    for k,v in pairs(CONTENT.CHARACTERHEADS) do
        if v.anim then
            GetAnim(v.anim)
        end
    end
end

local function GetUnlockTracks()
    return CONTENT.UNLOCK_TRACKS
end

local function GetUnlockTrack( id )
    assert( id )
    return CONTENT.UNLOCK_TRACKS[id or "SAL"] or table.empty
end

local function AddUnlock( unlock_id )
    CONTENT.UNLOCKS[unlock_id] = true
    return unlock_id
end

local function IsUnlock( unlock_id )
    return CONTENT.UNLOCKS[unlock_id]
end

local function AddUnlockTrack( id, track )
    assert( is_instance( track, UnlockTrackDef ))
    CONTENT.UNLOCK_TRACKS[id] = track
end

local function AddAchievement( def )
    for i, v in ipairs( CONTENT.ACHIEVEMENTS ) do
        assert( def.id ~= v.id, v.id )
    end

    table.insert( CONTENT.ACHIEVEMENTS, def )
end

local function GetAllAchievements()
    return CONTENT.ACHIEVEMENTS
end

local function FindAchievement( id )
    for i, def in ipairs( CONTENT.ACHIEVEMENTS ) do
        if def.id == id then
            return def
        end
    end
end

local function AddGrift( def )
    for i, v in ipairs( CONTENT.GRIFTS ) do
        assert( def.id ~= v.id, v.id )
    end

    table.insert( CONTENT.GRIFTS, def )
end

local function GetAllGrifts()
    return CONTENT.GRIFTS
end

local function FindGrift( id )
    for i, def in ipairs( CONTENT.GRIFTS ) do
        if def.id == id then
            return def
        end
    end
end

local function AddScoreComponent(data)
    data.policy = data.policy or SCORE_POLICY.ACCUMULATE
    CONTENT.SCORE_COMPONENTS[data.id] = data
end

local function GetScoreComponent(id)
    return CONTENT.SCORE_COMPONENTS[id]
end

local function GetScoreComponents()
    return CONTENT.SCORE_COMPONENTS
end

function GetPlayerActData(id)
    return CONTENT.ACT_DATA[id]
end

function AddPlayerActData(act_data)
    assert( is_instance( act_data, PlayerAct ), tostr(act_data))
    assert( act_data.id )
    CONTENT.ACT_DATA[ act_data.id ] = act_data
end

function AddPlayerCharacter(dat)
    CONTENT.PLAYER_DATA[dat.id] = dat
    table.insert(CONTENT.SORTED_PLAYER_DATA, dat)

    for k,act_data in ipairs(dat.acts) do
        CONTENT.ACT_DATA[act_data.id] = act_data
        act_data:SetModID( CURRENT_MOD_ID )
    end

    dat:SetModID( CURRENT_MOD_ID )

    table.sort(CONTENT.SORTED_PLAYER_DATA, function(a,b) return (a.data.sort_key or math.huge ) < (b.data.sort_key or math.huge) end)
end

function GetAllPlayerBackgrounds()
    return CONTENT.SORTED_PLAYER_DATA
end

function GetPlayerBackground(id)
    return CONTENT.PLAYER_DATA[id]
end

local function GetAllBattleFlourishes( character_id )
    local def = character_id and GetCharacterDef( character_id )
    local all_cards = GetAllBattleCards()
    local ret = {}
    for i,card in ipairs(all_cards) do
        if card.flourish then
            if not def or not def.card_series or table.contains( def.card_series, card.series ) then
                table.insert(ret, card)
            end
        end
    end
    return ret
end

local function GetAllNegotiationFlourishes( character_id )
    local def = character_id and GetCharacterDef( character_id )
    local all_cards = GetAllNegotiationCards()
    local ret = {}
    for i,card in ipairs(all_cards) do
        if card.flourish then
            if not def or not def.card_series or table.contains( def.card_series, card.series ) then
                table.insert(ret, card)
            end
        end
    end
    return ret
end

---------------------------------------------------------------------
-- Master Content loader.

local function OnWorkshopItemChanged( item )
    print( "OnWorkshopItemChanged", tostring(item.workshop_id))
    local mod = FindMod( tostring(item.workshop_id) )
    if not mod then
        if not CheckBits( item.flags, engine.steam.EItemStateSubscribed ) then
            -- Not subscribed, so we don't care?
        elseif not item.installPath or #item.installPath == 0 then
            -- Not installed, don't even track it yet cause we have no modinit to run.
        else
            mod =
            {
                id = tostring(item.workshop_id),
                workshop_id = item.workshop_id,
                title = item.title,
                installPath = item.installPath,
                workshop = item,
                status = MOD_STATUS.WORKSHOP,
            }
            RegisterMod( mod )
        end
    else
        -- Merely update workshop data.
        mod.workshop = item
        mod.title = mod.title or item.title

        if not CheckBits( item.flags, engine.steam.EItemStateSubscribed ) then
            print( "\tItem no longer subscribed." )
            -- If the mod was already enabled and finalized, keep it for now.
            if not mod.finalized then
                UnregisterMod( mod )
            end
        end
    end

    if TheGame:GetEvents() and mod then
        TheGame:BroadcastEvent( "mod_changed", mod )
    end
end

local function MountContent()
    if BRANCH == "dev" then
        -- Automatically load data_wip in DEV only.
        engine.inst:Mount( "../data_wip" )
        engine.inst:Mount( "../data_loc" )
    end
end

-- check if check_table contains an alias of mod_table
-- if no alias is provided for mod, fall back to its id.
local function TableContainAlias(check_table, mod_table)
    if check_table then
        if check_table == (mod_table.alias or mod_table.id) then
            return true
        elseif type(check_table) == "table" and table.contains(check_table, mod_table.alias or mod_table.id) then
            return true
        end
    end
    return false
end

-- Takes in a list of mods and re-orders them in place.
-- Each mod table may contain two fields: load_before, load_after which are either
-- strings referencing a mod by a alias or id, or a table of such strings.
-- This function uses Kahn’s Algorithm, except it is stable(tries to preserve original order).
-- credit to @RageLeague for this logic.
local function SortRegisteredMods( mod_list )
    -- a list that keeps track of edges so we can do topo sort later
    local edge_list = {}
    for i, mod_table in ipairs(mod_list) do
        -- contains the node for this mod
        local this_mod_node = {
            mod_table = mod_table, 
            load_before = {}, 
            load_after = {}, 
            preloads = 0,
        }
        table.insert(edge_list, this_mod_node)
        -- check dependencies for other mods.
        for j, other_table in ipairs(mod_list) do
            if other_table ~= mod_table then
                if TableContainAlias(mod_table.load_before, other_table) or TableContainAlias(other_table.load_after, mod_table) then
                    this_mod_node.load_before[j] = true
                end
                if TableContainAlias(mod_table.load_after, other_table) or TableContainAlias(other_table.load_before, mod_table) then
                    if not this_mod_node.load_after[j] then
                        this_mod_node.preloads = this_mod_node.preloads + 1
                    end
                    this_mod_node.load_after[j] = true
                end
            end
        end
    end
    -- use a heap to optimize the inserting from O(n) to O(log(n))
    -- this keeps track of which index has all their prereqs loaded already.
    local no_preloads = Heap(function(a,b)return a < b end)
    for i, edge_table in ipairs(edge_list) do
        if edge_table.preloads == 0 then
            no_preloads:push(i)
        end
    end

    local sorted_list = {}
    -- while you still have unassigned items, repeat the following
    while #sorted_list < #edge_list do
        if no_preloads:isempty() then
            -- assert(false, "Topo Sort Error: Circular requirement.")
            return false
        end
        local selected_index = no_preloads:pop()
        table.insert(sorted_list, selected_index)
        for idx, val in pairs(edge_list[selected_index].load_before) do
            edge_list[idx].preloads = edge_list[idx].preloads - 1
            if edge_list[idx].preloads == 0 then
                no_preloads:push(idx)
            end
        end
    end

    table.clear(mod_list)
    for i, idx in ipairs(sorted_list) do
        table.insert(mod_list, edge_list[idx].mod_table)
    end
    return true
end

local function MountMods()
    local function OnListFiles( filenames )
        for k, t in pairs( filenames ) do
            if t.type == "dir" then
                local folder_name = t.name
                local mod =
                {
                    id = folder_name,
                    installPath = engine.inst:ResolveFileName( string.format( "SETTINGS:%s/%s", MOD_FOLDER, folder_name ), false ),
                    status = MOD_STATUS.LOCAL,
                }
                RegisterMod( mod )
            end
        end
    end
   
   engine.settings:ListFiles( MOD_FOLDER.."/*", OnListFiles )

    if BRANCH == "dev" then
        for k, filename in pairs( engine.inst:ListFiles( "../"..MOD_FOLDER.."/*", engine.inst.FileType_Directory )) do
            local mod =
            {
                id = filename,
                installPath = string.format( "%s/%s", "../"..MOD_FOLDER, filename ),
                status = MOD_STATUS.DEV,
            }
            RegisterMod( mod )
        end
    end

    if engine.steam then
        engine.steam:EnumerateWorkshopItems( OnWorkshopItemChanged )
    end

    if not SortRegisteredMods( CONTENT.MODS ) then
        LOGWARN( "Error sorting mod table: circular requirement detected" )
    end

    -- Finalize all enabled mods.
    for i, mod in AllMods() do
        if IsModEnabled( mod ) then
            mod.finalized = true
            if mod.alias then
                engine.inst:Mount( mod.installPath, mod.alias )
            end
        end
    end
end

local function FindAllLocalizations()
    for k, filename in pairs( engine.inst:ListFiles( 'localizations/*.lua' )) do
        -- Run the lua file:
        local name = filename:match( "^(.+)[.]lua$" )
        local fullname = 'localizations/'..name
        print("Loading localization " .. fullname );
        require(fullname)
    end
end

function LoadLanguage( language )
    print("Loading Language "..language)

    -- Grab the language:
    local po_filenames
    local font_settings

    local locinfo = CONTENT.LOCALIZATIONS[language] or CONTENT.LOCALIZATIONS["english"]
    po_filenames = locinfo.po_filenames
    font_settings = locinfo.font_settings

    local loc_translator = require "util/loc_translator"
    local translator = loc_translator.Translator()
    
    if po_filenames then
        table.add( CONTENT.STRINGS, translator:LoadTranslationPoFiles( po_filenames ))
    else
        print( "No po files: defaulting to english string table.")
    end

    if font_settings then
        for alias, fontinfo in pairs( font_settings ) do
            local filename = fontinfo.font
            print( "Map Font:", alias, filename, " SDF threshold: ", fontinfo.sdfthreshold)
            print ( filename, alias, fontinfo.sdfthreshold, fontinfo.sdfboldthreshold, fontinfo.sdfshadowthreshold, fontinfo.supportsitalics )
            engine.inst:RegisterFont( filename, alias, fontinfo.sdfthreshold, fontinfo.sdfboldthreshold, fontinfo.sdfshadowthreshold, fontinfo.supportsitalics)
        end
    end

    print( "PLURALITY:", locinfo.plurality )
    loc_translator.overridePlurality( locinfo.plurality )
end


local function DoLoad( game )
    engine.inst:ProfilerPush("Content Load", 0xCCAA77FF )
      
    MountContent()

    FindAllLocalizations()

    if not CheckAnyBits( SPECIAL_BUILD_TYPE, 128 | 1024 ) then
        MountMods()
    end

    InvokeModAPI( "OnPreLoad" )

    --------------------------------------------------------------------------
    -- Load the strings table first, so that content can localize as it loads.

    game:InitLanguage()

    LoadLanguage( game:GetLanguage() )

    --------------------------------------------------------------------------

    engine.inst:CreateCursor("CURSOR_1", "cursors/cursor_1.bmp", 3, 3)
    engine.inst:CreateCursor("CURSOR_2", "cursors/cursor_2.bmp", 3, 3)
    engine.inst:CreateCursor("CURSOR_3", "cursors/cursor_3.bmp", 3, 3)
    engine.inst:CreateCursor("CURSOR_4", "cursors/cursor_4.bmp", 3, 3)
    engine.inst:CreateCursor("CURSOR_5", "cursors/cursor_5.bmp", 3, 3)
    engine.inst:CreateCursor("CURSOR_6", "cursors/cursor_6.bmp", 3, 3)

    --------------------------------------------------------------------------

    AddNamePool( "NAMES", "scripts/content/names/names.txt" )
    AddNamePool( "SNAILS", "scripts/content/names/snailnames.txt" )
    AddNamePool( "MECH", "scripts/content/names/mechnames.txt" )
    AddNamePool( "PET", "scripts/content/names/petnames.txt" )

    LoadAllPlax()

    do
        engine.inst:ProfilerPush("aspects")
        for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/character_aspects/*.lua')) do
            local id = filename:match( "^(.+)[.]lua$" )
            local aspect_class = require( "content/character_aspects/"..id )
            if is_class( aspect_class, Aspect ) then
                AddAspect( id, aspect_class ) -- Auto-register.
            end
        end
        engine.inst:ProfilerPop()
    end

    -- This includes battle, negotiation, and items
    require "cards/card_collection"
    require "graftcollection"
    require "content/heads"
    require "scoring"

    engine.inst:ProfilerPush("factions")
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/factions/*.lua')) do
        local name = filename:match( "^(.+)[.]lua$" )
        require ( "content/factions/"..name )
    end
    engine.inst:ProfilerPop()

    local codex = Codex()
    for k, filename in pairs(engine.inst:ListFiles( 'codex/*.yaml')) do
        codex:AddFilename( "codex/"..filename )
    end 
    Content.AddCodex( codex )

    Content.ReloadConversation()
    
    engine.inst:ProfilerPush("locations")
    do
        for k, filepath in ipairs( filepath.list_files( "scripts/content/locations/", "*.lua", true )) do
            local name = filepath:match( "scripts/(.+)[.]lua$" )
            require ( name )
        end
    end
    engine.inst:ProfilerPop()

    do
        for k, filepath in ipairs( filepath.list_files( "scripts/content/autoload/", "*.lua", true )) do
            local name = filepath:match( "scripts/(.+)[.]lua$" )
            require ( name )
        end
    end

    -- engine.inst:ProfilerPush("preload character anims")
    -- PreloadCharacterAnims()
    -- engine.inst:ProfilerPop()

    -- Only precache if not in DEBUG, to improve reload times.
    if not TheGame:GetLocalSettings().DEBUG then
        local paperdoll = require "paperdoll_util"
        paperdoll.PrecacheAnims()
    end

    Content.ReloadParticles()

    Content.ReloadSlideShows()

    local strings = require "lang/english"
    Content.AddStringTable( "GLOBAL", strings )
    Content.AddStringTable( "CONVO", require "content/convo_loc_common" )
    
    engine.inst:ProfilerPop()

    -- Specially Handle OnLoad callbacks for all mods.These can return a followup function, so that mods can 
    -- defer behaviour based on what OTHER mods have loaded.
    local mod_fns = {}
    for i, mod in ipairs( CONTENT.MODS ) do
        mod_fns[ mod ] = mod.OnLoad
        LocalizeModOptions( mod )
    end
    while next(mod_fns) ~= nil do
        for i, mod in ipairs( CONTENT.MODS ) do
            local fn = InvokeModAPIForMod( mod, mod_fns[ mod ] )
            if type(fn) == "function" then
                mod_fns[ mod ] = fn
            else
                mod_fns[ mod ] = nil
            end
        end
    end

    -- We load these for all languages, as a fallback for missing strings.
    local loc_translator = require "util/loc_translator"
    loc_translator.HarvestStrings( CONTENT.ENGLISH_STRINGS )
    assert( table.count( CONTENT.ENGLISH_STRINGS ) > 0 )
    if game:GetLanguage() == "english" then
        CONTENT.STRINGS = CONTENT.ENGLISH_STRINGS
    end

    collectgarbage()
end

----------------------------------------------------------------------

return 
{
    DoLoad = DoLoad,
    
    RegisterMod = RegisterMod,
    FindMod = FindMod,
    IsModEnabled = IsModEnabled,
    SetModEnabled = SetModEnabled,
    GetModSetting = GetModSetting,
    SetModSetting = SetModSetting,
    SaveWorkshopID = SaveWorkshopID,
    AllMods = AllMods,
    GetInstalledMods = GetInstalledMods,
    InvokeModAPI = InvokeModAPI,

    GetFaction = GetFaction,
    AddFaction = AddFaction,
    AllFactions = AllFactions,

    AddWorldRegion = AddWorldRegion,
    GetWorldRegion = GetWorldRegion,
    AllWorldRegions = AllWorldRegions,


    GetLocationContent = GetLocationContent,
    GetLocationContents = GetLocationContents,
    
    AddCharacterHeads = AddCharacterHeads,
    AddCharacterHead = AddCharacterHead,
    GetCharacterHead = GetCharacterHead,
    GetCharacterHeads = GetCharacterHeads,
   
    AddAspect = AddAspect,
    GetAspect = GetAspect,
    SortedAspects = SortedAspects,

    GetAllQuests = GetAllQuests,
    AddQuestDef = AddQuestDef,
    GetQuestDef = GetQuestDef,
    ReloadConversation = ReloadConversation,

    AddQuips = AddQuips,

    GetCharacterBuild = GetCharacterBuild,
    AddCharacterBuild = AddCharacterBuild,
    AddCharacterBuilds = AddCharacterBuilds,
    GetAllCharacterBuilds = GetAllCharacterBuilds,

    AddNamePool = AddNamePool,
    ExtendNamePool = ExtendNamePool,
    GenerateCharacterName = GenerateCharacterName,
    GetCharacterName = GetCharacterName,

    AddCharacterDef = AddCharacterDef,
    GetCharacterDef = GetCharacterDef,
    GetAllCharacterDefs = GetAllCharacterDefs,

    AddCharacterSkin = AddCharacterSkin,
    AddCharacterSkins = AddCharacterSkins,
    GetCharacterSkin = GetCharacterSkin,
    GetCharacterSkinByAlias = GetCharacterSkinByAlias,
    GetCharacterForSkin = GetCharacterForSkin,
    GetAllCharacterSkins = GetAllCharacterSkins,
    SaveAllCharacterSkins = SaveAllCharacterSkins,
    CreateCharacterSkin = CreateCharacterSkin,

    LookupQuip = LookupQuip,

    AddCodex = AddCodex,
    LookupCodexEntry = LookupCodexEntry,
    SubstituteCodexFormatting = SubstituteCodexFormatting,

    HatchNegotiationParasite = HatchNegotiationParasite,
    HatchBattleParasite = HatchBattleParasite,

    GetBattleCard = GetBattleCard,
    GetAllBattleCards = GetAllBattleCards,
    AddBattleCard = AddBattleCard,
    GetAllBattleFlourishes = GetAllBattleFlourishes,

    AddNegotiationCard = AddNegotiationCard,
    GetAllNegotiationCards = GetAllNegotiationCards,
    GetNegotiationCard = GetNegotiationCard,
    GetAllNegotiationFlourishes = GetAllNegotiationFlourishes,

    GetLocationsThatUsePlax = GetLocationsThatUsePlax,
    LoadAllPlax = LoadAllPlax,
    AllPlax = AllPlax,
    GetAllPlaxSorted = GetAllPlaxSorted,
    AddPlaxData = AddPlaxData,
    RemovePlaxData = RemovePlaxData,
    GetPlaxPath = GetPlaxPath,
    GetPlaxData = GetPlaxData,

    AddConvoStateGraph = AddConvoStateGraph,
    GetConvoStateGraph = GetConvoStateGraph,
    IterateConvoStateGraphs = IterateConvoStateGraphs,
    
    AddLocationContent = AddLocationContent,
    -- Expose access for debug browsing.
    
    GetGraft = GetGraft,
    GetAllGrafts = GetAllGrafts,
    AddGraft = AddGraft,
    AddSocialGraft = AddSocialGraft,
    AddMutatorGraft = AddMutatorGraft,
    AddStoryGraft = AddStoryGraft,
    AddFeatGraft = AddFeatGraft,
    AddPerkGraft = AddPerkGraft,
    AddMettleGraft = AddMettleGraft,
    GetMettleGrafts = GetMettleGrafts,

    AddOpinionEvent = AddOpinionEvent,

    AddNegotiationModifier = AddNegotiationModifier,
    GetNegotiationModifier = GetNegotiationModifier,
    GetAllNegotiationModifiers = GetAllNegotiationModifiers,

    AddNegotiationCardFeature = AddNegotiationCardFeature,
    GetNegotiationCardFeature = GetNegotiationCardFeature,
    GetNegotiationCardFeatures = GetNegotiationCardFeatures,

    AddBattleCardFeature = AddBattleCardFeature,
    GetBattleCardFeature = GetBattleCardFeature,
    GetAllBattleFeatures = GetAllBattleFeatures,

    AddBattleCondition = AddBattleCondition,
    GetBattleCondition = GetBattleCondition,
    GetAllBattleConditions = GetAllBattleConditions,

    AddUnlock = AddUnlock,
    IsUnlock = IsUnlock,
    AddUnlockTrack = AddUnlockTrack,
    GetUnlockTracks = GetUnlockTracks,
    GetUnlockTrack = GetUnlockTrack,

    AddGrift = AddGrift,
    GetAllGrifts = GetAllGrifts,
    FindGrift = FindGrift,

    AddAchievement = AddAchievement,
    GetAllAchievements = GetAllAchievements,
    FindAchievement = FindAchievement,

    GetAllParticleParams = GetAllParticleParams,
    GetParticleParams = GetParticleParams,
    ReloadParticles = ReloadParticles,

    AddDeck = AddDeck,
    GetDeck = GetDeck,
    GetBattleDecks = GetBattleDecks,
    GetNegotiationDecks = GetNegotiationDecks,
    
    AddOutfit = AddOutfit,
    GetOutfit = GetOutfit,
    GetOutfitsForCharacter = GetOutfitsForCharacter,

    ReloadSlideShows = ReloadSlideShows,
    AddSlideShow = AddSlideShow,
    GetSlideShow = GetSlideShow,
    GetSlideShows = GetSlideShows,

    HarvestStrings = HarvestStrings,
    AddLocalization = AddLocalization,
    AddPOFileToLocalization = AddPOFileToLocalization,
    GetLocalizations = GetLocalizations,

    LookupString = LookupString,
    AddStringTable = AddStringTable,

    AddRumble = AddRumble,
    GetRumble = GetRumble,

    AddScoreComponent = AddScoreComponent,
    GetScoreComponent = GetScoreComponent,
    GetScoreComponents = GetScoreComponents,

    internal = CONTENT,
}
