local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

------------------------------------------------------------------------------------------------
AddOpinionEvent("CELEB_GOSSIP", 
{
    delta = OPINION_DELTAS.MAJOR_BAD,
    txt = "Affected by Celebrity Gossip",
})

local SOCIAL_GRAFTS =
{
    thin_skin =
    {
        name = "Thin Skin",
        desc = "The first time you take damage each fight, gain 2 {WOUND}.",

        battle_condition = 
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated  )
                    if fighter == self.owner and stat == COMBAT_STAT.HEALTH and delta < 0 and not self.used then
                        self.used = true
                        self.owner:AddCondition("WOUND", 2 * self.stacks, self)
                        self.owner:RemoveCondition(self.id)
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    fragile_health =
    {
        name = "Fragile Health",
        desc = "Whenever you gain a debuff, take 2 damage.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                    if fighter == self.owner and condition.ctype == CTYPE.DEBUFF then
                        self.owner:ApplyDamage(2, self.owner, self)
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    elitist =
    {
        name = "Elitist",
        desc = "Feud Citizens that you meet for the first time will dislike you.",

        event_priorities =
        {
            [ "start_convo" ] = 1, -- Prioritized before first_impressions
        },

        event_handlers =
        {
            [ "start_convo" ] = function( self, agent )
                if not agent:KnowsPlayer() and agent:GetFactionID() == "FEUD_CITIZEN" then
                    agent:OpinionEvent( OPINION.ELITIST_FACTION_DISLIKE )
                end
            end,
        }
    },

    whispered_rumors =
    {
        name = "Whispered Rumors",
        desc = "Opponents start with an additional argument each negotiation.",

        negotiation_modifier =
        {
            desc_fn = function( self, fmt_str, engine, widget )
                if self.enemy and widget and widget.PostPortrait then
                    local txt = loc.format( self.def:GetLocalizedString( "RUMORS" ), self.enemy )
                    widget:PostPortrait( self.enemy, txt )
                end
            end,

            max_resolve = 5,
            max_stacks = 1,

            modifier_type = MODIFIER_TYPE.ARGUMENT,
            loc_strings =
            {
                RUMORS = "{1#agent} has been spreading bad rumors about you!",
                DIALOG1 = [[
                    right:
                        {1#agent} warned me about you.
                ]],
                DIALOG2 = [[
                    right:
                        {1#agent} told me you're not to be trusted.
                ]],
                DIALOG3 = [[
                    right:
                        Why should I listen to you and not {1#agent}?
                ]],
            },

            target_enemy = TARGET_ANY_RESOLVE,
            min_persuasion = 1,
            max_persuasion = 5,

            bonus_damage = { 0, 1, 1, 2 },

            OnInit = function( self, enemy )
                self:SetResolve( 3, MODIFIER_SCALING.MED )
                assert( is_instance( enemy, Agent ))
                self.enemy = enemy
                self.engine:BroadcastEvent( EVENT.CUSTOM, self.OnQuipBadRumor, self )
                self.min_persuasion = self.min_persuasion + self.bonus_damage[math.min(#self.bonus_damage, self.engine:GetDifficulty())]
                self.max_persuasion = self.max_persuasion + self.bonus_damage[math.min(#self.bonus_damage, self.engine:GetDifficulty())]
            end,

            OnQuipBadRumor = function( panel, self )
                local dialogid = table.arraypick{ "DIALOG1", "DIALOG2", "DIALOG3" }
                local txt = loc.format( self.def:GetLocalizedString( dialogid ), self.enemy )
                panel:Dialog( txt )
            end,

            OnBeginTurn = function( self, minigame )
                local r = math.random()
                if r < 1 / math.max(1, minigame:GetTurns()) then
                    minigame:BroadcastEvent( EVENT.CUSTOM, self.OnQuipBadRumor, self )
                end
                self:ApplyPersuasion()
            end,
        },

        OnStartNegotiation = function( self, minigame )
            local enemy
            if self.userdata and self.userdata.agents then
                local agents = shallowcopy( self.userdata.agents )
                table.arrayremove( agents, minigame:GetOpponent())
                enemy = table.arraypick( agents )
            end
            if enemy ~= nil then
                local modifier_id = self:GetDef().negotiation_modifier
                local modifier = minigame:GetOpponentNegotiator():CreateModifier( modifier_id, self.stacks or 1, enemy )
                modifier.graft = self
                return modifier
            end
        end,
    },

    blacklisted =
    {
        name = "Blacklisted",
        desc = "At the start of each negotiation, discard a random card.",

        negotiation_modifier = 
        {
            hidden = true,
            event_handlers= 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    if minigame:GetTurns() == 1 then
                        local hand = minigame:GetHandDeck()
                        local card = hand:PeekRandom()
                        if card then minigame:DiscardCard(card) end
                    end                    
                    self.negotiator:RemoveModifier( self )
                end
            },
        },
    },

    hunted =
    {
        name = "Hunted",
        desc = "All enemy Admiralty gain 1 {POWER} at the start of battle.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter:GetTeam() ~= self.owner:GetTeam() and fighter.agent:GetFactionID() == "ADMIRALTY" then
                        fighter:AddCondition("POWER", self.stacks, self)
                    end
                end,
            },
        },
    },

    --BANDITS  

    shakey = 
    {
        name = "Shakey",
        desc = "-2 max resolve.",
        
        OnAdded = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve(-2)
        end,

        OnRemoved = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve(2)
        end,
    },

    curse_of_hesh = 
    {
        name = "Curse of Hesh",
        desc = "Lose 1 action and draw a card on your first turn of battle.",
        
        battle_condition =
        {
            ctype = CTYPE.DEBUFF,
            desc = "Lose <#HILITE>{1} {1*action|actions}</> on your first turn of battle.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.stacks or 1)
            end,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        if self.owner.player_controlled then
                            self.battle:ModifyExertedCount( self.stacks )
                            self.battle:DrawCards(1)
                        end
                    end
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    bad_faith =
    {
        name = "Bad Faith",
        desc = "At the start of each fight, add a {gnarled_effigy} to your draw pile.",

        battle_condition =
        {
            hidden = true,
            card_id = "gnarled_effigy",
            event_handlers =
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    for i=1,self.stacks do
                        local card = Battle.Card( self.card_id, self.owner )
                        battle:DealCard(card)
                    end
                    self.owner:RemoveCondition(self.id, self.stacks)
                end
            },
        },
    },

    bad_credit = 
    {
        name = "Bad Credit",
        desc = "Buying anything from shopkeepers costs an additional 20 shills.",

        loc_strings =
        {
            MODIFIER = "<#PENALTY>{1} shill surcharge from {2}</>",
        },

        event_handlers = 
        {
            [ "calc_payment" ] = function( self, agent, modifiers, reasons, params )
                if params and params.is_shop then
                    modifiers.generic_modifier = (modifiers.generic_modifier or 0) + 20
                    table.insert(reasons, loc.format( self:GetDef():GetLocalizedString( "MODIFIER" ), 20, self:GetName()))
                end
            end
        },
    },

    spiked_drink =
    {
        name = "Spiked Drink",
        desc = "Gain an {negotiation.ulcer} card whenever you drink.",

        event_handlers = 
        {
            ["do_drink"] = function( self, drink_effects )
                table.insert( drink_effects.cards, "ulcer" )
            end
        },
    },

    tainted_supply =
    {
        name = "Tainted Supply",
        desc = "Drinking adds an additional drunk card.",
        
        event_handlers =
        {
            [ "do_drink" ] = function( self, drink_effects )
                table.insert( drink_effects.cards, "drunk_player" )
                table.insert( drink_effects.cards, "drunk" )
            end
        },
    }, 

    intimidated =
    {
        name = "Bully",
        desc = "Lose 1 {DEFEND} on all cards that apply {DEFEND}.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                    if condition_id == "DEFEND" and is_instance( source, Battle.Card ) and source.owner == self.owner then
                        if acc.value > 0 then
                            acc:AddValue( math.max(-acc.value, -self.stacks), self )
                        end
                    end
                end,
            },
        },
    },

    red_tape =
    {
        name = "Red Tape",
        desc = "Members of the Admiralty will not help you in battle or negotiation.",
    
        event_handlers = 
        {
            [ "manipulate_chat" ] = function( self, bribe_params, cxt )
                if cxt:GetAgent():GetFactionID() == "ADMIRALTY" then
                    bribe_params.disable = true
                    bribe_params.disable_reason = loc.format( "{1#graft} : {2}", self.id, self:GetDesc() )
                end
            end
        },
    },

    lowered_rank =
    {
        name = "Lowered Rank",
        desc = "Allies have 10% less max health in battle.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() then
                        local max_health = fighter:GetMaxHealth()
                        local loss = math.min( max_health - 1, math.ceil( max_health * 0.1*self.stacks ) )
                        fighter:DeltaMaxHealth( -loss )
                    end
                end,
            },
        },   
    },

    terrorized = 
    {
        name = "Terrorized",
        desc = "When gained, 2 random Spree dislike you.",
        
        OnAdded = function(self, agent)
            local candidates = ObtainWorkTable()

            for k,agent in pairs(TheGame:GetGameState().agents) do
                if not AgentUtil.HasPlotArmour(agent) and not agent:IsPlayer() and agent:GetFactionID() == "BANDITS" then
                    table.insert(candidates, agent)
                end
            end    

            for i=1, math.min( #candidates, 2 ) do
                local chosen = table.arraypick(candidates)
                table.remove(candidates, candidates.chosen)
                chosen:OpinionEvent( OPINION.TERRORIZED )
            end
            
            ReleaseWorkTable(candidates)
        end,
    },

    marked = 
    {
        name = "Marked",
        desc = "Enemy Spree join battle with less {SURRENDER}.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter:GetTeam() ~= self.owner:GetTeam() and fighter.agent:GetFactionID() == "BANDITS" then
                        local morale = fighter:GetMorale()
                        if morale then
                            local new_morale = morale/(0.5 + self.stacks)
                            fighter:DeltaMorale( new_morale - morale )
                        end
                    end
                end,
            },
        },
    },

    outwitted =
    {
        name = "Outwitted",
        desc = "Pin a random card at the start of battle.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    for i=1,self.stacks do
                        local incepted_card = Battle.Card( "status_pinned", self.owner)
                        incepted_card:DoPinning( battle, self.owner )
                    end
                    self.owner:RemoveCondition(self.id)
                end,
            },
        },
    },

    brutality = 
    {
        name = "Brutality",
        desc = "Status cards cost 1 additional action in battle.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if CheckBits(card.flags, battle_defs.CARD_FLAGS.STATUS) then
                        cost_acc:AddValue(self.stacks, self)
                    end
                end,
            },
        },
    },

    faulty_armor = 
    {
        name = "Faulty Armor",
        desc = "Start each battle with 2 {EXPOSED}.",
        
        battle_condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.owner:AddCondition("EXPOSED", self.stacks * 2, self)
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    crisis =
    {
        name = "Crisis",
        desc = "All allies gain a small amount of {SURRENDER} at the start of your turn.",

        battle_condition =
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    local fighters = shallowcopy( battle:GetPlayerTeam():GetFighters() )
                    for i, fighter in ipairs( fighters ) do
                        fighter:DeltaMorale(0.06)
                    end
                end
            },
        }
    },

    sketchy_equipment =
    {
        name = "Sketchy Equipment",
        desc = "Take 1 damage whenever you use an item.",
        
        battle_condition =
        {
            hidden = true,
            event_priorities =
            {
                -- After quality_materials
                [ BATTLE_EVENT.POST_RESOLVE ] = 11,
            },
            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.card:IsFlagged( battle_defs.CARD_FLAGS.ITEM ) then
                        battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        self.owner:ApplyDamage( 1* self.stacks, self.owner, self, false )

                        -- ccc: there is already a hit reaction via DAMAGE_APPLIED, so this one is redudnant and causing Smith to hang in smith_regeneration.
                        -- battle:BroadcastEvent( BATTLE_EVENT.PLAY_ANIM, self.owner, "hit_mid")
                    end
                end
            },
        },
    },

    civil_unrest =
    {
        name = "Civil Unrest",
        desc = "All allies start battle with additional {SURRENDER}.",
        
        battle_condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() then
                        fighter:DeltaMorale(.15)
                    end
                end,
            },
        },
    },

    hemophiliac = 
    {
        name = "Hemophiliac",
        desc = "Bleed cards deal 2 bonus damage to you.",

        battle_condition =
        {
            ctype = CTYPE.DEBUFF,
            event_handlers = 
            {
                [ battle_defs.BATTLE_EVENT.CARD_DEALT ] = function( self, card, deck )
                    if card.id == "status_bleed" then
                        card.bleed_damage = (card.bleed_damage or 2) + 2
                    end
                end
            },
        },
    },


    fumbling_fingers =
    {
        name = "Fumbling Fingers",
        desc = "Every 3rd turn in battle, discard a random card from your hand.",

        display_number = function( self )
            return self.userdata.counter or 0
        end,
        
        battle_condition =
        {
            event_handlers = 
            {
                [ battle_defs.BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, card, deck )
                    self.graft.userdata.counter = (self.graft.userdata.counter or 0) + 1
                    if self.graft.userdata.counter >= 3 then
                        local card = self.battle:GetHandDeck():PeekRandom()
                        if card and card.id ~= "execute" then
                            self.battle:DiscardCard( card )
                        end
                        self.graft.userdata.counter = 0
                    end
                end
            },
        },
    },

    short_circuit =
    {
        name = "Short Circuit",
        desc = "Take 2 damage whenever you gain {POWER}.",

        battle_condition = 
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                    if condition.id == "POWER" and fighter == self.owner then
                        self.owner:ApplyDamage(2, self.owner, self, false)
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    drone_supplier =
    {
        name = "Drone Supplier",
        desc = "Combat Drone cards cost 1 additional action.",

        battle_condition =
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if card.id == "combat_drone" then
                        cost_acc:AddValue(self.stacks, self)
                    end
                end,
            },
        },
    },

    slander =
    {
        name = "Slander",
        desc = "When you murder a member of the Rise, lose 2 max resolve.",

        event_handlers =
        {
            [ "resolve_battle" ] = function( self, battle, primary_enemy, repercussions )
                for i, fighter in battle:AllFighters() do
                    local agent = fighter.agent
                    if agent:IsDead() and agent:GetFactionID() == "RISE" and fighter:GetKiller() and fighter:GetKiller():IsPlayer() then
                        --
                        TheGame:GetGameState():GetCaravan():DeltaMaxResolve( -2 )
                    end
                end
            end,
        }
    },

    depleted =
    {
        name = "Depleted",
        desc = "On your first turn of battle, draw one less card.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_TURN_DRAW ] = function( self, acc )
                    if self.battle:GetTurns() <= 1 then
                        acc:AddValue( -self.stacks, self )
                    end
                end
            }
        },
    },

    informed_enemies =
    {
        name = "Informed Enemies",
        desc = "Take 2 damage every time you play a card on the first turn of battle.",

        battle_condition =
        {
            ctype = CTYPE.DEBUFF,

            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id, self.stacks, self)
                end,

                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.attacker == self.owner then
                        self.owner:ApplyDamage(2, self.owner, self)
                    end
                end
            },
        },
    },

    faulty_connections =
    {
        name = "Faulty Connections",
        desc = "All friendly arguments come into play with 1 less resolve.",

        negotiation_modifier =
        {
            hidden = true,
            OnApply = function( self, minigame )
                -- Apply to existing modifiers to fix order-of-operation issues with frontal_goad not spawning first.
                for i, modifier in self.negotiator:Modifiers() do
                    if modifier.modifier_type ~= MODIFIER_TYPE.CORE and modifier:GetResolve() then
                        modifier:ModifyResolve(-1, self)
                    end
                end
            end,

            event_handlers = 
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.max_resolve ~= nil then
                        self.engine:PushPostHandler( function()
                            modifier:ModifyResolve(-1, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end )
                    end
                end
            },
        },
    },

    wrenched =
    {
        name = "Wrenched",
        desc = "The first card played every battle costs 1 additional action to play.",

        battle_condition = 
        {
            ctype = CTYPE.DEBUFF,

            event_handlers =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if card.owner == self.owner then
                        cost_acc:AddValue(self.stacks, self)
                    end
                end,

                [ BATTLE_EVENT.START_RESOLVE ] = function( self, battle, card )
                    if card.owner == self.owner then
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end,
            },
        },
    },

    wrong_side =
    {
        name = "Wrong Side",
        desc = "At the start of each negotiation, gain 1 {negotiation.DOUBT}.",

        OnStartNegotiation = function( self, minigame )
            minigame:GetPlayerNegotiator():InceptModifier( "DOUBT", 1 )
        end,
    },

    gremlins = 
    {
        name = "Gremlins",
        desc = "At the end of your first turn each battle, take 2 damage per {CHARGE}.",
 
        battle_condition = 
        {
            ctype = CTYPE.DEBUFF,
            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    local tracker = self.owner:GetCondition("lumin_tracker")
                    if tracker then
                        self.owner:ApplyDamage(tracker:GetCharges() * 2, self.owner, self)
                    end
                    self.owner:RemoveCondition( self.id )
                end
            },
        },
    },

    concussed = 
    {
        name = "Concussed",
        desc = "If you draw more than 4 cards on your turn, your core argument takes 3 damage.",

        display_number = function( self )
            return self.userdata.counter or 0
        end,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                    if card.negotiator == self.negotiator and not start_of_turn then
                        self.graft:IncrementCounter()
                        if self.graft:GetCounter() == 4 then
                            self.negotiator:AttackResolve(3, self)
                            minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                        minigame:BroadcastEvent( EVENT.REFRESH_GRAFTS )
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.graft:ResetCounter()
                    minigame:BroadcastEvent( EVENT.REFRESH_GRAFTS )
                end
            },
        },
    },

    bog_sensitive =
    {
        name = "Bog Sensitive",
        desc = "Whenever you draw a Parasite card in negotiation or battle, take 2 damage.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.owner == self.owner and target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND and CheckBits(card.flags, battle_defs.CARD_FLAGS.PARASITE) then
                        self.owner:ApplyDamage(2 * self.stacks, self.owner, self)
                    end
                end
            },
        },

        negotiation_modifier =
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.negotiator == self.negotiator and target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND and CheckBits(card.flags, negotiation_defs.CARD_FLAGS.PARASITE) then
                        self.negotiator:AttackResolve(2 * self.stacks, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    bogling =
    {
        name = "Bogling",
        desc = "Whenever you draw a Parasite card in negotiation or battle, discard a random card from your hand.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.owner == self.owner and target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND
                        and (source_deck == nil or source_deck:GetDeckType() ~= DECK_TYPE.IN_HAND)
                        and CheckBits(card.flags, battle_defs.CARD_FLAGS.PARASITE) then
                        --
                        local valid_cards = shallowcopy(self.battle:GetHandDeck().cards)
                        for i,card in ipairs(valid_cards) do
                            if card == self then
                                table.remove(valid_cards, i)
                                break
                            end
                        end
                        if #valid_cards > 0 then
                            self.battle:DiscardCard(table.arraypick(valid_cards))
                        end
                    end
                end
            },
        },

        negotiation_modifier =
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.negotiator == self.negotiator and target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND
                        and (source_deck == nil or source_deck:GetDeckType() ~= DECK_TYPE.IN_HAND)
                        and CheckBits(card.flags, negotiation_defs.CARD_FLAGS.PARASITE) then
                        --
                        local valid_cards = shallowcopy(self.engine:GetHandDeck().cards)
                        for i,card in ipairs(valid_cards) do
                            if card == self then
                                table.remove(valid_cards, i)
                                break
                            end
                        end
                        if #valid_cards > 0 then
                            self.engine:DiscardCard(table.arraypick(valid_cards))
                        end
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    --Nadan
    kill_list =
    {
        name = "Kill List",
        desc = "Enemy Spree gain {1} {POWER} at the start of battle.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().power_amt)
        end,

        power_amt = 2,

        battle_condition = 
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() ~= self.owner:GetTeam() and fighter.agent:GetFactionID() == "BANDITS" then
                        fighter:AddCondition("POWER", self.graft:GetDef().power_amt, self)
                    end
                end
            },
        },
    },

    -- Oolo
    wrong_papers =
    {
        name = "Wrong Papers",
        desc = "At the start of every negotiation, gain {1} {FLUSTERED}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().flustered_amt)
        end,

        flustered_amt = 2,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:AddModifier("FLUSTERED", self.graft:GetDef().flustered_amt * self.stacks, self)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier(self.id, self.stacks, self)
                end,
            },
        },
    },

    -- Frizz
    treachery = 
    {
        name = "Treachery",
        desc = "You can no longer gain any {DEFEND} on your first turn of battle.",

        battle_condition =
        {
            hidden = true,

            event_priorities =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = 999,
            },

            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition(self.id, self.stacks, self)
                end,

                [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                    if condition_id == "DEFEND" and source and source.owner == self.owner then
                        acc:ModifyValue(0, self)
                    end
                end,
            },
        },
    },

    -- Rook
    contagious =
    {
        name = "Contagious",
        desc = "Gain a random {PARASITE} card.",

        OnAdded = function(self, agent)
            agent:GetAspect("battler"):LearnCard("twig")
        end,
    },

    -- Fellemo
    scandal = 
    {
        name = "Scandal",
        desc = "At the beginning of every negotiation, lose 1 maximum resolve permanently.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local heart = self.negotiator:FindCoreArgument()
                    local cur, max = heart:GetResolve()
                    heart:SetResolve(cur, nil, max - self.stacks)
                    TheGame:GetGameState():GetCaravan():DeltaMaxResolve(-self.stacks)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier(self.id, self.stacks, self)
                end,
            },
        },
    },

    -- Kalandra
    buried =
    {
        name = "Buried",
        desc = "At the start of each battle, gain 2 {IMPAIR}.",

        battle_condition =
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.owner:AddCondition("IMPAIR", 2 * self.stacks, self)
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    heshian_obscenity =
    {
        name = "Heshian Obscenity",
        desc = "Your maximum hand size in negotiations is reduced by 2.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    minigame:GetHandDeck():SetMaxSize(minigame:GetHandDeck():GetMaxSize() - (2 * self.stacks))
                    self.negotiator:RemoveModifier(self, self.stacks, self)
                end
            },
        },
    },

    back_of_the_line =
    {
        name = "Back of the Line",
        desc = "All items lose 1 use when added to your deck. (minimum of 1 use)",

        event_handlers =
        {
            [ "card_added" ] = function( self, card )
                if CheckBits(card.flags, battle_defs.CARD_FLAGS.ITEM) and card.userdata.charges then
                    card.userdata.charges = math.max(card.userdata.charges - 1, 1)
                end
            end,
        }
    },

    misinformation =
    {
        name = "Misinformation",
        desc = "Opponent arguments deal 2 bonus damage in negotiation.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if is_instance( source, Negotiation.Modifier ) and source.negotiator ~= self.negotiator then
                        persuasion:AddPersuasion( 2, 2, self )
                    end
                end,
            },
        },
    },

    affiliation = 
    {
        name = "Affiliation",
        desc = "At the start of each negotiation, gain 2 {VULNERABILITY} if your opponent is a Heshian.",

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local opponent = minigame:GetOpponentNegotiator().agent
                        if opponent:GetFactionID() == "CULT_OF_HESH" then
                            self.negotiator:AddModifier("VULNERABILITY", self.stacks * 2, self.graft)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                        self.negotiator:RemoveModifier(self.id, 1, self)
                    end
                end
            },
        },
    },

    bad_bookie = 
    {
        name = "Bad Bookie",
        desc = "Lose {1#money} every morning.",
        desc_fn = function ( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().money_loss)
        end,

        loc_strings =
        {
            DIALOG = [[
                * Lost {1#money} from <b>Bad Bookie</>.
            ]], 
        },

        money_loss = 30,

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now and self.userdata.agents and self.userdata.agents[1] then
                    cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), self:GetDef().money_loss )
                    cxt.enc:LoseMoney( self:GetDef().money_loss, self.userdata.agents[1] )
                    self.userdata.last_mail = cxt.now
                end
            end,
        },
    },

    troubled_past =
    {
        name = "Troubled Past",
        desc = "Lose 2 maximum resolve and 2 maximum health.",

        OnAdded = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve(-2)
            agent.health:AddStatModifier( "BANE_TROUBLED_PAST", -2 )
        end,

        OnRemoved = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve(2)
            agent.health:AddStatModifier( "BANE_TROUBLED_PAST", 2 )
        end,
    },

    punchy = 
    {
        name = "Punchy",
        desc = "The first Maneuver card you play each fight is {EXPEND|Expended}.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_CARD_FLAGS ] = function( self, acc, card, target )
                    if card.owner == self.owner and CheckBits(acc.value, battle_defs.CARD_FLAGS.SKILL) then
                        acc:ModifyValue( acc.value | battle_defs.CARD_FLAGS.EXPEND, self )
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if card.owner == self.owner and card:IsFlagged(battle_defs.CARD_FLAGS.SKILL) then
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    butt_kicked = 
    {
        name = "Butt Kicked",
        desc = "Whenever you draw an item card, your core argument takes 2 resolve damage.",

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                    if card:IsFlagged( negotiation_defs.CARD_FLAGS.ITEM ) then
                        self.negotiator:AttackResolve( 2, self )
                    end
                end
            },
        },
    },

    shellshock =
    {
        name = "Shellshock",
        desc = "At the end of every turn, lose 2 {DEFEND}.",

        battle_condition =
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.END_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        if self.owner:HasCondition("DEFEND") then
                            self.owner:RemoveCondition("DEFEND", 2, self)
                            self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                    end
                end
            },
        },
    },

    hounded =
    {
        name = "Hounded",
        desc = "Regain 30% less hp when sleeping.",

        event_handlers =
        {
            [ "do_sleep" ] = function( self, player, sleep_data )
                sleep_data.health_gain = math.round(sleep_data.health_gain * 0.7)
            end,
        }
    },

    bad_mouthed =
    {
        name = "Bad Mouthed",
        desc = "At the start of every negotiation, insert 2 {fabrication} cards into your draw pile.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local cards = {}
                        for i=1,2*self.stacks do
                            table.insert(cards, Negotiation.Card("fabrication", minigame:GetPlayer()))
                        end
                        minigame:DealCards( cards )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    ostracized =
    {
        name = "Ostracized",
        desc = "At the start of each negotiation, you gain 2 {negotiation.VULNERABILITY}.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("VULNERABILITY", 2, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    scorned =
    {
        name = "Scorned",
        desc = "The first Diplomacy card you play each negotiation costs 1 additional action.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if card.negotiator == self.negotiator and card:IsFlagged(negotiation_defs.CARD_FLAGS.DIPLOMACY) then
                        cost_acc:AddValue( 1, self )
                    end
                end,

                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.negotiator == self.negotiator and card:IsFlagged(negotiation_defs.CARD_FLAGS.DIPLOMACY) then
                        self.negotiator:RemoveModifier(self, 1, self)
                    end
                end
            },
        },
    },

    disapproval =
    {
        name = "Disapproval",
        desc = "At the start of any negotiation against a Pearlie or Heshian, gain 1 {negotiation.FLUSTERED}.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local opponent = minigame:GetOpponent()
                        if opponent:GetContentID() == "PEARLIE" or opponent:GetFactionID() == "CULT_OF_HESH" then
                            self.negotiator:AddModifier("FLUSTERED", 1, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    rattled =
    {
        name = "Rattled",
        desc = "Enemy attacks against you deal 1 bonus damage if you have {battle.IMPAIR}.",
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                    if target == self.owner and self.owner:HasCondition("IMPAIR") then
                        dmgt:AddDamage( 1, 1, self )
                    end
                end,
            },
        },
    },

    undermined =
    {
        name = "Undermined",
        desc = "At the start of each negotiation, gain 5 {PLANTED_EVIDENCE}.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("PLANTED_EVIDENCE", 5, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    shunned = 
    {
        name = "Shunned",
        desc = "At the end of your first turn in negotiations, lose all {RENOWN} and {INFLUENCE}.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:RemoveModifier("INFLUENCE", self.negotiator:GetModifierStacks("INFLUENCE"), self)
                        self.negotiator:RemoveModifier("RENOWN", self.negotiator:GetModifierStacks("RENOWN"), self)
                        self.negotiator:RemoveModifier(self.id, 1, self)
                    end
                end
            },
        },
    },

    disorder =
    {
        name = "Disorder",
        desc = "At the start of each battle, gain 4 {disrupt}.",
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:AddCondition("disrupt", 4, self)
                        local deck = self.battle:GetHandDeck()
                        deck:SetMaxSize( self.battle.MAX_HAND_SIZE - 4 )
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end,
            },
        },
    },

    heathen =
    {
        name = "Heathen",
        desc = "At the start of each negotiation, your opponent gains 1 {prayer_of_hesh}.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.anti_negotiator:AddModifier("prayer_of_hesh", 1, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    watched =
    {
        name = "Watched",
        desc = "The first time you deploy an argument in each negotiation, it takes 1 damage.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                    if modifier.negotiator == self.negotiator and source and source.negotiator == self.negotiator then
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        modifier:AttackResolve(self.stacks, self)
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    inconvenience =
    {
        name = "Inconvenience",
        desc = "{FATIGUED} occurs 1 turn earlier in both battle and negotiation.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local cur = minigame:GetFatiguedTurn()
                        minigame:SetFatiguedTurn(cur - self.stacks)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    local cur = battle:GetFatiguedTurn()
                    battle:SetFatiguedTurn(cur - self.stacks)
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    self.owner:RemoveCondition(self.id)
                end,
            },
        }
    },

    celebrity_gossip =
    {
        name = "Celebrity Gossip",
        desc = "At the start of each day, a random person will hate you.",

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now then
                    local candidates = ObtainWorkTable()
                    local gamestate = TheGame:GetGameState()
                    for k,agent in pairs(gamestate.agents) do
                        if not AgentUtil.HasPlotArmour(agent) and
                            not agent:IsInPlayerParty() and
                            agent:GetRelationship() < RELATIONSHIP.LOVED and
                            agent:IsSentient() then
                            table.insert(candidates, agent)
                        end
                    end

                    for i = #candidates + 1, 1 do
                        if TheGame:GetGameState() then
                            local new_agent = gamestate:AddSkinnedAgent()
                            table.insert( candidates, new_agent )
                        end
                    end

                    table.shuffle( candidates )

                    for i=1, math.min( #candidates, 1 ) do
                        local chosen = table.remove( candidates )
                        chosen:OpinionEvent( OPINION.CELEB_GOSSIP )
                    end
                    
                    ReleaseWorkTable(candidates)
                    self.userdata.last_mail = cxt.now
                end
            end,
        },
    },

    sway =
    {
        name = "Sway",
        desc = "At the start of any negotiation against Admiralty or Heshians, your opponent gains 1 {INTERROGATE}.",
        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local opponent = minigame:GetOpponentNegotiator().agent
                        if opponent:GetFactionID() == "CULT_OF_HESH" or opponent:GetFactionID() == "ADMIRALTY" then
                            self.anti_negotiator:AddModifier("INTERROGATE", 1, self.graft)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                        self.negotiator:RemoveModifier(self.id, 1, self)
                    end
                end
            },
        },
    },
}

AddOpinionEvent("TERRORIZED", 
{
    delta = OPINION_DELTAS.DISLIKE,
    txt = "Affected by Terrorized",
})

---------------------------------------------------------------------------------------------

for id, graft in pairs( SOCIAL_GRAFTS ) do
    graft.is_bane = true
    Content.AddSocialGraft( id, graft )
end
