local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

------------------------------------------------------------------------------------------------

local SOCIAL_GRAFTS =
{
    bolstered_health =
    {
        name ="Bolstered Health",
        desc = "+{1} Max Health.",
        is_good = true,
        desc_fn = function( self, fmt_str ) return loc.format( fmt_str, self:GetDef().health_bonus ) end,

        health_bonus = 2,

        OnAdded = function(self, agent)
            local health = self.owner:GetAspect("health")
            local cur, max = health:Get()
            health:AddStatModifier("BOON_BOLSTER", self:GetDef().health_bonus)
        end,
        OnRemoved = function(self, agent)
            local health = self.owner:GetAspect("health")
            local cur, max = health:Get()
            health:AddStatModifier("BOON_BOLSTER", -self:GetDef().health_bonus)
        end,
    },
    
    bolstered_resolve =
    {
        name ="Bolstered Resolve",
        desc = "+{1} Max Resolve.",
        is_good = true,
        desc_fn = function( self, fmt_str ) return loc.format( fmt_str, self:GetDef().resolve_bonus or 0 ) end,

        resolve_bonus = 2,

        OnAdded = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( self:GetDef().resolve_bonus )
        end,
        
        OnRemoved = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( self:GetDef().resolve_bonus )
        end,

    },

    --JAKES
    -- Dockhand
    bargaining = 
    {
        name = "Bargaining",
        desc = "20 shill discount from shopkeepers.",

        loc_strings =
        {
            MODIFIER = "<#BONUS>{1} shill discount from {2}</>",
        },

        event_handlers = 
        {
            [ "calc_payment" ] = function( self, agent, modifiers, reasons, params )
                if params and params.is_shop then
                    modifiers.generic_modifier = (modifiers.generic_modifier or 0) - 20
                    table.insert(reasons, loc.format( self:GetDef():GetLocalizedString( "MODIFIER" ), 20, self:GetName()))
                end
            end
        },
    },

    -- Smuggler
    first_aid = 
    {
        name = "First Aid",
        desc = "Whenever you remove a debuff, {HEAL} 5 health.",

        battle_condition =
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
                    if self.battle:GetFighterFromSource( source ) == self.owner and source ~= condition and condition.ctype == CTYPE.DEBUFF then
                        self.owner:HealHealth(5 * self.stacks, self)
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,
            },
        },
    },

    -- LABORER
    voice_of_the_people =
    {
        name = "Voice of the People",
        desc = "Start all negotiations with a {voice_of_the_people} argument.",

        negotiation_modifier = 
        {
            desc = "This argument's resolve damage doubles for each stack.",
            target_enemy = TARGET_ANY_RESOLVE,
            modifier_type = MODIFIER_TYPE.ARGUMENT,

            OnSetStacks = function( self, old_stacks )
                self.min_persuasion = math.floor(math.pow(2, self.stacks))
                self.max_persuasion = self.min_persuasion
            end,

            max_resolve = 4,

            min_persuasion = 2,
            max_persuasion = 2,

            target_enemy = TARGET_ANY_RESOLVE,
            modifier_type = MODIFIER_TYPE.ARGUMENT,

            event_handlers =
            {
                [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self:ApplyPersuasion()
                    end
                end
            },
        },
    },

    -- HEAVY LABORER
    heavy_lifting =
    {
        name = "Heavy Lifting",
        desc = "At the start of each battle, gain 4 {DEFEND} and 2 {TEMP_POWER}.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    self.owner:AddCondition("DEFEND", 4 * self.stacks, self)
                    self.owner:AddCondition("POWER", 2 * self.stacks, self)
                    self.owner:AddCondition("POWER_LOSS", 2 * self.stacks, self)
                    self.owner:RemoveCondition(self.id)
                end,
            },
        },
    },

    -- FOREMAN
    authority =
    {
        name = "Authority",
        desc = "At the start of each negotiation, add a random authority card to your draw pile.",

        negotiation_modifier =
        {
            hidden = true,
            cards = {"by_the_book", "proper_channels", "permits"},
            event_handlers =
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    for i=1,self.stacks do
                        local chosen_card = self.cards[math.random(#self.cards)]
                        local card = Negotiation.Card( chosen_card, minigame:GetPlayer() )
                        card:ScaleToDifficulty(minigame)
                        minigame:DealCard(card)
                    end
                    self.negotiator:RemoveModifier(self.id, self.stacks)
                end
            },
        },
    },


    --ADMIRALTY

    confidence = 
    {
        name ="Confidence",
        desc = "+5 Max Health.",
        
        bonus = 5,
        OnAdded = function(self, agent)
            local health = self.owner:GetAspect("health")
            health:AddStatModifier("BOON_CONFIDENCE", self:GetDef().bonus)
        end,
        OnRemoved = function(self, agent)
            local health = self.owner:GetAspect("health")
            health:AddStatModifier("BOON_CONFIDENCE", -self:GetDef().bonus)
        end,
    },

    -- Goon
    piece_of_the_pie = 
    {
        name = "Piece of the Pie",
        desc = "Gain {1#money} every morning.",
        desc_fn = function ( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().money_gain)
        end,

        loc_strings =
        {
            DIALOG = [[* Someone has left an envelope for you. It carries the Admiralty crest, stamped on in red wax.
* Scrawled on the front is a hasty thanks from {1}, and inside you find {2#money}.]], 
        },

        money_gain = 75,

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now then
                    if self.userdata.agents and self.userdata.agents[1] then
                        cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), self.userdata.agents[1]:GetName(), self:GetDef().money_gain )
                        cxt.enc:GainMoney( self:GetDef().money_gain )
                    end
                    self.userdata.last_mail = cxt.now
                end
            end,
        },
    },

    --HESH
    -- Luminari

    bell_of_hesh =
    {
        name = "Bell of Hesh",
        desc = "Draw 1 extra card at the start of negotiation.",
        
        negotiation_modifier =
        {
            event_handlers =
            {
                [ EVENT.CALC_TURN_DRAW ] = function( self, acc )
                    acc:AddValue( self.stacks or 1, self )
                end,

                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:RemoveModifier(self)
                end,
            },
        },
    },

    -- Priest
    bio_feedback =
    {
        name = "Bio-Feedback",
        desc = "Status Cards cost 1 less action in battle.",
        
        battle_condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if CheckBits(card.flags, battle_defs.CARD_FLAGS.STATUS) then
                        cost_acc:AddValue(-self.stacks, self)
                    end
                end,
            },
        },
    },

    --MERCHANT
    haggle_badge =
    {
        name = "Haggle Badge",
        desc = "You can negotiate for greater quest rewards.",
    },

    --Chemist
    doze_bug =
    {
        name = "Doze Bug",
        desc = "Whenever you sleep, increase your max health by {1}.",
        desc_fn = function( self, fmt_str ) return
            loc.format( fmt_str, self:GetDef().health_bonus )
        end,
        
        health_bonus = 5,

        loc_strings =
        {
            DIALOG = "* Doze Bug increased max health by {1}",
        },

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_sleep or -1) < cxt.now then
                    local health = self.owner:GetAspect("health")
                    health:AddStatModifier("BOON_DOZE_BUG", self:GetDef().health_bonus)
                    cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), self:GetDef().health_bonus )
                    self.userdata.last_sleep = cxt.now
                end
            end,
        },
    },

    --BARTENDER
    mixology =
    {
        name = "Mixology",
        desc = "Whenever you drink to restore resolve, also restore 4 health.",
        
        event_handlers = 
        {
            ["do_drink"] = function( self, drink_effects )
                drink_effects.health_gain = (drink_effects.health_gain or 0) + 4
            end
        },
    },

    top_shelf =
    {
        name = "Top Shelf",
        desc = "Restore an additional 6 resolve when drinking.",
        
        event_handlers = 
        {
            [ "do_drink" ] = function( self, drink_effects )
                drink_effects.resolve_gain = drink_effects.resolve_gain + 6
            end
        },
    },

    torpor_fiend =
    {
        name = "Torpor Fiend",
        desc = "When you sleep, gain full experience on a random card.",
        desc_fn = function( self, fmt_str ) return loc.format( fmt_str, self:GetDef().xp_amount) end,
        
        loc_strings =
        {
            DIALOG = "* You gained {1} experience on {2#card}",
        },

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now then
                    local negotiation_cards = TheGame:GetGameState():GetPlayerAgent().negotiator:GetCards()
                    local battle_cards = TheGame:GetGameState():GetPlayerAgent().battler:GetCards()
                    local options = {}
                       
                    for i=1, #negotiation_cards do
                        local cur, max = negotiation_cards[i]:GetXP()
                        if negotiation_cards[i]:HasUpgrades() and cur and max and max > 0 and not negotiation_cards[i]:UpgradeReady() and not negotiation_cards[i]:IsHatched() then
                            table.insert(options, negotiation_cards[i])
                        end                        
                    end
                    for i=1, #battle_cards do
                        local cur, max = battle_cards[i]:GetXP()
                        if battle_cards[i]:HasUpgrades() and cur and max and max > 0  and not battle_cards[i]:UpgradeReady() and not battle_cards[i]:IsHatched() then
                            table.insert(options, battle_cards[i])
                        end                        
                    end
                    local card
                    if #options > 1 then
                        card = options[math.random(#options)]
                    else
                        card = options[1]
                    end
                    if card then
                        local cur, max = card:GetXP()
                        local xp_amount = max - cur
                        card:AddXP(xp_amount)
                        cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), xp_amount, card )
                    end
                    self.userdata.last_mail = cxt.now
                end
            end,
        },
    },

    accelerator =
    {
        name = "Accelerator",
        desc = "The first card played in negotiation or battle gains +2 XP.",
        
        negotiation_modifier =
        {
            event_handlers =
            {
                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.owner == self.owner then
                        if not card:IsHatched() then
                            card:AddXP( 2 * self.stacks )
                        end
                        self.negotiator:RemoveModifier( self )
                    end
                end,
            }
        },

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack and attack.attacker == self.owner and attack.card then
                        if not attack.card:IsHatched() then
                            attack.card:AddXP( 2 * self.stacks )
                        end
                        self.owner:RemoveCondition( self.id )
                    end
                end,
            },
        },
    },

    -- Admiralty Guard
    sparring =
    {
        name = "Sparring",
        desc = "Gain 1 bonus {DEFEND} on all cards that apply {DEFEND}.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                    if condition_id == "DEFEND" and is_instance( source, Battle.Card ) and source.owner == self.owner then
                        if acc.value > 0 then
                            acc:AddValue( self.stacks, self )
                        end
                    end
                end,
            },
        },
    },

    -- Admiralty clerk
    authorization =
    {
        name = "Authorization",
        desc = "Once a day, you can ask a nearby member of the Admiralty to help you in battle or negotiation.",
        desc_fn = function( self, fmt_str )
            if self:GetDef().OnCooldown( self ) then
                return loc.format( "{1}\n<#PENALTY>{2}</>", fmt_str, "On Cooldown" )
            else
                return fmt_str
            end
        end,
        
        OnCooldown = function( self )
            return self:HasMemory( "COOLDOWN", 1 )
        end,

        StartCooldown = function( self )
            self:Remember( "COOLDOWN" )
        end,

        event_handlers = 
        {
            [ "manipulate_chat" ] = function( self, bribe_params, cxt )
                if cxt:GetAgent():GetFactionID() == "ADMIRALTY" then
                    -- Disable default manipulate options; see authorization_graft_chat.
                    if not self:GetDef().OnCooldown( self ) then
                        bribe_params.hidden = true
                    end
                end
            end
        },
    },

    -- Admiralty Patrol Leader
    commander =
    {
        name = "Commander",
        desc = "All allies gain +5 health in battle.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() then
                        fighter:DeltaMaxHealth(5 * self.stacks)
                        fighter:HealHealth(5 * self.stacks, self)
                    end
                end,
            },
        },        
    },

    -- Bandit Captain
    savagery =
    {
        name = "Savagery",
        desc = "+1 {BLEED} to all cards that apply {BLEED}.",
        
        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                    if fighter ~= self.owner and source and source.owner == self.owner and condition.id == "BLEED" then
                        fighter:AddCondition("BLEED", self.stacks)
                    end
                end,
            },
        },
    },

    -- Bandit Goon
    sucker_punch = 
    {
        name ="King Hit",
        desc = "+{1} Damage on the first attack in battle.",
        desc_fn = function( self, fmt_str ) return loc.format( fmt_str, 4) end,

        battle_condition =
        {
            desc = "+{1} Damage on the first attack in battle.",
            desc_fn = function( self, fmt_str ) return loc.format( fmt_str, 4 * self.stacks) end,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                    if card.owner == self.owner then
                        dmgt:AddDamage( 4 * self.stacks, 4 * self.stacks, self )
                    end
                end,

                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if self.owner == attack.owner and attack.card:IsAttackCard() and attack.card.min_damage and attack.card.min_damage > 0 then
                        self.owner:RemoveCondition(self.id)
                    end
                end
            },
        },

    },

    -- Bandit Goon2
    surprise_attack = 
    {
        name ="Surprise Attack",
        desc = "Gain 1 action at the start of battle.",

        battle_condition =
        {
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:ModifyActionCount(self.stacks)
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    -- Bandit Raider
    regimented =
    {
        name = "Regimented",
        desc = "All allied Spree have +10 Health in battle.",
        
        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() and fighter.agent:GetFactionID() == "BANDITS" then
                        fighter:DeltaMaxHealth(10 * self.stacks)
                        fighter:DeltaHealth(10 * self.stacks)
                    end
                end,
            },
        },
    },

    -- Spark Baron Professional
    spark_visor =
    {
        name = "Spark Visor",
        desc = "At the start of battle, mark an enemy. Attacking that enemy grants you 8 {DEFEND}.",
        
        battle_condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_COMBAT ] = function( self )
                    local battle = self.battle
                    local target_fighters = {}
                    battle:CollectRandomTargets( target_fighters, self.owner:GetEnemyTeam().fighters, self.stacks )

                    for i=1, #target_fighters do
                        local target = target_fighters[i]
                        target:AddCondition("SPARK_VISOR_MARK", 1, self)
                    end
                end,
            },
        },
    },

    -- Spark baron goon
    weapons_supplier =
    {
        name = "Weapons Supplier",
        desc = "Receive a random item card every morning.",
        loc_strings =
        {
            DIALOG = [[* You notice a large box with the Spark Barons logo on it.
* Upon closer inspection, you see {1}'s name on it. You open it to find a {2#card}.]],
            OPT_TAKE = "Take {1#card}",
            OPT_LEAVE_CARD = "Leave {1#card}",
        },

        available_items = {"stun_gun", "spark_grenade", "shrapnel_grenade"},

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now then
                    self.userdata.chosen_item = self.userdata.chosen_item or self:GetDef().available_items[math.random(#self:GetDef().available_items)]
                    cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), self.userdata.agents and self.userdata.agents[1] and self.userdata.agents[1]:GetName() or "your friend", self.userdata.chosen_item )
                    cxt:Opt(self:GetDef():GetLocalizedString( "OPT_TAKE" ), self.userdata.chosen_item)
                        :GainCards{self.userdata.chosen_item}
                        :Fn(function(cxt)
                            self.userdata.last_mail = cxt.now
                            self.userdata.chosen_item = nil
                        end)
                end
            end,
        },
    },

    -- Rise Rebel
    morale_boost =
    {
        name = "Morale Boost",
        desc = "At the end of their turn, surrendered allies gain 10% of their maximum health.",

        battle_condition =
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
                    if fighter:GetTeam() == self.owner:GetTeam() and condition.id == "SURRENDER" and condition.stacks > 0 then
                        fighter:HealHealth(math.round(fighter:GetMaxHealth() * 0.1 * self.stacks))
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,
            },
        },
    },

    -- Zyn
    sharpened_blades =
    {
        name = "Sharpened Blades",
        desc = "All your attacks apply 2 {BLEED} on your first turn of battle.",

        battle_condition =
        {
            event_handlers =
            {
                [ battle_defs.BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.card:IsAttackCard() and attack.attacker == self.owner then
                        attack:AddCondition( "BLEED", 2 * self.stacks, attack.card )
                    end
                end,

                [ battle_defs.BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:RemoveCondition( self.id )
                end,
            },
        },
    },

    -- Hanbi
    steady_hands =
    {
        name = "Steady Hands",
        desc = "Every 3rd turn in battle, draw 3 additional cards.",

        battle_condition =
        {
            counter = 0,
            event_handlers = 
            {
                [ battle_defs.BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, card, deck )
                    self.counter = self.counter + 1
                    if self.counter >= 3 then
                        self.battle:DrawCards(3)
                        self.counter = 0
                    end
                end
            },
        },
    },

    -- Sparky
    system_shock =
    {
        name = "System Shock",
        desc = "If you didn't attack last turn, gain 2 {TEMP_POWER}.",

        battle_condition =
        {
            hidden = true,
            event_handlers = 
            {
                [ battle_defs.BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    if self.no_attacks then
                        battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        self.owner:AddCondition("POWER", 2, self)
                        self.owner:AddCondition("POWER_LOSS", 2, self)
                    end
                    self.no_attacks = true
                end,

                [ battle_defs.BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack and attack.card and attack.card:IsAttackCard() and attack.card.owner == self.owner then
                        self.no_attacks = false
                    end
                end
            },
        },
    },

    -- Jeol
    drone_mastery =
    {
        name = "Drone Mastery",
        desc = "Allied mechs have 10 additional max Health.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() and fighter.agent:GetSpecies() == SPECIES.MECH then
                        fighter:DeltaMaxHealth(10 * self.stacks)
                        fighter:DeltaHealth(10 * self.stacks)
                    end
                end,
            },
        },
    },

    -- Rise Pamphleteer
    propaganda_machine =
    {
        name = "Propaganda Machine",
        desc = "Start each negotiation with 1 {negotiation.INFLUENCE}.",

        OnStartNegotiation = function( self, minigame )
            minigame:GetPlayerNegotiator():AddModifier( "INFLUENCE", 1 )
        end,
    },

    -- Rise Radical
    war_machine =
    {
        name = "War Machine",
        desc = "Receive a free {card.rpg} once a day from a member of the Rise.",
    },

    rentorian_battle_armor =
    {
        name = "Rentorian Battle Armor",
        desc = "Gain 3 {POWER} at the start of each battle. Lose 3 {POWER} the first time you take any damage.",

        battle_condition = 
        {
            hidden = true,
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    if not self.applied_power then
                        self.owner:AddCondition("POWER", 3 * self.stacks, self)
                        self.applied_power = true
                    end
                end,

                [ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
                    if fighter == self.owner and self.applied_power then
                        self.owner:RemoveCondition("POWER", 3 * self.stacks, self)
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    on_the_list =
    {
        name = "On The List",
        desc = "All friendly arguments come into play with 1 additional resolve.",

        negotiation_modifier =
        {
            hidden = true,
            OnApply = function( self, minigame )
                -- Apply to existing modifiers to fix order-of-operation issues with frontal_goad not spawning first.
                for i, modifier in self.negotiator:Modifiers() do
                    if modifier.modifier_type ~= MODIFIER_TYPE.CORE and modifier:GetResolve() then
                        modifier:ModifyResolve(1, self)
                    end
                end
            end,

            event_handlers = 
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.max_resolve ~= nil then
                        self.engine:PushPostHandler( function()
                            modifier:ModifyResolve(1, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end )
                    end
                end
            },
        },
    },

    gear_head =
    {
        name = "Gear Head",
        desc = "Heal 1 health at the start of your turn if you took no damage last turn.",

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    if self.damage_taken == false then
                        self.owner:HealHealth(1)
                    end
                end,

                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if hit.target == self.owner and not (hit.defended or hit.evaded) then
                        self.damage_taken = true
                    end
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.damage_taken = false
                end
            },
        },
    },

    people_person =
    {
        name = "People Person",
        desc = "At the start of each negotiation, gain {negotiation.INFLUENCE 2} and {negotiation.DOMINANCE 2}.",

        OnStartNegotiation = function( self, minigame )
            minigame:GetPlayerNegotiator():AddModifier( "INFLUENCE", 2 )
            minigame:GetPlayerNegotiator():AddModifier( "DOMINANCE", 2 )
        end,
    },

    heshian_armature =
    {
        name = "Heshian Armature",
        desc = "Whenever you gain {PRIMED}, gain 2 {SURGE}.",

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.PRIMED_GAINED ] = function( self, battle )
                    self.owner:AddCondition("SURGE", 2, self)
                end,
            },
        },
    },

    bogger_pauldrons = 
    {
        name = "Bogger Pauldrons",
        desc = "At the end of your turn, gain 1-3 {DEFEND}.",

        battle_condition =
        {
            desc = "At the end of your turn, gain <#HILITE>{1}-{2}</> {DEFEND}.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks, self.stacks * 3)
            end,

            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle)
                    self.owner:AddCondition("DEFEND", self.stacks * math.random(1, 3), self)
                end
            },
        },
    },

    inoculated =
    {
        name = "Inoculated",
        desc = "Whenever you draw a Parasite card, heal 2 health or restore 2 resolve.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.DRAW_CARD ] = function( self, battle, card, start_of_turn )
                    if card.owner == self.owner and CheckBits(card.flags, battle_defs.CARD_FLAGS.PARASITE) then
                        self.owner:HealHealth(2 * self.stacks, self)
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,
            },
        },

        negotiation_modifier =
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.negotiator == self.negotiator and (target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND) and CheckBits(card.flags, negotiation_defs.CARD_FLAGS.PARASITE) then
                        self.negotiator:RestoreResolve(2 * self.stacks, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    talisman =
    {
        name = "Talisman",
        desc = "Whenever you draw a {PARASITE} card, draw a card.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.owner == self.owner and target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND and CheckBits(card.flags, battle_defs.CARD_FLAGS.PARASITE) then
                        self.battle:DrawCards(1)
                    end
                end
            },
        },
        negotiation_modifier =
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if card.negotiator == self.negotiator and (target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND) and CheckBits(card.flags, negotiation_defs.CARD_FLAGS.PARASITE) then
                        self.engine:DrawCards(1)
                    end
                end
            },
        },
    },

    --Nadan
    chieftain = 
    {
        name = "Chieftain",
        desc = "All allied Spree gain {1} {POWER} at the start of battle.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().power_amt)
        end,

        power_amt = 2,

        battle_condition = 
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
                    if fighter ~= self.owner and fighter:GetTeam() == self.owner:GetTeam() and fighter.agent:GetFactionID() == "BANDITS" then
                        fighter:AddCondition("POWER", self.graft:GetDef().power_amt, self)
                    end
                end
            },
        },
    },

    -- Oolo
    admiralty_medals = 
    {
        name = "Admiralty Medals",
        desc = "All friendly arguments gain 1 {COMPOSURE} at the start of your turn.",

        negotiation_modifier = 
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    for i,arg in self.negotiator:Modifiers() do
                        arg:DeltaComposure(1, self)
                    end
                end
            },
        },
    },

    -- Frizz
    weapons_expert =
    {
        name = "Weapons Expert",
        desc = "The first attack card you play each fight costs 0.",

        battle_condition =
        {
            hidden = true,

            event_priorities =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_SETTOR,
            },

            event_handlers = 
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if card.owner == self.owner and card:IsAttackCard() then
                        cost_acc:ModifyValue(0, self)
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if card.owner == self.owner and card:IsAttackCard() then
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                end
            },
        },
    },

    -- Rook
    grumpy_scowl =
    {
        name = "Grumpy Scowl",
        desc = "At the start of each negotiation, gain {1} {HEATED}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().heated_amt)
        end,

        heated_amt = 2,

        negotiation_modifier = 
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:AddModifier("HEATED", self.graft:GetDef().heated_amt * self.stacks, self)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier(self.id, self.stacks, self)
                end,
            },
        },
    },

    -- Fellemo
    scheme = 
    {
        name = "Scheme",
        desc = "At the start of every negotiation gain either {INFLUENCE 3} or {DOMINANCE 3}. Chosen randomly.",

        negotiation_modifier = 
        {
            hidden = true,

            event_handlers = 
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:AddModifier(math.random() < 0.5 and "INFLUENCE" or "DOMINANCE", 3, self)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier(self.id, self.stacks, self)
                end,
            },
        },
    },

    -- Kalandra
    excavator =
    {
        name = "Excavator",
        desc = "The first item card you play in battle is played twice.",
        icon_override = "digger",
        battle_condition =
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_COMBAT ] = function( self, minigame )
                    self.ready_to_fire = true
                end,

                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )                    
                    if self.ready_to_fire and attack.attacker == self.owner and attack.card:IsFlagged(battle_defs.CARD_FLAGS.ITEM) then
                        self.ready_to_fire = false
                        battle:PlayCard( attack.card )
                    end
                end
            },
        },
    },

    grace_of_hesh =
    {
        name = "Grace of Hesh",
        desc = "{HEAL} 3 health whenever you win a negotiation.",

        event_handlers =
        {
            [ "award_loot" ] = function( self, loot, data )
                if data.negotiation and not data.defeat then
                    loot.health = (loot.health or 0) + 3
                end
            end,
        }
    },

    quality_materials =
    {
        name = "Quality Materials",
        desc = "Whenever you play an Item card, gain 2 {DEFEND}.",

        battle_condition =
        {
            hidden = true,
            event_priorities =
            {
                -- Before sketchy_equipment
                [ BATTLE_EVENT.POST_RESOLVE ] = 10,
            },
            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.card:IsFlagged( battle_defs.CARD_FLAGS.ITEM ) then
                        battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        self.owner:AddCondition("DEFEND", self.stacks * 2, self)
                    end
                end
            },
        },
    },

    boosted_cybernetics =
    {
        name = "Boosted Cybernetics",
        desc = "Gain {1} {MENDING} at the start of each battle.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().mending_amt)
        end,

        mending_amt = 3,

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    self.owner:AddCondition("MENDING", self.graft:GetDef().mending_amt, self)
                    self.owner:RemoveCondition(self.id)
                end,
            },
        },
    },

    throng =
    {
        name = "Throng",
        desc = "The first argument you create each negotiation has 2 bonus resolve.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                    if source and source.negotiator and source.negotiator == self.negotiator and source ~= self and
                        not (is_instance(source, Negotiation.Card) and CheckBits(source.flags, negotiation_defs.CARD_FLAGS.BYSTANDER)) and
                        modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN and modifier.modifier_type ~= MODIFIER_TYPE.INCEPTION then
                        modifier:ModifyResolve(2 * self.stacks, self.graft)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, 1)
                    end
                end
            },
        },
    },

    adhesive =
    {
        name = "Adhesive",
        desc = "At the start of each battle, choose a card in your hand and give it {STICKY}.",

        battle_condition = 
        {
            hidden = true,

            loc_strings =
            {
                CHOOSE_ADHESIVE_CARD = "Choose 1 card and give it Sticky",
            },

            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    local cards = battle:ChooseCardsFromTable( battle:GetHandDeck(), 0, self.stacks, nil, self:GetLocalizedString("CHOOSE_ADHESIVE_CARD"))
                    if cards then
                        for i,card in ipairs(cards) do
                            card:SetFlags(battle_defs.CARD_FLAGS.STICKY)
                        end
                    end
                    self.owner:RemoveCondition(self.id, self.stacks, self)
                end
            },
        },
    },

    cozy =
    {
        name = "Cozy",
        desc = "Gain 2 maximum resolve and 2 maximum health.",

        OnAdded = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( 2 )
            agent.health:AddStatModifier( "GRAFT_COZY", 2 )
        end,

        OnRemoved = function(self, agent)
            local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( -2 )
            agent.health:AddStatModifier( "GRAFT_COZY", -2 )
        end,
    },

    wrist_rocket = 
    {
        name = "Wrist Rocket",
        desc = "After your first attack each turn, deal 2 additional damage to your target.",

        battle_condition = 
        {
            hidden = true,

            event_handlers = 
            {
                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if attack.attacker == self.owner and attack.card and attack.card:IsAttackCard() and self.active then
                        hit.target:ApplyDamage( 2, self.owner, self )
                        battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        self.active = false
                    end
                end,

                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.active = true
                    end
                end
            },
        },
    },

    hard_knocks = 
    {
        name = "Hard Knocks",
        desc = "After playing 10 cards, your next attack deals 8 bonus damage.",

        display_number = function( self )
            return self.userdata.counter or 0
        end,

        battle_condition = 
        {
            hidden = true,

            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.attacker == self.owner then
                        if self.graft:GetCounter() < 10 then
                            self.graft:IncrementCounter()
                            battle:BroadcastEvent( BATTLE_EVENT.REFRESH_GRAFTS )
                            if self.graft:GetCounter() >= 10 then
                                self.graft:ResetCounter()
                                self.owner:AddCondition("HARDKNOCKS", 1, self)
                                self.owner:RemoveCondition(self.id, self.stacks, self)
                            end
                        else
                            -- Fail-safe to fix saves with the graft stuck at 10
                            self.graft:ResetCounter()
                        end
                    end
                end
            },
        },
    },

    hard_headed =
    {
        name = "Hard Headed",
        desc = "Whenever you play a manipulate card, apply 1 {COMPOSURE} to a random friendly argument.",

        negotiation_modifier = 
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                    if card:IsFlagged( negotiation_defs.CARD_FLAGS.MANIPULATE ) then
                        local options = {}
                        minigame:CollectAlliedTargets( self.negotiator, options )
                        local target = table.arraypick(options)
                        target:DeltaComposure(1, self)
                    end
                end
            },
        },
    },

    conditioning =
    {
        name = "Conditioning",
        desc = "At the start of each battle, gain 3 {POWER}. Whenever you take damage from an enemy, lose 1 {POWER}. You can only lose 3 total {POWER} from this.",

        num_stacks = 3,

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner and not self.applied_power then
                        self.owner:AddCondition("POWER", 3, self)
                        self.applied_power = true
                    end
                end,

                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if hit.target == self.owner and not hit:CheckHitResult( "defended" ) and not hit:CheckHitResult( "evaded" )  and hit.damage and hit.damage > 0 then
                        self.owner:RemoveCondition( "POWER", 1, self )
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                end,
            },
        },
    },

    gluey_acquaintance =
    {
        name = "Gluey Acquaintance",
        desc = "At the start of each battle, apply 1 {gluey_antennae} to all enemies.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if self.owner == fighter then
                        for i,enemy in ipairs(self.battle:GetEnemyTeam():GetFighters()) do
                            enemy:AddCondition("gluey_antennae", self.stacks, self)
                        end
                        self.owner:RemoveCondition(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    con_man =
    {
        name = "Con Man",
        desc = "At the end of every negotiation, recover 3 Resolve.",

        event_handlers = 
        {
            [ "resolve_negotiation" ] = function( self, minigame, repercussions )
                if repercussions.result == minigame.RESULT.WIN then
                    repercussions.loot.resolve = (repercussions.loot.resolve or 0) + 3
                end
            end
        }
    },

    celebrity =
    {
        name = "Celebrity",
        desc = "At the start of each negotiation, gain 3 {negotiation.RENOWN} and 2 {negotiation.INFLUENCE}.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("RENOWN", 3, self)
                        self.negotiator:AddModifier("INFLUENCE", 2, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    envoy_pendant =
    {
        name = "Envoy Pendant",
        desc = "The first Diplomacy card you play each negotiation costs 1 less action.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if card.negotiator == self.negotiator and card:IsFlagged(negotiation_defs.CARD_FLAGS.DIPLOMACY) then
                        cost_acc:AddValue( -1, self )
                    end
                end,

                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.negotiator == self.negotiator and card:IsFlagged(negotiation_defs.CARD_FLAGS.DIPLOMACY) then
                        self.negotiator:RemoveModifier(self, 1, self)
                    end
                end
            },
        },
    },

    local_chatter =
    {
        name = "Local Chatter",
        desc = "At the start of any negotiation against a Pearlie or Heshian, gain 3 {negotiation.RENOWN}.",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local opponent = minigame:GetOpponent()
                        if opponent:GetContentID() == "PEARLIE" or opponent:GetFactionID() == "CULT_OF_HESH" then
                            self.negotiator:AddModifier("RENOWN", 3, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    capitalization =
    {
        name = "Capitalization",
        desc = "Your attacks deal 2 additional damage to targets affected by {battle.IMPAIR}.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_DAMAGE ] = function( self, card, target, dmgt )
                    if card.owner == self.owner and target and target:HasCondition("IMPAIR") then
                        dmgt:AddDamage( 2, 2, self )
                    end
                end,
            },
        },
    },

    in_on_it =
    {
        name = "In On It",
        desc = "At the start of each negotiation, gain 1 {INTERROGATE}.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("INTERROGATE", 1, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    family_favor = 
    {
        name = "Family Favor",
        desc = "At the start of each negotiation, gain {1} {RENOWN}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetDef().renown_amt )
        end,
        renown_amt = 3,
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("RENOWN", self.graft:GetDef().renown_amt, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    covert_supplier =
    {
        name = "Covert Supplier",
        desc = "The first item card you play in each battle costs 1 less.",

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if CheckBits(card.flags, battle_defs.CARD_FLAGS.ITEM) then
                        cost_acc:AddValue(-1, self)
                    end
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    if CheckBits(card.flags, battle_defs.CARD_FLAGS.ITEM) and card.cost > 0 then
                        self.owner:RemoveCondition(self.id, 1, self)
                    end
                end
            },
        },
    },

    anointed =
    {
        name = "Anointed",
        desc = "At the start of every negotiation, gain 1 {negotiation.wrath_of_hesh}.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:AddModifier("wrath_of_hesh", self.stacks, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    informant =
    {
        name = "Informant",
        desc = "At the start of every negotiation, incept 5 {PLANTED_EVIDENCE} for your opponent.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.anti_negotiator:AddModifier("PLANTED_EVIDENCE", self.stacks * 5, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },
    },

    convenience =
    {
        name = "Convenience",
        desc = "{FATIGUED} occurs 1 turn later in both battle and negotiation.",
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local cur = minigame:GetFatiguedTurn()
                        minigame:SetFatiguedTurn(cur + self.stacks)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.negotiator:RemoveModifier(self.id, self.stacks, self)
                    end
                end
            },
        },

        battle_condition =
        {
            hidden = true,
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    local cur = battle:GetFatiguedTurn()
                    battle:SetFatiguedTurn(cur + self.stacks)
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    self.owner:RemoveCondition(self.id)
                end,
            },
        }
    },

    puppeteer =
    {
        name = "Puppeteer",
        desc = "Giving gifts to Heshians costs half as much.",

        event_handlers = 
        {
            ["calc_love_gift"] = function( self, loved_gift_data )
                if loved_gift_data.agent:GetFactionID() == "CULT_OF_HESH" then
                    loved_gift_data.cost = math.floor(loved_gift_data.cost * 0.5)
                end
            end
        },
    },

    fan_mail =
    {
        name = "Fan Mail",
        desc = "At the start of each day receive 100 shills.",

        loc_strings =
        {
            DIALOG = [[* Someone has left an envelope for you. You open it to find a fan letter addressed to both you and Theroux and a wad of {1#money}.]], 
        },

        money_gain = 100,

        event_handlers = 
        {
            ["morning_mail"] = function( self, cxt )
                if (self.userdata.last_mail or -1) < cxt.now then
                    cxt:Dialog( self:GetDef():GetLocalizedString( "DIALOG" ), self:GetDef().money_gain )
                    cxt.enc:GainMoney( self:GetDef().money_gain )
                    self.userdata.last_mail = cxt.now
                end
            end,
        },
    },
}

---------------------------------------------------------------------------------------------

for id, graft in pairs( SOCIAL_GRAFTS ) do
    graft.is_boon = true
    Content.AddSocialGraft( id, graft )
end
