local battle_defs = require "battle/battle_defs"

local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

--------------------------------------------------------------------

local GRAFT_IDS = {}
local ITEM_IDS = {}
local CONDITION_LOOKUPS = {}

local BATTLE_GRAFTS =
{
    afterburner_gloves =
    {
        name = "Afterburner Gloves",
        desc = "All your attacks apply 1 {BURN}.",
        
        rarity = CARD_RARITY.UNIQUE,
        flavour = "'The heat is on with these stylish and dangerous gloves, be sure to make sure that your district's incendiary regulation allows for the handling and ownership of these items.'",

        kashio_condition = "kashio_afterburner_gloves",
        auction_tier = 1,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_afterburner_gloves.tex"),

        battle_condition =
        {
            hidden = false,
            event_handlers =
            {
                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if attack.card:IsAttackCard() and attack.card.id ~= "counter_attack" and attack.attacker == self.owner then
                        if not hit.evaded and not hit.defended then
                            hit.target:AddCondition( "BURN", 1, attack.card )
                        end
                    end
                end,
            },
        },
    },

    blast_chamber = 
    {
        name = "Blast Chamber",
        flavour = "'Lot 42 is a very strong and very stable Spark battery! No leakage or exploding here folks, built by master craftsmen with the finest of materials.'",
        desc = "Gain 1 action at the start of your turn.",

        kashio_condition = "kashio_blast_chamber",
        auction_tier = 3,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_blast_chamber.tex"),
        rarity = CARD_RARITY.UNIQUE,

        battle_condition = 
        {
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        if self.owner.player_controlled then
                            self.battle:ModifyActionCount( self.stacks )
                            self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                    end
                end,
            },
        },
    },

    macro_rasp = 
    {
        name = "Macro Rasp",
        flavour = "'Straight from the highest ranking bloodstriker, it's the infamous Macro Rasp!'",
        desc = "Whenever you apply {BLEED}, deal 1 damage to the target.",

        kashio_condition = "kashio_macro_rasp",

        rarity = CARD_RARITY.UNIQUE,
        auction_tier = 3,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_macro_rasp.tex"),
        battle_condition = 
        {
            event_handlers = 
            {
                [ BATTLE_EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
                    if source and source.owner == self.owner and condition.id == "BLEED" then
                        fighter:ApplyDamage( self.stacks, self.owner, self )
                    end
                end
            },
        },
    },

    hammerthyst_ring = 
    {
        name = "Hammerthyst Ring",
        flavour = "'One of the smallest spark generators is set inside this obsidian casing. Fits on the finger but makes a huge impression!'",
        desc = "At the end of your turn, gain 2 {DEFEND} for every card in your hand.",

        kashio_condition = "kashio_hammerthyst_ring",

        rarity = CARD_RARITY.UNIQUE,
        auction_tier = 1,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_hammerthyst_ring.tex"),
        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.owner:AddCondition("DEFEND", battle:GetHandDeck():CountCards() * 2 )
                end
            },
        },
    },

    sage_conversion =
    {
        name = "Sage Conversion",
        flavour = "'Linked directly into the brain stem this is an exciting addition to anyone's cybernetic load out!'",
        desc = "At the start of battle, gain health equal to your current resolve.",

        kashio_condition = "kashio_sage_conversion",
        auction_tier = 2,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_sage_conversion.tex"),
        rarity = CARD_RARITY.UNIQUE,

        battle_condition = 
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    local current_resolve = TheGame:GetGameState():GetCaravan():GetResolve()
                    self.owner:DeltaMaxHealth( current_resolve )
                    self.owner:DeltaHealth( current_resolve )
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    dermal_cast = 
    {
        name = "Dermal Cast",
        flavour = "'This rare graft rejuvenates the body and mind! Well the body for sure...'",
        desc = "At the start of your turn, {HEAL} 2 health.",
        auction_tier = 2,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_dermal_cast.tex"),
        kashio_condition = "kashio_dermal_cast",

        rarity = CARD_RARITY.UNIQUE,

        battle_condition = 
        {
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    self.owner:HealHealth(2)
                end
            },
        },
    },
}

---------------------------------------------------------------------------------------------

for i, id, graft in sorted_pairs( BATTLE_GRAFTS ) do

    table.insert(GRAFT_IDS, id)
    CONDITION_LOOKUPS[id] = graft.kashio_condition

    graft.card_defs = battle_defs
    graft.series = "GENERAL"
    graft.type = GRAFT_TYPE.COMBAT
    if graft.battle_condition and graft.battle_condition.hidden == nil then
        graft.battle_condition.hidden = true
    end

    Content.AddGraft( id, graft )
end

local ITEMS = 
{
    hologram_projection_belt =
    {
        name = "Hologram Projection Belt",
        anim = "taunt",
        flavour = "'The technology of yesterday today! Great for entertaining!'",
        desc = "Play your next card twice.",

        kashio_condition = "kashio_hologram_belt",

        cost = 0,

        flags = CARD_FLAGS.SKILL,
        rarity = CARD_RARITY.UNIQUE,
        target_type = TARGET_TYPE.SELF,
        auction_tier = 2,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_hologram_belt.tex"),
        max_charges = 2,

        OnPostResolve = function( self, battle, attack)
            local modifier = self.owner:AddCondition( "hologram_projection_belt", 1, self )
            modifier.skip_card = attack.card
        end,

        condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if attack.card ~= self.skip_card and attack.attacker == self.owner then
                        self.owner:RemoveCondition( "hologram_projection_belt", 1 )
                        battle:PlayCard( attack.card )
                    end
                end
            },
        },
    },

    suitcase_grenades =
    {
        name = "Suitcase Grenades",
        anim = "throw",
        flavour = "'Easy to bring along to any altercation, these beauties glow like the sun while melting through almost anything!'",
        desc = "Hits all enemy targets.",
        anims = { "anim/weapon_grenade_suitcase.zip"},
        hit_tags = {"spark"},

        kashio_condition = "kashio_suitcase_grenades",
        
        cost = 1,

        flags = CARD_FLAGS.RANGED,
        rarity = CARD_RARITY.UNIQUE,

        max_charges = 3,
        auction_tier = 3,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_suitcase_grenades.tex"),
        target_mod = TARGET_MOD.TEAM,

        min_damage = 6,
        max_damage = 8,
    },

    eh1076_discharge_pistol =
    {
        name = "EH1076 Discharge Pistol",
        anim = "shoot",
        anims = { "anim/weapon_blaster_discharge.zip"},
        flavour = "'A heavily modified EH1076 sporting incendiary augments and a steep clipping range. One of a lost pair.'",
        hit_tags = { "spark" },
        
        kashio_condition = "kashio_pistol",
        auction_tier = 2,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_discharge_pistol.tex"),
        cost = 1,

        flags = CARD_FLAGS.RANGED | CARD_FLAGS.PIERCING,
        rarity = CARD_RARITY.UNIQUE,

        max_charges = 2,

        min_damage = 6,
        max_damage = 8,

        features = 
        {
            WOUND = 2,
        },
    },

    rentorian_force_glaive =
    {
        name = "Rentorian Force Glaive",
        anim = "throw",
        flavour = "'Sal wasn't too sure where the release was, or how to hold the Heshing thing at all.'",
        desc = "Apply 8 {BLEED}.",
        anims = { "anim/weapon_grenade_force_glaive.zip"},
        hit_tags = { "spark" },

        kashio_condition = "kashio_glaive",
        auction_tier = 1,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_force_glaive.tex"),
        cost = 1,

        flags = CARD_FLAGS.RANGED | CARD_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        min_damage = 8,
        max_damage = 8,

        OnPostResolve = function( self, battle, attack )
            attack:AddCondition( "BLEED", 8, self )
        end,
    },

    support_beacon =
    {
        name = "Support Beacon",
        desc = "Summon a Combat Drone to fight for you.",

        kashio_condition = "kashio_robot_backup",
        flavour = "'DRONE NUMBER 0056 ACTIVATED. LUXURY PACKAGE ENGAGED.'",
        auction_tier = 3,
        auction_img = engine.asset.Texture( "negotiation/modifiers/auction_support_beacon.tex"),
        cost = 0,

        flags = CARD_FLAGS.SKILL | CARD_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        target_type = TARGET_TYPE.SELF,

        CanPlayCard = function( self, battle, target )
            return self.owner:GetTeam():NumActiveFighters() < self.owner:GetTeam():GetMaxFighters(), CARD_PLAY_REASONS.TEAM_FULL
        end,

        OnPostResolve = function( self, battle, attack )
            local agent = Agent("COMBAT_DRONE")
            local fighter = Fighter.CreateFromAgent( agent, 1 )
            self.owner:GetTeam():AddFighter( fighter )
            self.owner:GetTeam():ActivateNewFighters()
        end
    },
}


for i, id, data in sorted_pairs( ITEMS ) do

    table.insert(ITEM_IDS, id)
    CONDITION_LOOKUPS[id] = data.kashio_condition

    data.item_tags = (data.item_tags or 0) | ITEM_TAGS.COMBAT
    data.flags = (data.flags or 0) | CARD_FLAGS.ITEM 
    assert(data.rarity, loc.format("Rarity missing on {1}", data.name))
    data.series = CARD_SERIES.GENERAL

    Content.AddBattleCard( id, data )
end

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local NEGOTIATION_FLAGS = negotiation_defs.CARD_FLAGS

local CARDS =
{
    bid = 
    {
        name = "Bid",
        desc = "{IMPROVISE} a special card to help you in the auction.",
        flavour = "'Yes, the young lady in the back.'",

        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME | NEGOTIATION_FLAGS.STICKY,
        rarity = CARD_RARITY.UNIQUE,
        pool_cards = {"grade", "hammer_price", "bait_and_switch", "compete", "lean", "appraisal" },

        OnPostResolve = function( self, minigame, targets )
            local cards = {
                Negotiation.Card( "grade", self.owner ),
                Negotiation.Card( "hammer_price", self.owner ),
                Negotiation.Card( "bait_and_switch", self.owner ),
                Negotiation.Card( "compete", self.owner ),
                Negotiation.Card( "lean", self.owner ),
                Negotiation.Card( "appraisal", self.owner ),

            }

            cards = table.multipick( cards, 3 )
            minigame:ImproviseCards( cards, 1, nil, nil, nil, self )
        end,
    },

    grade =
    {
        name = "Grade",
        desc = "Steal all composure from a target.",
        flavour = "'He came all the way from Supplicant City!'",

        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,
        target_enemy = TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY | TARGET_FLAG.CORE,

        money_cost = 100,

        event_handlers =
        {
            [ EVENT.PRE_RESOLVE ] = function( self, minigame, card, targets )
                if card == self then
                    for i, target in ipairs( targets ) do
                        local composure = target:GetComposure()
                        target:DeltaComposure( -composure, self )
                        self.negotiator:DeltaComposure( composure, self )
                    end
                end
            end
        }
    },

    hammer_price = 
    {
        name = "Hammer Price",
        flavour = "'Going once, twice, SOLD!'",
        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        money_cost = 250,

        min_persuasion = 15,
        max_persuasion = 15,
    },

    bait_and_switch = 
    {
        name = "Bait and Switch",
        flavour = "'If you look at your catalog on page... wait, something is wrong.'",
        desc = "Remove a random bidder argument.",
        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,
        target_mod = TARGET_MOD.TEAM,

        money_cost = 100,

        OnPostResolve = function( self, minigame, targets )
            for i,target in minigame:GetOpponentNegotiator():Modifiers() do
                if target.id == "BIDDER" then
                    target.negotiator:RemoveModifier(target)
                    break
                end
            end
        end
    },

    compete = 
    {
        name = "Compete",
        desc = "Reflect all incoming damage until your next turn.",
        flavour = "'This is going to be a close one.'",
        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        money_cost = 250,

        OnPostResolve = function( self, minigame, targets )
            self.negotiator:AddModifier("compete")
        end,

        modifier =
        {
            modifier_type = MODIFIER_TYPE.ARGUMENT,
            event_handlers =
            {
                [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                    if target.negotiator == self.negotiator then
                        if is_instance( source, Negotiation.Modifier ) then
                            source:AttackResolve(damage, self)
                        else
                            source.negotiator:AttackResolve(damage, self)
                        end
                        target.composure = target.composure + damage
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:RemoveModifier(self.id, self.stacks)
                end
            },
        },
    },

    lean =
    {
        name = "Lean",
        desc = "Target argument loses 8 resolve.",
        flavour = "'I came prepared but this isn't going to be a cheap endeavour.'",

        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        money_cost = 50,

        target_self = TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY,
        target_enemy = TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY | TARGET_FLAG.CORE,

        CanTarget = function( self, target )
            if is_instance( target, Negotiation.Modifier ) and target:GetResolve() == nil then
                return false, CARD_PLAY_REASONS.NO_TARGET_RESOLVE
            end
            return true
        end,

        OnPostResolve = function( self, minigame, targets )
            for i, target in ipairs( targets ) do
                local resolve = target:GetResolve()
                resolve = math.min( 8, resolve )
                target:ModifyResolve( -resolve, self )
            end
        end,
    },

    appraisal =
    {
        name = "Appraisal",
        flavour = "'Yeah he's been doing this for a while, one of the more honest ones.'",
        desc = "Double target bidder's damage this turn.",

        cost = 0,

        flags = NEGOTIATION_FLAGS.MANIPULATE | NEGOTIATION_FLAGS.CONSUME,
        rarity = CARD_RARITY.UNIQUE,

        money_cost = 100,
        target_enemy = TARGET_FLAG.ARGUMENT,

        CanPlayCard = function( self, card, engine, target )
            if card == self then
                return target == nil or target.id == "BIDDER"
            end
        end,

        modifier =
        {
            hidden = true,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_MULTIPLIER,
            },

            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if self.bidder and table.contains(self.bidder, source) then
                        local min = math.round( persuasion.min_persuasion * 2 )
                        local max = math.round( persuasion.max_persuasion * 2 )
                        persuasion:ModifyPersuasion( min, max, self )
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:RemoveModifier( self )
                end,
            }
        },

        OnPostResolve = function( self, minigame, targets )
            local modifier = self.negotiator:AddModifier( "appraisal", 1 )
            modifier.bidder = targets
        end,
    },
}

for i, id, carddef in sorted_pairs( CARDS ) do
    carddef.item_tags = (carddef.item_tags or 0) | ITEM_TAGS.NEGOTIATION
    carddef.flags = (carddef.flags or 0) | NEGOTIATION_FLAGS.ITEM
    carddef.rarity = carddef.rarity or CARD_RARITY.COMMON
    carddef.hide_in_cardex = true
    carddef.series = "SAL"
    Content.AddNegotiationCard( id, carddef )
end

local function PickAuctionLots(num_grafts, num_items) 
    local ret = {grafts = {}, items = {}}

    local grafts = table.shuffle(shallowcopy(GRAFT_IDS))
    local items = table.shuffle(shallowcopy(ITEM_IDS))
    
    for k = 1, math.min(num_grafts, #grafts) do 
        table.insert( ret.grafts, grafts[k]) 
    end
    
    for k = 1, math.min(num_items, #items) do 
        table.insert( ret.items, items[k]) 
    end

    return ret
end

return {
    CONDITION_LOOKUPS = CONDITION_LOOKUPS,
    PickAuctionLots = PickAuctionLots,
}