local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local CARD_FLAGS = negotiation_defs.CARD_FLAGS

local NEGOTIATION_GRAFTS =
{
    -- COMMON
    myelin_buffer =
    {
        name = "Myelin Buffer",
        desc = "Gain 6 max resolve.",
        flavour = "'It just keeps the membrane a little cooler and more supple.'",

        rarity = CARD_RARITY.COMMON,
        resolve_bonus = 6,

        OnAdded = function(self, agent)
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( self:GetDef().resolve_bonus )
        end,
        
        OnUpgraded = function(self, agent)
            TheGame:GetGameState():GetCaravan():DeltaMaxResolve( self:GetDef().resolve_bonus - self:GetDef().base_def.resolve_bonus )
        end,

        OnRemoved = function( self, agent )
            local caravan = TheGame:GetGameState():GetCaravan()
            caravan:DeltaMaxResolve( -self:GetDef().resolve_bonus )
            local resolve, max_resolve = caravan:GetResolve()
            if resolve > max_resolve then
                caravan:SetResolvePercent( 1 )
            end
        end,
    },

    myelin_buffer_plus =
    {
        name = "Boosted Myelin Buffer",
        desc = "Gain <#UPGRADE>10</> max resolve.",
        resolve_bonus = 10,
    },

    pocket_bomb =
    {
        name = "Pocket Bomb",
        desc = "Whenever you shuffle your deck, deal {1} damage to a random opponent argument.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().damage)
        end,
        flavour = "'No this is the new model, you have nothing to worry about.'",

        rarity = CARD_RARITY.COMMON,

        damage = 3,
        negotiation_modifier = 
        {
            target_enemy = TARGET_ANY_RESOLVE,
            min_persuasion = 2,
            max_persuasion = 2,
            event_handlers =
            {
                [ EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
                    self.min_persuasion = self.graft:GetDef().damage
                    self.max_persuasion = self.graft:GetDef().damage
                    self:ApplyPersuasion()
                end
            }            
        },
    },

    pocket_bomb_plus =
    {
        name = "Boosted Pocket Bomb",
        desc = "Whenever you shuffle your deck, deal <#UPGRADE>{1}</> damage to a random opponent argument.",
        damage = 4,
    },

    lightning_rod =
    {
        name = "Lightning Rod",
        desc = "At the start of each negotiation, insert a {card.conductor} card into your hand that redirects opponent intents.",
        flavour = "'Really helps out your personality, they only focus on one of your many flaws.'",

        rarity = CARD_RARITY.COMMON,
        card_id = "conductor",

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local card = Negotiation.Card(self.graft:GetDef().card_id, self.owner)
                    minigame:DealCard(card, minigame:GetHandDeck())
                    minigame:GetPlayerNegotiator():RemoveModifier(self.id, self.stacks, self)
                end
            },
        },
    },

    lightning_rod_plus =
    {
        name = "Boosted Lightning Rod",
        desc = "At the start of each negotiation, insert a <#UPGRADE>{card.conductor_upgraded}</> card into your hand that redirects enemy intents.",
        card_id = "conductor_upgraded",
    },

    -- UNCOMMON
    locus_breaker =
    {
        name = "Locus Breaker",
        desc = "Whenever you draw {1} or more cards after your initial hand in a single turn, gain 1 {BAIT}.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetDef().draw_amt)
        end,
        flavour = "'It’s a small change but you get to keep your wits.'",

        draw_amt = 5,

        display_number = function( self )
            return self:GetCounter() or 0
        end,

        rarity = CARD_RARITY.UNCOMMON,
        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                    if not start_of_turn and not self.used then
                        self.graft:IncrementCounter()
                        if self.graft:GetCounter() >= self.graft:GetDef().draw_amt then
                            minigame:GetPlayerNegotiator():AddModifier("BAIT", 1, self)
                            self.used = true
                        end
                        minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.used = false
                    self.graft:ResetCounter()
                    self:NotifyTriggered()
                end,
            },
        },
    },

    locus_breaker_plus =
    {
        name = "Boosted Locus Breaker",
        desc = "Whenever you draw <#UPGRADE>{1}</> or more cards after your initial hand in a single turn, gain 1 {BAIT}.",
        draw_amt = 4,
    },

    psimulacrum =
    {
        name = "Psimulacrum",
        desc = "Duplicate the first argument you create every negotiation.",
        flavour = "'I’m not too sure if we should be using this metal but it should be fine.'",

        rarity = CARD_RARITY.UNCOMMON,

        uses = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                    if source and source.negotiator and source.negotiator == self.negotiator and source ~= self and
                        not (is_instance(source, Negotiation.Card) and CheckBits(source.flags, CARD_FLAGS.BYSTANDER)) and
                        modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN and modifier.modifier_type ~= MODIFIER_TYPE.INCEPTION then
                        --
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.uses = (self.uses or 0) + 1
                        if self.uses >= self.graft:GetDef().uses then
                            self.negotiator:RemoveModifier(self.id, 1)
                        end

                        modifier.negotiator:CreateModifier(modifier.id, modifier.stacks, self)
                    end
                end
            },
        },
    },

    psimulacrum_plus =
    {
        name = "Enduring Psimulacrum",
        desc = "Duplicate the first <#UPGRADE>2</> arguments you create every negotiation.",
        uses = 2,
    },

    shenanigans =
    {
        name = "Shenanigans",
        desc = "{COMPOSURE} cards apply 1 additional {COMPOSURE}.",
        flavour = "'I've had it!'",

        rarity = CARD_RARITY.UNCOMMON,

        mod_xp = 3,

        bonus_comp = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.CALC_DELTA_COMPOSURE ] = function( self, composure_acc, target, source )
                    if source and source.negotiator == self.negotiator and is_instance(source, Negotiation.Card) and composure_acc.value > 0 then
                        composure_acc:AddValue(self.graft:GetDef().bonus_comp, self)
                    end
                end,
            },
        },
    },

    shenanigans_plus =
    {
        name = "Boosted Shenanigans",
        desc = "{COMPOSURE} cards apply <#UPGRADE>2</> additional {COMPOSURE}.",
        bonus_comp = 2,
    },

    high_pitched_whine =
    {
        name = "High Pitched Whine",
        desc = "At the start of each negotiation, gain 1 {high_pitched_whine}.",
        flavour = "'Eventually it’s the only thing they can focus on.'",

        rarity = CARD_RARITY.UNCOMMON,

        negotiation_modifier = 
        {
            hidden = false,
            name = "Whine",
            desc = "At the end of {1}'s turn, this argument gains <#HILITE>{2}</> resolve.",
            desc_fn = function( self, fmt_str, minigame )
                return loc.format( fmt_str, self:GetOwnerName(), self.stacks)
            end,

            modifier_type = MODIFIER_TYPE.ARGUMENT,

            max_resolve = 2,

            event_handlers =
            {   
                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    self:ModifyResolve(self.stacks, self)
                end,
            },
        },
    },

    high_pitched_whine_plus =
    {
        name = "Boosted High Pitched Whine",
        desc = "At the start of each negotiation, gain 2 {high_pitched_whine}.",
        num_stacks = 2,
    },

    psymitar =
    {
        name = "Psymitar",
        desc = "All cards gain 1 minimum damage.",
        flavour = "'For that razor sharp wit.'",

        rarity = CARD_RARITY.UNCOMMON,

        bonus_damage = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if source.owner == self.owner then
                        if is_instance( source, Negotiation.Card ) and source.min_persuasion ~= nil then
                            persuasion:AddPersuasion( self.graft:GetDef().bonus_damage or 1, 0, self )
                        end
                    end
                end,
            },
        },
    },

    psymitar_plus =
    {
        name = "Sharpened Psymitar",
        desc = "All cards gain <#UPGRADE>2</> minimum damage.",

        bonus_damage = 2,
        mod_xp = 2,
    },

    absorber =
    {
        name = "Absorber",
        desc = "Whenever you play 4 Diplomacy cards in a row, apply 1 {COMPOSURE} to all friendly arguments.",
        flavour = "'Consistency is key. The same pathways used over and over help this thing really sing.'",

        rarity = CARD_RARITY.UNCOMMON,

        streak_count = 4,

        display_number = function( self )
            return self:GetCounter() or 0
        end,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.negotiator == self.negotiator then
                        if CheckBits( card.flags, CARD_FLAGS.DIPLOMACY ) then
                            self.graft:IncrementCounter()
                            if self.graft:GetCounter() >= self.graft:GetDef().streak_count then
                                minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                                for i, modifier in self.negotiator:Modifiers() do
                                    modifier:DeltaComposure(1, self)
                                end
                                self.graft:ResetCounter()
                            end
                            self:NotifyTriggered()

                        elseif (self.graft:GetCounter() or 0) > 0 then
                            self.graft:ResetCounter()
                            self:NotifyTriggered()
                        end
                    end
                end,
            },
        },
    },

    absorber_plus =
    {
        name = "Swift Absorber",
        desc = "Whenever you play <#UPGRADE>3</> Diplomacy cards in a row, apply 1 {COMPOSURE} to all friendly arguments.",
        streak_count = 3,
    },

    brain_barrage =
    {
        name = "Brain Barrage",
        desc = "Whenever you play 3 Manipulate cards in a row, deal 4 damage to a random opponent argument.",
        flavour = "'This here will make quick work out of anyone that steps up.'",

        rarity = CARD_RARITY.UNCOMMON,

        streak_count = 3,
        min_persuasion = 4,
        max_persuasion = 4,

        display_number = function( self )
            return self:GetCounter() or 0
        end,

        negotiation_modifier =
        {
            hidden = true,
            target_enemy = TARGET_ANY_RESOLVE,
            event_handlers = 
            {
                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.negotiator == self.negotiator then
                        if CheckBits( card.flags, CARD_FLAGS.MANIPULATE ) then
                            self.graft:IncrementCounter()
                            if self.graft:GetCounter() >= self.graft:GetDef().streak_count then
                                minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                                self.min_persuasion = self.graft:GetDef().min_persuasion
                                self.max_persuasion = self.graft:GetDef().max_persuasion
                                self:ApplyPersuasion()
                                self.graft:ResetCounter()
                            end
                            self:NotifyTriggered()

                        elseif (self.graft:GetCounter() or 0) > 0 then
                            self.graft:ResetCounter()
                            self:NotifyTriggered()
                        end
                    end
                end,
            },
        },
    },

    brain_barrage_plus = 
    {
        name = "Boosted Brain Barrage",
        desc = "Whenever you play 3 Manipulate cards in a row, deal <#UPGRADE>6</> damage to a random opponent argument.",
        min_persuasion = 6,
        max_persuasion = 6,
    },

    hangunder =
    {
        name = "Hangunder",
        desc = "Whenever you play a {drunk_player} card, restore 3 resolve.",
        flavour = "'Here, take this. Don’t ask what’s in it.'",

        rarity = CARD_RARITY.UNCOMMON,

        resolve_amt = 5,
        cost_reduction = 0,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.POST_RESOLVE ] = function ( self, minigame, card )
                    if card.negotiator == self.negotiator and card.id == "drunk_player" then
                        self.negotiator:RestoreResolve(self.graft:GetDef().resolve_amt, self)
                        minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,

                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.graft:GetDef().cost_reduction > 0 and card.id == "drunk_player" then
                        cost_acc:AddValue(-self.graft:GetDef().cost_reduction, self)
                    end
                end
            },
        },
    },

    hangunder_plus =
    {
        name = "Fancy Hangunder",
        desc = "Whenever you play a {drunk_player} card, restore 3 resolve.\n<#UPGRADE>{drunk_player} cards cost 1 less.</>",
        cost_reduction = 1,
    },

    chemical_regulator =
    {
        name = "Chemical Regulator",
        desc = "When installed, upgrade 3 random negotiation cards.",
        flavour = "'I’m not saying that the initial readout is that bad, but everyone could probably use help with this.'",

        rarity = CARD_RARITY.UNCOMMON,

        OnAdded = function( self, agent )
            local cards = agent.negotiator:GetCards()
            local candidates = {}
            for i=1,#cards do
                if cards[i]:HasUpgrades() and not cards[i]:UpgradeReady() and not cards[i]:IsHatched() then
                    table.insert(candidates, cards[i])
                end
            end
            for i=1, math.min( #candidates, 3 ) do
                local pick = math.random(#candidates)
                candidates[pick]:AddXP(candidates[pick].max_xp)
                local popup = Screen.UpgradeCardPopup( candidates[pick], agent )
                TheGame:FE():PushScreen( popup )
                table.remove(candidates, pick)         
            end
        end
    },

    chemical_regulator_plus =
    {
        name = "Boosted Chemical Regulator",
        desc = "When upgraded, upgrade 3 additional negotiation cards.",

        OnUpgraded = function( self, agent )
            self:GetDef().OnAdded( self, agent )
        end
    },

    jugularnaut =
    {
        name = "Jugularnaut",
        desc = "Whenever your damage spills over, apply 2 {COMPOSURE} to a random argument.",
        flavour = "'Fits just under the jawline, has profound effects on one's delivery.'",

        rarity = CARD_RARITY.UNCOMMON,
        comp_amt = 2,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                    if modifier.owner ~= self.owner then
                        local targets = self.engine:CollectAlliedTargets(self.negotiator)
                        if targets and #targets > 0 then
                            targets[math.random(#targets)]:DeltaComposure(self.graft:GetDef().comp_amt, self)
                        end
                    end
                end,
            }    
        }
    },

    jugularnaut_plus =
    {
        name = "Boosted Jugularnaut",
        desc = "Whenever your damage spills over, apply <#UPGRADE>3</> {COMPOSURE} to a random argument.",
        comp_amt = 3,
    },

    -- RARE
    memorization_implant =
    {
        name = "Memorization Implant",
        desc = "At the start of each negotiation, choose a card in your hand and duplicate it.",
        flavour = "'Strengthens those opinions and takes away the stress of remembering your stronger points.'",

        rarity = CARD_RARITY.RARE,

        card_count = 1,
        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local txt = loc.format( LOC"CARD_ENGINE.CHOOSE_MAX_CARDS_DUPLICATE", self.graft:GetDef().card_count )
                    local cards = minigame:ChooseCardsFromTable( minigame:GetHandDeck(), 0, self.graft:GetDef().card_count, nil, txt)
                    if cards then
                        for i,card in ipairs(cards) do
                            local copy = card:Duplicate()
                            copy.show_dealt = false
                            minigame:DealCard( copy, minigame:GetHandDeck())
                        end
                    end
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier(self.id, 1)
                end
            },
        },
    },

    memorization_implant_plus =
    {
        name = "Boosted Memorization Implant",
        desc = "At the start of each negotiation, choose <#UPGRADE>up to 2</> cards in your hand and duplicate them.",
        card_count = 2,
    },

    -- BOSS
    three_fingers = 
    {
        name = "Three Fingers",
        desc = "Gain 1 action at the start of your turn. When installed, 2 random people will hate you.",
        flavour = "'Gross.'",

        rarity = CARD_RARITY.BOSS,
        card_draw = 0,

        OnAdded = function(self, agent)
            local candidates = ObtainWorkTable()

            for k,agent in pairs(TheGame:GetGameState().agents) do
                if not AgentUtil.HasPlotArmour(agent) and
                    not agent:IsInPlayerParty() and
                    agent:GetRelationship() > RELATIONSHIP.HATED and
                    agent:IsSentient() then
                    table.insert(candidates, agent)
                end
            end

            for i = #candidates + 1, 2 do
                if TheGame:GetGameState() then
                    local new_agent = TheGame:GetGameState():AddSkinnedAgent()
                    table.insert( candidates, new_agent )
                end
            end

            table.shuffle( candidates )

            for i=1, math.min( #candidates, 2 ) do
                local chosen = table.remove( candidates )
                chosen:OpinionEvent( OPINION.THREE_FINGERS )
            end
            
            ReleaseWorkTable(candidates)
        end,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
        },
    },

    three_fingers_plus =
    {
        name = "Visionary Three Fingers",
        desc = "Gain 1 action <#UPGRADE>and draw a card</> at the start of your turn. When installed, 2 random people will hate you.",
        card_draw = 1,
    },

    roller_drum =
    {
        name = "Roller Drum",
        desc = "Gain 1 action at the start of your turn. All Hostility cards deal 50% less damage.",
        flavour = "'It’s very zen, so you’ll be too calm to come off all that fierce.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_MULTIPLIER,
            },

            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if source.owner == self.owner then
                        if is_instance( source, Negotiation.Card ) and CheckBits( source.flags, CARD_FLAGS.HOSTILE ) then
                            persuasion:AddPersuasion( - math.floor(persuasion.min_persuasion * .5), - math.floor(persuasion.max_persuasion * .5), self )
                        end
                    end
                end,
                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
            
        },
    },

    roller_drum_plus =
    {
        name = "Visionary Roller Drum",
        desc = "Gain 1 action <#UPGRADE>and draw a card</> at the start of your turn. All Hostility cards deal 50% less damage.",
        card_draw = 1,
    },

    distributed_processing = 
    {
        name = "Distributed Processing",
        desc = "Whenever you draw a Status card, draw a card.",
        flavour = "'All your baggage, this thing will take care of it. Or at least leave it organized.'",

        expend_chance = 0,
        rarity = CARD_RARITY.UNCOMMON,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if CheckBits( card.flags, CARD_FLAGS.STATUS ) and target_deck == self.engine:GetHandDeck() then
                        self.engine:DrawCards(1)
                        local expend_chance = self.graft:GetDef().expend_chance
                        if math.random() < expend_chance then
                            self.engine:PushPostHandler( function()
                                card:NotifyTriggeredPre()
                                self.engine:ExpendCard(card)
                                card:NotifyTriggeredPost()
                            end )
                        end
                        self:NotifyTriggered()
                    end
                end,
            },
        },
    },

    distributed_processing_plus = 
    {
        name = "Boosted Distributed Processing",
        desc = "Whenever you draw a Status card, draw a card.\n<#UPGRADE>50% chance to {EXPEND} the Status card.</>",
        expend_chance = .5,
    },

    coactive_orbslug = 
    {
        name = "Coactive Orbslug",
        desc = "Whenever you gain {INFLUENCE}, deal 2 damage to a random opponent argument.",
        flavour = "'Sure about this? If I ever have to remove it you’re probably going to forget a few birthdays.'",

        rarity = CARD_RARITY.RARE,

        damage = 2,

        negotiation_modifier =
        {
            hidden = true,
            
            target_enemy = TARGET_ANY_RESOLVE,
            min_persuasion = 2,
            max_persuasion = 2,

            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.id == "INFLUENCE" then
                        self.min_persuasion = self.graft:GetDef().damage
                        self.max_persuasion = self.graft:GetDef().damage
                        self:ApplyPersuasion()
                    end
                end,

                [ EVENT.MODIFIER_CHANGED ] = function( self, modifier, delta, clone )
                    if delta and delta > 0 and modifier.negotiator == self.negotiator and modifier.id == "INFLUENCE" then
                        self.min_persuasion = self.graft:GetDef().damage
                        self.max_persuasion = self.graft:GetDef().damage
                        self:ApplyPersuasion()
                    end
                end,            
            },
        },
    },

    coactive_orbslug_plus =
    {
        name = "Boosted Coactive Orbslug",
        desc = "Whenever you gain {INFLUENCE}, deal <#UPGRADE>3</> damage to a random opponent argument.",
        damage = 3,
    },

    contractor = 
    {
        name = "Contractor",
        desc = "Whenever you gain {DOMINANCE}, deal 1 damage to all opponent arguments.",
        flavour = "'You’ll feel a little pinch.'",

        damage = 1,
        rarity = CARD_RARITY.RARE,

        negotiation_modifier =
        {
            hidden = true,

            min_persuasion = 1,
            max_persuasion = 1,
            target_enemy = TARGET_ANY_RESOLVE,
            target_mod = TARGET_MOD.TEAM,

            event_handlers =
            {
               [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.id == "DOMINANCE" then
                        self.min_persuasion = self.graft:GetDef().damage
                        self.max_persuasion = self.graft:GetDef().damage
                        self:ApplyPersuasion()
                    end
                end,

                [ EVENT.MODIFIER_CHANGED ] = function( self, modifier, delta, clone )
                    if delta and delta > 0 and modifier.negotiator == self.negotiator and modifier.id == "DOMINANCE" then
                        self.min_persuasion = self.graft:GetDef().damage
                        self.max_persuasion = self.graft:GetDef().damage
                        self:ApplyPersuasion()
                    end
                end,
            },
        },
    },

    contractor_plus =
    {
        name = "Boosted Contractor",
        desc = "Whenever you gain {DOMINANCE}, deal <#UPGRADE>2</> damage to all targets.",
        damage = 2,
    },

    interior_drill = 
    {
        name = "Interior Drill",
        desc = "At the start of your turn, reduce the cost of a random card in your hand by 1 until played.",
        flavour = "'You probably don't need what it’s getting rid of. Probably.'",

        cost_reduction = 1,
        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            CanReduceCost = function( card )
                return card.engine:CalculateActionCost( card ) > 0 and not card:IsFlagged( CARD_FLAGS.UNPLAYABLE )
            end,

            event_handlers = 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local hand = minigame:GetHandDeck()
                    local card = hand:PickCardIf( self.CanReduceCost )
                    if card then
                        if self.chosen_cards == nil then
                            self.chosen_cards = {}
                        end
                        table.insert(self.chosen_cards, card)
                        minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,

                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.chosen_cards and table.contains(self.chosen_cards, card) then
                        cost_acc:AddValue(-self.graft:GetDef().cost_reduction, self)
                    end
                end,

                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if self.chosen_cards then
                        table.arrayremove( self.chosen_cards, card )
                    end
                end,

            },
        },
    },

    interior_drill_plus =
    {
        name = "Boosted Interior Drill",
        desc = "At the start of your turn, reduce the cost of a random card in your hand by <#UPGRADE>2</> until played.",
        cost_reduction = 2,
    },

    jarackle_janglers =
    {
        name = "Jarackle Janglers",
        desc = "Gain 1 action at the start of your turn. Your core argument takes 2 damage at the end of your turn.",
        flavour = "'Those guys do NOT care if you hear them coming. They kind of live for that.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_priorities =
            {
                [ EVENT.END_PLAYER_TURN ] = 11,
            },
            event_handlers =
            {
                [ EVENT.END_PLAYER_TURN ] = function ( self, minigame )
                    self.negotiator:AttackResolve( 2, self )
                end,
                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
            
        },
    },

    jarackle_janglers_plus =
    {
        name = "Visionary Jarackle Janglers",
        desc = "Gain 1 action<#UPGRADE> and draw a card</> at the start of your turn. Your core argument takes 2 damage at the end of your turn.",
        card_draw = 1,
    },

    brain_harp =
    {
        name = "Brain Harp",
        desc = "Gain 1 action at the start of your turn, damage that spills over to your core is increased by 2.",
        flavour = "'You’re going to have to get used to that humming.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,

                [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                    if params.splashed_modifier == self.negotiator:FindCoreArgument() then
                        params.bonus_splash = params.bonus_splash + 2
                    end
                end,
            },
            
        },

    },

    brain_harp_plus =
    {
        name = "Visionary Brain Harp",
        desc = "Gain 1 action<#UPGRADE> and draw a card</> at the start of your turn, damage that spills over to your core is increased by 2.",
    },

    the_red =
    {
        name = "The Red",
        desc = "Gain 1 action at the start of your turn. All Diplomacy cards deal 50% less damage.",
        flavour = "'Surprisingly delicious while being low in sugar and rage inducing.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_MULTIPLIER,
            },

            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if source.owner == self.owner then
                        if is_instance( source, Negotiation.Card ) and CheckBits( source.flags, CARD_FLAGS.DIPLOMACY ) then
                            persuasion:AddPersuasion( - math.floor(persuasion.min_persuasion * .5), - math.floor(persuasion.max_persuasion * .5), self )
                        end
                    end
                end,
                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
            
        },
    },

    the_red_plus =
    {
        name = "The Visionary Red",
        desc = "Gain 1 action <#UPGRADE>and draw a card</> at the start of your turn. All Diplomacy cards deal 50% less damage.",
        card_draw = 1,
    },

    shrokian_glamour_stones =
    {
        name = "Shrokian Glamour Stones",
        desc = "Gain 1 action at the start of your turn, bystanders and pets no longer assist you in negotiations.",
        flavour = "'Impressive to those rough-skinned shrokes but just absolute Hesh on your posture.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {

                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
            
        },

        event_handlers =
        {
            -- Find any helpers and suppress them
            [ "get_negotiation_support" ] = function( self, agent, subject, params )
                -- print("shrokian_glamour_stones.get_negotiation_support", agent, params.sup, params.sup == SUPPORT_ACTION.HELP)
                if params.sup == SUPPORT_ACTION.HELP then
                    params.sup = SUPPORT_ACTION.SUPPRESSED
                end
            end
        }
    },

    shrokian_glamour_stones_plus =
    {
        name = "Visionary Glamour Stones",
        desc = "Gain 1 action<#UPGRADE> and draw a card</> at the start of your turn, bystanders and pets no longer assist you in negotiations.",
        card_draw = 1,
    },
}
AddOpinionEvent("THREE_FINGERS", 
{
    delta = OPINION_DELTAS.MAJOR_BAD,
    txt = "Affected by Three Fingers",
})

Content.AddNegotiationCard("conductor", 
{
    name = "Conductor",
    flavour = "'Grab the conversation by the throat!'",
    desc = "Force all enemy intents and arguments to target a specific argument this turn.",

    cost = 0,
    flags = negotiation_defs.CARD_FLAGS.EXPEND | negotiation_defs.CARD_FLAGS.STICKY | negotiation_defs.CARD_FLAGS.ITEM,
    max_xp = 0,

    rarity = CARD_RARITY.UNIQUE,
    series = CARD_SERIES.GENERAL,

    target_self = TARGET_ANY_RESOLVE,

    OnPostResolve = function( self, minigame, targets )
        for i,intent in ipairs(minigame:GetOtherNegotiator(self.negotiator):GetIntents()) do
            if intent.target then
                intent.target = targets[1]
            end
        end
        for i,argument in minigame:GetOtherNegotiator(self.negotiator):Modifiers() do
            if argument.target then
                argument.target = targets[1]
            end
        end
        if not targets[1].force_target then
            targets[1].force_target = true
            local mod = self.negotiator:AddModifier("conductor")
            mod.chosen = targets[1]
        end
    end,

    modifier =
    {
        hidden = true,

        event_handlers = 
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator ~= self.negotiator then
                    if self.chosen then
                        self.chosen.force_target = nil
                    end
                    self.negotiator:RemoveModifier(self.id, self.stacks)
                end
            end
        },
    },
})

Content.AddNegotiationCard("conductor_upgraded", 
{
    name = "Super Conductor",
    desc = "Force all enemy intents and arguments to target a specific argument this turn. Give that argument 5 {COMPOSURE}.",
    icon =  "negotiation/conductor.tex",

    cost = 0,
    flags = negotiation_defs.CARD_FLAGS.EXPEND | negotiation_defs.CARD_FLAGS.STICKY | negotiation_defs.CARD_FLAGS.ITEM,

    rarity = CARD_RARITY.UNIQUE,
    series = CARD_SERIES.GENERAL,
    target_self = TARGET_ANY_RESOLVE,

    composure_amt = 5,

    OnPostResolve = function( self, minigame, targets )
        for i,target in ipairs(targets) do
            target:DeltaComposure(self.composure_amt, self)
        end
        for i,intent in ipairs(minigame:GetOtherNegotiator(self.negotiator):GetIntents()) do
            if intent.target then
                intent.target = targets[1]
            end
        end
        for i,argument in minigame:GetOtherNegotiator(self.negotiator):Modifiers() do
            if argument.target then
                argument.target = targets[1]
            end
        end
    end,
})

---------------------------------------------------------------------------------------------

for i, id, graft in sorted_pairs( NEGOTIATION_GRAFTS ) do
    graft.card_defs = negotiation_defs
    graft.series = "GENERAL"
    
    if graft.negotiation_modifier and graft.negotiation_modifier.modifier_type == nil then
        graft.negotiation_modifier.modifier_type = MODIFIER_TYPE.HIDDEN
    end
    graft.type = GRAFT_TYPE.NEGOTIATION
    Content.AddGraft( id, graft )
end
