local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local CARD_FLAGS = negotiation_defs.CARD_FLAGS


local NEGOTIATION_GRAFTS =
{
    -----------------------------------------------------------------
    -- COMMONS
    parietal_integrator =
    {
        name = "Parietal Integrator",
        desc = "At the start of each negotiation, discard up to 2 cards and draw that many new cards.",
        flavour = "'It kind of filters out thoughts that are too distracting. Don’t worry, they probably weren’t too important anyways.'",

        rarity = CARD_RARITY.COMMON,
        card_count = 2,

        negotiation_modifier =
        {
            event_handlers = 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local discarded_cards = minigame:DiscardCards(0, self.graft:GetDef().card_count, self)
                    if #discarded_cards > 0 then
                        minigame:DrawCards(#discarded_cards)
                    end
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    self.negotiator:RemoveModifier( self )
                end
            },
        },
    },

    parietal_integrator_plus =
    {
        name = "Boosted Parietal Integrator",
        desc = "At the start of each negotiation, discard up to <#UPGRADE>4</> cards and draw that many new cards.",
        card_count = 4,
    },

    reflexor_snails = 
    {
        name = "Reflexor: Snails",
        desc = "At the start of each negotiation, {RIG} 2 {SNAILS}.",
        flavour = "'You just have to tilt it a little to the right.'",

        rig_amt = 2,
        rarity = CARD_RARITY.COMMON,
        restrictions = function( agent )
            return not agent.graft_owner:HasGraft("reflexor_heads")
        end,
        
        OnStartNegotiation = function(self, minigame)
            minigame:GetPlayerNegotiator():AddModifier("RIG_SNAILS", self:GetDef().rig_amt)
        end,
    },

    reflexor_snails_plus =
    {
        name = "Boosted Reflexor: Snails",
        desc = "At the start of each negotiation, {RIG} <#UPGRADE>4</> {SNAILS}.",
        rig_amt = 4,
    },

    reflexor_heads = 
    {
        name = "Reflexor: Heads",
        desc = "At the start of each negotiation, {RIG} 2 {HEADS}.",
        flavour = "'You just have to tilt it a little to the left.'",

        rarity = CARD_RARITY.COMMON,
        restrictions = function( agent )
            return not agent.graft_owner:HasGraft("reflexor_snails")
        end,
        rig_amt = 2,
        OnStartNegotiation = function(self, minigame)
            minigame:GetPlayerNegotiator():AddModifier("RIG_HEADS", self:GetDef().rig_amt)
        end,
    },

    reflexor_heads_plus =
    {
        name = "Boosted Reflexor: Heads",
        desc = "At the start of each negotiation, {RIG} <#UPGRADE>4</> {HEADS}.",
        rig_amt = 4,
    },

    stinger = 
    {
        name = "Stinger",
        desc = "Whenever you discard a card, deal 2 damage to a random opponent argument.",
        flavour = "'For those who love making strong points.'",

        rarity = CARD_RARITY.RARE,
        damage = 2,

        negotiation_modifier = 
        {
            target_enemy = TARGET_ANY_RESOLVE,
            min_persuasion = 2,
            max_persuasion = 2,
            event_handlers = 
            {
                [ EVENT.CARD_DISCARDED ] = function( self, card, minigame )
                    if card.owner == self.owner then
                        self.min_persuasion = self.graft:GetDef().damage
                        self.max_persuasion = self.graft:GetDef().damage
                        self:ApplyPersuasion()
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,
            },
        },
    },

    stinger_plus =
    {
        name = "Boosted Stinger",
        desc = "Whenever you discard a card, deal <#UPGRADE>3</> damage to a random opponent argument.",
        damage = 3,
    },

    -----------------------------------------------------------------
    -- UNCOMMONS
    thought_reverser =
    {
        name = "Thought Reverser",
        desc = "Every turn, the first card you draw after your initial hand is {PREPARE|Prepared}.",
        flavour = "'Changing your mind comes really naturally with this here. Makes you maintain that level of confidence.'",

        rarity = CARD_RARITY.UNCOMMON,

        num_cards = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                    if not start_of_turn then
                        if (self.count or 0) < self.graft:GetDef().num_cards then
                            minigame:PrepareCard(card, self)
                            self.count = (self.count or 0) + 1
                        end
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.count = 0
                end
            },
        },
    },

    thought_reverser_plus = 
    {
        name = "Boosted Thought Reverser",
        desc = "Every turn, the first <#UPGRADE>two cards</> you draw after your initial hand are {PREPARE|Prepared}.",
        num_cards = 2,
    },

    cranial_patch = 
    {
        name = "Cranial Patch",
        desc = "Every time you {PREPARE} a card, {GAMBLE}.",
        flavour = "'Yeah there’s some damage there. This will fix it and keep your eyes on the prize.'",

        rarity = CARD_RARITY.UNCOMMON,
        mod_xp = 3,

        negotiation_modifier =
        {
            event_handlers = 
            {
                [ EVENT.CARDS_SWAPPED ] = function( self, minigame, card )
                    local coin = self.negotiator:FindModifier("LUCKY_COIN")
                    if coin then coin:Gamble(self) end
                    if self.graft:GetDef().composure_amt then
                        self.negotiator:DeltaComposure(self.graft:GetDef().composure_amt, self)
                    end
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                end,
            },
        }
    },

    cranial_patch_plus =
    {
        name = "Stone Cranial Patch",
        desc = "Every time you {PREPARE} a card, {GAMBLE}<#UPGRADE> and gain 2 {COMPOSURE}</>.",
        composure_amt = 2,
    },

    memory_lock =
    {
        name = "Memory Lock",
        desc = "At the start of each turn, {PREPARE} a card.",
        flavour = "'You aren’t going to miss a thing with this little guy imbedded between your hemispheres.'",

        rarity = CARD_RARITY.UNCOMMON,
        card_draw = 0,

        negotiation_modifier =
        {
            event_priorities =
            {
                [ EVENT.HAND_DRAWN ] = 10, -- parietal_integrator
            },
            event_handlers = 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    local draw = self.graft:GetDef().card_draw 
                    if draw > 0 then minigame:DrawCards(draw) end
                    minigame:PrepareCard(nil, self)
                end
            },
        },
    },

    memory_lock_plus =
    {
        name = "Visionary Memory Lock",
        desc = "At the start of each turn, <#UPGRADE>draw a card, then</> {PREPARE} a card.",
        card_draw = 1,
    },

    kickoff =
    {
        name = "Kickoff",
        desc = "At the start of each turn, {GAMBLE}.",
        flavour = "'Let’s start this conversation off right.'",

        rarity = CARD_RARITY.UNCOMMON,
        gamble_count = 1,

        negotiation_modifier = 
        {
            event_handlers = 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local coin = self.negotiator:FindModifier("LUCKY_COIN")
                    if coin then 
                        for i=1, self.graft:GetDef().gamble_count do
                            coin:Gamble(self)
                        end
                    end
                end,
            },
        },
    },

    kickoff_plus = 
    {
        name = "Mirrored Kickoff",
        desc = "At the start of each turn, {GAMBLE} <#UPGRADE>twice</>.",
        gamble_count = 2,
    },

    tumbler = 
    {
        name = "Tumbler",
        desc = "Whenever you {PREPARE} a card, deal 2 damage to a random opponent argument.",
        flavour = "'It’s a consistent vibration that keeps your stronger points resonating at a higher frequency.'",

        rarity = CARD_RARITY.UNCOMMON,

        damage = 2,

        negotiation_modifier = 
        {
            hidden = true,
            min_persuasion = 2,
            max_persuasion = 2,
            target_enemy = TARGET_ANY_RESOLVE,
            target_mod = TARGET_MOD.RANDOM1,
            event_handlers = 
            {
                [ EVENT.CARDS_SWAPPED ] = function( self, minigame, card )
                    self.min_persuasion = self.graft:GetDef().damage
                    self.max_persuasion = self.graft:GetDef().damage
                    self:ApplyPersuasion()
                    self:NotifyTriggered()
                end,
            },
        },
    },

    tumbler_plus = 
    {
        name = "Boosted Tumbler",
        desc = "Whenever you {PREPARE} a card, deal <#UPGRADE>3</> damage to a random opponent argument.",
        damage = 3,
    },

    reinforcer =
    {
        name = "Reinforcer",
        desc = "Whenever you draw a card after your initial hand, apply 1 {COMPOSURE} to a random argument.",
        flavour = "'Wow, those are some strong arguments.'",

        rarity = CARD_RARITY.UNCOMMON,

        composure_amt = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.DRAW_CARD ] = function( self, engine, card, start_of_turn )
                    if not start_of_turn then
                        local targets = engine:CollectAlliedTargets(self.negotiator)
                        if #targets > 0 then
                            local target = targets[math.random(#targets)]
                            target:DeltaComposure(self.graft:GetDef().composure_amt, self)
                        end
                        self:NotifyTriggered()
                    end           
                end,
            },
        },
    },

    reinforcer_plus =
    {
        name = "Stone Reinforcer",
        desc = "Whenever you draw a card after your initial hand, apply <#UPGRADE>2</> {COMPOSURE} to a random argument.",
        composure_amt = 2,
    },

    ------------------------------------------------------------------
    -- RARE
    weighted_coin =
    {
        name = "Weighted Coin",
        desc = "Every 5th time you {GAMBLE}, gain 1 action.",
        flavour = "'Yeah, I would call it cheating.'",

        rarity = CARD_RARITY.RARE,

        gamble_count = 5,

        display_number = function( self )
            return self.userdata.counter or 0
        end,

        negotiation_modifier = 
        {
            event_handlers = 
            {
                [ EVENT.GAMBLE ] = function( self, result, source )
                    if self.graft then
                        self.graft.userdata.counter = (self.graft.userdata.counter or 0) + 1
                        if self.graft.userdata.counter >= self.graft:GetDef().gamble_count then
                            self.engine:ModifyActionCount(1)
                            self.graft.userdata.counter = 0
                        end
                        self:NotifyTriggered()
                    end
                end,
            },
        },
    },

    weighted_coin_plus = 
    {
        name = "Boosted Weighted Coin",
        desc = "Every <#UPGRADE>4th</> time you {GAMBLE}, gain 1 action.",
        gamble_count = 4,
    },
    
    slug_nook = 
    {
        name = "Slug Nook",
        desc = "The first time you {GAMBLE} and get Snails, a random card in your hand costs 1 less until the end of your turn. Resets at the start of your turn.",
        flavour = "'It’s subtle but the coins landing sure feels good in your hand doesn’t it?'",

        rarity = CARD_RARITY.RARE,
        cost_reduction = 1,

        negotiation_modifier = 
        {
            cards = {},
            CanReduceCost = function( card )
                return card.engine:CalculateActionCost( card ) > 0 and not card:IsFlagged( CARD_FLAGS.UNPLAYABLE )
            end,

            event_handlers = 
            {
                [ EVENT.GAMBLE ] = function( self, result, source )
                    if not self.used and result == "SNAILS" then
                        local hand = self.engine:GetHandDeck()
                        local card = hand:PickCardIf( self.CanReduceCost )
                        if card then 
                            table.insert(self.cards, card)
                        end
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        self.used = true
                    end
                end,

                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.cards and table.contains(self.cards, card) then
                        cost_acc:AddValue(-self.graft:GetDef().cost_reduction, self)
                    end
                end,

                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if table.contains(self.cards, card) then
                        table.remove(self.cards, self.cards[card])
                    end
                end,

                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    self.cards = {}
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.used = false
                end
            },
        },
    },

    slug_nook_plus =
    {
        name = "Boosted Slug Nook",
        desc = "The first time you {GAMBLE} and get Snails, a random card in your hand costs <#UPGRADE>2</> less until the end of your turn. Resets at the start of your turn.",
        cost_reduction = 2,
    },

    warding_trinket =
    {
        name = "Warding Trinket",
        desc = "Whenever you {GAMBLE}, apply 1 {COMPOSURE} to a random friendly argument.",
        flavour = "'As long as they are looking where they can, they aren’t looking where they should.'",

        rarity= CARD_RARITY.RARE,
        composure_amt = 1,
        argument_count = 1,

        negotiation_modifier = 
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.GAMBLE ] = function( self, result, source  )
                    local targets = self.engine:CollectAlliedTargets(self.negotiator)
                    if #targets > 0 then
                        targets = table.multipick(targets, self.graft:GetDef().argument_count)
                        for i,target in ipairs(targets) do
                            target:DeltaComposure(self.graft:GetDef().composure_amt, self)
                        end                        
                    end
                end,
            },
        },
    },

    warding_trinket_plus =
    {
        name = "Boosted Warding Trinket",
        desc = "Whenever you {GAMBLE}, apply 1 {COMPOSURE} to <#UPGRADE>2</> random friendly arguments.",
        argument_count = 2,
    },

    silver_tongue =
    {
        name = "Silver Tongue",
        desc = "Diplomacy cards deal 2 bonus damage while you have any {INFLUENCE}.",
        flavour = "'Everything just tastes so much better.'",

        rarity = CARD_RARITY.RARE,
        bonus_damage = 2,

        negotiation_modifier = 
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if self.negotiator:HasModifier("INFLUENCE") and is_instance( source, Negotiation.Card ) and CheckBits(source.flags, CARD_FLAGS.DIPLOMACY ) then
                        persuasion:AddPersuasion( self.graft:GetDef().bonus_damage, self.graft:GetDef().bonus_damage, self )
                    end
                end
            },
        },
    },

    silver_tongue_plus =
    {
        name = "Boosted Silver Tongue",
        bonus_damage = 3,
        desc = "Diplomacy cards deal <#UPGRADE>3</> bonus damage while you have any {INFLUENCE}.",
    },

    temporal_wire =
    {
        name = "Temporal Wire",
        desc = "Your {PREPARED} card deals 2 bonus damage.",
        flavour = "'Gives you a bit of space when concentrating, feels a little like you are cheating time.'",

        rarity = CARD_RARITY.RARE,
        bonus_damage = 2,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers = 
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if is_instance(source, Negotiation.Card) and source.owner == self.owner and (source:IsPrepared() or (self.buffed_card and self.buffed_card == source )) then
                        persuasion:AddPersuasion( self.graft:GetDef().bonus_damage, self.graft:GetDef().bonus_damage, self)
                    end
                end,

                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if target_deck == self.engine.resolve_deck then
                        if source_idx == 1 then
                            self.buffed_card = card
                        end
                    end
                end,

                [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                    if self.buffed_card and self.buffed_card == card then
                        self.buffed_card = nil
                    end
                end,
            },
        },
    },

    temporal_wire_plus =
    {
        name = "Boosted Temporal Wire",
        desc = "Your {PREPARED} card deals <#UPGRADE>4</> bonus damage.",
        bonus_damage = 4,
    },

    swamp_haunt =
    {
        name = "Swamp Haunt",
        desc = "Gain 1 action at the start of your turn. Gain a random negotiation parasite.",
        flavour = "'Crawling through your ear canal, and moving through your mind. The spined and twisting claw-backed worm, will leave your cares behind.'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,
        options = {"earworm", "tinnitus", "drowsiness"},

        OnAdded = function( self, agent )
            local chosen = self:GetDef().options[math.random(#self:GetDef().options)]
            TheGame:GetGameState():GetPlayerAgent().negotiator:LearnCard(chosen)
        end,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {

                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                end,
            },
            
        },
    },

    swamp_haunt_plus =
    {
        name = "Visionary Swamp Haunt",
        desc = "Gain 1 action<#UPGRADE> and draw a card</> at the start of your turn. Gain a random negotiation parasite.",
        card_draw = 1,
        OnAdded = function( self, agent )
        end,
    },
}


for i, id, graft in sorted_pairs( NEGOTIATION_GRAFTS ) do
    graft.card_defs = negotiation_defs
    graft.series = "ROOK"
    
    if graft.negotiation_modifier and graft.negotiation_modifier.modifier_type == nil then
        graft.negotiation_modifier.modifier_type = MODIFIER_TYPE.HIDDEN
    end
    graft.type = GRAFT_TYPE.NEGOTIATION
    Content.AddGraft( id, graft )
end