local battle_defs = require "battle/battle_defs"

local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local NEG_CARD_FLAGS = negotiation_defs.CARD_FLAGS

--------------------------------------------------------------------

local BATTLE_GRAFTS =
{
    void_chain =
    {
        name = "Void Chain",
        desc = "At the start of each battle, a random creature joins your team for the duration of the battle.",

        def_options = {"CRAYOTE_SPITTER", "CRAYOTE_UPGRADED", "VROC_UPGRADED", "FLEAD_UPGRADED", "SHROOGLET_UPGRADED", "CRAYOTE", "VROC", "FLEAD", "SHROOGLET", "MERCHANT_SNAIL"},

        battle_condition = 
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_COMBAT ] = function(self, battle)
                    self.spawned_agent = Agent( self.graft:GetDef().def_options[math.random(#self.graft:GetDef().def_options)] )
                    local screen = TheGame:FE():FindScreen( Screen.FightScreen )
                    if screen then
                        screen:PreloadAgentAnims(self.spawned_agent)
                    end
                end,

                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    local fighter = Fighter.CreateFromAgent( self.spawned_agent, battle:GetScenario():GetAllyScale() )
                    self.owner:GetTeam():AddFighter( fighter )
                    self.owner:GetTeam():ActivateNewFighters()
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    ether_hook =
    {
        name = "Ether Hook",
        desc = "At the start of each battle, {IMPROVISE_PLUS} a random {ABILITY} card, it costs 0 until played.",

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    local card_defs = BattleCardCollection():IsUnlocked():Filter(
                        function(card_def) 
                            local desc = card_def.desc or ""
                            local is_ability = desc:match("ABILITY") and true or false
                            return AgentUtil.CanUseCard(self:GetOwner().agent, card_def) and is_ability and not CheckBits(card_def.flags, CARD_FLAGS.UPGRADED)
                        end):Pick(5)

                    local cards = table.map(card_defs, 
                            function(card_def) 
                                local card = Battle.Card(card_def.id, self:GetOwner()) 
                                card.flags = SetBits( card.flags, CARD_FLAGS.FREEBIE )
                                card:ClearXP()
                                return card
                            end)
                    battle:ImproviseCards( cards, nil, nil, nil, nil, self )
                    self.owner:RemoveCondition(self.id)
                end
            },
        },
    },

    annihilation_cage =
    {
        name = "Annihilation Cage",
        desc = "At the end of every turn, choose any number of cards in your hand to gain {STICKY} for this battle.",

        battle_condition =
        {
            loc_strings =
            {
                CHOOSE_STICKY_STRING = "Choose cards and apply <b>Sticky</> to them.",
            },
            event_handlers =
            {
                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    local chosen_cards = battle:ChooseCards(nil, nil, nil, self:GetLocalizedString( "CHOOSE_STICKY_STRING" ))
                    for i,card in battle:GetHandDeck():Cards() do
                        if table.contains(chosen_cards, card) then
                            card:SetFlags(CARD_FLAGS.STICKY)
                        end
                    end
                end,
            },
        },
    },

    veterans_wire =
    {
        name = "Veteran's Wire",
        desc = "Whenever you add a card to your deck, it is immediately upgraded. (chosen randomly)",

        event_handlers =
        {
            [ "card_added" ] = function( self, card )
                if is_instance( card, Battle.Card ) and card:HasUpgrades() then
                    card:UpgradeCard()
                end
            end,
        }
    },

    despair_cables =
    {
        name = "Despair Cables",
        desc = "Whenever you fully defend an attack, increase the attacker's panic by 2.",

        battle_condition = 
        {
            event_handlers =
            {
                [ BATTLE_EVENT.ON_HIT ] = function( self, battle, attack, hit )
                    if hit.target == self.owner and (hit:CheckHitResult( "defended" ) or hit:CheckHitResult( "evaded" )) then
                        attack.attacker:ApplyMoraleDamage(2, self)
                    end
                end,
            },
        },
    },

    fused_collider =
    {
        name = "Fused Collider",
        desc = "All cards have a random cost between 0 and 2 and change cost when played or at the end of your turn.",

        battle_condition =
        {
            OnApply = function ( self, battle )
                self.affected_cards = {}
            end,
            event_priorities =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = 1,
            },
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.affected_cards[card] == nil then
                        self.affected_cards[card] = math.random(0,2)
                    end
                    cost_acc:ModifyValue(self.affected_cards[card], self)
                end,

                [ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
                    self.affected_cards[card] = nil
                end,

                [ BATTLE_EVENT.END_PLAYER_TURN ] = function( self, battle )
                    self.affected_cards = {}
                end
            },
        },
    },

    merged_glass =
    {
        name = "Merged Glass",
        desc = "{DEFEND} is halved instead of removed at the start of your turn.",

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.CALC_MODIFY_STACKS ] = function( self, acc, condition_id, fighter, source )
                    if fighter == self.owner and condition_id == "DEFEND" and source == self.owner:GetCondition("DEFEND") then
                        if acc.value < 0 then
                            acc:AddValue( - math.floor(acc.value * 0.5), self )
                        end
                    end
                end,
            },
        },
    },

    tomb_relic =
    {
        name = "Tomb Relic",
        desc = "At the start of each turn, gain 1 action for every 6 cards in your discard pile.",
        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        local count = math.floor(self.battle:GetDiscardDeck():CountCards() / 6)
                        self.battle:ModifyActionCount(count)
                    end
                end,
            },
        },
    },

    torrid_clone =
    {
        name = "Torrid Clone",
        desc = "The first card you play each turn is played twice.",

        battle_condition =
        {
            event_handlers =
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.active = true
                    end
                end,

                [ BATTLE_EVENT.POST_RESOLVE ] = function( self, battle, attack )
                    if self.active and attack.card and attack.card.owner == self.owner and attack.card.id ~= "execute" then
                        self.active = false
                        battle:PlayCard(attack.card)
                    end
                end
            },
        },
    },

    maelstrom_spine =
    {
        name = "Maelstrom Spine",
        desc = "Whenever you shuffle your deck, deal 4 damage to all enemies.",

        battle_condition = 
        {
            event_handlers =
            {
                [ BATTLE_EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
                    for i,enemy in self.battle:GetEnemyTeam():Fighters() do
                        enemy:ApplyDamage(4, self.owner, self)
                    end
                    self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                    if not self.battle.resolving then
                        self.battle:CheckAllSurrenders()
                    end
                end
            },
        },
    },

    wraith_needle =
    {
        name = "Wraith Needle",
        desc = "Whenever you {EXPEND} a card, {IMPROVISE} a random uncommon or rare card.",
        restrictions = function( agent )
            return agent:GetContentID() ~= "SMITH"
        end,

        battle_condition =
        {
            filter = function( card_def, self )
                return not (card_def.rarity == CARD_RARITY.UNIQUE or card_def.rarity == CARD_RARITY.BOSS) and (card_def.rarity == CARD_RARITY.UNCOMMON or card_def.rarity == CARD_RARITY.RARE)
            end,
            event_handlers =
            {
                [ BATTLE_EVENT.CARD_EXPENDED ] = function( self, battle, card )
                    if card.owner == self.owner then
                        self.battle:BroadcastEvent( BATTLE_EVENT.GRAFT_TRIGGERED, self.graft )
                        local collection = BattleCardCollection.AllUsableCards()
                        collection:Filter( self.filter, self )

                        local card_defs = collection:Pick(3)
                        local cards = table.map(card_defs, 
                                function(card_def) 
                                    local card = Battle.Card(card_def.id, self.owner) 
                                    card:ClearXP()
                                    return card
                                end)
                        local chosen_cards = battle:ImproviseCards( cards, self.draw_count, nil, nil, nil, self )
                        self.affected_cards = self.affected_cards or {}
                        for i,v in ipairs(chosen_cards) do
                            table.insert(self.affected_cards, v)
                        end
                    end
                end
            },  
        },
    },
}

---------------------------------------------------------------------------------------------

for i, id, graft in sorted_pairs( BATTLE_GRAFTS ) do
    graft.card_defs = battle_defs
    graft.series = "GENERAL"
    graft.rarity = CARD_RARITY.COSMIC
    graft.type = GRAFT_TYPE.COMBAT
    graft.brawl = true
    -- graft.unremovable = true
    if graft.battle_condition and graft.battle_condition.hidden == nil then
        graft.battle_condition.hidden = true
    end

    Content.AddGraft( id, graft )
end

local NEG_GRAFTS =
{
    critical_expanse =
    {
        name = "Critical Expanse",
        desc = "At the start of each turn, draw one extra card for each card in your hand at the end of last turn.",

        negotiation_modifier =
        {
            event_priorities =
            {
                [ EVENT.END_PLAYER_TURN ] = 999, -- jsut last.
            },
            event_handlers =
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    if (self.draw or 0) > 0 then
                        minigame:DrawCards( self.draw )
                    end
                end,

                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    self.draw = minigame:GetHandDeck():CountCards()
                end,
            }
        },
    },

    force_compression =
    {
        name = "Force Compression",
        desc = "All cards that cost more than 1 action cost 1 action instead.",

        negotiation_modifier =
        {
            event_priorities =
            {
                [ EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_CLAMP,
            },
            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if not card:IsFlagged(negotiation_defs.CARD_FLAGS.VARIABLE_COST) and cost_acc.value > 1 then
                        cost_acc:ModifyValue(1, self)
                    end
                end,
            },
        },
    },

    dynamic_reset =
    {
        name = "Dynamic Reset",
        desc = "At the start of each turn, discard any number of cards and draw that many new cards.",

        negotiation_modifier =
        {
            event_handlers = 
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local discarded_cards = minigame:DiscardCards(nil, nil, self)
                    if #discarded_cards > 0 then
                        minigame:DrawCards(#discarded_cards)
                    end
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                end
            },
        },
    },

    flawless_casualty =
    {
        name = "Flawless Casualty",
        desc = "While you have no arguments, all your cards deal double damage.",

        negotiation_modifier =
        {
            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_MULTIPLIER,
            },
            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if source and source.negotiator == self.negotiator and is_instance(source, Negotiation.Card) and self.negotiator:GetModifierCount(MODIFIER_TYPE.ARGUMENT) + self.negotiator:GetModifierCount(MODIFIER_TYPE.BOUNTY) == 0 then
                        persuasion:AddPersuasion( persuasion.min_persuasion, persuasion.max_persuasion, self )
                    end
                end,
            },
        },
    },

    collapsing_defeat =
    {
        name = "Collapsing Defeat",
        desc = "Whenever you fully mitigate damage, {INCEPT} 1 {DOUBT}.",

        negotiation_modifier =
        {
            event_handlers =
            {
                [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                    if target.negotiator == self.negotiator then
                        if damage <= defended then
                            self.anti_negotiator:InceptModifier("DOUBT", 1, self)
                        end
                    end
                end,
            },
        },
    },

    twined_conquest =
    {
        name = "Twinned Conquest",
        desc = "Duplicate the first argument you create each turn.",

        display_number = function( self )
            return self:GetCounter() or 0
        end,

        negotiation_modifier =
        {
            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                    if self.graft and self.graft:GetCounter() == 0 and source and source.negotiator and source.negotiator == self.negotiator and source ~= self and
                        not (is_instance(source, Negotiation.Card) and CheckBits(source.flags, NEG_CARD_FLAGS.BYSTANDER)) and
                        modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN and modifier.modifier_type ~= MODIFIER_TYPE.INCEPTION then
                        --
                        self.graft:IncrementCounter()
                        modifier.negotiator:CreateModifier(modifier.id, modifier.stacks, self)

                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    self.graft:ResetCounter()
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                end
            },
        },
    },

    immaculate_swarm =
    {
        name = "Immaculate Swarm",
        desc = "All cards deal 1 bonus damage per friendly argument.",

        negotiation_modifier =
        {
            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
            },
            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if source and is_instance(source, Negotiation.Card) and source.negotiator == self.negotiator and self.mod_count and self.mod_count > 0 then
                        persuasion:AddPersuasion( self.mod_count, self.mod_count, self )
                    end
                end,

                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN then
                        self.mod_count = (self.mod_count or 0) + 1
                    end
                end,

                [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN then
                        self.mod_count = (self.mod_count or 1) - 1
                    end
                end,
            },
        },
    },

    parlous_concept =
    {
        name = "Parlous Concept",
        desc = "Whenever you shuffle your deck, {IMPROVISE} a random uncommon or rare card. It costs 1 less.",

        negotiation_modifier =
        {
            filter = function( card_def, self )
                return (card_def.rarity == CARD_RARITY.UNCOMMON or card_def.rarity == CARD_RARITY.RARE) and not CheckBits(card_def.flags, NEG_CARD_FLAGS.ITEM)
            end,

            event_priorities =
            {
                [ EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_ADDITIVE,
            },
            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                    if self.affected_cards and table.contains(self.affected_cards, card) then
                        cost_acc:AddValue(-1, self)
                    end
                end,

                [ EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
                    local collection = NegotiationCardCollection.AllUsableCards()
                    collection:Filter( self.filter, self )

                    local card_defs = collection:Pick(3)
                    local cards = table.map(card_defs, 
                            function(card_def) 
                                local card = Negotiation.Card(card_def.id, self.negotiator.agent) 
                                card:ClearXP()
                                return card
                            end)
                    local chosen_cards = self.engine:ImproviseCards( cards, self.draw_count, nil, nil, nil, self )
                    self.affected_cards = self.affected_cards or {}
                    for i,v in ipairs(chosen_cards) do
                        table.insert(self.affected_cards, v)
                    end
                end
            },
        },
    },
}

---------------------------------------------------------------------------------------------

for i, id, graft in sorted_pairs( NEG_GRAFTS ) do
    graft.card_defs = negotiation_defs
    graft.series = "GENERAL"
    graft.rarity = CARD_RARITY.COSMIC
    graft.brawl = true
    -- graft.unremovable = true
    if graft.negotiation_modifier and graft.negotiation_modifier.modifier_type == nil then
        graft.negotiation_modifier.modifier_type = MODIFIER_TYPE.HIDDEN
    end
    graft.type = GRAFT_TYPE.NEGOTIATION
    Content.AddGraft( id, graft )
end