local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local CARD_FLAGS = negotiation_defs.CARD_FLAGS


local NEGOTIATION_GRAFTS =
{
    -----------------------------------------------------------------
    -- COMMONS

    shrieker =
    {
        name = "Shrieker",
        desc = "At the start of your turn, deal 1 damage to all opponent arguments.",
        flavour = "'Don’t worry, where it’s going you won’t be able to see any of it.'",
        rarity = CARD_RARITY.COMMON,

        min_persuasion = 1,
        max_persuasion = 1,

        negotiation_modifier =
        {
            name = "Shrieker",

            min_persuasion = 1,
            max_persuasion = 1,
            target_enemy = TARGET_ANY_RESOLVE,
            target_mod = TARGET_MOD.TEAM,

            event_priorities =
            {
                -- After: cludger
                [ EVENT.BEGIN_TURN ] = 11,
            },

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.min_persuasion = self.graft:GetDef().min_persuasion
                        self.max_persuasion = self.graft:GetDef().max_persuasion
                        self:ApplyPersuasion()
                    end
                end,
            }
        }
    },

    shrieker_plus =
    {
        name = "Boosted Shrieker",
        desc = "At the start of your turn, deal <#UPGRADE>2</> damage to all opponent arguments.",
        min_persuasion = 2,
        max_persuasion = 2,
    },

    distresser =
    {
        name = "Distresser",
        desc = "{VULNERABILITY} is twice as effective against your opponent.",
        flavour = "'Ugh, I hate these things, easiest way to clear a room.'",

        rarity = CARD_RARITY.COMMON,

        dmg_mult = 2,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.id == "VULNERABILITY" and modifier.negotiator ~= self.negotiator then
                        modifier.dmg_buff = modifier.dmg_buff * self.graft:GetDef().dmg_mult
                        modifier:NotifyChanged( nil, self )
                    end
                end
            },
        },
    },

    distresser_plus =
    {
        name = "Boosted Distresser",
        desc = "{VULNERABILITY} is <#UPGRADE>three times</> as effective against your opponent.",
        dmg_mult = 3,
    },

    cludger =
    {
        name = "Cludger",
        desc = "At the start of your turn, steal up to 4 {COMPOSURE} from a random opponent argument.",
        flavour = "'Mainly dealing with high frequency brain waves, the Cludger can literally drain away one’s confidence.'",
        rarity = CARD_RARITY.COMMON,

        composure_amt = 4,

        negotiation_modifier =
        {
            event_priorities =
            {
                -- Precede: shrieker
                [ EVENT.BEGIN_TURN ] = 10,
            },

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local targets = {}
                        minigame:CollectAlliedTargets(minigame:GetOpponentNegotiator(), targets, function(modifier) return modifier:GetComposure() >= 1 end)
                        table.sort( targets, function( a, b ) return a:GetComposure() > b:GetComposure() end )
                        if #targets > 0 then
                            local stolen = targets[1]:DeltaComposure(-self.graft:GetDef().composure_amt, self)
                            self.negotiator:DeltaComposure(-stolen, self)
                            minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end
                    end
                end,
            }
        }
    },

    cludger_plus =
    {
        name = "Stone Cludger",
        desc = "At the start of your turn, steal up to <#UPGRADE>6</> {COMPOSURE} from a random opponent argument.",
        composure_amt = 6,
    },

    cerebral_stimulants =
    {
        name = "Cerebral Stimulants",
        desc = "Draw one extra card per turn.",
        flavour = "'Oh yeah, I can feel them for sure. How much sleep am I supposed to be losing?'",
        rarity = CARD_RARITY.COMMON,
        draw_count = 1,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.CALC_TURN_DRAW ] = function( self, acc )
                    acc:AddValue( self.graft:GetDef().draw_count, self )
                end,

                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    cerebral_stimulants_plus =
    {
        name = "Visionary Cerebral Stimulants",
        desc = "Draw <#UPGRADE>2</> extra cards per turn.",
        draw_count = 2,
    },

    brain_spur = 
    {
        name = "Brain Spur",
        desc = "Your first card played every turn also damages an additional random opponent argument.",
        flavour = "'Brain Spurs are frowned upon in some communities because of the biting aggression that comes with the concentration boost.'",

        rarity = CARD_RARITY.RARE,
        attack_count = 1,

        negotiation_modifier =
        {
            hidden = true,

            used = false,

            target_enemy = TARGET_ANY_RESOLVE,

            event_handlers =
            {
                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if not self.used then
                        if card:IsAttack() and card.negotiator == self.negotiator then
                            for i=1,self.graft:GetDef().attack_count do
                                local new_target = self.engine:CollectPrimaryTarget( card )
                                if new_target then
                                    self.engine:ApplyPersuasion( card, new_target )
                                end
                            end
                        end
                        self.used = true
                    end
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function ( self, minigame )
                    self.used = false
                end
            },
        },
    },

    brain_spur_plus =
    {
        name = "Wide Brain Spur",
        desc = "Your first card played every turn also damages <#UPGRADE>2</> additional random opponent arguments.",
        attack_count = 2,
    },

    frontal_goad =
    {
        name = "Frontal Goad",
        desc = "All friendly arguments come into play with 2 bonus resolve.",
        flavour = "'Frontal goads used to be Admiralty guard standard before the market collapse.'",

        rarity = CARD_RARITY.UNCOMMON,
        
        bonus_resolve = 2,

        negotiation_modifier =
        {
            hidden = true,

            OnApply = function( self, minigame )
                -- Apply to existing modifiers to fix order-of-operation issues with frontal_goad not spawning first.
                for i, modifier in self.negotiator:Modifiers() do
                    if modifier.modifier_type ~= MODIFIER_TYPE.CORE and modifier:GetResolve() then
                        modifier:ModifyResolve(2, self)
                    end
                end
            end,

            event_handlers = 
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.max_resolve ~= nil and modifier.modifier_type == MODIFIER_TYPE.ARGUMENT then
                        self.engine:PushPostHandler( function()
                            modifier:ModifyResolve(2, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end )
                    end
                end
            },
        },
    },

    frontal_goad_plus = 
    {
        name = "Boosted Frontal Goad",
        desc = "All friendly arguments come into play with <#UPGRADE>3</> bonus resolve.",
        bonus_resolve = 3,
        negotiation_modifier =
        {
            hidden = true,

            OnApply = function( self, minigame )
                -- Apply to existing modifiers to fix order-of-operation issues with frontal_goad not spawning first.
                for i, modifier in self.negotiator:Modifiers() do
                    if modifier.modifier_type ~= MODIFIER_TYPE.CORE and modifier:GetResolve() then
                        modifier:ModifyResolve(3, self)
                    end
                end
            end,

            event_handlers = 
            {
                [ EVENT.MODIFIER_ADDED ] = function ( self, modifier, source )
                    if modifier.negotiator == self.negotiator and modifier.max_resolve ~= nil then
                        self.engine:PushPostHandler( function()
                            modifier:ModifyResolve(3, self)
                            self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                        end )
                    end
                end
            },
        },
    },

    keenskin = 
    {
        name = "Keenskin",
        desc = "Gain {COMPOSURE 2} whenever an opponent's argument is destroyed or removed by a card.",
        flavour = "'Heavily electrified epidermal implant. Keenskin has been around forever but not too many people get it installed.",

        rarity = CARD_RARITY.COMMON,
        composure_amt = 2,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.MODIFIER_REMOVED ] = function ( self, modifier, source )
                    if modifier.owner == self.anti_negotiator.agent and
                        (modifier.modifier_type == MODIFIER_TYPE.ARGUMENT or modifier.modifier_type == MODIFIER_TYPE.BOUNTY) and
                        is_instance( source, Negotiation.Card ) then
                        --
                        self.negotiator:DeltaComposure(self.graft:GetDef().composure_amt, self)
                        self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end
            },
        },
    },

    keenskin_plus =
    {
        name = "Stone Keenskin",
        desc = "Gain <#UPGRADE>{COMPOSURE 4}</> whenever an opponent's argument is destroyed or removed by a card.",
        composure_amt = 4,
    },

    -----------------------------------------------------------------
    -- UNCOMMONS

    luminjectors = 
    {
        name = "Luminjectors",
        desc = "A random friendly argument gains 2 resolve at the beginning of your turn.",
        flavour = "'Introducing lumin properly into the nervous system affects the willpower of the recipient. Properly being the key word here.'",

        rarity = CARD_RARITY.UNCOMMON,
        resolve_gain = 2,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local modifiers = {}
                    for i, modifier in self.negotiator:ModifierSlots() do
                        if modifier:GetResolve() ~= nil and modifier.modifier_type ~= MODIFIER_TYPE.CORE then
                            table.insert(modifiers, modifier)
                        end
                    end
                    
                    if #modifiers > 0 then
                        local mod = table.arraypick( modifiers )
                        mod:ModifyResolve( self.graft:GetDef().resolve_gain, self )
                        minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    end
                end,
            },

        },
    },

    luminjectors_plus =
    {
        name = "Boosted Luminjectors",
        desc = "A random friendly argument gains <#UPGRADE>3</> resolve at the beginning of your turn.",
        resolve_gain = 3,
    },

    midbrain_padding =
    {
        name = "Midbrain Padding",
        desc = "Apply 1 {COMPOSURE} on all arguments at the end of your turn.",
        flavour = "'Look, the change of hat size is worth the trouble, trust me.'",

        rarity = CARD_RARITY.UNCOMMON,
        composure_amt = 1,

        negotiation_modifier = 
        {
            event_priorities =
            {
                [ EVENT.END_PLAYER_TURN ] = 10,
            },
            event_handlers = 
            {
                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    local targets = minigame:CollectAlliedTargets(self.negotiator)
                    for i=1, #targets do
                        if (targets[i]:GetResolve() or 0) > 0 then
                            targets[i]:DeltaComposure(self.graft:GetDef().composure_amt, self)
                        end
                    end
                end
            },
        },
    },

    midbrain_padding_plus =
    {
        name = "Stone Midbrain Padding",
        desc = "Apply <#UPGRADE>2</> {COMPOSURE} on all arguments at the end of your turn.",
        composure_amt = 2,
    },

    disruptive_glanworm = 
    {
        name = "Disruptive Glanworm",
        desc = "Start each negotiation with {DOMINANCE 2}.",
        flavour = "'There’s a growing group of people that believe that the glanworms make you a lot more disagreeable when under stress.'",

        rarity = CARD_RARITY.UNCOMMON,
        
        OnStartNegotiation = function(self, minigame)
            minigame:GetPlayerNegotiator():AddModifier("DOMINANCE", 2)
        end,
    },

    disruptive_glanworm_plus =
    {
        name = "Assertive Glanworm",
        desc = "Start each negotiation with <#UPGRADE>{DOMINANCE 3}</>.",
        OnStartNegotiation = function(self, minigame)
            minigame:GetPlayerNegotiator():AddModifier("DOMINANCE", 3)
        end,
    },

    -----------------------------------------------------------------
    -- RARES

    mind_spool =
    {
        name = "Mind Spool",
        desc = "Draw 3 extra cards at the start of every turn. At the end of your turn, take 1 damage for every card in your hand above 3.",
        flavour = "'A very powerful cognitive accelerant, but one can get extremely carried away and turn into a rambling idiot.'",

        rarity = CARD_RARITY.RARE,
        draw_count = 3,

        negotiation_modifier = 
        {
            hidden = true,

            target_self = TARGET_FLAG.CORE,
            
            event_handlers = 
            {
                [ EVENT.BEGIN_PLAYER_TURN ] = function ( self, minigame )
                    minigame:DrawCards(self.graft:GetDef().draw_count)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft)
                end,

                [ EVENT.END_PLAYER_TURN ] = function ( self, minigame )
                    local damage = minigame:GetHandDeck():CountCards()
                    if damage > 3 and not minigame:CheckGameOver() then
                        self.min_persuasion = damage - 3
                        self.max_persuasion = damage - 3
                        self:ApplyPersuasion()
                    end
                end
            },
        },
    },

    mind_spool_plus =
    {
        name = "Visionary Mind Spool",
        desc = "Draw <#UPGRADE>4</> extra cards at the start of every turn. At the end of your turn, take 1 damage for every card in your hand above 3.",
        draw_count = 4,
    },

    ocular_surrogate =
    {
        name = "Ocular Surrogate",
        desc = "At the start of your turn, deal 4 damage to a random opponent argument.",
        flavour = "'No, you don’t get to pick the color.'",

        rarity = CARD_RARITY.RARE,
        min_persuasion = 4,
        max_persuasion = 4,
        negotiation_modifier =
        {
            hidden = true,
            target_enemy = TARGET_ANY_RESOLVE,
            min_persuasion = 4,
            max_persuasion = 4,

            event_priorities =
            {
                [ EVENT.BEGIN_TURN ] = 11, -- After Cludger
            },

            event_handlers = 
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if self.negotiator == negotiator then
                        self.min_persuasion = self.graft:GetDef().min_persuasion
                        self.max_persuasion = self.graft:GetDef().max_persuasion
                        self:ApplyPersuasion()
                    end
                end
            },
        },
    },

    ocular_surrogate_plus =
    {
        name = "Boosted Ocular Surrogate",
        desc = "At the start of your turn, deal <#UPGRADE>6</> damage to a random opponent argument.",
        min_persuasion = 6,
        max_persuasion = 6,
    },

    pons_mite = 
    {
        name = "Pons Mite",
        desc = "Increase the maximum damage of all cards by 1.",
        flavour = "'You’ll get used to the itchiness. And be sure to drink more water.'",

        bonus_damage = 1,

        rarity = CARD_RARITY.RARE,

        negotiation_modifier =
        {
            hidden = true,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
            },

            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion )
                    if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                        persuasion:AddPersuasion( 0, self.graft:GetDef().bonus_damage, self)
                    end
                end
            },
        },
    },

    pons_mite_plus =
    {
        name = "Boosted Pons Mite",
        desc = "Increase the maximum damage of all cards by <#UPGRADE>2</>.",
        bonus_damage = 2,
    },

    impulse_jack =
    {
        name = "Impulse Jack",
        desc = "At the start of every negotiation, {IMPROVISE} a 0 cost card from your deck.",
        flavour = "'Keeps your strongest ideas at the forefront for a fair installation price.'",

        rarity = CARD_RARITY.RARE,
        card_count = 1,

        negotiation_modifier =
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.HAND_DRAWN ] = function( self, minigame )
                    local cards = ObtainWorkTable()
                    for i, card in minigame:GetDrawDeck():Cards() do
                        if minigame:CalculateActionCost( card ) == 0 and not card:IsFlagged( CARD_FLAGS.UNPLAYABLE ) then
                            table.insert( cards, card )
                        end
                    end

                    cards = table.multipick(cards, 5)
                    minigame:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self )
                    minigame:ImproviseCards(cards, self.graft:GetDef().card_count, nil, nil, nil, self )
                    self.negotiator:RemoveModifier(self, self.stacks, self)
                end
            },
        },
    },

    impulse_jack_plus =
    {
        name = "Visionary Impulse Jack",
        desc = "At the start of every negotiation, {IMPROVISE} <#UPGRADE>2</> 0 cost cards from your deck.",
        card_count = 2,
    },

    scholars_clip =
    {
        name = "Scholar's Clip",
        desc = "Gain 1 action at the start of your turn. At the start of every turn remove 1 {INFLUENCE} and 1 {DOMINANCE}.",
        flavour = "'The constant cramping in your trapezius is just one of the many side effects!'",

        card_draw = 0,

        rarity = CARD_RARITY.BOSS,

        negotiation_modifier =
        {
            hidden = true,

            event_handlers =
            {

                [ EVENT.CALC_ACTIONS_PER_TURN ] = function( self, acc )
                    self.engine:BroadcastEvent( EVENT.GRAFT_TRIGGERED, self.graft )
                    acc:AddValue( 1 )
                end,

                [ EVENT.BEGIN_PLAYER_TURN ] = function( self, minigame )
                    local draw = self.graft:GetDef().card_draw
                    if draw > 0 then minigame:DrawCards(draw) end
                    self.negotiator:RemoveModifier("INFLUENCE", 1, self)
                    self.negotiator:RemoveModifier("DOMINANCE", 1, self)
                end,
            },
            
        },
    },

    scholars_clip_plus =
    {
        name = "Visionary Scholar's Clip",
        desc = "Gain 1 action<#UPGRADE> and draw a card</> at the start of your turn. At the start of every turn remove 1 {INFLUENCE} and 1 {DOMINANCE}.",
        card_draw = 1,
    },
}

---------------------------------------------------------------------------------------------

for i, id, graft in sorted_pairs( NEGOTIATION_GRAFTS ) do
    graft.card_defs = negotiation_defs
    graft.series = "SAL"
    
    if graft.negotiation_modifier and graft.negotiation_modifier.modifier_type == nil then
        graft.negotiation_modifier.modifier_type = MODIFIER_TYPE.HIDDEN
    end
    graft.type = GRAFT_TYPE.NEGOTIATION
    Content.AddGraft( id, graft )
end
