local fun = require "util/fun"
local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.BATTLE_EVENT

------------------------------------------------------------------------------------------------

local hesh_actions = {
    
    function(self, fighter) 
        local amt = 4 + 2 * fighter.battle.turns
        fighter:AddCondition( "NEXT_TURN_DEFENSE", amt, self )
    end,
    
    function(self, fighter) 
        local amt = 1 + math.ceil(fighter.battle.turns / 4)
        fighter:AddCondition("ADRENALINE", amt, self)
    end,
    
    function(self, fighter) 
        fighter:AddCondition("BONUS_ACTION", 1, self)
    end,

    function(self, fighter) 
        fighter:AddCondition("EVASION", 1, self)                
    end,
}


local GRAFTS =
{

    claimed_by_hesh = {
        name = "Claimed by Hesh",
        hidden = true,

        flavor = "Hesh has taken an interest in your affairs.",

        OnActivateFighter = function(self, fighter)
            self:AddBattleCondition(fighter)
        end,

        battle_condition = 
        {
            desc = "At the start of your turn, gain a random beneficial effect, take {1} damage, and increase by 1 up to a maximum of 6 stacks.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.stacks )
            end,

            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_TURN ] = function( self, fighter )
                    if fighter == self.owner then
                        self.owner:ApplyDamage( self.stacks )
                        if self.owner:IsAlive() then
                            fighter.battle:BroadcastEvent( BATTLE_EVENT.DELAY, 0.5 )
                            local action = table.arraypick(hesh_actions)
                            action(self, self.owner)
                            if fighter.battle.turns > 1 and self.stacks < 6 then
                                self.owner:AddCondition(self.id, 1)
                            end
                        end
                    end
                end,
            },
            CalculatePreviewDamage = function( self )
                return self.stacks or 0
            end,
        },
    },

    grout_bog_burrs =
    {
        name = "Burrs",
        hidden = true,
        desc = "Burrs will sometimes spawn during outdoor fights.",
        OnAdded = function(self)
            self.ignore_burr_count = math.random(2) - 1 --0,1
        end,

        OnActivateFighter = function(self, fighter)
            local battle = fighter:GetBattle()
            if not battle:GetScenario().location:IsIndoors() and not CheckBits(battle:GetScenario():GetFlags(), BATTLE_FLAGS.NO_BURRS) then
                local force_burrs =  battle:GetScenario():GetPrimary(TEAM.RED):GetFactionID() == "BOGGERS"
                
                if not force_burrs then
                    self.ignore_burr_count = (self.ignore_burr_count and self.ignore_burr_count - 1) or 0
                end
                
                if force_burrs or (self.ignore_burr_count or 0) <= 0 then
                    self:AddBattleCondition(fighter)
                end
            end
        end,            

        battle_condition = 
        {
            hidden = true,
            OnApply = function(self, battle) 
                
                self.spawn_schedule = {}
                self.num_to_spawn = 0

                self.num_to_spawn = math.random(2) -- 1,2
                
                local spawn_time = 2
                
                for k = 1, self.num_to_spawn do
                    local add = math.random(3)-1 -- 0,1,2
                    spawn_time = spawn_time + add
                    table.insert(self.spawn_schedule, spawn_time)
                end

                if #self.spawn_schedule > 0 then
                    print ("BURRS COMING IN:")
                for k,v in ipairs(self.spawn_schedule) do
                        print ("\tTURN ", v)
                    end
                else
                    print ("NO BURRS COMING")
                end
            end,
            
            hasRoomForGrout = function(self, battle, team_id, max_grout)
                max_grout = max_grout or 1
                if not battle:GetTeam(team_id):IsFull() then
                    local grout_count = fun(battle:GetTeam(team_id):GetFighters())
                                            :count(function(fighter) return fighter.agent.is_grout end)
                    return grout_count < max_grout
                end
                return false
            end,

            trySpawnGrout = function(self, battle)
                
                local max_grout_per_team = 2
                local spawn_team_id = fun{TEAM.RED, TEAM.BLUE}
                                        :filter(function(team_id) return self:hasRoomForGrout(battle, team_id, max_grout_per_team) end)
                                        :randomPick()
                                        
                if spawn_team_id then
                    local grout_id = fun(Content.GetAllCharacterDefs())
                                    :filter(function(def) return def.is_grout and (def.grout_team_restrict == nil or def.grout_team_restrict == spawn_team_id) end)
                                    :values(function(def) return def.id end)
                                    :randomPick()

                    local team = battle:GetTeam(spawn_team_id)
                    team:AddFighter( Fighter.CreateFromAgent( Agent(grout_id ) ) )
                    team:ActivateNewFighters()

                    if not self.spawned_burr then
                        self.spawned_burr = true
                        if self.graft.ignore_burr_count == nil or self.graft.ignore_burr_count <= 0 then
                            self.graft.ignore_burr_count = math.random(4)-1 -- 0,1,2,3
                        end
                    end

                    return true
                end
            end,

            event_priorities =
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = -100,
            },
            
            event_handlers = 
            {
                [ BATTLE_EVENT.BEGIN_PLAYER_TURN ] = function( self, battle )
                    if not battle:GetEnemyTeam():IsSurrendering() then
                        local turn = battle:GetTurns()
                        local did_spawn = false
                        while self.spawn_schedule and self.spawn_schedule[1] and self.spawn_schedule[1] <= turn do
                            local spawned = self:trySpawnGrout(battle)
                            if spawned then
                                table.remove(self.spawn_schedule, 1)
                                did_spawn = true
                            else
                                break
                            end
                        end

                        if did_spawn then
                            battle:RePrepareTurn()
                        end
                    end
                end
            },
        },
    },

    story_mode_graft =
    {
        hidden = true,
        event_handlers =
        {
            [ "new_game" ] = function( self, gamestate )
                gamestate:GetCaravan():UpgradeHealth( 15, "STORY_MODE" )
                gamestate:GetCaravan():UpgradeResolve( 10 )
            end,
        }
    },
}


---------------------------------------------------------------------------------------------

for id, graft in pairs( GRAFTS ) do
    Content.AddStoryGraft( id, graft )
end
