local negotiation_defs = require "negotiation/negotiation_defs"
local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT


local CARDS =
{
    default =
    {
        name = "Unimpressed",
        min_persuasion = 2,
        max_persuasion = 2,
        quip = "opponent_card",
        sound = "event:/sfx/battle/cards/neg/enemy_sends_damage",
        
        quips = {
            {
                [[
                    !crossed
                    * {self} has {self.hisher} doubts.
                ]],
                [[
                    !crossed
                    * {self} doesn't even respond to that.
                ]],
                [[
                    !hips
                    * {self} clearly wants this conversation to end.
                ]],
                [[
                    !hips
                    * {self} is not buying it.
                ]],
                [[
                    !hips
                    * You're going to have to try harder to convince {self} of that.
                ]],
                [[
                    !crossed
                    * {self} is holding out for a better argument.
                ]],
                [[
                    You're not half as convincing as you think you are.
                ]],
                [[
                    You expect me to believe that?
                ]],
                [[
                    I'm not convinced.
                ]],
                [[
                    No, I don't think so...
                ]],
                [[
                    What are you even talking about?
                ]],
                [[
                    Can you even hear yourself talking? 
                ]],
                [[
                    Why are we still talking?
                ]],
                [[
                    !dubious
                    Are you done yet?
                ]],
                [[
                    I'm not going to say you're wrong, but... you aren't right.
                ]],
                [[
                    Hrmmmph.
                ]],
            }
        }
    },

    impatience = 
    {
        name = "Impatience",

        quip = "impatience_card",

        max_delay = 3,

        OnPostResolve= function( self, minigame, targets )
            self.negotiator:AddModifier("IMPATIENCE", 1, self)
        end,

        quips = {
            {
                [[
                    !crossed
                    * {self} is getting impatient
                ]],
                [[
                    I don't have time for this!
                ]],
                [[
                    Just get to the point.
                ]],
                [[
                    I haven't got all day!
                ]],
                [[
                    You're starting to get on my nerves.
                ]],
            }
        }
    },

    npc_argument =
    {
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.modifier_id )
        end,

        flags = CARD_FLAGS.SPECIAL,
        cost = 0,
        stacks = 1,

        SetModifier = function( self, modifier_id )
            self.modifier_id = modifier_id
            self.debug_id = string.format( "npc_argument(%s)", tostring(modifier_id) )
        end,

        CanPlayCard = function( self, card, engine, target )
            if card == self then
                local def = Content.GetNegotiationModifier( self.modifier_id )
                assert_warning( def, self.modifier_id )
                local max_instances = math.min( def.max_instances or math.huge, self.max_instances or math.huge )
                if max_instances <= self.negotiator:GetModifierInstances( self.modifier_id ) then
                    return false, "Already exists"
                end
            end
            return true
        end,

        OnPostResolve = function( self, engine, targets )
            if self.modifier_id then
                engine:ApplyFeature( self.modifier_id, self.stacks, self, targets )
            end
        end,
    },

    heckler =
    {
        name = "Heckler",
        quips = {
            {
                [[ 
                    !angry
                    Don't listen to {player.himher}!
                ]],
                [[ 
                    !dismiss
                    Who are you trying to kid?
                ]],
                [[ 
                    !point
                    This one's not trustworthy, {agent}.
                ]],
                [[ 
                    !point
                    Don't believe a word {player.heshe} says!
                ]],
            },
        },

        flags = CARD_FLAGS.BYSTANDER | CARD_FLAGS.EXPEND,
        show_dealt = false,

        OnPostResolve = function( self, engine )
            self.negotiator:CreateModifier( "heckler", self.owner:GetRenown(), self.owner )
        end,

        modifier =
        {
            loc_strings =
            {
                SABOTAGE_DESC = "{1#agent} is sabotaging your negotiation!\nIncreases {2#agent}'s damage by {3}."
            },
            desc_fn = function( self, fmt_str, engine, widget )
                if self.enemy and widget and widget.PostPortrait then
                    local txt = loc.format( self.def:GetLocalizedString( "SABOTAGE_DESC" ), self.enemy, self.owner, self.stacks ) 
                    widget:PostPortrait( self.enemy, txt )
                end
            end,

            max_resolve = 2,
            max_stacks = 1,

            modifier_type = MODIFIER_TYPE.ARGUMENT,

            OnInit = function( self, enemy )
                self:SetResolve( 3, MODIFIER_SCALING.MED )
                assert( enemy == nil or is_instance( enemy, Agent ))
                self.enemy = enemy
            end,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
            },

            event_handlers =
            {
                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                        persuasion:AddPersuasion( self.stacks, self.stacks, self )
                    end
                end,
            }
        },
    },

    -----------------------------------
    -- JAKES --------------------------
    -----------------------------------

    ploy =
    {
        name = "Ploy",
        desc = "At the end of each turn, double the damage this argument deals.",
        max_instances = 2,

        OnPostResolve = function( self )
            self.negotiator:CreateModifier( self.id, nil, self )
        end,

        modifier =
        {
            OnInit = function( self )
                self:SetResolve( 2, MODIFIER_SCALING.LOW )
            end,

            min_persuasion = 1,
            max_persuasion = 1,
            modifier_type = MODIFIER_TYPE.ARGUMENT,
            target_enemy = TARGET_ANY_RESOLVE,

            OnSetStacks = function( self, old_stacks )
                if self.stacks > 0 then
                    self.min_persuasion = math.floor( 2^(self.stacks - 1) )
                else
                    self.min_persuasion = 0
                end

                self.max_persuasion = self.min_persuasion
            end,

            OnBeginTurn = function( self, minigame )
                self:ApplyPersuasion()

                -- FIXME: if owned by the player, acquiare a new target immediately even if it's
                -- for NEXT turn, because it's confusing to be unable to see the preview.
                if self.negotiator:IsPlayer() then
                    self.target_cleared_turn=  nil
                    self:PrepareTurn()
                end
            end,

            OnEndTurn = function( self, minigame )
                self.negotiator:DeltaModifier( self, 1 )
                self:NotifyTriggered()
            end,
        }
    },

    -----------------------------------
    -- FEUDALITES ---------------------
    -----------------------------------

    bone_tired =
    {
        name = "Bone Tired",
        desc = "When destroyed, {1} loses {2} resolve.",
        desc_fn = function( self, fmt_str )
            local name = self.negotiator and self.negotiator:GetName() or "the opponent"
            return loc.format( fmt_str, name, (self.bounty_resolve or 0) * (self.stacks or 1))
        end,

        OnDealt = function( self )
            if self.engine then
                self.bounty_resolve = 5 + self.engine:GetDifficulty()
            end
        end,

        OnPostResolve = function( self )
            self.negotiator:AddModifier( self.id, nil, self )
        end,

        quips = {
            {
                [[ 
                    !sigh
                    $neutralResigned
                    I'm so tired.
                ]],
                [[ 
                    !sigh
                    $neutralResigned
                    I just want to finish this.
                ]],
            },
        },
        modifier =
        {
            max_resolve = 1,
            modifier_type = MODIFIER_TYPE.BOUNTY,
            sound = "event:/sfx/battle/cards/neg/create_argument/bone_tired",

            OnInit = function( self, source )
                self:SetResolve( 2, MODIFIER_SCALING.LOW )
                self.bounty_resolve = self.resolve * 3
            end,

            OnBounty = function( self )
                self.negotiator:ModifyResolve( -self.bounty_resolve * self.stacks, self )
            end,
        },
    },


    intuition =
    {
        name = "Intuition",

        OnPostResolve = function( self )
            self.negotiator:CreateModifier( self.id, nil, self )
        end,

        modifier =
        {
            desc = "When you deploy a new argument, increase this argument's damage and resolve by {1}.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.bonus )
            end,

            OnInit = function( self )
                self:SetResolve( 4, MODIFIER_SCALING.MED )
                self.bonus = 1 + math.ceil( self.engine:GetDifficulty() / 2 )

                self.min_persuasion = 1
                self.max_persuasion = 1
            end,

            hide_stacks = true,
            target_enemy = TARGET_ANY_RESOLVE,
            modifier_type = MODIFIER_TYPE.ARGUMENT,
            sound = "event:/sfx/battle/cards/neg/create_argument/intuition",

            OnBeginTurn = function( self, minigame )
                self:ApplyPersuasion()
            end,

            event_handlers =
            {
                [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                
                    if modifier.negotiator ~= self.negotiator and (modifier.modifier_type == MODIFIER_TYPE.ARGUMENT or modifier.modifier_type == MODIFIER_TYPE.BOUNTY) then
                        self.min_persuasion = self.min_persuasion + self.bonus
                        self.max_persuasion = self.max_persuasion + self.bonus
                        self:ModifyResolve( self.bonus )
                        self:NotifyChanged()
                    end
                end
            }
        }
    },

    swindle =
    {
        name = "Swindle",
        desc = "Adds {1} {bad_deal} {1*card|cards} to your draw pile.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.count )
        end,

        count = 1,

        OnPostResolve  = function( self, minigame )
            local cards = {}
            for i = 1, self.count do
                local card = Negotiation.Card( "bad_deal", minigame:GetPlayer() )
                table.insert( cards, card )
            end
            minigame:InceptCards( cards, self )
        end,
    },
    
    auctioneer_defend_items =
    {
        name = "Call for Bids",

        OnPostResolve = function( self )
            
            local possible_targets = {}
            for i, mod in self.negotiator:Modifiers() do
                if mod.id == "auction_item" then
                    table.insert(possible_targets, mod)
                    
                end
            end
            table.sort(possible_targets, function(a,b) return a:GetResolve() < b:GetResolve() end)
            
            local num_to_give = math.min(#possible_targets, 1+math.random(2))
            if num_to_give > 0 then
                local each_amt = 30/num_to_give
            
                for k = 1, num_to_give do
                    possible_targets[k]:DeltaComposure( math.max(2, math.round( each_amt*(.5 + math.random()*.5))), self)
                end
            end

        end,

        quips = {
            {
                [[ 
                    Hesh does not sell so cheaply!
                    Can you afford such luxury?
                ]],
            },
        },
    },

    auctioneer_add_item =
    {
        name = "Add Item",

        CanPlayCard = function( self, card, engine, target )
            local core = self.negotiator:FindCoreArgument()
            return core and core.items_to_add and #core.items_to_add > 0
        end,

        OnPostResolve = function( self )
            local core = self.negotiator:FindCoreArgument()
            if core then
                core:AddAuctionItem()
            end
        end,

        quips = {
            pre = "!happy",
            {
                [[A new item is on the block!]],
                [[And now, up for bid! ]],
                [[Ladies and oshnus, I present...!]],
                [[Feast your eyes on this item!]],
            },
        },
    },    

    ------------------------------------------------------------------
    -- ADMIRALTY
    ------------------------------------------------------------------

    frisk =
    {
        name = "Frisk",
        desc = "Steal {1#money}",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.money )
        end,

        --sound = SoundEvents.conversation_accept_money,
        fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

        money = 5,
        stacks = 1,

        OnPostResolve = function( self, engine )
            self.negotiator:CreateModifier( self.id, self.stacks, self )
        end,

        modifier =
        {
            desc = "At the start of {1}'s turn, steal {2#money} from {3} and remove 1 <b>Frisk</b>.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, loc.cap( self:GetOwnerName() ), self.money, self:GetOpponentName() )
            end,
    
            sound = "event:/sfx/battle/cards/neg/create_argument/frisk",
            fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

            money = 10,
            max_resolve = 1,

            OnInit = function( self )
                self:SetResolve( 1, MODIFIER_SCALING.LOW )
            end,

            OnBeginTurn = function( self, minigame )
                if self.owner == minigame:GetPlayer() then
                    minigame:ModifyMoney( self.money )
                    self:NotifyTriggered()
                else
                    if minigame:GetMoney() > 0 then
                        minigame:ModifyMoney( -self.money)
                        self:NotifyTriggered()
                    end
                end
                self.negotiator:RemoveModifier( self, 1, self )
            end
        }
    },

    paper_pusher =
    {
        name = "Paper Pusher",

        OnPostResolve  = function( self, minigame )
            if GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ARGUMENT_PLUS ) then
                self.negotiator:CreateModifier( self.id, 2, self )
            else
                self.negotiator:CreateModifier( self.id, 1, self )
            end
        end,

        modifier =
        {
            modifier_type = MODIFIER_TYPE.ARGUMENT,
            sound = "event:/sfx/battle/cards/neg/create_argument/paper_pusher",

            desc = "Insert {1} {baffled} card into {2}'s draw pile at the beginning of {3}'s turn.",
            alt_desc = "Insert {1} {baffled} card.",

            desc_fn = function( self, fmt_str )
                if self.owner and self.owner:IsPlayer() then
                    fmt_str = (self.def or self):GetLocalizedString( "alt_desc" )
                    return loc.format( fmt_str, self.stacks )
                else
                    return loc.format( fmt_str, self.stacks, self.anti_negotiator:GetName(), self.negotiator.agent:GetName() )
                end
            end,


            OnInit = function( self )
                self:SetResolve( 3, MODIFIER_SCALING.MED )
            end,

            OnBeginTurn = function ( self, minigame )
                local cards = {}
                for i = 1, self.stacks do
                    local card = Negotiation.Card( "baffled", self.anti_negotiator.agent )
                    table.insert( cards, card )
                end
                minigame:InceptCards( cards, self )
            end
        }
    },

    -----------------------------------
    -- SPARK_BARONS---------------------
    -----------------------------------


    straw_man =
    {
        name = "Straw Man",
        desc = "When this argument is destroyed, gain {1} {VULNERABILITY}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, (self.stacks or 1) * 3 )
        end,

        OnPostResolve = function( self, minigame )
            self.anti_negotiator:CreateModifier( self.id, 1, self )
        end,

        modifier =
        {
            max_resolve = 1,

            modifier_type = MODIFIER_TYPE.BOUNTY,

            sound = "event:/sfx/battle/cards/neg/create_argument/strawman",

            OnBounty = function( self )
                self.negotiator:AddModifier( "VULNERABILITY", 3 * self.stacks )
            end,
        }
    },

    fearless =
    {
        name = "Fearless",
        desc = "Targets the opponent argument with the least resolve.\n\nIf the argument is destroyed, duplicate <b>Fearless</b>, otherwise, increase its stacks by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks )
        end,

        stacks = 1,

        OnPostResolve = function( self, minigame )
            local modifier = self.negotiator:CreateModifier( self.id, self.stacks, self )
        end,

        modifier =
        {
            modifier_type = MODIFIER_TYPE.ARGUMENT,
            sound = "event:/sfx/battle/cards/neg/create_argument/fearless",

            OnInit = function( self )
                self:SetResolve( 2, MODIFIER_SCALING.MED )
            end,

            modifier_type = MODIFIER_TYPE.ARGUMENT,

            IsValidTarget = function( target )
                return target:GetResolve() ~= nil
            end,

            min_persuasion = 1,
            max_persuasion = 1,
            target_enemy = TARGET_ANY_RESOLVE,

            OnSetStacks = function( self )
                self.min_persuasion = self.stacks
                self.max_persuasion = self.stacks
            end,

            EvaluateTargetWeight = function( self, target, targets )
                local lo_target
                for i, modifier in ipairs( targets ) do
                    local resolve = modifier:GetResolve()
                    if resolve ~= nil and (lo_target == nil or resolve < lo_target:GetResolve()) then
                        lo_target = modifier
                    end
                end
                return target == lo_target and 1 or 0
            end,

            OnBeginTurn = function( self, minigame )
                local target = self.target -- ApplyPersuasion() clears the target.
                if self:ApplyPersuasion() then
                    if target and not self.anti_negotiator:HasModifier( target ) then
                        self.negotiator:CreateModifier( self.id, self.stacks, self )
                    else
                        self.negotiator:AddModifier( self, 1 )
                    end
                end
            end,

        }
    },

    ai_drunk = 
    {
        name = "Drunk",

        flags = CARD_FLAGS.BYSTANDER,
        show_dealt = false,
        
        OnPostResolve = function( self, minigame )
            self.negotiator:AddModifier("DRUNK")
        end,
    },

    task_diplomacy = 
    {
        name = "Night Shift",
        desc = "If you gain {INFLUENCE}, {CONSUME} this and draw a card.",
        flavour = "'I prefer it, gives me time to work on my novella.'",

        icon = "negotiation/night_shift.tex",
        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.STATUS,

        Trigger = function(self)
            self:NotifyTriggeredPre()
            self.engine:ExpendCard(self)
            self.engine:DrawCards(1)
            self:NotifyTriggeredPost()
        end,

        event_handlers = 
        {
            [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                if modifier.id == "INFLUENCE" then
                    self:Trigger()
                end
            end,

            [ EVENT.MODIFIER_CHANGED ] = function( self, modifier, delta, clone, source )
                if modifier.id == "INFLUENCE" and delta > 0 then
                    self:Trigger()
                end
            end,
        },
    },

    task_argument =
    {
        name = "Day Shift",
        desc = "If any argument is removed or destroyed, {CONSUME} this and draw a card.",
        flavour = "'Come on, cover for me.'",
        icon = "negotiation/day_shift.tex",
        
        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.STATUS,
 
        event_handlers = 
        {
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, card )
                if modifier.modifier_type == MODIFIER_TYPE.ARGUMENT or modifier.modifier_type == MODIFIER_TYPE.BOUNTY then
                    self:NotifyTriggeredPre()
                    self.engine:ExpendCard(self)
                    self.engine:DrawCards(1)
                    self:NotifyTriggeredPost()
                end
            end,
        },
    },

    baffled =
    {
        name = "Baffled",
        desc = "At the end of your turn, take 2 resolve damage and {EXPEND} this card.",
        flavour = "'Uhhhh...'",

        cost = 0,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.UNPLAYABLE,

        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                self:NotifyTriggeredPre()
                minigame:ApplyPersuasion( self, self.negotiator, 2, 2 )
                self:NotifyTriggeredPost()
                minigame:ExpendCard(self)
            end
        },
    },

    silence =
    {
        name = "Silence",
        flavour = "'...'",
        cost = 1,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,
    },

    kickback =
    {
        name = "Kickback",
        flavour = "'Here’s a little something extra.'",
        quip = "bribery",

        hit_emote = table.empty,

        cost = 0,

        money_cost = 5,

        money_per_difficulty = {0, 0, 5, 5, 10},

        min_persuasion = 2,
        max_persuasion = 5,

        flags = CARD_FLAGS.STATUS,
        deck_handlers = ALL_DECKS,
        event_handlers =
        {
            [ EVENT.CALC_MONEY_COST ] = function ( self, acc, card )
                if card == self then
                    acc:AddValue(self.money_per_difficulty[self.engine:GetDifficulty()] or 5)
                end
            end,
        },
    },

    hush_money = 
    {
        name = "Hush Money",
        desc = "Dismiss a random enemy intent.",
        flavour = "'WOW! THIS IS A LOT OF MONEY!'",
        quip = "bribery",
        cost = 0,

        flags = CARD_FLAGS.MANIPULATE | CARD_FLAGS.EXPEND,

        OnPostResolve = function( self, minigame, targets )
            local intents = minigame:GetOpponentNegotiator():GetIntents()
            if #intents > 0 then
                minigame:GetOpponentNegotiator():DismissIntent(intents[math.random(#intents)])
            end
        end
    },

    bad_deal =
    {
        name = "Bad Deal",
        desc = "Pay {1#money}.\nIf you end your turn with this card, take {2} damage.",
        flavour = "'Sorry pal, you got taken for a ride.'",
        desc_fn = function( self, fmt_str )
            if self.engine then
                local cost = self.engine:CalculateMoneyCost( self, true )
                return loc.format( fmt_str, cost, self.resolve_damage )
            else
                return loc.format( fmt_str, self.money_cost, self.resolve_damage )
            end
        end,
        manual_money_desc = true,

        cost = 1,
        money_cost = 15,
        resolve_damage = 8,

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,

        event_handlers =
        {
            [ EVENT.END_TURN ] = function( self, engine, negotiator )
                if negotiator == self.negotiator and engine:GetResult() == nil then
                    self:NotifyTriggeredPre()
                    engine:ApplyPersuasion( self, self.negotiator, self.resolve_damage, self.resolve_damage )
                    self:NotifyTriggeredPost()
                    engine:ExpendCard(self)
                end
            end
        }
    },


    propaganda =
    {
        name = "Propaganda",
        desc = "{PREPARE} a card. Swap its cost with this card's cost.",
        flavour = "'I’ve heard great things.'",
        cost = 0,

        flags = CARD_FLAGS.MANIPULATE,
        rarity = CARD_RARITY.BASIC,

        OnPostResolve = function( self, minigame )
            local card = minigame:PrepareCard( nil, self )
            if card then
                local modifier = self.negotiator:FindModifier( self.id )
                if modifier == nil then
                    modifier = self.negotiator:AddModifier( self.id, 1 )
                end

                modifier:SwapCost( card, self )
            end
        end,

        modifier =
        {
            hidden = true,

            SwapCost = function( self, card1, card2 )
                local cost1 = self.engine:CalculateActionCost( card1 )
                local cost2 = self.engine:CalculateActionCost( card2 )

                if self.costs == nil then
                    self.costs = {}
                end
                self.costs[ card1 ] = cost2
                self.costs[ card2 ] = cost1
            end,

            event_priorities = {
                [ EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_SETTOR,
            },

            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, acc, card, target )
                    if self.costs and self.costs[ card ] then
                        acc:ModifyValue( self.costs[ card ], self )
                    end
                end,
            },
        }
    },

    moment_of_weakness =
    {
        name = "Moment of Weakness",
        desc = "When this argument is destroyed, deal {1} damage to the owner's core argument and spawn two new <b>Moment of Weakness</>.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, 3 )
        end,

        OnPostResolve = function( self, minigame )
            self.anti_negotiator:CreateModifier( self.id, 1, self )
        end,

        modifier =
        {
            max_resolve = 2,

            modifier_type = MODIFIER_TYPE.BOUNTY,

            OnBounty = function( self )
                self.engine:ApplyPersuasion( self, self.negotiator, 3, 3 )
                for i=1,2 do
                    if self.negotiator:GetSlotCount() + 1 < self.negotiator:GetMaxSlots() then
                        self.negotiator:CreateModifier(self.id, 1, self)
                    end
                end
            end,
        }
    },

    bog_boil =
    {
        name = "Bog Boil",
        desc = "Restore resolve to a random friendly argument equal to the resolve lost from <b>Bog Boil's</> attack.",
        
        OnPostResolve = function( self, minigame )
            self.negotiator:CreateModifier( self.id, 1, self)
        end,

        modifier =
        {
            modifier_type = MODIFIER_TYPE.ARGUMENT,

            max_resolve = 4,

            min_persuasion = 3,
            max_persuasion = 3,

            OnSetStacks = function( self )
                self.min_persuasion = self.stacks * 3
                self.max_persuasion = self.stacks * 3
            end,

            OnApply = function( self )
                self:PrepareTurn()
            end,

            target_enemy = TARGET_ANY_RESOLVE,

            event_handlers = 
            {
                [ EVENT.DELTA_RESOLVE ] = function( self, modifier, resolve, max_resolve, delta, source, params )
                    if source == self and delta < 0 then
                        self.negotiator:RestoreResolve(-delta, self)
                    end
                end,

                [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        if self.skipped_first or self.negotiator == minigame:GetPlayerNegotiator() then
                            self:ApplyPersuasion()
                        else
                            self.skipped_first = true
                        end
                    end
                end
            },
        },
    },

    bog_blast = 
    {
        name = "Bog Blast",
        desc = "Deals {1} less damage per friendly argument.",
        flavour = "'I have seen the half moon eye! You will all be made by the mud!'",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.reduction_amt )
        end,

        cost = 1,
        reduction_amt = 2,

        flags = CARD_FLAGS.MANIPULATE | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.BASIC,

        min_persuasion = 7,
        max_persuasion = 9,

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source == self then
                    local count = 0
                    for i, modifier in self.negotiator:Modifiers() do
                        if modifier:GetResolve() ~= nil and modifier.modifier_type ~= MODIFIER_TYPE.CORE then
                            count = count + 1
                        end
                    end
                    persuasion:AddPersuasion(-count*self.reduction_amt, -count*self.reduction_amt, self)
                end
            end,
        },
    },

    nadan_schmuck = 
    {
        OnPostResolve = function( self, minigame, targets )
            self.negotiator:CreateModifier("BANDIT_SCHMUCK", 1, self)
        end
    },

    bossbit_gas =
    {
        OnPostResolve = function( self, minigame )
            self.anti_negotiator:CreateModifier( "GAS", 1, self )
        end,
    },

    grout_bog_rash =
    {
        OnPostResolve = function( self, minigame )
            self.anti_negotiator:InceptModifier( "BOG_RASH", 1, self, true )
        end,
    },

    grout_bog_boil =
    {
        OnPostResolve = function( self, minigame )
            self.anti_negotiator:InceptModifier( "BOG_BOSS_BOIL", 1, self, true )
        end,
    },

    grout_bog_blister =
    {
        OnPostResolve =function( self, minigame )
            self.negotiator:CreateModifier("BLISTER", 1, self)
        end
    },

    grout_bog_fume =
    {
        OnPostResolve =function( self, minigame )
            self.negotiator:CreateModifier("FUME_BUBBLE", 1, self)
        end
    },

    grout_bog_tentacle =
    {
        OnPostResolve =function( self, minigame )
            self.negotiator:CreateModifier("WRITHING_APPENDAGES", 1, self)
        end
    },

    oshnu_slime =
    {
        name = "Oshnu Slime",
        desc = "You draw 1 less card at the start of your turn. At the end of your turn, remove 1 <b>Oshnu Slime</>.",

        OnPostResolve = function( self, minigame )
            self.anti_negotiator:CreateModifier( self.id, 2, self )
        end,

        modifier =
        {
            modifier_type = MODIFIER_TYPE.INCEPTION,

            event_handlers = 
            {
                [ EVENT.CALC_TURN_DRAW ] = function( self, acc )
                    acc:AddValue(-1, self)
                end,

                [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self.negotiator:RemoveModifier(self.id, 1, self)
                    end
                end
            },
        }
    },

    deception_inception =
    {
        stacks = 1,
        OnPostResolve = function( self, minigame )
            self.anti_negotiator:InceptModifier( "DECEPTION", self.stacks, self )
        end,
    },

    -- Special cards for showing choices to the player
    choose_diplomacy = 
    {
        name = "Diplomacy",
        desc = "",
        cost = 0,
        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.DIPLOMACY,
        rarity = CARD_RARITY.UNIQUE,
        hide_in_cardex = true,
    },

    choose_hostile =
    {
        name = "Hostility",
        desc = "",
        cost = 0,
        flags = CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.HOSTILE,
        rarity = CARD_RARITY.UNIQUE,
        hide_in_cardex = true,
    },
}

local MODIFIERS =
{

    APPROPRIATOR = 
    {
        name = "Appropriator",
        desc = "At the start of {1}'s turn, steal {2} {2*card|cards} from {3} and hold {2*it|them} hostage.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self:GetStealCount(), self:GetOpponentName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        GetStealCount = function( self )
            return GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ARGUMENT_PLUS ) and 2 or 1
        end,

        event_handlers = 
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local MAX_STOLEN = self:GetStealCount() * 3
                    local MIN_TO_LEAVE = 5

                    local num_stolen = 0
                    for i,modifier in self.negotiator:ModifierSlots() do
                        if modifier.id == "APPROPRIATED" then
                            num_stolen = num_stolen + modifier:GetStolenCount()
                        end
                    end

                    if num_stolen < MAX_STOLEN then
                        local cards = self.engine:GetAllPlayerCards(function(card) return not CheckBits( card.flags, CARD_FLAGS.STATUS ) end)
                        if #cards > MIN_TO_LEAVE then
                            local count = math.min( #cards - MIN_TO_LEAVE, self:GetStealCount() )
                            cards = table.multipick(cards, count)
                            local approp
                            if count > 1 then
                                approp = self.negotiator:CreateModifier("APPROPRIATED_plus", 1, self )
                            else
                                approp = self.negotiator:CreateModifier("APPROPRIATED", 1, self )
                            end
                            for i, card in ipairs( cards ) do
                                if approp:IsApplied() then -- veryify that it still exists
                                    print( self.negotiator, "appropriated", card, "from", card.deck )
                                    approp:AppropriateCard( card )
                                end
                            end
                            self:NotifyTriggered()
                        end
                    end
                end
            end
        },
    },

    APPROPRIATED =
    {
        name = "Appropriated",
        desc = "When this argument is destroyed, all cards are returned to {2}'s hand. {2} gains 2 {VULNERABILITY}.",
        desc_fn = function( self, fmt_str, minigame, widget )
            if widget and widget.PostCard then
                for i, card in ipairs( self.stolen_cards) do
                    widget:PostCard( card.id, card, minigame )
                end
            end
            return loc.format( fmt_str, self:GetOwnerName(), self:GetOpponentName() )
        end,

        max_resolve = 1,
        max_stacks = 1,

        modifier_type = MODIFIER_TYPE.BOUNTY,
        removed_sound = "event:/sfx/battle/cards/neg/appropriator_cardreleased",
        --sound = "event:/sfx/battle/cards/neg/create_argument/strawman",

        OnInit = function( self )
            self.stolen_cards = {}
            self:SetResolve( 1, MODIFIER_SCALING.LOW )
        end,

        GetStolenCount = function( self )
            return #self.stolen_cards
        end,

        AppropriateCard = function( self, card )
            table.insert( self.stolen_cards, card )
            card:RemoveCard()

            self:NotifyChanged()
           
            self.engine:BroadcastEvent( EVENT.CARD_STOLEN, card, self )
        end,

        OnBounty = function( self )
            for i, card in ipairs( self.stolen_cards ) do
                self.engine:BroadcastEvent( EVENT.CUSTOM, function( panel )
                    local slot_widget = panel:FindSlotWidget( self )
                    if slot_widget then
                        local w = panel.cards:CreateCardWidget( card )
                        local x, y = w.parent:TransformFromWidget( slot_widget, 0, 0 )
                        w:SetPos( x, y )
                    end
                end )
                card.show_dealt = false
                self.engine:InsertCard( card )
            end

            self.anti_negotiator:InceptModifier("VULNERABILITY", 2)
        end,
    },

    APPROPRIATED_plus =
    {
        -- Used to indicate multiple cards are picked; see APPROPRIATOR.
    },

    PERMIT =
    {
        name = "Permit",
        desc = "Whenever this argument takes damage, deal that much damage to a random opponent argument. Gains 2 resolve for each additional stack.",
        max_resolve = 0,

        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnSetStacks = function( self, old_stacks )
            if self.engine then
                local delta = self.stacks - old_stacks
                self:ModifyResolve( delta * 2 )
            end
        end,

        event_handlers = 
        {
            [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                if target == self then
                    self.engine:PushPostHandler( function()
                        self.min_persuasion = damage
                        self.max_persuasion =  damage
                        self.target_enemy = TARGET_ANY_RESOLVE
                        self.engine:AssignPrimaryTarget( self )

                        self:ApplyPersuasion()
                        self.target_enemy = nil
                        self.min_persuasion = nil
                        self.max_persuasion = nil
                    end )
                end
            end,
        },
    },
    PROPER_CHANNEL =
    {
        name = "Proper Channel",
        desc = "Increase by 1 whenever your core argument takes damage.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.max_persuasion)
        end,

        modifier_type = MODIFIER_TYPE.ARGUMENT,
        max_resolve = 4,

        min_persuasion = 1,
        max_persuasion = 1,

        target_enemy = TARGET_ANY_RESOLVE,

        OnSetStacks = function( self, old_stacks )
            self.min_persuasion = self.stacks
            self.max_persuasion = self.stacks
        end,

        OnApply = function( self )
            self:PrepareTurn()
        end,

        event_handlers = 
        {
            [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                if target.modifier_type == MODIFIER_TYPE.CORE and target.negotiator == self.negotiator and damage > defended then
                    self.negotiator:DeltaModifier(self, 1)
                end
            end,

            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                self:ApplyPersuasion()
            end
        },
    },

    DRUNK =
    {
        name = "Drunk",
        desc = "When destroyed, remove a random intent from {1}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.BOUNTY,

        max_resolve = 3,
        max_stacks = 1,

        OnBounty = function( self )
            local intents = self.negotiator:GetIntents()
            if #intents > 0 then
                self.negotiator:DismissIntent(intents[math.random(#intents)])
            end
        end,
    },

    ALL_BUSINESS =
    {
        name = "All Business",
        desc = "{1} gains {COMPOSURE {2}} for each Hostility card you draw.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, loc.cap( self:GetOwnerName() ), self.bonus )
        end,

        max_resolve = 1,
        max_stacks = 1,
        bonus = 1,

        sound = "event:/sfx/battle/cards/neg/create_argument/all_business",

        OnInit = function( self )
            if self.engine then
                local difficulty = self.engine:GetDifficulty()
                self:SetResolve( 1, MODIFIER_SCALING.MED )
                self.bonus = difficulty
            end
        end,

        event_handlers =
        {
            [ EVENT.DRAW_CARD ] = function( self, engine, card, start_of_turn )
                if card:IsFlagged( CARD_FLAGS.HOSTILE ) then
                    self.negotiator:DeltaComposure( self.bonus, self )
                    self:NotifyTriggered()
                end           
            end,
        },
    },

    TRIBUTE =
    {
        name = "Tribute",
        desc = "When destroyed, {2} gains: draw {1} extra {1*card|cards} at the start of your turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1, self:GetOpponentName() or "player" )
        end,

        modifier_type = MODIFIER_TYPE.BOUNTY,
        sound = "event:/sfx/battle/cards/neg/create_argument/tribute",

        OnInit = function( self )
            self:SetResolve( self.engine:GetDifficulty() )
        end,

        OnBounty = function( self )
            self.engine:GetPlayerNegotiator():AddModifier( "CARD_DRAW", 1 )
        end,
    },

    INDOCTRINATION =
    {
        name = "Indoctrination",
        desc = "When destroyed, gain 1 {DOMINANCE} per stack. At the end of each turn, gain 1 {INFLUENCE} and reduce this argument by 1.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        modifier_type = MODIFIER_TYPE.BOUNTY,
        sound = "event:/sfx/battle/cards/neg/create_argument/indoctrination",

        OnEndTurn = function( self, minigame )
            self.engine:GetPlayerNegotiator():AddModifier( "INFLUENCE", 1 )
            self.negotiator:DeltaModifier( self, -1 )
            self:NotifyTriggered()
        end,

        OnInit = function( self )
            self:SetResolve( (self.engine and self.engine:GetDifficulty() or 1)*2 )
        end,

        OnBounty = function( self )
            self.engine:GetPlayerNegotiator():AddModifier( "DOMINANCE", self.stacks )
        end,
    },

    CRAFTY =
    {
        name = "Crafty",
        desc = "Intents and target previews are hidden.",
        max_stacks = 1,
        modifier_type = MODIFIER_TYPE.ARGUMENT,
    
        flags = CARD_FLAGS.EXPEND,

        sound = "event:/sfx/battle/cards/neg/create_argument/crafty",

        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.MED )
        end,

        OnUnapply = function ( self )
            -- Need to trigger reveal for targets.
            self.engine:BroadcastEvent( EVENT.TARGETS_CHANGED )
        end,

        event_handlers =
        {
            [ EVENT.CALC_BOOLEAN ] = function( self, acc, key )
                if not self.negotiator:IsPlayer() and self.engine:GetTurns() > self.turn_applied then
                    if key == "HIDE_INTENT" then
                        acc:ModifyValue( true, self )
                    end
                end
            end
        }
    },

    CRAFTY_plus =
    {
        desc = "Intents and target previews are hidden.\n\nWhen this argument takes damage without being destroyed, {1} gains {ploy}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        event_handlers =
        {
            [ EVENT.CALC_BOOLEAN ] = function( self, acc, key )
                if not self.negotiator:IsPlayer() and self.engine:GetTurns() > self.turn_applied then
                    if key == "HIDE_INTENT" then
                        acc:ModifyValue( true, self )
                    end
                end
            end,

            [ EVENT.DELTA_RESOLVE ] = function( self, modifier, resolve, max_resolve, delta, source, params )
                if modifier == self and delta < 0 and resolve > 0 then
                    self:NotifyTriggered()
                    self.negotiator:CreateModifier( "ploy" )
                end
            end,
        },
    },

    MONEY_TALKS = 
    {
        name = "Money Talks",
        desc = "At the start of each turn, a random card in hand costs an additional {1#money} this turn.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.money )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/money_talks",
        fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

        money = 10,
        per_turn = true,
        max_stacks = 1,

        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnInit = function( self )
            self:SetResolve( 5, MODIFIER_SCALING.MED )
            self.card_costs = {}
        end,

        event_priorities =
        {
            [ EVENT.CALC_MONEY_COST ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers = 
        {
            [ EVENT.HAND_DRAWN ] = function( self, minigame )
                local card = minigame:GetHandDeck():PeekRandom( Negotiation.Card.IsPlayable )
                if card then
                    table.insert( self.card_costs, card )
                    -- card:NotifyTriggeredPre()
                    -- self:NotifyTriggered()
                    -- card:NotifyTriggeredPost()
                end
            end,

            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if self.per_turn and negotiator == self.anti_negotiator then
                    table.clear( self.card_costs )
                end
            end,

            [ EVENT.CALC_MONEY_COST ] = function ( self, acc, card )
                local count = table.count( self.card_costs, card )
                local cost = count * self.money
                if cost > 0 then
                    acc:AddValue( cost, self )
                    if not acc:IsPreview() then
                        self:NotifyTriggered()
                    end
                end
            end,
        },
    },

    MONEY_TALKS_plus =
    {
        desc = "At the start of each turn, a random card in hand costs an additional {1#money} this negotiation.",
        per_turn = false,
    },

    PENNY_PINCHER = 
    {
        name = "Penny Pincher",
        desc = "At the end of {1}'s turn, {INCEPT} {CATCHPENNY}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/money_talks",

        stacks = 1,

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers = 
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    if self.anti_negotiator:GetSlotCount() < self.anti_negotiator:GetMaxSlots() then
                        self.anti_negotiator:CreateModifier( "CATCHPENNY", 1, self )
                    end
                end
            end,
        },
    },

    CATCHPENNY =
    {
        name = "Catchpenny",
        desc = "At the start of your turn, double <b>Catchpenny</>. When destroyed, lose {1#money} per stack ({2#money} total).",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.money, self.money * self.stacks )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/frisk",
        fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

        money = 1,
        max_resolve = 2,

        event_handlers =
        {
            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator and minigame:GetTurns() > self.turn_applied + 1 then
                    negotiator:DeltaModifier( self, self.stacks )
                end
            end
        },
        OnBounty = function( self )
            if self.negotiator:IsPlayer() then
                self.engine:ModifyMoney( -self.money * self.stacks )
            end
        end,
    },

    GREEDY = 
    {
        name = "Greedy",
        desc = "When destroyed, lose {1#money} and you double your actions. Can only be targeted by cards.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.money )
        end,
        remove_fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

        sound = "event:/sfx/battle/cards/neg/create_argument/greedy",

        max_resolve = 3,
        money = 10,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.LOW )
        end,

        CanPlayCard = function( self, card, engine, target )
            if target == self and (card.negotiator ~= self.negotiator or card:IsAttack()) then
                return is_instance(card, Negotiation.Card)
            end
            return true
        end,
        OnBounty = function( self )
            if self.engine:GetMoney() >= self.money then
                self.engine:ModifyMoney( -self.money )
                self.engine:ModifyActionCount( self.engine:GetActionCount() )
            end
        end
    },

    SUSPICION =
    {
        name = "Suspicion",
        desc = "Diplomacy cards deal {1%+d} damage, with a minimum of 1 damage.\n\nReduce <b>Suspicion</b> by 1 at the beginning of {2}'s turn.",
        desc_fn = function( self, fmt_str, minigame )
            return loc.format(fmt_str, -self.malus, self:GetOwnerName() )
        end,

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.MED )
        end,

        malus = 1,

        modifier_type = MODIFIER_TYPE.ARGUMENT,
        sound = "event:/sfx/battle/cards/neg/create_argument/suspicion",

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if is_instance( source, Negotiation.Card ) and CheckBits( source.flags, CARD_FLAGS.DIPLOMACY ) then
                    local bonus = 
                    persuasion:AddPersuasion( - math.min(self.malus, persuasion.min_persuasion - 1), - math.min(self.malus, persuasion.max_persuasion - 1), self )
                end
            end,

            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    negotiator:RemoveModifier( self, 1 )
                end
            end,
        },
    },

    SUSPICION_plus =
    {
        malus = 2,
    },

    KINGPIN = 
    {
        name = "Kingpin",
        desc = "Intents gain 1 bonus damage for each stack of {INFLUENCE} owned by the opponent.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.card_dmg, self.return_dmg)
        end,

        max_delay = 2,
        modifier_type = MODIFIER_TYPE.ARGUMENT,

        sound = "event:/sfx/battle/cards/neg/create_argument/kingpin",

        OnInit = function( self )
            self:SetResolve( 6, MODIFIER_SCALING.MED )
        end,

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                    if self:GetTurnApplied() < self.engine:GetTurns() then
                        local influence = self.anti_negotiator:GetModifierStacks( "INFLUENCE" )
                        if influence > 0 then
                            persuasion:AddPersuasion( influence, influence, self )
                        end
                    end
                end
            end,
        },
    },

    SNARE =
    {
        name = "Snare",
        desc = "Must be targeted before anything else. This argument is removed next turn.",
        max_stacks = 1,
        max_instances = 1,
        modifier_type = MODIFIER_TYPE.ARGUMENT,
        force_target = true,

        sound = "event:/sfx/battle/cards/neg/create_argument/snare",

        OnInit = function( self )
            self:SetResolve( 0, MODIFIER_SCALING.MED )
        end,
        
        CanPlayCard = function( self, source, engine, target )
            -- Only verify forced targets if you are not targetting yourself
            if source:IsAttack() and target:GetNegotiator() == self.negotiator then
                if source.modifier_type == MODIFIER_TYPE.INCEPTION or source:GetNegotiator() ~= self.negotiator then
                    if not target.force_target then
                        return false, loc.format( LOC "CONVO_COMMON.MUST_TARGET", self:GetName() )
                    end
                end
            end

            return true
        end,

        OnBeginTurn = function( self )
            self.negotiator:RemoveModifier( self )
        end,
    },
    
    ECCENTRICITY =
    {
        name = "Eccentricity",
        desc = "All cards have a random cost between 0 and {1}.",
        desc_fn = function( self, fmt_str, widget, minigame )
            return loc.format( fmt_str, math.max( table.unpack( self.cost_table )))
        end,
        sound = "event:/sfx/battle/cards/neg/create_argument/eccentricity",

        modifier_type = MODIFIER_TYPE.ARGUMENT,
        max_stacks = 1,

        cost_table = { 0, 0, 1, 1, 2, 2 },

        OnInit = function( self )
            self:SetResolve( 4, MODIFIER_SCALING.MED )
            self:ReshuffleCosts()
        end,

        ReshuffleCosts = function( self )
            self.cost_values = shallowcopy( self.cost_table )
            table.shuffle( self.cost_values )
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_TURN ] = function( self, negotiator, card )
                if negotiator ~= self.negotiator then
                    self.randomized_costs = {}
                end
            end,

            [ EVENT.CALC_ACTION_COST ] = function( self, cost_acc, card, target )
                if self.owner ~= card.owner and self.randomized_costs then
                    local cost = self.randomized_costs[ card ]
                    if cost == nil then
                        cost = table.remove( self.cost_values )
                        self:ReshuffleCosts()
                        self.randomized_costs[ card ] = cost
                    end
                    cost_acc:ModifyValue( cost, self )
                end
            end,
        }
    },

    ECCENTRICITY_plus =
    {
        cost_table = { 0, 0, 1, 1, 2, 2, 3 },
    },

    ESCALATION =
    {
        name = "Escalation",
        desc = "When a Hostility card is played, increase stacks by 1.\nAdds {1%+d} damage to all intents.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/escalation",
        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.MED )
        end,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers =
        {
            [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                if card:IsFlagged( CARD_FLAGS.HOSTILE ) then
                    self.negotiator:AddModifier( self.id, 1 )
                end
            end,

            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner and is_instance( source, Negotiation.Card ) and self.engine:GetTurns() > self.turn_applied then
                    persuasion:AddPersuasion( self.stacks, self.stacks, self )
                end
            end,
        }
    },

    ESCALATION_plus =
    {
        desc = "When a Hostility card is played, increase stacks by 1 and gain 2 resolve.\nAdds {1%+d} damage to all intents.",
        event_handlers =
        {
            [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                if card:IsFlagged( CARD_FLAGS.HOSTILE ) then
                    self.negotiator:AddModifier( self.id, 1 )
                    self:ModifyResolve( 2, self )
                end
            end,

            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner and is_instance( source, Negotiation.Card ) and self.engine:GetTurns() > self.turn_applied then
                    persuasion:AddPersuasion( self.stacks, self.stacks, self )
                end
            end,
        }
    },

    TITHE =
    {
        name = "Tithe",
        desc = "When destroyed, restore {1} resolve for every opponent argument.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 1 )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/tithe",
        modifier_type = MODIFIER_TYPE.BOUNTY,
        resolve_bonus = 3,

        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.LOW )
        end,

        OnBounty = function( self )
            local count = 1 + self.negotiator:GetModifierCount( MODIFIER_TYPE.ARGUMENT )
            self.anti_negotiator:RestoreResolve( count * self.resolve_bonus, self )
        end,
    },

    IRRITABLE =
    {
        name = "Irritable",
        desc = "Gain 1 stack whenever a card is played.\n\nAt {1} stacks, {3} is dealt {2} resolve damage and this argument is removed.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.counter, self.resolve_loss or 1, self:GetOpponentName() )
        end,

        max_resolve = 3,

        counter = 10,

        sound = "event:/sfx/battle/cards/neg/create_argument/irritable",

        OnInit = function( self )
            if self.engine then
                self:SetResolve( 4 + self.engine:GetDifficulty() )
            end
            self.resolve_loss = self.engine:GetDifficulty() * 5
        end,

        event_handlers =
        {
            [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                if card:GetNegotiator() ~= self.negotiator and self.negotiator:FindCoreArgument() then
                    self.negotiator:AddModifier( self, 1 )

                    if self.stacks >= self.counter then
                        minigame:ApplyPersuasion( self, self.anti_negotiator, self.resolve_loss, self.resolve_loss )
                        self.negotiator:RemoveModifier( self )
                    end
                end
            end
        },
    },

    RUMOR_MONGER = 
    {
        name = "Rumormonger",
        desc = "When this argument is destroyed, permanently acquire {rumour}.",

        max_resolve = 5,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        sound = "event:/sfx/battle/cards/neg/create_argument/rumor_monger",
    },

    SCRUPLE =
    {
        name = "Scruple",
        desc = "When destroyed, {1} loses half their resolve.\nThis bounty lasts for {2} more {2*turn|turns}.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.stacks or 1 )
        end,

        sound = "event:/sfx/battle/cards/neg/create_argument/scruple",
        modifier_type = MODIFIER_TYPE.BOUNTY,

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.LOW )
        end,

        OnBounty = function( self )
            local resolve = self.negotiator:GetResolve() or 0
            resolve = math.floor( resolve / 2 )
            self.negotiator:AttackResolve( resolve, self )
        end,

        OnBeginTurn = function( self )
            self.negotiator:RemoveModifier( self, 1 )
        end,
    },

    CHASER =
    {
        name = "Chaser",
        desc = "Restore 10 resolve to {1}'s core argument. Add 2 {drunk_player} cards to your draw pile.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetOpponentName())
        end,

        max_resolve = 2,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        sound = "event:/sfx/battle/cards/neg/create_argument/apathetic",

        OnBounty = function( self )
            self.anti_negotiator:RestoreResolve( 10, self )
            local cards = {}
            for i = 1, 2 do
                table.insert( cards, Negotiation.Card( "drunk_player", self.engine:GetPlayer() ))
            end
            self.engine:DealCards( cards )
        end
    },

    -- Permanent arguments
    --Admiralty
    POWER_ABUSE = 
    {
        name = "Abuse of power",
        desc = "At the start of {1}'s turn, {INCEPT} a {PLANTED_EVIDENCE} argument if one does not already exist.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers = 
        {
            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local has_evidence = false
                    for i,modifier in self.anti_negotiator:ModifierSlots() do
                        if modifier.id == "PLANTED_EVIDENCE" then
                            has_evidence = true
                        end
                    end

                    if not has_evidence then
                        self.anti_negotiator:InceptModifier("PLANTED_EVIDENCE", self.engine:GetDifficulty() * 2 or 2, self )
                        self:NotifyTriggered()
                    end
                end
            end
        },
    },

    PLANTED_EVIDENCE =
    {
        name = "Planted Evidence",
        desc = "When this argument is destroyed, deal {1} damage to {2}'s core argument.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.stacks or 2, self:GetOwnerName() )
        end,

        max_resolve = 1,

        modifier_type = MODIFIER_TYPE.BOUNTY,

        sound = "event:/sfx/battle/cards/neg/create_argument/strawman",

        OnBounty = function( self )
            self.engine:ApplyPersuasion( self, self.negotiator, self.stacks, self.stacks )
        end,
    },

    -- Admiralty Investigators
    INVESTIGATOR = 
    {
        name = "Investigator",
        desc = "At the end of {1}'s turn, {INCEPT} {PLANTED_EVIDENCE} arguments until {2} has 2.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self:GetOpponentName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers = 
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local evidence_count = self.anti_negotiator:GetModifierInstances("PLANTED_EVIDENCE")

                    if evidence_count < 2 then
                        for i=1,2-evidence_count do
                            self.anti_negotiator:InceptModifier("PLANTED_EVIDENCE", self.engine:GetDifficulty() * 2 or 2, self, true )
                            self:NotifyTriggered()
                        end
                    end
                end
            end
        },
    },

    THOROUGH_SEARCH =
    {
        name = "Thorough Search",
        damage_tt = "UI.NEGOTIATION_PANEL.ARGUMENT_AOE_DAMAGE_TT", 
        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.MED )
        end,

        min_persuasion = 2,
        max_persuasion = 2,

        hide_stacks = true,
        max_stacks = 1,
        target_enemy = TARGET_ANY_RESOLVE,
        target_mod = TARGET_MOD.TEAM,
        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnBeginTurn = function( self, minigame )
            self:ApplyPersuasion()
        end,
    },

    --Merchants
    PROPOSITION = 
    {
        name = "Proposition",
        desc = "Add a <b>Kickback</b> card to your draw pile at the end of {1}'s turn.",
        desc_fn = function( self, fmt_str, minigame, widget )
            if widget and widget.PostCard and self.kickback then
                widget:PostCard( self.kickback.id, self.kickback, minigame )
            end
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        OnInit = function( self )
            self.kickback = Negotiation.Card( "kickback", self.anti_negotiator.agent )
            self.engine:GetDeck( DECK_TYPE.AI ):InsertCard( self.kickback )
        end,

        event_handlers = 
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local card = Negotiation.Card("kickback", minigame:GetPlayer()) 
                    minigame:InceptCard( card, self )
                end
            end
        },
    },

    INTERROGATE =
    {
        name = "Interrogate",
        sound = "event:/sfx/battle/cards/neg/create_argument/interrogate",

        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.MED )
            local difficulty = self.engine and self.engine:GetDifficulty() or 1
            self.min_persuasion = 1 + math.ceil(difficulty/2)
            self.max_persuasion = 3 + difficulty
        end,

        hide_stacks = true,
        max_stacks = 1,
        target_enemy = TARGET_ANY_RESOLVE,
        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnBeginTurn = function( self, minigame )
            self:ApplyPersuasion()
        end,
    },

    INTERROGATE_plus =
    {
        desc = "When this argument destroys another argument, restore {1} core argument resolve.",
        desc_fn = function (self, fmt_str)
            return loc.format( fmt_str, 3 )
        end,

        event_handlers =
        {
            [ EVENT.MODIFIER_REMOVED ] = function ( self, modifier, source )
                if source == self then
                    local core = self.negotiator:FindCoreArgument()
                    if core then
                        core:RestoreResolve( 3, self )
                    end
                end
            end
        }
    },

    --Laborers
    DEFENSIVE =
    {
        name = "Defensive",
        desc = "Gains {1} {COMPOSURE} whenever {2} creates a new argument.",
        desc_fn = function ( self, fmt_str )
            if self.engine then
                return loc.format( fmt_str, self.composure_gain + (self.comp_per_dif * self.engine:GetDifficulty()), self.engine:GetPlayerNegotiator():GetName())
            else
                return loc.format( fmt_str, self.composure_gain + (self.comp_per_dif * 1), "the opponent")
            end
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        composure_gain = 1,
        comp_per_dif = 1,

        event_handlers =
        {
            [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source )
                if modifier.owner ~= self.owner then
                    if (modifier.modifier_type == MODIFIER_TYPE.ARGUMENT or modifier.modifier_type == MODIFIER_TYPE.BOUNTY) then
                        self.negotiator:DeltaComposure( self.composure_gain + (self.comp_per_dif * self.engine:GetDifficulty()) )
                        self:NotifyChanged()
                    end
                end
            end
        }
    },

    -- --Foreman
    DELEGATOR = 
    {
        name = "Delegator",
        desc = "When your damage spills over against {1}'s arguments, it targets your core argument instead.",
        desc_fn = function ( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers = 
        {
            [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                if self.owner == modifier.owner and params.splashed_modifier == self.negotiator:FindCoreArgument() and (modifier.modifier_type == MODIFIER_TYPE.ARGUMENT or modifier.modifier_type == MODIFIER_TYPE.BOUNTY) then
                    params.splashed_modifier = self.anti_negotiator:FindCoreArgument()
                end
            end
        },
    },

    --Spree
    SHORT_FUSE =
    {
        name = "Short Fuse",
        desc = "Whenever you shuffle your deck, {1} gains {IMPATIENCE}.",
        desc_fn = function ( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers = 
        {
            [ EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
                self.negotiator:AddModifier("IMPATIENCE")
            end
        },
    },

    --Chemist
    PURPLE_HAZE = 
    {
        name = "Purple Haze",
        desc = "Whenever you play a card, draw a card and then discard a random card in your hand.",

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers =
        {
            [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                if card:GetNegotiator() == self.engine:GetPlayerNegotiator() then
                    if minigame:GetTurns() > 0 then

                        self.engine:DrawCards(1)

                        local card = self.engine:GetHandDeck():PeekRandom()
                        if card then
                            minigame:BroadcastEvent( EVENT.CUSTOM, function( screen, ... )
                                local w = screen:FindSlotWidget( self )
                                if w then
                                    local particles = w:AddChild( Widget.ParticleSystem() )
                                    particles:LoadParams( "card_discard_purple_haze" )
                                    particles:SetScale( 0.35, 0.35 )
                                    particles:StartCoroutine( function()
                                        while not particles:IsDone() do
                                            coroutine.yield()
                                        end
                                        particles:Remove()
                                    end )
                                end

                                AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/card_destroyed" )
                            end )

                            self.engine:DiscardCard( card )
                        end
                    end
                end
            end
        },
    },

    --Jakes
    DOUBLE_EDGE =
    {
        name = "Double Edge",
        desc = "Whenever {1} deals damage, they gain {COMPOSURE} equal to resolve lost.\nIf no resolve is lost, {1} takes {2} damage.",
        desc_fn = function ( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.resolve_loss)
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        resolve_loss = 3,

        target_self = TARGET_ANY_RESOLVE,

        OnBeginTurn = function( self, minigame )
            self.composure_gain = 0
        end,

        OnEndTurn = function( self, minigame )
            if self.composure_gain > 0 then
                local targets = {}
                for i, modifier in self.negotiator:ModifierSlots() do
                    if modifier:GetResolve() ~= nil then
                        table.insert( targets, {modifier=modifier, count=0} )
                    end
                end

                while self.composure_gain > 0 and #targets > 0 do
                    for i, target in ipairs(targets) do
                        if self.composure_gain > 0 then
                            target.count = target.count + 1
                            self.composure_gain = self.composure_gain - 1
                        end
                    end
                end

                for i, target in ipairs(targets) do
                    target.modifier:DeltaComposure( target.count, self)
                end
            end
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_ENEMY_TURN ] = function( self )
                self.active = true
            end,
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    self.active = false
                end
            end,
            [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                self.composure_gain = self.composure_gain or 0

                if self.active and source and source.negotiator == self.negotiator  then
                    if damage > defended then
                        local gain = math.min( target:GetResolve(), damage - defended)
                        self.composure_gain = self.composure_gain + gain
                    else
                        if self.negotiator:FindCoreArgument():GetShieldStatus() or CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.NO_CORE_RESOLVE) then
                            local targets = self.engine:CollectAllTargets(self)
                            table.arrayremove( targets, self.negotiator:FindCoreArgument() )
                            local edge_target = table.arraypick( targets )
                            if edge_target then
                                self.engine:ApplyPersuasion( target, edge_target, self.resolve_loss, self.resolve_loss )
                            end
                        else
                            self.engine:ApplyPersuasion( target, self.negotiator, self.resolve_loss, self.resolve_loss )
                        end
                    end
                end
            end,
        },
    },

   -- Hesh
    ZEAL =
    {
        name = "Zeal",
        desc = "When an argument of {1}'s is removed or destroyed, all other arguments gain {2} resolve.",
        desc_fn = function ( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.resolve_bonus)
        end,

        modifier_type = MODIFIER_TYPE.CORE,
        resolve_bonus = 1,

        OnInit = function( self )
            self.resolve_bonus = (self.engine and self.engine:GetDifficulty() or 1)
        end,

        event_handlers =
        {
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, card )
                if modifier.modifier_type == MODIFIER_TYPE.ARGUMENT and modifier.owner == self.owner then
                    for i, mod in self.negotiator:Modifiers() do
                        if mod ~= modifier and mod.modifier_type == MODIFIER_TYPE.ARGUMENT then
                            mod:ModifyResolve( self.resolve_bonus, self )
                        end
                    end
                end
            end,
        }
    },

    UNCERTAINTY =
    {
        name = "Impassive",
        desc = "Whenever {1}'s arguments are destroyed, splash damage is converted to {COMPOSURE} instead.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.CORE,
        
        alt_textures = {
            HESH_BOSS = engine.asset.Texture("negotiation/modifiers/uncertainty2.tex"),
            MERCENARY_BOSS = engine.asset.Texture("negotiation/modifiers/uncertainty3.tex")
        },
        
        OnInit = function(self)
            if self.alt_textures[self.owner:GetContentID()] then
                self.icon = self.alt_textures[self.owner:GetContentID()]
                self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
            end
        end,


        event_handlers =
        {
            [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                if modifier.owner == self.owner and params.splashed_modifier then
                    params.splashed_modifier = nil -- No splash damage.
                    self:DeltaComposure( math.abs(overflow), self )
                end
            end,
        }        
    },

    FEAR = 
    {
        name = "Fear",
        desc = "When destroyed, gives {1} {IMPATIENCE 1}. {1} will obtain {battle.DEFECT} when you enter battle.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.negotiator and self.negotiator:GetName() or "your opponent")
        end,

        modifier_type = MODIFIER_TYPE.BOUNTY,

        max_resolve = 6,

        OnBounty = function( self, card )
            if card then
                if self.result_table then
                    table.insert(self.result_table, 1)
                end
            end
            self.negotiator:AddModifier("IMPATIENCE")
            local mod = self.negotiator:CreateModifier("FEAR")
            mod.result_table = self.result_table
            mod:ModifyResolve( self.max_resolve - 2 )
        end,
    },


    --auction stuff 
    AUCTIONEER = 
    {
        name = "Auctioneer",
        desc = "Introduces items for bidding. There {1*is|are} {1} {1*item|items} left.\n\nYou will get to keep the first item that you dismiss and the negotiation ends.\n\n<#PENALTY>If another bidder dimisses an item, it is removed from play.</>",
        desc_fn = function(self, fmt_str )
            return loc.format(fmt_str, self.items_to_add and #self.items_to_add or 0)
        end,
        loc_strings =
        {
            AUCTION_OVER = [[
                Agent:
                    !clap
                    And that's all there is! The auction is over!
            ]],
        },

        modifier_type = MODIFIER_TYPE.CORE,

        SetItems = function (self, grafts, items)
            self.items_to_add = {}
            for k,v in ipairs(grafts) do
                table.insert(self.items_to_add, {graft_id = v} )
            end

            for k,v in ipairs(items) do
                table.insert(self.items_to_add, {item_id = v} )
            end
        end,

        GetItemCount = function(self)
            return #self.items_to_add
        end,

        AddAuctionItem = function( self )
            if self.items_to_add and #self.items_to_add > 0 then
                local idx = math.random(#self.items_to_add)
                local item = table.remove(self.items_to_add, idx)

                local modifier = self.negotiator:CreateModifier( "auction_item", 1, self )
                if item.item_id then
                    modifier:SetItem(item.item_id)
                elseif item.graft_id then
                    modifier:SetGraft(item.graft_id)
                end
            end
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_PLAYER_TURN ] = function( self, result, source )
                
                local has_bid = false
                for k,v in pairs(self.engine:GetHandDeck().cards) do
                    if v.id == "bid" then
                        has_bid = true
                    end
                end
                if not has_bid then
                    local card = Negotiation.Card( "bid", self.engine:GetPlayer() )
                    self.engine:DealCards( {card}, self.engine:GetHandDeck() )
                end
            end,
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                if modifier.id == "auction_item" then
                    
                    local other_bidder = source and source.id and source.id == "BIDDER"
                    if other_bidder then
                        local items = 0, 0
                        
                        for i, mod in self.negotiator:Modifiers() do
                            if mod.id == "auction_item" then
                                items = items + 1   
                            end
                        end
                        if self.engine:GetResult() == nil and items == 0 and #self.items_to_add == 0 then
                            self.engine:BroadcastEvent( EVENT.DIALOG, self.def:GetLocalizedString( "AUCTION_OVER" ))
                            self.engine:Lose()
                        end
                    else
                        if modifier.item_id then
                            self.won_item = self.won_item or {}
                            table.insert(self.won_item, modifier.item_id )
                        end
                        if modifier.graft_id then
                            self.won_graft = self.won_graft or {}
                            table.insert(self.won_graft, modifier.graft_id)
                        end
                        if self.engine:GetResult() == nil then
                            self.engine:Win()
                        end
                    end
                end
            end,
        },
    },

   
    BIDDER = 
    {
        name = "Anonymous Bid",
        desc = "A bid from another patron. When destroyed, another Anonymous Bid is created to replace it.\n\n<#PENALTY>If this argument kills an auction item, it is removed from play!</>",
        
        OnBeginTurn = function( self, minigame )
            self:ApplyPersuasion()
            self.negotiator:DeltaModifier( self, 2 )
            self:NotifyTriggered()
        end,

        min_resolve = 14,
        max_resolve = 20,

        OnDealt = function( self, engine )
            local difficulty = self.engine:GetTurns()
            self.min_resolve = self.min_resolve + difficulty
            self.max_resolve = self.max_resolve + difficulty
        end,

        OnBounty = function( self )
            self.negotiator:CreateModifier( "BIDDER" )
        end,

        EvaluateTargetWeight = function( self, target, targets )
            if target.negotiator == self.negotiator then
                return #targets
            else
                return 1
            end
        end,

        min_persuasion = 5,
        max_persuasion = 5,
        
        modifier_type = MODIFIER_TYPE.ARGUMENT,
        target_self = TARGET_FLAG.BOUNTY,
        target_enemy = TARGET_ANY_RESOLVE,

        max_stacks = 1,
    
        OnEndTurn = function( self, minigame )
            self.min_persuasion = self.min_persuasion + 5
            self.max_persuasion = self.max_persuasion + 5
            self:NotifyChanged()
        end,
    }, 
    

    auction_item =
    {
        name = "Auction Item",

        desc = "When you dismiss this argument, gain {{2}.{1}} and end the auction.\n\n<#PENALTY>If another bidder dismisses this argument, it is removed from play.</>",
        
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.item_id or self.graft_id, self.item_id and "card" or "graft")
        end,

        AUCTION_TIER_RESOLVE = {
            [1] = 40,
            [2] = 50,
            [3] = 60
        },
        max_resolve = 15,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        --sound = "event:/sfx/battle/cards/neg/create_argument/discovers_cave",
        SetGraft = function(self, id)
            self.custom_name = loc.format("{1#graft}", id)
            self.icon = Content.GetGraft(id).auction_img or Content.GetGraft(id).img
            self.graft_id = id
            local tier = Content.GetGraft(id).auction_tier or 2
            local res = self.AUCTION_TIER_RESOLVE[tier or 2] or self.AUCTION_TIER_RESOLVE[2]
            self:SetResolve(math.round(res*(.6+math.random()*.4)))
            self:NotifyChanged()
            self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
        end,

        SetItem = function(self, id)
            self.custom_name = loc.format("{1#card}", id)
            self.icon = (Content.GetBattleCard(id) or Content.GetNegotiationCard(id)).auction_img or (Content.GetBattleCard(id) or Content.GetNegotiationCard(id)).icon
            self.item_id = id
            local tier = (Content.GetBattleCard(id) or Content.GetNegotiationCard(id)).auction_tier or 2
            local res = self.AUCTION_TIER_RESOLVE[tier or 2] or self.AUCTION_TIER_RESOLVE[2]
            self:SetResolve(math.round(res*(.6+math.random()*.4)))
            self:NotifyChanged()
            self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
        end,

        OnBounty = function( self )
            --self.engine.start_params.quest.param.have_codes = (0 or self.engine.start_params.quest.param.have_codes) + 1
        end,

        OnRemoveWidget = function( self, widget, panel )
            --panel.encounter:Dialog( self.dialog )
            --panel:Delay( 1.0 )
        end,
    },

    FLEKFIS_WEAKNESS = 
    {
        name = "Flekfis' Weaknesses",
        desc = "Whenever a Weakness argument is destroyed, you gain a powerful card for the ensuing battle. After gaining 4 cards, the negotiation ends. {1} {1*card|cards} remaining.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.cards_left or 0)
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        SetCards = function(self, card_ids, cards_won)
            self.cards_to_create = {}
            for i,id in ipairs(card_ids) do
                table.insert(self.cards_to_create, id)
            end
            self.cards_left = 4
            self.cards_won = cards_won
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_PLAYER_TURN ] = function( self, result, source )
                if not self.added_cards and self.cards_to_create then
                    for i,card_id in ipairs(self.cards_to_create) do
                        local mod = self.negotiator:CreateModifier("weakness_argument")
                        mod:SetCard(card_id, self)
                    end
                    self.added_cards = true
                end
            end,
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                if self.engine:GetResult() == nil and modifier.id == "weakness_argument" then
                    if self.cards_left == 1 then
                        self.engine:Win()              
                    else 
                        self.cards_left = self.cards_left - 1
                    end
                end
            end,
        },
    },

    NPC_DOMINANT =
    {
        name = "Dominant",
        --sound = "event:/sfx/battle/cards/neg/create_argument/interrogate",
        desc = "Attacks deal an additional {1%+d} damage. Remove 1 stack at the start of {2}'s turn.",
        --icon = "negotiation/npc_dominant.tex",
        desc_fn = function( self, fmt_str, minigame )
            return loc.format( fmt_str, self.stacks or 1, self:GetOwnerName() )
        end,
        modifier_type = MODIFIER_TYPE.ARGUMENT,
        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.MED )
        end,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                    persuasion:AddPersuasion( self.stacks, self.stacks, self )
                end
            end,
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    self.negotiator:RemoveModifier( self, 1)
                end
            end,

        }        
    },

    NPC_INFLUENTIAL =
    {
        name = "Influential",
        --sound = "event:/sfx/battle/cards/neg/create_argument/interrogate",
        
        desc = "If stacks reach {1}: Attacks deal {2%+d} damage. Add 1 stack at the start of {3}'s turn.",
        --icon = "negotiation/npc_influential.tex",
        bonus = 8,
        thresh = 4,
        desc_fn = function( self, fmt_str, minigame )
            return loc.format( fmt_str, self.thresh, self.bonus, self:GetOwnerName() )
        end,

        modifier_type = MODIFIER_TYPE.ARGUMENT,
        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.HIGH )
            self.icon = self.alt_textures.INACTIVE
        end,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        alt_textures = {
            INACTIVE = engine.asset.Texture("negotiation/modifiers/influential.tex"),
            ACTIVE = engine.asset.Texture("negotiation/modifiers/influential_active.tex"),
        },

        OnSetStacks = function( self, old_stacks )
            local was_active = old_stacks >= self.thresh
            local is_active = self.stacks >= self.thresh
            if was_active ~= is_active then
                self.icon = is_active and self.alt_textures.ACTIVE or self.alt_textures.INACTIVE
                self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
                self:NotifyTriggered()
            end
        end,

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if self.stacks >= self.thresh and source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                    persuasion:AddPersuasion( self.bonus, self.bonus, self )
                end
            end,
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                
                if negotiator == self.negotiator and self.stacks < self.thresh then
                    self.negotiator:AddModifier( self, 1)
                end
            end,

        }        
    },

    --[[NPC_SAL_INFLUENCE =
    {
        name = "Influence",
        --sound = "event:/sfx/battle/cards/neg/create_argument/interrogate",
        
        OnInit = function( self )
            self:SetResolve( 1, MODIFIER_SCALING.MED )
        end,

        OnBeginTurn = function( self, minigame )
            self.negotiator:RemoveModifier(self, 1)
        end,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                    persuasion:AddPersuasion( self.stacks, self.stacks, self )
                end
            end,
        }        
    },--]]


    ARINT_WEAKNESS = 
    {
        name = "Arint's Weaknesses",
        desc = "Whenever a Weakness argument is destroyed, you gain a powerful card for the ensuing battle. After gaining 4 cards, the negotiation ends. {1} {1*card|cards} remaining.",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.cards_left or 0)
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        SetCards = function(self, card_ids, cards_won)
            self.cards_to_create = {}
            for i,id in ipairs(card_ids) do
                table.insert(self.cards_to_create, id)
            end
            self.cards_left = 4
            self.cards_won = cards_won
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_PLAYER_TURN ] = function( self, result, source )
                if not self.added_cards and self.cards_to_create then
                    for i,card_id in ipairs(self.cards_to_create) do
                        local mod = self.negotiator:CreateModifier("weakness_argument")
                        mod:SetCard(card_id, self)
                    end
                    self.added_cards = true
                end
            end,
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                if self.engine:GetResult() == nil and modifier.id == "weakness_argument" then
                    if self.cards_left == 1 then
                        self.engine:Win()              
                    else 
                        self.cards_left = self.cards_left - 1
                    end
                end
            end,
        },
    },

    weakness_argument =
    {
        name = "Weakness",
        desc = "When you dismiss this argument, gain a {{1}} card in battle.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self.card_id or "")
        end,

        max_resolve = 30,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        SetCard = function(self, id, core)
            self.custom_name = loc.format("{1#card}", id)
            self.icon = engine.asset.Texture(Content.GetBattleCard(id).argument_icon, true)
            self.core = core
            self.card_id = id
            self:NotifyChanged()
            self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
        end,

        OnBounty = function( self )
            table.insert(self.core.cards_won, self.card_id)
            local mod = self.negotiator:CreateModifier("weakness_argument")
            mod:SetCard(self.card_id, self.core)
        end,
    },

    UNSHAKABLE = 
    {
        name = "Unshakable",
        desc = "At the end of every turn, lose {INFLUENCE 1} and {DOMINANCE 1}.",

        modifier_type = MODIFIER_TYPE.CORE,

        event_handlers =
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if negotiator:HasModifier("INFLUENCE") then
                    negotiator:RemoveModifier("INFLUENCE", 1, self)
                end
                if negotiator:HasModifier("DOMINANCE") then
                    negotiator:RemoveModifier("DOMINANCE", 1, self)
                end
            end
        },
    },

    wrath_of_hesh =
    {
        name = "Wrath of Hesh",
        damage_tt = "UI.NEGOTIATION_PANEL.ARGUMENT_AOE_DAMAGE_TT", 

        max_stacks = 1,
        hide_stacks = true,
        modifier_type = MODIFIER_TYPE.ARGUMENT,
        target_enemy = TARGET_ANY_RESOLVE,
        target_mod = TARGET_MOD.TEAM,

        sound = "event:/sfx/battle/cards/neg/create_argument/wrath_of_hesh",

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.MED )
            self.min_persuasion = math.floor((self.engine and self.engine:GetDifficulty() or 1) / 2) + 1
            self.max_persuasion = (self.engine and self.engine:GetDifficulty() or 1) + 1
        end,

        OnBeginTurn = function( self, minigame )
            self:ApplyPersuasion()
            minigame:AssignPrimaryTarget( self )
        end,
    },


    prayer_of_hesh =
    {
        name = "Prayer of Hesh",
        desc = "Add {1} resolve and {COMPOSURE {2}} to a friendly argument at the start of {3}'s turn.",
        desc_fn = function( self, fmt_str )
            local bonus = self.bonus or 0
            return loc.format( fmt_str, bonus * (self.stacks or 1), bonus * (self.stacks or 1), self:GetOwnerName()) 
        end,

        modifier_type = MODIFIER_TYPE.ARGUMENT,
        sound = "event:/sfx/battle/cards/neg/create_argument/prayer_of_hesh",

        OnInit = function( self )
            self:SetResolve( 2, MODIFIER_SCALING.LOW )
            if self.owner and self.owner:IsPlayer() then
                self.bonus = 2
            elseif self.engine then
                self.bonus = math.ceil(self.engine:GetDifficulty()/2)
            else
                self.bonus = 1
            end
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local targets = {}
                    for i, modifier in self.negotiator:ModifierSlots() do
                        if modifier:GetResolve() ~= nil then
                            table.insert( targets, modifier )
                        end
                    end
                    local target = table.arraypick( targets )
                    if target then
                        target:ModifyResolve( self.bonus * self.stacks, self )
                        target:DeltaComposure( self.bonus * self.stacks, self )
                        self:NotifyTriggered() 
                    end
                end
            end
        }
    },

    prayer_of_hesh_plus =
    {
        desc = "Add {1} resolve and {SHIELDED|Shield} to a friendly argument at the start of {2}'s turn.",
        desc_fn = function( self, fmt_str )
            local bonus = self.bonus or 0
            return loc.format( fmt_str, bonus * (self.stacks or 1), self:GetOwnerName()) 
        end,

        ShieldArgument = function( self, target )
            if self.last_shield and self.last_shield:IsApplied() then
                self.last_shield:SetShieldStatus( nil )
            end

            if target and target:IsApplied() then
                target:SetShieldStatus( true )
                self.last_shield = target
            end
        end,

        OnUnapply = function( self )
            self:ShieldArgument( nil )
        end,

        event_handlers =
        {
            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local targets = {}
                    for i, modifier in self.negotiator:ModifierSlots() do
                        if modifier:GetResolve() ~= nil then
                            table.insert( targets, modifier )
                        end
                    end

                    local target = table.arraypick( targets )
                    if target then
                        target:ModifyResolve( self.bonus * self.stacks, self )
                        self:NotifyTriggered()
                    end

                    self:ShieldArgument( target )
                end
            end,
        }
    },

    ROWDY_PATRON = 
    {
        name = "Rowdy Patron",
        desc = "Whenever one of {1}'s arguments is destroyed, {1} deals {2} damage to your core argument.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.damage_amt )
        end,

        modifier_type = MODIFIER_TYPE.CORE,

        alt_textures = {
            BRUT = engine.asset.Texture("negotiation/modifiers/rowdy_patron_brut.tex"),
            WEEZIL = engine.asset.Texture("negotiation/modifiers/rowdy_patron_weezil.tex")
        },

        damage_amt = 2,
        
        OnInit = function(self)
            if self.alt_textures[self.owner:GetContentID()] then
                self.icon = self.alt_textures[self.owner:GetContentID()]
                self.engine:BroadcastEvent( negotiation_defs.EVENT.UPDATE_MODIFIER_ICON, self)
            end
            if GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ARGUMENT_PLUS )  then
                self.damage_amt = self.damage_amt + 2
            end
        end,


        event_handlers =
        {
            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                if modifier.negotiator == self.negotiator and source and source ~= modifier then
                    self.anti_negotiator:AttackResolve(self.damage_amt, self)
                end
            end
        }  
    },

    STUBBORN =
    {
        name = "Stubborn",
        desc = "At the start of {1}’s turn, remove this argument and apply {2} composure to all friendly arguments.",
        desc_fn = function( self, fmt_str )
            return loc.format( fmt_str, self:GetOwnerName(), self.composure_amt )
        end,

        modifier_type = MODIFIER_TYPE.ARGUMENT,

        max_resolve = 10,
        composure_amt = 6,
        target_self = TARGET_ANY_RESOLVE,

        event_handlers =
        {
            [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                if negotiator == self.negotiator then
                    local options = minigame:CollectAllTargets( self )
                    for i,arg in ipairs(options) do
                        arg:DeltaComposure(self.composure_amt, self)
                    end
                    self.negotiator:RemoveModifier(self, self.stacks, self)
                end
            end
        },
    },
}

-------------------------------------------------------------------------------


for i, id, carddef in sorted_pairs( CARDS ) do
    if not CheckAnyBits( carddef.flags or 0, CARD_FLAGS.STATUS | CARD_FLAGS.MANIPULATE | CARD_FLAGS.DIPLOMACY | CARD_FLAGS.HOSTILE ) then
        carddef.flags = (carddef.flags or 0) | CARD_FLAGS.OPPONENT
    end
    carddef.rarity = carddef.rarity or CARD_RARITY.UNIQUE
    carddef.cost = (carddef.cost or 0)
    carddef.series = carddef.series or CARD_SERIES.NPC

    Content.AddNegotiationCard( id, carddef )
end

for i, id, def in sorted_pairs( MODIFIERS ) do
    Content.AddNegotiationModifier( id, def )
end
