local negotiation_defs = require "negotiation/negotiation_defs"
local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT

local CARDS =
{
    drunk_player = 
    {
        name = "Slurred Speech",
        flavour = "'I’m fiiiii-iiiine.'",
        remove_on_rest = true,
        rarity = CARD_RARITY.BASIC,
        cost = 2,       
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME | CARD_FLAGS.SLEEP_IT_OFF,
    },

    gassy = 
    {
        name = "Gassy",
        rarity = CARD_RARITY.BASIC,
        flavour = "'It wasn’t me.'",
        remove_on_rest = true,
        cost = 1,       
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME | CARD_FLAGS.SLEEP_IT_OFF,
    },

    aggressive = 
    {
        name = "Aggressive",
        desc = "While this is in your hand, gain 1 less {COMPOSURE} from Manipulate cards.",
        flavour = "'Alright already! Forget I even brought it up!'",

        cost = 1,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME,

        event_handlers =
        {
            [ EVENT.CALC_DELTA_COMPOSURE ] = function( self, composure_acc, target, source )
                if is_instance( source, Negotiation.Card ) and source.negotiator == self.negotiator and source:IsFlagged( CARD_FLAGS.MANIPULATE ) and composure_acc.value > 0 then
                    composure_acc:AddValue( -1, self )
                end
            end,
        },
    },


    coward = 
    {
        name = "Coward",
        desc = "While this is in your hand, Hostility cards deal no damage.",
        flavour = "'Yeah, I’m not buying that tough guy attitude...'",
        rarity = CARD_RARITY.BASIC,
        cost = 2,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_SETTOR,
        },

        event_handlers = 
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source.owner == self.owner then
                    if is_instance( source, Negotiation.Card ) and CheckBits( source.flags, CARD_FLAGS.HOSTILE ) then
                        persuasion:ModifyPersuasion(0, 0, self )
                    end
                end
            end,
        }
    },


    unconvincing = 
    {
        name = "Unconvincing",
        flavour = "'Come on! Please?'",
        desc = "At the end of your turn, all other cards in your hand lose {1} XP.",
        
        desc_fn = function(self, fmt_str, minigame)
            return loc.format(fmt_str, self.xp_damage)
        end,
        
        rarity = CARD_RARITY.BASIC,
        xp_damage = 4,
        cost = 0,

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME | CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.REPLENISH,

        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                self:NotifyTriggeredPre()
                for i, card in minigame:GetHandDeck():Cards() do
                    card:RemoveXP(self.xp_damage)
                end
                self:NotifyTriggeredPost()
                self:TransferCard( minigame.trash_deck )
                self:Consume()
            end
        },
    },



    spooked = 
    {
        name = "Spooked",
        flavour = "'Did you hear something?'",
        remove_on_rest = true,
        rarity = CARD_RARITY.BASIC,
        desc = "When drawn, discard a random card.",
        cost = 2,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME | CARD_FLAGS.SLEEP_IT_OFF,

        event_handlers = 
        {
            [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                if card == self then
                    local card = self.engine:GetHandDeck():PeekRandom(function(card) return card.id ~= "spooked" end)
                    if card then
                        --self:NotifyTriggered()
                        self.engine:DiscardCard( card )
                    end
                end
            end
        },

    },


    killer = 
    {
        name = "Killer",
        desc = "While this is in your hand, gain {DOMINANCE 1}.",
        flavour = "'Dynamite with a laser beam!'",
        murder_card = true,
        must_loot = true,

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.REPLENISH,

        deck_handlers = ALL_DECKS,

        event_handlers =
        {
            [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                if card == self then
                    if target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND and not (source_deck and source_deck:GetDeckType() == DECK_TYPE.IN_HAND) then
                        self.negotiator:AddModifier( "DOMINANCE", 1 )

                    elseif source_deck and source_deck:GetDeckType() == DECK_TYPE.IN_HAND and not (target_deck and target_deck:GetDeckType() == DECK_TYPE.IN_HAND) then
                        self.negotiator:RemoveModifier( "DOMINANCE", 1 )
                    end
                end
            end,
        }
    },


    take_supporters = 
    {
        name = "Receive Support",
        flavour = "'I’ll take them.'",
        desc_fn = function(self, fmt_str, minigame)
            local allies
            if self.engine then
                allies = {}
                for i, modifier in self.engine:GetPlayerNegotiator():Modifiers() do
                    if modifier.id == "FIGHT_ALLY_WON" and modifier.ally_agent then
                        table.insert( allies, modifier.ally_agent )
                    end
                end

                table.sort(allies, function(a, b) return a:GetContentID() < b:GetContentID() end)
            end
            
            if allies and #allies > 0 then
                return loc.format(LOC"UI.CARDS.TAKE_SUPPORTERS_NONEMPTY", allies)
            else
                return LOC"UI.CARDS.TAKE_SUPPORTERS_EMPTY"
            end
        end,

        flags = CARD_FLAGS.MANIPULATE,
        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        
        OnPostResolve = function( self, minigame, targets )
            minigame:Win()
        end,
        
        CanPlayCard = function ( self, card, engine, target )
            if not self.engine:GetPlayerNegotiator():FindModifier("FIGHT_ALLY_WON") then
                return false, "You need to convince allies first"
            end
            return true
        end,

    },    

    insult = 
    {
        name = "Insult",
        flavour = "'Go cram yourself up an Oshnu.'",
        desc = "{EXPEND} all non-Hostility cards in hand.\nGain 1 {DOMINANCE} per card expended.",

        min_persuasion = 1,
        max_persuasion = 5,
        
        desc_fn = function(self, fmt_str, minigame )
            return loc.format(fmt_str, self.incept_amount )
        end,
        --quip = "insult",
        incept_amount = 3,
        flags = CARD_FLAGS.HOSTILE,
        rarity = CARD_RARITY.UNIQUE,
        cost = 1,

        OnPostResolve = function( self, minigame, targets )
            local cards = {}
            for i, card in minigame:GetHandDeck():Cards() do
                if not card:IsFlagged( CARD_FLAGS.HOSTILE ) then
                    table.insert( cards, card )
                end
            end
            if #cards > 0 then
                for i, card in ipairs( cards ) do
                    minigame:ExpendCard( card )
                end
                self.negotiator:AddModifier( "DOMINANCE", #cards, self )
            end
        end,
    },

    intimidate = 
    {
        name = "Intimidate",
        flavour = "'Oh this isn’t me angry, you’ll know when I get angry.'",
        desc = "At the end of your turn, {INCEPT} 2 {intimidated}.",
        flags = CARD_FLAGS.HOSTILE | CARD_FLAGS.UNPLAYABLE,
        rarity = CARD_RARITY.UNIQUE,
        cost = 0,

        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                self:NotifyTriggeredPre()
                self.anti_negotiator:AddModifier("intimidated", 2, self )
                self:NotifyTriggeredPost()
            end
        },
        
        modifier = 
        {
            id = "intimidated",
            name = "Intimidated",
            desc = "Hostility cards deal an additional {3%+d} damage. Remove 1 stack when you play a non-Hostility card.",

            desc_fn = function(self, fmt_str, minigame )
                local target = self.anti_negotiator or ( minigame and minigame:GetPlayerNegotiator() )
                return loc.format(fmt_str, self.bounty_damage, target and target:GetName(), self.stacks or 1 )
            end,

            bounty_damage = 10,
            modifier_type = MODIFIER_TYPE.INCEPTION,

            event_priorities =
            {
                [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
            },

            event_handlers = 
            {
                [ EVENT.POST_RESOLVE ] = function(self, minigame, card)
                    if self.owner ~= card.owner and not card:IsFlagged( CARD_FLAGS.HOSTILE ) then
                        self.negotiator:DeltaModifier(self, -1)
                        self:NotifyTriggered()
                    end
                end,

                [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                    if self.owner ~= source.owner then
                        if is_instance( source, Negotiation.Card ) and CheckBits( source.flags, CARD_FLAGS.HOSTILE ) then
                            persuasion:AddPersuasion(self.stacks, self.stacks, self )
                        end
                    end
                end,
            }

        },

    },

    bounty_hunter = 
    {
        name = "Bounty Hunter",
        desc = "{INCEPT} {1} {BOUNTY_REWARD}",
        flavour = "'Have you seen this ridiculous idiot?'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.num_to_incept)
        end,
        cost = 1,
        num_to_incept = 1,
        rarity = CARD_RARITY.UNIQUE,
        flags = CARD_FLAGS.MANIPULATE,
        OnPostResolve = function( self, minigame, targets )
            for k = 1, self.num_to_incept do
                self.anti_negotiator:CreateModifier("BOUNTY_REWARD", 1, self )
            end
        end
    },

    bounty_hunter_plus = 
    {
        name = "Pale Bounty Hunter",
        cost = 0,
    },

    bounty_hunter_plus2 = 
    {
        name = "Double-Barreled Bounty Hunter",
        num_to_incept = 2,
    },

    known_thug =
    {
        name = "Known Thug",
        desc = "Deal 1 resolve damage to a random opponent argument whenever you draw a card.",
        flavour = "'I’ve seen him around. Avoid him if you can.'",
        murder_card = true,
        must_loot = true,

        cost = 0,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.UNPLAYABLE,
        target_enemy = TARGET_ANY_RESOLVE,

        event_handlers =
        {
            [ EVENT.DRAW_CARD ] = function( self, engine, card, start_of_turn )
                if card.negotiator == self.negotiator and self ~= card then
                    self.min_persuasion, self.max_persuasion = 1,1
                    self.engine:ApplyPersuasion( self )
                    self.min_persuasion, self.max_persuasion = nil, nil
                end
            end,

        },
    },

    reputation_preceding = 
    {
        name = "Reputation Preceding",
        desc = "At the end of your turn, {INCEPT} {apprehensive 2} and {CONSUME} this card.",
        flavour = "'Oh we’ve heard about you.'",
        murder_card = true,
        must_loot = true,

        rarity = CARD_RARITY.BASIC,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.UNPLAYABLE,
        cost = 0,


        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                self:NotifyTriggeredPre()
                self.anti_negotiator:AddModifier("apprehensive", 2, self )
                self:NotifyTriggeredPost()
                self:TransferCard( minigame.trash_deck )
                self:Consume()
            end
        },

        modifier = 
        {
            id = "apprehensive",
            name = "Apprehensive",
            desc = "Apply <#HILITE>{1}</> {COMPOSURE} to a random argument at the end of turn.",

            desc_fn = function(self, fmt_str, minigame )
                return loc.format(fmt_str, self.stacks or 1 )
            end,

            max_resolve = 3,
            modifier_type = MODIFIER_TYPE.ARGUMENT,

            event_handlers = 
            {
                [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        local targets = self.engine:CollectAlliedTargets(self.negotiator)
                        if #targets > 0 then
                            local target = targets[math.random(#targets)]
                            target:DeltaComposure(self.stacks, self)
                        end
                    end
                end,
            }
        },    
    },

    dangerous = 
    {
        name = "Dangerous",
        desc = "While this is in your hand, Hostility cards deal 1 bonus damage and Diplomacy cards deal 1 less damage.",
        flavour = "'Hold on tight.'",
        murder_card = true,
        must_loot = true,

        cost = 2,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME,

        event_priorities =
        {
            [ EVENT.CALC_PERSUASION ] = EVENT_PRIORITY_ADDITIVE,
        },

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function ( self, source, persuasion, minigame )
                    if source.owner == self.owner and is_instance( source, Negotiation.Card ) then
                         if CheckBits( source.flags, CARD_FLAGS.DIPLOMACY ) then
                            persuasion:AddPersuasion(-1, -1, self )
                        elseif CheckBits( source.flags, CARD_FLAGS.HOSTILE ) then
                            persuasion:AddPersuasion(1, 1, self )
                        end
                    end
                end,
        }
    },

    unstable = 
    {
        name = "Unstable",
        flavour = "'Anything will set him off, I’m getting out of here.'",
        desc = "Draw a card.\nPlay it with a random target.",
        murder_card = true,
        must_loot = true,

        cost = 0,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,

        OnPostResolve = function( self, minigame, targets )
            local cards = minigame:DrawCards(1)
            local card = cards[1]
            if card then
                minigame:PlayCard( card, nil, 1.0 )
            end
        end,
    },

    -- Authority (Foreman Boon) cards
    by_the_book =
    {
        name = "By The Book",
        desc = "Cannot target core arguments.",
        flavour = "'Clear your schedule.'",
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.BASIC,
        series = CARD_SERIES.GENERAL,
        cost = 1,
        min_persuasion = 8,
        max_persuasion = 8,
        min_scaling = { 8, 10, 10, 12},
        max_scaling = { 8, 10, 10, 12},
        ScaleToDifficulty = function( self, minigame )
            self.min_persuasion = self.min_scaling[math.min(#self.min_scaling, minigame:GetDifficulty())]
            self.max_persuasion = self.max_scaling[math.min(#self.max_scaling, minigame:GetDifficulty())]
        end,
        CanTarget = function ( self, target )
            if is_instance( target, Negotiation.Modifier ) and target.modifier_type == MODIFIER_TYPE.CORE then
                return false, CARD_PLAY_REASONS.INVALID_TARGET
            end
            return true
        end,
    },

    proper_channels = 
    {
        name = "Proper Channels",
        desc = "Gain {1} {PROPER_CHANNEL}.",
        flavour = "'Hey back up, we are doing this the right way or not at all.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.count)
        end,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.BASIC,
        series = CARD_SERIES.GENERAL,

        cost = 1,

        count = 2,
        count_scaling = { 2, 2, 3, 4 },

        ScaleToDifficulty = function( self, minigame )
            self.count = self.count_scaling[math.min(#self.count_scaling, minigame:GetDifficulty())]
        end,

        OnPostResolve = function( self, minigame, targets )
            self.negotiator:AddModifier("PROPER_CHANNEL", self.count, self)
        end,
    },

    permits =
    {
        name = "Permits",
        desc = "Gain {1} {PERMIT}.",
        flavour = "'Papers please.'",
        desc_fn = function( self, fmt_str )
            return loc.format(fmt_str, self.count)
        end,
        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.BASIC,
        series = CARD_SERIES.GENERAL,

        cost = 1,

        count = 2,
        count_scaling = { 1, 2, 2, 3, 4 },

        ScaleToDifficulty = function( self, minigame )
            self.count = self.count_scaling[math.min(#self.count_scaling, minigame:GetDifficulty())]
        end,

        OnPostResolve = function( self, minigame, targets )
            self.negotiator:AddModifier("PERMIT", self.count, self)
        end,
    },

    sick =
    {
        name = "Sick",
        cost = 1,
        
        desc = "Lose 2 resolve.",
        flavour = "'Hesh, you look like garbage.'",
        remove_on_rest = true,
        
        OnPostResolve = function( self, minigame )
            self.negotiator:ModifyResolve(-2)
        end,

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND | CARD_FLAGS.SLEEP_IT_OFF,

        rarity = CARD_RARITY.UNIQUE,
        series = CARD_SERIES.GENERAL,
    },

    ulcer =
    {
        name = "Ulcer",
        cost = 2,
        flavour = "'I forgot my pills, this is going to be awful.'",
        quip = "sick_card",
        remove_on_rest = true,
        

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.CONSUME | CARD_FLAGS.SLEEP_IT_OFF,

        rarity = CARD_RARITY.UNIQUE,
        series = CARD_SERIES.GENERAL,
    },

    start_battle = 
    {
        name = "Start Fight",
        flavour = "'Let’s get this over with.'",
        desc = "End the negotiation and start the fight.",

        flags = CARD_FLAGS.MANIPULATE,
        rarity = CARD_RARITY.UNIQUE,
        cost = 0,
        
        OnPostResolve = function( self, minigame, targets )
            minigame:Win()
        end,
    },

    fabrication =
    {
        name = "Fabrication",
        name = "Fabrication",
        desc = "At the end of your turn, this card deals 2 damage to your core argument.",
        flavour = "'I don’t care if you didn’t do it. Everyone is already against you.'",

        cost = 1,

        flags = CARD_FLAGS.STATUS | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        event_handlers =
        {
            [ EVENT.END_TURN ] = function( self, minigame, negotiator )
                if self.negotiator == negotiator and self.deck == minigame:GetHandDeck() then
                    minigame:ApplyPersuasion( self, self.negotiator, 2, 2 )
                end
            end
        },
    },

    ad_lib =
    {
        name = "Ad-Lib",
        flavour = "'Wait a minute!'",
        cost = 0,

        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.MANIPULATE | CARD_FLAGS.DUD,
        rarity = CARD_RARITY.UNIQUE,

        min_persuasion = 2,
        max_persuasion = 2,
    },

    fake_it =
    {
        name = "Fake It",
        flavour = "'I can take him.'",
        cost = 0,

        flags = CARD_FLAGS.EXPEND | CARD_FLAGS.HOSTILE | CARD_FLAGS.DUD,
        rarity = CARD_RARITY.UNIQUE,

        min_persuasion = 1,
        max_persuasion = 3,
    },
}

local MODIFIERS = {
    
    FIGHT_ALLY = 
    {
        name = "Potential Ally",
        desc = "<b>{1.fullname}</> will fight on your side if you win the negotiation with this argument on your side.",
        loc_strings =
        {
            NOT_READY =  "{1#agent} is not ready to fight!"
        },
        desc_fn = function(self, fmt_str, minigame, widget)
            if self.ally_agent and widget and widget.PostPortrait then
                local txt = loc.format( self.def:GetLocalizedString( "NOT_READY" ), self.ally_agent )
                widget:PostPortrait( self.ally_agent, txt )
                return loc.format( fmt_str, self.ally_agent:LocTable() )
            end
        end,
        max_resolve = 10,
        modifier_type = MODIFIER_TYPE.BOUNTY,

        OnBounty = function( self )
            local mod = self.anti_negotiator:CreateModifier( "FIGHT_ALLY_WON" )
            mod:SetAgent(self.ally_agent)
            mod:NotifyChanged()
        end,


        SetAgent = function (self, agent)
            self.ally_agent = agent
            self.max_resolve = math.floor(5 + agent:GetCombatStrength() * 4)
            self:SetResolve(self.max_resolve)
            
            if agent.negotiation_ally_image then
                self.icon = agent.negotiation_ally_image
                self.engine:BroadcastEvent( EVENT.UPDATE_MODIFIER_ICON, self)
                self:NotifyTriggered()
            end
            self:NotifyChanged()
        end,
    },        

    FIGHT_ALLY_WON = 
    {
        name = "Ally",
        desc = "<b>{1.fullname}</> will fight on your side if you win this negotiation.",
        loc_strings =
        {
            READY =  "{1#agent} is ready to fight!"
        },

        desc_fn = function(self, fmt_str, minigame, widget)
            
            if self.ally_agent and widget and widget.PostPortrait then
                local txt = loc.format( self.def:GetLocalizedString("READY"), self.ally_agent )
                widget:PostPortrait( self.ally_agent, txt )
            end
            if self.ally_agent then
                return loc.format( fmt_str, self.ally_agent:LocTable() )
            end

            return "???"
        end,
        max_resolve = 1,

        modifier_type = MODIFIER_TYPE.ARGUMENT,

        OnBounty = function( self )
            local mod = self.anti_negotiator:CreateModifier( "FIGHT_ALLY" )
            mod:SetAgent(self.ally_agent)
            mod:NotifyChanged()
        end,

        SetAgent = function (self, agent)
            self.ally_agent = agent
            self.max_resolve = math.floor(2 + agent:GetCombatStrength())
            self:SetResolve(self.max_resolve)

            if agent.negotiation_ally_image then
                self.icon = agent.negotiation_ally_image
                self.engine:BroadcastEvent( EVENT.UPDATE_MODIFIER_ICON, self)
                self:NotifyTriggered()
            end
        end,
    },

    FIGHT_ALLY_SCARE = 
    {
        name = "Potential Enemy",
        desc = "<b>{1.fullname}</> will fight against you in the coming battle if you don't scare {1.himher} off first.\n\n{2.name} gains 1 {IMPATIENCE} when destroyed.",
        desc_fn = function(self, fmt_str, minigame, widget)
            if self.ally_agent and widget and widget.PostPortrait then
                --local txt = loc.format( "{1#agent} is not ready to fight!", self.ally_agent )
                widget:PostPortrait( self.ally_agent )
            end
            return loc.format( fmt_str, self.ally_agent and self.ally_agent:LocTable(), self.negotiator and self.negotiator.agent:LocTable() )
        end,
        icon = "negotiation/modifiers/recruit_admiralty_goon.tex", -- so this modifier doesn't show up as missing art.

        max_resolve = 10,
        max_stacks = 1,

        OnBounty = function(self)
            self.negotiator:AddModifier("IMPATIENCE")
        end,

        OnUnapply = function( self, minigame )
            local core = self.negotiator:FindCoreArgument()
            if core then
                core:SetShieldStatus( false)
            end
        end,
        
        SetAgent = function (self, agent)
            self.ally_agent = agent
            self.max_resolve = 4 + math.floor( agent:GetCombatStrength() * 2 )
            self:SetResolve(self.max_resolve)

            if agent.negotiation_ally_image then
                self.icon = agent.negotiation_ally_image
                self.engine:BroadcastEvent( EVENT.UPDATE_MODIFIER_ICON, self)
                self:NotifyTriggered()
            end
            self:NotifyChanged()
        end,
    },      


    FIGHT_SUPPORTER = 
    {
        name = "Supporter",
        desc = "<b>{1.fullname}</> will support {2.name} in battle - destroy this argument to undermine that. \n\n{2.name} gains 1 {IMPATIENCE} when destroyed.",
        desc_fn = function(self, fmt_str, minigame, widget)
            if self.ally_agent and widget and widget.PostPortrait then
                --local txt = loc.format( "{1#agent} is not ready to fight!", self.ally_agent )
                widget:PostPortrait( self.ally_agent )
            end
            return loc.format( fmt_str, self.ally_agent and self.ally_agent:LocTable(), self.negotiator and self.negotiator.agent:LocTable() )
        end,
        icon = "negotiation/modifiers/recruit_admiralty_goon.tex", -- so this modifier doesn't show up as missing art.

        max_resolve = 10,
        max_stacks = 1,

        OnBounty = function(self)
            self.negotiator:AddModifier("IMPATIENCE")
        end,

        OnUnapply = function( self, minigame )
            local core = self.negotiator:FindCoreArgument()
            if core then
                core:SetShieldStatus( false)
            end
            if self.destroy_dialog then
                self.engine:BroadcastEvent( negotiation_defs.EVENT.DIALOG, self.destroy_dialog )
            end

        end,
        
        SetAgent = function (self, agent)
            self.ally_agent = agent
            self.max_resolve = 4 + math.floor( agent:GetCombatStrength() * 2 )
            self:SetResolve(self.max_resolve)

            if agent.negotiation_ally_image then
                self.icon = agent.negotiation_ally_image
                self.engine:BroadcastEvent( EVENT.UPDATE_MODIFIER_ICON, self)
                self:NotifyTriggered()
            end
            self:NotifyChanged()
        end,
    },      




}

for i, id, carddef in sorted_pairs( CARDS ) do
    if not id:find("_plus") then
        carddef.rarity = carddef.rarity or CARD_RARITY.BASIC
        carddef.series = CARD_SERIES.GENERAL
    end

    Content.AddNegotiationCard( id, carddef )
end


for id, def in pairs( MODIFIERS ) do
    Content.AddNegotiationModifier( id, def )
end
