local negotiation_defs = require "negotiation/negotiation_defs"
local NEGOTIATION_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT
local battle_defs = require "battle/battle_defs"
local CARD_FLAGS = battle_defs.CARD_FLAGS
local BATTLE_EVENT = battle_defs.EVENT

local function FindCoinGrafts( number, rarity )
    local owner = TheGame:GetGameState():GetPlayerAgent()
    local collection = GraftCollection.RewardableCoin(owner):Rarity(rarity)
    local grafts = collection:Generate(number)
    
    if grafts then
        for k,graft in ipairs(grafts) do
            TheGame:GetGameProfile():SetSeenGraft( graft.id )
        end
    end
    
    return grafts
end

local RARITY_TABLE = {
    CARD_RARITY.COMMON,
    CARD_RARITY.UNCOMMON,
    CARD_RARITY.RARE
}

local NEGOTIATION_FLOURISHES = 
{
	-- SAL
	trounce =
	{
		name = "Trounce",
		desc = "Dismiss all enemy intents.",
		flavour = "'I forgot what I was trying to say.'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SAL",
		basic_flourish = true,

		OnPostResolve = function( self, minigame, targets )
			minigame:GetOtherNegotiator(self.negotiator):ClearIntents()
		end,
	},

	pure_diplomacy =
	{
		name = "Pure Diplomacy",
		desc = "Gain 5 {INFLUENCE} and apply 5 {COMPOSURE} to it.",
		flavour = "'I’m sure we can both leave here happy.'",

		flags = NEGOTIATION_FLAGS.DIPLOMACY,
        target_self = TARGET_ANY_RESOLVE,
        auto_target = true,
		series = "SAL",

		OnPostResolve = function( self, minigame, targets )
			local mod = self.negotiator:AddModifier( "INFLUENCE", 5, self )
			if mod then
				mod:DeltaComposure( 5, self )
			end
		end,
	},

	pure_diplomacy_ii =
	{
		name = "Pure Diplomacy II",
		desc = "Gain 5 {INFLUENCE} and apply 5 {COMPOSURE} to all friendly arguments.",

		flags = NEGOTIATION_FLAGS.DIPLOMACY,
		series = "SAL",
		target_mod = TARGET_MOD.TEAM,
		auto_target = true,

		OnPostResolve = function( self, minigame, targets )
			local mod = self.negotiator:AddModifier( "INFLUENCE", 5, self )
			local applied_to_mod = false
			for i,target in ipairs(targets) do
				target:DeltaComposure( 5, self )
				if target == mod then
					applied_to_mod = true
				end
			end
			if mod and not applied_to_mod then
				mod:DeltaComposure( 5, self )
			end
		end,
	},

	rage = 
	{
		name = "Rage",
		desc = "Gain {DOMINANCE 3} and {IMPROVISE} a Hostility card from your deck.",
		flavour = "'Oh, she looks mad.'",

		flags = NEGOTIATION_FLAGS.HOSTILE,
		series = "SAL",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("DOMINANCE", 3, self)
			local deck = table.merge( minigame:GetDrawDeck().cards, minigame:GetDiscardDeck().cards )
			local hostile_cards = ObtainWorkTable()
			for i,card in ipairs(deck) do
				if card:IsFlagged(NEGOTIATION_FLAGS.HOSTILE) then
					table.insert(hostile_cards, card)
				end
			end
			if #hostile_cards > 0 then
				minigame:ImproviseCards(table.multipick(hostile_cards, 3), 1, nil, nil, nil, self)
			end
			ReleaseWorkTable(hostile_cards)
		end
	},

	rage_ii = 
	{
		name = "Rage II",
		desc = "Gain {DOMINANCE 5} and {IMPROVISE_PLUS} a Hostility card from your deck.",

		flags = NEGOTIATION_FLAGS.HOSTILE,
		series = "SAL",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("DOMINANCE", 5, self)
			local deck = table.merge( minigame:GetDrawDeck().cards, minigame:GetDiscardDeck().cards )
			local hostile_cards = ObtainWorkTable()
			for i,card in ipairs(deck) do
				if card:IsFlagged(NEGOTIATION_FLAGS.HOSTILE) then
					table.insert(hostile_cards, card)
				end
			end
			if #hostile_cards > 0 then
				minigame:ImproviseCards(table.multipick(hostile_cards, 5), 1, nil, nil, nil, self)
			end
			ReleaseWorkTable(hostile_cards)
		end
	},

	clemency =
	{
		name = "Clemency",
		desc = "Restore 10 resolve to your core argument.",
		flavour = "'Don’t worry about it, we’ll get them next time.'",


		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SAL",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:RestoreResolve( 10, self )
		end
	},

	clemency_ii =
	{
		name = "Clemency II",
		desc = "Restore 15 resolve and apply 5 {COMPOSURE} to your core argument.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SAL",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:RestoreResolve( 15, self )
			self.negotiator:DeltaComposure( 5, self )
		end
	},

	duality =
	{
		name = "Duality",
		desc = "Discard your hand. Choose Diplomacy or Hostility, then draw cards of only that type until you have 6 cards.",
		flavour = "'I’m good for it, when have I ever steered you wrong?'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SAL",
		OnPostResolve = function( self, minigame, targets )
			local tbl = ObtainWorkTable()
            for i, card in minigame:GetHandDeck():Cards() do
                table.insert(tbl, card)
            end

            for i, card in ipairs(tbl) do
                minigame:DiscardCard(card)
            end
            ReleaseWorkTable(tbl)

            local choice
            local choices = {
                Negotiation.Card( "choose_diplomacy", self.owner ),
                Negotiation.Card( "choose_hostile", self.owner ),
            }
            choice = self.engine:ImproviseCards( choices, 1, nil, nil, nil, self )[1]
            minigame:ExpendCard(choice)
            choice = choice:IsFlagged(NEGOTIATION_FLAGS.DIPLOMACY) and NEGOTIATION_FLAGS.DIPLOMACY or NEGOTIATION_FLAGS.HOSTILE
            local to_draw = {}
            for i=1, #minigame:GetDrawDeck().cards do
            	local card = minigame:GetDrawDeck():Peek( i )
            	if card:IsFlagged(choice) then
            		table.insert(to_draw, card)
            		if #to_draw >= 6 then
            			break
            		end
            	end
            end
            if #to_draw < 6 then
            	minigame:ShuffleDiscardToDraw()
            	for i=1, #minigame:GetDrawDeck().cards do
	            	local card = minigame:GetDrawDeck():Peek( i )
	            	if card:IsFlagged(choice) and not table.contains(to_draw, card) then
	            		table.insert(to_draw, card)
	            		if #to_draw >= 6 then
	            			break
	            		end
	            	end
	            end
            end
            -- Manually drawing instead of just inserting to ensure any "when drawn", effects are triggered
            for i,card in ipairs(to_draw) do
            	minigame:GetHandDeck():InsertCard( card )
            	minigame:BroadcastEvent( EVENT.DRAW_CARD, minigame, card, false )
            end
		end
	},

	duality_ii =
	{
		name = "Duality II",
		desc = "Discard your hand. Choose Diplomacy or Hostility, then draw cards of only that type until your hand is full.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SAL",
		OnPostResolve = function( self, minigame, targets )
			local tbl = ObtainWorkTable()
            for i, card in minigame:GetHandDeck():Cards() do
                table.insert(tbl, card)
            end

            for i, card in ipairs(tbl) do
                minigame:DiscardCard(card)
            end
            ReleaseWorkTable(tbl)

            local choice
            local choices = {
                Negotiation.Card( "choose_diplomacy", self.owner ),
                Negotiation.Card( "choose_hostile", self.owner ),
            }
            choice = self.engine:ImproviseCards( choices, 1, nil, nil, nil, self )[1]
            minigame:ExpendCard(choice)
            choice = choice:IsFlagged(NEGOTIATION_FLAGS.DIPLOMACY) and NEGOTIATION_FLAGS.DIPLOMACY or NEGOTIATION_FLAGS.HOSTILE
            local max_size = minigame:GetHandDeck():GetMaxSize()
            local to_draw = {}
            for i=1, #minigame:GetDrawDeck().cards do
            	local card = minigame:GetDrawDeck():Peek( i )
            	if card:IsFlagged(choice) then
            		table.insert(to_draw, card)
            		if #to_draw >= max_size then
            			break
            		end
            	end
            end
            if #to_draw < max_size then
            	minigame:ShuffleDiscardToDraw()
            	for i=1, #minigame:GetDrawDeck().cards do
	            	local card = minigame:GetDrawDeck():Peek( i )
	            	if card:IsFlagged(choice) and not table.contains(to_draw, card) then
	            		table.insert(to_draw, card)
	            		if #to_draw >= max_size then
	            			break
	            		end
	            	end
	            end
            end
            -- Manually drawing instead of just inserting to ensure any "when drawn", effects are triggered
            for i,card in ipairs(to_draw) do
            	minigame:GetHandDeck():InsertCard( card )
            	minigame:BroadcastEvent( EVENT.DRAW_CARD, minigame, card, false )
            end
		end
	},

	-- ROOK
	trickster = 
	{
		name = "Trickster",
		desc = "{GAMBLE} 5 times.",
		flavour = "'Oh you want to twist the scoring now? Not a problem, that’ll just be 3 more shills.'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",
		basic_flourish = true,

		OnPostResolve = function( self, minigame, targets )
			local coin = self.negotiator:FindModifier("LUCKY_COIN")
            if coin then
            	for i=1,5 do
            		coin:Gamble(self)
            	end
            end
		end
	},

	backup_collection =
	{
		name = "Backup Collection",
		desc = "Change your Coin until the end of this negotiation.\n{GAMBLE}.",
		flavour = "'I brought another variant if you’d prefer the halfmoon bog rules.'",


		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
            local coin = self.negotiator:FindModifier("LUCKY_COIN")
            if coin then
    			local options = FindCoinGrafts(3, RARITY_TABLE[math.min(#RARITY_TABLE, minigame:GetDifficulty())])
    			local popup = Screen.PickCoinScreen(options, false, function(graft) 
                    TheGame:FE():FindScreen( Screen.ConversationScreen ).negotiation_panel:ResumePlay( graft )
    			end, true, true)
    			TheGame:FE():InsertScreen( popup )

                local graft = minigame:YieldPlay( minigame.STATUS.UI )
                if graft then
                    local coin_mod = self.negotiator:AddModifier(graft.id)
                    coin:SwapCoins(coin_mod)
                    self.negotiator:RemoveModifier(coin_mod, 1, self)
                    self.engine:BroadcastEvent( EVENT.UPDATE_COIN, coin, math.random() < 0.5)
                end
            end
		end,
	},

	backup_collection_ii =
	{
		name = "Backup Collection II",
		desc = "Change your Coin permanently.\n{GAMBLE}.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
            local coin = self.negotiator:FindModifier("LUCKY_COIN")
            if coin then
    			local options = FindCoinGrafts(3, RARITY_TABLE[math.min(#RARITY_TABLE, minigame:GetDifficulty())])
    			local popup = Screen.PickCoinScreen(options, false, function(graft) 
                    --this protects against resets during the coin select screen
                    if TheGame:FE():FindScreen( Screen.ConversationScreen ) and TheGame:FE():FindScreen( Screen.ConversationScreen ).negotiation_panel then
                    	TheGame:FE():FindScreen( Screen.ConversationScreen ).negotiation_panel:ResumePlay( graft )
                    end
    			end, true)
    			TheGame:FE():InsertScreen( popup )

                local graft = minigame:YieldPlay( minigame.STATUS.UI )
                if graft then
                    local coin_mod = self.negotiator:AddModifier(graft.id)
                    coin:SwapCoins(coin_mod)
                    self.negotiator:RemoveModifier(coin_mod, 1, self)
                    self.engine:BroadcastEvent( EVENT.UPDATE_COIN, coin, math.random() < 0.5)
                end
            end
		end,
	},

	expert_tactician =
	{
		name = "Expert Tactician",
		desc = "Discard any number of cards, then draw that many cards.",
		flavour = "'This opening gambit scores more if you take the game in 4 moves.'",


		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			local cards = minigame:DiscardCards(nil, nil, self)
			minigame:DrawCards(#cards)
		end
	},

	expert_tactician_ii =
	{
		name = "Expert Tactician II",
		desc = "Discard any number of cards, then draw cards until your hand is full.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			local cards = minigame:DiscardCards(nil, nil, self)
			minigame:DrawCards( 10 - minigame:GetHandDeck():CountCards() )
		end
	},

	always_prepared =
	{
		name = "Always Prepared",
		desc = "{always_prepared|}For the next 2 turns, {PREPARE} a card after every card you play.",
		flavour = "'Don’t ask me for help when we’re in trouble, I only have enough for myself.'",


		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("always_prepared", 2, self)
		end,

		modifier =
		{
			desc = "For the next {1} {1*turn|turns}, {PREPARE} a card after every card you play.",
			desc_fn = function( self, fmt_str )
				return loc.format( fmt_str, self.stacks )
			end,

			max_resolve = 4,
			modifier_type = MODIFIER_TYPE.ARGUMENT,

			event_handlers =
			{
				[ EVENT.END_TURN ] = function( self, minigame, negotiator )
					if negotiator == self.negotiator then
						self.negotiator:RemoveModifier(self.id, 1, self)
					end
				end,

				[ EVENT.POST_RESOLVE ] = function(self, minigame, card)
					if card.negotiator == self.negotiator then
						minigame:PrepareCard( nil, self )
					end
				end
			}
		}
	},

	always_prepared_ii =
	{
		name = "Always Prepared II",
		desc = "{always_prepared_ii|}For the rest of this negotiation, {PREPARE} a card after every card you play.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("always_prepared_ii", 1, self)
		end,

		modifier =
		{
			desc = "{PREPARE} a card after every card you play.",
			desc_fn = function( self, fmt_str )
				return loc.format( fmt_str, self.stacks )
			end,
			icon = "negotiation/modifiers/always_prepared.tex",

			max_resolve = 6,
			modifier_type = MODIFIER_TYPE.ARGUMENT,

			event_handlers =
			{
				[ EVENT.POST_RESOLVE ] = function(self, minigame, card)
					if card.negotiator == self.negotiator then
						minigame:PrepareCard( nil, self )
					end
				end
			}
		}
	},

	coin_juggler =
	{
		name = "Coin Juggler",
		desc = "For the next 2 turns, {GAMBLE} once after every card you play.",
		flavour = "'It just takes a bit of practice.'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier( "coin_juggler", 2, self )
		end,

		modifier =
		{
			desc = "For the next {1} {1*turn|turns}, {GAMBLE} once after every card you play.",
			desc_fn = function( self, fmt_str )
				return loc.format( fmt_str, self.stacks )
			end,

			max_resolve = 6,
			modifier_type = MODIFIER_TYPE.ARGUMENT,

			event_handlers =
			{
				[ EVENT.END_TURN ] = function( self, minigame, negotiator )
					if negotiator == self.negotiator then
						self.negotiator:RemoveModifier(self.id, 1, self)
					end
				end,

				[ EVENT.POST_RESOLVE ] = function(self, minigame, card)
					if card.negotiator == self.negotiator then
						local coin = self.negotiator:FindModifier("LUCKY_COIN")
						if coin then
							coin:Gamble()
						end
					end
				end,
			}
		}
	},

	coin_juggler_ii =
	{
		name = "Coin Juggler II",
		desc = "For the rest of this negotiation, {GAMBLE} once after every card you play.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "ROOK",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier( "coin_juggler_ii", 1, self )
		end,

		modifier =
		{
			desc = "{GAMBLE} once after every card you play.",
			desc_fn = function( self, fmt_str )
				return loc.format( fmt_str, self.stacks )
			end,
			icon = "negotiation/modifiers/coin_juggler.tex",

			max_resolve = 6,
			modifier_type = MODIFIER_TYPE.ARGUMENT,

			event_handlers =
			{
				[ EVENT.POST_RESOLVE ] = function(self, minigame, card)
					if card.negotiator == self.negotiator then
						local coin = self.negotiator:FindModifier("LUCKY_COIN")
						if coin then
							coin:Gamble()
						end
					end
				end,
			}
		}
	},

	-- SMITH
	claim_to_fame =
	{
		name = "Claim to Fame",
		desc = "Gain 8 {RENOWN}.",
		flavour = "'It helps if you are the best dressed in the room.'",

		flags = NEGOTIATION_FLAGS.DIPLOMACY,
		series = "SMITH",
		basic_flourish = true,

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("RENOWN", 8, self)
		end,
	},

	rebuild = 
	{
		name = "Rebuild",
		desc = "Remove a friendly argument and create a new version of it with twice as much max resolve.",
		flavour = "'Lets make sure we start from a good foundation.'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SMITH",

		target_self = TARGET_FLAG.BOUNTY | TARGET_FLAG.ARGUMENT,
 
        OnPostResolve = function( self, minigame, targets )
            local id, stacks, cur, max = targets[1].id, targets[1].stacks, targets[1]:GetResolve()
            self.negotiator:RemoveModifier( targets[1], stacks, self )
            local new_mod = self.negotiator:CreateModifier( id, stacks, self )
        	new_mod:ModifyResolve( (max * 2) - new_mod.resolve, self )
        end
	},

	rebuild_ii =
	{
		name = "Rebuild II",
		desc = "Remove a friendly argument and create two new versions of it with twice as much max resolve.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SMITH",

		target_self = TARGET_FLAG.BOUNTY | TARGET_FLAG.ARGUMENT,
 
        OnPostResolve = function( self, minigame, targets )
            local id, stacks, cur, max = targets[1].id, targets[1].stacks, targets[1]:GetResolve()
            self.negotiator:RemoveModifier( targets[1], stacks, self )
            for i=1,2 do
            	local new_mod = self.negotiator:CreateModifier( id, stacks, self )
	           	new_mod:ModifyResolve( (max * 2) - new_mod.resolve, self )
            end
        end
	},

	flatten =
	{
		name = "Flatten",
		desc = "Deal 6 damage to all opponent arguments.",
		flavour = "'You're going to be looking like this when I'm finished here.'",

		flags = NEGOTIATION_FLAGS.DIPLOMACY,
		series = "SMITH",

		target_mod = TARGET_MOD.TEAM,
		auto_target = true,

		min_persuasion = 6,
		max_persuasion = 6,
	},

	flatten_ii =
	{
		name = "Flatten II",
		desc = "Apply 3 {COMPOSURE} to all friendly arguments, then deal 6 damage to all opponent arguments.",

		flags = NEGOTIATION_FLAGS.DIPLOMACY,
		series = "SMITH",

		target_mod = TARGET_MOD.TEAM,
		auto_target = true,

		min_persuasion = 6,
		max_persuasion = 6,

		OnPreResolve = function( self, minigame, targets )
			local options = minigame:CollectAlliedTargets( self.negotiator )
			for i,mod in ipairs(options) do
				mod:DeltaComposure( 3, self )
			end
            if #options > 0 then
                minigame:Delay( 1.0 )
            end
		end
	},

	curve_ball =
	{
		name = "Curve Ball",
		desc = "Apply 4 {COMPOSURE} to all friendly arguments, then attack them all for 3 damage.",
		flavour = "'Well THAT was unexpected.'",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SMITH",

		target_mod = TARGET_MOD.TEAM,
		target_self = TARGET_ANY_RESOLVE,
        auto_target = true,

		OnPostResolve = function( self, minigame, targets )
			for i,mod in ipairs(targets) do
				mod:DeltaComposure( 4, self )
			end
			for i,mod in ipairs(targets) do
				mod:AttackResolve( 3, self )
			end
		end,
	},

	curve_ball_ii =
	{
		name = "Curve Ball II",
		desc = "Apply 6 {COMPOSURE} to all friendly arguments, then attack them all for 5 damage.",

		flags = NEGOTIATION_FLAGS.MANIPULATE,
		series = "SMITH",

		target_mod = TARGET_MOD.TEAM,
		target_self = TARGET_ANY_RESOLVE,
        auto_target = true,

		OnPostResolve = function( self, minigame, targets )
			for i,mod in ipairs(targets) do
				mod:DeltaComposure( 6, self )
			end
			for i,mod in ipairs(targets) do
				mod:AttackResolve( 5, self )
			end
		end,
	},

	inner_rage =
	{
		name = "Inner Rage",
		desc = "For the rest of this turn, Hostility cards are played twice.",
		flavour = "'Wait, there's MORE rage?'",

		flags = NEGOTIATION_FLAGS.HOSTILE,
		series = "SMITH",

		OnPostResolve = function( self, minigame, targets )
			local mod = self.negotiator:AddModifier("inner_rage", 1, self)
			mod.ignore = mod.ignore or {}
            table.insert(mod.ignore, self)
		end,

		modifier = 
		{
			desc = "For the rest of this turn, Hostility cards are played {1} additional {1*time|times}.",
			desc_fn = function( self, fmt_str )
				return loc.format( fmt_str, self.stacks or 1)
			end,
			max_resolve = 4,
			modifier_type = MODIFIER_TYPE.ARGUMENT,
			event_handlers =
			{
				[ EVENT.START_RESOLVE ] = function( self, minigame, card )
					if card.negotiator == self.negotiator and card:IsFlagged(NEGOTIATION_FLAGS.HOSTILE) then
						if self.ignore and table.contains(self.ignore, card) then
                            -- skip
                        else
                            self.ignore = self.ignore or {}
                            table.insert(self.ignore, card)
                            for i = 1, self.stacks do
        						minigame:PlayCard(card)
                            end
                        end
					end
				end,

                [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
                    if self.ignore and table.contains(self.ignore, card) then
                        if target_deck == self.engine:GetHandDeck() then
                            table.arrayremove( self.ignore, card)
                        end
                    end
                end,

				[ EVENT.END_TURN ] = function( self, minigame, negotiator )
					if negotiator == self.negotiator then
						self.negotiator:RemoveModifier(self.id, self.stacks, self)
					end
				end,
			},
		},
	},

	inner_rage_ii =
	{
		name = "Inner Rage II",
		desc = "Gain 3 {DOMINANCE}.\nFor the rest of this turn, Hostility cards are played twice.",

		flags = NEGOTIATION_FLAGS.HOSTILE,
		series = "SMITH",

		OnPostResolve = function( self, minigame, targets )
			self.negotiator:AddModifier("DOMINANCE", 3, self)
			local mod = self.negotiator:AddModifier("inner_rage", 1, self)
			mod.ignore = mod.ignore or {}
            table.insert(mod.ignore, self)
		end,
	},
}

for i, id, data in sorted_pairs( NEGOTIATION_FLOURISHES ) do
    assert(data.series, loc.format("Series missing on {1}", data.name))
    if id:match( "(.*)_ii.*$" ) then
        data.icon = string.format( "negotiation/%s.tex", id:match( "(.*)_ii.*$" ) )
        data.base_id = id:match( "(.*)_ii.*$" )
        data.pp_cost = 15
    elseif data.basic_flourish then
    	data.pp_cost = 0
    else
    	data.pp_cost = 5
    end
    if not data.basic_flourish then
        data.max_xp = 0
    end
	data.flags = data.flags | NEGOTIATION_FLAGS.FLOURISH
	data.rarity = CARD_RARITY.UNIQUE
	data.cost = 0
	data.flourish = true
    Content.AddNegotiationCard( id, data )
end

local BATTLE_FLOURISHES =
{
	-- SAL
	hail_of_blades =
	{
		name = "Hail of Blades",
		desc = "Attack all enemies twice.",
		anim = "targetpractice1",
		flavour = "'Just stay down!'",

		hit_anim = true,

		flags = CARD_FLAGS.RANGED,
		series = "SAL",
		basic_flourish = true,

		min_damage = 3,
		max_damage = 3,

		target_count = 2,
		target_mod = TARGET_MOD.TEAM,
	},

	squash = 
	{
		name = "Squash",
		desc = "Apply 4 {IMPAIR}.",
		anim = "stomp",
		flavour = "'I didn’t even see you there.'",

		flags = CARD_FLAGS.MELEE,
		series = "SAL",

		min_damage = 10,
		max_damage = 10,

		OnPostResolve = function( self, battle, attack )
			attack:AddCondition("IMPAIR", 4, self)
		end
	},

	squash_ii = 
	{
		name = "Squash II",
		anim = "stomp",

		flags = CARD_FLAGS.MELEE,
		series = "SAL",

		min_damage = 10,
		max_damage = 10,

		features =
        {
            STUN = 1,
        }
	},

	hemomancy =
	{
		name = "Hemomancy",
		desc = "Apply 16 {BLEED} split evenly among all enemies.",
		anim = "taunt3",
		flavour = "'I’ve never seen that much. I get sick just thinking about it.'",

		flags = CARD_FLAGS.RANGED,
		series = "SAL",
		target_mod = TARGET_MOD.TEAM,

		OnPostResolve = function( self, battle, attack )
			local bleed_amt = math.floor(16/#attack.hits)
			attack:AddCondition("BLEED", bleed_amt, self)
		end
	},

	hemomancy_ii =
	{
		name = "Hemomancy II",
		desc = "Apply 20 {BLEED} split evenly among all enemies.",
		anim = "taunt3",


		flags = CARD_FLAGS.RANGED,
		series = "SAL",
		target_mod = TARGET_MOD.TEAM,

		OnPostResolve = function( self, battle, attack )
			local bleed_amt = math.floor(20/#attack.hits)
			attack:AddCondition("BLEED", bleed_amt, self)
		end
	},

	martial_master =
	{
		name = "Martial Master",
		desc = "Gain 8 {COMBO} and {IMPROVISE} a {COMBO_FINISHER} from your deck.",
		anim = "hero_taunt2",
		flavour = "'You’re out of your league.'",

		flags = CARD_FLAGS.SKILL,
		series = "SAL",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("COMBO", 8, self)
			local deck = table.merge( battle:GetDrawDeck().cards, battle:GetDiscardDeck().cards )
			local finishers = ObtainWorkTable()
			for i,card in ipairs(deck) do
				if card:IsFlagged(CARD_FLAGS.COMBO_FINISHER) then
					table.insert(finishers, card)
				end
			end
			battle:ImproviseCards(table.multipick(finishers, 3), 1, "phantom_finisher", nil, nil, self)

			ReleaseWorkTable(finishers)
		end
	},

	martial_master_ii =
	{
		name = "Martial Master II",
		desc = "Gain 12 {COMBO} and {IMPROVISE_PLUS} a {COMBO_FINISHER} from your deck.",
		anim = "hero_taunt2",

		flags = CARD_FLAGS.SKILL,
		series = "SAL",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("COMBO", 12, self)
			local deck = table.merge( battle:GetDrawDeck().cards, battle:GetDiscardDeck().cards )
			local finishers = ObtainWorkTable()
			for i,card in ipairs(deck) do
				if card:IsFlagged(CARD_FLAGS.COMBO_FINISHER) then
					table.insert(finishers, card)
				end
			end
			battle:ImproviseCards(table.multipick(finishers, 5), 1, "phantom_finisher", nil, nil, self)
			ReleaseWorkTable(finishers)
		end
	},

	beast_master =
	{
		name = "Beast Master",
		desc = "Call in Crayotes to attack all enemies.",
		flavour = "'It takes a lot of training and I could never get over the smells.'",
		anim = "call_in",

		flags = CARD_FLAGS.SKILL,
		series = "SAL",
		target_type = TARGET_TYPE.SELF,

		on_duplicated = function( self )
			self.yotes = nil
			self.trigger_card = nil
		end,

		OnPostResolve = function( self, battle, attack )
            local targets = battle:GetEnemyTeam():GetFighters()
            self.yotes = {}
            local card
            for i,enemy in ipairs(targets) do
            	local yote = Fighter.CreateFromAgent( Agent("CRAYOTE"))
	            yote.team = self.owner:GetTeam()
	            yote.call_in = true
	            yote.home_x = (FIGHT_MAX_DIST + 12)*(self.owner:GetTeam() == TEAM.RED and 1 or -1)
	            yote:OnActivateFighter( battle )
	            table.insert(self.yotes, yote)
	            card = Battle.Card("yote_call_in", yote)
	            battle:PlayCard(card, enemy)
            end
            self.trigger_card = card
        end,

        deck_handlers = ALL_DECKS,
        event_handlers =
        {
        	-- Handling departure this way since doing it directly in yote_call_in will cause a delay between each yote's attack
        	[ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
        		if self.trigger_card and card == self.trigger_card then
        			for i,yote in ipairs(self.yotes) do
        				yote:Depart()
        			end
        			self.trigger_card = nil
        		end
        	end,
        },
	},

	beast_master_ii =
	{
		name = "Beast Master II",
		desc = "Call in Horned Yotes to attack all enemies.",
		anim = "call_in",

		flags = CARD_FLAGS.SKILL,
		series = "SAL",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
            local targets = battle:GetEnemyTeam():GetFighters()
            self.yotes = {}
            local card
            for i,enemy in ipairs(targets) do
            	local yote = Fighter.CreateFromAgent( Agent("HORNED_CRAYOTE"))
	            yote.team = self.owner:GetTeam()
	            yote.call_in = true
	            yote.home_x = (FIGHT_MAX_DIST + 12)*(self.owner:GetTeam() == TEAM.RED and 1 or -1)
	            yote:OnActivateFighter( battle )
	            table.insert(self.yotes, yote)
	            card = Battle.Card("horned_yote_call_in", yote)
	            battle:PlayCard(card, enemy)
            end
            self.trigger_card = card
        end,

        deck_handlers = ALL_DECKS,
        event_handlers =
        {
        	-- Handling departure this way since doing it directly in yote_call_in will cause a delay between each yote's attack
        	[ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
        		if self.trigger_card and card == self.trigger_card then
        			for i,yote in ipairs(self.yotes) do
        				yote:Depart()
        			end
        			self.trigger_card = nil
        		end
        	end,
        },
	},

	-- ROOK
	salvo = 
	{
		name = "Salvo",
		desc = "Hits all enemies.",
		flavour = "'It looks like a whole battalion came through here. What do you mean it was just one guy?'",
		anim = "execute",

		flags = CARD_FLAGS.RANGED,
		series = "ROOK",
		target_mod = TARGET_MOD.TEAM,
		basic_flourish = true,

		min_damage = 8,
		max_damage = 8,
	},

	guns_blazing =
	{
		name = "Guns Blazing",
		desc = "Apply 8 {BURN}, split evenly between all enemies.",
		flavour = "'Alright you can have them, but just don’t fire them indoors alright?.'",

		anim = "burner",

		flags = CARD_FLAGS.RANGED,
		series = "ROOK",
		target_mod = TARGET_MOD.TEAM,

		OnPostResolve = function( self, battle, attack )
			local burn_per = math.ceil(8/#attack.hits)
			attack:AddCondition("BURN", burn_per, self)
		end
	},

	guns_blazing_ii =
	{
		name = "Guns Blazing II",
		desc = "Apply 12 {BURN}, split evenly between all enemies.",
		anim = "burner",

		flags = CARD_FLAGS.RANGED,
		series = "ROOK",
		target_mod = TARGET_MOD.TEAM,

		OnPostResolve = function( self, battle, attack )
			local burn_per = math.ceil(12/#attack.hits)
			attack:AddCondition("BURN", burn_per, self)
		end
	},

	reset =
	{
		name = "Reset",
		desc = "Spend all {CHARGE}, then gain {CHARGE} until you are {PRIMED}.",
		flavour = "'Venting the chamber from here allows a cleaner and faster recharge.'",
		anim = "reload_02",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local tracker = self.owner:GetCondition("lumin_tracker")
			local cur, max = tracker:GetCharges()
			tracker:RemoveCharges( cur, self )
			tracker:AddLuminCharges( max, self )
		end
	},

	reset_ii =
	{
		name = "Reset II",
		desc = "Spend all {CHARGE}, then gain {CHARGE} until you are {PRIMED}.\nGain 5 {SURGE}.",
		anim = "reload_02",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local tracker = self.owner:GetCondition("lumin_tracker")
			local cur, max = tracker:GetCharges()
			tracker:RemoveCharges( cur, self )
			tracker:AddLuminCharges( max, self )
			self.owner:AddCondition("SURGE", 5, self)
		end
	},

	power_cycle = 
	{
		name = "Power Cycle",
		desc = "Gain {CHARGE} until you are {PRIMED}, then spend all your {CHARGE}.",
		flavour = "'It’s an older model with some changes I’ve made. Should do the trick.'",

		anim = "reload_01",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local tracker = self.owner:GetCondition("lumin_tracker")
			local cur, max = tracker:GetCharges()
			tracker:AddLuminCharges( max - cur, self )
			tracker:RemoveCharges( max, self )
		end
	},

	power_cycle_ii = 
	{
		name = "Power Cycle II",
		desc = "Gain {CHARGE} until you are {PRIMED}, then spend all your {CHARGE}.\nGain 6 {RIPOSTE}.",
		anim = "reload_01",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local tracker = self.owner:GetCondition("lumin_tracker")
			local cur, max = tracker:GetCharges()
			tracker:AddLuminCharges( max - cur, self )
			tracker:RemoveCharges( max, self )
			self.owner:AddCondition("RIPOSTE", 6, self)
		end
	},

	total_focus =
	{
		name = "Total Focus",
		desc = "Gain 6 {CONCENTRATION} and 10 {DEFEND}.",
		flavour = "'I’ve never seen someone so collected during a firefight.'",
		anim = "psionic",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("CONCENTRATION", 6, self)
			self.owner:AddCondition("DEFEND", 10, self)
		end
	},

	total_focus_ii =
	{
		name = "Total Focus II",
		desc = "Gain 10 {CONCENTRATION} and 20 {DEFEND}.",
		anim = "psionic",

		flags = CARD_FLAGS.SKILL,
		series = "ROOK",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("CONCENTRATION", 10, self)
			self.owner:AddCondition("DEFEND", 20, self)
		end
	},

	-- SMITH
	all_rounder =
	{
		name = "All Rounder",
		desc = "{DRINK} 2.\nTake 1 damage. Gain 3 {smith_regeneration}. Apply 3 {TRAUMA}.",
		flavour = "'Let’s get this party started!'",
		anim = "cheers_smash",

		flags = CARD_FLAGS.MELEE,
		series = "SMITH",
		basic_flourish = true,

		CanPlayCard = function( self, battle, target )
		    if self.owner:GetHealth() + self.owner:GetConditionStacks("DEFEND") > 1 then
		        return true
		    else
		        return false, CARD_PLAY_REASONS.NOT_ENOUGH_HEALTH
		    end
        end,

		min_damage = 6,
		max_damage = 6,

		OnPostResolve = function( self, battle, attack )
			local cards = {}
			for i=1,2 do
				local card = Battle.Card( "empty_bottle", self.owner )
		    	card.show_dealt = false
		    	table.insert(cards, card)
			end
		    battle:DealCards(cards, battle:GetDiscardDeck() )
		    battle:BroadcastEvent( BATTLE_EVENT.DRINK, battle, 2, cards )
		    self.owner:SelfDamage( 1, self )
		    local healing = self.owner:GetCondition("smith_regeneration")
			if healing then healing:IncreaseMax(2, self) end
			attack:AddCondition("TRAUMA", 3, self)
		end
	},

	shortcut =
	{
		name = "Shortcut",
		desc = "Apply {TRAUMATIZED}.",
		flavour = "'Hey that’s a cheap shot!'",
		anim = "punch",

		flags = CARD_FLAGS.MELEE,
		series = "SMITH",

		OnPostResolve = function( self, battle, attack )
			attack:AddCondition("TRAUMATIZED", 1, self)
		end
	},

	shortcut_ii =
	{
		name = "Shortcut II",
		desc = "Apply {TRAUMATIZED} to all enemies.",
		anim = "super_smash",

		flags = CARD_FLAGS.MELEE,
		series = "SMITH",
		target_mod = TARGET_MOD.TEAM,

		OnPostResolve = function( self, battle, attack )
			attack:AddCondition("TRAUMATIZED", 1, self)
		end
	},

	bio_boost =
	{
		name = "Bio-boost",
		desc = "Gain 8 {smith_regeneration}.",
		flavour = "'Have you been working out?'",
		anim = "moxie_high",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local healing = self.owner:GetCondition("smith_regeneration")
			if healing then healing:IncreaseMax(8, self) end
		end
	},

	bio_boost_ii =
	{
		name = "Bio-boost II",
		desc = "Gain 14 {smith_regeneration}.",
		anim = "moxie_high",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local healing = self.owner:GetCondition("smith_regeneration")
			if healing then healing:IncreaseMax(14, self) end
		end
	},

	secret_collection = 
	{
		name = "Secret Collection",
		desc = "{DRINK} 6 and gain 2 {ADRENALINE}.",
		flavour = "'Never leave home without it.'",
		anim = "drink2",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local cards = { }
		    for i=1, 6 do
		        local card = Battle.Card( "empty_bottle", self.owner )
		        card.auto_deal = true
		        table.insert(cards, card)
		    end
		    battle:DealCards( cards, battle:GetDiscardDeck() )
		    battle:BroadcastEvent( BATTLE_EVENT.DRINK, battle, 6, cards )
		    self.owner:AddCondition("ADRENALINE", 2, self)
		end
	},

	secret_collection_ii = 
	{
		name = "Secret Collection II",
		desc = "{DRINK} 10 and gain 2 {ADRENALINE}.",
		anim = "drink2",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			local cards = { }
		    for i=1, 10 do
		        local card = Battle.Card( "empty_bottle", self.owner )
                card.auto_deal = true
		        table.insert(cards, card)
		    end
		    battle:DealCards( cards, battle:GetDiscardDeck() )
		    battle:BroadcastEvent( BATTLE_EVENT.DRINK, battle, 10, cards )
		    self.owner:AddCondition("ADRENALINE", 2, self)
		end
	},

	innate_instinct =
	{
		name = "Innate Instinct",
		desc = "Gain 6 {ADRENALINE}.",
		flavour = "'Things are getting crazy!'",
		anim = "taunt",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("ADRENALINE", 6, self)
		end
	},

	innate_instinct_ii =
	{
		name = "Innate Instinct II",
		desc = "Gain 12 {ADRENALINE}.",
		anim = "taunt",

		flags = CARD_FLAGS.SKILL,
		series = "SMITH",
		target_type = TARGET_TYPE.SELF,

		OnPostResolve = function( self, battle, attack )
			self.owner:AddCondition("ADRENALINE", 12, self)
		end
	},
}

for i, id, data in sorted_pairs( BATTLE_FLOURISHES ) do
	assert(data.series, loc.format("Series missing on {1}", data.name))
	if id:match( "(.*)_ii.*$" ) then
        data.icon = string.format( "battle/%s.tex", id:match( "(.*)_ii.*$" ) )
        data.base_id = id:match( "(.*)_ii.*$" )
        data.pp_cost = 15
    elseif data.basic_flourish then
    	data.pp_cost = 0
    else
    	data.pp_cost = 5
    end
    if not data.basic_flourish then
        data.max_xp = 0
    end
	data.rarity = CARD_RARITY.UNIQUE
	data.flags = data.flags | CARD_FLAGS.FLOURISH
	data.cost = 0
	data.flourish = true
    Content.AddBattleCard( id, data )
end

Content.AddStoryGraft( "flourish_tracker", 
{
	name = "Flourish",
	hidden = true,

	default_battle_gains =
	{
		-- If new ones are added, update the FLOURISH_DESC_2_FIGHT string
		CARD_PLAY = 1,
    	DAMAGE_DEALT = 0.5,
    	DEFENDED = 0.5,
    	KILL = 10,
	},
	default_negotiation_gains = 
	{
		-- If new ones are added, update the FLOURISH_DESC_2_NEGOTIATION string
		CARD_PLAY = 1,
    	DAMAGE_DEALT = 0.5,
    	MITIGATED = 0.5,
    	DESTROY = 2,
	},

	OnAdded = function(self, agent)
        self.userdata.battle_gains = self.userdata.battle_gains or {}
        for k,v in pairs(self:GetDef().default_battle_gains) do
        	self.userdata.battle_gains[k] = self.userdata.battle_gains[k] or v
        end
        self.userdata.battle_flourish = 0
        self.userdata.battle_max = self.userdata.battle_max or 200

        self.userdata.negotiation_gains = self.userdata.negotiation_gains or {}
        for k,v in pairs(self:GetDef().default_negotiation_gains) do
        	self.userdata.negotiation_gains[k] = self.userdata.negotiation_gains[k] or v
        end
        self.userdata.negotiation_flourish = 0
        self.userdata.negotiation_max = self.userdata.negotiation_max or 200
    end,

    AddNegotiationFlourish = function( self, engine, amount )
        self:GetDef().SetNegotiationFlourish( self, engine, self.userdata.negotiation_flourish + amount )
    end,
    
    AddBattleFlourish = function( self, engine, amount )
        self:GetDef().SetBattleFlourish( self, engine, self.userdata.battle_flourish + amount )
    end,

    SetBattleFlourish = function(self, engine, flourish)
        local prev_flourish = self.userdata.battle_flourish 
        flourish = math.min(self.userdata.battle_max, flourish)
        if flourish ~= prev_flourish then
            self.userdata.battle_flourish = flourish
    		engine:BroadcastEvent( BATTLE_EVENT.FLOURISH_CHANGED, self, flourish, prev_flourish )    
    	end
    end,

    SetNegotiationFlourish = function(self, engine, flourish)
        local prev_flourish = self.userdata.negotiation_flourish
        flourish = math.min(self.userdata.negotiation_max, flourish)
        if flourish ~= prev_flourish then
			self.userdata.negotiation_flourish = flourish
			engine:BroadcastEvent( EVENT.FLOURISH_CHANGED, self, flourish, prev_flourish )
    	end
    end,
    

    negotiation_modifier =
    {
    	hidden = true,
    	event_handlers =
    	{
    		[ EVENT.POST_RESOLVE ] = function( self, minigame, card )
    			if card.negotiator == self.negotiator and self.graft.userdata.negotiation_gains["CARD_PLAY"] then
    				self.graft:GetDef().AddNegotiationFlourish( self.graft, minigame, self.graft.userdata.negotiation_gains["CARD_PLAY"] )
    			end 
    		end,

    		[ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
                if target.negotiator == self.negotiator then
                    if defended > 0 and self.graft.userdata.negotiation_gains["MITIGATED"] then
                    	self.graft:GetDef().AddNegotiationFlourish( self.graft, self.engine, self.graft.userdata.negotiation_gains["MITIGATED"] * defended )
                    end
                elseif source.negotiator == self.negotiator then
                	if damage > 0 and self.graft.userdata.negotiation_gains["DAMAGE_DEALT"] then
                		self.graft:GetDef().AddNegotiationFlourish( self.graft, self.engine, self.graft.userdata.negotiation_gains["DAMAGE_DEALT"] * damage )
                    end
                end
            end,

            [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
                if source and source.negotiator == self.negotiator and self.graft.userdata.negotiation_gains["DESTROY"] then
                    self.graft:GetDef().AddNegotiationFlourish( self.graft, self.engine, self.graft.userdata.negotiation_gains["DESTROY"] )
                end
            end,
    	},
    },

    battle_condition =
    {
    	hidden = true,
    	event_handlers =
    	{
    		[ BATTLE_EVENT.END_RESOLVE ] = function( self, battle, card )
    			if card.owner == self.owner and self.graft.userdata.battle_gains["CARD_PLAY"] then
    				self.graft:GetDef().AddBattleFlourish( self.graft, battle, self.graft.userdata.battle_gains["CARD_PLAY"] )
    			end
    		end,

    		[ BATTLE_EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source )
    			if fighter == self.owner then
                    if delta > 0 then
                        -- Yes, this is flourish for RECEIVING damage, not defending it, despite the name in the battle_gains table.
                        -- Can't change that key now due to backwards compatability.
    					self.graft:GetDef().AddBattleFlourish( self.graft, self.battle, self.graft.userdata.battle_gains["DEFENDED"] * delta )
                    end

    			elseif source and self.battle:GetFighterFromSource(source) == self.owner then
    				if delta > 0 then
    					self.graft:GetDef().AddBattleFlourish( self.graft, self.battle, self.graft.userdata.battle_gains["DAMAGE_DEALT"] * delta )
    				end
    			end
    		end,

    		[ BATTLE_EVENT.FIGHTER_KILLED ] = function ( self, fighter )
    			if fighter:GetTeam() ~= self.owner:GetTeam() and fighter.killer and fighter.killer == self.owner then
    				self.graft:GetDef().AddBattleFlourish( self.graft, self.battle, self.graft.userdata.battle_gains["KILL"] )
    			end
    		end,
    	},
    },
})