local fun = require "util/fun"

local negotiation_def = require "negotiation/negotiation_defs"
local CARD_FLAGS = negotiation_def.CARD_FLAGS
local EVENT = negotiation_def.EVENT

local function CreateDrawnHatcher(fn)
    local ret = function( self, minigame, card, start_of_turn )
                 if card == self then
                    self.engine:PushPostHandler( function()
                        self:NotifyTriggeredPre()
                        fn(self)
                        --self.owner:ApplyDamage(1, self.owner, self)
                        if self.engine then
                            self:AddXP(1)
                            if card:UpgradeReady() then
                                if card.hatch_fn then 
                                    if not card.hatched then
                                        card:hatch_fn(self.engine)
                                        card.hatched = true
                                    else
                                        card:TransferCard( self.engine.trash_deck )
                                        card:Consume()
                                    end
                                end
                            end
                        end
                        self:NotifyTriggeredPost()
                    end )
                end
            end
    return ret  
end

local PARASITES =
{
    stem = 
    {
        name = "Stem",
        desc = "When this card is drawn, increase the cost of a random card in your hand until the end of this turn.",
        cost = 0,
        flavour = "'You can barely notice it, just wear a scarf or something.'",
        max_xp = 6,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.UNPLAYABLE,
        rarity = CARD_RARITY.BASIC,

        event_handlers = 
        {
            [ EVENT.DRAW_CARD ] = CreateDrawnHatcher(function(self) 
                    local valid_cards = {}
                    for i,card in self.engine:GetHandDeck():Cards() do
                        if card ~= self and not card:IsFlagged( CARD_FLAGS.VARIABLE_COST ) and not card:IsFlagged( CARD_FLAGS.UNPLAYABLE ) then
                            table.insert(valid_cards, card)
                        end
                    end
                    if #valid_cards > 0 then
                        local mod = self.negotiator:AddModifier("stem", 1, self)
                        mod.chosen_card = valid_cards[math.random(#valid_cards)]
                    end
            end)
        },

        modifier =
        {
            hidden = true,
    
            event_priorities =
            {
                [ EVENT.CALC_ACTION_COST ] = EVENT_PRIORITY_ADDITIVE,
            },

            event_handlers =
            {
                [ EVENT.CALC_ACTION_COST ] = function( self, acc, card, target )
                    if self.chosen_card and card == self.chosen_card then
                        acc:AddValue(1, self)
                    end
                end,

                [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                    self.negotiator:RemoveModifier(self.id, 1)
                end
            },
        },
    },

    -- TIER 1
    earworm =
    {
        name = "Earworm",
        desc = "If this is in your hand at the end of your turn, add another <b>Earworm</> to your draw pile.",
        flavour = "'I keep hearing it, I'm sure you know the tune...'",
        cost = 1,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.COMMON,

        event_handlers =
        {
            [ EVENT.END_PLAYER_TURN ] = function( self, minigame )
                if self.deck == minigame:GetHandDeck() then
                    local clone = self:Duplicate()
                    clone:TransferCard( minigame:GetDrawDeck() )
                end
            end
        },
    },

    tinnitus =
    {
        name = "Tinnitus",
        desc = "When this card is drawn, deal 2 damage to a random friendly argument.",
        flavour = "'Eeeeeeee.'",

        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.UNPLAYABLE,
        rarity = CARD_RARITY.COMMON,


        event_handlers = 
        {
            [ EVENT.DRAW_CARD ] = CreateDrawnHatcher(function(self) 
                    local targets = self.engine:CollectAlliedTargets(self.negotiator)
                    if #targets > 0 then
                        local target = targets[math.random(#targets)]
                        target:AttackResolve(2, self)
                    end
            end)
        },
    },

    drowsiness = 
    {
        name = "Drowsiness",
        desc = "When this card is drawn, lose 1 action and {EXPEND} this card.",
        flavour = "'Never have I been this tired. It feels as if I could sleep for a hundred years.'",
        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.UNPLAYABLE,
        rarity = CARD_RARITY.COMMON,


        event_handlers = 
        {
            [ EVENT.DRAW_CARD ] = function( self, minigame, card, start_of_turn )
                 if card == self then
                    self.engine:PushPostHandler( function()
                        self:NotifyTriggeredPre()
                        if self.engine then
                            self.engine:ModifyActionCount(-1)
                            self:AddXP(1)
                            if card:UpgradeReady() then
                                if card.hatch_fn then 
                                    card:hatch_fn(self.engine)
                                end
                            end
                        end
                        self:NotifyTriggeredPost()
                        self.engine:ExpendCard(self)
                    end )
                end
            end
        },
    },

    -- TIER 2

    memory_lapse =
    {
        name = "Memory Lapse",
        desc = "Play a random card in your hand for free then {EXPEND} it.",

        flavour = "'I have to take good care of the clusters, when I took the first few off I forgot my mother's name.'",    
        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        OnPostResolve = function( self, minigame, targets )
            local valid_cards = shallowcopy(self.engine:GetHandDeck().cards)
            for i,card in ipairs(valid_cards) do
                if card == self or card:IsFlagged( CARD_FLAGS.UNPLAYABLE ) then
                    table.remove(valid_cards, i)
                end
            end
            if #valid_cards > 0 then
                local card = table.arraypick(valid_cards)
                card:SetFlags(CARD_FLAGS.FREEBIE)
                card:SetFlags(CARD_FLAGS.EXPEND)
                self.engine:PlayCard(card)
            end
        end
    },

    voices =
    {
        name = "Voices",
        desc = "Draw a card and duplicate it, both cards cost 0 and {EXPEND}.",
        flavour = "'A thick gurgling that is punctuated by a husky laugh, I can never tell what they are trying to say.'",
        cost = 1,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        OnPostResolve = function( self, minigame, targets )
            local cards = minigame:DrawCards(1)
            for i,card in ipairs(cards) do
                card:SetFlags(CARD_FLAGS.FREEBIE | CARD_FLAGS.EXPEND)
                local clone = card:Duplicate()
                clone:TransferCard( minigame:GetHandDeck() )
            end
        end
    },


    scrub =
    {
        name = "Scrub",
        desc = "{CONSUME} all Status cards in your deck, lose 1 resolve per card {CONSUME|Destroyed}.",
        flavour = "'A dark twisting in the guts with no relief in sight.'",

        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNCOMMON,

        GetStatusCards = function( self, minigame )
            local status_cards = {}
            -- Always take from draw pile first, then hand and finally discard.
            local all_cards = table.merge( minigame:GetDrawDeck().cards, minigame:GetHandDeck().cards, minigame:GetDiscardDeck().cards )
            for i,card in ipairs(all_cards) do
                if CheckBits(card.flags, CARD_FLAGS.STATUS) then
                    table.insert(status_cards, card)
                end
            end
            return status_cards
        end,

        CanPlayCard = function( self, card, engine, target )
            if card == self then
                return #self:GetStatusCards(engine) < self.negotiator:GetResolve()
            end
        end,

        OnPostResolve = function( self, minigame, targets )
            local cards = self:GetStatusCards(minigame)
            if #cards < self.negotiator:GetResolve() then
                for i,card in ipairs(cards) do
                    card:TransferCard( minigame.trash_deck )
                    card:Consume()
                end
                self.negotiator:LoseResolve(#cards, self)
            end
        end,
    },

    -- TIER 3
    flower =
    {
        name = "Flower",
        desc = "Play the top three cards of your draw pile for free. Lose resolve equal to twice their combined cost.",

        flavour = "'The pressure inside my head had left in a beautiful rush, my mind had finally unfurled.'",
        cost = 1,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,

        OnPostResolve = function( self, minigame, targets )
            if minigame:GetDrawDeck():CountCards() < 3 then
                minigame:ShuffleDiscardToDraw()
            end
            local cards = minigame:DrawCards(3)
            local combined_cost = 0
            for i,card in ipairs(cards) do
                combined_cost = combined_cost + minigame:CalculateActionCost(card)
                card:SetFlags(CARD_FLAGS.FREEBIE)
                minigame:PlayCard(card)
            end
            self.negotiator:ModifyResolve(-combined_cost * 2, self)
        end,
    },
    
    grandiosity = 
    {
        name = "Grandiosity",
        desc = "Upgrade all cards in your hand for the rest of this negotiation. Lose 2 resolve per card upgraded.",
        flavour = "'Can you hear it? A wonderful cacophony, nothing like it!'",
        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,

        OnPostResolve = function( self, minigame, targets )
            local cards = minigame:GetHandDeck().cards
            local count = 0
            for i,card in ipairs(cards) do
                if card:HasUpgrades() then
                    card:UpgradeCard()
                    count = count + 1
                end
            end
            self.negotiator:ModifyResolve( -count * 2, self )
        end
    },

    stutter =
    {
        name = "Stutter",
        desc = "Play your next card twice. Gain 2 {VULNERABILITY} and lose 4 resolve.",
        flavour = "'Something is wrong with my tongue, I'm not sure if it's mine anymore.'",
        cost = 1,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH,
        rarity = CARD_RARITY.RARE,

        OnPostResolve = function( self, minigame, targets )
            self.negotiator:AddModifier("VULNERABILITY", 2, self)
            self.negotiator:ModifyResolve( -4, self)
            self.negotiator:AddModifier("stutter", 1, self)
        end,

        modifier = 
        {
            hidden = true,
            event_handlers =
            {
                [ EVENT.POST_RESOLVE ] = function( self, minigame, card )
                    if card.owner == self.owner and card.id ~= "stutter" then
                        self.negotiator:RemoveModifier( self, 1 )
                        minigame:PlayCard( card )
                    end
                end,
            },
        },
    },

    hyperactive = 
    {
        name = "Hyperactive",
        desc = "While this is in your hand, all cards deal 2 bonus damage. Lose 1 resolve whenever you play a card.",
        flavour = "'What do you mean? I feel fine!'",
        cost = 0,
        max_xp = 5,
        flags = CARD_FLAGS.HATCH | CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.RARE,

        event_handlers =
        {
            [ EVENT.CALC_PERSUASION ] = function( self, source, persuasion )
                if source and is_instance(source, Negotiation.Card) and source.negotiator == self.negotiator then
                    persuasion:AddPersuasion( 2, 2, self )
                end
            end,

            [ EVENT.POST_RESOLVE ] = function( self, minigame, card, targets )
                if card.negotiator == self.negotiator and card ~= self then
                    self.negotiator:ModifyResolve( -1, self)
                end
            end
        },
    },

    -- TIER 4

    bog_enlightenment =
    {
        name = "Bog Enlightenment",
        desc = "Create 1 {bog_boil}.",
        flavour = "'And I became the antennae to attend to even more of the swamp, from places impossibly far reaching.'",

        cost = 1,
        flags = CARD_FLAGS.EXPEND,
        rarity = CARD_RARITY.UNIQUE,

        OnPostResolve = function( self, minigame, targets )
            self.negotiator:CreateModifier("bog_boil", 1, self)
        end,
    },
}

for i, id, data in sorted_pairs( PARASITES ) do
    data.series = "ROOK"
    if data.rarity ~= CARD_RARITY.UNIQUE then
        data.max_xp = data.max_xp or 4
    end

    data.flags = (data.flags or 0) | CARD_FLAGS.PARASITE
    assert(data.rarity, loc.format("Rarity missing on {1}", data.name))
    
    Content.AddNegotiationCard( id, data )
end



