local negotiation_defs = require "negotiation/negotiation_defs"
local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT
local MODIFIER_TYPE = negotiation_defs.MODIFIER_TYPE

local CARDS =
{
    friendly_support = 
    {
        name = "Friendly Support",
        quip = "support",
        
        min_persuasion = 3,
        max_persuasion = 6,
    },

    untrained_pet =
    {
        OnPostResolve = function( self )
            local id = string.format( "untrained_pet_plus_%s", self.owner:GetContentID():lower() )
            id = id:gsub("_upgraded", "" )
            self.negotiator:CreateModifier( id, nil, self )
        end,

        modifier =
        {
            name = "Untrained Pet",
            desc = "Destroy this argument at the start of your next turn.",
            sound = "event:/sfx/battle/cards/neg/create_argument/unruly_vroc",
            icon = "negotiation/modifiers/untrained_crayote.tex", -- Just so this argument doesn't show up as 'missing'.

            max_resolve = 3,
            damage = 3,
            target_enemy = TARGET_ANY_RESOLVE,
            
            OnSetStacks = function( self, old_stacks )
                self.min_persuasion, self.max_persuasion = self.stacks * self.damage, self.stacks * self.damage
            end,

            OnEndTurn = function( self, minigame )
                self:ApplyPersuasion()
            end,

            OnBeginTurn = function( self, minigame )
                if minigame:GetTurns() > 1 then
                    self.negotiator:RemoveModifier( self )
                end
            end,
        }
    },

    trained_pet =
    {
        OnPostResolve = function( self )
            local id = string.format( "trained_pet_plus_%s", self.owner:GetContentID():lower() )
            id = id:gsub("_upgraded", "" )
            self.negotiator:CreateModifier( id, nil, self )
        end,

        modifier =
        {
            name = "Trained Pet",
            sound = "event:/sfx/battle/cards/neg/create_argument/unruly_vroc",
            icon = "negotiation/modifiers/trained_crayote.tex", -- Just so this argument doesn't show up as 'missing'.

            max_resolve = 3,
            damage = 3,
            target_enemy = TARGET_ANY_RESOLVE,

            OnSetStacks = function( self, old_stacks )
                self.min_persuasion, self.max_persuasion = self.stacks * self.damage, self.stacks * self.damage
            end,

            OnEndTurn = function( self, minigame )
                self:ApplyPersuasion()
            end,
        }
    },

    bartender_support = 
    {
        quips = 
        {
            {
                [[I'll vouch for you, {player.name}.]],
                [[Want to be served in my bar? Better listen up.]],
            }
        },
        name = "Bartender Support",

        OnPostResolve = function( self )
            self.negotiator:AddModifier( "INFLUENCE", 3, self )
        end,
    },

    foreman_support =
    {
        name = "Foreman Support",

        OnPostResolve = function( self )
            self.negotiator:AddModifier( "FREE_ACTION", 1, self )
            self.engine:BroadcastEvent( EVENT.CUSTOM, function( panel )
                panel:ShowCenterFloater( loc.format( LOC "UI.NEGOTIATION_PANEL.FREE_ACTION", 1 ), UICOLOURS.BONUS )
                end )
        end,
    },

    chemist_support = 
    {
        quips =
        {
            {
                pre = "!point",
                [[Open your mind!]],
                [[We're not here to brew up any trouble.]],
            }
        },
        name = "Chemist Support",
        
        OnPostResolve  = function( self, minigame )
            minigame:GetPlayerNegotiator():ModifyResolve( minigame:GetDifficulty(), self )
        end,
    },

    spree_support = 
    {
        quips = 
        {
            {
                [[{player.name} knows how things work around here.]],
                [[Pay attention if you don't want trouble coming your way.]],
            }
        },
        name = "Spree Support",
        
        OnPostResolve = function( self )
            self.negotiator:AddModifier( "DOMINANCE", 1, self )
        end,
    },

    jakes_support =
    {
        name = "Jakes' Support",
        
        OnPostResolve = function( self )
            self.anti_negotiator:AddModifier( "FLUSTERED", 2, self )
        end
    },

    hesh_support =
    {
        name = "Hesh Support",
        
        OnPostResolve = function( self )
            self.negotiator:AddModifier( "prayer_of_hesh", 1, self )
        end
    },

    admiralty_support = 
    {
        name = "Wealth Extractor",
    
        OnPostResolve = function( self )
            self.anti_negotiator:CreateModifier( self.id, nil, self )
        end,

        modifier =
        {
            desc = "When destroyed, receive {1#money}.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.money )
            end,

            sound = SoundEvents.conversation_accept_money,
            remove_fx = { animfile = "anim/fx_coin_use.zip", anim = "idle" },

            max_resolve = 5,
            money = 30,
            modifier_type = MODIFIER_TYPE.BOUNTY,

            sound = "event:/sfx/battle/cards/neg/create_argument/wealth_extractor",

            OnBounty = function( self )
                self.engine:ModifyMoney( self.money )
            end,
        }
    },

    -- rise_support
    rabble_rouser =
    {
        name = "Rabble Rouser",

        OnPostResolve = function( self )
            self.negotiator:AddModifier(self.id, 1, self)
        end,

        modifier = 
        {
            desc = "Draw <#HILITE>{1}</> additional {1*card|cards} every 2nd turn.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self.stacks)
            end,

            max_resolve = 3,

            event_handlers =
            {
                [ EVENT.CALC_TURN_DRAW ] = function( self, acc )
                    if self.engine:GetTurns() % 2 == 0 then
                        acc:AddValue( self.stacks or 1, self )
                    end
                end,
            },
        },
    },

    -- spark_baron_support
    double_duty =
    {
        name = "Double Duty",

        OnPostResolve = function( self )
            self.negotiator:AddModifier( self.id, 2, self )
        end,
 
        modifier =
        {
            desc = "The first card played at the start of your turn is played twice.",
            max_resolve = 2,

            event_priorities =
            {
                [ EVENT.POST_RESOLVE ] = EVENT_PRIORITY_CLAMP,
            },
            
            event_handlers =
            {
                [ EVENT.START_RESOLVE ] = function( self, minigame, card )
                    if card.owner == self.owner and #minigame.cards_played == 0 then
                        -- Play it again.
                        self.negotiator:RemoveModifier( self, 1 )
                        minigame:PlayCard( card )
                    end
                end,
            },
        }
    },

    --laborer_support
    stoic =
    {
        name = "Stoic",

        OnPostResolve = function( self )
            self.negotiator:AddModifier( self.id, nil, self )
        end,
        
        modifier =
        {
            desc = "Gain {COMPOSURE {1}} at the beginning of each turn.\nDamage spills over to this argument before the core argument.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.stacks )
            end,

            max_resolve = 4,

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        self:DeltaComposure( self.stacks, self )
                    end
                end,

                [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                    if modifier.owner == self.owner and params.splashed_modifier and not self:GetShieldStatus() then
                        params.splashed_modifier = self
                    end
                end,
            }                    
        }
    },

    deep_pockets = 
    {
        name = "Deep Pockets",
        quips =
        {
            {
                [[We can all get what we want here.]],
                [[Any problems can be solved with shills, yes?]],
                [[Business. It's just business.]],
            }
        },

        OnPostResolve = function( self )
            self.negotiator:AddModifier( self.id, 2, self )
        end,
        
        modifier =
        {
            desc = "At the start of your turn, add one {hush_money} card to your hand and reduce this by one.",
            desc_fn = function( self, fmt_str )
                return loc.format( fmt_str, self.stacks )
            end,

            max_resolve = 3,

            event_handlers =
            {
                [ EVENT.BEGIN_TURN ] = function( self, minigame, negotiator )
                    if negotiator == self.negotiator then
                        if self.skipped_turn then
                            minigame:DealCard(Negotiation.Card("hush_money", self.owner), minigame:GetHandDeck())  
                            self.negotiator:RemoveModifier(self.id, 1, self)
                        else
                            self.skipped_turn = true
                        end
                    end
                end
            }
        }
    },

    bog_support =
    {
        name = "Bogger Support",
        
        OnPostResolve  = function( self, minigame )
            local card = Negotiation.Card( "bog_blast", minigame:GetPlayer() )

            minigame:GetDrawDeck():InsertCard( card, 1 )
        end,
    },

    formality =
    {
        name = "Formality",

        OnPostResolve = function( self, minigame )
            self.negotiator:AddModifier( self.id, 1, self )
        end,

        modifier = 
        {
            name = "Formality",
            desc = "Whenever {1}'s arguments are destroyed, splash damage is ignored.",
            desc_fn = function( self, fmt_str )
                return loc.format(fmt_str, self:GetOwnerName() )
            end,

            OnSetStacks = function( self, old_stacks )
                if self.engine then
                    local delta = self.stacks - old_stacks
                    self:ModifyResolve( delta * 2 )
                end
            end,

            max_resolve = 2,

            event_handlers =
            {
                [ EVENT.SPLASH_RESOLVE ] = function( self, modifier, overflow, params )
                    if modifier.owner == self.owner and params.splashed_modifier then
                        params.splashed_modifier = nil -- No splash damage.
                    end
                end,
            }   
        },
    },
}

for id, carddef in pairs( CARDS ) do
    carddef.show_dealt = false
    carddef.quip = "support"
    carddef.rarity = carddef.rarity or CARD_RARITY.UNIQUE
    carddef.flags = (carddef.flags or 0) | CARD_FLAGS.BYSTANDER
    carddef.series = CARD_SERIES.GENERAL

    Content.AddNegotiationCard( id, carddef )
end
