-----------------------------------------------------------
-- Data associated for logged GameState.notifications.
-- These tables are responsible for generating the necessary UI information
-- for presenting a notification at run-time.

-- Existence of the following fields determine what kind of presentation applies.

-- Widget.NotificationLog
--  log_txt : the string shown
--  img: the associated icon
--  details: the tooltip text for the log entry

-- Widget.InlineNotification
--  banner_txt : the title
--  details : the body text
--  delta : the "delta" text
--  sfx : the accompanying SoundEvent to play.
--  img : the "img" shown (see Widget.InlineNotification)

-------------------------------------------------------------
NOTIFY = {}
function AddNotification(id, t)
    t.id = id

    if t.auto_dismiss_time == nil and not t.manual_dismiss then
        t.auto_dismiss_time = INLINE_NOTIFICATION_DISMISS_TIME
    end
    assert( t.banner_txt or t.log_txt or t.FormatNotification, id )
    if NOTIFY[id] then
        print ("RE-DEFININING NOTIFICATION", id)
    end
    NOTIFY[id] = t
end


local common_notifications =
{
    DEBUG =
    {
        FormatNotification = function( self, notification, txt )
            notification.log_txt = txt or "DEBUG_NOTIFICATION"
            notification.banner_txt = txt or "DEBUG_NOTIFICATION"
        end,
    },

    AGENT_MSG =
    {
        FormatNotification = function( self, notification, agent, txt )
            notification.log_txt = txt
            notification.banner_txt = txt
            notification.img = agent
        end,
    },

    GRAFT_ADDED = {
	
		sfx = SoundEvents.notification_graft_installed,
	
        FormatNotification = function( self, notification, agent, graft )
            if graft:GetType() == GRAFT_TYPE.COIN then
                notification.banner_txt = LOC"NOTIFICATION.COIN_ADDED"
                notification.details = loc.format( LOC"NOTIFICATION.COIN_ADDED_DESC", graft.id )
                notification.img = graft:GetImage()
            
            elseif graft:GetType() == GRAFT_TYPE.COMBAT or graft:GetType() == GRAFT_TYPE.NEGOTIATION then
                notification.banner_txt = LOC"NOTIFICATION.GRAFT_ADDED"
                notification.details = loc.format( LOC"NOTIFICATION.GRAFT_ADDED_DESC", graft.id )
                notification.img = graft:GetImage()
            end
        end,
    },
    
    GRAFT_REMOVED = {
	
		sfx = SoundEvents.notification_graft_removed,
	
        FormatNotification = function( self, notification, agent, graft )
            if graft:GetType() ~= GRAFT_TYPE.SOCIAL and graft:GetType() ~= GRAFT_TYPE.MUTATOR and graft:GetType() ~= GRAFT_TYPE.COIN then
                notification.banner_txt = LOC"NOTIFICATION.GRAFT_REMOVED"
                notification.details = loc.format( LOC"NOTIFICATION.GRAFT_REMOVED_DESC", graft.id )
                notification.img = graft:GetImage()
            end
        end,
    },

    GRAFT_UPGRADED = {
    
        sfx = SoundEvents.notification_graft_installed,
    
        FormatNotification = function( self, notification, agent, graft )
            if graft:GetType() ~= GRAFT_TYPE.SOCIAL and graft:GetType() ~= GRAFT_TYPE.MUTATOR and graft:GetType() ~= GRAFT_TYPE.COIN then
                notification.banner_txt = LOC"NOTIFICATION.GRAFT_UPGRADED"
                notification.details = loc.format( LOC"NOTIFICATION.GRAFT_UPGRADED_DESC", graft.id )
                notification.img = graft:GetImage()
            end
        end,
    },

    RELATIONSHIP_CHANGED =
    {
        sfx = SoundEvents.notification_relationship_new,
        
        FormatNotification = function( self, notification, agent, old_rel, new_rel )
                
                notification.img = agent
                local rel_strings = {
                    [RELATIONSHIP.HATED] = LOC"NOTIFICATION.OPINION_NEW_RELATIONSHIP_DESC_HATED",
                    [RELATIONSHIP.DISLIKED] = LOC"NOTIFICATION.OPINION_NEW_RELATIONSHIP_DESC_DISLIKED",
                    [RELATIONSHIP.NEUTRAL] = LOC"NOTIFICATION.OPINION_NEW_RELATIONSHIP_DESC_NEUTRAL",
                    [RELATIONSHIP.LIKED] = LOC"NOTIFICATION.OPINION_NEW_RELATIONSHIP_DESC_LIKED",
                    [RELATIONSHIP.LOVED] = LOC"NOTIFICATION.OPINION_NEW_RELATIONSHIP_DESC_LOVED",
                }
                notification.banner_txt = loc.format( rel_strings[new_rel], agent )
                

                if new_rel == RELATIONSHIP.LOVED then
                    notification.sfx = SoundEvents.notification_relationship_loved
                elseif new_rel == RELATIONSHIP.HATED then
                    notification.sfx = SoundEvents.notification_relationship_hated
                elseif new_rel > old_rel then
                    notification.sfx = SoundEvents.notification_relationship_friendship_up
                elseif old_rel > new_rel then
                    notification.sfx = SoundEvents.notification_relationship_friendship_down
                end

                local new_graft = agent:GetSocialGraft(new_rel) and Content.GetGraft(agent:GetSocialGraft(new_rel))
                local old_graft = agent:GetSocialGraft(old_rel) and Content.GetGraft(agent:GetSocialGraft(old_rel))

                local graft = new_graft or old_graft
                if graft then
                    local is_good = graft.is_good
                    local removed = graft == old_graft 
                    notification.boon_bane_id = graft.id
                    if removed then
                        notification.details = loc.format( is_good and LOC"NOTIFICATION.GOOD_SOCIAL_GRAFT_REMOVED_DESC" or LOC"NOTIFICATION.BAD_SOCIAL_GRAFT_REMOVED_DESC", graft.id) 
                    else
                        notification.details = loc.format( is_good and LOC"NOTIFICATION.GOOD_SOCIAL_GRAFT_ADDED_DESC" or LOC"NOTIFICATION.BAD_SOCIAL_GRAFT_ADDED_DESC", graft.id) 
                    end
                    --notification.banner_txt = is_good and LOC"NOTIFICATION.GOOD_SOCIAL_GRAFT_REMOVED" or LOC"NOTIFICATION.BAD_SOCIAL_GRAFT_REMOVED"
                    --notification.details = loc.format( is_good and LOC"NOTIFICATION.GOOD_SOCIAL_GRAFT_REMOVED_DESC" or LOC"NOTIFICATION.BAD_SOCIAL_GRAFT_REMOVED_DESC", graft:GetName(), graft:GetDesc() )
                end

        end,
    },

    FRIEND_KILLED =
    {
        FormatNotification = function( self, notification, agent )
            notification.log_txt = loc.format(LOC"NOTIFICATION.FRIEND_KILLED", agent)
            notification.banner_txt = notification.log_txt
            notification.img = agent
        end,
    },
    
    ENEMY_KILLED =
    {
        FormatNotification = function( self, notification, agent )
            notification.log_txt = loc.format(LOC"NOTIFICATION.ENEMY_KILLED", agent)
            notification.banner_txt = notification.log_txt
            notification.img = agent
        end,
    },

    PARTY_MEMBER_ADDED =
    {
		sfx = SoundEvents.notification_partymember_added,
			
        FormatNotification = function( self, notification, agent )
            if not agent:IsCaptiveMember() then
                notification.banner_txt = LOC"NOTIFICATION.PARTY_MEMBER_ADDED"
            end
            notification.details = loc.format(LOC"NOTIFICATION.PARTY_MEMBER_ADDED_FMT", agent)
            notification.img = agent
        end,
    },

    PARTY_MEMBER_REMOVED =
    {
        sfx = SoundEvents.notification_partymember_removed,
            
        FormatNotification = function( self, notification, agent )
            notification.banner_txt = LOC"NOTIFICATION.PARTY_MEMBER_REMOVED"
            notification.details = loc.format(LOC"NOTIFICATION.PARTY_MEMBER_REMOVED_FMT", agent)
            notification.img = agent
        end,
    },

    DELTA_MONEY =
    {
        img = global_images.money,

        OnCreate = function( self, data, gs )
            data.total_money = gs:GetCaravan():GetMoney()
        end,
		
		sfx = SoundEvents.notification_money_gained,
        fx = { animfile = "anim/fx_coin_use.zip", anim = "idle", scale = 0.6, bloom = 0.05 },

        FormatNotification = function( self, notification, amt )
            if amt > 0 then
                notification.banner_txt = LOC"NOTIFICATION.GAINED_MONEY"
                notification.details = loc.format(LOC"NOTIFICATION.GAINED_MONEY_FMT", amt)
				notification.sfx = SoundEvents.notification_money_gained
            elseif amt < 0 then
                notification.banner_txt = loc.format(LOC"NOTIFICATION.LOST_MONEY", -amt)
                notification.details = loc.format(LOC"NOTIFICATION.LOST_MONEY_FMT", -amt)
				notification.sfx = SoundEvents.notification_money_lost
            end
            notification.log_txt = notification.details
            notification.delta = loc.format("{1%+d}", amt)
            notification.extra_txt = loc.format(LOC"NOTIFICATION.TOTAL_MONEY", notification.data.total_money )
        end
    },

    GAINED_METTLE =
    {
        img = global_images.mettle,

        OnCreate = function( self, data, gs )
            data.total = TheGame:GetGameProfile():GetMettlePoints()
        end,
        
        sfx = "event:/ui/mettle_shop/gain_mettlepoints",
        --fx = { animfile = "anim/mettle_gain_fx.zip", anim = "idle", scale = 0.6, bloom = 0.05 },
        FormatNotification = function( self, notification, amt )
            notification.banner_txt = LOC"NOTIFICATION.GAINED_METTLE"
            notification.details = loc.format(LOC"NOTIFICATION.GAINED_METTLE_FMT", amt)
            notification.delta = loc.format("{1%+d}", amt)
            notification.extra_txt = loc.format(LOC"NOTIFICATION.TOTAL_METTLE", notification.data.total )
        end
    },


    COMPLETED_JOB =
    {
	    sfx = SoundEvents.notification_job_success,
		
        FormatNotification = function( self, notification, quest )
            notification.banner_txt = LOC"NOTIFICATION.COMPLETED_JOB"
            notification.details = loc.format(LOC"NOTIFICATION.COMPLETED_JOB_DESC", quest )
            notification.img = quest:GetIcon()
        end
    },

    FAILED_JOB =
    {
		sfx = SoundEvents.notification_job_failure,
			
        FormatNotification = function( self, notification, quest )
            local fail_string = quest:GetCompletionString()
            if fail_string and fail_string ~= "" then
                notification.banner_txt = LOC"NOTIFICATION.FAILED_JOB_QUEST_REASON"
                notification.details = loc.format(LOC"NOTIFICATION.FAILED_JOB_QUEST_REASON_DESC", quest, fail_string)
            else
                notification.banner_txt = LOC"NOTIFICATION.FAILED_JOB_QUEST"
                notification.details = loc.format(LOC"NOTIFICATION.FAILED_JOB_QUEST_DESC", quest)
            end
            notification.img = quest:GetIcon()
        end
    },

    NEGOTIATED =
    {
        --details = "You completed a negotiation with {1#agent}",
        FormatNotification = function( self, notification, agent, result )
            local negotiation_defs = require "negotiation/negotiation_defs"
            if result == negotiation_defs.RESULT.WIN then
                notification.sfx = "event:/ui/notification/negotiation_success"
                notification.banner_txt = LOC"UI.NEGOTIATION_PANEL.NEGOTIATION_SUCCESS"
                notification.log_txt = loc.format( LOC"NOTIFICATION.NEGOTIATED", agent, "Success" )

            elseif result == negotiation_defs.RESULT.LOSE then
                notification.sfx = "event:/ui/notification/negotiation_failure"
                notification.banner_txt = LOC"UI.NEGOTIATION_PANEL.NEGOTIATION_FAILURE"
                notification.log_txt = loc.format( LOC"NOTIFICATION.NEGOTIATED", agent, "Failure" )
            end
            --notification.details = loc.format( self.details, agent )
            notification.img = agent
        end,
    },

    RESOLVE_UPGRADED =
    {
        sfx = SoundEvents.opinion_raise,
        img = global_images.resolve,

        FormatNotification = function( self, notification, amount )
            notification.banner_txt = loc.format( LOC"NOTIFICATION.RESOLVE_UPGRADED", amount )
            notification.details = loc.format( LOC"NOTIFICATION.RESOLVE_UPGRADED_DETAILS", amount )
        end,
    },

    HEALTH_UPGRADED =
    {
        sfx = SoundEvents.opinion_raise,
        img = global_images.health,

        FormatNotification = function( self, notification, amount )
            notification.banner_txt = loc.format( LOC"NOTIFICATION.HEALTH_UPGRADED", amount )
            notification.details = loc.format( LOC"NOTIFICATION.HEALTH_UPGRADED_DETAILS", amount )
        end,
    },

    GRAFT_SLOT_UPGRADED =
    {
        sfx = SoundEvents.notification_aspect_gained,
        img = global_images.graft,

        FormatNotification = function( self, notification, graft_type )
            if graft_type == GRAFT_TYPE.COMBAT then
                notification.banner_txt = loc.format( LOC"NOTIFICATION.COMBAT_GRAFT_SLOT" )
                notification.details = loc.format( LOC"NOTIFICATION.COMBAT_GRAFT_SLOT_DETAILS" )
            elseif graft_type == GRAFT_TYPE.NEGOTIATION then
                notification.banner_txt = loc.format( LOC"NOTIFICATION.NEGOTIATION_GRAFT_SLOT" )
                notification.details = loc.format( LOC"NOTIFICATION.NEGOTIATION_GRAFT_SLOT_DETAILS" )
            end
        end,
    },
}

-----------------------------------------------------------

for k,v in pairs(common_notifications) do
    AddNotification(k, v)
end
