local memutil = require "util/memutil"

---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
-- Console Functions -- These are simple helpers made to be typed at the console.
-- These defined within the scope of a special debug environment, they are NOT
-- accessible via _G.
--
-- The function prefix denotes the general usage class for the benefit of readability
-- and autocomplete.
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------

--------------------------------------------------------------
-- General console miscellanea

function c_givecards()
	if gs == nil then
		return
	end
	local negotiation_defs = require "negotiation/negotiation_defs"
	local negotiator = gs:GetPlayerAgent().negotiator

	local battle_defs = require "battle/battle_defs"
	local battler = gs:GetPlayerAgent().battler

	local player = gs:GetPlayerAgent()
	gs:GetNotifications():EnableNotifications( false )

	for i, carddef in ipairs( Content.GetAllNegotiationCards() ) do
		local ok = AgentUtil.CanUseCard( player, carddef )
		ok = ok and not CheckAnyBits( carddef.flags, negotiation_defs.CARD_FLAGS.UPGRADED | negotiation_defs.CARD_FLAGS.BYSTANDER )
		if ok and negotiator:GetCardCount( carddef.id ) == 0 then
			negotiator:LearnCard( carddef.id )
		end
	end

	for i, carddef in ipairs( Content.GetAllBattleCards() ) do
		local ok = AgentUtil.CanUseCard( player, carddef )
		ok = ok and not CheckAnyBits( carddef.flags, battle_defs.CARD_FLAGS.UPGRADED | battle_defs.CARD_FLAGS.SPECIAL )
		if ok and battler:GetCardCount( carddef.id ) == 0 then
			battler:LearnCard( carddef.id )
		end
	end

	gs:GetNotifications():EnableNotifications( true )
	gs:LogNotification( NOTIFY.DEBUG, "You got all the cards!" )
end

function c_export_agent_defs()
	local tabular = DataUtil.AgentDefs()
	tabular:SaveToCSV( "agentdefs.csv" )

	print( "Exported to agentdefs.csv in your root directory.")
	print( "See data_util.lua for related code." )
end


local CSV_COLUMNS = {
    { name="UUID", fn = function(skin) return skin, skin.uuid end },
    { name="Name", fn = function(skin) return skin, skin.name end },
    { name="Def", fn = function(skin) return skin, skin.id end },
    { name="Gender", fn = function(skin) return skin, skin.gender end },
    { name="Species", fn = function(skin) return skin, skin.species end },
    { name="Bio", fn = function(skin) return skin, skin.bio or "" end },
    { name="Loved Bio", fn = function(skin) return skin, skin.loved_bio or "" end },
    { name="Hated Bio", fn = function(skin) return skin, skin.hated_bio or "" end },
    { name="Small Talk", fn = function(skin) return skin, skin.smalltalk or "" end },
}

local function ProcessSkinProperty( results, skin, value, key )
	if value then
		value = value:gsub( "\\n", "\n" )
		-- This processing is so to ensure that if we import the output from c_export_skins()
		-- directly, we apply an identity transformation.
		-- Data processed from a spreadsheet will typically already perform these edits.

		-- (1) Unwrap enclosing quotes that are used to preserve columns.
		value = value:match( "^%s*\"(.+)\"%s*$" ) or value
		-- (2) condense double-quotations
		value = value:gsub( "\"\"", "\"" )
		-- (3) finally, let's trim
		value = value:trim()
	else
		value = ""
	end
	
	if skin[ key ] ~= value and #value > 0 then
		local result = { skin = skin, key = key, value = value }
		table.insert( results, result )
	end
end

function c_import_skins( filepath )
	local str
	if filepath then
		local lines = {}
		for line in io.lines( filepath ) do
			table.insert( lines, line )
		end
		str = table.concat( lines, "\n" )
	else
		str = engine.inst:GetClipboardText()
	end

	local function tab_split( s )
	   local sep, fields = "\t", {}
	   local pattern = string.format("%s?([^%s]*)%s-", sep, sep, sep)
	   s:gsub(pattern, function(c) fields[#fields+1] = c end)
	   return fields
	end

	local function line_split( s )
		local lpeg = require "lpeg"
	 	local sep = lpeg.P( "\n" )
		local uuid = (lpeg.locale().alnum + lpeg.S("-"))^1
		local quotes = lpeg.P{ "\"" * ((1 - lpeg.S"\"") + lpeg.V(1))^0 * "\"" }
		local elem = lpeg.C(uuid * (quotes + (1 - sep))^1)
	 	local p = lpeg.Ct(elem * (sep * elem)^0)
	 	return lpeg.match(p, s)
	end

	local lines = line_split( str )
	local results = { string.format( "%d lines read.", #lines )}
	local line_count = 0
	for i, line in ipairs( lines ) do
		local cols = tab_split( line )
		local uuid = cols[1]
		local skin = Content.GetCharacterSkin( uuid )
		if skin then
			line_count = line_count + 1
			ProcessSkinProperty( results, skin, cols[2], "name" )
			ProcessSkinProperty( results, skin, cols[6], "bio" )
			ProcessSkinProperty( results, skin, cols[7], "loved_bio" )
			ProcessSkinProperty( results, skin, cols[8], "hated_bio" )
			ProcessSkinProperty( results, skin, cols[9], "smalltalk" )
		else
			if uuid ~= "UUID" then -- This dumb check just ignores the 'header' row in case you pasted that.
				table.insert( results, string.format( "Invalid skin: %s (%d columns recognized)", tostring(uuid), #cols ))
			end
		end
	end

	local function RenderImport( self, ui, panel )
		local edit_count = 0
		for i, result in ipairs( results ) do
			if type(result) == "string" then
				ui:Text( result )

			elseif result.skin then
				ui:Text( string.format( "[%s] %s's '%s' changed to:", tostring(result.skin.id), result.skin.name, result.key ))
				ui:SameLine( nil, 10 )
				ui:TextColored( 0x00CCCCFF, tostring(result.value))
				edit_count = edit_count + 1
			end
		end

		if edit_count > 0 then
			if ui:Button( "Commit" ) then
				for i, result in ipairs( results ) do
					if result.skin then
						result.skin[ result.key ] = result.value
					end
				end
				Content.SaveAllCharacterSkins()
				table.clear( results )
				table.insert( results, string.format( "Saved and committed %d edits.", edit_count ))
			end
		else
			ui:Text( "No edits." )
		end
	end

	DBG( RenderImport )
end

function c_export_skins( filepath )
    local all_skins = {}
    for content_id, skins in pairs( Content.GetAllCharacterSkins() ) do
        for i, skin in ipairs( skins ) do
            table.insert( all_skins, skin )
        end
    end

    local tab = DebugTabular( CSV_COLUMNS, all_skins )
    if filepath then
		tab:SaveToCSV( filepath )
		DBG( "Saved to csv: ".. filepath )
	else
		DBG( string.format( "Saved skins to clipboard: %d skins", #all_skins ))
	    local str = tab:GetString( "\t" )
		engine.inst:SetClipboardText( str )
	end
end


function c_export_cards()
	local tabular = DataUtil.AllCards()
	tabular:SaveToCSV( "cards.csv" )

	print( "Exported to cards.csv in your root directory.")
	print( "See data_util.lua for related code." )
end

function c_export_grafts()
	local tabular = DataUtil.AllGrafts()
	tabular:SaveToCSV( "grafts.csv" )

	print( "Exported to grafts.csv in your root directory.")
	print( "See data_util.lua for related code." )
end
function c_unlockall()
	TheGame:GetGameProfile():DeltaPoints( nil, math.huge )

	TheGame:FE():PushScreen(Screen.InfoPopup( "All Unlocks Acquired", "", "Okay"))
end

function c_count_perks()
	local total_points = 0
	local total_grifts = 0
	for i, def in ipairs( Content.GetAllGrifts() ) do
		total_points = total_points + (def.perk_points or 0)
		total_grifts = total_grifts + 1
	end

	local total_cost = 0
	local total_perks = 0
	for i, def in ipairs( Content.GetAllGrafts() ) do
		if def.type == GRAFT_TYPE.PERK then
			total_cost = total_cost + (def.unlock_cost or 0)
			total_perks = total_perks + 1
		end
	end

	print( string.format( "Total Points Available across %d grifts: %d", total_grifts, total_points ))
	print( string.format( "Total Cost for %d Perks: %d", total_perks, total_cost ))
end

function c_givemoney( amount )
	amount = amount or 1000
	local gamestate = game:GetGameState()
	local inv = gamestate:GetCaravan():AddMoney(amount)
	
	TheGame:GetEvents():BroadcastEvent( "update_overlay" )
end

function c_learncard( id )
	local def = Content.GetBattleCard( id )
	if def then
		gs:GetPlayerAgent().battler:LearnCard( id )
	end

	local def = Content.GetNegotiationCard( id )
	if def then
		gs:GetPlayerAgent().negotiator:LearnCard( id )
	end
end

function c_simrate( delta )
	if sim then
		sim:ChangeSimRate( delta )
	end
end

function c_decode_save( filename )
	local to_filename = filename .. ".lua"

	local function OnSave( ok )
		print( string.format( "Saved '%s': %s", to_filename, ok and "SUCCESS" or "FAIL" ))
	end

    local function OnLoad(success, filedata)
	    if success and filedata then
	    	print( string.format( "Loaded '%s': SUCCESS (%d bytes)", filename, #filedata ))
	    	engine.savegamestorage:SaveFile( to_filename, filedata, OnSave )
	    else
	    	print( string.format( "Loaded '%s': FAILED", filename ))
	    end
	end

    engine.savegamestorage:LoadFile( filename, OnLoad, true )
end

function c_dump_save( filename )
    local function OnLoad(success, filedata)
	    if success and filedata then
	    	print( string.format( "Loaded '%s': SUCCESS (%d bytes)", filename, #filedata ))

            local header
            load_ok, header = pcall(loadstring(filedata))
            
            if not load_ok then
            	print( "loadstring() failed: ", tostring(header))

            else
            	print( string.format( "Version: %s (Current: %s)", tostring(header.save_version), SAVE_VERSION ))
            	if header.time then
	            	print( "Time:", os.date( nil, header.time ))
	            else
	            	print( "Time: nil")
	            end
	            print( "Build ID:", header.build_id )
	            print( "Desc:", header.desc )
	            print( "Options:", tostr(header.options) )

            	if type(header.gamestate) ~= "function" then
	            	print( "Corrupted file; missing gamestate loader")
	            end
			end
	
	    else
	    	print( string.format( "Loaded '%s': FAILED", filename ))
	    end
	end

    engine.savegamestorage:LoadFile( filename, OnLoad, true )
end

function c_cotd( time )
	time = time or os.time()
	time = time - (time % (60*60*24))
	local rng = engine.Random( time )
	
	local out = {}
	table.insert( out, string.format( "Date: %s", os.date( nil, time )))

	local battle_defs = require "battle/battle_defs"
	local battle_cards = {}
	for i, def in ipairs( Content.GetAllBattleCards() ) do
		if not CheckBits( def.flags, battle_defs.CARD_FLAGS.UPGRADED ) and not CheckBits( def.flags, battle_defs.CARD_FLAGS.NPC ) then
			table.insert( battle_cards, def.id )
		end
	end
	local id = battle_cards[ rng:Random( #battle_cards ) ]
	local card = Battle.Card( id )
	table.insert( out, string.format( "## Battle Card: %s" , id ))
	table.insert( out, loc.format( "\t{1} (Cost: {2}) - {3} {4}", card:GetName(), card.cost, card.rarity, GetCardType( card ) ))
	if card.min_damage then
		table.insert( out, loc.format( "\tDamage: {1}-{2}", card.min_damage, card.max_damage ))
	end
	table.insert( out, "\t".. SetFormattedText( nil, card:GetDesc() ))

	local negotiation_defs = require "negotiation/negotiation_defs"
	local negotiation_cards = {}
	for i, def in ipairs( Content.GetAllNegotiationCards() ) do
		if not CheckBits( def.flags, negotiation_defs.CARD_FLAGS.UPGRADED ) and not CheckBits( def.flags, negotiation_defs.CARD_FLAGS.OPPONENT ) then
			table.insert( negotiation_cards, def.id )
		end
	end

	print( "Total battle/negotiation cards:", #battle_cards, #negotiation_cards )

	local id = negotiation_cards[ rng:Random( #negotiation_cards ) ]
	local card = Negotiation.Card( id )
	table.insert( out, string.format( "## Negotiation Card: %s" , id ))
	table.insert( out, loc.format( "\t{1} (Cost: {2}) - {3} {4}", card:GetName(), card.cost, card.rarity, GetCardType( card ) ))
	if card.min_persuasion then
		table.insert( out, loc.format( "\tDamage: {1}-{2}", card.min_persuasion, card.max_persuasion ))
	end
	table.insert( out, "\t".. SetFormattedText( nil, card:GetDesc() ))

	print(table.concat( out, "\n" ))
end

--------------------------------------------------------------
-- Content functions

function cn_allplax( match )
    if type(match) == "string" then
	    print( string.format( "Listing all plax matching: '%s'", match ))
	else
		print( "Listing all plax:" )
	end
	local count = 0
    for k, filename in Content.AllPlax() do
    	local plax_data = Content.GetPlaxData(k)
    	if match == nil or k:match( match ) then
    		count = count + 1
	    	printf( "%s (%d layers)", k, #plax_data.layers )
	    end
    end
    print( string.format( "%d total matches.", count ))
end


--------------------------------------------------------------
-- Localization functions

function loc_lang( language )
	engine.inst:SetSetting( "general", "language", language or "english" )
	engine.inst:SaveSettings()
	engine.inst:Reset()
end

local function TestGlyphsOfFont(language, fontname, allchars, missing)
    local locinfo = Content.GetLocalizations()[language]

    local missingglyphs = engine.inst:FindMissingGlyphs(locinfo.font_settings[fontname].font, allchars)
    if #missingglyphs > 0 then
        for _, chr in utf8.codes(missingglyphs) do
            local c = utf8.char(chr)
            table.insert( missing, string.format( "Missing: [%s, %s]\tCodepoint: %s (U+%X)\tChar: %s", language, fontname, utf8.codepoint(c), chr, c ))
        end
    end
end


function loc_testglyphs( all )
    local loc_translator = require "util/loc_translator"
    local translator = loc_translator.Translator()    
    local current_language = TheGame:GetLanguage()

    local locinfo = Content.GetLocalizations()
    local missing = {}
    for language, v in pairs(locinfo) do
    	-- Load translated strings.
    	if all or language == current_language then
		    local strings = translator:LoadTranslationPoFiles( v.po_filenames ) or Content.internal.ENGLISH_STRINGS

	    	local ranges = loc_translator.getCodepoints( strings )
	        local all_characters_string = {}
	        for i, range in ipairs( ranges ) do
	        	for j = range[1], range[2] do
	        		table.insert( all_characters_string, utf8.char( j ))
	        	end
	        end
	        all_characters_string = table.concat( all_characters_string, "" )
	        print( "\t", #all_characters_string, " total codepoints" )

		    -- TestGlyphsOfFont( language, "title", all_characters_string )

	    	for alias, fontinfo in pairs( v.font_settings ) do
	    		-- Let's just do 1 font, since in practice they're all the same and it takes a long time.
	    		if alias == "body" then
		            local filename = fontinfo.font
		            print( "\tMap Font:", alias, filename )
		            engine.inst:RegisterFont( filename, alias, fontinfo.sdfthreshold, fontinfo.sdfboldthreshold, fontinfo.sdfshadowthreshold, fontinfo.supportsitalics )
				    TestGlyphsOfFont( language, alias, all_characters_string, missing )
		        end
		    end
	    end
    end

    if all then
	    -- Reload the current localization and its corresponding fonts.
	    TheGame:LoadLocalization( TheGame:GetLanguage() )
	end

	if #missing == 0 then
		print( "No glyphs are missing!" )
	else
	    for i, str in ipairs( missing ) do
	    	print( str )
	    end
	end
end


function loc_countwords()
	local total_keys = 0
	local total_words = 0

	for id, str in pairs( Content.internal.STRINGS ) do
		local _,n = str:gsub("[%a'-_]+","")
		total_words = total_words + n
		total_keys = total_keys + 1
	end

	print( string.format( "There are %d total strings, containing about %d total words.", total_keys, total_words ))
end

-- Dumps all codepoints used by the current language's strings to codepoints.txt
-- Output is suitable for pasting into a .bmfc config file.
-- (MAY TAKE A MOMENT!)
function loc_dumpcodepoints()
	local loc_translator = require "util/loc_translator"
	local ranges = loc_translator.getCodepoints( Content.internal.STRINGS )

	local count = 0
	local strs = {}
	local last_str = ""
	for i, range in ipairs (ranges) do
		if #last_str == 0 then
			last_str = "chars="
		else
			last_str = last_str..","
		end
		if range[1] == range[2] then
			last_str = last_str .. tostring(range[1])
		else
			last_str = last_str .. string.format( "%d-%d", range[1], range[2] )
		end

		if #last_str > 80 then
			table.insert( strs, last_str )
			last_str = ""
		end
		count = count + (range[2] - range[1] ) + 1
	end
	if #last_str > 0 then
		table.insert( strs, last_str )
	end

	local file = io.open( "../codepoints.txt", "w")
	file:write( table.concat( strs, "\n" ))
	file:close()

	print( "TOTAL CODEPOINTS:", count )
end
--------------------------------------------------------------
-- Debug framework functions

function dbg_toggle()
	dbg:ToggleDebugFlags( DBG_FLAGS.DETECT_ALL )
	if sim then
	    sim:ToggleDebugController()
	end
end

function dbg_diffcount()
	memutil.CountObjects(true)
end

function dbg_countobj(filter)
	memutil.CountObjects(false, filter)
end

function dbg_find(obj, allow_partial)
	memutil.FindObject(obj, allow_partial)
end

function dbg_counttype(typestr)
	memutil.GetObjectTypeCount(typestr)
end

function dbg_difftrace(objtype)
	objtype = objtype or "table"
	memutil.TraceNewObjects(objtype)
end

function dbg_tracealloc(enable)
	print (enable and "ENABLE TRACE ALLOC" or "DISABLE TRACE ALLOC")
	engine.inst:SetTraceAlloc(enable)
end

function c_countents()
	if not sim then
		print( "NO SIM" )
		return
	end

	local entcounts = {}
	for k,v in pairs(sim.entities) do
		local pref = v.prefab or "NOPRE"
		entcounts[pref] = entcounts[pref] and entcounts[pref] + 1  or 1
	end

	local sorted = {}
	for k,v in pairs(entcounts) do
		table.insert(sorted, {prefab = k, count = v})
	end


	print ("***ENTITY COUNTS******")
	table.sort(sorted, function(a, b) return a.count > b.count end)
	for k,v in ipairs(sorted) do
		print (v.prefab.name, v.count)
	end
	print ("**********************")

end


-- Create a UI panel for a DebugEntity node.
function dbg_entity( guid )
	local entity = sim and sim:GetEntityByGUID( guid )
	if entity then
		entity = entity -- Track the entity in the debug environment.
		dbg:CreatePanel( DebugEntity( entity ))
	else
		print( "No entity with GUID", guid )
	end
end

-- Create a generic table panel for a table.
function dbg_table( t )
	if type(t) == "table" then
		dbg:CreatePanel( DebugTable( t ))
	else
		print( "Not a table:", t )
	end
end

function dbg_panel( value )
	local DebugUtil = require "debug/debug_util"
	local node = DebugUtil.CreateDebugNode( value )
	if node == nil and type(value) == "table" then
		node = DebugTable( value )
	end
	if node then
		dbg:CreatePanel( node )
	else
		print( "Not a debuggable value:", value )
	end
end

function dbg_test( filename )
	filename = "debug/tests/"..filename
	package.loaded[ filename ] = nil
	require( filename ) 
end

-----------------------------------------------------------------------------

-- Shows all functions in the dbg_env.
function help()
	local str, funcs = {}, {}
	for k, v in pairs(dbg:GetDebugEnv()) do
		if k ~= "__index" then
			local clr
			if type(v) == "number" then
				clr = "ff5577"
			elseif type(v) == "string" then
				clr = "777700"
			elseif type(v) == "table" then
				clr = "555555"
			elseif type(v) == "function" then
				table.insert( funcs, k )
			else
				clr = "00ffff"
			end
			if clr then
				table.insert( str, string.format( "%s (%s)", tostring(k), clr, tostring(v) ))
			end
		end
	end
	table.sort( funcs )
	print( "\n\n---FUNCTIONS---\n"..table.concat( funcs, ", ").."\n\n---VARS---\n"..table.concat( str, ", " ))
end