require "debug/debug_panel"
local DebugUtil = require "debug/debug_util"

---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
-- Central debug manager exposing all unified debugging support and options for the
-- game. As much as possible, this is decoupled from the core game code.
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------



local dbg_env =
{
	__index = function( t, k )
		local v = rawget( _G, k )--rawget( t, k ) or rawget( _G, k )
		return v
	end
}
setmetatable( dbg_env, dbg_env )

local MAX_PRINT_HISTORY = 1024
local MAX_CMD_HISTORY = 32
local MAX_DEBUG_PANEL_HISTORY = 5

-------------------------------------------------------------------
--

local DebugManager = class( "DebugManager" )

function DebugManager:init( game )
	assert( game )
	self.debug_flags = 0
	--self.debug_atlas = engine.asset.Atlas('atlases/debug.xml')
	self.game = game
	self.imgui = engine.debug.ImGUI()
	self.debug_bindings = {}
	self.disabled_bindings = {} -- Array of binding IDs that are disabled
	self.debug_panels = {}
	self.closed_debug_panels = {}

	self:LoadDebugInspectors()
	self:LoadConsoleSettings()

	self.print_history = {}
	self.print_history_idx = 1 -- Ringbuffer index.

	local function LOGPRINT(str)
		local history, idx = self.print_history, self.print_history_idx
	    for s in string.gmatch(str.."\n", "(.-)\n") do
		    history[ idx ] = s
		    idx = (idx % MAX_PRINT_HISTORY) + 1
	    end
		self.last_print_output = os.clock()
		self.scroll_to_bottom = true
		self.print_history_idx = idx
	end
	AddPrintListener( LOGPRINT )
end

function DebugManager:Shutdown()
	print( "DebugManager:Shutdown()")
end

function DebugManager:GetDebugEnv()
	dbg_env.dbg = self
    dbg_env.game = self.game
    dbg_env.gs = self.game:GetGameState()
    if dbg_env.gs then
	    dbg_env.caravan = dbg_env.gs:GetCaravan()
	    dbg_env.now = dbg_env.gs:GetDateTime()
	    for i, member in dbg_env.caravan:Members() do
	    	if member:HasAlias() then
		    	dbg_env[ member:GetAlias():lower() ] = member
		    end
		    if member:IsPlayer() then
		    	dbg_env.player = member
		    end
	    end
	end
    dbg_env.fe = self.game:FE()
    if dbg_env.fe then
	    dbg_env.W, dbg_env.H = dbg_env.fe:GetScreenDims()
	    dbg_env.screen = dbg_env.fe:GetTopScreen()
	    if dbg_env.screen then
	    	dbg_env.tt = dbg_env.fe.tooltip_widget
	    	dbg_env.focus = dbg_env.fe:GetFocusWidget()
	    end
	    dbg_env.overlay = dbg_env.fe:FindScreen( Screen.MainOverlay )
		if dbg_env.screen and dbg_env.screen.GetDebugEnv then
			dbg_env.screen:GetDebugEnv( dbg_env )
		end

	    dbg_env.mx, dbg_env.my = self.game:GetInput():GetMousePos()
	end
    return dbg_env
end

function DebugManager:GetDebugUI()
	return self.imgui
end

function DebugManager:ExecuteDebugFile( filename )
	print( "Executing: ", filename )
	if filename then
		local f,e = loadfile( filename, nil, self:GetDebugEnv() )
		if not f then
			error(e, 2)
		end
		return f()
	end
end

function DebugManager:DoRender()
	self.game:FE():DoDebugRender( self )

	if self.console_open then
		self:RenderDebugConsole()
	end

	if not self.console_open or not self.console_settings.docked then
		self:RenderDebugPanels()
	end
end

function DebugManager:LoadDebugInspectors()
	require "debug/debug_nodes"

    for k, filename in pairs(engine.inst:ListFiles( 'scripts/debug/inspectors/*.lua')) do
        local name = filename:match( "^(.+)[.]lua$" )
        if name then
        	filename = string.format( "debug/inspectors/%s", name )
        	require( filename )
        end
    end 
end

function DebugManager:LoadConsoleSettings()
	self.console_settings = DebugUtil.GetLocalDebugData("CONSOLE")
	if self.console_settings == nil then
		self.console_settings = { docked = true }
		DebugUtil.SetLocalDebugData("CONSOLE", self.console_settings)
	end
	if self.console_settings.history == nil then
		self.console_settings.history = {}
	end
	self.history = self.console_settings.history
    self.history_idx = #self.history + 1
end

function DebugManager:SubmitConsoleCommand( str )
	if str ~= self.history[ #self.history ] then
		table.insert( self.history, str )
		while #self.history > MAX_CMD_HISTORY do
			table.remove( self.history, 1 )
		end
		DebugUtil.SaveLocalDebugData()
	end
	self.history_idx = #self.history + 1

	local env = self:GetDebugEnv()
	local fn, err = loadstring(str, "INP", "t", env )
	if not fn then
		print ("ERROR: ", err)
	else
		local ok, res = xpcall( fn, generic_error )
		if not ok then
			print ("\n\n*****ERROR*****\n"..res.."\n\n")
		end
	end		
end

function DebugManager:ToggleDebugConsole()
	self.console_open = not self.console_open
	if self.console_settings.docked then
		self.imgui:ToggleInputPassthrough()
	end
end

function DebugManager:IsConsoleDocked()
	return self.console_open and self.console_settings.docked
end

local function InputTextCallback( self, flags, key, str )
	if flags == self.imgui.TextFlags_CallbackHistory then
		if key == 3 then -- ImGuiKey_UpArrow
            self.history_idx = math.max(1, self.history_idx - 1)
		elseif key == 4 then -- ImGuiKey_DownArrow
            self.history_idx = math.min(#self.history, self.history_idx + 1)
		end
		return self.history[ self.history_idx ]

	elseif flags == self.imgui.TextFlags_CallbackCompletion then
		local AutoComplete = require "util/autocomplete"
		local t = AutoComplete( str, self:GetDebugEnv() )
		return t and t[1]

	elseif flags == self.imgui.TextFlags_CallbackAlways then
		self.last_console_input = str
	end
end

function DebugManager:RenderDebugConsole()
	local ui = self.imgui
	local w, h = ui:DisplaySize()

    local flags = 0
    if self.console_settings.docked then
    	flags = ui.NoTitleBar | ui.NoResize | ui.NoMove | ui.NoScrollbar
		ui:SetNextWindowSize( w, h*(self.console_settings.height or 0.6) )
		ui:SetNextWindowPos( 0, 0 )
    else
		ui:SetNextWindowSize( w*0.8, h*(self.console_settings.height or 0.6), ui.ImGuiSetCond_Once )
		ui:SetNextWindowPos( 0, 0, ui.ImGuiSetCond_Once )
    end


    local show = ui:Begin( "DBG_CONSOLE", flags )
    if not show then
    	self:ToggleDebugConsole()

    elseif show then
    	local offset = -30
    	if self.print_filter then
    		offset = -50
    	end
	    ui:BeginChild("ScrollingRegion", nil, offset, false, ui.WindowFlags_HorizontalScrollbar )
	    -- print_history_idx points at the next line to be replaced in the buffer (ie. the current oldest entry)
	    for i = 0, MAX_PRINT_HISTORY-1 do
	    	local idx = (self.print_history_idx + i) % MAX_PRINT_HISTORY
	    	local line = self.print_history[ idx ]
	    	if line and (self.print_filter == nil or line:find( self.print_filter, 1, true )) then
		        ui:TextWrapped( line )
		    end
	    end
	    if self.scroll_to_bottom then
	    	ui:SetScrollHere()
	    	self.scroll_to_bottom = false
	    end
	    ui:EndChild()

	    local should_focus = self.force_console_focus
	    if self.console_settings.docked then
	    	should_focus = should_focus or not ui:IsAnyItemActive()
	    else
	    	should_focus = should_focus or ui:IsItemClicked()
	    end
        if ui:BeginPopupContextItem( "CXT_MENU" ) then
        	should_focus = false
        	if ui:BeginMenu( "Preferences..." ) then
        		if ui:Checkbox( "Docked", self.console_settings.docked ) then
        			if self.console_open then
        				self:ToggleDebugConsole() -- Turn off first.
        			end
        			self.console_settings.docked = not self.console_settings.docked
        			self:ToggleDebugConsole() -- Turn back on to properly set input passthru.
					DebugUtil.SaveLocalDebugData()
        		end

        		local height = ui:SliderFloat( "Height", self.console_settings.height or 0.6, 0, 1.0 ) or self.console_settings.height
        		if height ~= self.console_settings.height then
        			self.console_settings.height = height
					DebugUtil.SaveLocalDebugData()
				end
        		ui:EndMenu()
        	end
        	if ui:MenuItem( "Scroll to Bottom" ) then
        		self.scroll_to_bottom = true
        	end
        	if ui:MenuItem( "Clear History" ) then
	            table.clear( self.print_history )
	            self.print_history_idx = 1
	        end
	        if ui:BeginMenu( "Filter..." ) then
	            local filter = ui:InputText( "Filter", self.print_filter )
	            if ui:Button( "Clear" ) then
	            	filter = ""
	            end
	            if filter then
	            	if #filter == 0 then
	            		filter = nil
	            	end
		            self.print_filter = filter
	            end
	            ui:EndMenu()
	        end
            ui:EndPopup()
        end

        if self.print_filter then
        	ui:TextColored( 0x00FFFFFF, string.format( "Filtered by: '%s'", self.print_filter ))
        end

	    ui:Separator()
	    ui:PushItemWidth( -10 )
	    local flags = ui.TextFlags_EnterReturnsTrue | ui.TextFlags_CallbackHistory | ui.TextFlags_CallbackCompletion | ui.TextFlags_CallbackAlways
	    local console_input = ui:InputText( "", self.last_console_input, flags, InputTextCallback, self )
	    if console_input then
	        self:SubmitConsoleCommand( console_input )
		    self.scroll_to_bottom = true
		    self.last_console_input = ""
		    self.force_console_focus = true
	    end
	    if should_focus then
	    	self.force_console_focus = nil
	    	ui:SetKeyboardFocusHere(-1)
	    end
	    ui:PopItemWidth()
	end
	ui:End()
end

function DebugManager:RenderDebugPanels()
	local notifications_enabled
    if TheGame:GetGameState() then
    	notifications_enabled = TheGame:GetGameState():GetNotifications():IsEnabled()
        TheGame:GetGameState():GetNotifications():EnableNotifications( false )
    end

	local i = 1
	while i <= #self.debug_panels do
		local panel = self.debug_panels[i]

		local ok, result = xpcall( panel.RenderPanel, generic_error, panel, self.imgui )
		if ok and panel._wants_to_close then
			result = false
		end
		if not ok or not result then
			panel:OnClose()
			table.remove( self.debug_panels, i )
			if not ok then
				print( tostring(result) )
				engine.inst:ShowWarning( tostring(result) )
				break
			end
		else
			i = i + 1
		end
	end

	if dbg_env.show_context then
		self.imgui:OpenPopup( "DBG_CXT" )
		self.debug_contextmenu = dbg_env.show_context
		dbg_env.show_context = nil
	end

	if self.debug_contextmenu and self.imgui:BeginPopup( "DBG_CXT" ) then
		local ok, result = xpcall(  self.debug_contextmenu.RenderContextMenu, generic_error, self.debug_contextmenu, self.imgui )
		if not ok then
			engine.inst:ShowWarning( tostring(result) )
		end
		self.imgui:EndPopup()
	end

    if TheGame:GetGameState() then
        TheGame:GetGameState():GetNotifications():EnableNotifications( notifications_enabled )
    end
end

function DebugManager:WantCaptureMouse()
	return self.imgui:WantCaptureMouse()
end

function DebugManager:Clear()
	self:ClearPanels()
	for k, v in pairs(dbg_env) do
		if type(v) ~= "function" then
			dbg_env[k] = nil
		end
	end
end

function DebugManager:ClearPanels()
	while #self.debug_panels > 0 do
		self:ClosePanel( self.debug_panels[ #self.debug_panels ] )
	end

	if self.debug_contextmenu then
		self.debug_contextmenu:OnClose()
		self.debug_contextmenu = nil
	end

	table.clear( self.closed_debug_panels )
end

function DebugManager:TogglePanel()
	if #self.debug_panels > 0 then
		self:ClosePanel( self.debug_panels[ #self.debug_panels ] )

	else
		self:CreatePanel()
	end
end

function DebugManager:CreatePanel( debug_node )
	if debug_node == nil then
		local screen = TheGame:FE():GetTopScreen()
		local node_class = screen and DebugUtil.FindRegisteredClass( screen )
		if node_class then
			debug_node = node_class( screen )
		else
			debug_node = DebugRoot( self.game )
		end
	end

	assert( is_instance( debug_node, DebugNode ))
	assert( debug_node.RenderPanel )
	local panel = DebugPanel( debug_node )
	table.insert( self.debug_panels, panel )

	return panel
end

function DebugManager:ClosePanel( panel )
	local idx = table.arrayfind( self.debug_panels, panel )
	assert( idx )

	table.remove( self.debug_panels, idx )
	panel:OnClose()

	table.insert( self.closed_debug_panels, panel )
	while #self.closed_debug_panels > MAX_DEBUG_PANEL_HISTORY do
		table.remove( self.closed_debug_panels, 1 )
	end
end

function DebugManager:RestorePanel( panel )
	local idx = table.arrayfind( self.closed_debug_panels, panel )
	assert( idx )

	table.remove( self.closed_debug_panels, idx )
	table.insert( self.debug_panels, panel )
end

function DebugManager:FindPanelByClass( class )
	for i, panel in ipairs( self.debug_panels ) do
		if is_instance( panel:GetNode(), class ) then
			return panel
		end
	end
end

function DebugManager:DebugPanelsOpen()
	return #self.debug_panels > 0 
end

function DebugManager:DebugPanels()
	return ipairs( self.debug_panels )
end

function DebugManager:ToggleDebugFlags( flags )
    self.debug_flags = self.debug_flags ~ flags 
	
	if self.game:FE() then
		local is_debugrender = self.debug_flags & (DBG_FLAGS.RENDER | DBG_FLAGS.UI) ~= 0
		self.game:FE():SetDebugRender( is_debugrender )

		if (flags & DBG_FLAGS.UI) ~= 0 then
	        self.game:FE():SetDebugMode( self.debug_flags & DBG_FLAGS.UI ~= 0 )
		end
	end
end

function DebugManager:IsDebugFlagged( flag )
    return (self.debug_flags or 0) & flag ~= 0
end

function DebugManager:RunBinding( binding, ... )
	local ok, result = false, "Illegal binding"
	if type(binding) == "function" then
		ok, result = xpcall( binding, generic_error, ... )
	elseif type(binding) == "table" then
		local fn = binding[1]
		ok, result = xpcall( fn, generic_error, table.unpack( binding, 2 ))
	end
	if not ok then
		engine.inst:ShowWarning( "DEBUG-ERR: "..tostring(result) )
	else
		return result
	end
end

function DebugManager:AddBindingGroup( bindings )
	assert( type(bindings) == "table" )
	assert( table.arrayfind( self.debug_bindings, bindings ) == nil)
	table.insert( self.debug_bindings, bindings )
end

function DebugManager:RemoveBindingGroup( bindings )
	table.arrayremove( self.debug_bindings, bindings )
end

function DebugManager:SetBindingEnabled( id, enabled )
	if enabled then
		table.arrayremove( self.disabled_bindings, id )
	else
		table.insert_unique( self.disabled_bindings, id )
	end
end

function DebugManager:ProcessDebugBinding( input, key, debug_menu, params )
    for i, debug_option in ipairs(debug_menu) do
        if type(debug_option.Menu) == "table" then
            if self:ProcessDebugBinding( input, key, debug_option.Menu, params ) then
                return true
            end
        else
            local ok = debug_option.Binding and debug_option.Binding.key == key and input:CheckModifiers( debug_option.Binding )
            ok = ok and (debug_option.Enabled == nil or debug_option.Enabled( table.unpack( params or table.empty )))
			ok = ok and (debug_option.ID == nil or not table.contains( self.disabled_bindings, debug_option.ID ))

            if not ok and debug_option.Bindings then
                for i, binding in ipairs(debug_option.Bindings) do
                    if binding and binding.key == key and input:CheckModifiers( binding ) then
                        ok = true
                        break
                    end
                end
            end
            if ok then
                local result = TheGame:GetDebug():RunBinding( debug_option.Do, table.unpack( params or table.empty ))
                return true
            end
        end
    end

    return false
end

function DebugManager:RawKeyDown( input, key )
	for j = #self.debug_panels, 1, -1 do
		local panel = self.debug_panels[j]
		if panel:RawKeyDown( input, key ) then
			return true
		end
	end

    for j = #self.debug_bindings, 1, -1 do
		if self:ProcessDebugBinding( input, key, self.debug_bindings[j] ) then
			return true
		end
	end
	return false
end

function DebugManager:OnControlDown( control, device_type )
	if control:Has( Controls.Digital.MENU_ALT ) and is_instance( TheGame:FE():GetTopScreen(), Screen.FightScreen ) then
		local dbg_env = self:GetDebugEnv()
		if not dbg_env.sim then
			return
		end
	end

	if control:Has( Controls.Digital.MENU_ALT ) and TheGame:GetInput():IsModifierControl() then
		local top_screen = TheGame:FE():GetTopScreen()
		if top_screen then
			local debug_class = DebugUtil.FindRegisteredClass( top_screen )
			if debug_class and debug_class.RenderContextMenu then
				dbg_env.show_context = DebugPanel( debug_class( top_screen ))
			end
		end
	end
end

return DebugManager
