-------------------------------------------------------------------
-- A debug node is simply an adapter that returns debug text for
-- a "source" object, which is literally whatever it is you want to
-- be debugging.
-- It also has a notion of child nodes, and a parent node, for easy
-- traversing of data hierarchies.

local DebugUtil = require "debug/debug_util"
local debug_menus = require( "debug/debug_menus" )
local DebugNode = class( "DebugNode" )
DebugNode.can_reload = true

 -- Debug nodes may have a registered class, which automatically creates this node
 -- if linked as a table and retrieved via GetTableLink().

DebugNode.REGISTERED_CLASS = nil

function DebugNode:GetName()
    return self._classname
end

function DebugNode:GetUID()
    return self._classname -- Unique identifier for ImGUI and equating recent nodes.
end

function DebugNode:GetDesc( sb )
    sb:Append( "NO NODE" )
end


--------------------------------------------------------------------
-- A debug root level node.

local DebugRoot = class( "DebugRoot", DebugNode )

DebugRoot.MENU_BINDINGS = { debug_menus.DEBUG_TOGGLES }

function DebugRoot:init( game )
    assert( game )
    self.game = game
    self.frame_times = {}
    self.frame_offset = 1
end

function DebugRoot:RenderFPS( ui, panel )
    local MAX_FRAMES = 60 -- SAMPLE_FRAME * MAX_FRAMES duration (roughly...)

    local frame_time = TheGame.times.frame
    if frame_time > 0 then
        self.frame_times[ self.frame_offset ] = frame_time * 1000
        self.frame_offset = (self.frame_offset % MAX_FRAMES) + 1
    end

    if #self.frame_times > 0 then
        local offset = 0
        if #self.frame_times >= MAX_FRAMES then
            offset = self.frame_offset - 1
        end
        local max = math.max( table.unpack( self.frame_times ))
        ui:PlotLines("Frame (ms)", string.format( "%.1f (max %.1f)", self.frame_times[#self.frame_times], max ), self.frame_times, offset, 0, 60, 50)
    end
end

function DebugRoot:RenderPanel( ui, panel )
    ui:Text( string.format( "%s\n\n", BUILD_ID_STRING ))

    self:RenderFPS( ui, panel, 300 )

    if panel:CollapsingHeader( ui, "SUBSYSTEMS" ) then
        if ui:TreeNode( "Game", ui.TreeFlags_DefaultOpen ) then
            ui:Indent( 20 )
            if ui:Selectable( "Game" ) then
                panel:PushDebugValue( self.game )
            end
            if ui:Selectable( "FE" ) then
                panel:PushDebugValue( self.game:FE() )
            end
            if self.game:GetGameState() then
                if ui:Selectable( "GameState" ) then
                    panel:PushDebugValue( self.game:GetGameState() )
                end
                if ui:Selectable( "Agents" ) then
                    DataUtil.Agents()
                end
            end
            ui:Unindent( 20 )
            ui:TreePop()
        end

        if ui:TreeNode( "Render", ui.TreeFlags_DefaultOpen ) then
            ui:Indent( 20 )
            if ui:Selectable( "Debug Render" ) then
                panel:PushDebugValue( DebugRendering() )
            end
            if ui:Selectable( "Colour Settings" ) then
                panel:PushDebugValue( DebugColourSettings() )
            end
            if ui:Selectable( "Textures" ) then
                panel:PushDebugValue( DebugTextures() )
            end
            ui:Unindent( 20 )
            ui:TreePop()
        end

        if ui:TreeNode( "Content", ui.TreeFlags_DefaultOpen ) then
            ui:Indent( 20 )

            if ui:Selectable( "Content.internal" ) then
                panel:PushDebugValue( Content.internal )
            end
            if ui:Selectable( "Build Counts" ) then
                panel:PushDebugValue( DebugCharacterBuilds() )
            end
            if ui:Selectable( "Battle Cards" ) then
                local battle_defs = require "battle/battle_defs"
                panel:PushDebugValue( DebugCardCollection( battle_defs ) )
            end
            if ui:Selectable( "Negotiation Cards" ) then
                local negotiation_defs = require "negotiation/negotiation_defs"
                panel:PushDebugValue( DebugCardCollection( negotiation_defs ) )
            end
            if ui:Selectable( "Character Skins" ) then
                panel:PushDebugValue( DebugCharacterSkins() )
            end
            if ui:Selectable( "Quests" ) then
                panel:PushDebugValue( Content.GetAllQuests() )
            end

            ui:Unindent( 20 )
            ui:TreePop()
        end

        if ui:TreeNode( "Misc", ui.TreeFlags_DefaultOpen ) then
            ui:Indent( 20 )
            if ui:Selectable( "Fight Tester" ) then
                panel:PushDebugValue( DebugFightTest() )
            end
            if ui:Selectable( "Audio" ) then
                panel:PushDebugValue( DebugAudio() )
            end

            if ui:Selectable( "Tuning" ) then
                panel:PushDebugValue( DebugTuning() )
            end

            if ui:Selectable( "Easing" ) then
                panel:PushDebugValue( DebugEasing() )
            end

            ui:Unindent( 20 )
            ui:TreePop()
        end
    end
end

--------------------------------------------------------------------
-- Whenever attempting to create a panel for the value 'nil'

local DebugNil = class( "DebugNil", DebugNode )

DebugNil.PANEL_WIDTH = 200
DebugNil.PANEL_HEIGHT = 80

function DebugNil:RenderPanel( ui, panel )
    ui:TextColored( 0xCC0000FF, "nil" )
end
 
 --------------------------------------------------------------------
-- A debug source for a generic table

local DebugTable = class( "DebugTable", DebugNode )

function DebugTable:init( t, name, offset )
    self.t = t
    self.offset = offset
    self.name = name
    self.menu_params = { t }
end

function DebugTable:GetSubject()
    return self.t
end

function DebugTable:GetName()
    return self.name or rawstring(self.t)
end

function DebugTable:GetUID()
    return rawstring(self.t)
end

function DebugTable:RenderPanel( ui, panel )
 
    if self.t.RenderDebugUI then
        self.t:RenderDebugUI( ui, panel )
        ui:Separator()
    end

    ui:Text( string.format( "%d fields", table.count( self.t ) ))
    ui:Separator()
    
    ui:Columns(2, "mycolumns3", false)
    local mt = getmetatable(self.t)
    if mt then
        ui:Text( "getmetatable()" )
        ui:NextColumn()

        if ui:Selectable( mt._class and mt._classname or rawstring(mt) ) then
            panel:PushNode( DebugUtil.CreateDebugNode( mt ))
        end
        ui:NextColumn()
    end
    ui:Columns(1)

    panel:AppendKeyValues( ui, self.t, self.offset )
end

--------------------------------------------------------------------
-- Just show a plain old value.

local DebugValue = class( "DebugValue", DebugNode )

function DebugValue:init( value )
    self.value = value
end

function DebugValue:RenderPanel( ui, panel )
    ui:TextColored( 0x55CCCCFF, type(self.value))
    ui:Text( tostring(self.value) )
end


--------------------------------------------------------------------
-- Dynamic, custom inspector

local DebugCustom = class( "DebugCustom", DebugNode )

function DebugCustom:init( fn )
    self.fn = fn
end

function DebugCustom:RenderPanel( ui, panel )
    self:fn( ui, panel )
end
 