local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

-------------------------------------------------
-- Renders a debug panel with immediate mode GUI.

local DebugPanel = class( "DebugPanel" )
local next_uid = 1

function DebugPanel:init( node )
    self.nodes = {} -- stack of debug nodes
    self.idx = 0 -- Current index into self.nodes
    self.uid = next_uid
    next_uid = next_uid + 1
    self.dbg = TheGame:GetDebug()
    self.init_traceback = debug.traceback()

    if node then
        self:PushNode( node )
    end
end

function DebugPanel:Close()
    self._wants_to_close = true
end

function DebugPanel:GetNode()
    return self.nodes[ self.idx ]
end

function DebugPanel:OnClose()
    local node = self:GetNode()
    if node then
        if node.OnDeactivate then
            node:OnDeactivate( self )
        end
    end
end

function DebugPanel:RenderPanel( ui )

    self.frame_uid = 0 -- For ensuring ID uniqueness of different widgets with the same values.
    local node = self:GetNode()
    if node ~= self.last_node then
        if self.last_node and self.last_node.OnDeactivate then
            self.last_node:OnDeactivate()
        end

        self.last_node = node

        if node and node.OnActivate then
            node:OnActivate( self )
        end
    end

    local show
    if node.BeginRenderPanel then
        show = node:BeginRenderPanel( ui, self )
    end

    if show == nil then
        local flags = ui.MenuBar
        local title = string.format( "%s (%d/%d)###%d", node:GetName(), self.idx, #self.nodes, self.uid )

        if self.is_maximized then
            flags = flags | ui.NoTitleBar | ui.NoResize | ui.NoMove
            local w, h = ui:DisplaySize()
            ui:SetNextWindowSize( w, h, ui.ImGuiSetCond_Always )
            ui:SetNextWindowPos( 0, 0, ui.ImGuiSetCond_Always )

        elseif node.PANEL_AUTOSIZE then
            flags = flags | ui.AlwaysAutoResize

        elseif self.set_width or self.set_height then
            ui:SetNextWindowSize( self.set_width or -1, self.set_height or -1, ui.ImGuiSetCond_Always )
            self.set_width, self.set_height = nil, nil
        end 

        show = ui:Begin( title, flags )
    end

    local did_maximize = false   
    if show then
        if ui:BeginMenuBar() then
            if node.can_reload and node._file then
                if ui:Button( "*" ) then
                    local filename = node._file:match( "^scripts/(.+)[.]lua$" )
                    if filename then
                        package.loaded[ filename ] = nil
                        local ok, err = pcall( require, filename )
                        if not ok then
                            LOGWARN( "WARNING: Failed to load debug inspector '%s'\n%s", filename, err )
                        end
                    end
                end
                if ui:IsItemHovered() then
                    ui:SetTooltip( string.format( "Reload %s", node._file ))
                end
                ui:SameLine( nil, 5 )
            end

            if node.GetSubject then
                local subject = node:GetSubject()
                if subject and subject.can_reload and subject._file then
                    local filename = subject._file:match( "^scripts/(.+)[.]lua$" )
                    if filename and ui:Button("**") then
                        package.loaded[ filename ] = nil
                        local ok, err = pcall( require, filename )
                        if not ok then
                            LOGWARN( "WARNING: Failed to load debug inspector '%s'\n%s", filename, err )
                        end
                    end
                end
                if ui:IsItemHovered() then
                    ui:SetTooltip( string.format( "Reload %s", subject._file ))
                end
                ui:SameLine( nil, 5 )
            end

            if ui:Button( " < ", nil, nil, self.idx <= 1 ) then
                self:GoBack()
            end
            ui:SameLine( nil, 5 )
            if ui:Button( " > ", nil, nil, self.idx >= #self.nodes ) then
                self:GoForward()
            end
            ui:SameLine( nil, 20 )

            if ui:BeginMenu( "Menu" ) then
                if ui:MenuItem( "Maximize", nil, self.is_maximized ) then
                    did_maximize = true
                end

                ui:Separator()

                if ui:MenuItem( "Back", nil, nil, self.idx > 1 ) then
                    self:GoBack()
                end
                if ui:MenuItem( "Forward", "", nil, self.idx < #self.nodes ) then
                    self:GoForward()
                end

                if ui:MenuItem( "View Root" ) then
                    self:PushNode( DebugRoot( TheGame ) )
                end

                local closed_panels = TheGame:GetDebug().closed_debug_panels
                if ui:BeginMenu( "View Recently Closed", #closed_panels > 0 ) then
                    for i = #closed_panels, 1, -1 do
                        local panel = closed_panels[i]
                        if ui:MenuItem( panel:GetNode():GetName() ) then
                            TheGame:GetDebug():RestorePanel( panel )
                            break
                        end
                    end
                    ui:EndMenu()
                end

                if ui:MenuItem( "View This Panel" ) then
                    self:PushNode( DebugTable( self ) )
                end

                ui:Separator()
                if ui:MenuItem( "Close" ) then
                    show = false
                end
                ui:EndMenu()
            end

            if node.MENU_BINDINGS then
                for i, menu in ipairs( node.MENU_BINDINGS ) do
                    if ui:BeginMenu( menu.name or "???" ) then
                        self:AddDebugMenu( ui, menu, node.menu_params )
                        ui:EndMenu()
                    end
                end
            end

            if node.menu_params and ui:BeginMenu( debug_menus.TABLE_BINDINGS.name ) then
                self:AddDebugMenu( ui, debug_menus.TABLE_BINDINGS, node.menu_params )
                ui:EndMenu()
            end

            for i, bindings in ipairs(TheGame:GetDebug().debug_bindings) do
                if ui:BeginMenu( bindings.name or "???" ) then
                    self:AddDebugMenu( ui, bindings, node.menu_params )
                    ui:EndMenu()
                end
            end

            if ui:BeginMenu( "Help" ) then
                if ui:MenuItem( "Demo" ) then
                    self.show_test_window = true
                end
                ui:EndMenu()
            end

            ui:EndMenuBar()
        end

        if node.RenderPanel then
            node:RenderPanel( ui, self )
        else
            ui:TextColored( 0xFFFF00FF, "NOT YET IMPLEMENTED" )
        end

        if did_maximize then
            self.is_maximized = not self.is_maximized
            if self.is_maximized then
                -- Keep track of original size
                self.set_width, self.set_height = ui:GetWindowSize()
            end
        end
    end
    
    ui:End()

    if self.show_test_window then
        self.show_test_window = ui:ShowTestWindow()
    end

    return show
end

function DebugPanel:RenderContextMenu( ui )
    self.frame_uid = 0 -- For ensuring ID uniqueness of different widgets with the same values.
    local node = self:GetNode()
    if node.RenderContextMenu then
        node:RenderContextMenu( ui, self )
    end
end

function DebugPanel:RawKeyDown( input, key )
    local node = self:GetNode()
    if node.MENU_BINDINGS then
        for j = #node.MENU_BINDINGS, 1, -1 do
            if self.dbg:ProcessDebugBinding( input, key, node.MENU_BINDINGS[j], node.menu_params ) then
                return true
            end
        end
    end
    return false
end

function DebugPanel:PushNode( node )
    if not is_instance( node, DebugNode ) then
        node = DebugUtil.CreateDebugNode( node )
    end

    -- Clear forward history.
    while #self.nodes > self.idx do
        table.remove( self.nodes )
    end
    table.insert( self.nodes, node )
    self.idx = #self.nodes

    self.set_width, self.set_height = node.PANEL_WIDTH, node.PANEL_HEIGHT
end

function DebugPanel:PushDebugValue( v )
    local node
    if is_instance( v, DebugNode ) then
        node = v
    else
        node = DebugUtil.CreateDebugNode( v )
    end
    if TheGame:GetInput():IsModifierShift() then
        TheGame:GetDebug():CreatePanel( node )
    else
        self:PushNode( node )
    end
end

function DebugPanel:GoBack()
    self.idx = self.idx - 1
    assert( self.idx >= 1 )
end

function DebugPanel:GoForward()
    self.idx = self.idx + 1
    assert( self.idx <= #self.nodes )
end

----------------------------------------------------------------------
-- Helper wrappers for composite imgui rendering.

function DebugPanel:CollapsingHeader( ui, title, flags )
    return ui:CollapsingHeader( title, flags or ui.TreeFlags_DefaultOpen )
end

function DebugPanel:SetLockedValues( locked_values )
    self.locked_values = locked_values
end

function DebugPanel:AddDebugMenu( ui, menu, menu_params )
    menu_params = menu_params or table.empty
    for i, option in ipairs( menu ) do
        if option.Visible == false or (type(option.Visible) == "function" and not option.Visible( table.unpack( menu_params ))) then
            -- Invalid for this thingy.
        else
            local txt
            if type(option.Text) == "string" then
                txt = option.Text
            elseif type(option.Text) == "function" then
                txt = option.Text( table.unpack( menu_params ) )
            end
            if option.Binding then
                txt = (txt or "") .. string.format( " (%s)", TheGame:GetInput():GetBindingString( option.Binding ))
            end

            if txt == nil then
                ui:Separator()
            else
                local checked = type(option.Checked) == "function" and option.Checked( table.unpack( menu_params ))
                local enabled, tt = true
                if type(option.Enabled) == "function" then
                    enabled, tt = option.Enabled( table.unpack( menu_params ))
                end
                if option.Menu then
                    local menu = type(option.Menu) == "function" and option.Menu( table.unpack( menu_params) ) or option.Menu
                    if ui:BeginMenu( txt, enabled ) then
                        self:AddDebugMenu( ui, menu, menu_params )
                        ui:EndMenu()
                    end

                elseif option.CustomMenu then
                    if ui:BeginMenu( txt, enabled ) then
                        option:CustomMenu( ui, table.unpack( menu_params ))
                        ui:EndMenu()
                    end

                elseif ui:MenuItem( txt, nil, checked, enabled ) then
                    self.dbg:RunBinding( option.Do, table.unpack(menu_params) )
                end
                if tt and ui:IsItemHovered() then
                    ui:SetTooltip( tt )
                end
            end
        end
    end
end

function DebugPanel:AppendTable( ui, v, name, clr )
    -- Add a Button that allows pushing a new debug node for the table.
    if v == table.empty then
        ui:PushStyleColor( ui.Style_Button, 0x000000AA )
    else
        ui:PushStyleColor( ui.Style_Button, clr or 0x0055CC55 )
    end
    ui:PushID( self.frame_uid )
    if ui:Button( tostring(name or v ) or "??" ) and v then
        self:PushDebugValue( v )
    end

    if ui:BeginPopupContextItem( "cxt" ) then
        ui:TextColored( 0x00FFFFFF, name or tostring(v) )

        ui:Separator()
        self:AddDebugMenu( ui, debug_menus.TABLE_BINDINGS, { v } )

        local debug_class = DebugUtil.FindRegisteredClass( v )
        if debug_class and debug_class.MENU_BINDINGS then
            for i, menu in ipairs( debug_class.MENU_BINDINGS ) do
                ui:Separator()
                self:AddDebugMenu( ui, menu, { v } )
            end
        end
        ui:EndPopup()
    end
    ui:PopID()
    ui:PopStyleColor()
    self.frame_uid = self.frame_uid + 1
end

function DebugPanel:AppendTableInline( ui, t, name )
    ui:PushID( self.frame_uid )
    if ui:TreeNode( name or tostring(t) ) then
        for k, v in pairs(t) do
            if type(v) == "table" then
                self:AppendTableInline( ui, v, tostring(k) )
            else
                -- Key
                ui:Text( tostring(k)..":" )
                ui:SameLine( nil, 10 )
                -- Value
                if type(v) == "string" then
                    ui:TextColored( 0x7777FFFF, v )

                elseif type(v) == "function" then
                    ui:TextColored( 0xFF7755FF, tostring(v) )

                elseif type(v) == "userdata" then
                    ui:TextColored( 0x55CCBBFF, tostring(v) )

                else
                    ui:Text( tostring(v) )
                end
            end
        end
        ui:TreePop()
    end
    ui:PopID()
    self.frame_uid = self.frame_uid + 1
end

function DebugPanel:AppendValue( ui, v, owning_table, owning_table_key )
    if type(v) == "table" then
        self:AppendTable( ui, v )

    elseif type(v) == "string" then
        ui:TextColored( 0x7777FFFF, v )

    elseif type(v) == "function" then
        ui:TextColored( 0xFF7755FF, tostring(v) )

    elseif type(v) == "userdata" then
        ui:TextColored( 0x55CCBBFF, tostring(v) )
    
    elseif type(v) == "number" and owning_table and owning_table_key and not self.locked_values and v < 2^31 then
        local new_v = ui:DragFloat("##"..tostring(owning_table)..tostring(owning_table_key), v )
        if new_v ~= v then
            owning_table[owning_table_key] = new_v
        end
    else
        ui:Text( tostring(v) )
    end
end


function DebugPanel:AppendKeyValue( ui, key, v, t )
    -- Key
    self:AppendValue( ui, key )
    ui:NextColumn()

    -- Value
    self:AppendValue( ui, v, t, key )
    ui:NextColumn()
end

function DebugPanel:AppendKeyValues( ui, t, offset )
    local MAX_ENTRIES = 128
    offset = offset or 0 

    ui:Columns( 2, "keyvalues", false )

    local n = table.count( t )
    local more
    for i, k, v in sorted_pairs( t ) do
        if i >= offset then
            self:AppendKeyValue( ui, k, v, t )

            if i >= offset + MAX_ENTRIES then
                more = i
                break
            end
        end
    end

    ui:Columns(1)

    if more then
        ui:Separator()
        ui:TextColored( 0xFF00FFFF, string.format( "%d more entries", n - more ))
        ui:SameLine( nil, 10 )
        
        if offset > 0 then
            ui:SameLine( nil, 10 )
            if ui:Button( "Prev..." ) then
                self:PushNode( DebugTable( t, nil, math.max( 0, offset - MAX_ENTRIES )))
            end
        end

        ui:SameLine( nil, 10 )
        if ui:Button( "Next..." ) then
            self:PushNode( DebugTable( t, nil, more ))
        end
    end
end


function DebugPanel:AppendTabularData( ui, column_headers, data)
    ui:Columns( #column_headers, "tabular", true )
    for k, v in ipairs(column_headers) do
        ui:TextColored(0x55CCBBFF, tostring(v))
        ui:NextColumn()
    end
    ui:Separator()
    for k, v in ipairs(data) do
        for c = 1, #column_headers do
            if c <= #v then
                if type(v[c]) == "table" and v[c].name and v[c].table then
                    self:AppendTable(ui, v[c].table, v[c].name)
                else
                    ui:Text(tostring(v[c]))
                end
            end
            ui:NextColumn()
        end
    end
    ui:Separator()
    ui:Columns(1)
end


function DebugPanel:AppendBitField( ui, flags, bit_field, bit_field_strings )
    ui:PushID( self.frame_uid )
    flags = DebugUtil.RenderBitField( ui, flags, bit_field, bit_field_strings )
    ui:PopID()
    self.frame_uid = self.frame_uid + 1

    return flags
end