local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

local function HasFX( t )
	return t.fx ~= nil
end

local function HasSound( t )
	return t.sound ~= nil
end

--------------------------------------------------------------------
-- Shows missing FX/sound markup for all character def attack anims.

local DebugAttackAnims = class( "DebugAttackAnims", DebugNode )
DebugAttackAnims.PANEL_WIDTH = 800
DebugAttackAnims.PANEL_HEIGHT = 500

DebugAttackAnims.can_reload = true

function DebugAttackAnims:init()
	self.defs = {}
	for content_id, def in pairs(Content.GetAllCharacterDefs()) do
		if not def.is_template then
			table.insert( self.defs, { content_id = content_id, def = def, agent = Agent( content_id ) })
		end
	end
	table.sort( self.defs, function( a, b ) return a.content_id < b.content_id end )

	self.attacks = {}
	for i, def in ipairs( Content.GetAllBattleCards() ) do
		if def.anim and (def.min_damage or def.generate_damage_fn) then
			self.attacks[ def.anim ] = true
		end
	end
end

function DebugAttackAnims:GetMissingStuff( anim, key, name )
	local missing = ObtainWorkTable()
    local meta, meta_filename = GetAnimMeta( key, name )

    if not self.filter_sound then
		if self.attacks[ name ] then
	    	if meta == nil or meta.hit_frames == nil or #meta.hit_frames == 0 then
	    		table.insert( missing, "HITS" )
	    	end
	    end

		if meta == nil or meta.fx_tracks == nil or not table.findif( meta.fx_tracks, HasFX ) then
	    	table.insert( missing, "FX" )
	    end
	end
    
    local has_sound = false
    if meta then
		if meta.fx_tracks and table.findif( meta.fx_tracks, HasSound ) then
			has_sound = true
		end
		if meta.sound_frames and table.findif( meta.sound_frames, HasSound ) then
			has_sound = true
		end
    end

	if not has_sound then
		table.insert( missing, "SOUND" )
	end
    local missing_str = table.concat( missing, " " )
    ReleaseWorkTable( missing )
    return missing_str, meta_filename
end

function DebugAttackAnims:ShowAnims( ui, panel, content_id, anims, req_source )

	if req_source then
	    for i, anim_path in pairs( anims or table.empty ) do
		    local anim = GetAnim( anim_path )
		    local key = GetAnimMetaFilename( anim )
			local anim_meta = GetAnimMeta( key )
			if anim_meta._parent == nil then
				ui:Text( content_id )
				ui:NextColumn()

				ui:NextColumn()

				if ui:Button( string.format( "%s##META", key )) then
					DBG( anim_meta )
				end
				ui:NextColumn()

	        	ui:TextColored( 0xFF0000FF, "UNSOURCED" )
	        	ui:NextColumn()
			end
		end
	end

    for i, anim_path in pairs( anims or table.empty ) do
	    local anim = GetAnim( anim_path )
	    local key = GetAnimMetaFilename( anim )
	    for k = 1, anim:GetNumAnims() do
	        local name = anim:GetAnimName(k-1)
	        -- Only about attacks?
    		if self.attacks[ name ] or name == "defend" or name == "death" then
    			local missing, meta_filename = self:GetMissingStuff( anim, key, name )
		        if #missing > 0 then
	    			ui:Text( content_id )
	    			ui:NextColumn()

			        ui:Text( name )
			        if ui:IsItemHovered() then
						local tt = loc.format( "{1} ({2} anims)", anim_path, math.floor( anim:GetNumAnims()))
						ui:SetTooltip( tt )
			        end
			        ui:NextColumn()

					if meta_filename and ui:Button( string.format( "%s##%s", meta_filename, anim )) then
					    local meta, meta_filename = GetAnimMeta( key, name )
						DBG( meta )
					end
		        	ui:NextColumn()

		        	ui:TextColored( 0xFF0000FF, missing )
		        	ui:NextColumn()
		        end
	        end
	    end
	end
end

function DebugAttackAnims:RenderPanel( ui, panel )

	if ui:Checkbox( "Sound Only", self.filter_sound ) then
		self.filter_sound = not self.filter_sound
	end
	ui:Spacing()

	ui:Columns(4)
	ui:Text( "CharacterDef" )
	ui:NextColumn()
	ui:Text( "Anim" )
	ui:NextColumn()
	ui:Text( "Meta File" )
	ui:NextColumn()
	ui:Text( "Missing" )
	ui:NextColumn()
	ui:Separator()


	for i, t in ipairs( self.defs ) do
		local def, agent = t.def, t.agent
	    if not def.is_template then
    		local build = agent:GetBuild()
    		if build and build.combatanims then
    			self:ShowAnims( ui, panel, t.content_id, build.combatanims )
    		end
    		-- self:ShowAnims( ui, panel, t.content_id, def.anims )

    		self:ShowAnims( ui, panel, t.content_id, def.combat_anims, true )
	    end
	end

	ui:Columns(1)
end
