local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------

local COLS =
{
    ID = { "ID", "id",
        function( self, carddef )
            return carddef.id
        end
    },

    NAME = { "Name", "name",
        function( self, carddef )
            return carddef.name
        end
    },

    DESC = { "Desc", "desc",
        function( self, carddef )
            return tostring(carddef.desc)
        end
    },

    FLAGS = { "Flags",
        function( self, ui, panel, carddef )
            ui:Text( StringizeBitField( carddef.flags, panel:GetNode().defs.CARD_FLAG_BITSTRINGS ))
        end,
        function( self, carddef )
            return carddef.flags
        end
    },

    COST = { "Cost", "cost",
        function( self, carddef )
            return carddef.cost
        end
    },

    CHARGES = { "Charges",
        function( self, ui, panel, carddef )
            if carddef.max_charges then
                ui:TextColored( 0x77FFFFFF, tostring(carddef.max_charges) )
            end
        end,
        function( self, carddef )
            return carddef.max_charges
        end,
    },

    RARITY = { "Rarity", "rarity",
        function( self, carddef )
            return CARD_RARITY_ORDINALS[ carddef.rarity ] or -1
        end,
    },

    DAMAGE = { "Damage",
        function( self, ui, panel, carddef )
            if carddef.min_damage and carddef.max_damage then
                ui:Text( string.format( "%d - %d", carddef.min_damage, carddef.max_damage ))
            end
        end,
        function( self, carddef )
            return carddef.min_damage or 0
        end,
    },

    ART = { "Art",
        function( self, ui, panel, carddef )
            ui:Image( carddef.icon, 64, 42 )
            if ui:IsItemHovered() then
                ui:SetTooltip( tostring(carddef.path_str) .. "\n" ..tostring(carddef.desc) )
            end
        end,
        function( self, carddef )
            return tostring(carddef.path_str)
        end,
    },
}

local DebugCardCollection = class( "DebugCardCollection", DebugNode )

DebugCardCollection.MENU_BINDINGS =
{
    {
        name = "Cards",
        {
            Text = "Filter tags",
            CustomMenu = function( option, ui, self )
                for k, data in ipairs( GetAllPlayerBackgrounds() ) do
                    local id = data.id
                    local checked = table.contains( self.settings.show_tags, id )
                    if ui:Selectable( id, checked, ui.DontClosePopups ) then
                        if checked then
                            table.arrayremove( self.settings.show_tags, id )
                        else
                            table.insert( self.settings.show_tags, id )
                        end
                        self:SaveSettings()
                    end
                end
            end
        },
        {
            Text = "Filter CARD_FLAGS",
            CustomMenu = function ( option, ui, self )
                for i, flag in pairs( self.defs.CARD_FLAG_BITSTRINGS ) do
                    local bit = 1 << (i - 1 )
                    local txt, clr
                    if self.settings.filter_flags[ bit ] == "-" then
                        txt = string.format( "-%s", flag )
                    elseif self.settings.filter_flags[ bit ] == "+" then
                        txt = string.format( "+%s", flag )
                    else
                        clr = 0x777777FF
                        txt = flag
                    end
                    if clr then
                        ui:PushStyleColor( ui.Style_Text, clr )
                    end

                    if ui:Selectable( txt, nil, ui.DontClosePopups ) then
                        if self.settings.filter_flags[ bit ] == "-" then
                            self.settings.filter_flags[ bit ] = nil
                        elseif self.settings.filter_flags[ bit ] == "+" then
                            self.settings.filter_flags[ bit ] = "-"
                        else
                            self.settings.filter_flags[ bit ] = "+"
                        end
                        self:SaveSettings()
                    end

                    if clr then
                        ui:PopStyleColor()
                    end
                end
            end
        },
        {
            Text = "Filter Rarity",
            CustomMenu = function( option, ui, self )
                for i, _, rarity in sorted_pairs( CARD_RARITY ) do
                    if ui:MenuItem( tostring(rarity), nil, self.settings.show_rarity == rarity ) then
                        self.settings.show_rarity = rarity
                        self:SaveSettings()
                    end
                end
            end,
        }
    }
}


function DebugCardCollection:init( defs )
    assert( defs )
    self.defs = defs
    self.menu_params = { self }

    self:LoadSettings()

    self:RefreshCards()
end

function DebugCardCollection:RefreshCards()
    local battle_defs = require "battle/battle_defs"
    local negotiation_defs = require "negotiation/negotiation_defs"    

    self.cards = {}

    local all_defs
    if battle_defs == self.defs then
        all_defs = Content.GetAllBattleCards()
    else
        all_defs = Content.GetAllNegotiationCards()
    end

    for i, carddef in ipairs( all_defs ) do
        table.insert( self.cards, carddef )
    end
end

function DebugCardCollection:LoadSettings()
    local battle_defs = require "battle/battle_defs"
    local negotiation_defs = require "negotiation/negotiation_defs"    

    if self.defs == negotiation_defs then
        self.settings = DebugUtil.GetLocalDebugData("DEBUG_LIST_NEGOTIATION_CARDS")
        if self.settings == nil then
            self.settings = {}
        end
        if self.settings.columns == nil then
            self.settings.columns = { "ART", "COST", "CHARGES", "RARITY" }
        end
        if self.settings.filter_flags == nil then
            self.settings.filter_flags = 
            {
                [ self.defs.CARD_FLAGS.DIPLOMACY ] = "+",
                [ self.defs.CARD_FLAGS.MANIPULATE ] = "+",
                [ self.defs.CARD_FLAGS.HOSTILE ] = "+",
            }
        end
        table.arrayremove( self.settings.columns, "ID" )

    else
        self.settings = DebugUtil.GetLocalDebugData("DEBUG_LIST_BATTLE_CARDS")
        if self.settings == nil then
            self.settings = {}
        end
        if self.settings.columns == nil then
            self.settings.columns = { "ART", "COST", "FLAGS", "DAMAGE", "RARITY" }
        end
        if self.settings.filter_flags == nil then
            self.settings.filter_flags = 
            {
                [ self.defs.CARD_FLAGS.MELEE ] = "+",
                [ self.defs.CARD_FLAGS.RANGED ] = "+",
                [ self.defs.CARD_FLAGS.SKILL ] = "+",
                [ self.defs.CARD_FLAGS.ITEM ] = "+",
            }
        end
        table.arrayremove( self.settings.columns, "ID" )
    end

    if self.settings.show_tags == nil then
        self.settings.show_tags = {}
    end
end

function DebugCardCollection:SaveSettings()
    local battle_defs = require "battle/battle_defs"
    local negotiation_defs = require "negotiation/negotiation_defs"    

    if self.defs == negotiation_defs then
        DebugUtil.SetLocalDebugData("DEBUG_LIST_NEGOTIATION_CARDS", self.settings)
    else
        DebugUtil.SetLocalDebugData("DEBUG_LIST_BATTLE_CARDS", self.settings)
    end
    DebugUtil.SaveLocalDebugData()
end


function DebugCardCollection:SortByColumn( column_id )
    if self.sort_column == column_id then
        if self.sort_ascending then
            self.sort_ascending = false
        else
            self.sort_ascending = true
        end
    else
        self.sort_ascending = true
        self.sort_column = column_id
    end

    local column = COLS[ column_id ]
    local fn = column[3]

    if self.sort_ascending then
        local function SortAscending( a, b )
            return fn( column, a ) < fn( column, b )
        end
        table.sort( self.cards, SortAscending )
    else
        local function SortDescending( a, b )
            return fn( column, a ) > fn( column, b )
        end
        table.sort( self.cards, SortDescending )
    end
end

function DebugCardCollection:IsCardShown( carddef )
    local CARD_FLAGS = self.defs.CARD_FLAGS

    if self.settings.show_rarity and carddef.rarity ~= self.settings.show_rarity then
        return false
    end

    if #self.settings.show_tags > 0 then
        if not table.contains( self.settings.show_tags, carddef.series ) then
            return false
        end
    end

    local ok = false
    for flag, v in pairs( self.settings.filter_flags ) do
        if v == "-" and CheckBits( carddef.flags, flag ) then
            return false
        elseif v == "+" and CheckBits( carddef.flags, flag ) then
            ok = true
        end
    end
    if not ok then
        return false
    end

    return true
end

function DebugCardCollection:RenderPanel( ui, panel )

    if self.last_count then
        ui:TextColored( 0x55CCCCFF, string.format( "%d cards listed", self.last_count ))
    end
    ui:Separator()

    ui:Columns( #self.settings.columns + 1 )
    for i = 0, #self.settings.columns do
        local column_id
        if i == 0 then
            column_id = "ID"
        else
            column_id = self.settings.columns[i]
        end
        if column_id == nil then
            break
        end
        local column = COLS[ column_id ]
        local header, fn = column[1], column[3]
        if ui:Selectable( header ) and fn then
            self:SortByColumn( column_id )
        end
        ui:NextColumn()

        if ui:BeginPopupContextItem( header ) then
            for j, column_id, column in sorted_pairs( COLS ) do
                if column_id ~= "ID" then
                    local is_shown = table.arrayfind( self.settings.columns, column_id ) ~= nil
                    if ui:Checkbox( column[1], is_shown ) then
                        if is_shown then
                            table.arrayremove( self.settings.columns, column_id )
                        else
                            table.insert( self.settings.columns, i + 1, column_id )
                        end
                        self:SaveSettings()
                    end
                end
            end
            ui:EndPopup()
        end
    end

    ui:Separator()

    local count = 0
    for i, carddef in ipairs( self.cards ) do
        if self:IsCardShown( carddef ) then
            if ui:Selectable( carddef.id, nil, ui.Selectable_SpanAllColumns ) then
                DBG( carddef )
            end
            ui:NextColumn()

            for j = 1, #self.settings.columns do
                local column_id = self.settings.columns[j]
                local column = COLS[ column_id ]
                local key = column[2]

                if type(key) == "string" then
                    ui:Text( tostring(carddef[ key ]) )
                elseif type(key) == "function" then
                    key( column, ui, panel, carddef )
                end
                ui:NextColumn()
            end
            count = count + 1
        end
    end

    ui:Columns(1)

    ui:Separator()
    ui:TextColored( 0x55CCCCFF, string.format( "%d cards listed", count ))
    self.last_count = count

end
