local DebugUtil = require "debug/debug_util"
local chargen = require "charactergen"
require "ui/screens/characteratlasscreen"

-------------------------------------------------------------------
-- Cached array lists for Imgui arguments.

local GENDER_COMBO = copykeys( GENDER )
table.sort( GENDER_COMBO)

local SPECIES_COMBO = copykeys( SPECIES )
table.sort( SPECIES_COMBO )

local ATLAS_BG_COLOURS = {
	0x000000FF, 0xFFFFFFFF, 0x2C0E2DFF, 0x2D240BFF, 0x03011DFF, 0x640700FF
}


---------------------------------------------------------------------
-- Column definitions.

local function ShowStatCount( key )
	return function(t)
		local count = t.counts[ key ] or 0
		return count, count > 0 and tostring(count) or ""
	end
end

local COLUMNS_BY_SPECIES =
{
    { name="ID", fn = function(t) return t.content_id, t.content_id end },
    { name="Human", fn = ShowStatCount( SPECIES.HUMAN ) },
    { name="Jarackle", fn = ShowStatCount( SPECIES.JARACKLE ) },
    { name="Kradeshi", fn = ShowStatCount( SPECIES.KRADESHI ) },
    { name="Shroke", fn = ShowStatCount( SPECIES.SHROKE) },
    { name="Phicket", fn = ShowStatCount( SPECIES.PHICKET ) },
    { name="Total", fn = function(t) return t.total_count, t.total_count end }
}


local COLUMNS_BY_GENDER =
{
    { name="ID", fn = function(t) return t.content_id, t.content_id end },
    { name="Male", fn = ShowStatCount( GENDER.MALE ) },
    { name="Female", fn = ShowStatCount( GENDER.FEMALE ) },
    { name="Undisclosed", fn = ShowStatCount( GENDER.UNDISCLOSED ) },
    { name="Total", fn = function(t) return t.total_count, t.total_count end }
}

local COLUMNS_UNUSED_HEADS =
{
	{ name="ID", fn = function(t) return t.head_id, t.head_id end },
	{ name="Gender", key = "gender" },
	{ name="Species", key = "species" },
	{ name="Faction", key = "faction_id" },
	{ name="Priority", key = "priority" },
	{ name="Match Fn", key = "match_fn" },
}

local COLUMNS =
{
	EDIT = { name = "Edit", width = 30,
		ui = function( column, self, ui, panel, skin )
			local is_expanded = table.contains( self.expanded_skins, skin.uuid )
			if ui:Checkbox( "##EDIT", is_expanded ) then
				if is_expanded then
					table.arrayremove( self.expanded_skins, skin.uuid )
				else
					table.insert( self.expanded_skins, skin.uuid )
				end
			end
		end,
		value = function( column, self, skin )
			return table.contains( self.expanded_skins, skin.uuid )
		end,
	},
	NAME = { name = "Name", width = 100, property = "name",
		ui = function( column, self, ui, panel, skin )
			ui:PushItemWidth( ui:GetColumnWidth() - 3 )
			local name = ui:InputText( "##NAME", skin.name, ui.TextFlags_EnterReturnsTrue )
			ui:PopItemWidth()
			if name and name ~= skin.name then
				skin.name = name
				return true
			end
		end,
		value = function( column, self, skin )
			return skin.name
		end
	},
	NICKNAME = { name = "Nickname", width = 100, property = "nickname",
		ui = function( column, self, ui, panel, skin )
			ui:PushItemWidth( ui:GetColumnWidth() - 3 )
			local nickname = ui:InputText( "##NICKNAME", skin.nickname, ui.TextFlags_EnterReturnsTrue )
			ui:PopItemWidth()
			if nickname and nickname ~= skin.nickname then
				skin.nickname = nickname
				return true
			end
		end,
		value = function( column, self, skin )
			return skin.nickname
		end
	},
	DEF = { name = "Def", width = 90,
		ui = function( column, self, ui, panel, skin )
			local content_id = Content.GetCharacterForSkin( skin )
			ui:Text( content_id )
		end,
		value = function( column, self, skin )
			return Content.GetCharacterForSkin( skin )
		end,
	},
	ALIAS = { name = "Alias", width = 120, property = "alias",
		ui = function( column, self, ui, panel, skin )
			ui:PushItemWidth( ui:GetColumnWidth() - 3 )
			local alias = ui:InputText( "##ALIAS", skin.alias or "", ui.TextFlags_EnterReturnsTrue )
			ui:PopItemWidth()
			if alias and alias ~= skin.alias then
				skin.alias = alias
				return true
			end
		end,
		value = function( column, self, skin )
			return skin.alias or ""
		end,
	},
	GENDER = { name = "Gender", width = 60, property = "gender",
		ui = function( column, self, ui, panel, skin )
		 	local modified = false
			if ui:Selectable( skin.gender ) then
		 		ui:OpenPopup( "GENDER" )
		 	end
			if ui:BeginPopup( "GENDER" ) then
				for i, gender in ipairs( GENDER_COMBO ) do
					if ui:MenuItem( gender ) then
						skin.gender = gender
						modified = true
					end
				end
				ui:EndPopup()
			end
			return modified
		end,
		value = function( column, self, skin )
			return skin.gender
		end,
	},

	SPECIES = { name = "Species", width = 80, property = "species",
		ui = function( column, self, ui, panel, skin )
			local modified = false
			if ui:Selectable( skin.species ) then
				ui:OpenPopup( "SPECIES" )
			end
			if ui:BeginPopup( "SPECIES" ) then
				for i, species in ipairs( SPECIES_COMBO ) do
					if ui:MenuItem( species ) then
						skin.species = species
						modified = true
					end
				end
				ui:EndPopup()
			end
			return modified
		end,
		value = function( column, self, skin )
			return skin.species
		end
	},

	BUILD = { name = "Build", width = 200, property = "build",
		ui = function( column, self, ui, panel, skin )
			local modified = false
			if ui:Selectable( skin.build or "" ) then
				ui:OpenPopup( "BUILD" )
			end
			if ui:BeginPopup( "BUILD" ) then
				if self.builds == nil then
					self.builds = copykeys( Content.GetAllCharacterBuilds() )
					table.sort( self.builds )
				end

				for i, build in ipairs( self.builds ) do
					if ui:MenuItem( build ) then
						skin.build = build
						modified = true
					end
				end

				ui:EndPopup()
			end
			return modified
		end,
		value = function( column, self, skin )
			return skin.build
		end
	},

	HEAD = { name = "Head", width = 200, property = "head",
		ui = function( column, self, ui, panel, skin )
			local modified = false
			if ui:SmallButton( ">" ) then
				local content_id = Content.GetCharacterForSkin( skin )
				local def = Content.GetCharacterDef( content_id )
				if self.show_all_heads then
					chargen.QueryHeads( def, skin, self.heads )
				else
					chargen.QueryUnusedHeads( def, skin, self.heads )
				end
				table.sort( self.heads )

				local idx = table.arrayfind( self.heads, skin.head )
				local new_head
				if idx == nil or idx >= #self.heads then
					new_head = self.heads[ 1 ]
				else
					new_head = self.heads[ idx + 1 ]
				end
				if new_head and new_head ~= skin.head then
					skin.head = new_head
					modified = true
				end
			end
	
			ui:SameLine( nil, 5 )
			if ui:Selectable( tostring(skin.head) ) then
				ui:OpenPopup( "HEADS" )
			end
			if ui:BeginPopup( "HEADS" ) then
				local content_id = Content.GetCharacterForSkin( skin )
				local def = Content.GetCharacterDef( content_id )
				if self.show_all_heads then
					chargen.QueryHeads( def, skin, self.heads )
				else
					chargen.QueryUnusedHeads( def, skin, self.heads )
				end
				table.sort( self.heads )

				if #self.heads == 0 then
					ui:TextColored( 0xCC0000FF, "No valid heads available" )
				else
					ui:TextColored( 0x00CCCCFF, "Generic heads are cyan" )
					ui:Separator()

					for i, head_id in ipairs( self.heads ) do
						local head = Content.GetCharacterHead( head_id )
						if head.generic then
							ui:PushStyleColor( ui.Style_Text, 0x00CCCCFF )
						end
						if ui:MenuItem( head_id ) then
							skin.head = head_id
							modified = true
						end
						if head.generic then
							ui:PopStyleColor()
						end
					end
				end
				ui:EndPopup()
			end
			return modified
		end,
		value = function( column, self, skin )
			return skin.head
		end
	},

	SKIN = { name = "Skin", width = 40, property = "skin_colour", property_is_color = true,
		ui = function( column, self, ui, panel, skin )
			local modified = false
			local head = Content.GetCharacterHead( skin.head )
			local palette = SPECIES_PROPS[ skin.species ].SKIN_PALETTE
			local palette_map = SPECIES_PROPS[ skin.species ].SKIN_TONES
			local name
			if head and head.skin_tone and palette_map then
				palette = palette_map[ head.skin_tone ]
				name = string.format( "%s %s skin palette", skin.species, head.skin_tone )
			else
				name = string.format( "%s skin palette", skin.species )
			end
			if palette then
				ui:PushID( "SKIN" )
				local skin_clr = self:ShowPalette( ui, skin.skin_colour, palette, name )
				if skin_clr then
			        skin.skin_colour = skin_clr
			        modified = true
			    end
			    ui:PopID()
			end
		    return modified
		end,
		value = function( column, self, skin )
			return skin.skin_colour or UICOLOURS.WHITE
		end
	},
	HAIR = { name = "Hair", width = 40, property = "hair_colour", property_is_color = true,
		ui = function( column, self, ui, panel, skin )
			local modified = false
			local palette = SPECIES_PROPS[ skin.species ].HAIR_PALETTE
			local name = string.format( "%s hair palette", skin.species )
			if palette then
				ui:PushID( "HAIR" )
				local hair_clr = self:ShowPalette( ui, skin.hair_colour, palette, name )
				if hair_clr then
			        skin.hair_colour = hair_clr
			        modified = true
			    end
			    ui:PopID()
			end
		    return modified
		end,
		value = function( column, self, skin )
			return skin.hair_colour or UICOLOURS.WHITE
		end
	},
	VOICE = { name = "Voice", width = 60, property = "voice_actor",
		ui = function( column, self, ui, panel, skin )
			local modified = false
			if ui:Selectable( skin.voice_actor or "" ) then
				ui:OpenPopup( "VOICE_ACTOR" )
			end
			if ui:BeginPopup( "VOICE_ACTOR" ) then
				local voice_actors = chargen.QueryVoiceActors( skin )
				for i, vo in ipairs( voice_actors ) do
					if ui:MenuItem( vo ) then
						skin.voice_actor = vo
						modified = true
					end
				end
				ui:EndPopup()
			end
			return modified
		end,
		value = function( column, self, skin )
			return skin.voice_actor or ""
		end
	},
	-- TRAITS = { name = "Traits", width = 200,
	-- 	ui = function( column, self, ui, panel, skin )
	-- 		local modified = false
	-- 		if ui:Selectable( table.concat( skin.traits or table.empty, "," )) then
	-- 			ui:OpenPopup( "TRAITS" )
	-- 		end
	-- 		if ui:BeginPopup( "TRAITS" ) then
	-- 			modified = self:ShowTraitsPopup( ui, panel, skin )
	-- 			ui:EndPopup()
	-- 		end
	-- 		return modified
	-- 	end,
	-- 	value = function( column, self, skin )
	-- 		return table.concat( skin.traits or table.empty, "," )
	-- 	end
	-- },
}
local COLUMN_ORDERING =
{
	"EDIT", "NAME", "NICKNAME", "DEF", "ALIAS", "GENDER", "SPECIES", "BUILD", "HEAD", "HAIR", "SKIN", "VOICE"
}
local DEFAULT_COLUMNS =
{
	"NAME", "DEF", "GENDER", "SPECIES", "HEAD", "HAIR", "SKIN",
}

local function SortColumns( a, b )
	return table.arrayfind( COLUMN_ORDERING, a ) < table.arrayfind( COLUMN_ORDERING, b )
end

------------------------------------------------------------------------------
-- Inspector for viewing and editing CONTENT.CHARACTER_SKINS.

local DebugCharacterSkins = class( "DebugCharacterSkins", DebugNode )
DebugCharacterSkins.can_reload = true
DebugCharacterSkins.PANEL_WIDTH = 800
DebugCharacterSkins.PANEL_HEIGHT = 500
DebugCharacterSkins.REGISTERED_CLASS = Screen.CharacterAtlasScreen

DebugCharacterSkins.MENU_BINDINGS =
{
	{
		name = "Skins",
		{
			Text = "View Stats By Species",
			Do = function( self )
			    local filter = function( t )
					return self.filter_name_txt == nil or t.content_id:find( self.filter_name_txt )
			    end

			    local data = {}
				local all_skins = Content.GetAllCharacterSkins()
				for content_id, skins in pairs( all_skins ) do
					local datum = { content_id = content_id, total_count = #skins, counts = {} }
					for i, species in pairs( SPECIES ) do
						datum.counts[ species ] = 0
					end
					for i, skin in ipairs( skins ) do
						if skin.species then
							datum.counts[ skin.species ] = datum.counts[ skin.species ] + 1
						end
					end
					table.insert( data, datum )
				end

			    local tab = DebugTabular( COLUMNS_BY_SPECIES, data, filter )
			    DBG( tab )
			end
		},
		{
			Text = "View Stats By Gender",
			Do = function( self )
			    local filter = function( t )
					return self.filter_name_txt == nil or t.content_id:find( self.filter_name_txt )
			    end

			    local data = {}
				local all_skins = Content.GetAllCharacterSkins()
				for content_id, skins in pairs( all_skins ) do
					local datum = { content_id = content_id, total_count = #skins, counts = {} }
					for i, gender in pairs( GENDER ) do
						datum.counts[ gender ] = 0
					end
					for i, skin in ipairs( skins ) do
						if skin.gender then
							datum.counts[ skin.gender ] = datum.counts[ skin.gender ] + 1
						end
					end
					table.insert( data, datum )
				end

			    local tab = DebugTabular( COLUMNS_BY_GENDER, data, filter )
			    DBG( tab )
			end
		},
		{
			Text = "View Unused Heads",
			Do = function( self )
				local data = {}
			    local all_heads = Content.GetCharacterHeads()
			    local priority = 0
			    for head_id, head in pairs( all_heads ) do
	               table.insert( data, { head_id = head_id, head = head,
	               		  gender = head.gender,
	               		  species = head.species,
	               		  faction_id = head.faction_id,
	               		  priority = head.priority,
	               		  generic = head.generic,
	               		  match_fn = head.match ~= nil and "YES" or "" } )
			    end

			    local filter = function( t )
			    	return not t.head.generic and chargen.IsUnusedHead( t.head_id, t.head )
				end

			    local tab = DebugTabular( COLUMNS_UNUSED_HEADS, data, filter )
			    DBG( tab )
			end
		},
		{},
		{
			Text = "Show All Valid Heads",
			Checked = function( self )
				return self.show_all_heads == true
			end,
			Do = function( self )
				self.show_all_heads = not self.show_all_heads
			end
		},
		{},
		{ 
			Text = "Save",
			Do = function( self )
				Content.SaveAllCharacterSkins()
			end
		},
		{
			Text = "Redo",
			Enabled = function( self )
				return self:CanRedo()
			end,
	        Binding = {key ="Y", CTRL = true },
			Do = function( self )
				self:Redo()
			end,
		},
		{
			Text = "Undo",
			Enabled = function( self )
				return self:CanUndo()
			end,
	        Binding = {key ="Z", CTRL = true },
			Do = function( self )
				self:Undo()
			end,
		},		
		{},
		{
			Text = "Export to .csv",
			Do = function( self )
				local env = TheGame:GetDebugEnv()
				env.c_export_skins( "../character_skins.csv" )
			end,
		},
		{
			Text = "Export to clipboard",
			Do = function( self )
				local env = TheGame:GetDebugEnv()
				env.c_export_skins()
			end,
		},
		{
			Text = "Import from clipboard",
			Do = function( self )
				local env = TheGame:GetDebugEnv()
				env.c_import_skins()
			end,
		},
		{
			Text = "Background Colour",
			Enabled = function( self )
				return TheGame:FE():FindScreen( Screen.CharacterAtlasScreen ) ~= nil
			end,
			CustomMenu = function( option, ui, self )
				local screen = TheGame:FE():FindScreen( Screen.CharacterAtlasScreen )
				local col = self:ShowPalettePopup( ui, screen:GetBackgroundColour(), ATLAS_BG_COLOURS, "Background Palette" )
				if col and col ~= screen:GetBackgroundColour() then
					screen:SetBackgroundColour( col )
				end
			end,
		},
		{
			Text = "Bigger Characters",
			Binding = {key = "EQUALS"},
			Enabled = function( self )
				return TheGame:FE():FindScreen( Screen.CharacterAtlasScreen ) ~= nil
			end,
			Do = function( self )
				local screen = TheGame:FE():FindScreen( Screen.CharacterAtlasScreen )
				screen:ChangeScale( 0.05 )
			end,
		},
		{
			Text = "Smaller Characters",
			Enabled = function( self )
				return TheGame:FE():FindScreen( Screen.CharacterAtlasScreen ) ~= nil
			end,
			Binding = {key = "MINUS"},
			Do = function( self )
				local screen = TheGame:FE():FindScreen( Screen.CharacterAtlasScreen )
				screen:ChangeScale( -0.05 )
			end,
		}
	}
}

function DebugCharacterSkins:init( game )
	self.menu_params = { self }

	self:LoadSettings()

	self.CHARACTER_DEFS_ARRAY = {}
	for content_id, def in pairs( Content.GetAllCharacterDefs() ) do
		if not def.is_template then
			table.insert( self.CHARACTER_DEFS_ARRAY, content_id )
		end
	end
	table.sort( self.CHARACTER_DEFS_ARRAY )


	self.reset_column_widths= true

	self.edit_index = 0
	self.edit_history = {}
	self:PushState()

	self.heads = {}
	self.voice_actors = {}
	self.head_counts = {} -- for filter_dupe_heads

	self.skins = {} -- Shown array of filtered character skin tables
	self.expanded_skins = {} -- Array of skin uuids, indicating skins showing detailed info
	self:PopulateSkins()
end

function DebugCharacterSkins:LoadSettings()
    self.settings = DebugUtil.GetLocalDebugData("DEBUG_CHARACTER_SKINS") or {}
    if self.settings.columns == nil then
        self.settings.columns = shallowcopy( DEFAULT_COLUMNS )
    end
    if self.settings.sort_columns == nil then
    	self.settings.sort_columns = deepcopy( self.settings.columns )
    end
end

function DebugCharacterSkins:SaveSettings()
    DebugUtil.SetLocalDebugData("DEBUG_CHARACTER_SKINS", self.settings)
    DebugUtil.SaveLocalDebugData()
end

function DebugCharacterSkins:PushState()
	self.edit_index = self.edit_index + 1
	self.edit_history[ self.edit_index ] = deepcopy( Content.GetAllCharacterSkins() )

	while #self.edit_history > self.edit_index do
		table.remove( self.edit_history )
	end
end

function DebugCharacterSkins:IsExpanded( skin )
	return skin and table.arrayfind( self.expanded_skins, skin.uuid ) ~= nil
end

function DebugCharacterSkins:CanRedo()
	return self.edit_index < #self.edit_history
end

function DebugCharacterSkins:Redo()
	assert( self.edit_index < #self.edit_history )
	self.edit_index = self.edit_index + 1

	local all_skins = Content.GetAllCharacterSkins()
	table.clear( all_skins )
	table.add( all_skins, deepcopy( self.edit_history[ self.edit_index ] ))

	self:OnModifiedSkins()
	self:PopulateSkins()
end

function DebugCharacterSkins:CanUndo()
	return self.edit_index > 1
end

function DebugCharacterSkins:Undo()
	assert( self.edit_index > 1 )
	self.edit_index = self.edit_index - 1

	local all_skins = Content.GetAllCharacterSkins()
	table.clear( all_skins )
	table.add( all_skins, deepcopy( self.edit_history[ self.edit_index ] ))

	self:OnModifiedSkins()
	self:PopulateSkins()
end

function DebugCharacterSkins:OnModifiedSkins( skin )
	Content.SaveAllCharacterSkins()
	TheGame:BroadcastEvent( "character_skins_changed", skin )
end

function DebugCharacterSkins:CreateSkins( content_id, count )
	local chargen = require "charactergen"

	for i = 1, count do
		Content.CreateCharacterSkin( content_id )
	end

	self:PushState()
	self:OnModifiedSkins()
	self:PopulateSkins()
end

function DebugCharacterSkins:ShowTraitsPopup( ui, panel, skin )
	local modified = false
    for i, aspect_name, aspect in Content.SortedAspects() do
        local filtered = not aspect:IsTrait()

        if not filtered then
            local has_aspect = skin.traits and table.contains( skin.traits, aspect_name )
            if ui:Checkbox( aspect_name, has_aspect ) then
                if has_aspect then
                	table.arrayremove( skin.traits, aspect_name )
                else
                	if skin.traits == nil then
                		skin.traits = {}
                	end
                	table.insert( skin.traits, aspect_name )
                end
				modified = true
            end
        end
    end

    return modified
end

function DebugCharacterSkins:ShowPalette( ui, clr, palette, name )
	clr = clr or UICOLOURS.WHITE
	local new_clr
	if ui:ColorButton( clr ) then
        ui:OpenPopup( "PALETTE" )
    end
    if ui:BeginPopup( "PALETTE" ) then
    	new_clr = self:ShowPalettePopup( ui, clr, palette, name )
        ui:EndPopup()
    end

    return new_clr
end

function DebugCharacterSkins:ShowPalettePopup( ui, clr, palette, name )
    local new_clr = ui:ColorEdit3( "Current", clr )
    if new_clr == clr then
    	new_clr = nil
    end

    ui:Separator()
    ui:Text( name )
    ui:Spacing()

	for i, palette_clr in ipairs( palette ) do
		if (i - 1) % 10 ~= 0 then
			ui:SameLine( nil, 5 )
		end
		ui:PushID( i )
		if ui:ColorButton( palette_clr ) then
			new_clr = palette_clr
			ui:CloseCurrentPopup()
		end
		ui:PopID()
	end

	return new_clr
end

function DebugCharacterSkins:PopulateSkins()
	table.clear( self.skins )

	table.clear( self.head_counts )
	for content_id, skins in pairs( Content.GetAllCharacterSkins() ) do
		for i, skin in ipairs( skins ) do
			local count = self.head_counts[ skin.head or "" ] or 0
			self.head_counts[ skin.head or "" ] = count + 1
		end
	end

	for content_id, skins in pairs( Content.GetAllCharacterSkins() ) do
		for i, skin in ipairs( skins ) do
			if self:FilterSkin( skin, content_id ) then
				table.insert( self.skins, skin )
			end
		end
	end

	self:RefreshAtlasSkins()
end

function DebugCharacterSkins:RefreshAtlasSkins()
	local screen = TheGame:FE():FindScreen( Screen.CharacterAtlasScreen )
	if screen then
		screen:RefreshSkins( self.skins )
	end
end


function DebugCharacterSkins:SetFilterText( txt )
	self.filter_name_txt = txt
	self:PopulateSkins()
end

function DebugCharacterSkins:FilterSkin( skin, content_id )
	if self.filter_dupe_heads and self.head_counts[ skin.head or "" ] <= 1 then
		return false
	end

	if self.filter_name_txt then
		if skin.name and skin.name:find( self.filter_name_txt ) then
			return true
		end

		if content_id:find( self.filter_name_txt ) then
			return true
		end

		return false
	else
		return true
	end
end

function DebugCharacterSkins:RenderCharacterSkin( ui, panel, skin )
	local modified = false
	
	ui:PushID( tostring(skin) )

	local content_id = Content.GetCharacterForSkin( skin )
	local def = Content.GetCharacterDef( content_id )

    for j = 1, #self.settings.columns do
        local column_id = self.settings.columns[j]
        local column = COLUMNS[ column_id ]
        if column.property and def and def[ column.property ] ~= nil then
        	if column.property_is_color then
        		ui:ColorButton( def[ column.property ] )
        	else
				ui:TextColored( 0xCCCCCCFF, def[ column.property ] )
			end
    		if ui:IsItemHovered() then
    			ui:SetTooltip( string.format( "This property is specified by %s and is locked", content_id ))
    		end
		else
	        modified = modified or column:ui( self, ui, panel, skin )
	    end
        ui:NextColumn()
    end

	if table.arrayfind( self.expanded_skins, skin.uuid ) then
		ui:Columns( 1 )
		ui:Indent( 40 )

		local new_bio = ui:InputTextMultiline( "Bio", skin.bio or "" )
		if new_bio then
			skin.bio = new_bio
		end
		if self.bio_edit == nil and ui:IsItemActive() then
			self.bio_edit_string = skin.bio or ""
			self.bio_edit = skin

		elseif self.bio_edit == skin and not ui:IsItemActive() then
			modified = self.bio_edit_string ~= skin.bio
			self.bio_edit = nil
			self.bio_edit_string = nil
		end

		if ui:Button( "DBG" ) then
			DBG( skin )
		end

		ui:SameLine( nil, 10 )
		ui:PushStyleColor( ui.Style_Button, 0x770000FF )
		if ui:Button( "Delete" ) then
			local content_id, skins = Content.GetCharacterForSkin( skin )
			table.arrayremove( self.skins, skin )
			table.arrayremove( skins, skin )
			modified = true
		end
		ui:PopStyleColor()

		ui:Unindent( 40 )
		ui:Separator()
		self:RenderColumns( ui, panel, self.reset_column_widths )
	end
	ui:PopID()

	if modified then
		self:PushState()
		self:OnModifiedSkins( skin )
	end
end

function DebugCharacterSkins:RenderColumnHeaders( ui, panel, set_widths )
	ui:Spacing()
	ui:Separator()
	self:RenderColumns( ui, panel, set_widths )

	local toggle_column
    for j = 1, #self.settings.columns do
        local column_id = self.settings.columns[j]
        local column = COLUMNS[ column_id ]

		if ui:Selectable( column.name ) then
            self:SortByColumn( column_id, COLUMNS, self.skins )
		end

        if ui:BeginPopupContextItem( column_id ) then
			if ui:Button( "Reset widths" ) then
				self.reset_column_widths = true
			end
			ui:Separator()

			for i, column_id in ipairs( COLUMN_ORDERING ) do
				local column = COLUMNS[ column_id ]
				local is_visible = table.arrayfind( self.settings.columns, column_id ) ~= nil
				if ui:Checkbox( column.name.."##TOGGLE", is_visible ) then
					toggle_column = column_id
				end
			end

        	ui:EndPopup()
        end
		ui:NextColumn()
	end

	if toggle_column then
		if not table.arrayremove( self.settings.columns, toggle_column ) then
			table.insert( self.settings.columns, toggle_column )
			table.sort( self.settings.columns, SortColumns )
		end
		self:SaveSettings()
	end

	ui:Columns(1)
	ui:Separator()
end

function DebugCharacterSkins:SortByColumn( column_id, columns, values )
	if self.settings.sort_columns[1] == column_id then
        if self.sort_ascending then
            self.sort_ascending = false
        else
            self.sort_ascending = true
        end
    else
    	-- Prioritize this column in the sort order.
		table.arrayremove( self.settings.sort_columns, column_id )
		table.insert( self.settings.sort_columns, 1, column_id )
    end

    if self.sort_ascending then
        local function SortAscending( a, b )
        	for i, column_id in ipairs( self.settings.sort_columns ) do
        		local column = COLUMNS[ column_id ]
	            local va = column:value( self, a )
	            local vb = column:value( self, b )
	            if va and vb then
		            if va < vb then
		            	return true
		            elseif va > vb then
		            	return false
		            end
		        end
	        end
	        return false -- All columns equal, apparently.
        end
        table.sort( values, SortAscending )
    else
        local function SortDescending( a, b )
        	for i, column_id in ipairs( self.settings.sort_columns ) do
        		local column = COLUMNS[ column_id ]
	            local va = column:value( self, a )
	            local vb = column:value( self, b )
	            if va and vb then
		            if va > vb then
		            	return true
		            elseif va < vb then
		            	return false
		            end
		        end
	        end
	        return false -- All columns equal, apparently.
        end
        table.sort( values, SortDescending )
    end
end

function DebugCharacterSkins:RenderColumns( ui, panel, set_widths )
	ui:Columns( #self.settings.columns )

	if set_widths then
		local offset = 0
		for i, column_id in ipairs( self.settings.columns ) do
			local column = COLUMNS[ column_id ]
			offset = offset + column.width
			ui:SetColumnOffset( i, offset )
		end
	end
end

function DebugCharacterSkins:RenderFilter( ui, panel )
    local filter_txt = ui:InputText( "Filter string", self.filter_name_txt, ui.TextFlags_EnterReturnsTrue )
    if filter_txt and filter_txt ~= self.filter_name_txt then
    	self.filter_name_txt = filter_txt
    	self:PopulateSkins()
    end

    if ui:Button( "Choose Character..." ) then
    	ui:OpenPopup( "CHARACTER_FILTER" )
    end
    if ui:BeginPopup( "CHARACTER_FILTER" ) then
    	for i, id in ipairs( self.CHARACTER_DEFS_ARRAY ) do
    		if ui:Selectable( id ) then
    			self.filter_name_txt = id
    			self:PopulateSkins()
    		end
    	end
    	ui:EndPopup()
    end

    ui:SameLine( nil, 20 )
    if ui:Checkbox( "Filter for duplicate heads", self.filter_dupe_heads ) then
    	self.filter_dupe_heads = not self.filter_dupe_heads
    	self:PopulateSkins()
    end

    if ui:Button( "Clear Filter" ) then
    	self.filter_name_txt = nil
    	self.filter_dupe_heads = nil
    	self:PopulateSkins()
    end
end

function DebugCharacterSkins:RenderPanel( ui, panel )
	self:RenderFilter( ui, panel )
    ui:Separator()

	ui:PushStyleColor( ui.Style_Button, 0x007700955 )
	if ui:Button( "Create Skins..." ) then
		ui:OpenPopup( "CREATE_SKINS" )
	end
	ui:PopStyleColor()

	local screen = TheGame:FE():FindScreen( Screen.CharacterAtlasScreen )
	if screen == nil then
		ui:SameLine( nil, 10 )
		if ui:Button( "Show Atlas" ) then
			screen = Screen.CharacterAtlasScreen( self.skins )
			TheGame:FE():PushScreen( screen )
			self:RefreshAtlasSkins()
		end
	end

	if ui:BeginPopup( "CREATE_SKINS" ) then
		self.create_content_idx = ui:Combo( "Character Def", self.create_content_idx or 1, self.CHARACTER_DEFS_ARRAY )
		self.create_count = ui:SliderInt( "Count", self.create_count or 1, 1, 12 )

		local content_id = self.CHARACTER_DEFS_ARRAY[ self.create_content_idx ]
		local txt = string.format( "Create %d %s!", self.create_count, content_id )
	
		if ui:Button( txt ) then
			self:CreateSkins( content_id, self.create_count )
 			ui:CloseCurrentPopup()
 		end
		ui:EndPopup()
	end
	ui:Spacing()

	-- Render a filtered list of all character skins
	self:RenderSkinsTable( ui, panel )
end

function DebugCharacterSkins:RenderSingleColumn( ui, skin )
    self:RenderColumnHeaders( ui, nil, self.reset_column_widths )
    self:RenderColumns( ui, nil, self.reset_column_widths )
    self:RenderCharacterSkin( ui, nil, skin )
    ui:Columns(1)
    self.reset_column_widths = nil
end

function DebugCharacterSkins:RenderStatsTable( ui, panel, columns )
	ui:Columns( 1 + #columns )

	ui:Separator()
	for i, column in ipairs( columns ) do
		if ui:Selectable( column.name ) then
			self:SortByColumn( i, columns, self.CHARACTER_DEFS_ARRAY )
		end
		ui:NextColumn()
	end

	ui:Text( "Total" )
	ui:NextColumn()
	ui:Separator()

	for i, content_id in ipairs( self.CHARACTER_DEFS_ARRAY ) do
		if self.filter_name_txt == nil or content_id:find( self.filter_name_txt ) then
			for i, column in ipairs( columns ) do
				column:ui( self, ui, panel, content_id )
				ui:NextColumn()
			end

			-- Totals
			ui:Text( tostring(#(Content.GetAllCharacterSkins()[ content_id ] or table.empty)))
			ui:NextColumn()
		end
	end
end

function DebugCharacterSkins:RenderSkinsTable( ui, panel )
	ui:Spacing()
	ui:TextColored( 0x00CCCCFF, string.format( "%d skins shown", #self.skins ))

	if panel:CollapsingHeader( ui, "Characters" ) then
		ui:BeginChild( "SKINS" )
		self:RenderColumnHeaders( ui, panel, self.reset_column_widths )
		ui:Columns(1)

		-- Render every skin that hasn't been filtered.
		self:RenderColumns( ui, panel, self.reset_column_widths )
		for i, skin in ipairs( self.skins ) do
	 		self:RenderCharacterSkin( ui, panel, skin )
		end
		ui:Columns( 1 )
		ui:EndChild()
	end
	
	self.reset_column_widths = nil
end
