local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------

local DebugCollection = class( "DebugCollection", DebugNode )

function DebugCollection:init()
    self.menu_params = { self }

    self:LoadSettings()
    self.items = self:RefreshItems()

    assert( self.items )
    assert( self.COLS and self.DEFAULT_COLS )
end


function DebugCollection:LoadSettings()
    if self.SETTINGS_ID then
        self.settings = DebugUtil.GetLocalDebugData( self.SETTINGS_ID )
    end
    if self.settings == nil then
        self.settings = {}
    end
    if self.settings.columns == nil then
        self.settings.columns = self.DEFAULT_COLS
    end
    table.arrayremove( self.settings.columns, "ID" )
end

function DebugCollection:SaveSettings()
    if self.SETTINGS_ID then
        DebugUtil.SetLocalDebugData( self.SETTINGS_ID, self.settings )
        DebugUtil.SaveLocalDebugData()
    end
end

function DebugCollection:SortByColumn( column_id )
    if self.sort_column == column_id then
        if self.sort_ascending then
            self.sort_ascending = false
        else
            self.sort_ascending = true
        end
    else
        self.sort_ascending = true
        self.sort_column = column_id
    end

    local column = self.COLS[ column_id ]
    local fn = column[3]

    if self.sort_ascending then
        local function SortAscending( a, b )
            return fn( column, a ) < fn( column, b )
        end
        table.sort( self.items, SortAscending )
    else
        local function SortDescending( a, b )
            return fn( column, a ) > fn( column, b )
        end
        table.sort( self.items, SortDescending )
    end
end

function DebugCollection:IsShown( item )
    return true
end

function DebugCollection:RenderPanel( ui, panel )

    ui:Separator()

    ui:Columns( #self.settings.columns + 1 )
    for i = 0, #self.settings.columns do
        local column_id
        if i == 0 then
            column_id = "ID"
        else
            column_id = self.settings.columns[i]
        end
        if column_id == nil then
            break
        end
        local column = self.COLS[ column_id ]
        local header, fn = column[1], column[3]
        if ui:Selectable( header ) and fn then
            self:SortByColumn( column_id )
        end
        ui:NextColumn()

        if ui:BeginPopupContextItem( header ) then
            for j, column_id, column in sorted_pairs( self.COLS ) do
                if column_id ~= "ID" then
                    local is_shown = table.arrayfind( self.settings.columns, column_id ) ~= nil
                    if ui:Checkbox( column[1], is_shown ) then
                        if is_shown then
                            table.arrayremove( self.settings.columns, column_id )
                        else
                            table.insert( self.settings.columns, i + 1, column_id )
                        end
                        self:SaveSettings()
                    end
                end
            end
            ui:EndPopup()
        end
    end

    ui:Separator()

    local count = 0
    for i, item in ipairs( self.items ) do
        if self:IsShown( item ) then
            if ui:Selectable( item.id, nil, ui.Selectable_SpanAllColumns ) then
                DBG( item )
            end
            ui:NextColumn()

            for j = 1, #self.settings.columns do
                local column_id = self.settings.columns[j]
                local column = self.COLS[ column_id ]
                local key = column[2]

                if type(key) == "string" then
                    ui:Text( tostring(item[ key ]) )
                elseif type(key) == "function" then
                    key( column, ui, panel, item )
                end
                ui:NextColumn()
            end
            count = count + 1
        end
    end

    ui:Columns(1)

    ui:Separator()
    ui:TextColored( 0x55CCCCFF, string.format( "%d items listed", count ))
end

