local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------
-- A debug source for uh, colours?

local DebugColourSettings = class( "DebugColourSettings", DebugNode )

function DebugColourSettings:init(  )
    package.loaded[ "content/colour_settings" ] = nil
    self.colour_snapshots = require "content/colour_settings"

    if TheGame:FE():GetTopScreen().GetCurrentColourName then
        self.current_snapshot = TheGame:FE():GetTopScreen():GetCurrentColourName()
    end
    
    self.current_snapshot = self.current_snapshot or (self.colour_snapshots and "DEFAULT") or next(self.colour_snapshots)
    self:UpdateSnapshotList()
end

function DebugColourSettings:UpdateSnapshotList()
    self.snapshot_list = {}
    for k,v in pairs(self.colour_snapshots) do
        table.insert(self.snapshot_list, k)
    end
    table.sort (self.snapshot_list)

end

function DebugColourSettings:HSVAdjust(ui, name, vals)

    local h,s,v = table.unpack(vals.hsv)
    local r,g,b = table.unpack(vals.colour)
    
    ui:Text(name)
    ui:PushID(name)

	ui:PushItemWidth(-60)	-- Use full column width

    local colour = MakeHexColour(r,g,b,1)
    local new_colour = ui:ColorSelector( name, colour, false ) 
    if new_colour ~= colour then
        local r,g,b,a = HexColour(new_colour)
        vals.colour[1] = r
        vals.colour[2] = g
        vals.colour[3] = b
        self.apply = true
    end


    local new_h = ui:SliderFloat("H", h, 0, 1)
    local new_s = ui:SliderFloat("S", s, 0, 3)
    local new_v = ui:SliderFloat("V", v, 0, 3)

	ui:PopItemWidth()

    ui:PopID()
    
    if new_h ~= h or new_s ~= s or new_v ~= v then
        vals.hsv[1] = new_h
        vals.hsv[2] = new_s
        vals.hsv[3] = new_v
        self.apply = true
    end
end


function DebugColourSettings:Save()
    local filename = 'scripts/content/colour_settings.lua'

    local dumps = serpent.dump(self.colour_snapshots, { indent = '  ', sortkeys = true, compact = false, sparse = false } )

    engine.data:SaveFile( filename, dumps, function(ok) 
            print( string.format( "Write colour_settings file %s : '%s'", ok and "OK" or "FAILED", filename))
        end )

end


function DebugColourSettings:Reload()
    package.loaded[ "content/colour_settings" ] = nil
    self.colour_snapshots = require "content/colour_settings"
end

function DebugColourSettings:LoadSetting(name)
    if self.colour_snapshots[name] then
        self.current_snapshot = name
        self.apply = true
    end
end

local function CreateNewSettings()
  return {
        shadows = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
        midtones = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
        highlights = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
    }
end

function DebugColourSettings:RenderPanel( ui, panel )

    local buttonColumnWidth = 80

	ui:PushItemWidth(-60)	-- Use full column width

    local snapshot_idx = ui:Combo( "Snapshot", table.arrayfind(self.snapshot_list, self.current_snapshot) or 1, self.snapshot_list )
    
    if self.snapshot_list[snapshot_idx] and self.snapshot_list[snapshot_idx] ~= self.current_snapshot then
        self:LoadSetting(self.snapshot_list[snapshot_idx])
    end

    local buttonColumnWidth = (ui:GetContentRegionAvail() - 60 - 15) / 4;   -- 60 = border, 15 = 3 * gap of 5

    if ui:Button("New", buttonColumnWidth) then
        self.new_name = ""
        ui:OpenPopup("NEW_SNAPSHOT_POPUP")
    end

    ui:SameLine(nil, 5)
    if ui:Button("Reset", buttonColumnWidth) then
        self.colour_snapshots[self.current_snapshot] = CreateNewSettings()
        self.apply = true
    end

    
    ui:SameLine(nil, 5 )
    if ui:Button("Save", buttonColumnWidth) then
        self:Save()
    end

    ui:SameLine(nil, 5 )
    if ui:Button("Revert", buttonColumnWidth) then
        self:Reload()
    end



    if self.colour_snapshots[self.current_snapshot] then
        ui:Spacing()
        self:HSVAdjust(ui, "Shadows", self.colour_snapshots[self.current_snapshot].shadows)
        ui:Spacing()
        self:HSVAdjust(ui, "Midtones", self.colour_snapshots[self.current_snapshot].midtones)
        ui:Spacing()
        self:HSVAdjust(ui, "Highlights", self.colour_snapshots[self.current_snapshot].highlights)
        ui:Spacing()
    end

    

    if ui:BeginPopup("NEW_SNAPSHOT_POPUP", ui.AlwaysAutoResize) then
        ui:Text("Name of new snapshot...")
        local new_name = ui:InputText("", self.new_name)
        
        if new_name then
            self.new_name = new_name
        end

        if ui:Button("Create New",nil,nil, not (self.new_name and #self.new_name > 0 and not self.colour_snapshots[self.new_name])) then
            self.colour_snapshots[self.new_name] = CreateNewSettings()
            self:UpdateSnapshotList()
            self:LoadSetting(self.new_name)
            ui:CloseCurrentPopup()
        end

        if self.new_name and self.colour_snapshots[self.new_name] then
            ui:SameLine()
            ui:Text("Already Exists...")
        elseif self.new_name == nil or self.new_name == "" then
            ui:SameLine()
            ui:Text("Enter a name first")
        end

        if ui:Button("Cancel") then
            --ui:CloseCurrentPopup()
        end

        ui:EndPopup()
    end


    if ui:BeginPopup("SaveCheck", ui.AlwaysAutoResize) then
        
        ui:Text("Save current changes first?")
        if ui:Button("Save") then
            self:Save()
            self:LoadSetting(self.snapshot_list[snapshot_idx])
            ui:CloseCurrentPopup()

        end
        if ui:Button("Discard") then
            self:Reload()
            self:LoadSetting(self.snapshot_list[snapshot_idx])
            ui:CloseCurrentPopup()
        end
        ui:EndPopup()
    end



    if self.apply then
        if self.colour_snapshots[self.current_snapshot] then
            local vals = self.colour_snapshots[self.current_snapshot]
--            if TheGame:FE():GetTopScreen().ApplyColourCorrectionValues then
--                TheGame:FE():GetTopScreen():ApplyColourCorrectionValues(vals)
--            end
        end
        self.apply = false

        --apply to the shader values, if they are present
    end


end