local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"
require "ui/screens/fightscreen"
--------------------------------------------------------------------
-- A debug source for the main map simulation

local TEAM_ORDER = {"BLUE", "RED"}

local DebugFightScreen = class( "DebugFightScreen", DebugNode )
DebugFightScreen.REGISTERED_CLASS = Screen.FightScreen

DebugFightScreen.MENU_BINDINGS =
{
    {
        name = "Battle",
        {
            Text = "Condensed View",
            Binding = {key ="W", CTRL = true },
            Visible = function( screen, node )
                return node ~= nil
            end,
            Checked = function( screen, node )
                return node.settings.condensed
            end,
            Do = function( screen, node )
                node.settings.condensed = not node.settings.condensed
                node:SaveDebugSettings()
            end
        },
        {
            Text = "Render Flags",
            CustomMenu = function( self, ui, screen )
                for i, k in sorted_pairs(FightSim.DBG_RENDER) do
                    local flag = FightSim.DBG_RENDER[k]
                    if flag ~= FightSim.DBG_RENDER.NONE and flag ~= FightSim.DBG_RENDER.ANY then
                        local is_checked = CheckBits( screen.fightsim.debug_render, flag )
                        if ui:MenuItem( tostring(k), nil, is_checked ) then
                            screen.fightsim.debug_render = ToggleBits( screen.fightsim.debug_render, flag )
                        end
                    end
                end
            end
        },
        {
            Text = "Pause",
            Binding = {key = "SPACE" },
            Do = function( screen, node )
                local DebugUtil = require "debug/debug_util"
                DebugUtil.TogglePauseTime( node )
            end,
        }
    }
}

function DebugFightScreen:init( screen )
    assert( is_instance( screen, Screen.FightScreen ))
    self.screen = screen
    self.sim = screen.fightsim
    self.hover_fighter = screen.hover_fighter -- For RenderContextMenu
    self.menu_params = {screen, self}
    self.settings = self:LoadDebugSettings()
end

function DebugFightScreen:LoadDebugSettings()
    local data = DebugUtil.GetLocalDebugData( "DEBUG_FIGHT_SCREEN" ) or {}
    return data
end

function DebugFightScreen:SaveDebugSettings()
    DebugUtil.SetLocalDebugData( "DEBUG_FIGHT_SCREEN", self.settings )
    DebugUtil.SaveLocalDebugData()
end

function DebugFightScreen:BeginRenderPanel( ui, panel )
    if self.settings.condensed then
        local flags = ui.MenuBar| ui.NoTitleBar | ui.NoResize | ui.NoMove
        local w, h = ui:DisplaySize()
        ui:SetNextWindowSize( 400, 125, ui.ImGuiSetCond_Always )
        ui:SetNextWindowPos( w/2 - 200, 0, ui.ImGuiSetCond_Always )
        return ui:Begin( "Battle Debug", flags )
    end
end

function DebugFightScreen:EndRenderPanel( ui, panel )
    if self.settings.condensed then
        ui:End()
    end
end

function DebugFightScreen:RenderPanel( ui, panel )
    if self.settings.condensed then
        self:RenderCondensedPanel( ui, panel )
    else
        self:RenderFullPanel( ui, panel )
    end
end

function DebugFightScreen:RenderContextMenu( ui, panel )
    if ui:MenuItem( "Win", nil, nil, self.screen.battle:GetBattleResult() == nil ) then
        self.screen.battle:Win()
        self.screen:ResumePlay()
    end

    if ui:MenuItem("Lose", nil, nil, self.screen.battle:GetBattleResult() == nil) then
        self.screen.battle:Lose()
        self.screen:ResumePlay()
    end

    if self.hover_fighter then
        ui:Separator()
        ui:TextColored( 0x00FFFFFF, tostring(self.hover_fighter))
        ui:Separator()

        panel:AddDebugMenu( ui, debug_menus.FIGHTER_BINDINGS, { self.hover_fighter } )
    end

    ui:Separator()

    self.screen.battle:GetHandDeck():RenderDebugUI( ui, panel, self.screen.battle )
end

function DebugFightScreen:RenderCondensedPanel( ui, panel )
    panel:AppendTable( ui, self.screen, self.screen.name )
    if self.screen.fightsim then
        ui:SameLine( nil, 10 )
        panel:AppendTable( ui, self.screen.fightsim, "Sim" )
        ui:SameLine( nil, 10 )
        panel:AppendTable( ui, self.screen.battle, "Battle" )
    end

    local dbg_env = TheGame:GetDebugEnv()
    if dbg_env.wx then
        ui:Text( string.format( "World: <%2.1f, %2.1f>", dbg_env.wx, dbg_env.wz ))
    end

    ui:Separator()

    DebugUtil.RenderTimeScale( self, ui, panel )

    self:RenderWinLoseButtons( ui, panel )
end

function DebugFightScreen:RenderWinLoseButtons( ui, panel )
    ui:PushStyleColor( ui.Style_Button, 0x007700FF )
    if ui:Button( "Win", nil, nil, self.screen.battle:GetBattleResult() ~= nil ) then
        self.screen.battle:Win()
        self.screen:ResumePlay()
    end
    ui:PopStyleColor()

    ui:SameLine(nil, 10)
    if ui:Button("Lose", nil, nil, self.screen.battle:GetBattleResult() ~= nil) then
        self.screen.battle:Lose()
        self.screen:ResumePlay()
    end
end

function DebugFightScreen:RenderFullPanel( ui, panel )

    self:RenderCondensedPanel( ui, panel )

    -------------------------------------------------------------

    if ui:CollapsingHeader( "PRESETS" ) then
        if self.presets == nil then
            self.presets = DebugFightTest( function( preset ) self:ApplyPreset( preset ) end )
        end
        self.presets:RenderPanel( ui, panel )
    end

    if panel:CollapsingHeader( ui, "BATTLE" ) then
        local new_phase = ui:SliderInt("Music Phase", self.screen.music_phase or 1, 1, 4)
        if new_phase ~= self.screen.music_phase then
            self.screen:SetMusicPhase(new_phase)
        end

        if self.screen.battle then
            self.screen.battle:RenderDebugUI( ui, panel )
        end 
    end

    if panel:CollapsingHeader( ui, "PREVIEW" ) then
        self.screen.damage_preview:RenderDebugUI( ui, panel )
    end

    if panel:CollapsingHeader( ui, "VIZ" ) then
        if ui:Checkbox("VIZ STEP", self.screen.viz_break ~= nil) then
            if self.screen.viz_break ~= nil then
                self.screen.viz_break = nil
            else
                self.screen.viz_break = true
            end
        end
        if self.screen.viz_break ~= nil then
            ui:SameLine( nil, 10 )
            if ui:SmallButton( "Step" ) then
                self.screen.viz_break = false
            end
        end

        for key, coro in pairs( self.screen.coros ) do
            if coro == self.screen.viz_coro then
                key = "VIZ"
            elseif coro == self.screen.wait_coro then
                key = "WAIT"
            end
            DebugUtil.RenderCoroUI( ui, panel, key, coro )
        end
        
        local hover_fighter = self.screen.hover_fighter

        if self.screen.current_event then
            local clr = hover_fighter and table.arrayfind( self.screen.current_event, hover_fighter ) and 0xCC00CCFF
            panel:AppendTable( ui, self.screen.current_event, "*"..tostring(self.screen.current_event[1]), clr )
            if clr then
                ui:SameLine( nil, 20 )
                ui:Text( serpent.line( self.screen.current_event ))
            end
        end
        for i, ev in ipairs( self.screen.event_queue or table.empty ) do
            local clr = hover_fighter and table.arrayfind( ev, hover_fighter ) and 0xCC00CCFF
            panel:AppendTable( ui, ev, ev[1], clr )
            if clr then
                ui:SameLine( nil, 20 )
                ui:Text( serpent.line( ev ))
            end
        end
    end
    
     -- Debug teams
    ---------------------------
    if panel:CollapsingHeader( ui, "TEAMS" ) then
        for idx, team_name in ipairs(TEAM_ORDER) do
            ui:PushID(team_name)

            if idx > 1 then
                ui:Separator()
            end
     
            local team = self.screen.battle.teams[team_name]
            self:RenderTeam( ui, panel, idx, team, team_name )

            ui:PopID()
        end
    end

    DebugUtil.FightCamDebug(ui, self.sim:GetCamera())
end

function DebugFightScreen:ApplyPreset( scenario )
    if self.screen and self.screen:IsOnStack() then
        TheGame:FE():PopScreen( self.screen )
    end

    self.screen = Screen.FightScreen()
    TheGame:FE():PushScreen( self.screen )
end

function DebugFightScreen:RenderTeam( ui, panel, idx, team, team_name )
    panel:AppendTable( ui, team, string.format( "Team %s", team_name ) )
    
    if self.screen:IsOnStack() then
        for i, fighter in ipairs(team.fighters) do
            ui:Indent( 10 )
            ui:Bullet()
            panel:AppendTable( ui, fighter, fighter.name )

            local ent = self.screen.fightsim and self.screen.fightsim.mgr:GetAnimFighterByFighter(fighter)

            ui:SameLine( nil, 10 )
            if ent == nil then
                ui:TextColored( 0xCC0000FF, "NOT SPAWNED" )
            else
                panel:AppendTable( ui, ent )

                local settled, reason = ent.cmp.AnimFighter:IsSettled()
                if reason then
                    ui:SameLine( nil, 10 )
                    ui:TextColored( 0xCC0000FF, reason )
                end
            end
            ui:Unindent( 10 )
        end

        ui:Spacing()
        
        for i, fighter in ipairs(team.reserve_fighters) do
            ui:Indent( 10 )
            ui:Bullet()
            panel:AppendTable( ui, fighter, fighter.name )

            ui:SameLine( nil, 10 )
            ui:Text( string.format( "Joins in %d turns", fighter:GetStartTurn() - self.screen.battle:GetTurns() ))
            ui:Unindent( 10 )
        end
    end

end