local DebugUtil = require "debug/debug_util"
require "battle/battle_engine"
local DebugFightTest = class("DebugFightTest", DebugNode )

local BASE_FIGHT = 
{
    LOCATION = "OUTDOORS",
    NAME = "Untitled Fight",
    FLAGS = BATTLE_FLAGS.NONE,
    TEAMS =
    {
        [TEAM.RED] = {members = {}, },
        [TEAM.BLUE] = {members = {}, },
    },
}

local BASE_DATA =
{
    last_selected = nil,
    fights = {},
}

function DebugFightTest:init( start_fn )
    self.fight_presets = {}
    self.start_fn = start_fn
    self.new_fight_name = "Untitled Fight"
    self.agent_search = {}

    self.fight_locations = {}
    for id, loc in pairs(Content.GetLocationContents()) do
        table.insert(self.fight_locations, id)
    end
    table.sort( self.fight_locations )

    self:LoadFightPresets()
    self:SelectFight( self:GetFightPresetData( self.fight_presets.last_selected))
end

function DebugFightTest:LoadFightPresets()
    self.fight_presets = deepcopy(DebugUtil.GetLocalDebugData("DEBUG_FIGHTS")) or deepcopy(BASE_DATA)
end

function DebugFightTest:SaveFightPresets()
    DebugUtil.SetLocalDebugData("DEBUG_FIGHTS", self.fight_presets)
    DebugUtil.SaveLocalDebugData()
end

function DebugFightTest:SaveFightAs(fight)
    local fight_data = self:GetFightPresetData( fight.NAME )
    if fight_data == nil then
        fight_data = deepcopy( BASE_FIGHT )
        table.insert( self.fight_presets.fights, fight_data )
    end

    for team, data in pairs(fight.TEAMS) do
        table.clear( fight_data.TEAMS[ team ].members )
        for i, member in ipairs(data.members) do
            local t =
            {
                agent_def = member:GetContentID(),
                skin_uuid = member:GetSkinID(),
                start_turn = member.start_turn,
                player_controlled = member.player_controlled,
            }
            if member.health then
                t.health = member.health:Get()
            end
            if member.player_controlled then
                t.build = member.build
            end
            if member.battler then
                t.battle_cards = {}
                for i, card in ipairs( member.battler:GetCards() ) do
                    table.insert( t.battle_cards, card.id )
                end
            end
            fight_data.TEAMS[team].members[i] = t
        end
    end

    fight_data.LOCATION = fight.LOCATION
    fight_data.FLAGS = fight.FLAGS
    fight_data.NAME = fight.NAME
    fight_data.ADVANTAGE = fight.ADVANTAGE
    
    fight_data.ENEMY_SCALE = fight.ENEMY_SCALE
    fight_data.ALLY_SCALE = fight.ALLY_SCALE

    self:SaveFightPresets()
end

function DebugFightTest:GetFightPresetData(name)
    name = name or self.fight_presets.last_selected

    for i, fight in ipairs( self.fight_presets.fights ) do
        if fight.NAME == name then
            return fight
        end
    end
end

function DebugFightTest:GenerateFightPreset( fight_data )
    local fight = deepcopy(BASE_FIGHT)
    fight.LOCATION = fight_data.LOCATION
    fight.FLAGS = fight_data.FLAGS
    fight.NAME = fight_data.NAME
    fight.ADVANTAGE = fight_data.ADVANTAGE
    fight.ENEMY_SCALE = fight_data.ENEMY_SCALE
    fight.ALLY_SCALE = fight_data.ALLY_SCALE

    for team, team_data in pairs(fight_data.TEAMS) do
        for idx, member in ipairs(team_data.members) do
            local agent
            if member.agent_data then
                -- old?
                local load_ok, ret = pcall(loadstring(member.agent_data))
                if load_ok and is_instance( ret, Agent ) then
                    agent = ret
                end
            elseif member.agent_def then
                local skin_table = Content.GetCharacterSkin( member.skin_uuid )
                agent = Agent( member.agent_def, skin_table )
                if member.health and agent.health then
                    agent.health:Set( member.health )
                end
                if member.build then
                    agent.build = member.build
                end
                if member.battle_cards and agent.battler then
                    for i, card_id in ipairs( member.battle_cards ) do
                        agent.battler:LearnCard( card_id )
                    end
                end
            end

            if agent then
                fight.TEAMS[ team ].members[ idx ] = agent
                agent.DEBUG_FIGHT_TEST = true
                agent.start_turn = member.start_turn
                agent.player_controlled = member.player_controlled
            end
        end
    end

    return fight
end

function DebugFightTest:GenerateScenario( fight )
    local location = Location( fight.LOCATION )
    local scenario = Scenario( location, { flags = fight.FLAGS } )

    for team_colour, team_data in pairs(fight.TEAMS) do
        for i, member in ipairs(team_data.members) do
            scenario:AddAgentToTeam(member, team_colour )
            if member.player_controlled then
                scenario:SetPlayerControlled(member)
            end
        end
    end

    scenario:SetAdvantage( fight.ADVANTAGE )

    scenario.enemy_scale = fight.ENEMY_SCALE or 1
    scenario.ally_scale = fight.ALLY_SCALE or 1
    return scenario
end

function DebugFightTest:SelectFight( fight_data )
    if fight_data == nil then
        self.current_fight = nil

    else
        self.fight_presets.last_selected = fight_data.NAME
        self.current_fight = self:GenerateFightPreset( fight_data )
        self:SaveFightPresets()
    end
end

function DebugFightTest:MakeNewPreset(ui, panel)
    if ui:BeginPopup("New Preset", ui.AlwaysAutoResize) then
        local combat_parties = require "content/combat_parties"
        for k, fn in pairs( combat_parties ) do
            if ui:Selectable( k, self.template_name == k, ui.DontClosePopups ) then
                self.template_name = k
                self.new_fight_name = string.format( "Copy of %s", k )
            end
        end
        for i, preset in ipairs( self.fight_presets.fights ) do
            if ui:Selectable( preset.NAME, self.template_name == preset.NAME, ui.DontClosePopups ) then
                self.template_name = preset.NAME
                self.new_fight_name = string.format( "Copy of %s", preset.NAME )
            end
        end

        self.template_difficulty = ui:SliderInt( "Difficulty", self.template_difficulty or 1, 1, 4 ) or self.template_difficulty

        ui:Separator()

        local name = ui:InputText("", self.new_fight_name)
        if name and #name > 0 and self.new_fight_name ~= name then
            self.new_fight_name = name
        end

        ui:SameLine(nil, 10 )
        if self.template_name == nil then
            ui:TextColored( 0xCC0000FF, "Select a template" )
        elseif self.template_difficulty == nil then
            ui:TextColored( 0xCC0000FF, "Specify a preset difficulty" )
        elseif combat_parties[self.new_fight_name] ~= nil then
            ui:TextColored( 0xCC0000FF, "A combat party already has that name" )
        elseif self:GetFightPresetData( self.new_fight_name ) then
            ui:TextColored( 0xCC0000FF, "A preset already has that name" )
        elseif ui:Button("Ok", 100 ) then
            local fight
            
            if self.template_name and combat_parties[ self.template_name ] then
                fight = deepcopy( BASE_FIGHT )
                local members = combat_parties[ self.template_name ]( self.template_difficulty )
                for i, id in ipairs( members ) do
                    table.insert( fight.TEAMS[ TEAM.RED ].members, Agent( id ))
                end
            else
                fight = self:GenerateFightPreset( self:GetFightPresetData( self.template_name ))
            end

            fight.NAME = self.new_fight_name

            self:SaveFightAs( fight )
            self:SelectFight( self:GetFightPresetData( fight.NAME ))

            ui:CloseCurrentPopup()
        end

        ui:EndPopup()
    end
end


function DebugFightTest:CustomizeFight(ui, panel)
    local invalid_start = nil
    for team_id, t in pairs( self.current_fight.TEAMS ) do
        if t.members == nil or #t.members == 0 then
            invalid_start = string.format( "Team %s has no members", team_id )
        end
    end

    if ui:Button("Start Fight", nil, nil, invalid_start ~= nil ) then
        local scenario  = self:GenerateScenario( self.current_fight )
        if self.start_fn then
            self.start_fn( scenario )
        else
            self:StartFight( scenario )
        end
    end
    if ui:IsItemHovered() then
        ui:SetTooltip( invalid_start or "Start the fight!" )
    end

    -- Current fight customization
    ui:PushStyleColor(ui.Style_Button, 0x228B22FF)
    ui:SameLine(nil, 20)
    if ui:Button("Save") then
        self:SaveFightAs(self.current_fight)
    end
    ui:PopStyleColor()
    ui:SameLine( nil, 10 )

    ui:PushStyleColor(ui.Style_Button, 0x8B0000FF)
    if ui:Button("Delete") then
        local fight_data = self:GetFightPresetData( self.fight_presets.last_selected )
        if fight_data then
            table.arrayremove( self.fight_presets.fights, fight_data )
        end
        self:SaveFightPresets()
        self.current_fight = nil
    end

    ui:PopStyleColor()

    if self.current_fight == nil then return end

    local new_name = ui:InputText( "Name", self.current_fight.NAME or "" )
    if new_name and new_name ~= self.current_fight.NAME then
        local combat_parties = require "content/combat_parties"
        if combat_parties[new_name] ~= nil then
            ui:TextColored( 0xCC0000FF, "A combat party already has that name" )
        elseif self:GetFightPresetData( new_name ) then
            ui:TextColored( 0xCC0000FF, "A preset already has that name" )
        else
            self.current_fight.NAME = new_name
            self:SaveFightPresets()
        end
    end

    local current_loc = table.arrayfind(self.fight_locations, self.current_fight.LOCATION)
    local new_loc = ui:Combo("Location", current_loc or 1, self.fight_locations)
    if new_loc ~= current_loc then
        self.current_fight.LOCATION = self.fight_locations[new_loc]
        self:SaveFightPresets()
    end

    do
        self.current_fight.ALLY_SCALE = ui:SliderInt( "Ally Scale", self.current_fight.ALLY_SCALE or 1, 1, 10 ) or 1
        self.current_fight.ENEMY_SCALE = ui:SliderInt( "Enemy Scale", self.current_fight.ENEMY_SCALE or 1, 1, 10 ) or 1

        ui:Text( "Advantage:" )
        ui:SameLine( nil, 20 )
        local radio_idx = 1
        if self.current_fight.ADVANTAGE == TEAM.BLUE then
            radio_idx = 2
        elseif self.current_fight.ADVANTAGE == TEAM.RED then
            radio_idx = 3
        end

        if ui:RadioButton( "None", radio_idx, 1 ) then
            self.current_fight.ADVANTAGE = nil
            self:SaveFightPresets()
        end

        ui:SameLine( nil, 10 )
        if ui:RadioButton( "Blue", radio_idx, 2 ) then
            self.current_fight.ADVANTAGE = TEAM.BLUE
            self:SaveFightPresets()
        end

        ui:SameLine( nil, 10 )
        if ui:RadioButton( "Red", radio_idx, 3 ) then
            self.current_fight.ADVANTAGE = TEAM.RED
            self:SaveFightPresets()
        end
    end

    local new_flags = panel:AppendBitField( ui, self.current_fight.FLAGS or 0, BATTLE_FLAGS, BATTLE_FLAGS_STRINGS )
    if new_flags ~= self.current_fight.FLAGS then
        self.current_fight.FLAGS = new_flags
        self:SaveFightPresets()
    end

    ui:Separator()
    ui:Text("Blue Team: (Player)")
    self:CustomizeTeam(ui, panel, self.current_fight, TEAM.BLUE)

    ui:Separator()
    ui:Text("Red Team: (Enemy)")
    self:CustomizeTeam(ui, panel, self.current_fight, TEAM.RED)
    ui:Separator()
end

function DebugFightTest:CustomizeTeam(ui, panel, data, team_id)
    ui:PushID(team_id)
    local team = data.TEAMS[team_id]

    ui:Indent( 20 )
    if #team.members == 0 then
        ui:Text("This team is empty.")
    end

    for idx, agent in ipairs(team.members) do
        self:CustomizeAgent(ui, panel, idx, team.members)
    end

    ui:PushStyleColor(ui.Style_Button, 0x228B22FF)
    if ui:SmallButton("+") then
        ui:OpenPopup("Pick Agent")
    end
    ui:PopStyleColor()

    if ui:BeginPopup("Pick Agent", ui.AlwaysAutoResize) then
        local selection, skin_table = DebugUtil.ShowAgentSearch(self.agent_search, ui)
        if selection then
            local member = {}
            member = Agent(selection, skin_table)
            member.DEBUG_FIGHT_TEST = true 
            member.player_controlled = member:HasTag("PC")
            table.insert(team.members, member)

            self:SaveFightAs(self.current_fight, self.fight_presets.last_selected)
            ui:CloseCurrentPopup()
        end
        ui:EndPopup()
    end
    ui:Unindent( 20 )


    ui:PopID()
end

function DebugFightTest:CustomizeAgent(ui, panel, idx, members)
    ui:PushID(idx)
    ui:PushStyleColor(ui.Style_Button, 0x8B0000FF)

    local removed
    if ui:Button("-", 0, 14) then
        table.remove(members, idx)
        removed = true
    end

    if ui:IsItemHovered() then
        ui:SetTooltip("Remove Member")
    end
    ui:PopStyleColor()

    if removed then
        ui:PopID()
        self:SaveFightPresets()
        return
    end

    local member = members[idx]

    ui:SameLine( nil, 10 )
    if ui:Button("*", 0, 14) then
        ui:OpenPopup("MEMBER_PROPERTIES")
    end
    if ui:IsItemHovered() then
        ui:SetTooltip( "Set fighter properties" )
    end
    if ui:BeginPopup( "MEMBER_PROPERTIES" ) then
        if ui:Checkbox( "AI", not member.player_controlled ) then
            member.player_controlled = not member.player_controlled
        end
        local outfits = Content.GetOutfitsForCharacter( member:GetContentID() )
        if outfits and member.player_controlled then
            local idx = 1
            for i, outfit in ipairs( outfits ) do
                if member.build == outfit.build then
                    idx = i
                    print( idx )
                end
                outfits[i] = outfit.id
            end
            local outfit_idx = ui:Combo( "Outfits", idx or 1, outfits )
            if outfit_idx and outfit_idx ~= idx then
                member.build = Content.GetOutfit( outfits[ outfit_idx ]).build
            end
        end
        local start_turn = ui:SliderInt( "Start Turn", member.start_turn or 0, 1, 9 )
        if start_turn and start_turn ~= member.start_turn then
            member.start_turn = start_turn
        end
        ui:EndPopup()
    end

    if not is_instance( member, Agent ) then
        ui:Text("???? INVALID MEMBER")
    else
        local agent = member
        ui:SameLine( nil, 20 )
        local txt
        if agent:GetSkinID() ~= nil then
            txt = loc.format( "{1} ({2})", agent:GetContentID(), agent:GetName() )
        else
            txt = agent:GetContentID()
        end
        if ui:Button( txt ) then
            ui:OpenPopup( "EDIT_FIGHTER" )
        end
        if ui:BeginPopup( "EDIT_FIGHTER" ) then
            local current, max = agent:GetHealth()
            if current and max then
                local hp = ui:SliderInt( "Health", current, 1, max )
                if hp ~= current then
                    agent.health:Set( hp )
                end
            end

            if agent.battler then
                if self.battle_cards_ui == nil then
                    self.battle_cards_ui = {}
                end
                DebugUtil.RenderBattleCards( self.battle_cards_ui, ui, agent )
            end
            ui:EndPopup()
        end
    end

    ui:PopID()
end


function DebugFightTest:RenderPanel(ui, panel)
    ui:Indent( 20 )
    for i, data in ipairs(self.fight_presets.fights) do
        if ui:Selectable( data.NAME, self.fight_presets.last_selected == data.NAME ) then
            self:SelectFight( data )
        end
    end
    if ui:Selectable( "+New Preset" ) then
        ui:OpenPopup("New Preset")
    end
    self:MakeNewPreset(ui, panel)

    ui:Unindent( 20 )

    if self.current_fight then
        ui:Separator()
        ui:PushID(self.fight_presets.last_selected)
        self:CustomizeFight(ui, panel)
        ui:PopID()
    end
end

function DebugFightTest:StartFight( scenario )
    local function OnBattleDone( result )
        TheGame:FE():PushScreen(Screen.InfoPopup( "FIGHT OVER", loc.format( "Status: {1}", result), "Okay"))
    end

    TheGame:FE():PushScreen( Screen.FightScreen( nil, OnBattleDone ))
end