local DebugUtil = require "debug/debug_util"

----------------------------------------------------------------------

local DebugLister = class( "DebugLister", DebugNode )
DebugLister.REFRESH_TIME = 0.5
DebugLister.PANEL_WIDTH = 500
DebugLister.PANEL_HEIGHT = 500

local SEARCH_FLAGS, SEARCH_FLAGS_ARRAY = MakeBitField{ "AGENTS", "AGENTS_IN_LIMBO", "QUESTS", "LOCATIONS"}

local SEARCH_KEYVALUES = 0
local SEARCH_LUA = 1

function DebugLister:init()
	self.results = {}
	self.pinned_results = {}
	self.settings =
	{
		search_flags = SEARCH_FLAGS.NONE,
		search_values = true,
		search_plain = true,
		ignore_case = true,
		filter = "",
		lua_filter = "",
		search_type = SEARCH_KEYVALUES
	}
end

function DebugLister:OnActivate()
	self:LoadListerSettings()
	self:DoSearch()
end

function DebugLister:SaveListerSettings()
	DebugUtil.SetLocalDebugData( "LISTER_PANEL", self.settings )
	DebugUtil.SaveLocalDebugData()
end

function DebugLister:LoadListerSettings()
	local data = DebugUtil.GetLocalDebugData( "LISTER_PANEL" )
	if data then
		self.settings = data
		self.settings.search_flags = self.settings.search_flags or SEARCH_FLAGS.NONE
		self.settings.search_type = self.settings.search_type or SEARCH_KEYVALUES
		self.settings.search_plain = self.settings.search_plain or true
	end
end

function DebugLister:SearchItem( root, pred, obj, prefix )
	obj = obj or root
	if obj == nil then
		return
	end
	
	local result = { pcall( pred, self, root, obj, prefix ) }
	local ok = table.remove( result, 1 )
	local match = result[1]
	if not ok or match then
		result.obj = root
		table.insert( self.results, result )
		self.max_match_outputs = math.max( self.max_match_outputs, #result )
		return result
	end
end

local function MatchKeyValues( self, obj, t, prefix )
	local txt
	if self.settings.ignore_case then
		txt = self.settings.filter:lower()
	else
		txt = self.settings.filter
	end
	for k, v in pairs( t ) do
		if self.settings.search_values then
			local ok = false
			if type(v) == "string" then
				local strv = self.settings.ignore_case and v:lower() or v
				ok = strv:find( txt, 1, self.settings.search_plain ) ~= nil
			elseif (type(v) == "number" and v == tonumber(txt)) then
				ok = true
			end
			if ok then
				if prefix then
					return string.format( "%s: %s = %s", prefix, tostring(k), tostring(v))
				else
					return string.format( "%s = %s", tostring(k), tostring(v))
				end
			end
		end
		if self.settings.search_keys then
			local ok = false
			if type(k) == "string" then
				local strk = self.settings.ignore_case and k:lower() or k
				ok = strk:find( txt, 1, self.settings.search_plain ) ~= nil
			elseif (type(k) == "number" and k == tonumber(txt)) then
				ok = true
			end
			if ok then
				if prefix then
					return string.format( "%s: %s = %s", prefix, tostring(k), tostring(v))
				else
					return string.format( "%s = %s", tostring(k), tostring(v))
				end
			end
		end
	end
end


function DebugLister:DoSearchForAgents( gs, pred, in_world, search_subtables )
	for guid, agent in pairs( gs.agents ) do
		if (in_world and not agent:InLimbo()) or (not in_world and agent:InLimbo()) then
			local result = self:SearchItem( agent, pred )
			if search_subtables then
				result = result or self:SearchItem( agent, pred, agent.tags, "tags" )
				result = result or self:SearchItem( agent, pred, agent.loc_table, "loc_table" )
				for i = #agent.def, 1, -1 do
					result = result or self:SearchItem( agent, pred, agent.def[i].tags, string.format( "def[%d].tags", i ))
					result = result or self:SearchItem( agent, pred, agent.def[i], string.format( "def[%d]", i ))
				end
				result = result or self:SearchItem( agent, pred, agent.aspects, "aspects" )
			end
		end
	end
end

function DebugLister:DoSearch()
	table.clear( self.results )
	self.max_match_outputs = 0
	self.search_err = nil

	local gs = TheGame:GetGameState()
	if gs == nil then
		return
	end

	local output_fn
	if self.settings.match_outputs_str then
		local chunk = "local _, obj, t = ...\nreturn " .. self.settings.match_outputs_str
		local f,e = load( chunk, "lister_outputs", nil, TheGame:GetDebugEnv() )
		if not f then
			output_fn = function() end
			self.search_err = tostring(e)
		else
			output_fn = f
		end
	end

	local pred_fn
	if self.settings.search_type == SEARCH_LUA then
		local chunk = "local _, obj, t = ...\n" .. (self.settings.lua_filter or "")
		local f,e = load( chunk, "lister_pred", nil, TheGame:GetDebugEnv() )
		if not f then
			pred_fn = function() return false end
			self.search_err = tostring(e)
		else
			pred_fn = f
		end

	elseif self.settings.search_type == SEARCH_KEYVALUES then
		if self.settings.filter and #self.settings.filter > 0 then
			pred_fn = MatchKeyValues
		end
	end

	if pred_fn == nil then
		return
	end

	local function pred( ... )
		local match = pred_fn( ... )
		if match then
			if output_fn then
				return match, output_fn( ... )
			else
				return match
			end
		end
	end

	-- Now search the stuff we should be searching.
	if CheckBits( self.settings.search_flags, SEARCH_FLAGS.AGENTS ) then
		self:DoSearchForAgents( gs, pred, true, pred_fn == MatchKeyValues )
	end
	if CheckBits( self.settings.search_flags, SEARCH_FLAGS.AGENTS_IN_LIMBO ) then
		self:DoSearchForAgents( gs, pred, false, pred_fn == MatchKeyValues )
	end

	if CheckBits( self.settings.search_flags, SEARCH_FLAGS.QUESTS ) then
		for k, quest in gs:ActiveQuests() do
			local result = self:SearchItem( quest, pred )
			if result then
				result.name = quest:GetContentID()
			end
		end
	end

	if CheckBits( self.settings.search_flags, SEARCH_FLAGS.LOCATIONS ) then
		for k, location in gs:AllLocations() do
			local result = self:SearchItem( location, pred )
			result = result or (location.tags and self:SearchItem( location, pred, location.tags, "tags" ))
			if result then
				result.name = tostring(location)
			end
		end
	end

	return self.results
end

function DebugLister:RenderSearchResult( ui, panel, result )

	-- The matched object
	ui:SameLine( 50 )
	panel:AppendTable( ui, result.obj, result.name )
	if result.addref then
		ui:NewLine()
		ui:SameLine( 50 )
		panel:AppendTable( ui, result.addref )
	end
	ui:NextColumn()

	-- Additional match outputs
	for i = 1, self.max_match_outputs do
		panel:AppendValue( ui, result[i] )
		ui:NextColumn()
	end
end

function DebugLister:RenderPanel( ui, panel )
	if not self.set_focus then
		ui:SetKeyboardFocusHere()
		self.set_focus = true
	end

	if self:RenderOptions( ui, panel ) then
		self.do_search = true
	end
	
	if self.settings.search_type == SEARCH_LUA then
		ui:Text( "local obj" )
		local filter = ui:InputTextMultiline( "", self.settings.lua_filter, -1 )
		if type(filter) == "string" and filter ~= self.settings.lua_filter then
			self.settings.lua_filter = filter
			self.do_search = true
		end

	else
		local filter = ui:InputText( "", self.settings.filter or "" )
		if type(filter) == "string" and filter ~= self.settings.filter then
			self.settings.filter = filter
			self.do_search = true
		end
	end

	local match_outputs = ui:InputText( "Outputs", self.settings.match_outputs_str or "", ui.TextFlags_EnterReturnsTrue )
	if type(match_outputs) == "string" then
		self.settings.match_outputs_str = match_outputs
		self.do_search = true
		self:SaveListerSettings()
	end

	ui:Spacing()

	if self.do_search then
		self.do_search = nil
		local st = engine.inst:QueryPreciseTime()
		self:DoSearch()
		self:SaveListerSettings()
		self.search_time = (engine.inst:QueryPreciseTime() - st) * 1000
	end

	if panel:CollapsingHeader( ui, "RESULTS" ) then
		if ui:Button( "Refresh" ) then
			self.do_search = true
		end

		if self.search_err then
			ui:TextColored( 0xFF0000FF, self.search_err )

		elseif self.search_time then
			if self.settings.search_type == SEARCH_LUA then
				ui:TextColored( 0x00FFFFFF, string.format( "%d results for '%s' (took %.1f ms)", #self.results, self.settings.lua_filter, self.search_time ))
			else
				ui:TextColored( 0x00FFFFFF, string.format( "%d results for '%s' (took %.1f ms)", #self.results, self.settings.filter, self.search_time ))
			end
		end

		ui:Separator()

		ui:Columns( 1 + self.max_match_outputs )
		for i = 1, 1 + self.max_match_outputs do
			local hilite = self.sort_idx == i
			if hilite then
				ui:PushStyleColor(ui.Style_Button, 0x228B22FF)
			end
			if ui:SmallButton( i == 1 and "Match" or string.format( "Result %d", i - 1 ), 200 ) then
				if self.sort_idx == i then
					self.sort_ascending = not self.sort_ascending
				else
					self.sort_idx = i
				end
				local function SortResults( a, b )
					local v1 = self.sort_idx == 1 and a.obj or a[self.sort_idx - 1]
					local v2 = self.sort_idx == 1 and b.obj or b[self.sort_idx - 1]
					if type(v1) ~= "number" or type(v2) ~= "number" then
						v1, v2 = tostring(v1), tostring(v2)
					end
					if self.sort_ascending then
						return v1 > v2
					else
						return v1 < v2
					end
				end
				table.sort( self.results, SortResults )
				table.sort( self.pinned_results, SortResults )
			end
			if hilite then
				ui:PopStyleColor()
			end
			ui:NextColumn()
		end

		ui:Separator()

		for i = #self.pinned_results, 1, -1 do
			local result = self.pinned_results[i]
			if ui:Checkbox( "Pin##pin"..tostring(i), true ) then
				table.remove( self.pinned_results, i )
			else
				self:RenderSearchResult( ui, panel, result )
			end
		end
		if #self.pinned_results > 0 then
			ui:Separator()
		end
		for i = #self.results, 1, -1 do
			local result = self.results[i]
			if ui:Checkbox( "Pin##"..tostring(i), false ) then
				table.insert( self.pinned_results, table.remove( self.results, i ))
			else
				self:RenderSearchResult( ui, panel, result )
			end
		end
		ui:Columns( 1 )
	end
end

function DebugLister:RenderOptions( ui, panel )
	local changed = false

	if ui:RadioButton("Key-Values", self.settings.search_type, SEARCH_KEYVALUES ) then
		self.settings.search_type = SEARCH_KEYVALUES
		changed = true
	end

	ui:SameLine( nil, 20 )
	if ui:RadioButton( "Interpreted Lua", self.settings.search_type, SEARCH_LUA ) then
		self.settings.search_type = SEARCH_LUA
		changed = true
	end
	if ui:IsItemHovered() then
		ui:SetTooltip( "Run the search text as interpreted lua.\nRefer to the searched item as 'obj'. eg.\nreturn obj.guid == 10001" )
	end

	ui:Separator()
	do
		if ui:Button( StringizeBitField( self.settings.search_flags, SEARCH_FLAGS_ARRAY, ", " )) then
			ui:OpenPopup( "SEARCH_FLAGS" )
		end
		if self.settings.search_flags ~= SEARCH_FLAGS.NONE then
			ui:SameLine( nil, 10 )
			if ui:Button( "Clear Flags" ) then
				self.settings.search_flags = 0
				changed = true
			end
		end
		if ui:BeginPopup( "SEARCH_FLAGS" ) then
			for i, v in ipairs( SEARCH_FLAGS_ARRAY ) do
				local bit = SEARCH_FLAGS[ v ]
				if ui:MenuItem( v, nil, CheckBits( self.settings.search_flags, bit )) then
					self.settings.search_flags = ToggleBits( self.settings.search_flags, bit )
					changed = true
				end
			end
			ui:EndPopup()
		end

		if ui:Checkbox( "Ignore Case", self.settings.ignore_case ) then
			self.settings.ignore_case = not self.settings.ignore_case
			changed = true
		end

		ui:SameLine( nil, 10 )
		if ui:Checkbox( "Search Keys", self.settings.search_keys ) then
			self.settings.search_keys = not self.settings.search_keys
			changed = true
		end
		if ui:IsItemHovered() then
			ui:SetTooltip( "Match table keys of type string or number." )
		end

		ui:SameLine( nil, 10 )
		if ui:Checkbox( "Search Values", self.settings.search_values ) then
			self.settings.search_values = not self.settings.search_values
			changed = true
		end
		if ui:IsItemHovered() then
			ui:SetTooltip( "Match table values of type string or number." )
		end

		ui:SameLine( nil, 10 )
		if ui:Checkbox( "Plain", self.settings.search_plain ) then
			self.settings.search_plain = not self.settings.search_plain
			changed = true
		end
		if ui:IsItemHovered() then
			ui:SetTooltip( "Plain search does not use lua pattern matching." )
		end
	end

	return changed
end

return DebugLister

