local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------
-- A debug source for Mods
local MOD_BINDINGS =
{
    name = "Mods",
    {
    }
}

local DebugMods = class( "DebugMods", DebugNode )

DebugMods.MENU_BINDINGS = { MOD_BINDINGS }

function DebugMods:BeginRenderPanel( ui, panel )
    return ui:Begin( string.format( "Mod List###%d", panel.uid ), ui.MenuBar | ui.AlwaysAutoResize )
end

function DebugMods:PostMessage( msg, clr )
    self.msg = tostring(msg)
    self.msg_time = os.clock()
    self.msg_clr = clr or 0xFFFFFFFF
end

local function CanUploadMod( mod )
    if engine.steam == nil then
        return false, "Workshop not enabled"
    end
    
    if mod.title == nil then
        return false, "Missing 'title'"
    end

    if mod.description == nil then
        return false, "Missing 'description'"
    end

    return true
end

function DebugMods:RenderWorkshopPopup( mod, ui, panel )
    if (mod.status == MOD_STATUS.LOCAL or mod.status == MOD_STATUS.DEV) then
        local ok, reason = CanUploadMod( mod )
        local txt
        if mod.workshop_id then
            txt = string.format( "Upload item %d to workshop", mod.workshop_id )
        else
            txt = "Upload NEW item to workshop"
        end
        if ui:MenuItem( txt, nil, nil, ok ) then
            local function OnSubmitted( item, msg )
                print( "OnSubmitted", msg, tostr(item))
                if item.workshop_id and item.workshop_id ~= 0 then
                    Content.SaveWorkshopID( mod, item.workshop_id )
                end

                if item.lastResult == engine.steam.EResultOK then
                    self:PostMessage( msg or "Submit succeeded!", 0x00CC00FF )
                else
                    -- From the workshop docs:
                    -- k_EResultFail (2) - Generic failure.
                    -- k_EResultInvalidParam (8) - Either the provided app ID is invalid or doesn't match the consumer app ID of the item or, you have not enabled ISteamUGC for the provided app ID on the Steam Workshop Configuration App Admin page.
                    -- The preview file is smaller than 16 bytes.
                    -- k_EResultAccessDenied (15) - The user doesn't own a license for the provided app ID.
                    -- k_EResultFileNotFound (9) - Failed to get the workshop info for the item or failed to read the preview file.
                    -- k_EResultFileNotFound (9) - The provided content folder is not valid. (eg. this workshop_id has previously been deleted)
                    -- k_EResultLockingFailed (33) - Failed to aquire UGC Lock.
                    -- k_EResultLimitExceeded (25) - The preview image is too large, it must be less than 1 Megabyte; or there is not enough space available on the users Steam Cloud.

                    msg = msg or "Submit failed!"
                    msg = msg .. string.format( " (error=%s)", tostring(item.lastResult))
                    self:PostMessage( msg, 0xCC0000FF )
                end
            end

            print( "Submitting to workshop:", tostr(mod))
            self:PostMessage( "Submitting to workshop..." )
            engine.steam:SubmitItem( mod, OnSubmitted )
        end
        if ui:IsItemHovered() and reason then
            ui:SetTooltip( reason )
        end

    elseif mod.status == MOD_STATUS.WORKSHOP and engine.steam ~= nil then
        if CheckBits( mod.workshop.flags, engine.steam.EItemStateNeedsUpdate ) then
            if ui:MenuItem( "Update" ) then
                self:PostMessage( "Updating..." )
                engine.steam:DownloadItem( mod.workshop_id )
            end
        elseif not CheckBits( mod.workshop.flags, engine.steam.EItemStateInstalled ) then
            if CheckAnyBits( mod.workshop.flags, engine.steam.EItemStateDownloading | engine.steam.EItemStateDownloadPending ) then
                ui:TextColored( 0x00FF00FF, string.format( "%d/%d bytes downloaded", mod.workshop.bytesProcessed or 0, mod.workshop.bytesTotal or 0 ))
            else
                if ui:MenuItem( "Install" ) then
                    engine.steam:DownloadItem( mod.workshop_id )
                    self:PostMessage( "Installing..." )
                end
            end
        end


        if CheckBits( mod.workshop.flags, engine.steam.EItemStateSubscribed ) then
            if ui:MenuItem( "Unsubscribe" ) then
                self:PostMessage( "Unsubscribing..." )
                engine.steam:UnsubscribeItem( mod.id )
            end
        else
            if ui:MenuItem( "Subscribe" ) then
                self:PostMessage( "Subscribing..." )
                engine.steam:SubscribeItem( mod.id )
            end                    
        end

        -- You can do this through the Steam Workshop page.
        -- Make sure to also delete the associated steam_workshop.txt file.
        -- if mod.workshop_id and ui:MenuItem( "Delete" ) then
        --     self:PostMessage( "Deleting from workshop..." )
        --     engine.steam:DeleteItem( mod.workshop_id )
        -- end

        -- Not necessary, ideally.
        -- if ui:MenuItem( "Refresh" ) then
        --     engine.steam:RefreshItem( mod.id )
        -- end
    end
end


function DebugMods:RenderPanel( ui, panel )
    -- ui:Separator()
    ui:Dummy( 800, 1 )

    ui:Columns( 5 )
    
    ui:Text( "ID" )
    ui:NextColumn()
    
    ui:Text( "Status" )
    ui:NextColumn()

    ui:Text( "Install Path" )    
    ui:NextColumn()

    ui:Text( "User Settings" )
    ui:NextColumn()

    ui:Text( "Workshop" )
    ui:NextColumn()

    ui:Separator()

    local profile = TheGame:GetGameProfile()

    for i, mod in Content.AllMods() do
        if mod.status == MOD_STATUS.BUILTIN then
            -- don't show
        else
            ui:PushID( rawstring(mod))

            ui:Text( tostring(mod.title or "Untitled Mod" ) )

            if ui:IsItemHovered() and mod.description then
                ui:SetTooltip( tostring(mod.description) )
            end

            local enabled = Content.IsModEnabled( mod )
            if ui:Checkbox( "", enabled ) then
                Content.SetModEnabled( mod, not enabled )
                self:PostMessage( string.format( "'%s' enabled == %s! You must restart the game for this to take effect.", tostring(mod.id), not enabled ))
            end
            ui:SameLine( nil, 10 )

            ui:TextColored( 0x00CCCCFF, string.format( "id=%s", tostring(mod.id)))
            ui:SameLine( nil, 5 )
            if ui:SmallButton( "?" ) then
                DBG(mod)
            end

            if mod.version then
                ui:TextColored( 0x00CCCCFF, string.format( "version=%s", tostring(mod.version)))
            end

            if mod.error then
                ui:TextColored( 0xFF0000FF, "LOAD ERROR" )
                if ui:IsItemHovered() and type(mod.error) == "string" then
                    ui:SetTooltip( mod.error )
                end
            end

            ui:Spacing()

            ui:NextColumn()

            ui:TextColored( 0xCCCC00FF, tostring(mod.status))
            if mod.workshop then
                local flags = {}
                if CheckBits( mod.workshop.flags, 1 ) then
                    table.insert( flags, "subscribed" )
                end
                if CheckBits( mod.workshop.flags, 4 ) then
                    table.insert( flags, "installed" )
                end
                if CheckBits( mod.workshop.flags, 8 ) then
                    table.insert( flags, "needs_update" )
                end
                if CheckAnyBits( mod.workshop.flags, 16 | 32 ) then
                    table.insert( flags, "downloading" )
                end
                if #flags > 0 then
                    ui:Text( table.concat( flags, ", " ))
                end
            end

            ui:NextColumn()

            if mod.installPath and #mod.installPath > 0 then
                ui:Text( tostring(mod.installPath))
            else
                ui:TextColored( 0x777777FF, "Not installed" )
            end
            ui:NextColumn()

            local mod_settings = profile:GetModSettings()[ mod.id ]
            if mod_settings then
                for k, v in pairs( mod_settings ) do
                    ui:Text( tostring(k)..":" )
                    ui:SameLine( nil, 5 )
                    ui:TextColored( 0xCCCC00FF, tostring(v))
                end
            end

            ui:NextColumn()

            if ui:Button( "Workshop", nil, nil, engine.steam == nil ) then
                ui:OpenPopup( "WORKSHOP" )
            end

            if ui:BeginPopup( "WORKSHOP" ) then
                self:RenderWorkshopPopup( mod, ui, panel )
                ui:EndPopup()
            end
            ui:NextColumn()

            ui:PopID()
            ui:Separator()
        end
    end
    ui:Columns( 1 )

    ui:Separator()

    local now = os.clock()
    if self.msg and (now - self.msg_time) < 3.0 then
        ui:TextColored( self.msg_clr or 0xFFFFFFFF, tostring(self.msg ))
    end
end
