local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

-------------------------------------------------------------------------------------
-- Helper panel for debugging quests.

local QLIST_BY_ID = 0
local QLIST_BY_TYPE = 1

local DebugQuests = class( "DebugQuests", DebugNode )
DebugQuests.MENU_BINDINGS = { debug_menus.QUEST_BINDINGS }
DebugQuests.PANEL_HEIGHT = 600
DebugQuests.PANEL_WIDTH = 600

function DebugQuests:init()
	
	self.spawn_quests = {}
	self.spawn_quest_params = {}
	self.sort_order = { QLIST_BY_ID, QLIST_BY_TYPE }
	self.when_options = MakeEnumArray( QWHEN )
	self.dbg_quest = nil
    self.deck_names = {"NONE"}
    for k, filename in pairs(engine.inst:ListFiles( 'scripts/saved_decks/*.lua')) do
        local root_name = filename:match( "^(.+)[.]lua$" )
        table.insert(self.deck_names, root_name)
    end
end

function DebugQuests:OnActivate()
	self:LoadQuestSettings()

end

function DebugQuests:LoadQuestSettings()
	local data = DebugUtil.GetLocalDebugData( "DEBUG_QUEST" )

	if data then
		self.dbg_quest = data.dbg_quest
		self.sort_order = data.sort_order or {}
		self:RefreshSpawnList( data.dbg_quest )
	end
end

function DebugQuests:SaveQuestSettings()
	local save_data = {}
	save_data.dbg_quest = self.dbg_quest
	save_data.sort_order = self.sort_order
	DebugUtil.SetLocalDebugData( "DEBUG_QUEST", save_data )
	DebugUtil.SaveLocalDebugData()
end

function DebugQuests:SortQuest( q1, q2 )
	for i = 1, #self.sort_order do
		local order = self.sort_order[i]
		if order == QLIST_BY_ID then
			if q1.id < q2.id then
				return true
			elseif q1.id > q2.id then
				return false
			end

		elseif order == QLIST_BY_TYPE then
			local at, bt = q1:GetType(), q2:GetType()
			if at < bt then
				return true
			elseif at > bt then
				return false
			end
		end
	end

	return tostring(q1) < tostring(q2)
end

function DebugQuests:RefreshSortOrder( t, sort_order )
	if sort_order then
    	table.arrayremove( self.sort_order, sort_order )
    	table.insert( self.sort_order, 1, sort_order )
		self:SaveQuestSettings()
    end
	table.sort( self.spawn_quests, function( q1, q2 ) return self:SortQuest( q1, q2 ) end )
end

function DebugQuests:RefreshSpawnList( filter_txt )
	table.clear( self.spawn_quests )
	for id, qdef in pairs( Content.GetAllQuests() ) do
		if filter_txt == nil or id:upper():find(filter_txt:upper(), 1, true) then
			table.insert( self.spawn_quests, qdef )
		end
	end
	self:RefreshSortOrder( self.spawn_quests )

	if #self.spawn_quests == 1 then
		self:SelectQuest(self.spawn_quests[1]:GetID())
	end
end

function DebugQuests:SelectQuest(id)
	self.spawn_quest_id = id
	self.param_lists = {}
	
	self.quest_fn_lists = {}
	
	self.game_fn_lists = {}


	local test_data = DebugUtil.GetLocalDebugData( "QUEST_TEST_PARAMS" ) or {}	
	self.quest_test_selections = test_data[self.spawn_quest_id] or {}
	
	--DebugUtil.SetLocalDebugData( "QUEST_TEST_PARAMS", test_data )


end

function DebugQuests:GenerateParams()
	
	local qdef = self.spawn_quest_id and Content.GetQuestDef( self.spawn_quest_id ) 
	if not qdef then
		return table.empty
	end

	local params = {}

	if qdef.debug_options then
		for k,v in ipairs(qdef.debug_options) do
			if v.test == nil or v.test(params) then
				params[v.name] = v.opt[ self.quest_test_selections[v.name] or 1 ]
			end
		end
	end
	return params
end


function DebugQuests:RenderSpawnQuestID( ui, panel )
	local qdef = self.spawn_quest_id and Content.GetQuestDef( self.spawn_quest_id ) 
	if not qdef then
		return
	end

	if type(qdef.icon) == "userdata" then
		ui:Image( qdef.icon, 24, 24 )
		ui:SameLine( nil, 10 )
	end
	panel:AppendTable( ui, qdef, qdef.id )

	ui:Spacing()
	ui:PushStyleColor( ui.Style_Button, 0x227722FF )

	if TheGame:GetGameState() then
		if ui:Button( "Spawn Quest" ) then
			self.quest_spawn, self.failed_quest_spawn = QuestUtil.SpawnQuest( self.spawn_quest_id, self.spawn_quest_params, true )
		end
	end

	if ui:Button( "Test Quest" ) then
		
		local deck_idx = self.quest_test_selections["DECK_OVERRIDE"] or 1
		local deck_name = deck_idx ~= 1 and self.deck_names[deck_idx]
		local deck
		if deck_name then
			local filename = "saved_decks/" .. deck_name
        	package[ filename ] = nil
        	deck = require(filename)			
		end

		local adv = self.quest_test_selections["ADVANCEMENT_LEVEL"] or 0
		qdef:IsolatedTest(self:GenerateParams(), self.spawn_quest_params.qrank, deck, adv)
	end

	if ui:Button( "RoboQuest" ) then

		qdef:RunTests(function() print ("TEST COMPLETE")  TheGame:Reset() end)
	end
	
	ui:PopStyleColor()
	
	ui:Separator()

    local qrank_min, qrank_max = qdef:GetRankRange()
    self.spawn_quest_params.qrank = ui:SliderInt( "Rank", self.spawn_quest_params.qrank or qrank_min, qrank_min, qrank_max )

	self.param_idx = self.param_idx or {}

	if qdef.debug_spawn_fn then
		qdef.debug_spawn_fn(self, qdef)
	end

    
    local dirty = false

	
    self.quest_test_selections["ADVANCEMENT_LEVEL"] = ui:SliderInt("Advancement Level", self.quest_test_selections["ADVANCEMENT_LEVEL"] or 0, 0, 11)
    local new_deck_idx = ui:Combo("Player deck override", self.quest_test_selections["DECK_OVERRIDE"] or 1, self.deck_names)
    if self.quest_test_selections["DECK_OVERRIDE"] ~= new_deck_idx then
    	self.quest_test_selections["DECK_OVERRIDE"] = new_deck_idx
    	dirty = true
    end


	if qdef.debug_options then
		local params = self:GenerateParams()
		
		for k,dat in ipairs(qdef.debug_options) do
			if dat.test == nil or dat.test(params) then
				local name = (dat.name or "").."###param"..tostring(k)
				local new_idx = ui:Combo(name, self.quest_test_selections[dat.name] or 1, dat.opt)
				if new_idx ~= self.quest_test_selections[dat.name] then
					self.quest_test_selections[dat.name] = new_idx
					dirty = true
				end
			end
		end
	end

	if dirty then
		local test_data = DebugUtil.GetLocalDebugData( "QUEST_TEST_PARAMS") or {}
		test_data[self.spawn_quest_id] = self.quest_test_selections
		DebugUtil.SetLocalDebugData( "QUEST_TEST_PARAMS", test_data )
		DebugUtil.SaveLocalDebugData()
	end

    ui:Separator()

	if self.quest_spawn then
	    ui:Spacing()
		ui:TextColored( 0x00CC00FF, "Spawned:" )
		ui:SameLine( nil, 10 )
		panel:AppendTable( ui, self.quest_spawn )

	elseif self.failed_quest_spawn then
	    ui:Spacing()
		ui:TextColored( 0xCC0000FF, "Failed spawn: " )
		ui:SameLine( nil, 10 )
		panel:AppendTable( ui, self.failed_quest_spawn )
	end

	ui:Separator()
end

function DebugQuests:RenderDebugSpawn( ui, panel )
	self:RenderSpawnQuestID( ui, panel )

	-- Save last quest selected, use on load.
	local filter_txt = ui:InputText( "Filter", self.dbg_quest or "" )
	if ui:BeginPopupContextItem( "FILTER" ) then
		if ui:MenuItem( "Clear" ) then
			filter_txt = ""
		end
		if ui:MenuItem( "Refresh List" ) then
			self:RefreshSpawnList( self.dbg_quest )
		end
		ui:EndPopup()
	end

	if filter_txt and filter_txt ~= self.dbg_quest then
		self:RefreshSpawnList( filter_txt )
		self.dbg_quest = filter_txt
		self:SaveQuestSettings()
	end


	ui:TextColored( 0x00CCCCFF, string.format( "%d quests shown", #self.spawn_quests ))
	ui:Columns( 2, "quests" )
--    ui:SetColumnOffset(1, 60)
    --ui:SetColumnOffset(2, 120)

    if ui:Selectable( "Type" ) then
    	self:RefreshSortOrder( self.spawn_quests, QLIST_BY_TYPE )
	end
    ui:NextColumn()
    if ui:Selectable( "ID" ) then
    	self:RefreshSortOrder( self.spawn_quests, QLIST_BY_ID )
	end
    ui:NextColumn() 

    ui:Separator()

	for i, qdef in ipairs( self.spawn_quests ) do
		ui:TextColored( 0xCC00CCFF, qdef:GetType() )
		ui:NextColumn()

		local id = qdef:GetID()
		if ui:Selectable( id, self.spawn_quest_id == id ) then
			self:SelectQuest(id)
		end
		if ui:BeginPopupContextItem( id ) then
			ui:TextColored( 0x777777FF, id )
			ui:Separator()

			if ui:MenuItem( "View QuestDef" ) then
				panel:PushNode( qdef )
			end

			ui:EndPopup()
		end
		ui:NextColumn()
	end
	ui:Columns( 1, "reset" )
end

local function SortQuest( q1, q2 )
	if q1:IsActive() and not q2:IsActive() then
		return true
	elseif not q1:IsActive() and q2:IsActive() then
		return false
	else
		return q1:GetContentID() < q2:GetContentID() -- lexographical by content ID.
	end
end

function DebugQuests:RenderPanel( ui, panel )
	if ui:CollapsingHeader( "Spawn", ui.TreeFlags_DefaultOpen ) then
		self:RenderDebugSpawn( ui, panel )
	end

	local gs = TheGame:GetGameState()
	if gs and panel:CollapsingHeader( ui, string.format("%d Spawned Quests", table.count( gs.available_quests ))) then
		local obj_ids = ObtainWorkTable()
		ui:Columns( 2, "debug spawn" )

	    ui:Text( "Quest" )
	    ui:NextColumn()
	    ui:Text( "Status" )
	    ui:NextColumn()
	    ui:Separator()

		for guid, quest in pairs( gs.available_quests ) do
			-- Quest Table
			panel:AppendTable( ui, quest )
			if ui:IsItemHovered() then
				for i, obj in quest:GetQuestDef():Objectives() do
					table.insert( obj_ids, obj.id )
				end
				ui:SetTooltip( table.concat( obj_ids, ", " ))
			end
			ui:NextColumn()

			-- Active & tracked status
			if quest:GetStatus() == QSTATUS.ACTIVE then
				ui:TextColored( 0x00CC00FF, quest:GetStatus() )
			else
				ui:Text( quest:GetStatus() )
			end
			ui:NextColumn()
		end
		ReleaseWorkTable( obj_ids )
		ui:Columns( 1, "reset2" )
	end
end

--------------------------------------------------------------------
-- A debug source for Quests.

local DebugQuest = class( "DebugQuest", DebugNode )
DebugQuest.REGISTERED_CLASS = QuestState

function DebugQuest:init( quest )
	self.t = quest
	self.menu_params = { quest }
end

function DebugQuest:GetName()
	return self.t.content_id
end

function DebugQuest:GetUID()
	return rawstring(self.t)
end

local debug_statuses = 
{
	QSTATUS.ACTIVE,
	QSTATUS.INACTIVE,
	QSTATUS.COMPLETE,
	QSTATUS.FAILED,
	QSTATUS.CANCELLED,
}

function DebugQuest:RenderPanel( ui, panel )
	local quest = self.t
	local qdef = quest:GetQuestDef()

	panel:AppendTable( ui, qdef, "QuestDef" )

	ui:Spacing()
	local status = quest:GetStatus()
	ui:Text( loc.format( "{1} (Rank {2})", quest:GetTitle(), quest:GetRank() ))
	ui:TextColored(QuestUtil.GetStatusColour(status), status)

	if not quest:IsDone() then
		ui:SameLine( nil, 10 )
		if ui:Button("Complete") then
			quest:Complete()
		end
		ui:SameLine( nil, 10 )
		if ui:Button("Fail") then
			quest:Fail()
		end
		ui:SameLine( nil, 10 )
		if ui:Button("Cancel") then
			quest:Cancel()
		end
	end

	if quest.is_hidden then
		ui:TextColored( 0xCC0000FF, "Hidden" )
	end


	if panel:CollapsingHeader( ui, "CAST" ) then
		ui:Columns( 2, "cast" )
		ui:SetColumnOffset( 1, 100 )
		for i, cast_def in qdef:CastDefs() do
			local cast_id = cast_def.cast_id
			panel:AppendTable( ui, cast_def, cast_id )
			ui:NextColumn()
			
			local cast = quest:GetCastMember( cast_id )
			if cast == nil then
				ui:TextColored( 0x777777FF, "uncasted" )
			else
				panel:AppendTable( ui, cast )
			end
			ui:NextColumn()
		end
		ui:Columns( 1, "reset3" )
	end


	if quest.param and next(quest.param) ~= nil and panel:CollapsingHeader( ui, "PARAMS" ) then
		panel:AppendKeyValues( ui, quest.param )
	end

	if qdef.objectives and panel:CollapsingHeader( ui, "OBJECTIVES" ) then
		for i, obj in ipairs( qdef.objectives ) do
			local state = quest:GetObjectiveState(obj.id)
			local lbl = string.format( "%s (%s)", obj.id, state)

			ui:PushStyleColor( ui.Style_Text, QuestUtil.GetStatusColour(state))
			if ui:TreeNode( lbl, 0, "obj_"..tostring(i) ) then -- ui.TreeFlags_DefaultOpen
				ui:PopStyleColor()

				local title = quest:FormatString( obj.title or "NO TITLE" )
				ui:TextColored(0xFFD700FF, title)

				local debug_status_idx = table.arrayfind(debug_statuses, state)
				local new_debug_status_idx = ui:Combo( "Objective Status##state_"..tostring(i), debug_status_idx, debug_statuses )
				if new_debug_status_idx ~= debug_status_idx then
					quest:SetObjectiveState(obj.id, debug_statuses[new_debug_status_idx])
				end

				local quest_obj = quest:FindObjective(obj.id)
				if quest_obj then
					for k, v in pairs(quest_obj) do
						if k ~= "id" and k ~= "state" then
							ui:Text( string.format( "\t%s = %s\n", tostring(k), tostring(v) ))
						end
					end
				end

				ui:TreePop()
			else
				ui:PopStyleColor()
			end
		end
	end

	if ui:CollapsingHeader( "Find", ui.TreeFlags_DefaultOpen ) then
		if quest.trace and ui:CollapsingHeader( "TRACE" ) then
	        local filter = ui:InputText( "Filter", self.trace_filter )
	        if filter and filter ~= self.trace_filter then
	            self.trace_filter = filter:lower()
	        end
	        for i = 1, #quest.trace do
	            local str = quest.trace[i]
	            if self.trace_filter == nil or str:lower():find( self.trace_filter, 1, true ) then
	                ui:Text( str )
	            end
	        end
		end
	end
end

