local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------
-- A debug source for Agents.

local DebugTabular = class( "DebugTabular", DebugNode )
DebugTabular.PANEL_WIDTH = 1000

function DebugTabular:init(cols, data, filter)
    assert( #cols > 0 ) -- Empty columns.
    self.cols = cols
    self.data = data
    self.filter = filter
    self:Refresh(1)
end


function DebugTabular:Refresh(sort_key_idx)
    if sort_key_idx and sort_key_idx == self.sort_key_idx and self.resolved_data then
        table.reverse(self.resolved_data)        
        return
    end


    self.resolved_data = {}
    for _,datum in pairs(self.data) do
        if not self.filter or self.filter(datum) then
            local res_d = {}
            for _,col in ipairs(self.cols) do
                local val, str
                if col.fn then
                    val, str = col.fn(datum)
                    if type(str) == "string" then
                        -- str = str:gsub( "\n", "\\n" )
                    end
                elseif col.key then
                    val = datum[col.key]
                    if val ~= nil then
                        str = tostring(val)
                    end
                else
                    val = datum
                end

                table.insert(res_d, {val=val, str=str})
            end
            table.insert(self.resolved_data, res_d)
        end
    end
    
    if sort_key_idx then
        self.sort_key_idx = sort_key_idx
    end

    table.sort(self.resolved_data, function(dat_a,dat_b) 
            local a = dat_a[ self.sort_key_idx].val
            local b = dat_b[ self.sort_key_idx].val
            
            if type(a) == type(b) and type(a) == "number" then
                return a < b
            else
                return (dat_a[ self.sort_key_idx].str or "") < (dat_b[ self.sort_key_idx].str or "")
            end
        end)
end

function DebugTabular:GetString( delimiter )
    delimiter = delimiter or ","
    local t = {}
    local header = {}
    for k,col in ipairs(self.cols) do
        table.insert(header, col.name or tostring(k))
    end
    table.insert(t, table.concat(header, delimiter))

    for k,linedat in ipairs(self.resolved_data) do
        local line = {}

        for k,val in ipairs(linedat) do
            local str = tostring(val.str or "")
            -- Single quotes will be removed when importing a .csv, they must be doubled-up.
            str = str:gsub( [["]], [[""]] )
            -- Cells with commas, quotes, or new-lines imbedded must be quote-enclosed to preserve columns.
            if str:find(",") or str:find("\"") or str:find( "\n" ) then
                str = [["]]..str..[["]]
            end
            table.insert(line, str)
        end

        table.insert(t, table.concat(line, delimiter))
    end
    local str =table.concat(t, "\n")
    return str
end

function DebugTabular:SaveToCSV(filepath, delimiter)
    if BRANCH == "dev" then
        return engine.data:SaveFile( "../"..filepath, self:GetString(delimiter))
    else
        return engine.settings:SaveFile( filepath, self:GetString(delimiter))
    end
end

function DebugTabular:RenderPanel( ui, panel )

    if ui:Button("Copy to clipboard") then
        engine.inst:SetClipboardText(self:GetString())
    end
    if ui:Button("Save output.csv") then
        local filename = "../output.csv"
        self:SaveToCSV( filename )
        self.save_msg = string.format( "Saved to '%s'", tostring(filename))
    end
    if self.save_msg then
        ui:SameLine( nil, 10 )
        ui:TextColored( 0x55CC55FF, self.save_msg )
    end

    if ui:Button("Refresh") then
        self:Refresh()
    end
    ui:Columns( #self.cols )
    for k,col in ipairs(self.cols) do
        if ui:Selectable(col.name or tostring(k)) then
            self:Refresh(k)
            return
        end
        ui:NextColumn()
    end
    
    ui:Separator()
    for _, data in ipairs(self.resolved_data) do
        for k, val in ipairs(data) do
            if val.str == nil then
                --
            elseif type(val.val) == "table" then
                panel:AppendTable(ui, val.val, val.str)
            else
                ui:Text(tostring(val.str))
            end
            ui:NextColumn()
        end
    end
end