local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

--------------------------------------------------------------------
-- A debug source for UI widgets


local function RenderTree(self, ui, panel, widget, child)
    if not child then
        self.widget_id = 0
    else
        self.widget_id = self.widget_id+1
    end
    local flags = 0

    -- if widget._class == Widget or widget.is_screen then
    --     flags = flags | ui.TreeFlags_DefaultOpen
    -- end

    if widget.children == nil or #widget.children == 0 then
        flags = flags | ui.TreeFlags_Leaf
    end

    if ui:TreeNode( tostring(widget) .. "##" .. tostring(self.widget_id), flags ) then
        ui:SameLine( nil, 10 )
        panel:AppendTable(ui, widget, rawstring(widget))
        
        if widget.hover then
            ui:SameLine() ui:TextColored(0xffff00ff, "  HOVER")
        end

        if widget.children then
            for _, child in ipairs(widget.children) do
                RenderTree(self, ui, panel, child, true)
            end
        end
        ui:TreePop()
    else
        ui:SameLine( nil, 10 )
        panel:AppendTable(ui, widget, rawstring(widget) )
        if widget.hover then
            ui:SameLine() ui:TextColored(0xffff00ff, "  HOVER")
        end
    end
end


local DebugWidget = class( "DebugWidget", DebugNode )
DebugWidget.REGISTERED_CLASS = Widget
DebugWidget.MENU_BINDINGS =  { debug_menus.WIDGET_BINDINGS }

function DebugWidget:init( widget )
    assert( is_instance( widget, Widget ))
    self.widget = widget
    self.menu_params = { widget }
end

function DebugWidget:GetName()
    return tostring(self.widget)
end

function DebugWidget:GetUID()
    return rawstring(self.widget)
end

function DebugWidget:RenderPanel( ui, panel )
    if self.widget.RenderDebugUI == nil or self.widget:RenderDebugUI( ui, panel ) then
        if self.widget.removed then
            ui:TextColored( 0xFF0000FF, "REMOVED" )
            ui:Spacing()
        end

        if self.widget.parent then
            ui:Text( "Parent:" )
            ui:SameLine( nil, 10 )
            panel:AppendTable(ui, self.widget.parent )
        end

        if ui:TreeNode( "Hierarchy" ) then
            RenderTree(self, ui, panel, self.widget)
            ui:TreePop()
        end

        if ui:TreeNode( "Properties" ) then
            local current_tint = MakeHexColour(self.widget.tint_r or 1, self.widget.tint_g or 1, self.widget.tint_b or 1, self.widget.tint_a or 1)
            local tint = ui:ColorEdit4( "Tint", current_tint )
            if current_tint ~= tint then
                self.widget:SetTintColour(tint)
            end
            
            local sat = ui:SliderFloat( "Sat", self.widget.saturation or 1, 0, 3 )
            if sat ~= (self.widget.saturation or 1) then
                self.widget:SetSaturation(sat)
            end

            local ox, oy = self.widget:GetPos()
            local x, y = ui:DragFloat3( "Position", ox, oy, 0, -2000, 2000, 1)
            if x then
                self.widget:SetPos(x,y)
            end

            local osx, osy = self.widget:GetScale()
            local sx, sy = ui:DragFloat3( "Scale", osx, osy, 1, 0, 10, .1)
            if sx then
                self.widget:SetScale(sx,sy)
            end

            ui:TreePop()
        end
        ui:Separator()

        panel:AppendKeyValues( ui, self.widget )
    end
end

function DebugWidget:RenderContextMenu( ui, panel )
    if self.widget.RenderContextMenu then
        self.widget:RenderContextMenu( ui, panel )
    end
end

-------------------------------------------------------------------------------------
-- Helper panel for debugging widgets and the FE.

local DebugUIDebug = class( "DebugUIDebug", DebugNode )
DebugUIDebug.REGISTERED_CLASS = FrontEnd

function DebugUIDebug:RenderPanel( ui, panel )
    local fe = TheGame:FE()
    if panel:CollapsingHeader( ui, "FE" ) then
        if ui:Checkbox( "UI Freeze", fe.debugfreeze ) then
            fe:ToggleDebugFreeze()
        end

        ui:Columns( 2 )
        do
            ui:Text( "Debug Root") ui:NextColumn()
            panel:AppendTable( ui, fe.debugroot ) ui:NextColumn()

            ui:Text( "Focus Widget") ui:NextColumn()
            panel:AppendTable( ui, fe:GetFocusWidget() ) ui:NextColumn()

            ui:Text( "Hover Widget") ui:NextColumn()
            panel:AppendTable( ui, fe:GetHoverWidget() ) ui:NextColumn()
        end
        ui:Columns( 1 )

        ui:TextColored( 0xFF00FFFF, string.format( "Mouse: %d, %d", TheGame:GetInput():GetMousePos() ))

        ui:Text( "Tooltip:" )
        ui:SameLine( nil, 10 )

        local tt = fe.tooltip_widget
        if tt then
            local x, y = tt:GetPos()
            local x1, y1, x2, y2 = tt:GetBoundingBox()
            x1, y1 = tt:TransformToWorld( x1, y1 )
            x2, y2 = tt:TransformToWorld( x2, y2 )
            panel:AppendTable( ui, tt )
            ui:Text( "Focus:" )
            ui:SameLine( nil, 10 )
            panel:AppendTable( ui, fe.tooltip_focus )
            ui:SameLine()
            ui:Text( string.format( "Pos: <%.1f, %.1f>, Bounds: BL<%.1f, %.1f> TR<%.1f, %.1f>", x, y, x1, y1, x2, y2))
        else
            ui:Text( "none" )
        end
    end

    
    if panel:CollapsingHeader( ui, "SCREENS" ) then
        for i = #fe.screenstack, 1, -1 do
            local screen = fe.screenstack[i]
            panel:AppendTable( ui, screen, screen.name )
            if screen:IsPopup() then
                ui:SameLine( nil, 10 )
                ui:TextColored( 0xCCCCCCFF, "(Popup)" )
            elseif screen:IsOverlay() then
                ui:SameLine( nil, 10 )
                ui:TextColored( 0xCCCCCCFF, "(Overlay)" )
            end

            if screen == fe:GetTopScreen() then
                ui:SameLine( nil, 10 )
                ui:Text( "(Topmost)" )
            end


            RenderTree(self, ui, panel, screen)
        end
    end

    if panel:CollapsingHeader( ui, "FE" ) then
        panel:AppendKeyValues( ui, fe )
    end
end


