local negotiation_defs = require "negotiation/negotiation_defs"

-------------------------------------------------------------------------------
-- Representation of a conversation option

local ConvoOption = class( "ConvoOption" )
ConvoOption.can_reload = true

function ConvoOption:init( text )
    self.text = text
    self.handlers = {}
end

function ConvoOption:AttachHub( hub )
    assert( is_instance( hub, BaseHub ))
    assert( self.hub == nil )
    self.hub = hub
end

function ConvoOption:SetHoverLinkParameters( ... )
    self.hover_link_params = {...}
end

function ConvoOption:SetQuestMark( quest )
    quest = quest or self.hub.quest
    if quest then
        local icon = quest:GetMarkData()
        self:PreIcon( icon )
        self.quest = quest
    end
    return self
end

function ConvoOption:Robot(tag)
    if type(tag) == "string" then
        self.robot_tags = {tag}
    elseif type(tag) == "table" then
        self.robot_tags = shallowcopy(tag)
    end
    return self
end

function ConvoOption:MakeUnder()
    self.underfooter = true
    return self
end

function ConvoOption:SetSFX( sfx )
    if sfx then
        self.data = self.data or {}
        self.data.sfx = sfx
    end
    return self
end

function ConvoOption:SetPortrait( agent )
    if self.option_widget then
        self.option_widget:SetPortrait(agent)
    else
        self.portrait = agent
    end
    return self
end

function ConvoOption:DeltaHealth(delta, ignore_full)

    local current_health, max_health = TheGame:GetGameState():GetPlayerAgent():GetHealth()

    delta = math.floor(delta)

    self:PushHandler(function() 
        ConvoUtil.DoHealthDelta(self.hub, delta)
    end)

    if delta > 0 then
        self:PostIcon( global_images.heal )
        self:PostText("TT_HEALING", delta )
        if not ignore_full then self:ReqCondition( current_health < max_health, "REQ_AT_FULL_HEALTH" ) end
    else
        --self:PostIcon( global_images.hurt )
        self:PostText("TT_HURTING", math.abs(delta) )
        self:ReqCondition( current_health + delta > 0, "REQ_NOT_ENOUGH_HEALTH" )
    end
    return self
end

function ConvoOption:DeltaResolve(delta)

    local cur, max = TheGame:GetGameState():GetCaravan():GetResolve()

    delta = math.floor(delta)
    
    self:PushHandler(function() 
        ConvoUtil.DoResolveDelta(self.hub, delta)
    end)

    if delta > 0 then
        self:PostIcon( global_images.restoreresolve )
        self:PostText("TT_GAIN_RESOLVE", delta )
        -- self:ReqCondition( cur < max, "REQ_AT_FULL_RESOLVE" )
    else
        --self:PostIcon( global_images.hurt )
        self:PostText("TT_LOSE_RESOLVE", math.abs(delta) )
    end
    return self
end

function ConvoOption:GetLocString( txt, ... )
    if self.hub then
        txt = self.hub:GetLocString( txt, ... )
    end
    return txt
end

function ConvoOption:PostText( txt, ... )
    if txt and #txt > 0 then
        if self.post_text == nil then
            self.post_text = {}
        end
        txt = self:GetLocString( txt, ... )
        table.insert( self.post_text, txt )
    end
    return self
end

function ConvoOption:FadeOut()
    self:PushHandler(function()
        self.hub:FadeOut()
    end)
    return self
end

function ConvoOption:FadeIn()
    self:PushHandler(function()
        self.hub:FadeIn()
    end)
    return self
end

function ConvoOption:PostCard( card_id, no_keywords )
    if no_keywords ~= nil then
        self.no_keywords = no_keywords
    end
    if self.post_cards == nil then
        self.post_cards = {}
    end
    table.insert_unique( self.post_cards, card_id )
    return self
end

function ConvoOption:PreIcon( icon, colour )
    if self.option_widget then
        self.option_widget:PostIcon( icon, colour )
    else
        self.preicon = icon
        self.colour = colour
    end
    return self
end

function ConvoOption:PostIcon( icon, colour )
    if self.option_widget then
        self.option_widget:PostIcon( icon, colour )
    else
        self.posticon = self.posticon or {}
        self.colour = colour
        table.insert(self.posticon,icon)
    end
    return self
end

function ConvoOption:CreateNegotiationPreview( option_widget, negotiation_params )
    option_widget:SetColour( UICOLOURS.NEGOTIATION )

    local function OnGainHover()
        if not TheGame:FE():IsGamepadMode() then
            self.hub.encounter:GetScreen():PresentNegotiationPreviewPanel( negotiation_params, function() multi:OnSelectOption( idx ) end )
        end
    end

    local function OnLoseHover()
        self.hub.encounter:GetScreen():NegotiationPreviewPanelLoseHover()
    end

    option_widget.OnGainHover = function()
        Widget.IconButton.OnGainHover( option_widget )
        OnGainHover()
    end
    option_widget.OnLoseHover = function()
        Widget.IconButton.OnLoseHover( option_widget )
        OnLoseHover()
    end
    option_widget.OnGainFocus = function()
        Widget.IconButton.OnGainFocus( option_widget )
        OnGainHover()
    end
    option_widget.OnLoseFocus = function()
        Widget.IconButton.OnLoseFocus( option_widget )
        OnLoseHover()
    end
end

function ConvoOption:CreateBattlePreview( option_widget, battle_scenario )
    option_widget:SetColour( UICOLOURS.FIGHT )

    local function OnGainHover()
        if not TheGame:FE():IsGamepadMode() then
            self.hub.encounter:GetScreen():PresentBattlePreviewPanel( battle_scenario, function() error() multi:OnSelectOption( idx ) end )
        end
    end

    local function OnLoseHover()
        self.hub.encounter:GetScreen():BattlePreviewPanelLoseHover()
    end

    option_widget.OnGainHover = function()
        Widget.IconButton.OnGainHover( option_widget )
        OnGainHover()
    end
    option_widget.OnLoseHover = function()
        Widget.IconButton.OnLoseHover( option_widget )
        OnLoseHover()
    end
    option_widget.OnGainFocus = function()
        Widget.IconButton.OnGainFocus( option_widget )
        OnGainHover()
    end
    option_widget.OnLoseFocus = function()
        Widget.IconButton.OnLoseFocus( option_widget )
        OnLoseHover()
    end
end

function ConvoOption:CreateWidget( enc, screen, multi, idx )
    local option_widget
    local text = self.text
    text = Content.SubstituteCodexFormatting( text )
    

    if self.hidden then
        -- Not shown.
        return

    elseif self.underfooter and multi.AddUnderfooterOption then
        option_widget = multi:AddUnderfooterOption( text, self.data )
    else
        option_widget = multi:AddOption( text, self.data )

        if self.hover_link_params then
            option_widget:SetHoverLinkParameters( table.unpack( self.hover_link_params ))
        end

        if self.preicon then
            option_widget:PreIcon(self.preicon, self.colour)
        end

        if self.posticon then
            for k,v in ipairs(self.posticon) do
                option_widget:PostIcon(v)
            end
        end

        if self.portrait then
            option_widget:SetPortrait(self.portrait)
        end

        if self.hub_option then
            option_widget:PostIcon( engine.asset.Texture("UI/characterselect_deck_arrow.tex"), 14, 14 )
        end

        if self.robot_tags then
            option_widget.robot_tags = self.robot_tags
        end

    end
    
    if self.negotiation_params then
        self:CreateNegotiationPreview( option_widget, self.negotiation_params )

    elseif self.battle_scenario then
        self:CreateBattlePreview( option_widget, self.battle_scenario )
        
        if self.battle_scenario:IsAggressive() then
            self:PostText(loc.format( LOC"UI.AGGRESSIVE_ACTION", "aggressive"))
            self:PostCard("aggressive", true)
        end
    end

    option_widget:SetEnabled( not self.disabled )
    self.option_widget = option_widget

    self:ApplyTooltip(option_widget)

    return option_widget
end

function ConvoOption:ApplyTooltip(button)
    if self.disabled_reasons or
        (self.post_text and #self.post_text > 0) or
        self.relationship_changes ~= nil or
        self.negotiation_params or
        self.quest ~= nil then
        --
        button:SetToolTip( self )
        button:SetToolTipClass( Widget.TooltipConvoOption )
    else
        button:SetToolTip( nil )
    end

end

function ConvoOption:ShowQuestAsInactive()
    self.show_quest_as_inactive = true
    return self
end

function ConvoOption:RunOption()
    if self.hub then
        self.hub.enc:PushHub( self.hub )
    end
    self:RunHandlers()
    if self.hub then
        self.hub.enc:PopHub()
    end
end

function ConvoOption:RunHandlers(...)
    for i, handler in ipairs( self.handlers ) do
        local args = {}
        for k,v in ipairs(handler) do
            if k > 1 then
                table.insert(args, v)
            end
        end
        for k, v in ipairs{...} do
          table.insert(args, v)
        end

        handler[1]( table.unpack( args ) )
    end
end

--[[function ConvoOption:RunHandlers(...)
    for i, handler in ipairs( self.handlers ) do
        print (table.unpack( handler, 2 ) )
        
        handler[1](table.unpack( handler, 2 ))
    end
end--]]

function ConvoOption:ClearHandlers()
    table.clear( self.handlers )
end

function ConvoOption:PushHandler( fn, ... )
    assert( fn )
    table.insert( self.handlers, { fn, ... } )
end

function ConvoOption:Quip(agent, ...)
    self:PushHandler( self.hub.Quip, self.hub, agent, ...)
    return self
end

function ConvoOption:Fn(fn, ...)
    self:PushHandler( self.hub.Run, self.hub, fn, ...)
    return self
end

function ConvoOption:LoopingFn(fn, ...)
    self:PushHandler( self.hub.RunLoopingFn, self.hub, fn, ...)
    return self
end

function ConvoOption:GoToQuest(quest, state_id, ...)
    self:PushHandler( self.hub.GoToQuest, self.hub, quest, state_id, ...)
    return self
end

function ConvoOption:GoToExt(convo_id, state_id, ...)
    self:PushHandler( self.hub.GoToExt, self.hub, convo_id, state_id, ...)
    return self
end

function ConvoOption:GoTo(state_id, ...)
    self:PushHandler( self.hub.GoTo, self.hub, state_id, ...)
    return self
end

function ConvoOption:Dialog(id, ...)
    self:PushHandler( self.hub.Dialog, self.hub, id, ...)
    return self
end

function ConvoOption:Pop()
    self:PushHandler( self.hub.Pop, self.hub )
    return self
end

function ConvoOption:PopConvo()
    self:PushHandler(self.hub.PopConvo, self.hub )
    return self
end

function ConvoOption:IsDisabled()
    return self.disabled == true
end

function ConvoOption:ReqEnabled()
    if self.disabled then
        self.hub.enc:RemoveOption( self )
    end
    return self
end


function ConvoOption:DoneConvo()
    self:Fn(function() 
        StateGraphUtil.AddEndOption(self.hub)
    end)

    return self
end

function ConvoOption:End()
    self:PushHandler(function() 
        self.hub:End()
    end)
    return self
end

function ConvoOption:AddScore(id, val)
    
    self:PushHandler(function()
        if TheGame:GetGameState():GetMainQuest() then
            TheGame:GetGameState():GetMainQuest():AddScore(id, val)
        end
    end)
    
    local score_def = TheGame:GetGameState():GetMainQuest() and TheGame:GetGameState():GetMainQuest():GetQuestDef():GetScoreDef(id)
    if score_def == nil then
       score_def = Content.GetScoreComponent(id)
    end

    assert_warning("missing score component '%s'", id )
    local raw_points = (score_def and score_def.value or 0)

    self:PostText( "{1#points}", raw_points)

    return self
end
     
function ConvoOption:GainCard(card_id, fn)
    self:PushHandler(function()
        local cards = self.hub:GainCards{card_id}
        if fn then
            for k,v in ipairs(cards) do
                fn(v)
            end
        end

    end)
    return self
end

function ConvoOption:GainCards(cards, fn)
    self:PushHandler(function()
        local cards = self.hub:GainCards(cards)
        if fn then
            for k,v in ipairs(cards) do
                fn(v)
            end
        end
    end)
    return self
end

function ConvoOption:Travel()
    self:Fn(function() 
        StateGraphUtil.AddLeaveLocation(self.hub)
    end)
    return self
end

function ConvoOption:ReqCondition( ok, reason, ... )
    self.disabled = self.disabled or (not ok)
    if not ok and reason then
        local text = self.hub and self.hub.GetLocString and self.hub:GetLocString( reason, ... ) or reason
        if self.disabled_reasons == nil then
            self.disabled_reasons = {}
        end
        table.insert( self.disabled_reasons, text )
    end
    return self
end

function ConvoOption:IsHubOption( bool )
    self.hub_option = bool
    return self
end


function ConvoOption:DeliverMoney( amount, params, agent )

    local _, newamount = self:RequireMoney( amount, params, agent or self.hub:GetAgent() )
    if newamount then
        self:PushHandler( function()
            self.hub.caravan:PayMoney( newamount, agent or self.hub:GetAgent(), params and params.no_scale )
        end )
    end

    return self, newamount
end

function ConvoOption:RequireMoney( amount, params, agent )

    --get discount / surcharge
    local amount, reasons = CalculatePayment(agent or self.hub:GetAgent(), amount, 1, params)
    if amount <= 0 then
        return
    end

    self:PostIcon( global_images.shills )
    
    if amount <= TheGame:GetGameState():GetCaravan():GetMoney() then
        self:PostText( "DELIVER_MONEY", amount )
        self:PostText( table.concat( reasons, "\n" ))
    else
        self:ReqCondition( false, "REQUIRES_MONEY", amount )
        self:PostText( table.concat( reasons, "\n" ))
    end

    return self, amount
end

function ConvoOption:RequireResolve( amount, params )

    self:PostIcon( global_images.resolve )
    
    if amount <= TheGame:GetGameState():GetCaravan():GetResolve() then
        --self:PostText( "LOSE_RESOLVE", amount )
    else
        self:ReqCondition( false, "REQUIRES_RESOLVE", amount )
    end

    return self, amount
end

function ConvoOption:ReqTimer(timer_name, time, agent, set_on_select)
    agent = agent or self.hub:GetAgent()
    local mem, duration = agent:HasMemory(timer_name, time)
    self:ReqCondition( not mem, "ON_COOLDOWN", duration or 1 )

    if set_on_select then
        self:PushHandler( function()
            agent:Remember(timer_name)
        end )
    end

    return self
end

function ConvoOption:ReqRelationship(min_rel, agent)
    agent = agent or self.hub:GetAgent()
    self:ReqCondition( agent:GetRelationship() >= min_rel, "REQUIRES_RELATIONSHIP", GetRelationshipString(min_rel) )
    return self
end
    
function ConvoOption:RecruitMember( member_type, agent )
    agent = agent or self.hub:GetAgent()

    self:PushHandler(function()
        agent:Recruit( member_type )
    end)

    if member_type == PARTY_MEMBER_TYPE.CAPTIVE then
        self:ReceiveOpinion( OPINION.TOOK_CAPTIVE )
    end

    return self
end

function ConvoOption:ReceiveMoney( amount )
    self:PostText( "RECEIVE_MONEY", amount )

    self:PushHandler( function()
        AUDIO:PlayEvent(SoundEvents.conversation_accept_money)
        self.hub.encounter:GainMoney( amount )
    end )
    return self
end


function ConvoOption:ReceiveOpinion( opinion_data, params, agent )
    if type(opinion_data) == "string" then
        opinion_data = self.hub.quest and self.hub.quest:GetQuestDef():GetOpinionEvent(opinion_data) or OPINION[opinion_data]
    end

    if opinion_data == nil then
        LOGWARN( "Missing opinion_data" )
        return
    end

    
    agent = agent or self.hub:GetAgent()
    if type(agent) == "string" then
        agent = self.hub.enc:GetCastAgent(agent)
    end

    if agent and not agent:IsDead() then
        if self.relationship_changes == nil then
            self.relationship_changes = {}
        end
        local found = false
        for i, t in ipairs( self.relationship_changes ) do
            if t.agent == agent then
                -- Modify relationship.
                t.relationship = opinion_data.delta.relationship_delta[ t.relationship ] or t.relationship
                found = true
            end
        end
        if not found then
            local relationship = agent:GetRelationship()
            local new_relationship = opinion_data.delta.relationship_delta[ relationship ] or relationship
            table.insert( self.relationship_changes, { agent = agent, relationship = new_relationship })
        end

        if params and params.only_show then
            -- just for show
        else
            self:PushHandler( function() 
                self.hub.enc:WaitOnLine()
                agent:OpinionEvent(opinion_data, params)
                self.hub.enc:ShowRelationshipChanges()    
            end)
        end
    end

    return self
end

function ConvoOption:Battle(params)
    params = params or {}
    params.is_battle = true

    if params.on_runaway == nil then
        if self.hub.quest and ( self.hub.quest:GetType() ~= QTYPE.STORY ) then
            params.on_runaway = StateGraphUtil.DoRunAway
        end
    end

    self.battle_scenario = self.hub.encounter:CreateScenario( params, self.hub.quest )

    if params.on_runaway == nil then
        self.battle_scenario:SetFlags( BATTLE_FLAGS.NO_FLEE )
    end


    self:PreIcon( global_images.combat, UICOLOURS.FIGHT )
    if not self.battle_scenario:IsSelfDefence() then
        local primary = self.battle_scenario:GetPrimary( TEAM.RED )
        if primary and primary:IsSentient() and primary:GetSocialConnections() then
            self:ReceiveOpinion( OPINION.ATTACKED, nil, primary )
        end
    end

    self.data = self.data or {}
    self.data.sfx = "event:/ui/main/gen/pre_battle_click"
    self.data.preclick_fn = function( option )
        local screen = self.hub.encounter:GetScreen()
        if not screen:IsBattlePreviewPanelShowing() then
            screen:PresentBattlePreviewPanel( self.battle_scenario )
            return false
        end
        return true
    end
    
    self:Fn(function()
        self.hub:DoBattle(params)   
    end)

    self.last_param = params
    return self
end    
    

function ConvoOption:Negotiation( params )
    params = params or {}
    params.is_negotiation = true
    if self.hub.caravan:GetResolve() <= 0 then
        if params.flags and CheckBits( params.flags, NEGOTIATION_FLAGS.NO_AUTOFAIL ) then
            self:ReqCondition( false, "NO_RESOLVE_DISABLED" )
            self.last_param = params
            return self
        end
    end

    local agent = params.target_agent or self.hub:GetAgent()
    if agent == nil then
        self:ReqCondition( false )
        self.last_param = params
        return self
    end
    
    params.reason = self.hub:GetLocString(params.reason or (params.flags and CheckBits( params.flags, NEGOTIATION_FLAGS.INTIMIDATION ) and "DEFAULT_NEGOTIATION_REASON_INTIMIDATE" or "DEFAULT_NEGOTIATION_REASON"), agent)

    -- Negotiation difficulty, unless overridden, is determined by quest rank.
    if params and params.difficulty == nil then
        if self.hub.quest then
            params.difficulty = self.hub.quest:GetRank()
        else
            params.difficulty = TheGame:GetGameState():GetCurrentBaseDifficulty()
        end
    end

    self.negotiation_params = Negotiation.MiniGame.ParamsFromParty( self.hub.caravan, agent, params, self.hub.encounter )

    if self.hub.caravan:GetResolve() <= 0 then
        self:PostText( "NO_RESOLVE" )
        self:PreIcon( global_images.warning, UICOLOURS.NEGOTIATION )
    else
        self:PreIcon( global_images.negotiation, UICOLOURS.NEGOTIATION )
    end
    self.data = self.data or {}
    self.data.sfx = "event:/ui/main/gen/pre_negotiation_click"
    self.data.preclick_fn = function( option )
        local screen = self.hub.encounter:GetScreen()
        if not screen:IsNegotiationPreviewPanelShowing() then
            screen:PresentNegotiationPreviewPanel( self.negotiation_params )
            return false
        end
        return true
    end

    local timer_name = self.hub:GetStateID()..(params.timer_id or self.loc_id or self.text or "")
    local cooldown = params.cooldown or 1
    if cooldown > 0 then
        local mem, duration = agent:HasMemoryOfNegotiation(timer_name, cooldown)
        self:ReqCondition( not mem, "ON_COOLDOWN", cooldown )
    end

    --DBG(params.resolve_modifiers)
    if self.negotiation_params.resolve_modifiers then
        for _, modifier in ipairs(self.negotiation_params.resolve_modifiers) do
            if modifier.value then
                local txt = self.hub.encounter:LocFormat( modifier.text )
                if modifier.value > 0 then

                    self:PostText( "SITUATION_MODIFIER_POSITIVE", modifier.value, txt )
                else
                    self:PostText( "SITUATION_MODIFIER_NEGATIVE", modifier.value, txt )
                end
            end
        end
    end

    self:Fn(function(cxt)
        local result = cxt:DoNegotiation( params )
        if result ~= negotiation_defs.RESULT.CANCELLED then
            agent:RememberNegotiation( timer_name )
        end
    end )

    if self.negotiation_params.flags & NEGOTIATION_FLAGS.INTIMIDATION ~= 0 then
        -- NOTE: only_show because the actual opinion event needs to happen in Encounter:RunNegotiation,
        -- to ensure it occurs before any subsequence option handlers are run.
       self:ReceiveOpinion(OPINION.ATTEMPT_INTIMIDATION, { only_show = true }, agent)
    end

    self.last_param = params
    return self
end



function ConvoOption:OnWin(p)
    assert_warning( p == nil, "Hey, this isn't how OnWin() works" )

    if self.parent_option then
        return self.parent_option:OnWin()
    end


    if self.last_param and self.last_param.is_battle then
        self.last_param.on_win = self.hub:Handler()
        self.last_param.on_win.parent_option = self
    else
        assert(nil, "Bad OnWin")
    end

    return self.last_param.on_win
end

function ConvoOption:OnSuccess(p)
    assert_warning( p == nil, "Hey, this isn't how OnSuccess() works" )

    if self.parent_option then
        return self.parent_option:OnSuccess()
    end


    if self.last_param and self.last_param.is_negotiation then
        if not self.last_param.on_success then
            self.last_param.on_success = self.hub:Handler()
            self.last_param.on_success.parent_option = self
        end
    else
        error("Bad OnSuccess: "..tostr(self.last_param))
    end

    return self.last_param.on_success
end

function ConvoOption:OnFailure()
    if self.parent_option then
        return self.parent_option:OnFailure()
    end
    
    if self.last_param and self.last_param.is_negotiation then
        if not self.last_param.on_fail then
            self.last_param.on_fail = self.hub:Handler()
            self.last_param.on_fail.parent_option = self
        end
    else
        assert(nil, "Bad OnFailure")
    end

    return self.last_param.on_fail
end



function ConvoOption:CompleteQuest(objective)
    self:PushHandler(function() 
        assert_warning( self.hub.quest ~= nil, self.hub )
        if self.hub.quest then
            self.hub.quest:Complete(objective)
            if not objective then
                ConvoUtil.GiveQuestRewards(self.hub)
            end
        end
    end)
    return self
end

function ConvoOption:ActivateQuest(objective)
    self:PushHandler(function() 
        if self.hub.quest then
            self.hub.quest:Activate(objective)
        end
    end)
    return self
end

function ConvoOption:FailQuest(objective)
    self:PushHandler(function() 
        if self.hub.quest then
            self.hub.quest:Fail(objective)
        end
    end)
    return self
end

function ConvoOption:CancelQuest(objective)
    self:PushHandler(function() 
        if self.hub.quest then
            self.hub.quest:Cancel(objective)
        end
    end)
    return self
end

function ConvoOption:CheckDemo()
    if TheGame:IsDemo() then
        self:PushHandler(function()
            TheGame:DeleteSave(TheGame:GetGameState().save_file_name, function() print ("DELETED", TheGame:GetGameState().save_file_name)  end)
            UIHelpers.ShowDemoMessage()
            self.hub.encounter:YieldEncounter()
        end)
    end

    return self
end
