require "encounter/base_hub"
local ExpressionParser = require "util/expression_parser"

CONVO_FLAGS = MakeBitField{ "DISABLED", "QUEST", "NO_PRESENT" }

function GetLocalizedConvoCommonString(option)
    local key = "CONVO_COMMON." .. option
    local str = Content.LookupString( key )
    if str then
        return str
    else
        return option
    end
end


function GenerateQuestConvoID( quest_id )
    -- Must create unique names, since there are multiple convos per quest!
    -- This grossness hides the grossness of having to manually construct these unique names.
    local num = 1
    local uid = quest_id
    while Content.GetConvoStateGraph( uid ) ~= nil do
        num = num + 1
        uid = string.format( "%s_%d", quest_id, num )
    end

    return uid
end


-----------------------------------------------------------------------
--

local State = class("State")
State.can_reload = true

function State:init(id, convodef)
    self.id = id
    self.convo = convodef
    self.flags = {}
    self.fns = {}
end

function State:GetLocPrefix()
	assert(self.convo)
	assert(self.id)
	local key = self.convo:GetLocPrefix() .. "." .. string.upper(self.id)
	return key
end

function State:GetLocKey(loc_id)
	return self:GetLocPrefix() .. "." .. string.upper(loc_id)
end

function State:HarvestStrings(t)
	assert(self.convo, "Convo states must have references to their parent in order to generate a unique id")

    if self.loc then
        for id, str in pairs(self.loc) do
            local key = self:GetLocKey(id)
            t[key] = str:gsub( "\n( +)", "\n" ):trim()
        end
    end

    if self.loc_tables then
        
        for id, tbl in pairs(self.loc_tables) do
            for k, str in ipairs(tbl) do
                local key = self:GetLocKey(id) .. "_" .. k
                -- Trim all intervening whitespace between newlines.
                t[key] = str:gsub( "\n( +)", "\n" ):trim()
            end
        end
    end

end

function State:HasFlag(id)
    return self.flags[id] == true
end

function State.RunState( cxt, self, ... )
    if #self.fns == 0 then
        LOGWARN( "No function defined for state '%s.%s'", self.convo.id, self.id )
    else
        for k,fn in ipairs(self.fns) do
            fn( cxt, ... )
        end
    end
end

-------------------------------------------------------------------------------
-- Encounter driver based in raw lua script.
-- Accepts a function table with named keys -> function( encounter )

local ConvoStateGraph = class( "ConvoStateGraph", BaseHub )
ConvoStateGraph.can_reload = true

function ConvoStateGraph:init( encounter, convodef, default_state_id, quest, mask )
    assert( convodef )
    BaseHub.init( self, encounter )

    self.enc = encounter
    self.quest = quest
    self.convodef = convodef
    self.game_state = TheGame:GetGameState()
    self.now = TheGame:GetGameState():GetDateTime()
    self.act_id = self.game_state and self.game_state:GetCurrentActID()
    self.mod_id = convodef:GetModID()
    self.scratch = self.enc.scratch --just alias this
    self.cached_hook_state = {}
    self.loop_context = 0
    self.default_state_id = default_state_id or "START"
    --what we are talking about
    self.mask = mask or 0
end

function ConvoStateGraph:GetConvoMask()
    return self.mask
end

function ConvoStateGraph:OnStopConvo()
    if self.quest and not self.quest:IsDone() and CheckBits( self.quest:GetQuestDef():GetFlags(), QFLAGS.AUTO_COMPLETE ) then
        self.quest:Complete()
    end
end


function ConvoStateGraph:GetCastMember( cast_id )
    local cast_member = ConvoStateGraph._base.GetCastMember( self, cast_id )
    if cast_member == nil and self.quest then
        cast_member = self.quest:GetCastMember( cast_id )
    end
    return cast_member
end

function ConvoStateGraph:GetRNG( stream_id )
    return TheGame:GetGameState():GetRNG( stream_id )
end

function ConvoStateGraph:Clone()
    -- Somewhat shady: clone'd hubs want to share some state (cast, scratch,...) while remaining distinct instances.
    local clone = ConvoStateGraph( self.enc, self.convodef, nil, self.quest )
    clone.cast = self.cast
    clone.cast_aliases = self.cast_aliases
    clone.scratch = self.scratch
    return clone
end

function ConvoStateGraph:OnStartConvo( enc )
    -- Conversations reloaded: refresh to new definition.
    if self.quest then
        self.quest:LoadCastAnims()
    end


    self.convodef = Content.GetConvoStateGraph( self.convodef.id )

    if self.convodef.cooldown then
        self.enc:GetCaravan():StartCooldown( self.convodef.cooldown.id, self.convodef.cooldown.duration )
    end

    if not CheckBits( self.convodef.flags, CONVO_FLAGS.NO_PRESENT ) then
        enc:PresentDefault()
    end
end

function ConvoStateGraph:GetPriority()
    return self.convodef.priority or 0
end

function ConvoStateGraph:RunHub( enc, hub_state, ... )
    if hub_state == nil then
        hub_state = self.default_state_id
    end
    if hub_state == HUB_STATE.CONFRONT then
        self.enc:GetCaravan():StartTimer( HUB_STATE.CONFRONT )
    end

    if IsHubState(hub_state) then
        if self.convodef.hooks[hub_state] then
            self.convodef.hooks[hub_state](self, enc:GetPrimaryCast(), ...)
            return true
        else
            LOGWARN( "NO HUB STATE '%s.%s'", self.convodef.id, hub_state )
            return false
        end

    elseif self.cached_hook_state[ hub_state ] then
        hub_state = self.cached_hook_state[ hub_state ]
    end

    if hub_state then
        return self:RunState( hub_state, ... )
    else
        return false
    end
end




function ConvoStateGraph:ValidateHub( primary_cast, hub_state, ... )
    assert( hub_state )
    if self.convodef.cooldown then
        local time_left = self.enc:GetCaravan():CheckTimer( self.convodef.cooldown.id, self.convodef.cooldown.duration )
        if time_left then
            return false
        end
    end

    if IsHubState(hub_state) then
        return self.convodef.hooks[ hub_state ] ~= nil

    else
        local hook = self.convodef.hooks[hub_state]
        if hook then
            local new_state = hook(self, primary_cast, ...)
            self.cached_hook_state[ hub_state ] = new_state
            hub_state = new_state
        end
    end

    return hub_state and self.convodef.states[ hub_state ] ~= nil
end

function ConvoStateGraph:Wait()
    self.enc:Wait()
end

function ConvoStateGraph:End()
    self.enc:End()
end

function ConvoStateGraph:GetAgent()
    return self.enc:GetAgent()
end

function ConvoStateGraph:SeededRandom( n1, n2 )
    return TheGame:GetGameState():SeededRandom( n1, n2 )
end


function ConvoStateGraph:GetLocStringTable(id)
    if id == nil then
        return table.empty
    end
    
    if self.running_state and self.running_state.loc_tables and self.running_state.loc_tables[id] then
        return self.running_state.loc_tables[id]
    end

    if self.convodef and self.convodef.loc_tables and self.convodef.loc_tables[id] then
        return self.convodef.loc_tables[id]
    end
    return table.empty
end

function ConvoStateGraph:GetLocString(id, ...)
    if id == nil then
        return ""
    end
    
    local loc_str = nil
    local loc_key = nil

    if self.running_state and self.running_state.loc and self.running_state.loc[id] then
		loc_key = self.running_state:GetLocKey(id)
        loc_str = LOC(loc_key)

    elseif self.convodef.loc and self.convodef.loc[id] then
        loc_key = self.convodef:GetLocPrefix() .. "." .. id
        loc_str = LOC(loc_key)

    else
        loc_key = "CONVO_COMMON." .. id
        loc_str = Content.LookupString( loc_key )
    end

    if loc_str == nil then
        loc_key = nil
        loc_str = self.enc:LocFormat( id, ... )
    else
        self.fresh_dialog = not TheGame:GetGameState():HasDialogMemory(loc_key)
        loc_str = self.enc:LocFormat( loc_str,... )
        self.fresh_dialog = nil

    end

    return loc_str, loc_key
end

function ConvoStateGraph:GetContentID()
    return self.convodef.id
end

function ConvoStateGraph:GetStateID()
    return self.running_state and self.running_state.id or self.convodef.id
end

function ConvoStateGraph:FadeOut()
    self.enc:GetScreen():Fade( .5 )
    self.enc:Wait(.5)
end

function ConvoStateGraph:FadeIn()
    self.enc:GetScreen():Fade( .25, true )
    self.enc:Wait(.25)
end


function ConvoStateGraph:TalkToParty(id, reuse, filter_fn)
    if reuse == nil then
        reuse = true
    end
    local fighters = CreateCombatParty( id, self.quest:GetRank(), self.location, reuse, filter_fn )
    self.enc:SetPrimaryCast(fighters[1])
    
    self.enc:PresentAgent( TheGame:GetGameState():GetPlayerAgent(), SIDE.LEFT )
    self.enc:PresentAgent( fighters[1], SIDE.RIGHT )
    return fighters
end

function ConvoStateGraph:TalkToSomeone( agents )
    for i, agent in ipairs( agents ) do
        if not agent:IsRetired() then
            self:TalkTo( agent )
            return true
        end
    end

    return false
end

function ConvoStateGraph:LoseCards(cards)
    self.enc:GetScreen():ForceWaitOnLine()

    for k, card in ipairs(cards) do
        if is_instance(card, Battle.Card) then
            self.player.battler:RemoveCard( card )
        else
            self.player.negotiator:RemoveCard( card )
        end
    end

    self.enc:GetScreen():ShowLostCards(cards, function() self.enc:ResumeEncounter() end)
    self.enc:YieldEncounter()
end

function ConvoStateGraph:GainCards(cards_to_give)
    self.enc:GetScreen():ForceWaitOnLine()

    local cards_given = {}
    for k,card_id in ipairs(cards_to_give) do
        local player = TheGame:GetGameState():GetPlayerAgent()
        if Content.GetBattleCard( card_id ) then
            table.insert(cards_given, player.battler:LearnCard( card_id ))
        elseif Content.GetNegotiationCard( card_id ) then
            table.insert(cards_given, player.negotiator:LearnCard( card_id ))
        else
            LOGWARN( "Invalid card_id: '%s'", tostring(card_id))
        end
    end

    self.enc:GetScreen():ShowGainCards(cards_given, function() self.enc:ResumeEncounter() end)
    self.enc:YieldEncounter()

    if #cards_given > 1 then
        self:Dialog("DIALOG_GET_CARDS", cards_given)
    elseif #cards_given == 1 then
        if is_instance( cards_given[1], Negotiation.Card ) then
            self:Dialog( "DIALOG_GET_CARD_PRES_NEGOTIATION", cards_given[1].id)
        else
            self:Dialog( "DIALOG_GET_CARD_PRES_BATTLE", cards_given[1].id)
        end
    end
    return cards_given
end

function ConvoStateGraph:ForceTakeCards(card_list)
    self:RunLoop(function()
        local opt
        if #card_list > 1 then
            opt = self:Opt("OPT_FORCE_GAIN_CARDS", card_list)
        else
            if Content.GetNegotiationCard( card_list[1] ) then
                opt = self:Opt( "OPT_FORCE_GAIN_NEGOTIATION_CARD", card_list[1])
            else
                opt = self:Opt( "OPT_FORCE_GAIN_BATTLE_CARD", card_list[1])
            end
        end
        opt:GainCards(card_list):Pop()
    end)
end



function ConvoStateGraph:DoBattle(params)
    params = params or {}
    local scenario = self.enc:CreateScenario( params, self.quest )
	scenario.content_id = self.quest and self.quest.content_id
    if params.on_runaway == nil then
        scenario:SetFlags( BATTLE_FLAGS.NO_FLEE )
    end
    
    local result, battle, repercussions = self.enc:DoBattle( self, scenario )
    local handler

    if params.IS_EXPERIMENT then
        self.enc:GetScreen():ForceWaitOnLine()
        handler = params.on_experiment_done
    else
        if result == BATTLE_RESULT.WON then
            self.enc:GetScreen():ForceWaitOnLine()

            if repercussions.forced_cards and #repercussions.forced_cards > 0 then
                local did_murder = repercussions.murders and next (repercussions.murders)
                self:Dialog(did_murder and "DIALOG_COMBAT_REPERCUSSIONS_MURDER" or "DIALOG_COMBAT_REPERCUSSIONS")
                self:ForceTakeCards(repercussions.forced_cards)
            end

            handler = scenario.params.on_win

        else
            handler = scenario.params.on_runaway
        end
    end


    self.enc.scratch.any_enemies_died = false
    self.enc.scratch.all_enemies_died = true

    local team = scenario:GetTeam(TEAM.RED)
    
    for k,v in ipairs(team) do
        if v:IsSentient() and v:IsDead() then
            self.enc.scratch.any_enemies_died = true
        else
            self.enc.scratch.all_enemies_died = false
        end
    end
    if #team > 1 then
        self.enc.scratch.multiple_enemies = true
    end



    if not handler then
        if self:GetAgent() then
            if self:GetAgent():InLimbo() then
                self:End()
            else
                self:Quip( self:GetAgent(), "greeting" )
            end
        end

    elseif type(handler) == "function" then
        handler( self, battle )

    elseif is_instance(handler, ConvoOption) then
        handler:RunHandlers(battle)
    end
end


function ConvoStateGraph:BasicNegotiation(name, params)
    return self:Opt("OPT_"..name)
                :Dialog("DIALOG_"..name)
                :Negotiation(params)
                    :OnSuccess()
                        :Dialog("DIALOG_"..name.."_SUCCESS")
                    :OnFailure()
                        :Dialog("DIALOG_"..name.."_FAILURE")
end

function ConvoStateGraph:DoNegotiation( params )
    self.enc:GetScreen():ForceWaitOnLine()
    
    local negotiation_defs = require "negotiation/negotiation_defs"

    local agent = params.target_agent or self:GetAgent()
    
    if params.subject then
        self:ReassignCastMember("subject", params.subject)
    end
    if params.quest == nil then
        params.quest = self.quest
    end

    local start_params = Negotiation.MiniGame.ParamsFromParty( self.caravan, agent, params, self.enc )
    local minigame = Negotiation.MiniGame( start_params, self.caravan )

    local result = self.enc:RunNegotiation( minigame )

    if result == negotiation_defs.RESULT.WIN then

        if params.on_success then
            if type(params.on_success) == "function" then
                params.on_success( self, minigame )
            elseif is_instance(params.on_success, ConvoOption) then
                params.on_success:RunHandlers(minigame)
            end
        end
        

    elseif result == negotiation_defs.RESULT.LOSE then
        
        if minigame.force_take_cards and #minigame.force_take_cards > 0 then
            self:ForceTakeCards(minigame.force_take_cards)
        end

        if params.on_fail then
            if type(params.on_fail) == "function" then
                params.on_fail( self, minigame )
            elseif is_instance(params.on_fail, ConvoOption) then
                params.on_fail:RunHandlers(minigame)
            end
        end
    end


    return result
end

function ConvoStateGraph:Handler()
    local opt = ConvoOption()
    opt:AttachHub( self.encounter.active_hub )
    return opt
end

function ConvoStateGraph:Opt(title_id, ...)
    local str = self:GetLocString(title_id, ...)
    
    if str == title_id then
        assert_warning(nil, loc.format( "UNLOCALIZED DIALOG {1}:{2}:{3}", self.convodef.id or "???", self.running_state and self.running_state.id or "???", title_id ))
    end

    local opt = ConvoOption( str )
    self.encounter:AddOption( opt )

    opt.loc_id = title_id
    
    if self.quest and #self.enc.hub_stack == 1 and self.enc.hub_stack[1].hub_state == HUB_STATE.HUB then
        opt:SetQuestMark()
    end

    return opt
end

function ConvoStateGraph:QuestOpt( quest, obj_id )
    quest = quest or self.quest
    local provider = quest:GetProvider()
    local opt = ConvoOption( loc.format( LOC"CONVO_OPTION.BASIC_OPPORTUNITY", quest:GetTitle() ))

    if obj_id then
        local title = quest:GetObjectiveTitle( obj_id )
        if title then
            opt:PostText( title )
        end
    end
    opt:SetQuestMark( quest, self:GetAgent() )
    self.encounter:AddOption( opt )
    return opt
end



function ConvoStateGraph:GoTo(state_id, ...)
    local new_hub = self:Clone()
    self.enc:RunHub( new_hub, state_id, ...)
end

function ConvoStateGraph:GoToExt(convo_id, state_id, ...)
    local convodef = Content.GetConvoStateGraph(convo_id)
    assert_warning( convodef, "CAN'T FIND CONVO %s", convo_id )
    if convodef then
        local new_hub = ConvoStateGraph(self.enc, convodef )
        self.enc:RunHub( new_hub, state_id, ... )
    end
end

function ConvoStateGraph:GoToQuest(quest, state_id, ...)
    assert( is_instance( quest, QuestState ))

    local convodef = Content.GetConvoStateGraph( quest:GetContentID() )
    assert_warning( convodef, "CAN'T FIND CONVO for quest %s", tostring(quest) )
    if convodef then
        local new_hub = ConvoStateGraph(self.enc, convodef, nil, quest )
        self.enc:RunHub( new_hub, state_id, ... )
    end
end

function ConvoStateGraph:Quip(agent, quip, ...)
    self.enc:SayQuip( agent, quip, ... )
end

function ConvoStateGraph:InsertQuestions(id)
    local ret = 0
    if self.convodef.question_data and self.convodef.question_data[id] then
        for k,v in ipairs(self.convodef.question_data[id]) do
            local added = self:Question(v.opt_id, v.dialog_id)
            if added then
                ret = ret + 1
            end

        end
    end
    return ret
end

function ConvoStateGraph:Dialog(dialog_id, ...)
    local str, id = self:GetLocString( dialog_id, ... )

    if id then
        if Content.LookupString(id) == nil then
            LOGWARN( "NO STRING: %s (%s)", id, str )
        else
            TheGame:GetGameState():RememberDialog(id)
        end
    end

    if str == dialog_id then
        assert_warning(nil, loc.format( "UNLOCALIZED DIALOG {1}:{2}:{3}", self.convodef.id or "???", self.running_state and self.running_state.id or "???", dialog_id ))
    end

    self.enc:Dialog( str )
end


function ConvoStateGraph:AwardUnlock(unlock_id, character_to_show, title, body)
    
    if not TheGame:GetGameProfile():HasUnlock(unlock_id) and not TheGame:IsDemo() then
        self.enc:WaitOnLine()
        self:FadeOut()
        UIHelpers.ShowUnlockMessage(unlock_id, character_to_show, title, body, function() 
            self.enc:ResumeEncounter()
        end )
        self.enc:YieldEncounter()
        self:FadeIn()
    end
end


function ConvoStateGraph:ProgressDialog(progress_id, dialog_id, ...)
    local progress = TheGame:GetGameProfile():GetNarrativeProgress(progress_id)
    local t = self.convodef and self:GetLocStringTable( dialog_id ) or table.empty

    local idx = math.min( progress+1, #t)
    local loc_key = self.running_state:GetLocKey(dialog_id) .. "_" .. tostring(idx)

    local str = LOC(loc_key)

    self.fresh_dialog = not TheGame:GetGameState():HasDialogMemory(loc_key)
    str = self.enc:LocFormat( loc.format( str, ... ) )
    self.fresh_dialog = nil

    TheGame:GetGameProfile():IncNarrativeProgress(progress_id)
    
    self.enc:Dialog(str)
end

function ConvoStateGraph:RunFn(fn, ...)
    self.enc:ClearOptions()
    
    fn(self, ...)

    if not self.enc:IsEnded() and self.enc:CountOptions() > 0 then

        local opt = self.enc:ChooseOption( )
        if opt then
            opt:RunOption()
        end
    end
end

function ConvoStateGraph:RunLoopingFn(fn, ...)
    self.loop_context = self.loop_context + 1
    local loop_context = self.loop_context

    self.num_loops = 0

    repeat
        self:RunFn( fn, ... )
        self.num_loops = self.num_loops + 1
        assert( self.num_loops < 1024 )
    until self.loop_context ~= loop_context or self.enc:IsEnded()
end


function ConvoStateGraph:Run(to_run, ...)
    if type(to_run) == "function" then
        self:RunFn(to_run, ...)
    elseif type(to_run) == "string" then
        self:RunState(to_run, ...)
    else
        self.enc:Wait()
    end
end

function ConvoStateGraph:Pop()
    -- ccc: :( lots of places issue redundant Pops whether or not they're 'looping'.
    --assert_warning( self.loop_context > 0 )
    self.loop_context = math.max( 0, self.loop_context - 1 )
end

function ConvoStateGraph:__tostring()
    if self.quest then
        return string.format( "SG %s: %s.%s", tostring(self.quest), self.convodef.id, self.running_state and self.running_state.id or "???" )
    else
        return string.format("SG %s.%s", self.convodef.id, self.running_state and self.running_state.id or "???")
    end
end

function ConvoStateGraph:PopConvo()
    local convodef = self.convodef

    for i = #self.enc.hub_stack, 1, -1 do
        local hub = self.enc.hub_stack[i]
        if hub.convodef == convodef then
            hub.loop_context = 0 -- Unwind completely.
        else
            break
        end
    end
end

function ConvoStateGraph:FirstLoop()
    return (self.num_loops or 0) == 0
end

function ConvoStateGraph:GetNumLoops()
    return self.num_loops or 0
end

function ConvoStateGraph:RunState(state_id, ...)   

    state_id = state_id or self.default_state_id

    local state = self.convodef.states[state_id]
    self.running_state = state

    if not state then
        LOGWARN( "Trying to play missing state: %s.%s", self.convodef.id, state_id )
        return false
    end

    if TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY then
        if state_id ~= "CONF" then -- Yes, hacking this in against the clock.
            DoAutoSave()
        end
    end

    if state.is_looping then
        self:RunLoopingFn( state.RunState, state, ... )
    else
        self:RunFn( state.RunState, state, ... )
    end
    self.running_state = nil

    return true
end

function ConvoStateGraph:AskAbout(state)
    return self:Opt("OPT_ASK_ABOUT")
        :IsHubOption(true)
        :GoTo(state or "STATE_QUESTIONS")
end

function ConvoStateGraph:TalkTo(who)
    local agent
    if type(who) == "string" then
        agent = self:GetCastMember(who)
    else
        agent = who
    end
    
    if agent then
        self.enc:SetPrimaryCast(agent)
    end
end

function ConvoStateGraph:QST(id)
    self:Question("OPT_"..id, "DIALOG_"..id)
end

function ConvoStateGraph:Question(opt, dialog, fn)
    self.enc.scratch.asked_questions = self.enc.scratch.asked_questions or {}
    if not self.enc.scratch.asked_questions[opt] then
        self:Opt(opt)
            :Dialog(dialog)
            :Fn(function() 
                self.enc.scratch.asked_questions[opt] = true 
                if fn then
                    fn()
                end
            end)
        return true
    end
end

function ConvoStateGraph:RunLoop(fn)
    self:RunLoopingFn(fn, self)
    --self:PushHandler( self.hub.RunLoopingFn, self.hub, fn, ...)
end

function ConvoStateGraph:LookupQuip( tags, seed )
    if self.convodef.quip_db then
        return self.convodef.quip_db:LookupQuip( tags, seed )
    end
end


function ConvoStateGraph:SubstituteFormatter( text, rest )
    if rest and string.find(rest, "^%s*%?") then
        local tags = {}
        if self:GetAgent() then
            self:GetAgent():FillOutQuipTags(tags)
        end
        local val = ExpressionParser.Evaluate(text, function(id) 
            if id == "first_time" then
                return self.fresh_dialog
            end

            return self.enc.scratch[id] or (self.quest and self.quest.param[id]) or (table.arrayfind(tags, id) ~= nil) 
        end)
        return val == true
    end

    if self.quest then
        local cast_member = self.quest:GetCastMember( text )
        if cast_member then
            return cast_member
        end

        local v = self.quest:GetParameter( text )
        if v then
            return v
        end
    end

    local cast_member = self:GetCastMember( text )
    if cast_member then
        return cast_member
    end

end

function ConvoStateGraph:Travel()
    self.enc:DoTravel(true)
end


function ConvoStateGraph:PlayQuestConvo(quest, hook, ...)
    local convodef = quest:GetQuestDef():GetRegisteredConvo(hook)
    if convodef then
        local new_hub = ConvoStateGraph(self.enc, convodef, nil, quest )
        self.enc:RunHub( new_hub, nil ,... )
        return true
    end

    return false
end

function ConvoStateGraph:Opinion(what, who, userdata)
    
    if type(who) == "string" then
        who = self:GetCastMember(who)
    end
    if who then
        if type(what) == "string" then
            what = self.quest and self.quest:GetQuestDef():GetOpinionEvent(what)
        end

        if what then
            who:OpinionEvent(what, userdata)
        end
    end
end


--Wraps definition of a ConvoStateGraph
---------------------------------------------------------------------------

local Convo = class("Convo")
Convo.can_reload = true

function Convo:init(id, flags)
    assert( id )
    
    
    local n = 2
    
    while true do

        local t = debug.getinfo( n )
        if not t then
            break
        end
        local path = t.source -- Ergh.

        if (string.find(path, "content") and not string.find(path, "content.lua")) then
            
            self.def_path = engine.inst:ResolveFileName(path:gsub("@", ""))
            self.def_line = t.currentline
            break
        end
        n = n + 1
    end

    self.states = {}
    self.id = id
    self.flags = flags or 0
    self.hooks = {}
    
    self.req_mask = CONVERSATION_FLAGS.ROOT
    self.dis_mask = 0
    self.any_mask = 0

    Content.AddConvoStateGraph(self)
end

function Convo:GetLocPrefix()
	return "CONVO." .. string.upper(self.id)
end

function Convo:SetModID( mod_id )
    self.mod_id = mod_id
end

function Convo:GetModID()
    return self.mod_id
end

function Convo:HarvestStrings(t)

    if self.quest_id then
        local qdef = Content.GetQuestDef( self.quest_id )
        if qdef == nil then
            return 
        end
    end

	local prefix = self:GetLocPrefix()
    if self.loc then
        for id, str in pairs(self.loc) do
			local key = prefix  .. "." .. id
            t[key] = str:gsub( "\n( +)", "\n" ):trim()
        end
    end

	if self.states ~= nil then
		for _, state in pairs(self.states) do
			state:HarvestStrings(t)
		end
	end

    if self.quip_db then
        self.quip_db:HarvestStrings( t )
    end
end


function Convo:Priority(p)
    self.priority = p
    return self
end

function Convo:Cooldown( duration, id )
    assert( duration > 0 )
    self.cooldown = { duration = duration, id = id or self.id }
    return self
end

function Convo:SetLooping()
    assert( self.default_state, "NO DEFAULT STATE SET")
    self.default_state.is_looping = true
    return self
end

function Convo:State(id)
    assert(self.states[id] == nil, "DUPLICATE STATE")
    local state = State(id, self)
    self.states[id] = state
    self.default_state = state
    return self
end

function Convo:ProgressionLoc(dict)
    for string_id, v in pairs(dict) do
        local t= {}
        for k, v in ipairs(v) do
            table.insert(t, v)
        end
        self:AddLocStringTable(string_id, t)
    end
    return self
end

function Convo:AddLocStringTable(string_id, t)
    --assert( string.len( str ) ~= nil or error( string.format( "Mal-formed utf8 string: %s. Is your file the wrong encoding?", string_id )))
    if self.default_state then
        self.default_state.loc_tables = self.default_state.loc_tables or {}
        self.default_state.loc_tables[string_id] = t
    else
        self.loc_tables = self.loc_tables or {}
        self.loc_tables[string_id] = t
    end
end


function Convo:Loc(dict)
    for k,v in pairs(dict) do
        if type(v) == "string" then
            self:AddLocString(k, v)
        end
    end
    return self
end

function Convo:Quips(dict)
    if self.quip_db == nil then
        self.quip_db = QuipDatabase( self:GetLocPrefix() )
    end
    self.quip_db:ParseTable( dict )
    return self
end

function Convo:RunLoopingFn(fn)
    self:SetLooping()
    return self:Fn(fn)
end

function Convo:Fn(fn)
    assert(self.default_state, "NO STATE PUSHED")
    table.insert(self.default_state.fns, fn)
    return self
end

function Convo:SetFlag(id)
    if self.default_state then
        self.default_state.flags[id] = true
    else
        self.flags = SetBits( self.flags, id )
    end
    return self
end

--these are our entry points for conversations
function Convo:Hub( arg1, arg2 )
    local hub_state, fn
    if IsEnum( arg1, HUB_STATE ) then
        hub_state, fn = arg1, arg2
    else
        hub_state, fn = HUB_STATE.HUB, arg1
    end
    self.hooks[hub_state] = fn
    Content.AddConvoStateGraph(self, hub_state)
    return self
end


function Convo:Hub_Location( fn )
    self.hooks[HUB_STATE.LOCATION] = fn
    Content.AddConvoStateGraph(self, HUB_STATE.LOCATION)
    return self
end

function Convo:Confront(fn)
    self.hooks[HUB_STATE.CONFRONT] = fn
    Content.AddConvoStateGraph(self, HUB_STATE.CONFRONT)
    return self
end

function Convo:Attract(fn)
    self.hooks[HUB_STATE.ATTRACT] = fn
    Content.AddConvoStateGraph(self, HUB_STATE.ATTRACT)
    return self
end

function Convo:AttractState(id, condition)
    self:Attract(function(cxt) if condition == nil or condition(cxt) then return id end end)
    self:State(id)
    return self
end

function Convo:ConfrontState(id, condition)
    id = id or "CONF"
    self:Confront(function(cxt, who) if condition == nil or condition(cxt, who) then return id end end)
    self:State(id)
    return self
end

function Convo:TravelConfront(id, condition)
    
    self:Confront(function(cxt) if cxt.location:HasTag("in_transit") and (condition == nil or condition(cxt)) then return id end end)

    self:State(id)
    return self
end

function Convo:Validator( name, fn )
    self.hooks[ name ] = fn
    Content.AddConvoStateGraph(self, name)
    return self
end

function Convo:AddLocString(string_id, str)
    assert( string.len( str ) ~= nil or error( string.format( "Mal-formed utf8 string: %s. Is your file the wrong encoding?", string_id )))
    if self.default_state then
        self.default_state.loc = self.default_state.loc or {}
        self.default_state.loc[string_id] = str
    else
        self.loc = self.loc or {}
        self.loc[string_id] = str
    end
end

function Convo:Aside(args)
    assert(args.loc, "Missing loc id in aside")
    assert(args.opt, "Missing opt text in aside")
    assert(args.txt, "Missing txt in aside")
    assert(self.default_state, "No state pushed")

    local loc_id_opt = loc.format("ASIDE_{1}_OPT", args.loc)
    local loc_id_dialog = loc.format("ASIDE_{1}_DIALOG", args.loc)
    self:AddLocString(loc_id_opt, args.opt)
    self:AddLocString(loc_id_dialog, args.txt)
    self:Fn(function(cxt) 
            if not args.condition or args.condition(cxt) then
                cxt:Opt(loc_id_opt):Dialog(loc_id_dialog)
            end
        end)

    return self
end

function Convo:Dialog(args)
    assert(args.loc, "Missing loc id in dialog")
    assert(args.txt, "Missing txt dialog")
    assert(self.default_state, "No state pushed")

    local loc_id_dialog = loc.format("DIALOG_{1}", args.loc)
    self:AddLocString(loc_id_dialog, args.txt)
    self:Fn(function(cxt) 
            if not args.condition or args.condition(cxt) then
                cxt:Dialog(loc_id_dialog)
            end
        end)
    return self
end

function Convo:Do(args)
    assert(args.loc, "Missing loc id in dialog")
    assert(args.opt, "Missing opt dialog")
    assert(args.fn, "Missing fn")
    assert(self.default_state, "No state pushed")

    local loc_id_opt = loc.format("OPT_{1}", args.loc)
    self:AddLocString(loc_id_opt, args.opt)

    local loc_id_dialog
    if args.txt then
        loc_id_dialog = loc.format("DIALOG_{1}", args.loc)
        self:AddLocString(loc_id_dialog, args.txt)
    end

    self:Fn(function(cxt) 
            if not args.condition or args.condition(cxt) then
                if loc_id_dialog then
                    cxt:Opt(loc_id_opt):Dialog(loc_id_dialog):Fn(args.fn)
                else
                    cxt:Opt(loc_id_opt):Fn(args.fn)
                end
            end
        end)
    return self
end

function Convo:GoTo(args)
    assert(args.loc, "Missing loc id in dialog")
    assert(args.opt, "Missing opt dialog")
    assert(args.state, "Missing dest state")
    assert(self.default_state, "No state pushed")

    local loc_id_opt = loc.format("OPT_{1}", args.loc)
    self:AddLocString(loc_id_opt, args.opt)

    local loc_id_dialog
    if args.txt then
        loc_id_dialog = loc.format("DIALOG_{1}", args.loc)
        self:AddLocString(loc_id_dialog, args.txt)
    end

    self:Fn(function(cxt) 
            if not args.condition or args.condition(cxt) then
                if loc_id_dialog then
                    cxt:Opt(loc_id_opt):Dialog(loc_id_dialog):GoTo(args.state)
                else
                    cxt:Opt(loc_id_opt):GoTo(args.state)
                end
            end
        end)
    return self
end

function Convo:Questions(id, args)
    self.question_data = self.question_data or {}


    local data = {}
    local loctable = {}
    local n = 1
    for k = 1, #args, 2 do
        local t= {
            opt_id = "OPT_QUESTION_"..tostring(n)..id,
            opt_txt = args[k],
            dialog_id = "DIALOG_QUESTION_"..tostring(n)..id,
            dialog_txt = args[k+1]
        }
        table.insert(data, t)
        loctable[t.opt_id] = t.opt_txt
        loctable[t.dialog_id] = t.dialog_txt
        n = n + 1
    end

    self.question_data[id] =  data
    self:Loc(loctable)

    return self
end


function Convo:AskAboutHub(state_id, args)

    local opt = {}
    local loctable = {}
    local n = 1
    for k = 1, #args, 2 do
        local t= {
            opt_id = "ASK_ABOUT_OPT_"..n,
            opt_txt = args[k],
            dialog_id = "ASK_ABOUT_DIALOG_"..n,
            dialog_txt = args[k+1]
        }
        
        loctable[t.opt_id] = t.opt_txt
        loctable[t.dialog_id] = t.dialog_txt
        table.insert(opt, t)
        n = n + 1
    end
    
    local scratchid = state_id .. "_HISTORY"
    self:State(state_id)  
        :Loc(loctable)
        :SetLooping(true)
        :Fn(function(cxt) 
            cxt.enc.scratch[scratchid] = cxt.enc.scratch[scratchid] or {}
            local history = cxt.enc.scratch[scratchid]
            for _,optdata in ipairs(opt) do

                cxt:Opt(optdata.opt_id)
                    :Dialog(optdata.dialog_id)
                    :ReqCondition(history[optdata] == nil)
                    :Fn(function(cxt) history[optdata] = true end)

            end
            StateGraphUtil.AddBackButton(cxt)
        end)

    return self
end

function Convo:AskAboutHubConditions(state_id, args)

    local opt = {}
    local loctable = {}
    local n = 1
    for k = 1, #args, 4 do
        local t= {
            opt_id = "OPT_"..n,
            condition = args[k],
            opt_txt = args[k+1],
            dialog_id = "DIALOG_"..n,
            dialog_txt = args[k+2],
            side_effect = args[k+3],
        }
        
        assert( t.condition == nil or type(t.condition) == "function", string.format( "%s.%s", state_id, k ) )
        assert( t.side_effect == nil or type(t.side_effect) == "function", string.format( "%s.%s", state_id, k ) )

        loctable[t.opt_id] = t.opt_txt
        loctable[t.dialog_id] = t.dialog_txt
        table.insert(opt, t)
        n = n + 1
    end
    
    local scratchid = state_id .. "_HISTORY"
    self:State(state_id)  
        :Loc(loctable)
        :SetLooping(true)
        :Fn(function(cxt) 
            cxt.enc.scratch[scratchid] = cxt.enc.scratch[scratchid] or {}
            local history = cxt.enc.scratch[scratchid]
            for _,optdata in ipairs(opt) do
                if not optdata.condition or optdata.condition(cxt) then
                    cxt:Opt(optdata.opt_id)
                        :Dialog(optdata.dialog_id)
                        :ReqCondition(history[optdata] == nil)
                        :Fn(function(cxt) 
                            history[optdata] = true 
                            if optdata.side_effect then
                                optdata.side_effect(cxt)
                            end
                        end)
                end

            end
            StateGraphUtil.AddBackButton(cxt)
        end)

    return self
end

function Convo:RequireFlags(mask)
    self.req_mask = mask or 0
    return self
end

function Convo:AnyFlags(mask)
    self.any_mask = mask or 0
    return self
end

function Convo:DisallowFlags(mask)
    self.dis_mask = mask or 0
    return self
end


--Wraps a ConvoStateGraph used by a quest, triggered only for certain objective/cast members
----------------------------------------------------------------------------------------

local QuestConvo = class("QuestConvo", Convo)
QuestConvo.can_reload = true

function QuestConvo:init( quest_id, objective_id, cast_id)
    QuestConvo._base.init( self, GenerateQuestConvoID( quest_id ), CONVO_FLAGS.QUEST )

    self.quest_id = quest_id

    local qdef = Content.GetQuestDef( quest_id )
    if qdef == nil then
        return -- Could be disabled.
    end

    qdef:RegisterConvoStateGraph( self )

    self.objective_id = objective_id
    assert_warning( objective_id == nil or qdef:CheckObjective( objective_id ), objective_id )

    self.cast_id = cast_id
    if cast_id and not qdef:GetCast( cast_id ) then
        assert( false, loc.format( "Unknown cast id: {1} (quest {2})", cast_id, quest_id ))
    end

    self:Priority( CONVO_PRIORITY_HIGHEST )
end




--Wraps a ConvoStateGraph used by a quest triggered only post-completion.
----------------------------------------------------------------------------------------

local QuestPostConvo = class("QuestPostConvo", Convo)
QuestPostConvo.can_reload = true

function QuestPostConvo:init( quest_id, cast_id )
    QuestPostConvo._base.init( self, GenerateQuestConvoID( quest_id ), CONVO_FLAGS.QUEST )

    local qdef = Content.GetQuestDef( quest_id )
    if qdef == nil then
        return -- Could be disabled.
    end

    qdef:RegisterConvoStateGraph( self )

    self.cast_id = cast_id
    assert_warning( cast_id == nil or qdef:GetCast( cast_id ), cast_id )

    self.post_convo = true
end











