local paperdoll = require "paperdoll_util"
require "encounter/default_hub"
require "battle/scenario"
require "battle/battle_engine"
require "ui/screens/fightscreen"
require "ui/screens/lootscreen"
require "ui/screens/pickgraftscreen"
require "ui/screens/pickcoinscreen"
require "ui/screens/gameoverscreen"
require "ui/screens/socialgraftscreen"
require "ui/screens/dayoverscreen"
require "encounter/convostategraph"
require "ui/screens/jobpopup"
local battle_defs = require "battle/battle_defs"
SIDE = MakeEnum{"LEFT", "LEFT_AUX", "RIGHT", "RIGHT_AUX" }

-------------------------------------------------------------------------------
-- Runs generic, scriptable encounters.
--  Maintains a context consisting of a WHAT is being encountered ('primary_cast') and a WHERE ('location')
--  Hubs are controllers which call Encounter routines to drive a sequence.
--  There is one active_hub at a time.

local Encounter = class( "Encounter" )

function Encounter:init( primary_cast, location )

    if primary_cast and location == nil then
        self.location = primary_cast:GetLocation()
        assert( self.location )
    else
        self.location = location
        assert( self.location )
    end

    self.presented_agents = {}
    self.scratch = {}
    self.hub_stack = {}
    self.current_options = {}
    self.quip_tags = {}
    self.caravan = TheGame:GetGameState():GetCaravan()

    self.left_agents = {}
    self.right_agents = {}

    self.primary_cast = primary_cast

    self.pending_relationship_changes = {}
end

function Encounter:SetPrimaryCast( cast )
    if self.current_speaker == self.primary_cast then
        self:ClearCurrentSpeaker()
    end

    self.primary_cast = cast

    if is_instance( cast, Agent ) then
        assert_warning( not cast:IsDead(), "Assigning dead primary_cast" )
        self:UpdateCurrentSpeaker( cast )

        paperdoll.LoadAgentAnims(cast)

        if cast:GetLocation() ~= self.location then
            print( "Moving new primary cast", cast, " to ", self.location )
            if cast:GetParty() then
                cast:GetParty():MoveToLocation( self.location )
            else
                cast:MoveToLocation( self.location )
            end
        end
    end

    for i, hub in ipairs( self.hub_stack ) do
        hub:ReassignCastMember( "agent", cast )
    end   
end

function Encounter:SetInitialHub( initial_hub, ... )
    assert( initial_hub == nil or is_instance( initial_hub, BaseHub ))
    self.initial_hub = initial_hub
    self.initial_params = { ... }
end

local function RunHubCoro( encounter, hub, ... )
    hub:RunHub( encounter, ... )

    -- Implicit End to wait for final user acknowledgement.
    encounter:End()
end

function Encounter:StartConvo( screen )
 
    self.screen = screen

    TheGame:GetEvents():ListenForEvents( self, "agent_relationship_changed")

    self:SetPrimaryCast( self.primary_cast )
    self:PushHub( self.initial_hub or DefaultHub( self ))

    if self.active_hub.OnStartConvo then
        self.active_hub:OnStartConvo( self )
    end

    print( "StartConvo", self.active_hub, self.primary_cast, self.location, "Params:", table.unpack( self.initial_params or table.empty ))

    assert( self.coro == nil )
    self.coro = coroutine.create( RunHubCoro )
    self:ResumeEncounter( self, self.active_hub, table.unpack( self.initial_params or table.empty ))

    if self.primary_cast then
        TheGame:BroadcastEvent( "start_convo", self.primary_cast )
    end

end

function Encounter:StopConvo()
    table.clear( self.left_agents )
    table.clear( self.right_agents )
    self.right_agent, self.left_agent = nil, nil
    self.coro = nil
    self.wait_delay = nil
    self.wait_fn = nil

    while #self.hub_stack > 0 do
        self:PopHub()
    end
    self.active_hub = nil
    self.screen = nil

    table.clear( self.current_options )

    self.caravan:SetCurrentEncounter( nil )

    TheGame:GetEvents():RemoveListener( self )
end

function Encounter:YieldEncounter( wait_fn )
    if coroutine.running() == self.coro then
        -- We check this to allow external callers to invoke Encounter API without crashing (like debug functions)
        return coroutine.yield( wait_fn )
    else
        print( "WARNING: attempting to yield outside Encounter coroutine" )
        print( debug.traceback() )
    end
end

function Encounter:ResumeEncounter( ... )
    self.wait_delay = nil
    self.wait_fn = nil

    if self.coro == nil then
        return
    end

    local ok, result = coroutine.resume( self.coro, ... )
    if not ok then
        error( "ERR running driver:" ..tostring(result).."\n"..debug.traceback( self.coro) )
    end
    if self.coro then
        local status = coroutine.status( self.coro )
        if status ~= "suspended" then
            self.coro = nil
            self:PopHub()
        else
            self.wait_fn = result
        end
    end
end

function Encounter:IsRunning()
    return self.coro ~= nil
end

function Encounter:RunHub( hub, ... )
    assert( is_instance( hub, BaseHub ))
    self:PushHub( hub )
    local ret = hub:RunHub( self, ... )
    self:PopHub()

    return ret
end

-- Evaluate all these hubs for execution.  Return if handled.
function Encounter:RunHubs( hubs, state, ... )
    local handled = false
    if hubs == nil then
        hubs = self:CollectHubs( {}, state )
    end

    for i, context in ipairs( hubs ) do
        handled = self:RunHub( context, state, ... )
        if handled then
            break
        end
    end
    return handled
end


function Encounter:RunNegotiation( minigame, no_loot )

    minigame:PrintDump()

    
    local music_event = TheGame:GetGameState():GetDayPhase() == DAY_PHASE.DAY and "negotiation_music" or "negotiation_music_night"
    self:SetMusicEvent( TheGame:LookupPlayerMusic( music_event ) or "event:/music/adaptive_negotiation_barter" )

    local left_agent = self:GetLeftAgent()
    if not left_agent or not left_agent:IsPlayer() then
        self:PresentAgent( self:GetPlayer(), SIDE.LEFT )
    end

    local initial_player_state = TheGame:GetGameState():GetPlayerState()

    local result
    if TheGame:GetLocalSettings().ROBOTICS and TheGame:GetLocalSettings().AUTO_RESOLVE_NEGOTIATION then
        if TheGame:GetRoboticsSettings().negotiation_policy == "WIN" then
            result = minigame.RESULT.WIN
        elseif TheGame:GetRoboticsSettings().negotiation_policy == "LOSE" then
            result = minigame.RESULT.LOSE
        else
            result = table.arraypick{ minigame.RESULT.LOSE, minigame.RESULT.WIN }
        end

    elseif TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
        print( "AUTO-RESOLVE" )
        result = minigame.RESULT.WIN

    elseif minigame:GetPlayerNegotiator():GetResolve() == 0 then
        result = minigame.RESULT.LOSE
    else
        self.negotiation = minigame
        result = self:GetScreen():RunNegotiation( minigame )
        self.negotiation = nil

        if left_agent and not left_agent:IsPlayer() and left_agent:GetLocation() == self.location then
            self:PresentAgent( left_agent, SIDE.LEFT )
        end
    end

    local heart = minigame:GetPlayerNegotiator():FindCoreArgument()

    -- Permit mods, grafts and other external systems modify this result.
    local repercussions =
    {
        resolve = heart and heart:GetResolve() or 0,
        result = result,
    }
    if not CheckBits( minigame:GetFlags(), NEGOTIATION_FLAGS.PROVOCATION ) and result == minigame.RESULT.WIN  and not CheckBits( minigame:GetFlags(), NEGOTIATION_FLAGS.NO_LOOT ) then
        repercussions.loot = { negotiation_card_draft_difficulty = minigame:GetRewardDifficulty() or TheGame:GetGameState():GetCurrentBaseDifficulty()  }
        if TheGame:GetGameState():CanDoMettle() then
            repercussions.loot.mettle = NEGOTIATION_METTLE_REWARD[minigame:GetDifficulty()] or 0

            local bonus_mettle = minigame:GetPlayerNegotiator():GetModifierStacks("METTLE_BANK")
            if bonus_mettle and bonus_mettle > 0 then
                repercussions.loot.bonus_mettle = bonus_mettle
            end

        end

    else
        repercussions.loot = {}
    end
    TheGame:BroadcastEvent( "resolve_negotiation", minigame, repercussions )

    -- This may have changed, depending on handlers.
    result = repercussions.result

    if result == minigame.RESULT.WIN then
        TheGame:GetGameState():AddScore("NEGOTIATIONS_WON", 1)
        TheGame:AddGameplayStat( "negotiations_won", 1 )

    else
        -- TheGame:GetGameState():AddScore("NEGOTIATIONS_LOST", 1)
    end
	
    self.caravan:SetResolve( repercussions.resolve )

    if minigame.start_params.flags & NEGOTIATION_FLAGS.INTIMIDATION ~= 0 then
        minigame:GetOpponent():OpinionEvent(OPINION.ATTEMPT_INTIMIDATION)
    end

	local player_state = TheGame:GetGameState():GetPlayerState()
    local enemy_def = minigame:GetOpponent() and minigame:GetOpponent():GetContentID()
    if player_state and initial_player_state then
    	local json_t = {
    		PLAYER_DATA = player_state,
    		NEGOTIATION_DATA = {
    			DIFFICULTY = minigame.start_params.difficulty,
    			RESOLVE = minigame.start_params.enemy_resolve_required,
    			RESOLVE_DELTA = player_state.RESOLVE - initial_player_state.RESOLVE,
    			TURNS = minigame.turns,
    			CONTENT_ID = self.active_hub and self.active_hub.quest and self.active_hub.quest.content_id,
                OPPONENT_DEF = enemy_def,
    		}
    	}
    	local evt_type = (result == minigame.RESULT.WIN) and "NEGOTIATION_WON" or "NEGOTIATION_LOST"
    	SendMetricsData(evt_type, json_t)
    end

    local has_negotiation_upgrades = #self:GetPlayer().negotiator:GetCardsReadyForUpgrade() > 0
    local has_battle_upgrades = #self:GetPlayer().battler:GetCardsReadyForUpgrade() > 0

    if result == minigame.RESULT.WIN then
        for i,graft in ipairs(TheGame:GetGameState():GetPlayerAgent().graft_owner:GetGrafts(GRAFT_TYPE.NEGOTIATION)) do
            if graft:HasUpgrades() and not graft:GetDef().manual_xp and not graft:UpgradeReady() then
                graft:AddXP(1)
            end
        end
    end

    if not repercussions.no_loot_screen then
        if (repercussions.loot ~= nil and next(repercussions.loot) ~= nil) or has_negotiation_upgrades or has_battle_upgrades then 
            local loot_display_data = {}
            loot_display_data.negotiation = true
            loot_display_data.defeat = not ( result == minigame.RESULT.WIN )
            self:AwardLoot( repercussions.loot or table.empty, loot_display_data )
        end
    end

    --this is later so that the negotiation music plays through the loot screens
    self:SetMusicEvent( nil )

    TheGame:GetGameState():Action()

    return result
end

function Encounter:SetMusicEvent( music_event )
    self.music_event = music_event
    TheGame:FE():RefreshMusic()
end

function Encounter:GetMusicEvent()
    return self.music_event
end

function Encounter:PushHub( hub )
    table.insert( self.hub_stack, hub )
    self.active_hub = hub
end

function Encounter:PopHub()
    local hub = table.remove( self.hub_stack )
    hub:OnStopConvo()
    self.active_hub = self.hub_stack[ #self.hub_stack ]
end

function Encounter:OnUpdate( dt )
    if self.wait_delay then
        self.wait_delay = self.wait_delay - dt
        if self.wait_delay <= 0 then
            self.wait_delay = nil
            self:ResumeEncounter()
        end

    elseif self.wait_fn and self.wait_fn() then
        self:ResumeEncounter()
    end
end

function Encounter:IsEnded()
    return self.is_ended
end

function Encounter:CreateScenario( params, quest )
    params = params or table.empty
    
    params.ally_scale = params.ally_scale or GetDefaultAllyScale()
    params.enemy_scale = params.enemy_scale or GetDefaultEnemyScale(quest and quest:GetDifficulty())


    local scenario = Scenario( self.location, params, quest )
    scenario:AddAgentToTeam( self:GetPlayer(), TEAM.BLUE, SUPPORT_REASON.PARTY)
    scenario:SetPlayerControlled(self:GetPlayer())
    
    local non_combat = {}
    if params.noncombatants then
        for k,v in pairs(params.noncombatants) do
            non_combat[v] = true
        end
    end

    if params.allies then
        for i, cast in ipairs( params.allies ) do
            local agent = is_instance(cast, Agent) and cast or  self:GetCastAgent( cast )
            if agent and scenario:GetTeamForAgent( agent ) == nil  then
                scenario:AddAgentToTeam( agent, scenario:GetPlayerTeam(), SUPPORT_REASON.ALLY, LOC"SUPPORT_DESC.ALLY")
            end
        end
    end

    local enemy_leader
    if params.enemies then
        for i, cast in ipairs( params.enemies ) do
            local agent
            if is_instance(cast, Agent) then
                agent = cast
            else
                agent = self:GetCastAgent( cast )
                if agent == nil then
                    LOGWARN( "Unknown agent in enemies table: %s", cast )
                end
            end
            if agent and not agent:IsRetired() and scenario:GetTeamForAgent( agent ) == nil and not non_combat[agent] then
                enemy_leader = enemy_leader or agent
                scenario:AddAgentToTeam( agent, scenario:GetEnemyTeam() )

                if not agent:IsInPlayerParty() and agent:IsLeader() then
                    for k,party_agent in agent:Party() do
                        if party_agent ~= agent and party_agent:IsCrewMember() and party_agent:IsWithAgent( agent ) and not non_combat[party_agent] and  scenario:GetTeamForAgent( party_agent ) == nil then
                            scenario:AddAgentToTeam(party_agent, TEAM.RED, SUPPORT_REASON.PARTY, LOC"SUPPORT_DESC.SAME_PARTY")
                        end
                    end
                end

            end
        end
 
    elseif is_instance( self.primary_cast, Agent ) then
        scenario:AddAgentToTeam( self.primary_cast, TEAM.RED)
        enemy_leader = self.primary_cast
    end


    if scenario:AllowsBackup() then

        --figure out the relative strength of the teams before any joiners
        local blue_strength = 0
        for k,v in ipairs( scenario:GetTeam(TEAM.BLUE) ) do
            local str = v:GetCombatStrength() 
            blue_strength = blue_strength + str
        end
        local red_strength = 0
        for k,v in ipairs( scenario:GetTeam(TEAM.RED) ) do
            local str = v:GetCombatStrength() 
            red_strength = red_strength + str
        end
        local red = {}

        for k, agent in self.location:Agents() do
            if not non_combat[agent] and not scenario:GetTeamForAgent( agent ) then
                local self_defence = params.flags and CheckBits( params.flags, BATTLE_FLAGS.SELF_DEFENCE ) or false
                local no_bystanders = params.flags and CheckBits( params.flags, BATTLE_FLAGS.NO_BYSTANDERS ) or false
                local support, reason, txt = AgentUtil.GetCombatSupport(agent, enemy_leader, self_defence, no_bystanders)
                
                if support == SUPPORT_ACTION.HELP then
                    --don't add people that we've paid off to our team strength, because then others might add themselves in opposition
                    if reason ~= SUPPORT_REASON.MODIFIED then
                        blue_strength =  blue_strength + agent:GetCombatStrength()
                    end
                    scenario:AddAgentToTeam(agent, TEAM.BLUE, reason, txt)
                elseif support == SUPPORT_ACTION.HINDER then
                    table.insert(red, {agent, reason, txt})
                elseif support == SUPPORT_ACTION.SUPPRESSED then
                    table.insert(red, {agent, reason, txt})
                    scenario:AddBystander(agent, reason, txt)
                end
            end
        end
        
        
        local MAX_DIFF = params.oppo_limit or math.min( 5, 2 + TheGame:GetGameState():GetCurrentBaseDifficulty()) --this is a tuning value. maybe this should change as the game progresses, and more card picks have happened?
        local max_strength = blue_strength + MAX_DIFF
        local max_to_add = math.max(0, max_strength - red_strength)
        
        --we don't want to totally overwhealm the player, so stop adding after a point
        if max_to_add > 0 then
            --do a total sort ordering, so that this is the same every time this runs
            
            table.sort(red, function(a,b) 
                    if a[2] ~= b[2] then
                        --some reasons are more important than others
                        --print ("!!!!", a[2], b[2], CompareSupportReason(a[2],b[2]))
                        return CompareSupportReason(a[2],b[2])
                    else
                        local a_str, b_str = a[1]:GetCombatStrength(), b[1]:GetCombatStrength()
                        if a_str == b_str then
                            --default to guid, just so that we get a consistent ordering
                            return a[1]:GetID() < b[1]:GetID()
                        else
                            --add weak people first?
                            return a_str < b_str
                        end
                    end
                end)

            --now try to add people until we can't any more
            for k,info in ipairs(red) do
                local str = info[1]:GetCombatStrength()
                if params.no_oppo_limit or (max_to_add >= str and #scenario:GetTeam(TEAM.RED) < 4) then                
                    max_to_add = max_to_add - str
                    
                    if info[2] ~= SUPPORT_REASON.MODIFIED then
                        scenario:AddAgentToTeam(info[1], TEAM.RED, info[2], info[3])
                    end
                end
            end
        end
    end

    return scenario
end

function Encounter:GetCaravan()
    return self.caravan
end

function Encounter:GetDateTime()
    return Now()
end

function Encounter:GetConvoStateID()
    if is_instance( self.active_hub, ConvoStateGraph ) then
        return self.active_hub:GetContentID(), self.active_hub:GetStateID()
    end
end

function Encounter:GetScreen()
    if self.screen then
        return self.screen
    end

    -- ccc: This is an ugly fallback can I delete it?
    LOGWARN("Why is this fallback here?  Report repro steps")
    return TheGame:FE():FindScreen( Screen.ConversationScreen )
end

function Encounter:GetLocation()
    return self.location 
end

function Encounter:SetLocation( location )
    if self.location ~= location then
        self.location = location
        for i, hub in ipairs( self.hub_stack ) do
            hub.location = location
        end

        if self.screen and self.screen.RefreshPlax then
            self.screen:RefreshPlax()
        end
    end
end

function Encounter:GetLocationContent()
    return self.location:GetContent()
end

function Encounter:GetPrimaryCast()
    return self.primary_cast
end

function Encounter:GetAgent( cast )
    assert( cast == nil ) -- Encounter does not lookup cast members, Hubs do.
    if is_instance( self.primary_cast, Agent ) then
        return self.primary_cast
    end
end

function Encounter:GetCastAgent( cast )
    local agent = self:GetCastMember( cast )
    if is_instance( agent, Agent ) then
        return agent
    end
end

function Encounter:GetCastMember( cast )
    if cast:lower() == "other" then
        if self.current_speaker == self.left_agent then
            return self.right_agent
        elseif self.current_speaker == self.right_agent then
            return self.left_agent
        end

    elseif cast:lower() == "self" then
        return self.current_speaker

    elseif cast:lower() == "player" then
        return self:GetPlayer()
    end

    return self.active_hub:GetCastMember( cast )
end

function Encounter:GetRightAgent()
    return self.right_agent
end

function Encounter:RightAgents()
    return ipairs( self.right_agents )
end

function Encounter:GetLeftAgent()
    return self.left_agent
end

function Encounter:IsAgentPresent(agent)
    return table.arrayfind(self.left_agents, agent) or table.arrayfind(self.right_agents, agent)
end

function Encounter:LeftAgents()
    return ipairs( self.left_agents )
end

function Encounter:ShowText( txt, style )
    self:ShowRelationshipChanges()
    self:GetScreen():AddText( txt, style )
end

function Encounter:ShowImageBox( texpath )
    self:GetScreen():AddBubbleWidget( Widget.Image( engine.asset.Texture( texpath, true )))
end


function Encounter:PresentAgent( agent, side )
    
    local screen = self:GetScreen()
    if screen and screen.WaitOnLine then
        screen:WaitOnLine()
    end
    assert( agent == nil or is_instance( agent, Agent ))
    assert( side == nil or IsEnum( side, SIDE ))

    if agent then

        local prev_side = self:GetAgentSide( agent )
        local is_left = side == SIDE.LEFT or side == SIDE.LEFT_AUX
        local was_left = prev_side == SIDE.LEFT or prev_side == SIDE.LEFT_AUX

        if prev_side == side then
            return -- Already here.
        elseif (is_left ~= was_left) or (prev_side and side == nil) then
            
            self:ExitAgent( agent )
        end


        
        if side == SIDE.LEFT or side == SIDE.RIGHT then
            if not agent:IsPlayer() then
                TheGame:GetGameProfile():SetMetCharacter( agent:GetUniqueID() )
                self.presented_agents[agent] = true
            end
        end

        if side == SIDE.LEFT then
            -- Present as left-side speaker.
            table.insert_unique( self.left_agents, agent )
            self.left_agent = agent        
            for i, hub in ipairs( self.hub_stack ) do
                hub:ReassignCastMember( "left", agent )
            end
            if screen then
                screen:AddCharacter( agent, true )
            end

        elseif side == SIDE.LEFT_AUX then
            -- Present on left side.
            table.insert_unique( self.left_agents, agent )
            if self.left_agents == agent then
                self.left_agent = nil
            end
            if screen then
                screen:Refresh()
            end

        elseif side == SIDE.RIGHT then
            -- Present as right-side speaker.
            if self.right_agent ~= agent then
                table.insert_unique( self.right_agents, agent )
                self.right_agent = agent
                for i, hub in ipairs( self.hub_stack ) do
                    hub:ReassignCastMember( "right", agent )
                end
                if screen then
                    screen:AddCharacter( agent, false )
                end
            end

        elseif side == SIDE.RIGHT_AUX then
            -- Present on right side.
            table.insert_unique( self.right_agents, agent )
            if screen then
                screen:Refresh()
            end
        end

    elseif side == SIDE.LEFT then
        if self.left_agent then
            self:ExitAgent( self.left_agent )
        end
 
    elseif side == SIDE.RIGHT then
        if self.right_agent then
            self:ExitAgent( self.right_agent )
        end

    else
        --error( side ) -- Illegal parameters.
    end
end

function Encounter:RemoveAllAgents()
    local all_agents = {}
    for k,v in ipairs(self.left_agents) do
        table.insert(all_agents, v)
    end
    for k,v in ipairs(self.right_agents) do
        table.insert(all_agents, v)
    end
    for k,v in ipairs(all_agents) do
        self:ExitAgent(v)
    end


end

function Encounter:ExitAgent( agent )
    assert( agent )
    local screen  = self:GetScreen()

    table.arrayremove( self.left_agents, agent )
    table.arrayremove( self.right_agents, agent )

    if self.left_agent == agent then
        if screen then
            screen:RemoveCharacter( nil, true )
        end
        self.left_agent = nil
        for i, hub in ipairs( self.hub_stack ) do
            hub:ReassignCastMember( "left", nil )
        end

    elseif self.right_agent == agent then
        if screen then
            screen:RemoveCharacter( nil, false )
        end
        self.right_agent = nil
        for i, hub in ipairs( self.hub_stack ) do
            hub:ReassignCastMember( "right", nil )
        end

    else
        if screen then
            screen:Refresh()
        end
    end
end

function Encounter:RemoveDeadAgents()
    local deads = {}
    for k,v in pairs(self.left_agents) do
        if v:InLimbo() then
            table.insert(deads, v)
        end
    end
    for k,v in pairs(self.right_agents) do
        if v:InLimbo() then
            table.insert(deads, v)
        end
    end

    for k,v in ipairs(deads) do
        if v == self.right_agent or v == self.left_agent then
            self:Emote( v, "injured" )
        end
        self:ExitAgent(v)
    end

    -- NOTE: do not clear primary cast if they are dead, many convos rely on it still being non-nil.
end

function Encounter:GetAgentSide( agent )
    assert( agent )
    if self.left_agent == agent then
        return SIDE.LEFT
    elseif self.right_agent == agent then
        return SIDE.RIGHT
    elseif table.arrayfind( self.left_agents, agent ) then
        return SIDE.LEFT_AUX
    elseif table.arrayfind( self.right_agents, agent ) then
        return SIDE.RIGHT_AUX
    end
end

-- Join the left side and become the left speaker.
function Encounter:PresentLeft( agent )
    self:PresentAgent( agent, SIDE.LEFT )
end

-- Join the right side and become the right speaker.
function Encounter:PresentRight( agent )
    self:PresentAgent( agent, SIDE.RIGHT )
end

-- Join the left side; do not affect left speaker.
function Encounter:JoinLeft( agent )
    self:PresentAgent( agent, SIDE.LEFT_AUX )
end

-- Join the right side; do not affect right speaker.
function Encounter:JoinRight( agent )
    self:PresentAgent( agent, SIDE.RIGHT_AUX )
end

-- Present left and right agents; either may be nil to clear the side.
function Encounter:PresentSides( left_agent, right_agent )
    self:PresentAgent( left_agent, SIDE.LEFT )
    self:PresentAgent( right_agent, SIDE.RIGHT )
end

function Encounter:PresentDefault()
    if is_instance( self.primary_cast, Agent ) then
        self:PresentSides( self:GetPlayer(), self.primary_cast )
    end
end 

function Encounter:Wait(time)
    if time then
        if time > 0 then
            self.wait_delay = time
            self:YieldEncounter()
        end
    else
        self:GetScreen():WaitOnLine()
    end
end

function Encounter:WaitOnLine()
    if self.screen then
        if self.screen.WaitOnLine then
            self.screen:WaitOnLine()
        end
    end
end

function Encounter:WaitOnWidget( widget )
    self:YieldEncounter( function() return not widget:IsUpdating() end )
end

function Encounter:DoFade( duration )
    if TheGame:GetLocalSettings().ROBOTICS then
        return
    end
    
    self:WaitOnLine()

    local done = false
    TheGame:FE():FadeTransition( function() done = true end, duration or 1.0 )
    self:YieldEncounter( function() return done end )
end

function Encounter:ClearOptions()
    table.clear( self.current_options )
    self:GetScreen():ClearOptions()
end

function Encounter:CountOptions()
    return #self.current_options
end

function Encounter:AddOption( option, data )
    option:AttachHub( self.active_hub )
    option.data = data
    table.insert( self.current_options, option )
end

function Encounter:RemoveOption( option )
    table.arrayremove( self.current_options, option )
end

function Encounter:ChooseOption( options_type )
    if #self.current_options == 0 then
        --print( "WARN: no options", debug.traceback() )
        return nil
    end

    local function OnSelectOption( idx )
        local opt = idx and self.current_options[ idx ] 

        -- ccc: This is critical to clear options once a given hub has executed them:
        --  fallback to a previous hub will otherwise re-use the old options redundantly
        -- (examine carefully the flow from ConvoStateGraph:RunState when calling the state handler 'fn')
        self:ClearOptions()
        self:ResumeEncounter( opt )
    end
    self:GetScreen():PresentOptions( self.current_options, options_type, OnSelectOption )

    return self:YieldEncounter()
end

function Encounter:AddHubOptions( mask, hub_state )
    hub_state = hub_state or HUB_STATE.HUB
    mask = mask or 0
    
    local hubs = {} -- this is not an obtainworktable beacuse when using location hubs, we sometimes do crazy things which leaves them stranded.
    self:CollectHubs( hubs, hub_state, mask )

    for i, context in ipairs( hubs ) do
        self:RunHub( context, hub_state )
    end
    
    if hub_state == HUB_STATE.LOCATION then
        if LocationUtil.CanLeaveLocation(self:GetLocation()) then        
            local exit = self:GetLocation():GetDefaultExit()
            
            local leave_txt
            local leave_sfx
            if exit and exit:GetTo() then
                -- Leaving a character convo
                leave_txt = LOC"ENCOUNTER.LEAVE"
                leave_sfx = SoundEvents.leave_location
            else
                -- Leaving a location
                leave_txt = LOC"ENCOUNTER.TRAVEL"
                leave_sfx = SoundEvents.leave_location
            end
            local leave_opt = ConvoOption( leave_txt )
            leave_opt.underfooter = true

            self:AddOption( leave_opt, { sfx = leave_sfx } )
            -- HAX for being invoked from convoStateGraph.
            if leave_opt.hub.Pop then
                leave_opt:Pop()
            end
            
            return leave_opt
        else
            print( "HUB_STATE.LOCATION : cannot leave", self:GetLocation() )
        end
    end
end

function Encounter:ClearCurrentSpeaker( style )
    self.current_speaker = nil
    self.textbox_style = style
end

function Encounter:UpdateCurrentSpeaker( agent )
    if agent then
        assert( is_instance( agent, Agent ))
        self.current_speaker = agent
    end
    return self.current_speaker
end

function Encounter:LookupQuip( agent, ... )
    table.clear( self.quip_tags )
    local tags = self.quip_tags

    for i = 1, select( "#", ... ) do
        local tag = select( i, ... )
        if tag then
            table.insert( tags, tag:lower() )
        end
    end

    if agent then
        agent:FillOutQuipTags(tags)
    end
    
    local seed = table.contains( tags, "seeded" ) and agent and agent:GetID()
    local quip

    if self.negotiation then
        quip = self.negotiation:LookupQuip(tags, seed)
    end

    if quip == nil and agent then
        quip = agent:LookupQuip( tags, seed )
    end

    if quip == nil and self.active_hub then
        quip = self.active_hub:LookupQuip( tags, seed )
    end

    if quip == nil then
        quip = Content.LookupQuip( tags, seed )
    end

    return quip
end

function Encounter:SayQuip( agent, ... )
    agent = self:UpdateCurrentSpeaker( agent )

    local quip = self:LookupQuip( agent, ... )

    if quip then
        if agent and not (self.left_agent == agent or self.right_agent == agent) then
            if table.arrayfind(self.left_agents, agent) then
                self:PresentLeft(agent)
            else
                self:PresentRight(agent)
            end
        end

        self:Dialog( self:LocFormat( quip ))
        return true
    else
        print( "MISSING QUIP:", agent, ... )
        return false
    end
end

function Encounter:LocFormat( text, ... )
    local params = {...}
    local function Formatter( word, operator, rest )
        local locobj = self.active_hub and self.active_hub:SubstituteFormatter( word, operator )
        if locobj == nil then
            locobj = self:GetCastMember( word )
        end
        if not locobj then
            local num = tonumber(word)
            if num and params[num] then
                locobj = params[num]
            end
        end
        
        if locobj == true then
            return rest
        elseif locobj == false then
            return ""
        elseif locobj ~= nil then
            local inner_params = shallowcopy(params)
            
            if is_instance( locobj, Agent ) then
                if operator == "#" then
                    table.insert(inner_params, locobj)
                    return loc.format( string.format( "{%d%s%s}", #inner_params, operator, rest ), table.unpack(inner_params)  )
                elseif operator then
                    table.insert(inner_params, locobj:LocTable())
                    return loc.format( string.format( "{%d%s%s}", #inner_params, operator, rest ), table.unpack(inner_params) )
                else
                    table.insert(inner_params, locobj)
                    return loc.format( string.format("{%d#agent}", #inner_params), table.unpack(inner_params) )
                end
            elseif is_instance( locobj, Location ) then
                table.insert(inner_params,locobj)
                return loc.format( string.format("{%d#location}", #inner_params),table.unpack(inner_params) )
            else
                table.insert(inner_params,locobj)
                return loc.format( string.format("{%d%s%s}", #inner_params,operator, rest), table.unpack(inner_params))
            end
        else
            if operator == ".gender:" then
                return loc.format( "{1:"..rest.."}", 1 )
            end
            -- print( "NO LOCOBJ:", word, operator, rest )
        end
    end

    return loc.custom_format( text, Formatter )
end

function Encounter:Dialog( str )
        
    local lines = str:split("\n")
    for i, line in ipairs(lines) do
        line = line:trim()
        local speaker = string.match(line, "^([_%w]*):$")
        local narration = string.match(line, "^[*]%s*(.+)") -- shorthand for Narrator: that doesn't change current speaker
        local tutorial = string.match(line, "^[*][*]([^*]%s*.+)") -- shorthand for Tutorial: that doesn't change current speaker
        local tldr = string.match(line, "^[*][*][*]%s*(.+)") -- shorthand for TLDR: that doesn't change current speaker
        local emote = string.match(line, "^!([^%s]+)")
        local voice = string.match(line, "^$([^%s]+)")
        local quip = string.match(line, "^[%%](.+)")
        local sound_event = string.match(line, "^[#]%s*(.+)%s*")

        if speaker then
            speaker = speaker:lower()
            if speaker == "narrator" then
                self:ClearCurrentSpeaker( Widget.TextBox.STYLE.PLAIN )
            elseif speaker == "tutorial" then
                self:ClearCurrentSpeaker( Widget.TextBox.STYLE.TUTORIAL )
            elseif speaker == "tutorial" then
                self:ClearCurrentSpeaker( Widget.TextBox.STYLE.TLDR )
            else
                self:UpdateCurrentSpeaker( self:GetCastAgent( speaker ))
            end
        elseif voice then
            self:SetVoice( nil, voice)
        elseif tldr then
            if self:GetScreen():IsAutoSkip() then
                self:ShowText( tldr, Widget.TextBox.STYLE.TLDR )
            end
        elseif tutorial then
            tutorial = tutorial:trim()
            self:ShowText( tutorial, Widget.TextBox.STYLE.TUTORIAL )
        elseif narration then
            self:ShowText( narration )
        elseif emote then
            self:Emote( nil, emote )

        elseif quip then
            self:SayQuip( nil, table.unpack( quip:split() ))

        elseif sound_event then
            if string.match(sound_event, "^event:/music") then
                TheGame:GetMusic():PlayMusic(sound_event)
            else
                AUDIO:PlayEvent(sound_event)
            end
        elseif #line > 0 then
            self:SayText( nil, line )
        end
        
    end
end

function Encounter:SayText(agent, text)
    agent = self:UpdateCurrentSpeaker( agent )
    if agent then
        self:GetScreen():AddSpeech( agent, text )
    else
        self:GetScreen():AddText( text, self.textbox_style )
    end
end

function Encounter:ShowPreformattedText(text)
    self:ClearCurrentSpeaker( Widget.TextBox.STYLE.PLAIN )
    self:ShowText( text )
end

function Encounter:SetVoice( agent, voice )
    agent = self:UpdateCurrentSpeaker( agent )
    self:GetScreen():SetVoice(agent, voice)
end


function Encounter:Emote( agent, emote )
    agent = self:UpdateCurrentSpeaker( agent )
    
    if emote == "left" then
        self:PresentAgent( agent, SIDE.LEFT )
    elseif emote == "join_left" then
        self:PresentAgent( agent, SIDE.LEFT_AUX )
    elseif emote == "right" then
        self:PresentAgent( agent, SIDE.RIGHT )
    elseif emote == "join_right" then
        self:PresentAgent( agent, SIDE.RIGHT_AUX )
    elseif emote == "exit" then
        self:PresentAgent( agent )
    elseif emote == "reset" then
        self:PresentDefault()
    elseif emote == "fade" then
        self:DoFade( 1.0 )
    elseif emote == "wait" then
        self:Wait( 1.0 )
    elseif emote == "shake" then
        self:GetScreen():ScreenShake( CONVO_SHAKE )
    else
        self:GetScreen():WaitOnLine()
        self:GetScreen():Emote( agent, emote )
    end
end

function Encounter:PlayerDeath(killed_by)
    TheGame:Lose(killed_by)
    self:End()
end

function Encounter:End()
    self:WaitOnLine()
    self.is_ended = true
    
    for agent in pairs(self.presented_agents) do
        if agent.social_connections then
            if not agent.social_connections:FindOpinionEvent( TheGame:GetGameState():GetPlayerAgent(), OPINION.MET ) then
                agent:OpinionEvent(OPINION.MET)
            end
        end

    end

end

function Encounter:AwardLoot(loot, data)
    data = data or table.empty

    self:WaitOnLine()

    TheGame:BroadcastEvent( "award_loot", loot, data )
    
    local function OnDone()
        self:ResumeEncounter()
    end
    local screen = Screen.LootScreen( OnDone, data )

    if not data.defeat and not loot.no_draft then
        if loot.is_boss_fight then
            local battle_cards = RewardUtil.GetBossCards( TheGame:GetGameState():GetPlayerAgent(), BattleCardCollection)
            local negotiation_cards = RewardUtil.GetBossCards( TheGame:GetGameState():GetPlayerAgent(), NegotiationCardCollection)
            screen:AddBattleCardChoice( battle_cards )
            screen:AddNegotiationCardChoice( negotiation_cards )
        else
            if loot.battle_card_draft_difficulty then
                local cards = RewardUtil.GetBattleCards( loot.battle_card_draft_difficulty, 3, TheGame:GetGameState():GetPlayerAgent())
                screen:AddBattleCardChoice( cards )
            end
            if loot.negotiation_card_draft_difficulty then
                local cards = RewardUtil.GetNegotiationCards( loot.negotiation_card_draft_difficulty, 3, TheGame:GetGameState():GetPlayerAgent())
                screen:AddNegotiationCardChoice( cards )
            end
            if loot.battle_card_swap and loot.battle_card_swap_difficulty then
                local drafts = {}
                for i=1, loot.battle_card_swap do
                    table.insert(drafts, RewardUtil.GetBattleCards( loot.battle_card_swap_difficulty, 3, TheGame:GetGameState():GetPlayerAgent()))
                end
                screen:AddBattleCardSwap( loot.battle_card_swap, drafts )
            end
        end
    end
    
    screen:AddUpgradeReminders()

    if loot.mettle and loot.mettle > 0 then
        screen:AddMettle(loot.mettle)
    end

    if loot.bonus_mettle and loot.bonus_mettle > 0 then
        screen:AddMettle(loot.bonus_mettle, true)
    end

    if loot.items then
        for k,v in ipairs(loot.items) do
            screen:AddCard(v)
        end
    end
    if loot.grafts then
        for k,v in ipairs(loot.grafts) do
            screen:AddGraft(v)
        end
    end

    if (loot.money or 0) > 0 then
        screen:AddMoney( loot.money )
    end
    if (loot.stolen_money or 0) > 0 then
        screen:AddMoney(loot.stolen_money, loc.format( LOC "UI.LOOT_SCREEN.LOOT_STOLEN_MONEY", loot.stolen_money ))
    end

    if loot.resolve and loot.resolve > 0 then
        screen:AddResolve(loot.resolve)
    end

    if loot.health and loot.health > 0 then
        screen:AddHealth(loot.health)
    end

    if not screen:IsEmpty() then
        TheGame:FE():InsertScreen( screen )
        self:YieldEncounter()
    end
end

function Encounter:DoBattle( hub, scenario )
    
    self:ShowRelationshipChanges()
    self:GetScreen():ForceWaitOnLine()
    
	local initial_player_state = TheGame:GetGameState():GetPlayerState()

    scenario = scenario or self:CreateScenario()
    print(scenario:GetStringDump())
    TheGame:GetGameState():SetLastBattleScenario(scenario)

    local battle = BattleEngine( scenario )

    for k,fighter in battle:GetTeam(scenario:GetEnemyTeam()):AllFighters() do
        if fighter.agent:IsInPlayerParty() and not fighter.agent:IsCaptiveMember() then
            fighter.agent:Dismiss(true)
        end
    end

    local other_primary = battle:GetTeam(TEAM.RED):Primary():GetAgent()
    local num_enemies = battle:GetTeam(scenario:GetEnemyTeam()):NumFighters()

    TheGame:GetGameState():GetNotifications():SetBannersEnabled(false)
    local overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if overlay then
        overlay:SetModeOverride( OVERLAY_MODE.BATTLE )
    end


    -- ccc: this doesn't work with PostProcessed plaxes. Also we need to fade anyway, so that we can do loading can happen during darkness.
    -- TheGame:FE():InkTransition(5.5)

    -- INITIATE BATTLE --
    ---------------------
    local result, repercussions = nil, table.empty
    local screen

    if TheGame:GetLocalSettings().ROBOTICS then
        result = BattleUtil.ResolveBattleAutomatically( battle, TheGame:GetRoboticsSettings().battle_policy )

    elseif TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
        print( "AUTO-RESOLVE" )
        result = BattleUtil.ResolveBattleAutomatically( battle, "SPARE" )

    else
        local function OnDoneBattle( result )
            self:ResumeEncounter( result )
        end
        
        screen = Screen.FightScreen( battle, OnDoneBattle )
        TheGame:FE():FadeTransition( function()
            TheGame:FE():InsertScreen( screen )
        end )
        screen.plax:SetDayPhase( TheGame:GetGameState():GetDayPhase() )
        
        -- Yield until battle is complete.
        result = self:YieldEncounter()
    end

    -- POST BATTLE HANDLING --
    --------------------------

    if overlay then
        overlay:SetModeOverride( nil )
    end
    TheGame:GetGameState():GetNotifications():SetBannersEnabled(true)        

    -- Reset overlay mode before FightScreen actually gets closed, to permit things like upgrading.
    if scenario.params.IS_EXPERIMENT then
        if screen then
            TheGame:FE():PopScreen( screen )
        end

    elseif result == BATTLE_RESULT.LOST then
        self:PlayerDeath( other_primary )

    else
        repercussions = BattleUtil.ResolveBattle( battle, other_primary )
        
        self:ShowRelationshipChanges()

        if self.right_agent == other_primary and other_primary:IsDead() then
            self:ExitAgent( self.right_agent )
        end

        if result == BATTLE_RESULT.WON then
            TheGame:GetGameState():AddScore("BATTLES_WON", 1)
            
            if not CheckBits( scenario:GetFlags(), BATTLE_FLAGS.NO_REWARDS ) then
                local difficulty = scenario:GetRewardDifficulty() or
                    TheGame:GetGameState():GetCurrentBaseDifficulty()

                repercussions.loot.battle_card_draft_difficulty = difficulty
                repercussions.loot.is_boss_fight = scenario:IsBossFight()

                if TheGame:GetGameState():CanDoMettle() then
                    repercussions.loot.mettle = BATTLE_METTLE_REWARD[difficulty] or 0
                    local bonus_mettle = battle:GetPlayerFighter():GetConditionStacks("METTLE_BANK")
                    if bonus_mettle and bonus_mettle > 0 then
                        repercussions.loot.bonus_mettle = bonus_mettle
                    end
                end
                
                local resolve, resolve_max = TheGame:GetGameState():GetCaravan():GetResolve()
                repercussions.loot.resolve = math.ceil( (resolve_max - resolve) * GetAdvancementModifier( ADVANCEMENT_OPTION.BATTLE_RESOLVE_RESTORE ))
                repercussions.loot.no_draft = CheckBits( scenario:GetFlags(), BATTLE_FLAGS.NO_DRAFT )
                self:AwardLoot( repercussions.loot, { battle=true } )
            end

            TheGame:AddGameplayStat( "battles_won", 1 )

        elseif result == BATTLE_RESULT.RUNAWAY then
            if self.right_agent then
                self:ExitAgent( self.right_agent )
            end
            if self.left_agent then
                self:ExitAgent( self.left_agent )
            end
        end

        if screen then
            if scenario.params.post_fade then
                TheGame:FE():FadeTransition(function()
                    TheGame:FE():PopScreen( screen )
                    self:ResumeEncounter() end, 1.0 )
                self:YieldEncounter()
            else
                TheGame:FE():PopScreen( screen )
            end
        end

        assert( not self.screen.removed )
        self:GetScreen():Refresh()
        self:GetScreen():ClearHistory()
        self:GetScreen():ForceWaitOnLine() -- Must go before resolving, to clear fastmode.
        self:RemoveDeadAgents()
    end

    -- POST BATTLE METRICS --
    --------------------------

	local player_team = {}
    local enemy_team = {}
    for k,v in ipairs(scenario.teams[TEAM.BLUE]) do
        table.insert(player_team, v:GetContentID())
    end

    for k,v in ipairs(scenario.teams[TEAM.RED]) do
        table.insert(enemy_team, v:GetContentID())
    end

    local player_state = TheGame:GetGameState():GetPlayerState()
	local json_t = {
		PLAYER_DATA = TheGame:GetGameState():GetPlayerState(),
		BATTLE_DATA = {
			TURNS = battle.turns,
			NUM_ENEMIES = num_enemies,
			CONTENT_ID = scenario.content_id,
			--DIFFICULTY = scenario:CalculateDifficulty(),
			HEALTH_DELTA = player_state.HEALTH - initial_player_state.HEALTH,
            PLAYER_TEAM = player_team,
            ENEMY_TEAM = enemy_team,
		}
	}

	local evt_type = (result == BATTLE_RESULT.WON) and "BATTLE_WON" or "BATTLE_LOST"
	SendMetricsData(evt_type, json_t )
    
    TheGame:GetGameState():Action()

    return result, battle, repercussions
end

function Encounter:DoTravel(must_leave)
    
    local screen, idx = TheGame:FE():FindScreen( self.screen )
    if self.screen and self.screen.plax then
        self.screen.plax:PlayExitSound()
    end

    
    TheGame:FE():FadeTransition(function()
        TheGame:FE():InsertScreen( Screen.TravelScreen(must_leave), idx )
        
        --if self.coro then
            self:ResumeEncounter()
        --end

    end, .5)

    self:YieldEncounter()
    self:End()
end



function Encounter:ShowSlides( slide_deck, flags, on_done )
    if on_done == nil then
        on_done = function() self:ResumeEncounter() end
    end
    TheGame:FE():FadeTransition( function()
        TheGame:FE():PushScreen( Screen.SlideshowScreen( slide_deck, on_done, flags) )
    end, 1.5 )
    self:YieldEncounter()
end

function Encounter:GetPlayer()
    return TheGame:GetGameState():GetPlayerAgent()
end

function Encounter:GetPlayerMoney()
    return TheGame:GetGameState():GetCaravan():GetMoney()
end

function Encounter:LoseMoney( amount, to_whom )
    if amount ~= 0 then
        local caravan = TheGame:GetGameState():GetCaravan() 
        caravan:PayMoney( amount, to_whom )
    end
end

function Encounter:GainMoney( amount )
    if amount ~= 0 then
        AUDIO:PlayEvent(SoundEvents.conversation_accept_money)
        local caravan = TheGame:GetGameState():GetCaravan() 
        caravan:AddMoney( amount )
    end
end

function Encounter:ShowLoseAmount( id, amount )
    TheGame:GetGameState():LogNotification( NOTIFY.LOST_ITEM, id, amount )
end


function Encounter:OpinionEvent( opinion, data, agent )
    assert( opinion )
    agent = agent or self:GetAgent()
    agent:OpinionEvent( opinion, data )    
end

--------------------------------------------------------------------------

function Encounter:CollectQuestHubs( hubs, state, mask)
    
    --asides are mechanical submenus (hostile action, manipulate). We want to show quest actions at the root, so filter them out here
    if state == HUB_STATE.HUB and ( mask & CONVERSATION_FLAGS.ASIDE ~= 0) then
        return
    end

    for i, quest in TheGame:GetGameState():ActiveQuests() do
        local hub
        local cast_id
        if quest:GetQuestDef() then
            for i, questdef in ipairs( quest:GetQuestDef():GetConvoStateGraph() ) do

                if state ~= HUB_STATE.HUB or ( (questdef.req_mask & mask == questdef.req_mask) and (questdef.dis_mask & mask == 0) and  (questdef.any_mask == 0 or questdef.any_mask & mask ~= 0 )) then
                    if questdef.objective_id == nil or quest:GetObjectiveState( questdef.objective_id ) == QSTATUS.ACTIVE then
                        if questdef.cast_id == nil or quest:GetCastMember(questdef.cast_id) == self.primary_cast then
                            local hub = ConvoStateGraph( self, questdef, nil, quest, mask )
                            if hub:TryValidateHub( self.primary_cast, state ) then
                                table.insert( hubs, hub )
                            end
                        end
                    end
                end
            end
        end
    end

    table.sort( hubs, function( a, b ) return a:GetPriority() > b:GetPriority() end )
end

function Encounter:CollectConvoHubs( hubs, state, mask, ... )
    local gs = TheGame:GetGameState()
    local act_data = gs:GetCurrentAct()
    local filter_fn = act_data.data.convo_filter_fn

    for i, convodef in Content.IterateConvoStateGraphs( state ) do
        if CheckBits( convodef.flags, CONVO_FLAGS.QUEST ) then
            -- Don't add quest convos here.
        else
            --print ( mask, convodef.req_mask, convodef.dis_mask, (convodef.req_mask & mask == convodef.req_mask) , (convodef.dis_mask & mask == 0))
            if state ~= HUB_STATE.HUB or
                    ( (convodef.req_mask & mask == convodef.req_mask) and
                    (convodef.dis_mask & mask == 0) and
                    (convodef.any_mask == 0 or convodef.any_mask & mask ~= 0) ) then
                --
                if filter_fn == nil or filter_fn( convodef, gs ) then
                    local hub = ConvoStateGraph( self, convodef, nil, nil, mask )
                    if hub:TryValidateHub( self.primary_cast, state, ... ) then
                        table.insert( hubs, hub )
                    end
                end
            end
        end
    end

    table.sort( hubs, function( a, b ) return a:GetPriority() > b:GetPriority() end )
end

function Encounter:CollectHubs( hubs, state, mask )
    self:CollectQuestHubs( hubs, state, mask )
    self:CollectConvoHubs( hubs, state, mask )
    return hubs
end

function Encounter:DoLocationTransition(loc)
    if is_instance( self:GetScreen(), Screen.ConversationScreen ) then
        self:GetScreen():ForceWaitOnLine() 
        self:GetScreen():ClearHistory()
    end

    self:DoFade(.4)

    self:SetPrimaryCast( nil )
    self:SetLocation( loc )
    self.caravan:MoveToLocation( loc )

    local screen = TheGame:FE():FindScreen( Screen.LocationScreen )
    if screen then
        screen:UpdateDayPhase()
        screen:RefreshPlaxClickables(true)
    end
    
    return loc
end


function Encounter:ShowJobOffer(quest, objective_id, must_accept)
    self:WaitOnLine()

    quest:VerifyRewards()

    local popup = Screen.AcceptJobPopup( quest, must_accept )
    popup:SetFn( function(accepted) 
        if self:GetScreen().SetAutoSkip then
            self:GetScreen():SetAutoSkip( false )
        end
        self:ResumeEncounter( accepted ) 
        self:GetScreen():NextLine()
    end )
    TheGame:FE():InsertScreen( popup )
        
    local accepted = self:YieldEncounter()

    if accepted then
        if objective_id then
            quest:SetObjectiveState( objective_id, QSTATUS.ACTIVE )
        end


        if quest:HasUpfrontReward() then
            self:Wait()

            local draft_popup = Screen.DraftChoicePopup()
            local function OnDone()
                self:ResumeEncounter()
            end

            local cards = {}
            local player = self:GetPlayer()

            if quest:GetFocus() == QUEST_FOCUS.COMBAT then
                cards = RewardUtil.GetBattleCards( 1, 3, player )
                draft_popup:DraftCards( player, Battle.Card, cards, OnDone )

            elseif quest:GetFocus() == QUEST_FOCUS.NEGOTIATION then
                cards = RewardUtil.GetNegotiationCards( 1, 3, player )
                draft_popup:DraftCards( player, Negotiation.Card, cards, OnDone )
            else
                cards = RewardUtil.GetMixedCards( 1, 3, player )
                draft_popup:DraftCards( player, nil, cards, OnDone )
            end

            TheGame:FE():InsertScreen( draft_popup )

            self:YieldEncounter()

            self:GetScreen():ForceWaitOnLine()
        end
    end
    
    return accepted
end


function Encounter:OfferCards(cards, title, subtitle, no_skip)
    self:WaitOnLine()
    local draft_popup = Screen.DraftChoicePopup()
    draft_popup:DraftCards( nil, nil, cards, function() self:ResumeEncounter() end, title, subtitle, no_skip )
    TheGame:FE():InsertScreen( draft_popup )
    self:YieldEncounter()
end


function Encounter:TryInstallGraft(graft)
    self:WaitOnLine()

    if TheGame:GetGameState():GetPlayerAgent().graft_owner:CanInstallGraft(graft.id) then
        TheGame:GetGameState():GetPlayerAgent().graft_owner:AddGraft(graft)
        return true
    else
        local screen = Screen.ReplaceGraftScreen( graft, 
            function(replaced) 
                self:ResumeEncounter(replaced)
            end )
        TheGame:FE():InsertScreen( screen )

        local installed = self:YieldEncounter()
        return installed
    end
end


function Encounter:OnGlobalEvent( event_name, ... )
    if event_name == "agent_relationship_changed" then
        local agent, old_rel, new_rel = ...
        if new_rel == RELATIONSHIP.LOVED then
            table.insert(self.pending_relationship_changes, agent)
        elseif new_rel == RELATIONSHIP.HATED then
            table.insert(self.pending_relationship_changes, agent)
        end
    end
end

function Encounter:ShowRelationshipChanges()
    if #self.pending_relationship_changes > 0 then
        local agents = shallowcopy(self.pending_relationship_changes)
        table.clear(self.pending_relationship_changes)

        if coroutine.running() == self.coro then
            local screen = Screen.SocialGraftScreen( agents, function() self:ResumeEncounter() end )
            TheGame:FE():InsertScreen( screen )
            self:YieldEncounter()
        else
            local screen = Screen.SocialGraftScreen( agents )
            TheGame:FE():InsertScreen( screen )
        end

    end
end
