class("Entity")

require "components/colourtweener"
require "eventsystem"

local entity_impl_mt =
{
    -- Looks up the first component implementing a function k, and caches a thunk if found.
    __index = function( t, k )
        for i, cmp in ipairs(t.cmp_sorted) do
            if type(cmp[k]) == "function" then
                rawset( t, k, function( entity, ... ) return cmp[k]( cmp, ... ) end )
                return t[k]
            end
        end
        -- Doesn't exist.
        rawset( t, k, false )
    end
}

local function ComponentPrioritySort(a, b)
    local pa, pb = a.Priority or DEFAULT_COMPONENT_PRIORITY, b.Priority or DEFAULT_COMPONENT_PRIORITY
    if pa == pb then
        return a._classname < b._classname
    else
        return pa < pb
    end
end

local function CreateEntityInterface( entity )
    local t = { cmp_sorted = entity:GetSortedComponents() }
    return setmetatable( t, entity_impl_mt )
end


function Entity:init(sim, prefab, spawn_data)
    assert(spawn_data, "spawn_data is required, may be empty")

    self.sim = sim
    self.prefab = prefab
    self.spawn_data = spawn_data
    self.guid = spawn_data.guid or sim:GetNextGUID()
    self.entity_name = spawn_data.entity_name
    self.cmp = {}
    self.cmp_sorted = {}
    self.oldest_event = 1
        
    if spawn_data.tags then
        
        if type(spawn_data.tags) == "string" then
            self:SetTagsFromString(spawn_data.tags)
        else
            for k, v in pairs(spawn_data.tags) do
                self:AddTag(v)    
            end
        end
    end

    if self.prefab.def and self.prefab.def.tags then
        for tag, val in pairs( self.prefab.def.tags ) do
            self.sim:AddToBucket( self, tag )
        end
    end

    self.spawnTime = self.sim:GetTime()

end

function Entity:GetSaveData()
    error()
end

function Entity:IsPersistent()
    return self.prefab.def and self.prefab.def.savable
end

function Entity:Hide()
    if not self.hidden then 
        self.hidden = true
        if self.transform_provider then
            self.transform_provider:Hide()
        end
    end
end

function Entity:Show(show)
    if show == false then
        return self:Hide()
    end
    if self.hidden then
        self.hidden = false
        if self.transform_provider then
            self.transform_provider:Show()
        end
    end
end

function Entity:IsHidden()
    return self.hidden
end

function Entity:GetName()
    -- a name is an assigned lookup label, not a localized string.
    return self.entity_name
end

function Entity:SetName( name )
    assert_warning(name == nil, "Entity:SetName() is deprecated [%s]. If you have a reason to use it, remove this warning.", name)
    -- a name is an assigned lookup label, not a localized string.
    self.entity_name = name
end

function Entity:AddTag(tag)
    self.tags = self.tags or {}
    self.tags[tag] = tag
    if self.sim then
        self.sim:AddToBucket( self, tag )
    end
end

function Entity:ClearTags()
    if self.tags then
        for k,v in pairs(self.tags) do
            self.sim:RemoveFromBucket( self, k )
        end

        table.clear(self.tags)
    end
end

function Entity:RemoveTag(tag)
    if self.tags then
        self.tags[tag] = nil
        if self.sim then
            self.sim:RemoveFromBucket( self, tag )
        end
    end
end

function Entity:HasTag(tag)
    if self.tags and self.tags[tag] then
        return true
    elseif (self.prefab.def and self.prefab.def.tags and self.prefab.def.tags[tag] ~= nil) then
        return true
    end
    return false
end

local tag_string = {}
function Entity:GetTagString()
    if self.tags then
        for tag, _ in pairs( self.tags ) do
            table.insert( tag_string, tag)
        end
    end

    local s = table.concat(tag_string, " ")
    table.clear(tag_string)

    return s
end

function Entity:SetTagsFromString( str )
    -- print(self, "Entity:SetTagsFromString( )", str )
    self:ClearTags()

    if str then
        for i, tag in ipairs(str:split()) do
            if #tag > 0 then
                self:AddTag(tag)
            end
        end
    end
end

function Entity:GetGUID()
    return self.guid
end

function Entity:__tostring()
    local name = self.prefab and self.prefab.name or self.prefab_name_override or "NO PREFAB"
    if self.entity_name then
        return string.format("[%d (%s) - %s]", self.guid, self.entity_name, name)
    else
        return string.format("[%d - %s]", self.guid, name)
    end
end


function Entity:IsReady()
    return self.entity_ready and not self.removed
end



function Entity:GetImpl()
    if not self.impl then
        self.impl = CreateEntityInterface( self )
    end
    return self.impl
end

function Entity:GetSortedComponents()
    if self.sort_dirty then
        table.clear(self.cmp_sorted)
        for k,v in pairs(self.cmp) do
            table.insert(self.cmp_sorted, v)
        end
        table.sort(self.cmp_sorted, ComponentPrioritySort)
        
        self.sort_dirty = nil
    end
    return self.cmp_sorted
end

function Entity:Despawn()
    -- Don't implement ANYTHING in this function; this is just a shortcut to make the call into the simulation.
    self.sim:DespawnEntity( self )
end

function Entity:OnDespawn()
    assert(not self.removed, string.format("Entity %s was already Removed", self))
    self.removed = true

    if self.tags then
        for tag, val in pairs( self.tags ) do
            self.sim:RemoveFromBucket( self, tag )
        end
    end
    if self.prefab.def and self.prefab.def.tags then
        for tag, val in pairs( self.prefab.def.tags ) do
            self.sim:RemoveFromBucket( self, tag )
        end
    end

    self:CancelAllTasks()

    -- Reverse order destruction, as only makes proper sense. :p    
    for i = #self:GetSortedComponents(), 1, -1 do
        local v = self:GetSortedComponents()[i]
        if v.OnDespawn then
            v:OnDespawn()
        end

        v:StopUpdating()

        self.sim:RemoveEntCmp( self, v )
        v:StopCoroutines()
    end

    --this will be processed at the end of the update. It is for other things listening to us
    self:PushLocalImmediateEvent("removed")

    
    if self.parent then
        self.parent:RemoveChild(self)
    end
    self:DestroyChildren()
end

function Entity:DoDebugRender(scene)
    for k,v in pairs(self.cmp) do
        if v.DoDebugRender then
            v:DoDebugRender(scene)
        end
    end
end

function Entity:DestroyChildren()
    if self.children and #self.children > 0 then
        while #self.children > 0 do
            local child = table.remove( self.children )
            if not child.removed then
                child:Despawn()
            end
        end
    end
end

function Entity:CancelAllTasks()
    if self.tasks and next(self.tasks) then
        --print(self, "CancelAllTasks")
        for k,v in pairs(self.tasks) do
            --print("", k)
            k:Cancel()
        end
    end
end

function Entity:CleanTasks()
    if self.tasks then
        for k,v in pairs(self.tasks) do
            if not k:Running() then
                self.tasks[k] = nil
            end
        end
    end
end

function Entity:DoCoroutine(delay, fn, ... )
    self:CleanTasks()
    local task = self.sim.scheduler:DoCoroutine(delay, fn, ...)
    self.tasks = self.tasks or {}
    self.tasks[task] = true
    return task
end

function Entity:DoTaskInTime(name, delay, fn, ... )
    self:CleanTasks()
    local task = self.sim.scheduler:DoTaskInTime(name, delay, fn, ...)
    self.tasks = self.tasks or {}
    self.tasks[task] = true
    return task
end

function Entity:DoPeriodicTask(name, inital_delay, period, fn, ... )
    self:CleanTasks()
    local task = self.sim.scheduler:DoPeriodicTask(name, inital_delay, period, fn, ...)
    self.tasks = self.tasks or {}
    self.tasks[task] = true
    return task
end

function Entity:DoTaskNextUpdate(name, fn, ...)
    return self:DoTaskInTime(name, 0, fn, ...)
end

function Entity:AddComponent(cmp)
    assert(self.cmp[cmp._classname] == nil or error( string.format("%s already has component [%s]", tostring(self), cmp._classname)))

    self.cmp[cmp._classname] = cmp
    cmp:DoAddedToEntity(self)
    if cmp.SetLocalPosition then
        assert(self.transform_provider == nil, "Only allowed one transform provider!")
        self.transform_provider = cmp
    end 

    self.sort_dirty = true
    self.impl = nil

    --ADDING TO AN ALREADY READY ENTITY
    if self:IsReady() then
        if cmp.OnEntityReady then
            cmp:OnEntityReady(self.spawn_data and self.spawn_data.cmp and self.spawn_data.cmp[cmp._classname])
        end

        if self.sim and self.sim:IsReady() then
            if cmp.OnSimReady then
                cmp:OnSimReady(self.spawn_data and self.spawn_data.cmp and self.spawn_data.cmp[cmp._classname])
            end
        end
    end

    return cmp
end

function Entity:RemoveComponent(cmp)
    --assert(false, "Congratulations! You're the first person to use this code path. You should verify it works correctly.")
    if type(cmp) == "string" then
        cmp = self.cmp[cmp]
    end

    if cmp then
        self.cmp[cmp._classname] = nil
        self.sort_dirty = true
        self.impl = nil

        cmp:DoRemovedFromEntity()

        if self.transform_provider == cmp then
            self.transform_provider = nil
        end
        --[[if cmp.DoDebugRender then
            ent.sim:UnregisterDebugRender( cmp )
        end
        --]]
    end
end

----------------Transform stuff... this is fundamental enough to be in entity

function Entity:TransformPoint(x,y,z)
    if not self.transform_provider then return x,y,z end
    return self.transform_provider:TransformPoint(x,y,z)
end

function Entity:TransformVector(x,y,z)
    if not self.transform_provider then return x,y,z end
    return self.transform_provider:TransformVector(x,y,z)
end

function Entity:GetLocalPosition()
    if not self.transform_provider then return 0,0,0 end
    return self.transform_provider:GetLocalPosition()
end

function Entity:GetLocalRotation()
    if not self.transform_provider then return 0 end
    return self.transform_provider:GetLocalRotation()
end

function Entity:GetLocalScale()
    if not self.transform_provider then return 1,1,1 end
    return self.transform_provider:GetLocalScale()
end

function Entity:SetLocalPosition(x,y,z)
    if self.transform_provider then
        self.transform_provider:SetLocalPosition(x,y,z)
    end
end

function Entity:SetLocalRotation(x, y, z)
    if self.transform_provider then 
        self.transform_provider:SetLocalRotation(x, y, z)
    end
end

-- only use SetLocalTransform if you actually want to change the scale
-- see SetPosRot
function Entity:SetLocalTransform(x,y,z, sx, sy, sz, r)
    if self.transform_provider then
        self.transform_provider:SetLocalTransform(x,y,z, sx, sy, sz, r)
    end
end

function Entity:SetPosRot(x,y,z, r)
    if self.transform_provider then
        self.transform_provider:SetPosRot(x,y,z, r)
    end
end

function Entity:SetLocalScale(x,y,z)
    if self.transform_provider then
        self.transform_provider:SetLocalScale(x,y,z)
    end
end

function Entity:GetRenderModel()
    return self.cmp.Transform and self.cmp.Transform:GetModel()
end



function Entity:GetWorldTransform()
    if self.transform_provider then
        return self.transform_provider:GetWorldTransform()
    end
end


function Entity:AddChild(child)
    local old_parent = child.parent

    if child.parent then
        child.parent:RemoveChild(child)
    end
    
    child.parent = self
    self.children = self.children or {}
    --assert( table.arrayfind( self.children, child ) == nil )
    table.insert(self.children, child)

    if self.cmp.Transform and child.cmp.Transform then
        return self.cmp.Transform:AddChild(child)
    end

    return child
end

function Entity:RemoveChild(child)
    --print (self, "Entity:RemoveChild", child)
    child.parent = nil
    table.arrayremove(self.children, child)

    for k,v in ipairs(self:GetSortedComponents()) do
        if v.OnRemoveChild then
            v:OnRemoveChild(child)
        end
    end

    if self.cmp.Transform and child.cmp.Transform then
        local child = self.cmp.Transform:RemoveChild(child)
        return child
    end
end

function Entity:GetNumChildren()
    return self.children and #self.children or 0
end

function Entity:GetChildren()
    return self.children
end

function Entity:RemoveAllChildren()
    if self.children then
        table.clear(self.children)
        if self.transform_provider then
            return self.transform_provider:RemoveAllChildren()
        end
    end
end


function Entity:ScreenToLocal(screen_x, screen_y)
    if self.transform_provider then
        return self.transform_provider:ScreenToLocal( self.sim:GetCameraInfo(), screen_x, screen_y)
    end
end

function Entity:ScreenToGround(screen_x, screen_y)
    if self.transform_provider then
        return self.transform_provider:ScreenToGround( self.sim:GetCameraInfo(), screen_x, screen_y)
    end
end

function Entity:LocalToScreen(x,y,z)
    if self.transform_provider then
        return self.transform_provider:LocalToScreen( self.sim:GetCameraInfo(), x,y,z)
    end
end

function Entity:WorldToLocal(x,y,z)
    if self.transform_provider then
        return self.transform_provider:WorldToLocal(x,y,z)
    end
end

function Entity:LocalToWorld(x,y,z)
    if self.transform_provider then
        return self.transform_provider:LocalToWorld(x,y,z)
    end
end

function Entity:GetWorldPosition()
    if self.transform_provider then
        return self.transform_provider:GetWorldPosition()
    end
end 

function Entity:ApplyEntityTransform(entity)
    if self.transform_provider then
        self:SetLocalPosition(entity:GetLocalPosition())
        self:SetLocalScale(entity:GetLocalScale())
        self:SetLocalRotation(entity:GetLocalRotation())
    end
end

function Entity:GetEvents()
    if self.events == nil then
        self.events = QueuedEventSystem()
        for k, v in ipairs(self:GetSortedComponents()) do
            if v.OnLocalEvent then
                self.events:ListenForAny( v, v.OnLocalEvent )
            end
        end     
    end
    return self.events
end

function Entity:StopListeningToLocalEvent(event, listener)
    if self.events then
        self.events:StopListening(event, listener)
    end
end

function Entity:RemoveListener( listener )
    if self.events then
        self.events:RemoveListener( listener )
    end
end

function Entity:ListenForLocalEvent(event, fn, listener)
    local events = self:GetEvents()
    self.events:ListenForEvent( event, listener, fn )
end

function Entity:ListenForLocalEvents(fn, listener)
    local events = self:GetEvents()
    self.events:ListenForAny( listener, fn )
end

local MAX_RECENT_EVENTS = 10
function Entity:PushLocalEvent(event, ...)
    local events = self:GetEvents()
    events:QueueEvent( event, self, ... )

    self.sim.entities_with_events[self] = true

    --keep track of our last couple events, for debugging
    self.recent_events = self.recent_events or {}
    self.recent_event_times = self.recent_event_times or {}
    
    if #self.recent_events < MAX_RECENT_EVENTS then
        table.insert(self.recent_events, event)
        table.insert(self.recent_event_times, self.sim:GetTime())
    else
        self.oldest_event = self.oldest_event or 1
        self.recent_events[self.oldest_event] = event
        self.recent_event_times[self.oldest_event] = self.sim:GetTime()
        self.oldest_event = self.oldest_event + 1
        if self.oldest_event > MAX_RECENT_EVENTS then
            self.oldest_event = 1
        end
    end 
end



function Entity:ConsumeEvents()
    self.events:BroadcastQueuedEvents()
end

function Entity:PushLocalImmediateEvent(event, ...)
    local events = self:GetEvents()
    events:BroadcastEvent(event, self, ...)
end

function Entity:IsOnCamera()
    local wx, wy, wz = self:GetWorldPosition()
    if wx then
        local x, y = self.sim:GetCamera():WorldToScreen( wx, wy, wz )
        local w, h = TheGame.frontend:GetScreenDims()

        local safex = w*TheGame.frontend.horizontal_safe_zone*.5
        local safey = h*TheGame.frontend.vertical_safe_zone*.5


        if x and y and x <= w-safex and x >= safex and y <= h - safey and y >= safey then 
            return true
        end
    end
    return false
end

function Entity:TweenColour(startColour, endColour, time)
    self:AddComponent(ColourTweener())
    self.cmp.ColourTweener:StartTween(startColour, endColour, easing.linear, time, self.OnColourTweenEnd)
end

function Entity:OnColourTweenEnd()
    self:RemoveComponent("ColourTweener")
end

function Entity:OnSimReady()   
    assert(self.simready == nil, self)
    self.simready = true
    local temp_table = {}
    for k,v in pairs(self.cmp) do
        temp_table[k] = v
    end
    for k,v in pairs(temp_table) do
        if v.OnSimReady then
            v:OnSimReady(self.spawn_data and self.spawn_data.cmp and self.spawn_data.cmp[k])
        end
    end
end

function Entity:MakeEntityReady()
    if not self.entity_ready then
        self.entity_ready = true

        if self.spawn_data then
            if self.spawn_data.x and self.spawn_data.z then
                self.cmp.Transform:SetLocalPosition( self.spawn_data.x, self.spawn_data.y or 0, self.spawn_data.z )
            end
        end

        local temp_table = {}
        for i, cmp in ipairs( self:GetSortedComponents() ) do
            table.insert( temp_table, cmp )
        end


        for i,cmp in ipairs(temp_table) do
            if cmp.OnEntityReady then
                cmp:OnEntityReady(self.spawn_data and self.spawn_data.cmp and self.spawn_data.cmp[cmp._classname])
            end
        end
    end
end


