require "simulation"
require "battle/fightermanager"
require "map/fx_mgr"

local paperdoll = require "paperdoll_util"

local DBG_RENDER, DBG_RENDER_STRINGS = MakeBitField{ "CAM_POS", "CAM_BOUNDS", "RANKS" }

class("FightSim", Simulation)
FightSim.can_reload = true
FightSim.DBG_RENDER = DBG_RENDER
FightSim.DBG_RENDER_STRINGS = DBG_RENDER_STRINGS

function FightSim:init()
    FightSim._base.init(self)
    self.cam = self:SpawnPrefab("fight_camera")
    
    self:PushCamera(self.cam.cmp.Camera)
    self.mgr = FighterManager(self)
    self.fx_mgr = FXManager(self)
    self.debug_render = DBG_RENDER.RANKS

    self.target_ents = {}

    self:MakeSimReady()
end

function FightSim:SetBattleSpeed( speed )
    self:SetTimeScale( speed )

    for guid, cmp in pairs( self:GetAllComponents( AnimController )) do
        cmp:SetAnimSpeed( "idle", 1 / speed )
    end
end

function FightSim:GetCamera()
    return self.cam.cmp.Camera
end

function FightSim:ShakeCamera( shake_params )
    local cam = self:GetCamera()
    local mag, speed, time = table.unpack( shake_params )
    cam:Shake(mag, speed, time) 
end

function FightSim:ShowPos(x, z, dist, gain)
    local cam = self:GetCamera()
    x = clamp(x, -FIGHT_CAM_X_RANGE, FIGHT_CAM_X_RANGE)
    z = clamp(z, -FIGHT_CAM_Z_RANGE, FIGHT_CAM_Z_RANGE)
    cam:MoveTo(x, FIGHT_CAM_HEIGHT + 1, z, dist, gain) 
end

function FightSim:IsHidingFX()
    return self.hide_fx == true
end

function FightSim:SetHidingFX( hide_fx )
    self.hide_fx = hide_fx
end

function FightSim:GetTargetFX( idx )
    local ent = self.target_ents[ idx ]
    if ent == nil then
        local target_ui = self:SpawnPrefab( "fight_target_arrow" )
        target_ui:Hide()

        self.target_ents[ idx ] = target_ui
    end

    return self.target_ents[ idx ]
end



function FightSim:GetPerspectiveScale( z )
    return (1+ clamp(z / FIGHT_FIELD_DEPTH, -1, 1)*FIGHT_PERSPECTIVE_SCALE)
end

function FightSim:ShowTargetFX( ... )
    local n = select( "#", ... )
    local count = math.floor( n / 5 )
    for i = 1, count do
        local x0, z0, x1, z1, clr = select( (i - 1) * 5 + 1, ... )
        local target_ui = self:GetTargetFX( i )

        local HEIGHT = 0.01
        local WIDTH = 2.0
        local ORIGIN_RADIUS = 2.0
        local TARGET_RADIUS = 1.9

        local total_len = math.max( 0.01, distance(x0, z0, x1, z1))
        local directionx = (x1-x0) / total_len
        local directionz = (z1-z0) / total_len

        target_ui:SetLocalPosition( x0 + (directionx * ORIGIN_RADIUS), HEIGHT, z0 + (directionz * ORIGIN_RADIUS) )

        local perspectiveScaleAmount = 1.5
        local startScale = math.pow(self:GetPerspectiveScale(z0), perspectiveScaleAmount)
        local endScale = math.pow(self:GetPerspectiveScale(z1), perspectiveScaleAmount)

        local len = total_len - TARGET_RADIUS - ORIGIN_RADIUS
        target_ui.cmp.FightArrowRenderer:SetWidthAndLength(WIDTH * startScale, len, WIDTH * endScale)

        local angle = math.atan2( x1 - x0, z1 - z0 )
        target_ui:SetLocalRotation( math.pi/2, 0, 2 * math.pi - angle )

        target_ui.cmp.FightArrowRenderer:SetColour( clr )
 
        target_ui:Show()
    end

    for i = count + 1, #self.target_ents do
        local target_ui = self.target_ents[ i ]
        target_ui:Hide()
    end
end

function FightSim:SetRenderSize(w, h)
    self._base.SetRenderSize(self, w, h)
    self.cam.cmp.Camera:SetRenderSize(w, h)
end

function FightSim:OnControlUp(control)
    if FightSim._base.OnControlUp(self, control) then return true end

    if self.controller then
        return self.controller:OnControlUp(control)
    end
end

function FightSim:OnControlDown(control)
    if FightSim._base.OnControlDown(self, control) then return true end

    if self.controller then
        return self.controller:OnControlDown(control)
    end
end

function FightSim:DoTick()
    FightSim._base.DoTick( self )
    self.fx_mgr:OnUpdate(SIM_TICK_TIME)

    if TheGame.local_settings.NO_SOUND then
        local cam = self:GetCamera()
        local x = cam:GetCurrentBasePos()
        engine.sound:SetListenerPosUpFwd(x, 2, -FIGHT_FIELD_DEPTH - 3, 0, 1, 0,  0, 0, -1)
    end

end

function FightSim:GetGroundPosUnderMouse()
    local sx, sy = TheGame:GetInput():GetMousePos()
    return self:GetGroundPosAt( sx, sy )
end

function FightSim:GetGroundPosAt( sx, sy )
    return self.root:ScreenToGround( self:GetCameraInfo(), sx, sy )
end

function FightSim:WorldToScreen( wx, wy, wz )
    return self.root:LocalToScreen( self:GetCameraInfo(), wx, wy, wz )
end

function FightSim:ScreenToWorld( sx, sy )
    return self.root:ScreenToLocal( self:GetCameraInfo(), sx, sy )
end

function FightSim:DoDebugRender()
    FightSim._base.DoDebugRender(self)
    
    if CheckBits( self.debug_render, DBG_RENDER.CAM_POS ) then
        local camx, camy, camz = self:GetCamera():GetCurrentBasePos()
        self.scene:DrawDebugLine(camx, 0, camz, camx, 1, camz, 1, 0, 1)
        self.scene:DrawDebugLine(camx, 0, camz, camx + 0.2, 0.5, camz, 1, 0, 1)
        self.scene:DrawDebugLine(camx, 0, camz, camx - 0.2, 0.5, camz, 1, 0, 1)
    end

    if CheckBits( self.debug_render, DBG_RENDER.RANKS ) then
        for fighter, ent in pairs(self.mgr.ent_by_fighter) do
            if fighter:IsAlive() then
                local x, z = ent.cmp.AnimFighter:GetHomePosition()
                self.scene:DrawDebugLine(x, 0, z, x, 1, z, 1, 0, 1)
                self.scene:DrawDebugLine(x, 0, z, x + 0.2, 0.5, z, 1, 0, 1)
                self.scene:DrawDebugLine(x, 0, z, x - 0.2, 0.5, z, 1, 0, 1)
            end
        end
    end

    if CheckBits( self.debug_render, DBG_RENDER.CAM_BOUNDS ) then
        self.scene:DrawDebugLine(-FIGHT_CAM_X_RANGE, 0, -FIGHT_CAM_Z_RANGE, FIGHT_CAM_X_RANGE, 0, -FIGHT_CAM_Z_RANGE, 1, 1, 1)
        self.scene:DrawDebugLine(-FIGHT_CAM_X_RANGE, 0, FIGHT_CAM_Z_RANGE, FIGHT_CAM_X_RANGE, 0, FIGHT_CAM_Z_RANGE, 1, 1, 1)
        self.scene:DrawDebugLine(-FIGHT_CAM_X_RANGE, 0, -FIGHT_CAM_Z_RANGE, -FIGHT_CAM_X_RANGE, 0, FIGHT_CAM_Z_RANGE, 1, 1, 1)
        self.scene:DrawDebugLine(FIGHT_CAM_X_RANGE, 0, -FIGHT_CAM_Z_RANGE, FIGHT_CAM_X_RANGE, 0, FIGHT_CAM_Z_RANGE, 1, 1, 1)
    end
end