require "input/input"
require "ui/frontend"
require "ui/screens/savingscreen"
require "ui/screens/signinscreen"
require "ui/screens/mainmenu"
require "ui/screens/slideshowscreen"
require "ui/screens/locationscreen"
require "ui/screens/mainoverlay"
require "ui/screens/feedbackscreen"
require "ui/screens/dailyoverscreen"
require "ui/screens/picklegacyscreen"
require "ui/screens/pausemenu"
require "ui/uihelpers"
require "gamestate"
require "game_profile"
require "input_settings"
require "streaming_profile"
require "run_history"
require "game_stats"
require "eventsystem"
require "util/scheduler"
require "ambience"
require "music"

require "imageprocessor"
local path = require "util/filepath"

-- Prefix assigned to missing localization strings.
local MISSING_STRING = "" --"STR:"


----------------------------------------------------------------
-- Topmost Game class.

class("Game")

function Game:init( local_settings )
    engine.inst:ProfilerPush( "Game()" )
    -- sigh. FIXME. Child constructors reference this Game object through the global accessor, thus requiring this here.
    rawset( _G, "TheGame", self )

    -- local_settings are DEV-only settings
    self.local_settings = local_settings
    
    if local_settings.DEBUG then
        engine.inst:SetDebugMode(true)
        engine.inst:ProfilerPush( "DebugManager()", 0x55CC55FF )

        local DebugManager = require "debug/debug_mgr"
        local debug_menus = require "debug/debug_menus"
        self.debug = DebugManager( self )
        self.debug:AddBindingGroup( debug_menus.GAME_BINDINGS )
        self.debug:ExecuteDebugFile( "scripts/debug/consolecommands.lua" )
        self.debug:GetDebugUI():SetDisplayScale( self.local_settings.DEBUG_IMGUI_SCALE or 1.0 )

        engine.inst:SetSetting("DEBUG", "font", self.local_settings.DEBUG_FONT)
        engine.inst:ProfilerPop()
    else
        self.debug = nil
    end

    -- GameProfile stores gameplay settings, including meta progression.
    -- On certain builds, this requires user sign-in to occur first.  See OnSignIn.
    if not CheckBits( SPECIAL_BUILD_TYPE, 128 ) then
        self:OnSignIn()
    end

    self.wall_time_debug_scale = 1
    self.walltime = 0

    Content.DoLoad( self )
 
    self.input = Input(self, self.input_settings)
    self.frontend = FrontEnd(self)

    if SPECIAL_BUILD_TYPE == (1024+64) and self.debug and CONFIG ~= "PRODUCTION" then
        self.debug:ToggleDebugFlags( DBG_FLAGS.RENDER | DBG_FLAGS.SIM )
    end

    self.prefabdefs = {}
    self.music = Music()
    self.ambience = Ambience()
    
    self:LoadPrefabs("prefabs" )
    self.scheduler = Scheduler()
    self.scheduler:SetTimeFn(function(t) self.walltime = t end)

    self.eventsystem = EventSystem()
    self.robotics_settings = {}
    self.times = {}
    self.smoothtimes = {}
    self.maxtimes = {}

    engine.inst:ProfilerPush( "scene stuff" )
    self.ui_render_node = engine.rendering.SceneRenderNode("UI Render Node")        


    local w, h = engine.inst:GetScreenSize()
    self.pending_w, self.pending_h = w, h

    self.frontend:OnScreenResize(w, h)
    self.input:OnScreenResize(w, h)


    --local msaa = engine.inst:GetSetting("graphics", "msaa", 0)
    local bloom = engine.inst:GetSetting("graphics", "bloom", true)

    self.msaa = 0
    self:SetBloomEnabled(bloom) -- triggers the first postprocess framebuffer creation.

    local lod_level = engine.inst:GetSetting("graphics", "lod_level", 0)
    if lod_level == 0 then 
        lod_level = engine.inst:GetRecommendedLevelOfDetail()
    end
    self:SetLevelOfDetail(lod_level);


    local max_w, max_h = engine.inst:GetMaxDesktopResolution()
    
    local default_cursor = "CURSOR_1"
    

    if max_h <= 1080 then
        default_cursor = "CURSOR_2"    
    elseif max_h <= 1440 then
        default_cursor = "CURSOR_3"    
    else
        default_cursor = "CURSOR_4"
    end

    local cursor = engine.inst:GetSetting("graphics", "cursor", default_cursor)
    engine.inst:SetCursor(cursor)
    engine.inst:SetSetting("graphics", "cursor", cursor)

    self.tex_routing_dirty = true

    self.imageprocessor = ImageProcessor("improc1")
    self.imageprocessor2 = ImageProcessor("improc2")
    engine.inst:ProfilerPop()


    engine.inst:ProfilerPop()
end

function Game:Shutdown()
    if self.debug then
        self.debug:Shutdown()
        self.debug = nil
    end
    engine.inst:Quit()
end

function Game:BlurImage(img, amount)

    local w, h = img:GetSize()
    --w, h = math.floor(w/2), math.floor(h/2)
    self.imageprocessor:SetConstant("RESOLUTION", w,h)
    self.imageprocessor:SetConstant("direction", amount or 1, 0)

    local blurshader = global_shaders.UI_BLUR_NICE
    if self:GetLevelOfDetail() == 1 then -- low level of detail
        blurshader = global_shaders.UI_BLUR_FAST
    end
    
    local half_blurred = self.imageprocessor:Process(blurshader, img, w,h)
    --do return half_blurred end    

    self.imageprocessor2:SetConstant("RESOLUTION", w,h)
    self.imageprocessor2:SetConstant("direction", 0, amount or 1)
    local blurred = self.imageprocessor2:Process(blurshader, half_blurred, w,h)
    return blurred
end

function Game:SetWallTimeScale(scale)
    self.wall_time_debug_scale = scale
end

function Game:GetWallTimeScale()
    return self.wall_time_debug_scale
end

function Game:GetWallTime()
    return self.walltime
end

function Game:IsDemo()
    return self.local_settings.DEMO == true or engine.inst:IsDemo()
end

function Game:GetEvents() 
    return self.eventsystem
end

function Game:BroadcastEvent( eventname, ... )
    self.eventsystem:BroadcastEvent( eventname, ... )
end

function Game:GetLocalSettings()
    return self.local_settings
end

function Game:GetGameProfile()
    return self.game_profile
end

function Game:GetStreamingProfile()
    return self.streaming_profile
end

function Game:GetInputSettings()
    return self.input_settings
end

function Game:GetGameStats()
    return self.game_stats
end

function Game:AddGameplayStat( stat_id, amount )
    if not self.gamestate then
        return
    end

    if self.gamestate:GetOptions().experiment then
        return
    end

    if self.gamestate:GetCurrentAct() == nil then
        return
    end

    if self.local_settings.ROBOTICS then
        return
    end

    -- print( "Stat:", stat_id, amount )
    self.game_stats:AddStat( stat_id, amount )

    self:CheckGriftProgress()
end

function Game:GetInput()
    return self.input
end

function Game:CreateLocKey(...)
    local num_specifiers = select('#', ... )
	local loc_key = nil
	for k = 1, num_specifiers do
		local specifier = select(k, ... )
		if loc_key then
			loc_key = loc_key .. "." .. specifier
		else
			loc_key = specifier
		end
	end

    return loc_key
end

function Game:LookupStr(id)
    return Content.LookupString( id )
end

function Game:Str(id)
    if id then
        return Content.LookupString( id ) or MISSING_STRING..id
    end

    return ""
end

function Game:GetLanguage()
    return self.language 
end

function Game:GetLanguageInfo()
    local language = self:GetLanguage()
    return Content.GetLocalizations()[ language ]
end

function Game:IsDefaultLanguage()
    return self:GetLanguage() == "english"
end


function Game:InitLanguage()

    local userlocalename = engine.inst:GetUserLocaleName()
    local language = engine.inst:GetSetting("general", "language", "")
    local settings_language = language


    if language == "" or Content.GetLocalizations()[language] == nil then

        if userlocalename ~= "" then
            for language_name,localization in pairs(Content.GetLocalizations()) do
                if localization.default_languages and table.arrayfind(localization.default_languages, userlocalename) then
                    language = language_name
                    break
                end
            end
        end

        if language == "" then
            language = engine.inst:GetPreferredLanguage()
        end
    end

    print( "Game:InitLanguage", userlocalename, settings_language, language )
    
    if settings_language ~= language then
        engine.inst:SetSetting( "general", "language", language )
        engine.inst:SaveSettings()
    end


    self.language = language or "english"
end

function Game:FE()
    return self.frontend
end

function Game:GetDebug()
    return self.debug
end

function Game:GetDebugEnv()
    if self.debug then
        return self.debug:GetDebugEnv()
    end
end

function Game:IsDebugFlagged( flags )
    return self.debug and self.debug:IsDebugFlagged( flags ) or false
end

function Game:RawKeyUp(key)
    if self.frontend:RawKeyUp(key) then return true end
end

function Game:RawKeyDown(key)
    if self.debug == nil or not self.debug:IsConsoleDocked() then
        if self.frontend:RawKeyDown(key) then
            return true
        end
    end

    if self.debug and self.debug:RawKeyDown( self.input, key) then
        return true
    end
end

function Game:OnTextInput(text)
    if self.frontend:TextInput(text) then
        return true
    end
end

function Game:OnTextEdit(text)
    if self.frontend:TextEdit(text) then
        return true
    end
end


function Game:OnControlDown(control, device_type, device_id)


    if self.debug and device_type == "gamepad" then
        if control:Has( Controls.Digital.TOGGLE_DEBUG_MENU ) then 
            self.debug:TogglePanel()
            
            self.debug_using_gamepad = self.debug:DebugPanelsOpen()
            return true
        end

        if self.debug_using_gamepad and self.debug:DebugPanelsOpen() then
            return true
        end
    end        
    

    if self.debug == nil or not self.debug:IsConsoleDocked() then
        if self.frontend:OnControlDown(control, device_type, nil, device_id) then
            return true
        end
    end

    if self.debug and self.debug:OnControlDown( control, device_type ) then
        return true
    end

    if control:Has( Controls.Digital.FEEDBACK ) then 
       self:StartFeedback()
        return true
    end

end

function Game:IsDebugUsingGamepad()
    return self.debug and self.debug:DebugPanelsOpen() and self.debug_using_gamepad
end


function Game:OnControlUp(control, device_type, device_id)

    if self.debug and self.debug_using_gamepad and device_type == "gamepad" and self.debug:DebugPanelsOpen() then
        return true
    end        

    if self.debug == nil or not self.debug:IsConsoleDocked() then
        if self.frontend:OnControlUp(control, device_type, device_id) then
            return true
        end
    end
end

function Game:LookupPlayerMusic( id )
    local background = self.gamestate and self.gamestate:GetCurrentBackground()
    if background then
        return background.data[ id ]
    end
end

function Game:GetMusic()
    return self.music
end

function Game:OnSignIn()
    -- Need to reload all user data.
    self.game_profile = GameProfile()
    self.game_stats = GameStats()
    self.run_history = RunHistory()
    self.input_settings = InputSettings()
    self.streaming_profile = StreamingProfile()

    local screen = self.frontend and self.frontend:FindScreen( Screen.SignInScreen )
    if screen then
        screen:OnSignIn()
    end
end

function Game:PromptSignInWrongAccount()
    if self.gamestate then
        local ret
        -- In Game.  Need to re-prompt sign in, as they cannot proceed with the wrong profile.
        TheGame:FE():PushScreen( Screen.YesNoPopup(
            LOC"UI.SIGN_IN_SCREEN.POPUP_WRONG_USER_TITLE", LOC"UI.SIGN_IN_SCREEN.POPUP_WRONG_USER_DESC", nil, nil, LOC"UI.SIGN_IN_SCREEN.ABANDON" )
            :SetYesText( LOC"UI.SIGN_IN_SCREEN.POPUP_SIGNIN" )
            :SetFn( function(ret)
                if ret == Screen.YesNoPopup.YES then
                    engine.inst:InitiateSignIn( false )

                else
                    TheGame:Reset( function() 
                        TheGame:FE():InsertScreen( Screen.SignInScreen() )
                    end)
                end
            end ))

    else
        UIHelpers.InfoPopup( LOC"UI.SIGN_IN_SCREEN.POPUP_WRONG_USER_TITLE", LOC"UI.SIGN_IN_SCREEN.POPUP_WRONG_USER_DESC" )
    end
end

function Game:PromptSignIn()
    if self.gamestate then
        local ret
        -- In Game.  Need to ask if they want to continue without being able to Save.
        TheGame:FE():PushScreen( Screen.YesNoPopup(
            LOC"UI.SIGN_IN_SCREEN.POPUP_TITLE", LOC"UI.SIGN_IN_SCREEN.POPUP_DESC", LOC"UI.SIGN_IN_SCREEN.ABANDON" )
            :SetFn( function(ret)
                if ret == Screen.YesNoPopup.YES then
                    -- Continue without save.

                elseif ret == Screen.YesNoPopup.CANCEL then
                    engine.inst:InitiateSignIn( false )

                else
                    TheGame:Reset( function() 
                        TheGame:FE():InsertScreen( Screen.SignInScreen() )
                    end)
                end
            end ))

    else
        -- Shut everything down and go to sign in screen.
        TheGame:Reset( function() 
            TheGame:FE():InsertScreen( Screen.SignInScreen() )
        end)
    end
end

function Game:Reset( fn )
    print( "########## Game:Reset() ############\n" )

    engine.inst:ProfilerPush( "Game:Reset()" )
	engine.inst:EndMetricsSession()

    AUDIO:Reset()
    self:ApplyVolumeSettings()
    
    self:Clear()
    
    engine.inst:SetRecentSaveFile( nil )

    local settings = self.local_settings or table.empty
    TheGame:SetWallTimeScale(settings.TIME_SCALE or 1)

    if fn then
        fn()
    elseif settings.SCREEN_OVERRIDE then
        local path =  "ui/screens/".. string.lower(settings.SCREEN_OVERRIDE)
        require( path )
        self.frontend:InsertScreen( Screen[settings.SCREEN_OVERRIDE]() )

    elseif settings.SKIP_MENUS then
        if settings.LOAD_FILE then
            self:LoadGame(settings.LOAD_FILE)
        else
            self:StartNewGame()
        end
    
    elseif CheckBits( SPECIAL_BUILD_TYPE, 128 ) then
        self.frontend:PushScreen( Screen.SignInScreen() )

    else
        -- A function that shows the main menu screen
        local show_main_menu_fn = function() self.frontend:FadeTransition(function() self.frontend:PushScreen(Screen.MainMenu()) end, .3) end
        -- If we haven't shown the player the save-game screen, do that, and then main menu. Otherwise, send them directly to the main menu
        if BRANCH ~= "dev"
        and SPECIAL_BUILD_TYPE == (1024+64)
        and not self.shown_save_screen then
            self.shown_save_screen = true
            self.frontend:PushScreen( Screen.SavingScreen( show_main_menu_fn ) )
        else
            show_main_menu_fn()
        end
    end

    if self.debug then
        local startup_file = settings.LOCAL_STARTUP or "scripts/local_startup.lua"
        local ok, err = pcall( self.debug.ExecuteDebugFile, self.debug, startup_file )
        if not ok then print( "\n\n*****local_startup*********************\n\n",err,"\n\n**************************************\n" ) end
        
    end

    engine.inst:ProfilerPop()
end

function Game:SetBloomEnabled(val)
    if val ~= self.bloom_enabled then
        self.post_dirty = true
        self.bloom_enabled = val

        self.ui_post_process_node = engine.rendering.PostProcessRenderNode("PostProcess")
        self.ui_post_process_node:SetBloomEnabled(self.bloom_enabled)
        self.ui_post_process_node:SetBloomScale(2.5)
        self.ui_post_process_node:SetVignetteStrength( UI_VIGNETTE )
--        self.ui_post_process_node:SetFilmGrainStrength( UI_FILMGRAIN )
        
    end
    collectgarbage()
end

function Game:SetLevelOfDetail(val)
    if val ~= self.lod_level then
        self.lod_level = val
    end
    collectgarbage()
end

function Game:GetLevelOfDetail()
    return self.lod_level or 4  -- 4 = high level of detail 
end

function Game:SetVignetteStrength( val )
    self.ui_post_process_node:SetVignetteStrength( val )
end

--function Game:SetFilmGrainStrength( val )
--    self.ui_post_process_node:SetFilmGrainStrength( val )
--end

function Game:Render(dt)
    
    engine.inst:ProfilerPush( "SETUP AND CLEAR")
    self:CreateRenderBuffers()

    self.ui_render_node:ClearTexture()
    self.ui_post_process_node:ClearTexture()

    engine.inst:ProfilerPop()


    engine.inst:ProfilerPush( "OnRENDER")
    self.frontend:OnRender(dt)
    if self.debug then
        self.debug:DoRender(dt)
    end
    engine.inst:ProfilerPop()

    engine.inst:ProfilerPush( "PRESENT")
    
    self.ui_render_node:SetTreeScene(self.frontend.scene, self.frontend.root, self.frontend.camera, (1 << RENDER_PASS_OPAQUE) | (1 << RENDER_PASS_ALPHA))
    self.ui_render_node:RenderToTexture(self:GetLevelOfDetail())
    

    if self.tex_routing_dirty then
        self.ui_post_process_node:SetInputTexture( self.ui_render_node:GetOutputTexture(0) )
        local bright = self.ui_render_node:GetOutputTexture(1)
    
        if bright then
            self.ui_post_process_node:SetBrightInputTexture( bright )
        end
        self.tex_routing_dirty = false
    end


    self.ui_post_process_node:RenderToScreen(self:GetLevelOfDetail())
    
    engine.inst:ProfilerPop()
end


function Game:CaptureRenderOutput(tex)
 
    self:CreateRenderBuffers()

    -- ccc: disabling this because it triggers ImGUI calls, but ImGUI msut be called under Game:OnRender.
    -- I think it's fine, because this should just use the last rendered scene_render_node texture.
    --self.frontend:OnRender()

    if self.frontend.tooltip_widget then
        self.frontend.tooltip_widget:Hide()
    end
    self.ui_render_node:SetTreeScene(self.frontend.scene, self.frontend.root, self.frontend.camera, (1 << RENDER_PASS_OPAQUE) | (1 << RENDER_PASS_ALPHA))
    self.ui_render_node:RenderToTexture(self:GetLevelOfDetail())
    self.ui_post_process_node:SetInputTexture( self.ui_render_node:GetOutputTexture(0) )
    self.ui_post_process_node:SetVignetteStrength(0)
--    self.ui_post_process_node:SetFilmGrainStrength(0)

    local bright = self.ui_render_node:GetOutputTexture(1)
    if bright then
        self.ui_post_process_node:SetBrightInputTexture( bright )
    end
    

    self.ui_post_process_node:CaptureOutput(tex, self:GetLevelOfDetail())
    self.ui_post_process_node:SetVignetteStrength( UI_VIGNETTE )
--    self.ui_post_process_node:SetFilmGrainStrength( UI_FILMGRAIN )
end



function Game:Update(dt)
    if self.debug and self.debug:IsConsoleDocked() then
        return
    end

    if self.gamestate and not self.frontend:FindScreen( Screen.PauseMenu ) then
        self.gamestate:AccumulatePlayTime(dt)
    end

    dt = math.min(dt, 1/30)

    local wt = self.walltime
    self.times.frame, self.times.update, self.times.render, self.times.swap_time, self.times.update_lua = engine.inst:GetTimings()
    local alpha = 1/20
    for k,v in pairs(self.times) do
        self.maxtimes[k] = self.maxtimes[k] and math.max(self.maxtimes[k], v) or v
        self.smoothtimes[k] = self.smoothtimes[k] and ( (alpha*v) + (1-alpha)*self.smoothtimes[k]) or v
    end
    engine.inst:ProfilerPush("Wall Scheduler Update")
    self.walltime = wt + dt
    self.scheduler:Update(dt)
    engine.inst:ProfilerPop()

    engine.inst:ProfilerPush("Input Update")
    self.input:Update(dt)
    engine.inst:ProfilerPop()

    engine.inst:ProfilerPush("FE Update")
    self.frontend:Update(dt)
    engine.inst:ProfilerPop()

    if self.test_update_fn then
        self.test_update_fn(dt)
    end

end

function Game:SetTestUpdateFn(fn)
    self.test_update_fn = fn
    return self
end

function Game:CanLoadSave( save_data )
    if not save_data or not save_data.options then
        return false
    end

    local act_data = GetPlayerActData( save_data.options.start_act )
    if not act_data then
        return false
    end

    if (save_data.save_version or 0) < SAVE_VERSION then
        return false, LOC"UI.NEW_GAME_SCREEN.SAVE_TOO_OLD_BODY"
    end
    
    if act_data.data and act_data.data.invalidate_saves_before_revision then
        if save_num < act_data.data.invalidate_saves_before_revision then
            return false, LOC"UI.NEW_GAME_SCREEN.SAVE_TOO_OLD_BODY"
        end
    end

    local def = Content.GetCharacterDef( act_data.background.data.player_agent )
    if not def then
        return false
    end

    if save_data.options.required_mods then
        for i, v in ipairs( save_data.options.required_mods ) do
            if not Content.IsModEnabled( Content.FindMod( v.id ), v.version ) then
                return false, LOC"UI.NEW_GAME_SCREEN.SAVE_MISSING_MODS"
            end
        end
    end

    return true
end

function Game:DeleteSave(filename, cb)
    print ("### Game:DeleteSave", filename)
    engine.savegamestorage:DeleteFile(filename, cb)
end

function Game:Clear()

    self.frontend:ClearScreens()
    self.ambience:Clear()
    if self.debug then
        self.debug:Clear()
    end
    self.eventsystem:Clear()

    if self.gamestate then
        self.gamestate:OnDestroyed()
        self.gamestate = nil
    end
    
    UnloadFXTracks()

    serpent.clear()
    collectgarbage()
end



function Game:LoadSaveHeader( filename, fn )

    local full_filename = filename
        
    local function OnLoad(success, filedata)
        local save_header

        if success and filedata then
            local load_ok, header = pcall(loadstring(filedata))

            if load_ok and header.save_version == SAVE_VERSION then
                save_header = header
            end
        end
        if fn then
            fn( save_header )
        end
    end
    engine.savegamestorage:LoadFile( full_filename, OnLoad, true )
end

function Game:LoadGame( filename, autosave_filename, fn )
    
    local full_filename = filename
    
    self:Clear()
	engine.inst:BeginMetricsSession()
    AUDIO:PlayEvent( SoundEvents.start_game )

    print( "=============================")
    print ("LOAD GAME", full_filename)
    print( "=============================")

    local load_ok, err
    local function OnLoad(success, filedata)
        if success and filedata then
            local header
            load_ok, header = pcall(loadstring(filedata))
            
            if not load_ok then
                err = tostring(header)

            elseif header.save_version ~= SAVE_VERSION then
                load_ok, err = false, "Sorry - incompatible save version v."..tostring(header.save_version).."\nPlease restart."

            elseif type(header.gamestate) ~= "function" then
                load_ok, err = false, "Corrupted file; missing gamestate loader"

            else
                assert( self.gamestate == nil )
                self.gamestate = GameState( header.options, header.gamestate() )
                self.gamestate:SetSaveFileName(autosave_filename or filename)
                self.gamestate:PostLoad()
                
                if fn then
                    fn(self.gamestate)
                end

                self:ResumeGame()
            end
        else
            load_ok, err = false, "Failed to load file"
        end

        if not load_ok then
            engine.inst:ShowWarning( err or "Unknown error" )
        end
    end

    engine.inst:SetRecentSaveFile( full_filename )
    TheGame:GetGameProfile():SetCurrentSaveGame(filename)

    engine.savegamestorage:LoadFile( full_filename, OnLoad, true )

    -- make sure that the voices are loaded.
    AUDIO:FinalizeVoiceLoad()

    while load_ok == nil do end -- Technically correct, but LoadFile currently isn't REALLY asynchronous. :P
    return load_ok
end

--get the headers for all valid save games. It would be great to do this without actually loading everything...
function Game:ListSaveGames( cb )
    local ret = {}
    
    local function OnListFiles( files ) 
        for k,v in pairs(files) do
            if v.type == "file" then
                ret[v.name] = {name = v.name, filename = v.name, status = "PENDING", data = nil}
            end
        end
        if next(ret) then
            for k,v in pairs(ret) do
                local is_encoded = v.filename:match( "[.]lua$") == nil -- only a .lua file is non-encoded
                
                local function OnLoadFile( success, filedata )
                    local load_ok, data = pcall(loadstring(filedata))
                    if success and load_ok and data ~= nil then
                        v.data = data
                        v.status = "LOADED"
                    else
                        ret[k] = nil
                    end

                    local done = true
                    for k,v in pairs(ret) do
                        if v.status ~= "LOADED" then
                            done = false
                            break
                        end
                    end
                    if done then
                        cb(ret)
                    end

                end
                engine.savegamestorage:LoadFile( v.filename, OnLoadFile, is_encoded )
            end
        else
            cb(ret)
        end

    end
    engine.savegamestorage:ListFiles("*", OnListFiles )
end


local function SaveGameInternal(filename, gamestate, cb)
    -- Save out the header and gamestate in separate 'chunks', wrapping the gamestate tables in a function call.
    -- The purpose of this is to allow the load() of the savegame string to always return the header table successfully for versioning
    -- purposes, in case the game/sim tables contain missing/incompatible class/upvalue references that would otherwise fail load() outright.
    -- Note that a syntactically incorrect savefile is still considered a "bug" (not gracefully handled).
    local percent, title = gamestate:GetNarrativeProgress()
    local header = 
    {
        save_version = SAVE_VERSION,
        play_time = gamestate.play_time or 0,
        afk_time = gamestate.afk_time or 0,
        desc = title,
        progress = percent or 0,
        options = gamestate:GetOptions(),
        mods = Content.GetInstalledMods(),
        time = os.time(),
        creation_time = gamestate.creation_time,
        creation_build_id = gamestate.build_id,
        build_id = BUILD_ID_STRING,
    }
    
    local s = engine.LuaSerializer()
    local savestr = table.concat
    {
        s:Serialize( header, "HEADER" ),
       "\nHEADER.gamestate = function()\n",
        s:Serialize( gamestate:GetSaveData() ),
        "\nend\n",
        "return HEADER"
    }

    --print("writing save...", filename)

    engine.savegamestorage:SaveFile(filename, savestr, cb, true )
    engine.inst:SetRecentSaveFile( filename )
    TheGame:GetGameProfile():SetCurrentSaveGame( filename )
end

function Game:SaveGame(filename, cb)
    assert( filename )
    print ("Saving", filename)

    -- Display saving icon
    TheGame:FE():ShowSaving( 0.4 )

    -- Also update our stats
    self.game_stats:SaveSettings()
    self:GetGameProfile():SaveSettings()
    SaveGameInternal( filename, self:GetGameState(), cb )
end

function Game:StartFeedback()
    local feed = TheGame:FE():FindScreen( Screen.FeedbackScreen )
    if not feed then
        engine.inst:RequestScreenshot()
        TheGame:SetWallTimeScale(0)
    end
end
function Game:ScreenShotReady(texture, texture_string )
    self.frontend:PushScreen( Screen.FeedbackScreen( self.gamestate, texture, texture_string ))
end

function Game:LoadPrefabDefs(file)
    local prefabs = require(file)
    for k,v in pairs(prefabs) do
        self:AddPrefab(v)
    end
end

function Game:AddPrefab(prefab)
    if self.prefabdefs[prefab.name] then
        error("Duplicate prefab " .. prefab.name)
    end
    self.prefabdefs[prefab.name] = prefab
end

function Game:GetPrefabData(name)
    return self.prefabdefs[name]
end

function Game:LoadPrefabs(subdir)
    engine.inst:ProfilerPush( "LoadPrefabs" )
    local prefabs_base = "scripts/"
    local prefabs = engine.inst:ListFiles(prefabs_base .. subdir .. '/*.lua')
    for k,p in pairs(prefabs) do
        if string.len(p) > 5 and string.sub(p, -4) == '.lua' then
            local prefab = string.sub(p, 1, -5)
            self:LoadPrefabDefs(subdir .. '/' .. prefab)
        end
    end
    engine.inst:ProfilerPop()
end

local function OnNewGame( self, save_file_name )

    -- make sure that the voices are loaded, BEFORE we play any slides!
    AUDIO:FinalizeVoiceLoad()

    if not self:GetLocalSettings().SKIP_SLIDE_SHOWS and not GetSkipMovies() and not self.local_settings.NO_MOVIES then
        local act = self.gamestate:GetCurrentAct()
        local progress_id = "INTRO_SLIDES_SEEN_" .. act.id
        local idx = self:GetGameProfile():GetNarrativeProgress(progress_id) + 1
        
        local slides
        if act.data.slides and type(act.data.slides) == "string" then
            slides = act.data.slides
        elseif act.data.slides and type(act.data.slides) == "table" then
            slides = act.data.slides[math.min(idx, #act.data.slides)]
        end

        if slides then
            self:GetGameProfile():IncNarrativeProgress(progress_id)

            local coro = coroutine.running()
            local function OnDone()
                coroutine.resume( coro )
            end
            self:FE():PushScreen( Screen.SlideshowScreen( slides, OnDone ))
            coroutine.yield()
        end
    end

    print( "**** NEW GAME:", save_file_name )
    print( "FAST_MODE:", self:GetLocalSettings().FAST_STARTUP )
    print( "Options:", tostr(self.gamestate.options, 2))
    self.gamestate:OnNewGame( save_file_name )

    -- Only save once NewGame has completely initialized.
    -- ccc:  I actually DONT want a new save right after start -- the player hasn't done anyhting.
    -- DoAutoSave()
end


function Game:StartNewGame(options, save_file_name)
    options = options or GameState.CreateOptions()
    
    self:Clear()

    engine.inst:ProfilerPush("Start new game")
    do
        self.gamestate = GameState( options )
        self.gamestate:InitializeAct()

        Content.InvokeModAPI( "OnNewGame", self.gamestate )
    end
    engine.inst:ProfilerPop()
	
    engine.inst:BeginMetricsSession()
	SendMetricsData("NEW_GAME")
   
    self:ResumeGame()

    -- New game can result in a sequence of UI (slide shows, mutator popups, etc.)
    -- and thus is run in a coroutine to avoid callback hell.
    -- The newly created game is auto-saved ONLY at the end.
    local coro = coroutine.create( OnNewGame )
    local ok, err = coroutine.resume( coro, self, save_file_name )
    if not ok then
        print( "ERROR starting new game:", err )
        print( debug.traceback( coro ))
    end
end

function Game:OnDropFile(txt)
    local filename = txt:match( ".-([^\\]+[.]sav)$")
    if filename then
        local f = io.open( txt, "r" )
        local savestr = f:read( "*all" )
        local function OnSave( success, filedata )
            self.scheduler:DoTaskInTime( "DROP-LOAD", 0.1, self.LoadGame, self, filename, "autosave.sav" )
        end
        engine.savegamestorage:SaveFile( filename, savestr, OnSave )

    else
        self.eventsystem:BroadcastEvent( "file_drop", txt )
    end
end

function Game:SetMSAA(msaa)
    msaa = math.floor(msaa)
    if msaa >= 0 and msaa <= 8 then
        self.pending_msaa = msaa
    end
end


function Game:CreateRenderBuffers()

    if (self.pending_w and self.w ~= self.pending_w) or (self.pending_h and self.h ~= self.pending_h) or (self.pending_msaa and self.msaa ~= self.pending_msaa) or self.post_dirty then
		
        collectgarbage()

        self.post_dirty = false
        self.w = self.pending_w or self.w
        self.h = self.pending_h or self.h
        self.msaa = self.pending_msaa or self.msaa

        if self.h > 1080 then
            
        end

        local w = self.w
        local h = self.h

	local rdrw = math.round(self.w)
	local rdrh = math.round(self.h)

	local fmt0 = PIXEL_FORMAT.A8R8G8B8;
	local fmt1 = -1
	if self.bloom_enabled then
		fmt1 = fmt0
	end

	self.ui_render_node:SetRenderFormat(rdrw, rdrh, fmt0, fmt1,  true,  self.msaa)
        self.ui_post_process_node:SetRenderFormat(rdrw, rdrh, fmt0,   -1, false,      false)

        self.pending_w = nil
        self.pending_h = nil
        collectgarbage()
        self.tex_routing_dirty = true
        self.eventsystem:BroadcastEvent( "new_render_buffers" )
    end
end


function Game:OnScreenResize(w, h)
    
    self.pending_w = w
    self.pending_h = h

    if self.w ~= w or self.h ~ h then    
        self.frontend:OnScreenResize(w,h)
        self.input:OnScreenResize(w, h)
    end
end


function Game:ResumeGame()
    assert(self.gamestate, "You need a valid gamestate to resume!")

    self.frontend:ClearScreens()
    self.music:StopMusic() -- FE.

    PreloadFXTracks()

    self.frontend:InsertScreen( Screen.LocationScreen( self.gamestate:GetPlayerAgent():GetLocation() ))
    self.frontend:PushScreen( Screen.MainOverlay( ) )

    self.frontend:FadeIn( 0.5 )  
    if self.gamestate:GetPlayerAgent() then
        TheGame:GetGameProfile():SetLastCharacterPlayed(self.gamestate:GetPlayerAgent():GetContentID())
    end

    Content.InvokeModAPI( "OnResumeGame", self.gamestate )
end

function Game:GetGameState()
    return self.gamestate
end

function Game:SetAmbience( ambience )
    self.ambience:SetAmbience( ambience)
end

function Game:SetAmbienceParam( param, val )
    self.ambience:SetParameter( param, val )
end

local vca_names = 
{
    master = "vca:/master",
    music = "vca:/music",
    sfx = "vca:/sfx_master",
    vo = "vca:/vo",
    ambience = "vca:/sfx_ambient",
    actions = "vca:/sfx_actions",
    cards = "vca:/sfx_card",
    ui = "vca:/sfx_ui",
}

function Game:SetVolume(ID, v)
    if self.local_settings.NO_SOUND then
        return
    end

    if ID == "music_volume" then
        if AUDIO:GetVCA( vca_names.music ) then
            AUDIO:GetVCA(vca_names.music):SetFaderLevel(v)
        end    
    elseif ID == "master_volume" then
        if AUDIO:GetVCA( vca_names.master ) then
            AUDIO:GetVCA(vca_names.master):SetFaderLevel(v)
        end
    elseif ID == "fx_volume" then
        if AUDIO:GetVCA( vca_names.sfx ) then
            AUDIO:GetVCA(vca_names.sfx):SetFaderLevel(v)
        end
    elseif ID == "vo_volume" then
        if AUDIO:GetVCA( vca_names.vo ) then
            AUDIO:GetVCA(vca_names.vo):SetFaderLevel(v)
        end
    elseif ID == "ambience_volume" then
        if AUDIO:GetVCA( vca_names.ambience ) then
            AUDIO:GetVCA(vca_names.ambience):SetFaderLevel(v)
        end
    elseif ID == "actions_volume" then
        if AUDIO:GetVCA( vca_names.actions ) then
            AUDIO:GetVCA(vca_names.actions):SetFaderLevel(v)
        end
    elseif ID == "cards_volume" then
        if AUDIO:GetVCA( vca_names.cards ) then
            AUDIO:GetVCA(vca_names.cards):SetFaderLevel(v)
        end
    elseif ID == "ui_volume" then
        if AUDIO:GetVCA( vca_names.ui ) then
            AUDIO:GetVCA(vca_names.ui):SetFaderLevel(v)
        end
    else
        LOGWARN("There is no volume setting [%s]", ID)
    end
end

function Game:ApplyVolumeSettings()
    for i, key in ipairs( VOLUME_SETTINGS ) do
        self:SetVolume( key, engine.inst:GetSetting( "audio", key, 1.0 ))
    end
end

function Game:Lose(killed_by)
    self:GetGameState():SetGameOver( { result = GAMEOVER.DEATH, killed_by = killed_by } )
    self:FinalizeRun()
    self:ShowGameOverScreen()
end

function Game:AbandonGame()
    self:GetGameState():SetGameOver( { result = GAMEOVER.ABANDONED } )
    self:FinalizeRun()
    self:ShowGameOverScreen()
end

function Game:Win( result )
    assert( result == nil or IsEnum( result, GAMEOVER ))
    self:GetGameState():SetGameOver( { result = result or GAMEOVER.VICTORY } )
    self:FinalizeRun()
    self:ShowGameOverScreen()
end

function Game:FinalizeRun()
    local gamestate = self.gamestate
    local game_type = gamestate:GetGameType()

    if game_type == GAME_TYPE.DAILY then
        gamestate.score = CalculateDailyScore( gamestate )
        if gamestate:IsVictory() then
            self:AddGameplayStat( "complete_daily", 1 )
        end

    elseif game_type == GAME_TYPE.BRAWL and gamestate:IsVictory() then
        self:AddGameplayStat( "complete_brawl", 1 )

    elseif game_type == GAME_TYPE.CAMPAIGN and gamestate:IsVictory() then
        self:AddGameplayStat( "complete_campaign", 1 )
    end

    -- Add this game to the history
    if not gamestate.options.experiment then
        self.run_history:AddRun( gamestate )
    end

    if gamestate:IsVictory() then
        if (game_type == GAME_TYPE.BRAWL or game_type == GAME_TYPE.CAMPAIGN) then
            if StatUtil.CountStat( "times_upgraded" ) == 0 then
                self:UnlockAchievement( "no_upgrades" )
            end

            if AgentUtil.IsFullyUpgraded( gamestate:GetPlayerAgent() ) then
                self:UnlockAchievement( "total_upgrades" )
            end
 
            if StatUtil.CountStat( "times_installed" ) == 0 and #(gamestate:GetOptions().perks or table.empty) == 0 then
                self:UnlockAchievement( "just_cards" )
            end

            local def = Content.FindAchievement( "efficiency" )
            if def and gamestate:GetPlayerAgent().battler:GetCardCount() <= def.count then
                self:UnlockAchievement( def.id )
            end
        end

        if game_type == GAME_TYPE.CAMPAIGN then
            local def = Content.FindAchievement( "to_the_oshnudrome" )
            if def and gamestate:GetPlayTime() <= def.count then
                self:UnlockAchievement( def.id )
            end
        end

        if game_type == GAME_TYPE.BRAWL and gamestate:GetOptions().advancement_level == #gamestate:GetCurrentBackground().data.advancement then
            self:UnlockAchievement( "brawler" )
        end

        -- Track which perks were used to win.
        TheGame:GetGameProfile():SetPlayedPerks( gamestate )
    end

    return self
end

function Game:ShowGameOverScreen()
    if self:IsDemo() then
        UIHelpers.ShowDemoMessage()
    else
        -- We may be already be fading out, if we're coming out of a slideshow.  But not always (eg. Brawl)
        if TheGame:FE():IsFading() then
            if self:GetGameState():GetGameType() == GAME_TYPE.DAILY then
                local screen = Screen.DailyOverScreen( self:GetGameState() )
                TheGame:FE():PushScreen( screen )
            else
                local game_over = self:GetGameState():GetGameOver()
                if game_over and game_over.result == GAMEOVER.DEATH and self.gamestate:GetRollbackSave() ~= nil then
                    local screen = Screen.PickLegacyScreen( self:GetGameState():GetSaveFileName(), self.gamestate)
                    self.frontend:PushScreen( screen )
                else
                    local screen = Screen.GameOverScreen( self:GetGameState() )
                    self.frontend:PushScreen( screen )
                end
            end
        else
            TheGame:FE():FadeTransition( function() self:ShowGameOverScreen() end )
        end
    end
    return self
end

function Game:GetRunHistory()
    return self.run_history:GetHistory()
end

function Game:SubmitLog(msg, extra_txt)
-- No longer useful for us, so turned off
--     if self:GetGameState() then
--         if self:GetGameState().save_file_name and self:GetGameState().save_file_name ~= "autosave" then
--             local gamestatus = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerStateString() or "---"
--             engine.inst:CreateFeedback(msg, gamestatus .. "\n" .. (extra_txt or "" ), TheGame:GetGameProfile():GetFeedbackName() or engine.inst:GetUserName(), "GAMEPLAY", gamestatus, 0, true, false, false, "checkpoints")
--         end
--     end
end

function Game:GetRoboticsSettings()
    return self.robotics_settings
end

function Game:CamShakeEnabled()
    return not engine.inst:GetSetting( "graphics", "shake_disabled", false )
end

function Game:ShowCardToolTipsOnFocus()
    if self.card_tooltip_on_focus == nil then
        self.card_tooltip_on_focus = engine.inst:GetSetting( "gameplay", "card_tooltip_on_focus", false )
    end
    return self.card_tooltip_on_focus == true
end

function Game:ScreenFlashEnabled()
    return not engine.inst:GetSetting( "graphics", "flash_disabled", false )
end

function Game:ScreenSwayEnabled()
    if not self.frontend or self.frontend:IsGamepadMode() then
        return false
    else
        return not engine.inst:GetSetting( "graphics", "sway_disabled", false )
    end
end

function Game:UnlockAchievement( id )
    -- if BRANCH ~= "dev" then return end

    if not self.gamestate then
        return
    end

    local options = self.gamestate:GetOptions()
    if options.experiment or (options.mutators and #options.mutators > 0) then
        return false
    end

    if self:IsDemo() then
        return false
    end
    
    if self.gamestate:GetCurrentAct() == nil then
        return
    end

    local def = Content.FindAchievement( id )
    assert_warning( def, "Missing achievement definition: '%s'", tostring(id) )
    if def and not self.game_profile:IsAchievementUnlocked( id ) then
        print( "Achievement Complete:", id )
        self.game_profile:UnlockAchievement( id )
        self.game_profile:SaveSettings()
        if engine.steam then
            engine.steam:UnlockAchievement( id )
        end

        local tex = def.icon
        local title = loc.format( LOC"NOTIFICATION.GRIFT_COMPLETE", def:GetLocalizedName() )
        local desc = def:GetLocalizedDesc()

        TheGame:FE():ShowAchievementToast( tex, title, desc )
    end
end

function Game:CheckGriftProgress()
    local count = 0
    for i, def in ipairs( Content.GetAllGrifts() ) do
        if not def:IsComplete() and def:CanComplete() then
            def:Complete()
            count = count + 1
        end
    end
    if count > 0 then
        self.game_stats:SaveSettings()
        self.game_profile:SaveSettings()
    end
end
