local UNLOCK_THRESH = 100

---------------------------------------------------------------
-- Meta-game progression management.

local GameProfile = class( "GameProfile", UserSettings )

function GameProfile:init()
    UserSettings.init( self, "profile.lua", engine.settings )
    self:SetDefaults()

    if not self:GetSetting("first_build_played") then
        local build = tonumber(BUILD_REVISION)
        self:ApplySetting("first_build_played", build)
        self:SaveSettings()
    end

end

function GameProfile:SetDefaults()
    self:VerifySetting( "seen_plaxes", {} )
    self:VerifySetting( "unlock_data", {} )
    self:VerifySetting( "acquired_unlocks", {} )
    self:VerifySetting( "advancement", {} )
    self:VerifySetting( "selected_mutators", {} )
    self:VerifySetting( "played_perks", {} )
    self:VerifySetting( "selected_perks", {} )
    self:VerifySetting( "outfits", {} )
    self:VerifySetting( "played_advancement", {} )
    self:VerifySetting( "selected_advancement", {} )
    self:VerifySetting( "has_seen_messages", {} )
    self:VerifySetting( "narrative_progress", {} )
    self:VerifySetting( "toggled_off_unlocks", {} )
    self:VerifySetting( "character_unlocks", {} )
    self:VerifySetting( "graft_unlocks", {} )
    self:VerifySetting( "grifts", {} )
    self:VerifySetting( "achievements", {} )
    self:VerifySetting( "perk_points", 0 )
    self:VerifySetting( "perk_unlocks", {} )
    self:VerifySetting( "incidents", {} )
    self:VerifySetting( "mod_settings", {} )
    self:VerifySetting( "no_streak_random", {})
    self:VerifySetting( "draft_bounties", {})
    self:VerifySetting( "mettle_points", {} )
    self:VerifySetting( "mettle_upgrades", {} )
    self:VerifySetting( "unlocked_flourishes", {
        SAL = {
          BATTLE = {},
          NEGOTIATION = {},
        },
        ROOK = {
          BATTLE = {},
          NEGOTIATION = {},
        },
        SMITH = {
          BATTLE = {},
          NEGOTIATION = {},
        }
    })
    self:VerifySetting( "prestige_points", 0 )
    self:VerifySetting( "boss_advancement", {} )
end

function GameProfile:GetFirstBuild()
    return self:GetSetting("first_build_played") or 0
end

function GameProfile:GetNumUnlocked( pool )
    pool = pool or "SAL"
    local track = Content.GetUnlockTrack(pool)
    local pts = self:GetPoints(pool)

    if track then
        for k, unlock in ipairs(track) do
            if unlock.total_pts > pts then
                return k-1
            end
        end
        return #track
    end

    return 0
end

function GameProfile:GetLastCharacterPlayed()
    return self:GetSetting( "last_character_used" )
end

function GameProfile:SetLastCharacterPlayed(id)
    if self:GetSetting("last_character_used") ~= id then
        self:ApplySetting("last_character_used", id)
        self.dirty = true
    end
end

function GameProfile:IncreaseDraftBounty(id)
    local bounties = self:GetSetting( "draft_bounties" )
    bounties[id] = (bounties[id] or 0) + 1
    self.dirty = true
end

function GameProfile:ClearDraftBounty(id)
    local bounties = self:GetSetting( "draft_bounties" )
    bounties[id] = 0
end

function GameProfile:GetDraftBounty(id)
    local bounties = self:GetSetting( "draft_bounties" )
    return bounties[id] or 0
end

function GameProfile:GetCurrentSaveGame()
    return self:GetSetting( "current_save_game" )
end

function GameProfile:SetCurrentSaveGame(id)
    self:ApplySetting("current_save_game", id)
    self:SaveSettings()
end

function GameProfile:SeePlax( plax_id )
    local seen_plaxes = self:GetSetting( "seen_plaxes" )
    if not seen_plaxes[plax_id] then
        seen_plaxes[plax_id] = true
        self.dirty = true
    end
    --self:SaveSettings()
end

function GameProfile:HasSeenPlax( plax_id )
    local seen_plaxes = self:GetSetting( "seen_plaxes" )
    return seen_plaxes[plax_id]
end


function GameProfile:IsCharacterLocked( character_id )
    local character_data = GetPlayerBackground(character_id)
    if character_data == nil then
        return false
    end
    -- assert( character_data, tostring(character_id) )
    
    local unlock_id = character_data.data and character_data.data.unlock_id
    if unlock_id and not TheGame:GetGameProfile():HasUnlock(unlock_id) then
        return true
    end
    return false
end

function GameProfile:IsActLocked(act_id)
    local act_data = GetPlayerActData(act_id)
    local unlock_id = act_data.data and act_data.data.unlock_id and act_data.data.unlock_id
    if unlock_id and not TheGame:GetGameProfile():HasUnlock(unlock_id) then
        return true
    end
    return false
end


function GameProfile:GetFeedbackName()
    return self:GetSetting( "feedback_name") 
end


function GameProfile:SetFeedbackName(name)
    name = name or ""
    if type(name) == "string" and self:GetSetting( "feedback_name ") ~= name then
        self:ApplySetting("feedback_name", name)
        self:SaveSettings()
    end
end

function GameProfile:GetPoints( pool )
    local data = self:GetSetting( "unlock_data" )
    return data and data[pool] and data[pool].pts or 0
end


function GameProfile:GetPerkPoints()
    return self:GetSetting( "perk_points" ) or 0
end

function GameProfile:AddPerkPoints( points )
    self:ApplySetting( "perk_points", self:GetPerkPoints() + math.max( 0, points ) )
end

function GameProfile:BuyPerk( id )
    local perk_unlocks = self:GetSetting( "perk_unlocks" )
    local def = Content.GetGraft(id)
    
    if def and not perk_unlocks[id] then
        local cost = def.unlock_cost or 0
        if cost <= self:GetPerkPoints() then
            self:ApplySetting( "perk_points", self:GetPerkPoints() - cost )
            perk_unlocks[id] = true
            self:SaveSettings()            
            return true
        end
    end
    return false
end

function GameProfile:GetPerkUnlocks()
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        local all_perks = {}
        for i, def in ipairs( Content.GetAllGrafts()) do
            if def.type == GRAFT_TYPE.PERK then
                all_perks[ def.id ] = true
            end
        end
        return all_perks

    else
        return shallowcopy(self:GetSetting( "perk_unlocks" ))
    end
end

function GameProfile:HasPerkUnlocked( id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    local data = self:GetSetting( "perk_unlocks" )
    return data[ id ] == true
end


function GameProfile:IsGriftComplete( id )
    local data = self:GetSetting( "grifts" )
    return data[ id ] == true
end

function GameProfile:CompleteGrift( id )
    local data = self:GetSetting( "grifts" )
    data[ id ] = true
    self.dirty = true
end

function GameProfile:IsAchievementUnlocked( id )
    local data = self:GetSetting( "achievements" )
    return data[ id ] == true
end

function GameProfile:GetAchievementUnlocks()
    local data = copykeys( self:GetSetting( "achievements" ) or {} )
    return data
end

function GameProfile:UnlockAchievement( id )
    local data = self:GetSetting( "achievements" )
    data[ id ] = true
    self.dirty = true
end

function GameProfile:GetNextUnlock(pool)
    local start_pts = self:GetPoints(pool)
    local track = Content.GetUnlockTrack(pool)
    if track then
        for k, unlock in ipairs(track) do
            if unlock.total_pts > start_pts then
                return unlock, k
            end
        end
    end
end

function GameProfile:DeltaPoints( pool, delta )
    local unlocks_awarded = {}
    local next_award

    local track = Content.GetUnlockTrack(pool)
    local start_pts = self:GetPoints(pool)
    local end_pts = start_pts + delta

    local total_pts = 0

    if track then
        for k, unlock in ipairs(track) do
            total_pts = total_pts + unlock.pts
            if total_pts > end_pts then
                next_award = unlock
                break
            elseif total_pts > start_pts then
                table.insert(unlocks_awarded, unlock)
                self:AcquireUnlockGroup( pool, unlock )
            end
        end
    end

    local all_points = self:GetSetting( "unlock_data" )
    all_points[pool] = all_points[pool] or {}
    all_points[pool].pts = end_pts
    self.dirty = true

    self:SaveSettings()

    
    return unlocks_awarded, next_award
end 

function GameProfile:AcquireUnlockGroup( pool, unlock_group )
    local unlocks = self:GetSetting( "acquired_unlocks" )

    if unlock_group.items then
        for k,v in ipairs(unlock_group.items) do
            if table.insert_unique( unlocks, v.unlock_id ) then
                self.dirty = true
            end
        end
    end
end

function GameProfile:CheckUnlocks()
    local all_points = self:GetSetting( "unlock_data" )
    for character_id, data in pairs(all_points) do
        local track = Content.GetUnlockTrack(character_id)
        
        local pts = self:GetPoints(character_id)
        if track and pts and pts > 0 then
            for k, unlock in ipairs(track) do
                if unlock.total_pts > pts then
                    break
                end
                self:AcquireUnlockGroup( character_id, unlock )
            end
        end
    end
    self:SaveSettings()
end

function GameProfile:IsUnlocked( card_def )
    return not Content.IsUnlock( card_def.id ) or self:HasUnlock( card_def.id )
end

function GameProfile:HasUnlock( unlock_id, gamestate )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    local unlocks = self:GetSetting( "acquired_unlocks" )
    if table.arrayfind( unlocks, unlock_id ) ~= nil then
        -- The player has unlocked this.. but they might have they toggled it off..
        if gamestate then
            -- DBG( gamestate )
            local act = 0
            
            -- List of unlocks the user doesn't want to use.
            local toggled_off_unlocks = gamestate.options.toggled_off_unlocks
            -- Go through all the unlocks
            if gamestate:GetCurrentAct() then
                local unlock_track = Content.GetUnlockTrack( gamestate:GetPlayerBackgroundID() )
                for k, unlock in ipairs( unlock_track ) do
                    -- Check if the current unlock is on the toggled_off list
                    if table.arrayfind( toggled_off_unlocks, unlock.id ) ~= nil then
                        -- It is, so any items here are a no-go
                        for j, item in ipairs( unlock.items ) do
                            -- Look for the current card in the list
                            if ( item.battle_card or item.negotiation_card ) and item.unlock_id == unlock_id then
                                -- We found the card! It means the player doesn't want to use this one
                                return false
                            end
                        end
                    end
                end
            end
            -- The item wasn't there, I guess it is allowed then!
            return true
        else
            -- No list of toggled-off unlocks was sent, and the user has unlocked this, so yeah, they have this!
            return true
        end
    else
        -- Not found on the fancy list of unlocks, tsk tsk
        return false
    end
end

function GameProfile:AcquireUnlock( unlock_id )
    local unlocks = self:GetSetting( "acquired_unlocks" )
    if table.insert_unique( unlocks, unlock_id ) then
        self.dirty = true
    end
    self:SaveSettings()
end

-- Returns the unlocks the player has toggled off
function GameProfile:GetToggledOffUnlocks( act_id, gamestate )
    
    if gamestate and gamestate.options then
        return gamestate.options.toggled_off_unlocks or {}
    end

    local toggled_off_unlocks = self:GetSetting( "toggled_off_unlocks" ) 
    toggled_off_unlocks = toggled_off_unlocks and toggled_off_unlocks[act_id] or table.empty
    local unlocks_track = Content.GetUnlockTrack( act_id )
    local num_unlocked = TheGame:GetGameProfile():GetNumUnlocked( act_id )
    
    local retval = {}
    
    for k, unlock in ipairs( unlocks_track ) do
        local toggled_off = table.arrayfind(toggled_off_unlocks, unlock.id) ~= nil
        --print (k, unlock.id, toggled_off, num_unlocked)
        if k <= num_unlocked and toggled_off then 
            table.insert( retval, unlock.id ) 
        end
    end

    return retval
end

-- Returns the unlocks the player has toggled on
function GameProfile:GetToggledOnUnlocks( act_id, gamestate )
    local unlocks_track = Content.GetUnlockTrack( act_id )
    local num_unlocked = TheGame:GetGameProfile():GetNumUnlocked( act_id )
    local toggled_off_unlocks = self:GetToggledOffUnlocks( act_id, gamestate )
    local toggled_on_unlocks = {}
    for k, unlock in ipairs( unlocks_track ) do
        local toggled_off = table.arrayfind( toggled_off_unlocks, unlock.id ) ~= nil
        if k <= num_unlocked and not toggled_off then table.insert( toggled_on_unlocks, unlock.id ) end
    end
    return toggled_on_unlocks or table.empty
end

function GameProfile:IsUnlockToggledOff( act_id, unlock_id, gamestate )
    local toggled_off_unlocks = self:GetToggledOffUnlocks( act_id, gamestate )
    if toggled_off_unlocks then
        return table.arrayfind( toggled_off_unlocks, unlock_id ) ~= nil
    end
    return false
end

function GameProfile:SetUnlockToggledOff( act_id, unlock_id, toggled_off )
    local toggled_off_unlocks = self:GetSetting( "toggled_off_unlocks" )
    if toggled_off then
        toggled_off_unlocks[act_id] =  toggled_off_unlocks[act_id] or {}
        if table.insert_unique( toggled_off_unlocks[act_id], unlock_id ) then
            self.dirty = true
        end
        --self:SaveSettings()

    elseif toggled_off_unlocks and toggled_off_unlocks[act_id] then
        if table.arrayremove( toggled_off_unlocks[act_id], unlock_id ) then
            self.dirty = true
            --self:SaveSettings()
        end
    end
end

function GameProfile:GetNarrativeProgress( unlock_id )
    local narrative_progress = self:GetSetting( "narrative_progress" )
    return narrative_progress[unlock_id] or 0
end

function GameProfile:IncNarrativeProgress( unlock_id )
    local narrative_progress = self:GetSetting( "narrative_progress" )
    narrative_progress[unlock_id] = (narrative_progress[unlock_id] or 0) + 1
    self.dirty = true
end

function GameProfile:SetSelectedPerks( act_id, perk_ids )
    local perks_data = self:GetSetting( "selected_perks" )
    perks_data[act_id] = perk_ids
    self.dirty = true
end

function GameProfile:GetNumPerksUnlocked()
    local perk_unlocks = self:GetPerkUnlocks()
    return table.count( perk_unlocks )
end


function GameProfile:GetNumPerkSlots()
    local num_perks = self:GetNumPerksUnlocked()
    local num_unlocked = 0
    
    for k,v in ipairs(PERK_SLOT_UNLOCK_THRESHOLDS) do
        if num_perks >= v then
            num_unlocked = k
        end
    end
    return num_unlocked
end


function GameProfile:GetSelectedPerks( act_id )
    local perks_data = self:GetSetting( "selected_perks" )
    return perks_data[act_id] or table.empty
end

-- When winning a run, call this to save what perks you've completed the game with
function GameProfile:SetPlayedPerks( gamestate )

    -- Get played-perks list
    local perks_data = self:GetSetting( "played_perks" )

    -- Get currently active perks, if any
    if gamestate and gamestate:GetPlayerAgent() then
        local player_agent = gamestate:GetPlayerAgent()
        local graft_owner = player_agent:GetAspect("graft_owner")
        local grafts = graft_owner:GetGrafts( GRAFT_TYPE.PERK )
        if #grafts > 0 then
            for k, graft in ipairs( grafts ) do
                table.insert_unique( perks_data, grafts[k].id )
            end
        end
    end
    self.dirty = true
    self:SaveSettings()
end

function GameProfile:SetPlayedPerk( perk_id )

    -- Get played-perks list
    local perks_data = self:GetSetting( "played_perks" )
    
    if perk_id then
        table.insert_unique( perks_data, perk_id )
    end

    self.dirty = true
    self:SaveSettings()
end

function GameProfile:IsPerkPlayed( perk_id )
    local perks_data = self:GetSetting( "played_perks" )
    return table.arrayfind( perks_data, perk_id ) ~= nil
end

function GameProfile:SetSelectedMutators( act_id, mutator_ids )
    local mutators_data = self:GetSetting( "selected_mutators" )
    mutators_data[act_id] = mutator_ids
    self.dirty = true
    --self:SaveSettings()
end

function GameProfile:GetSelectedMutators( act_id )
    if self:HasUnlock("MUTATORS_UNLOCKED") then
        local mutators_data = self:GetSetting( "selected_mutators" )
        return mutators_data[act_id]
    else
        return table.empty
    end
end

function GameProfile:SetSelectedOutfit( act_id, outfit_id )
    local outfits_data = self:GetSetting( "outfits" )
    outfits_data[act_id] = outfits_data[act_id] or {}
    outfits_data[act_id].selected_outfit_id = outfit_id
    self.dirty = true
    --self:SaveSettings()
end

function GameProfile:GetSelectedOutfit( act_id )
    local outfits_data = self:GetSetting( "outfits" )
    outfits_data[act_id] = outfits_data[act_id] or {}
    return outfits_data[act_id].selected_outfit_id
end

function GameProfile:SetSeenOutfit( act_id, outfit_id )
    local outfits_data = self:GetSetting( "outfits" )
    outfits_data[act_id] = outfits_data[act_id] or {}
    outfits_data[act_id].seen_outfits = outfits_data[act_id].seen_outfits or {}
    if not table.arrayfind( outfits_data[act_id].seen_outfits, outfit_id ) then
        table.insert_unique( outfits_data[act_id].seen_outfits, outfit_id )
        self.dirty = true
    end
end


function GameProfile:IsOutfitSeen( act_id, outfit_id )
    local outfits_data = self:GetSetting( "outfits" )
    outfits_data[act_id] = outfits_data[act_id] or {}
    outfits_data[act_id].seen_outfits = outfits_data[act_id].seen_outfits or {}
    return table.arrayfind( outfits_data[act_id].seen_outfits, outfit_id ) ~= nil
end


function GameProfile:AcquireAdvancement( act_id, played_level )
    local advancement = self:GetSetting( "advancement" ) -- Advancement level the player's unlocked
    local last_advancement = advancement[act_id]
    local max_advancement = self:GetMaxAdvancement( act_id )

    -- The max prestige level that is saved is 1 + the current max advancement (to indicate
    -- the player has defeated the final prestige)
    advancement[act_id] = clamp( played_level + 1, advancement[act_id] or 0, max_advancement + 1 )

    self:SetHighestPlayedAdvancement( act_id, played_level )
end

function GameProfile:SetHighestPlayedAdvancement( act_id, advancement )
    local played_advancement = self:GetSetting( "played_advancement" ) -- Advancement level the player's played
    if played_advancement[act_id] == nil or advancement > played_advancement[act_id] then
        played_advancement[act_id] = advancement
        self.dirty = true
        self:SaveSettings()
    end
end

function GameProfile:GetHighestPlayedAdvancement( act_id )
    local played_advancement = self:GetSetting( "played_advancement" ) -- Advancement level the player's played
    return played_advancement[act_id] or 0
end

function GameProfile:GetMaxAdvancement( act_id )
    local act_data = GetPlayerActData( act_id )
    return #act_data.background.data.advancement
end


function GameProfile:GetAdvancementLevel( act_id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        -- Return the maximum advancement for this act
        return self:GetMaxAdvancement( act_id )
    end

    local advancement = self:GetSetting( "advancement" )
    if advancement[act_id] then
        return math.min( advancement[act_id], self:GetMaxAdvancement( act_id ) )
    end
    
    return 0
end

function GameProfile:HasSeenMessage(id)
    return self:GetSetting( "has_seen_messages" )[id]
end

function GameProfile:SetHasSeenMessage(id)
    if not self:HasSeenMessage(id) then
        self:GetSetting( "has_seen_messages" )[id] = true
        self.dirty = true
        self:SaveSettings()
    end
end

function GameProfile:GetSelectedAdvancement( act_id )
    local max_advancement = TheGame:GetGameProfile():GetMaxAdvancement( act_id )
    local selected_advancement = self:GetSetting( "selected_advancement" )
    return math.min(max_advancement, selected_advancement[act_id] or 0)
end



function GameProfile:UpdateSelectedAdvancement( act_id, advancement )
    local selected_advancement = self:GetSetting( "selected_advancement" )
    if selected_advancement[act_id] ~= advancement then
        selected_advancement[act_id] = advancement
        self.dirty = true
    end
end


function GameProfile:Reset()
    self:ClearSettings()
    self:SetDefaults()
    self:SaveSettings()
end

-------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------
-- AGENT META UNLOCKS
-------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Player has met this agent
function GameProfile:HasMetCharacter( skin_id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    if skin_id ~= nil then
        return self:GetSetting( "character_unlocks" )[skin_id] ~= nil
    end
end
function GameProfile:SetMetCharacter( skin_id )
    if skin_id ~= nil then
        local character_unlocks = self:GetSetting( "character_unlocks" )
        if not character_unlocks[skin_id] then
            character_unlocks[skin_id] = character_unlocks[skin_id] or {}
            self.dirty = true
        end
    end
end

-- Player has triggered a custom agent unlock
function GameProfile:SetCustomAgentUnlock( skin_id, unlock_id )
    if skin_id ~= nil then
        local character_unlocks = self:GetSetting( "character_unlocks" )
        if not character_unlocks[skin_id] then
            character_unlocks[skin_id] = character_unlocks[skin_id] or {}
        end
        character_unlocks[skin_id][unlock_id] = true
        self.dirty = true
    end
end

function GameProfile:HasCustomAgentUnlock( skin_id, unlock_id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    if skin_id ~= nil then
        local character_unlocks = self:GetSetting( "character_unlocks" )
        if character_unlocks[skin_id] and character_unlocks[skin_id][unlock_id] then return true end
    end
    return false
end

-- Player was loved by this agent
function GameProfile:SetLovedByAgent( skin_id )
    self:SetCustomAgentUnlock( skin_id, "loved_by" )
end
function GameProfile:HasLovedByAgent( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "loved_by" )
end

-- Player was hated by this agent
function GameProfile:SetHatedByAgent( skin_id )
    self:SetCustomAgentUnlock( skin_id, "hated_by" )
end
function GameProfile:HasHatedByAgent( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "hated_by" )
end

-- Player was killed by this agent
function GameProfile:SetKilledByAgent( skin_id )
    self:SetCustomAgentUnlock( skin_id, "killed_by" )
end
function GameProfile:HasKilledByAgent( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "killed_by" )
end

-- Player killed this agent
function GameProfile:SetKilledAgent( skin_id )
    self:SetCustomAgentUnlock( skin_id, "killed" )
end
function GameProfile:HasKilledAgent( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "killed" )
end

-- Player had this agent's social boon
function GameProfile:SetUnlockedBoon( skin_id )
    self:SetCustomAgentUnlock( skin_id, "unlocked_boon" )
end
function GameProfile:HasUnlockedBoon( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "unlocked_boon" )
end

-- Player had this agent's social bane
function GameProfile:SetUnlockedBane( skin_id )
    self:SetCustomAgentUnlock( skin_id, "unlocked_bane" )
end
function GameProfile:HasUnlockedBane( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "unlocked_bane" )
end

-- Player had this agent's social bane
function GameProfile:SetDrankWith( skin_id )
    self:SetCustomAgentUnlock( skin_id, "drank_with" )
end
function GameProfile:HasDrankWith( skin_id )
    return self:HasCustomAgentUnlock( skin_id, "drank_with" )
end

function GameProfile:GetModSettings()
    return self:GetSetting( "mod_settings")
end


-------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------
-- GRAFT META UNLOCKS
-------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------

-- Player has met this agent
function GameProfile:HasSeenGraft( graft_id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    if graft_id ~= nil then
        return self:GetSetting( "graft_unlocks" )[graft_id] ~= nil
    end
end
function GameProfile:SetSeenGraft( graft_id )
    if graft_id ~= nil then
        local graft_unlocks = self:GetSetting( "graft_unlocks" )
        if not graft_unlocks[graft_id] then
            graft_unlocks[graft_id] = graft_unlocks[graft_id] or {}
            self.dirty = true
        end
    end
end

function GameProfile:SetCustomGraftUnlock( graft_id, unlock_id )
    if graft_id ~= nil then
        local graft_unlocks = self:GetSetting( "graft_unlocks" )
        if not graft_unlocks[graft_id] then
            graft_unlocks[graft_id] = graft_unlocks[graft_id] or {}
        end
        graft_unlocks[graft_id][unlock_id] = true
        self.dirty = true
    end
end

function GameProfile:HasCustomGraftUnlock( graft_id, unlock_id )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        return true
    end

    if graft_id ~= nil then
        local graft_unlocks = self:GetSetting( "graft_unlocks" )
        if graft_unlocks[graft_id] and graft_unlocks[graft_id][unlock_id] then return true end
    end
    return false
end

-- Player has installed this graft
function GameProfile:SetInstalledGraft( graft_id )
    self:SetCustomGraftUnlock( graft_id, "installed" )
end

function GameProfile:HasInstalledGraft( graft_id )
    return self:HasCustomGraftUnlock( graft_id, "installed" )
end

function GameProfile:RecordIncident( category_id, incident_id )
    local MAX_HISTORY = 100
    local incidents = self:GetSetting( "incidents" )
    incidents.history = incidents.history or {}
    incidents.occurances = incidents.occurances or {}
    
    incidents.history[category_id] = incidents.history[category_id] or {}
    table.insert(incidents.history[category_id], 1, incident_id)
    if #incidents.history[category_id] > MAX_HISTORY then
        table.remove(incidents.history[category_id])
    end
    
    incidents.occurances[category_id] = incidents.occurances[category_id] or {}
    incidents.occurances[category_id][incident_id] = (incidents.occurances[category_id][incident_id] or 0) + 1
    self.dirty = true
end

function GameProfile:HasEverSeenIncident( category_id, incident_id )
    local incidents = self:GetSetting( "incidents" )
    return incidents.occurances and incidents.occurances[category_id] and incidents.occurances[category_id][incident_id] and incidents.occurances[category_id][incident_id] > 0 
end

function GameProfile:HasSeenIncidentRecently( category_id, incident_id, recent_limit )

    local incidents = self:GetSetting( "incidents" )
    incidents.history = incidents.history or {}
    if incidents.history[category_id] then
        math.min(recent_limit, #incidents.history[category_id])
        for k = 1, math.min(recent_limit, #incidents.history[category_id]) do
            if incidents.history[category_id][k] == incident_id then
                return true
            end
        end
    end
    return false
end


function GameProfile:GetNoStreakRandom( id, options_in, max_streak )
    
    local options = shallowcopy(options_in)

    local memory = self:GetSetting( "no_streak_random" )
    memory[id] = memory[id] or {}

    local value     

    local unseen_options = {}
    memory[id].seen = memory[id].seen or {}
    for k,v in ipairs(options) do
        if not memory[id].seen[v] then
            table.insert(unseen_options, v)
        end
    end

    if #unseen_options > 0 then
        value = table.arraypick(unseen_options)
    end

    if not value then
        if memory[id].current_streak_length and memory[id].current_streak_length >= max_streak then
            table.arrayremove(options, memory[id].current_streak_value)
        end
        value = table.arraypick(options) or table.arraypick(options_in)
    end

    local new_value = value ~= memory[id].current_streak_value
    memory[id].current_streak_value = value
    memory[id].current_streak_length = new_value and 1 or memory[id].current_streak_length + 1
    memory[id].seen = memory[id].seen or {}
    memory[id].seen[value] = (memory[id].seen[value] or 0 ) + 1
    self.dirty = true

    return value
end

function GameProfile:GetMettlePoints( character_id )
    local mettle_points = self:GetSetting( "mettle_points" )
    return mettle_points[character_id] or 0
end

function GameProfile:AddMettlePoints( character_id, points, source )
    local mettle_points = self:GetSetting( "mettle_points" )
    mettle_points[character_id] = (mettle_points[character_id] or 0) + points
    self.dirty = true
    --self:SaveSettings() --specifically don't save here, to prevent reliable save-scum mettle farming
    local json_t = { PLAYER_DATA = TheGame:GetGameState():GetPlayerState(), METTLE_DELTA = points, METTLE_SOURCE = type(source) == "string" and source or "" }
    local json_data = json.encode( json_t )
    engine.inst:SendMetricsData("GAINED_METTLE", json_data)
    TheGame:GetEvents():BroadcastEvent( "mettle_gained" )
    if TheGame:GetGameState() then
        TheGame:GetGameState():LogNotification( NOTIFY.GAINED_METTLE, points )
    end
end

function GameProfile:SpendMettlePoints( character_id, points, source )
    local mettle_points = self:GetSetting( "mettle_points" )
    mettle_points[character_id] = (mettle_points[character_id] or 0) - points
    self.dirty = true
    self:SaveSettings()
    local json_t = { PLAYER_DATA = TheGame:GetGameState():GetPlayerState(), METTLE_DELTA = points, METTLE_SOURCE = type(source) == "string" and source or "" }
    local json_data = json.encode( json_t )
    engine.inst:SendMetricsData("SPENT_METTLE", json_data)
    TheGame:GetEvents():BroadcastEvent( "mettle_spent" )
end

function GameProfile:HasMettleUnlocked( character_id )
    local mettle_points = self:GetSetting( "mettle_points" )
    return mettle_points[character_id] ~= nil
end

function GameProfile:UnlockMettle( character_id )
    local mettle_points = self:GetSetting( "mettle_points" )
    if mettle_points[character_id] == nil then
        mettle_points[character_id] = 0
        self.dirty = true
        self:SaveSettings()
    end
end

function GameProfile:BuyMettleUpgrade( character_id, unlock_id, level )
    local mettle_upgrades = self:GetSetting( "mettle_upgrades" )
    mettle_upgrades[character_id] = mettle_upgrades[character_id] or {}
    mettle_upgrades[character_id][unlock_id] = level
    self.dirty = true
    self:SaveSettings()
end

function GameProfile:GetMettleUpgrades( character_id )
    local all_upgrades = shallowcopy(self:GetSetting( "mettle_upgrades" ))
        
    return all_upgrades[character_id] or table.empty
end

function GameProfile:UnlockFlourish( character_id, game_type, flourish_id )
    local unlocked_flourishes = self:GetSetting( "unlocked_flourishes" )
    unlocked_flourishes[character_id] = unlocked_flourishes[character_id] or {}
    unlocked_flourishes[character_id][game_type] = unlocked_flourishes[character_id][game_type] or {}
    table.insert(unlocked_flourishes[character_id][game_type], flourish_id)
    self.dirty = true
    self:SaveSettings()
end

function GameProfile:GetUnlockedFlourishes( character_id, game_type )
    if TheGame and TheGame:GetLocalSettings().ALL_UNLOCKS then
        local t 
        if game_type == "BATTLE" then
            t = Content.GetAllBattleFlourishes( character_id )
        else
            t = Content.GetAllNegotiationFlourishes( character_id )
        end
        return table.map(t, function(elem) return elem.id end)
    else
        local all_flourishes = shallowcopy(self:GetSetting( "unlocked_flourishes" ))
        if not all_flourishes[character_id] then
            return table.empty
        end

        return all_flourishes[character_id][game_type]
    end
end

function GameProfile:GetUsableFlourishes( character_id, game_type )
    local all_flourishes = shallowcopy(self:GetUnlockedFlourishes(character_id, game_type))
    local def = Content.GetCharacterDef( character_id )
    local basic_flourish = def and def.basic_flourishes and def.basic_flourishes[game_type]
    if basic_flourish then
        local to_remove = {}
        for i,id in ipairs(all_flourishes) do
            if id:match( "(.*)_ii.*$" ) then
                local base_id = string.gsub(id, "_ii", "")
                table.insert(to_remove,  base_id)
            end
            if id == basic_flourish then
                table.insert(to_remove, id)
            end
        end
        for i,id in ipairs(to_remove) do
            table.arrayremove(all_flourishes, id)
        end
        table.insert(all_flourishes, 1, basic_flourish)
    end
    return all_flourishes
end

function GameProfile:GetPrestigePoints()
    return self:GetSetting( "prestige_points" ) or 0
end

function GameProfile:AddPrestigePoints( points )
    self:ApplySetting( "prestige_points", self:GetPrestigePoints() + math.max( 0, points ) )
end

function GameProfile:SpendPrestigePoints( cost )
    self:ApplySetting( "prestige_points", math.max( 0, self:GetPrestigePoints() - cost) )
end

function GameProfile:VerifyPrestigePoints()
    local current_pp = self:GetPrestigePoints()
    local spent_pp = 0
    local earned_pp = 0
    local has_invalid = false
    for id,game_modes in pairs(self:GetSetting( "unlocked_flourishes" )) do
        for game,flourishes in pairs(game_modes) do
            local invalid_ids = {}
            for k,flourish_id in pairs(flourishes) do
                if game == "BATTLE" then
                    local carddef = Content.GetBattleCard(flourish_id)
                    if carddef then
                        spent_pp = spent_pp + (carddef.pp_cost or 0)
                    else
                        table.insert(invalid_ids, flourish_id)
                    end
                elseif game == "NEGOTIATION" then
                    local carddef = Content.GetNegotiationCard(flourish_id)
                    if carddef then
                        spent_pp = spent_pp + (carddef.pp_cost or 0)
                    else
                        table.insert(invalid_ids, flourish_id)
                    end
                end
            end
            if #invalid_ids > 0 then
                for i,v in ipairs(invalid_ids) do
                    table.arrayremove(flourishes, v)
                end
                has_invalid = true
            end
        end
    end
    if has_invalid then
        self.dirty = true
        self:SaveSettings()
    end
    

    for k,v in pairs(self:GetSetting( "boss_advancement" )) do
        for i=0, v do
            earned_pp = earned_pp + math.max(i, 1)
        end
    end
    if current_pp + spent_pp < earned_pp then
        self:AddPrestigePoints(earned_pp - (current_pp + spent_pp))
        self.dirty = true
        self:SaveSettings()
    end
end

function GameProfile:GetBossAdvancement( boss_id )
    local boss_advancement = self:GetSetting( "boss_advancement" )
    return boss_advancement[boss_id]
end

function GameProfile:SetBossAdvancement( boss_id, advancement )
    local boss_advancement = self:GetSetting( "boss_advancement" )
    local current = boss_advancement[boss_id]
    if current == nil or advancement > current then
        local gained = 0
        for i=(current or 0), advancement do
            gained = gained + math.max(i, 1)
        end
        self:AddPrestigePoints( gained )
        boss_advancement[boss_id] = advancement
        self.dirty = true
        self:SaveSettings()
    end
end

function GameProfile:VerifyPerkPoints()
    local current_pp = self:GetPerkPoints()
    local spent_pp = 0
    local earned_pp = 0
    for id,v in pairs(self:GetSetting( "perk_unlocks" )) do
        local def = Content.GetGraft( id )
        if def then
            spent_pp = spent_pp + def.unlock_cost
        end
    end
    for id,v in pairs(self:GetSetting( "grifts" )) do
        local def = Content.FindGrift( id )
        if def then
            earned_pp = earned_pp + def.perk_points
        end
    end
    if current_pp + spent_pp < earned_pp then
        print("ADDING MISSING PERK POINTS.", earned_pp - (current_pp + spent_pp))
        self:AddPerkPoints(earned_pp - (current_pp + spent_pp))
        self.dirty = true
        self:SaveSettings()
    end
end