
require "constants"
require "helpers"
require "quest"
require "region"
require "quest_util"
require "statutil"
require "map/caravan"
require "map/faction_mgr"
require "agent"
require "location"
require "notifications"
require "tracker"
require "scorekeeper"


---------------------------------------------------------------------------
-- GameState is the "in memory" portion of the player's save game state
-- that is outside the simulation.

----------------------------------------------------------------------------
SAVE_VERSION = 12 -- WARNING: Changing this will make save games incompatible!

----------------------------------------------------------------------------
AGENT_BUCKET = MakeEnum
{
    "LIMBO", -- Valid agent. Not available for spawn.
    "RESERVE", -- Valid agent. Availablity pool for spawning.
    "KILLEDBYPLAYER",

    -- Programmatic keys:
    --FactionID: Denotes faction membership.
    --RELATION: All agents with this player relationship status
}

---------------------------------------------------------------------------------

local GameState = class("GameState")
GameState.can_reload = true

function GameState.CreateOptions( start_act, advancement_level, character_config, toggled_off_unlocks, story_mode )
    
    local options =
    {
        advancement_level = advancement_level or 0,
        start_act = start_act or TheGame:GetLocalSettings().ACT_OVERRIDE or "SAL",
        custom = character_config,
        toggled_off_unlocks = toggled_off_unlocks or {},
        mutators = TheGame:GetLocalSettings().MUTATORS or {},
        required_mods = {},
        story_mode = story_mode or false
    }

    return options
end


local function GetScoreMultiplier( self )
    if self:GetGameType() == GAME_TYPE.DAILY then
        local prestige_level = self:GetAdvancementLevel()
        local difficulty_level = 0
        for difficulty, difficulty_data in ipairs(DAILY_DIFFICULTY_OPTIONS) do
            if difficulty_data.prestige == prestige_level then difficulty_level = difficulty end
        end
        local score_mult = DAILY_DIFFICULTY_OPTIONS[difficulty_level] and DAILY_DIFFICULTY_OPTIONS[difficulty_level].score_mult
        return score_mult or 1
    end
    return 1
end

function GameState:init(options, data)
    
    --don't do this, because we call this with pcall, and end up c crashing on mispaired profiler frames if we bail.
    --engine.inst:ProfilerPush( "GameState()" )

    assert( not ( data and not options )) -- "weird init. don't do that."

    if options == nil then
        options = GameState.CreateOptions()
    else
        options.toggled_off_unlocks = options.toggled_off_unlocks or {}
        options.advancement_level = options.advancement_level or 0
        options.story_mode = options.story_mode or false
    end

    self.options = options 

    self.scorekeeper = data and data.scorekeeper or ScoreKeeper( GetScoreMultiplier( self ) )
    self.uuid = data and data.uuid or engine.inst:NewGUID()
    self.world_region_id = data and data.world_region_id
    self.build_id = data and data.build_id or BUILD_ID_STRING
    self.creation_time = data and data.creation_time or os.time()
    self.play_time = data and data.play_time or 0
    self.afk_time = data and data.afk_time or 0
    self.current_act = data and data.current_act
    self.datetime = data and data.datetime or 0
    self.action_clock = data and data.action_clock or 0
    self.last_action_location = data and data.last_action_location
    self.last_action_datetime = data and data.last_action_datetime

    self.day_phase = data and data.day_phase or DAY_PHASE.DAY
    self.encounter_state = data and data.encounter_state
    
    self.current_difficulty = data and data.current_difficulty or 1
    self.card_skip_count = data and data.card_skip_count
    self.last_battle_scenario = data and data.last_battle_scenario
    if data and data.rollback_thresh then
        self.rollback_thresh = data.rollback_thresh
    elseif (options.advancement_level or 0) == 0 then
        self.rollback_thresh = 0
    else
        self.rollback_thresh = math.huge -- No rollback.
    end
    self.rollback_count = data and data.rollback_count or 0

    self.character_skins = data and data.character_skins
    if self.character_skins == nil then
        self.character_skins = {}
        for content_id, skins in pairs( Content.GetAllCharacterSkins() ) do
            self.character_skins[ content_id ] = {}
            for i, skin in ipairs(skins) do
                table.insert( self.character_skins[ content_id ], skin.uuid )
            end
        end
    end

    self.rollback_save = data and data.rollback_save
    self.next_id = data and data.next_id or 1
    if data and data.rng then
        self.rng = engine.Random( data.rng[1], data.rng[2] )
    elseif options.worldseed then
        self.rng = engine.Random( options.worldseed[1], options.worldseed[2] )
    else
        self.rng = engine.Random()
        self.rng:SetSeedFromTime()
    end

    if data and data.rngs then
        self.rngs = {}
        for id, rng in pairs( data.rngs ) do
            self.rngs[id] = engine.Random( rng[1], rng[2] )
        end
    else
        self:CreateRNG( "GRAFTS", self.rng:Random(2^32-1))
        self:CreateRNG( "CARDS", self.rng:Random(2^32-1))
        self:CreateRNG( "CARD_UPGRADES", self.rng:Random(2^32-1))        
    end

    self.notifications = data and data.notifications or Notifications()

    if data and data.factions then
        self.factions = data.factions
    else
        self.factions = FactionMgr( self )
        self.factions:AddFaction( PLAYER_FACTION, "DEFAULT_PLAYER" )
        self.factions:InitialPopulate()
    end

    self.player_agent = data and data.player_agent
    self.agents = data and data.agents or {}
    self.removed_agents = data and data.removed_agents or {}
    self.agent_buckets = data and data.agent_buckets or {}

    self.locations = data and data.locations or {}

    self.locations_by_uuid = {}
    for guid, location in pairs( self.locations ) do
        self.locations_by_uuid[ location:GetUUID() ] = location
    end

    self.seen_history = data and data.seen_history or {}
    self.act_progress_sub = data and data.act_progress_sub or 0
    self.act_progress = data and data.act_progress or 0
    if data == nil then
        self.act_progress_datetime = self.datetime
    else
        self.act_progress_datetime = data.act_progress_datetime

        if self.act_progress_datetime == nil then
            -- ccc: backwards save compability: derive starting datetime for current progress
            -- because there are always 2 phases per act_progress increment.
            self.act_progress_datetime = 2 * (self.act_progress - 1)
        end
    end

    self.caravan = data and data.caravan or Caravan()

    self.finished_quests = data and data.finished_quests or {}
    for guid, quest in pairs( self.finished_quests ) do
        if quest:GetQuestDef() == nil then
            print( string.format( "WARNING: %s does not exist", tostring(quest)))
            self.finished_quests[ guid ] = nil
        end
    end

    self.available_quests = data and data.available_quests or {}

    for guid, quest in pairs( self.available_quests ) do
        if quest:GetQuestDef() == nil then
            print( string.format( "WARNING: %s does not exist", tostring(quest)))
            self.available_quests[ guid ] = nil
        elseif quest:IsDone() then
            print( string.format( "WARNING: %s is already done; removing", tostring(quest)))
            self.available_quests[ guid ] = nil
        end
    end


    self.region = data and data.region 
    self.quest_stats = data and data.quest_stats or {}

    self:InitAgentTables( self.agents )

    self.tracker = Tracker( self )

    self.pending_events = data and data.pending_events or {}
    self.main_quest = data and data.main_quest

    self.card_stats = data and data.card_stats or {}
    

    --backwards compatibility
    --self.scoring = data and data.scoring or {}
    if data and data.scoring then
        for k,v in pairs(data.scoring) do
            self:AddScore(k, v)
        end
    end
    
    self.dialog_memory = data and data.dialog_memory or {}
end

function GameState.ConvertDatetimeToDay( datetime )
    return math.ceil( datetime/2 )
end

function GameState:ConvertAdvancement( level, old_build, new_build )
    print("converting advancement from ", old_build, "to", new_build)
    return ADVANCEMENT_CONVERSION[level]
end

function GameState:SetLastBattleScenario(scenario)
    self.last_battle_scenario = scenario
end

function GameState:RequireMod( mod, version )
    if type(mod) ~= "table" or type(mod.id) ~= "string" then
        LOGWARN( "RequireMod( mod, version ) expects mod '%s' as a mod table", tostring(mod))
        return
    end
    if version ~= nil and type(version) ~= "string" then
        LOGWARN( "RequireMod( mod_id, version ) expects version '%s' as nil or a string", tostring(version))
        return
    end

    table.insert( self.options.required_mods, { id = mod.id, title = mod.title, version = version })
end

function GameState:PostLoad()
    self.factions:PostLoad()
    self.caravan:PostLoad( self )
    
    -- TODO: the fact that these PostLoads occur in different orders due to pairs() is not great.
    for id, agent in pairs( self.agents ) do
        agent:PostLoad( self )
    end

    for id, location in pairs( self.locations ) do
        location:PostLoad( self )
    end
    
    local available_quests = shallowcopy(self.available_quests)
    for k, quest in pairs(available_quests) do
        quest:PostLoad()
    end
    self.tracker:PostLoad()

    -- Adjust the advancement for any players that created their save before the change to prestiges
    local options = self.options
    local creation_build_id  = self:GetCreationBuild()
    if options.advancement_level > 0 and creation_build_id <= 423983 and creation_build_id > 0 and options.advancement_build_id == nil then
        local _, _, current_build_id = string.find(BUILD_ID_STRING,  ".* r(%d+) .*" )
        print(">>> Converting advancement from ", creation_build_id, "to", current_build_id)
        options.advancement_level = ADVANCEMENT_CONVERSION[ options.advancement_level ] or 1
        options.advancement_build_id = tonumber(current_build_id)
    end
end

function GameState:GetCreationBuild()
    local _, _, creation_build_id  = string.find(self.build_id,  ".* r(%d+) .*" )
    creation_build_id = tonumber(creation_build_id) or 0
    return creation_build_id

end

function GameState:TryRollbackSave(thresh)
    if TheGame:IsDemo() then
        return
    end
    if not self.save_file_name then
        print( "Cannot save rollback, no save_file_name" )
        return
    end

    if self.rollback_thresh >= thresh then
        -- Already checkpointed.
        return
    end

    self.rollback_save = self.save_file_name .. "_rollback"
    self.rollback_thresh = thresh

    TheGame:SaveGame( self.rollback_save )
end


function GameState:UseRollback()
    -- Note: just track that a rollback was used, but don't delete it.  This has more flexibility (eg. regaining rollbacks, debugging, etc.)
    self.rollback_count = self.rollback_count + 1
end

function GameState:CanRollback()
    local max_rollbacks = 1
    return self.rollback_save ~= nil and self.rollback_count < max_rollbacks
end

function GameState:GetRollbackSave()
    return self.rollback_save
end

function GameState:GetGameType()
    local act_data = GetPlayerActData( self.options.start_act )
    return act_data and act_data.data.game_type
end

function GameState:SetSaveFileName(name)
    self.save_file_name = name
    return self
end

function GameState:GetSaveFileName()
    return self.save_file_name
end

function GameState:SetGameOver( results )
    assert_warning( self.game_over_results == nil, tostr(self.game_over_results) )

    -- Validate results.
    if not IsEnum( results.result, GAMEOVER ) then
        LOGWARN( "Invalid game over result: %s", tostring(results.result))
        results.result = GAMEOVER.DEATH
    end

    if results.killed_by and not is_instance( results.killed_by, Agent ) then
        LOGWARN( "Invalid killed_by parameter to game over: %s", tostring(results.killed_by))
        results.killed_by = nil
    end

    self.game_over_results = results

    TheGame:GetEvents():BroadcastEvent( "GAME_OVER", self, results.result )

    self:DoGameEndScoring()

    print( "GameOver!", tostr(results))
    print( tostr(self.options, 2))
end


function GameState:DoGameEndScoring()
    self:AddScore("LIKED", #self:GetPlayerAgent():GetAspect("social_connections"):GetConnectionsOfLevel(RELATIONSHIP.LIKED))
    self:AddScore("LOVED", #self:GetPlayerAgent():GetAspect("social_connections"):GetConnectionsOfLevel(RELATIONSHIP.LOVED))
    self:AddScore("LEFTOVER_MONEY", self:GetPlayerAgent():GetMoney())
    self:AddScore("GAINED_MONEY", self:GetPlayerAgent().money_gained or 0)
end


function GameState:GetGameOver()
    return self.game_over_results
end

function GameState:IsVictory()
    return self.game_over_results and self.game_over_results.result == GAMEOVER.VICTORY
end

function GameState:GetTracker()
    return self.tracker
end


function GameState:GetScoreCount(category)
    category = loc.toupper( category )
    return self.scorekeeper:GetCount(category)
end


function GameState:GetScoreValue(category)
    category = loc.toupper( category )
    return self.scorekeeper:GetValue(category)
end

function GameState:GetScore()
    return self.scorekeeper:GetScore()
end



function GameState:AddScore(category, count, userdata)
    count = count or 1
    category = loc.toupper( category )
    self.scorekeeper:AddScore(category, count, userdata)
end


function GameState:InitAgentTables( agents )
    -- Init run-time table structures from the complete agents list.
    self.agents_by_alias = {}

    for k, agent in pairs(agents) do
        if agent.alias then
            self.agents_by_alias[ agent.alias ] = agent
        end
    end
end


function GameState:ClearPendingEvents()
    table.clear(self.pending_events)
end

function GameState:AddPendingEvent(type_of_event)
    table.insert(self.pending_events, type_of_event or "")
end


function GameState:CanSpawnTravelEvent()

    for k, quest in pairs( self.available_quests ) do
        if quest:PreventsRandomEvents() then
            return false
        end
    end

    if self.caravan:HasEscort() then
        
        return false
    end

    return true
end


function GameState:TrySpawnNewEvent(trigger)

    if not self:CanSpawnTravelEvent() then
        return
    end

    if #self.pending_events == 0 then
        print ("OUT OF PENDING EVENTS")
        return
    end

    --don't let things stack up too much

    if self.skipped_event or math.random() < 0.5 then
        self.skipped_event =  false
        local overrides = { cast = { here = self:GetPlayerAgent():GetLocation() } }
        for idx, quest_type in ipairs(self.pending_events) do
            local q = QuestUtil.SpawnEventQuest( trigger, overrides, quest_type )
            if q then
                print ("SPAWNED EVENT", q )
                table.remove(self.pending_events, idx)
                return q
            end
        end
        print ("COULD NOT SPAWNED EVENT")
    else
        print ("DECIDED TO SKIP_EVENT")
        self.skipped_event = true
    end
    
    
end

function GameState:AdvanceTime( new_phase )
    print ("PASS TIME:", self.datetime, new_phase )
    self.datetime = self.datetime + 1
    
    if not new_phase then
        new_phase = self.day_phase == DAY_PHASE.DAY and DAY_PHASE.NIGHT or DAY_PHASE.DAY
    end

    self.day_phase = new_phase
    self:GetNotifications():SetBannersEnabled( false )
    self:UpdateWorld()
    self:GetNotifications():SetBannersEnabled( true )
    
    self:ClearAllAgentStatusEffects()

    local player_location = self:GetPlayerAgent():GetLocation()
    for k,agent in pairs(self.agents) do
        agent:OnTimePass()
        if agent:GetLocation() ~= player_location and agent:GetRoleAtLocation() ~= CHARACTER_ROLES.WAITING_PET then
            agent:MoveToLimbo()        
        end
    end

    TheGame:BroadcastEvent( "phase_change", new_phase)

end

function GameState:GetOptions()
    return self.options
end

function GameState:GetDateTime()
    return self.datetime
end

function GameState:GenerateID()
    self.next_id = self.next_id + 1
    return self.next_id
end

function GameState:CreateRNG( stream_id, seed )
    if self.rngs == nil then
        self.rngs = {}
    end
    self.rngs[ stream_id ] = engine.Random( seed )
end

function GameState:GetRNG( stream_id )
    if stream_id then
        assert_warning( "missing RNG for '%s'", tostring(stream_id))
        return self.rngs and self.rngs[ stream_id ]
    else
        return self.rng
    end
end

function GameState:SeededRandom( n1, n2 )
    -- Same API as math.random.
    return self.rng:Random( n1, n2 )
end

function GameState:AllLocations()
    return pairs( self.locations )
end

function GameState:GetLocation( id )
    if type(id) == "number" then
        return self.locations[ id ]

    elseif type(id) == "string" then
        if self.locations_by_uuid[ id ] then
            return self.locations_by_uuid[ id ]
        end

        return self:GetLocationByContentID( id )
    end
end

function GameState:GetLocationByContentID( content_id )
    for k,location in pairs(self.locations) do
        if location:GetContentID() == content_id then
            return location
        end
    end
    return nil
end

function GameState:LogNotification( notification_def, ... )
    self.notifications:LogNotification( notification_def, ... )
end

function GameState:GetNotifications()
    return self.notifications
end

function GameState:AccumulatePlayTime(dt)
    self.play_time = (self.play_time or 0) + dt
    
    local AFK_THRESH = 30
    if TheGame:GetInput():GetNoInputTime() > AFK_THRESH then
        if not self.afk then
            self.afk = true
            self.afk_time = self.afk_time + AFK_THRESH
        else
            self.afk_time = self.afk_time + dt
        end
    end

end

function GameState:GetPlayTime()
    return self.play_time, self.afk_time
end


function GameState:AddLocation( location )
    assert( is_instance(location, Location) )
    assert( location.guid == nil )

    location:PopulateGameState( self )
    assert(self.locations[location.guid] == nil, location )
    self.locations[location.guid] = location
    self.locations_by_uuid[location:GetUUID()] = location
    return location
end

function GameState:RemoveLocation( location )
    assert( is_instance(location, Location) )
    assert( self.locations[ location.guid ], location )
    assert( self.locations_by_uuid[ location:GetUUID() ], location )

    self.locations[location.guid] = nil
    self.locations_by_uuid[location:GetUUID()] = nil

    location:Depopulate()
end

function GameState:Agents()
    return pairs( self.agents )
end

function GameState:GetAgent( id )
    -- For convenience, we support lookup by guid (numeric id) or alias (string)
    if type(id) == "number" then
        return self.agents[ id ]
    else
        return self.agents_by_alias[ id ]
    end
end

function GameState:GetAgentOrMemento( id )
    return self:GetAgent( id ) or self:GetAgentMemento( id )
end

function GameState:GetAgentMemento( id )
    if type(id) == "number" then
        return self.removed_agents[ id ]
    else
        for guid, agent in pairs( self.removed_agents ) do
            if agent:HasAlias( id ) then
                return agent
            end
        end
    end
end

function GameState:GetAgentBucket( bucket_id )
    return self.agent_buckets[ bucket_id ] or table.empty
end

function GameState:AddAgentToBucket( agent, bucket_id )
    local bucket = self.agent_buckets[ bucket_id ]
    if bucket == nil then
        bucket = {}
        self.agent_buckets[ bucket_id ] = bucket
    end
    table.insert( bucket, agent )
end

function GameState:RemoveAgentFromBucket( agent, bucket_id )
    local bucket = self.agent_buckets[ bucket_id ]
    table.arrayremove( bucket, agent )
end

function GameState:OnDestroyed()
    self.tracker:DestroyTracker()
end


function GameState:GetOrCreateSpecificSkinnedAgent( skin_id )

    for k, agent in pairs(self.agents) do
        if agent:GetSkinID() == skin_id then
            return agent
        end
    end

    local skin = Content.GetCharacterSkin(skin_id)
    if skin then
        local content_id = Content.GetCharacterForSkin( skin )
        if content_id then
            local skins = self.character_skins[ content_id ]
            if skins then
                table.arrayremove( skins, skin_id )
            end
            return self:AddAgent( Agent( content_id, skin ) )
        end
    end

end

function GameState:CreateSkinnedAgent( content_id, invalid_uuids )
    if content_id == nil then
        -- Pick an appropriate character ID based on this region.
        local population = self.region and self.region:GetContent().population
        if population == nil then
            return
        end
        content_id = table.arraypick( population )
    end

    -- Remap Promoted units.
    local p = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_PROMOTION_CHANCE )
    if p > 0 then
        local def = Content.GetCharacterDef( content_id )
        if def.promotion_def and self:SeededRandom() < p then
            -- print( ">>>", content_id, " was promoted to ", def.promotion_def )
            content_id = def.promotion_def
        end
    end
    
    local skins = self.character_skins[ content_id ]
    local is_copy = false
    local skin

    -- Try picking a unique agent if one is left.
    while skin == nil and skins and #skins > 0 do
        local skin_uuid = table.arraypick( skins )
        if invalid_uuids and table.arrayfind( invalid_uuids, skin_uuid ) then
            -- Invalid pick.
            if not is_copy then
                skins = shallowcopy( skins )
                is_copy = true
            end
            table.arrayremove( skins, skin_uuid )
        else
            skin = Content.GetCharacterSkin( skin_uuid )
            if skin == nil then
                -- Invalid skin: purge, and repick.
                print( "WARNING: Invalid skin uuid in GameState; purging" )
                table.arrayremove( skins, skin_uuid )
            end
        end
    end

    if skin == nil then
        local def = Content.GetCharacterDef( content_id )
        -- Faction may not be what we want here; sentient is closer but that's a feature of the SPECIES, which is
        -- typically associated with the skin, not the CharacterDef.
        assert( def, content_id )
        if def.faction_id ~= MONSTER_FACTION then
            print( "WARNING: ran out of skins for:", content_id )
            local json_t = { PLAYER_DATA = TheGame:GetGameState():GetPlayerState(), MISSING_SKIN_DEF = content_id }
            SendMetricsData("RAN_OUT_OF_SKINS", json_t)
        end
    end

    return Agent( content_id, skin )
end



function GameState:AddSkinnedAgent( content_id, invalid_uuids )
    local agent = self:CreateSkinnedAgent( content_id, invalid_uuids )
    if agent then
        self:AddAgent( agent )
        return agent
    end
end

function GameState:IsSkinAvailable( skin_uuid )
    for content_id, skins in pairs( self.character_skins ) do
        if table.arrayfind( skins, skin_uuid ) then
            return true
        end
    end
    return false
end

function GameState:AddPlayerAgent( agent )
    assert( self.player_agent == nil )
    self.player_agent = agent
    self:AddAgent( agent )

    agent:Recruit( PARTY_MEMBER_TYPE.LEADER )

    if self.options.mutators then
        for i, graft_id in ipairs( self.options.mutators ) do
            if Content.GetGraft( graft_id ) then
                local graft = GraftInstance( graft_id )
                agent.graft_owner:AddGraft( graft )
            end
        end
    end

    if self.options.perks then
        for i, graft_id in ipairs( self.options.perks ) do
            if Content.GetGraft( graft_id ) then
                local graft = GraftInstance( graft_id )
                agent.graft_owner:AddGraft( graft )
            end
        end
    end


    if self.options.feats then
        for i, graft_id in ipairs( self.options.feats ) do
            if Content.GetGraft( graft_id ) then
                local graft = GraftInstance( graft_id )
                agent.graft_owner:AddGraft( graft )
            end
        end
    end
    
    if self:IsStoryMode() then
        agent.graft_owner:AddGraft(GraftInstance("story_mode_graft"))
    end

    --self:GetGameType() ~= GAME_TYPE.DAILY 
    if agent.mettle_handler then
        agent.mettle_handler:ApplyUpgrades()
    end

    return agent
end

function GameState:GetPlayerBackgroundID()
    return self:GetCurrentAct().background.id
end

function GameState:GetPlayerAgent()
    return self.player_agent
end

function GameState:AddAgent( agent )
    assert( is_instance(agent, Agent) )
    assert( agent.guid == nil )
    
    agent:PopulateGameState( self )

    self.agents[agent.guid] = agent

    if agent.alias then
        assert_warning( self.agents_by_alias[ agent.alias ] == nil, agent.alias )
        self.agents_by_alias[agent.alias] = agent
    end

    -- Remove this skin from availability
    local skin_uuid = agent:GetSkinID()
    if skin_uuid then
        local skins = self.character_skins[ agent:GetContentID() ]
        if skins then
            table.arrayremove( skins, skin_uuid )
        else
            print ( "NO SKINS FOR", agent:GetContentID())
        end
    end

    local faction = agent:GetFaction()
    if faction and faction.default_opinion_event_id then
        local op_dat = OPINION[faction.default_opinion_event_id]
        if op_dat then
            agent:OpinionEvent(op_dat)
        end
    end



    return agent
end

function GameState:RemoveAgent( agent )
    assert( not agent.removed, "Agent already removed" )
    assert( agent:InLimbo(), "Agent must be in limbo" )

    if agent.alias then
        self.agents_by_alias[agent.alias] = nil
    end

    if agent.guid then
        self.agents[agent.guid] = nil
        self.removed_agents[agent.guid] = agent
    end
end

function GameState:GetAgentByAlias( alias )
    local agent = self.agents_by_alias[ alias ]
    if agent then
        return agent
    else
        for guid, agent in pairs( self.removed_agents ) do
            if agent:HasAlias( alias ) then
                return agent
            end
        end
    end
end


function GameState:GetWorldRegion()
    return self.region
end

function GameState:GetWorldRegionID()
    return self.world_region_id
end

function GameState:SetHasSeen( id )
    self.seen_history[ id ] = true
end

function GameState:HasSeen( id )
    return self.seen_history[ id ]
end

function GameState:GetCaravan()
    return self.caravan
end

function GameState:GetFactions()
    return self.factions
end

function GameState:GetFaction( faction_id )
    return self.factions:GetFaction( faction_id )
end

function GameState:AddActiveQuest( quest )

    print ("GameState:AddActiveQuest", quest.content_id, quest:GetRank())
    self.available_quests[ quest.id ] = quest
    if not quest:PopulateGameState( self ) then
        self.available_quests[ quest.id ] = nil
        return false
    end
    self:RecordQuestActivated( quest:GetContentID() )
    quest:NotifyChanged()
    
end

function GameState:GetActiveQuestWithContentID( quest_id )
    local quests = {}
    for k,quest in pairs(self.available_quests) do
        if quest.content_id == quest_id then
            table.insert(quests, quest)
        end
    end
    return quests
end

function GameState:GetFinishedQuest( quest_id )
    return self.finished_quests[ quest_id ]
end

function GameState:RemoveActiveQuest( quest )
    self.available_quests[ quest.id ] = nil
    self.finished_quests[ quest.id ] = quest
end

function GameState:ActiveQuests()
    return pairs( self.available_quests )
end

function GameState:FinishedQuests()
    return pairs( self.finished_quests )
end

function GameState:GetQuestStats( quest_id, create )
    local stats = self.quest_stats[ quest_id ]
    if stats == nil and create then
        stats = {}
        self.quest_stats[ quest_id ] = stats
    end
    return stats
end

function GameState:RecordQuestActivated( quest_id )
    local stats = self:GetQuestStats( quest_id, true )
    stats.activated = (stats.activated or 0) + 1
    stats.activated_timestamps = stats.activated_timestamps or {}
    table.insert( stats.activated_timestamps, self.datetime )    
end


function GameState:GetQuestActivatedCount( quest_id )
    local stats = self:GetQuestStats( quest_id )
    return stats and stats.activated_timestamps and #stats.activated_timestamps or 0
end

function GameState:UpdateQuests()
    -- Quests may complete and trigger new quests (eg. job boards), so need to shallowcopy.
    local available_quests = ObtainWorkTable()
    for k,v in pairs(self.available_quests) do
        table.insert(available_quests, v)
    end
   
    for k,quest in pairs( available_quests ) do
        quest:UpdateObjectiveState()
    end

    ReleaseWorkTable(available_quests)
end

function GameState:UpdateAgents()
    -- This loop is multipass because agents can be generated within (see Patrol:OnStart)
    local update_count
    repeat
        update_count = 0

        for k, agent in pairs(self.agents) do
            if agent:UpdateAgent() then
                update_count = update_count + 1
            end
        end
    until update_count == 0
end


function GameState:MarkLocationForDeletion(location)
    location:SetFlags( LOCFLAGS.TRANSIENT )
end

function GameState:UpdateLocations( dt, now )   
   
    local loc_iteration_list = shallowcopy(self.locations)    --make a copy because they can be deleted
    for k,loc in pairs(loc_iteration_list) do
        loc:UpdateLocation()
    end
end

function GameState:InitializeAct()
    engine.inst:ProfilerPush("InitializeAct")

    self.current_act = self.options.start_act
    local act_data = GetPlayerActData( self.current_act )
 
    self.notifications:EnableNotifications( false )

    act_data:InitializeAct( self, self.options and self.options.custom)

    local region_id = act_data.data.world_region   
    self.world_region_id = region_id

    local world_region = Content.GetWorldRegion(region_id)
    assert(world_region, region_id)
    
    self.region = Region(region_id)
    self.region:PopulateGameState( self )

    local max_resolve = act_data.data.max_resolve or self.player_agent.max_resolve
    if max_resolve then
        local advancement_mod = GetAdvancementModifier( ADVANCEMENT_OPTION.LOWER_MAX_RESOLVE )
        if advancement_mod then
            max_resolve = math.round(max_resolve * advancement_mod)
        end
        -- This is to account for any grafts, mettle, mutators or similar things that have affected max resolve
        local cur, current_max = self.caravan:GetResolve()
        self.caravan:SetMaxResolve( max_resolve + current_max )
    end

    self:UpdateWorld()

    if not self.options.no_act_quests then
        self:GetCurrentAct():SpawnQuests( self.options )
    end
    
    self.notifications:Clear()
    self.notifications:EnableNotifications( true )

    engine.inst:ProfilerPop()
end

function GameState:OnNewGame( save_file_name )
    -- Mutator and other new game hooks, etc.
    TheGame:BroadcastEvent( "new_game", self )

    self:SetSaveFileName(save_file_name or "autosave")
end


function GameState:UpdateWorld()
    self:UpdateQuests()
    self:UpdateAgents()
    self:UpdateLocations()
end

function GameState:GetSummaryData()
    if table.count( self.finished_quests ) == 0 and (self.game_over_results == nil or self.game_over_results.result == GAMEOVER.ABANDONED) then
        -- Don't bother saving any runs where you did nothing, and abandond the game.
        return nil
    end

    local data = {
        uuid = self.uuid,
        build_id = self.build_id,
        current_act = self.current_act,
        play_time = self.play_time,
        afk_time = self.afk_time,
        datetime = self.datetime,
        creation_time = self.creation_time,

        options = self.options,
    }

    data.resolve, data.max_resolve = self.caravan:GetResolve()
    data.health, data.max_health = self.player_agent:GetHealth()

    data.battle_deck = {}
    for i, card in ipairs( self.player_agent.battler:GetCards()) do
        table.insert( data.battle_deck, card:GetSummaryData() )
    end

    data.negotiation_deck = {}
    for i, card in ipairs( self.player_agent.negotiator:GetCards()) do
        table.insert( data.negotiation_deck, card:GetSummaryData() )
    end

    data.grafts = {}
    for i, graft in ipairs( self.player_agent.graft_owner:GetAllGrafts() ) do
        data.grafts[i] = graft:GetSummaryData()
    end

    data.social_connections = {}
    for agent, relationship in self.player_agent:GetSocialConnections():Relationships() do
        table.insert( data.social_connections, 
        {
            agent = agent:GetSummaryData(),
            relationship = relationship,
        })
    end

    data.dead_agents = {}
    for k, agent in pairs( self.removed_agents or {} ) do
        if agent:IsSentient() and agent:IsDead() then 
            table.insert( data.dead_agents, agent:GetSummaryData() )
        end
    end

    data.score = self.score
    data.quest_stats = self.quest_stats
    data.negotiations_won = self:GetScoreCount("NEGOTIATIONS_WON")
    data.battles_won = self:GetScoreCount("BATTLES_WON")
    data.jobs_completed = self:GetScoreCount("JOBS_COMPLETED")
    data.money = self:GetScoreCount("GAINED_MONEY")
    data.progress = self:GetScoreCount("PROGRESS")

    if self.game_over_results then
        data.result = self.game_over_results.result
        if self.game_over_results.killed_by then
            data.killed_by = self.game_over_results.killed_by:GetSummaryData()
        end
    end

    return data
end

function GameState:GetSaveData()
    local data =
    {
        uuid = self.uuid,
        build_id = self.build_id,
        world_region_id = self.world_region_id,
        current_act = self.current_act,
        action_clock = self.action_clock,
        last_action_location = self.last_action_location,
        last_action_datetime = self.last_action_datetime,

        play_time = self.play_time,
        afk_time = self.afk_time,
        next_id = self.next_id,
        rng = { self.rng:GetSeeds() },
        caravan = self.caravan,
        factions = self.factions,
        seen_history = self.seen_history,
        available_quests = self.available_quests,
        finished_quests = self.finished_quests,
        player_agent = self.player_agent,
        agents = self.agents,
        agent_buckets = self.agent_buckets,
        locations = self.locations,
        removed_agents = self.removed_agents,
        quest_stats = self.quest_stats,
        notifications = self.notifications,
        datetime = self.datetime,
        pending_events = self.pending_events,
        region = self.region,

        main_quest = self.main_quest,
        card_stats = self.card_stats,
        dialog_memory = self.dialog_memory,
        character_skins = self.character_skins,
        current_difficulty = self.current_difficulty,
        day_phase = self.day_phase,
        act_progress = self.act_progress,
        act_progress_sub = self.act_progress_sub,
        act_progress_datetime = self.act_progress_datetime,
        rollback_thresh = self.rollback_thresh,
        card_skip_count = self.card_skip_count,
        rollback_count = self.rollback_count,
        rollback_save = self.rollback_save,

        last_battle_scenario = self.last_battle_scenario,
        scorekeeper = self.scorekeeper,
    }

    if self.rngs then
        data.rngs = {}
        for id, rng in pairs( self.rngs ) do
            data.rngs[id] = { rng:GetSeeds() }
        end
    end
    
    if self:GetGameType() == GAME_TYPE.DAILY then
        -- Dailys have a specific structure revolving around a single conversation.
        local encounter = self.caravan:GetCurrentEncounter()
        if encounter then
            local convo_id, state_id = encounter:GetConvoStateID()
            data.encounter_state =
            {
                convo_id = convo_id,
                state_id = state_id,
                scratch = shallowcopy( encounter.scratch ),
            }
        end
    end

	return data
end

function GameState:GetEncounterState()
    return self.encounter_state
end

function GameState:GetCurrentActID()
    return self.current_act
end

function GameState:GetCurrentAct()
    return self.current_act and GetPlayerActData(self.current_act)
end

function GameState:GetCurrentBackground()
    return self.current_act and GetPlayerActData(self.current_act).background
end

function GameState:SetDifficulty(diff)
    self.current_difficulty = diff
    return self
end

function GameState:Action()
    local player_loc = self:GetPlayerAgent():GetLocation()
    local datetime = Now()
    
    if player_loc ~= self.last_action_location or datetime ~= self.last_action_datetime then
        self.action_clock = (self.action_clock or 0) + 1
        self.last_action_datetime = datetime
        self.last_action_location = player_loc

        TheGame:BroadcastEvent( "action_clock_advance" )
        print ("TIME ADVANCE", self.action_clock)
    end
end

function GameState:GetActionClock()
    return self.action_clock
end

function GameState:GetCurrentBaseDifficulty()
    return self.current_difficulty
end

function GameState:GetDefaultQuestRank()
    return TheGame:GetGameState():GetCurrentBaseDifficulty() 
end


function GameState:DoPlayerRest()
    self:FlushSideJobs()
    self:ClearAllAgentStatusEffects()
    self:ClearPendingEvents()

    for k, agent in pairs(self.agents) do
        agent:OnPlayerRest()
    end
end




function GameState:SideJobsPresent()
    for k,quest in self:ActiveQuests() do
        if quest:GetType() == QTYPE.SIDE then
            return true
        end
    end
    return false
end

function GameState:FlushSideJobs()
    local to_cancel = {}
    for k,quest in self:ActiveQuests() do
        if (quest:GetType() == QTYPE.SIDE or quest:GetType() == QTYPE.WANDERING_SHOP) and quest:GetQuestDef().can_flush ~= false then
            table.insert(to_cancel, quest)
        end
    end
    for k,quest in pairs(to_cancel) do
        print( "Flushing:", quest )
        if quest:IsActive() then
            quest:Fail()
        else
            quest:Cancel()
        end
    end
end

function GameState:ClearAllAgentStatusEffects()
    for k,agent in pairs(self.agents) do
        agent:ClearStatusEffects()
    end
end


function GameState:SetDecks(deck_data)
    
    if self.player_agent and self.player_agent:GetContentID() == deck_data.character then
        self:GetNotifications():SetBannersEnabled( false )
        self.player_agent.aspects.battler:SetCards(deck_data.battle_cards)
        self.player_agent.aspects.negotiator:SetCards(deck_data.negotiation_cards)
        self.player_agent.aspects.graft_owner:SetGrafts(deck_data.grafts)
        self:GetNotifications():SetBannersEnabled( true )
        return true
    end
    
end

function GameState:GetDeckState()
    local player = self.player_agent
    if player then
        local ret = {character = player:GetContentID(), battle_cards = {}, negotiation_cards = {}, grafts = {}}
        
        for k,v in ipairs( player.aspects.battler.cards.cards) do
            table.insert(ret.battle_cards, v.id)
        end

        for k,v in ipairs( player.aspects.negotiator.cards.cards) do
            table.insert(ret.negotiation_cards, v.id)
        end

        for _, graft in ipairs(player.aspects.graft_owner.grafts) do
            table.insert(ret.grafts, graft.id)
        end

        return ret
    end
end

function GameState:GetDeckStateString()
    local ret = self:GetDeckState()
    if ret == nil then
        return ""
    else
        return "return ".. serpent.block(ret, {comment = false})
    end
end

function GameState:GetPlayerStateString()
    local t= {}
    local percent, title, subtitle = self:GetNarrativeProgress()
    table.insert(t, loc.format( "User Name: {1}", engine.inst:GetUserName() or "???"))

    local data= self:GetPlayerState()
    if data then
        table.insert(t, loc.format( "{1} {2} {3#percent}", data.CHARACTER_ID, title, percent or 0))
        table.insert(t, loc.format( "Health:{1}/{2} Resolve:{3}/{4}", data.HEALTH, data.HEALTH_MAX, data.RESOLVE, data.MAX_RESOLVE))
    end
    table.insert(t, loc.format( "Prestige Level: {1}", self.options.advancement_level or 0 ))

    if self:GetCurrentAct() ~= nil then
        table.insert(t, loc.format( "Toggled-on Unlocks: {1}", tostr( TheGame:GetGameProfile():GetToggledOnUnlocks(self:GetPlayerBackgroundID(),self) ) ))
        table.insert(t, loc.format( "Toggled-off Unlocks: {1}", tostr( self.options.toggled_off_unlocks or {} ) ))
    end

    table.insert(t, loc.format( "Battles won: {1}", self:GetScoreCount("BATTLES_WON") ))
    table.insert(t, loc.format( "Negotiations won: {1}", self:GetScoreCount("NEGOTIATIONS_WON") ))
    table.insert(t, loc.format( "Negotiations lost: {1}", self:GetScoreCount("NEGOTIATIONS_LOST") ))


    --table.insert(t, loc.format( "Save UUID: {1}", self.uuid))
    table.insert(t, loc.format( "Playtime: {1} min", math.floor(self.play_time/60)))
    table.insert(t, loc.format( "AFK time: {1} min", math.floor((self.afk_time or 0)/60)))

    if GetIsFastMode() then
        table.insert(t, "FAST MODE IS ON")
    end

    if data then    
        local rarity_count = {}
        local bcards = {}
        for k,v in ipairs(data.BATTLE_CARDS) do
            
            local def = Content.GetBattleCard(v.ID)
            rarity_count[def.rarity] = rarity_count[def.rarity] and rarity_count[def.rarity] + 1 or 1
            if v.XP and v.XP > 0  then
                table.insert(bcards, loc.format("\t{2} '{1}' xp:{3}", def and def.name or "", v.ID, v.XP))
            else
                table.insert(bcards, loc.format("\t{2} '{1}'", def and def.name or "", v.ID))
            end
        end


        table.sort(bcards)
        table.insert(t, "\nBATTLE CARDS:")
        
        table.insert(t, loc.format("basic:{2}, common:{3} uncommon:{4} rare:{5} unique:{6} boss:{7} TOTAL:{8}",
            0,
            rarity_count[CARD_RARITY.BASIC] or 0,
            rarity_count[CARD_RARITY.COMMON] or 0,
            rarity_count[CARD_RARITY.UNCOMMON] or 0,
            rarity_count[CARD_RARITY.RARE] or 0,
            rarity_count[CARD_RARITY.UNIQUE] or 0,
            rarity_count[CARD_RARITY.BOSS] or 0,
            #data.BATTLE_CARDS))

        table.insert(t, table.concat(bcards, "\n"))


        table.clear(rarity_count)
        local ncards = {}
        for k,v in ipairs(data.NEGOTIATION_CARDS) do
            local def = Content.GetNegotiationCard(v.ID)
            rarity_count[def.rarity] = rarity_count[def.rarity] and rarity_count[def.rarity] + 1 or 1
            if v.XP and v.XP > 0  then
                table.insert(ncards, loc.format("\t{2} '{1}' xp:{3}", def and def.name or "", v.ID, v.XP))
            else
                table.insert(ncards, loc.format("\t{2} '{1}'", def and def.name or "", v.ID))
            end
        end
        table.sort(ncards)
        table.insert(t, "\nNEGOTIATION CARDS:")

        table.insert(t, loc.format("basic:{2}, common:{3} uncommon:{4} rare:{5} unique:{6} boss:{7} TOTAL:{8}",
            0,
            rarity_count[CARD_RARITY.BASIC] or 0,
            rarity_count[CARD_RARITY.COMMON] or 0,
            rarity_count[CARD_RARITY.UNCOMMON] or 0,
            rarity_count[CARD_RARITY.RARE] or 0,
            rarity_count[CARD_RARITY.UNIQUE] or 0,
            rarity_count[CARD_RARITY.BOSS] or 0,
            #data.NEGOTIATION_CARDS))

        table.insert(t, table.concat(ncards, "\n"))


        table.insert(t, loc.format("\nBATTLE GRAFTS:"))

        for k,v in ipairs(data.GRAFTS) do
            local def = Content.GetGraft(v.ID)
            if def.type == GRAFT_TYPE.COMBAT then
                table.insert(t, loc.format("\t{1}",v.ID))
            end
        end

        table.insert(t, loc.format("\nNEGOTIATION GRAFTS:"))
        for k,v in ipairs(data.GRAFTS) do
            local def = Content.GetGraft(v.ID)
            if def.type == GRAFT_TYPE.NEGOTIATION then
                table.insert(t, loc.format("\t{1}",v.ID))
            end
        end

        table.insert(t, loc.format("\nSOCIAL GRAFTS:"))
        for k,v in ipairs(data.GRAFTS) do
            local def = Content.GetGraft(v.ID)
            if def.type == GRAFT_TYPE.SOCIAL then
                table.insert(t, loc.format("\t{1}",v.ID))
            end
        end

        table.insert(t, loc.format("\nMUTATORS:"))
        for k,v in ipairs(data.GRAFTS) do
            local def = Content.GetGraft(v.ID)
            if def.type == GRAFT_TYPE.MUTATOR then
                table.insert(t, loc.format("\t{1}",v.ID))
            end
        end
    end

    if self.last_battle_scenario then
        table.insert(t, "\nMOST RECENT BATTLE:\n")
        table.insert(t, self.last_battle_scenario:GetStringDump())
    end


    return table.concat(t, "\n")
end




function SummarizeCards(cards)
	local result = {}
	for _, card in ipairs(cards) do
		table.insert(result, { ID = card.id, XP = card.userdata.xp })
	end
	return result
end

function GameState:GetPlayerState()
	if not self.player_agent or not self:GetCurrentAct() then
		return nil
	end
    
    local percent, title, subtitle = self:GetNarrativeProgress()

	local player = self.player_agent
    local resolve, max_resolve = self:GetCaravan():GetResolve()

	local data = {
		CHARACTER_ID = player.def.id,
		WORLD_REGION_ID = player.world_region_id,
		LOCATION_ID = player.location and player.location.content_id,
		HEALTH = math.min( 9999, player.aspects.health.current ),
		HEALTH_MAX = math.min( 9999, player.aspects.health.max ),

		RESOLVE = math.min( 9999, resolve ),
		MAX_RESOLVE = math.min( 9999, max_resolve ),
	}

	local inventory_data = {
		MONEY = player:GetMoney(),
	}

    data["SAVE_UUID"] = self.uuid

    data["PLAYTIME"] = self.play_time
    data["AFKTIME"] = self.afk_time or 0
	data["INVENTORY"] = inventory_data
    --data["MONEY"] = player:GetMoney()
    data["BATTLE_CARDS"] = SummarizeCards(player.aspects.battler.cards.cards)
    data["NEGOTIATION_CARDS"] = SummarizeCards(player.aspects.negotiator.cards.cards)
	data["STORY_ACT"] = self.current_act
    data["PROGRESS_PERCENT"] = percent or 0
    data["ADVANCEMENT_LEVEL"] = self.options.advancement_level or 0
    
    data["TOGGLED_ON_UNLOCKS"] = TheGame:GetGameProfile():GetToggledOnUnlocks(self:GetPlayerBackgroundID(),self) or {}
    data["TOGGLED_OFF_UNLOCKS"] = self.options.toggled_off_unlocks or {}

	data["PROGRESS_MAJOR"] = self.act_progress
	--data["PROGRESS_MINOR"] = self.act_progress_sub
    data["METTLE"] = TheGame:GetGameProfile():GetMettlePoints()

	local graft_data = {}
	for _, graft in ipairs(player.aspects.graft_owner.grafts) do
		table.insert(graft_data, { ID = graft.id })
	end
	data["GRAFTS"] = graft_data

	return data
end

function GameState:GetDayPhase()
    return self.day_phase
end

function GameState:SetDayPhase(phase)
    self.day_phase = phase
end


function GameState:GetNarrativeProgress()
    if self.main_quest then
        return self.main_quest:GetNarrativeProgress()
    end
end

function GameState:GetQuestLocationForAgent(agent)
    for guid, quest in pairs( self.available_quests ) do
        local location, role = quest:GetQuestLocationForAgent(agent)
        if location then
            return location, role
        end
    end
end

function GameState:GetQuestAgentsForLocation(location, t)
    for guid, quest in pairs( self.available_quests ) do
        quest:GetQuestAgentsForLocation(location, t)
    end
end


function GameState:IsHiringAvailable()
    for guid, quest in pairs( self.available_quests ) do
        if quest:IsReadyForTurnIn() or (quest.is_hiring_available and quest.is_hiring_available() == false) then
            return false
        end
    end
    return true
end

function GameState:GetCardStats()
    return self.card_stats
end

function GameState:RememberDialog(id)
    self.dialog_memory[id] = 1 + (self.dialog_memory[id] or 0)
end

function GameState:HasDialogMemory(id)
    return self.dialog_memory[id] and self.dialog_memory[id] > 0 
end

function GameState:IncreaseCardSkipReward()
    self.card_skip_count = (self.card_skip_count or 0) + 1
end

function GameState:GetCardSkipReward( )
    local card_skip_details = {money = SKIP_CARD_REWARD}
    TheGame:BroadcastEvent( "calc_card_skip_reward", card_skip_details )
    return card_skip_details["money"] 
end

function GameState:SetActProgress(num)
    self.act_progress = num or 0
    self.act_progress_sub  = 0
    self.act_progress_datetime = self.datetime
end

function GameState:GetActProgress()
    return self.act_progress, self.act_progress_sub, self.act_progress_datetime
end

--if there are people in impossible situatiuons, send them to limbo and let them re-enter the world.
function GameState:EnforceAvoids()
    local player_loc = self:GetPlayerAgent():GetLocation()
    local to_limbo = ObtainWorkTable()
    for k,agent in pairs(self.agents) do
        if agent:GetLocation() and agent:GetLocation() ~= player_loc then
            if AgentUtil.IsAgentBarredFromLocation(agent, agent:GetLocation()) then
                table.insert(to_limbo, agent)
            end
        end
    end
    for k,v in pairs(to_limbo) do
        v:MoveToLimbo()
    end

    ReleaseWorkTable(to_limbo)
end

function GameState:PutImportantAgentsInWorld()
    for k,agent in pairs(self.agents) do
        if agent:InLimbo() and agent:IsCurrentlyImportant() then
            AgentUtil.PutAgentInWorld( agent )
        elseif not agent:IsInPlayerParty() then
            --make sure agents who are supposed to be in a specific location byut
            local location, role = TheGame:GetGameState():GetQuestLocationForAgent(agent)
            if location and (location ~= agent:GetLocation() or role ~= agent:GetRoleAtLocation()) then
                agent:ClearParty()
                agent:MoveToLocation(location)
                agent:SetLocationRole(role)
            end
        end
    end
end

function GameState:PopulateAgentsInWorld()
    if self.populating then
        return
    end
    
    self.populating = true
    self:EnforceAvoids()
    self:PutImportantAgentsInWorld()

    self.populating = false
end

function GameState:GetAdvancementModifier( option )
    assert( IsEnum( option, ADVANCEMENT_OPTION ))
    if self:IsStoryMode() then
        local modifier = STORY_ADVANCEMENT_OPTION[ option ]
        if modifier then
            return modifier
        end
    elseif self.options and self.options.advancement_level then
        local act = GetPlayerActData(self.options.start_act)
        local level = math.min(self.options.advancement_level, #act.background.data.advancement)
        for i = level, 1, -1 do
            local modifier = act.background.data.advancement[ i ][ option ]
            if modifier then
                return modifier
            end
        end
    end

    return DEFAULT_ADVANCEMENT_OPTION[ option ]
end

function GameState:GetAdvancementLevel()
    if self.options then
        return self.options.advancement_level and self.options.advancement_level or 0
    end

    return 0
end

function GameState:SetAdvancementLevel( level )
    self.options.advancement_level = level
end

function GameState:IsStoryMode()
    return self.options and self.options.story_mode or false
end

function GameState:IsCustomMode()
    if self.options.mutators and #self.options.mutators > 0 then
        return true
    end

    if self.options.toggled_off_unlocks and #self.options.toggled_off_unlocks > 0 then
        return true
    end

    return false
end

function GameState:SetMainQuest(quest)
    self.main_quest = quest
end

function GameState:GetMainQuest()
    return self.main_quest
end

function GameState:ClearOpportunities(keep_quests)
    local to_clear = {}
    for _,quest in pairs(self.available_quests) do
        if quest:GetType() == QTYPE.OPPORTUNITY then
            if keep_quests == nil or not table.arrayfind(keep_quests, quest) then
                table.insert(to_clear, quest)
            end
        end
    end

    for k,v in pairs(to_clear) do
        v:Fail()
    end

    return to_clear
end

function GameState:CancelInactiveSideJobs()
    local to_clear = {}
    for _,quest in pairs(self.available_quests) do
        if quest:GetType() == QTYPE.SIDE and quest:IsInactive() then
            table.insert(to_clear, quest)
        end
    end

    for k,v in pairs(to_clear) do
        v:Cancel()
    end
end

function GameState:GetDecks()
    local player = self:GetPlayerAgent()
    
    local battle_card_ids, negotiation_card_ids = {}, {}
    local battle_graft_ids, negotiation_graft_ids = {}, {}

    for k,v in pairs(player.battler.cards.cards) do
        table.insert(battle_card_ids, v.id)
    end

    for k,v in pairs(player.negotiator.cards.cards) do
        table.insert(negotiation_card_ids, v.id)
    end

    for k,graft in pairs(player.graft_owner.grafts) do
        if graft:GetType() == GRAFT_TYPE.COMBAT then
            table.insert(battle_graft_ids, graft.id)
        elseif graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            table.insert(negotiation_graft_ids, graft.id)
        end
    end


    local last_decks = {negotiation_cards = negotiation_card_ids, 
                        battle_cards = battle_card_ids, 
                        negotiation_grafts = negotiation_graft_ids,
                        battle_grafts = battle_graft_ids
                    }
    return last_decks              
end

function GameState:GetGraftDraftDetails()
    local graft_draft_details = { count = 3 }
    TheGame:BroadcastEvent( "calc_graft_draft", graft_draft_details )
    return graft_draft_details
end



function GameState:ShouldDoDraftBounties()
    if not TheGame:GetLocalSettings().ROBOTICS then
        return self:GetGameType() ~= GAME_TYPE.DAILY
    end

    return false
end

function GameState:CanDoMettle()
    return self:GetPlayerAgent() and 
            not GetAdvancementModifier( ADVANCEMENT_OPTION.METTLE_OFF ) and
            TheGame:GetGameProfile():HasMettleUnlocked(self:GetPlayerAgent():GetContentID()) and
            not TheGame:GetGameState():GetPlayerAgent().graft_owner:HasGraft("bare_mettle")
end
