require "collection"
local battle_defs = require "battle/battle_defs"

local GraftCollection = class( "GraftCollection", Collection )

function GraftCollection:init( fn, ...)
    GraftCollection._base.init(self)
    self:AddItems(Content.GetAllGrafts(), fn, ...)
end

function GraftCollection:NotInstalled(owner)
    return self:Filter(function(graft_def) return not owner.graft_owner:HasGraft(graft_def.id) end)
end

function GraftCollection:NotUnique()
    return self:Filter(function(graft_def) return graft_def.rarity ~= CARD_RARITY.UNIQUE end)        
end

function GraftCollection:NotBoss()
    return self:Filter(function(graft_def) return graft_def.rarity ~= CARD_RARITY.BOSS end)        
end

function GraftCollection:NotRestricted(owner)
    return self:Filter(function(graft_def) return AgentUtil.CanUseGraft(owner, graft_def) end)        
end

function GraftCollection:NotUpgraded()
    return self:Filter(function(graft_def) return graft_def.flags == nil or not CheckBits(graft_def.flags, battle_defs.CARD_FLAGS.UPGRADED) end)        
end

function GraftCollection:NotLocked()
	return self:Filter( function(graft_def)
	 	if Content.IsUnlock( graft_def.id ) and not TheGame:GetGameProfile():HasUnlock( graft_def.id ) then
			return false
		end
		return true
	end )
end

function GraftCollection:Rarity(rarity)
    return self:Filter(function(graft_def) return rarity == nil or graft_def.rarity == rarity end)        
end

function GraftCollection:Type(type)
    return self:Filter(function(graft_def) return type == nil or graft_def.type == type end)        
end

function GraftCollection:HasTag(tag)
    return self:Filter(function(graft_def) return graft_def.tags and table.arrayfind(graft_def.tags, tag) end)        
end

function GraftCollection:Generate(num, rng)
    if rng then
        return table.map( self:SeededPick(num, rng), function(graft_def) return GraftInstance(graft_def.id) end)
    else
        return table.map( self:Pick(num), function(graft_def) return GraftInstance(graft_def.id) end)
    end
end

function GraftCollection:GenerateIDs()
    return table.map( self.items, function(graft_def) return graft_def.id end )
end

function GraftCollection.Rewardable( owner, fn )
    local function Filter( graft_def )
        return graft_def.type == GRAFT_TYPE.COMBAT or graft_def.type == GRAFT_TYPE.NEGOTIATION
    end

    local collection = GraftCollection(fn):NotUnique():NotLocked():NotBoss():NotUpgraded():Filter( Filter )
    if owner then
        collection:NotInstalled(owner)
        collection:NotRestricted(owner)
    end
    return collection
end

function GraftCollection.RewardableCoin( owner, fn )
    local collection = GraftCollection(fn):NotUnique():NotLocked():NotBoss():Filter(function(graft_def) return graft_def.type == GRAFT_TYPE.COIN end)
    if owner then
        collection:NotInstalled(owner)
        collection:NotRestricted(owner)
    end
    return collection
end

function GraftCollection.BossGrafts( owner, fn )
    local function Filter( graft_def )
        return graft_def.type == GRAFT_TYPE.COMBAT or graft_def.type == GRAFT_TYPE.NEGOTIATION
    end

    local collection = GraftCollection(fn):NotUnique():NotLocked():NotUpgraded():Filter( Filter )
    if owner then
        collection:NotInstalled(owner)
        collection:NotRestricted(owner)
    end
    return collection
end

function GraftCollection.GraftPool( owner, fn, rng )

    if rng then
        push_random( function( n, m )
            return rng:Random( n, m )
        end )
    end

    local rarity_pools = {}
    for i, rarity in ipairs( CARD_RARITY_ORDER ) do
        local collection = GraftCollection():NotUnique():NotBoss():NotUpgraded():Rarity( rarity ):Filter( fn )
        if owner then
            collection:NotInstalled(owner)
            collection:NotRestricted(owner)
        end
        rarity_pools[ rarity ] = collection:GenerateIDs()

        table.shuffle( rarity_pools[ rarity ] )
    end

    if rng then
        pop_random()
    end

    return rarity_pools
end
