require "util/class"
require "input/rumble"

local ControlSet = class("ControlSet")

function ControlSet:init()
    self.controls = {}
end

function ControlSet:Clear()
    table.clear( self.controls )
end

function ControlSet:IsEmpty()
    return #self.controls == 0
end

function ControlSet:Controls()
    return ipairs( self.controls )
end

function ControlSet:AddControl( control )
    table.insert( self.controls, control )
end

function ControlSet:Has( ... )
    for i = 1, select( "#", ... ) do
        if table.arrayfind( self.controls, select( i, ... )) then
            return true
        end
    end
    return false
end

function ControlSet:Dump()
    local txt = ""
    for i, ctrl in ipairs( self.controls ) do
        txt = txt .. tostr(ctrl) .. " | "
    end
    return txt
end

------------------------------------------------------------------------
--

local Input = class("Input")
Controls = require "input/controls" 

function Input:init(game, input_settings)
    --self.mouse_atlas = engine.asset.Atlas('atlases/icons_mouse.xml')
    --self.keyboard_atlas = engine.asset.Atlas('atlases/icons_keyboard.xml')
    
    self.ignore_gamepad = engine.inst:HasCommandLineOption("ignore_gamepad")

    if SPECIAL_BUILD_TYPE < 1024 then
        self.mouse_images = 
        {
            LEFT = engine.asset.Texture("icons_mouse/lmb.tex"),
            RIGHT = engine.asset.Texture("icons_mouse/rmb.tex"),
            MIDDLE = engine.asset.Texture("icons_mouse/mmb.tex"),
            SCROLLUP = engine.asset.Texture("icons_mouse/scroll_up.tex"),
            SCROLLDOWN = engine.asset.Texture("icons_mouse/scroll_down.tex"),
        }
        self.key_images =
        {
            ALT = engine.asset.Texture("icons_keyboard/key_alt.tex"),
            CTRL = engine.asset.Texture("icons_keyboard/key_ctrl.tex"),
            SHIFT = engine.asset.Texture("icons_keyboard/key_shift.tex"),
        }
    end

    self.game = game
    self.input_settings = input_settings

    self.playing_rumbles = {}

    self.keys = {}
    self.gamepads = {}

    self.control_set = ControlSet() -- Just exists so we don't have to recreate the table for every input

    self.control_state = {}
    self:ResetControlState()

    self:ApplyInputBindings()
    self:SetLeftJoystickDigitalForwarding( true )
    self:SetRightJoystickDigitalForwarding( true )

    self.mouse_x, self.mouse_y = 0, 0
end

function Input:ApplyInputBindings()
    local default_binds = require "input/bindings"
    self:SetKeyBinds(default_binds.keyboard)
    self:SetMouseBinds(default_binds.mouse)
    self:SetGamepadBinds(default_binds.gamepad)

    if self.input_settings then
        self.input_settings:ApplyInputBindings( self )
    end
end

function Input:ResetControlState()
    table.clear(self.control_state)

    for k,v in pairs(Controls.Digital) do
        self.control_state[v] = {down = false, t = 0, rep_time = 0}
    end
end

function Input:GetMousePos()
    return self.mouse_x, self.mouse_y
end

-- Used in touch mode when a finger is lifted from the screen and we don't want
-- the mouse position to stay the same, giving focus to the element under it until the next tap
function Input:ResetMousePos()
    self.mouse_x = -1000000
    self.mouse_y = -1000000
    return self.mouse_x, self.mouse_y
end

function Input:IsMousePosReset()
    return self.mouse_x == -1000000 or self.mouse_y == -1000000
end

function Input:IsControlDown(control)
    assert(control, "Invalid control")
    return self.control_state[control] and self.control_state[control].down
end

function Input:OnScreenResize(w,h)
    self.w = w
    self.h = h
end


function Input:Update(dt)
    
    
    engine.inst:ProfilerPush("Input:Update")
    
    self.no_input_time = (self.no_input_time or 0) + dt

    --[[for k,v in pairs(self.control_state) do
        
        if v.down and k.repeat_rate then
            v.rep_time = v.rep_time - dt
            if v.rep_time <= 0 then
                v.rep_time = 1/k.repeat_rate
                self:DoControlRepeat(k)
            end
        end
    end--]]
    self:UpdateRumble(dt)

    self:UpdateDebugControls()
    
    engine.inst:ProfilerPop()
end

function Input:UpdateDebugControls()
    self.debug_controls = self.debug_controls or {}
    --[[    
    ImGuiNavInput_Activate,      // activate / open / toggle / tweak value       // e.g. Circle (PS4), A (Xbox), B (Switch), Space (Keyboard)
    ImGuiNavInput_Cancel,        // cancel / close / exit                        // e.g. Cross  (PS4), B (Xbox), A (Switch), Escape (Keyboard)
    ImGuiNavInput_Input,         // text input / on-screen keyboard              // e.g. Triang.(PS4), Y (Xbox), X (Switch), Return (Keyboard)
    ImGuiNavInput_Menu,          // tap: toggle menu / hold: focus, move, resize // e.g. Square (PS4), X (Xbox), Y (Switch), Alt (Keyboard)
    ImGuiNavInput_DpadLeft,      // move / tweak / resize window (w/ PadMenu)    // e.g. D-pad Left/Right/Up/Down (Gamepads), Arrow keys (Keyboard)
    ImGuiNavInput_DpadRight,     // 
    ImGuiNavInput_DpadUp,        // 
    ImGuiNavInput_DpadDown,      // 
    ImGuiNavInput_LStickLeft,    // scroll / move window (w/ PadMenu)            // e.g. Left Analog Stick Left/Right/Up/Down
    ImGuiNavInput_LStickRight,   // 
    ImGuiNavInput_LStickUp,      // 
    ImGuiNavInput_LStickDown,    // 
    ImGuiNavInput_FocusPrev,     // next window (w/ PadMenu)                     // e.g. L1 or L2 (PS4), LB or LT (Xbox), L or ZL (Switch)
    ImGuiNavInput_FocusNext,     // prev window (w/ PadMenu)                     // e.g. R1 or R2 (PS4), RB or RT (Xbox), R or ZL (Switch) 
    ImGuiNavInput_TweakSlow,     // slower tweaks                                // e.g. L1 or L2 (PS4), LB or LT (Xbox), L or ZL (Switch)
    ImGuiNavInput_TweakFast,     // faster tweaks                                // e.g. R1 or R2 (PS4), RB or RT (Xbox), R or ZL (Switch)
--]]
    if TheGame:IsDebugUsingGamepad() then
        self.debug_controls[1] = self:IsControlDown(Controls.Digital.GAMEPAD_A) and 1.0 or 0.0
        self.debug_controls[2] = self:IsControlDown(Controls.Digital.GAMEPAD_B) and 1.0 or 0.0
        self.debug_controls[3] = self:IsControlDown(Controls.Digital.GAMEPAD_Y) and 1.0 or 0.0
        self.debug_controls[4] = self:IsControlDown(Controls.Digital.GAMEPAD_X) and 1.0 or 0.0

        self.debug_controls[5] = self:IsControlDown(Controls.Digital.GAMEPAD_DPAD_LEFT) and 1.0 or 0.0
        self.debug_controls[6] = self:IsControlDown(Controls.Digital.GAMEPAD_DPAD_RIGHT) and 1.0 or 0.0
        self.debug_controls[7] = self:IsControlDown(Controls.Digital.GAMEPAD_DPAD_UP) and 1.0 or 0.0
        self.debug_controls[8] = self:IsControlDown(Controls.Digital.GAMEPAD_DPAD_DOWN) and 1.0 or 0.0

        self.debug_controls[9] = self:IsControlDown(Controls.Digital.GAMEPAD_LS_LEFT) and 1.0 or 0.0
        self.debug_controls[10] = self:IsControlDown(Controls.Digital.GAMEPAD_LS_RIGHT) and 1.0 or 0.0
        self.debug_controls[11] = self:IsControlDown(Controls.Digital.GAMEPAD_LS_UP) and 1.0 or 0.0
        self.debug_controls[12] = self:IsControlDown(Controls.Digital.GAMEPAD_LS_DOWN) and 1.0 or 0.0

        self.debug_controls[13] = self:IsControlDown(Controls.Digital.GAMEPAD_LB) and 1.0 or 0.0
        self.debug_controls[14] = self:IsControlDown(Controls.Digital.GAMEPAD_RB) and 1.0 or 0.0

        self.debug_controls[15] = self:IsControlDown(Controls.Digital.GAMEPAD_LT) and 1.0 or 0.0
        self.debug_controls[16] = self:IsControlDown(Controls.Digital.GAMEPAD_RT) and 1.0 or 0.0
    else
        table.clear( self.debug_controls )
    end
    engine.inst:SetDebugControllerState(self.debug_controls)
end

function Input:GetNoInputTime()
    return self.no_input_time or 0
end

function Input:GetDefaultKeyBindings( control )
    local bindings = {}
    local default_binds = require "input/bindings"
    for i, default_binding in ipairs(default_binds.keyboard) do
        if default_binding.control == control then
            table.insert( bindings, default_binding )
        end
    end
    return bindings
end

function Input:RemoveKeyBind( control )
    local bindings = self.key_lookup[ control ]
    if bindings then
        for j = #bindings, 1, -1 do
            local binding = bindings[j]
            -- print( "Removing binding ", serpent.line(binding))
            assert( table.arrayremove( self.keybindings[ binding.key ], binding ))
            table.remove( bindings, j )
        end
    end
end

function Input:RebindKey( binding )
    self:RemoveKeyBind( binding.control )
    self:AddKeyBind( binding )
end

function Input:RebindGamepad( binding )
    self:RemoveGamepadBind( binding.control )
    self:AddGamepadBind( binding )
end

function Input:RemoveGamepadBind( control )
    local bindings = self.gamepad_lookup[ control ]
    if bindings then
        for j = #bindings, 1, -1 do
            local binding = bindings[j]
            print( "Removing binding ", serpent.line(binding))
            assert( table.arrayremove( self.gamepadbindings[ binding.key ], binding ))
            table.remove( bindings, j )
        end
    end
end

function Input:AddKeyBind(binding)
    local bindings = self.keybindings[binding.key]
    if bindings == nil then
        bindings = {}
        self.keybindings[binding.key] = bindings
    end
    table.insert(bindings, binding)
    if #bindings > 1 then
        table.sort(bindings, function(a,b) return (a.control.priority or 0) > (b.control.priority or 0) end)
    end

    self.key_lookup[binding.control] = self.key_lookup[binding.control] or {}
    table.insert(self.key_lookup[binding.control], binding)
end

function Input:SetKeyBinds(bindings)
    self.keybindings = {}
    self.key_lookup = {} -- Maps of Control -> list of bindings   
   
    for k,binding in pairs(bindings) do
        self:AddKeyBind(binding)
    end
end

function Input:SetMouseBinds(bindings)
    self.mousebindings = {}
    self.mouse_lookup = {}

    for k,v in pairs(bindings) do
        if v.control then
            self.mousebindings[v.button] = self.mousebindings[v.button] or {}
            table.insert(self.mousebindings[v.button], v)

            self.mouse_lookup[v.control] = self.mouse_lookup[v.control] or {}
            table.insert(self.mouse_lookup[v.control], v)
        end
    end

    for k, v in pairs(self.mousebindings) do
        table.sort(v, function(a,b) return (a.control.priority or 0) > (b.control.priority or 0) end)
    end
end

function Input:AddGamepadBind(binding)
    local bindings = self.gamepadbindings[binding.button]
    if bindings == nil then
        bindings = {}
        self.gamepadbindings[binding.button] = bindings
    end
    table.insert(bindings, binding)
    if #bindings > 1 then
        table.sort(bindings, function(a,b) return (a.control.priority or 0) > (b.control.priority or 0) end)
    end

    self.gamepad_lookup[binding.control] = self.gamepad_lookup[binding.control] or {}
    table.insert(self.gamepad_lookup[binding.control], binding)
end

function Input:SetGamepadBinds(bindings)
    self.gamepadbindings = {}
    self.gamepad_lookup = {}

    for k,binding in pairs(bindings) do
        self:AddGamepadBind(binding)
    end
end


function Input:OnTextInput(text)
    self.no_input_time = 0
    self.game:OnTextInput(text)
end

function Input:OnTextEdit(text)
    self.no_input_time = 0
    self.game:OnTextEdit(text)
end


function Input:IsKeyDown(key)
    local keyid = InputConstants.Keys[ key ]
    if keyid then
        return self.keys[ keyid ] == true
    else
        return false
    end
end

function Input:IsModifierControl()
    return self:IsKeyDown("LCTRL") or self:IsKeyDown("RCTRL")
end

function Input:IsModifierShift()
    return self:IsKeyDown("LSHIFT") or self:IsKeyDown("RSHIFT")
end

function Input:IsModifierAlt()
    return self:IsKeyDown("LALT") or self:IsKeyDown("RALT")
end

function Input:IsModifierAny()
    return self:IsModifierControl() or self:IsModifierShift() or self:IsModifierAlt()
end


function Input:IsModifierKey(key)
    local is = key == "LCTRL" or key == "RCTRL" or
            key == "LALT" or key == "RALT" or
            key == "LSHIFT" or key == "RSHIFT"
    --print("IsModifierKey()", key, is)
    return is
end

function Input:CheckModifiers( binding )
    if binding.ANYMOD then
        return true
    end

    local ctrl = self:IsModifierControl()
    local alt = self:IsModifierAlt()
    local shift = self:IsModifierShift()

    local match = ((binding.CTRL ~= nil) == ctrl) and
            ((binding.SHIFT ~= nil) == shift) and
            ((binding.ALT ~= nil) == alt)
    return match
end

function Input:OnKeyDown(keyid, modifiers)
    self.no_input_time = 0
    self.keys[keyid] = true

    local key = InputConstants.KeyById[keyid]
    if not key then print( "Invalid key:", keyid) return end

    if self.game:RawKeyDown(key) then
        return true
    end
    
    local bindings = self.keybindings[key]
    self.control_set:Clear()
    if bindings then
        for k,v in ipairs(bindings) do
            if (self:IsModifierKey(key) or self:CheckModifiers(v)) then
                self.control_set:AddControl( v.control )
            end
        end
    end
    if not self.control_set:IsEmpty() then
        self:DoControlDown( self.control_set, "key" )
    end
    self.last_input_device = "key"
end

function Input:OnKeyRepeat(keyid, modifiers)
    -- for now, just call OnKeyDown
    self:OnKeyDown(keyid, modifiers)
end


function Input:OnKeyUp(keyid, modifiers)
    self.no_input_time = 0
    if self.keys[keyid] then
        self.keys[keyid] = false

        local key = InputConstants.KeyById[keyid]
        if not key then return end

        if self.game:RawKeyUp(key) then
            return true
        end

        local bindings = self.keybindings[key]
        self.control_set:Clear()
        if bindings then
            for k,v in ipairs(bindings) do
                self.control_set:AddControl( v.control )
            end
        end
    
        if not self.control_set:IsEmpty() then
            self:DoControlUp( self.control_set, "key" )
        end
    end
end

function Input:OnMouseMove(x,y,device)
    self.no_input_time = 0
    self.mouse_x = x
    self.mouse_y = (self.h or 1080) - y
    self.last_mouse_device = device or "mouse"
end

function Input:GetLastMouseDevice()
    return self.last_mouse_device or "mouse"
end

function Input:GetLastInputDevice()
    return self.last_input_device or "mouse"
end

function Input:OnMouseWheel(wheel)
    if wheel > 0 then
        self:OnMouseButtonDown(self.mouse_x, self.mouse_y, 1003, "mouse")
        self:OnMouseButtonUp(self.mouse_x, self.mouse_y, 1003, "mouse")
    elseif wheel < 0 then
        self:OnMouseButtonDown(self.mouse_x, self.mouse_y, 1004, "mouse")
        self:OnMouseButtonUp(self.mouse_x, self.mouse_y, 1004, "mouse")
    end
end

function Input:OnMouseButtonDown(x, y, button, device)
    self:OnMouseButtonDownInternal(x, y, button, TheGame:GetLocalSettings().SIMULATE_TOUCH_MODE and "touch" or device or "mouse")
end

function Input:OnMouseButtonUp(x, y, button, device)
    self:OnMouseButtonUpInternal(x, y, button, TheGame:GetLocalSettings().SIMULATE_TOUCH_MODE and "touch" or device or "mouse")
end

function Input:OnMouseButtonDownInternal(x, y, button, device_type)
    self:OnMouseMove(x, y, device_type)

    local button_id = InputConstants.MouseButtonById[button]
    local bindings = button_id and self.mousebindings[button_id]
    self.control_set:Clear()
    if bindings then
        for k,v in ipairs(bindings) do
            if self:CheckModifiers(v) then
                self.control_set:AddControl( v.control )
            end
        end
    end
    
    if not self.control_set:IsEmpty() then
        self:DoControlDown( self.control_set, device_type )
    end

    self.last_input_device = "mouse"
end

function Input:OnMouseButtonUpInternal(x, y, button, device_type)
    self:OnMouseMove(x, y, device_type)

    local button_id = InputConstants.MouseButtonById[button]
    local bindings = button_id and self.mousebindings[button_id]
    self.control_set:Clear()
    if bindings then
        for k,v in pairs(bindings) do
            if self:CheckModifiers(v) then
                self.control_set:AddControl( v.control )
            end
        end
    end
    if not self.control_set:IsEmpty() then
        self:DoControlUp( self.control_set, device_type )
    end
end

function Input:OnTouchBegan(x, y)
    self:OnMouseButtonDownInternal(x, y, InputConstants.MouseButtons.LEFT, "touch")
    self.last_input_device = "touch"
end

function Input:OnTouchMove(x, y)
    self:OnMouseMove(x, y, "touch")
end

function Input:OnTouchEnded(x, y)
    self:OnMouseButtonUpInternal(x, y, InputConstants.MouseButtons.LEFT, "touch")
end

function Input:OnGamepadConnected(gamepad_id, gamepad_name)
    self.gamepads[gamepad_id] = {}
    self.gamepads[gamepad_id].name = gamepad_name;
    self:SetActiveGamepad( gamepad_id )
end

function Input:OnGamepadDisconnected(gamepad_id)
    self.gamepads[gamepad_id] = nil

    if self.active_gamepad == gamepad_id then
        local k, v = next(self.gamepads)
        if k then
            self.active_gamepad = k
        else
            self.active_gamepad = nil
        end
    end
end

function Input:SetActiveGamepad(gamepad_id)
    self.active_gamepad = gamepad_id
end

function Input:GetActiveGamepad()
    return self.active_gamepad
end

local LEFT_STICK_BUTTONS = { LS_UP=true, LS_RIGHT=true, LS_DOWN=true, LS_LEFT=true }

-- If enabled, the left joystick will trigger digital MENU_LEFT, etc like the dpad
function Input:SetLeftJoystickDigitalForwarding( enabled )
    self.left_stick_digital_forwarding = enabled
    return self
end

local RIGHT_STICK_BUTTONS = { RS_UP=true, RS_RIGHT=true, RS_DOWN=true, RS_LEFT=true }

-- If enabled, the right joystick will trigger digital MENU_LEFT, etc like the dpad
function Input:SetRightJoystickDigitalForwarding( enabled )
    self.right_stick_digital_forwarding = enabled
    return self
end

function Input:OnGamepadButtonDown(gamepad_id, button)
    self.no_input_time = 0
    if self.gamepads[gamepad_id] then
        self.gamepads[gamepad_id][InputConstants.GamepadButtonById[button]] = true;

        if self.active_gamepad ~= gamepad_id then
            self:SetActiveGamepad( gamepad_id )
        end

        local button_id = InputConstants.GamepadButtonById[button]
        local bindings = button_id and self.gamepadbindings[button_id]
        self.control_set:Clear()

        if bindings then
            for k,v in pairs(bindings) do

                if LEFT_STICK_BUTTONS[ button_id ] then
                    -- This is a left stick button
                    if self.left_stick_digital_forwarding then
                        -- If digital forwarding is enabled, trigger the digital binding
                        self.control_set:AddControl( v.control )
                    end
                elseif RIGHT_STICK_BUTTONS[ button_id ] then
                    -- This is a right stick button
                    if self.right_stick_digital_forwarding then
                        -- If digital forwarding is enabled, trigger the digital binding
                        self.control_set:AddControl( v.control )
                    end
                else
                    self.control_set:AddControl( v.control )
                end

            end
        end

        if not self.control_set:IsEmpty() then
            self:DoControlDown( self.control_set, "gamepad", self.gamepads[self.active_gamepad].name )
        end
    end
    self.last_input_device = "gamepad"
end

function Input:OnGamepadButtonRepeat(gamepad_id, button)
    -- Just forward it to Button Down:
    self:OnGamepadButtonDown(gamepad_id, button);
end

function Input:OnGamepadButtonUp(gamepad_id, button)
    self.no_input_time = 0    

    if self.gamepads[gamepad_id] then
        self.gamepads[gamepad_id][InputConstants.GamepadButtonById[button]] = false;

        if self.active_gamepad == gamepad_id then
            local button_id = InputConstants.GamepadButtonById[button]
            local bindings = button_id and self.gamepadbindings[button_id]
            self.control_set:Clear()

            -- Do we need to check modifiers?
            if bindings then
                for k,v in pairs(bindings) do
                    self.control_set:AddControl( v.control )
                end
            end
            if not self.control_set:IsEmpty() then
                self:DoControlUp( self.control_set, "gamepad", self.gamepads[self.active_gamepad].name )
            end
        end
    end
end


function Input:GetActiveGamepadStickVals()
    if self.active_gamepad and self.gamepads[self.active_gamepad] then
        return self.gamepads[self.active_gamepad].ls_x or 0, self.gamepads[self.active_gamepad].ls_y or 0, self.gamepads[self.active_gamepad].rs_x or 0, self.gamepads[self.active_gamepad].rs_y or 0, self.gamepads[self.active_gamepad].lt or 0, self.gamepads[self.active_gamepad].rt or 0
    end
    return 0,0,0,0,0,0
end

function Input:OnGamepadAnalogInput(gamepad_id, ls_x, ls_y, rs_x, rs_y, lt, rt)

    if self.gamepads[gamepad_id] then
        self.gamepads[gamepad_id].ls_x = ls_x;
        self.gamepads[gamepad_id].ls_y = ls_y;
        self.gamepads[gamepad_id].rs_x = rs_x;
        self.gamepads[gamepad_id].rs_y = rs_y;
        self.gamepads[gamepad_id].lt = lt;
        self.gamepads[gamepad_id].rt = rt;
    end

    if self.active_gamepad == gamepad_id then
        -- TODO!
    end
end

function Input:DoControlUp( control_set, device_type, device_id )
    if device_type == "gamepad" and self.ignore_gamepad then
        return false
    end

    for i = #control_set.controls, 1, -1 do
       local state = self.control_state[ control_set.controls[i] ]
         -- we need to make sure the control was down, not just the key
         -- as there may be multiple controls bound to one key, with different modifiers
        local wasdown = state and state.down
        if state then
            state.down = false
        end
        if not wasdown then
            table.remove( control_set.controls, i )
        end
    end

    if not control_set:IsEmpty() then
        return self.game:OnControlUp( control_set, device_type, device_id )
    end
end


function Input:DoControlDown( control_set, device_type, device_id )
    if device_type == "gamepad" and self.ignore_gamepad then
        return false
    end

    for i, control in ipairs( control_set.controls ) do
        local state = self.control_state[control]
        if state then
            state.down = true
        end
    end

    if not control_set:IsEmpty() then
        return self.game:OnControlDown( control_set, device_type, device_id )
    end
end

function Input:GetPrimaryKeyBinding( control )
    return (self.key_lookup[control] and self.key_lookup[control][1]) or (self.mouse_lookup[control] and self.mouse_lookup[control][1])
end

function Input:GetBoundKeyString( control )
    local binding = (self.key_lookup[control] and self.key_lookup[control][1]) or (self.mouse_lookup[control] and self.mouse_lookup[control][1])
    return binding and self:GetBindingString( binding )
end

function Input:GetBoundButtonString( control )
    local binding = (self.gamepad_lookup[control] and self.gamepad_lookup[control][1])
    return binding and self:GetBindingString( binding )
end

function Input:GetBindingString( binding )
    local str = ""
    if binding.CTRL then
        str = str.."CTRL-"
    end
    if binding.SHIFT then
        str = str.."SHIFT-"
    end
    if binding.ALT then
        str = str.."ALT-"
    end
    if binding.key then
        str = str .. binding.key
    elseif binding.button then
        str = str .. binding.button
    end
    return str
end

-- returns true if x and y are equivalent key bindings (in terms of the button/modifiers they require)
-- This function is like shallowcompare, except it permits equivalency of x.CTRL == false and x.CTRL == nil, for example.
function Input:IsEqualBinding( x, y )
    if x.key ~= y.key or x.button ~= y.button then
        return false
    end

    if x.CTRL and not y.CTRL then
        return false
    end

    if x.SHIFT and not y.SHIFT then
        return false
    end

    if x.ALT and not y.ALT then
        return false
    end

    return true
end

function Input:UpdateRumble(dt)
    
    local active_gamepad = self:GetActiveGamepad()
    if active_gamepad then
        local small_rumble_value = 0
        local large_rumble_value = 0

        local to_remove
        for k,rumble_instance in ipairs(self.playing_rumbles) do
            rumble_instance.time = rumble_instance.time + (dt * rumble_instance.speed)
            if rumble_instance.rumble:IsDoneAtTime(rumble_instance.time) then
                to_remove = to_remove or {}
                table.insert(to_remove, rumble_instance)
            else
                local small, large = rumble_instance.rumble:GetValues(rumble_instance.time)
                small_rumble_value = small_rumble_value + small*rumble_instance.amp
                large_rumble_value = large_rumble_value + large*rumble_instance.amp
            end
        end

        if to_remove then
            for k,v in ipairs(to_remove) do
                self:KillRumble(v)
            end
        end

        engine.inst:GamepadRumble(active_gamepad, small_rumble_value, large_rumble_value)

        self.small_rumble_value, self.large_rumble_value = small_rumble_value, large_rumble_value
    else
        self:ClearRumble()
    end

    for k,v in pairs(self.gamepads) do
        if k ~= active_gamepad then
            engine.inst:GamepadRumble(k, 0, 0)
        end
    end

end

function Input:PlayRumble(rumble, speed, amp)
    rumble = type(rumble) == "string" and Content.GetRumble(rumble) or rumble
    if rumble then
        local rumble_instance = {rumble = rumble, time = 0, speed = speed or 1, amp = amp or 1}
        table.insert(self.playing_rumbles, rumble_instance) 

        return rumble_instance
    end
end

function Input:KillRumble(rumble_instance)
    table.arrayremove(self.playing_rumbles, rumble_instance)
end

function Input:ClearRumble()
    table.clear(self.playing_rumbles)
end

