require "locationutil"
local CastMember = require "castmember"

local memtokens = require "memtokens"

-----------------------------------------
-- Location bit flags.

LOCFLAGS, LOCFLAGS_STRINGS = MakeBitField{
    "CLOSED", -- just closed
    "HOME", -- Is a home
    "DOOR", -- is a temporary door adjacent to a single location
    "TRANSIENT", -- Removed during UpdateWorld if no owner and no occupants.
    "MAP", -- An outdoors map location; will create an entity with a Squad.
    "HIDDEN", --not shown on the travel screen
}

-----------------------------------------

local function ResolveContentID( content_id )
    local def = Content.GetLocationContent( content_id )
    if def == nil then
        def = Content.GetLocationContent( "MISSING_LOCATION" )
        LOGWARN( "Missing location data for '%s'", content_id )
    end
    return def
end

-----------------------------------------

local Location = class("Location")

CastMember.AddInterface( Location )

Location.can_reload = true
 
function Location:init( content_id, region )
    self.content_id = content_id
    self.region = region

    self.location_data = Content.GetLocationContent(content_id)
    if self.location_data == nil then
        self.location_data = Content.GetLocationContent( "MISSING_LOCATION" )
        LOGWARN( "Missing location data for '%s'", content_id )
    end
    self.agents = {}
    self.exits = {}
    self.tags = {}
    self.flags = 0

end

function Location:OnDoViolence(victim)
    if not self:HasMemory("DID_VIOLENCE", 1) then
        self:Remember("DID_VIOLENCE")
        if self:GetContent().on_violence then
            self:GetContent().on_violence(self, victim)
        end
    end
end

function Location:SetPlaxElementVisiblity( element_id, valid )
    if self.plax_element_ids == nil then
        self.plax_element_ids = {}
    end
    self.plax_element_ids[ element_id ] = valid

    self:BroadcastEvent( "plax_changed", self, element_id )
end

function Location:GetPlax()
    if self.plax then
        return self.plax
    end

    if self.location_data.plax then
        return self.location_data.plax
    end
    
    return "PLACEHOLDER"
end

function Location:SetPlax( plax_id )
    if self.plax ~= plax_id then
        self.plax = plax_id
        
        for _,agent in pairs(self.agents) do
            agent.plax_slot_id = nil
            agent.plax_name = nil
        end
        self:BroadcastEvent( "plax_changed", self)
    end
end

function Location:RefreshPlax( plax )
    local plaxname = self:GetPlax()
    if plaxname then
        plax:SetPlaxData( plaxname )
        plax:SetWindy( not self:IsIndoors() )
        plax:SetPlaxElementVisibility( self.plax_element_ids )
    end

    return plaxname
end

function Location:FillOutQuipTags()
    local tags = { self.content_id:lower() }
    for k, v in pairs( self.tags ) do
        table.insert( tags, k )
    end

    if self.location_data.tags then
        for k, v in pairs(self.location_data.tags) do
            table.insert( tags, v )
        end
    end
    for k, v in pairs(self.location_data) do
        if v == true then
            table.insert( tags, k )
        end
    end

    return tags
end

function Location:GetFactionID()

    if self.location_data.faction_id then
        return self.location_data.faction_id

    elseif CheckAnyBits( self:GetFlags(), LOCFLAGS.HOME ) then
        local exit = self:GetDefaultExit()
        if exit then
            return exit:GetTo():GetFactionID()
        end
        
    else
        -- Not sure the point of this dynamic faction evaluation.
        -- We don't want it to happen for homes, because otherwise we can't query
        -- whether the home is valid for a given agent (if it always returns the occupant's
        -- faction, then it will never become invalid)
        local agent = self:GetProprietor()
        if agent then
            return agent:GetFactionID()
        end
    end

    return NEUTRAL_FACTION
end

function Location:GetID()
    return self:GetGUID()
end

function Location:GetGUID()
    return self.guid
end

function Location:GetUUID()
    return self.uuid
end

function Location:IsIndoors()
    return self.location_data.indoors == true
end

function Location:GetContent()
    return self.location_data
end

function Location:GetContentID()
    return self.content_id
end



function Location:WorkPositions()
    return pairs( self.work or table.empty )
end

function Location:SetMapPos(x,y)
    self.map_x, self.map_y = x, y
end

function Location:GetMapPos()
    return self.map_x, self.map_y
end


function Location:RecurseExits( fn, visited, ... )
    if visited == nil then
        visited = {}
    end
    if visited[ self ] then
        return
    end
    visited[ self ] = true

    fn( self, ... )

    for i, exit in ipairs( self.exits ) do
        if not exit.is_fasttravel then
            exit:GetTo():RecurseExits( fn, visited, ... )
        end
    end
end

function Location:HasExits()
    return #self.exits > 0
end

function Location:GetExits()
    return self.exits
end

local function IterateExits( state, i )
    local exit
    i, exit = next( state.exits, i )
    if i then
        return i, exit, exit:GetTo()
    end
end

function Location:Exits()
    return IterateExits, self
end

function Location:AddExit( exit, idx )
    assert( is_instance( exit, Location.Exit ))
    table.insert( self.exits, idx or (#self.exits + 1), exit )

    return #self.exits
end

function Location:RemoveExit( idx )
    assert( self.exits[ idx ] )
    table.remove( self.exits, idx )
end

function Location:FindExit( to_location )
    for i, exit, to in self:Exits() do
        if to == to_location then
            return i, exit
        end
    end
end


function Location:GetRootLocation()
    if self:GetDefaultExit() then
        return self:GetDefaultExit():GetTo():GetRootLocation()
    end

    return self
end

function Location:GetDefaultExit()
    for i, exit in ipairs( self.exits ) do
        if exit.is_default then
            return exit
        end
    end
end

function Location:IsReachable( location )
    -- FIXME: this could be much faster if we assign all reachable locations a shared identifier.
    local found = false
    local function Test( t )
        found = found or (t == location)
    end
    self:RecurseExits( Test )
    return found
end

function Location:Save()

    local data = 
    {
        content_id = self.content_id,
        uuid = self.uuid,
        flags = self.flags,
        name_idx = self.name_idx,
        map_x = self.map_x,
        map_y = self.map_y,
        plax_element_ids = self.plax_element_ids,
    }

    data.tags = next(self.tags) and self.tags

    if next(self.exits) then
        data.exits = {}
        for k, exit in pairs(self.exits) do
            local dest = exit:GetTo()
            if dest then
                table.insert( data.exits, exit )
            end
        end
    end

    if self.quest_membership and next(self.quest_membership) then
        data.quest_membership = self.quest_membership
    end

    data.parent_location = self.parent_location
    data.guid = self.guid
    data.memory = self.memory
    data.work = self.work
    data.room_occupants = self.room_occupants
    data.removed = self.removed

    data.agents = next(self.agents) and self.agents
    data.plax = self.plax
    data.region = self.region
    
    return data
end

function Location:GetWorldRegion()
    return self.region
end

function Location:GetWorkPosition(id)
    return self.work and self.work[ id ] 
end

function Location:GetWorkers(workers)
    workers = workers or {}
    for k, wp in pairs( self.work ) do
        if wp:GetAgent() then
            table.insert( workers, wp:GetAgent() )
        end
    end
    return workers
end

function Location:Load()
    self.location_data = ResolveContentID( self.content_id )
    self.agents = self.agents or {}
    self.exits = self.exits or {}
    self.tags = self.tags or {}
    self.flags = self.flags or 0
    self.work = self.work or {}

    if self.plax == nil then
        self:InitializePlax()
    end
end

function Location:PostLoad()
    for id, wp in pairs( self.work ) do
        if not wp.content then
            print( "Clearing invalid work position:", id )
            self.work[ id ] = nil
        end
    end
end

function Location:GenerateLocName()
    local content = self:GetContent()
    if content.codex_entry == nil then
		local name_key = content:GetLocNameKey()
		local loc_name = LOC(name_key)

        if content.name_fn then
			self.name = content.name_fn(self, loc_name)
		else
			self.name = loc_name
        end
    end
end

function Location:GetName()
    self:GenerateLocName()
    local name
    if self.name then
        name = self.name
    elseif self:GetContent().codex_entry then
        local codex_entry = Content.LookupCodexEntry( self:GetContent().codex_entry )
        if codex_entry then
            name = codex_entry:GetLocalizedTitle()
        end
    end

    local is_door = CheckBits( self:GetFlags(), LOCFLAGS.DOOR )
    if is_door then
        local exit = self.exits[1]
        local location = exit and exit:GetTo()
        return loc.format("Outside {1}", location and location:GetName() or "????")
    end
    return name
end

function Location:GetIcon()
    return self.location_data.icon or engine.asset.Texture("UI/location_default.tex")
end

function Location:__tostring()
    -- ccc: DO NOT USE GetName().  It has lots of runtime dependencies.
    return string.format("[%d:%s]", self.guid or -1, self:GetContentID() )
end


function Location:IsClosed()
    
    local flags = self:GetFlags()
    
    if CheckBits( flags, LOCFLAGS.CLOSED ) then
        return true, "Closed"
    end

    return false
end

function Location:HasMemory(token, duration)
    return memtokens.HasMemory( self.memory, token, duration)
end

function Location:Remember(token, val)
    if self.memory == nil then
        self.memory = {}
    end
    return memtokens.Remember( self.memory, token, val )
end

function Location:Forget(token)
    return memtokens.Forget( self.memory, token )
end


function Location:FindDoorLocation()
    for k,exit in pairs(self.exits) do
        local to = exit:GetTo()
        if to:CheckFlag( LOCFLAGS.DOOR ) then
            return exit:GetTo()
        end
    end
end

function Location:GetDoorLocation( from )
    local ext_location = self:FindDoorLocation( from )
    if ext_location then
        return ext_location
    end

    -- Not yet generated: auto-generate.
    local is_home = CheckBits( self:GetFlags(), LOCFLAGS.HOME )

    local content_id = self:GetContent().exterior_loc
    if content_id then
        local ext_location, entrance_exit, room_exit =
            self:GenerateSublocation( content_id, LOCFLAGS.TRANSIENT | LOCFLAGS.DOOR, false )

        room_exit.can_break_in = true
        room_exit.can_knock = is_home

        if from then
            -- ext_location -> from
            local exit1 = Location.Exit( from )
            exit1.is_default = true
            ext_location:AddExit( exit1 )

            -- 'from' --> ext_location
            local exit2 = Location.Exit( ext_location )
            from:AddExit( exit2 )
        end

        return ext_location
    end
end

function Location:GenerateSublocation( content_id, flags, is_default_exit )
    assert( content_id )
    local new_room = Location( content_id, self:GetWorldRegion() )
    new_room.parent_location = self
    if flags then
        new_room.flags = new_room.flags | flags
    end
    
    local game_state = TheGame:GetGameState()
    game_state:AddLocation(new_room)

    local room_exit = Location.Exit( self )
    room_exit.is_default = is_default_exit
    room_exit.can_hear_through = true
    new_room:AddExit(room_exit)

    
    local entrance_exit = Location.Exit(new_room)
    self:AddExit(entrance_exit)

    return new_room, entrance_exit, room_exit
end


function Location:AddAgent( agent )
    assert( self.guid, self )
    assert( table.arrayfind( self.agents, agent ) == nil )
    table.insert(self.agents, agent)
    self:BroadcastEvent( "agents_changed", self, agent )
end

function Location:RemoveAgent( agent )
    table.arrayremove( self.agents, agent )
    self:BroadcastEvent( "agents_changed", self, agent )

    if #self.agents == 0 then
        if CheckBits( self:GetFlags(), LOCFLAGS.TRANSIENT ) then
            TheGame:GetGameState():RemoveLocation( self )
        end
    end
end

function Location:Agents()
    return ipairs(self.agents)
end

function Location:GetAgentByIndex( i )
    return self.agents[i]
end

function Location:FindAgent( agent )
    return table.arrayfind( self.agents, agent )
end

function Location:FindAgentIf( pred, ... )
    for i, agent in ipairs( self.agents ) do
        if pred( agent, ... ) then
            return agent
        end
    end
end

function Location:GetAgentCount()
    return #self.agents
end

function Location:IsEmpty()
    return #self.agents == 0
end

function Location:GetDesc()
	local content = self:GetContent()
	local desc = content:GetLocalizedDesc()
    return desc
end

function Location:AddTag( tag )
    self.tags[ tag ] = true
end

function Location:RemoveTag( tag )
    self.tags[ tag ] = nil
end

function Location:HasTag( tag )
    if self.tags[ tag ] == true then
        return true
    end
    if self.location_data.tags and table.arrayfind( self.location_data.tags, tag ) then
        return true
    end
    if self.location_data[ tag ] == true then
        return true
    end
    return false
end

function Location:GetFlags()
    return self.flags | (self.location_data.flags or 0)
end

function Location:SetFlags( flags )
    self.flags = SetBits( self.flags, flags )
end

function Location:CheckFlag( flags )
    return CheckBits( self:GetFlags(), flags )
end

function Location:Tags()
    return pairs( self.tags )
end

function Location:GetAgentByRole( role )
    for i, agent in ipairs( self.agents ) do
        local v = agent:GetRoleAtLocation( self )
        if v == role then
            return agent
        end
    end
end

function Location:PopulateGameState( game_state )
    if self.guid == nil then
        self.guid = game_state:GenerateID()
    end

    if self.uuid == nil then
        self.uuid = engine.inst:NewGUID()
    end

    local content = self:GetContent()
    assert( content )

    self:GenerateLocName()

    if content.sublocations then
        if #self.exits == 0 then
            for k, content_id in pairs(content.sublocations) do
                local v = Content.GetLocationContent( content_id )
                for i = 1, v.num or 1 do
                    local room, entrance, exit = self:GenerateSublocation( content_id, nil, true, v.requires_admittance )
                end
            end
        else
            -- Sublocations already created: just add to game state.
            for i, exit in ipairs( self.exits ) do
                local to = exit:GetTo()
                if to and to.guid == nil then
                    TheGame:GetGameState():AddLocation( to )
                end
            end
        end
    end

    if content.work then
        self.work = self.work or {}
        
        for k,v in pairs( content.work ) do
            self:AddWorkPosition(k)
        end
    end
    
    self:InitializePlax()

end

function Location:GetWorkData( id )
    if self.location_data == nil then
        self.location_data = ResolveContentID( self.content_id )
    end

    return self.location_data and self.location_data.work and self.location_data.work[ id ]
end

function Location:RemoveWorkPosition(local_name)
    if self.work[local_name] then
        if self.work[local_name]:GetAgent() then
            self.work[local_name]:Fire()
        end
        self.work[local_name] = nil
    end
end

function Location:AddWorkPosition(work_id)
    assert( work_id )
    assert(self.work[work_id] == nil, "Already have a job of that name")
   
    local wp = WorkPosition( self, work_id )
    self.work[work_id] = wp

    wp:DoHiring()

    return wp
end

function Location:DoHiring()
    if self.work then
        for k,v in pairs(self.work) do
            v:DoHiring()
        end
    end
end


function Location:Depopulate()
    print ("Location:Depopulate", self)

    -- Remove transience, so that clearing agents doesn't trigger a recursive Depopulate.
    local is_transient = CheckBits( self:GetFlags(), LOCFLAGS.TRANSIENT )
    self.flags = ClearBits( self.flags, LOCFLAGS.TRANSIENT )

    -- Move all occupying agents into limbo.
    while #self.agents > 0 do
        local agent = self.agents[ #self.agents ]
        assert_warning( not agent:IsInPlayerParty() )
        agent:MoveToLimbo()
        assert( not table.arrayfind( self.agents, agent )) -- Sanity check
    end

    if self.work then
        for work_id, wp in pairs( self.work ) do
            if wp:GetAgent() then
                wp:Fire()
            end
        end
        self.work = nil
    end

    -- Means we're no longer attached to GameState.
    self.guid = nil

    -- Recursively Depopulate exit locations if this isn't a TRANSIENT location
    if not is_transient then
        for i, exit in ipairs( self.exits ) do
            local to = exit:GetTo()
            if to and to.guid ~= nil then
                TheGame:GetGameState():RemoveLocation( to )
            end
        end
    end

    while #self.exits > 0 do
        local exit = table.remove( self.exits )
        local to = exit:GetTo()
        if to then
            local i, rev_exit = to:FindExit( self )
            if i then
                to:RemoveExit( i )
            end
        end
    end    
end

function Location:GetProprietor()
    if self.work then
        for work_id, work in pairs( self.work ) do
            local agent = work:GetAgent()
            if agent and work:GetRole() == CHARACTER_ROLES.PROPRIETOR then
                return agent
            end
        end

    elseif CheckBits( self:GetFlags(), LOCFLAGS.HOME ) then
        return self:GetPrimaryRoomOccupant()
    end
end



function Location:AddRoomOccupant(new_occupant)
    if self.room_occupants == nil then
        self.room_occupants = {}
    end
    assert( table.arrayfind( self.room_occupants, new_occupant ) == nil )
    table.insert( self.room_occupants, new_occupant )
end

function Location:RemoveRoomOccupant( agent )
    table.arrayremove( self.room_occupants, agent )
    if #self.room_occupants <= 0 then
        TheGame:GetGameState():MarkLocationForDeletion(self)
    end
end

function Location:GetRoomOccupants()
    if self.room_occupants then
        return self.room_occupants
    end
    return nil
end

function Location:GetPrimaryRoomOccupant()
    if self.room_occupants then
        return self.room_occupants[1]
    end
    return nil
end

function Location:ClearAgents()
    local t = ObtainWorkTable()

    local function Inner(location)
        for k,agent in pairs(location.agents) do           
            if agent:IsCurrentlyImportant() then
                --
            elseif agent:IsInPlayerParty() then
                --
            else
                table.insert(t, agent)
            end
            
        end
    end
    self:RecurseExits(Inner)
    
    for k, agent in pairs(t) do
        agent:MoveToLimbo()
    end
    ReleaseWorkTable(t)
end

function Location:SetCurrentPatronCapacity( capacity )
    self.patron_capacity = capacity
end

function Location:GetCurrentPatronCapacity()

    if self:IsClosed() then
        return 0
    end
    
    if not self:GetContent().patron_data then
        return 0
    end

    if self.patron_capacity then
        return self.patron_capacity
    end

    if type(self:GetContent().patron_data.num_patrons) == "table" then
        local phase = TheGame:GetGameState():GetDayPhase()
        return self:GetContent().patron_data.num_patrons[phase] or 0
    else
        return self:GetContent().patron_data.num_patrons or 0
    end
    
end

function Location:GetNumCurrentPatrons()
    local num = 0
    
    for k,agent in pairs(self.agents) do
        if agent:GetBrain():IsPatronizing() then
            num = num + 1
        end
    end

    return num
end

function Location:InitializePlax()
    local content = self:GetContent()
    if type(content.plax) == "function" then
        self.plax = content.plax(self)
    end
end

function Location:IsCurrentlyImportant()
    if not TheGame:GetGameState() then
        return false
    end

    local is_important = false
    TheGame:GetGameState():GetTracker():RecurseTrackedObjects( self, function( obj, sources )
        if not is_instance( obj, Agent ) or not obj:IsPlayer() then
            is_important = is_important or (sources ~= nil)
            return true
        end
    end )
    return is_important
end

function Location:UpdateLocation()
    -- Patrons can respawn here.
    self.patron_capacity = nil

    if CheckBits( self:GetFlags(), LOCFLAGS.TRANSIENT ) then
        if #self.agents == 0 and self:GetProprietor() == nil and not self:IsCurrentlyImportant() then
            TheGame:GetGameState():RemoveLocation( self )
            return
        end
    end
end

function Location:ListenForEvent( ... )
    if self.events == nil then
        self.events = EventSystem()
    end
    self.events:ListenForEvent( ... )
end

function Location:ListenForAny(listener, fn)
    if self.events == nil then
        self.events = EventSystem()
    end
    self.events:ListenForAny( listener, fn )
end

function Location:BroadcastEvent( event, ... )
    if self.events then
        self.events:BroadcastEvent( event, ... )
    end
end

function Location:RemoveListener( listener )
    if self.events then
        self.events:RemoveListener( listener )
        if not self.events:HasListeners() then
            self.events = nil
        end
    end
end

--------------------------------------------------------------
-- Represents an edge between two Locations, a and b.

local Exit = class( "Location.Exit" )

function Exit:init( to )
    self.to = to
end

function Exit:GetTo()
    -- self.to is either a raw Location reference, or a GUID to a Location in another area.
    -- This is because Locations are saved per Area and we don't want Exits to save outer-of-area Lcoations.
    if is_instance( self.to, Location ) then
        return self.to
    else
        return TheGame:GetGameState():GetLocation( self.to )
    end
end

function Exit:Save()
    return self
end


