require "party"

-----------------------------------------------------------
-- Manages the state of the caravan while traveling.

local Caravan = class ("Caravan" )

function Caravan:init()
    self.party = Party()
    self.cooldowns = {}
    self.max_resolve = 0
    self.resolve = 0
end

function Caravan:Save()
    local data =
    {
        party = self.party,
        cooldowns = self.cooldowns,
        resolve  = self.resolve,
        max_resolve = self.max_resolve,
        last_travel_destination = self.last_travel_destination,
    }
    return data
end

function Caravan:GetResolve()
    return self.resolve, self.max_resolve
end

function Caravan:SetResolve(val)
    val = clamp( val, 0, self.max_resolve )
    if val ~= self.resolve then
        local old_resolve, old_max = self.resolve, self.max_resolve
        self.resolve = val
        TheGame:BroadcastEvent( "resolve_change", self.resolve, old_resolve, self.max_resolve, old_max )
    end
end
function Caravan:SetResolvePercent(p)
    self:SetResolve( math.round( p*self.max_resolve) )
end

function Caravan:DeltaMaxResolve( delta )
    local old_max = self.max_resolve
    local old_resolve = self.resolve
    self.max_resolve = math.max( 0, self.max_resolve + delta )
    if delta > 0 then
        self.resolve = self.resolve + delta
    end
    self.resolve = math.min(self.resolve, self.max_resolve)
    TheGame:BroadcastEvent( "resolve_change", self.resolve, old_resolve, self.max_resolve, old_max )
end

function Caravan:UpgradeResolve( delta, recover, gain_empty, no_audio )
    if not no_audio then
        AUDIO:PlayEvent("event:/ui/char/gen/max_resolve_increase")
    end
    local old_max = self.max_resolve
    local old_resolve = self.resolve
    self.max_resolve = self.max_resolve + delta
    if not gain_empty then
        self.resolve = self.resolve + delta
    end
    if recover then
        self.resolve = self.max_resolve
    end
    TheGame:BroadcastEvent( "resolve_change", self.resolve, old_resolve, self.max_resolve, old_max )

    TheGame:GetGameState():LogNotification( NOTIFY.RESOLVE_UPGRADED, delta )
end

function Caravan:SetMaxResolve( new_max )
    self.max_resolve = new_max
    self.resolve = new_max

    TheGame:BroadcastEvent( "resolve_change", self.resolve, 0, self.max_resolve, 0 )
end

function Caravan:DeltaResolve( amt )
    local old = self.resolve
    self.resolve = clamp(self.resolve + amt, 0, self.max_resolve)
    if self.resolve ~= old then
        TheGame:BroadcastEvent( "resolve_change", self.resolve, old, self.max_resolve, self.max_resolve )
    end
    return self.resolve - old
end

function Caravan:UpgradeHealth( amt, id )
    id = id or "MISC"
    AUDIO:PlayEvent("event:/ui/char/gen/max_health_increase")
    self:GetPlayer().health:AddStatModifier( id, amt )
    TheGame:GetGameState():LogNotification( NOTIFY.HEALTH_UPGRADED, amt )
end

function Caravan:SetCurrentEncounter(enc)
    if not ( self.current_encounter == nil or enc == nil ) then
        print( "WARNING: stomping existing current_encounter with:", self.current_encounter:GetAgent())
        print( debug.traceback() )
    end

    self.current_encounter = enc

    if enc then
        print( "*** NEW ENCOUNTER:", self.current_encounter.initial_hub or "DefaultHub" )
        TheGame:BroadcastEvent( "encounter_changed", self.current_encounter )
    else
        -- Ensure we re-check confronts when this encounter completes.
        self:RefreshConfronts()
    end
end

function Caravan:GetCurrentEncounter()
    return self.current_encounter
end

function Caravan:StartCooldown( key, duration )
    assert( duration >= 0 )
    local now = Now()
    self.cooldowns[ key ] = now + duration
end

function Caravan:StartTimer( key )
    local now = Now()
    self.cooldowns[ key ] = now
end

function Caravan:CheckCooldown( key )
    local cooldown = self.cooldowns[ key ]
    if cooldown then
        local now = Now()
        if now < cooldown then
            return cooldown - now -- Time remaining
        else
            self.cooldowns[ key ] = nil -- Lazy clear.
        end
    end
end

function Caravan:CheckTimer( key, duration )
    local cooldown = self.cooldowns[ key ]
    if cooldown then
        local now = Now()
        if now < (cooldown + duration) then
            return (cooldown + duration) - now -- Time remaining
        else
            self.cooldowns[ key ] = nil -- Lazy clear.
        end
    end
end

function Caravan:CollectNegotiationCards( all_cards, filterfn )
    all_cards = all_cards or {}

    local player = self:GetPlayer()
    for i, card in ipairs( player.negotiator:GetCards() ) do
        if not filterfn or filterfn(card) then
            table.insert( all_cards, card:Clone() )
        end
    end

    return all_cards
end

function Caravan:PostLoad( game_state )
    for k, agent in ipairs( self.party:GetMembers() ) do
        agent:ListenForAny( self, self.OnAgentEvent )
    end
    
    -- In case any confronts should run, ensure we re-evaluate on load.
    self:RefreshConfronts()
end

function Caravan:CanOwnPet( pet )
    local pet_details = { max_pets = 1 }
    TheGame:BroadcastEvent( "calc_max_pets", pet_details )

    if pet_details.max_pets <= 0 then
        return false, pet_details.reason or LOC "UI.REQ_CANT_OWN_PET"
    else
        local pet_count = 0
        for i, agent in self.party:Members() do
            if agent:IsPet() then
                pet_count = pet_count + 1
            end
        end

        if pet_count >= pet_details.max_pets then
            return false, loc.format( LOC "UI.REQ_MAX_PETS", pet_details.max_pets )
        end
    end

    return true
end

function Caravan:GetPet( )
    for i, agent in ipairs( self.party:GetMembers() ) do
        if agent:IsPet() then
            return agent
        end
    end
end

function Caravan:GetPets( )
    local pets = {}
    for i, agent in ipairs( self.party:GetMembers() ) do
        if agent:IsPet() then
            table.insert(pets, agent)
        end
    end

    return pets
end

function Caravan:AddMember( agent, type )
    assert( agent.guid )
    self.party:AddMember(agent, type)
    agent:ListenForAny( self, self.OnAgentEvent )
    TheGame:GetEvents():BroadcastEvent( "caravan_member_add", agent )
    TheGame:GetGameState():LogNotification( NOTIFY.PARTY_MEMBER_ADDED, agent )

    return agent
end

function Caravan:RemoveMember( agent )
    -- use agent:Dismiss() instead of calling this directly.
    self.party:RemoveMember(agent)
    agent:RemoveListener( self )
    TheGame:GetEvents():BroadcastEvent( "caravan_member_remove", agent )
    TheGame:GetGameState():LogNotification( NOTIFY.PARTY_MEMBER_REMOVED, agent )
end

function Caravan:RemoveFollowers( )
    local mem = shallowcopy(self.party:GetMembers())    
    for k, agent in ipairs(mem) do
        if not agent:IsPlayer() then
            if agent:IsPet() then
                QuestUtil.RunAwayPet(agent)
            else
                agent:Dismiss()
            end
        end
    end
end

function Caravan:OnAgentEvent( event_name, ... )
    -- Forward all agent events.
    -- This allows interested observers to handle all party member events without manually subscribing to each agent.
    TheGame:GetEvents():BroadcastEvent( "caravan_member_event", event_name, ... )
end


function Caravan:HasEscort()
    return self.party:GetMemberType( PARTY_MEMBER_TYPE.ESCORT ) ~= nil
end

function Caravan:HasHireling()
    return self.party:GetMemberType( PARTY_MEMBER_TYPE.HIRED ) ~= nil
end

function Caravan:GetHireling()
    return self.party:GetMemberType( PARTY_MEMBER_TYPE.HIRED )
end

function Caravan:GetCaptive()
    return self.party:GetMemberType( PARTY_MEMBER_TYPE.CAPTIVE )
end

function Caravan:PickMember( pred_fn )
    local options = {}
    for i, member in pairs( self.party:GetMembers() ) do
        if pred_fn == nil or pred_fn( member ) then
            table.insert( options, member )
        end
    end
    return table.arraypick(options)
end

function Caravan:GetParty()
    return self.party
end

function Caravan:GetPartyCount()
    return #self.party:GetMembers()
end

function Caravan:GetPartyMember( idx )
    return self.party:GetMembers()[ idx ]
end

function Caravan:GetPlayer()
    return self.party:GetMembers()[1]--TheGame:GetGameState():GetPlayerAgent()
end

function Caravan:Members()
    return ipairs( self.party:GetMembers() )
end

function Caravan:EvaluateConfronts()

    local location = self:GetPlayer():GetLocation()
    if location == nil then
        return
    end

    local hub_state = HUB_STATE.CONFRONT
    local hubs = ObtainWorkTable()
    local enc = Encounter( nil, location )
    enc:CollectHubs( hubs, hub_state )

    for i, agent in location:Agents() do
        if not agent:IsPlayer() then
            local enc = Encounter( agent, location )
            enc:CollectHubs( hubs, hub_state )
        end
    end

    -- NOTE: There are many ConfrontStates() which do not conditionalize on the confronted agent (ie. whether
    -- cxt:GetAgent() == nil, or not), even though they SHOULD be specifying this.
    -- When sorting hubs with identical priority, sometimes a hub with a nil primary cast will come first,
    -- but sometimes not, resulting in occasionaly inadvertant primary casts showing up at the beginning of a convo.
    -- To impose a total ordering, ensure hubs with no cast always appear first in the sorted listing.

    table.sort( hubs, function( a, b )
        if a:GetPriority() == b:GetPriority() then
            return b:GetAgent() and not a:GetAgent()
        else
            return a:GetPriority() > b:GetPriority()
        end
    end )
    
    local hub = hubs[1]
    ReleaseWorkTable( hubs )

    return hub, hub_state
end

function Caravan:RefreshConfronts()
    self.refresh_confronts = true
end

function Caravan:CheckConfronts()
    if not self.refresh_confronts then
        return
    end
    if self:GetCurrentEncounter() and not self:GetCurrentEncounter():IsEnded() then
        print( "WARNING: skipping DoEnterConvos (already in an Encounter" )
        print( debug.traceback() )
        return
    end
    if TheGame:FE():FindScreen( Screen.ConversationScreen ) then
        print( "WARNING: CheckConfronts skipped due to ConversationScreen" )
        print( debug.traceback() )
        return
    end

    self.refresh_confronts = false

    local hub, hub_state = self:EvaluateConfronts()
    if hub then
        hub.encounter:SetInitialHub( hub, hub_state )
        self:SetCurrentEncounter( hub.encounter )
        return true
    end
end

function Caravan:SetMoney(amt)
    local money = self:GetPlayer():GetMoney()
    local delta = amt - money
    if delta ~= 0 then
        TheGame:GetGameState():GetPlayerAgent():DeltaMoney( delta )
        TheGame:GetGameState():LogNotification( NOTIFY.DELTA_MONEY, delta )
    end
end

function Caravan:AddMoney(amt)
    assert_warning( amt >= 0, tostring(amt) )
    amt = math.floor( amt )
    TheGame:GetGameState():GetPlayerAgent():DeltaMoney( amt )
    TheGame:GetGameState():LogNotification( NOTIFY.DELTA_MONEY, amt )
    TheGame:AddGameplayStat( "money_gained", amt )
end

function Caravan:PayMoney( total_amt, agent )
    assert( total_amt >= 0 )

    local delta = math.min( self:GetMoney(), total_amt )
    
    TheGame:GetGameState():GetPlayerAgent():DeltaMoney( -delta )
    TheGame:GetGameState():LogNotification( NOTIFY.DELTA_MONEY, -delta )

    if agent then
        agent:DeltaMoney( delta )
    end
end

function Caravan:GetMoney()
    return self:GetPlayer():GetMoney()
end

function Caravan:MoveToLocation( location )
    assert( location )
    local player = self:GetPlayer()
    
    local old_location = player:GetLocation()
    if location ~= old_location then
        player:GetParty():MoveToLocation( location )
        
        TheGame:GetGameState():PopulateAgentsInWorld()
        LocationUtil.PopulateLocation( location )
        TheGame:BroadcastEvent( "caravan_move_location", location )
        
        
        if old_location then
            local is_transient = CheckBits( old_location:GetFlags(), LOCFLAGS.TRANSIENT )
            if is_transient then
                TheGame:GetGameState():RemoveLocation( old_location )
            end
        end
    end
end
