require "map/faction"

local FactionMgr = class("FactionMgr")

function FactionMgr:init()
    -- List of <faction_id1, faction_id2, relationship> tuples (faction_id1 < faction_id2, ordered lexographically)
    -- NOTE: Relationships are bidrectional, hence they are stored centrally in the FactionMgr.
    self.relationships = {}
    self.factions = {}
end

function FactionMgr:InitialPopulate()
    for faction_id, faction_data in Content.AllFactions() do
        -- NPC factions only: player faction can vary, and is therefore added when the player agent is
        if self.factions[ faction_id ] == nil and not faction_data.player then
            self:AddFaction( faction_id, faction_id )
        end
    end

    for faction_id,faction in pairs(self.factions) do
        local content = faction:GetContent()
        
        if content.hostile then
            for other_id, other_faction in pairs(self.factions) do
                if other_id ~= faction_id then
                    self:SetFactionRelationship( faction_id, other_id, RELATIONSHIP.HATED )
                end
            end
        elseif content.relationships then
            for other_id, rel in pairs( content.relationships ) do
                self:SetFactionRelationship( faction_id, other_id, rel )
            end
        end
    end
end

function FactionMgr:AddFaction( id, content_id )
    assert( self.factions[ id ] == nil, id )
    local new_faction = Faction( self, id, content_id )
    self.factions[ id ] = new_faction
end


function FactionMgr:GetFactionRelationship( faction_id1, faction_id2 )
    -- 'nil' factions are treated as impartial to everything.
    if not faction_id1 or not faction_id2 then
        return RELATIONSHIP.NEUTRAL
    elseif faction_id1 == faction_id2 then
        return RELATIONSHIP.LOVED
    end

    assert( self.factions[ faction_id1 ] and self.factions[ faction_id2 ])
    if faction_id1 > faction_id2 then
        faction_id1, faction_id2 = faction_id2, faction_id1
    end

    local t = self.relationships[ faction_id1 ]
    
    local rep = t and t[ faction_id2 ] or RELATIONSHIP.NEUTRAL
    return rep
end

function FactionMgr:SetFactionRelationship( faction_id1, faction_id2, rel )
    assert( faction_id1 and faction_id2 )
    assert( faction_id1 ~= faction_id2 )
    assert( rel )
    

    if faction_id1 > faction_id2 then
        faction_id1, faction_id2 = faction_id2, faction_id1
    end

    local old_rel = self:GetFactionRelationship(faction_id1, faction_id2)
    if old_rel ~= rel then
        local t = self.relationships[ faction_id1 ]
        if t == nil then
            t = {}
            self.relationships[ faction_id1 ] = t
        end
        t[ faction_id2 ] = rel

    end
    
    return rel
end

function FactionMgr:GetFaction( faction_id )
    local faction = self.factions[ faction_id ]
    assert(faction or faction_id == nil, string.format("Invalid faction id [%s]", faction_id))
    return faction
end

function FactionMgr:FindFaction( faction_id )
    local faction = self.factions[ faction_id ]
    return faction
end

function FactionMgr:Factions()
    return pairs( self.factions )
end

function FactionMgr:Save()
    return self
end

function FactionMgr:PostLoad()
    for faction_id, faction in pairs( self.factions ) do
        faction:PostLoad( faction_id, self )
    end

    for faction_id, faction_data in Content.AllFactions() do
        -- NPC factions only: player faction can vary, and is therefore added when the player agent is
        if self.factions[ faction_id ] == nil and not faction_data.player then
            self:AddFaction( faction_id, faction_id )
        end
    end
end

