local FXManager = class("FXManager")
local POOL_SIZE = 100

function FXManager:init( sim )
    self.sim = sim

    self.playing_fx = {}
    self.fx_pool = {}

    self.root = engine.rendering.SceneGraphNode()
    --self.root:SetName("FX ROOT")
    sim:GetRoot():AddChild(self.root)

    for k = 1, POOL_SIZE do
        local fx = {}
        fx.model = engine.model.AnimModel()
        fx.tint_val = nil
        fx.node = engine.rendering.SceneGraphNode(fx.model)
        table.insert(self.fx_pool, fx)
    end
end

function FXManager:PlayingFX()
    return ipairs( self.playing_fx )
end

function FXManager:IsFXPlaying(fx)
    for k,v in ipairs(self.playing_fx) do
        if v == fx then
            return true
        end
    end
    return false
end

function FXManager:OnUpdate(dt)
    local time = self.sim:GetTime()
    for k = #self.playing_fx, 1, -1 do
        local fx = self.playing_fx[k]
        if fx.over_time and fx.over_time < time then
            if fx.looping then
                fx.over_time = time + fx.model:GetDuration()
            elseif fx.queue and #fx.queue > 0 then
                local t = table.remove( fx.queue, 1 )
                local anim_asset, animdat = GetAnim( fx.anim, t.state )
                assert( animdat, t.state )
                fx.over_time = time + animdat.duration
                fx.looping = t.loop == true

                fx.model:SetLooping( t.loop == true )
                fx.model:SetStartTime( time )
                fx.model:SetAnim( anim_asset, animdat.idx )
            else
                local n = #self.playing_fx
                self.playing_fx[k], self.playing_fx[ n ] = self.playing_fx[ n ], self.playing_fx[k]
                table.remove(self.playing_fx)
                self:DisposeOfFX(fx)
            end

        elseif fx.fxdata ~= nil then
            if fx.fxdata.end_colour ~= nil then
                local duration = fx.model:GetDuration()
                local t = ((time- fx.model:GetStartTime())% duration ) / duration
                local sr, sg, sb, sa = 1.0, 1.0, 1.0, 1.0
                if fx.fxdata.colour then
                    sr, sg, sb, sa = HexColour(fx.fxdata.colour)
                end
                local er, eg, eb, ea = HexColour(fx.fxdata.end_colour)
                local nr, ng, nb, na = LerpColour({sr, sg, sb, sa}, {er, eg, eb, ea}, t)
                
                if fx.tint_val then
                    fx.tint_val:SetValue(nr, ng, nb, na)
                else
                    fx.tint_val = engine.rendering.ShaderConstantValue("TINT_MULT", nr, ng, nb, na)
                    fx.model:SetShaderConstant(fx.tint_val)
                end

            elseif fx.fxdata.colour then
                local nr, ng, nb, na = HexColour(fx.fxdata.colour)
                if fx.tint_val then
                    fx.tint_val:SetValue(nr, ng, nb, na)
                else
                    fx.tint_val = engine.rendering.ShaderConstantValue("TINT_MULT", nr, ng, nb, na)
                    fx.model:SetShaderConstant(fx.tint_val)
                end
            end           
        end
    end
end

function FXManager:DisposeOfFX(fx)
    if fx.on_dispose then
        fx.on_dispose( fx )
    end

    fx.on_dispose = nil
    fx.queue = nil
    fx.tint_val = nil

    if #self.fx_pool < 100 then
        fx.node:Orphan()
        fx.model:Clear()
        fx.over_time = nil
        table.insert( self.fx_pool, fx )
    else
        fx.node:SetModel(nil)
        fx.node:Remove()
        table.clear(fx)
    end
end

function FXManager:KillFX(fx)
    for k,v in ipairs(self.playing_fx) do
        if v == fx then

            self:DisposeOfFX(fx)
            local last_idx = #self.playing_fx
            self.playing_fx[k], self.playing_fx[last_idx] = self.playing_fx[last_idx], self.playing_fx[k]            
            table.remove(self.playing_fx)
            return
        end
    end

    print( "KillFX -- missing fx", fx )
end

function FXManager:PlayBillboardFX(anim, state, x, y, z, sx, sy, sz, r, flip, loop, speed)
    
    local fx = self:PlayFXInternal(anim, state, x, y, z, sx or 1, sy or sx or 1, sy or sx or sz or 1, 0, 0, r or 0, flip, loop, speed)
    if fx then
        fx.model:SetLayer( RENDER_LAYER_WORLD ) 
    end
    return fx
end

function FXManager:PlayFXOnGround(anim, state, x, z, sx, sy, r, flip, loop, speed )
    local fx = self:PlayFXInternal(anim, state, x, .01, z, sx*(flip and -1 or 1), sy, 1, -90*DEGREES, r, 0, false, loop, speed)
    if fx then
        fx.model:SetLayer( RENDER_LAYER_GROUNDCOVER ) 
    end
    return fx
end

function FXManager:PlayFXInternal(anim, state, x, y, z, sx,sy,sz, rx,ry,rz, flip, loop, speed )
    sx,sy,sz, rx,ry,rz = sx or 1, sy or 1, sz or 1, rx or 0, ry or 0, rz or 0
    speed = speed or 1.0
    
    local time = self.sim:GetTime()
    local anim_asset, animdat = GetAnim( anim, state )

    if animdat then
        local fx
        if #self.fx_pool > 0 then
            fx = table.remove(self.fx_pool)
            fx.fxdata = nil
        else
            engine.inst:ProfilerPush("NEWFX")
            fx = {}
            fx.model = engine.model.AnimModel()
            fx.tint_val = nil
            fx.node = engine.rendering.SceneGraphNode(fx.model)
            engine.inst:ProfilerPop()
        end
        self.root:AddChild(fx.node)

        fx.anim = anim
        fx.over_time = time + animdat.duration / speed
        fx.looping = loop == true

        fx.model:SetPlaySpeed( speed )
        fx.model:SetPassMask((1 << RENDER_PASS_ALPHA))
        fx.model:SetStartTime(time)
        fx.model:SetLooping(loop == true)
        fx.model:SetBuild(anim_asset)
        fx.model:SetBlendMode(BLEND_MODE.BLENDED)
        fx.model:SetAnim(anim_asset, animdat.idx)
        fx.model:SetBloom( 0 )
        fx.model:SetWallTime(false)
        fx.model:SetCanFlip(false)
        fx.node:SetTransform(x, y, z, sx,sy,sz, rx,ry,rz, flip)

        table.insert(self.playing_fx, fx)
        -- print( "PLAY", anim, state, os.clock(), tostring(flip), x, y, z, sx, sy, sz )
        return fx
    end
end

function FXManager:StopLooping( fx )
    fx.looping = false
    if fx.queue then
        for i, t in ipairs( fx.queue ) do
            t.loop = false
        end
    end
end

function FXManager:QueueFX( fx, state, loop )
    if fx.queue == nil then
        fx.queue = {}
    end
    table.insert( fx.queue, { state = state, loop = loop } )
end
