require "cards/card_picker"
local negotiation_defs = require "negotiation/negotiation_defs"

---------------------------------------------------------------------------
-- Manages AI for playing negotiation.

local IMPATIENCE_START_DELAY = { 3, 2, 2, 1, 0}

local NegotiationBehaviour = class("NegotiationBehaviour")
NegotiationBehaviour.can_reload = true

function NegotiationBehaviour:init(negotiator, data)
    for k, v in pairs(data) do
        assert( rawget( self, k ) == nil, k )
        self[k] = v
    end

    self.engine = negotiator.engine
    self.difficulty = self.engine:GetDifficulty()



    self.negotiator = negotiator
    self.agent = negotiator.agent
    self.owner = self.agent
    self.player_negotiator = self.engine:GetPlayerNegotiator()

    self.hand = self.engine:GetDeck( DECK_TYPE.AI )

    self.prepared_cards = {}

    if self.event_handlers then
        for event_name, fn in pairs( self.event_handlers ) do
            self.engine:ListenForEvent( event_name, self, self.OnNegotiationEvent )
        end
    end

    if data.OnInit then
        data.OnInit( self, self.difficulty )
    end
end

function NegotiationBehaviour:MakePicker()
    local picker = CardEngine.CardPicker( self, self.engine )
    return picker
end

function NegotiationBehaviour:SetPattern( pattern )
    assert(type(pattern) == "function")
    self.current_pattern = pattern

    if type(self.current_pattern) ~= "function" then
        LOGWARN( "Unknown pattern for %s: '%s'", self.negotiator, pattern )
        self.current_pattern = nil
    end
end

function NegotiationBehaviour:AddCard( id )
    local card = Negotiation.Card( id, self.negotiator.agent )
    card.show_dealt = false
    self.engine:DealCard( card, self.hand )
    return card
end

function NegotiationBehaviour:AddArgument( modifier_id )
    local card = self:AddCard( "npc_argument" )

    local modifier_id_plus = string.format( "%s_plus", modifier_id )
    if GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ARGUMENT_PLUS ) and Content.GetNegotiationModifier( modifier_id_plus ) then
        card:SetModifier( modifier_id_plus )
    else
        card:SetModifier( modifier_id )
    end
 
    return card
end

function NegotiationBehaviour:AddCards( t )
    local cards = {}
    for i, card_id in ipairs( card_ids ) do
        local card = self:AddCard( card_id )
        table.insert( cards, card )
    end
    return cards
end

function NegotiationBehaviour:OnNegotiationEvent( eventname, ... )
    if self.event_handlers and self.event_handlers[ eventname ] then
        self.event_handlers[ eventname ]( self, ... )
    end
end

function NegotiationBehaviour:ChooseCard( card )
    if self.negotiator:HasIntent( card ) then
        return false
    end
    
    card:SetRandSeed( math.random( 1, 2^32 ) )

    local ok, reason = self.engine:CanPlayCard( card )
    if not ok then
        return false
    end

    if not self.engine:AssignPrimaryTarget( card ) then
        return false
    end

    table.insert( self.prepared_cards, card )
    return true
end

function NegotiationBehaviour:ChooseCards( cards )
    for i, card in ipairs( cards ) do
        self:ChooseCard( card )
    end
end

function NegotiationBehaviour:ChooseGrowingNumbers( count, k0, scaling, min_damage )

    local DEFAULT_SCALING = 1.75
    local VARIANCE = 0.25
    local MIN_DAMAGE = min_damage or 4

    k0 = k0 or 1
    scaling = scaling or DEFAULT_SCALING

    local adv_scale = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_RESOLVE_DAMAGE )
    if adv_scale then
        scaling = scaling * adv_scale
    end

    local total = k0 + math.floor( scaling * (self.difficulty) ) -- + self.engine:GetTurns() ))
    total = math.random( math.floor(total*(1-VARIANCE)), math.floor(total*(1+VARIANCE)) )
    total = math.max(total, MIN_DAMAGE)

    self:ChooseNumbersFromTotal( count, total )
end

function NegotiationBehaviour:ChooseFlatNumbers( count, k0 )
    self:ChooseNumbersFromTotal( count, k0 + self.difficulty )
end

function NegotiationBehaviour:ChooseNumbers( count, k0, scaling )
    k0 = k0 or 0
    scaling = scaling or 1.0
    local total = k0 + math.floor( scaling * (self.difficulty))-- * self.engine:GetTurns() ))
    self:ChooseNumbersFromTotal( count, total )
end

local function GetDefaultCard( self )
    if self.default_cards == nil then
        self.default_cards = {}
    end

    local card = self.default_cards[ #self.prepared_cards ] 
    if card == nil then
        card = self:AddCard( "default" )
        self.default_cards[ #self.prepared_cards ] = card 
    end

    card.features = nil
    card.min_persuasion = nil
    card.max_persuasion = nil
    return card
end

function NegotiationBehaviour:ChooseComposure( count, min_composure, max_composure )
    for i = 1, count do
        local composure = math.random( min_composure, max_composure )
        if composure > 0 then
            local card = GetDefaultCard( self )
            card.features =
            {
                COMPOSURE = composure,
            }
            if CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.NO_CORE_RESOLVE) then
                card.target_self = TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY
            else
                card.target_self = TARGET_FLAG.CORE | TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY
            end
            card.target_enemy = nil

            self:ChooseCard( card )
        end
    end
end

function NegotiationBehaviour:ChooseNumbersFromTotal( count, total )

    total = math.max( count, total )
    
    for i = 1, count do
        if total <= 0 then
            break
        end

        local card = GetDefaultCard( self )

        local persuasion
        if i == count then
            persuasion = total
        else
            persuasion = math.random( 1, total - (count - i) )
        end
        total = total - persuasion

        card.target_enemy = TARGET_FLAG.CORE | TARGET_FLAG.ARGUMENT | TARGET_FLAG.BOUNTY
        card.target_self = nil
        card.min_persuasion = persuasion
        card.max_persuasion = persuasion

        self:ChooseCard( card )
    end
end

function NegotiationBehaviour:RunBehaviour()
    local turn_count = self.engine:GetTurns()
    if turn_count == self.last_turn or turn_count == 0 then
        return -- NO new intents.
    end

    table.clear( self.prepared_cards )
    self.last_turn = turn_count
    
    if not CheckBits(self.engine.start_params.flags, NEGOTIATION_FLAGS.NO_IMPATIENCE) then
        if CheckBits(self.engine.start_params.flags, NEGOTIATION_FLAGS.WORDSMITH) then
            if self.impatience_delay == nil then
                self.impatience_delay = 4
            end
            self.impatience_delay = self.impatience_delay - 1

            if self.impatience_delay <= 0 then
                local impatience_card = self:AddCard("impatience")
                self:ChooseCard(impatience_card)
                self.impatience_delay = 2
            end
        else
            -- on turn 3, create a 2 turn intent for impatient
            -- when impatient fires, start a new 2 turn impatient
            if self.impatience_delay == nil then
                self.impatience_delay = IMPATIENCE_START_DELAY[self.difficulty] or 3
            end
            self.impatience_delay = self.impatience_delay - 1

            if self.impatience_delay <= 0 then
                local impatience_card = self:AddCard("impatience")
                self:ChooseCard(impatience_card)
                self.impatience_delay = 1
            end
        end
    end

    local ok, result
    if self.current_pattern then
        ok, result = xpcall( self.current_pattern, generic_error, self, turn_count )
    else
        ok = true
    end

    if not ok then
        LOGWARN( result )
    end

    return self.prepared_cards
end

function NegotiationBehaviour:Dialog( str )
    self.engine:BroadcastEvent( negotiation_defs.EVENT.DIALOG, str )
end

function NegotiationBehaviour:Quip( ... )
    self.engine:BroadcastEvent( negotiation_defs.EVENT.QUIP, self.agent, ... )
end
