local fun = require "util/fun"
require "cards/card_engine"
require "negotiation/negotiation_behaviour"
require "negotiation/persuasion_accumulator"
require "negotiation/negotiation_modifier"
local negotiation_defs = require "negotiation/negotiation_defs"

local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local STATUS = negotiation_defs.STATUS
local EVENT = negotiation_defs.EVENT
local RESULT = negotiation_defs.RESULT
local Modifier = Negotiation.Modifier

local MAX_MODIFIER_SLOTS = 13 -- This includes the core argument.
local INFINITE_THRESHOLD = 15

--------------------------------------------------------------------------------
-- A single card.

local Card = class( "Negotiation.Card", CardEngine.Card )
Card.can_reload = true
Card.CARD_FLAGS = CARD_FLAGS

function Card:init( id, owner, userdata )
    assert( owner == nil or is_instance( owner, Agent ))

    local data = Content.GetNegotiationCard( id )
    if data == nil then
        LOGWARN( "Unknown negotiation card: %s. Converting to fast_talk", id )
        data = Content.GetNegotiationCard( "fast_talk" )
        id = "fast_talk"
    end

    for k,v in pairs(data) do 
        self[k] = v
    end  
    
    self.id = id
    self.def = data
    self.owner = owner
    self.userdata = userdata or {}

    self:RollUpgrades()

    if self:GetMaxXP() ~= nil and self.userdata.xp == nil then
        self.userdata.xp = 0
    end

    -- Charges are stored in user-data as they are persistent.
    if (self.userdata.max_charges or self.max_charges) and self.userdata.charges == nil then
        self.userdata.charges = self.userdata.max_charges or self.max_charges
    end

    if data.on_init then
        data.on_init(self)
    end
end

function Card:CanTrash()
    if CheckAnyBits( self.flags, CARD_FLAGS.ITEM ) then
        if is_instance( self.owner, Agent ) then
            return self.owner.negotiator and self.owner.negotiator:HasCard( self )
        end
    end
    return false
end

function Card:IsAutoplay()
    return CheckBits( self.flags, CARD_FLAGS.AUTOPLAY )
end

function Card:IsAttack()
    return self.min_persuasion ~= nil
end

function Card:IsItemCard()
    return CheckBits( self.flags, CARD_FLAGS.ITEM )
end

function Card:Load()
    self.def = Content.GetNegotiationCard( self.id )
    if self.def then
        for k,v in pairs(self.def) do 
            self[k] = v
        end
    end
end

function Card:CanGainXP()
    return self.negotiator == nil or self.negotiator:FindModifier("FATIGUED") == nil
end


function Card:OnAdded( minigame )
    if table.arrayfind( minigame.start_params.cards, self ) then
        self.negotiator = minigame:GetPlayerNegotiator()
    elseif table.arrayfind( minigame.start_params.helper_cards, self ) then
        self.negotiator = minigame:GetPlayerNegotiator()
    elseif table.arrayfind( minigame.start_params.hinder_cards, self ) then
        self.negotiator = minigame:GetOpponentNegotiator()
    else
        self.negotiator = minigame:GetNegotiator( self.owner )
    end
    assert( self.negotiator, tostring(self) )

    self.anti_negotiator = minigame:GetOtherNegotiator( self.negotiator )
    self.engine = minigame

    if self.event_handlers then
        for event_name, fn in pairs( self.event_handlers ) do
            local priority = self.event_priorities and self.event_priorities[ event_name ]
            minigame:ListenForEvent( event_name, self, self.OnEvent, priority )
        end
    end

    -- If this card doesn't exist in the permanent deck, flag it as temporary.
    if self.owner.negotiator and not self.owner.negotiator:FindCardByID( self.id, self.userdata ) then
        self:MakeTemporary()
    end

    if self.OnDealt then
        self:OnDealt( minigame )
    end
end

function Card:OnRemoved( minigame )
    if self.event_handlers then
        minigame:RemoveListener( self )
    end
end

function Card:IsPlayerOwner()
    return self.negotiator and self.negotiator:IsPlayer()
end

function Card:IsUpgraded()
    return CheckBits( self.flags, CARD_FLAGS.UPGRADED ) 
end

function Card:IsHatched()
    return CheckBits( self.flags, CARD_FLAGS.HATCH )
end

function Card:NotifyChanged()
    local clone = self:Clone()
    if self.engine then
        -- Lock in the current damage and modifier string for this card, so that the UI
        -- will have the correct values generated from this point in time.
        local min_persuasion, max_persuasion, details = self.engine:PreviewPersuasion( self )
        clone.CalculatePersuasion = function( self )
            return min_persuasion, max_persuasion, details
        end

        self.engine:BroadcastEvent( EVENT.CARD_CHANGED, self, clone )
    end

    return clone
end

function Card:NotifyTriggeredPre()
    self.engine:BroadcastEvent( EVENT.CARD_TRIGGERED_PRE, self )
end

function Card:NotifyTriggeredPost()
    self.engine:BroadcastEvent( EVENT.CARD_TRIGGERED_POST, self, self.deck, self.deck and self.deck:FindCard( self ))
end

function Card:GetOwner( owner )
    return self.owner
end

function Card:GetNegotiator()
    return self.negotiator
end

function Card:Consume( minigame )
    if self.owner.negotiator and self.owner.negotiator:GetCardCount( self.id ) > 0 then
        self.owner.negotiator:RemoveCardByID( self.id, self.userdata )
    end
end

function Card:NeedsTarget()
    if self.delay and self.delay > 1 then
        return false
    end

    return (self.target_self or 0) ~= 0 or (self.target_enemy or 0) ~= 0
end

function Card:AssignTarget( target )
    assert( target == nil or is_instance( target, Negotiation.Negotiator ) or is_instance( target, Negotiation.Slot ))
    self.target = target
end

function Card:GetTarget()
    return self.target
end

function Card:SetRandSeed( seed )
    self.rand_seed = seed
end

function Card:GetRandSeed()
    return self.rand_seed
end

function Card:IsPlayable()
    return self.deck.engine:CanPlayCard( self )
end

function Card:CalculateComposureText( delta )
    if self.engine then
        local modified_delta = self.engine:CalculateComposure( delta, self.target, self )
        if modified_delta < delta then
            return string.format( "<#PENALTY_CARD_TEXT>%d</>", modified_delta )
        elseif modified_delta > delta then
            return string.format( "<#BONUS_CARD_TEXT>%d</>", modified_delta )       
        end
    end

    return tostring(delta)
end

function Card:CalculatePersuasion()
    local min_persuasion, max_persuasion = self.min_persuasion, self.max_persuasion
    if self.userdata.min_persuasion_bonus then
        min_persuasion = min_persuasion + self.userdata.min_persuasion_bonus
    end
    if self.userdata.max_persuasion_bonus then
        max_persuasion = max_persuasion + self.userdata.max_persuasion_bonus
    end
    return min_persuasion, max_persuasion
end

function Card:Evoke( max_count )
    local deck_type = self.deck and self.deck:GetDeckType()
    if (deck_type == DECK_TYPE.IN_HAND or deck_type == DECK_TYPE.DRAW or deck_type == DECK_TYPE.DISCARDS) and not self.engine:IsQueued( self ) then
        self.evoke_count = (self.evoke_count or 0) + 1

        -- *Exactly* equal, so that we evoke at most once until ResetEvoke().
        if self.evoke_count == max_count then
            self:SetFlags( CARD_FLAGS.FREEBIE )
            self.engine:PlayCard( self )
        end
    end
end

function Card:ResetEvoke()
    self.evoke_count = nil
end

function Card:CanEvoke()
    return self.evoke_count ~= nil
end

function Card:ResetTurnDelay()
    self.delay = self.max_delay or 1
end

function Card:ClearTurnDelay()
    self.delay = nil
end

function Card:GetTurnDelay()
    return self.delay
end

function Card:DeltaTurnDelay( delta )
    self.delay = self.delay + delta
    self.engine:BroadcastEvent( EVENT.INTENT_CHANGED, self, self:Clone() )
end

function Card:CheckTurnDelay()
    if self.delay then
        self.delay = self.delay - 1
        if self.delay <= 0 then
            self.delay = nil
        end

        self.engine:BroadcastEvent( EVENT.INTENT_CHANGED, self, self:Clone() )
    end

    return self.delay == nil
end

function Card:IsPrepared()
    if self.deck and self.deck:GetDeckType() == DECK_TYPE.IN_HAND then
        return self.deck:FindCard( self ) == 1
    end
end

function Card:WasPrepared()
    return self.was_prepared == true
end

function Card:RenderDebugUI( ui, panel )
    ui:Text( "Flags: ".. StringizeBitField( self.flags, negotiation_defs.CARD_FLAG_BITSTRINGS ))

    if self.target_self then
        ui:Text( "Target Self: ".. StringizeBitField( self.target_self, TARGET_FLAG_STRINGS ))
    end

    if self.target_enemy then
        ui:Text( "Target Enemy: ".. StringizeBitField( self.target_enemy, TARGET_FLAG_STRINGS ))
    end
end

function Card:GetDesc( minigame )
	local desc = CardEngine.GenerateCardDesc( self, nil, minigame or self.engine, negotiation_defs )
    return desc 
end

function Card:LookupCardDef( content_id )
    return Content.GetNegotiationCard( content_id )
end

function Card:__tostring()
    return string.format( "Card[%s]", tostring(self.debug_id or self.id))
end

----------------------------------------------------------------------------
-- A single deck of cards.  Could be a draw pile, a discard pile, etc..

local Deck = class("Negotiation.Deck", CardEngine.Deck )
Deck.can_reload = true

function Deck:init( engine, deck_type, is_event_handler )
    Deck._base.init( self, deck_type, is_event_handler )
    self.engine = engine
end

function Deck:RenderDebugUI( ui, panel, minigame )
    if ui:Button( "Insert Card" ) then
        ui:OpenPopup( "INSERT_CARD" )
    end

    if ui:BeginPopup( "INSERT_CARD" ) then
        local card_id = self:RenderCardPopup( ui, panel, minigame:GetPlayer() )
        if card_id then
            local card = Card( card_id, minigame:GetPlayer()  )
            card.show_dealt = false
            print( "Debug dealing:", card )
            minigame:DealCard( card, self )
        end
        ui:EndPopup()
    end

    for i, card in ipairs( self.cards ) do
        ui:PushID( "CARD"..tostring(i) )
        ui:Indent( 20 )
        panel:AppendTable( ui, card )
        ui:Unindent( 20 )
        ui:PopID()
    end
end

function Deck.RenderCardPopup( self, ui, panel, agent, filter_fn )
    local card_id
    
    self.filter_txt = ui:InputText( "Filter", self.filter_txt or "" ) or self.filter_txt
    ui:Separator()

    if ui:Checkbox( "Show All", self.show_all ) then
        self.show_all = not self.show_all
    end

    ui:BeginChild( "Results", 400, 400 )
    for i, carddef in ipairs( Content.GetAllNegotiationCards()) do
        if self.show_all or agent == nil or AgentUtil.CanUseCard( agent, carddef ) then
            if filter_fn == nil or filter_fn( carddef ) then
                if self.filter_txt == nil or carddef.id:find( self.filter_txt, nil, true ) then
                    if ui:Selectable( carddef.id, nil,  ui.DontClosePopups ) then
                        card_id = carddef.id
                    end
                end
            end
        end
    end
    ui:EndChild()

    return card_id
end

------------------------------------------------------------\--------------
-- Negotiation participant.

local Negotiator = class( "Negotiation.Negotiator" )
Negotiator.can_reload = true

function Negotiator:init( agent, engine, resolve, max_resolve )
    self.engine = engine
    self.agent = agent

    self.resolve, self.max_resolve = resolve, max_resolve

    self.all_modifiers = {} -- Array of ALL modifiers

    self.modifier_slots = {} -- Array of Negotiation.Modifier
    self.max_modifiers = MAX_MODIFIER_SLOTS

    self.prepared_cards = {} -- Array of Negotiation.Card

end

function Negotiator:InitBehaviour()
    if self.agent.negotiation_data and self.agent.negotiation_data.behaviour then
        self.behaviour = NegotiationBehaviour( self, self.agent.negotiation_data.behaviour )
    else
        self.behaviour = NegotiationBehaviour( self, table.empty )
    end
end

function Negotiator:OnStartNegotiation( engine )
    for k,aspect in pairs(self.agent.aspects) do
        if aspect.OnStartNegotiation then
            aspect:OnStartNegotiation( engine )
        end
    end

    if not self:FindCoreArgument() then
        self:AddModifier( "HEART", 1 )
    end

    if not self:IsPlayer() and TheGame:GetGameState() and TheGame:GetGameState():CanDoMettle() and not CheckBits(self.engine:GetFlags(), NEGOTIATION_FLAGS.NO_LOOT) then
        local METTLE_CHANCE = { 0, .08, .1, .15, .2 }
        if math.random() < METTLE_CHANCE[math.min(#METTLE_CHANCE, self.engine and self.engine:GetDifficulty() or 1)] then
            local AMT = { 3, 4, 6, 8, 16 }
            local stacks = AMT[math.min(#AMT, self.engine and self.engine:GetDifficulty() or 1)]
            self:AddModifier( "METTLE_BOUNTY", stacks )
        end
    end

end

function Negotiator:GetName()
    return self.agent:GetName() or self.agent:GetShortTitle()
end

function Negotiator:IsPlayer()
    return self.agent:IsPlayer()
end

function Negotiator:GetComposure()
    local heart = self:FindCoreArgument()
    if heart then
        return heart:GetComposure()
    end
    
    return 0
end

function Negotiator:DeltaComposure( delta, source )
    local heart = self:FindCoreArgument()
    if heart then
        local changed = heart:DeltaComposure( delta, source )
        if changed ~= nil then
            return changed
        end
    end

    local t = {}
    for i, modifier in self:Modifiers() do
        if modifier:GetResolve() then
            table.insert( t, modifier )
        end
    end
    local modifier = table.arraypick( t )
    if modifier then
        return modifier:DeltaComposure( delta, source )
    end
end

function Negotiator:GetResolve()
    local heart = self:FindCoreArgument()
    if heart then
        return heart:GetResolve()
    end

    return 0, self.max_resolve
end

function Negotiator:LoseResolve( resolve, source )
    self:ModifyResolve( -resolve, source )
end

function Negotiator:RestoreResolve( resolve, source )
    local heart = self:FindCoreArgument()
    if heart then
        heart:RestoreResolve( resolve, source )
    end
end

function Negotiator:ModifyResolve( delta, source )
    local heart = self:FindCoreArgument()
    if heart then
        heart:ModifyResolve( delta, source )
    end
end

function Negotiator:AttackResolve( delta, source, params )
    local heart = self:FindCoreArgument()
    if heart then
        heart:AttackResolve( delta, source, params )
    end
end

function Negotiator:GetIntents()
    return self.prepared_cards
end

function Negotiator:ClearIntents()
    while #self.prepared_cards > 0 do
        self:DismissIntent( self.prepared_cards[ #self.prepared_cards ])
    end
end

function Negotiator:DismissIntent( card )
    local idx = table.find( self.prepared_cards, card )
    if idx then
        card:ClearTurnDelay()
        table.remove( self.prepared_cards, idx )
        self.engine:BroadcastEvent( EVENT.INTENT_REMOVED, card )
    end
end

function Negotiator:InceptModifier( modifier_id, stacks, source, create )
    if create then
        self:CreateModifier(modifier_id, stacks, source)
    else
        self:AddModifier(modifier_id, stacks, source)
    end
end

function Negotiator:HasIntent( intent )
    for i, card in ipairs( self.prepared_cards ) do
        if card == intent or card.id == intent then
            return card
        end
    end
end

function Negotiator:GetSlot( idx )
    return self.modifier_slots[ idx ]
end

function Negotiator:GetSlotCount()
    return #self.modifier_slots
end

function Negotiator:GetMaxSlots()
    return self.max_modifiers
end

function Negotiator:ModifierSlots()
    return ipairs( self.modifier_slots )
end

function Negotiator:CreateModifier( modifier_id, stacks, source )
    local def, current_modifier

    if is_instance( modifier_id, Negotiation.Modifier ) then
        def = modifier_id.def
    else
        def = Content.GetNegotiationModifier( modifier_id )
        if not def then
            assert( def, "unknown modifier: " .. (modifier_id or "") )
        end
    end

    local is_hidden = def.modifier_type == MODIFIER_TYPE.HIDDEN
    if not is_hidden then
        if #self.modifier_slots >= self.max_modifiers then
            -- Destroy the oldest one.
            local arg = self:FindOldestArgument()
            if arg then
                self:DestroyModifier( arg, source )
            end
        end
    end

    if is_instance( modifier_id, Negotiation.Modifier ) then
        current_modifier = modifier_id
    else
        current_modifier = Negotiation.Modifier( modifier_id, self, stacks, source )
    end
    table.insert( self.all_modifiers, current_modifier )

    if not is_hidden then
        assert( current_modifier.modifier_type ~= MODIFIER_TYPE.HIDDEN )
        table.insert( self.modifier_slots, current_modifier )
    end

    current_modifier:ApplyModifier()

    self.engine:BroadcastEvent( EVENT.MODIFIER_ADDED, current_modifier, source, current_modifier:Clone() )

    return current_modifier
end

function Negotiator:DestroyModifier( modifier, card )
    local i = table.arrayfind( self.all_modifiers, modifier )
    if i == nil then
        LOGWARN( "DestroyModifier called on non-existent modifier %s (card=%s)", tostring(modifier), tostring(card) )
    else
        modifier:UnapplyModifier()

        table.remove( self.all_modifiers, i )
        table.arrayremove( self.modifier_slots, modifier )

        if modifier.OnBounty and card ~= nil and modifier:GetStacks() > 0 then
            -- Bounty comes first so that the slot isn't cleared before any bounty presentation.
            modifier:OnBounty( card )
        end

        self.engine:BroadcastEvent( EVENT.MODIFIER_REMOVED, modifier, card )

        if self:IsPlayer() then
            self.engine:GetOtherNegotiator( self ):PrepareTurn()
        else
            self.engine:GetPlayerNegotiator():PrepareTurn()
        end
    end
end

function Negotiator:HasModifier( modifier )
    if type(modifier) == "string" then
        return self:FindModifier( modifier ) ~= nil
    else
        return table.contains( self.all_modifiers, modifier )
    end
end

function Negotiator:FindCoreArgument()
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.modifier_type == MODIFIER_TYPE.CORE then
            return modifier
        end
    end
end

function Negotiator:FindOldestArgument()
    -- Only searches for non-hidden arguments.
    local oldest_modifier
    for i, modifier in ipairs( self.modifier_slots ) do
        if modifier.modifier_type ~= MODIFIER_TYPE.CORE then
            if oldest_modifier == nil or modifier.uid < oldest_modifier.uid then
                oldest_modifier = modifier
            end
        end
    end
    return oldest_modifier
end

function Negotiator:FindModifier( modifier_id )
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.id == modifier_id then
            return modifier
        end
    end
    return nil
end

function Negotiator:FindModifierByUID( uid )
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.uid == uid then
            return modifier
        end
    end
    return nil
end

function Negotiator:RemoveModifier( modifier, stacks, card )
    if type(modifier) == "string" then
        -- Must take delta. To remove an entire stack, use RemoveModifier / DeltaModifier.
        assert_warning( stacks ~= nil )

        while stacks and stacks > 0 do
            local modifier, change = self:DeltaModifier( modifier, -stacks, card )
            if change < 0 then
                stacks = stacks + change
            else
                break -- No infinite looping pls.
            end
        end

    else
        local delta_stacks = stacks == nil and modifier.stacks or stacks
        self:DeltaModifier( modifier, -delta_stacks, card )
    end
end

function Negotiator:CanAddModifier( modifier_id )
    local modifier = self:FindModifier( modifier_id )
    if modifier == nil then
        local def = Content.GetNegotiationModifier( modifier_id )
        if def and def.modifier_type == MODIFIER_TYPE.HIDDEN then
            -- Hidden modifiers have no slot restrictions.
        end
    end

    return true
end

function Negotiator:AddModifier( modifier, stacks, card )
    return self:DeltaModifier( modifier, stacks or 1, card )
end

function Negotiator:DeltaModifier( modifier, delta, card )
    assert( modifier )
    assert( delta )

    local current_modifier
    if type(modifier) == "string" then
        current_modifier = self:FindModifier( modifier )
    else
        current_modifier = modifier
        assert( is_instance( modifier, Negotiation.Modifier ))
    end

    local applied_delta = 0 -- The actual change in stacks.

    if current_modifier then
       if delta ~= 0 then
            applied_delta = current_modifier:ApplyStacks( delta )
            -- If something hooks Apply, this modifier may not exist anymore.
            if table.contains( self.all_modifiers, current_modifier ) then
                if current_modifier:GetStacks() == 0 and not current_modifier.is_counter then
                    self:DestroyModifier( current_modifier, card )
                else
                    current_modifier:NotifyChanged( delta, card )
                end
            end
        end


    else
        local def = Content.GetNegotiationModifier( modifier )
        if delta > 0 or (delta == 0 and def.is_counter) then
            -- NOTE: Applying zero stacks can create a modifier if the modifier "is_counter"
            current_modifier = self:CreateModifier( modifier, delta, card )
            if current_modifier then
                applied_delta = current_modifier:GetStacks()
            end

        elseif delta < 0 then
            if def.min_stacks and def.min_stacks < 0 then
                -- Applying negative stacks can, strangely, CREATE a modifier if that modifier allows negative stacks.
                current_modifier = self:CreateModifier( modifier, delta, card )
                if current_modifier then
                    applied_delta = current_modifier:GetStacks()
                end
            end
        end
    end

    return current_modifier, applied_delta
end

function Negotiator:GetModifierInstances( modifier_id )
    local count = 0
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.id == modifier_id then
            count = count + 1
        end
    end
    return count
end

function Negotiator:GetModifierStacks( modifier_id )
    local count = 0
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.id == modifier_id then
            count = count + modifier.stacks
        end
    end
    return count
end

function Negotiator:Modifiers()
    return ipairs( self.all_modifiers )
end

function Negotiator:GetModifierCount( modifier_type )
    local count = 0
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.modifier_type == modifier_type then
            count = count + 1
        end
    end
    return count
end

function Negotiator:CanTarget( source, target )
    assert( target )

    if source:GetNegotiator() == self then
        local ok = false

        if is_instance( target, Negotiation.Card ) then
            ok = CheckBits( source.target_enemy or 0, TARGET_FLAG.INTENT )

        elseif is_instance( target, Negotiation.Modifier ) then
            if target:GetNegotiator() == self then
                ok = CheckBits( source.target_self or 0, TARGET_FLAG[ target.modifier_type ] or 0)
            else
                ok = CheckBits( source.target_enemy or 0, TARGET_FLAG[ target.modifier_type ] or 0)
            end
        end

        if not ok then
            return false, CARD_PLAY_REASONS.INVALID_TARGET
        end
    end

    if source:IsAttack() and target:GetResolve() == nil then
        return false, "NO TARGET RESOLVE"
    end

    if target.GetShieldStatus and target:GetShieldStatus() and source:IsAttack() then
        return false, "SHIELDED"
    end


    -- All my modifiers have a say, too.
    for i, modifier in ipairs( self.all_modifiers ) do
        if modifier.CanPlayCard then
            local ok, reason = modifier:CanPlayCard( source, engine, target )
            if not ok then
                return false, reason
            end
        end
    end

    return true
end

function Negotiator:PrepareTurn()
    local new_intents
    if self.behaviour then
        new_intents = self.behaviour:RunBehaviour( self.engine )
    end

    if new_intents then
        for i, card in ipairs( new_intents ) do
            card:ResetTurnDelay()

            table.insert( self.prepared_cards, card )
            self.engine:BroadcastEvent( EVENT.INTENT_ADDED, card )
        end

        self.engine:BroadcastEvent( EVENT.INTENTS_CHANGED )
    end

    -- Check re-targetting for any intents.
    for i, card in ipairs( self.prepared_cards ) do
        self.engine:AssignPrimaryTarget( card )
    end

    for i, modifier in ipairs( self.modifier_slots ) do
        modifier:PrepareTurn()
    end
end

function Negotiator:BeginTurn()

    local modifiers = shallowcopy( self.all_modifiers )
    for i, modifier in ipairs( modifiers ) do
        if table.arrayfind( self.all_modifiers, modifier ) then
            modifier:BeginTurn()
        end
    end
    self.engine:BroadcastEvent( EVENT.BEGIN_TURN, self.engine, self )
end

function Negotiator:EndTurn()

    local modifiers = shallowcopy( self.all_modifiers )
    for i, modifier in ipairs( modifiers ) do
        if modifier.OnEndTurn then
            if table.arrayfind( self.all_modifiers, modifier ) then
                modifier:OnEndTurn( self.engine )
            end
        end
    end

    self.engine:BroadcastEvent( EVENT.END_TURN, self.engine, self )
end

function Negotiator:__tostring()
    return tostring(self.agent) .. "(negotiator)"
end

--------------------------------------------------------------------------
-- Negotiation minigame state.

local MiniGame = class( "Negotiation.MiniGame", CardEngine )
MiniGame.can_reload = true

MiniGame.MAX_HAND_SIZE = 10
MiniGame.EVENT = EVENT
MiniGame.STATUS = STATUS
MiniGame.CARD_FLAGS = CARD_FLAGS
MiniGame.RESULT = RESULT

local DEFAULT_PARAMS =
{
    flags = NEGOTIATION_FLAGS.NONE,
    resolve = BASE_PLAYER_RESOLVE,
    max_resolve = BASE_PLAYER_RESOLVE,
    enemy_resolve = 0,
    enemy_resolve_required = 0,
    hand_size = 5,
    actions_per_turn = 3,
    cards = table.empty,
    modifiers = table.empty,
}

function MiniGame:init( start_params, caravan )
    assert( start_params )
    MiniGame._base.init( self )

    self.decks = {}
    self.draw_deck = self:CreateDeck( DECK_TYPE.DRAW )
    self.resolve_deck = self:CreateDeck( DECK_TYPE.RESOLVE, true )
    self.discard_deck = self:CreateDeck( DECK_TYPE.DISCARDS )
    self.hand_deck = self:CreateDeck( DECK_TYPE.IN_HAND, true )
    self.hand_deck:SetMaxSize( self.MAX_HAND_SIZE )
    self.trash_deck = self:CreateDeck( DECK_TYPE.TRASH )
    self.ai_deck = self:CreateDeck( DECK_TYPE.AI )

    self.start_params = start_params
    self.seed = os.time()

    assert( is_instance( caravan, Caravan ))
    self.caravan = caravan
    self.player = caravan:GetPlayer()

    assert( is_instance( start_params.agent, Agent ))
    self.agent = start_params.agent
    assert_warning( self.agent and not self.agent:IsDead(), "%s is an invalid opponent!", tostring(self.agent) )

    self.turns = 0
    self.next_uid = 100
    self.resolve_count = 0
    self.cards_played = {} -- this turn
    self.all_cards_played = {} -- since battle began
    self.cards_drawn = 0
    self.fatigue_turn = 6

    self.action_count = 0
    self.actions_per_turn = start_params.actions_per_turn
    self.hand_size = start_params.hand_size
    self.resolve_queue = {}

    self.flags = start_params.flags
    
    self.persuasion_amount = Negotiation.PersuasionAccumulator( self, EVENT.CALC_PERSUASION )
    self.cost_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_ACTION_COST )
    self.draw_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_TURN_DRAW )
    self.boolean_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_BOOLEAN )
    self.actions_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_ACTIONS_PER_TURN )
    self.money_cost = CardEngine.ScalarAccumulator( self, EVENT.CALC_MONEY_COST )
    self.resolve_restore = CardEngine.ScalarAccumulator( self, EVENT.CALC_RESTORE_RESOLVE )
    self.target_mod_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_TARGET_MOD )

    self.player_negotiator = Negotiator( self.player, self,
            start_params.resolve, start_params.max_resolve )

    self.opponent_negotiator = Negotiator( self.agent, self,
            start_params.enemy_resolve_required, start_params.enemy_resolve_required )

    self.player_negotiator:InitBehaviour()
    self.opponent_negotiator:InitBehaviour()

    self.min_enemy_resolve_achieved = start_params.enemy_resolve_required

    self.coro = coroutine.create( self.Play )
    self.force_take_cards = {}
end

function MiniGame:CreateDeck( deck_type, ... )
    assert( self.decks[ deck_type ] == nil )
    local deck = Deck( self, deck_type, ... )
    self.decks[ deck_type ] = deck
    return deck
end

function MiniGame:GetDeck( deck_type )
    return self.decks[ deck_type ]
end

function MiniGame:GenerateUID()
    self.next_uid = self.next_uid + 1
    return self.next_uid
end

function MiniGame.GetPersuasionRequired( difficulty )
    local tbl = { 23, 35, 46, 58, 70 }
    local val = tbl[ clamp( difficulty, 1, #tbl ) ]
    local modifiers = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_RESOLVE )
    if modifiers and modifiers[ difficulty ] then
        val = math.round( val + modifiers[ difficulty ] )
    end

    return val
end

local function AddHelperCards( start_params, agent )
    local card_id = agent.support_negotiation 
    if card_id == nil and agent:IsSentient() then
        card_id = "friendly_support"
    end
    
    if card_id then
        local card = Card( card_id, agent )
        if not card.no_duplicate_support or not table.findif( start_params.helper_cards, function( card ) return card.id == card_id end ) then
            table.insert( start_params.helper_cards, card )
        end
    end
end

local function AddHinderCards( start_params, agent, reason )
    local card = Card("heckler", agent )
    if reason then
        card.desc = reason
    end
    table.insert( start_params.hinder_cards, card )
end

local function CheckBystanderList( t, agent )
    for i, v in ipairs( t ) do
        if v == agent then
            return true
        end
    end
    return false
end

function MiniGame.ParamsFromParty( caravan, agent, params, enc )
    params = params or table.empty

    local difficulty = params.difficulty or 1
    local flags = params.flags or 0
    local is_hostile = CheckBits( flags, NEGOTIATION_FLAGS.HOSTILE ) 

    local subject = params.subject or agent
    local acting_against = {}
    local resolve_modifiers = {}

    if params.acting_against then
        for k,v in ipairs(params.acting_against) do
            table.insert_unique(acting_against, v)
        end
    end
    if is_hostile then
        table.insert_unique(acting_against, agent)
    end

    if params.situation_modifiers then
        for _, modifier in ipairs(params.situation_modifiers) do
            table.insert(resolve_modifiers, modifier)
        end
    end

    if agent:GetRelationship() <= RELATIONSHIP.HATED then
        table.insert( resolve_modifiers, { value = 20, text = loc.format( LOC"UI.NEGOTIATION_PANEL.MODIFIER_HATE", agent:GetName() ) } )
    elseif agent:GetRelationship() <= RELATIONSHIP.DISLIKED then
        table.insert( resolve_modifiers, { value = 10, text = loc.format( LOC"UI.NEGOTIATION_PANEL.MODIFIER_DISLIKE", agent:GetName() ) } )
    elseif agent:GetRelationship() >= RELATIONSHIP.LOVED then
        table.insert( resolve_modifiers, { value = -20, text = loc.format( LOC"UI.NEGOTIATION_PANEL.MODIFIER_LOVE", agent:GetName() ) } )
    elseif agent:GetRelationship() >= RELATIONSHIP.LIKED then
        table.insert( resolve_modifiers, { value = -10, text = loc.format( LOC"UI.NEGOTIATION_PANEL.MODIFIER_LIKE", agent:GetName() ) } )
    end

    local start_params = deepcopy( DEFAULT_PARAMS )
    start_params.reward_difficulty = params.reward_difficulty
    start_params.difficulty = math.max( 1, difficulty )
    start_params.reason_txt = params.reason or ""
    start_params.reason_fn = params.reason_fn


    start_params.fight_allies = params.fight_allies
    start_params.cards = params.cards or start_params.cards
    start_params.subject = subject
    start_params.quest = params.quest
    start_params.resolve_modifiers = resolve_modifiers
    start_params.agent = agent

    start_params.on_start_negotiation = params.on_start_negotiation 

    start_params.enemy_resolve_required = params.enemy_resolve_required or MiniGame.GetPersuasionRequired( start_params.difficulty )
    for i, mod in ipairs( resolve_modifiers ) do
        if mod.value then
            start_params.enemy_resolve_required = start_params.enemy_resolve_required + mod.value
        end
    end
    start_params.enemy_resolve_required = math.max(start_params.enemy_resolve_required, 1)

    start_params.flags = SetBits( start_params.flags, flags )

    -- Add cards from party members.
    caravan:CollectNegotiationCards( start_params.cards )
    for i, card in ipairs( start_params.cards ) do
        card.show_dealt = false
    end

    start_params.player_modifiers = {}
    start_params.opponent_modifiers = deepcopy( params.opponent_modifiers ) or {}
    
    local helpers, hinders, suppressed = {}, {}, {}
    
    if params.helpers then
        for k,v in ipairs(params.helpers) do
            local agent
            if type(v) == "string" then
                agent = enc and enc:GetCastAgent(v)
            elseif type(v) == "table" then
                if is_instance(v, Agent) then
                    agent = v
                elseif v.agent then
                    agent = v.agent
                end
            end
            if agent and not agent:IsDead() then
                table.insert(helpers, agent)
            end
        end
    end
    
    if params.hinders then
        for k,v in ipairs(params.hinders) do
            local agent
            
            if type(v) == "string" then
                agent = enc and enc:GetCastAgent(v)
            elseif type(v) == "table" then
                if is_instance(v, Agent) then
                    agent = v
                elseif v.agent then
                    agent = v.agent
                end
            end
            if agent and not agent:IsDead() then
                table.insert(hinders, agent)
            end
        end

    end

    if params.suppressed then
        for k,v in ipairs(params.suppressed) do
            local agent
            if type(v) == "string" then
                agent = enc and enc:GetCastAgent(v)
            elseif type(v) == "table" then
                if is_instance(v, Agent) then
                    agent = v
                elseif v.agent then
                    agent = v.agent
                end
            end
            if agent and not agent:IsDead() then
                table.insert(suppressed, agent)
            end
        end

    end

    start_params.helpers = helpers 
    start_params.hinders = hinders
    start_params.suppressed = suppressed 


    start_params.helper_cards = {}
    start_params.hinder_cards = {}
    

    -- Determine bystander help or hinder.
    local bystanders = params.bystanders or caravan:GetPlayer():GetLocation().agents

    if not CheckBits(start_params.flags, NEGOTIATION_FLAGS.NO_BACKUP) then
        for _, bystander in ipairs( bystanders ) do
            local is_suppressed = false
            for k,v in pairs(start_params.suppressed) do
                if v == bystander then
                    is_suppressed = true
                end
            end

            local has_role = CheckBystanderList( start_params.helpers, bystander ) or CheckBystanderList( start_params.hinders, bystander ) or CheckBystanderList( start_params.suppressed, bystander )

            if not has_role and not is_suppressed and bystander ~= agent and not bystander:IsPlayer() and bystander ~= subject then
                
                local is_acting_against = table.arrayfind(acting_against, bystander)
                
                local result, reason, txt = AgentUtil.GetNegotiationSupport( bystander, subject, is_hostile, CheckBits( start_params.flags, NEGOTIATION_FLAGS.NO_BYSTANDERS) )

                if result == SUPPORT_ACTION.HELP and not is_acting_against then
                    table.insert( start_params.helpers, bystander)
                elseif result == SUPPORT_ACTION.HINDER then
                    table.insert( start_params.hinders, bystander )
                elseif result == SUPPORT_ACTION.SUPPRESSED then
                    table.insert(start_params.suppressed, bystander)
                end
            end
        end
    end

    table.sort(start_params.helpers, function(a,b) return a:GetRenown() > b:GetRenown() end)
    table.sort(start_params.hinders, function(a,b) return a:GetRenown() > b:GetRenown() end)
    table.sort(start_params.suppressed, function(a,b) return a:GetRenown() > b:GetRenown() end)

    for i, v in ipairs( start_params.helpers ) do
        AddHelperCards( start_params, v )
    end
    for i, v in ipairs( start_params.hinders ) do
        AddHinderCards( start_params, v )
    end
    if subject:HasAspect("drunk") then
        table.insert(start_params.hinder_cards, Card("ai_drunk", subject))
    end
         
    table.shuffle( start_params.helper_cards )
    table.shuffle( start_params.hinder_cards )

    start_params.resolve, start_params.max_resolve = caravan:GetResolve()
    start_params.resolve = math.max( 0, start_params.resolve )
    

    return start_params
end

function MiniGame:PrintDump()
    print("-------NEGOTIATION-------")
    print( "Quest:", self.start_params.quest )
    print( "Player Resolve:", loc.format( "{1}: {2}/{3}", self.player, self.start_params.resolve, self.start_params.max_resolve ))
    print( "Resolve:", self.start_params.enemy_resolve_required )
    print( "Difficulty", self.start_params.difficulty )
    print( "Opponent:", tostring(self.agent))
    local grafts = {}
    for i, v in ipairs( self.player.graft_owner:GetAllGrafts() ) do
        if v:GetType() == GRAFT_TYPE.NEGOTIATION or v:GetType() == GRAFT_TYPE.SOCIAL or v:GetType() == GRAFT_TYPE.COIN then
            table.insert( grafts, tostring( v ))
        end
    end
    if #grafts > 0 then
        print( loc.format( "Grafts: {1#listing}", grafts ))
    end
    for i, v in ipairs( self.start_params.helpers ) do
        print( "Helper: ", v )
    end
    for i, v in ipairs( self.start_params.hinders ) do
        print( "Hinders: ", v )
    end
    print("-----------------------")
end

function MiniGame:GetDifficulty()
    return self.start_params.difficulty
end

function MiniGame:GetRewardDifficulty()
    return self.start_params.reward_difficulty
end

function MiniGame:GetPlayer()
    return self.player
end

function MiniGame:GetOpponent()
    return self.agent
end


function MiniGame:GetPlayerNegotiator()
    return self.player_negotiator
end

function MiniGame:GetOpponentNegotiator()
    return self.opponent_negotiator
end

function MiniGame:GetOtherNegotiator( negotiator )
    if negotiator == self.player_negotiator then
        return self.opponent_negotiator
    elseif self.opponent_negotiator == negotiator then
        return self.player_negotiator
    end
end

function MiniGame:GetNegotiator( agent )
    if self.player == agent then
        return self.player_negotiator
    elseif self.agent == agent then
        return self.opponent_negotiator
    end
end

function MiniGame:GetCurrentNegotiator()
    return self.current_negotiator
end

function MiniGame:GetTurns()
    return self.turns
end

function MiniGame:GetFlags()
    return self.flags
end

function MiniGame:Win()
    assert_warning( self.result == nil )
    self.result = RESULT.WIN
    print( "Negotiation Won!" )
end

function MiniGame:Lose()
    assert_warning( self.result == nil )
    self.result = RESULT.LOSE
    print( "Negotiation Lost!" )
end

function MiniGame:GetHandDeck()
    return self.hand_deck
end

function MiniGame:GetDrawDeck()
    return self.draw_deck
end

function MiniGame:GetDiscardDeck()
    return self.discard_deck
end

function MiniGame:GetTrashDeck()
    return self.trash_deck
end

function MiniGame:GetActionCount()
    return self.action_count, self.actions_per_turn
end

function MiniGame:CountCardsDrawn()
    return self.cards_drawn
end

function MiniGame:GetAllCardsPlayed()
    return self.all_cards_played
end

function MiniGame:CountCardsPlayed( fn )
    if fn then
    return table.count_if( self.cards_played, fn )
    else
        return #self.cards_played
    end
end

function MiniGame:InsertCard( card, deck_type )
    if type(card) == "string" then
        card = Battle.Card( card, player )
    end
    local deck = self:GetDeck( deck_type or DECK_TYPE.IN_HAND )
    if not deck:InsertCard( card ) then
        if deck:GetDeckType() == DECK_TYPE.IN_HAND then
            -- Insert into discards instead.
            card:TransferCard( self.discard_deck )
        else
            -- Just trash it, though this can't ever happen atm.
            card:TransferCard( self.trash_deck )
        end
    end

    return card
end

function MiniGame:ModifyActionCount( delta, turn_start )
    local action_count = clamp( self.action_count + delta, 0, 99 )
    if action_count ~= self.action_count then
        delta = action_count - self.action_count
        self.action_count = action_count
        
        self:BroadcastEvent( EVENT.DELTA_ACTION_COUNT, self.action_count, self.actions_per_turn, delta, turn_start == true )
    end
end


function MiniGame:GetAllPlayerCards(filter)
    local ret = {}
    for i, card in self.hand_deck:Cards() do
        if filter == nil or filter(card) then
            table.insert(ret, card)
        end
    end
    for i, card in self.discard_deck:Cards() do
        if filter == nil or filter(card) then
            table.insert(ret, card)
        end
    end
    for i, card in self.draw_deck:Cards() do
        if filter == nil or filter(card) then
            table.insert(ret, card)
        end
    end
    return ret
end


function MiniGame:CanPlayAny()
    for i, card in self.hand_deck:Cards() do
        if self:CanPlayCard( card ) then
            return true
        end
    end
    return false
end

function MiniGame:CanTarget( source, target )
    if is_instance( target, Negotiation.Modifier ) then
        if target.negotiator and not target.negotiator:HasModifier( target ) then
            return false
        end
    end

    if source.CanTarget then
        local ok, reason = source:CanTarget( target )
        if not ok then
            return false, reason or CARD_PLAY_REASONS.INVALID_TARGET
        end
    end
    
    local ok, reason = self.player_negotiator:CanTarget( source, target )
    if not ok then
        return false, reason or CARD_PLAY_REASONS.INVALID_TARGET
    end

    local ok, reason = self.opponent_negotiator:CanTarget( source, target )
    if not ok then
        return false, reason or CARD_PLAY_REASONS.INVALID_TARGET
    end

    return true
end

function MiniGame:CanPlayCard( card, target )

    if card.negotiator == nil then
        return false

    elseif card.negotiator == self.opponent_negotiator then
        if not (self.ai_deck:HasCard( card )) then
            return false --, "Not in hand"
        end

        if card:GetTurnDelay() then
            return false, "Turn countdown"
        end

    elseif card.negotiator == self.player_negotiator then
        if card:IsFlagged( CARD_FLAGS.BYSTANDER ) then
            -- These are played from your draw.
        elseif not self.hand_deck:HasCard( card ) and not card:IsFlaggedAny( CARD_FLAGS.SPECIAL | CARD_FLAGS.FLOURISH ) then
            return false --, "Not in hand"
        end
    end

    local charges = card:GetCharges()
    if charges and charges <= 0 then
        return false, CARD_PLAY_REASONS.NO_CHARGE
    end

    if self.hand_deck:HasCard( card ) then
        if not CheckAnyBits( card.flags, CARD_FLAGS.VARIABLE_COST | CARD_FLAGS.FREEBIE ) then
            local cost = self:CalculateActionCost( card )
            if cost > self.action_count then
                return false, CARD_PLAY_REASONS.NO_ACTIONS
            end
        end
    end

    local has_money = self:GetMoney()
    local cost = self:CalculateMoneyCost( card, true )
    if cost > 0 then
        if has_money < cost then
            return false, CARD_PLAY_REASONS.REQ_MONEY
        end
    end

    if CheckAnyBits( card.flags, CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.AUTOPLAY ) then
        return false, CARD_PLAY_REASONS.UNPLAYABLE
    end

    local negotiator = card:GetNegotiator()

    if target then
        if is_instance( target, Negotiation.Card ) then
            return false, CARD_PLAY_REASONS.INVALID_TARGET
        end

        if card.CanPlayCard then
            local ok, reason = card:CanPlayCard( card, self, target )
            if not ok then
                return false, reason
            end
        end

        local ok, reason = self:CanTarget( card, target )
        if not ok then
            return false, reason
        end

    -- Otherwise, playable if this card requires no targets, or is playable against SOME target
    else
        if card.CanPlayCard then
            local ok, reason = card:CanPlayCard( card, self, nil )
            if not ok then
                return false, reason
            end
        end

        if card:NeedsTarget() then
            local t = ObtainWorkTable()
            self:CollectAllTargets( card, t, 1 )
            local target_count = #t
            ReleaseWorkTable( t )

            if target_count == 0 then
                return false, CARD_PLAY_REASONS.NO_VALID_TARGETS
            end
        end
    end

    return true
end


function MiniGame:CalculateTurnDraw()
    local hand_size, details = self.draw_accumulator:PreviewValue( self.hand_size )
    hand_size = math.max( 1, hand_size )
    return hand_size, details
end

function MiniGame:ModifyTurnDraw( delta )
    self.hand_size = clamp( self.hand_size + delta, 0, 10 )
end

function MiniGame:GetMoney()
    return self.caravan:GetMoney()
end

function MiniGame:ModifyMoney( delta, source )
    if delta > 0 then
        self.caravan:AddMoney( delta )

    elseif delta < 0 then
        delta = math.min( -delta, self.caravan:GetMoney() )
        self.caravan:PayMoney( delta, self.agent )
    end

    self:BroadcastEvent( EVENT.DELTA_MONEY, delta, source )
end

function MiniGame:OnStartNegotiation()
    if self.start_params.on_start_negotiation then
        self.start_params.on_start_negotiation(self)
    end

    if self.start_params.modifiers then
        for modifier_id, count in pairs( self.start_params.player_modifiers ) do
            self.player_negotiator:AddModifier( modifier_id, count )
        end
        for modifier_id, count in pairs( self.start_params.opponent_modifiers ) do
            self.opponent_negotiator:AddModifier( modifier_id, count )
        end
    end

    self.player_negotiator:OnStartNegotiation( self )
    self.opponent_negotiator:OnStartNegotiation( self )

    if CheckBits( self.start_params.flags, NEGOTIATION_FLAGS.ALLY_GRAB ) then
        assert_warning(self.start_params.fight_allies and #self.start_params.fight_allies > 0, "no allies specified")
        local card = Negotiation.Card( "take_supporters", self:GetPlayerNegotiator().agent )
        card.show_dealt = true
        card:TransferCard(self:GetDrawDeck())

        for k, ally in ipairs(self.start_params.fight_allies or table.empty) do
            if k > 1 then
                local mod = Negotiation.Modifier( "FIGHT_ALLY", self:GetOpponentNegotiator() )
                mod:SetAgent(ally)
                self:GetOpponentNegotiator():CreateModifier( mod )
            end
        end
    end
    
    if CheckBits( self.start_params.flags, NEGOTIATION_FLAGS.ALLY_SCARE ) then
        assert_warning(self.start_params.fight_allies and #self.start_params.fight_allies > 0, "no allies specified")
        
        for k, ally in ipairs(self.start_params.fight_allies or table.empty) do
            local mod = Negotiation.Modifier( "FIGHT_ALLY_SCARE", self:GetOpponentNegotiator() )
            mod:SetAgent(ally)
            self:GetOpponentNegotiator():CreateModifier( mod )
        end

        local core = self.opponent_negotiator:FindCoreArgument()
        core:SetShieldStatus( true, LOC"UI.NEGOTIATION_ALLY_SCARE_SHIELD" )
    end



    if CheckBits( self.start_params.flags, NEGOTIATION_FLAGS.INTIMIDATION ) then
        local n = math.max(1, math.round( self.player_negotiator.agent.negotiator:GetCardCount() / 5 ))
        for k = 1, n do
            local card = Negotiation.Card( "intimidate", self.player_negotiator.agent )
            card:TransferCard(self:GetDrawDeck())
        end
    end 
    if CheckBits( self.start_params.flags, NEGOTIATION_FLAGS.PROVOCATION ) then
        local n = math.max(1, math.round( self.player_negotiator.agent.negotiator:GetCardCount() / 5 ))
        for k = 1, n do
            local card = Negotiation.Card( "insult", self.player_negotiator.agent )
            card:TransferCard(self:GetDrawDeck())
        end
    end 

    -------------------------------------------------
    -- Deal only after starting modifiers are applied so they can hook initial hands.

    self:DealCards( self.start_params.cards )

    -- DEAL_CARDS handler needs to query current turns
    self:YieldPlay( STATUS.SYNC )

    self.draw_deck:Shuffle()

    -- Play bystander cards.
    for i, card in ipairs( self.start_params.helper_cards ) do
        -- So this is kinda weird, but we need the card to be registered to a deck.
        self.trash_deck:InsertCard( card )
        self:PlayCard( card )
    end
    for i, card in ipairs( self.start_params.hinder_cards ) do
        -- So this is kinda weird, but we need the card to be registered to a deck.
        self.trash_deck:InsertCard( card )
        self:PlayCard( card )
    end

    self:BroadcastEvent( EVENT.BEGIN_NEGOTIATION, self )
end

function MiniGame:OnEndNegotiation()
    self:BroadcastEvent( EVENT.END_NEGOTIATION, self )
end

function MiniGame:Play()

    self:OnStartNegotiation()

     -- ccc: can this be removed?  everything should use PlayCard.
    assert_warning( self.resolve_deck:CountCards() == 0, tostr(self.resolve_deck.cards) )

    while not self:CheckGameOver() do

        self.turns = self.turns + 1
        local actions_per_turn = self:CalculateActionsPerTurn()
        self:ModifyActionCount( actions_per_turn - self.action_count, true )

        self:RunBeginPlayerTurn()
        
        self.opponent_negotiator:PrepareTurn()

        while not self:CheckGameOver() do

            -- Wait for user choice...
            local card = self:YieldPlay( STATUS.PLAY_CARD )
            assert( card == nil or is_instance( card, Card ))

            if not card then
                -- End turn.
                break

            else
                self:PayCosts( card )
                self:PlayCard( card )
            end

            -- Resolve cards played.
            self:RunResolvePhase()
        end

        if not self:CheckGameOver() then
            self:RunEndPlayerTurn()
        end

        if not self:CheckGameOver() then
            self:RunOpponentTurn()
        end
    end

    self:OnEndNegotiation()
end

function MiniGame:RunOpponentTurn()
    self.current_negotiator = self.opponent_negotiator
    self:BroadcastEvent( EVENT.BEGIN_ENEMY_TURN, self )

    self.opponent_negotiator:BeginTurn()

    -- Playing intents can potentially remove intents through side effects, so we must copy the list and
    -- ensure each entry is still valid before resolving it.
    local intents = shallowcopy( self.opponent_negotiator:GetIntents())
    table.shuffle( intents )

    for i, card in ipairs( intents ) do
        if card:CheckTurnDelay() then
            local idx = table.arrayfind( self.opponent_negotiator:GetIntents(), card )
            if idx then
                table.remove( self.opponent_negotiator:GetIntents(), idx )
                if self:AssignPrimaryTarget( card ) then
                    self:PlayCard( card )
                end

                self:BroadcastEvent( EVENT.INTENT_REMOVED, card )
                self:YieldPlay( STATUS.SYNC )
            end
        end
    end

    self.opponent_negotiator:EndTurn()

    self:BroadcastEvent( EVENT.END_ENEMY_TURN, self )

    self:YieldPlay( STATUS.SYNC )
end

function MiniGame:RunBeginPlayerTurn()
    self.current_negotiator = self.player_negotiator
    -- Player draw.
    self:BroadcastEvent( EVENT.BEGIN_PLAYER_TURN, self )

    self.player_negotiator:BeginTurn()

    if self:GetTurns() == self:GetFatiguedTurn() then
        self.player_negotiator:AddModifier("FATIGUED")
    end

    do
        if self.turns == 1 then
            -- Ambush cards go directly into hand on turn 1 only.
            local function IsAmbush( card )
                return card:IsFlagged( CARD_FLAGS.AMBUSH )
            end
            self.draw_deck:TransferCardsIf( self.hand_deck, IsAmbush )
        end

        local hand_size = self:CalculateTurnDraw()
        self:DrawCards( hand_size, true )
    end
    self:BroadcastEvent( EVENT.HAND_DRAWN, self )

    self.player_negotiator:PrepareTurn()
end

function MiniGame:CollectRandomTargets( targets, candidates, n, source )
    local seed = source.GetRandSeed and source:GetRandSeed() or self.resolve_count
    local rand = engine.Random( self.seed, seed )
    if #candidates == 0 then
        return
    end
    local idxs = table.range( 1, #candidates )
    for i = #idxs, 1, -1 do
        if table.arrayfind( targets, candidates[i] ) then
            table.remove( idxs, i )

        elseif source and not self:CanTarget( source, candidates[i] ) then
            table.remove( idxs, i )
        end
    end

    for i = 1, math.min( n, #idxs ) do
        local idx = table.remove( idxs, rand:Random( 1, #idxs ))
        table.insert(targets, candidates[ idx ] )
    end
end

function MiniGame:CollectAlliedTargets( primary_target, targets, fn, card )
    targets = targets or {}

    local negotiator
    if is_instance( primary_target, Negotiation.Negotiator ) then
        negotiator = primary_target
    elseif is_instance( primary_target, Negotiation.Modifier ) then
        negotiator = primary_target.negotiator
    end

    for i, modifier in negotiator:ModifierSlots() do
        if fn == nil or fn( modifier ) then
            local target
            if card == nil or self:CanTarget( card, modifier ) then
                target = modifier
            end
            if target and target.resolve then
                table.insert( targets, target )
            end
        end
    end

    return targets
end

function MiniGame:CollectAllTargets( source, t, max_count )
    local negotiator = source:GetNegotiator()
    local opponent = self:GetOtherNegotiator( negotiator )
    max_count = max_count or math.huge
    t = t or {}

    for _, modifier in negotiator:ModifierSlots() do
        if self:CanTarget( source, modifier ) then
            table.insert( t, modifier )
            if #t >= max_count then
                return t
            end
        end
    end

    for _, modifier in opponent:ModifierSlots() do
        if self:CanTarget( source, modifier ) then
            table.insert( t, modifier )
            if #t >= max_count then
                return t
            end
        end
    end

    for _, intent in ipairs( opponent:GetIntents() ) do
        if self:CanTarget( source, intent ) then
            table.insert( t, intent )
            if #t >= max_count then
                return t
            end
        end
    end

    return t
end


function MiniGame:CollectTargets( source )
    if not source:NeedsTarget() then
        return table.empty
    end

    local primary_target = source:GetTarget()

    if primary_target == nil or (primary_target:GetResolve() or 1) <= 0 then
        return table.empty
        -- self:CollectAllTargets( source, targets )
        -- print( "Finding targets:", source, tostr(targets))
        -- primary_target = table.arraypick( targets )
        -- table.clear( targets )
    end

    local targets = {}
    local target_mod = self.target_mod_accumulator:CalculateValue( source.target_mod or TARGET_MOD.SINGLE, source, primary_target )

    if primary_target == nil then
        -- Nothing to mod.

    elseif target_mod == TARGET_MOD.SINGLE then
        table.insert( targets, primary_target )

    elseif target_mod == TARGET_MOD.RANDOM1 then
        local candidates = self:CollectAlliedTargets( primary_target )
        self:CollectRandomTargets( targets, candidates, 1, source )

    elseif target_mod == TARGET_MOD.RANDOM2 then
        local candidates = self:CollectAlliedTargets( primary_target )
        self:CollectRandomTargets( targets, candidates, 2, source )

    elseif target_mod == TARGET_MOD.RANDOM3 then
        local candidates = self:CollectAlliedTargets( primary_target )
        self:CollectRandomTargets( targets, candidates, 3, source )

    elseif target_mod == TARGET_MOD.TEAM then
        self:CollectAlliedTargets( primary_target, targets, nil, source )

    elseif target_mod == TARGET_MOD.ALL then
        self:CollectAlliedTargets( self.player_negotiator, targets, nil, source )
        self:CollectAlliedTargets( self.opponent_negotiator, targets, nil, source )

    elseif target_mod == TARGET_MOD.CUSTOM then
        source:target_fn( self, primary_target, targets, source )
    end

    return targets
end

function MiniGame:CollectPrimaryTarget( source, target )
    if is_instance( target, Negotiation.Negotiator ) then
        target = target:FindCoreArgument()
    end

    if not source:NeedsTarget() then
        return target
    end

    local t = {}
    self:CollectAllTargets( source, t )

    local weighted_targets = {}
    for i, v in ipairs( t ) do
        if source.EvaluateTargetWeight then
            weighted_targets[ v ] = source:EvaluateTargetWeight( v, t )
        else
            weighted_targets[ v ] = 1
        end
    end

    -- Verify the optional target argument; it takes precedence.
    if target and (not self:CanTarget( source, target ) or (weighted_targets[ target ] or 0) <= 0) then
        print( "Retargetting:", source, "from", target )
        target = nil
    end

    if target == nil then
        target = weightedpick( weighted_targets )
    end

    return target
end


function MiniGame:AssignPrimaryTarget( source, target )
    target = self:CollectPrimaryTarget( source, target or source:GetTarget() )
    source:AssignTarget( target )
    return not source:NeedsTarget() or target ~= nil
end

function MiniGame:IsBeingTargettedBy( target, negotiator, sources )
    for i, intent in ipairs( negotiator:GetIntents() ) do
        if (intent:GetTurnDelay() or 0) <= 1 then
            local targets = self:CollectTargets( intent )
            if table.contains( targets, target ) then
                if sources then
                    table.insert( sources, intent )
                else
                    return true
                end
            end
        end
    end
    for i, modifier in negotiator:ModifierSlots() do
        local targets = self:CollectTargets( modifier )
        if table.contains( targets, target ) then
            if sources then
                table.insert( sources, modifier )
            else
                return true
            end
        end
    end

    return sources and #sources > 0
end

function MiniGame:IsBeingTargetted( target, sources )
    local targetted = false
    if target.anti_negotiator then
        if self:IsBeingTargettedBy( target, target.anti_negotiator, sources ) then
            targetted = true
        end
    end

    if target.negotiator then
        if self:IsBeingTargettedBy( target, target.negotiator, sources ) then
            targetted = true
        end
    end

    return targetted
end

function MiniGame:IsResolving( card )
    return card.deck == self.resolve_deck
end

function MiniGame:RunResolvePhase()

    if self:CheckGameOver() then
        return
    end

    self.resolving = true

    while #self.resolve_queue > 0 do
        local t = table.remove( self.resolve_queue, 1 )
        local resolve_card = t.card
        self:AssignPrimaryTarget( resolve_card, t.target )

        resolve_card.was_prepared = t.prepared

        print( "Resolving: ", resolve_card, resolve_card:GetTarget(), t.delay )

        resolve_card:TransferCard( self.resolve_deck )

        self:BroadcastEvent( EVENT.START_RESOLVE, self, resolve_card )

        self:AssignPrimaryTarget( resolve_card )
        local targets = self:CollectTargets( resolve_card )

        if resolve_card.OnPreResolve then
            resolve_card:OnPreResolve( self, targets )
        end

        self:BroadcastEvent( EVENT.PRE_RESOLVE, self, resolve_card, targets, t.delay )

        for i, target in ipairs( targets ) do
            if resolve_card.min_persuasion and resolve_card.max_persuasion then
                self.persuasion_amount:ApplyPersuasion( resolve_card, target )
            end
        end

        if resolve_card.features then
            for feature_name, stacks in pairs( resolve_card.features ) do
                self:ApplyFeature( feature_name, stacks, resolve_card, targets )
            end
        end

        if resolve_card.OnPostResolve then
            resolve_card:OnPostResolve( self, targets )
        end

        self:BroadcastEvent( EVENT.POST_RESOLVE, self, resolve_card, targets )

        if resolve_card.owner and resolve_card.owner:IsPlayer() then
            table.insert( self.cards_played, resolve_card )
            table.insert( self.all_cards_played, resolve_card )
            if #self.cards_played >= INFINITE_THRESHOLD and not self.reported_infinite then
                self.reported_infinite = true
                local player_state = TheGame:GetGameState():GetPlayerState()
                if player_state then
                    local json_t = {
                        PLAYER_DATA = player_state,
                    }
                    SendMetricsData("NEGOTIATION_INFINITE_DETECTED", json_t )
                end
            end
        end

        self.resolve_count = self.resolve_count + 1

        resolve_card:ConsumeCharge()
        if resolve_card:IsFlagged( CARD_FLAGS.ITEM ) and resolve_card:IsSpent() then
            resolve_card:SetFlags( CARD_FLAGS.CONSUME )
        end

        if resolve_card.owner and resolve_card.owner:IsPlayer() then
            StatUtil.PlayedCard(resolve_card.id)
        end

        resolve_card:AddXP(1)

        if resolve_card:IsFlagged( CARD_FLAGS.HATCH ) and resolve_card:UpgradeReady() then
            if resolve_card.hatch_fn then 
                if not resolve_card.hatched then
                    resolve_card:hatch_fn(self)
                    resolve_card.hatched = true
                else
                    resolve_card:TransferCard( self.trash_deck )
                    resolve_card:Consume()
                end
            elseif resolve_card.userdata.hatch_id then
                self:HatchCard( resolve_card, resolve_card.userdata.hatch_id )
            end
        end

        resolve_card:AssignTarget( nil )
        
        resolve_card.was_prepared = nil
        resolve_card.flags = ClearBits( resolve_card.flags, CARD_FLAGS.FREEBIE )
        
        -- Where does this resolved card go now?
        if CheckBits( resolve_card.flags, CARD_FLAGS.BYSTANDER ) then
            -- These are special implementation-only cards: they're so special they are removed entirely.
            resolve_card:RemoveCard()

        elseif CheckBits( resolve_card.flags, CARD_FLAGS.ITEM ) and resolve_card.userdata.charges == 0 then
            --
            resolve_card:TransferCard( self.trash_deck )
            resolve_card:Consume()

        elseif CheckBits( resolve_card.flags, CARD_FLAGS.CONSUME ) then
            resolve_card:TransferCard( self.trash_deck )
            resolve_card:Consume()

        -- Bystander cards poof, but don't trigger the EXPEND mechanic.
        elseif CheckAnyBits( resolve_card.flags, CARD_FLAGS.BYSTANDER ) then
            resolve_card:TransferCard( self.trash_deck )

        -- Expended: go directly to the trash.
        elseif CheckAnyBits( resolve_card.flags, CARD_FLAGS.EXPEND ) or CheckAnyBits( resolve_card.flags, CARD_FLAGS.FLOURISH ) then
            self:ExpendCard( resolve_card )

        elseif resolve_card.deck == self.resolve_deck then
            -- If the card is no longer in the resolve deck, then something moved it
            -- somewhere else and we should not attempt any post-resolve transfer.
            if not resolve_card:GetNegotiator():IsPlayer() then
                resolve_card:TransferCard( self.ai_deck )

            elseif CheckBits( resolve_card.flags, CARD_FLAGS.MULTISHOT ) then
                -- Multi-shots get re-inserted into your hand.
                resolve_card:TransferCard( self.hand_deck )
            else
                -- By default, most things go into the discards after being resolved.
                resolve_card:TransferCard( self.discard_deck )
            end
        end

        assert( self.resolve_deck:IsEmpty() )

        if resolve_card.negotiator:IsPlayer() then
            self.opponent_negotiator:PrepareTurn()
        end

        self:BroadcastEvent( EVENT.END_RESOLVE, self, resolve_card )

        self:CheckGameOver()

        if self.result ~= nil then
            table.clear( self.resolve_queue )
            break
        end
    end

    self.resolving = false
end

function MiniGame:RunEndPlayerTurn()

    -- Resolve all AUTOPLAY cards remaining in hand.
    self:PlayCards( self.hand_deck, Card.IsAutoplay )

    self.player_negotiator:EndTurn()

    self:BroadcastEvent( EVENT.END_PLAYER_TURN, self )

    local wt = ObtainWorkTable()
    table.arrayadd( wt, self.hand_deck.cards )
    for k,card in ipairs(wt) do
        if card:IsFlagged(CARD_FLAGS.BURNOUT) or card:IsFlagged(CARD_FLAGS.FLOURISH) then
            self:ExpendCard(card)
        end
    end
    ReleaseWorkTable(wt)

    -- Lastly, discard all cards in hand.
    -- I changed the order of end turn/ transfer cards for sound reasons.
    -- If this creates bugs, we can talk about how to solve the sound issue in another way.
    -- (wanted to not play discard sound after end turn had been pressed)

    local CheckIfSticky = function( card )
        return not card:IsFlagged( CARD_FLAGS.STICKY )
    end

    -- Lastly, discard all cards in hand.
    if not self.no_discard then
        self.hand_deck:TransferCardsIf( self.discard_deck, CheckIfSticky)
    end

    table.clear( self.cards_played )
    self.cards_drawn = 0
end

function MiniGame:PayCosts( card )
    if card.owner:IsPlayer() then
        local cost = self:CalculateActionCost( card )
        self:ModifyActionCount( -cost )

        local cost = self:CalculateMoneyCost( card )
        if cost > 0 then
            self:ModifyMoney( -cost, card )
        end

        if card:IsFlagged( CARD_FLAGS.FLOURISH ) then
            local flourish_graft = self:GetPlayer().graft_owner:GetGraft("flourish_tracker")
            if flourish_graft then
                flourish_graft:GetDef().SetNegotiationFlourish( flourish_graft, self, 0 )
            end
        end
    end
end

function MiniGame:IsQueued( card )
    for i, t in ipairs( self.resolve_queue ) do
        if t.card == card then
            return true
        end
    end
    return false
end

function MiniGame:Delay( duration )
    self:BroadcastEvent( EVENT.CUSTOM, function( panel ) panel:Delay(duration) end )
end

function MiniGame:PlayCard( card, target, delay )
    if self.result ~= nil then
        return
    end
    if card:IsFlagged(CARD_FLAGS.UNPLAYABLE) and not card:CanEvoke() then
        return
    end
    local t = { card = card, target = target or card:GetTarget(), delay = delay, prepared = card:IsPrepared() or card:WasPrepared() }
    table.insert( self.resolve_queue, t )

    if not self.resolving then
        self:RunResolvePhase()
    end
end

function MiniGame:HatchCard( card, card_id )
    card:TransferCard( self.trash_deck )
    card:Consume()

    local new_card = Negotiation.Card( card_id, card.owner )
    new_card.owner.negotiator:AddCard(new_card)

    local negotiation_card = new_card:Clone()
    negotiation_card.owner = card.owner
    self:DealCard(negotiation_card, self:GetHandDeck())
end

function MiniGame:CheckGameOver()
    if self.result == nil then
        local opponent_resolve = self.opponent_negotiator:GetResolve()
        local player_resolve = self.player_negotiator:GetResolve()

        if opponent_resolve and opponent_resolve <= 0 then
            self:Win()
        elseif player_resolve and player_resolve <= 0 then
            self:Lose()
        end
    end

    return self.result ~= nil
end

function MiniGame:GetResult()
    return self.result
end

function MiniGame:AllModifiers()
    local modifiers = {}
    for i, modifier in self.player_negotiator:Modifiers() do
        table.insert( modifiers, modifier )
    end
    for i, modifier in self.opponent_negotiator:Modifiers() do
        table.insert( modifiers, modifier )
    end
    return ipairs( modifiers )
end

function MiniGame:FindModifierByUID( uid )
    return self.player_negotiator:FindModifierByUID( uid ) or
            self.opponent_negotiator:FindModifierByUID( uid )
end

function MiniGame:CalculateActionsPerTurn()
    local actions = self.actions_accumulator:CalculateValue( self.actions_per_turn )
    actions = math.max( 0, actions )
    return actions
end

function MiniGame:CalculateActionCost( card )
    if CheckBits( card.flags, CARD_FLAGS.FREEBIE ) then
        return 0
    elseif card.cost == nil then
        return 0 -- FIXME: possibly return nil
    else
        local cost, details = self.cost_accumulator:PreviewValue( card.cost, card )
        cost = math.max( 0, cost )
        return cost, details
    end
end

function MiniGame:CalculateMoneyCost( card, preview )
    local cost, details
    if preview then
        cost, details = self.money_cost:PreviewValue( card.money_cost or 0, card )
    else
        cost = self.money_cost:CalculateValue( card.money_cost or 0, card )
    end

    cost = math.max( 0, cost )
    return cost, details
end

function MiniGame:CalculateRestoreResolve( resolve, source )
    local resolve = self.resolve_restore:CalculateValue( resolve, source )
    return math.max( 0, resolve )
end

function MiniGame:CalculateBoolean( key, source )
    local value, details = self.boolean_accumulator:CalculateValue( false, key, source )
    assert( type(value) == "boolean", tostring(value))
    return value
end

function MiniGame:CalculateComposure( delta, target, card )
    if self.delta_composure_accumulator == nil then
        self.delta_composure_accumulator = CardEngine.ScalarAccumulator( self, EVENT.CALC_DELTA_COMPOSURE )
    end
    local stacks = self.delta_composure_accumulator:CalculateValue( delta, target, card )
    return stacks
end

function MiniGame:PreviewPersuasion( source, no_rand )
    return self.persuasion_amount:PreviewPersuasion( source, no_rand )
end

function MiniGame:ApplyPersuasion( source, target, min_persuasion, max_persuasion )
    if target == nil then
        target = self:CollectPrimaryTarget( source, source:GetTarget() )
    end
    self.persuasion_amount:ApplyPersuasion( source, target, min_persuasion, max_persuasion )
end

function MiniGame:ApplyFeature( feature_name, stacks, resolve_card, targets )
    local feature = Content.GetNegotiationCardFeature( feature_name )
    if feature == nil then
        -- TODO: validate load-time.
        LOGWARN( "Unknown feature on card: %s.%s", tostring(resolve_card), feature_name )
    end
    if feature and feature.apply then
        return feature:apply( self, resolve_card, targets, stacks )
    end
end

function MiniGame:InceptCard( card, source )
    self:InceptCards( { card }, source )
end

function MiniGame:InceptCards( cards, source )
    for i, card in ipairs( cards ) do
        card.show_dealt = false
    end

    self:BroadcastEvent( EVENT.CARDS_INCEPTED, cards, source )
    self:DealCards( cards, self.draw_deck )
end

function MiniGame:ImproviseCards( cards, count, txt, replacement_id, can_skip, source )
    assert( cards )

    count = count or 1
    local min_count = can_skip and 0 or count


    if #cards < count and replacement_id then
        local bonus = count - #cards
        for i=1, bonus do
            table.insert(cards, Negotiation.Card(replacement_id, self:GetPlayer()))            
        end

    elseif #cards > count then
        AUDIO:PlayEvent("event:/sfx/battle/cards/neg/improvised_card_presented")
        cards = self:ChooseCardsFromTable( cards, min_count, count, nil, txt, source )
    end

    for i, card in ipairs( cards ) do
        if card.deck == nil then
            -- Because you chose it, seeing the CARD_DEALT viz is annoying.
            card.show_dealt = false
            self:DealCard( card, self:GetHandDeck() )
        else
            card.deck:TransferCard( card, self:GetHandDeck() )
        end
        if card.deck ~= self:GetHandDeck() then
            if card.deck then
                card:RemoveCard()
            end
            card:TransferCard( self:GetDiscardDeck() )
        else
            AUDIO:PlayEvent("event:/sfx/battle/cards/neg/improvised_goes_to_deck")
            self:BroadcastEvent( EVENT.CARD_IMPROVISED, card, self )
        end
    end
    return cards
end

function MiniGame:PrepareCard( hand_card, source )
    local card = hand_card or self:ChooseCard( nil, LOC"CARD_ENGINE.CHOOSE_CARD_PREPARE", source)
    if card then
        if not card:IsPrepared() then
            card:TransferCard( self:GetHandDeck(), 1 )
            self:BroadcastEvent( EVENT.CARDS_SWAPPED, self, {card}, 1)
        end
        return card
    end
end

function MiniGame:GetFatiguedTurn()
    return self.fatigue_turn
end

function MiniGame:SetFatiguedTurn( val )
    self.fatigue_turn = val
end

function MiniGame:LookupQuip(tags, seed)
    if self.current_quip_db then
        return self.current_quip_db:LookupQuip(tags, seed)
    end
end


function MiniGame:__tostring()
    return string.format( "Negotiation[%s]", tostring(self.status))
end

function MiniGame:RenderModifiersDebugUI( ui, panel, negotiator )
    self.filter_txt = ui:InputText( "Filter", self.filter_txt or "" ) or self.filter_txt
    ui:Separator()

    for i, modifier in negotiator:Modifiers() do
        local new_stacks = ui:SliderInt( modifier.id, modifier.stacks, modifier.min_stacks or 0, modifier.max_stacks )
        if new_stacks and new_stacks ~= modifier.stacks then
            if new_stacks == 0 and not modifier.is_counter then
                negotiator:RemoveModifier( modifier )
            else
                local delta = new_stacks - modifier.stacks
                negotiator:DeltaModifier( modifier, delta )
            end
        end

        if modifier then
            ui:SameLine( nil, 10 )
            panel:AppendTable( ui, modifier )
        end
    end

    ui:Separator()

    for i, id, def in sorted_pairs( Content.GetAllNegotiationModifiers() ) do
        if self.filter_txt and id:upper():find( self.filter_txt:upper() ) then
            local new_stacks = ui:SliderInt( id, 0, def.min_stacks or 0, def.max_stacks )
            if new_stacks then
                negotiator:AddModifier( id, new_stacks )
            end
        end
    end
end

function MiniGame:RenderDebugUI( ui, panel )
    ui:TextColored( 0x00CCCCFF, tostring(self.status) )
    if ui:IsItemHovered() then
        ui:SetTooltip( tostr(self.wait_params) )
    end

    ui:SameLine( nil, 50 )
    ui:Text( string.format( "Turn: %d", self.turns ))

    local action_count = ui:SliderInt( LOC"UI.CARDS.ACTIONS", self.action_count, 0, 9 )
    if action_count and action_count ~= self.action_count then
        self:ModifyActionCount( action_count - self.action_count )
    end

    local resolve, max_resolve = self.player_negotiator:GetResolve()
    if resolve then
        local new_resolve = ui:SliderInt( "Resolve", resolve, 0, max_resolve )
        if new_resolve and new_resolve ~= resolve then
            self.player_negotiator:ModifyResolve( new_resolve - resolve )
        end
    end

    local composure = self.player_negotiator:GetComposure()
    if composure then
        local new_composure = ui:SliderInt( "Composure", composure, 0, 20 )
        if new_composure and new_composure ~= composure then
            self.player_negotiator:DeltaComposure( new_composure - composure )
        end
    end

    local resolve, max_resolve = self.opponent_negotiator:GetResolve()
    if resolve then
        local new_resolve = ui:SliderInt( "Enemy Resolve", resolve, 0, max_resolve )
        if new_resolve and new_resolve ~= resolve then
            self.opponent_negotiator:ModifyResolve( new_resolve - resolve )
        end
    end

    local composure = self.opponent_negotiator:GetComposure()
    if composure then
        local new_composure = ui:SliderInt( "Enemy Composure", composure, 0, 20 )
        if new_composure and new_composure ~= composure then
            self.opponent_negotiator:DeltaComposure( new_composure - composure )
        end
    end

    local money = self:GetMoney()
    local new_money = ui:SliderInt( "Money", money, 0, 1000 )
    if new_money and new_money ~= money then
        self:ModifyMoney( new_money - money )
    end

    local flourish_graft = self.player_negotiator.agent.graft_owner and self.player_negotiator.agent.graft_owner:GetGraft( "flourish_tracker" )
    if flourish_graft then
        local new_value = ui:SliderFloat("FLOURISH", flourish_graft.userdata.negotiation_flourish or 0, 0, flourish_graft.userdata.negotiation_max or 200)
        if new_value and new_value ~= flourish_graft.userdata.negotiation_flourish then
            flourish_graft:GetDef().SetNegotiationFlourish(flourish_graft, self.player_negotiator.engine, new_value)
        end
    end


    ui:Spacing()

    panel:AppendTable( ui, self.player_negotiator )
    ui:SameLine( nil, 10 )
    panel:AppendTable( ui, self.opponent_negotiator)

    if ui:TreeNode( "Decks" ) then
        self.debug_deck = self.debug_deck or table.arrayfind( DECK_TYPE_ARRAY, DECK_TYPE.IN_HAND )
        local new_deck = ui:Combo( "Deck", self.debug_deck, DECK_TYPE_ARRAY )
        if new_deck and new_deck ~= self.debug_deck then
            self.debug_deck = new_deck
        end

        local deck = self:GetDeck( DECK_TYPE_ARRAY[ self.debug_deck ])
        deck:RenderDebugUI( ui, panel, self )
        ui:TreePop()
    end

    if ui:TreeNode( "Player Modifiers" ) then
        self:RenderModifiersDebugUI( ui, panel, self.player_negotiator )   
        ui:TreePop()
    end

    if ui:TreeNode( "Opponent Modifiers" ) then
        self:RenderModifiersDebugUI( ui, panel, self.opponent_negotiator )   
        ui:TreePop()
    end 

    panel:AppendTable( ui, self.start_params, "Start Params" )

    ui:SameLine( nil, 10 )
    if self.coro and ui:Button( "Negotiation Coroutine" ) then
        TheGame:GetDebug():CreatePanel( DebugTraceback( self.coro ))
    end

    ui:SameLine( nil, 10 )
    panel:AppendTable( ui, Content.internal.QUIPS, "Quips" )
end
