local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

--------------------------------------------------------------------------------
-- A negotiation modifier, either hidden or within a Negotiation.Slot.

local Modifier = class("Negotiation.Modifier")
Modifier.can_reload = true

function Modifier:init( modifier_id, owner, stacks, source )
    local params = Content.GetNegotiationModifier( modifier_id )
    
    assert( params, modifier_id )

    for k,v in pairs(params) do 
        self[k] = v
    end

    self.def = params
    self.stacks = 0
    self.composure = 0
    self.rand_seed = math.random( 1, 2^32 )

    local minval = self.min_resolve or self.max_resolve
    local maxval = self.max_resolve
    
    if minval ~= maxval then
        self.resolve = math.round( math.randomGauss(minval, maxval) )
        self.max_resolve = self.resolve
    else
        self.resolve = maxval
        self.max_resolve = maxval
    end

    self:ApplyOwner( owner )

    if self.engine then
        self.uid = self.engine:GenerateUID()
        self.turn_applied = self.engine:GetTurns()
    end

    self:ApplyStacks( stacks or 1 )


    if self.OnInit then
        self:OnInit( source )
    end
end

function Modifier:GetUID()
    -- Unique identifying ordinal for this modifier (within the parent negotiation)
    return self.uid
end

function Modifier:GetTurnApplied()
    return self.turn_applied
end

function Modifier:GetOwnerName()
    return self.negotiator and self.negotiator:GetName() or LOC "UI.CARDS.OWNER"
end

function Modifier:GetOpponentName()
    return self.anti_negotiator and self.anti_negotiator:GetName() or LOC "UI.CARDS.OPPONENT"
end

function Modifier:ApplyOwner( owner )
    if is_instance( owner, CardEngine ) then
        self.engine = owner

    elseif is_instance( owner, Negotiation.Negotiator ) then
        self.negotiator = owner
        self.owner = self.negotiator.agent
        self.engine = owner.engine
        self.anti_negotiator = self.engine:GetOtherNegotiator( self.negotiator )
    end
end

function Modifier:Clone()
    return setmetatable( shallowcopy( self ), Modifier )
end

function Modifier:Duplicate( owner )
    local clone = self:Clone()
    clone.uid = self.engine:GenerateUID()
    if owner and owner ~= self.negotiator then
        self:ApplyOwner( owner )
    end
    return self.negotiator:CreateModifier( clone )
end

function Modifier:IsPlayerOwner()
    return self.negotiator and self.negotiator:IsPlayer()
end

function Modifier:IsAttack()
    return self.min_persuasion ~= nil
end

function Modifier:GetRandSeed()
    return self.rand_seed
end

function Modifier:ApplyModifier()
    if self.event_handlers then
        for event_name, fn in pairs( self.event_handlers ) do
            local priority = self.event_priorities and self.event_priorities[ event_name ]
            self.engine:ListenForEvent( event_name, self, self.OnNegotiationEvent, priority )
        end
    end

    self.turn_applied = self.engine:GetTurns()

    if self.modifier_type == MODIFIER_TYPE.CORE then
        if self.negotiator:IsPlayer() or not CheckBits( self.engine.flags, NEGOTIATION_FLAGS.NO_CORE_RESOLVE ) then
            self.resolve, self.max_resolve = self.negotiator.resolve, self.negotiator.max_resolve
        end
    end

    if self.OnApply then
        self:OnApply( self.engine )
    end
end

function Modifier:IsApplied()
    return self.turn_applied and not self.turn_unapplied
end

function Modifier:UnapplyModifier()
    if self.OnUnapply then
        self:OnUnapply( self.engine )
    end
    if self.event_handlers then
        self.engine:RemoveListener( self )
    end

    self.turn_unapplied = self.engine:GetTurns()
end

function Modifier:NotifyTriggered()
    self.engine:BroadcastEvent( EVENT.MODIFIER_TRIGGERED, self )

    -- If this YIELD occurs midway through a CALC_ handler, that's bad.
    -- self.engine:YieldPlay( self.engine.STATUS.SYNC )
end

function Modifier:NotifyChanged( delta, source )
    if self.negotiator:HasModifier( self ) then
        self.engine:BroadcastEvent( EVENT.MODIFIER_CHANGED, self, delta, self:Clone(), source )
    end
end

function Modifier:OnNegotiationEvent( eventname, ... )
    if self.event_handlers and self.event_handlers[ eventname ] then
        self.event_handlers[ eventname ]( self, ... )
    end
end

function Modifier:PrepareTurn()
    if not self:NeedsTarget() or self.min_persuasion == nil then
        -- no target needed, or not a damaging modifier.
    elseif self.target_cleared_turn and self.engine:GetTurns() <= self.target_cleared_turn then
        print( self, "Already cleared target: not prepping" )
    else
        self.engine:AssignPrimaryTarget( self )
    end
end

function Modifier:NeedsTarget()
    return (self.target_self or 0) ~= 0 or (self.target_enemy or 0) ~= 0
end

function Modifier:AssignTarget( target )
    self.target = target
end

function Modifier:ClearTarget()
    self.target = nil
    self.target_cleared_turn = self.engine:GetTurns()
    self.rand_seed = math.random( 1, 2^32 )
end

function Modifier:GetTarget()
    return self.target
end

function Modifier:IsBeingTargetted()
    -- In case we are a UI clone, need to search with the actual modifier.
    local modifier = self.engine:FindModifierByUID( self.uid )
    if modifier == nil then
        return false
    end

    return self.engine:IsBeingTargetted( modifier )
end

function Modifier:ApplyPersuasion()

    self.target_cleared_turn = nil
    self:PrepareTurn()

    if self.min_persuasion and self.max_persuasion and self.target then
        local targets = self.engine:CollectTargets( self )
        for i, target in ipairs( targets ) do
            self.engine:ApplyPersuasion( self, target )
        end

        self:ClearTarget()
        self:NotifyTriggered()
        return true
    end
end

function Modifier:BeginTurn()
    if self.composure > 0 then
        if self.composure_accumulator == nil then
            self.composure_accumulator = CardEngine.ScalarAccumulator( self.engine, EVENT.CALC_COMPOSURE_DECAY )
        end

        local delta = self.composure_accumulator:CalculateValue( self.composure, self )
        if delta > 0 then
            self:DeltaComposure( -delta, nil, true )
        end
    end

    if self.OnBeginTurn then
        self:OnBeginTurn( self.engine )
    end
end

function Modifier:GetNegotiator()
    return self.negotiator
end

function Modifier:GetComposure()
    return self.composure
end

function Modifier:DeltaComposure( delta, source, start_of_turn )
    if self.resolve == nil or not self:IsApplied() then
        return
    end
    
    local old_composure = self.composure
    delta = self.engine:CalculateComposure(delta, self, source)
    self.composure = math.max( 0, self.composure + delta )

    self.engine:BroadcastEvent( EVENT.DELTA_COMPOSURE, self, self.composure, old_composure, source, start_of_turn )
    return self.composure - old_composure 
end

function Modifier:GetResolve()
    return self.resolve, self.max_resolve
end

function Modifier:LoseResolve( resolve, source )
    self:ModifyResolve( -resolve, source )
end

function Modifier:RestoreResolve( delta, source )
    if self.resolve == nil then
        return
    end

    delta = self.engine:CalculateRestoreResolve( delta, source )
    delta = math.min( self.max_resolve - self.resolve, delta )

    self.resolve = clamp( self.resolve + delta, 0, self.max_resolve )
    
    self.engine:BroadcastEvent( EVENT.RESTORE_RESOLVE, self, self.resolve, self.max_resolve, delta, source )
end

function Modifier:GetShieldStatus()
    return self.is_shielded, self.shield_desc
end

function Modifier:SetShieldStatus(shield_status, shield_desc)
    if not self:IsApplied() then
        print( self, "SetShieldStatus", shield_status, "while not applied" )
        return
    end

    if (self.is_shielded == true) ~= (shield_status == true) then
        self.is_shielded = shield_status
        self.shield_desc = shield_desc
        self.engine:BroadcastEvent( EVENT.SHIELD_STATUS_CHANGE, self, shield_status )

        if shield_status then
            local sources = {}
            if self.engine:IsBeingTargetted( self, sources ) then
                print( "BEING TARGETTED by", tostr(sources))
                for i, v in ipairs(sources) do
                    print( i, v.target )
                    self.engine:AssignPrimaryTarget( v )
                    print( "\t now targetting", v.target )
                end
            end
        end
    end
end

function Modifier:ModifyResolve( delta, source, params )
    if self.resolve == nil then
        return
    end

    local old_resolve = self.resolve

    self.resolve = math.max( self.resolve + delta, 0 )
    self.max_resolve = math.max( self.resolve, self.max_resolve)

    self.engine:BroadcastEvent( EVENT.DELTA_RESOLVE, self, self.resolve, self.max_resolve, delta, source, params )
    -- This event might in fact destroy us.
    if not self:IsApplied() then
        return
    end

    if self.resolve <= 0 then
        self.negotiator:DestroyModifier( self, source )
    end

    -- FIXME: shouldn't this go in AttackResolve only?
    -- (but then this modifier may be destroyed by then)
    local overflow = delta - (self.resolve - old_resolve)
    if overflow < 0 and self.modifier_type ~= MODIFIER_TYPE.CORE then
        -- Suppress 'flinch' emote on leftovers.
        params = params or {}
        params.hit_emote = table.empty

        -- By default, splash to any force_target.
        -- for i, modifier in self.negotiator:Modifiers() do
        --     if modifier.force_target then
        --         params.splashed_modifier = modifier
        --         break
        --     end
        -- end

        -- If none, hit core.
        local core = self.negotiator:FindCoreArgument()
        if core and not core:GetShieldStatus() and params.splashed_modifier == nil then
            params.splashed_modifier = core
        end

        if params.splashed_modifier then
            params.bonus_splash = 0
            -- Let listeners modify the splash target, prevent it, or DO ANYTHING.
            self.engine:BroadcastEvent( EVENT.SPLASH_RESOLVE, self, overflow, params )

            -- Event handlers may modify params.
            if params.splashed_modifier then
                params.splashed_modifier:AttackResolve(-overflow + params.bonus_splash, self, params)
            end
        end
    end
end

function Modifier:AttackResolve( delta, source, params )
    if self.resolve == nil or (self.GetShieldStatus and self:GetShieldStatus()) then
        return
    end

    delta = math.max( 0, delta )
    local defended = math.min( delta, self.composure )

    -- Suppress 'flinch' emote if fully defended
    if defended >= delta then
        params = params or {}
        params.hit_emote = table.empty
        params.shake_params = false
    end

    self.engine:BroadcastEvent( EVENT.ATTACK_RESOLVE, source, self, delta, params, defended )

    if source and source.negotiator and source.negotiator:IsPlayer() then
        TheGame:AddGameplayStat( "resolve_damage_dealt", delta )
    end

    -- Guard against ATTACK_RESOLVE handlers removing us.
    if not self.negotiator:HasModifier( self ) then
        return
    end

    local defended = math.min( delta, self.composure )
    if defended > 0 then
        self:DeltaComposure( -defended, source )
        delta = delta - defended
    end

    if delta > 0 then
        local old_resolve = self.resolve

        -- Suppress floater (issued by EVENT.ATTACK_RESOLVE above)
        params = params or {}
        params.no_floater = true

        self:ModifyResolve( -delta, source, params )
    end
end

function Modifier:SetResolve( resolve, scaling, max_resolve )
    if self.engine == nil then
        return
    end

    if resolve == nil then
        self.resolve, self.max_resolve = nil, nil
        self:NotifyChanged()
        return
    end
    
    local difficulty = self.engine:GetDifficulty()

    -- Don't immediately scale on diff 1
    -- if difficulty == 1 then
    --     turn_count = math.max(0, turn_count - 4)
    -- end

    local modifiers = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_ARGUMENT_RESOLVE )
    if modifiers and modifiers[ difficulty ] then
        resolve = resolve + modifiers[ difficulty ]
    end

    if scaling == MODIFIER_SCALING.LOW then
        resolve = resolve + math.floor(difficulty/2) --+ math.floor(turn_count/2)
    elseif scaling == MODIFIER_SCALING.MED then
        resolve = resolve + difficulty-- + turn_count
    elseif scaling == MODIFIER_SCALING.HIGH then
        resolve = resolve + difficulty * 2 --+ turn_count
    end

    self.resolve = self.resolve or 0
    self.max_resolve = max_resolve or resolve

    local delta
    if self.resolve > self.max_resolve then
        delta = self.max_resolve - self.resolve
    else
        delta = resolve - self.resolve
    end

    self:ModifyResolve( delta )
end

function Modifier:GetStacks()
    return self.stacks, self.max_stacks
end

function Modifier:ApplyStacks(num)
    return self:SetStacks( self.stacks + num )
end

function Modifier:SetStacks(num)
    local old_stacks = self.stacks

    self.stacks = clamp(num, self.min_stacks or 0, self.max_stacks)

    if self.stacks ~= old_stacks and self.OnSetStacks and self.engine then
        self:OnSetStacks( old_stacks )
    end
    return  self.stacks - old_stacks
end

function Modifier:GetName()
    -- ccc: imo this is dumb.  This should just be 'return self.name', where self.name is a localized string
    if self.custom_name then
        return self.custom_name
	elseif self.def then
		return self.def:GetLocalizedName()
	end
end

function Modifier:GetFeatureDesc( minigame, widget, features )
    local desc = self:GetDesc( minigame, widget, features )

    if self.resolve then
        -- Show how much resolve this argument can have if we are accessing teh description through a feature tooltip.
        if desc then
            desc = desc .. "\n\n" .. loc.format( LOC "UI.CARDS.ARGUMENT_RESOLVE", self:GetName(), self.resolve )
        else
            desc = loc.format( LOC "UI.CARDS.ARGUMENT_RESOLVE", self:GetName(), self.resolve )
        end
    end

    return desc
end

function Modifier:GetDesc( minigame, widget, features )

    minigame = minigame or self.engine
    local desc
	if self.def then
		desc = self.def:GetLocalizedDesc()
	end

    if self.desc_fn then
        desc = self:desc_fn(desc, minigame, widget, features )
    elseif desc == nil then
        desc = self.desc
    end

    if self.min_persuasion and self.max_persuasion and (self.target_enemy ~= nil or self.target_self ~= nil) then
        local min_persuasion, max_persuasion, details = self.min_persuasion, self.max_persuasion
        if minigame then
            min_persuasion, max_persuasion, details = minigame:PreviewPersuasion( self, true )
        end
        local tt
        if self.damage_tt then
            tt = loc.format( LOC( self.damage_tt ), min_persuasion, max_persuasion )

        elseif not self.no_damage_tt then
            if self.target_self ~= nil then
                tt = loc.format( LOC"UI.NEGOTIATION_PANEL.ARGUMENT_SELF_DAMAGE_TT", min_persuasion, max_persuasion )
            else
                tt = loc.format( LOC"UI.NEGOTIATION_PANEL.ARGUMENT_DAMAGE_TT", min_persuasion, max_persuasion )
            end
        end

        if details then
            tt = (tt or "") .. loc.format( LOC"UI.NEGOTIATION_PANEL.EFFECT_DETAILS_TOOLTIP", details )
        end
        if tt then
            if desc then
                desc = tt .. "\n\n" .. desc
            else
                desc = tt
            end
        end
    end
    
    if self.modifier_type == MODIFIER_TYPE.CORE and self.negotiator and self.resolve ~= nil then
        -- local tt
        -- if self.negotiator:IsPlayer() then
        --     tt = LOC"UI.NEGOTIATION_PANEL.YOUR_RESOLVE_DESC"
        -- else
        --     tt = LOC"UI.NEGOTIATION_PANEL.THEIR_RESOLVE_DESC"
        -- end

        -- if desc then
        --     desc = tt .. "\n\n" .. desc
        -- else
        --     desc = tt
        -- end
    end

    local is_shielded, shield_desc = self:GetShieldStatus()
    if is_shielded then
        local feature = Content.GetNegotiationCardFeature( "SHIELDED" )
        local tt = shield_desc
        if tt == nil and feature then
            tt = loc.format( "{1}. {2}", feature:GetLocalizedName(), feature:GetLocalizedDesc() )
        end
        if tt then
            if desc then
                desc = loc.format( "<#PENALTY>{1}</>\n\n{2}", tt, desc )
            else
                desc = "<#PENALTY>"..tt.."</>"
            end
        end
    end

    local oldest = self.negotiator and self.negotiator:FindOldestArgument()
    if oldest and oldest.uid == self.uid and self.negotiator:GetSlotCount() >= self.negotiator.max_modifiers then
        local tt = loc.format( LOC "UI.NEGOTIATION_PANEL.ARGUMENT_OLDEST", self.negotiator:GetName(), self.negotiator.max_modifiers - 1 )
        if desc then
           desc = desc .. "\n\n" .. tt
       else
            desc = tt
       end
    end

    if desc then
        desc = CardEngine.ExtractKeywords( desc, widget, minigame, negotiation_defs, features )
    end

    return desc
end

function Modifier:__tostring()
    return string.format( "%s:%d (x%d)", self.id, self.uid or -1, self.stacks )
end
