local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT

local MAX_DAMAGE = 999

---------------------------------------------------
-- Wrapper for tracking resolve damage (persuasion).

local PersuasionAccumulator = class( "Negotiation.PersuasionAccumulator" )
PersuasionAccumulator.can_reload = true

function PersuasionAccumulator:init( minigame, event_name )
    assert( event_name )
    self.minigame = minigame
    self.event_name = event_name
end

function PersuasionAccumulator:IsMinPersuasionModified()
    return self.min_persuasion_modified == true
end

function PersuasionAccumulator:IsMaxPersuasionModified()
    return self.max_persuasion_modified == true
end

function PersuasionAccumulator:CalculatePersuasion( source, target, min_persuasion, max_persuasion )

    if min_persuasion and max_persuasion then
        self.min_persuasion, self.max_persuasion = min_persuasion, max_persuasion
    elseif source.CalculatePersuasion then
        local details
        self.min_persuasion, self.max_persuasion, details = source:CalculatePersuasion()
        if self.details and details then
            table.insert( self.details, details )
        end
    else
        self.min_persuasion, self.max_persuasion = source.min_persuasion, source.max_persuasion
    end

    self.min_persuasion = math.round( self.min_persuasion or 0 )
    self.max_persuasion = math.round( self.max_persuasion or 0 )

    self.minigame:BroadcastEvent( self.event_name, source, self, self.minigame, target )
    self.min_persuasion = math.max( 0, self.min_persuasion )
    self.max_persuasion = math.max( 0, self.max_persuasion ) 
end


function PersuasionAccumulator:PreviewPersuasion( card, no_rand )
    self.details = ObtainWorkTable()
    self.preview = true
    self.min_persuasion_modified = nil
    self.max_persuasion_modified = nil

    self:CalculatePersuasion( card, card:GetTarget() )

    local str
    if #self.details > 0 then
        str = loc.format( "{1#listing}", self.details ) -- table.concat( self.details, "\n" )
    else
        str = nil
    end

    ReleaseWorkTable( self.details )
    self.details = nil

    if not no_rand and card:GetRandSeed() then
        local rand = engine.Random( self.seed, card:GetRandSeed() )
        local damage = rand:Random( self.min_persuasion, self.max_persuasion )
        return damage, damage, str, self
    else
        return self.min_persuasion, self.max_persuasion, str, self
    end
end

function PersuasionAccumulator:ApplyPersuasion( card, target, min_persuasion, max_persuasion )
    self.preview = false

     -- Check target resolve here in case this target has "died"
     -- Ideally CollectTargets would not accrue such targets, but there's some cards
     -- apply persuasion damage but target a non-resolve target. (eg. Bluster)
    if target == nil or (target:GetResolve() or 0) <= 0 then
        return
    end

    self:CalculatePersuasion( card, target, min_persuasion, max_persuasion )

    local rand
    if card:GetRandSeed() then
        rand = engine.Random( self.seed, card:GetRandSeed() )
    end

    -- Roll the persuasion...
    local persuasion
    if rand then
        persuasion = rand:Random( self.min_persuasion, self.max_persuasion )
    else
        persuasion = math.random( self.min_persuasion, self.max_persuasion )
    end

    local params
    if persuasion == self.max_persuasion and self.min_persuasion ~= self.max_persuasion then
        params = { hit_emote = "neutral_wince_critical", shake_params = NEGOTIATION_SHAKE_CRIT }
    end

    target:AttackResolve( persuasion, card, params )
end

function PersuasionAccumulator:AddPersuasion( min_delta, max_delta, source )
    self:ModifyPersuasion( self.min_persuasion + min_delta, self.max_persuasion + max_delta, source )
end

function PersuasionAccumulator:ClampPersuasion( min_persuasion, max_persuasion, source )
    local min_delta = min_persuasion - self.min_persuasion
    local max_delta = max_persuasion - self.max_persuasion
    self:ModifyPersuasion( min_delta, max_delta, source )
end

function PersuasionAccumulator:ModifyPersuasion( min_persuasion, max_persuasion, source )
    if min_persuasion ~= self.min_persuasion then
        self.min_persuasion = clamp( min_persuasion, 0, MAX_DAMAGE )
        self.min_persuasion_modified = true
    end
    if max_persuasion ~= self.max_persuasion then
        self.max_persuasion_modified = true
    end
    self.max_persuasion = clamp( max_persuasion, self.min_persuasion, MAX_DAMAGE )

    if self.details then
        assert_warning( source ~= nil, "Missing source for modified persuasion" )
        local detail
        if is_instance( source, Negotiation.Card ) then
            detail = source:GetName() --loc.format( "<b><#TITLE>{1}</></> {2}", source.name, source:GetDesc() )
        elseif is_instance( source, Negotiation.Modifier ) then
            detail = source:GetName() --loc.format( "<b><#TITLE>{1}</></> {2}", source.name, source:GetDesc() )
        else
            detail = tostring(source)
        end

        if detail then
            table.insert( self.details, detail )
        end
    end
end

