require "util/rewardutil"
local paperdoll = require "paperdoll_util"

--------------------------------------------------------------------------------
-- Stateful instantiation of an active quest.
-- Quests maintain a completion status (QSTATUS) and active objectives.

QSTATUS = MakeEnum{ "INACTIVE", "ACTIVE", "COMPLETE", "FAILED", "CANCELLED" }

QTYPE = MakeEnum{   "STORY",        --tied to the act
                    "SIDE",         --optional ! over the head opportunity that shows up on someone's head and follows them around
                    "EVENT",        --something that happens when you go somewhere, and is presented at the time.
                    "CONTRACT",     --things that your handler gives you
                    "OPPORTUNITY", --one-time only ops that clear out if you don't visit them
                    "WANDERING_SHOP",
                    "SCENARIO",         --only manually spawned
                    "EXPERIMENT",
                    "DAILY",
                    "FOLLOWUP",
}

QREWARDSIZE = MakeEnum{ "SMALLER", "NORMAL", "LARGER"}


QWHEN = MakeEnum{ "DEFAULT", "MANUAL"}
QWHAT = MakeEnum{ "AGENT", "LOCATION", "PARAMETER" }
QFLAGS = MakeBitField{
    "AUTO_COMPLETE" -- Auto-completed after finishing an encounter with this quest.
}

QEVENT_PRIORITY = 
{
    HIGH = 10,
    NORMAL = 0,
    LOW = -10,
}

QEVENT_TRIGGER, QEVENT_TRIGGER_STRINGS = MakeBitField{
        "FIRST_LOCATION_VISIT",
        "TRAVEL",
        "MANUAL",
    }

QUEST_MARK_CONTEXT = MakeEnum{"PLAX", "WORLD", "MAP"}

QLIMIT = MakeEnum{
    "NOLIMIT", -- The cast member is not limited by how many quests they are assigned to.
}

QCARD_TYPE = MakeEnum{
    "BATTLE",
    "NEGOTIATION"
}

QUEST_MAX_RANK = 10

local QuestState = class( "QuestState" )

function QuestState:init( content_id, rank, spawn_params )
    
    local qdef = Content.GetQuestDef( content_id )
    assert( qdef, content_id )

    self.id = TheGame:GetGameState():GenerateID()
    self.content_id = content_id
    self.status = QSTATUS.INACTIVE
    self.cast = {}
    self.objectives = {}
    self.trace = {}
    self.param = {}
    self.jobs = {}
    self.clean_up_locations = {}
    self.spawn_params = spawn_params or table.empty
    self.hide_in_overlay = false
    self.valid = false
    
    self.qrank = rank

    if spawn_params and spawn_params.parameters then
        for param,val in pairs(spawn_params.parameters) do
            self:SetParameter(param, val)
        end
    end

    self.quest_provider = nil 

    self.reward_random = math.randomGauss(.8, 1.2)

end


function QuestState:CollectTooltipItems(t)
    local qdef = self:GetQuestDef() 
    if qdef.collect_tooltip_items then
        qdef.collect_tooltip_items(self, t)
    end

    if self.extra_reward == EXTRA_QUEST_REWARD.FREE_ITEM then
        table.insert_unique(t, self.extra_reward_data)
    end
end

function QuestState:ProcessFighter(fighter)
    local qdef = self:GetQuestDef() 
    if qdef.process_fighter then
        qdef.process_fighter(self, fighter)
    end

    for i, obj_def in qdef:Objectives() do
        if self:IsActive(obj_def.id) and obj_def.process_fighter then
            obj_def.process_fighter(self, fighter)
        end
    end
end

function QuestState:GetMettleReward()
    if (self:GetType() == QTYPE.CONTRACT or self:GetType() == QTYPE.SIDE) and TheGame:GetGameState():CanDoMettle() then
        return QUEST_METTLE_REWARD[self.qrank] or 0
    end

    return 0
end

function QuestState:GetRewardSize()
    return self:GetQuestDef().reward_size or QREWARDSIZE.NORMAL
end

function QuestState:BlocksSideJobs()
    if self:IsActive() then
        local qdef = self:GetQuestDef() 

        if qdef.block_side_jobs then
            if type(qdef.block_side_jobs) == "function" then
                if qdef.block_side_jobs(self) then
                    return true
                end
            else
                return true
            end
        end

        for i, obj_def in qdef:Objectives() do
            if self:IsActive(obj_def.id) and obj_def.block_side_jobs then
                if type(obj_def.block_side_jobs) == "function" then
                    if qdef.block_side_jobs(self) then
                        return true
                    end
                else
                    return true
                end
            end
        end
    end
end

function QuestState:Trace( fmt, ... )
    if self.trace then
        table.insert( self.trace, string.format( fmt, ... ))
    else
        print( tostring(self), string.format( fmt, ... ))
    end
end

function QuestState:IsValid()
    -- We are valid if we've been setup via PopulateGameState, and have not yet been Destroy()ed.
    return self.valid
end

function QuestState:LoadCastAnims()
    for k,cast_member in pairs(self.cast) do
        if is_instance( cast_member, Agent ) then
            paperdoll.LoadAgentAnims(cast_member)
        end
    end
end

function QuestState:HasUpfrontReward()
    return self.upfront_reward
    --return self:GetType() ~= QTYPE.STORY and TheGame:GetGameState():GetCurrentAct().data.gives_upfront_rewards
end

function QuestState:PopulateGameState( gs )

    self.valid = true

    local def = self:GetQuestDef()

    for i, castdef in def:CastDefs() do
        local cast_member = self.cast[ castdef.cast_id ]
        if cast_member then
            self:OnAssignCastMember( castdef.cast_id, cast_member )
        end
    end
    
    self:ListenForEvents()

    for i, obj_def in def:Objectives() do
        if obj_def.state then
            self:SetObjectiveState( obj_def.id, obj_def.state )
        end
    end

    self:RunScriptHandler( def.on_init )

    -- Return whether we are still valid: yes it's possible to be destroyed during setup!
    return self:IsValid()
end


function QuestState:OnStart()
    assert ( self.status == QSTATUS.INACTIVE, tostring(self.status) )

    self.status = QSTATUS.ACTIVE

    if self:GetQuestDef().on_start then
        self:GetQuestDef().on_start(self)
    end

    self:NotifyChanged()
    
    if self:GetType() == QTYPE.SIDE or self:GetType() == QTYPE.CONTRACT then
        TheGame:GetGameState():ClearOpportunities()
    end
end

function QuestState:OnDebugStart()
    if self:GetQuestDef().on_debug_start then
        self:GetQuestDef().on_debug_start(self)
    end

    self:NotifyChanged()
end


function QuestState:PreventsRandomEvents()
    
    if self:IsActive() then

        if self:GetType() == QTYPE.OPPORTUNITY then
            return true
        end
        
        if self:GetQuestDef().prevent_random_events then
            if type(self:GetQuestDef().prevent_random_events) == "function" then
                if self:GetQuestDef().prevent_random_events(self) then
                    return true
                end
            else
                return true
            end
        end
        if self.objectives then
            for i,obj in ipairs(self.objectives) do
                if obj.state == QSTATUS.ACTIVE then
                    
                    if self:GetQuestDef():GetObjective(obj.id).prevent_random_events then
                        if type(self:GetQuestDef():GetObjective(obj.id).prevent_random_events) == "function" then
                            if self:GetQuestDef():GetObjective(obj.id).prevent_random_events(self) then
                                return true
                            end
                        else
                            return true
                        end

                    end
                end
            end
        end
    end
end



function QuestState:OnFinish( status )


    assert( status == QSTATUS.FAILED or status == QSTATUS.COMPLETE or status == QSTATUS.CANCELLED )
    if self:IsDone() then
        if self.status ~= status then
            print( string.format( "[%s] Already complete: %s; assigning %s\nDestroyed at %s", self.content_id, self.status, status, self.destroy_traceback ))
        end
        return
    end

    local was_active = self.status == QSTATUS.ACTIVE

    self.status = status

    -- Finish any outstanding objectives.
    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                self:SetObjectiveState( obj.id, status )
            end
        end
    end

    if was_active then

        if self:GetType() == QTYPE.CONTRACT or self:GetType() == QTYPE.SIDE then
            
            TheGame:GetGameState():AddScore("JOBS_COMPLETED", 1)
            TheGame:AddGameplayStat( "quest_complete", 1 )

            if status == QSTATUS.COMPLETE then
                TheGame:GetGameState():LogNotification( NOTIFY.COMPLETED_JOB, self )
            elseif status == QSTATUS.FAILED then
                TheGame:GetGameState():LogNotification( NOTIFY.FAILED_JOB, self )
            end

            local who_is_happy = self:GetProvider() or self:GetCastMember("handler")    
            if who_is_happy then
                if status == QSTATUS.FAILED then
                    who_is_happy:OpinionEvent(OPINION.JOB_FAILED_MINOR, { quest = self })
                elseif status == QSTATUS.COMPLETE then
                    if self.param.poor_performance then
                        who_is_happy:OpinionEvent( OPINION.JOB_MINOR_POOR, { quest = self } )
                    else
                        who_is_happy:OpinionEvent( OPINION.JOB_MINOR, { quest = self } )
                    end

                end
            end
        end
    end

    local qdef = self:GetQuestDef()
    if status == QSTATUS.FAILED then
        if qdef.on_fail then
            self:RunScriptHandler( qdef.on_fail )
        end
    elseif status == QSTATUS.COMPLETE then
        if qdef.on_complete then
            self:RunScriptHandler( qdef.on_complete )
        end
    elseif status == QSTATUS.CANCELLED then
        if qdef.on_cancel then
            self:RunScriptHandler( qdef.on_cancel )
        end
    end

    local event_type = nil
    if status == QSTATUS.FAILED then
        event_type = "QUEST_FAILED"
    elseif status == QSTATUS.COMPLETE then
        event_type = "QUEST_COMPLETED"
    elseif status == QSTATUS.CANCELLED then
        event_type = "QUEST_CANCELLED"
    end

    if event_type then
        local json_t = { PLAYER_DATA = TheGame:GetGameState():GetPlayerState(), QUEST_ID = self.content_id }
        SendMetricsData(event_type, json_t)
    end

    TheGame:GetEvents():BroadcastEvent( "quest_finished", self )

    if not self:GetQuestDef():HasPostConvos() and self:IsValid() then
        self:Destroy()
    end

    if self:GetType() == QTYPE.OPPORTUNITY and status == QSTATUS.COMPLETE then
        TheGame:GetGameState():ClearOpportunities()
    end

end

function QuestState:Destroy()
    assert( self.valid )
    if not self:IsDone() then
        LOGWARN( "[%s] Destroying unfinished quest", self.content_id )
    end

    self.destroy_traceback = debug.traceback()

    TheGame:GetEvents():RemoveListener( self )

    TheGame:GetGameState():RemoveActiveQuest( self )

    local qdef = self:GetQuestDef()
    self:RunScriptHandler( qdef.on_destroy )

    self:CleanupCastMembers()

    if self.jobs then
        for k,job in pairs(self.jobs) do
            job:RetireJob()
        end
    end

    self:NotifyChanged()
    TheGame:GetEvents():BroadcastEvent( "quest_destroyed", self )
    
    for k,loc in ipairs(self.clean_up_locations) do
        TheGame:GetGameState():MarkLocationForDeletion(loc)
    end

    self.valid = false
end

function QuestState:GetStatus()
    return self.status
end

function QuestState:IsActive( objective_id )
    local status = objective_id and self:GetObjectiveState( objective_id ) or self.status
    return status == QSTATUS.ACTIVE
end

function QuestState:IsInactive( objective_id )
    local status = objective_id and self:GetObjectiveState( objective_id ) or self.status
    return status == QSTATUS.INACTIVE
end

function QuestState:IsFailed( objective_id )
    local status = objective_id and self:GetObjectiveState( objective_id ) or self.status
    return status == QSTATUS.FAILED
end

function QuestState:IsComplete( objective_id )
    local status = objective_id and self:GetObjectiveState( objective_id ) or self.status
    return status == QSTATUS.COMPLETE
end

function QuestState:IsDone( objective_id )
    local status = objective_id and self:GetObjectiveState( objective_id ) or self.status
    return status == QSTATUS.FAILED or status == QSTATUS.COMPLETE or status == QSTATUS.CANCELLED
end

function QuestState:Activate( objective_id )
    if objective_id then
        self:SetObjectiveState( objective_id, QSTATUS.ACTIVE )
    else
        self:OnStart()
    end
end

function QuestState:Fail( objective_id )
    if objective_id then
        self:SetObjectiveState( objective_id, QSTATUS.FAILED )
    else
        self:OnFinish( QSTATUS.FAILED )
    end
end

function QuestState:Complete( objective_id )
    if objective_id then
        self:SetObjectiveState( objective_id, QSTATUS.COMPLETE )
    else
        self:OnFinish( QSTATUS.COMPLETE )
    end
end

function QuestState:CompleteIfActive( objective_id )
    if self:IsActive( objective_id ) then
        self:Complete( objective_id )
    end
end


function QuestState:Cancel( objective_id )
    if objective_id then
        self:SetObjectiveState( objective_id, QSTATUS.CANCELLED )
    else
        self:OnFinish( QSTATUS.CANCELLED )
    end
end

function QuestState:SetProvider( agent )
    self.quest_provider = agent
end

function QuestState:GetProvider()
    return self.quest_provider
end

function QuestState:GetID()
    -- unique identifier: this will be nil if this quest has not yet been added to the GameState
    return self.id
end

function QuestState:GetContentID()
    return self.content_id
end

function QuestState:SetTitle(title)
    self.title = title
end

function QuestState:GetTitle()
    local result = self.title
    if self.title == nil then
        local quest_def = self:GetQuestDef()
        result = quest_def:GetTitle(self)
        if result then
            result = self:FormatString(result)
        end
    end
    return result
end

function QuestState:GetQuestFaction()
    local quest_def = self:GetQuestDef()
    if quest_def.quest_faction then
        return quest_def:quest_faction(self)
    end
end

function QuestState:SetHideInOverlay(val)
    self.hide_in_overlay = val
    self:NotifyChanged()
    return self
end

function QuestState:IsQuestVisibleInOverlay()
    if not self:IsActive() then
        return false
    end

    if self.hide_in_overlay then
        return false
    end

    if not self.is_hidden and self:GetTitle() then
        local qdef = self:GetQuestDef()
        for i, obj in ipairs( self.objectives ) do
            local obj_def = qdef:GetObjective( obj.id )
            local obj_title = qdef:GetObjectiveTitle(obj.id)
            local obj_desc = qdef:GetObjectiveDesc(obj.id)
            if obj.state == QSTATUS.ACTIVE and (obj_title ~= nil or obj_desc ~= nil) and not obj_def.hide_in_overlay then
                return true
            end
        end
    end

    return false
end


-- will this quest show up in your journal?
function QuestState:IsQuestVisible()
    if not self:IsActive() then
        return false
    end

    if not self.is_hidden and self:GetTitle() then
        local qdef = self:GetQuestDef()
        for i, obj in ipairs( self.objectives ) do
            --local obj_def = qdef:GetObjective( obj.id )
            local obj_title = qdef:GetObjectiveTitle(obj.id)
            local obj_desc = qdef:GetObjectiveDesc(obj.id)
            if obj.state == QSTATUS.ACTIVE and (obj_title ~= nil or obj_desc ~= nil) then
                return true
            end
        end
        local qdesc = qdef:GetDesc(self)
        if qdesc then
            return true
        end
end

    return false
end

function QuestState:CanAbandon()
    return self:GetQuestDef():CanAbandon()
end


function QuestState:IsQuestHidden()
    return self.is_hidden
end

function QuestState:SetQuestVisible( vis )
    self.is_hidden = not vis
    self:NotifyChanged()
end

function QuestState:RunScriptHandler( handler, ... )
    if handler then
        assert( type(handler) == "function", handler )
        handler( self, ... )
    end
end

function QuestState:GetIcon(obj)

    local icon
    if type( self:GetQuestDef().icon ) == "function" then
        icon = self:GetQuestDef().icon( self, obj )
    else
        icon = self:GetQuestDef().icon or global_images.default_quest_icon
    end

    return icon
end

function QuestState:GetObjectiveTitle( objective_id )
    local result = nil

    local qdef = self:GetQuestDef()
    local obj_def = qdef:GetObjective( objective_id )
    if obj_def.title then
        result = qdef:GetObjectiveTitle( objective_id )
        if obj_def.title_fn then
            result = obj_def.title_fn(self, result)
        end
        result = self:FormatString( result )
    end

    return result
end

function QuestState:GetObjectiveDesc( objective_id )
    local qdef = self:GetQuestDef()
    local obj_def = qdef:GetObjective( objective_id )

    if obj_def.desc then
        local obj_desc = qdef:GetObjectiveDesc( objective_id )
        if obj_def.desc_fn then
            obj_desc = obj_def.desc_fn(self, obj_desc)
        end
        return self:FormatString( obj_desc )
    end

    return nil--"MISSING OBJECTIVE DESCRIPTION"
end

function QuestState:GetDifficulty()
    if self:GetType() ~= QTYPE.STORY then
        return self.qrank
    end
end

function QuestState:GetDesc()
    local quest_def = self:GetQuestDef()
    local str = quest_def:GetDesc(self)
    if str then
        return self:FormatString( str )
    end
end

function QuestState:GetRank()
    return self.qrank
end

function QuestState:SetRank( qrank )
    self.qrank = clamp( qrank, 1, QUEST_MAX_RANK )
end

function QuestState:GetType()
    return self:GetQuestDef():GetType()
end


function QuestState:NotifyChanged()
    -- Hey WHOLE WORLD, we changed :p
    TheGame:GetEvents():BroadcastEvent( "quests_changed", self )
    TheGame:GetGameState():GetCaravan():RefreshConfronts()
end

function QuestState:AssignCastMember( cast_id, cast_member )
    if self.cast[ cast_id ] then
        LOGWARN( "[%s] has cast slot [%s] already assigned to %s", self.content_id, cast_id, tostring(self.cast[ cast_id]))
        return
    end

    assert( self:GetQuestDef():GetCast( cast_id ), cast_id )

    if cast_member ~= nil then
        self.cast[ cast_id ] = cast_member

        if self:IsValid() then
            -- Only do this once we are officially a populated quest.
            self:OnAssignCastMember( cast_id, cast_member )
        end
    else
        local qdef = self:GetQuestDef()
        local cast_def = qdef:GetCast( cast_id )
        local registrant = qdef:RegisterCastMember( self, cast_def )
        if registrant then
            self:AssignCastMember( cast_id, registrant )

        elseif not cast_def.optional then
            self:Trace( "[%s] Non-optional cast registration failed", cast_id )
        end
    end
end

function QuestState:OnAssignCastMember( cast_id, cast_member )
    local qdef = self:GetQuestDef()
    local cast_def = qdef:GetCast( cast_id )

    if is_instance( cast_member, Agent ) then
        if cast_member:GetID() == nil then
            TheGame:GetGameState():AddAgent( cast_member )
        end

        -- It is possible to cast dead people, but we don't want them to track anything.
        if not cast_member:IsRetired() then
            cast_member:AssignMembership( self )
            cast_member:ListenForAny( self, self.OnAgentEvent, self )
        end

        if cast_def.provider then
            self:SetProvider( cast_member )
        end

        local location = cast_member:GetLocation()
        -- If they're for some reason in a transient location, then send them home (unless the player is also there)
        if location ~= nil and CheckBits( location:GetFlags(), LOCFLAGS.TRANSIENT ) and location ~= TheGame:GetGameState():GetPlayerAgent():GetLocation() then
            cast_member:GetBrain():MoveToHome()
        end

    elseif is_instance( cast_member, Location ) then

        cast_member:AssignMembership( self )
        cast_member:ListenForAny(self, self.OnLocationEvent, self)
    end

    self:RunScriptHandler( cast_def.on_assign, cast_member )
end

function QuestState:UnassignCastMember( cast_id )
    local cast = self:GetCastMember( cast_id )
    local cast_def = self:GetQuestDef():GetCast( cast_id )
    
    if is_instance( cast, Agent ) then
        local agent = cast
        
        if not agent:IsRetired() then
            agent:RemoveListener( self )
            agent:UnassignMembership( self )

            if cast_def.cleanup then
                print( "Cleaning up:", self.content_id, cast_id, agent )
                agent:MoveToLimbo()
            end
        end

    elseif is_instance(cast, Location) then
        cast:UnassignMembership( self )
        cast:RemoveListener(self)

    end

    if not self:IsDone() then
        -- It's useful to maintain cast references after completion
        self.cast[ cast_id ] = nil
    end

    if cast then
        self:RunScriptHandler( cast_def.on_unassign, cast )
    end
end

function QuestState:IsCastMember( cast )
    assert( type(cast) == "table" ) -- Cast members are always tables of some thing.
    for cast_id, v in pairs(self.cast) do
        if v == cast then
            return cast_id
        end
    end
end

function QuestState:Cast()
    return pairs( self.cast )
end

function QuestState:GetCastMember( cast_id ) 
    return self.cast[ cast_id ]
end

function QuestState:GetCastAgent( cast_id )
    local cast_member = self.cast[ cast_id ]
    if is_instance( cast_member, Agent ) then
        return cast_member
    end
end

function QuestState:CreateSkinnedAgent( content_id )
    -- print ( "QuestState:CreateSkinnedAgent", self, content_id)
    -- Create an Agent with an un-used skin.
    local used_skin_uuids
    for cast_id, cast_member in pairs( self.cast ) do
        if is_instance( cast_member, Agent ) then
            local skin_uuid = cast_member:GetSkinID()
            if skin_uuid then
                if used_skin_uuids == nil then
                    used_skin_uuids = {}
                end
                table.insert( used_skin_uuids, skin_uuid )
            end
        end
    end

    return TheGame:GetGameState():CreateSkinnedAgent( content_id, used_skin_uuids )
end


function QuestState:FormatString( text )
    if text == nil then
        return nil
    end
    
    local function QuestFormatter( word, operator, rest )
        local cast_member = self:GetCastMember( word ) or self:GetParameter( word )
        if is_instance( cast_member, Agent ) then
            if operator and operator ~= "#" then
                return loc.format( string.format( "{1%s%s}", operator, rest ), cast_member:LocTable() )
            else
                return loc.format( "{1#agent}", cast_member )
            end

        elseif is_instance( cast_member, Location ) then
            return loc.format( "{1#location}", cast_member )
            
        elseif cast_member then
            if operator and rest then
                return loc.format( string.format( "{1%s%s}", operator, rest ), cast_member )
            else
                return tostring(cast_member)
            end
        end
    end

    return loc.custom_format( text, QuestFormatter )
end

-- Creates a table entry in 'markings' for each marked cast_member.
-- The table contains all objective_ids the cast_member is marked for.
function QuestState:CollectMarkedCast( markings, in_location )
    local qdef = self:GetQuestDef()
    if not qdef then
        return
    end

    if self:IsQuestHidden() then
        return
    end
    
    if self.status ~= QSTATUS.ACTIVE then
        if qdef:GetType() == QTYPE.SIDE and QuestUtil.CanTakeSideJob() then

            for i, cast_def in qdef:CastDefs() do
                local cast_member = self:GetCastMember( cast_def.cast_id )
                if cast_member then
                    if cast_def.provider then
                        -- Providers for inactive quests are marked.
                        markings[ cast_member ] = table.empty
                    end
                end
            end
        end

    else
        local quest_def = self:GetQuestDef()
        for _, obj in ipairs( self.objectives ) do
            if obj.state == QSTATUS.ACTIVE then
                local obj_def = quest_def:GetObjective( obj.id )
                if obj_def.mark then
                    local list = obj_def.mark

                    if type(obj_def.mark) == "function" then
                        local list = ObtainWorkTable()
                        obj_def.mark(self, list, in_location)
                        for _, agent in ipairs(list) do
                            local objective_ids = markings[ agent ]
                            if objective_ids == nil then
                                objective_ids = {}
                                markings[ agent ] = objective_ids
                            end
                            table.insert( objective_ids, obj.id )
                        end
                        ReleaseWorkTable(list)
                    else
                        for i, mark_id in ipairs( obj_def.mark ) do
                            local cast_member = self:GetCastMember( mark_id ) or self:GetParameter( mark_id )
                            if cast_member then
                                local objective_ids = markings[ cast_member ]
                                if objective_ids == nil then
                                    objective_ids = {}
                                    markings[ cast_member ] = objective_ids
                                end
                                table.insert( objective_ids, obj.id )
                            end
                        end
                    end
                end
            end
        end
    end
end

function QuestState:LookupQuip(tags, seed)
    local quest_def = self:GetQuestDef()
    if quest_def.quips then
        return quest_def.quips:LookupQuip( tags, seed )
    end
end


function QuestState:SetObjectiveState( objective_id, state )
    assert(IsEnum(state, QSTATUS))

    print ("QuestState:SetObjectiveState", self.content_id, objective_id, state)
    local quest_def = self:GetQuestDef()
    local obj_def = quest_def:GetObjective(objective_id)
    assert_warning( obj_def, "%s has no objective: '%s'", self:GetContentID(), tostring(objective_id))

    if obj_def then
        local obj = self:FindObjective(objective_id) or self:AddObjective(objective_id)
        if obj.state ~= state then
            --print("QuestState:SetObjectiveState()", objective_id, state)

            local last_state = obj.state
            obj.state = state

            if last_state == QSTATUS.ACTIVE then
                self:RunScriptHandler( obj_def.on_deactivate )
            end

            if obj.state == QSTATUS.ACTIVE then
                if self.status == QSTATUS.INACTIVE then
                    self:OnStart()
                end

                -- Mark all appropriate cast members.
                --SetMarks( self, obj_def, true )

                self:RunScriptHandler( obj_def.on_activate )

            elseif obj.state == QSTATUS.COMPLETE then
                assert_warning( last_state == QSTATUS.ACTIVE, "Objective %s not active", objective_id )
                self:RunScriptHandler( obj_def.on_complete )

            elseif obj.state == QSTATUS.FAILED then
                assert_warning( last_state == QSTATUS.ACTIVE, "Objective %s not active", objective_id )
                self:RunScriptHandler( obj_def.on_fail )

            elseif obj.state == QSTATUS.CANCELLED then
                if last_state == QSTATUS.ACTIVE then
                    self:RunScriptHandler( obj_def.on_cancel )
                else
                    print( string.format( "[%s] WARNING: Objective %s not active", self.content_id, objective_id ))
                    print( debug.traceback() )
                end
            end



            -- If activating an objective, automatically close any other active exclusive objectives.
            if state == QSTATUS.ACTIVE then
                for i, v in ipairs( self.objectives ) do
                    if obj ~= v and v.state == QSTATUS.ACTIVE and quest_def:GetObjective( v.id ).exclusive then
                        self:SetObjectiveState( v.id, QSTATUS.COMPLETE )
                    end
                end

            elseif state ~= QSTATUS.ACTIVE then
                -- Unmark all appropriate cast members.
                --SetMarks( self, obj_def, false )
                if obj_def.terminal and not self:IsDone() then
                    -- Terminal objective: finish quest with objective result.
                    self:OnFinish( state )
                end
            end

            -- dont' redundantly notify; being Done means we would have already
            if not self:IsDone() then
                self:NotifyChanged()
            end
        end
    end

    --the player did something of note to progress the plot. Advance the action clock
    TheGame:GetGameState():Action()
end

function QuestState:GetObjectiveState( objective_id )
    assert( objective_id )
    local obj = self:FindObjective(objective_id)
    if obj then
        return obj.state
    end
    return QSTATUS.INACTIVE
end


function QuestState:FindObjective( objective_id )
    for i,v in ipairs(self.objectives) do
        if v.id == objective_id then
            return v
        end
    end
end

function QuestState:AddObjective( objective_id )
    local obj = { id = objective_id, state = QSTATUS.INACTIVE, quest = self}
    table.insert(self.objectives, obj)
    return obj
end

function QuestState:GetQuestDef( )
    assert(self.content_id)
    return Content.GetQuestDef(self.content_id)
end

function QuestState:OnLocationEvent( eventname, location, ... )

    for cast_id, v in pairs(self.cast) do
        if v == location then
            local cast_def = self:GetQuestDef():GetCast( cast_id )
            local events = cast_def and cast_def.events and cast_def.events[ eventname ]
            if events then
                self:RunScriptHandler( events, location, ... )
            end
        end
    end

end



function QuestState:OnAgentEvent( eventname, agent, ... )
    if not is_instance( agent, Agent ) then
        return -- Agent events don't always include the agent as the first parameter :/
    end


    local cleanup = true
    for cast_id, v in pairs(self.cast) do
        if v == agent then
            local cast_def = self:GetQuestDef():GetCast( cast_id )
            local events = cast_def and cast_def.events and cast_def.events[ eventname ]
            if events then
                self:RunScriptHandler( events, agent, ... )
                cleanup = false

            elseif cast_def and cast_def.no_cleanup then
                cleanup = false
            end

            if not self:IsActive() and cast_def and (eventname == "agent_retired") and not cast_def.no_cleanup then
                self:Fail()
                return
            end
        end
    end
    
    if (eventname == "agent_retired") and cleanup then
        -- Do the default thing.  Which is to fail this entire quest.
        if agent == self:GetProvider() then
            self:Fail()
        end
    end
end


function QuestState:CleanupCastMembers()
    local def = self:GetQuestDef()
    local castdefs = def:GetCastDefs()

    -- Unassign in reverse order.
    for i = #castdefs, 1, -1 do
        local cast_id = castdefs[i].cast_id
        local cast_member = self.cast[ cast_id ]
        if cast_member then
            self:UnassignCastMember( cast_id, cast_member )
        end
    end
end

function QuestState:SetParameter( name, value )
    local def = self:GetQuestDef()
    assert_warning(def, "Setting parameter [%s] to quest [%s] with no definition", name, self.content_id)

    self.param[name] = value

    self:NotifyChanged() --er....
end

function QuestState:GetParameter( name )
    return self.param[ name ]
end

function QuestState:UpdateObjectiveState()
    local quest_def = self:GetQuestDef()


    for i,obj in ipairs(self.objectives) do
        if obj.state == QSTATUS.ACTIVE then

            local obj_def = quest_def:GetObjective( obj.id )
            if obj_def then
                self:RunScriptHandler( obj_def.on_update)
            end
        end
    end

    self:RunScriptHandler( quest_def.on_update )
end


function QuestState:OnGlobalEvent( event_name, ... )
    local def = self:GetQuestDef()
    if def.events[ event_name ] then
        self:RunScriptHandler( def.events[ event_name ], ... )
    end

    if self.objectives then
        local objectives = shallowcopy(self.objectives)
        for i,obj in ipairs(objectives) do
            if obj.state == QSTATUS.ACTIVE then
                if self:GetQuestDef():GetObjective(obj.id).events and self:GetQuestDef():GetObjective(obj.id).events[event_name] then
                   self:RunScriptHandler( self:GetQuestDef():GetObjective(obj.id).events[event_name], ... )
                end
            end
        end
    end
end


function QuestState:ListenForEvents()
    local def = self:GetQuestDef()
    local events = {}
    for event_name, handler in pairs( def.events ) do
        table.insert_unique(events, event_name)
    end

    if self:GetQuestDef().objectives then
        for k, def in pairs(self:GetQuestDef().objectives) do
            if def.events then
                for event_name, handler in pairs( def.events ) do
                    table.insert_unique(events, event_name)
                    
                end
            end
        end
    end

    for k, event_name in ipairs(events) do
        TheGame:GetEvents():ListenForEvent( event_name, self )
    end

end

function QuestState:AddJob(job)
    self.jobs = self.jobs or {}
    table.insert(self.jobs, job)
end

function QuestState:Load()
    self.param = self.param or {}
    self.objectives = self.objectives or {}
    self.jobs = self.jobs or {}
end

function QuestState:PostLoad()

    if self.rng then
        self.rng = engine.Random( self.rng[1], self.rng[2] )
    end

    self:ListenForEvents()

    for cast_id, cast in pairs( self.cast ) do
        if is_instance( cast, Agent ) then
           cast:ListenForAny( self, self.OnAgentEvent, self )
        elseif is_instance( cast, Location ) then
           cast:ListenForAny( self, self.OnLocationEvent, self )
        end
    end

    local def = self:GetQuestDef()

    for i = #self.objectives, 1, -1 do
        local obj = self.objectives[i]
        if not def:CheckObjective( obj.id ) then
            print( "## Sanitizing missing objective", self.content_id, obj.id )
            table.remove( self.objectives, i )
        end
    end

    self:RunScriptHandler( def.on_post_load )
end


function QuestState:Save()
    local data =
    {
        id = self.id,
        valid = self.valid,
        content_id = self.content_id,
        status = self.status,
        qrank = self.qrank,
        is_hidden = self.is_hidden,
        quest_provider = self.quest_provider,
        objectives = next(self.objectives) and self.objectives,
        param = next(self.param) and self.param,
        jobs  = next(self.jobs) and self.jobs,
        cast = self.cast,
        hide_in_overlay = self.hide_in_overlay,
        extra_reward = self.extra_reward,
        extra_reward_data = self.extra_reward_data,
        upfront_reward = self.upfront_reward,
        
        rewarded = self.rewarded,
        clean_up_locations = self.clean_up_locations,
        reward_random = self.reward_random,
    }

    if self.rng then
        data.rng = { self.rng:GetSeeds() }
    end

    return data
end

function QuestState:__tostring()
    return string.format( "[%d - %s]", self.id, self.content_id )
end


-------------------REWARD STUFF


--our relationship may have changed since we generated the rewards.
function QuestState:VerifyRewards( )
    self:GenerateExtraReward()
    self:GenerateMoneyRewards()
end

function QuestState:GenerateMoneyRewards( )

    if (self:GetQuestDef():GetType() == QTYPE.CONTRACT or self:GetQuestDef():GetType() == QTYPE.SIDE) then
        local reward, rel_modifier, giver = QuestUtil.CalculateReward( self, self:GetParameter("reward_negotiated_bonus") or 0)
        self:SetParameter( "rewards", reward )
        self:SetParameter( "reward_relationship", giver and giver:GetRelationship() or RELATIONSHIP.NEUTRAL )
        self:SetParameter( "reward_modifier_relationship", rel_modifier )
     end
end



function QuestState:GetExtraReward()
    return self.extra_reward, self.extra_reward_data
end


function QuestState:GetFocus()
    return self:GetQuestDef().focus or QUEST_FOCUS.BALANCED
end




function QuestState:GenerateExtraReward()
    if not self.extra_reward then
        local current_difficulty = TheGame:GetGameState():GetCurrentBaseDifficulty()

        if self:GetType() == QTYPE.SIDE and self:GetRank() > current_difficulty then
            self.extra_reward  = EXTRA_QUEST_REWARD.GRAFT_CHOICE

        elseif self:GetType() == QTYPE.CONTRACT and (self:GetRewardSize() ~= QREWARDSIZE.SMALLER or self:GetRank() > current_difficulty) then
            self.extra_reward  = EXTRA_QUEST_REWARD.GRAFT_CHOICE

        elseif self:GetType() == QTYPE.SIDE or self:GetType() == QTYPE.CONTRACT then
            self.extra_reward, self.extra_reward_data = QuestUtil.GenerateExtraReward( self:GetRank() )
        end
    end
end


function QuestState:AddScore(id, value)
    local score_def = self:GetQuestDef():GetScoreDef(id) 
    if score_def then
        TheGame:GetGameState():AddScore(score_def.id, value)
    else
        TheGame:GetGameState():AddScore(id, value)
    end

    return self
end


function QuestState:OpinionEvent(cast_id, opinion_id)
    
    local who = type(cast_id) == "string" and self:GetCastMember(cast_id) or cast_id
    if who then
        local op = self:GetQuestDef():GetOpinionEvent(opinion_id)
        if op then
            who:OpinionEvent(op)
        else
            LOGWARN( "Missing opinion '%s'", opinion_id )
        end
    end
end

function QuestState:GetMoneyReward()
    local reward = self:GetParameter("rewards")
    return reward or 0 
end

function QuestState:SpawnFollowQuest(id)
    local qdef = Content.GetQuestDef(id)
    assert(qdef and qdef.parent_quest == self.content_id, "Invalid folloup quest")
    local overrides = {cast = {}, qrank = self:GetRank(), parameters = {}}
    for id, agent in pairs(self.cast) do
        overrides.cast[id] = agent
    end
    

    if self.param then
        for k,v in pairs(self.param) do
            overrides.parameters[k] = v
        end
    end

    return QuestUtil.SpawnQuest(id, overrides)
end

function QuestState:GetCompletionString()
    if self:IsFailed() and self:GetQuestDef().get_fail_string then
        return self:FormatString( self:GetQuestDef().get_fail_string(self))
    elseif self:IsComplete() and self:GetQuestDef().get_complete_string then
        return self:FormatString( self:GetQuestDef().get_complete_string(self))
    end
    return ""
end

function QuestState:IsReadyForTurnIn()
    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                
                if self:GetQuestDef():GetObjective(obj.id).ready_for_completion then
                    return true
                end
            end
        end
    end
end

function QuestState:IsHighPriority()
    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                
                if self:GetQuestDef():GetObjective(obj.id).high_priority then
                    return true
                end
            end
        end
    end
end

function QuestState:SpawnTempLocation(id, cast_id)
    
    local def = self:GetQuestDef():GetLocationDef(id) or Content.GetLocationContent(id)
    if not def then
        assert(false, "DEF NOT FOUND " .. id )
    end

    local location = Location( def.id )
    assert(location)
    TheGame:GetGameState():AddLocation(location)
    if cast_id then
        self:AssignCastMember(cast_id, location )
    end
    table.insert( self.clean_up_locations, location )
    return location
end

function QuestState:GetFightSupport(agent, target_agent)
    if self:GetQuestDef().get_fight_support then
        return self:GetQuestDef().get_fight_support(self, agent, target_agent)
    end
end

function QuestState:GetNegotiationSupport(agent, target_agent, is_hostile)
    if self:GetQuestDef().get_negotiation_support then
        return self:GetQuestDef().get_negotiation_support(self, agent, target_agent, is_hostile)
    end
end


function QuestState:IsNegotiationTarget(agent)
    if self:GetQuestDef().is_negotiation_target then
        if self:GetQuestDef().is_negotiation_target(self, agent) then
            return true
        end
    end

    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                local obj_def = self:GetQuestDef():GetObjective( obj.id )             
                
                if obj_def.is_negotiation_target then
                    if obj_def.is_negotiation_target(self, agent) then
                        return true
                    end
                end

                if obj_def.negotiation_targets then
                    for k,v in ipairs(obj_def.negotiation_targets) do
                        if self:GetCastMember(v) == agent then
                            return true
                        end
                    end
                end
            end
        end
    end

    return false
end


function QuestState:IsInHiding(agent)
    if self:GetQuestDef().is_in_hiding then
        if self:GetQuestDef().is_in_hiding(self, agent) then
            return true
        end
    end
    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                local obj_def = self:GetQuestDef():GetObjective( obj.id )             
                if obj_def.is_in_hiding then
                    if obj_def.is_in_hiding(self, agent) then
                        return true
                    end
                end
            end
        end
    end

    return false
end


function QuestState:IsCombatTarget(agent)
    local combat
    if self:GetQuestDef().is_combat_target then
        if self:GetQuestDef().is_combat_target(self, agent) then
            return true
        end
    end

    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE then
                local obj_def = self:GetQuestDef():GetObjective( obj.id )             
                
                if obj_def.is_combat_target then
                    if obj_def.is_combat_target(self, agent) then
                        return true
                    end
                end

                if obj_def.combat_targets then
                    for k,v in ipairs(obj_def.combat_targets) do
                        if self:GetCastMember(v) == agent then
                            return true
                        end
                    end
                end

            end
        end
    end

    return false
end

function QuestState:GetLocalizedStr(str_id)
    local qdef = self:GetQuestDef()
    local loc_key = qdef:GetLocKey(str_id)
    local str = LOC(loc_key)
    local result = self:FormatString(str)
    return result
end


function QuestState:GetQuestLocationForAgent(agent)
    local location, role
    if not self:IsActive() then
        return
    end
    

    local t = ObtainWorkTable()
    if self:GetQuestDef().collect_agent_locations then
        self:GetQuestDef().collect_agent_locations(self, t)
    end

    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE and self:GetQuestDef():GetObjective(obj.id).collect_agent_locations then
                self:GetQuestDef():GetObjective(obj.id).collect_agent_locations(self, t)
            end
        end
    end

    for k,v in ipairs(t) do
        if v.agent == agent then
            location = v.location
            role = v.role
            break
        end
    end


    ReleaseWorkTable(t)

    return location, role
end

function QuestState:GetQuestAgentsForLocation(location, agents)
    if not self:IsActive() then
        return
    end
    local t = ObtainWorkTable()

    if self:GetQuestDef().collect_agent_locations then
        self:GetQuestDef().collect_agent_locations(self, t)
    end

    if self.objectives then
        for i,obj in ipairs(self.objectives) do
            if obj.state == QSTATUS.ACTIVE and self:GetQuestDef():GetObjective(obj.id).collect_agent_locations then
                self:GetQuestDef():GetObjective(obj.id).collect_agent_locations(self, t)
            end
        end
    end

    for k,v in ipairs(t) do
        assert_warning( v.agent, tostring(self))
        if v.location == location and v.agent and not v.agent:IsRetired() then
            agents[v.agent] = v.role
        end
    end

    ReleaseWorkTable(t)
end

function QuestState:IsCastIDBarredFromLocation(cast_id, location)
    
    if self:GetQuestDef().avoids then
        for _,data in ipairs(self:GetQuestDef().avoids) do
            
            if cast_id == data[1] or cast_id == data[2] then
                local other_cast_id = cast_id == data[1] and data[2] or data[1]
                local other_member = self:GetCastMember(other_cast_id)
                
                local check_location
                if is_instance(other_member, Agent) then
                    if other_member:GetBrain() and other_member:GetBrain():GetWorkPosition() and other_member:GetBrain():GetWorkPosition():ShouldBeWorking() then
                        check_location = other_member:GetBrain():GetWorkPosition():GetLocation()
                    else
                        check_location = other_member:GetLocation()
                    end
                elseif is_instance(other_member, Location) then
                    check_location = other_member
                end
                
                
                if check_location and check_location == location then
                    if data[3] and type(data[3]) == "table" then
                        for _, obj_id in pairs(data[3]) do
                            if self:IsActive(obj_id) then
                                return true
                            end
                        end
                    elseif data[3] and type(data[3] == "function") then
                        return data[3](self)
                    else
                        
                        return true --all objectives!
                    end
                end
            end
        end
    end
end

function QuestState:IsAgentBarredFromLocation(agent, location)
    if self:GetQuestDef().avoids_fn then
        if self:GetQuestDef().avoids_fn(self, agent, location) then
            return true
        end
    end

    if self:GetQuestDef().avoids then
        for cast_id, cast_member in pairs( self.cast ) do
            if agent == cast_member then
                if self:IsCastIDBarredFromLocation(cast_id, location) then
                    return true
                end
            end
        end
    end
    return false
end


--returns %complete, title, subtitle
function QuestState:GetNarrativeProgress()
    if self:GetQuestDef().get_narrative_progress then
        return self:GetQuestDef().get_narrative_progress(self)
    end
end

function QuestState:GetRatingString()
    local focus  = self:GetFocus()
    if focus == QUEST_FOCUS.COMBAT then
        return LOC"UI.JOB_POPUP.FOCUS_COMBAT", UICOLOURS.FIGHT, LOC"UI.JOB_POPUP.FOCUS_COMBAT_TT"
    elseif focus == QUEST_FOCUS.NEGOTIATION then
        return LOC"UI.JOB_POPUP.FOCUS_NEGOTIATION", UICOLOURS.NEGOTIATION, LOC"UI.JOB_POPUP.FOCUS_NEGOTIATION_TT"
    else
        return LOC"UI.JOB_POPUP.FOCUS_MIXED", UICOLOURS.TITLE, LOC"UI.JOB_POPUP.FOCUS_MIXED_TT"
    end
end


function QuestState:GetMarkData(objectives)
    if objectives then
        for i,obj in ipairs(self.objectives) do
            if self:IsActive(obj.id) and table.arrayfind(objectives, obj.id) then
                local obj_def = self:GetQuestDef():GetObjective( obj.id )
                if obj_def.icon then
                    return obj_def.icon
                end
            end
        end
    end

    return self:GetIcon()
end

function QuestState:DefFn(fn, ...)
    return self:GetQuestDef()[fn](self, ...)
end

