require "quest"
require "agentutil"

--------------------------------------------------------------------------------
-- These are helpers for instantiating job types.

local MAX_QUESTS = 6 -- Maximum non-story quests allowed.

local QuestUtil = class( "QuestUtil" )


function QuestUtil.CalculateRewardValue( quest )
    
    local main_quest = TheGame:GetGameState():GetMainQuest()
    if main_quest and main_quest:GetQuestDef().quest_reward_fn then
        return main_quest:GetQuestDef().quest_reward_fn(main_quest, quest)
    end

    return 100
end


-- A positive-valued cost function with domain [1, QUEST_MAX_RANK].
-- Should correlate roughly with CalculateWealthFromRenown( 1... 10 )
function QuestUtil.CalculateReward( quest, bonus_amt )
    
    local qtype = quest:GetType()

    local giver = quest:GetProvider() or quest:GetCastMember("giver") or quest:GetCastMember("handler")
    local rel = giver and giver:GetRelationship() or RELATIONSHIP.NEUTRAL
    local value_mod = OPINION_REWARD_VALUE_MOD[rel]
   
    local reward = QuestUtil.CalculateRewardValue( quest )

    local value = math.round( reward * value_mod * (quest:GetQuestDef().reward_mod or 1) + (bonus_amt or 0))
    return value, value_mod, giver
end

function QuestUtil.GenerateExtraReward( rank )
    rank = rank or TheGame:GetGameState():GetCurrentBaseDifficulty()
    
    local active_extras = {}
    for i, quest in TheGame:GetGameState():ActiveQuests() do
        if quest:GetType() == QTYPE.SIDE and quest.extra_reward then
            active_extras[quest.extra_reward] = true
        end
    end

    local t = {}
    for k in pairs(EXTRA_QUEST_REWARD) do
        if k ~= EXTRA_QUEST_REWARD.NONE and k ~= EXTRA_QUEST_REWARD.GRAFT_CHOICE and k ~= EXTRA_QUEST_REWARD.METTLE_POINTS and not active_extras[k] then
            table.insert(t, k)
        end
    end
    
    if TheGame:GetGameState():CanDoMettle() then
        table.insert(t, EXTRA_QUEST_REWARD.METTLE_POINTS)
    end
    
    local extra_reward = table.arraypick(t)
    
    if extra_reward == EXTRA_QUEST_REWARD.EXTRA_CASH then
        return extra_reward, rank * 30
    
    elseif extra_reward == EXTRA_QUEST_REWARD.FREE_ITEM then
        local t = {}
        local def1 = BattleCardCollection.AllLocalItems():Pick(1)
        if def1 and def1[1] then
            table.insert(t, def1[1].id)
        end
        local def2 = NegotiationCardCollection.AllLocalItems():Pick(1)
        if def2 and def2[1] then
            table.insert(t, def2[1].id)
        end
        return extra_reward, table.arraypick(t) or "portable_bunker"
 
    elseif extra_reward == EXTRA_QUEST_REWARD.RESTORATION then
        return extra_reward, 15
    elseif extra_reward == EXTRA_QUEST_REWARD.METTLE_POINTS then
        return extra_reward, rank * 4
    else
        return extra_reward
    end
end


function QuestUtil.GetStatusColour( state )
    if state == QSTATUS.INACTIVE then
        return 0xAAAAAAAA
    elseif state == QSTATUS.COMPLETE then
        return 0x228B22FF
    elseif state == QSTATUS.FAILED then
        return 0x8B0000FF
    else
        return 0xFFFFFFFF
    end
end

function QuestUtil.CountQuests( agent, fn )
    local count = 0
    for i, quest in ipairs( agent:GetQuestMembership() ) do
        if fn == nil or fn( quest ) then
            count = count + 1
        end
    end
    return count
end




function QuestUtil.GetQuestSpawnPenalty( quest_id )
    return TheGame:GetGameState():GetQuestActivatedCount(quest_id) 
end

function QuestUtil.SpawnInactiveQuest(quest_id, overrides)
    local ok, result, quest_state = QuestUtil.TrySpawnQuest( quest_id, overrides )
        
    if ok and result then
        return quest_state
    end

    return nil, quest_state
end

function QuestUtil.TrySpawnQuest( quest_id, overrides, try_force )
    engine.inst:ProfilerPush("QuestUtil.TrySpawnQuest")
    local quest_def = Content.GetQuestDef( quest_id )
    if not quest_def then
        assert(false, "MISSING QUEST: " .. tostring(quest_id))
    end

    local ok, result, quest_state = xpcall( quest_def.CreateQuest, generic_error, quest_def, overrides, try_force )
    engine.inst:ProfilerPop()
    return ok, result, quest_state
end

function QuestUtil.SpawnQuest( quest_id, overrides, try_force )
    local ok, result, quest_state = QuestUtil.TrySpawnQuest( quest_id, overrides, try_force )

    if not ok then
        LOGWARN( "[%s] failed spawn:\n%s", quest_id, tostring(result) )
        return nil, quest_state
    elseif result then
        TheGame:GetGameState():AddActiveQuest( quest_state )
        return quest_state
    else
        return nil, quest_state
    end
end



function QuestUtil.CalcQuestSpawnScore(category, history_length, quest_id)
    if TheGame:GetGameState():GetQuestActivatedCount(quest_id) > 0 then
        return 10
    else
        if TheGame:GetGameProfile():HasEverSeenIncident(category, quest_id) then
            return TheGame:GetGameProfile():HasSeenIncidentRecently(category, quest_id, history_length) and 2 or 1
        else
            return 0
        end
    end
end




function QuestUtil.SpawnEventQuest( event_trigger, overrides, event_type )
    if event_type == "" then
        event_type = nil
    end

    local act_id = TheGame:GetGameState():GetCurrentAct().id
    local rank = TheGame:GetGameState():GetCurrentBaseDifficulty() 


    
    local attempt_quests = ObtainWorkTable()
    local quest_scores = ObtainWorkTable()

    for id, qdef in pairs( Content.GetAllQuests() ) do
        local valid = qdef:GetType() == QTYPE.EVENT and CheckBits( qdef:GetSpawnEventMask(), event_trigger) and event_type == qdef.event_type and qdef:FilterForAct( act_id )
        if valid then
            table.insert( attempt_quests, id)
        end
    end

    local event_id = act_id .. "_EVENT_SPAWN"
    for k,v in ipairs(attempt_quests) do
        quest_scores[v] = QuestUtil.CalcQuestSpawnScore(event_id, 10, v)
    end
    
    table.shuffle(attempt_quests)
    table.stable_sort(attempt_quests, 
        function(a,b) 
            if quest_scores[a] == quest_scores[b] then
                return Content.GetQuestDef(a):GetPriority() > Content.GetQuestDef(b):GetPriority()
            else
                return quest_scores[a] < quest_scores[b]    
            end
        end)


    local quest_state
    for _, quest_id in ipairs(attempt_quests) do
        quest_state = QuestUtil.SpawnQuest( quest_id, overrides ) 
        if quest_state then
            break
        end
    end

    ReleaseWorkTable( quest_scores )
    ReleaseWorkTable( attempt_quests )
    
    if quest_state then
        TheGame:GetGameProfile():RecordIncident(event_id, quest_state:GetContentID())
    end

    return quest_state

end


local function DefaultQuestSort( a, b )
    if a:GetType() ~= b:GetType() then
        return a:GetType() == QTYPE.STORY
    end
    return a:GetTitle() < b:GetTitle()
end

function QuestUtil.CountQuestActive(quest_id, duration)
    local count = 0
    if duration == nil then
        -- Count quests active RIGHT NOW
        for i, quest in TheGame:GetGameState():ActiveQuests() do
            if quest:GetContentID() == quest_id then
                count = count + 1
            end
        end
    else
        -- Count quests active since 'duration' ago.
        local since = math.max(0, Now() - duration)
        local quest_stats = TheGame:GetGameState():GetQuestStats( quest_id )

        if quest_stats and quest_stats.activated_timestamps then
            for i, datetime in ipairs( quest_stats.activated_timestamps ) do
                if datetime >= since then
                    count = count + 1
                end
            end
        end
    end

    return count
end


function QuestUtil.CreateQuestPhaseAttract(args)
    QuestConvo(args.quest_id, args.obj_id, args.cast_id)
        :Attract(function(cxt, agent)
            return "QUEST_PHASE_ATTRACT"
        end)
        :State("QUEST_PHASE_ATTRACT")
            :Loc{
                DIAL_TXT = args.txt
            }
            :Fn(function(cxt)
                cxt:Dialog("DIAL_TXT")
                if args.fn then
                    args.fn(cxt)
                end
            end)
end

--these are intentionally too broad until we get more specific def quests
function QuestUtil.CanDoPoorCivvieQuest(agent) return agent.def.id == "LABORER" or agent.def.id == "HEAVY_LABORER" end
function QuestUtil.CanDoRichCivvieQuest(agent) return agent.def.id == "WEALTHY_MERCHANT" end
function QuestUtil.CanDoAdmiraltyGoonQuest(agent) return agent.def.id == "ADMIRALTY_GOON" or agent.def.id == "ADMIRALTY_GUARD" end
function QuestUtil.CanDoJakesGoonQuest(agent) return agent.def.id == "JAKES_RUNNER" or agent.def.id == "JAKES_SMUGGLER" end
function QuestUtil.CanDoBanditGoonQuest(agent) return agent.def.id == "BANDIT_GOON" or agent.def.id == "BANDIT_GOON2" end

function QuestUtil.FilterPromoted( def )
    local p = GetAdvancementModifier( ADVANCEMENT_OPTION.NPC_PROMOTION_CHANCE )
    if p <= 0 then
        local base_def = def.base_def and Content.GetCharacterDef( def.base_def )
        if base_def and base_def.promotion_def == def.id then
            -- This is a promoted def, but is not permitted at this advancement.
            return false
        end
    end

    return true
end


--assume that you can either fight or negotiate, unless something 
function QuestUtil.FilterAgentForDifficulty(rank, agent, can_fight, can_negotiate)

    local target_renown, combat_strength = agent:GetRenown(), agent:GetCombatStrength()
    if not can_fight and not can_negotiate then
        return false, "not allowed to do anything"
    end

    local renown_diff = target_renown - rank
    local combat_diff = combat_strength - rank

    local can_win_a_fight = combat_diff <= 1
    local can_win_a_negotiation = renown_diff <= 1

    if renown_diff >= 2 then
        return false, "Too high renown"
    end
    
    if combat_diff >= 2 then
        return false, "Too high combat"
    end

    if combat_diff >= 1 and renown_diff >= 1 then
        return false, "Generally too high"
    end

    if combat_diff <= -1 and renown_diff <= -1 then
        return false, "Generally too low"
    end


    
    if not can_negotiate then
        --we have to fight. Can we do it?
        if combat_diff > 1 then
            return false, "Too hard to fight"
        end

        if combat_diff <-1 then
            return false, "Too easy to fight"
        end

        return true, "Can Fight"
    else


        --it's probably too hard
        if renown_diff >= 1 then
            --can I just beat them up?
            if can_fight then
                if combat_diff > 1 then
                    return false, "can't negotiate, too hard to intimidate"
                end
                return true, "Can Intimidate, then negotiate"
            else
                return false, "Too high renown to negotiate, not allowed to intimidate"
            end
        elseif renown_diff < -1 then
            return false, "Too easy to negotiate"
        else
            return true, "Can negotiate"
        end
    end

end

function QuestUtil.CountSideJobs()
    local num_side_jobs = 0

    for i, quest in TheGame:GetGameState():ActiveQuests() do
        if quest:IsActive() and (quest:GetType() == QTYPE.CONTRACT or quest:GetType() == QTYPE.SIDE) then
            num_side_jobs = num_side_jobs + 1
        end

    end

    return num_side_jobs
end

function QuestUtil.CanTakeSideJob()
    
    if not TheGame:GetGameState() then
        return false
    end

    for i, quest in TheGame:GetGameState():ActiveQuests() do
        if quest:BlocksSideJobs() then
            return false, "Can't do a side job right now"
        end
    end

    if QuestUtil.CountSideJobs() >= 1 then
        return false, "You have too many active jobs"
    end

    return true
end


function QuestUtil.GetValidQuests( quests )
    -- returns the quests that can actually be spawned
    local valid_quests = {}

    for _, id in ipairs(quests) do
        local ok, result, quest_state = QuestUtil.TrySpawnQuest(id)
        if ok then
            table.insert(valid_quests, id)
        else
            print(ok, result, quest_state)
            print(loc.format("{1} is not valid {2}", id, tostr(quest_state.trace, 2)))
        end
    end

    return valid_quests
end




function QuestUtil.TrySpawnSideJob(rank_bias, quest_def_filter)
    --print ("TRY SPAWN SIDES JOBS", #list_of_agents, num_to_spawn, debugstack())
    
    local diff = TheGame:GetGameState():GetCurrentBaseDifficulty() 
    local rank = clamp( math.round(diff) + (rank_bias or 0), 1, 5)
    local now = Now()

    local act_id = TheGame:GetGameState():GetCurrentAct().id

    local jobs = {}
    local scores = {}

    for id, def in pairs( Content.GetAllQuests() ) do
        if def.qtype == QTYPE.SIDE and not def:HasTag("manual_spawn") and (quest_def_filter == nil or quest_def_filter(def)) then
            local defmin, defmax = def:GetRankRange()
            if rank >= defmin and rank <= defmax then
                if def:FilterForAct( act_id ) then
                    table.insert(jobs, def.id)
                end
            end
        end
    end


    local event_id = act_id .. "_SIDE_JOB_SPAWN" 
    for k,v in ipairs(jobs) do
        scores[v] = QuestUtil.CalcQuestSpawnScore(event_id, math.floor(#jobs), v)
    end
    print (tostr(scores))
    
    table.shuffle(jobs)
    table.stable_sort(jobs, function(a,b) return scores[a] < scores[b] end )

    for _, jobid in ipairs(jobs) do
        local overrides = { qrank = rank }
        local quest, err = QuestUtil.SpawnQuest( jobid, overrides ) 
        if quest then
            TheGame:GetGameProfile():RecordIncident(event_id, jobid)
            local agent = quest:GetCastMember( "giver" )
            return quest
        else
            print ("FAIL TO SPAWN", jobid)
        end
    end
end



function QuestUtil.AddSideJobChain(QDEF, num_jobs_to_spawn, num_jobs_to_do, end_objective, spawn_oppos, give_starting_bonus, quest_def_filter)


    QDEF:AddObjective{
        id = "pick_a_side_job",
        --desc = [[{fssh} knows of some people looking for help in Havaria. Doing so is a good way to establish yourself here.]],
        title = "Pick a job",
        desc = "Gain money and notoriety by doing jobs for the denizens of Havaria",
        on_activate = function(quest) 
            quest.param.job_options = quest.param.job_options or {}
            quest.param.side_jobs = quest.param.side_jobs or {}

            quest.param.current_job = nil
            local to_spawn = math.max(0, num_jobs_to_spawn - #quest.param.job_options)
            for k = 1, to_spawn do
                
                --this is super ugly
                local bias = 0
                if #quest.param.side_jobs > 0 or num_jobs_to_do == 1 then
                    if k == 1 then
                        bias = 1
                    end
                end

                local new_quest = QuestUtil.TrySpawnSideJob(bias, quest_def_filter)
                new_quest.upfront_reward = give_starting_bonus 
                table.insert( quest.param.job_options, new_quest )
            end
            
            for k, q in ipairs(quest.param.job_options) do
                q:SetQuestVisible(true)
                --local cast_name = "giver" .. k
                --quest:UnassignCastMember(cast_name)
                --quest:AssignCastMember(cast_name, q:GetProvider())    
            end
            TheGame:GetGameState():AddPendingEvent()

            for k,v in pairs(quest.param.job_options) do
                if quest:GetProvider() and quest:GetProvider():GetBrain() then
                    quest:GetProvider():GetBrain():SendToWork()
                end
            end

        end,

        on_complete = function(quest)
            quest:Activate("do_a_job")
        end
    }
    QDEF:AddObjective{
        id = "do_a_job",
        hide_in_overlay = true,
        on_complete = function(quest)
            if #quest.param.side_jobs >= num_jobs_to_do then
                quest:Activate(end_objective)

                for k,v in pairs( quest.param.job_options) do
                    if v:IsInactive() then
                        v:Cancel()    
                    end
                end
            else
                if spawn_oppos then
                    QuestUtil.SpawnOpportunities()
                end
                quest:Activate("pick_a_side_job")
            end
        end
    }

    --[[for k = 1, num_jobs_to_spawn do
        QDEF:AddCast{
            cast_id = "giver" .. k,
            when = QWHEN.MANUAL,
            no_validation = true,
            no_cleanup = true,
        }
    end--]]
     
    local on_quest_changed = function(quest, event_quest) 
        if quest:IsActive("do_a_job") then
            if quest.param.current_job == event_quest then
                if (event_quest:IsComplete() or event_quest:IsFailed()) then
                    quest:Complete("do_a_job")
                    quest.param.current_job = nil
                end
            end

        elseif event_quest:IsActive() then
            local idx = quest.param.job_options and table.arrayfind(quest.param.job_options, event_quest)
            if idx then
                if event_quest:IsActive() then
                    
                    local rewards = { MONEY = event_quest.param.rewards, GRAFT = event_quest.graft_rewards and true or false }
                    local json_t = {
                        PLAYER_DATA = TheGame:GetGameState():GetPlayerState(),
                        CONTENTID = event_quest.content_id,
                        DIFFICULTY = event_quest:GetDifficulty(),
                        REWARDS = rewards,
                    }
                    SendMetricsData("QUEST_ACCEPTED", json_t )

                    table.remove(quest.param.job_options, idx)

                    local to_cancel = shallowcopy(quest.param.job_options)

                    for k,v in ipairs(to_cancel) do
                        v:Cancel()
                    end
                    table.clear(quest.param.job_options)
                    quest.param.current_job = event_quest
                    quest:Complete("pick_a_side_job")
                    
                    quest.param.side_jobs = quest.param.side_jobs or {}
                    table.insert(quest.param.side_jobs, event_quest )                    

                end
            end
        end        
    end

    QDEF.events = QDEF.events or {}
    if QDEF.events.quests_changed then
        local old = QDEF.events.quests_changed
        QDEF.events.quests_changed = function(...) old(...) on_quest_changed(...) end
    else
        QDEF.events.quests_changed = on_quest_changed
    end
end



function QuestUtil.SpawnOpportunities(num_to_spawn)
    num_to_spawn = num_to_spawn or 2

    TheGame:GetGameState():ClearOpportunities()

    local possible_spawns = {}
    local now = Now()
    local act_id = TheGame:GetGameState():GetCurrentAct().id
    for id, def in pairs( Content.GetAllQuests() ) do
        if def:GetType() == QTYPE.OPPORTUNITY and id ~= "OPPO_FRIEND_BOON" and def:FilterForAct( act_id ) and not def:HasTag("manual_spawn") then
            local stats = TheGame:GetGameState():GetQuestStats( def.id )
            local time_stamp = stats and stats.activated_timestamps and stats.activated_timestamps[#stats.activated_timestamps]
            local score = 0
            if time_stamp then
                score = math.max( 100 - (now - time_stamp), 0)
            end
            table.insert(possible_spawns, {def = def, score = score})
        end
    end
    
    table.shuffle(possible_spawns)
    table.stable_sort(possible_spawns, function(a,b) return a.score < b.score end )

    local num_spawned = 0
    local friend = QuestUtil.SpawnQuest("OPPO_FRIEND_BOON")
    if friend then
        num_spawned = num_spawned + 1
    end

    for _, jobinfo in ipairs(possible_spawns) do
        
        if num_spawned >= num_to_spawn then
            break
        end
       
        local quest, err = QuestUtil.SpawnQuest( jobinfo.def.id ) 
        if quest then
            num_spawned = num_spawned + 1
        end
    end

end

ERR_QUEST = "!"
function QuestUtil.StartDayQuests(day_schedule, quest) 
    quest.param.all_day_quests = quest.param.all_day_quests or {}        
    for k,schedule in ipairs(day_schedule) do
        if TheGame:GetGameState():GetQuestActivatedCount( schedule.quest ) == 0 then
            local day_quest, err = QuestUtil.SpawnQuest( schedule.quest, {qrank = schedule.difficulty, parameters = {main_quest = quest, home_base = quest.param.home_base }})
            if err then
                ERR_QUEST = err
            end
            assert_warning(day_quest, tostring(err))
            table.insert(quest.param.all_day_quests, day_quest)
        end
    end
end

function QuestUtil.DoNextDay(schedule, quest, day)
    quest.param.day = day or ( quest.param.day and quest.param.day + 1 or 1 )
    if day then
        for k = 1, day-1 do
            if quest.param.all_day_quests[k] and quest.param.all_day_quests[k]:IsInactive() then
                quest.param.all_day_quests[k]:Complete()
            end
        end
    end
    
    local diff = schedule[quest.param.day] and schedule[quest.param.day].difficulty
    TheGame:GetGameState():SetDifficulty(diff)
    TheGame:GetGameState():SetActProgress( quest.param.day )
    
    for k, v in ipairs(quest.param.all_day_quests) do
        if v:IsActive() then
            v:Cancel()
        end
    end
    
    TheGame:GetGameState():AddScore("PROGRESS", quest.param.day-1)

    if quest.param.all_day_quests[quest.param.day] then
        quest.param.all_day_quests[quest.param.day]:Activate()
        quest.param.day_quest = quest.param.all_day_quests[quest.param.day]
    end

    local json_t = {
        PLAYER_DATA = TheGame:GetGameState():GetPlayerState(),
    }
    SendMetricsData("DAY_START", json_t )
    
    TheGame:GetGameState():TryRollbackSave(quest.param.day)
end


function QuestUtil.PresentJobChoice(cxt, quest_options, force_take, on_picked_fn)
    for k,v in ipairs(quest_options) do
        v:SetHideInOverlay(true)
    end

    if #quest_options > 0 then 
        cxt:RunLoopingFn( function()
            for k,job in ipairs(quest_options) do
                
                cxt:QuestOpt( job )
                    :ShowQuestAsInactive()
                    :Fn(function(cxt) 
                        StateGraphUtil.PresentQuestOffer(cxt, job, nil, 
                            function() 

                                for k,v in pairs(quest_options) do
                                    if v ~= job then
                                        v:Cancel()
                                    end
                                end
                                cxt:PlayQuestConvo(job, QUEST_CONVO_HOOK.ACCEPTED)
                                cxt:Pop()
                                job:SetHideInOverlay(false)
                                if on_picked_fn then
                                    on_picked_fn(cxt, quest_options, job)
                                end
                            end,
                            function()
                                if not cxt:PlayQuestConvo(job, QUEST_CONVO_HOOK.DECLINED) then
                                    cxt:Quip( cxt:GetAgent(), "job_decline")
                                end
                            end,
                            #quest_options == 1)
                    end)
            end

            if not force_take then
                StateGraphUtil.AddBackButton(cxt)
            end
        end )
    end
end


function QuestUtil.GetDifficultyFromPrestige( prestige )
    prestige = prestige or TheGame:GetGameState():GetAdvancementLevel()
    for difficulty, difficulty_data in ipairs(DAILY_DIFFICULTY_OPTIONS) do
        if difficulty_data.prestige == prestige then return difficulty end
    end
    return 1
end

function QuestUtil.RunAwayPet(agent)
    agent:Dismiss()
    agent:MoveToLimbo()
    local overrides = { cast = { pet = agent } }
    local q = QuestUtil.SpawnQuest("STORY_PET_RETURN", overrides)
    q:Activate()
end
