local lyaml = require "lyaml"

-------------------------------------------------------------------------------
-- A system for looking up randomized quips by a set of criterion tags.

local QuipDatabase = class( "QuipDatabase" )

function QuipDatabase:init( loc_prefix )
    self.db = {}
    self.recent_lookups = {}
    self.tag_scores = {}
    self.stats = {}
    self.loc_prefix = loc_prefix or "QUIPS"
    --self.files = {}
end

function QuipDatabase:SetModID( mod_id )
    self.mod_id = mod_id
end

function QuipDatabase:GetModID()
    return self.mod_id
end

function QuipDatabase:AddFilename( filename )
    local id = filename:match( "([%w_]+)[.]yaml$" )

    local text = engine.inst:LoadTextFile( filename )
    assert( text, filename )
    
    local ok, yaml = pcall( lyaml.load, text )
    if not ok then
        LOGWARN( "Failed YAML parsing: %s\n%s", filename, tostring(yaml))
    end

    if ok then
        -- Just for debugging.
        if self.files then
            self.files[id] = yaml
        end
        local ok, result = pcall( self.ParseYAML, self, yaml )
        if not ok then
            LOGWARN( "Failed parsing: '%s'\n%s", filename, result )
        end
    end
end

function QuipDatabase:ParseYAML( yaml )
    self:AddLines( yaml.lines )

    if yaml.tag_scores then
        for tag, score in pairs (yaml.tag_scores) do
            assert( self.tag_scores[ tag ] == nil or error( string.format( "Redundant tag score %s : %d", tag, tostring(score))))
            self.tag_scores[ tag ] = score
        end
    end
end

function QuipDatabase:ParseTable( t, primary_tag, additional_tags)
    local processed = {}
    for _,quip_dat in ipairs(t) do
        local quip = {tags = {}, dialog = {}, pre = quip_dat.pre, post = quip_dat.post }
        if quip_dat.tags then
            for outer_tag in string.gmatch(quip_dat.tags, '([^,]+)') do
                
                local inner_tags = {}
                for inner_tag in string.gmatch(outer_tag, '([^|]+)') do
                    table.insert(inner_tags, string.trim(inner_tag))
                end
                if #inner_tags > 1 then
                    --print (serpent.block(inner_tags))
                end
                table.insert(quip.tags, #inner_tags > 1 and inner_tags or inner_tags[1])
            end
        end
        
        if primary_tag then
            table.insert(quip.tags, 1, primary_tag)
        end
        if additional_tags then
            for _,v in ipairs(additional_tags) do
                table.insert(quip.tags, v)
            end
        end

        for _, quiptxt in ipairs(quip_dat) do
            table.insert(quip.dialog, quiptxt)
        end

        table.insert(processed, quip)
    end
    self:AddLines( processed )
end

function QuipDatabase:AddLines( lines )
    assert( type(lines) == "table", "Missing root table 'lines'" )
    local tag_context = ObtainWorkTable()
    for i, line in ipairs( lines ) do
        if line.push then
            assert( type(line.push) == "table" )
            table.insert( tag_context, line.push )

        elseif line.pop then
            table.remove( tag_context )

        else
            assert( type(line.tags) == "table" )
            for i, tags in ipairs( tag_context ) do
                table.arrayadd( line.tags, tags )
            end

            assert( line.notags == nil or type(line.notags) == "table" )

            if line.dialog == nil then
                line.dialog = {}
            elseif type(line.dialog) == "table" then
                for j, dialog in ipairs(line.dialog) do
                    assert(type(dialog) == "string" or error( string.format( "Invalid dialog #%d: %s", i, tostr(line,2))))
                end
            else
                assert(type(line.dialog) == "string" or error( string.format( "Invalid dialog #%d: %s", i, tostr(line,2))))
                line.dialog = { line.dialog }
            end

            if line.include then
                for i, t in ipairs(line.include) do
                    assert( type(t[1]) == "string")
                    table.arrayadd( line.dialog, t )
                end
            end
            if line.pre or line.post then
                for i, str in ipairs( line.dialog ) do
                    assert( line.pre == nil or type(line.pre) == "string", "invalid pre: "..tostr(line,2))
                    if line.pre then
                        str = line.pre .. "\n" .. str
                    end
                    if line.post then
                        str = str .. "\n" .. line.post
                    end
                    line.dialog[i] = str
                end
            end

            table.insert( self.db, line )
        end
    end
    ReleaseWorkTable( tag_context )
end

function QuipDatabase:EvaluateTags( match_tags, tags, notags )
    if notags then
        for i = 1, #notags do
            if table.arrayfind( match_tags, notags[i] ) then
                return nil
            end
        end
    end

    local count, score = 0, 0
    for i = 1, #match_tags do
        local match_tag = match_tags[i]
        local idx
        for i = 1, #tags do
            local tag = tags[i]
            if tag == match_tag then
                idx = i
                break
            elseif type(tag) == "table" and table.arrayfind( tag, match_tag ) then
                idx = i
                break
            end
        end

        if idx then
            count = count + 1
            score = score + (self.tag_scores[ match_tag ] or 1)
            table.remove( tags, idx )
            if #tags == 0 then
                return score
            end
        elseif i == 1 then
            return nil -- First match_tag is special: *must* be present.
        end
    end
end

local function cmp( x, y )
    return x[1] > y[1]
end

function QuipDatabase:GenerateMatches( match_tags, entries )
    if entries == nil then
        entries = {}
    end
    local tmp = ObtainWorkTable()
    for i, entry in ipairs( self.db ) do
        table.arrayadd( tmp, entry.tags )
        local score = self:EvaluateTags( match_tags, tmp, entry.notags )
        if score then
            table.binsert( entries, { entry.score or score, entry }, cmp )
        else
            table.clear( tmp )
        end
    end
    ReleaseWorkTable( tmp )
    return entries
end

function QuipDatabase:LookupEntries( match_tags, entries )
    entries = self:GenerateMatches( match_tags, entries )

    table.insert( self.recent_lookups, 1, { shallowcopy(entries), shallowcopy(match_tags) })
    while #self.recent_lookups > 5 do
        table.remove( self.recent_lookups )
    end

    return entries
end

function QuipDatabase:LookupQuipInEntries( entries )
    -- Go through entries and remove ones with lesser score.
    if #entries > 0 then
        local hi_score = entries[1][1]
        for i = #entries, 1, -1 do
            local match = entries[i]
            local score, probability = match[1], match[2].probability
            if score < hi_score or (probability and math.random() > probability) then
                table.remove( entries, i ) -- Skip this set.
            end
        end
    end

    local candidates = {}
    for k,v in ipairs(entries) do
        for j, str in ipairs( v[2].dialog ) do
            str = LOC( self:GetLocKey( v[2], j ))
            self.stats[str] = self.stats[str] or 0
            table.insert(candidates, str)
        end
    end
    table.shuffle(candidates)
    table.stable_sort(candidates, function(a,b) return self.stats[a] < self.stats[b] end)

    local entry = candidates[1]
    if entry then
        self.stats[entry] = self.stats[entry] + 1
        return entry, candidates
    end
end

function QuipDatabase:LookupQuip( match_tags )
    local entries = self:LookupEntries( match_tags )    
    return self:LookupQuipInEntries( entries )
end

function QuipDatabase:RecentLookups()
    return ipairs( self.recent_lookups )
end

function QuipDatabase:GetLocKey( entry, index )
    return string.format( "%s.%s.%d", self.loc_prefix, table.concat( entry.tags, "." ), index )
end

function QuipDatabase:HarvestStrings( t )
    for i, v in ipairs( self.db ) do
        for j, line in ipairs( v.dialog ) do
            local key = self:GetLocKey( v, j )
            t[ key ] = line:gsub( "\n( +)", "\n" ):trim()
        end
    end
end

function QuipDatabase:AllEntries()
    return ipairs( self.db )
end
