local ScoreEvent = class("ScoreEvent")

function ScoreEvent:init(keeper, id, count, userdata)  
    self.keeper = keeper
    self.id = id
    assert(self:GetDef(), id)

    self.userdata = userdata
    self.count = count
end

function ScoreEvent:GetDef()
    return Content.GetScoreComponent(self.id)
end

function ScoreEvent:GetName()
    return self:GetDef():GetName()
end

function ScoreEvent:GetText()
    local def = self:GetDef()   
    if def then
        local value = self:GetValue()
        local name = def:GetName()
        if self.count and self.count > 1 then
            return loc.format( "{1} (x{2})", name, self.count)
        else
            return loc.format( "{1}", name)
        end
    end
    return ""
end

function ScoreEvent:GetValue()

    local value = 0
    local def = self:GetDef()   
    if def.GetValue then
        value = math.round( (self.count or 1) * def.GetValue(def, self) )
    else
        value = math.round( (self.count or 1) * (def.value or 0) )
    end

    local multiplied_value = math.round( value * self.keeper.multiplier)

    return multiplied_value, value, self.keeper.multiplier
end

function ScoreEvent:Save()
    return {userdata = self.userdata, count = self.count, id = self.id}
end

---------------------------------------------------------


local ScoreKeeper = class("ScoreKeeper")

function ScoreKeeper:init(multiplier)
    self.score = 0
    self.score_events = {}
    self.multiplier = multiplier
end

function ScoreKeeper:GetScore()
    return self.score
end

function ScoreKeeper:GetScoreEvents()
    return self.score_events
end

function ScoreKeeper:AddScore( id, count, ... )

    if not Content.GetScoreComponent(id) then
        print ("UNKNOWN SCORE ID", id)
        return
    end

    local score_event = ScoreEvent(self, id, count, ...)
    
    local policy = score_event:GetDef().policy
    
    local delta = 0
    
    if policy == SCORE_POLICY.KEEP_LATEST then
        if self.score_events[id] and self.score_events[id][1] then
            delta = score_event:GetValue() - self.score_events[id][1]
        end
        self.score_events[id] = { score_event }
    elseif policy == SCORE_POLICY.KEEP_HIGHEST then
        
        local old_event = self.score_events[id] and self.score_events[id][1]
        if old_event and old_event:GetValue() < score_event:GetValue() then
            delta = score_event:GetValue() - old_event:GetValue()
            self.score_events[id] = { score_event }
        else
            self.score_events[id] = { score_event }
            delta = score_event:GetValue()
        end


    elseif policy == SCORE_POLICY.ACCUMULATE then
        self.score_events[id] = self.score_events[id] or {}
        
        delta = score_event:GetValue()

        local add = true
        if not score_event:GetDef().no_combine then
            for k,old_event in ipairs(self.score_events[id]) do
                if old_event.userdata == score_event.userdata then
                    old_event.count = old_event.count + score_event.count
                    add = false
                    break
                end
            end
        end

        if add then
            table.insert(self.score_events[id], score_event)
        end
    end

    local new_score = math.round(math.max(0, self.score + delta))
    local old_score = self.score 

    self.score = new_score
    if delta ~= 0 then
        TheGame:BroadcastEvent( "score_change", old_score, new_score, score_event)
    end
    
end


function ScoreKeeper:GetCount( id )
    local count = 0
    if self.score_events[id] then
        for k,v in ipairs(self.score_events[id]) do
            count = count + (v.count or 1)
        end
    end
    return count
end

function ScoreKeeper:GetValue( id )
    local value = 0
    if self.score_events[id] then
        for k,v in ipairs(self.score_events[id]) do
            value = value + v:GetValue()
        end
    end
    return value
end

function ScoreKeeper:Save()
    return {score_events = self.score_events, score = self.score, multiplier = self.multiplier}
end


function ScoreKeeper:GetScoreComponents()
    local approach = {}
    
    for score_id, event_list in pairs(self.score_events) do
        for _,score_event in ipairs(event_list) do
            local val, raw, multiplier = score_event:GetValue()
            if val > 0 then
                table.insert(approach, { txt = score_event:GetText(), raw = raw, total = val, multipler = multiplier})
            end
        end
    end

    table.sort(approach, function(a,b) return a.total > b.total end)
    return approach
end

function ScoreKeeper:GetScoreMultiplier()
    return self.multiplier or 1
end

function ScoreKeeper:Load()
    self.multiplier = self.multiplier or 1
    for _,v in pairs(self.score_events) do
        for _,ev in pairs(v) do
            ev.keeper = self
        end
    end

end
