class("Simulation")

require "simupdater"
require "util/scheduler"
require "entity"
require "prefab"

function Simulation:init( nextguid )
    self.entities = {}
    self.num_entities = 0
    
    self.scene = engine.rendering.Scene()
    self.scene:SetName( "SIM" )
    self.root = engine.rendering.SceneGraphNode()
    self.root:SetScene(self.scene)


    self.time = 0
    self.ticks= 0
    self.dtaccum = 0
    self.speed = 1

    self.scheduler = Scheduler()
    self.scheduler:SetTimeFn(function(t) self.time = t end)

    self.camera_stack = {}

    self.nextguid = nextguid or 1

    self.debug_renderers = {}

    self.newents = {}
    self.entity_bucket = {}

    self.entities_with_events = {}
    self.eventsystem = EventSystem()
    self.component_lists = {}
    self.updatehandlers = {
        wall = SimUpdater(self, "wall", "OnUpdate"),
        tick = SimUpdater(self, "tick", "OnTickUpdate"),
        late = SimUpdater(self, "late", "OnLateUpdate"),
    }

    self.scene:SetDebugRender( TheGame:IsDebugFlagged( DBG_FLAGS.RENDER ))

    self.timeconstant = engine.rendering.ShaderConstantValue("TIME", 0)
    

    self:SetWorldShaderConstant(self.timeconstant)
end

   
function Simulation:GetSaveData()
    local simstate = {
        nextguid = self.nextguid,
        entities={},
    }

    for guid,entity in pairs(self.entities) do
        if entity:IsPersistent() then
            local save_data = entity:GetSaveData()
            simstate.entities[guid] = save_data
        end
    end
    for guid,entity in pairs(self.newents) do
        if entity:IsPersistent() then
            local save_data = entity:GetSaveData()
            simstate.entities[guid] = save_data
        end
    end

    return simstate
end

function Simulation:GetScene()
    return self.scene
end

function Simulation:CreateEntityBucket( tag )
    if self.entity_bucket[ tag ] == nil then
        self.entity_bucket[ tag ] = {}
        for guid, entity in pairs(self.entities) do
            if entity:HasTag( tag ) then
                self:AddToBucket( entity, tag )
            end
        end
    end
end

function Simulation:AddToBucket( entity, tag )
    local bucket = self.entity_bucket[ tag ]
    if bucket then
        bucket[ entity.guid ] = entity
    end
end

function Simulation:RemoveFromBucket( entity, tag )
    local bucket = self.entity_bucket[ tag ]
    if bucket then
        bucket[ entity.guid ] = nil
    end
end

function Simulation:EntitiesInBucket( tag )
    return pairs( self.entity_bucket[ tag ] )
end

local _TEMP_ENTS_ON_SCREEN = {}
function Simulation:DoDebugRender()
    self.scene:DrawDebugLine(0, 0, 0, 1, 0, 0, 1, 0, 0)
    self.scene:DrawDebugLine(0, 0, 0, 0, 1, 0, 0, 1, 0)
    self.scene:DrawDebugLine(0, 0, 0, 0, 0, 1, 0, 0, 1)

    for i, renderer in ipairs( self.debug_renderers ) do
        renderer:DoDebugRender( self.scene )
    end

    table.clear(_TEMP_ENTS_ON_SCREEN)
    for class_name, cmps in pairs(self.component_lists) do
        if CLASSES[class_name].DoDebugRender then
            for guid, cmp in pairs(cmps) do
                
                if not _TEMP_ENTS_ON_SCREEN[cmp.entity] and cmp.entity:IsOnCamera() then
                    _TEMP_ENTS_ON_SCREEN[cmp.entity] = cmp.entity
                end
            end
        end

        if CLASSES[class_name].StaticDebugRender then
            CLASSES[class_name].StaticDebugRender(self.component_lists[class_name])
        end
    end

    local num_deb = 0
    for k,v in pairs(_TEMP_ENTS_ON_SCREEN) do
        v:DoDebugRender(self.scene)
        num_deb = num_deb + 1
    end
    table.clear(_TEMP_ENTS_ON_SCREEN)
end

function Simulation:RegisterDebugRender( renderer )
    assert( table.arrayfind( self.debug_renderers, renderer ) == nil )
    table.insert( self.debug_renderers, renderer )
end

function Simulation:UnregisterDebugRender( renderer )
    assert( table.arrayfind( self.debug_renderers, renderer ) ~= nil )
    table.arrayremove( self.debug_renderers, renderer )
end


function Simulation:GetTimeScale()
    return self.speed
end

function Simulation:SetTimeScale(scale)
    self.speed = scale
    self:BroadcastEvent( "rate_changed", scale)
end

function Simulation:GetEntityByGUID(guid)
    --print("Simulation:GetEntityByGUID", guid, self.newents[guid])
    
    local ent = self.entities[guid] or self.newents[guid]
    --assert(ent, "Could not find entity " .. tostring(guid))
    return ent
end

function Simulation:GetEntityByName(name)
    -- Can easily convert this into a table lookup, at the cost of maintenance and garbage.
    for guid, ent in pairs(self.entities) do
        if ent:GetName() == name then
            return ent
        end
    end
    for guid, ent in pairs(self.newents) do
        if ent:GetName() == name then
            return ent
        end
    end
end

function Simulation:GetRoot()
    return self.root
end

function Simulation:SpawnPrefab(prefabname, spawn_data, x, y, z)
    assert(prefabname)

    local prefab = TheGame:GetPrefabData(prefabname)
    if not prefab then
        -- Making this non-fatal so that we delete prefabs and maintain area backwards compat.
        print( "WARNING: Could not find prefab named " .. prefabname)
        print( debug.traceback() )
        return
    end

    spawn_data = spawn_data or {}

    local ent = Entity(self, prefab, spawn_data)

    if prefab.fn then
        prefab.fn( ent, spawn_data )
    end

    assert( self.newents[ ent.guid ] == nil, ent )
    assert( self.entities[ ent.guid ] == nil )
    self.newents[ent.guid] = ent

    if ent.cmp.Transform then
        self:GetRoot():AddChild(ent.cmp.Transform.scenegraphnode)
        if x and y and z then
            ent.cmp.Transform:SetLocalPosition(x, y, z)        
        end
    end

    ent:MakeEntityReady()
    
    if self.sim_ready then
        ent:OnSimReady()
    end

    return ent
end

function Simulation:DespawnEntity(ent)
    if self.newents[ ent.guid ] then
        ent:OnDespawn()
        self.newents[ ent.guid ] = nil

    else
        assert( self.entities[ ent.guid ] == ent or error( string.format( "%s - removed=%s", tostring(ent), tostring(ent.removed) )) )
        ent:OnDespawn()
        self.entities[ent.guid] = nil
        self.num_entities = self.num_entities - 1
    end
end

function Simulation:GetTime()
    return self.time
end

function Simulation:GetNextGUID()
    self.nextguid = self.nextguid + 1
    return self.nextguid
end

function Simulation:GetEntityCount()
    return self.num_entities
end

function Simulation:DoFullSimUpdate( dt, is_paused )
    engine.inst:ProfilerPush("Simulation:DoFullSimTick")
    
    self.timeconstant:SetValue(self.time)
    
    self.updatehandlers.wall:Update(dt)

    if #self.camera_stack > 0 then
        self.camera_stack[ #self.camera_stack ]:OnUpdate( dt*self.speed )
    end

    if self.OnUpdate then
        self:OnUpdate(dt*self.speed, is_paused)
    end

    local num_ticks = 0

    if not is_paused then   
        self.dtaccum = self.dtaccum + dt*self.speed
        
        -- while (self.dtaccum - SIM_TICK_TIME*.95) > 0 and num_ticks < MAX_SIM_TICKS do
        while self.dtaccum >= SIM_TICK_TIME do
            
            num_ticks = num_ticks + 1
            -- self.dtaccum = math.max(0, self.dtaccum - SIM_TICK_TIME)
            self.dtaccum = self.dtaccum - SIM_TICK_TIME

            self:DoTick()
        end

        --self.dtaccum = math.min(self.dtaccum, SIM_TICK_TIME)
    end
  

    engine.inst:ProfilerPush("Late Updates")
    self.updatehandlers.late:Update(dt)
    engine.inst:ProfilerPop()
    
    self:HandleEvents() 
    self:DoPostUpdate()
    engine.inst:ProfilerPop()

    return num_ticks

end

function Simulation:DoPostUpdate()
    engine.inst:ProfilerPush('Entity List Management')

    while next(self.newents) do
        local guid, entity = next(self.newents)
        assert( self.entities[ guid ] == nil )
        self.newents[ guid ] = nil
        self.entities[ guid ] = entity
        self.num_entities = self.num_entities + 1
    end

    engine.inst:ProfilerPop()
end


function Simulation:HandleEvents()
    engine.inst:ProfilerPush('Events')

    local passes = 0
    while passes < 100 do
        if next(self.entities_with_events) then
            self.alt_entities_with_events = self.alt_entities_with_events or {}
            self.entities_with_events, self.alt_entities_with_events = self.alt_entities_with_events, self.entities_with_events
            table.clear(self.entities_with_events)

            for k, v in pairs(self.alt_entities_with_events) do
                if not k.removed then
                    k:ConsumeEvents()
                end
            end

            table.clear(self.alt_entities_with_events)

        else
            break
        end
        passes = passes + 1
    end

    engine.inst:ProfilerPop()
end


function Simulation:GetCameraInfo()
    assert(#self.camera_stack > 0)
    local cam =  self.camera_stack[#self.camera_stack]
    return cam.caminfo
end


--the sim is based around fixed ticks (of 1/60 or 1/30)
function Simulation:DoTick()
    engine.inst:ProfilerPush("Simulation:DoTick")

    self.ticks = self.ticks + 1
    local dt = SIM_TICK_TIME

    engine.inst:ProfilerPush('Scheduler')
    do
        self.scheduler:Update(dt)
    end
    engine.inst:ProfilerPop()

    self.scene:SetSceneTime(self.time )
    self.updatehandlers.tick:Update(dt)
    engine.inst:ProfilerPop()
end



function Simulation:PushCamera(cam)
    local old_top = #self.camera_stack > 0 and self.camera_stack[#self.camera_stack]
    cam:UpdatePosition()
    table.insert(self.camera_stack, cam)
end

function Simulation:PopCamera(cam)
    local old_top = #self.camera_stack > 0 and self.camera_stack[#self.camera_stack]
    local rem_cam = nil

    if cam then
        for k, v in ipairs(self.camera_stack) do
            if v == cam then
                rem_cam = table.remove(self.camera_stack, k)
                break
            end
        end
    else
        rem_cam = table.remove(self.camera_stack)
    end
end


function Simulation:GetActiveCamera()
    if #self.camera_stack > 0 then
        return self.camera_stack[#self.camera_stack]
    end
end

function Simulation:GetRenderSize()
    return self.renderw, self.renderh
end

function Simulation:SetRenderSize(w, h)

    self.renderw, self.renderh = w, h    
    for k,v in ipairs(self.camera_stack) do
        v.caminfo:SetScreenSize(w,h)
    end
end

function Simulation:StartUpdatingComponent(cmp, type)
    self.updatehandlers[type]:StartUpdating(cmp)
end

function Simulation:StopUpdatingComponent(cmp, type)
    self.updatehandlers[type]:StopUpdating(cmp)
end


function Simulation:OnDestroy()
    local guid, ent
    -- Clear out all entities and pending newents.
    -- We iterate this way in case despawning an entity somehow triggers spawning another one.
    while true do
        guid, ent = next( self.entities )
        if ent == nil then
            guid, ent = next( self.newents )
        end
        if ent == nil then
            break
        end
        ent:Despawn()
    end
    assert( next( self.entities ) == nil )
    assert( next( self.newents ) == nil )
end

function Simulation:OnControlUp(control)
end

function Simulation:OnControlDown(control)
end

function Simulation:GetEvents()
    return self.eventsystem
end

function Simulation:BroadcastEvent( ... )
    return self.eventsystem:BroadcastEvent( ... )
end


function Simulation:SetWorldShaderConstant(constant)
    self.scene:SetShaderConstant(constant)
    return constant
end

function Simulation:AddEntCmp(ent, cmp)
    self.component_lists[cmp._classname] = self.component_lists[cmp._classname] or {}
    self.component_lists[cmp._classname][ent.guid] = cmp
end

function Simulation:RemoveEntCmp(ent, cmp)
    self.component_lists[cmp._classname][ent.guid] = nil
end

function Simulation:GetAllComponents(component_class)
    return self.component_lists[component_class._classname] or table.empty
end

function Simulation:IsReady()
    return self.sim_ready
end



function Simulation:MakeSimReady()
    assert( self.sim_ready == nil)

    if self.OnSimReady then
        self:OnSimReady()
    end
    
    for k,v in pairs(self.entities) do
        v:OnSimReady()
    end
    self.sim_ready = true

    --we might spawn some new entities from within this loop, so copy the list out. Since the sim is now "ready", the new new ents will receieve their onsimready callbacks in-line
    local temp_table = {}
    for k,v in pairs(self.newents) do
        temp_table[k] = v
    end

    for k,v in pairs(temp_table) do
        v:OnSimReady()
    end
end


function Simulation:RaycastNodeList( nodes, screenx, screeny, filter_fn )

    engine.inst:ProfilerPush("RaycastNodeList")

    local cam_info = self:GetCameraInfo()
    
    local hit_node
    local nearest = math.huge

    for guid, node in pairs( nodes ) do
        
        local model = node:GetModel()
        if model and model.RayTest then
            local hit, dist = model:RayTest( cam_info, screenx, screeny )
            if hit and dist < nearest and (not filter_fn or filter_fn(node)) then
                hit_node = node
                nearest = dist
            end
        end
    end
    
    engine.inst:ProfilerPop()

    return hit_node
end

function Simulation:RaycastEntityList( entities, screenx, screeny, filter_fn )

    engine.inst:ProfilerPush("RaycastEntityList")

    local cam_info = self:GetCameraInfo()

    local hit

    for guid, entity in pairs( entities ) do
        -- print("RaycastEntityList", entity, entity:GetImpl().RayTest)
        if entity:GetImpl().RayTest and entity:GetImpl():RayTest( screenx, screeny ) and (not filter_fn or filter_fn(entity)) then
            hit = entity
            break
        end
    end
    
    engine.inst:ProfilerPop()

    return hit
end

function Simulation:ToggleDebugController()
end
