class("FrontEnd")
require "input/input"
require "colours"
require "ui/widget"
require "ui/widgets/debugrenderoverlay"
require "ui/widgets/tooltip"
require "ui/widgets/solidbox"
require "ui/widgets/toastwidget"

CONTROL_MODE = MakeEnum{"MOUSE_KEYBOARD", "GAMEPAD", "TOUCH", "DIRECTIONS_ONLY"}

local TTDELAY = 0--.2

function FrontEnd:init(game)
    self.game = game
    self.dirtytransforms = {}
    self.base_scale = 1

    -- Set the default screen mode
    self.screen_mode = SCREEN_MODE.MONITOR
    -- Check if the engine sets a screen mode different than the default
    local engine_screen_mode = engine.inst:GetPreferredDisplaySize() ~= SCREEN_MODE.MONITOR and engine.inst:GetPreferredDisplaySize()
    -- Check if a local setting has a set screen mode
    local local_settings_screen_mode = TheGame:GetLocalSettings().SCREEN_MODE
    -- Check if the user has set a preferred size
    local user_settings_screen_mode = engine.inst:GetSetting( "graphics", "layoutscale_mode", SCREEN_MODE.MONITOR )
    if tonumber(user_settings_screen_mode) then user_settings_screen_mode = SCREEN_MODE_INDEX[user_settings_screen_mode] end
    -- If this mode is different than the current one, it'll trigger a change through all live screens and widgets
    self:SetPendingScreenMode( engine_screen_mode or local_settings_screen_mode or user_settings_screen_mode or SCREEN_MODE.MONITOR ) 

    -- Set control mode
    self.control_mode = CONTROL_MODE.MOUSE_KEYBOARD
    local control_mode = (TheGame:GetLocalSettings().SIMULATE_TOUCH_MODE and "TOUCH") or engine.inst:GetPreferredInputType() or CONTROL_MODE.MOUSE_KEYBOARD
    self:SetPendingControlMode( control_mode )

    self.horizontal_safe_zone = 0.0
    self.vertical_safe_zone = 0.0

    self.screenstack = {}
    self.updatingwidgets = {}
    self.newupdatingwidgets = {}

    self.enable = true
    self.hud_flags = 0

    self.scene = engine.rendering.Scene()
    self.scene:SetName( "FE" )
    
    self.camera = engine.rendering.CameraInfo()
    self.camera:SetUI()


    self.root = engine.rendering.SceneGraphNode()
    self.root:SetScene(self.scene)
    
    self.fe_root = Widget():SetFE(self):SetName("input root")
    self.fe_root:SetForceViz()
    self.root:AddChild(self.fe_root.node) 
    
    self.screenroot = self.fe_root:AddChild(Widget())
    self.screenroot:SetName("screen root")
    
    local w, h = self:GetScreenDims()

    self.bottom_bar = self.fe_root:AddChild( Widget.BottomBarOverlay( self ) )
        :SetAnchors( "center", "bottom" )
        :Hide()
    self.inker = Widget.Presentation(w, h):Hide():SetPos(w/2, h/2)
    self.fader = self.fe_root:AddChild(Widget.SolidBox(w, h , 1, 0, 0, 0):Hide():SetPos(w/2,h/2))
    self.reticule_root = self.fe_root:AddChild(Widget())
    self.saving_root = self.fe_root:AddChild(Widget():SetAnchors("left", "center"))
    self.saving_anim = self.saving_root:AddChild(Widget.SaveGameAnim())
        :LayoutBounds( "left", "center" )
        :Offset( 60, -RES_Y*0.4 )
        :Hide()

    self.debugroot = self.fe_root:AddChild(Widget())
    self.debugroot:SetName("debug root")
    self.debugroot.nodebug = true

    self.dragroot = self.fe_root:AddChild(Widget())

    local BRANCH_STR = 
    {
        testing = "Experimental branch",
        staging = "Experimental branch (staging)",
        release = "Release",
        dev = "Dev branch",
        demo = "Demo",
    }

    local build_str = BRANCH_STR[BRANCH] or "Unknown Branch"

    
    
    local build_txt = loc.format( "Griftlands {3} Build (r{1}) - {2}", BUILD_REVISION, build_str, IsEarlyAccess() and "Early Access" or "Launch" )
    self.build_version = self.fe_root:AddChild( Widget.Label( "body", 22, build_txt )
        :SetGlyphColour(0xffffffaa))

    if BRANCH ~= "dev" then
        self.build_version:Hide()
    end

    self.tooltip_widgets = {}
    self.tooltip_delay = TTDELAY

    self.timeconstant = engine.rendering.ShaderConstantValue("TIME", 0)
    self:SetShaderConstant(self.timeconstant)

    self:RefreshColourblindMode()

    self:OnScreenResize(w,h)
end

-- Shows an animation indicating the game is saving
-- If no parameters are set, it'll stay looping in place until :HideSaving() is called
-- If duration is set, it'll call callback and HideSaving() automatically after that delay (in seconds)
-- If show_info is set, it'll also display the saving icon description (only for the Screen.SavingScreen)
function FrontEnd:ShowSaving( duration, callback, show_info )
    self.saving_anim:AnimateIn( show_info )
    if duration then self.saving_anim:DoTaskInTime( "hide_saving", duration, function()
        if callback then callback() end
        self:HideSaving()
    end ) end
    return self
end

function FrontEnd:HideSaving()
    self.saving_anim:AnimateOut()
    return self
end

function FrontEnd:RefreshControlOptions()
    local handled, left, right = ObtainWorkTable(), ObtainWorkTable(), ObtainWorkTable()
    local main_overlay = self:FindScreen( Screen.MainOverlay )
    if main_overlay and main_overlay:TopBarHasFocus() then
        -- If the focus is on the main overlay's top bar itself, call it
        main_overlay:CollectControlHints(left, right, handled)
        self.bottom_bar:SetControlOptions( left, right )
            :SetPos( 0, nil )
    else
        -- If not, query the top screen
        local top_screen = self:GetTopScreen()
        if top_screen then
            top_screen:CollectControlHints(left, right, handled)
            self.bottom_bar:SetControlOptions( left, right )
                :SetPos( 0, nil )
        end
    end


    ReleaseWorkTable(left)
    ReleaseWorkTable(right)
    ReleaseWorkTable(handled)
end

function FrontEnd:GetControlMode()
    return self.control_mode
end

function FrontEnd:GetScreenMode()
    return self.screen_mode
end

function FrontEnd:GetControlDeviceID()
    return self.device_id
end

function FrontEnd:SetPendingControlMode( cm, device_id )
    -- Notify the engine if we're in touch-input mode
    if TheGame:GetLocalSettings().SIMULATE_TOUCH_MODE then
        engine.inst:EnableTouchScreenSimulation()
    else
        engine.inst:DisableTouchSimulation()
    end

    self.pending_control_mode = cm
    self.pending_device_id = device_id
end

function FrontEnd:SetPendingScreenMode( sm )
    if IsEnum( sm, SCREEN_MODE ) then
        self.pending_screen_mode = sm
    else
        LOGWARN ("invalid screen mode: %s", tostring(sm))
    end
end

function FrontEnd:AddDirtyTransformWidget(w)
    if not self.dirtytransforms[w] then
        self.dirtytransforms[w] = w
    end
end

function FrontEnd:RefreshColourblindMode()

    -- Get the current colours
    local colourblind_mode = engine.inst:GetSetting( "graphics", "colourblind_mode", 1 )
    local colours = GetColourblindModeColours()

    -- Update the colours in UICOLOURS
    UICOLOURS.NEGOTIATION_HOSTILE = colours[CARD_TYPE.NEGOTIATION_HOSTILE][colourblind_mode]
    UICOLOURS.NEGOTIATION_DIPLOMACY = colours[CARD_TYPE.NEGOTIATION_DIPLOMACY][colourblind_mode]
    UICOLOURS.NEGOTIATION_MANIPULATE = colours[CARD_TYPE.NEGOTIATION_MANIPULATE][colourblind_mode]
    UICOLOURS.FIGHT_ATTACK = colours[CARD_TYPE.FIGHT_MELEE][colourblind_mode]
    UICOLOURS.FIGHT_SKILL = colours[CARD_TYPE.FIGHT_SKILL][colourblind_mode]
    UICOLOURS.FIGHT_STATUS = colours[CARD_TYPE.FIGHT_STATUS][colourblind_mode]
    UICOLOURS.FIGHT_ITEM = colours[CARD_TYPE.ITEM][colourblind_mode]

    -- Notify any objects that might want to update their colours
    if TheGame and TheGame:GetEvents() then TheGame:GetEvents():BroadcastEvent( "colourblind_mode_changed" ) end

end

function FrontEnd:IsGamepadMode()
    return self:GetControlMode() == CONTROL_MODE.GAMEPAD 
end

function FrontEnd:IsTouchMode()
    return self:GetControlMode() == CONTROL_MODE.TOUCH 
end
   
function FrontEnd:ToggleDebugFreeze()
    self.debugfreeze = not self.debugfreeze
end

function FrontEnd:SetDebugMode(enable)
    if enable then
        TheGame:GetDebug():ClearPanels()
        TheGame:GetDebug():CreatePanel( DebugUIDebug())
    else
        self.debugfreeze = nil
        self.debugwidget = nil
    end
end

function FrontEnd:SetDebugWidget(w)
    if not self.debugfreeze then
        self.debugwidget = w
    end
end

function FrontEnd:DoDebugRender( dbg )
    if self.debugwidget then
        self.debugwidget:DrawDebugBounds(0xFF000080)
    end

    if dbg:IsDebugFlagged( DBG_FLAGS.UI ) and self.debugfreeze then
        if self.debugwidget then
            self.debugwidget:DoDebugRender(dbg, 0, true)
        end
    else
        local screen = self:GetTopScreen()
        if screen then
            screen:DoDebugRender( dbg )
        end
    end
end

function FrontEnd:DrawDebugText(str, x, y, sz, col)
    sz = sz or 10
    local r,g,b,a = 1,1,1,1
    if col then
        r,g,b,a = HexColour(col)
    end
    self.scene:DrawDebugText( str, x, y, 0, sz, r,g,b,a )
end

function FrontEnd:DrawDebugLine(x1, y1, x2, y2, col)
    local r,g,b,a = 1,0,0,1
    if col then
        r,g,b,a = HexColour(col)
    end
    self.scene:DrawDebugLine(x1,y1,0,x2,y2,0, r,g,b,a)    
end

function FrontEnd:DrawDebugTriangle(x1, y1, z1, x2, y2, z2, x3, y3, z3, col)
    local r,g,b,a = 1,0,0,1
    if col then
        r,g,b,a = HexColour(col)
    end
    self.scene:DrawDebugTriangle(x1,y1,z1, x2,y2,z2, x3,y3,z3, r,g,b,a)    
end


function FrontEnd:GetDragWidget()
    return self.drag and self.drag.w, self.drag and self.drag.id
end


function FrontEnd:CancelDrag()

    if self.drag and self.drag.cancel_fn then
        self.drag.cancel_fn()
    end

    self:ClearDragWidget()

end
function FrontEnd:ClearDragWidget()
    if self.drag then
        if self.drag.clear_fn then
            self.drag:clear_fn()
        elseif not self.drag.w.removed then
            self.drag.w:Remove()
        end
        self.drag = nil
    end
end

function FrontEnd:SetDragWidget(w, cancel_fn, clear_fn)
    assert( is_instance( w, Widget ))

    if self.drag then
        if self.drag.clear_fn then
            self.drag:clear_fn()
        elseif not self.drag.w.removed then
            self.drag.w:Remove()
        end
        self.drag.w = w
    else
        self.drag = 
        {
            w = w,
            clear_fn = clear_fn,
            cancel_fn = cancel_fn,
        }
        self.drag.orig_x, self.drag.orig_y = w:GetPos()
    
        local x, y = w:TransformFromWorld(TheGame:GetInput():GetMousePos())
        if w.parent then
            self.drag.drag_offset_x, self.drag.drag_offset_y = x, y
        else
            self.drag.drag_offset_x, self.drag.drag_offset_y = 0, 0
        end
    end
    
    w.hover = false

    w:ClearHover()
    w:ClearFocus()
   
    self.dragroot:AddChild(w)
    self:UpdateDragPos()
end


function FrontEnd:InkTransition(time_to_take)
    if TheGame:GetLevelOfDetail() <= 1 then
        -- astolk: InkTranstions cause Vulkan Validation Errors. Should be fixed more permanently, but for now, disable on NX.
        return
    end

    local sw, sh = FrontEnd:GetScreenDims()
    if self.capture_image then
        local w, h = self.capture_image:GetSize()
        
        if math.ceil(sw/2) ~= w or math.ceil(sh/2) ~= h then
            self.capture_image = nil
        end
    end
    if not self.capture_image then
        self.capture_image = engine.asset.Texture(math.ceil(sw/2), math.ceil(sh/2), PIXEL_FORMAT.A8R8G8B8)
    end

    TheGame:CaptureRenderOutput(self.capture_image)
    if self.inker:GetParent() == self.screenroot then
        self.screenroot:RemoveChild( self.inker )
    end

    self.screenroot:AddChild( self.inker, 2 )

    self.inker:Show():Dissolve(self.capture_image, time_to_take, function()
        self.screenroot:RemoveChild( self.inker )
    end)
end

function FrontEnd:FadeScreen(fn, screen, time_to_take )
    if self.fading then
        self.fading = false
        if self.fade_fn then
            self.fade_fn()
        end

    end

    assert( not self.fading ) -- Not supported.
    assert( is_instance( screen, Widget.Screen ))

    self.fade_out = true
    self.fading = true

    self.fade_screen = screen
    screen:SetTintAlpha( 1.0 )

    self.fader:Hide()
    self.fade_time_total = time_to_take or 0.2
    self.fade_time = 0
    self.fade_fn = fn
end


function FrontEnd:FadeTransition(fn, time_to_take )
    time_to_take = time_to_take or .2

    if self.fading and self.fade_out then
        self.fade_fn = fn
    else
        self.fade_out = true
        self.fading = true

        self.fader:Show()
        self.fader:SetColour(0,0,0,0)
        self.fade_time_total = time_to_take
        self.fade_time = 0
        self.fade_fn = fn
    end
end

function FrontEnd:FadeIn(time_to_take)
    self.fade_out = false
    self.fading = true

    self.fader:Show()
    self.fader:SetColour(0,0,0,1.0)
    self.fade_time_total = time_to_take
    self.fade_time = 0
    self.fade_fn = nil
end

function FrontEnd:SetDebugRender(val)

    if val then
        self.debugrenderoverlay = self.debugrenderoverlay or self.debugroot:AddChild(Widget.DebugRenderOverlay(self))
        self.debugrenderoverlay:Show()
    else
        if self.debugrenderoverlay then
            self.debugrenderoverlay:Remove()
            self.debugrenderoverlay = nil
        end
    end
    self.debugrender = val
    self.scene:SetDebugRender( val )
end

function FrontEnd:GetCamera()
    return self.camera
end

function FrontEnd:OnScreenResize(w,h)
    self.camera:SetScreenSize(w, h)
    local aspect = w/h
    local author_aspect = RES_X/RES_Y

    if aspect >= author_aspect then
        self.base_scale = h/RES_Y
    elseif aspect < author_aspect then
        self.base_scale = w/RES_X
    end

    local sw,sh = w, h
    local x,y = 0,0
    --too wide!
    if w/h > 21/9 then
        sw = h*(21/9)
        x = (w-sw)/2
    --too tall!
    elseif w/h < 4/3 then
        sh = w / (4/3)
        y = (h-sh)/2
    end
    self.scissor_w = sw
    self.scissor_h = sh
    self.screenroot:SetScissor(x, y, sw, sh)

    for k,v in pairs(self.screenstack) do
        v:OnScreenResize(w,h)
    end

    self.fader:SetSize(w, h):SetPos(w/2, h/2)
    self.inker:SetSize(w, h):SetPos(w/2, h/2)
    self.bottom_bar:OnScreenResize(w, h)
        :SetPos( 0, nil )

    self.debugroot:OnScreenResize(w,h)
    self.dragroot:SetScale(self.base_scale)

    self.reticule_root:SetScale(self.base_scale, self.base_scale):SetPos(w/2, h/2)

    self.build_version:SetScale( self.base_scale ):LayoutBounds( "left", "bottom", 0, 0 )

end

function FrontEnd:GetScreenScissorDims()
    if self.scissor_w and self.scissor_h then
        return self.scissor_w, self.scissor_h
    else
        return self:GetScreenDims()
    end
end

function FrontEnd:Enable()
    self.enable = true
end

function FrontEnd:Disable()
    self.enable = false
end

function FrontEnd:OnInputEvent( fn, control, device_type, trace, device_id )
    if ( device_type == "mouse" or device_type == "key" ) then
        self:SetPendingControlMode( CONTROL_MODE.MOUSE_KEYBOARD, nil )
    elseif device_type == "touch" then
        self:SetPendingControlMode( CONTROL_MODE.TOUCH, device_type )
    elseif device_type == "gamepad" then
        self:SetPendingControlMode( CONTROL_MODE.GAMEPAD, device_id )
    end

    for i = #self.screenstack, 1, -1 do
        local screen = self.screenstack[i]
        if fn( screen, control, device_type, trace, device_id ) then
            return true
        elseif screen:SinksInput() then
            break
        end
    end
    return false
end

function FrontEnd:RawKeyDown(key) 
    if not self.enable or (self.fading and self.fade_out) then
        return false
    end
    return self:OnInputEvent( Widget.OnRawKeyDown, key )
end

function FrontEnd:RawKeyUp(key) 
    if not self.enable or (self.fading and self.fade_out) then
        return false
    end
    return self:OnInputEvent( Widget.OnRawKeyUp, key )
end

function FrontEnd:TextInput(text) 
    if not self.enable or (self.fading and self.fade_out) then
        return false
    end
    return self:OnInputEvent( Widget.OnTextInput, text )
end

function FrontEnd:TextEdit(text) 
    if not self.enable or (self.fading and self.fade_out) then
        return false
    end
    return self:OnInputEvent( Widget.OnTextEdit, text )
end

function FrontEnd:OnControlUp( control, device_type, device_id )
    if self.drag and ( device_type == "mouse" or device_type == "touch" ) then
        return false
    end

    -- local txt = ""
    -- for k, v in ipairs(control.controls) do
    --     txt = txt .. (v.displayname or "") .. " "
    -- end
    -- print( "CONTROL UP", device_type, txt )

    if not self.enable or (self.fading and self.fade_out) then
        return false
    end

    if not self.reticule_active then
        if control:Has( Controls.Digital.MENU_UP, Controls.Digital.MENU_RIGHT, Controls.Digital.MENU_DOWN, Controls.Digital.MENU_LEFT ) then
            local focus_widget = self:GetFocusWidget()
            -- print( "-- focus_widget:",focus_widget )
            if focus_widget == nil then
                local default_widget = self.top_screen and self.top_screen:GetDefaultFocus()
                -- print( "-- focus_widget:",focus_widget,"default_widget:",default_widget)
                if default_widget then
                    default_widget:SetFocus()
                    return true
                end
            end
        end
    end

    local ret = self:OnInputEvent( Widget.OnControlUp, control, device_type, nil, device_id )
    
	-- When ending a click/tap, only clear hover after notifying all widgets, so that the focused/hovered ones get to process input
    if self:IsTouchMode() then
        self.game:GetInput():ResetMousePos()
        self.fe_root:ClearHover()
        self:ClearFocusWidget()
    end

    return ret
end

function FrontEnd:OnControlDown( control, device_type, trace, device_id )
    if not self.enable or (self.fading and self.fade_out) then
        return false
    end

    if ( device_type == "mouse" or device_type == "touch" ) then
        if not control:Has( Controls.Digital.SCROLL_DOWN ) and not control:Has( Controls.Digital.SCROLL_UP ) then
            self:ClearFocusWidget()

            local x, y = self.game:GetInput():GetMousePos()
            -- Checking mouse hover here specifically for touch screens.
            -- Updating the hover in the update is 1 frame too late. It needs to happen immediately when the touch starts.
            self:CheckMouseHover( x, y )
        end
    end

    if self.drag then
        if (device_type == "mouse" and control:Has( Controls.Digital.MENU_ALT )) or control:Has( Controls.Digital.MENU_CANCEL )  then
            self:CancelDrag()
            return true
        end
    end

    return self:OnInputEvent( Widget.OnControlDown, control, device_type, trace, device_id )
end

function FrontEnd:GetFocusWidget()
    return self.focus_widget
end

function FrontEnd:SetFocusWidget( widget )
    if widget ~= self.focus_widget then

        assert( is_instance( widget, Widget ))
        assert( not widget.removed, tostring(widget) )
        -- assert( widget.can_focus, widget )

        -- print( "FOCUS:", widget, debug.traceback() )

        local prev_focus = self.focus_widget
        if self.focus_widget then
            self.focus_widget:LoseFocus( widget )
        end

        self.focus_widget = widget

        -- Somewhat arbitrary according to the feature I'm implementing, but if focus changes, the tooltip override should be cleared.
        self.tooltip_focus_override = nil
        
        widget:GainFocus()

        self.game:BroadcastEvent( "focus_changed", prev_focus, widget )
        self.wants_control_option_refresh = true
    else
        self.game:BroadcastEvent( "focus_changed", self.focus_widget, nil )
    end
end

function FrontEnd:ClearFocusWidget()
    local widget = self.focus_widget
    self.game:BroadcastEvent( "focus_changed", widget, nil )
    if widget then
        -- print( "CLEAR FOCUS:", self.focus_widget, debug.traceback() )
        self.focus_widget = nil
        self.tooltip_focus_override = nil
        self.wants_control_option_refresh = true
        widget:LoseFocus()
    end
end

function FrontEnd:GetHoverWidget()
    return self.fe_root:GetDeepestHover()
end

function FrontEnd:SetScreen(screen)
    self:ClearScreens()
    self:PushScreen(screen)
end

function FrontEnd:FindScreen( screen_name )
    assert( screen_name ~= nil )
    for i, screen in ipairs(self.screenstack) do
        if screen == screen_name or screen._classname == screen_name or is_instance( screen, screen_name ) then
            return screen, i
        end
    end
end

function FrontEnd:Screens()
    return ipairs( self.screenstack )
end

function FrontEnd:ClearScreens()

    while #self.screenstack > 0 do
        self:PopScreen(self.screenstack[1])
    end

    for _, w in pairs( self.tooltip_widgets ) do
        w:Remove()
    end
    table.clear( self.tooltip_widgets )
    
    self.tooltip_focus = nil
    self.tooltip_widget = nil
    self.tooltip_data = nil

    self.drag_test_widget = nil

    self:ClearFade()
end

local function FindTopScreen( screenstack )
    for i = #screenstack, 1, -1 do
        local screen = screenstack[i]
        if not screen:IsOverlay() then
            return screen
        end
    end
    return nil
end

function FrontEnd:PushScreen( s, idx )
    self:InsertScreen( s, #self.screenstack + 1 )
    return s
end

function FrontEnd:RefreshTopScreen()
    local new_top = FindTopScreen( self.screenstack )
    if new_top ~= self.top_screen then
        if self.top_screen then
            self.top_screen:OnLoseTop()
        end

        self.top_screen = new_top

        -- Notify any overlays on top, and the new top screen itself.
        for i = #self.screenstack, 1, -1 do
            local screen = self.screenstack[i]
            screen:OnGainTop( new_top )
            self:RequestControlRefresh()
            if screen == new_top then
                break
            end
        end
    end
end

function FrontEnd:InsertScreen( s, idx )
    self:CancelDrag()

    s:SetFE(self)


    -- Calculate top screen instead of using self.top_screen, which may be out of date
    -- if we are called recursively.
    local top_screen = FindTopScreen( self.screenstack )
    

    if idx == nil then
        
        if #self.screenstack == 1 and self.screenstack[1]:IsOverlay() then
            idx = 1 --this is a hack for when we *only* have an overlay on the stack 
        elseif top_screen == nil or s:IsOverlay() then
            idx = #self.screenstack + 1
        else
            idx = table.arrayfind( self.screenstack, top_screen ) + 1
        end
    end


    local next_screen = self.screenstack[ idx ]
    table.insert(self.screenstack, idx, s)

    self.screenroot:AddChild(s, self.screenroot:IndexOf( next_screen ) )

    self:UpdateScreenViz()
    s:OnOpen()

    -- These must come last due to complications if screens push more screens during OnOpen().
    -- (1) OnGainTop() should only ever be called after OnOpen()
    -- (2) OnLoseTop() should be called if and only if the screen ever received OnGainTop()
    if not s:IsOverlay() then
        self:RefreshTopScreen()
    end

    self:RefreshMusic()

    return s
end

function FrontEnd:RefreshMusic()
    local music_event, character_music_event
    for i = #self.screenstack, 1, -1 do
        local screen = self.screenstack[i]
        music_event, character_music_event = screen:GetMusicEvent()
        if music_event or character_music_event then
            break
        end
    end

    if character_music_event then
        TheGame:GetMusic():PlayCharacterMusic( character_music_event )
    else
        TheGame:GetMusic():PlayMusic( music_event )
    end

end

function FrontEnd:UpdateScreenViz()
    local viz = true

    local show_overlays = true
    for k = #self.screenstack, 1, -1 do
        local screen = self.screenstack[k]
        local screen_idx = self.screenroot:IndexOf( screen )

        if screen:IsOverlay() and not show_overlays then
            viz = false
        end
        
        if viz then
            screen:Show()
        else
            screen:Hide()
        end
        
        if screen:IsOpaque() then
            viz = false
        end

        if not screen:IsOverlay() then
            show_overlays = false
        end
    end
end

function FrontEnd:PopScreen( screen )
    self:CancelDrag()

    assert( #self.screenstack > 0 )

    local idx
    if screen then
        idx = table.arrayfind( self.screenstack, screen )
        if not idx then
            error( "Screen '" .. screen._classname .. "' is not on stack" )
        end
    elseif self.top_screen then
        idx = table.arrayfind( self.screenstack, self.top_screen )
    else
        idx = #self.screenstack
    end
    screen = table.remove( self.screenstack, idx )
    
    screen:OnClose()

    if screen == self.top_screen then
        screen:OnLoseTop()

        local new_top = FindTopScreen( self.screenstack )
        if new_top then
            self.top_screen = new_top

            -- Notify any overlays on top, and the new top screen itself.
            for i = #self.screenstack, 1, -1 do
                local screen = self.screenstack[i]
                screen:OnGainTop( new_top )
                
                if not screen.removed then
                    self:RequestControlRefresh()
                end

                if screen == new_top then
                    break
                end
            end
        else
            self.top_screen = nil
        end
    end
    
    screen:Remove()
   
    screen = nil

    self:RefreshMusic()
    self:UpdateScreenViz()
end

function FrontEnd:GetTopScreen()
    return self.top_screen
end

function FrontEnd:ClearFade()
    self.fader:Hide()
    self.fading = false
    self.last_fade_percent = nil
    self.fade_screen = nil
    self.fade_time = nil
    self.fade_time_total = nil
    self.fade_out = nil
end

function FrontEnd:UpdateFade( dt )
	local alpha = 0

	-- If starting a fade-in, delay the actual fade until all async loading is completed
--	if self.fade_time == 0 and self.fade_out == false and not engine.inst:IsAsyncLoadingComplete() then	
	if not self.last_fade_percent and not self.fade_out and not engine.inst:IsAsyncLoadingComplete() then	
		alpha = 1
	else
		self.fade_time = self.fade_time + math.min(dt, 1/30)
		local fade_percent = self.fade_time / self.fade_time_total
    
		if (self.last_fade_percent or 0) < 1 then
			alpha = self.fade_out and easing.outQuad(fade_percent, 0, 1, 1) or easing.inQuad(fade_percent, 1, -1, 1)
			self.last_fade_percent = fade_percent
		else
        
			if self.fade_out then
				local fade_fn = self.fade_fn
				self.fade_fn = nil
				self.fade_time = 0
				self.last_fade_percent = nil
				alpha = 1
				self.fade_out = nil
				if fade_fn then
					fade_fn( self )
				end

			else
				self:ClearFade()
			end
		end
	end

	self.fader:SetColour(0,0,0,alpha)
end

function FrontEnd:UpdateScreenFade( dt )

    self.fade_time = self.fade_time + math.min(dt, 1/30)
    local fade_percent = self.fade_time / self.fade_time_total

    if fade_percent < 1 then
        local alpha = easing.outQuad(fade_percent, 1, -1.0, 1)
        self.fade_screen:SetTintAlpha( alpha )

    else
        local fade_fn = self.fade_fn
        self:ClearFade()

        if fade_fn then
            fade_fn( self )
        end
    end
end

function FrontEnd:CheckMouseHover( x, y, trace )
    local mouse_blocked, hover_widget
    for i = #self.screenstack, 1, -1 do
        local screen = self.screenstack[ i ]
        mouse_blocked, hover_widget = screen:CheckMouseHover(x,y,trace)
        if mouse_blocked or hover_widget or screen:SinksInput() then
            break
        end
    end
    if hover_widget then
        hover_widget:SetHover()
    else
        self.fe_root:ClearHover()
    end
end

function FrontEnd:Update(dt)
    engine.inst:ProfilerPush("FE:Update")
    if TheGame:GetLocalSettings().ROBOTICS then
        if TheGame:GetRoboticsSettings().status_string then
            self.build_version:SetTintColour(0xffff00ff)
            self.build_version:SetText(TheGame:GetRoboticsSettings().status_string):LayoutBounds( "left", "bottom", 0, 0 )
        end
    end


    if self.pending_screen_mode then
        if self.pending_screen_mode ~= self.screen_mode then
            print ("SWITCHING SCREEN MODE:", self.screen_mode, " -> ", self.pending_screen_mode )
            self.screen_mode = self.pending_screen_mode
            self.bottom_bar:OnScreenModeChange( self.screen_mode )
                :SetPos( 0, nil )
            self.saving_anim:OnScreenModeChange( self.screen_mode )
                :LayoutBounds( "left", "center" )
                :Offset( 60, -RES_Y*0.4 )
            for k,v in pairs(self.screenstack) do
                v:OnScreenModeChange( self.screen_mode )
            end
            self.pending_screen_mode = nil
        end
    end
    
    local sx, sy = TheGame:GetInput():GetMousePos()
    local last_mouse_device = TheGame:GetInput():GetLastMouseDevice()
    if self.last_mouse_x -- There is a past cursor position
        and (self.last_mouse_x ~= sx or self.last_mouse_y ~= sy or self.last_mouse_device ~= last_mouse_device) then -- And the new one is different
        self:SetPendingControlMode( last_mouse_device == "touch" and CONTROL_MODE.TOUCH or CONTROL_MODE.MOUSE_KEYBOARD, nil )
    end
    self.last_mouse_x, self.last_mouse_y = sx, sy
    self.last_mouse_device = last_mouse_device

    if self.pending_control_mode or self.pending_device_id then
        if self.pending_control_mode ~= self.control_mode or self.pending_device_id ~= self.device_id then
            print ("SWITCHING CONTROL MODE:", self.control_mode, self.device_id, " -> ", self.pending_control_mode, self.pending_device_id)
            
            self.control_mode = self.pending_control_mode
            self.device_id = self.pending_device_id
            self.bottom_bar:OnControlModeChange( self.control_mode, self.device_id )
                :SetPos( 0, nil )

            -- Any screens down to the top-most one should be given a chance to refresh.
            -- Screens below that will have OnControlModeChange called when they re-GainTop.
            local top = FindTopScreen( self.screenstack )
            for i = #self.screenstack, 1, -1 do
                local screen = self.screenstack[i]
                screen:OnControlModeChange( self.control_mode, self.device_id )
                if screen == top then
                    break
                end
            end
            engine.inst:SetCursorVisible(self.control_mode ~= CONTROL_MODE.GAMEPAD)

            if self.control_mode ~= CONTROL_MODE.MOUSE_KEYBOARD then
                self.fe_root:ClearHover()
            end
            if self.control_mode == CONTROL_MODE.GAMEPAD then
                local screen = self:GetTopScreen()
                if screen then
                    screen:SetDefaultFocus()
                end
            else
                TheGame:GetInput():ClearRumble()
            end
            self:RequestControlRefresh()
        end
        self.pending_control_mode = nil
        self.pending_device_id = nil
    end

    self:UpdateScreenViz()
    local t = TheGame:GetWallTime()
    self.timeconstant:SetValue(t)


    if self.fading then
        engine.inst:ProfilerPush("Screen Fade")
        if self.fade_screen then
            self:UpdateScreenFade( dt )
        else
            self:UpdateFade( dt )
        end
        engine.inst:ProfilerPop()
    end

        
    if self.enable then
        local x, y = self.game:GetInput():GetMousePos()
        
        engine.inst:ProfilerPush("Hover Checking")
        if not (TheGame.debug and TheGame.debug:WantCaptureMouse()) and not self.fading then
            if self.control_mode == CONTROL_MODE.MOUSE_KEYBOARD or self.control_mode == CONTROL_MODE.TOUCH then
                self:CheckMouseHover( x, y )
            end
        end
        engine.inst:ProfilerPop()

        engine.inst:ProfilerPush("Screen Update")
        for i = #self.screenstack, 1, -1 do
            local screen = self.screenstack[i]
            screen:OnUpdate( dt )
            if not screen:IsOverlay() then
                break
            end
        end
        engine.inst:ProfilerPop()

        if TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) then
            if not self.drag_root_x then
                self.drag_test_widget = self:GetHoverWidget()
                self.drag_root_x, self.drag_root_y = x, y
            
            elseif self.drag_test_widget and self.drag_test_widget.enabled then
                if math.abs(self.drag_root_x - x) + math.abs(self.drag_root_y - y) > 5 then
                    if self.drag_test_widget.StartDrag then
                        self.drag_test_widget:StartDrag()
                    end
                    self.drag_test_widget = nil
                end
            end
        else
            self.drag_root_x, self.drag_root_y = nil, nil
            self.drag_test_widget = nil
        end
    end


    engine.inst:ProfilerPush("Updating widgets")


    if self.drag then
        local drag = self.drag
        if TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) then
            self:UpdateDragPos()
        else
            local w = self:GetHoverWidget()
            local handled = false
            
            while w and not handled do
                if w.OnDrop then
                    handled = w:OnDrop(drag.w)
                end            
                w = w.parent
            end
            if handled then
                if drag.w.OnDropped then
                    drag.w.OnDropped(drag.w)
                end

                self:ClearDragWidget()
            else
                self:CancelDrag()
            end
        end
    end

    if next(self.newupdatingwidgets) then
        for k,v in pairs(self.newupdatingwidgets) do
            assert( not k.removed )
            self.updatingwidgets[k] = true
        end
        table.clear(self.newupdatingwidgets)
    end

    for k,v in pairs(self.updatingwidgets) do
        assert( not k.removed )
        k:OnUpdate(dt)
    end

    if not self.debugfreeze then
        self:UpdateToolTip(dt)
    end

    if self.wants_control_option_refresh then
        self:RefreshControlOptions()
        self.wants_control_option_refresh = false
    end

    self:UpdateReticule(dt)
    engine.inst:ProfilerPop()

    engine.inst:ProfilerPop()    
    
end


function FrontEnd:RequestControlRefresh()
    self.wants_control_option_refresh = true
end

function FrontEnd:ReticuleActive()
    return self.reticule_active
end

function FrontEnd:SetReticulePos(x, y)
    self.reticule_x, self.reticule_y = self.reticule_root:TransformFromWorld(x, y)    
    self.reticule_speed = 0
end

function FrontEnd:GetReticulePos(x, y)
    if self.reticule then
        return self.reticule:GetWorldPos()
    end
end

function FrontEnd:UpdateReticule(dt)
    
    local main_overlay = self:FindScreen( Screen.MainOverlay )
    local overlay_focus = main_overlay and main_overlay:TopBarHasFocus()

    local top = self:GetTopScreen()
    self.reticule_active = top and self.control_mode == CONTROL_MODE.GAMEPAD and top:WantsReticuleMode() and not self:IsFading() and not overlay_focus


    if self.reticule_active then
        if not self.reticule then
            local size = 110
            self.reticule = self.reticule_root:AddChild( Widget() )
                :SetTintColour( UICOLOURS.FOCUS )
            self.reticule_shadow = self.reticule:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
                :SetTintAlpha( 0 )
                :SetHiddenBoundingBox( true )
                :SetSize( size*6, size*6 )
            self.reticule_image = self.reticule:AddChild( Widget.Image( engine.asset.Texture( "UI/reticule.tex") ) )
                :SetSize( size, size )
                :Bloom( 0.1 )
            local aura_size = size*1.18
            self.reticule_layer_1 = self.reticule:AddChild( Widget.Image( engine.asset.Texture("icons/role_aura_layer1_0.tex"), aura_size, aura_size )):SetHiddenBoundingBox(true):SetBloom(0.5)
                :RotateIndefinitely( -0.01 )
                :Bloom( 0.1 )
            self.reticule_layer_2 = self.reticule:AddChild( Widget.Image( engine.asset.Texture("icons/role_aura_layer2_0.tex"), aura_size, aura_size )):SetHiddenBoundingBox(true):SetBloom(0.5)
                :RotateIndefinitely( 0.03 )
                :Bloom( 0.1 )
            self.reticule_layer_3 = self.reticule:AddChild( Widget.Image( engine.asset.Texture("icons/role_aura_layer3_0.tex"), aura_size, aura_size )):SetHiddenBoundingBox(true):SetBloom(0.5)
                :RotateIndefinitely( -0.025 )
                :Bloom( 0.1 )
        end

        if self.reticule_x == nil or self.reticule_y == nil then
            if top.GetReticulePosition then
                self:SetReticulePos( top:GetReticulePosition()  )
            else
                self.reticule_x, self.reticule_y = 0, 0
            end
        end

         self.reticule:Show()

        local _, _, lx, ly = TheGame:GetInput():GetActiveGamepadStickVals()
        
        local MIN_RAD = .1
        local MAX_RAD = .8

        local inner = lx*lx+ly*ly
        local rad = math.abs(inner) < .0001 and 0 or math.sqrt(inner)
        
        local wx,wy = self.reticule:TransformToWorld(0,0)
        local hover_widget = top:TestReticule(wx, wy)
        local speed_factor = hover_widget ~= nil and 0.6 or 1.0
        self.reticule:ScaleTo( nil, hover_widget and 1 or .55, 0.1 )

        if hover_widget then
            self:SetFocusWidget(hover_widget)

            self.reticule_image:Bloom( 0.15 )
            self.reticule_image:AlphaTo( 0.35, 0.1 )
            self.reticule_layer_1:RotateIndefinitely( -0.015 )
            self.reticule_layer_2:RotateIndefinitely( 0.05 )
            self.reticule_layer_3:RotateIndefinitely( -0.04 )
            self.reticule_layer_2:ScaleTo( nil, 1.25, 0.3 )
            self.reticule_layer_3:ScaleTo( nil, 1.35, 0.2 )
            self.reticule_shadow:AlphaTo( 0, 0.1 )
        else
            self:ClearFocusWidget()

            self.reticule_image:Bloom( 0.1 )
            self.reticule_image:AlphaTo( 1, 0.1 )
            self.reticule_layer_1:RotateIndefinitely( -0.01 )
            self.reticule_layer_2:RotateIndefinitely( 0.03 )
            self.reticule_layer_3:RotateIndefinitely( -0.025 )
            self.reticule_layer_2:ScaleTo( nil, 1, 0.3 )
            self.reticule_layer_3:ScaleTo( nil, 1, 0.4 )
            self.reticule_shadow:AlphaTo( 0.2, 0.3 )
        end

        local turbo = not (self.game:GetInput():IsControlDown(Controls.Digital.GAMEPAD_LT) or self.game:GetInput():IsControlDown(Controls.Digital.GAMEPAD_RT))
        if turbo then
            self.reticule_image:SetBloom( 0.3 )
        else
            self.reticule_image:SetBloom( 0.1 )
        end

        if rad > MIN_RAD then
            
            local MAX_SPEED = turbo and RES_X or RES_X/2
            local FORCE = turbo and 6000 or 3000
            local MASS = 1
            local ACCEL = FORCE/MASS
            local throttle = clamp(.2+ rad / MAX_RAD, 0, 1)
            local angle = math.atan2(ly, lx)
            self.reticule_speed = self.reticule_speed or 0
            self.reticule_speed = clamp( self.reticule_speed + throttle*ACCEL*dt, 0, throttle*MAX_SPEED*speed_factor)

            local x_speed = math.cos(angle)*self.reticule_speed*dt
            local y_speed = math.sin(angle)*self.reticule_speed*dt

            local min_x, max_x, min_y, max_y
            
            if top.GetReticuleBounds then
                min_x, min_y, max_x, max_y = top:GetReticuleBounds()
            else
                -- NOTE: Default reticule bounds excludes top bar area for MainOverlay.
                local screen_w, screen_h = TheGame:FE():GetScreenDims()
                local w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
                local h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )
                min_x, max_x = -w/2, w/2
                min_y, max_y = -h/2 + 80, h/2 - 80
            end

            self.reticule_x = clamp(self.reticule_x + x_speed, min_x, max_x)
            self.reticule_y = clamp(self.reticule_y + y_speed, min_y, max_y)
        else
            self.reticule_speed = 0

            if self.tooltip_focus and not self.tooltip_focus.no_reticule_snap then
                if hover_widget and (self.tooltip_focus == hover_widget or self.tooltip_focus:IsAncestorOf( hover_widget )) then
                    self.reticule_x, self.reticule_y = self.reticule_root:TransformFromWidget( hover_widget, 0, 0 )
                end
            end
        end

        self.reticule:SetPos(self.reticule_x, self.reticule_y)        

    else
        if self.reticule then
            self.reticule:Hide()
        end
        -- It's really annoying that the reticule position isn't preserved just because eg. a popup appears.
        -- self.reticule_x, self.reticule_y = nil, nil
    end
end

function FrontEnd:IsFading()
    return self.fading
end

function FrontEnd:UpdateToolTip(dt)
    if self.fading then
        if self.tooltip_widget then
            self.tooltip_widget:Hide()
        end
        return
    
    end
    local focus
    if self.tooltip_focus_override then
        if self.tooltip_focus_override.removed then
            self.tooltip_focus_override = nil
        else
            focus = self.tooltip_focus_override
        end
    end

    
    if focus == nil then
        -- Get hover widget
        focus = self:GetHoverWidget()
        -- If the current focused widget should show tooltip on focus, do that
        local focus_widget = self:GetFocusWidget()
        if focus_widget then
            if focus_widget.show_tooltip_on_focus then
                focus = focus_widget
            else
                -- If any ancestor of focus_widget is flagged 'show_child_tooltip_on_focus', then also show tooltip.
                local focus_parent = focus_widget.parent
                while focus_parent do
                    if focus_parent.show_child_tooltip_on_focus then
                        focus = focus_widget
                        break
                    end
                    focus_parent = focus_parent.parent
                end
            end
        end
    end

    local tt_class = Widget.Tooltip
    local tt

    if focus then
        while tt == nil and focus do
            tt = focus:GetToolTip()
            tt_class = focus:GetToolTipClass() or Widget.Tooltip
            if not tt then
                focus = focus.parent
            end
        end
    end
    

    if tt_class and self.tooltip_widgets[ tt_class ] == nil then
        --print( "Creating new shared tooltip: ", tt_class._classname )
        self.tooltip_widgets[ tt_class ] = self.fe_root:AddChild( tt_class():IgnoreInput( true ) )
    end
    local tooltip_widget = self.tooltip_widgets[ tt_class ]

    -- Should we hide the old tooltip?
    if self.tooltip_widget ~= nil and (self.tooltip_widget ~= tooltip_widget or not tt) then
        
        self.tooltip_delay = self.tooltip_delay - dt
        
        
        if self.tooltip_delay <= 0 or (self.tooltip_focus.removed or not self.tooltip_focus.shown) then
            if self.tooltip_widget.OnHideToolTip then
                self.tooltip_widget:OnHideToolTip()
            else
                self.tooltip_widget:Hide()
            end
            self.tooltip_widget, self.tooltip_focus, self.tooltip_data = nil, nil, nil
            self.tooltip_delay = TTDELAY
        end
    end

    -- Should we show a new tooltip?
    if tt and (focus ~= self.tooltip_focus or tt ~= self.tooltip_data or focus:GetToolTipDirty()) then
        self.tooltip_delay = self.tooltip_delay - dt

        if self.tooltip_delay <= 0 or (self.tooltip_widget and self.tooltip_widget.shown) then
            local sm = self:GetScreenMode()
            local layout_scale
            if tooltip_widget.LAYOUT_SCALE then
                layout_scale = tooltip_widget.LAYOUT_SCALE[ sm ] or LAYOUT_SCALE[ sm ]
            else
                layout_scale = LAYOUT_SCALE[ sm ]
            end
            tooltip_widget:SetLayoutScale( layout_scale )
            if tooltip_widget:DoLayout( tt ) ~= false then
                self.tooltip_widget = tooltip_widget
                self.tooltip_widget.nodebug = true
                self.tooltip_widget:Show()
                self.tooltip_widget:SetScale(self.base_scale,self.base_scale)
                self.tooltip_focus = focus
                self.tooltip_data = tt
                self.tooltip_delay = TTDELAY
                focus:SetToolTipDirty( nil )
            else
                tooltip_widget:Hide()
            end
        end
    end

    if self.tooltip_widget and self.tooltip_focus and self.tooltip_widget.shown and not self.tooltip_focus.removed then
        self:UpdateToolTipPos( self.tooltip_widget )
        -- if self.tooltip_focus.LayoutToolTip then
        --     self.tooltip_focus:LayoutToolTip( self.tooltip_widget )
        -- end
        if self.tooltip_focus.tooltiplayoutfn then
            self.tooltip_focus.tooltiplayoutfn( self.tooltip_focus, self.tooltip_widget )
        end
    end

end

function FrontEnd:SetToolTipOverride( tooltip_focus )
    self.tooltip_focus_override = tooltip_focus
end

function FrontEnd:GetToolTipOverride()
    return self.tooltip_focus_override
end

function FrontEnd:UpdateToolTipPos( tooltip_widget )
    if tooltip_widget and tooltip_widget.shown then
        local screenw, screenh = self:GetScreenDims()
        local xmin, ymin, xmax, ymax = tooltip_widget:GetWorldBoundingBox()
        local tw, th = xmax - xmin, ymax - ymin

        local xmin, ymin, xmax, ymax = self.tooltip_focus:GetWorldBoundingBox()

        if screenw - xmax >= tw and ymax >= th then
            tooltip_widget:LayoutBounds( "after", "top", self.tooltip_focus )

        elseif xmin >= tw and ymax >= th then
            tooltip_widget:LayoutBounds( "before", "top", self.tooltip_focus )

        elseif screenw - xmax >= tw and screenh - ymin >= th then
            tooltip_widget:LayoutBounds( "after", "bottom", self.tooltip_focus )

        elseif xmin >= tw and screenh - ymin >= th then
            tooltip_widget:LayoutBounds( "before", "bottom", self.tooltip_focus )

        elseif screenh - ymax >= th then
            tooltip_widget:LayoutBounds( "center", "above", self.tooltip_focus )

        else
            tooltip_widget:LayoutBounds( "center", "below", self.tooltip_focus )
            self:ConstrainToolTipPos( tooltip_widget )
        end
    end
end

function FrontEnd:ConstrainToolTipPos( tooltip_widget )
    UIHelpers.KeepOnScreen(tooltip_widget)
end

function FrontEnd:UpdateDragPos()
    if self.drag then
        local mx, my = self.drag.w:TransformFromWorld(TheGame:GetInput():GetMousePos())
        local x,y = self.drag.w:GetPos()
        local new_y = y + my - self.drag.drag_offset_y
        local new_x = x + mx - self.drag.drag_offset_x
        self.drag.w:SetPos(new_x, new_y)
    end

end

function FrontEnd:GetScreenDims()
    return engine.inst:GetScreenSize()
end

-- Returns the closest aspect ratio to the current window size
function FrontEnd:GetAspectRatio()
    -- Get the current screen size
    local w, h = self:GetScreenDims()
    local aspects = {}
    -- Go through our preset aspect ratios, and check their difference to our current one
    for k,v in pairs( ASPECT_RATIO_VALUE ) do
        local difference = math.abs( v - (w/h) )
        table.insert( aspects, { aspect_ratio = k, diff = difference } )
        -- print( "Aspect Ratio:", k, "Current Aspect:", v, "Difference:", difference )
    end
    -- Sort all the differences
    table.sort( aspects, function( a, b ) return a.diff < b.diff end)
    return( aspects[1].aspect_ratio )
end

function FrontEnd:StartUpdatingWidget(w)

    assert( not w.removed )
    -- screens have their own update.
    if not is_instance (w, Widget.Screen) then 
        self.newupdatingwidgets[w] = true
    end
end

function FrontEnd:StopUpdatingWidget(w)
    -- nil-check before assignment; even assignment to nil will cause a rehash and
    -- invalidate self.updatingwidgets iteration during OnUpdate!
    if self.updatingwidgets[w] then
        self.updatingwidgets[w] = nil
    end
    self.newupdatingwidgets[w] = nil
end

function FrontEnd:GetBaseWidgetScale()
    return self.base_scale
end

function FrontEnd:OnRender(dt)
    engine.inst:ProfilerPush( "FrontEnd:OnRender")

    engine.inst:ProfilerPush( "FrontEnd:OnRender: Updating Transforms")
    for k,v in pairs(self.dirtytransforms) do
        if not v.removed then
            v:UpdateTransform()
        end
    end
    table.clear(self.dirtytransforms)
    engine.inst:ProfilerPop()

    local top_opaque = 1
    for k = #self.screenstack, 1, -1 do
        local screen = self.screenstack[k]
        if screen:IsOpaque() then
            top_opaque = k
            break
        end
    end

    for k = top_opaque, #self.screenstack do
        local screen = self.screenstack[k]
        if screen.shown and screen.DoRender then
            screen:DoRender(dt)
        end
    end

    engine.inst:ProfilerPop()
end

function FrontEnd:IsHUDVisible( flag )
    if flag then
        -- Visible if this flag is on, OR we are not hidden.
        return CheckBits( self.hud_flags, flag ) or not CheckBits( self.hud_flags, HUD_FLAGS.HIDDEN )
    else
        return not CheckBits( self.hud_flags, HUD_FLAGS.HIDDEN )
    end
end

function FrontEnd:IsHUDFlagged( flag )
    return CheckBits( self.hud_flags, flag )
end

function FrontEnd:ToggleHUD( flags )
    self.hud_flags = ToggleBits( self.hud_flags, flags )
    print( "ToggleHUD:", StringizeBitField( self.hud_flags, HUD_FLAGS_STRINGS ))

    local visible = self:IsHUDVisible()

    for i = #self.screenstack, 1, -1 do
        local screen = self.screenstack[i]
        if screen.ToggleHUD then
            screen:ToggleHUD( visible, self.hud_flags )
        end
    end
end

function FrontEnd:SetShaderConstant(constant)
    self.scene:SetShaderConstant(constant)
    return constant
end

function FrontEnd:ShowAchievementToast( icon, title, description )
    self.queued_toasts = self.queued_toasts or {}

    -- Add toast to the queue (last element on queue)
    table.insert( self.queued_toasts, { icon = icon, title = title, description = description } )
    
    if #self.queued_toasts == 1 then
        -- If we're not showing a toast, show one
        self:_DisplayToast()
    end

    return self
end

function FrontEnd:_DisplayToast()
    -- Get first toast on the list
    local toast_data = self.queued_toasts[1]

    -- Prepare toast widget
    local toast_widget = self.fe_root:AddChild( Widget.ToastWidget( toast_data.icon, toast_data.title, toast_data.description ) )
        :SetBackgroundColor( UICOLOURS.SUBTITLE )
        :SetAnchors( "center", "top" )
        :SetTintAlpha( 0 )
        :SetLayoutScale( LAYOUT_SCALE[self.screen_mode] )
        :SetScale( 0.5, 0.5 )

    -- Display toast widget in all its glory
    local timing = 0.3
    toast_widget:RunUpdater(
        Updater.Series{
            Updater.Parallel{
                -- Move in the toast
                Updater.Ease( function(v) toast_widget:SetScale(v) end, 0.5, 1, timing, easing.outQuad),
                Updater.Ease( function(v) toast_widget:SetTintAlpha(v) end, 0, 1, timing, easing.outQuad),
                Updater.Ease2D( function(x,y) toast_widget:SetPos(x,y) end, 280*LAYOUT_SCALE[self.screen_mode], -40*LAYOUT_SCALE[self.screen_mode], 0, -80*LAYOUT_SCALE[self.screen_mode], timing, easing.outQuad),
            },
            Updater.Do( function() toast_widget:Animate() end ),
            Updater.Wait( 3 ),
            Updater.Parallel{
                -- Move out the toast
                Updater.Ease( function(v) toast_widget:SetScale(v) end, 1, 0.5, timing, easing.inQuad),
                Updater.Ease( function(v) toast_widget:SetTintAlpha(v) end, 1, 0, timing, easing.inQuad),
                Updater.Ease2D( function(x,y) toast_widget:SetPos(x,y) end, 0, -80*LAYOUT_SCALE[self.screen_mode], -280*LAYOUT_SCALE[self.screen_mode], -40*LAYOUT_SCALE[self.screen_mode], timing, easing.inQuad),
            },
            Updater.Do( function()
                -- Remove the toast widget
                toast_widget:Remove()
                -- And the toast data from the queue (first element on queue)
                table.remove( self.queued_toasts, 1 )
                -- If there are toasts in the queue, show the next one
                if #self.queued_toasts > 0 then
                    self:_DisplayToast()
                end
            end )
        }
    )
end

------------------------------------------------------------------------------------------
-- 

local BottomBarOverlay = class( "Widget.BottomBarOverlay", Widget )

local BAR_HEIGHT = 50

function BottomBarOverlay:init( fe )
    BottomBarOverlay._base.init( self )
    self.fe = fe
    self.h = 50

    self.bg = self:AddChild( Widget.Image( engine.asset.Texture("UI/controls_bottom_bar.tex") ) )

    -- LEFT
    self.left_label = self:AddChild( Widget.Label( "title", 30 ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    -- RIGHT
    self.right_label = self:AddChild( Widget.Label( "title", 30 ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )

    -- Refresh control mode display
    self:OnControlModeChange( fe:GetControlMode(), fe:GetControlDeviceID() )

    self:SetSize()
end

function BottomBarOverlay:OnControlModeChange( cm, device_id )
    BottomBarOverlay._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.GAMEPAD then
        if self.has_content then self:FadeIn() end
    else
        self:FadeOut()
    end
    return self
end

function BottomBarOverlay:OnScreenModeChange( sm )
    BottomBarOverlay._base.OnScreenModeChange( self, sm )

    self.left_label:SetFontSize( 30*LAYOUT_SCALE[sm] )
    self.right_label:SetFontSize( 30*LAYOUT_SCALE[sm] )

    self.h = BAR_HEIGHT*LAYOUT_SCALE[sm]
    
    self:SetSize()
    return self
end

function BottomBarOverlay:FadeIn()
    self.visible = true
    self:Layout()
        :SetTintAlpha(0)
        :Show()
        :AlphaTo( 1, 0.2, easing.outQuad )
        :MoveTo( nil, self.h/2, 0.2, easing.outQuad )
    return self
end

function BottomBarOverlay:FadeOut()
    self.visible = false
    self:AlphaTo( 0, 0.1, easing.outQuad, function() self:Hide() end )
        :MoveTo( nil, -self.h, 0.2, easing.outQuad )
    return self
end

function BottomBarOverlay:OnScreenResize( w, h )
    Widget.OnScreenResize( self, w, h )
    self:SetSize()
    return self
end

function BottomBarOverlay:SetSize()
    local screen_w, screen_h = self.fe:GetScreenScissorDims()
    self.w = math.ceil( screen_w / self.fe:GetBaseWidgetScale() )
    self.h = self.h or BAR_HEIGHT

    self.bg:SetSize( self.w, self.h )
    self:SetPos( 0, self.h/2 )

    self:Layout()
    return self
end

function BottomBarOverlay:Layout()

    -- LEFT
    self.left_label:LayoutBounds( "left", "center", self.bg )
        :Offset( 30, 0 )

    -- RIGHT
    self.right_label:LayoutBounds( "right", "center", self.bg )
        :Offset( -30, 0 )

    return self
end

function BottomBarOverlay:SetControlOptions( left, right )
    self.has_content = false

    if left then
        if type( left ) == "string" and left ~= "" then
            self.has_content = true
            self.left_label:SetText( left )
        elseif type( left ) == "table" and next(left) ~= nil then
            self.has_content = true
            self.left_label:SetText( table.concat( left, "    " ) )
        else
            self.left_label:SetText( "" )
        end
    else
        self.left_label:SetText( "" )
    end

    if right then
        if type( right ) == "string" and right ~= "" then
            self.has_content = true
            self.right_label:SetText( right )
        elseif type( right ) == "table" and next(right) ~= nil then
            self.has_content = true
            self.right_label:SetText( table.concat( right, "    " ) )
        else
            self.right_label:SetText( "" )
        end
    else
        self.right_label:SetText( "" )
    end

    if self.has_content then
        self:Layout()
        if not self.visible and TheGame:FE():IsGamepadMode() then self:FadeIn() end
    else
        if self.visible then self:FadeOut() end
    end
    return self
end
