require "ui/widgets/checkbox"


-------------------------------------------------------------------------
--

local MutatorPopupOption = class( "Widget.MutatorPopupOption", Widget )

function MutatorPopupOption:init( def, checked )
    Widget.init( self )
    self.features = {}

    self.graft = GraftInstance( def.id )

    local icon = self:AddChild( Widget.Image( def.img, 48, 48 ))
    local check = self:AddChild( Widget.CheckBox( 32 ):SetText( def.name ))
        :LayoutBounds( "after", "top", icon ):Offset( 16, 0 ):SetValue( checked, true )
    local label = self:AddChild( Widget.Label( "title", 24, self.graft:GetDesc( self )))
        :LayoutBounds( "left", "below", check ):Offset( 32, 0 )
    
    self.check = check
    self:SetToolTipClass( Widget.TooltipGraft ):SetToolTip( self.graft )
end

function MutatorPopupOption:IsChecked()
    return self.check:GetValue()
end

function MutatorPopupOption:SetValue( state )
    return self.check:SetValue( state, true )
end

function MutatorPopupOption:SetFn( fn )
    self.check:SetFn( fn )
end

function MutatorPopupOption:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function MutatorPopupOption:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

-------------------------------------------------------------------------
--

local ChooseMutatorsPopup = class("Screen.ChooseMutatorsPopup", Widget.Screen)

local PANEL_PADDING = 30

function ChooseMutatorsPopup:init( mutators, fn )
    Widget.Screen.init(self)

    self.is_popup = true
    self.on_close = fn

    -- Get screen dimensions
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )
    -- Content sizes
    local default_w = RES_X*.36
    local title_sz = 42
    
    -- A container for all the background elements
    self.scrim = self:AddChild( Widget.SolidBox( RES_X, RES_Y, 0x021A18F0 ) )--:BlockMouse()
    self.panel_bg = self:AddChild( Widget.Image( engine.asset.Texture("large/dialog_bg.tex" ) ) )

    self.label_title = self:AddChild( Widget.Label("title", title_sz) ):SetAutoSize( default_w ):SetText( "Choose Mutators" ):SetGlyphColour(UICOLOURS.SUBTITLE)

    self.widget_content = self:AddChild( Widget.ScrollPanel() )

    self.checkboxes = {}
    for i, def in ipairs( Content.GetAllGrafts() ) do
        if def.type == GRAFT_TYPE.MUTATOR then
            local checked = mutators and table.arrayfind( mutators, def.id )
            self:AddOption( def.id, def, checked )
        end
    end

    self.random_btn = self:AddChild( Widget.PanelButton( "Randomize 3" ))
    self.random_btn:SetFn( function()
        self:RandomizeMutators( 3 )
    end )

    self.button_ok = self:AddChild( Widget.PanelButton( "Okay" ))
    self.button_ok:SetFn( function()
        self:GetFE():PopScreen( self )
    end )

    self:SetAnchors( "center", "center" )
    self:SetSounds( SoundEvents.tut_popup, SoundEvents.tut_popout )
    self:RegisterHotkeyButton( Controls.Digital.MENU_CANCEL, self.button_ok )

    self:Layout()
end

function ChooseMutatorsPopup:AddOption( id, def, checked )
    local w = Widget.MutatorPopupOption( def, checked )

    self.widget_content:AddScrollChild( w ):LayoutBounds( "left", "below" ):Offset( 0, -20 )

    self.checkboxes[ id ] = w
    w:SetFn( function( state ) self:OnOptionChanged( id, state ) end )
end

function ChooseMutatorsPopup:OnOptionChanged( id, state )
    if state then
        -- Turn off mutually exclusive options.
        local def = Content.GetGraft( id )
        if def.exclusion_ids then
            for i, id in ipairs( def.exclusion_ids ) do
                self.checkboxes[ id ]:SetValue( false )
            end
        end
    end
end

function ChooseMutatorsPopup:RandomizeMutators( count )

    -- Clear existing.
    for id, w in pairs( self.checkboxes ) do
        w:SetValue( false )
    end

    local mutator_defs = {}
    for i, def in ipairs( Content.GetAllGrafts() ) do
        if def.type == GRAFT_TYPE.MUTATOR then
            table.insert( mutator_defs, def )
        end
    end

    -- This is not completely correct, as depending on the selection it is possible
    -- to pick LESS than 3 due to mutually exclusive options.
    while #mutator_defs > 0 and count > 0 do
        local def = table.remove( mutator_defs, math.random( #mutator_defs ))
        self.checkboxes[ def.id ]:SetValue( true )
        count = count - 1

        if def.exclusion_ids then
            for i, id in ipairs( def.exclusion_ids ) do
                local idx = table.findif( mutator_defs, function( x ) return x.id == id end )
                if idx then
                    table.remove( mutator_defs, idx )
                end
            end
        end
    end

end

function ChooseMutatorsPopup:OnClose()
    if self.on_close then
        self.on_close( self:GenerateMutatorsTable() )
    end
end

function ChooseMutatorsPopup:GenerateMutatorsTable()
    local t = {}
    for id, w in pairs( self.checkboxes ) do
        if w:IsChecked() then
            table.insert( t, id )
        end
    end

    table.sort( t )
    return t
end

function ChooseMutatorsPopup:Layout()
    -- Get screen dimensions
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    self.scrim:SetSize( self.screen_w, self.screen_h )
    self.widget_content:SetSize( self.screen_w * 0.5, self.screen_h * 0.6 )
    self.widget_content:RefreshView()

    -- Size up the background
    local _, h = self.widget_content:GetSize()
    h = h + PANEL_PADDING*2
    self.panel_bg:SetSize( self.screen_w, h )

    self.label_title:LayoutBounds( "center", "above", self.panel_bg )
    -- Layout the content
    self.button_ok:LayoutBounds("center","below", self.panel_bg ):Offset( 128, -16 )
    self.random_btn:LayoutBounds( "center", "below", self.panel_bg ):Offset( -128, -16 )
end

function ChooseMutatorsPopup:OnScreenResize(w, h)
    ChooseMutatorsPopup._base.OnScreenResize(self, w, h)
    self:Layout()
end
