require "ui/widgets/screen"

----------------------------------------------------------------------
-- Shows a list of all the achievements

local AchievementsScreen = class("Screen.AchievementsScreen", Widget.Screen)

local ACHIEVEMENT_WIDTH = 500 -- Width of a single achievement widget
local ACHIEVEMENT_HEIGHT = 140 -- Height of a single achievement widget

AchievementsScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_CANCEL,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end,
        fn = function(self)
            self.close_button:Click()    
            return true
        end,
    }
}

function AchievementsScreen:init()
	Widget.Screen.init( self )

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Header
    self.header = self:AddChild( Widget() ):SetAnchors( "left", "top" )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE, LOC"UI.ACHIEVEMENTS_SCREEN.TITLE" ) )
        :SetGlyphColour( UICOLOURS.QUEST )
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.1 )
        :LayoutBounds( "left", "top", 60, -60 )
    self.subtitle = self.header:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.QUEST_LIGHT )
        :LeftAlign()
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )

    -- Backing
    self.widepanel = self:AddChild( Widget.WidePanel( 800, UICOLOURS.QUEST ) )

    -- Scroll panel for all AchievementWidgets
    self.content_scroll = self:AddChild( Widget.ScrollPanel() )
        :SetAnchors( "center", "center" )
        :SetVirtualMargin( SPACING.M1 )
        :SetScrollBarOuterMargin( 0 )
    self.content = self.content_scroll:AddScrollChild( Widget() )

    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ))
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
end

function AchievementsScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function AchievementsScreen:OnAdded()
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    self:RefreshAchievements()
end

function AchievementsScreen:OnGainTop()
    AchievementsScreen._base.OnGainTop(self)
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function AchievementsScreen:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function AchievementsScreen:OnScreenResize(w, h)
    AchievementsScreen._base.OnScreenResize(self, w, h)
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function AchievementsScreen:OnControlModeChange( cm, device_id )
    AchievementsScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_left:Show()
    end

    if cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
    end

    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
    end
end

function AchievementsScreen:OnScreenModeChange( sm )
    AchievementsScreen._base.OnScreenModeChange( self, sm )

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Scale contents
    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Scale achievements
    for i, child in self.content:Children() do
        if is_instance( child, Widget.AchievementWidget ) then
            child:SetLayoutScale( LAYOUT_SCALE[sm] )
        end
    end

    self:Layout()
end

function AchievementsScreen:GetDefaultFocus()
    return self.content.children and #self.content.children>0 and self.content.children[1]
end

function AchievementsScreen:RefreshAchievements()

    -- Get all achievements
    local all_achievements = Content.GetAllAchievements()
    local unlocked_achievements = TheGame:GetGameProfile():GetAchievementUnlocks() or {}
    if #unlocked_achievements == #all_achievements then
        -- All achievements complete!
        self.subtitle:SetText( LOC"UI.ACHIEVEMENTS_SCREEN.DESCRIPTION_ALL_COMPLETED" )
    elseif #unlocked_achievements > 0 then
        self.subtitle:SetText( loc.format( LOC"UI.ACHIEVEMENTS_SCREEN.DESCRIPTION_SOME_COMPLETED", #unlocked_achievements, #all_achievements ) )
        -- self.subtitle:SetText( LOC"UI.ACHIEVEMENTS_SCREEN.DESCRIPTION_NONE_COMPLETED" )
        -- self.subtitle:SetText( LOC"UI.ACHIEVEMENTS_SCREEN.DESCRIPTION_ALL_COMPLETED" )
    else
        self.subtitle:SetText( LOC"UI.ACHIEVEMENTS_SCREEN.DESCRIPTION_NONE_COMPLETED" )
    end

	self.content:DestroyAllChildren()

    for i, achievement in ipairs(all_achievements) do
        self.content:AddChild( Widget.AchievementWidget( achievement ) )
    end

    if self.content:IsEmpty() then
        if not self.empty_text then
            self.empty_text = self:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT, LOC"UI.ACHIEVEMENTS_SCREEN.EMPTY_TEXT" ) )
                :SetGlyphColour( UICOLOURS.DISABLED )
                :SetTintAlpha( 0.7 )
                :SetAutoSize( 600 )
                :SetWordWrap( true )
                :Bloom( 0.05 )
        end
        self.empty_text:Show()
    elseif self.empty_text then
        self.empty_text:Hide()
    end

    self:Layout()

    -- Position scrollable content
    self.content_scroll:RefreshView()
end

function AchievementsScreen:Layout()

    -- Layout header
    self.title:LayoutBounds( "left", "top", 60, -60 )
    self.subtitle:LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )

    -- Layout achievements
    local columns = 2
    local spacing = SPACING.M1/2
    local max_content_w = self.screen_w - 120*LAYOUT_SCALE[TheGame:FE():GetScreenMode()] -- Add some extra spacing for the left edge of the scroll panel
    local column_w = max_content_w/columns - (columns-1)*spacing
    column_w = column_w/LAYOUT_SCALE[TheGame:FE():GetScreenMode()] -- Compensate for the panel's UI scaling
    for i, child in self.content:Children() do
        if is_instance( child, Widget.AchievementWidget ) then
            child:SetWidth(column_w)
        end
    end
    self.content:LayoutChildrenInGrid( columns, spacing )
        :LayoutBounds( nil, "top", self.widepanel )
        :LayoutBounds( "left", nil, self.title )
        :Offset( 0, -SPACING.M1 )

    -- Header height
    local w, h = self.header:GetSize()
    local header_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[TheGame:FE():GetScreenMode()]
    -- Footer height
    local w, h = self.bottom_left:GetSize()
    local footer_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[TheGame:FE():GetScreenMode()]

    if self.empty_text then
        self.empty_text:LayoutBounds( "center", "top", self.widepanel )
            :Offset( 0, -SPACING.M05 )
    end

    -- Update containers
    self.widepanel:SetHeight( self.screen_h - header_height - footer_height )
        :LayoutBounds( nil, "below", self.header )
        :Offset( 0, -SPACING.M1 )
    self.content_scroll:SetSize( self.screen_w - SPACING.M1, self.screen_h - header_height - footer_height )
        :LayoutBounds( "left", "center", self.widepanel )
        :Offset( 0, 3 )

    return self
end

function AchievementsScreen:OnClickClose()
    self:FadeOut()
end

----------------------------------------------------------------------------------------------------------

local AchievementWidget = class( "Widget.AchievementWidget", Widget.Clickable )

function AchievementWidget:init( achievement_data )
	AchievementWidget._base.init( self )
    self:DisableAnimateScales()

	self.achievement_data = achievement_data
    self.unlocked = TheGame:GetGameProfile():IsAchievementUnlocked( self.achievement_data.id )

    -- Left badge
    self.bg_border = 20
    self.bg_border_scale = 0.5
    self.left = self:AddChild( Widget() )
    self.left.bg = self.left:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonright.tex") ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetScale( -1, -1 )
        :SetTintColour( self.unlocked and UICOLOURS.QUEST or UICOLOURS.SUBTITLE )
        :Bloom( 0.05 )
    self.left.circle = self.left:AddChild( Widget.Image( engine.asset.Texture ("images/circle.tex") ) )
        :SetTintColour( UICOLOURS.DARK_BODYTEXT )
        :SetTintColour( 0x492C4Fff )
    self.left.checkmark = self.left:AddChild( Widget.Image( engine.asset.Texture ("UI/ic_check.tex") ) )
        :SetTintColour( self.unlocked and UICOLOURS.QUEST or 0x2D1A31ff )
        :Bloom( 0.15 )

    -- Background panel
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsbutton_bg.tex") ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( self.unlocked and UICOLOURS.SUBTITLE or UICOLOURS.TITLE )
        :SetTintAlpha( self.unlocked and 0.1 or 0.1 )
        :Bloom( 0.05 )
    
    -- Achievement icon
    self.icon = self:AddChild( Widget.Image() )
        :SetTexture( self.unlocked and self.achievement_data.icon or self.achievement_data.locked_icon )
    -- Text
    self.text_container = self:AddChild( Widget() )
    -- Achievement title
    self.title = self.text_container:AddChild( Widget.Label("title", FONT_SIZE.ITEM_SUBTITLE, self.achievement_data:GetLocalizedName() or self.achievement_data.id ) )
        :LeftAlign()
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom(0.1)
    -- Achievement description
    self.text = self.text_container:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.achievement_data:GetLocalizedDesc() ) )
        :LeftAlign()
        :TopAlign()
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC*LINE_HEIGHT_TITLE_MULT )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
end

function AchievementWidget:SetWidth( w )
    self.width = w or self.width or 400 
    self.height = self.height or 90

    -- Layout left badge
    local left_width = self.height*0.9
    local nineslice_w, nineslice_h = left_width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.left.bg:SetSize( nineslice_w, nineslice_h )
    self.left:SetPos( 0, 0 )
    local circle_size = left_width*0.5
    self.left.circle:SetSize( circle_size, circle_size )
        :SetPos( -2.5, 0 )
    local check_size = circle_size*0.6
    self.left.checkmark:SetSize( check_size, check_size )
        :LayoutBounds( "center", "center", self.left.circle )

    -- Layout center panel
    local bg_width = self.width - left_width
    nineslice_w, nineslice_h = bg_width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.bg:SetSize( nineslice_w, nineslice_h )
        :LayoutBounds( "after", nil, self.left)

    -- Size up achievement icon
    local icon_size = left_width*0.7
    self.icon:SetSize( icon_size, icon_size )
        :LayoutBounds( "left", "center", self.bg )
        :Offset( SPACING.M1/2, 0 )

    -- Size up text
    local text_w = bg_width - SPACING.M1*1.5 - icon_size
    self.title:SetAutoSize( text_w )
    self.text:SetAutoSize( text_w )
        :LayoutBounds( "left", "below", self.title )
        :Offset( -1, 2 )
    self.text_container:LayoutBounds( "after", "center", self.icon )
        :Offset( SPACING.M1/2, 1 )

    return self
end

function AchievementWidget:UpdateImage()
    if ( self.hover or self.focus ) then
        if not self.earned then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.left.bg:Bloom( 0.08 )
        self.bg:Bloom( 0.08 )
    else
        self.left.bg:Bloom( 0.05 )
        self.bg:Bloom( 0.05 )
    end
end

function AchievementWidget:OnGainFocus()
    AchievementWidget._base.OnGainFocus(self)
    self:UpdateImage()
end

function AchievementWidget:OnLoseFocus()
    AchievementWidget._base.OnLoseFocus(self)
    self:UpdateImage()
end

function AchievementWidget:OnGainHover()
    AchievementWidget._base.OnGainHover(self)
    self:UpdateImage()
end

function AchievementWidget:OnLoseHover()
    AchievementWidget._base.OnLoseHover(self)
    self:UpdateImage()
end
