local path = require "util/filepath"
require "ui/widgets/screen"
require "ui/screens/fightscreenbase"
require "fightsim"
require "battle/scenario"
require "ui/widgets/unitframe"
require "ui/screens/debugtoolsmenu"
local battle_defs = require "battle/battle_defs"

local DebugUtil = require "debug/debug_util"
local debug_menus = require "debug/debug_menus"

local TARGET_DIST = 5
local MAX_UNDO = 100
local DATA_DIR = path.normslashes(path.normpath( path.join(BASE_PATH, "../data")))

-------------------------------------------------------------------

local DEFAULT_SCALE = 1
local blendmodes = {BLEND_MODE.BLENDED, BLEND_MODE.EMISSIVE, BLEND_MODE.MULTIPLY, BLEND_MODE.ADDITIVE, BLEND_MODE.PREMULTIPLIED}
local blendmodes_txt = {"BLENDED", "EMISSIVE", "MULTIPLY", "ADD (No!)", "PREMULT (No!)"}

local basic_anims = 
{
    "idle",
    "surrender",
    "death",
    "run_forward",
    "run_forward_pre",
    "run_forward_pst",
    "step_forward",
    "step_back",
    "taunt",
    "defend",
    "dodge",
    "stunned",
    "hit_mid",
    "stunned_pst",
    "run_forward_pre_turn",
    "run_forward_pst_turn",
    "hit_mid_pst_surrender",
}

local Y_POS = { 0, 4, -4, 8 }

local assets =
{
    [ VIS_STATUS.DEFAULT ] = global_images.black,
    [ VIS_STATUS.HIDE_IN_EDITOR ] = engine.asset.Texture("Tools/treeview_hidden.tex"),
    [ VIS_STATUS.HIDE_IN_GAME ] = engine.asset.Texture("Tools/treeview_box.tex"),
    [ VIS_STATUS.ALWAYS_SHOW ] = engine.asset.Texture("Tools/treeview_visible.tex"),
}

local function GetNextVisStatus( status )
    if status == VIS_STATUS.DEFAULT or status == nil then
        return VIS_STATUS.HIDE_IN_EDITOR
    elseif status == VIS_STATUS.HIDE_IN_EDITOR then
        return VIS_STATUS.HIDE_IN_GAME
    elseif status == VIS_STATUS.HIDE_IN_GAME then
        return VIS_STATUS.ALWAYS_SHOW
    else
        return nil
    end
end

local function GetClipboardColour()
    local txt = engine.inst:GetClipboardText()
    if type(txt) == "string" then
        return ParseColourString( txt )
    end
end

local function SetClipboardColour( clr )
    if clr then
        engine.inst:SetClipboardText( "0x"..MakeColourString( HexColour(clr) ))
    end
end

-------------------------------------------------------------------

local AttackTester = class("Screen.AttackTester", Screen.FightScreenBase)
AttackTester.IS_EDITOR = true

function AttackTester:init()
    AttackTester._base.init(self)
    
    self.loop_extra_time = 0
    self.target_dist = TARGET_DIST
    self.target_depth= 0
    self.num_targets = 1
    self.dodge_chance = 0

    self.target_entities = {}
    self.selected_values = {}
    self.dirty_files = {}

    self.vo_events = require "combat_vo_events"

    self:GenerateSoundEvents()
    self:GenerateHitTags()

    self:PopulateAgents()

    self.anim_files = self:GenerateAnimList()

    self.teams = {}
    self.teams[TEAM.RED] = Team(TEAM.RED)
    self.teams[TEAM.BLUE] = Team(TEAM.BLUE)

    self.solid_bg = Widget.SolidBox( RES_X, RES_Y, 0xFF0000FF ):Hide()
    self:AddChild( self.solid_bg:SetAnchors("fill", "fill") )
    self.solid_bg:SendBehind( self.img )
end


function AttackTester:GenerateAnimList( filter_str )
    local files = {}
    for k, filename in pairs(engine.inst:ListFiles( 'anim/*.zip')) do
        local filtered =   string.match(filename, "^head_") or
                string.match(filename, "^weapon_") or
                string.match(filename, "^med_") or
                string.match(filename, "_build$")

        filtered = filtered or (filter_str and filename:find( filter_str, nil, true ) == nil)

        if not filtered then
            table.insert( files, loc.format("anim/{1}", filename) )
        end 
    end

    return files
end

function AttackTester:GenerateSoundEvents( filter_str, is_looping )
    local sound_defs = require "sound_event_names"
    self.sound_events = {}
    for event, path in pairs( sound_defs ) do
        if filter_str == nil or event:find( filter_str, nil, true ) then
            if AUDIO:IsOneShot( path ) ~= is_looping then
                table.insert( self.sound_events, event )
            end
        end
    end
    table.sort( self.sound_events )
end

function AttackTester:GenerateHitTags()
    local all_attack_tags = { defended = true, small = true, large = true, evaded = true, ricochet = true }
    local function AddTags( t )
        if t then
            for k, tag in ipairs(t) do
                all_attack_tags[tag] = true
            end
        end
    end

    for i,v in ipairs(Content.GetAllBattleCards()) do
        AddTags( v.hit_tags )
    end

    for animfile, anims in pairs( GetAnimMetaTable() ) do
        for anim, v in pairs( anims ) do
            AddTags( v.hit_tags )
            if v.fx_tracks then
                for i, v in ipairs( v.fx_tracks ) do
                    AddTags( v.must_have_tags )
                    AddTags( v.cant_have_tags )
                end
            end
        end
    end

    self.all_attack_tags = {}
    for k,v in pairs(all_attack_tags) do
        table.insert(self.all_attack_tags, k)
    end

    table.sort(self.all_attack_tags)

    -- Add variant tags for current meta only
    self.current_attack_tags = shallowcopy( self.all_attack_tags )
end

function AttackTester:StripTags( tag )
    if self.presentation and self.presentation.anim_track then
        for i, v in ipairs( self.presentation.anim_track ) do
            while table.arrayremove( v.must_have_tags or table.empty, tag ) do end
            while table.arrayremove( v.cant_have_tags or table.empty, tag ) do end
        end
    end
end

function AttackTester:OnAdded()
    TheGame:GetDebug():RemoveBindingGroup( debug_menus.GAME_BINDINGS )
    TheGame:GetDebug():AddBindingGroup( debug_menus.TOOL_BINDINGS )

    DBG( DebugAttackAnims() )
end

function AttackTester:OnRemoved()
    TheGame:GetDebug():RemoveBindingGroup( debug_menus.TOOL_BINDINGS )
    TheGame:GetDebug():AddBindingGroup( debug_menus.GAME_BINDINGS )
end

function AttackTester:OnOpen()
    AttackTester._base.OnOpen(self)

    AUDIO:FinalizeVoiceLoad()
    PreloadFXTracks()
    
    self:ApplyColourCorrectionValues( (require "content/colour_settings")[self:GetColourTintName()] )

    local data = self:LoadDebugSettings()
    self.settings = data

    self:RefreshFighter( table.unpack( data.attack_agent ))
    self:RefreshTargetFighter( table.unpack( data.target_agent ))

    local attack_idx
    if data and data.attack_name then
        attack_idx = table.arrayfind( self.valid_attacks, data.attack_name )
    end
    self:SetAttackIDX( attack_idx or 1 )

    if data.plax_id then
        self.plax:SetPlaxData( data.plax_id )
    end

    self.solid_bg:SetColour( self.settings.bg_colour )
    self.solid_bg:SetShown( self.settings.bg_shown )

    self:ClearUndo()

    TheGame:SetWallTimeScale( self.paused and 0 or self.speed or 1.0 )
end

function AttackTester:GetDebugEnv( debug_env )
    debug_env.scenario = self.scenario
    debug_env.sim = self.fightsim
    debug_env.attacker = self.attack_entity
    debug_env.anim = self.attack_entity and self.attack_entity.cmp.AnimController
end


function AttackTester:LoadDebugSettings()
    local data = DebugUtil.GetLocalDebugData( "ATTACK_TESTER" ) or {}

    if data.attack_agent == nil then
        data.attack_agent = { "SAL" }
    end
    if data.attack_agent[2] then
        data.attack_agent[2] = Content.GetCharacterSkin( data.attack_agent[2] )
    end
    if data.target_agent == nil then
        data.target_agent = { "CRAYOTE" }
    end
    if data.target_agent[2] then
        data.target_agent[2] = Content.GetCharacterSkin( data.target_agent[2] )
    end

    if data.bg_colour == nil then
        data.bg_colour = 0xFFFFFFFF
    end

    self.num_targets = data.num_targets or 1
    self.target_dist = data.target_dist or TARGET_DIST
    self.target_depth = data.target_depth or 0
    self.looping = data.looping
    self.paused = data.paused
    self.hide_fx_gnomons = data.hide_fx_gnomons
    self.plax_id = data.plax_id
    self.editor_hit_tags = data.editor_hit_tags or {}

    return data
end

function AttackTester:SaveDebugSettings()
    local save_data = {}
    save_data.attack_agent = { self.attack_agent:GetContentID(), self.attack_agent:GetSkinID() }
    save_data.target_agent = { self.target_agent:GetContentID(), self.target_agent:GetSkinID() }
    
    save_data.num_targets = self.num_targets
    save_data.target_dist = self.target_dist
    save_data.target_depth = self.target_depth
    save_data.editor_hit_tags = self.editor_hit_tags

    save_data.bg_shown = self.settings.bg_shown
    save_data.bg_colour = self.settings.bg_colour
    
    save_data.looping = self.looping
    save_data.paused = self.paused
    save_data.hide_fx_gnomons = self.hide_fx_gnomons

    save_data.attack_name = self.valid_attacks[ self.attack_idx ]
    save_data.plax_id = self.plax_id
    save_data.show_card = self.cardw and self.cardw:IsShown()

    DebugUtil.SetLocalDebugData( "ATTACK_TESTER", save_data )
    DebugUtil.SaveLocalDebugData()
end


function AttackTester:PopulateAgents()
    if self.aliased_agents == nil then
        self.aliased_agents = {}
    else
        table.clear( self.aliased_agents )
    end

    self.agent_defs = {}
    self.player_defs = {}

    for content_id, v in pairs(Content.GetAllCharacterDefs()) do
        if v.base_def == "PLAYER_BASE" then
            table.insert( self.player_defs, content_id )
        elseif v.alias then
            table.insert( self.aliased_agents, content_id )
        elseif not v.is_template then
            table.insert( self.agent_defs, content_id )
        end
    end
    table.sort( self.aliased_agents )
    table.sort( self.agent_defs )
    table.sort( self.player_defs )
end

function AttackTester:ClearUndo()
    self.dirty = self:IsDirty()
    self.undo_states = {}
    self.undo_head = 0
    self:PushState()
    self.save_head = self.undo_head
end

function AttackTester:PushState( msg )
    assert_warning( self.presentation )

    local meta, meta_filename, anim_name = self.presentation.meta, self.presentation.meta_filename, self.presentation.anim

    -- Table references will be changing, so save selection indices.
    local sel
    if self.presentation then
        for k,v in ipairs(self.selected_values) do
            if sel == nil then
                sel = {}
            end
            table.insert(sel, table.find( self.presentation.anim_track, v ))
        end
    end
        
    if self.save_head and self.save_head > self.undo_head then
        -- Since we're losing a bunch of undo states, need to save the files that they had modified.
        for i, filename in ipairs( self:GetDirtyFiles() ) do
            table.insert_unique( self.dirty_files, filename )
        end
        self.save_head = nil
    end
    while #self.undo_states > self.undo_head do
        table.remove( self.undo_states )
    end
    while #self.undo_states > MAX_UNDO do
        table.remove( self.undo_states, 1 )
        if (self.save_head or 1) > 1 then
            self.save_head = self.save_head - 1
        end
    end

    if msg == nil then
        msg = string.format( "%s.%s", meta_filename, anim_name )
    end

    local state = {
        meta_filename = meta_filename,
        anim_name = anim_name,
        meta = deepcopy(meta),
        selection = sel,
        msg = msg
    }
    table.insert(self.undo_states, state)

    print( "Save:", #self.undo_states, meta_filename, anim_name, state.meta, msg )

    self.undo_head = #self.undo_states

    if self.undo_head > 1 then
        if self.edit_time and os.clock() - self.edit_time < 0.1 then
            self.fast_edits = (self.fast_edits or 0) + 1
        else
            self.fast_edits = 0
        end
        if self.fast_edits > 20 then
            LOGWARN( "Whoa there, %d fast edits detected", self.fast_edits )
            -- Disable state saving as a precauation.
            self.PushState = function( self ) end 
        end

        self.time_to_pending = nil
        self.pending = nil
        self.pending_msg = nil
        self.edit_time = os.clock()
        self.edit_msg = string.format( "%d] EDIT: %s", self.undo_head, msg )
    end
end

function AttackTester:RestoreState( state )
    if state.meta_filename == GetFXTrackFilename() then
        GetFXTrack()[ state.anim_name ] = deepcopy( state.meta )
    else
        local meta_tables = GetAnimMeta( state.meta_filename )
        meta_tables[ state.anim_name ] = deepcopy( state.meta )
    end

    print( "Restore:", state.meta_filename, state.anim_name, state.meta )

    self:RefreshAnimTrack()

    if state.anim_name == self.presentation.anim and state.meta_filename == self.presentation.meta_filename then
        self:SetSelection( nil )
        if state.selection then
            for i, idx in ipairs( state.selection ) do
                local v = self.presentation.anim_track[ idx ]
                self:UnionSelection( v )
            end
        end
    end
end

function AttackTester:Redo()
    if self.undo_head < #self.undo_states then
        self.undo_head = self.undo_head + 1
        local state = self.undo_states[ self.undo_head ]
        self:RestoreState( state )
    
        self.edit_time = os.clock()
        self.edit_msg = string.format( "%d] REDO (%s): %s", self.undo_head, state.anim_name, state.msg )
    end
end

function AttackTester:Undo()
    if self.undo_head > 1 then
        local state = self.undo_states[ self.undo_head ]
        self.edit_time = os.clock()
        self.edit_msg = string.format( "%d] UNDO (%s): %s", self.undo_head, state.anim_name, state.msg )

        self.undo_head = self.undo_head - 1
        self:RestoreState( self.undo_states[ self.undo_head ] )
    end
end

function AttackTester:GetDirtyFiles()
    local dirty_files = shallowcopy( self.dirty_files )
    if self.save_head then
        if self.save_head > self.undo_head then
            for i = self.save_head - 1, self.undo_head, -1 do
                table.insert_unique( dirty_files, self.undo_states[ i ].meta_filename )
            end

        elseif self.save_head < self.undo_head then
            for i = self.save_head + 1, self.undo_head do
                table.insert_unique( dirty_files, self.undo_states[ i ].meta_filename )
            end
        end
    end
    return dirty_files
end

function AttackTester:IsDirty()
    return self.save_head ~= self.undo_head or self.dirty
end

function AttackTester:PushPending( id, msg )
    if self.pending ~= nil and self.pending ~= id then
        self:PushState()
    end
    self.time_to_pending = 0.5
    self.pending_msg = msg
    self.pending = id
end

function AttackTester:HandleControlDown(control, device)
    if self.top then
        if control:Has( Controls.Digital.EDITOR_SAVE ) then
            self:Save()
            return true

        elseif control:Has( Controls.Digital.EDITOR_UNDO ) then
            self:Undo()
            return true

        elseif control:Has( Controls.Digital.EDITOR_REDO ) then
            self:Redo()
            return true


        elseif control:Has( Controls.Digital.MENU_DELETE ) then
            self:DeleteSelection()
            return true
        end
    end
    
    return AttackTester._base.HandleControlDown( self, control, device )
end

function AttackTester:DebugRenderFX( entity, k, v, match_tags, clr )
    local show = true
    if v.vis_status == VIS_STATUS.HIDE_IN_EDITOR or v.vis_status == VIS_STATUS.HIDE_IN_GAME then
        show = false
    elseif v.vis_status == nil or v.vis_status == VIS_STATUS.DEFAULT then
        if not FXTrack.MatchFxTrackTags( v, match_tags ) then
            show = false
        end
    end

    if show then
        local node = entity.cmp.AnimController.animnode
        local lengthx = (v.scalex or 1)*.2
        local lengthy = (v.scaley or 1)*.2
        local tx, ty, tz = v.x or 0, v.y or 0, v.z or 0
        local rot = v.rot or 0
        local x,y,z = node:LocalToWorld( tx, ty, tz )
        local rx,ry,rz
        if v.flip then
            rx,ry,rz = node:LocalToWorld(tx + lengthx*math.cos( rot+ 180*DEGREES ),ty + lengthx*math.sin( rot + 180*DEGREES ), tz)
        else
            rx,ry,rz = node:LocalToWorld(tx + lengthx*math.cos( rot ),ty + lengthx*math.sin( rot ), tz)
        end
        local ux,uy,uz = node:LocalToWorld(tx + lengthy*math.cos( rot  + 90*DEGREES),ty + lengthy*math.sin( rot+90*DEGREES), tz)
        
        if v.ground then
            y = .01            
            ry = .01
            uy = .01
        end

        self.fightsim:GetScene():DrawDebugLine(x, y, z, rx, ry, rz, 1, 0, 0)
        self.fightsim:GetScene():DrawDebugLine(x, y, z, ux, uy, uz, 0, 1, 0)
        self.fightsim:GetScene():DrawDebugText(tostring(k), x, y, z, 0.5, HexColour(clr))
    end
end

function AttackTester:ShouldDebugRender()
    return not self.hide_fx_gnomons
end

function AttackTester:DoRender(dt)
    if not self.hide_fx_gnomons then
        if self.presentation and self.current_entity then
            for i, t in ipairs( self.presentation.anim_track ) do
                if t.fx then
                    self:DebugRenderFX( self.current_entity, i, t, self.presentation.match_tags, t.colour_tag or 0xFFFFFFFF )
                end
            end
        end
    end

    AttackTester._base.DoRender(self, dt)
    self:RenderEditorPanel( TheGame:GetDebug():GetDebugUI() )
end


function AttackTester:OnUpdate( dt )
    AttackTester._base.OnUpdate(self, dt)

    if self.time_to_pending then
        self.time_to_pending = self.time_to_pending - dt
        if self.time_to_pending <= 0 then
            self:PushState( self.pending_msg )
        end
    end

    if self.looping and self.current_entity then
        if not self.pres_coro then
            self.looping_dt = (self.looping_dt or 0) - dt
            if self.looping_dt <= 0 then
                self:Play( self.looping )
            end
        end
    end

end

function AttackTester:GetPrimaryTargets()
    local targets = {}
    if self.current_entity == self.attack_entity then
        for i, entity in ipairs( self.target_entities ) do
            table.insert( targets, entity.cmp.AnimFighter.fighter )
        end
    else
        targets[1] = self.attack_entity.cmp.AnimFighter.fighter
    end
    return targets
end

function AttackTester:SwitchCurrentEntity()
    if self.current_entity == self.attack_entity then
        self.current_entity = self.target_entities[1]
    else
        self.current_entity = self.attack_entity
    end
    self:RefreshAttacks()
end

function AttackTester:RefreshAttacks()
    --print("AttackTester:RefreshAttacks()")
    local old_anim = self.all_anims and self.all_anims[self.anim_idx]
    self.all_anims = {}
    self.anim_idx = nil

    for anim_name, anim_info in pairs( self.current_entity.cmp.AnimController.anims) do
        table.insert(self.all_anims, 
            {
                anim = anim_name,
                txt = loc.format("ANIM:{1}", anim_name),
            })
    end

    table.sort(self.all_anims, function(a,b) return a.txt < b.txt end)

    self.anim_txt = {}
    for k, v in ipairs(self.all_anims) do
        table.insert(self.anim_txt, v.txt)
    end
   
    ----------------------------
    -- Update valid attacks list.

    local selected_attack = self.valid_attacks and self.valid_attacks[ self.attack_idx ]
    self.valid_attacks = {}

    local fighter = self.current_entity.cmp.AnimFighter.fighter
    local cards = fighter.agent.fight_data.cards
    if cards then
        table.arrayadd( self.valid_attacks, cards )
    end

    if #self.valid_attacks == 0 and fighter.agent:GetFactionID()== PLAYER_FACTION then
        for i, carddef in ipairs(Content.GetAllBattleCards()) do
            if not CheckAnyBits( carddef.flags, battle_defs.CARD_FLAGS.UPGRADED | battle_defs.CARD_FLAGS.SPECIAL ) then
                local ok = AgentUtil.CanUseCard( fighter.agent, carddef )
                if ok then
                    table.insert_unique(self.valid_attacks, carddef.id)
                end
            end
        end
    end
    table.sort(self.valid_attacks)

    local attack_idx = table.arrayfind( self.valid_attacks, selected_attack )
    self:SetAttackIDX( attack_idx or 1 )
end

function AttackTester:PresFrameEdit(ui, pres, v )

    ui:Text( string.format( "Time: %.2f seconds", ((v.frame or 1) - 1) * FRAMES ))

    local anim_info = pres.animdat
    local max_frame
    if anim_info then
        local _
        _, max_frame = anim_info.anim:GetAnimDuration( anim_info.idx )
    else
        -- No limit to max_frame, we are not tied to a specific anim.
        max_frame = (v.frame or 1) + 10
    end
    local new_frame = ui:SliderInt("Frame", v.frame or 1, 0, max_frame-1 )
    if new_frame ~= (v.frame or 1) then
        v.frame = new_frame
    end
    if self.active_frame_slider == v then
        if not ui:IsItemActive() then
            self:RefreshAnimTrack()
            self.active_frame_slider = nil
            self:PushState( string.format( "frame=%d", new_frame ))
        end
    elseif ui:IsItemActive() then
        self.active_frame_slider = v
    end

    if v.command == "merge" then
        if ui:Button( v.filename or "<NONE>" ) then
            -- Source this anim file.
            ui:OpenPopup( "MERGE_SOURCE_POPUP" )
        end

        if ui:BeginPopup( "MERGE_SOURCE_POPUP" ) then
            for i, filename, meta in sorted_pairs( GetAnimMetaTable() ) do
                if ui:MenuItem( filename, nil, nil, filename ~= v.filename and filename ~= pres.meta_filename ) then
                    v.filename = filename
                    self:PushState( "set merge source:"..filename )
                    break
                end
            end
            ui:EndPopup()
        end

        if ui:Checkbox( "No FX", v.no_fx == true ) then
            v.no_fx = not v.no_fx
            self:PushState( "merge no_fx=%s", tostring(v.no_fx))
        end

        if ui:Checkbox( "No Sound", v.no_sound == true ) then
            v.no_sound = not v.no_sound
            self:PushState( "merge no_sound=%s", tostring(v.no_sound))
        end

        local anim_name = ui:InputText( "Anim", v.anim or "", ui.TextFlags_EnterReturnsTrue )
        if anim_name and anim_name ~= v.anim then
            v.anim = anim_name
            self:PushState( string.format("merge anim:%s", anim_name ))
        end

        ui:Separator()

        -- Merge translations.
        ui:PushItemWidth(100)
        local x = ui:DragFloat("X", v.offsetx or 0, -5, 5, "%0.2f", .01)
        ui:SameLine(nil,10)
        local y = ui:DragFloat("Y", v.offsety or 0, -5, 5, "%0.2f", .01)
        ui:SameLine(nil,10)
        local z = ui:DragFloat("Z", v.offsetz or 0, -5, 5, "%0.2f", .01)
        
        if x and x ~= (v.offsetx or 0) then
            v.offsetx = x
            self:PushPending( "MergeOffsetX", string.format( "merge offsetx: %.2f", x ))
        end
        if y and y ~= (v.offsety or 0) then
            v.offsety = y
            self:PushPending( "MergeOffsetY", string.format( "merge offsety: %.2f", y ))
        end
        if z and z ~= (v.offsetz or 0) then
            v.offsetz = z
            self:PushPending( "MergeOffsetZ", string.format( "merge offsetz: %.2f", z ))
        end

        local rot = ui:DragFloat("Rotation", (v.rot or 0)/DEGREES, -360, 360, "%0.2f", 1)
        if rot then
            rot = rot*DEGREES
            if rot ~= v.rot then
                v.rot = rot
                self:PushPending( "MergeRotation", string.format( "merge rotation: %.2f", rot ))
            end
        end

        local scale = ui:SliderFloat("Scale", v.scale or 1.0, 0.1, 5.0, "%0.2f", .01)
        if scale and scale ~= v.scale then
            v.scale = scale
            self:PushPending( "MergeScale", string.format( "merge scale: %.2f", scale ))
        end
    end
end

function AttackTester:PresSoundEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    ui:PushStyleColor( ui.Style_Button, 0x770033FF )
    local txt = v.sound
    if not txt or #txt == 0 then
        txt = "<no sound>"
    end
    if ui:Button( txt ) then
        ui:OpenPopup( "SOUND" )
        self:GenerateSoundEvents( self.filter_str, v.looping )
    end
    ui:PopStyleColor()

    if ui:BeginPopup( "SOUND" ) then
        local filter_str = ui:InputText( "Filter", self.filter_str or "" )
        if filter_str and filter_str ~= self.filter_str and #filter_str > 0 then
            self.filter_str = filter_str
            self:GenerateSoundEvents( filter_str, v.looping )
        end

        if ui:Button( "Clear" ) then
            v.sound = ""
            v.meta_sound = nil
            self:PushState( string.format( "sound: %s", v.sound ))
            ui:CloseCurrentPopup()
        end

        ui:Separator()
        ui:TextColored( 0x00CCCCFF, string.format( "%d results for '%s'", #self.sound_events, self.filter_str ))    
        ui:BeginChild( "Results", 400, 300 )
        for i, sound_ev in ipairs( self.sound_events or table.empty ) do
            if ui:Selectable( sound_ev ) then
                v.sound = sound_ev
                AUDIO:PlayEvent( SoundEvents[ sound_ev ] )
                v.meta_sound = nil
                self:PushState( string.format( "sound: %s", v.sound ))
                ui:CloseCurrentPopup()
            end
        end
        ui:EndChild()

        ui:EndPopup()
    end


    v.must_have_tags = v.must_have_tags or {}
    v.cant_have_tags = v.cant_have_tags or {}

    -- if ui:Checkbox( "Looping", v.looping ) then
    --     v.looping = not v.looping
    --     self:PushState( string.format( "looping sound: %s", v.looping ))
    --     self:GenerateSoundEvents( self.filter_str, v.looping )
    -- end

    self:EditTagRequirements( ui, v )
end

function AttackTester:PreviewVO(vo)
    if self.attack_agent and self.attack_agent.voice_actor then
        local event_name = self.attack_agent:GetCombatVOEvent(vo)
        if event_name then
            AUDIO:PlayEvent( event_name, "sound/vo.bank" )
        end
        AUDIO:PlayEvent( event_name, "sound/vo.bank" )
    end
end

function AttackTester:PreviewShake(shake_params)
    if self.ScreenShake then
        self:ScreenShake( shake_params )
    end
end


function AttackTester:PresShakeEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    v.shake = v.shake or {0.25, 55, 0.3}
    --Magnitude, Speed, Duration
    local m, s, d = table.unpack(v.shake)
    v.shake[1] = ui:SliderFloat( "Magnitude", v.shake[1] or 0, 0, 5 )    
    v.shake[2] = ui:SliderFloat( "Speed", v.shake[2] or 0, 0, 100 )    
    v.shake[3] = ui:SliderFloat( "Duration", v.shake[3] or 0, 0, 5 )    

    if ui:Button( "Preview Shake" ) then
        self:PreviewShake(v.shake)
    end

    if m ~= v.shake[1] or s ~= v.shake[2] or d ~= v.shake[3] then
        self:PushPending( "MergeShake", string.format( "merge shake: %f %f %f", v.shake[1], v.shake[2], v.shake[3] ))
    end

    v.must_have_tags = v.must_have_tags or {}
    v.cant_have_tags = v.cant_have_tags or {}

    self:EditTagRequirements( ui, v )
end

function AttackTester:PresVOEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    local idx = table.arrayfind( self.vo_events, v.vo or "" )
    local new_idx = ui:Combo( "VO Event", idx or 1, self.vo_events )
    if ui:Button( "Preview VO" ) then
        self:PreviewVO(v.vo)
    end


    if new_idx ~= idx and new_idx then
        v.vo = self.vo_events[ new_idx ]
        self:PreviewVO(v.vo)
        self:PushState( string.format( "vo: %s", v.vo ))
    end

    v.must_have_tags = v.must_have_tags or {}
    v.cant_have_tags = v.cant_have_tags or {}

    self:EditTagRequirements( ui, v )
end

function AttackTester:PresHitFrameEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    local new_type = ui:InputText( "Hit Type", v.type ) or v.type
    if new_type ~= v.type then
        v.type = new_type
        self:PushState( string.format( "hit_type: %s", new_type ))
    end

    local MAX_KNOCKBACK = 5.0
    local knockback = ui:SliderFloat( "Knockback", v.dist or 0, 0, MAX_KNOCKBACK )
    if knockback == 0 then
        knockback = nil
    end
    if knockback ~= v.dist then
        v.dist = knockback
        self:PushState( string.format( "knockback: %.2f", v.dist or 0.0 ))
    end

    self:EditTagRequirements(ui, v)
end

function AttackTester:PresFlipFrameEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    if ui:Checkbox( "Is Flipped", v.is_flipped == true ) then
        v.is_flipped = not v.is_flipped
        self:PushState( string.format( "flipped: %s", tostring(v.is_flipped) ))
    end
end

function AttackTester:PresActiveTrackEdit(ui, pres, k, v )
    self:PresFrameEdit( ui, pres, v )

    if v.start_bloom then
        local new_bloom = ui:SliderFloat("Start Bloom", v.start_bloom or 0, 0, 2)
        if new_bloom ~= (v.start_bloom or 0) then
            v.start_bloom = new_bloom
            self:PushState( string.format( "start_bloom: %.2f", new_bloom ))
        end

        local new_bloom = ui:SliderFloat("End Bloom", v.end_bloom or 0, 0, 2)
        if new_bloom ~= (v.end_bloom or 0) then
            v.end_bloom = new_bloom
            self:PushState( string.format( "end_bloom: %.2f", new_bloom ))
        end
    end

    local new_duration = ui:SliderFloat("Duration", v.duration or 0, 0, 10)
    if new_duration ~= (v.duration or 0) then
        v.duration = new_duration
        self:PushState( string.format( "duration: %.2f", new_duration ))
    end
end

local function HasTransform( t )
    return t.x ~= nil and t.y ~= nil
end

local function HasRotation( t )
    return t.rot ~= nil
end

local function HasColour( t )
    return t.fx ~= nil -- Yeah... t.colour is often nil, if it is default
end

local function ColorEdit4( ui, label, clr )
    if ui:Button( "...##"..tostring(label) ) then
        ui:OpenPopup( "CLR_CONTEXT"..tostring(label) )
    end

    ui:SameLine( nil, 10 )
    ui:PushItemWidth(300)
    local old_colour = clr or 0xFFFFFFFF
    local new_colour = ui:ColorEdit4( label, old_colour )
    if new_colour == 0xFFFFFFFF then
        new_colour = nil -- Default white is nil.
    end
    ui:PopItemWidth()

    local clear
    if ui:BeginPopup( "CLR_CONTEXT"..tostring(label) ) then
        if ui:MenuItem( "Clear" ) then
            new_colour = nil
        end

        ui:Separator()

        local dropper_clr = GetClipboardColour()
        if dropper_clr then
            if ui:MenuItem( "Paste Colour" ) then
                new_colour = dropper_clr
            end
            ui:SameLine( nil, 5 )
            ui:ColorButton( dropper_clr, true )
        end

        if ui:MenuItem( "Copy Colour" ) then
            SetClipboardColour( new_colour or 0xFFFFFFFF )
        end

        ui:EndPopup()
    end


    if clear then
        return nil
    else
        return new_colour
    end
end

function AttackTester:PresMultiEdit(ui, pres, values)
    if self.multi_edit == nil then
        self.multi_edit = {}
    end
    local edit_count = 0

    ui:Separator()

    if table.findif( values, HasTransform ) then
        ui:PushItemWidth(100)
        self.multi_edit.dx = ui:DragFloat("DX", self.multi_edit.dx or 0, -10, 10, "%0.2f", .01)
        ui:SameLine(nil,10)
        self.multi_edit.dy = ui:DragFloat("DY", self.multi_edit.dy or 0, -10, 10, "%0.2f", .01)
        ui:SameLine(nil,10)
        self.multi_edit.dz = ui:DragFloat("DZ", self.multi_edit.dz or 0, -10, 10, "%0.2f", .01)

        if ui:Button( "Group Shift" ) then
            for i, value in ipairs( values ) do
                local edited = false
                if value.x then
                    value.x = value.x + self.multi_edit.dx
                    edited = true
                end
                if value.y then
                    value.y = value.y + self.multi_edit.dy
                    edited = true
                end
                if value.z then
                    value.z = value.z + self.multi_edit.dz
                    edited = true
                end
                if edited then
                    edit_count = edit_count + 1
                end
            end
        end
    end

    if table.findif( values, HasRotation ) then
        ui:Spacing()
    
        local rot = ui:DragFloat("Rotation", (self.multi_edit.rot or 0)/DEGREES, -360, 360, "%0.2f", 1)
        self.multi_edit.rot = rot*DEGREES

        if ui:Button( "Group Rotate" ) then
            for i, value in ipairs( values ) do
                if value.rot then
                    value.rot = normalizeRads( value.rot + self.multi_edit.rot )
                    edit_count = edit_count + 1
                end
            end
        end
    end

    if table.findif( values, HasColour ) then
        ui:Spacing()

        self.multi_edit.colour = ColorEdit4( ui, "Colour", self.multi_edit.colour )

        self.multi_edit.end_colour = ColorEdit4( ui, "Colour End", self.multi_edit.end_colour )

        if ui:Button( "Group Colour" ) then
            for i, value in ipairs( values ) do
                if value.fx ~= nil then
                    local edited = false
                    if (value.colour or 0xFFFFFFFF) ~= self.multi_edit.colour then
                        value.colour = self.multi_edit.colour
                        edited = true
                    end
                    if (value.end_colour or 0xFFFFFFFF) ~= self.multi_edit.end_colour then
                        value.end_colour = self.multi_edit.end_colour
                        edited = true
                    end
                    if edited then
                        edit_count = edit_count + 1
                    end
                end
            end
        end

    end
    
    if edit_count > 0 then
        self.multi_edit.msg = string.format( "Changed %d values.", edit_count )
        self.multi_edit.msg_time = os.clock()
        self:PushState( self.multi_edit_msg )
    end

    if self.multi_edit.msg_time then
        local dt = os.clock() - self.multi_edit.msg_time
        local a = 1.0 - clamp( dt / 2, 0, 1 )
        ui:Spacing()
        ui:TextColored( MakeHexColour( 0, 1, 0, a ), tostring(self.multi_edit.msg))

        if a <= 0 then
            self.multi_edit.msg_time = nil
            self.multi_edit.msg = nil
        end
    end
end

function AttackTester:EditTags(ui, name, tag_list, possible_tags)
    local changed = false

    if ui:TreeNode( name ) then               
        ui:Columns(5)
        for k,tag in ipairs(possible_tags) do
            local found = tag_list and table.contains( tag_list, tag )          
            if ui:Checkbox( tag, found ) then
                if found then
                    table.arrayremove( tag_list, tag )
                    self:PushState( string.format( "remove hit_tag '%s'", tag ))
                    self:GenerateHitTags()
                else
                    if tag_list == nil then
                        tag_list = {}
                    else
                        self:PushState( string.format( "add hit_tag '%s'", tag ))
                    end
                    table.insert(tag_list, tag)
                end
            end
            ui:NextColumn()
        end
        ui:Columns(1)

        if tag_list and #tag_list > 0 and ui:Button("Clear") then
            table.clear( tag_list )
            self:GenerateHitTags()
            self:PushState( "clear hit_tags")
        end

        local new_tag = ui:InputText( "Custom Tag", "", ui.TextFlags_EnterReturnsTrue )
        if new_tag and #new_tag > 0 then
            if tag_list == nil then
                tag_list = {}
            end
            table.insert( tag_list, new_tag )
            self:GenerateHitTags()
            self:PushState( string.format( "custom hit_tag '%s'", new_tag ))
        end

        ui:TreePop()
    end

    return tag_list
end

function AttackTester:EditTagRequirements(ui, v)
    ui:Spacing()

    local txt = {}
    for i, tag in ipairs( v.must_have_tags or table.empty ) do
        table.insert( txt, tag )
    end
    if #txt > 0 then
        ui:Text( "Required Tags: ")
        ui:SameLine( nil, 10 )
        ui:TextColored( 0x00CC00FF, table.concat( txt, ", " ))
    end

    local txt = {}
    for i, tag in ipairs( v.cant_have_tags or table.empty ) do
        table.insert( txt, tag )
    end
    if #txt > 0 then
        ui:Text( "Invalid Tags: ")
        ui:SameLine( nil, 10 )
        ui:TextColored( 0xCC0000FF, table.concat( txt, ", " ))
    end

    if ui:Button( "Edit Tags" ) then
        ui:OpenPopup( "EDIT_TAGS" )
    end

    if ui:BeginPopup( "EDIT_TAGS" ) then
        for i, tag in ipairs( self.current_attack_tags ) do
            if table.contains( v.must_have_tags or table.empty, tag ) then
                ui:PushStyleColor( ui.Style_Text, 0x00CC00FF )
                if ui:Selectable( tag, nil,  ui.DontClosePopups ) then
                    table.arrayremove( v.must_have_tags, tag )
                    if #v.must_have_tags == 0 then
                        v.must_have_tags = nil
                    end
                    if v.cant_have_tags == nil then
                        v.cant_have_tags = {}
                    end
                    table.insert( v.cant_have_tags, tag )
                    self:PushState( "tag requirements" )
                end

            elseif table.contains( v.cant_have_tags or table.empty, tag ) then
                ui:PushStyleColor( ui.Style_Text, 0xCC0000FF )
                if ui:Selectable( tag, nil,  ui.DontClosePopups ) then
                    table.arrayremove( v.cant_have_tags, tag )
                    if #v.cant_have_tags == 0 then
                        v.cant_have_tags = nil
                    end
                    self:PushState( "tag requirements" )
                end

            else
                ui:PushStyleColor( ui.Style_Text, 0xAAAAAAFF )
                if ui:Selectable( tag, nil,  ui.DontClosePopups ) then
                    if v.must_have_tags == nil then
                        v.must_have_tags = {}
                    end
                    table.insert( v.must_have_tags, tag )
                    self:PushState( "tag requirements" )
                end
            end
            ui:PopStyleColor()
        end
        ui:EndPopup()
    end

    -- self:EditTags(ui, "ALL tags:", v.must_have_tags, self.all_attack_tags)
    -- self:EditTags(ui, "ANY tags:", v.one_of_tags, self.all_attack_tags)
    -- self:EditTags(ui, "NO tags:", v.cant_have_tags, self.all_attack_tags)
end

function AttackTester:PresFXEdit( ui, pres, k, v )
   
    self:PresFrameEdit( ui, pres, v )

    ui:PushID( "FXNAME" )
    if ui:Button( tostring( v.fx ), ui:GetItemWidth() ) then
        ui:OpenPopup( "FX" )
    end
    ui:SameLine( nil, 5 )
    ui:Text( "FX" )

    if ui:BeginPopup( "FX" ) then
        local filter_str = ui:InputText( "Filter", self.filter_str or "" )
        if filter_str and filter_str ~= self.filter_str and #filter_str > 0 then
            self.filter_str = filter_str
            self.anim_files = self:GenerateAnimList( filter_str )
        end

        ui:Separator()
        ui:TextColored( 0x00CCCCFF, string.format( "%d results for '%s'", #self.anim_files, self.filter_str ))    
        ui:BeginChild( "Results", 250, 300 )
        for i, result in ipairs( self.anim_files or table.empty ) do
            if ui:Selectable( result, v.fx == result ) then
                v.fx = result
                self:PushState( string.format( "fx: %s", result ))
            end
        end
        ui:EndChild()

        ui:EndPopup()
    end
    ui:PopID()

    local afile = #v.fx > 0 and GetAnim(v.fx)
    if afile then
        local states = {}
        for k = 1, afile:GetNumAnims() do
            local name = afile:GetAnimName(k-1)
            table.insert(states, name)
        end
        if #states > 0 then
            local state_idx = table.arrayfind(states, v.state) or 0
            local new_state_idx = ui:Combo("State", state_idx, states)                    
            if new_state_idx and new_state_idx ~= state_idx then
                v.state = states[new_state_idx]
                self:PushState( string.format( "state: %s", v.state ))
            end
        end
    end

    if ui:Button("Build File: " .. (v.buildfile or "")) then
        self:OnClickBuildFile(v) 
    end
    ui:SameLine(nil, 10)
    if ui:Button("Clear##1") then
        if v.buildfile then
            self:PushState("clear build file")
            v.buildfile = nil
        end
    end

    v.must_have_tags = v.must_have_tags or {}
    v.cant_have_tags = v.cant_have_tags or {}

    self:EditTagRequirements(ui, v)
    

    --v.hit_tag = self.all_attack_tags[ui:Combo("Hit Tag", table.arrayfind(self.all_attack_tags, v.hit_tag or "") or 1, self.all_attack_tags)]

    ui:PushItemWidth(100)
    local x = ui:DragFloat("X", v.x or 0, -10, 10, "%0.2f", .01)
    ui:SameLine(nil,10)
    local y = ui:DragFloat("Y", v.y or 0, -10, 10, "%0.2f", .01)
    ui:SameLine(nil,10)
    local z = ui:DragFloat("Z", v.z or 0, -10, 10, "%0.2f", .01)
    
    if x and x ~= (v.x or 0) then
        v.x = x
        self:PushPending( "X", string.format( "x: %.2f", x ))
    end
    if y and y ~= (v.y or 0) then
        v.y = y
        self:PushPending( "Y", string.format( "y: %.2f", y ))
    end
    if z and z ~= (v.z or 0) then
        v.z = z
        self:PushPending( "Z", string.format( "z: %.2f", z ))
    end

    local scalex = ui:SliderFloat("X Scale", v.scalex or v.scale or DEFAULT_SCALE, -20, 20.0, "%0.2f", .01)
    if ui:BeginPopupContextItem( "XSCALE" ) then
        if ui:MenuItem( "Reset Scale" ) then
            scalex = 1.0
        end
        ui:EndPopup()
    end
    ui:SameLine(nil,10)
    local scaley = ui:SliderFloat("Y Scale", v.scaley or v.scale or DEFAULT_SCALE, -20, 20.0, "%0.2f", .01)
    if ui:BeginPopupContextItem( "YSCALE" ) then
        if ui:MenuItem( "Reset Scale" ) then
            scaley = 1.0
        end
        ui:EndPopup()
    end
    
    ui:SameLine( nil, 15 )
    if ui:Checkbox( "Model Scale", v.model_scale == true ) then
        v.model_scale = not v.model_scale
    end

    if scalex and (v.scalex or v.scale or DEFAULT_SCALE) ~= scalex then
        v.scalex = scalex
        self:PushPending( "scalex", string.format( "scalex: %.2f", scalex ))
    end
    if scaley and (v.scaley or v.scale or DEFAULT_SCALE) ~= scaley then
        v.scaley = scaley
        self:PushPending( "scaley", string.format( "scaley: %.2f", scaley ))
    end

    if ui:Checkbox("Ground", v.ground ) then
        v.ground = not v.ground
        self:PushState( string.format( "ground: %s", v.ground ))
    end

    ui:SameLine( nil, 50 )
    if ui:Checkbox("Flip", v.flip ) then
        v.flip = not v.flip
        self:PushState( string.format( "flip: %s", v.flip ))
    end
    
    ui:SameLine( nil, 10 )
    local rot = ui:DragFloat("Rotation", v.rot and v.rot/DEGREES or 0, -360, 360, "%0.2f", 1)
    if rot and (v.rot or 0) ~= rot*DEGREES then
        v.rot = rot*DEGREES
        self:PushPending( "rot", string.format( "rot: %.2f", v.rot ))
    end

    local blend_idx = table.arrayfind(blendmodes, v.blendmode or BLEND_MODE.BLENDED)
    local new_blend_idx = ui:Combo("Blend Mode", blend_idx, blendmodes_txt)
    if blend_idx ~= new_blend_idx then
        v.blendmode = blendmodes[new_blend_idx]
        self:PushState( string.format( "blendmode: %s", v.blendmode ))
    end

    local new_clr = ColorEdit4( ui, "Colour", v.colour)
    if new_clr ~= v.colour then
        v.colour = new_clr
        self:PushPending( "colour", string.format( "colour: %08X", v.colour or 0xFFFFFFFF ))
    end

    local end_new_clr = ColorEdit4( ui, "Colour End", v.end_colour or v.colour)
    if end_new_clr ~= (v.end_colour or v.colour) then
        v.end_colour = end_new_clr
        self:PushPending( "end_colour", string.format( "end_colour: %08X", v.end_colour or 0xFFFFFFFF ))
    end

    if ui:Checkbox("Erode", v.erode ) then
        v.erode = not v.erode
        self:PushState( string.format( "erode: %s", v.erode ))
    end
    
    ui:SameLine(nil, 10)
    local new_bloom = ui:SliderFloat("Bloom", v.bloom or 0, 0, 2)
    if new_bloom ~= (v.bloom or 0) then
        v.bloom = new_bloom
        self:PushPending( "bloom", string.format( "bloom: %.2f", v.bloom ))
    end

    if ui:Checkbox("Follow", v.follow) then
        v.follow = not v.follow
        self:PushState( string.format( "follow: %s", v.follow ))
    end

    ui:SameLine( nil, 10 )
    local symbol = ui:InputText( "Attach to Symbol", v.symbol or "", ui.TextFlags_EnterReturnsTrue )
    if symbol and symbol ~= v.symbol then
        v.symbol = symbol
        self:PushPending( "symbol", string.format( "symbol: %s", symbol ))
    end

    ui:SameLine( nil, 10 )
    if ui:Checkbox( "Model Rotation", v.model_rotation ) then
        v.model_rotation = not v.model_rotation
        self:PushPending( "model_rotation", string.format( "model_rotation: %s", v.model_rotation ))
    end

    if ui:Checkbox("Linked to Anim", v.anim_linked) then
        v.anim_linked = not v.anim_linked
        self:PushState( string.format( "anim_linked: %s", v.anim_linked ))
    end

    if ui:Checkbox("Loop", v.loop) then
        v.loop = not v.loop
        self:PushState( string.format( "loop: %s", v.loop ))
    end

    ui:PopItemWidth()
end

function AttackTester:OnClickBuildFile(fx_elem)

    local base = path.normpath( path.join(BASE_PATH, "anim"))
    local file = engine.inst:OpenDialog(nil, base)--"zip", path)
    if file then
        fx_elem.buildfile = path.relpath (file,DATA_DIR)
        self:PushState("Set fx build file")
    end

end



local function GenerateAnimMeta( filename, anim )
    assert( filename )
    if filename == GetFXTrackFilename() then
        local meta = GetFXTrack( anim )
        if meta == nil then
            meta = {}
            GetFXTrack()[ anim ] = meta
        end
        return meta
    else
        local ANIM_META = GetAnimMetaTable()
        if ANIM_META[ filename ] == nil then
            ANIM_META[ filename ] = {}
        end
        local meta = ANIM_META[ filename ][ anim ]
        if meta == nil then
            meta = {}
            ANIM_META[ filename ][ anim ] = meta
        end
        return meta
    end
end

local function GenerateTracks( pres, key )
    local meta = pres.meta
    if meta == nil then
        meta = GenerateAnimMeta( pres.meta_filename, pres.anim )
    end
    if meta[ key ] == nil then
        meta[ key ] = {}
    end
    return meta[ key ]
end

local function GenerateFXTracks( pres )
    return GenerateTracks( pres, "fx_tracks" )
end

local function GenerateShakeTracks( pres )
    return GenerateTracks( pres, "shake_frames" )
end

local function GenerateHitTracks( pres )
    return GenerateTracks( pres, "hit_frames" )
end

local function GenerateSoundFrames( pres )
    return GenerateTracks( pres, "sound_frames" )
end

function AttackTester:CollectFolderNames()
    if self.folder_names then
        table.clear( self.folder_names )
    else
        self.folder_names = {}
    end

    for i, v in ipairs( self.presentation.anim_track ) do
        if v.folder_name then
            table.insert_unique( self.folder_names, v.folder_name )
        end
    end

    return self.folder_names
end

function AttackTester:CreateNewFolderName()
    local name = "New Folder"
    local names = self:CollectFolderNames()
    local found
    local count = 1
    repeat
        found = false
        for i, v in ipairs( names ) do
            if v == name then
                found = true
                break
            end
        end
        if found then
            name = string.format( "New Folder %d", count )
            count = count + 1
        end
    until not found or count > 99

    return name
end

function AttackTester:BeginEditGenericFX()

    self.all_fx = {}
    for i, name, v in sorted_pairs( GetFXTrack() or table.empty ) do
        table.insert( self.all_fx, name )
    end

    self.fx_idx = 1

    self:RefreshAnimTrack()
end

function AttackTester:EndEditGenericFX()
    self.fx_idx = nil
    self:RefreshAnimTrack()
end

function AttackTester:PresGenericFXEdit( ui )
    local fx = GetFXTrack()

    local new_idx = ui:Combo( "Track", self.fx_idx, self.all_fx )
    if new_idx and new_idx ~= self.fx_idx then
        self.fx_idx = new_idx
        self:SetSelection( nil )
        self:RefreshAnimTrack()
    end

    self.new_fx_name = ui:InputText( "", self.new_fx_name or "" ) or self.new_fx_name
    ui:SameLine( nil, 10 )
    local disabled = self.new_fx_name == nil or #self.new_fx_name <= 1 or table.contains( self.all_fx, self.new_fx_name )
    if ui:Button( "New Track", nil, nil, disabled ) then
        table.binsert( self.all_fx, self.new_fx_name )
        self.fx_idx = table.find( self.all_fx, self.new_fx_name )
        self.new_fx_name = ""
        self:SetSelection( nil )
        self:RefreshAnimTrack()
    end

    if ui:Button( "Play" ) then
        self.current_entity.cmp.AnimFighter:PlayOneShotFX( self.presentation.anim, self.editor_hit_tags )
    end
    ui:SameLine( nil, 10 )
    ui:PushStyleColor( ui.Style_Button, 0xCC0000FF )
    if ui:Button( "Delete" ) then
        -- TODO: impl
    end
    ui:PopStyleColor()

    local fx_tracks = self.current_entity.cmp.AnimFighter.fx_tracks
    if #fx_tracks > 0 then
        ui:Text( string.format( "%d tracks playing", #fx_tracks ))
        ui:SameLine( nil, 10 )
        if ui:Button( "Clear" ) then
            for i = #fx_tracks, 1, -1 do
                fx_tracks[i]:Kill()
            end
        end
    end

    self:PresEdit( ui, self.presentation, "GENERIC FX" )
end

function AttackTester:PresEditContextMenu( ui, pres )

    local current_folder
    if #self.selected_values > 0 then
        for i, value in ipairs( self.selected_values ) do
            ui:TextColored( 0x00CCCCFF, tostring( value.fx or value.sound or value.meta_sound or value.type ))
        end
        ui:Separator()

        local colour_tag = self.selected_values[1].colour_tag or 0x777777FF          
        local new_clr = ui:ColorEdit3( "Colour Tag", colour_tag )

        ui:SameLine( nil, 15 )
        if new_clr and (ui:Button( "Set" ) or new_clr ~= colour_tag) then
            for i, value in ipairs( self.selected_values ) do
                value.colour_tag = new_clr
            end
            self:PushPending( "colour tag" )
        end

        current_folder = self.selected_values[1].folder_name
        for i = 2, #self.selected_values do
            if self.selected_values[i].folder_name ~= current_folder then
                current_folder = nil -- Multi-valued
            end
        end

        if ui:BeginMenu( "Assign Folder..." ) then
            local folder_names = self:CollectFolderNames()
            for i, name in ipairs( folder_names ) do
                if ui:MenuItem( name, nil, name == current_folder ) then
                    for i, value in ipairs( self.selected_values ) do
                        value.folder_name = name
                    end
                    self:PushState( string.format( "assign folder: %s", name ))
                end
            end

            if #folder_names > 0 then
                ui:Separator()

                if table.findif( self.selected_values, function(v) return v.folder_name ~= nil end ) then
                    if ui:MenuItem( "None" ) then
                        for i, value in ipairs( self.selected_values ) do
                            value.folder_name = nil
                        end
                        self:PushState( "unassign folder" )
                    end
                end
            end

            local name = self:CreateNewFolderName()
            if ui:MenuItem( tostring(name) ) then
                for i, value in ipairs( self.selected_values ) do
                    value.folder_name = name
                end
                self:PushState( string.format( "new folder: %s", name ))
            end
            ui:EndMenu()
        end

        if current_folder then
            local new_folder = ui:InputText( "Folder", current_folder, ui.TextFlags_EnterReturnsTrue )
            if new_folder and new_folder ~= current_folder then
                for i, value in ipairs( self.selected_values ) do
                    value.folder_name = new_folder
                end
                self:PushState( "edit folder" )
            end
        end

        local status = self.selected_values[1].vis_status or VIS_STATUS.DEFAULT
        ui:Image( assets[ status ], 16, 16 )

        if ui:IsMouseClicked() and ui:IsItemHovered() then
            for i, value in ipairs( self.selected_values ) do
                value.vis_status = GetNextVisStatus( value.vis_status )
            end
            self:PushState( "vis_status" )
        end
        ui:SameLine( nil, 5 )
        if status == VIS_STATUS.DEFAULT then
            ui:Text( "Default Visibility" )
        elseif status == VIS_STATUS.HIDE_IN_EDITOR then
            ui:Text( "Hide in Editor" )
        elseif status == VIS_STATUS.HIDE_IN_GAME then
            ui:Text( "Hide in Game" )
        elseif status == VIS_STATUS.ALWAYS_SHOW then
            ui:Text( "Always Show" )
        end

        ui:Separator()
    end

    if ui:MenuItem( "Add FX" ) then
        local fx_track = GenerateFXTracks( pres )
        table.insert( fx_track, {frame=1, follow = true, fx = "", folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add fx" )
    end
    if ui:MenuItem( "Add Shake" ) then
        local fx_track = GenerateShakeTracks( pres )
        table.insert( fx_track, {frame=1,  shake = { 0.25, 55, 0.3 }, folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add shake" )
    end
    if ui:MenuItem( "Add Sound" ) then
        local sound_frames = GenerateSoundFrames( pres )
        table.insert( sound_frames, {frame=1, sound = "", folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add sound" )
    end
    if ui:MenuItem( "Add VO" ) then
        local sound_frames = GenerateSoundFrames( pres )
        table.insert( sound_frames, {frame=1, vo = "", folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add vo" )
    end
    if ui:MenuItem( "Add Hit" ) then
        local hit_frames = GenerateHitTracks( pres )
        table.insert( hit_frames, {frame=1, type = "hit_mid", folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add hit" )
    end
    if ui:MenuItem( "Add Flip" ) then
        local flip_frames = GenerateTracks( pres, "flip_frames" )
        table.insert( flip_frames, {frame=1, is_flipped= true, folder_name = current_folder})
        self:RefreshAnimTrack()
        self:PushState( "add flip" )
    end
    if ui:MenuItem( "Add Bloom" ) then
        local fx_track = GenerateFXTracks( pres )
        table.insert( fx_track, {frame=1, start_bloom = 0, end_bloom = 0, duration = 0, folder_name = current_folder })
        self:RefreshAnimTrack()
        self:PushState( "add bloom" )
    end
    if ui:BeginMenu( "Add Command..." ) then
        if ui:MenuItem( "Loop" ) then
            local fx_track = GenerateFXTracks( pres )
            table.insert( fx_track, {frame=1, command = "loop", folder_name = current_folder })
            self:RefreshAnimTrack()
            self:PushState( "add loop command" )
        end
        if ui:MenuItem( "Merge" ) then
            local fx_track = GenerateFXTracks( pres )
            table.insert( fx_track, {frame=1, command = "merge", filename = nil, anim = pres.anim, folder_name = current_folder })
            self:RefreshAnimTrack()
            self:PushState( "add merge command" )
        end
        ui:EndMenu()
    end

    ui:Separator()

    if #self.selected_values > 0 and ui:MenuItem( "Select All" ) then
        self.selected_values = shallowcopy(pres.anim_track)
        self.selected_value = nil
    end
    if #self.selected_values > 0 and ui:MenuItem( string.format( "Copy %d", #self.selected_values )) then
        self.allcopy = deepcopy(self.selected_values)
    end
    if ui:MenuItem("Copy All", nil, nil, pres.anim_track and #pres.anim_track > 0 ) then
        self.allcopy = deepcopy(pres.anim_track)
    end
    if ui:MenuItem("Paste", nil, nil, self.allcopy ~= nil ) then
        self:PasteTrack( self.allcopy )
    end

    ui:Separator()

    if ui:MenuItem("Delete") then
        self.delete_selection = true
    end
end

function AttackTester:PasteTrack( track )
    for i, v in ipairs( track ) do
        if v.type then
            local hit_frames = GenerateHitTracks( self.presentation )
            table.insert( hit_frames, deepcopy(v))

        elseif v.is_flipped ~= nil then
            local flip_frames = GenerateTracks( pres, "flip_frames" )
            table.insert( flip_frames, deepcopy(v))

        elseif v.sound_frames ~= nil then
            local sound_frames = GenerateTracks( pres, "sound_frames" )
            table.insert( sound_frames, deepcopy(v))

        else
            local fx_track = GenerateFXTracks( self.presentation )
            table.insert( fx_track, deepcopy(v))
        end
    end
    
    self:PushState( "paste" )
    self:RefreshAnimTrack()
end

function AttackTester:DeleteTrackProperty( t )
    local meta = self.presentation.meta
    if meta then
        if meta.fx_tracks then
            table.arrayremove( meta.fx_tracks, t )
            if #meta.fx_tracks == 0 then
                meta.fx_tracks = nil
            end
        end
        if meta.hit_frames then
            table.arrayremove( meta.hit_frames, t )
            if #meta.hit_frames == 0 then
                meta.hit_frames = nil
            end
        end
        if meta.shake_frames then
            table.arrayremove( meta.shake_frames, t )
            if #meta.shake_frames == 0 then
                meta.shake_frames = nil
            end
        end
        if meta.flip_frames then
            table.arrayremove( meta.flip_frames, t )
            if #meta.flip_frames == 0 then
                meta.flip_frames = nil
            end
        end
        if meta.sound_frames then
            table.arrayremove( meta.sound_frames, t )
            if #meta.sound_frames == 0 then
                meta.sound_frames = nil
            end
        end
        if table.count( meta ) == 0 then
            local anim_meta = GetAnimMeta( self.presentation.meta_filename )
            assert_warning( anim_meta[ self.presentation.anim ] == meta )
            anim_meta[ self.presentation.anim ] = nil
            print( "Deleting entry for", self.presentation.meta_filename, self.presentation.anim )
        end
    end
end

function AttackTester:EditWaypoints( ui, meta )
    for i, t in ipairs( meta.waypoints or table.empty ) do
        ui:PushID( i )
        local x, y = ui:InputFloat2( "Offset", t[1] or 0, t[2] or 0)
        if x and y then
            t[1], t[2] = x, y
            self:PushState( "edit waypoint" )
        end

        ui:SameLine( nil, 20 )
        if ui:Button( "Delete" ) then
            table.remove( meta.waypoints, i )
            if #meta.waypoints == 0 then
                meta.waypoints = nil
            end
            self:PushState( "delete waypoint" )
        end
        ui:PopID()
    end
    if ui:Button( "Add Waypoint" ) then
        if meta.waypoints == nil then
            meta.waypoints = {}
        end
        table.insert( meta.waypoints, { 0, 0 } )
        self:PushState( "add waypoint" )
    end
end

function AttackTester:PresEditAnimMeta(ui, meta)
    if ui:TreeNode( "Attack Waypoints" ) then
        self:EditWaypoints( ui, meta )
        ui:TreePop()
    end

    local tagstr = ""
    if meta.hit_tags then
        for i, tag in ipairs( meta.hit_tags ) do
            tagstr = tagstr .. tostring(tag) .. " "
        end
    end

    local hit_tags = self:EditTags(ui, "Hit Tags", meta.hit_tags, self.all_attack_tags)
    if hit_tags ~= meta.hit_tags then
        meta.hit_tags = hit_tags
        self:GenerateHitTags()
        self:PushState( "new hit_tags" )
    end
    if meta.hit_tags and #meta.hit_tags == 0 then
        meta.hit_tags = nil
    end

    ui:Separator()
end


function AttackTester:RenderAnimTrackFolder( ui, pres, folder_name )
    local popup
    if folder_name == nil or ui:TreeNode( tostring(folder_name), ui.TreeFlags_DefaultOpen  ) then
        for i,v in ipairs(pres.anim_track) do
            if v.folder_name == folder_name then
                ui:PushID( tostring(i) )
                ui:BeginGroup( tostring(i) )

                ui:ColorButton( v.colour_tag or 0x000000FF, true, true )

                ui:SameLine( 20 )
                local status = v.vis_status or VIS_STATUS.DEFAULT
                ui:Image( assets[ status ], 16, 16 )

                if ui:IsMouseClicked() and ui:IsItemHovered() then
                    v.vis_status = GetNextVisStatus( v.vis_status )
                    self:PushState( string.format( "vis_status: %s", tostring(v.vis_status) ))
                end

                ui:SameLine( 30 )
                local text_clr
                if status == VIS_STATUS.HIDE_IN_EDITOR or status == VIS_STATUS.HIDE_IN_GAME then
                    text_clr = 0x777777FF
                elseif status == VIS_STATUS.DEFAULT and v.type == nil and pres.match_tags and not FXTrack.MatchFxTrackTags( v, pres.match_tags ) then
                    text_clr = 0x777777FF
                elseif (v.fx and #v.fx == 0) or (v.sound and #v.sound == 0) or (v.meta_sound and #v.meta_sound == 0) then
                    text_clr = 0xCC0000FF
                end
                if text_clr then
                    ui:PushStyleColor( ui.Style_Text, text_clr )
                end

                -- Index
                ui:SameLine( 50 )
                local is_selected = table.arrayfind( self.selected_values, v ) ~= nil
                if ui:Selectable( string.format( "%2d)", i ), is_selected ) then
                    if TheGame:GetInput():IsModifierShift() then
                        self:ToggleSelection( v )
                    else
                        self:SetSelection( v )
                    end
                    self:RefreshSelectionBoxes()
                end

                -- Frame #
                ui:SameLine( 90 )
                ui:Text( tostring(v.frame))

                -- Type
                ui:SameLine( 130 )
                if v.fx then
                    ui:Text( "fx" )
                elseif v.sound then
                    ui:Text( "sound" )
                elseif v.vo then
                    ui:Text( "vo" )
                elseif v.shake then
                    ui:Text( "shake" )
                elseif v.meta_sound then
                    ui:Text( "meta_sound" )
                elseif v.type then
                    ui:Text( "hit" )
                elseif v.is_flipped ~= nil then
                    ui:Text( "flip" )
                elseif v.duration ~= nil then
                    ui:Text( "lerp" )
                elseif v.command then
                    ui:Text( "command" )
                else
                    ui:Text( "???" )
                end

                -- Effect
                ui:SameLine( 200 )
                if v.fx then
                    ui:Text( loc.format("{1} [{2}]", tostring(v.fx), v.state or "effect" ))
                elseif v.sound then
                    ui:Text( tostring(v.sound) )
                elseif v.meta_sound then
                    ui:Text( tostring(v.meta_sound) )
                elseif v.vo then
                    ui:Text( v.vo )
                elseif v.shake then
                    ui:Text( string.format( "Shake: %2.2f, %2.2f, %2.2f", v.shake[1], v.shake[2], v.shake[3]) )
                elseif v.type then
                    ui:Text( tostring(v.type ))
                elseif v.is_flipped ~= nil then
                    ui:Text( tostring(v.is_flipped) )
                elseif v.duration ~= nil and v.start_bloom then
                    ui:Text( string.format( "bloom %.1f - %.1f", v.start_bloom, v.end_bloom ))
                elseif v.command then
                    if v.command == "merge" then
                        ui:TextColored( 0x00FFFFFF, loc.format( "{1} ({2}:{3})", v.command, v.filename or "<NONE>", v.anim ))
                    else
                        ui:TextColored( 0x00FFFFFF, v.command )
                    end
                else
                    ui:Text( tostr(v) )
                end

                if text_clr then
                    ui:PopStyleColor()
                end

                ui:EndGroup()
                if ui:BeginPopupContextItem( i ) then
                    if self.popup_value == nil then
                        -- Ensure on popup that this entry is selected.
                        self.popup_value = v
                        if TheGame:GetInput():IsModifierShift() then
                            self:UnionSelection( v )
                        else
                            self:SetSelection( v )
                        end
                    end

                    self:PresEditContextMenu( ui, pres )
                    ui:EndPopup()
                    popup = true

                elseif self.popup_value == v then
                    self.popup_value = nil
                end

                ui:PopID()
            end
        end
        if folder_name ~= nil then
            ui:TreePop()
        end
    end

    return popup
end

function AttackTester:RenderAnimTrack( ui, pres )
    local folder_names = self:CollectFolderNames()

    local popup = false
    if self:RenderAnimTrackFolder( ui, pres, nil ) then
        popup =true
    end

    for i, name in ipairs( folder_names ) do
        if self:RenderAnimTrackFolder( ui, pres, name ) then
            popup = true
        end
    end

    return popup
end

function AttackTester:PresEditSources(ui, pres)
    ui:Text( "Parent:" )
    ui:SameLine( nil, 10 )
    local meta_filename = GetAnimMetaFilename( pres.animdat.anim )
    local meta_file = GetAnimMeta( meta_filename )
    if ui:Button( meta_file._parent or "<NONE>" ) then
        -- Source this anim file.
        ui:OpenPopup( "SOURCE_POPUP" )
    end

    if ui:BeginPopup( "SOURCE_POPUP" ) then
        for i, filename, v in sorted_pairs( GetAnimMetaTable() ) do
            if v._parent == nil then
                if ui:MenuItem( filename ) then
                    meta_file._parent = filename
                    self:PushState( "set parent" )
                end
            end
        end

        ui:Separator()

        if meta_file._parent and ui:MenuItem( "Clear Source" ) then
            meta_file._parent = nil
            self:PushState( "clear parent" )
        end

        local name = ui:InputText( "New Source", "", ui.TextFlags_EnterReturnsTrue )
        if name then
            -- This will instantiate a new empty anim meta table.
            local new_meta = GetAnimMeta( name )
            meta_file._parent = name
            self:PushState( "set new parent" )
        end
        ui:EndPopup()
    end


    if pres.meta_filename then
        local meta_filenames = ObtainWorkTable()

        -- Legal sources include this anim file and its parents.
        table.insert( meta_filenames, pres.meta_filename )
        local default_filename = GetAnimMetaFilename( pres.animdat.anim )
        local meta_filename = default_filename
        while meta_filename do
            table.insert_unique( meta_filenames, meta_filename )
            local meta_file = GetAnimMeta( meta_filename )
            meta_filename = meta_file._parent
        end
        table.sort( meta_filenames )

        local idx = table.find( meta_filenames, pres.meta_filename )
        local new_idx = ui:Combo( "Source", idx, meta_filenames )

        if new_idx and new_idx ~= idx then
            local meta = GenerateAnimMeta( default_filename, pres.anim )
            if meta_filenames[ new_idx ] == default_filename then
                pres.meta_filename = default_filename
            else
                pres.meta_filename = meta_filenames[ new_idx ]
            end

            self:RefreshAnimTrack()
            for i, ent in ipairs( self.target_entities ) do
                self:RefreshAnimTrack( ent )
            end
        end
        ReleaseWorkTable( meta_filenames )
    end
end

function AttackTester:DeleteSelection()
    if #self.selected_values > 0 then
        while #self.selected_values > 0 do
            self:DeleteTrackProperty( table.remove( self.selected_values ))
        end
        self:RefreshAnimTrack()
        self.delete_selection = nil
        self:PushState( "delete selection" )
    end
end

function AttackTester:PresEdit(ui, pres, cxt)

    local popup = false
    if #pres.anim_track == 0 then
        ui:Separator()
        local meta = GetAnimMeta( pres.meta_filename )
        local parent_meta = meta._parent and GetAnimMeta( meta._parent )
        ui:Text( string.format( "No data in %s/%s.\nRight-click to add effects.", tostring(pres.meta_filename), tostring(pres.anim) ))

    elseif ui:CollapsingHeader( pres.txt, ui.TreeFlags_DefaultOpen ) then
        ui:BeginChild( "LIST", -20, math.min( #pres.anim_track * 24, 300 ) )
        popup = self:RenderAnimTrack( ui, pres )
        ui:EndChild()
    end

    if not popup and ui:BeginPopupContextItem( cxt or "cxt" ) then
        self:PresEditContextMenu( ui, pres )
        ui:EndPopup()
    end

    if pres.anim_track then
        local idx = table.arrayfind( pres.anim_track, self.selected_value )
        if idx then
            ui:Spacing()
            ui:Separator()
            if self.selected_value.fx then
                self:PresFXEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.sound or self.selected_value.meta_sound then
                self:PresSoundEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.vo then
                self:PresVOEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.shake then
                self:PresShakeEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.type then
                self:PresHitFrameEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.is_flipped ~= nil then
                self:PresFlipFrameEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.duration then
                self:PresActiveTrackEdit( ui, pres, idx, self.selected_value )
            elseif self.selected_value.command then
                self:PresFrameEdit( ui, pres, self.selected_value )
            end

        elseif #self.selected_values > 0 then
            self:PresMultiEdit( ui, pres, self.selected_values )
        end

        if self.delete_selection then
            self:DeleteSelection()
        end
    end
end

function AttackTester:Play(loop)
    if self.pres_coro then
        self:Stop()
    end

    self.looping = loop
    self.looping_dt = self.loop_extra_time

    local attack_name = self.valid_attacks[ self.attack_idx ]
    local attack_entity, attack_fighter = self.current_entity, self.current_entity.cmp.AnimFighter.fighter
    local target_fighters = self:GetPrimaryTargets()
    local primary_target = target_fighters[1]

    if attack_name and primary_target then
        local card = Battle.Card( attack_name, attack_fighter )
        card:AssignTarget( target_fighters[1], { target_fighters } )

        local attack = Battle.Attack( card )
        attack:GenerateHits()

        for i, tag in ipairs( self.editor_hit_tags ) do
            for j, hit in attack:Hits() do
                hit:SetHitResult( tag )
            end
            table.insert( attack.hit_tags, tag )
        end

        local function PlayAttack( anim_fighter, target_fighter, attack )
            print( "Attack:", attack )
            anim_fighter:DoAttack( attack )
            if self.presentation then
                self.presentation.match_tags = attack.hit_tags
            end
            print( "HitTags: ", tostr(attack.hit_tags ))
            
            -- the post attack actually happens internally in a coro, so wait for it.
            anim_fighter:WaitForSequences()
            for i, ent in ipairs( self.target_entities ) do
                ent.cmp.AnimFighter:ResetFighter()
            end

            self.pres_coro = nil
        end

        local anim_fighter = attack_entity.cmp.AnimFighter
        self.pres_coro = anim_fighter:StartCoroutine( PlayAttack, anim_fighter, self.target_entities[1].cmp.AnimFighter, attack )
    
    elseif self.presentation then
        local function PlayPres( cmp, anim )
            local hit_tags = self.editor_hit_tags
            cmp.entity.cmp.AnimController:SetFXTrackTags( hit_tags )
            print(tostr(hit_tags))

            if anim == "death" then
                cmp:DoKill( true )
                cmp:RemoveIdleState( IDLE_STATE.DEAD )
                cmp.entity:Show()
                cmp.entity.cmp.AnimController:SetTintMult(1,1,1,1)
                if cmp.dropshadow then
                    cmp.dropshadow.cmp.PlaneRenderer:TintColour(0,0,0,.5)
                end
                cmp:PlayIdleAnim()

                if cmp.fight_data.fx then
                    for name, fx_def in pairs( cmp.fight_data.fx ) do
                        if fx_def.auto then
                            cmp:PlayUpdatingFX( name )
                        end
                    end
                end
            else
                cmp:StopMoving()
                cmp.entity.cmp.AnimController:Play( anim )
                cmp.entity.cmp.AnimController:CoroWaitForDoneAnims()
                cmp:PlayIdleAnim()
                cmp.coro = nil
            end

            cmp:WaitForSequences()
            self.pres_coro = nil
        end

        local anim_fighter = self.current_entity.cmp.AnimFighter
        if self.all_anims[ self.anim_idx ] then
            local anim = self.all_anims[ self.anim_idx ].anim
            self.pres_coro = anim_fighter:StartCoroutine( PlayPres, anim_fighter, anim )
        else
            DBG( self.presentation )
        end
    end
end

function AttackTester:Stop()
    if self.pres_coro then
        self.pres_coro:Stop()
        self.pres_coro = nil
    end

    self.current_entity.cmp.AnimFighter:AbortSequence()

    self.looping = false
    self:ResetPositions()    
end

function AttackTester:RenderPlaytimePanel(ui)
    ui:Separator()

    if ui:Checkbox( "Loop", self.looping ) then
        if self.looping then
            self:Stop()
        else
            self:Play(true)
        end
        self:SaveDebugSettings()
    end
    if ui:BeginPopupContextItem( "LOOP_CXT" ) then
        self.loop_extra_time = ui:SliderFloat( "Extra Loop Time", self.loop_extra_time, 0, 10.0 )
        ui:EndPopup()
    end

    ui:SameLine( nil, 20 )
    if ui:Checkbox( "Pause", self.paused ) then
        self.paused = not self.paused
        TheGame:SetWallTimeScale( self.paused and 0 or self.speed or 1.0 )
        self:SaveDebugSettings()
    end


    ui:SameLine( nil, 20 )

    if self.paused and ui:Button("Step", 100) then
        self.fightsim:DoFullSimUpdate( 1 * FRAMES, false )
    elseif not self.paused and ui:Button("Play!", 100 ) then
        self:Play()
    end

    ui:SameLine( nil, 10 )
    ui:PushStyleColor( ui.Style_Button, 0x555500FF )
    if ui:Button( string.format( "Speed... (%.2fx)", self.speed or 1.0 )) then
        ui:OpenPopup( "SPEED_CXT" )
    end
    ui:PopStyleColor()

    ui:BeginGroup()
    ui:Columns(2)
    local cmp = self.attack_entity.cmp.AnimController
	if cmp and cmp.animmodel then
        ui:Text(string.format("Attacker: %2d/%d %s (x%.2f)", cmp:GetFrame()+1, cmp.animmodel:GetAnimFrameLength(), tostring(cmp.animName), self.speed or 1.0 ))
        if ui:IsItemHovered() then
            ui:SetTooltip( string.format( "Time in anim: %f", cmp:GetTimeInAnim() ))
        end
	end
	ui:NextColumn()
    cmp = self.target_entities[1] and self.target_entities[1].cmp.AnimController
    if cmp and cmp.animmodel then
        ui:Text(string.format("Target: %2d/%d %s (x%.2f)", cmp:GetFrame()+1, cmp.animmodel:GetAnimFrameLength(), tostring(cmp.animName), self.speed or 1.0 ))
        if ui:IsItemHovered() then
            ui:SetTooltip( string.format( "Time in anim: %f", cmp:GetTimeInAnim() ))
        end
    end
    ui:Columns(1)
    ui:EndGroup()

    if ui:BeginPopup( "SPEED_CXT" ) then
        if ui:MenuItem( "Reset Speed" ) then
            self.speed = 1.0
        end
        self.speed =  ui:SliderFloat("SPEED", self.speed or 1, 0.01, 2)
        if not self.paused then
            TheGame:SetWallTimeScale(self.speed)
        end
        ui:EndPopup()
    end
end

function AttackTester:ResetPositions()
    local attack_entity = self.attack_entity

    attack_entity.cmp.AnimFighter:AbortSequence()
    attack_entity.cmp.AnimFighter:SetHomePosition( 0, self.target_depth )
    attack_entity.cmp.AnimFighter:SetPos( attack_entity.cmp.AnimFighter:GetHomePosition() )
    attack_entity.cmp.AnimFighter:StopMoving()
    attack_entity.cmp.AnimController:StopAnim()
    attack_entity.cmp.AnimFighter:PlayIdleAnim()
    attack_entity.cmp.AnimController:SetXFlip(false)

    local dy = 2 * FIGHT_FIELD_DEPTH / #self.target_entities
    local y = -FIGHT_FIELD_DEPTH + dy/2

    for i, entity in ipairs(self.target_entities) do
        entity.cmp.AnimFighter:AbortSequence()
        entity.cmp.AnimFighter:SetPos( self.target_dist, y )
        entity.cmp.AnimFighter:SetHomePosition( self.target_dist, y )
        entity.cmp.AnimController:SetXFlip(true)
        entity.cmp.AnimController:StopAnim()
        entity.cmp.AnimFighter:StopMoving()
        entity.cmp.AnimFighter:PlayIdleAnim()
        y = y + dy
    end

    if self.looping then
        self:Play(true)
    end
end

local function ShowMenuForCharacterDef( ui, content_id )
    local character_id, character_skin
    local skins = Content.GetAllCharacterSkins()[ content_id ]
    if skins and #skins > 1 then
        if ui:BeginMenu( content_id ) then
            for i, skin_table in ipairs( skins ) do
                local str
                if skin_table.alias then
                    str = string.format( "%s (%s)", skin_table.name, skin_table.alias )
                else
                    str = tostring(skin_table.name)
                end
                if ui:MenuItem( str ) then
                    character_id, character_skin = content_id, skin_table
                end
            end
            ui:EndMenu()
        end

    elseif ui:MenuItem( content_id ) then
        character_id = content_id
        character_skin = skins and skins[1]
    end

    return character_id, character_skin
end

function AttackTester:RenderAgentsPopup( ui )
    local character_id, character_skin

    if ui:BeginMenu( "Special Agents" ) then
        for i, content_id in ipairs( self.aliased_agents ) do
            character_id, character_skin = ShowMenuForCharacterDef( ui, content_id )
            if character_id then
                break
            end
        end
        ui:EndMenu()
    end

    if character_id == nil then
        for i, content_id in ipairs( self.player_defs ) do
            character_id, character_skin = ShowMenuForCharacterDef( ui, content_id )
            if character_id then
                break
            end
        end
    end

    if character_id == nil then
        for i, content_id in ipairs( self.agent_defs ) do
            character_id, character_skin = ShowMenuForCharacterDef( ui, content_id )
            if character_id then
                break
            end
        end
    end

    return character_id, character_skin
end


function AttackTester:SwitchScreens(fn)
    self:Save()
    fn()
end

function AttackTester:ShowRecentEdit( ui )
    if self.edit_msg then
        local a = 1.0 - clamp( (os.clock() - (self.edit_time or 0)) / 2.0, 0, 1.0 )
        if a > 0 then
            local w, h = ui:DisplaySize()
            ui:SetNextWindowSize( w / 5, 36, ui.ImGuiSetCond_Once )
            ui:SetNextWindowPos( w - w / 5, h - 36, ui.ImGuiSetCond_Once)

            local show = ui:Begin( "EDIT", ui.NoTitleBar | ui.NoResize | ui.NoMove | ui.NoInputs | ui.NoFocusOnAppearing )
            local clr = MakeHexColour( 0, 0.8, 0.8, a )
            ui:TextColored( clr, tostring(self.edit_msg))
            ui:End()
        end
    end
end

function AttackTester:RenderEditorPanel(ui)
    local w, h = ui:DisplaySize()
    ui:SetNextWindowSize( w / 3, h, ui.ImGuiSetCond_Once )
    ui:SetNextWindowPos( 0, 0, ui.ImGuiSetCond_Once)

    local show = ui:Begin( "ATTACK EDITOR", ui.MenuBar | ui.NoMove )
    if show then
        if ui:BeginMenuBar() then
            if ui:BeginMenu( "File" ) then
                if ui:MenuItem( "Reload anim meta data", "R" ) then
                    self:ReloadAnimMeta()
                end
                if ui:MenuItem( "Save" ) then
                   self:Save()
                end
                if ui:BeginMenu( "Save All" ) then
                    if ui:MenuItem( "Confirm" ) then
                        self:SaveAll()
                    end
                    if ui:MenuItem( "Dry Run" ) then
                        self:SaveAll( true )
                    end
                    ui:EndMenu()
                end
                if ui:MenuItem( "Exit" ) then
                   TheGame:FE():PopScreen()
                end
                ui:EndMenu()
            end
            if ui:BeginMenu( "Edit" ) then
                if ui:MenuItem( "Edit FX", nil, self.fx_idx ~= nil ) then
                    if self.fx_idx ~= nil then
                        self:EndEditGenericFX()
                    else
                        self:BeginEditGenericFX()
                    end
                end
                ui:Separator()

                if ui:MenuItem( "Undo", nil, nil, self.undo_head > 1 ) then
                    self:Undo()
                end
                if ui:MenuItem( "Redo", nil, nil, self.undo_head < #self.undo_states ) then
                    self:Redo()
                end
                ui:Separator()
                if ui:MenuItem( string.format( "Clear History (%d/%d edits)", self.undo_head - 1, #self.undo_states - 1 )) then
                    self:ClearUndo()
                end
                ui:EndMenu()
            end
            if ui:BeginMenu( "Scene" ) then
                local target_dist = ui:SliderFloat( "Separation", self.target_dist, 0, 20 )
                if target_dist ~= self.target_dist then
                    self.target_dist = target_dist
                    self:ResetPositions()
                    self:SaveDebugSettings()
                end

                local target_depth = ui:SliderFloat( "Depth", self.target_depth, -FIGHT_FIELD_DEPTH, FIGHT_FIELD_DEPTH )
                if target_depth ~= self.target_depth then
                    self.target_depth = target_depth
                    self:ResetPositions()
                    self:SaveDebugSettings()
                end

                local num_targets = ui:SliderInt("# Targets", self.num_targets, 1, 4)
                if num_targets ~= self.num_targets then
                    self.num_targets = num_targets
                    self:RefreshTargetFighter()
                    self:SaveDebugSettings()
                end

                ui:Separator()

                if ui:BeginMenu( "Background" ) then
                    ui:BeginChild( "PLAX", 400, 400 )
                    for i,plax_id in ipairs( Content.GetAllPlaxSorted() ) do
                        if ui:MenuItem( plax_id ) then
                            self.plax:SetPlaxData( plax_id )
                            self.plax_id = plax_id
                            self:SaveDebugSettings()
                        end
                    end
                    ui:EndChild()
                    ui:EndMenu()
                end

 
                local toggle_bg = ui:Checkbox( "Solid Background", self.solid_bg:IsShown() )
                if toggle_bg then
                    self.solid_bg:SetShown( not self.solid_bg:IsShown() )
                    self.settings.bg_shown = self.solid_bg:IsShown()
                    self:SaveDebugSettings()
                end

                local new_clr = ui:ColorEdit3( "Current", self.settings.bg_colour or 0 )
                if new_clr and new_clr ~= self.settings.bg_colour then
                    self.settings.bg_colour = new_clr
                    self.solid_bg:SetColour( new_clr )
                    self:SaveDebugSettings()
                end

                ui:EndMenu()

            end
            if ui:BeginMenu( "Debug" ) then
                if ui:MenuItem( "This" ) then
                    TheGame:GetDebug():CreatePanel( DebugTable( self ) )
                end
                if ui:MenuItem( "Presentation" ) then
                    TheGame:GetDebug():CreatePanel( DebugTable( self.presentation ) )
                end
                if self.attack_entity then
                    if ui:MenuItem( "Attack Fighter" ) then
                        TheGame:GetDebug():CreatePanel( DebugTable( self.attack_entity.cmp.AnimFighter.fighter ) )
                    end
                    if ui:MenuItem( "Attack Entity" ) then
                        TheGame:GetDebug():CreatePanel( DebugTable( self.attack_entity ) )
                    end
                end

                for i, entity in ipairs(self.target_entities) do
                    if ui:MenuItem( loc.format( "Target Entity ({1})", i) ) then
                        TheGame:GetDebug():CreatePanel( DebugTable( entity ) )
                    end
                end
                if ui:MenuItem( "Audio" ) then
                    DBG( DebugAudio() )
                end

                ui:Separator()
                if ui:MenuItem( "View Missing Files Checker..." ) then
                    DBG( DebugAttackAnims() )
                end
                if ui:MenuItem( "View Anim Counts..." ) then
                    local anim_counts = {}
                    for i, id in ipairs( self.valid_attacks ) do
                        local carddef = Content.GetBattleCard( id )
                        if carddef.anim then
                            anim_counts[ carddef.anim ] = (anim_counts[ carddef.anim ] or 0) + 1
                        end
                    end
                    local panel = DBG( anim_counts )
                    panel:SetLockedValues( true )
                end

                ui:Separator()

                if ui:Checkbox( "Hide Gnomons", self.hide_fx_gnomons ) then
                    self.hide_fx_gnomons = not self.hide_fx_gnomons
                    self:SaveDebugSettings()
                end

                ui:EndMenu()
            end

            ShowDebugToolsMenu(self:GetFE(), ui, function(fn) self:SwitchScreens(fn) end )

            local dt = self.save_time and (os.clock() - self.save_time)
            if dt and dt < 3.0 then
                ui:SameLine( ui:GetWindowSize() - 60 )
                local alpha = easing.outCubic( dt, 1.0, -1.0, 3.0 )
                ui:TextColored( MakeHexColour( 0, 1, 0, alpha ), "Saved!" )
            
            elseif self:IsDirty() then
                ui:SameLine( ui:GetWindowSize() - 50 )
                ui:TextColored( 0xFF0000FF, "Dirty" )
                if ui:IsItemHovered() then
                    ui:BeginTooltip()
                    ui:Text( "Dirty files:" )
                    local dirty_files = self:GetDirtyFiles()
                    ui:TextColored( 0xFF0000FF, table.concat( dirty_files, "\n" ))
                    if dt then
                        ui:Text( loc.format( "Last save: {1} {1*min|mins} ago", math.floor( dt / 60 )))
                    end
                    ui:EndTooltip()
                end
            end

            ui:EndMenuBar()
        end

        if self:IsOnStack() then
            self:ShowRecentEdit( ui )

            do
                ui:Columns(2)
                if ui:Button( self.attack_agent:GetContentID() ) then
                    ui:OpenPopup( "LEFT_AGENT" )
                end
                if ui:BeginPopup( "LEFT_AGENT" ) then
                    local content_id, skin_table = self:RenderAgentsPopup( ui )
                    if content_id then
                        self:RefreshFighter( content_id, skin_table )
                        self:SaveDebugSettings()

                        if self.looping then
                            self:Play(true)
                        end
                    end
                    ui:EndPopup()
                end
                if ui:Checkbox( "Primary##LEFT", self.current_entity == self.attack_entity ) then
                    self:SwitchCurrentEntity()
                end

                ui:NextColumn()

                if ui:Button( string.format( "%s##RIGHT", self.target_agent:GetContentID() )) then
                    ui:OpenPopup( "RIGHT_AGENT" )
                end
                if ui:BeginPopup( "RIGHT_AGENT" ) then
                    local content_id, skin_table = self:RenderAgentsPopup( ui )
                    if content_id then
                        self:RefreshTargetFighter( content_id, skin_table )
                        self:SaveDebugSettings()

                        if self.looping then
                            self:Play(true)
                        end
                    end
                    ui:EndPopup()
                end
                if ui:Checkbox( "Primary##RIGHT", self.current_entity == self.target_entities[1] ) then
                    self:SwitchCurrentEntity()
                end

                ui:NextColumn()

                ui:Columns(1)
                ui:Spacing()

                local attack_name = self.valid_attacks[ self.attack_idx ] or "<NO ATTACK SET>"
                ui:PushStyleColor( ui.Style_Button, 0x770033FF )
                if ui:Button( attack_name, ui:GetItemWidth() ) then
                    ui:OpenPopup( "ATTACKS_POPUP" )
                end
                ui:PopStyleColor()

                if ui:BeginPopup( "ATTACKS_POPUP" ) then
                    -- ui:BeginChild( "Attacks", 500, 400 )
                    self.filter_txt = ui:InputText( "Filter", self.filter_txt or "" ) or self.filter_txt
                    ui:Separator()
                    
                    for i, id in ipairs( self.valid_attacks ) do
                        local carddef = Content.GetBattleCard( id )
                        local anim_name = carddef and tostring(carddef.anim )
                        if carddef and (self.filter_txt == nil or id:find( self.filter_txt ) or (anim_name and anim_name:find(self.filter_txt))) then
                            ui:Text( tostring( i ))
                            ui:SameLine( 40 )

                            if ui:Selectable( id ) then
                                self:SetAttackIDX( i )
                                self:SaveDebugSettings()
                                if self.looping then
                                    self:Play(true)
                                end
                            end
                            if ui:IsItemHovered() then
                                ui:SetTooltip( carddef.name )
                            end

                            ui:SameLine( 200 )
                            ui:TextColored( 0xCCCCCCFF, tostring(carddef.anim ))

                            ui:SameLine( 300 )
                            local bitstring = StringizeBitField( carddef.flags, battle_defs.CARD_FLAG_BITSTRINGS )
                            ui:TextColored( 0x00AACCFF, tostring(carddef.rarity ))
                            ui:SameLine( nil, 5 )
                            ui:TextColored( 0x00FFFFFF, bitstring )
                        end
                    end

                    ui:EndPopup()
                end

                local new_attack_idx = ui:Combo( "Anim", self.anim_idx or 1, self.anim_txt )
                if new_attack_idx ~= self.anim_idx and self.all_anims[new_attack_idx] then
                    self:SetAttackIDX( nil )
                    self:SetAnimIDX(new_attack_idx)
                    if self.looping then
                        self:Play(true)
                    end
                end
                
                self:EditTags(ui, "Editor Tags:", self.editor_hit_tags, self.all_attack_tags)

                if self.cardw and ui:Checkbox( "Show Card", self.cardw:IsShown() ) then
                    self.cardw:SetShown( not self.cardw:IsShown() )
                    self:SaveDebugSettings()
                end
                if self.attack_image then
                    ui:Image( self.attack_image, self.attack_image:GetSize() )
                end
            end

            self:RenderPlaytimePanel( ui )

            ui:PushID("attacker")
            ui:Spacing()

            if self.fx_idx then
                self:PresGenericFXEdit( ui )

            elseif self.presentation then
                self:PresEditSources(ui, self.presentation)
                ui:Spacing()

                if self.attack_idx ~= nil and self.presentation.meta and ui:CollapsingHeader( "Attack Data" ) then
                    self:PresEditAnimMeta( ui, self.presentation.meta )
                end

                self:PresEdit(ui, self.presentation, "ATTACKER" )
            end
            ui:PopID()
        end
    end

    ui:End()
end

function AttackTester:UnionSelection( v )
    table.insert_unique( self.selected_values, v )
    self.selected_value = #self.selected_values == 1 and self.selected_values[1] or nil
end

function AttackTester:ToggleSelection( v )
    if not table.arrayremove( self.selected_values, v ) then
        table.insert( self.selected_values, v )
    end
    self.selected_value = #self.selected_values == 1 and self.selected_values[1] or nil
end

function AttackTester:SetSelection( v )
    table.clear( self.selected_values )
    table.insert( self.selected_values, v )
    self.selected_value = v
end

function AttackTester:RefreshSelectionBoxes()
    -- Refresh 'selection' bounding box
    for i, fx in self.fightsim.fx_mgr:PlayingFX() do
        if fx.fxdata == self.selected_value then
            fx.model:SetTag( engine.rendering.SceneGraphModel.TAG_DRAW_BOUNDING_BOX )
        else
            fx.model:ClearTag( engine.rendering.SceneGraphModel.TAG_DRAW_BOUNDING_BOX )
        end
    end
end


function AttackTester:TriggerFXTrack( fxtrack, fxdata, pointer, match_tags )
    local status = fxdata.vis_status or VIS_STATUS.DEFAULT
    if status == VIS_STATUS.ALWAYS_SHOW or
            (status == VIS_STATUS.DEFAULT and FXTrack.MatchFxTrackTags( fxdata, match_tags ) and fxtrack:IsCurrentVariant( fxdata )) then
        --
        local result = fxtrack:TriggerFXTrack( fxdata, pointer )
        self:RefreshSelectionBoxes()
        return result
    end
end

function AttackTester:RefreshFighter( content_id, skin_table )
    local is_current = self.attack_entity == self.current_entity

    if self.attack_entity then
        self.attack_entity:Despawn()
        self.attack_entity = nil
    end

    if content_id then
        if not Content.GetCharacterDef( content_id ) then
            content_id = "SAL"
        end
        self.attack_agent = Agent( content_id, skin_table )
    end

    local fighter = Fighter.CreateFromAgent( self.attack_agent )
    fighter.team = self.teams[TEAM.BLUE]

    self.attack_entity = self.fightsim.mgr:AddFighter( fighter )
    self.attack_entity.cmp.AnimFighter.team = TEAM.BLUE
    self.attack_entity.cmp.AnimController:SetFXTrackFn( function( fxtrack, fxdata, pointer, match_tags )
        return self:TriggerFXTrack( fxtrack, fxdata, pointer, match_tags )
    end )

    if is_current then
        self.current_entity = self.attack_entity
    end

    self:RefreshAttacks()
    self:ResetPositions()

    self.attack_entity:Show()
end

function AttackTester:RefreshTargetFighter( content_id, skin_table )
    local is_current = table.contains( self.target_entities, self.current_entity )

    for i, entity in ipairs(self.target_entities) do
        entity:Despawn()
    end
    self.target_entities = {}

    if self.attack_entity then
        -- Any attack sequence probably stashed a reference to the old target.
        self.attack_entity.cmp.AnimFighter:AbortSequence()
    end
    
    for i = 1, self.num_targets do
        if content_id then
            if not Content.GetCharacterDef( content_id ) then
                content_id = "SAL"
            end
            self.target_agent = Agent( content_id, skin_table )
        end

        local fighter = Fighter.CreateFromAgent( self.target_agent )
        fighter.team = self.teams[TEAM.RED]

        local ent = self.fightsim.mgr:AddFighter( fighter )
        table.insert(self.target_entities, ent)
        ent.cmp.AnimFighter.team = TEAM.RED
        ent.cmp.AnimController:SetFXTrackFn( function( cmp, fxdata, pointer, match_tags )
            return self:TriggerFXTrack( cmp, fxdata, pointer, match_tags )
        end )

        self.target_entities[i].cmp.AnimFighter:PlayIdleAnim()
        
        self:ResetPositions()
        
        self.target_entities[i]:Show()

        if is_current then
            self.current_entity = self.target_entities[1]

            self:RefreshAttacks()
        end
    end
end

function AttackTester:SetAttackIDX( attack_idx )
    self.attack_idx = attack_idx
    if attack_idx == nil then
        self.attack_image = nil
        return
    end

    local attack_name = self.valid_attacks[ self.attack_idx ]
    local attack = Content.GetBattleCard( attack_name )
    if attack then
        local cmp = self.current_entity.cmp.AnimFighter
        local anim = cmp:GetPresentation( attack.anim )

        for i, v in ipairs( self.all_anims ) do
            if v.anim == anim then
                self:SetAnimIDX( i )
                break
            end
        end
    end

    if self.anim_idx == nil then
        self:SetAnimIDX( 1 )
    end

    if attack and not CheckBits( attack.flags, battle_defs.CARD_FLAGS.NPC ) then
        local card = Battle.Card( attack_name )
        if self.cardw == nil then
            self.cardw = Widget.BattleCard( card )
            self:AddChild( self.cardw ):SetAnchors( "right", "top" )
            local w, h = self.cardw:GetSize()
            self.cardw:Offset( -w/2, -h/2 )
            self.cardw:SetShown( self.settings.show_card )
        end
        self.cardw:RefreshCard( card )

        -- ccc: The text rendering on the captured image is broken for some reason.
        -- local cap =  Widget.CaptureImage():SetSize( 370/2, 542/2 )
        -- cap:RenderToScene( self.cardw )
        -- self.attack_image = cap:GetTexture()
    else
        self.attack_image = nil
        if self.cardw then
            self.cardw:Remove()
            self.cardw = nil
        end
    end
end

function AttackTester:SetAnimIDX(idx, match_tags)
    self.anim_idx = idx
    self:RefreshAnimTrack()
end

function AttackTester:SortTrack( track )
    local indices = {}
    for i, v in ipairs( track ) do
        indices[v] = i
    end
    local function TrackCompare( a, b )
        if a.frame == b.frame then
            return indices[a] < indices[b]
        else
            return a.frame < b.frame
        end
    end
    table.sort( track, TrackCompare )
end

local function SortByFrame( t1, t2 )
    return t1.frame < t2.frame
end

-- Takes an anim meta table (which typically has subfields like fx_track, hit_frames, etc.) and merges them into a
-- singular array sorted by frame, for UI presentation.

function AttackTester:GenerateUnifiedTrack( meta )
    if meta then
        local anim_track = {}
        if meta.fx_tracks then
            self:SortTrack( meta.fx_tracks )
            for i, t in ipairs( meta.fx_tracks ) do
                table.binsert( anim_track, t, SortByFrame )
            end
        end
        if meta.shake_frames then
            self:SortTrack( meta.shake_frames )
            for i, t in ipairs( meta.shake_frames ) do
                table.binsert( anim_track, t, SortByFrame )
            end
        end

        if meta.hit_frames then
            self:SortTrack( meta.hit_frames )
            for i, t in ipairs( meta.hit_frames ) do
                table.binsert( anim_track, t, SortByFrame )
            end
        end

        if meta.flip_frames then
            self:SortTrack( meta.flip_frames )
            for i, t in ipairs( meta.flip_frames ) do
                table.binsert( anim_track, t, SortByFrame )
            end
        end

        if meta.sound_frames then
            self:SortTrack( meta.sound_frames )
            for i, t in ipairs( meta.sound_frames ) do
                table.binsert( anim_track, t, SortByFrame )
            end
        end

        return anim_track
    else
        return table.empty
    end
end

function AttackTester:RefreshAnimTrack( entity )
    entity = entity or self.current_entity

    local anim = self.anim_idx and self.all_anims[ self.anim_idx ] and self.all_anims[ self.anim_idx ].anim
    local animdat = entity.cmp.AnimController:GetAnimInfo( anim )
    if animdat == nil then
        return
    end

    local anim_filename = GetAnimMetaFilename( animdat.anim )
    local meta, meta_filename = GetAnimMeta( anim_filename, anim )

    -- If we are viewing an alternate source meta, check that it is a valid parent here.
    if self.presentation then
        local meta_filename_tmp, meta_tmp = anim_filename, GetAnimMeta( anim_filename )
        while meta_tmp do
            meta_filename_tmp = meta_tmp._parent
            meta_tmp = meta_filename_tmp and GetAnimMeta( meta_filename_tmp )
            if self.presentation.meta_filename == meta_filename_tmp then
                meta, meta_filename = GetAnimMeta( meta_filename_tmp, anim )
                break
            end
        end
    end

    entity.cmp.AnimController:RegenerateAnimInfo( anim )

    if entity == self.current_entity then
        if self.fx_idx then
            anim = self.all_fx[ self.fx_idx ]
            meta, meta_filename = GetFXTrack( anim ), GetFXTrackFilename()

            local anim_track = self:GenerateUnifiedTrack( meta )

            self.presentation =
            {
                anim = anim,
                animdat = nil,
                meta = meta,
                meta_filename = meta_filename,
                anim_track = anim_track,
                txt = string.format( "FX - %s", anim ),
                match_tags = table.empty,
            }
        else
            local anim_track = self:GenerateUnifiedTrack( meta )

            self.presentation =
            {
                anim = anim,
                animdat = animdat,
                meta = meta,
                meta_filename = meta_filename,
                anim_track = anim_track,
                txt = string.format( "%s - %s", self.all_anims[ self.anim_idx ].txt, meta_filename ),
                match_tags = table.empty,
            }
        end
    end
end

function AttackTester:ReloadAnimMeta()
    print( "Reloaded anim meta data" )
    LoadAnimMeta()

    -- AnimControllers need to re-reference the meta tables.
    self:RefreshFighter()
    self:RefreshTargetFighter()
end

function AttackTester:Save( )

    local filenames = self:GetDirtyFiles()

    local st = os.clock()
    self.save_time = st
    self.save_head = self.undo_head

    for i, filename in ipairs( filenames ) do
        self:AmendAnimData( filename )

        local outs = SaveAnimMeta( filename )

        filename = string.format( "scripts/content/anim/%s.lua", filename )

        engine.data:SaveFile( filename, outs, function(ok) 
            print( string.format( "Write file %s (%.1f ms) : '%s'", ok and "OK" or "FAILED", (os.clock() - st) * 1000.0, filename))
        end )
    end

    table.clear( self.dirty_files )
end

function AttackTester:AmendAnimData( anim_filename, anims )
    if anims == nil then
        if anim_filename == GetFXTrackFilename() then
            anims = GetFXTrack()
        else
            anims = GetAnimMetaTable()[ anim_filename ]
        end
    end

    for anim, anim_meta in pairs(anims) do
        -- if anim == "death" then
        --     if anims._parent == "med_combat_humanoid" then
        --         print( anim, anim_filename, anims._parent )
        --         for i, v in ipairs( anim_meta.fx_tracks ) do
        --             if v.command == "merge" and v.no_sound == nil then
        --                 print( "\tADDED no_sound" )
        --                 v.no_sound = true
        --                 break
        --             end
        --         end
        --         -- table.insert( anim_meta.fx_tracks, 1, { command = "merge", filename = "med_combat_humanoid", anim = "death", frame = 0 } )
        --     end
        -- end

        if anim_meta.filename then
            print( "#### WARNING:", anim_filename, anim, " referencing deprecated filename field", anim_meta.filename )
        end

        if anim_meta.fx_tracks then

            for i = #anim_meta.fx_tracks, 1, -1 do
                local t = anim_meta.fx_tracks[i]

                t.one_of_tags = nil
                if t.must_have_tags and #t.must_have_tags == 0 then
                    t.must_have_tags = nil
                end
                if t.cant_have_tags and #t.cant_have_tags == 0 then
                    t.cant_have_tags = nil
                end

                -- if t.vo then
                --     -- Move to the sound frame.
                --     if anim_meta.sound_frames == nil then
                --         anim_meta.sound_frames = {}
                --     end

                --     print( "Moved VO:", t.vo )
                --     table.insert( anim_meta.sound_frames, 1, t )
                --     table.remove( anim_meta.fx_tracks, i )
                --     if #anim_meta.fx_tracks == 0 then
                --         anim_meta.fx_tracks = nil
                --     end

                -- elseif t.sound then
                --     if not SoundEvents[ t.sound ] then
                --         print( "MISSING ENTRY FOR:", anim, t.sound )

                --     elseif not AUDIO:FindEvent( SoundEvents[ t.sound ] ) then
                --         print( "MISSING SOUND:", anim, t.sound, SoundEvents[ t.sound ] )
                --     end

                --     -- Move to the sound frame.
                --     if anim_meta.sound_frames == nil then
                --         anim_meta.sound_frames = {}
                --     end

                --     print( "Moved sound:", t.sound )
                --     table.insert( anim_meta.sound_frames, 1, t )
                --     table.remove( anim_meta.fx_tracks, i )
                --     if #anim_meta.fx_tracks == 0 then
                --         anim_meta.fx_tracks = nil
                --     end
                -- end
            end
        end

        if anim_meta.sound_frames then
            for i = #anim_meta.sound_frames, 1, -1 do
                local t = anim_meta.sound_frames[i]
                t.one_of_tags = nil
                if t.must_have_tags and #t.must_have_tags == 0 then
                    t.must_have_tags = nil
                end
                if t.cant_have_tags and #t.cant_have_tags == 0 then
                    t.cant_have_tags = nil
                end
            end
        end
    end
end

function AttackTester:SaveAll( dry_run )
    local st = os.clock()
    local ANIM_META = GetAnimMetaTable()

    for filename, v in pairs( ANIM_META ) do
        self:AmendAnimData( filename, v )

        local outs = SaveAnimMeta( filename )

        filename = string.format( "scripts/content/anim/%s.lua", filename )

        if not dry_run then
            engine.data:SaveFile( filename, outs, function(ok) 
                print( string.format( "Write file %s : '%s'", ok and "OK" or "FAILED", filename))
            end )
        end
    end

    do
        local filename = GetFXTrackFilename()
        self:AmendAnimData( filename, GetFXTrack() )
        local outs = SaveAnimMeta( filename )

        filename = string.format( "scripts/content/anim/%s.lua", filename )

        if not dry_run then
            engine.data:SaveFile( filename, outs, function(ok) 
                print( string.format( "Write file %s : '%s'", ok and "OK" or "FAILED", filename))
            end )
        end
    end


    print( string.format( "Took: %.1f ms",  (os.clock() - st) * 1000.0 ))
end


function AttackTester:RenderDebugUI( ui, panel )
    DebugUtil.FightCamDebug(ui, self.fightsim:GetCamera())
end