require "ui/widgets/installed_grafts"
require "ui/widgets/chargecell"

local BlasterScreen = class( "Screen.BlasterScreen", Widget.Screen )

local GRAFT_SLOT_SIZE = 150
local GRAFT_SLOT_CONTENT_W = 540
local DETAILS_W = 540

BlasterScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.character_animation_done and TheGame:FE():IsGamepadMode()
        end,
        fn = function(self)
            self.character_animation_done = false
            self.character:Emote( "gun_idle", true )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.BLASTER_SCREEN.CONTROLS_FLIP_BLASTER", Controls.Digital.MENU_ACCEPT))
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self)
            self:OnClickClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end,
    },
}

function BlasterScreen:init( owner, target_graft_type )
    BlasterScreen._base.init( self )

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.OBJECTIVE )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )


    -- Setup character anim display
    self.owner = owner or TheGame:GetGameState():GetPlayerAgent()
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.owner )
    -- self.character:SetXRayOutline( 0x0a4434FF, UICOLOURS.OBJECTIVE, 0.0, 0.2 )
    self.character:SetHoverOutlineColour( UICOLOURS.OBJECTIVE )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "gun_idle", true )
        end
    end )
    self.character:SetAutoLoopEnabled( false )
    self.character:SetPos( RES_X*0.25, -350 )
        :SetScale( -1, 1 )
    self.character:Emote( "gun_idle", true )

    -- Blaster display!
    self.content = self:AddChild(Widget())
    self.title = self.content:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( loc.upper( loc.format( LOC"UI.BLASTER_SCREEN.TITLE", self.owner:GetName() ) ) )
        :SetGlyphColour( UICOLOURS.OBJECTIVE )
        :SetAutoSize(DETAILS_W)
        :LeftAlign()
        :SetWordWrap(true)
        :Bloom(0.15)
    self.blaster_widget = self.content:AddChild( Widget.BlasterScreenSlot( self.graft_type, GRAFT_SLOT_SIZE ) )
    self.blaster_charges_widget = self.content:AddChild( Widget.BlasterChargesDisplay() )
    self.blaster_overcharge_widget = self.content:AddChild( Widget.BlasterOverchargeDisplay() )
    self.subtitle = self.content:AddChild( Widget.Label("title", FONT_SIZE.BODY_TEXT ) )
        :SetText( loc.format( LOC"UI.BLASTER_SCREEN.SUBTITLE", self.owner:GetName() ) )
        :SetGlyphColour( UICOLOURS.OBJECTIVE )
        :SetAutoSize(DETAILS_W)
        :LeftAlign()
        :SetWordWrap(true)
        :SetTintAlpha( 0.8 )
        :Bloom(0.1)

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    
    self:Refresh()
end

function BlasterScreen:Refresh()

    self.owner = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    self.graft_type = GRAFT_TYPE.BLASTER
    self.max_grafts = self.owner:GetAspect("graft_owner"):GetMaxGrafts( self.graft_type )
    local grafts = ObtainWorkTable()
    self.owner:GetAspect("graft_owner"):GetGrafts( self.graft_type, grafts )

    -- Add new
    self.blaster_widget:SetGraft( grafts[1] )
        :Refresh()
    self.blaster_charges_widget:SetGraft( grafts[1] )
    self.blaster_overcharge_widget:SetGraft( grafts[1] )

    ReleaseWorkTable(grafts)

    -- Layout
    self.blaster_widget:LayoutBounds( "left", "below", self.title )
        :Offset( 0, -SPACING.M1 )
    self.blaster_charges_widget:LayoutBounds( "left", "below", self.blaster_widget )
        :Offset( 26, -SPACING.M05 )
    self.blaster_overcharge_widget:LayoutBounds( "left", "below", self.blaster_charges_widget )
        :Offset( 0, -SPACING.M05 )
    self.subtitle:LayoutBounds( "left", "below", self.blaster_overcharge_widget )
        :Offset( 0, -SPACING.M1 )
    self.content:LayoutBounds( "right", "center", -RES_X*0, 0 )

    return self
end

function BlasterScreen:OnGainTop()
    BlasterScreen._base.OnGainTop( self )

    local anim_duration = 0.7
    local horizontal_movement = 90

    -- Prepare for animation by getting current values
    local content_x, content_y = self.content:GetPos()

    -- Hide objects until they're animated in
    self.content:SetTintAlpha(0)

    -- Setup and start animation
    self.fadein_anim = self:RunUpdater(
        Updater.Parallel{
            -- Fade and move in the coin container
            Updater.Ease( function(v) self.content:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) self.content:SetPos(x,y) end, content_x - horizontal_movement, content_y, content_x, content_y, anim_duration, easing.outQuad),
        }
    )

    self:RefreshStreamerMode()
end

function BlasterScreen:OnControlModeChange( cm, device_id )
    BlasterScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
    end
end

function BlasterScreen:OnScreenModeChange( sm )
    BlasterScreen._base.OnScreenModeChange( self, sm )

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    self.title:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.blaster_widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.blaster_charges_widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.blaster_overcharge_widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.subtitle:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Layout
    self.blaster_widget:LayoutBounds( "left", "below", self.title )
        :Offset( 0, -SPACING.M1 )
    self.blaster_charges_widget:LayoutBounds( "left", "below", self.blaster_widget )
        :Offset( 26, -SPACING.M05 )
    self.blaster_overcharge_widget:LayoutBounds( "left", "below", self.blaster_charges_widget )
        :Offset( 0, -SPACING.M05 )
    self.subtitle:LayoutBounds( "left", "below", self.blaster_overcharge_widget )
        :Offset( 0, -SPACING.M1 )
    self.content:LayoutBounds( "right", "center", RES_X*0.05, 0 )
end

function BlasterScreen:GetDefaultFocus()
    return self.content or self.close_button
end

function BlasterScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function BlasterScreen:OnOpen()
    BlasterScreen._base.OnOpen( self )
    self.screen_loop = AUDIO:CreateEventInstance("event:/ui/grafts_menu/open")
    self.screen_loop:Start()
end

function BlasterScreen:OnClickClose()
    AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
    self:FadeOut()
end

function BlasterScreen:OnClose()
    BlasterScreen._base.OnClose( self )

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil
end

------------------------------------------------------------------------------------------
-- Shows graft slot

local BlasterScreenSlot = class( "Widget.BlasterScreenSlot", Widget.GraftSlot )

function BlasterScreenSlot:init( type, size )
    self._base.init( self, type, size )

    self.slot_size = size or 64
    self.graft_type = type

    self.features = {}

    if not self.bg then
        self.bg = self:AddChild( Widget.Image() )
    end
    self.bg:SetTexture( engine.asset.Texture ("UI/grafts_slot.tex" ) )
    self.bg:SetSize(self.slot_size,self.slot_size)
    self.bg:Bloom(0.2)
    self.bg:SetTintColour( UICOLOURS.OBJECTIVE )

    self.img = self:AddChild( Widget.Image():SetSize( self.slot_size*0.9, self.slot_size*0.9 ) )

    self.text_content = self:AddChild( Widget() )
    self.name = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_NAME ) )
        :SetWordWrap( true )
        :SetAutoSize( GRAFT_SLOT_CONTENT_W )
        :LeftAlign()
        :OverrideLineHeight( FONT_SIZE.ITEM_NAME*LINE_HEIGHT_TITLE_MULT )
        :SetGlyphColour( UICOLOURS.OBJECTIVE )
        :Bloom(0.1)
    self.subtitle = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ):SetAutoSize( GRAFT_SLOT_CONTENT_W ):LeftAlign())
    self.rarity_icon = self.text_content:AddChild( Widget.Image( engine.asset.Texture ("UI/graft_rarity.tex") ))
        :SetSize( 16, 16 )
    self.desc = self.text_content:AddChild( Widget.Label( "body", FONT_SIZE.BODY_TEXT ) )
        :SetWordWrap( true )
        :SetAutoSize( GRAFT_SLOT_CONTENT_W )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TITLE )
        :SetTintAlpha( 0.7 )
        :Bloom(0.1)

    self:SetFocusable()
end

function BlasterScreenSlot:SetGraft( graft )
    if self.graft ~= graft then
        self.graft = graft
        self:Refresh()
    end

    -- self:SetToolTipClass( Widget.TooltipGraft )
    -- self:SetToolTip(graft)
    -- self:ShowToolTipOnFocus()
    return self
end

function BlasterScreenSlot:Refresh()
    if self.graft then
        -- Set values
        self.img:SetTexture(self.graft:GetImage())
        self.name:SetText( loc.upper( self.graft:GetName() ) or LOC"UI.GRAFT_COLLECTION.SLOT_EMPTY" )
        self.desc:SetText( self.graft:GetDesc( self ) )
        -- local rarity_color = MakeColourString( HexColour( CARD_RARITY_COLOURS[ self.graft:GetRarity() or CARD_RARITY.UNIQUE ] ) )
        -- local rarity_name = GetCardRarityString( self.graft:GetRarity() )
        -- local type_name = LOC"UI.GRAFT_COLLECTION.SLOT_TYPE_COIN"
        -- local subtitle_text = loc.format( LOC"UI.GRAFT_COLLECTION.SLOT_SUBTITLE", rarity_color, rarity_name, type_name )
        -- self.subtitle:SetText( string.upper( subtitle_text ) )
        -- self.rarity_icon:SetTintColour( CARD_RARITY_COLOURS[ self.graft:GetRarity() ] )
        -- self.subtitle:SetTintAlpha(1)
        -- self.rarity_icon:SetTintAlpha(1)
        self.subtitle:Hide()
        self.rarity_icon:Hide()

        -- Layout
        -- self.rarity_icon:LayoutBounds( "left", "below", self.name ):Offset( 0, -3 )
        -- self.subtitle:LayoutBounds( "after", "center", self.rarity_icon ):Offset( 6, 2 )
        self.desc:LayoutBounds( "left", "below", self.name ):Offset( 0, -SPACING.TITLE_DESC )
    else
        self.img:SetTexture( nil )
        self.name:SetText( LOC"UI.GRAFT_COLLECTION.SLOT_EMPTY" )
        self:SetToolTip(LOC"UI.GRAFTSLOT.EMPTY_COIN")
        self.desc:SetText( "" )
        local subtitle_text = loc.format( LOC"UI.GRAFT_COLLECTION.SLOT_EMPTY_SUBTITLE" )
        self.subtitle:SetText( subtitle_text )
        self.rarity_icon:SetTintColour( UICOLOURS.WHITE )
        self.subtitle:SetTintAlpha(0.2)
        self.rarity_icon:SetTintAlpha(0.2)

        -- Layout
        self.name:LayoutBounds( "after", "center", self.img ):Offset( 15, 10 )
        self.rarity_icon:LayoutBounds( "left", "below", self.name ):Offset( 0, -3 )
        self.subtitle:LayoutBounds( "after", "center", self.rarity_icon ):Offset( 6, 2 )
        self.desc:LayoutBounds( "left", "top", self.name ):Offset( 0, 0 )
    end
    self.text_content:LayoutBounds( "after", "center", self.img):Offset( 40, 1 )
end

function BlasterScreenSlot:SetFn( fn )
    self:SetOnClickFn( function() fn( self.graft, self ) end )
end

function BlasterScreenSlot:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        self.bg:ScaleTo( nil, 1.02, 0.2, easing.outQuad )
            :Bloom(0.15)
        if self.connector then self.connector:Bloom(0.5) end
        if self.name then self.name:Bloom(0.2) end
        self.img:ScaleTo( nil, 1.1, 0.3, easing.outQuad )
    else
        self.bg:ScaleTo( nil, 1, 0.2, easing.outQuad )
            :Bloom(0.1)
        if self.connector then self.connector:Bloom(0.3) end
        if self.name then self.name:Bloom(0.1) end
        self.img:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function BlasterScreenSlot:OnGainFocus()
    BlasterScreenSlot._base.OnGainFocus(self)
    self:UpdateImage()
end

function BlasterScreenSlot:OnLoseFocus()
    BlasterScreenSlot._base.OnLoseFocus(self)
    self:UpdateImage()
end

function BlasterScreenSlot:OnGainHover()
    BlasterScreenSlot._base.OnGainHover(self)
    self:UpdateImage()
end

function BlasterScreenSlot:OnLoseHover()
    BlasterScreenSlot._base.OnLoseHover(self)
    self:UpdateImage()
end

function BlasterScreenSlot:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function BlasterScreenSlot:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

------------------------------------------------------------------------------------------
-- Shows blaster charges

local BlasterChargesDisplay = class( "Widget.BlasterChargesDisplay", Widget )

local ICON_SIZE = 80

function BlasterChargesDisplay:init()
    self._base.init( self )

    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame.tex" ) ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetBloom( 0.2 )
    self.contents = self:AddChild( Widget() )
    self.charges_label = self.contents:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetAutoSize( DETAILS_W )
        :SetWordWrap( true )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.15 )
    self.charges_text = self.contents:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ))
        :SetAutoSize( DETAILS_W )
        :SetWordWrap( true )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
    self.charges_icons = self.contents:AddChild( Widget() )
end

function BlasterChargesDisplay:SetGraft( graft )
    self.graft = graft
    self.graft_def = self.graft and self.graft:GetDef()
    self.charges_icons:DestroyAllChildren()

    local charges = self.graft.userdata.max_charges or self.graft_def.charges or 0
    if charges and charges > 0 then
        self.charges_label:SetText( loc.format( LOC"UI.BLASTER_SCREEN.CHARGES_TITLE", charges ) )
        self.charges_text:SetText( LOC"UI.BLASTER_SCREEN.CHARGES_TEXT" )
        for k = 1, charges do
            self.charges_icons:AddChild( Widget.SingleCharge() )
                :Bloom( 0.1 )
                :SetCharged( k==1, false )
        end
        self.charges_icons:StackChildrenRow( -8, 0 )
        self.charges_label:LayoutBounds( "after", "center", self.charges_icons ):Offset( 8, 0 )
        self.charges_text:LayoutBounds( "left", "below", self.charges_icons ):Offset( 8, -10 )
    else
        -- No charges
        self.charges_label:SetText( LOC"UI.BLASTER_SCREEN.CHARGES_TITLE_EMPTY" )
        self.charges_text:SetText( "" )
    end
    local w, h = self.contents:GetSize()
    self.bg:SetSize( DETAILS_W, h )
    self.contents:LayoutBounds( "left", "center", self.bg ):Offset( 20, 0 )
end

------------------------------------------------------------------------------------------
-- Shows blaster surge

local BlasterOverchargeDisplay = class( "Widget.BlasterOverchargeDisplay", Widget )

local ICON_SIZE = 80

function BlasterOverchargeDisplay:init()
    self._base.init( self )

    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame.tex" ) ) )
        :SetTintColour( UICOLOURS.SURGE )
        :SetBloom( 0.2 )
    self.contents = self:AddChild( Widget() )
    self.charges_label = self.contents:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetAutoSize( DETAILS_W )
        :SetWordWrap( true )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.SURGE )
        :Bloom( 0.15 )
    self.charges_text = self.contents:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ))
        :SetAutoSize( DETAILS_W )
        :SetWordWrap( true )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
    self.overcharge_icon = self.contents:AddChild( Widget.SurgeCounter() )
        :SetMuted( true )
        :UpdateSurge( 3 )

    self.overcharge_icon.label:Hide()
end

function BlasterOverchargeDisplay:SetGraft( graft )
    self.graft = graft
    self.graft_def = self.graft and self.graft:GetDef()

    local overcharge_limit = self.graft_def.overcharge_limit
    if overcharge_limit then
        self.charges_label:SetText( loc.format( LOC"UI.BLASTER_SCREEN.OVERCHARGE_TITLE", overcharge_limit ) )
    else
        -- No limit
        self.charges_label:SetText( LOC"UI.BLASTER_SCREEN.OVERCHARGE_TITLE_NOLIMIT" )
    end

    self.charges_text:SetText( LOC"UI.BLASTER_SCREEN.OVERCHARGE_TEXT" )
    self.charges_label:LayoutBounds( "after", "center", self.overcharge_icon ):Offset( 14, -3 )
    self.charges_text:LayoutBounds( "left", "below", self.overcharge_icon ):Offset( 8, -4 )
    local w, h = self.contents:GetSize()
    self.bg:SetSize( DETAILS_W, h-4 )
    self.contents:LayoutBounds( "left", "center", self.bg ):Offset( 20, 4 )
end