require "ui/widgets/widepanel"

local CardShopScreen = class("Screen.CardShopScreen", Widget.Screen)

local PANEL_H = 640
local CARD_SCALE = 0.70
local CARD_SCALE_HOVER = 0.75

CardShopScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            local focus_card = UIHelpers.GetFocusCard(self)
            return focus_card and focus_card.purchaseable and not self.show_reticule
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.BUY", Controls.Digital.MENU_ACCEPT))
        end,
        fn = function(self)
            local focus_card = UIHelpers.GetFocusCard(self)
            if focus_card then
                self:OnClickContinue(focus_card)
            end
            return true
        end
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            return not self.show_reticule
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end,
        fn = function(self)
            self.close_button:Click()    
            return true
        end,
    }
}

function CardShopScreen:init( agent, on_done, free_items )
	Widget.Screen.init( self )

	self.is_popup = true
    self:SetupUnderlay( true )
	self:SetAnchors("center", "center")

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Backing panel
    self.panel = self:AddChild( Widget.WidePanel( 650 ) )
        :SetIcon( engine.asset.Texture( "UI/ic_shop_icon.tex" ) )
        :SetTopTitlesMargin( SPACING.M1 )

    -- Titles
    self.panel:SetTitles( LOC"UI.SHOP_SCREEN.CARD_SHOP_TITLE", loc.format( LOC"UI.SHOP_SCREEN.CARD_SHOP_SUBTITLE", agent:GetName() ) )

    self.on_done = on_done
    self.select_then_continue_mode = false -- If true, you'll have to click a graft, then click the continue button

    -- Card container
    self.free_items = 0
    self.card_widgets = {}
    self.card_container = self:AddChild( Widget() )
    self.price_container = self:AddChild( Widget() )
    self.shadow_container = self:AddChild( Widget() )
        :SetAnchors( "center", "bottom" )

    self.upgrade_info_label = self:AddChild( Widget.Label( "body", 22 ) )
        :SetGlyphColour( UICOLOURS.OBJECTIVE )
        :SetAutoSize(800)
        :SetWordWrap(true)
        :SetText( LOC"UI.DRAFT_SCREEN.UPGRADES_PREVIEW_INFO" )
        :SetTintAlpha( 0.9 )
        :Bloom( 0.08 )

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ))
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.CARDSHOP.BUY_BTN_BUY", function() self:OnClickContinue(self.selected_card) end ) )
        :SetIcon( global_images.money )
        :LayoutBounds( "right", "above", -60, 60 )
        :Hide()
        :SetToolTip( LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_TOOLTIP" )

    -- Add our cards
    if free_items ~= nil then
        self:GiveFreeItems( free_items )
    end

    self:SetShop( agent:GetAspect("cardshop") )

    self:DoLayout()
end

function CardShopScreen:OnOpen()
    CardShopScreen._base.OnOpen( self )

    self.screen_loop = AUDIO:CreateEventInstance("event:/ui/store/menu_open")
    self.screen_loop:Start()
end

function CardShopScreen:OnClickClose()
    self:FadeOut()
end

function CardShopScreen:OnClose()
    CardShopScreen._base.OnClose( self )

    if self.on_done then
        self.on_done()
    end

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil
    AUDIO:PlayEvent( "event:/ui/store/menu_close" )
end

function CardShopScreen:OnGainTop()
    CardShopScreen._base.OnGainTop(self)
    self.card_container:SetTintAlpha(0)

    self.panel:FadeIn( function()
        self.card_container:AlphaTo( 1, 0.1 )
    end )

    -- If a card is selected, unselect it
    self.selected_card = nil
    -- Unselect options
    for k, widget in ipairs( self.card_container.children ) do
        widget.selected = false
        self:UpdateCardImageState( widget )
    end
end

function CardShopScreen:RefreshRobotOptions()
    self:ClearRobotOptions()
    self:AddRobotOption(self.close_button)
    for k,v in ipairs(self.card_container.children) do  
        if v.purchaseable then
            self:AddRobotOption(v)
        end
    end  
end


function CardShopScreen:OnControlModeChange( cm, device_id )
    CardShopScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.select_then_continue_mode = engine.inst:GetSetting("gameplay", "confirm_select", true)
        self.close_button:Show()
        self.bottom_right:SetShown( self.select_then_continue_mode )
        self.upgrade_info_label:Show()
        -- Unselect options
        for k, widget in ipairs( self.card_container.children ) do
            widget.selected = false
        end
        self.selected_card = nil
    end
    if cm == CONTROL_MODE.TOUCH then
        self.select_then_continue_mode = true
        self.close_button:Show()
        self.bottom_right:SetShown( self.select_then_continue_mode )
        self.upgrade_info_label:Hide()
        -- Unselect options
        for k, widget in ipairs( self.card_container.children ) do
            widget.selected = false
        end
        self.selected_card = nil
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.select_then_continue_mode = engine.inst:GetSetting("gameplay", "confirm_select", true)
        self.close_button:Hide()
        self.bottom_right:Hide()
        self.upgrade_info_label:Hide()
        -- Unselect options
        for k, widget in ipairs( self.card_container.children ) do
            widget.selected = false
        end
        self.selected_card = nil
    end

    self:DoLayout()
end

function CardShopScreen:OnScreenModeChange( sm )
    CardShopScreen._base.OnScreenModeChange( self, sm )

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )

    self.upgrade_info_label:SetLayoutScale( LAYOUT_SCALE[sm] )
    
    local new_panel_h, new_cardcontainer_scale
    if sm == SCREEN_MODE.MONITOR then
        new_panel_h = PANEL_H
        new_cardcontainer_scale = 1
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        new_panel_h = PANEL_H + 130
        new_cardcontainer_scale = 1.2
    end
    self.card_container:SetLayoutScale( new_cardcontainer_scale )

    -- Resize prices
    for k, widget in ipairs( self.price_container.children ) do
        widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    end
    -- Resize container panel
    self.panel:SetHeight( new_panel_h )

    self:DoLayout()
end

function CardShopScreen:GetDefaultFocus()
    if self.last_focus and self.last_focus:GetScreen() == self then
        return self.last_focus
    end
    return self.card_widgets and #self.card_widgets > 0 and self.card_widgets[1]
end

function CardShopScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

function CardShopScreen:GiveFreeItems( items )
    self.free_items = self.free_items + items
    self:DoLayout()
end

function CardShopScreen:SetShop( card_shop )
    self.card_shop = card_shop
    self.has_cards_with_upgrades = false
    
    self:Refresh()

    self:DoLayout()

    return self
end

function CardShopScreen:Refresh()
    -- Go through the cards the owner is selling, mark sold cards as sold, etc.
    for i, item in ipairs(self.card_shop:GetStock()) do
        local widget = self:RefreshCardWidget( i, item )
    
        -- Check if there are upgrades
        if widget and widget.card and widget.card:HasUpgrades() then self.has_cards_with_upgrades = true end
    end
    self:RefreshRobotOptions()
    return self
end

function CardShopScreen:RefreshCardWidget( idx, data )
    local caravan = TheGame:GetGameState():GetCaravan()

    local widget = self.card_widgets[idx]
    if not widget then
        StatUtil.PresentedAtShop(data.card_id)
        local card = data.card
        if is_instance( card, Negotiation.Card ) then
            widget = Widget.NegotiationCard( card )
        else
            widget = Widget.BattleCard( card )
        end

        widget:SetScale( CARD_SCALE )
        self.card_container:AddChild( widget )
        self.card_widgets[idx] = widget
        widget.idx = idx
        widget:ShowToolTipOnFocus()
        widget:SetAcceptHintShown( false )

        -- Add sub card info labels
        widget.show_class_label = true
        widget.sub_card_info = widget:AddChild( Widget.SubCardInfo( widget ) )
        widget.sub_card_info:SetHiddenBoundingBox( true )
        widget.sub_card_info:LayoutBounds( "center", "below", widget.bg ):Offset( 0, -8 )

        -- Handle events
        widget.OnGainFocus = function() widget._base.OnGainFocus(widget) self:UpdateCardImageState( widget, card.id ) self.last_focus = widget end
        widget.OnLoseFocus = function() widget._base.OnLoseFocus(widget) self:UpdateCardImageState( widget, card.id ) end
        widget.OnGainHover = function() widget._base.OnGainHover(widget) self:UpdateCardImageState( widget, card.id ) end
        widget.OnLoseHover = function() widget._base.OnLoseHover(widget) self:UpdateCardImageState( widget, card.id ) end

        -- Handle clicks
        widget:SetOnClick( function()
            self:OnCardSelect( widget )
        end )

        -- Add the price
        widget.price = self.price_container:AddChild( Widget.MoneyIndicator( 40, 44, false ) )
            :SetSpacing( 6 )
            :LayoutBounds( "center", "below", widget )
            :Offset( 0, -40 )

        widget:SetOutline( widget.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT, true )

        -- Add a shadow
        widget.shadow = self.shadow_container:AddChild( Widget.Image( engine.asset.Texture("UI/negotiation_iconglow.tex") ) )
            :SetSize( 360, 70 )
            :SetTintColour( card.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
            :SetTintAlpha( 0.8 )

    end

    local price, reasons, old_price = self.card_shop:CalculatePrice( idx )
    local tt = LOC"UI.MONEY_INDICATOR.TOOLTIP"
    if reasons and #reasons > 0 then
        tt = tt .. "\n\n" .. table.concat(reasons, "\n")
    end
    widget.price:SetToolTip( tt )

    if self.free_items > 0 and not data.sold then
        widget.price:FormatMoney(LOC"UI.CARDSHOP.FREE")
        widget.purchaseable = true
        widget.price_value = 0

    elseif data.sold then
        widget:SetTintAlpha( 0.20 )
        widget.price:FormatMoney(LOC"UI.CARDSHOP.SOLD")
        widget.purchaseable = false
        widget:SetOnRightClickFn( nil )
        widget:SetOnClick( nil )

    elseif caravan:GetMoney() >= price then
        local old_price_str = old_price and old_price ~= price and tostring(old_price)
        widget.price:FormatMoney( loc.format("<#MONEY>{1}</>", price), old_price_str )
        widget.price_value = price
        widget.purchaseable = true
    else
        local old_price_str = old_price and old_price ~= price and tostring(old_price)
        widget.price:FormatMoney( loc.format("<#ENEMY_COLOUR>{1}</>", price), old_price_str )
        widget.price_value = price
        widget.purchaseable = false
    end

    return widget
end

function CardShopScreen:DoLayout()
    -- Layout things!
    self.card_container:StackChildrenRow( 0 )
    self.upgrade_info_label:LayoutBounds( "center", "bottom", self.panel )
        :Offset( 0, 30 )
    self.card_container:LayoutBounds( "center", "center", self.panel )
        :Offset( 0, self.upgrade_info_label:IsShown() and 5 or -15 )
    for k,v in ipairs(self.card_container.children) do
        local card_x, card_y = v:GetPos()
        v.shadow:SetPos( card_x, 0 )
        -- Get widget scale
        local scale = v:GetScale()
        -- Reset the scale
        v:SetScale(CARD_SCALE)
        -- Align price
        v.price:LayoutBounds( "center", "below", v ):Offset( 0, -45 )
        -- Set the scale back
        v:SetScale(scale)
    end
    local grid_w, grid_h = self.grid:GetSize()
    self.shadow_container:LayoutBounds( "center", "above", 0, grid_h*0.2 )
end

function CardShopScreen:UpdateCardImageState( widget )

    if widget.enabled and widget.selected then
        widget:SetOutline( UICOLOURS.WHITE, true )
        widget:ScaleTo( nil, CARD_SCALE_HOVER, 0.2, easing.outQuad )
        if widget.shadow then
            local r, g, b, a = HexColour( UICOLOURS.WHITE )
            widget.shadow:ColourFromCurrentTo( r, g, b, a, 0.2 )
            widget.shadow:ScaleTo( nil, 1.4, 0.2 )
        end
    elseif widget.enabled and ( widget.hover or widget.focus ) then
        widget:SetOutline( UICOLOURS.FOCUS, true )
        widget:ScaleTo( nil, CARD_SCALE_HOVER, 0.2, easing.outQuad )
        if widget.shadow then
            local r, g, b, a = HexColour( UICOLOURS.FOCUS )
            widget.shadow:ColourFromCurrentTo( r, g, b, a, 0.2 )
            widget.shadow:ScaleTo( nil, 1.4, 0.2 )
        end
    else
        widget:SetOutline( widget.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT, true )
        widget:ScaleTo( nil, CARD_SCALE, 0.2, easing.outQuad )
        if widget.shadow then
            local r, g, b, a = HexColour( widget.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
            a = 0.8
            widget.shadow:ColourFromCurrentTo( r, g, b, a, 0.2 )
            widget.shadow:ScaleTo( nil, 1, 0.2 )
        end
    end
end

function CardShopScreen:OnCardSelect( widget )
    AUDIO:PlayEvent(SoundEvents.button_click)
    if self.card_shop:CanBuyCard( widget.idx ) or self.free_items > 0 then

        if self.select_then_continue_mode then
            -- We're in select-then-continue mode, and the selected graft isn't the clicked one

            if self:GetFE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD then
                self.selected_card = widget
    
                -- Unselect options
                for k, widget in ipairs( self.card_container.children ) do
                    widget.selected = ( widget == self.selected_card )
                    self:UpdateCardImageState( widget )
                end
                -- And show the confirm button
                self.button_continue:Show()
                self:RefreshRobotOptions()

            else
                local screen = Screen.YesNoPopup( LOC"UI.CARDSHOP.CONFIRM_TITLE", loc.format( LOC"UI.CARDSHOP.CONFIRM_BUY", widget.card, widget.price_value ) )
                    :SetFn( function(v)
                        if v == Screen.YesNoPopup.YES then 
                            self.selected_card = widget
                            self:OnClickContinue()                           
                        end 
                    end )

                local card_widget
                if is_instance( widget.card, Negotiation.Card ) then
                    card_widget = Widget.NegotiationCard( widget.card )
                else
                    card_widget = Widget.BattleCard( widget.card )
                end
                local sm = self:GetFE():GetScreenMode()
                screen:AddChild( card_widget ):SetScale( 0.8 ):SetLayoutScale( LAYOUT_SCALE[ sm ] ):LayoutBounds( "before", "center", screen:GetTextWidget() ):Offset( -30, 0 )

                TheGame:FE():PushScreen( screen )
            end

        elseif widget then
            -- Not select-then-continue mode, so let's just buy this one straight up!
            self.selected_card = widget
            self:OnClickContinue()
        end

    else

        self.selected_card = nil
        -- Unselect options
        for k, widget in ipairs( self.card_container.children ) do
            widget.selected = false
            self:UpdateCardImageState( widget )
        end
        -- Hide the confirm button
        self.button_continue:Hide()
        self:RefreshRobotOptions()

    end
end

function CardShopScreen:OnClickContinue( card_widget )
    card_widget = card_widget or self.selected_card
    if card_widget then
        local on_success = function()
            if self.free_items then
                self.free_items = self.free_items - 1
            end
            AUDIO:PlayEvent(SoundEvents.buy)
            card_widget:AlphaTo( 0.20, 0.5, easing.outQuad )
            self:Refresh()
            self:DoLayout()
            
            local card_id = self.card_shop:GetStock()[card_widget.idx].card_id
            if card_id then
                local card = Content.GetBattleCard(card_id) and Battle.Card(card_id) or Negotiation.Card(card_id)
                local x, y = card_widget:GetWorldPos()
                UIHelpers.ShowGainCards({card}, x, y, nil, 1/CARD_SCALE_HOVER)
            end
        end
        -- Buy card
        self.card_shop:DoBuyCard( card_widget.idx, on_success, self.free_items > 0 )
        -- Unselect card
        self.selected_card = nil
        -- Unselect other cards too
        for k, widget in ipairs( self.card_container.children ) do
            widget.selected = false
            self:UpdateCardImageState( widget )
        end
        -- Hide the confirm button
        self.button_continue:Hide()
        self:RefreshRobotOptions()
    end
end

function CardShopScreen:PreviewUpgrade( upgrade_card, agent )
    if not self.coro then
        self.coro = self:StartCoroutine( self.PreviewUpgradeCoro, self, upgrade_card, agent )
    end
end

function CardShopScreen:PreviewUpgradeCoro( upgrade_card, agent )
    
    local preview_only = true
    local popup = Screen.UpgradeCardPopup( upgrade_card, agent, preview_only )
    TheGame:FE():PushScreen( popup )

    while TheGame:FE():FindScreen( popup ) do
        coroutine.yield()
    end

    self.coro = nil
end



local ReticuleMixin = require "ui/screens/ReticuleMixin"
ReticuleMixin:Mixin( CardShopScreen )
