require "ui/widgets/screen"
require "ui/widgets/cardwidget"
require "ui/widgets/widepanel"

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.EVENT

local CARD_SCALE = 0.8
local CARD_SCALE_HOVER = 0.84

-------------------------------------------------------------------------------
local ChooseFlourishPopup = class( "Screen.ChooseFlourishPopup", Widget.Screen )

ChooseFlourishPopup.CONTROL_MAP = 
{
    {
        --this one's just a hint, because the generic card widget handles the pick
        control = Controls.Digital.MENU_ACCEPT,
        hint = function(self, left, right)
            table.insert(right, loc.format ( LOC"UI.CONTROLS.ACCEPT", Controls.Digital.MENU_ACCEPT))
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self) 
            self:Close()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL) )
        end,
    },
}

function ChooseFlourishPopup:init( flourishes, can_play_card, card_widget_class, minigame, on_close )
    ChooseFlourishPopup._base.init( self )
    
    self.is_popup = true
    self:SetAnchors( "center", "center" )
    self.can_play_card = can_play_card
    self.card_widget_class = card_widget_class
    self.minigame = minigame
    self.on_close = on_close
    self.owner = TheGame:GetGameState():GetPlayerAgent()

	-- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Background darkening
    self.bg = self:AddChild( Widget.SolidBox( nil, nil, 0x1D0D1EA0 ) )
        :SetAnchors( "fill", "fill" )
        :SetSize( RES_X, RES_Y )
    
    self.header_gradient = self:AddChild( Widget.Panel( engine.asset.Texture ("UI/motd_gradient.tex") ) )
        :SetTintColour( 0x421D44ff )
        :SetAnchors( "fill", "center" )
        :SetSize( RES_X, 500 )

    -- Panel
    self.widepanel = self:AddChild( Widget.WidePanel( 400, UICOLOURS.FLOURISH ) )
        :SetPos( 0, 40 )
        :PrepareAnimation()

    -- Cards panel
	self.cards_container_width = self.screen_w * 0.6
	self.cards_container_height = 600
	self.card_container = self:AddChild( Widget.CardListBlock( self.cards_container_width, self.cards_container_height ) )
        :HideBackground()
        :DisableCardRotation()
        :SetCardScale( CARD_SCALE )
        :SetUpdateCardImageFn( function(...) self:UpdateCardImageState(...) end )

	-- Above the panel
    self.header = self:AddChild( Widget() )
    	:SetAnchors( "left", "top" )
    self.flourish_widget = self.header:AddChild( Widget.FlourishProgressWidget() )
        :SetToolTipClass( nil )
        :SetToolTip( nil )
        :SetScale( 1.1 )
        :SetOnClickFn( nil )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :SetAutoSize( RES_X*0.6 )
        :SetWordWrap( true )
        :Bloom( 0.1 )
        :LayoutBounds( "left", "top", 0, -60 )
    self.subtitle = self.header:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT ) )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :LeftAlign()
        :SetAutoSize( RES_X*0.6 )
        :SetWordWrap( true )
        :Bloom( 0.15 )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )
    self.header_divider = self.header:AddChild( Widget.SolidBox( 2, 100, UICOLOURS.TOOLTIP_TEXT ) )
        :Bloom( 0.15 )
    self.criteria = self.header:AddChild( Widget.Label( "title", 22 ) )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :SetText( loc.format( LOC"UI.FLOURISH_POPUP.BATTLE_CRITERIA", 0 ) )
        :OverrideLineHeight( 18 )
        :LeftAlign()
        :SetAutoSize( RES_X*0.6 )
        :SetWordWrap( true )
        :Bloom( 0.15 )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )
    -- Set text
    if is_class( self.card_widget_class, Widget.BattleCard ) then
    	-- Set battle titles
    	self.title:SetText( loc.format( LOC"UI.FLOURISH_POPUP.TITLE", self.owner:GetName() ) )
    	self.subtitle:SetText( loc.format( LOC"UI.FLOURISH_POPUP.BATTLE_SUBTITLE", self.owner:GetName() ) )
        -- Get agent's flourish graft
        local graft_owner = self.owner and self.owner:GetAspect( "graft_owner" )
        self.flourish_graft = graft_owner and graft_owner:GetGraft( "flourish_tracker" )
        if self.flourish_graft then
	        -- Get point multipliers from the graft
	        local card_played = self.flourish_graft.userdata.battle_gains["CARD_PLAY"]
	        local damage_dealt = self.flourish_graft.userdata.battle_gains["DAMAGE_DEALT"]
	        local damage_received = self.flourish_graft.userdata.battle_gains["DEFENDED"]
	        local enemy_killed = self.flourish_graft.userdata.battle_gains["KILL"]
	    	self.criteria:SetText( loc.format( LOC"UI.FLOURISH_POPUP.BATTLE_CRITERIA", card_played, damage_dealt, damage_received, enemy_killed ) )
	    	-- Set the current flourish
			self.max_flourish = self.flourish_graft.userdata.battle_max
			self.current_flourish = self.flourish_graft.userdata.battle_flourish
			-- Update meter
			self.flourish_widget:SetBattle( self.owner, self.minigame )
			-- Start listening for changes
			self.minigame:ListenForEvent( BATTLE_EVENT.FLOURISH_CHANGED, self, self.OnBattleEvent )
    	end
    else
		-- Set negotiation titles
    	self.title:SetText( loc.format( LOC"UI.FLOURISH_POPUP.TITLE", self.owner:GetName() ) )
    	self.subtitle:SetText( loc.format( LOC"UI.FLOURISH_POPUP.NEGOTIATION_SUBTITLE", self.owner:GetName() ) )
        -- Get agent's flourish graft
        local graft_owner = self.owner and self.owner:GetAspect( "graft_owner" )
        self.flourish_graft = graft_owner and graft_owner:GetGraft( "flourish_tracker" )
        if self.flourish_graft then
	        -- Get point multipliers from the graft
	        local card_played = self.flourish_graft.userdata.negotiation_gains["CARD_PLAY"]
	        local damage_dealt = self.flourish_graft.userdata.negotiation_gains["DAMAGE_DEALT"]
	        local damage_mitigated = self.flourish_graft.userdata.negotiation_gains["MITIGATED"]
	        local argument_destroyed = self.flourish_graft.userdata.negotiation_gains["DESTROY"]
	    	self.criteria:SetText( loc.format( LOC"UI.FLOURISH_POPUP.NEGOTIATION_CRITERIA", card_played, damage_dealt, damage_mitigated, argument_destroyed ) )
	    	-- Set the current flourish
			self.max_flourish = self.flourish_graft.userdata.negotiation_max
			self.current_flourish = self.flourish_graft.userdata.negotiation_flourish
			-- Update meter
			self.flourish_widget:SetNegotiation( self.owner, self.minigame )
			-- Start listening for changes
			self.minigame:ListenForEvent( EVENT.FLOURISH_CHANGED, self, self.OnNegotiationEvent )
    	end
    end

    -- Check if it's possible to activate a card
    self.flourish_active = self.current_flourish >= self.max_flourish
    self.selected_card = nil

    -- Show particles
    self:RefreshParticleEmitter()

	self.select_then_continue_mode = false -- If true, you'll have to click a choice, then click the continue button

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:Close() end ) )
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
    self.close_button.shadow = self.close_button:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( 0x021A18ff )
        :SetTintAlpha( 1 )
        :SetHiddenBoundingBox( true )
        :SetSize( 600, 250 )
        :SendToBack()

    -- Continue button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.FLOURISH_POPUP.CONFIRM_BTN", function() self:OnClickContinue(self.selected_card) end ) )
        :SetIcon( global_images.accept )
        :LayoutBounds( "right", "above", -60, 60 )
        :Hide()
        :SetToolTip( LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_TOOLTIP" )
    self.button_continue.shadow = self.button_continue:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( 0x021A18ff )
        :SetTintAlpha( 1 )
        :SetHiddenBoundingBox( true )
        :SetSize( 600, 250 )
        :SendToBack()

    
    self:AddFlourishWidgets( flourishes )
    self:Layout()

    self:PrepareAnimation()
end

function ChooseFlourishPopup:OnRemoved()
    self.minigame:RemoveListener( self )
end

function ChooseFlourishPopup:AddFlourishWidgets( flourishes )
    self.flourishes = flourishes
    self.card_container:SetCards( self.flourishes )

    -- Go through cards in the container and set some properties
    local cards = self.card_container:GetFocusables()
    self.card_w = 0
    self.card_h = 0
    for i, card_container in ipairs( cards ) do
    	-- Get actual card size
    	self.card_w, self.card_h = card_container.card_w:GetSize()

    	-- Check if the card can be played
    	if self.can_play_card and not self.can_play_card( card_container.card_w.card ) then
            card_container.card_w:Disable()
            card_container.card_w:SetSaturation( 0 ) 
        else
            card_container.card_w:Enable()
            card_container.card_w:SetSaturation( 1 ) 
        end

        -- Set tooltip
        card_container:SetToolTip( { CARD = card_container.card_w.card, FEATURES = card_container.card_w.card_features } )
        card_container:SetToolTipClass( Widget.FullCardPanel )

        -- Set right click
        card_container:SetOnRightClickFn( function() card_container.card_w:ShowCardInfoPopup() end )

        -- Set gamepad hint
        card_container.CONTROL_MAP = {
            {
                control = Controls.Digital.GAMEPAD_RT,
                fn = function(self)
                    card_container.card_w:ShowCardInfoPopup()
                    return true
                end,
                hint = function(self, left, right)
                    table.insert(right, loc.format(LOC"UI.CONTROLS.CARD_INFO", Controls.Digital.GAMEPAD_RT))
                end
            },
        }

        -- Set minigame color
        card_container.card_w:SetOutline( card_container.card_w.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT, true )

    end
	return self
end

function ChooseFlourishPopup:OnUpdate(dt)
    ChooseFlourishPopup._base.OnUpdate(self, dt)

    -- self.select_timer = self.select_timer - dt

    -- if TheGame:GetLocalSettings().ROBOTICS and self.select_timer < -1.0 then
    --     local cards = {}
    --     for i, w in ipairs( self.card_widgets ) do
    --         table.insert( cards, w.card )
    --     end
    --     table.arrayadd( self.flourishes, table.multipick( cards, self.min_count ))
    --     self:Close()
    -- end

    if self.shaker then
        if self.shaker:IsDone() then
            self.shakex, self.shakey = nil, nil
            self.shaker = nil
        else
            self.shakex, self.shakey = self.shaker:UpdateShake( dt )
        end
        self.header:SetPos( self.header_x + (self.shakex or 0), self.header_y + (self.shakey or 0) )
        self.bottom_left:SetPos( self.bottom_left_x + (self.shakex or 0), self.bottom_left_y + (self.shakey or 0) )
        self.bottom_right:SetPos( self.bottom_right_x + (self.shakex or 0), self.bottom_right_y + (self.shakey or 0) )
        self:SetPos( self.shakex or 0, self.shakey or 0 )
    end

end

function ChooseFlourishPopup:OnBattleEvent( eventname, graft, flourish_level, prev_flourish )
    -- Bail if a card has been chosen already
    if self.card_chosen then return self end

    -- Check if it's possible to activate a card
    self.flourish_active = flourish_level >= self.max_flourish
    -- Show particles
    self:RefreshParticleEmitter()
    -- If not, unselect whatever card is selected
    self.selected_card = self.flourish_active and self.selected_card or nil
    -- Refresh all cards' visual state
    for k, card_container in ipairs( self.card_container:GetFocusables() ) do
        self:UpdateCardImageState( card_container.card_w, card_container )
    end
end

function ChooseFlourishPopup:OnNegotiationEvent( eventname, graft, flourish_level, prev_flourish )
    -- Bail if a card has been chosen already
    if self.card_chosen then return self end

    -- Check if it's possible to activate a card
    self.flourish_active = flourish_level >= self.max_flourish
    -- Show particles
    self:RefreshParticleEmitter()
    -- If not, unselect whatever card is selected
    self.selected_card = self.flourish_active and self.selected_card or nil
    -- Refresh all cards' visual state
    for k, card_container in ipairs( self.card_container:GetFocusables() ) do
        self:UpdateCardImageState( card_container.card_w, card_container )
    end
end

function ChooseFlourishPopup:RefreshParticleEmitter()
    if self.flourish_active then
        -- Show particles
        if not self.particles_top then
            self.particles_top = self:AddChild( Widget.ParticleSystem() )
                :LoadParams("flourish_popup_edge")
                :SendBehind( self.widepanel )
        end
        self.particles_top:Restart()
        if not self.particles_bottom then
            self.particles_bottom = self:AddChild( Widget.ParticleSystem() )
                :LoadParams("flourish_popup_edge")
                :SendBehind( self.widepanel )
        end
        self.particles_bottom:Restart()
    else
        -- Hide particles
        if self.particles_top then self.particles_top:Stop() end
        if self.particles_bottom then self.particles_bottom:Stop() end
    end
    return self
end

function ChooseFlourishPopup:PrepareAnimation()

    -- Hide header contents
    self.bg:SetTintAlpha( 0 )
    self.header:SetTintAlpha( 0 )
    self.header_gradient:SetTintAlpha( 0 )

	-- Hide cards and prepare their animations
	for k, card_container in ipairs( self.card_container:GetFocusables() ) do

        -- Fade card out preemptively
        card_container:SetTintAlpha( 0 )
            :SetScale( 1.1 )

        -- Setup the circular animation around the card
        local animname = "anim/UI_negotiation_square_fx.zip"
        local anim_scale = 1.3
        local anim = Widget.Anim()
            :SetHiddenBoundingBox(true)
            :SetScale( anim_scale * (math.random()<0.5 and -1 or 1), anim_scale * (math.random()<0.5 and -1 or 1) )
            :SetTintColour( UICOLOURS.FLOURISH )
            :SetBloom( 0.2 )
        anim:AddAnims( GetAnim( animname ) )
        anim:AddOverrideFile( animname )
        card_container.flourish_fx_card_drop = card_container:AddChild( anim )
            :SendToBack()
            :Hide()

        -- Setup the energy pulse animation
        local animname = "anim/card_flourish_fx.zip"
        local anim = Widget.Anim()
            :SetHiddenBoundingBox(true)
            :SetScale( 0.5 )
            -- :SetTintColour( UICOLOURS.FLOURISH )
            :SetBloom( 0.2 )
        anim:AddAnims( GetAnim( animname ) )
        anim:AddOverrideFile( animname )
        card_container.flourish_fx_energy_container = card_container:AddChild( Widget() )
            :SendToBack()
            :Hide()
            :SetTintAlpha( 0 )
        card_container.flourish_fx_energy = card_container.flourish_fx_energy_container:AddChild( anim )

    end
end

function ChooseFlourishPopup:OnGainTop()
    ChooseFlourishPopup._base.OnGainTop( self )

    -- Hide top-left bar
    local main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if main_overlay then
        local show_pause_btn = true
        main_overlay:FadeOutTopLeft( 0.1 )
            :FadeOutTopRight( 0.1, show_pause_btn )
    end

end

function ChooseFlourishPopup:OnLoseTop()
    ChooseFlourishPopup._base.OnLoseTop( self )

    -- Show top-left bar
    local main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if main_overlay then
        main_overlay:FadeInTopLeft( 0.1 )
            :FadeInTopRight( 0.1 )
    end

end

function ChooseFlourishPopup:OnOpen()
    ChooseFlourishPopup._base.OnOpen( self )
    self.select_timer = 0.1

    -- Animate stuff in
    self.widepanel:FadeIn( function()

        -- Create updater
        local timing = 0.08
        if self.card_widget_class == Widget.BattleCard then
            timing = timing / engine.inst:GetSetting( "gameplay", "battle_speed", 1.0 )
        else
            timing = timing / engine.inst:GetSetting( "gameplay", "negotiation_speed", 1.0 )
        end
        local intro = Updater.Parallel()

        -- Background!
        intro:Add( Updater.Do( function() self.bg:AlphaTo( 1, timing*10 ) end ) )

        -- Header!
        local header_x, header_y = self.header:GetPos()
        local header_gradient_x, header_gradient_y = self.header_gradient:GetPos()
        intro:Add( Updater.Do( function() self.header:AlphaTo( 1, timing*2 ) end ) )
        intro:Add( Updater.Ease( function(v) self.header:SetPos(header_x, v) end, header_y+30, header_y, timing, easing.outQuad ) )
        intro:Add( Updater.Do( function() self.header_gradient:AlphaTo( 1, timing ) end ) )
        intro:Add( Updater.Ease( function(v) self.header_gradient:SetPos(header_gradient_x, v) end, header_gradient_y-300, header_gradient_y, timing, easing.outQuad ) )
        intro:Add( Updater.Wait( timing ) )

	    -- Cards!
        local intro_cards = Updater.Series()
        intro_cards:Add( intro )
        intro_cards:Add( Updater.Wait( timing*3 ) )

        local num_cards = #self.card_container:GetFocusables()
	    for i, card_container in ipairs( self.card_container:GetFocusables() ) do
	    	local pos_x, pos_y = card_container.card_w:GetPos()

            -- Create updater for this card
            card_container.updater_in = Updater.Series{

                -- Fade card in
                Updater.Parallel{
                    Updater.Ease( function(v) card_container.card_w:SetPos(pos_x,v) end, pos_y+40, pos_y, timing*1.5, easing.outQuad ),
                    Updater.Ease( function(v) card_container:SetTintAlpha(v) end, 0, 1, timing*1.5, easing.outQuad ),
                },

                -- Scale card down
                Updater.Ease( function(v) card_container:SetScale(v) end, 1.1, 0.85, timing, easing.inOutQuad ),
                
                -- Screen shake
                Updater.Do( function()
                    self:ScreenShake()
                    AUDIO:PlayEvent( CARD_TYPE_SOUNDS[ CARD_TYPE.FLOURISH ] )
                end ),

                -- Show the explosion animation
                Updater.Do( function()
                    card_container.flourish_fx_card_drop:Play( "card_drop", false )
                        :Show()
                end ),

                -- Scale card back up
                Updater.Ease( function(v) card_container:SetScale(v) end, 0.85, 1, timing/3, easing.inOutQuad ),
            }

            -- Start card updater within main updater
            intro_cards:Add( Updater.Do( function() card_container:RunUpdater( card_container.updater_in ) end ) )

            -- Introduce per-card delay
            local delay = 1-easing.outQuad( i, 0, num_cards, num_cards )/num_cards
            delay = delay*timing*3 + timing*3
            intro_cards:Add( Updater.Wait( delay ) )
	    end

    	self:RunUpdater( intro_cards )
    end )
end

function ChooseFlourishPopup:OnScreenModeChange( sm )
    ChooseFlourishPopup._base.OnScreenModeChange( self, sm )
    self:Layout()
end

function ChooseFlourishPopup:OnControlModeChange( cm, device_id )
    ChooseFlourishPopup._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
    	self.select_then_continue_mode = false or engine.inst:GetSetting("gameplay", "confirm_select", true)
        self.bottom_left:Show()
        self.bottom_right:Show()
        -- Unselect options
    	self.selected_card = nil
        for k, card_container in ipairs( self.card_container:GetFocusables() ) do
            self:UpdateCardImageState( card_container.card_w, card_container )
        end
    end
    if cm == CONTROL_MODE.TOUCH then
        self.select_then_continue_mode = true
        self.bottom_left:Show()
        self.bottom_right:SetShown( self.select_then_continue_mode )
        -- Unselect options
        self.selected_card = nil
        for k, card_container in ipairs( self.card_container:GetFocusables() ) do
            self:UpdateCardImageState( card_container.card_w, card_container )
        end
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.select_then_continue_mode = false
        self.bottom_left:Hide()
        self.bottom_right:Hide()
        -- Unselect options
        self.selected_card = nil
        for k, card_container in ipairs( self.card_container:GetFocusables() ) do
            self:UpdateCardImageState( card_container.card_w, card_container )
        end
    end
end

function ChooseFlourishPopup:GetDefaultFocus()
    return self.card_container:GetDefaultFocus()
end

function ChooseFlourishPopup:Close( card_widget )
    AUDIO:PlayEvent( "event:/ui/select_cards/close" )

    -- Restore default events
    if card_widget then
        card_widget.OnGainFocus = function() card_widget._base.OnGainFocus(card_widget) end
        card_widget.OnLoseFocus = function() card_widget._base.OnLoseFocus(card_widget) end
        card_widget.OnGainHover = function() card_widget._base.OnGainHover(card_widget) end
        card_widget.OnLoseHover = function() card_widget._base.OnLoseHover(card_widget) end
        card_widget:SetOnClick(nil)
    end

    -- Callback
    if self.on_close then
        self.on_close( self, card_widget )
    end

    -- Remove screen
    self:GetFE():PopScreen( self )
end

function ChooseFlourishPopup:ScreenShake()
    if TheGame:CamShakeEnabled() then
        local mag, speed, time = 5+(math.random()*3), 80+(math.random()*30), 0.20
        self.shaker = CamShake( mag, speed, time )
    end
end

function ChooseFlourishPopup:OnCardClick( card_widget, card_container )
    if card_container.updater_in and not card_container.updater_in:IsDone() then
        return
    end

    AUDIO:PlayEvent(SoundEvents.button_click)

	if not TheGame:GetLocalSettings().ROBOTICS and self.select_then_continue_mode then
        if self.selected_card == card_widget then
            -- This card was already selected, and was confirmed again. Just draft it
            self:OnClickContinue()
        else
            -- We're in select-then-continue mode, and the selected card isn't the clicked one
            -- Select this one
            self.selected_card = card_widget
            -- Unselect options
            for k, card_container in ipairs( self.card_container:GetFocusables() ) do
                self:UpdateCardImageState( card_container.card_w, card_container )
            end
            -- And show the confirm button
            self.button_continue:Show()
        end
    elseif card_widget then
        -- Not select-then-continue mode, so let's just draft this one straight up!
        self.selected_card = card_widget
        self:OnClickContinue()
    end
end

-- Picks the self.selected_card card
function ChooseFlourishPopup:OnClickContinue( selected_card_widget )
    selected_card_widget = selected_card_widget or self.selected_card
    if selected_card_widget then
        -- Prevent events from making the cards grey
        self.card_chosen = true
		-- Prevent clicks on other cards
        for k,w in ipairs( self.card_container:GetFocusables() ) do
            w:SetOnClickFn( nil )
            w:SetOnRightClickFn( nil )
            w:SetToolTip( nil )
            w:SetToolTipClass( nil )
            w.hover_check = false
        end
        self.button_continue:SetFn( nil )
        selected_card_widget:PlayPositionalEvent("event:/ui/select_cards/aquired_flourish")
        
        -- Pick card
	    self:Close( selected_card_widget )
    end
end

function ChooseFlourishPopup:Layout()
	-- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

	-- Scale contents
    local sm = TheGame:FE():GetScreenMode()
    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
	self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    
    -- Calculate content width
	local content_w = self.screen_w - 120
    local scaled_content_w = content_w/LAYOUT_SCALE[sm]

    -- Layout header
    local flourish_w, flourish_h = self.flourish_widget:GetScaledSize() -- Flourish icon width
    local header_w = content_w - 60 - SPACING.M1*4 -- Total header width (60 is for the right screen padding)
    local scaled_header_w = scaled_content_w - 60 - SPACING.M1*4 -- Total header width (60 is for the right screen padding)
    local subtitle_criteria_ratio = 0.65 -- Ratio between the width of the subtitle and the criteria
    local _text_w = scaled_header_w - flourish_w - SPACING.M1 -- How much space for the text after the icon
    local subtitle_w = (_text_w-SPACING.M1*2 ) * subtitle_criteria_ratio
    local criteria_w = (_text_w-SPACING.M1*2 ) * (1-subtitle_criteria_ratio)
	self.title:SetAutoSize( scaled_header_w )
		:LayoutBounds( "left", "top", 0, -60 )
	self.flourish_widget:LayoutBounds( "left", "below", self.title )
		:Offset( 0, -5 )
    self.subtitle:SetAutoSize( subtitle_w )
        :LayoutBounds( "after", "top", self.flourish_widget )
        :Offset( SPACING.M1, 0 )
    self.header_divider:LayoutBounds( "after", "top", self.subtitle )
    	:Offset( SPACING.M1, 0 )
    self.criteria:SetAutoSize( criteria_w )
    	:LayoutBounds( "after", "top", self.header_divider )
    	:Offset( SPACING.M1, 0 )
   	-- Get the height of the header elements and size the divider to match
   	local subtitle_w, subtitle_h = self.subtitle:GetSize()
   	local criteria_w, criteria_h = self.criteria:GetSize()
   	local max_h = math.max( flourish_h, subtitle_h, criteria_h )
   	self.header_divider:SetSize( 2, max_h )
   		:LayoutBounds( "after", "top", self.subtitle )
    	:Offset( SPACING.M1, 0 )

	-- Header height
    local w, h = self.header:GetSize()
    local header_height = h*LAYOUT_SCALE[sm]
    -- Footer height
    local w, h = self.bottom_left:GetSize()
    local footer_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[sm]
    -- Content height
    self.content_h = self.card_h - 40

	-- Update containers
    self.widepanel:SetHeight( self.content_h )
        :LayoutBounds( "center", "center", 0, -SPACING.M1*2.5 )
    self.header_gradient:LayoutBounds( nil, "above", self.widepanel )
        :Offset( 0, -2 )
    self.card_container:SetSize( content_w, self.content_h )
    	:LayoutBounds( "center", nil, self.header_bg )
    	:LayoutBounds( nil, "center", self.widepanel )
    	:Offset( 0, 30 )
    self.header:LayoutBounds( "center", "above", self.widepanel )
        :Offset( 0, SPACING.M1*1.3 )
    if self.particles_top then self.particles_top:LayoutBounds( "center", "above", self.widepanel ) end
    if self.particles_bottom then self.particles_bottom:SetScale(1,1):LayoutBounds( "center", "below", self.widepanel ):SetScale(1,-1) end

    -- Refresh card states
    for k, card_container in ipairs( self.card_container:GetFocusables() ) do
        self:UpdateCardImageState( card_container.card_w, card_container )
    end

    -- Get positions before screen shake
    self.header_x, self.header_y = self.header:GetPos()
    self.bottom_left_x, self.bottom_left_y = self.bottom_left:GetPos()
    self.bottom_right_x, self.bottom_right_y = self.bottom_right:GetPos()

    return self
end

function ChooseFlourishPopup:UpdateCardImageState( card_widget, card_container )

	if self.flourish_active then
        card_widget:Enable()
        card_widget:SetSaturation( 1 )
        -- card_container:Enable()
		if card_container.enabled and card_widget == self.selected_card then
	        card_widget:SetOutline( UICOLOURS.WHITE, true )
	        card_widget:ScaleTo( nil, CARD_SCALE_HOVER, 0.2, easing.outQuad )
	    elseif card_container.enabled and ( card_container.hover or card_container.focus ) then
	        card_widget:SetOutline( UICOLOURS.FOCUS, true )
	        card_widget:ScaleTo( nil, CARD_SCALE_HOVER, 0.2, easing.outQuad )
	    elseif card_container.enabled == false then
	        card_widget:SetOutline( nil )
	        card_widget:ScaleTo( nil, CARD_SCALE, 0.2, easing.outQuad )
	        card_widget:Disable()
	        card_widget:SetSaturation( 0 )
	    else
	        card_widget:SetOutline( card_widget.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT, true )
	        card_widget:ScaleTo( nil, CARD_SCALE, 0.2, easing.outQuad )
	    end

        -- Activate cards
        self.card_container:SetOnClickFn( function(...) self:OnCardClick(...) end )
	else
		-- Not enough flourish to play anything
        if card_container.enabled and card_container.focus then
            card_widget:ScaleTo( nil, CARD_SCALE_HOVER, 0.2, easing.outQuad )
        else
            card_widget:ScaleTo( nil, CARD_SCALE, 0.2, easing.outQuad )
        end
        card_widget:SetOutline( nil )
        card_widget:Disable()
        card_widget:SetSaturation( 0 )

        -- Disable cards
        self.card_container:SetOnClickFn( function() end )
	end

end

function ChooseFlourishPopup:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end

    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end

    self:Layout()
end

function ChooseFlourishPopup:GetDebugEnv( debug_env )
    debug_env.cards = self.flourishes
end
