require "ui/widgets/screen"
require "ui/widgets/pointsfloater"
require "ui/widgets/feats"
local daily = require "daily_util"

local DailyChallengeScreen = class("Screen.DailyChallengeScreen", Widget.Screen)

local SCREEN_INSET = 60
local SIDEBAR_W = 500
local SIDEBAR_CHARACTERS_W = 350
local SIDEBAR_LEFT_PADDING = 60 -- The sidebar has this padding on the left, so that glows aren't visibly clipped off. And then the whole sidebar is moved left this same amount, so the contents display in the same spot
local SCROLLPANEL_RIGHT_PADDING = 600 -- Extra space on the right of the sidebars to prevent stuff from getting clipped

local function GetRemainingTime()
    local dt = os.date("!*t")
    local time_left = (dt.hour * -3600 - dt.min * 60 - dt.sec) % 86400
    time_left = time_left / 60
    local hours_left = time_left / 60
    local minutes_left = math.floor(( hours_left - math.floor(hours_left) )*60)
    return math.floor(hours_left), minutes_left
end

function DailyChallengeScreen:GetRemainingTimeString()
    local hours_left, minutes_left = GetRemainingTime()
    if hours_left > 0 then
        return loc.format( LOC"UI.MAINMENU.DAILY_TIMER_HOURS_MINUTES", hours_left, minutes_left )
    else
        return loc.format( LOC"UI.MAINMENU.DAILY_TIMER_MINUTES", minutes_left )
    end
end

DailyChallengeScreen.CONTROL_MAP = 
{
    -- {
    --     control = Controls.Digital.GAMEPAD_X,
    --     fn = function(self)
    --         self:OnClickStart()
    --         return true
    --     end,
    --     hint = function(self, left, right)
    --         table.insert(right, loc.format(LOC"UI.CONTROLS.START_GAME", Controls.Digital.GAMEPAD_X))
    --     end,
    -- },
    {
        control = Controls.Digital.GAMEPAD_Y,
        fn = function(self)
            self:OnClickLeaderboards()
            return true
        end,
        test = function(self)
            return self.highscores_is_long_list
        end,
        hint = function(self, left, right)
            table.insert(left, loc.format(LOC"UI.DAILY_CHALLENGE_SCREEN.CONTROLLER_TOGGLE_LEADERBOARDS", Controls.Digital.GAMEPAD_Y))
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end,
        fn = function(self)
            self.close_button:Click()    
            return true
        end,
    },
    {
        test = function(self)
            return TheGame:FE():IsGamepadMode()
        end,
        hint = function(self, left, right) 
            local txt = LOC"UI.CONTROLS.SCROLL"
            txt = txt:gsub( "<c img='lt'>", "" ) -- To avoid retranslating new string, just cut out the LT control image.
            table.insert( left, txt )
        end,
    },
}

function DailyChallengeScreen:init()
	Widget.Screen.init( self )

	self.is_popup = true
    self:SetupUnderlay( true )
	self:SetAnchors("center", "center")

    self.sidebar_h = 600
    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Add our background plax
    self.plax = self:AddChild( Widget.Plax() )
        :SetAnchors( "center", "center" )
        :TrackMouse( true )
        :PostProcess( true )
        :SetTintColour( 0x85979Aff )

    -- Character anims
    self.characters = self:AddChild( Widget() )
        :SetAnchors( "center", "center" )
        :SetPos( 100, 0 )
    self.secondary_character = self.characters:AddChild( Widget.AnimatedCharacter() )
        :SetTintColour( 0xADBEC0ff )
        :SetFacing( false )
        :SetPhase( math.random() )
        :SetScale( 0.85 )
        :SetPos( -250, -310 )
    self.primary_character = self.characters:AddChild( Widget.AnimatedCharacter() )
        :SetFacing( false )
        :SetPhase( math.random() )
        :SetPos( 0, -310 )

    -- Backing gradient for the sidebar
    self.left_gradient_container = self:AddChild( Widget() )
        :SetAnchors( "left", "fill" )
    self.left_gradient = self.left_gradient_container:AddChild( Widget.Image( engine.asset.Texture("UI/newgame_sidebar_gradient.tex") ) )
        :SetSize( SIDEBAR_W*2, self.screen_h )
        :LayoutBounds( "left", "center", -5, 0 )
    -- Backing gradient for the right side contents
    self.right_gradient_container = self:AddChild( Widget() )
        :SetAnchors( "right", "fill" )
    self.right_gradient = self.right_gradient_container:AddChild( Widget.Image( engine.asset.Texture("UI/newgame_sidebar_gradient.tex") ) )
        :SetSize( SIDEBAR_W*1.2, self.screen_h )
        :LayoutBounds( "right", "center", 5, 0 )
        :SetScale( -1, 1 )

    -- Sidebar left
    self.sidebar_left_root = self:AddChild( Widget() )
        :SetAnchors( "left", "top" )
    self.sidebar_left_scroll = self.sidebar_left_root:AddChild( Widget.ScrollPanel() )
        :SetVirtualTopMargin( 60 )
        :SetVirtualBottomMargin( 250 )
        :SetScrollBarOuterMargin( 0 )
        :LeftScrollBar()
        :ShowBar( SCROLLBAR.NEVER )
        :SetSize( SIDEBAR_W+SCROLLPANEL_RIGHT_PADDING, self.sidebar_h )
    -- Sidebar contents
    self.sidebar_left_content = self.sidebar_left_scroll:AddScrollChild( Widget() )
    self.title = self.sidebar_left_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE, LOC"UI.DAILY_CHALLENGE_SCREEN.TITLE" ) )
        :SetGlyphColour( UICOLOURS.DAILY )
        :SetAutoSize( SIDEBAR_W )
        :SetWordWrap( true )
        :Bloom( 0.1 )
    self.subtitle = self.sidebar_left_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_SUBTITLE, "SUBTITLE: Into the Caverns" ) )
        :SetGlyphColour( UICOLOURS.DAILY_LIGHT )
        :SetAutoSize( SIDEBAR_W )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    self.text = self.sidebar_left_content:AddChild( Widget.Label( "body", FONT_SIZE.BODY_TEXT, "TEXT: Brill is going into the caves and has hired Rook to keep him alive. Brill is going into the caves and has hired Rook to keep him alive. Brill is going into the caves and has hired Rook to keep him alive." ) )
        :SetAutoSize( SIDEBAR_W )
        :SetWordWrap( true )
        :LeftAlign()
        :LeftAlign()
        :SetGlyphColour( 0xCBFFEFff )
        :Bloom( 0.1 )
    self.progression = self.sidebar_left_content:AddChild(Widget.DailyProgress())
    self.expiry_info = self.sidebar_left_content:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT, loc.format( LOC"UI.MAINMENU.DAILY_TIMER", 1, 2) ) )
        :SetGlyphColour( UICOLOURS.OBJECTIVE )
        :SetAutoSize( SIDEBAR_W )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    -- Modifier info
    self.mutators_block = self.sidebar_left_content:AddChild( Widget.DailyMutators( SIDEBAR_W ) )

    
    -- Sidebar right
    self.sidebar_right_root = self:AddChild( Widget() )
        :SetAnchors( "right", "top" )
    self.sidebar_right_content = self.sidebar_right_root:AddChild( Widget() )
    self.daily_high_score = self.sidebar_right_content:AddChild( Widget.DailyHighScore() )

    --self.top_right = self:AddChild( Widget() ):SetAnchors( "right", "top" )
    
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    -- Difficulty selector
    self.difficulty_block = self.bottom_right:AddChild( Widget.DailyDifficultySelector( 240 ) )
        :SetFn( function( difficulty ) self:OnClickStart( difficulty ) end )
        :LayoutBounds( "right", "above", -60, 60 )
    -- Feats info
    self.feats_block = self.bottom_right:AddChild( Widget.DailyFeatsList( SIDEBAR_W ) )
        :LayoutBounds( "right", "above", self.difficulty_block )
        :Offset( 0, SPACING.M1 )

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ))
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
    self.close_button.shadow = self.close_button:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( 0x021A18ff )
        :SetTintAlpha( 1 )
        :SetHiddenBoundingBox( true )
        :SetSize( 600, 250 )
        :SendToBack()
    -- self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.START", function() self:OnClickStart() end ) )
    --     :SetIcon( global_images.accept )
    --     :LayoutBounds( "right", "above", -60, 60 )

    self:Refresh()

    -- Update the time left every minute
    self:DoPeriodicTask("update_daily_challenge_screen", 15, 60, function()
        self:Refresh()
    end )
end

function DailyChallengeScreen:GetCurrentHighScore()
    -- Get current date
    local day = GetEpochDay()

    -- Go through the history and find runs with this starting date
    local high_score = 0
    local history = TheGame:GetRunHistory()
    for k, gamestate in ipairs( history ) do
        -- Get matching act
        local act = gamestate.options and GetPlayerActData( gamestate.options.start_act )
        -- Check if this run is a daily challenge
        if act and act.data.game_type == GAME_TYPE.DAILY and gamestate.creation_time and gamestate.score then
            -- Check if it started on this day
            local game_day = GetEpochDay(gamestate.creation_time)
            if game_day == day then
                -- Now check its score, and see if it's larger than our current highest
                high_score = math.max( high_score, gamestate.score )
            end
        end
    end
    return high_score > 0 and high_score or nil
end

function DailyChallengeScreen:OnClickClose()
    self:FadeOut()
end

function DailyChallengeScreen:OnClickStart( difficulty )
    if self.daily_challenge then
        self.StartDaily( self.daily_challenge, DAILY_DIFFICULTY_OPTIONS[difficulty].prestige )
    end
end

function DailyChallengeScreen:RefreshRobotOptions()
    self:ClearRobotOptions()
    self:AddRobotOption(self.close_button)
--     for k,v in ipairs(self.card_container.children) do  
--         if v.purchaseable then
--             self:AddRobotOption(v)
--         end
--     end  
end

function DailyChallengeScreen:OnGainTop()
    DailyChallengeScreen._base.OnGainTop( self )

    if self.plax then
        self.plax:PlaySounds()
    end
end

function DailyChallengeScreen:GetMusicEvent()
    return self.plax:GetMusicEvent() or "event:/music/character_select"
end


function DailyChallengeScreen:OnControlModeChange( cm, device_id )
    DailyChallengeScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.close_button:Show()
        if self.highscores_button then self.highscores_button:Show() end
        -- self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.close_button:Show()
        if self.highscores_button then self.highscores_button:Show() end
        -- self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.close_button:Hide()
        if self.highscores_button then self.highscores_button:Hide() end
        -- self.bottom_right:Hide()
    end

    self:Layout()
end

function DailyChallengeScreen:OnScreenModeChange( sm )
    DailyChallengeScreen._base.OnScreenModeChange( self, sm )

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    self.sidebar_h = self.screen_h

    self.sidebar_left_content:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.sidebar_right_content:SetLayoutScale( (LAYOUT_SCALE[sm]+1)/2 )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end

function DailyChallengeScreen:GetDefaultFocus()
    -- return self.card_widgets and #self.card_widgets > 0 and self.card_widgets[1]
end

function DailyChallengeScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
    -- self.bottom_right:SetPos( 0, 0 )
    -- if GetStreamerMode() == 2 then
    --     -- Nudge the right side widgets
    --     self.bottom_right:SetPos( -360, 0 )
    -- end
    self:Layout()
end

function DailyChallengeScreen:Refresh()
    self.daily_challenge = daily.GenerateDaily()
    self.mutators_block:Refresh( self.daily_challenge )    
    self.progression:Refresh(self.daily_challenge.progression, true)
    self.difficulty_block:Refresh( self.daily_challenge )
    self.feats_block:Refresh( self.daily_challenge )
    if self.daily_challenge.params.secondary then
        -- Only update the agent if it changed
        if not self.secondary_agent or ( self.secondary_agent and self.secondary_agent:GetUniqueID() ~= self.daily_challenge.params.secondary ) then
            local skin_table = Content.GetAllCharacterSkins(self.daily_challenge.params.secondary)
            self.secondary_agent = Agent( self.daily_challenge.params.secondary, skin_table[1] )
            self.secondary_character:Refresh( self.secondary_agent )
            self.secondary_character:SetFacing( false )
            self.secondary_character:Show()
            self.characters:SetPos( 150, 0 )
        end
    else
        self.secondary_agent = nil
        self.secondary_character:Hide()
        self.characters:SetPos( 50, 0 )
    end
    -- Only update the agent if it changed
    if not self.primary_agent or ( self.primary_agent and self.primary_agent:GetUniqueID() ~= self.daily_challenge.character ) then
        self.primary_agent = Agent( self.daily_challenge.character )
        self.primary_character:Refresh( self.primary_agent )
        self.primary_character:SetFacing( false )
    end
    self.subtitle:SetText(self.daily_challenge.def:GetTitle())
    self.text:SetText(self.daily_challenge.def:GetDesc())
    self.expiry_info:SetText( self:GetRemainingTimeString() )
    self.plax:SetPlaxData(self.daily_challenge.def.plax or "INT_Bog_Cave_01")

    if engine.steam then
        local function OnDownloadLeaderboard( t )
            if not self.removed then
                self:RefreshLeaderboards(t)
            end
        end
        engine.steam:DownloadDailyLeaderboard( OnDownloadLeaderboard )

    elseif engine.epic then
        local function OnDownloadLeaderboard( success, response )
           local localKU = engine.inst:GetUserName()
            if not self.removed and success then
                local t = json.decode( response )
                if t.Entries then
                    local scores = {}
                    for i, v in ipairs(t.Entries) do
                        if v.Score then
                            table.insert( scores, { score = v.Score, name = engine.epic:GetFriendName(v.User.EpicID), you = (v.KU == localKU) })
                        end
                    end
                    self:RefreshLeaderboards( scores )
                end
            end
        end
        engine.epic:DownloadDailyLeaderboard( OnDownloadLeaderboard )
    else
        self.daily_high_score:SetLocalScore()
    end

    self:Layout()

    self:RefreshRobotOptions()
    return self
end

function DailyChallengeScreen:RefreshLeaderboards( t )
    if t then
        -- Save this
        self.leaderboards_list = t

        -- If there isn't a highscores list, add one
        if not self.highscores then
            self.highscores = self.sidebar_right_content:AddChild( Widget() )
        end

        -- Clear it
        self.highscores:DestroyAllChildren()

        -- Sort list
        local short_list = self:GetShortLeaderboards( self.leaderboards_list, 4 )

        -- Check if we should have a sidebar list
        self.highscores_is_long_list = #self.leaderboards_list > #short_list

        -- Get player's score
        local player_index = nil
        for k, h in ipairs(short_list) do
            if h.you then player_index = k end
        end

        if player_index then
            local player_score = short_list[player_index]
            self.daily_high_score:SetVal( player_score.score, player_score.name, player_score.rank )
        end
        self.daily_high_score:LayoutBounds( "right", "top", -60, -20 )

        -- Add new ones
        for k, h in ipairs(short_list) do
            if not h.you then
                self.highscores:AddChild( Widget.RankedHighScore(h.rank, h.name, h.score) )
                    :LayoutBounds( "right", "below" )
                    :Offset( 0, 0 )
            end
        end
        self.highscores:LayoutBounds( "right", "below", self.daily_high_score )
            :Offset( 0, -15 )

        -- Add highscores button
        if not self.highscores_button then
            self.highscores_button = self.sidebar_right_content:AddChild( Widget.IconButton( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_BUTTON", function() self:OnClickLeaderboards() end ) )
                :SetIcon( engine.asset.Texture("UI/ic_score.tex") )
                :SetColour( UICOLOURS.QUEST )
        end
        self.highscores_button:LayoutBounds( "right", "below", self.highscores )
            :Offset( 0, -20 )
            :SetShown( self.highscores_is_long_list )

    else
        -- No leaderboards
        self.leaderboards_list = nil
        if self.highscores then
            self.highscores:Remove()
            self.highscores = nil
        end
        if self.highscores_button then
            self.highscores_button:Remove()
            self.highscores_button = nil
        end

        self.daily_high_score:SetLocalScore()
            :LayoutBounds( "right", "top", -60, -20 )
        self.highscores_is_long_list = false
    end

    self:Layout()
    return self
end

function DailyChallengeScreen:SortLeaderboards( list )
    -- Sort list by rank
    return table.stable_sort( list, function(h1,h2) return h1.rank < h2.rank end )
end

-- Goes through <list> and returns a list with <count> elements in it, including the player
function DailyChallengeScreen:GetShortLeaderboards( list, count )
    
    -- Sort list by rank
    local sorted = self:SortLeaderboards( list )

    -- Check where the player is on the list
    local player_index = nil
    for k, h in ipairs(sorted) do
        if h.you then player_index = k end
    end

    if player_index then -- The player is on the list! Yay
        -- If the player is in first, return them and the next top scores
        -- If the player is 2nd, show the top score, the player, and then scores below
        if player_index == 1 or player_index == 2 then
            local result = {}
            for k, h in ipairs(sorted) do
                if k <= count then
                    table.insert( result, h)
                else
                    break
                end
            end
            return result
        end

        -- If the player is 3rd or lower, show the top score, the score above the player and then the player and others below
        if player_index >= 3 then
            local result = {}
            for k, h in ipairs(sorted) do
                if k == 1 then
                    table.insert( result, h)
                elseif k == player_index-1 then
                    table.insert( result, h)
                elseif k >= player_index and #result < count then
                    table.insert( result, h)
                end
            end
            return result
        end

    else
        -- The player is not on the list. Return the top scores
        local result = {}
        for k, h in ipairs(sorted) do
            if k <= count then
                table.insert( result, h)
            else
                break
            end
        end
        return result
    end

end

-- Shows a sidebar with a list of scores
function DailyChallengeScreen:OnClickLeaderboards()
    local sidebar = Screen.LeaderboardsSidebar()
        :Refresh( self:SortLeaderboards(self.leaderboards_list) )
    TheGame:FE():InsertScreen( sidebar )
end

function DailyChallengeScreen:Layout()
    -- Layout things!
    self.subtitle:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_SUBTITLE )
    self.text:LayoutBounds( "left", "below", self.subtitle ):Offset( 0, -SPACING.M1 )
    self.progression:LayoutBounds( "left", "below", self.text ):Offset( 0, -SPACING.M1/2 )
    self.expiry_info:LayoutBounds( "left", "below", self.progression ):Offset( 0, -SPACING.M1/2 )
    self.mutators_block:LayoutBounds( "left", "below", self.expiry_info ):Offset( 0, -SPACING.M1 )
    
    self.sidebar_left_content:LayoutBounds( "left", "top", SIDEBAR_LEFT_PADDING, 0 )
    self.difficulty_block:LayoutBounds( "right", "above", -60, 60 )
    self.feats_block:LayoutBounds( "right", "above", self.difficulty_block ):Offset( 0, SPACING.M1 )

    self.sidebar_left_scroll:SetSize( SIDEBAR_W+SCROLLPANEL_RIGHT_PADDING, self.sidebar_h )
        :LayoutBounds( "left", "top", SCREEN_INSET-SIDEBAR_LEFT_PADDING, 0 )

    self.sidebar_left_scroll:RefreshView()

    self.daily_high_score:LayoutBounds( "right", "top", -60, -20 )
end

function DailyChallengeScreen:GetDefaultFocus()
    return self.difficulty_block:GetDefaultFocus() or self.close_button
end

function DailyChallengeScreen:OnUpdate(dt)
    DailyChallengeScreen._base.OnUpdate(self, dt)

    if TheGame:FE():IsGamepadMode() then
        local _,_,_,ry = TheGame:GetInput():GetActiveGamepadStickVals()
        local SPD = 50
        if math.abs(ry) > 0.1 then
            self.sidebar_left_scroll:Scroll(-SPD*ry)
        end
    end

end

function DailyChallengeScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.sidebar_left_scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.sidebar_left_scroll:ScrollUp()
        return true
    end
end

function DailyChallengeScreen.StartDaily( daily_data, advancement )
    local act_id = daily_data.character
    local options = GameState.CreateOptions(act_id .. "_DAILY", advancement)
    options.mutators = daily_data.mutators
    options.feats = daily_data.feats
    options.advancement_level = advancement
    options.worldseed = daily_data.worldseed
    options.main_quest = daily_data.id
    options.main_quest_params = {parameters = daily_data.params}

    local save_name = "slot_" .. act_id .. "_DAILY"
    TheGame:FE():FadeTransition( function()
        TheGame:StartNewGame( options, save_name)
    end, 0.5 )
end

------------------------------------------------------------------------------------------
-- A panel with a list of the modifiers on this Daily Challenge

local DailyMutators = class("Widget.DailyMutators", Widget)

local title_top_margin_fudge_factor = -6

function DailyMutators:init( width )
    DailyMutators._base.init(self)

    self.width = width or 200
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetSize( self.width, 250 )
        :SetTintColour( UICOLOURS.MUTATOR )
        :SetBloom(0.1)
    self.content = self:AddChild( Widget() )
    self.title = self.content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_LABEL, LOC"UI.DAILY_CHALLENGE_SCREEN.MUTATORS_TITLE" ) )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :SetAutoSize( self.width - SPACING.M1*2 )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    self.mutators_container = self.content:AddChild( Widget() )
    
end

function DailyMutators:Refresh( daily_challenge )

    -- Remove old mutator widgets
    self.mutators_container:DestroyAllChildren()

    -- Add new ones
    local mutators = daily_challenge.mutators
    if mutators then
        for k, id in ipairs(mutators) do
            local def = Content.GetGraft(id)
            if not def.hidden then
                self.mutators_container:AddChild( Widget.DailyMutator( self.width - SPACING.M1*2, id ) )
                    :LayoutBounds( "left", "below" )
                    :Offset( 0, -10 )
            end
        end
    end

    -- Layout things
    self.mutators_container:LayoutBounds( "left", "below", self.title )
        :Offset( 0, -SPACING.M1/2 )

    -- Resize background
    local w, h = self.content:GetSize()
    w = self.width
    h = h + SPACING.M1*2+title_top_margin_fudge_factor
    self.bg:SetSize( w, h )
    self.content:LayoutBounds( "left", "bottom", self.bg )
        :Offset( SPACING.M1, SPACING.M1 )

    return self
end

------------------------------------------------------------------------------------------
-- A single mutator on the list

local DailyMutator = class("Widget.DailyMutator", Widget)

local icon_spacing = 10

function DailyMutator:init( width, mutator_id )
    DailyMutator._base.init(self)

    self.width = width or 200
    self.icon_size = 50
    self.mutator_id = mutator_id
    self.graft = GraftInstance( self.mutator_id )

    self.features = {}
    self:SetToolTipClass( Widget.TooltipKeywords ):SetToolTip( self.features )

    -- Add contents
    self.icon = self:AddChild( Widget.Image( self.graft:GetImage() ) )
        :SetSize( self.icon_size, self.icon_size )
        :Bloom( 0.15 )
    self.text_content = self:AddChild( Widget() )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, self.graft:GetName() ) )
        :LeftAlign()
        :SetAutoSize( self.width-self.icon_size-icon_spacing )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.graft:GetDesc( self ) ) )
        :LeftAlign()
        :SetAutoSize( self.width-self.icon_size-icon_spacing )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )

    -- Layout stuffs
    self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    self.text_content:LayoutBounds( "after", "top", self.icon ):Offset( icon_spacing, 4 )
end

function DailyMutator:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function DailyMutator:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

------------------------------------------------------------------------------------------
-- A panel with a list of selectable difficulty levels on this Daily Challenge

local DailyDifficultySelector = class("Widget.DailyDifficultySelector", Widget)

function DailyDifficultySelector:init( width )
    DailyDifficultySelector._base.init(self)

    self.width = width or 200

    -- Add contents
    self.difficulty_container = self:AddChild( Widget() )
    -- Add difficulties
    for i, v in ipairs(DAILY_DIFFICULTY_OPTIONS) do
        self.difficulty_container:AddChild( Widget.IconButton( DAILY_DIFFICULTY_OPTIONS[i].title ) )
            :SetText( DAILY_DIFFICULTY_OPTIONS[i].title(), DAILY_DIFFICULTY_OPTIONS[i].desc() )
            :SetIcon( DAILY_DIFFICULTY_OPTIONS[i].btn_ic() )
            :LayoutBounds( "after", "center" )
            :Offset( SPACING.M1, 0 )
    end
end

function DailyDifficultySelector:SetFn( fn )
    self.on_click_fn = fn

    for i, widget in ipairs( self.difficulty_container.children ) do
        -- widget:SetOnClickFn( function() self.on_click_fn( widget:GetDifficultyLevel() ) end )
        widget:SetFn( function() self.on_click_fn( i ) end )
    end

    return self
end

function DailyDifficultySelector:Refresh()
    return self
end

function DailyDifficultySelector:GetDefaultFocus()
    return self.difficulty_container.children and self.difficulty_container.children[1]
end

DAILY_DIFFICULTY_OPTIONS = {
    {
        title = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_1_TITLE" end,
        desc = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_1_DESC" end,
        icon = function() return engine.asset.Texture("UI/daily_difficulty_1.tex") end,
        btn_ic = function() return engine.asset.Texture("UI/ic_daily_difficulty_1.tex") end,
        score_mult = 1,
		prestige = 0,
    },
    {
        title = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_2_TITLE" end,
        desc = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_2_DESC" end,
        icon = function() return engine.asset.Texture("UI/daily_difficulty_2.tex") end,
        btn_ic = function() return engine.asset.Texture("UI/ic_daily_difficulty_2.tex") end,
        score_mult = 1.5,
		prestige = 1,
    },
    {
        title = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_3_TITLE" end,
        desc = function() return LOC"UI.DAILY_CHALLENGE_SCREEN.DIFFICULTY_3_DESC" end,
        icon = function() return engine.asset.Texture("UI/daily_difficulty_3.tex") end,
        btn_ic = function() return engine.asset.Texture("UI/ic_daily_difficulty_3.tex") end,
        score_mult = 2,
		prestige = 4,
    }
}

------------------------------------------------------------------------------------------
-- A widget showing your current high score for this challenge

local DailyHighScore = class("Widget.DailyHighScore", Widget)

function DailyHighScore:init()
    DailyHighScore._base.init(self)

    -- Add widgets
    self.score_value = self:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE*1.6 ) )
        :RightAlign()
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom( 0.1 )
    self.score_desc = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_DESC, LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_DESC" ) )
        :RightAlign()
        :SetGlyphColour( UICOLOURS.DAILY_LIGHT )
        :Bloom( 0.15 )

    -- Get local high score
    self:SetLocalScore()
end

function DailyHighScore:SetLocalScore()
    local score = DailyChallengeScreen.GetCurrentHighScore()
    if score then
        self:SetVal( score )
    else
        self:Hide()
    end
    return self
end

function DailyHighScore:SetVal( score, name, rank )
    self.score_value:SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_VALUE", score ))
    
    if name then
        self.score_desc:SetText( name )
    else
        self.score_desc:SetText( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_DESC" )
    end

    self.score_desc:LayoutBounds( "right", "below", self.score_value )
        :Offset( 0, 10 )

    if rank then
        if not self.rank then
            self.rank = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_DESC*0.8, LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_DESC" ) )
                :RightAlign()
                :SetGlyphColour( UICOLOURS.WHITE )
                :Bloom( 0.15 )
        end
        self.rank:SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_RANK", rank ))
            :LayoutBounds( "before", "bottom", self.score_desc ):Offset( -8, 0.8 )
    else
        if self.rank then
            self.rank:Remove()
            self.rank = nil
        end
    end

    self:Show()

    return self
end

function DailyHighScore:Layout()
    if #(self.score_value:GetText() or "") > 0 then
        self:Show()
    else
        self:Hide()
    end

    return self
end

------------------------------------------------------------------------------------------
-- A widget showing your a high score with a rank

local RankedHighScore = class("Widget.RankedHighScore", Widget)

function RankedHighScore:init( rank, name, score)
    RankedHighScore._base.init(self)

    self.score = self:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE*0.7 ) )
        :RightAlign()
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom( 0.1 )
        :SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_VALUE", score ))
    self.rank = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_DESC*0.8 ) )
        :RightAlign()
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom( 0.1 )
        :SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_RANK", rank ))
    self.name = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_DESC ) )
        :RightAlign()
        :SetGlyphColour( UICOLOURS.DAILY_LIGHT )
        :Bloom( 0.15 )
        :SetText( name )

    self.name:LayoutBounds( "before", "bottom", self.score ):Offset( -4, 2 )
    self.rank:LayoutBounds( "before", "bottom", self.name ):Offset( -8, 0.8 )
end

------------------------------------------------------------------------------------------
-- Shows a sliding sidebar with a leaderboards list

local LeaderboardsSidebar = class( "Screen.LeaderboardsSidebar", Screen.SlidingSidebarScreen )

function LeaderboardsSidebar:init()
    LeaderboardsSidebar._base.init( self )

    self.scroll:SetVirtualTopMargin( 3 )
        :SetVirtualBottomMargin( SPACING.M1*4 )
end

function LeaderboardsSidebar:Refresh( leaderboards_list )
    self.leaderboards_list = leaderboards_list or {}

    -- Remove any old contents
    self.contents:DestroyAllChildren()

    local text_size = FONT_SIZE.ITEM_DESC
    local row_w = Screen.SlidingSidebarScreen.SIDEBAR_WIDTH - 74

    -- Add header
    local header = self.contents:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TABS ) )
        :SetGlyphColour( UICOLOURS.QUEST )
        :Bloom( 0.15 )
        :SetText( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_HEADER" )
        :LayoutBounds( "left", "top", 0, 0 )

    -- Add new ones
    for k, h in ipairs( self.leaderboards_list ) do
        -- Create row widget
        local row = self.contents:AddChild( Widget() )
        row.rank = row:AddChild( Widget.Label( "title", text_size ) )
            :SetGlyphColour( h.you and UICOLOURS.QUEST or UICOLOURS.WHITE )
            :Bloom( 0.1 )
            :SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_RANK", h.rank ))
        row.name = row:AddChild( Widget.Label( "title", text_size ) )
            :SetGlyphColour( h.you and UICOLOURS.QUEST or UICOLOURS.DAILY_LIGHT )
            :Bloom( 0.15 )
            :SetText( h.name )
        row.score = row:AddChild( Widget.Label( "title", text_size ) )
            :SetGlyphColour( h.you and UICOLOURS.QUEST or UICOLOURS.WHITE )
            :Bloom( 0.1 )
            :SetText( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_VALUE", h.score ))

        -- Layout row
        row.name:LayoutBounds( "left", nil, row.rank ):Offset( 70, 0 )
        row.score:LayoutBounds( "before", nil, row.rank ):Offset( row_w, 0 )
        
        row:LayoutBounds( "left", "below" ):Offset( 0, 0 )
    end
 
    self:Layout()
    return self
end

function LeaderboardsSidebar:Layout()

    self.contents:LayoutBounds( "left", "top", 4, 0 )

    LeaderboardsSidebar._base.Layout( self )
    return self
end