require "ui/widgets/screen"
require "ui/widgets/unlocktrack"
require "ui/screens/unlockinspectorscreen"
require "ui/screens/picklegacyscreen"
require "ui/screens/deckscreen"
require "ui/screens/dailychallengescreen"





function CalculateDailyScore( gamestate )
    local score_table = {}
    local won = false

    -- Check if this daily run is over
    if gamestate:GetGameOver() then
        -- It has ended, get the result
        local result = gamestate:GetGameOver().result
        won = result == GAMEOVER.ALPHA_VICTORY or result == GAMEOVER.VICTORY
    else
        -- It is in progress, ignore the feats
        won = false
    end

    local score = gamestate:GetScore()
    local components = gamestate.scorekeeper:GetScoreComponents()

    for k,v in ipairs(components) do
        table.insert( score_table, loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.HIGH_SCORE_TOOLTIP_LINE", v.txt, v.total ) )
    end

    -- Go through the various feats
    local score_mult = gamestate.scorekeeper:GetScoreMultiplier()

    local feats = gamestate:GetPlayerAgent().graft_owner:GetGrafts( GRAFT_TYPE.FEAT )
    for k, feat in ipairs(feats) do
        local pts = feat:GetDef().points or 0
        score = score + math.round( score_mult * ( feat.userdata.status == FEAT_STATUS.COMPLETE and pts or 0 ) )
    end

    -- print( "DAILY SCORE", score )
    return math.round(score), table.concat( score_table, "\n" )
end

local DailyOverScreen = class( "Screen.DailyOverScreen", Widget.Screen )

DailyOverScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.animation_started and not self.animation_ended
        end,
        fn = function(self) 
            -- Skip animation
            if self.animation_started then
                self.fast = true
                for k, w in ipairs(self.feats.children) do
                    w:SkipAnimation()
                end
                self.approach:SkipAnimation()
            end
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GAME_OVER_SCREEN.CONTROLS_RIGHT_SKIP", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.SKIP,
        test = function(self)
            return self.animation_started and not self.animation_ended
        end,
        fn = function(self) 
            -- Skip animation
            if self.animation_started then
                self.fast = true
                for k, w in ipairs(self.feats.children) do
                    w:SkipAnimation()
                end
                self.approach:SkipAnimation()
            end
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.animation_ended and TheGame:FE():IsGamepadMode()
        end,
        fn = function(self)
            -- Next screen
            self.button_continue:Click()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GAME_OVER_SCREEN.CONTROLS_RIGHT_CONTINUE", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            return TheGame:FE():GetFocusWidget()
        end,
        fn = function(self)
            TheGame:FE():ClearFocusWidget()
            return true
        end,
    },
}

local PANEL_H = 426
local CONTENT_W = RES_X/4 * 2.2
local PADDING = 20

function DailyOverScreen:init(gamestate)
    Widget.Screen.init(self)
    self.gamestate = gamestate
    local result = self.gamestate:GetGameOver().result
    self.won = result == GAMEOVER.ALPHA_VICTORY or result == GAMEOVER.VICTORY
    self.daily_date = self.gamestate.creation_time - (self.gamestate.creation_time % (60*60*24))

    self.is_abandon_mode = result == GAMEOVER.ABANDONED
    self.killed_by = self.gamestate:GetGameOver().killed_by

    if self.killed_by then TheGame:GetGameProfile():SetKilledByAgent( self.killed_by:GetUniqueID() ) end
    -- self.killed_by = Agent( "LUMINTHIAN" )
    self.is_popup = true
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
    self:SetSounds( SoundEvents.tut_popup, SoundEvents.tut_popout )

    -- Meta progression UI.
    ----------------------------

    self.previous_pts = 0

    if not self.gamestate:GetOptions().experiment then
        -- Contribute to card statistics
        local feat_stats = {}
        local feats = gamestate:GetPlayerAgent().graft_owner:GetGrafts( GRAFT_TYPE.FEAT )
        for k, feat in ipairs(feats) do
            table.insert(feat_stats, {ID = feat.id, RESULT = feat.userdata.status == FEAT_STATUS.COMPLETE})
        end
        TheGame:GetGameStats():AddCardStats(self.gamestate:GetCardStats())
        local daily_data = 
        {
            DATE = self.daily_date,
            POINTS = CalculateDailyScore(TheGame:GetGameState()),
            FEATS = feat_stats,
        }

        if daily_data.POINTS > 0 then
            if engine.steam then
                engine.steam:UploadDailyScore( daily_data.POINTS )
            elseif engine.epic then
                engine.epic:UploadDailyScore( daily_data.POINTS )
            end
        end

        -- Send metrics
        local json_t = 
        { 
            PLAYER_DATA = self.gamestate:GetPlayerState(),
            DAILY_DATA = daily_data,
        }
        if self.won then
            SendMetricsData("DAILY_OVER_WON", json_t)
        else
            SendMetricsData("DAILY_OVER_LOST", json_t)
        end
    end

    -- Remove this save file! Rogue!
    if self.gamestate.save_file_name and self.gamestate.save_file_name ~= "autosave" and self.gamestate:GetGameOver().result ~= GAMEOVER.ALPHA_VICTORY then
        self.save_file_name = self.gamestate.save_file_name
        TheGame:DeleteSave(self.gamestate.save_file_name, function() end)
    end

    self.current_act = self.gamestate.current_act and self.gamestate.current_act or "SAL"
    self.total_pts = 0

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )
    
    -- An animated character that killed the player
    self.enemy = self:AddChild( Widget() ):Hide()
    self.enemy_glow = self.enemy:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), 800, 600 ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.HEALTH )
        :SetTintAlpha( 0.7 )
    self.enemy_anim_root = self.enemy:AddChild(Widget())
        :Hide()
    self.enemy_anim = self.enemy_anim_root:AddChild( Widget.AnimatedCharacter() )
    self.enemy_text_container = self.enemy:AddChild( Widget() )
    local enemy_text_width = 280
    self.enemy_title = self.enemy_text_container:AddChild( Widget.Label("title", FONT_SIZE.ITEM_SUBTITLE ))
        :SetGlyphColour( UICOLOURS.HEALTH )
        :RightAlign()
        :SetWordWrap( true )
        :SetAutoSize( enemy_text_width )
        :Bloom( 0.09 )
    self.enemy_subtitle = self.enemy_text_container:AddChild( Widget.Label("body", FONT_SIZE.ITEM_DESC ))
        :SetGlyphColour( UICOLOURS.HEALTH )
        :SetTintAlpha( 0.9 )
        :RightAlign()
        :SetWordWrap( true )
        :SetAutoSize( enemy_text_width )
        :Bloom( 0.05 )

    -- A container for all the background elements
    self.panel = self:AddChild( Widget.WidePanel( PANEL_H, self.won and UICOLOURS.OBJECTIVE or UICOLOURS.PENALTY ) )

    -- Our fancy character anim!
    self.anim_root = self:AddChild(Widget()):SetHiddenBoundingBox( true )
    self:RefreshBigCharacter( self.gamestate:GetPlayerAgent() )

    -- Our titles and difficulty above the background panel
    self.header = self:AddChild( Widget() )
    self.title = self.header:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE) )
        :SetText( self.won and LOC"UI.GAME_OVER_SCREEN.TITLE_SUCCESS" or LOC"UI.GAME_OVER_SCREEN.TITLE_FAIL" )
        :SetGlyphColour( self.won and UICOLOURS.OBJECTIVE or UICOLOURS.PENALTY )
        :Bloom( 0.05 )
    self.subtitle = self.header:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TABS ))
        :SetText( loc.format( LOC"UI.DAILY_OVER_SCREEN.SUBTITLE", self.gamestate:GetMainQuest():GetTitle(), os.date("%x",self.daily_date) ) )
        :Bloom( 0.1 )
        :LayoutBounds( "center", "below", self.title )
        :Offset( 0, SPACING.TITLE_SUBTITLE )

    -- Difficulty label
    self.difficulty_widget = self.header:AddChild( Widget() )
    local prestige_level = self.gamestate:GetAdvancementLevel()
    local difficulty_level = 1
    for difficulty, difficulty_data in ipairs(DAILY_DIFFICULTY_OPTIONS) do
        if difficulty_data.prestige == prestige_level then difficulty_level = difficulty end
    end
    self.difficulty_backing = self.difficulty_widget:AddChild( Widget.Image( engine.asset.Texture("UI/gameover_progress_bg.tex") ) )
        :SetSize( 400, 60 )
        :SetTintColour( UICOLOURS.TITLE )
    self.difficulty_label = self.difficulty_widget:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_SUBTITLE, DAILY_DIFFICULTY_OPTIONS[difficulty_level].title() ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :Bloom( 0.1 )
        :LayoutBounds( "center", "center", self.difficulty_backing )
        :Offset( FONT_SIZE.ITEM_NAME/2, 0 )
    self.difficulty_icon = self.difficulty_widget:AddChild( Widget.Image( DAILY_DIFFICULTY_OPTIONS[difficulty_level].icon() ) )
        :SetSize( 40, 40 )
        :Bloom( 0.15 )
        :LayoutBounds( "before", "center", self.difficulty_label )
        :Offset( -6, 0 )
    self.difficulty_backing:SetScale( 1, -1 )
    self.difficulty_widget:LayoutBounds( "center", "below", self.subtitle )
        :Offset( 0, -8 )

    self.header:LayoutBounds( "center", "above", self.panel )
        :Offset( 0, 0 )

    -- Hitbox for content alignment
    self.hitbox = self:AddChild( Widget.SolidBox( CONTENT_W, PANEL_H, 0xFF000000 ) )
        :LayoutBounds( "left", "center", -350, 0 )

    -- Container for the info blocks
    self.panel_contents = self:AddChild( Widget() )

    -- Feats section
    local total_h = PANEL_H - PADDING*2 -- The total height available within the panel
    local buttons_h = 46
    local total_feats_w = CONTENT_W * 0.7 -- The three feats take up 70% of the space
    local total_feats_h = total_h - PADDING - buttons_h
    local feats = self.gamestate:GetPlayerAgent().graft_owner:GetGrafts( GRAFT_TYPE.FEAT )
    -- Count feats
    local feats_count = 0
    for k, feat in ipairs(feats) do
        feats_count = feats_count + 1
    end
    local feats_widgets_w = ( total_feats_w - PADDING*(feats_count-1) ) / feats_count -- Width of each feat
    self.feats = self.panel_contents:AddChild( Widget() )
    for k, feat in ipairs(feats) do
        
        self.feats:AddChild( Widget.FeatBlockDisplay( feats_widgets_w, total_feats_h, feat, self.won, gamestate.scorekeeper:GetScoreMultiplier()) )
    end
    self.feats:StackChildrenRow( PADDING )
        :LayoutBounds( "left", "top", self.hitbox ):Offset( 0, -PADDING*1 )


    local approach = gamestate.scorekeeper:GetScoreComponents()

    -- Approach section
    local total_approach_w = CONTENT_W * 0.3 -- The approach list takes up 30% of the space
    self.approach = self.panel_contents:AddChild( Widget.ApproachList( total_approach_w, total_h ) )
        :LayoutBounds( "after", "top", self.feats )
        :Offset( PADDING, 0 )
        :SetShown( #approach > 0 )
    -- Add approach blocks
    self.approach:AddBlocks( approach )

    -- Add info buttons
    self.battle_deck_btn = self.panel_contents:AddChild( Widget.AdvancedMenuButton( LOC"UI.DAILY_OVER_SCREEN.BATTLE_DECK_BTN" ) )
        :SetSize( feats_widgets_w, buttons_h )
        :SetIcon( engine.asset.Texture("UI/ic_subcard_fight.tex") )
        :SetColour( UICOLOURS.FIGHT )
        :SetOnClickFn( function()
                -- Show the player's battle cards
                local cards = self.gamestate:GetPlayerAgent():GetAspect("battler"):GetCards()
                local screen = Screen.DeckScreen( cards, nil, Widget.BattleCard, nil, true )
                screen:SetTitles( LOC"UI.GAME_OVER_SCREEN.STATS_BATTLE_SCREEN_TITLE", LOC"UI.SHOW_DECK_SCREEN.BATTLE_DECK_UPPERCASE" )
                TheGame:FE():InsertScreen( screen )
            end )
        :LayoutBounds( "left", "below", self.feats )
        :Offset( 0, -PADDING )
    self.negotiation_deck_btn = self.panel_contents:AddChild( Widget.AdvancedMenuButton( LOC"UI.DAILY_OVER_SCREEN.NEGOTIATION_DECK_BTN" ) )
        :SetSize( feats_widgets_w, buttons_h )
        :SetIcon( engine.asset.Texture("UI/ic_subcard_negotiation.tex") )
        :SetColour( UICOLOURS.NEGOTIATION )
        :SetOnClickFn( function()
                -- Show the player's negotiation cards
                local cards = self.gamestate:GetPlayerAgent():GetAspect("negotiator"):GetCards()
                local screen = Screen.DeckScreen( cards, nil, Widget.NegotiationCard, nil, true )
                screen:SetTitles( LOC"UI.GAME_OVER_SCREEN.STATS_NEGOTIATION_SCREEN_TITLE", LOC"UI.SHOW_DECK_SCREEN.NEGOTIATION_DECK_UPPERCASE" )
                TheGame:FE():InsertScreen( screen )
            end )
        :LayoutBounds( "after", "bottom", self.battle_deck_btn )
        :Offset( PADDING, 0 )
    self.graft_deck_btn = self.panel_contents:AddChild( Widget.AdvancedMenuButton( LOC"UI.DAILY_OVER_SCREEN.GRAFT_DECK_BTN" ) )
        :SetSize( feats_widgets_w, buttons_h )
        :SetIcon( engine.asset.Texture("UI/ic_graftscompendium.tex") )
        :SetColour( UICOLOURS.GRAFT )
        :SetOnClickFn( function()
                -- Show the player's grafts
                local agent = self.gamestate:GetPlayerAgent()
                local screen = Screen.GraftCollectionScreen( agent, true )
                TheGame:FE():InsertScreen( screen )
            end )
        :LayoutBounds( "after", "bottom", self.negotiation_deck_btn )
        :Offset( PADDING, 0 )

    -- Add total score counter widget
    self.pts_block = self:AddChild( Widget() )
        :SendBehind( self.panel )
    -- Add points counter
    self.pts_backing = self.pts_block:AddChild( Widget.Image( engine.asset.Texture("UI/gameover_progress_bg.tex") ) )
        :SetSize( 500, 74 )
        :SetTintColour( UICOLOURS.OBJECTIVE )
    self.pts_widget = self.pts_block:AddChild( Widget.PtsFloater( 0, 60 ) )
        :SetPts( 0 )
        :SetHiddenBoundingBox( true )
        :Bloom( 0.14 )
        :LayoutBounds( "center", "center", self.pts_backing )
        :Offset( 0, 4 )
    -- Add difficulty bonus info
    --[[
    self.difficulty_bonus_mult = DAILY_DIFFICULTY_OPTIONS[difficulty_level].bonus
    self.bonus_backing = self.pts_block:AddChild( Widget.Image( engine.asset.Texture("UI/gameover_progress_bg.tex") ) )
        :SetSize( 400, 50 )
        :SetTintColour( UICOLOURS.OBJECTIVE_LIGHT )
        :LayoutBounds( "center", "below", self.pts_backing )
        :SetShown( self.difficulty_bonus_mult > 0 )
    local bonus_text = loc.format( LOC"UI.DAILY_OVER_SCREEN.DIFFICULTY_BONUS", self.difficulty_bonus_mult )
    self.bonus_label = self.pts_block:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_SUBTITLE, bonus_text ) )
        :SetGlyphColour( UICOLOURS.OBJECTIVE_LIGHT )
        :Bloom( 0.1 )
        :LayoutBounds( "center", "center", self.bonus_backing )
        :Offset( 0, 2 )
        :SetShown( self.difficulty_bonus_mult > 0 )
    --]]        
    self.pts_block:LayoutBounds( "center", "below", self.panel:GetPanelWidget() )

    -- Continue button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    
    self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.GAME_OVER_SCREEN.BUTTON_CONTINUE", 
        function()
            self:GetFE():FadeTransition(
                    function()
                        self:GetFE():PopScreen(self) 

                        TheGame:Reset( function() 
                            TheGame:FE():InsertScreen( Screen.MainMenu() )
                        end)

                    end, .4)
        end ) )
        :SetIcon( global_images.accept )
        :LayoutBounds( "right", "above", -60, 60 )
        :Hide()
end

function DailyOverScreen:GetDefaultFocus()
    for k, w in ipairs(self.feats.children) do
        return w
    end
end

function DailyOverScreen:OnControlModeChange( cm, device_id )
    DailyOverScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_right:Hide()
    end
end

function DailyOverScreen:OnScreenModeChange( sm )
    DailyOverScreen._base.OnScreenModeChange( self, sm )

    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    
    self.panel:SetHeight( PANEL_H*LAYOUT_SCALE[sm] )

    self.enemy_text_container:SetLayoutScale( LAYOUT_SCALE[sm] )
    self:LayoutEnemyCharacter()

    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "center", "above", self.panel )
        :Offset( 0, 0 )
    self.pts_block:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "center", "below", self.panel:GetPanelWidget() )

    self.panel_contents:SetLayoutScale( LAYOUT_SCALE[sm] )

    if sm == SCREEN_MODE.MONITOR then
        self.anim_root:SetPos( 0, 0 )
        self.panel_contents:SetPos( 0, 0 )
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        self.anim_root:SetPos( -160, 0 )
        self.panel_contents:SetPos( -50, 0 )
    end

end

function DailyOverScreen:RefreshStreamerMode()
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

function DailyOverScreen:OnOpen()
    DailyOverScreen._base.OnOpen( self )

    -- Kick-off the animation!
    self:StartCoroutine( DailyOverScreen.ShowScoreCoro, self )

    -- Show the enemy, if any
    self:StartCoroutine( self.RefreshEnemyCharacter, self )

    if TheGame:GetLocalSettings().ROBOTICS then
        engine.inst:Quit()
        return
    end
end

function DailyOverScreen:OnPtsChange( delta )

    -- Add difficulty bonus
    --if self.difficulty_bonus_mult then delta = delta + self.difficulty_bonus_mult*delta end

    -- Add it up!
    self.total_pts = self.total_pts + math.round( delta )

    -- Update the progress bar
    local earned = self.total_pts - self.previous_pts
    
    if earned >= 1 then
        self.pts_widget:SetPts( self.total_pts )
    end
end

function DailyOverScreen:OnNewUnlock()
    self.biganim:Emote("clap")
end

function DailyOverScreen:ShowScoreCoro()
    for k, feat in ipairs(self.feats.children) do
        feat:PrepareAnimation()
    end
    self.battle_deck_btn:SetTintAlpha(0)
    self.negotiation_deck_btn:SetTintAlpha(0)
    self.graft_deck_btn:SetTintAlpha(0)
    self.approach:PrepareAnimation()

    self.animation_started = true
    for k, feat in ipairs(self.feats.children) do
        feat:AnimateIn( function( pts_delta ) self:OnPtsChange( pts_delta ) end )
    end
    self.approach:AnimateIn( function( pts_delta ) self:OnPtsChange( pts_delta ) end )
    if self.fast then self.battle_deck_btn:SetTintAlpha(1) else self.battle_deck_btn:AlphaTo( 1, 0.1 ) self:Delay( 0.1 ) end
    if self.fast then self.negotiation_deck_btn:SetTintAlpha(1) else self.negotiation_deck_btn:AlphaTo( 1, 0.1 ) self:Delay( 0.1 ) end
    if self.fast then self.graft_deck_btn:SetTintAlpha(1) else self.graft_deck_btn:AlphaTo( 1, 0.1 ) self:Delay( 0.1 ) end
    
    self:Delay( 0.2 )
    self.button_continue:Show()
    self.animation_ended = true
end

function DailyOverScreen:FadeInWidget( widget, timing)
    timing = timing or 1

    while widget:GetAlpha() < 1 do
        local dt = coroutine.yield()
        widget:SetTintAlpha( widget:GetAlpha() + timing*dt )
        if self.fast then
            break
        end
    end
    self.is_done_animating = true
    widget:SetTintAlpha( 1 )

    return self
end

function DailyOverScreen:RefreshBigCharacter( agent )
    local OFFSET = RES_X/4
    
    if agent then
        self.biganim = self.anim_root:AddChild( Widget.AnimatedCharacter() )
        self.biganim:Refresh( agent )

        local x = -RES_X/4-150
        local y = -270
        if self.won then
            self.biganim:SetMood( MOOD.HAPPY )
            x = -RES_X/4-150
        else
            self.biganim:SetMood( MOOD.INJURED )
            x = -RES_X/4-250
        end
        self.biganim:SetScale( 1 )
        self.biganim:SetPos( x-OFFSET, y )
        self.biganim:MoveTo( x, y, .6, easing.outQuad)
    end
end

function DailyOverScreen:RefreshEnemyCharacter()
    if self.won then return end
    if self.killed_by then

        -- Update the info text
        local loctable = self.killed_by:LocTable()
        loctable.name = loctable.name or loctable.fullname or loctable.Who
        self.enemy_title:SetText( string.upper( loc.format( LOC"UI.GAME_OVER_SCREEN.ENEMY_TITLE", loctable ) ) )
        self.enemy_subtitle:SetText( loc.format( LOC"UI.GAME_OVER_SCREEN.ENEMY_SUBTITLE", loctable ) )
        self.enemy:Show()

        -- Update the anim
        local scale = 0.27 --* ( fight_data.battle_scale or 1 )
        self.enemy_anim:SetScale( -scale, scale )
            :Refresh( self.killed_by )
            :SetOutline( 5, UICOLOURS.HEALTH )

        self:Delay( 0.1 )
        self:LayoutEnemyCharacter()
    end
end

function DailyOverScreen:LayoutEnemyCharacter()
    if self.won then return end
    if self.killed_by then

        -- Align the enemy info text
        self.enemy_subtitle:LayoutBounds( "right", "below", self.enemy_title )
            :Offset( 0, SPACING.TITLE_DESC )
        self.enemy_text_container
            :LayoutBounds( "right", nil, self.hitbox )
            :LayoutBounds( nil, "above", self.panel )
            :Offset( -60, 20 )
        self.enemy_glow:LayoutBounds( "center", "center", self.enemy_text_container )
            :Offset( 80, -200 )

        -- And the enemy relative to the text
        local text_x, text_y = self.enemy_text_container:GetPos()
        local text_w, text_h = self.enemy_text_container:GetScaledSize()
        local text_right_x, text_top_y = text_x+text_w/2, text_y+text_h/2

        -- Update the anim
        self.enemy_anim_root:SetPos( 0, 0 )
        local fight_data = self.killed_by.fight_data
        local head_offset_x, head_offset_y = 60, 0
        local anim_w, anim_h = self.enemy_anim_root:GetSize()
        if anim_w < 200 then head_offset_x = head_offset_x + 20 end -- Things the width of a regular humanoid (but not as wide as a Luminthian) should move right a bit
        if anim_w > 600 then head_offset_x = head_offset_x + 70 end -- Things the size of the Shroog should move right a bit
        if anim_h < 300 then head_offset_y = head_offset_y - 30 end -- Things the height of a Crayote should move down a bit

        -- Get the anim's own size (it has its own weird scale)
        local local_size_w, local_size_h = self.enemy_anim.anim:GetSize()
        -- Get the head's position within the anim's own size
        local local_head_x, local_head_y, local_head_z = self.enemy_anim.anim.model:SymbolToLocal( "head_swap", 0, 0, 0 )
        if fight_data.track_symbol then
            local_head_x, local_head_y, local_head_z = self.enemy_anim.anim.model:SymbolToLocal(fight_data.track_symbol, 0, 0, 0)
            head_offset_y = head_offset_y + 30 -- Drones should move up a bit
        end
        -- Calculate the proportional position of the head within the anim
        local relative_head_x, relative_head_y = local_head_x/local_size_w, local_head_y/local_size_h

        -- Get the widget's current size within the screen
        local scaled_size_w, scaled_size_h = self.enemy_anim:GetScaledSize()
        -- Calculate the head's position relative to that size
        local scaled_head_x, scaled_head_y = relative_head_x*scaled_size_w, relative_head_y*scaled_size_h

        -- Position the widget so the head is aligned with the top right corner of the text box
        self.enemy_anim_root:Show()
            :SetPos( text_right_x+scaled_head_x+head_offset_x, text_top_y-scaled_head_y+head_offset_y )

        -- local anim_x, anim_y, anim_z = self.enemy_anim:TransformToWorld(0,0,0)
        -- local dx, dy, dz = anim_x - head_x, anim_y - head_y, anim_z - head_z
        -- print ( "local size", local_size_w, local_size_h, "local head pos", local_head_x, local_head_y, "rel", relative_head_x, relative_head_y )
        -- print ( "- scaled size", scaled_size_w, scaled_size_h, "scaled head pos", scaled_head_x, scaled_head_y )
        -- print (dx, dy, "WWWWW")
        --if fight_data.track_symbol then
            --head_x, head_y, head_z = self.enemy_anim.anim.model:SymbolToWorld(fight_data.track_symbol, 0, 0, 0)
            -- head_offset_y = head_offset_y - 120
        --end
        -- local dx, dy, dz  = self.enemy_anim:TransformFromWorld( head_x, head_y, head_z )
        -- print (dx, dy,dz)
        -- self.enemy_anim_root:Show()
            --:SetPos( text_right_x - head_x + head_offset_x, text_top_y - head_y + head_offset_y )
            -- :SetPos(text_right_x, text_top_y)
        -- print("Offset", text_right_x, text_top_y)
        -- self.enemy_anim_root:LayoutBounds( "after", "top", self.enemy_text_container )
        --     :Offset( 0, 0 )
    end
end

------------------------------------------------------------------------------------------
-- Shows a large stats block

local FeatBlockDisplay = class( "Widget.FeatBlockDisplay", Widget )

local PADDING = 20
local ICON_SIZE = 90
local COMPLETED_HEIGHT = 30

function FeatBlockDisplay:init( width, height, feat_data, victory, score_mult )
    FeatBlockDisplay._base.init(self)
    score_mult = score_mult or 1
    -- Get feat info
    local colour = UICOLOURS.MUTATOR
    self.feat_data = feat_data
    self.feat_def = self.feat_data:GetDef()
    
    
    self.pts = math.round( (self.feat_def.points or 0)*score_mult )
    self.victory = self.feat_data.userdata.status == FEAT_STATUS.COMPLETE

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_block.tex") ) )
    self.content_bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( self.victory and colour or UICOLOURS.PENALTY )

    -- Size up our content block
    local nineslice_w, nineslice_h = width + self.bg_border*2*self.bg_border_scale, height + self.bg_border*2*self.bg_border_scale
    self.content_bg:SetSize( nineslice_w, nineslice_h )

    -- Add some glow!
    self.glow = self:AddChild( Widget.Image( engine.asset.Texture( "images/glow.tex" ), width+50, height+50 ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( colour )
        :SetTintAlpha( 0.2 )

    -- Our content block
    self.content = self:AddChild( Widget() )
    self.icon = self.content:AddChild( Widget.Image( engine.asset.Texture( self.feat_def.feat_icon ), ICON_SIZE, ICON_SIZE ) )
        :Bloom( 0.08 )
    self.title = self.content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, self.feat_data:GetName() ) )
        :SetGlyphColour( colour )
        :SetWordWrap( true )
        :SetAutoSize( width - PADDING*2 )
        :Bloom( 0.1 )
        :LayoutBounds( "center", "below", self.icon ):Offset( 0, -5 )
    self.subtitle = self.content:AddChild( Widget.Label( "body", FONT_SIZE.SMALL_TEXT, self.feat_data:GetDesc() ) )
        :SetGlyphColour( colour )
        :SetWordWrap( true )
        :SetAutoSize( width - PADDING*2 )
        :SetTintAlpha( 0.7 )
        :Bloom( 0.08 )
        :LayoutBounds( "center", "below", self.title ):Offset( 0, 5 )
    self.content:LayoutBounds( "center", "top", self.content_bg ):Offset( 0, -SPACING.M1 )

    -- Completed label
    self.completed = self:AddChild( Widget() )
    self.completed_bg_border = 20
    self.completed_bg_border_scale = 0.5
    local nineslice_w, nineslice_h = width - 1 + self.completed_bg_border*2*self.completed_bg_border_scale, COMPLETED_HEIGHT + self.completed_bg_border*2*self.completed_bg_border_scale
    self.completed_bg = self.completed:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_bg.tex") ) )
        :SetNineSliceCoords( self.completed_bg_border, self.completed_bg_border, self.completed_bg_border, self.completed_bg_border )
        :SetNineSliceBorderScale( self.completed_bg_border_scale )
        :SetTintColour( self.victory and colour or UICOLOURS.PENALTY )
        :Bloom( 0.03 )
        :SetSize( nineslice_w, nineslice_h )
    self.completed_text = self.completed:AddChild( Widget.Label( "body", FONT_SIZE.SMALL_TEXT, self.victory and LOC"UI.DAILY_OVER_SCREEN.FEAT_COMPLETE" or LOC"UI.DAILY_OVER_SCREEN.FEAT_MISSED" ) )
        :SetGlyphColour( UICOLOURS.DARK_BODYTEXT )
        :LayoutBounds( "center", "center", self.completed_bg ):Offset( 0, 1 )
    self.completed:LayoutBounds( "right", "bottom", self.content_bg )

    -- EXP floater!
    self.pts_floater = self:AddChild( Widget.PtsFloater( self.pts ) )
        :LayoutBounds( "center", "above", self.completed )
        :Offset( 0, 4 )
        :Hide()

    self:SetFocusable(true)
end

function FeatBlockDisplay:PrepareAnimation()
    self.is_done_animating = false
    self:SetTintAlpha( 0 )
    self.pts_floater:SetPts( 0 )
        :LayoutBounds( "center", "below", self.content ):Offset( 0, 10 )
        :SetTintAlpha( 0 )

    return self
end

function FeatBlockDisplay:AnimateIn( on_done_fn )
    local fade_in_rate = 3
    local count_up_rate = 8

    -- Fade in
    AUDIO:PlayParamEvent( "event:/ui/game_over_screen/stat_appear", "position", self:GetPanPos() )
    while self:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
    end

    if self.victory then
        self.pts_floater:SetPts( self.pts )
            :LayoutBounds( "center", "above", self.completed )
            :Offset( 0, 4 )
            :AnimateIn()
    end
    if on_done_fn then on_done_fn( self.victory and self.pts or 0 ) end

    self.is_done_animating = true
    self:SetTintAlpha( 1 )

    return self
end

function FeatBlockDisplay:SkipAnimation()
    self.fast = true
end

function FeatBlockDisplay:IsDoneAnimating()
    return self.is_done_animating
end

function FeatBlockDisplay:SetButtonText( text )
    self.button_text:SetText( text )
        :LayoutBounds( "left", "center", self.button_bg ):Offset( 15, 1 )
    return self
end

function FeatBlockDisplay:UpdateImage()
    if self.enabled and self.focus then
        self.icon:ScaleTo( nil, 1.1, 0.3, easing.inOutQuad )
        self.title:Bloom( 0.15 )
        self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
    else
        self.icon:ScaleTo( nil, 1, 0.3, easing.inOutQuad )
        self.title:Bloom( 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function FeatBlockDisplay:OnGainFocus()
    FeatBlockDisplay._base.OnGainFocus(self)
    self:UpdateImage()
end

function FeatBlockDisplay:OnLoseFocus()
    FeatBlockDisplay._base.OnLoseFocus(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a smaller stats display widget

local ApproachList = class("Widget.ApproachList", Widget)

local BLOCK_SPACING = 10

function ApproachList:init( width, height )
    ApproachList._base.init(self)

    self.width = width or 200
    self.height = height or 200

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = self.width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_wideblock.tex") ) )
    self.content_bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )

    -- Scroll panel!
    self.scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualTopMargin( PADDING )
        :SetVirtualBottomMargin( PADDING )
        :SetScrollBarOuterMargin( 0 )
        :ShowBar( SCROLLBAR.IF_NEEDED )
        :SetSize( self.width - 10, self.height )
        :LayoutBounds( "left", "top", self.content_bg )

    -- Add a container for our blocks
    self.list_container = self.scroll:AddScrollChild( Widget() )

    -- Calculate item width
    self.block_w = self.width - PADDING*2 - self.scroll:GetScrollBarWidth()
    self.text_w = self.block_w - 60 - 10 -- 60 is the width of the pts_floater with a 3 digit score (+110pts, for instance)
end

function ApproachList:OnAdded()
    TheGame:GetEvents():ListenForEvent( "focus_changed", self, self.OnFocusChanged )
end

function ApproachList:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function ApproachList:OnGainFocus()
    local focused_idx = self:GetFocusedIndex()
end

function ApproachList:GetFocusedIndex()
    for k, box in ipairs( self.list_container.children ) do
        if box:IsShown() and box.focus then return k end
    end
    return 1
end

function ApproachList:AddBlocks( approach_items )
    for k, approach in ipairs( approach_items ) do
        self:AddBlock( approach )
    end
    self:Layout()
    return self
end
function ApproachList:AddBlock( approach_data )
    local block = self.list_container:AddChild( Widget() )
    block.data = approach_data
    
    local total = approach_data.total or 0
    block.total = total

    block.title = block:AddChild( Widget.Label( "title", FONT_SIZE.BUTTON, approach_data.txt ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :SetWordWrap( true )
        :SetAutoSize( self.text_w )
        :OverrideLineHeight( 18 )
        :LeftAlign()
        :Bloom( 0.08 )

    block.exp_floater = block:AddChild( Widget.PtsFloater( total ) )
        :LayoutBounds( "before", "bottom", block.title )
        :Offset( self.block_w, 0 )

    -- Handle events
    block:SetFocusable( true )
    block.OnGainFocus = function() block.title:SetGlyphColour( UICOLOURS.TITLE ) end
    block.OnLoseFocus = function() block.title:SetGlyphColour( UICOLOURS.SUBTITLE ) end
end

function ApproachList:Layout()
    self.list_container:StackChildren( BLOCK_SPACING )
        :LayoutBounds( "left", "top", self.content_bg )
        :Offset( PADDING, -PADDING )

    self.scroll:RefreshView()
end

function ApproachList:PrepareAnimation()
    self.is_done_animating = false
    self:SetTintAlpha( 0 )
    for k, block in ipairs( self.list_container.children ) do
        block:SetTintAlpha( 0 )
        if block.exp_floater then block.exp_floater:SetTintAlpha( 0 ) end
    end

    return self
end

function ApproachList:AnimateIn( on_done_fn )
    local fade_in_rate = 8
    local block_fade_in_rate = 3

    -- Fade in
    AUDIO:PlayEvent("event:/ui/game_over_screen/approach_list_appears")
    while self:GetAlpha() < 1 do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
        if self.fast then
            break
        end
    end
    -- Not the blocks yet
    for k, block in ipairs( self.list_container.children ) do
        block:SetTintAlpha( 0 )
    end

    -- Fade in each approach block
    local page = 1
    local alpha = 0
    for k, block in ipairs( self.list_container.children ) do

        -- Fade in this block
        AUDIO:PlayParamEvent( "event:/ui/game_over_screen/approach_stat_appears", "position", block:GetPanPos() )
        alpha = 0
        while alpha < 1 and not self.fast do
            local dt = coroutine.yield()
            alpha = alpha + block_fade_in_rate*dt
            block:SetTintAlpha( alpha )
        end
        block:SetTintAlpha( 1 )
        self.scroll:EnsureVisible( block ) -- Scroll to this block
        if block.exp_floater then block.exp_floater:AnimateIn() end
        if on_done_fn then on_done_fn( block.total ) end
    end

    self.is_done_animating = true
    self:SetTintAlpha( 1 )

    return self
end

function ApproachList:SkipAnimation()
    self.fast = true
end

function ApproachList:IsDoneAnimating()
    return self.is_done_animating
end
