require "ui/widgets/screen"
require "ui/screens/gameoverscreen"

local DayOverScreen = class( "Screen.DayOverScreen", Widget.Screen )

local PANEL_H = 530
local CONTENT_W = RES_X/4 * 2.3
local PADDING = 20
local TITLE_PADDING = 6

DayOverScreen.CONTROL_MAP = 
{
    -- {
    --     control = Controls.Digital.MENU_ACCEPT,
    --     test = function(self)
    --         return self.animation_started and not self.animation_ended
    --     end,
    --     fn = function(self) 
    --         return self:AnimationNext()
    --     end
    -- },
    {
        control = Controls.Digital.GAMEPAD_X,
        test = function(self)
            return self.animation_started and not self.animation_ended
        end,
        fn = function(self) 
            return self:AnimationSkip()
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GAME_OVER_SCREEN.CONTROLS_RIGHT_SKIP", Controls.Digital.GAMEPAD_X))
        end
    },
    {
        control = Controls.Digital.SKIP,
        test = function(self)
            return self.animation_started and not self.animation_ended
        end,
        fn = function(self) 
            return self:AnimationSkip()
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.animation_ended and TheGame:FE():IsGamepadMode()
        end,
        fn = function(self)
            -- Next screen
            self:OnClickContinue()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GAME_OVER_SCREEN.CONTROLS_RIGHT_CONTINUE", Controls.Digital.MENU_ACCEPT))
        end
    },
}

function DayOverScreen:init( initial_state, sleep_data, fn )
    Widget.Screen.init(self)

    self.state = self:GenerateState( initial_state )
    self.sleep_data = sleep_data

    self.fn = fn

    self.is_popup = true
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
    self:SetSounds( SoundEvents.tut_popup, SoundEvents.tut_popout )

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- A container for all the background elements
    self.panel = self:AddChild( Widget.WidePanel( PANEL_H ) )

    -- Our fancy character anim!
    self.anim_root = self:AddChild(Widget()):SetHiddenBoundingBox( true )
    self:RefreshBigCharacter( TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent() )

    -- Our titles above the background panel
    self.header = self:AddChild( Widget() )
    self.subtitle = self.header:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TABS ) )
        :SetText( self.state.subtitle )
        :Bloom( 0.1 )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( LOC"UI.DAY_OVER_SCREEN.TITLE" )
        :SetGlyphColour(UICOLOURS.OBJECTIVE)
        :Bloom( 0.05 )
    self.subtitle:LayoutBounds( "center", "below", self.title )
        :Offset( 0, SPACING.TITLE_SUBTITLE )
    self.header:LayoutBounds( "center", "above", self.panel )
        :Offset( 0, SPACING.M1 )

    -- Hitbox for content alignment
    self.hitbox = self:AddChild( Widget.SolidBox( CONTENT_W, PANEL_H, 0xFF000000 ) )
        :LayoutBounds( "left", "center", -350, 0 )

    local bars_w = (CONTENT_W - PADDING)
    local half_columns_w = ( bars_w - PADDING ) / 2
    local quarter_columns_w = ( bars_w - PADDING*3 ) / 4
    local three_quarters_w = quarter_columns_w*3 + PADDING
    local card_columns_w = three_quarters_w / 2


    -- Container for the info blocks
    self.panel_contents = self:AddChild( Widget() )

    -- Add health bar
    self.health_bar = self.panel_contents:AddChild( Widget.BarBlockDisplay( half_columns_w, global_images.health, UICOLOURS.HEALTH, self.state.health, self.state.max_health, self.state.health_healed, LOC"UI.DAY_OVER_SCREEN.TITLE_HEALTH" ) )
    	:LayoutBounds( "left", "top", self.hitbox )
    	:Offset( 0, -PADDING/2 )

    -- Add resolve bar
    self.resolve_bar = self.panel_contents:AddChild( Widget.BarBlockDisplay( half_columns_w, global_images.resolve, UICOLOURS.RENOWN, self.state.resolve, self.state.max_resolve, self.state.resolve_healed, LOC"UI.DAY_OVER_SCREEN.TITLE_RESOLVE" ) )
    	:LayoutBounds( "after", "center", self.health_bar )
    	:Offset( PADDING, 0 )

   	-- Add friends gained
   	self.loved_block = self.panel_contents:AddChild( Widget.PeopleListBlock( quarter_columns_w, 180, LOC"UI.DAY_OVER_SCREEN.LOVED_GAINED", self.state.loved_gained, RELATIONSHIP.LOVED ) )
   		:LayoutBounds( "left", "below", self.health_bar )
    	:Offset( 0, -PADDING*1.5 )
    -- Add dislikes gained
    self.liked_block = self.panel_contents:AddChild( Widget.PeopleListBlock( quarter_columns_w, 180, LOC"UI.DAY_OVER_SCREEN.LIKED_GAINED", self.state.liked_gained, RELATIONSHIP.LIKED ) )
        :LayoutBounds( "after", "bottom", self.loved_block )
        :Offset( PADDING, 0 )
    -- Add dislikes gained
    self.disliked_block = self.panel_contents:AddChild( Widget.PeopleListBlock( quarter_columns_w, 180, LOC"UI.DAY_OVER_SCREEN.DISLIKED_GAINED", self.state.disliked_gained, RELATIONSHIP.DISLIKED ) )
        :LayoutBounds( "after", "bottom", self.liked_block )
        :Offset( PADDING, 0 )
    -- Add enemies gained
    self.hated_block = self.panel_contents:AddChild( Widget.PeopleListBlock( quarter_columns_w, 180, LOC"UI.DAY_OVER_SCREEN.HATED_GAINED", self.state.hated_gained, RELATIONSHIP.HATED ) )
        :LayoutBounds( "after", "bottom", self.disliked_block )
        :Offset( PADDING, 0 )

    -- Add negotiation cards gained
    self.negotiation_block = self.panel_contents:AddChild( Widget.CardListBlock( card_columns_w, 120, LOC"UI.DAY_OVER_SCREEN.NEGOTIATION_GAINED", self.state.negotiation_cards_received, UICOLOURS.NEGOTIATION ) )
        :LayoutBounds( "left", "below", self.loved_block )
        :Offset( 0, -PADDING )
   	
   	-- Add battle cards gained
   	self.battle_block = self.panel_contents:AddChild( Widget.CardListBlock( card_columns_w, 120, LOC"UI.DAY_OVER_SCREEN.BATTLE_GAINED", self.state.battle_cards_received, UICOLOURS.FIGHT ) )
        :LayoutBounds( "after", "bottom", self.negotiation_block )
        :Offset( PADDING, 0 )
    
    -- Add grafts gained
    self.grafts_block = self.panel_contents:AddChild( Widget.GraftListBlock( quarter_columns_w, 120, LOC"UI.DAY_OVER_SCREEN.GRAFTS_GAINED", self.state.grafts_received, UICOLOURS.OBJECTIVE ) )
        :LayoutBounds( "after", "bottom", self.battle_block )
        :Offset( PADDING, 0 )

   	-- Add hint
   	if self.state.hint then
	    self.hint_text = self:AddChild( Widget.Label("tooltip", FONT_SIZE.BODY_TEXT, loc.format( LOC"UI.DAY_OVER_SCREEN.HINT_TEXT", self.state.hint or "") ) )
	        :SetAutoSize( CONTENT_W*0.4 )
	        :SetWordWrap( true )
	        :CenterH()
	        :OverrideLineHeight( 19 )
	        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
	        :Bloom(0.1)
	        :SetTintAlpha(0)
            :LayoutBounds( "center", "below", self.panel )
            :Offset( 0, -SPACING.M1/2 )
    end

    -- Continue button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.DAY_OVER_SCREEN.BUTTON_CONTINUE", function() self:OnClickContinue() end ) )
        :SetIcon( global_images.accept )
        :LayoutBounds( "right", "above", -60, 60 )
        :Hide()

    -- Center of the screen
    self.center = self:AddChild( Widget() ):SetAnchors("center", "center")

    -- Set any necessary navigation overrides
    self:ReNav()
end

function DayOverScreen:GenerateState( initial_state )

    local player = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    if player == nil then
        return self:GenerateDefaultState()
    end

    -- Get friendos/enemios
    local loved_gained = {}
    local liked_gained = {}
    local disliked_gained = {}
    local hated_gained = {}
    local current_act_day, _, current_act_datetime = TheGame:GetGameState():GetActProgress()
    local player_expected_relationships = GetPlayerActData( TheGame:GetGameState():GetOptions().start_act ).data.score_modifiers and GetPlayerActData( TheGame:GetGameState():GetOptions().start_act ).data.score_modifiers.expected_relationships or table.empty

    if player.social_connections then
        -- Go through all the player's relationships
        for other_agent, agent_rel in pairs(player.social_connections.relationships) do
            -- Get the reason for the relationship
            local opinion_evt = other_agent.social_connections:GetRelationshipReason(player)
            if opinion_evt and opinion_evt.time then
                -- Check if this agent is not one of the default relationships for this act
                if table.arrayfind( player_expected_relationships, other_agent.id ) == nil then
                    if opinion_evt.time >= current_act_datetime then
                        if agent_rel == RELATIONSHIP.LOVED and not other_agent:IsDead() then
                            table.insert(loved_gained, other_agent)
                        end
                        if agent_rel == RELATIONSHIP.LIKED and not other_agent:IsDead() then
                            table.insert(liked_gained, other_agent)
                        end
                        if agent_rel == RELATIONSHIP.DISLIKED and not other_agent:IsDead() then
                            table.insert(disliked_gained, other_agent)
                        end
                        if agent_rel == RELATIONSHIP.HATED and not other_agent:IsDead() then
                            table.insert(hated_gained, other_agent)
                        end
                    end
                end
            end
        end
    end

    -- Get new cards
    local battle_cards_received = {}
    for i, card in ipairs( player.battler:GetCards() ) do
        if card.userdata and card.userdata.time_added ~= nil then
            if card.userdata.time_added >= current_act_datetime then
                table.insert(battle_cards_received, card)
            end
        end
    end

    local negotiation_cards_received = {}
    for i, card in ipairs( player.negotiator:GetCards() ) do
        if card.userdata and card.userdata.time_added ~= nil then
            if card.userdata.time_added >= current_act_datetime then
                table.insert(negotiation_cards_received, card)
            end
        end
    end

    -- Get new grafts
    local grafts_received = {}
    for i, graft in ipairs( player:GetAspect("graft_owner"):GetAllGrafts() ) do
        if graft.userdata and graft.userdata.time_added ~= nil and graft:GetType() ~= GRAFT_TYPE.SOCIAL and graft:GetType() ~= GRAFT_TYPE.MUTATOR then
            if graft.userdata.time_added >= current_act_datetime and not ( graft:IsHidden() or graft.hide_on_day_summary ) then
                table.insert(grafts_received, graft)
            end
        end
    end

    local percent, title, subtitle = TheGame:GetGameState():GetNarrativeProgress()
    local hp, maxhp = player:GetHealth()
    local resolve, maxresolve = TheGame:GetGameState():GetCaravan():GetResolve()

    if initial_state == nil then
        initial_state = {
            health = hp,
            max_health = maxhp,
            resolve = resolve,
            max_resolve = maxresolve
        }
    end

    local data = {
        character_id = player.id,
        subtitle = title,

        health = initial_state.health,
        max_health = maxhp,
        health_healed = hp - initial_state.health,

        resolve = initial_state.resolve,
        max_resolve = maxresolve,
        resolve_healed = resolve - initial_state.resolve,

        loved_gained = loved_gained,
        liked_gained = liked_gained,
        disliked_gained = disliked_gained,
        hated_gained = hated_gained,
        battle_cards_received = battle_cards_received,
        negotiation_cards_received = negotiation_cards_received,
        grafts_received = grafts_received,

        hint = UIHelpers.GetHint()
    }

    return data
end

function DayOverScreen:GenerateDefaultState()
    return
    {
        character_id = "SAL",
        subtitle = "Day 1: Welcome to Havaria",
        health = 44,
        max_health = 50,
        health_healed = 6,
        resolve = 11,
        max_resolve = 55,
        resolve_healed = 30,

        loved_gained = {
            Agent("RISE_SECOND"),
            Agent("FSSH"),
        },
        liked_gained = {
            Agent("RAKE"),
            Agent("ENDO"),
        },
        disliked_gained = {
            -- Agent("WANDERING_CHEF"),
        },
        hated_gained = {
            Agent("PLOCKA"),
        },

        battle_cards_received = { 
            Battle.Card("stab"),
            Battle.Card("muscle_memory"),
        },
        negotiation_cards_received = {
            Negotiation.Card("menacing_air"),
            Negotiation.Card("oppress"),
            Negotiation.Card("ipso_facto"),
            Negotiation.Card("fast_talk"),
            -- Negotiation.Card("oppress"),
            -- Negotiation.Card("ipso_facto"),
            -- Negotiation.Card("fast_talk"),
        },

        grafts_received = {
            GraftInstance("quick_strap"),
            GraftInstance("luminjectors"),
            GraftInstance("cerebral_stimulants"),
        },

        hint = UIHelpers.GetHint()
    }
end

function DayOverScreen:OnClickContinue()
    self:GetFE()
        :FadeTransition( 
            function()
                self:GetFE():PopScreen(self) 
                if self.fn then self.fn() end
            end, .4 )
end

function DayOverScreen:OnControlModeChange( cm, device_id )
    DayOverScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_right:Hide()
    end
end

function DayOverScreen:OnScreenModeChange( sm )
    DayOverScreen._base.OnScreenModeChange( self, sm )

    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    
    self.panel:SetHeight( PANEL_H*LAYOUT_SCALE[sm] )

    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "center", "above", self.panel )
        :Offset( 0, SPACING.M1 )

    self.panel_contents:SetLayoutScale( LAYOUT_SCALE[sm] )

    if sm == SCREEN_MODE.MONITOR then
        self.anim_root:SetPos( 0, 0 )
        self.panel_contents:SetPos( 0, 0 )
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        self.anim_root:SetPos( -200, 0 )
        self.panel_contents:SetPos( -130, 0 )
    end

    if self.hint_text then
        self.hint_text:SetLayoutScale( LAYOUT_SCALE[sm] )
            :LayoutBounds( "center", "below", self.panel )
            :Offset( 0, -SPACING.M1/2 )
    end

end

function DayOverScreen:GetDefaultFocus()
    if self._last_focus_widget and not self._last_focus_widget.removed then
        return self._last_focus_widget
    end
end

function DayOverScreen:ReNav()

    -- FIRST ROW
    -- ccc: This is pointless.
    -- self.health_bar:SetFocusDir( "left", self.resolve_bar, true )
    -- self.health_bar:SetFocusDir( "right", self.resolve_bar, true )

    -- SECOND ROW
    local second_row = {}
    second_row = self:AddFocusables( second_row, self.loved_block:GetFocusables() )
    second_row = self:AddFocusables( second_row, self.liked_block:GetFocusables() )
    second_row = self:AddFocusables( second_row, self.disliked_block:GetFocusables() )
    second_row = self:AddFocusables( second_row, self.hated_block:GetFocusables() )
    if #second_row > 0 then
        self._last_focus_widget = second_row[1]

        for k, w in ipairs(second_row) do

            if k == 1 then w:SetFocusDir( "left", second_row[#second_row] ) end -- Loop around to the last one
            -- if k == 1 then w:SetFocusDir( "left", w ) end -- Don't go left from the first one

            if k < #second_row then w:SetFocusDir( "right", second_row[k+1], true ) end -- Connect to the next one

            if k == #second_row then w:SetFocusDir( "right", second_row[1] ) end -- Loop around to the first one
            -- if k == #second_row then w:SetFocusDir( "right", w ) end -- Don't go right from the last one

        end
    end

    -- THIRD ROW
    local third_row = {}
    third_row = self:AddFocusables( third_row, self.negotiation_block:GetFocusables() )
    third_row = self:AddFocusables( third_row, self.battle_block:GetFocusables() )
    third_row = self:AddFocusables( third_row, self.grafts_block:GetFocusables() )
    if #third_row > 0 then
        self._last_focus_widget = self._last_focus_widget or third_row[1]

        for k, w in ipairs(third_row) do

            if k == 1 then w:SetFocusDir( "left", third_row[#third_row] ) end -- Loop around to the last one
            -- if k == 1 then w:SetFocusDir( "left", w ) end -- Don't go left from the first one

            if k < #third_row then w:SetFocusDir( "right", third_row[k+1], true ) end -- Connect to the next one

            if k == #third_row then w:SetFocusDir( "right", third_row[1] ) end -- Loop around to the first one
            -- if k == #third_row then w:SetFocusDir( "right", w ) end -- Don't go right from the last one

        end
    end
end

function DayOverScreen:AddFocusables( source_array, add_array )
    for k, v in ipairs( add_array ) do
        table.insert( source_array, v )
    end
    return source_array
end

function DayOverScreen:RefreshStreamerMode()
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

local OFFSET = RES_X/4

function DayOverScreen:RefreshBigCharacter( agent )
    
    if agent then
        self.biganim = self.anim_root:AddChild( Widget.AnimatedCharacter() )
        self.biganim:Refresh( agent )

        local x = -RES_X/4-150
        local y = -270
        if self.state.health/self.state.max_health < 0.5 then
            self.current_emote = "injured"
            self.biganim:Emote( self.current_emote )
            x = -RES_X/4-250
        elseif self.state.resolve/self.state.max_resolve < 0.5 then
            self.current_emote = "tired"
            self.biganim:Emote( self.current_emote )
            x = -RES_X/4-150
        else
            self.current_emote = "happy"
            self.biganim:Emote( self.current_emote )
            x = -RES_X/4-150
        end
        self.biganim:SetScale( 1 )
        self.biganim:SetPos( x-OFFSET, y )
        self.biganim:MoveTo( x, y, .6, easing.outQuad)
    end
end

function DayOverScreen:UpdateBigCharacterEmote( current_health, current_resolve )
    if not self.biganim then return end
    if current_health/self.state.max_health < 0.5 then
        if self.current_emote ~= "injured" then
            self.current_emote = "injured"
            self.biganim:Emote( self.current_emote )
        end
    elseif current_resolve/self.state.max_resolve < 0.5 then
        if self.current_emote ~= "tired" then
            self.current_emote = "tired"
            self.biganim:Emote( self.current_emote )
        end
    else
        if self.current_emote ~= "happy" then
            self.current_emote = "happy"
            self.biganim:Emote( self.current_emote )
        end
    end
end

function DayOverScreen:OnOpen()
    DayOverScreen._base.OnOpen( self )

    -- Kick-off the animation!
    self:StartCoroutine( DayOverScreen.ShowStatsCoro, self )
end

function DayOverScreen:ShowStatsCoro()
	self.health_bar:PrepareAnimation()
	self.resolve_bar:PrepareAnimation()
	self.loved_block:PrepareAnimation()
    self.liked_block:PrepareAnimation()
    self.disliked_block:PrepareAnimation()
	self.hated_block:PrepareAnimation()
	self.negotiation_block:PrepareAnimation()
	self.battle_block:PrepareAnimation()
    self.grafts_block:PrepareAnimation()

	self.animation_started = true
	self.health_bar:AnimateIn()
    self:UpdateBigCharacterEmote( self.state.health+self.state.health_healed, self.state.resolve )
	self.resolve_bar:AnimateIn()
    self:UpdateBigCharacterEmote( self.state.health+self.state.health_healed, self.state.resolve+self.state.resolve_healed )
	self.loved_block:AnimateIn()
    self.liked_block:AnimateIn()
    self.disliked_block:AnimateIn()
	self.hated_block:AnimateIn()
	self.negotiation_block:AnimateIn()
	self.battle_block:AnimateIn()
    self.grafts_block:AnimateIn()

    if self.sleep_data and #self.sleep_data.cards_removed > 0 then
        UIHelpers.ShowLostCards( self.sleep_data.cards_removed, self )
    end

	if self.state.hint then
    	if not self.fast then self:Delay(0.3) end
		self:FadeInWidget( self.hint_text, 4 )
    end

    self.animation_ended = true

	self.button_continue:Show()
    self:AddRobotOption(self.button_continue)

    self:GetFE():RefreshControlOptions()
end

function DayOverScreen:FadeInWidget( widget, timing)
    timing = timing or 1

    while widget:GetAlpha() < 1 do
        local dt = coroutine.yield()
        widget:SetTintAlpha( widget:GetAlpha() + timing*dt )
        if self.fast then
            break
        end
    end
    self.is_done_animating = true
    widget:SetTintAlpha( 1 )

    return self
end

function DayOverScreen:AnimationNext()
    self.fast = true
    if not self.health_bar.fast then
        self.health_bar:SkipAnimation()
        return true
    end
    if not self.resolve_bar.fast then
        self.resolve_bar:SkipAnimation()
        return true
    end
    if not self.loved_block.fast then
        self.loved_block:SkipAnimation()
        return true
    end
    if not self.liked_block.fast then
        self.liked_block:SkipAnimation()
        return true
    end
    if not self.disliked_block.fast then
        self.disliked_block:SkipAnimation()
        return true
    end
    if not self.hated_block.fast then
        self.hated_block:SkipAnimation()
        return true
    end
    if not self.negotiation_block.fast then
        self.negotiation_block:SkipAnimation()
        return true
    end
    if not self.battle_block.fast then
        self.battle_block:SkipAnimation()
        return true
    end
    if not self.grafts_block.fast then
        self.grafts_block:SkipAnimation()
        self.animation_ended = true
        return true
    end
end

function DayOverScreen:AnimationSkip()
    if self.animation_started then
        self:DoTaskInTime( "refresh_gamepad_hints", 0.1, function() TheGame:FE():RefreshControlOptions() end )
        self.fast = true
        if not self.grafts_block.fast then
            self.health_bar:SkipAnimation()
            self.resolve_bar:SkipAnimation()
            self.loved_block:SkipAnimation()
            self.liked_block:SkipAnimation()
            self.disliked_block:SkipAnimation()
            self.hated_block:SkipAnimation()
            self.negotiation_block:SkipAnimation()
            self.battle_block:SkipAnimation()
            self.grafts_block:SkipAnimation()
            self.animation_ended = true
            return true
        else
            self.animation_ended = true
            self.button_continue:SetFocus()
            return true
        end
    end
end

function DayOverScreen:OnUpdate( dt )
    DayOverScreen._base.OnUpdate(self, dt)
end


------------------------------------------------------------------------------------------
-- Shows a progress bar block

local BarBlockDisplay = class("Widget.BarBlockDisplay", Widget)

local ICON_SIZE = 70
local ICON_OVERLAP = -10
local BAR_WIDTH = 0
local BAR_HEIGHT = 12

local CLR_RESOLVE_BAR = UICOLOURS.NEGOTIATION
local CLR_RESOLVE_BG = 0x120D32FF

function BarBlockDisplay:init( width, icon, colour, value, max_value, delta, name )
    BarBlockDisplay._base.init(self)

    self.value = value
    self.max_value = max_value
    self.delta = delta
    self.heal_rate = delta / math.max( 1, max_value )
    self.name = name

    -- Add progress bar
    BAR_WIDTH = width - ICON_SIZE + ICON_OVERLAP
    self.bar = self:AddChild( Widget.ProgressBar( BAR_WIDTH, BAR_HEIGHT ) )
    	:OrientH()
    	:SetOutline(4)
    	:SetBarColour( colour )
    	:SetBGColour( CLR_RESOLVE_BG )
    	:SetOutlineColour( 0x00000000 )
    	:SetProgress( value/max_value )
    	:Bloom(0.1)

   	-- And progress bar framing
    local bg_border = 12
    local bg_border_scale = 0.5
    local nineslice_w, nineslice_h = BAR_WIDTH + bg_border*2*bg_border_scale, BAR_HEIGHT + bg_border*2*bg_border_scale
    self.bar_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/argument_bar_bg.tex") ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
    	:SetSize( nineslice_w, nineslice_h )
    	:Bloom(0.15)

    -- Icon
    self.icon = self:AddChild( Widget.Image( icon ) )
        :SetSize( ICON_SIZE, ICON_SIZE )
    	:LayoutBounds( "before", "center", self.bar_bg )
    	:Offset( ICON_OVERLAP, 6 )
    	:Bloom( 0.05 )

    self.current_value_label = self:AddChild( Widget.Label( "body", 50 ) )
        :SetText( tostring( math.round( self.value ) ) )
        :SetGlyphColour(colour)
        :Bloom( 0.1 )
        :LayoutBounds( "left", "above", self.bar_bg )
        :Offset( 0, 0 )
    self.max_value_label = self:AddChild( Widget.Label("title", 32 ) )
        :SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.BAR_MAX_VALUE", self.max_value, name ) )
        :SetGlyphColour(colour)
        :Bloom( 0.1 )
        :LayoutBounds( "after", "bottom", self.current_value_label )
        :Offset( 0, 5 )

    self.restoration_rate = self:AddChild( Widget.Label("title", 24 ) )
        :SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.BAR_RESULT_HEALED", self.heal_rate ) )
    	:SetGlyphColour( UICOLOURS.BONUS )
        :Bloom( 0.1 )
		:LayoutBounds( "left", "below", self.bar_bg )
        :Offset( 0, -4 )

    -- Add our result label
	self.result = self:AddChild( Widget.Label("title", 24 ) )
		:SetGlyphColour( colour )
        :Bloom( 0.1 )
        :SetHiddenBoundingBox( true )
    if delta == 0 then
    	self.result:SetText( LOC"UI.DAY_OVER_SCREEN.BAR_RESULT_MAX" )
    		:SetGlyphColour( UICOLOURS.SUBTITLE )
    		:LayoutBounds( "right", "below", self.bar_bg )
        	:Offset( 0, -4 )
    end
end

function BarBlockDisplay:PrepareAnimation()
    self.is_done_animating = false
    self:SetTintAlpha( 0 )
    self.restoration_rate:SetTintAlpha( 0 )
    self.restoration_rate:Hide()
    self.result:SetTintAlpha( 0 )
    return self
end

function BarBlockDisplay:AnimateIn( on_done_fn )
    local fade_in_rate = 3
    local count_up_rate = 20

    -- Fade in
    AUDIO:PlayParamEvent( "event:/ui/game_over_screen/stat_appear", "position", self:GetPanPos() )
    while self:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
    end
    self:SetTintAlpha( 1 )

    -- Show the rate
    self.restoration_rate:SetTintAlpha( 0 )
    self.restoration_rate:Show()
    local timing = 10
    while self.restoration_rate:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self.restoration_rate:SetTintAlpha( self.restoration_rate:GetAlpha() + timing*dt )
        if self.fast then
            break
        end
    end
    self.restoration_rate:SetTintAlpha( 1 )
    if not self.fast then self:Delay(0.2) end

    -- Count up the value
    if self.delta > 0 then
	    self.result:SetTintAlpha( 1 )
	    local current_value = self.value
	    local target_value = self.value + self.delta
	    while current_value < target_value and not self.fast do
	        local dt = coroutine.yield()
	        current_value = current_value + count_up_rate*dt
	        -- Update the value and align it
	        AUDIO:PlayParamEvent( "event:/ui/game_over_screen/xp_tick_up", "position", self:GetPanPos() )
	        self.current_value_label:SetText( tostring( math.round( current_value ) ) )
	        	:LayoutBounds( "left", "above", self.bar_bg )
	        self.max_value_label:LayoutBounds( "after", "bottom", self.current_value_label )
	        	:Offset( 0, 5 )
	        self.bar:SetProgress( current_value/self.max_value )
	        self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", math.round( current_value-self.value ) ) )
	    		:LayoutBounds( "right", "below", self.bar_bg )
	        	:Offset( 0, -4 )
	    end
	    self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", math.round( target_value-self.value ) ) )
    		:LayoutBounds( "right", "below", self.bar_bg )
        	:Offset( 0, -4 )
	    self.current_value_label:SetText( tostring( math.round( target_value ) ) )
	    	:LayoutBounds( "left", "above", self.bar_bg )
	    self.max_value_label:LayoutBounds( "after", "bottom", self.current_value_label )
	    	:Offset( 0, 5 )
	    self.bar:SetProgress( target_value/self.max_value )
	else
		-- Show the result
	    self.result:SetTintAlpha( 0 )
	    self.result:Show()
	    local timing = 10
	    while self.result:GetAlpha() < 1 and not self.fast do
	        local dt = coroutine.yield()
	        self.result:SetTintAlpha( self.result:GetAlpha() + timing*dt )
	        if self.fast then
	            break
	        end
	    end
	    self.result:SetTintAlpha( 1 )
	end

    if not self.fast then self:Delay(0.3) end

    return self
end

function BarBlockDisplay:SkipAnimation()
    self.fast = true
end

function BarBlockDisplay:UpdateImage()
    if self.enabled and self.focus then
        self.icon:Bloom( 0.1 )
        self.current_value_label:Bloom( 0.15 )
        self.max_value_label:Bloom( 0.15 )
        self:ScaleTo( nil, 1.02, 0.2, easing.outQuad )
    else
        self.icon:Bloom( 0.05 )
        self.current_value_label:Bloom( 0.1 )
        self.max_value_label:Bloom( 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function BarBlockDisplay:OnGainFocus()
    BarBlockDisplay._base.OnGainFocus(self)
    self:UpdateImage()
end

function BarBlockDisplay:OnLoseFocus()
    BarBlockDisplay._base.OnLoseFocus(self)
    self:UpdateImage()
end

function BarBlockDisplay:OnGainHover()
    BarBlockDisplay._base.OnGainHover(self)
    self:UpdateImage()
end

function BarBlockDisplay:OnLoseHover()
    BarBlockDisplay._base.OnLoseHover(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a list of agent portraits

local PeopleListBlock = class( "Widget.PeopleListBlock", Widget.Clickable )

local PADDING = 20
local CHARACTER_SCALE = 0.25

function PeopleListBlock:init( width, height, title, agents, relationship )
    PeopleListBlock._base.init(self)

    self.relationship = relationship or RELATIONSHIP.LOVED
    self.agents = agents

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = width + self.bg_border*2*self.bg_border_scale, height + self.bg_border*2*self.bg_border_scale
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_block.tex") ) )
    	:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( RELATIONSHIP_COLOURS[self.relationship] )
		:SetSize( nineslice_w, nineslice_h )

	self.anims_container = self:AddChild( Widget() )

	-- Create our agent widgets
	if agents and #agents > 0 then
        local _, bgheight = self.content_bg:GetScaledSize()
		for k, agent in pairs( agents ) do
            local w = self.anims_container:AddChild( Widget.PeopleListBlockAgent( agent, relationship ) )
            local width, height = w:GetScaledSize()
            w:SetScissor( -width/2, height/2 - bgheight, width, bgheight * 1.5 )
            w:InvalidateBBox()
            w:LayoutBounds( "center", "top", self.content_bg ):Offset( 0, -8 )
		end
	end

	-- Check if portraits are wider than container
	local w, h = self.anims_container:GetSize()
	if true then-- w > width*0.8 then
		local anim_spacing = width*0.8 / #self.anims_container.children
		self.anims_container:ArrangeChildrenRow( anim_spacing )
	end

    -- Set the container's scissoring
    local x1, y1, x2, y2 = self.content_bg:GetBoundingBox()
    -- Adjust for container's bg texture
    x1 = x1 + 2.5
    x2 = x2 - 0.5
    self.anims_container:SetScissor( x1, y1, x2 - x1, y2 - y1 )

    -- And our title label
    self.title = self:AddChild( Widget.Label( "title", 24, title or "" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

	-- Add our result label
	self.result = self:AddChild( Widget.Label( "title", 24, "" ) )
        :SetGlyphColour( RELATIONSHIP_COLOURS[self.relationship] )
        :Bloom( 0.1 )
        :LayoutBounds( "right", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

    self:SetFocusable( false )

end

function PeopleListBlock:GetFocusables()
    local list = {}
    for k, w in ipairs( self.anims_container.children ) do
        table.insert( list, w )
    end
    return list
end

function PeopleListBlock:PrepareAnimation()
    self:SetTintAlpha( 0 )
    for k, w in ipairs( self.anims_container.children ) do
		w:DisableToolTip( true )
	end
    return self
end

function PeopleListBlock:AnimateIn( on_done_fn )
    local fade_in_rate = 6
    local anim_fade_in_rate = 4

    -- Fade in
    AUDIO:PlayParamEvent( "event:/ui/game_over_screen/stat_appear", "position", self:GetPanPos() )
    while self:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
    end
    self:SetTintAlpha( 1 )

    -- Fade in the anims
    for k, anim in ipairs( self.anims_container.children ) do
    	if not self.fast then self:Delay(0.05) end
    	AUDIO:PlayEvent( "event:/ui/game_over_screen/xp_tick_up" )
    	anim:SetTintAlpha( 1 )
    	self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", k ) )
    		:LayoutBounds( "right", "above", self.content_bg )
    		:Offset( 0, TITLE_PADDING )
	end

	if self.agents == nil or #self.agents == 0 then
		self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", 0 ) )
    		:LayoutBounds( "right", "above", self.content_bg )
    		:Offset( 0, TITLE_PADDING )
    end

	-- Add tooltips back
	for k, w in ipairs( self.anims_container.children ) do
		w:DisableToolTip( false )
	end

    if not self.fast then
        self:Delay(0.3)
        self.fast = true
    end

    return self
end

function PeopleListBlock:SkipAnimation()
    self.fast = true
end

function PeopleListBlock:UpdateImage()
    if self.enabled and self.focus then
        self.title:Bloom( 0.1 )
        self.result:Bloom( 0.15 )
        self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
    else
        self.title:Bloom( 0.05 )
        self.result:Bloom( 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function PeopleListBlock:OnGainFocus()
    PeopleListBlock._base.OnGainFocus(self)
    self:UpdateImage()
end

function PeopleListBlock:OnLoseFocus()
    PeopleListBlock._base.OnLoseFocus(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows an agent within a people block

local PeopleListBlockAgent = class( "Widget.PeopleListBlockAgent", Widget.Clickable )

PeopleListBlockAgent.CONTROL_MAP = {
    {
        control = Controls.Digital.GAMEPAD_RT,
        fn = function(self)
            local details = Screen.PeopleCompendiumDetailsScreen()
            self:GetFE():PushScreen( details )
            details:ShowAgent( self.agent )
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.INSPECT", Controls.Digital.GAMEPAD_RT, self.agent ))
        end
    },
}

function PeopleListBlockAgent:init( agent, relationship )
    PeopleListBlockAgent._base.init(self)

    self.agent = agent
    self.relationship = relationship or RELATIONSHIP.LOVED

    -- Show the anim
    self.anim = self:AddChild( Widget.AnimatedCharacter() )
        :SetScale( ( self.relationship == RELATIONSHIP.LOVED or self.relationship == RELATIONSHIP.LIKED ) and CHARACTER_SCALE or -CHARACTER_SCALE, CHARACTER_SCALE )
        :SetOutline( 5, RELATIONSHIP_COLOURS[self.relationship] )
        :LayoutBounds( "after", "center" )
        :SetToolTip( nil )
        :Refresh( agent )
    local w, h = self.anim:GetScaledSize()

    -- Use a hitbox so we can control the width of each character, instead of relying on their anim's potentially huge bounding box
    self.hitbox = self:AddChild( Widget.SolidBox( 70, h, 0xff00ff00 ) )
        :SetPos( ( self.relationship == RELATIONSHIP.LOVED or self.relationship == RELATIONSHIP.LIKED ) and 10 or -10, 0 )

    -- Set emote
    if self.relationship == RELATIONSHIP.LOVED then self.anim:Emote( "happy" ) end
    if self.relationship == RELATIONSHIP.LIKED then self.anim:Emote( "neutral" ) end
    if self.relationship == RELATIONSHIP.DISLIKED then self.anim:Emote( "crossed" ) end
    if self.relationship == RELATIONSHIP.HATED then self.anim:Emote( "angry" ) end

    -- Make them breathe at different rates
    self.anim.anim:SetPhase( math.random() )

    -- Show this agent's graft
    if self.relationship == RELATIONSHIP.LOVED or self.relationship == RELATIONSHIP.HATED then
        local boon_id = agent:GetSocialGraft( self.relationship )
        local boon_graft = boon_id and GraftInstance( boon_id )
        self.graft = self:AddChild( Widget.RelationshipGraftIcon( 60, boon_graft ):SetOpaque() )
            :SetPos( ( self.relationship == RELATIONSHIP.LOVED ) and 8 or -24, 230 )
        self.graft:SetToolTip( nil )
    end

    self:SetFocusable( true )
    self:ShowToolTipOnFocus( true )
    self:SetToolTipClass( Widget.TooltipAgentSocialGrafts )
    self:SetToolTip( { AGENT = self.agent, RELATIONSHIP = self.relationship } )
    self.agent = agent
end

function PeopleListBlockAgent:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        AUDIO:PlayParamEvent( SoundEvents.card_mouseover, "position", self:GetPanPos() )
        self.anim:SetOutline( 8, RELATIONSHIP_COLOURS[self.relationship] )
        -- ccc: this messes up gamepad navigation
        -- self.anim:MoveTo( 0, 5, 0.2, easing.outQuad )
        -- if self.focus then self:ScaleTo( nil, 1.02, 0.2, easing.outQuad ) end
    else
        AUDIO:PlayParamEvent( SoundEvents.card_mouseoff, "position", self:GetPanPos() )
        self.anim:SetOutline( 5, RELATIONSHIP_COLOURS[self.relationship] )
        -- ccc: this messes up gamepad navigation
        -- self.anim:MoveTo( 0, 0, 0.2, easing.outQuad )
        -- self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function PeopleListBlockAgent:OnGainFocus()
    PeopleListBlockAgent._base.OnGainFocus(self)
    self:UpdateImage()
end

function PeopleListBlockAgent:OnLoseFocus()
    PeopleListBlockAgent._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PeopleListBlockAgent:OnGainHover()
    PeopleListBlockAgent._base.OnGainHover(self)
    self:UpdateImage()
end

function PeopleListBlockAgent:OnLoseHover()
    PeopleListBlockAgent._base.OnLoseHover(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a list of card previews

local CardListBlock = class( "Widget.CardListBlock", Widget.Clickable )

local PADDING = 20
local CARD_SCALE = 0.3

function CardListBlock:init( width, height, title, cards, colour )
    CardListBlock._base.init(self)

    self.width = width or 400
    self.height = height or 200
    self.scale_on_focus = true
    self.rotate_cards = true

    -- Add hitbox
    self.hitbox = self:AddChild( Widget.SolidBox( self.width, self.height, 0xff00ff30 ) )
        :SetTintAlpha( 0 )

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = self.width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_block.tex") ) )
    	:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
		:SetSize( nineslice_w, nineslice_h )

	-- Add our cards container
	self.cards_mask = self:AddChild( Widget() )
        :SetScissor( -self.width/2, -self.height/2, self.width, self.height )
        :SetHiddenBoundingBox( true )
    self.cards_container = self.cards_mask:AddChild( Widget() )

    -- And our title label
    self.title = self:AddChild( Widget.Label( "title", 24, title or "" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

	-- Add our result label
	self.result = self:AddChild( Widget.Label("title", 24 ) )
        :Bloom( 0.1 )
        :LayoutBounds( "right", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

    self:SetCardScale( CARD_SCALE )

    if cards then self:SetCards( cards, nil, true ) end
    if colour then self:SetColour( colour ) end

    self:SetFocusable( false )
end

function CardListBlock:SetSize( width, height )
    self.width = width or self.width or 400
    self.height = height or self.height or 200

    self.hitbox:SetSize( self.width, self.height )
    local nineslice_w, nineslice_h = self.width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.content_bg:SetSize( nineslice_w, nineslice_h )
    self.cards_mask:SetScissor( -self.width/2, -self.height/2, self.width, self.height )

    self:LayoutCards()

    return self
end

function CardListBlock:DisableCardRotation()
    self.rotate_cards = false
    self:LayoutCards()
    return self
end

function CardListBlock:SetCardScale( normal_scale, focus_scale )
    self.normal_scale = normal_scale or CARD_SCALE
    self.focus_scale = focus_scale or self.normal_scale*1.05
    return self
end

function CardListBlock:SetOnFocusFn( fn )
    self.on_focus_fn = fn
    return self
end

function CardListBlock:SetOnClickFn( fn )
    self.on_click_fn = fn
    for k, card_container in ipairs( self.cards_container.children ) do
        card_container:SetOnClickFn( function() if self.on_click_fn then self.on_click_fn( card_container.card_w, card_container ) else card_container.card_w:ShowCardInfoPopup() end end )
    end
    return self
end

function CardListBlock:SetCardIDs( card_ids, style, owner )
    local cards = {}
    for k, id in ipairs( card_ids ) do
        if Content.GetBattleCard(id) then
            table.insert( cards, Battle.Card(id,owner) )
        elseif Content.GetNegotiationCard(id) then
            table.insert( cards, Negotiation.Card(id,owner) )
        end
    end
    self:SetCards( cards, style )
    return self
end

function CardListBlock:SetCards( cards, style, no_card_focus )
    self.cards = cards

    -- Clear old cards
    self.cards_container:DestroyAllChildren()

    -- Create our card widgets
    if self.cards and #self.cards > 0 then
        for k, card in pairs( self.cards ) do
            local card_container = self.cards_container:AddChild( Widget.Clickable() ) -- Card containing widget
            if is_instance( card, Battle.Card ) then
                card_container.card_w = card_container:AddChild( Widget.BattleCard() )
                    :RefreshCard( card, nil, style )
            elseif is_instance( card, Negotiation.Card ) then
                card_container.card_w = card_container:AddChild( Widget.NegotiationCard() )
                    :RefreshCard( card, nil, style )
            end
            card_container.hitbox = card_container:AddChild( Widget.SolidBox( 140, 200, 0xff00ff00 ) ) -- Card container hitbox
            if card_container.card_w then
                card_container:SetOnRightClickFn( function() card_container.card_w:ShowCardInfoPopup() end )
                card_container.card_w:SetHiddenBoundingBox( true )
                card_container.card_w:SetScale( self.normal_scale, self.normal_scale )
                card_container.card_w:SetFocusable( true )
                card_container.card_w:SetToolTip( nil )
                card_container.card_w:SetToolTipClass( nil )
                card_container.card_w:ShowInfoOnClick()
                card_container.card_w:SetOutline( UICOLOURS.FOCUS, false )
                card_container.card_w:SetOutline( nil, false )
                if not no_card_focus then
                    card_container.OnGainFocus = function() card_container.card_w._base.OnGainFocus(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( card_container.card_w.card ) end end
                    card_container.OnLoseFocus = function() card_container.card_w._base.OnLoseFocus(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( nil ) end end
                    card_container.OnGainHover = function() card_container.card_w._base.OnGainHover(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( card_container.card_w.card ) end end
                    card_container.OnLoseHover = function() card_container.card_w._base.OnLoseHover(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( nil ) end end
                end
                card_container:LayoutBounds( "after", "center" )
                    :Offset( 0, 0 )


                -- Set gamepad hint
                card_container.CONTROL_MAP = {
                    {
                        control = Controls.Digital.GAMEPAD_RT,
                        fn = function(self)
                            card_container.card_w:ShowCardInfoPopup()
                            return true
                        end,
                        hint = function(self, left, right)
                            table.insert(right, loc.format(LOC"UI.CONTROLS.CARD_INFO", Controls.Digital.GAMEPAD_RT))
                        end
                    },
                }
            end
        end
    end

    self:LayoutCards()

    return self
end

function CardListBlock:LayoutCards()

    -- Get card size
    local first_cardcontainer = self.cards_container and self.cards_container.children and #self.cards_container.children > 0 and self.cards_container.children[1]
    if not first_cardcontainer then return self end
    first_cardcontainer.card_w:SetHiddenBoundingBox( false )
    local card_w, card_h = first_cardcontainer.card_w:GetScaledSize()
    first_cardcontainer.card_w:SetHiddenBoundingBox( true )
    -- Resize hitboxes to match
    for k, card_container in ipairs( self.cards_container.children ) do
        card_container.hitbox:SetSize( card_w, card_h ):SetPos( 0, 0 )
    end
    self.cards_container:StackChildrenRow( 0 )
    -- Check if the cards are wider than the container
    local w, h = self.cards_container:GetSize()
    local max_width = self.width - (self.content_bg:IsShown() and 100 or 0)
    if w > max_width then
        local card_spacing = max_width / #self.cards_container.children
        self.cards_container:ArrangeChildrenRow( card_spacing )
        -- Resize hitboxes
        for k, card_container in ipairs( self.cards_container.children ) do
            if k == 1 and k == #self.cards_container.children then -- Only one card
                card_container.hitbox:SetSize( card_spacing, card_h )
                    :SetPos( 0, 0 )
            elseif k == 1 then -- The first card has a wider hitbox to the left, so it's easier to hover
                local hitbox_w = card_spacing/2 + card_w/2
                card_container.hitbox:SetSize( hitbox_w, card_h )
                    :SetPos( -(card_w-hitbox_w)/2, 0 )
            elseif k == #self.cards_container.children then -- The last card has a wider hitbox to the right
                local hitbox_w = card_spacing/2 + card_w/2
                card_container.hitbox:SetSize( hitbox_w, card_h )
                    :SetPos( (card_w-hitbox_w)/2, 0 )
            else
                card_container.hitbox:SetSize( card_spacing, card_h )
                    :SetPos( 0, 0 )
            end
        end
    end

    -- Align card container
    self.cards_container:LayoutBounds( "center", "center", self.hitbox )

    -- Rotate cards
    if self.rotate_cards then
        for k, card_container in ipairs( self.cards_container.children ) do
            card_container:SetPos( nil, -34 )
            card_container.card_w:SetRotation( -5*DEGREES )
        end
    end

    return self
end

function CardListBlock:GetKeywords()
    local keywords = {}
    for k, card_container in ipairs( self.cards_container.children ) do
        table.arrayadd( keywords, card_container.card_w:GetCardFeatures() )
    end
    return keywords
end

function CardListBlock:HideBackground()
    self.content_bg:Hide()
    return self
end

function CardListBlock:SetColour( colour )
    self.mycolour = colour or self.mycolour or UICOLOURS.SUBTITLE
    self.content_bg:SetTintColour( self.mycolour )
    self.result:SetGlyphColour( self.mycolour )
    return self
end

function CardListBlock:GetFocusables()
    local list = {}
    for k, w in ipairs( self.cards_container.children ) do
        table.insert( list, w )
    end
    return list
end

function CardListBlock:HasCards()
    return self.cards_container and self.cards_container.children and #self.cards_container.children > 0
end

function CardListBlock:GetDefaultFocus()
    for k, w in ipairs( self.cards_container.children ) do
        return w
    end
end

function CardListBlock:PrepareAnimation()
    self:SetTintAlpha( 0 )
	for k, card in ipairs( self.cards_container.children ) do
		card:Hide()
	end
    return self
end

function CardListBlock:AnimateIn( on_done_fn )
    local fade_in_rate = 6
    local count_up_rate = 20

    -- Fade in
    AUDIO:PlayParamEvent( "event:/ui/game_over_screen/stat_appear", "position", self:GetPanPos() )
    while self:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
    end
    self:SetTintAlpha( 1 )

	-- Fade in the cards
    for k, card in ipairs( self.cards_container.children ) do
    	if not self.fast then self:Delay(0.05) end
    	AUDIO:PlayEvent( "event:/ui/game_over_screen/xp_tick_up" )
    	card:Show()
    	self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", k ) )
    		:LayoutBounds( "right", "above", self.hitbox )
    		:Offset( 0, TITLE_PADDING )
	end

	if self.cards == nil or #self.cards == 0 then
		self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", 0 ) )
    		:LayoutBounds( "right", "above", self.hitbox )
    		:Offset( 0, TITLE_PADDING )
    end

    -- Add tooltips
    for k, card_container in ipairs( self.cards_container.children ) do
    	if is_instance( card_container.card_w, Widget.BattleCard ) then
			card_container.card_w:SetToolTipClass( Widget.TooltipBattleCard )
		elseif is_instance( card_container.card_w, Widget.NegotiationCard ) then
			card_container.card_w:SetToolTipClass( Widget.TooltipNegotiationCard )
		end
        card_container.OnGainFocus = function() card_container.card_w._base.OnGainFocus(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( card_container.card_w.card ) end end
        card_container.OnLoseFocus = function() card_container.card_w._base.OnLoseFocus(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( nil ) end end
        card_container.OnGainHover = function() card_container.card_w._base.OnGainHover(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( card_container.card_w.card ) end end
        card_container.OnLoseHover = function() card_container.card_w._base.OnLoseHover(card_container.card_w) self:UpdateCardImageState( card_container ) if self.on_focus_fn then self.on_focus_fn( nil ) end end
	end

    if not self.fast then self:Delay(0.3) end

    self.fast = true

    return self
end

function CardListBlock:SkipAnimation()
    self.fast = true
end

function CardListBlock:SetUpdateCardImageFn( fn )
    self.update_image_state_fn = fn
    return self
end

function CardListBlock:UpdateCardImageState( card_container )
    local reverse = true
    self.cards_container:ArrangeChildrenByX( reverse )

    if self.update_image_state_fn then
        self.update_image_state_fn( card_container.card_w, card_container )
        return self
    end

    if card_container.enabled and ( card_container.hover or card_container.focus ) then
        card_container:SendToFront()

        if is_instance( card_container.card_w, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseover, "position", card_container:GetPanPos())
        elseif is_instance( card_container.card_w, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseover, "position", card_container:GetPanPos())
        end

        card_container.card_w:SetOutline( UICOLOURS.FOCUS, true )
        card_container.card_w:ScaleTo( nil, self.focus_scale, 0.2, easing.outQuad )
        card_container.card_w:MoveTo( nil, 6, 0.3, easing.outQuad )
    else

        if is_instance( card_container.card_w, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseoff, "position", card_container:GetPanPos())
        elseif is_instance( card_container.card_w, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseoff, "position", card_container:GetPanPos())
        end

        card_container.card_w:SetOutline( nil, true )
        card_container.card_w:ScaleTo( nil, self.normal_scale, 0.2, easing.outQuad )
        card_container.card_w:MoveTo( nil, 0, 0.3, easing.outQuad )
    end

end

function CardListBlock:SetScaleOnFocus( scale_on_focus )
    self.scale_on_focus = scale_on_focus or false
    return self
end

function CardListBlock:UpdateImage()
    if self.enabled and self.focus then
        self.title:Bloom( 0.1 )
        self.result:Bloom( 0.15 )
        if self.scale_on_focus then self:ScaleTo( nil, 1.05, 0.2, easing.outQuad ) end
    else
        self.title:Bloom( 0.05 )
        self.result:Bloom( 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function CardListBlock:OnGainFocus()
    CardListBlock._base.OnGainFocus(self)
    self:UpdateImage()
end

function CardListBlock:OnLoseFocus()
    CardListBlock._base.OnLoseFocus(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a list of graft previews

local GraftListBlock = class( "Widget.GraftListBlock", Widget.Clickable )

local PADDING = 20
local CARD_SCALE = 0.3

function GraftListBlock:init( width, height, title, grafts, colour )
    GraftListBlock._base.init(self)

    self.grafts = grafts

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = width + self.bg_border*2*self.bg_border_scale, height + self.bg_border*2*self.bg_border_scale
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_block.tex") ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( colour )
        :SetSize( nineslice_w, nineslice_h )

    -- Add our cards container
    self.grafts_container = self:AddChild( Widget() )
        :SetHiddenBoundingBox( true )


    -- Create our graft widgets
    if grafts and #grafts > 0 then
        for k, graft in pairs( grafts ) do
            local graft_w = self.grafts_container:AddChild( Widget.RelationshipGraftIcon( 80, graft ) )
                :SetOpaque()
                :LayoutBounds( "after", "center" )
                :Offset( 0, 0 )
                
            graft_w:SetFocusable( true )
            graft_w:ShowToolTipOnFocus( true )
            graft_w.OnGainFocus = function() graft_w._base.OnGainFocus(graft_w) self:UpdateGraftImageState( graft_w ) end
            graft_w.OnLoseFocus = function() graft_w._base.OnLoseFocus(graft_w) self:UpdateGraftImageState( graft_w ) end
            graft_w.OnGainHover = function() graft_w._base.OnGainHover(graft_w) self:UpdateGraftImageState( graft_w ) end
            graft_w.OnLoseHover = function() graft_w._base.OnLoseHover(graft_w) self:UpdateGraftImageState( graft_w ) end
        end
    end

    -- Check if grafts are wider than container
    local w, h = self.grafts_container:GetSize()
    if w > width*0.8 then
        local graft_spacing = width*0.8 / #self.grafts_container.children
        self.grafts_container:ArrangeChildrenRow( graft_spacing )
    end

    -- Align grafts container
    self.grafts_container:LayoutBounds( "center", "center", self.content_bg )

    -- Rotate cards
    for k, graft in ipairs( self.grafts_container.children ) do
        graft:SetRotation( -5*DEGREES )
    end

    -- And our title label
    self.title = self:AddChild( Widget.Label( "title", 24, title or "" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

    -- Add our result label
    self.result = self:AddChild( Widget.Label("title", 24 ) )
        :SetGlyphColour( colour )
        :Bloom( 0.1 )
        :LayoutBounds( "right", "above", self.content_bg )
        :Offset( 0, TITLE_PADDING )

    self:SetFocusable( false )

end

function GraftListBlock:GetFocusables()
    local list = {}
    for k, w in ipairs( self.grafts_container.children ) do
        table.insert( list, w )
    end
    return list
end

function GraftListBlock:PrepareAnimation()
    self:SetTintAlpha( 0 )
    for k, graft in ipairs( self.grafts_container.children ) do
        graft:Hide()
    end
    return self
end

function GraftListBlock:AnimateIn( on_done_fn )
    local fade_in_rate = 6
    local count_up_rate = 20

    -- Fade in
    AUDIO:PlayParamEvent( "event:/ui/game_over_screen/stat_appear", "position", self:GetPanPos() )
    while self:GetAlpha() < 1 and not self.fast do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
    end
    self:SetTintAlpha( 1 )

    -- Fade in the cards
    for k, graft in ipairs( self.grafts_container.children ) do
        if not self.fast then self:Delay(0.05) end
        AUDIO:PlayEvent( "event:/ui/game_over_screen/xp_tick_up" )
        graft:Show()
        self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", k ) )
            :LayoutBounds( "right", "above", self.content_bg )
            :Offset( 0, TITLE_PADDING )
    end

    if self.grafts == nil or #self.grafts == 0 then
        self.result:SetText( loc.format( LOC"UI.DAY_OVER_SCREEN.COUNTER_LABEL", 0 ) )
            :LayoutBounds( "right", "above", self.content_bg )
            :Offset( 0, TITLE_PADDING )
    end

    if not self.fast then self:Delay(0.3) end

    self.fast = true

    return self
end

function GraftListBlock:SkipAnimation()
    self.fast = true
end

function GraftListBlock:UpdateGraftImageState( widget )

    if widget.enabled and ( widget.hover or widget.focus ) then
        AUDIO:PlayParamEvent( SoundEvents.card_mouseover, "position", widget:GetPanPos() )
        widget:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
        widget.icon:MoveTo( nil, 4, 0.2, easing.outQuad )
        widget.icon:ScaleTo( nil, 1.1, 0.3, easing.outQuad )
        widget.icon_glow:ScaleTo( nil, 1.1, 0.4, easing.outQuad )
        widget.icon_frame:Bloom( 0.15 )
    else
        AUDIO:PlayParamEvent( SoundEvents.card_mouseoff, "position", widget:GetPanPos() )
        widget:ScaleTo( nil, 1, 0.2, easing.outQuad )
        widget.icon:MoveTo( nil, 0, 0.2, easing.outQuad )
        widget.icon:ScaleTo( nil, 1, 0.3, easing.outQuad )
        widget.icon_glow:ScaleTo( nil, 1, 0.4, easing.outQuad )
        widget.icon_frame:Bloom( 0.08 )
    end

end

function GraftListBlock:UpdateImage()
    if self.enabled and self.focus then
        self.title:Bloom( 0.1 )
        self.result:Bloom( 0.15 )
        self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
    else
        self.title:Bloom( 0.05 )
        self.result:Bloom( 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function GraftListBlock:OnGainFocus()
    GraftListBlock._base.OnGainFocus(self)
    self:UpdateImage()
end

function GraftListBlock:OnLoseFocus()
    GraftListBlock._base.OnLoseFocus(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a character details panel

local RelationshipGraftIcon = class( "Widget.RelationshipGraftIcon", Widget.Clickable )

RelationshipGraftIcon.CONTROL_MAP = {
    {
        control = Controls.Digital.GAMEPAD_RT,
        fn = function(self)
            self:ShowDetails()
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.GRAFT_INFO", Controls.Digital.GAMEPAD_RT ))
        end
    },
}

function RelationshipGraftIcon:init( size, graft )
    self._base.init( self )

    self:SetOnRightClickFn( function() self:ShowDetails() end )   

    -- Contents!
    self.icon_frame = self:AddChild( Widget.Panel( engine.asset.Texture("UI/empty_inner_frame_border_selected_clear.tex") ) )
        :Bloom( 0.08 )
    self.icon_glow = self:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), size*1.5, size*1.5 ) )
        :SetHiddenBoundingBox( true )
    self.icon = self:AddChild( Widget.Image() )
        :Bloom( 0.05 )

    self:SetSize( size )
    self:Refresh( graft )
end

function RelationshipGraftIcon:ShowDetails()
    local details = Screen.GraftsCompendiumDetailsPopup()
    TheGame:FE():PushScreen( details )
    details:ShowGraft( self.graft )
end

function RelationshipGraftIcon:SetSize( size )
    -- Layout our contents
    self.icon_frame:SetSize( size, size )
    self.icon:SetSize( size*0.9, size*0.9 )

    -- Refresh
    self:Refresh( self.graft )
end

function RelationshipGraftIcon:SetOpaque()
    self.icon_frame:SetTexture( engine.asset.Texture("UI/empty_inner_frame_border_selected_flat.tex") )
    return self
end

function RelationshipGraftIcon:Refresh( graft )
    if graft then
        local graft_def = graft:GetDef() or graft
        -- Set appropriate colors!
        if graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            self.icon_glow:SetTintColour( UICOLOURS.NEGOTIATION )
                :SetTintAlpha( 0.4 )
            self.icon_frame:SetTintColour( UICOLOURS.NEGOTIATION )
        elseif graft:GetType() == GRAFT_TYPE.COMBAT then
            self.icon_glow:SetTintColour( UICOLOURS.FIGHT )
                :SetTintAlpha( 0.4 )
            self.icon_frame:SetTintColour( UICOLOURS.FIGHT )
        elseif graft:GetType() == GRAFT_TYPE.SOCIAL and graft_def.is_good then
            self.icon_glow:SetTintColour( UICOLOURS.BONUS )
                :SetTintAlpha( 0.4 )
            self.icon_frame:SetTintColour( UICOLOURS.BONUS )
        elseif graft:GetType() == GRAFT_TYPE.SOCIAL and not graft_def.is_good then
            self.icon_glow:SetTintColour( UICOLOURS.PENALTY )
                :SetTintAlpha( 0.4 )
            self.icon_frame:SetTintColour( UICOLOURS.PENALTY )
        elseif graft:GetType() == GRAFT_TYPE.COIN then
            self.icon_glow:SetTintColour( UICOLOURS.MONEY )
                :SetTintAlpha( 0.4 )
            self.icon_frame:SetTintColour( UICOLOURS.MONEY )
        end
        self.graft = graft
        self.icon:SetTexture( graft:GetImage() )
        self:SetToolTipClass( Widget.TooltipGraft )
		self:SetToolTip( graft )
        self:Show()
    else
        self:SetToolTipClass( nil )
		self:SetToolTip( nil )
        self:Hide()
    end

    return self
end
