local fun = require "util/fun"
require "ui/widgets/cardgrid"
require "ui/widgets/negotiationpanel"
require "ui/screens/fightscreen"
require "ui/screens/slidingsidebarscreen"

local DeckScreen = class( "Screen.DeckScreen", Widget.Screen )

DeckScreen.CONTROL_MAP = 
{

    {
        control = Controls.Digital.GAMEPAD_RB,
        fn = function(self)
            self:ShowOtherDeck()
            return true
        end,
        test = function(self)
            return self.other_deck_data and self.other_deck_button
        end,
        
        hint = function(self, left, right)
            table.insert(left, loc.format(LOC"UI.CONTROLS.OTHER_DECK", Controls.Digital.GAMEPAD_RB))
        end
    },

    {
        control = Controls.Digital.GAMEPAD_LT,
        fn = function(self)
            self.nav_left.fn( self )
        end,
        test = function(self)
            return self.nav_left ~= nil
        end,        
        hint = function(self, left, right)
            table.insert(left, self.nav_left.hint_text)
        end
    },

    {
        control = Controls.Digital.GAMEPAD_RT,
        fn = function(self)
            self.nav_right.fn( self )
        end,
        test = function(self)
            return self.nav_right ~= nil
        end,
        
        hint = function(self, left, right)
            table.insert(left, self.nav_right.hint_text)
        end
    },

    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self) return self.close_button and self.close_button:IsShown() end,
        fn = function(self)
            if not self.animating_out then self:OnClickClose() end
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end
    },
}

function DeckScreen:init( cards, select_fn, card_class, on_end_fn, locked, no_close )
    DeckScreen._base.init( self )
    self.locked = locked
    assert( is_class( card_class, Widget.CardWidget ))
    
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )


    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )


    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.MONEY )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )


    -- Setup character anim display
    self.owner = TheGame:GetGameState():GetPlayerAgent()
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.owner )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "give", true )
        end
    end )
    self.character:SetAutoLoopEnabled( false )
    self.character:SetPos( -RES_X*0.35, -350 )
    self.character:Emote( "give", true )

    -- Save arguments
    self.card_class = card_class
    self.cards = cards or {}
    self.on_end_fn = on_end_fn
    self.select_fn = select_fn

    -- Cards panel
    self.card_grid = self:AddChild( Widget.CardGrid() )
        :SetTopPadding( 100 )
        :SetOnCardClickFn( function(w) self:OnCardClick(w) end )
        :SetCardInfoPopupMode( locked and CARD_DETAILS_MODE.LOCKED or CARD_DETAILS_MODE.DECK )
        :SetFilterButton( LOC"UI.SHOW_DECK_SCREEN.FILTERS_BUTTON_TT", function() self:ShowFilters() end )

    -- Customize colors and icons to the current card type
    if card_class == Widget.NegotiationCard then
        self.card_grid:SetColour( UICOLOURS.NEGOTIATION )
        self.grid:SetTintColour( UICOLOURS.NEGOTIATION )
        self.character:SetHoverOutlineColour( UICOLOURS.NEGOTIATION )
    elseif card_class == Widget.BattleCard then
        self.card_grid:SetColour( UICOLOURS.FIGHT )
        self.grid:SetTintColour( UICOLOURS.FIGHT )
        self.character:SetHoverOutlineColour( UICOLOURS.FIGHT )
    end

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    if not no_close then
        self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ) )
        self.close_button:SetIcon( global_images.close )
        self.close_button:LayoutBounds( "left", "above", 60, 60 )
    end
    
    self:Refresh( cards )
end

function DeckScreen:ShowFilters()
    if not self.filter_container or self.filter_container:IsEmpty() then
        -- If we haven't created all the filters necessary, do it now

        -- A container for all the filters to wait while the sidebar isn't open
        self.filter_container = self:AddChild( Widget() ):Hide()

        -- List of our filter widgets
        self.filters_widgets = {}

        -- Add a text search bar
        -- if SPECIAL_BUILD_TYPE < 1024 then
        --     local header = self.filter_container:AddChild( Widget.FilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_SEARCH" ) )
        --     table.insert( self.filters_widgets, header )
        --     local filter = self.filter_container:AddChild( Widget.CardSearchFilterBlock() )
        --         :SetOnToggleFn( function() self:OnFiltersChanged() end )
        --     table.insert( self.filters_widgets, filter )
        -- end

        -- Add a rarity filter
        local header = self.filter_container:AddChild( Widget.FilterHeader( LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY" ) )
        table.insert( self.filters_widgets, header )
        local filter = self.filter_container:AddChild( Widget.FilterOptionalRadioGroup() )
            :SetOnToggleFn( function() self:OnFiltersChanged() end )
            :SetRowMode()
        table.insert( self.filters_widgets, filter )
        filter:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.BASIC, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_BASIC_TT" ) )
        filter:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.COMMON, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_COMMON_TT" ) )
        filter:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.UNCOMMON, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_UNCOMMON_TT" ) )
        filter:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.RARE, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_RARE_TT" ) )
        filter:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.UNIQUE, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_UNIQUE_TT" ) )
        filter:SetSelected( 0 )

        -- Add a card-type filter
        local header = self.filter_container:AddChild( Widget.FilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_TYPE" ) )
        table.insert( self.filters_widgets, header )
        local filter = self.filter_container:AddChild( Widget.FilterOptionalRadioGroup() )
            :SetOnToggleFn( function() self:OnFiltersChanged() end )
            :SetRowMode( 3 )
        table.insert( self.filters_widgets, filter )
        if self.card_class == Widget.NegotiationCard then
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_HOSTILE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_HOSTILE_TT" ) )
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_DIPLOMACY, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_DIPLOMACY_TT" ) )
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_MANIPULATE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_MANIPULATE_TT" ) )
        elseif self.card_class == Widget.BattleCard then
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_MELEE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_MELEE_TT" ) )
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_SKILL, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_SKILL_TT" ) )
        end
        filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_STATUS, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_STATUS_TT" ):AddType( CARD_TYPE.FIGHT_STATUS ) )
        filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.ITEM, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_ITEM_TT" ) )
        filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.PARASITE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_PARASITE_TT" ) )
        filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FLOURISH, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FLOURISH_TT" ) )
        if TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY then
            filter:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.SCORE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_SCORE_TT" ) )
        end
        filter:SetSelected( 0 )

        -- Add a sorting selector
        local header = self.filter_container:AddChild( Widget.FilterHeader(LOC"UI.CARDCOMPENDIUM.CARD_FILTER_SORT" ) )
        table.insert( self.filters_widgets, header )
        self.sorting_widget = self.filter_container:AddChild( Widget.FilterRadioGroup() )
        -- And its options
        local sorting_options, selected_idx = self:GetSortingOptions()
        for k, option in ipairs( sorting_options ) do
            local row = self.sorting_widget:AddFilter( Widget.SortingOptionFilterRow( option.txt ) )
                :SetOnToggleFn( function( filter, is_toggled )
                    self.sorting_widget:FilterToggled( filter, is_toggled )
                    self:OnSortOrderChanged( option )
                end )
        end
        self.sorting_widget:SetSelected( selected_idx or 1 )

    end
    self.sidebar = self.card_class == Widget.NegotiationCard and Screen.NegotiationDeckFiltersSidebar() or Screen.BattleDeckFiltersSidebar()
    self.sidebar:Refresh( self.filters_widgets, self.sorting_widget )
        :SetOnCloseFunction( function(...) self:OnSidebarClosed(...) end )
    TheGame:FE():InsertScreen( self.sidebar )
    return self
end

function DeckScreen:CloseFilters()
    if self.sidebar then self.sidebar:Close() end
    return self
end

function DeckScreen:OnFiltersChanged()
    local filtered_cards = fun( self.cards )
        :filter(function(card) 
            for j, filter in ipairs( self.filters_widgets ) do
                -- If this filter rejects the card, hide the card widget
                if filter:Filter( card ) == false then return false end
            end
            return true
        end)
        :value()
    self.card_grid:Refresh( filtered_cards )
    return self
end

function DeckScreen:OnSortOrderChanged( option )
    if option then self.sorting_fn = option.fn end
    self.card_grid:SetSortingFn( self.sorting_fn )
    self:OnFiltersChanged()
    return self
end

function DeckScreen:OnSidebarClosed()
    -- Ths sidebar closed, and we have to store these widgets meanwhile, in case it opens again
    -- Add all the filter widgets
    for k, w in ipairs( self.filters_widgets ) do
        w:Reparent( self.filter_container )
    end
    -- And the sorting one
    self.sorting_widget:Reparent( self.filter_container )
end

function DeckScreen:GetSortingOptions()
    local sorting_options = {
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_DEFAULT", fn = nil },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_NAME", fn = function( w1, w2 ) return w1.card_data:GetTitle() < w2.card_data:GetTitle() end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_COST_ASC", fn = function( w1, w2 ) return w1.card_data.cost < w2.card_data.cost end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_COST_DES", fn = function( w1, w2 ) return w1.card_data.cost > w2.card_data.cost end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_DAMAGE_ASC", fn = 
            function( w1, w2 )
                local max1, max2 = 0, 0
                if is_instance( w1.card_data, Battle.Card ) then
                    max1 = w1.card_data.max_damage or 0
                elseif is_instance( w1.card_data, Negotiation.Card ) then
                    max1 = w1.card_data.max_persuasion or 0
                end
                if is_instance( w2.card_data, Battle.Card ) then
                    max2 = w2.card_data.max_damage or 0
                elseif is_instance( w2.card_data, Negotiation.Card ) then
                    max2 = w2.card_data.max_persuasion or 0
                end
                return max1 < max2
            end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_DAMAGE_DES", fn = 
            function( w1, w2 )
                local max1, max2 = 0, 0
                if is_instance( w1.card_data, Battle.Card ) then
                    max1 = w1.card_data.max_damage or 0
                elseif is_instance( w1.card_data, Negotiation.Card ) then
                    max1 = w1.card_data.max_persuasion or 0
                end
                if is_instance( w2.card_data, Battle.Card ) then
                    max2 = w2.card_data.max_damage or 0
                elseif is_instance( w2.card_data, Negotiation.Card ) then
                    max2 = w2.card_data.max_persuasion or 0
                end
                return max1 > max2
            end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_TYPE", fn = function( w1, w2 )
            local type1 = GetCardType(w1.card_data)
            local type2 = GetCardType(w2.card_data)

            if type1 == type2 then 
                return w1.card_data:GetTitle() < w2.card_data:GetTitle()
            end

            return type1 < type2
        end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_RARITY", fn = function( w1, w2 )
            local rarity1 = 0
            if w1.card_data.rarity == CARD_RARITY.BASIC then
                rarity1 = 1
            elseif w1.card_data.rarity == CARD_RARITY.COMMON then
                rarity1 = 2
            elseif w1.card_data.rarity == CARD_RARITY.UNCOMMON then
                rarity1 = 3
            elseif w1.card_data.rarity == CARD_RARITY.RARE then
                rarity1 = 4
            elseif w1.card_data.rarity == CARD_RARITY.UNIQUE then
                rarity1 = 5
            end

            local rarity2 = 0
            if w2.card_data.rarity == CARD_RARITY.BASIC then
                rarity2 = 1
            elseif w2.card_data.rarity == CARD_RARITY.COMMON then
                rarity2 = 2
            elseif w2.card_data.rarity == CARD_RARITY.UNCOMMON then
                rarity2 = 3
            elseif w2.card_data.rarity == CARD_RARITY.RARE then
                rarity2 = 4
            elseif w2.card_data.rarity == CARD_RARITY.UNIQUE then
                rarity2 = 5
            end
            
            if rarity1 == rarity2 then 
                return w1.card_data:GetTitle() > w2.card_data:GetTitle()
            end

            return rarity1 < rarity2
        end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_NOTUPGRADED", fn = function( w1, w2 )
            local xp1, xp2
            if w1.card_data:HasUpgrades() then
                if w1.card_data:UpgradeReady() then
                    xp1 = 1000
                elseif w1.card_data:GetMaxXP() then
                    local xp, max_xp = w1.card_data:GetXP()
                    xp1 = xp/max_xp + 1
                else
                    xp1 = 0
                end
            else
                xp1 = 0
            end
            if w2.card_data:HasUpgrades() then
                if w2.card_data:UpgradeReady() then
                    xp2 = 1000
                elseif w2.card_data:GetMaxXP() then
                    local xp, max_xp = w2.card_data:GetXP()
                    xp2 = xp/max_xp + 1
                else
                    xp2 = 0
                end
            else
                xp2 = 0
            end
            return xp1 > xp2
        end },
    }
    local selected_idx = 1 -- Default
    return sorting_options, selected_idx
end

function DeckScreen:Refresh( cards )
    -- Update card grid
    self.card_grid:Refresh( cards )
    self.card_widgets = self.card_grid:GetCardWidgets()
end

function DeckScreen:OnUpdate(dt)
    DeckScreen._base.OnUpdate(self, dt)

    self:ClearRobotOptions()
    for i, widget in ipairs( self.card_widgets ) do
        self:AddRobotOption(widget)
    end

    if TheGame:FE():IsGamepadMode() then
        local _,_,_,ry = TheGame:GetInput():GetActiveGamepadStickVals()
        local SPD = 50
        if math.abs(ry) > 0.1 then
            self.card_grid:Scroll(-SPD*ry)
        end
    end

end

function DeckScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.card_grid:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.card_grid:ScrollUp()
        return true
    end
end

function DeckScreen:SetHoverFn( on_hover )
end

function DeckScreen:OnCardClick( widget )
    if widget and self.select_fn then
        -- If we're selecting a card for an impending coroutine, run that
        widget:SetFocus()
        -- self:StartCoroutine( self.select_fn, self, widget, widget.card )
        self.select_fn( self, widget, widget.card )
        --self.select_fn = nil

    else
        if not TheGame:GetLocalSettings().ROBOTICS then
            if widget.card and widget.card:UpgradeReady() and not widget.card:IsHatched() then
                -- If this card is ready to upgrade, show that
                self:DoUpgrade( widget )
            else
                -- If not, show the info popup
                widget:ShowCardInfoPopup()
            end
        end
    end
end

function DeckScreen:TrashCard( widget )
    if is_instance( widget.card, Negotiation.Card ) then
        widget.card:GetOwner().negotiator:RemoveCardByID( widget.card.id, widget.card.userdata )
        SendMetricsData("REMOVED_NEGOTIATION_CARD", {SELECTED_ID = widget.card.id})
    else
        widget.card:GetOwner().battler:RemoveCardByID( widget.card.id, widget.card.userdata )
        SendMetricsData("REMOVED_BATTLE_CARD", {SELECTED_ID = widget.card.id})
    end
    AUDIO:PlayEvent("event:/ui/select_cards/remove_card")
    self.card_grid:RemoveCard( widget, function()
        self.card_grid:LayoutCards()
    end )
end

function DeckScreen:SetCancelText( text )
    self.close_button:SetText(text):Show()
end

function DeckScreen:SetTitles( title, subtitle, text )
    self.title = title
    self.subtitle = subtitle
    self.text = text
    self.card_grid:SetTitles( loc.upper(title), subtitle, text )
    return self
end

function DeckScreen:SetNavLeft( hint_text, fn )
    self.nav_left = { hint_text = hint_text, fn = fn }
end

function DeckScreen:SetNavRight( hint_text, fn )
    self.nav_right = { hint_text = hint_text, fn = fn }
end

function DeckScreen:SetOtherDeckButton( cards, title, subtitle, text )
    -- For when we want the screen to have a link to a different deck
    self.other_deck_data = {
        CARDS = cards,
        TITLE = title,
        SUBTITLE = subtitle,
        TEXT = text
    }
    if self.other_deck_data then
        if not self.other_deck_button then
            self.other_deck_button = self.bottom_left:AddChild( Widget.IconButton( self.other_deck_data.TITLE, function() self:ShowOtherDeck() end ) )
            self.other_deck_button:SetIcon( self.card_class == Widget.NegotiationCard and global_images.buynegotiation or global_images.buycombat )
            self.other_deck_button:LayoutBounds( "after", "center", self.close_button )
                :Offset( SPACING.M1, 0 )
            self.bottom_left:LayoutBounds( "center", "above", 0, 60 )
        end
    end
end

function DeckScreen:ShowOtherDeck()
    if self.other_deck_data then
        local cards = self.other_deck_data.CARDS
        local select_fn = self.select_fn
        local card_class = self.card_class
        local on_end_fn = self.on_end_fn
        local locked = self.locked
        local title = self.other_deck_data.TITLE
        local subtitle = self.other_deck_data.SUBTITLE
        local text = self.other_deck_data.TEXT
        local screen = (self._class)( cards, select_fn, card_class, on_end_fn, locked )
        screen:Refresh( cards )
        screen:SetTitles( title, subtitle, text )

        local own_cards = self.cards
        local own_title = self.title
        local own_subtitle = self.subtitle
        local own_text = self.text
        screen:SetOtherDeckButton( own_cards, own_title, own_subtitle, own_text )
        TheGame:FE():InsertScreen( screen )
        self:OnClickClose()
    end
end

function DeckScreen:OnControlModeChange( cm, device_id )
    DeckScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.card_grid:ShowBar( SCROLLBAR.IF_NEEDED )
        self.card_grid:SetDesc( LOC"UI.DRAFT_SCREEN.UPGRADES_PREVIEW_INFO" )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.card_grid:ShowBar( SCROLLBAR.NEVER )
        self.card_grid:SetDesc( nil )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.card_grid:ShowBar( SCROLLBAR.NEVER )
        self.card_grid:SetDesc( nil )
        self.bottom_left:Hide()
        local focus = self:GetDefaultFocus()
        if focus then focus:SetFocus() end
    end
end

function DeckScreen:OnScreenModeChange( sm )
    DeckScreen._base.OnScreenModeChange( self, sm )

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end

function DeckScreen:GetDefaultFocus()
    return self.card_grid:GetDefaultFocus() or self.close_button
end

function DeckScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function DeckScreen:OnOpen()
    DeckScreen._base.OnOpen( self )

    self.screen_loop = AUDIO:CreateEventInstance(SoundEvents.card_select_screen_open)
    self.screen_loop:Start()

    -- Animate in the cards after a delay
    self.card_grid:PrepareAnimation()
    self:DoTaskInTime( "fade_in_cards", 0.1, function()
        self.card_grid:AnimateIn()

        local focus = self:GetDefaultFocus()
        if TheGame:FE():GetControlMode() == CONTROL_MODE.GAMEPAD and focus then focus:SetFocus() end
    end )
end

function DeckScreen:OnClose()
    DeckScreen._base.OnClose( self )

    if self.sidebar and self.sidebar:IsOnStack() then
        self:GetFE():PopScreen( self.sidebar )
    end

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil

    if self.on_end_fn then
        self.on_end_fn( self )
    end
    AUDIO:PlayEvent("event:/ui/select_cards/close_decks")
end

function DeckScreen:OnClickClose()
    AUDIO:PlayEvent( SoundEvents.card_select_screen_close )

    if self.select_fn then
        self.select_fn( self )
    end

    self:Close()
end

function DeckScreen:Close()
    if self:IsOnStack() then
        self:FadeOut()
    end
end

function DeckScreen:PreviewUpgrade( widget )
    local preview_only = true
    local popup = Screen.UpgradeCardPopup( widget.card, widget.card.owner, preview_only, function( upgraded_card )
        
    end )
    TheGame:FE():PushScreen( popup )
end

function DeckScreen:DoUpgrade( widget )
    local preview_only = self.locked == true
    local popup = Screen.UpgradeCardPopup( widget.card, widget.card.owner, preview_only, function( upgraded_card )
        if upgraded_card then
            self:ShowUpgrade( widget, upgraded_card )
        end
    end )
    TheGame:FE():PushScreen( popup )
end

function DeckScreen:ShowUpgrade( widget, new_card, on_done_fn )
    
    if self.removed then --the user can click really fast and dismiss the screen early. in which case, it has closed out from under us, so skip the next presentation.
        return
    end

    if not is_instance( widget, Widget.CardWidget ) then
        -- This is a Card, not a CardWidget
        -- Find this widget in the grid
        local widgets = self.card_grid:GetCardWidgets()
        for k, c_w in ipairs( widgets ) do
            if c_w.card == widget then
                widget = c_w
                break
            end
        end
    end

    local duration = 0.2
    local card_scale = widget:GetScale()

    -- Add a glow layer behind the card
    local glow_bg = widget.contents:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_glow.tex" ) ) )
        :SetHiddenBoundingBox( true )
        :SendBehind( widget.bg )
        :SetTintColour( UICOLOURS.TITLE )
        :SetTintAlpha( 0 )

    -- Show the card animating into its upgraded counterpart
    
    
    widget:RunUpdater(
        Updater.Series{
            Updater.Parallel{
                Updater.Do( function() AUDIO:PlayEvent(SoundEvents.level_up) end ),
                Updater.Ease( function(v) glow_bg:SetTintAlpha(v) end, 0, 0.6, duration, easing.linear),
                Updater.Ease( function(v) widget:SetScale(v,v) end, card_scale*1, card_scale*1.1, duration, easing.linear),
            },
            Updater.Wait( duration/2 ),
            Updater.Parallel{
                Updater.Ease( function(v) widget:SetScale(v,v) end, card_scale*1.2, card_scale*1, duration/3, easing.linear),
                Updater.Ease( function(v) glow_bg:SetScale(v,v) end, 1, 1.3, duration/3, easing.linear),
                Updater.Ease( function(v) glow_bg:SetTintAlpha(v) end, 0.6, 1, duration/3, easing.linear),
                Updater.Do( function() widget:RefreshCard( new_card, nil ) end ),
            },
            Updater.Do( function()
                glow_bg:ScaleTo( 1.5, 1, duration )
                glow_bg:AlphaTo( 0, duration*4, easing.linear, function() glow_bg:Remove() end )
            end ),
            Updater.Do( function() if widget.sub_card_info then widget.sub_card_info:Update(widget) end end ),
            Updater.Wait( 1.0 ),
            Updater.Do( function() if on_done_fn then on_done_fn() end end ),
        }
    )
end

function DeckScreen:ShowUpgradeAndClose( widget, new_card )
    self.close_button:Hide()
    if widget then widget:SetOnClick(nil) end

    self:ShowUpgrade( widget, new_card, 
        function()
            self:Close()
        end )
end

function DeckScreen:ShowRemoval( widget )
    self.animating_out = true
    if self.close_button then
        self.close_button:Hide()
    end
    self.card_grid:SetOnCardClickFn( nil )
    self.card_grid:RemoveCard( widget, function()
        self:Close()
    end )
end

function DeckScreen:OnScreenResize(w,h)
    DeckScreen._base.OnScreenResize(self,w,h)

    self:Layout()
end

function DeckScreen:Layout()
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Update the floor grid
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize()
    local target_h = target_w/g_w*g_h -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Position the character
    self.character:SetPos( -RES_X*0.35, -350 )

    -- Get MainOverlay's top bar height
    local overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    local overlay_topbar_height = overlay and overlay:GetTopBarHeight() or 0

    -- Calculate card grid size
    local left_offset = RES_X*0.13 -- How much the grid should cross to the left of the screen center
    local left_padding = 100 -- How much space should there be between the left edge of the scroll area and the first card (allows space for glow/scaling without clipping)
    local left_screen_padding = 15
    self.card_grid:SetLeftPadding( left_padding )
    self.card_grid:SetSize( self.screen_w/2 + left_offset + left_padding, self.screen_h - overlay_topbar_height )
    self.card_grid:SetPos( self.screen_w/4 - left_offset/2 - left_padding/2 - left_screen_padding, -overlay_topbar_height/2 )
end

function DeckScreen:GetDebugEnv( debug_env )
    debug_env.cards = self.cards
end

------------------------------------------------------------------------------------------
-- Shows informative labels under a card, displaying if the card is ready to upgrade, 
-- if it's an incepted card, or if it's out of energy

local SubCardInfo = class( "Widget.SubCardInfo", Widget )

local label_w = 290
local label_h = 48

function SubCardInfo:init( card_widget )
    SubCardInfo._base.init( self )

    self:Update( card_widget )
end

function SubCardInfo:Update( card_widget )

    self.card_widget = card_widget
    local first_visible_label

    -- Check if the card is out of charges
    local charges, max_charges = self.card_widget.card:GetCharges()
    if charges and charges == 0 then
        if self.charges_label then
            self.charges_label:Show()
        else
            self.charges_label = self:AddChild( 
                Widget.SubCardLabel( label_w, label_h, 
                    engine.asset.Texture( "UI/ic_charges.tex" ), 
                    UICOLOURS.PENALTY, 
                    LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFO_CHARGES" ) )
            self.charges_label:SetToolTip( LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFOTOOLTIP_CHARGES" )
            self.charges_label:SendToBack()
        end
        if not first_visible_label then first_visible_label = self.charges_label end
    else
        if self.charges_label then self.charges_label:Hide() end
    end

    -- Check if the card is ready to upgrade
    if card_widget.card:UpgradeReady() then
        if self.upgrade_label then
            self.upgrade_label:Show()
            self.upgrade_label.icon:RunUpdater(
                Updater.Loop{
                    Updater.Ease( function(y) self.upgrade_label.icon:SetPos(nil, y) end, 3, 0, 0.7, easing.inOutQuad),
                    Updater.Ease( function(y) self.upgrade_label.icon:SetPos(nil, y) end, 0, 3, 0.2, easing.inOutQuad),
                }
            )
        else
            self.upgrade_label = self:AddChild( 
                Widget.SubCardLabel( label_w, label_h, 
                    engine.asset.Texture( "UI/ic_upgrade.tex" ), 
                    UICOLOURS.OBJECTIVE, 
                    LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFO_UPGRADE" ) )
            self.upgrade_label:SetToolTip( LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFOTOOLTIP_UPGRADE" )
            self.upgrade_label:SendToBack()
            self.upgrade_label.icon:RunUpdater(
                Updater.Loop{
                    Updater.Ease( function(y) self.upgrade_label.icon:SetPos(nil, y) end, 3, 0, 0.7, easing.inOutQuad),
                    Updater.Ease( function(y) self.upgrade_label.icon:SetPos(nil, y) end, 0, 3, 0.2, easing.inOutQuad),
                }
            )
        end
        if not first_visible_label then first_visible_label = self.upgrade_label end
        card_widget:SetOutline( UICOLOURS.OBJECTIVE )
    else
        if self.upgrade_label then 
            self.upgrade_label:Hide()

        end
        card_widget:SetOutline( nil )
    end

    -- Check if the card should show its card_class
    if card_widget.show_class_label then
        if self.cardclass_label then
            self.cardclass_label:Show()
        else
            local icon, colour, text, tooltip
            if is_instance( card_widget.card, Negotiation.Card ) then
                icon = engine.asset.Texture ("UI/ic_subcard_negotiation.tex")
                colour = UICOLOURS.NEGOTIATION
                text = LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFO_CLASS_NEGOTIATION"
                tooltip = LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFOTOOLTIP_CLASS_NEGOTIATION"
            elseif is_instance( card_widget.card, Battle.Card ) then
                icon = engine.asset.Texture ("UI/ic_subcard_fight.tex")
                colour = UICOLOURS.FIGHT
                text = LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFO_CLASS_BATTLE"
                tooltip = LOC"UI.SHOW_DECK_SCREEN.SUBCARDINFOTOOLTIP_CLASS_BATTLE"
            end
            self.cardclass_label = self:AddChild( Widget.SubCardLabel( label_w, label_h, icon, colour, text ) )
            self.cardclass_label:SetToolTip( tooltip )
        end
        if not first_visible_label then first_visible_label = self.cardclass_label end
    else
        if self.cardclass_label then self.cardclass_label:Hide() end
    end

    -- Layout labels
    if first_visible_label then
        first_visible_label:SetPos( 0, 0 )
        local previous_widget = first_visible_label
        for i, widget in ipairs( self.children ) do
            if widget ~= first_visible_label and widget:IsShown() then
                widget:SetHiddenState( false )
                widget:LayoutBounds( "center", "below", previous_widget ):Offset( 0, -4 )
                previous_widget = widget
            end
        end
    end

end

------------------------------------------------------------------------------------------
-- Shows the currently selected card's details

local SubCardLabel = class( "Widget.SubCardLabel", Widget )

local icon_size = 32
local text_size = 28
local padding = 8

function SubCardLabel:init( w, h, icon, color, text )
    SubCardLabel._base.init( self )

    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture( "UI/subcard_info_label.tex" ) ) )
        :SetTintColour( color )
        :SetSize( w, h )
        :Bloom( 0.15 )
        :Hide()
    self.icon = self:AddChild( Widget.Image( icon ) ):SetSize( icon_size, icon_size )
        :SetTintColour( color )
        :Bloom( 0.15 )
    self.label = self:AddChild( Widget.Label( "title", text_size ) )
        :SetText( text )
        :SetTintColour( color )
        :SetAutoSize( w - icon_size - padding*3 )
        :Bloom( 0.15 )

    self.icon:LayoutBounds( "left", "center", self.bg ):Offset( padding+4, 0 )
    self.label:LayoutBounds( "after", "center", self.icon ):Offset( padding, 0 )
end

function SubCardLabel:Refresh( text )
    self.label:SetText( text )
        :LayoutBounds( "after", "center", self.icon ):Offset( padding, 0 )
end

function SubCardLabel:SetHiddenState( hidden )
    if hidden then
        self.icon:SetTintAlpha(0)
        self.label:SetTintAlpha(0)
    else
        self.icon:SetTintAlpha(1)
        self.label:SetTintAlpha(1)
    end
end

------------------------------------------------------------------------------------------
-- Shows the player's current negotiation deck

local NegotiationDeckScreen = class( "Screen.NegotiationDeck", DeckScreen)

function NegotiationDeckScreen:init( locked )
    local negotiation_defs = require "negotiation/negotiation_defs"
    local cards = TheGame:GetGameState():GetPlayerAgent().negotiator:GetCards()

    NegotiationDeckScreen._base.init( self, cards, nil, Widget.NegotiationCard, nil, locked )
    self.owner = self.owner or TheGame:GetGameState():GetPlayerAgent()
    self:SetTitles( loc.format( LOC"UI.SHOW_DECK_SCREEN.NEGOTIATION_DECK_TITLE", self.owner:GetName() ) )
end

------------------------------------------------------------------------------------------
-- Shows the player's current battle deck

local BattleDeckScreen = class( "Screen.BattleDeck", DeckScreen)

function BattleDeckScreen:init( locked )
    local battle_defs = require "battle/battle_defs"
    local cards = TheGame:GetGameState():GetPlayerAgent().battler:GetCards()

    BattleDeckScreen._base.init( self, cards, nil, Widget.BattleCard, nil, locked )
    self.owner = self.owner or TheGame:GetGameState():GetPlayerAgent()
    self:SetTitles( loc.format( LOC"UI.SHOW_DECK_SCREEN.BATTLE_DECK_TITLE", self.owner:GetName() ) )
end

------------------------------------------------------------------------------------------
-- Shows the player's current draw deck

local DrawDeckScreen = class( "Screen.DrawDeckScreen", Screen.DeckScreen)

DrawDeckScreen.CONTROL_MAP = shallowcopy( DeckScreen.CONTROL_MAP )

table.arrayadd( DrawDeckScreen.CONTROL_MAP,
{
    {
        control = Controls.Digital.VIEW_DRAW_DECK,
        fn = function(self)
            self:OnClickClose()
            return true
        end,
    },
})

------------------------------------------------------------------------------------------
-- Shows the player's current discard deck

local DiscardDeckScreen = class( "Screen.DiscardDeckScreen", Screen.DeckScreen)

DiscardDeckScreen.CONTROL_MAP = shallowcopy( DeckScreen.CONTROL_MAP )

table.arrayadd( DiscardDeckScreen.CONTROL_MAP,
{
    {
        control = Controls.Digital.VIEW_DISCARDS_DECK,
        fn = function(self) 
            self:OnClickClose()
            return true
        end,
    },
})

------------------------------------------------------------------------------------------
-- Shows the player's current battle deck

local CardDetailsPopup = class( "Screen.CardDetailsPopup", Widget.Screen )

CardDetailsPopup.CONTROL_MAP = 
{
    {
        control = Controls.Digital.DECK_CARD_UPGRADE,
        test = function(self) return self.card:UpgradeReady() and not self.card:IsHatched() end,
        fn = function(self) self:OnClickUpgrade() return true end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.SHOW_DECK_SCREEN.CONTROLS_RIGHT_UPGRADE", Controls.Digital.DECK_CARD_UPGRADE))
        end
    },
    {
        control = Controls.Digital.DECK_CARD_PREVIEWUPGRADE,
        test = function(self) return self.card:HasUpgrades() and (self.details_mode == CARD_DETAILS_MODE.LOCKED or ( self.details_mode == CARD_DETAILS_MODE.DECK and not self.card:UpgradeReady() ) ) end,
        fn = function(self) self:OnClickPreview() return true end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.SHOW_DECK_SCREEN.CONTROLS_RIGHT_PREVIEW", Controls.Digital.DECK_CARD_PREVIEWUPGRADE))
        end
    },
    {
        control = Controls.Digital.DECK_CARD_TRASH,
        test = function(self) return self.card:CanTrash() end,
        fn = function(self) self:OnClickTrash() return true end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.SHOW_DECK_SCREEN.CONTROLS_RIGHT_TRASH", Controls.Digital.DECK_CARD_TRASH))
        end
    },
    {
        -- If there's something in focus
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            local focus = TheGame:FE():GetFocusWidget()
            return focus and focus:GetDescendant( Widget.CardWidget )
        end,
        fn = function(self) TheGame:FE():ClearFocusWidget() return true end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CANCEL", Controls.Digital.MENU_CANCEL))
        end
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self) return self.close_button and self.close_button:IsShown() end,
        fn = function(self) self:OnClickClose() return true end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end
    },
    {
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },
    {
        test = function(self)
            return TheGame:FE():IsGamepadMode() and self.scroll:HasScrolling()
        end,
        hint = function(self, left, right) 
            table.insert( left, loc.format( LOC"UI.CONTROLS.SCROLL" ))
        end,
    },
}

-- DECK shows the current card's status and allows you to upgrade it, etc
-- LOCKED shows the current card's status, but doesn't allow upgrades (shown in battle/negotiations)
-- CODEX shows generic card info, and an option to see all its upgrades
CARD_DETAILS_MODE = MakeEnum{ "DECK", "LOCKED", "CODEX" }

function CardDetailsPopup:init( card, details_mode )
    CardDetailsPopup._base.init( self )

    AUDIO:PlayEvent("event:/ui/card_upgrade/upgrade_popup")

    self.card = card
    self.details_mode = details_mode or CARD_DETAILS_MODE.CODEX
    
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
    self.is_popup = true

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Prepare sizes
    self.height = 760 -- This is max height
    self.card_scale = 1.2
    self.contents_w = 840
    self.desc_w = (self.contents_w-SPACING.M1)*0.7
    self.keywords_w = (self.contents_w-SPACING.M1)*0.3
    self.keywords = {}
    self.keyword_card_scale = 0.62

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.MONEY )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Background
    self.panel = self:AddChild( Widget.WidePanel( self.height, UICOLOURS.FIGHT ) )
        :PrepareAnimation()

    -- Contents
    self.scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( 0 )
        :SetVirtualTopMargin( SPACING.M1 )
        :SetVirtualBottomMargin( SPACING.M1 )
        :SetScrollBarOuterMargin( 0 )
        :ShowBar( SCROLLBAR.NEVER )
        :SetSize( self.contents_w+SPACING.M1*4, self.height )
    self.content = self.scroll:AddScrollChild( Widget() )
        :SetTintAlpha( 0 )

    -- Titles
    self.header_block = self.content:AddChild( Widget() )
    self.hitbox = self.header_block:AddChild( Widget.SolidBox( self.contents_w, 1, 0xffff000 ))
    self.title = self.header_block:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_NAME, "TITLE" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :SetAutoSize( self.contents_w )
        :SetWordWrap( true )
        :Bloom( 0.15 )
    self.rarity_icon = self.header_block:AddChild( Widget.Image( engine.asset.Texture ("UI/graft_rarity.tex") ) )
        :SetSize( FONT_SIZE.ITEM_SUBTITLE, FONT_SIZE.ITEM_SUBTITLE )
        :Bloom( 0.05 )
    self.subtitle = self.header_block:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, "SUBTITLE" ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :SetAutoSize( self.contents_w )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    self.xp_bar = self.header_block:AddChild( Widget.CardDetailsUpgradeBar( self.contents_w ) )
        :SetOnFocusFn( function(card) self:OnCardFocus(card) end)
    self.charges_block = self.header_block:AddChild( Widget.CardDetailsCharges( self.contents_w ) )

    -- Description panel
    self.description_block = self.content:AddChild( Widget() )
    self.description_background = self.description_block:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetTintColour( UICOLOURS.FIGHT )
        :SetBloom( 0.15 )
    self.description_container = self.description_block:AddChild( Widget() )
    self.desc = self.description_container:AddChild( Widget.Label( "body", FONT_SIZE.BODY_TEXT, "" ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :SetHoverLinkGlyphColour( UICOLOURS.TITLE )
        :SetAutoSize( self.desc_w - SPACING.M1*2 )
        :SetWordWrap( true )
        :LeftAlign()
        :Bloom( 0.1 )
    self.cards_container_height = 270
    self.cards_container_overlap = 40 -- How much the upgrades will overlap on the header content
    self.cards_container_block = self.description_container:AddChild( Widget.CardListBlock( self.desc_w-5, self.cards_container_height ) )
        :HideBackground()
        :SetCardScale( 0.45 )
        :SetScaleOnFocus( false )
    self.flavour_text = self.description_container:AddChild( Widget.Label( "body", FONT_SIZE.BODY_TEXT, "" ) )
        :SetAutoSize( self.desc_w - SPACING.M1*2 )
        :SetWordWrap( true )
        :LeftAlign()
        :Bloom( 0.05 )

    -- Keywords panel
    self.keywords_block = self.content:AddChild( Widget.KeywordsList( self.keywords_w ) ) -- Widget with a list of keywords

    -- Card
    self.battle_card_widget = self:AddChild( Widget.BattleCard() )
        :SetScale( self.card_scale )
        :SetOutline( UICOLOURS.FIGHT )
        :ShowToolTipOnFocus()
    self.negotiation_card_widget = self:AddChild( Widget.NegotiationCard() )
        :SetScale( self.card_scale )
        :SetOutline( UICOLOURS.NEGOTIATION )
        :ShowToolTipOnFocus()

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    -- Right buttons
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.buttons = self.bottom_right:AddChild( Widget() )
    self.trash_button = self.buttons:AddChild( Widget.IconButton( LOC"UI.SHOW_DECK_SCREEN.TRASH", function() self:OnClickTrash() end, 300 ) ):SetIcon( engine.asset.Texture( "UI/ic_nav_trash.tex") )

    self:Refresh( self.card )
    self:Layout()
end

function CardDetailsPopup:OnGainTop()
    CardDetailsPopup._base.OnGainTop( self )

    self:AddRobotOption(self.close_button)

    if not self.animated_in then
        self.animated_in = true
        -- Animate in
        self.panel:FadeIn( function() self:AnimateIn() end )
    end

    -- To keep track of gamepad scroll
    self:StartUpdating()
end

function CardDetailsPopup:OnLoseTop()
    CardDetailsPopup._base.OnLoseTop( self )

    -- Stop keeping track of gamepad scroll
    self:StopUpdating()
end

function CardDetailsPopup:AnimateIn()
    self.content:AlphaTo( 1, 0.3, easing.inOutQuad )
end

function CardDetailsPopup:OnControlModeChange( cm, device_id )
    CardDetailsPopup._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD
    or cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
        self.bottom_right:Show()
        self.buttons:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
        self.bottom_right:Hide()
        self.buttons:Hide()
    end
end

function CardDetailsPopup:OnScreenModeChange( sm )
    CardDetailsPopup._base.OnScreenModeChange( self, sm )

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    if sm == SCREEN_MODE.MONITOR then
    end
    if sm == SCREEN_MODE.SMALL
    or sm == SCREEN_MODE.TV then
    end

    self.content:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.scroll:SetSize( (self.contents_w+SPACING.M1*4)*LAYOUT_SCALE[sm], self.height )

    self:Layout()
end

function CardDetailsPopup:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

function CardDetailsPopup:OnCardFocus( card )
    if card then
        if is_instance( card, Battle.Card ) then
            self.battle_card_widget:RefreshCard( card, nil, Widget.CardWidget.STYLE.UPGRADE )
            :SetOutline( nil )
            :ShowParticles( false )
            return self
        elseif is_instance( card, Negotiation.Card ) then
            self.negotiation_card_widget:RefreshCard( card, nil, Widget.CardWidget.STYLE.UPGRADE )
            :SetOutline( nil )
            :ShowParticles( false )
            return self
        end
    end
    -- If no card was focused, or the sent card wasn't a card, show the default one
    if is_instance( self.card, Battle.Card ) then
        self.battle_card_widget:RefreshCard( self.card )
            :SetOutline( UICOLOURS.FIGHT )
            :ShowParticles( true )
            :DisableToolTip( true )
            :SetOnRightClickFn( nil )
    elseif is_instance( self.card, Negotiation.Card ) then
        self.negotiation_card_widget:RefreshCard( self.card )
            :SetOutline( UICOLOURS.NEGOTIATION )
            :ShowParticles( true )
            :DisableToolTip( true )
            :SetOnRightClickFn( nil )
    end
end

function CardDetailsPopup:OnClickUpgrade()
    local preview_only = false
    local popup = Screen.UpgradeCardPopup( self.card, self.card.owner, preview_only, function( upgraded_card )
        self:Close()
        if upgraded_card then
            local deck_screen = TheGame:FE():FindScreen( Screen.DeckScreen )
            if deck_screen then deck_screen:ShowUpgrade( self.card, upgraded_card ) end
        end
    end )
    TheGame:FE():PushScreen( popup )
end

function CardDetailsPopup:OnClickPreview()
    local preview_only = true
    local popup = Screen.UpgradeCardPopup( self.card, self.card.owner, preview_only )
    TheGame:FE():PushScreen( popup )
end

function CardDetailsPopup:OnClickTrash()
    TheGame:FE():PushScreen( 
        Screen.YesNoPopup(loc.format(LOC"UI.SHOW_DECK_SCREEN.TRASH_CONFIRM_TITLE", self.card:GetName()), LOC"UI.SHOW_DECK_SCREEN.TRASH_CONFIRM_BODY")
            :SetFn(
                function(res) 
                    if res == Screen.YesNoPopup.YES then 
                        self.panel:FadeOut( function()
                            self:Close()

                            if is_instance( self.card, Negotiation.Card ) then
                                self.card:GetOwner().negotiator:RemoveCardByID( self.card.id, self.card.userdata )
                                SendMetricsData("REMOVED_NEGOTIATION_CARD", {SELECTED_ID = self.card.id})
                            else
                                self.card:GetOwner().battler:RemoveCardByID( self.card.id, self.card.userdata )
                                SendMetricsData("REMOVED_BATTLE_CARD", {SELECTED_ID = self.card.id})
                            end
                            AUDIO:PlayEvent("event:/ui/select_cards/remove_card")

                            local deck_screen = TheGame:FE():FindScreen( Screen.DeckScreen )
                            if deck_screen then deck_screen:Refresh( deck_screen.cards ) end
                        end
                    )
                end
            end
        )
    )
end

function CardDetailsPopup:OnClickClose()
    AUDIO:PlayEvent( SoundEvents.card_select_screen_close )

    self.panel:FadeOut( function() self:Close() end )
end

function CardDetailsPopup:Close()
    if self:IsOnStack() then
        -- self:FadeOut()
        self:GetFE():PopScreen( self )
    end
end

function CardDetailsPopup:GetDefaultFocus()
    -- Prevent a card from being selected by default, since they change the master card on focus.
    -- But still let the player focus on one after the screen is open
    if not self.screen_initialized then
        self.screen_initialized = true
        return nil
    else
        return self.xp_bar:GetDefaultFocus() or self.cards_container_block:GetDefaultFocus()
    end
end
 
function CardDetailsPopup:Refresh( card )
    self.card = card

    -- Update card preview
    if is_instance( self.card, Battle.Card ) then
        self.negotiation_card_widget:Hide()
        self.battle_card_widget:Show()
            :RefreshCard( card )
            :SetOutline( UICOLOURS.FIGHT )
            :ShowParticles( true )
            :DisableToolTip( true )
            :SetOnRightClickFn( nil )
        self.panel:SetColour( UICOLOURS.FIGHT )
            :SetIcon( engine.asset.Texture("ui/card_type_fight.tex") )
        self.title:SetGlyphColour( UICOLOURS.FIGHT )
        self.description_background:SetTintColour( UICOLOURS.FIGHT )
        self.grid:SetTintColour( UICOLOURS.FIGHT )
    end
    if is_instance( self.card, Negotiation.Card ) then
        self.battle_card_widget:Hide()
        self.negotiation_card_widget:Show()
            :RefreshCard( card )
            :SetOutline( UICOLOURS.NEGOTIATION )
            :ShowParticles( true )
            :DisableToolTip( true )
            :SetOnRightClickFn( nil )
        self.panel:SetColour( UICOLOURS.NEGOTIATION )
            :SetIcon( engine.asset.Texture("ui/card_type_negotiation.tex") )
        self.title:SetGlyphColour( UICOLOURS.NEGOTIATION )
        self.description_background:SetTintColour( UICOLOURS.NEGOTIATION )
        self.grid:SetTintColour( UICOLOURS.NEGOTIATION )
    end

    self:ShowCardInfo()
    self:Layout()

    TheGame:FE():RefreshControlOptions()
    return self      
end

function CardDetailsPopup:ShowCardInfo()
    -- Set rarity and type
    if self.card then
        -- Update title and subtitle
        local rarity_color = MakeColourString( HexColour( CARD_RARITY_COLOURS[ self.card.rarity or CARD_RARITY.UNIQUE ] ) )
        local rarity_name = GetCardRarityString( self.card.rarity )
        local card_type = GetCardType( self.card )
        local type_color = MakeColourString( HexColour( GetCardTypeColour( card_type ) ) )
        local type_name = GetCardTypeString( card_type )
        local subtitle_text = loc.format( LOC"UI.SHOW_DECK_SCREEN.DETAILS_SUBTITLE", rarity_color, rarity_name, type_color, type_name )
        if self.card.base_def then
            subtitle_text = loc.format( LOC"UI.SHOW_DECK_SCREEN.DETAILS_SUBTITLE_UPGRADED", rarity_color, rarity_name, type_color, type_name )
        else
            subtitle_text = loc.format( LOC"UI.SHOW_DECK_SCREEN.DETAILS_SUBTITLE", rarity_color, rarity_name, type_color, type_name )
        end
        local name_key = self.card.def:GetLocNameKey()
        self.title:SetText( loc.upper( LOC( name_key ) ) )
        self.subtitle:SetText( subtitle_text )
        self.rarity_icon:SetTexture( global_images.card_rarity_icons[self.card.rarity] )

        -- Update xp/upgrades.
        if self.card:HasUpgrades() and not self.card:IsHatched() then
            self.xp_bar:Refresh( self.card ):Show()

            -- Set the button to upgrade or preview upgrades
            if self.card:UpgradeReady() and self.details_mode ~= CARD_DETAILS_MODE.LOCKED then
                self.xp_bar:SetButtonFn( function() self:OnClickUpgrade() end )
            elseif self.card:HasUpgrades() then
                self.xp_bar:SetButtonFn( function() self:OnClickPreview() end )
            end
        else
            self.xp_bar:Hide()
        end

        -- Update charges
        if self.card:GetCharges() then
            self.charges_block:Refresh( self.card ):Show()
        else
            self.charges_block:Hide()
        end

        -- Update description
        self.desc:SetText( self.battle_card_widget:IsShown() and self.battle_card_widget:GetDescText() or self.negotiation_card_widget:GetDescText() or "" )

        -- Update flavour text
        local flavour = self.card:GetFlavourText()
        if flavour then self.flavour_text:SetText( loc.format( "<i><#FLAVOUR_TEXT>{1}</></>", flavour ) ) end

        -- Update keywords from the current card
        local card_keywords = self.battle_card_widget:IsShown() and self.battle_card_widget:GetCardFeatures() or self.negotiation_card_widget:GetCardFeatures() or {}
        
        -- If this card references other cards on its keywords, get their ids to display them on the desc panel
        local card_ids = {}
        for id, keyword in pairs( card_keywords ) do
            if keyword.card then table.insert( card_ids, keyword.card.id ) end
        end

        -- If this card spawns other cards, get their ids too
        if self.card.pool_cards then
            for k, card_id in ipairs( self.card.pool_cards ) do
                table.insert( card_ids, card_id )
            end
        end

        -- Display all of those cards on the card block
        self.cards_container_block:SetCardIDs( card_ids )

        -- Get keywords for all the cards on the card block, and add them to our existing keywords
        table.keywordsaddunique( card_keywords, self.cards_container_block:GetKeywords() )

        -- List our keywords on the side panel
        self.keywords_block:Refresh( card_keywords )

        -- Update buttons
        self.trash_button:SetShown( self.card:CanTrash() )

        self:LayoutContents()
    end
    return self
end

function CardDetailsPopup:UpdateCardImageState( widget )
    if widget.enabled and ( widget.hover or widget.focus ) then
        if is_instance( widget, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseover, "position", widget:GetPanPos())
        elseif is_instance( widget, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseover, "position", widget:GetPanPos())
        end
        widget:SetOutline( UICOLOURS.FOCUS, true )
        widget:ScaleTo( nil, self.keyword_card_scale*1.05, 0.2, easing.outQuad )
    else
        if is_instance( widget, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseoff, "position", widget:GetPanPos())
        elseif is_instance( widget, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseoff, "position", widget:GetPanPos())
        end
        widget:SetOutline( widget.card:UpgradeReady() and UICOLOURS.OBJECTIVE or nil, true )
        widget:ScaleTo( nil, self.keyword_card_scale, 0.2, easing.outQuad )
    end
end

function CardDetailsPopup:Layout()
    -- Calculate content width
    local card_w = self.battle_card_widget:GetScaledSize()
    local scroll_w = self.scroll:GetScaledSize()
    local total_w = card_w + scroll_w + SPACING.M1
    local vertical_offset = 20

    -- Position cards
    self.battle_card_widget:LayoutBounds( "left", "center", -total_w/2, vertical_offset+10 )
    self.negotiation_card_widget:LayoutBounds( "left", "center", -total_w/2, vertical_offset+10 )

    -- Position scroll
    self.scroll:LayoutBounds( "right", nil, total_w/2, 0 )
    self.content:LayoutBounds( "left", "top", 0, 0 )

    -- Position buttons
    self.buttons:StackChildrenRow( SPACING.M05 )
        :LayoutBounds( "right", "above", -60, 60 )

    -- Resize the panel if there's little content
    local c_w, c_h = self.content:GetScaledSize()
    c_h = c_h + SPACING.M1*2
    c_h = math.min( c_h, self.height )
    self.scroll:SetSize( nil, c_h )
        :SetPos( nil, vertical_offset )
    self.panel:SetHeight( c_h )
        :SetPos( nil, vertical_offset )

    return self
end

function CardDetailsPopup:LayoutContents()
    -- Layout header
    self.title:LayoutBounds( "left", "center", self.hitbox ):Offset( 0, 0 )
    self.rarity_icon:LayoutBounds( "left", "below", self.title ):Offset( 0, -3 )
    self.subtitle:LayoutBounds( "after", "center", self.rarity_icon ):Offset( 6, 2 )
    self.xp_bar:LayoutBounds( "left", "below", self.rarity_icon ):Offset( 0, -15 )
    if self.no_xp_label then
        self.no_xp_label:LayoutBounds( "left", "below", self.rarity_icon ):Offset( 0, -15 )
    end
    self.charges_block:LayoutBounds( "left", "below", self.xp_bar:IsShown() and self.xp_bar or self.rarity_icon ):Offset( 0, -15 )

    -- Layout flavour text
    self.flavour_text:LayoutBounds( "left", "below", self.desc ):Offset( 0, -SPACING.M1/2 )
    -- Layout description cards
    self.cards_container_block:LayoutBounds( "left", "below", self.flavour_text ):Offset( 0, self.cards_container_overlap )
        :SetShown( self.cards_container_block:HasCards() )
    -- Adjust background
    local w, h = self.description_container:GetSize()
    self.description_background:SetSize( self.desc_w, h + SPACING.M1*1.5 )
    self.description_container:LayoutBounds( "left", "top", self.description_background )
        :Offset( SPACING.M1, -SPACING.M1/2-5 )
    -- Layout description cards
    self.cards_container_block:LayoutBounds( "center", "bottom", self.description_background ):Offset( 3, 0 )
    -- Position description block
    self.description_block:LayoutBounds( "left", "below", self.header_block ):Offset( 0, -SPACING.M1 )

    -- Layout keywords
    self.keywords_block:SetShown( self.keywords_block:HasKeywords() )
        :LayoutBounds( "right", "below", self.header_block )
        :Offset( 0, -SPACING.M1 )

    self.content:LayoutBounds( "left", "top", 0, 0 )

    return self
end

function CardDetailsPopup:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.scroll:ScrollUp()
        return true
    end
end

------------------------------------------------------------------------------------------
-- Shows upgrade status

local CardDetailsUpgradeBar = class( "Widget.CardDetailsUpgradeBar", Widget )

function CardDetailsUpgradeBar:init( width )
    CardDetailsUpgradeBar._base.init( self )

    self.width = width or 300
    self.content_w = self.width - SPACING.M1*2
    self.upgrade_ready = false

    -- Sizes
    self.upgrades_height = 270
    self.upgrades_overlap = 40 -- How much the upgrades will overlap on the header content
    self.bg_height = 70 + self.upgrades_height - self.upgrades_overlap

    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetSize( self.width, self.bg_height )
        :SetBloom( 0.15 )

    self.bar = self:AddChild( Widget.SegmentedBar( self.content_w, 8 ) )
        :SetOnTexture( engine.asset.Texture("UI/cardupgrade_bar.tex"), 6, 6, 12, 12 )
        :SetOffTexture( engine.asset.Texture("UI/cardupgrade_bar.tex"), 6, 6, 12, 12 )
        :SetSpaceBetweenPips( 4 )
        :Bloom( 0.3 )

    self.upgrades_block = self:AddChild( Widget.CardListBlock( self.width-5, self.upgrades_height ) )
        :HideBackground()
        :SetCardScale( 0.45 )
        :SetScaleOnFocus( false )

    self.title_label = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :Bloom( 0.05 )
    self.value_label = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :Bloom( 0.05 )
    self.upgrade_button = self:AddChild( Widget.ThinButton( 250, 28, LOC"UI.UPGRADE_CARD_POPUP.TITLE" ) )

end

function CardDetailsUpgradeBar:SetButtonFn( fn )
    self.upgrade_button:SetOnClickFn( fn )
    return self
end

function CardDetailsUpgradeBar:SetOnFocusFn( fn )
    self.on_focus_fn = fn
    self.upgrades_block:SetOnFocusFn( self.on_focus_fn )
    return self
end

function CardDetailsUpgradeBar:Refresh( card )
    local current_xp, max_xp = card:GetXP()
    if max_xp then
        -- Card may be a 'clone' and not have xp, but in general it would.
        current_xp = current_xp or 0

        self.bar:SetProgress( max_xp, current_xp ):Show()

        if current_xp >= max_xp then
            -- Use bright colors if this card is ready to upgrade
            self.upgrade_ready = true
            local r, g, b, a = HexColour( UICOLOURS.QUEST )
            self.bar:SetOnTint( r, g, b, a )
                :SetOffTint( r, g, b, a )
            self.bg:SetTintColour( UICOLOURS.QUEST )
            self.title_label:SetGlyphColour( UICOLOURS.QUEST )
                :SetText( LOC"UI.CARD_WIDGET.UPGRADE_LABEL_READY" )
            self.value_label:SetGlyphColour( UICOLOURS.QUEST )
                :SetText( loc.format( LOC"UI.CARD_WIDGET.UPGRADE_VALUE", current_xp, max_xp ) )
            self.upgrade_button:SetColour( UICOLOURS.QUEST )
                :SetText( LOC"UI.SHOW_DECK_SCREEN.UPGRADE" )
                :Show()
        else
            self.upgrade_ready = false
            local r, g, b, a = HexColour( UICOLOURS.OBJECTIVE )
            self.bar:SetOnTint( r, g, b, a )
                :SetOffTint( r, g, b, 0.1 )
            self.bg:SetTintColour( UICOLOURS.OBJECTIVE )
                :SetTintAlpha( 0.6 )
            self.title_label:SetGlyphColour( UICOLOURS.OBJECTIVE )
                :SetText( LOC"UI.CARD_WIDGET.UPGRADE_LABEL" )
            self.value_label:SetGlyphColour( UICOLOURS.OBJECTIVE )
                :SetText( loc.format( LOC"UI.CARD_WIDGET.UPGRADE_VALUE", current_xp, max_xp ) )
            self.upgrade_button:SetColour( UICOLOURS.OBJECTIVE )
                :SetText( LOC"UI.SHOW_DECK_SCREEN.PREVIEW" )
                :Hide()
        end
    else
        self.upgrade_ready = false
        self.bar:Hide()
        self.bg:SetTintColour( UICOLOURS.OBJECTIVE )
            :SetTintAlpha( 0.6 )
        self.title_label:SetGlyphColour( UICOLOURS.OBJECTIVE )
            :SetText( LOC"UI.SHOW_DECK_SCREEN.DETAILS_NO_XP" )
        self.value_label:Hide()
        self.upgrade_button:Hide()
    end

    -- Get all the ids for this card's potential upgrades
    local upgrade_ids = card.def.upgrade_ids or table.empty
    -- Show them on the upgrades block
    self.upgrades_block:SetCardIDs( upgrade_ids, Widget.CardWidget.STYLE.UPGRADE )
        :SetShown( #upgrade_ids > 0 )

    -- Layout
    self:Layout()

    return self
end

function CardDetailsUpgradeBar:Layout()
    self.title_label:LayoutBounds( "left", "top", self.bg ):Offset( SPACING.M1, -8 )
    if self.upgrade_button:IsShown() then
        self.upgrade_button:LayoutBounds( "right", "top", self.bg ):Offset( -SPACING.M1, -11 )
        self.value_label:LayoutBounds( "before", "center", self.upgrade_button ):Offset( -9, 1 )
    else
        self.value_label:LayoutBounds( "right", "top", self.bg ):Offset( -SPACING.M1, -8 )
    end
    self.bar:LayoutBounds( "center", nil, self.bg )
        :LayoutBounds( nil, "below", self.title_label )
        :Offset( 0, -6 )
    self.upgrades_block:LayoutBounds( "center", "bottom", self.bg )
        :Offset( 4, 0 )

    return self
end

function CardDetailsUpgradeBar:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
end

function CardDetailsUpgradeBar:GetDefaultFocus()
    return self.upgrades_block:GetDefaultFocus()
end

function CardDetailsUpgradeBar:OnControlModeChange( cm, device_id )
    CardDetailsUpgradeBar._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.upgrade_button:SetShown(self.upgrade_ready)
    end
    if cm == CONTROL_MODE.TOUCH then
        self.upgrade_button:SetShown(self.upgrade_ready)
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.upgrade_button:Hide()
    end

    self:Layout()
end

------------------------------------------------------------------------------------------
-- Shows charges

local ThinButton = class( "Widget.ThinButton", Widget.Clickable )

function ThinButton:init( width, height, text )
    ThinButton._base.init( self )

    self.width = width or 250
    self.height = height or 50
    self.text = text or ""

    -- Button bg
    self.bg_border = 16
    self.bg_border_scale = 0.5
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/dropdown_inner.tex") ) )
    self.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.03 )
    local nineslice_w, nineslice_h = self.width + self.bg_border*2*self.bg_border_scale, self.height + self.bg_border*2*self.bg_border_scale
    self.bg:SetSize( nineslice_w, nineslice_h )

    -- Button label
    self.text = self:AddChild( Widget.Label( "body", FONT_SIZE.SMALL_TEXT, text ) )
        :SetGlyphColour( UICOLOURS.DARK_BODYTEXT )
        :LayoutBounds( "left", "center", self.bg ):Offset( 10, 1 )

    -- Button icon
    self.icon = self:AddChild( Widget.Image( engine.asset.Texture("UI/cardxp_viewupgrade.tex"), 12, 12 ) )
        :LayoutBounds( "right", "center", self.bg ):Offset( -10, 0 )
end

function ThinButton:SetText( text )
    self.text:SetText( text or "" )
        :LayoutBounds( "left", "center", self.bg ):Offset( 10, 1 )
    return self
end

function ThinButton:SetColour( colour )
    self.bg:SetTintColour( colour or UICOLOURS.SUBTITLE )
    return self
end

function ThinButton:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
        self.bg:Bloom( 0.05 )
    else
        self:ScaleTo( nil, 1.0, 0.2, easing.outQuad )
        self.bg:Bloom( 0.03 )
    end
    return self
end

function ThinButton:OnGainFocus()
    ThinButton._base.OnGainFocus(self)
    self:UpdateImage()
end

function ThinButton:OnLoseFocus()
    ThinButton._base.OnLoseFocus(self)
    self:UpdateImage()
end

function ThinButton:OnGainHover()
    ThinButton._base.OnGainHover(self)
    self:UpdateImage()
end

function ThinButton:OnLoseHover()
    ThinButton._base.OnLoseHover(self)
    self:UpdateImage()
end



------------------------------------------------------------------------------------------
-- Shows charges

local CardDetailsCharges = class( "Widget.CardDetailsCharges", Widget )

function CardDetailsCharges:init( width )
    CardDetailsCharges._base.init( self )

    self.width = width or 300
    self.content_w = self.width - SPACING.M1*2

    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetTintColour( UICOLOURS.CARD_CHARGES )
        :SetTintAlpha( 0.6 )
        :SetSize( self.width, 50 )
        :SetBloom( 0.15 )

    self.title_label = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetGlyphColour( UICOLOURS.CARD_CHARGES )
        :Bloom( 0.05 )
    self.value_label = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetGlyphColour( UICOLOURS.CARD_CHARGES )
        :Bloom( 0.05 )
    self.icon = self:AddChild( Widget.Image( engine.asset.Texture("UI/ic_charges_full.tex") ) )
        :SetSize( 30, 30 )
        :Bloom( 0.05 )
end

function CardDetailsCharges:Refresh( card )
    local charges, max_charges = card:GetCharges()
    self.title_label:SetText( LOC"UI.CARD_WIDGET.CHARGES_LABEL" )
    self.value_label:SetText( loc.format( LOC"UI.CARD_WIDGET.CHARGES_VALUE", charges, max_charges ) )

    self.title_label:LayoutBounds( "left", "center", self.bg ):Offset( SPACING.M1, 1 )
    self.icon:LayoutBounds( "right", "center", self.bg ):Offset( -SPACING.M1*0.6, 0 )
    self.value_label:LayoutBounds( "before", "center", self.icon ):Offset( 0, 1 )

    self.icon:SetRotation(-90*DEGREES)

    return self
end

------------------------------------------------------------------------------------------
-- Shows a vertical list of keywords, with a background

local KeywordsList = class( "Widget.KeywordsList", Widget )

function KeywordsList:init( width )
    KeywordsList._base.init( self )

    self.width = width or 250
    self.background = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetBloom( 0.15 )
    self.keywords_container = self:AddChild( Widget() ) -- Vertical list of keywords
    self.keywords_list = {} -- Will contain keywords objects for the listed ones

end

function KeywordsList:SetWidth( width )
    self.width = width or self.width or 250
    self:Layout()
    return self
end

-- Clears the list and adds the keywords provided
function KeywordsList:Refresh( keywords_list )
    -- Make sure the input list only has unique elements
    local clean_list = {}
    table.keywordsaddunique( clean_list, keywords_list )

    -- Clear old keywords
    self.keywords_container:DestroyAllChildren()
    self.keywords_list = {}

    -- Go through the keywords and list them
    for id, keyword in pairs( clean_list ) do
        if keyword.name and keyword.desc then
            self.keywords_container:AddChild( Widget.Label( "tooltip", 24 ) )
                :SetAutoSize( self.width - SPACING.M1*2 )
                :OverrideLineHeight( 20 )
                :SetWordWrap( true )
                :LeftAlign()
                :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
                :SetText( loc.format( LOC"UI.CARD_WIDGET.KEYWORD_ENTRY", keyword.name, keyword.desc ) )
                :Bloom(0.05)
            table.insert( self.keywords_list, keyword )
        end
    end

    -- Layout all the things!
    self:Layout()
    return self
end

-- Does not clear existing keywords, and adds unique ones in the provided list
function KeywordsList:AddKeywords( keywords_list )
    table.keywordsaddunique( self.keywords_list, keywords_list )
    self:Refresh( self.keywords_list )
    return self
end

function KeywordsList:HasKeywords()
    return self.keywords_container and self.keywords_container.children and #self.keywords_container.children > 0
end

function KeywordsList:Layout()
    -- Layout keywords
    self.keywords_container:StackChildren( SPACING.M1/2 )

    -- Adjust background
    local w, h = self.keywords_container:GetSize()
    self.background:SetSize( self.width, h + SPACING.M1*1.5 )
    self.keywords_container:LayoutBounds( "left", "top", self.background )
        :Offset( SPACING.M1, -SPACING.M1/2-5 )

    return self
end

function table.keywordsaddunique( dst, src )
    for i, keyword in pairs(src) do
        local exists = false
        for k, v in pairs(dst) do
            if keyword.id == v.id then exists = true end
        end
        if not exists then table.insert( dst, keyword ) end
    end
end

------------------------------------------------------------------------------------------
-- Shows a sliding sidebar with filters. Don't use this directly.
-- Use the Negotiation or Battle variants, so that the MainOverlay nav buttons work correctly

local DeckFiltersSidebar = class( "Screen.DeckFiltersSidebar", Screen.SlidingSidebarScreen )

function DeckFiltersSidebar:init()
    DeckFiltersSidebar._base.init( self )

end

function DeckFiltersSidebar:Refresh( filter_widgets, sorting_widget )
    self.filter_widgets = filter_widgets or {}
    self.sorting_widget = sorting_widget or nil

    -- Remove any old contents
    self.contents:DestroyAllChildren()

    -- Add all the filter widgets
    for k, w in ipairs( self.filter_widgets ) do
        if w.parent then
            w:Reparent( self.contents )
                :SetScale( 1 )
        else
            self.contents:AddChild( w )
        end
    end

    -- And the sorting widget
    if self.sorting_widget then
        if self.sorting_widget.parent then
            self.sorting_widget:Reparent( self.contents )
                :SetScale( 1 )
        else
            self.contents:AddChild( self.sorting_widget )
        end
    end
 
    self:Layout()
    return self
end

function DeckFiltersSidebar:SetOnCloseFunction( on_close )
    self.on_close_fn = on_close
    return self
end

function DeckFiltersSidebar:Layout()
    -- Align all
    for k, w in ipairs( self.contents.children ) do
        w:SetWidth( Screen.SlidingSidebarScreen.SIDEBAR_WIDTH - SPACING.M1*4.5 )
            :LayoutBounds( "left", "below" )
            :Offset( 0, -SPACING.M1 )
    end

    self.contents:LayoutBounds( "left", "top", 0, 0 )
        :Offset( SPACING.M1, 0 )

    DeckFiltersSidebar._base.Layout( self )
    return self
end

function DeckFiltersSidebar:Close()
    -- Reparent all the filter widgets back
    if self.on_close_fn then self.on_close_fn() end

    DeckFiltersSidebar._base.Close( self )
end

------------------------------------------------------------------------------------------
-- Shows a sliding sidebar with filters. Negotiation variant.
-- Use this instead of the generic one, so the MainOverlay nav buttons stay highlighted when this opens
local NegotiationDeckFiltersSidebar = class( "Screen.NegotiationDeckFiltersSidebar", Screen.DeckFiltersSidebar )

function NegotiationDeckFiltersSidebar:init()
    NegotiationDeckFiltersSidebar._base.init( self )
end

------------------------------------------------------------------------------------------
-- Shows a sliding sidebar with filters. Battle variant.
-- Use this instead of the generic one, so the MainOverlay nav buttons stay highlighted when this opens
local BattleDeckFiltersSidebar = class( "Screen.BattleDeckFiltersSidebar", Screen.DeckFiltersSidebar )

function BattleDeckFiltersSidebar:init()
    BattleDeckFiltersSidebar._base.init( self )
end
