require "ui/widgets/tooltip"
require "ui/widgets/conditionimage"
require "ui/widgets/peoplecompendium"
require "ui/screens/relationshipsscreen"

---------------------------------------------------------------------------------------------

local ConditionEntry = class( "Widget.ConditionEntry", Widget )

function ConditionEntry:init( width, fighter )
    ConditionEntry._base.init( self )
    self.width = width or DEFAULT_TT_WIDTH
    self.fighter = fighter
    self.text = self:AddChild( Widget.Label("tooltip", 24) )
        :SetAutoSize( self.width )
        :SetWordWrap( true )
        :LeftAlign()
        :OverrideLineHeight( 20 )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom(0.05)
end

function ConditionEntry:RefreshCondition( condition, tooltip_widget )
    assert( self.icon == nil )

    local padding = 8
    local icon_w = 32

    local stacks = self.fighter:GetConditionStacks( condition.id )
    self.icon = self:AddChild( Widget.ConditionImage( icon_w ) )
        :SetCondition( condition.id, stacks )

    local tt = loc.format( "<#TITLE>{1}</>\n{2}", condition:GetName(), condition:GetDesc( nil, tooltip_widget ))
    self.text:SetText( tt )
        :SetAutoSize( self.icon:HasImage() and self.width-padding-icon_w or self.width )
        :LayoutBounds( "after", "top", self.icon ):Offset( self.icon:HasImage() and padding, 4 )

    return self
end

function ConditionEntry:RefreshCard( card, fighter )
    local battle_defs = require "battle/battle_defs"

    local icon
    if card.preview_icon then
        icon = card.preview_icon
    elseif card:IsFlagged( battle_defs.CARD_FLAGS.BUFF ) then
        icon = engine.asset.Texture( "UI/fightscreen_preview_buff.tex" )
    elseif card:IsFlagged( battle_defs.CARD_FLAGS.DEBUFF ) then
        icon = engine.asset.Texture( "UI/fightscreen_preview_debuff.tex" )
    else
        icon = engine.asset.Texture( "UI/fightscreen_damagepreview.tex" )
    end

    assert( self.icon == nil )
    self.icon = self:AddChild( Widget.Image( icon, 32, 32 ))

    local tt = loc.format( "<#TITLE>{1}</b>\n{2}", card:GetName(), card:GetDesc( fighter.battle ))
    self.text:SetText( tt ):LayoutBounds( "after", "top", self.icon ):Offset( 8, 4 )

    return self
end

function ConditionEntry:RefreshText( tt )
    self.text:SetText( tt )
    
    return self
end

------------------------------------------------------------------------------------------
-- Displays the health and conditions for the clicked fighter

local FighterInfoPopup = class( "Screen.FighterInfoPopup", Widget.Screen )

FighterInfoPopup.CONTROL_MAP =
{
    {
        control = Controls.Digital.GAMEPAD_RT,
        test = function(self)
            return self.death_loot:IsShown()
        end,
        fn = function(self)
            AUDIO:PlayEvent( "event:/ui/battle_deck/click_card_in_menu" )
            self.death_loot:ShowCardInfoPopup()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CARD_INFO", Controls.Digital.GAMEPAD_RT))
        end
    },
    {
        controls = { Controls.Digital.MENU_CANCEL, Controls.Digital.MENU_ALT },
        fn = function(self) 
            self:Close()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.BACK", Controls.Digital.MENU_CANCEL) )
        end,
    }
}

local MAX_PANEL_HEIGHT = 700

function FighterInfoPopup:init( selected_fighter )
    FighterInfoPopup._base.init( self )

    self.is_popup = true
    -- Set custom underlay
    self.clickable_bg = self:AddChild( Widget.Clickable() )
        :SetOnClickFn( function() self:Close() end )
        :SetOnRightClickFn( function() self:Close() end )
    self.underlay = self.clickable_bg:AddChild( Widget.Image( engine.asset.Texture( "splash/underlay.tex" )), 1 )
        :SetAnchors( "fill", "fill" )
        :SetTintAlpha( 0 )
        :AlphaTo( 0.95, 0.3, easing.outQuad )
    self:SetAnchors( "center", "center" )
    
    self.height = 540
    self.contents_w = 750
    self.conditions_w = self.contents_w*0.6-SPACING.M1*1.5
    self.keywords_w = self.contents_w*0.4-SPACING.M1*1.5
    self.character_scale = 1
    self.keywords = {}

    -- Background
    self.bg = self:AddChild( Widget.WidePanel( self.height ) )
        :BlockMouse()
    -- Anim
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :SetOutline( 8, UICOLOURS.SUBTITLE )
        :Hide()
    self.vertical_offset = 800 -- How much vertical clearance the scissor should leave for the character's head to overlap the panel edge
    -- Contents
    self.title = self:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TABS ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :LeftAlign()
        :Bloom( 0.1 )
    self.difficulty = self:AddChild( Widget() )
        :SetToolTip( LOC"UI.FIGHT_SCREEN.FIGHTER_INFO_DIFFICULTY" )
    for k=1, 5 do
        self.difficulty:AddChild( Widget.Image( engine.asset.Texture( "UI/teammember_star_full.tex" ) ) )
            :SetSize( 22, 22 )
            :Bloom( 0.1 )
            :LayoutBounds( "after", "center" )
            :Offset( 1, 0 )
    end
    self.scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( 0 )
        :SetVirtualTopMargin( 20 )
        :SetVirtualBottomMargin( 100 )
        :SetScrollBarOuterMargin( 0 )
        :ShowBar( SCROLLBAR.NEVER )
        :SetSize( self.contents_w+SPACING.M1*4, self.height )
    self.contents = self.scroll:AddScrollChild( Widget() )
    self.health_block = self.contents:AddChild( Widget.FighterInfoHealthBlock( self.contents_w, selected_fighter ) ) -- Widget with a custom labeled health bar
    self.custom_conditions_block = self.contents:AddChild( Widget() ) -- Contains custom condition widgets this fighter might have
    self.conditions_block = self.contents:AddChild( Widget() ) -- Widget with a background and list of conditions
    self.conditions_background = self.conditions_block:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetTintColour( UICOLOURS.FIGHT )
        :SetBloom( 0.15 )
    self.conditions_container = self.conditions_block:AddChild( Widget() ) -- Vertical list of fighter conditions
    self.keywords_block = self.contents:AddChild( Widget.KeywordsList( self.keywords_w ) ) -- Widget with a background and list of keywords
    self.death_loot = self.contents:AddChild( Widget.FighterInfoDeathLoot( self.conditions_w ) )
    self.agent_boon = self.contents:AddChild( Widget.FighterInfoSocialGraft( self.conditions_w, 170 ) )
    self.agent_bane = self.contents:AddChild( Widget.FighterInfoSocialGraft( self.conditions_w, 170 ) )

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.cancel_btn = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.DIALOGS.BACK", function() self:Close() end ) )
    self.cancel_btn:SetIcon( global_images.close )
    self.cancel_btn:LayoutBounds( "left", "above", 60, 60 )

    self:Refresh( selected_fighter )
end

function FighterInfoPopup:Refresh( fighter )

    self.selected_fighter = fighter

    self.contents:SetTintAlpha(0)

    -- Reset keywords
    self.keywords = {}

    -- Update fighter's name
    self.title:SetText( fighter:GetName() )
        :SetGlyphColour( fighter.team.id == "BLUE" and UICOLOURS.TITLE or UICOLOURS.PENALTY )
        :LayoutBounds( "left", nil, self.contents )

    -- Update difficulty
    for k, star in ipairs( self.difficulty.children ) do
        star:SetTintColour( self.selected_fighter.team.id == "BLUE" and UICOLOURS.TITLE or UICOLOURS.PENALTY )
        if self.selected_fighter.agent.boss then
            star:SetTexture( engine.asset.Texture( "UI/teammember_star_full.tex" ) )
        elseif ( self.selected_fighter.agent:GetCombatStrength() + 0.5 ) == k then
            star:SetTexture( engine.asset.Texture( "UI/teammember_star_half.tex" ) )
        elseif self.selected_fighter.agent:GetCombatStrength() >= k then
            star:SetTexture( engine.asset.Texture( "UI/teammember_star_full.tex" ) )
        else
            star:SetTexture( engine.asset.Texture( "UI/teammember_star_empty.tex" ) )
        end
    end
    -- Only show stars for the non-player characters
    self.difficulty:SetShown( self.selected_fighter.agent ~= TheGame:GetGameState():GetPlayerAgent() )

    -- Update anim
    local fight_mode = false
    self.character:Refresh( self.selected_fighter.agent, fight_mode )
        :Play( "idle_neutral", true )
        :SetScale( (self.selected_fighter.team.id == "BLUE" and 1 or -1)*self.character_scale, self.character_scale )
        :SetOutline( 4, self.selected_fighter.team.id == "BLUE" and UICOLOURS.TITLE or UICOLOURS.PENALTY )
    -- Calculate where the character scissor should go
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )
    local scroll_w, scroll_h = self.scroll:GetSize()
    local scroll_x, scroll_y = self.scroll:GetPos()
    self.bg:SetColour( self.selected_fighter.team.id == "BLUE" and UICOLOURS.TITLE or UICOLOURS.PENALTY )

    -- Clear old conditions
    self.conditions_container:DestroyAllChildren()
    self.custom_conditions_block:DestroyAllChildren()

    -- Get fighter's conditions
    local conditions = self.selected_fighter:GetConditions()
    for id, condition in pairs( conditions ) do
        if condition.fighter_info_widget then
            -- This condition has a custom widget to display itself
            self.custom_conditions_block:AddChild( condition.fighter_info_widget(self.selected_fighter, condition, self.conditions_w, true ) )
                :Layout()
        elseif not condition.hidden then
            self.conditions_container:AddChild( Widget.ConditionEntry(self.conditions_w - SPACING.M1*2, self.selected_fighter) )
                :RefreshCondition( condition, self )
        end
    end

    -- Layout conditions
    self.conditions_container:StackChildren( SPACING.M1/2 )
        :Offset( 0, -SPACING.M1/2 )
    -- Adjust background
    local w, h = self.conditions_container:GetSize()
    self.conditions_background:SetSize( self.conditions_w, h + SPACING.M1*1.5 )
    self.conditions_container:LayoutBounds( "left", "top", self.conditions_background )
        :Offset( SPACING.M1, -SPACING.M1/2-5 )

    -- Layout custom widgets
    self.custom_conditions_block:StackChildren( SPACING.M1/2 )

    -- Death loot
    self.death_loot:Refresh( self.selected_fighter.agent )

    -- Refresh keywords
    self.keywords_block:Refresh( {} )
    -- Add health block keywords
    self.keywords_block:AddKeywords( self.health_block:GetKeywords() )
    -- Add death loot keywords
    self.keywords_block:AddKeywords( self.death_loot:GetKeywords() )
    -- Add other keywords
    self.keywords_block:AddKeywords( self.keywords )

    -- Get this agent's relationship to the player
    local player = TheGame:GetGameState():GetPlayerAgent()
    local agent = self.selected_fighter.agent
    local rel = self.selected_fighter.agent:GetRelationship( player )
    -- Social grafts
    local boon_id = agent:GetSocialGraft( RELATIONSHIP.LOVED )
    local boon_graft
    if player and rel == RELATIONSHIP.LOVED then
        boon_graft = player.graft_owner:GetSocialGraft( agent )
    else
        boon_graft = boon_id and GraftInstance( boon_id )
    end
    self.agent_boon:Refresh( boon_graft, rel == RELATIONSHIP.LOVED, agent )
        :SetShown( player ~= agent and boon_graft and agent:IsSentient() )
    local bane_id = agent:GetSocialGraft( RELATIONSHIP.HATED )
    local bane_graft
    if player and rel == RELATIONSHIP.HATED then
        bane_graft = player.graft_owner:GetSocialGraft( agent )
    else
        bane_graft = bane_id and GraftInstance( bane_id )
    end
    self.agent_bane:Refresh( bane_graft, rel == RELATIONSHIP.HATED, agent )
        :SetShown( player ~= agent and bane_graft and agent:IsSentient() )

    -- Layout scroll
    self.health_block:LayoutBounds( "left", "top", 0, 0 )
    self.custom_conditions_block:SetShown( #self.custom_conditions_block.children > 0 )
        :LayoutBounds( "left", "below", self.health_block )
        :Offset( SPACING.M1, -SPACING.M1 )
    self.conditions_block:SetShown( #self.conditions_container.children > 0 )
        :LayoutBounds( "left", "below", self.custom_conditions_block:IsShown() and self.custom_conditions_block or self.health_block )
        :Offset( self.custom_conditions_block:IsShown() and 0 or SPACING.M1, -SPACING.M1 )
    self.keywords_block:SetShown( self.keywords_block:HasKeywords() )
        :LayoutBounds( "right", "below", self.health_block )
        :Offset( -SPACING.M1, -SPACING.M1 )
    self.death_loot:LayoutBounds( "left", "below", self.conditions_block )
        :Offset( 0, -SPACING.M1 )
    self.agent_boon:LayoutBounds( "left", "below", self.death_loot )
        :Offset( 0, -SPACING.M1 )
    self.agent_bane:LayoutBounds( "left", "below", self.agent_boon )
        :Offset( 0, -SPACING.M1 )
    self.contents:LayoutBounds( "left", "top", 0, 0 )
        :Offset( 40, 0 )

    return self
end

function FighterInfoPopup:ShowFighter( fighter )
    local x, y = self.character:GetPos()
    self.character:SetPos( self.selected_fighter.team.id == "BLUE" and x-100 or x+100, nil )
        :Show()
        :MoveTo( x, nil, 0.2, easing.outQuad )
        
    self.contents:AlphaTo( 1, 0.1, easing.outQuad )

    return self
end


function FighterInfoPopup:Close()
    self:AlphaTo( 0, 0.1, easing.outQuad, function() self:GetFE():PopScreen(self) end )
    return self
end

function FighterInfoPopup:PostFeature( feature_id, name, desc )
    table.insert( self.keywords, { id = feature_id, name = name, desc = desc } )
end

function FighterInfoPopup:PostCard( feature_id, card )
    -- self.card_keywords[ feature_id ] = { card = card }
end

function FighterInfoPopup:OnOpen()
    FighterInfoPopup._base.OnOpen( self )

    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    self.bg:FadeIn( function() self:ShowFighter() end )
end

function FighterInfoPopup:OnGainTop()
    FighterInfoPopup._base.OnGainTop( self )

    self.scroll:SetFocus()
end

function FighterInfoPopup:OnControlModeChange( cm, device_id )
    FighterInfoPopup._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD
    or cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
    end
end

function FighterInfoPopup:OnScreenModeChange( sm )
    FighterInfoPopup._base.OnScreenModeChange( self, sm )

    self.contents:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Calculate panel size
    local w, h = self.contents:GetScaledSize()
    h = h + SPACING.M1*2
    self.height = math.min( MAX_PANEL_HEIGHT, h )

    self.bg:SetHeight( self.height )
    self.scroll:SetSize( self.contents_w*LAYOUT_SCALE[sm]+SPACING.M1*4, self.height )
        :RefreshView()
    self.title:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "left", nil, self.contents )
        :LayoutBounds( nil, "above", self.bg )
        :Offset( 0, SPACING.M1*0.4 )
    self.difficulty:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "right", nil, self.contents )
        :LayoutBounds( nil, "above", self.bg )
        :Offset( 0, SPACING.M1*0.6 )
    self.contents:LayoutBounds( "left", "top", 0, 0 )
        :Offset( 40, 0 )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Position character anim
    self.character:SetPos( 0, 0 )
        :LayoutBounds( self.selected_fighter.team.id == "BLUE" and "before" or "right", nil, self.scroll )
        :Offset( self.selected_fighter.team.id == "BLUE" and -60 or 60, -RES_Y/2+200 )
end

function FighterInfoPopup:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

------------------------------------------------------------------------------------------
-- Shows a character's social graft

local FighterInfoHealthBlock = class( "Widget.FighterInfoHealthBlock", Widget )

function FighterInfoHealthBlock:init( width, fighter )
    FighterInfoHealthBlock._base.init( self )

    -- Backing
    self.background = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetTintColour( UICOLOURS.HEALTH )
        :SetBloom( 0.2 )

    self.contents = self:AddChild( Widget() )

    -- Get stats
    local health = fighter:GetStat( COMBAT_STAT.HEALTH ) or 0
    local min_health, max_health = fighter:GetStatBounds( COMBAT_STAT.HEALTH )
    local morale = fighter:GetStat( COMBAT_STAT.MORALE ) or 0
    local defend = fighter:GetConditionStacks( "DEFEND" ) or 0

    -- Header title
    self.header = self.contents:AddChild( Widget() )
    self.icon = self.header:AddChild( Widget.Image( engine.asset.Texture("UI/health_transparent.tex") ) )
        :SetSize( 30, 30 )
        :Bloom(0.1)
    self.health_label = self.header:AddChild( Widget.Label("body", FONT_SIZE.BODY_TEXT ) )
        :SetGlyphColour(UICOLOURS.HEALTH)
        :SetText( loc.format( LOC"UI.FIGHT_SCREEN.FIGHTER_INFO_HEALTH", health, max_health ) )
        :Bloom(0.1)

    -- Health bar
    self.healthbar_container = self.contents:AddChild( Widget() ) -- Contains the healthbar and the info labels
    self.healthbar = self.healthbar_container:AddChild( Widget.HealthBar(fighter) )
        :SetNoTextMode()
        :SetVals( health, morale )
        :SetDefend( defend )
        :SetFixedSize( width-SPACING.M1*3, 20 )

    -- Health bar info labels
    local label_w = 400
    local surrender_dmg
    if fighter:HasMorale() and fighter:IsActive() then
        local battle_defs = require "battle/battle_defs"
        local health = fighter:GetHealth()
        local surrender_health = fighter:GetSurrenderHealth()
        if surrender_health < health and surrender_health > 0 then
            surrender_dmg = math.max( 1, health - surrender_health )
        end
    end
    self.panic_label = self.healthbar_container:AddChild( Widget.FighterInfoHealthLabel( label_w ) )
        :SetColor( UICOLOURS.MORALE )
        :SetIcon( engine.asset.Texture("UI/ic_info_panic.tex") )
        :SetText( loc.format( LOC"UI.FIGHT_SCREEN.FIGHTER_INFO_PANIC_LABEL", fighter.agent:GetName(), surrender_dmg ) )
        :SetShown( surrender_dmg )
    self.panic_connector = self.healthbar_container:AddChild( Widget.SolidBox( 3, 20, UICOLOURS.MORALE ) )
        :Bloom( 0.1 )
        :Hide()
    self.defend_label = self.healthbar_container:AddChild( Widget.FighterInfoHealthLabel( label_w ) )
        :SetColor( UICOLOURS.DEFEND )
        :SetIcon( engine.asset.Texture("UI/ic_info_defend.tex") )
        :SetText( loc.format( LOC"UI.FIGHT_SCREEN.FIGHTER_INFO_DEFEND_LABEL", fighter.agent:GetName(), defend ) )
        :SetShown( defend and defend > 0 )
    self.defend_connector = self.healthbar_container:AddChild( Widget.SolidBox( 3, 20, UICOLOURS.DEFEND ) )
        :Bloom( 0.1 )
        :Hide()

    -- Position labels
    local icon_offset = self.panic_label:GetIconSize()/2
    if self.panic_label:IsShown() then
        local panic_bar_widget = self.healthbar:GetMoraleBar()
        -- Try aligning the label to the right side of the bar
        self.panic_label:LayoutBounds( "right", nil, panic_bar_widget )
            :LayoutBounds( nil, "below", self.healthbar )
            :Offset( icon_offset-2, -SPACING.M1/2 )
        -- Check if we're past the left edge of the healthbar
        local label_x1, label_y1, label_x2, label_y2 = self.panic_label:GetWorldBoundingBox()
        local bar_x1, bar_y1, bar_x2, bar_y2 = self.healthbar:GetWorldBoundingBox()
        if label_x1 < bar_x1 then
            -- Make this label align to the left instead, and reposition it
            self.panic_label:SetLeft()
                :LayoutBounds( "after", nil, panic_bar_widget )
                :LayoutBounds( nil, "below", self.healthbar )
                :Offset( -icon_offset-2, -SPACING.M1/2 )
        end
        -- Align the connector
        local icon_widget = self.panic_label:GetIcon()
        local icon_x1, icon_y1, icon_x2, icon_y2 = icon_widget:GetWorldBoundingBox()
        local bar_x1, bar_y1, bar_x2, bar_y2 = panic_bar_widget:GetWorldBoundingBox()
        local connector_height = math.abs( bar_y1 - icon_y2 )
        self.panic_connector:Show()
            :SetSize( 3, connector_height )
            :LayoutBounds( nil, "above", icon_widget )
            :LayoutBounds( "right", nil, panic_bar_widget )
            :Offset( 0, 4 )
    end
    if self.defend_label:IsShown() then
        local defend_bar_widget = self.healthbar:GetDefendBar()
        -- Try aligning the label to the right side of the bar
        self.defend_label:LayoutBounds( "right", nil, defend_bar_widget )
            :LayoutBounds( nil, "above", self.healthbar )
            :Offset( icon_offset-2, SPACING.M1/2 )
        -- Check if we're past the left edge of the healthbar
        local label_x1, label_y1, label_x2, label_y2 = self.defend_label:GetWorldBoundingBox()
        local healthbar_x1, healthbar_y1, healthbar_x2, healthbar_y2 = self.healthbar:GetWorldBoundingBox()
        if label_x1 < healthbar_x1 then
            -- Make this label align to the left instead, and reposition it
            self.defend_label:SetLeft()
                :LayoutBounds( "after", nil, defend_bar_widget )
                :LayoutBounds( nil, "above", self.healthbar )
                :Offset( -icon_offset-2, SPACING.M1/2 )
        end
        -- Align the connector
        local icon_widget = self.defend_label:GetIcon()
        local icon_x1, icon_y1, icon_x2, icon_y2 = icon_widget:GetWorldBoundingBox()
        local bar_x1, bar_y1, bar_x2, bar_y2 = defend_bar_widget:GetWorldBoundingBox()
        local connector_height = math.abs( bar_y2 - icon_y1 )
        self.defend_connector:Show()
            :SetSize( 3, connector_height )
            :LayoutBounds( nil, "below", icon_widget )
            :LayoutBounds( "right", nil, defend_bar_widget )
            :Offset( 0, -4 )
    end

    -- Layout
    self.healthbar_container:SetPos( 0, 0 )
    self.health_label:LayoutBounds( "after", "center", self.icon )
        :Offset( 4, 1 )
    self.header:LayoutBounds( "left", "above", self.healthbar_container )
        :Offset( -4, SPACING.M1/2 )
    local w, h = self.contents:GetSize()
    self.background:SetSize( width, h + SPACING.M1*2 )
        :LayoutBounds( "center", "center", self.contents )
end

function FighterInfoHealthBlock:GetKeywords()
    local keywords = {}
    if self.panic_label:IsShown() then
        local condition = Content.GetBattleCardFeature( "SURRENDER" )
        table.insert( keywords, { id = condition.id, name = condition:GetLocalizedName(), desc = condition:GetFeatureTooltip() } )
    end
    if self.defend_label:IsShown() then
        local condition = Content.GetBattleCardFeature( "DEFEND" )
        table.insert( keywords, { id = condition.id, name = condition:GetLocalizedName(), desc = condition:GetLocalizedDesc() } )
    end
    return keywords
end

------------------------------------------------------------------------------------------
-- Shows a character's social graft

local FighterInfoHealthLabel = class( "Widget.FighterInfoHealthLabel", Widget )

function FighterInfoHealthLabel:init( width, color, icon, text )
    FighterInfoHealthLabel._base.init( self )

    self.width = width or 200
    self.padding = 5
    self.icon_size = 20

    self.icon = self:AddChild( Widget.Image( engine.asset.Texture("UI/health_transparent.tex") ) )
        :SetTintColour( color or UICOLOURS.SUBTITLE )
        :SetSize( self.icon_size, self.icon_size )
        :Bloom(0.1)
    self.label = self:AddChild( Widget.Label( "body", FONT_SIZE.SMALL_TEXT ) )
        :SetGlyphColour( color or UICOLOURS.SUBTITLE )
        :SetWordWrap( true )
        :SetAutoSize( self.width - self.padding - self.icon_size )
        :SetText( text or "" )
        :Bloom(0.1)

    self:Layout()
end

function FighterInfoHealthLabel:SetWidth( width )
    self.width = width or 200
    self.label:SetAutoSize( self.width - self.padding - self.icon_size )
    self:Layout()
    return self
end

function FighterInfoHealthLabel:SetColor( color )
    self.icon:SetTintColour( color or UICOLOURS.SUBTITLE )
    self.label:SetGlyphColour( color or UICOLOURS.SUBTITLE )
    return self
end

function FighterInfoHealthLabel:SetIcon( icon )
    self.icon:SetTexture( icon )
    return self
end

function FighterInfoHealthLabel:GetIcon()
    return self.icon
end

function FighterInfoHealthLabel:SetText( text )
    self.label:SetText( text or "" )
    self:Layout()
    return self
end

function FighterInfoHealthLabel:SetLeft()
    self.icon_on_left = true
    self:Layout()
    return self
end

function FighterInfoHealthLabel:GetIconSize()
    return self.icon_size
end

function FighterInfoHealthLabel:Layout()
    if self.icon_on_left then
        self.label:LeftAlign()
            :LayoutBounds( "after", "top", self.icon )
            :Offset( self.padding, 2 )
    else
        self.label:RightAlign()
            :LayoutBounds( "before", "top", self.icon )
            :Offset( -self.padding, 2 )
    end
    return self
end

------------------------------------------------------------------------------------------
-- Shows a character's death loot

local FighterInfoDeathLoot = class( "Widget.FighterInfoDeathLoot", Widget.AgentDeathLoot )

function FighterInfoDeathLoot:init( width )
    FighterInfoDeathLoot._base.init( self, width, 200 )

    self.background = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetSize( self.width, self.height+20 )
        :SetPos( -20, 10 )
        :SetTintColour( 0xFFCA78ff )
        :SetBloom( 0.15 )
        :SendToBack()

end

function FighterInfoDeathLoot:SetSize( width, height )
    FighterInfoDeathLoot._base.SetSize( self, width, height )

    if self.background then
        self.background:SetSize( self.width, self.height )
            :SetPos( -20, 10 )
    end

    return self
end

------------------------------------------------------------------------------------------
-- Shows a character's boon

local FighterInfoSocialGraft = class( "Widget.FighterInfoSocialGraft", Widget.RelationshipsScreenBoon )

function FighterInfoSocialGraft:init( width, height )
    FighterInfoSocialGraft._base.init( self, width, height )
    
    self.bg:Remove()
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetSize( self.width, self.height )
        :SetBloom( 0.15 )
        :SendToBack()

    self:SetIconSize( 70 )
end

function FighterInfoSocialGraft:Layout()
    FighterInfoSocialGraft._base.Layout( self )
    
    -- Background
    self.bg:SetSize( self.width, self.height )
        :SetTexture( self.active_boon and engine.asset.Texture("UI/left_border_frame_faint_border.tex" ) or engine.asset.Texture("UI/left_border_frame_faint.tex" ) )

    -- Layout our title
    self.label:LayoutBounds( "left", "top", self.bg )
        :Offset( SPACING.M1, -10 )

    -- Layout the Active label and icon
    self.active_icon:LayoutBounds( "right", nil, self.bg )
        :LayoutBounds( nil, "center", self.label )
        :Offset( -SPACING.M1+4, -1 )
    self.active_label:LayoutBounds( "before", "center", self.active_icon )
        :Offset( -6, 1 )

    self.icon_frame:LayoutBounds( "left", "below", self.label )
        :Offset( 0, -SPACING.M1*0.5 )
    self.icon:SetSize( self.icon_size, self.icon_size )
        :LayoutBounds( "center", "center", self.icon_frame )
        
    -- Align text
    self.subtitle
        :LayoutBounds( "left", "below", self.title )
        :Offset( 0, 4 )
    self.text_content
        :LayoutBounds( "after", "top", self.icon_frame )
        :Offset( SPACING.M1, 8 )

    return self
end