require "ui/widgets/screen"
require "fightsim"
require "battle/scenario"
require "ui/widgets/cardwidget"
require "ui/screens/fightscreenbase"
require "ui/screens/infopopup"
require "ui/widgets/cardarrangement"
require "ui/widgets/card_deck"
require "ui/widgets/timeindicator"
require "ui/widgets/targetarrow"
require "battle/battle_engine"
require "battle/damage_preview"
require "ui/widgets/unitframe"
require "ui/widgets/cardspanel"
require "ui/widgets/battlecard"
require "ui/widgets/animatedcharacterui"
require "ui/screens/fighterinfopopup"

local SMOOTH_MOUSE_TARGETING = true

local battle_defs = require "battle/battle_defs"

local CARD_FLAGS = battle_defs.CARD_FLAGS
local EVENT = battle_defs.BATTLE_EVENT

------------------------------------------------------------------------------------------
--
local EVENT_HANDLERS =
{
    [ EVENT.ACTIVATED_FIGHTER ] = function( self, fighter )
        self:AddFighter( fighter )
    end,

    [ EVENT.SETTLE_FIGHTERS ] = function( self )
        self:DealCards()
    end,

    [ EVENT.BEGIN_COMBAT ] = function( self )
        self:Delay( 0.3 )
    end,

    [ EVENT.DEACTIVATED_FIGHTER ] = function( self, fighter, cards )
        self:RemoveFighter( fighter )

        if cards then
            for i, card in ipairs( cards ) do
                local w = self:DetachCardWidget( card )
                if w then
                    self.floating_cards:AttachCardWidget( card, w )
                    w:AnimateToTrash()
                end
            end
        end
    end,

    [ EVENT.ADDED_FIGHTER ] = function( self, fighter )
        self.main_overlay = self.main_overlay or TheGame:FE():FindScreen( Screen.MainOverlay )
        if fighter:GetTeamID() == TEAM.BLUE then
            self.main_overlay:SetFriendReserves( self.battle, TEAM.BLUE )
        else
            self.main_overlay:SetEnemyReserves( self.battle, TEAM.RED )
        end
    end,

    [ EVENT.FIGHTER_REMAP_ANIMS ] = function( self, fighter, mapping )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:SetAnimMapping( mapping )
        end
    end,

    [ EVENT.UPDATE_HOME_POSITIONS ] = function( self, team )
        if team then
            self.fightsim.mgr:CalculateHomePositions( team, true )
        end
    end,
    
    [ EVENT.FIGHTER_PREVIEW ] = function( self, fighter )
        self:RefreshDamagePreviews()
    end,

    [ EVENT.CONDITION_ADDED ] = function( self, fighter, condition, stacks, source )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:AddCondition( condition, stacks, source )
        end

        if self.current_event.is_debug then
            self.cards:RefreshCards( self.battle )
        end
 
        self:RefreshDamagePreviews()
        self:SetToolTipDirty( true )
    end,

    [ EVENT.CONDITION_BLOCKED ] = function( self, fighter, condition )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:BlockedCondition()
        end
    end,

    [ EVENT.CONDITION_REMOVED ] = function( self, fighter, condition, stacks, source )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:RemoveCondition( condition, stacks, source )
        end
 
        if self.current_event.is_debug then
            self.cards:RefreshCards( self.battle )
        end

        self:RefreshDamagePreviews()
        self:SetToolTipDirty( true )
     end,

    [ EVENT.GRAFT_TRIGGERED ] = function( self, graft )
        local graft_widget = self.grafts:FindWidget(graft) or self.boons:FindWidget(graft) or self.banes:FindWidget(graft) or self.perks:FindWidget(graft)
        if graft_widget then
            graft_widget:Triggered()
        end
    end,

    [ EVENT.REFRESH_GRAFTS ] = function( self )
        self.grafts:Refresh()
        self.boons:Refresh()
        self.banes:Refresh()
    end,

    [ EVENT.DELTA_ACTION_COUNT ] =  function( self, actions, max_actions, delta, no_sound )
        self.time_indicator:AnimateNumberChange( actions, max_actions )

        if self.current_event.is_debug then
            self.cards:RefreshCards( self.battle ) -- For debug: changes card playability.
        end
    end,

    [ EVENT.DELTA_MONEY ] = function( self, fighter, delta )
        self.status_widgets[ fighter ]:ShowDeltaMoney( delta )
    end,

    [ EVENT.DELTA_STAT ] = function( self, fighter, stat, delta, value, mitigated, new_cap, delta_cap )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            local cmp = ent.cmp.AnimFighter
            if cmp.self_damaging then
                local busy, coro = ent.cmp.AnimFighter:IsBusy()
                if busy then
                    ent.cmp.AnimFighter:WaitForSequences()
                    cmp.self_damaging = nil
                end
            end
            ent.cmp.AnimFighter:ShowDeltaStat( stat, delta, value, mitigated, new_cap, delta_cap )
        end
    end,

    [ EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
        self.cardspanel:AnimateShuffle( num_cards )
    end,

    [ EVENT.START_RESOLVE ] = function( self, battle, card )
        if not card.owner.player_controlled and not card:IsFlagged( CARD_FLAGS.SPECIAL ) and not card.combo_followup then
            self:DelayAndAnimate( nil, true )
        end

        if CheckBits(card.flags, CARD_FLAGS.FLOURISH) and self.battle:GetPlayerFighter() then
            local ent = self:GetAnimFighterEntity( card.owner )
            if ent then
                ent.cmp.AnimFighter:PlayFlourish()
            end

        elseif CheckBits(card.flags, CARD_FLAGS.SKILL) then
            AUDIO:PlayEvent(SoundEvents.battle_play_maneuver_card)
        elseif CheckBits(card.flags, CARD_FLAGS.MELEE) then
            AUDIO:PlayEvent(SoundEvents.battle_play_attack_card)
        end
    end,

    [ EVENT.END_RESOLVE ] = function( self, battle, card )
    end,

    [ EVENT.CARD_TRIGGERED ] = function( self, card, deck, idx )
        local w = self:DetachCardWidget( card )
        if w then
            self.floating_cards:AttachCardWidget( card, w )
            w:AnimateChosen( self ):SendToFront()
            w:Delay( 0.5 )
            
            if card.deck then
                self:UpdateCardPlacement( card, deck, idx, deck, idx )
            end
        end
    end,
    [ EVENT.CARD_TRIGGERED_PRE ] = function( self, card )
        local w, idx = self:DetachCardWidget( card )
        if w then
            self.floating_cards:AttachCardWidget( card, w )
            w:AnimateChosen( self ):SendToFront()
        end
    end,
    [ EVENT.CARD_TRIGGERED_POST ] = function( self, card, deck, deck_idx )
        if card.deck then
            self:UpdateCardPlacement( card, nil, nil, deck, deck_idx )
        end
    end,

    [ EVENT.PRE_RESOLVE ] = function( self, battle, attack )
        if not attack.anim and not attack.card.hit_anim then
            return
        end

        self:PrepareAttackEvents( attack )

        local ent = self:GetAnimFighterEntity( attack.attacker )
        if ent == nil then
            return -- It's possible the fighter was killed during PRE_RESOLVE or START_RESOLVE.
        end

        local cmp = ent.cmp.AnimFighter

        if attack.card.combo_starter then
            cmp:DoSequence( cmp.DoAttack, cmp, attack, self)
        else
            cmp:DoAttack( attack, self )
        end
    end,

    [ EVENT.ON_RIPOSTE ] = function( self, attack, first_hit, riposted_attack, last_hit, delay_for_target )
        if first_hit and delay_for_target then
            self:DelayForFighter(attack.target)
        end
        -- target_ent is the target of the riposte (ie. the original attacker)
        self:PrepareAttackEvents( attack )

        if first_hit then
            -- Stop the riposte target from moving
            self:DelayForAttackAnim( attack.target,  riposted_attack.anim )

            local target_ent = self:GetAnimFighterEntity(attack.target)
            if target_ent and not target_ent.cmp.AnimFighter:InHomePosition() and CheckBits(attack.card.flags, CARD_FLAGS.MELEE) then
                target_ent.cmp.AnimFighter:StopMoving()
                target_ent.cmp.AnimFighter:FaceFighter( attack.attacker )
                target_ent.cmp.AnimFighter:PlayIdleAnim()
            end
        end

        local riposter_ent = self:GetAnimFighterEntity( attack.attacker )
        riposter_ent.cmp.AnimFighter:DoAttack( attack, self )
        if last_hit then
            -- riposter_ent.cmp.AnimFighter:WaitAnim()
            -- riposter_ent.cmp.AnimFighter:PlayIdleAnim()
        end
    end,

    [ EVENT.POST_RESOLVE ] = function( self, battle, attack )  
        -- Clear previews for this card.
        self.player_preview:RemovePreview( attack.card )
        self:RefreshDamagePreviews()

        self:DealCards()
    end,

    [ EVENT.DAMAGE_APPLIED ] = function( self, fighter, damage, delta, source, hit_tags )
        if not is_instance( source, Battle.Hit ) then
            if hit_tags and table.contains(hit_tags, "no_reaction") then
                -- Consume the stat change, initiate delayed floater.
                local ev = self:ConsumeEvent( function(ev) return ev[1] == EVENT.DELTA_STAT and ev[2] == fighter and ev[3] == COMBAT_STAT.HEALTH end )
                local ent = self:GetAnimFighterEntity( fighter )
                if ent and ev then
                    local cmp = ent.cmp.AnimFighter
                    cmp:StartCoroutine( function()
                        -- Waiting until not busy makes these floaters stack up and play simultaneously, which isn't great.
                        cmp.self_damaging = true
                        if cmp:IsBusy() then
                            cmp:CoroDelay( 1.5 )
                        end
                        cmp:PlayOneShotFX( "self_damage" )
                        cmp:ShowDeltaStat( table.unpack(ev, 3) )
                        cmp.self_damaging = nil
                    end)
                end

            else
                self:DelayForFighter( fighter )
                local target = self:GetAnimFighterEntity(fighter)
                if target then
                    local cmp = target.cmp
                    local function DoHitReaction()
                        if hit_tags then
                            cmp.AnimController:SetFXTrackTags( hit_tags )
                        end
                        cmp.AnimFighter:PlayAnim("hit_mid")
                        cmp.AnimFighter:WaitAnim()
                        cmp.AnimFighter:PlayIdleAnim()
                    end
                    
                    if cmp then
                        cmp.AnimFighter:DoSequence(DoHitReaction)
                    end
                end
            end
        end
    end,

    [ EVENT.END_PLAYER_TURN ] = function( self )
        self:RefreshDamagePreviews()
        self.concede_btn:AnimateOut()
        self.end_turn_btn:AnimateOut()
        self.end_turn_btn:HideFocusGlow()
    end,

    [ EVENT.BEGIN_ENEMY_TURN ] = function( self )
        self:DealCards()

        if self.battle:GetTurns() == 1 and self.battle:GetScenario():GetAdvantageTeam() == self.battle:GetEnemyTeam():GetID() then
            self.cardspanel:ShowBanner( LOC"UI.FIGHT_SCREEN.ENEMY_ADVANTAGE", nil, nil, nil, UICOLOURS.HEALTH, true )
            self:DelayAndAnimate( 1.0, true )
        else
            self:DelayAndAnimate( nil, true )
        end
    end,

    [ EVENT.BEGIN_PLAYER_TURN ] = function( self )
        self:DealCards()
        self.main_overlay:RefreshScenario( self.battle )
        self:RestoreTimeLabel()
        self:RefreshDamagePreviews()

        self.end_turn_btn:SetText( LOC"UI.CARDS.END_TURN", loc.format( LOC"UI.CARDS_PANEL.TURNS_COUNTER", self.battle:GetTurns() ) )
        self:DoTaskInTime( "delay_show_buttons", 0.7, function()
            if not self.accept_surrender_button:IsShown() then
                local can_flee = self.battle:GetScenario():CanFlee()
                if can_flee then self.concede_btn:AnimateIn() else self.concede_btn:AnimateOut() end
                self.flourish_bar:AnimateIn()
                self.end_turn_btn:AnimateIn()
                self.end_turn_btn:HideFocusGlow()
            end
        end )

        self:DelayAndAnimate( nil, true )

        if self.battle:GetTurns() == 1 and self.battle:GetScenario():GetAdvantageTeam() == self.battle:GetPlayerTeam():GetID() then
            self.cardspanel:ShowBanner( LOC"UI.FIGHT_SCREEN.PLAYER_ADVANTAGE", nil, nil, UICOLOURS.SUBTITLE, nil )
        end
        
        if TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent() then
            UIHelpers.ShowTutorials(
                    {"battle_tutorial", TheGame:GetGameState():GetPlayerAgent().def.battle_tutorial},
                    TheGame:GetGameState():GetPlayerAgent():GetContentID() == "SAL" and LOC"UI.FIGHT_SCREEN.TUTORIAL_TITLE" or loc.format( LOC"UI.FIGHT_SCREEN.CHARACTER_TUTORIAL_TITLE", TheGame:GetGameState():GetPlayerAgent():GetName()), 
                    TheGame:GetGameState():GetPlayerAgent():GetContentID() == "SAL" and LOC"UI.FIGHT_SCREEN.TUTORIAL_BODY" or loc.format( LOC"UI.FIGHT_SCREEN.CHARACTER_TUTORIAL_BODY", TheGame:GetGameState():GetPlayerAgent():GetName()))
        end
end,

    [ EVENT.END_ENEMY_TURN ] = function( self )
        self:RefreshDamagePreviews()
        self.end_turn_btn:SetText( LOC"UI.CARDS.END_TURN" )
        self:DelayAndAnimate( nil, true )
    end,

    [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
        if self.battle:GetBattleResult() ~= nil and self:FindCardWidget( card ) == nil then
            -- Suppress animation for any cards that don't already have widgets (specifically,
            -- don't animate cards expending from Draw/Discard)
            return
        end

        self:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )

        -- if source_deck ~= target_deck then
        --     if source_deck then
        --         self:RefreshDeck( source_deck:GetDeckType() )
        --     end
        --     if target_deck then
        --         self:RefreshDeck( target_deck:GetDeckType() )
        --     end
        -- end
    end,

    [ EVENT.CARD_DISCARDED ] = function( self, card, battle )
        local w = self:FindCardWidget( card )
        if w then
            w:AnimateDiscard( card )
        end
    end,

    [ EVENT.CARD_PINNED ] = function( self, card, status_pinned )
        local w = self:FindCardWidget( card )
        local pinned_w = self:FindCardWidget( status_pinned )
        if w then
            if pinned_w then
                w:AnimateToWidget( pinned_w )
            else
                w:AnimateToTrash()
            end
        else
            -- TODO: would be nice to animate the card instance from its source deck.
        end
    end,

    [ EVENT.CARDS_STOLEN ] = function( self, cards, fighter )
        self:Delay(.6)
        local card_widgets = {}
        for i,card in ipairs(cards) do
            local w = self:DetachCardWidget( card )
            if w == nil then
                w = self.dealt_cards:CreateCardWidget( card )
            else
                self.dealt_cards:AttachCardWidget( card, w )
            end

            local card_dest = self:GetAnimFighterEntity( fighter )
            card_dest = card_dest.cmp.AnimFighter.status_widget
            card_widgets[w] = card_dest
        end
        self:Delay(1)

        for w, dest_w in pairs(card_widgets) do
            if w and dest_w then
                w:AnimateToWidget( dest_w )
            end
        end
    end,

    [ EVENT.CARD_UPGRADED ] = function( self, card, old_snapshot )
        local w = self:FindCardWidget( card )
        if w then
            w:AnimateUpgraded( card, old_snapshot )
            
        end
    end,

    [ EVENT.CARD_CHANGED ] = function( self, card, snapshot, hint )
        local w = self:FindCardWidget( card )
        if w then
            w:RefreshCard( snapshot, self.battle )
            w.card = card -- FIXME: Uh, don't save the snapshot into the widget.

            if hint == "restore_charges" then
                w:AnimateUpgraded()
            end
        end
    end,

    [ EVENT.CARD_DEALT ] = function( self, card, deck )
        -- cards is a transient table; can modify.
        if card.show_dealt == false then
            return -- Force suppressed.
        elseif card.show_dealt == true then
            -- Force show!
        elseif CheckBits( card.flags, CARD_FLAGS.NPC ) then
            -- No NPC cards!
            return
        end

        if not self:FindCardWidget( card ) then
            self.dealt_cards:CreateCardWidget( card )
        end
    end,

    [ EVENT.DECK_TOO_LARGE ] = function( self, card, deck )
        self.cardspanel:ShowBanner( LOC"CARD_ENGINE.DECK_TOO_LARGE", 1.0, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
    end,
 
    [ EVENT.STATUS_CHANGED ] = function( self, fighter, status )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            self:DelayForFighter( fighter, true )
            ent.cmp.AnimFighter:DoSequence( AnimFighter.OnStatusChanged, ent.cmp.AnimFighter, status )
            -- ent.cmp.AnimFighter:OnStatusChanged( status )
            self:RefreshDamagePreviews()
        end
    end,

    [ EVENT.REFRESH_CONDITIONS ] = function( self, fighter, condition )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter.status_widget:RefreshCondition( condition )
        end
    end,

    [ EVENT.QUIP ] = function( self, fighter, ... )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:SayQuip( ... )
        end
    end,

    [ EVENT.SPEECH ] = function( self, fighter, delay, txt )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:SaySomething( txt )
            self:Delay( delay )
        end
    end,

    [ EVENT.SOUND ] = function( self, sound )
        AUDIO:PlayEvent( sound )
    end,

    [ EVENT.DELAY ] = function( self, duration )
        self:Delay( duration )
    end,

    [ EVENT.PLAY_ANIM ] = function( self, fighter, anim, wait, idle, overrides )
        self:DelayForFighter( fighter )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            if overrides then
                for k,v in pairs( overrides ) do
                    ent.cmp.AnimController:AddOverrideFile(v) 
                end
            end
            ent.cmp.AnimFighter:PlayAnim( anim )
            if wait then
                ent.cmp.AnimFighter:WaitAnim()
            end
            if idle then
                ent.cmp.AnimFighter:QueueIdleAnim()
            end
            if overrides then
                for k,v in pairs( overrides ) do
                    ent.cmp.AnimController:RemoveOverrideFile(v) 
                end
            end
        end
    end,

    [ EVENT.PLAY_FX ] = function( self, fighter, fx_name )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:PlayOneShotFX( fx_name )
        end
    end,

    [ EVENT.PLAY_FLOATER ] = function( self, fighter, txt, duration, colour )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            ent.cmp.AnimFighter:ShowAttackFloater( txt, duration or 0.8, colour or UICOLOURS.WHITE )
        end
    end,
    
    [ EVENT.FLOURISH_CHANGED ] = function( self, graft, flourish, prev_flourish )
        if flourish >= (graft.userdata.battle_max or math.huge) then
            local ent = self:GetAnimFighterEntity( self.battle:GetPlayerFighter() )
            if ent then
                ent.cmp.AnimFighter:PlayFlourish()
            end
        end
    end,

    [ EVENT.CUSTOM ] = function( self, fighter, fn, custom_ent )
        local ent = self:GetAnimFighterEntity( fighter )
        if ent then
            fn( self, ent, fighter )
        end
        if custom_ent then
            fn( self, custom_ent )
        end
    end,

    [ EVENT.FIGHTER_REPLACED ] = function( self, old_fighter )
        self:Delay( 0.5 )
        self:DelayForFighter( old_fighter )
        local ent = self:GetAnimFighterEntity( old_fighter )
        if ent then
            ent.cmp.AnimFighter:DoRun()
        end
    end,

    [ EVENT.EXECUTES_ADDED ] = function( self )
        AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/hand_expend" )
        AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/discard_expend" )
    end,
}

------------------------------------------------------------------------------------------
--
local FightScreen = class("Screen.FightScreen", Screen.FightScreenBase)

FightScreen.CONTROL_MAP =
{
    {
        test = function(self)
            return not self.show_reticule
        end,
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },
    {
        -- For the Widget.ChooseCardsFromHand
        controls = { Controls.Digital.MENU_ACCEPT, Controls.Digital.MENU_CANCEL },
        test = function(self)
            return self.cardspanel:IsChoosingCards() and not self.cardspanel:GetChoosingCardsPanel():IsVisible()
        end,
        fn = function(self)
            self.cardspanel:GetChoosingCardsPanel():RestorePopup()
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self)
            self.cardspanel.wait_for_click = false
            TheGame:FE():ClearFocusWidget() -- prevents the dealt card from staying focused
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CONTINUE", Controls.Digital.MENU_ACCEPT) )
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return not self.show_reticule and self.battle:GetStatus() ~= battle_defs.STATUS.CHOOSE_CARDS
        end,
        fn = function(self)
            --
            if self.hover_fighter and self:GetChosenCard() then
                if not self:OnClickFighter( self.hover_fighter ) and TheGame:FE():IsTouchMode() then
                    self:ShowFighterInfoPopup( self.hover_fighter )
                end
            else
                self:SetCardFocus( true )
            end
            return true
        end,

        hint = function(self, left, right)
            if self.hover_target or self:GetChosenCard() then
                table.insert(right, loc.format(LOC"UI.CONTROLS.PLAY_CARD", Controls.Digital.MENU_ACCEPT) )
            end
        end,
    },

    {
        control = Controls.Digital.CARDS_CHOOSE_FLOURISH,
        test = function(self)
            if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self.show_reticule then
                return self.flourish_bar:IsShown() and self.end_turn_btn:IsShown() and self.end_turn_btn:IsEnabled() and not self.fighter_focus and not self.cardspanel:IsChoosingCards()
            end
        end,
        fn = function(self)
            self:ActivateFlourish()
            return true
        end,
        hint = function(self, left, right)
            if self.flourish_bar:IsFull() then
                table.insert(right, loc.format(LOC"UI.CONTROLS.PLAY_FLOURISHES", Controls.Digital.CARDS_CHOOSE_FLOURISH) )
            else
                table.insert(right, loc.format(LOC"UI.CONTROLS.VIEW_FLOURISHES", Controls.Digital.CARDS_CHOOSE_FLOURISH) )
            end
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
           return self:GetChosenCard() and not self.cardspanel:IsWaitingForClick() and not self.show_reticule
        end,
        fn = function(self)
            return self:CancelChosenCard()
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CANCEL", Controls.Digital.MENU_CANCEL) )
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            if not self.choosing_cards_from_hand and not self.cardspanel:IsWaitingForClick() then
                return self.cardspanel:IsCardSelected()
            end
        end,
        fn = function(self)
            TheGame:FE():ClearFocusWidget()
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            return self.cardspanel:IsChoosingCards() and self.cardspanel:ChooseCards():GetCardCount() > 0
        end,
        fn = function(self)
            local cards = self.cardspanel:ChooseCards()
            local last_card = cards:GetCardAt( cards:GetCardCount() )
            cards:TryChooseCard( last_card )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format ( LOC"UI.CONTROLS.UNSELECT", Controls.Digital.MENU_CANCEL ) )
        end,
    },
    {
        control = Controls.Digital.MENU_ALT,
        test = function(self)
            return ( self:GetChosenCard() or self.cardspanel:IsCardSelected() ) and not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self)
            return self:CancelChosenCard()
        end,
    },
    {
        control = Controls.Digital.CARDS_END_TURN,
        test = function(self)
            if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self.show_reticule then
                return self.end_turn_btn:IsShown() and self.end_turn_btn:IsEnabled() and not self.cardspanel:IsChoosingCards() and not self.cardspanel:IsWaitingForClick() and not self:GetChosenCard()
            end
        end,
        fn = function(self)
            -- Unselect selected card
            local card = self:GetChosenCard()
            if card then
                self.cards:ChooseCard( card )
                card:AssignTarget( nil )
                AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
            end
            if self.cardspanel:IsCardSelected() then
                TheGame:FE():ClearFocusWidget()
            end
            -- End turn
            self.end_turn_btn:Click()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.FIGHT_SCREEN.CONTROLS_RIGHT_ENDTURN", Controls.Digital.CARDS_END_TURN, self.battle:GetTurns() ) )
        end,
    },
    {
        control = Controls.Digital.CARDS_RUN,
        test = function(self)
            if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self.show_reticule then
                return self.concede_btn:IsShown() and self.concede_btn:IsEnabled() and not self.cardspanel:IsChoosingCards() and not self.cardspanel:IsWaitingForClick() and not self:GetChosenCard()
            end
        end,
        fn = function(self)
            self:OnClickRun()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.FIGHT_SCREEN.CONTROLS_RIGHT_RUNAWAY", Controls.Digital.CARDS_RUN ) )
        end,
    },
    {
        control = Controls.Digital.CARDS_ACCEPT_SURRENDER,
        test = function(self)
            return self.accept_surrender_button:IsShown() and self.accept_surrender_button:IsEnabled() and not self.cardspanel:IsChoosingCards() and not self.cardspanel:IsWaitingForClick() and not self.show_reticule
        end,
        fn = function(self) 
            self:AcceptSurrender()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.FIGHT_SCREEN.CONTROLS_RIGHT_ACCEPTSURRENDER", Controls.Digital.CARDS_ACCEPT_SURRENDER ) )
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_RT,
        test = function(self)
            return self.show_reticule and self:GetFighterUnderMouse()
        end,
        fn = function(self)
            local focused_fighter = self:GetFighterUnderMouse()
            self:ShowFighterInfoPopup( focused_fighter )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.FIGHTER_INFO", Controls.Digital.GAMEPAD_RT))
        end
    },
    {
        controls = { Controls.Digital.MENU_CANCEL, Controls.Digital.MENU_UP, Controls.Digital.MENU_RIGHT, Controls.Digital.MENU_DOWN, Controls.Digital.MENU_LEFT },
        test = function(self)
            return self.show_reticule == true
        end,
        fn = function(self)
            self:SetReticuleMode( false )
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT_RELEASE", Controls.Digital.MENU_CANCEL ) )
        end,
    },
    {
        test = function(self)
            return not self.cardspanel:IsChoosingCards()
        end,
        hint = function(self, left, right)
            if self.show_reticule then
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_MOVE_INSPECT" )
            else
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT" )
            end
        end,
    },
    {
        control = Controls.Digital.VIEW_DRAW_DECK,
        test = function(self)
            return not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self) 
            self.draw_deck:ViewDeck()
            return true
        end,
    },
    {
        control = Controls.Digital.VIEW_DISCARDS_DECK,
        test = function(self)
            return not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self) 
            self.discard_deck:ViewDeck()
            return true
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_X,
        test = function(self)
            return self.cardspanel:IsChoosingCards() and self.cardspanel:GetChoosingCardsPanel():HasConfirmButton()
        end,
        fn = function(self)
            self.cardspanel:GetChoosingCardsPanel():Confirm()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format ( LOC"UI.CONTROLS.ACCEPT_CARDS", Controls.Digital.GAMEPAD_X ) )
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_Y,
        test = function(self)
            return self:GetFE():FindScreen( Screen.ChooseCardsPopup ) ~= nil
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.VIEW_CHOICES", Controls.Digital.GAMEPAD_Y))
        end,
    }

}

function FightScreen:init( battle, on_done )
    FightScreen._base.init(self)

    -- self.hover_check = true
    self.on_done = on_done
    
    if battle == nil and TheGame:GetDebug() then
        battle = self:CreateBattle()
    end

    self.battle = battle
    self.battle:ListenForAny( self, self.OnBattleEvent, -1 )
    self.fightsim.mgr:AssignBattle( battle )
    self.scenario = battle:GetScenario()

    self.music_phase = 1
    self.event_queue = {}

    -- Total damage preview.
    self.damage_preview = Battle.DamagePreview( battle ) 
    -- Damage preview for the player only.
    self.player_preview = Battle.DamagePreview( battle ) 

    self.crit_overlay_image = engine.asset.Texture("large/screen_flash1.tex")

    self.overlay_images = {
        engine.asset.Texture("large/hit_overlay.tex"),
        engine.asset.Texture("large/hit_overlay2.tex"),
    }
    self.hit_overlay = Widget.Image()
    
    self.status_root = Widget()

    self.speech_bubbles = {}
    self.speech_bubble_root = Widget()

    self.float_root = Widget()

    self.target_arrow = self:AddChild( Widget.TargetArrow() )

    local VPAD = 10
    
    self:AddChildren{
        self.hit_overlay:SetAnchors("fill", "fill"):Hide(),
        self.float_root:AddChildren{
            self.status_root,
            self.speech_bubble_root,
        },
    }
    self.status_widgets = {}

    self.darken_overlay = Widget.SolidBox( RES_X, RES_Y, UICOLOURS.BLACK ):BlockMouse():SetAnchors( "fill", "fill" )
    self:AddChild( self.darken_overlay ):SetTintAlpha( 0 ):Hide()

    local function OnCardDraggingChanged( card, old_card )
        if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
            if not self.show_reticule then
                self.cards:DisablePreview( card ~= nil )
            end
        end
        self:RefreshChosenCard()

        if old_card then
            old_card:AssignTarget( nil )
            local w = self:FindCardWidget( old_card )
            if w then
                w:RefreshCard( old_card, self.battle )
            end
            self.player_preview:RemovePreview( old_card )
            self:RefreshDamagePreviews()
        end
    end

    self.main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if self.main_overlay == nil then
        self.main_overlay = TheGame:FE():PushScreen( Screen.MainOverlay() )
    end

    self.grafts = self.main_overlay:GetBattleGraftsWidget()
    self.boons = self.main_overlay:GetBoonGraftsWidget()
    self.banes = self.main_overlay:GetBaneGraftsWidget()
    self.perks = self.main_overlay:GetPerksGraftsWidget()

    -- Cards Panel
    ----------------------------------
    self.cardspanel = self:AddChild( Widget.CardsPanel( Widget.BattleCard, self.battle ) )
    self.time_indicator = self.cardspanel:ActionsWidget()
    self.flourish_bar = self.cardspanel:FlourishBar():SetBattle( TheGame and TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent(), self.battle )
    self.end_turn_btn = self.cardspanel:EndTurnButton( function() self:OnClickEndTurn() end, loc.format( LOC"UI.FIGHT_SCREEN.END_TURN_TOOLTIP", Controls.Digital.CARDS_END_TURN ))
    self.concede_btn = self.cardspanel:ConcedeButton( function() self:OnClickRun() end, LOC"UI.FIGHT_SCREEN.RUN_AWAY" )
    self.accept_surrender_button = self.cardspanel:AcceptSurrenderButton( function() self:AcceptSurrender() end, LOC"UI.FIGHT_SCREEN.ACCEPT_SURRENDER_TOOLTIP" )
    self.draw_deck = self.cardspanel:DrawDeckButton( self.battle:GetDrawDeck() )
    self.discard_deck = self.cardspanel:DiscardDeckButton( self.battle:GetDiscardDeck() )
    self.trash_deck = self.cardspanel:TrashDeckButton( self.battle:GetTrashDeck() )
    
    local function OnCardChoiceChanged( card, is_chosen )
        if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
            if not self.show_reticule then
                self.cards:DisablePreview( is_chosen )
            end
        end

        if is_chosen then
            self:GetFE():ClearFocusWidget()
            self:SetFighterFocus()
            self:RefreshChosenCard()
        else
            self:GetFE():ClearFocusWidget()
            self:SetCardFocus()
            self.player_preview:RemovePreview( card )
            self:RefreshDamagePreviews()

            card:AssignTarget( nil )
            local w = self:FindCardWidget( card )
            if w then
                w:RefreshCard( card, self.battle )
            end

            self:RefreshDamagePreviews()
        end
    end

    local function OnTryChooseCard( card )
        local ok, reason = self.battle:CanPlayCard( card )
        if not ok then
            if reason == CARD_PLAY_REASONS.NO_ACTIONS then
                AUDIO:PlayEvent(SoundEvents.card_not_enough_actions)
            else
                AUDIO:PlayEvent(SoundEvents.card_invalid)
            end

            if reason then
                reason = GetLocalizedConvoCommonString( reason )

                self:NoActions( reason )
            end

            return true -- sink it
        end

        local auto_target = self:AssignAutoTarget( card )
        if not auto_target then
            self.cards:ChooseCard( card )
            return true
        end
    end

    self.cards = self.cardspanel:Cards()
    self.floating_cards = self.cardspanel:FloatingCards()
    self.cards:SetOnChoose( OnCardChoiceChanged, OnTryChooseCard )
    self.cards:SetOnDrag( OnCardDraggingChanged )
    self.dealt_cards = self.cardspanel:DealtCards()
    self.choose_cards = self.cardspanel:ChooseCards()
    self.choose_cards:Hide()

    self:SetToolTipLayoutFn( function( w, tooltip_widget ) self:LayoutToolTip( tooltip_widget ) end )
end

function FightScreen:FocusOnValidFighter()
    -- Go through the fighters and focus on the first valid one
    -- Used on gamepad mode only
    local card = self:GetChosenCard()
    if card and self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
        local best_fighter, best_widget
        for fighter, widget in pairs( self.status_widgets ) do
            local can_target, reason = self.battle:CanPlayCard( card, fighter )
            if can_target and (best_fighter == nil or fighter == self.last_ally_target or fighter == self.last_enemy_target) then
                best_fighter, best_widget = fighter, widget
            end
        end

        if best_fighter and best_widget then
            best_widget:SetFocus()
            self:SetHoverFighter( best_fighter )
        end
    end
end

function FightScreen:OnClickEndTurn()
    if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self:GetChosenCard() then
        if self.battle:CanPlayAny( true ) and engine.inst:GetSetting("gameplay", "confirm_end_turn", true) then
            local popup = Screen.YesNoPopup( LOC"UI.END_TURN_CONFIRM_BTN", LOC"UI.END_TURN_CONFIRM_BODY", LOC"UI.END_TURN_CONFIRM_SKIP" ):
            SetFn( function( result )
                if result == Screen.YesNoPopup.YES or result == Screen.YesNoPopup.NO then
                    -- End the turn
                    self.end_turn = true
                    AUDIO:PlayEvent(SoundEvents.card_end_turn)
                    -- Also save this preference for next time
                    if result == Screen.YesNoPopup.NO then
                        -- Tell the player that
                        UIHelpers.InfoPopup( LOC"UI.END_TURN_CONFIRM_INFO_TITLE", LOC"UI.END_TURN_CONFIRM_INFO_BODY" )
                        -- Actually save it to the settings
                        engine.inst:SetSetting("gameplay", "confirm_end_turn", false)
                    end
                end
            end )
            self:GetFE():PushScreen( popup )

        else
            self.end_turn = true
            AUDIO:PlayEvent(SoundEvents.card_end_turn)
        end
    end
end

function FightScreen:OnClickRun()
    if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self:GetChosenCard() then
        local popup = Screen.YesNoPopup(LOC"UI.FIGHT_SCREEN.RUN_CONFIRM_BTN", LOC"UI.FIGHT_SCREEN.RUN_CONFIRM_BTN_TOOLTIP" ):
        SetFn( function( result )
            if result == Screen.YesNoPopup.YES then
                AUDIO:PlayEvent(SoundEvents.negotiation_concede)
                self.battle:GetPlayerFighter():AddCondition( "RUNNING" )
                self.end_turn = true
            end
        end )
        self:GetFE():InsertScreen( popup )
    end
end

function FightScreen:CreateBattle()
     local battle

    -- Fresh GameState so that local_battle can create skinned agents indefinitely.
    TheGame.gamestate = GameState()

    -- DONT use require so that we can pcall without messing up the Profiler stack.
    package.loaded[ "local_battle" ] = nil
    local ok, result = pcall( dofile, "scripts/local_battle.lua" )
    if is_instance( result, BattleEngine ) then
        battle = result
    elseif is_instance( result, Scenario ) then
        battle = BattleEngine( result )
    else
        if not ok then
            if result == nil or (not result:find( "No such file" ) and not result:find( "Bad file" )) then
                LOGWARN( tostring(result) )
            else
                print( "Loading local_battle.lua:", result )
            end
        end

        local fight_tester = DebugFightTest()
        local fight_data = fight_tester:GetFightPresetData()
        fight_data = fight_data and fight_tester:GenerateFightPreset( fight_data )
        if fight_data then
            local scenario = fight_tester:GenerateScenario( fight_data )
            battle = BattleEngine( scenario )
        end
    end

    if battle == nil then
        local scenario = Scenario()
        scenario:AddAgentToTeam( Agent("SAL"), TEAM.BLUE, 0, true )
        battle = BattleEngine( scenario )
    end

    if TheGame:GetGameState():GetPlayerAgent() == nil then
        for i, fighter in battle:AllFighters() do
            if fighter.agent:GetFactionID() == PLAYER_FACTION then
                TheGame:GetGameState():AddPlayerAgent( fighter.agent )
                break
            end
        end
    end

    return battle
end

function FightScreen:ShowHit( attacker, hit_tags )
    if attacker:GetTeamID() ~= TEAM.BLUE then
        self.hit_overlay:SetScale(-1,1)
    end

    local is_threshold = table.arrayfind( hit_tags, "threshold" )
    local is_crit = table.arrayfind( hit_tags, "critical" )

    if is_threshold then
        self.hit_overlay:SetTexture(self.crit_overlay_image)
    else
        self.hit_overlay:SetTexture(self.overlay_images[math.random(#self.overlay_images)])
    end
    
    if Game:ScreenFlashEnabled() then
        self.hit_overlay:Show():SetTintColour(1,1,1,0):AlphaTo(.5, .1, easing.outQuad, function() self.hit_overlay:AlphaTo(0, .2, easing.inQuad, function() self.hit_overlay:Hide() end)end)
    end

    if is_crit then
        AUDIO:PlayEvent(SoundEvents.battle_crit_hit)
        self:ScreenShake( FIGHT_SHAKE_CRIT )
    else
        self:ScreenShake( FIGHT_SHAKE_NORMAL )
    end
end

function FightScreen:_DoRefreshDamagePreviews()
    if self.previews_dirty then
        
        --print ("DID PREVIEW", self.preview_count)
        --self.preview_count = 0

        self.previews_dirty = false
    
        self.damage_preview:Clear()
        self.damage_preview:AddPreview( self.player_preview )

        for i, fighter in self.battle:ActiveFighters() do
            local attacks = fighter.prepared_attacks
            if attacks then
                for k, attack in ipairs(attacks) do
                    self.damage_preview:PreviewAttack( attack )
                end            
            end
            local conditions = fighter:GetConditions()
            if conditions then
                for k,condition in pairs(conditions) do
                    if condition.PreviewFn then
                        self.damage_preview:PreviewCondition( condition )
                    end
                end            
            end
        end

        for fighter, widget in pairs( self.status_widgets ) do
            widget:RefreshDamagePreview( self.damage_preview:GetPreviews( fighter ))
        end
    end

end
function FightScreen:RefreshDamagePreviews()
    self.previews_dirty = true
    --self.preview_count = (self.preview_count or 0) + 1
end

local function AddTargettingArrow( coords, x0, z0, x1, z1, card )
    table.insert( coords, x0 )
    table.insert( coords, z0 )
    table.insert( coords, x1 )
    table.insert( coords, z1 )

    if card and card:IsFlagged( CARD_FLAGS.BUFF ) then
        table.insert( coords, UICOLOURS.FRIENDLY_TARGETTING_COLOUR )
    else
        table.insert( coords, UICOLOURS.ENEMY_TARGETTING_COLOUR )
    end
end

function FightScreen:RefreshDamagePreviewArrows( fighter )
    engine.inst:ProfilerPush("FightScreen:RefreshDamagePreviewArrows")
    local ent = self:GetAnimFighterEntity( fighter )
    local coords = ObtainWorkTable()
    if ent then
        local previews = self.damage_preview:GetPreviews( fighter )
        local attackers, victims = ObtainWorkTable(), ObtainWorkTable()
        local x0, _, z0 = ent:GetLocalPosition()

        for target, previews in self.damage_preview:Targets() do
            for i, preview in ipairs( previews ) do
                if target ~= fighter and preview.owner == fighter then
                    -- Targetting someone else.
                    local target_ent = self:GetAnimFighterEntity( target )
                    if target_ent and not table.contains( victims, target_ent ) then
                        local x1, _, z1 = target_ent:GetLocalPosition()
                        table.insert( victims, target_ent )
                        AddTargettingArrow( coords, x0, z0, x1, z1, preview.card )
                    end
                end
            end
        end
        ReleaseWorkTable( attackers )
        ReleaseWorkTable( victims )
    end

    engine.inst:ProfilerPush("_2")
    for _, w in pairs( self.status_widgets ) do
        w:HilightDamagePreview( fighter )
    end
    engine.inst:ProfilerPop()
    engine.inst:ProfilerPush("_3")
    self.fightsim:ShowTargetFX( table.unpack( coords ))
    engine.inst:ProfilerPop()

    ReleaseWorkTable( coords )
    engine.inst:ProfilerPop()
end

function FightScreen:RefreshActions( actions, actions_per_turn )
    if actions == nil then
        actions = self.battle:GetActionCount()
    end
    if actions_per_turn == nil then
        actions_per_turn = self.battle:GetActionsPerTurn()
    end

    local exertion = self.battle:GetExertedCount()
    if exertion > 0 then
        self.time_indicator:FormatTime( loc.format("{1}/{2} <#STATUS_HEALTH_CRIT>({3})</>", actions, actions_per_turn, exertion ))
    else
        self.time_indicator:SetTime( actions, actions_per_turn )
    end
end

function FightScreen:RefreshDrawDeck()
    local tt
    if self.fe:IsGamepadMode() then
        tt = loc.format( TheGame:Str( "UI.CARDS.DRAW_DECK_GAMEPAD_TT" ), self.battle:CalculateTurnDraw(), Controls.Digital.VIEW_DRAW_DECK )
    else
        tt = loc.format( TheGame:Str( "UI.CARDS.DRAW_DECK_TT" ), self.battle:CalculateTurnDraw(), Controls.Digital.VIEW_DRAW_DECK )
    end
    self.draw_deck:RefreshDeck()
    self.draw_deck:SetToolTip( tt )
end

function FightScreen:RefreshDeck( deck_type )
    if deck_type == DECK_TYPE.DRAW then
        self:RefreshDrawDeck()

    elseif deck_type == DECK_TYPE.IN_HAND then
        self.cards:RefreshCards( self.battle )

    elseif deck_type == DECK_TYPE.DISCARDS then
        self.discard_deck:RefreshDeck()

    elseif deck_type == DECK_TYPE.TRASH then
        local deck = self.trash_deck:RefreshDeck()
        self.discard_deck:SetOtherDeckButton( deck, LOC"UI.CARDS.TRASH_DECK_DIALOG_TITLE", LOC"UI.CARDS.TRASH_DECK_DIALOG_SUBTITLE" )
    end
end

function FightScreen:RefreshDecks()

    self:RefreshDrawDeck()

    local tt
    if self.fe:IsGamepadMode() then
        tt = loc.format( LOC"UI.CARDS.DISCARDS_DECK_GAMEPAD_TT", Controls.Digital.VIEW_DISCARDS_DECK )
    else
        tt = loc.format( LOC"UI.CARDS.DISCARDS_DECK_TT", Controls.Digital.VIEW_DISCARDS_DECK )
    end
    self.discard_deck:RefreshDeck()
    self.discard_deck:SetToolTip( tt )

    self:RefreshDeck( DECK_TYPE.TRASH )
    self:RefreshDeck( DECK_TYPE.IN_HAND )
end


function FightScreen:PreloadBuild( build_id )
    local build = Content.GetCharacterBuild(build_id)
    if build then
        CacheAnim( build.build )
        table.insert( self.preloaded_anims, build.build )

        for k,v in pairs(build.combatanims) do
            CacheAnim(v)
            table.insert( self.preloaded_anims, v )
        end
    end
end


function FightScreen:PreloadDefAnims( def_id )
    if self.preloaded_defs[def_id] then
        return
    end
    self.preloaded_defs[def_id] = true

    local def = Content.GetCharacterDef(def_id)
    if def then
        if def.build then
            self:PreloadBuild(def.build)
        end
        if def.base_builds then
            for k,v in pairs(def.base_builds) do
                self:PreloadBuild(v)
            end
        end

        if def.combat_anims then
            for k,v in pairs(def.combat_anims) do
                CacheAnim(v)
                table.insert( self.preloaded_anims, v )
            end
        end

        if def.anims then
            for k,v in pairs(def.anims) do
                CacheAnim(v)
                table.insert( self.preloaded_anims, v )
            end
        end

        if def.preload_defs then
            for k,v in ipairs(def.preload_defs) do
                self:PreloadDefAnims(v)
            end

        end

    end
end

function FightScreen:PreloadAgentAnims( char )
    self:PreloadDefAnims(char:GetContentID())
end

function FightScreen:PreloadAnims()
    self.preloaded_anims = {}
    self.preloaded_defs = {}

    for i, fighter in self.battle:AllFighters() do
        self:PreloadAgentAnims( fighter:GetAgent() )
    end
end

function FightScreen:UnloadAnims()
    for i, anim in ipairs( self.preloaded_anims ) do
        ForgetAnim(anim)
    end
end

function FightScreen:RefreshDisplay()
    self:RefreshDecks()
    self:RefreshActions()

    self.main_overlay = self.main_overlay or TheGame:FE():FindScreen( Screen.MainOverlay )
    self.main_overlay:SetFriendReserves( self.battle, TEAM.BLUE )
    self.main_overlay:SetEnemyReserves( self.battle, TEAM.RED )
    self.main_overlay:RefreshScenario( self.battle )
end

function FightScreen:FindCardWidget( card )
    return self.cardspanel:FindCardWidget( card )
end

function FightScreen:DetachCardWidget( card )
    return self.cardspanel:DetachCardWidget( card )
end

function FightScreen:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )
    return self.cardspanel:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )
end


function FightScreen:DelayAndAnimate( delay, wait_for_settled )
    local t = 0
    while true do 
        if delay == nil or t >= delay then
            if not wait_for_settled then
                break
            else
                for i, cmp in pairs( self.fightsim:GetAllComponents( AnimFighter )) do
                    self:DelayForFighter( cmp.fighter )
                end
                break
            end
        end

        local dt = coroutine.yield()
        t = t + dt
    end
    
    return t
end

function FightScreen:DelayForFighter( fighter )
    local t = 0
    while true do
        local ent = self:GetAnimFighterEntity( fighter )
        if ent == nil or ent.cmp.AnimFighter:IsSettled() then
            break
        elseif not ent.cmp.AnimFighter:IsBusy() then
            ent.cmp.AnimFighter:DoSequence( AnimFighter.ResetFighter, ent.cmp.AnimFighter )
        end

        local dt = coroutine.yield()
        t = t + dt
        if t >= 5.0 then
            print( "WARNING: timeout for fighter", fighter, fighter:GetStatus() )
            print( debug.traceback() )
            break
        end
    end
    
    return t
end

function FightScreen:DelayForAttackAnim( fighter, anim )
    local t = 0
    while true do
        local ent = self:GetAnimFighterEntity( fighter )
        if ent == nil or ent.cmp.AnimController:GetAnimName() ~= anim or not ent.cmp.AnimFighter:IsBusy() then
            break
        end

        local dt = coroutine.yield( )
        t = t + dt
        if t >= 5.0 then
            print( "WARNING: timeout for fighter", fighter, fighter:GetStatus() )
            print( debug.traceback() )
            break
        end
    end
    
    return t
end

function FightScreen:OnBattleEvent( eventname, ... )
    if type(eventname) == "string" and eventname:find( "CALC_" ) then
        -- Don't even queue.
    elseif self.wait_coro == nil then
        table.insert( self.event_queue, { eventname, ... } )
    else
        -- This is an out-of-line eventname, probably injected via a debug command or other.
        -- Flag it so we can conditionally do certain refreshes we could otherwise avoid.
        table.insert( self.event_queue, { is_debug = true, eventname, ... })
    end
end

function FightScreen:PumpAllEvents( evs )
    while true do
        local ev = not self.viz_break and table.remove( evs, 1 )
        if ev then
            local event_name = ev[1]
            local st = os.clock()
            self.current_event = ev
            if EVENT_HANDLERS[ event_name ] then
                --print( "Serial processing", table.unpack( ev ) )
                EVENT_HANDLERS[ event_name ]( self, table.unpack( ev, 2, table.maxn( ev ) ))
                --print( "\ttook ", (os.clock() - st) * 1000 )
            end
            Content.InvokeModAPI( "OnBattleEvent", self, table.unpack( ev ))
            self.current_event = nil

            if self.viz_break ~= nil then
                self.viz_break = true -- Step.
                print( event_name, " took ", (os.clock() - st), " seconds" )
            end
        else
            coroutine.yield()
        end
        if evs ~= self.event_queue and #evs == 0 then
            break
        end
    end
end

local function PrioritizeStatus( status, prev_status )
    if status == FIGHT_STATUS.DEAD then
        return status
    elseif prev_status ~= FIGHT_STATUS.DEAD then
        return status
    end
end

function FightScreen:ConsumeEvent( fn )
    for i = 1, #self.event_queue do
        local ev = self.event_queue[ i ]
        if fn( ev ) then
            table.remove( self.event_queue, i )
            return ev
        end
    end
end

function FightScreen:PrepareAttackEvents( attack )

    attack.status_changed = {}

    local current_hit = attack:FindNextHit()

    local i = 1
    local events_moved = 0
    while i <= #self.event_queue do
        local ev = self.event_queue[i]
        local event_name = ev[1]

        if event_name == EVENT.ON_HIT then
            local battle, hit_attack, hit = table.unpack( ev, 2 )
            if hit_attack == attack then
                current_hit = attack:FindNextHit( hit )
                table.remove( self.event_queue, i )
            else
                i = i + 1
            end

        elseif event_name == EVENT.DELTA_STAT then
            -- All stat changes caused by this attack are removed, so that their timing with the
            -- animations of the hit results.
            if current_hit then
                if current_hit.deltas == nil then
                    current_hit.deltas = {}
                end
                table.insert( current_hit.deltas, ev )
                table.remove( self.event_queue, i )
            else
                -- Re-insert after POST_RESOLVE.
                local _, j = self:PeekVizEvent( EVENT.POST_RESOLVE )
                if j then
                    table.insert( self.event_queue, j + 1 + events_moved, ev )
                    table.remove( self.event_queue, i )
                    events_moved = events_moved + 1
                else
                    i = i + 1
                end                
            end

        elseif event_name == EVENT.ON_DODGE then
            if current_hit then
                current_hit.dodge = ev
                table.remove( self.event_queue, i )
            else
                -- Re-insert after POST_RESOLVE.
                local _, j = self:PeekVizEvent( EVENT.POST_RESOLVE )
                if j then
                    table.insert( self.event_queue, j + 1 + events_moved, ev )
                    table.remove( self.event_queue, i )
                    events_moved = events_moved + 1
                else
                    i = i + 1
                end        
            end

        elseif event_name == EVENT.CONDITION_ADDED or event_name == EVENT.CONDITION_REMOVED then
            if current_hit then
                if current_hit.conditions == nil then
                    current_hit.conditions = {}
                end
                table.insert( current_hit.conditions, ev )
                table.remove( self.event_queue, i )
            else
                 i = i + 1
            end

        elseif event_name == EVENT.STATUS_CHANGED then
            local fighter, status = table.unpack( ev, 2 )
            if attack:IsTarget( fighter ) then
                table.remove( self.event_queue, i )
                attack.status_changed[ fighter ] = PrioritizeStatus( status, attack.status_changed[ fighter ] )
            else
                i = i + 1
            end

        elseif event_name == EVENT.POST_RESOLVE or event_name == EVENT.ON_RIPOSTE then
            break

        elseif event_name == EVENT.CUSTOM_BEGIN then
            -- Do not reearrange anything within the event block CUSTOM_BEGIN .. CUSTOM_END.
            local _, j = self:PeekVizEvent( EVENT.CUSTOM_END, nil, i )
            i = j and (j + 1) or (i + 1)

        else
            i = i + 1
        end
    end
end

function FightScreen:Breakpoint( txt )
    local current = coroutine.running()
    local current_coro

    for k, coro in pairs( self.coros ) do
        if coro.c == current then
           coro.breakpoint = txt 
           current_coro = coro
           break
        end
    end

    assert( current_coro, txt )
    while current_coro and current_coro.breakpoint == txt do
        coroutine.yield()
    end
end

function FightScreen:WaitForEvents()
    while #self.event_queue > 0 or self.current_event do
        coroutine.yield()
    end

    self:DealCards()
end

function FightScreen:PeekVizEvent( eventname, fn, start )
    for i = (start or 2), #self.event_queue do
        local ev = self.event_queue[i]
        if ev[1] == eventname and (fn == nil or fn(ev)) then
            return ev, i
        end
    end
end

function FightScreen:ToggleHUD( visible, hud_flags )
    -- if self.fightsim then
    --     self.fightsim:SetHidingFX( not show )
    -- end

    if visible or CheckBits( hud_flags, HUD_FLAGS.SHOW_CARDS ) then
        -- ccc: corey's new requirements are to hide the cardspanel entirely.
        self.cardspanel:Show()

        self.float_root:SetShown( visible )

        self.RefreshDamagePreviewArrows = nil
        self.target_arrow:Show()
    else
        -- ccc: corey's new requirements are to hide the cardspanel entirely.
        self.cardspanel:Hide() -- self.cardspanel:SetHUDViz( false )

        self.float_root:Hide()

        -- Hide targetting arrow
        self.RefreshDamagePreviewArrows = function() end
        self.target_arrow:Hide()
    end
end

function FightScreen:SetDemoMode( plaxdata, day_phase )
    self.demo_mode = true
    if plaxdata then
        self.plax:SetLayers( plaxdata.layers )
        self.plax:SetColourNames( plaxdata.day_colour, plaxdata.night_colour )
        self.plax:SetDayPhase(day_phase)
    end
end

function FightScreen:RestoreTimeLabel()
    local actions, max_actions = self.battle:GetActionCount()
    self.time_indicator:RestoreTime( actions, max_actions )
end

function FightScreen:NoActions( txt )
    self.cardspanel:ShowBanner( txt, nil, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
    self.time_indicator:AnimateOutOfTime()
end

function FightScreen:RefreshCardSpeed()
    local card_speed = engine.inst:GetSetting( "gameplay", "battle_speed", 1.0 )
    card_speed = 1.0 + (card_speed - 1.0) * 0.5
    self:SetUpdateSpeed( card_speed )
    self.hit_overlay:SetUpdateSpeed( 1 / card_speed )
    self.cards:SetUpdateSpeed( 1 / card_speed )
    self.fightsim:SetBattleSpeed( card_speed )
end

function FightScreen:OnOptionsChanged()
    self:RefreshCardSpeed()
end

function FightScreen:OnOpen()
    FightScreen._base.OnOpen(self)

    self:RefreshReason()
    self:RefreshDisplay()

    self:PreloadAnims()

    if self.demo_mode then
        self:SetupDemo()
    else
        self:Setup()
    end
    self:SetCardFocus()

    self:RefreshCardSpeed()
    TheGame:GetEvents():ListenForEvent( "options_changed", self, self.OnOptionsChanged )

    TheGame:SetAmbienceParam("playing_cards", 1)
end

function FightScreen:OnControlModeChange( cm, device_id )
    FightScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD or cm == CONTROL_MODE.TOUCH then
        self:SetReticuleMode( false )
    end

    self:RefreshDecks()
end

function FightScreen:OnScreenModeChange( sm )
    FightScreen._base.OnScreenModeChange( self, sm )

    -- Update the fighter outlines
    -- This value is set initially on animfighter.lua:OnScreenModeChange() since when the fight screen initializes, there aren't any animfighters
    local fighter_outline_width = 3
    local fighter_selected_outline_width = 8
    if sm == SCREEN_MODE.MONITOR then
        fighter_outline_width = 3
        fighter_selected_outline_width = 8
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        fighter_outline_width = 7
        fighter_selected_outline_width = 16
    end

    for i, fighter in self.battle:AllFighters() do
        local animfighter = self:GetAnimFighterEntity(fighter)
        if animfighter then
            animfighter.cmp.AnimFighter:SetOutlineWidth( fighter_outline_width )
            animfighter.cmp.AnimFighter:SetSelectedOutlineWidth( fighter_selected_outline_width )
        end
    end

    for fighter, speech_bubble in pairs( self.speech_bubbles ) do
        speech_bubble:SetLayoutScale( LAYOUT_SCALE[sm] )
    end

    -- widget:SetLayoutScale( LAYOUT_SCALE[sm] )
end

function FightScreen:TestReticule(wx, wy)
    local mouse_blocked, hover_widget = self:CheckMouseHover(wx, wy)
    if hover_widget then
        return hover_widget
    end

    if self.main_overlay then
        local mouse_blocked, hover_widget = self.main_overlay:CheckMouseHover(wx, wy)
        return hover_widget
    end
end

function FightScreen:WantsReticuleMode()
    return self.show_reticule == true
end

function FightScreen:SetReticuleMode( show_reticule )
    if show_reticule ~= self.show_reticule then
        self.show_reticule = show_reticule
        -- self.cardspanel:RefreshCardsInHandPos( false )

        -- Keep track of currently targetted hover fighter
        if show_reticule then
            self.target_fighter = self.hover_fighter

            self.cards:DisablePreview( true )
            self.cards:SetCardPreview( nil )

            if self.fe:GetFocusWidget() then
                local x, y = self.fe:GetFocusWidget():GetWorldPos()
                self.fe:SetReticulePos( x, y )
            end
        else   
            self.cards:DisablePreview( false )
            if self.target_fighter then
                local w = self.status_widgets[ self.target_fighter ]
                if w then
                    w:SetFocus()
                end
                self.target_fighter = nil
            else
                self.fe:ClearFocusWidget()
            end
        end

        self.status_root.show_child_tooltip_on_focus = show_reticule
    end
end

function FightScreen:GetDefaultFocus()
    if self.hover_fighter then
        local w = self.status_widgets[ self.hover_fighter ]
        if w then
            return w
        end
    end

    if self.fighter_focus then
        return self.status_root and self.status_root.children and #self.status_root.children>0 and self.status_root.children[1]
    end
    return self.cardspanel:GetDefaultFocus()
end

function FightScreen:SetFighterFocus()

    self.fighter_focus = true
    self.card_focus = false

    -- Make fighters focusable
    self.status_root:SetFocusLock( true )
    for k, v in ipairs( self.status_root.children ) do
        v:SetFocusable( true )
    end

    -- Make cards un-focusable
    self.cardspanel:SetFocusable( false )
    TheGame:FE():ClearFocusWidget()

    if self.cards:IsGamepadMode() then self:FocusOnValidFighter() end
end

function FightScreen:SetCardFocus( force_default )

    self.fighter_focus = false
    self.card_focus = true

    -- Make fighters un-focusable
    self.status_root:SetFocusLock( false )
    for k, v in ipairs( self.status_root.children ) do
        v:SetFocusable( false )
    end

    -- Make cards focusable
    self.cardspanel:SetFocusable( true )
    if not self.show_reticule then
        self.cards:DisablePreview( false )
    end

    TheGame:FE():ClearFocusWidget()

    if self.cards:IsGamepadMode() and self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
        if force_default or self.fe:GetScreenMode() == SCREEN_MODE.MONITOR then
            local widget = self.cards:GetDefaultFocus()
            if widget then
                widget:SetFocus()
            end
        end
    end
end

function FightScreen:OnClose()
    FightScreen._base.OnClose( self )
    self.battle:RemoveListener( self )
    TheGame:GetEvents():RemoveListener( self )

    if self.select_event then
        self.select_event:Stop()
        self.select_event:Release()
    end

    self.main_overlay = self.main_overlay or TheGame:FE():FindScreen( Screen.MainOverlay )
    self.main_overlay:SetModeOverride( nil )

    -- Need to clear the debug environment to release battle resources.
    local dbg = TheGame:GetDebug()
    if dbg then
        dbg:Clear()
        dbg:GetDebugEnv().viz_break = self.viz_break
    end

    TheGame:SetAmbienceParam("playing_cards", 0)

    self:UnloadAnims()
end

function FightScreen:RepositionFloaterWidget( widget, ent )
    local wx, wy, wz = ent.cmp.AnimController:GetSymbolWorldPos("head_swap", 0, 0, 0)
    local dx, dy = 0, 0
    if ent.cmp.AnimController.flip then
        dx = - dx
    end

    local x, y = self.fightsim:GetCamera():WorldToScreen( wx + dx, wy + dy, wz )
    widget:SetPos(x,y)
    widget:SetShown( self:GetFE():IsHUDVisible() )
    return x, y
end

function FightScreen:RepositionStatusWidget( widget, ent )
    local feet_wx, feet_wy, feet_wz, head_wx, head_wy, head_wz = ent.cmp.AnimFighter:GetStatusWidgetPosition()
    local feet_x, feet_y = self.fightsim:GetCamera():WorldToScreen( feet_wx, feet_wy, feet_wz )
    local head_x, head_y = self.fightsim:GetCamera():WorldToScreen( head_wx, head_wy, head_wz )
    widget:SetScale(self:GetFE():GetBaseWidgetScale())
    widget:SetPos( feet_x, feet_y )
    widget:SetHeadPos( head_x, head_y )
    widget:SetShown( self:GetFE():IsHUDVisible() )
    return feet_x, feet_y
end

function FightScreen:RepositionSpeechWidget( widget, ent, fighter )
    local wx, wy, wz = ent.cmp.AnimController:GetSymbolWorldPos("head_swap", 0, 0, 0)
    local x, y = self.fightsim:GetCamera():WorldToScreen( wx, wy, wz )
    local base_scale = self:GetFE():GetBaseWidgetScale()

    widget:SetScale( base_scale )

    local dx = 60 * base_scale
    if fighter:GetTeamID() == TEAM.RED then
        widget:LayoutBounds("before", "bottom", x - dx, y )
    else
        widget:LayoutBounds("after", "bottom", x + dx, y )
    end
    widget:SetShown( self:GetFE():IsHUDVisible() )
    return x, y
end

function FightScreen:LayoutToolTip( tooltip_widget )
    local widget = self:GetToolTip()
    if is_instance( widget, Widget.UnitFrame ) then
        if widget.fighter:GetTeamID() == TEAM.BLUE then
            tooltip_widget:LayoutBounds( "after", "top", widget ):Offset( 32, 0 )
        else
            tooltip_widget:LayoutBounds( "before", "top", widget ):Offset( -32, 0 )
        end
        self:GetFE():ConstrainToolTipPos( tooltip_widget )
    end
end

function FightScreen:DoRender(dt)

    local debugrendersim = TheGame:IsDebugFlagged( DBG_FLAGS.SIM ) and TheGame:IsDebugFlagged( DBG_FLAGS.RENDER )
    self.fightsim:GetScene():SetDebugRender( debugrendersim )

    if debugrendersim then
        self.fightsim:DoDebugRender()
    end

    FightScreen._base.DoRender(self, dt)

    for fighter,widget in pairs(self.status_widgets) do
        local animfighter = self:GetAnimFighterEntity(fighter)
        if animfighter == nil then
            widget:Hide()
            self.speech_bubbles[ fighter ]:Hide()
        else
            self:RepositionStatusWidget( widget, animfighter )
        end
    end

    for fighter, bubble in pairs(self.speech_bubbles) do
        local animfighter = self:GetAnimFighterEntity(fighter)
        if animfighter == nil then
            bubble:Hide()
        else
            local talk_txt = animfighter.cmp.AnimFighter:GetTalkText()
            if talk_txt then
                local x, y = self:RepositionSpeechWidget( bubble, animfighter, fighter )
                bubble:SetText( talk_txt )
            
            else
                bubble:Hide()
            end
        end
    end
end

function FightScreen:DoDebugRender( dbg )
    FightScreen._base.DoDebugRender( self, dbg )

    local W, H = self:GetFE():GetScreenDims()
    local cx, cy = W/2, H/2
    self:GetFE():DrawDebugLine(cx-50,cy,cx+50,cy)
    self:GetFE():DrawDebugLine(cx,cy-50,cx,cy+50)

    if dbg:IsDebugFlagged( DBG_FLAGS.RENDER ) then
        local dbg_env = dbg:GetDebugEnv()
        if dbg_env.wx then
            local sx, sy = self.fightsim:GetCamera():WorldToScreen( dbg_env.wx, dbg_env.wy, dbg_env.wz )
            local FONT_SIZE = 16
            dbg_env.fe.scene:DrawDebugText( string.format( "<%.1f, %.1f>", dbg_env.wx, dbg_env.wz), sx + 16, sy, 0, FONT_SIZE, 1, 1, 1, 1 )
        end
    end

    local mx, my = TheGame:GetInput():GetMousePos()
    for guid, cmp in pairs( self.fightsim:GetAllComponents( AnimController )) do
        local bx1, by1, bz1, bx2, by2, bz2 = cmp:GetBoundingBox()
        local x1, y1 = self.fightsim:WorldToScreen( bx1, by1, bz1 )
        local x2, y2 = self.fightsim:WorldToScreen( bx2, by2, bz2 )

        local hit = cmp:RayTest( mx, my )
        if hit then
            self:GetFE():DrawDebugLine( x1, y1, x1, y2, 0xFFFFFFFF )
            self:GetFE():DrawDebugLine( x1, y2, x2, y2, 0xFFFFFFFF )
            self:GetFE():DrawDebugLine( x2, y2, x2, y1, 0xFFFFFFFF )
            self:GetFE():DrawDebugLine( x2, y1, x1, y1, 0xFFFFFFFF )
        else
            self:GetFE():DrawDebugLine( x1, y1, x1, y2 )
            self:GetFE():DrawDebugLine( x1, y2, x2, y2 )
            self:GetFE():DrawDebugLine( x2, y2, x2, y1 )
            self:GetFE():DrawDebugLine( x2, y1, x1, y1 )
        end
    end
end

function FightScreen:DealCards()
    self.cardspanel:DealCards()
end

function FightScreen:RefreshReason()
    self.main_overlay = self.main_overlay or TheGame:FE():FindScreen( Screen.MainOverlay )
    local title = ""
    local desc = self.scenario.reason_fn and self.scenario.reason_fn(self.battle) or self.scenario.fight_reason_txt or ""

    if title ~= self.title or self.desc ~= desc then
        self.title = title
        self.desc = desc
        desc = CardEngine.ExtractKeywords( desc, nil, self.battle, battle_defs )
        self.cardspanel:SetMinigameObjective( title, desc )
    end

end

function FightScreen:OnUpdate( dt )
    FightScreen._base.OnUpdate(self, dt)
    if self.fightsim == nil then
        return
    end

    self.plax:SetOverrideTime(self.fightsim:GetTime())

    
    self:RefreshReason()

    self.cards:AnimateCards( dt )

    self.dealt_cards:AnimateCards( dt )

    
    if self.choose_cards:IsShown() then
        self.choose_cards:AnimateCards( dt )
    end

    if self.battle:GetBattleResult() == nil then
    	self:UpdateMusicPhase()
    end

    

    if (self.battle:GetStatus() == battle_defs.STATUS.CHOOSE_CARDS and self.cardspanel.choose_cards:IsShown()) or self.cardspanel:IsWaitingForClick() or self.darken_overlay:IsShown() or self.cards:GetDragCard() then
        self:SetHoverFighter( nil )
    else
        if TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD then
            
            local mx, my = TheGame:GetInput():GetMousePos()
            local sw, sh = TheGame:FE():GetScreenDims()

            local new_hover_fighter = self:GetFighterUnderMouse()


            local apply_change = false
            
            
            if SMOOTH_MOUSE_TARGETING and TheGame:GetInput():GetLastInputDevice() ~= "key" then
    
                self.keep_hover_x = self.keep_hover_x or 0
                self.keep_hover_y = self.keep_hover_y or 0

                if self.hover_fighter == new_hover_fighter then
                    self.hover_keep_timer = (self.hover_keep_timer or 0) + dt
                    self.hover_change_timer = nil
                    self.keep_hover_x, self.keep_hover_y = mx, my
                else
                    self.hover_change_timer = (self.hover_change_timer or 0) + dt
                    self.hover_keep_timer = nil
                end
                
                local SMALL_MOVE_THRESH = .01
                local LARGE_MOVE_THRESH = .02
                
                local dx_p, dy_p = math.abs(self.keep_hover_x - mx)/sw, math.abs(self.keep_hover_y - my)/sh
                local has_moved = dx_p > SMALL_MOVE_THRESH or dx_p > SMALL_MOVE_THRESH
                local has_moved_a_lot = dx_p > LARGE_MOVE_THRESH or dx_p > LARGE_MOVE_THRESH

                local HOVER_TIMER_SOMETHING_TO_NOTHING = has_moved_a_lot and 0 or has_moved and .5 or 1
                local HOVER_TIMER_NOTHING_TO_SOMETHING = 0
                local HOVER_TIMER_SOMETHING_TO_SOMETHING = has_moved_a_lot and 0 or has_moved and .25 or .5

                if self.hover_change_timer then
                    if self.hover_fighter and self.hover_fighter:IsDead() then
                        apply_change = true
                    elseif new_hover_fighter and not self.hover_fighter and self.hover_change_timer >= HOVER_TIMER_NOTHING_TO_SOMETHING then
                        apply_change = true
                    elseif new_hover_fighter == nil and self.hover_fighter and self.hover_change_timer >= HOVER_TIMER_SOMETHING_TO_NOTHING then
                        apply_change = true
                    elseif new_hover_fighter and self.hover_fighter and self.hover_change_timer >= HOVER_TIMER_SOMETHING_TO_SOMETHING then
                        apply_change = true
                    end
                end
            else
                apply_change = true
            end

           
            if apply_change then
                self:SetHoverFighter( new_hover_fighter )
            end

        elseif TheGame:FE():GetControlMode() == CONTROL_MODE.GAMEPAD then
            local new_hover_fighter = self:GetFighterUnderMouse()
            self:SetHoverFighter( new_hover_fighter )
        end
    
    end


    self:UpdateTargetFX()
    
    local card = self:GetChosenCard() or self.cards:GetDragCard()
    if card and card:GetTarget() and card:IsAttackCard() then
        self:RefreshDamagePreviewArrows( card:GetOwner() )
    else
        self:RefreshDamagePreviewArrows( self.hover_fighter )
    end

    if card and self:GetFE():IsHUDVisible() and not card:IsFlagged( CARD_FLAGS.FLOURISH ) then
        self.cardspanel:ShowCancelInfoLabel()
    else
        self.cardspanel:HideCancelInfoLabel()
    end

    self:SettleFighters()

    if self.no_fight_popup_timer then
        self.no_fight_popup_timer = self.no_fight_popup_timer - dt
    end

    self:_DoRefreshDamagePreviews()

    if not self.cardspanel:IsChoosingCards() then
        local lx, ly, rx, ry = TheGame:GetInput():GetActiveGamepadStickVals()
        if math.length2d( rx, ry ) > 0.5 then
            self:SetReticuleMode( true )
        elseif math.length2d( lx, ly ) > 0.5 then
            self:SetReticuleMode( false )
        end
    end
end

function FightScreen:SettleFighters()
    if #self.event_queue > 0 or self.current_event then
        return
    end

    -- Failsafe: ensure fighters are settled if there are no events pending.
    for i, cmp in pairs( self.fightsim:GetAllComponents( AnimFighter )) do
        if not cmp:IsBusy() and not cmp.entity.cmp.AnimController:IsAnimQueued() then
            if not cmp:InHomePosition() or not cmp:IsIdle() then
                cmp:DoSequence( AnimFighter.ResetFighter, cmp )
            end
        end
    end
end

function FightScreen:ResumePlay( ... )
    if self.wait_coro then
        -- This should only happen if debug-interrupting.
        self.wait_coro:Stop()
        self.wait_coro = nil
    end

    -- Resume backend. This runs the combat engine until we reach a yield point (eg. when the player
    -- needs to choose a card or make a UI decision, or select points such as an AI turn.)
    local wait_status, wait_params = self.battle:ResumePlay( ... )

    local handler
    if wait_status == battle_defs.STATUS.PLAY_CARD then
        handler = self.HandlePlayCard
    elseif wait_status == battle_defs.STATUS.CHOOSE_CARDS then
        handler = self.HandleChooseCards
    elseif wait_status == battle_defs.STATUS.SYNC then
        handler = self.HandleSyncPoint
    elseif self.battle:IsDone() then
        handler = self.HandleBattleResult
    end

    -- Most UI must immediately reflect the current battle state, so that
    -- enabled status of all UI options (cards, run...) are correct.
    self:RefreshDecks()
    

    if handler then
        local n = table.maxn( wait_params )
        self.wait_coro = self:StartCoroutine( handler, self, table.unpack( wait_params, 1, n ))
    end
    if self.viz_coro == nil then
        self.viz_coro = self:StartCoroutine( self.PumpAllEvents, self, self.event_queue )
    end
end

function FightScreen:HandleSyncPoint( delay )
    if delay then
        self:Delay( delay )
    end

    self:WaitForEvents()

    self:ResumePlay()
end

function FightScreen:HandleChooseCards( deck, min_count, max_count, txt, pred, source )

    self:WaitForEvents()

    local cards
    if deck ~= self.battle:GetHandDeck() then
        cards = self:HandleChooseCardsFromDeck( deck, min_count, max_count, txt, pred, source )
    else
        cards = self:HandleChooseCardsFromHand( deck, min_count, max_count, txt, pred, source )
    end
    self.choosing_cards_from_hand = false

    self:ResumePlay( cards )
end

function FightScreen:HandleChooseCardsFromDeck( deck, min_count, max_count, txt, pred, source )
    local cards
    if is_instance( deck, CardEngine.Deck ) then
        cards = deck:GetCards()
    else
        cards = deck
    end
    cards = shallowcopy( cards )
    self.sim_paused = true
    self.floating_cards:Hide()
    self.cardspanel:HideObjective()

    local function OnClose( popup, cards )
        for i, card in ipairs( cards ) do
            local widget = popup:GetWidgetForCard( card )
            if self:FindCardWidget( card ) then
                widget:Remove()
            else
                self.floating_cards:AttachCardWidget( card, widget )
            end
        end
    end

    local popup = Screen.ChooseCardsPopup( cards, min_count, max_count, pred, Widget.BattleCard, nil, nil, OnClose )
    popup:SetTitle( txt, source )
    self:GetFE():InsertScreen( popup )

    while popup:IsOnStack() do
        coroutine.yield()
    end
    self.sim_paused = false
    self.floating_cards:Show()
    self.cardspanel:ShowObjective()


    return cards
end

function FightScreen:HandleChooseCardsFromHand( deck, min_count, max_count, txt, pred, source )
    self.choosing_cards_from_hand = true
    self:SetCardFocus()
    return self.cardspanel:DoCardChoosing( self.battle, self.cards, txt, min_count, max_count, pred, source )
end


function FightScreen:HandlePlayCard()
    while self.demo_mode do
        -- Stall here indefinitely.
        coroutine.yield()
    end

    self:SetPlayerInterfaceViz(true)

    if self.battle:CanPlayAny() then
        self.end_turn_btn:HideFocusGlow()
    elseif not TheGame:FE():IsGamepadMode() then
        self.end_turn_btn:ShowFocusGlow()
        self.end_turn_btn:SetFocus()
    end

    if TheGame:GetLocalSettings().ROBOTICS and self.accept_surrender_button:IsShown() then
        local r = math.random()
        if r < 0.5 then
            self:AcceptSurrender()
            return
        else
            local ent = self:GetAnimFighterEntity( self.battle:GetPlayerFighter() )
            if ent then
                ent.cmp.AnimFighter:SayQuip( "last_stand" )
            end
        end
    end

    local width, height = self:GetFE():GetScreenDims()
    self.cards:BeginChoosingForPlay( self.battle, 1, 1 )
    do
        -- Choosing a target will resume with the given card.
        while not self.end_turn and not self.play_card do
            coroutine.yield()
            self:DealCards()

            if TheGame:GetLocalSettings().ROBOTICS or self.battle:GetPlayerFighter() == nil then
                self:Delay( 1.0 )
                if self.battle:CountCardsPlayed() > 30 then
                    self.play_card = nil
                else
                    self.play_card = self.battle:GetHandDeck():PickCardIf( function( card ) return self.battle:CanPlayCard( card ) end )
                end
                self.end_turn = self.play_card == nil
            end
        end
    end
    self.cards:EndChoosing( self.battle )
    if not self.show_reticule then
        self.cards:DisablePreview( false )
    end
    
    local play_card = self.play_card
    if play_card then
        local can_play, reason = self.battle:CanPlayCard( play_card )
        assert_warning( can_play, "Can't play %s: %s", tostring(play_card), reason )

        -- Shake the screen if the card demands it
        if play_card.def.on_play_screen_shake then
            local shake_data = FIGHT_SHAKE_ABILITY
            if type(play_card.def.on_play_screen_shake) == "table" then
                -- The card itself details what shake to do
                shake_data = play_card.def.on_play_screen_shake
            end
            self:ScreenShake( shake_data )
        end

        -- Ensure this card doesn't animate back to hand if it is queued.
        local w = self:DetachCardWidget( play_card )
        if w then
            w:DisableToolTip( true )
            self.floating_cards:AttachCardWidget( play_card, w )
            w:SendToBack()
        end
    end

    if play_card and play_card:GetTarget() then
        if play_card:GetTarget():GetTeamID() == TEAM.BLUE then
            self.last_ally_target = play_card:GetTarget()
        else
            self.last_enemy_target = play_card:GetTarget()
        end
    end

    self.end_turn = nil
    self.play_card = nil
    self.chosen_card = nil

    self:SetPlayerInterfaceViz( false )
    self:GetFE():ClearFocusWidget()

    self:ResumePlay( play_card )
end

function FightScreen:HandleBattleResult()

    self:WaitForEvents()

    print( "HandleBattleResult()" )
    
    self.dealt_cards:ClearCards()

    self:DelayAndAnimate( nil, true )

    local winning_team
    if self.battle:GetBattleResult() == BATTLE_RESULT.WON then
        winning_team = self.battle:GetPlayerTeam()
        self:SetMusicPhase( 5 )

    else
        winning_team = self.battle:GetEnemyTeam()
        AUDIO:PlayEvent(SoundEvents.battle_lost )
        self:SetMusicPhase( 6 )
    end

    local victory = winning_team == self.battle:GetPlayerTeam()
    local victory_txt
    if victory and self.scenario:GetSurvivalTurns() ~= nil then
        victory_txt = LOC"UI.FIGHT_SCREEN.SURVIVED"
    end
    self.cardspanel:ShowEndBanner( victory, victory_txt )

    for k, fighter in self.battle:ActiveFighters() do
        local animfighter = self:GetAnimFighterEntity( fighter )
        if animfighter then
            if fighter:GetTeam() == winning_team then
                animfighter.cmp.AnimFighter:Celebrate()
            else
                animfighter.cmp.AnimFighter:Humiliate()
            end
        end
    end

    self:Delay( 1.5 )

    if self.main_overlay then self.main_overlay:SetModeOverride( nil ) end

    if self.on_done then
        self.on_done( self.battle:GetBattleResult() )
    end
end

function FightScreen:GetDebugEnv( debug_env )
    debug_env.scenario = self.scenario
    debug_env.sim = self.fightsim
    debug_env.battle = self.battle
    debug_env.player = self.battle:GetPlayerFighter()
    debug_env.enemy = self.battle:GetEnemyTeam():Primary()
    debug_env.plax = self.plax
    debug_env.CARD_FLAGS = self.battle.CARD_FLAGS

    if self.fightsim then
        debug_env.cam = self.fightsim:GetCamera()
        debug_env.wx, debug_env.wy, debug_env.wz = self.fightsim:GetGroundPosUnderMouse()
    end
end

function FightScreen:SetSimTimeScale(scale)
    self.fightsim:SetTimeScale(scale or 1)
end

function FightScreen:GetChosenCard()
    return self.chosen_card or self.cards:GetChosenCard()
end

function FightScreen:UpdateTargetFX()
    engine.inst:ProfilerPush("FightScreen:UpdateTargetFX")
    local card = self:GetChosenCard()
    if card and self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
        
        
        --[[if TheGame:FE():GetControlMode() == CONTROL_MODE.TOUCH or TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD then
            self:SetSimTimeScale(0)
        else
            self:SetSimTimeScale(1)
        end-]]
        
        
        local auto_target = self:AssignAutoTarget( card )
        if auto_target then
            self:OnClickFighter()
        else
            local source_widget = self:FindCardWidget( card )
            if not source_widget then
                return
            end
            
            self.target_arrow:AttachSource( source_widget )

            local target_fighter = self.target_fighter or self.hover_fighter
            if not target_fighter then
                self.target_arrow:AttachMouseTarget( TARGET_STATE.NONE )
            else
                local can_target, reason = self.battle:CanPlayCard( card, target_fighter )
                local status_widget = self.status_widgets[ target_fighter ]
                if status_widget then
                    if can_target then
                        local x, y = status_widget:GetMidpoint()
                        x, y = self:TransformFromWidget( status_widget, x, y )
                        self.target_arrow:AttachToPosition( x, y, TARGET_STATE.VALID )

                    else
                        local x, y = status_widget:GetMidpoint()
                        x, y = self:TransformFromWidget( status_widget, x, y )
                        self.target_arrow:AttachToPosition( x, y, TARGET_STATE.INVALID )
                    end
                end
            end
        end
    else
        self.target_arrow:Detach()
        --self:SetSimTimeScale(1)
    end
    engine.inst:ProfilerPop()
end

function FightScreen:SetHoverFighter( new_hover_fighter )
    self.hover_change_timer = nil
    self.hover_keep_timer = nil

    if new_hover_fighter == self.hover_fighter then
        return
    end

    if self.hover_fighter then
        local ent = self:GetAnimFighterEntity( self.hover_fighter )
        if ent then
            ent.cmp.AnimFighter:HideSelectFX()
        end
    end

    self.hover_fighter = new_hover_fighter

    if self.hover_fighter then
        local ent = self:GetAnimFighterEntity( new_hover_fighter )
        ent.cmp.AnimFighter:PlaySelectFX()
        AUDIO:PlayEvent(SoundEvents.battle_gen_mouseover_battleview)
    end


    for k, v in ipairs( self.status_root.children ) do
        v:SetHighlighted( v.fighter == self.hover_fighter )
    end

    self:RefreshChosenCard()
    self.fe:RefreshControlOptions()
end

function FightScreen:CancelChosenCard()
    if self.chosen_card then
        local w = self:FindCardWidget( self.chosen_card )
        if w then
            w:StartCoroutine(function()
                w:AnimateToWidget( self.flourish_bar )
            end )
        end
        self.player_preview:RemovePreview( self.chosen_card )
        self:RefreshDamagePreviews()
        self.chosen_card:AssignTarget( nil )
        self.chosen_card = nil
        self:SetCardFocus()
        AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
        return true
    end

    -- Check for fighter CTRL+Clicks
    local card = self:GetChosenCard()
    if card then 
        -- Unselect selected card
        self.cards:ChooseCard( card )
        card:AssignTarget( nil )
        AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
        return true
    end

    if self.cardspanel:IsCardSelected() then
        TheGame:FE():ClearFocusWidget()
        return true
    end

    return false
end

function FightScreen:RefreshChosenCard()
    local card = self:GetChosenCard() or self.cards:GetDragCard()
    if card then
        local auto_target = self:AssignAutoTarget( card )
        if auto_target then
            self.battle:AssignPrimaryTarget( card )
        elseif self.hover_fighter and self.battle:CanPlayCard( card, self.hover_fighter ) then
            self.battle:AssignPrimaryTarget( card, self.hover_fighter )
            -- card:AssignTarget( self.hover_fighter )
        else
            card:AssignTarget( nil )
        end

        self.player_preview:RemovePreview( card )
        -- only preview this card if it doesn't need a target
        if auto_target or card:GetTarget() then
            self.player_preview:PreviewCard( card )
        end
        self:RefreshDamagePreviews()

        local w = self:FindCardWidget( card )
        if w then
            w:RefreshCard( card, self.battle )
        else
            print( "Could not find card", card, card.deck )
        end
    end
end

function FightScreen:SetPlayerInterfaceViz(viz)
    if viz then
        if self.battle:GetScenario():CanFlee() then
            if self.battle:GetTurns() < MIN_RUN_TURN then
                self.concede_btn:Disable():SetToolTip(loc.format( LOC"UI.FIGHT_SCREEN.RUN_TOO_EARLY", MIN_RUN_TURN))
            else
                self.concede_btn:Enable():SetToolTip(LOC"UI.FIGHT_SCREEN.RUN")
            end
        end
    
        if not CheckBits( self.battle:GetFlags(), BATTLE_FLAGS.NO_SURRENDER ) then
            local enemy_team = self.battle:GetEnemyTeam()
            if enemy_team:GetTeamStatus() == FIGHT_STATUS.SURRENDER then
                self.end_turn_btn:AnimateOut()
                self.concede_btn:AnimateOut()
                self.flourish_bar:AnimateOut()
                self.accept_surrender_button:AnimateIn()
                    :ShowFocusGlow()
            else
                self.accept_surrender_button:AnimateOut()
                    :HideFocusGlow()
            end

        else
            local enemy_team = self.battle:GetEnemyTeam()
            if enemy_team:GetTeamStatus() == FIGHT_STATUS.SURRENDER then
                self.end_turn_btn:AnimateOut()
                self.concede_btn:AnimateOut()
            end
        end

    else
        self.accept_surrender_button:Hide()
            :HideFocusGlow()
    end
end

function FightScreen:GetAnimFighterEntity(fighter)
    return self.fightsim.mgr:GetAnimFighterByFighter(fighter)
end

function FightScreen:AddFighter(fighter)
    assert( not self:GetAnimFighterEntity( fighter ) )

    local ent = self.fightsim.mgr:AddFighter(fighter)
    local cmp = ent.cmp.AnimFighter

    local w = Widget.UnitFrame(fighter, ent):SetTintAlpha( 0 ) -- revealed in JoinBattle
    self.status_root:AddChild( w )
        :SetFocusable()
        :ShowToolTipOnFocus( true )
    self.status_widgets[fighter] = w

    -- Used for tag-in cards
    if fighter.home_x then
        cmp:SetHomePosition(fighter.home_x, 0)
    end

    cmp:DoSequence( cmp.JoinBattle, cmp, w, self )

    local bubble = Widget.TextBox():SetColour(UICOLOURS.COMIC_PANEL_BG):SetMaxWidth(250):SetFontSize(26):SetLineHeight(18):SetPadding(-10)
    bubble:SetLayoutScale( LAYOUT_SCALE[ self.fe:GetScreenMode() ] )
    bubble:SetSpeechBubble( fighter.agent, fighter:GetTeamID() == TEAM.BLUE and bubble.STYLE.SPEECH_LEFT or bubble.STYLE.SPEECH_RIGHT )
    bubble:Hide()
    self.speech_bubble_root:AddChild( bubble )
    self.speech_bubbles[fighter] = bubble

    -- Not sure WHY this is necessary since the widget starts out hidden and only is shown after its
    -- x, y is updated in DoRender, but without initializing there's a frame of the widget at screen
    -- origin ???!
    self:RepositionStatusWidget( w, ent )

    self.main_overlay = self.main_overlay or TheGame:FE():FindScreen( Screen.MainOverlay )
    if fighter:GetTeamID() == TEAM.BLUE then
        self.main_overlay:SetFriendReserves( self.battle, TEAM.BLUE )
    else
        self.main_overlay:SetEnemyReserves( self.battle, TEAM.RED )
    end

    return ent
end

function FightScreen:RemoveFighter(fighter)
    
    local ent = self:GetAnimFighterEntity( fighter )
    if ent then
        ent.cmp.AnimFighter:SetStatusWidget( nil )
    end

    self.status_widgets[ fighter ]:Remove()
    self.status_widgets[ fighter ] = nil

    if self.hover_fighter == fighter then
        self.hover_fighter = nil
    end
end

function FightScreen:SetMusicPhase(phase)
    if phase ~= self.music_phase then
        self.music_phase = phase
        TheGame:GetMusic():SetParameter("phase", self.music_phase)
    end
end

function FightScreen:UpdateMusicPhase()
    local phase = 1
    local player_team, enemy_team = self.battle:GetPlayerTeam(), self.battle:GetEnemyTeam()
    local strength_player, strength_player_max = player_team:CalculateStrength()
    local strength_enemy, strength_enemy_max = enemy_team:CalculateStrength()
    local player_p = strength_player / strength_player_max
    local enemy_p = strength_enemy / strength_enemy_max

    
    if player_p < .333 then
        phase = 3
    elseif enemy_p < .333 then
        phase = 4
    elseif player_p < .67 or enemy_p < .67 then
        phase = 2
    end

    if self.music_phase == 2 then
        if phase == 1 then
            phase = self.music_phase
        end
    elseif self.music_phase == 3 or self.music_phase == 4 then
        if phase <= 4 then
            phase = self.music_phase
        end
    end
    self:SetMusicPhase(phase)
end

function FightScreen:Setup()
    TheGame:GetMusic():SetParameter("phase", 1)
    
    if self.scenario:GetLocation() then
        self.plax:SetPlaxData( self.scenario:GetLocation():GetPlax())
    end


    self:ApplyColourCorrectionValues((require "content/colour_settings")[self:GetColourTintName()])
 
    if TheGame:GetDebug() then
        self.viz_break = TheGame:GetDebugEnv().viz_break
    end

    Content.InvokeModAPI( "OnBattleSetup", self, self.battle )
    
    self:StartCoroutine( function()
        self.cardspanel:Setup()
        self.cardspanel:AnimateIn()
        self.cardspanel:DelayIfAnimating()
        self:ResumePlay( self.battle )
    end )
end

function FightScreen:SetupDemo()
    local function OnClickExit()
        self:GetFE():PopScreen( self )
    end

    local exit_btn = Widget.PanelButton( LOC"UI.FIGHT_SCREEN.EXIT_DEMO", OnClickExit, 100, 40 )
    exit_btn:SetAnchors( "center", "bottom" ):SetPos( 0, 100 )
    self:AddChild( exit_btn )

    self:StartCoroutine( function()
        self.cardspanel:Setup()
        self.end_turn_btn:AnimateOut()
        self.time_indicator:Hide()
        self.cardspanel:AnimateIn()
        self.cardspanel:DelayIfAnimating()
        self:ResumePlay( self.battle )
    end )
end

function FightScreen:GetMusicEvent()
    local music_event
    if self.scenario:IsBossFight() then
        music_event = TheGame:LookupPlayerMusic( "boss_music" )
    else
        music_event = TheGame:LookupPlayerMusic( "battle_music" )
    end
    return music_event or "event:/music/adaptive_battle"
end

local function SortByDepth( ent1, ent2 )
    local x0, y0, z0 = ent1:GetLocalPosition()
    local x1, y1, z1 = ent2:GetLocalPosition()
    return z0 < z1
end

-- 
-- Returns the fighter that's currently hovered, or focused on
function FightScreen:GetFighterUnderMouse()
    
        -- Try and get the focused one
    local hover = self:GetFE():GetFocusWidget()
    local frame
    if hover then
        frame = hover:GetAncestor( Widget.UnitFrame )
    end

    if frame == nil then
        hover = self:GetFE():GetHoverWidget()
        frame = hover and hover:GetAncestor( Widget.UnitFrame )
    end

    if frame then
        local card = self:GetChosenCard() or self.cards:GetDragCard()
        if card == nil or is_instance( hover, Widget.UnitFrame ) then
            local hover_fighter
            if hover and hover.card then
                hover_fighter = hover.card.owner
            else
                hover_fighter = frame.fighter
            end
            
            local ent = hover_fighter and self:GetAnimFighterEntity( hover_fighter )
            if ent then
                return hover_fighter
            end
        end
    end

    if not TheGame:FE():IsGamepadMode() or self.fe:ReticuleActive() then
        local mx, my
        if self.fe:ReticuleActive() then
            mx, my = self.fe:GetReticulePos()
        else
            mx, my = TheGame:GetInput():GetMousePos()
        end
        
        local t = ObtainWorkTable()
        for guid, cmp in pairs( self.fightsim:GetAllComponents( AnimFighter )) do
            if cmp.fighter:GetStatus() ~= FIGHT_STATUS.DEAD then
                table.insert( t, cmp.entity )
            end
        end
        table.sort( t, SortByDepth )
        
        local hover_ent, best_score
        for i, ent in ipairs( t ) do
            local cmp = ent.cmp.AnimController       
            local fighter = ent.cmp.AnimFighter.fighter
            local bx1, by1, bz1, bx2, by2, bz2 = cmp:GetBoundingBox()
            local x1, y1 = self.fightsim:WorldToScreen( bx1, by1, bz1 )
            local x2, y2 = self.fightsim:WorldToScreen( bx2, by2, bz2 )
            if x1 > x2 then
                x1, x2 = x2, x1
            end
            if y1 > y2 then
                y1, y2 = y2, y1
            end
            
            
            local hit = cmp:RayTest( mx, my, fighter.fight_data.ignore_anim_symbols )
            if hit then
                local score = distance( mx, my, (x1 + x2)/2, (y1 + y2)/2 ) / i
                if best_score == nil or score < best_score then
                    hover_ent = ent
                    best_score = score
                end
            end
        end
        ReleaseWorkTable( t )
        
        if hover_ent then
            return hover_ent.cmp.AnimFighter.fighter
        end
    end
end

-- NOTE: fighter can be nil if this is an auto_target card.
function FightScreen:OnClickFighter(fighter)

    if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD then
        local card = self:GetChosenCard()
        if card then
            local can_target, reason = self.battle:CanPlayCard( card, fighter )
            if can_target then
                AUDIO:PlayEvent( SoundEvents.battle_card_commit )
                self.play_card = card
                self.battle:AssignPrimaryTarget( card, fighter )
                self:SetCardFocus()
            else
                if reason then
                    reason = GetLocalizedConvoCommonString( reason )
                    self.cardspanel:ShowBanner( reason, nil, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
                end

                AUDIO:PlayEvent( SoundEvents.battle_gen_click_invalid )
            end
            self:AvoidPopup()
            return true
        end
    end
end

function FightScreen:AvoidPopup()
    self.no_fight_popup_timer = .5
end

function FightScreen:CanShowFighterPopup() 
    return self.no_fight_popup_timer == nil or self.no_fight_popup_timer <= 0
end

function FightScreen:ShowFighterInfoPopup( selected_fighter )
    local fighter_popup = Screen.FighterInfoPopup( selected_fighter )
    TheGame:FE():InsertScreen( fighter_popup )
end

function FightScreen:AssignAutoTarget( card )
    if card then
        local auto_target = card.auto_target
        local current_mod = self.battle.target_mod_accumulator:CalculateValue( card.target_mod, card, card:GetTarget())
        auto_target = auto_target or (card.target_type == TARGET_TYPE.SELF)
        auto_target = auto_target or (current_mod == TARGET_MOD.TEAM)
        auto_target = auto_target or (card.target_type == TARGET_TYPE.FRIENDLY_OR_SELF and current_mod ~= TARGET_MOD.SINGLE and current_mod ~= TARGET_MOD.CUSTOM)
        auto_target = auto_target or (card.target_type == TARGET_TYPE.ENEMY and current_mod ~= TARGET_MOD.SINGLE and current_mod ~= TARGET_MOD.CUSTOM)

        return auto_target
    end
end

function FightScreen:HandleControlDown(control, device)

    local focused_fighter = self:GetFighterUnderMouse()
    if control:Has( Controls.Digital.MENU_ACCEPT ) and focused_fighter then
        if TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
            local ent = self:GetAnimFighterEntity( focused_fighter )
            DBG( ent.cmp.AnimFighter )
            return true
        end
    end

    local hotkey_idx = UIHelpers.FindHotkeyIdx( control )
    if hotkey_idx then
        local card = self.cardspanel:GetCardAt( hotkey_idx )
        if card and self.cardspanel:TryChooseCard( card ) then
            return true
        end
    end

    return FightScreen._base.HandleControlDown( self, control, device )
end

function FightScreen:HandleControlUp(control, device)

    if control:Has( Controls.Digital.MENU_DELETE ) then
        if not self:GetFE():FindScreen( "Screen.LocationScreen" ) then
            -- Fixes post-close clear hover crash.
            self:ClearHover()

            TheGame:FE():PopScreen( self )
            TheGame:GetDebug():ClearPanels()

            package.loaded[ "ui/screens/fightscreen" ] = nil
            package.loaded[ "ui/widgets/cardwidget" ] = nil
            package.loaded[ "ui/widgets/cardarrangement" ] = nil
            package.loaded[ "ui/widgets/targetarrow" ] = nil
            package.loaded[ "ui/widgets/healthbar" ] = nil
            package.loaded[ "ui/widgets/unitframe" ] = nil
            package.loaded[ "ui/widgets/battlecard" ] = nil
            package.loaded[ "components/animfighter" ] = nil
            package.loaded[ "fightsim" ] = nil
            package.loaded[ "battle/fighter_behaviour" ] = nil
            package.loaded[ "battle/damage_accumulator" ] = nil
            package.loaded[ "battle/damage_preview" ] = nil
            package.loaded[ "battle/fighter" ] = nil
            package.loaded[ "battle/attack" ] = nil
            package.loaded[ "battle/team" ] = nil
            package.loaded[ "cards/scalar_accumulator" ] = nil
            package.loaded[ "cards/card_engine" ] = nil
            package.loaded[ "cards/card_picker" ] = nil

            Content.ReloadConversation()

            require "fightsim"
            require "battle/fighter"
            require "battle/team"
            require "ui/widgets/cardwidget"
            require "ui/screens/fightscreen"
            require "components/animfighter"

            TheGame:FE():InsertScreen( FightScreen() )
        else
            self:GetFE():InsertScreen(Screen.PauseMenu())
        end

        return true
    end

    if control:Has( Controls.Digital.MENU_ACCEPT ) and not TheGame:GetInput():IsModifierAny() then
        if  TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD or TheGame:FE():IsTouchMode() then
            local popup = self:GetFE():FindScreen( Screen.ChooseCardsPopup )
            if popup and popup:IsOverlay() then
                popup:TogglePopup()
                return true
            end
        end

    elseif control:Has( Controls.Digital.MENU_ALT ) then
        if self.hover_fighter and self:CanShowFighterPopup() then
            self:ShowFighterInfoPopup(self.hover_fighter)
        end
    end

    if TheGame:GetDebug() and TheGame:GetInput():IsControlDown( Controls.Digital.PLAX_EDIT_PAN ) then
        if control:Has( Controls.Digital.ZOOM_IN_STEP) then
            self.zoom = (self.zoom or 1.0) + 0.1
            self.plax:SetScale( self.zoom, self.zoom )
            return true

        elseif control:Has( Controls.Digital.ZOOM_OUT_STEP) then
            self.zoom = (self.zoom or 1.0) - 0.1
            self.plax:SetScale( self.zoom, self.zoom )
            return true
        end
    end

    if self.cards:HandleControlUp( control, device ) then
        return true
    end

    if control:Has(Controls.Digital.PAUSE) then
        if TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent() then
            self:GetFE():InsertScreen(Screen.PauseMenu())
        end
        return true
    end
    
    return FightScreen._base.HandleControlUp( self, control, device )
end

function FightScreen:AcceptSurrender()
    AUDIO:PlayEvent(SoundEvents.battle_accept_surrender)
    self.battle:Win()
    self.accept_surrender_button:AnimateOut()
    self:ResumePlay()
end

function FightScreen:ActivateFlourish()
    if self.battle:GetStatus() == battle_defs.STATUS.PLAY_CARD and not self:GetChosenCard() then
        self.sim_paused = true
        local flourishes = TheGame:GetGameProfile():GetUsableFlourishes( self.battle:GetPlayerFighter().agent:GetContentID(), "BATTLE")
        for i,id in ipairs(flourishes) do
            flourishes[i] = Battle.Card(id, self.battle:GetPlayerFighter())
        end

        self.cardspanel:ActivateFlourish( flourishes, function( card )
            self.chosen_card = card
            if card then
                self:SetFighterFocus()
            end
            self.sim_paused = false
        end )
    end
end
