require "ui/widgets/screen"
require "fightsim"

-------------------------------------------------------------------

local FightScreenBase = class("Screen.FightScreenBase", Widget.Screen)

function FightScreenBase:init()
    Widget.Screen.init(self)

    self.fightsim = FightSim()

    self.plax = Widget.Plax():SetName("FIGHTSCREEN_PLAX"):PostProcess( true ):SetGameMode( GAME_MODE.BATTLE ):SetAnchors( "center", "center" )
    self:AddChild( self.plax )
    self.plax.ignore_input = true

    self.img = Widget.Image():SetSize(RES_X, RES_Y):SetAnchors("fill", "fill"):SetEffect(global_shaders.UI_BLOOMMAP_HSVADJUST)
    self.img:SetBloom( 1.0 )
    self:AddChildren{
        self.img,
    }
    self.colour_correct_constants = 
    {
        shadows = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST", 0,1,1),
        midtones = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST", 0,1,1),
        highlights = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST", 0,1,1),
        shadows_col = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST_COL", 1,1,1),
        midtones_col = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST_COL", 1,1,1),
        highlights_col = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST_COL", 1,1,1),
    }

    for k,v in pairs(self.colour_correct_constants) do
        self.img.model:SetShaderConstant(v)
    end    

    self.scene_render_node = engine.rendering.SceneRenderNode("FightScreenBase")

    local w, h = TheGame:FE():GetScreenDims()
    self:SetSize(w,h)
end



function FightScreenBase:SetSize(w, h)
    self.pending_w = w
    self.pending_h = h
end


function FightScreenBase:UpdateRenderTarget()

    if self.pending_w ~= self.w or self.pending_h ~= self.h or self.bloom_enabled ~= TheGame.bloom_enabled then
        self.w = self.pending_w or self.w
        self.h = self.pending_h or self.h

        self.bloom_enabled = TheGame.bloom_enabled
 
        local texw, texh = self.w, self.h

		local fmt0 = PIXEL_FORMAT.A8R8G8B8;
		local fmt1 = -1;
		if self.bloom_enabled then
			fmt1 = fmt0
		end

        self.scene_render_node:SetRenderFormat(math.round(texw), math.round(texh), fmt0, fmt1, true, TheGame.msaa)

        self.fightsim:SetRenderSize(texw, texh)
    end

    return self
end


function FightScreenBase:OnScreenResize(w, h)
    FightScreenBase._base.OnScreenResize(self, w, h)
    
    self:SetSize(w, h)
end


function FightScreenBase:ShouldDebugRender()
    return TheGame:IsDebugFlagged( DBG_FLAGS.SIM ) and TheGame:IsDebugFlagged( DBG_FLAGS.RENDER )
end

function FightScreenBase:DoRender(dt)

    self:UpdateRenderTarget()
    
    if self:ShouldDebugRender() then
        self.fightsim:GetScene():SetDebugRender( true )
        self.fightsim:DoDebugRender()
    else
        self.fightsim:GetScene():SetDebugRender( false )
    end

    local render_node = self.scene_render_node
    render_node:SetFlatScene(self.fightsim:GetScene(), self.fightsim:GetRoot(), self.fightsim:GetCameraInfo(), (1 << RENDER_PASS_OPAQUE) | (1 << RENDER_PASS_ALPHA) )
    engine.inst:SetClearColour( 0, 0, 0, 0 )
    render_node:ClearTexture()
    render_node:RenderToTexture( TheGame:GetLevelOfDetail() )

    local bright = render_node:GetOutputTexture(1)
    self.img:SetTexture(render_node:GetOutputTexture(), 0)

    if engine.inst:IsScreenVerticallyFlipped() then
        self.img:SetScale(1,-1)
    end
    
    self.img:SetBlendMode(3)    -- Premultiplied_Premultiplied. This has to be the blend mode because all characters and effects are rendered on top of a black screen, resulting in everything being blended to black, essentially pre-multiplying the colour components
    if bright then
        self.img:SetTexture(bright, 1)
    end

    for guid, cmp in pairs( self.fightsim:GetAllComponents( AnimFighter )) do
        cmp:OnRender( dt )
    end
    for guid, cmp in pairs( self.fightsim:GetAllComponents( AnimController )) do
        cmp:OnRender( dt )
    end
    
    if self.top then
        self.plax:RenderToScene()
    end
end

function FightScreenBase:ScreenShake( shake_params )
    if TheGame:CamShakeEnabled() then
        local mag, speed, time = table.unpack( shake_params )
        self.shaker = CamShake( mag, speed, time )
    end
end

function FightScreenBase:PanToScreen( mx, my, zoom )
    local px0, py0 = self.plax:TransformFromWorld( mx, my )
    self.plax:SetPos( -px0 * (self.zoom or 1), -py0 * (self.zoom or 1))
    if zoom then
        self.plax:SetScale( zoom, zoom )
        self.zoom = zoom
    end
end

function FightScreenBase:PanTo( wx, wy, wz, zoom, duration )
    zoom = zoom or self.plax:GetScale()

    if duration then
        local wx0, wy0 = self.plax:GetPos()
        local zoom0 = self.plax:GetScale()
        self.pan_start = { wx0, wy0, zoom0 }

        local sx, sy = self.fightsim:GetCamera():WorldToScreen( wx, wy, wz )
        local xplax, yplax = self.plax:TransformFromWorld( sx, sy )
        self.pan_end = { -xplax * zoom, -yplax * zoom, zoom }

        self.pan_duration = duration or 1.0
        self.camera_dt = 0
    else
        local sx, sy = self.fightsim:GetCamera():WorldToScreen( wx, wy, wz )
        local xplax, yplax = self.plax:TransformFromWorld( sx, sy )

        self.plax:SetPos( -xplax * zoom, -yplax * zoom ):SetScale( zoom )
    end
end

function FightScreenBase:ResetCamera( duration )
    local basey = FIGHT_CAM_ORTHO_DEFAULTS.basey
    self:PanTo( 0, basey, 0, 1.0, duration )
end

function FightScreenBase:WaitForCamera()
    while self.pan_duration do
        coroutine.yield()
    end
end

function FightScreenBase:SyncCamera()
    -- sim cam needs to track the plax transform.
    local cam = self.fightsim:GetCamera()
    local plax_scale = self.plax:GetResolvedScale()
    cam:SetOrthoScale( FIGHT_CAM_ORTHO_DEFAULTS.orthoscale * plax_scale )
    
    -- World coordinates at cam origin.
    cam:MoveTo( 0, 0, 0, 0 )
    cam:WarpToTargets()

    -- <xplax, yplax> is the plax origin in screen space
    local dx, dy = self.plax:GetPlaxShift()
    local xplax, yplax = self.plax:TransformToWorld( 0, 0 )

    -- Project that to the ground using the default cam to find where teh cam SHOULD be.
    local wx1, wy1, wz1 = self.fightsim:GetGroundPosAt( xplax, yplax )
    if wx1 then
        local basey = FIGHT_CAM_ORTHO_DEFAULTS.basey
        cam:MoveTo( -wx1 + dx, -wy1 + basey + dy, -wz1, 0 )
        cam:WarpToTargets()
    end
end

function FightScreenBase:UpdatePanning( dt )
    self.camera_dt = (self.camera_dt or 0) + dt

    local t = clamp( self.camera_dt / self.pan_duration, 0, 1.0 )

    local easefn = easing.outQuad
    local x0, y0, zoom0 = table.unpack( self.pan_start )
    local x1, y1, zoom1 = table.unpack( self.pan_end )
    local xplax = easefn( t, x0, x1 - x0, 1.0 )
    local yplax = easefn( t, y0, y1 - y0, 1.0 )
    local zoom = easefn( t, zoom0, zoom1 - zoom0, 1.0 )

    self.plax:SetPos( xplax, yplax ):SetScale( zoom )

    if t >= 1.0 then
        self.pan_duration = nil
        self.pan_start, self.pan_end = nil, nil
    end
end

function FightScreenBase:OnUpdate( dt )
    self.fightsim:DoFullSimUpdate(dt, self.sim_paused )

    if self.pan_duration then
        self:UpdatePanning( dt )
    end

    if self.shaker then
        if self.shaker:IsDone() then
            self.shakex, self.shakey = nil, nil
            self.shaker = nil
        else
            self.shakex, self.shakey = self.shaker:UpdateShake( dt )
        end
        self.plax:SetPlaxShift( self.shakex or 0, self.shakey or 0 )
    end

    self:SyncCamera()

    return FightScreenBase._base.OnUpdate( self, dt )
end


function FightScreenBase:GetColourTintName()
    return "DEFAULT"
end




function FightScreenBase:ApplyColourCorrectionValues(vals)
    if vals then
        self.colour_correct_constants.shadows:SetValue(table.unpack(vals.shadows.hsv))
        self.colour_correct_constants.midtones:SetValue(table.unpack(vals.midtones.hsv))
        self.colour_correct_constants.highlights:SetValue(table.unpack(vals.highlights.hsv))
        self.colour_correct_constants.shadows_col:SetValue(table.unpack(vals.shadows.colour))
        self.colour_correct_constants.midtones_col:SetValue(table.unpack(vals.midtones.colour))
        self.colour_correct_constants.highlights_col:SetValue(table.unpack(vals.highlights.colour))
    else
        self.colour_correct_constants.shadows:SetValue(0,1,1)
        self.colour_correct_constants.midtones:SetValue(0,1,1)
        self.colour_correct_constants.highlights:SetValue(0,1,1)
        self.colour_correct_constants.shadows_col:SetValue(1,1,1)
        self.colour_correct_constants.midtones_col:SetValue(1,1,1)
        self.colour_correct_constants.highlights_col:SetValue(1,1,1)
    end
end

function FightScreenBase:OnClose()
    self.fightsim:OnDestroy()
    self.fightsim = nil
end
