require "ui/widgets/installed_grafts"
require "ui/widgets/segmentedbar"
require "ui/widgets/radialprogress"
require "ui/screens/upgradegraftpopup"

local GraftCollectionScreen = class( "Screen.GraftCollectionScreen", Widget.Screen )

local PANEL_W = RES_X * 0.95
local PANEL_H = RES_Y * 0.70
local PANEL_PADDING = 8
local CONTAINER_W = RES_X * 0.75
local CONTAINER_H = PANEL_H - PANEL_PADDING*2
local GRAFT_SLOT_SIZE = 110
local DETAILS_W = RES_X*0.6

GraftCollectionScreen.CONTROL_MAP = {
    {
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self)
            AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
            self:FadeOut()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end
    }
}


function GraftCollectionScreen:init( owner, locked )
    GraftCollectionScreen._base.init( self )

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
    
    self.locked = locked

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.GRAFT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Setup character anim display
    self.owner = owner or TheGame:GetGameState():GetPlayerAgent()
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :SetScale( 1, 1 )
        :Refresh( self.owner )
    self.character:SetHoverOutlineColour( UICOLOURS.GRAFT )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "thought", true )
        end
    end )
    self.character:SetAutoLoopEnabled( false )
    self.character:Emote( "thought", true )
    self.character:SetPos( -RES_X*0.35, -350 )

    -- Our scroll container
    self.scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( 0 )
        :SetVirtualTopMargin( SPACING.M1*4 )
        :SetVirtualBottomMargin( SPACING.M1*4 )
        :SetScrollBarOuterMargin( 0 )
        :ShowBar( SCROLLBAR.NEVER )
        :SetSize( DETAILS_W+SPACING.M1*4, self.height )
    self.content = self.scroll:AddScrollChild( Widget() )

    -- Show text info
    self.text_content = self.content:AddChild(Widget())
    self.title = self.text_content:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( loc.upper( loc.format( LOC"UI.GRAFT_COLLECTION.TITLE", self.owner:GetName() ) ) )
        :SetGlyphColour( UICOLOURS.GRAFT )
        :SetAutoSize(DETAILS_W)
        :LeftAlign()
        :SetWordWrap(true)
        :Bloom(0.15)
    self.subtitle = self.text_content:AddChild( Widget.Label("title", FONT_SIZE.BODY_TEXT ) )
        :SetText( loc.format( LOC"UI.GRAFT_COLLECTION.DESC", self.owner:GetName() ) )
        :SetGlyphColour( UICOLOURS.GRAFT )
        :SetAutoSize(DETAILS_W)
        :LeftAlign()
        :SetWordWrap(true)
        :SetTintAlpha( 0.8 )
        :Bloom(0.1)

    -- Setup grafts display
    self.graft_root = self.content:AddChild(Widget())
    self.graft_widgets = {
        [GRAFT_TYPE.COMBAT] = self.graft_root:AddChild( Widget.GraftScreenSlotContainer(GRAFT_TYPE.COMBAT, self.owner, GRAFT_SLOT_SIZE, self.locked) ),
        [GRAFT_TYPE.NEGOTIATION] = self.graft_root:AddChild( Widget.GraftScreenSlotContainer(GRAFT_TYPE.NEGOTIATION, self.owner, GRAFT_SLOT_SIZE, self.locked) ),
    }
    self.graft_widgets[GRAFT_TYPE.COMBAT]:SetWidth( (DETAILS_W-SPACING.M1)/2 )
    self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:SetWidth( (DETAILS_W-SPACING.M1)/2 )

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", 
        function()
            AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
            self:FadeOut()
        end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    
    self:Refresh()
end

function GraftCollectionScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.scroll:ScrollUp()
        return true
    end
end

function GraftCollectionScreen:Refresh()
    -- Update tab counts
    local installed_grafts, max_grafts
    installed_grafts, max_grafts = self.graft_widgets[GRAFT_TYPE.COMBAT]:GetGraftCount()
    installed_grafts, max_grafts = self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:GetGraftCount()

    -- Show grafts
    self.graft_widgets[GRAFT_TYPE.COMBAT]:Refresh():Layout()
    self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:Refresh():Layout()

    self:Layout()
end

function GraftCollectionScreen:OnControlModeChange( cm, device_id )
    GraftCollectionScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_left:Show()
        self.scroll:ShowBar( SCROLLBAR.IF_NEEDED )
    end
    if cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
        self.scroll:ShowBar( SCROLLBAR.NEVER )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
        self.scroll:ShowBar( SCROLLBAR.NEVER )

        local first = self:GetDefaultFocus()
        if first then first:SetFocus() end
    end
end

function GraftCollectionScreen:OnScreenModeChange( sm )
    GraftCollectionScreen._base.OnScreenModeChange( self, sm )

    local content_w = DETAILS_W
    if sm == SCREEN_MODE.MONITOR then
    elseif sm == SCREEN_MODE.TV then
        content_w = content_w * 0.9
    elseif sm == SCREEN_MODE.SMALL then
        content_w = content_w * 0.8
    end

    self.title:SetAutoSize(content_w)
    self.subtitle:SetAutoSize(content_w)
    self.graft_widgets[GRAFT_TYPE.COMBAT]:SetWidth( (content_w-SPACING.M1)/2 )
    self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:SetWidth( (content_w-SPACING.M1)/2 )

    self.content:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end

function GraftCollectionScreen:GetDefaultFocus()
    return ( self.graft_widgets
            and self.graft_widgets[GRAFT_TYPE.NEGOTIATION]
            and self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:IsShown()
            and self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:GetDefaultFocus() )
        or ( self.graft_widgets
            and self.graft_widgets[GRAFT_TYPE.COMBAT]
            and self.graft_widgets[GRAFT_TYPE.COMBAT]:IsShown()
            and self.graft_widgets[GRAFT_TYPE.COMBAT]:GetDefaultFocus() )
        or self.close_button
end


function GraftCollectionScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function GraftCollectionScreen:OnOpen()
    GraftCollectionScreen._base.OnOpen( self )

    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    self.screen_loop = AUDIO:CreateEventInstance("event:/ui/grafts_menu/open")
    self.screen_loop:Start()

    local anim_duration = 0.7
    local horizontal_movement = 90

    -- Prepare for animation by getting current values
    local negotiation_grafts = self.graft_widgets[GRAFT_TYPE.NEGOTIATION]
    local negotiation_grafts_x, negotiation_grafts_y = negotiation_grafts:GetPos()
    local combat_grafts = self.graft_widgets[GRAFT_TYPE.COMBAT]
    local combat_grafts_x, combat_grafts_y = combat_grafts:GetPos()

    -- Hide objects until they're animated in
    negotiation_grafts:SetTintAlpha(0)
    combat_grafts:SetTintAlpha(0)

    -- Setup and start animation
    self.fadein_anim = self:RunUpdater(
        Updater.Parallel{
            -- Fade and move in the graft container
            Updater.Ease( function(v) negotiation_grafts:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) negotiation_grafts:SetPos(x,y) end, negotiation_grafts_x - horizontal_movement, negotiation_grafts_y, negotiation_grafts_x, negotiation_grafts_y, anim_duration, easing.outQuad),
        
            Updater.Ease( function(v) combat_grafts:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) combat_grafts:SetPos(x,y) end, combat_grafts_x + horizontal_movement, combat_grafts_y, combat_grafts_x, combat_grafts_y, anim_duration, easing.outQuad),
        }
    )
end

function GraftCollectionScreen:OnClose()
    GraftCollectionScreen._base.OnClose( self )

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil
end

function GraftCollectionScreen:Layout()

    self.text_content:SetPos( 0, 0 )
    self.subtitle:LayoutBounds( "left", "below", self.title )
        :Offset( 0, SPACING.TITLE_DESC )
    self.graft_widgets[GRAFT_TYPE.NEGOTIATION]:LayoutBounds( "left", "below", self.text_content ):Offset( 0, -30 )
    self.graft_widgets[GRAFT_TYPE.COMBAT]:LayoutBounds( "after", "top", self.graft_widgets[GRAFT_TYPE.NEGOTIATION] ):Offset( SPACING.M1, 0 )

    -- Get MainOverlay's top bar height
    local overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    local overlay_topbar_height = overlay and overlay:GetTopBarHeight() or 0

    -- Calculate scroll panel size
    local left_offset = RES_X*0.15 -- How much the grid should cross to the left of the screen center
    local left_padding = 100 -- How much space should there be between the left edge of the scroll area and the first graft (allows space for glow/scaling without clipping)
    local left_screen_padding = 15
    -- self.scroll:SetLeftPadding( left_padding )
    self.content:LayoutBounds( "left", "top", 0, 0 )
    self.scroll:SetSize( self.screen_w/2 + left_offset + left_padding, self.screen_h - overlay_topbar_height )
        :SetPos( self.screen_w/4 - left_offset/2 - left_padding/2 - left_screen_padding, -overlay_topbar_height/2 )
        :RefreshView()

    return self
end

------------------------------------------------------------------------------------------
-- Shows grafts

local GraftScreenSlotContainer = class( "Widget.GraftScreenSlotContainer", Widget )

function GraftScreenSlotContainer:init( graft_type, owner, size, locked )
    self._base.init( self )

    self.locked = locked

    self.width = 300
    self.graft_type = graft_type
    self.owner = owner or TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    self.grafts = {}

    self.contents = self:AddChild( Widget() )
    self.type_label = self.contents:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TABS ) )
        :SetText( self.graft_type == GRAFT_TYPE.COMBAT and LOC"UI.GRAFT_COLLECTION.COMBAT_GRAFTS" or LOC"UI.GRAFT_COLLECTION.NEGOTIATION_GRAFTS" )
        :SetGlyphColour( self.graft_type == GRAFT_TYPE.COMBAT and UICOLOURS.FIGHT or UICOLOURS.NEGOTIATION )
        :LeftAlign()
        :Bloom(0.1)
    self.slot_widgets = self.contents:AddChild( Widget() )

    if self.owner then
        self.owner:ListenForAny( self, self.OnAgentEvent )
        self:Refresh()
    end
end

function GraftScreenSlotContainer:SetWidth( width )
    self.width = width or self.width or 300
    self:Layout()
    return self
end

-- Set if we're replacing grafts on the Screen.ReplaceGraftScreen and these are grafts that can be replaced
function GraftScreenSlotContainer:SetReplacingGraft()
    self.replacing_graft = true

    -- Update all the grafts
    for k, w in ipairs( self.slot_widgets.children ) do
        w:SetReplacingGraft()
    end

    return self
end

function GraftScreenSlotContainer:GetGraftCount()
    return self.num_grafts or 0, self.max_grafts or 0
end

function GraftScreenSlotContainer:Refresh()
    self.max_grafts = self.owner:GetAspect("graft_owner"):GetMaxGrafts( self.graft_type )
    local grafts = ObtainWorkTable()
    self.owner:GetAspect("graft_owner"):GetGrafts( self.graft_type, grafts )

    -- Clear old graft widgets
    self.grafts = {}
    self.slot_widgets:DestroyAllChildren()

    -- Add slots
    for i = 1, self.max_grafts do
        local widget = self.slot_widgets:AddChild( Widget.GraftDetails( self.width, self.owner, self.locked ) )
        if self.replacing_graft then widget:SetReplacingGraft() end
        if self.hide_tt then widget:DisableToolTip(true) end
    end

    -- Set grafts to slots
    for k, graft in ipairs( grafts ) do
        if k <= #self.slot_widgets.children then
            local widget = self.slot_widgets.children[k]
            widget:Refresh( graft )
            self.grafts[graft] = widget
        if self.replacing_graft then widget:SetReplacingGraft() end
            if self.hide_tt then widget:DisableToolTip(true) end
        end
    end

    self.num_grafts = #grafts

    self.type_label:SetText( loc.format( self.graft_type == GRAFT_TYPE.COMBAT and LOC"UI.GRAFT_COLLECTION.COMBAT_GRAFTS" or LOC"UI.GRAFT_COLLECTION.NEGOTIATION_GRAFTS", self.num_grafts, self.max_grafts ) )

    ReleaseWorkTable(grafts)
    self:Layout()
    return self
end

function GraftScreenSlotContainer:Layout()

    -- Update the size on the grafts
    for k, w in ipairs( self.slot_widgets.children ) do
        w:SetWidth( self.width )
    end

    -- Layout grafts in a column
    self.slot_widgets:StackChildren( SPACING.M1 )
        :LayoutBounds( "left", "below", self.type_label ):Offset( 0, -SPACING.M1*0.7 )

    return self
end

function GraftScreenSlotContainer:OnRemoved()
    if self.owner then
        self.owner:RemoveListener( self )
    end
end

function GraftScreenSlotContainer:OnAgentEvent( event, ... )
    
    if event == "graft_added" then
        self:Refresh()
        local graft = select(1, ...)
        local widget = self.grafts[graft]
        self:Layout()
        if widget then
            widget:ShowGained()
        end

    elseif event == "graft_removed" then
        self:Refresh()
        self:Layout()
    
    elseif event == "graft_changed" then
        local graft = select(1, ...)
        local widget = self.grafts[graft]
        if widget then
            widget:Refresh(graft)
            self:Layout()
        end
    end
end

function GraftScreenSlotContainer:SetFn(fn)
    for k, widget in ipairs(self.slot_widgets.children) do
        widget:SetOnClickFn( function() fn( widget.graft, widget ) end )
    end
    return self
end

function GraftScreenSlotContainer:GetDefaultFocus()
    return self.slot_widgets and self.slot_widgets.children and #self.slot_widgets.children>0 and self.slot_widgets.children[1]
end

------------------------------------------------------------------------------------------
-- Shows the selected graft's details

local GraftDetails = class( "Widget.GraftDetails", Widget.Clickable )

GraftDetails.CONTROL_MAP = {
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.graft and self.graft:HasUpgrades() and self.graft:UpgradeReady() and not self.locked and not self.replacing_graft
        end,
        fn = function(self)
            self:DoUpgrade()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GRAFT_COLLECTION.CONTROLS_INSTALL", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.graft and self.graft:HasUpgrades() and not self.graft:UpgradeReady() and not self.locked and not self.replacing_graft
        end,
        fn = function(self)
            local preview = true
            self:DoUpgrade( preview )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GRAFT_COLLECTION.CONTROLS_PREVIEW", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.onclickfn and self.replacing_graft
        end,
        fn = function(self)
            self.onclickfn()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.GRAFT_COLLECTION.CONTROLS_REPLACE", Controls.Digital.MENU_ACCEPT))
        end
    }
}

function GraftDetails:init( content_w, owner, locked )
    GraftDetails._base.init( self )
    
    self.locked = locked
    self.icon_size = 80
    self.max_width = content_w or 400 -- Max width, including the icon, if visible
    self.text_width = self.max_width - self.icon_size - 10 - SPACING.M1*2
    self.owner = owner

    self.hitbox = self:AddChild( Widget.SolidBox( 100, 100, 0xffff0030 ) )
        :SetTintAlpha( 0 )

    -- Add background
    self.background = self:AddChild( Widget.Panel( engine.asset.Texture("UI/left_border_frame_faint.tex" ) ) )
        :SetBloom( 0.15 )

    -- Animatable contents
    self.contents = self:AddChild( Widget() )

    -- Add graft icon
    self.graft_icon = self.contents:AddChild( Widget.GraftSlot( nil, self.icon_size ) )
        :DisableToolTip(true)

    -- Add graft details
    self.details = self.contents:AddChild( Widget() )
    self.name = self.details:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_NAME, LOC"UI.GRAFT_COLLECTION.SLOT_EMPTY" ) )
        :SetWordWrap( true )
        :SetAutoSize( self.text_width )
        :LeftAlign()
        :Bloom(0.1)
    self.subtitle = self.details:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetWordWrap( true )
        :SetAutoSize( self.text_width )
        :LeftAlign()
        :Bloom(0.05)
    self.desc = self.details:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ) )
        :SetWordWrap( true )
        :SetAutoSize( self.text_width )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :LeftAlign()
        :Bloom(0.05)
    self.xp_bar = self.details:AddChild( Widget.GraftSlotUpgradeBar( self.text_width ) )
        :Hide()
    self.upgrade_available = self.details:AddChild( Widget.GraftSlotUpgradeAvailable() )
        :Hide()

    self.card_keywords = {}
end

-- Set if we're replacing grafts on the Screen.ReplaceGraftScreen and this is the new one
function GraftDetails:SetNewGraft()
    self.new_graft = true
    self.new_graft_background = self:AddChild( Widget.Panel( engine.asset.Texture("UI/empty_corner_frame_border.tex" ) ) )
        :SendToBack()
    return self
end

-- Set if we're replacing grafts on the Screen.ReplaceGraftScreen and this is a graft that can be replaced
function GraftDetails:SetReplacingGraft()
    self.replacing_graft = true
    return self
end

function GraftDetails:SetWidth( width )
    self.max_width = width or self.max_width or 400
    self.text_width = self.max_width - self.icon_size - 6 - SPACING.M1*2
    self.name:SetAutoSize( self.text_width )
    self.subtitle:SetAutoSize( self.text_width )
    self.desc:SetAutoSize( self.text_width )
    self.xp_bar:SetSize( self.text_width )
    self:Layout()
    return self
end

function GraftDetails:SetOwner( agent )
    self.owner = agent
    return self
end

function GraftDetails:HideUpgradeLabel()
    self.hide_upgrade_label = true
    return self
end

function GraftDetails:PostFeature( feature_id, name, desc )
    self.card_keywords[ feature_id ] = { name = name, desc = desc }
end

function GraftDetails:PostCard( feature_id, card )
    self.card_keywords[ feature_id ] = { card = card }
end

function GraftDetails:Refresh(graft)
    self.graft = graft
    table.clear( self.card_keywords )

    if self.graft then
        -- Update icon
        self.graft_icon:Refresh(self.graft)

        -- Update text
        self.name:SetText( loc.upper( self.graft:GetName() ) )
        if self.graft:GetType() == GRAFT_TYPE.COMBAT then
            self.name:SetGlyphColour( UICOLOURS.FIGHT )
            self.background:SetTintColour( UICOLOURS.FIGHT )
            if self.new_graft_background then self.new_graft_background:SetTintColour( UICOLOURS.FIGHT ):SetTintAlpha( 0.4 ) end
        elseif self.graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            self.name:SetGlyphColour( UICOLOURS.NEGOTIATION )
            self.background:SetTintColour( UICOLOURS.NEGOTIATION )
            if self.new_graft_background then self.new_graft_background:SetTintColour( UICOLOURS.NEGOTIATION ):SetTintAlpha( 0.4 ) end
        end

        local rarity_color = CARD_RARITY_COLOURS[ self.graft:GetRarity() or CARD_RARITY.UNIQUE ]
        local rarity_icon_string = GetCardRarityIconString( self.graft:GetRarity() )
        local subtitle_text = TheGame:Str( string.format( "UI.GRAFT_COLLECTION.%s.%s", self.graft:GetRarity(), self.graft:GetType() ))
        self.subtitle:SetText( rarity_icon_string .. subtitle_text )
        self.subtitle:SetGlyphColour( rarity_color )
        self.subtitle:SetTintAlpha(1)
        if self.desc then self.desc:SetText( self.graft:GetDesc(self) ) end

        -- XP
        if self.graft:HasUpgrades() and self.xp_bar and self.upgrade_available then
            local current_xp, max_xp = self.graft:GetXP()
            self.xp_bar:Refresh( math.floor(current_xp), max_xp )
                :Show()
            self.upgrade_available:SetShown( self.graft:UpgradeReady() and not self.hide_upgrade_label )
            if self.graft:UpgradeReady() and not self.locked then
                self.name:SetGlyphColour( UICOLOURS.MONEY )
                    :Pulse( 0.8, 1.0, 0.003 )
                self.background:SetTintColour( UICOLOURS.MONEY )
                    :Pulse( 0.8, 1.0, 0.003 )
            end
        else
            if self.xp_bar then self.xp_bar:Hide() end
            if self.upgrade_available then self.upgrade_available:Hide() end
        end

        -- Go through all the keywords for this graft
        self:SetToolTipClass( Widget.TooltipKeywords )
        self:SetToolTip( self.card_keywords )
    else
        self:SetOnClickFn( nil )
    end

    self:Layout()
    return self
end

function GraftDetails:Layout()

    self.subtitle:LayoutBounds( "left", "below", self.name ):Offset( 0, 4 )
    if self.xp_bar and self.xp_bar:IsShown() then self.xp_bar:LayoutBounds( "left", "below", self.subtitle ):Offset( 0, -10 ) end
    if self.upgrade_available and self.upgrade_available:IsShown() then self.upgrade_available:LayoutBounds( "left", "below", self.xp_bar ):Offset( 0, 0 ) end
    self.desc:LayoutBounds( "left", "below", self.upgrade_available:IsShown() and self.upgrade_available or ( self.xp_bar:IsShown() and self.xp_bar or self.subtitle ) ):Offset( 0, -10 )
    
    local details_w, details_h = self.details:GetSize()
    details_h = math.max( self.icon_size, details_h )
    details_h = details_h + SPACING.M1
    self.hitbox:SetSize( self.max_width, details_h )
    self.background:SetSize( self.max_width, details_h )
    if self.new_graft_background then self.new_graft_background:SetSize( self.max_width, details_h ) end
    self.graft_icon:LayoutBounds( "left", "top", self.hitbox )
        :Offset( SPACING.M1/2, -SPACING.M1*0.4 )
    self.details:LayoutBounds( "after", "top", self.graft_icon )
        :Offset( 6, 0 )

    if not self.graft then
        -- Empty graft slot
        self.graft_icon:LayoutBounds( "left", "center", self.hitbox )
            :Offset( SPACING.M1/2, 0 )
        self.name:LayoutBounds( "after", "center", self.graft_icon )
            :Offset( 6, 2 )
    end

    return self
end

function GraftDetails:ShowGained( fn )
    
    local anim_duration = 0.2
    local icon_glow_1 = self:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), self.icon_size*1.5, self.icon_size*1.5 ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( self.graft:GetType() == GRAFT_TYPE.COMBAT and UICOLOURS.FIGHT or UICOLOURS.NEGOTIATION )
        :SendToBack()
        :Bloom(0.15)
        :LayoutBounds( "center", "center", self.graft_icon )
    local icon_glow_2 = self:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), self.icon_size, self.icon_size ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( self.graft:GetType() == GRAFT_TYPE.COMBAT and UICOLOURS.FIGHT or UICOLOURS.NEGOTIATION )
        :SendToBack()
        :Bloom(0.01)
        :LayoutBounds( "center", "center", self.graft_icon )

    self:RunUpdater(
        Updater.Series{
            Updater.Wait( anim_duration ),
            Updater.Parallel{
                Updater.Series{
                    Updater.Ease( function(v) icon_glow_1:SetTintAlpha(v) end, 0, 0.4, anim_duration, easing.outQuad),
                    Updater.Ease( function(v) icon_glow_1:SetTintAlpha(v) end, 0.4, 0, anim_duration, easing.outQuad),
                },
                Updater.Ease( function(v) icon_glow_1:SetScale(v,v) end, 1, 1.5, anim_duration*2, easing.outQuad),
                Updater.Series{
                    Updater.Ease( function(v) icon_glow_2:SetTintAlpha(v) end, 0, 0.3, anim_duration*1.5, easing.outQuad),
                    Updater.Ease( function(v) icon_glow_2:SetTintAlpha(v) end, 0.3, 0, anim_duration*1.5, easing.outQuad),
                },
                Updater.Ease( function(v) icon_glow_2:SetScale(v,v) end, 1, 3, anim_duration*3, easing.outQuad),
            },
            Updater.Do( function()
                icon_glow_1:Remove()
                icon_glow_2:Remove()
                if fn then fn() end
            end )
        }
    )

    return self
end

function GraftDetails:DoUpgrade( preview )
    if not self.upgrade_coro and self.owner then
        self.upgrade_coro = self:StartCoroutine( self.DoUpgradeCoro, self, preview )
    end
end

function GraftDetails:DoUpgradeCoro( preview )
    local popup = Screen.UpgradeGraftPopup( self.graft, self.owner, preview )
    TheGame:FE():PushScreen( popup )

    while TheGame:FE():FindScreen( popup ) do
        coroutine.yield()
    end

    if popup.chosen_graft and self.graft:UpgradeReady() then
        self.graft:Upgrade()
    end
    self.upgrade_coro = nil
end

function GraftDetails:HandleControlDown(control, device)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() and self.graft then
            DBG( self.graft )
            return true
        end
    end

    return GraftDetails._base.HandleControlDown( self, control, device )
end

function GraftDetails:UpdateImage()
    if ( self.enabled and self.onclickfn and self.hover ) or self.focus then
        self.graft_icon:ScaleTo( nil, 1.02, 0.2, easing.outQuad )
        self.background:SetTexture( engine.asset.Texture( "UI/left_border_frame_faint_border.tex" ) )
            :MoveTo( 6, 0, 0.2, easing.outQuad )
        if self.new_graft_background then self.new_graft_background:MoveTo( 6, 0, 0.2, easing.outQuad ) end
        self.contents:MoveTo( 12, 0, 0.2, easing.outQuad )
        if self.name then self.name:Bloom(0.15) end
    else
        self.graft_icon:ScaleTo( nil, 1, 0.2, easing.outQuad )
        self.background:SetTexture( engine.asset.Texture( "UI/left_border_frame_faint.tex" ) )
            :MoveTo( 0, 0, 0.2, easing.outQuad )
        if self.new_graft_background then self.new_graft_background:MoveTo( 0, 0, 0.2, easing.outQuad ) end
        self.contents:MoveTo( 0, 0, 0.2, easing.outQuad )
        if self.name then self.name:Bloom(0.1) end
    end
end

function GraftDetails:OnGainFocus()
    GraftDetails._base.OnGainFocus(self)
    self:UpdateImage()
end

function GraftDetails:OnLoseFocus()
    GraftDetails._base.OnLoseFocus(self)
    self:UpdateImage()
end

function GraftDetails:OnGainHover()
    GraftDetails._base.OnGainHover(self)
    self:UpdateImage()
end

function GraftDetails:OnLoseHover()
    GraftDetails._base.OnLoseHover(self)
    self:UpdateImage()
end


------------------------------------------------------------------------------------------
-- Shows upgrade status

local GraftSlotUpgradeBar = class( "Widget.GraftSlotUpgradeBar", Widget.SegmentedBar )

function GraftSlotUpgradeBar:init( width )
    GraftSlotUpgradeBar._base.init( self, width, 8 )

    self.width = width
    self.current_xp = 0
    self.max_xp = 0

    self:SetOnTexture( engine.asset.Texture("UI/cardupgrade_bar.tex"), 6, 6, 12, 12 )
    self:SetOffTexture( engine.asset.Texture("UI/cardupgrade_bar.tex"), 6, 6, 12, 12 )
    self:SetSpaceBetweenPips( 4 )

    self:Bloom( 0.3 )
end

function GraftSlotUpgradeBar:SetSize( dw )
    GraftSlotUpgradeBar._base.SetSize( self, dw, 8 )
    return self
end

function GraftSlotUpgradeBar:Refresh( current_xp, max_xp )

    self:SetProgress( max_xp, current_xp )

    if current_xp >= max_xp then
        -- Use bright colors if this graft is ready to upgrade
        local r, g, b, a = HexColour( UICOLOURS.QUEST )
        self:SetOnTint( r, g, b, a )
        self:SetOffTint( r, g, b, a )
    else
        local r, g, b, a = HexColour( UICOLOURS.OBJECTIVE )
        self:SetOnTint( r, g, b, a )
        self:SetOffTint( r, g, b, 0.1 )
    end

    self.current_xp = current_xp
    return self
end

------------------------------------------------------------------------------------------
-- Shows that an upgrade is available

local GraftSlotUpgradeAvailable = class( "Widget.GraftSlotUpgradeAvailable", Widget )

local icon_size = 32
local text_size = 28

function GraftSlotUpgradeAvailable:init()
    GraftSlotUpgradeAvailable._base.init( self )

    self.icon = self:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_upgrade.tex" ) ) )
        :SetSize( icon_size, icon_size )
        :SetTintColour( UICOLOURS.MONEY )
        :Bloom( 0.15 )
    self.label = self:AddChild( Widget.Label( "title", text_size ) )
        :SetText( LOC"UI.GRAFT_COLLECTION.SLOT_UPGRADE_AVAILABLE" )
        :SetTintColour( UICOLOURS.MONEY )
        :Bloom( 0.15 )

    self.label:LayoutBounds( "after", "center", self.icon ):Offset( 8, 0 )

    self.icon:RunUpdater(
        Updater.Loop{
            Updater.Ease( function(y) self.icon:SetPos(nil, y) end, 3, 0, 0.7, easing.inOutQuad),
            Updater.Ease( function(y) self.icon:SetPos(nil, y) end, 0, 3, 0.2, easing.inOutQuad),
        }
    )
end

------------------------------------------------------------------------------------------
-- Shows progress radial

local GraftUpgradeProgressRadial = class( "Widget.GraftUpgradeProgressRadial", Widget )

function GraftUpgradeProgressRadial:init( size )
    GraftUpgradeProgressRadial._base.init( self )

    self.size = size
    self.current_xp = 0
    self.max_xp = 0

    self.circle_bg = self:AddChild( Widget.Image( engine.asset.Texture( "UI/graft_radial_progress.tex") ) )
        :SetTintColour( 0x000000aa )
        :SetSize( self.size, self.size )
        :Bloom( 0.1 )
    self.circle_progress = self:AddChild( Widget.RadialProgress( self.size, self.size, engine.asset.Texture( "UI/graft_radial_progress.tex") ) )
        :Bloom( 0.2 )
    self.text_container = self:AddChild( Widget() )
    self.label_value = self.text_container:AddChild( Widget.Label() )
        :SetFont( "title" )
        :SetFontSize( self.size*0.65 )
        :Bloom( 0.2 )
    self.label_max = self.text_container:AddChild( Widget.Label() )
        :SetFont( "title" )
        :SetFontSize( self.size*0.4 )
        :Bloom( 0.2 )
end

function GraftUpgradeProgressRadial:SetColour( colour )
    self.circle_progress:SetTintColour( colour )
    self.label_value:SetGlyphColour( colour )
    self.label_max:SetGlyphColour( colour )
end

function GraftUpgradeProgressRadial:Refresh( current_xp, max_xp )
    self.label_value:SetText( loc.format( "{1}", current_xp ) )
    self.label_max:SetText( loc.format( "/{1}", max_xp ) )
    self.circle_progress:SetProgress( current_xp and current_xp > 0 and max_xp and max_xp > 0 and current_xp/max_xp or 0.005 )

    self.label_max:LayoutBounds( "after", "bottom", self.label_value ):Offset( 0, 4 )
    self.text_container:LayoutBounds( "center", "center", self.circle_bg ):Offset( 0, 3 )
end

