require "ui/widgets/screen"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/panelbutton"
require "ui/widgets/plax_nameplate"
require "ui/widgets/imagebutton"
require "ui/widgets/locationbutton"
require "ui/widgets/brawlprogress"
require "ui/widgets/plax"
require "encounter/encounter"

---------------------------------------------------------------------
local BOTTOM_MARGIN = 65
local CONTROL_BAR_OFFSET = 50


local LocationScreen = class("Screen.LocationScreen", Widget.Screen)


LocationScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.OPEN_MAP, 
        fn = function(self)
            local close_opt = self.multi:GetCloseOption()
            if close_opt then close_opt:Click() end
            return true
        end,
        test = function(self)
            local close_opt = self.multi:GetCloseOption()
            return close_opt and close_opt:IsShown()
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.TRAVEL", Controls.Digital.OPEN_MAP))
        end
    },
    {
        controls = { Controls.Digital.MENU_CANCEL, Controls.Digital.MENU_UP, Controls.Digital.MENU_RIGHT, Controls.Digital.MENU_DOWN, Controls.Digital.MENU_LEFT },
        test = function(self)
            return self.show_reticule == true and self.multi:GetConvoOption(1) ~= nil
        end,
        fn = function(self)
            self:SetReticuleMode( false )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT_RELEASE", Controls.Digital.MENU_CANCEL ) )
        end,
    },
    {
        hint = function(self, left, right)
            if self.show_reticule then
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_MOVE_INSPECT" )
            else
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT" )
            end
        end,
    },
    {
        test = function(self) 
            return self.show_reticule == true and not TheGame:GetLanguageInfo().hide_cursor_hint
        end,
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.CURSOR_SPEEDUP")
        end
    },
    {
        test = function(self) 
            return self.show_reticule == true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.OPEN_OVERLAY_MENU", Controls.Digital.OPEN_OVERLAY_MENU ))
        end
    },
    {
        control = Controls.Digital.PAUSE, 
        fn = function(self)
            self:GetFE():InsertScreen(Screen.PauseMenu())
            return true
        end,
    },


}


function LocationScreen:init(location)
    Widget.Screen.init(self)

    self.start_location = location
    self.hover_check = true

    self.multi = Widget.MultiChoice():SetOnSelectOptionFn(function(idx, ...) self.on_select( idx ) end)

    self.plax = Widget.Plax( PLAX_FLAGS.NONE ):SetName("LOCATION_" .. (location and location.content_id or "")):TrackMouse( true ):PostProcess( true )

    self.interface = Widget():SetName("interface"):AddChildren{
        self.multi,
    }:SetAnchors("center", "bottom")

    self:AddChildren
    {
        self.plax,
        self.interface,
    }

    self.main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )

    self:SetAnchors( "center", "center" )
end

function LocationScreen:GetMusicEvent()
    return self.plax:GetMusicEvent()
end

function LocationScreen:OnGainTop()
    LocationScreen._base.OnGainTop(self)


    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    collectgarbage()
    
    self.interface:SetShown( TheGame:FE():IsHUDVisible() )
    
    if self.location ~= TheGame:GetGameState():GetPlayerAgent():GetLocation() then
        self:PushLocation(TheGame:GetGameState():GetPlayerAgent():GetLocation())
    end

    if self.brawl_progress then
        self.brawl_progress:Refresh()
    end

    if self.needs_refresh then
        self:Refresh()
    end
    self.plax:PlaySounds()
	self.should_evalulate_confronts = true

    self:SetReticuleMode( true )
end

function LocationScreen:OnLoseTop()
    LocationScreen._base.OnLoseTop(self)

    if self.main_overlay then
        self.main_overlay:FadeInTopLeft( 0.1 ) 
        self.main_overlay:FadeInTopRight( 0.1 )
    end

    self.interface:Hide()
end

function LocationScreen:OnOpen()
    LocationScreen._base.OnOpen(self)
    TheGame:GetEvents():ListenForEvents( self, "phase_change", "encounter_changed", "quests_changed" )
    TheGame:GetGameState():GetPlayerAgent():ListenForAny( self, self.OnPlayerEvent )

    if TheGame:GetGameState():GetGameType() == GAME_TYPE.BRAWL then self.brawl_progress = self.interface:AddChild(Widget.BrawlProgress()) end

    self:PushLocation( self.start_location )
    self.should_evalulate_confronts = true
end


function LocationScreen:PositionMulti()
    local offset = BOTTOM_MARGIN
    if TheGame:FE():IsGamepadMode() then
        offset = offset + CONTROL_BAR_OFFSET
    end
    if self.brawl_progress then
        self.brawl_progress:LayoutBounds( "center", "bottom", 0, 0 )
            :Offset(0, offset)
        
        self.multi:LayoutBounds( "center", "above", self.brawl_progress )
            :Offset( 0, 20 )
    else
        self.multi:LayoutBounds( "center", "bottom", 0, 0 )
            :Offset(0, offset)
    end
end

function LocationScreen:OnControlModeChange( cm, device_id )
    LocationScreen._base.OnControlModeChange( self, cm, device_id )
    
    if cm == CONTROL_MODE.GAMEPAD then
        self.multi:HideCloseOption()
    elseif cm == CONTROL_MODE.TOUCH then
        self.multi:ShowCloseOption()
        self.multi:HideControllerHint()
    else
        self.multi:ShowCloseOption()
        self.multi:HideControllerHint()
    end

    self:PositionMulti()
end

function LocationScreen:OnScreenModeChange( sm )
    LocationScreen._base.OnScreenModeChange( self, sm )

    self.multi:SetLayoutScale( LAYOUT_SCALE[sm] )
    if self.brawl_progress then self.brawl_progress:SetLayoutScale( LAYOUT_SCALE[sm] ) end
    self:PositionMulti()

end

function LocationScreen:HandleControlUp( control, device )
    if self.multi:HandleControlUp( control, device ) then
        return true
    end

    if control:Has(Controls.Digital.MENU_CANCEL) then
        
        local close_opt = self.multi:GetCloseOption()
        if close_opt then
            close_opt:Click()
            return true
        end
    end

    return LocationScreen._base.HandleControlUp( self, control, device )
end


function LocationScreen:HandleControlDown(control, device)

    -- If we're in touch mode, tapping the map will toggle the quests overlay on and off
    -- When we toggle the quests off, we show the agent names instead
    if TheGame:FE():GetControlMode() == CONTROL_MODE.TOUCH then
        self.main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
        if not self.hiding_todo_list then
            self.hiding_todo_list = true
            -- Hide overlays
            if self.main_overlay then
                self.main_overlay:FadeOutTopLeft( 0.3 ) 
                self.main_overlay:FadeOutTopRight( 0.3 )
            end
            -- Show agent names
            if self.plax and self.plax:Slots() then
                for k, v in self.plax:Slots() do
                    if v.occupant and v.label then
                        v.label:ShowName( true )
                    end
                end
            end
        else
            self.hiding_todo_list = false
            -- Show overlays
            if self.main_overlay then
                self.main_overlay:FadeInTopLeft( 0.1 ) 
                self.main_overlay:FadeInTopRight( 0.1 )
            end
            -- Hide agent names
            if self.plax and self.plax:Slots() then
                for k, v in self.plax:Slots() do
                    if v.occupant and v.label then
                        v.label:ShowName( false )
                    end
                end
            end
        end
    end

    if self.multi:HandleControlDown( control, device ) then
        return true
    end

    return LocationScreen._base.HandleControlDown( self, control, device )
end

--[[function LocationScreen:GetDefaultFocus()
    if self.plax and self.plax:Slots() then
        for k, v in self.plax:Slots() do
            if v.occupant and v.label then
                return v.label
            end
        end
    end
end
--]]

function LocationScreen:ToggleHUD( visible, hud_flags )
    self.interface:SetShown( visible and self:IsOnTop())
end

function LocationScreen:OnClose()
    
    TheGame:GetEvents():RemoveListener( self )
    TheGame:GetGameState():GetPlayerAgent():RemoveListener( self )

    if self.location then
        self:ClearLocation()
    end
end

function LocationScreen:OnGlobalEvent( event_name, ... )
    if event_name == "encounter_changed" then
        local encounter = ...
        self:EnterEncounter( encounter )
    elseif self.location then
        if self:IsShowing() then
            self:Refresh()
        else
            self.needs_refresh = true
        end
    end
end


function LocationScreen:IsShowing()
    if TheGame:FE():GetTopScreen() == self then
        return true
    end

    --this is probably correct?
    local convo_screen = TheGame:FE():FindScreen( Screen.ConversationScreen )
    if convo_screen and TheGame:FE():GetTopScreen() == convo_screen then
        return true
    end

    return false
end

function LocationScreen:OnPlayerEvent( event_name, ... )
    if event_name == "agent_location_changed" then
       
        if self:IsShowing() then
            local player, old_location, new_location = ...
            self:PushLocation( new_location )
        end
        self.needs_refresh = true
        self.should_evalulate_confronts = true
    end
end

function LocationScreen:OnLocationEvent( event_name, ... )
    if event_name == "plax_changed" then
        local location, element_id = select(1, ...)
        if location == self.location then
            if element_id then
                self.plax:RefreshVisibility( element_id )
            else
                self:Refresh()
                self.plax:PlaySounds()
                TheGame:FE():RefreshMusic()
            end
        end
    elseif event_name == "agents_changed" or event_name == "agent_role_changed" then
        -- FIXME: A little overkill, because multiple agents_changed can happen in a frame.
        if self:IsShowing() then
            self:Refresh( true )
        else
            self.needs_refresh = true
        end
    end
end

function LocationScreen:GetDebugEnv( debug_env )
    debug_env.location = self.location
    debug_env.plax = self.plax
end

function LocationScreen:GetDefaultFocus()
    return self.multi:GetConvoOption(1)
end

function LocationScreen:AddAgentToSlot( agent )
    
    local screen = self:GetFE():FindScreen( Screen.ConversationScreen )
    if screen and (screen:IsCharacterLeft( agent ) or screen:IsCharacterRight( agent )) then
        -- On the convo stage, so they're not on our plax.
        return
    end

    local character_widget
    local slot = self.plax:GetAgentSlot( agent )
    if slot == nil then
        local slot_id = agent:GetPlaxSlotID( self.plaxname )
        local role = agent:GetRoleAtLocation(self.location)
        
        if slot_id and not self.plax:GetSlotOccupant( slot_id ) and not self.plax:ShouldMoveAgentToMoreSpecificSlot(agent, role, slot_id) then
            character_widget, slot = self.plax:FillSlot(agent, slot_id)
        else
            character_widget, slot = self.plax:FillSlot(agent, role, true)
        end
    else
        character_widget = slot.character
    end

    if slot and character_widget then
        agent:SetPlaxSlotID( slot.slot_id )

        local nameplate = self.plax:GetSlotLabel( slot.slot_id )
        
        if nameplate == nil then
            nameplate = Widget.PlaxNameplate():SetName( "NAMEPLATE" )
            self.plax:SetSlotLabel( slot.slot_id, nameplate )
        end
        nameplate:RefreshNameplate( agent, math.abs( character_widget.parent:GetScale() ) )
        if not agent:IsPlayer() and agent:CanTalk() then
            nameplate:SetOnClickFn( function() self:EnterConvo( agent ) end )
        end 

        self.plax:UpdateAllHSV()    -- Force updating the HSV for the whole plax. Shotgun approach to try to fix the bug where one character has the wrong HSV setting (day instead of night)

        character_widget:SetHoverOutlineColour( UICOLOURS.SUBTITLE )
        return character_widget
    end
    
    print ("WARNING - could not find plax slot for agent", agent)
end

function LocationScreen:UpdateDayPhase()
    self.locked_time_of_day = TheGame:GetGameState():GetDayPhase()
    self.plax:SetDayPhase( self.locked_time_of_day )
end

function LocationScreen:Refresh( no_populate, force )
    --print ("REFRESH", debugstack())
    if not no_populate then
        TheGame:GetGameState():PopulateAgentsInWorld()
    end

    self.needs_refresh = false
    
    self:ProcessHoverAgent()

    local plaxname = self.location:RefreshPlax( self.plax )
    if plaxname then
        TheGame:GetGameProfile():SeePlax(plaxname)
        self.locked_time_of_day = self.locked_time_of_day or TheGame:GetGameState():GetDayPhase()
        self.plax:SetDayPhase( self.locked_time_of_day )

        self:RefreshPlaxClickables(force)
    else
        self.plax:Clear()
        collectgarbage()
    end

    if self.plax then
        self.plax:UpdateAllHSV()    -- Force updating the HSV for the whole plax. Shotgun approach to try to fix the bug where one character has the wrong HSV setting (day instead of night)
    end
end


local function GetAgentSortPriority(a)
    if a.quest_membership and next(a.quest_membership) then
        return 3
    end

    if a:GetBrain():IsOnDuty() then
        return 2.5
    end
    
    if a:GetRoleAtLocation() then
        return 2
    end

    if a:IsSentient() then
        return 1
    end

    return 0.5
end


local function agent_sort_fn(a,b)
    return GetAgentSortPriority(b) < GetAgentSortPriority(a)
end

function LocationScreen:RefreshPlaxClickables( force )

    local convo = TheGame:FE():FindScreen( Screen.ConversationScreen )
    local content = self.location:GetContent()
    if not force and convo and not content.show_agents then
        return
    end

    -- Remove invalid plax slots.
    for slot_id, slot in self.plax:Slots() do
        if slot.agent then
            local agent_role = slot.agent:GetRoleAtLocation()
            if slot.agent:GetLocation() ~= self.location or slot.role ~= agent_role then
                self.plax:ClearSlot( slot_id )
            end
        end
    end
    
    local can_show_new_agents = content.show_agents or (convo == nil)
    if can_show_new_agents then
        local sorted_agents = {}

        for i, agent in self.location:Agents( ) do
            local should_show = true
            if agent:IsInPlayerParty() and not self.location:GetContent().show_player then
                should_show = false
            end
            if not agent:IsInPlayerParty() and not agent:CanTalk() then
                should_show = false
            end

            if should_show then
                table.insert(sorted_agents, agent)
            end
        end
        
        table.sort(sorted_agents, agent_sort_fn)

        for _,agent in ipairs(sorted_agents) do
            local character_widget = self:AddAgentToSlot( agent )
            if character_widget and not agent:IsPlayer() and agent:CanTalk() then
                local fn = function() self:EnterConvo( agent ) end
                character_widget:SetClickFn( fn )
            	local slot = self.plax:GetAgentSlot( agent )
            	local nameplate = slot and self.plax:GetSlotLabel( slot and slot.slot_id )
            	if nameplate then
                        character_widget:SetOnHoverFn( function(hovered) 
                            if not nameplate.removed then
                                nameplate:SetCharacterFocused( hovered or self.current_hover_agent == agent ) 
                            end
                        end)
            	end
			end
        end
    end
end

function LocationScreen:HarvestRobotOptions()
    self:RefreshLocationActions()

    for k, slot in self.plax:Slots() do
        if slot.character and slot.occupant and not slot.occupant:IsPlayer() then
            self:AddRobotOption(slot.character)
        end
    end

    for i, opt in ipairs( self.multi:GetOptions() ) do
        if opt:IsEnabled() then
            self:AddRobotOption(opt)
        end
    end
end

function LocationScreen:ProcessHoverAgent()

    -- Find out what character should be highlighted
    local hover_agent = nil
    -- Go through every agent and check if they're hovered
    for k, v in self.plax:Slots() do
        if v.occupant and v.label then
            if v.character.anim.hover then
                hover_agent = v.occupant
            end
        end
    end

    -- Check if this agent is different than the previously hovered/focused one
    if hover_agent ~= self.current_hover_agent then
        
        -- Stop previously playing sound
        if self.looping_hover then
            if self.current_widget then
                if self.current_widget.removed then
                    DBG(self.current_widget)
                else
                    AUDIO:PlayParamEvent(SoundEvents.npc_mouseover_off, "position", self.current_widget:GetPanPos())
                end
            end
            self.looping_hover:Stop()
        end


        -- Clear previous highlight
        if self.current_hover_agent then
            local slot = self.plax:GetAgentSlot( self.current_hover_agent )
            if slot then
                slot.character:SetSelected( false )
                slot.character:HideOutline()
            end
            local nameplate = slot and self.plax:GetSlotLabel( slot and slot.slot_id )
            if nameplate then
                nameplate:Show()
                nameplate:SetCharacterFocused( false ) 
            end
        end


        -- If there's a new focused/hovered agent, display it accordingly
        local agent = hover_agent
        if agent then
            local slot = self.plax:GetAgentSlot( agent )
            if slot then
                slot.character:SetSelected( true )
                slot.character:ShowOutline()
            end
            local nameplate = slot and self.plax:GetSlotLabel( slot and slot.slot_id )
            if nameplate then
                nameplate:SetCharacterFocused( true )
                nameplate.parent:SendToFront()

                AUDIO:PlayParamEvent(SoundEvents.npc_mouseover_on, "position", nameplate:GetPanPos())
                self.looping_hover = AUDIO:CreateEventInstance(SoundEvents.npc_mouseover_loop)
                if self.looping_hover then
                    self.looping_hover:SetPos(nameplate:GetPanPos())
                    self.looping_hover:Start()
                end
                self.current_widget = nameplate
            end 
        else
            self.current_widget = nil        
        end

        -- Set this as the current hovered/focused agent
        self.current_hover_agent = hover_agent
    end
end

function LocationScreen:EnterConvo( agent )
    local encounter = Encounter( agent, self.location )
    self.current_widget = nil
    self.selected_agent = nil
    self:ProcessHoverAgent()
    TheGame:GetGameState():GetCaravan():SetCurrentEncounter( encounter )
end

function LocationScreen:DoSpecificConvo(convo_id, state_id, opt_agent, opt_quest, scratch)
    local encounter = Encounter( opt_agent, self.location)
    
    if scratch then
        for k,v in pairs(scratch) do
            encounter.scratch[k] = v
        end
    end

    local convo_def = Content.GetConvoStateGraph(convo_id)
    if convo_def then
        state_id = state_id or "START"
        
        local hub = ConvoStateGraph(encounter, convo_def, state_id, opt_quest)
        if not hub:TryValidateHub( opt_agent, state_id ) then
            LOGWARN( "Failed validation for convo %s.%s", convo_id, state_id )
        else
            encounter:SetInitialHub( hub )
            TheGame:GetGameState():GetCaravan():SetCurrentEncounter( encounter )
        end
    else
        LOGWARN( "MISSING_CONVO: %s", convo_id )
    end
end


function LocationScreen:EnterEncounter( encounter, on_finish )
    local function OnDone( screen )
        print ("ENCOUNTER DONE:", self.location, encounter:GetAgent() )
        self.needs_refresh = true
        if on_finish then
            on_finish( screen )
        else
            TheGame:FE():PopScreen( screen )
        end
        
        self.last_agent_talked_to = encounter:GetAgent()

        if not self.removed then
            local player = TheGame:GetGameState():GetPlayerAgent()
            if player:GetLocation() == self.location then
                self:Refresh()
            else
                if player:GetLocation() then
                    self:PushLocation( player:GetLocation() )
                end
            end
        end
        self.should_evalulate_confronts = true
    end

    local player = TheGame:GetGameState():GetPlayerAgent()
    if player:GetLocation() == self.location then
        self:Refresh()
    else
        self:PushLocation( player:GetLocation() )
    end

    local screen = self:GetFE():FindScreen( Screen.ConversationScreen )
    if screen == nil then
        screen = Screen.ConversationScreen( self.plax )
        TheGame:FE():InsertScreen( screen )
    end

    self:ClearOptions()
    screen:StartConvo( encounter, OnDone )

end

function LocationScreen:ClearLocation()
    self.location:RemoveListener( self )
end

function LocationScreen:PushLocation( location )
    if location == self.location then
        return
    end

    engine.inst:ProfilerPush( "PushLocation" )
    if self.location then
        self:ClearLocation()
    end

    if location == nil then
        LOGWARN( "Missing location!!" )
    else
        local content = location:GetContent()

        self.location = location

        location:ListenForAny( self, self.OnLocationEvent )

        self:Refresh()
        self.plax:PlayEnterSound()
        self.plax:PlaySounds()
        TheGame:FE():RefreshMusic()
        
        -- FEH.
        -- Ideally this would be in Agent.MoveToLocationInternal, except if it's flagged
        -- there, encounters are initiated but aborted if the location triggers TravelToArea.
        TheGame:GetGameState():GetCaravan():RefreshConfronts()
    end
    engine.inst:ProfilerPop()
end

function LocationScreen:OnSelectLocationAction( idx, ... )
    local opt = self.current_options[ idx ]
    opt:RunOption()
end

function LocationScreen:ClearOptions()
    self.multi:ClearOptions()
    self.fe:RefreshControlOptions()
end

function LocationScreen:PresentOptions( options, options_type, on_select )
    
    for i, opt in ipairs( options ) do
        local opt_widget = opt:CreateWidget( self.encounter, self, self.multi )
    end
    self.multi:LayoutOptions()
    self:PositionMulti()
    self.fe:RefreshControlOptions()
    self.on_select = on_select
end


function LocationScreen:RefreshLocationActions()
    self:ClearOptions()

    if self.location then
       
        local encounter = Encounter( nil, self.location )
        self.encounter = encounter

        local hub = DefaultHub( encounter, true, HUB_STATE.LOCATION )
        hub.Pop = function()
            
            self.plax:PlayExitSound()
            self:GetFE():FadeTransition( 
                function()
                    local exit = self.location:GetDefaultExit()
                    if exit and exit:GetTo() then
                        TheGame:GetGameState():GetCaravan():MoveToLocation(exit:GetTo())
                    else
                        TheGame:FE():InsertScreen( Screen.TravelScreen() )
                    end
                end,.25)        
        end
        encounter:SetInitialHub( hub )
        encounter:StartConvo( self )
    end
end

function LocationScreen:DoRender()
    self.plax:RenderToScene()
end


function LocationScreen:SetReticuleMode( show_reticule )
    if show_reticule ~= self.show_reticule then
        self.show_reticule = show_reticule

        if not self.show_reticule then
            self:SetDefaultFocus()
        end

        self.fe:RefreshControlOptions()
    end
end

function LocationScreen:WantsReticuleMode()
    return self.show_reticule
end

function LocationScreen:GetReticulePosition()
    if self.plax and self.plax:Slots() then
        local candidates = {}
        for k, v in self.plax:Slots() do
            if v.occupant and v.label then
                table.insert(candidates, v)
            end
        end
        if #candidates > 0 then
            table.sort(candidates, function(a,b)
                local score_a = a.occupant:HasActiveQuestMembership() and 1 or 0
                local score_b = b.occupant:HasActiveQuestMembership() and 1 or 0
                
                if self.last_agent_talked_to then
                    if a.occupant == self.last_agent_talked_to then
                        score_a = score_a + 10
                    end

                    if b.occupant == self.last_agent_talked_to then
                        score_b = score_b + 10
                    end
                end

                return score_a < score_b
             end)
            return candidates[#candidates].label:TransformToWorld(0,0)
        end
    end
    return 0, 0
end


function LocationScreen:TestReticule(x, y)

    if self.plax and self.plax:Slots() then
        for k, v in self.plax:Slots() do
            if v.occupant and v.label then
                if v.character.anim:CheckHit(x,y) then
                    return v.label
                end
            end
        end
        for k, v in self.plax:Slots() do
            if v.occupant and v.label then
                local scale = math.abs(v.character.parent:GetScale())
                local wx, wy = v.label:TransformToWorld(0,0)
                local distsq = (wx-x)*(wx-x) + (wy-y)*(wy-y)
                local dist = distsq == 0 and 0 or math.sqrt(distsq)
                

                if dist < (300*scale+32) then
                    return v.label
                end
            end
        end
    end

    local main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if main_overlay then
        local mouse_blocked, hover_widget = main_overlay:CheckMouseHover(x, y)
        return hover_widget
    end
end



function LocationScreen:OnUpdate( dt )

    LocationScreen._base.OnUpdate( self, dt )

    self:ProcessHoverAgent()

    if self.needs_refresh and self:IsShowing() then
        self:Refresh()
    end

    if self.should_evalulate_confronts and self:IsOnTop() then
        local did_confront = TheGame:GetGameState():GetCaravan():CheckConfronts()
        self:RefreshLocationActions()
        if did_confront then
            self:Refresh(nil, true)
        end


        self.should_evalulate_confronts = false
    end
        
        
    if self.encounter then
        self.encounter:OnUpdate( dt )
    end

    if TheGame:FE():IsGamepadMode() then
        local _,_,rx,ry = TheGame:GetInput():GetActiveGamepadStickVals()
        if math.length2d( rx, ry ) > 0.5 then
            self:SetReticuleMode( true )
        end
    end
end

function LocationScreen:ApplyColourCorrectionValues(v)
    self.plax:ApplyColourCorrectionValues(v)
end

function LocationScreen:GetCurrentColourName()
    return self.plax:GetCurrentColourName()
end

function LocationScreen:OnRemoved()
    self.plax = nil
end

------------------------------------------------------------------------------------------
-- Shows the current act and location, along with mainline quest progress

local LocationScreenHeader = class( "Widget.LocationScreenHeader", Widget )

function LocationScreenHeader:init( )
    LocationScreenHeader._base.init( self )

    self.hitbox = self:AddChild( Widget.SolidBox( 200, 100, 0x00000000 ) )
    self.title = self:AddChild( Widget.SpacedLabel( "body", 20, nil, 8 ) )
    self.title:SetGlyphColour( UICOLOURS.PROGRESS_BAR_OUTLINE )
    self.subtitle = self:AddChild( Widget.SpacedLabel( "body", 24, nil, 25 ) )
    self.subtitle:SetGlyphColour( UICOLOURS.PROGRESS_BAR_OUTLINE )

    -- Get mainline quest data
    local character_id = TheGame:GetGameState():GetCaravan():GetPlayer().id
    local act_data = GetPlayerActData( TheGame:GetGameState().current_act )

    self.title:SetText( string.upper( act_data:GetLocalizedTitle() ) )
    self.subtitle:SetText( "" )

    self.title:LayoutBounds( "center", "center", self.hitbox ):Offset( 0, -35 )
    self.subtitle:LayoutBounds( "center", "below", self.title ):Offset( 0, 0 )

end

function LocationScreenHeader:AnimateIn()
    --self.tracker:SetTintAlpha( 0 )

    --local tracker_x, tracker_y = self.tracker:GetPos()

    self:RunUpdater(
        Updater.Parallel{
            -- Move in the tracker
            --Updater.Ease( function(v) self.tracker:SetTintAlpha(v) end, 0, 1, 0.8, easing.outQuad),
        
            -- Move in the titles
            Updater.Do( function()
                self.title:Animate( 10, 8, nil, nil, 0, 1, 2.5, easing.outQuad )
                self.subtitle:Animate( 29, 25, nil, nil, 0, 1, 2, easing.outQuad )
            end ),
        }
    )
    
end

function LocationScreenHeader:SetLocation( location )
    self.title:SetText( string.upper( location:GetName() or "" ) )
end

