require "ui/widgets/screen"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/panelbutton"
require "util/rewardutil"

-------------------------------------------------------------------------
--

local LootScreen = class("Screen.LootScreen", Widget.Screen)

local PANEL_H = 500
local CONTENT_W = RES_X*.45
local CONTENT_X = -350

function LootScreen:init( ondonefn, data ) -- set negotiation to false for battles
    Widget.Screen.init(self)

    self.is_popup = true
    self:SetupUnderlay( true )
    self.ondonefn = ondonefn
    self:SetAnchors( "center", "center" )
    self.options = {}
    self.negotiation = data and data.negotiation or false
    self.battle = data and data.battle or false
    self.found = data and data.found or false
    self.defeat = data and data.defeat or false

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- A container for all the background elements
    self.panel = self:AddChild( Widget.WidePanel(PANEL_H) )

    -- Our fancy character anim!
    self.anim_root = self:AddChild(Widget()):SetHiddenBoundingBox( true )
    self:RefreshBigCharacter()

    -- Our titles above the background panel
    self.text_content = self:AddChild(Widget())
    self.subtitle = self.text_content:AddChild( Widget.Label("title", FONT_SIZE.BODY_TEXT ))
        :SetText( LOC"UI.LOOT_SCREEN.SUBTITLE" )
        :Bloom( 0.1 )
    local title_text = ""
    if self.negotiation and not self.defeat then title_text = LOC"UI.LOOT_SCREEN.TITLE_NEGOTIATION_VICTORY" end
    if self.negotiation and self.defeat then title_text = LOC"UI.LOOT_SCREEN.TITLE_NEGOTIATION_DEFEAT" end
    if self.battle then title_text = LOC"UI.LOOT_SCREEN.TITLE_BATTLE" end
    if self.found then title_text = LOC"UI.LOOT_SCREEN.TITLE_FOUND" end
    self.title = self.text_content:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE)
        :SetText( title_text )
        :SetGlyphColour(UICOLOURS.SUBTITLE) )
        :Bloom( 0.05 )
        :LayoutBounds( "center", "above", self.subtitle )
        :Offset( 0, -10 )
    self.text_content:LayoutBounds( "center", "above", self.panel )
        :Offset( 0, 15 )

    -- Loot section
    self.loot_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( SPACING.M1/2 )
        :SetSize( CONTENT_W, PANEL_H )
        :LayoutBounds( "left", "center", CONTENT_X, 0 )
    self.loot_scroll_root = self.loot_scroll_panel:AddScrollChild( Widget() )

    -- Continue button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -380, 0 )
    end
    self.btn_close = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.LOOT_SCREEN.LEAVE", function() self:OnClickClose() end ) )
        :SetIcon( global_images.accept )
        :LayoutBounds( "right", "above", -60, 60 )
    self:RegisterHotkeyButton( Controls.Digital.MENU_CANCEL, self.btn_close )
end


LootScreen.CONTROL_MAP = {
    {
        control = Controls.Digital.SKIP,
        test = function(self)
            if TheGame:FE():IsGamepadMode() then
                local must_loot = false
                for i, opt in ipairs( self.options ) do
                    must_loot = must_loot or opt.must_loot
                end
                return not must_loot and #self.options > 0 and not self.show_reticule
            end
        end,
        fn = function(self)
            self:OnClickClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.LEAVE_REWARDS", Controls.Digital.SKIP))
        end
    },
}

function LootScreen:OnClickClose()
    local confirm = false
    for i, opt in ipairs( self.options ) do
        if opt.confirm_close then
            confirm = true
            break
        end
    end
    if confirm then
        -- We still have stuff. Sure you want to let go of it?
        TheGame:FE():PushScreen( Screen.YesNoPopup( LOC"UI.LOOT_SCREEN.CONFIRM_TITLE", LOC"UI.LOOT_SCREEN.CONFIRM_TEXT" ) )
            :SetFn( function(v) if v == Screen.YesNoPopup.YES then 
                self:DoClose()
            end end )
    else
        self:DoClose()
    end
end

function LootScreen:RefreshClose()
    local must_loot = false
    for i, opt in ipairs( self.options ) do
        must_loot = must_loot or opt.must_loot
    end

    if must_loot then
        self.btn_close:Disable():SetToolTip( LOC"UI.LOOT_SCREEN.LOOT_REMAINING" )
    else
        self.btn_close:Enable():SetToolTip( nil )
    end
end

function LootScreen:RefreshBigCharacter()
    local OFFSET = RES_X/4
    
    local agent = TheGame:GetGameState():GetPlayerAgent()
    self.character = self.anim_root:AddChild( Widget.AnimatedCharacter() )
    self.character:Refresh(agent)

    local x = -RES_X*0.33
    local y = -310
    if not self.defeat then
        self.character:SetMood( MOOD.HAPPY )
    else
        self.character:Emote( "disappoint" )
    end

    self.character:SetScale( 1 )
    self.character:SetPos( x-OFFSET, y )
    self.character:MoveTo( x, y, .6, easing.outQuad)
end

function LootScreen:AddMoney( amount, title )
    assert( amount > 0 )

    local data = {
        money = amount
    }
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5,
        title )

    self.loot_scroll_root:AddChild( opt )
    
    table.insert( self.options, opt )

    opt:SetOnClickFn( function()
        AUDIO:PlayEvent( SoundEvents.conversation_accept_money )
        TheGame:GetGameState():GetCaravan():AddMoney( amount )
        self:RemoveOption( opt )
        self:Evaluate()
    end )

    self:LayoutOptions()
end

function LootScreen:AddResolve( amount )
    assert( amount > 0 )

    local data = {
        resolve = amount
    }
    local txt = loc.format( "Regain <#TITLE>{1} Resolve</>", amount )
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5 )
    self.loot_scroll_root:AddChild( opt )
    
    table.insert( self.options, opt )

    opt:SetOnClickFn( function()
        local resolve_gained = TheGame:GetGameState():GetCaravan():DeltaResolve( amount )
        if resolve_gained > 0 then
            AUDIO:PlayEvent("event:/ui/char/gen/gain_resolve")
        end
        self:RemoveOption( opt )
        self:Evaluate()
    end )

    self:LayoutOptions()
end


function LootScreen:AddMettle( amount, is_bonus )
    assert( amount > 0 )

    local data = {}

    if is_bonus then
        data.mettle_bonus = amount
    else
        data.mettle = amount
    end

    --local txt = loc.format( "Gain <#TITLE>{1} Mettle</>", amount )
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5 )
    self.loot_scroll_root:AddChild( opt )
    
    table.insert( self.options, opt )

    opt:SetOnClickFn( function()
        TheGame:GetGameProfile():AddMettlePoints(TheGame:GetGameState():GetPlayerAgent():GetContentID(), amount, "CARD_BONUS")
        self:RemoveOption( opt )
        self:Evaluate()
    end )

    self:LayoutOptions()
end

function LootScreen:AddHealth( amount )
    assert( amount > 0 )

    local data = {
        health = amount
    }
    --local txt = loc.format( "Regain <#HEALTH>{1} Health</>", amount )
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5 )
    self.loot_scroll_root:AddChild( opt )
    
    table.insert( self.options, opt )

    opt:SetOnClickFn( function()
        local health_gained = TheGame:GetGameState():GetPlayerAgent():GetAspect("health"):Delta(amount)
        if health_gained > 0 then
            AUDIO:PlayEvent("event:/ui/char/gen/gain_health")
        end
        self:RemoveOption( opt )
        self:Evaluate()
    end )

    self:LayoutOptions()
end

function LootScreen:AddGraft( graft_id )
    local data = {
        graft_id = graft_id
    }
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5 )
    self.loot_scroll_root:AddChild(opt)
    table.insert( self.options, opt )

    opt:SetOnClickFn( function()
        AUDIO:PlayEvent("event:/ui/select_cards/aquired_graft")
        local player = TheGame:GetGameState():GetPlayerAgent()
        local graft = GraftInstance( graft_id )
        if player.graft_owner:CanInstallGraft(graft.id) then
            player.graft_owner:AddGraft(graft)
            self:RemoveOption( opt )
            self:Evaluate()
        else
            local screen = Screen.ReplaceGraftScreen( graft, 
                function(replaced) 
                    if replaced then
                        self:RemoveOption( opt )
                        self:Evaluate()
                    end
                end )
            TheGame:FE():InsertScreen( screen )
        end
    end )

    self:LayoutOptions()
end

function LootScreen:AddCard( card_id )
    local data = {
        card_id = card_id
    }
    local opt = Widget.LootOption( 
        data,
        CONTENT_W-SPACING.M1*2.5 )
    self.loot_scroll_root:AddChild(opt)
    table.insert( self.options, opt )

    opt:SetOnClickFn( function( card )
        AUDIO:PlayEvent("event:/ui/select_cards/aquired_item")
        local player = TheGame:GetGameState():GetPlayerAgent()
        local carddef = Content.GetBattleCard( card_id ) or Content.GetNegotiationCard( card_id )
        local card = carddef.is_negotiation and Negotiation.Card( card_id ) or Battle.Card( card_id )
        if is_instance( card, Negotiation.Card ) then
            player.negotiator:LearnCard( card_id )
        elseif is_instance( card, Battle.Card ) then
            player.battler:LearnCard( card_id )
        end
        

        opt:SetToolTip(nil)
        local w, h = TheGame:FE():GetScreenDims()
        local scale = TheGame:FE():GetBaseWidgetScale()
        UIHelpers.ShowGainCards({card}, w-scale*300, h*.5)
        

        self:RemoveOption( opt )
        self:Evaluate()
    end )

    self:LayoutOptions()
end

function LootScreen:AddBattleCardChoice( cards )
    if #cards > 0 then
        local data = {
            battle_cards = cards
        }
        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            self:RemoveOption( opt )
            local draft_popup = Screen.DraftChoicePopup()
            draft_popup:SetMusicEvent( nil )
            draft_popup:DraftCards( nil, Battle.Card, cards, function() self:Evaluate() end )
            TheGame:FE():InsertScreen( draft_popup )
        end )

        self:LayoutOptions()
    end
end

function LootScreen:AddNegotiationCardChoice( cards )
    if #cards > 0 then
        local data = {
            negotiation_cards = cards
        }
        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            self:RemoveOption( opt )
            local draft_popup = Screen.DraftChoicePopup()
            draft_popup:SetMusicEvent( nil )
            draft_popup:DraftCards( nil, Negotiation.Card, cards, function() self:Evaluate() end )
            TheGame:FE():InsertScreen( draft_popup )
        end )

        self:LayoutOptions()
    end
end

function LootScreen:AddBattleCardSwap( count, drafts )
    if count > 0 then
        local data = {
            battle_card_swap = count,
        }
        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            for i,cards in ipairs(drafts) do
                local draft_popup = Screen.DraftChoicePopup()
                draft_popup:SetMusicEvent( nil )
                local title = loc.format( LOC "UI.DRAFT_SCREEN.DRAFT_CARDS", i, count )
                draft_popup:DraftCards( nil, Battle.Card, cards, function() self:Evaluate() end, title, nil, true )
                TheGame:FE():InsertScreen( draft_popup )
            end
            local removed = 0
            local function OnDone(card) 
                if card then
                    removed = removed + 1
                    if removed >= count then
                        self:RemoveOption( opt )
                    else
                        AgentUtil.RemoveBattleCard( TheGame:GetGameState():GetPlayerAgent(), OnDone, true)
                    end
                    self:Evaluate()
                end
            end
            AgentUtil.RemoveBattleCard( TheGame:GetGameState():GetPlayerAgent(), OnDone, true)
        end )

        self:LayoutOptions()
    end
end

function LootScreen:_CheckUpgradeOption( key, values )
    for k,v in ipairs(self.options) do
        if v.data and v.data[ key ] ~= nil then
            if shallowcompare( v.data[ key ], values ) then
                -- Already exists
                return false
            else
                -- Exists, but is different, so we need to remove and re-add.
                self:RemoveOption( v )
                break
            end
        end
    end

    return #values > 0
end

function LootScreen:AddUpgradeReminders()

    -- Check if there are battle cards ready to upgrade
    local battle_cards = TheGame:GetGameState():GetPlayerAgent().battler:GetCardsReadyForUpgrade()
    if self:_CheckUpgradeOption( "upgrade_battle_cards", battle_cards ) then
        local data = {
            upgrade_battle_cards = battle_cards
        }

        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            self:RemoveOption( opt )
            local draft_popup = Screen.BattleDeck()
            TheGame:FE():InsertScreen( draft_popup )
        end )
    end

    -- Check if there are negotiation cards ready to upgrade
    local negotiation_cards = TheGame:GetGameState():GetPlayerAgent().negotiator:GetCardsReadyForUpgrade()
    if self:_CheckUpgradeOption( "upgrade_negotiation_cards", negotiation_cards ) then
        local data = {
            upgrade_negotiation_cards = negotiation_cards
        }

        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            self:RemoveOption( opt )
            local draft_popup = Screen.NegotiationDeck()
            TheGame:FE():InsertScreen( draft_popup )
        end )
    end

    -- Check if there are grafts ready to upgrade
    local grafts = TheGame:GetGameState():GetPlayerAgent().graft_owner:GetGraftsReadyForUpgrade()
    if self:_CheckUpgradeOption( "upgrade_grafts", grafts ) then
        local data = {
            upgrade_grafts = grafts
        }

        local opt = Widget.LootOption(
            data,
            CONTENT_W-SPACING.M1*2.5 )
        self.loot_scroll_root:AddChild(opt)
        table.insert( self.options, opt )

        opt:SetOnClickFn( function()
            self:RemoveOption( opt )
            TheGame:FE():InsertScreen( Screen.GraftCollectionScreen() )
        end )
    end

    -- Layout all the loot rows
    self:LayoutOptions()

end

function LootScreen:Evaluate()
    if #self.options == 0 then
        self:DoClose()
    else
        self:AddUpgradeReminders()

        for k,v in ipairs(self.options) do
            self:AddRobotOption(v)
        end

    end

end

function LootScreen:RemoveOption( opt )
    -- Look for this row in the list
    local opt_index = table.arrayfind( self.options, opt )
    -- Remove it from the list
    table.arrayremove( self.options, opt )
    -- Remove the widget
    opt:Remove()
    -- Layout everything!
    self:LayoutOptions()
    -- Focus on an analogous row
    if self.loot_scroll_root.children and TheGame:FE():IsGamepadMode() then
        if self.loot_scroll_root.children[opt_index] then
            self.loot_scroll_root.children[opt_index]:SetFocus()
        elseif self.loot_scroll_root.children[opt_index-1] then
            self.loot_scroll_root.children[opt_index-1]:SetFocus()
        elseif self.loot_scroll_root.children[opt_index+1] then
            self.loot_scroll_root.children[opt_index+1]:SetFocus()
        end
    end
end

function LootScreen:LayoutOptions()
    self.loot_scroll_root:StackChildren( SPACING.M1/2 )
        :LayoutBounds( "left", "top", SPACING.M1, 0 )
    self.loot_scroll_panel:RefreshView()
    self:RefreshClose()
end

function LootScreen:IsEmpty()
    return #self.loot_scroll_root.children == 0
end

function LootScreen:OnGainTop()
    LootScreen._base.OnGainTop( self )
    self:Evaluate()
end

function LootScreen:OnControlModeChange( cm, device_id )
    LootScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.bottom_right:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_right:Hide()
    end

    self.loot_scroll_panel:ShowBar( cm == CONTROL_MODE.TOUCH and SCROLLBAR.NEVER or SCROLLBAR.IF_NEEDED )
end

function LootScreen:OnScreenModeChange( sm )
    LootScreen._base.OnScreenModeChange( self, sm )

    if sm == SCREEN_MODE.MONITOR then
        PANEL_H = 540
        CONTENT_X = -360
        self.character:SetPos(-RES_X*0.33, -310)
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        PANEL_H = 660
        CONTENT_X = -460
        self.character:SetPos(-RES_X*0.38, -310)
    end
    self.panel:SetHeight( PANEL_H )
    self.loot_scroll_root:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.loot_scroll_panel:SetSize( CONTENT_W*LAYOUT_SCALE[sm], PANEL_H-2 )
        :LayoutBounds( "left", "center", CONTENT_X, -1 )

    self.text_content:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )

    self.text_content:LayoutBounds( "center", "above", self.panel )
        :Offset( 0, 15 )

    -- self:Layout()
end

function LootScreen:RefreshStreamerMode()
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

function LootScreen:GetDefaultFocus()
    for k, c in ipairs( self.loot_scroll_root.children ) do
        return c
    end
    return self.btn_close
end

function LootScreen:DoClose()
    if not self:IsOnStack() then
        -- This can open if a screen opened by the LootScreen is closed, and LootScreen:OnGainTop calls Evaluate redundantly.
        return
    end

    AUDIO:PlayEvent(SoundEvents.lootscreen_close)
    self:GetFE():PopScreen( self )
    if self.ondonefn then
        self.ondonefn()
    end
end

local ReticuleMixin = require "ui/screens/ReticuleMixin"
ReticuleMixin:Mixin( LootScreen )

-----------------------------------------------------------------------------------
-- A single loot reward button

local LootOption = class( "Widget.LootOption", Widget.Clickable )

local BADGE_W = 160
local BADGE_PADDING = -1
local ROW_H = 90
local MASK_TOP_OVERFLOW = 30
local CARD_SCALE = 0.26

function LootOption:init( data, width, title )
    LootOption._base.init( self )

    self.hitbox = self:AddChild( Widget.SolidBox( width, ROW_H, 0xff000070 ))
        :Hide()

    self:ShowToolTipOnFocus( true )

    -- REWARD BADGE
    self.reward = self:AddChild( Widget() )
    local bg_border = 33
    local bg_border_scale = 0.5
    self.reward_bg = self.reward:AddChild( Widget.Panel( engine.asset.Texture("UI/rewardwidget_bg.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border*2, bg_border*2 )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( BADGE_W, ROW_H )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.reward_content = self.reward:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
        :SetScissor( -BADGE_W/2, -(ROW_H)/2, BADGE_W, ROW_H+MASK_TOP_OVERFLOW )
    self.reward_gradient = self.reward:AddChild( Widget.Panel( engine.asset.Texture("UI/rewardwidget_gradient.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border*2, bg_border*2 )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( BADGE_W, ROW_H )
        :SetTintColour( UICOLOURS.SUBTITLE )
    self.reward_overlay = self.reward:AddChild( Widget.Panel( engine.asset.Texture("UI/rewardwidget_overlay.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border*2, bg_border*2 )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( BADGE_W, ROW_H )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.reward:LayoutBounds( "left", "bottom", self.hitbox )
    
    -- BUTTON
    self.button = self:AddChild( Widget() )
    local bg_border = 20
    local bg_border_scale = 0.5
    local button_w = width - BADGE_PADDING - BADGE_W
    local nineslice_w, nineslice_h = button_w + bg_border*2*bg_border_scale, ROW_H + bg_border*2*bg_border_scale
    self.button_bg = self.button:AddChild( Widget.Panel( engine.asset.Texture("UI/loot_button.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetTintAlpha( 0.6 )
        :Bloom( 0.05 )
    self.text_content = self.button:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label("title", FONT_SIZE.ITEM_SUBTITLE ))
        :SetGlyphColour( UICOLOURS.BUTTON_TEXT )
        :RightAlign()
        :SetWordWrap( true )
        :OverrideLineHeight( 22 )
        :SetAutoSize( button_w - SPACING.M05 )
        :Bloom( 0.0 )
    self.subtitle = self.text_content:AddChild( Widget.Label("body", FONT_SIZE.ITEM_DESC ))
        :SetGlyphColour( UICOLOURS.BUTTON_TEXT )
        :SetTintAlpha( 0.8 )
        :RightAlign()
        :SetWordWrap( true )
        :OverrideLineHeight( 18 )
        :SetAutoSize( button_w - SPACING.M05 )
        :Bloom( 0.0 )
    self.button:LayoutBounds( "right", "bottom", self.hitbox )

    -- If this option is untaken, confirm with the player the screen can close.
    self.confirm_close = true
    self.data = data

    if data.card_id then
        -- Card
        local carddef = Content.GetBattleCard(data.card_id) or Content.GetNegotiationCard(data.card_id)
        local card = carddef.is_negotiation and Negotiation.Card(data.card_id) or Battle.Card(data.card_id)
        local card_widget
        if is_instance( card, Negotiation.Card ) then
            card_widget = self.reward_content:AddChild( Widget.NegotiationCard( card ) )
                :SetOutline( UICOLOURS.NEGOTIATION, true )

            if card.must_loot then
                self.title:SetText( LOC"UI.LOOT_SCREEN.MANDATORY_CARD_NEGOTIATION_TITLE" )
                self.must_loot = true
            else
                self.title:SetText( LOC"UI.LOOT_SCREEN.CARD_NEGOTIATION_TITLE" )
            end
            self.reward_bg:SetTintColour( UICOLOURS.NEGOTIATION )
            self.reward_gradient:SetTintColour( UICOLOURS.NEGOTIATION )
            self.reward_overlay:SetTintColour( UICOLOURS.NEGOTIATION )
            self:SetToolTipClass( Widget.TooltipNegotiationCard )
            self:SetToolTip( { CARD = card } )

        elseif is_instance( card, Battle.Card ) then
            card_widget = self.reward_content:AddChild( Widget.BattleCard( card ) )
                :SetOutline( UICOLOURS.FIGHT, true )

            if card.must_loot then
                self.title:SetText( LOC"UI.LOOT_SCREEN.MANDATORY_CARD_BATTLE_TITLE" )
                self.must_loot = true
            else
                self.title:SetText( LOC"UI.LOOT_SCREEN.CARD_BATTLE_TITLE" )
            end
            self.reward_bg:SetTintColour( UICOLOURS.FIGHT )
            self.reward_gradient:SetTintColour( UICOLOURS.FIGHT )
            self.reward_overlay:SetTintColour( UICOLOURS.FIGHT )
            self:SetToolTipClass( Widget.TooltipBattleCard )
            self:SetToolTip( { CARD = card } )
        end
        self.subtitle:SetText( card:GetName() )
        card_widget:SetScale( CARD_SCALE )
            :SetRotation( -10*DEGREES )
            :SetPos( -5, -20 )

    elseif data.graft_id then
        -- Graft
        local graft = GraftInstance( data.graft_id )
        local image = self.reward_content:AddChild( Widget.Image( graft:GetImage() ) )
            :SetSize( ROW_H, ROW_H )
            :SetPos( 0, 10 )
        if graft:GetType() == GRAFT_TYPE.NEGOTIATION then
            self.title:SetText( LOC"UI.LOOT_SCREEN.GRAFT_NEGOTIATION_TITLE" )
            self.reward_bg:SetTintColour( UICOLOURS.NEGOTIATION )
            self.reward_gradient:SetTintColour( UICOLOURS.NEGOTIATION )
            self.reward_overlay:SetTintColour( UICOLOURS.NEGOTIATION )
        else
            self.title:SetText( LOC"UI.LOOT_SCREEN.GRAFT_BATTLE_TITLE" )
            self.reward_bg:SetTintColour( UICOLOURS.FIGHT )
            self.reward_gradient:SetTintColour( UICOLOURS.FIGHT )
            self.reward_overlay:SetTintColour( UICOLOURS.FIGHT )
        end
        TheGame:GetGameProfile():SetSeenGraft( graft.id )
        self:SetToolTipClass( Widget.TooltipGraft )
        self:SetToolTip( graft )
        self.subtitle:SetText( graft:GetName() )

    elseif data.battle_cards then
        -- Battle cards choice
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_battle.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.FIGHT )
        self.reward_gradient:SetTintColour( UICOLOURS.FIGHT )
        self.reward_overlay:SetTintColour( UICOLOURS.FIGHT )
        self.title:SetText( LOC"UI.LOOT_SCREEN.CARDS_BATTLE_TITLE" )
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_BATTLE_SUBTITLE", #data.battle_cards ) )
        self:SetToolTip( loc.format( LOC"UI.LOOT_SCREEN.CARDS_BATTLE_SUBTITLE", #data.battle_cards ) )

    elseif data.negotiation_cards then
        -- Negotiation cards choice
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_negotiation.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.NEGOTIATION )
        self.reward_gradient:SetTintColour( UICOLOURS.NEGOTIATION )
        self.reward_overlay:SetTintColour( UICOLOURS.NEGOTIATION )
        self.title:SetText( LOC"UI.LOOT_SCREEN.CARDS_NEGOTIATION_TITLE" )
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_NEGOTIATION_SUBTITLE", #data.negotiation_cards ) )
        self:SetToolTip( loc.format( LOC"UI.LOOT_SCREEN.CARDS_NEGOTIATION_SUBTITLE", #data.negotiation_cards ) )
    
    elseif data.upgrade_battle_cards then
        -- Battle upgrades available
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_upgrade_battle.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.FIGHT )
        self.reward_gradient:SetTintColour( UICOLOURS.FIGHT )
        self.reward_overlay:SetTintColour( UICOLOURS.FIGHT )
        self.title:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_BATTLE_TITLE", #data.upgrade_battle_cards ) )
        local card_titles = {}
        for k, card in ipairs( data.upgrade_battle_cards ) do
            table.insert( card_titles, card:GetName() )
        end
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_BATTLE_SUBTITLE", card_titles ) )
        self.confirm_close = false
    
    elseif data.upgrade_negotiation_cards then
        -- Negotiation upgrades available
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_upgrade_negotiation.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.NEGOTIATION )
        self.reward_gradient:SetTintColour( UICOLOURS.NEGOTIATION )
        self.reward_overlay:SetTintColour( UICOLOURS.NEGOTIATION )
        self.title:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_NEGOTIATION_TITLE", #data.upgrade_negotiation_cards ) )
        local card_titles = {}
        for k, card in ipairs( data.upgrade_negotiation_cards ) do
            table.insert( card_titles, card:GetName() )
        end
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_NEGOTIATION_SUBTITLE", card_titles ) )
        self.confirm_close = false
    
    elseif data.upgrade_grafts then
        -- Negotiation upgrades available
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_upgrade_graft.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.GRAFT )
        self.reward_gradient:SetTintColour( UICOLOURS.GRAFT )
        self.reward_overlay:SetTintColour( UICOLOURS.GRAFT )
        self.title:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_GRAFT_TITLE", #data.upgrade_grafts ) )
        local card_titles = {}
        for k, card in ipairs( data.upgrade_grafts ) do
            table.insert( card_titles, card:GetName() )
        end
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_UPGRADE_GRAFT_SUBTITLE", card_titles ) )
        self.confirm_close = false

    elseif data.money then
        -- Shills
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_shills.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 12 )
        self.reward_bg:SetTintColour( UICOLOURS.MONEY )
        self.reward_gradient:SetTintColour( UICOLOURS.MONEY )
        self.reward_overlay:SetTintColour( UICOLOURS.MONEY )
        self.title:SetText( title or LOC"UI.LOOT_SCREEN.MONEY_TITLE" )
        self.subtitle:SetText( loc.format( "{1#money}", data.money ) )
            :SetTintAlpha( 1 )
    elseif data.mettle then
        -- Shills
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_mettle.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 12 )
        self.reward_bg:SetTintColour( UICOLOURS.METTLE )
        self.reward_gradient:SetTintColour( UICOLOURS.METTLE )
        self.reward_overlay:SetTintColour( UICOLOURS.METTLE )
        self.title:SetText( title or LOC"UI.LOOT_SCREEN.METTLE_TITLE" )
        self.subtitle:SetText( loc.format( "{1#mettle}", data.mettle ) )
            :SetTintAlpha( 1 )
    elseif data.mettle_bonus then
        -- Shills
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_mettle.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 12 )
        self.reward_bg:SetTintColour( UICOLOURS.METTLE )
        self.reward_gradient:SetTintColour( UICOLOURS.METTLE )
        self.reward_overlay:SetTintColour( UICOLOURS.METTLE )
        self.title:SetText( title or LOC"UI.LOOT_SCREEN.METTLE_BONUS_TITLE" )
        self.subtitle:SetText( loc.format( "{1#mettle}", data.mettle_bonus ) )
            :SetTintAlpha( 1 )
    elseif data.resolve then
        -- Shills
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_resolve.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 12 )
        self.reward_bg:SetTintColour( UICOLOURS.RESOLVE )
        self.reward_gradient:SetTintColour( UICOLOURS.RESOLVE )
        self.reward_overlay:SetTintColour( UICOLOURS.RESOLVE )
        self.title:SetText( LOC"UI.LOOT_SCREEN.RESOLVE_TITLE" )
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.RESOLVE_SUBTITLE", data.resolve ) )
            :SetTintAlpha( 1 )
    elseif data.health then
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_health.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 12 )
        self.reward_bg:SetTintColour( UICOLOURS.HEALTH )
        self.reward_gradient:SetTintColour( UICOLOURS.HEALTH )
        self.reward_overlay:SetTintColour( UICOLOURS.HEALTH )
        self.title:SetText( LOC"UI.LOOT_SCREEN.HEALTH_TITLE" )
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.HEALTH_SUBTITLE", data.health ) )
            :SetTintAlpha( 1 )
    elseif data.battle_card_swap then
        -- Battle cards swap
        local image = self.reward_content:AddChild( Widget.Image( engine.asset.Texture( "UI/ic_loot_battle.tex" ) ) )
            :SetSize( ROW_H+20, ROW_H+20 )
            :SetPos( 0, 5 )
        self.reward_bg:SetTintColour( UICOLOURS.FIGHT )
        self.reward_gradient:SetTintColour( UICOLOURS.FIGHT )
        self.reward_overlay:SetTintColour( UICOLOURS.FIGHT )
        self.title:SetText( loc.format(LOC"UI.LOOT_SCREEN.CARDS_SWAP_BATTLE_TITLE", data.battle_card_swap ) )
        self.subtitle:SetText( loc.format( LOC"UI.LOOT_SCREEN.CARDS_SWAP_BATTLE_SUBTITLE", data.battle_card_swap ) )
        self:SetToolTip( loc.format( LOC"UI.LOOT_SCREEN.CARDS_SWAP_BATTLE_SUBTITLE", data.battle_card_swap ) )
    end

    -- Align text
    self.subtitle
        :LayoutBounds( "left", "below", self.title )
        :Offset( 0, 4 )
    self.text_content
        :LayoutBounds( "left", "center", self.button_bg )
        :Offset( SPACING.M1, 0 )
    self.reward_gradient:SetTintAlpha( 0.9 )
end

function LootOption:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        self.reward_overlay:Bloom( 0.3 )
        self.button_bg:AlphaTo( 1, 0.1 )
        if self.focus then self:ScaleTo( nil, 1.01, 0.2, easing.outQuad ) end
    else
        self.reward_overlay:Bloom( 0.1 )
        self.button_bg:AlphaTo( 0.6, 0.1 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function LootOption:OnGainFocus()
    LootOption._base.OnGainFocus(self)
    self:UpdateImage()
end

function LootOption:OnLoseFocus()
    LootOption._base.OnLoseFocus(self)
    self:UpdateImage()
end

function LootOption:OnGainHover()
    LootOption._base.OnGainHover(self)
    self:UpdateImage()
end

function LootOption:OnLoseHover()
    LootOption._base.OnLoseHover(self)
    self:UpdateImage()
end
