require "ui/widgets/installed_grafts"

local assets = {
    mask = engine.asset.Texture("UI/mettle_mask.tex"),
    available = engine.asset.Texture("UI/mettle_available.tex"),
    cost = engine.asset.Texture("UI/mettle_cost.tex"),
    selected = engine.asset.Texture("UI/mettle_selected.tex"),
    grid = engine.asset.Texture( "UI/grid.tex" ),
    dropdown_inner = engine.asset.Texture("UI/dropdown_inner.tex" ),
}

local MettleScreen = class( "Screen.MettleScreen", Widget.Screen )

local CELL_W = 480
local CELL_H = 112
local CONTAINER_H = 600

MettleScreen.CONTROL_MAP = Widget.MergeControlMap( Widget.Screen.CONTROL_MAP,
{
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self) 
            self:DoClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL) )
        end,
    },
})

function MettleScreen:init( owner, on_close )
    MettleScreen._base.init( self )
    self.on_close = on_close
    
    self.owner = owner or TheGame:GetGameState():GetPlayerAgent()

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( assets.grid ) )
    self.grid:SetTintColour( UICOLOURS.METTLE )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Backing
    self.widepanel = self:AddChild( Widget.WidePanel( 800, UICOLOURS.SUBTITLE ) )

    -- Header
    self.header = self:AddChild( Widget() ):SetAnchors( "left", "top" )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( loc.format( LOC"UI.METTLE_SCREEN.METTLE_SCREEN_TITLE", self.owner:GetName() ) )
        :SetGlyphColour( UICOLOURS.METTLE )
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.1 )
        :LayoutBounds( "left", "top", 0, -60 )
    self.subtitle = self.header:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.METTLE )
        :LeftAlign()
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )
    
    -- List scroll
    self.mettle_scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualTopMargin( 10 )
        :SetVirtualBottomMargin( 10 )
        :SetScrollBarOuterMargin( 0 )
        :SetSize( 600, CONTAINER_H )
    self.mettle_container = self.mettle_scroll:AddScrollChild( Widget() )

    -- Setup character anim display
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.owner )
    self.character:SetHoverOutlineColour( UICOLOURS.METTLE )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "eureka", true )
        end
    end )
    self.character:SetPos( -RES_X*0., -250 )
    self.character:Emote( "eureka", true )
    
    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:DoClose() end ) )
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
    
    self:AddRobotOption(self.close_button)
    self:Refresh()
end

function MettleScreen:DoClose()
    AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
    self:FadeOut()
end

function MettleScreen:Refresh()

    self.owner = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    self.mettle_container:DestroyAllChildren()

    -- Get owned upgrades
    local upgrades = TheGame:GetGameProfile():GetMettleUpgrades( self.owner.id )

    -- Get all mettle grafts
    self.grafts = Content.GetMettleGrafts( self.owner.id )
    table.sort(self.grafts, function(a,b) return (a.sort_key or 0) < (b.sort_key or 0) end)
    for i, graft_def in ipairs( self.grafts ) do
        local upgraded_level = upgrades[graft_def.id] or 0
        local widget = self.mettle_container:AddChild( Widget.MettleRowWidget( CELL_W, CELL_H, graft_def, upgraded_level, function(...) self:OnMettleUpgraded(...) end ) )
    end

    -- Get available mettle
    local mettle = TheGame:GetGameProfile():GetMettlePoints( self.owner.id ) or 0
    self.subtitle:SetText( loc.format( LOC"UI.METTLE_SCREEN.METTLE_SCREEN_SUBTITLE", mettle, self.owner:LocTable(), self.owner ) )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )

    -- Update rows
    for k, row in ipairs( self.mettle_container.children ) do
        row:SetCurrentMettle(mettle)
    end

    self:Layout()

    return self
end

function MettleScreen:OnMettleUpgraded( graft_id, level, cost, widget )

    -- Confirm acquisition
    TheGame:FE():PushScreen( 
        Screen.YesNoPopup( LOC"UI.METTLE_SCREEN.CONFIRM_TITLE", loc.format(LOC"UI.METTLE_SCREEN.CONFIRM_BODY", level, cost) )
            :SetFn(
                function(res) 
                    if res == Screen.YesNoPopup.YES then 
                        
                        -- Acquire upgrade
                        TheGame:GetGameProfile():BuyMettleUpgrade( self.owner.id, graft_id, level )

                        -- Remove mettle
                        TheGame:GetGameProfile():SpendMettlePoints( self.owner.id, cost, graft_id )
                        local handler = self.owner:GetAspect("mettle_handler")
                        
                        if handler then
                            handler:VerifyUpgrades()
                        end

                        -- Get available mettle
                        local mettle = TheGame:GetGameProfile():GetMettlePoints( self.owner.id )
                        self.subtitle:SetText( loc.format( LOC"UI.METTLE_SCREEN.METTLE_SCREEN_SUBTITLE", mettle, self.owner:LocTable(), self.owner ) )
                            :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )

                        -- Get owned upgrades
                        local upgrades = TheGame:GetGameProfile():GetMettleUpgrades( self.owner.id )

                        -- Refresh all rows
                        for k, row in ipairs( self.mettle_container.children ) do
                            local upgraded_level = upgrades[row.def.id] or 0
                            row:SetCurrentMettle(mettle)
                                :SetCurrentLevel(upgraded_level)
                        end

                    end
                end
            )
        )
end

function MettleScreen:OnOpen()
    MettleScreen._base.OnOpen( self )

    local anim_duration = 0.7
    local horizontal_movement = -90

    -- Prepare for animation by getting current values
    local content_x, content_y = self.content:GetPos()

    -- Hide objects until they're animated in
    self.content:SetTintAlpha(0)

    -- Setup and start animation
    self.fadein_anim = self:RunUpdater(
        Updater.Parallel{
            -- Fade and move in the coin container
            Updater.Ease( function(v) self.content:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) self.content:SetPos(x,y) end, content_x - horizontal_movement, content_y, content_x, content_y, anim_duration, easing.outQuad),
        }
    )

end

function MettleScreen:OnScreenResize(w, h)
    MettleScreen._base.OnScreenResize(self, w, h)
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
    self:Layout()
end

function MettleScreen:GetDefaultFocus()
    return self.mettle_container and self.mettle_container.children and #self.mettle_container.children>0 and self.mettle_container.children[1]:GetDefaultFocus()
end

function MettleScreen:OnControlModeChange( cm, device_id )
    MettleScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.mettle_scroll:ShowBar( SCROLLBAR.IF_NEEDED )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.mettle_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.mettle_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Hide()
    end

end

function MettleScreen:OnScreenModeChange( sm )
    MettleScreen._base.OnScreenModeChange( self, sm )

    -- Scale contents
    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )

    for k, v in pairs(self.mettle_container.children) do
        v:SetLayoutScale( LAYOUT_SCALE[sm] )
    end

    self:Layout()
end

function MettleScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end

    self:Layout()
end

function MettleScreen:Layout()

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    local sm = TheGame:FE():GetScreenMode()

    -- Update grid
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Header height
    local w, h = self.header:GetSize()
    local header_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[sm]
    -- Footer height
    local w, h = self.bottom_left:GetSize()
    local footer_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[sm]
    -- Content height
    self.content_h = self.screen_h - header_height - footer_height

    -- Update containers
    self.widepanel:SetHeight( self.content_h )
        :LayoutBounds( nil, "below", self.header )
        :Offset( 0, -SPACING.M1 )

    self.mettle_scroll:SetSize( self.screen_w-10, self.content_h )
        :LayoutBounds( "left", "center", self.widepanel )
        :Offset( 0, 3 )

    local left_offset = self.screen_w*0.18 -- How much the scroll panel should cross to the left of the screen center

    CELL_W = self.screen_w/2 + left_offset - 40
    CELL_W = CELL_W/LAYOUT_SCALE[TheGame:FE():GetScreenMode()]
    for k, v in pairs(self.mettle_container.children) do
        v:SetSize( CELL_W, CELL_H )
    end
    self.mettle_container:StackChildren( 10 )
        :LayoutBounds( "left", "top", 0, 0 )
        :Offset( self.screen_w/2-left_offset, 0 )

    self.character:SetPos( 0, -250 )
        :LayoutBounds( "before", nil, self.mettle_container )
        :Offset( -140, 0 )

    self.header:LayoutBounds( "left", nil, self.widepanel )
        :Offset( self.screen_w/2-left_offset+13*LAYOUT_SCALE[TheGame:FE():GetScreenMode()], 0 )

    self.mettle_scroll:RefreshView()

    return self
end

function MettleScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.mettle_scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.mettle_scroll:ScrollUp()
        return true
    end
end

function MettleScreen:HandleControlUp( control, device )
    if control:Has( Controls.Digital.MENU_CANCEL ) then
        AUDIO:PlayEvent( "event:/ui/mettle_shop/close" )
        self:FadeOut()
    end

    return MettleScreen._base.HandleControlUp( self, control, device )
end

function MettleScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function MettleScreen:OnOpen()
    MettleScreen._base.OnOpen( self )
    AUDIO:PlayEvent( "event:/ui/mettle_shop/open" )
end

function MettleScreen:OnClose()
    MettleScreen._base.OnClose( self )

    if self.on_close then
        self.on_close()
    end

end

-------------------------------------------------------------------------
-- A single Mettle row, with multiple buttons

local MettleRowWidget = class( "Widget.MettleRowWidget", Widget )

local PADDING = 26
local BUTTON_W = 90

function MettleRowWidget:init( w, h, def, upgraded_level, on_upgrade_fn, option_widget_override )
    MettleRowWidget._base.init( self )
    self.features = {}

    self.def = def
    self.graft = GraftInstance( def.id )
    self.upgraded_level = upgraded_level
    self.on_upgrade_fn = on_upgrade_fn

    self.contents = self:AddChild( Widget() )
    self.bg = self.contents:AddChild( Widget.Panel( assets.dropdown_inner ) )
        :SetTintColour( UICOLOURS.METTLE )
        :SetTintAlpha( 0 )
        :SetBloom( 0.1 )
    self.text_content = self.contents:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, self.graft:GetDesc( self ) ) )
        :LeftAlign()
        :SetAutoSize( w-PADDING )
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.ITEM_NAME*LINE_HEIGHT_TITLE_MULT )
        :SetGlyphColour( UICOLOURS.METTLE )
        :Bloom( 0.1 )
    self.buttons = self.contents:AddChild( Widget() )
    
    local option_widget = option_widget_override or Widget.MettleOptionButton
    for k, value in ipairs(self.def.values) do
        self.buttons:AddChild( option_widget( self.graft, self.def.values[k], self.def.upgrade_costs[k], function() self:OnUpgrade(k) end ) )
            :SetSize( BUTTON_W )
    end
    self:SetCurrentLevel( self.upgraded_level )

    self:SetSize( w, h )
    self.hover_check = true
end

function MettleRowWidget:GetDef()
    return self.def
end

function MettleRowWidget:GetGraft()
    return self.graft
end

-- Returns how many are unlocked and how many are unlockable (ie: 2, 6 )
function MettleRowWidget:GetProgress()
    return self.upgraded_level, #self.buttons.children
end

function MettleRowWidget:SetCurrentMettle( mettle )
    for k, button in ipairs( self.buttons.children ) do
        button:SetExpensive( button.mettle_cost > mettle )
            :Refresh()
    end
    return self
end

function MettleRowWidget:SetCurrentLevel( upgraded_level )
    self.upgraded_level = upgraded_level
    for k, button in ipairs( self.buttons.children ) do
        button:SetAcquired( k <= self.upgraded_level )
            :SetSelected( k == self.upgraded_level )
            :SetAvailable( k == self.upgraded_level+1 )
            :Refresh()
    end
    return self
end

function MettleRowWidget:OnUpgrade( option_index )
    AUDIO:PlayEvent("event:/ui/mettle_shop/unlock_mettle")
    if self.on_upgrade_fn then
        local widget = self.buttons.children[option_index]
        self.on_upgrade_fn( self.def.id, option_index, widget.mettle_cost, widget )
    end
end

function MettleRowWidget:SetSize( w, h )
    self.title:SetAutoSize( w-PADDING )
    -- self.text:SetAutoSize( w-PADDING )
    -- self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    self.bg:SetSize( w, h )
    self.text_content:LayoutBounds( "left", "top", self.bg ):Offset( PADDING/2, -PADDING/2 )
    self.buttons:StackChildrenRow( PADDING*0.6 )
        :LayoutBounds( "left", "below", self.text_content )
        :Offset( 0, -8 )
    return self
end

function MettleRowWidget:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function MettleRowWidget:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

function MettleRowWidget:GetDefaultFocus()
    for k, button in ipairs( self.buttons.children ) do
        return button
    end
end

function MettleRowWidget:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.bg:AlphaTo( 0.09, 0.1 )
    else
        self.bg:AlphaTo( 0, 0.1 )
    end
end

function MettleRowWidget:OnGainFocus()
    MettleRowWidget._base.OnGainFocus(self)
    self:UpdateImage()
end

function MettleRowWidget:OnLoseFocus()
    MettleRowWidget._base.OnLoseFocus(self)
    self:UpdateImage()
end

function MettleRowWidget:OnGainHover()
    MettleRowWidget._base.OnGainHover(self)
    self:UpdateImage()
end

function MettleRowWidget:OnLoseHover()
    MettleRowWidget._base.OnLoseHover(self)
    self:UpdateImage()
end


-------------------------------------------------------------------------
-- A single Mettle option button

local MettleOptionButton = class( "Widget.MettleOptionButton", Widget.IconButton )

MettleOptionButton.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.onclickfn
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.METTLE_SCREEN.CONTROLS_UNLOCK", Controls.Digital.MENU_ACCEPT, self.mettle_cost))
        end
    },
}

function MettleOptionButton:init( mettle_graft, mettle_value, mettle_cost, onclick_fn )
    MettleOptionButton._base.init( self )

    self.mettle_graft = mettle_graft
    self.mettle_value = mettle_value or 0
    self.mettle_cost = mettle_cost or 0
    self.acquired = false
    self.selected = false
    self.expensive = false
    self.available = false
    self.onclick_fn = onclick_fn
    self:SetOnClickFn( onclick_fn )
    self:HideIcon()
    self:HideGradient()

    -- Update textures
    self.bg_panel:SetTexture( assets.mask )
    self.bg_outline:SetTexture( assets.available )
    self.label:SetFontSize( FONT_SIZE.BODY_TEXT )

    -- Add cost badge
    self.cost_badge = self:AddChild( Widget.Panel( assets.cost ) )
        :SetHiddenBoundingBox( true )
    self.subtitle:SetHiddenBoundingBox( true )

    self:Refresh()
end

-- If it has been acquired. Can be the selected one simultaneously
function MettleOptionButton:SetAcquired( acquired )
    self.acquired = acquired
    return self
end

-- If it's the currently active level
function MettleOptionButton:SetSelected( selected )
    self.selected = selected
    return self
end

-- If its cost is higher than this character's amount of mettle
function MettleOptionButton:SetExpensive( expensive )
    self.expensive = expensive
    return self
end

-- If it can be purchased. Can be expensive simultaneously
function MettleOptionButton:SetAvailable( available )
    self.available = available
    return self
end

function MettleOptionButton:Refresh()
    self:SetText( self.mettle_value )
    if self.selected then
        self.bg_outline:SetTexture( assets.selected )
        self.bg_panel:SetTintColour( 0xC6AB3Eff )
        self:SetColour( UICOLOURS.QUEST )
        self:SetTextColour( UICOLOURS.DARK_BODYTEXT )
        self:SetOnClickFn(nil)
    elseif self.acquired then
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( 0x6B5A16ff )
        self:SetColour( UICOLOURS.QUEST )
        self:SetTextColour( 0xC6AB3Eff )
        self:SetOnClickFn(nil)
    elseif self.available and self.expensive then
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( UICOLOURS.DARK_BODYTEXT )
        self:SetColour( 0xFF636380 )
        self:SetTextColour( 0x666666ff )
        self.cost_badge:SetTintColour( UICOLOURS.PENALTY )
        self:SetOnClickFn(nil)
    elseif self.available and not self.expensive then
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( UICOLOURS.DARK_BODYTEXT )
        self:SetColour( 0x40D988ff )
        self:SetTextColour( 0x40D988ff )
        self.cost_badge:SetTintColour( UICOLOURS.BONUS )
        self:SetOnClickFn( self.onclick_fn )
    else
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( UICOLOURS.DARK_BODYTEXT )
        self:SetColour( 0x666666ff )
        self:SetTextColour( 0x666666ff )
        self.cost_badge:SetTintColour( 0x666666ff )
        self:SetOnClickFn(nil)
    end

    -- Update cost badge
    if self.acquired then
        self.cost_badge:Hide()
        self.subtitle:Hide()
    else
        self.cost_badge:Show()
        self.subtitle:Show()
            :SetText( loc.format(LOC"UI.METTLE_SCREEN.METTLE_COST",self.mettle_cost) )
        local w, h = self.subtitle:GetSize()
        self.cost_badge:SetSize( w + 10, h + 0 )
        self.text_container:SendToFront()
    end

    self:UpdateImage()
    return self
end

function MettleOptionButton:Layout()
    MettleOptionButton._base.Layout( self )

    if self.cost_badge then
        self.label:LayoutBounds( "center", "center", 0, 0  )
            :Offset( 0, self.selected and 0 or 1 )
        self.cost_badge:LayoutBounds( "right", "bottom", self.width/2, -self.height/2 )
            :Offset( 4, -4 )
        self.subtitle:LayoutBounds( "center", "center", self.cost_badge )
            :Offset( 0, 0 )
    end

    return self
end

function MettleOptionButton:UpdateImage()
    MettleOptionButton._base.UpdateImage( self )

    if self.hover or self.focus then
        self:ScaleTo( nil, self.selected and 1.1 or 1.05, 0.2, easing.outQuad )
    else
        self:ScaleTo( nil, self.selected and 1.08 or 1, 0.2, easing.outQuad )
    end
    self.subtitle:SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.DARK_BODYTEXT )

    return self
end
