require "ui/widgets/screen"
require "ui/widgets/checkbox"

local MOD_WIDTH = 300 -- Width of a single Widget.ModWidget
local MOD_HEIGHT = 380 -- Height of a single Widget.ModWidget

-- Returns true if this mod is up to date.
local function IsModUpdated( mod )
    return mod.workshop and not CheckBits( mod.workshop.flags, engine.steam.EItemStateNeedsUpdate )
end

local function IsModSubscribed( mod )
    return mod.workshop and CheckBits( mod.workshop.flags, engine.steam.EItemStateSubscribed )
end

----------------------------------------------------------------------
-- Shows a list of installed mods (local mods, and workshop mods)

local ModManagerScreen = class("Screen.ModManagerScreen", Widget.Screen)

ModManagerScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.GAMEPAD_LT,
        hint = function(self, left, right)
            table.insert(left, loc.format( LOC"UI.MOD_SCREEN.CONTROLS_WORKSHOP", Controls.Digital.GAMEPAD_LT))
        end,
        test = function()
            return engine.steam ~= nil
        end,
        fn = function(self)
            self:OnClickBrowseWorkshop()   
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end,
        fn = function(self)
            self.close_button:Click()    
            return true
        end,
    }
}

function ModManagerScreen:init()
	Widget.Screen.init( self )

    -- Keep track of original enabled status of mods, to detect changes on close.
    -- This is a map of mod_id -> boolean
    self.enabled_mods = {}

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Header
    self.header = self:AddChild( Widget() ):SetAnchors( "left", "top" )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE, LOC"UI.MOD_SCREEN.TITLE" ) )
        :SetGlyphColour( UICOLOURS.MOD )
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.1 )
        :LayoutBounds( "left", "top", 60, -60 )
    self.subtitle = self.header:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT, LOC"UI.MOD_SCREEN.DESCRIPTION" ) )
        :LeftAlign()
        :SetAutoSize( RES_X/2 )
        :SetWordWrap( true )
        :Bloom( 0.05 )
        :LayoutBounds( "left", "below", self.title ):Offset( 0, 0 )

    -- Backing
    self.widepanel = self:AddChild( Widget.WidePanel( 800, UICOLOURS.MOD ) )

    -- Scroll panel for all ModWidgets
    self.content_scroll = self:AddChild( Widget.ScrollPanel() )
        :SetAnchors( "center", "center" )
        :SetVirtualMargin( SPACING.M1 )
        :SetScrollBarOuterMargin( 0 )
    self.content_scroll.HandlePreControlDown = function( self, control, device )
        -- Only scroll the panel if not hovering one of the mod widgets
        local hover = TheGame:FE():GetHoverWidget()
        local mod_widget = hover and hover:GetAncestor( Widget.ModWidget )
        if not mod_widget then return Widget.ScrollPanel.HandlePreControlDown( self, control, device ) end
        return false
    end
    self.content_scroll.HandleGamepadSticks = function( self )
        -- Stop the main scroll panel from reacting to the gamepad sticks, so that we can use that input for the description text
        -- within the mod widgets themselves
        return self
    end
    self.content = self.content_scroll:AddScrollChild( Widget() )

    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnClickClose() end ))
        :SetIcon( global_images.close )
        :LayoutBounds( "left", "above", 60, 60 )
    self.workshop_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.MOD_SCREEN.BROWSE_WORKSHOP", function() self:OnClickBrowseWorkshop() end ))
        :SetIcon( engine.asset.Texture("UI/ic_mods.tex") )
        :LayoutBounds( "after", "top" ):Offset( 20 )
        :SetShown( engine.steam ~= nil )
end

function ModManagerScreen:OnAdded()
    TheGame:GetEvents():ListenForEvent( "mod_changed", self, self.OnModChanged )
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function ModManagerScreen:OnGainTop()
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function ModManagerScreen:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function ModManagerScreen:OnModChanged( event_name, mod )
    if Content.FindMod( mod.id ) then
        for i, child in self.content:Children() do
            if is_instance( child, Widget.ModWidget ) and child.mod == mod then
                child:Refresh()
                return
            end
        end
    end

    -- If we got here, a mod was added or removed.  So just refresh everything.
    self:RefreshMods()
end

function ModManagerScreen:OnScreenResize(w, h)
    ModManagerScreen._base.OnScreenResize(self, w, h)
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function ModManagerScreen:OnControlModeChange( cm, device_id )
    ModManagerScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.bottom_left:Show()
    end

    if cm == CONTROL_MODE.TOUCH then
        self.bottom_left:Show()
    end

    if cm == CONTROL_MODE.GAMEPAD then
        self.bottom_left:Hide()
    end
end

function ModManagerScreen:OnScreenModeChange( sm )
    ModManagerScreen._base.OnScreenModeChange( self, sm )

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Scale contents
    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.content:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Set a small number of columns
    local columns = 2
    local content_width = 0
    local max_content_w = self.screen_w - 120 -- Add some extra spacing for the left edge of the scroll panel
    -- Now keep increasing the number of columns until they don't fit anymore
    repeat
        columns = columns + 1
        self.content:LayoutChildrenInGrid( columns, {h=15, v=15})
        content_width = self.content:GetScaledSize()
    until content_width > max_content_w or columns == 10
    -- Then scale down so it fits
    self.content:LayoutChildrenInGrid( columns-1, {h=15, v=15})
    self.columns = columns - 1

    -- Header height
    local w, h = self.header:GetSize()
    local header_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[sm]
    -- Footer height
    local w, h = self.bottom_left:GetSize()
    local footer_height = ( 60 + h + SPACING.M1 )*LAYOUT_SCALE[sm]

    -- Update containers
    self.widepanel:SetHeight( self.screen_h - header_height - footer_height )
        :LayoutBounds( nil, "below", self.header )
        :Offset( 0, -SPACING.M1 )
    self.content_scroll:SetSize( self.screen_w - SPACING.M1, self.screen_h - header_height - footer_height )
        :LayoutBounds( "left", "center", self.widepanel )
        :Offset( 0, 3 )

    self:RefreshMods()
end

function ModManagerScreen:GetDefaultFocus()
    return self.content.children and #self.content.children>0 and self.content.children[1]
end

function ModManagerScreen:RefreshMods()
	self.content:DestroyAllChildren()

    for i, mod in Content.AllMods() do
        if mod.status == MOD_STATUS.LOCAL or mod.status == MOD_STATUS.WORKSHOP then
            self.content:AddChild( Widget.ModWidget( mod ) )
            if self.enabled_mods[ mod.id ] == nil then
                self.enabled_mods[ mod.id ] = Content.IsModEnabled( mod )
            end
        end
    end

    self.content:LayoutChildrenInGrid( self.columns, {h=15, v=15})
        :LayoutBounds( "left", nil, self.header )
        :LayoutBounds( nil, "top", 0, 0 )

    if self.content:IsEmpty() then
        if not self.empty_text then
            self.empty_text = self:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT, LOC"UI.MOD_SCREEN.EMPTY_TEXT" ) )
                :SetGlyphColour( UICOLOURS.DISABLED )
                :SetTintAlpha( 0.7 )
                :SetAutoSize( 600 )
                :SetWordWrap( true )
                :Bloom( 0.05 )
        end
        self.empty_text:Show()
            :LayoutBounds( "center", "top", self.widepanel )
            :Offset( 0, -SPACING.M05 )
    elseif self.empty_text then
        self.empty_text:Hide()
    end

    -- Position scrollable content
    self.content_scroll:RefreshView()
end

function ModManagerScreen:AskToRestart()
    TheGame:FE():PushScreen( Screen.YesNoPopup( LOC"UI.MOD_SCREEN.RESTART_TITLE", LOC "UI.MOD_SCREEN.RESTART_BODY" ) )
        :SetFn( function(v)
            if v == Screen.YesNoPopup.YES then 
                engine.inst:Reset()
            else
                self:FadeOut()
            end 

        end )    
end

function ModManagerScreen:OnClickBrowseWorkshop()
    engine.steam:ViewURL( "https://steamcommunity.com/app/601840/workshop/" )
end

function ModManagerScreen:OnClickClose()

    for i, mod in Content.AllMods() do
        if mod.status == MOD_STATUS.LOCAL or mod.status == MOD_STATUS.WORKSHOP then
            if self.enabled_mods[ mod.id ] ~= Content.IsModEnabled( mod ) then
                self:AskToRestart()
                return
            end
        end
    end

    self:FadeOut()
end

----------------------------------------------------------------------------------------------------------

local ModWidget = class( "Widget.ModWidget", Widget.Clickable )

ModWidget.CONTROL_MAP = 
{
    {
        control = Controls.Digital.GAMEPAD_A,
        hint = function(self, left, right)
            table.insert(right, loc.format( Content.IsModEnabled( self.mod ) and LOC"UI.MOD_SCREEN.CONTROLS_DISABLE" or LOC"UI.MOD_SCREEN.CONTROLS_ENABLE", Controls.Digital.GAMEPAD_A))
        end,
        fn = function(self)
            self:OnClickEnabled( not Content.IsModEnabled( self.mod ) )
            TheGame:FE():RefreshControlOptions()
            return true
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_X,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.MOD_SCREEN.CONTROLS_VIEW_IN_WORKSHOP", Controls.Digital.GAMEPAD_X))
        end,
        test = function(self)
            return self.mod.workshop ~= nil
        end,
        fn = function(self)
            self:OnClickView()
            return true
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_Y,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.MOD_SCREEN.CONTROLS_UNSUBSCRIBE", Controls.Digital.GAMEPAD_Y))
        end,
        test = function(self)
            return true
        end,
        fn = function(self)
            self:OnClickUnsubscribe()
            return true
        end,
    }
}

function ModWidget:init( mod )
	ModWidget._base.init( self )
    self:DisableAnimateScales()

	self.mod = mod

    -- Background panel
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = MOD_WIDTH + 45*2*self.bg_border_scale, MOD_HEIGHT + 245*self.bg_border_scale + 25*self.bg_border_scale
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/mod_bg_normal.tex") ) )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetNineSliceCoords( 45, 245, 45, 245 )
        :SetSize( nineslice_w, nineslice_h )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( 0x09E1DCff )
        :Bloom( 0.05 )
        :SetPos( 0, 0 )
    self.texture_left_padding = 10
    self.padding = 15
    
    -- Text
    self.text_w = MOD_WIDTH - self.texture_left_padding - self.padding*2
    -- Mod title
    self.title = self:AddChild( Widget.Label("title", FONT_SIZE.ITEM_SUBTITLE, mod.title or mod.id ) )
        :LeftAlign()
        :SetAutoSize( self.text_w )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom(0.1)
    -- Mod description
    self.description_scroll = self:AddChild( Widget.ScrollPanel() )
        :ShowBar( SCROLLBAR.NEVER )
        :SetScrollBarOuterMargin( 0 )
    self.description_scroll.HandleGamepadSticks = function( self )
        if self.hover or self.focus or self.parent.hover or self.parent.focus then Widget.ScrollPanel.HandleGamepadSticks( self ) end
        return self
    end
    self.text = self.description_scroll:AddScrollChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, mod.description ) )
        :LeftAlign()
        :TopAlign()
        :SetAutoSize( self.text_w )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
        :LayoutBounds( "left", "top", self.padding, 0 )
    self.description_scroll:SetSize( self.text_w+self.padding*2, 60 )
        :RefreshView()
        :LayoutBounds( "left", "below", self.title )
        :Offset( -self.padding, 0 )

    -- Mod images
    self.mod_image_frame = self:AddChild( Widget.Image( engine.asset.Texture( "UI/mod_frame.tex" ) ) )
        :SetSize( 110, 110 )
        :SetTintColour( 0x09E1DCff )
        :SetTintAlpha( 0.5 )
        :Bloom( 0.2 )
        :LayoutBounds( "left", "top", self.bg )
        :Offset( 8, -8 )
    self.mod_image = self:AddChild( Widget.Image() )
        :SetSize( 75, 75 )
        :Bloom( 0.05 )
        :LayoutBounds( "center", "center", self.mod_image_frame )
    self.mod_chips_top = self:AddChild( Widget.Image( engine.asset.Texture( "UI/mod_btn_chips.tex" ) ) )
        :SetSize( 110, 15 )
        :SetTintColour( 0x09E1DCff )
        :SetTintAlpha( 0.5 )
        :Bloom( 0.2 )
    self.mod_chips_bottom = self:AddChild( Widget.Image( engine.asset.Texture( "UI/mod_btn_chips.tex" ) ) )
        :SetSize( 110, 15 )
        :SetTintColour( 0x09E1DCff )
        :SetTintAlpha( 0.5 )
        :Bloom( 0.2 )
    local enable_panel_w = MOD_WIDTH-134
    local enable_panel_h = 46
    self.mod_enable_border = 33
    self.mod_enable_gradient_w = 200
    local nineslice_w, nineslice_h = enable_panel_w - self.mod_enable_border*2*self.bg_border_scale, enable_panel_h - self.mod_enable_border*2*self.bg_border_scale
    self.mod_enable_panel = self:AddChild( Widget.Panel( engine.asset.Texture("UI/mod_enable_panel.tex") ) )
        :SetNineSliceCoords( self.mod_enable_border, self.mod_enable_border, self.mod_enable_border, self.mod_enable_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetInnerSize( nineslice_w, nineslice_h )
        :SetTintAlpha( 0.4 )
    self.mod_enable_checkbox = self:AddChild( Widget.CheckBox( FONT_SIZE.BUTTON ) )
        :SetText( LOC"UI.MOD_SCREEN.ENABLED" )
        :SetFn( function() self:OnClickEnabled( not Content.IsModEnabled( self.mod ) ) end )

    -- Mod type
    self.type_label = self:AddChild( Widget.Label( "title", 20 ))
        :CenterH()
        :BottomAlign()
        :SetRegionSize( 64, 20 )
        :SetTintColour( 0x09E1DCff )
        :SetTintAlpha( 0.5 )
        :Bloom( 0.3 )
    if self.mod.status == MOD_STATUS.WORKSHOP then
        self.type_label:SetText( LOC "UI.MOD_SCREEN.WORKSHOP_MOD" )
    elseif self.mod.status == MOD_STATUS.LOCAL then
        self.type_label:SetText( LOC "UI.MOD_SCREEN.LOCAL_MOD" )
    else
        error()
    end
    self.type_label:SetFontSize( self.type_label:TryFitToRegion( 11, 20 ) )
        :LayoutBounds( "center", "above", self.mod_image_frame )
        :Offset( 0, -10 )

    -- Buttons
    self.buttons_container = self:AddChild( Widget() )
    self.update_btn = self.buttons_container:AddChild( Widget.IconButton( LOC"UI.MOD_SCREEN.UPDATE", function() self:OnClickUpdate() end, self.text_w ))
        :SetIcon( engine.asset.Texture( "UI/ic_update.tex" ) )
    self.view_btn = self.buttons_container:AddChild( Widget.IconButton( LOC"UI.MOD_SCREEN.VIEW_IN_WORKSHOP", function() self:OnClickView() end, self.text_w ))
        :SetIcon( engine.asset.Texture( "UI/ic_link.tex" ) )
    self.unsubscribe_btn = self.buttons_container:AddChild( Widget.IconButton( LOC"UI.MOD_SCREEN.UNSUBSCRIBE", function() self:OnClickUnsubscribe() end, self.text_w ))
        :SetIcon( engine.asset.Texture( "UI/ic_unsubscribe.tex" ) )

    self:Refresh()
end

function ModWidget:OnClickEnabled( state )
    Content.SetModEnabled( self.mod, state == true )
end

function ModWidget:OnClickUpdate()
    engine.steam:DownloadItem( self.mod.workshop_id )
end

function ModWidget:OnClickView()
    engine.steam:ViewItem( self.mod.workshop_id )
end

function ModWidget:OnClickUnsubscribe()
    engine.steam:UnsubscribeItem( self.mod.workshop_id )
end

function ModWidget:OnControlModeChange( cm, device_id )
    ModWidget._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.buttons_container:Show()
    end

    if cm == CONTROL_MODE.TOUCH then
        self.buttons_container:Show()
    end

    if cm == CONTROL_MODE.GAMEPAD then
        self.buttons_container:Hide()
    end

    self:Refresh()
end

function ModWidget:Refresh()
    if Content.IsModEnabled( self.mod ) then
        -- Mod is active
        self.bg:SetTexture( engine.asset.Texture("UI/mod_bg_active.tex") )
            :SetTintColour( UICOLOURS.MOD )
        self.mod_image_frame:SetTintColour( UICOLOURS.MOD )
            :SetTintAlpha( 0.9 )
        self.mod_chips_top:SetTintColour( UICOLOURS.MOD )
            :SetTintAlpha( 0.9 )
        self.mod_chips_bottom:SetTintColour( UICOLOURS.MOD )
            :SetTintAlpha( 0.9 )
        self.type_label:SetTintColour( UICOLOURS.MOD )
            :SetTintAlpha( 1 )
            :Bloom( 0.3 )
        self.mod_enable_checkbox:SetText( LOC"UI.MOD_SCREEN.ENABLED" )
            :SetValue( true, true )
    else
        -- Mod is inactive
        self.bg:SetTexture( engine.asset.Texture("UI/mod_bg_normal.tex") )
            :SetTintColour( 0x5B8181ff )
        self.mod_image_frame:SetTintColour( 0x09E1DCff )
            :SetTintAlpha( 0.5 )
        self.mod_chips_top:SetTintColour( 0x09E1DCff )
            :SetTintAlpha( 0.3 )
        self.mod_chips_bottom:SetTintColour( 0x09E1DCff )
            :SetTintAlpha( 0.3 )
        self.type_label:SetTintColour( 0x09E1DCff )
            :SetTintAlpha( 0.5 )
            :Bloom( 0.3 )
        self.mod_enable_checkbox:SetText( LOC"UI.MOD_SCREEN.DISABLED" )
            :SetValue( false, true )
    end

    self.update_btn:SetShown( self.mod.workshop ~= nil and not IsModUpdated( self.mod ) )
    self.view_btn:SetShown( self.mod.workshop ~= nil )
    self.unsubscribe_btn:SetShown( self.mod.workshop ~= nil and IsModSubscribed( self.mod ))

    -- Layout
    self.title:LayoutBounds( "left", "top", self.bg )
        :Offset( self.texture_left_padding + self.padding, -120 )
    self.mod_enable_panel:LayoutBounds( "after", "center", self.mod_image_frame )
        :Offset( 4, 0 )
    self.mod_enable_checkbox:LayoutBounds( "left", "center", self.mod_enable_panel )
            :Offset( 10, 1 )
    self.mod_chips_top:SetScale( 1, 1 )
        :LayoutBounds( "center", "above", self.mod_enable_panel )
        :SetScale( 1, -1 )
    self.mod_chips_bottom:LayoutBounds( "center", "below", self.mod_enable_panel )

    -- Layout buttons
    self.buttons_container:StackChildren( 10 )
        :LayoutBounds( "left", "bottom", self.bg )
    if self.buttons_container:IsShown() then self.buttons_container:Offset( self.texture_left_padding + self.padding, self.padding ) end

    -- Get buttons height to make the description scroll to take the rest of the space
    local title_x, title_y = self.title:GetPos()
    local title_w, title_h = self.title:GetSize()
    local btns_x, btns_y = self.buttons_container:GetPos()
    local btns_w, btns_h = self.buttons_container:GetSize()
    local description_h
    if self.buttons_container:IsShown() then
        -- Allow description to take up the space between the title and the buttons
        description_h = math.abs(( title_y - title_h/2 ) - btns_y) - self.padding/2
    else
        -- Allow description to take up the space between the title and the bottom of the panel
        description_h = math.abs(( title_y - title_h/2 ) - ( btns_y - btns_h )) - self.padding/2
    end
    self.description_scroll:SetSize( self.text_w + self.padding*2, description_h )
        :RefreshView()
        :LayoutBounds( "left", "below", self.title )
        :Offset( -self.padding, -self.padding/4 )
end

function ModWidget:UpdateImage()
    if ( self.hover or self.focus ) then
        if not self.earned then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.bg:Bloom( 0.08 )
        self:ScaleTo( nil, 1.03, 0.2, easing.outQuad )
    else
        self.bg:Bloom( 0.05 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function ModWidget:OnGainFocus()
    ModWidget._base.OnGainFocus(self)
    self:UpdateImage()
end

function ModWidget:OnLoseFocus()
    ModWidget._base.OnLoseFocus(self)
    self:UpdateImage()
end

function ModWidget:OnGainHover()
    ModWidget._base.OnGainHover(self)
    self:UpdateImage()
end

function ModWidget:OnLoseHover()
    ModWidget._base.OnLoseHover(self)
    self:UpdateImage()
end
