require "ui/widgets/installed_grafts"

local MutatorsScreen = class( "Screen.MutatorsScreen", Widget.Screen )

local COLUMNS = 3
local CELL_W = 340
local CELL_H = 115
local CONTAINER_W = CELL_W*COLUMNS + 100
local CONTAINER_H = 600

MutatorsScreen.CONTROL_MAP = Widget.MergeControlMap( Widget.Screen.CONTROL_MAP,
{
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self) 
            self:DoClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL) )
        end,
    },
})

function MutatorsScreen:init( owner, target_graft_type )
    MutatorsScreen._base.init( self )

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.MUTATOR )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Setup character anim display
    self.owner = owner or TheGame:GetGameState():GetPlayerAgent()
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.owner )
    self.character:SetHoverOutlineColour( UICOLOURS.MUTATOR )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "threaten", true )
        end
    end )
    self.character:SetPos( -RES_X*0.3, -350 )
    self.character:Emote( "threaten", true )

    -- Coin display!
    self.mutators_scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( 0 )
        :SetVirtualTopMargin( 100 )
        :SetVirtualBottomMargin( 100 )
        :SetScrollBarOuterMargin( 0 )
        :SetSize( CONTAINER_W, CONTAINER_H )
    self.title = self.mutators_scroll:AddScrollChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( string.upper( loc.format( LOC"UI.MUTATORS_SCREEN.TITLE", self.owner:GetName() ) ) )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :LeftAlign()
        :Bloom(0.15)
    self.mutators_container = self.mutators_scroll:AddScrollChild( Widget() )
    
    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:DoClose() end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    
    self:Refresh()
end

function MutatorsScreen:DoClose()
    AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
    self:FadeOut()
end

function MutatorsScreen:Refresh()

    self.owner = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    self.mutators_container:DestroyAllChildren()

    local mutators = {}
    local grafts = self.owner.graft_owner:GetGrafts( GRAFT_TYPE.MUTATOR, mutators )
    for i, graft in ipairs( grafts ) do
        if not graft:GetDef().hidden then
            local widget = self.mutators_container:AddChild( Widget.MutatorWidget( CELL_W, CELL_H, graft:GetDef() ) )
        end
    end

    self:Layout()

    return self
end

function MutatorsScreen:Layout()
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    local left_offset = RES_X*0.1 -- How much the scroll panel should cross to the left of the screen center
    local left_padding = 100 -- How much space should there be between the left edge of the scroll area and the first mutator (allows space for glow/scaling without clipping)
    self.title:LayoutBounds( "left", "top", 0, 0 )
        :Offset( left_padding, 0 )
    self.mutators_container:LayoutChildrenInGrid( COLUMNS, SPACING.M1 )
        :LayoutBounds( "left", "below", self.title )
        :Offset( -SPACING.M1*0.75, -SPACING.M1/2 )

    -- Get MainOverlay's top bar height
    local overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    local overlay_topbar_height = overlay and overlay:GetTopBarHeight() or 0

    self.mutators_scroll:SetSize( RES_X/2 + left_offset + left_padding, self.screen_h - overlay_topbar_height )
    self.mutators_scroll:SetPos( RES_X/4 - left_offset/2 - left_padding/2, -overlay_topbar_height/2 )

    return self
end

function MutatorsScreen:OnOpen()
    MutatorsScreen._base.OnOpen( self )

    local anim_duration = 0.7
    local horizontal_movement = -90

    -- Prepare for animation by getting current values
    local content_x, content_y = self.content:GetPos()

    -- Hide objects until they're animated in
    self.content:SetTintAlpha(0)

    -- Setup and start animation
    self.fadein_anim = self:RunUpdater(
        Updater.Parallel{
            -- Fade and move in the coin container
            Updater.Ease( function(v) self.content:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) self.content:SetPos(x,y) end, content_x - horizontal_movement, content_y, content_x, content_y, anim_duration, easing.outQuad),
        }
    )

end

function MutatorsScreen:GetDefaultFocus()
    return self.mutators_container and self.mutators_container.children and #self.mutators_container.children>0 and self.mutators_container.children[1]
end

function MutatorsScreen:OnControlModeChange( cm, device_id )
    MutatorsScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.mutators_scroll:ShowBar( SCROLLBAR.IF_NEEDED )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.mutators_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.mutators_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Hide()
    end

end

function MutatorsScreen:OnScreenModeChange( sm )
    MutatorsScreen._base.OnScreenModeChange( self, sm )

    if sm == SCREEN_MODE.MONITOR then
        COLUMNS = 3
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        COLUMNS = 2
    end

    -- self.character:SetPos( -RES_X*0.27*LAYOUT_SCALE[sm], -350 )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.title:SetLayoutScale( LAYOUT_SCALE[sm] )
    for k, v in pairs(self.mutators_container.children) do
        v:SetLayoutScale( LAYOUT_SCALE[sm] )
    end

    self:Layout()
end

function MutatorsScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.mutators_scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.mutators_scroll:ScrollUp()
        return true
    end
end

function MutatorsScreen:HandleControlUp( control, device )
    if control:Has( Controls.Digital.MENU_CANCEL ) then
        AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
        self:FadeOut()
    end

    return MutatorsScreen._base.HandleControlUp( self, control, device )
end

function MutatorsScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function MutatorsScreen:OnOpen()
    MutatorsScreen._base.OnOpen( self )
    self.screen_loop = AUDIO:CreateEventInstance("event:/ui/grafts_menu/open")
    self.screen_loop:Start()
end

function MutatorsScreen:OnClose()
    MutatorsScreen._base.OnClose( self )

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil
end

-------------------------------------------------------------------------
-- A single mutator

local MutatorWidget = class( "Widget.MutatorWidget", Widget )

local ICON_W = 44
local TEXT_LEFT_MARGIN = 68

function MutatorWidget:init( w, h, def, checked )
    MutatorWidget._base.init( self )
    self.features = {}

    self.def = def
    self.graft = GraftInstance( def.id )
    
    self.hitbox = self:AddChild( Widget.SolidBox( w, h, 0x00FFFF00 ) )
    self.contents = self:AddChild( Widget() )
    self.bg = self.contents:AddChild( Widget.Panel( engine.asset.Texture("UI/dropdown_inner.tex" ) ) )
        :SetTintColour( UICOLOURS.MUTATOR )
        :SetTintAlpha( 0 )
        :SetBloom( 0.1 )
    self.glow = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( UICOLOURS.MUTATOR )
        :SetTintAlpha( 0.2 )
        :SetHiddenBoundingBox( true )
        :SetSize( ICON_W*2, ICON_W*2 )
    self.icon = self.contents:AddChild( Widget.Image( def.img ) )
        :SetSize( ICON_W, ICON_W )
        :Bloom( 0.15 )
    self.text_content = self.contents:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, def:GetLocalizedName() ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.ITEM_NAME*LINE_HEIGHT_TITLE_MULT )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.graft:GetDesc( self ) ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )

    self:SetSize( w, h )
    self:SetFocusable( true )

    self:SetToolTipClass( Widget.TooltipGraft )
        :SetToolTip( self.graft )
        :ShowToolTipOnFocus( true )
end

function MutatorWidget:SetSize( w, h )
    self.title:SetAutoSize( w-TEXT_LEFT_MARGIN-12 )
    self.text:SetAutoSize( w-TEXT_LEFT_MARGIN-12 )
    self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    self.hitbox:SetSize( w, h )
    self.bg:SetSize( w, h )
    self.icon:LayoutBounds( "left", "top", self.hitbox ):Offset( 12, -12 )
    self.glow:LayoutBounds( "center", "center", self.icon ):Offset( 0, 0 )
    self.text_content:LayoutBounds( "left", "top", self.hitbox ):Offset( TEXT_LEFT_MARGIN, -8 )
    return self
end

function MutatorWidget:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function MutatorWidget:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

function MutatorWidget:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.bg:AlphaTo( 0.09, 0.1 )
        self.glow:ScaleTo( nil, 1.2, 0.1 )
        self.contents:ScaleTo( nil, 1.05, 0.2 )
    else
        self.bg:AlphaTo( 0, 0.1 )
        self.glow:ScaleTo( nil, 1, 0.1 )
        self.contents:ScaleTo( nil, 1, 0.3 )
    end
end

function MutatorWidget:OnGainFocus()
    MutatorWidget._base.OnGainFocus(self)
    self:UpdateImage()
end

function MutatorWidget:OnLoseFocus()
    MutatorWidget._base.OnLoseFocus(self)
    self:UpdateImage()
end

function MutatorWidget:OnGainHover()
    MutatorWidget._base.OnGainHover(self)
    self:UpdateImage()
end

function MutatorWidget:OnLoseHover()
    MutatorWidget._base.OnLoseHover(self)
    self:UpdateImage()
end
