local DebugUtil = require "debug/debug_util"
require "ui/screens/conversationscreen"
require "ui/widgets/negotiationpanel"
require "ui/screens/mainoverlay"
require "encounter/base_hub"

------------------------------------------------------------------
--

local NegotiationHub = class( "NegotiationHub", BaseHub )
NegotiationHub.can_reload = true

function NegotiationHub:init( encounter, preset )
    NegotiationHub._base.init( self, encounter )
    self.preset = preset
end

function NegotiationHub:RunHub( enc, state )
    TheGame:GetGameState():GetNotifications():SetBannersEnabled( false )

    local preset = self.preset
    local caravan = TheGame:GetGameState():GetCaravan()
    local player = caravan:GetPlayer()
    local resolve, max_resolve = caravan:GetResolve()
    caravan:DeltaResolve( max_resolve )
    player:MoveToLocation( enc:GetLocation() )
    if self.preset.money then
        player:SetMoney( self.preset.money )
    end
    -- player should be recreated, not necessary to clear grafts (which would clear starting grafts erroneously)
    -- player.graft_owner:ClearGrafts()

    for graft_id, count in pairs( preset.starting_grafts ) do
        local graft = GraftInstance( graft_id, nil, player )
        if type(count) ~= "number" then
            count = 1 -- backwards compatability: values used to be 'true'.
        end
        for i = 1, count do
            player.graft_owner:AddGraft( graft )
        end
    end
    
    local params = {
        difficulty = preset.difficulty,
        helpers = {},
        hinders = {},
        cards = {},
        flags = (preset.flags or 0) | NEGOTIATION_FLAGS.NO_LOOT, -- No loot.
        on_start_negotiation = preset.on_start_negotiation,
    }
    for i, v in ipairs( preset.helpers or table.empty ) do
        local id, skin_uuid = table.unpack( v )
        local skin_table = Content.GetCharacterSkin( skin_uuid )
        local agent = Agent( id, skin_table )
        table.insert( params.helpers, agent )
    end
    for i, v in ipairs( preset.hinders or table.empty ) do
        local id, skin_uuid = table.unpack( v )
        local skin_table = Content.GetCharacterSkin( skin_uuid )
        local agent = Agent( id, skin_table )
        table.insert( params.hinders, agent )
    end
    for card_id, count in pairs( preset.starting_cards or table.empty ) do
        for i = 1, count do
            local card = Negotiation.Card( card_id, enc:GetPlayer() )
            table.insert( params.cards, card )
        end
    end

    local agent = enc:GetAgent()
    local start_params = Negotiation.MiniGame.ParamsFromParty( caravan, agent, params, enc )
    self.minigame = Negotiation.MiniGame( start_params, caravan )

    enc:PresentDefault()

    enc:RunNegotiation( self.minigame, true )
end

-------------------------------------------------------------
-- SCREEN IS NOT USED IN GAME, ONLY FOR DEBUGGING USE.

local NegotiationScreen = class( "Screen.NegotiationScreen", Screen.ConversationScreen )

function NegotiationScreen:init( preset_fn )
    NegotiationScreen._base.init( self )

    self.agent_search = {}

    TheGame.gamestate = GameState()

    self.preset_fn = preset_fn
    self:LoadPreset( preset_fn )

    self.plax = Widget.Plax()
    self:AddChild( self.plax, 1 ):SetAnchors( "center", "center" )
end

function NegotiationScreen:GetDefaultFocus()
    return self.negotiation_panel and self.negotiation_panel:GetDefaultFocus()
end

function NegotiationScreen:OnOpen()
    NegotiationScreen._base.OnOpen( self )

    local overlay = self:GetFE():FindScreen( Screen.MainOverlay )
    if not overlay and not self.preset.hide_overlay then
        overlay = Screen.MainOverlay( )
        self:GetFE():PushScreen( overlay )
    end

    local location = self:SetLocationID( self.preset.location_settings or table.empty )
    local encounter = Encounter( self.opponent, location )

    encounter:SetInitialHub( NegotiationHub( encounter, self.preset ) )

    local function OnDone()
        self:GetFE():PopScreen( self )
        collectgarbage()
        TheGame:FE():InsertScreen( Screen.NegotiationScreen( self.preset_fn ) )
    end
    self:StartConvo( encounter, OnDone )

    if self.preset.hide_grafts then
        self.negotiation_panel.grafts:Hide()
    end

    self.plax:PostProcess( true )
end

function NegotiationScreen:DoRender()
    self.plax:RenderToScene()
end

function NegotiationScreen:LoadLocalNegotiation()
    package.loaded[ "local_negotiation" ] = nil

    local env =
    {
        gs = TheGame:GetGameState(),
        caravan = TheGame:GetGameState():GetCaravan(),
        negotiation_defs = require "negotiation/negotiation_defs",
    }
    setmetatable( env, { __index = _G })

    -- DONT use require so that we can pcall without messing up the Profiler stack.
    local ok, result = pcall( loadfile, "scripts/local_negotiation.lua", nil, env )
    if ok and result then
        ok, result = pcall( result )
    end
    if not ok then
        if (result == nil or not result:find( "No such file" )) then
            LOGWARN( tostring(result) )
        else
            print( "Loading local_negotiation.lua:", result )
        end
    else
        return result
    end
end

function NegotiationScreen:SetLocationID( location_settings )
    local location = Location( location_settings.location_id or "GROG_N_DOG" )

    self.plax:Clear()
    self.plax:SetPlaxData( location:GetPlax() )

    for slot_id, t in pairs( location_settings.plax_slots or table.empty ) do
        local id, skin_uuid = table.unpack( t )
        local skin_table = Content.GetCharacterSkin( skin_uuid )
        local agent = Agent( id, skin_table )
        self.plax:FillSlot( agent, slot_id )
    end

    if location_settings.day_phase then
        local idx = table.arrayfind( PHASE_MASK_STRINGS, location_settings.day_phase )
        if idx then
            self.plax:SetDayPhase( 1 << (idx - 1) )
        end
    end

    TheGame:GetGameState():AddLocation( location )
    self.opponent:MoveToLocation( location )

    if self.encounter then
        self.encounter:SetLocation( location )
        self.encounter:PresentDefault()
    end

    return location
end

function NegotiationScreen:LoadPreset( preset_fn )
    local preset

    if preset_fn then
        preset = preset_fn()
    end

    preset = preset or self:LoadLocalNegotiation()
    if preset == nil then
        local data = DebugUtil.GetLocalDebugData("DEBUG_NEGOTIATION")
        local load_ok
        if data then
            local fn, err = loadstring(data)
            if err then
                LOGWARN( "Failed to load preset: %s", err )
            else
                load_ok, preset = pcall( fn )
            end
        end

        if type(preset) == "string" then
            LOGWARN( "Failed to load preset: %s", preset )
            preset = nil
        end
    end

    self.preset = preset or {}

    if self.preset.starting_cards == nil then
        self.preset.starting_cards = 
        {
            fast_talk = 5,
        }
    end
    if self.preset.starting_grafts == nil then
        self.preset.starting_grafts = {}
    end

    if self.preset.opponent_id then
        self.opponent = Agent( table.unpack( self.preset.opponent_id) )
    else
        self.opponent = Agent( "BANDIT_GOON" )
    end

    if self.preset.location_settings == nil then
        self.preset.location_settings = {}
    end

    if self.preset.player_id and Content.GetCharacterDef( self.preset.player_id[1] ) then
        TheGame:GetGameState():AddPlayerAgent( Agent( self.preset.player_id[1] ))

    elseif self.preset.player then
        TheGame:GetGameState():AddPlayerAgent( self.preset.player )
        self.preset.player_id = { self.preset.player:GetContentID() }
        self.preset.player = nil
    else
        self.preset.player_id = { "SAL" }
        TheGame:GetGameState():AddPlayerAgent( Agent( self.preset.player_id[1] ))
    end

    TheGame:GetGameState():GetCaravan():SetMaxResolve( TheGame:GetGameState():GetPlayerAgent().max_resolve or 1 )
    
    if self.preset.prestige then
        TheGame:GetGameState():GetOptions().advancement_level = self.preset.prestige
    end
end

function NegotiationScreen:SavePreset()
    local s = engine.LuaSerializer()
    local data = s:Serialize( self.preset )
    DebugUtil.SetLocalDebugData("DEBUG_NEGOTIATION", data )
    DebugUtil.SaveLocalDebugData()
end

function NegotiationScreen:RenderPresetCards( ui, panel )
    if next(self.preset.starting_cards) == nil then
        ui:TextColored( 0xCCCCCCFF, "No starting cards!" )
    else
        for i, card_id, count in sorted_pairs( self.preset.starting_cards ) do
            local new_count = ui:SliderInt( card_id, count, 0, 9 )
            if new_count and new_count ~= count then
                if new_count == 0 then
                    self.preset.starting_cards[ card_id ] = nil
                else
                    self.preset.starting_cards[ card_id ] = new_count
                end
                self:SavePreset()
            end
        end
    end

    if ui:Button( "Add Card..." ) then
        ui:OpenPopup( "INSERT_CARD" )
    end
    ui:SameLine( nil, 10 )
    if ui:Button( "Clear Cards" ) then
        self.preset.starting_cards = {}
        self:SavePreset()
    end

    ui:SameLine( nil, 10 )
    if ui:Button( "Assign Deck" ) then
        ui:OpenPopup( "SET_POPUP" )
    end
    if ui:BeginPopup( "SET_POPUP" ) then
        for i, deck in pairs( Content.GetNegotiationDecks() ) do
            local txt = string.format( "%s - %s", deck:GetOwner(), deck:GetContentID() )
            if ui:MenuItem( txt ) then
                self.preset.starting_cards = deepcopy( deck:GetCardCounts() )
                self:SavePreset()
            end
        end

        if self.debug_decks == nil then
            self.debug_decks = DebugDecks()
        end
        for i, t in ipairs( self.debug_decks.sorted_decks ) do
            if ui:MenuItem( t.name.."###_IDENTIFIER"..i ) then
                local deck = Deck( t.name )
                deck:AddCardArray( t.negotiation_cards )
                self.preset.starting_cards = deck:GetCardCounts()
                self:SavePreset()
                self.preset.starting_grafts = {}
                for i, graft_id in ipairs( t.grafts ) do
                    self.preset.starting_grafts[ graft_id ] = (self.preset.starting_grafts[ graft_id ] or 0) + 1
                end
            end
            if ui:IsItemHovered() then
                ui:BeginTooltip()
                ui:Text( tostr( t.negotiation_cards ))
                ui:EndTooltip()
            end
        end

        ui:EndPopup()
    end

    if ui:BeginPopup( "INSERT_CARD" ) then
        local function FilterCard( carddef )
            local negotiation_defs = require "negotiation/negotiation_defs"
            return self.preset.starting_cards[ carddef.id ] == nil
        end
        if self.encounter then
            local card_id = Negotiation.Deck.RenderCardPopup( self.preset, ui, panel, self.encounter:GetPlayer(), FilterCard )
            if card_id then
                self.preset.starting_cards[ card_id ] = (self.preset.starting_cards[ card_id ] or 0) + 1
                self:SavePreset()
            end
        end
        ui:EndPopup()
    end
end

function NegotiationScreen:RenderPresetGrafts( ui, panel )
    if next(self.preset.starting_grafts) == nil then
        ui:TextColored( 0xCCCCCCFF, "No starting grafts!" )
    else
        for i, graft_id, count in sorted_pairs( self.preset.starting_grafts ) do
            if type(count) ~= "number" then
                count = 1
            end
            local new_count = ui:SliderInt( graft_id, count, 0, 9 )
            if new_count and new_count ~= count then
                if new_count == 0 then
                    self.preset.starting_grafts[ graft_id ] = nil
                else
                    self.preset.starting_grafts[ graft_id ] = new_count
                end
                self:SavePreset()
            end
        end
    end

    if ui:Button( "Add Graft..." ) then
        ui:OpenPopup( "INSERT_GRAFT" )
    end
    ui:SameLine( nil, 10 )
    if ui:Button( "Clear Grafts" ) then
        table.clear( self.preset.starting_grafts )  
        self:SavePreset()
    end

    if ui:BeginPopup( "INSERT_GRAFT" ) then
        local filter_str = ui:InputText( "Filter", self.filter_str or "" )
        if filter_str and filter_str ~= self.filter_str then
            self.filter_str = filter_str
        end
        ui:Separator()

        local graft_id
        for i, graft_def in ipairs( Content.GetAllGrafts() ) do
            local id = graft_def.id
            if self.preset.starting_grafts[ id ] == nil and (self.filter_str == nil or id:find(self.filter_str, 1, true)) then
                if ui:Selectable( id, nil, ui.DontClosePopups ) then
                    graft_id = id
                    break
                end
            end
        end
        if graft_id then
            self.preset.starting_grafts[ graft_id ] = 1
            self:SavePreset()
        end
        ui:EndPopup()
    end
end

function NegotiationScreen:RenderPresetBystanders( ui, panel )
    ui:PushStyleColor( ui.Style_Button, 0x007700FF )
    for i, v in ipairs( self.preset.helpers or table.empty ) do
        if ui:Button( tostring(v[1]) ) then
            table.remove( self.preset.helpers, i )
            break
        end
        ui:SameLine( nil, 10 )
    end
    ui:PopStyleColor()
    ui:PushStyleColor( ui.Style_Button, 0x770000FF )
    for i, v in ipairs( self.preset.hinders or table.empty ) do
        if ui:Button( tostring(v[1]) ) then
            table.remove( self.preset.hinders, i )
            break
        end
        ui:SameLine( nil, 10 )
    end
    ui:PopStyleColor()

    ui:NewLine()

    if ui:Button( "Add Helper..." ) then
        ui:OpenPopup( "ADD_BYSTANDER" )
        if self.preset.helpers == nil then
            self.preset.helpers = {}
        end
        self.bystanders = self.preset.helpers
    end

    ui:SameLine( nil, 10 )
    if ui:Button( "Add Hinder..." ) then
        ui:OpenPopup( "ADD_BYSTANDER" )
        if self.preset.hinders == nil then
            self.preset.hinders = {}
        end
        self.bystanders = self.preset.hinders
    end

    ui:SameLine( nil, 10 )
    if ui:Button( "Clear" ) then
        self.preset.helpers = nil
        self.preset.hinders = nil
        self:SavePreset()
    end

    if self.bystanders and ui:BeginPopup( "ADD_BYSTANDER" ) then
        local DebugUtil = require "debug/debug_util"
        local id, skin_table = DebugUtil.ShowAgentSearch(self.agent_search, ui)
        if id then
            table.insert( self.bystanders, { id, skin_table and skin_table.uuid } )
            self.bystanders = nil
            self:SavePreset()
        end
        ui:EndPopup()
    end
end

function NegotiationScreen:RenderDebugUI( ui, panel )
    if self.negotiation_panel then
        panel:AppendTable( ui, self.negotiation_panel )
        ui:SameLine( nil, 10 )
        panel:AppendTable( ui, self.negotiation_panel.minigame, "Negotiation" )
    end

    ui:Separator()


    local current_difficulty = self.preset.difficulty or 1
    local diff = ui:SliderInt( "Difficulty", current_difficulty, 1, 5 )
    if diff ~= current_difficulty and diff then
        self.preset.difficulty = diff
        self:SavePreset()
    end

    local current_prestige = math.min( self.preset.prestige or 0, #DEFAULT_ADVANCEMENT )
    local prestige = ui:SliderInt( "Prestige", current_prestige, 0, #DEFAULT_ADVANCEMENT )
    if (prestige ~= current_prestige and prestige) or current_prestige > #DEFAULT_ADVANCEMENT then
        self.preset.prestige = prestige
        self:SavePreset()
    end
    if ui:IsItemHovered() then
        local tt = {}
        for i = 1, current_prestige do
            table.insert( tt, loc.format( "{1%2d}) {2}", i, LOC( DEFAULT_ADVANCEMENT[i].desc )))
        end
        ui:SetTooltip( table.concat( tt, "\n" ))
    end

    ui:Indent( 20 )
    ui:Text( string.format( "Opponent resolve: %d", Negotiation.MiniGame.GetPersuasionRequired( current_difficulty )))
    ui:Unindent( 20 )

    ui:NewLine()

    ui:Columns( 2 )
    if ui:Button( self.preset.player_id and self.preset.player_id[1] or "<NONE>" ) then
        ui:OpenPopup( "PICK_PLAYER" )
    end
    ui:NextColumn()

    if ui:Button( (self.preset.opponent_id and self.preset.opponent_id[1] or "<NONE>") .. "##OPPO" ) then
        ui:OpenPopup( "PICK" )
    end
    ui:NextColumn()

    ui:Columns( 1 )

    ui:Separator()

    self:RenderPresetCards( ui, panel )
    ui:NewLine()

    local new_flags = DebugUtil.RenderBitField( ui, self.preset.flags or 0, NEGOTIATION_FLAGS, NEGOTIATION_FLAGS_STRINGS )
    if new_flags ~= self.preset.flags then
        self.preset.flags =  new_flags
        self:SavePreset()
    end
    
    local money = ui:SliderInt( "Money", self.preset.money or 0, 0, 1000 )
    if money and money ~= self.preset.money then
        self.preset.money = money
        self:SavePreset()
    end
    ui:Spacing()

    self:RenderPresetGrafts( ui, panel )
    ui:NewLine()

    self:RenderPresetBystanders( ui, panel )
    ui:NewLine()

    if DebugUtil.RenderLocationSettings( self.preset.location_settings, self.plax, ui, panel ) then
        self:SetLocationID( self.preset.location_settings )
        self:SavePreset()
    end

    if ui:TreeNode( "HUD Visibility" ) then
        local overlay = self:GetFE():FindScreen( Screen.MainOverlay )
        local is_shown = overlay ~= nil
        if ui:Checkbox( "Hide Overlay", not is_shown ) then
            self.preset.hide_overlay = is_shown
            if self.preset.hide_overlay and overlay then
                self:GetFE():PopScreen( overlay )
            elseif not self.preset.hide_overlay and not overlay then
                overlay = Screen.MainOverlay( )
                self:GetFE():PushScreen( overlay )
                overlay:SetModeOverride( OVERLAY_MODE.NEGOTIATION )
            end
            self:SavePreset()
        end

        local is_shown = self.negotiation_panel.grafts:IsShown()
        if ui:Checkbox( "Hide Grafts", not is_shown ) then
            self.negotiation_panel.grafts:SetShown( not is_shown )
            self.preset.hide_grafts = is_shown
            self:SavePreset()
        end
        ui:TreePop()
    end

    if ui:BeginPopup("PICK_PLAYER", ui.AlwaysAutoResize) then
        local function IsPlayer( def )
            return def.base_def == "PLAYER_BASE"
        end
        local selection, skin_table = DebugUtil.ShowAgentSearch(self.agent_search, ui, IsPlayer )
        if selection then
            self.preset.player_id = { selection, skin_table }
            ui:CloseCurrentPopup()
            self:SavePreset()
        end
        ui:EndPopup()
    end

    if ui:BeginPopup("PICK", ui.AlwaysAutoResize) then
        local function HasNegotiation( def )
            return def.negotiation_data ~= nil
        end
        local selection, skin_table = DebugUtil.ShowAgentSearch(self.agent_search, ui, HasNegotiation )
        if selection then
            self.preset.opponent_id = { selection, skin_table }
            ui:CloseCurrentPopup()
            self:SavePreset()
        end
        ui:EndPopup()
    end
end

function NegotiationScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.MENU_DELETE ) then
        TheGame:FE():PopScreen( self )
        TheGame:GetDebug():ClearPanels()

        package.loaded[ "ui/screens/negotiationscreen" ] = nil
        package.loaded[ "ui/widgets/cardwidget" ] = nil
        package.loaded[ "ui/widgets/cardarrangement" ] = nil
        package.loaded[ "ui/widgets/negotiationcard" ] = nil
        package.loaded[ "ui/widgets/negotiationslot" ] = nil
        package.loaded[ "ui/widgets/negotiationpanel" ] = nil
        package.loaded[ "ui/widgets/intentcontainer" ] = nil
        package.loaded[ "ui/widgets/negotiationresolvebar" ] = nil
        package.loaded[ "ui/widgets/cardspanel" ] = nil
        package.loaded[ "cards/scalar_accumulator" ] = nil
        package.loaded[ "cards/card_engine" ] = nil
        package.loaded[ "cards/card_picker" ] = nil
        package.loaded[ "negotiation/negotiation_behaviour" ] = nil
        package.loaded[ "negotiation/persuasion_accumulator" ] = nil
        package.loaded[ "negotiation/negotiation_modifier" ] = nil

        Content.ReloadConversation()
        require( "ui/screens/negotiationscreen" )

        TheGame:FE():InsertScreen( NegotiationScreen() )
        return true
    end

    return NegotiationScreen._base.HandleControlDown( self, control, device )
end
