require "ui/widgets/screen"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/slider"
require "ui/widgets/menu"
require "ui/widgets/dropdown"
require "ui/widgets/solidbox"
require "ui/widgets/panelbutton"
require "ui/widgets/tabnavigation"
require "ui/widgets/spinner"

require "ui/screens/creditsscreen"

--------------------------------------------------------------------------------
-- Dialog for acessing (loading/saving, depending on save_mode) save games

local OptionScreen = class("Screen.OptionScreen", Widget.Screen)

local PANEL_H = 540
local TABS_W = RES_X*.7
local CONTENT_W = RES_X*.7
local TAB = MakeEnum{ "GAMEPLAY", "GRAPHICS", "AUDIO", "CONTROLS", "OTHER", "MODS" }
local LEFT_OFFSET = 10
local CONTENT_X = -460
local TAB_HEIGHT = 60

local KEY_BINDS =
{
    Controls.Digital.SKIP, -- Tab, right-click
    Controls.Digital.NEXT_LINE, -- Space.
    Controls.Digital.TOGGLE_CONTRACTS,
    Controls.Digital.TOGGLE_NEGOTIATION_DECK,
    Controls.Digital.TOGGLE_BATTLE_DECK,
    Controls.Digital.TOGGLE_GRAFTS,
    Controls.Digital.TOGGLE_COIN,
    Controls.Digital.TOGGLE_BLASTER,
    Controls.Digital.TOGGLE_MUTATORS,
    Controls.Digital.TOGGLE_RELATIONSHIPS,

    Controls.Digital.CARDS_END_TURN,
    Controls.Digital.VIEW_DRAW_DECK,
    Controls.Digital.VIEW_DISCARDS_DECK,
}

-- local GAMEPAD_BINDS =
-- {
--     {button="A", control=Controls.Digital.GAMEPAD_A},
--     {button="A", control=Controls.Digital.MENU_ACCEPT},
--     {button="B", control=Controls.Digital.GAMEPAD_B},
--     {button="B", control=Controls.Digital.MENU_CANCEL},

--     {button="B", control=Controls.Digital.OPEN_MAP},    
--     {button="Y", control=Controls.Digital.SKIP},
--     {button="X", control=Controls.Digital.NEXT_LINE},
-- }

local function FormatPercentage( num )
    return tostring( math.floor( num * 100 ))
end

OptionScreen.CONTROL_MAP = {
    
    {
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },

    {   
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self) self:OnCancel() return true end,
        hint = function(self, left, right) table.insert(right, loc.format(self.dirty and LOC"UI.CONTROLS.CANCEL" or LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL)) end,
    },

    {   
        control = Controls.Digital.MENU_PREV,
        fn = function(self) self.navigation:SetCurrent( self.navigation:GetCurrentIdx() - 1 ) return true end,
    },
    {   
        control = Controls.Digital.MENU_NEXT,
        fn = function(self) self.navigation:SetCurrent( self.navigation:GetCurrentIdx() + 1 ) return true end,
    },

    {   
        control = Controls.Digital.GAMEPAD_LT,
        test = function(self) return self.dirty end,
        fn = function(self) self:Revert() return true end,
        hint = function(self, left, right) table.insert(right, loc.format(LOC"UI.CONTROLS.REVERT", Controls.Digital.GAMEPAD_LT)) end,
    },
    {   
        control = Controls.Digital.GAMEPAD_X,
        test = function(self) return self.dirty end,
        fn = function(self) self:OnApply() return true end,
        hint = function(self, left, right) table.insert(right, loc.format(LOC"UI.CONTROLS.APPLY", Controls.Digital.GAMEPAD_X)) end,
    },


}

function OptionScreen:init(save_mode, cb)
    OptionScreen._base.init(self)

    self.simplified_mode = SPECIAL_BUILD_TYPE >= 1024 -- If true, no tabs are shown, and only select options show on the list

    self.audio_device_options = { {default = true, name = LOC"UI.OPTIONS_SCREEN.SYSTEM_DEFAULT_AUDIO"}}
    
    local audio_devices = engine.sound and engine.sound:GetOutputDevices()
    if audio_devices then    
        for _, audio_device in ipairs(audio_devices) do
            table.insert( self.audio_device_options, {
                default = false,
                name = audio_device.name,
                data = audio_device.id,
            })
        end
    end


    self.is_popup = true
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
    
    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )
    
    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- A container for all the background elements
    self.panel = self:AddChild( Widget.WidePanel(PANEL_H) )

    -- Top navigation tabs
    self.navigation = self:AddChild( Widget.TabNavigation( 200, TAB_HEIGHT, function( tab ) self:NavigationChanged( tab.tab_id ) end ) )
        :SetFocusable( false )
    local tab_gameplay = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_gameplay.tex" ), loc.format( LOC"UI.OPTIONS_SCREEN.TAB_GAMEPLAY" )  ) )
    tab_gameplay.tab_id = TAB.GAMEPLAY

    if not self.simplified_mode then
        local tab_graphics = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_graphics.tex" ), loc.format( LOC"UI.OPTIONS_SCREEN.TAB_GRAPHICS" ) ) )
        tab_graphics.tab_id = TAB.GRAPHICS
    end
    local tab_audio = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_audio.tex" ), loc.format( LOC"UI.OPTIONS_SCREEN.TAB_AUDIO" ) ) )
    tab_audio.tab_id = TAB.AUDIO

    if not self.simplified_mode then
        local tab_controls = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_controls.tex" ), LOC"UI.OPTIONS_SCREEN.TAB_CONTROLS" ))
        tab_controls.tab_id = TAB.CONTROLS

        local tab_other = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_other.tex" ), loc.format( LOC"UI.OPTIONS_SCREEN.TAB_OTHER") ) )
        tab_other.tab_id = TAB.OTHER

        local tab_mods = self.navigation:AddTab( Widget.IconTextTab( engine.asset.Texture( "UI/ic_options_mods.tex" ), loc.format( LOC"UI.OPTIONS_SCREEN.TAB_MODS") ) )
        tab_mods.tab_id = TAB.MODS
    end

    self.navigation:SetTotalWidth( TABS_W )
    -- Navigation tab labels
    self.navigation_label_left = self:AddChild( Widget.Label( "title", 30, LOC"UI.OPTIONS_SCREEN.CONTROLS_LABEL_TAB_LEFT" ) )
        :Bloom( 0.1 )
    self.navigation_label_right = self:AddChild( Widget.Label( "title", 30, LOC"UI.OPTIONS_SCREEN.CONTROLS_LABEL_TAB_RIGHT" ) )
        :Bloom( 0.1 )
    
    -- Move the panel over the tabs
    self.panel:SendToFront()

    self.gameplay_scroll_panel, self.gameplay_content = self:CreateGameplayPanel()

    self.graphics_scroll_panel, self.graphics_content = self:CreateGraphicsPanel()

    self.audio_scroll_panel, self.audio_content = self:CreateAudioPanel()

    self.controls_scroll_panel, self.controls_content = self:CreateControlsPanel()

    self.other_scroll_panel, self.other_content = self:CreateOtherPanel()

    self.mod_scroll_panel, self.mod_content = self:CreateModPanel()


    -- Our fancy character anim!
    self.anim_root = self:AddChild(Widget()):SetHiddenBoundingBox( true )

    --- Left buttons
    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:OnCancel() end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    --- Right buttons
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    --- Center buttons
    self.bottom_center = self:AddChild( Widget() ):SetAnchors( "center", "bottom" )
    self.apply_button = self.bottom_center:AddChild( Widget.IconButton( LOC"UI.OPTIONS_SCREEN.APPLY", function() self:OnApply() end ) )
    self.apply_button:SetIcon( engine.asset.Texture( "UI/ic_options_apply.tex" ) ):Disable()
    self.apply_button:LayoutBounds( "right", "above", -10, 60 )
    self.revert_button = self.bottom_center:AddChild( Widget.IconButton( LOC"UI.OPTIONS_SCREEN.REVERT", function() self:Revert() end ) )
    self.revert_button:SetIcon( engine.asset.Texture( "UI/ic_options_revert.tex" ) ):Disable()
    self.revert_button:LayoutBounds( "left", "above", 10, 60 )

    -- Backup settings for the revert button
    self.control_settings =
    {
        key_binds = shallowcopy( TheGame:GetInputSettings():GetSetting( "key_binds" )),
    }

    -- Get values
    self:SyncValues()

    self:Layout()
    self:RefreshBigCharacter()

    self.navigation:SetCurrent( 1 )

    -- Play screen sound
    self.screen_loop = AUDIO:CreateEventInstance( "event:/ui/options_menu/open" )
    self.screen_loop:Start()

    self:RefreshTwitchMode()
end

function OptionScreen:CreateGraphicsPanel()

    self.option_resolution = {
        title = LOC"UI.OPTIONS_SCREEN.RESOLUTION",
        widget_name = "FS_RES",
        show_in_simplified_mode = false,
        spinner = true,
        values = {
            { name="1280x720", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=1280, h=720} },
            { name="1440x810", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=1440, h=810} },
            { name="1600x900", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=1600, h=900} },
            { name="1760x990", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=1760, h=990} },
            { name="1920x1080", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=1920, h=1080} },
            { name="2240x1260", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=2240, h=1260} },
            { name="2560x1440", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=2560, h=1440} },
            { name="2880x1620", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=2880, h=1620} },
            { name="3200x1800", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=3200, h=1800} },
            { name="3840x2160", desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=3840, h=2160} },
        },
        fn = function(...)
            self:OnSelectResolution(...)
        end
    }

    local OPTIONS_GRAPHICS = {
        {
            title = LOC"UI.OPTIONS_SCREEN.FULLSCREEN",
            widget_name = "WINDOW_MODE",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.FULLSCREEN_ON", desc=LOC"UI.OPTIONS_SCREEN.FULLSCREEN_ON_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.FULLSCREEN_OFF", desc=LOC"UI.OPTIONS_SCREEN.FULLSCREEN_OFF_DESC", data=0 },
            },
            fn = function(...)
                self:OnSetWindowMode(...)
            end
        },
        self.option_resolution,
        {
            title = LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE",
            widget_name = "LAYOUTSCALE_MODE",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_MONITOR", desc=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_MONITOR_DESC", data=SCREEN_MODE.MONITOR },
                { name=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_TV", desc=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_TV_DESC", data=SCREEN_MODE.TV },
                { name=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_SMALL", desc=LOC"UI.OPTIONS_SCREEN.LAYOUTSCALE_SMALL_DESC", data=SCREEN_MODE.SMALL },
            },
            fn = function(...)
                self:MakeDirty()
            end
        }, 
        {
            title = LOC"UI.OPTIONS_SCREEN.CURSOR",
            widget_name = "CURSOR",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_1", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_1_DESC", data="CURSOR_1" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_2", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_2_DESC", data="CURSOR_2" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_3", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_3_DESC", data="CURSOR_3" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_4", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_4_DESC", data="CURSOR_4" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_5", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_5_DESC", data="CURSOR_5" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_6", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_6_DESC", data="CURSOR_6" },
                { name=LOC"UI.OPTIONS_SCREEN.CURSOR_SYSTEM", desc=LOC"UI.OPTIONS_SCREEN.CURSOR_SYSTEM_DESC", data="SYSTEM" },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },


        {
            title = LOC"UI.OPTIONS_SCREEN.FRAMELIMITER",
            widget_name = "FRAMELIMITER_MODE",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_OFF", desc=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_OFF_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_60HZ", desc=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_60HZ_DESC", data=2 },
                { name=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_30HZ", desc=LOC"UI.OPTIONS_SCREEN.FRAMELIMITER_30HZ_DESC", data=3 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },        
        {
            title = LOC"UI.OPTIONS_SCREEN.COLOURBLIND",
            widget_name = "COLOURBLIND_MODE",
            show_in_simplified_mode = true,
            colourblind = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_DEFAULT", desc=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_DEFAULT_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_PROTANOPIA", desc=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_PROTANOPIA_DESC", data=2 },
                { name=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_DEUTERANOPIA", desc=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_DEUTERANOPIA_DESC", data=3 },
                { name=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TRITANOPIA", desc=LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TRITANOPIA_DESC", data=4 },
            },
            fn = function( val )
                self:OnSelectColorblindMode( val )
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.BLOOM",
            widget_name = "BLOOM_CHECK",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.BLOOM_ON", desc=LOC"UI.OPTIONS_SCREEN.BLOOM_ON_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.BLOOM_OFF", desc=LOC"UI.OPTIONS_SCREEN.BLOOM_OFF_DESC", data=0 },
            },
            fn = function(...)
                self:OnSetBloom(...)
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.DETAILLEVEL",
            widget_name = "LOD",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_HIGH", desc=LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_HIGH_DESC", data=4 },
                { name=LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_LOW", desc=LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_LOW_DESC", data=1 },
            },
            fn = function(...)
                self:OnSelectLOD(...)
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.SCREEN_SHAKE",
            widget_name = "SCREEN_SHAKE",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_SHAKE_ON", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_SHAKE_ON_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_SHAKE_OFF", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_SHAKE_OFF_DESC", data=0 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.SCREEN_FLASH",
            widget_name = "SCREEN_FLASH",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_FLASH_ON", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_FLASH_ON_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_FLASH_OFF", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_FLASH_OFF_DESC", data=0 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
    }

    if false and not CheckAnyBits( SPECIAL_BUILD_TYPE, 64 | 128 | 256 ) then
        table.insert( OPTIONS_GRAPHICS,
        {
            title = LOC"UI.OPTIONS_SCREEN.SCREEN_SWAY",
            widget_name = "SCREEN_SWAY",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_SWAY_ON", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_SWAY_ON_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.SCREEN_SWAY_OFF", desc=LOC"UI.OPTIONS_SCREEN.SCREEN_SWAY_OFF_DESC", data=0 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        } )
    end


    self.MAXY_OPTIONS = {
        { name="4320p", desc=loc.format( LOC"UI.OPTIONS_SCREEN.MAXY_DESC", "4320" ), data=4320 },
        { name="2160p", desc=loc.format( LOC"UI.OPTIONS_SCREEN.MAXY_DESC", "2160" ), data=2160 },
        { name="1440p", desc=loc.format( LOC"UI.OPTIONS_SCREEN.MAXY_DESC", "1440" ), data=1440 },
        { name="1080p", desc=loc.format( LOC"UI.OPTIONS_SCREEN.MAXY_DESC", "1080" ), data=1080 },
        { name="720p", desc=loc.format( LOC"UI.OPTIONS_SCREEN.MAXY_DESC", "720" ), data=720 },
    }

    if engine.inst:ShouldShowMaxYOption() then
        table.insert( OPTIONS_GRAPHICS, 3, {
            title = LOC"UI.OPTIONS_SCREEN.MAXY",
            widget_name = "MAXY",
            show_in_simplified_mode = false,
            spinner = true,
            values = self.MAXY_OPTIONS,
            fn = function(...)
                self:OnSelectMaxYRes(...)
            end
        } )
    end

    -- Add our graphics panel
    local graphics_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local graphics_content = graphics_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )

    if not self.simplified_mode then
        for k, option in ipairs( OPTIONS_GRAPHICS ) do
            local spinner = graphics_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, option.title, k, option ) )
            spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
            spinner:SetOptions( option.values )
            spinner:SetName( option.widget_name or "" )
        end

    else
        for k, option in ipairs( OPTIONS_GRAPHICS ) do
            local spinner = self.gameplay_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, option.title, k, option ) )
            spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
            spinner:SetOptions( option.values )
            spinner:SetName( option.widget_name or "" )
            spinner:SetShown( option.show_in_simplified_mode )
        end
    end

    graphics_scroll_panel:SetVirtualMargin( 15 )
    graphics_scroll_panel:RefreshView()
    graphics_content:SetPos( LEFT_OFFSET, 0 )

    return graphics_scroll_panel, graphics_content
end

function OptionScreen:CreateAudioPanel()

    local OPTIONS_AUDIO = {
        {
            title = LOC"UI.OPTIONS_SCREEN.AUDIO_DEVICE",
            widget_name = "AUDIO_DEVICE",
            spinner = true,
            show_in_simplified_mode = false,
            values = self.audio_device_options,
            fn = function(v)
                self:MakeDirty()
            end,
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.FORCE_MONO",
            widget_name = "FORCE_MONO",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.FORCE_MONO_OFF", desc=LOC"UI.OPTIONS_SCREEN.FORCE_MONO_OFF_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.FORCE_MONO_ON", desc=LOC"UI.OPTIONS_SCREEN.FORCE_MONO_ON_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.MUTE_ON_LOSE_FOCUS",
            widget_name = "MUTE_ON_LOSE_FOCUS",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.MUTE_ON_LOSE_FOCUS_OFF", desc=LOC"UI.OPTIONS_SCREEN.MUTE_ON_LOSE_FOCUS_OFF_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.MUTE_ON_LOSE_FOCUS_ON", desc=LOC"UI.OPTIONS_SCREEN.MUTE_ON_LOSE_FOCUS_ON_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.MASTER",
            widget_name = "AUDIO_MASTER",
            slider = true,
            values = {
                display_format = FormatPercentage,
                range_min = 0,
                range_max = 1,
                step_size = 0.05,
                desc = LOC"UI.OPTIONS_SCREEN.MASTER_DESC",
            },
            fn = function(v)
                self:VolumeChange("master_volume",v)
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.MUSIC",
            widget_name = "AUDIO_MUSIC",
            slider = true,
            values = {
                display_format = FormatPercentage,
                range_min = 0,
                range_max = 1,
                step_size = 0.05,
                desc = LOC"UI.OPTIONS_SCREEN.MUSIC_DESC",
            },
            fn = function(v)
                self:VolumeChange("music_volume",v)
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.VOICE",
            widget_name = "AUDIO_VO",
            slider = true,
            values = {
                display_format = FormatPercentage,
                range_min = 0,
                range_max = 1,
                step_size = 0.05,
                desc = LOC"UI.OPTIONS_SCREEN.VOICE_DESC",
            },
            fn = function(v)
                self:VolumeChange("vo_volume",v)
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.FX",
            widget_name = "AUDIO_FX",
            slider = true,
            values = {
                display_format = FormatPercentage,
                range_min = 0,
                range_max = 1,
                step_size = 0.05,
                desc = LOC"UI.OPTIONS_SCREEN.FX_DESC",
            },
            fn = function(v)
                self:VolumeChange("fx_volume",v)
            end,
            dropdown = {
                {
                    title = LOC"UI.OPTIONS_SCREEN.FX_AMBIENCE",
                    widget_name = "AUDIO_AMBIENCE",
                    slider = true,
                    values = {
                        display_format = FormatPercentage,
                        range_min = 0,
                        range_max = 1,
                        step_size = 0.05,
                        desc = LOC"UI.OPTIONS_SCREEN.FX_AMBIENCE_DESC",
                    },
                    fn = function(v)
                        self:VolumeChange("ambience_volume",v)
                    end
                },
                {
                    title = LOC"UI.OPTIONS_SCREEN.FX_ACTIONS",
                    widget_name = "AUDIO_ACTIONS",
                    slider = true,
                    values = {
                        display_format = FormatPercentage,
                        range_min = 0,
                        range_max = 1,
                        step_size = 0.05,
                        desc = LOC"UI.OPTIONS_SCREEN.FX_ACTIONS_DESC",
                    },
                    fn = function(v)
                        self:VolumeChange("actions_volume",v)
                    end
                },
                {
                    title = LOC"UI.OPTIONS_SCREEN.FX_CARDS",
                    widget_name = "AUDIO_CARDS",
                    slider = true,
                    values = {
                        display_format = FormatPercentage,
                        range_min = 0,
                        range_max = 1,
                        step_size = 0.05,
                        desc = LOC"UI.OPTIONS_SCREEN.FX_CARDS_DESC",
                    },
                    fn = function(v)
                        self:VolumeChange("cards_volume",v)
                    end
                },
                {
                    title = LOC"UI.OPTIONS_SCREEN.FX_UI",
                    widget_name = "AUDIO_UI",
                    slider = true,
                    values = {
                        display_format = FormatPercentage,
                        range_min = 0,
                        range_max = 1,
                        step_size = 0.05,
                        desc = LOC"UI.OPTIONS_SCREEN.FX_UI_DESC",
                    },
                    fn = function(v)
                        self:VolumeChange("ui_volume",v)
                    end
                },
            }
        },
    }

        -- Add our audio panel
    local audio_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local audio_content = audio_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )
    
    -- And its spinners
    for k, option in ipairs( OPTIONS_AUDIO ) do
        local spinner = audio_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, option.title, k, option ) )
        spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
        spinner:SetOptions( option.values )
        spinner:SetName( option.widget_name or "" )
        if option.dropdown then
            -- This row will have hidden rows in it
            local rows = {}
            local dropdown_row_width_delta = -50
            for j, dropdown_option in ipairs( option.dropdown ) do
                local dropdown_spinner = audio_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40+dropdown_row_width_delta, dropdown_option.title, k, dropdown_option ) )
                dropdown_spinner:LayoutBounds( "right", "below" ):Offset( 0, -15 )
                dropdown_spinner:SetOptions( dropdown_option.values )
                dropdown_spinner:SetName( dropdown_option.widget_name or "" )
                table.insert( rows, dropdown_spinner )
            end
            spinner:SetDropdownRows( rows )
            spinner:SetDropdownFn( function( expanded )
                -- Re-layout the scroll panel when things get added or removed
                audio_scroll_panel:RefreshView()
            end )
        end
    end
    audio_scroll_panel:SetVirtualMargin( 15 )
    audio_scroll_panel:RefreshView()
    audio_content:SetPos( LEFT_OFFSET, 0 )

    return audio_scroll_panel, audio_content
end

function OptionScreen:CreateControlsPanel()

    -- Create UI options from current settings.
    local input = TheGame:GetInput()
    local OPTIONS_CONTROLS = {
        {
            name = LOC "UI.OPTIONS_SCREEN.CONTROL_REVERT_BINDINGS",
            desc = LOC "UI.OPTIONS_SCREEN.CONTROL_REVERT_BINDINGS_DESC",
            button = true,
            widget_name = "REVERT_BINDINGS",
            disabled = function()
                return next(self.control_settings.key_binds) == nil 
            end,
            on_click = function()
                if next(self.control_settings.key_binds) ~= nil then
                    table.clear( self.control_settings.key_binds )
                    self:MakeDirty()
                    self:SyncControlValues()
                end
            end
        }
    }
    for i, control in ipairs( KEY_BINDS ) do
        local control_name = table.find( Controls.Digital, control )
        local lockey = string.format( "CONTROLS.DIGITAL.%s", control_name )
        local option =
        {
            name = LOC( lockey ),
            desc = "",
            button = true,
            widget_name = control_name,
        }
        option.on_click = function()
            local binding = self.control_settings.key_binds[ control_name ]
            local default_bindings = input:GetDefaultKeyBindings( control )
            local title = loc.format( LOC "UI.DIALOGS.REBIND_TITLE", control.displayname )

            local screen = Screen.RebindInputPopup( title, binding, function(screen)
                local rebinding = screen:GetRebinding()
                if rebinding then
                    -- check if we are just rebinding to a default value
                    for i, binding in ipairs( default_bindings ) do
                        if input:IsEqualBinding( binding, rebinding ) then
                            rebinding = nil
                            break
                        end
                    end
                end
                if rebinding ~= binding then
                    if rebinding then
                        self.control_settings.key_binds[ control_name ] = rebinding
                    else
                        self.control_settings.key_binds[ control_name ] = nil
                    end
                    self:MakeDirty()
                    self:SyncControlValues()
                end
            end )
            TheGame:FE():PushScreen( screen )
        end
        table.insert( OPTIONS_CONTROLS, option )
    end

    local controls_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local controls_content = controls_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )

    if not self.simplified_mode then
        for k, option in ipairs( OPTIONS_CONTROLS ) do
            local button = controls_content:AddChild( Widget.OptionsScreenButton( CONTENT_W-40, option.name, nil, option.desc, option.on_click ))
            button:LayoutBounds( "left", "below" ):Offset( 0, -15 )
            button:SetName( option.widget_name )
            button.option = option
        end
    end
    controls_scroll_panel:SetVirtualMargin( 15 )
    controls_scroll_panel:RefreshView()
    controls_content:SetPos( LEFT_OFFSET, 0 )

    return controls_scroll_panel, controls_content
end

function OptionScreen:CreateOtherPanel()
    -- Generate language options
    local language_options = {}

    local translator = require("util/loc_translator").Translator()
    local languages = Content.GetLocalizations()

    for language_id, language_settings in pairs(languages) do
        table.insert( language_options, {
            name = language_settings.name,
            desc = language_settings.incomplete and LOC"UI.OPTIONS_SCREEN.LANGUAGES_INCOMPLETE_DESC" or LOC"UI.OPTIONS_SCREEN.LANGUAGES_DESC",
            data = language_id,
        })
    end
    -- Sort languages
    table.stable_sort( language_options, function( option1, option2 ) return option1.data < option2.data end )

    local OPTIONS_OTHER = {
        {
            title = LOC"UI.OPTIONS_SCREEN.METRICS",
            widget_name = "METRICS",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.METRICS_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.METRICS_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.METRICS_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.METRICS_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.LANGUAGES",
            widget_name = "LANGUAGES",
            show_in_simplified_mode = true,
            spinner = true,
            values = language_options,
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            name = LOC"UI.OPTIONS_SCREEN.CREDITS",
            title = LOC"UI.OPTIONS_SCREEN.CREDITS_TITLE",
            desc = LOC"UI.OPTIONS_SCREEN.CREDITS_DESC",
            button = true,
            show_in_simplified_mode = true,

            on_click = function()
                TheGame:FE():PushScreen( Screen.CreditsScreen() )
            end
        },
        {
            name = LOC"UI.OPTIONS_SCREEN.RESET_PROFILE",
            title = LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_TITLE",
            desc = TheGame:GetGameState() and LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_DESC_LOCKED" or LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_DESC",
            button = true,
            show_in_simplified_mode = true,

            on_click = function()
                local main_menu = TheGame:FE():FindScreen( Screen.MainMenu )
                if main_menu then
                    TheGame:FE():PushScreen(Screen.YesNoPopup(LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_TITLE", LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_BODY")
                        :SetFn(function(res) 
                            if res == Screen.YesNoPopup.YES then 
                                TheGame:GetGameProfile():Reset()
                                TheGame:GetGameStats():Reset()
                                TheGame.run_history:Reset()

                                local function OnDelete( success )
                                    if not success then
                                        print( "Failed to delete save game:", success )
                                    end
                                end
                                local function DeleteAll( t )
                                    for name, v in pairs(t) do
                                        TheGame:DeleteSave( v.filename, OnDelete )
                                    end
                                end
                                TheGame:ListSaveGames( DeleteAll )

                                local main_menu = TheGame:FE():FindScreen( Screen.MainMenu )
                                if main_menu then
                                    main_menu:SetCurrentSaveDirty()
                                end

                                TheGame:FE():PushScreen(Screen.InfoPopup( LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_CONFIRM_TITLE", LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_CONFIRM_BODY"))
                            end
                        end ))
                else
                    UIHelpers.InfoPopup( LOC"UI.OPTIONS_SCREEN.RESET_PROFILE", LOC"UI.OPTIONS_SCREEN.RESET_PROFILE_DISABLED" )
                end

            end
        },
        {
            name = LOC"UI.OPTIONS_SCREEN.RESET_SETTINGS",
            title = LOC"UI.OPTIONS_SCREEN.RESET_SETTINGS_TITLE",
            desc = LOC"UI.OPTIONS_SCREEN.RESET_SETTINGS_DESC",
            button = true,
            show_in_simplified_mode = true,

            on_click = function()
                self:Reset()
            end
        }

    }

    local other_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local other_content = other_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )

    for k, option in ipairs( OPTIONS_OTHER ) do
        local parent
        if not self.simplified_mode then
            parent = other_content
        else
            parent = self.gameplay_content
        end
        if option.button then
            local button = parent:AddChild( Widget.OptionsScreenButton( CONTENT_W-40, option.name, option.title, option.desc, option.on_click ))
            button:LayoutBounds( "left", "below" ):Offset( 0, -15 )
            button:SetName( option.widget_name )
            button:SetShown( option.show_in_simplified_mode or not self.simplified_mode )
        else
            local spinner = parent:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, option.title, k, option ) )
            spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
            spinner:SetOptions( option.values )
            spinner:SetName( option.widget_name or "" )
            spinner:SetShown( option.show_in_simplified_mode or not self.simplified_mode )
        end
    end

    other_scroll_panel:SetVirtualMargin( 15 )
    other_scroll_panel:RefreshView()
    other_content:SetPos( LEFT_OFFSET, 0 )

    return other_scroll_panel, other_content
end

function OptionScreen:CreateModPanel()
    local mod_scroll_panel = self:AddChild( Widget.ScrollPanel() )
            :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local mod_content = mod_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )

    for i, id in pairs( Content.GetInstalledMods() ) do
        local mod = Content.FindMod( id )
        for k, option in ipairs( mod.mod_options or table.empty ) do
            option.widget_name = string.format( "MODOPTION_%s_%s", id, option.key )
            option.fn = function( value )
                self:MakeDirty()
            end

            local title = loc.format( "{1} ({2})", option.title, mod.title )
            if option.spinner or option.slider then
                local spinner = mod_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, title, k, option ) )
                spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
                spinner:SetOptions( option.values )
                spinner:SetName( option.widget_name )
            else
                local button = mod_content:AddChild( Widget.OptionsScreenButton( CONTENT_W-40, title, nil, option.desc, option.on_click ))
                button:LayoutBounds( "left", "below" ):Offset( 0, -15 )
                button:SetName( option.widget_name )
            end
        end
    end
    mod_scroll_panel:SetVirtualMargin( 15 )
    mod_scroll_panel:RefreshView()
    mod_content:SetPos( LEFT_OFFSET, 0 )

    return mod_scroll_panel, mod_content
end

function OptionScreen:CreateGameplayPanel()
    local OPTIONS_GAMEPLAY = {
        {
            title = LOC"UI.OPTIONS_SCREEN.AUTO_SKIP",
            widget_name = "AUTO_SKIP",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.AUTO_SKIP_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.AUTO_SKIP_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.AUTO_SKIP_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.AUTO_SKIP_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.FAST_MODE",
            widget_name = "FAST_MODE",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.FAST_MODE_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.FAST_MODE_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.FAST_MODE_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.FAST_MODE_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.NEGOTIATION_SPEED",
            widget_name = "NEGOTIATION_SPEED",
            show_in_simplified_mode = false,
            slider = true,
            values = {
                display_format = function( num ) return loc.format( "{1%.1f}x", num ) end,
                range_min = 0.5,
                range_max = 4.0,
                step_size = 0.1,
                desc = LOC"UI.OPTIONS_SCREEN.NEGOTIATION_SPEED_DESC",
            },
            fn = function(v)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.BATTLE_SPEED",
            widget_name = "BATTLE_SPEED",
            show_in_simplified_mode = false,
            slider = true,
            values = {
                display_format = function( num ) return loc.format( "{1%.1f}x", num ) end,
                range_min = 0.5,
                range_max = 4.0,
                step_size = 0.1,
                desc = LOC"UI.OPTIONS_SCREEN.BATTLE_SPEED_DESC",
            },
            fn = function(v)
                self:MakeDirty()
            end
        },
        --[[{
            title = LOC"UI.OPTIONS_SCREEN.SKIP_MOVIES",
            widget_name = "SKIP_MOVIES",
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.SKIP_MOVIES_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.SKIP_MOVIES_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.SKIP_MOVIES_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.SKIP_MOVIES_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },--]]
        {
            title = LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT",
            widget_name = "CONFIRM_SELECT",
            show_in_simplified_mode = false,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.CONFIRM_ENDTURN",
            widget_name = "CONFIRM_ENDTURN",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.CONFIRM_ENDTURN_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.CONFIRM_ENDTURN_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.CONFIRM_ENDTURN_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.CONFIRM_ENDTURN_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:MakeDirty()
            end
        },
        {
            title = LOC"UI.OPTIONS_SCREEN.STREAMER_MODE",
            widget_name = "STREAMER_MODE",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_ENABLED_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_ENABLED_RIGHT", desc=LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_ENABLED_DESC_RIGHT", data=2 },
            },
            fn = function(...)
                self:RefreshStreamerMode( true )
                self:MakeDirty()
            end
        },
    }
    if BRANCH == "dev" then
        table.insert( OPTIONS_GAMEPLAY, {
            title = LOC"UI.OPTIONS_SCREEN.TWITCH_MODE",
            widget_name = "TWITCH_MODE",
            show_in_simplified_mode = true,
            twitch_refresh_fn = function() self:ConnectToTwitch() end,
            twitch_refresh_tt = LOC"UI.TWITCH.OPTIONS_REFRESH_CONNECTION_TT",
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.TWITCH_MODE_DISABLED", desc=LOC"UI.OPTIONS_SCREEN.TWITCH_MODE_DISABLED_DESC", data=0 },
                { name=LOC"UI.OPTIONS_SCREEN.TWITCH_MODE_ENABLED", desc=LOC"UI.OPTIONS_SCREEN.TWITCH_MODE_ENABLED_DESC", data=1 },
            },
            fn = function(...)
                self:RefreshTwitchMode()
                self:MakeDirty()
            end
        })
    end


    local gameplay_scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
        :LayoutBounds( "center", "center", self.loot_bg )
    local gameplay_content = gameplay_scroll_panel:AddScrollChild( Widget.OptionsScreenContentWidget() )
    
    -- And its spinners
    if self.simplified_mode then
        local battery_saver_option = {
            title = LOC"UI.OPTIONS_SCREEN.POWERSAVER",
            widget_name = "POWERSAVER_MODE",
            show_in_simplified_mode = true,
            spinner = true,
            values = {
                { name=LOC"UI.OPTIONS_SCREEN.POWERSAVER_OFF", desc=LOC"UI.OPTIONS_SCREEN.POWERSAVER_OFF_DESC", data=1 },
                { name=LOC"UI.OPTIONS_SCREEN.POWERSAVER_ON", desc=LOC"UI.OPTIONS_SCREEN.POWERSAVER_ON_DESC", data=0 },
            },
            fn = function(...)
                self:OnSetPowersaverMode(...)
            end
        }
        table.insert( OPTIONS_GAMEPLAY, 1, battery_saver_option )
    end
    for k, option in ipairs( OPTIONS_GAMEPLAY ) do
        local spinner = gameplay_content:AddChild( Widget.OptionsScreenSpinner( CONTENT_W-40, option.title, k, option ) )
        spinner:LayoutBounds( "left", "below" ):Offset( 0, -15 )
        spinner:SetOptions( option.values )
        spinner:SetName( option.widget_name or "" )
        spinner:SetShown( ( self.simplified_mode and option.show_in_simplified_mode ) or not self.simplified_mode )
    end

    gameplay_scroll_panel:SetVirtualMargin( 15 )
    gameplay_scroll_panel:RefreshView()
    gameplay_content:SetPos( LEFT_OFFSET, 0 )

    return gameplay_scroll_panel, gameplay_content
end


function OptionScreen:OnClose()
    OptionScreen._base.OnClose(self)

    -- Stop the screen sound
    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil

    if engine.sound then
        engine.sound:ClearDevicesDirty()
    end

end

function OptionScreen:GetDefaultFocus()
    if self.current_tab == TAB.GRAPHICS then return self:FirstChild( self.graphics_content ) end
    if self.current_tab == TAB.AUDIO then return self:FirstChild( self.audio_content ) end
    if self.current_tab == TAB.OTHER then return self:FirstChild( self.other_content ) end
    if self.current_tab == TAB.GAMEPLAY then return self:FirstChild( self.gameplay_content ) end
    if self.current_tab == TAB.MODS then return self:FirstChild( self.mod_content ) end
    if self.current_tab == TAB.CONTROLS then return self:FirstChild( self.controls_content ) end
end

function OptionScreen:FirstChild( parent )
    for k, w in ipairs( parent.children ) do
        if w:IsShown() then return w end
    end
end

function OptionScreen:Layout()
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Size up the background
    self.panel:SetHeight( PANEL_H )
    
    -- Layout the panels
    self.graphics_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )
    self.audio_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )
    self.other_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )
    self.controls_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )
    self.gameplay_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )
    self.mod_scroll_panel:LayoutBounds( "left", "center", CONTENT_X - LEFT_OFFSET, 0 )

    -- Top navigation tabs
    self.navigation:LayoutBounds( "left", nil, CONTENT_X, 0 )
    self.navigation:LayoutBounds( nil, "above", self.panel ):Offset( 0, 0 )
    self.navigation_label_left:LayoutBounds( "before", "center", self.navigation ):Offset( -15, 0 )
    self.navigation_label_right:LayoutBounds( "after", "center", self.navigation ):Offset( 15, 0 )

    return self
end

function OptionScreen:NavigationChanged( tab )
    self.current_tab = tab

    self.graphics_scroll_panel:SetShown( tab == TAB.GRAPHICS )
    self.audio_scroll_panel:SetShown( tab == TAB.AUDIO )
    self.other_scroll_panel:SetShown( tab == TAB.OTHER )
    self.controls_scroll_panel:SetShown( tab == TAB.CONTROLS )
    self.gameplay_scroll_panel:SetShown( tab == TAB.GAMEPLAY )
    self.mod_scroll_panel:SetShown( tab == TAB.MODS )

    self:RefreshStreamerMode( tab == TAB.GAMEPLAY )
end

function OptionScreen:RefreshBigCharacter()
    if self.biganim == nil then
        local agent = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
        if agent == nil then
            agent = Agent( "SAL" )
        end

        self.biganim = self.anim_root:AddChild( Widget.AnimatedCharacter() )
        self.biganim:Refresh(agent)
        self.biganim:SetScale(1)
    end
    self.biganim:SetPos(-RES_X*0.33-150, -270)
end

function OptionScreen:TryApplySettings()
    local window_mode = self:FindChild("WINDOW_MODE"):GetValue().data
    local resolution = self:FindChild("FS_RES"):GetValue().data

    local bloom = self:FindChild("BLOOM_CHECK"):GetValue().data == 1
    local metrics_enabled = self:FindChild("METRICS"):GetValue().data == 1
    if self.simplified_mode then
        local powersaver_mode = self:FindChild("POWERSAVER_MODE"):GetValue().data == 1
    end
    local auto_skip_enabled = self:FindChild("AUTO_SKIP"):GetValue().data == 1
    local fast_mode_enabled = self:FindChild("FAST_MODE"):GetValue().data == 1
    local battle_speed = self:FindChild("BATTLE_SPEED"):GetValue()
    local negotiation_speed = self:FindChild("NEGOTIATION_SPEED"):GetValue()
    local streamer_mode = self:FindChild("STREAMER_MODE"):GetValue().data -- 0 is disabled, 1 is left, 2 is right
    local twitch_mode = BRANCH == "dev" and self:FindChild("TWITCH_MODE"):GetValue().data
    local colourblind_mode = self:FindChild("COLOURBLIND_MODE"):GetValue().data
    local layoutscale_mode = self:FindChild("LAYOUTSCALE_MODE"):GetValue().data
    local framelimiter_mode = self:FindChild("FRAMELIMITER_MODE"):GetValue().data
    local cursor_mode = self:FindChild("CURSOR"):GetValue().data

    local sway = self:FindChild("SCREEN_SWAY") and self:FindChild("SCREEN_SWAY"):GetValue().data == 1
    local flash = self:FindChild("SCREEN_FLASH"):GetValue().data == 1
    local shake = self:FindChild("SCREEN_SHAKE"):GetValue().data == 1

    local audio_device = self:FindChild("AUDIO_DEVICE"):GetValue().option
    
    if AUDIO and engine.sound and audio_device then
        if audio_device.data then
            AUDIO:SetOutputDeviceByID(audio_device.data)
        else
            AUDIO:SetDefaultOutputDevice()
        end
    end

    local force_mono = self:FindChild("FORCE_MONO"):GetValue().data == 1
    
    if AUDIO then
        AUDIO:SetForceMono(force_mono)
    end

    local mute_on_lose_focus = self:FindChild("MUTE_ON_LOSE_FOCUS"):GetValue().data == 1

    --local skip_movies_enabled = self:FindChild("SKIP_MOVIES"):GetValue().data == 1
    local confirm_select_enabled = self:FindChild("CONFIRM_SELECT"):GetValue().data == 1
    local confirm_endturn_enabled = self:FindChild("CONFIRM_ENDTURN"):GetValue().data == 1
    local lod_level = self:FindChild("LOD"):GetValue().data
   
    local maxy
    local maxy_widget = self:FindChild("MAXY")
    if maxy_widget then maxy = maxy_widget:GetValue().data end

    if window_mode == 0 then
        engine.inst:SetWindowMode(resolution.w, resolution.h, 0, 666666,666666); -- position is not altered 
    else
        engine.inst:SetWindowMode(0, 0, 1, 666666,666666); -- resolution and position are not used
    end

    if not self.simplified_mode then
        TheGame:FE():SetPendingScreenMode( layoutscale_mode or SCREEN_MODE.MONITOR )    
    end

	engine.inst:SetFramerateLimiter(framelimiter_mode);

    engine.inst:SetCursor(cursor_mode)

    if maxy_widget then engine.inst:SetMaxYResolution(maxy) end
    TheGame:SetBloomEnabled(bloom)
    TheGame:SetLevelOfDetail(lod_level);

    engine.inst:SetSetting("graphics", "bloom", bloom)
    engine.inst:SetSetting("graphics", "lod_level", lod_level)

    engine.inst:SetSetting("graphics", "sway_disabled", not sway)
    engine.inst:SetSetting("graphics", "shake_disabled", not shake)
    engine.inst:SetSetting("graphics", "flash_disabled", not flash)

    engine.inst:SetSetting("graphics", "cursor", cursor_mode)

    engine.inst:SetAgreementsSetting( "METRICS", "enabled", metrics_enabled )

    engine.inst:SetSetting("gameplay", "powersaver", fast_mode_enabled)
    engine.inst:SetSetting("gameplay", "auto_skip", auto_skip_enabled)
    engine.inst:SetSetting("gameplay", "fast_mode", fast_mode_enabled)
    engine.inst:SetSetting("gameplay", "battle_speed", battle_speed)
    engine.inst:SetSetting("gameplay", "negotiation_speed", negotiation_speed)
    --engine.inst:SetSetting("gameplay", "skip_movies", skip_movies_enabled)
    engine.inst:SetSetting("gameplay", "confirm_select", confirm_select_enabled)
    engine.inst:SetSetting("gameplay", "confirm_end_turn", confirm_endturn_enabled)
    engine.inst:SetSetting("gameplay", "streamer_mode", streamer_mode)
    if BRANCH == "dev" then
        engine.inst:SetSetting("gameplay", "twitch_mode", twitch_mode)
    end
    engine.inst:SetSetting("graphics", "colourblind_mode", colourblind_mode)
    engine.inst:SetSetting("graphics", "layoutscale_mode", layoutscale_mode)
    engine.inst:SetSetting("graphics", "framelimiter_mode", framelimiter_mode)


    engine.inst:SetSetting("audio", "force_mono", force_mono)
    engine.inst:SetSetting("audio", "mute_on_lose_focus", mute_on_lose_focus)


    engine.inst:SetSetting( "audio", "master_volume", self:FindChild("AUDIO_MASTER"):GetValue() )
    engine.inst:SetSetting( "audio", "music_volume", self:FindChild("AUDIO_MUSIC"):GetValue() )
    engine.inst:SetSetting( "audio", "fx_volume", self:FindChild("AUDIO_FX"):GetValue() )
    engine.inst:SetSetting( "audio", "vo_volume", self:FindChild("AUDIO_VO"):GetValue() )

    engine.inst:SetSetting( "audio", "ambience_volume", self:FindChild("AUDIO_AMBIENCE"):GetValue() )
    engine.inst:SetSetting( "audio", "actions_volume", self:FindChild("AUDIO_ACTIONS"):GetValue() )
    engine.inst:SetSetting( "audio", "cards_volume", self:FindChild("AUDIO_CARDS"):GetValue() )
    engine.inst:SetSetting( "audio", "ui_volume", self:FindChild("AUDIO_UI"):GetValue() )

    self:ApplyModSettings()

    self:ApplyControlSettings()

    TheGame.frontend:RefreshColourblindMode()

    local language_selected = self:FindChild("LANGUAGES"):GetValue().data

	local settingsLang = TheGame:GetLanguage()
    if (settingsLang ~= language_selected) and (Content.GetLocalizations()[language_selected] ~= nil) then
	    engine.inst:SetSetting( "general", "language", language_selected )
		engine.inst:SaveSettings()
		engine.inst:Reset()
    end

--    self:SyncGraphicsValues() NOTE: Don't sync here. Need to wait for kiln to send us events that the windowmode actually changed.
end

function OptionScreen:SyncValues()
    self:SyncGameplayValues()
    self:SyncGraphicsValues()
    self:SyncControlValues()
    self:SyncModValues()
    self:SyncVolumeValues()
    self:MakeClean()
end

function OptionScreen:SyncModValues()
    for i, id in pairs( Content.GetInstalledMods() ) do
        local mod = Content.FindMod( id )
        for k, option in ipairs( mod.mod_options or table.empty ) do
            local w = self:FindChild( option.widget_name )
            local value = Content.GetModSetting( mod, option.key )
            if value ~= nil then
                if option.spinner then
                    w:SetIndexByData( value, nil, true )
                elseif option.slider then
                    w:SetValue( value, true )
                end
            end
        end
    end
end

function OptionScreen:ApplyModSettings()
    for i, id in pairs( Content.GetInstalledMods() ) do
        local mod = Content.FindMod( id )
        for k, option in ipairs( mod.mod_options or table.empty ) do
            local w = self:FindChild( option.widget_name )
            local value

            if option.spinner then
                value = w:GetValue().data                
            elseif option.slider then
                value = w:GetValue()
            end

            Content.SetModSetting( mod, option.key, value )
        end
    end
end

-- Updates the UI to reflect the underlying data in self.control_settings.
function OptionScreen:SyncControlValues()
    local input = TheGame:GetInput()
    for i, control in ipairs( KEY_BINDS ) do
        local control_name = table.find( Controls.Digital, control )
        local widget = self.controls_content:FindChild( control_name )
        if widget then
            local default_bindings = input:GetDefaultKeyBindings( control )
            local binding = self.control_settings.key_binds[ control_name ]
            local title = input:GetBindingString( binding or default_bindings[1] )
            widget:SetTitle( title )

            if binding ~= nil then
                local desc = loc.format( LOC "UI.OPTIONS_SCREEN.CONTROL_DEFAULT", input:GetBindingString( default_bindings[1] ))
                widget:SetDesc( desc )
            else
                widget:SetDesc( nil )
            end
        end
    end

    local widget = self.controls_content:FindChild("REVERT_BINDINGS")
    if widget then
        widget:SetEnabled( not widget.option or not widget.option.disabled or not widget.option.disabled() )
    end
end

function OptionScreen:ApplyControlSettings()
    local settings = TheGame:GetInputSettings()
    settings:AssignInputBindings( self.control_settings )
    TheGame:GetInput():ApplyInputBindings()
    settings:SaveSettings()
end

function OptionScreen:GetSetting( category, key, default_value )
    if self.grab_defaults then
        return default_value
    else
        return engine.inst:GetSetting( category, key, default_value )
    end
end

function OptionScreen:SyncGameplayValues()
    if self.simplified_mode then
        local powersaver_mode = self:GetSetting("gameplay", "powersaver", true) and 1 or 0
        self:FindChild("POWERSAVER_MODE"):SetIndexByData( powersaver_mode, nil, true)
    end
    local auto_skip_enabled = self:GetSetting("gameplay", "auto_skip", false) and 1 or 0
    self:FindChild("AUTO_SKIP"):SetIndexByData( auto_skip_enabled, nil, true)
	local fast_mode_enabled = self:GetSetting("gameplay", "fast_mode", false) and 1 or 0
    self:FindChild("FAST_MODE"):SetIndexByData( fast_mode_enabled, nil, true)
    local battle_speed = self:GetSetting("gameplay", "battle_speed", 1.0)
    self:FindChild("BATTLE_SPEED"):SetValue( battle_speed )
    local negotiation_speed = self:GetSetting("gameplay", "negotiation_speed", 1.0)
    self:FindChild("NEGOTIATION_SPEED"):SetValue( negotiation_speed )
    --local skip_movies_enabled = self:GetSetting("gameplay", "skip_movies", false) and 1 or 0
    --self:FindChild("SKIP_MOVIES"):SetIndexByData( skip_movies_enabled, nil, true)
    local confirm_select_enabled = self:GetSetting("gameplay", "confirm_select", true) and 1 or 0
    self:FindChild("CONFIRM_SELECT"):SetIndexByData( confirm_select_enabled, nil, true)
    local confirm_endturn_enabled = self:GetSetting("gameplay", "confirm_end_turn", true) and 1 or 0
    self:FindChild("CONFIRM_ENDTURN"):SetIndexByData( confirm_endturn_enabled, nil, true)
    local streamer_mode = self:GetSetting("gameplay", "streamer_mode", 0)
    self:FindChild("STREAMER_MODE"):SetIndexByData( streamer_mode, nil, true )
    if BRANCH == "dev" then
        local twitch_mode = self:GetSetting("gameplay", "twitch_mode", 0)
        self:FindChild("TWITCH_MODE"):SetIndexByData( twitch_mode, nil, true )
    end
    local metrics_enabled = engine.inst:GetAgreementsSetting("METRICS", "enabled", true) and 1 or 0
    self:FindChild("METRICS"):SetIndexByData( metrics_enabled, nil, true)


end

-- Show the streamer_mode_overlay if we're on the Gameplay tab
function OptionScreen:RefreshStreamerMode( show )
    local streamer_mode = self:FindChild("STREAMER_MODE"):GetValue().data
    if not self.streamer_mode_overlay and streamer_mode > 0 then
        local w, h = 400, 300
        self.streamer_mode_overlay = self:AddChild( Widget() )
            :SetHiddenBoundingBox( true )
            :UnblockMouse()
        self.streamer_mode_overlay.frame = self.streamer_mode_overlay:AddChild( Widget.Panel( engine.asset.Texture("ui/empty_slot_frame.tex") ) )
            :SetNineSliceCoords( 30, 30, 44, 44)
            :SetSize( w, h )
            :Bloom( 0.2 )
            :UnblockMouse()
        self.streamer_mode_overlay.label = self.streamer_mode_overlay:AddChild( Widget.Label( "body", 24, LOC"UI.OPTIONS_SCREEN.STREAMER_MODE_OVERLAY_TEXT" ) )
            :SetGlyphColour( UICOLOURS.SUBTITLE )
            :SetAutoSize( w*0.55 )
            :Bloom( 0.25 )
            :SetWordWrap( true )
            :UnblockMouse()
    end
    if streamer_mode == 0 then
        -- Disabled
        self.bottom_left:SetPos( 0, 0 )
        self.bottom_right:SetPos( 0, 0 )
    elseif streamer_mode == 1 then
        -- Left side
        self.streamer_mode_overlay:SetAnchors( "left", "bottom" )
        self.streamer_mode_overlay.frame:SetScale( 1, 1 )
        self.streamer_mode_overlay.frame:LayoutBounds( "left", "bottom", 0, 0 )
        self.streamer_mode_overlay.label:LayoutBounds( "center", "center", self.streamer_mode_overlay.frame )
        self.streamer_mode_overlay.frame:SetScale( 1, -1 )
        -- Layout buttons too
        self.bottom_left:SetPos( 360, 0 )
        self.bottom_right:SetPos( 0, 0 )
    elseif streamer_mode == 2 then
        -- Right side
        self.streamer_mode_overlay:SetAnchors( "right", "bottom" )
        self.streamer_mode_overlay.frame:SetScale( 1, 1 )
        self.streamer_mode_overlay.frame:LayoutBounds( "right", "bottom", 0, 0 )
        self.streamer_mode_overlay.label:LayoutBounds( "center", "center", self.streamer_mode_overlay.frame )
        self.streamer_mode_overlay.frame:SetScale( -1, -1 )
        -- Layout buttons too
        self.bottom_left:SetPos( 0, 0 )
        self.bottom_right:SetPos( -360, 0 )
    end
    local streamer_mode = self:FindChild("STREAMER_MODE"):GetValue().data
    if self.streamer_mode_overlay then self.streamer_mode_overlay:SetShown( show and streamer_mode > 0 ) end
end

function OptionScreen:RefreshTwitchMode()
    local twitch_spinner = self:FindChild("TWITCH_MODE")
    if twitch_spinner then
        local twitch_mode = twitch_spinner:GetValue().data
        -- Update the refresh button
        if twitch_mode == 1 then
            twitch_spinner.twitch_refresh:Show()
            -- If this was disabled before and now it's enabled, connect to twitch
            if self:GetSetting("gameplay", "twitch_mode", 0) == 0 then self:ConnectToTwitch() end
        else
            twitch_spinner.twitch_refresh:Hide()
        end
    end
end

function OptionScreen:ConnectToTwitch()
    local twitch_spinner = self:FindChild("TWITCH_MODE")
    local twitch_mode = twitch_spinner:GetValue().data
    if twitch_mode == 1 then
        -- Disable the button
        twitch_spinner.twitch_refresh:Disable()
        -- Make the request
        Twitch.GameConnect( function(response) self:ConnectToTwitchResponse(response) end )
    end
end

function OptionScreen:ConnectToTwitchResponse( response )
    local twitch_spinner = self:FindChild("TWITCH_MODE")
    -- Re-enable the button
    twitch_spinner.twitch_refresh:Enable()
    -- Handle response
    if response and response.error then
        if response.error == "NO_TWITCH_ID" then
            -- Tell user to connect their twitch account to their Klei account
            TheGame:FE():PushScreen(Screen.YesNoPopup(
                LOC"UI.TWITCH.NO_TWITCH_POPUP_TITLE", 
                LOC"UI.TWITCH.NO_TWITCH_POPUP_TEXT" ))
                :SetFn( function(v)
                    if v == Screen.YesNoPopup.YES then 
                        engine.inst:OpenUrlInBrowser( "https://accounts.klei.com/account/info" )
                    end
                end )
        else
            -- Some other error. Show the user that too
            TheGame:FE():PushScreen(Screen.InfoPopup( LOC"UI.TWITCH.ERROR_TWITCH_POPUP_TITLE", response.error ))
        end
    elseif response then
        TheGame:FE():PushScreen(Screen.InfoPopup( LOC"UI.TWITCH.CONNECTED_POPUP_TITLE", LOC"UI.TWITCH.CONNECTED_POPUP_TEXT" ))
    end
end

function OptionScreen:SyncGraphicsValues()
--    local is_vs = engine.inst:GetVsync()
    local bloom = TheGame.bloom_enabled
    local maxy = engine.inst:GetMaxYResolution()
    local lod_level = TheGame:GetLevelOfDetail()

    local res_x, res_y, fs_mode, pos_x, pos_y = engine.inst:GetWindowMode();   -- current windowed mode settings
    self:FindChild("WINDOW_MODE"):SetIndexByData( fs_mode, nil, true )


    self:FindChild("CURSOR"):SetIndexByData( self:GetSetting("graphics", "cursor", "CURSOR_3"))
    self:UpdateResolutions()

    local colourblind_mode = self:GetSetting( "graphics", "colourblind_mode", 1 )
    self:FindChild("COLOURBLIND_MODE"):SetIndexByData( colourblind_mode, nil, true )
    self:OnSelectColorblindMode( colourblind_mode )

    local layoutscale_mode = self:GetSetting( "graphics", "layoutscale_mode", SCREEN_MODE.MONITOR )
    if tonumber(layoutscale_mode) then layoutscale_mode = SCREEN_MODE_INDEX[layoutscale_mode] end
    self:FindChild("LAYOUTSCALE_MODE"):SetIndexByData( layoutscale_mode, nil, true )

    local framelimiter_mode = self:GetSetting( "graphics", "framelimiter_mode", 1 )
    self:FindChild("FRAMELIMITER_MODE"):SetIndexByData( framelimiter_mode, nil, true )
    
    self:FindChild("BLOOM_CHECK"):SetIndexByData(bloom and 1 or 0, nil, true)

    local sway = self:FindChild("SCREEN_SWAY")
    if sway then
        sway:SetIndexByData(TheGame:ScreenSwayEnabled() and 1 or 0, nil, true)
    end
    self:FindChild("SCREEN_SHAKE"):SetIndexByData(TheGame:CamShakeEnabled() and 1 or 0, nil, true)
    self:FindChild("SCREEN_FLASH"):SetIndexByData(TheGame:ScreenFlashEnabled() and 1 or 0, nil, true)

    self:FindChild("LOD"):SetIndexByData(lod_level, nil, true)

    local language = TheGame:GetLanguage()
    self:FindChild("LANGUAGES"):SetIndexByData( language )

    -- Find the closest index to the current scale:
    local maxy_idx = self:GetClosestMaxYOptionIndex( maxy )

    local maxy_widget = self:FindChild("MAXY")
    if maxy_widget then maxy_widget:SetIndex(maxy_idx, true) end
end

function OptionScreen:GetClosestMaxYOptionIndex( maxy )
    -- Receives a screen height and returns the closest MAXY option index
    local dsidx = -1
    local mindist = 100000.0

    local curridx = 1
    for k, data in ipairs(self.MAXY_OPTIONS) do
        local dist = math.abs(maxy - data.data)
        if (dist < mindist) then
            mindist = dist
            dsidx = curridx
        end
        curridx = curridx + 1
    end
    return dsidx
end

function OptionScreen:OnSelectMaxYRes(val)
    self:MakeDirty()
end

function OptionScreen:OnSelectLOD(val)
    if not self.shown_lod_warning then
        TheGame:FE():PushScreen( Screen.InfoPopup( LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_POPUP_TITLE", LOC"UI.OPTIONS_SCREEN.DETAILLEVEL_POPUP_TEXT" ) )
        self.shown_lod_warning = true
    end
    self:MakeDirty()
end

function OptionScreen:OnSelectColorblindMode( val )
    local colours = GetColourblindModeColours()

    local spinner = self:FindChild("COLOURBLIND_MODE")
    spinner.colourblind_icons[CARD_TYPE.NEGOTIATION_HOSTILE]:SetTintColour( colours[CARD_TYPE.NEGOTIATION_HOSTILE][val] )
    spinner.colourblind_icons[CARD_TYPE.NEGOTIATION_DIPLOMACY]:SetTintColour( colours[CARD_TYPE.NEGOTIATION_DIPLOMACY][val] )
    spinner.colourblind_icons[CARD_TYPE.NEGOTIATION_MANIPULATE]:SetTintColour( colours[CARD_TYPE.NEGOTIATION_MANIPULATE][val] )
    spinner.colourblind_icons[CARD_TYPE.FIGHT_MELEE]:SetTintColour( colours[CARD_TYPE.FIGHT_MELEE][val] )
    spinner.colourblind_icons[CARD_TYPE.FIGHT_SKILL]:SetTintColour( colours[CARD_TYPE.FIGHT_SKILL][val] )
    spinner.colourblind_icons[CARD_TYPE.FIGHT_STATUS]:SetTintColour( colours[CARD_TYPE.FIGHT_STATUS][val] )
    spinner.colourblind_icons[CARD_TYPE.ITEM]:SetTintColour( colours[CARD_TYPE.ITEM][val] )
    spinner.colourblind_icons[CARD_TYPE.PARASITE]:SetTintColour( colours[CARD_TYPE.PARASITE][val] )
    spinner.colourblind_icons[CARD_TYPE.FLOURISH]:SetTintColour( colours[CARD_TYPE.FLOURISH][val] )
end

function OptionScreen:OnSetBloom(val)
    if ( self:FindChild("BLOOM_CHECK"):GetValue().data == 1 ) then
        self.biganim:Emote("clap")
    else
        self.biganim:Emote("disappoint")
    end
    self:MakeDirty()
end


function OptionScreen:OnSetWindowMode(idx)
    self:UpdateResolutions()
    self:MakeDirty()
end

function OptionScreen:OnSetPowersaverMode(idx)
    -- Check if the user set it to High Quality (1), or Power Saver (0)
    local high_quality = idx == 1

    -- Set the framerate limit
    -- High Quality sets the limit to 60fps (2)
    -- Power Saver sets the limit to 30fps (3)
    self:FindChild("FRAMELIMITER_MODE"):SetIndexByData( high_quality and 2 or 3, nil, true )

    -- Set the LOD
    -- High Quality sets it to High (4)
    -- Power Saver sets it to Low (1)
    self:FindChild("LOD"):SetIndexByData( high_quality and 4 or 1, nil, true)

    -- Set the Max Target Size
    -- High Quality sets it to the device's screen resolution
    -- Power Saver sets it to 70% of the device's screen resolution
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    -- Look for the option closest to this resolution
    local maxy_idx = self:GetClosestMaxYOptionIndex( high_quality and screen_h or screen_h*0.7 )
    self:FindChild("MAXY"):SetIndex(maxy_idx, true)

    self:MakeDirty()
end

function OptionScreen:OnSelectResolution(idx)
    self:MakeDirty()
end

function OptionScreen:UpdateResolutions()
    local window_mode = self:FindChild("WINDOW_MODE"):GetValue().data

    local filtered_resolution_options = {}
    local selected_filtered_resolution_option = 0;
    local desk_x, desk_y = engine.inst:GetDesktopSize();

    if window_mode == 0 then    -- windowed
        local win_res_x, win_res_y, fs_mode, win_pos_x, win_pos_y = engine.inst:GetWindowMode(0);   -- windowed mode settings

        -- If the current window resolution is none of the choices, don't add a custom size
        for _, value in ipairs(self.option_resolution.values) do
            if (value.data.w <= desk_x) and (value.data.h <= desk_y) then
                table.insert( filtered_resolution_options, value )
                if (value.data.w == win_res_x) and (value.data.h == win_res_y) then
                    selected_filtered_resolution_option = #filtered_resolution_options;
                end
            end
        end

        if (selected_filtered_resolution_option == 0) then
            -- current resolution doesn't show in the various resolution options, so create a custom one:
            table.insert(filtered_resolution_options, 1, { name=loc.format(LOC"UI.OPTIONS_SCREEN.CUSTOM_RESOLUTION", win_res_x, win_res_y), desc=LOC"UI.OPTIONS_SCREEN.RESOLUTION_DESC", data = {w=win_res_x, h=win_res_y} } )    -- Insert at the top of the list
            selected_filtered_resolution_option = 1 -- TODO
        end

    else -- fullscreen
        -- Get the resolution of the desktop, and add that as the only option.

        table.insert( filtered_resolution_options, { name=loc.format("{1}x{2}", desk_x, desk_y), desc=LOC"UI.OPTIONS_SCREEN.CUSTOM_RESOLUTION_DESC", data = {w=desk_x, h=desk_y } } )
        selected_filtered_resolution_option = 1
    end   

    self:FindChild("FS_RES"):SetOptions(filtered_resolution_options, selected_filtered_resolution_option)
end

function OptionScreen:OnScreenResize(w,h)
    self._base.OnScreenResize(self, w, h)

    -- Only sync the graphics values if the current fullscreen mode is the same as what is selected in the options screen.
    local res_x, res_y, fs_mode, pos_x, pos_y = engine.inst:GetWindowMode();   -- current windowed mode settings
    local window_mode = self:FindChild("WINDOW_MODE"):GetValue().data

    if window_mode == fs_mode then
        self:SyncGraphicsValues();
    end

    self:Layout()
end

function OptionScreen:OnOpen()
    OptionScreen._base.OnOpen( self )

    local x, y = self.biganim:GetPos()
    self.biganim:SetPos(x-200, y)
    self.biganim:MoveTo(x, y, .3, easing.outQuad)
end

function OptionScreen:OnControlModeChange( cm, device_id )
    OptionScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.close_button:Show()
        self.apply_button:Show()
        self.revert_button:Show()
        self.navigation_label_left:Hide()
        self.navigation_label_right:Hide()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.close_button:Show()
        self.apply_button:Show()
        self.revert_button:Show()
        self.navigation_label_left:Hide()
        self.navigation_label_right:Hide()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.close_button:Hide()
        self.apply_button:Hide()
        self.revert_button:Hide()
        self.navigation_label_left:Show()
        self.navigation_label_right:Show()
    end

    -- Update the panels
    self.graphics_scroll_panel:ShowBar( cm == CONTROL_MODE.TOUCH and SCROLLBAR.NEVER or SCROLLBAR.ALWAYS )
    self.audio_scroll_panel:ShowBar( cm == CONTROL_MODE.TOUCH and SCROLLBAR.NEVER or SCROLLBAR.ALWAYS )
    self.other_scroll_panel:ShowBar( cm == CONTROL_MODE.TOUCH and SCROLLBAR.NEVER or SCROLLBAR.ALWAYS )
    self.gameplay_scroll_panel:ShowBar( cm == CONTROL_MODE.TOUCH and SCROLLBAR.NEVER or SCROLLBAR.ALWAYS )

    -- If we've hidden the scroll bar, then resize the tabs to match the width of the spinners instead of the scroll panel
    if cm == CONTROL_MODE.TOUCH then
        self.navigation:SetTotalWidth( CONTENT_W-40 )
    else
        self.navigation:SetTotalWidth( TABS_W )
    end
    self.navigation:LayoutBounds( "left", nil, CONTENT_X, 0 )

    self.navigation_label_left:LayoutBounds( "before", "center", self.navigation ):Offset( -15, 0 )
    self.navigation_label_right:LayoutBounds( "after", "center", self.navigation ):Offset( 15, 0 )
end

function OptionScreen:OnScreenModeChange( sm )
    OptionScreen._base.OnScreenModeChange( self, sm )

    if sm == SCREEN_MODE.MONITOR then
        PANEL_H = 540
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        PANEL_H = 660
    end

    -- Update the panels
    self.graphics_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )
    self.audio_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )
    self.other_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )
    self.controls_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )
    self.gameplay_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )
    self.mod_content:SetLayoutScale( LAYOUT_SCALE[sm] )
        :SetWidth( CONTENT_W/LAYOUT_SCALE[sm] -30 )
        :LayoutBounds( "left", "top", 10, -15 )

    self.graphics_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
    self.audio_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
    self.other_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
    self.gameplay_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
    self.controls_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )
    self.mod_scroll_panel:SetSize( CONTENT_W+LEFT_OFFSET, PANEL_H )

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_center:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end

function OptionScreen:SyncVolumeValues()
    
    local audio_device_override = AUDIO and AUDIO:GetAudioDeviceOverrideName()
    local device_data
    if audio_device_override then
        for k,v in ipairs(self.audio_device_options) do
            if audio_device_override == v.name then
                device_data = v.data
                break
            end
        end
    end

    if device_data then
        self:FindChild("AUDIO_DEVICE"):SetIndexByData( device_data, nil, true)
    else
        self:FindChild("AUDIO_DEVICE"):SetIndex(1, true)
    end

    self:FindChild("FORCE_MONO"):SetIndexByData( AUDIO and AUDIO:GetForceMono() and 1 or 0, nil, true)

    
    local mute_on_lose_focus = self:GetSetting("audio", "mute_on_lose_focus", false)
    self:FindChild("MUTE_ON_LOSE_FOCUS"):SetIndexByData( mute_on_lose_focus and 1 or 0, nil, true)

    self:FindChild("AUDIO_MASTER"):SetValue( self:GetSetting( "audio", "master_volume", 1.0), true)
    self:FindChild("AUDIO_MUSIC"):SetValue( self:GetSetting( "audio", "music_volume", 1.0), true)
    self:FindChild("AUDIO_FX"):SetValue( self:GetSetting( "audio", "fx_volume", 1.0), true)
    self:FindChild("AUDIO_VO"):SetValue( self:GetSetting( "audio", "vo_volume", 1.0), true)

    self:FindChild("AUDIO_AMBIENCE"):SetValue( self:GetSetting( "audio", "ambience_volume", 1.0), true)
    self:FindChild("AUDIO_ACTIONS"):SetValue( self:GetSetting( "audio", "actions_volume", 1.0), true)
    self:FindChild("AUDIO_CARDS"):SetValue( self:GetSetting( "audio", "cards_volume", 1.0) or 1, true)
    self:FindChild("AUDIO_UI"):SetValue( self:GetSetting( "audio", "ui_volume", 1.0), true)

    self.volume_off = self:IsVolumeOff()
end

function OptionScreen:IsVolumeOff()
    return self:FindChild("AUDIO_MASTER"):GetValue() == 0
        or ( self:FindChild("AUDIO_FX"):GetValue() == 0 and self:FindChild("AUDIO_MUSIC"):GetValue() == 0 and self:FindChild("AUDIO_VO"):GetValue() == 0 )
end

function OptionScreen:VolumeChange(id, v)

    TheGame:SetVolume(id, v)

    local volume_off = self:IsVolumeOff()

    -- Update this to make Sal frown whenever all the audio is off
    if volume_off and not self.volume_off then
        self.biganim:Emote("disappoint")
    elseif self.volume_off and not volume_off then
        self.biganim:Emote("clap")
    end

    self.volume_off = volume_off

    if not self.dirty then
        self:MakeDirty()
    end
end

function OptionScreen:Revert()
    if self.dirty then

        self.control_settings =
        {
            key_binds = shallowcopy( TheGame:GetInputSettings():GetSetting( "key_binds" )),
        }

        TheGame:ApplyVolumeSettings()

        self:SyncValues()
    end
end

function OptionScreen:Reset()
    self.grab_defaults = true

    table.clear( self.control_settings.key_binds )

    for i, key in ipairs( VOLUME_SETTINGS ) do
        TheGame:SetVolume( key, 1.0 )
    end

    self:SyncValues()
    self:MakeDirty()
    self.grab_defaults = nil
end

function OptionScreen:OnCancel()
    if self.dirty then
        TheGame:FE():PushScreen( Screen.YesNoPopup(LOC"UI.OPTIONS_SCREEN.CONFIRM_CLOSE", LOC"UI.OPTIONS_SCREEN.CONFIRM_CLOSE_DETAILS" ):SetFn(
			function(v) if v == Screen.YesNoPopup.YES then self:Revert() self:GetFE():FadeTransition(function() self:GetFE():PopScreen() end) end end ) )
    else
        self:GetFE():FadeTransition(function() 
            self:GetFE():PopScreen()
        end)
    end
end

function OptionScreen:MakeClean()
    if self.dirty then
        self.dirty = false  
        self.apply_button:Disable()
        self.revert_button:Disable()
        
        if self.fe then
            self.fe:RequestControlRefresh()
        end
    end
end

function OptionScreen:MakeDirty()
    if not self.dirty then
        self.dirty = true
        self.apply_button:Enable()
        self.revert_button:Enable()

        if self.fe then
            self.fe:RequestControlRefresh()
        end
    end
end

function OptionScreen:OnApply()
    AUDIO:PlayEvent("event:/ui/main/gen/back_general")
    if self.dirty then
        self:TryApplySettings()

        engine.inst:SaveSettings()
		engine.inst:SaveAgreementsSettings()

        TheGame:BroadcastEvent( "options_changed" )
        
        self:MakeClean()
    end
end

------------------------------------------------------------------------------------------
-- Shows a single list-button

local OptionsScreenButton = class( "Widget.OptionsScreenButton", Widget.Clickable )

local HEIGHT = 90
local BUTTON_W = 50
local ICON_SIZE = 24
local BUTTONS_X_OFFSET = 10

function OptionsScreenButton:init( width, name, title, desc, fn )
    self._base.init( self )

    self.width = width
    self:SetOnClickFn( fn )

    -- Hitbox
    self.bg = self:AddChild( Widget.SolidBox( self.width, HEIGHT, 0x00ff0000 ) )

    -- Add our title
    self.title = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_LABEL, string.upper( name ) ) )

    -- Add our content block
    self.bg_border = 20
    self.bg_border_scale = 0.5
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsbutton_bg.tex") ) )
    self.content_bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.content_bg:SetNineSliceBorderScale( self.bg_border_scale ):Bloom( 0.05 )
    self.content_outline = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsbutton_border.tex") ) )
    self.content_outline:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.content_outline:SetNineSliceBorderScale( self.bg_border_scale ):Bloom( 0.2 )

    -- Add our content
    self.text_content = self:AddChild( Widget() )
    self.value_label = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, title ) )
        :SetWordWrap( true )
        :OverrideLineHeight( 24 )
        :LeftAlign()
        :Bloom(0.05)
    self.desc_label = self.text_content:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC, desc ) )
        :SetWordWrap( true )
        :OverrideLineHeight( 18 )
        :LeftAlign()

    -- Add our button
    self.right = self:AddChild( Widget.Clickable() )
    self.right:SetOnClickFn( fn )
    self.right:SetHoverFn( function( hover ) self.right.bg:Bloom( hover and 0.1 or 0.07 ) end )
    self.right.Disable = function() self.right.bg:SetTintColour( UICOLOURS.TITLE ):SetTintAlpha( 0.2 ) self.right.icon:SetTintAlpha( 0.7 ) self.right.enabled = false return self end
    self.right.Enable = function() self.right.bg:SetTintColour( UICOLOURS.SUBTITLE ):SetTintAlpha( 0.7 ) self.right.icon:SetTintAlpha( 1 ) self.right.enabled = true return self end
    self.right.bg = self.right:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonright.tex") ) )
    self.right.bg:SetTintColour( UICOLOURS.SUBTITLE )
    self.right.bg:SetTintAlpha( 0.7 ):Bloom( 0.07 )
    self.right.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.right.bg:SetNineSliceBorderScale( self.bg_border_scale )
    self.right.bg:SetSize( BUTTON_W, HEIGHT )
    self.right.icon = self.right:AddChild( Widget.Image( engine.asset.Texture("UI/ic_detailsspinner_right.tex"), ICON_SIZE, ICON_SIZE ) )
    self.right.icon:SetPos( 2, 0 )

    self.initialized = true

    -- Layout everything
    self:Layout()
    self:SetColour( UICOLOURS.SUBTITLE )
end

function OptionsScreenButton:SetWidth( width )
    self.width = width
    self.bg:SetSize( self.width, HEIGHT )
    self:Layout()
    return self
end

function OptionsScreenButton:Layout()
    if not self.initialized then return end

    local buttons_count = 1
    local button_spacing = 0
    local content_w = self.width - ( buttons_count*BUTTON_W ) - ( buttons_count-1 )*button_spacing
    
    -- Size up our right button
    local nineslice_w, nineslice_h = BUTTON_W + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
    self.right.bg:SetSize( nineslice_w, nineslice_h )
    self.right:LayoutBounds( "right", "center", self.bg )

    -- Size up our content block
    local nineslice_w, nineslice_h = content_w + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
    self.content_bg:SetSize( nineslice_w, nineslice_h )
    self.content_outline:SetSize( nineslice_w, nineslice_h )
    self.content_bg:LayoutBounds( "before", "center", self.right )
    self.content_outline:LayoutBounds( "center", "center", self.content_bg )

    -- Position our title
    self.title:LayoutBounds( "left", "above", self.bg ):Offset( 0, 4 )

    -- Calculate text width
    local padding = 20
    local text_w = content_w - padding*2
    self.value_label:SetAutoSize( text_w )
    self.desc_label:SetAutoSize( text_w )

    -- Position our content
    self.desc_label:LayoutBounds( "left", "below", self.value_label ):Offset( 0, 4 )
    self.text_content:LayoutBounds( "left", "center", self.content_bg ):Offset( padding, 0 )
    
end

function OptionsScreenButton:UpdateState()
    if self.focus or self.hover then
        self.title:SetGlyphColour( UICOLOURS.WHITE )
        self.value_label:SetGlyphColour( UICOLOURS.WHITE )
        self.content_bg:Bloom( 0.1 )
        self.content_outline:Bloom( 0.3 )
        self:ScaleTo( nil, 1.01, 0.2, easing.outQuad )   
    else
        self.title:SetGlyphColour( UICOLOURS.SUBTITLE )
        self.value_label:SetGlyphColour( UICOLOURS.SUBTITLE )
        self.content_bg:Bloom( 0.05 )
        self.content_outline:Bloom( 0.2 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function OptionsScreenButton:OnGainFocus()
    self:UpdateState()

    -- Make sure we're visible within the parent scrollpanel
    local scrollpanel = self:GetAncestor( Widget.ScrollPanel )
    if scrollpanel then scrollpanel:EnsureVisible(self) end
end

function OptionsScreenButton:OnLoseFocus()
    self:UpdateState()
end

function OptionsScreenButton:OnGainHover()
    self:UpdateState()
end

function OptionsScreenButton:OnLoseHover()
    self:UpdateState()
end

function OptionsScreenButton:SetTitle( title )
    self.value_label:SetText( title )
    self:Layout()
end

function OptionsScreenButton:SetDesc( desc )
    self.desc_label:SetText( desc )
    self:Layout()
end

function OptionsScreenButton:SetColour( colour )
    self.title:SetGlyphColour( colour )
    self.value_label:SetGlyphColour( colour )
    local r,g,b,a = HexColour( colour )
    a = 0.2
    self.content_bg:SetTintColour( r,g,b,a )
    self.content_outline:SetTintColour( UICOLOURS.SUBTITLE )
end

function OptionsScreenButton:Disable()
    self._base.Disable( self )
    self:SetSaturation( 0.1 )
    return self
end

function OptionsScreenButton:Enable()
    self._base.Enable( self )
    self:SetSaturation( 1.0 )
    return self
end

------------------------------------------------------------------------------------------
-- Shows a single spinner

local OptionsScreenSpinner = class( "Widget.OptionsScreenSpinner", Widget.Spinner )

local HEIGHT = 90
local IMAGE_H = 110
local BUTTON_W = 50
local ICON_SIZE = 24
local BUTTONS_X_OFFSET = 10

function OptionsScreenSpinner:init( width, name, id, option )
    self._base.init( self )

    -- Save data
    self.width = width
    self.name = name
    self.id = id
    self:SetFn( option.fn )
    self.option = option

    -- Hitbox
    self.bg:SetSize( self.width, HEIGHT )
    self.bg:SetColour( 0xFF000000 )

    -- Add our title
    self.title = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_LABEL, string.upper( self.name ) ) )

    -- Add our content block
    self.bg_border = 20
    self.bg_border_scale = 0.5
    self.content_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_bg.tex") ) )
    self.content_bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.content_bg:SetNineSliceBorderScale( self.bg_border_scale ):Bloom( 0.05 )
    self.content_outline = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_border.tex") ) )
    self.content_outline:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.content_outline:SetNineSliceBorderScale( self.bg_border_scale ):Bloom( 0.2 )

    -- Add our content
    self.text_content = self:AddChild( Widget() )
    self.value_label = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetWordWrap( true )
        :OverrideLineHeight( 24 )
        :LeftAlign()
        :Bloom(0.05)
    self.desc_label = self.text_content:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ) )
        :SetWordWrap( true )
        :OverrideLineHeight( 18 )
        :LeftAlign()

    -- Override the default spinner buttons with fancier ones!
    self.left:Remove()
    self.left = self:AddChild( Widget.Clickable() )
    self.left:SetOnClickFn(
        function()
            AUDIO:PlayEvent( "event:/ui/options_menu/arrow" )
            self:Left()
        end )
    self.left:SetHoverFn( function( hover ) self.left.bg:Bloom( hover and 0.1 or 0.07 ) end )
    self.left.Disable = function() self.left.bg:SetTintColour( UICOLOURS.TITLE ):SetTintAlpha( 0.2 ) self.left.icon:SetTintAlpha( 0.7 ) self.left.enabled = false return self end
    self.left.Enable = function() self.left.bg:SetTintColour( UICOLOURS.SUBTITLE ):SetTintAlpha( 0.7 ) self.left.icon:SetTintAlpha( 1 ) self.left.enabled = true return self end
    self.left.bg = self.left:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonleft.tex") ) )
    self.left.bg:SetTintColour( UICOLOURS.SUBTITLE )
    self.left.bg:SetTintAlpha( 0.7 ):Bloom( 0.07 )
    self.left.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.left.bg:SetNineSliceBorderScale( self.bg_border_scale )
    self.left.bg:SetSize( BUTTON_W, HEIGHT )
    self.left.icon = self.left:AddChild( Widget.Image( engine.asset.Texture("UI/ic_detailsspinner_left.tex"), ICON_SIZE, ICON_SIZE ) )
    self.left.icon:SetPos( -2, 0 )
    self.right:Remove()
    self.right = self:AddChild( Widget.Clickable() )
    self.right:SetOnClickFn( 
        function()
            AUDIO:PlayEvent( "event:/ui/options_menu/arrow" )
            self:Right()
        end )
    self.right:SetHoverFn( function( hover ) self.right.bg:Bloom( hover and 0.1 or 0.07 ) end )
    self.right.Disable = function() self.right.bg:SetTintColour( UICOLOURS.TITLE ):SetTintAlpha( 0.2 ) self.right.icon:SetTintAlpha( 0.7 ) self.right.enabled = false return self end
    self.right.Enable = function() self.right.bg:SetTintColour( UICOLOURS.SUBTITLE ):SetTintAlpha( 0.7 ) self.right.icon:SetTintAlpha( 1 ) self.right.enabled = true return self end
    self.right.bg = self.right:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonright.tex") ) )
    self.right.bg:SetTintColour( UICOLOURS.SUBTITLE )
    self.right.bg:SetTintAlpha( 0.7 ):Bloom( 0.07 )
    self.right.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    self.right.bg:SetNineSliceBorderScale( self.bg_border_scale )
    self.right.bg:SetSize( BUTTON_W, HEIGHT )
    self.right.icon = self.right:AddChild( Widget.Image( engine.asset.Texture("UI/ic_detailsspinner_right.tex"), ICON_SIZE, ICON_SIZE ) )
    self.right.icon:SetPos( 2, 0 )

    -- Set the preview button
    if self.option.dropdown then
        self.dropdown = self:AddChild( Widget.Clickable() )
        self.dropdown:SetOnClickFn(
            function()
                AUDIO:PlayEvent( "event:/ui/relationships_menu/character_whoosh_in" )
                self:TriggerDropdown()
            end )
        self.dropdown:SetHoverFn( function( hover ) self.dropdown.bg:Bloom( hover and 0.1 or 0.07 ) end )
        self.dropdown.Disable = function() self.dropdown.bg:SetTintColour( UICOLOURS.TITLE ):SetTintAlpha( 0.2 ) self.dropdown.icon:SetTintAlpha( 0.7 ) self.dropdown.enabled = false return self end
        self.dropdown.Enable = function() self.dropdown.bg:SetTintColour( UICOLOURS.SUBTITLE ):SetTintAlpha( 0.7 ) self.dropdown.icon:SetTintAlpha( 1 ) self.dropdown.enabled = true return self end
        self.dropdown:SetToolTip( LOC"UI.OPTIONS_SCREEN.FX_DROPDOWN_TT" )
        self.dropdown.bg = self.dropdown:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonright.tex") ) )
        self.dropdown.bg:SetTintColour( UICOLOURS.SUBTITLE )
        self.dropdown.bg:SetTintAlpha( 0.7 ):Bloom( 0.07 )
        self.dropdown.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        self.dropdown.bg:SetNineSliceBorderScale( self.bg_border_scale )
        self.dropdown.bg:SetSize( BUTTON_W, HEIGHT )
        self.dropdown.icon = self.dropdown:AddChild( Widget.Image( engine.asset.Texture("UI/ic_detailsspinner_dropdown.tex"), ICON_SIZE, ICON_SIZE ) )
        self.dropdown.icon:SetPos( 2, 0 )
    end

    -- Set the refresh button
    if self.option.twitch_refresh_fn then
        self.twitch_refresh = self:AddChild( Widget.Clickable() )
        self.twitch_refresh:SetOnClickFn(
            function()
                AUDIO:PlayEvent( "event:/ui/relationships_menu/character_whoosh_in" )
                self.option.twitch_refresh_fn()
            end )
        self.twitch_refresh:SetHoverFn( function( hover ) self.twitch_refresh.bg:Bloom( hover and 0.1 or 0.07 ) end )
            :SetToolTip( self.option.twitch_refresh_tt )
        self.twitch_refresh.Disable = function() self.twitch_refresh.bg:SetTintColour( UICOLOURS.TITLE ):SetTintAlpha( 0.2 ) self.twitch_refresh.icon:SetTintAlpha( 0.7 ) self.twitch_refresh.enabled = false return self end
        self.twitch_refresh.Enable = function() self.twitch_refresh.bg:SetTintColour( UICOLOURS.SUBTITLE ):SetTintAlpha( 0.7 ) self.twitch_refresh.icon:SetTintAlpha( 1 ) self.twitch_refresh.enabled = true return self end
        self.twitch_refresh.bg = self.twitch_refresh:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_buttonright.tex") ) )
            :SetTintColour( UICOLOURS.SUBTITLE )
            :SetTintAlpha( 0.7 ):Bloom( 0.07 )
            :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
            :SetNineSliceBorderScale( self.bg_border_scale )
            :SetSize( BUTTON_W, HEIGHT )
        self.twitch_refresh.icon = self.twitch_refresh:AddChild( Widget.Image( engine.asset.Texture("UI/ic_refresh.tex"), ICON_SIZE, ICON_SIZE ) )
            :SetTintColour( UICOLOURS.DARK_BODYTEXT )
            :SetPos( 2, 0 )
    end

    -- If this is a slider instead of spinner. Let's add a slider input
    if self.option.slider then
        self.slider_widget = self:AddChild( Widget.Slider(200) )
            :SetRange( self.option.values.range_min, self.option.values.range_max )
            :SetStepSize( self.option.values.step_size )
            :SetFn( function(v) self:SetValue( v ) end )
            :HideButtons()
        self.Left = function()
                AUDIO:PlayEvent( "event:/ui/options_menu/arrow" )
                self.slider_widget:Decrement()
                return true
            end
        self.Right = function()
                AUDIO:PlayEvent( "event:/ui/options_menu/arrow" )
                self.slider_widget:Increment()
                return true
            end
    end

    -- If this is the colourblind spinner, show the coloured card icons
    if self.option.colourblind then
        self.colourblind_bg_border = 45
        self.colourblind_bg_border_scale = 0.5
        self.colourblind_bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/detailsspinner_label.tex") ) )
            :SetNineSliceCoords( self.colourblind_bg_border, self.colourblind_bg_border, self.colourblind_bg_border, self.colourblind_bg_border )
            :SetNineSliceBorderScale( self.colourblind_bg_border_scale )
            :SetTintColour( UICOLOURS.BLACK )
            :SetTintAlpha( 0.6 )
            :SendToBack()

        self.colourblind_container = self:AddChild( Widget() )
        local icon_size = 26
        self.colourblind_icons = {}
        self.colourblind_icons[CARD_TYPE.NEGOTIATION_HOSTILE] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_negotiation_hostility.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT",  GetCardTypeString( CARD_TYPE.NEGOTIATION_HOSTILE ) ) )
        self.colourblind_icons[CARD_TYPE.NEGOTIATION_DIPLOMACY] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_negotiation_diplomacy.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.NEGOTIATION_DIPLOMACY ) ) )
        self.colourblind_icons[CARD_TYPE.NEGOTIATION_MANIPULATE] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_negotiation_manipulation.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.NEGOTIATION_MANIPULATE ) ) )
        self.colourblind_icons[CARD_TYPE.FIGHT_MELEE] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_combat_attack.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.FIGHT_MELEE ) ) )
        self.colourblind_icons[CARD_TYPE.FIGHT_SKILL] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_combat_maneuver.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.FIGHT_SKILL ) ) )
        self.colourblind_icons[CARD_TYPE.FIGHT_STATUS] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_combat_status.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.FIGHT_STATUS ) ) )
        self.colourblind_icons[CARD_TYPE.ITEM] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_item.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.ITEM ) ) )
        self.colourblind_icons[CARD_TYPE.PARASITE] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_parasite.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.PARASITE ) ) )
        self.colourblind_icons[CARD_TYPE.FLOURISH] = self.colourblind_container:AddChild( Widget.Image( engine.asset.Texture( "UI/cardview_typeicon_flourish.tex" ), icon_size, icon_size ) ):SetToolTip( loc.format( LOC"UI.OPTIONS_SCREEN.COLOURBLIND_TT", GetCardTypeString( CARD_TYPE.FLOURISH ) ) )
    end

    self.initialized = true

    -- Layout everything
    self:Layout()
    self:SetColour( UICOLOURS.SUBTITLE )
end

function OptionsScreenSpinner:SetWidth( width )
    self.width = width
    self.bg:SetSize( self.width, HEIGHT )
    self:Layout()
    return self
end

function OptionsScreenSpinner:Layout()
    if not self.initialized then return end

    local buttons_count = self.preview and 3 or 2
    local button_spacing = 0
    local content_w = self.width - ( buttons_count*BUTTON_W ) - ( buttons_count-1 )*button_spacing
    
    -- Size up our left button
    local nineslice_w, nineslice_h = BUTTON_W + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
    self.left.bg:SetSize( nineslice_w, nineslice_h )
    self.left:LayoutBounds( "left", "center", self.bg )
    
    -- Size up our right button
    local nineslice_w, nineslice_h = BUTTON_W + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
    self.right.bg:SetSize( nineslice_w, nineslice_h )
    self.right:LayoutBounds( "right", "center", self.bg )
    
    -- Size up our dropdown button
    if self.option.dropdown then
        local nineslice_w, nineslice_h = BUTTON_W + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
        self.dropdown.bg:SetSize( nineslice_w, nineslice_h )
        self.dropdown:LayoutBounds( "before", "center", self.right )
    end

    -- Size up our refresh button
    if self.option.twitch_refresh_fn then
        local nineslice_w, nineslice_h = BUTTON_W + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
        self.twitch_refresh.bg:SetSize( nineslice_w, nineslice_h )
        self.twitch_refresh:LayoutBounds( "before", "center", self.right )
    end

    -- Size up our content block
    local nineslice_w, nineslice_h = content_w + self.bg_border*2*self.bg_border_scale, HEIGHT + self.bg_border*2*self.bg_border_scale
    self.content_bg:SetSize( nineslice_w, nineslice_h )
    self.content_outline:SetSize( nineslice_w, nineslice_h )
    self.content_bg:LayoutBounds( "after", "center", self.left )
    self.content_outline:LayoutBounds( "center", "center", self.content_bg )

    -- Position our title
    self.title:LayoutBounds( "left", "above", self.bg ):Offset( 0, 4 )

    -- Calculate text width
    local padding = 20
    local text_w = content_w - padding*2
    self.value_label:SetAutoSize( text_w )
    self.desc_label:SetAutoSize( text_w )

    -- Position our content
    self.desc_label:LayoutBounds( "left", "below", self.value_label ):Offset( 0, 4 )
    self.text_content:LayoutBounds( "left", "center", self.content_bg ):Offset( padding, 0 )

    -- Position the slider
    if self.slider_widget then
        self.slider_widget:SetWidth( content_w - ( self.option.dropdown and 150 or 105 ) )
        self.slider_widget:LayoutBounds( nil, "center", self.value_label )
        self.slider_widget:LayoutBounds( "left", nil, self.content_bg ):Offset( 80, 0 )
    end

    -- Position the colourblind icons
    if self.colourblind_container then
        self.colourblind_container:StackChildrenRow( 6 )
        local w, h = self.colourblind_container:GetSize()
        w, h = w + 40, h
        local nineslice_w, nineslice_h = w + self.colourblind_bg_border*2*self.colourblind_bg_border_scale, h + self.colourblind_bg_border*2*self.colourblind_bg_border_scale
        self.colourblind_bg:SetSize( nineslice_w, nineslice_h )
            :LayoutBounds( "right", "top", self.content_bg )
            :Offset( 0, 0 )
        self.colourblind_container
            :LayoutBounds( "center", "center", self.colourblind_bg )
            :Offset( 5, 3 )
            
    end
    
end

function OptionsScreenSpinner:OnGainFocus()
    self.title:SetGlyphColour( UICOLOURS.WHITE )
    self.value_label:SetGlyphColour( UICOLOURS.WHITE )
    self.content_bg:Bloom( 0.1 )
    self.content_outline:Bloom( 0.3 )
    self:ScaleTo( nil, 1.01, 0.2, easing.outQuad )

    -- Make sure we're visible within the parent scrollpanel
    local scrollpanel = self:GetAncestor( Widget.ScrollPanel )
    if scrollpanel then scrollpanel:EnsureVisible(self) end
end

function OptionsScreenSpinner:OnLoseFocus()
    self.title:SetGlyphColour( UICOLOURS.SUBTITLE )
    self.value_label:SetGlyphColour( UICOLOURS.SUBTITLE )
    self.content_bg:Bloom( 0.05 )
    self.content_outline:Bloom( 0.2 )
    self:ScaleTo( nil, 1, 0.2, easing.outQuad )
end

function OptionsScreenSpinner:HandleControlDown( control, device )
    if self.focus and self.dropdown and control:Has( Controls.Digital.MENU_ACCEPT ) then
        self:TriggerDropdown()
        return true
    end
    if self.focus and self.twitch_refresh and self.twitch_refresh:IsShown() and control:Has( Controls.Digital.MENU_ACCEPT ) then
        self.option.twitch_refresh_fn()
        return true
    end

    return OptionsScreenSpinner._base.HandleControlDown( self, control, device )
end

function OptionsScreenSpinner:TriggerDropdown()
    -- This is called whenever someone clicks the dropdown button
    local x, y = self:GetPos()
    if self.expanded then
        -- Collapse rows
        for k, row in ipairs( self.dropdown_rows ) do
            row:MoveTo( nil, y, 0.1 )
            row:AlphaTo( 0, 0.1, easing.linear, function() row:Hide() end )
        end
        self.dropdown.icon:SetScale( 1, 1 )
        self.expanded = false
        self:AlphaTo( 1, 0.1, easing.linear, function() 
            if self.dropdown_fn then self.dropdown_fn( self.expanded ) end
        end )
    else
        -- Expand rows
        local previous_row = self
        local target_x, target_y = previous_row:GetPos()
        local row_w, row_y = previous_row:GetSize()
        target_y = target_y - row_y - 15
        for k, row in ipairs( self.dropdown_rows ) do
            row:Show()
                :MoveTo( nil, target_y, 0.1 )
                :AlphaTo( 1, 0.1, easing.linear )
            previous_row = row
            target_y = target_y - row_y - 15
        end
        self.dropdown.icon:SetScale( 1, -1 )
        self.expanded = true
        self:AlphaTo( 1, 0.1, easing.linear, function() 
            if self.dropdown_fn then self.dropdown_fn( self.expanded ) end
        end )
    end
end

function OptionsScreenSpinner:SetDropdownRows( rows )
    self.dropdown_rows = rows
    -- Collapse rows
    local x, y = self:GetPos()
    for k, row in ipairs( self.dropdown_rows ) do
        row:LayoutBounds( "right", "center" )
            :SetPos( nil, y )
            :SetTintAlpha( 0 )
            :Hide()
    end
end

function OptionsScreenSpinner:SetDropdownFn( fn )
    self.dropdown_fn = fn
end

function OptionsScreenSpinner:SetOptions( options, selected_idx )
    OptionsScreenSpinner._base.SetOptions( self, options, selected_idx )
end

function OptionsScreenSpinner:SetColour( colour )
    self.title:SetGlyphColour( colour )
    self.value_label:SetGlyphColour( colour )
    local r,g,b,a = HexColour( colour )
    a = 0.2
    self.content_bg:SetTintColour( r,g,b,a )
    self.content_outline:SetTintColour( UICOLOURS.SUBTITLE )
end


function OptionsScreenSpinner:UpdateState()
    OptionsScreenSpinner._base.UpdateState( self )

    if self.enabled then
        if self.option.spinner then
            -- self.options is a list of the possible values of this spinner
            if self.options[self.idx] and type(self.options[self.idx]) == "table" then
                local value = self.options[self.idx]
                self.value_label:SetText( string.upper( value.name or "" ) )
                self.desc_label:SetText( value.desc or "" )
            end
        elseif self.option.slider then
            -- self.options is a table with the ranges between which this spinner can change
            local txt
            if self.options.display_format then
                txt = self.options.display_format( self.idx )
            else
                txt = tostring(self.idx)
            end
            self.value_label:SetText( txt )
            self.desc_label:SetText( self.options.desc or "" )
            self.slider_widget:SetValue( self.idx, true )
            if self.idx >= self.options.range_max then
                self.right:Disable()
            else
                self.right:Enable()
            end
            if self.idx <= self.options.range_min then
                self.left:Disable()
            else
                self.left:Enable()
            end
        end
        self:Layout()
    end

end

function OptionsScreenSpinner:GetValue()
    if self.option.spinner then
        if self.options[ self. idx ] then
            return { index = self.idx, option = self.options[self.idx], data = self.options[self.idx].data }
        else
            -- This can happen if there are NO valid spinner options assigned.
            return table.empty
        end
    else
        return self.idx
    end
end

function OptionsScreenSpinner:SetValue( val, nofn )
    self.idx = val
    if not nofn then self:OnSetIndex() end
    self:UpdateState()
end

function OptionsScreenSpinner:SetIndexByData(data, key_name, nofn)
    for k, option in pairs(self.options) do
        if option == data or (type(option) == "table" and option[key_name or "data"] == data) then
            self:SetIndex(k, nofn)
            return true
        end
    end
end

function OptionsScreenSpinner:Disable()
    self._base.Disable( self )
    self:SetSaturation( 0.1 )
    return self
end

------------------------------------------------------------------------------------------
-- Shows a single content panel

local OptionsScreenContentWidget = class( "Widget.OptionsScreenContentWidget", Widget )

function OptionsScreenContentWidget:init()
    self._base.init( self )

end

function OptionsScreenContentWidget:SetWidth( width )
    
    if self.children then
        for k = 1, #self.children do
            if self.children[k].SetWidth then
                self.children[k]:SetWidth( width )
            else
                print( self.children[k], "Doesn't have the SetWidth function")
            end
        end
    end

    return self
end
