require "ui/screens/debugtoolsmenu"


local default_params = 
{
  emitters = {}
}


local DraggableEmitter = class("Widget.DraggableEmitter", Widget)

function DraggableEmitter:init()
    DraggableEmitter._base.init(self)
    self.hover_check = true
    self:StartUpdating()
    self.img = self:AddChild( Widget.Image(global_images.circle):SetTintColour(0xffaaffff):SetSize(32, 32))

    self.particle_system = self:AddChild(Widget.ParticleSystem())
end

function DraggableEmitter:LoadParams(id)
    self.particle_system:Remove()
    self.particle_system = self:AddChild(Widget.ParticleSystem())
    self.particle_system:LoadParams(id)
    return self
end

function DraggableEmitter:OnGainHover()
    self.img:SetSize(36,36)
end

function DraggableEmitter:OnLoseHover()
    self.img:SetSize(32,32)
end


function DraggableEmitter:HandleControlDown(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        self.dragging = true        
    end
end

function DraggableEmitter:OnUpdate(dt)
    DraggableEmitter._base.OnUpdate(self, dt)
    if self.dragging then
        local x, y = self.parent:TransformFromWorld(TheGame:GetInput():GetMousePos())
        self:SetPos(x, y)
        if not TheGame:GetInput():IsControlDown( Controls.Digital.MENU_ACCEPT ) then
            self.dragging = false
        end
    end
end



local ParticleEditorScreen = class("Screen.ParticleEditorScreen", Widget.Screen)



function ParticleEditorScreen:init()
    ParticleEditorScreen._base.init(self)
    self:SetAnchors("center", "center")
    self.plax = self:AddChild( Widget.Plax() )

    self.particles = self:AddChild(DraggableEmitter())
    self.particles:SetPos(RES_X/4, 0)

    self.undo_stack = {}
    self.undo_ptr = 1
end



function ParticleEditorScreen:DoRender()
    local param_list = {}
    for k,v in pairs(Content.GetAllParticleParams()) do
        table.insert(param_list, k)            
    end
    table.sort(param_list)

    self.plax:RenderToScene()
    
    local ui = TheGame:GetDebug():GetDebugUI()
    self:RenderPanel( ui )

end


function ParticleEditorScreen:SetCurrentEffect( name )
    self.param_id = name
    self.particles:LoadParams(self.param_id)
    self:ClearState()
    self:PushState()
end


function ParticleEditorScreen:SetCurrentBackground( bgname )
    if bgname then 
        self.plax_name = bgname
        self.plax:SetPlaxData(self.plax_name)
    end
end

function ParticleEditorScreen:GetFilename( param_id )
    return string.format( "scripts/content/particles/%s.lua", param_id )
end

function ParticleEditorScreen:Save()
    local filename = self:GetFilename( self.param_id )
    local save_str = "return\n"..serpent.block(self.particles.particle_system.params, {comment=false})
    engine.data:SaveFile( filename, save_str, function(ok) 
        print( string.format( "Write file %s: '%s'", ok and "OK" or "FAILED", filename))
    end )
    local reqname = "content/particles/"..self.param_id
    package.loaded[ reqname ] = nil
    Content.internal.PARTICLE_PARAMS[reqname] = require( reqname )
    Content.ReloadParticles()   -- Reload so that saved values get stored
end

function ParticleEditorScreen:FindUniqueName()
    local count = 1
    local param_id
    while count < 99 do
        param_id = string.format( "%s_copy%s", self.param_id, count > 1 and tostring(count) or "" )
        local filename = self:GetFilename( param_id )
        if not engine.inst:FileExists( filename ) then
            break
        end
        count = count + 1
    end

    return param_id
end

function ParticleEditorScreen:SwitchScreens(fn)
    self:Save()
    fn()
end

function ParticleEditorScreen:RenderPanel( ui )
    local param_list = {}
    for k,v in pairs(Content.GetAllParticleParams()) do
        table.insert(param_list, k)            
    end
    table.sort(param_list)


    local w, h = ui:DisplaySize()
    ui:SetNextWindowSize( w * .333, h, ui.ImGuiSetCond_Once )
    ui:SetNextWindowPos( 0, 0, ui.ImGuiSetCond_Once )

    if ui:Begin( "Particle Editor", ui.MenuBar ) then

        if ui:BeginMenuBar() then
            if ui:BeginMenu( "File" ) then
                
                if ui:MenuItem( "Exit" ) then
                    TheGame:Reset()
                end
                ui:EndMenu()
            end

            ShowDebugToolsMenu(self:GetFE(), ui, function(fn) self:SwitchScreens(fn) end )
            
            ui:EndMenuBar()
        end

        if self:IsOnStack() then
            local plax_ids = Content.GetAllPlaxSorted()
            local snapshot_idx = ui:Combo( "Background", table.arrayfind( plax_ids, self.plax_name) or 1, plax_ids )
            if snapshot_idx and snapshot_idx ~= self.plax_idx then
                self.plax_name = plax_ids[snapshot_idx]
                self.plax:SetPlaxData(self.plax_name)
            end
            ui:Separator()


            local idx = table.arrayfind(param_list, self.param_id)
            local new_idx = ui:Combo("Params", idx or 1, param_list)
            if new_idx and new_idx ~= idx then
                self.param_id = param_list[new_idx]
                self.particles:LoadParams(self.param_id)
                self:ClearState()
                self:PushState()
            end

            if ui:Button("New") then
                self.new_name = ""
                self.new_particles = true
                ui:OpenPopup("NEW_PARTY_POPUP")
            end
            ui:SameLine(nil, 10)

            if ui:Button("Save") then
                self:Save()
            end

            ui:SameLine(nil, 10)

            if ui:Button("Save As...") then
                self.new_name = self:FindUniqueName()
                self.new_particles = false
                ui:OpenPopup("NEW_PARTY_POPUP" )

            end

            ui:Separator()

            if ui:Button("Restart") then
                self.particles.particle_system:Reset()
            end

            self.particles.particle_system:RenderDebugUI(ui)


            if ui:BeginPopup("NEW_PARTY_POPUP", ui.AlwaysAutoResize) then
                ui:Text("Name of new system...")
                local new_name = ui:InputText("", self.new_name)
                
                if new_name then
                    self.new_name = new_name
                end

                if ui:Button("Create New",nil,nil, not (self.new_name and #self.new_name > 0 and not table.arrayfind(param_list, self.new_name))) then
                    
                    local filename = string.format( "scripts/content/particles/%s.lua", self.new_name )
                    local save_str = "return\n"..serpent.block(default_params, {comment=false})
                    engine.data:SaveFile( filename, save_str, function(ok) 
                        print( string.format( "Write file %s: '%s'", ok and "OK" or "FAILED", filename))
                    end )
                    Content.ReloadParticles()
                    self.param_id = self.new_name
                    if self.new_particles then
                        self.particles.particle_system:LoadParams(self.param_id)
                    else
                        self:Save()
                    end

                    self:ClearState()
                    self:PushState()
                    self.new_name = nil

                    ui:CloseCurrentPopup()
                end

                if self.new_name and table.arrayfind(param_list, self.new_name) then
                    ui:SameLine()
                    ui:Text("Already Exists...")
                elseif self.new_name == nil or self.new_name == "" then
                    ui:SameLine()
                    ui:Text("Enter a name first")
                end

                if ui:Button("Cancel") then
                    ui:CloseCurrentPopup()
                end

                ui:EndPopup()
            end
        end
    end
    ui:End()    -- ALWAYS call this, regardless of ui:Begin failing


    local pending = false
    for k,v in ipairs(self.particles.particle_system.emitters) do
        if v.pending then
            pending = true
        end
    end
    local mouse_down = ui:IsMouseDown()
    if not mouse_down and pending then
        for k,v in ipairs(self.particles.particle_system.emitters) do
            v.pending = false
        end
        self:PushState()
    end
end

function ParticleEditorScreen:ClearState( )
    self.undo_stack = {}
    self.undo_ptr = 1
    
end

function ParticleEditorScreen:PushState( )
    local old_sz = #self.undo_stack
    local params = deepcopy(self.particles.particle_system.params)
    self.undo_stack[self.undo_ptr] = params
    self.undo_ptr = self.undo_ptr + 1
    for k = self.undo_ptr, old_sz do
        self.undo_stack[k] = nil
    end
end

function ParticleEditorScreen:Undo()
    if self.undo_ptr > 2 then
        self.undo_ptr = self.undo_ptr - 1
        local params = self.undo_stack[self.undo_ptr-1]
        self.particles.particle_system:SetParams(deepcopy(params))
    end
end

function ParticleEditorScreen:Redo()
    
    if self.undo_ptr < #self.undo_stack +1  then
        
        local params = self.undo_stack[self.undo_ptr]
        self.particles.particle_system:SetParams(deepcopy(params))
        self.undo_ptr = self.undo_ptr + 1
    end
end


function ParticleEditorScreen:HandleControlDown(control)

    if control:Has( Controls.Digital.PLAX_EDIT_REDO ) then
        self:Redo()
        return true

    elseif control:Has( Controls.Digital.PLAX_EDIT_UNDO ) then
        self:Undo()
        return true
    end
end