require "ui/widgets/installed_grafts"

local PerkScreen = class( "Screen.PerkScreen", Widget.Screen )

local COLUMNS = 3
local CELL_W = 480
local CELL_H = 136
local CONTAINER_H = 600

PerkScreen.CONTROL_MAP = Widget.MergeControlMap( Widget.Screen.CONTROL_MAP,
{
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self) 
            self:DoClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL) )
        end,
    },
})

function PerkScreen:init( owner, target_graft_type )
    PerkScreen._base.init( self )

    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.PERK )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Setup character anim display
    self.owner = owner or TheGame:GetGameState():GetPlayerAgent()
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.owner )
    self.character:SetHoverOutlineColour( UICOLOURS.PERK )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "eureka", true )
        end
    end )
    self.character:SetPos( -RES_X*0.45, -350 )
    self.character:Emote( "eureka", true )

    self.perk_scroll = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualMargin( 0 )
        :SetVirtualTopMargin( 100 )
        :SetVirtualBottomMargin( 100 )
        :SetScrollBarOuterMargin( 0 )
        :SetSize( 600, CONTAINER_H )
    self.title = self.perk_scroll:AddScrollChild( Widget.Label("title", FONT_SIZE.SCREEN_TITLE ) )
        :SetText( string.upper( loc.format( LOC"UI.PERK_SCREEN.PERK_SCREEN_TITLE", self.owner:GetName() ) ) )
        :SetGlyphColour( UICOLOURS.PERK )
        :LeftAlign()
        :Bloom(0.15)
    self.perk_container = self.perk_scroll:AddScrollChild( Widget() )
    
    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:DoClose() end ) )
    self.close_button:SetIcon( global_images.close )
    self.close_button:LayoutBounds( "left", "above", 60, 60 )
    
    self:Refresh()
end

function PerkScreen:DoClose()
    AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
    self:FadeOut()
end

function PerkScreen:Refresh()

    self.owner = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    self.perk_container:DestroyAllChildren()

    local grafts = self.owner.graft_owner:GetGrafts( GRAFT_TYPE.PERK )
    for i, graft in ipairs( grafts ) do
        local widget = self.perk_container:AddChild( Widget.PerkInfoWidget( CELL_W, CELL_H, graft:GetDef() ) )
    end

    self:Layout()

    return self
end

function PerkScreen:Layout()
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    local left_offset = RES_X*0.3 -- How much the scroll panel should cross to the left of the screen center
    local left_padding = 100 -- How much space should there be between the left edge of the scroll area and the first mutator (allows space for glow/scaling without clipping)
    self.title:LayoutBounds( "left", "top", 0, 0 )
        :Offset( left_padding, 0 )
    self.perk_container:LayoutChildrenInGrid( COLUMNS, SPACING.M1 )
        :LayoutBounds( "left", "below", self.title )
        :Offset( -SPACING.M1*0.75, -SPACING.M1/2 )

    -- Get MainOverlay's top bar height
    local overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    local overlay_topbar_height = overlay and overlay:GetTopBarHeight() or 0

    self.perk_scroll:SetSize( RES_X/2 + left_offset + left_padding, self.screen_h - overlay_topbar_height )
    self.perk_scroll:SetPos( RES_X/4 - left_offset/2 - left_padding/2, -overlay_topbar_height/2 )

    return self
end

function PerkScreen:OnOpen()
    PerkScreen._base.OnOpen( self )

    local anim_duration = 0.7
    local horizontal_movement = -90

    -- Prepare for animation by getting current values
    local content_x, content_y = self.content:GetPos()

    -- Hide objects until they're animated in
    self.content:SetTintAlpha(0)

    -- Setup and start animation
    self.fadein_anim = self:RunUpdater(
        Updater.Parallel{
            -- Fade and move in the coin container
            Updater.Ease( function(v) self.content:SetTintAlpha(v) end, 0, 1, anim_duration, easing.outQuad),
            Updater.Ease2D( function(x,y) self.content:SetPos(x,y) end, content_x - horizontal_movement, content_y, content_x, content_y, anim_duration, easing.outQuad),
        }
    )

end

function PerkScreen:GetDefaultFocus()
    return self.perk_container and self.perk_container.children and #self.perk_container.children>0 and self.perk_container.children[1]
end

function PerkScreen:OnControlModeChange( cm, device_id )
    PerkScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.perk_scroll:ShowBar( SCROLLBAR.IF_NEEDED )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.perk_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.perk_scroll:ShowBar( SCROLLBAR.NEVER )
        self.bottom_left:Hide()
    end

end

function PerkScreen:OnScreenModeChange( sm )
    PerkScreen._base.OnScreenModeChange( self, sm )

    if sm == SCREEN_MODE.MONITOR then
        COLUMNS = 3
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        COLUMNS = 2
    end

    -- self.character:SetPos( -RES_X*0.27*LAYOUT_SCALE[sm], -350 )
    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.title:SetLayoutScale( LAYOUT_SCALE[sm] )
    for k, v in pairs(self.perk_container.children) do
        v:SetLayoutScale( LAYOUT_SCALE[sm] )
    end

    self:Layout()
end

function PerkScreen:HandleControlDown(control, device)
    if control:Has( Controls.Digital.SCROLL_DOWN ) then
        self.perk_scroll:ScrollDown()
        return true
    elseif control:Has( Controls.Digital.SCROLL_UP ) then
        self.perk_scroll:ScrollUp()
        return true
    end
end

function PerkScreen:HandleControlUp( control, device )
    if control:Has( Controls.Digital.MENU_CANCEL ) then
        AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
        self:FadeOut()
    end

    return PerkScreen._base.HandleControlUp( self, control, device )
end

function PerkScreen:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function PerkScreen:OnOpen()
    PerkScreen._base.OnOpen( self )
    self.screen_loop = AUDIO:CreateEventInstance("event:/ui/grafts_menu/open")
    self.screen_loop:Start()
end

function PerkScreen:OnClose()
    PerkScreen._base.OnClose( self )

    self.screen_loop:Stop()
    self.screen_loop:Release()
    self.screen_loop = nil
end

-------------------------------------------------------------------------
-- A single mutator

local PerkInfoWidget = class( "Widget.PerkInfoWidget", Widget )

local ICON_W = 168
local TEXT_LEFT_MARGIN = 192

function PerkInfoWidget:init( w, h, def, checked )
    PerkInfoWidget._base.init( self )
    self.features = {}

    self.def = def
    self.graft = GraftInstance( def.id )
    
    self.hitbox = self:AddChild( Widget.SolidBox( w, h, 0x00FFFF00 ) )
    self.contents = self:AddChild( Widget() )
    self.bg = self.contents:AddChild( Widget.Panel( engine.asset.Texture("UI/dropdown_inner.tex" ) ) )
        :SetTintColour( UICOLOURS.PERK )
        :SetTintAlpha( 0 )
        :SetBloom( 0.1 )
    self.glow = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( UICOLOURS.PERK )
        :SetTintAlpha( 0.2 )
        :SetHiddenBoundingBox( true )
        :SetSize( ICON_W*1.5, ICON_W*1.5 )
    self.icon = self.contents:AddChild( Widget.PerkButton() )
        :SetWidth( ICON_W )
        :Refresh( def.id, true )
        -- :Bloom( 0.15 )
    self.text_content = self.contents:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, def:GetLocalizedName() ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.ITEM_NAME*LINE_HEIGHT_TITLE_MULT )
        :SetGlyphColour( UICOLOURS.PERK )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.graft:GetDesc( self ) ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )

    self:SetSize( w, h )
    self:SetFocusable( true )

    self:SetToolTipClass( Widget.TooltipGraft )
        :SetToolTip( self.graft )
        :ShowToolTipOnFocus( true )
end

function PerkInfoWidget:SetSize( w, h )
    self.title:SetAutoSize( w-TEXT_LEFT_MARGIN-12 )
    self.text:SetAutoSize( w-TEXT_LEFT_MARGIN-12 )
    self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    self.hitbox:SetSize( w, h )
    self.bg:SetSize( w, h )
    self.icon:LayoutBounds( "left", "top", self.hitbox ):Offset( 12, -12 )
    self.glow:LayoutBounds( "center", "center", self.icon ):Offset( 0, 0 )
    self.text_content:LayoutBounds( "left", "top", self.hitbox ):Offset( TEXT_LEFT_MARGIN, -8 )
    return self
end

function PerkInfoWidget:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function PerkInfoWidget:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

function PerkInfoWidget:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.bg:AlphaTo( 0.09, 0.1 )
        self.glow:ScaleTo( nil, 1.2, 0.1 )
        self.contents:ScaleTo( nil, 1.05, 0.2 )
    else
        self.bg:AlphaTo( 0, 0.1 )
        self.glow:ScaleTo( nil, 1, 0.1 )
        self.contents:ScaleTo( nil, 1, 0.3 )
    end
end

function PerkInfoWidget:OnGainFocus()
    PerkInfoWidget._base.OnGainFocus(self)
    self:UpdateImage()
end

function PerkInfoWidget:OnLoseFocus()
    PerkInfoWidget._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PerkInfoWidget:OnGainHover()
    PerkInfoWidget._base.OnGainHover(self)
    self:UpdateImage()
end

function PerkInfoWidget:OnLoseHover()
    PerkInfoWidget._base.OnLoseHover(self)
    self:UpdateImage()
end

------------------------------------------------------------------------------------------
-- Shows a group of Perks

local PerkGroup = class("Widget.PerkGroup", Widget)

function PerkGroup:init(label, perks, fn, locked)
    Widget.init(self)
    
    self.perks = perks or {}
    self.perk_widgets = {}
    self.locked = locked
    self.columns = 7
    self.fn = fn

    self.text_container = self:AddChild( Widget() )
    self.label = self.text_container:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_SUBTITLE, label ) )
        :SetGlyphColour( UICOLOURS.PERK )
        :Bloom( 0.1 )
    self.locktext = self.text_container:AddChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT ) )
        :LeftAlign()
        :Bloom( 0.1 )
    
    self.perk_root = self:AddChild(Widget())
    
    for k, perk in ipairs(self.perks) do
        local earned = TheGame:GetGameProfile() and TheGame:GetGameProfile():HasPerkUnlocked(perk.id)
        local perk_widget = Widget.PerkWidget( perk.id )
            :PrepareGlowAnimation() -- Prevents the glow from animating straight away
            :Refresh( self.locked )

        self:AddWidget( perk_widget )
    end
    self:SetLocked( self.locked )

    self:Layout()
end

function PerkGroup:ClearWidgets()
    self.perks = {}
    self.perk_widgets = {}

    self.perk_root:DestroyAllChildren()

    return self
end

function PerkGroup:SetWidgets( perk_widgets )
    for k, perk_widget in ipairs( perk_widgets ) do
        self:AddWidget( perk_widget )
    end
    return self
end

function PerkGroup:AddWidget( perk_widget )
    self.perk_root:AddChild( perk_widget )
    self.perk_widgets[ perk_widget:GetPerkId() ] = perk_widget
    perk_widget:SetClickSoundEvent( self.locked and SoundEvents.button_invalid or SoundEvents.button_click )
    return perk_widget
end

function PerkGroup:HasPerks()
    return not self.perk_root:IsEmpty()
end

function PerkGroup:IsLocked() 
    return self.locked
end

function PerkGroup:SetLocked( locked ) 
    self.locked = locked

    for k, perk_widget in pairs( self.perk_widgets ) do
        perk_widget:SetClickSoundEvent( self.locked and SoundEvents.button_invalid or SoundEvents.button_click )
        perk_widget:SetOnClickFn( not self.locked and function() self.fn(perk_widget:GetPerk()) end )
        perk_widget:Refresh( self.locked )
    end

    return self
end

function PerkGroup:SetLockedText( locked_text ) 
    if locked_text then
        self.locktext:SetText(locked_text):Show()
        self.locktext:LayoutBounds( "left", "below", self.label ):Offset( 0, SPACING.TITLE_DESC )
    else
        self.locktext:Hide()
    end
    return self
end

function PerkGroup:SetColumns( num )
    self.columns = num
    self:Layout()
    return self
end

function PerkGroup:Layout()
    self.perk_root:LayoutChildrenInGrid( self.columns, {h=15, v=15})
    if self.locktext:IsShown() then
        self.locktext:LayoutBounds( "left", "below", self.label ):Offset( 0, SPACING.TITLE_DESC )
    else
        self.locktext:LayoutBounds( "left", "center", self.label )
    end
    self.perk_root:LayoutBounds( "left", "below", self.text_container ):Offset( 0, -SPACING.M1 )
    return self
end

function PerkGroup:GetDefaultFocus()
    for k, w in ipairs( self.perk_root.children ) do
        return w
    end
end

function PerkGroup:PrepareAnimation()

    self.label:SetTintAlpha( 0 )
    self.locktext:SetTintAlpha( 0 )

    for i, widget in ipairs( self.perk_root.children ) do
        widget:SetTintAlpha( 0 )
    end

    return self
end

function PerkGroup:GetAnimateUpdater( initial_delay )

    -- Go through the perks and animate each of them in
    local timing = 0.03
    local spacing = 30
    local label_x, label_y = self.label:GetPos()
    local locktext_x, locktext_y = self.locktext:GetPos()
    local sequence = Updater.Series{
        Updater.Wait( initial_delay or 0 ),
        Updater.Parallel{
            Updater.Ease( function(v) self.label:SetPos(v, label_y) end, label_x+spacing, label_x, timing*3, easing.outQuad ),
            Updater.Ease( function(v) self.label:SetTintAlpha(v) end, 0, 1, timing*3, easing.outQuad )
        },
        Updater.Parallel{
            Updater.Ease( function(v) self.locktext:SetPos(v, locktext_y) end, locktext_x+spacing*2, locktext_x, timing*3, easing.outQuad ),
            Updater.Ease( function(v) self.locktext:SetTintAlpha(v) end, 0, 1, timing*3, easing.outQuad )
        }
    }
    for i, widget in ipairs( self.perk_root.children ) do
        local pos_x, pos_y = widget:GetPos()
        sequence:Add( Updater.Parallel{
            Updater.Ease( function(v) widget:SetPos(v, nil) end, pos_x+spacing, pos_x, timing, easing.outQuad ),
            Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, timing, easing.outQuad )
        } )

        -- Then animate in the glows for the completed perks
        if widget:IsPlayed() then
            sequence:Add( Updater.Do( function() widget:AnimateInGlow() end ) )
            sequence:Add( Updater.Ease( function(v) widget:SetScale(v,v) end, 1, 1.1, 0.1, easing.outQuad ) )
            sequence:Add( Updater.Ease( function(v) widget:SetScale(v,v) end, 1.2, 1, 0.1, easing.outQuad ) )
        end
    end

    return sequence
end
