require "ui/widgets/optionpicker"
require "ui/widgets/screen"
require "ui/widgets/tooltip_graft"
require "ui/screens/replacegraftscreen"

-----------------------------------------------------------------------
--

local PickGraftScreen = class("Screen.PickGraftScreen", Widget.Screen)

local PANEL_H = 700
local SHADOW_OFFSET = 480

local SLOT_W = 440
local SLOT_H = 460

local function GetFocusGraft()
    local focus = TheGame:FE():GetFocusWidget()
    if is_instance( focus, Widget.PickGraftOption ) then
        return focus.graft
    end
end

PickGraftScreen.CONTROL_MAP = {
    {
        control = Controls.Digital.GAMEPAD_X,
        
        test = function(self)
            return GetFocusGraft() ~= nil and not self.show_reticule
        end,
        
        fn = function(self)
            self:OnSelectGraft( GetFocusGraft() )
            return true
        end,
        
        hint = function(self, left, right)
            local graft_id = GetFocusGraft().id
            table.insert(right, loc.format ( LOC"UI.CONTROLS.DRAFT_GRAFT", Controls.Digital.GAMEPAD_X, graft_id))
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_B,
        test = function(self)
            if self.money_btn and not self.show_reticule then
                return true
            end
        end,
        fn = function(self)
            self.money_btn:Click()
            return true
        end,
        hint = function(self, left, right)
            if self.skip_shills_reward and self.skip_shills_reward > 0 then
                local is_daily = TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY 
                if is_daily then
                    table.insert(right, loc.format(LOC"UI.CONTROLS.TAKE_POINTS", Controls.Digital.MENU_CANCEL, self.skip_shills_reward))
                else
                    table.insert(right, loc.format(LOC"UI.CONTROLS.TAKE_CARD_MONEY", Controls.Digital.MENU_CANCEL, self.skip_shills_reward))
                end
            else
                table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
            end
        end
    },
    {
        control = Controls.Digital.GAMEPAD_RT,
        test = function(self)
            local focus = self:GetFE():GetFocusWidget() 
            return focus and focus.ShowDetails ~= nil
        end,
        fn = function(self)
            local focus = self:GetFE():GetFocusWidget()
            focus:ShowDetails()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.GRAFT_INFO", Controls.Digital.GAMEPAD_RT))
        end
    },
    {
        control = Controls.Digital.PAUSE, 
        fn = function(self)
            self:GetFE():InsertScreen(Screen.PauseMenu())
            return true
        end,
    },
}

function PickGraftScreen:init( grafts, no_skip, on_done, no_money_for_skip, money )
    PickGraftScreen._base.init( self )
    self.screen_mix_event = "event:/mix_events/Popup"
    
    self.is_popup = true
    self:SetupUnderlay( true )
    self:SetAnchors("center", "center")

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Backing panel
    self.panel = self:AddChild( Widget.WidePanel( PANEL_H ) )
        :SetColour( UICOLOURS.GRAFT )
        :SetIcon( engine.asset.Texture( "UI/ic_draft_graft.tex") )
        :SetTitles( LOC"UI.GRAFT_SCREEN.TITLE", "" )

    self.chosen_graft = nil
    self.grafts = grafts
    self.on_done = on_done
    self.select_then_continue_mode = false -- If true, you'll have to click a graft, then click the continue button

    self.options = self:AddChild( Widget() )
    self.shadows = self:AddChild( Widget() )
    self.warnings = self:AddChild(Widget())
    self.buttons = {}
    for i, graft in ipairs(grafts) do
        local widget = self.options:AddChild( Widget.PickGraftOption( graft, self ) )

        -- Add a shadow
        widget.shadow = self.shadows:AddChild( Widget.Image( engine.asset.Texture("UI/negotiation_iconglow.tex") ) )
            :SetHiddenBoundingBox( true )
            :SetSize( 500, 100 )
            :SetTintColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
            :SetPos( 0, -SHADOW_OFFSET )
            :SetTintAlpha( 0.8 )

        TheGame:GetGameProfile():SetSeenGraft( graft.id )
        table.insert(self.buttons, widget)
    end
    self.options:StackChildrenRow( 30 )

    -- Align the shadows
    for k, widget in ipairs( self.options.children ) do
        widget.shadow:LayoutBounds( "center", nil, widget )
    end

    if not no_skip then
        self.no_money_for_skip = no_money_for_skip
        self:SetMoneyButton( money )
    end

    -- Continue button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.button_continue = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.DAY_OVER_SCREEN.BUTTON_CONTINUE", function() self:OnClickContinue() end, 300 ) )
        :SetIcon( global_images.accept )
        :LayoutBounds( "right", "above", -60, 60 )
        :Hide()
        :SetToolTip( LOC"UI.OPTIONS_SCREEN.CONFIRM_SELECT_TOOLTIP" )

    self:Layout()

    self:SetSounds( "event:/ui/choose_grafts_screen/open", "event:/ui/choose_grafts_screen/close" )
    self:SetMusicEvent( TheGame:LookupPlayerMusic( "deck_music" ))
end

function PickGraftScreen:OnClickContinue( graft )
    graft = graft or self.selected_graft
    if graft then
        if TheGame:GetGameState():GetPlayerAgent().graft_owner:CanInstallGraft(graft.id) then
            self.chosen_graft = graft
            TheGame:GetGameState():GetPlayerAgent().graft_owner:AddGraft(graft)
            self:GetFE():PopScreen( self )
        else
            if graft:GetDef().type == GRAFT_TYPE.COIN then
                local graft_owner = TheGame:GetGameState():GetPlayerAgent().graft_owner
                graft_owner:ReplaceGraft( graft_owner:FindGraft(function(graft) return graft:GetDef().type == GRAFT_TYPE.COIN end), graft )
                self.chosen_graft = graft
                self:GetFE():PopScreen( self )
            else
                local screen = Screen.ReplaceGraftScreen( graft, 
                    function(replaced) 
                        if replaced then
                            self.chosen_graft = graft
                            self:GetFE():PopScreen( self )
                        end
                    end )
                TheGame:FE():InsertScreen( screen )
            end
        end
    end
end

function PickGraftScreen:OnSelectGraft( graft )
    AUDIO:PlayEvent( "event:/ui/choose_grafts_screen/select_graft" )
    if self.select_then_continue_mode then
        if graft == self.selected_graft then
            -- This graft was already selected, and was confirmed again. Just draft it
            self:OnClickContinue()
        else
            -- We're in select-then-continue mode, and the selected graft isn't the clicked one
            -- Select this one
            self.selected_graft = graft
            -- Unselect options
            for k, widget in ipairs( self.options.children ) do
                widget:SetSelected( widget.graft == self.selected_graft )
            end
            -- And show the confirm button
            self.button_continue:SetText( loc.format( LOC "UI.GRAFT_SCREEN.BUTTON_CONTINUE", graft:GetName() ))
            self.button_continue:SizeToLabel( 10, 10 )
            self.button_continue:LayoutBounds( "right", "above", -60, 60 ):Show()
        end
    elseif graft then
        self.selected_graft = graft
        self:OnClickContinue()
    end
end

function PickGraftScreen:OnOpen()
    PickGraftScreen._base.OnOpen( self )
    self.options:SetTintAlpha( 0 )
    self.shadows:SetTintAlpha( 0 )

    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    -- self.panel:FadeIn( function()
        self.options:AlphaTo( 1, 0.3 )
        self.shadows:AlphaTo( 1, 0.3 )
    -- end )
end

function PickGraftScreen:OnControlModeChange( cm, device_id )
    PickGraftScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.select_then_continue_mode = false or engine.inst:GetSetting("gameplay", "confirm_select", true)
        self.options:LayoutBounds( "center", "center", self.panel )
            :Offset( 0, 20 )
        if self.money_btn then self.money_btn:Show() end
        self.bottom_right:Show()
        -- Unselect options
        for k, widget in ipairs( self.options.children ) do
            widget:SetSelected( false )
            widget:SetOnClickFn( function() self:OnSelectGraft( widget.graft ) end )
        end
        self.selected_graft = nil
    end
    if cm == CONTROL_MODE.TOUCH then
        self.select_then_continue_mode = true
        self.options:LayoutBounds( "center", "center", self.panel )
            :Offset( 0, 20 )
        if self.money_btn then self.money_btn:Show() end
        self.bottom_right:Show()
        -- Unselect options
        for k, widget in ipairs( self.options.children ) do
            widget:SetSelected( false )
            widget:SetOnClickFn( function() self:OnSelectGraft( widget.graft ) end )
        end
        self.selected_graft = nil
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.select_then_continue_mode = false
        self.options:LayoutBounds( "center", "center", self.panel )
            :Offset( 0, -20 )
        if self.money_btn then self.money_btn:Hide() end
        self.bottom_right:Hide()
        -- Unselect options
        for k, widget in ipairs( self.options.children ) do
            widget:SetSelected( false )
            widget:SetOnClickFn( nil )
        end
        self.selected_graft = nil
    end
end

function PickGraftScreen:OnScreenModeChange( sm )
    PickGraftScreen._base.OnScreenModeChange( self, sm )

    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    
    local new_option_h, new_panel_h
    if sm == SCREEN_MODE.MONITOR then
        new_option_h = SLOT_H
        new_panel_h = PANEL_H
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        new_option_h = SLOT_H + 60
        new_panel_h = PANEL_H + 120
    end

    -- Resize options
    for k, widget in ipairs( self.options.children ) do
        widget:SetHeight( new_option_h )
    end
    -- Resize container panel
    self.panel:SetHeight( new_panel_h )

    if self.money_btn then self.money_btn:SetLayoutScale( LAYOUT_SCALE[sm] ) end

    self:Layout()
end

function PickGraftScreen:HandleControlUp(control, device)
    local hotkey_idx = UIHelpers.FindHotkeyIdx( control )
    if hotkey_idx and self.options.children[ hotkey_idx ] then
        for k, widget in ipairs( self.options.children ) do
            if k == hotkey_idx then
                self:OnSelectGraft( widget.graft )
                return true
            end
        end
    end
end

function PickGraftScreen:RefreshStreamerMode()
    self.bottom_right:SetPos( 0, 0 )
    if GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        self.bottom_right:SetPos( -360, 0 )
    end
end

function PickGraftScreen:Layout()
    if self.money_btn then self.money_btn:LayoutBounds( "center", "bottom", self.panel ):Offset( 0, 43 ) end
end

function PickGraftScreen:OnClose()
    PickGraftScreen._base.OnClose( self )
    
    if self.chosen_graft then
        self:SendMetrics( self.chosen_graft.id )
    end

    if self.on_done then
        self.on_done( self.chosen_graft )
    end
end

function PickGraftScreen:SendMetrics( selected_id )
    local graft_choices = {}
    for i, graft in ipairs(self.grafts) do
        table.insert(graft_choices, {ID = graft.id})
    end

    SendMetricsData("DRAFTED_GRAFT", {
        SELECTED_ID = selected_id,
        GRAFT_CHOICES = graft_choices,
        PLAYER_DATA = TheGame:GetGameState():GetPlayerState(),
    })
end

function PickGraftScreen:SetMoneyButton( money )
    if self.no_money_for_skip then
        self.skip_shills_reward = 0
        
        self.money_btn = self:AddChild( Widget.IconButton( LOC"UI.GRAFT_SCREEN.SKIP_GRAFT", nil, 230 ) )
        self.money_btn:SetFn( function( )
            self:GetFE():PopScreen( self )
        end )

    else
        self.skip_shills_reward = money or GRAFT_SKIP_REWARD

        local title = LOC"UI.DRAFT_SCREEN.MONEY_BTN_TITLE"
        local subtitle

        local is_daily = TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY 
        if is_daily then
            subtitle = loc.format( LOC"UI.DRAFT_SCREEN.MONEY_BTN_SUBTITLE_POINTS", Content.GetScoreComponent("SKIPPED_GRAFT_DRAFT").value )
        else
            subtitle = loc.format( LOC"UI.DRAFT_SCREEN.MONEY_BTN_SUBTITLE", self.skip_shills_reward )
        end

        self.money_btn = self:AddChild( Widget.IconButton( LOC"UI.GRAFT_SCREEN.SKIP_GRAFT", nil, 230 ) )
        self.money_btn:SetFn( function( )
            if is_daily then
                TheGame:GetGameState():AddScore("SKIPPED_GRAFT_DRAFT")
            else
                AUDIO:PlayEvent( "event:/ui/select_cards/aquired_money" )
                TheGame:GetGameState():GetCaravan():AddMoney( self.skip_shills_reward )                
            end

            self:SendMetrics( "MONEY" )

            self:GetFE():PopScreen( self )
        end )

        self.money_btn:SetIcon( global_images.shills )
            :SetSize( 300, 50 )
            :SetTextSize( 24 )
            -- :SetScale( 1.4, 1.4 )
        self.money_btn:SetText( title, subtitle )
    end
    self.money_btn:LayoutBounds( "center", "bottom", self.panel )
        :Offset( 0, 43 )
end

function PickGraftScreen:HarvestRobotOptions()
    self:AddRobotOption(self.money_btn)
    if self.button_continue:IsShown() then
        self:AddRobotOption(self.button_continue)
    else
        for k, widget in ipairs(self.buttons) do
            self:AddRobotOption(widget)
        end
    end
end

function PickGraftScreen:GetDefaultFocus()
    if self.last_focus and self.last_focus:GetScreen() == self then
        return self.last_focus
    end
    return self.options and self.options.children and #self.options.children > 0 and self.options.children[1]
end

-----------------------------------------------------------------------
-- A single clickable graft

local PickGraftOption = class("Widget.PickGraftOption", Widget.Clickable)

local PADDING = 20
local GRAFT_SIZE = 180
local bg_border = 70
local bg_border_scale = 0.5

function PickGraftOption:init(graft)
    PickGraftOption._base.init( self )
    self.close_sound = "event:/ui/choose_grafts_screen/close"
    
    self.width = SLOT_W
    self.height = SLOT_H
    self.graft = graft
    self.features = {}
    self:SetToolTipClass( Widget.TooltipKeywords ):SetToolTip( self.features )
    self:SetOnRightClickFn( function() self:ShowDetails() end )   
    self:DisableAnimateScales()

    -- Background panel
    local nineslice_w, nineslice_h = self.width + bg_border*2*bg_border_scale, self.height + bg_border*2*bg_border_scale
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/graft_bg.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )

    -- Gradient
    self.gradient = self:AddChild( Widget.Panel( engine.asset.Texture("UI/graft_gradient.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
        :SetTintAlpha( 1 )

    -- Graft icon
    self.graft_frame = self:AddChild( Widget.Panel( engine.asset.Texture("UI/grafts_slot.tex") ) )
        :SetSize( GRAFT_SIZE, GRAFT_SIZE )
        :SetTintColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
        :Bloom( 0.25 )
        :Offset( 0, 30 )
    self.graft_img = self:AddChild( Widget.Image( graft:GetImage() ) )
        :SetSize( GRAFT_SIZE*0.9, GRAFT_SIZE*0.9 )
        :Bloom( 0.08 )
        :Offset( 0, 30 )

    -- Graft details
    self.title = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_NAME ) )
        :SetWordWrap( true )
        :SetAutoSize( self.width - PADDING*2 )
        :LeftAlign()
        :Bloom( 0.1 )
        :SetText( graft:GetName() )
        :SetGlyphColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )

    if graft:GetType() == GRAFT_TYPE.COMBAT then
        self.title:SetGlyphColour( UICOLOURS.FIGHT )
    elseif graft:GetType() == GRAFT_TYPE.NEGOTIATION then
        self.title:SetGlyphColour( UICOLOURS.NEGOTIATION )
    end

    local rarity_color = CARD_RARITY_COLOURS[ self.graft:GetRarity() or CARD_RARITY.UNIQUE ]
    local rarity_icon_string = GetCardRarityIconString( self.graft:GetRarity() )
    local subtitle_text = TheGame:Str( string.format( "UI.GRAFT_COLLECTION.%s.%s", self.graft:GetRarity(), self.graft:GetType() ))

    self.subtitle = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE - 4 ) )
        :SetWordWrap( true )
        :SetAutoSize( self.width - PADDING*2 )
        :LeftAlign()
        :SetGlyphColour( rarity_color )
        :SetText( rarity_icon_string .. subtitle_text )

    self.desc = self:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ) )
        :SetWordWrap( true )
        :SetAutoSize( self.width - PADDING*2 )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom(0.1)
        :SetText( graft:GetDesc( self ) )

    -- Overlay panel
    self.overlay = self:AddChild( Widget.Panel( engine.asset.Texture("UI/graft_overlay.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( graft:GetType() == GRAFT_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT )
        :Bloom( 0.2 )

    -- Class label under graft
    local icon, colour, text
    if graft:GetType() == GRAFT_TYPE.NEGOTIATION then
        icon = engine.asset.Texture ("UI/ic_subcard_negotiation.tex")
        colour = UICOLOURS.NEGOTIATION
        text = LOC"UI.GRAFT_SCREEN.CLASS_LABEL_NEGOTIATION"
    else
        icon = engine.asset.Texture ("UI/ic_subcard_fight.tex")
        colour = UICOLOURS.FIGHT
        text = LOC"UI.GRAFT_SCREEN.CLASS_LABEL_BATTLE"
    end
    self.class_label = self:AddChild( Widget.SubCardLabel( 290, 48, icon, colour, text ) )
        :SetHiddenBoundingBox( true )

    self:Layout()
end

function PickGraftOption:SetSelected( selected )
    self.selected = selected
    self:UpdateImage()
    return self
end

function PickGraftOption:ShowDetails()
    local details = Screen.GraftsCompendiumDetailsPopup()
    TheGame:FE():PushScreen( details )
    details:ShowGraft( self.graft )
end

function PickGraftOption:UpdateImage()
    local timing = 0.2
    if self.enabled and ( self.hover or self.focus or self.selected ) then
        self:ScaleTo( nil, self.selected and 1.08 or 1.05, 0.2, easing.outQuad )
        if self.shadow then self.shadow:AlphaTo( 1, timing ) self.shadow:ScaleTo( nil, 1.05, timing ) end
        self.title:Bloom(0.2)
        self.overlay:Bloom(0.4)
        self.graft_img:ScaleTo( nil, 1.05, 0.1, easing.outQuad )
        self.graft_frame:ScaleTo( nil, 1.1, 0.1, easing.outQuad )
        self.graft_frame:Bloom(0.3)
    else
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
        if self.shadow then self.shadow:AlphaTo( 0.6, timing ) self.shadow:ScaleTo( nil, 1, timing ) end
        self.title:Bloom(0.1)
        self.overlay:Bloom(0.2)
        self.graft_img:ScaleTo( nil, 1, 0.1, easing.outQuad )
        self.graft_frame:ScaleTo( nil, 1, 0.1, easing.outQuad )
        self.graft_frame:Bloom(0.25)
    end
end

function PickGraftOption:OnGainFocus()
    PickGraftOption._base.OnGainFocus(self)
    self:UpdateImage()
    self:GetScreen().last_focus = self
end

function PickGraftOption:OnLoseFocus()
    PickGraftOption._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PickGraftOption:OnGainHover()
    PickGraftOption._base.OnGainHover(self)
    self:UpdateImage()
end

function PickGraftOption:OnLoseHover()
    PickGraftOption._base.OnLoseHover(self)
    self:UpdateImage()
end

function PickGraftOption:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function PickGraftOption:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

function PickGraftOption:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function PickGraftOption:OnControlModeChange( cm, device_id )
    PickGraftOption._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
    end
    if cm == CONTROL_MODE.TOUCH then
    end
    if cm == CONTROL_MODE.GAMEPAD then
    end
end

function PickGraftOption:OnScreenModeChange( sm )
    PickGraftOption._base.OnScreenModeChange( self, sm )
    
    -- ccc: too big for other languages
    -- self.title:SetFontSize( FONT_SIZE.ITEM_NAME*LAYOUT_SCALE[sm] )
    -- self.subtitle:SetFontSize( FONT_SIZE.ITEM_SUBTITLE*LAYOUT_SCALE[sm] )
    self.desc:SetFontSize( FONT_SIZE.ITEM_DESC*LAYOUT_SCALE[sm] )
    -- self.header:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end

function PickGraftOption:SetHeight( height )
    self.height = height or SLOT_H
    local nineslice_w, nineslice_h = self.width + bg_border*2*bg_border_scale, self.height + bg_border*2*bg_border_scale
    self.bg:SetSize( nineslice_w, nineslice_h )
    self.gradient:SetSize( nineslice_w, nineslice_h )
    self.overlay:SetSize( nineslice_w, nineslice_h )
    self:Layout()
    return self
end

function PickGraftOption:Layout()
    self.title:LayoutBounds( "left", "top", self.bg ):Offset( PADDING, -PADDING )
    self.subtitle:LayoutBounds( "left", "below", self.title ):Offset( 0, 4 )
    self.desc:LayoutBounds( "left", "bottom", self.bg ):Offset( PADDING, PADDING )
    self.class_label:LayoutBounds( "center", "below", self.overlay ):Offset( 0, -8 )
    return self
end

local ReticuleMixin = require "ui/screens/ReticuleMixin"
ReticuleMixin:Mixin( PickGraftScreen )
