require "ui/widgets/screen"
require "ui/widgets/plax"
require "ui/widgets/checkbox"
require "ui/widgets/polyline"

require "ui/widgets/slider"
require "ui/screens/plaxdemoscreen"
require "ui/screens/fightscreen"
require "ui/screens/debugtoolsmenu"

local debug_menus = require "debug/debug_menus"
local DebugUtil = require "debug/debug_util"
local posedefs = require "content/character_pose_defs"
local path = require "util/filepath"
local sColourSettings = require "content/colour_settings"

local GRIFTLANDS_MOD = "Griftlands"

local NODE_TYPE = Widget.Plax.NODE_TYPE
local DEFAULT_PROP_TEX = "debug/default_plax_prop.tex"

local day_phases = {DAY_PHASE.DAY, DAY_PHASE.NIGHT }
local day_phase_txt = {"Day", "Night"}

local game_modes = {GAME_MODE.CONVERSATION, GAME_MODE.BATTLE }
local game_mode_txt = {"Conversation", "Battle"}

local ASPECT_FRAMING = MakeBitField{ "CONVO", "BATTLE" }



---------------------------------------------------------------------
-- Plax Editor.

local PlaxEditor = class("Screen.PlaxEditor", Widget.Screen)
PlaxEditor.IS_EDITOR = true

local DATA_DIR = path.normslashes(path.normpath( path.join(BASE_PATH, "../data")))

local assets = 
{
    layer = engine.asset.Texture(   "Tools/treeview_layer.tex"   ),
    anim = engine.asset.Texture(   "Tools/treeview_anim.tex"   ),
    img = engine.asset.Texture(   "Tools/treeview_image.tex"   ),
    character = engine.asset.Texture(   "Tools/treeview_character.tex"   ),

    hidden = engine.asset.Texture("Tools/treeview_hidden.tex"),
    visible = engine.asset.Texture("Tools/treeview_visible.tex"),
    locked = engine.asset.Texture("Tools/treeview_locked.tex"),
    unlocked = engine.asset.Texture("Tools/treeview_unlocked.tex"),
    delete = engine.asset.Texture("Tools/treeview_delete.tex"),

    day = engine.asset.Texture("Tools/treeview_day.tex"),
    night = engine.asset.Texture("Tools/treeview_night.tex"),

    battle = engine.asset.Texture("Tools/treeview_battle.tex"),
    conversation = engine.asset.Texture("Tools/treeview_conversation.tex"),
}

local function TransformToPlatform(fname)
    -- TODO: convert this to whatever the platform needs?
    return fname:gsub("/", "\\")
end


local function CreateFolder(tgtdir)
    -- TODO: platform specific
    return os.execute("mkdir \"" .. TransformToPlatform(tgtdir) .. "\"") == 0
end

local function CopyFile(from, to)
    -- TODO: platform specific
    print("Copying file from " .. from .. " to " .. to)

    local execstr = "copy /Y \"" .. TransformToPlatform(from) .. "\" \"" .. TransformToPlatform(to) .."\""
    local result = os.execute(execstr)
end

local function MoveFile(from, to)
    -- TODO: platform specific
    print("Moving file from " .. from .. " to " .. to)

    local execstr = "move /Y \"" .. TransformToPlatform(from) .. "\" \"" .. TransformToPlatform(to) .. "\""
    local result = os.execute(execstr)
end

local function DeleteFileSVN(fname)
    -- TODO: platform specific
    print("Deleting file from SVN: " .. fname)

    local execstr = "svn delete -q \"" .. TransformToPlatform(fname) .. "\""
    local result = os.execute(execstr)
end

local function RemoveFolderIfEmptySVN(tgtdir)
    print("Removing folder from SVN: " .. tgtdir)
    -- TODO: platform specific


    print("rmdir /Q " .. TransformToPlatform(tgtdir))

    local result = os.execute("rmdir /Q \"" .. TransformToPlatform(tgtdir)) .. "\""
    print(result)

    if result == true then
        -- Folder was able to be deleted, so it must be empty. delete it from SVN as well:

        print("svn delete -q \"" .. TransformToPlatform(tgtdir) .. "\"")
        os.execute("svn delete -q \"" .. TransformToPlatform(tgtdir) .. "\"")
    end
end


local function RemoveFolderSVN(tgtdir)
    print("Removing folder from SVN: " .. tgtdir)

    os.execute("svn delete -q \"" .. TransformToPlatform(tgtdir) .. "\"")
end

-- usage: local path,file,extension = SplitPath("C:\\1\\test.png")
local function SplitPath(url)
    return string.match(url, "(.-)([^\\/]-)%.?([^%.\\/]*)$")
end


local function SavePlax(plax_name, plax_data, onsavedfn)
    local save_str = serpent.block(plax_data, {comment=false})
    save_str = loc.format("--THIS FILE IS AUTOGENERATED BY THE EDITOR. DO NOT CHANGE MANUALLY\nreturn {1}", save_str)

    local function OnSave( ok, msg )
        if ok then 
            print("Plax " .. plax_name .. " saved successfully.")
        else
            print( "ERROR: Failed to save plax", plax_name )
        end
        if onsavedfn then
            onsavedfn(ok, msg)
        end
    end

    if plax_data.mod_id then
        local mod = Content.FindMod( plax_data.mod_id )
        if not mod or mod.status == MOD_STATUS.WORKSHOP then
            OnSave( false, "Unable to edit plax for Workshop mod." )
            return
        end

        local fname = Content.GetPlaxPath( plax_name )
        if fname == nil then
            if mod.status == MOD_STATUS.DEV then
                -- MOD_STATUS.DEV means the path is something like: "../mods/ModName/plax.lua".
                fname = string.format( "%s/%s.lua", mod.installPath, plax_name )
            else
                -- MOD_STATUS.LOCAL means the path is something like: "C:\Users\UserName\AppData\Roaming\Klei\Griftlands\mods\ModName\plax.lua"
                -- The root portion of the path representing the alias needs to be shorn so that:
                -- fname = mods/ModName/plax.lua
                local relativePath = mod.installPath:match( "^.+[/\\](mods/.+)$") or mod.installPath
                fname = string.format( "%s/%s.lua", relativePath, plax_name )
            end
        else
            -- A LOCAL mod will be of the form ALIAS:plax_path
            -- The ALIAS needs to be replaced with the modPath.
            local relativePath = mod.installPath:match( "^.+[/\\](mods/.+)$") or mod.installPath
            fname = string.format( "%s/%s.lua", relativePath, plax_name )
        end
        
        print( string.format( "Saving %s mod plax to: %s", tostring(mod.status), fname ))
        if mod.status == MOD_STATUS.DEV then
            engine.data:SaveFile( fname, save_str, OnSave )
        else
            engine.settings:SaveFile( fname, save_str, OnSave )
        end

        if Content.GetPlaxPath( plax_name ) == nil then
            local path = string.format( "%s:%s", mod.alias, plax_name )
            print( "Adding new plax content:", plax_name, " => ", path )
            Content.AddPlaxData( plax_name, path )
        end

    else
        -- Save for the base game.
        local fname = 'scripts/content/plax/' .. plax_name .. '.lua'
        print("Saving plax data to " .. fname)
        engine.data:SaveFile( fname, save_str, OnSave )

        if Content.GetPlaxPath( plax_name ) == nil then
            local path = string.format( "content/plax/%s", plax_name )
            print( "Adding new plax content:", plax_name, " => ", path )
            Content.AddPlaxData( plax_name, path )
        end
    end
end

local function UpdateFileReferences(elem, fileremapping)
    local changes = {}

    for k,v in pairs(elem) do
        if (type(v) == "table") then
            UpdateFileReferences(v, fileremapping)
        else
            if (k == "img") then
                local remapping = fileremapping[v];
                if remapping then
                    changes[k] = remapping
                end
            end
        end
    end

    -- Actually replace the references to the images:
    for k,v in pairs(changes) do
        elem[k] = v
    end
  
end



local function FindTemplateImages(elem)
    local changes = {}
    local visitedTables = {}

    local function RecFindTemplateImages(elem)
        for k,v in pairs(elem) do
            if type(v) == "table" then
                if not visitedTables[v] then
                    visitedTables[v] = true
                    RecFindTemplateImages(v)
                end
            else
                if (k == "img") then
                    if string.find(v, "plax_templates") then
                      --  print("Found image: " .. v)
                        table.insert(changes, v)
                    end
                end
            end
        end
    end

    RecFindTemplateImages(elem)
    return changes
end



local function FindAllImages(elem)
    local changes = {}
    local visitedTables = {}
    local visitedImages = {}

    local function RecFindAllImages(elem)
        for k,v in pairs(elem) do
            if type(v) == "table" then
                if not visitedTables[v] then
                    visitedTables[v] = true
                    RecFindAllImages(v)
                end
            else
                if (k == "img") then
                    if not visitedImages[v] then
                        visitedImages[v] = true -- Only add images if they weren't added before
                        table.insert(changes, v)
                    end
                end
            end
        end
    end

    RecFindAllImages(elem)
    return changes
end

local function FilterElems(elems, elemToFind, plaxid)
    for k,v in pairs(elems) do
        if (v:lower() == elemToFind) then
            -- not a unique image! remove it:
            print("image " .. v .. " is also used in plax " .. plaxid)

            elems[k] = nil
            return true
        end
    end

    return false
end


local function FindAllUniqueImages(all_elems, plaxid)
    local images = FindAllImages(all_elems[plaxid])
    print("Find All Images for plax " .. plaxid .. " returned " .. #images .. " images")

    -- images holds all images found in elem. Now, see if any of these images are used in other Plaxes:
    for k,v in pairs(all_elems) do
        if not (k == plaxid) then
            local otherimages = FindAllImages(v)

            for otherk,otherv in pairs(otherimages) do
                FilterElems(images, otherv:lower(), k)
            end
        end
    end

    return images
end




local function FindCommonPart(strA, strB)
    if strA and strB then
        local index = 1;
       
--        print(strA)
--        print(strB)

        while strA:sub(1, index) == strB:sub(1, index) do
            index = index + 1
        end

        return strA:sub(1, index-1)
    end

    return ""
end


local function FindCommonFilenameSubstring(filenames)
    if #filenames > 0 then
        local path, file, ext = SplitPath(filenames[1])

        local commonpart = file

        for count = 2, #filenames do
            path, file, ext = SplitPath(filenames[count])

            commonpart = FindCommonPart(commonpart, file)
        end

        return commonpart
    end
end




local function GetElementDisplayName(elem)
    if elem then
        if elem.node_type == NODE_TYPE.IMAGE then
            return elem.img
        elseif elem.node_type == NODE_TYPE.SLOT then
            return elem.name
        elseif elem.node_type == NODE_TYPE.ANIMATION then
            return loc.format("{1}:{2}", elem.animfile, elem.anim or "anim")
        elseif elem.node_type == NODE_TYPE.PATH then
            return "PATH"
        elseif elem.node_type == NODE_TYPE.LOCATION then
            return loc.format("LOC: {1}", elem.location_name or elem.location_tags)
        elseif elem.node_type == NODE_TYPE.PARTICLES then
            if elem.particles then return "Particles: " .. elem.particles end
            return "Particles: Empty"
        else
            return "OTHER??"
        end
    end
end

function PlaxEditor:init() 
    PlaxEditor._base.init(self)

    self.ambience_txt = {"NONE"}
    self.ambience_events = {""}

    for k,v in ipairs( AUDIO.ambience_events ) do
        table.insert(self.ambience_txt, v.txt)
        table.insert(self.ambience_events, v.event_id)
    end

    local door_events = AUDIO:MatchEvents("sfx/general/door/(.*)")
    self.door_event_txt = {"NONE"}
    self.door_events = {""}
    if door_events then
        for k,v in ipairs(door_events) do
            table.insert(self.door_event_txt, v.txt)
            table.insert(self.door_events, v.event_id)
        end
    end

    self.music_txt = {"NONE"}
    self.music_events = {""}

    for k,v in ipairs( AUDIO.music_events ) do
        table.insert(self.music_txt, v.txt)
        table.insert(self.music_events, v.event_id)
    end
    --"event:/sfx/ambience/"


    local prev_agents = {}
    self.agent_data_by_name = {}
    for k,v in pairs(Content.GetAllCharacterDefs()) do
        if not v.is_template then
            if v.alias then
                local val = {txt= k, name=v.alias, agent = Agent(k), order = 1}
                self.agent_data_by_name[k] = val
                table.insert(prev_agents, val)
            else
                local val = {txt = k, name =k, agent = Agent(k), order = 2}
                self.agent_data_by_name[k] = val
                table.insert(prev_agents, val)
            end
        end
    end

    table.sort(prev_agents,
        function(a,b)
            if a.order == b.order then
                return a.txt < b.txt
            else
                return a.order < b.order
            end
        end )
    self.preview_agent_options = {}
    for k,v in ipairs(prev_agents) do
        table.insert(self.preview_agent_options, v.name)
    end

    self.posedef_options = {}
    for k,v in pairs(posedefs) do
        table.insert(self.posedef_options, k)
    end
    table.sort(self.posedef_options)


    self.day_phase = DAY_PHASE.DAY

    local strip_h =8
    self.plax_val = 0

    self.plax = Widget.Plax()

    self:AddChildren{
        Widget():SetName("SCALE"):SetAnchors("center", "center"):AddChildren{
            self.plax
        },
    }

    self.drag_element_state = {}
    self.selected_elements = {}

    local plax_ids = Content.GetAllPlaxSorted()
    if #plax_ids > 0 then
        local last_plax = DebugUtil.GetLocalDebugData("LAST_PLAX_EDIT")
        local id = last_plax and table.contains(plax_ids, last_plax) and last_plax or plax_ids[1]
        self.save_id = id
        self.title = id
        self:SetPlax( Content.GetPlaxData( id, true ))
    end
    self.dirty = false
    self:ClearUndo()


end

function PlaxEditor:OnAdded()
    TheGame:GetDebug():RemoveBindingGroup( debug_menus.GAME_BINDINGS )
    TheGame:GetDebug():AddBindingGroup( debug_menus.TOOL_BINDINGS )

    -- Conflicts with Deselect hotkey.
    TheGame:GetDebug():SetBindingEnabled( "EXECUTE_DEBUG", false ) 
end

function PlaxEditor:OnRemoved()
    TheGame:GetDebug():RemoveBindingGroup( debug_menus.TOOL_BINDINGS )
    TheGame:GetDebug():AddBindingGroup( debug_menus.GAME_BINDINGS )
    TheGame:GetDebug():SetBindingEnabled( "EXECUTE_DEBUG", true )
end


local function CreateNewSettings()
  return {
        shadows = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
        midtones = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
        highlights = {
            colour = {1,1,1},
            hsv = {0,1,1},
        },
    }
end

function PlaxEditor:DoRender(dt)

    self.plax:EditorRenderToScene()

    local ui = TheGame:GetDebug():GetDebugUI()

    self:RenderEditorPanel( ui )

    if not self:IsOnStack() then
        return -- screen may have popped via the editor menu
    end
 
     if #self.selected_elements == 1 then
        self:RenderElementPanel( ui )
    end

    if self.open_select_popup then
        ui:OpenPopup( self.open_select_popup )
        self.open_select_popup = false
    end

    if self.errorSaving then
        ui:OpenPopup( "SAVE_ERROR" )  
        self.errorSaving = false
    end

    if ui:BeginPopup( "PATH_MENU" ) then
        
        local path = self.selected_elements[1]
        local widget = self.plax:GetWidgetByElement(path)
        local node = widget and widget:GetSelectedNode()
        local edge = widget and widget:GetSelectedEdge()

        if edge then
            if ui:MenuItem("Delete Edge") then
                self:PushState()
                widget:RemoveEdge(edge)
            end

            if ui:MenuItem("Insert Node") then
                self:PushState()
                widget:BisectEdge(edge)
            end

        end

        if node then
            if ui:MenuItem("Extend") then
                self:PushState()
                widget:ExtendNode(node)
            end
            if ui:MenuItem("Delete Node") then
                self:PushState()
                widget:RemoveNode(node)
            end
            local other_node = widget:CanWeldNode(node)
            if other_node then
                if ui:MenuItem("Weld") then
                    self:PushState()
                    widget:Weld(node, other_node)
                end
            end
        end

        ui:EndPopup()
    end

    if ui:BeginPopup( "SELECT_UNDER_MOUSE" ) then
        for k = #self.items_under_mouse , 1, -1 do
            local v = self.items_under_mouse[k]
            local str = GetElementDisplayName(v.element)
            if str then
                local elem = v.element
                str = elem.parent_layer.name .. "/" .. str
                
                if ui:MenuItem( str ) then
                    self:ClearSelection()
                    self:SelectElement(elem, true)    
                end
            end
        end

        ui:EndPopup()
    end

    if ui:BeginPopupModal( "OPEN_PLAX", ui.AlwaysAutoResize ) then
        self:RenderOpenPlax( ui )
        ui:EndPopup()
    end
    if ui:BeginPopupModal( "NEW_PLAX", ui.AlwaysAutoResize ) then
        self:RenderNewPlax( ui )
        ui:EndPopup()
    end
    
    if ui:BeginPopupModal( "DELETE_PLAX", ui.AlwaysAutoResize ) then
        self:RenderDeletePlax( ui )
        ui:EndPopup()
    end


    if ui:BeginPopupModal( "Plax Editor Settings", ui.AlwaysAutoResize ) then
        self:RenderSettingsMenu( ui )
        ui:EndPopup()
    end

    if ui:BeginPopupModal( "SAVE_ERROR", ui.AlwaysAutoResize ) then
        self:RenderSaveError( ui )
        ui:EndPopup()
    end
        
    if ui:BeginPopupModal("NEW_SNAPSHOT_POPUP", ui.AlwaysAutoResize) then
        ui:Text("Name of new colour setting...")
        local new_name = ui:InputText("", self.new_name)
        
        if new_name then
            self.new_name = new_name
        end

        ui:Columns(2, "", false)	-- false to turn off manual resizing

        local totalWidth = 106
        if ui:Button("Create New",totalWidth,nil, not (self.new_name and #self.new_name > 0 and not sColourSettings[self.new_name])) then
            sColourSettings[self.new_name] = CreateNewSettings()
            self:UpdateSnapshotList()
            ui:CloseCurrentPopup()
        end

        ui:NextColumn()

        if ui:Button("Cancel", totalWidth) then
            ui:CloseCurrentPopup()
        end
        ui:Columns(1)

        if self.new_name and sColourSettings[self.new_name] then
            ui:Text("Already Exists...")
        elseif self.new_name == nil or self.new_name == "" then
            ui:Text("Enter a name first")
        end

        ui:EndPopup()
    end
end

function PlaxEditor:RenderOpenPlax( ui )
    ui:Text( "Select a PLAX to open:" )
    ui:Spacing()
    ui:Separator()

    local regen_plax = false
    self.filter_str = ui:InputText( "Filter", self.filter_str or "" ) or self.filter_str

    ui:BeginChild( "Plaxes", 400, 300 )
    for i, str in ipairs( Content.GetAllPlaxSorted() ) do
        if self.filter_str == nil or string.lower(str):find( string.lower(self.filter_str) ) then
            if ui:Selectable( str ) then
                DebugUtil.SetLocalDebugData("LAST_PLAX_EDIT", str)
                DebugUtil.SaveLocalDebugData()

                self.save_id = str
                self:ClearUndo()
                self.dirty = false
                table.clear(self.selected_elements)
                self.selected_layer = nil
                self:SetPlax( Content.GetPlaxData( str, true ) )
                ui:CloseCurrentPopup()
            end
        end
    end
    ui:EndChild()

    ui:Separator()
    ui:Spacing()
    if ui:Button( "Cancel" ) then
        ui:CloseCurrentPopup()
    end
end



function PlaxEditor:DeletePlax( plax_id )
    -- Deletes the plax with the given id
    local plax_data = Content.GetPlaxData( plax_id )
    if plax_data then
        print("Deleting Plax '" .. plax_id .. "'")

        -- Just remove the Plax folder itself:
        -- contentsrc/images/plax/plax_id or contentsrc/images/plax_template/plax_id
        -- data/plax/plax_id or data/plax_template/plax_id

        local datadir = ""
        if plax_data.is_template then
            print("Deleting Plax Template '" .. plax_id .. "'")
            datadir = "plax_templates/"
        else
            print("Deleting Plax '" .. plax_id .. "'")
            datadir = "plax/"
        end

        RemoveFolderSVN("../contentsrc/images/" .. datadir .. plax_id);
        RemoveFolderSVN(datadir .. plax_id);
        DeleteFileSVN("scripts/content/plax/" .. plax_id .. ".lua")

        Content.RemovePlaxData( plax_id )
    end
end


function PlaxEditor:RenderDeletePlax( ui )
    ui:Text( "Select a PLAX to delete:" )
    ui:Spacing()
    ui:Separator()

    local regen_plax = false
    self.filter_str = ui:InputText( "Filter", self.filter_str or "" ) or self.filter_str

    ui:BeginChild( "Plaxes", 400, 300 )
    for i, str in ipairs( Content.GetAllPlaxSorted() ) do
        if self.filter_str == nil or string.lower(str):find( string.lower(self.filter_str) ) then

            if ui:Selectable( str, (self.plaxToDelete == str) ) then
                self.plaxToDelete = str
            end
        end
    end
    ui:EndChild()

    ui:Separator()
    ui:Spacing()

    if self.plaxToDelete and (self.filter_str == nil or string.lower(self.plaxToDelete):find( string.lower(self.filter_str) )) then
        if ui:Button( "Delete " .. self.plaxToDelete ) then
            self:DeletePlax( self.plaxToDelete )

            if self.save_id == self.plaxToDelete then
                -- Find any other plax that doesn't have the same name:
                for id, v in Content.AllPlax() do
                    if id ~= self.save_id then
                        self.save_id = id
                        break
                    end
                end
                self:SetPlax( Content.GetPlaxData( self.save_id, true ))

                table.clear(self.selected_elements)
                self.selected_layer = nil
                self:ClearUndo()
            end
        
            self.dirty = true
            self:Refresh()
            self:DoSave()

            self.plaxToDelete = nil
        end    

        ui:SameLine(nil, 5)
    end

    if ui:Button( "Cancel", 80 ) then
        self.plaxToDelete = nil
        ui:CloseCurrentPopup()
    end
end


function PlaxEditor:RenderNewPlax( ui )
    ui:Text( "Enter a unique identifier for this new plax" )
    ui:Spacing()
    ui:Separator()

    self.new_plax_id = ui:InputText( "ID", self.new_plax_id or "" ) or self.new_plax_id

    local mods = Content.GetInstalledMods()
    if CONFIG ~= "PRODUCTION" then
        table.insert( mods, 1, GRIFTLANDS_MOD )
    end

    -- Strip Workshop Mods, you can't edit those.
    for i = #mods, 1, -1 do
        local mod = Content.FindMod( mods[i] )
        if mod and (mod.status ~= MOD_STATUS.DEV and mod.status ~= MOD_STATUS.LOCAL) then
            table.remove( mods, i )
        end
    end

    local idx = table.arrayfind( mods, self.new_plax_mod or self.plaxdata.mod_id )
    local new_idx = ui:Combo( "Mod", idx or 1, mods )
    if new_idx and new_idx ~= idx then
        self.new_plax_mod = mods[ new_idx ]
    elseif self.new_plax_mod == nil then
        self.new_plax_mod = self.plaxdata.mod_id
    end

    local disabled_txt
    if #(self.new_plax_id or "") == 0 then
        disabled_txt = "Enter a valid name."
    elseif Content.GetPlaxData( self.new_plax_id ) then
        disabled_txt = "That name is already in use by another plax."
    end
    if disabled_txt then
        ui:TextColored( 0xCC0000FF, disabled_txt )
    end

    if ui:Checkbox("Is Region", self.new_plax_is_region) then
        self.new_plax_is_region = not self.new_plax_is_region
    end

    if ui:Checkbox("Is Template", self.new_plax_is_template) then
        self.new_plax_is_template = not self.new_plax_is_template
    end

    ui:Separator()
    if ui:TreeNode( "Clone From..." ) then
        local regen_plax = false
        self.filter_str = ui:InputText( "Filter", self.filter_str or "" ) or self.filter_str

        ui:BeginChild( "Plaxes", 400, 200 )
        for i, str in ipairs( Content.GetAllPlaxSorted() ) do
            if self.filter_str == nil or string.lower(str):find( string.lower(self.filter_str) ) then
                if ui:Selectable( str, self.new_clone_id == str ) then
                    if self.new_clone_id == str then
                        self.new_clone_id = nil
                    else
                        self.new_clone_id = str
                    end
                end
            end
        end
        ui:EndChild()
        ui:TreePop()
    end

    ui:Separator()
    ui:Spacing()

    local txt
    if self.new_clone_id then
        txt = string.format( "Clone '%s'", self.new_clone_id )
    else
        txt = "New Plax"
    end

    if ui:Button( txt, nil, nil, disabled_txt ~= nil ) then
        self.save_id = self.new_plax_id
        local new_plax_data
        
        if self.new_clone_id then
            new_plax_data = deepcopy( Content.GetPlaxData( self.new_clone_id, true ) )
            local filesToCopy = FindTemplateImages( new_plax_data )
            local commonFname = FindCommonFilenameSubstring(filesToCopy);

            new_plax_data.is_template = self.new_plax_is_template

            local plaxtgtpath = "plax/"
            if self.new_plax_is_template then 
                plaxtgtpath = "plax_templates/"
            end

			if commonFname then
				print ("Common filename part = " .. commonFname)

				--
				-- Copy the data folder textures:
				--

				-- Make sure the target folder exists:
				local tgtdir = plaxtgtpath .. self.new_plax_id .. "/";

				if not CreateFolder(tgtdir) then 
					print("mkdir returned error code! Trying to create folder " .. tgtdir );
				end

				local fileremapping = {}

				-- Copy:
				for k,v in pairs(filesToCopy) do
					local path, file, ext = SplitPath(v)

					local renamedFile = string.gsub(file, commonFname, self.new_plax_id .. "_")

					local tgtdata = tgtdir .. renamedFile .. "." .. ext;

					fileremapping[v] = tgtdata
					CopyFile(v, tgtdata);
				end

				-- Update the references:
				UpdateFileReferences(new_plax_data, fileremapping);

				-- Now copy the source folder textures:
				local tgtcontentsrcdir = "../contentsrc/images/" .. plaxtgtpath .. self.new_plax_id .. "/";
				if not CreateFolder(tgtcontentsrcdir) then 
					print("mkdir returned error code! Trying to create folder " .. tgtcontentsrcdir );
				end

				for k,v in pairs(filesToCopy) do
					local path, file, ext = SplitPath(v)

					local sourcefile = "../contentsrc/images/" .. path .. "/" .. file .. ".png"

					local renamedFile = string.gsub(file, commonFname, self.new_plax_id .. "_")
					local targetfile = tgtcontentsrcdir .. renamedFile .. ".png"

					CopyFile(sourcefile, targetfile);
				end
			end                        
        else
            new_plax_data = { plax=150, is_region = self.new_plax_is_region, is_template = self.new_plax_is_template, layers = {}}
        end

        if self.new_plax_mod ~= GRIFTLANDS_MOD then
            new_plax_data.mod_id = self.new_plax_mod
        else
            new_plax_data.mod_id = nil
        end

        self:SetPlax( new_plax_data )

        table.clear(self.selected_elements)
        self.selected_layer = nil
        
        self:ClearUndo()
        self.dirty = true
        self.new_plax_id = nil
        self.new_clone_id = nil
        self:Refresh()
        ui:CloseCurrentPopup()
    end

    ui:SameLine( nil, 20 )
    if ui:Button( "Cancel" ) then
        ui:CloseCurrentPopup()
    end
end

function PlaxEditor:SetDayPhase(phase)
    self.day_phase = phase
    self.plax:SetDayPhase(phase)
end

function PlaxEditor:SetGameMode(mode)
    self.game_mode = mode
    self.plax:SetGameMode(mode)
end

function PlaxEditor:DoDemo(no_characters)
    self:GetFE():PushScreen(Screen.PlaxDemoScreen(self.plaxdata, self.day_phase, no_characters))
end

function PlaxEditor:DoFightDemo()
    local scenario = Scenario()
    local sal = Agent("SAL")

    scenario:AddAgentToTeam( sal, TEAM.BLUE, 0, true )
    for i = 1, (self.blue_fighters or 1) -1 do
        scenario:AddAgentToTeam( Agent( "BANDIT_GOON" ), TEAM.BLUE )
    end

    scenario:AddAgentToTeam( Agent("BANDIT_CAPTAIN"), TEAM.RED )
    for i = 1, (self.red_fighters or 1) -1 do
        scenario:AddAgentToTeam( Agent( "BANDIT_GOON" ), TEAM.RED )
    end

    local battle = BattleEngine( scenario )
    local screen = Screen.FightScreen( battle )
    screen:SetDemoMode( self.plaxdata, self.day_phase )
    self:GetFE():InsertScreen( screen )
end

function PlaxEditor:SetElemBloom(elem, b)
    if not math.fequal(b, elem.bloom or 0) then
        b = clamp(b, 0, 1)
        elem.bloom = b
        self:SetPendingChange("BLOOM")
        local w = self.plax:GetWidgetByElement(elem)
        if w and w.SetBloom then
            w:SetBloom(b)
        end
    end
end

function PlaxEditor:SetElemColour(elem, r,g,b,a)
    r,g,b,a = tonumber(r) or 255,tonumber(g) or 255,tonumber(b) or 255,tonumber(a) or 255
    elem.tint_r = clamp(r, 0, 255)
    elem.tint_g = clamp(g, 0, 255)
    elem.tint_b = clamp(b, 0, 255)
    elem.tint_a = clamp(a, 0, 255)
    self:SetPendingChange("COL")
    local w = self.plax:GetWidgetByElement(elem)
    if w then 
        self.plax:GetWidgetByElement(elem):SetTintColour( 
            elem.tint_r/255, 
            elem.tint_g/255,
            elem.tint_b/255,
            elem.tint_a/255)
    end
end


function PlaxEditor:DoPaste(x, y)
    x = x or 0
    y = y or 0

    if self.copied_elems then

        self:ClearSelection()
        
        if self.cut_elems then
            for k,elem in pairs( self.cut_elems ) do
                for k,v in pairs(self.plaxdata.layers) do
                    if table.arrayremove(v.elements, elem) then
                        break
                    end
                end
            end
            self.cut_elems = nil
        end

        for k,v in pairs(self.copied_elems) do
            if v.node_type ~= "LAYER" then
                local new_element = deepcopy_circ(v)
                if #self.plaxdata.layers == 0 then
                    self:AddNewLayer()
                end
                local parent_layer = self.selected_layer or  self.plaxdata.layers[1]
                new_element.parent_layer = parent_layer
                new_element.id = self.next_id
                self.next_id = self.next_id + 1
                table.insert( parent_layer.elements, new_element)
                self:SelectElement(new_element, true)
            end
        end


        self:PushState()
        self:Refresh()
    end 
end


function PlaxEditor:DoCopy()
    if #self.selected_elements > 0 then
        -- Shallowcopy so we can order these elemnets according to their original layers.
        table.sort( self.selected_elements, function( e1, e2 )
                local layer1, layer2 = e1.parent_layer, e2.parent_layer
                if layer1 == layer2 then
                    local idx1, idx2 = table.find( e1.parent_layer.elements, e1 ), table.find( e2.parent_layer.elements, e2 )
                    return idx1 < idx2
                else
                    local idx1, idx2 = table.find( self.plaxdata.layers, layer1 ), table.find( self.plaxdata.layers, layer2 )
                    return idx1 < idx2
                end
            end )

        self.copied_elems = deepcopy_circ(self.selected_elements)
        self.cut_elems = nil
    end
end

function PlaxEditor:DoCut()
    if #self.selected_elements > 0 then
        self:DoCopy()
        self.cut_elems = shallowcopy(self.selected_elements)
    end
end


function PlaxEditor:AddProp()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.SLOT, id =self.next_id, name = "SLOT" }
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        --elem.preview_texture = DEFAULT_PROP_TEX
        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end
end

function PlaxEditor:AddPath()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.PATH, id = self.next_id, name = "PATH", nodes = { {x=0,y=0}, {x=100, y = 100}}, edges = {{node1=1, node2=2}} }
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end
end

function PlaxEditor:AddLocationNode()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.LOCATION, id = self.next_id, location_name = "", location_tags = "" }
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end
end


function PlaxEditor:AddCharacter()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.SLOT, id =self.next_id, name = "CHARACTER"}
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        elem.preview_agent = "ROOK"
        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end        
end

function PlaxEditor:AddAnimation()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.ANIMATION, id =self.next_id, name = "ANIM"}
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end        
end


function PlaxEditor:AddParticles()
    local layer = self.selected_layer or self.plaxdata.layers[1]
    if layer then
        layer.elements = layer.elements or {}
        local elem = {x = 0, y = 0, parent_layer = layer, node_type = NODE_TYPE.PARTICLES, id =self.next_id, name = "PARTICLES"}
        elem.plax_flags = PLAX_DEFAULT_FLAGS;

        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
        self:Refresh()
        self:SelectElement( elem )
    end        
end


function PlaxEditor:SetElemPoseDef(elem, posedef)
    if elem.posedef ~= posedef then
        elem.posedef = posedef
        self:RefreshSlot( elem )
        self:PushState()
    end
end

function PlaxEditor:SetAnimState(elem, anim)
    if elem.anim ~= anim then
        assert( elem.node_type == NODE_TYPE.ANIMATION )
        elem.anim = anim
        local w = self.plax:GetWidgetByElement(elem)
        if w and anim then
            self.plax:PlayRepeatingAnim(w, elem.anim or "anim", elem.anim_repeat_delay_min, elem.anim_repeat_delay_max)
            self:UpdateWidgetBounds(w)
        end

        self:PushState()
    end
end


function PlaxEditor:SetElemLock(elem, viz)
    elem.lock = viz
    self:PushState()
end


function PlaxEditor:SetElemFlip(elem, flip)
    elem.flip = flip
    local w = self.plax:GetWidgetByElement(elem)
    if w then
        w:SetScale((elem.scale or 1)*(flip and -1 or 1), (elem.scale or 1))
    end
    self:PushState()

end


function PlaxEditor:SetLayerPlax(layer, p)
    layer.plax = p
    self:SetPendingChange("PLAX")
    self:SetPlaxPercent(self.plax_val)  
end

function PlaxEditor:SuggestSave(fn)
    
    TheGame:FE():PushScreen( Screen.YesNoPopup("Save?", "You have unsaved changes. Save first?", true):SetFn(
        function(res) 
            if res == Screen.YesNoPopup.YES then 
                self:DoSave( fn )
            elseif res == Screen.YesNoPopup.NO then
                fn(true)
            elseif res == Screen.YesNoPopup.CANCEL then
                fn(false)
            end
        end))
end

function PlaxEditor:DoNew(force, data)
    if not force and self.dirty then
        self:SuggestSave(function(v) if v then self:DoNew(true, data) end end)
        return
    end

    self.open_select_popup = "NEW_PLAX"
end


function PlaxEditor:AddNewLayer()

    table.insert( self.plaxdata.layers, { node_type = "LAYER", name = "LAYER", plax = .5, id = self.next_id, elements = {}} )
    self.next_id = self.next_id + 1
    self:PushState()
    self:Refresh()

end

function PlaxEditor:DoOpen(force)
    if not force and self.dirty then
        self:SuggestSave(function(v) if v then self:DoOpen(true) end end)
        return
    end

    self.open_select_popup = "OPEN_PLAX"
end

function PlaxEditor:DoDelete(force)
    if not force and self.dirty then
        self:SuggestSave(function(v) if v then self:DoDelete(true) end end)
        return
    end

    self.open_select_popup = "DELETE_PLAX"
end




function PlaxEditor:DoClean()

    local used_plaxes = {}

    local all_plax = {}

    local plax_data = require "content/plax_defs"
    for str, plax in pairs ( plax_data ) do
        table.insert_unique(all_plax, str)
    end

    for _, location in pairs(Content.GetLocationContents()) do
        table.insert_unique(used_plaxes, location.plax)
    end

    local ad_hoc = {
        "cyoa_slide_card",
        "title_screen",
        "REGION_MURDERBAY1",
        "region_grout_bog",
    }

    for k,v in ipairs(ad_hoc) do
        table.insert(used_plaxes, v)
    end

    table.sort(all_plax)
    table.sort(used_plaxes)

    local unused_plaxes = {}

    for k,v in ipairs(all_plax) do
        if not table.arrayfind(used_plaxes, v) then
            table.insert(unused_plaxes, v)
        end
    end

    for k,v in pairs(unused_plaxes) do
        self:DeletePlax(v)
    end
end


function PlaxEditor:DoSave(onsavedfn)
    
    if self.pending then
        self:PushState()
        self.pending = false
    end

    if self.dirty then
        
        local save_table = {
            plax = self.plaxdata.plax, 
            mod_id = self.plaxdata.mod_id,
            is_region = self.plaxdata.is_region,
            is_template = self.plaxdata.is_template,
            day_colour = self.plaxdata.day_colour, 
            night_colour = self.plaxdata.night_colour, 
            ambient_event = self.plaxdata.ambient_event,
            music_event = self.plaxdata.music_event,
            entry_event = self.plaxdata.entry_event,
            exit_event = self.plaxdata.exit_event,
            scripts = self.plaxdata.scripts,
            layers = {}}

        for k, layer in ipairs(self.plaxdata.layers) do
            local layer_save_table =
            {
                plax = layer.plax,
                name = layer.name,
                elements={},
                guide = layer.guide,
                plax_flags = layer.plax_flags,
                hidden = layer.hidden_in_editor,
                
                day_colour = layer.day_colour,
                night_colour = layer.night_colour, 
                
                day_colour_ignore_parent = layer.day_colour_ignore_parent,
                night_colour_ignore_parent = layer.night_colour_ignore_parent, 
            }
            for k, element in ipairs(layer.elements) do
                local elem_table = {}
                for k,v in pairs(element) do
                    if type(v) == "number" or type(v) == "boolean" or type(v) == "string" then
                        elem_table[k] = v
                    elseif (k == "day_colour") or (k == "night_colour") then
                        elem_table[k] = v
                    end
                end
                if element.node_type == NODE_TYPE.PATH then
                    elem_table.nodes = element.nodes
                    elem_table.edges = element.edges
                end

                table.insert(layer_save_table.elements, elem_table)
            end
            table.insert(save_table.layers, layer_save_table)
        end

        self.saving = true
        SavePlax(self.save_id, save_table,
            function(ok, msg) 
                self.saving = false
                if ok then 
                    self.dirty = false 
                else
                    self.errorSaving = true
                    self.errorSavingMsg = msg
                end
                if onsavedfn then onsavedfn(ok) end
            end)
    else
        if onsavedfn then onsavedfn(true) end
    end
end

function PlaxEditor:DoExit(force)
    if not force and self.dirty then
        self:SuggestSave(function(v) if v then self:DoExit(true) end end)
        return
    end

    TheGame.scheduler:DoTaskInTime( "editor-reset", 0.01, TheGame.Reset, TheGame )
end


function PlaxEditor:ClearUndo()
    self.undo_states = {{data=deepcopy_circ(self.plaxdata)}}
    self.undo_head = 1
    self.redo_max = 1
end

function PlaxEditor:PushState()
    local data = deepcopy_circ(self.plaxdata)

    local sel = {}
    for k,v in ipairs(self.selected_elements) do
        table.insert(sel, v.id)
    end
    
    table.insert(self.undo_states, self.undo_head+1, {data = data, selection = sel})
    self.undo_head = self.undo_head+1
    self.redo_max = self.undo_head
    self.dirty = true
end

function PlaxEditor:Redo()

    if self.pending then
        self:PushState()
        self.pending = false
    end

    if self.undo_head < self.redo_max then
        self.undo_head = self.undo_head + 1
        self:SetPlax(self.undo_states[self.undo_head].data, self.undo_states[self.undo_head].selection)
    end
end

function PlaxEditor:Undo()
    if self.pending then
        self:PushState()
        self.pending = false
    end

    if self.undo_head > 1 then
        self.undo_head = self.undo_head - 1
        self:SetPlax(self.undo_states[self.undo_head].data, self.undo_states[self.undo_head].selection)
    end
end

function PlaxEditor:OnGainTop()
    PlaxEditor._base.OnGainTop(self)
    TheGame:GetEvents():ListenForEvents( self, "file_drop" )
end

function PlaxEditor:OnLoseTop()
    PlaxEditor._base.OnLoseTop(self)
    TheGame:GetEvents():RemoveListener( self )
end


function PlaxEditor:OnDropFile(filename)
    
    filename= path.normslashes(filename)
    
    local basepath, fname = path.splitpath(filename)
    local fname, extension = path.splitext(fname)
    local common = path.common_prefix(DATA_DIR, filename)
    local under_data_dir = common == DATA_DIR
    local fname = path.relpath (filename,DATA_DIR)
    
    print("fname: " .. fname)
    
    local isTexture = extension == ".tex"
    local isAnim = extension == ".zip"
    local particlespath = "scripts/content/particles/"
    local isParticles = extension == ".lua" and (fname:sub(1, #particlespath) == particlespath)

    if isTexture then 
		print("Texture Dropped")
        if not under_data_dir then
            TheGame:FE():PushScreen( Screen.InfoPopup("Invalid Texture", "Texture needs to be located under the data folder." ))
            return
        end
    elseif isAnim then 
		print("Animation Dropped")
        if not under_data_dir then
            TheGame:FE():PushScreen( Screen.InfoPopup("Invalid Anim", "Anim needs to be located under the data folder." ))
            return
        end
    elseif isParticles then 
		print("Particles Dropped")
    else
        TheGame:FE():PushScreen( Screen.InfoPopup("Invalid File", "Unrecognized file dropped onto the window. (need to be .tex, .zip, or .lua)" ))
        return
    end

    local elem
    if #self.plaxdata.layers == 0 then
        self:AddNewLayer()
    end

    local layer = self.selected_layer or self.plaxdata.layers[#self.plaxdata.layers]
    if layer then
        layer.elements = layer.elements or {}
        local x, y = self.plax:GetWidgetByElement(layer):TransformFromWorld(TheGame:GetInput():GetMousePos())

        local imgscale = 1.0
        local settings = DebugUtil.GetLocalDebugData( "PLAX_SETTINGS" )
        if settings then
            imgscale = settings.DefaultImageScale or 1.0
        end


        if isTexture then
            elem = {img = fname, x = x, y = y, scale = imgscale, parent_layer = layer, node_type = NODE_TYPE.IMAGE, id =self.next_id}
        elseif isAnim then
            elem = {animfile = fname, x = x, y = y, parent_layer = layer, node_type = NODE_TYPE.ANIMATION, id =self.next_id}
        elseif isParticles then
            elem = {particles = fname, x = x, y = y, parent_layer = layer, node_type = NODE_TYPE.PARTICLES, id =self.next_id}
        end

        elem.plax_flags = PLAX_DEFAULT_FLAGS;
        
        self.next_id = self.next_id+ 1
        table.insert( layer.elements, elem)
        self:PushState()
    end        
    self:Refresh()

end



function PlaxEditor:OnGlobalEvent( eventname, ... )
    if eventname == "file_drop" then
        self:OnDropFile(...)
    end
end


function PlaxEditor:UpdateWidgetBounds(w)
    if w then
        if not w.bounding_widget then
            w.bounding_widget = Widget.PolyLine():CreateRect(1,1, true):SetTintColour(0xC9D32BFF)
            w:AddChild(w.bounding_widget)
        end

        local width, height = w:GetSize()
        local x,y = w:GetCentroid()    
        w.bounding_widget:SetScale(width,height):SetPos(x,y)

    end
end

function PlaxEditor:SetBoundingWidget(w)
    self.bounding_widget = w
    if w then
        self.bounding:Hide()
        self.bounding:SetScale(self.bounding_widget:GetSize())
        self.bounding_widget:AddChild(self.bounding)
        self.bounding:Show()
    else
        self.bounding:Hide()
    end
end


function PlaxEditor:UpdateElementViz(element)
    self.plax:RefreshVisibility( element )
end

function PlaxEditor:SetElemFlags(element, flags)
    if flags == 0 then
        element.plax_flags = nil
    else
        element.plax_flags = flags
    end
    self:UpdateElementViz(element)
    self:PushState()
end

function PlaxEditor:SetViz(element, viz)
    element.hidden_in_editor = not viz
    self:UpdateElementViz(element)
    self:PushState()
end

function PlaxEditor:IsElementVisibleForCurrentSetting(elem)
    local visible = true

    if (self.day_phase == DAY_PHASE.DAY) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.DAY ) then visible = false end
    if (self.day_phase == DAY_PHASE.NIGHT) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.NIGHT ) then visible = false end

    if (self.game_mode == GAME_MODE.CONVERSATION) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.CONVERSATION ) then visible = false end
    if (self.game_mode == GAME_MODE.BATTLE) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.BATTLE ) then visible = false end

    return visible
end

function PlaxEditor:SetGuide(element, guide)
    element.guide = guide
    self:PushState()
end



function PlaxEditor:ToggleFlags(element, flag)
    element.plax_flags = ToggleBits(element.plax_flags, flag)
    self:UpdateElementViz(element)
    self:PushState();
end




function PlaxEditor:MoveLayer(delta)
    local moved = false

    for k,sel_elem in pairs(self.selected_elements) do
        if sel_elem.parent_layer then
            local layer_idx = table.arrayfind(self.plaxdata.layers, sel_elem.parent_layer) 
            if layer_idx then
                local new_layer_idx = clamp(layer_idx + delta, 1, #self.plaxdata.layers)
                if new_layer_idx ~= layer_idx then
                    
                    table.arrayremove(self.plaxdata.layers[layer_idx].elements, sel_elem)
                    
                    if delta < 0 then
                        table.insert(self.plaxdata.layers[new_layer_idx].elements, sel_elem)
                    else
                        table.insert(self.plaxdata.layers[new_layer_idx].elements, 1, sel_elem)
                    end
                    sel_elem.parent_layer = self.plaxdata.layers[new_layer_idx]
                    moved = true
                end
            end
        end
    end

    if moved then
        self:PushState()
        self:Refresh()
    end


end

function PlaxEditor:MoveLayerOrder(delta)
    local moved = false

    local move_items = {}
    for k,sel_elem in pairs(self.selected_elements) do
        
        if sel_elem.parent_layer then
            local idx = table.arrayfind(sel_elem.parent_layer.elements, sel_elem)
            if idx then
                table.insert(move_items, {elem = sel_elem, idx = idx})
            end
        end
    end
    if #move_items > 0 then
        
        if delta < 0 then
            table.sort(move_items, function(a,b) assert(a.idx and b.idx) return a.idx < b.idx end)
        else
            table.sort(move_items, function(a,b) assert(a.idx and b.idx) return a.idx > b.idx end)
        end

        for k,v in pairs(move_items) do

            local other_idx = clamp(v.idx + delta, 1, #v.elem.parent_layer.elements)
            if other_idx ~= v.idx then
                moved = true
                local w1 = self.plax:GetWidgetByElement( v.elem )
                local w2 = self.plax:GetWidgetByElement( v.elem.parent_layer.elements[other_idx] )
                if w1 and w2 then
                    w1.parent:SwapChildOrder(w1, w2)
                end
                v.elem.parent_layer.elements[v.idx], v.elem.parent_layer.elements[other_idx] = v.elem.parent_layer.elements[other_idx], v.elem.parent_layer.elements[v.idx]
            end
        end

        if moved then
            self:PushState()
        end
    end
end


local node_images = {
    CHARACTER = assets.character,
    IMAGE = assets.img,
    ANIMATION = assets.anim
}


function PlaxEditor:Refresh()

    if self.plaxdata then
        self.plax:SetLayers(self.plaxdata.layers, true)
        self.plax:SetColourNames(self.plaxdata.day_colour, self.plaxdata.night_colour)

        -- Add preview agents.        
        for slot_id, slot in self.plax:Slots() do
            if slot.data.preview_agent == nil then
                slot.data.preview_agent = "ROOK"
            end

            self:RefreshSlot( slot.data )
        end
    end
           
    if self.selected_layer then
        if not table.arrayfind(self.plaxdata.layers, self.selected_layer) then
            self.selected_layer = nil
        end
    end
    
    self.plax:SetPlaxShift(self.plax_val)
end

function PlaxEditor:SelectLayer(layer)
    self.selected_layer = layer
end


function PlaxEditor:GetImportantRolesForCurrentLocation()
    self.important_roles = self.important_roles or {}
    table.clear(self.important_roles)

    local loc = ObtainWorkTable()
    
    Content.GetLocationsThatUsePlax(self.save_id, loc)
    for k,location_id in pairs(loc) do
        local content = Content.GetLocationContent(location_id)
        if content.plax_roles then
            for k,v in pairs(content.plax_roles) do
                table.insert_unique(self.important_roles, v)
            end
        end
        
        if content.work then
            for _, position in pairs(content.work) do
                
                if position.job and position.job.role then
                    table.insert_unique(self.important_roles, position.job.role)
                end
            end
        end
    end
    ReleaseWorkTable(loc)
    table.sort(self.important_roles)
    table.insert( self.important_roles, 1, "[OPEN SLOT]")

    for role in pairs(CHARACTER_ROLES) do
        table.insert_unique(self.important_roles, role)
    end
    return self.important_roles
end


function PlaxEditor:SetPlax(plaxdata, selected_elements)
    self.plaxdata = deepcopy_circ(plaxdata)
    local max_id = 0
    local new_selected_elements = {}

    --condition the data a bit so that we can tell what things are without context
    for k,layer in pairs(self.plaxdata.layers) do
        
        layer.id = selected_elements and layer.id or max_id
        max_id = math.max(max_id, layer.id + 1)
        layer.node_type = "LAYER"
    
        if selected_elements and (table.arrayfind(selected_elements, layer.id) ~= nil) then
            table.insert(new_selected_elements, layer)
        end


        for k,v in pairs(layer.elements) do
            
            v.x = v.x and math.floor(v.x) or 0
            v.y = v.y and math.floor(v.y) or 0
            v.id = selected_elements and v.id or max_id
            max_id = math.max(max_id, v.id + 1)
            
            if not IsEnum( v.node_type, NODE_TYPE ) then
                assert( nil , "UNKNOWN NODE TYPE: " .. tostring(v.node_type))
            end
            v.node_type = v.node_type
            v.parent_layer = layer

            if selected_elements and (table.arrayfind(selected_elements, v.id) ~= nil) then
                
                table.insert(new_selected_elements, v)
            end

        end
    end

    self.next_id = max_id
    self:Refresh()
    self:ClearSelection()
    
    for k,v in pairs(new_selected_elements) do
        self:SelectElement(v, true)
    end

    local invalid_elements = self.plax:GetInvalidElements()
    if #invalid_elements > 0 then
        local err = { string.format( "%d missing textures referenced.  See log.", #invalid_elements) }
        for i = 1, math.min( 10, #invalid_elements ) do
            local v = invalid_elements[i]
            table.insert( err, string.format( "\t%s", tostring(v.img)))
        end
        LOGWARN( table.concat(err, "\n") )
    end

    TheGame:FE():RefreshMusic()
end

function PlaxEditor:SetPlaxPercent(p)
    self.plax_val = clamp(p, -1, 1)
    self.plax:SetPlaxShift(self.plax_val)
end

function PlaxEditor:SetPendingChange(id)
    if self.last_id ~= id or id == nil then
        self:PushState()
    end
    self.last_id = id
    self.time_to_pending = 2
    self.pending = true
    self.dirty = true
end


function PlaxEditor:SetPanCoordinates(pan_x, pan_y)
    local zoom = self.zoom_scale or 1
    local border = 200
    local sizex = (RES_X + border) * zoom
    local sizey = (RES_Y + border) * zoom

    self.pan_x = clamp(pan_x or 0, -sizex, sizex)
    self.pan_y = clamp(pan_y or 0, -sizey, sizey)
    local x = self.pan_x
    local y = self.pan_y
    if CheckBits(self.aspect_framing or 0, ASPECT_FRAMING.BATTLE) then
        y = y - FIGHT_PLAX_VOFFSET
    end
    self:FindChild("SCALE"):SetPos(x, y):SetScale(zoom)
end



function PlaxEditor:OnUpdate( dt )

    self.plax:SetPlaxShift(self.plaxpanval or 0)
    if self.saving then
        return false
    end


    local mouse_down = TheGame:GetInput():IsControlDown( Controls.Digital.MENU_ACCEPT )

    if not mouse_down and self.pending then
        self.time_to_pending = self.time_to_pending - dt
        if self.time_to_pending <= 0 then
            self.pending = false
            self:PushState()
        end
    end

    local x, y = TheGame:GetInput():GetMousePos()
    
    
    if not self.dragging and self.start_mouse_x and self.start_mouse_y and self.can_drag then
        if mouse_down and ( math.abs(x - self.start_mouse_x) > 4 or math.abs(y - self.start_mouse_y) > 4) then
            self:StartDrag()
        end
    end

    if self.dragging then
        if not mouse_down then
            self:StopDrag()
        else
            local dx, dy = x - self.start_mouse_x, y - self.start_mouse_y
            local scale = self.zoom_scale and 1 / self.zoom_scale or 1
            

            if TheGame:GetInput():IsControlDown( Controls.Digital.PLAX_EDIT_PAN ) then
                
                self:SetPanCoordinates(self.start_pan_x + dx, self.start_pan_y + dy)

            else
                
                local select_center_x, select_center_y = 0, 0
                if TheGame:GetInput():IsModifierControl() then
                    for k,v in pairs(self.drag_element_state) do
                        select_center_x = select_center_x + v.x
                        select_center_y = select_center_y + v.y
                    end
                    select_center_x = select_center_x / #self.drag_element_state
                    select_center_y = select_center_y / #self.drag_element_state
                end

                
                for k,v in ipairs(self.selected_elements) do
                    if v.node_type ~= "LAYER" then
                        local is_path = v.node_type == NODE_TYPE.PATH
                        local is_location = v.node_type == NODE_TYPE.LOCATION
                        if TheGame:GetInput():IsModifierControl() and not is_path and not is_location then
                            local scale = self.drag_element_state[k].scale or 1
                            self:SetElemRelScale(v, (1+dy/100), scale, self.drag_element_state[k].x, self.drag_element_state[k].y, select_center_x, select_center_y)
                        elseif TheGame:GetInput():IsModifierAlt() and not is_path and not is_location then
                            local rot = self.drag_element_state[k].rot or 0
                            self:SetElemRotation(v, rot + dy*.005)
                        else
                            local x, y = self.drag_element_state[k].x, self.drag_element_state[k].y
                            
                            local set_sub_pos = false
                            if v.node_type == NODE_TYPE.PATH then
                                
                                local widget = self.plax:GetWidgetByElement(v)
                                local node = widget and widget:GetSelectedNode()
                                if node then
                                    self:SetSubElemPos(v, x + dx*scale, y + dy*scale)
                                    set_sub_pos = true
                                end
                            end
                            if not set_sub_pos then
                                self:SetElemPos(v, x + dx*scale, y + dy*scale)
                            end
                        end
                    end
                    
                end
            end
        end
    end

    if self.popup and self.popup.removed then
        self.popup = nil
    end
end

function PlaxEditor:IsEditingPath()
    return #self.selected_elements == 1 and self.selected_elements[1].node_type == NODE_TYPE.PATH
end

function PlaxEditor:SetElemScale(elem, scale)
    
    if not math.fequal(scale, elem.scale or 1) then
        scale = math.max(.01, scale)
        if elem and elem.node_type ~= "LAYER" then
            if elem.scale ~= scale then
                elem.scale = scale
                local widget = self.plax:GetWidgetByElement(elem)
                if widget then
                    widget:SetScale(elem.scale * (elem.flip and -1 or 1), elem.scale)
                end
                self:SetPendingChange("SCALE")
            end
        end
    end
end

function PlaxEditor:SetUVSpeed(elem, uspeed, vspeed)

    if elem and elem.node_type == "IMAGE" then

        local set_u = uspeed or elem.uspeed or 0
        local set_v = vspeed or elem.vspeed or 0

        if elem.uspeed ~= set_u or elem.vspeed ~= set_v then
            elem.uspeed = set_u
            elem.vspeed = set_v
            local widget = self.plax:GetWidgetByElement(elem)
            if widget and widget.SetUVSpeed then
                widget:SetUVSpeed(elem.uspeed, elem.vspeed)
            end
            self:SetPendingChange("UVSPEED")
        end
    end
end


function PlaxEditor:ResetUVAnim(elem)

    if elem and elem.node_type == "IMAGE" then
        local widget = self.plax:GetWidgetByElement(elem)
        if widget and widget.ResetUVAnim then
            widget:ResetUVAnim()
        end

        elem.uspeed = 0
        elem.vspeed = 0

        self:SetPendingChange("UVSPEED")
    end
end



function PlaxEditor:SetElemRelScale(elem, rel_scale, base_scale, base_x, base_y, center_x, center_y)
    --local elem = self.plax:GetElementForWidget(self.selected_widget)
    
    local widget = self.plax:GetWidgetByElement(elem)
    assert(widget)    
    if widget then

        local dx, dy = ((base_x or 0) - center_x)*rel_scale, ((base_y or 0) - center_y)*rel_scale
        elem.x, elem.y = center_x + dx, center_y + dy
        widget:SetPos(elem.x, elem.y)
        elem.scale = math.max(.01, base_scale*rel_scale)
        widget:SetScale((elem.flip and -1 or 1)*elem.scale, elem.scale)
        self:SetPendingChange("SCALE")
    end
end

function PlaxEditor:SetElemRotation(elem, rot)
    --local elem = self.plax:GetElementForWidget(self.selected_widget)
    if not math.fequal(rot, elem.rot or 0) then
        local widget = self.plax:GetWidgetByElement(elem)
        if widget then
            
            if elem and elem.node_type ~= "LAYER" then
                if elem.rot ~= rot then
                    elem.rot = rot
                    widget:SetRotation(rot)
                    self:SetPendingChange("ROT")
                end
            end
        end
    end
end

function PlaxEditor:SetSubElemPos(elem, x, y)
    local widget = self.plax:GetWidgetByElement(elem)
    if widget and widget.SetSubWidgetPos then
        if elem then
            local set_x = math.floor(x or 0)
            local set_y = math.floor(y or 0)


            if widget:SetSubWidgetPos(set_x,set_y) then
                elem.nodes, elem.edges = widget.graph:GetSaveTables()
                self:SetPendingChange("SUB_POS")
            end
        end
    end
end

function PlaxEditor:SetElemPos(elem, x, y)
    --local elem = self.plax:GetElementForWidget(self.selected_widget)
    local widget = self.plax:GetWidgetByElement(elem)
    if widget then
        if elem then
            if elem.node_type ~= "LAYER" then
                local set_x = math.floor(x or elem.x or 0)
                local set_y = math.floor(y or elem.y or 0)

                if elem.x ~= set_x or elem.y ~= set_y then
                    elem.x = set_x
                    elem.y = set_y
                    widget:SetPos(set_x,set_y)
                    --self.plax:UpdateClickableElement(elem)
                    self:SetPendingChange("POS")
                end
            end
        end
    end
end

function PlaxEditor:SetElemEnterPos(elem, x, y)
    --local elem = self.plax:GetElementForWidget(self.selected_widget)
    local widget = self.plax:GetWidgetByElement(elem)
    if widget then
        if elem then
            if elem.node_type ~= "LAYER" then
                local set_x = math.floor(x or elem.enter_x or 0)
                local set_y = math.floor(y or elem.enter_y or 0)

                if elem.enter_x ~= set_x or elem.enter_y ~= set_y then
                    
                    elem.enter_x = set_x
                    elem.enter_y = set_y
                    self:SetPendingChange("ENTER_POS")
                end

                if not widget.entry_handle then
                    widget.entry_handle = widget:AddChild(Widget.Image(global_images.circle, 16,16):SetTintColour(0xff0000af))
                end
                widget.entry_handle:SetPos(elem.enter_x, elem.enter_y)
            end
        end
    end
end


function PlaxEditor:Nudge(dx,dy)
    
    if TheGame:GetInput():IsModifierShift() then
        dx = dx * 10
        dy = dy * 10
    end
    for k,elem in pairs( self.selected_elements ) do
        --local elem = self.plax:GetElementForWidget(self.selected_widget)
        if elem and elem.node_type ~= "LAYER" then
            local widget = self.plax:GetWidgetByElement(elem)
            if widget then
                local x = math.floor((elem.x or 0) + dx)
                local y = math.floor((elem.y or 0) + dy)
                elem.x = x
                elem.y = y

                widget:SetPos(x,y)
                self:SetPendingChange("POS")
            end
        end
    end
    
end

function PlaxEditor:DeleteSelected()

    if #self.selected_elements > 0 then
        

        for k,elem in pairs( self.selected_elements ) do
                
                if table.arrayremove(self.plaxdata.layers, elem) then
                    self.selected_layer = nil
                else
                for k,v in pairs(self.plaxdata.layers) do
                    if table.arrayremove(v.elements, elem) then
                        break
                    end
                end
            end
        end
        table.clear(self.selected_elements)
        self:PushState()
        self:Refresh()
        return
    end
end


function PlaxEditor:UpdateElemShader( elem, w )
    if elem.node_type == NODE_TYPE.ANIMATION then
        if elem.erode then
            w:SetEffect( global_shaders.FX_ANIM_ERODE )
        elseif elem.blend_mode == BLEND_MODE.MULTIPLY then
            w:SetEffect( global_shaders.FX_ANIM_PREMULT )
        elseif w:HasNonIdentityHSVSettings() then
            w:SetEffect(global_shaders.UI_ANIM_HSV)
        else
            w:SetEffect( global_shaders.UI_ANIM )
        end
    end
end

function PlaxEditor:SetElemBlendMode(elem, blend_mode)
    elem.blend_mode = blend_mode
    local w = self.plax:GetWidgetByElement(elem)
    if w then
        w:SetBlendMode(blend_mode)
        self:UpdateElemShader( elem, w )
    end
    self:PushState()
end

function PlaxEditor:SetElemErode(elem, b)
    elem.erode = b
    local w = self.plax:GetWidgetByElement(elem)
    if w then
        self:UpdateElemShader(elem, w)
    end
    self:PushState()
end


function PlaxEditor:RefreshSlot(elem)
    for slot_id, slot in self.plax:Slots() do
        if slot.data == elem then
            self.plax:ClearSlot( slot_id )
            if slot.data.preview_agent and self.agent_data_by_name[ slot.data.preview_agent ] then
                local agent = self.agent_data_by_name[ slot.data.preview_agent ].agent
                self.plax:FillSlot(agent, slot_id)
            end
            break
        end
    end
end

function PlaxEditor:SetPreviewCharacter(elem, name)
    if elem.preview_agent ~= name then
        elem.preview_agent = name
        elem.preview_texture = nil
        self:RefreshSlot(elem)
        self:PushState()
    end
end


function PlaxEditor:SetAnimSpeed(elem, v)
    if elem.anim_speed ~= v then
        elem.anim_speed = v
        self:SetPendingChange("SPEED")
        local w = self.plax:GetWidgetByElement(elem)
        if w then
            w:SetSpeed(elem.anim_speed/100)
        end
    end
end

function PlaxEditor:SetAnimPhase(elem, v)
    if elem.anim_phase ~= v then
        elem.anim_phase = v
        self:SetPendingChange("PHASE")
        local w = self.plax:GetWidgetByElement(elem)
        if w then
            w:SetPhase(elem.anim_phase / 100)
        end
    end
end

function PlaxEditor:SetAnimRepeatDelayMin(elem, v)
    if elem.anim_repeat_delay_min ~= v then
        elem.anim_repeat_delay_min = v
        self:SetPendingChange("REPEATDELAYMIN")
        local w = self.plax:GetWidgetByElement(elem)
        if w then
            self.plax:PlayRepeatingAnim(w, elem.anim or "anim", elem.anim_repeat_delay_min, elem.anim_repeat_delay_max)
        end
    end
end

function PlaxEditor:SetAnimRepeatDelayMax(elem, v)
    if elem.anim_repeat_delay_max ~= v then
        elem.anim_repeat_delay_max = v
        self:SetPendingChange("REPEATDELAYMAX")
        local w = self.plax:GetWidgetByElement(elem)
        if w then
            self.plax:PlayRepeatingAnim(w, elem.anim or "anim", elem.anim_repeat_delay_min, elem.anim_repeat_delay_max)
        end
    end
end

function PlaxEditor:SetAnimTimeOffset(elem, v)
    if (elem.anim_initial_delay or 0) ~= v then
        elem.anim_initial_delay = v
        self:SetPendingChange("TIMEOFFSET")
        local w = self.plax:GetWidgetByElement(elem)
        if w then
            self.plax:PlayRepeatingAnim(w, elem.anim or "anim", elem.anim_repeat_delay_min, elem.anim_repeat_delay_max, elem.anim_initial_delay)
        end
    end
end

function PlaxEditor:OnClickAnimFile()
    
    local base = path.normpath( path.join(BASE_PATH, "anim"))
    local file = engine.inst:OpenDialog(nil, base)--"zip", path)
    if file then
        self.selected_elements[1].animfile = path.relpath (file,DATA_DIR)

        self.selected_elements[1].anim = nil
        if self.selected_elements[1].animfile then
            local animdat = GetAnim(self.selected_elements[1].animfile)
            if animdat and animdat:GetNumAnims() > 0 then
                self.selected_elements[1].anim = animdat:GetAnimName(0)
            end
        end
        self:PushState()
        self:Refresh()
    end

end
function PlaxEditor:OnClickBuildFile()
    
    local base = path.normpath( path.join(BASE_PATH, "anim"))
    local file = engine.inst:OpenDialog(nil, base)--"zip", path)
    if file then
        self.selected_elements[1].buildfile = path.relpath (file,DATA_DIR)
        self:PushState()
        self:Refresh()
    end

end



function PlaxEditor:OnClickPropFile(elem, idx)
    local base = path.normpath( path.join(BASE_PATH, "anim"))
    local file = engine.inst:OpenDialog(nil, base)--"zip", path)
    if file then
        
        if idx == 1 then
            elem.prop_file = path.relpath (file,DATA_DIR)
        else
            elem.prop_file2 = path.relpath (file,DATA_DIR)
        end
        self:PushState()
        
        self:RefreshSlot(elem)
    end
end

local minZoom = 0.125
local maxZoom = 3.0
local zoom_sensitivity = 0.1

function PlaxEditor:DeltaZoom(delta)

    -- Zoom in and out to the current mouse cursor position:
    local mx, my = TheGame:GetInput():GetMousePos()

    local lx, ly = self:FindChild("SCALE"):TransformFromWorld(mx, my)
    --print("local pos = " .. lx .. ", " .. ly );


    local zoomfactor = clamp(1.0 + (delta * zoom_sensitivity), 0.5, 2.0);   -- Allow only 50%-200% size changes in one step
    self.zoom_scale = clamp((self.zoom_scale or 1.0) * zoomfactor, minZoom, maxZoom)    -- Clamp the modified zoom scale to minZoom and maxZoom
    self:FindChild("SCALE"):SetScale(self.zoom_scale)

    local lx2, ly2 = self:FindChild("SCALE"):TransformFromWorld(mx, my)
    local dx = (lx2 - lx) * self.zoom_scale;
    local dy = (ly2 - ly) * self.zoom_scale;

    self:SetPanCoordinates((self.pan_x or 0) + dx, (self.pan_y or 0) + dy)
end


function PlaxEditor:HandleControlUp(control)
    if self.saving then
        return false
    end

    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        self.can_drag = false
        local ignore_up = self.ignore_select_up
        self.ignore_select_up = nil
        if not ignore_up and not self.dragging and not TheGame:GetInput():IsControlDown( Controls.Digital.PLAX_EDIT_PAN )  then
            self:SelectNextUnderPoint(TheGame:GetInput():GetMousePos())
            return true
        end
    end

end


function PlaxEditor:HandleControlDown(control)
    if self.saving then
        return false
    end

    if control:Has( Controls.Digital.ZOOM_IN_STEP ) then
        self:DeltaZoom(1)
        return true
    elseif control:Has( Controls.Digital.ZOOM_OUT_STEP ) then
        self:DeltaZoom(-1)
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_VIEW_NATURAL) then
        self:SetPanCoordinates(0, 0)

    elseif control:Has( Controls.Digital.MENU_ALT ) then
        if self:IsEditingPath() then
            self:PathMenuAtPoint(TheGame:GetInput():GetMousePos())
        else
            self:SelectionMenuAtPoint(TheGame:GetInput():GetMousePos())
        end
        
        return true
    elseif control:Has( Controls.Digital.MENU_ACCEPT ) then
        self.can_drag = true
        self.start_mouse_x, self.start_mouse_y = TheGame:GetInput():GetMousePos()
        local popup_valid = self.popup
        if self.popup then
            self.ignore_select_up = true
        end
        
        if not self.dragging and not popup_valid and not TheGame:GetInput():IsControlDown( Controls.Digital.PLAX_EDIT_PAN ) then
            if self:SelectNextUnderPoint(TheGame:GetInput():GetMousePos()) then
                self.can_drag = false   -- If the selection changed with this click, don't allow dragging.
            end

            self.ignore_select_up = true
        end
        
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_SAVE ) then
        if self.dirty then
            self:DoSave()
        end
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_DESELECT ) then
        self:DoDeselect()
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_SELECT_ALL ) then
        self:DoSelectAll()
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_COPY ) then
        self:DoCopy(TheGame:GetInput():GetMousePos())
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_CUT ) then
        self:DoCut(TheGame:GetInput():GetMousePos())
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_PASTE ) then
        self:DoPaste()
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_REDO ) then
        self:Redo()
        return true
    elseif control:Has( Controls.Digital.PLAX_EDIT_UNDO ) then
        self:Undo()
        return true
    elseif control:Has( Controls.Digital.MENU_LEFT ) then
        self:Nudge(-1, 0)
        return true
    elseif control:Has( Controls.Digital.MENU_RIGHT ) then
        self:Nudge(1, 0)
        return true
    elseif control:Has( Controls.Digital.MENU_UP ) then
        if TheGame:GetInput():IsModifierAlt() then
            self:MoveLayer(1)
        elseif TheGame:GetInput():IsModifierControl() then
            self:MoveLayerOrder(1)
        else
            self:Nudge(0, 1)
        end
        return true
    elseif control:Has( Controls.Digital.MENU_DOWN ) then
        if TheGame:GetInput():IsModifierAlt() then
            self:MoveLayer(-1)
        elseif TheGame:GetInput():IsModifierControl() then
            self:MoveLayerOrder(-1)
        else
            self:Nudge(0, -1)
        end
        return true
    elseif control:Has( Controls.Digital.MENU_DELETE ) then
        self:DeleteSelected()
        return true
    end
end

function PlaxEditor:ClearSelection()
    for k,v in pairs(self.selected_elements) do
        local w = self.plax:GetWidgetByElement(v)
        if w and w.bounding_widget then
            w.bounding_widget:Hide()
        end
    end

    table.clear(self.selected_elements)


end

function PlaxEditor:SelectElement(element, val)
    if not val then
        table.arrayremove(self.selected_elements, element)
    else
        table.insert_unique(self.selected_elements, element)
    end

    local w = self.plax:GetWidgetByElement(element)
    if w then
        self:UpdateWidgetBounds(w)
        if val then
            w.bounding_widget:Show()
        else
            w.bounding_widget:Hide()
        end
    end

    if val then
        self:SelectLayer(element.parent_layer)
    end
end

function PlaxEditor:SelectRange(element, union)
    local last_selected = self.selected_elements[ #self.selected_elements ]
    if last_selected == nil then
        self:SelectElement(element, true)
        return
    end

    if not union then
        self:ClearSelection()
    end

    local from_layer_idx = table.arrayfind( self.plaxdata.layers, last_selected.parent_layer )
    local to_layer_idx = table.arrayfind( self.plaxdata.layers, element.parent_layer )
    if from_layer_idx and to_layer_idx then
        local delta = from_layer_idx > to_layer_idx and -1 or 1

        -- Loop through each layer and element therein, selecting as we go.
        for layer_idx = from_layer_idx, to_layer_idx, delta do
            local layer = self.plaxdata.layers[ layer_idx ]
            local from_idx, to_idx
            if last_selected.parent_layer == layer then
                from_idx = table.arrayfind( layer.elements, last_selected )
            end
            if element.parent_layer == layer then
                to_idx = table.arrayfind( layer.elements, element )
            end

            if from_idx == nil and to_idx == nil then
                from_idx, to_idx = 1, #layer.elements
            elseif from_idx == nil then
                from_idx = delta > 0 and 1 or #layer.elements
            elseif to_idx == nil then
                to_idx = delta > 0 and #layer.elements or 1
            end

            local delta = from_idx > to_idx and -1 or 1
            for i = from_idx, to_idx, delta do
                self:SelectElement( layer.elements[ i ], true)
            end
        end
    end
end

function PlaxEditor:DoSelectAll()
    self:ClearSelection()
    for i, layer in pairs(self.plaxdata.layers) do
        for k,v in pairs(layer.elements) do
            self:SelectElement(v, true)
        end
    end
end

function PlaxEditor:DoDeselect()
    self:ClearSelection()
end

function PlaxEditor:PathMenuAtPoint(x, y)    
    self.open_select_popup = "PATH_MENU"
end

function PlaxEditor:SelectionMenuAtPoint(x, y)    
    self.open_select_popup = "SELECT_UNDER_MOUSE"
    self.items_under_mouse = {}
    self.plax:GetHitsUnderPoint(self.items_under_mouse, x,y)
end


function PlaxEditor:SelectNextUnderPoint(x, y)
    local t = {}
    self.plax:GetHitsUnderPoint(t, x,y)

    local to_select

    -- If elements are currently selected, do NOT change the selection:
    local smallestBBArea
    for k,v in pairs(t) do
        if table.arrayfind(self.selected_elements, v.element) then
            -- Normally it would return here right away. However, because paths have a sub selection thing going on, forward the select click to the path (kinda hacky):
            for k,v in ipairs(self.selected_elements) do
                if v.node_type == NODE_TYPE.PATH then
                    local widget = self.plax:GetWidgetByElement(v)
                    widget:SetSubSelection(x,y)
                end
            end

            return
        end        

        local area = v.widget:GetBoundingBoxArea()
        if not smallestBBArea or area < smallestBBArea then
            smallestBBArea = area
            to_select = v.element
        end
    end


    -- If the none of the elements under the mouse are in the current selection, select the one with the smallest AABB        
    if to_select then
        local was_selected = table.arrayfind(self.selected_elements, to_select)
        
        if not TheGame:GetInput():IsModifierShift() and not TheGame:GetInput():IsModifierControl() then
            self:ClearSelection()
        end

        if TheGame:GetInput():IsModifierControl() then
            self:SelectElement(to_select, not was_selected)
        else
            self:SelectElement(to_select, true)
            if to_select.node_type == NODE_TYPE.PATH then
                local widget = self.plax:GetWidgetByElement(to_select)
                widget:SetSubSelection(x,y)
            end
        end
    else
        self:ClearSelection()
    end

    return true;
end


function PlaxEditor:StartDrag()
    local x, y = TheGame:GetInput():GetMousePos()

    table.clear(self.drag_element_state)
    if #self.selected_elements > 0 then
        for k,v in ipairs(self.selected_elements) do
            
            local w = self.plax:GetWidgetByElement(v)
            local x,y = v.x, v.y
            if w and w.GetSubWidgetPos then
                local sx,sy = w:GetSubWidgetPos()
                x = x + (sx or 0)
                y = y + (sy or 0)
            end
            table.insert(self.drag_element_state, {scale = v.scale or 1, x = x or 0, y = y or 0, rot = v.rot or 0})
        end
    end
    
    self.start_pan_x = self.pan_x or 0
    self.start_pan_y = self.pan_y or 0
    self.dragging = true

end

function PlaxEditor:StopDrag()
    self.dragging = false
    table.clear(self.drag_element_state)

    if self.pending then
        self:PushState()
        self.pending = false
    end
end

function PlaxEditor:ShowAspectFraming( framing )
    if self.aspect_framing ~= framing then
        if self.frame_widget_convo then
            self.frame_widget_convo:Remove()
            self.frame_widget_convo = nil
        end
        if self.frame_widget_battle then
            self.frame_widget_battle:Remove()
            self.frame_widget_battle = nil
        end

        self.aspect_framing = framing

        if CheckBits(framing, ASPECT_FRAMING.CONVO) then
            self.frame_widget_convo = Widget():AddChildren{
                    Widget.PolyLine():CreateRect(RES_X, RES_X*(3/4)):SetTintColour(0xC92BD3FF),
                    Widget.PolyLine():CreateRect(RES_Y*(21/9), RES_Y):SetTintColour(0x2BC9D3FF),
                    Widget.PolyLine():CreateRect(RES_X, RES_X*(10/16)):SetTintColour(0xC92BD3FF),
                    Widget.PolyLine():CreateRect(RES_X, RES_Y):SetTintColour(0xC9D32BFF),
            }

            self:FindChild("SCALE"):AddChild( self.frame_widget_convo )
        end

        if CheckBits(framing, ASPECT_FRAMING.BATTLE) then
            self.frame_widget_battle = Widget():AddChildren{
                    Widget.PolyLine():CreateRect(RES_X, RES_X*(3/4)):SetTintColour(0xC92BD3FF),
                    Widget.PolyLine():CreateRect(RES_Y*(21/9), RES_Y):SetTintColour(0x2BC9D3FF),
                    Widget.PolyLine():CreateRect(RES_X, RES_X*(10/16)):SetTintColour(0xC92BD3FF),
                    Widget.PolyLine():CreateRect(RES_X, RES_Y):SetTintColour(0xC9D32BFF),
            }

            self.frame_widget_battle:Offset( 0, FIGHT_PLAX_VOFFSET )

            self:FindChild("SCALE"):AddChild( self.frame_widget_battle )
        end
    end
end

function PlaxEditor:RenderSettingsMenu(ui)

    ui:Spacing()

    ui:BeginChild( "Import Settings", 300, 50 )

    local val = 1.0
	local data = DebugUtil.GetLocalDebugData( "PLAX_SETTINGS" )
	if data then
        val = data.DefaultImageScale or 1.0
	end

    local scale = ui:DragFloat("Import Scale", val or 1.0, 0.0, 10.0, "%0.2f", .01)
    if math.abs(scale - val) > .001 then
        data = {}
        data.DefaultImageScale = clamp(scale, 0.0, 10.0)
        DebugUtil.SetLocalDebugData("PLAX_SETTINGS", data)
    end
    ui:EndChild()

    ui:Spacing()
    
    if ui:Button( "OK", 100 ) then
        DebugUtil.SaveLocalDebugData()
        ui:CloseCurrentPopup()
    end

end

function PlaxEditor:RenderSaveError(ui)

    ui:Spacing()

    ui:BeginChild( "Save Error", 250, 50 )

    if type(self.errorSavingMsg) == "string" then
        ui:Text( self.errorSavingMsg )
    else
        ui:Text( "There was a problem saving the file!" )
    end

    ui:EndChild()

    if ui:Button( "OK", 100 ) then
        ui:CloseCurrentPopup()
    end
end


function PlaxEditor:UpdateSnapshotList()
    self.snapshot_list = {}
    for k,v in pairs(sColourSettings) do
        table.insert(self.snapshot_list, k)
    end
    table.sort (self.snapshot_list)
end



-- returns true if changed
function PlaxEditor:HSVAdjust(ui, name, vals)
    local returnvalue = false

    local h,s,v = table.unpack(vals.hsv)
    local r,g,b = table.unpack(vals.colour)
    
    ui:Text(name)
    ui:PushID(name)

	ui:PushItemWidth(-60)	-- Use full column width

    local colour = MakeHexColour(r,g,b,1)
    local new_colour = ui:ColorSelector( name, colour, false ) 
    if new_colour ~= colour then
        local r,g,b,a = HexColour(new_colour)
        self:SetPendingChange(name .. "RGB")
        vals.colour[1] = r
        vals.colour[2] = g
        vals.colour[3] = b
        returnvalue = true
    end


    local new_h = ui:SliderFloat("H", h, 0, 1)
    local new_s = ui:SliderFloat("S", s, 0, 3)
    local new_v = ui:SliderFloat("V", v, 0, 3)

	ui:PopItemWidth()

    ui:PopID()
    
    if new_h ~= h or new_s ~= s or new_v ~= v then
        self:SetPendingChange(name .. "HSV")
        vals.hsv[1] = new_h
        vals.hsv[2] = new_s
        vals.hsv[3] = new_v
        returnvalue = true
    end

    return returnvalue
end



function PlaxEditor:ReloadColourSettings()
    package.loaded[ "content/colour_settings" ] = nil
    sColourSettings = require "content/colour_settings"
    self:UpdateSnapshotList()
end

function PlaxEditor:SaveColourSettings()
    local filename = 'scripts/content/colour_settings.lua'

    local dumps = serpent.dump(sColourSettings, { indent = '  ', sortkeys = true, compact = false, sparse = false } )

    engine.data:SaveFile( filename, dumps, function(ok) 
            print( string.format( "Write colour_settings file %s : '%s'", ok and "OK" or "FAILED", filename))
        end )

end



function PlaxEditor:LightingPicker(ui, colour_name, ignore_parent, comboName, editButtonWidth, showIgnoreParent )
    colour_name = colour_name or "DEFAULT"
    local ignoreParent = ignore_parent or false
  
    if not self.snapshot_list then
        self:UpdateSnapshotList()
    end
        
    local snapshot_idx = ui:Combo( "HSV Setting##"..comboName, table.arrayfind(self.snapshot_list, colour_name) or 1, self.snapshot_list )
    local new_colour
    local new_ignoreParent

    if self.snapshot_list[snapshot_idx] ~= colour_name then
         new_colour = self.snapshot_list[snapshot_idx]
    end

    if showIgnoreParent then          
        ui:SameLine(nil, 5.0)
        if ui:Checkbox("Ignore Parent##"..comboName, ignoreParent) then
            new_ignoreParent = not ignoreParent
        end
    end

    local colname = new_colour or colour_name

    local buttonColumnWidth = 70

    if ui:Button("New##"..comboName, buttonColumnWidth) then
        self.new_name = ""
        self.open_select_popup = "NEW_SNAPSHOT_POPUP"
        --ui:OpenPopup("NEW_SNAPSHOT_POPUP")     -- This doesn't work for some reason
    end

    if colname and colname ~= "DEFAULT" then    -- Don't allow changing the default colour setting
        local changedHSV

        ui:SameLine(nil, 5)
        if ui:Button("Reset##"..comboName, buttonColumnWidth) then
            if colname then
                sColourSettings[colname] = CreateNewSettings()
            end
            changedHSV = true
        end

    
        ui:SameLine(nil, 5 )
        if ui:Button("Save##"..comboName, buttonColumnWidth) then
            self:SaveColourSettings()
        end

        ui:SameLine(nil, 5 )
        if ui:Button("Revert##"..comboName, buttonColumnWidth) then
            self:ReloadColourSettings()
            changedHSV = true
        end
        
        
        if sColourSettings[colname] then
            ui:Spacing()

            if self:HSVAdjust(ui, "Shadows", sColourSettings[colname].shadows) then
                changedHSV = true
            end
            
            ui:Spacing()
            
            if self:HSVAdjust(ui, "Midtones", sColourSettings[colname].midtones) then
                changedHSV = true
            end

            ui:Spacing()

            if self:HSVAdjust(ui, "Highlights", sColourSettings[colname].highlights) then
                changedHSV = true
            end

            ui:Spacing()

        end

        if changedHSV then
            self.plax:UpdateAllHSV()
        end    
    end

    return new_colour, new_ignoreParent
end



function PlaxEditor:SwitchScreens(fn, force)
    if force then
        print("SwitchScreens is called with force = true");
    else
        print("SwitchScreens is called with force = false");
    end

    if not force and self.dirty then
        self:SuggestSave(function(v) if v then self:SwitchScreens(fn, true) end end)
        return
    end

    fn()
end


function PlaxEditor:RenderPlaxScriptsPanel( ui )
    for i, v in ipairs( self.plaxdata.scripts or table.empty ) do
        local filename = string.format( "scripts/content/plax_scripts/%s.lua", v.filename )
        local env = { plax = self.plax }
        local f, err = loadfile( filename, nil, env )
        if not f then
            ui:TextColored( 0xFF0000FF, v.filename )
            if ui:IsItemHovered() then
                ui:SetTooltip( tostring(err) )
            end

        else
            ui:Text( v.filename )
        end

        ui:SameLine( nil, 20 )
        if ui:Button( "Delete" ) then
            table.remove( self.plaxdata.scripts, i )
            self:SetPendingChange()
            break
        end
    end

    if ui:Button( "New Script..." ) then
        ui:OpenPopup( "NEW_SCRIPT" )
    end

    if self.plaxdata.scripts and #self.plaxdata.scripts > 0 and ui:Button( "Run" ) then
        self.plax:RunScripts( self.plaxdata.scripts )
    end
    if self.plax.script_coros and #self.plax.script_coros > 0 then
        ui:SameLine( nil, 10 )
        ui:Text( string.format( "%d coros running", #self.plax.script_coros ))
    end

    if ui:BeginPopup( "NEW_SCRIPT" ) then
        local filenames = {}
        for k, filename in pairs(engine.inst:ListFiles( 'scripts/content/plax_scripts/*.lua')) do
            table.insert( filenames, filename:match( "^(.+)[.]lua$" ))
        end

        local idx = ui:Combo( "Filename", 0, filenames )
        if idx and filenames[ idx ] then
            if self.plaxdata.scripts == nil then
                self.plaxdata.scripts = {}
            end

            table.insert( self.plaxdata.scripts, { filename = filenames[ idx ] } )
            self:SetPendingChange()
        end
        ui:EndPopup()
    end
end

function PlaxEditor:RenderMainPlaxPanel(ui)
    ui:PushItemWidth(-155)  -- Use full column width

    self.plaxpanval = ui:SliderFloat("PLAX", self.plaxpanval or 0, -1, 1)
              
    if ui:Checkbox( "Post Process", self.plax:IsPostProcess() ) then
        self.plax:PostProcess( not self.plax:IsPostProcess() )
    end


    if #self.ambience_events > 1 then
        ui:Separator()
        ui:Spacing()

        local ambience_idx = self.plaxdata.ambient_event and table.arrayfind(self.ambience_events, self.plaxdata.ambient_event) or 1
        local new_ambient_idx = ui:Combo("Ambience", ambience_idx, self.ambience_txt)
        
        if new_ambient_idx and new_ambient_idx ~= ambience_idx then
            self.plaxdata.ambient_event = self.ambience_events[new_ambient_idx]
            self.plax.ambience = self.plaxdata.ambient_event
            self.plax:PlaySounds()
            self:PushState()
        end
    end

    if #self.music_events > 1 then
        ui:Separator()
        ui:Spacing()

        local music_idx = self.plaxdata.music_event and table.arrayfind(self.music_events, self.plaxdata.music_event) or 1
        local new_music_idx = ui:Combo("Scene Music", music_idx, self.music_txt)
        if new_music_idx and new_music_idx ~= music_idx then
            self.plaxdata.music_event = self.music_events[new_music_idx]
            self.plax.music = self.plaxdata.music_event
            TheGame:FE():RefreshMusic()
            self:PushState()
        end
    end

    if #self.door_events > 1 then
        ui:Separator()
        ui:Spacing()

        local door_idx = self.plaxdata.entry_event and table.arrayfind(self.door_events, self.plaxdata.entry_event) or 1
        local new_door_idx = ui:Combo("Entry event", door_idx, self.door_event_txt)
        
        if new_door_idx and new_door_idx ~= door_idx then
            self.plaxdata.entry_event = self.door_events[new_door_idx]
            self:PushState()
        end
        ui:Separator()
        ui:Spacing()

        local door_idx = self.plaxdata.exit_event and table.arrayfind(self.door_events, self.plaxdata.exit_event) or 1
        local new_door_idx = ui:Combo("Exit event", door_idx, self.door_event_txt)
        
        if new_door_idx and new_door_idx ~= door_idx then
            self.plaxdata.exit_event = self.door_events[new_door_idx]
            self:PushState()
        end

    end


    ui:Separator()
    ui:Spacing()

    local mode_idx = table.arrayfind(game_modes, self.game_mode or GAME_MODE.CONVERSATION)
    local new_mode_idx = ui:Combo("Game Mode", mode_idx, game_mode_txt)
    if mode_idx ~= new_mode_idx then
        self:SetGameMode(game_modes[new_mode_idx])
    end


    local phase_idx = table.arrayfind(day_phases, self.day_phase or DAY_PHASE.DAY)
    local new_phase_idx = ui:Combo("Day Phase", phase_idx, day_phase_txt)
    if phase_idx ~= new_phase_idx then
        self:SetDayPhase(day_phases[new_phase_idx])
    end

    ui:Spacing()
    
    local lighting_name, lighting_ignore_parent = self:GetDayPhaseName()

    local selected_lighting = self.plaxdata[lighting_name] or "DEFAULT"
    if ui:TreeNode("HSV - " .. selected_lighting .. "###LightingPlax" ) then
   
        local new_colour = self:LightingPicker(ui, self.plaxdata[lighting_name], false, "WholePlax", 150, false) 
        if new_colour then
            self:PushState()
            self.plaxdata[lighting_name] = new_colour
            self.dirty = true
            self.plax:SetColourNames(self.plaxdata.day_colour, self.plaxdata.night_colour)
        end 
        ui:TreePop()
    end

    if ui:TreeNode( "Scripts" ) then
        self:RenderPlaxScriptsPanel( ui )
        ui:TreePop()
    end

    ui:Separator()

    local filter = ui:InputText( "Filter", self.element_filter_txt or "" )
    if filter and filter ~= self.element_filter_txt then
        self.element_filter_txt = filter
    end

    ui:PopItemWidth()

    ui:Spacing()

    for k = #self.plaxdata.layers, 1, -1 do
        ui:PushID("LAYER"..k)
        local layer = self.plaxdata.layers[k]

        local buttonColumnWidth = 150
        
        if ui:CollapsingHeader( layer.name.."##"..k, ui.TreeFlags_DefaultOpen ) then

            if ui:TreeNode("Layer Properties") then
                local totalWidth = ui:GetContentRegionAvail();
                ui:Columns(2, "", false)    -- false to turn off manual resizing
            
                ui:SetColumnWidth(0, totalWidth-buttonColumnWidth) -- hack to make the left column resize with the window, but the right column be fixed

                ui:PushItemWidth(-70)   -- Use full column width

                local new_name = ui:InputText("Layer name ##"..k, layer.name)
                if new_name then
                    self:PushState()
                    layer.name = new_name
                end

                if ui:Checkbox("Visible", not layer.hidden_in_editor) then
                    self:SetViz(layer, layer.hidden_in_editor)
                end

                ui:SameLine(nil, 10)
                if ui:Checkbox("Guide", layer.guide) then
                    self:SetGuide(layer, not layer.guide)
                end

                local plax = ui:SliderFloat("Plax pixels", layer.plax, 0, 300)
                if math.abs(plax - layer.plax) > .001 then
                    self:SetLayerPlax(layer, plax)
                end

                ui:PopItemWidth()

                ui:NextColumn()

                if ui:Button("Delete", buttonColumnWidth) then
                    self.selected_elements = {layer}
                    self:DeleteSelected()
                end

                if ui:Button("Move Layer Up", buttonColumnWidth) then
                    if k < #self.plaxdata.layers then
                        self:PushState()
                        self:Refresh()
                        self.plaxdata.layers[k], self.plaxdata.layers[k+1] = self.plaxdata.layers[k+1], self.plaxdata.layers[k]
                    end
                end
                if ui:Button("Move Layer Down", buttonColumnWidth) then
                    if k > 1 then
                        self:PushState()
                        self:Refresh()
                        self.plaxdata.layers[k], self.plaxdata.layers[k-1] = self.plaxdata.layers[k-1], self.plaxdata.layers[k]
                    end
                end

                ui:Columns(1)

                ui:PushItemWidth(-(buttonColumnWidth + 5) - 20) -- Use full column width

                local selected_lighting = layer[lighting_name] or "DEFAULT"
                if layer[lighting_ignore_parent] then
                    selected_lighting = selected_lighting .. " (Ignore Parent)"
                end

                if ui:TreeNode("HSV - " .. selected_lighting .. "###LightingLayer" .. k ) then

                    local new_layercolour, new_ignoreParent = self:LightingPicker(ui, layer[lighting_name], layer[lighting_ignore_parent], "Layer"..k, buttonColumnWidth, true ) 
                    if (new_layercolour and (layer[lighting_name] ~= new_layercolour)) or ((new_ignoreParent ~= nil) and (layer[lighting_ignore_parent] ~= new_ignoreParent)) then
                        self:PushState()
                        if new_layercolour then
                            layer[lighting_name] = new_layercolour
                            self.dirty = true
                        end
                        if new_ignoreParent ~= nil then
                            layer[lighting_ignore_parent] = new_ignoreParent
                            self.dirty = true
                        end

                        self.plax:UpdateCurrentHSVForElement(layer)
                    end 
                    ui:TreePop()
                end

                ui:PopItemWidth()
                
                ui:Spacing()
                ui:TreePop()
            end

            ui:Separator()
            
            if layer.elements then
                ui:Indent( 10 )
                for k = #layer.elements, 1, -1 do
                    local v = layer.elements[k]
                    local display_name = GetElementDisplayName( v ) or ""
                    if self.element_filter_txt == nil or display_name:find( self.element_filter_txt, 1, true ) then
                        ui:PushID("ELEMENT"..v.id)    
                        local is_selected = table.arrayfind(self.selected_elements, v) ~= nil
                        
                        if ui:DragCheckbox(assets.locked, 16, 16, 1, v.lock) then 
                            self:SetElemLock(v, not v.lock)
                        end             

                        ui:SameLine(nil, 4)

                        local visible = not v.hidden_in_editor
                        if ui:DragCheckbox(assets.visible, 16, 16, 2, visible) then 
                            self:SetViz(v, not visible)
                        end
                        ui:SameLine(nil, 8)

                        local clr = 0xffffffff

                        if not self:IsElementVisibleForCurrentSetting(v) then
                            clr = 0x808080FF
                        end

                        if v.node_type == NODE_TYPE.IMAGE then
                            local w = self.plax:GetWidgetByElement( v )
                            if w == nil or w:IsDefaultTexture() then
                                clr = 0xCC00000FF
                            end
                        elseif v.guide then
                            clr = 0xB5C4DDFF
                        end

                        local FlagsColumnWidth = 500

                        ui:PushStyleColor( ui.Style_Text, clr )
                        if ui:Selectable( display_name, is_selected, 0, FlagsColumnWidth - 75, 0.0) then
                            if TheGame:GetInput():IsModifierShift() then
                                self:SelectRange(v, TheGame:GetInput():IsModifierControl())
                            elseif TheGame:GetInput():IsModifierControl() then
                                self:SelectElement(v, not is_selected)
                            else
                                self:ClearSelection()
                                self:SelectElement(v, true)
                            end
                        end
                        ui:PopStyleColor()
                        if ui:IsItemHovered() then
                            ui:SetTooltip( display_name .. "\nDEL to delete\nCTRL+UP/DOWN to re-order\nALT+UP/DOWN to change layers")
                        end
                        if ui:BeginPopupContextItem( "CXT_MENU" ) then
                            if #self.selected_elements > 1 then
                                ui:TextColored( 0x777777FF, string.format( "%d selected elements", #self.selected_elements ))
                            else
                                ui:TextColored( 0x777777FF, GetElementDisplayName(self.selected_elements[1]) or "???")
                            end
                            ui:Separator()
                            if ui:MenuItem( "Lock" ) then
                                for i, elem in ipairs( self.selected_elements ) do
                                    self:SetElemLock( elem, true )
                                end
                            end
                            if ui:MenuItem( "Unlock" ) then
                                for i, elem in ipairs( self.selected_elements ) do
                                    self:SetElemLock( elem, false )
                                end
                            end
                            if ui:MenuItem( "Hide" ) then
                                for i, elem in ipairs( self.selected_elements ) do
                                    self:SetViz( elem, false )
                                end
                            end
                            if ui:MenuItem( "Show" ) then
                                for i, elem in ipairs( self.selected_elements ) do
                                    self:SetViz( elem, true )
                                end
                            end
                            ui:Separator()
                            if ui:MenuItem( "Mirror" ) then
                                for i, elem in ipairs( self.selected_elements ) do
                                    self:SetElemFlip(elem, not elem.flip)
                                    self:SetElemPos(elem, -elem.x, elem.y)
                                end
                            end
                            ui:Separator()
                            if ui:MenuItem( "DBG" ) then
                                DBG( self.selected_elements )
                            end
                            ui:EndPopup()
                        end

                        local is_path = v.node_type == NODE_TYPE.PATH
                        local is_location = v.node_type == NODE_TYPE.LOCATION
                        if not is_path and not is_location then 
                            ui:SameLine(FlagsColumnWidth)
                            if ui:DragCheckbox(assets.day, 16, 16, 3, CheckBits( v.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.DAY )) then 
                                self:ToggleFlags(v, PLAX_FLAGS.DAY)
                            end

                            ui:SameLine(nil, 4)
                            if ui:DragCheckbox(assets.night, 16, 16, 5, CheckBits( v.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.NIGHT )) then 
                                self:ToggleFlags(v, PLAX_FLAGS.NIGHT)
                            end
                        
                            ui:SameLine(nil, 8)
                            if ui:DragCheckbox(assets.conversation, 16, 16, 6, CheckBits( v.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.CONVERSATION )) then 
                                self:ToggleFlags(v, PLAX_FLAGS.CONVERSATION)
                            end

                            ui:SameLine(nil, 4)
                            if ui:DragCheckbox(assets.battle, 16, 16, 7, CheckBits( v.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.BATTLE )) then 
                                self:ToggleFlags(v, PLAX_FLAGS.BATTLE)
                            end
                        end

                        ui:PopID()
                    end
                end
                ui:Unindent( 10 )
            end
            
            ui:Spacing()

        end
        
        ui:PopID(k)
    end
end

function PlaxEditor:RenderEditorPanel(ui)
    local title = "PLAX EDITOR: " .. self.save_id
    if self.plaxdata.mod_id then
        title = title .. string.format( "(%s)", tostring(self.plaxdata.mod_id))
    end
    if self.dirty then
        title = title .. " *"
    end

    local flags = ui.MenuBar
    local w, h = ui:DisplaySize()
    local panel_w = 350
    ui:SetNextWindowSize( panel_w, h, ui.ImGuiSetCond_Once )
    ui:SetNextWindowPos( 0, 0, ui.ImGuiSetCond_Once)

    local show = ui:Begin( title.."###MAINWINDOWTITLE", flags )
    if show then
        
        if ui:BeginMenuBar() then
            if ui:BeginMenu( "File" ) then
                
                if ui:MenuItem( "New" ) then
                   self:DoNew()
                end

                if ui:MenuItem( "Open" ) then
                   self:DoOpen()
                end

                if ui:MenuItem( "Save" ) then
                   self:DoSave()
                end
                
                ui:Separator()

                if ui:MenuItem( "Delete" ) then
                   self:DoDelete()
                end

                --[[if ui:MenuItem( "Clean" ) then
                   self:DoClean()
                end--]]

                ui:Separator()

                if ui:MenuItem( "Exit" ) then
                   self:DoExit()
                end
                ui:EndMenu()
            end

            if ui:BeginMenu( "Edit" ) then
                if ui:MenuItem( "Copy" ) then
                   self:DoCopy()
                end
                if ui:MenuItem( "Paste" ) then
                   self:DoPaste()
                end
                ui:Separator()
                if ui:MenuItem( "Redo") then
                    self:Redo()
                end
                if ui:MenuItem( "Undo" ) then
                    self:Undo()
                end
                ui:Separator();
                
                if ui:MenuItem( "Settings..." ) then
                    self.open_select_popup = "Plax Editor Settings" -- See below for implementation
                end

                ui:EndMenu()
            end
            if ui:BeginMenu( "Layers" ) then
                if ui:MenuItem( "Add New" ) then
                   self:AddNewLayer()
                end
                ui:EndMenu()
            end
            if ui:BeginMenu( "Elements" ) then
                if not self.plaxdata.is_region then
                    if ui:MenuItem( "Add Character" ) then
                        self:AddCharacter()
                    end
                    if ui:MenuItem( "Add Prop" ) then
                        self:AddProp()
                    end
                else
                    if ui:MenuItem( "Add Path" ) then
                        self:AddPath()
                    end
                    if ui:MenuItem( "Add Location Node" ) then
                        self:AddLocationNode()
                    end
                end

                if ui:MenuItem( "Add Animation" ) then
                   self:AddAnimation()
                end
                if ui:MenuItem( "Add Particles" ) then
                   self:AddParticles()
                end
                ui:EndMenu()
            end
            if ui:BeginMenu( "Tools" ) then

                if ui:MenuItem( "Demo" ) then
                   self:DoDemo()
                end
                if ui:MenuItem( "Demo (no characters)" ) then
                   self:DoDemo(true)
                end

                if ui:MenuItem( "Fight Demo" ) then
                    self:DoFightDemo()
                end

                if ui:BeginMenu( "Fight Demo Parameters" ) then
                    local blue_fighters = ui:SliderInt( "Team Blue", (self.blue_fighters or 1), 1, MAX_BATTLE_SIZE )
                    if blue_fighters ~= self.blue_fighters then
                        self.blue_fighters = blue_fighters
                    end
                    local red_fighters = ui:SliderInt( "Team Red", (self.red_fighters or 1), 1, MAX_BATTLE_SIZE )
                    if red_fighters ~= self.red_fighters then
                        self.red_fighters = red_fighters
                    end
                    ui:EndMenu()
                end

                ui:Separator()

                if ui:MenuItem( "Convo Frames", nil, CheckBits(self.aspect_framing or 0, ASPECT_FRAMING.CONVO ) ) then
                    self:ShowAspectFraming( ToggleBits(self.aspect_framing or 0, ASPECT_FRAMING.CONVO) )
                end

                if ui:MenuItem( "Battle Frames", nil, CheckBits(self.aspect_framing or 0, ASPECT_FRAMING.BATTLE ) ) then
                    self:ShowAspectFraming( ToggleBits(self.aspect_framing or 0, ASPECT_FRAMING.BATTLE) )
                end

                ui:EndMenu()
            end

            ShowDebugToolsMenu(self:GetFE(), ui, function(fn) self:SwitchScreens(fn) end )

            ui:EndMenuBar()
        end
    end

    if self:IsOnStack() and self.plaxdata then
        self:RenderMainPlaxPanel( ui )
    end
    ui:End()
end

local blendmodes = {BLEND_MODE.BLENDED, BLEND_MODE.EMISSIVE, BLEND_MODE.MULTIPLY, BLEND_MODE.ADDITIVE, BLEND_MODE.PREMULTIPLIED}
local blendmodes_txt = {"BLENDED", "EMISSIVE", "MULTIPLY", "ADD (No!)", "PREMULT (No!)"}


local clipBoardHSV = nil


-- Returns the key names for day/night colour and ignore parent
function PlaxEditor:GetDayPhaseName()
    if self.day_phase == DAY_PHASE.DAY then
        return "day_colour", "day_colour_ignore_parent"
    elseif self.day_phase == DAY_PHASE.NIGHT then
        return "night_colour", "night_colour_ignore_parent"
    end
end


function PlaxEditor:EditTags(ui, owner, key_name)

    local in_string = owner[key_name] or ""
    local all_tags = in_string:split(",")
    local out_tags = shallowcopy(all_tags)

    ui:Columns(2)
    for k,tag in ipairs(all_tags) do
        ui:Text(tag)
        ui:NextColumn()
        if ui:Button("REMOVE##TAGREMOVE"..k) then
            table.arrayremove(out_tags, tag)
        end
        ui:NextColumn()
    end
    ui:Columns(1)

    local new_string = ui:InputText("New Tag", self.new_tag_text or "")
    if new_string then
        self.new_tag_text = new_string
    end

    local is_invalid = self.new_tag_text == "" or string.find(self.new_tag_text or "", ",") ~= nil

    ui:SameLine(nil, 10)
    if ui:Button("Add",nil,nil, is_invalid) then
        
        table.insert_unique(out_tags, self.new_tag_text)
        table.sort(out_tags)
        self.new_tag_text = ""
    end

    local new_tag_packed = #out_tags > 0 and table.concat( out_tags, ",") or ""
    if new_tag_packed ~= (owner and owner[key_name] or "") then
        self:PushState()
        owner[key_name] = new_tag_packed
        return true
    end
end

function PlaxEditor:RenderAnimationElementPanel(ui, elem)
    if ui:Button("Anim File: " .. (elem.animfile or "")) then
        self:OnClickAnimFile() 
    end

    if ui:Button("Build File: " .. (elem.buildfile or "")) then
        self:OnClickBuildFile() 
    end
    ui:SameLine(nil, 10)
    if ui:Button("Clear##1") then
        if elem.buildfile then
            self:PushState()
            elem.buildfile = nil
            self:Refresh()
        end
    end
    local anims = { "<NONE>" }
    if elem.animfile then
        local animdat = GetAnim(elem.animfile)
        if animdat then
            for k = 1, animdat:GetNumAnims() do
                local anim_name = animdat:GetAnimName(k-1)
                table.insert(anims, anim_name )
            end
        end
    end

    local anim_idx = table.arrayfind(anims, elem.anim)
    local new_anim_idx = ui:Combo("Anim State", anim_idx or 1, anims)
    if anim_idx ~= new_anim_idx then
        if new_anim_idx == 1 then
            self:SetAnimState(elem, nil )
        else
            self:SetAnimState(elem, anims[new_anim_idx])
        end
    end

    local w = self.plax:GetWidgetByElement( elem )
    if w then
        ui:Indent( 20 )
        ui:Text( string.format( "Current anim: %s (%.2f)", w:GetCurrentAnimName(), w:GetNormalizedTime() ))
        ui:Text( string.format( "Duration: %.2f", w.model and w.model:GetDuration() or -1 ))
        ui:Unindent( 20 )
    end

    local timeoffset = ui:SliderFloat("Initial Delay", elem.anim_initial_delay or 0, 0, 60)
    if timeoffset ~= (elem.anim_initial_delay or 0) then
        self:SetAnimTimeOffset(elem, timeoffset)
    end

    local phase = ui:SliderFloat("Phase", elem.anim_phase or 0, 0, 100)
    if phase ~= elem.anim_phase then
        self:SetAnimPhase(elem, phase)
    end

    local delaymin = ui:SliderFloat("Repeat (Min)", elem.anim_repeat_delay_min or 0, 0, 60)
    if delaymin ~= elem.anim_repeat_delay_min then
        self:SetAnimRepeatDelayMin(elem, delaymin)
    end

    local delaymax = ui:SliderFloat("Repeat (Max)", elem.anim_repeat_delay_max or 0, 0, 60)
    if delaymax ~= elem.anim_repeat_delay_max then
        self:SetAnimRepeatDelayMax(elem, delaymax)
    end

    local animspeed = ui:SliderFloat("Speed", elem.anim_speed or 100, 0, 300)
    if animspeed ~= elem.anim_speed then
        self:SetAnimSpeed(elem, animspeed)
    end
end

function PlaxEditor:RenderElementWidget( ui, elem, w )
    ui:Text( string.format( "%s (%s)", w._classname, w:IsShown() and "Visible" or "Hidden" ))

    if is_instance( w, Widget.Anim ) then
        ui:Text( "Current anim:" )
        ui:SameLine( nil, 10 )
        ui:TextColored( 0xCC00FFFF, string.format( "%s (%.2f)", w:GetCurrentAnimName(), w:GetNormalizedTime() ))
    end
end

function PlaxEditor:RenderElementPanel(ui)
    local title = "Element"
    local w, h = ui:DisplaySize()
    local panel_w = 350
    ui:SetNextWindowSize( panel_w, h, ui.ImGuiSetCond_Once )
    ui:SetNextWindowPos( w - panel_w, 0, ui.ImGuiSetCond_Once)

    local elem = self.selected_elements[1]
    local is_path = elem.node_type == NODE_TYPE.PATH
    local is_image = elem.node_type == NODE_TYPE.IMAGE
    local is_location = elem.node_type == NODE_TYPE.LOCATION

    local show = ui:Begin( title, ui.MenuBar )
    if show then
        if ui:BeginMenuBar() then
            if ui:BeginMenu( "Element" ) then
                if ui:MenuItem( "Debug Element" ) then
                    DBG( elem )
                end
                if ui:MenuItem( "Debug Widget" ) then
                    DBG( self.plax:GetWidgetByElement(elem) )
                end
                ui:EndMenu()
            end
            ui:EndMenuBar()
        end

        ui:PushItemWidth(panel_w*.67)
        if is_path then
            ui:PushItemWidth(100)
            local x = ui:DragFloat("X", elem.x or 0, -4000, 4000, "%0.2f", .01)
            ui:SameLine(nil,10)
            local y = ui:DragFloat("Y", elem.y or 0, -4000, 4000, "%0.2f", .01)
            if x or y then
                self:SetElemPos(elem, x, y)
            end
            ui:PopItemWidth()

            local path_widget = self.plax:GetWidgetByElement(elem)
            local edge = path_widget:GetSelectedEdge()
            local node = path_widget:GetSelectedNode()
            if edge or node then
                local pathpiece = edge or node

                pathpiece.userdata = pathpiece.userdata or {}
                if self:EditTags(ui, pathpiece.userdata, "tag_list") then
                    path_widget:Refresh()
                end
            end

        else
            local new_id = ui:InputText("Element ID", elem.element_id)
            if new_id then
                self:PushState()
                elem.element_id = new_id
            end

            ui:PushItemWidth(100)
            local x = ui:DragFloat("X", elem.x or 0, -4000, 4000, "%0.2f", .01)
            ui:SameLine(nil,10)
            local y = ui:DragFloat("Y", elem.y or 0, -4000, 4000, "%0.2f", .01)
            if x or y then
                self:SetElemPos(elem, x, y)
            end
        
                
            local s = ui:DragFloat("Scale", elem.scale or 1, .01, 20, "%0.2f", .01)
            if s then
                self:SetElemScale(elem, s)
            end
            local r = ui:DragFloat("Rotation", elem.rot or 0, -180, 180, "%0.2f", .01)
            if r then
                self:SetElemRotation(elem, r)
            end
            if ui:Checkbox("Flip", elem.flip) then
                self:SetElemFlip(elem, not elem.flip)
            end
        


            ui:PopItemWidth()
    
            ui:Spacing()

            if is_image then
                if ui:TreeNode("UV Animation") then
                    if ui:Button("Reset UV Anim") then
                        self:ResetUVAnim(elem)
                        elem.uspeed = 0
                        elem.vspeed = 0
                    end

                    local u = ui:DragFloat("U Speed", elem.uspeed or 0, -1.0, 1.0, "%0.4f", .0001)
                    local v = ui:DragFloat("V Speed", elem.vspeed or 0, -1.0, 1.0, "%0.4f", .0001)

                    if u or v then
                        self:SetUVSpeed(elem, u, v)
                    end

                    ui:TreePop()
                end
            end

            if ui:TreeNode("Colour / Blend mode") then

                --there are too many ways to represent a colour. sigh.
                local col_in = ((elem.tint_r or 255) << 24) | ((elem.tint_g or 255) << 16) | ((elem.tint_b or 255) << 8) | ((elem.tint_a or 255))
                local col = ui:ColorEdit4("Colour", col_in)
                if col and col ~= col_in then
                    local r, g, b, a = HexColour(col)
                    self:SetElemColour(elem, math.floor(r*255), math.floor(g*255), math.floor(b*255), math.floor(a*255))
                end

                local bloom = ui:SliderFloat("Bloom",elem.bloom or 0, 0, 1)
                if bloom and bloom ~= elem.bloom then
                    self:SetElemBloom(elem, bloom)
                end

                local blend_idx = table.arrayfind(blendmodes, elem.blend_mode or BLEND_MODE.BLENDED)
                local new_blend_idx = ui:Combo("Blend Mode", blend_idx, blendmodes_txt)
                if blend_idx ~= new_blend_idx then
                    self:SetElemBlendMode(elem, blendmodes[new_blend_idx])
                end

                if elem.node_type == NODE_TYPE.ANIMATION then
                    if ui:Checkbox("Erode", elem.erode) then
                        self:SetElemErode(elem, not elem.erode)
                    end
                end


                if ui:TreeNode("HSV") then
                    local lighting_name, lighting_ignore_parent = self:GetDayPhaseName()

                    local hsvButtonWidth = 70

                    local changedHSV = false

                    if elem[lighting_name] then
                        if ui:Button("Reset", hsvButtonWidth) then
                            self:PushState()
                            elem[lighting_name] = nil
                            changedHSV = true
                        end
                    else
                        if ui:Button("Add", hsvButtonWidth) then
                            self:PushState()
                            elem[lighting_name] = { highlights = { colour = { 1, 1, 1 }, hsv = { 0, 1, 1 } },
                                                    midtones = { colour = { 1, 1, 1 }, hsv = { 0, 1, 1 } },
                                                    shadows = { colour = { 1, 1, 1 }, hsv = { 0, 1, 1 } } }
                            changedHSV = true
                        end
                    end    

                    if elem[lighting_name] then
                        ui:SameLine(nil, 5)
                        if ui:Button("Copy", hsvButtonWidth) then 
                            clipBoardHSV = deepcopy(elem[lighting_name])
                        end
                    end

                    if clipBoardHSV then
                        ui:SameLine(nil, 5)
                        if ui:Button("Paste", hsvButtonWidth) then 
                            self:PushState()
                            elem[lighting_name] = deepcopy(clipBoardHSV)
                            changedHSV = true
                        end
                    end
           
                    if elem[lighting_name] then
                        local ignoreParent = elem[lighting_ignore_parent] or false
                        if ui:Checkbox("Ignore Parent Lighting", ignoreParent) then
                            elem[lighting_ignore_parent] = not ignoreParent
                            changedHSV = true
                        end
                        
                        ui:Spacing()
                        
                        if self:HSVAdjust(ui, "Shadows", elem[lighting_name].shadows) then
                            changedHSV = true
                        end
                        
                        ui:Spacing()

                        if self:HSVAdjust(ui, "Midtones", elem[lighting_name].midtones) then
                            changedHSV = true
                        end
                        
                        ui:Spacing()
                        
                        if self:HSVAdjust(ui, "Highlights", elem[lighting_name].highlights) then
                            changedHSV = true
                        end
                        
                        ui:Spacing()    
                    end

                    if changedHSV then
                        self.plax:UpdateCurrentHSVForElement(elem)
                    end
                    ui:Separator()
                    ui:TreePop()
                end

                ui:TreePop()
            end


            ui:Spacing()

            if ui:TreeNode("Editor Flags") then
                if ui:Checkbox("guide", elem.guide) then
                    self:SetGuide(elem, not elem.guide)
                end
                if ui:Checkbox("Expensive", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.EXPENSIVE )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.EXPENSIVE)
                end
                ui:TreePop()
            end

            ui:Spacing()
   
            if ui:TreeNode("Visibility Flags") then
                local default_visibility = elem.default_visibility
                if default_visibility == nil then
                    default_visibility = true
                end
                if ui:Checkbox( "Default Visibility", default_visibility ) then
                    if default_visibility then
                        elem.default_visibility = false
                    else
                        elem.default_visibility = nil -- nil means true, so don't bother saving it.
                    end
                    self:PushState()
               end

                if ui:Checkbox("Day", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.DAY )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.DAY)
                end

                ui:SameLine(200)
                if ui:Checkbox("Night", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.NIGHT )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.NIGHT)
                end

                ui:Separator()

                if ui:Checkbox("Conversation", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.CONVERSATION )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.CONVERSATION)
                end

                ui:SameLine(200)
                if ui:Checkbox("Battle", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.BATTLE )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.BATTLE)
                end

                ui:TreePop()
            end

            ui:Spacing()
        
        end

        if elem.node_type == NODE_TYPE.ANIMATION then
            if ui:TreeNode("Animation Settings") then
                self:RenderAnimationElementPanel( ui, elem )
                ui:TreePop()
            end
        elseif elem.node_type == NODE_TYPE.SLOT then
            if ui:TreeNode("Slot Settings") then          
                local valid_roles = self:GetImportantRolesForCurrentLocation()
                local role_idx = ((elem.character_role == nil or elem.character_role == "") and 1 or table.arrayfind(valid_roles, elem.character_role)) or 1
                local new_role_idx = ui:Combo("Used Roles", role_idx, valid_roles)
                if new_role_idx ~= role_idx then
                    self:PushState()
                    if new_role_idx == 1 then
                        elem.character_role = nil
                    else
                        elem.character_role = valid_roles[new_role_idx]
                    end
                end

                local priority = ui:SliderInt( "Priority", elem.priority or 1, 1, 6 )
                if priority and priority ~= elem.priority then
                    self:PushState()
                    elem.priority = priority
                end

                
                if ui:Checkbox("Allow Snails", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.ALLOW_SNAILS )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.ALLOW_SNAILS)
                end

                ui:SameLine(140)
                if ui:Checkbox("Allow Monsters", CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.ALLOW_MONSTERS )) then
                    self:ToggleFlags(elem, PLAX_FLAGS.ALLOW_MONSTERS)
                end

                -- Preview texture
                if ui:TreeNode( "Preview Texture" ) then
                    local ok = engine.inst:TextureExists( elem.preview_texture or "" )
                    if not ok then
                        ui:PushStyleColor( ui.Style_Text, 0xFF0000FF )
                    end

                    local preview_texture = ui:InputText( "preview", elem.preview_texture or "", ui.TextFlags_EnterReturnsTrue )
                    if not ok then
                        ui:PopStyleColor()
                    end
                    if ui:Button( "Make Default" ) then
                        preview_texture = DEFAULT_PROP_TEX
                    end

                    if preview_texture and preview_texture ~= elem.preview_texture then
                        self:PushState()
                        elem.preview_texture = preview_texture
                        elem.preview_agent = nil
                        self:RefreshSlot( elem )
                    end
                    ui:TreePop()
                end

                if ui:TreeNode( "Preview Agent" ) then
                    local agent_idx = table.arrayfind(self.preview_agent_options, elem.preview_agent )
                    local new_agent_idx = ui:Combo("agent", agent_idx or 0, self.preview_agent_options)
                    if agent_idx ~= new_agent_idx then
                        self:SetPreviewCharacter(elem, self.preview_agent_options[new_agent_idx])
                    end

                    local pose_idx = table.arrayfind(self.posedef_options, elem.posedef or "standing")
                    local new_pose_idx = ui:Combo("Pose", pose_idx or 1, self.posedef_options)
                    if pose_idx ~= new_pose_idx then
                        self:SetElemPoseDef(elem, self.posedef_options[new_pose_idx])
                    end
                
                    if ui:Button("Propfile:" ..(elem.prop_file or "")) then
                        self:OnClickPropFile(elem, 1)
                    end
                    ui:SameLine(nil, 10)
                    if ui:Button("Clear##1") then
                    
                        self:PushState()
                        elem.prop_file = nil
                        self:RefreshSlot(elem)
                    end

                    if ui:Button("Propfile2:" ..(elem.prop_file2 or "")) then
                        self:OnClickPropFile(elem, 2)
                    end
                    ui:SameLine(nil, 10)
                    if ui:Button("Clear##2") then
                        self:PushState()
                        elem.prop_file2 = nil
                        self:RefreshSlot(elem)
                    end
                    ui:TreePop()
                end

                ui:TreePop()
            end
        elseif elem.node_type == NODE_TYPE.PARTICLES then
            if ui:TreeNode("Particles Settings") then

                local param_list = {}
                for k,v in pairs(Content.GetAllParticleParams()) do
                    table.insert(param_list, k)            
                end
                table.sort(param_list)               
                
                local param_id = elem.particles or ""

                -- Combo box with all known particle types:
                local idx = table.arrayfind(param_list, param_id)

                local new_idx = ui:Combo("Particles", idx or 1, param_list)
                if new_idx and new_idx ~= idx then
                    elem.particles = param_list[new_idx]
                    self:PushState()
                    self:Refresh()
                    self:SelectElement( elem, true )    -- Force rebuilding of the widget bounding box
                end

                if elem.particles and idx then
                    if ui:Button("Edit " .. elem.particles) then
                        self:SwitchScreens(function() DebugTools_ParticleEditor(self:GetFE(), elem.particles, self.save_id) end)
                    end
                end


                ui:TreePop()
            end
        elseif is_location then
            local new_name = ui:InputText("Name", elem.location_name)
            if new_name then
                if new_name == "" then
                    elem.location_name = nil
                else
                    elem.location_name = new_name
                end
                self.plax:GetWidgetByElement(elem):Refresh()
            end

            --elem.location_tags = elem.location_tags or {}
            if self:EditTags(ui, elem, "location_tags") then
                self.plax:GetWidgetByElement(elem):Refresh()
                --self.plax:GetWidgetByElement(elem):Refresh()
            end

            ui:PushItemWidth(100)
            local x = ui:DragFloat("Enter_X", elem.enter_x or 0, -800, 800, "%0.2f", .5)
            ui:SameLine(nil,10)
            local y = ui:DragFloat("Enter_Y", elem.enter_y or 0, -800, 800, "%0.2f", .5)
            if x or y then
                self:SetElemEnterPos(elem, x, y)
            end
            ui:PopItemWidth()                

        end

        ui:Separator()

        ui:PopItemWidth()
    end

    ui:End()
end


function PlaxEditor:ApplyColourCorrectionValues(v)
    self.plax:ApplyColourCorrectionValues(v)
end

function PlaxEditor:GetCurrentColourName()
    return self.plax:GetCurrentColourName()
end

function PlaxEditor:GetMusicEvent()
    return self.plaxdata and self.plaxdata.music_event
end