require "ui/widgets/screen"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/movieplayer"
require "ui/widgets/panelbutton"
require "ui/widgets/solidbox"

---------------------------------------------------------------------------------
-- Plays a slideshow of movies and still frames.

local FADE_HOLD = .5 -- Time to hold at black before fading into the next slide.
local FADE_DURATION = .5 -- Time to fade to or from black (therefore, full duration is 2*FADE_DURATION + FADE_HOLD)
local MAX_SLIDES = 16 -- How many MoviePlayers to preallocate.
local BLEND_THRESHOLD = 0.5

local assets = 
{
    pip = engine.asset.Texture('UI/movie_pip.tex'),
}

local SlideshowScreen = class("Screen.SlideshowScreen", Widget.Screen)


SlideshowScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            -- This test prevents inadvertant skipping if you happen to click right after slides auto-advance.
            return (self.slide_elapsed_time or math.huge) > 1.0
        end,
        fn = function(self) 
            self:Advance()
        end,
        hint = function(self, left, right) 
            table.insert( right, loc.format( self.step < #self.script and LOC"UI.CONTROLS.NEXT_SLIDE" or LOC"UI.CONTROLS.DONE_SLIDES", Controls.Digital.MENU_ACCEPT))
        end,
    },
    {
        control = Controls.Digital.REWIND,
        test = function(self)
            return #self.movie_widgets > 1
        end,
        fn = function(self) 
            self:Rewind()
        end,
        
        hint = function(self, left, right) 
            table.insert( left, loc.format( LOC"UI.CONTROLS.REWIND", Controls.Digital.REWIND))
        end,
    },

    {
        control = Controls.Digital.SKIP,
        
        fn = function(self) 
            if self.showing_quit then
                self:Finish()
            else
                self.showing_quit = true
                if self.vo_sound then
                    self.vo_sound:SetPaused( true )
                end
                self.next:Show():SetTintAlpha(1.0)
            end
            return true
        end,
        
        hint = function(self, left, right) 
            if self.step < #self.script then
                table.insert( right, loc.format( LOC"UI.CONTROLS.SKIP_SLIDES", Controls.Digital.SKIP))
            end
        end,
    }
}

function SlideshowScreen:init( content_id, ondone, flags)
    Widget.Screen.init(self)

    AUDIO:FinalizeVoiceLoad()
    
    if content_id == nil then
        content_id = "sal_ending_slideshow"
        flags = { nadan_is_alive = true }
    end
    local data = Content.GetSlideShow( content_id )
    self.content = data

    flags = flags or table.empty
    self.script = {}
    for k,v in ipairs(data.script) do
        if v.flag == nil or
            (type(v.flag) == "string" and flags[v.flag]) or
            (type(v.flag) == "function" and v.flag( flags )) then
            --
            local slide = shallowcopy(v)
            slide.txt = self.content:GetLocalizedTxt( k )
            table.insert(self.script, slide)
        end
    end

    self.music_name = data.music
    self.tutorial_mode = data.is_tutorial
    self.ondone = ondone
    self.auto_advance = true

    if self.tutorial_mode then
        self.screen_mix_event = "event:/mix_events/TutorialMusicDown"
    end

    self:SetAnchors( "center", "center" )

    -- Get a standard slideshow image to calculate the proper size for the movie
    local defaultimage = "images/white.tex"
    
    self.lastimg = defaultimage

    local MOVIE_ASPECT = 1920/672
    self.movie_width = 2560
    self.movie_height = 2560 / MOVIE_ASPECT

    local panel_padding = 60
    self.text_width = RES_X*.6

    self.next = Widget.Label("body", 48):SetAnchors("right", "bottom"):SetPos(-200, 60):Hide()
    -- Click-to-skip widget
    self.next = Widget()
    self.next.bg = self.next:AddChild( Widget.Image( engine.asset.Texture( "large/conversation_banner.tex") ) ):SetHiddenBoundingBox( true )
        :Bloom( 0.05 )
        :SetSize( 650, 250 )
    self.next.text = self.next:AddChild( Widget.Label() )
        :SetText( loc.format(LOC"UI.SLIDESHOW_SCREEN.SKIP", Controls.Digital.SKIP) )
        :SetFont("title")
        :SetFontSize( FONT_SIZE.SCREEN_SUBTITLE )
        :SetGlyphColour(UICOLOURS.SUBTITLE)
        :SetPos( 0, 1 )
        :Bloom( 0.05 )
    self.next:Hide()
    self.body_text = Widget.Label( "body", FONT_SIZE.BODY_TEXT )
        :SetAutoSize(self.text_width)
        :SetWordWrap(true)
        :SetGlyphColour(0xcbffefff)
        :LeftAlign()
        :TopAlign()
        :SetBloom(.05)

    self.progress_root = Widget()

    self.movie_root = Widget()
    self.panel_bg = self.movie_root:AddChild( Widget.SolidBox( self.movie_width, self.movie_height, 0x00000000 ) )
    self.framing_elements = Widget()
    self.panel_glow_top = self.framing_elements:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetTintAlpha( 0.4 )
    self.panel_glow_bottom = self.framing_elements:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetTintAlpha( 0.2 )
    self.panel_border_top = self.framing_elements:AddChild( Widget.SolidBox( RES_X, 6, 0xffffffff ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.panel_border_bottom = self.framing_elements:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )

    self.image_widget = self.movie_root:AddChild(Widget.Image( engine.asset.Texture(defaultimage) ))
    self.image_widget:SetSize( self.movie_width, self.movie_height )

    self.movie_widgets = {}
    for i=1, MAX_SLIDES do
        local movie_widget = self.movie_root:AddChild( Widget.MoviePlayer( nil, self.movie_width, self.movie_height ))
        movie_widget:SetBlendTexture(engine.asset.Texture("splash/inkBleed.tex"))
        movie_widget:SetBlendMode(BLEND_MODE.BLENDED)
        movie_widget:Hide()
        table.insert(self.movie_widgets, movie_widget)
    end

    self.text_bg = Widget.Image( engine.asset.Texture("UI/teammember_gradient.tex" ) )

    self:AddChildren{
        self.framing_elements,
        self.movie_root,
        self.text_bg,
        self.body_text,
        self.progress_root,
        self.next,
    }

    self.progress_pips = {}
    for k = 1, #self.script do
        local w = Widget.Clickable()
        w.hitbox = w:AddChild( Widget.SolidBox( 110, 80, 0xff00ff00 ) )
        w.img = w:AddChild( Widget.Image(assets.pip) )
        w:SetHighlightWidget(w.img)
        w:Bloom(.1)
        w:SetOnClickFn( function()
            self:ChangeSlide( k )
        end )
        table.insert(self.progress_pips, w)
        self.progress_root:AddChild(w)
        w:LayoutBounds("after", "center"):Offset(0, 0)
    end
    

    self:SetupUnderlay( true )
    self:DoLayout()

end

function SlideshowScreen:OnControlModeChange( cm, device_id )
    SlideshowScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.TOUCH then
        self.next.text:SetText( LOC"UI.MOVIE_SCREEN.SKIP_TOUCH" )
    else
        self.next.text:SetText( loc.format(LOC"UI.MOVIE_SCREEN.SKIP", Controls.Digital.SKIP) )
    end
end

function SlideshowScreen:OnScreenModeChange( sm )
    SlideshowScreen._base.OnScreenModeChange( self, sm )

    self.progress_root:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.text_bg:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.body_text:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.next:SetLayoutScale( LAYOUT_SCALE[sm] )
    self:DoLayout()
end

function SlideshowScreen:SetAutoAdvance( auto )
    self.auto_advance = auto
    return self
end

function SlideshowScreen:OnUpdate(dt)
    
    SlideshowScreen._base.OnUpdate(self, dt)

    if self.slide_elapsed_time then
        self.slide_elapsed_time = self.slide_elapsed_time + dt
    end

    if self.ignore_input_time then
        self.ignore_input_time = self.ignore_input_time - dt
        if self.ignore_input_time <= 0 then
            self.ignore_input_time = nil
        end
    end


    if self.vo_sound then
        if self.vo_sound:GetPlaybackState() == engine.sound.STOPPED then
            if self.auto_advance then self.auto_advance_timer = 1 end
            self.vo_sound:Stop()
            self.vo_sound:Release()
            self.vo_sound = nil

        end
    end

    if self.auto_advance_timer then
        self.auto_advance_timer = self.auto_advance_timer - dt
        if self.auto_advance_timer <= 0 then -- and self.step < #self.script then
            self.auto_advance_timer = nil
            self.slide_elapsed_time = 0
            self:Advance()
        end
    end

    if TheGame:GetLocalSettings().ROBOTICS then
        self:Advance()
    end

end

function SlideshowScreen:DoLayout()
    
    local sw, sh = engine.inst:GetScreenSize()
    
    local base_scale = TheGame:FE():GetBaseWidgetScale()
    local screen_ratio = sw/sh
    local normal_ratio = 16/9
    local wide_ratio = 21/9
    local tall_ratio = 4/3

    local scale
    
    if screen_ratio > normal_ratio then
        scale = math.min( sw/sh, wide_ratio) / wide_ratio
    elseif screen_ratio < normal_ratio then
        scale = tall_ratio/normal_ratio
    else
        scale = normal_ratio/wide_ratio
    end

    self.panel_bg:SetScale( scale, scale )
    self.image_widget:LayoutBounds("center", "center", self.panel_bg):SetScale(scale, scale)
    for k, v in ipairs(self.movie_widgets) do
        v:LayoutBounds("center", "center", self.panel_bg):SetScale(scale, scale)
    end

    -- Size up the framing elements
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.panel_glow_top:SetSize( self.screen_w*1.3, 300 )
    self.panel_glow_bottom:SetSize( self.screen_w*0.9, 300 )
    self.panel_border_top:SetSize( self.screen_w, 3 )
    self.panel_border_bottom:SetSize( self.screen_w, 2 )
    
    self.text_bg:Hide()

    --special layouts for different aspect ratios
    if (scale*self.movie_height*base_scale) > (sh - 300) then
        self.movie_root:LayoutBounds("center", "center", 0, 0)
        self.body_text:SetFontSize( FONT_SIZE.BODY_TEXT )
        self.body_text:LayoutBounds("center", "bottom", self.image_widget):Offset(0,46)
        local w,h = self.body_text:GetSize()
        self.text_bg:SetSize(self.screen_w , h+200):Show():LayoutBounds("center", "bottom", self.image_widget )

    else
        local PADDING = 50
        local text_height = (sh - scale*self.movie_height*base_scale) - 100 - PADDING -- Accomodate for bottom bar height
        self:FitBodyText( text_height )

        if screen_ratio < normal_ratio then
            self.movie_root:LayoutBounds("center", "top", 0, (sh/2-150)*(1/base_scale))
        else
            self.movie_root:LayoutBounds("center", "top", 0, (sh/2-50)*(1/base_scale))
        end

        self.body_text:LayoutBounds("center", "below", self.movie_root):Offset(0,-PADDING)
    end

    self.panel_border_top:LayoutBounds( nil, "above", self.panel_bg )
    self.panel_glow_top:LayoutBounds( nil, "top", self.panel_bg ):Offset( 0, 100 )
    self.panel_glow_bottom:LayoutBounds( nil, "bottom", self.panel_bg ):Offset( 0, -150 )
    self.panel_border_bottom:LayoutBounds( nil, "below", self.panel_bg )
    self.progress_root:LayoutBounds("center", "center", self.panel_border_bottom):Offset( 0, -7.3 )
    self.next:LayoutBounds("center", "above", self.progress_root ):Offset( 0, 60 )
end

function SlideshowScreen:OnScreenResize(w, h)
    SlideshowScreen._base.OnScreenResize(self, w, h)
    self:DoLayout()
end

function SlideshowScreen:OnOpen()
    SlideshowScreen._base.OnOpen(self)
    
    TheGame:SetVignetteStrength( 0 )

    self:ChangeSlide( 1 )
    self.slide_elapsed_time = 0
    self.ignore_input_time = .75
end

function SlideshowScreen:GetMusicEvent()
    return self.music_name
end

function SlideshowScreen:Finish(time_override)
    self.closing = true
    self.slide_elapsed_time = nil

    if self.vo_sound then
        self.vo_sound:SetPaused( true )
    end

    TheGame:FE():FadeTransition(function() 
        self:GetFE():PopScreen()
        if self.ondone then
            AUDIO:PlayEvent(SoundEvents.cutscene_end)
            self.ondone()
        end
    end, time_override or FADE_DURATION )
end

function SlideshowScreen:OnClose()
    SlideshowScreen._base.OnClose(self)

    if self.vo_sound then
        self.vo_sound:Stop()
        self.vo_sound:Release()
        self.vo_sound = nil
    end

    if self.amb_sound then
        self.amb_sound:Stop()
        self.amb_sound:Release()
        self.amb_sound = nil
    end
    
    TheGame:SetVignetteStrength( UI_VIGNETTE )
end

function SlideshowScreen:RefreshPips()
    for k = 1, #self.script do
        if self.step == k then
            self.progress_pips[k]:SetTintAlpha(1)
        else
            self.progress_pips[k]:SetTintAlpha(0.3)
        end
    end
end

function SlideshowScreen:ChangeSlide( idx )
    if idx == self.step then
        return
    end
    self.auto_advance_timer = nil

    self.step = idx

    local txt = self.script[self.step].txt 
    local img = self.script[self.step].img
    local mov = self.script[self.step].mov
    local vo = self.script[self.step].vo
    self:RefreshPips()

    if self.vo_sound then
        self.vo_sound:Stop()
        self.vo_sound:Release()
        self.vo_sound = nil
    end

    if vo then
        self.vo_sound = AUDIO:CreateEventInstance(vo, "sound/vo.bank")
        
        if self.vo_sound:IsValid() then
            self.vo_sound:Start()
        else
            self.vo_sound = nil
        end
    end


    if self.amb_sound then
        self.amb_sound:Stop()
        self.amb_sound:Release()
        self.amb_sound = nil
    end
    
    if self.script[self.step].amb then
        
        self.amb_sound = AUDIO:CreateEventInstance(self.script[self.step].amb)
        
        if self.amb_sound:IsValid() then
            self.amb_sound:Start()
        else
            self.amb_sound = nil
        end
    end



    if self.lastimg ~= img then
        if img then
            self.image_widget:Show():SendToFront()
            self.image_widget:SetTexture( engine.asset.Texture( img ) )
        else
            self.image_widget:Hide()
        end
    end

    -- Put the activating movie on top 
    -- Stop the updaters on the other movies
    -- Blend in the activating movie
    -- When it is done blending in, stop and hide all the other movies.
    if self.lastmov ~= mov then
        if mov then
            if self.movie_widgets[idx] then
                self.movie_widgets[idx]:SendToFront()
            end

            for k, v in ipairs(self.movie_widgets) do
                if v.blend_updater then
                    v.blend_updater:Stop()
                end

                if v:IsPlaying() then
                    v:SetBlendFactors(1.0, BLEND_THRESHOLD)
                end
            end

            self.movie_widgets[idx]:Show()
            self.movie_widgets[idx]:SetBlendFactors(0.0, BLEND_THRESHOLD)
            
            local loop = not self.script[self.step].no_loop
            self.movie_widgets[idx]:PlayMovie( mov, loop )
        end

    
        self.movie_widgets[idx].blend_updater = self.movie_widgets[idx]:RunUpdater(
                Updater.Series{
                    Updater.Ease( function(v) self.movie_widgets[idx]:SetBlendFactors(v, BLEND_THRESHOLD) end, 0, 1, 1.0, easing.linear),
                    Updater.Do( 
                        function() 
                            for k, v in ipairs(self.movie_widgets) do 
                                if k ~= idx then 
                                    v:StopMovie() 
                                    v:Hide() 
                                end 
                            end 
                        end )
                }
            )
    end

    self.lastimg = img
    self.lastmov = mov

    self.next:Hide():SetTintAlpha(0)

    self.showing_quit = false
    self.body_text:SetText( txt )
    self.body_text:Spool()
    
    AUDIO:PlayEvent(SoundEvents.page_turn)
    collectgarbage()
    self:DoLayout()

end


function SlideshowScreen:FitBodyText( text_height )
    local smallest_text_size, largest_text_size = 20, 32

    local size = math.round( (smallest_text_size + largest_text_size)/2)
    while smallest_text_size < largest_text_size do
        self.body_text:SetFontSize(size)

        local textw, texth = self.body_text:GetScaledSize()

        local new_size
        if texth > text_height then
            largest_text_size = size
            new_size = math.floor( (smallest_text_size + size)/2)

        else
            smallest_text_size = size
            new_size = math.floor( (largest_text_size + size)/2)
        end
        if new_size == size then
            break
        else
            size = new_size
        end
    end
end

function SlideshowScreen:Advance()
    if self.step == #self.script then
        self:Finish()

    else
        self.body_text:SetTintAlpha( 1.0 )
        self:ChangeSlide( self.step + 1 )
    end
    self.fe:RequestControlRefresh()
end

function SlideshowScreen:Rewind()
    if self.step == 1 then
        -- At beginning.
    else
        self.body_text:SetTintAlpha( 1.0 )
        self:ChangeSlide( self.step - 1 )
    end
    self.fe:RequestControlRefresh()
end
