local UnlockInspectorScreen = class("Screen.UnlockInspectorScreen", Widget.Screen)

local PANEL_W = RES_X * 0.5
local PANEL_H = 640
local CARD_SCALE = .6
local PADDING = 20


UnlockInspectorScreen.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self)
            self:DoClose()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end
    }
}


function UnlockInspectorScreen:init( unlock, animate, fn )
	Widget.Screen.init( self )

    assert(fn == nil or type(fn) == "function")
    unlock = unlock or Content.GetUnlockTrack("SAL")[1]
	self.is_popup = true
    self.fn = fn
    self.animate = animate or false
    self:SetupUnderlay( true )
	self:SetAnchors("center", "center")

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- A container for all the background elements
    self.panel = self:AddChild( Widget.WidePanel( PANEL_H ) )
        :SetTopTitlesMargin( SPACING.M1 )

    if self.animate then
        self.panel:SetTitles( LOC"UI.UNLOCK_INSPECTOR.TITLE_NEW", unlock:GetLocalizedName() )
        -- Continue button
        self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
        self.continue_button = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.UNLOCK_INSPECTOR.CONTINUE_BTN", function() self:DoClose() end ) )
            :SetIcon( global_images.accept )
            :LayoutBounds( "right", "above", -60, 60 )
            :Hide()
        self:RegisterHotkeyButton( Controls.Digital.MENU_CANCEL, self.continue_button )
    else
        self.panel:SetTitles( LOC"UI.UNLOCK_INSPECTOR.TITLE_UNLOCKED", unlock:GetLocalizedName() )
        -- Back button
        self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
        self.close_button = self.bottom_left:AddChild( Widget.IconButton( LOC"UI.OVERLAYS.CLOSE", function() self:DoClose() end ) )
            :SetIcon( global_images.close )
            :LayoutBounds( "left", "above", 60, 60 )
        
        self:RegisterHotkeyButton( Controls.Digital.MENU_CANCEL, self.close_button )
    end

    -- Rewards block
    self.rewards_block = self:AddChild( Widget() )

    -- Negotiation cards block
    self.negotiation_cards_widget = self.rewards_block:AddChild( Widget.UnlocksInspectorBlock(UICOLOURS.NEGOTIATION) ):Hide()

    -- Battle cards block
    self.battle_cards_widget = self.rewards_block:AddChild( Widget.UnlocksInspectorBlock(UICOLOURS.FIGHT) ):Hide()

    -- Outfit block
    self.outfit_widget = self.rewards_block:AddChild( Widget.UnlocksInspectorBlock(UICOLOURS.OBJECTIVE) ):Hide()

    --for now we will generally have 3 battle cards and 3 negotiation cards, and maybe an outfit. layout will get complex, I guess?
    for k, v in ipairs(unlock.items) do
        local w
        if v.battle_card then
            self.battle_cards_widget:AddContent( v )
                :SetTitles( unlock:GetLocalizedBattleName(), LOC"UI.UNLOCK_INSPECTOR.COMBAT_CARDS_TITLE" )
                :Show()
        elseif v.negotiation_card then
            self.negotiation_cards_widget:AddContent( v )
                :SetTitles( unlock:GetLocalizedNegotiationName(), LOC"UI.UNLOCK_INSPECTOR.NEGOTIATION_CARDS_TITLE" )
                :Show()
        elseif v.outfit_id then
            local outfit = Content.GetOutfit(v.outfit_id)
            self.outfit_widget:AddContent( v )
                :SetTitles( outfit:GetName(), LOC"UI.UNLOCK_INSPECTOR.OUTFIT_TITLE" )
                :Show()
        end
    end

    self:Layout()
    if self.animate then
        self:ShowAnimation()
    end
end

function UnlockInspectorScreen:OnControlModeChange( cm, device_id )
    UnlockInspectorScreen._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        if self.bottom_left then self.bottom_left:Show() end
        if self.bottom_right then self.bottom_right:Show() end
        self.panel:SetDesc( LOC"UI.DRAFT_SCREEN.UPGRADES_PREVIEW_INFO" )
    end
    if cm == CONTROL_MODE.TOUCH then
        if self.bottom_left then self.bottom_left:Show() end
        if self.bottom_right then self.bottom_right:Show() end
        self.panel:SetDesc( nil )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        if self.bottom_left then self.bottom_left:Hide() end
        if self.bottom_right then self.bottom_right:Hide() end
        self.panel:SetDesc( nil )
    end
end

function UnlockInspectorScreen:OnScreenModeChange( sm )
    UnlockInspectorScreen._base.OnScreenModeChange( self, sm )

    if self.bottom_left then self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] ) end
    if self.bottom_right then self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] ) end
    
    if self.negotiation_cards_widget then 
        self.negotiation_cards_widget.header:SetLayoutScale( LAYOUT_SCALE[sm] )
            :LayoutBounds( "left", "top", self.negotiation_cards_widget.bg )
            :Offset( PADDING*1, -PADDING*0.8 )
    end

    local panel_height = PANEL_H
    local panel_y_offset = 0
    if sm == SCREEN_MODE.MONITOR then
        panel_height = PANEL_H
        panel_y_offset = 0
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        panel_height = PANEL_H + 80
        panel_y_offset = 60
    end
    
    self.panel:SetHeight( panel_height )
        :SetPos( 0, panel_y_offset )
    self:Layout()
end

function UnlockInspectorScreen:GetDefaultFocus()
    return self.negotiation_cards_widget:GetDefaultFocus()
    or self.battle_cards_widget:GetDefaultFocus()
    or self.outfit_widget:GetDefaultFocus()
end

function UnlockInspectorScreen:RefreshStreamerMode()
    if self.bottom_left then self.bottom_left:SetPos( 0, 0 ) end
    if self.bottom_right then self.bottom_right:SetPos( 0, 0 ) end
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        if self.bottom_left then self.bottom_left:SetPos( 360, 0 ) end
    elseif GetStreamerMode() == 2 then
        -- Nudge the right side widgets
        if self.bottom_right then self.bottom_right:SetPos( -360, 0 ) end
    end
end

function UnlockInspectorScreen:Layout()
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Layout the grid
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Get the height of every container
    local max_height = self.battle_cards_widget:GetHeight()
    max_height = math.max( max_height, self.negotiation_cards_widget:GetHeight() )
    max_height = math.max( max_height, self.outfit_widget:GetHeight() )
    -- Set them all to the same height
    self.battle_cards_widget:RefreshSize( max_height )
    self.negotiation_cards_widget:RefreshSize( max_height )
    self.outfit_widget:RefreshSize( max_height )
    -- Position them properly
    self.battle_cards_widget:LayoutBounds( "after", "top", self.negotiation_cards_widget ):Offset( 20, 0 )
    self.outfit_widget:LayoutBounds( "after", "top", self.battle_cards_widget ):Offset( 20, 0 )
    self.rewards_block:LayoutBounds( "center", "bottom", self.panel ):Offset( 0, PADDING*2 )

    -- Layout the buttons
    if self.close_button then self.close_button:LayoutBounds( "left", "above", 60, 60 ) end
    if self.continue_button then self.continue_button:LayoutBounds( "right", "above", -60, 60 ) end
end

function UnlockInspectorScreen:DoClose()
    if self.fn then self.fn() end
    self:FadeOut()
end

function UnlockInspectorScreen:ShowAnimation( )
    self.battle_cards_widget.title:SetTintAlpha(0)
    self.battle_cards_widget.subtitle:SetTintAlpha(0)
    self.battle_cards_widget.bg:SetTintAlpha(0)
    self.negotiation_cards_widget.title:SetTintAlpha(0)
    self.negotiation_cards_widget.subtitle:SetTintAlpha(0)
    self.negotiation_cards_widget.bg:SetTintAlpha(0)
    self.outfit_widget.title:SetTintAlpha(0)
    self.outfit_widget.subtitle:SetTintAlpha(0)
    self.outfit_widget.bg:SetTintAlpha(0)

    local duration = 0.1

    -- Prepare battle card animation
    local card_anim_duration_total = 0.3
    local card_number = math.min( #self.battle_cards_widget.content.children, 14 )
    local card_anim_duration = 0.1
    local delay_per_card = card_anim_duration_total / card_number
    local battle_card_adder = Updater.Parallel()
    for i, widget in ipairs( self.battle_cards_widget.content.children ) do
        widget:SetTintAlpha( 0 )
        battle_card_adder:Add( 
            Updater.Series{
                Updater.Wait(delay_per_card*i),
                Updater.Parallel{
                    Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/approach_list_appears") end ),
                    Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, card_anim_duration, easing.inQuad),
                    Updater.Ease( function(v) widget:SetScale(v,v) end, CARD_SCALE * 1.2, CARD_SCALE, card_anim_duration, easing.inQuad),
                }
            }
        )
    end
    battle_card_adder:Add( 
        Updater.Series{
            Updater.Wait(delay_per_card*(#self.battle_cards_widget.content.children+1)),
            Updater.Parallel{
                Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/new_card_unlocked") end ),
                Updater.Ease( function(v) self.battle_cards_widget.title:SetTintAlpha(v) end, 0, 1, card_anim_duration*2, easing.inQuad),
                Updater.Ease( function(v) self.battle_cards_widget.subtitle:SetTintAlpha(v) end, 0, 1, card_anim_duration*1.5, easing.inQuad),
                Updater.Ease( function(v) self.battle_cards_widget.bg:SetTintAlpha(v) end, 0, 1, card_anim_duration*2, easing.inQuad),
                Updater.Ease( function(v) self.battle_cards_widget.bg:SetScale(v,v) end, 1.1, 1.0, card_anim_duration, easing.inQuad),
            }
        }
    )

    -- Prepare negotiation card animation
    local card_anim_duration_total = 0.3
    local card_number = math.min( #self.negotiation_cards_widget.content.children, 14 )
    local card_anim_duration = 0.1
    local delay_per_card = card_anim_duration_total / card_number
    local negotiation_card_adder = Updater.Parallel()
    for i, widget in ipairs( self.negotiation_cards_widget.content.children ) do
        widget:SetTintAlpha( 0 )
        negotiation_card_adder:Add(
            Updater.Series{
                Updater.Wait(delay_per_card*i),
                Updater.Parallel{
                    Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/approach_list_appears") end ),
                    Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, card_anim_duration, easing.inQuad),
                    Updater.Ease( function(v) widget:SetScale(v,v) end, CARD_SCALE * 1.2, CARD_SCALE, card_anim_duration, easing.inQuad)
                }
            }
        )
    end
    negotiation_card_adder:Add( 
        Updater.Series{
            Updater.Wait(delay_per_card*(#self.negotiation_cards_widget.content.children+1)),
            Updater.Parallel{
                Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/new_card_unlocked") end ),
                Updater.Ease( function(v) self.negotiation_cards_widget.title:SetTintAlpha(v) end, 0, 1, card_anim_duration*2, easing.inQuad),
                Updater.Ease( function(v) self.negotiation_cards_widget.subtitle:SetTintAlpha(v) end, 0, 1, card_anim_duration*1.5, easing.inQuad),
                Updater.Ease( function(v) self.negotiation_cards_widget.bg:SetTintAlpha(v) end, 0, 1, card_anim_duration*2, easing.inQuad),
                Updater.Ease( function(v) self.negotiation_cards_widget.bg:SetScale(v,v) end, 1.1, 1.0, card_anim_duration, easing.inQuad),
            }
        }
    )

    -- Prepare outfit animation
    local outfit_anim_duration_total = 0.3
    local outfit_number = math.min( #self.outfit_widget.content.children, 14 )
    local outfit_anim_duration = 0.1
    local delay_per_outfit = outfit_anim_duration_total / outfit_number
    local outfit_adder = Updater.Parallel()
    if #self.outfit_widget.content.children > 0 then
        for i, widget in ipairs( self.outfit_widget.content.children ) do
            widget:SetTintAlpha( 0 )
            outfit_adder:Add(
                Updater.Series{
                    Updater.Wait(delay_per_outfit*i),
                    Updater.Parallel{
                        Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/new_card_unlocked") end ),
                        Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, outfit_anim_duration, easing.inQuad),
                        Updater.Ease( function(v) widget:SetScale(v,v) end, 1.2, 1, outfit_anim_duration, easing.inQuad)
                    }
                }
            )
        end
        outfit_adder:Add( 
            Updater.Series{
                Updater.Wait(delay_per_outfit*(#self.outfit_widget.content.children+1)),
                Updater.Parallel{
                    Updater.Do( function() AUDIO:PlayEvent("event:/ui/game_over_screen/approach_list_appears") end ),
                    Updater.Ease( function(v) self.outfit_widget.title:SetTintAlpha(v) end, 0, 1, outfit_anim_duration*2, easing.inQuad),
                    Updater.Ease( function(v) self.outfit_widget.subtitle:SetTintAlpha(v) end, 0, 1, outfit_anim_duration*1.5, easing.inQuad),
                    Updater.Ease( function(v) self.outfit_widget.bg:SetTintAlpha(v) end, 0, 1, outfit_anim_duration*2, easing.inQuad),
                    Updater.Ease( function(v) self.outfit_widget.bg:SetScale(v,v) end, 1.1, 1.0, outfit_anim_duration, easing.inQuad),
                }
            }
        )
    end

    self:RunUpdater(
        Updater.Series{
            negotiation_card_adder,
            battle_card_adder,
            outfit_adder,
            Updater.Wait( 0.5 ),
            Updater.Do( function()
                self.continue_button:Show()
            end ),
        }
    )

end

------------------------------------------------------------------------------------------
-- Shows a rewards block

local UnlocksInspectorBlock = class("Widget.UnlocksInspectorBlock", Widget)

function UnlocksInspectorBlock:init( colour )
    UnlocksInspectorBlock._base.init(self)

    self.colour = colour or UICOLOURS.SUBTITLE
    self.width = 200
    self.height = 200

    self.bg_border = 45
    self.bg_border_scale = 0.5
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/gameover_block.tex") ) )
    self.bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetTintColour( self.colour )
    
    self.header = self:AddChild( Widget() )
    self.title = self.header:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_NAME ) )
        :SetGlyphColour( self.colour )
        :Bloom( 0.1 )
    self.subtitle = self.header:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_DESC, LOC"UI.UNLOCK_INSPECTOR.NEGOTIATION_CARDS_TITLE" ) )
        :SetGlyphColour( self.colour )
        :SetTintAlpha( 0.7 )
        :Bloom( 0.05 )
    self.content = self:AddChild( Widget() )

end

function UnlocksInspectorBlock:SetTitles( title, subtitle )
    if title then self.title:SetText( title ) end
    if subtitle then self.subtitle:SetText( subtitle ) end
    self:Layout()
    return self
end

function UnlocksInspectorBlock:AddContent( unlock_item )
    if unlock_item.battle_card then
        self.content:AddChild( Widget.BattleCard( Battle.Card(unlock_item.battle_card) ) )
            :SetScale(CARD_SCALE)
            :SetFocusable(true)
            :ShowToolTipOnFocus(true)
            :LayoutBounds( "after", "center" )
    elseif unlock_item.negotiation_card then
        self.content:AddChild( Widget.NegotiationCard( Negotiation.Card(unlock_item.negotiation_card) ) )
            :SetScale(CARD_SCALE)
            :SetFocusable(true)
            :ShowToolTipOnFocus(true)
            :LayoutBounds( "after", "center" )
    elseif unlock_item.outfit_id then
        local outfit = Content.GetOutfit(unlock_item.outfit_id)
        self.content:AddChild( Widget.Image( outfit.icon ) )
            :SetToolTip(loc.format( "<#TITLE><b>{1}</></>\n{2}", outfit:GetName(), outfit:GetDesc()))
            :SetFocusable(true)
            :ShowToolTipOnFocus(true)
            :LayoutBounds( "after", "center" )
    end
    return self
end

function UnlocksInspectorBlock:RefreshSize( height )
    local content_w, content_h = self.content:GetSize()
    content_w = content_w + PADDING*1
    local nineslice_w, nineslice_h = content_w + self.bg_border*2*self.bg_border_scale, height + self.bg_border*2*self.bg_border_scale
    self.bg:SetSize( nineslice_w, nineslice_h )
    self:Layout()
    return self
end

function UnlocksInspectorBlock:Layout()
    self.subtitle:LayoutBounds( "left", "below", self.title )
        :Offset( 0, SPACING.TITLE_DESC )
    self.header:LayoutBounds( "left", "top", self.bg )
        :Offset( PADDING, -PADDING )
    local header_width, header_height = self.header:GetScaledSize()
    self.content:LayoutBounds( "center", "center", self.bg )
        :Offset( 0, -header_height/2 )
    return self
end

function UnlocksInspectorBlock:GetHeight()
    local header_width, header_height = self.header:GetScaledSize()
    local content_width, content_height = self.content:GetSize()
    return header_height+content_height+PADDING*3
end

function UnlocksInspectorBlock:GetDefaultFocus()
    return self.content and self.content.children and #self.content.children > 0 and self.content.children[1]
end

function UnlocksInspectorBlock:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function UnlocksInspectorBlock:OnControlModeChange( cm, device_id )
    UnlocksInspectorBlock._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
    end
    if cm == CONTROL_MODE.TOUCH then
    end
    if cm == CONTROL_MODE.GAMEPAD then
    end
end

function UnlocksInspectorBlock:OnScreenModeChange( sm )
    UnlocksInspectorBlock._base.OnScreenModeChange( self, sm )
    
    self.header:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:Layout()
end


