require "ui/widgets/screen"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/menu"
require "ui/widgets/solidbox"
require "ui/widgets/panelbutton"
require "ui/screens/deckscreen"

local ORIGINAL_CARD_SCALE = 0.8
local CARD_SCALE = 0.7
local PANEL_H = RES_Y * 0.55

local UpgradeGraftPopup = class("Screen.UpgradeGraftPopup", Widget.Screen)

UpgradeGraftPopup.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return not self.preview_only and TheGame:FE():IsGamepadMode()
        end,
        fn = function(self)
            self.chosen_graft = self.upgraded_graft
            self:OnCancel()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.UPGRADE_GRAFT_POPUP.CONTROLS_UPGRADE", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        fn = function(self)
            AUDIO:PlayEvent( "event:/ui/grafts_menu/close" )
            self:OnCancel()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CLOSE", Controls.Digital.MENU_CANCEL))
        end
    }
}


function UpgradeGraftPopup:init( graft, agent, preview_only, fn )
	UpgradeGraftPopup._base.init(self)
	self.preview_only = preview_only
    self.is_popup = true
    self:SetupUnderlay( true )
    self:SetAnchors( "center", "center" )
	
	self.original_graft = graft
	self.agent = agent
    self.fn = fn

    -- Start assembling our screen
    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Setup the grid
    self.grid = self:AddChild( Widget.Image( engine.asset.Texture( "UI/grid.tex" ) ) )
    self.grid:SetTintColour( UICOLOURS.ACCENT )
    self.grid:SetTintAlpha( 0.8 )
    local target_w = self.screen_w*1.1
    local g_w, g_h = self.grid:GetSize();
    local target_h = target_w/g_w*g_h; -- Scale the grid to match the width, mantaining aspect ratio
    self.grid:SetSize( target_w, target_h )
    self.grid:SetPos( nil, -self.screen_h/2 + target_h/2 )

    -- Setup character anim display
    self.character = self:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.agent )
    self.character:SetHoverOutlineColour( UICOLOURS.GRAFT )
    self.character_animation_done = true
    self.character.anim:SetAnimDoneFn( function() self.character_animation_done = true end )
    self.character:SetClickFn( function()
        if self.character_animation_done then
            self.character_animation_done = false
            self.character:Emote( "thought", true )
        end
    end )
    self.character:SetAutoLoopEnabled( false )
    self.character:Emote( "thought", true )
    self.character:SetPos( -400, -350 )

    -- Add contents!
    self.grafts_container = self:AddChild( Widget() )

    -- Show starting graft
    self.original_graft_container = self.grafts_container:AddChild( Widget() )
    self.original_graft_header = self.original_graft_container:AddChild( Widget() )
    self.original_graft_title = self.original_graft_header:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_SUBTITLE ))
        :SetText( LOC"UI.UPGRADE_GRAFT_POPUP.ORIGINAL_GRAFT_TITLE" )
        :SetGlyphColour( UICOLOURS.GRAFT )
        :Bloom(0.1)
    self.original_graft_subtitle = self.original_graft_header:AddChild( Widget.Label("title", FONT_SIZE.BODY_TEXT ) )
        :SetText( self.preview_only and loc.format( LOC"UI.UPGRADE_GRAFT_POPUP.UPGRADED_GRAFT_SUBTITLE", self.original_graft:GetXPDesc() ) or LOC"UI.UPGRADE_GRAFT_POPUP.UPGRADED_GRAFT_READY_SUBTITLE" )
        :SetGlyphColour( UICOLOURS.GRAFT )
        :SetAutoSize( 660 )
        :LeftAlign()
        :SetWordWrap( true )
        :SetTintAlpha( 0.8 )
        :Bloom( 0.1 )
    self.original_graft_widget = self.original_graft_container:AddChild( Widget.GraftDetails( 600, agent, true ) )
        :HideUpgradeLabel()
        :Refresh( self.original_graft )
        :SetOnClickFn( nil )

    -- Show upgraded graft
    self.upgraded_graft_container = self.grafts_container:AddChild( Widget() )
    self.upgraded_graft_header = self.upgraded_graft_container:AddChild( Widget() )
    self.upgraded_graft = GraftInstance( self.original_graft.userdata.upgrade_ids[1] )
    self.upgraded_graft_title = self.upgraded_graft_header:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_SUBTITLE ))
        :SetText( LOC"UI.UPGRADE_GRAFT_POPUP.UPGRADED_GRAFT_TITLE" )
        :SetGlyphColour( UICOLOURS.MONEY )
        :Bloom(0.1)
    self.upgrade_available = self.upgraded_graft_header:AddChild( Widget.GraftSlotUpgradeAvailable() )
        :SetShown( not self.preview_only )
    self.upgraded_graft_widget = self.upgraded_graft_container:AddChild( Widget.GraftDetails( 600, agent, true ) )
        :Refresh( self.upgraded_graft )

    -- Keywords panel
    self.graft_keywords = {}
    self.original_graft:GetDesc( self )
    self.upgraded_graft:GetDesc( self )
    self.keywords_block = self.grafts_container:AddChild( Widget.KeywordsList() )
        :Refresh( self.graft_keywords )

    -- Back button
    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.cancel_btn = self.bottom_left:AddChild( Widget.IconButton( preview_only and LOC"UI.DIALOGS.BACK" or LOC"UI.DIALOGS.CANCEL", function() self:OnCancel() end ) )
    self.cancel_btn:SetIcon( global_images.close )
    self.cancel_btn:LayoutBounds( "left", "above", 60, 60 )

    -- Upgrade button
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    if not self.preview_only then
        self.upgrade_btn = self.bottom_right:AddChild( Widget.IconButton( LOC"UI.UPGRADE_GRAFT_POPUP.UPGRADE_BTN", function() self.chosen_graft = self.upgraded_graft self:OnCancel() end ) )
            :SetIcon( engine.asset.Texture( "UI/ic_nav_upgrade.tex") )
            :SetSize( 250 )
            :LayoutBounds( "right", "above", -60, 60 )
    end
end

function UpgradeGraftPopup:PostFeature( feature_id, name, desc )
    self.graft_keywords[ feature_id ] = { name = name, desc = desc }
end

function UpgradeGraftPopup:PostCard( feature_id, card )
    self.graft_keywords[ feature_id ] = { card = card }
end

function UpgradeGraftPopup:GetDefaultFocus()
    return nil
end

function UpgradeGraftPopup:OnGainTop()
	UpgradeGraftPopup._base.OnGainTop( self )

	-- Prepare for animation by getting current values
    local anim_pacing = 0.3
    local character_x, character_y = self.character:GetPos()
    local original_graft_container_x, original_graft_container_y = self.original_graft_container:GetPos()
    local upgraded_graft_container_x, upgraded_graft_container_y = self.upgraded_graft_container:GetPos()

	-- Hide objects until they're animated in
    self.upgraded_graft_container:SetTintAlpha(0)
    self.original_graft_container:SetTintAlpha(0)

    AUDIO:PlayEvent(SoundEvents.upgrade_card_popup)

    -- Our main animation sequence
    local animation_series = Updater.Parallel {
        Updater.Ease( function(v) self.character:SetPos(v,character_y) end, 0, character_x, anim_pacing, easing.inQuad),
        Updater.Series {
            Updater.Wait( anim_pacing*0.3 ),
            Updater.Parallel {
                Updater.Ease( function(v) self.original_graft_container:SetPos(v,original_graft_container_y) end, original_graft_container_x -80, original_graft_container_x, anim_pacing, easing.inQuad),
                Updater.Ease( function(v) self.original_graft_container:SetTintAlpha(v) end, 0, 1, anim_pacing, easing.inQuad),
            },
        },
        Updater.Series {
            Updater.Wait( anim_pacing*0.6 ),
            Updater.Parallel {
                Updater.Ease( function(v) self.upgraded_graft_container:SetPos(v,upgraded_graft_container_y) end, upgraded_graft_container_x -40, upgraded_graft_container_x, anim_pacing, easing.inQuad),
                Updater.Ease( function(v) self.upgraded_graft_container:SetTintAlpha(v) end, 0, 1, anim_pacing, easing.inQuad),
            }
        },
    }

    -- Start animation
    self.fadein_anim = self:RunUpdater( animation_series )
end

function UpgradeGraftPopup:OnControlModeChange( cm, device_id )
    UpgradeGraftPopup._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.cancel_btn:Show()
        if self.upgrade_btn then self.upgrade_btn:Show() end
    end
    if cm == CONTROL_MODE.TOUCH then
        self.cancel_btn:Show()
        if self.upgrade_btn then self.upgrade_btn:Show() end
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.cancel_btn:Hide()
        if self.upgrade_btn then self.upgrade_btn:Hide() end
    end
end

function UpgradeGraftPopup:OnScreenModeChange( sm )
    UpgradeGraftPopup._base.OnScreenModeChange( self, sm )

    local grafts_y_offset = 0
    if sm == SCREEN_MODE.MONITOR then
        self.character_x_pos = 400
        grafts_y_offset = 20
    elseif sm == SCREEN_MODE.SMALL then
        self.character_x_pos = 650
        grafts_y_offset = 40
    elseif sm == SCREEN_MODE.TV then
        self.character_x_pos = 650
        grafts_y_offset = 40
    end

    self.bottom_left:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.bottom_right:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.original_graft_header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.original_graft_widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.upgraded_graft_header:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.upgraded_graft_widget:SetLayoutScale( LAYOUT_SCALE[sm] )
    self.keywords_block:SetLayoutScale( LAYOUT_SCALE[sm] )

    self.character:SetPos( -self.character_x_pos, -350 )
    self.original_graft_subtitle:LayoutBounds( "left", "below", self.original_graft_title ):Offset( 0, SPACING.TITLE_DESC )
    self.original_graft_widget:LayoutBounds( "left", "below", self.original_graft_header ):Offset( 0, -SPACING.M1 )
    self.upgrade_available:LayoutBounds( "left", "below", self.upgraded_graft_title ):Offset( -5, 4 )
    self.upgraded_graft_widget:LayoutBounds( "left", "below", self.upgraded_graft_header ):Offset( 0, -SPACING.M1 )
    self.keywords_block:LayoutBounds( "after", "top", self.original_graft_widget ):Offset( SPACING.M1, 0 )
        :SetShown( self.keywords_block:HasKeywords() )

    self.upgraded_graft_container:LayoutBounds( "left", "below", self.original_graft_container ):Offset( 0, -SPACING.M05 )
    self.grafts_container:LayoutBounds( "after", nil, self.character )
        :LayoutBounds( nil, "center", 0, grafts_y_offset )
        :Offset( SPACING.M1, 0 )

end

function UpgradeGraftPopup:RefreshStreamerMode()
    self.bottom_left:SetPos( 0, 0 )
    if GetStreamerMode() == 1 then
        -- Nudge the left side widgets
        self.bottom_left:SetPos( 360, 0 )
    end
end

function UpgradeGraftPopup:OnCancel()

    -- Prepare for animation by getting current values
    local anim_pacing = 0.3
    local character_x, character_y = self.character:GetPos()
    local original_graft_container_x, original_graft_container_y = self.original_graft_container:GetPos()
    local upgraded_graft_container_x, upgraded_graft_container_y = self.upgraded_graft_container:GetPos()

    self.character:Emote( "neutral", true )

    -- Our main animation sequence
    local animation_series = Updater.Parallel {
        Updater.Ease( function(v) self.character:SetPos(v,character_y) end, character_x, 0, anim_pacing, easing.linear),
        Updater.Series {
            Updater.Wait( anim_pacing*0.3 ),
            Updater.Parallel {
                Updater.Ease( function(v) self.original_graft_container:SetPos(v,original_graft_container_y) end, original_graft_container_x, original_graft_container_x + 200, anim_pacing, easing.inQuad),
                Updater.Ease( function(v) self.original_graft_container:SetTintAlpha(v) end, 1, 0, anim_pacing, easing.inQuad),
            },
        },
        Updater.Series {
            Updater.Wait( anim_pacing*0.6 ),
            Updater.Parallel {
                Updater.Ease( function(v) self.upgraded_graft_container:SetPos(v,upgraded_graft_container_y) end, upgraded_graft_container_x, upgraded_graft_container_x + 160, anim_pacing, easing.inQuad),
                Updater.Ease( function(v) self.upgraded_graft_container:SetTintAlpha(v) end, 1, 0, anim_pacing, easing.inQuad),
            },
            Updater.Do( function()
                self:GetFE():PopScreen( self )
                if self.fn then
                    self.fn()
                end
            end )
        },
    }

    -- Start animation
    self.fadein_anim = self:RunUpdater( animation_series )
end

