require "ui/widget"
require "ui/widgets/characterportrait"
require "ui/widgets/label"
require "ui/widgets/solidbox"
require "ui/widgets/textbox"
require "ui/widgets/tooltipagent"

local AgentStatusPortrait = class( "Widget.AgentStatusPortrait", Widget )

local ICON_SIZE = 26
local ITEM_SPACING_X = 10
local ITEM_SPACING_Y = 0

local PORTRAIT_H = 100

function AgentStatusPortrait:init( facing_right )
    Widget.init( self )

    self.blocks_mouse = true
    self.hover_check = true

    local height = PORTRAIT_H
    local max_bubble_width = 300

    self.facing_right = facing_right

    self.root = self:AddChild(Widget())

    self.portrait = Widget.CharacterPortrait(nil, facing_right):SetHeight(height)
    self.portrait:SetFacing( self.facing_right )
    
    self.name = Widget.Label( "title", 40 ):RightAlign():SetGlyphColour(UICOLOURS.TITLE)
    self.desc = Widget.Label("title", 24 ):RightAlign():SetGlyphColour(UICOLOURS.SUBTITLE)
    if self.facing_right then
        self.name:LeftAlign()
        self.desc:LeftAlign()
    end
    
    self.status = Widget():Hide()

    self.root:AddChildren{
        self.portrait,
        self.name,
        self.desc,
        self.status,
    }
    self.default_scale = 1

    self.status_icons = {}
    self.status_count = 0

    self.portrait:SetFn( function() self:OnClick() end)
end

function AgentStatusPortrait:OnClick()
    if self.character then
        local details = Screen.PeopleCompendiumDetailsScreen()
        self:GetFE():PushScreen( details )
        details:ShowAgent( self.character )
    end
end

function AgentStatusPortrait:OnRemoved()
    if self.character then
        self.character:RemoveListener( self )
    end
end

function AgentStatusPortrait:ShowName( show )
    if show then
        self.name:Show()
        self.desc:Show()
    else
        self.name:Hide()
        self.desc:Hide()
    end

    self:DoLayout()

    return self
end

function AgentStatusPortrait:ShowOpinion( show )
    if show then
        -- self.opinion_item:Show()
    else
        -- self.opinion_item:Hide()
    end

    self:DoLayout()

    return self
end

function AgentStatusPortrait:ShowHealth( show )
    if self.health_bar == nil and show then
        local portraitw, portraith = self.portrait:GetSize()
        self.health_bar = Widget.ProgressBar( portraitw - 10, 10 ):OrientH():SetBarColour(UICOLOURS.HEALTH ):SetBGColour( 0x000000FF )
        self.health_bar:SetOutline(3)
        self:AddChild( self.health_bar ):LayoutBounds( "left", "below", self.portrait ):Offset( 5, 8 )

    elseif self.health_bar and not show then
        self.health_bar:Remove()
        self.health_bar = nil
    end

    self:UpdateStatusIcons()
    return self
end

function AgentStatusPortrait:ShowStatusIcon( texture )
    self.status_count = self.status_count + 1

    local icon = self.status_icons[ self.status_count ]
    if icon == nil then
        icon = Widget.Image():SetSize( ICON_SIZE, ICON_SIZE )
        table.insert(self.status_icons, icon)
        self.status:AddChild(icon)

        -- Make it obvious
        icon:ScaleTo(1.2, 1, 0.3)
        icon:ColourTo(3,3,3,3, 1,1,1,1, 0.5)
    end

    icon:SetTexture( texture )
    icon:SetToolTip( nil )

    icon:Show()

    self:DoLayout()

    return icon
end

function AgentStatusPortrait:UpdateStatusIcons()
    if self.health_bar then
        local health, max_health = self.character:GetHealth()
        self.health_bar:SetProgress( health / max_health )
        self.health_bar:SetToolTip( loc.format(LOC"UI.AGENT_STATUS_PORTRAIT.HEALTH_TOOLTIP", health, max_health ) )
    end

    for k, v in ipairs(self.status_icons) do
        v:Hide()
    end
    self.status_count = 0

    if self.character then
        for i, aspect in self.character:Aspects( "status_effect" ) do
            local show, texture = aspect:ShouldShow( self.character )
            if show then
                local icon = self:ShowStatusIcon( texture or aspect.texture )
                icon:SetToolTip( aspect:FormatDesc() )
            end
        end
    end

    self:DoLayout()
end

function AgentStatusPortrait:SetMood(mood)
    -- self.portrait:GetAnim():SetMood(mood)
    return self
end


function AgentStatusPortrait:GetCharacter( )
    return self.character
end

function AgentStatusPortrait:SetCharacter( char )
    assert( is_instance( char, Agent ))

    if self.character then
        self.character:RemoveListener( self )
    end

    self.character = char
    self.character:ListenForAny( self, self.OnAgentEvent )
    
    self.portrait:SetCharacter( self.character )

    self:UpdateName()

    self:UpdateStatusIcons()
    
    self.status:Show()

    self:DoLayout()

    return self
end

function AgentStatusPortrait:OnStatChange(eventname, member, ...)
    if self.character.guid == member.guid then
        self:UpdateStatusIcons()
    end
end

function AgentStatusPortrait:OnAgentEvent(eventname, ...)
    if eventname == "equipment_changed" then
        self.portrait:SetCharacter(self.character)

    elseif eventname == "loc_changed" then
        self:UpdateName()
        self:DoLayout()

    elseif eventname == "aspects_changed" then
        self:UpdateStatusIcons()
        
    elseif eventname == "health_change" then
        if self.health_bar then
            self:UpdateStatusIcons()
        end
    elseif eventname == "delta_renown" then
        local delta = ...
        self.portrait:ShowFloater( loc.format( "{1%+d}", delta ), 0, 20, THexColour( UICOLOURS.TEXT ))
    end
end

function AgentStatusPortrait:UpdateName()
    self.name:SetText( self.character:GetName() or self.character:GetShortTitle() )
    self.desc:SetText( self.character:GetIdentity() )
    self:SetToolTip( self.character:GetBio() )
end

function AgentStatusPortrait:SetInnerScale(s)
    self.default_scale = s
    self.root:SetScale(s,s)
    return self
end

function AgentStatusPortrait:DoLayout()
    self.name:LayoutBounds( self.facing_right and "after" or "before", "top", self.portrait ):Offset( self.facing_right and 10 or -10, -5 )
    self.desc:LayoutBounds( self.facing_right and "left" or "right", "below", self.name ):Offset( 0, 7 )

    -- Show opinion icon
    if self.opinion then
        if not self.character or self.character:IsPlayer() then
            self.opinion:Hide()
        else
            if self.character.species == SPECIES.BEAST then
                self.opinion:Hide()
            else
                self.opinion:Show()
            end
        end
    end

    self.status:ArrangeChildrenRow( ICON_SIZE + ITEM_SPACING_X )
    self.status:LayoutBounds( self.facing_right and "left" or "right", "below", self.desc ):Offset( 0, -8 )
end

function AgentStatusPortrait:OnGainFocus()
    self:ScaleTo( nil, 1.1, 0.2, easing.outQuad )
end

function AgentStatusPortrait:OnLoseFocus()
    self:ScaleTo( nil, 1, 0.2, easing.outQuad )
end