require "ui/widget"
require "helpers"

local Anim = class("Widget.Anim", Widget)

function Anim:init()
    self.model = engine.model.AnimModel()
    self.model:SetWallTime(true)
    self.model:SetCanFlip(false)
    Anim._base.init(self, self.model)
    self.anims = {}    
    self:SetScale(1)
    self.nodebug = true
    self.model:SetEffect(global_shaders.UI_ANIM)
    self.model:SetOutlineEffect(global_shaders.ANIM_OUTLINE)
    self.model:SetDepthState(DEPTH_TEST.LESSEQUAL, false, 0)
end


function Anim:ApplyHSV(hsvtable)
    Anim._base.ApplyHSV(self, hsvtable)

    if not self.special_effect then
        if self:HasNonIdentityHSVSettings() then
            self.model:SetEffect(global_shaders.UI_ANIM_HSV)
        else
            self.model:SetEffect(global_shaders.UI_ANIM)
        end
    end
end

function Anim:SetMultiplyShader()
    self.special_effect = true
    self:SetEffect( global_shaders.FX_ANIM_PREMULT )
end

function Anim:SetErodeShader()
    self.special_effect = true
	self:SetEffect( global_shaders.FX_ANIM_ERODE )
end



function Anim:__tostring()
    local bldname = self.build and self.build:GetFileName() or "null"
    if self.name then
        return string.format("Anim[%s: %s]", self.name, bldname )
    else
        return string.format("Anim[%s]", bldname )
    end
end


function Anim:SetScale(x,y)
    Anim._base.SetScale(self, x/ ANIM_IMPORT_SCALE,(y or x)/ANIM_IMPORT_SCALE)
    return self
end


function Anim:SetBuild(build_asset)
	self.model:SetBuild(build_asset)
    self.build = build_asset
    return self
end

function Anim:Reset()
    self:Clear()
    return self
end

function Anim:Clear()
    
    if self.ondonetask then
        self.ondonetask:Cancel()
        self.ondonetask = nil
    end
    if self.queue then
        table.clear(self.queue)
    end

    if self.anims then
        table.clear(self.anims)
    end
    self.build = nil
    self.animName = nil
    if self.subanims then
        table.clear(self.subanims)
    end
    self.model:Reset()
    self.special_effect = false
    return self
end

function Anim:GetCurrentAnimName()
    return self.animName
end

function Anim:AddAnims(anim_asset)
    for k = 1, anim_asset:GetNumAnims() do
        local name = anim_asset:GetAnimName(k-1)
        self.anims[name] = {anim=anim_asset, idx = k-1, name=name, duration=anim_asset:GetAnimDuration(k-1)}
    end
    return self
end

function Anim:SetSpeed(v)
    self.model:SetPlaySpeed(v)
    return self
end

function Anim:SetPhase(v)
    self.phase = clamp(v,0,1)
    if self.start_time then
        local dur = self.model:GetDuration()
        self.model:SetStartTime(self.start_time - self.phase*dur)
    end
    return self
end

function Anim:AddSymbolAnim(symbol, anim_name, anim_asset, idx)
    self.subanims = self.subanims or {}
    self.subanims[symbol] = self.subanims[symbol] or {}
    self.subanims[symbol][anim_name] = {anim=anim_asset, idx = idx, name=anim_name, duration=anim_asset:GetAnimDuration(idx)}
    return self
end

function Anim:SetMask()
    --self.model:SetColourWrite(false)
    --self.model:SetEffect(global_shaders.UI_ANIM_MASK)
    self:SetStencilWrite(STENCIL_MODES.SET)
    return self
end

function Anim:SetMaskOnly()
    self.model:SetColourWrite(false)
    self:SetStencilWrite(STENCIL_MODES.SET)
    return self
end

function Anim:GetBoundingBox()
    local x1, y1, x2, y2 = self.model:GetBoundingBox()
    if x1 then
        return x1, y1, x2, y2
    else
        return 0, 0, 0, 0
    end
end


function Anim:GetNormalizedTime()
    if self.anim then
        local dur = self.model:GetDuration()
        local t = GetWallTime() - self.start_time
        return t / dur
    end
    return 0
end

function Anim:OnQueueDone(fn)
    if self:HasQueuedAnims() then
        self.onqueuedonefn = fn
    else
        fn()
    end
    return self
end

function Anim:GetAnimDuration(anim)
    local animdat = self.anims[anim]
    if animdat then
        return animdat.duration
    end

    return 0
end
    

function Anim:Push(anim, loop_at_end)
    
    if not self.ondonetask then
        self:Play(anim, loop_at_end)
        return self
    end

    self.queue = self.queue or {}
    self.loop_at_end_of_queue = loop_at_end

    table.insert(self.queue, anim)
    return self
end

function Anim:HasQueuedAnims()
    return self.queue and #self.queue > 0
end

function Anim:SetAutoRemove()
    self.autoremove = true
    return self
end

function Anim:SetAutoHide()
    self.autohide = true
    return self
end

function Anim:SetAnimDoneFn( fn )
    self.on_anim_done_fn = fn
    return self
end

function Anim:OnAnimDone(animdat)
    self.ondonetask = nil
    if self.on_anim_done_fn then self.on_anim_done_fn( animdat ) end
    if self.queue and #self.queue > 0 then
        local anim = table.remove( self.queue, 1)
        self:Play(anim, self.loop_at_end_of_queue and #self.queue == 0, 0, true)
    else
        if self.autoremove then
            self:Remove()
        elseif self.autohide then
            self:Hide()
        end
    end
    return self
end

function Anim:PushSubAnim(symbol, anim, loop_at_end)
    
    if not self.ondonesubtask and not self.ondonesubtask[symbol] then
        self:PlaySubAnim(symbol, anim, loop_at_end, 0)
        return
    end

    self.subqueue = self.subqueue or {}
    self.subqueue[symbol] = self.subqueue[symbol] or {}
    self.loop_at_end_of_subqueue = self.loop_at_end_of_subqueue or self.loop_at_end_of_subqueue
    self.loop_at_end_of_subqueue[symbol] = loop_at_end
    table.insert(self.subqueue[symbol], anim)
    return self
end

function Anim:CheckHit(x,y)
    local lx, ly = self:TransformFromWorld(x,y)
    local hit, symbolhash = self.model:TestPoint(lx,ly)
    return hit, symbolhash
end


function Anim:OnSubAnimDone(animdat, symbol)
    if self.removed then
        return
    end
    self.ondonesubtask[symbol] = nil
    if self.subqueue and self.subqueue[symbol] and #self.subqueue[symbol] > 0 then
        local anim = table.remove( self.subqueue[symbol], 1)
        self:PlaySubAnim(anim, self.loop_at_end_of_subqueue and self.loop_at_end_of_subqueue[symbol] and #self.subqueue[symbol] == 0)
    else
        self:ClearSubAnim(symbol)
    end
    return self
end

function Anim:GetAnimStartTime()
    return self.start_time
end

function Anim:Play(anim, looping, timeoffset, preserve_queue)
    --print ("PLAY", anim, debugstack())
    if self.onqueuedonefn and not self:HasQueuedAnims() then
        self.onqueuedonefn()
        self.onqueuedonefn = nil
    end

    local animdat = self.anims[anim]

    if not preserve_queue then
        if self.ondonetask then
            self.ondonetask:Cancel()
            self.ondonetask = nil
        end

        if self.queue then
            table.clear(self.queue)
        end
    end
    
    timeoffset = timeoffset or 0
    if animdat then
        self.looping = looping == true
        self.animName = anim
        self.anim = animdat.anim
        self.model:SetLooping(self.looping)
        self.start_time = GetWallTime() - timeoffset or 0
        self.model:SetStartTime(self.start_time - (self.phase or 0)*animdat.duration)
        self.model:SetAnim(animdat.anim, animdat.idx)
        
        if not looping then
            self.ondonetask = self:DoTaskInTime("DONE", animdat.duration, Anim.OnAnimDone, self, animdat)
        end

    else
        --print("Missing UI anim " .. tostring(anim), debugstack() )
        --assert(animdat, "Missing UI anim " .. anim)
    end
    return self
end


function Anim:SetAutoLoopEnabled(v)
    self.model:SetAutoLoopEnabled(v and true or false)
    return self
end


function Anim:IsDone()
    local t, dur = self.model:GetPlayHeadTime()
    return t >= dur
end

function Anim:ClearSubAnim(symbol)
    if self.subqueue and self.subqueue[symbol] then
        table.clear(self.subqueue[symbol])
    end

    if self.ondonesubtask and self.ondonesubtask[symbol] then
        self.ondonesubtask[symbol]:Cancel()
        self.ondonesubtask[symbol] = nil
    end

    self.model:ClearSymbolAnim(symbol)
    return self
end


function Anim:PlaySubAnim(symbol, anim, looping, timeoffset, hold)
    --[[if symbol == "mouth_swap" then
        print (symbol, anim, looping, timeoffset) 
        print (self.subanims, self.subanims[symbol], self.subanims[symbol] and self.subanims[symbol][anim])
        print (serpent.block(self.subanims[symbol]))
        assert(nil)
    end--]]

    if self.subanims and self.subanims[symbol] and self.subanims[symbol][anim] then
        local animdat = self.subanims[symbol][anim] 
        self.ondonesubtask = self.ondonesubtask or {}

        if self.ondonesubtask[symbol] then
            self.ondonesubtask[symbol]:Cancel()
            self.ondonesubtask[symbol] = nil
        end

        if self.subqueue and self.subqueue[symbol] then
            table.clear(self.subqueue[symbol])
        end
        assert(animdat)
        if animdat then
            
            if self.model then
                assert(animdat.anim)
                --print (symbol, anim, animdat.idx)
                self.model:PlaySymbolAnim(symbol, animdat.anim, animdat.idx, looping or false, timeoffset or 0)
            end

            if not looping and not hold then
                self.ondonesubtask[symbol] = self:DoTaskInTime("SUBDONE", animdat.duration, Anim.OnSubAnimDone, self, animdat, symbol)
            end

        else
            print("Missing UI Sub anim " .. anim)
        end
    end
    return self
end


function Anim:HideSymbol(name)
    self.model:SetSymbolVisible(name, false)
    return self
end

function Anim:ShowSymbol(name)
    self.model:SetSymbolVisible(name, true)
    return self
end

function Anim:SetSymbolTint(symbol, r,g,b,a)
    self.model:SetSymbolTint(symbol, r, g, b, a)
    return self
end

function Anim:SetLayerTint(layer, r,g,b,a)
    self.model:SetLayerTint(layer, r, g, b, a)
    return self
end

function Anim:SetBloom(b)
    self.model:SetBloom(b)
    return self
end

function Anim:SetOutline(size, colour)
    self.model:SetOutlineSize(size)    
    local r,g,b,a = HexColour(colour)
    if not self.outline_colour_constant then
        self.outline_colour_constant = engine.rendering.ShaderConstantValue("OUTLINE_COLOUR", r, g, b, a)
        self.model:SetShaderConstant(self.outline_colour_constant)
    else
        self.outline_colour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end
    return self
end

function Anim:HideOutline()
    self.model:SetOutlineSize(0)  
    return self  
end

function Anim:AddOverrideFile(anim_file_name, symbol)
    self.model:AddOverrideFile(anim_file_name, symbol)
    return self
end

function Anim:SetClickFn( fn )
    self.click_fn = fn
    self.hover_check = fn ~= nil
    return self
end

function Anim:HandleControlDown(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) and self.hover then
        if self.enabled then
            self.down = true
            return true
        end
    end
end

function Anim:HandleControlUp(control, dev)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if self.down then
            self.down = false

            if self.hover or self.focus then
                if self.enabled then
                    if self.click_fn then
                        self.click_fn(dev)
                    end
                end
                return true
            end
        end
    end
end

function Anim:OnRemoved()
    if self.ondonetask then
        self.ondonetask:Cancel()
    end

    if self.ondonesubtask then
        for k,v in pairs( self.ondonesubtask) do
            v:Cancel()
        end
    end
    self:Clear()
end

function Anim:SetXRayOutline(fillColour, lineColour, thresholdMin, thresholdMax)
    self.special_effect = true
    self.model:SetEffect(global_shaders.UI_ANIM_XRAY)

    local minthreshold = math.min( thresholdMin or 0.05, thresholdMax or 0.1)
    local maxthreshold = math.max( thresholdMin or 0.05, thresholdMax or 0.1)

    if not self.xray_threshold_constant then
        self.xray_threshold_constant = engine.rendering.ShaderConstantValue("XRAY_THRESHOLD", minthreshold, maxthreshold)
        self.model:SetShaderConstant(self.xray_threshold_constant)
    else
        self.xray_threshold_constant:SetValue( minthreshold, maxthreshold )
    end

    local r,g,b,a = HexColour(fillColour)
    if not self.xray_fillcolour_constant then
        self.xray_fillcolour_constant = engine.rendering.ShaderConstantValue("XRAY_FILL_COLOUR", r, g, b, a)
        self.model:SetShaderConstant(self.xray_fillcolour_constant)
    else
        self.xray_fillcolour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end

    r,g,b,a = HexColour(lineColour)
    if not self.xray_linecolour_constant then
        self.xray_linecolour_constant = engine.rendering.ShaderConstantValue("XRAY_LINE_COLOUR", r, g, b, a)
        self.model:SetShaderConstant(self.xray_linecolour_constant)
    else
        self.xray_linecolour_constant:SetValue(r or 1, g or 1, b or 1, a or 1)
    end
    return self
end

function Anim:SetSelected(enabled)
    self.model:SetSelected(enabled)
    return self
end

function Anim:SetSelectedLayerSpeeds(layer1Speed, layer2Speed, layer3Speed)
    self.model:SetSelectedLayerSpeeds(layer1Speed, layer2Speed, layer3Speed)
    return self
end

function Anim:SetSelectedLayerScale(scale)
    self.model:SetSelectedLayerScale(scale)
    return self
end

function Anim:SetSelectedLayerAlphas(layer1Alpha, layer2Alpha, layer3Alpha)
    self.model:SetSelectedLayerAlphas(layer1Alpha, layer2Alpha, layer3Alpha)
    return self
end

function Anim:SetSelectedFadeDuration(fadeDuration)
    self.model:SetSelectedFadeDuration(fadeDuration)
    return self
end

function Anim:SetSelectedColor(colour)
    local r,g,b,a = HexColour(colour)
    self.model:SetSelectedColor(r or 1, g or 1, b or 1, a or 1)
    return self
end

function Anim:SetGlitch(enabled)
    self.model:SetGlitch(enabled)
    return self
end

function Anim:SetGlitchSpeedAndScale(speed, scale)
    self.model:SetGlitchSpeedAndScale(speed, scale)
    return self
end

function Anim:SetGlowThresholdAndAmount( threshold, amount)
    self.model:SetGlowThresholdAndAmount(threshold, amount)
    return self
end

function Anim:SetGlowColour(colour)
    local r,g,b,a = HexColour(colour)
    self.model:SetGlowColour(r, g, b, a)
    return self
end