require "ui/widget"
require "ui/widgets/anim"

local paperdoll = require "paperdoll_util"
local posedefs = require "content/character_pose_defs"

local AnimatedCharacter = class( "Widget.AnimatedCharacter", Widget )
local assets =
{
    shadow = engine.asset.Texture('images/blob.tex'),
}

function AnimatedCharacter:init( )
    Widget.init( self )
    
    self.shadow = Widget.Image(assets.shadow)
    self.shadow:SetTintColour(0,0,0,.5)
        :SetHiddenBoundingBox(true)
        :Hide()

    self.anim = Widget.Anim()
    self:AddChildren{
        self.shadow,
        self.anim
    }

    self.emote_queue = {}
    self.mouth_queue = {}
    self.mood = MOOD.NEUTRAL
    self.anim.model:SetAutoLoopBaseTime(math.random())
    self.foley_sounds = {}
    self.emote_fx = {}
end

function AnimatedCharacter:OnVizChange(viz)
    if viz then  
        self:StartUpdating()
    else
        self:StopUpdating()
    end
    return self
end

function AnimatedCharacter:OnAdded()
    TheGame:GetEvents():ListenForEvent( "character_skins_changed", self )
end

function AnimatedCharacter:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
    
    if self.char then
        self.char:RemoveListener(self)
    end
end

local MOOD_VOICES = 
{
    [MOOD.ANGRY] = "angryDefault",
    [MOOD.NEUTRAL] = "neutralDefault",
    [MOOD.HAPPY] = "happyDefault",
    [MOOD.SCARED] = "scaredDefault",
    [MOOD.INJURED] = "neutralDefault",
    [MOOD.DRUNK] = "neutralDefault",
}
--expose this to modding?
local idle_anims =
{
    [MOOD.ANGRY] = "idle_angry",
    [MOOD.NEUTRAL] = "idle_neutral",
    [MOOD.HAPPY] = "idle_happy",
    [MOOD.SCARED] = "idle_scared",
    [MOOD.INJURED] = "idle_injured",
    [MOOD.DRUNK] = "idle_drunk",
}

local portrait_idle_anims = 
{
    [MOOD.NEUTRAL] = "portrait_neutral",
    [MOOD.ANGRY] = "portrait_angry",
    [MOOD.HAPPY] = "portrait_happy",
    [MOOD.SCARED] = "portrait_scared",
    [MOOD.INJURED] = "portrait_injured",
    [MOOD.DRUNK] = "portrait_drunk",
}

local transition_anims = 
{
    [MOOD.NEUTRAL] =
    {
        [MOOD.HAPPY] = "neutral_happy",
        [MOOD.ANGRY] = "neutral_angry",
        [MOOD.SCARED] = "neutral_scared",
        [MOOD.DRUNK] = "neutral_drunk",
        [MOOD.INJURED] = "neutral_injured",
    },
    [MOOD.ANGRY] =
    {
        [MOOD.NEUTRAL] = "angry_neutral",
        [MOOD.HAPPY] = "angry_happy",
        [MOOD.SCARED] = "angry_scared",
        [MOOD.DRUNK] = "angry_drunk",
        [MOOD.INJURED] = "angry_injured",

    },
    [MOOD.SCARED] =
     {
        [MOOD.NEUTRAL] = "scared_neutral",
        [MOOD.HAPPY] = "scared_happy",
        [MOOD.ANGRY] = "scared_angry",
        [MOOD.DRUNK] = "scared_drunk",
        [MOOD.INJURED] = "scared_injured",

    },
     [MOOD.HAPPY] =
     {
        [MOOD.NEUTRAL] = "happy_neutral",
        [MOOD.SCARED] = "happy_scared",
        [MOOD.ANGRY] = "happy_angry",
        [MOOD.DRUNK] = "happy_drunk",
        [MOOD.INJURED] = "happy_injured",

    },
    [MOOD.DRUNK] =
    {
        [MOOD.NEUTRAL] = "drunk_neutral",
        [MOOD.SCARED] = "drunk_scared",
        [MOOD.HAPPY] = "drunk_happy",
        [MOOD.ANGRY] = "drunk_angry",
        [MOOD.INJURED] = "drunk_injured",
    },
    [MOOD.INJURED] =
    {
        
        [MOOD.NEUTRAL] = "injured_neutral",
        [MOOD.HAPPY] = "injured_happy",
        [MOOD.SCARED] = "injured_scared",
        [MOOD.ANGRY] = "injured_angry",
        [MOOD.DRUNK] = "injured_drunk",
    },
}


local emotes = require "content/convo_emotes"


local face_anims = 
{
    [MOOD.ANGRY] = 
    { 
        blink = "blink_angry",
        talk = "dial_neutral",
    },
    [MOOD.SCARED] = 
    { 
        blink = "blink_scared",
        talk = "dial_neutral",
    },
    [MOOD.HAPPY] = 
    { 
        blink = "blink_happy",
        talk = "dial_happy",
    },
    [MOOD.NEUTRAL] =
    {
        blink = "blink_neutral",
        talk = "dial_neutral",
    },
    [MOOD.INJURED] =
    {
        blink = "blink_scared",
        talk = "dial_scared",
    },
    [MOOD.DRUNK] =
    {
        blink = "blink_happy",
        talk = "dial_happy",
    }
}

function AnimatedCharacter:GetIdleAnim(mood_override)
    local mood = mood_override or self.mood

    if self.portrait then
        return portrait_idle_anims[mood] or "portrait_neutral"
    end

    local idle_anim = idle_anims[mood]
    if idle_anim and (self.anim:GetAnimDuration(idle_anim) or 0) > 0 then
        return idle_anim
    else
        return "idle_neutral"
    end
end

local function GetFaceAnim(mood, action)
    local anim = face_anims[mood] and face_anims[mood][action]
    return anim 
end

function AnimatedCharacter:SetPortraitMode( bool )
    self.portrait = (bool == nil and true) or bool
    return self
end



function AnimatedCharacter:CheckHit(x,y)
    return self.anim:CheckHit(x,y)
end

function AnimatedCharacter:SetVoice(voice)
    self.current_voice = voice
end

function AnimatedCharacter:GetTransitionAnim( mood, to_mood )
    return transition_anims[ mood ][ to_mood ]
end

function AnimatedCharacter:SetMood(mood)
    self.posedef = nil
    if self.mood ~= mood then

        --self:ClearMouthAnim()
        self:ClearEyeAnim()
    
        local transition_anim =  not self.portrait and (transition_anims[self.mood] and transition_anims[self.mood][mood])
        self.mood = mood
        self.facemood = nil

        if transition_anim then
            self:Play(transition_anim)
            self:Push(self:GetIdleAnim(), true)
        else
            self:Play(self:GetIdleAnim(), true)
        end
        
        self.current_voice = nil
    end
    
    return self
end


function AnimatedCharacter:Talk(is_question, voice_delay)
    if self.talk_sound_inst then
        self.talk_sound_inst:Stop()
        self.talk_sound_inst:Release()
        self.talk_sound_inst = nil
    end

    if self.char.voice_event then
        AUDIO:PlayEvent(self.char.voice_event)
    elseif not self.char.voice_actor then
        print ("NO VOICE ACTOR SET")
        return
    end

    local voice = self.current_voice or MOOD_VOICES[self.mood] or "neutralDefault"

    self:QueueVoice(voice, is_question, voice_delay or 0 )

    self.current_voice = nil

end

function AnimatedCharacter:QueueVoice(voice, is_question, dt)
    if self.voice_coro then
        self.voice_coro:Stop()
        self.voice_coro = nil
    end        
    self.voice_coro = self:StartCoroutine(AnimatedCharacter.DelayedVoice, self, voice, is_question, dt)
end

function AnimatedCharacter:DelayedVoice(voice, is_question, dt)

    self:Delay(dt)
    
    local chosen_event, sound_inst = AUDIO:PlaySpeechEvent(self.char.voice_actor, voice, is_question, self.char:GetSpecies())
    self.talk_sound_inst = sound_inst

    local phonemes = chosen_event and AUDIO:GetPhonemes(chosen_event.id)
    if phonemes then
        self:QueueLipsynch(phonemes)
    else
        self:ClearMouthAnim()
    end

end

function AnimatedCharacter:TalkSpecific(event_info)
    if self.talk_sound_inst then
        self.talk_sound_inst:Release()
        self.talk_sound_inst = nil
    end
    

    self.talk_sound_inst = AUDIO:PlayEvent(event_info.event_name, event_info.bank_name) 
    local phonemes = AUDIO:GetPhonemes(event_info.id)
    if phonemes then
        self:QueueLipsynch(phonemes)
    else
        self:ClearMouthAnim()
    end
end

function AnimatedCharacter:SetPreferredMood(mood)
    self.preferred_mood = mood
    return self
end

function AnimatedCharacter:Idle()
    self:KillEmoteEffects()
    self.current_emote = nil
    self:Play(self:GetIdleAnim(), true)
    return self
end

function AnimatedCharacter:QueueLipsynch(phonemes)
    table.clear(self.mouth_queue)
    local now = GetWallTime()
    for k = 1, #phonemes, 2 do
        table.insert( self.mouth_queue,  {shape=phonemes[k+1], time=phonemes[k]+now})
    end

end


function AnimatedCharacter:Emote(emote_name, preserve_queue)
    self:KillEmoteEffects()

    if not preserve_queue then
        table.clear(self.emote_queue)
    end
    self.posedef = nil
    self.posedef_t = nil
    
    local emo = emotes[emote_name]
    if not emo then
        print( "INVALID EMOTE:", emote_name )
        return
    end
    self.emote_hold_time = emo.default_hold_time or 3

    local old_emote = self.current_emote
    self.current_emote = emo
    --self:ClearMouthAnim()
    self:ClearEyeAnim()

    table.clear(self.foley_sounds)
    table.clear(self.emote_fx)
    if emo.foley then
        for k,v in pairs(emo.foley) do
            self.foley_sounds[k] = v
        end
        if self.char.emote_foley_overrides and self.char.emote_foley_overrides[emote_name] then
            for k,v in pairs( self.char.emote_foley_overrides[emote_name]) do
                self.foley_sounds[k] = v
            end
        end

    elseif emo.sound then
        if emo.pre then
            self.foley_sounds[emo.pre] = emo.sound
        end
    end

    if emo.fx then
        self.emote_fx[ emo.pre ] = emo.fx
    end

    local emo_trans_anim = transition_anims[self.mood] and transition_anims[self.mood][emo.mood]
    --print (self.char, emote_name, emo_trans_anim, self.mood, emo.mood)
    
    self.mood = emo.end_mood or emo.mood or self.mood
    --DBG(emo)

    local anims = {}
    
    
    local let_finish = false
    if old_emote and old_emote.pst then
        if self.anim:GetCurrentAnimName() == old_emote.pst then
            let_finish = true
        else
            table.insert(anims, old_emote.pst)
        end
    end

    if emo_trans_anim then
        table.insert(anims, emo_trans_anim)
    end
    
    if emo.pre then
        table.insert(anims, emo.pre)
    end



    
    if emo.hold then
        table.insert(anims, emo.hold)
    else
        table.insert(anims, self:GetIdleAnim(self.mood))
    end
    
    self.facemood = emo.hold_mood
    
    if #anims == 0 then
        if let_finish then
            self:Push(self:GetIdleAnim(), true)
        else
            self:Play(self:GetIdleAnim(), true)
        end
    end
    
    --this isn't right, considering our queue. It might work well enough? Otherwise the anim queue has to be a list of anims AND overrides
    -- ALSO, we should probably clear them up after? Won't be a problem for most use cases, but could get weird.
    if emo.builds then
        for k,v in pairs(emo.builds) do
            self.anim.model:AddOverrideFile(v)
        end
    end

    for k, v in ipairs(anims) do
        
        if k == 1 and not let_finish then
            self:Play(v, k == #anims)
        else
            self:Push(v, k == #anims)
        end
    end

    if emo.voice then
        self.current_voice = emo.voice
    end

    return self
end

function AnimatedCharacter:StopTalking()
    if self.talk_sound_inst then
        self.talk_sound_inst:Stop()
        self.talk_sound_inst:Release()
        self.talk_sound_inst = nil
    end

    if self.voice_coro then
        self.voice_coro:Stop()
        self.voice_coro = nil
    end        


    self:ClearMouthAnim()
end

function AnimatedCharacter:KillEmotes()
    table.clear(self.emote_queue)
end

function AnimatedCharacter:SetColour(colname, colour)
    paperdoll.SetColour(self.anim, colname, colour)
    return self
end

function AnimatedCharacter:Push(anim, loop)
    self.anim:Push(anim, loop)
    return self
end

function AnimatedCharacter:AddAnims(...)
    self.anim:AddAnims(...)
    return self
end

function AnimatedCharacter:Play(anim, loop, offset)
    if anim then
        self.loop = loop
        self.anim:Play(anim, loop, offset or 0)
    end
    return self
end

function AnimatedCharacter:ClearPose()
    self.posedef = nil
    self.posedef_t = nil
    return self
end

function AnimatedCharacter:SetPoseDef(posedef_name)
    self:ClearEyeAnim()
    self:ClearMouthAnim()
    self.current_emote = nil

    if posedefs[posedef_name] then
        self.posedef = posedefs[posedef_name]
        self.posedef_t = nil
        if self.posedef.funny_time then
            self.posedef_t = self.posedef.funny_time[1] + math.random()*(self.posedef.funny_time[2] - self.posedef.funny_time[1])
        end
        local anim = type(self.posedef.anim) == "function" and self.posedef.anim(self.char) or self.posedef.anim
        self:Play(anim, true, math.random())


        if self.posedef.shadow then
            self.shadow:Show()

            if self.posedef.shadow.img then
                self.shadow:SetImage(self.posedef.shadow.img)
            end
            self.shadow:SetPos(self.posedef.shadow.x or 0, self.posedef.shadow.y or 0)
            if self.posedef.shadow.w and self.posedef.shadow.h then
                self.shadow:SetSize(self.posedef.shadow.w, self.posedef.shadow.h)
            end
            
        else
            self.shadow:Hide()
        end

    end
    return self
end


function AnimatedCharacter:OnAgentEvent( event_name, ... )
    if event_name == "build_changed" then
        self:Refresh(self.char)
    end

end


function AnimatedCharacter:OnGlobalEvent( event_name, ... )
    if event_name == "character_skins_changed" then
        if self.char then
            self:Refresh( self.char )
        end
    end
end


function AnimatedCharacter:Refresh( char, fight_mode )
    
    if self.char ~= char then
        
        if self.char then
            self.char:RemoveListener(self)
        end        

        char:ListenForEvent("build_changed", self, AnimatedCharacter.OnAgentEvent)
    end

    self.char = char
    self.anim:Clear()
    self.anim:SetName(char and char.name)
    paperdoll.SetCharacterData(self.anim, char, fight_mode or false, 1)
    

    if char:HasAspect("drunk") then
        self:SetMood( MOOD.DRUNK )
    end

    self:Idle()

    self.time_to_blink = 2+math.random()*3
    return self
end

function AnimatedCharacter:ClearMouthAnim()
    table.clear(self.mouth_queue)
    self.anim:ClearSubAnim("mouth_swap")
end

function AnimatedCharacter:SetPhase( v )
    self.anim:SetPhase( v )
    return self
end

function AnimatedCharacter:SetFacing( facing_right )
    self.anim:SetScale( facing_right and 1 or -1, 1 )
    return self
end

function AnimatedCharacter:PlayMouthAnim(anim, loop, preserve_queue, hold)
    
    if not preserve_queue then
        table.clear(self.mouth_queue)
    end
    self.anim:PlaySubAnim("mouth_swap", anim, loop, 0, hold)
    return self
end

function AnimatedCharacter:ClearEyeAnim()
    self.anim:ClearSubAnim("eye_right_swap")
    self.anim:ClearSubAnim("eye_left_swap")
end

function AnimatedCharacter:PlayEyeAnim(anim, loop)
    self.anim:PlaySubAnim("eye_right_swap", anim, loop)
    self.anim:PlaySubAnim("eye_left_swap", anim, loop)
    return self
end

function AnimatedCharacter:KillEmoteEffects()
    if self.current_emote_foley_sound then
        self.current_emote_foley_sound:Stop()
        self.current_emote_foley_sound:Release()
        self.current_emote_foley_sound = nil
    end
    if self.current_emote_anim then
        if not self.current_emote_anim.removed then
            self.current_emote_anim:Remove()
        end
        self.current_emote_anim = nil
    end
end

function AnimatedCharacter:OnUpdate(dt)
    AnimatedCharacter._base.OnUpdate(self, dt)
    
    if self.talk_sound_inst then
        self.talk_sound_inst:SetParameterValue("x_position", self:GetPanPos()*100)
    end


    --certain moods are 'sticky'
    if self.mood == MOOD.INJURED or self.mood == MOOD.DRUNK then
        self.preferred_mood = self.mood
    end


    if self.removed or not self.anim.model then
        return
        --assert(nil)
    end
    
    local time = GetWallTime()

    if #self.emote_queue > 0 then
        if self.emote_queue[#self.emote_queue].time < time then
            self:Emote(self.emote_queue[#self.emote_queue].emote, true)
            table.remove(self.emote_queue)
        end
    end

    if #self.mouth_queue > 0 then
        
        if self.mouth_queue[1].time < time then
            local shp = self.mouth_queue[1].shape
            

            if shp then

                local face_mood = self.current_emote and self.current_emote.face_mood or self.mood
                local anim = MOUTH_SHAPE[shp] and MOUTH_SHAPE[shp][(face_mood == MOOD.HAPPY or face_mood == MOOD.DRUNK) and MOOD.HAPPY or MOOD.NEUTRAL]
                
                --print (self.mouth_queue[1].time, shp, anim)    
                
                if anim then
                    self:PlayMouthAnim(anim, false, true, #self.mouth_queue ~= 1 )
                else
                    self.anim:ClearSubAnim("mouth_swap")
                end
            end
            table.remove(self.mouth_queue, 1)
        end
    end
    
    --play the funny idles!
    if self.posedef_t then
        self.posedef_t = self.posedef_t - dt
        if self.posedef_t <= 0 and self.posedef.funny then
            self.posedef_t = self.posedef.funny_time[1] + math.random()*(self.posedef.funny_time[2] - self.posedef.funny_time[1])
            
            local total = 0
            local idx = 1
            
            for k,v in pairs(self.posedef.funny) do
                total = total + v.weight or 1
            end

            local rng = math.random()*total
            for k,v in pairs(self.posedef.funny) do
                rng = rng - (v.weight or 1)
                if rng <= 0 then
                    idx = k
                    break
                end
            end
            
            local queue = {}
            if self.posedef.funny[idx] then
                for k,v in ipairs(self.posedef.funny[idx].anims) do
                    local num = 1
                    if v.loops then
                        num = math.random(v.loops[1], v.loops[2])
                    end
                    
                    for k=1,num do
                        table.insert(queue, v.anim)
                    end
                end
            end
            for k,v in ipairs(queue) do
                if k ==1 then
                    self:Play(v)
                else
                    self:Push(v)
                end
            end
            self:Push(self.posedef.anim, true)
        end
    end

    local can_anim_eyes = true
    
    local current_anim = self.anim:GetCurrentAnimName()
    if current_anim ~= self.last_seen_anim then
        local fx = self.emote_fx[current_anim]
        if self.foley_sounds[current_anim] or fx then
            self:KillEmoteEffects()
        end

        if self.foley_sounds[current_anim] then
            --print ( self.foley_sounds[current_anim] )
            self.current_emote_foley_sound =  AUDIO:CreateEventInstance(self.foley_sounds[current_anim])
            if self.current_emote_foley_sound then
                self.current_emote_foley_sound:Start()
            end
        end
        if fx then
            local anim = UIHelpers.PlayAnimOnce( self, fx.animfile, fx.anim )
            if anim and fx.scale then
                anim:SetScale( fx.scale )
            end
            if anim and fx.bloom then
                anim:SetBloom( fx.bloom )
            end
            self.current_emote_anim = anim
            self.emote_fx[current_anim] = nil
        end
    end
    self.last_seen_anim = current_anim


    if self.current_emote then
        if self.current_emote.noeyes then
            can_anim_eyes = false
        end


        if self.current_emote.pst and self.current_emote.pst == self.anim:GetCurrentAnimName() and self.anim:IsDone() then
            self.facemood = nil
            self.current_emote = nil
            --self:KillEmoteEffects()
            self:Play(self:GetIdleAnim(), true)
        elseif self.current_emote.hold and self.current_emote.hold == self.anim:GetCurrentAnimName() then
            self.emote_hold_time = self.emote_hold_time - dt
            
            if self.emote_hold_time <= 0 then
                local idle = self:GetIdleAnim()
                if self.current_emote.pst then
                    self:Play(self.current_emote.pst)
                else
                    self.facemood = nil
                    self.current_emote = nil
                    self:KillEmoteEffects()
                    self:Play(idle, true)
                end
            end
        end

    end

    if self.current_emote == nil and #self.emote_queue == 0 and self.preferred_mood and self.mood ~= self.preferred_mood then
        self:SetMood(self.preferred_mood)
    end

    
    if self.mood ~= MOOD.INJURED and self.mood ~= MOOD.DRUNK then
        if self.time_to_blink and not self.noblink then
            self.time_to_blink = self.time_to_blink - dt
            if self.time_to_blink <= 0 then
                if can_anim_eyes then 
                    
                    if self.posedef then
                        if self.posedef.blink then
                            self:PlayEyeAnim(self.posedef.blink)
                        end
                    else
                        self:PlayEyeAnim(GetFaceAnim(self.facemood or self.mood, "blink"))
                    end
                end
                self.time_to_blink = 2+math.random()*3
            end
        end
    end


    if self.current_emote_foley_sound then
        self.current_emote_foley_sound:SetParameterValue("x_position", self:GetPanPos()*100)
    end


end

function AnimatedCharacter:IsEmoting()
    return not self.anim:IsDone()
end

function AnimatedCharacter:DisableBlink()
    self.noblink = true
    return self
end

function AnimatedCharacter:SetSpeed(v)
    self.anim.model:SetPlaySpeed(v)
    return self
end

-- You can turn this to false if it's not windy, on interior scenes
function AnimatedCharacter:SetAutoLoopEnabled(enabled)
    self.anim.model:SetAutoLoopEnabled(enabled)
    return self
end

function AnimatedCharacter:SetClickFn(fn)
    self.fn = fn
    self.isbutton = true
    self.hover_check = true
    self.blocks_mouse = true
    return self
end

function AnimatedCharacter:SetSelected( enabled, fade_duration )
    self.selected = enabled 
    self.anim:SetSelectedLayerSpeeds(0.025, 0.04, 0.4)
    self.anim:SetSelectedLayerScale(1.5)
    self.anim:SetSelectedLayerAlphas(0.1, 0.1, 0.7)
    self.anim:SetSelectedFadeDuration( fade_duration or 0.2 )
    self.anim:SetSelected(enabled)   
end

function AnimatedCharacter:SetSelectedColor( colour ) 
    self.anim:SetSelectedColor( colour )
end

function AnimatedCharacter:OnGainFocus()
    self:UpdateButtonState()
end 

function AnimatedCharacter:OnGainHover()
    self:UpdateButtonState()
    if self.onhoverfn then
        self.onhoverfn(true)
    end
end

function AnimatedCharacter:OnLoseHover()
    self.down = false
    self:UpdateButtonState()
    if self.onhoverfn then
        self.onhoverfn(false)
    end
end

function AnimatedCharacter:SetOnHoverFn(fn)
    self.hover_check = true
    self.onhoverfn = fn
    return self
end

function AnimatedCharacter:OnLoseFocus()
    self.down = false
    self:UpdateButtonState()
end

function AnimatedCharacter:UpdateButtonState()
    if self.isbutton then
        self:SetSaturation(1)
        
        if not self.enabled then
            self:SetSaturation(0)
        elseif self.down or self.hover or self.selected then
            self.anim:SetOutline( self:GetOutlineSize(), self.hover_outline_colour or UICOLOURS.LIGHT_PANEL_BG )
        else
            self.anim:HideOutline()
        end
    end
end

function AnimatedCharacter:GetOutlineSize()
    local scale = self:GetScale()
    if scale == 1 then scale = self.parent:GetScale() or 1 end
    scale = math.abs( scale )

    -- Calculate an outline size inversely proportional to the scale.
    -- Larger scale, smaller outline
    local outline = remap( scale, 0.2, 0.6, 11, 3 )
    --print( "scale", scale, outline )

    return outline
end

function AnimatedCharacter:SetHoverOutlineColour( colour )
    self.hover_outline_colour = colour or UICOLOURS.LIGHT_PANEL_BG
    return self
end

function AnimatedCharacter:SetOutline( size, colour )
    self.anim:SetOutline( size, colour )
    return self
end

function AnimatedCharacter:ShowOutline()
    self.anim:SetOutline( self:GetOutlineSize(), self.hover_outline_colour or UICOLOURS.LIGHT_PANEL_BG )
    return self
end

function AnimatedCharacter:HideOutline()
    self.anim:HideOutline()
    return self
end

function AnimatedCharacter:HandleControlDown(control)
    if not self.isbutton then
        return false
    end

    if control:Has( Controls.Digital.MENU_ACCEPT ) and self.hover then
        if self.enabled then
            self.down = true
            self:UpdateButtonState()
            return true
        end
    end
end

function AnimatedCharacter:Click()
    if self.fn then self.fn() end
end

function AnimatedCharacter:HandleControlUp(control, dev)
    if not self.isbutton then
        return false
    end

    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if self.down then
            self.down = false
            self:UpdateButtonState()

            if self.hover or self.focus then
                if self.enabled then
                    AUDIO:PlayEvent(SoundEvents.button_click)
                    if self.fn then self.fn(dev) end
                end
                return true
            end
        end
    end
end

function AnimatedCharacter:__tostring()
    return loc.format( "ANIMATED CHARACTER({1})", self.char and self.char:GetName() or "??")
end


function AnimatedCharacter:SetXRayOutline(fillColour, lineColour, thresholdMin, thresholdMax)
    self.anim:SetXRayOutline(fillColour, lineColour, thresholdMin, thresholdMax)
end


function AnimatedCharacter:SetBloom(b)
    self.anim:SetBloom(b)
end
