
-------------------------------------------------------------------------------
-- Widget containing Negotiation.Modifier widgets
-- 
-- Shows an animated wheel with slots for modifier widgets

local ArgumentContainer = class( "Widget.ArgumentContainer", Widget )

local TEXTURE_RADIUS_PADDING = 30
local MAX_ROTATING_SPEED = 0.1

function ArgumentContainer:init( size )
    ArgumentContainer._base.init( self )

    -- Default values
    self.left = true
    self.wheel_angle = 0
    self.layout_slots = true
    self.was_animating = false

    self.rotating_mode = false
	self.rotating_speed = MAX_ROTATING_SPEED
	self.target_speed = MAX_ROTATING_SPEED
    self.rotating_angle = 0
	self.slot_widgets = {}
	self.widget_size = size or 550
	self.radius = ( self.widget_size - TEXTURE_RADIUS_PADDING )/2

	-- Container for widgets with hover checking reasons
	self.contents = Widget()
	self.hitbox = self.contents:AddChild( Widget.SolidBox( self.widget_size, self.widget_size, 0xff000000 ) )

    -- Prepare background wheel
    self.wheel = Widget()
        :SetHiddenBoundingBox( true )
    self.wheel_mask_container = Widget() -- Contains all the mask widgets
    self.wheel_mask_bg = Widget.Image( engine.asset.Texture("images/white.tex"), self.widget_size, self.widget_size )
    self.wheel_gradient_container = Widget() -- Contains all the mask overlays
    self.wheel_container = Widget()
    self.wheel1 = Widget.Image( engine.asset.Texture("UI/arguments_wheel.tex"), self.widget_size, self.widget_size )
        :SetTintColour(UICOLOURS.SUBTITLE)
        :SetTintAlpha(0.7)
        :Bloom( 0.3 )
        :RotateIndefinitely( 0.02 )
        :SetHiddenBoundingBox( true )
    self.wheel2 = Widget.Image( engine.asset.Texture("UI/arguments_wheel.tex"), self.widget_size, self.widget_size )
        :SetTintColour(UICOLOURS.SUBTITLE)
        :SetTintAlpha(0.5)
        :Bloom( 0.2 )
        :RotateIndefinitely( -0.035 )
        :SetHiddenBoundingBox( true )
    self.wheel_pointer = Widget.Image( engine.asset.Texture( "UI/arguments_direction.tex" ), self.widget_size*1.3, self.widget_size*1.3 )
        :SetTintColour(UICOLOURS.SUBTITLE)
        :SetTintAlpha(0)
        :Bloom( 0.3 )
        :SetHiddenBoundingBox( true )

    -- The hit animation
    self.hit_anim = Widget.Anim()
    local animname = "anim/UI_negotiation_hit_halfcircle.zip"
    self.hit_anim:AddAnims( GetAnim( animname ) )
    self.hit_anim:AddOverrideFile( animname )
    self.hit_anim:SetSpeed( 1.0 )
    self.hit_anim
        :SetHiddenBoundingBox(true)
        :SetBloom( 0.5 )
        :SetTintColour( UICOLOURS.NEGOTIATION )
        :SetTintAlpha( 0.9 )
        :SetScale( self.left and -1.05 or 1.05 )

    -- Setup hierarchy
    self:AddChildren
    {
        self.wheel:SetStencilContext( true ):AddChildren
        {
            self.wheel_mask_container:AddChildren
            {
                self.wheel_mask_bg:SetMask(),
            },
            self.wheel_container:AddChildren 
            {
                self.wheel1:SetStencilTest( true ),
                self.wheel2:SetStencilTest( true ),
            },
            self.wheel_gradient_container,
        },
        self.wheel_pointer,
        self.hit_anim:SetPos( self.left and self.radius+4 or -self.radius-4, 0 ),
        self.contents
    }

    -- self.DBG_center = self:AddChild( Widget.Label( "body", 32 ) )

    -- Contains the masking/overlay widgets necessary to make the "cuts" in the wheel
    self.slot_container_widgets = {}

	self:StartUpdating()
end

function ArgumentContainer:SetSize( diameter )
    self.widget_size = diameter or 550
    self.radius = ( self.widget_size - TEXTURE_RADIUS_PADDING )/2
    self.hitbox:SetSize( self.widget_size, self.widget_size )
    self.wheel_mask_bg:SetSize( self.widget_size, self.widget_size )
    self.wheel1:SetSize( self.widget_size, self.widget_size )
    self.wheel2:SetSize( self.widget_size, self.widget_size )
    self.wheel_pointer:SetSize( self.widget_size*1.3, self.widget_size*1.3 )
    self.hit_anim:SetPos( self.left and self.radius+4 or -self.radius-4, 0 )
    for k, v in ipairs( self.wheel_gradient_container.children ) do
        v:SetSize( self.widget_size, self.widget_size )
    end
    self:Layout()
end

function ArgumentContainer:OnScreenModeChange( sm )
    ArgumentContainer._base.OnScreenModeChange( self, sm )

    if self.flourish_bar then self.flourish_bar:SetLayoutScale( LAYOUT_SCALE[sm] ) end
end

function ArgumentContainer:GetDefaultFocus()
    for k, v in ipairs(self.slot_widgets) do
        if v.modifier and v.modifier.modifier_type == MODIFIER_TYPE.CORE then
            return v
        end
    end
    return self.slot_widgets and #self.slot_widgets > 0 and self.slot_widgets[1]
end

function ArgumentContainer:PlayAnim( anim_name )
    self.hit_anim:Play( anim_name, false )
    return self
end

-- If true, the first argument goes on the left, if false, it goes on the right
function ArgumentContainer:SetStartingSide( start_left )
    self.left = start_left

    self.hit_anim
        :SetScale( self.left and -1.05 or 1.05 )
        :SetPos( self.left and self.radius+4 or -self.radius-4, 0 )
    return self
end

function ArgumentContainer:SetWheelColour( tint_colour )
	self.wheel:SetTintColour( tint_colour )
	return self
end

function ArgumentContainer:SetIntentsMode()
	self.wheel1:SetTexture( engine.asset.Texture("UI/arguments_wheel_intents.tex") )
	self.wheel1:RotateIndefinitely( 0.01 )
	self.wheel2:Hide()
    self:SetRotatingMode( true )
	return self
end

function ArgumentContainer:SetRotatingMode( on )
	self.rotating_mode = (on == true)
	self.rotating_speed = MAX_ROTATING_SPEED
	return self
end

function ArgumentContainer:SetAgentName( name )
    self.name_label = self:AddChild( Widget.Label( "title", 40, name ) )
        :SetHiddenBoundingBox( true )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
        :EnableOutline( SMALL_TEXT_OUTLINE )
        :LayoutBounds( "center", "below", self.contents )
        :Offset( 0, -20 )
        :SetTintAlpha( 0 )
    return self
end

function ArgumentContainer:SetLayoutSlots( on )
    self.layout_slots = (on == true)
end

function ArgumentContainer:GetSlotCount()
    return #self.slot_widgets
end

function ArgumentContainer:AddSlotWidget( widget )
    local slot_idx = #self.slot_widgets + 1
    
    -- Set angle values
    widget.current_angle = self.left and 0 or math.pi
    widget.target_angle = widget.current_angle

    widget = self.contents:AddChild( widget )
    table.insert( self.slot_widgets, widget )

    return widget
end

function ArgumentContainer:SetWheelAngle( slot, angle )
    slot.target_angle = angle
end

function ArgumentContainer:SnapWheelAngle( slot )
    slot.current_angle = slot.target_angle
    local x, y = math.cos( slot.current_angle ) * self.radius, math.sin( slot.current_angle ) * self.radius
    slot:SetPos( x, y )
end

function ArgumentContainer:RemoveModifierSlot( slot_widget )
    slot_widget:Remove()
    table.arrayremove( self.slot_widgets, slot_widget )
end

function ArgumentContainer:GetWidgets( slot_widget )
    return self.slot_widgets
end

function ArgumentContainer:UpdateWheelSlots()

    -- Check if we have the right number of mask/overlay widgets for the slots we're displaying
    local count = self:GetSlotCount()
    if count > #self.slot_container_widgets then
        -- We need to add more widgets
        while count > #self.slot_container_widgets do
            local mask = Widget.Image( engine.asset.Texture("UI/arguments_mask.tex"), self.widget_size, self.widget_size ):SetMaskClear()
            local gradient = Widget.Image( engine.asset.Texture("UI/arguments_gradient.tex"), self.widget_size, self.widget_size )
                :SetTintColour( UICOLOURS.SUBTITLE )
            table.insert( self.slot_container_widgets, {
                mask = mask,
                gradient = gradient
            })
            self.wheel_mask_container:AddChild( mask )
            self.wheel_gradient_container:AddChild( gradient )
        end
    elseif count < #self.slot_container_widgets then
        -- We need to remove widgets
        while count < #self.slot_container_widgets do
            local data = table.remove( self.slot_container_widgets )
            data.mask:Remove()
            data.gradient:Remove()
        end
    end

    -- Position our masks/overlays correctly
    local angle = 360/count
    for k, data in ipairs(self.slot_container_widgets) do
        data.mask:SetRotation( angle*DEGREES*(k-1) )
        data.gradient:SetRotation( angle*DEGREES*(k-1) )
    end

    self.wheel_angle = -math.pi/2 * (self.left and 1 or -1)
    self.wheel:SetRotation( self.wheel_angle )
end

function ArgumentContainer:Layout()
    local slot_count = self:GetSlotCount()
    local angle_per_slot = 360 / slot_count
 
    local prev_widget
    for i, widget in ipairs( self.slot_widgets ) do
        local slot = i - 1
        local angle = (angle_per_slot * slot + ( self.left and 0 or 180 ) ) * DEGREES
        self:SetWheelAngle( widget, angle )
    end

    self:UpdateWheelSlots()
end

function ArgumentContainer:OnUpdate( dt )
    ArgumentContainer._base.OnUpdate( self, dt )

    local t = self.elapsed_time or 0
    self.elapsed_time = t + dt

    if self.rotating_mode then
    	-- If the widget gained or lost hover, we should adjust the speed accordingly
    	if self.rotating_speed ~= self.target_speed then
    		self.rotating_speed = Lerp( self.rotating_speed, self.target_speed, 0.1 )
    	end

    	-- Update the rotation for the widgets and wheel
    	self.rotating_angle = self.rotating_angle + self.rotating_speed*dt
    end

    if self.layout_slots then
        local animating = false
        for i, widget in ipairs( self.slot_widgets ) do
        	local angle = widget.current_angle
    	    if widget.current_angle ~= widget.target_angle and widget.target_angle ~= nil then
    	        local DT_MAX = 0.5
    	        widget.angle_dt = (widget.angle_dt or 0) + dt
    	        if widget.angle_dt >= DT_MAX then
    	            widget.angle_dt = 0
    	            widget.current_angle = widget.target_angle
    	            angle = widget.current_angle
    	        else
    	            angle = easing.inQuad( widget.angle_dt, widget.current_angle, widget.target_angle - widget.current_angle, DT_MAX )
    	        end
                animating = true
    	    end
            local x = math.cos( angle + self.rotating_angle ) * self.radius
            local y = math.sin( angle + self.rotating_angle ) * self.radius
            widget:SetPos( x, y )
    	end
        if animating ~= self.was_animating then
            self.was_animating = animating
            self:SortSlots()
        end
    end

	self.wheel:SetRotation( self.wheel_angle + self.rotating_angle )

    self:UpdateRadialCursor()

end

function ArgumentContainer:UpdateRadialCursor()
    if not self.focus then return end
    if self.fe:ReticuleActive() then return end

    local lx,ly,rx,ry = TheGame:GetInput():GetActiveGamepadStickVals()
    local length = math.sqrt(lx*lx+ly*ly)
    if length > 0.1 then
        -- Rotate the wheel based on the vector angle
        local angle = math.atan2(ly,lx)
        angle = angle>=0 and angle or 2*math.pi+angle
        angle = angle - math.pi/2
        self.wheel_pointer:SetRotation( angle )

        -- Fade the wheel in based on the vector length
        self.wheel_pointer_faded_in = true
        self.wheel_pointer:SetTintAlpha( length )

        -- Give focus to the closest argument, if close enough
        local argument = self:GetClosestToAngle( angle )
        if argument then argument:SetFocus() end
    else
        -- Fade the wheel out if it was shown
        if self.wheel_pointer_faded_in then
            self.wheel_pointer_faded_in = false
            self.wheel_pointer:AlphaTo( 0, 0.1 )
        end
    end

end

function ArgumentContainer:GetClosestToAngle( selector_angle )

    local closest_widget, closest_angle_dist

    for i, widget in ipairs( self.slot_widgets ) do
        -- Get widget's angle
        local widget_angle = self:GetArgumentAngle( widget )

        -- Get angle distance
        local angle_dist = math.abs( angledist( selector_angle, widget_angle ) )

        -- If this is the new smallest distance, save it
        if not closest_widget or angle_dist < closest_angle_dist then
            closest_widget = widget
            closest_angle_dist = angle_dist
        end
                    -- widget.composure_widget:Show()
                    --     :SetTintAlpha( 1 )
                    -- widget.composure_label:SetText( loc.format( "angle\n{1%.2f}\ndistance\n{2%.2f}", widget_angle*(180/math.pi), angle_dist*(180/math.pi) ) )
                    --     :SetGlyphColour( UICOLOURS.QUEST )
    end

    -- Make one third of the distance between arguments be a dead zone
    local threshold = (2*math.pi)/#self.slot_widgets/3 

                -- self.DBG_center:SetText( loc.format( "selector angle\n{1%.2f}\nthreshold\n{2%.2f}",selector_angle*(180/math.pi), threshold*(180/math.pi) ) )
                --     :EnableOutline( 0.2 )
                --     :SetGlyphColour( UICOLOURS.QUEST )
                --     :Bloom( 0.1 )

    if closest_angle_dist and closest_angle_dist < threshold then
        return closest_widget
    end
end

function ArgumentContainer:GetArgumentAngle( argument_widget )
    local pos_x, pos_y = argument_widget:GetPos()

    local angle = math.atan2( pos_y, pos_x )
    angle = angle>=0 and angle or 2*math.pi+angle
    angle = angle - math.pi/2

    return angle
end

function ArgumentContainer:DestroyAllSlots()
    while #self.slot_widgets > 0 do
        local w = table.remove( self.slot_widgets )
        AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/ui_component_fades" )
        w:ScaleTo( 1.0, 2.0, 0.3, easing.outCubic )
        w:AlphaTo( 0, 0.3, nil, function() w:Hide() end )
        w:Delay( 0.1 )
    end

    self.wheel:AlphaTo( 0, 0.5 )
end


function ArgumentContainer:SortSlots()
    local core = self.slot_widgets[1]
    local cores = 0
    for k,slot in pairs(self.slot_widgets) do
        if slot.modifier and slot.modifier.modifier_type == MODIFIER_TYPE.CORE then
            cores = cores + 1
            table.arrayremove(self.slot_widgets, slot)
            table.insert(self.slot_widgets, cores, slot )
        end
    end

    local slot_widgets = shallowcopy( self.slot_widgets )
    table.sort( slot_widgets, function( w1, w2 ) return w1.y < w2.y end )

    if #slot_widgets > 0 then
        slot_widgets[ 1 ]:SendToFront()
        for i = 2, #slot_widgets do
            local prev = slot_widgets[ i - 1 ]
            local this = slot_widgets[ i ] 
            this:SendBehind( prev )
        end
    end
end

-- Overrides set SetToolTip, so that the tooltip is set to the wheel itself
-- And then we use the wheel radius as a hit check, instead of the regular bounding box
function ArgumentContainer:SetToolTip( tt )
    self.wheel:SetToolTip( tt )
    self.wheel.parent = self

    self.wheel.CheckHit = function( self, x, y )
        local lx, ly = self:TransformFromWorld(x,y)
        local x1,y1,x2,y2 = self:GetBoundingBox()

        if x1 and y1 and x2 and y2 then
            local center_x, center_y = x2/2 + x1, y2/2 + y1
            local dist = distance( 0, 0, lx, ly )
            if dist > self.parent.radius*0.90 and dist < self.parent.radius*1.2 then
                self:AlphaTo( 2.5, 0.1 )
                return true
            end
        end

        self:AlphaTo( 1, 0.2 )
        return false
    end
end

function ArgumentContainer:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.name_label then
            self.name_label:AlphaTo( 1, 0.1, easing.OutQuad )
        end
        -- if self.focus then
        --     self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
        -- end
    else
        if self.name_label then
            self.name_label:AlphaTo( 0, 0.1, easing.OutQuad )
        end
        -- self:ScaleTo( nil, 1, 0.2, easing.outQuad )

        self.wheel_pointer_faded_in = false
        self.wheel_pointer:AlphaTo( 0, 0.1 )
    end
end

function ArgumentContainer:OnGainFocus()
    -- Disable the conversion of left stick to digital directions so that the radial selection works
    TheGame:GetInput():SetLeftJoystickDigitalForwarding( false )

    ArgumentContainer._base.OnGainFocus(self)
    self:UpdateImage()
end

function ArgumentContainer:OnLoseFocus()
    -- Re-enable the conversion of left stick to digital directions so that the radial selection works
    TheGame:GetInput():SetLeftJoystickDigitalForwarding( true )

    ArgumentContainer._base.OnLoseFocus(self)
    self:UpdateImage()
end

function ArgumentContainer:OnGainHover()
    ArgumentContainer._base.OnGainHover(self)
    self:UpdateImage()
end

function ArgumentContainer:OnLoseHover()
    ArgumentContainer._base.OnLoseHover(self)
    self:UpdateImage()
end