require "ui/widgets/cardwidget"
local battle_defs = require "battle/battle_defs"

local CARD_FLAGS = battle_defs.CARD_FLAGS
local EVENT = battle_defs.BATTLE_EVENT

local assets =
{
    type_icons =
    {
        [ CARD_TYPE.FIGHT_MELEE ] = engine.asset.Texture( "UI/cardview_typeicon_combat_attack.tex" ),
        [ CARD_TYPE.FIGHT_SKILL ] = engine.asset.Texture( "UI/cardview_typeicon_combat_maneuver.tex" ),
        [ CARD_TYPE.FIGHT_STATUS ] = engine.asset.Texture( "UI/cardview_typeicon_combat_status.tex" ),
        [ CARD_TYPE.FIGHT_SPECIAL ] = engine.asset.Texture( "UI/cardview_typeicon_combat_status.tex" ),
        [ CARD_TYPE.ITEM ] = engine.asset.Texture( "UI/cardview_typeicon_item.tex" ),
        [ CARD_TYPE.PARASITE ] = engine.asset.Texture( "UI/cardview_typeicon_parasite.tex" ),
        [ CARD_TYPE.FLOURISH ] = engine.asset.Texture( "UI/cardview_typeicon_flourish.tex" ),
        [ CARD_TYPE.SCORE ] = engine.asset.Texture( "UI/cardview_typeicon_score.tex" ),
    },

    pip_open = engine.asset.Texture( "UI/ic_charges_empty.tex" ),
    pip_closed = engine.asset.Texture( "UI/ic_charges_full.tex" ),
}

------------------------------------------------------------------------------------------

local CardWidget = class( "Widget.BattleCard", Widget.CardWidget )

function CardWidget:init( card, battle )
    CardWidget._base.init( self, assets, battle_defs )

    self.shadow:SetTintColour( UICOLOURS.FIGHT )
    self.shadow:SetTintAlpha(0.3)

    if card then
        self:RefreshCard( card, battle )
    end
end

function CardWidget:RefreshCard( card, battle, style )
    assert_warning( is_instance( card, Battle.Card ), card.id )

    battle = battle or card.engine

    CardWidget._base.RefreshCard( self, card, battle, style )
    self.card = card

    local card_back_tint, type_icon, text_colour, tt = self:RefreshCardType( card, battle )

    if not (type_icon and text_colour and tt) then
        return
    end

    self.bg:SetTintColour( card_back_tint )

    self.type_icon:SetTexture( type_icon )
        :SetTintColour( text_colour )

    return self
end

function CardWidget:RefreshCardType( card, battle )
    local card_type = GetCardType( card )

    return
        GetCardTypeColour( card_type ),
        assets.type_icons[ card_type ],
        GetCardTypeColour( card_type ),
        GetCardTypeString( card_type )
end

function CardWidget:RefreshCostLabel( card, battle )
    local cost, details
    if battle then
        cost, details = battle:CalculateActionCost( card, card:GetTarget() )
    else
        cost = card.cost
    end

    local cost_txt
    if CheckAnyBits( card.flags, CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.AUTOPLAY ) then
        cost_txt = "-"
    elseif CheckBits( card.flags, CARD_FLAGS.VARIABLE_COST ) then
        cost_txt = LOC("X")
    else
        cost_txt = tostring(cost)

        if battle and cost > battle:GetActionCount() then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetDisabledColour( card ), cost_txt )
        elseif self.style == self.STYLE.UPGRADE then
            if cost < card.base_def.cost or CheckBits( card.base_def.flags, CARD_FLAGS.UNPLAYABLE ) then
                cost_txt = loc.format( "<#UPGRADE>{1}</>", cost_txt )
            elseif cost > card.base_def.cost then
                cost_txt = loc.format( "<#DOWNGRADE>{1}</>", cost_txt )
            end
        elseif cost < card.cost then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), cost_txt )
        elseif cost > card.cost then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), cost_txt )
        else
            cost_txt = tostring(cost)
        end
    end

    if details then
        details = LOC"UI.FIGHT_SCREEN.COST_TT" ..
            loc.format( LOC"UI.FIGHT_SCREEN.COST_TT_MODIFIED", details )
    else
        details = LOC"UI.FIGHT_SCREEN.COST_TT"
    end


    return cost_txt, details
end

function CardWidget:RefreshEffectLabel( card, battle )
    if card.min_damage == nil or card.max_damage == nil then
        return
    end

    local min_damage, max_damage, details, acc
    if battle then
        local target
        local current_mod = battle.target_mod_accumulator:CalculateValue( card.target_mod, card, card:GetTarget())
        if current_mod == TARGET_MOD.SINGLE then
            target = card:GetTarget()
        end
        min_damage, max_damage, details, acc = battle:PreviewDamage( card, target )
        if card.threshold then
           battle :CalculateThreshold(card)
        end
    else
        min_damage, max_damage, details = card.min_damage, card.max_damage
    end
    local min_txt

    if self.style == self.STYLE.UPGRADE and min_damage ~= card.base_def.min_damage then
        if card.base_def.min_damage and min_damage < card.base_def.min_damage then
            min_txt = loc.format( "<#DOWNGRADE>{1}</>", min_damage )
        else
            min_txt = loc.format( "<#UPGRADE>{1}</>", min_damage )
        end
    elseif min_damage < card.min_damage then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), min_damage )
    elseif min_damage > card.min_damage then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), min_damage )
    elseif min_damage == card.min_damage and acc and acc:IsMinDamagePenalized() then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), min_damage )
    else
        min_txt = tostring(min_damage)
    end

    local max_txt
    if self.style == self.STYLE.UPGRADE and max_damage ~= card.base_def.max_damage then
        if card.base_def.max_damage and max_damage < card.base_def.max_damage then
            max_txt = loc.format( "<#DOWNGRADE>{1}</>", max_damage )
        else
            max_txt = loc.format( "<#UPGRADE>{1}</>", max_damage )
        end
    elseif max_damage < card.max_damage then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), max_damage )
    elseif max_damage > card.max_damage then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), max_damage )
    elseif max_damage == card.max_damage and acc and acc:IsMaxDamagePenalized() then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), max_damage )
    else
        max_txt = tostring(max_damage)
    end

    local txt
    if min_txt == max_txt then
        txt = min_txt
    else
        txt = loc.format( "{1}-{2}", min_txt, max_txt )
    end
    local tt = LOC"UI.FIGHT_SCREEN.DAMAGE_INFLICTED"
    if details then
        tt = tt .. loc.format( LOC"UI.FIGHT_SCREEN.MODIFIERS", details )
    end

    return txt, tt, min_damage, max_damage
end

function CardWidget:AnimatePreResolve( screen, idx )
    self.anim = nil
    self:DisableToolTip( true )

    self:Enable() -- Shouldn't appear disabled once resolved.
    self:SetTintAlpha( 1.0 )

    if self.card.played_sound then
        AUDIO:PlayEvent( self.card.played_sound )
    end

    AUDIO:PlayEvent(SoundEvents.negotiation_cardplayed_player)

    if self.card:IsFlagged( CARD_FLAGS.AUTOPLAY ) then
        -- Animate card down to the centre of the screen
        local w, h = self:GetFE():GetScreenDims()
        local ypos = h/2.2
        local xpos = w/2
        local x, y = self:GetPos()
        local tx, ty = self.parent:TransformFromWorld( xpos, ypos )
        local CARD_DROP_TIMING = 0.1
        local CARD_DROP_EASING = easing.outQuad

        self:ScaleTo( self.sx, 0.6, CARD_DROP_TIMING, CARD_DROP_EASING )
        self:MoveTo( tx, ty, CARD_DROP_TIMING, CARD_DROP_EASING )
        self:RotateTo( 0, CARD_DROP_TIMING, CARD_DROP_EASING )

        self:Delay( CARD_DROP_TIMING*2 )

    elseif self.card:IsFlaggedAny( CARD_FLAGS.FLOURISH ) then
        self:ShowErosion( "expend_flourish" )

    elseif self.card.deck and self.card.deck:GetDeckType() ~= DECK_TYPE.RESOLVE then
        -- Send the target immediately to its final destination, even before we received any CARD_MOVED event.
        screen:UpdateCardPlacement( self.card )
    end
end

function CardWidget:PlaySelectSound( card )
    AUDIO:PlayEvent(SoundEvents.battle_card_select)
end

function CardWidget:PlayDeselectSound( card )
    AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
end
