require "ui/widget"


local CaptureImage = class("Widget.CaptureImage", Widget)

function CaptureImage:init(scene, name)
    CaptureImage._base.init(self)

    self.scene = scene or TheGame:FE().scene
    self.scene_render_node = engine.rendering.SceneRenderNode(name or "Capture")

    self.camera = engine.rendering.CameraInfo()
    self.camera:SetOrtho()
    self.camera:SetDir( 0, 0, -1 )

    self.img = Widget.Image(global_images.white):SetEffect(global_shaders.UI_BLOOMMAP)
    self.shader_dirty = true    
    if engine.inst:IsScreenVerticallyFlipped() then
        self.img:SetScale(1,-1)
    end

    self.img:SetBloom( 1.0 ) -- with BLOOMMAP, this is a bloom scaling factor.
    self:AddChild( self.img )
    
    self.colour_correct_constants = 
    {
        shadows = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST", 0,1,1),
        midtones = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST", 0,1,1),
        highlights = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST", 0,1,1),
        shadows_col = engine.rendering.ShaderConstantValue("SHADOWS_ADJUST_COL", 1,1,1),
        midtones_col = engine.rendering.ShaderConstantValue("MIDTONES_ADJUST_COL", 1,1,1),
        highlights_col = engine.rendering.ShaderConstantValue("HIGHLIGHTS_ADJUST_COL", 1,1,1),
        vignette = engine.rendering.ShaderConstantValue("VIG_STRENGTH", 0),
    }
    for k,v in pairs(self.colour_correct_constants) do
        self.img.model:SetShaderConstant(v)
    end

    local w, h = TheGame:FE():GetScreenDims()
    self:SetSize(w,h, w,h)

end

function CaptureImage:SetBloom( bloom )
    self.img:SetBloom( bloom )
end

function CaptureImage:SetBlur( blur )
    self.blur_amt = blur
end

function CaptureImage:SetSize(sw, sh, texw, texh)
    self.pending_sw = sw
    self.pending_sh = sh
    self.pending_texw = texw or sw
    self.pending_texh = texh or sh
    return self
end

function CaptureImage:GetTexture()
    return self.img:GetTexture()
end

function CaptureImage:UpdateRenderTarget()

    if self.pending_texw ~= self.texw or self.pending_texh ~= self.texh or self.bloom_enabled ~= TheGame.bloom_enabled then
        self.texw = self.pending_texw or self.texw
        self.texh = self.pending_texh or self.texh
        self.sw = self.pending_sw or self.sw
        self.sh = self.pending_sh or self.sh
        self.bloom_enabled = TheGame.bloom_enabled
 
        self.img:SetSize( self.sw, self.sh )
        local texw, texh = self.texw or self.sw, self.texh or self.sh

	local fmt0 = PIXEL_FORMAT.A8R8G8B8
	local fmt1 = -1
	if self.bloom_enabled then
		fmt1 = fmt0
	end
	local rw = math.round(texw)
	local rh = math.round(texh)
        self.scene_render_node:SetRenderFormat(rw, rh, fmt0, fmt1, true, false)
    end

    return self
end

function CaptureImage:RenderToScene( capture, x1, y1, x2, y2 )
    self:UpdateRenderTarget()   -- Update the size before getting the bounding box.
    if self.shader_dirty then
        if self.vig_active and self.hsv_active then
            self.img:SetEffect(global_shaders.UI_BLOOMMAP_HSVADJUST_VIG)
        elseif self.hsv_active then
            self.img:SetEffect(global_shaders.UI_BLOOMMAP_HSVADJUST)
        elseif self.vig_active then
            self.img:SetEffect(global_shaders.UI_BLOOMMAP_VIG)
        else
            self.img:SetEffect(global_shaders.UI_BLOOMMAP)
        end
        self.shader_dirty = false
    end

    if x1 == nil then
        x1, y1, x2, y2 = capture:GetBoundingBox()
    else
        capture:GetBoundingBox() -- Need to ensure bb is refreshed. (??)
    end

    self.camera:SetOrthoProjection( x1, x2, y2, y1 )


    self.scene_render_node:ClearTexture()
    self.scene_render_node:SetTreeScene(self.scene, capture.node, self.camera, (1 << RENDER_PASS_OPAQUE) | (1 << RENDER_PASS_ALPHA))
    self.scene_render_node:RenderToTexture(TheGame:GetLevelOfDetail())
        
    local bright = self.scene_render_node:GetOutputTexture(1)   
    if bright then
        self.img:SetTexture(bright, 1)
    end
    local output
    if self.blur_amt and self.blur_amt > 0 then
        output = TheGame:BlurImage(self.scene_render_node:GetOutputTexture(), self.blur_amt)
    else
        output = self.scene_render_node:GetOutputTexture()
    end
    self.img:SetTexture(output, 0)
    return self
end

function CaptureImage:SetVignette( val )
    self.vig_active = val > 0 
    self.colour_correct_constants.vignette:SetValue( val )
    self.shader_dirty = true
end

function CaptureImage:ApplyColourCorrectionValues(vals)
    if vals then
        self.colour_correct_constants.shadows:SetValue(table.unpack(vals.shadows.hsv))
        self.colour_correct_constants.midtones:SetValue(table.unpack(vals.midtones.hsv))
        self.colour_correct_constants.highlights:SetValue(table.unpack(vals.highlights.hsv))
        self.colour_correct_constants.shadows_col:SetValue(table.unpack(vals.shadows.colour))
        self.colour_correct_constants.midtones_col:SetValue(table.unpack(vals.midtones.colour))
        self.colour_correct_constants.highlights_col:SetValue(table.unpack(vals.highlights.colour))
    else
        self.colour_correct_constants.shadows:SetValue(0,1,1)
        self.colour_correct_constants.midtones:SetValue(0,1,1)
        self.colour_correct_constants.highlights:SetValue(0,1,1)
        self.colour_correct_constants.shadows_col:SetValue(1,1,1)
        self.colour_correct_constants.midtones_col:SetValue(1,1,1)
        self.colour_correct_constants.highlights_col:SetValue(1,1,1)
    end

    self.hsv_active = vals and not IsIdentityHSV(vals)
    self.shader_dirty = true
    
end

function CaptureImage:OnRemoved()
    -- print ("REMOVED CAPTURE NODE")
    self.scene_render_node:Dispose()
    self.scene_render_node = nil
end

