require "ui/widget"
require "ui/widgets/screen"

local DEFAULT_CARD_SCALE = 0.45
local PREVIEW_SCALE = 1.9

local DRAG_SCALE = 1.5
local DRAG_TIME_THRESHOLD = 0.3 -- Duration in seconds after which a click becomes a drag
local DRAG_TIME_THRESHOLD_SMALL = 0.3 -- Duration in seconds after which a click becomes a drag in small screen mode
local TOUCH_DRAG_MIN_DIST = 0 -- Minimum distance the user must drag up on a card before picking it up. Calculated later
local TOUCH_INITIAL_CARD_Y -- The first y position of a touch on a card. When the user drags up TOUCH_DRAG_MIN_DIST from this, it picks up the card

local DRAG_ZONE_HEIGHT = 0.28 -- Percentage of screen height where the drag zone begins; the drag zone dictates where a card can be dropped to play it.

local CARD_WIDTH = 140 -- Rough card width for purposes of hand layout, dictates spacing.
local MAX_HAND_WIDTH = 900
local MAX_PREVIEW_HAND_WIDTH = 1000 -- Used when there is a preview card

local PREVIEW_WIDTH = CARD_WIDTH * (PREVIEW_SCALE - 0.1) -- How much width is used for the preview card.

-------------------------------------------------------------------------------
-- Handles an arrangement of card widgets.

local CardArrangement = class("Widget.CardArrangement", Widget )
CardArrangement.can_reload = true

CardArrangement.CONTROL_MAP =
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self:IsCardSelected() and not self.card_dragging and self.fe:FindScreen( Screen.ChooseCardsPopup ) == nil
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.SELECT", Controls.Digital.MENU_ACCEPT) )
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self:IsInDragZone()
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.PLAY_CARD", Controls.Digital.MENU_ACCEPT) )
        end,
    },
}

function CardArrangement:init( card_class, card_scale, preview_offset_override, preview_scale_multiplier )
    Widget.init(self)

    assert( card_class == nil or is_class( card_class, Widget ))
    self.card_class = card_class
    self.default_card_scale = card_scale or DEFAULT_CARD_SCALE
    self.card_scale = self.default_card_scale
    self.preview_scale_multiplier = preview_scale_multiplier or 1.0
    self.preview_scale = self.preview_scale_multiplier * PREVIEW_SCALE
    self.small_screen_focus_card = nil -- On small screens, this card takes most of the height of the screen, and replaces the card_preview visually
    self.mask_glow_y = -74

    self.preview_offset_override = preview_offset_override or 0
    self.preview_offset = self.preview_offset_override or 0

    self.in_drag_zone = false
    self.drag_time_threshold = DRAG_TIME_THRESHOLD
    self.cards_chosen = {}

    self.card_widgets = {} -- Table of [Card] -> [CardWidget]
    self.card_order = {} -- Array of [CardWidget] (for ordered iterating)
    self.card_root = Widget()
    self:AddChild( self.card_root )
end

function CardArrangement:OnAdded()
    TheGame:GetEvents():ListenForEvent( "focus_changed", self, self.OnFocusChanged )
    self:StartUpdating()
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function CardArrangement:OnRemoved()
    self:StopUpdating()
    TheGame:GetEvents():RemoveListener( self )
end

function CardArrangement:OnControlModeChange( cm, device_id )
    CardArrangement._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.TOUCH then
    else
    end
end

function CardArrangement:OnScreenModeChange( sm )
    CardArrangement._base.OnScreenModeChange( self, sm )

    self.preview_offset = self.centered_mode and 0 or self.preview_offset_override or 0
    self.preview_scale = self.preview_scale_multiplier * PREVIEW_SCALE -- this is multiplied by self.card_scale
    local card_scale_mult = 1
    if sm == SCREEN_MODE.MONITOR then
        self.drag_time_threshold = DRAG_TIME_THRESHOLD
        self:RemoveSmallScreenFocusCard()
        card_scale_mult = 1
    elseif sm == SCREEN_MODE.TV then
        self.drag_time_threshold = DRAG_TIME_THRESHOLD_SMALL
        self:RemoveSmallScreenFocusCard()
        card_scale_mult = 1.4
        self.preview_offset = self.preview_offset + 80
    elseif sm == SCREEN_MODE.SMALL then
        self.drag_time_threshold = DRAG_TIME_THRESHOLD_SMALL
        self:CreateSmallScreenFocusCard()
        card_scale_mult = 1.5
    end
    self.card_scale = self.default_card_scale * card_scale_mult
    PREVIEW_WIDTH = CARD_WIDTH * self.preview_scale * 1.2 -- How much width is used for the preview card.
    
    local width, height = TheGame:FE():GetScreenDims()
    TOUCH_DRAG_MIN_DIST = height*0.2
end

function CardArrangement:CreateSmallScreenFocusCard()
    if self.small_screen_focus_card == nil then
        self.small_screen_focus_card = self:AddChild( self.card_class() )
            :SetScale( self.centered_mode and 1.1 or 1.25 )
            :IgnoreInput(true)
            :Hide()
    end
end

function CardArrangement:RemoveSmallScreenFocusCard()
    if self.small_screen_focus_card ~= nil then
        self.small_screen_focus_card:Remove()
        self.small_screen_focus_card = nil
    end
end

function CardArrangement:RefreshSmallScreenFocusCard()
    -- If we're supposed to show a small-screen focus card, do it
    if self.small_screen_focus_card then
        if self.card_preview then
            if self.card_preview ~= self.small_screen_focus_card_last then
                local card_preview_widget = self.card_widgets[ self.card_preview ]
                self.small_screen_focus_card:RefreshCard( self.card_preview )
                    :Show()
                    :LayoutBounds( "center", nil, card_preview_widget )
                    :LayoutBounds( nil, self.centered_mode and "center" or "above", 0, self.centered_mode and -50 or -80 )
                    :MoveTo( nil, self.small_screen_focus_card.y+30, 0.1, easing.outQuad )

                local idx = table.arrayfind( self.card_order, card_preview_widget )
                local prev_card = self.card_order[ idx - 1 ]
                self.small_screen_focus_card:SetFocusDir( "left", prev_card or self.small_screen_focus_card )

                local after_card = self.card_order[ idx + 1 ]
                self.small_screen_focus_card:SetFocusDir( "right", after_card or self.small_screen_focus_card )

                self.small_screen_focus_card_last = self.card_preview
            end
        else
            self.small_screen_focus_card:Hide()
            self.small_screen_focus_card_last = nil
        end
    end
end

function CardArrangement:LayoutPreviewToolTip( card_widget, tooltip_widget )
    if self.small_screen_focus_card then
        -- We're displaying a large card, so let's hide its small cards and align the tooltip to the right of the large card
        tooltip_widget:HideCardEntries()
            :LayoutBounds( "after", "top", self.small_screen_focus_card ):Offset( 0, -20 )

        -- And now check if there isn't enough room on the right side of the screen for the tooltip
        local screenw, screenh = TheGame:FE():GetScreenDims()
        local tooltip_xmin, tooltip_ymin, tooltip_xmax, tooltip_ymax = tooltip_widget:GetWorldBoundingBox()
        if tooltip_xmax > screenw then
            -- The tooltip is going over the right edge of the screen. Let's align it to the left of the card instead
            tooltip_widget:LayoutBounds( "before", "top", self.small_screen_focus_card ):Offset( 0, -20 )
        end
    else
        -- No large card, let's just leave the tooltip aligned to the preview then
    end
    return self
end

function CardArrangement:OnScreenResize(w, h)
    CardArrangement._base.OnScreenResize(self, w, h)

    self:RefreshMaskSize()

    local width, height = TheGame:FE():GetScreenDims()
    TOUCH_DRAG_MIN_DIST = height*0.2
end

function CardArrangement:OnFocusChanged( event_name, prev_focus, new_focus )
    if TheGame:FE():IsTouchMode() then

    elseif prev_focus and self:IsAncestorOf( prev_focus ) and prev_focus.card and prev_focus.card == self.card_dragging then
        -- The dragging card just lost focus, which means the user pressed direction keys
        -- Lets unselect it then
        self:SetCardDragging( nil )

    elseif prev_focus and prev_focus.card then
        -- The direction keys were pressed, let's prevent the mouse from hovering on one of the focused cards (stopping the focus from leaving it)
        --- ccc: I'm disabling this because forcing the mouse position causes weird side effects; eg. it causes a pending control mode change
        -- on the subsequent frame because "the mouse position changed".
        -- Frankly I don't think changing focus using mouse & KB is very standard anyway.
        -- TheGame:GetInput():ResetMousePos()
    end
end

function CardArrangement:SetFanningRadius( radius )
    self.fan_radius = radius
    return self
end

function CardArrangement:BeginChoosingForPlay( engine, min_count, max_count )
    self:BeginChoosing( engine, min_count, max_count )
    self.drag_zone = true
end


function CardArrangement:BeginChoosing( engine, min_count, max_count, pred )
    self.is_choosing = true
    self.min_choose = min_count
    self.max_choose = max_count
    table.clear( self.cards_chosen )

    if pred then
        for card, w in pairs( self.card_widgets ) do
            w:SetChooseFn( pred )
        end
        self:RefreshCards( engine )
    end

    return self
end

function CardArrangement:IsChoosing()
    return self.is_choosing == true
end

function CardArrangement:EndChoosing( engine )
    self.is_choosing = nil
    self.min_choose = nil
    self.max_choose = nil
    self:SetCardDragging( nil )
    self.drag_zone = nil
    self:SetCardPreview( nil )
    table.clear( self.cards_chosen )

    for card, w in pairs( self.card_widgets ) do
        w:SetChooseFn( nil )
    end
    self:RefreshCards( engine )
end

function CardArrangement:DisablePreview( value )
    self.no_preview = value
    return self
end

-- Showing the cards in the center of the screen, instead of the bottom. Preview card scales up in place.
function CardArrangement:SetCenteredMode()
    self.centered_mode = true
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
    return self
end

function CardArrangement:IsEmpty()
    return #self.card_order == 0
end

function CardArrangement:GetCardCount()
    return table.count( self.card_widgets )
end

function CardArrangement:GetCardAt( i )
    local w = self.card_order[ i ]
    return w and w.card
end

function CardArrangement:Cards()
    return pairs( self.card_widgets )
end

function CardArrangement:SetFocusable( focusable )
    for k, w in pairs( self.card_widgets ) do
        w:SetFocusable( focusable )
    end
end

function CardArrangement:RefreshCards( ... )
    for card, w in pairs( self.card_widgets ) do
        w:RefreshCard( card, ... )
    end
end

function CardArrangement:RefreshCardOrder()
    -- Order hand
    local lastw
    for i = #self.card_order, 1, -1 do

        local w = self.card_order[i]
        w:SetToolTipLayoutFn( nil ) -- Reset the tooltip layout fn override. It is only for the preview card
        if lastw == nil then
            w:SendToFront()
        else
            w:SendBehind( lastw )
        end

        lastw = w
    end

    -- Put preview frontmost
    local frontmost = self.card_dragging or self.card_preview
    if frontmost then
        local preview_widget = self.card_widgets[ frontmost ]
        if self.small_screen_focus_card then
            preview_widget:SetToolTipLayoutFn( function( card_widget, tooltip_widget ) self:LayoutPreviewToolTip( card_widget, tooltip_widget ) end )
        end
        if preview_widget then preview_widget:SendToFront() end
    end

    self:ReNav()
end

-- Overrides the focusing order for the cards
function CardArrangement:ReNav()
    local lastw = nil
    local first = nil
    for i = 1, #self.card_order, 1 do
        local v = self.card_order[i]
        first = first or v
        v:SetFocusDir( "left", lastw, true )
        v:SetFocusDir( "up", v )
        v:SetFocusDir( "down", v )
        v:SetFocusDir( "right", v )
        lastw = v
    end
    -- Uncomment this if we want to loop around
    -- if first then first:SetFocusDir( "left", lastw, true ) end
end

function CardArrangement:AnimateCards( dt )
    if not self.no_preview and not self.card_dragging then
        self:CheckCardPreview( dt )
    end
    -- Animate each card.
    for card, w in pairs( self.card_widgets ) do
        self:AnimateCard( dt, card )
    end

    self:UpdateMaskGlow()

    -- Need to refresh preview outlines when transitioning to/from the drag zone.
    local in_drag_zone = self:IsInDragZone()
    if in_drag_zone ~= self.in_drag_zone then
        self.in_drag_zone = in_drag_zone
        local w = self.card_widgets[ self.card_dragging ]
        if w then
            w:SetHighlight( in_drag_zone )
        end
    end
end

function CardArrangement:AnimateCard( dt, card )
    local w = self.card_widgets[ card ]
    local x, y, alpha, scale, rot, speed = self:FindCardPosition( card )
    
    if dt and speed then
        w:AnimateToPlacement( dt * speed, x, y, alpha, scale * self.card_scale, rot )
    else
        w:ClearAnim()
        w:SetPos( x, y )
        w:SetTintAlpha( alpha )
        w:SetScale( scale * self.card_scale )
        w:SetRotation( rot )
    end

end

function CardArrangement:GetCardPreview()
    return self.card_preview
end

function CardArrangement:CheckCardPreview( dt )
    local preview = nil
    -- Figure out the current preview card.
    for i, w in ipairs( self.card_order ) do
        if w.dead_time then
            w.dead_time = w.dead_time - dt
            if w.dead_time <= 0 then
                w.dead_time = nil
            end
        elseif w.hover or w.focus then
            preview = w.card
        end
    end

    self:SetCardPreview( preview )
end

function CardArrangement:OnRightClickCard( widget )
    if not self:GetChosenCard() and not self.card_dragging then
        widget:ShowCardInfoPopup()
        return true
    else
        return false -- Don't sink input.
    end
end

function CardArrangement:CreateCardWidget( card )
    local w = self.card_widgets[ card ]
    if w == nil then
        w = self.card_class( card )
        if self.card_scale then
            w:SetScale( self.card_scale )
        end
        self.card_root:AddChild( w )
        self.card_widgets[ card ] = w
        w:SetParticlesEnabled( TheGame:GetLevelOfDetail() > 1 )
        w:SetShineEnabled(true)
        w:SetOnRightClickFn( function( widget ) return self:OnRightClickCard( widget ) end )

        w.OnGainFocus = function() w._base.OnGainFocus(w) self:UpdateCardImageState( w, card.id ) end
        w.OnLoseFocus = function() w._base.OnLoseFocus(w) self:UpdateCardImageState( w, card.id ) end
        w.OnGainHover = function() w._base.OnGainHover(w) self:UpdateCardImageState( w, card.id ) end
        w.OnLoseHover = function() w._base.OnLoseHover(w) self:UpdateCardImageState( w, card.id ) end

        w:SetToolTipLayoutFn( function( self, tooltip_widget )
            -- Position the tooltip_widget in relation to the card
            if TheGame:FE():GetScreenMode() == SCREEN_MODE.TV
            or TheGame:FE():GetScreenMode() == SCREEN_MODE.SMALL then
                if tooltip_widget.GetWorldCardSize then
                    -- Get card widget size
                    local tooltip_card_w, tooltip_card_h = tooltip_widget:GetWorldCardSize()
                    local fx1, fy1, fx2, fy2 = w:GetWorldBoundingBox()
                    local dx = (fx2 - fx1)/2 - tooltip_card_w/2
                    -- Get screen height
                    local screenw, screenh = self:GetFE():GetScreenDims()
                    -- Position tooltip so that its giant card preview is right above the focused card widget
                    tooltip_widget:LayoutBounds( "left", "bottom", w ):Offset( dx, screenh*0.1 )
                    -- Check if the info on the right of the card preview is off-screen
                    local tx1, ty1, tx2, ty2 = tooltip_widget:GetWorldBoundingBox()
                    if tx2+10 > screenw then
                        -- The tooltip info is past the right edge of the screen. Let's show it on the left side of the card instead then
                        tooltip_widget:SetLeft( true )
                        tooltip_widget:LayoutBounds( "right", "bottom", w ):Offset( -dx, screenh*0.1 )
                    end
                end
            end
        end )

        -- Hack to prevent preview while a card is initially settling into position
        w.dead_time = 0.2
        w:ShowToolTipOnFocus()

        table.insert( self.card_order, w )
        self:RefreshCardOrder()
        return w, true
    else
        return w, false
    end
end

function CardArrangement:UpdateCardImageState( widget, card_id )
    if widget.focus then
        widget.bg:Bloom(0.03)
        widget.art:Bloom(0.03)
    else
        widget.bg:Bloom(0)
        widget.art:Bloom(0)
    end
end

function CardArrangement:RemoveCardWidget( w )
    local card = w.card
    assert( self.card_widgets[ card ] == w )
    assert( table.contains( self.card_order, w ))
    table.arrayremove( self.card_order, w )
    self.card_widgets[ card ] = nil
    w:Remove()
    if self.card_preview == card then
        self:SetCardPreview( nil )
    end
end

function CardArrangement:SwapCardWidget( card, idx )
    local w = self.card_widgets[ card ]
    local old_idx = table.arrayfind( self.card_order, w )
    if old_idx ~= idx then
        table.remove( self.card_order, old_idx )
        table.insert( self.card_order, idx, w )
        self:RefreshCardOrder()
    end
end

function CardArrangement:FindCardWidget( card )
    return self.card_widgets[ card ]
end

function CardArrangement:AttachCardWidget( card, w, idx )
    if self.card_widgets[ card ] == nil then

        self.card_widgets[ card ] = w
        w:Reparent( self.card_root )
        w:SetOnRightClickFn( function( widget ) return self:OnRightClickCard( widget ) end )
        w.dead_time = 0.2

        if idx then
            table.insert( self.card_order, math.min(idx, #self.card_order+1), w )
        else
            table.insert( self.card_order, w )
        end
        self:RefreshCardOrder()
    end
end

function CardArrangement:DetachCardWidget( card )
    local w = self.card_widgets[ card ]
    while w do
        w:WaitOnCoros()
        if self.card_widgets[ card ] == w then
            break -- Still attached to us, continue.
        else
            w = self.card_widgets[ card ]
        end
    end

    local idx
    if w then
        if self.card_preview == card then
            self:SetCardPreview( nil )
        end
        if self.card_dragging == card then
            self:SetCardDragging( nil )
        end
        if self:IsChosen( card ) then
            self:ChooseCard( card )
        end
        w:SetChooseFn( nil )
        w:ClearFocusDirs()
        
        idx = table.arrayfind( self.card_order, w )
        table.remove( self.card_order, idx )
        self.card_widgets[ card ] = nil
        self:ReNav()
        
        w.anim = nil
    end
    return w, idx
end

function CardArrangement:SetArrangementWidth( width )
    MAX_HAND_WIDTH = width or 900
    return self
end

function CardArrangement:GetArrangementWidth()
    local n = table.count( self.card_widgets )
    local max_width = math.min( MAX_HAND_WIDTH, n * CARD_WIDTH )

    if self.card_preview then
        return math.min( MAX_PREVIEW_HAND_WIDTH, n * CARD_WIDTH )
    else
        return math.min( MAX_HAND_WIDTH, n * CARD_WIDTH )
    end
end

function CardArrangement:TransformToFan( x, y, r )
    local n = table.count( self.card_widgets )
    local max_width = self:GetArrangementWidth()
    local half_width = max_width*0.5

    local radius = math.sqrt( r^2 + half_width^2 )
    local theta_min = math.atan( r / half_width )
    local t = 1.0 - math.abs(x) / half_width
    local theta = Lerp( theta_min, math.pi/2, t )

    local tan_theta = math.tan( theta )
    local y1 = (radius + y) * tan_theta / math.sqrt( tan_theta^2 + 1 )
    local x1 = math.sqrt( (radius + y)^2 - y1^2 )
    if x < 0 then
        return -x1, y1 - r, math.pi/2 - theta
    elseif x == 0 then
        return 0, radius - r + y, 0
    else
        return x1, y1 - r, theta - math.pi/2
    end
end

function CardArrangement:FindCardPosition( card )
    if card == self.card_dragging then
        local mx, my
        if self:IsGamepadMode() then
            local screenw, screenh = self:GetFE():GetScreenDims()
            mx, my = screenw/2, screenh/2
        else
            mx, my = TheGame:GetInput():GetMousePos()
        end
        local w = self.card_widgets[ card ]
        local x, y = w.parent:TransformFromWorld( mx, my )
        local alpha = 1
        return x, y, alpha, DRAG_SCALE, 0, 5.0, 1
    else
        local x, y, scale, rot, speed, alpha = self:FindCardPositionInArrangement( card )
        return x, y, alpha, scale, rot, speed
    end
end

function CardArrangement:FindCardIndexInArrangement( card )
    local w = self.card_widgets[ card ]
    return table.arrayfind( self.card_order, w )
end

function CardArrangement:FindCardPositionInArrangement( card )
    local n = table.count( self.card_widgets )
    local card_widget = self.card_widgets[ card ]

    local idx = self:FindCardIndexInArrangement( card )
    assert( idx, card ) -- Why are we here if this card isn't in hand?

    local x, y, scale, rot
    local preview_widget = self.card_widgets[ self.card_preview ]

    if preview_widget ~= nil then
        -- The preview slot should stay in place, and its slot is now sized PREVIEW_WIDTH.
        local preview_idx = self:FindCardIndexInArrangement( self.card_preview )
        local max_width = self:GetArrangementWidth()
        local slot_width = max_width / n
        local preview_x = preview_idx * slot_width - slot_width/2

        -- Left slots
        if idx < preview_idx then
            slot_width = (preview_x - PREVIEW_WIDTH/2) / (preview_idx - 1)
            x = idx * slot_width - slot_width/2
        elseif idx == preview_idx then
            x = preview_x
        else
            slot_width = (max_width - (preview_x + PREVIEW_WIDTH/2)) / (n - preview_idx)
            x = (preview_x + PREVIEW_WIDTH/2) + (idx - preview_idx) * slot_width - slot_width/2
        end
        x = x - max_width/2

    elseif n > 1 then
        -- No card previews: layout cards in 'slots' of identical width
        local max_width = self:GetArrangementWidth()
        local slot_width = max_width / n
        x = idx * slot_width - slot_width/2
        x = x - max_width/2
    else
        x = 0
    end

    if self:IsChosen( card ) then
        y = 60
    else
        y = 0
    end

    if self.fan_radius then
        x, y, rot = self:TransformToFan( x, y, self.fan_radius)
    else
        rot = 0
    end

    -- x, y, scale, rot, speed, alpha
    local alpha = 1
    if card == self.card_preview then
        alpha = self.small_screen_focus_card == nil and 1 or 0
        return x, self.preview_offset, self.preview_scale, 0, nil, alpha
    else
        return x, y, 1, rot, 1, alpha
    end
end

function CardArrangement:SetCardPreview( card )
    if card ~= self.card_preview then
        self.card_preview = card
        self:RefreshCardOrder()
        if card then
            self.focus_hint = table.arrayfind(self.card_order, self.card_widgets[ card ] )
        end
    end
    self:RefreshSmallScreenFocusCard()

    TheGame:FE():RefreshControlOptions()
end

function CardArrangement:SetOnDrag( fn )
    self.on_drag = fn
end

function CardArrangement:SetCardDragging( card )
    if card == self.card_dragging then
        return
    end

    if card and self.card_preview then
        -- Unpreview any card if we are starting to drag.
        self:SetCardPreview( nil )
    end

    local old_card = self.card_dragging
    if old_card then
        -- Cancel state on currently dragged card.
        local w = self.card_widgets[ old_card ]
        w:DisableToolTip( false )
        w:SetHighlight( false )
        w.dead_time = 0.2
    end

    self.card_dragging = card

    if card then
        local w = self.card_widgets[ card ]
        w:DisableToolTip( true )
        w:SetHighlight( self:IsInDragZone() )
        self.drag_time = os.clock()
    else
        self.drag_time = nil
    end

    self:RefreshCardOrder()

    if self.on_drag then
        self.on_drag( card, old_card )
    end

    TheGame:FE():RefreshControlOptions()
end

function CardArrangement:IsInDragZone()
    if self.card_dragging then
        local w = self.card_widgets[ self.card_dragging ]
        local cx, cy = w:TransformToWorld( 0, 0 )
        if not self.drag_zone then
            return false
        end
        local w, h = self:GetFE():GetScreenDims()
        local y = DRAG_ZONE_HEIGHT * h
        return cy >= y and cy <= h
    end
end

function CardArrangement:SetCardOutline( card, tint )
    local w = self.card_widgets[ card ]
    if w then
        w:SetOutline( tint )
    end
end

function CardArrangement:ClearCards()
    self:EndChoosing()
    for card, w in pairs( self.card_widgets ) do
        self:RemoveCardWidget( w )
    end
end

function CardArrangement:SetOnChoose( fn, try_fn )
    self.on_choose = fn
    self.on_try_choose = try_fn
end

function CardArrangement:GetOnChoose()
    return self.on_choose, self.on_try_choose
end

function CardArrangement:ChooseCard( card )
    self:SetCardDragging( nil )

    if not self:IsChoosing() then
        return
    end

    if table.arrayfind( self.cards_chosen, card ) then
        table.arrayremove( self.cards_chosen, card )
        self:SetCardOutline( card, Widget.CardWidget.CLR_OUTLINE)

        if self.on_choose then
            self.on_choose( card, false )
        end
        self.gamepad_mode = nil

    else
        if card == self.card_preview then
            self:SetCardPreview( nil )
        end

        table.insert( self.cards_chosen, card )
        while #self.cards_chosen > self.max_choose do
            self:ChooseCard( self.cards_chosen[1] ) -- Unchoose last in list.
        end
        self:SetCardOutline( card, UICOLOURS.WHITE )
        self.touch_dragging_over_deck = false

        if self.on_choose then
            self.on_choose( card, true)
        end
    end
end

function CardArrangement:GetChosenCards()
    local can_confirm = #self.cards_chosen >= self.min_choose and #self.cards_chosen <= self.max_choose
    return self.cards_chosen, can_confirm
end

function CardArrangement:GetChosenCard()
    return self.cards_chosen[1]
end

function CardArrangement:IsCardSelected()
    return self:GetCardPreview() or self:GetDragCard()
end

function CardArrangement:GetDragCard()
    return self.card_dragging
end

function CardArrangement:IsChosen( card )
    assert( card )
    return table.arrayfind( self.cards_chosen, card ) ~= nil
end

function CardArrangement:IsGamepadMode()
    return TheGame:FE():IsGamepadMode() or self.gamepad_mode
end

function CardArrangement:TryChooseCard( card, device, gamepad_mode )
    local w = self.card_widgets[ card ]
    if w == nil then
        return false
    end

    if self:IsChosen( card ) then
        -- Toggle choice.
        w:PlayDeselectSound()
        w:DisableToolTip( false )
        self:ChooseCard( card )
        return true

    else
        w:PlaySelectSound()

        self.gamepad_mode = gamepad_mode

        if self.on_try_choose and self.on_try_choose( card ) then
            -- handled
        elseif not self.is_choosing then
            -- Not currently choosing.
            return false
        elseif not self.drag_zone or device == "keyboard" or device == "gamepad" then
            -- Click chooses immediately.
            self:ChooseCard( card )
        else
            -- Start dragging this card.
            while #self.cards_chosen > 0 do
                self:ChooseCard( self.cards_chosen[ #self.cards_chosen ] )
            end
            self:SetCardDragging( card )
        end
        return true
    end

    return false
end

function CardArrangement:HandleControlDown(control, device)
    if TheGame:FE():IsTouchMode() then

        if self.is_choosing and not self.drag_zone then
            -- When choosing cards from the hand to be discarded or such
            -- Do just a tap to select each card
            local mx, my = TheGame:GetInput():GetMousePos()
            for card, w in pairs( self.card_widgets ) do
                if w:CheckHit( mx, my ) then
                    self:TryChooseCard( card )
                end
            end
        elseif self.card_dragging then
            if self:IsInDragZone() then
                self:ChooseCard( self.card_dragging )
            else
                self.card_widgets[ self.card_dragging ]:PlayDeselectSound()
                self:SetCardDragging( nil )
            end
        else
            -- Check if we're above any of the deck cards
            local mx, my = TheGame:GetInput():GetMousePos()
            for card, w in pairs( self.card_widgets ) do
                if w:CheckHit( mx, my ) then
                    if not w.focus then TOUCH_INITIAL_CARD_Y = my end
                    w:SetFocus()
                    self.touch_dragging_over_deck = true
                end
            end
        end

    elseif control:Has( Controls.Digital.MENU_ACCEPT ) and device == "mouse" then
        -- If the mouse pressed down on a card, select it. But not if it was selected with Enter on the keyboard
        if self.card_preview then
            if self:TryChooseCard( self.card_preview, device ) then
                return true
            end
        end
    end
end

function CardArrangement:HandleControlUp(control, device)
    if control:Has( Controls.Digital.MENU_ACCEPT ) and (device == "gamepad" or device == "key") then
        -- If the card was selected with a keyboard, only select it on control up
        if self.card_preview then
            if self:TryChooseCard( self.card_preview, device, true ) then -- Always gamepad, even keyboard is meant to emulate gamepad here.
                return true
            end
        end
        if self.card_widgets[ self.card_dragging ] and self:IsInDragZone() then
            local card = self.card_dragging
            self:ChooseCard( self.card_dragging )
            return true
        end
    elseif control:Has( Controls.Digital.MENU_ACCEPT ) and ( device == "mouse" or device == "touch" ) then
        local drag_card = self.card_widgets[ self.card_dragging ]
        if drag_card then
            if self:IsInDragZone() then
                local card = self.card_dragging
                self:ChooseCard( self.card_dragging )
                return true
            elseif (os.clock() - self.drag_time) > self.drag_time_threshold then
                drag_card:PlayDeselectSound()
                self:SetCardDragging( nil )
                return true
            end
        elseif not self.is_choosing and not self.preview then
            -- We're being dealt a card. A tap/click on it should continue
            local mx, my = TheGame:GetInput():GetMousePos()
            for card, w in pairs( self.card_widgets ) do
                if w:CheckHit( mx, my ) then
                    return false
                end
            end
        elseif TheGame:FE():IsTouchMode() then
            -- Any touch end causes the cards to unselect
            self:SetCardDragging( nil )
            if self:GetCardPreview() then self:ChooseCard( self:GetCardPreview() ) end
            if self:GetChosenCard() then self:ChooseCard( self:GetChosenCard() ) end
            return true
        else
            -- Necessary for keyboard/mouse mode to click on a card and it staying selected until a target is clicked
            return true
        end

    elseif control:Has( Controls.Digital.MENU_CANCEL, Controls.Digital.MENU_ALT ) then
        if self.card_dragging then
            self.card_widgets[ self.card_dragging ]:PlayDeselectSound()
            self.card_widgets[ self.card_dragging ]:DisableToolTip( false )
            self:SetCardDragging( nil )
            return true
        end

    end

    if TheGame:FE():IsTouchMode() then
        self.touch_dragging_over_deck = false
        TheGame:FE():ClearFocusWidget()
    end
end

function CardArrangement:OnUpdate( dt )
    CardArrangement._base.OnUpdate( self, dt )
    if self.touch_dragging_over_deck then
        local mx, my = TheGame:GetInput():GetMousePos()
        for card, w in pairs( self.card_widgets ) do

            -- Check if the user slid their finger onto this card
            if w:CheckHit( mx, my ) then
                -- They did. Give this new card focus
                if not w.focus then TOUCH_INITIAL_CARD_Y = my end
                w:SetFocus()
            end

            -- Check if the user dragged up from a card enough to pick it up
			-- self.card_dragging checks if a card is being dragged
			-- self:GetChosenCard() checks if a card is selected (meaning there's a target arrow going on). Without this, the TryChooseCard will trigger after the card has been picked up
            if w.focus and my > TOUCH_INITIAL_CARD_Y + TOUCH_DRAG_MIN_DIST and not ( self.card_dragging or self:GetChosenCard() ) then
                -- We're dragging this card! Action time!
                self:TryChooseCard( w.card )
            end
        end

    end

end

function CardArrangement:RenderDebugUI( ui, panel )

    ui:Columns( 3 )
    for i, w in ipairs( self.card_order ) do
        local card = table.find( self.card_widgets, w )
        panel:AppendTable( ui, card )
        ui:NextColumn()

        panel:AppendTable( ui, w )
        ui:NextColumn()

        local x, y, alpha, scale, rot, speed = self:FindCardPosition( card )
        ui:Text( string.format( "<%s, %s>, %.2f, %.2f, %.2f", x, y, alpha, scale, rot ))
        ui:NextColumn()
    end
    ui:Columns( 1 )

    ui:Separator()
    ui:Spacing()

    return true
end

function CardArrangement:ShowMask()
    self.is_masking = true
    self:RefreshMaskSize()

    self.particles = Widget.ParticleSystem()
    self:AddChild( self.particles )
    self.particles:LoadParams( "cardarrangement_glow" )
    self.particles_emitter = self.particles:GetEmitter( 1 )
    self.particles:SetTintAlpha( 0 )

    -- Hide the seam where the cards are masked off
    self.mask_glow = self:AddChild( Widget.Image( engine.asset.Texture("UI/cardarrangement_glow.tex") ) )
    self.mask_glow_w, self.mask_glow_h = self.mask_glow:GetSize()
    self.mask_glow:SetPos( 0, self.mask_glow_y )
    self.mask_glow:SetSize( self.mask_glow_w*0.6, self.mask_glow_h )
    self.mask_glow:SetTintAlpha( 0 )

    -- Make a blocking region, preventing the mouse from making a card blink in and out of focus
    -- Change color to 0x000000da to see its area
    self.mask_block = self:AddChild( Widget.SolidBox( self.mask_glow_w*0.6, 20, 0x00000000 ) )
    self.mask_block:BlockMouse()
    self.mask_block:SetPos( 0, self.mask_glow_y )

    self.particles:SetPos( 0, self.mask_glow_y )
    return self
end

function CardArrangement:RefreshMaskSize()
    if self.is_masking then
        -- Get the screen size
        local screen_w, screen_h = TheGame:FE():GetScreenDims()
        self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
        self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )
        self.card_root:SetScissor( -self.screen_w/2, self.mask_glow_y, self.screen_w, self.screen_h )
    end
    return self
end

--this is quite expensive
function CardArrangement:UpdateMaskGlow()
    if self.mask_glow then
        local xmin, xmax = math.huge, -math.huge
        for card, widget in pairs( self.card_widgets ) do
            if card ~= self.card_dragging then
                local x1, y1, x2, y2 = widget:GetBoundingBox()
                x1, y1 = self:TransformFromWidget( widget, x1, y1 )
                x2, y2 = self:TransformFromWidget( widget, x2, y2 )
                xmin = math.min( xmin, x1 )
                xmax = math.max( xmax, x2 )
            end
        end
        local cards_w = xmax - xmin
        local cards_x = ( xmax + xmin ) / 2
        self.mask_glow:SetSize( cards_w+300, self.mask_glow_h )
        self.mask_glow:SetPos( cards_x, self.mask_glow_y )
        self.mask_block:SetSize( cards_w+50, 20 )
        self.mask_block:SetPos( cards_x, self.mask_glow_y )
        local x1, y1, x2, y2 = self.mask_glow:GetBoundingBox()
        self.particles_emitter:SetEmitterBounds( 0, 0, cards_w, 20 )
        
        -- Hide or show the glow depending on whether there are cards or not
        if cards_w > 50 then
            self.mask_glow:SetTintAlpha( 1 )
            self.particles:SetTintAlpha( 1 )
        else
            self.mask_glow:SetTintAlpha( 0 )
            self.particles:SetTintAlpha( 0 )
        end
    end
end

function CardArrangement:GetDefaultFocus()
    local idx = self.focus_hint or (1 + math.floor( #self.card_order / 2 ))
    for i = idx, #self.card_order do
        local widget = self.card_order[ i ]
        if widget:IsEnabled() then
            return widget
        end
    end
    for i = math.min( #self.card_order, idx - 1 ), 1, -1 do
        local widget = self.card_order[ i ]
        if widget:IsEnabled() then
            return widget
        end
    end
end
