local fun = require "util/fun"
local CONTENT_LEFT_MARGIN = 15 -- Extra size added on the left of scroll boxes, so that the contents can scale up on hover without clipping


require "ui/widgets/compendiumwidget"
------------------------------------------------------------------------------------------
-- Shows the cards panel

local CardCompendium = class( "Widget.CardCompendium", Widget.CompendiumWidget )

local assets = {
    selection = engine.asset.Texture("UI/card_selection.tex"),
    wideblock = engine.asset.Texture("UI/gameover_wideblock.tex"),
    gameover_arrow_left = engine.asset.Texture("UI/gameover_arrow_left.tex"),
    gameover_arrow = engine.asset.Texture("UI/gameover_arrow.tex"),
    gameover_divider = engine.asset.Texture("UI/gameover_divider.tex"),
    motd_page = engine.asset.Texture("UI/motd_page.tex"),
}

local CARD_SCALE = 0.45
local CARD_PREVIEW_SCALE = 0.5
local battle_defs = require "battle/battle_defs"
local negotiation_defs = require "negotiation/negotiation_defs"

function CardCompendium:init( width, height, sidebar_width )
    CardCompendium._base.init( self, width, height, sidebar_width )

    self.scroll_root:Hide()
    self.card_grid = self:AddChild( Widget.CardGrid() )
        :SetSize( self.content_w+CONTENT_LEFT_MARGIN, self.widget_h-2 )
        :SetLeftPadding( 60 )
        :SetTopPadding( 20 )
        :SetFocusLock(true)
        :SetOnCardClickFn( function(card_widget) card_widget:ShowCardInfoPopup() end )

    self.card_grid.scroll_panel:SetIgnoreStickScroll()

    self.card_grid.scroll_panel:ShowBar( SCROLLBAR.IF_NEEDED )

    self.card_defs = {}

    local function FilterBattleCard( card_def )
        return not card_def.hide_in_cardex and not card_def.base_id and not CheckAnyBits( card_def.flags, battle_defs.CARD_FLAGS.SPECIAL | battle_defs.CARD_FLAGS.NPC ) 
    end

    -- Get all the battle cards
    local battle_cards = BattleCardCollection():Filter( FilterBattleCard )
    for k, card_def in ipairs( battle_cards.items ) do
        self.card_defs[card_def.id] = card_def
    end

    local function FilterNegotiationCard( card_def )
        return not card_def.hide_in_cardex and not card_def.base_id and not CheckAnyBits( card_def.flags, negotiation_defs.CARD_FLAGS.SPECIAL | negotiation_defs.CARD_FLAGS.OPPONENT | negotiation_defs.CARD_FLAGS.BYSTANDER ) 
    end

    -- Get all the negotiation cards
    local negotiation_card_options = {}
    local negotiation_cards = NegotiationCardCollection():Filter( FilterNegotiationCard )
    for k, card_def in ipairs( negotiation_cards.items ) do
        self.card_defs[card_def.id] = card_def
    end
end

function CardCompendium:OnAdded()

    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function CardCompendium:OnControlModeChange( cm, device_id )
    CardCompendium._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.card_grid:ShowBar( SCROLLBAR.IF_NEEDED )
    end
    if cm == CONTROL_MODE.TOUCH then
        self.card_grid:ShowBar( SCROLLBAR.NEVER )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.card_grid:ShowBar( SCROLLBAR.NEVER )
    end
end

function CardCompendium:OnScreenModeChange( sm )
    CardCompendium._base.OnScreenModeChange( self, sm )
end

function CardCompendium:HasSidebar()
    return true
end

function CardCompendium:SetScrollEnabled( enabled )
    self.scroll_root:SetEnabled( enabled )
    self.card_grid:SetScrollEnabled( enabled )
end

function CardCompendium:OnControlModeChange( cm, device_id )
    CardCompendium._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.sidebar:SetScrollEnabled( true )
        self.card_grid:SetScrollEnabled( true )
    end
    if cm == CONTROL_MODE.TOUCH then
        self.sidebar:SetScrollEnabled( true )
        self.card_grid:SetScrollEnabled( true )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        -- So that using the gamepad's joystick doesn't scroll both scroll panels
        self.sidebar:SetScrollEnabled( false )
        self.card_grid:SetScrollEnabled( true )
    end
end

function CardCompendium:AddSidebarFilters()
    -- Add character filters
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_CHARACTERS" )
    self.series_filter_group = self.sidebar:AddFilterWidget( Widget.FilterRadioGroup() )
    self.agent_filters = {}

    local player = TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent()
    local selected_idx = 1
    for k, data in ipairs(GetAllPlayerBackgrounds()) do
        if data.id then
            local row = self.series_filter_group:AddFilter( Widget.CardSeriesFilterRow( data.data.player_agent, { data.id } ) )
            self.agent_filters[data.id] = row
            if player and player:GetContentID() == data.id then
                selected_idx = k
            end
            -- Lock filters for locked characters
            local is_character_locked = not Screen.Compendium.ShowAllUnlocks() and TheGame:GetGameProfile():IsCharacterLocked( data.id )
            self.agent_filters[data.id]:SetLocked( is_character_locked )
        end
    end

    local npc_character = table.arraypick( { "DRONE_GOON", "JAKES_ASSASSIN", "JAKES_ASSASSIN2" } )
    self.series_npc = self.series_filter_group:AddFilter( Widget.CardSeriesFilterRow( npc_character, { CARD_SERIES.NPC } ))
    local general_character = table.arraypick( { "BANDIT_GOON", "JAKES_SMUGGLER", "ADMIRALTY_CLERK" } )
    self.series_general = self.series_filter_group:AddFilter( Widget.CardSeriesFilterRow( general_character, { CARD_SERIES.GENERAL, CARD_SERIES.DAILY } ))
    self.series_filter_group:SetSelected( selected_idx )

    -- Add a text search bar
    if SPECIAL_BUILD_TYPE < 1024 then
        self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_SEARCH" )
        self.sidebar:AddFilterWidget( Widget.CardSearchFilterBlock() )
    end

    -- Add a negotiation/battle filter
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.CARD_FILTER_MINIGAME" )
    self.minigame_filter_group = self.sidebar:AddFilterWidget( Widget.FilterRadioGroup() )
    self.series_battle = self.minigame_filter_group:AddFilter( Widget.CardMinigameFilterRow( Battle.Card, "event:/ui/card_codex/select_battle_cards" ) )
    self.series_negotiation = self.minigame_filter_group:AddFilter( Widget.CardMinigameFilterRow( Negotiation.Card, "event:/ui/card_codex/select_negotiation_cards" ) )
    self.minigame_filter_group:SetSelected( 1 )

    -- Add a rarity filter
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY" )
    self.rarity_filter_group = self.sidebar:AddFilterWidget( Widget.FilterOptionalRadioGroup() )
        :SetRowMode()
    self.rarity_filter_group:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.BASIC, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_BASIC_TT" ) )
    self.rarity_filter_group:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.COMMON, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_COMMON_TT" ) )
    self.rarity_filter_group:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.UNCOMMON, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_UNCOMMON_TT" ) )
    self.rarity_filter_group:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.RARE, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_RARE_TT" ) )
    self.rarity_filter_group:AddFilter( Widget.CardRarityFilterRow( CARD_RARITY.UNIQUE, LOC"UI.CARDCOMPENDIUM.CARD_FILTER_RARITY_UNIQUE_TT" ) )
    self.rarity_filter_group:SetSelected( 0 )

    -- Add a card-type filter
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_TYPE" )
    self.type_filter_group = self.sidebar:AddFilterWidget( Widget.FilterOptionalRadioGroup() )
        :SetRowMode( 3 )
    local filter_hostile = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_HOSTILE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_HOSTILE_TT" ) )
    local filter_diplomacy = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_DIPLOMACY, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_DIPLOMACY_TT" ) )
    local filter_manipulate = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_MANIPULATE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_MANIPULATE_TT" ) )
    -- local filter_opponent = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_OPPONENT, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_OPPONENT_TT" ) )
    self.negotiation_toggles = { filter_hostile, filter_diplomacy, filter_manipulate }
    local filter_melee = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_MELEE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_MELEE_TT" ) )
    local filter_skill = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_SKILL, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_SKILL_TT" ) )
    -- local filter_special = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_SPECIAL, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_SPECIAL_TT" ) )
    -- local filter_status = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FIGHT_STATUS, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FIGHT_STATUS_TT" ) )
    self.battle_toggles = { filter_melee, filter_skill }
    local filter_status = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.NEGOTIATION_STATUS, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_NEGOTIATION_STATUS_TT" ):AddType( CARD_TYPE.FIGHT_STATUS ) )
    local filter_item = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.ITEM, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_ITEM_TT" ) )
    local filter_parasite = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.PARASITE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_PARASITE_TT" ) )
    local filter_flourish = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.FLOURISH, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_FLOURISH_TT" ) )

    local filter_score = self.type_filter_group:AddFilter( Widget.CardTypeFilterRow( CARD_TYPE.SCORE, LOC"UI.CARDCOMPENDIUM.FILTER_TYPE_SCORE_TT" ) )
    self.shared_toggles = { filter_status, filter_item, filter_parasite, filter_score }
    self.type_filter_group:SetSelected( 0 )

    -- Add a sorting filter
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.CARD_FILTER_SORT" )
    self:AddSortingWidget()
end

function CardCompendium:AddSortingWidget()
    local sorting_options = {
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_NAME", fn = function( w1, w2 ) return w1.card_data:GetName() < w2.card_data:GetName() end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_COST_ASC", fn = function( w1, w2 ) return w1.card_data.cost < w2.card_data.cost end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_COST_DES", fn = function( w1, w2 ) return w1.card_data.cost > w2.card_data.cost end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_DAMAGE_ASC", fn = 
            function( w1, w2 )
                local max1 = w1.card_data.max_damage or w1.card_data.max_persuasion or 0
                local max2 = w2.card_data.max_damage or w2.card_data.max_persuasion or 0
                return max1 < max2
            end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_DAMAGE_DES", fn = 
            function( w1, w2 )
                local max1 = w1.card_data.max_damage or w1.card_data.max_persuasion or 0
                local max2 = w2.card_data.max_damage or w2.card_data.max_persuasion or 0
                return max1 > max2
            end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_TYPE", fn = function( w1, w2 )
            local type1 = GetCardType(w1.card_data)
            local type2 = GetCardType(w2.card_data)
            return type1 < type2
        end },
        { txt = LOC"UI.CARDCOMPENDIUM.SORT_OPTION_RARITY", fn = function( w1, w2 )
            local rarity1 = 0
            if w1.card_data.rarity == CARD_RARITY.BASIC then
                rarity1 = 1
            elseif w1.card_data.rarity == CARD_RARITY.COMMON then
                rarity1 = 2
            elseif w1.card_data.rarity == CARD_RARITY.UNCOMMON then
                rarity1 = 3
            elseif w1.card_data.rarity == CARD_RARITY.RARE then
                rarity1 = 4
            elseif w1.card_data.rarity == CARD_RARITY.UNIQUE then
                rarity1 = 5
            end

            local rarity2 = 0
            if w2.card_data.rarity == CARD_RARITY.BASIC then
                rarity2 = 1
            elseif w2.card_data.rarity == CARD_RARITY.COMMON then
                rarity2 = 2
            elseif w2.card_data.rarity == CARD_RARITY.UNCOMMON then
                rarity2 = 3
            elseif w2.card_data.rarity == CARD_RARITY.RARE then
                rarity2 = 4
            elseif w2.card_data.rarity == CARD_RARITY.UNIQUE then
                rarity2 = 5
            end
            
            if rarity1 == rarity2 then 
                return w1.card_data:GetName() < w2.card_data:GetName()
            end

            return rarity1 < rarity2
        end },
    }
    local selected_idx = 7 -- Rarity
    self.sidebar:AddSortingWidget( sorting_options, function( idx, option ) self:SortOrderSelected( option ) end, selected_idx )
end

function CardCompendium:SortOrderSelected( option )
    self.sorting_fn = option.fn
    if self.card_grid then
        self:LayoutContent()
    end
end

local function IsSeriesLocked( series )
    if series == nil then
        return false
    end

    -- This series is only locked if is used by at least one character, but not any unlocked characters use this series.
    local is_player_series = false
    for k, data in ipairs(GetAllPlayerBackgrounds()) do
        local def = Content.GetCharacterDef( data.data.player_agent )
        local uses_series = def and def.card_series and table.contains( def.card_series, series )
        is_player_series = is_player_series or uses_series

        if uses_series and not TheGame:GetGameProfile():IsCharacterLocked( data.id ) then
            -- Found an unlocked character that uses this series.
            return false
        end
    end

    return is_player_series
end

-- Gets all relevant cards and starts displaying them
function CardCompendium:RefreshContent( force_refresh )
    CardCompendium._base.RefreshContent()

    if self.cards ~= nil and not force_refresh then
        return -- Already constructed.
    end

    self.cards = {}
    self.last_focus = nil

    -- Refresh locked status
    for background_id, row in pairs(self.agent_filters) do
        local is_character_locked = not Screen.Compendium.ShowAllUnlocks() and TheGame:GetGameProfile():IsCharacterLocked( background_id )
        row:SetLocked( is_character_locked )
    end

    for background_id, row in pairs(self.agent_filters) do
        if self.card_counts[ background_id ] then
            row:SetProgress( self.card_counts[background_id] )
        end
    end
    self.series_battle:SetProgress( self.card_counts["BattleCardDef"] )
    self.series_negotiation:SetProgress( self.card_counts["NegotiationCardDef"] )

    self.cards = fun(self.card_defs)
        :map(function(card_def) 
            if card_def._classname == "BattleCardDef" then
                return Battle.Card( card_def.id )
            end
            if card_def._classname == "NegotiationCardDef" then
                return Negotiation.Card( card_def.id )
            end
        end)
        :value()

    -- Update lock status on the cards
    for id, card in pairs( self.cards ) do
        local is_locked = true
    
        if Screen.Compendium.ShowAllUnlocks() then
            is_locked = false
        elseif Content.IsUnlock(card.def.id) then
            is_locked = not TheGame:GetGameProfile():HasUnlock( card.def.id )
        else
            is_locked = IsSeriesLocked( card.def.series )
        end

        self.cards[id].card_locked = is_locked
    end

    -- Update filtering
    self.sidebar:RefreshFilters()

    if force_refresh then

    end

    return self
end

function CardCompendium:UpdateCount()
    -- Reset counts
    self.card_counts = {}
    self.card_counts["UNLOCKED"] = 0
    self.card_counts["UNLOCKABLE"] = 0

    for k, card_def in pairs( self.card_defs ) do

        -- If this series isn't being counted yet, do it
        if not self.card_counts[card_def.series] then 
            self.card_counts[card_def.series] = {
                UNLOCKED = 0,
                UNLOCKED_CARDS = {},
                UNLOCKABLE = 0,
                UNLOCKABLE_CARDS = {},
                STARTING = 0,
                STARTING_CARDS = {},
                TOTAL = 0,
                TOTAL_CARDS = {},
            }
        end

        -- If this card's minigame isn't being counted yet, do it
        if not self.card_counts[card_def._classname] then
            self.card_counts[card_def._classname] = {
                UNLOCKED = 0,
                UNLOCKED_CARDS = {},
                UNLOCKABLE = 0,
                UNLOCKABLE_CARDS = {},
                STARTING = 0,
                STARTING_CARDS = {},
                TOTAL = 0,
                TOTAL_CARDS = {},
            }
        end

        -- Add this to the totals
        self.card_counts[card_def.series].TOTAL = self.card_counts[card_def.series].TOTAL + 1
        table.insert( self.card_counts[card_def.series].TOTAL_CARDS, card_def.id )
        self.card_counts[card_def._classname].TOTAL = self.card_counts[card_def._classname].TOTAL + 1
        table.insert( self.card_counts[card_def._classname].TOTAL_CARDS, card_def.id )

        -- Check if this is an unlockable card
        local is_unlock = Content.IsUnlock( card_def.id )
        local is_unlocked = false
        if is_unlock then
            self.card_counts["UNLOCKABLE"] = self.card_counts["UNLOCKABLE"] + 1
            self.card_counts[card_def.series].UNLOCKABLE = self.card_counts[card_def.series].UNLOCKABLE + 1
            table.insert( self.card_counts[card_def.series].UNLOCKABLE_CARDS, card_def.id )
            self.card_counts[card_def._classname].UNLOCKABLE = self.card_counts[card_def._classname].UNLOCKABLE + 1
            table.insert( self.card_counts[card_def._classname].UNLOCKABLE_CARDS, card_def.id )
            is_unlocked = TheGame:GetGameProfile():HasUnlock( card_def.id ) or Screen.Compendium.ShowAllUnlocks()
            if is_unlocked then
                self.card_counts["UNLOCKED"] = self.card_counts["UNLOCKED"] + 1
                self.card_counts[card_def.series].UNLOCKED = self.card_counts[card_def.series].UNLOCKED + 1
                table.insert( self.card_counts[card_def.series].UNLOCKED_CARDS, card_def.id )
                self.card_counts[card_def._classname].UNLOCKED = self.card_counts[card_def._classname].UNLOCKED + 1
                table.insert( self.card_counts[card_def._classname].UNLOCKED_CARDS, card_def.id )
            end
        else
            self.card_counts[card_def.series].STARTING = self.card_counts[card_def.series].STARTING + 1
            table.insert( self.card_counts[card_def.series].STARTING_CARDS, card_def.id )
            self.card_counts[card_def._classname].STARTING = self.card_counts[card_def._classname].STARTING + 1
            table.insert( self.card_counts[card_def._classname].STARTING_CARDS, card_def.id )
        end

    end
end

-- Returns the number of unlocked cards and the number of unlockable cards, to be displayed on the nav bar
function CardCompendium:GetCount()
    if not self.card_counts then
        self:UpdateCount()
    end

    return Screen.Compendium.ShowAllUnlocks() and self.card_counts["UNLOCKABLE"] or self.card_counts["UNLOCKED"], self.card_counts["UNLOCKABLE"]
end

function CardCompendium:RefreshBottomText()
    self.parent:SetBottomText( loc.format( LOC "UI.CARDCOMPENDIUM.CARD_COUNT", self.card_grid:GetCardCount() ))
end

-- Lays its content out
function CardCompendium:LayoutContent()
    CardCompendium._base.LayoutContent()

    local MARGIN = 20
    local SPACING = 20
    local CARD_Y_SPACING = 50

    local CARD_WIDTH = 370
    local CARD_HEIGHT = 542

    local pos = 0
    local first_visible_card

    local RMARGIN = 60 -- for the scroll bar
    local SLOT_W, SLOT_H = CARD_WIDTH * CARD_SCALE, CARD_HEIGHT * CARD_SCALE
    
    
    local xslots = math.min( #self.card_grid:GetCardWidgets(), math.floor( (self.content_w-RMARGIN) / ( SLOT_W + SPACING ) ) )
    local xstart = (-self.content_w/2) + (self.content_w - RMARGIN - ((xslots - 1) * (SLOT_W + SPACING))) / 2

    self.card_grid:SetSortingFn( self.sorting_fn )

    -- Layout things!
    self.card_grid:LayoutCards()
        :LayoutBounds( "after", "top", self.sidebar )
        :Offset( 8-CONTENT_LEFT_MARGIN, -1 )

    self:RefreshBottomText()
end


function CardCompendium:GetDefaultFocus( last_focus )
    if self.sidebar_focus then
        return self.sidebar:GetDefaultFocus()
    elseif is_instance( last_focus, Widget.CardWidget ) then
        return last_focus
    end
    return self.card_grid:GetDefaultFocus()
end

function CardCompendium:OnFiltersChanged( filters )

    if self.series_battle:IsToggled() then
        -- Hide the negotiation-type filters
        for k, w in ipairs( self.negotiation_toggles ) do w:Hide() end
        for k, w in ipairs( self.battle_toggles ) do w:Show() end
        for k, w in ipairs( self.shared_toggles ) do w:Show() end
    end
    if self.series_negotiation:IsToggled() then
        -- Hide the battle-type filters
        for k, w in ipairs( self.negotiation_toggles ) do w:Show() end
        for k, w in ipairs( self.battle_toggles ) do w:Hide() end
        for k, w in ipairs( self.shared_toggles ) do w:Show() end
    end
    self.type_filter_group:Layout()
    self.sidebar:Layout()

    self:FilterCardWidgets(filters )
    self:LayoutContent()
end

function CardCompendium:OnSearch( filter, is_toggled )
    self.sidebar:SetEnabled( not is_toggled )
    self:FilterCardWidgets { filter }
    self:LayoutContent()
end

function CardCompendium:FilterCardWidgets( filters )

    local filtered_cards = fun(self.cards)
        :filter(function(card) 
            for j, filter in ipairs( filters ) do
                -- If this filter rejects the card, hide the card widget
                if filter:Filter( card ) == false then return false end
            end
            return true
        end)
        :value()

    self.card_grid:Refresh(filtered_cards)
end

------------------------------------------------------------------------------------------
-- Shows a paginated list of card upgrades

local CardUpgradesBlock = class("Widget.CardUpgradesBlock", Widget)

local SIDE_SPACING = 40

    -- self.upgrades_widget = self:AddChild( Widget.CardUpgradesBlock( CONTENT_W, 240, UPGRADE_COLUMNS, 20, 0.35 ) )
    -- self.upgrades_widget:Refresh( upgrades )

function CardUpgradesBlock:init( width, height, columns, spacing, card_scale )
    CardUpgradesBlock._base.init(self)

    -- Add our content block
    self.bg_border = 45
    self.bg_border_scale = 0.5
    self.content_bg = self:AddChild( Widget.Panel( assets.wideblock ) )
    self.content_bg:SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )

    -- Size up our content block
    local nineslice_w, nineslice_h = width + self.bg_border*2*self.bg_border_scale, height + self.bg_border*2*self.bg_border_scale
    self.content_bg:SetSize( nineslice_w, nineslice_h )

    self.column_count = columns or 3
    self.card_spacing = spacing or 10
    self.block_width = ( width - (self.column_count-1)*self.card_spacing - SIDE_SPACING*2 )/self.column_count
    self.block_height = height - self.card_spacing*2
    self.card_scale = card_scale or 0.25

    -- And left/right arrows
    local arrow_size = 32
    self.arrow_left = self:AddChild( Widget.ImageButton( assets.gameover_arrow_left, function() self:SetPage( self.current_page - 1 ) end ) )
        :SetHiddenBoundingBox( true )
        :SetSize( arrow_size, arrow_size )
        :LayoutBounds( "left", "center", self.content_bg )
        :Bloom( 0.1 )
        :Hide()
    self.arrow_right = self:AddChild( Widget.ImageButton( assets.gameover_arrow, function() self:SetPage( self.current_page + 1 ) end ) )
        :SetHiddenBoundingBox( true )
        :SetSize( arrow_size, arrow_size )
        :LayoutBounds( "right", "center", self.content_bg )
        :Bloom( 0.1 )
        :Hide()

    -- Container for the pagination dots
    self.pagination = self:AddChild( Widget() )
        :Hide()

    -- Divider below the title
    self.divider_top = self:AddChild( Widget.Image( assets.gameover_divider, width*0.4, 2 ) )
        :LayoutBounds( "left", "top", self.content_bg )
        :SetTintAlpha( 0.4 )
        :Offset( 10, 0 )
        :Bloom( 0.1 )

    -- Divider above the pagination dots
    self.divider_bottom = self:AddChild( Widget.Image( assets.gameover_divider, width*0.7, 2 ) )
        :LayoutBounds( "center", "bottom", self.content_bg )
        :Offset( 0, 0 )
        :Bloom( 0.1 )
        :Hide()

    -- Add a container for our blocks
    self.upgrades_container = self:AddChild( Widget() )
end

function CardUpgradesBlock:OnAdded()
    TheGame:GetEvents():ListenForEvent( "focus_changed", self, self.OnFocusChanged )
end

function CardUpgradesBlock:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function CardUpgradesBlock:GetFocusedIndex()
    for k, box in ipairs( self.upgrades_container.children ) do
        if box:IsShown() and box.focus then return k end
    end
    return 1
end

function CardUpgradesBlock:SetFocusedIndex( idx )
    if idx < 1 then idx = #self.upgrades_container.children end
    if idx > #self.upgrades_container.children then idx = 1 end
    for k, box in ipairs( self.upgrades_container.children ) do
        if k == idx then box:SetFocus() end
    end
end

function CardUpgradesBlock:OnFocusChanged( event_name, prev_focus, new_focus )
    if new_focus and self:IsAncestorOf( new_focus ) then
        if new_focus and new_focus.index and new_focus.page and new_focus.page ~= self.current_page then
            self:SetPage( new_focus.page )
        end
    end
end

function CardUpgradesBlock:OnFocusMove( dir )
    if dir == "left" or dir == "right" then
        local focus = self:GetFE():GetFocusWidget()
            if focus and self:IsAncestorOf( focus ) then
            local next_focus = focus:GetFocusDir( dir )
            if next_focus then
                next_focus:SetFocus()
                return true
            end
        end
    else
        self._base.OnFocusMove( self, dir )
    end
end

function CardUpgradesBlock:Refresh( upgrades )

    -- Remove old Upgrades
    self.upgrades_container:DestroyAllChildren()

    if upgrades and #upgrades>0 then
        -- Add new ones
        for k, upgrade in ipairs( upgrades ) do self:AddUpgrade( upgrade ) end

        self:Show()
    else
        self:Hide()
    end

    self:Layout()
end

function CardUpgradesBlock:AddUpgrade( card_def )
    -- Create card widget and set its scale
    local card_class, card_widget_class, card_tooltip_class
    if card_def._classname == "BattleCardDef" then
        card_class = Battle.Card
        card_widget_class = Widget.BattleCard
        card_tooltip_class = Widget.TooltipBattleCard
    end
    if card_def._classname == "NegotiationCardDef" then
        card_class = Negotiation.Card
        card_widget_class = Widget.NegotiationCard
        card_tooltip_class = Widget.TooltipNegotiationCard
    end
    local card = card_class( card_def.id )
    local widget = self.upgrades_container:AddChild( card_widget_class() )
        :RefreshCard( card, nil, Widget.CardWidget.STYLE.UPGRADE )
        :SetScale( self.card_scale )
        :ShowToolTipOnFocus( true )
        :SetShineEnabled(true)

    -- Handle events
    widget.OnGainFocus = function() widget._base.OnGainFocus(widget) self:UpdateCardImageState( widget ) end
    widget.OnLoseFocus = function() widget._base.OnLoseFocus(widget) self:UpdateCardImageState( widget ) end
    widget.OnGainHover = function() widget._base.OnGainHover(widget) self:UpdateCardImageState( widget ) end
    widget.OnLoseHover = function() widget._base.OnLoseHover(widget) self:UpdateCardImageState( widget ) end
end

function CardUpgradesBlock:UpdateCardImageState( widget )
    if ( widget.hover or widget.focus ) then
        widget:ScaleTo( nil, self.card_scale*1.5, 0.2, easing.outQuad )
        widget:MoveTo( nil, 45, 0.2, easing.outQuad )
        widget:SetOutline( widget.is_negotiation and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT, true )
    else
        widget:ScaleTo( nil, self.card_scale, 0.3 )
        widget:MoveTo( nil, 0, 0.2 )
        widget:SetOutline( nil, true )
    end
end

function CardUpgradesBlock:ReNav()
    local previous_v = nil
    for k, v in ipairs( self.upgrades_container.children ) do
        if previous_v then v:SetFocusDir( "left", previous_v, true ) end
        previous_v = v
    end
    if self.upgrades_container.children[1] then
        self.upgrades_container.children[1]:SetFocusDir( "left", previous_v, true )
    end
end

function CardUpgradesBlock:Layout()

    -- Align the blocks
    for k, block in ipairs( self.upgrades_container.children ) do
        -- Assign this item a spot. If there are three items per page,
        -- the first will get 0, the second will get 1, the third will get 2, 
        -- the fourth will get 0 again, and so on
        local spot = ( k - 1 ) % self.column_count
        -- Calculate the x-position based on what spot this block is at, including spacings
        local pos_x = self.block_width * spot + self.card_spacing * spot
        -- Set the position
        block:SetPos( pos_x, 0 )
        -- Inform the block of what page it's on
        block.page = math.ceil( k / self.column_count )
        block.index = k
    end

    -- Clear any old pagination markers
    self.pagination:DestroyAllChildren()

    local page_count = math.ceil ( #self.upgrades_container.children / self.column_count )
    if page_count > 1 then
        -- Add brand new ones
        for k = 1, page_count do
            local dot = self.pagination:AddChild( Widget.Clickable( ) )
            dot.hitbox = dot:AddChild( Widget.SolidBox( 67, 30, 0xFF000000 ))
            dot.image = dot:AddChild( Widget.Image( assets.motd_page, 67, 8 ) ):Bloom( 0.1 )
            dot:SetOnClickFn( function() self:SetPage( k ) end )
            dot:SetHoverFn( function( hover ) if hover then dot:ScaleTo( nil, 1.05, 0.1 ) dot:Bloom(0.2) else dot:ScaleTo( nil, 1, 0.1 ) dot:Bloom(0.1) end end )
        end
        -- Align them
        self.pagination:StackChildrenRow( 15 )
        self.pagination:LayoutBounds( "center", "below", self.content_bg ):Offset( 0, 12 )

        self.pagination:Show()
        self.arrow_left:Show()
        self.arrow_right:Show()
        self.divider_bottom:Show()
    else
        self.pagination:Hide()
        self.arrow_left:Hide()
        self.arrow_right:Hide()
        self.divider_bottom:Hide()
    end

    self:ReNav()

    -- Set the current page
    self:SetPage( 1 )
end

function CardUpgradesBlock:SetPage( page )
    local page_count = math.ceil ( #self.upgrades_container.children / self.column_count )
    page = page or 1
    if page > page_count then page = 1 end
    if page < 1 then page = page_count end

    -- Update the pagination dots
    for k, dot in ipairs( self.pagination.children ) do
        if k == page then
            -- Current page
            dot:SetTintAlpha( 1 )
        else
            -- Not the current page..
            dot:SetTintAlpha( 0.3 )
        end
    end

    -- Update page widgets visibility
    for k, block in ipairs( self.upgrades_container.children ) do
        if k > ( page*self.column_count - self.column_count ) and k <= ( page*self.column_count ) then
            block:Show()
        else
            block:Hide()
        end
    end

    -- Align vertically
    self.upgrades_container:LayoutBounds( "left", "center", self.content_bg ):Offset( SIDE_SPACING, 0 )

    self.current_page = page
end

function CardUpgradesBlock:PrepareAnimation()
    self.is_done_animating = false
    self:SetTintAlpha( 0 )
    for k, block in ipairs( self.upgrades_container.children ) do
        block:SetTintAlpha( 0 )
        if block.exp_floater then block.exp_floater:SetTintAlpha( 0 ) end
    end

    return self
end

function CardUpgradesBlock:AnimateIn( on_done_fn )
    local fade_in_rate = 8
    local block_fade_in_rate = 3

    -- Fade in
    AUDIO:PlayEvent("event:/ui/game_over_screen/approach_list_appears")
    while self:GetAlpha() < 1 do
        local dt = coroutine.yield()
        self:SetTintAlpha( self:GetAlpha() + fade_in_rate*dt )
        if self.fast then
            break
        end
    end
    -- Not the blocks yet
    for k, block in ipairs( self.upgrades_container.children ) do
        block:SetTintAlpha( 0 )
    end

    -- Fade in each approach block
    local page = 1
    local alpha = 0
    for k, block in ipairs( self.upgrades_container.children ) do

        -- Fade in this block
        AUDIO:PlayParamEvent( "event:/ui/game_over_screen/approach_stat_appears", "position", block:GetPanPos() )
        alpha = 0
        while alpha < 1 and not self.fast do
            local dt = coroutine.yield()
            alpha = alpha + block_fade_in_rate*dt
            block:SetTintAlpha( alpha )
        end
        block:SetTintAlpha( 1 )
        if block.exp_floater then block.exp_floater:AnimateIn() end
        if on_done_fn then on_done_fn( block.exp ) end

        -- Check if we should change page
        if k%self.column_count == 0 then 
            self:SetPage( page+1, true )
        end

    end

    self.is_done_animating = true
    self:SetTintAlpha( 1 )

    return self
end

function CardUpgradesBlock:SkipAnimation()
    self.fast = true
end

function CardUpgradesBlock:IsDoneAnimating()
    return self.is_done_animating
end
