require "ui/widgets/lazyloadwidget"
require "ui/widgets/cardwidget"
local fun = require "util/fun"

local CardGrid = class( "Widget.CardGrid", Widget )

CardGrid.CONTROL_MAP = 
{
    {
        control = Controls.Digital.GAMEPAD_Y,
        fn = function(self)
            self.filter_button:Click()
            return true
        end,
        test = function(self)
            return self.filter_button ~= nil
        end,
        hint = function(self, left, right)
            table.insert(left, loc.format(LOC"UI.CONTROLS.OPEN_FILTERS", Controls.Digital.GAMEPAD_Y))
        end
    },

}

local DEFAULT_CARD_SCALE = 0.55
local DEFAULT_SPACING_HORIZONTAL = 0
local DEFAULT_SPACING_VERTICAL = 30
local DEFAULT_HEADER_LEFT_PADDING = 15

function CardGrid:init( w, h )
    CardGrid._base.init( self )
    self.loaders = {}

    self.width = w or 200
    self.height = h or 200
	self.top_padding = 60 -- Scroll panel's top virtual margin. Contents will be offset from this, but still scroll to the clipping area
	self.right_padding = 0 -- Space between the left edge of the clipping area and the first bit of content
	self.bottom_padding = 250 -- Scroll panel's bottom virtual margin. Scrolling contents will get this extra empty space at the bottom of the scroll area
	self.left_padding = 0 -- Space between the right edge of the content and the clipping area
	self.card_scale = DEFAULT_CARD_SCALE
	self.card_spacing_horizontal = DEFAULT_SPACING_HORIZONTAL
	self.card_spacing_vertical = DEFAULT_SPACING_VERTICAL
	self.header_left_padding = DEFAULT_HEADER_LEFT_PADDING -- So the titles visually align with the cards

    -- self.hitbox = self:AddChild( Widget.SolidBox( self.width, self.height, 0xFF00FF20 ) )

    self.scroll_panel = self:AddChild( Widget.ScrollPanel() )
        :SetVirtualTopMargin( self.top_padding )
        :SetVirtualBottomMargin( self.bottom_padding )
        :SetScrollBarOuterMargin( 0 )
        :ShowBar( SCROLLBAR.NEVER )

    -- Header, which contains text elements above the cards
    self.header_content = self.scroll_panel:AddScrollChild( Widget() )
    self.title = self.header_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE, "TITLE" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :SetAutoSize( self.width )
        :SetWordWrap( true )
		:Bloom( 0.15 )
    self.title_count = self.header_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TABS ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :SetAutoSize( self.width )
        :Bloom( 0.1 )
    self.subtitle = self.header_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_SUBTITLE, "SUBTITLE" ) )
        :SetGlyphColour( UICOLOURS.TITLE )
        :SetAutoSize( self.width )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    self.text = self.header_content:AddChild( Widget.Label( "body", FONT_SIZE.BODY_TEXT ) )
		:SetAutoSize( self.width )
		:SetWordWrap( true )
		:LeftAlign()
		:SetGlyphColour( 0xCBFFEFff )
        :Bloom( 0.1 )
    
    -- Container for the cards, shown under the header
    self.cards_container = self.scroll_panel:AddScrollChild( Widget() )
	self.card_layout_scale = 1
    
    self:SetTitles()
end

function CardGrid:SetFilterButton( tooltip_text, on_click_fn )
    self.filter_button = self.header_content:AddChild( Widget.ImageButton( engine.asset.Texture("UI/btn_filter.tex"), on_click_fn ) )
        :SetSize( FONT_SIZE.SCREEN_TITLE*0.8, FONT_SIZE.SCREEN_TITLE*0.8 )
        :SetTintColour( self.header_colour or UICOLOURS.SUBTITLE )
        :Bloom( 0.15 )
    if tooltip_text then self.filter_button:SetToolTip( tooltip_text ) end

    self:LayoutCards()
    return self
end

function CardGrid:SetColour( colour )
	self.header_colour = colour or UICOLOURS.SUBTITLE
	self.title:SetGlyphColour( self.header_colour )
    self.title_count:SetGlyphColour( self.header_colour )
	self.subtitle:SetGlyphColour( self.header_colour )
    if self.filter_button then self.filter_button:SetTintColour( self.header_colour ) end
	return self
end

function CardGrid:SetTitles( title, subtitle, desc )
	if title then
        self.title:SetText( title ):Show()
        self.title_count:SetText( loc.format( LOC "UI.SHOW_DECK_SCREEN.CARD_COUNT", #self.cards )):Show()
    else
        self.title_count:Hide()
        self.title:Hide()
    end
	if subtitle then self.subtitle:SetText( subtitle ):Show() else self.subtitle:Hide() end
	if desc then self.text:SetText( desc ):Show() else self.text:Hide() end
    self:LayoutCards()
    return self
end

function CardGrid:GetTitleWidget()
    return self.title
end

function CardGrid:SetTitle( desc )
    if title then self.title:SetText( title ):Show() else self.title:Hide() end
    self:LayoutCards()
    return self
end

function CardGrid:GetSubtitleWidget()
    return self.subtitle
end

function CardGrid:SetSubtitle( desc )
    if subtitle then self.subtitle:SetText( subtitle ):Show() else self.subtitle:Hide() end
    self:LayoutCards()
    return self
end

function CardGrid:GetDescWidget()
    return self.desc
end

function CardGrid:SetDesc( desc )
    if desc then self.text:SetText( desc ):Show() else self.text:Hide() end
    self:LayoutCards()
    return self
end

function CardGrid:ShowBar( viz )
	self.scroll_panel:ShowBar( viz )
    return self
end

function CardGrid:Scroll(...)
    self.scroll_panel:Scroll(...)
    return self
end

function CardGrid:ScrollUp()
    self.scroll_panel:ScrollUp()
    return self
end

function CardGrid:ScrollDown()
    self.scroll_panel:ScrollDown()
    return self
end

function CardGrid:SetScrollEnabled( enabled )
    self.scroll_panel:SetEnabled( enabled )
    return self
end

function CardGrid:SetCardScale( scale )
    self.card_scale = scale or DEFAULT_CARD_SCALE
	for k, w in ipairs( self.cards_container.children ) do
    	w:SetScale( self.card_scale )
    end
    return self
end

function CardGrid:GetListHeight()
    local h_w, h_h = self.header_content:GetScaledSize()
    local c_w, c_h = self.cards_container:GetScaledSize()
    return h_h+c_h
end

function CardGrid:SetSize( w, h )
	if self.width ~= w or self.height ~= h then
	    self.width = w or self.width or 200
	    self.height = h or self.height or 200
	    if self.hitbox then self.hitbox:SetSize( self.width, self.height ) end
	    self.scroll_panel:SetSize( self.width, self.height )
	    self:LayoutCards()
	end
    return self
end

function CardGrid:SetTopPadding( padding )
	self.top_padding = padding or 60
	self.scroll_panel:SetVirtualTopMargin( self.top_padding )
    return self
end

function CardGrid:SetRightPadding( padding )
	self.right_padding = padding or 0
	self:SetSize( self.width, self.height )
    return self
end

function CardGrid:SetBottomPadding( padding )
	self.bottom_padding = padding or 250
	self.scroll_panel:SetVirtualBottomMargin( self.bottom_padding )
    return self
end

function CardGrid:SetLeftPadding( padding )
	self.left_padding = padding or 0
	self:SetSize( self.width, self.height )
    return self
end

function CardGrid:OnVizChange(viz)
    if viz then
        self:StartUpdating()
    else
        self:StopUpdating()
    end
end

-- Receives a list of card objects as a parameter
function CardGrid:Refresh( cards, style )
	-- Destroy old cards, if any
	self.cards_container:DestroyAllChildren()

	-- Clear lists
    self.cards = {}
    self.card_widgets = {}
    self.loaders = {}
    self.selected_card = nil

    -- Go through the new set
    if cards then
        for i, card in pairs( cards ) do
            self:AddCard(card, style)
        end
    end
    if self.title:IsShown() then
        self.title_count:SetText( loc.format( LOC "UI.SHOW_DECK_SCREEN.CARD_COUNT", #self.cards )):Show()
    end

    self:LayoutCards()
    return self
end

function CardGrid:AddCard(card, style)
    if table.arrayfind(self.cards, card) ~= nil then
        return
    end

    -- Add card to the list
    table.insert( self.cards, card )

    -- Create loader widget
    local loader = self.cards_container:AddChild( Widget.LazyLoadWidget( Widget.CardWidget.Width, Widget.CardWidget.Height, function( loader_widget )
        -- Load the card when the loading widget becomes visible
        local widget
        if is_instance( card, Battle.Card ) then
            -- Battle Card!
            widget = loader_widget:AddChild( Widget.BattleCard() )
                :RefreshCard( card, nil, style )
        else
            -- Negotiation Card!
            widget = loader_widget:AddChild( Widget.NegotiationCard() )
                :RefreshCard( card, nil, style )
        end

        -- Create references to the card
        table.insert( self.card_widgets, widget )
        loader_widget.card_widget = widget

        if widget then
            widget:ShowToolTipOnFocus()

            -- Set the card to scroll mode
            widget:SetScrollListMode( true )

            -- Add the widget that surfaces relevant info under the card
            widget.sub_card_info = widget:AddChild( Widget.SubCardInfo( widget ) )
            widget.sub_card_info:SetHiddenBoundingBox( true )
            widget.sub_card_info:LayoutBounds( "center", "below", widget.bg ):Offset( 0, -8 )

            -- Set the widget's state functions
            widget.OnGainFocus = function() widget._base.OnGainFocus(widget) self.selected_card = widget self:UpdateCardImageState( widget ) end
            widget.OnLoseFocus = function() widget._base.OnLoseFocus(widget) self:UpdateCardImageState( widget ) end
            widget.OnGainHover = function() widget._base.OnGainHover(widget) self:UpdateCardImageState( widget ) end
            widget.OnLoseHover = function() widget._base.OnLoseHover(widget) self:UpdateCardImageState( widget ) end

             -- Don't set a callback unless one is defined, otherwise the card will have a MENU_ACCEPT hint in gamepad.
            if self.card_click_fn then
                widget:SetOnClick( function() self:OnCardClick( widget ) end )
            end
            widget:SetCardInfoPopupMode( self.card_info_popup_mode )
    
            -- Prepare this widget's outline
            if widget.card:UpgradeReady() then
                widget:SetOutline( UICOLOURS.OBJECTIVE, true )
            else
                -- This primes the outline to green, so it doesn't focus from white to green on hover later
                widget:SetOutline( UICOLOURS.FOCUS, false )
                widget:SetOutline( nil, false )
            end

            if loader_widget.card_locked or card.card_locked then
                card.card_locked = true
                loader_widget.card_locked = true
                widget:SetLocked( true )
            end
        end
    end )
    )
    table.insert(self.loaders, loader)
    loader:SetScale( self.card_scale )-- Set correct scale for the current screen mode
    loader.card_data = card
end

function CardGrid:GetCardLoaders()
	return self.cards_container.children
end

function CardGrid:GetCardWidgets()
	return self.card_widgets
end

function CardGrid:GetCardCount()
    return #self.cards
end

function CardGrid:RemoveCard( widget, fn )
	for k, w in ipairs( self.cards_container.children ) do
		if w.card_data == widget.card then
			local duration = 0.4
			local pos_x, pos_y = w:GetPos()
			self:RunUpdater( Updater.Series{
				Updater.Wait( 0.1 ),
				Updater.Parallel{
					Updater.Ease( function(v) w:SetTintAlpha(v) end, 1, 0, duration, easing.outQuad),
                    Updater.Ease2D( function(x,y) w:SetPos(x,y) end, pos_x, pos_y, pos_x, pos_y+40, duration, easing.outQuad),
				},
				Updater.Wait( 0.6 ),
				Updater.Do( function()
					if w and not w.removed then w:Remove() end
					if fn then fn() end
				end ),
			} )
		end
	end
end

function CardGrid:OnUpdate(dt)
    CardGrid._base.OnUpdate(self, dt)
    local to_remove = {}
    for k,v in ipairs(self.loaders) do
        if v:CheckForLoad() then
            to_remove[v] = true
        end
    end

    if next(to_remove) then
        self.loaders = fun(self.loaders)
                            :filter(function(v) return not to_remove[v] end)
                            :value()
    end

end

-- Define a function that gets a card widget as argument whenever the hover/focus state of that widget changes 
function CardGrid:SetCardImageStateFn( fn )
	self.card_image_state_fn = fn
	return self
end

function CardGrid:UpdateCardImageState( widget )
    if widget.enabled and ( widget.hover or widget.focus ) then

        --if widget.focus then self:OnCardSelect( widget, true ) end

        if is_instance( widget, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseover, "position", widget:GetPanPos())
        elseif is_instance( widget, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseover, "position", widget:GetPanPos())
        end

        widget:SetOutline( UICOLOURS.FOCUS, true )
        widget.parent:ScaleTo( nil, self.card_scale*1.05, 0.2, easing.outQuad )
    else

        if is_instance( widget, Widget.BattleCard ) then
            AUDIO:PlayParamEvent(SoundEvents.battle_deck_card_mouseoff, "position", widget:GetPanPos())
        elseif is_instance( widget, Widget.NegotiationCard ) then
            AUDIO:PlayParamEvent(SoundEvents.negotiation_deck_card_mouseoff, "position", widget:GetPanPos())
        end

        widget:SetOutline( widget.card:UpgradeReady() and UICOLOURS.OBJECTIVE or nil, true )
        widget.parent:ScaleTo( nil, self.card_scale, 0.2, easing.outQuad )
    end

	if self.card_image_state_fn then self.card_image_state_fn( widget ) end
end

function CardGrid:SetOnCardClickFn( fn )
	self.card_click_fn = fn
	return self
end

function CardGrid:SetCardInfoPopupMode( mode )
    self.card_info_popup_mode = mode
    for k, w in ipairs( self.cards_container.children ) do
        widget:SetCardInfoPopupMode( self.card_info_popup_mode )
    end
    return self
end

function CardGrid:OnCardClick( widget )
	if widget and self.card_click_fn then
    	self.card_click_fn( widget )
    end
end

function CardGrid:SetSortingFn( fn )
	self.sorting_fn = fn
	return self
end

function CardGrid:LayoutCards()
    -- Calculate available width
    local available_w = self.width - self.left_padding - self.right_padding

	-- Layout titles
    self.title:SetAutoSize( available_w )
    self.title_count:LayoutBounds( "after", "center", self.title ):Offset( 6, 0 )
	self.subtitle:SetAutoSize( available_w ):LayoutBounds( "left", self.subtitle:IsShown() and "below" or "bottom", self.title ):Offset( 0, self.subtitle:IsShown() and SPACING.TITLE_SUBTITLE or 0 )
	self.text:SetAutoSize( available_w ):LayoutBounds( "left", self.text:IsShown() and "below" or "bottom", self.subtitle )
    self.header_content:LayoutBounds( "left", "top", self.left_padding, 0 )

	-- Sort cards appropriately first
    local function SortDeck( w1, w2 )
        if w1.sort_index and w2.sort_index then
            return w1.sort_index < w2.sort_index
        end
        -- Upgrade-ready cards should show at the top
        if w1.card_data:UpgradeReady() and not w2.card_data:UpgradeReady() then return true end
        if w2.card_data:UpgradeReady() and not w1.card_data:UpgradeReady() then return false end

        -- Otherwise, order them by ID, so they're in alphabetical order, with same name cards together
        return w1.card_data.id < w2.card_data.id
    end
    table.stable_sort( self.cards_container.children, self.sorting_fn or SortDeck )

    -- Maintain sort order, so if cards get upgraded, they don't shift around.
    for i, w in ipairs( self.cards_container.children) do
        w.sort_index = i
    end

    -- Scale cards
    for k, w in ipairs( self.cards_container.children ) do
    	w:SetLayoutScale( self.card_layout_scale )
    end

    -- Figure out how many card columns fit in our container
    local card_width, column_w, columns
    if self.cards_container.children and #self.cards_container.children > 0 then
    	card_width = self.cards_container.children[1]:GetScaledSize()
        column_w = card_width+self.card_spacing_horizontal
    	columns = math.max( math.floor( available_w / column_w ), 1 )

    	-- Layout cards
    	self.cards_container:LayoutChildrenInGrid( columns, { h = self.card_spacing_horizontal, v = self.card_spacing_vertical } )
   	end
    self.cards_container:LayoutBounds( "left", "below", self.header_content ):Offset( -self.header_left_padding, -SPACING.M1 )

    -- Position filter button
    if self.filter_button and column_w and columns then
        self.filter_button:LayoutBounds( "before", "center", self.title )
            :Offset( column_w*columns/LAYOUT_SCALE[TheGame:FE():GetScreenMode()] - 40, -2 )
            :SetShown( TheGame:FE():GetControlMode() ~= CONTROL_MODE.GAMEPAD )
    elseif self.filter_button then
        self.filter_button:Hide()
    end

    self.scroll_panel:RefreshView()
    
	return self
end

function CardGrid:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function CardGrid:OnControlModeChange( cm, device_id )
    CardGrid._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD
    or cm == CONTROL_MODE.TOUCH then
        if self.filter_button then self.filter_button:Show() end
    end
    if cm == CONTROL_MODE.GAMEPAD then
        if self.filter_button then self.filter_button:Hide() end
        if self.selected_card and self:IsVisible() then self.selected_card:SetFocus() end
    end
end

function CardGrid:OnScreenModeChange( sm )
    CardGrid._base.OnScreenModeChange( self, sm )

    if not self.screen_mode or self.screen_mode ~= sm then
    	self.screen_mode = sm
	    self.header_content:SetLayoutScale( LAYOUT_SCALE[sm] )
	    self.card_layout_scale = LAYOUT_SCALE[sm]
	    self.header_left_padding = DEFAULT_HEADER_LEFT_PADDING * LAYOUT_SCALE[sm]
		self.card_spacing_horizontal = DEFAULT_SPACING_HORIZONTAL * LAYOUT_SCALE[sm]
		self.card_spacing_vertical = DEFAULT_SPACING_VERTICAL * LAYOUT_SCALE[sm]

	    self:LayoutCards()
	end
end

function CardGrid:GetDefaultFocus()
    if self.selected_card then return self.selected_card end
    for i, widget in ipairs( self.cards_container.children ) do
        if widget.card_widget then
            return widget.card_widget
        end
    end
end

function CardGrid:PrepareAnimation()
	for i, widget in ipairs( self.cards_container.children ) do
    	widget:SetTintAlpha( 0 )
    end
	return self
end

function CardGrid:AnimateIn()
    -- Prepare card fade-in sequence
    local card_anim_duration_total = 0.2
    local card_number = math.min( #self.cards_container.children, 10 )
    local card_anim_duration = 0.1
    local delay_per_card = card_anim_duration_total / card_number
    local card_adder = Updater.Parallel()
    local spacing = 40
    for i, widget in ipairs( self.cards_container.children ) do
        local pos_x, pos_y = widget:GetPos()
    	widget:SetTintAlpha( 0 )
        card_adder:Add( 
            Updater.Series{
                Updater.Wait(delay_per_card*i),
                Updater.Parallel{
                    Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, card_anim_duration, easing.inOutQuad),
                    Updater.Ease( function(v) widget:SetScale(v,v) end, self.card_scale * 0.9, self.card_scale, card_anim_duration, easing.inOutQuad),
                    Updater.Ease2D( function(x,y) widget:SetPos(x,y) end, pos_x, pos_y-spacing*i, pos_x, pos_y, card_anim_duration, easing.inOutQuad),
                }
            }
        )
    end
    self:RunUpdater( card_adder )

	return self
end

