require "ui/widget"

local CARD_SCALE = 0.6
local PREVIEW_SCALE = 0.8
local ARRANGEMENT_WIDTH = 450

------------------------------------------------------------------------------------------

local CardPack = class( "Widget.CardPack", Widget )
CardPack.can_focus = true

function CardPack:init( pack )
    Widget.init( self )

    self.hover_check = true

    self.pack = pack

    local card_class, card_scale = is_instance( pack[1], Battle.Card ) and Widget.BattleCard or Widget.NegotiationCard, CARD_SCALE

    self.content = self:AddChild( Widget.CardArrangement( card_class, card_scale, nil, PREVIEW_SCALE ) )
    self.content:SetFanningRadius( 1500 )
    self.content:SetArrangementWidth( ARRANGEMENT_WIDTH )

    self.cards = {}
    for i, card in ipairs( self.pack ) do
        self.content:CreateCardWidget( card )
    end

    self:StartUpdating()
end

function CardPack:OnUpdate( dt )
    Widget.OnUpdate( self, dt )
    self.content:AnimateCards( dt )
end

function CardPack:GetBoundingBox()
    local x1, y1, x2, y2 =self.content:GetBoundingBox()
    x1 = -ARRANGEMENT_WIDTH/2
    x2 = ARRANGEMENT_WIDTH/2
    return x1, y1, x2, y2
end


function CardPack:TransformToFan( x, y, r, max_width )
    local half_width = max_width*0.5

    local radius = math.sqrt( r^2 + half_width^2 )
    local theta_min = math.atan( r / half_width )
    local t = 1.0 - math.abs(x) / half_width
    local theta = Lerp( theta_min, math.pi/2, t )

    local tan_theta = math.tan( theta )
    local y1 = (radius + y) * tan_theta / math.sqrt( tan_theta^2 + 1 )
    local x1 = math.sqrt( (radius + y)^2 - y1^2 )
    if x < 0 then
        return -x1, y1 - r, math.pi/2 - theta
    elseif x == 0 then
        return 0, radius - r + y, 0
    else
        return x1, y1 - r, theta - math.pi/2
    end
end

function CardPack:SetOutline( clr )
    self.outline_clr = clr

    for i, card in ipairs( self.pack ) do
        self.content:SetCardOutline( card, clr )
    end
end

function CardPack:OnGainHover()
    if not self.outline_clr then
        for i, card in ipairs( self.pack ) do
            self.content:SetCardOutline( card, UICOLOURS.FOCUS )
        end
    end
    AUDIO:PlayParamEvent(SoundEvents.card_mouseover, "position", self:GetPanPos())
end

function CardPack:OnLoseHover()
    for i, card in ipairs( self.pack ) do
        self.content:SetCardOutline( card, self.outline_clr )
    end

    AUDIO:PlayParamEvent(SoundEvents.card_mouseoff, "position", self:GetPanPos())
end

function CardPack:OnGainFocus()
    self:OnGainHover()
end

function CardPack:OnLoseFocus()
    self:OnLoseHover()
end

function CardPack:SetOnClick(fn)
    self.on_click_handler = fn
    return self
end

function CardPack:Click()
    if self.on_click_handler then
        self.on_click_handler()
        self.down = false
    end
end

function CardPack:HandleControlUp( control, device )
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if ((device == "mouse" or device == "touch" ) and self.down) then
            self:Click()
        end
    end
end


function CardPack:HandleControlDown(control, device)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
            DBG( self )
            return true
        elseif control:Has( Controls.Digital.MENU_ACCEPT ) then
            self.down = true
        end
    end
end