require "ui/widgets/card_deck"
require "ui/widgets/bottlesindicator"
require "ui/widgets/largeroundbutton"
require "ui/widgets/flourishprogresswidget"
require "ui/screens/chooseflourishpopup"

--------------------------------------------
-- Shows a current objective/tip

local MinigameObjective = class( "Widget.MinigameObjective", Widget )

function MinigameObjective:init( title, desc )
    MinigameObjective._base.init( self )

    self.title = self:AddChild( Widget.Label( "body", 24, LOC"UI.MINIGAME_OBJECTIVE" ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
    self.desc = self:AddChild( Widget.Label( "title", 32 ) )
        :SetWordWrap( true )
        :SetAutoSize( 450 )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TITLE )
        :SetHoverLinkGlyphColour( UICOLOURS.TITLE )

    self:SetText( title, desc )
end

function MinigameObjective:GetText()
    return self.title:GetText(), self.desc:GetText()
end

function MinigameObjective:SetText( title, desc )
    self:Show()
    self.title:SetText( title or LOC"UI.MINIGAME_OBJECTIVE")

    if desc then
        self:Show()
        self.desc:SetText( desc )
    else
        self:Hide()
        self.desc:SetText( nil )
    end
    self.desc:LayoutBounds( "left", "below", self.title )
        :Offset( 0, 3 )

    return self
end

--------------------------------------------

local CardsPanel = class("Widget.CardsPanel", Widget )

local REFLECTION_HEIGHT = 35
local BUTTON_WIDTH, BUTTON_HEIGHT = 160, 40
local PREVIEW_Y_OFFSET = 150
local SCREEN_MODE_DECK_Y = 0 -- How much the cards deck should move up based on the screen size
local INPUT_MODE_DECK_Y = 0 -- How much the cards deck should move up based on the input mode (for instance, on gamepad mode there's a controls bar at the bottom)
local CURRENT_DECK_Y = 0 -- Sum of SCREEN_MODE_DECK_Y and INPUT_MODE_DECK_Y
local LOWERED_DECK_Y = -95 -- Offset when lowering cards during fighter inspection
local DEALT_CARD_SIZE = 0.75
local BUTTON_Y_OFFSET = 100

-- Delay after drawing a card into hand; causes them to animate out in sequence.
local DELAY_CARD_DRAW = 0.1

CardsPanel.BANNER = MakeEnum{ "HIGH", "MID", "LOW" }


local BANNER_PLACEMENT =
{
    [CardsPanel.BANNER.HIGH] = { 0, 415 },
    [CardsPanel.BANNER.MID] = { 0, 315 },
    [CardsPanel.BANNER.LOW] = { 0, 0 },
}

function CardsPanel:init( card_widget_class, card_engine )
	CardsPanel._base.init(self)

    self.engine = card_engine
    self.card_widget_class = card_widget_class

    self.minigame_objective = self:AddChild( Widget.MinigameObjective() ):SetAnchors( "center", "top" )

    self.animated_grid = Widget.AnimatedGrid( RES_X )
    self:AddChild( self.animated_grid ):SetAnchors( "fill", "bottom" ):LayoutBounds("center", "bottom", 0, 0 )

    self.floating_cards = Widget.CardArrangement( card_widget_class ):SetAnchors( "center", "center" ):SetName( "FLOAT" )
    self:AddChild( self.floating_cards )

    self.bottom_left = self:AddChild( Widget() ):SetAnchors( "left", "bottom" )
    self.bottom_right = self:AddChild( Widget() ):SetAnchors( "right", "bottom" )
    self.buttons_right = self.bottom_right:AddChild( Widget() )
    self.center_right = self:AddChild( Widget() ):SetAnchors( "right", "center" )

    -- Concede button
    self.concede_btn = self.buttons_right:AddChild( Widget.LargeRoundButton(140) )
        :SetIcon( engine.asset.Texture("UI/ic_concede.tex") )
        :SetClickSoundEvent( "event:/sfx/battle/cards/neg/concede" )
        :ShowToolTipOnFocus()
        :SetFocusable( false )
    -- Accept surrender button
    self.acceptsurrender_btn = self.buttons_right:AddChild( Widget.LargeRoundButton(200) )
        :SetText( LOC"UI.CARDS_PANEL.ACCEPT_SURRENDER", nil )
        :SetIcon( engine.asset.Texture("UI/ic_concede.tex") )
        :SetupFocusGlow()
        :ShowToolTipOnFocus()
        :SetFocusable( false )
    -- End-turn button
    self.end_turn_btn = self.buttons_right:AddChild( Widget.LargeRoundButton(200) )
        :SetText( LOC"UI.CARDS.END_TURN", nil )
        :SetIcon( engine.asset.Texture( "UI/characterselect_deck_arrow.tex") )
        :SetupFocusGlow()
        :ShowToolTipOnFocus()
        :SetFocusable( false )

    -- Resource Indicators
    local icon_size = 26
    local bloom = 0.08
    local reflection_w = 110
    local reflection_h = REFLECTION_HEIGHT*0.9
    local reflection_offset = -40

    -- Actions widget
    self.actions_widget = self.bottom_left:AddChild( Widget.TimeIndicator() )
    self.actions_widget:ShowReflection( REFLECTION_HEIGHT, -13 )
    -- Draw deck
    self.draw_deck = self.bottom_left:AddChild( Widget.DeckWidget( nil, LOC"UI.CARDS_PANEL.DECK", card_widget_class, Screen.DrawDeckScreen ) )
        :SetDialogTitle( LOC "UI.CARDS.DRAW_DECK_DIALOG_TITLE", LOC"UI.CARDS.DRAW_DECK_DIALOG_SUBTITLE" )
        :ShowReflection( REFLECTION_HEIGHT, -34 )
        :ShowToolTipOnFocus()
    self.bottles_draw_indicator = self.bottom_left:AddChild( Widget.BottlesIndicator() ):OrientLeft():Hide()
        :SetToolTip( LOC"UI.CARDS.BOTTLES_INDICATOR_TT" )
        :ShowToolTipOnFocus()
    -- Discard deck
	self.discard_deck = self.bottom_right:AddChild( Widget.DeckWidget( nil, LOC"UI.CARDS_PANEL.DISCARD", card_widget_class, Screen.DiscardDeckScreen ) )
        :SetDialogTitle( LOC"UI.CARDS.DISCARDS_DECK_DIALOG_TITLE", LOC"UI.CARDS.DISCARDS_DECK_DIALOG_SUBTITLE" )
        :ShowReflection( REFLECTION_HEIGHT, -34  )
        :ShowToolTipOnFocus()
    self.bottles_discard_indicator = self.bottom_right:AddChild( Widget.BottlesIndicator() ):OrientRight():Hide()
        :SetToolTip( LOC"UI.CARDS.BOTTLES_INDICATOR_TT" )
        :ShowToolTipOnFocus()
    -- Trash deck
    self.trash_deck = self.bottom_right:AddChild( Widget.DeckWidget( nil, LOC"UI.CARDS_PANEL.TRASH", card_widget_class, Screen.DiscardDeckScreen ) )
        :SetDialogTitle( LOC"UI.CARDS.TRASH_DECK_DIALOG_TITLE", LOC"UI.CARDS.TRASH_DECK_DIALOG_SUBTITLE" )
        :ShowReflection( REFLECTION_HEIGHT, -34  )
        :ShowToolTipOnFocus()
        :Hide()
    -- Flourish bar
    self.flourish_bar = self.bottom_right:AddChild( Widget.FlourishProgressWidget() )
        :SetupFlourishEffects()

    -- Cards in the middle of the screen when the player is choosing cards from their hand
    self.choose_cards = Widget.CardArrangement( card_widget_class, 0.5, nil, 0.7 ):SetAnchors( "center", "center" ):Offset( 0, 200 ):SetName( "CHOOSE" )
        :SetCenteredMode() -- Showing the cards in the center of the screen, instead of the bottom. Preview card scales up in place.
    self:AddChild( self.choose_cards )

    -- Cards at the bottom of the screen. This is the player's hand
    self.cards = self:AddChild( Widget.CardArrangement( card_widget_class, nil, PREVIEW_Y_OFFSET ) )
        :SetAnchors( "center", "bottom" )
        :SetPos( 0, CURRENT_DECK_Y )
        :SetName( "HAND" )
        :SetFanningRadius( CARD_FAN_HEIGHT )
        :SetFocusLock( true )
        :ShowMask() -- This clips the cards and shows the glow at the bottom

    self.click_overlay = self:AddChild( Widget.SolidBox( RES_X, RES_Y, 0x00000099 ) )
        :BlockMouse():SetAnchors( "fill", "fill" )
        :Hide()

    -- Cards in the middle of the screen that were just dealt to the player. Shown with a "Click to continue" banner
    self.dealt_cards = Widget.CardArrangement( card_widget_class, DEALT_CARD_SIZE, PREVIEW_Y_OFFSET ):SetAnchors( "center", "center" ):SetName( "DEALT" )
        :DisablePreview( true )

    self:AddChild( self.dealt_cards )

    self.flourish_cards = Widget.CardArrangement( card_widget_class, 0.55 ):SetAnchors( "center", "center" ):SetName( "FLOURISH" )
    self:AddChild( self.flourish_cards )

    -- Info banner
    self.banner = Widget.CardsPanelBannerWidget( card_widget_class ):SetPos( 0, 415 ):SetAnchors( "center", "center" )
    self:AddChild( self.banner )
    -- Ending banner
    self.ending_banner = Widget.CardsPanelEndingBanner():SetPos(0, 315):SetAnchors( "center", "center" )
    self:AddChild( self.ending_banner )

    -- Cancel info label..
    self.cancel_info_label_container = self:AddChild( Widget() )
        :SetAnchors( "center", "bottom" )
    self.cancel_info_label = self.cancel_info_label_container:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_SUBTITLE, LOC"UI.CARDS_PANEL.CANCEL_INFO_LABEL" ) )
        :SetHiddenBoundingBox( true )
        :Bloom( 0.05 )
        :Hide()

    self.anim_pacing = 0.3
    self.streamer_deck_width_multiplier = 1
    self.streamer_deck_offset_left = 0
    self.streamer_deck_offset_right = 0
    self.choose_cards_dialog_bottom_offset = 10
    
    if TheGame:GetLocalSettings().DEBUG then
        self.debug_container = self:AddChild(Widget())
            :SetName("DEBUG_WIDGET")
            :SetAnchors("left","center")
        self.debug_label = self.debug_container:AddChild( Widget.Label("title",30,""))
    end
end

function CardsPanel:OnControlModeChange( cm, device_id )
    CardsPanel._base.OnControlModeChange( self, cm, device_id )

    self.choose_cards_dialog_bottom_offset = 10
    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self:ShowButtons()
        BUTTON_Y_OFFSET = 100
        INPUT_MODE_DECK_Y = 0
        self.draw_deck:SetTitle( LOC"UI.CARDS_PANEL.DECK" )
        self.discard_deck:SetTitle( LOC"UI.CARDS_PANEL.DISCARD" )
        self.click_to_continue_string = LOC"UI.CARDS_PANEL.KEYBOARD_CONTINUE"
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self:HideButtons()
        BUTTON_Y_OFFSET = 140
        INPUT_MODE_DECK_Y = 34
        self.draw_deck:SetTitle( LOC"UI.CARDS_PANEL.CONTROLS_DECK" )
        self.discard_deck:SetTitle( LOC"UI.CARDS_PANEL.CONTROLS_DISCARD" )
        self.click_to_continue_string = LOC"UI.CARDS_PANEL.CONTROLS_CONTINUE"
        self.choose_cards_dialog_bottom_offset = 60
    end
    if cm == CONTROL_MODE.TOUCH then
        self:ShowButtons()
        BUTTON_Y_OFFSET = 100
        INPUT_MODE_DECK_Y = 0
        self.draw_deck:SetTitle( LOC"UI.CARDS_PANEL.DECK" )
        self.discard_deck:SetTitle( LOC"UI.CARDS_PANEL.DISCARD" )
        self.click_to_continue_string = LOC"UI.CARDS_PANEL.TOUCH_CONTINUE"
    end

    if self.choose_cards_dialog then
        self.choose_cards_dialog:SetPos( 0, self.choose_cards_dialog.height/2 + self.choose_cards_dialog_bottom_offset )
        self.buttons_right:Hide()
    end

    self:RefreshCardsInHandPos()

    if self.wait_for_click then
        self.banner:UpdateText( self.click_to_continue_string )
    end

    self:LayoutButtons()
end

function CardsPanel:OnScreenModeChange( sm )
    CardsPanel._base.OnScreenModeChange( self, sm )

    self.streamer_deck_offset_left = 0 -- How much the deck should move to the left when in right-streamer mode
    self.streamer_deck_offset_right = 0 -- How much the deck should move to the right when in left-streamer mode
    local cancel_info_label_y = 330
    if sm == SCREEN_MODE.MONITOR then
        SCREEN_MODE_DECK_Y = 0
        self.streamer_deck_width_multiplier = 1 -- Adjusts the width of the deck for certain screen UI sizes
        self.streamer_deck_offset_left = -170
        self.streamer_deck_offset_right = 170
        cancel_info_label_y = 330
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        SCREEN_MODE_DECK_Y = 0
        self.streamer_deck_width_multiplier = 0.7
        self.streamer_deck_offset_left = -230
        self.streamer_deck_offset_right = 270
        cancel_info_label_y = 380
    end

    self:RefreshCardsInHandPos()

    self.cancel_info_label:SetLayoutScale( LAYOUT_SCALE[sm] )
        :LayoutBounds( "center", "center", 0, cancel_info_label_y )

    self.bottom_left:SetLayoutScale( (LAYOUT_SCALE[sm]+1)/2 )
    self.bottom_right:SetLayoutScale( (LAYOUT_SCALE[sm]+1)/2 )
    self.center_right:SetLayoutScale( (LAYOUT_SCALE[sm]+1)/2 )

    self.minigame_objective:SetLayoutScale( LAYOUT_SCALE[sm] )

    self:LayoutButtons()
end

function CardsPanel:SetHUDViz( show )
    if not show then
        self.banner.Show = function( self ) return self end
        self.banner:Hide()
        self.bottom_right:Hide()
        self.bottom_left:Hide()
    else
        self.banner.Show = nil
        self.bottom_right:Show()
        self.bottom_left:Show()
    end
end

function CardsPanel:RefreshCardsInHandPos( lowered )
    if lowered ~= nil then
        local sm = TheGame:FE():GetScreenMode()
        self.cards_lowered = lowered and (sm == SCREEN_MODE.TV or sm == SCREEN_MODE.SMALL)
    end

    CURRENT_DECK_Y = 75 + SCREEN_MODE_DECK_Y + INPUT_MODE_DECK_Y + (self.cards_lowered and LOWERED_DECK_Y or 0)
    self.cards:SetPos( nil, CURRENT_DECK_Y )
end

function CardsPanel:SetMinigameObjective( title, desc )
    local old_title, old_desc = self.minigame_objective:GetText()
    if old_title ~= title or old_desc ~= desc then
        self.minigame_objective:SetText( title, desc )
        local cx, cy = self.minigame_objective:GetCentroid()
        self.minigame_objective:SetPos( -cx, -cy - 135 )
    end
    return self
end

function CardsPanel:ShowObjective()
    self.minigame_objective:Show()
end

function CardsPanel:HideObjective()
    self.minigame_objective:Hide()
end

function CardsPanel:GetDefaultFocus()
    return ( self.wait_for_click and self.dealt_cards and self.dealt_cards.card_order[1] )
        or self.nextfocus
        or ( self.cards and self.cards:GetDefaultFocus() )
end

function CardsPanel:SetFocusable( focusable )
    self.cards:SetFocusable( focusable )
end

function CardsPanel:ShowButtons()
    self.buttons_right:Show()
end

function CardsPanel:HideButtons()
    self.buttons_right:Hide()
end

function CardsPanel:LayoutButtons()

    -- Set the hand width
    self.cards:SetArrangementWidth(900 * self.streamer_deck_width_multiplier)
        :SetPos( 0, nil )

    -- Position the left side widgets
    self.draw_deck:SetPos( 80, BUTTON_Y_OFFSET )
    self.actions_widget
        :ShowLabel( true )
        :SetReflectionShown( true )
        :ShowToolTipOnFocus( true )
        :LayoutBounds( "after", "bottom", self.draw_deck ):Offset( 30, 0 )

    -- Position the right side widgets
    self.discard_deck:SetPos( -80, BUTTON_Y_OFFSET )
    self.end_turn_btn:LayoutBounds( "before", "center", self.discard_deck )
        :Offset( -SPACING.M1*1, 0 )
        :SetToolTipLayoutFn( function( w, tooltip_widget ) tooltip_widget:LayoutBounds( "right", "above", w ):Offset( 0, SPACING.M1 ) end )
    self.flourish_bar:LayoutBounds( "before", "bottom", self.end_turn_btn )
        :Offset( 0, SPACING.M1*0.4 )
        :SetToolTipLayoutFn( function( w, tooltip_widget ) tooltip_widget:LayoutBounds( "right", "above", w ):Offset( 0, SPACING.M1 ) end )
    self.concede_btn:LayoutBounds( "right", "above", self.flourish_bar ):Offset( 30, -5 )

    if GetStreamerMode() == 1 then
        -- Left side should be free
        -- Nudge the right side widgets
        self.end_turn_btn:LayoutBounds( "before", "center", self.discard_deck )
            :Offset( 0, 0 )
            :SetToolTipLayoutFn( function( w, tooltip_widget ) tooltip_widget:LayoutBounds( "right", "above", w ):Offset( 0, SPACING.M1 ) end )
        if self.bottles_discard_indicator:IsShown() then
            self.concede_btn:LayoutBounds( "center", "above", self.end_turn_btn )
                :Offset( SPACING.M1*2, -SPACING.M1*0.5 )
            self.flourish_bar:LayoutBounds( "before", "center", self.concede_btn ):Offset( 0, -8 )
        else
            self.concede_btn:LayoutBounds( "center", "above", self.discard_deck )
                :Offset( 0, SPACING.M1*0.5 )
            self.flourish_bar:LayoutBounds( "before", "center", self.concede_btn ):Offset( 11, -2 )
        end

        -- Nudge the left side widgets
        self.draw_deck:SetPos( 440, BUTTON_Y_OFFSET )
        self.actions_widget
            :ShowLabel( true )
            :SetReflectionShown( true )
            :LayoutBounds( "after", "bottom", self.draw_deck ):Offset( 30, 0 )
        -- Position the cards
        self.cards:SetArrangementWidth(640 * self.streamer_deck_width_multiplier)
            :SetPos( self.streamer_deck_offset_right, nil )
    end

    if GetStreamerMode() == 2 then
        -- Right side should be free
        -- Nudge the right side widgets
        self.discard_deck:SetPos( -440, BUTTON_Y_OFFSET )
        self.end_turn_btn:LayoutBounds( "before", "center", self.discard_deck )
            :Offset( 0, 0 )
            :SetToolTipLayoutFn( function( w, tooltip_widget ) tooltip_widget:LayoutBounds( "right", "above", w ):Offset( 0, SPACING.M1 ) end )
        if self.bottles_discard_indicator:IsShown() then
            self.concede_btn:LayoutBounds( "center", "above", self.end_turn_btn )
                :Offset( SPACING.M1*2, -SPACING.M1*0.5 )
            self.flourish_bar:LayoutBounds( "before", "center", self.concede_btn ):Offset( 0, -8 )
        else
            self.concede_btn:LayoutBounds( "center", "above", self.discard_deck )
                :Offset( 0, SPACING.M1*0.5 )
            self.flourish_bar:LayoutBounds( "before", "center", self.concede_btn ):Offset( 10, -2 )
        end
        -- Position the cards
        self.cards:SetArrangementWidth(640 * self.streamer_deck_width_multiplier)
            :SetPos( self.streamer_deck_offset_left, nil )
    end

    self.acceptsurrender_btn:LayoutBounds( "center", "center", self.end_turn_btn )
        :SetToolTipLayoutFn( function( w, tooltip_widget )  tooltip_widget:LayoutBounds( "right", "above", w ):Offset( 0, SPACING.M1 ) end )

    -- If we're in gamepad mode, readjust the flourish bar
    if TheGame:FE():IsGamepadMode() then
        self.flourish_bar:LayoutBounds( "before", "center", self.discard_deck ):Offset( -10, 4 )
    end

    self.bottles_draw_indicator:LayoutBounds( "center", "above", self.draw_deck ):Offset( -30, 8 )
    self.bottles_discard_indicator:LayoutBounds( "center", "above", self.discard_deck ):Offset( 30, 8 )
end

function CardsPanel:Setup()
    TheGame:GetEvents():ListenForAny( self )

    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    -- Hide objects until they're animated in
    self.concede_btn:Hide()
    self.acceptsurrender_btn:Hide()
    self.end_turn_btn:Hide()
    self.flourish_bar:Hide()

	self.draw_deck:PrepareAnimation()
    self.actions_widget:PrepareAnimation()
    self.end_turn_btn:PrepareAnimation()
    self.concede_btn:PrepareAnimation()
    self.discard_deck:PrepareAnimation()
    self.flourish_bar:PrepareAnimation()
end

function CardsPanel:AnimateIn()
	-- Prepare for animation by getting current values
    local animated_grid_x, animated_grid_y = self.animated_grid:GetPos()
    
	-- Setup and start animation
    self.fadein_anim = self:RunUpdater(
		Updater.Series{
            Updater.Parallel{
                -- Fade and move in the grid
                Updater.Ease2D( function(x,y) self.animated_grid:SetPos(x,y) end, animated_grid_x, animated_grid_y-20, animated_grid_x, animated_grid_y, self.anim_pacing, easing.linear),
                Updater.Ease( function(v) self.animated_grid:SetTintAlpha(v) end, 0, 1, self.anim_pacing, easing.linear),
            },
            Updater.Do( function()
                self.draw_deck:AnimateIn()
                self.discard_deck:AnimateIn()
                -- self.concede_btn:AnimateIn() -- These will show on BEGIN_TURN
                self.actions_widget:AnimateIn()
                -- self.end_turn_btn:AnimateIn() -- These will show on BEGIN_TURN
            end )
        }
    )
end

function CardsPanel:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function CardsPanel:FlourishBar()
	return self.flourish_bar
end

function CardsPanel:ActionsWidget()
	return self.actions_widget
end

function CardsPanel:SetActions( actions, max_actions )
	self.actions_widget:SetTime( actions, max_actions )
end

function CardsPanel:ShowCancelInfoLabel()
    self.cancel_info_label:SetShown( TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD and not self:IsChoosingCards() )
    return self
end

function CardsPanel:HideCancelInfoLabel()
    self.cancel_info_label:Hide()
    return self
end

function CardsPanel:EndTurnButton( fn, tooltip )
	if fn then self.end_turn_btn:SetOnClickFn( fn ) end
	if tooltip then self.end_turn_btn:SetToolTip( tooltip ) end
	return self.end_turn_btn
end

function CardsPanel:ConcedeButton( fn, txt, tooltip )
	if fn then self.concede_btn:SetOnClickFn( fn ) end
	self.concede_btn:SetText( txt ):SetToolTip( tooltip )
	return self.concede_btn
end

function CardsPanel:AcceptSurrenderButton( fn, tooltip )
	if fn then self.acceptsurrender_btn:SetOnClickFn( fn ) end
	if tooltip then self.acceptsurrender_btn:SetToolTip( tooltip ) end
	return self.acceptsurrender_btn
end

function CardsPanel:DrawDeckButton( deck )
	self.draw_deck:RefreshDeck( deck )
	return self.draw_deck
end

function CardsPanel:DiscardDeckButton( deck )
	self.discard_deck:RefreshDeck( deck )
	return self.discard_deck
end

function CardsPanel:TrashDeckButton( deck, fn )
    if deck then
        self.trash_deck:RefreshDeck( deck )
        self.discard_deck:SetOtherDeckButton( deck, LOC"UI.CARDS.TRASH_DECK_DIALOG_TITLE", LOC"UI.CARDS.TRASH_DECK_DIALOG_SUBTITLE" )
    end
    -- if fn then self.trash_deck:SetOnClickFn( fn ) end
    return self.trash_deck
end

function CardsPanel:FloatingCards()
    return self.floating_cards
end

function CardsPanel:Cards()
	return self.cards
end

function CardsPanel:DealtCards()
	return self.dealt_cards
end

function CardsPanel:ChooseCards()
	return self.choose_cards
end

function CardsPanel:AnimateShuffle( num_cards )
    AUDIO:PlayEvent(SoundEvents.card_discard_reshuffle)

    local x0, y0 = self:TransformFromWidget( self.discard_deck, 0, 0 )
    local x1, y1 = self:TransformFromWidget( self.draw_deck, 0, 0 )
    for i = 1, num_cards do
        local card = Widget.Image( engine.asset.Texture( "UI/ic_deck_card.tex") ):SetScale( 0.6 )
        self:AddChild( card ):SetPos( x0, y0 )
        card:MoveTo( x1, y1, 0.3, easing.inQuad, function() card:Remove() end )
        self:Delay( 0.05 )
    end
end

-- Finds the card's associated widget.
function CardsPanel:FindCardWidget( card )
    local w = self.floating_cards:FindCardWidget( card )

    if w == nil then
        w = self.cards:FindCardWidget( card )
    end

    if w == nil then
        w = self.dealt_cards:FindCardWidget( card )
    end

    if w == nil then
        w = self.choose_cards:FindCardWidget( card )
    end

    if w == nil then
        w = self.flourish_cards:FindCardWidget( card )
    end

    return w
end

-- Removes the card from its existing arrangement.
function CardsPanel:DetachCardWidget( card )
    local w, idx

    if w == nil then
        w, idx = self.floating_cards:DetachCardWidget( card )
    end

    if w == nil then
        w, idx = self.cards:DetachCardWidget( card )
    end

    if w == nil then
        w, idx = self.dealt_cards:DetachCardWidget( card )
    end

    if w == nil then
        w, idx = self.choose_cards:DetachCardWidget( card )
    end

    if w == nil then
        w, idx = self.flourish_cards:DetachCardWidget( card )
    end

    return w, idx
end

function CardsPanel:GetCardAt( i )
    return self.cards:GetCardAt( i )
end

function CardsPanel:TryChooseCard( card, device )
    return self.cards:TryChooseCard( card, device )
end

function CardsPanel:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )
    local deck_type = target_deck and target_deck:GetDeckType()

    local w = self.dealt_cards:FindCardWidget( card ) -- and deck_type ~= DECK_TYPE.RESOLVE then
    if w then
        -- Store a placement memento so that when the card does its dealt animation, it knows where to go.
        w.card_placement = { source_deck, source_idx, target_deck, target_idx }
        return
    end

    if source_deck == target_deck and source_idx and target_idx and deck_type == DECK_TYPE.IN_HAND then
        -- Just swapping positions.
        local w = self.cards:FindCardWidget( card )
        if w then
            self.cards:SwapCardWidget( card, target_idx )
            w:RefreshCard( card, self.engine )
            return
        end
    end

    local w = self:DetachCardWidget( card )
    if w then
        w:WaitOnCoros()
        if w.removed then
            w = nil
        end
    end

    if deck_type == DECK_TYPE.IN_HAND then
       if w == nil then
            w = self.cards:CreateCardWidget( card )

            local x, y
            if source_deck == self.engine:GetDrawDeck() then
                x, y = w.parent:TransformFromWidget( self.draw_deck, 0, 0 )

            elseif source_deck == self.engine:GetDiscardDeck() then
                x, y = w.parent:TransformFromWidget( self.discard_deck, 0, 0 )

            else
                -- FIXME: unhandled deck source
                x, y = w.parent:TransformFromWidget( self.draw_deck, 0, 0 )
            end
            w:SetPos( x, y )
            w:SetTintAlpha( 0 )
        else
            self.cards:AttachCardWidget( card, w, target_idx )
            w:RefreshCard( card, self.engine )
        end
        
        AUDIO:PlayEvent(SoundEvents.card_draw_single)

         -- Delay cards attached in hand so they come out in sequence.
        self:Delay( DELAY_CARD_DRAW )

    elseif deck_type == DECK_TYPE.DISCARDS then
        if w == nil then
            if source_deck == self.engine:GetDrawDeck() then
                w = self.floating_cards:CreateCardWidget( card )
                local x, y = self.floating_cards:TransformFromWidget( self.draw_deck, 0, 0 )
                w:SetPos( x, y ):SetScale( 0.5 )
                self:Delay( 0.1 )

            elseif source_deck == nil then
                w = self.floating_cards:CreateCardWidget( card )
                w:SetScale( 0.8 )
                self:Delay( 0.5 )
            end
        else
            self.floating_cards:AttachCardWidget( card, w )
        end

        if w then
            AUDIO:PlayEvent(SoundEvents.card_discard)
            w:AnimateToDeck( self.discard_deck )
        end

    elseif deck_type == DECK_TYPE.DRAW then
        if w then
            self.floating_cards:AttachCardWidget( card, w )
            AUDIO:PlayEvent(SoundEvents.card_add_to_deck)
            w:AnimateToDeck( self.draw_deck )
        end

    elseif deck_type == DECK_TYPE.RESOLVE then
        if w then
            self.floating_cards:AttachCardWidget( card, w )
            w:AnimatePreResolve( self )
        
        elseif card.owner:IsPlayer() then
            w = self.floating_cards:CreateCardWidget( card )

            local x, y
            if source_deck == self.engine:GetDiscardDeck() then
                x, y = self.floating_cards:TransformFromWidget( self.discard_deck, 0, 0 )
            else
                x, y = self.floating_cards:TransformFromWidget( self.draw_deck, 0, 0 )
            end
            w:SetPos( x, y )

            w:AnimatePreResolve( self )
        end

    elseif deck_type == DECK_TYPE.TRASH then
        if w == nil then
            local screenw, screenh = self:GetFE():GetScreenDims()
            local x, y, tx, ty
            if source_deck == self.engine:GetDrawDeck() then
                x, y = self.floating_cards:TransformFromWidget( self.draw_deck, 0, 0 )
                tx, ty = self.floating_cards:TransformFromWorld( screenw/4, screenh/3 )

            elseif source_deck == self.engine:GetDiscardDeck() then
                x, y = self.floating_cards:TransformFromWidget( self.discard_deck, 0, 0 )
                tx, ty = self.floating_cards:TransformFromWorld( screenw*3/4, screenh/3 )
            end

            if x and y then
                w = self.floating_cards:CreateCardWidget( card )
                w:SetPos( x, y ):MoveTo( tx, ty, 0.3, easing.outQuad )
            end
        else
            self.floating_cards:AttachCardWidget( card, w )
        end
 
        if w then
            w:AnimateToTrash()
        end

    else
        -- Card is not in any deck.  This is a special situation that happens due to some cards, eg. Pinned that
        -- extract other cards and hold them in stasis.  The card presentation at this point is therefore up to that
        -- mechanic to handle. Pinned, for example, broadcasts CARD_PINNED to make the card animate 'into' the status_pinned card.
        if w then
            self.floating_cards:AttachCardWidget( card, w )
        end
    end
end

function CardsPanel:DealCards()
    if self.dealt_cards:GetCardCount() > 0 then

        AUDIO:PlayEvent( "event:/sfx/battle/cards/gen/enemy_incepts_card" )
        self.dealt_cards:SetFocusLock( true )
        if TheGame:FE():IsGamepadMode() then TheGame:FE():ClearFocusWidget() end

        local auto_deal = true
        for card, card_widget in self.dealt_cards:Cards() do
            if not card.auto_deal then
                auto_deal = false
                break
            end
        end

        if not auto_deal then
            self:WaitForClick()
        else
            self:Delay( 0.5 )
        end

        while self.dealt_cards:GetCardCount() > 0 do
            local w = self.dealt_cards.card_order[1]

             -- Must pre-detach since UpdateCardPlacement skips dealt_cards
            self:DetachCardWidget( w.card )
            self.floating_cards:AttachCardWidget( w.card, w )

            -- If CARD_MOVED for this card has already been processed, then it should have a card_placement memento.
            -- If not, then CARD_MOVED is pending and will move this card to the appropriate place.
            if w.card_placement then
                self:UpdateCardPlacement( w.card, table.unpack( w.card_placement ))
                w.card_placement = nil
                self:Delay( 0.1 )
            end
        end
    end
end

function CardsPanel:ShowChooseCards( txt, cards_widget, min_count, max_count, fn, source )

    local height
    if self:GetFE():GetScreenMode() == SCREEN_MODE.SMALL then
        height = 360
    else
        height = 330
    end
    self.choose_cards_dialog = self:AddChild( Widget.ChooseCardsFromHand( height, txt, cards_widget, min_count, max_count, fn, source ) )
        :SetAnchors( "center", "bottom" )
        :SetPos( 0, height/2 + self.choose_cards_dialog_bottom_offset )
        :SendBehind( self.bottom_left )
        :SetOnInspectHandFn( function() self:OnInspectHand() end )
        :SetOnInspectOptionsFn( function() self:OnInspectOptions() end )
    self.nextfocus = cards_widget:GetDefaultFocus()
    self.buttons_right:Hide()

    self:HideCancelInfoLabel()
    return self.choose_cards_dialog:GetConfirmButton()
end

function CardsPanel:HideChooseCards()
    self.nextfocus = nil
    self.choose_cards_dialog:Remove()
    self.choose_cards_dialog = nil
    if not TheGame:FE():IsGamepadMode() then self.buttons_right:Show() end
end

function CardsPanel:IsChoosingCards()
    return self.choose_cards_dialog ~= nil
end

function CardsPanel:HasConfirmButton()
    return self.choose_cards_dialog and self.choose_cards_dialog:HasConfirmButton()
end

function CardsPanel:GetChoosingCardsPanel()
    return self.choose_cards_dialog
end

function CardsPanel:OnInspectHand()
    self.floating_cards:Hide()
    self.choose_cards:Hide()
    self.cards:DisablePreview( true )
end

function CardsPanel:OnInspectOptions()
    self.floating_cards:Show()
    self.choose_cards:Show()
    self.cards:DisablePreview( false )
end

function CardsPanel:IsWaitingForClick()
    return self.wait_for_click
end

function CardsPanel:WaitForClick(custom_msg)
    self.click_overlay:Show():SetTintAlpha( 0 ):AlphaTo( 1.0, 0.4 )

    -- ccc: always show this msg, why wouldn't we? (conditional breaks authority graft dealing authority card)
    -- if not self.banner:IsShown() then
        self:SetInstructionText( custom_msg or self.click_to_continue_string, Widget.CardsPanel.BANNER.MID )
    -- end

    self.wait_for_click = true

    if TheGame:GetLocalSettings().ROBOTICS then
        self:Delay( 0.5 )
    else
        while self.wait_for_click do
            coroutine.yield()
        end
    end

    self:SetInstructionText( nil )
    self.click_overlay:AlphaTo( 0, 0.25, easing.linear, function() self.click_overlay:Hide() end )
end

function CardsPanel:IsCardSelected()
    return self.cards:IsCardSelected()
end

function CardsPanel:GetCardSelected()
    return self.cards:GetCardPreview()
end

function CardsPanel:DoCardChoosing( card_engine, arrangement, txt, min_count, max_count, pred, source )

    self.choose_cards:Show()
    local confirmed_choice = false
    local card_order = shallowcopy( arrangement.card_order )
    local chosen_cards = {}

    local function OnClickConfirm()
        confirmed_choice = true
    end

    local button = self:ShowChooseCards( txt, arrangement, min_count, max_count, OnClickConfirm, source )
    arrangement:BeginChoosing( card_engine, min_count, max_count, pred or TrueFunction )
    -- Focus on the first card
    if TheGame:FE():IsGamepadMode() then
        for i = 1, arrangement:GetCardCount() do
            local card_at = arrangement:GetCardAt( i )
            local widget = arrangement:FindCardWidget( card_at )
            if widget.enabled then
                arrangement:SetCardPreview( widget.card )
                widget:SetFocus()
                break
            end
        end
    end

    local function RemoveChosenCard( card )
        local w = self.choose_cards:DetachCardWidget( card )
        local idx = table.arrayfind( card_order, w )

        -- Preserve original order.
        for i = 1, arrangement:GetCardCount() do
            local card_at = arrangement:GetCardAt( i )
            local this_idx = table.arrayfind( card_order, arrangement:FindCardWidget( card_at ))
            if this_idx > idx then
                -- Here!
                arrangement:AttachCardWidget( card, w, i )
                break
            end
        end

        arrangement:AttachCardWidget( card, w )
    end

    self.choose_cards:SetOnChoose( nil,
        function( card )
            RemoveChosenCard( card )
            table.arrayremove( chosen_cards, card )
            
            TheGame:FE():ClearFocusWidget()
            
            return true
        end )

    local f1, f2 = arrangement:GetOnChoose()
    arrangement:SetOnChoose( nil,
        function( card )
            local ok, reason
            if pred then
                ok, reason = pred( card )
            else
                ok = true
            end
            if not ok then
                return true -- sink it
            end
            local w = arrangement:DetachCardWidget( card )
            self.choose_cards:AttachCardWidget( card, w )

            while #chosen_cards >= max_count do
                RemoveChosenCard( table.remove( chosen_cards, 1 ))
            end

            table.insert( chosen_cards, card )

            -- Focus on the next card
            if TheGame:FE():IsGamepadMode() then
                for i = 1, arrangement:GetCardCount() do
                    local card_at = arrangement:GetCardAt( i )
                    local widget = arrangement:FindCardWidget( card_at )
                    if widget.enabled then
                        arrangement:SetCardPreview( widget.card )
                        widget:SetFocus()
                        break
                    end
                end
            end

            return true
        end )

    do
        while not confirmed_choice do
            local can_confirm = #chosen_cards >= min_count and #chosen_cards <= max_count
            if can_confirm and min_count == max_count then
                -- Auto-confirm if there is an explicit count.
                break
            else
                button:SetEnabled( can_confirm == true )

                if TheGame:GetLocalSettings().ROBOTICS then
                    self:Delay( 0.5 )
                    local cards = {}
                    for card, w in arrangement:Cards() do
                        if w:IsEnabled() then
                            table.insert( cards, card )
                        end
                    end
                    cards = table.multipick( cards, min_count )
                    for i, card in ipairs( cards ) do
                        arrangement:TryChooseCard( card )
                    end
                    confirmed_choice = true
                end
            end
            coroutine.yield()
        end
    end
    arrangement:EndChoosing( self.engine )

    self.choose_cards:SetOnChoose( nil, nil )
    arrangement:SetOnChoose( f1, f2 )

    for card, w in self.choose_cards:Cards() do
        RemoveChosenCard( card )
        w:RefreshCard( card, self.engine )
    end

    self:HideChooseCards()
    self:Delay( 0.25 )
    self.choose_cards:Hide()

    return chosen_cards
end

function CardsPanel:SetInstructionText( txt, placement )
    if self.banner_coro then
        self.banner_coro:Stop()
        self.banner_coro = nil
    end

    if txt then
        local x, y = table.unpack( BANNER_PLACEMENT[ placement or Widget.CardsPanel.BANNER.LOW ] )
        self.banner:SetPos( x, y )
        self.banner:ShowBanner( txt )
    else
        self.banner:HideBanner()
    end
end

function CardsPanel:ShowBanner( txt, duration, placement, color_top, color_bottom, icon_on_right )
    if self.banner_coro then
        self.banner_coro:Stop()
    end
    local x, y = table.unpack( BANNER_PLACEMENT[ placement or Widget.CardsPanel.BANNER.LOW ] )
    duration = duration or 2.5
    self.banner:SetPos( x, y )
    self.banner_coro = self:StartCoroutine( self.banner.ShowBanner, self.banner, txt, duration, color_top, color_bottom, icon_on_right )
end

function CardsPanel:HideBanner()
	if self.banner_coro then
        self.banner_coro:Stop()
        self.banner_coro = nil
    end
	self.banner:HideBanner()
end

function CardsPanel:ShowEndBanner( victory, txt )
    self:StartCoroutine( self.ending_banner.ShowBanner, self.ending_banner, victory, txt )
end

function CardsPanel:BannerAnimating()
	return self.banner_coro and not self.banner_coro:IsDone()
end

function CardsPanel:ActivateFlourish( flourishes, on_choose )

    self.cards:EndChoosing()

    local function OnClose( popup, card_widget )
        local card
        if card_widget then
            local ok, reason = self.engine:CanPlayCard( card_widget.card )
            if not ok then
                AUDIO:PlayEvent(SoundEvents.card_invalid)
                if reason then
                    reason = GetLocalizedConvoCommonString( reason )
                    self:ShowBanner( reason, nil, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
                end
            else
                local w, h = self:GetFE():GetScreenDims()
                self.flourish_cards:AttachCardWidget( card_widget.card, card_widget )
                local tx, ty = self.flourish_cards:TransformFromWorld( w/2, h/6 )
                card_widget:MoveTo( tx, ty, 0.2, easing.inQuad )
                card = card_widget.card
            end
        end

        self.cards:BeginChoosingForPlay( self.engine, 1, 1 )
        on_choose( card )
        self.cards:DisablePreview( self.parent.chosen_card ~= nil )
    end

    local function CanChooseFlourish( card )
        return true
    end

    local popup = Screen.ChooseFlourishPopup( flourishes, CanChooseFlourish, self.card_widget_class, self.engine, OnClose )
    self:GetFE():InsertScreen( popup )
end

--------------------------------------------------------------
-- Banner Widget

local CardsPanelBannerWidget = class("Widget.CardsPanelBannerWidget", Widget)

function CardsPanelBannerWidget:init( card_widget_class )
    CardsPanelBannerWidget._base.init(self)

    self.card_type = card_widget_class or Widget.NegotiationCard--Widget.BattleCard

    -- Get screen dimensions
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- A container for all the background elements
    self.panel = Widget():BlockMouse()
    self.panel_glow = self.panel:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), 1200, 300 ) )
    self.panel_glow:SetHiddenBoundingBox( true )
    self.panel_bg = self.panel:AddChild( Widget.Image( engine.asset.Texture("large/dialog_bg.tex" ) ) )
    self.panel_grid = self.panel:AddChild( Widget.Image( engine.asset.Texture("large/banner_grid.tex" ) ) )
    self.panel_border_top_right = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_top_left = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_bottom_right = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_bottom_left = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    
    -- Two lines that show different colors depending on context
    self.content_divider_top = Widget.Image( engine.asset.Texture("large/dialog_divider_top.tex" ) )
    self.content_divider_top:SetSize( self.screen_w*0.5, 6 )
    self.content_divider_top:Bloom( 0.2 )
    self.content_divider_top:SetTintColour( UICOLOURS.SUBTITLE )
    self.content_divider_bottom = Widget.Image( engine.asset.Texture("large/dialog_divider_top.tex" ) )
    self.content_divider_bottom:SetSize( self.screen_w*0.5, 6 )
    self.content_divider_bottom:Bloom( 0.2 )
    self.content_divider_bottom:SetTintColour( UICOLOURS.HEALTH )

    -- Our actual content!
    self.content = Widget()
    self.icon_container = Widget.Image( engine.asset.Texture("ui/banner_icon_container.tex" ), 190, 190 )
    self.icon_container:Bloom( 0.2 )
    self.icon_container:SetHiddenBoundingBox( true )
    self.icon = Widget.Image( engine.asset.Texture("ui/card_type_negotiation.tex"), 70, 70 ):SetBloom(0.1)
    self.label = Widget.Label( "title", 38 )

    self:AddChildren{
        self.panel,
        self.content_divider_top,
        self.content_divider_bottom,
        self.content:AddChildren{
            self.icon_container,
            self.icon,
            self.label
        }
    }

    -- Set colors
    if self.card_type == Widget.NegotiationCard then
        self.panel_bg:SetTintAlpha( 0.98 )
        self.icon:SetTexture( engine.asset.Texture("ui/card_type_negotiation.tex") )
        self.panel_glow:SetTintColour( UICOLOURS.NEGOTIATION )
        self.panel_glow:SetTintAlpha( 0.4 )
        self.icon_container:SetTintColour( UICOLOURS.NEGOTIATION )
        self.panel_border_top_right:SetTintColour( UICOLOURS.NEGOTIATION )
        self.panel_border_top_left:SetTintColour( UICOLOURS.NEGOTIATION )
        self.panel_border_bottom_right:SetTintColour( UICOLOURS.NEGOTIATION )
        self.panel_border_bottom_left:SetTintColour( UICOLOURS.NEGOTIATION )
        self.label:SetGlyphColour( UICOLOURS.NEGOTIATION )
    elseif self.card_type == Widget.BattleCard then
        self.panel_bg:SetTintAlpha( 0.98 )
        self.icon:SetTexture( engine.asset.Texture("ui/card_type_fight.tex") )
        self.panel_glow:SetTintColour( UICOLOURS.FIGHT )
        self.panel_glow:SetTintAlpha( 0.4 )
        self.icon_container:SetTintColour( UICOLOURS.FIGHT )
        self.panel_border_top_right:SetTintColour( UICOLOURS.FIGHT )
        self.panel_border_top_left:SetTintColour( UICOLOURS.FIGHT )
        self.panel_border_bottom_right:SetTintColour( UICOLOURS.FIGHT )
        self.panel_border_bottom_left:SetTintColour( UICOLOURS.FIGHT )
        self.label:SetGlyphColour( UICOLOURS.FIGHT )
    end
    
    self:Hide()
    self:SetTintAlpha( 0 )
end

function CardsPanelBannerWidget:ShowBanner( text, duration, color_top, color_bottom, icon_on_right )
    local h = 80

    -- Size up the background
    self.panel_bg:SetSize( self.screen_w, h )
    local grid_w, grid_h = self.panel_grid:GetSize()
    self.panel_grid:SetSize( grid_w * h/grid_h,  h )
    self.panel_border_top_right:SetSize( self.screen_w, 1 )
    self.panel_border_top_left:SetSize( self.screen_w, 1 )
    self.panel_border_bottom_right:SetSize( self.screen_w, 1 )
    self.panel_border_bottom_left:SetSize( self.screen_w, 1 )
    self.icon_on_right = icon_on_right

    self.label:SetText( loc.toupper( text ))
    if self.card_type == Widget.NegotiationCard then
        self.icon:SetTexture( engine.asset.Texture("ui/card_type_negotiation.tex") )
    elseif self.card_type == Widget.BattleCard then
        self.icon:SetTexture( engine.asset.Texture("ui/card_type_fight.tex") )
    end

    -- Enemy actions should have the icon on their side
    if self.icon_on_right then
        self.label:LayoutBounds( "before", "center", self.icon_container ):Offset( -30, 0 )
        self.content_divider_top:SetTexture( engine.asset.Texture("large/dialog_divider_bottom.tex" ) )
        self.content_divider_bottom:SetTexture( engine.asset.Texture("large/dialog_divider_bottom.tex" ) )
    else
        self.label:LayoutBounds( "after", "center", self.icon_container ):Offset( 30, 0 )
        self.content_divider_top:SetTexture( engine.asset.Texture("large/dialog_divider_top.tex" ) )
        self.content_divider_bottom:SetTexture( engine.asset.Texture("large/dialog_divider_top.tex" ) )
    end
    
    -- Layout the background
    self.icon:LayoutBounds( "center", "center", self.icon_container ):Offset( 0, 0 )
    self.content:LayoutBounds( "center", "center", self.panel_bg ):Offset( 0, 0 )
    self.content_divider_top:Hide()
    self.content_divider_bottom:Hide()
    if color_top then 
        self.content_divider_top:SetTintColour( color_top ):Show()
        self.content_divider_top:LayoutBounds( icon_on_right and "before" or "after", nil, self.icon_container ):Offset( icon_on_right and 140 or -170, 0 )
    end
    if color_bottom then
        self.content_divider_bottom:SetTintColour( color_bottom ):Show()
        self.content_divider_bottom:LayoutBounds( icon_on_right and "before" or "after", nil, self.icon_container ):Offset( icon_on_right and 140 or -170, 0 )
    end
    self.panel_border_top_right:LayoutBounds( nil, "above", self.panel_bg ):Offset( 0, -1 )
    self.panel_border_top_left:LayoutBounds( nil, "above", self.panel_bg ):Offset( 0, -1 )
    self.panel_border_bottom_right:LayoutBounds( nil, "below", self.panel_bg ):Offset( 0, 1 )
    self.panel_border_bottom_left:LayoutBounds( nil, "below", self.panel_bg ):Offset( 0, 1 )

    local gap = 40
    self.panel_border_top_right:LayoutBounds( "before", nil, self.icon_container ):Offset( gap, 0 )
    self.panel_border_top_left:LayoutBounds( "after", nil, self.icon_container ):Offset( -gap, 0 )
    self.panel_border_bottom_right:LayoutBounds( "before", nil, self.icon_container ):Offset( gap, 0 )
    self.panel_border_bottom_left:LayoutBounds( "after", nil, self.icon_container ):Offset( -gap, 0 )

    self.content_divider_top:LayoutBounds( nil, "above", self.panel_bg ):Offset( 0, -4 )
    self.content_divider_bottom:LayoutBounds( nil, "below", self.panel_bg ):Offset( 0, 0 )

    -- Animate in
    self:Show():SetTintAlpha( 0 )
    self:AlphaTo( 1, 0.2, easing.inOutQuad )

    if duration then
        self:Delay( duration * 0.75 )
        self:HideBanner()
    end
end

function CardsPanelBannerWidget:UpdateText( text )
    self.label:SetText( loc.toupper( text ))
    if self.icon_on_right then
        self.label:LayoutBounds( "before", "center", self.icon_container ):Offset( -30, 0 )
    else
        self.label:LayoutBounds( "after", "center", self.icon_container ):Offset( 30, 0 )
    end
end

function CardsPanelBannerWidget:HideBanner()
    self:AlphaTo( 0, 0.6, easing.inOutQuad, function() self:Hide() end )
end

--------------------------------------------------------------
-- CardsPanel Ending Banner Widget

local CardsPanelEndingBanner = class("Widget.CardsPanelEndingBanner", Widget)

function CardsPanelEndingBanner:init()
    CardsPanelEndingBanner._base.init(self)

    -- Get screen dimensions
    local screen_w, screen_h = TheGame:FE():GetScreenDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- A container for all the background elements
    self.panel = Widget():BlockMouse()
    self.panel_glow = self.panel:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), 1200, 300 ) )
    self.panel_glow:SetHiddenBoundingBox( true )
    self.panel_bg = self.panel:AddChild( Widget.Image( engine.asset.Texture("large/dialog_bg.tex" ) ) )
    self.panel_grid = self.panel:AddChild( Widget.Image( engine.asset.Texture("large/banner_grid.tex" ) ) )
    self.panel_border_top_right = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_top_left = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_bottom_right = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )
    self.panel_border_bottom_left = self.panel:AddChild( Widget.SolidBox( RES_X, 3, 0xffffffff ) )

    -- Our actual content!
    self.content = Widget()
    self.icon_container = Widget.Image( engine.asset.Texture("ui/banner_icon_container.tex" ), 190, 190 )
    self.icon_container:Bloom( 0.05 )
    self.icon_container:SetHiddenBoundingBox( true )
    self.icon = Widget.Image( engine.asset.Texture("ui/card_type_fight.tex"), 70, 70 ):SetBloom(0.1)
    self.label = Widget.Label( "title", 38 )

    self:AddChildren{
        self.panel,
        self.content:AddChildren{
            self.icon_container,
            self.icon,
            self.label
        }
    }
    
    self:Hide()
    self:SetTintAlpha( 0 )
end

function CardsPanelEndingBanner:ShowBanner( victory, txt )
    local h = 120

    -- Size up the background
    self.panel_bg:SetSize( self.screen_w, h )
    local grid_w, grid_h = self.panel_grid:GetSize()
    self.panel_grid:SetSize( grid_w * h/grid_h,  h )
    self.panel_border_top_right:SetSize( self.screen_w, 1 )
    self.panel_border_top_left:SetSize( self.screen_w, 1 )
    self.panel_border_bottom_right:SetSize( self.screen_w, 1 )
    self.panel_border_bottom_left:SetSize( self.screen_w, 1 )

    if txt == nil then
        if victory then
            txt = LOC"UI.FIGHT_SCREEN.VICTORY"
        else
            txt = LOC"UI.FIGHT_SCREEN.DEFEAT"
        end
    end
    self.label:SetText( txt )

    -- Set colors
    if victory then
        self.panel_bg:SetTintAlpha( 0.98 )
        self.panel_glow:SetTintColour( UICOLOURS.SUBTITLE )
        self.panel_glow:SetTintAlpha( 0.4 )
        self.icon_container:SetTintColour( UICOLOURS.SUBTITLE )
        self.panel_border_top_right:SetTintColour( UICOLOURS.SUBTITLE )
        self.panel_border_top_left:SetTintColour( UICOLOURS.SUBTITLE )
        self.panel_border_bottom_right:SetTintColour( UICOLOURS.SUBTITLE )
        self.panel_border_bottom_left:SetTintColour( UICOLOURS.SUBTITLE )
        self.label:SetGlyphColour( UICOLOURS.SUBTITLE )
    else
        self.panel_bg:SetTintAlpha( 0.98 )
        self.panel_glow:SetTintColour( UICOLOURS.HEALTH )
        self.panel_glow:SetTintAlpha( 0.4 )
        self.icon_container:SetTintColour( UICOLOURS.HEALTH )
        self.panel_border_top_right:SetTintColour( UICOLOURS.HEALTH )
        self.panel_border_top_left:SetTintColour( UICOLOURS.HEALTH )
        self.panel_border_bottom_right:SetTintColour( UICOLOURS.HEALTH )
        self.panel_border_bottom_left:SetTintColour( UICOLOURS.HEALTH )
        self.label:SetGlyphColour( UICOLOURS.HEALTH )
    end

    -- Victory shows the icon on the left, defeat shows the icon on the right
    if victory then
        self.label:LayoutBounds( "after", "center", self.icon_container ):Offset( 30, 0 )
    else
        self.label:LayoutBounds( "before", "center", self.icon_container ):Offset( -30, 0 )
    end

    -- Layout the background
    self.icon:LayoutBounds( "center", "center", self.icon_container ):Offset( 0, 0 )
    self.content:LayoutBounds( "center", "center", self.panel_bg ):Offset( 0, 0 )
    self.panel_border_top_right:LayoutBounds( nil, "above", self.panel_bg ):Offset( 0, -1 )
    self.panel_border_top_left:LayoutBounds( nil, "above", self.panel_bg ):Offset( 0, -1 )
    self.panel_border_bottom_right:LayoutBounds( nil, "below", self.panel_bg ):Offset( 0, 1 )
    self.panel_border_bottom_left:LayoutBounds( nil, "below", self.panel_bg ):Offset( 0, 1 )

    local gap = 40
    self.panel_border_top_right:LayoutBounds( "before", nil, self.icon_container ):Offset( gap, 0 )
    self.panel_border_top_left:LayoutBounds( "after", nil, self.icon_container ):Offset( -gap, 0 )
    self.panel_border_bottom_right:LayoutBounds( "before", nil, self.icon_container ):Offset( gap, 0 )
    self.panel_border_bottom_left:LayoutBounds( "after", nil, self.icon_container ):Offset( -gap, 0 )

    -- Animate in
    local pacing = 0.2
    local vertical_size_change = 40
    local horizontal_change = victory and 30 or -30
    local panel_bg_w, panel_bg_h = self.panel_bg:GetSize()
    local panel_grid_w, panel_grid_h = self.panel_grid:GetSize()
    local panel_border_top_right_x, panel_border_top_right_y = self.panel_border_top_right:GetPos()
    local panel_border_top_left_x, panel_border_top_left_y = self.panel_border_top_left:GetPos()
    local panel_border_bottom_right_x, panel_border_bottom_right_y = self.panel_border_bottom_right:GetPos()
    local panel_border_bottom_left_x, panel_border_bottom_left_y = self.panel_border_bottom_left:GetPos()
    local label_x, label_y = self.label:GetPos()
    
    self:Show():SetTintAlpha( 0 )
    self.panel_bg:SizeTo( panel_bg_w, panel_bg_w, panel_bg_h + vertical_size_change*2, panel_bg_h, pacing )
    self.panel_grid:SizeTo( panel_bg_w + vertical_size_change*6, panel_bg_w, panel_bg_h + vertical_size_change*2, panel_bg_h, pacing )
    self.panel_border_top_left:SetPos( panel_border_top_left_x, panel_border_top_left_y + vertical_size_change)
    self.panel_border_top_left:MoveTo( panel_border_top_left_x, panel_border_top_left_y, pacing )
    self.panel_border_top_right:SetPos( panel_border_top_right_x, panel_border_top_right_y + vertical_size_change)
    self.panel_border_top_right:MoveTo( panel_border_top_right_x, panel_border_top_right_y, pacing )
    self.panel_border_bottom_left:SetPos( panel_border_bottom_left_x, panel_border_bottom_left_y - vertical_size_change)
    self.panel_border_bottom_left:MoveTo( panel_border_bottom_left_x, panel_border_bottom_left_y, pacing )
    self.panel_border_bottom_right:SetPos( panel_border_bottom_right_x, panel_border_bottom_right_y - vertical_size_change)
    self.panel_border_bottom_right:MoveTo( panel_border_bottom_right_x, panel_border_bottom_right_y, pacing )
    self.label:SetPos( label_x + horizontal_change, label_y )
    self.label:MoveTo( label_x, label_y, pacing )
    self.label:StartCoroutine( function()
        local dt = 0
        while dt <= 1.0 do
            local bloom = 0.3 * math.sin( dt * math.pi )
            self.label:SetBloom( bloom )
            dt = dt + coroutine.yield()
        end
    end )

    self:AlphaTo( 1, pacing, easing.inOutQuad )
    self:Delay( 2 )

    -- Animate out
    self.panel_bg:SizeTo( panel_bg_w, panel_bg_w, panel_bg_h, panel_bg_h + vertical_size_change*2, pacing )
    self.panel_grid:SizeTo( panel_bg_w, panel_bg_w + vertical_size_change*6, panel_bg_h, panel_bg_h + vertical_size_change*2, pacing )
    self.panel_border_top_left:SetPos( panel_border_top_left_x, panel_border_top_left_y)
    self.panel_border_top_left:MoveTo( panel_border_top_left_x, panel_border_top_left_y + vertical_size_change, pacing )
    self.panel_border_top_right:SetPos( panel_border_top_right_x, panel_border_top_right_y)
    self.panel_border_top_right:MoveTo( panel_border_top_right_x, panel_border_top_right_y + vertical_size_change, pacing )
    self.panel_border_bottom_left:SetPos( panel_border_bottom_left_x, panel_border_bottom_left_y)
    self.panel_border_bottom_left:MoveTo( panel_border_bottom_left_x, panel_border_bottom_left_y - vertical_size_change, pacing )
    self.panel_border_bottom_right:SetPos( panel_border_bottom_right_x, panel_border_bottom_right_y)
    self.panel_border_bottom_right:MoveTo( panel_border_bottom_right_x, panel_border_bottom_right_y - vertical_size_change, pacing )
    self.label:SetPos( label_x, label_y )
    self.label:MoveTo( label_x + horizontal_change, label_y, pacing )

    self:AlphaTo( 0, pacing, easing.inOutQuad )
end

--------------------------------------------------------------
-- Panel that allows the user to select cards from their hand

local ChooseCardsFromHand = class( "Widget.ChooseCardsFromHand", Widget )

ChooseCardsFromHand.CONTROL_MAP = 
{
    
    -- {
    --     control = Controls.Digital.GAMEPAD_X,
    --     fn = function(self) 
    --         self:TogglePopup()
    --         return true
    --     end,
    --     hint = function(self, left, right)
    --         table.insert(right, loc.format ( self.is_visible and LOC"UI.CONTROLS.VIEW_HAND" or LOC"UI.CONTROLS.VIEW_CHOICES", Controls.Digital.GAMEPAD_X))
    --     end,
    -- }
}

function ChooseCardsFromHand:init( height, txt, cards_widget, min_count, max_count, fn, source )
    ChooseCardsFromHand._base.init(self)

    self.height = height or 330
    self.txt = txt
    self.cards_widget = cards_widget
    self.min_count = min_count
    self.max_count = max_count
    self.fn = fn

    AUDIO:PlayEvent( "event:/sfx/battle/cards/gen/choose_cards_prompt" )

    self.overlay = self:AddChild( Widget.SolidBox( RES_X, RES_Y, 0x00000099 ) )
        :SetTintAlpha(0)
        :BlockMouse()
        :SetAnchors( "fill", "fill" )
        :Show()
        :AlphaTo( 1.0, 0.4 )

    self.bg = self:AddChild( Widget.WidePanel( self.height ) )

    self.label = self:AddChild( Widget.Label( "body", 36 ) )
        :SetRegionSize( 700, 40 )
        :CenterH()
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
        :SetText( txt )
        :LayoutBounds( "center", "top", self.bg )
        :Offset( 0, -15 )

    self.subtitle = self:AddChild( Widget.Label( "body", 32 ) )
        :CenterH()
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
        :SetHoverLinkFn( UIHelpers.ItemTooltips )

    local subtitle
    if is_instance( source, CardEngine.Card ) then
        subtitle = loc.format( LOC "CARD_ENGINE.SOURCE_CARD", source.id )
    elseif is_instance( source, CombatCondition ) or is_instance( source, Negotiation.Modifier ) then
        if source.hidden and source.graft then
            subtitle = loc.format( LOC "CARD_ENGINE.SOURCE_GRAFT", source.graft.id)
        else
            subtitle = loc.format( LOC "CARD_ENGINE.SOURCE_OTHER", source:GetName() )
        end
    elseif is_instance( source, GraftInstance ) then
        subtitle = loc.format( LOC "CARD_ENGINE.SOURCE_GRAFT", source.id )
    end

    if subtitle ~= nil then
        self.subtitle:SetText( subtitle ):LayoutBounds( "center", "below", self.label )
    else
        self.subtitle:Hide()
    end

    self.button = self:AddChild( Widget.IconButton() )
        :SetText( LOC"UI.NEGOTIATION_PANEL.CONFIRM_BTN" )
        :SetIcon( global_images.accept )
        :SetShown( self.min_count ~= self.max_count )
        :SetAnchors( "right", "bottom" )
        -- :LayoutBounds( "right", nil, self.overlay )
        :Offset( -120, self.height - 40 )
        :SetFn( self.fn )

    self.hide_btn = self:AddChild( Widget.IconButton() )
        :SetText( LOC"CHOOSE_CARDS_POPUP.MINIMIZE_TITLE", LOC"CHOOSE_CARDS_POPUP.MINIMIZE_DESC" )
        :SetIcon( engine.asset.Texture("UI/ic_preview_open.tex" ) )
        :SetupFocusGlow()
        :SetAnchors( "left", "bottom" )
        :Offset( 120, self.height - 40 )
        :SetFn( function() self:TogglePopup() end )

    self.is_visible = true
end

function ChooseCardsFromHand:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
end

function ChooseCardsFromHand:OnControlModeChange( cm, device_id )
    ChooseCardsFromHand._base.OnControlModeChange( self, cm, device_id )

    self.choose_cards_dialog_bottom_offset = 10
    if cm == CONTROL_MODE.MOUSE_KEYBOARD
    or cm == CONTROL_MODE.TOUCH then
        self.hide_btn:Show()
        self.button:SetShown( self.is_visible and self.min_count ~= self.max_count )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.hide_btn:Hide()
        self.button:Hide()
    end
end

function ChooseCardsFromHand:SetOnInspectHandFn( on_inspect_hand )
    self.on_inspect_hand = on_inspect_hand
    return self
end

function ChooseCardsFromHand:SetOnInspectOptionsFn( on_inspect_options )
    self.on_inspect_options = on_inspect_options
    return self
end

function ChooseCardsFromHand:GetConfirmButton()
    return self.button
end

function ChooseCardsFromHand:HasConfirmButton()
    return self.min_count ~= self.max_count
end

function ChooseCardsFromHand:Confirm()
    if self.min_count ~= self.max_count then
        self.fn()
    end
    return self
end

function ChooseCardsFromHand:TogglePopup()
    if self.is_visible then
        AUDIO:PlayEvent( "event:/sfx/battle/cards/gen/cardselect_popup/show_hand" )
        self:HidePopup()
    else
        AUDIO:PlayEvent( "event:/sfx/battle/cards/gen/cardselect_popup/hide_hand" )
        self:RestorePopup()
    end
end

function ChooseCardsFromHand:HidePopup()
    if self.on_inspect_hand then self.on_inspect_hand( self ) end
    self.is_visible = false
    self.overlay:AlphaTo( 0, 0.1, easing.linear, function() self.overlay:Hide() end )
    self.bg:AlphaTo( 0, 0.15, easing.linear, function() self.overlay:Hide() end )
    self.button:Hide()
    self.hide_btn:SetText( LOC"CHOOSE_CARDS_POPUP.RESTORE_TITLE", LOC"CHOOSE_CARDS_POPUP.RESTORE_DESC" )
        :ShowFocusGlow()
        :SetIcon( engine.asset.Texture("UI/ic_preview_close.tex" ) )
end

function ChooseCardsFromHand:RestorePopup()
    if self.on_inspect_options then self.on_inspect_options( self ) end
    self.is_visible = true
    self.overlay:Show():AlphaTo( 1.0, 0.1 )
    self.bg:Show():AlphaTo( 1.0, 0.15 )
    self.button:SetShown( self.min_count ~= self.max_count 
        and ( TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD
        or TheGame:FE():GetControlMode() == CONTROL_MODE.TOUCH ) )
    self.hide_btn:SetText( LOC"CHOOSE_CARDS_POPUP.MINIMIZE_TITLE", LOC"CHOOSE_CARDS_POPUP.MINIMIZE_DESC" )
        :HideFocusGlow( true )
        :SetIcon( engine.asset.Texture("UI/ic_preview_open.tex" ) )
end

function ChooseCardsFromHand:IsVisible()
    return self.is_visible
end
