require "ui/widget"
require "ui/widgets/plax"

local CharacterStage = class("Widget.CharacterStage", Widget)

local NODE_TYPE = Widget.Plax.NODE_TYPE

function CharacterStage:init()
    CharacterStage._base.init(self)
    self.nodebug = true
    self.ignore_input = true
    self.plax = Widget.Plax():SetName("CHARACTERSTAGE_PLAX")
    -- self.plax:TrackMouse( true )

    self:AddChild( self.plax )

    self:SetPlax()
    self:StartUpdating()
end

function CharacterStage:SetPlax()
    local prev_characters = self.characters

    self.characters = {}
    self.remove_characters = {}

    self.plax:Clear()

    local scale = 1.0
    local x, y

    local aspect_ratio = TheGame:FE():GetAspectRatio()
    if aspect_ratio == ASPECT_RATIO.R4_3 then
        x, y = 630, -25
    elseif aspect_ratio == ASPECT_RATIO.R16_10 or aspect_ratio == ASPECT_RATIO.R16_9 then
        x, y = 660, -430
    elseif aspect_ratio == ASPECT_RATIO.R21_9 then
        x, y = 800, -430
    else
        x, y = 800, -430
    end

    local layer  = {elements = {}, name = "CHARACTERS", plax = 30}
    table.insert(layer.elements, {node_type=NODE_TYPE.SLOT,character_role = "LEFT", x=-x, y=y, scale=scale })
    table.insert(layer.elements, {node_type=NODE_TYPE.SLOT, flip = true, character_role = "RIGHT", x=x,y=y, scale=scale })

    local plax_layers = { layer }
    self.plax:SetLayers(plax_layers)

    if prev_characters then
        for agent, t in pairs( prev_characters ) do
            local body = self:AddCharacter( agent, t.left, 0 )
            if t.posedef then
                self:SetPoseDef( agent, t.posedef )
            end
        end
    end

    return self
end

function CharacterStage:OnScreenResize( w, h )
    CharacterStage._base.OnScreenResize( self, w, h )
    self:SetPlax()
end

function CharacterStage:SetWindy( windy )
    self.plax:SetWindy( windy )
end

local function AnimLerp(val, target, dt, c)
    c = c or 1
    val = val or 0
    target = target or 0
    local delta = target - val
    delta = math.min(1, c*dt)*delta
    if math.abs(delta) < 0.001 then
        return target
    else
        return val + delta
    end
end

function CharacterStage:OnUpdate(dt)
    CharacterStage._base.OnUpdate(self, dt)

    self.scale_val = AnimLerp(self.scale_val or 1, self.scale_target or 1, dt, 2)
    self.offset_val = AnimLerp(self.offset_val or 0, self.offset_target or 0, dt, 2)
    
    if self.scale_target then
        self.plax:SetScale(self.scale_val)
    end
    if self.offset_target then
        self.plax:SetPos( nil, self.offset_val )
    end
   
    for k,char in pairs(self.characters) do
        char.scale_val = AnimLerp(char.scale_val or 1 , char.scale_target or 1, dt, 3)
        char.x_val = AnimLerp(char.x_val or 0 , char.x_target or 0, dt, 3)
        if self.char_movement_fn then self.char_movement_fn( char.x_val ~= char.x_target, k ) end
        char.body:SetPos(char.x_val, char.y_val or 0):SetScale(char.scale_val, char.scale_val)
    end
    
    for k = #self.remove_characters, 1, -1 do
        local char = self.remove_characters[k]
        if not char.body.removed then
            char.x_val = AnimLerp(char.x_val or 0 , char.x_target or 0, dt, 2)
            if self.char_movement_fn then self.char_movement_fn( char.x_val ~= char.x_target, k ) end
            char.body:SetPos(char.x_val, char.y_val)
            if math.abs(char.x_val - char.x_target) <= 5 then
                char.body:Remove()
                table.remove(self.remove_characters, k)
            end
        end
    end
end

function CharacterStage:GetCharacterSlot( agent )
    return self.plax:GetAgentSlot( agent ).widget
end

function CharacterStage:GetCharacterPos( agent )
    if self.characters[agent] then return self.characters[agent].body:GetPos() end
    return 0
end

function CharacterStage:MoveCharacterToXPos( agent, target_x )
    if self.characters[agent] then 
        self.characters[agent].x_target = target_x
    end
end

-- Allows for a callback to be triggered when the character starts moving, and when it stops
-- Parameters: 
--- moving (true if it's moving, false if it's stopped)
--- agent (the respective character)
function CharacterStage:SetCharacterMovingFn( fn )
    self.char_movement_fn = fn
end

function CharacterStage:IsCharacterLeft( agent )
    return self.characters[agent] and self.characters[agent].left
end

function CharacterStage:IsCharacterRight( agent )
    return self.characters[agent] and not self.characters[agent].left
end

function CharacterStage:GetLeftAnim()
    for k, v in pairs( self.characters ) do
        if v.left then
            return v.body
        end
    end
end

function CharacterStage:GetRightAnim()
    for k, v in pairs( self.characters ) do
        if not v.left then
            return v.body
        end
    end
end


function CharacterStage:GetCharacter( is_left )
    for k, v in pairs( self.characters ) do
        if v.left == is_left then
            return v.data
        end
    end
end

function CharacterStage:AddCharacter(character_data, is_left, x_val)
    assert( is_instance( character_data, Agent ))

    -- Remove whatever is currently specified by this facing.
    for k, v in pairs( self.characters ) do
        if v.data == character_data and v.left == is_left then
            -- Already added here.
            return v.body

        elseif v.left == is_left then
            -- Somebody not us is in thie slot: remove em.
            self:RemoveCharacter( v.data )

        elseif v.data == character_data then
            -- We are in a different slot. Remove us from there!
            self:RemoveCharacter( v.data )
        end
    end
    
    --local body = Widget.AnimatedCharacter():SetScale(anim_scale * (is_left and 1 or -1), anim_scale):Refresh(character_data):Play("idle_neutral", true)
    local body, slot = self.plax:FillSlot(character_data, is_left and "LEFT" or "RIGHT")
    body:ClearPose()

    --print ("ADD CHARACTER TO STAGE", character_data.name, character_data)
    self.characters[ character_data ] = {
        body = body,    
        data= character_data,
        left= is_left == true,
        x_val = x_val or -500,
        x_target = 0,
        x_dest = 0,
        slot = slot.slot_id,
    }

    body:SetPos( self.characters[ character_data ].x_val, 0)

    AUDIO:PlayEvent(SoundEvents.switch)


    return body
end

function CharacterStage:Emote(agent, emote)
    local emotes = require "content/convo_emotes"
    if emote == nil or emotes[ emote ] == nil then
        print( "BAD EMOTE: ", emote )
    elseif not self.characters[agent] then
        print ("NO CHARACTER HERE", agent, emote)
    else
        self.characters[agent].body:Emote( emote )
    end
end



function CharacterStage:DoCharacterSpeech(agent, is_question, voice_delay)
    if self.characters[agent] then
        local character_widget = self.characters[ agent ].body
         
        character_widget:Talk(is_question, voice_delay)
    end
end

function CharacterStage:DoCharacterSpeechSpecific(agent, event_info)
    if self.characters[agent] then
        local character_widget = self.characters[ agent ].body
        character_widget:TalkSpecific(event_info)
    end
end


function CharacterStage:SetMood(agent, mood)
    if self.characters[agent] then
        local character_widget = self.characters[ agent ].body
        character_widget:SetMood(mood)
        character_widget.preferred_mood = nil
    else
        print( "missing character: ", agent.name, debug.traceback() )
    end
end


function CharacterStage:SetVoice(agent, voice)
    if self.characters[agent] then
        local character_widget = self.characters[ agent ].body
        character_widget:SetVoice(voice)
    else
        print( "missing character: ", agent and agent.name or "??", debug.traceback() )
    end
end

function CharacterStage:SetPoseDef(agent, posedef)
    local posedefs = require "content/character_pose_defs"
    if self.characters[agent] then
        local character_widget = self.characters[ agent ].body
        character_widget:SetPoseDef(posedef)

        self.characters[ agent ].posedef = posedef

        local y_val = 0

        -- Conversation pose has a different root point than the other poses. (?)
        local aspect_ratio = TheGame:FE():GetAspectRatio()
        if posedefs[posedef] and posedefs[posedef].convo then
            if aspect_ratio == ASPECT_RATIO.R4_3 then
                y_val = -RES_Y/6
            else
                y_val = 0
            end
        else
            y_val = -RES_Y/2
        end

        self.characters[ agent ].y_val = y_val

    else
        print( "missing character: ", agent.name, debug.traceback() )
    end
end

function CharacterStage:StopAllTalking()
    for k,v in pairs(self.characters) do
        v.body:StopTalking()
        v.body:KillEmotes()
    end
end

function CharacterStage:Clear()
    self:RemoveAllCharacters()
end

function CharacterStage:RemoveAllCharacters()
    for k,v in pairs(self.characters) do
        self:RemoveCharacter(k)
    end
end

function CharacterStage:RemoveCharacter(agent)
    -- print("CharacterStage:RemoveCharacter", agent)
    self.plax:UnhookAgentWidget(agent)
    local char = self.characters[agent]
    if char then
        char.x_target = -1000
        self.characters[agent] = nil
        table.insert(self.remove_characters, char)
    end

end

function CharacterStage:RemoveSide( side )
    for k, v in pairs(self.characters) do
        if v.left == (side == "left") then
            self:RemoveCharacter( k )
            return
        end
    end
end

function CharacterStage:StopTalk(agent)
    if not self.characters[agent] then
        return
    end

    local character_widget = self.characters[agent].body
    character_widget:StopTalking()

end

function CharacterStage:Talk(agent, emote, t)

    if not self.characters[agent] then
        return
    end

    local character_widget = self.characters[agent].body

    if self.characters[agent].left then
        self:Left(agent)
    else
        self:Right(agent)
    end

    if emote then
        character_widget:Emote(emote)
    end

end

function CharacterStage:IsSettled()
    return #self.remove_characters == 0
end

function CharacterStage:Center()
    self.plax_target = 0
    self.scale_target = 1
end

function CharacterStage:SetZoom( scale_target, offset_target )
    self.plax_target = 0
    self.scale_target = scale_target or 1
    self.offset_target = offset_target or 0
end

function CharacterStage:Left(char)
    self.plax_target = -1

    for k,v in pairs(self.characters) do
        v.scale_target = char == k and 1.025 or 1
    end
end

function CharacterStage:Right(char)
    self.plax_target = 1

    for k,v in pairs(self.characters) do
        v.scale_target = char == k and 1.025 or 1
    end

end
