require "ui/widget"


local Clickable = class( "Widget.Clickable", Widget )
Clickable.can_focus = true

Clickable.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.enabled and not self:IsAnimatingOut() and self.onclickfn
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.SELECT", Controls.Digital.MENU_ACCEPT))
        end
    },
}

local LONG_PRESS_DELAY = 0.4 -- How long to press down before the callback is called
local DOUBLE_CLICK_DELAY = 0.5 -- Max delay between two clicks to be considered a double click

function Clickable:init( )
    Clickable._base.init( self )
    self.hover_check = true
    self:SetScales(1,1,1)

    self.base_r = 1
    self.base_g = 1
    self.base_b = 1
    self.base_a = 1

    self.animate_scale = nil
    self.animate_scales = true
end

function Clickable:OnAdded()
end

function Clickable:OnRemoved()
    self:StopUpdating()
end

function Clickable:DisableAnimateScales()
    self.animate_scales = false
    return self
end

function Clickable:IsEnabled()
    return self.enabled and not self:IsAnimatingOut()
end

function Clickable:SetScales(normal, hover, down, animate)
    -- Send in a value in seconds if you want this to animate. Something like 0.1
    self.animate_scale_timing = animate or nil
    self.scale_normal = normal
    self.scale_hover = hover
    self.scale_down = down
end

function Clickable:SetOnClickFn(onclickfn)
    self.onclickfn = onclickfn
    return self
end

function Clickable:GetOnClickFn()
    return self.onclickfn
end

function Clickable:SetOnRightClickFn(onclickfn)
    self.onaltclickfn = onclickfn
    return self
end

function Clickable:SetOnDoubleClickFn(onclickfn)
    self.ondoubleclickfn = onclickfn
    return self
end

function Clickable:SetOnLongClickFn(onclickfn)
    self.onlongclickfn = onclickfn
    if self.onlongclickfn and self:IsShown() then self:StartUpdating() end
    return self
end

function Clickable:SetOnHighlight( fn )
    self.hilite_fn = fn
    return self
end

function Clickable:SetBGMult( m )
    if self.highlight_widget then
        self.highlight_widget:SetTintColour(m,m,m,1)
    end
    return self
end

function Clickable:HotkeyDown()
    if not self.down and self:IsEnabled() then
        self.down = true
        self:UpdateHighlight()
    end
    return true
end

function Clickable:HotkeyUp()
    if self.down then
        self.down = false
        self:UpdateHighlight()
        if self.onclickfn then self.onclickfn("hotkey") end        
        return true
    end
end

function Clickable:OnGainFocus()
    if self.enabled and not self:IsAnimatingOut() and not self.silent then AUDIO:PlayEvent(SoundEvents.mouseover) end
    self:UpdateHighlight()
end

function Clickable:OnLoseFocus()
    self:UpdateHighlight()
end

function Clickable:UpdateHighlight()
    if self.removed then
        return
    end
    if self.animate_scales then
        if self.down then
            if self.animate_scale_timing then self:ScaleTo(nil,self.scale_down,self.animate_scale_timing) else self:SetScale(self.scale_down) end
            self:SetBGMult(0.9)
        elseif self.hover or self.selected or self.focus then
            if self.animate_scale_timing then self:ScaleTo(nil,self.scale_hover,self.animate_scale_timing) else self:SetScale(self.scale_hover) end
            self:SetBGMult(1.8)
        else
            if self.animate_scale_timing then self:ScaleTo(nil,self.scale_normal,self.animate_scale_timing) else self:SetScale(self.scale_normal) end
            self:SetBGMult(1)
        end
    end
    if self.hilite_fn then
        self.hilite_fn( self.down, self.hover, self.selected, self.focus )
    end
end

function Clickable:SetHighlightWidget(w)
    self.highlight_widget = w
    return self
end

function Clickable:IsSelected()
    return self.selected
end

function Clickable:Select()
    if not self.selected then
        self.selected = true
        if self.HandleSelectChange then
            self:HandleSelectChange(true)
        end
        if self.onselectfn then
            self.onselectfn()
        end
    end

    self:UpdateHighlight()
    return self
end

function Clickable:SetOnSelectFn(onselectfn, ondeseelctfn)
    self.onselectfn = onselectfn
    self.ondeselectfn = ondeseelctfn
    return self
end

function Clickable:Deselect()
    if self.selected then

        if self.HandleSelectChange then
            self:HandleSelectChange(false)
        end

        self.selected = false
        
        if self.ondeselectfn then
            self.ondeselectfn()
        end
    end
    self:UpdateHighlight()
    return self
end

function Clickable:MakeSilent()
    self.silent = true
    return self
end

function Clickable:IsAnimatingOut()
    return self.animating_out and self.animating_out == self.colour
end

function Clickable:AnimateIn()
    self:Show():AlphaTo( 1, self.pacing, easing.outQuad )
    return self
end

function Clickable:AnimateOut()
    self:AlphaTo( 0, self.pacing, easing.outQuad, function() 
        self.animating_out = nil
        self:Hide()
    end )
    self.animating_out = self.colour
    return self
end


function Clickable:OnLoseHover()
    --print (self, "LOSE HOVER")
    self.down = false
    self:UpdateHighlight()
    if self.onhoverfn then
        self.onhoverfn(false)
    end
end

function Clickable:OnGainHover()
    --print (self, "GAIN HOVER")
    if self.enabled and not self:IsAnimatingOut() then
        if not self.silent then AUDIO:PlayEvent(SoundEvents.mouseover) end
        self:UpdateHighlight()
        if self.onhoverfn then
            self.onhoverfn(true)
        end
    end
end

function Clickable:HandleControlDown(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if self.enabled and not self:IsAnimatingOut() and ( self.onclickfn or self.ondoubleclickfn or self.onlongclickfn ) then
            self.down = true
            if self.onlongclickfn then
                self.down_time = os.clock() -- How long has a click/tap been held on this widget
            end

            self:UpdateHighlight()
            return true
        end

    elseif control:Has( Controls.Digital.MENU_ALT ) then
        if self.enabled and not self:IsAnimatingOut() and self.onaltclickfn then
            self.down = true
            if self.onlongclickfn then
                self.down_time = os.clock() -- How long has a click/tap been held on this widget
            end

            self:UpdateHighlight()
            return true
        end
    end
end

function Clickable:OnUpdate( dt )
    Clickable._base.OnUpdate( self, dt )
    if self.onlongclickfn and self.down and self.down_time then
        if ( os.clock() - self.down_time ) > LONG_PRESS_DELAY then
            -- Long press!
            self.down = false
            self:UpdateHighlight()
            self.onlongclickfn()
        end
    end
end

function Clickable:SetClickSoundEvent(snd)
    self.click_sound = snd
    return self
end

-- For when we want a widget to trigger on ControlUp without the ControlDown first
-- Useful when we want this to be a drag target
function Clickable:SetClickWithoutDown()
    self.click_without_down = true
    return self
end

function Clickable:HandleControlUp(control, dev)
    if control:Has( Controls.Digital.MENU_ALT ) then
        if self.down and ( self.onclickfn or self.onaltclickfn ) then
            if not self.silent and self.onaltclickfn then 
                AUDIO:PlayEvent(self.click_sound or SoundEvents.button_click)
            end
            self.down = false
            self:UpdateHighlight()

            if self.onaltclickfn then
                self.onaltclickfn()
            end
            return true
        end
    
    elseif control:Has( Controls.Digital.MENU_ACCEPT ) then    
        if ( self.down or self.click_without_down ) and ( self.onclickfn or self.ondoubleclickfn or self.onaltclickfn ) then
            
            self.down = false
            
            if self.selectable and not self.selected then
                if not self.silent then
                    AUDIO:PlayEvent(self.click_sound or SoundEvents.button_click)
                end

                self:Select()
            else
                if self.ondoubleclickfn and not self.one_click_complete then --  If we had a first click, save its start time
                    self.one_click_complete = true
                    self.double_click_time = os.clock()

                elseif self.ondoubleclickfn and self.one_click_complete then -- If we had a second click
                    if ( os.clock() - self.double_click_time ) < DOUBLE_CLICK_DELAY then
                        -- Double click!
                        if not self.silent then
                            AUDIO:PlayEvent(self.click_sound or SoundEvents.button_click)
                        end

                        self.ondoubleclickfn()
                        self.one_click_complete = false
                    else
                        -- Too long. Cancel out and start a new first click
                        self.one_click_complete = true
                        self.double_click_time = os.clock()
                    end
                end

                if self.onclickfn then
                    if not self.silent then
                        AUDIO:PlayEvent(self.click_sound or SoundEvents.button_click)
                    end

                    self.onclickfn()
                end
            end
            self:UpdateHighlight()
            return true
        end
    end
end

function Clickable:SetHoverFn(fn)
    self.onhoverfn = fn
    return self
end

function Clickable:SetSelectable(val)
    self.selectable = val
    return self
end

function Clickable:Click()
    if self.onclickfn then
        self.onclickfn()
    end
    return self
end

function Clickable:Show()
    Clickable._base.Show(self)
    --if self.onlongclickfn then self:StartUpdating() end
    return self
end

function Clickable:Hide()
    Clickable._base.Hide(self)
    --self:StopUpdating()
    return self
end
