local CONTENT_LEFT_MARGIN = 15 -- Extra size added on the left of scroll boxes, so that the contents can scale up on hover without clipping
local SIDEBAR_LEFT_MARGIN = 10 -- Extra size added on the left of scroll boxes, so that the contents can scale up on hover without clipping
------------------------------------------------------------------------------------------
-- Shows a compendium panel

local CompendiumWidget = class( "Widget.CompendiumWidget", Widget )

function CompendiumWidget:init( width, height, sidebar_width )
    CompendiumWidget._base.init( self )

    self.widget_w = width
    self.widget_h = height
    self.sidebar_width = sidebar_width
    self.content_w = self.widget_w - self.sidebar_width

    self.hitbox = self:AddChild( Widget.SolidBox( self.widget_w, self.widget_h, 0xFFFF0030 ) )
        :Hide()

    -- Add sidebar for filters
    self.sidebar = self:AddChild( Widget.CompendiumSidebar( self.sidebar_width, self.widget_h, function( filters ) self:OnFiltersChanged( filters ) end ) )
        :LayoutBounds( "left", "center", self.hitbox )
        :Offset( -SIDEBAR_LEFT_MARGIN, nil )
        :SetFocusLock( true )
        :SetFocusGroup( FOCUS_GROUP.LAST_FOCUS )

    self:AddSidebarFilters()

    -- Add scrollable area for the contents
    self.scroll_root = self:AddChild( Widget.ScrollPanel() )
        :SetSize( self.content_w+CONTENT_LEFT_MARGIN, self.widget_h-2 )
        :SetFocusLock(true)
        :SetFocusGroup( FOCUS_GROUP.LAST_FOCUS )
        :SetIgnoreStickScroll()
end


function CompendiumWidget:GetItemWidgets()
    return self.agent_widgets or self.graft_widgets or self.card_widgets
end

function CompendiumWidget:ReNav()
    -- Reset all the focus dirs on all the widgets
    local widgets = self:GetItemWidgets()
    for k, widget in ipairs( widgets ) do
        widget:ClearFocusDirs()
    end

    -- Go through each horizontal panel
    if self.visibility_panels then
        for row, vispanel in ipairs( self.visibility_panels ) do
            if not vispanel.children or #vispanel.children < 1 then break end
            -- Go through each person in the panel
            for col, widget in ipairs( vispanel.children ) do
                -- Set move up dir
                if row > 1 then
                    local previous_row = self.visibility_panels[row-1]
                    local widget_above = previous_row.children[col]
                    if widget_above then widget:SetFocusDir( "up", widget_above, true ) end
                end
                -- Set move down dir
                if row < #self.visibility_panels then
                    local next_row = self.visibility_panels[row+1]
                    local widget_below = #next_row.children >= col and next_row.children[col] or nil
                    if widget_below then widget:SetFocusDir( "down", widget_below, true ) end
                end
            end
        end
    end
end

-- Gets all relevant content and starts displaying it
function CompendiumWidget:RefreshContent()
    return self
end

-- Returns the number of unlocked items and the number of unlockable items, to be displayed on the nav bar
function CompendiumWidget:GetCount()
    return 0, 0
end

-- Lays itself out
function CompendiumWidget:Layout()
    return self
end

-- Lays its content out
function CompendiumWidget:LayoutContent()
    return self
end

function CompendiumWidget:OnAdded()
    TheGame:GetEvents():ListenForEvent( "focus_changed", self, self.OnFocusChanged )

    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
end

function CompendiumWidget:OnRemoved()
    TheGame:GetEvents():RemoveListener( self )
end

function CompendiumWidget:OnControlModeChange( cm, device_id )
    CompendiumWidget._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.sidebar:SetScrollEnabled( true )
        self:SetScrollEnabled( true )
    end
    if cm == CONTROL_MODE.TOUCH then
        self.sidebar:SetScrollEnabled( true )
        self:SetScrollEnabled( true )
    end
    if cm == CONTROL_MODE.GAMEPAD then
        -- So that using the gamepad's joystick doesn't scroll both scroll panels
        self.sidebar:SetScrollEnabled( false )
        self:SetScrollEnabled( true )
    end
end

function CompendiumWidget:CanNav()
    return true
end

function CompendiumWidget:SetScrollEnabled( enabled )
    self.scroll_root:SetEnabled( enabled )
end

function CompendiumWidget:HasSidebar()
    return false
end

function CompendiumWidget:ToggleSideBar()
    local old_focus = TheGame:FE():GetFocusWidget()
    
    if self.sidebar_focus then
        if old_focus and self.sidebar:IsAncestorOf( old_focus ) then
            self.last_sidebar_focus_item = old_focus
        end
    else
        if old_focus and self.scroll_root:IsAncestorOf( old_focus ) then
            self.last_focus_item = old_focus
        end
    end
    
    self.sidebar_focus = not self.sidebar_focus

    local new_focus
    if self.sidebar_focus then
        new_focus = self.last_sidebar_focus_item
    else
        new_focus = self.last_focus_item
    end
    
    if not ( new_focus and new_focus:IsVisible() and not new_focus.removed ) then
        if self.sidebar_focus then
            new_focus = self.sidebar:GetDefaultFocus()
        else
            new_focus = self:GetDefaultFocus()
        end
    end
    
    if new_focus then
        new_focus:SetFocus()
    end

    -- So that using the gamepad's joystick doesn't scroll both scroll panels
    if TheGame:FE():IsGamepadMode() then
        self.sidebar:SetScrollEnabled( self.sidebar_focus )
        self:SetScrollEnabled( not self.sidebar_focus )
    end
end
