require "ui/widget"
require "ui/widgets/animatedcharacter"
require "ui/widgets/label"
require "ui/widgets/solidbox"
require "ui/widgets/particlesystem"
require "ui/widgets/tooltipcrew"

local CrewPortraits = class( "Widget.CrewPortraits", Widget )

function CrewPortraits:init( portrait_size )
    Widget.init( self )

    self.portrait_size = portrait_size or 60
    self.portraits = {}

    TheGame:GetEvents():ListenForAny( "loc_changed", self.OnAgentEvent )
    TheGame:GetEvents():ListenForEvent( "caravan_member_remove", self )
    TheGame:GetEvents():ListenForEvent( "caravan_member_add", self )
end

function CrewPortraits:GetPortrait(agent)
    for portrait_agent,v in pairs(self.portraits) do
        if portrait_agent == agent then
            return v
        end
    end
end

local function GetScore(agent)
    if agent:IsCrewMember() then
        return 1
    end
    if agent:IsEscortMember() then
        return 3
    end
    if agent:IsCaptiveMember() then
        return 4
    end
    return 5
end

local sortfn = function(a,b)
    return GetScore(a) < GetScore(b)
end

function CrewPortraits:Refresh()

    if TheGame:GetGameState() then
        local needs_sort = false
        for k, caravan_member in TheGame:GetGameState():GetCaravan():Members() do
            if not caravan_member:IsPlayer() and not self.portraits[caravan_member] then

                -- Portrait containing widget
                local portrait_widget = self:AddChild( Widget.Clickable() )
                    :SetToolTipClass( Widget.TooltipCrew )
                    :SetToolTip( caravan_member )
                    :ShowToolTipOnFocus( self.show_tooltip_on_focus )

                -- Portrait background
                self.bg_border = 16
                self.bg_border_scale = 0.5
                local nineslice_w, nineslice_h = self.portrait_size*1.1 + self.bg_border*2*self.bg_border_scale, self.portrait_size + self.bg_border*2*self.bg_border_scale
                portrait_widget.agent_bg = portrait_widget:AddChild( Widget.Panel( engine.asset.Texture("UI/mainoverlay_icons_bg.tex" ) ) )
                    :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
                    :SetNineSliceBorderScale( self.bg_border_scale )
                    :SetTintColour( UICOLOURS.SUBTITLE )
                    :SetSize( nineslice_w, nineslice_h )
                    :Bloom( 0.1 )

                -- Agent portrait image
                portrait_widget.agent_portrait = portrait_widget:AddChild( Widget.AgentPortrait( self.portrait_size*0.9, caravan_member ) )
                    :SetHiddenBoundingBox( true )
                    :LayoutBounds( "right", "bottom", portrait_widget.agent_bg )
                    :Offset( 0, 3 )
                    :SetScale( -1, 1 )

                -- Input animations!
                portrait_widget.OnGainHover = function( portrait, caravan_member ) self:OnPortraitUpdateImage( portrait, caravan_member ) end
                portrait_widget.OnLoseHover = function( portrait, caravan_member ) self:OnPortraitUpdateImage( portrait, caravan_member ) end
                portrait_widget.OnGainFocus = function( portrait, caravan_member ) self:OnPortraitUpdateImage( portrait, caravan_member ) end
                portrait_widget.OnLoseFocus = function( portrait, caravan_member ) self:OnPortraitUpdateImage( portrait, caravan_member ) end
                
                portrait_widget:SetOnClickFn( function()
                    local details = Screen.PeopleCompendiumDetailsScreen()
                    self:GetFE():PushScreen( details )
                    details:ShowAgent( caravan_member )
                end )

                if self.tooltiplayoutfn then portrait_widget:SetToolTipLayoutFn( self.tooltiplayoutfn ) end

                self.portraits[caravan_member] = portrait_widget
                needs_sort = true
            end
        end

        for agent, w in pairs(self.portraits) do
            if not agent:IsInPlayerParty() then
                self.portraits[agent] = nil
                w:Remove()
                needs_sort = true
            end
        end

        if needs_sort then
            local sorted = {}
            for agent,w in pairs(self.portraits) do
                table.insert(sorted, agent)
            end
            table.sort(sorted, sortfn)

            -- for k, agent in ipairs(sorted) do
            --     self.portraits[agent]:LayoutBounds( "after", "center" ):Offset( 10, 0 )
            -- end
            self:StackChildrenRow( 8 )
        end
        self:Show()
    else
        self:Hide()
    end

    return self
end

function CrewPortraits:HasPortraits()
    return #self.children > 0
end

function CrewPortraits:OnPortraitUpdateImage( portrait, caravan_member )
    if portrait.focus or portrait.hover then
        portrait:RunUpdater( Updater.Ease( function(v) portrait.agent_bg:Bloom( v ) portrait.current_bloom = v end, portrait.current_bloom or 0, 0.1, 0.2, easing.outQuad ) )
        portrait:ColourTo( nil,nil,nil,nil, 1.2,1.2,1.2,1.2, 0.1, easing.outQuad )
    elseif portrait.current_bloom and portrait.current_bloom > 0 then
        portrait:RunUpdater( Updater.Ease( function(v) portrait.agent_bg:Bloom( v ) portrait.current_bloom = v end, portrait.current_bloom or 0.1, 0, 0.2, easing.outQuad ) )
        portrait:ColourTo( nil,nil,nil,nil, 1,1,1,1, 0.05, easing.inOutQuad )
    end
end

function CrewPortraits:OnGlobalEvent( eventname, ... )
    if eventname =="caravan_member_add" or eventname == "caravan_member_remove" then
        self:Refresh()
    end
end

function CrewPortraits:OnAgentEvent( eventname, ... )
    if eventname == "loc_changed" then
        local agent = ...
        self:GetPortrait( agent )
            :SetCharacter( agent )
            :SetToolTip( agent )
    end
end

function CrewPortraits:ShowToolTipOnFocus( show )
    CrewPortraits._base.ShowToolTipOnFocus( self, show )
    for agent, w in pairs(self.portraits) do
        w:ShowToolTipOnFocus( self.show_tooltip_on_focus )
    end
    return self
end