------------------------------------------------------------------------------------------
-- A panel with a list of the feats available on this Daily Challenge

local DailyFeatsList = class("Widget.DailyFeatsList", Widget)

local title_top_margin_fudge_factor = -6

function DailyFeatsList:init( max_width )
    DailyFeatsList._base.init(self)

    self.max_width = max_width or 200
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture( "UI/left_border_frame_faint.tex" ) ) )
        :SetSize( self.max_width, 250 )
        :SetTintColour( UICOLOURS.MUTATOR )
        :SetBloom(0.1)
    self.content = self:AddChild( Widget() )
    self.title = self.content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_LABEL, LOC"UI.DAILY_CHALLENGE_SCREEN.FEATS_TITLE" ) )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :SetAutoSize( self.max_width - SPACING.M1*2 )
        :SetWordWrap( true )
        :Bloom( 0.05 )
    self.feats_container = self.content:AddChild( Widget() )
    
end

function DailyFeatsList:Refresh( daily_challenge )

    -- Remove old mutator widgets
    self.feats_container:DestroyAllChildren()

    -- Add new ones
    local feats = {}
    if daily_challenge.feats then
        for i,feat in ipairs(daily_challenge.feats) do
            local def = Content.GetGraft(feat)
            local instance = GraftInstance(feat)
            table.insert(feats, 
            {
                icon = engine.asset.Texture( def.feat_icon ),
                title = instance:GetName(),
                desc = instance:GetDesc(),
                points = def.points,
            })
        end
    end

    for k, feat in ipairs(feats) do
        self.feats_container:AddChild( Widget.DailyFeatListItem( self.max_width - SPACING.M1*2, feat.icon, feat.title, feat.desc, feat.points ) )
            :LayoutBounds( "left", "below" )
            :Offset( 0, -10 )
    end

    self:Layout()
    return self
end

function DailyFeatsList:Layout()
    -- Layout things
    self.feats_container:LayoutBounds( "left", "below", self.title )
        :Offset( 0, -SPACING.M1/2 )

    -- Resize background
    local w, h = self.content:GetSize()
    w = w + SPACING.M1*2
    h = h + SPACING.M1*2+title_top_margin_fudge_factor
    self.bg:SetSize( self.max_width, h )
        :SetScale( 1, 1 )
    self.content:LayoutBounds( "left", "bottom", self.bg )
        :Offset( SPACING.M1, SPACING.M1 )
    self.bg:SetScale( -1, 1 )

    return self
end



------------------------------------------------------------------------------------------
-- A single feat on the list

local DailyFeatListItem = class("Widget.DailyFeatListItem", Widget)

local icon_spacing = 10

function DailyFeatListItem:init( max_width, icon, title, desc, points )
    DailyFeatListItem._base.init(self)

    self.max_width = max_width or 200
    self.icon_size = 60
    -- Add contents
    self.icon = self:AddChild( Widget.Image( icon ) )
        :SetSize( self.icon_size, self.icon_size )
        :Bloom( 0.1 )
    self.pts_floater = self:AddChild( Widget.PtsFloater( points, FONT_SIZE.ITEM_DESC * 0.8 ) )
        :Bloom( 0.1 )
    self.text_content = self:AddChild( Widget() )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, title ) )
        :LeftAlign()
        :SetAutoSize( self.max_width-self.icon_size-icon_spacing )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.MUTATOR )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, desc ) )
        :LeftAlign()
        :SetAutoSize( self.max_width-self.icon_size-icon_spacing )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )

    -- Layout stuffs
    self.pts_floater:LayoutBounds( "center", "below", self.icon ):Offset( 0, 6 )
    self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    self.text_content:LayoutBounds( "after", "top", self.icon ):Offset( icon_spacing, 0 )
end

------------------------------------------------------------------------------------------
-- A horizontal icon list with the feats available in this challenge, and their state
-- Shown on the Main Overlay top bar during a Daily Challenge

local DailyFeatsStatus = class("Widget.DailyFeatsStatus", Widget)

function DailyFeatsStatus:init( icon_size )
    DailyFeatsStatus._base.init(self)
    
    self.icon_size = icon_size or 40
    self.icons = self:AddChild( Widget() )

    self:Refresh()
end

function DailyFeatsStatus:Refresh()

    if TheGame:GetGameState() and TheGame:GetGameState():GetGameType() == GAME_TYPE.DAILY then

        -- Remove old widgets
        self.icons:DestroyAllChildren()

        -- Get challenge feats and add new ones
        local feats = TheGame:GetGameState():GetPlayerAgent().graft_owner:GetGrafts( GRAFT_TYPE.FEAT )
        for k, feat in ipairs(feats) do
            self.icons:AddChild( Widget.DailyFeatStatusIcon( feat, self.icon_size ) )
        end

        -- Layout icons
        self.icons:StackChildrenRow( 0 )

        self:Show()
    else
        self:Hide()
    end

    return self
end

------------------------------------------------------------------------------------------
-- A single feat icon

local DailyFeatStatusIcon = class("Widget.DailyFeatStatusIcon", Widget)

function DailyFeatStatusIcon:init( feat, icon_size )
    DailyFeatStatusIcon._base.init(self)

    self.feat = feat
    self.padding = 6
    self.widget_size = icon_size or 60
    self.icon_size = self.widget_size - self.padding/2

    -- self.bg_border = 16
    -- self.bg_border_scale = 0.5
    -- self.bg = self:AddChild( Widget.Panel( engine.asset.Texture ("UI/mainoverlay_icons_bg.tex") ) )
    --     :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
    --     :SetNineSliceBorderScale( self.bg_border_scale )
    --     :SetTintColour( UICOLOURS.MUTATOR )
    --     :Bloom( 0.1 )
    --     :Hide()

    self.glow = self:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( UICOLOURS.MUTATOR )
        :SetSize( self.widget_size, self.widget_size )
        :Bloom( 0.1 )

    self.icon = self:AddChild( Widget.Image() )
        :SetSize( self.icon_size, self.icon_size )
        :Bloom( 0.1 )

    self:ShowToolTipOnFocus( true )
    self:SetToolTipLayoutFn( function( w, tooltip_widget ) tooltip_widget:LayoutBounds( "center", "below", w ):Offset( 0, -SPACING.M1 ) end )
    self:SetToolTipFn( function()
        if self.feat.userdata.status == FEAT_STATUS.COMPLETE then
            return loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.FEAT_TOOLTIP_SUCCESS", self.feat:GetName(), self.feat:GetDesc(), self.feat_def.points )

        elseif self.feat.userdata.status == FEAT_STATUS.INCOMPLETE then
            local tt = loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.FEAT_TOOLTIP_INCOMPLETE", self.feat:GetName(), self.feat:GetDesc(), self.feat_def.points)
            local txt = self.feat:GetProgressText()
            if txt then
                tt = tt .. "\n\n" .. loc.format( LOC "UI.DAILY_CHALLENGE_SCREEN.FEAT_TOOLTIP_PROGRESS", txt )
            end
            return tt

        elseif self.feat.userdata.status == FEAT_STATUS.FAILED then
            return loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.FEAT_TOOLTIP_FAIL", self.feat:GetName(), self.feat:GetDesc(), self.feat_def.points )
        else
            return nil
        end
    end )

    self:Refresh( feat )
end

function DailyFeatStatusIcon:Refresh( feat )
    self.feat = feat
    self.feat_def = feat:GetDef()
    self.icon:SetTexture( self.feat_def.img )
    if self.feat.userdata.status == FEAT_STATUS.COMPLETE then
        self.icon:SetSaturation( 1 )
        -- self.bg:SetTintColour( UICOLOURS.MUTATOR )
        self.glow:SetTintColour( UICOLOURS.MUTATOR )
    elseif self.feat.userdata.status == FEAT_STATUS.INCOMPLETE then
        self.icon:SetSaturation( 0 )
        -- self.bg:SetTintColour( UICOLOURS.PENALTY )
        self.glow:SetTintColour( UICOLOURS.PENALTY )
    elseif self.feat.userdata.status == FEAT_STATUS.FAILED then
        self.icon:SetSaturation( 1 )
        self.icon:SetTintColour(UICOLOURS.PENALTY)
        -- self.bg:SetTintColour( UICOLOURS.PENALTY )
        self.glow:SetTintColour( UICOLOURS.PENALTY )
        self:SetToolTip( loc.format( LOC"UI.DAILY_CHALLENGE_SCREEN.FEAT_TOOLTIP_FAIL", self.feat:GetName(), self.feat:GetDesc(), self.feat_def.points) )
    end
    -- local nineslice_w, nineslice_h = self.widget_size*1.1 + self.bg_border*2*self.bg_border_scale, self.widget_size + self.bg_border*2*self.bg_border_scale
    -- self.bg:SetSize( nineslice_w, nineslice_h )
    self.glow:SetSize( self.widget_size, self.widget_size )
        :SetTintAlpha( 0.4 )
    self.icon:LayoutBounds( "center", "center", self.glow )
        -- :LayoutBounds( "center", "center", self.bg )
        -- :Offset( 0, 2 )
    return self
end