require "ui/widgets/progressbar"
require "ui/widgets/tooltipflourish"

local negotiation_defs = require "negotiation/negotiation_defs"
local EVENT = negotiation_defs.EVENT
local battle_defs = require "battle/battle_defs"
local BATTLE_EVENT = battle_defs.EVENT

local FlourishProgressWidget = class( "Widget.FlourishProgressWidget", Widget )

function FlourishProgressWidget:init( fn )
    FlourishProgressWidget._base.init( self )
    
    -- Save data
    self.min_flourish = 0
    self.max_flourish = 200
    self.bar_w_min = 80 -- edge_width when the flourish value is at 0%
    self.bar_w_max = 80 -- Width when the flourish value is at 100%
    self.bar_w = 80 -- Default width
    self.bar_h = 16

    -- Activate button
    self.activate_button = self:AddChild( Widget.FlourishProgressWidgetButton( 140 ) )
        :SetOnClickFn( function() self:ActivateClicked() end )
        :SetIcon( engine.asset.Texture("UI/ic_flourish.tex") )
        :SetupFocusGlow()

    self:SetValue( 0 )
end

-- If this button is supposed to show the full-screen effects on the button when at 100%
function FlourishProgressWidget:SetupFlourishEffects()
    self.flourish_fx = true
    self.activate_button:SetupFlourishEffects()
    return self
end

function FlourishProgressWidget:SetBattle( agent, battle )
	self.battle = battle
	if self.battle then
        self.agent = agent
        local graft_owner = self.agent and self.agent:GetAspect( "graft_owner" )
        self.flourish_graft = graft_owner and graft_owner:GetGraft( "flourish_tracker" )
        if not self.flourish_graft then
            self:Disable()
            return self
        end
		-- Get initial flourish value
		local graft = self.flourish_graft.userdata
		self.max_flourish = graft.battle_max
		self:AnimValue( graft.battle_flourish )
		-- Start listening for changes
		self.battle:ListenForEvent( BATTLE_EVENT.FLOURISH_CHANGED, self, self.OnBattleEvent )

        self:SetToolTipClass( Widget.TooltipFlourish )
        self:SetToolTip( self )
	end
	return self
end

function FlourishProgressWidget:OnBattleEvent( eventname, graft, flourish_level, prev_flourish )
    self:AnimValue( flourish_level )
end

function FlourishProgressWidget:IsBattle()
    return self.battle ~= nil
end

function FlourishProgressWidget:SetNegotiation( agent, negotiation )
	self.negotiation = negotiation
	if self.negotiation then
        self.agent = agent
        local graft_owner = self.agent and self.agent:GetAspect( "graft_owner" )
        self.flourish_graft = graft_owner and graft_owner:GetGraft( "flourish_tracker" )
        if not self.flourish_graft then
            self:Disable()
            return self
        end
		-- Get initial flourish value
		local graft = self.flourish_graft.userdata
		self.max_flourish = graft.negotiation_max
		self:AnimValue( graft.negotiation_flourish )
		-- Start listening for changes
		self.negotiation:ListenForEvent( EVENT.FLOURISH_CHANGED, self, self.OnNegotiationEvent )

        self:SetToolTipClass( Widget.TooltipFlourish )
        self:SetToolTip( self )
	end
	return self
end

function FlourishProgressWidget:OnNegotiationEvent( eventname, graft, flourish_level, prev_flourish )
	self:AnimValue( flourish_level )
end

function FlourishProgressWidget:IsNegotiation()
    return self.negotiation ~= nil
end

function FlourishProgressWidget:SetMax( max )
    self.max_flourish = max
    return self
end

function FlourishProgressWidget:GetAgent()
    return self.agent
end

function FlourishProgressWidget:OnRemoved()
    if self.negotiation then
        self.negotiation:RemoveListener( self )
    end
    if self.battle then
        self.battle:RemoveListener( self )
    end
    
    if self.loop then
        self.loop:Stop()
        self.loop:Release()
        self.loop = nil
    end

end

function FlourishProgressWidget:Layout()
    
    if self.target_value == self.max_flourish and not self.activate_button:IsGlowing() then
    	-- Fade in button
		self.activate_button:ScaleTo( 1.4, 1, 0.3, easing.outQuad, function() self:Layout() end )
            :ShowFocusGlow()
        if self.flourish_fx then self.activate_button:ShowFlourishFx() end

        self:PlayPositionalEvent("event:/ui/main/gen/flourishbar_full")

    elseif self.target_value ~= self.max_flourish and self.activate_button:IsGlowing() then
    	-- Fade out button
        self.activate_button:HideFocusGlow()
        if self.flourish_fx then self.activate_button:HideFlourishFx() end
    end

    if self.on_layout_fn then self.on_layout_fn() end
end

function FlourishProgressWidget:OnLayout( fn )
    self.on_layout_fn = fn
end

-- From 0 to 100
-- Instantly sets the bar to this value
function FlourishProgressWidget:SetValue( value )
	self.current_value = value or self.current_value or 0
	self.target_value = self.target_value or self.current_value
	self.activate_button:SetText( loc.format(LOC"UI.FLOURISH_SCREEN.BAR_PERCENT", self.target_value / self.max_flourish) )
        :SetProgress( self.current_value / self.max_flourish )
	self:Layout()
	return self
end

-- Animates from whatever value we're at to the new one
function FlourishProgressWidget:AnimValue( value )
	self.current_value = self.current_value or 0
    self.target_value = clamp( value, 0, 200 )
    self.activate_button:SetText( loc.format(LOC"UI.FLOURISH_SCREEN.BAR_PERCENT", self.target_value / self.max_flourish) )
    if not self.animating then
        self.animating = true
        self:StartUpdating()
    end
end

function FlourishProgressWidget:GetValues()
    return self.target_value or 0, self.max_flourish or 0
end

function FlourishProgressWidget:IsFull()
    return self.target_value >= self.max_flourish
end

function FlourishProgressWidget:OnUpdate( dt )
    FlourishProgressWidget._base.OnUpdate( self, dt )

    local delta = self.target_value - self.current_value
    if delta ~= 0 then
        local dp
        if math.abs( delta ) < 100 / 60 then
            dp = delta
        else
            dp = delta / 30
        end
        self:SetValue( self.current_value + dp )
    else
        self:SetValue( self.current_value )
    end

    if delta == 0 and self.animating then
        self.animating = false
        self:StopUpdating()
    end
end

function FlourishProgressWidget:SetOnClickFn( fn )
    self.activate_button:SetOnClickFn( fn )
    self.activate_button.hover_check = false
    return self
end

function FlourishProgressWidget:ActivateClicked()
    if self.battle then
        local screen = self:GetScreen()
        if screen then screen:ActivateFlourish() end
    elseif self.negotiation then
        local screen = self:GetScreen()
        if screen then screen:ActivateFlourish() end
    end
end

function FlourishProgressWidget:GetGraft()
    return self.flourish_graft
end

function FlourishProgressWidget:ResetBattleFlourish()
    self:AnimValue( 0 )
    self.flourish_graft.userdata.battle_flourish = 0
end

function FlourishProgressWidget:ResetNegotiationFlourish()
    self:AnimValue( 0 )
    self.flourish_graft.userdata.negotiation_flourish = 0
end

function FlourishProgressWidget:PrepareAnimation()
    self.activate_button:PrepareAnimation()
    self:Show():SetTintAlpha(0)
    return self
end

function FlourishProgressWidget:AnimateIn()
    self.activate_button:AnimateIn()
    self:Show():AlphaTo( 1, 0.1, easing.inOutQuad )
    return self
end

function FlourishProgressWidget:AnimateOut()
    self.activate_button:AnimateOut()
    self:AlphaTo( 0, 0.1, easing.inOutQuad, function() self:Hide() end )
    return self
end

-------------------------------------------------------------------------
-- A button to trigger a flourish

local FlourishProgressWidgetButton = class( "Widget.FlourishProgressWidgetButton", Widget.LargeRoundButton )

function FlourishProgressWidgetButton:init( size )
    FlourishProgressWidgetButton._base.init( self, size )

    self.title:SetFontSize( math.ceil( size * 0.18 ))

    self:ShowToolTipOnFocus( true )

    -- Replace overlay texture
    self.icon_overlay:SetTexture( engine.asset.Texture( "UI/flourishbutton_overlay.tex") )

    -- Add radial progress
    self.radial_size = size*0.498
    self.radial_bg = self.contents:AddChild( Widget.Image( engine.asset.Texture( "UI/flourish_radial_bg.tex") ) )
        :SetSize( self.radial_size, self.radial_size )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.DARK_BODYTEXT )
        :Bloom( 0 )
    self.radial_progress = self.contents:AddChild( Widget.RadialProgress( self.radial_size, self.radial_size, engine.asset.Texture( "UI/flourish_radial_progress.tex") ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.FLOURISH )
        :Bloom( 0.3 )
end

function FlourishProgressWidgetButton:SetupFocusGlow()
    FlourishProgressWidgetButton._base.SetupFocusGlow( self )
    self.innerglow1:SetTexture( engine.asset.Texture("UI/flourish_button_innerglow.tex" ) )
        :SetTintColour(UICOLOURS.WHITE)
        :SetTintAlpha(0)
    self.innerglow2:SetTexture( engine.asset.Texture("UI/flourish_button_innerglow.tex" ) )
        :SetTintColour(UICOLOURS.WHITE)
        :SetTintAlpha(0)
    return self
end

-- If this button is supposed to show the full-screen effects on the button when at 100%
function FlourishProgressWidgetButton:SetupFlourishEffects()
    self.flourish_fx = true
    
    -- Setup the circular animation around the button
    local animname = "anim/UI_negotiation_round_fx.zip"
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetScale( 0.4 )
        :SetTintColour( UICOLOURS.FLOURISH )
        :SetBloom( 0.2 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.flourish_fx_round = self.contents:AddChild( anim )
        :SendToBack()
        :Hide()

    -- Setup the energy pulse animation
    local animname = "anim/card_flourish_fx.zip"
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetScale( 0.9 )
        -- :SetTintColour( UICOLOURS.FLOURISH )
        :SetBloom( 0.2 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.flourish_fx_energy_container = self.contents:AddChild( Widget() )
        :SendToBack()
        :Hide()
    self.flourish_fx_energy = self.flourish_fx_energy_container:AddChild( anim )

    return self
end

function FlourishProgressWidgetButton:ShowFlourishFx()
    if self.flourish_fx then

        -- Start the circular animation
        self.flourish_fx_round:Play( "map_difficulty_highlight", true ):Show()
            :RotateIndefinitely( 0.01 )

        -- Start the energy animation
        self.flourish_fx_energy:Play( "map_difficulty_highlight", true )
            :Play( "flourish_in", false )
            :Push( "flourish_loop", true )
            :RotateIndefinitely( 0.014 )
        self.flourish_fx_energy_container:Show()
            :SetScale( 1, 1 )
            :AlphaTo( 1, 0.8, easing.outQuad, function() self.flourish_fx_energy_container:ScaleTo( 1, 0.15, 0.6, easing.inOutQuad ) end)

    end
    return self
end
function FlourishProgressWidgetButton:HideFlourishFx()
    if self.flourish_fx then

        -- Stop the circular animation
        self.flourish_fx_round:Hide():StopSpin()

        -- Stop the energy animation
        self.flourish_fx_energy:Push( "flourish_out", false )
            :AlphaTo( 1, 1, easing.outQuad, function() self.flourish_fx_energy_container:Hide() end)

    end
    return self
end

function FlourishProgressWidgetButton:UpdateImage()
    FlourishProgressWidgetButton._base.UpdateImage( self )
    
    if not self.enabled then
        self.icon:SetSaturation( 1 )
        self.title:SetSaturation( 1 )
        self.subtitle:SetSaturation( 1 )
        self.icon_bg:SetTintColour( 0x061F22ff )
    end
    return self
end

function FlourishProgressWidgetButton:SetProgress( progress )
    self.radial_progress:SetProgress( progress )
    return self
end

function FlourishProgressWidgetButton:OnGainHover()
    FlourishProgressWidgetButton._base.OnGainHover(self)
    self.loop = self.loop or AUDIO:CreateEventInstance("event:/ui/main/gen/flourishbar_mouseover_LP")
    self.loop:Start()
end

function FlourishProgressWidgetButton:OnLoseHover()
    FlourishProgressWidgetButton._base.OnLoseHover(self)
    if self.loop then
        self.loop:Stop()
    end
end
