local CONTENT_LEFT_MARGIN = 15 -- Extra size added on the left of scroll boxes, so that the contents can scale up on hover without clipping
require "ui/widgets/compendiumwidget"
------------------------------------------------------------------------------------------
-- Shows the grafts panel

local assets = {
    gradient = engine.asset.Texture("UI/compendium_graft_gradient.tex"),
    bg = engine.asset.Texture("UI/compendium_graft_bg.tex"),
    white = engine.asset.Texture( "images/white.tex" ),
    glow = engine.asset.Texture( "UI/grifts_progress_glow.tex" ),
    arrrow = engine.asset.Texture( "UI/grifts_progress_arrow.tex" ),
    complete = engine.asset.Texture("UI/compendium_grift_complete.tex"),
}


local GriftsCompendiumWidget = class( "Widget.GriftsCompendiumWidget", Widget.CompendiumWidget )

GriftsCompendiumWidget.CONTROL_MAP = 
{
}

------------------------------------------------------------------------------------------
-- Shows a single grift

local GRIFT_COLUMNS = 2

function GriftsCompendiumWidget:init( width, height, sidebar_width )
    GriftsCompendiumWidget._base.init( self, width, height, sidebar_width )
    self.grifts_root = self.scroll_root:AddScrollChild( Widget() )
    self.sidebar:Hide()

    local scroll_w = self.content_w+sidebar_width+CONTENT_LEFT_MARGIN+6
    self.scroll_root:SetSize( scroll_w, self.widget_h-2 )
    self.scroll_inner_w = scroll_w - 320
end

function GriftsCompendiumWidget:CanNav()
    return true
end

function GriftsCompendiumWidget:HasSidebar()
    return false
end

function GriftsCompendiumWidget:SortOrderSelected( option )
    if option.fn then self.sorting_fn = option.fn end
    self:LayoutContent()
end

function GriftsCompendiumWidget:RefreshContent()
    GriftsCompendiumWidget._base.RefreshContent()

    if self.grifts_root:GetNumChildren() > 0 then
        return
    end

    -- Calculate the size of each widget
    local columns = GRIFT_COLUMNS
    local spacing = 20
    local graft_w = (self.scroll_inner_w-(spacing*(columns-1)))/columns
    local graft_h = 140

    local reviewed_grift_ids = {} -- The ids of the grifts we've checked
    local function GetNextGrift( current_grift, list, grifts )
        -- If there isn't a list, create one
        list = list or {}
        -- Add the current grift
        table.insert( list, current_grift )
        -- Add this to the reviewed grifts list, so we don't go down this chain again
        reviewed_grift_ids[ current_grift.id ] = true
        -- Check if there's another grift
        if current_grift.next_id then
            -- If so, call this function on it
            return GetNextGrift( grifts[current_grift.next_id], list, grifts )
        end
        return list
    end

    -- Organize all the grifts
    local all_grifts = Content.GetAllGrifts()
    -- Save them by id
    local grifts_id = {}
    for i, def in pairs( all_grifts ) do
        grifts_id[ def.id ] = def
    end
    -- Group them by id (win_battle_1, win_battle_2, win_battle_3, etc)
    local grouped_grifts = {}
    for i, grift in pairs( all_grifts ) do
        if not reviewed_grift_ids[grift.id] then
            grouped_grifts[grift.id] = GetNextGrift( grift, nil, grifts_id )
        end
    end

    -- Create widgets for them
    for id, grifts_group in pairs( grouped_grifts ) do
        self.grifts_root:AddChild( Widget.GriftProgress( graft_w, graft_h, grifts_group ) )
    end

    -- Sort them
    table.stable_sort( self.grifts_root.children, function( w1, w2 ) return w1.max_value > w2.max_value end )

    self:LayoutContent()
    return self
end

function GriftsCompendiumWidget:LayoutContent()
    self.grifts_root:LayoutChildrenInGrid( GRIFT_COLUMNS, 12 ):Offset( 12, 0 )
        :LayoutBounds( "center", "top", RES_X/2, 0 )

    -- Layout things!
    self.scroll_root:SetVirtualMargin( 20 )
    self.scroll_root:RefreshView()
    self.scroll_root:LayoutBounds( "left", "top", self.sidebar )
        :Offset( -CONTENT_LEFT_MARGIN + 12, -1 )

    return self
end

function GriftsCompendiumWidget:AddSidebarFilters()
end

function GriftsCompendiumWidget:UpdateCount()
    local unlocked, total = 0, 0
    for id, def in pairs(Content.GetAllGrifts()) do
        total = total + 1
        if def:IsComplete() then
            unlocked = unlocked + 1
        end
    end
    self.grifts_completed = unlocked
    self.grifts_total = total
end

-- Returns the number of unlocked grafts and the number of unlockable grafts, to be displayed on the nav bar
function GriftsCompendiumWidget:GetCount()
    if self.grifts_completed == nil then
        self:UpdateCount()
    end

    return self.grifts_completed, self.grifts_total
end

function GriftsCompendiumWidget:GetDefaultFocus()
    for k, v in ipairs( self.grifts_root.children ) do
        return v
    end
end


------------------------------------------------------------------------------------------
-- Shows a single grift

local GriftProgress = class( "Widget.GriftProgress", Widget.Clickable )

local OUTLINE_SIZE = 3

function GriftProgress:init( width, height, grift_defs )
    GriftProgress._base.init( self )

    self.widget_w = width or 240
    self.widget_h = height or 90
    self.grift_defs = grift_defs or {}

    self.bg_border = 30
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = self.widget_w + self.bg_border*2*self.bg_border_scale, self.widget_h + self.bg_border*2*self.bg_border_scale
    
    -- Gradient glow
    self.bg_gradient = self:AddChild( Widget.Image( assets.gradient ) )
        :SetSize( self.widget_w*0.75, self.widget_h )
        :SetHiddenBoundingBox( true )

    -- Background panel
    self.bg = self:AddChild( Widget.Panel( assets.bg ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :Bloom( 0.05 )

    -- Grift title
    self.title = self:AddChild( Widget.Label("title", FONT_SIZE.ITEM_SUBTITLE ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :LeftAlign():TopAlign():SetWordWrap( true )
        :Bloom(0.1)

    self.progress_w = self.widget_w - SPACING.M1*2 -- Width of the total widget, including outline
    self.progress_w_inner = self.progress_w - OUTLINE_SIZE*2 -- Width of the inner progress bar
    self.progress_h = 15
    -- Contains the progress bar widget
    self.progress_container = self:AddChild( Widget() )
    self.progress_outline = self.progress_container:AddChild( Widget() ) -- Dark outline
        :SetTintColour( UICOLOURS.DARK_BODYTEXT )
    self.progress_outline_rect = self.progress_outline:AddChild( Widget.SolidBox( self.progress_w, self.progress_h, UICOLOURS.WHITE ) )
    self.progress_outline_arrows = self.progress_outline:AddChild( Widget() ) -- Contains the arrow outlines
    self.progress_bar = self.progress_container:AddChild( Widget() ) -- Colored inner line
        :SetStencilContext( true )

    -- The masking parts of the progress bar
    self.progress_mask_rect = self.progress_bar:AddChild( Widget.Image( assets.white ) )
        :SetSize( self.progress_w_inner, self.progress_h-OUTLINE_SIZE*2 )
        :SetMask(true)
    self.progress_bar_arrows = self.progress_mask_rect:AddChild( Widget() ) -- Contains the arrows
    
    -- The visible parts of the progress bar, masked by the stuff above
    self.progress_shown_rect = self.progress_bar:AddChild( Widget.Image( assets.white ) )
        :SetSize( self.progress_w_inner, 40 )
        :SetTintColour( UICOLOURS.GRIFTS )
        :SetStencilTest( true )
        :Bloom( 0.15 )
    self.progress_shown_glow = self.progress_bar:AddChild( Widget.Image( assets.glow ) )
        :SetSize( 250, 40 )
        :SetStencilTest( true )
        :SetHiddenBoundingBox( true )
        :Bloom( 0.2 )

    self.progress_labels_container = self.progress_container:AddChild( Widget() ) -- The progress labels below the bar
    self.progress_labels = {}

    -- Go throught the levels of this grift
    self.id = self.grift_defs[1].id
    self.current_value = 0 -- Your current progress overall on this grift
    self.current_stage = 0 -- The current stage you're in on this grift
    self.current_stage_progress = 0 -- The progress within the current stage
    self.current_title = "" -- The title for your current stage on this grift
    self.current_reward = 0 -- The reward for your current stage on this grift
    self.max_value = 0 -- The max value on this grift
    local previous_max = 0
    for k, grift in ipairs( self.grift_defs ) do
        -- Get progress
        local progress, max_progress = grift:GetProgress()
        -- Update progress
        self.current_value = progress
        -- Update max
        self.max_value = math.max( self.max_value, max_progress )
        -- Check if you're in this stage
        if self.current_value >= previous_max and self.current_value <= self.max_value then
            self.current_stage = k
            self.current_stage_progress = (progress-previous_max)/(max_progress-previous_max)
            self.current_title = loc.format( grift:GetLocalizedName(), grift.count )
            self.current_reward = grift.perk_points
        end
        -- Update last max
        previous_max = self.max_value
        -- Add a progress label for this stage
        local label = self.progress_labels_container:AddChild( Widget.Label( "body", FONT_SIZE.ITEM_DESC ) )
        table.insert( self.progress_labels, label )
        -- Set its value
        if k == #self.grift_defs then
            -- Last or only label
            label:SetText( loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_PROGRESS_TOTAL", self.current_value, self.max_value ) )
                -- :SetTintColour( UICOLOURS.GRIFTS )
                :Bloom( 0.15 )
        else
            -- Progress label
            label:SetText( loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_PROGRESS", max_progress ) )
                -- :SetTintColour( UICOLOURS.GRIFTS )
                -- :SetTintAlpha( 0.9 )
                :Bloom( 0.1 )
        end
        -- Add arrow outline
        local arrow_size = 14
        label.arrow_outline = self.progress_outline_arrows:AddChild( Widget.Image( assets.arrow ) )
            :SetSize( arrow_size, arrow_size )
        -- Add arrow
        label.arrow_inner = self.progress_bar_arrows:AddChild( Widget.Image( assets.arrow ) )
            :SetSize( arrow_size, arrow_size )
            :SetMask(true)
    end

    -- Calculate the length of the progress bar
    local progress = (self.current_stage-1)/#self.grift_defs + self.current_stage_progress/#self.grift_defs
    self.progress_shown_rect:SetSize( self.progress_w_inner*progress )

    -- Reward value
    local reward_txt = loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_REWARD", self.current_reward or 0 )
    self.reward_label = self:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, reward_txt ))
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom( 0.15 )

    -- Check if the grift is complete
    self.complete = self.grift_defs[self.current_stage]:IsComplete()

    -- Change display accordingly
    if self.complete then
        self.reward_label:Hide()
        self.bg:SetTexture( assets.complete )
            :SetTintColour( UICOLOURS.GRIFTS )
            :Bloom( 0.2 )
        self.bg_gradient:SetTintColour( UICOLOURS.GRIFTS )
            :SetTintAlpha( 0.3 )
        self.title:SetText( loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_TITLE_COMPLETE", self.current_title ) )
            :Bloom( 0.02 )
    else
        self.bg:SetTexture( assets.bg )
        self.bg_gradient:Hide()
        local w, h = self.reward_label:GetRenderSize()
        self.title:SetSize( self.widget_w - self.bg_border * 2 - w, self.widget_h )
        self.title:SetText( self.current_title )
    end

    -- Set tooltip
    self:SetToolTipClass( Widget.TooltipGrift )
    self:SetToolTip( self )
    self:ShowToolTipOnFocus( true )

    self:Layout()
end

function GriftProgress:IsComplete()
    return self.complete
end

function GriftProgress:Layout()
    -- Layout text
    self.title:LayoutBounds( "left", "top", self.bg ):Offset( SPACING.M1, -8 )
    self.reward_label:LayoutBounds( "right", "top", self.bg ):Offset( -SPACING.M1+3, -8 )

    -- Layout bar
    local sector_size = self.progress_w_inner/#self.grift_defs -- The length of each sector on the bar
    -- Position the progress labels
    for k, label in ipairs( self.progress_labels ) do
        label.arrow_inner:LayoutBounds( "right", "below", self.progress_outline_rect )
            :Offset( -math.floor( sector_size*(#self.progress_labels - k) ) - OUTLINE_SIZE, 12 )
        label.arrow_outline:LayoutBounds( "right", "bottom", label.arrow_inner )
            :Offset( OUTLINE_SIZE*1, -OUTLINE_SIZE*2.5 )
        label:LayoutBounds( "right", "below", label.arrow_outline )
            :Offset( -10, 10 )
    end
    -- Position the progress rect
    self.progress_shown_rect:LayoutBounds( "left", "center", self.progress_outline_rect ):Offset( OUTLINE_SIZE, 0 )
    self.progress_shown_glow:LayoutBounds( "right", nil, self.progress_shown_rect )

    self.progress_container:LayoutBounds( "center", nil, self.bg )
        :LayoutBounds( nil, "bottom", self.bg )
        :Offset( 0, 10 )

    self.bg_gradient:LayoutBounds( "left", nil, self.bg ):Offset( 1, 0 )

    return self
end

function GriftProgress:UpdateImage()
    if ( self.hover or self.focus ) then
        self.title:Bloom( self.complete and 0.08 or 0.15 )
        self.bg:Bloom( self.complete and 0.25 or 0.1 )

        self:ScaleTo( nil, 1.02, 0.2, easing.outQuad )
    else
        self.title:Bloom( self.complete and 0.02 or 0.1 )
        self.bg:Bloom( self.complete and 0.2 or 0.05 )

        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function GriftProgress:OnGainFocus()
    GriftProgress._base.OnGainFocus(self)
    self:UpdateImage()
end

function GriftProgress:OnLoseFocus()
    GriftProgress._base.OnLoseFocus(self)
    self:UpdateImage()
end

function GriftProgress:OnGainHover()
    GriftProgress._base.OnGainHover(self)
    self:UpdateImage()
end

function GriftProgress:OnLoseHover()
    GriftProgress._base.OnLoseHover(self)
    self:UpdateImage()
end

---------------------------------------------------------------------------------------------

local TooltipGrift = class( "Widget.TooltipGrift", Widget )

function TooltipGrift:init()
    TooltipGrift._base.init( self )

    self.bg = self:AddChild( Widget.Panel( global_images.tooltipbg ) )
    self.bg:SetNineSliceBorderScale( 0.5 )

    self.content = self:AddChild( Widget() )

    -- Title and subtitle
    self.title = self.content:AddChild( Widget.Label( "title", 30 ) )
        :SetAutoSize( DEFAULT_TT_WIDTH  )
        :SetWordWrap( true )
        :LeftAlign()
        :SetHoverLinkGlyphColour( UICOLOURS.QUEST_LIGHT )
        :SetGlyphColour( UICOLOURS.WHITE )
        :Bloom( 0.05 )
    self.desc = self.content:AddChild( Widget.Label("tooltip", 22) )
        :SetAutoSize( DEFAULT_TT_WIDTH )
        :SetWordWrap( true )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :SetHoverLinkGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom(0.1)
end

function TooltipGrift:DoLayout( grift_widget )
    
    self.title:SetText( grift_widget.title:GetText() )
    local tt = ""
    for k, grift in ipairs( grift_widget.grift_defs ) do
        local progress, max_progress = grift:GetProgress()
        if max_progress == 1 and #grift_widget.grift_defs == 1 then
            -- This is a single objective goal (1/1)
            tt = tt .. loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_TT_SINGLEROW", grift.perk_points )
        else
            local bold, close_bold = "", ""
            if progress >= max_progress then
                bold, close_bold = "<b>", "</b>"
            end
            tt = tt .. loc.format( LOC"UI.CARDCOMPENDIUM.GRIFT_TT_ROW", max_progress, grift.perk_points, progress, bold, close_bold ) .. "\n"
        end
    end
    self.desc:SetText( tt )

    -- Layout
    self.desc:LayoutBounds( "left", "below", self.title )

    local w, h = self.content:GetSize()
    self.bg:SetSize( w+SPACING.M1*2, h+SPACING.M1*2 )
        :LayoutBounds( "center", "center", self.content )
end
