require "ui/widget"
require "ui/widgets/panelbutton"

local IconTouchSlideButton = class( "Widget.IconTouchSlideButton", Widget.IconButton )

-- This button behaves as a regular icon button when using mouse/keyboard, 
-- but as a slide-to-trigger button on touch devices
function IconTouchSlideButton:init( text, fn, w, h, text_size, img, img_hover )
	IconTouchSlideButton._base.init( self, text, fn, w, h, text_size, img, img_hover )

	self.touch_mode = false
    self.was_click_down = false -- If the click was down on the button last Update

	-- Add slide visual elements
	self.slider_pointer_h = self.height*1.05
	self.slider_pointer_w = self.height*1.8
	self.slider_pointer = self:AddChild( Widget.Image( engine.asset.Texture( "UI/btn_pointer.tex" ) ) ) -- A draggable pointer
		:SetSize( self.slider_pointer_w, self.slider_pointer_h )
        :SetTintColour( UICOLOURS.FOCUS )
        :SetTintAlpha( 0 )
        :Bloom( 0.3 )
        :SetHiddenBoundingBox( true )
        :SendBehind( self.label )
    self.slider_bg = self:AddChild( Widget.Image( engine.asset.Texture( "UI/btn_pointer_frame.tex") ) ) -- A background for the pointer
    	:SetSize( self.width, self.slider_pointer_h )
        :SetTintColour( UICOLOURS.FOCUS )
        :SetTintAlpha( 0 )
        :Bloom( 0.3 )
        :SetHiddenBoundingBox( true )
        :SendBehind( self.label )

    -- Set slide boundaries
    self.slide_min_x = self.slide_min_x or 0 -- This is the initial position of the pointer on the left. Calculated after we know our width
    self.slide_max_x = self.slide_max_x or 0 -- This is the final position of the pointer on the right. Calculated after we know our width
    self.slide_initial_click_area = -0.2 -- (multiplied by self.w) Initial clicks must be lower than this distance-from-the-center (0) for a drag to start
    self.slide_jump_animation_distance = 70 -- How much the pointer should jump on invalid clicks, to indicate where to start dragging from

    self:RefreshSliderSize()
end

function IconTouchSlideButton:SetSize( w, h )
    IconTouchSlideButton._base.SetSize( self, w, h )
    
    self:RefreshSliderSize()
    
    return self
end

function IconTouchSlideButton:RefreshSliderSize()
	if self.slider_pointer then
		self.slider_pointer_h = self.height*1.05
		self.slider_pointer_w = self.height*1.8
		self.slider_pointer:SetSize( self.slider_pointer_w, self.slider_pointer_h )
	    self.slide_min_x = -self.w/2 + self.icon_size/2 + self.slider_pointer_w/2
	    self.slide_max_x = self.w/2 - self.slider_pointer_w/2 - 4

	    self.slider_bg:SetSize( self.width-20, self.slider_pointer_h )
	end
	return self
end

function IconTouchSlideButton:RefreshBloom()
	self.slider_pointer:Bloom(0.3)
	self.slider_bg:Bloom(0.3)
	return self
end

function IconTouchSlideButton:OnAdded()
    -- Refresh control mode
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
end

function IconTouchSlideButton:OnControlModeChange( cm, device_id )
    IconTouchSlideButton._base.OnControlModeChange( self, cm, device_id )
    
    if cm == CONTROL_MODE.TOUCH then
    	self.touch_mode = true
    else
    	self.touch_mode = false
    	self.slider_pointer:SetTintAlpha(0)
    	self.slider_bg:SetTintAlpha(0)
    end
end

function IconTouchSlideButton:HandleControlDown( control, device )

	-- If we're in touch mode, trigger the OnUpdate so we can check for dragging
	if self.touch_mode and self.enabled then self:StartUpdating() end

	return IconTouchSlideButton._base.HandleControlDown( self, control, device )
end

function IconTouchSlideButton:OnUpdate(dt)
	IconTouchSlideButton._base.OnUpdate(self, dt)

	if not self.touch_mode or not self.enabled then return end

	-- This is only for touch mode!

	if TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) then

        local current_x, current_y = TheGame:GetInput():GetMousePos() -- Get current mouse position
        local click_within_bounds = self:CheckHit( current_x, current_y ) -- Check if the mouse is within the button's bounding box
		current_x, current_y = self:TransformFromWorld( current_x, current_y ) -- Convert the mouse position to local coordinates

		if not self.was_click_down and click_within_bounds then
            -- We just clicked down!

            -- Is this click within the initial area?
            if current_x < self.w*self.slide_initial_click_area then
	            -- Let's start dragging
	            self.slider_dragging = true
	            -- Fade in the pointer
	            self.slider_pointer:AlphaTo( 1, 0.1 )
	        else
	        	-- The player tapped outside the initial area. Let's animate the arrow to clue them in
	        	self.slider_pointer:SetPos( self.slide_min_x, nil )
	        		:AlphaTo( 1, 0.1, easing.inOutQuad )
	        		:MoveTo( self.slide_min_x+self.slide_jump_animation_distance, nil, 0.1, easing.inOutQuad, function()
	        			self.slider_pointer:MoveTo( self.slide_min_x, nil, 0.3, easing.inOutQuad )
	        				:AlphaTo( 0, 0.3, easing.inOutQuad )
	        		end )
	            self.slider_dragging = false
	        end
        end

        if self.slider_dragging then
            -- We're dragging the pointer with touch input
            self.slider_pointer:SetPos( clamp( current_x, self.slide_min_x, self.slide_max_x ), nil ) -- Actually move the pointer, preventing it from going past the limits
            -- Make the slider background fade in as we get closer to triggering the click
            self.slider_bg:SetTintAlpha( math.max( 0, current_x/(self.w/2) ) )
            if current_x >= self.slide_max_x then
            	-- Trigger click!
            	self:FinishDragging()
            	IconTouchSlideButton._base.Click( self )
            	return self
            end
        end
    else
        -- Mouse up. No more dragging over here
        self:FinishDragging()
    end
    self.was_click_down = TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) and self:CheckHit( TheGame:GetInput():GetMousePos() )

end

function IconTouchSlideButton:FinishDragging()
	if self.slider_dragging then
    	self.slider_dragging = false

        -- Fade out the pointer
        self.slider_pointer:AlphaTo( 0, 0.3, easing.inOutQuad )
        	:MoveTo( math.max( self.slide_min_x, self.slider_pointer.x-self.slide_jump_animation_distance ), nil, 0.3, easing.inOutQuad )
        self.slider_bg:AlphaTo( 0, 0.3, easing.inOutQuad )

        -- No more need for updating
        self:StopUpdating()
    end
	return self
end

-- We just want Clicks to be triggered if we're not in touch_mode. 
-- When in touch mode, the click should be triggered by sliding the pointer across the button
function IconTouchSlideButton:Click()
	if self.touch_mode then return end
	return IconTouchSlideButton._base.Click( self )
end