require "ui/widget"
require "ui/widgets/panelbutton"

local IconTrackSlideButton = class( "Widget.IconTrackSlideButton", Widget.IconButton )

-- This button behaves as a regular icon button when using mouse/keyboard, 
-- but as a slide-to-trigger button on touch devices
function IconTrackSlideButton:init( text, fn, w, h, text_size, img, img_hover )
	IconTrackSlideButton._base.init( self, text, fn, w, h, text_size, img, img_hover )

	self.touch_mode = false
    self.was_click_down = false -- If the click was down on the button last Update

    -- Encapsulate all the button elements into a widget that we'll drag around
    self.button_contents = self:AddChild( Widget() )
    -- Reparent every child
    while #self.children > 1 do
        if self.children[1] ~= self.button_contents then
            self.children[1]:Reparent( self.button_contents )
        end
    end

	-- Add slide visual elements
	self.slider_pointer_h = self.height*1
	self.slider_pointer_w = self.height*3
	self.slider_pointer = self:AddChild( Widget.Image( engine.asset.Texture( "UI/btn_track_slide.tex" ) ) ) -- A visual target for the button
        :SetTintColour( UICOLOURS.SUBTITLE )
		:SetSize( self.slider_pointer_w, self.slider_pointer_h )
        :Bloom( 0.1 )
        :SendToBack()
        :Hide()
    self.slider_bg = self.button_contents:AddChild( Widget.Image( engine.asset.Texture( "UI/btn_pointer_frame.tex") ) ) -- A highlight for the pointer
        :SetSize( self.width, self.slider_pointer_h )
        :SetTintColour( UICOLOURS.FOCUS )
        :SetTintAlpha( 0 )
        :Bloom( 0.3 )
        :SetHiddenBoundingBox( true )
        :SendBehind( self.label )

    -- Set slide boundaries
    self.slide_min_x = self.slide_min_x or 0 -- This is the initial position of the button on the left. Calculated after we know our width
    self.slide_max_x = self.slide_max_x or 0 -- This is the final position of the button on the right. Calculated after we know our width
    self.slide_jump_animation_distance = 40 -- How much the button should jump on clicks to indicate where to start dragging from
    self.target_distance = self.slider_pointer_w + 10 -- How far the target should peek from the edge of the button. self.width+self.target_distance will be the actual visual width of the button

    self:RefreshSliderSize()
end

function IconTrackSlideButton:SetSize( w, h )
    IconTrackSlideButton._base.SetSize( self, w, h )
    
    self:RefreshSliderSize()
    
    return self
end

function IconTrackSlideButton:RefreshSliderSize()
	if self.slider_pointer then
        self.slider_pointer_h = self.height*1
        self.slider_pointer_w = self.height*3
		self.slider_pointer:SetSize( self.slider_pointer_w, self.slider_pointer_h )
        self.slider_bg:SetSize( self.width-20, self.slider_pointer_h )
        -- Move button to initial position
        self.button_contents:SetPos( 0, 0 )
	    self.slide_min_x = 0
	    self.slide_max_x = self.w/2 + self.target_distance
        self.slider_pointer:LayoutBounds( "right", nil, self.button_contents )
            :Offset( self.target_distance )
        -- Find the drag limit
        self.button_contents:LayoutBounds( "right", nil, self.slider_pointer )
        self.slide_max_x = self.button_contents:GetPos()
        self.button_contents:SetPos( 0, 0 )
	end
	return self
end

function IconTrackSlideButton:RefreshBloom()
	self.slider_pointer:Bloom( 0.1 )
    self.slider_bg:Bloom(0.3)
	return self
end

function IconTrackSlideButton:OnAdded()
    -- Refresh control mode
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
end

function IconTrackSlideButton:SetIcon( icon, colour )
    IconTrackSlideButton._base.SetIcon( self, icon, colour )
    self.slider_pointer:SetTintColour( colour or UICOLOURS.SUBTITLE )
    return self
end

function IconTrackSlideButton:OnControlModeChange( cm, device_id )
    IconTrackSlideButton._base.OnControlModeChange( self, cm, device_id )
    
    if cm == CONTROL_MODE.TOUCH then
    	self.touch_mode = true
        self:ScaleOnFocus( false )
        self.slider_pointer:Show()
    else
    	self.touch_mode = false
        self:ScaleOnFocus( true )
    	self.slider_pointer:Hide()
        self.slider_bg:SetTintAlpha(0)
    end
end

function IconTrackSlideButton:HandleControlDown( control, device )

	-- If we're in touch mode, trigger the OnUpdate so we can check for dragging
	if self.touch_mode and self.enabled then
        self.control_down_time = GetWallTime()*1000
        self:StartUpdating()
    end

	return IconTrackSlideButton._base.HandleControlDown( self, control, device )
end

function IconTrackSlideButton:OnUpdate(dt)
	IconTrackSlideButton._base.OnUpdate(self, dt)

	if not self.touch_mode or not self.enabled then return end

	-- This is only for touch mode!

	if TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) then

        local current_x, current_y = TheGame:GetInput():GetMousePos() -- Get current mouse position
        local click_within_bounds = self:CheckHit( current_x, current_y ) -- Check if the mouse is within the button's bounding box
        current_x, current_y = self:TransformFromWorld( current_x, current_y ) -- Convert the mouse position to local coordinates

		if not self.was_click_down and click_within_bounds then
            -- We just clicked down!
            -- Let's start dragging
            self.slider_dragging = true
        end

        if self.slider_dragging and self.initial_x then
            -- We're dragging the button with touch input
            -- Calculate how much the finger has moved
            local delta_x = current_x - self.initial_x
            -- Move that
            self.button_contents:SetPos( clamp( delta_x, self.slide_min_x, self.slide_max_x ), nil ) -- Actually move the button, preventing it from going past the limits
            if self.slide_max_x - delta_x < 6 then
                -- Make the slider background fade in if we're at the end
                self.slider_bg_fading_out = false
                if not self.slider_bg_fading_in then self.slider_bg:AlphaTo( 1, 0.1 ) self.slider_bg_fading_in = true end
                self.click_ready = true
            	return self
            else
                self.slider_bg_fading_in = false
                if not self.slider_bg_fading_out then self.slider_bg:AlphaTo( 0, 0.3 ) self.slider_bg_fading_out = true end
                self.click_ready = false
            end
        elseif self.slider_dragging then
            self.initial_x = current_x
        end

    else
        if self.click_ready then
            -- Trigger click!
            self:FinishDragging()
            IconTrackSlideButton._base.Click( self )
        elseif self.control_down_time
        and GetWallTime()*1000 - self.control_down_time < 500 then
            -- The player clicked on and off quickly, so let's show them the animation, indicating they should drag the button instead of clicking it
            self.button_contents:MoveTo( self.slide_min_x+self.slide_jump_animation_distance, nil, 0.1, easing.inOutQuad, function()
                self.button_contents:MoveTo( self.slide_min_x, nil, 0.3, easing.inOutQuad )
            end )
            self.slider_dragging = false
            self.control_down_time = nil
        end
        -- Mouse up. No more dragging over here
        self:FinishDragging()
    end
    self.was_click_down = TheGame:GetInput():IsControlDown(Controls.Digital.MENU_ACCEPT) and self:CheckHit( TheGame:GetInput():GetMousePos() )

end

function IconTrackSlideButton:FinishDragging()
	if self.slider_dragging then
    	self.slider_dragging = false

        -- Reset the initial click position
        self.initial_x = nil

        -- Fade out the pointer
        self.button_contents:MoveTo( self.slide_min_x, nil, 0.3, easing.inOutQuad )

        self.slider_bg:AlphaTo( 0, 0.3 )

        -- No more need for updating
        self:StopUpdating()
    end
	return self
end

-- We just want Clicks to be triggered if we're not in touch_mode. 
-- When in touch mode, the click should be triggered by sliding the pointer across the button
function IconTrackSlideButton:Click()
	if self.touch_mode then return end
	return IconTrackSlideButton._base.Click( self )
end

function IconTrackSlideButton:UpdateImage()
    IconTrackSlideButton._base.UpdateImage( self )

    if self.slider_pointer then
        if self.enabled then
            self.slider_pointer:SetSaturation( 1 )
        else
            self.slider_pointer:SetSaturation( 0.1 )
            self.slider_bg:SetTintAlpha( 0 )
        end
    end

    return self
end