require "ui/widget"
require "ui/widgets/tooltip_intent"

local negotiation_defs = require "negotiation/negotiation_defs"
local SLOT_SIZE = 70
local assets =
{
    damage = engine.asset.Texture("UI/intent_damage.tex"),
    argument = engine.asset.Texture("UI/intent_argument.tex"),
    composure = engine.asset.Texture("UI/intent_composure.tex"),
}

-------------------------------------------------------------------------------
-- Widget wrapping a Negotiation.Modifier

local IntentSlot = class( "Widget.IntentSlot", Widget )

function IntentSlot:init( card )
    IntentSlot._base.init( self )

    self.hover_check = true

    -- Prepare the selector
    local bg_border = 76
    local bg_border_scale = 0.5
    self.selector = self:AddChild( Widget.Panel( engine.asset.Texture("UI/argument_selector.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :Bloom( 0.1 )
        :SetSize( SLOT_SIZE*2.35, SLOT_SIZE*2.35 )
        :SetTintColour( UICOLOURS.FOCUS )
        :SetTintAlpha( 0 )
        :Hide()
    
    -- Prepare the contents
    self.content = self:AddChild( Widget() )

    -- Show glow behind everything
    self.glow = self.content:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex") ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.NEGOTIATION )
        :SetSize( SLOT_SIZE*1.4, SLOT_SIZE*1.4 )
        :SetTintAlpha( 1 )

    -- This intent's image
    self.art_container = self.content:AddChild( Widget() )
    self.art_mask = Widget.Image( engine.asset.Texture("UI/argument_mask.tex"), SLOT_SIZE, SLOT_SIZE )
    self.art = Widget.Image( nil, SLOT_SIZE-4, SLOT_SIZE-4 )
    self.art_container:SetStencilContext( true ):AddChildren
    {
        self.art_mask:SetMask( true ),
        self.art:SetStencilTest( true ),
    }

    -- This intent's frame
    self.frame = self.content:AddChild( Widget.Image( engine.asset.Texture("UI/argument_bg_intent.tex") ) )
        :SetSize( SLOT_SIZE, SLOT_SIZE )
        :SetHiddenBoundingBox(true)
        :SetTintColour(UICOLOURS.SUBTITLE)
        :Bloom( 0.1 )

    self.label = Widget.Label( "title", 72 ):EnableOutline( true )
    self.content:AddChild( self.label )

    -- Show intent's damage
    self.damage_bg = self.content:AddChild( Widget.Image( engine.asset.Texture( "UI/argument_damage.tex" ) ) )
        :SetSize( SLOT_SIZE, SLOT_SIZE )
    self.damage_label = self.content:AddChild( Widget.Label( "title", 22 ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :EnableOutline( 0.2 )

    -- Show intent's countdown
    self.countdown = self.content:AddChild( Widget() )
    self.countdown_icon = self.countdown:AddChild( Widget.Image( engine.asset.Texture("UI/ic_argument_spinner.tex"), 32, 32 ) )
        :Bloom( 0.1 )
    self.countdown_label = self.countdown:AddChild( Widget.Label( "title", 42 ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :EnableOutline( true )

    self:AssignCard( card )

    self.selector:SendToFront()

    self:SetFocusable( false )
    self:ShowToolTipOnFocus( false )
end

function IntentSlot:OnAdded()
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function IntentSlot:OnScreenModeChange( sm )
    IntentSlot._base.OnScreenModeChange( self, sm )

    self:SetLayoutScale( LAYOUT_SCALE[sm] )
end

function IntentSlot:AssignCard( card )
    self.card = card
    self:RefreshIntent()
end

function IntentSlot:OnUpdate( dt )
    IntentSlot._base.OnUpdate( self, dt )
    self.elapsed_time = (self.elapsed_time or 0) + dt

    local t = self.elapsed_time
    self.label:SetTintAlpha( 0.7 + 0.4 * math.sin( t * 6 ) )

    if self.hilite then
        self.selector:SetScale( 1.0 + 0.05 * math.sin ( t * 6 ) )
        self.glow:SetTintColour( UICOLOURS.FOCUS )
    else
        self.glow:SetTintColour( UICOLOURS.NEGOTIATION )
    end
end

function IntentSlot:HighlightTarget( hilite )
    self.hilite = hilite
    if not hilite then
        self.selector:SetScale( 1.0 )
            :SetBloom( 0 )
            :AlphaTo( 0, 0.1, easing.linear, function() self.selector:Hide() end )
    else
        self.selector:SetBloom( 0.05 )
            :Show()
            :AlphaTo( 1, 0.1 )
    end
end

function IntentSlot:IsEmpty()
    return self.card ~= nil
end

function IntentSlot:ConsumeIntent()
    -- Not yet cleared, but we want to be "hidden" while consumed.
    self.consumed = true
    self.content:ScaleTo( 1.0, 1.5, 0.2 )
    self:Delay( 0.2 )
    self:AlphaTo( 0, 0.2 )
end

function IntentSlot:CreateLabel( txt, color )
    self.label:SetText( txt ):SetGlyphColour( color )
    self.label:Show()

    self.content:Show():SetTintAlpha( 1.0 )

    self.art:Hide()
    self.damage_label:Hide()
    self.damage_bg:Hide()
end

function IntentSlot:CreateImage( image )
    self.art:SetTexture( image )
        :ApplySaturation( 0.5 )
        :SetTintColour( 0xcecefeff )
    self.art:Show()

    self.content:Show():SetTintAlpha( 1.0 )

    self.label:Hide()
    self.damage_label:Hide()
    self.damage_bg:Hide()
end

function IntentSlot:GeneratePersuasionText( card )
    local min_persuasion, max_persuasion, details = card.engine:PreviewPersuasion( card, true )

    local min_txt
    if min_persuasion < card.min_persuasion then
        min_txt = loc.format( "<#PENALTY>{1}</>",  min_persuasion )
    elseif min_persuasion > card.min_persuasion then
        min_txt = loc.format( "<#BONUS>{1}</>", min_persuasion )
    else
        min_txt = tostring(min_persuasion)
    end

    local max_txt
    if max_persuasion < card.max_persuasion then
        max_txt = loc.format( "<#PENALTY>{1}</>",  max_persuasion )
    elseif max_persuasion > card.max_persuasion then
        max_txt = loc.format( "<#BONUS>{1}</>", max_persuasion )
    else
        max_txt = tostring(max_persuasion)
    end

    local txt
    if min_txt == max_txt then
        txt = min_txt
    else
        txt = loc.format( "{1}-{2}", min_txt, max_txt )
    end

    return txt
end

function IntentSlot:ShowTurnDelay( delay )
    if delay == self.turn_delay then
        return
    end

    if delay and (delay > 1 or self.turn_delay ~= nil) then
        self.countdown:Show()
        self.countdown_label:SetText( tostring(delay) )
        self.countdown_icon:LayoutBounds( "after", "center", self.countdown_label ):Offset( -2, -1 )
        self.countdown:LayoutBounds( "center", "center", self.frame ):Offset( 3, 1 )

        if self.damage_bg:IsShown() then
            self.countdown:Offset( 0, 6 )
            if delay > 1 then
                self.damage_label:SetTintAlpha( 0.5 )
            else
                self.damage_label:SetTintAlpha( 1.0 )
            end   
        end


        if delay == 1 then
            self.countdown:AlphaTo( 0, 1.0, nil, function()
                self.countdown:Hide():SetTintAlpha( 1.0 )
            end )
        end
    else
        self.countdown:Hide()
    end

    self.turn_delay = delay
end

function IntentSlot:RefreshIntent( card )
    if self.consumed then
        return
    end

    card = card or self.card

    local hide_intents = not card:IsFlagged( negotiation_defs.CARD_FLAGS.INCEPTED ) and card.engine:CalculateBoolean( "HIDE_INTENT" )

    if hide_intents then
        self:CreateLabel( "?", UICOLOURS.WHITE )
        self:SetToolTip( LOC "UI.NEGOTIATION_PANEL.INTENT_UNKNOWN" )
        self:SetToolTipClass( nil )
        self.countdown:Hide()
    else
        if (card.max_persuasion or 0) > 0 then
            self:CreateImage( assets.damage )

            local txt = self:GeneratePersuasionText( card )

            self.damage_bg:Show()
            self.damage_label:SetText( txt )
                :LayoutBounds( "center", "bottom", self.damage_bg )
                :Offset( 0, 8 )
                :Show()

        elseif card.features and card.features.COMPOSURE then
            self:CreateImage( assets.composure )
        elseif card:IsFlagged( negotiation_defs.CARD_FLAGS.INCEPTED ) then
            self:CreateImage( card.icon )
            self.frame:SetTintColour( UICOLOURS.PENALTY )
        else
            self:CreateImage( assets.argument )
        end

        self:SetToolTip( card )
        self:SetToolTipClass( Widget.TooltipIntent )

        -- Update the countdown
        local delay = card:GetTurnDelay()
        self:ShowTurnDelay( delay )
    end
end

function IntentSlot:OnGainHover()
    self.selector:Bloom( 0.1 )
    local parent = self:GetAncestor( Widget.NegotiationPanel )
    if parent then
        parent:SetHoverTarget( self.card )
    end
end

function IntentSlot:OnLoseHover()
    local parent = self:GetAncestor( Widget.NegotiationPanel )
    if parent then
        parent:SetHoverTarget( nil )
    end
    self.selector:Bloom( 0.05 )
end

function IntentSlot:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        
        local parent = self:GetAncestor( Widget.NegotiationPanel )
        if parent then parent:SetHoverTarget( self.card ) end

        self.selector:Bloom( 0.1 )

        -- In inspect mode with controller, show the target more clearly
        if not self.hilite and self.focus then
            self.selector:Show():SetScale( 1.1 )
            self.selector:SetScale( 1.1 )
            self.selector:SetTintColour( UICOLOURS.SUBTITLE )
            self.selector:AlphaTo( 1, 0.1 )
            self.glow:SetTintColour( UICOLOURS.SUBTITLE )
        end
    else

        local parent = self:GetAncestor( Widget.NegotiationPanel )
        if parent then parent:SetHoverTarget( nil ) end

        self.selector:Bloom( 0.05 )

        if not self.hilite then
            self.selector:AlphaTo( 0, 0.1, easing.linear, function()
                self.selector:Hide()
                self.selector:SetTintColour( UICOLOURS.FOCUS )
            end )
            self.glow:SetTintColour( UICOLOURS.NEGOTIATION )
        end
    end
end

function IntentSlot:OnGainFocus()
    IntentSlot._base.OnGainFocus(self)
    self:UpdateImage()
end

function IntentSlot:OnLoseFocus()
    IntentSlot._base.OnLoseFocus(self)
    self:UpdateImage()
end

function IntentSlot:OnGainHover()
    IntentSlot._base.OnGainHover(self)
    self:UpdateImage()
end

function IntentSlot:OnLoseHover()
    IntentSlot._base.OnLoseHover(self)
    self:UpdateImage()
end

-------------------------------------------------------------------------------
-- Widget containing IntentSlot widgets

local IntentContainer = class( "Widget.IntentContainer", Widget )

function IntentContainer:init( negotiator, diameter )
    IntentContainer._base.init( self )

    self.minigame = negotiator.engine
    self.negotiator = negotiator
    self.diameter = diameter or 360

    self.slot_wheel = self:AddChild( Widget.ArgumentContainer( self.diameter ) )
        :SetIntentsMode()
end

function IntentContainer:GetDefaultFocus()
    return self.slot_wheel:GetDefaultFocus()
end

function IntentContainer:FindIntentWidget( card )
    local slot_widgets = self.slot_wheel:GetWidgets()

    for i, slot in ipairs( slot_widgets ) do
        if slot.card == card then
            return slot
        end
    end
end

function IntentContainer:RefreshIntents()
    local slot_widgets = self.slot_wheel:GetWidgets()

    for i, slot in pairs( slot_widgets ) do
        slot:RefreshIntent()
    end
    self.slot_wheel:Layout()
end

function IntentContainer:ConsumeIntent( card )
    local slot = self:FindIntentWidget( card )
    if slot then
        slot:ConsumeIntent()
    end
end

function IntentContainer:FindAvailableSlot()
    local slot_widgets = self.slot_wheel:GetWidgets()
    for i, slot in ipairs( slot_widgets ) do
        if not slot:IsEmpty() then
            return slot
        end
    end
end

function IntentContainer:OnIntentsChanged()
    for i, slot in ipairs( self.slot_wheel:GetWidgets() ) do
        self.slot_wheel:SnapWheelAngle( slot )
    end
end

function IntentContainer:OnIntentRemoved( card )
    local slot = self:FindIntentWidget( card )
    self.slot_wheel:RemoveModifierSlot( slot )
    self.slot_wheel:Layout()
end

function IntentContainer:OnIntentChanged( card, clone )
    local slot = self:FindIntentWidget( card )
    if slot then
        slot:RefreshIntent( clone )
    end
end

function IntentContainer:OnIntentAdded( new_card, source )
    local slot = self:FindIntentWidget( new_card )
    if slot == nil then
        slot = Widget.IntentSlot( new_card )
        self.slot_wheel:AddSlotWidget( slot ):Hide()
        self.slot_wheel:Layout()

        if source then
            -- Wait for intents to animate to target positions.
            self:Delay( 0.5 )
        end
    end

    local source_widget
    if is_instance( source, Negotiation.Card ) then
        source_widget = self.parent:FindCardWidget( source )
    elseif is_instance( source, Negotiation.Modifier ) then
        source_widget = self.parent:FindSlotWidget( source )
    end

    local floater
    if source_widget then

        floater = self:AddChild( Widget.IntentSlot( new_card ))

        -- Position the floater over the source card
        local source_x, source_y = floater.parent:TransformFromWidget( source_widget, 0, 0 ) 

        -- Get target position
        local target_x, target_y = floater.parent:TransformFromWidget( slot, 0, 0 )

        -- Animate the floater from the card to the target
        floater:SetPos( source_x, source_y )

        local SPAWN_TIMING = 0.25
        floater:ScaleTo( 1.5, 1.0, SPAWN_TIMING, easing.inQuad )
        floater:DelayIfAnimating()
        floater:Delay( SPAWN_TIMING )

        floater:MoveTo( target_x, target_y, SPAWN_TIMING, easing.inCirc )
        floater:Delay( SPAWN_TIMING )
        AUDIO:PlayEvent( SoundEvents.negotiation_argument_destroyed )

        floater:Remove()
    end

    slot:AssignCard( new_card )
    slot:StartUpdating()
    slot:ScaleTo( 2.0, 1.0, 0.3, easing.outQuad ):Show()

    AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/intent_appears_single" )
end

function IntentContainer:SetFocusable( focusable )
    local slot_widgets = self.slot_wheel:GetWidgets()
    for i, slot in ipairs( slot_widgets ) do
        slot:SetFocusable( focusable )
    end
end

function IntentContainer:SetRotatingMode( on )
    self.slot_wheel:SetRotatingMode( on )
end

function IntentContainer:SetLayoutSlots( on )
    self.slot_wheel:SetLayoutSlots( on )
end

function IntentContainer:SetTargetDisplay( valid )
    self.slot_wheel:SetWheelColour( valid and UICOLOURS.FOCUS or UICOLOURS.SUBTITLE )
end

function IntentContainer:OnGainFocus()
    IntentContainer._base.OnGainFocus(self)
    self:UpdateImage()
end

function IntentContainer:OnLoseFocus()
    IntentContainer._base.OnLoseFocus(self)
    self:UpdateImage()
end

function IntentContainer:OnGainHover()
    IntentContainer._base.OnGainHover(self)
    self:UpdateImage()
end

function IntentContainer:OnLoseHover()
    IntentContainer._base.OnLoseHover(self)
    self:UpdateImage()
end

function IntentContainer:UpdateImage()
    self:SetRotatingMode( not ( self.hover or self.focus ) )
end