require "ui/widget"

local Label = class("Widget.Label", Widget)
function Label:init( font, sz, txt)
    Label._base.init(self, engine.rendering.TextNode())
    
    self.font = font or "body"
    self.model:SetFont(self.font)
    if sz then
        self:SetFontSize(sz)
    end
    self:SetGlyphColour( UICOLOURS.TEXT )

    self:SetHoverLinkGlyphColour( UICOLOURS.TEXT_LINK )
    
    if txt then
        self:SetText(txt)
    end
    
    --self:EnableOutline(true)
    if TheGame.local_settings.DEBUG_DRAW_LABELS then
        self.model:SetDebugRender(true)
    end
end

function Label:SetCursor(pos)
    self.model:SetCursor(pos)
    return self
end

function Label:SetWordWrap(val)
    self.model:SetWordWrap(val)
    return self
end

function Label:SnapSpool()

    if self.spool_time then
        self:StopUpdating()
        self.model:SetSpool(1)
        self.spool_time = nil
        self.total_spool_time = nil
    end


end

function Label:Spool(speed)
    
    local characters = string.len(self.text)
    if characters > 0 then
        self:StartUpdating()
        speed = speed or 200
        if speed > 0 then
            self.total_spool_time = characters / speed 
            self.spool_time = 0
            self.model:SetSpool(0)
        else
            self.spool_time = nil
            self.total_spool_time = nil
            self:StopUpdating()
            self.model:SetSpool(1)
        end
    else
        self.model:SetSpool(1)
    end
    return self
end

function Label:OnUpdate(dt)
    Label._base.OnUpdate(self, dt)
    if self.spool_time then
        self.spool_time = self.spool_time + dt
        
        if self.spool_time >= self.total_spool_time then
            self.spool_time = nil
            self.total_spool_time = nil
            self:StopUpdating()
            self.model:SetSpool(1)
        else
            local t = self.spool_time / self.total_spool_time
            self.model:SetSpool(t)
        end
    end
end

function Label:SetDebugRender(val)
    self.model:SetDebugRender(val)
    return self
end

function Label:SetFont(font)
    self.font = font
    self.model:SetFont(font)
    return self
end

function Label:SetUserData( userdata )
    self.userdata = userdata
    return self
end

function Label:GetUserData( idx )
    return self.userdata and self.userdata[ idx ]
end

function Label:ForceSetText(text)
    local text = self.text
    self.text = nil
    self:SetText( text )
end

function Label:GetLinks()
    self.links = self.links or self.model:GetLinks()
    return self.links, self.hover_fn
end

function Label:SetHoverOverride(idx) --nil to clear
    self.model:SetHoverSpanOverride(idx)
end

function Label:SetText(text)
    text = text or ""
    if text ~= self.text then
        self.links = nil
        self.text = text
        local formatted_text = SetFormattedText( self.model, text )

        if self.drop_shadow then
            self.model:AddMarkup( 0, #formatted_text, engine.rendering.TextNode.MARKUP_SHADOW )
        end

        self:UpdateTransform()
        self:InvalidateBBox()
    end
    return self
end

function Label:AppendText( text )
    if text and #text > 0 then
        self.links = nil
        text = ApplyFormatting( self.model, self.text .. text )
        self.text = text
        self.model:SetText( self.text )
        if not self.w then
            self:InvalidateBBox()
        end
    end
end

function Label:SetRawText(text)
    text = text or ""
    if text ~= self.text then
        self.links = nil
        self.text = text
        self.model:ClearMarkup()
        self.model:SetText( text )

        self:UpdateTransform()
        self:InvalidateBBox()
    end
    return self
end

function Label:SetValue( ... )
    return self:SetText(...)
end

function Label:GetText()
    return self.text
end

function Label:SetFontSize(sz)
    self.model:SetFontSize(sz)
    self.font_sz = sz
    self:OverrideLineHeight( sz * LINE_HEIGHT_TEXT_MULT )
    return self
end

function Label:OverrideLineHeight(h)
    self.model:OverrideLineHeight(h)
    return self
end

function Label:SetHoverLinkGlyphColour( r, g, b, a )
    if r and not g and not b then
        r,g,b,a = HexColour(r)
    end
    self.model:SetHoverLinkGlyphColour( r, g, b, a )
    self.model:SetHighlightColour( r, g, b, a )
    return self
end

function Label:SetGlyphColour( r, g, b, a )
    if r and not g and not b then
        r,g,b,a = HexColour(r)
    end
    self.model:SetGlyphColour( r, g, b, a )
    return self
end

function Label:SetOutlineColour( r, g, b, a )
    if r and not g and not b then
        r, g, b, a = HexColour(r)
    end
    self.model:SetOutlineColour( r, g, b, a )
    return self
end

function Label:SetShadowColour( r, g, b, a )
    if r and not g and not b then
        r, g, b, a = HexColour(r)
    end
    self.model:SetShadowColour( r, g, b, a )
    return self
end

function Label:SetShadowOffset( dx, dy )
    self.model:SetShadowOffset( dx, dy )
    return self
end

function Label:EnableShadow( enabled )
    self.drop_shadow = enabled
    if enabled and self.text then
        self:ForceSetText( self.text )
    end
    return self
end

function Label:EnableOutline( enable )
    self.model:EnableOutline( enable )
    return self
end

function Label:SetSDFThreshold( threshold, boldThreshold )
    self.model:SetSDFThreshold( threshold, boldThreshold or threshold * 0.9 )
    return self
end


function Label:TryFitToRegion(smallest_text_size, largest_text_size)
    assert(smallest_text_size <= largest_text_size, "Bad size arguments")
    assert(self.w and self.h, "No region bounds set")

    local size = math.round( (smallest_text_size + largest_text_size)/2)
    while smallest_text_size < largest_text_size do
        self.model:SetFontSize(size)
        local x1, y1, x2, y2 = self.model:GetRenderBounds()

        local new_size
        if math.abs(x1-x2) > self.w or math.abs(y1-y2) > self.h then
            largest_text_size = size
            new_size = math.floor( (smallest_text_size + size)/2)

        else
            smallest_text_size = size
            new_size = math.floor( (largest_text_size + size)/2)
        end
        if new_size == size then
            break
        else
            size = new_size
        end
    end

    return size
end

function Label:SetRegionSize(w, h)
    self.w = w
    self.h = h
    self.model:SetRegionSize(w,h)
    return self
end

function Label:SetSize(w, h)
    self:SetRegionSize( w, h )
    return self
end

function Label:SetAutoSize( wrap_width )
    self.wrap_width = wrap_width
    self.w = nil
    self.h = nil
    self.model:SetAutoSize( wrap_width )
    return self
end

function Label:GetAutoSize( wrap_width )
    return self.wrap_width
end

function Label:SetBloom(b)
    self.model:SetBloom(b)
    return self
end

function Label:LeftAlign()
    self.model:SetHAlign( engine.rendering.TextNode.HANCHOR_LEFT )
    return self
end

function Label:RightAlign()
    self.model:SetHAlign( engine.rendering.TextNode.HANCHOR_RIGHT )
    return self
end

function Label:CenterH()
    self.model:SetHAlign( engine.rendering.TextNode.HANCHOR_MIDDLE )
    return self
end

function Label:TopAlign()
    self.model:SetVAlign(1)
    return self
end

function Label:BottomAlign()
    self.model:SetVAlign(2)
    return self
end

function Label:CenterV()
    self.model:SetVAlign(0)
    return self
end

function Label:GetFont()
    return self.font    
end

function Label:GetBoundingBox()

    if self.w and self.h then
        return -self.w/2,  -self.h/2, self.w/2,  self.h/2
    end
    
    if self.text and self.text ~= "" then
        local x1, y1, x2, y2 = self.model:GetRenderBounds()
        return x1, y1, x2, y2
    end

    return 0, 0, 0, 0
end

function Label:GetRenderSize()
    local x1, y1, x2, y2 = self.model:GetRenderBounds()
    return math.abs(x2-x1), math.abs(y2-y1)
end

function Label:GetRenderHeight()
    local x1, y1, x2, y2 = self.model:GetRenderBounds()
    return math.abs(y2-y1)
end

function Label:SetClickLinkFn( fn, ... )
    self.fn = { fn, ... }
    self.hover_check = (fn ~= nil or self.hover_fn ~= nil)
    return self
end

function Label:SetHoverLinkFn( fn, ... )
    self.hover_fn = { fn, ... }
    self.hover_check = (fn ~= nil or self.fn ~= nil)
    return self
end

function Label:OnGainHover()
end

function Label:OnControlModeChange( cm, device_id )
    Label._base.OnControlModeChange( self, cm, device_id )
    self:ForceSetText( self.text )
end

function Label:OnLoseHover()
    self.hover_tag, self.hover_text = nil, nil
    self.model:HitTest() -- clear hover status on the model.
end

function Label:CheckMouseHover(x,y)
    if self.ignore_input then return end
    if not self.shown then return end
    
    if self.hover_check then
        local hover_tag, hover_text = self.model:HitTest( x, y, true )
        if hover_tag ~= self.hover_tag or hover_text ~= self.hover_text then
            self.hover_tag, self.hover_text = hover_tag, hover_text 
            if self.hover_fn then
                self.hover_fn[1]( self, self.hover_tag, self.hover_text, table.unpack( self.hover_fn, 2) )
            end
        end
        if self.hover_tag and self.hover_text then
            -- self:SetHover()
            return true, self
        end
    end

    return Label._base.CheckMouseHover(self,x,y)
end

function Label:HandleControlDown( control, device )
    if control:Has( Controls.Digital.MENU_ACCEPT, Controls.Digital.MENU_ALT ) and self.hover and self.fn then
        local x, y = TheGame:GetInput():GetMousePos()
        self.hover_tag, self.hover_text = self.model:HitTest( x, y, true )
        if self.hover_tag and self.hover_text then
            if control:Has( Controls.Digital.MENU_ACCEPT ) then
                self.down = Controls.Digital.MENU_ACCEPT
            else
                self.down = Controls.Digital.MENU_ALT
            end
            return true
        end
    end
end

function Label:HandleControlUp( control, device )
    if self.down and control:Has( self.down ) then
        if self.hover_tag and self.hover_text then
            if self.focus or self.hover then
                if self.fn then self.fn[1]( self, self.hover_tag, self.hover_text, self.down, table.unpack( self.fn, 2 )) end
            end
            self.down = nil
            return true
        end
    end
end

function Label:__tostring()
    local txt = tostring(self.text)
    return string.format("Label Widget (%s%s)", string.sub(txt:gsub( "\n", " " ), 1, 16), string.len(txt) > 16 and "..." or "")
end


function Label:GetCursorAtPoint(x,y, override_line)
    if override_line or self:CheckHit(x,y) then
        x,y = self:TransformFromWorld(x,y)
        local cursor = self.model:GetCursorAtLocalPoint(x,y, override_line)
        return cursor
    end
end
