require "ui/widget"
require "ui/widgets/menubutton"

local Menu = class("Widget.Menu", Widget)

function Menu:init( buttonclass, buttonw, buttonh )
	Menu._base.init(self)
	self.buttonw, self.buttonh = buttonw, buttonh
	self.options = {}
	self.buttons = {}
	self.font_size = 18
	self.buttonclass = buttonclass or Widget.MenuButton
	self.layoutx, self.layouty = "center", "below"
	self.hpad, self.vpad = 0, 0
	self.root = Widget()
	self:AddChild(self.root)
	
	self:SetFocusGroup( FOCUS_GROUP.LAST_FOCUS )
	self:SetFocusLock(true)
end

function Menu:SetFontSize(val)
	self.font_size = val
	return self
end

function Menu:SetLastDefault(val)
	self.last_default = val
	return self
end

function Menu:GetOptionsCount( )
	return #self.options
end

function Menu:PrepareAnimation()
	for k = #self.buttons, 1, -1 do
		self.buttons[k]:SetTintAlpha(0)
		AUDIO:PlayEvent(SoundEvents.mouseover)
	end
	return self
end

function Menu:AnimOptions()
	if self.anim_coro then
		self.anim_coro:Stop()
	end

	self.anim_coro = self:StartCoroutine( function()
		if TheGame:FE():GetControlMode() == CONTROL_MODE.GAMEPAD then
			self.buttons[1]:SetFocus(true)
		end
		-- Hide all options.  Use alpha so we don't mess up Focus.
		for k = #self.buttons, 1, -1 do
			self.buttons[k]:SetTintAlpha(0)
			AUDIO:PlayEvent(SoundEvents.mouseover)
		end
		-- Reveal one by one
		for k = 1, #self.buttons do
			self.buttons[k]:SetTintAlpha(1)
			AUDIO:PlayEvent(SoundEvents.mouseover)
			self:Delay(0.025)
		end
	end )
	return self
end

function Menu:GetButtons()
	return self.buttons
end

function Menu:SetOptions( opt )
	self:InvalidateBBox()
	self:Clear()

	if opt then
		for k,v in ipairs(opt) do
			local button = self:AddOption( v )
			if v.name then
				button:SetName( v.name )
			end

			if v.snd then
				button:SetClickSoundEvent(v.snd)
			end

		end

		if #self.buttons >= 2 then
			self.buttons[1]:SetFocusDir( "up", self.buttons[ #self.buttons ] )
			self.buttons[ #self.buttons ]:SetFocusDir( "down", self.buttons[ 1 ] )
		end
	end

	self:Layout()
	return self
end

function Menu:Enable()
	Menu._base.Enable(self)
	for k,v in pairs(self.buttons) do
		v:Enable()
	end
end

function Menu:Disable()
	Menu._base.Disable(self)
	for k,v in pairs(self.buttons) do
		v:Disable()
	end
end

function Menu:Clear()
	if self.anim_coro then
		self.anim_coro:Stop()
		self.anim_coro = nil
	end

	for k,v in pairs(self.buttons) do
		v:Remove()
	end

	table.clear(self.buttons)
	table.clear(self.options)
end

function Menu:AddOption( opt )
	table.insert( self.options, { txt = opt.txt, fn = opt.fn } )
	local button = opt.buttonclass and opt.buttonclass( opt.txt, opt.fn, opt ) or self.buttonclass( opt.txt, opt.fn )

	if opt.icon then
		button:SetIcon( opt.icon )
	end

	if opt.cta then
		if button.SetCta then
			button:SetCta( true )
		else
			button:SetTintColour( UICOLOURS.OBJECTIVE )
		end
	end

	if self.buttonw and self.buttonh then
		button:SetSize(self.buttonw, self.buttonh)
	elseif button.SizeToLabel then
		button:SizeToLabel( 4, 4 )
	end
	if opt.tt then
		button:SetToolTip(opt.tt)
	end

	if button.SetTextSize and self.font_size then
		button:SetTextSize( self.font_size )
	end

	if opt.disabled or not self.enabled then
		button:Disable()
	else
		button:Enable()
	end

	self.root:AddChild(button)
	table.insert(self.buttons, button)

	return button
end

function Menu:GetButton(idx)
	return #self.buttons >= idx and self.buttons[idx]
end

function Menu:FocusOption(idx)
	if idx < #self.buttons and idx > 0 then
		self.buttons[idx]:SetFocus()
	end
	return self
end

function Menu:SetHoverCheck( hover_check )
	for i, btn in ipairs( self.buttons ) do
		btn.hover_check = hover_check
	end
end

function Menu:SetFocus(from_child)
	if not from_child then
		if #self.buttons > 0 then
			print( "Menu.SetFocus", self.buttons[self.last_default and #self.buttons or 1] )
			self.buttons[self.last_default and #self.buttons or 1]:SetFocus()
		end
	else
		Menu._base.SetFocus(self, from_child)
	end
end

function Menu:SetLayout( layoutx, layouty, hpad, vpad )
	self.layoutx = layoutx or self.layoutx
	self.layouty = layouty or self.layouty
	self.hpad = hpad or self.hpad
	self.vpad = vpad or self.vpad
	return self
end

-- In case we don't want this menu to resize every button
function Menu:AllowMultiSizeButtons()
	self.allow_multisize_buttons = true
	return self
end

function Menu:Layout()
	local max_width = 0
	if self.buttonw == nil then
		-- If there is no explicit button width, use max_width across all buttons.
		for i, btn in ipairs( self.buttons ) do
			local w, h = btn:GetSize()
			max_width = math.max( max_width, w )
		end
	end

	for i, btn in ipairs(self.buttons) do
		if max_width > 0 and not self.allow_multisize_buttons then
			local w, h = btn:GetSize()
			btn:SetSize( max_width, h )
		end
		if i == 1 then
			btn:SetPos( 0, 0 )
		else
			btn:LayoutBounds( self.layoutx, self.layouty ):Offset( self.hpad, self.vpad )
		end
	end

	return self
end

function Menu.CreatePopUp(options, x, y, fn_closed, button_width, button_height)
	local popuproot = Widget()
	for k,v in pairs(options) do
		local fn = v.fn
		v.fn = function() if fn then fn() end if not popuproot.removed then popuproot:Remove() end end
	end

	local padding = 10
    local menu = Widget.Menu( Widget.PanelButton, button_width, button_height ):SetLayout( "center", "below", 0, -10 ):SetOptions( options )
    
    local w, h = menu:GetSize()
    local x1, y1, x2, y2 = menu:GetBoundingBox()
    menu:Offset( 0, -y2 * 2 ) -- to put the menu BELOW the <x,y> point.
    local cy =  (y1 + y2)/2 - y2*2 
    popuproot:AddChildren{
        Widget.SolidBox(w + padding*2, h + padding*2):SetPos( 0, cy ):BlockMouse():SetColour(0x00000077),
        menu,
    }
    --local screen = TheGame:FE():GetTopScreen()
    --popuproot:SetPos( screen:TransformFromWorld( x, y ))
    --screen:AddChild( popuproot )
    --popuproot:SetFocus()

    local doclose = function()
    	if not popuproot.removed then
    		popuproot:Remove()
    	end
    	if fn_closed then
    		fn_closed()
    	end
	end
    popuproot.OnLoseFocus = doclose

   	local old_fn = popuproot.HandleControlDown

   	popuproot.HandleControlDown = function(_, control) 
		
		if old_fn then
			if old_fn(_, control) then
				return true
			end
		end

		if control:Has( Controls.Digital.MENU_CANCEL )  then
			doclose()
			return true
		end
   	end
    return popuproot
end