local CONTENT_LEFT_MARGIN = 15 -- Extra size added on the left of scroll boxes, so that the contents can scale up on hover without clipping
require "ui/widgets/compendiumwidget"
require "ui/screens/mettlescreen"
------------------------------------------------------------------------------------------
-- Shows the grafts panel

local assets = {
    selected = engine.asset.Texture("UI/mettle_selected.tex"),
    divider = engine.asset.Texture("large/dialog_divider_top.tex" ),
    available = engine.asset.Texture("UI/mettle_available.tex"),
}

local MettleCompendiumWidget = class( "Widget.MettleCompendiumWidget", Widget.CompendiumWidget )

MettleCompendiumWidget.CONTROL_MAP = 
{
}

------------------------------------------------------------------------------------------
-- Shows mettle upgrades

local graft_w
local graft_h

function MettleCompendiumWidget:init( width, height, sidebar_width )
    MettleCompendiumWidget._base.init( self, width, height, sidebar_width )

    -- Add Mettle count header
    self.mettle_divider = self.scroll_root:AddScrollChild( Widget.Image( assets.divider ) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.mettle_count = self.scroll_root:AddScrollChild( Widget.Label( "title", FONT_SIZE.BODY_TEXT ) )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :LeftAlign()
        :SetAutoSize( graft_w )
        :SetWordWrap( true )
        :Bloom( 0.1 )

    -- Add container for Mettle upgrade rows
    self.graft_container = self.scroll_root:AddScrollChild( Widget() )

    -- Get all the grafts
    self.graft_defs = {}
    local grafts = Content.GetAllGrafts()
    for i, graft_def in ipairs( grafts ) do
        if graft_def.type == GRAFT_TYPE.METTLE_UPGRADE and not graft_def.hide_in_compendium then
            table.insert( self.graft_defs, graft_def )
        end
    end
end

function MettleCompendiumWidget:CanNav()
    return true
end

function MettleCompendiumWidget:HasSidebar()
    return true
end

function MettleCompendiumWidget:SortOrderSelected( option )
end

function MettleCompendiumWidget:RefreshContent()
    MettleCompendiumWidget._base.RefreshContent()

    if self.graft_widgets then
        return
    end

    self.graft_widgets = {}
    self.graft_container:DestroyAllChildren()
    self.last_focus = nil

    -- Calculate the size of each widget
    self.scroll_inner_w = self.content_w - 5
    graft_w = self.scroll_inner_w - 40
    graft_h = 112

    -- Resize the count label
    self.mettle_count:SetAutoSize( graft_w )

    -- Resize the banner
    self.mettle_divider:SetSize( graft_w*0.5, 4 )

	-- Get all the grafts
    for i, graft_def in ipairs(self.graft_defs) do
        local widget = self.graft_container:AddChild( Widget.MettleCompendiumRowWidget( graft_w, graft_h, graft_def ) )
        table.insert( self.graft_widgets, widget )
    end

    -- Refresh locked status
    for background_id, row in pairs(self.agent_filters) do
        local is_character_locked = not Screen.Compendium.ShowAllUnlocks() and TheGame:GetGameProfile():IsCharacterLocked( background_id)
        row:SetLocked( is_character_locked )
    end

    for background_id, row in pairs(self.agent_filters) do
        row:SetProgress( self.graft_counts[background_id] )
    end

    -- Layout all the grafts
    self:LayoutContent()

    -- Update filtering
    self.sidebar:RefreshFilters()


    return self
end

function MettleCompendiumWidget:UpdateCount()
    self.graft_upgrades = {}
    -- Get owned upgrades per character
    for k, data in ipairs(GetAllPlayerBackgrounds()) do
        if data.id then
            self.graft_upgrades[data.id] = TheGame:GetGameProfile():GetMettleUpgrades( data.id )
        end
    end
    
    -- Reset counts
    self.graft_counts = {}
    self.graft_counts["UNLOCKED"] = 0
    self.graft_counts["UNLOCKABLE"] = 0

    for i, graft_def in ipairs(self.graft_defs) do
        for background_id, upgrades in pairs(self.graft_upgrades) do

            if self.graft_counts[background_id] == nil then
                self.graft_counts[background_id] = {
                    UNLOCKED = 0,
                    UNLOCKABLE = 0,
                }
            end

            -- Get graft progress
            local unlocked, unlockable = upgrades[ graft_def.id ] or 0, #graft_def.values

        	-- Update total counts
            self.graft_counts["UNLOCKED"] = self.graft_counts["UNLOCKED"] + unlocked
            self.graft_counts["UNLOCKABLE"] = self.graft_counts["UNLOCKABLE"] + unlockable

            -- Update character-specific counts
            self.graft_counts[background_id]["UNLOCKED"] = self.graft_counts[background_id]["UNLOCKED"] + unlocked
            self.graft_counts[background_id]["UNLOCKABLE"] = self.graft_counts[background_id]["UNLOCKABLE"] + unlockable
        end
    end

end

function MettleCompendiumWidget:LayoutContent()

    self.mettle_count:LayoutBounds( "left", "top", 35, 0 )
    self.mettle_divider:LayoutBounds( "left", "below", self.mettle_count )
        :Offset( -110, -15 )
	self.graft_container:StackChildren( 10 )
        :LayoutBounds( "left", "below", self.mettle_count )
        :Offset( -13, -15 )
    self.scroll_root:SetVirtualMargin( 20 )
    	:RefreshView()
    	:LayoutBounds( "after", "top", self.sidebar )
        :Offset( 8-CONTENT_LEFT_MARGIN, -1 )

    return self
end

function MettleCompendiumWidget:AddSidebarFilters()
	-- Add character filters
    self.sidebar:AddFilterHeader( LOC"UI.CARDCOMPENDIUM.FILTER_CHARACTERS" )
    self.series_filter_group = self.sidebar:AddFilterWidget( Widget.FilterRadioGroup() )
    self.agent_filters = {}
    for k, data in ipairs(GetAllPlayerBackgrounds()) do
        if data.id then
            local row = self.series_filter_group:AddFilter( Widget.MettleSeriesFilterRow( data.data.player_agent, { data.id } ) )
            self.agent_filters[data.id] = row
            -- Lock filters for locked characters
            local is_character_locked = not Screen.Compendium.ShowAllUnlocks() and TheGame:GetGameProfile():IsCharacterLocked( data.data.player_agent )
            self.agent_filters[data.id]:SetLocked( is_character_locked )
        end
    end
    self.series_filter_group:SetSelected( 1 )
end

function MettleCompendiumWidget:OnFiltersChanged( filters )
    self:FilterGraftWidgets( self.graft_widgets, filters )
    self:LayoutContent()
end

function MettleCompendiumWidget:FilterGraftWidgets( graft_widgets, filters )
    -- Update the header count to match the current character
    local character = self.series_filter_group:GetSelected():GetAgent()
    local mettle = TheGame:GetGameProfile():GetMettlePoints( character.id ) or 0
    self.mettle_count:SetText( loc.format( LOC"UI.CARDCOMPENDIUM.METTLE_AMOUNT", character:GetName(), mettle ) )

    -- Go through each of the grafts
    for i, graft in pairs( graft_widgets ) do
        local upgrades = self.graft_upgrades[ character.id ] or table.empty
        graft:SetCurrentLevel( upgrades[ graft:GetDef().id ] or 0 )
    end
end

-- Returns the number of unlocked grafts and the number of unlockable grafts, to be displayed on the nav bar
function MettleCompendiumWidget:GetCount()
    if self.graft_counts == nil then
        self:UpdateCount()
    end

    return self.graft_counts["UNLOCKED"] or 0 , self.graft_counts["UNLOCKABLE"] or 0
end

function MettleCompendiumWidget:GetDefaultFocus()
    if self.sidebar_focus then
        return self.sidebar:GetDefaultFocus()
    end

    for k, v in pairs( self.graft_widgets ) do
        if v:IsShown() then return v end
    end
end

-------------------------------------------------------------------------
-- A single Mettle row, adjusted for the Compendium

local BUTTON_W = 90

local MettleCompendiumRowWidget = class( "Widget.MettleCompendiumRowWidget", Widget.MettleRowWidget )

function MettleCompendiumRowWidget:init( w, h, def )
    MettleCompendiumRowWidget._base.init( self, w, h, def, 0, nil, Widget.MettleCompendiumOptionButton )
end

-------------------------------------------------------------------------
-- A single Mettle button, adjusted for the Compendium

local MettleCompendiumOptionButton = class( "Widget.MettleCompendiumOptionButton", Widget.MettleOptionButton )

function MettleCompendiumOptionButton:init( mettle_graft, mettle_value, mettle_cost, onclick_fn )
    MettleCompendiumOptionButton._base.init( self, mettle_graft, mettle_value, mettle_cost, onclick_fn )
    self:DisableAnimateScales()
end

function MettleCompendiumOptionButton:Refresh()
    self:SetText( self.mettle_value )
    if self.selected then
        self.bg_outline:SetTexture( assets.selected )
        self.bg_panel:SetTintColour( 0xC6AB3Eff )
        self:SetColour( UICOLOURS.QUEST )
        self:SetTextColour( UICOLOURS.DARK_BODYTEXT )
        self.cost_badge:SetTintColour( UICOLOURS.QUEST )
        self:SetOnClickFn(nil)
    elseif self.acquired then
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( 0x6B5A16ff )
        self:SetColour( UICOLOURS.QUEST )
        self:SetTextColour( 0xC6AB3Eff )
        self.cost_badge:SetTintColour( UICOLOURS.QUEST )
        self:SetOnClickFn(nil)
    else
        self.bg_outline:SetTexture( assets.available )
        self.bg_panel:SetTintColour( UICOLOURS.DARK_BODYTEXT )
        self:SetColour( 0x666666ff )
        self:SetTextColour( 0x666666ff )
        self.cost_badge:SetTintColour( 0x666666ff )
        self:SetOnClickFn(nil)
    end

    -- Update cost badge
    self.cost_badge:Show()
    
    self.subtitle:Show()
        :SetText( loc.format(LOC"UI.METTLE_SCREEN.METTLE_COST",self.mettle_cost) )
    
    local w, h = self.subtitle:GetSize()
    self.cost_badge:SetSize( w + 10, h + 0 )
    self.text_container:SendToFront()

    self:UpdateImage()
    return self
end

function MettleCompendiumOptionButton:UpdateImage()
	Widget.IconButton.UpdateImage( self )
    self.subtitle:SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.DARK_BODYTEXT )
    return self
end