require "ui/widget"
require "ui/widgets/iconbutton"
require "ui/widgets/solidbox"
require "ui/widgets/label"
require "ui/widgets/tinyportrait"

local OPTION_HEIGHT = 36
local OPTION_TEXT_SIZE = 24
local OPTION_SPACING = 10

local ICON_SIZE = OPTION_HEIGHT - 8

local MultiChoiceOption = class( "Widget.MultiChoiceOption", Widget.IconButton )

MultiChoiceOption.show_child_tooltip_on_focus = true
MultiChoiceOption.no_reticule_snap = true

function MultiChoiceOption:init( txt, data, width )
    MultiChoiceOption._base.init( self, txt, nil, width, OPTION_HEIGHT, OPTION_TEXT_SIZE )
    self.width = width
    self.data = data
    self:SetIcon( global_images.question )
    
    if data and data.sfx then self:SetClickSoundEvent( data.sfx ) end
    self.label:SetHoverLinkFn( UIHelpers.ItemTooltips ):LeftAlign()
    self.label:EnableShadow( true )
    self.label.no_reticule_snap = true
    local _, label_h = self.label:GetRenderSize()
    if label_h > OPTION_HEIGHT then
        -- self:SetIconSize( OPTION_HEIGHT, OPTION_HEIGHT )
        self:SetSize( width, label_h )
    end

    self:ShowToolTipOnFocus()

    self.txt = txt
end

function MultiChoiceOption:SetHoverLinkParameters( ... )
    self.label:SetHoverLinkFn( UIHelpers.ItemTooltips, ... )
end

function MultiChoiceOption:GetLinks()
    return self.label:GetLinks()
end

function MultiChoiceOption:SetIndex( multi, idx )
    self:SetFn( function()
        if self.data and self.data.preclick_fn then
            if not self.data.preclick_fn( self ) then
                return
            end
        end

        multi:OnSelectOption( idx ) 
    end )
end

function MultiChoiceOption:PreIcon( icon, colour )
    assert( icon )
    self:SetIcon( icon, colour )
    return self
end

function MultiChoiceOption:SetPortrait( agent )
    self.portrait = self.portrait or self.post_images:AddChild( Widget.TinyPortrait( agent ) )
    self.portrait:SendToBack()
    self.portrait:SetHeight( ICON_SIZE )
    self.portrait:SetFacing(false)
    self:SetSize( self.width )
    self.portrait.blocks_mouse = false
    self.portrait.anim.blocks_mouse = false
    self.portrait:SetToolTip( agent.guid )
    self.portrait:SetToolTipClass( Widget.TooltipAgent )   
    self:Layout()
    return self
end

function MultiChoiceOption:PostIcon( icon, w, h )
    assert( icon )
    local icon = Widget.Image( icon ):SetSize( w or ICON_SIZE, h or ICON_SIZE )
    self.post_images:AddChild( icon )
    self:Layout()
    return self
end


-----------------------------------------------------------------------------------------------
--

local MultiChoice = class("Widget.MultiChoice", Widget)

function MultiChoice:init( )
    MultiChoice._base.init(self)

    self.opt = Widget():SetName("OPT")
    self.footer = Widget():SetName("FOOTER")
    self.underfooter = Widget():SetName("UNDERFOOTER")
    self.header = Widget():SetName("HEADER")
    self.header_select_bg = Widget.Panel( engine.asset.Texture ("UI/panel_selected.9.tex") ):SetBloom(0.3):Hide()

    self:AddChildren{
        self.header_select_bg,
        self.header,
        self.opt,
        self.footer,
        self.underfooter,
    }

    -- Array of weak refs to all widgets contained within footer, header, and opts.
    self.optionwidgets = {}
    -- Array of only convo option widgets. No footer/leave options
    self.convooptionwidgets = {}

    self.controller_toggle_label = self:AddChild( Widget.Label( "title", 30 ) )
        :SetHiddenBoundingBox( true )
        :Bloom( 0.1 )
        :Hide()

    self:SetSize(600, 512+128)
    self:SetFocusLock(true)
end

function MultiChoice:ShowControllerHint()
    if TheGame:FE():IsGamepadMode() then
        self.controller_toggle_label:Show()
            :SetText( loc.format(LOC"UI.CONVERSATION_SCREEN.CONTROLLER_TOGGLE_CONVO", Controls.Digital.GAMEPAD_LB) )
            :LayoutBounds( "before", "bottom", self )
            :Offset( -12, 7 )
    else
        self.controller_toggle_label:Hide()
    end
    return self
end

function MultiChoice:HideControllerHint()
    self.controller_toggle_label:Hide()
    return self
end

function MultiChoice:SetSize(w, h)
    self.w, self.h = w, h
    self.innerw = w - 8
    self.header_select_bg:SetSize( w, 58 )
    self:LayoutOptions()
    return self
end

function MultiChoice:GetHeight()
    local w, h = self:GetSize()
    return h
end

function MultiChoice:ClearOptions()
    self.header_select_bg:Hide()
    self.header_idx = nil

    self.header:DestroyAllChildren()
    self.footer:DestroyAllChildren()
    self.underfooter:DestroyAllChildren()
    self.opt:DestroyAllChildren()

    table.clear( self.optionwidgets )
    table.clear( self.convooptionwidgets )

    return self
end

function MultiChoice:LayoutOptions()

    local padding = 12
    local last_widget = nil

    -- Layout from the bottom upwards

    if not self.underfooter:IsEmpty() then
        
        for k,w in ipairs(self.underfooter.children) do
            w:LayoutBounds("after", "center"):Offset(8, 0)
        end


        self.underfooter:LayoutBounds("center", "below", 0, 0):Offset(0, -padding)
    end

    if not self.footer:IsEmpty() then
        last_widget = self.footer:LayoutBounds( "center", "above", 0, 0)
    end

    local cur_y = 0
    if not self.opt:IsEmpty() then
        local opts = self.opt.children
        for idx = #opts, 1, -1 do
            local widget = opts[ idx ]
            local w, h = widget:GetSize()

            cur_y = cur_y + h/2

            widget:SetPos( 0, cur_y )

            cur_y = cur_y + h/2 + OPTION_SPACING
        end

        if last_widget then
            last_widget = self.opt:LayoutBounds( "center", "above", last_widget ):Offset( 0, padding )
        else
            last_widget = self.opt:LayoutBounds( "center", "above", 0, 0)
        end
    end
    
    if not self.header:IsEmpty() then
        if last_widget then
            last_widget = self.header:LayoutBounds( "center", "above", last_widget ):Offset( 0, padding )
        else
            last_widget = self.header:LayoutBounds( "center", "above", 0, 0)
        end
    end
    
end


function MultiChoice:AddOption( text, data )
    local idx = #self.optionwidgets + 1
    local widget = Widget.MultiChoiceOption( text, data, self.innerw )
    widget:SetIndex( self, idx )
    table.insert( self.optionwidgets, widget )
    table.insert( self.convooptionwidgets, widget )
    self.opt:AddChild( widget )

    return widget
end

function MultiChoice:AddUnderfooterOption( text, data )
    local idx = #self.optionwidgets + 1
    local function OnClickFooter()
        self:OnSelectOption( idx )
    end

    local widget = Widget.IconButton( text, OnClickFooter )
    widget:SetIcon( global_images.close )
    widget.label:EnableShadow( true )
    if data and data.sfx then widget:SetClickSoundEvent( data.sfx ) end
    table.insert( self.optionwidgets, widget )
    self.underfooter:AddChild( widget )

    return widget
end

function MultiChoice:GetCloseOption()
    if self.underfooter and self.underfooter:GetNumChildren() > 0 then
        return self.underfooter.children[1]
    end
end

function MultiChoice:HideCloseOption()
    self.underfooter:Hide()
end

function MultiChoice:ShowCloseOption()
    self.underfooter:Show()
end

function MultiChoice:AnimateInCoro()
    self:Disable()
    local widgets = {}

    for k,widget in ipairs(self.optionwidgets) do
        widget:Hide()
        table.insert(widgets, widget)
    end

    table.sort(widgets, function(a, b) 
            local _, y1 = a:GetWorldPos()
            local _, y2 = b:GetWorldPos()
            return y1 < y2
        end)

    local n = 0
    for _, widget in ipairs(widgets) do
        self:Delay( (.04+(n)*.015)*.333 ) 
        widget:Show()
        n = n + 1 
    end

    self:Enable()
end

function MultiChoice:AnimateIn()
    self:StartCoroutine(MultiChoice.AnimateInCoro, self)
end


function MultiChoice:SetOptions(options)
    self:ClearOptions()
    
    for k,v in ipairs(options) do
        self:AddOption(v)
    end

    self:LayoutOptions()
    return self
end

function MultiChoice:GetOption(idx)
    if self.optionwidgets and #self.optionwidgets >= idx then
        return self.optionwidgets[idx]
    end
end

function MultiChoice:GetOptions()
    return self.optionwidgets
end

function MultiChoice:GetConvoOptions()
    return self.convooptionwidgets or {}
end

function MultiChoice:GetConvoOption(idx)
    if self.convooptionwidgets and #self.convooptionwidgets >= idx then
        return self.convooptionwidgets[idx]
    end
end

function MultiChoice:OnSelectOption(idx)
    if self.onSelectOptionFn then
        self.onSelectOptionFn(idx)
    end
end

function MultiChoice:SetOnSelectOptionFn(fn)
    self.onSelectOptionFn = fn
    return self
end

function MultiChoice:HandleControlDown(control, device)
    local hotkey_idx = UIHelpers.FindHotkeyIdx( control )
    if hotkey_idx then
        if self.optionwidgets[ hotkey_idx ] and self.optionwidgets[ hotkey_idx ].HotkeyDown then
            self.optionwidgets[ hotkey_idx ]:HotkeyDown()
            return true
        end
    end
end

function MultiChoice:HandleControlUp(control, device)
    local hotkey_idx = UIHelpers.FindHotkeyIdx( control )
    if hotkey_idx then
        if self.optionwidgets[ hotkey_idx ] and self.optionwidgets[ hotkey_idx ].HotkeyUp then
            self.optionwidgets[ hotkey_idx ]:HotkeyUp()
            return true
        end
    end
end

