require "ui/widgets/cardwidget"
local negotiation_defs = require "negotiation/negotiation_defs"

-------------------------------------------------------------------------------

local CARD_FLAGS = negotiation_defs.CARD_FLAGS

local SCREEN_SHAKE_TIMING = 0.1
local CARD_DROP_TIMING = 0.1
local CARD_DROP_EASING = easing.outQuad

local assets =
{
    type_icons = 
    {
        [ CARD_TYPE.NEGOTIATION_HOSTILE ] = engine.asset.Texture( "UI/cardview_typeicon_negotiation_hostility.tex" ),
        [ CARD_TYPE.NEGOTIATION_DIPLOMACY ] = engine.asset.Texture( "UI/cardview_typeicon_negotiation_diplomacy.tex" ),
        [ CARD_TYPE.NEGOTIATION_MANIPULATE ] = engine.asset.Texture( "UI/cardview_typeicon_negotiation_manipulation.tex" ),
        [ CARD_TYPE.NEGOTIATION_STATUS ] = engine.asset.Texture( "UI/cardview_typeicon_negotiation_status.tex" ),
        [ CARD_TYPE.NEGOTIATION_OPPONENT ] = engine.asset.Texture( "UI/cardview_typeicon_opponent.tex" ),
        [ CARD_TYPE.ITEM ] = engine.asset.Texture( "UI/cardview_typeicon_item.tex" ),
        [ CARD_TYPE.PARASITE ] = engine.asset.Texture( "UI/cardview_typeicon_parasite.tex" ),
        [ CARD_TYPE.FLOURISH ] = engine.asset.Texture( "UI/cardview_typeicon_flourish.tex" ),
    },

    pip_open = engine.asset.Texture( "UI/ic_charges_empty.tex" ),
    pip_closed = engine.asset.Texture( "UI/ic_charges_full.tex" ),
}

local NEGOTIATION_SOUND_EVENTS = 
{
    SELECT_CARD = 
    {
        [CARD_TYPE.NEGOTIATION_BYSTANDER] = "event:/sfx/battle/cards/neg/card_type/bystander/select_bystander",
        [CARD_TYPE.NEGOTIATION_DIPLOMACY] = "event:/sfx/battle/cards/neg/card_type/diplomacy/select_diplomacy",
        [CARD_TYPE.NEGOTIATION_HOSTILE] =  "event:/sfx/battle/cards/neg/card_type/hostility/select_hostility",
        [CARD_TYPE.ITEM] = "event:/sfx/battle/cards/neg/card_type/item/select_item",
        [CARD_TYPE.PARASITE] = "event:/sfx/battle/cards/neg/card_type/parasite/select_parasite",
        [CARD_TYPE.FLOURISH] = "event:/sfx/battle/cards/neg/card_type/parasite/select_flourish",
        [CARD_TYPE.NEGOTIATION_MANIPULATE] = "event:/sfx/battle/cards/neg/card_type/manipulate/select_manipulate",
        [CARD_TYPE.NEGOTIATION_STATUS] = "event:/sfx/battle/cards/neg/card_type/status/select_status",
    },
    PLAY_CARD =     
    {           
        [CARD_TYPE.NEGOTIATION_BYSTANDER] = "event:/sfx/battle/cards/neg/card_type/bystander/play_bystander",
        [CARD_TYPE.NEGOTIATION_DIPLOMACY] = "event:/sfx/battle/cards/neg/card_type/diplomacy/play_diplomacy",
        [CARD_TYPE.NEGOTIATION_HOSTILE] =  "event:/sfx/battle/cards/neg/card_type/hostility/play_hostility",
        [CARD_TYPE.ITEM] = "event:/sfx/battle/cards/neg/card_type/item/play_item",
        [CARD_TYPE.PARASITE] = "event:/sfx/battle/cards/neg/card_type/parasite/play_parasite",
        [CARD_TYPE.FLOURISH] = "event:/sfx/battle/cards/neg/card_type/parasite/play_flourish",
        [CARD_TYPE.NEGOTIATION_MANIPULATE] = "event:/sfx/battle/cards/neg/card_type/manipulate/play_manipulate",
        [CARD_TYPE.NEGOTIATION_STATUS] = "event:/sfx/battle/cards/neg/card_type/status/play_status",
    }
}

-------------------------------------------------------------------------------
-- Widget wrapping a Negotiation.Card

local CardWidget = class( "Widget.NegotiationCard", Widget.CardWidget )

function CardWidget:init( card, minigame )
    CardWidget._base.init( self, assets, negotiation_defs )

    self.shadow:SetTintColour( UICOLOURS.NEGOTIATION )
    self.shadow:SetTintAlpha(0.3)

    if card then
        self:RefreshCard( card, minigame )
    end
end

function CardWidget:RefreshCard( card, minigame, style )
    assert_warning( is_instance( card, Negotiation.Card ), card.id )

    minigame = minigame or card.engine
    local card_changed = self.card ~= card

    CardWidget._base.RefreshCard( self, card, minigame, style )

    local card_back_tint, type_icon, text_colour, tt = self:RefreshCardType( card, minigame )

    if not (type_icon and text_colour and tt) then
        return
    end
    self.bg:SetTintColour( card_back_tint )

    self.type_icon:SetTexture( type_icon )
        :SetTintColour( text_colour )

    if not self.card.manual_money_desc then
        local cost = card.money_cost or 0
        if minigame then
            cost = minigame:CalculateMoneyCost( card, true )
        end
        
        if self.last_money_cost ~= cost or card_changed then
            self.last_money_cost = cost
            self:UpdateDescription(minigame, true)
            if cost > 0 then
                self:PostText( loc.format( GetLocalizedConvoCommonString("OPT_HAVE_MONEY"), cost ) )
            end
        end
    end

    return self
end

function CardWidget:RefreshCardType( card, minigame )
    local card_type = GetCardType( card )

    return
        GetCardTypeColour( card_type ),
        assets.type_icons[ card_type ],
        GetCardTypeColour( card_type ),
        GetCardTypeString( card_type )
end

function CardWidget:RefreshCostLabel( card, minigame )
    local cost, details
    if minigame then
        cost, details = minigame:CalculateActionCost( card )
    else
        cost = card.cost
    end

    local cost_txt
    if CheckAnyBits( card.flags, CARD_FLAGS.UNPLAYABLE | CARD_FLAGS.AUTOPLAY ) then
        cost_txt = "-"
    elseif CheckBits( card.flags, CARD_FLAGS.VARIABLE_COST ) then
        cost_txt = LOC("X")
    else
        cost_txt = tostring(cost)

        if minigame and cost > minigame:GetActionCount() then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetDisabledColour( card ), cost_txt )
        elseif self.style == self.STYLE.UPGRADE then
            if cost < card.base_def.cost or CheckBits( card.base_def.flags, CARD_FLAGS.UNPLAYABLE ) then
                cost_txt = loc.format( "<#UPGRADE>{1}</>", cost_txt )
            elseif cost > card.base_def.cost then
                cost_txt = loc.format( "<#DOWNGRADE>{1}</>", cost_txt )
            end
        elseif cost < card.cost then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), cost_txt )
        elseif cost > card.cost then
            cost_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), cost_txt )
        end
    end

    if details then
        details = LOC"UI.NEGOTIATION_PANEL.ACTION_COST" ..
            loc.format( LOC"UI.NEGOTIATION_PANEL.ACTION_COST_MODIFIED", details )
    else
        details = LOC"UI.NEGOTIATION_PANEL.ACTION_COST"
    end

    return cost_txt, details
end

function CardWidget:RefreshEffectLabel( card, minigame )
    if card.min_persuasion == nil or card.max_persuasion == nil then
        return
    end

    local min_persuasion, max_persuasion, details, acc
    if minigame then
        min_persuasion, max_persuasion, details, acc = minigame:PreviewPersuasion( card )
    else
        min_persuasion, max_persuasion, details, acc = card:CalculatePersuasion()
    end
    local min_txt

    if self.style == self.STYLE.UPGRADE and min_persuasion ~= card.base_def.min_persuasion then
        if card.base_def.min_persuasion and min_persuasion < card.base_def.min_persuasion then
            min_txt = loc.format( "<#DOWNGRADE>{1}</>", min_persuasion )
        else
            min_txt = loc.format( "<#UPGRADE>{1}</>", min_persuasion )
        end
    elseif min_persuasion < card.min_persuasion and details then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), min_persuasion )
    elseif min_persuasion > card.min_persuasion and details then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), min_persuasion )
    elseif min_persuasion == card.min_persuasion and acc and acc:IsMinPersuasionModified() then
        min_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), min_persuasion )
    else
        min_txt = tostring(min_persuasion)
    end

    local max_txt
    if self.style == self.STYLE.UPGRADE and max_persuasion ~= card.base_def.max_persuasion then
        if card.base_def.max_persuasion and max_persuasion < card.base_def.max_persuasion then
            max_txt = loc.format( "<#DOWNGRADE>{1}</>", max_persuasion )
        else
            max_txt = loc.format( "<#UPGRADE>{1}</>", max_persuasion )
        end
    elseif max_persuasion < card.max_persuasion and details then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), max_persuasion )
    elseif max_persuasion > card.max_persuasion and details then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetBonusColour( card ), max_persuasion )
    elseif max_persuasion == card.max_persuasion and acc and acc:IsMaxPersuasionModified() then
        max_txt = loc.format( "<{1#colour}>{2}</>", self:GetPenaltyColour( card ), max_persuasion )
    else
        max_txt = tostring(max_persuasion)
    end

    local txt
    if min_txt == max_txt then
        txt = min_txt
    else
        txt = loc.format( "{1}-{2}", min_txt, max_txt )
    end

    local tt = LOC"UI.NEGOTIATION_PANEL.STRENGTH_TOOLTIP"
    if details then
        tt = tt .. loc.format( LOC"UI.NEGOTIATION_PANEL.EFFECT_DETAILS_TOOLTIP", details )
    end

    return txt, tt, min_persuasion, max_persuasion
end

function CardWidget:AnimatePreResolve( panel, idx )
    self.anim = nil
    self:DisableToolTip( true )

    self:Enable() -- Shouldn't appear disabled once resolved.



    local ev = NEGOTIATION_SOUND_EVENTS.PLAY_CARD[GetCardType(self.card)]
    if ev then
        AUDIO:PlayEvent(ev)
    else
        AUDIO:PlayEvent(SoundEvents.negotiation_cardplayed_player)
    end
    
    
    if self.card.owner and self.card.owner.def.negotiation_card_play_sound then
        AUDIO:PlayEvent(self.card.owner.def.negotiation_card_play_sound)
    end

    -- Animate card down to the centre of the screen
    local w, h = self:GetFE():GetScreenDims()
    local ypos = h/2.2
    local xpos = w/2
    local x, y = self:GetPos()
    local tx, ty = self.parent:TransformFromWorld( xpos, ypos )

    self:ScaleTo( self.sx, 0.6, CARD_DROP_TIMING, CARD_DROP_EASING )
    self:MoveTo( tx, ty, CARD_DROP_TIMING, CARD_DROP_EASING )
    self:RotateTo( 0, CARD_DROP_TIMING, CARD_DROP_EASING )

    self:Delay( CARD_DROP_TIMING*2 )
end

function CardWidget:PlaySelectSound( card )
    card = card or self.card
    local ev = NEGOTIATION_SOUND_EVENTS.SELECT_CARD[ GetCardType( card )]
    if ev then
        AUDIO:PlayEvent(ev)
    end
    if card.owner and card.owner.def.negotiation_card_play_sound then
        AUDIO:PlayEvent( card.owner.def.negotiation_card_select_sound )
    end
end

function CardWidget:PlayDeselectSound( card )
    AUDIO:PlayEvent(SoundEvents.negotiation_card_deselect)
end

