require "ui/widget"
require "ui/widgets/screen"
require "ui/widgets/animated_grid"
require "ui/widgets/card_deck"
require "ui/widgets/timeindicator"
require "ui/widgets/cardarrangement"
require "ui/widgets/cardwidget"
require "ui/widgets/progressimagebar"
require "ui/widgets/cardspanel"
require "ui/widgets/particlesystem"
require "ui/widgets/targetarrow"
require "ui/widgets/damagearrow"
require "ui/widgets/negotiationslot"
require "ui/widgets/intentcontainer"
require "ui/widgets/negotiationcard"
require "ui/widgets/negotiationresolvebar"
require "ui/screens/choosecardspopup"
require "negotiation/negotiation_engine"
local negotiation_defs = require "negotiation/negotiation_defs"
require "content/convo_loc_common"
require "util/camshake"

-------------------------------------------------------------------------------

local CARD_FLAGS = negotiation_defs.CARD_FLAGS
local EVENT = negotiation_defs.EVENT
local STATUS = negotiation_defs.STATUS

local RESOLVE_FLOATER_DURATION = 0.3

local LOSE_RESOLVE_FLOATER =
{
    color = 0xFF7777FF,
    outline_color = 0xAA5555FF,
    shake_params = NEGOTIATION_SHAKE,
    bloom_params = { 2.0, 0.4 },
    hit_emote = { "neutral_wince", "neutral_wince2" },
}

local INCEPT_HIT_RESULT =
{
    shake_params = NEGOTIATION_SHAKE,
    bloom_params = { 2.0, 0.4 },
    hit_emote = { "neutral_wince", "neutral_wince2" },
}

local GAIN_RESOLVE_FLOATER =
{
    color = 0x77FF77FF,
    outline_color = 0x55AA55FF,
}

local COMPOSURE_FLOATER =
{
    color = UICOLOURS.COMPOSURE,
    outline_color = UICOLOURS.BLACK
}


-------------------------------------------------------------------------------
-- Persuasion floater helpers

local function AnimateFloaterToTarget( panel, w, tx, ty, sound_event, time )
    
    local inst = nil

    if tx and ty then
        w:MoveTo( tx, ty, time, easing.inQuad )
        inst = AUDIO:CreateEventInstance(sound_event)
        
        if inst then
            inst:SetParameterValue("position", w:GetPanPos())
            inst:Start()
        end

        local t = time
        local max_bloom = 0.1
        while t > 0 do
            t = t - coroutine.yield()
            local bloom = easing.inQuad( 0.5 - t, 0, max_bloom, 0.5 )
            --local theta = ( 0.5 - t ) * math.pi * 4
            w:SetBloom( bloom )
            --w:SetRotation( -theta )
            if inst then
                inst:SetParameterValue("position", w:GetPanPos())
            end
        end
    end

    if inst then
        inst:Stop()
        inst:Release()
    end
end

local function CreatePersuasionFloater( panel, delta, source_widget, params )
    local w = panel:AddChild( Widget.Label( "title", 36 ) )
    if delta < 0 then
        w:SetText( tostring(delta) )
    else
        w:SetText( delta )
    end    
    if params then
        w:SetGlyphColour( params.color ):SetOutlineColour( params.outline_color ):EnableOutline( true )
    end

    local scale = 1 + math.abs( delta ) / ( math.abs( delta ) + 10 )
    w:SetScale( 2.0 * scale * panel:GetFE():GetBaseWidgetScale() )
    w:SetPos( panel:TransformFromWidget( source_widget, 0, 0 ) )

    return w
end

local function DoResolveFloat( w )
    w:SetRotation( 0 )
    local sx = w.sx
    w:ScaleTo( sx, sx * 1.5, 0.15, easing.outQuad )
    w:MoveTo( w.x, w.y + 100, 0.8, easing.outCubic )
    w:Delay( 0.2 )
    w:ScaleTo( sx * 1.5, sx, 0.15, easing.outQuad )
    w:Delay( 0.4 )
    w:AlphaTo( 0, 0.8 ):DelayIfAnimating()
    w:Remove()
end

local function DoResolveStrike( panel, source, source_widget, target, target_widget, delta, defended, params )
    assert( is_instance( target_widget, Widget.NegotiationSlot ))
    
    local blob = CreatePersuasionFloater(panel, math.abs(delta), source_widget, params)
    local tx, ty = target_widget:TransformToWorld( 0, 0 )
    local sound_event
    if target_widget == panel.opponent_modifiers then
        sound_event = SoundEvents.negotiation_morale_number_flyin_enemy
    else
        sound_event = SoundEvents.negotiation_morale_number_flyin_player
    end

    -- Keep the label in place...
    local t = .2
    while t > 0 do
        t = t - coroutine.yield()
    end

    AnimateFloaterToTarget( panel, blob, tx, ty, sound_event, params.floater_duration or RESOLVE_FLOATER_DURATION )
    blob:Remove()

    if source.shake_params then
        panel:GetScreen():ScreenShake( card.shake_params )
    elseif params and type(params.shake_params) == "table" then
        panel:GetScreen():ScreenShake( params.shake_params )
    end

    if source.bloom_params then
        panel:GetScreen():DoBloomBurst( table.unpack( params.bloom_params ) )
    elseif params and params.bloom_params then
        panel:GetScreen():DoBloomBurst( table.unpack( params.bloom_params ) )
    end

    local emote = source.hit_emote or params.hit_emote
    if type(emote) == "table" then
        emote = table.arraypick( emote )
    end
    if type(emote) == "string" then
        local body = panel:GetScreen():GetCharacterWidget( target:GetNegotiator().agent )
        if body then
            -- Need to bypass all mood transitions in order for this to time right.
            -- body:Emote( "wince" )
            body:Play( emote )
            body:Push( body:GetTransitionAnim( MOOD.NEUTRAL, body.mood ))
            body:Push( body:GetIdleAnim(), true )
        end
    end

    if target_widget and not target_widget.removed then
        local pst_delta = delta + defended
        local w = CreatePersuasionFloater( panel, pst_delta, target_widget, params )
        panel:StartCoroutine( DoResolveFloat, w )
    end
end

-------------------------------------------------------------------------------
-- Plays a negotiation card minigame.

local EVENT_HANDLERS =
{
    [ EVENT.ATTACK_RESOLVE ] = function( self, source, target, damage, params, defended )
        params = table.inherit( LOSE_RESOLVE_FLOATER, params )
        self:ShowAttackResolve( target, -damage, source, defended, params )
    end,

    [ EVENT.DELTA_ACTION_COUNT ] =  function( self, actions, max_actions, delta, no_sound )
        self.time_indicator:StartCoroutine( self.time_indicator.AnimateNumberChange, self.time_indicator, actions, max_actions, no_sound )

        self.cards:RefreshCards( self.minigame ) -- Changes playability of cards.
    end,

    [ EVENT.DELTA_MONEY ] =  function( self, actions, delta )
        AUDIO:PlayEvent( SoundEvents.putdown_shills )
    end,

    [ EVENT.DELTA_COMPOSURE ] =  function( self, modifier, new_value, old_value, source, start_of_turn )
        local widget = self:FindSlotWidget( modifier )
        if widget then
            if source == self.resolve_card then
                self.composure_delay = 0.3
            end

            self:StartCoroutine( function()
                widget:WaitOnCoros()
                widget:RegisterCoro()

                if widget.removed then
                    print( "DELTA_COMPOSURE on a removed widget:", modifier, new_value, old_value, source )
                end

                if new_value > old_value then
                    self:ShowAttackResolve( modifier, new_value - old_value, source, 0, COMPOSURE_FLOATER )
                    if widget and not widget.removed then
                        widget:AnimateComposure( new_value, old_value, start_of_turn )
                    end
                else
                    widget:AnimateComposure( new_value, old_value, start_of_turn )
                end
            end )
        end
    end,

    [ EVENT.SHIELD_STATUS_CHANGE] = function(self, modifier, shield_status)
        local widget = self:FindSlotWidget( modifier )
        if widget then
            widget:RunAfterCoros( self, function()
                widget:SetShieldStatus(shield_status, modifier)
            end)
        end
    end,

    [ EVENT.DELTA_RESOLVE ] = function( self, modifier, resolve, max_resolve, delta, source, params )
        local widget = self:FindSlotWidget( modifier )
        if widget then
            widget:RunAfterCoros( self, function()
                if source and (params == nil or not params.no_floater) then
                    if delta > 0 then
                        params = table.inherit( GAIN_RESOLVE_FLOATER, params )
                        self:ShowAttackResolve( modifier, delta, source, 0, params  )
                    else
                        params = table.inherit( LOSE_RESOLVE_FLOATER, params )
                        self:ShowAttackResolve( modifier, delta, source, 0, params )
                    end
                end

                widget:AnimateResolve( delta, resolve, max_resolve, source )
            end )
        end
    end,

    [ EVENT.RESTORE_RESOLVE ] = function( self, modifier, resolve, max_resolve, delta, source )
        local widget = self:FindSlotWidget( modifier )
        if widget then
            if source then
                self:ShowAttackResolve( modifier, delta, source, 0, GAIN_RESOLVE_FLOATER )
            end

            widget:AnimateResolve( delta, resolve, max_resolve, source )
        end
    end,

    [ EVENT.MODIFIER_CHANGED ] = function( self, modifier, delta, clone, source )
        local container = self:GetSlotContainer( modifier )
        if container then
            if source == self.resolve_card then
                self.modifier_delay = math.max( self.modifier_delay or 0, 0.5 )
            end

            self:StartCoroutine( container.OnModifierChanged, container, clone, delta, source )
        end

        self.cards:RefreshCards( self.minigame )
        self.intents:RefreshIntents()
        self.opponent_modifiers:UpdatePersuasionLabels()
        self.player_modifiers:UpdatePersuasionLabels()
    end,

    [ EVENT.MODIFIER_ADDED ] = function( self, modifier, source, clone )
        local container = self:GetSlotContainer( modifier )
        if container then
            if source == self.resolve_card then
                self.modifier_delay = math.max( self.modifier_delay or 0, 0.5 )
            end
            self:StartCoroutine( container.OnModifierAdded, container, clone, source )

            self:SayQuip( modifier.owner, "modifier_added", modifier.base_id or modifier.id )
        end

        self.cards:RefreshCards( self.minigame )
        self.intents:RefreshIntents()
        self.opponent_modifiers:UpdatePersuasionLabels()
        self.player_modifiers:UpdatePersuasionLabels()
    end,

    [ EVENT.MODIFIER_REMOVED ] = function( self, modifier, source )
        local container = self:GetSlotContainer( modifier )
        if container then
            if source == self.resolve_card then
                self.modifier_delay = math.max( self.modifier_delay or 0, 0.3 )
            end
           self:StartCoroutine( container.OnModifierRemoved, container, modifier, source )

           if source then
               self:SayQuip( modifier.owner, "modifier_removed", modifier.id )
           end
        end

        self.cards:RefreshCards( self.minigame )
        self.intents:RefreshIntents()
        self.opponent_modifiers:UpdatePersuasionLabels()
        self.player_modifiers:UpdatePersuasionLabels()
    end,

    [ EVENT.MODIFIER_TRIGGERED ] = function( self, modifier )
        local container = self:GetSlotContainer( modifier )
        if container then
            self:RefreshDamagePreview()
            container:PingModifier( modifier )

        elseif modifier.graft then
            local widget = self.grafts:FindWidget( modifier.graft )
            if widget then
                widget:Triggered()
            end
        end
    end,

    [ EVENT.UPDATE_MODIFIER_ICON ] = function( self, modifier, icon )
        local container = self:GetSlotContainer( modifier )
        if container then
            container:RefreshIcon(modifier, icon)
        end
    end,

    [ EVENT.UPDATE_COIN ] = function( self, modifier, got_heads )
        local container = self:GetSlotContainer( modifier )
        local slot_widget = container and container:FindSlotWidget(modifier)

        if slot_widget then
            local path = got_heads and modifier.heads_icon or modifier.snails_icon       
            if path then
                modifier.icon = engine.asset.Texture( path, true )
            end
            container:PlayPositionalEvent(got_heads and "event:/sfx/battle/cards/neg/rook_coinflips_heads" or "event:/sfx/battle/cards/neg/rook_coinflips_snails")
            slot_widget:FlipCoin(modifier, got_heads and "HEADS" or "SNAILS")
            container:PingModifier(modifier)
        end
    end,

    [ EVENT.GRAFT_TRIGGERED ] = function( self, graft )
        local widget = self.grafts:FindWidget(graft) or self.boons:FindWidget(graft) or self.banes:FindWidget(graft) or self.perks:FindWidget(graft)
        if widget then
           widget:Triggered()
       end
    end,

    [ EVENT.REFRESH_GRAFTS ] = function( self )
        self.grafts:Refresh()
        self.boons:Refresh()
        self.banes:Refresh()
    end,

    [ EVENT.CARD_TRIGGERED_PRE ] = function( self, card )
        local w = self:DetachCardWidget( card )
        if w then
            self.floating_cards:AttachCardWidget( card, w )
            w:AnimateChosen( self ):SendToFront()
            w:Delay( 0.2 )
            
            -- w:ScaleTo( self.sx, self.sx * 0.8, 0.1, easing.inCubic )
            self:Delay( 0.3 )
        end
    end,

    [ EVENT.CARD_TRIGGERED_POST ] = function( self, card, deck, deck_idx )
        if deck then
            self:UpdateCardPlacement( card, deck, deck_idx, deck, deck_idx )
        end
    end,

    [ EVENT.CARD_MOVED ] = function( self, card, source_deck, source_idx, target_deck, target_idx )
        -- print( "EVENT.CARD_MOVED", card, source_deck, source_idx, target_deck, target_idx )
        self:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )

        if source_deck ~= target_deck then
            if source_deck and source_deck:GetDeckType() ~= DECK_TYPE.IN_HAND then
                self:RefreshDeck( source_deck:GetDeckType() )
            end
            if target_deck and target_deck:GetDeckType() ~= DECK_TYPE.IN_HAND then
                self:RefreshDeck( target_deck:GetDeckType() )
            end
        end
    end,

    [ EVENT.CARD_DISCARDED ] = function( self, card, minigame )
        local w = self:FindCardWidget( card )
        if w then
            w:AnimateDiscard( card )
        end
    end,

    [ EVENT.CARD_UPGRADED ] = function( self, card, old_snapshot )
        local w = self:FindCardWidget( card )
        if w then
            w:AnimateUpgraded( card, old_snapshot )
        end
    end,

    [ EVENT.CARD_CHANGED ] = function( self, card, snapshot )
        local w = self:FindCardWidget( card )
        if w then
            w:RefreshCard( snapshot, self.minigame )
            w.card = card -- FIXME: Uh, don't save the snapshot into the widget.
        end
    end,

    [ EVENT.DECK_TOO_LARGE ] = function( self, card, deck )
        self.cardspanel:ShowBanner( LOC"CARD_ENGINE.DECK_TOO_LARGE", 1.0, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
    end,

    [ EVENT.CARDS_INCEPTED ] = function( self, cards, source )
        self:AnimateInceptedCards( cards, source )
    end,

    [ EVENT.CARD_DEALT ] = function( self, card, deck )
        if card.show_dealt == false then
            return -- Force suppressed.
        elseif card.show_dealt == true then
            -- Force show!
        end

        if not self:FindCardWidget( card ) then
            self.dealt_cards:CreateCardWidget( card )
        end

        -- Dragons:
        -- If there are multiple cards dealt IN A ROW, we want them to accumulate first before dealing them out.
        -- So we must peek into the event queue to see if this is the case; if there are any interevening events,
        -- that would imply a new "dealing".
        local deal_count = 0
        for i, ev in ipairs( self.event_queue ) do
            if ev[1] == EVENT.CARD_DEALT then
                deal_count = deal_count + 1
            elseif ev[1] ~= EVENT.CARD_MOVED then
                break
            end
        end

        if deal_count == 0 then -- No more dealt cards it seems, fire these off.
            self:DealCards()
        end
    end,

    [ EVENT.CARD_STOLEN ] = function( self, card, destination )
        local w = self:DetachCardWidget( card )
        if w == nil then
            w = self.dealt_cards:CreateCardWidget( card )
        else
            self.dealt_cards:AttachCardWidget( card, w )
        end

        AUDIO:PlayEvent("event:/sfx/battle/cards/neg/appropriator_stealscard")
        self.cardspanel:WaitForClick(loc.format( LOC"UI.NEGOTIATION_PANEL.STOLEN_CARD", card.id))

        local w = self.dealt_cards:DetachCardWidget( card )
        if destination then
            local dest_w = self:FindSlotWidget( destination )
            if dest_w then
                self.floating_cards:AttachCardWidget( card, w )
                w:AnimateToDeck( dest_w )
                AUDIO:PlayParamEvent("event:/sfx/battle/cards/neg/appropriator_cardlocked", dest_w:GetPanPos())
            else
                w:Remove()
            end
        end
    end,


    [ EVENT.BEGIN_ENEMY_TURN ] = function( self, minigame )
        self.encounter:PresentAgent( self.encounter:GetPlayer(), SIDE.LEFT )
        self.encounter:PresentAgent( self.minigame:GetOpponent(), SIDE.RIGHT )

        self.intents:SetRotatingMode( false )
        self.intents:SetLayoutSlots( false )

        AUDIO:PlayEvent("event:/sfx/battle/cards/neg/enemy_turn")
        self.cardspanel:ShowBanner( LOC"UI.NEGOTIATION_PANEL.ENEMY_TURN", 0.75, nil, nil, UICOLOURS.HEALTH, true )
        self:WaitForBanner()
        self:Delay( 0.25 )
        self.cards_played = 0
    end,

    [ EVENT.END_ENEMY_TURN ] = function( self, minigame )
        if minigame:GetResult() == nil and self.cards_played == 0 then
            self:SayQuip( self.minigame:GetOpponent(), "stymied" )
        end
        self.intents:SetRotatingMode( true )
        self.intents:SetLayoutSlots( true )
    end,

    [ EVENT.INTENTS_CHANGED ] = function( self, cards )
        self.intents:OnIntentsChanged( cards )
        self:RefreshDamagePreview()
    end,

    [ EVENT.INTENT_REMOVED ] = function( self, card )
        self.intents:OnIntentRemoved( card )
        self:RefreshDamagePreview()
        self:Delay( 0.5 )
    end,

    [ EVENT.INTENT_CHANGED ] = function( self, card, clone )
        self.intents:OnIntentChanged( card, clone )
        -- self:RefreshDamagePreview()
    end,

    [ EVENT.INTENT_ADDED ] = function( self, new_card, source )
        self.intents:OnIntentAdded( new_card, source )
        self:RefreshDamagePreview()
    end,

    [ EVENT.FLOURISH_CHANGED ] = function( self, graft, flourish, prev_flourish )
        if flourish >= (graft.userdata.negotiation_max or math.huge) then
            local body = self:GetScreen():GetCharacterWidget( self.minigame:GetPlayer() )
            if body then
                local anim  = UIHelpers.PlayAnimOnce( body, "anim/UI_flourish_fx.zip", "flourish" )
                if anim then
                    local r, g, b, a = HexColour( 0xFC4F4FFF )
                    body:PulseColour( 1, 1, 1, 1, r, g, b, 1, 0.4 )

                    anim:SetScale( 0.7 )
                    local wx, wy, wz = body.anim.model:SymbolToWorld( "head_swap", 0, 0, 0 )
                    wx, wy = body:TransformFromWorld( wx, wy )
                    anim:Offset( wx, wy ):Offset( 0, -100 ):SendToBack():SetBloom( 0.25 )
                    anim:SetAnimDoneFn( function()
                        body:StopPulse()
                        body:SetTintColour( 1, 1, 1, 1 )
                    end )
                end

                body:Play( "flourish" )
                if graft.owner.negotiation_flourish_gain_sound then
                    body:PlayPositionalEvent(graft.owner.negotiation_flourish_gain_sound)
                end

                body:Push( body:GetIdleAnim(), true )
            end
        end
    end,

    [ EVENT.BEGIN_PLAYER_TURN ] = function( self )
        AUDIO:PlayEvent("event:/sfx/battle/cards/neg/player_turn")
        self.encounter:PresentAgent( self.encounter:GetPlayer(), SIDE.LEFT )
        self.encounter:PresentAgent( self.minigame:GetOpponent(), SIDE.RIGHT )
        self:GetScreen():ClearHistory()

        -- because action_count gets assigned raw at beginning of turn.
        AUDIO:PlayEvent(SoundEvents.negotiation_deck_appear)

        self.cardspanel:ShowBanner( LOC"UI.NEGOTIATION_PANEL.PLAYER_TURN", 0.75, nil, UICOLOURS.SUBTITLE, nil )

        self.intents:RefreshIntents()

        self.end_turn_btn:SetText( LOC"UI.CARDS.END_TURN", loc.format( LOC"UI.CARDS_PANEL.TURNS_COUNTER", self.minigame:GetTurns() ) )
        self:DoTaskInTime( "delay_show_buttons", 0.7, function()
            self.end_turn_btn:AnimateIn()
            self.end_turn_btn:HideFocusGlow()
            self.flourish_bar:AnimateIn()
            self.concede_btn:AnimateIn()
            self.concede_btn:Enable()
        end )
        self.cards_played = 0
        if TheGame:GetGameState():GetPlayerAgent() then

            UIHelpers.ShowTutorials(
                    {"negotiation_tutorial", TheGame:GetGameState():GetPlayerAgent().def.negotiation_tutorial},
                    TheGame:GetGameState():GetPlayerAgent():GetContentID() == "SAL" and LOC"UI.NEGOTIATION_PANEL.TUTORIAL_TITLE" or loc.format( LOC"UI.NEGOTIATION_PANEL.CHARACTER_TUTORIAL_TITLE", TheGame:GetGameState():GetPlayerAgent():GetName()), 
                    TheGame:GetGameState():GetPlayerAgent():GetContentID() == "SAL" and LOC"UI.NEGOTIATION_PANEL.TUTORIAL_BODY" or loc.format( LOC"UI.NEGOTIATION_PANEL.CHARACTER_TUTORIAL_BODY", TheGame:GetGameState():GetPlayerAgent():GetName()))
        end
    end,

    [ EVENT.END_PLAYER_TURN ] = function( self )
        self.end_turn_btn:SetText( LOC"UI.CARDS.END_TURN" )
        self.concede_btn:Hide()
        self.end_turn_btn:Hide()
    end,

    [ EVENT.PRE_RESOLVE ] = function( self, minigame, card, targets, resolve_delay )

        self:SetCardFocus()

        if resolve_delay then
            self:Delay( resolve_delay )
        end

        if card.fx then
            local slot = self.opponent_modifiers:FindSlotWidget( card )
            if slot then
                UIHelpers.PlayAnimOnce( slot, card.fx.animfile, card.fx.anim )
            end
        end

        if card.sound then
            local cardwidget = self:FindCardWidget(card)
            if cardwidget then
                cardwidget:PlayPositionalEvent( card.sound )
            else
                AUDIO:PlayEvent( card.sound )
            end
            
        end


        if card:GetNegotiator() == self.minigame:GetOpponentNegotiator() then
            self.intents:ConsumeIntent( card )
        end

        self:SayCardQuip( card )

        if self.cards_played then
            self.cards_played = self.cards_played + 1
        end

        self.resolve_card = card
    end,

    [ EVENT.POST_RESOLVE ] = function ( self, minigame, card )
        
        minigame.current_quip_db = card.def.quips

        self.resolve_card = nil

        if card:GetNegotiator():IsPlayer() then
            local delay = (self.modifier_delay or 0) + (self.composure_delay or 0)
            if delay > 0 then
                self:Delay( delay )
                self.modifier_delay, self.composure_delay = nil, nil
            end
        end

        minigame.current_quip_db = nil

        self:RefreshDamagePreview()

        if self.minigame:CanPlayAny() then
            self.end_turn_btn:HideFocusGlow()
        elseif not TheGame:FE():IsGamepadMode() then
            if not self.end_turn_btn:IsGlowing() then
                AUDIO:PlayEvent("event:/sfx/battle/cards/gen/out_of_actions")
            end

            self.end_turn_btn:ShowFocusGlow()
        end

    end,

    [ EVENT.SHUFFLE_DISCARDS ] = function( self, num_cards )
        self.cardspanel:AnimateShuffle( num_cards )
    end,

    [ EVENT.CARDS_SWAPPED ] = function( self, minigame, card, idx )
        self.cards:RefreshCards( self.minigame )
    end,

    [ EVENT.QUIP ] = function( self, ... )
        self:SayQuip( ... )
    end,

    [ EVENT.DIALOG ] = function( self, str )
        self:Dialog( str )
    end,

    [ EVENT.CUSTOM ] = function( self, fn, ... )
        fn( self, ... )
    end,
}

------------------------------------------------------------------------------
-- Main parent for negotiation UI

local NegotiationPanel = class("Widget.NegotiationPanel", Widget )
NegotiationPanel.can_reload = true

NegotiationPanel.CONTROL_MAP =
{
    {
        test = function(self)
            return not self.show_reticule
        end,
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },
    {
        -- For the Widget.ChooseCardsFromHand
        controls = { Controls.Digital.MENU_ACCEPT, Controls.Digital.MENU_CANCEL },
        test = function(self)
            return self.cardspanel:IsChoosingCards() and not self.cardspanel:GetChoosingCardsPanel():IsVisible()
        end,
        fn = function(self)
            self.cardspanel:GetChoosingCardsPanel():RestorePopup()
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self)
            self.cardspanel.wait_for_click = false
            TheGame:FE():ClearFocusWidget() -- prevents the dealt card from staying focused
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.CONTINUE", Controls.Digital.MENU_ACCEPT) )
        end,
    },
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return not self.show_reticule and self.minigame:GetStatus() ~= STATUS.CHOOSE_CARDS
        end,
        fn = function(self)
            if self.hover_target and self:GetChosenCard() then
                self:OnTargetClicked(self.hover_target)
            else
                self:SetCardFocus( true )
            end
            return true
        end,
        hint = function(self, left, right)
            if self.hover_target or self:GetChosenCard() then
                table.insert(right, loc.format(LOC"UI.CONTROLS.PLAY_CARD", Controls.Digital.MENU_ACCEPT) )
            end
        end,
    },
    {
        control = Controls.Digital.CARDS_CHOOSE_FLOURISH,
        test = function(self)
            if self.minigame:GetStatus() == negotiation_defs.STATUS.PLAY_CARD and not self.show_reticule then
                return self.flourish_bar:IsShown() and self.end_turn_btn:IsShown() and self.end_turn_btn:IsEnabled() and not self.argument_focus and not self.cardspanel:IsChoosingCards()
            end
        end,
        fn = function(self)
            self:ActivateFlourish()
            return true
        end,
        hint = function(self, left, right)
            if self.flourish_bar:IsFull() then
                table.insert(right, loc.format(LOC"UI.CONTROLS.PLAY_FLOURISHES", Controls.Digital.CARDS_CHOOSE_FLOURISH) )
            else
                table.insert(right, loc.format(LOC"UI.CONTROLS.VIEW_FLOURISHES", Controls.Digital.CARDS_CHOOSE_FLOURISH) )
            end
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            return self:GetChosenCard() and not self.cardspanel:IsWaitingForClick() and not self.show_reticule
        end,
        fn = function(self)
            return self:CancelChosenCard()
        end,
        hint = function(self, left, right)
            if self:GetChosenCard() then
                table.insert(right, loc.format(LOC"UI.CONTROLS.CANCEL", Controls.Digital.MENU_CANCEL) )
            end
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            if not self.choosing_cards_from_hand and not self.cardspanel:IsWaitingForClick() then
                return self.cardspanel:IsCardSelected()
            end
        end,
        fn = function(self)
            TheGame:FE():ClearFocusWidget()
            return true
        end,
    },
    {
        control = Controls.Digital.MENU_CANCEL,
        test = function(self)
            return self.cardspanel:IsChoosingCards() and self.cardspanel:ChooseCards():GetCardCount() > 0
        end,
        fn = function(self)
            local cards = self.cardspanel:ChooseCards()
            local last_card = cards:GetCardAt( cards:GetCardCount() )
            cards:TryChooseCard( last_card )
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format ( LOC"UI.CONTROLS.UNSELECT", Controls.Digital.MENU_CANCEL ) )
        end,
    },
    {
        control = Controls.Digital.MENU_ALT,
        test = function(self)
            return ( self:GetChosenCard() or self.cardspanel:IsCardSelected() ) and not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self)
            return self:CancelChosenCard()
        end,
    },
    {
        control = Controls.Digital.CARDS_END_TURN,
        test = function(self)
            if not self.show_reticule then
                if self:GetFE():FindScreen( Screen.ChooseCardsPopup ) ~= nil then
                    return false
                end
                return self.end_turn_btn:IsShown() and self.end_turn_btn:IsEnabled() and not self.cardspanel:IsChoosingCards() and not self.cardspanel:IsWaitingForClick() and not self:GetChosenCard()
            end
        end,
        fn = function(self)
            -- Unselect selected card
            local card = self:GetChosenCard()
            if card then
                self.cards:ChooseCard( card )
                card:AssignTarget( nil )
                AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
            end
            if self.cardspanel:IsCardSelected() then
                TheGame:FE():ClearFocusWidget()
            end
            -- End turn
            self.end_turn_btn:Click()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEGOTIATION_PANEL.CONTROLS_RIGHT_ENDTURN", Controls.Digital.CARDS_END_TURN, self.minigame:GetTurns() ) )
        end,
    },
    {
        control = Controls.Digital.CARDS_RUN,
        test = function(self)
            if not self.show_reticule then
                if self:GetFE():FindScreen( Screen.ChooseCardsPopup ) ~= nil then
                    return false
                end
                return self.concede_btn:IsShown() and self.concede_btn:IsEnabled() and not self.cardspanel:IsChoosingCards() and not self.cardspanel:IsWaitingForClick() and not self:GetChosenCard()
            end
        end,
        fn = function(self) 
            self:OnClickConcede()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEGOTIATION_PANEL.CONTROLS_RIGHT_CONCEDE", Controls.Digital.CARDS_CONCEDE ) )
        end,
    },
    {
        controls = { Controls.Digital.MENU_CANCEL, Controls.Digital.MENU_UP, Controls.Digital.MENU_RIGHT, Controls.Digital.MENU_DOWN, Controls.Digital.MENU_LEFT },
        test = function(self)
            return self.show_reticule == true
        end,
        fn = function(self)
            self:SetReticuleMode( false )
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT_RELEASE", Controls.Digital.MENU_CANCEL ) )
        end,
    },
    {
        test = function(self)
            return not self.cardspanel:IsChoosingCards()
        end,
        hint = function(self, left, right)
            if self.show_reticule then
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_MOVE_INSPECT" )
            else
                table.insert(left, LOC"UI.NEGOTIATION_PANEL.CONTROLS_LEFT_INSPECT" )
            end
        end,
    },

    {
        control = Controls.Digital.VIEW_DRAW_DECK,
        test = function(self)
            return not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self) 
            self.draw_deck:ViewDeck()
            return true
        end,
    },
    {
        control = Controls.Digital.VIEW_DISCARDS_DECK,
        test = function(self)
            return not self.cardspanel:IsWaitingForClick()
        end,
        fn = function(self) 
            self.discard_deck:ViewDeck()
            return true
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_X,
        test = function(self)
            return self.cardspanel:IsChoosingCards() and self.cardspanel:GetChoosingCardsPanel():HasConfirmButton()
        end,
        fn = function(self)
            self.cardspanel:GetChoosingCardsPanel():Confirm()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format ( LOC"UI.CONTROLS.ACCEPT_CARDS", Controls.Digital.GAMEPAD_X ) )
        end,
    },
    {
        control = Controls.Digital.PAUSE, 
        fn = function(self)
            self:GetFE():InsertScreen(Screen.PauseMenu())
            return true
        end,
    },
    {
        control = Controls.Digital.GAMEPAD_Y,
        test = function(self)
            return self:GetFE():FindScreen( Screen.ChooseCardsPopup ) ~= nil
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format( LOC"UI.CONTROLS.VIEW_CHOICES", Controls.Digital.GAMEPAD_Y))
        end,
    }
}

local WHEEL_DIAMETER = 600
local INTENTS_DIAMETER = 360

function NegotiationPanel:init( minigame, encounter )
    assert( minigame )
    NegotiationPanel._base.init(self)

    self.event_queue = {}

    self.encounter = encounter
    self.minigame = minigame
    minigame:ListenForAny( self, self.OnNegotiationEvent, -1 )

    local function OnCardDraggingChanged( card, old_card )
        if self.minigame:GetStatus() == STATUS.PLAY_CARD then
            if not self.show_reticule then
                self.cards:DisablePreview( card ~= nil )
            end
        end
        self:RefreshChosenCard()
    end

    self.main_overlay = TheGame:FE():FindScreen( Screen.MainOverlay )
    if self.main_overlay == nil then
        self.main_overlay = TheGame:FE():PushScreen( Screen.MainOverlay() )
    end

    self.grafts = self.main_overlay:GetNegotiationGraftsWidget()
    self.boons = self.main_overlay:GetBoonGraftsWidget()
    self.banes = self.main_overlay:GetBaneGraftsWidget()
    self.perks = self.main_overlay:GetPerksGraftsWidget()

    self.damage_arrows_root = self:AddChild( Widget() )

    -- Modifiers
    ---------------------------------------

    self.modifier_wheels_container = self:AddChild( Widget() )
    self.modifier_wheels_container.GetDefaultFocus = function()
        return self.opponent_modifiers:GetDefaultFocus()
            or self.player_modifiers:GetDefaultFocus()
            or self.intents:GetDefaultFocus()
    end

    local player_negotiator = minigame:GetPlayerNegotiator()
    self.player_modifiers = self.modifier_wheels_container:AddChild( Widget.SlotContainer( player_negotiator, WHEEL_DIAMETER ))
        :SetPos( 0, 0 )
        :Layout()
    -- self.player_modifiers.slot_wheel:SetToolTip( LOC"UI.NEGOTIATION_PANEL.PLAYER_WHEEL_TT" )

    local opponent_negotiator = minigame:GetOpponentNegotiator()
    self.opponent_modifiers = self.modifier_wheels_container:AddChild( Widget.SlotContainer( opponent_negotiator, WHEEL_DIAMETER ))
        :SetPos( 0, 0 )
        :Layout()
    -- self.opponent_modifiers.slot_wheel:SetToolTip( loc.format( LOC"UI.NEGOTIATION_PANEL.ARGUMENT_WHEEL_TT", opponent_negotiator.agent:LocTable() ) )

    self.intents = self.modifier_wheels_container:AddChild( Widget.IntentContainer( opponent_negotiator, INTENTS_DIAMETER ))
        :SetPos( 0, 0 )
    -- self.intents.slot_wheel:SetToolTip( loc.format( LOC"UI.NEGOTIATION_PANEL.INTENT_WHEEL_TT", opponent_negotiator.agent:LocTable() ) )
    
    self.slot_container =
    {
        [ player_negotiator ] = self.player_modifiers,
        [ opponent_negotiator ] = self.opponent_modifiers,
    }

    self.target_arrow = self:AddChild( Widget.TargetArrow() )
    self.damage_arrows = {}

    -- Dark overlay
    self.darken_overlay = Widget.SolidBox( RES_X, RES_Y, UICOLOURS.BLACK ):BlockMouse():SetAnchors( "fill", "fill" )
    self:AddChild( self.darken_overlay ):SetTintAlpha( 0 ):Hide()

    -- Cards Panel
    ----------------------------------
    self.cardspanel = self:AddChild( Widget.CardsPanel( Widget.NegotiationCard, self.minigame ))
    self.time_indicator = self.cardspanel:ActionsWidget()
    self.flourish_bar = self.cardspanel:FlourishBar():SetNegotiation( TheGame and TheGame:GetGameState() and TheGame:GetGameState():GetPlayerAgent(), self.minigame )
    self.end_turn_btn = self.cardspanel:EndTurnButton( function() self:OnClickEndTurn() end, loc.format( LOC"UI.NEGOTIATION_PANEL.END_TURN", Controls.Digital.CARDS_END_TURN ))
    self.concede_btn = self.cardspanel:ConcedeButton( function() self:OnClickConcede() end, LOC"UI.CARDS_PANEL.CONCEDE_BTN", LOC"UI.NEGOTIATION_PANEL.CONCEDE"):SetToolTipClass( Widget.NegotiationTooltip )    
    

    self.draw_deck = self.cardspanel:DrawDeckButton( self.minigame:GetDrawDeck() )
    self.discard_deck = self.cardspanel:DiscardDeckButton( self.minigame:GetDiscardDeck() )
    self.trash_deck = self.cardspanel:TrashDeckButton( self.minigame:GetTrashDeck() )

    local function OnCardChoiceChanged( card, is_chosen )
        if self.minigame:GetStatus() == STATUS.PLAY_CARD then
            if not self.show_reticule then
                self.cards:DisablePreview( is_chosen )
            end
        end

        if is_chosen then
            self:SetArgumentFocus()
            self:RefreshChosenCard()

            if self:AssignAutoTarget( card ) then
                self:OnTargetClicked( card:GetTarget() )
            end
        else
            self:GetFE():ClearFocusWidget()
            self:SetCardFocus()
            card:AssignTarget( nil )
            local w = self:FindCardWidget( card )
            if w then
                w:RefreshCard( card, self.minigame )
            end
            self:UpdateTargetFX( nil )
        end
    end

    local function OnTryChooseCard( card )
        local ok, reason = self.minigame:CanPlayCard( card )
        if not ok then
            if reason == CARD_PLAY_REASONS.NO_ACTIONS then
                AUDIO:PlayEvent(SoundEvents.card_not_enough_actions)
            else
                AUDIO:PlayEvent(SoundEvents.card_invalid)
            end

            if reason then
                reason = GetLocalizedConvoCommonString( reason )

                self:NoActions( reason )
            end

            return true -- sink it
        end

        if not self:AssignAutoTarget( card ) then
            self.cards:ChooseCard( card )
            return true
        end
    end

    self.cards = self.cardspanel:Cards()
    self.floating_cards = self.cardspanel:FloatingCards()
    self.cards:SetOnChoose( OnCardChoiceChanged, OnTryChooseCard )
    self.cards:SetOnDrag( OnCardDraggingChanged )

    self.dealt_cards = self.cardspanel:DealtCards()
    self.choose_cards = self.cardspanel:ChooseCards()

    self:RefreshReason()
end

function NegotiationPanel:OnClickEndTurn()
    if self.minigame:GetStatus() == STATUS.PLAY_CARD and not self:GetChosenCard() then
        if self.minigame:CanPlayAny( true ) and engine.inst:GetSetting("gameplay", "confirm_end_turn", true) then
            local popup = Screen.YesNoPopup( LOC"UI.END_TURN_CONFIRM_BTN", LOC"UI.END_TURN_CONFIRM_BODY", LOC"UI.END_TURN_CONFIRM_SKIP" ):
            SetFn( function( result )
                if result == Screen.YesNoPopup.YES or result == Screen.YesNoPopup.NO then
                    -- End the turn
                    self.end_turn = true
                    AUDIO:PlayEvent(SoundEvents.card_end_turn)
                    -- Also save this preference for next time
                    if result == Screen.YesNoPopup.NO then
                        -- Tell the player that
                        UIHelpers.InfoPopup( LOC"UI.END_TURN_CONFIRM_INFO_TITLE", LOC"UI.END_TURN_CONFIRM_INFO_BODY" )
                        -- Actually save it to the settings
                        engine.inst:SetSetting("gameplay", "confirm_end_turn", false)
                    end
                end
            end )
            self:GetFE():PushScreen( popup )
        else
            self.end_turn = true
            AUDIO:PlayEvent(SoundEvents.card_end_turn)
        end
    end
end

function NegotiationPanel:OnClickConcede()
    local popup = Screen.YesNoPopup(LOC"UI.NEGOTIATION_PANEL.CONCEDE_CONFIRM_BTN", LOC"UI.NEGOTIATION_PANEL.CONCEDE_CONFIRM_BTN_TOOLTIP" ):
    SetFn( function( result )
        if result == Screen.YesNoPopup.YES then
            AUDIO:PlayEvent(SoundEvents.negotiation_concede)
            self.minigame:Lose()
            self.end_turn = true
            table.insert( self.minigame.force_take_cards, "unconvincing" )
        end
    end )
    self:GetFE():InsertScreen( popup )
end

function NegotiationPanel:RefreshReason()
    local title = ""
    local desc = self.minigame.start_params.reason_fn and self.minigame.start_params.reason_fn(self.minigame) or self.minigame.start_params.reason_txt or ""

    desc = CardEngine.ExtractKeywords( desc, nil, self.minigame, negotiation_defs )
    self.cardspanel:SetMinigameObjective( title, desc )
end

function NegotiationPanel:OnControlModeChange( cm, device_id )
    NegotiationPanel._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD or cm == CONTROL_MODE.TOUCH then
        self:SetReticuleMode( false )
    end

    self:RefreshDecks()
end

function NegotiationPanel:OnScreenModeChange( sm )
    NegotiationPanel._base.OnScreenModeChange( self, sm )
    
    -- Make the arguments and intents wheels smaller on small screens, so that their
    -- arguments don't overlap other widgets on screen (top bar, for instance)
    local wheel_scale = 1
    local intents_scale = 1
    if sm == SCREEN_MODE.MONITOR then
        wheel_scale = 1
        intents_scale = 1
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        wheel_scale = 0.9
        intents_scale = 0.76
    end

    self.player_modifiers.slot_wheel:SetSize( WHEEL_DIAMETER*wheel_scale )
    self.opponent_modifiers.slot_wheel:SetSize( WHEEL_DIAMETER*wheel_scale )
    self.intents.slot_wheel:SetSize( INTENTS_DIAMETER*intents_scale )
end



function NegotiationPanel:GetControlMap()
    return nil
end

function NegotiationPanel:WantsReticuleMode()
    return self.show_reticule == true
end

function NegotiationPanel:SetReticuleMode( show_reticule )
    if show_reticule ~= self.show_reticule then
        print( "SetReticuleMode", show_reticule )
        self.show_reticule = show_reticule
        -- self.cardspanel:RefreshCardsInHandPos( false )

        if show_reticule then

            self.cards:DisablePreview( true )
            self.cards:SetCardPreview( nil )

            if self.fe:GetFocusWidget() then
                local x, y = self.fe:GetFocusWidget():GetWorldPos()
                self.fe:SetReticulePos( x, y )
            end
        else
            self.cards:DisablePreview( false )
            local focus_widget = self.hover_target and self:FindSlotWidget( self.hover_target )
            if focus_widget then
                focus_widget:SetFocus()
            else
                self.fe:ClearFocusWidget()
            end
        end

        self.player_modifiers.show_child_tooltip_on_focus = show_reticule
        self.opponent_modifiers.show_child_tooltip_on_focus = show_reticule
        self.intents.show_child_tooltip_on_focus = show_reticule
   end
end

function NegotiationPanel:GetDefaultFocus()
    if self.argument_focus then
        local argument = self.modifier_wheels_container:GetDefaultFocus()
        if argument then return argument end
    end
    return self.cardspanel:GetDefaultFocus()
end

function NegotiationPanel:SetArgumentFocus()

    self.argument_focus = true
    self.card_focus = false

    -- Make arguments focusable
    self.modifier_wheels_container:SetFocusLock( true )
    self.player_modifiers:SetFocusable( true )
    self.opponent_modifiers:SetFocusable( true )
    self.intents:SetFocusable( true )

    -- Make cards un-focusable
    self.cardspanel:SetFocusable( false )
    TheGame:FE():ClearFocusWidget()

    if self.cards:IsGamepadMode() then self:FocusOnValidArgument() end
end

function NegotiationPanel:SetCardFocus( force_default )

    self.argument_focus = false
    self.card_focus = true

    -- Make arguments un-focusable
    self.modifier_wheels_container:SetFocusLock( false )
    self.player_modifiers:SetFocusable( false )
    self.opponent_modifiers:SetFocusable( false )
    self.intents:SetFocusable( false )

    -- Make cards focusable
    self.cardspanel:SetFocusable( true )
    if not self.show_reticule then
        self.cards:DisablePreview( false )
    end
    TheGame:FE():ClearFocusWidget()

    if self.cards:IsGamepadMode() and self.minigame:GetStatus() == negotiation_defs.STATUS.PLAY_CARD then
        -- Don't automatically focus next card unless in Monitor mode, they are too occluding.
        if force_default or self.fe:GetScreenMode() == SCREEN_MODE.MONITOR then
            local widget = self.cards:GetDefaultFocus()
            if widget then
                widget:SetFocus()
            end
        end
    end
end

function NegotiationPanel:FocusOnValidArgument()
    -- Go through the arguments and focus on the first valid one
    -- Used on gamepad mode only
    local card = self:GetChosenCard()
    if card and self.minigame:GetStatus() == STATUS.PLAY_CARD then
        local targets = ObtainWorkTable()
        if card then
            self.minigame:CollectAllTargets( card, targets )
        end
        -- Check if any of the targets is a Core argument
        for i, target in ipairs( targets ) do
            local slot = self:FindSlotWidget( target )
            local intent = self.intents:FindIntentWidget( target )
            local w = slot or intent
            if w and w.HighlightTarget and w.modifier and w.modifier.modifier_type == MODIFIER_TYPE.CORE then
                w:SetFocus()
                ReleaseWorkTable( targets )
                return
            end
        end
        -- Try highlighting something else
        for i, target in ipairs( targets ) do
            local slot = self:FindSlotWidget( target )
            local intent = self.intents:FindIntentWidget( target )
            local w = slot or intent
            if w and w.HighlightTarget then
                w:SetFocus()
                ReleaseWorkTable( targets )
                return
            end
        end

        ReleaseWorkTable( targets )
    end
end

function NegotiationPanel:ToggleHUD( visible, hud_flags )
    self.cardspanel:SetShown( visible or CheckBits( hud_flags, HUD_FLAGS.SHOW_CARDS ))
    self.player_modifiers:ToggleHUD( visible, hud_flags )
    self.opponent_modifiers:ToggleHUD( visible, hud_flags )
end

function NegotiationPanel:RefreshCardSpeed()
    local card_speed = engine.inst:GetSetting( "gameplay", "negotiation_speed", 1.0 )
    card_speed = math.min( 4.0, card_speed * (self.speedup_factor or 1.0))
    self.cards:SetUpdateSpeed( 1 / card_speed )
    self:SetUpdateSpeed( card_speed )
end

function NegotiationPanel:OnOptionsChanged()
    self:RefreshCardSpeed()
end

function NegotiationPanel:OnAdded()

    self:RefreshCardSpeed()
    TheGame:GetEvents():ListenForEvent( "options_changed", self, self.OnOptionsChanged )

    self:StartUpdating()
    self:StartCoroutine( function()
        self.cardspanel:Setup()
        -- self.concede_btn:Show()
        self.cardspanel:AnimateIn()
        self.cardspanel:DelayIfAnimating()
        self:RefreshTimeLabel()
        self:ResumePlay( self.minigame )
    end )

    self:RefreshDecks()

    self.player_modifiers:RefreshForNegotiator( self.minigame:GetPlayerNegotiator() )
    self.opponent_modifiers:RefreshForNegotiator( self.minigame:GetOpponentNegotiator() )

    -- Get notified as the characters move around the screen
    local character_stage = self:GetScreen():GetCharacterStage()
    character_stage:SetCharacterMovingFn( 
        function( moving, agent )
            if character_stage:IsCharacterLeft( agent ) then
                self.player_modifiers.is_character_moving = moving
            end
            if character_stage:IsCharacterRight( agent ) then
                self.opponent_modifiers.is_character_moving = moving
            end
        end
    )

    engine.sound:SetListenerPosUpFwd(0, 1, 5, 0, 1, 0,  0, 0, -1)

    if TheGame:GetDebug() then
        self.viz_break = TheGame:GetDebugEnv().viz_break
    end

    self.anim_pacing = 0.3

    local low_stakes = true

    self:SetMusicPhase(1)
    
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )

    self:SetCardFocus()
end

function NegotiationPanel:OnRemoved()
    self.minigame:RemoveListener( self )
    TheGame:GetEvents():RemoveListener( self )
end

function NegotiationPanel:HandleControlUp(control, device)

    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD or TheGame:FE():GetControlMode() == CONTROL_MODE.TOUCH then        
            if not self.hover_target then
                -- Releasing drag, cancel out.
                local card = self.cards:GetChosenCard()
                if card then 
                    -- Unselect selected card
                    self.cards:ChooseCard( card )
                    card:AssignTarget( nil )
                    AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
                    return true
                end
                if self.cardspanel:IsCardSelected() then
                    TheGame:FE():ClearFocusWidget()
                    return true
                end
            end

            local popup = self:GetFE():FindScreen( Screen.ChooseCardsPopup )
            if popup and popup:IsOverlay() then
                popup:TogglePopup()
                return true
            end
        end        
    end
end

function NegotiationPanel:HandleControlDown(control, device)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
            if self.hover_target then
                return true
            end
        end
    end

    local hotkey_idx = UIHelpers.FindHotkeyIdx( control )
    if hotkey_idx then
        local card = self.cardspanel:GetCardAt( hotkey_idx )
        if card and self.cardspanel:TryChooseCard( card ) then
            return true
        end
    end

    return false
end

function NegotiationPanel:OnNegotiationEvent( eventname, ... )
    if type(eventname) == "string" and eventname:find( "CALC_" ) then
        -- Don't even queue.
    else
        table.insert( self.event_queue, { eventname, ... } )
    end
end

function NegotiationPanel:PumpAllEvents( evs )
    while true do
        local ev = not self.viz_break and table.remove( evs, 1 )
        if ev then
            if self.last_ev_time == nil then
                self.last_ev_time = engine.inst:QueryPreciseTime()
            end
            self.last_ev_count = (self.last_ev_count or 0) + 1

            local event_name = ev[1]
            local st = os.clock()
            if EVENT_HANDLERS[ event_name ] then
                --print( "Serial processing", table.unpack( ev ) )
                self.current_event = ev
                EVENT_HANDLERS[ event_name ]( self, table.unpack( ev, 2, table.maxn(ev) ))
                self.current_event = nil
                --print( "\ttook ", (os.clock() - st) * 1000 )
            end

            if self.viz_break ~= nil then
                self.viz_break = true -- Step.
                print( event_name, " took ", (os.clock() - st), " seconds" )
            end
        else
            if self.last_ev_time then
                local now =  engine.inst:QueryPreciseTime()
                -- print( "Took", (now - self.last_ev_time) * 1000, " ms for last batch of events", self.last_ev_count )
                self.last_ev_count = nil
                self.last_ev_time = nil
            end
            coroutine.yield()
        end
        if evs ~= self.event_queue and #evs == 0 then
            break
        end
    end
end

function NegotiationPanel:SkipAllViz()
    self.viz_coro:Stop()
    self.viz_coro = nil
    table.clear( self.event_queue )
    self.current_event = nil
end

function NegotiationPanel:CountVizEvent( event_name )
    local count = 0
    for i, ev in ipairs( self.event_queue ) do
        if ev[1] == event_name then
            count = count + 1
        end
    end
    return count
end

function NegotiationPanel:IsDone()
    if self.minigame:GetResult() ~= nil then
        if self.wait_coro == nil or self.wait_coro:IsDone() then
            for i, coro in ipairs( self.coros ) do
                if coro ~= self.viz_coro then
                    return false
                end
            end
            return true
        end
    end
    return false
end

function NegotiationPanel:AssignAutoTarget( card )
    if not card:NeedsTarget() then
        card:AssignTarget( nil )
        return true
    else
        local auto_target = card.auto_target
        local current_mod = self.minigame.target_mod_accumulator:CalculateValue( card.target_mod, card, card:GetTarget())
        auto_target = auto_target or (current_mod == TARGET_MOD.RANDOM1 or current_mod == TARGET_MOD.RANDOM2 or current_mod == TARGET_MOD.RANDOM3)
        auto_target = auto_target or (current_mod == TARGET_MOD.TEAM)

        return auto_target
    end
end

function NegotiationPanel:OnTargetClicked( target )
    if TheGame:GetDebug() and TheGame:GetInput():IsModifierControl() then
        DBG( target )
        return
    end

    if self.minigame:GetStatus() == STATUS.PLAY_CARD then
        local card = self:GetChosenCard()
        if card then
            local can_target, reason = self.minigame:CanPlayCard( card, card:GetTarget() or target )
            if can_target then
                AUDIO:PlayEvent( SoundEvents.battle_card_commit )
                self.play_card = card
            else
                AUDIO:PlayEvent( SoundEvents.battle_gen_click_invalid )
            
                if reason then
                    reason = GetLocalizedConvoCommonString( reason )

                    self:NoActions( reason )
                end
            end
        end
    end
end


function NegotiationPanel:SetHoverTarget( target )
    if target == self.hover_target then
        return
    end
    if self.fe:ReticuleActive() then
        -- Target is locked in place while inspecting with the reticule.
        return
    end

    assert( target == nil or
        is_instance( target, Negotiation.Modifier ) or
        is_instance( target, Negotiation.Card ) or
        is_instance( target, Negotiation.Negotiator ))

    self.hover_target = target

    self:RefreshChosenCard()
end

function NegotiationPanel:CancelChosenCard()

    if self.chosen_card then
        local w = self:FindCardWidget( self.chosen_card )
        if w then
            w:StartCoroutine(function()
                w:AnimateToWidget( self.flourish_bar )
            end )
        end
        self.chosen_card:AssignTarget( nil )
        self.chosen_card = nil
        self:SetCardFocus()
        AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
        return true
    end

    local card = self:GetChosenCard()
    if card then 
        -- Unselect selected card
        self.cards:ChooseCard( card )
        card:AssignTarget( nil )
        AUDIO:PlayEvent(SoundEvents.battle_card_deselect)
        return true
    end

    if self.cardspanel:IsCardSelected() then
        TheGame:FE():ClearFocusWidget()
        return true
    end

    return false
end

function NegotiationPanel:RefreshChosenCard()
    local card = self:GetChosenCard()
    if card and self.play_card ~= card then
        if self:AssignAutoTarget( card ) then
            -- Auto target assigned.
        elseif self.hover_target and self.minigame:CanPlayCard( card, self.hover_target ) then
            card:AssignTarget( self.hover_target )
        else
            card:AssignTarget( nil )    
        end

        local w = self:FindCardWidget( card )
        w:RefreshCard( card, self.minigame )
    end

    self:UpdateTargetFX( card )
end

function NegotiationPanel:GetChosenCard()
    return self.chosen_card or self.cards:GetChosenCard()
end

function NegotiationPanel:UpdateTargetFX( card )

    if card and self.minigame:GetStatus() == STATUS.PLAY_CARD then
        local source_widget = self:FindCardWidget( card )
        self.target_arrow:AttachSource( source_widget )
        local target_widget
        if is_instance( self.hover_target, Negotiation.Modifier ) then
            target_widget = self:FindSlotWidget( self.hover_target )
        elseif is_instance( self.hover_target, Negotiation.Card ) then
            target_widget = self.intents:FindIntentWidget( self.hover_target )
        end
        if target_widget then
            local target_state
            if not self.hover_target then
                target_state = TARGET_STATE.NONE
            elseif is_instance( self.hover_target, Negotiation.Card ) then
                -- Cna never target intents.
                target_state = TARGET_STATE.INVALID
            else
                local can_target, reason = self.minigame:CanPlayCard( card, self.hover_target )
                if can_target then
                    target_state = TARGET_STATE.VALID
                else
                    target_state = TARGET_STATE.INVALID
                end
            end

            self.target_arrow:AttachTarget( target_widget, target_state )
        else
            self.target_arrow:AttachMouseTarget()
        end

        self:HighlightTargets( card )
    else
        self.target_arrow:Detach()
        self:HighlightTargets( nil )
    end
end

function NegotiationPanel:HighlightTargets( card )
    local targets = ObtainWorkTable()
    if card then
        self.minigame:CollectAllTargets( card, targets )
    end

    if self.highlight_targets == nil then
        self.highlight_targets = {}
    end
    local highlight_intent_wheel = false
    local highlight_enemy_wheel = false
    local highlight_player_wheel = false
    -- Try highlighting new targets
    for i, target in ipairs( targets ) do
        if not table.contains( self.highlight_targets, target ) then
            table.insert( self.highlight_targets, target )
            local slot = self:FindSlotWidget( target ) 
            if slot then
                if target.negotiator == self.minigame:GetPlayerNegotiator() then
                    highlight_player_wheel = true
                else
                    highlight_enemy_wheel = true
                end
            end

            local intent = self.intents:FindIntentWidget( target )
            local w = slot or intent
            if intent then
                highlight_intent_wheel = true    
            end
            if w and w.HighlightTarget then
                w:HighlightTarget( true )

            end
        end
    end

    -- Unhighlight old targets
    for i = #self.highlight_targets, 1, -1 do
        local target = self.highlight_targets[ i ]
        if not table.contains( targets, target ) then
            table.remove( self.highlight_targets, i )
            local w = self:FindSlotWidget( target ) or self.intents:FindIntentWidget( target )
            if w and w.HighlightTarget then
                w:HighlightTarget( false )
            end
        end
    end

    self:HighlightWheels( highlight_player_wheel, highlight_enemy_wheel, highlight_intent_wheel )

    ReleaseWorkTable( targets )
end

function NegotiationPanel:RefreshDamagePreview()
    self.previews_dirty = true
end

function NegotiationPanel:_DoRefreshDamagePreviews()
    self.player_modifiers:RefreshDamagePreview()
    self.opponent_modifiers:RefreshDamagePreview()
end

function NegotiationPanel:AddDamagePreviewArrow( source, target, arrow_idx )
    local targetw = target and self:FindSlotWidget( target )
    if targetw == nil then
        return arrow_idx
    end

    local w = self:FindSlotWidget( source ) or self.intents:FindIntentWidget( source )
    if w then
        local arrow = self.damage_arrows[ arrow_idx ]
        if arrow == nil then
            arrow = self.damage_arrows_root:AddChild( Widget.DamageArrow() )
            table.insert( self.damage_arrows, arrow )
        else
            arrow:Show()
        end
        arrow_idx = arrow_idx + 1

        if target.modifier_type == MODIFIER_TYPE.CORE then
            arrow:Attach( w, targetw, -16 )
        else
            arrow:Attach( w, targetw, -4)
        end

        if source:IsAttack() then
            arrow:SetColour( UICOLOURS.HEALTH_DAMAGE )
        else
            arrow:SetColour( 0xAA5500FF )
        end

        if source == self.damage_preview_target then
            targetw:HilightDamagePreview( source )
        end
    end

    return arrow_idx
end

function NegotiationPanel:RefreshDamagePreviewArrows( target )
    if target ~= self.damage_preview_target then
        for i, arrow in ipairs( self.damage_arrows ) do
            arrow:Hide()
        end
        local arrow_idx = 1

        if self.damage_preview_target then
            -- Clear highlighted damage previews.
            local targets = self.minigame:CollectTargets( self.damage_preview_target )
            for i, t in ipairs( targets ) do
                local targetw = self:FindSlotWidget( t )
                if targetw then
                    targetw:HilightDamagePreview( nil )
                end
            end

            local targetw = self:FindSlotWidget( self.damage_preview_target )
            if targetw then
                targetw:HilightDamagePreview( nil )
            end
        end

        self.damage_preview_target = target

        if target then
            local hide_intents = self.minigame:CalculateBoolean( "HIDE_INTENT" )
            local sources = {}
            self.minigame:IsBeingTargetted( target, sources )

            for i, source in ipairs( sources ) do
                if not hide_intents or source:IsPlayerOwner() then
                    arrow_idx = self:AddDamagePreviewArrow( source, target, arrow_idx )
                end
            end

            if target:GetTarget() and (not hide_intents or target:IsPlayerOwner()) then
                local targets = self.minigame:CollectTargets( target )
                for i, t in ipairs( targets ) do
                    arrow_idx = self:AddDamagePreviewArrow( target, t, arrow_idx )
                end
            end
        end
    end
end

function NegotiationPanel:HighlightWheels( player_wheel, enemy_wheel, intent_wheel )
    self.player_modifiers:SetTargetDisplay( player_wheel )
    self.opponent_modifiers:SetTargetDisplay( enemy_wheel )
    self.intents:SetTargetDisplay( intent_wheel )
end

function NegotiationPanel:OnUpdate(dt)
    assert( not self.removed )

    if self.last_ev_time and not self.cardspanel:IsWaitingForClick() then
        local now = engine.inst:QueryPreciseTime()
        local speedup_factor = math.min( 5.0, math.floor( now - self.last_ev_time ) * 0.5 )
        if speedup_factor > 1 and self.speedup_factor ~= speedup_factor then
            print( "Adaptive card speed is now:", speedup_factor, serpent.line( self.event_queue[ 1 ] ))
            self.speedup_factor = speedup_factor
            self:RefreshCardSpeed()
        end

    elseif self.speedup_factor ~= nil then
        print( "Resetting adaptive card speed." )
        self.speedup_factor = nil
        self:RefreshCardSpeed()
    end

    NegotiationPanel._base.OnUpdate( self, dt )

    self.cards:AnimateCards( dt )
    self.dealt_cards:AnimateCards( dt )
    self.choose_cards:AnimateCards( dt )

    self:RepositionArgumentWheels()

    if self.minigame:GetResult() == nil then
        self:UpdateMusicPhase()
    end

    self:RefreshReason()

    local card = self:GetChosenCard() or self.cards:GetDragCard()
    if card then
        self:RefreshDamagePreviewArrows( nil )
    else
        self:RefreshDamagePreviewArrows( self.hover_target )
    end

    if self.previews_dirty then
        self.previews_dirty = false
        self:_DoRefreshDamagePreviews()
    end

    if card and self:GetFE():IsHUDVisible() and not card:IsFlagged( CARD_FLAGS.FLOURISH ) then
        self.cardspanel:ShowCancelInfoLabel()
    else
        self.cardspanel:HideCancelInfoLabel()
    end

    if not self.cardspanel:IsChoosingCards() then
        local lx, ly, rx, ry = TheGame:GetInput():GetActiveGamepadStickVals()
        if math.length2d( rx, ry ) > 0.5 then
            self:SetReticuleMode( true )
        elseif math.length2d( lx, ly ) > 0.5 then
            self:SetReticuleMode( false )
        end
    end
end

function NegotiationPanel:OnScreenResize(w,h)
    self._base.OnScreenResize(self,w,h)
    
    -- This is a horrible countdown due to frame latency in updating anim positions.    
    self.warp_floaters = 3
end

function NegotiationPanel:RepositionArgumentWheels()

    local screen_scale = self:GetFE():GetBaseWidgetScale()
    local sw, sh = engine.inst:GetScreenSize()
    local lerp = 0.25
    if self.warp_floaters ~= nil or self:GetScreen():IsShaking() then
        lerp = 1.0
    end

    local w, h = screen_scale*RES_X/3, screen_scale*RES_X/3
    local OFFSET = -20 --10
    -- Position player's resolve widget. This is also done on screen resize
    local player_body = self:GetScreen():GetCharacterWidget( self.minigame:GetPlayer() )
    if player_body then
        local wx, wy, wz = player_body.anim.model:SymbolToWorld( "head_swap", 0, 0, 0 )
        

        -- Position player's modifiers
        local min_x = (w/2) + 60*screen_scale
        local max_y = sh - (h/2) - 140*screen_scale

        local old_x, old_y = self.player_modifiers:GetPos()
        local bx, by = math.max(min_x, wx + 20 * screen_scale), math.min( wy + OFFSET * screen_scale, max_y)
        if old_x ~= 0 and old_y ~= 0 then
            -- Only lerp if the aura widget has already been positioned around the head at least once
            if self.player_modifiers.is_character_moving == false then
                -- The character has already finished animating into the stage
                -- So we should now ease the aura's movements around it, so there's no jerky motions
                bx = Lerp( old_x, bx, lerp )
                by = Lerp( old_y, by, lerp )
            end
        end
        self.player_modifiers:SetPos( bx, by ):SetScale( screen_scale )
    end
    self.player_modifiers:SetShown( player_body ~= nil )

    -- Position enemy's resolve widget. This is also done on screen resize
    local enemy_body = self:GetScreen():GetCharacterWidget( self.minigame:GetOpponent() )
    if enemy_body then
        local wx, wy, wz = enemy_body.anim.model:SymbolToWorld( "head_swap", 0, 0, 0 )
        local offsets = self.minigame:GetOpponent().def.negotiation_data.ring_offset
        if offsets then
            wx = wx + (offsets.x or 0)
            wy = wy + (offsets.y or 0)
            wz = wz + (offsets.z or 0)
        end

        -- Position enemy's modifiers
        local max_x = sw-(w/2) - 60*screen_scale
        local max_y = sh - (h/2) - 140*screen_scale

        local old_x, old_y = self.opponent_modifiers:GetPos()
        local bx, by = math.min(max_x, wx - 20 * screen_scale), math.min( wy + OFFSET * screen_scale, max_y)
        if old_x ~= 0 and old_y ~= 0 then
            -- Only lerp if the aura widget has already been positioned around the head at least once
            if self.opponent_modifiers.is_character_moving == false  then
                -- The character has already finished animating into the stage
                -- So we should now ease the aura's movements around it, so there's no jerky motions
                bx = Lerp( old_x, bx, lerp )
                by = Lerp( old_y, by, lerp )
            end
        end
        self.opponent_modifiers:SetPos( bx, by ):SetScale( screen_scale )
        self.intents:SetPos( bx, by ):SetScale( screen_scale )
    end

    -- Reset the lerp to the original value to slow down the circles again after screen resize
    if self.warp_floaters then
        self.warp_floaters = self.warp_floaters - 1
        if self.warp_floaters <= 0 then
            self.warp_floaters = nil
        end
    end

    self.opponent_modifiers:SetShown( enemy_body ~= nil )
    self.intents:SetShown( enemy_body ~= nil )
end

function NegotiationPanel:DealCards()
    self.cardspanel:DealCards()
end

function NegotiationPanel:ConsumeIntent( card )
    return self.intents:ConsumeIntent( card )
end

function NegotiationPanel:ResumePlay( ... )
    if self.wait_coro then
        self.wait_coro:Stop()
        self.wait_coro = nil
    end

    -- needs to choose a card or make a UI decision)
    local wait_status, wait_params = self.minigame:ResumePlay( ... )

    local handler
    if wait_status == STATUS.PLAY_CARD then
        handler = self.HandlePlayCard
    elseif wait_status == STATUS.CHOOSE_CARDS then
        handler = self.HandleChooseCards
    elseif wait_status == STATUS.SYNC then
        handler = self.HandleSyncPoint
    elseif self.minigame:IsDone() then
        handler = self.HandleNegotiationResult
    end

    -- Cards in hand must refresh immediately to reflect CURRENT state to allow
    -- FIXME: enabled/disabled state, but what about other features?
    self.cards:RefreshCards( self.minigame )

    if handler then
        local n = table.maxn( wait_params )
        self.wait_coro = self:StartCoroutine( handler, self, table.unpack( wait_params, 1, n ))
    end
    if self.viz_coro == nil then
        self.viz_coro = self:StartCoroutine( self.PumpAllEvents, self, self.event_queue )
    end
end

function NegotiationPanel:HandleSyncPoint( delay )
    if delay then
        self:Delay( delay )
    end
    self:WaitForEvents()

    self:DealCards()

    self:ResumePlay()
end

function NegotiationPanel:HandleChooseCards( deck, min_count, max_count, txt, pred, source )

    self:WaitForEvents()

    local cards
    if deck ~= self.minigame:GetHandDeck() then
        cards = self:HandleChooseCardsFromDeck( deck, min_count, max_count, txt, pred, source )
    else
        cards = self:HandleChooseCardsFromHand( deck, min_count, max_count, txt, pred, source )
    end
    self.choosing_cards_from_hand = false

    self:ResumePlay( cards )
end

function NegotiationPanel:HandleChooseCardsFromHand( deck, min_count, max_count, txt, pred, source )
    self.choosing_cards_from_hand = true

    -- This ensures that all cards in hand are returned to the hand CardArrangement so that they can be chosen.
    -- They might be floating, due to CARD_TRIGGERED_PRE, etc.
    for i, card in deck:Cards() do
        self:UpdateCardPlacement( card, deck, i, deck, i )
    end

    self:SetCardFocus()
    return self.cardspanel:DoCardChoosing( self.minigame, self.cards, txt, min_count, max_count, pred, source )
end

function NegotiationPanel:HandleChooseCardsFromDeck( deck, min_count, max_count, txt, pred, source )
    local cards
    if is_instance( deck, CardEngine.Deck ) then
        cards = deck:GetCards()
    else
        cards = deck
    end
    cards = shallowcopy( cards )
    self.floating_cards:Hide()
    self.cardspanel:HideObjective()

    local function OnClose( popup, cards )
        for i, card in ipairs( cards ) do
            local widget = popup:GetWidgetForCard( card )
            if self:FindCardWidget( card ) then
                widget:Remove()
            else
                self.floating_cards:AttachCardWidget( card, widget )
            end
        end
    end

    local popup = Screen.ChooseCardsPopup( cards, min_count, max_count, pred, Widget.NegotiationCard, nil, nil, OnClose )
    popup:SetTitle( txt, source )
    self:GetFE():InsertScreen( popup )

    while popup:IsOnStack() do
        coroutine.yield()
    end
    self.floating_cards:Show()
    self.cardspanel:ShowObjective()

    return cards
end

function NegotiationPanel:HandlePlayCard()
    self.end_turn_btn:HideFocusGlow()

    local chosen_card = nil

    local width, height = self:GetFE():GetScreenDims()
    self.cards:BeginChoosingForPlay( self.minigame, 1, 1 )
    do
        -- Choosing a target will resume with the given card.
        while not self.end_turn and not self.play_card do
            coroutine.yield()

            if TheGame:GetLocalSettings().ROBOTICS then
                self:Delay( 1.0 )
                if self.minigame:CountCardsPlayed() > 30 then
                    self.play_card = nil
                else
                    self.play_card = self.minigame:GetHandDeck():PickCardIf( function( card ) return self.minigame:CanPlayCard( card ) end )
                end
                self.end_turn = self.play_card == nil
            end
        end
    end
    self.cards:EndChoosing( self.minigame )
    if not self.show_reticule then
        self.cards:DisablePreview( false )
    end
    self:UpdateTargetFX( nil )

    local play_card = self.play_card
    if play_card then
        local can_play, reason = self.minigame:CanPlayCard( play_card )
        assert_warning( can_play, "Can't play %s: %s", tostring(play_card), reason )

        -- Shake the screen if the card demands it
        if play_card.def.on_play_screen_shake then
            local shake_data = NEGOTIATION_SHAKE_ABILITY
            if type(play_card.def.on_play_screen_shake) == "table" then
                -- The card itself details what shake to do
                shake_data = play_card.def.on_play_screen_shake
            end
            self:ScreenShake( shake_data )
        end
        
        -- Ensure this card doesn't animate back to hand if it is queued.
        local w = self:DetachCardWidget( play_card )
        if w then
            self.floating_cards:AttachCardWidget( play_card, w )
            w:SendToBack()
        end
    end

    self.end_turn = nil
    self.play_card = nil
    self.chosen_card = nil

    self.concede_btn:Disable()

    self:ResumePlay( play_card )
end

function NegotiationPanel:HandleNegotiationResult()
    print( "NegotiationPanel:HandleNegotiationResult", self.minigame:GetResult())

    self:WaitForEvents()

    self.damage_arrows_root:DestroyAllChildren()
    table.clear( self.damage_arrows )

    if self.minigame:GetResult() == negotiation_defs.RESULT.LOSE then
        AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/negotiation_ends" )
        self:SetMusicPhase( 6 )

    elseif self.minigame:GetResult() == negotiation_defs.RESULT.WIN then
        self:SetMusicPhase( 5 )
    end

    -- Always re-present both sides
    self.encounter:PresentAgent( self.encounter:GetPlayer(), SIDE.LEFT )
    self.encounter:PresentAgent( self.minigame:GetOpponent(), SIDE.RIGHT )

    self.intents:AlphaTo( 0, 0.5, nil, function() self.intents:Hide() end )
    
    self.floating_cards:AlphaTo( 0, 0.4 )
    self.cards:AlphaTo( 0, 0.4 )

    self.opponent_modifiers:DestroyAllSlots()
    self.player_modifiers:DestroyAllSlots()

    self:Delay( 0.5 )
end

function NegotiationPanel:SayQuip( agent, ... )
    if agent and not agent:CanTalk() and self.minigame:GetOpponent() == agent then
        return
    end

    local quipt = { ... }
    for modifier_id, modifier in self.minigame:AllModifiers() do
        if modifier.quip then
            table.insert( quipt, modifier.quip )
        end
    end

    -- This double quip lookup is unfortunate: but we don't write for every quip combination.
    -- We don't want to ClearHistory for a quip lookup that fails.
    if self.encounter:LookupQuip( agent, "negotiation", table.unpack( quipt )) then
        self:GetScreen():ClearHistory()
        return self.encounter:SayQuip( agent, "negotiation", table.unpack( quipt ))
    end
end

function NegotiationPanel:SayText( agent, txt )
    if agent and not agent:CanTalk() then
        return
    end

    self:GetScreen():ClearHistory()
    self.encounter:SayText( agent, txt )
    self:Delay( 1.0 )
    -- self.cardspanel:WaitForClick()
    self:GetScreen():ClearHistory()
end

function NegotiationPanel:SayCardQuip( card )
    self.minigame.current_quip_db = card.def.quips

    local card_speaker = card.owner
    local card_quip = string.format( "%s_card", string.lower(card.base_id or card.id) )
    local additional_quip_tags = { card_quip, card.quip }

    if card:GetNegotiator() == self.minigame:GetOpponentNegotiator() then
        local on_cooldown = self.last_opponent_quip_time and ( TheGame:GetWallTime() - self.last_opponent_quip_time  < 2.5)
        local other_owner = card_speaker ~= self.minigame:GetOpponent()

        if other_owner or not on_cooldown then
            if other_owner then
                self.encounter:PresentAgent( self.encounter:GetPlayer(), SIDE.LEFT )
                self.encounter:PresentAgent( card_speaker, SIDE.RIGHT )
            end

            if self:SayQuip( card_speaker, table.unpack(additional_quip_tags) ) then
                self.last_opponent_quip_time = TheGame:GetWallTime()
            end

            if other_owner then
                self:Delay( 1.0 )
            end
        end

    else
        local on_cooldown = self.last_player_quip_time and ( TheGame:GetWallTime() - self.last_player_quip_time  < 2.5)
        local other_owner = card_speaker ~= self.minigame:GetPlayer() and card_speaker ~= self.minigame:GetOpponent()

        if other_owner or not on_cooldown then
            if other_owner then
                self.encounter:PresentAgent( card_speaker, SIDE.LEFT )
                self.encounter:PresentAgent( self.minigame:GetOpponent(), SIDE.RIGHT )
            end

            if self:SayQuip( card_speaker, table.unpack(additional_quip_tags) ) then
                self.last_player_quip_time = TheGame:GetWallTime()
            end

            if other_owner then
                self:Delay( 1.0 )
            end
        end
    end

    self.minigame.current_quip_db = nil
end

function NegotiationPanel:Dialog( str )
    self:GetScreen():ClearHistory()
    self.encounter:Dialog( str )
    self:Delay( 1.0 )
end

function NegotiationPanel:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )
    return self.cardspanel:UpdateCardPlacement( card, source_deck, source_idx, target_deck, target_idx )
end

function NegotiationPanel:Breakpoint( txt )
    local current = coroutine.running()
    local current_coro

    for k, coro in pairs( self.coros ) do
        if coro.c == current then
           coro.breakpoint = txt 
           current_coro = coro
           break
        end
    end

    assert( current_coro, txt )
    while current_coro and current_coro.breakpoint == txt do
        coroutine.yield()
    end
end

function NegotiationPanel:WaitForEvents()
    while #self.event_queue > 0 or self.current_event do
        coroutine.yield()
    end
end

function NegotiationPanel:WaitForBanner()
    while self.cardspanel:BannerAnimating() do
        coroutine.yield()
    end
end

function NegotiationPanel:Delay( t )
    if (self.viz_coro and self.viz_coro:IsRunning()) or (self.wait_coro and self.wait_coro:IsRunning()) then
        t = t or 0
        while (t > 0) do 
            local dt = coroutine.yield()
            t = t - dt
        end
    end
end

function NegotiationPanel:NoActions( txt )
    self.cardspanel:ShowBanner( txt, nil, nil, UICOLOURS.HEALTH, UICOLOURS.HEALTH )
    self.time_indicator:AnimateOutOfTime()
end

function NegotiationPanel:RefreshDrawDeck()
    local tt
    if self.fe:IsGamepadMode() then
        tt = loc.format( TheGame:Str( "UI.CARDS.DRAW_DECK_GAMEPAD_TT" ), self.minigame:CalculateTurnDraw(), Controls.Digital.VIEW_DRAW_DECK )
    else
        tt = loc.format( TheGame:Str( "UI.CARDS.DRAW_DECK_TT" ), self.minigame:CalculateTurnDraw(), Controls.Digital.VIEW_DRAW_DECK )
    end
    self.draw_deck:RefreshDeck()
    self.draw_deck:SetToolTip( tt )
end

function NegotiationPanel:RefreshDeck( deck_type )
    if deck_type == DECK_TYPE.DRAW then
        self:RefreshDrawDeck()

    elseif deck_type == DECK_TYPE.IN_HAND then
        self.cards:RefreshCards( self.minigame )

    elseif deck_type == DECK_TYPE.DISCARDS then
        self.discard_deck:RefreshDeck()

    elseif deck_type == DECK_TYPE.TRASH then
        local deck = self.trash_deck:RefreshDeck()
        self.discard_deck:SetOtherDeckButton( deck, LOC"UI.CARDS.TRASH_DECK_DIALOG_TITLE", LOC"UI.CARDS.TRASH_DECK_DIALOG_SUBTITLE" )
    end
end

function NegotiationPanel:RefreshDecks()

    self:RefreshDrawDeck()

    local tt
    if self.fe:IsGamepadMode() then
        tt = loc.format( LOC"UI.CARDS.DISCARDS_DECK_GAMEPAD_TT", Controls.Digital.VIEW_DISCARDS_DECK )
    else
        tt = loc.format( LOC"UI.CARDS.DISCARDS_DECK_TT", Controls.Digital.VIEW_DISCARDS_DECK )
    end

    self.discard_deck:RefreshDeck()
    self.discard_deck:SetToolTip( tt )

    self:RefreshDeck( DECK_TYPE.TRASH )
    self:RefreshDeck( DECK_TYPE.IN_HAND )
end

function NegotiationPanel:FindSlotWidget( target )
    local negotiator = target.GetNegotiator and target:GetNegotiator()
    local slots = self.slot_container[ negotiator ]
    if slots then
        return slots:FindSlotWidget( target )
    end
end

function NegotiationPanel:FindCardWidget( card )
    local w = self.intents:FindIntentWidget( card )
    if w == nil then
        w = self.cardspanel:FindCardWidget( card )
    end

    return w
end

function NegotiationPanel:DetachCardWidget( card )
    return self.cardspanel:DetachCardWidget( card )
end

function NegotiationPanel:ShowAttackResolve( target, delta, source, defended, params  )
    if delta == nil then
        return
    end

    local source_widget
    if is_instance( source, Negotiation.Card ) then
        source_widget = self:FindCardWidget( source )
        if source_widget == nil then
            source_widget = self.floating_cards
        end

    elseif is_instance( source, Negotiation.Modifier ) then
        source_widget = self:FindSlotWidget( source )
        if source_widget == nil and source.graft then
            source_widget = self.grafts:FindWidget( source.graft ) or self.boons:FindWidget(source.graft) or self.banes:FindWidget(source.graft)
        end
    end

    local target_widget
    if is_instance( target, Negotiation.Modifier ) then
        target_widget = self:FindSlotWidget( target )
    end

    if source_widget and target_widget then
        DoResolveStrike( self, source, source_widget, target, target_widget, delta, defended, params )
    end
end

function NegotiationPanel:ShowCenterFloater( txt, clr )
    local label = self:AddChild( Widget.Label( "title", 48, txt ):SetBloom( 0.2 ))
    local tx, ty = self:TransformFromWidget( self.time_indicator, 0, 0 )

    if clr then
        label:SetGlyphColour( clr )
    end
    local w, h = self:GetFE():GetScreenDims()
    label:SetPos( w/2, h/2 )
    label:ScaleTo( 1.0, 1.3, 0.2, easing.outQuad )
    self:Delay( 0.2 )
    label:ScaleTo( 1.3, 1.0, 0.2, easing.outQuad )
    self:Delay( 0.5 )

    local duration = 0.5
    label:MoveTo( tx, ty, duration, easing.outQuad )
    label:AlphaTo( 0, duration )
    label:ScaleTo( 1.0, 0.5, duration, easing.outQuad )
    label:Delay( duration )
    label:Remove()
end

function NegotiationPanel:RefreshTimeLabel()
    local actions, max_actions = self.minigame:GetActionCount()
    self.time_indicator:SetTime( actions, max_actions )
end

function NegotiationPanel:GetSlotContainer( modifier )
    if modifier.modifier_type == MODIFIER_TYPE.HIDDEN then
        return
    end
    if modifier.negotiator == self.minigame:GetPlayerNegotiator() then
        return self.player_modifiers
    else
        return self.opponent_modifiers
    end
end

function NegotiationPanel:AnimateInceptedCards( cards, source )
    local target = self.minigame:GetOtherNegotiator( source.negotiator ):FindCoreArgument()
    local targetw = target and self:FindSlotWidget( target )
    local sourcew = self:FindSlotWidget( source ) or self.intents:FindIntentWidget( source )
    if targetw == nil or sourcew == nil then
        return
    end

    targetw:WaitOnCoros()

    local floaters = {}
    for i, card in ipairs( cards ) do        
        local floater = self.floating_cards:CreateCardWidget( card )
        floater:SetScale( 0.25 )

        -- Position the floater over the source card
        local source_x, source_y = floater.parent:TransformFromWidget( sourcew, 0, 0 )
        floater:SetPos( source_x, source_y )

        -- Fade it in and wait
        floater:SetTintAlpha( 0 ):AlphaTo( 1.0, 0.25 )
        floaters[i] = floater
    end

    self:Delay( 0.5 )

    for i, card in ipairs( cards ) do
        -- Animate to target, and wait
        local floater = floaters[i]
        local target_x, target_y = floater.parent:TransformFromWidget( targetw, 0, 0 )
        floater:MoveTo( target_x, target_y, 0.3, easing.inQuad )
        floater:ScaleTo( floater.sx, 0.4, 0.3 )
    end

    self:Delay( 0.3 )
    AUDIO:PlayParamEvent( "event:/sfx/battle/cards/neg/resolve_change_player", "intensity", -10 )

    local screen = self:GetScreen()
    screen:ScreenShake( INCEPT_HIT_RESULT.shake_params )
    screen:DoBloomBurst( table.unpack( INCEPT_HIT_RESULT.bloom_params ))
    local body = screen:GetCharacterWidget( target:GetNegotiator().agent )
    if body then
        -- Need to bypass all mood transitions in order for this to time right.
        -- body:Emote( "wince" )
        local emote = table.arraypick( INCEPT_HIT_RESULT.hit_emote )
        body:Play( emote )
        body:Push( body:GetIdleAnim(), true )
    end

    for i, card in ipairs( cards ) do
        self.floating_cards:DetachCardWidget( card )
        self.dealt_cards:AttachCardWidget( card, floaters[i] )
    end

    self:DealCards()
end

function NegotiationPanel:GetDebugEnv( debug_env )
    debug_env.minigame = self.minigame
    debug_env.hand = self.minigame:GetHandDeck()
    debug_env.panel = self
    debug_env.left = self.minigame:GetPlayerNegotiator()
    debug_env.right = self.minigame:GetOpponentNegotiator()
    debug_env.behaviour = debug_env.right.behaviour
end

function NegotiationPanel:RenderDebugUI( ui, panel )
    local DebugUtil = require "debug/debug_util"

    if self.minigame:GetResult() == nil then
        if ui:Button( "Win" ) then
            self.minigame:Win()
            self:ResumePlay()
        end
        ui:SameLine( nil, 10 )
        if ui:Button( "Lose" ) then
            self.minigame:Lose()
            self:ResumePlay()
        end
    end
    
    DebugUtil.RenderTimeScale( self, ui, panel )

    ui:Separator()
    self.minigame:RenderDebugUI( ui, panel )
    
    if panel:CollapsingHeader( ui, "VIZ" ) then
        for key, coro in pairs( self.coros or table.empty ) do
            if coro == self.viz_coro then
                key = "VIZ"
            elseif coro == self.wait_coro then
                key = "WAIT"
            end
            DebugUtil.RenderCoroUI( ui, panel, key, coro )
        end

        if ui:Checkbox("VIZ STEP", self.viz_break ~= nil) then
            if self.viz_break ~= nil then
                self.viz_break = nil
            else
                self.viz_break = true
            end
        end
        if self.viz_break ~= nil then
            ui:SameLine( nil, 10 )
            if ui:SmallButton( "Step" ) then
                self.viz_break = false
            end
        end
        TheGame:GetDebugEnv().viz_break = self.viz_break

        if self.current_event then
            panel:AppendTable( ui, self.current_event, "*"..tostring(self.current_event[1]))
        end

        for i, ev in ipairs( self.event_queue ) do
            panel:AppendTable( ui, ev, ev[1] )
        end
    end

    ui:Separator()
    ui:Spacing()

    return true
end

function NegotiationPanel:RenderContextMenu( ui, panel )
    if ui:MenuItem( "Win", nil, nil, self.minigame:GetResult() == nil ) then
        self.minigame:Win()
        self:ResumePlay()
    end

    if ui:MenuItem("Lose", nil, nil, self.minigame:GetResult() == nil ) then
        self.minigame:Lose()
        self:ResumePlay()
    end

    ui:Separator()

    ui:TextColored( 0x00FFFFFF, tostring(self.minigame:GetPlayerNegotiator() ))
    ui:Indent( 20 )
    for i, w in pairs( self.player_modifiers.slot_wheel.slot_widgets ) do
        if ui:MenuItem( string.format( "DBG( %s )", tostring(w) )) then
            DBG( w )
        end
    end
    ui:Unindent( 20 )

    ui:TextColored( 0x00FFFFFF, tostring(self.minigame:GetOpponentNegotiator() ))
    ui:Indent( 20 )
    for i, w in pairs( self.opponent_modifiers.slot_wheel.slot_widgets ) do
        if ui:MenuItem( string.format( "DBG( %s )", tostring(w) )) then
            DBG( w )
        end
    end
    ui:Unindent( 20 )

    ui:Separator()

    local deck = self.minigame:GetHandDeck()
    deck:RenderDebugUI( ui, panel, self.minigame )
end

function NegotiationPanel:SetMusicPhase(phase)
    if phase ~= self.music_phase then
        self.music_phase = phase
        TheGame:GetMusic():SetParameter("phase", self.music_phase)
    end
end

function NegotiationPanel:UpdateMusicPhase()
    local player_resolve, player_max_resolve = self.minigame.player_negotiator:GetResolve()
    local player_p
    if player_resolve then
        player_resolve = player_max_resolve > 0 and (player_resolve / player_max_resolve) or 0
    end

    local enemy_resolve, enemy_max_resolve = self.minigame.opponent_negotiator:GetResolve()
    local enemy_p
    if enemy_resolve then
        enemy_p = enemy_max_resolve > 0 and (enemy_resolve / enemy_max_resolve) or 0
    end

    local phase = 1
    if (player_p and player_p < .333) or (enemy_p and enemy_p < .333) then
        phase = 4
    end

    if self.music_phase and self.music_phase > phase then
        phase = self.music_phase
    end

    self:SetMusicPhase(phase)
end

function NegotiationPanel:ActivateFlourish()
    if self.minigame:GetStatus() == self.minigame.STATUS.PLAY_CARD and not self:GetChosenCard() then
        local flourishes = TheGame:GetGameProfile():GetUsableFlourishes(self.minigame:GetPlayer():GetContentID(), "NEGOTIATION")
        for i,id in ipairs(flourishes) do
            flourishes[i] = Negotiation.Card(id, self.minigame:GetPlayer())
            flourishes[i].negotiator = self.minigame:GetPlayerNegotiator() -- this. is. a hack. needed for UI targetting.
            flourishes[i].anti_negotiator = self.minigame:GetOpponentNegotiator()
        end

        self.cardspanel:ActivateFlourish( flourishes, function( card )
            self.chosen_card = card
            if card then
                self:SetArgumentFocus()
                self:RefreshChosenCard()

                if self:AssignAutoTarget( card ) then
                    self:OnTargetClicked( card:GetTarget() )
                end
            end
        end )
    end
end
