require "ui/widget"
require "ui/widgets/negotiationresolvebar"
require "ui/widgets/tooltip_argument"
require "ui/widgets/argumentcontainer"
local negotiation_defs = require "negotiation/negotiation_defs"

local MODIFIER_TYPE = negotiation_defs.MODIFIER_TYPE
local MODIFIERS_Y = 60
local MODIFIER_SPAWN_TIMING = 0.4
local COMPOSURE_SLOT_SIZE = 220
local SLOT_SIZE = 90

FRAME_COLOURS =
{
    [ MODIFIER_TYPE.INCEPTION ] = UICOLOURS.PENALTY,
    [ MODIFIER_TYPE.BOUNTY ] = UICOLOURS.BONUS,
    [ MODIFIER_TYPE.ARGUMENT ] = UICOLOURS.SUBTITLE,
    [ MODIFIER_TYPE.PERMANENT ] = UICOLOURS.SUBTITLE,
    [ MODIFIER_TYPE.CORE ] = UICOLOURS.SUBTITLE,
}

SIZE_SCALE =
{
    [ MODIFIER_TYPE.INCEPTION ] = 1,
    [ MODIFIER_TYPE.BOUNTY ] = 1,
    [ MODIFIER_TYPE.ARGUMENT ] = 1,
    [ MODIFIER_TYPE.PERMANENT ] = 1.2,
    [ MODIFIER_TYPE.CORE ] = 1.6,
}

local assets =
{
    attack = engine.asset.Texture( "UI/fightscreen_damagepreview.tex" ),
}

-------------------------------------------------------------------------------
-- Widget wrapping a Negotiation.Modifier

local NegotiationSlot = class( "Widget.NegotiationSlot", Widget )
NegotiationSlot.TARGET_RADIUS = SLOT_SIZE/2

function NegotiationSlot:init( flipped, negotiator )
    NegotiationSlot._base.init( self )

    self.flipped = flipped

    self.minigame = negotiator and negotiator.engine
    self.negotiator = negotiator

    -- Set default values
    self.composure_value = 0

    self.hover_check = (negotiator ~= nil)

    -- Prepare the selector
    local bg_border = 76
    local bg_border_scale = 0.5
    self.selector = self:AddChild( Widget.Panel( engine.asset.Texture("UI/argument_selector.tex") ) )
        :SetNineSliceCoords( bg_border, bg_border, bg_border, bg_border )
        :SetNineSliceBorderScale( bg_border_scale )
        :SetSize( SLOT_SIZE * 2.4, SLOT_SIZE * 2.4 )
        :Bloom( 0.1 )
        :Hide()
        :SetTintColour( UICOLOURS.FOCUS )
        :SetHiddenBoundingBox( true )
        :SetTintAlpha( 0 )

    -- Prepare the contents
    self.content = self:AddChild( Widget() ):Hide():SetName("CONTENT")
    
    -- Show glow behind everything
    self.glow = self.content:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex") ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.NEGOTIATION )
        :SetSize( SLOT_SIZE*1.4, SLOT_SIZE*1.4 )
        :SetTintAlpha( 1 )

    -- Add composure widget
    self.composure_widget = self.content:AddChild( Widget() )
        :SetName( "composure_widget" )
        :Hide()
        :SetTintAlpha( 0 )
        :SetToolTip( LOC"UI.NEGOTIATION_PANEL.COMPOSURE_TT" )
    self.composure_bg = self.composure_widget:AddChild( Widget.Image( engine.asset.Texture("UI/argument_badge_defend.tex"), 50, 50 ) )
    self.composure_label = self.composure_widget:AddChild( Widget.Label( "title", 32 ) )
        :SetGlyphColour( UICOLOURS.QUEST )
        :EnableOutline( 0.2 )
        :Bloom( 0.1 )
        :SetPos( 0, 2 )

    -- Show health bar
    if negotiator then
        self.resolve = self.content:AddChild( Widget.NegotiationResolveBar( negotiator:IsPlayer() ))

        -- Get a callback when the resolve changes
        self.resolve:SetOnResolveChangeFn( function( resolve, max_resolve ) self:OnResolveBarUpdated( resolve, max_resolve ) end )
        self.resolve:SetOnComposureChangeFn( function( composure ) self:OnComposureBarUpdated( composure ) end )
    end

    -- The shield widget
    self.shield_widget = self:AddChild(Widget())
        :SetName( "shield_widget" )
        :Hide()
    local animname = "anim/UI_negotiation_dot_shield_square_fx.zip"
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetBloom( 0.3 )
        :SetTintColour( UICOLOURS.PENALTY )
        :SetScale( 0.59 )
        :SetPos( 0, -1 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.shield_anim = self:AddChild( anim )

    self.selector:SendToFront()

    self.damage_preview_root = self:AddChild( Widget() )
    self.damage_previews = {}

    self:SetToolTipClass( Widget.TooltipArgument )
    self:SetFocusable( false )
    self:ShowToolTipOnFocus( false )
    self:SetToolTipLayoutFn( function( w, tooltip_widget ) self:LayoutToolTip( tooltip_widget ) end )
end

function NegotiationSlot:OnAdded()
    self.panel = self:GetAncestor( Widget.NegotiationPanel )
    assert( self.panel )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
    self:StartUpdating()
end

function NegotiationSlot:ToggleHUD( visible, hud_flags )
    self.damage_preview_root:SetShown( visible )
end

function NegotiationSlot:OnScreenModeChange( sm )
    NegotiationSlot._base.OnScreenModeChange( self, sm )

    self:SetLayoutScale( LAYOUT_SCALE[sm] )
end

function NegotiationSlot:RegisterCoro()
    local c, is_main = coroutine.running()
    if c and not is_main then
        if self.registered_coros == nil then
            self.registered_coros = {}
        end
        table.insert( self.registered_coros, c )
    end
end

function NegotiationSlot:RunAfterCoros( panel, fn )
    if self.registered_coros and #self.registered_coros > 0 then
        panel:StartCoroutine( function()
            self:WaitOnCoros()
            if self.removed then
                LOGWARN( "already removed: %s", tostring(self.modifier))
                return
            end
            self:RegisterCoro()
            fn()
        end )
    else
        fn()
    end
end

function NegotiationSlot:WaitOnCoros()
    if self.registered_coros and #self.registered_coros > 0 then
        local running = coroutine.running()
        local done = false

        while not done and not self.removed do
            done = true
            for i = #self.registered_coros, 1, -1 do
                local c = self.registered_coros[i]
                if c ~= running then
                    if coroutine.status( c ) == "dead" then
                        table.remove( self.registered_coros, i )
                    else
                        done = false
                    end
                end
            end
            if not done then
                coroutine.yield()
            end
        end
    end
end

function NegotiationSlot:HiliteTargettedByEnemy( hilite )
end

function NegotiationSlot:HighlightTarget( hilite )
    self.hilite = hilite

    if not hilite then
        self.selector:SetScale( 1.0 )
            :SetBloom( 0 )
            :AlphaTo( 0, 0.1, easing.linear, function() self.selector:Hide() end )
    else
        self.selector:SetBloom( 0.05 )
            :Show()
            :AlphaTo( 1, 0.1 )
    end
end

function NegotiationSlot:PingLabel()
end

function NegotiationSlot:PingSlot()
end

function NegotiationSlot:PingModifier( modifier )
    if modifier then
        if modifier.sound then
            --AUDIO:PlayEvent( modifier.sound )
        end

        if modifier.fx then
            UIHelpers.PlayAnimOnce( self, modifier.fx.animfile, modifier.fx.anim )
        end
    end
    UIHelpers.PingImage( self.art, self )
end

function NegotiationSlot:GetUserData()
    return self.user_data
end

function NegotiationSlot:GetTarget()
    return self.minigame:FindModifierByUID( self.user_data )
end

function NegotiationSlot:IsOldestArgument()
    if self.negotiator:GetSlotCount() >= self.negotiator.max_modifiers then
        local oldest = self.negotiator:FindOldestArgument()
        return oldest and oldest.uid == self.user_data
    end

    return false
end

function NegotiationSlot:OnUpdate( dt )
    NegotiationSlot._base.OnUpdate( self, dt )

    local t = self.elapsed_time or 0
    self.elapsed_time = t + dt

    if self.hilite then
        self.selector:SetScale( 1.1 + 0.1 * math.sin ( t * 6 ) )
        self.glow:SetTintColour( UICOLOURS.FOCUS )
    else
        self.glow:SetTintColour( UICOLOURS.NEGOTIATION )
    end
end

function NegotiationSlot:HilightDamagePreview( source )
    for i, widget in ipairs( self.damage_previews ) do
        if widget.source == source then
            local scale_x, scale_y = widget:GetScale()
            widget:ScaleTo(scale_x, 1.4, 0.2)
            widget.img:SetBloom( 0.2 )
        else
            local scale_x, scale_y = widget:GetScale()
            widget:ScaleTo(scale_x, 1, 0.3)
            widget.img:SetBloom( 0.1 )
        end
    end
end

function NegotiationSlot:CreateDamagePreview( source )
    if source.min_persuasion == nil or source.max_persuasion == nil then
        return
    end
    
    local w = Widget()

    -- Preview bg
    local img = Widget.Image( assets.attack, 54, 54 )
    img:Bloom( 0.1 )
    w:AddChild( img )
    w.img = img
    w.source = source

    local label = Widget.Label( "title", 24 ):EnableOutline( true ):SetGlyphColour( UICOLOURS.WHITE )
    local min_delta, max_delta = self.minigame:PreviewPersuasion( source, true )

    if min_delta ~= max_delta then
        label:SetText( loc.format( "{1}-{2}", min_delta, max_delta ))
    else
        label:SetText( tostring( max_delta ))
    end
    w:AddChild( label ):LayoutBounds("center","center", img ):Offset( 0, 2 )
    w:SetToolTip( loc.format( LOC"UI.NEGOTIATION_PANEL.DAMAGE_RECEIVED_TOOLTIP", max_delta, source.owner:GetName() ) )

    self.damage_preview_root:AddChild( w ):LayoutBounds( "after", "center" ):Offset( -10, 0 )
    table.insert( self.damage_previews, w )
end

function NegotiationSlot:RefreshDamagePreview()
    while #self.damage_previews > 0 do
        table.remove( self.damage_previews ):Remove()
    end

    local show = false
    local modifier = self.negotiator:FindModifierByUID( self.user_data )
    if modifier then
        local hide_intents = self.minigame:CalculateBoolean( "HIDE_INTENT" )
        local sources = {}
        self.minigame:IsBeingTargetted( modifier, sources )

        for i, source in ipairs( sources ) do
            if not hide_intents or source:IsPlayerOwner() then
                self:CreateDamagePreview( source )
                show = show or source:IsAttack()
            end
        end
        self.damage_preview_root:LayoutBounds( "center", "above", self.selector )
    end

    self:HiliteTargettedByEnemy( show )
end

function NegotiationSlot:AttachModifier( modifier )
    -- so that the modiifer can be referenced later, but it should appear empty.
    self.user_data = modifier:GetUID()
end

function NegotiationSlot:CreateModifier( modifier )
    self.modifier = modifier
    self.user_data = modifier:GetUID()

    self.content:Show()

    -- Update resolve
    local resolve, max_resolve = modifier:GetResolve()
    if resolve and self.resolve then
        self.resolve:RefreshResolve( resolve, max_resolve )

        local composure = modifier:GetComposure()
        self.resolve:RefreshComposure( composure )
        self:OnComposureBarUpdated( composure )
        self.resolve:Show()

    elseif self.resolve then
        self.resolve:Hide()
    end

    if modifier:GetShieldStatus() then
        self:SetShieldStatus(true)
    end

    if self.negotiator then
        self:RefreshDamagePreview()
    end
end

--this should animage nicely and play some sounds!
function NegotiationSlot:SetShieldStatus( status, modifier )
    if status then
        self.shield_widget:Show()
        self.shield_anim:Play( "in", false )
            :Push( "loop", true )
    else
        self.shield_widget:AlphaTo(1, 1, easing.outQuad, function() self.shield_widget:Hide() end)
        self.shield_anim:Play( "out", false )
    end
    self:SetToolTip( modifier )
end


function NegotiationSlot:AnimateComposure( value, old_value, start_of_turn )
    if old_value > 0 and value <= 0 then
        --Differentiate between destroyed and expired
        if start_of_turn then
            AUDIO:PlayParamEvent( "event:/sfx/battle/cards/neg/composure_expires", "position", self:GetPanPos() )
        else
            AUDIO:PlayEvent( self.negotiator:IsPlayer() and "event:/sfx/battle/cards/neg/composure_destroyed_player" or "event:/sfx/battle/cards/neg/composure_destroyed_enemy" )
        end
    else
        AUDIO:PlayParamEvent( self.negotiator:IsPlayer() and "event:/sfx/battle/cards/neg/composure_change_player" or "event:/sfx/battle/cards/neg/composure_change_enemy", "intensity", value - old_value)
    end   

    self.resolve:AnimateComposure( value, old_value, start_of_turn )
end

function NegotiationSlot:AnimateResolve( delta, resolve, max_resolve )
    self.resolve:AnimateResolve( resolve, max_resolve )
end

function NegotiationSlot:OnResolveBarUpdated( resolve, max_resolve )
end

function NegotiationSlot:OnComposureBarUpdated( composure )

    -- Update the composure too
    if self.composure_value == 0 and composure > 0 then
        -- We just gained composure!
        self.composure_widget:Show()
            :AlphaTo( 1, 0.1, easing.outQuad )
            :ScaleTo( 1.2, 1, 0.1 )
        self.composure_label:SetText( loc.format( "{1}", composure ) )
    elseif self.composure_value > 0 and composure == 0 then
        -- We just lost all our composure!
        self.composure_widget:AlphaTo( 0, 0.1, easing.outQuad )
            :ScaleTo( 1, 1.2, 0.1, easing.outQuad, function() self.composure_widget:Hide() end )
        self.composure_label:SetText( loc.format( "{1}", composure ) )
    else
        -- Update the value
        self.composure_label:SetText( loc.format( "{1}", composure ) )
    end
    self.composure_value = composure
end

function NegotiationSlot:ShowParticles( show )
    if show then
        if self.particles == nil then
            self.particles = self:AddChild( Widget.ParticleSystem() ):SetPos(0,90)
            self.particles:LoadParams("negotiation_modifier")
        else
            self.particles:Restart()
        end

    elseif not show and self.particles then
        -- self.particles:Stop()
    end
end

function NegotiationSlot:LayoutToolTip( tooltip_widget )
    tooltip_widget:LayoutBounds( "center", "below", self ):Offset( 0, -20 )
    self:GetFE():ConstrainToolTipPos( tooltip_widget )
end

function NegotiationSlot:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        self:ScaleTo( nil, 1.05, 0.2, easing.outQuad )
        self.selector:Bloom( 0.1 )

        self.panel:SetHoverTarget( self:GetTarget() )

        -- In inspect mode with controller, show the target more clearly
        if not self.hilite and self.focus then
            self.selector:Show():SetScale( 1.1 )
            self.selector:SetScale( 1.1 )
            self.selector:SetTintColour( UICOLOURS.SUBTITLE )
            self.selector:AlphaTo( 1, 0.1 )
            self.glow:SetTintColour( UICOLOURS.SUBTITLE )
        end
    else
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
        self.selector:Bloom( 0.05 )

        self.panel:SetHoverTarget( nil )

        if not self.hilite then
            self.selector:AlphaTo( 0, 0.1, easing.linear, function()
                self.selector:Hide()
                self.selector:SetTintColour( UICOLOURS.FOCUS )
            end )
            self.glow:SetTintColour( UICOLOURS.NEGOTIATION )
        end
    end
end

function NegotiationSlot:OnGainFocus()
    NegotiationSlot._base.OnGainFocus(self)
    self:UpdateImage()
end

function NegotiationSlot:OnLoseFocus()
    NegotiationSlot._base.OnLoseFocus(self)
    self:UpdateImage()
end

function NegotiationSlot:OnGainHover()
    NegotiationSlot._base.OnGainHover(self)
    self:UpdateImage()
end

function NegotiationSlot:OnLoseHover()
    NegotiationSlot._base.OnLoseHover(self)
    self:UpdateImage()
end


-------------------------------------------------------------------------------
-- Widget displaying a standard modifier

local NegotiationArgument = class( "Widget.NegotiationArgument", Widget.NegotiationSlot )

function NegotiationArgument:init( flipped, negotiator )
    NegotiationArgument._base.init( self, flipped, negotiator )

    -- This modifier's frame
    self.frame_glow = self.content:AddChild( Widget.Image( engine.asset.Texture("UI/argument_glow.tex"), SLOT_SIZE, SLOT_SIZE ) )
        :SetHiddenBoundingBox(true)
        :SetTintColour(UICOLOURS.SUBTITLE)

    -- The upcoming damage animation
    local animname = "anim/UI_negotiation_square_fx.zip"
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetScale( 0.335 )
        :SetBloom( 0.2 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.targetted_anim = self.content:AddChild( anim )

    -- This modifier's image
    self.art_container = self.content:AddChild( Widget() )
    self.art_mask = Widget.Image( engine.asset.Texture("UI/argument_mask.tex"), SLOT_SIZE, SLOT_SIZE )
    self.art = Widget.Image( nil, SLOT_SIZE-25, SLOT_SIZE-25 )
    self.art_container:SetStencilContext( true ):AddChildren
    {
        self.art_mask:SetMask( true ),
        self.art:SetStencilTest( true ),
    }

    -- This modifier's frame
    self.frame = self.content:AddChild( Widget.Image( engine.asset.Texture("UI/argument_bg.tex") ) )
        :SetSize( SLOT_SIZE, SLOT_SIZE )
        :SetHiddenBoundingBox(true)
        :SetTintColour(UICOLOURS.SUBTITLE)
        :Bloom( 0.1 )

    -- Overlaying animation, for when it gets destroyed and such
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetScale( 0.335 )
        :SetBloom( 0.2 )
        :SetPos( -1, -1 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.center_anim = self.content:AddChild( anim )

    -- Display of this modifier's stack count
    self.stack_widget = self.content:AddChild( Widget() ):SetName( "stack_widget" )
        :SetHiddenBoundingBox( true )
    self.stack_bg = self.stack_widget:AddChild( Widget.Image( engine.asset.Texture("UI/argument_stack.tex") ) )
        :SetSize( SLOT_SIZE, SLOT_SIZE )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.stack_label = self.stack_widget:AddChild( Widget.Label( "title", 24 ) )
        :SetGlyphColour( UICOLOURS.DARK_BODYTEXT )
        :SetHiddenBoundingBox( true )
        :SetRegionSize( 40, 30 )
        :SetPos( -25, 32.5 )
        :SetBloom( 0.0 )
    self.stack_widget:SetPos( -2, 18 )

    -- Display of this modifier's damage value
    self.damage_bg = self.content:AddChild( Widget.Image( engine.asset.Texture("UI/argument_damage.tex"), SLOT_SIZE, SLOT_SIZE ) )
    self.damage_label = self.content:AddChild( Widget.Label( "title", 25 ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :EnableOutline( 0.2 )

    -- Position the resolve bar
    if self.resolve then
        self.resolve:LayoutBounds( "center", "center", self.frame ):Offset( 0, -SLOT_SIZE*0.66 )
    end

    -- Bring the composure_widget to the front
    self.composure_widget:SendToFront()
    -- And position it properly!
    self.composure_widget:SetPos( SLOT_SIZE*0.39, SLOT_SIZE*0.33 )

    -- Update the selector size
    self.selector:SetSize( SLOT_SIZE * 1.96, SLOT_SIZE * 1.96 )
    self.selector:SendToBack()
end

function NegotiationArgument:PingSlot()
    return UIHelpers.PingImage( self.art, self )
end

function NegotiationArgument:HiliteTargettedByEnemy( hilite )
    if hilite then
        self.targetted_anim:SetTintColour( 0xFF330077 )
        local modifier = self.negotiator:FindModifierByUID( self.user_data )
        local is_core_argument = modifier and (modifier.modifier_type == MODIFIER_TYPE.PERMANENT or modifier.modifier_type == MODIFIER_TYPE.CORE)
        self.targetted_anim:Play( is_core_argument and "highlight" or "radial1_loop", true ):Show()
    else
        self.targetted_anim:Hide()
    end
end

function NegotiationArgument:AnimateResolve( delta, resolve, max_resolve )
    if delta < 0 then
        self.center_anim:SetTintColour( UICOLOURS.NEGOTIATION ):Show()
        if resolve <= 0 then
            self.center_anim:Play( "break", false )
        elseif math.abs(delta) <= 3 then
            self.center_anim:Play( "hit_1", false )
        else
            self.center_anim:Play( "hit_2", false )
        end
    end

    self.resolve:AnimateResolve( resolve, max_resolve )
    local modifier = self.minigame:FindModifierByUID( self.user_data )
    self:SetToolTip( modifier )
end

function NegotiationArgument:CreateModifier( modifier )
    NegotiationArgument._base.CreateModifier( self, modifier )

    -- Update the image
    self.art:SetTexture( modifier.icon ):Show():SetTintAlpha( 1.0 )

    self.frame_tint = FRAME_COLOURS[ modifier.modifier_type or MODIFIER_TYPE.ARGUMENT ]
    self.frame:SetTintColour( self.frame_tint )
    self.frame_glow:SetTintColour( self.frame_tint )
    self.stack_bg:SetTintColour( self.frame_tint )
    self.damage_offset = 14

    -- Display stacks if appropos
    if modifier.hide_stacks then
        self.stack_widget:Hide()
    else
        local stacks, max_stacks = modifier:GetStacks()
        if max_stacks > 1 or modifier.is_counter then
            self.stack_widget:Show()
            self.stack_label:SetText( loc.format( "x{1}", tostring( stacks ) ) )
        else
            self.stack_widget:Hide()
        end
    end

    -- Check if this is a Bait modifier
    if modifier.force_target then
        self.frame:SetTexture( engine.asset.Texture("UI/argument_bg_bait.tex") )
        self.art_mask:SetTexture( engine.asset.Texture("UI/argument_mask_bait.tex") )
        self.damage_bg:SetTexture( engine.asset.Texture("UI/argument_bg_bait.tex") )
    end

    -- Check if this is a Def argument
    if modifier.modifier_type == MODIFIER_TYPE.PERMANENT or modifier.modifier_type == MODIFIER_TYPE.CORE then
        local new_size = SLOT_SIZE * (SIZE_SCALE[modifier.modifier_type] or 1)
        self.selector:SetSize( new_size * 1.5, new_size * 1.5 )
        self.glow:SetSize( new_size * 1.4, new_size * 1.4 )
        self.art_mask:SetTexture( engine.asset.Texture("UI/argument_mask_def.tex") )
        self.art_mask:SetSize( new_size, new_size )
        self.damage_bg:SetTexture( engine.asset.Texture("UI/argument_damage_def.tex") )
        self.damage_bg:SetSize( new_size, new_size )
        self.damage_offset = 20
        self.art:SetSize( new_size-25, new_size-25 )
        self.frame:SetTexture( engine.asset.Texture("UI/argument_bg_def.tex") )
        self.frame:SetSize( new_size, new_size )
        self.stack_bg:SetTexture( engine.asset.Texture("UI/argument_stack_def.tex") )
        self.stack_bg:SetSize( new_size, new_size )
        self.stack_bg:SetPos( nil, 8 )
        if modifier.modifier_type == MODIFIER_TYPE.CORE then
            self.stack_widget:SetPos( -2, 18.5 )
            self.stack_label:SetPos( -40, 60 )
        else
            self.stack_widget:SetPos( 0, 10.5 )
            self.stack_label:SetPos( -32, 48 )
        end
        self.composure_widget:SetPos( new_size*0.39, new_size*0.33 )
        self:AddShine( new_size )

        if self.resolve then
            self.resolve:LayoutBounds( "center", "below", self.frame ):Offset( 0, -6 )
            self.resolve.label:SetFontSize( 44 )
        end

        -- Core arguments use a different targetted animation
        local animname = "anim/UI_negotiation_round_fx.zip"
        self.targetted_anim
            :AddAnims( GetAnim( animname ) )
            :AddOverrideFile( animname )
            :SetScale( 0.78 )

        -- Core arguments use a different shield animation
        local animname = "anim/UI_negotiation_dot_shield_fx.zip"
        self.shield_anim
            :AddAnims( GetAnim( animname ) )
            :AddOverrideFile( animname )
            :SetScale( 0.52, 0.51 )
            :SetPos( 0, -1 )
    end

    -- Update damage
    if modifier.min_persuasion and modifier.max_persuasion then
        local txt = self:GeneratePersuasionText(modifier)
        self.damage_bg:Show()
        self.damage_label:SetText( txt )
            :LayoutBounds( "center", "bottom", self.damage_bg )
            :Offset( 0, self.damage_offset )
            :Show()
    else
        self.damage_bg:Hide()
        self.damage_label:Hide()
    end

    -- Update tooltip
    if self.hover_check then
        self:SetToolTip( modifier )
    end
end

function NegotiationArgument:RefreshIcon(modifier, icon)
    self.art:SetTexture( icon or modifier.icon ):Show():SetTintAlpha( 1.0 )
end

function NegotiationArgument:AddShine( slot_size )
    -- Animated reflection shine on the slot
    if not self.shine_coro then
        self.shine = self.art_container:AddChild( Widget.Image( engine.asset.Texture( "UI/argument_bg_shine.tex") ) )
            :SetStencilTest( true )
            :SetHiddenBoundingBox( true )
            :SetSize( slot_size, slot_size*2 )
            :SetTintColour( UICOLOURS.SUBTITLE )
            :SetTintAlpha( 0.2 )
            :Bloom( 0.4 )
            :SetPos( 0, 400 )

        self.shine_coro = self:StartCoroutine( self.ShowShine, self )
    end
end

function NegotiationArgument:ShowShine()
    local anim_pacing = 2.6

    while true do
        self:Delay( 1 + math.random() * 3.0 )

        self.shine:Show()

        self.shine:SetPos( 0, 400 )
            :MoveTo( 0, -400, anim_pacing, easing.outQuad )

        self:Delay( anim_pacing )
        self.shine:Hide()
    end
end

function NegotiationArgument:GeneratePersuasionText( modifier )
    local min_persuasion, max_persuasion, details = modifier.engine:PreviewPersuasion( modifier, true )

    local min_txt
    if min_persuasion < modifier.min_persuasion then
        min_txt = loc.format( "<#PENALTY>{1}</>",  min_persuasion )
    elseif min_persuasion > modifier.min_persuasion then
        min_txt = loc.format( "<#BONUS>{1}</>", min_persuasion )
    else
        min_txt = tostring(min_persuasion)
    end

    local max_txt
    if max_persuasion < modifier.max_persuasion then
        max_txt = loc.format( "<#PENALTY>{1}</>",  max_persuasion )
    elseif max_persuasion > modifier.max_persuasion then
        max_txt = loc.format( "<#BONUS>{1}</>", max_persuasion )
    else
        max_txt = tostring(max_persuasion)
    end

    local txt
    if min_txt == max_txt then
        txt = min_txt
    else
        txt = loc.format( "{1}-{2}", min_txt, max_txt )
    end

    return txt
end

function NegotiationArgument:UpdatePersuasionLabel()
    local modifier = self.minigame:FindModifierByUID( self.user_data )
    if modifier and modifier.min_persuasion and modifier.max_persuasion then
        local txt = self:GeneratePersuasionText(modifier)
        self.damage_bg:Show()
        self.damage_label:SetText( txt )
            :LayoutBounds( "center", "bottom", self.damage_bg )
            :Offset( 0, self.damage_offset )
            :Show()

        self:SetToolTip( modifier )
    end
end

function NegotiationArgument:OnComposureBarUpdated( composure )
    NegotiationArgument._base.OnComposureBarUpdated( self, composure )

    -- Update the composure too
    if self.composure_value == 0 and composure > 0 then
        -- We have composure!
        self.stack_bg:SetTintColour(UICOLOURS.QUEST)
        self.frame_glow:SetTintColour(UICOLOURS.QUEST)
            :Bloom( 0.1 )
        self.frame:SetTintColour(UICOLOURS.QUEST)
    else
        -- We have no composure!
        self.frame:SetTintColour( self.frame_tint )
        self.frame_glow:SetTintColour( self.frame_tint )
            :Bloom( 0 )
        self.stack_bg:SetTintColour( self.frame_tint )
    end
end


-------------------------------------------------------------------------------
-- Widget displaying a coin modifier

local NegotiationCoinArgument = class( "Widget.NegotiationCoinArgument", Widget.NegotiationArgument )

function NegotiationCoinArgument:init( flipped, negotiator )
    NegotiationCoinArgument._base.init( self, flipped, negotiator )
end

function NegotiationCoinArgument:CreateModifier( modifier )
    NegotiationCoinArgument._base.CreateModifier( self, modifier )

    self.modifier = modifier
    -- Display of the current coin state
    self.coin_indicator = self.content:AddChild( Widget() ):SetName( "coin_indicator" )
        :SetHiddenBoundingBox( true )
    self.coin_indicator_bg = self.coin_indicator:AddChild( Widget.Image( engine.asset.Texture("UI/argument_stack_def.tex") ) )
        :SetSize( SLOT_SIZE * (SIZE_SCALE[modifier.modifier_type] or 1), SLOT_SIZE * (SIZE_SCALE[modifier.modifier_type] or 1) )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.coin_indicator_icon = self.coin_indicator:AddChild( Widget.Image( engine.asset.Texture("UI/ic_coinindicator_heads.tex") ) )
        :SetSize( 25, 25 )
        :SetHiddenBoundingBox( true )
        :SetPos( -41, 51 )
        :SetBloom( 0.0 )
    self.coin_indicator:SetPos( 0, 25 )
        :SendBehind( self.composure_widget )
        :Hide()

    -- Overall animation speed
    self.coin_flip_anim_speed = 1.2 -- 1 is standard, 0.5 is slow

    -- Container for the animation layers
    self.coin_flip_anim = self.content:AddChild( Widget() )
        :Hide()

    -- The coin flip animation layer 1
    self.coin_anim_shield = Widget.Anim()
    local animname = "anim/UI_negotiation_coin_fx.zip"
    self.coin_anim_shield:AddAnims( GetAnim( animname ) )
    self.coin_anim_shield:AddOverrideFile( animname )
    self.coin_flip_anim:AddChild( self.coin_anim_shield )
        :SetHiddenBoundingBox(true)
        :SetScale( 0.55, 0.54 )
        :SetBloom( 0.1 )
        :SetTintAlpha( 0.9 )
        :SetPos( 2, 2 )
        :SetSpeed( self.coin_flip_anim_speed )

    -- The coin flip animation layer 2
    self.coin_anim_flip = Widget.Anim()
    local animname = "anim/UI_negotiation_coin_fx.zip"
    self.coin_anim_flip:AddAnims( GetAnim( animname ) )
    self.coin_anim_flip:AddOverrideFile( animname )
    

    self.coin_flip_anim:AddChild( self.coin_anim_flip )
        :SetHiddenBoundingBox(true)
        :SetScale( 0.55, 0.54 )
        :SetTintAlpha( 0.9 )
        :SetPos( 2, 2 )
        :SetSpeed( self.coin_flip_anim_speed )

    -- The coin flip animation layer 3
    self.coin_anim_shine = Widget.Anim()
    local animname = "anim/UI_negotiation_coin_fx.zip"
    self.coin_anim_shine:AddAnims( GetAnim( animname ) )
    self.coin_anim_shine:AddOverrideFile( animname )
    self.coin_flip_anim:AddChild( self.coin_anim_shine )
        :SetHiddenBoundingBox(true)
        :SetScale( 0.55, 0.54 )
        :SetBloom( 0.2 )
        :SetTintAlpha( 0.9 )
        :SetPos( 2, 2 )
        :SetSpeed( self.coin_flip_anim_speed )

    -- The coin flip animation layer 4
    self.coin_anim_sparks = Widget.Anim()
    local animname = "anim/UI_negotiation_coin_fx.zip"
    self.coin_anim_sparks:AddAnims( GetAnim( animname ) )
    self.coin_anim_sparks:AddOverrideFile( animname )
    self.coin_flip_anim:AddChild( self.coin_anim_sparks )
        :SetHiddenBoundingBox(true)
        :SetScale( 0.55, 0.54 )
        :SetBloom( 0.5 )
        :SetTintAlpha( 0.9 )
        :SetPos( 2, 2 )
        :SetSpeed( self.coin_flip_anim_speed )
end

function NegotiationCoinArgument:RefreshIcon( modifier )
    self.modifier = modifier
end

function NegotiationCoinArgument:FlipCoin( modifier, status )
    self.modifier_transition_queue = self.modifier_transition_queue or {}
    self.modifier = modifier or self.modifier
    if modifier then table.insert( self.modifier_transition_queue, status ) end

    -- If there are modifiers in the queue, transition them.
    -- When one finishes, it'll call this function again
    if #self.modifier_transition_queue > 0 and not self.animation_started then
        local status = table.remove( self.modifier_transition_queue, 1 )
        self:AnimateCoin( status )
    end
end

function NegotiationCoinArgument:AnimateCoin( status )

    if self.modifier.coin_build then
        self.coin_anim_flip:AddOverrideFile( self.modifier.coin_build )
    end

    self.animation_started = true
    self.starting_color = self.target_color or 0xBE4CD2FF
    local target_icon
    if status == "HEADS" then
        self.target_color = 0x4294C4FF
        target_icon = engine.asset.Texture( self.modifier.heads_icon, true )
    elseif status == "SNAILS" then
        self.target_color = 0x9E3030FF
        target_icon = engine.asset.Texture( self.modifier.snails_icon, true )
    end
    self:RunUpdater(
        Updater.Parallel{
            Updater.Do( function()
                -- Set the anim to the initial colour
                -- self.coin_flip_anim:SetTintColour( self.starting_color )
                -- Set the anim to the final colour
                self.coin_anim_sparks:SetTintColour( self.target_color )
                -- Start the animation
                self.coin_flip_anim:Show()
                self.coin_anim_shield:Play( "shield", false )
                self.coin_anim_flip:Play( status == "HEADS" and "coin_heads" or "coin_tails", false )
                self.coin_anim_shine:Play( status == "HEADS" and "shine_heads" or "shine_tails", false )
                self.coin_anim_sparks:Play( "sparks", false )
            end ),
            Updater.Series{
                -- Wait for frame 13 in the anim
                Updater.Wait( 1/30*13/self.coin_flip_anim_speed ),
                Updater.Do( function()
                    -- Hide the argument during the animation
                    self.glow:Hide()
                    self.frame_glow:Hide()
                    self.frame:Hide()
                    self.art_container:Hide()
                    self.coin_indicator:Hide()
                end ),
            },
            Updater.Series{
                -- Wait for frame 21 in the anim
                Updater.Wait( 1/30*21/self.coin_flip_anim_speed ),
                Updater.Do( function()
                end ),
            },
            Updater.Series{
                -- Wait for frame 30 in the anim
                Updater.Wait( 1/30*30/self.coin_flip_anim_speed ),
                Updater.Do( function()
                    -- Set the right image in the argument
                    self.art:SetTexture( target_icon ):Show():SetTintAlpha( 1.0 )
                    -- Show the argument again
                    self.glow:Show()
                    self.frame_glow:Show()
                    self.frame:Show()
                    self.art_container:Show()
                    -- Update coin indicator
                    self.coin_indicator:Show()
                        :SetScale( status == "HEADS" and 1 or -1, 1 )
                    self.coin_indicator_bg:SetTintColour( status == "HEADS" and UICOLOURS.SUBTITLE or UICOLOURS.PENALTY )
                    self.coin_indicator_icon:SetTexture( status == "HEADS" and engine.asset.Texture("UI/ic_coinindicator_heads.tex") or engine.asset.Texture("UI/ic_coinindicator_snails.tex") )
                    :SetScale( status == "HEADS" and 1 or -1, 1 )
                end ),
            },
            Updater.Series{
                -- Wait for end of the anim
                Updater.Wait( 1/30*49/self.coin_flip_anim_speed ),
                Updater.Do( function()
                    self.animation_started = false
                    self:FlipCoin()
                end )
            }
        }
    )
end

function NegotiationCoinArgument:PingModifier( modifier )
    self:Delay( modifier.ping_delay or 0.5 )
end

-------------------------------------------------------------------------------
-- Widget containing Negotiation.Modifier widgets

local SlotContainer = class( "Widget.SlotContainer", Widget )

function SlotContainer:init( negotiator, diameter )
    SlotContainer._base.init( self )

    self.minigame = negotiator.engine
    self.negotiator = negotiator

    if self.negotiator.agent:IsPlayer() then
        self.flipped = true
    else
        self.flipped = false
    end

    self.slot_wheel = self:AddChild( Widget.ArgumentContainer( diameter or 600 ) )
        :SetStartingSide( self.flipped )
        :SetWheelColour( UICOLOURS.SUBTITLE )
        -- :SetAgentName( self.negotiator:GetName() )

end

function SlotContainer:GetDefaultFocus()
    return self.slot_wheel:GetDefaultFocus()
end

function SlotContainer:ToggleHUD( visible, hud_flags )
    for i, w in ipairs( self.slot_wheel.slot_widgets ) do
        w:ToggleHUD( visible, hud_flags )
    end
end

function SlotContainer:AnimateModifierFloater( source, target, modifier, delta )
    target:WaitOnCoros()
    target:RegisterCoro()

    local scale = 1.0
    local x0, y0
    local panel = self:GetAncestor( Widget.NegotiationPanel )

    if source then
        local source_widget
        if delta and delta > 0 then
            if is_instance( source, Negotiation.Card ) then
                source_widget = self:GetAncestor( Widget.NegotiationPanel ):FindCardWidget( source )
                if not source.negotiator:IsPlayer() then
                    AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/argument_appear_gen" )
                end

                if source.flags and CheckBits( source.flags, negotiation_defs.CARD_FLAGS.BYSTANDER ) and source.negotiator:IsPlayer() then
                    AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/friendly_arguments_presented" )
                end

            elseif is_instance( source, Negotiation.Modifier ) then
                source_widget = self:GetAncestor( Widget.NegotiationPanel ):FindSlotWidget( source )
            end
        end

        if source_widget then
            x0, y0 = panel:TransformFromWidget( source_widget, 0, 0 )
        else
            local screenw, screenh = self:GetFE():GetScreenDims()
            x0, y0 = panel:TransformFromWorld( screenw/2, screenh/2 )
            scale = 1.5
        end
    end

    -- If there is a card, animate a floater modifier from the card to the tray
    local floater
    if x0 and y0 then
        floater = panel:AddChild( Widget.NegotiationArgument( self.flipped ) ):SetHiddenBoundingBox(true )
        floater:CreateModifier( modifier )

        -- Animate the floater from the card to the target
        floater:SetPos( x0, y0 )
        floater:SetTintAlpha( 0 )

        local scale_x, scale_y = floater:GetScale()

        floater:ScaleTo( scale_x * scale * 1.5, scale_x * scale, MODIFIER_SPAWN_TIMING/2, easing.inQuad )
        floater:AlphaTo( 1.0, MODIFIER_SPAWN_TIMING/2 )

        floater:DelayIfAnimating()
        floater:Delay( MODIFIER_SPAWN_TIMING/2 )

        -- wait til we're visible
        if not self:IsShown() then
            while not self:IsShown() and not self.destroying do
                coroutine.yield()
            end
            self:Delay( 0.25 )
        end

        -- Get target position
        AUDIO:PlayEvent( "event:/sfx/battle/cards/neg/argument_slide_into_place" )

        local target_x, target_y = floater.parent:TransformFromWidget( target, 0, 0 )
        floater:MoveTo( target_x, target_y, MODIFIER_SPAWN_TIMING/2, easing.inQuad )
        floater:Delay( MODIFIER_SPAWN_TIMING/2 )
        floater:Remove()
    end

    self:OnArrived( target, modifier, delta )
end

function SlotContainer:OnModifierChanged( modifier, delta, source )
    local widget = self:FindSlotWidget( modifier )
    if not widget then
        LOGWARN( "Missing modifier: %s", modifier )
    else
        widget:SetToolTip( modifier )
        if delta ~= nil then
            self:AnimateModifierFloater( source, widget, modifier, delta )
        end
    end
end

function SlotContainer:UpdateOldestModifier()
    local slot_widgets = self.slot_wheel:GetWidgets()
    for i, slot_widget in pairs( slot_widgets ) do
        if slot_widget:IsOldestArgument() then
            UIHelpers.PulseAlpha( slot_widget, 2, 0.75 )
        else
            UIHelpers.PulseAlpha( slot_widget, nil )
        end
    end
end

function SlotContainer:UpdatePersuasionLabels()
    local slot_widgets = self.slot_wheel:GetWidgets()
    for i, slot_widget in pairs( slot_widgets ) do
        if slot_widget.UpdatePersuasionLabel then
            slot_widget:UpdatePersuasionLabel()
        end
    end
    return self
end

local function CreateModifierWidget( self, modifier )
    local widget
    if modifier.widget_class then
        widget = modifier.widget_class( self.flipped, self.negotiator )
    -- elseif modifier.modifier_type and modifier.modifier_type == MODIFIER_TYPE.CORE then
    --     widget = Widget.NegotiationCore( self.flipped, self.negotiator )
    else
        widget = Widget.NegotiationArgument( self.flipped, self.negotiator )
    end
    widget = self.slot_wheel:AddSlotWidget( widget )
    return widget
end

function SlotContainer:OnModifierAdded( modifier, source )
    local widget = CreateModifierWidget( self, modifier )
    widget:AttachModifier( modifier )

    self:Layout()
    self:UpdateOldestModifier()

    self.slot_wheel:SnapWheelAngle( widget )
    self:AnimateModifierFloater( source, widget, modifier, modifier:GetStacks() )
    return self
end

function SlotContainer:OnArrived( target, modifier, delta )

    local sound
    if delta then
        if delta > 0 then
            sound = SoundEvents.negotiation_argument_tick_up

        elseif delta < 0 then
            sound = SoundEvents.negotiation_argument_tick_down
        end
    end

    if sound then
        target:PlayPositionalEvent( sound )
    end
    if modifier.sound then
        target:PlayPositionalEvent( modifier.sound )
    end

    target:CreateModifier( modifier )
    target:Show()
    
    if target.jumping == nil then
        target.num_jumps = 3

        -- Shake it!
        target.jumping = target:StartCoroutine( function()
            local _, y0 = target.content:GetPos()
            while target.num_jumps > 0 do

                local start_height = 120
                local end_height = 10
                local jump_height = easing.outQuad( 3 - target.num_jumps, start_height, end_height - start_height, 3 )
                local y1 = y0 + jump_height

                local start_timing = 0.15
                local end_timing = 0.05
                local jump_timing = easing.outQuad( 3 - target.num_jumps, start_timing, end_timing - start_timing, 3 )

                -- jump up
                target.content:MoveTo( nil, y1, jump_timing, easing.outQuad )
                target.content:DelayIfAnimating()

                -- jump down
                target.content:MoveTo( nil, y0, jump_timing, easing.inQuad )
                target.content:DelayIfAnimating()

                target.num_jumps = target.num_jumps - 1
            end
            target.jumping = nil

        end )
    else
        target.num_jumps = 3
    end

    if delta and delta ~= 1 then
        local txt = loc.format( "{1%+d}", delta )
        local label = target:ShowFloater( txt, 0, 180, THexColour( UICOLOURS.SUBTITLE ), 60, 0.8, 80 )
        label:Bloom( 0.2 )

        if delta > 0 then
            target:PingLabel()
        end
    end
end

function SlotContainer:OnModifierRemoved( modifier, source )
    local modifier_widget = self:FindSlotWidget( modifier )

    self:UpdateOldestModifier()
    
    if modifier_widget and not modifier_widget.is_removing then
        modifier_widget.is_removing = true
        
        modifier_widget:WaitOnCoros()
        modifier_widget:RegisterCoro()
        modifier_widget:RefreshDamagePreview()

        local sound
        if source then
            if source.is_converter then
                sound = SoundEvents.negotiation_argument_converted
            else
                sound = SoundEvents.negotiation_argument_destroyed
            end
        else
            sound = SoundEvents.negotiation_argument_expired
        end
        modifier_widget:PlayPositionalEvent( sound )

        if modifier.removed_sound then
            modifier_widget:PlayPositionalEvent( modifier.removed_sound )
        end

        local anim
        if modifier.remove_fx then
            anim = UIHelpers.PlayAnimOnce( modifier_widget, modifier.remove_fx.animfile, modifier.remove_fx.anim )
        end

        local x, y = modifier_widget:GetPos()
        local angle = math.random() * 2 * math.pi
         
        modifier_widget:ScaleTo( 1, 2, 0.5, easing.outQuad )
        modifier_widget:AlphaTo( 0, 0.5, easing.linear )
         
        x = x + 300 * math.cos( angle )
        x = y + 300 * math.sin( angle )

        modifier_widget:CurveTo( x, y, x/2, y/2, 0.5 )
        modifier_widget:RotateIndefinitely( 0.2 )
        
        self:Delay( 0.5 )

        while anim do
            if anim and anim.removed then
                anim = nil
            end
            coroutine.yield()
        end

        if modifier.OnRemoveWidget then
            modifier:OnRemoveWidget( modifier_widget, self:GetAncestor( Widget.NegotiationPanel ) )
        end

        self.slot_wheel:RemoveModifierSlot( modifier_widget )
        self:Layout()
    end
end

function SlotContainer:SetFocusable( focusable )
    local slot_widgets = self.slot_wheel:GetWidgets()
    for i, widget in ipairs( slot_widgets ) do
        widget:SetFocusable( focusable )
    end
end

function SlotContainer:RefreshForNegotiator( negotiator )
    for i, modifier in negotiator:ModifierSlots() do
        local widget = CreateModifierWidget( self, modifier )
        widget:CreateModifier( modifier )

        self.slot_wheel:SnapWheelAngle( widget )
    end
    self:Layout()
end

function SlotContainer:PingModifier( modifier )
    local widget = self:FindSlotWidget( modifier )
    if widget then
        widget:PingModifier( modifier )
    end
end

function SlotContainer:RefreshIcon( modifier )
    local widget = self:FindSlotWidget( modifier )
    if widget then
        widget:RefreshIcon( modifier )
    end
end

function SlotContainer:Layout()
    self.slot_wheel:Layout()
    return self
end

function SlotContainer:FindSlotWidget( modifier )
    local slot_widgets = self.slot_wheel:GetWidgets()

    if is_instance( modifier, Negotiation.Modifier ) then
        for i, widget in ipairs( slot_widgets ) do
            if widget:GetUserData() == modifier:GetUID() then
                return widget
            end
        end
    end
end

function SlotContainer:RefreshDamagePreview()
    for i, w in ipairs( self.slot_wheel.slot_widgets ) do
        w:RefreshDamagePreview()
    end
end

function SlotContainer:DestroyAllSlots()
    self.destroying = true
    
    for i, w in ipairs( self.slot_wheel.slot_widgets ) do
        w:WaitOnCoros()
    end

    self.slot_wheel:DestroyAllSlots()
end

function SlotContainer:SetTargetDisplay( valid )
    self.slot_wheel:SetWheelColour( valid and UICOLOURS.FOCUS or UICOLOURS.SUBTITLE )
end