require "ui/widget"


local OverlayEffect = class("Widget.OverlayEffect", Widget)

function OverlayEffect:init( texture )
    self.w, self.h = RES_X + 4, RES_Y + 4

    local texture_asset = texture or engine.asset.Texture("rgba/overlayeffect.tex")
    
    OverlayEffect._base.init(self, engine.model.PlaneModel(global_shaders.OVERLAY, texture_asset, false, self.w, self.h))

    self.model:SetDepthState(DEPTH_TEST.DISABLED, false)  
    self.model:SetTextureSampler(0, "linear","linear","linear","wrap", "wrap", false)
    self.model:SetBlendMode(BLEND_MODE.ADDITIVE)

    if texture_asset then
        assert( type(texture_asset) ~= "string" )
        self.texture = texture_asset
    end

    self.currenttime = 40.0
    self.duration = 1.0
    self:Hide()
end

function OverlayEffect:OnRemoved()
    self.model:ClearTexture(0)
    self.texture = nil
end

function OverlayEffect:__tostring()
    local texname = self.texture and self.texture:GetFileName() or "null"
    if self.name then
        return string.format("OverlayEffect[%s: %s]", self.name, texname )
    else
        return string.format("OverlayEffect[%s]", texname )
    end
end

function OverlayEffect:GetTexture()
    return self.texture
end

function OverlayEffect:SetTexture(texture_asset, sampler)
    
    sampler = sampler or 0
    
    if sampler == 0 and self.texture == texture_asset then
        return self
    end
    
    if texture_asset then
        self.model:SetTexture(sampler, texture_asset)
        self.texture = texture_asset
        self.model:SetSize(self.w, self.h)
    else
        self.model:ClearTexture(sampler)
        self.texture = nil      
    end
    
    --self.model:Rebuild()
    return self
end

function OverlayEffect:GetBoundingBox()
    local w, h = self.w, self.h
    if not w or not h then
        return 0,0,0,0
    end
    return -w/2, -h/2, w/2, h/2
end

function OverlayEffect:SetSize(dw,dh)
    if self.w ~= dw or self.h ~= dh then
        self.w, self.h = dw, dh
        if self.texture then
            self.model:SetSize(self.w, self.h)
        end
        self:InvalidateBBox()
    end

    return self
end

function OverlayEffect:GetSize()
    return self.w, self.h
end


function OverlayEffect:OnUpdate(dt)
    OverlayEffect._base.OnUpdate(self, dt)

    local walltime = math.fmod(GetWallTime(), 1000.0)

    local dt = clamp(walltime - (self.lastWallTime or 0), 0.0, 0.1);
    self.lastWallTime = walltime

    self.currenttime = self.currenttime + dt

    local time = -math.fmod(walltime, 10000.0) -- Negative so that the effect scrolls the right way by default

    local x = (self.layer1Speed or 0) * time
    local y = (self.layer2Speed or 0) * time
    local z = (self.layer3Speed or 0) * time
    local w = 0;

    local a1 = self.layer1Alpha or 1
    local a2 = self.layer2Alpha or 1
    local a3 = self.layer3Alpha or 1
    local a4 = self.layerScale or 1

    local r = self.tint_r or 1
    local g = self.tint_g or 1
    local b = self.tint_b or 1
    local a = self.tint_a or 1

    local factor = self.currenttime / self.duration

    local shutterh = self.shutterHeight or 0.5

    local shutterx = (factor * (1.0 + shutterh)) - shutterh
    local shuttery = (factor * (1.0 + shutterh))

    if not self.layerOffsets_constant then
        self.layerOffsets_constant = engine.rendering.ShaderConstantValue("OVERLAY_LAYER_OFFSETS", x, y, z, w)
        self.model:SetShaderConstant(self.layerOffsets_constant)
    else
        self.layerOffsets_constant:SetValue(x, y, z, w)
    end    

    if not self.layerAlphas_constant then
        self.layerAlphas_constant = engine.rendering.ShaderConstantValue("OVERLAY_LAYER_ALPHAS_SCALE", a1, a2, a3, a4)
        self.model:SetShaderConstant(self.layerAlphas_constant)
    else
        self.layerAlphas_constant:SetValue(a1, a2, a3, a4)
    end    

    if not self.tint_constant then
        self.tint_constant = engine.rendering.ShaderConstantValue("OVERLAY_LAYER_COLOR", r, g, b, a)
        self.model:SetShaderConstant(self.tint_constant)
    else
        self.tint_constant:SetValue(r, g, b, a)
    end      

    if not self.shutter_constant then
        self.shutter_constant = engine.rendering.ShaderConstantValue("OVERLAY_LAYER_SHUTTER", shutterx, shuttery-shutterx,0,0)
        self.model:SetShaderConstant(self.shutter_constant)
    else
        self.shutter_constant:SetValue(shutterx, shuttery-shutterx,0,0)
    end    

    if self.currenttime > self.duration then
        self:Hide()
        self:StopUpdating()
    end

end


function OverlayEffect:SetLayerSpeeds(layer1Speed, layer2Speed, layer3Speed)
    self.layer1Speed = layer1Speed
    self.layer2Speed = layer2Speed
    self.layer3Speed = layer3Speed
    return self
end

function OverlayEffect:SetLayerScale(scale)
    self.layerScale = scale
    return self
end

function OverlayEffect:SetLayerAlphas(layer1Alpha, layer2Alpha, layer3Alpha)
    self.layer1Alpha = layer1Alpha
    self.layer2Alpha = layer2Alpha
    self.layer3Alpha = layer3Alpha
    return self
end

function OverlayEffect:SetColour(colour)
    local r,g,b,a = HexColour(colour)
    self.tint_r = r;
    self.tint_g = g;
    self.tint_b = b;
    self.tint_a = a;
    return self
end


function OverlayEffect:SetBloom(b)
    self.model:SetBloom(b)
    return self
end


-- height should be between 0 and 1
function OverlayEffect:SetShutterHeight(height)
    self.shutterHeight = height
    return self
end

function OverlayEffect:SetDuration(duration)
    self.duration = math.max(duration, 0.1)
    return self
end



-- Fires the overlay effect
function OverlayEffect:Fire(colour)

    self.currenttime = 0

    if colour then
        self:SetColour(colour)
    end

    self:Show()
    self:StartUpdating()

    -- Just setting the alpha to 0 just in case we don't get an update before rendering
    if not self.tint_constant then
        self.tint_constant = engine.rendering.ShaderConstantValue("OVERLAY_LAYER_COLOR", 0,0,0,0)
        self.model:SetShaderConstant(self.tint_constant)
    else
        self.tint_constant:SetValue(0,0,0,0)
    end      
end