require "ui/widget"

local assets = {
    panel_bg = { engine.asset.Texture ("UI/convo_narration_box.tex"), 30, 30, 60, 60 },
}


local Panel = class("Widget.Panel", Widget)

function Panel:init( texture_asset, dw, dh, innerpw, innerph)

    Panel._base.init(self, engine.model.NineSliceModel())

    self.model:SetEffect(global_shaders.UI) 

    if innerpw and innerph then
        self.model:SetInteriorPercent(innerpw, innerph)
    end

    self.blocks_mouse = true

    if texture_asset then
        self:SetTexture(texture_asset)
    else
        self:SetNineSlice(assets.panel_bg)
    end

    if dw and dh then
        self:SetInnerSize(dw, dh)
    end
    --self.blocks_mouse = true
end

function Panel:OnRemoved()
    self.texture = nil
end


function Panel:__tostring()
    return string.format("Panel Widget (%2.2fx%2.2f)", self:GetSize())
end

function Panel:SetTexture(texture_asset)
    self.texture = texture_asset
    self.model:SetTexture(texture_asset)
    self:MarkTransformDirty()
    self:InvalidateBBox()
    
    return self
end

function Panel:FitChildren(padding)
    padding = padding or 0
    local xmin, ymin, xmax, ymax = math.huge, math.huge, -math.huge, -math.huge
    
    if self.children then
        for k,widget in pairs(self.children) do
            local x0, y0, x1, y1 = widget:GetBoundingBox()
            if y1 then
                x0, y0 = widget:TransformToParent( x0, y0 )
                x1, y1 = widget:TransformToParent( x1, y1 )
                xmin = math.min( x0, xmin )
                xmax = math.max( x1, xmax )
                ymin = math.min( y0, ymin )
                ymax = math.max( y1, ymax )
            end
        end
    end
    if ymax > ymin then
        self:SetInnerSize( (xmax - xmin) + padding, (ymax - ymin) + padding )
    end
    
    return self
end

function Panel:SetBloom(b)
    self.model:SetBloom(b)
    return self
end

function Panel:GetBoundingBox()
    return self.model:GetBoundingBox()
end

function Panel:SizeToWidgets( padding, ... )
    local xmin, ymin, xmax, ymax = self.parent:CalculateBoundingBox( ... )
    local hpad, vpad
    if type(padding) == "table" then
        hpad, vpad = padding[1], padding[2]
    else
        hpad, vpad = padding, padding
    end
    if ymax > ymin then
        self:SetInnerSize( (xmax - xmin) + hpad, (ymax - ymin) + vpad )
        self:SetPos( (xmin + xmax)/2, (ymin + ymax)/2 )
    end
end

function Panel:ExpandToWidgets( padding, ... )
    local w, h = self.model:GetInnerSize()
    local xmin, ymin, xmax, ymax = self.parent:CalculateBoundingBox( ... )
    xmin = math.min( xmin, self.x - w/2 )
    xmax = math.max( xmax, self.x + w/2 )
    ymin = math.min( ymin, self.y - h/2 )
    ymax = math.max( ymax, self.y + h/2 )
    if ymax > ymin then
        self:SetInnerSize( (xmax - xmin) + padding, (ymax - ymin) + padding )
        self:SetPos( (xmin + xmax)/2, (ymin + ymax)/2 )
    end
end

function Panel:ExpandToPoint( x, y )
    local w, h = self.model:GetInnerSize()
    local xmin = math.min( x, self.x - w/2 )
    local xmax = math.max( x, self.x + w/2 )
    local ymin = math.min( y, self.y - h/2 )
    local ymax = math.max( y, self.y + h/2 )
    if ymax > ymin then
        self:SetInnerSize( (xmax - xmin), (ymax - ymin) )
        self:SetPos( (xmin + xmax)/2, (ymin + ymax)/2 )
    end
end

function Panel:SetInnerSize(dw,dh)
    self.model:SetInnerSize(dw, dh)
    self:InvalidateBBox()
    return self
end

function Panel:GetInnerSize()
    return self.model:GetInnerSize()
end

function Panel:SetInnerUVs(minx, miny, maxx, maxy)
    -- print("Panel:SetInnerUVs()", minx, miny, maxx, maxy)
    self.model:SetInnerUVS(minx, miny, maxx, maxy)
    return self
end

-- really these pretty much always have to be set together
function Panel:SetNineSlice(asset)
    local texture_asset, minx, miny, maxx, maxy = table.unpack(asset)
    self.texture = texture_asset
    self.model:SetTexture(texture_asset)
    self:SetNineSliceCoords(minx, miny, maxx, maxy)
    self:InvalidateBBox()
    return self
end

-- in pixels from top left
function Panel:SetNineSliceCoords(minx, miny, maxx, maxy)
    local tx, ty = self.texture:GetSize()
    -- print("SetNineSliceCoords()", minx, miny, maxx, maxy, tx, ty)
    local uvminx = minx/tx
    local uvminy = miny/ty
    local uvmaxx = maxx/tx
    local uvmaxy = maxy/ty
    self:SetInnerUVs(uvminx, uvminy, uvmaxx, uvmaxy)
    self:InvalidateBBox()
    return self
end

-- scale = 1 by default.
function Panel:SetNineSliceBorderScale(scale)
    self.model:SetBorderScale(scale)
    self:InvalidateBBox()
    return self
end

function Panel:SetMask()
    self.model:SetColourWrite(false)
    self.model:SetEffect(global_shaders.UI_MASK)
    self:SetStencilWrite(STENCIL_MODES.SET)
    return self
end

function Panel:SetInnerPercents(dw,dh)
    self.model:SetInteriorPercent(dw, dh)
    return self
end

function Panel:SetSize(w, h)
    local current_w, current_h = self:GetSize()

    w = w or current_w
    h = h or current_h

    self.model:SetSize(w, h)
    self:MarkTransformDirty()
    self:InvalidateBBox()

    return self
end

function Panel:Expand( dw, dh )
    local w, h = self:GetSize()
    self:SetSize( w + (dw or 0), h + (dh or 0) )
    return self
end

function Panel:GetSize()
    return self.model:GetOuterSize()
end

function Panel:GetBorderSize()
    return self.model:GetBorderSize()
end

function Panel:__tostring()
    return "9 slice panel"
end

    