require "ui/widget"


local ParticleRenderer = class("Widget.ParticleRenderer", Widget)
function ParticleRenderer:init( )
    ParticleRenderer._base.init(self, engine.model.ParticleModel())
    self.model:SetBlendMode(BLEND_MODE.BLENDED)
    self.model:SetEffect(global_shaders.FX) 
    self.erode_bias = engine.rendering.ShaderConstantValue("ERODE_BIAS", 1)
    self.model:SetShaderConstant(self.erode_bias)    
    --self:InheritTransform(false)
end

------------------------------------------------------------------------------------



local ParticleEmitter = class("Widget.ParticleEmitter", Widget)

function ParticleEmitter:init( )
    ParticleEmitter._base.init(self)
    self.renderer = self:AddChild(ParticleRenderer())
end

function ParticleEmitter:OnVizChange(viz)
    if viz then
        self:StartUpdating()
    else
        self:StopUpdating()
    end
end



function ParticleEmitter:SetParams(params)
    self.renderer.model:SetEffect((params.erode_bias or 0) > 0 and global_shaders.FX_ERODE or global_shaders.FX) 
    --self.renderer.model:SetEffect(global_shaders.FX_ERODE) 
    self.params = params
    self.reference_params = deepcopy(params)
    self:SetPos(params.x or 0, params.y or 0)
    self:SetRotation(params.r or 0)

    if params.texture then
        if self.texture_name ~= params.texture then
            self.texture_name = params.texture
            self:SetTexture(engine.asset.Texture(params.texture))
        end
    end

    if params.emit_world_space then
        self.renderer:InheritTransform(false)
    else
        self.renderer:InheritTransform(true)
    end

    self.renderer.model:SetMaxParticles(params.max_particles)
    self.renderer.model:SetEmissionRate(params.emit_rate)

    if params.xbounds then
        self.renderer.model:SetXBounds(table.unpack(params.xbounds))
    else
        self.renderer.model:SetXBounds()
    end

    if params.ybounds then
        self.renderer.model:SetYBounds(table.unpack(params.ybounds))
    else
        self.renderer.model:SetYBounds()
    end

    if params.zbounds then
        self.renderer.model:SetZBounds(table.unpack(params.zbounds))
    else
        self.renderer.model:SetZBounds()
    end

    if params.lod then
        self.renderer.model:SetLevelOfDetail(params.lod)
    else
        self.renderer.model:SetLevelOfDetail(0xFF)
    end
    
    if params.spawn.box then
        self.renderer.model:SetSpawnAABB( table.unpack( params.spawn.box ) )
    end

    if params.spawn.emit_on_grid then
        self.renderer.model:SetSpawnOnGrid( true, self.params.spawn.emit_grid_rows or 1, self.params.spawn.emit_grid_colums or 1)
    else
        self.renderer.model:SetSpawnOnGrid( false )
    end

    if params.spawn.vel then
        self.renderer.model:SetSpawnVel( table.unpack( params.spawn.vel ) )
    end
    
    if params.spawn.circvel then
        self.renderer.model:SetSpawnCircVel( table.unpack( params.spawn.circvel ) )
    end
    
    if params.spawn.size then
        self.renderer.model:SetSpawnSize( table.unpack( params.spawn.size ) )
    end

    if params.spawn.rot then
        self.renderer.model:SetSpawnRot( table.unpack( params.spawn.rot ) )
    end

    if params.spawn.rotvel then
        self.renderer.model:SetSpawnRotVel( table.unpack( params.spawn.rotvel ) )
    end

    if params.spawn.ttl then
        self.renderer.model:SetSpawnTTL( table.unpack( params.spawn.ttl ) )
    end

    if params.spawn.colour then
        self.renderer.model:SetSpawnColour( HexColour(params.spawn.colour ))
    end

    if params.spawn.aspect then
        self.renderer.model:SetSpawnAspect( params.spawn.aspect )
    end

    local t = {}
    if params.curves.colour then
        for k = 1, params.curves.colour.num or 0 do
            local r,g,b,a = HexColour(params.curves.colour.data and params.curves.colour.data[k] or 0xffffffff)
            table.insert(t, r)
            table.insert(t, g)
            table.insert(t, b)
            table.insert(t, a)
        end
    end
    self.renderer.model:SetCurveColourOverTime(table.unpack(t))

    if params.curves and params.curves.scale and params.curves.scale.enabled then
        self.renderer.model:SetCurveScaleMinMax(params.curves.scale.min or 0, params.curves.scale.max or 1)
        self.renderer.model:SetCurveScaleOverTime(table.unpack(params.curves.scale.data))
    else
        self.renderer.model:SetCurveScaleOverTime()
    end

    if params.blendmode then
        self.renderer.model:SetBlendMode(params.blendmode)    
    end

    if params.bloom then
        self.renderer.model:SetBloom(params.bloom)
    end


    if params.velocity_inherit then
        self.renderer.model:SetVelocityInherit(params.velocity_inherit)
    end
    
    self.renderer.model:SetFriction(self.params.friction or 0)

    if params.curves.emission_rate and params.curves.emission_rate.enabled then
        self.renderer.model:SetCurveEmissionRate(table.unpack(self.params.curves.emission_rate.data))
        self.renderer.model:SetEmissionRateLoop(self.params.emission_rate_time, (self.params.emission_rate_loops == true) and true or false)
    else
        self.renderer.model:SetCurveEmissionRate()
    end

    if params.gravity_x and params.gravity_y and params.gravity_z then
        self.renderer.model:SetGravity(params.gravity_x, params.gravity_y, params.gravity_z)
    end
    
    self.renderer.model:SetBurst(params.burst_amt or 0, params.burst_time or 0)

    self.renderer.erode_bias:SetValue(params.erode_bias or 1)

    if params.bake_time then
        self.renderer.model:Update(params.bake_time)
    end


    return self
end

function ParticleEmitter:SetBloom(b)
    self.renderer.model:SetBloom(b)
    return self
end

function ParticleEmitter:SetLevelOfDetail(lod)
    self.renderer.model:SetLevelOfDetail(lod)
    return self
end

function ParticleEmitter:SetManualMode(val)
    self.manual_mode = val
end

function ParticleEmitter:UpdateSystem(dt)
    local x, y = self:GetWorldPos()
    local vx = self.old_x and (x - self.old_x)/dt or 0
    local vy = self.old_y and (y - self.old_y)/dt or 0
    --print (vx, vy)
    self.old_x, self.old_y = x, y

    if self.renderer.model then
        self.renderer.model:SetEmitterVelocity(vx, vy, 0)
        self.renderer.model:Update(dt)
    end
end

function ParticleEmitter:OnUpdate(dt)
    ParticleEmitter._base.OnUpdate(self, dt)
    if not self.manual_mode then
        self:UpdateSystem(dt)
    end
end

function ParticleEmitter:SetTexture(texture_asset)
    self.renderer.model:SetTexture(0, texture_asset)
    return self
end


local function ValidateParams(params)
    params.emission_rate_time = params.emission_rate_time or 5
    params.curves = params.curves or {}

    params.curves.colour = params.curves.colour or {}
    params.curves.colour.num = params.curves.colour.num or 0
    params.curves.colour.data = params.curves.colour.data or {}

    params.curves.scale = params.curves.scale or {enabled = false}
    params.curves.scale.data = params.curves.scale.data or {-1,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0,}    

    params.curves.emission_rate = params.curves.emission_rate or {enabled = false}
    params.curves.emission_rate.data = params.curves.emission_rate.data or {-1,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0,}    
end

function ParticleEmitter:Reset()
    self.renderer.model:Reset()

    if self.params.bake_time then
        self.renderer.model:Update(self.params.bake_time)
    end

end

function ParticleEmitter:RenderDebugUI( ui )
    if self.params then
        ValidateParams(self.params)

        local lod = self.params.lod or 0xFF

        if ui:Checkbox("Visible on Low Detail", CheckBits(lod, LEVEL_OF_DETAIL_LOW)) then
            self.params.lod = ToggleBits(lod, LEVEL_OF_DETAIL_LOW)
        end

        if ui:Checkbox("Visible on High Detail", CheckBits(lod, LEVEL_OF_DETAIL_HIGH) ) then
            self.params.lod = ToggleBits(lod, LEVEL_OF_DETAIL_HIGH)
        end

        local new_bake = ui:SliderFloat("BAKE TIME", self.params.bake_time or 0, 0, 5)
        if new_bake ~= self.params.bake_time then
            self.params.bake_time = new_bake
            self:Reset()
        end


        if ui:CollapsingHeader("TRANSFORM" ) then
            local ox, oy = self:GetPos()
            local rot = self:GetRotation()
            local x, y = ui:DragFloat3( "OFFSET", ox, oy, 0, -2000, 2000, 1)
            

            local rdeg = ui:DragFloat( "ROT", rot/DEGREES, -360, 360, "%0.1f", .01)

            if x then
                self:SetPos(x,y)
                self.params.x, self.params.y = x, y
            end

            if rdeg then
                self:SetRotation(rdeg*DEGREES)
                self.params.r = rdeg*DEGREES
            end
        end

        if ui:CollapsingHeader("EMISSION" ) then

            if ui:Checkbox("EMIT IN WORLD SPACE", self.params.emit_world_space) then
                self.params.emit_world_space = not self.params.emit_world_space
            end

            self.params.max_particles = ui:SliderInt("MAX PARTICLES:", self.params.max_particles, 1, 4096)
            self.params.emit_rate = ui:SliderFloat("EMIT RATE:", self.params.emit_rate, 0, 100)

            self.params.burst_amt = ui:SliderInt("BURST PARTICLES:", self.params.burst_amt or 0 , 0, 4096)
            self.params.burst_time = ui:SliderFloat("BURST TIME:", self.params.burst_time or 0, 0, 60)


            self.params.spawn.ttl[1] = ui:SliderFloat("TTL_MIN:", self.params.spawn.ttl[1] or 0 , 0, 60)
            self.params.spawn.ttl[2] = ui:SliderFloat("TTL_MAX:", self.params.spawn.ttl[2] or 0, 0, 60)
            
            if ui:CollapsingHeader("EMIT CURVE" ) then
                if ui:Checkbox("enabled##EMIT_CURVE_ENABLED", self.params.curves.emission_rate.enabled) then
                    self.params.curves.emission_rate.enabled = not self.params.curves.emission_rate.enabled
                end
                ui:CurveEditor("Curve##EMIT_CURVE", self.params.curves.emission_rate.data)
                self.params.emission_rate_time = ui:SliderFloat("Time:##EMITCURVETIME", self.params.emission_rate_time or 5 , .01, 10)
                if ui:Checkbox("Loops##EMITCURVELOOPS", self.params.emission_rate_loops) then
                    self.params.emission_rate_loops = not self.params.emission_rate_loops
                end
                ui:Separator()
            end

            ui:Separator()
        end


        if ui:CollapsingHeader("SIZE" ) then
            self.params.spawn.size[1] = ui:SliderFloat("SIZE_MIN:", self.params.spawn.size[1] or 0 , .01, 5000)
            self.params.spawn.size[2] = ui:SliderFloat("SIZE_MAX:", self.params.spawn.size[2] or 0, .01, 5000)
            ui:Separator()
            self.params.spawn.aspect = ui:SliderFloat("ASPECT RATIO:", self.params.spawn.aspect or 1 , .1, 10)
            ui:Separator()
        end

        if ui:CollapsingHeader("RENDERING" ) then
            if self.all_textures == nil then
                self.all_textures = {}
                for k, filename in pairs(engine.inst:ListFiles( 'particles/*.tex')) do
                    table.insert( self.all_textures, string.format( "particles/%s", filename ))
                end
                table.sort(self.all_textures, function (k1, k2) return string.lower(k1) < string.lower(k2) end)
            end

            local texture_idx = table.arrayfind(self.all_textures, self.params.texture)
            self.params.texture = self.all_textures[ui:Combo("Texture", texture_idx or 1, self.all_textures)]


            self.params.spawn.colour = self.params.spawn.colour or 0xffffffff
            self.params.spawn.colour = ui:ColorEdit4( "Colour",  self.params.spawn.colour )

            self.params.bloom = ui:SliderFloat("BLOOM:", self.params.bloom, 0, 5)

            self.params.erode_bias = ui:SliderFloat("LIFETIME_EROSION:", self.params.erode_bias or 1, 0, 1)

            local blendmodes = {BLEND_MODE.BLENDED, BLEND_MODE.EMISSIVE, BLEND_MODE.MULTIPLY, BLEND_MODE.ADDITIVE, BLEND_MODE.PREMULTIPLIED}
            local blendmodes_txt = {"BLENDED", "EMISSIVE", "MULTIPLY", "ADD (No!)", "PREMULT (No!)"}
            local blend_idx = table.arrayfind(blendmodes, self.params.blendmode or BLEND_MODE.BLENDED)
            self.params.blendmode = blendmodes[ui:Combo("Blend Mode", blend_idx, blendmodes_txt)]
            ui:Separator()
        end


        if ui:CollapsingHeader("VELOCITY" ) then
            self.params.spawn.vel[1] = ui:SliderFloat("XVEL_MIN:", self.params.spawn.vel[1] or 0 , -1000, 1000)
            self.params.spawn.vel[2] = ui:SliderFloat("XVEL_MAX:", self.params.spawn.vel[2] or 0, -1000, 1000)

            self.params.spawn.vel[3] = ui:SliderFloat("YVEL_MIN:", self.params.spawn.vel[3] or 0 , -1000, 1000)
            self.params.spawn.vel[4] = ui:SliderFloat("YVEL_MAX:", self.params.spawn.vel[4] or 0, -1000, 1000)
            self.params.spawn.vel[5], self.params.spawn.vel[6] = 0, 0

            self.params.velocity_inherit = ui:SliderFloat("VEL INHERIT:", self.params.velocity_inherit or 0, 0, 1)
            self.params.friction = ui:SliderFloat("FRICTION:", self.params.friction or 0, 0, 3)

            local gx,gy,gz = ui:DragFloat3( "GRAVITY", self.params.gravity_x or 0, self.params.gravity_y or 0, self.params.gravity_z or 0, -500, 500, 1.0)
            if gx then
                self.params.gravity_x, self.params.gravity_y, self.params.gravity_z = gx, gy, gz
            end
        end

        if ui:CollapsingHeader("ROTATION" ) then
            self.params.spawn.rot = self.params.spawn.rot or {0,0}
            self.params.spawn.rot[1] = ui:SliderFloat("ROT_MIN:", self.params.spawn.rot[1] or 0 , -10, 10)
            self.params.spawn.rot[2] = ui:SliderFloat("ROT_MAX:", self.params.spawn.rot[2] or 0, -10, 10)

            self.params.spawn.rotvel = self.params.spawn.rotvel or {0,0}
            self.params.spawn.rotvel[1] = ui:SliderFloat("ROTVEL_MIN:", self.params.spawn.rotvel[1] or 0 , -10, 10)
            self.params.spawn.rotvel[2] = ui:SliderFloat("ROTVEL_MAX:", self.params.spawn.rotvel[2] or 0, -10, 10)
        end



        local x1,x2, y1, y2 = table.unpack( self.params.spawn.box)
        self:DrawDebugLine(x1,y1,x1,y2,0xff00ffff)
        self:DrawDebugLine(x1,y1,x2,y1,0xff00ffff)
        self:DrawDebugLine(x2,y1,x2,y2,0xff00ffff)
        self:DrawDebugLine(x1,y2,x2,y2,0xff00ffff)


        
        self:DrawDebugLine(0, 0, 10, 0, 0xff0000ff )
        self:DrawDebugLine(0, 0, 0, 10, 0x00ff00ff)


        if ui:CollapsingHeader("EMITTER SHAPE" ) then
            local w = ui:SliderFloat("Emit W:", x2-x1, 0, 6000)    
            local h = ui:SliderFloat("Emit H:", y2-y1, 0, 6000)    
            local x = (x1+x2)/2
            local y = (y1+y2)/2
            x = ui:SliderFloat("Emit X:", x, -2000, 2000)    
            y = ui:SliderFloat("Emit Y:", y, -2000, 2000)    

            x1, x2 = x - w/2, x + w/2
            y1, y2 = y - h/2, y + h/2
            self.params.spawn.box = {x1, x2, y1, y2}

            if ui:Checkbox("Emit on a Grid", self.params.spawn.emit_on_grid) then
                self.params.spawn.emit_on_grid = not self.params.spawn.emit_on_grid
            end

            if self.params.spawn.emit_on_grid then
                self.params.spawn.emit_grid_rows = ui:SliderInt("Emit Rows:", self.params.spawn.emit_grid_rows or 10, 2, 30)
                self.params.spawn.emit_grid_colums = ui:SliderInt("Emit Columns:", self.params.spawn.emit_grid_colums or 10, 2, 30)
            end
        end

        if ui:CollapsingHeader("COLOUR CURVE" ) then
            self.params.curves.colour.num = ui:SliderInt("NUM_TO_USE##COLOURNUMTOUSE", self.params.curves.colour.num, 0, 8)
            for k = 1, self.params.curves.colour.num do
                self.params.curves.colour.data[k] = ui:ColorEdit4( "Colour##COLOURCURVE"..k,  self.params.curves.colour.data[k] or 0xffffffff )
            end
        end
        if ui:CollapsingHeader("SCALE CURVE" ) then
            if ui:Checkbox("enabled##SCALE_CURVE_ENABLED", self.params.curves.scale.enabled) then
                self.params.curves.scale.enabled = not self.params.curves.scale.enabled
            end

            if self.params.curves.scale.enabled then

                self.params.curves.scale.min = ui:SliderFloat("SCALE MIN:", self.params.curves.scale.min or 0, 0, 10)
                self.params.curves.scale.max = ui:SliderFloat("SCALE MAX:", self.params.curves.scale.max or 1, 0, 10)

                ui:CurveEditor("Curve##SCALECURVE", self.params.curves.scale.data)
            end

        end
    end   
    
    if not deepcompare(self.params, self.reference_params) then
        self:SetParams(self.params)
        self.pending = true
    end

    ui:Separator()     
    return true
end

function ParticleEmitter:IsDone()
    return self.renderer.model:IsDone()
end

function ParticleEmitter:Stop()
    self.renderer.model:SetEmissionRate(0)
end

function ParticleEmitter:Restart()
    self.renderer.model:SetEmissionRate(self.params.emit_rate)
end

function ParticleEmitter:SetEmitterBounds(x,y,w,h)
    local x1, x2 = x - w/2, x + w/2
    local y1, y2 = y - h/2, y + h/2
    self.params.spawn.box = {x1, x2, y1, y2}
    self.renderer.model:SetSpawnAABB( table.unpack( self.params.spawn.box ) ) 
    return self
end

function ParticleEmitter:SetEmitterSpawnOnGrid(enabled, rows, columns)
    self.renderer.model:SetSpawnOnGrid( enabled, rows, columns ) 
    return self
end




function ParticleEmitter:GetBoundingBox()
    local x1, x2, y1, y2 = table.unpack(self.params.spawn.box) -- For some reason the coords are stored X1, X2, and then Y1, Y2.. ?
    return x1, y1, x2, y2
end


--------------------------------------------------------------------------------------------



local ParticleSystem = class("Widget.ParticleSystem", Widget)

function ParticleSystem:init( )
    ParticleSystem._base.init(self)
    self.emitters = {}
end

function ParticleSystem:Reset()
    for k,v in ipairs(self.emitters) do
        v:Reset()
    end
end

function ParticleSystem:AddEmitter()

    local handle = Widget()
    local emitter = handle:AddChild(ParticleEmitter())
    self:AddChild(handle)
    table.insert( self.emitters, emitter)

    return emitter, handle
end

function ParticleSystem:IsDone()
    for k,v in ipairs(self.emitters) do
        if not v:IsDone() then
            return false
        end
    end

    return true
end

function ParticleSystem:Stop()
    for k,v in ipairs(self.emitters) do
        v:Stop()
    end    
end

function ParticleSystem:Restart()
    for k,v in ipairs(self.emitters) do
        v:Restart()
    end    
end


function ParticleSystem:SetParams(params)
    self.params = params
    self.reference_params = deepcopy(params)
    params.emitters = params.emitters or {}

    if #self.params.emitters > #self.emitters then
        local num_to_add = #self.params.emitters - #self.emitters
        for k = 1, num_to_add do
            self:AddEmitter()
        end

    elseif #self.emitters > #self.params.emitters then
        local num_to_cull = #self.emitters - #self.params.emitters
        for k = 1, num_to_cull do
            table.remove(self.emitters):Remove()
        end
    end

    for k,v in ipairs(params.emitters) do
        if not params.emitters[k] then
            self:AddEmitter()
        end
    end

    for k, params in ipairs(params.emitters) do
        self.emitters[k]:SetParams(params)
    end

    return self
end

function ParticleSystem:LoadParams(id)
    self.param_id = id
    local params = Content.GetParticleParams(id)
    assert(params, id)
    self:SetParams(params)
    return self
end

function ParticleSystem:GetEmitter( idx )
    return self.emitters[idx]
end

function ParticleSystem:SetManualMode(val)
    for k,emitter in ipairs(self.emitters) do
        emitter:SetManualMode(val)
    end
end

function ParticleSystem:UpdateSystem(dt)
    for k,emitter in ipairs(self.emitters) do
        emitter:UpdateSystem(dt)
    end
end


function ParticleSystem:GetBoundingBox()
    local x1, y1, x2, y2

    for k,emitter in ipairs(self.emitters) do
        local ex1, ey1, ex2, ey2 = emitter:GetBoundingBox()

        if x1 then
            x1 = math.min(x1, ex1)
            y1 = math.min(y1, ey1)
            x2 = math.max(x2, ex2)
            y2 = math.max(y2, ey2)
        else
            x1 = ex1
            y1 = ey1
            x2 = ex2
            y2 = ey2
        end
    end

    if not x1 then
        x1 = -10
        y1 = -10
        x2 = 10
        y2 = 10
    else
        -- enforce a minimum size of 20 pixels:
        local dx = math.max((20 - (x2-x1)) * 0.5, 0)
        local dy = math.max((20 - (y2-y1)) * 0.5, 0)
        x1 = x1 - dx
        y1 = y1 - dy
        x2 = x2 + dx
        y2 = y2 + dy
    end

    return x1, y1, x2, y2
end


function ParticleSystem:RenderDebugUI( ui )

    local delete_emitter = nil
    local duplicate_emitter = nil
    local move_up_emitter = nil
    local move_down_emitter = nil

    if not TheGame:GetDebug():IsDebugFlagged(DBG_FLAGS.RENDER) then
        TheGame:GetDebug():ToggleDebugFlags( DBG_FLAGS.RENDER )
    end

    if ui:Button("Add emitter") then
        table.insert( self.params.emitters, deepcopy(ParticleSystem.default_params))
        self:SetParams(self.params)
    end

    for k,emitter in ipairs(self.emitters) do

        if ui:TreeNode( "Emitter_"..k, ui.TreeFlags_DefaultOpen ) then

            local totalWidth = ui:GetContentRegionAvail();
            local buttonColumnWidth = 200
            ui:Columns(2, "", false)    -- false to turn off manual resizing
            ui:SetColumnWidth(0, totalWidth-buttonColumnWidth) -- hack to make the left column resize with the window, but the right column be fixed
            ui:PushItemWidth(-70)   -- Use full column width

            if ui:Checkbox("Visible", emitter:IsShown()) then
                emitter:SetShown(not emitter:IsShown())
            end

            ui:PopItemWidth()

            ui:NextColumn()

            if ui:Button("Delete", buttonColumnWidth) then
                delete_emitter = k
            end
            
            if ui:Button("Duplicate", buttonColumnWidth) then
                duplicate_emitter = k
            end

            if ui:Button("Move Up", buttonColumnWidth) then
                move_up_emitter = k
            end
            
            if ui:Button("Move Down", buttonColumnWidth) then
                move_down_emitter = k
            end

            ui:Columns(1)

            ui:PushItemWidth(-(buttonColumnWidth + 5) - 20) -- Use full column width
            --ui:Indent( 20 )
            ui:PushID( "EMITTER_" .. k )
            emitter:RenderDebugUI(ui)
            ui:PopID()
            --ui:Unindent( 20 )
            ui:TreePop()
        end
    end
    
    if delete_emitter then
        table.remove(self.params.emitters, delete_emitter)
        self:SetParams(self.params)
    end

    if duplicate_emitter then
        table.insert( self.params.emitters, deepcopy(self.params.emitters[duplicate_emitter]))
        self:SetParams(self.params)
    end

    if move_down_emitter then
        if move_down_emitter < #self.params.emitters -1 then
            self.params.emitters[move_down_emitter], self.params.emitters[move_down_emitter + 1] = self.params.emitters[move_down_emitter + 1], self.params.emitters[move_down_emitter]
            self:SetParams(self.params)
        end
    end

    if move_up_emitter then
        if move_up_emitter > 1 then
            self.params.emitters[move_up_emitter], self.params.emitters[move_up_emitter - 1] = self.params.emitters[move_up_emitter - 1], self.params.emitters[move_up_emitter]
            self:SetParams(self.params)
        end
    end

    return true

end


ParticleSystem.default_params = 
{

  blendmode = 1,
  bloom = 0,
  emit_rate = 4.0,
  emit_world_space = true,
  max_particles = 100,
  curves = {},
  spawn = {
    box = { 0,0,0,0 },
    colour = 0xffffff,
    size = {10,10},
    ttl = {4,4},
    vel = {0,0,10,10,0,0},
  },
  texture = "images/circle.tex"
}
