require "ui/widget"

local assets = {
    bg = engine.asset.Texture( "UI/perk_bg_back.tex"),
    earned = engine.asset.Texture( "UI/perk_bg_earned.tex"),
    unearned = engine.asset.Texture( "UI/perk_bg_normal.tex"),
}

-------------------------------------------------------------------------
-- Displays a perk, with a frame, image and backing

local PerkButton = class("Widget.PerkButton", Widget.Clickable)

function PerkButton:init(fn)
    PerkButton._base.init(self)
    self.back = self:AddChild( Widget.Image(assets.bg) )
    self.icon = self:AddChild( Widget.Image() )
        :Bloom( 0.07 )
    self.frame = self:AddChild( Widget.Image() )

    local animname = "anim/perk_border_UI_fx.zip"
    local anim = Widget.Anim()
        :SetHiddenBoundingBox(true)
        :SetBloom( 0.2 )
        :SetTintColour( UICOLOURS.QUEST )
        :SetTintAlpha( 0.7 )
        :SetPos( 0, 0 )
    anim:AddAnims( GetAnim( animname ) )
    anim:AddOverrideFile( animname )
    self.border_anim = self:AddChild( anim )
        :SetTintAlpha( 0.5 )
        :Hide()
    self.play_glow_on_create = true

    self:SetOnClickFn(fn)
    self:DisableAnimateScales()
    self:SetWidth( 168 )
end

function PerkButton:SetWidth( target_w )
    local source_w, source_h = 420, 280
    local target_h = target_w/source_w*source_h;
    self.back:SetSize( target_w, target_h )
    self.icon:SetSize( target_w, target_h )
    self.frame:SetSize( target_w, target_h )
    self.border_anim:SetScale( 1.44*target_w/source_w )
    return self
end

function PerkButton:GetPerk()
    return self.perk
end

function PerkButton:GetPerkId()
    return self.perk and self.perk.id or nil
end

function PerkButton:GetTier()
    return self.perk and self.perk.tier or 1
end

function PerkButton:Refresh(perk_id, earned)
    self.perk = Content.GetGraft(perk_id)
    self.earned = earned

    self.icon:SetSaturation( self.earned and 1 or 0.1 )
        :SetTintAlpha( self.earned and 1 or 0.25 )
    
    if self.perk then
        -- Get perk's image
        self.icon:SetTexture(self.perk.img)

        -- Check if the player has finished a game with this perk already
        self.played = TheGame:GetGameProfile():IsPerkPlayed( perk_id )
        if self.played and self.play_glow_on_create then
            self:AnimateInGlow()
        end

    end

    self.frame:SetTexture( self.played and assets.earned or assets.unearned )
        :Bloom( self.played and 0.17 or 0.05 )

    self.icon:LayoutBounds("center", "center", self.bg)
    self.frame:LayoutBounds("center", "center", self.bg)

    return self
end

-- Must be called before Refresh
function PerkButton:PrepareGlowAnimation()
    self.play_glow_on_create = false
    return self
end

function PerkButton:AnimateInGlow()
    self.border_anim:Show()
        :Play( "1", false )
        -- :Push( "idle", false )

    if self.glow_task == nil then
        self.glow_task = self:DoPeriodicTask( "glow", 5.0, 5.0 * math.random() + 5.0, function() self:AnimateInGlow() end )
    end

    return self
end

function PerkButton:IsPlayed()
    return self.played
end

function PerkButton:SetIconShown( shown )
    self.icon:SetShown( shown )
    return self
end

function PerkButton:UpdateImage()
    if ( self.hover or self.focus ) then
        if not self.earned then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.frame:Bloom( self.earned and 0.2 or 0.08 )
        self:ScaleTo( nil, 1.03, 0.2, easing.outQuad )
    else
        self.frame:Bloom( self.earned and 0.17 or 0.05 )
        self:ScaleTo( nil, 1, 0.2, easing.outQuad )
    end
end

function PerkButton:OnGainFocus()
    PerkButton._base.OnGainFocus(self)
    self:UpdateImage()
end

function PerkButton:OnLoseFocus()
    PerkButton._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PerkButton:OnGainHover()
    PerkButton._base.OnGainHover(self)
    self:UpdateImage()
end

function PerkButton:OnLoseHover()
    PerkButton._base.OnLoseHover(self)
    self:UpdateImage()
end


-------------------------------------------------------------------------
-- Used to display an empty slot, without an image, just a number counter

local PerkSlot = class("Widget.PerkSlot", Widget)

function PerkSlot:init()
    PerkSlot._base.init(self)
    local s = .3
    local w, h = 420*s, 280*s
    self.back = self:AddChild(Widget.Image(assets.bg)):SetSize(w,h)
    self.number = self:AddChild(Widget.Label("title", 50))
        :SetGlyphColour( UICOLOURS.PERK )
        :SetTintAlpha( 0.7 )
        :Bloom( 0.15 )
    self.frame = self:AddChild(Widget.Image()):SetSize(w,h)
end

function PerkSlot:Refresh(number, earned, tt)
    self.number:SetText( tostring(number) )
        :LayoutBounds( "center", "center", self.back )
        :Offset( 0, 2 )
    self.frame:SetTexture( earned and assets.earned or assets.unearned )
        :Bloom( earned and 0.17 or 0.05 )
        :LayoutBounds( "center", "center", self.back )
    if earned then
        self.number:SetGlyphColour( UICOLOURS.PERK )
    else
        self.number:SetGlyphColour( 0x4A6A82ff )
    end
    self:SetToolTip(tt)

    return self
end

-------------------------------------------------------------------------
-- Displays an empty slot's info: if it's unlocked or not, and how to unlock
-- it if locked

local PerkSlotInfo = class("Widget.PerkSlotInfo", Widget)

function PerkSlotInfo:init( w )
    PerkSlotInfo._base.init(self)
    
    self.w = w or 300
    self.slot_spacing = 10
    self.unlocked = false

    self.slot = self:AddChild( Widget.PerkSlot() )
    self.text_container = self:AddChild( Widget() )
    self.title = self.text_container:AddChild(Widget.Label( "button", FONT_SIZE.ITEM_SUBTITLE ) )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.WHITE )
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.BUTTON*LINE_HEIGHT_TEXT_MULT*0.9 )
    self.subtitle = self.text_container:AddChild(Widget.Label( "body", FONT_SIZE.ITEM_DESC ) )
        :LeftAlign()
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :SetWordWrap( true )
        :OverrideLineHeight( FONT_SIZE.BUTTON*LINE_HEIGHT_TEXT_MULT )
end

function PerkSlotInfo:IsUnlocked()
    return self.unlocked
end

function PerkSlotInfo:AnimateUnlock( number, earned, tt )
    -- Animate
    self:RunUpdater(
        Updater.Series{
            Updater.Do( function() AUDIO:PlayEvent(SoundEvents.perkslot_unlock) end ),
            Updater.Ease( function(v) self.slot:SetScale(v) end, 1, 1.1, 0.3, easing.inQuad),
            Updater.Do( function() self:Refresh( number, earned, tt ) end ),
            Updater.Ease( function(v) self.slot:SetScale(v) end, 1.2, 1, 0.2, easing.inQuad),
        }
    )
    return self
end

function PerkSlotInfo:Refresh( number, earned, tt )
    self.slot:Refresh( number, earned, tt )

    -- Set text size
    local text_w = self.w - self.slot_spacing - self.slot:GetSize()
    self.title:SetAutoSize( text_w )
    self.subtitle:SetAutoSize( text_w )

    if earned then
        self.unlocked = true
        self:SetSaturation(1)
        self.title:SetText( loc.format( LOC"UI.PERK_SCREEN.SLOT_TITLE", number ) )
    else
        self.unlocked = false
        self:SetSaturation(0.1)
        self.title:SetText( loc.format( LOC"UI.PERK_SCREEN.SLOT_TITLE_LOCKED", number ) )
    end
    self.subtitle:SetText( tt )

    self.subtitle:LayoutBounds( "left", "below", self.title )
        :Offset( 0, 4 )
    self.text_container:LayoutBounds( "after", "center", self.slot )
        :Offset( self.slot_spacing, 2 )

    return self
end

-------------------------------------------------------------------------
-- Displays a perk (PerkButton) with extra info: if it's locked or not, and how many 
-- points are required to unlock it

local PerkWidget = class("Widget.PerkWidget", Widget)

function PerkWidget:init(id, fn)
    Widget.init(self)
    self.perk_id = id
    
    self.button = self:AddChild(Widget.PerkButton(fn))
    self.cost = self:AddChild( Widget.Label( "title", 24 ) )
        :SetGlyphColour( UICOLOURS.WHITE )
    self.left_icons = self:AddChild(Widget())
    self.lock_icon = self.left_icons:AddChild( Widget.Image( engine.asset.Texture("UI/card_lock.tex" ) ) )
        :SetSize( 19, 19 )
        :Bloom( 0.05 )

    self.button:SetToolTipClass( Widget.TooltipGraft )
    self.button:ShowToolTipOnFocus( true )
end

function PerkWidget:Refresh( tier_locked )

    local earned = TheGame:GetGameProfile() and TheGame:GetGameProfile():HasPerkUnlocked(self.perk_id)
    self.button:Refresh(self.perk_id, earned)
    if tier_locked then
        -- This tier is locked
        self.lock_icon:SetSize( 60, 60 )
            :SetTintColour( UICOLOURS.WHITE )
            :Bloom( 0.05 )
            :LayoutBounds( "center", "center", self.button )
        self.cost:Hide()
        self.button:SetIconShown( false )

        self.button:SetToolTipClass( Widget.Tooltip )
        self.button:SetToolTip()
    else
        -- This tier is available
        if earned then
            self.lock_icon:SetSize( 19, 19 )
                :SetTintColour( UICOLOURS.WHITE )
                :Bloom( 0.05 )
                :SetTintAlpha( 0 )
            self.cost:Hide()
        else
            local perk = Content.GetGraft(self.perk_id)
            self.lock_icon:SetSize( 19, 19 )
                :SetTintColour( UICOLOURS.WHITE )
                :Bloom( 0.05 )
                :SetTintAlpha( 1 )
            self.cost:Show()
                :SetText(loc.format(LOC"UI.PERK_SCREEN.PERK_BALANCE", perk and perk.unlock_cost or 0))
                :LayoutBounds( "right", "below", self.button )
        end

        self.button:SetIconShown( true )
        self.left_icons:StackChildrenRow( 4 )
            :LayoutBounds( "left", "below", self.button )
            :Offset( 0, -4 )

        self.button:SetToolTipClass( Widget.TooltipGraft )
        self.button:SetToolTip(GraftInstance(self.perk_id))
    end

   
    return self
end

function PerkWidget:AnimateUnlock()
    self:StartCoroutine( function()
        -- Prevent clicks
        self.button:SetOnClickFn( nil )

        AUDIO:PlayEvent( SoundEvents.perk_unlock )

        self.lock_icon:Hide()
        UIHelpers.PingImage( self.button.icon, self.button )
    end )

    return self
end

function PerkWidget:GetPerkId()
    return self.button:GetPerkId()
end

function PerkWidget:GetPerk()
    return self.button:GetPerk()
end

function PerkWidget:IsPlayed()
    return self.button:IsPlayed()
end

function PerkWidget:PrepareGlowAnimation()
    self.button:PrepareGlowAnimation()
    return self
end

function PerkWidget:AnimateInGlow()
    self.button:AnimateInGlow()
    return self
end

function PerkWidget:SetClickSoundEvent(snd)
    self.button:SetClickSoundEvent(snd)
    return self
end

function PerkWidget:SetOnClickFn(...)
    self.button:SetOnClickFn(...)
    return self
end
