
require "ui/widget"
require "ui/widgets/plax_nameplate"
require "ui/widgets/navpath"
require "ui/widgets/locationnode"
require "ui/widgets/polyline"
require "ui/widgets/capture_image"
local filepath = require "util/filepath"

GAME_MODE = MakeBitField{ "CONVERSATION", "BATTLE"}

local NODE_TYPE = MakeEnum{ "IMAGE", "SLOT", "ANIMATION", "PARTICLES", "PATH", "LOCATION" }
local BG_PLAX_TINT_COL = 0xF0E3FFff
local BG_PLAX_SAT_LEVEL = .67
local BG_PLAX_TIME_IN = .20
local BG_PLAX_TIME_OUT = .33
local BG_PLAX_BLUR_PERCENT = 1

local BG_PLAX_VIG = .40
local BG_PLAX_VIG_DEFAULT = .35

local Plax = class("Widget.Plax", Widget)
Plax.NODE_TYPE = NODE_TYPE

function Plax:init( plax_flags )
    Plax._base.init(self)

    self.layer_roots = {}
    self.plax_flags = plax_flags or PLAX_DEFAULT_FLAGS

    self.vig = true
    
    self.xshift = 0
    self.yshift = 0

    self.labelroot = Widget()
    self.layers_root = Widget()

    self.invalid_elements = {} -- array of elements that are invalid (missing textures, etc.)
    self.element_lookup = {} -- map of element table -> widget
    self.element_id_lookup = {} -- map of element name -> element table
    self.selectable_elements = {}
    self.slot_roots = {}

    self.agent_to_slot = {}
    self.label_layer_roots = {}
    self.rollovers_by_element = {}
    self.layer_root_lookup_by_element = {}
    self.day_phase = DAY_PHASE.DAY
    self.game_mode = GAME_MODE.CONVERSATION
    self.is_windy = true

    self.all_anims = {}
    self.all_particles = {}
    
    self.location_slots = {}


    self:AddChildren{
        self.layers_root,
        self.labelroot,
    }
end


function Plax:PlayEnterSound()
    if self.entry_event then
        AUDIO:PlayEvent(self.entry_event)
    end
end

function Plax:PlayExitSound()
    if self.exit_event then
        AUDIO:PlayEvent(self.exit_event)
    end
end

function Plax:PlaySounds()
    TheGame:SetAmbience(self.ambience)
end

function Plax:GetMusicEvent()
    return self.music
end

function Plax:SetVigOn( val )
    self.vig = val
    self.img:SetVignette( self.vig and BG_PLAX_VIG_DEFAULT or 0 )
    return self
end

function Plax:OnAdded()
    self:StartUpdating()
    self:RenderToScene()
end

function Plax:ShowAspectRatios( show )
    if show and self.aspect_ratios == nil then
        self.aspect_ratios = self:AddChild( Widget():AddChildren{
                Widget.PolyLine():CreateRect(RES_X, RES_X*(3/4)):SetTintColour(0xC92BD3FF),
                Widget.PolyLine():CreateRect(RES_Y*(21/9), RES_Y):SetTintColour(0x2BC9D3FF),
                Widget.PolyLine():CreateRect(RES_X, RES_X*(10/16)):SetTintColour(0xC92BD3FF),
                Widget.PolyLine():CreateRect(RES_X, RES_Y):SetTintColour(0xC9D32BFF),
        })
        self.aspect_ratios:SetAnchors("center", "center")

    elseif not show and self.aspect_ratios then
        self.aspect_ratios:Remove()
        self.aspect_ratios = nil
    end
end

function Plax:IsPostProcess()
    return self.img ~= nil
end

function Plax:PostProcess( post )
    if self.img then
        self.img:Remove()
        self.img = nil
    end

    if not post then
        self.layers_root:InheritTransform( true )
        self.layers_root:SetRenderTerminator( false )

    else
        self.img = Widget.CaptureImage(TheGame:FE().scene, self.name)
        self.img:SetVignette( self.vig and BG_PLAX_VIG_DEFAULT or 0 )
        self:AddChild( self.img, 1 )
    end

    if self.game_mode == GAME_MODE.BATTLE and self.img == nil then
        self.layers_root:SetPos( 0, -FIGHT_PLAX_VOFFSET )
    else
        self.layers_root:SetPos( 0, 0 )
    end

    return self
end

function Plax:SetBloom( bloom )
    -- Only supported for post-process plaxes.
    if self.img then
        self.img:SetBloom( bloom )
    end
end



function Plax:GetCurrentColourName()
    return (self.day_phase == DAY_PHASE.DAY and self.layers_root.day_colour) or (self.day_phase == DAY_PHASE.NIGHT and self.layers_root.night_colour)
end


-- Resolves a colour setting name or an HSV table to a valid HSV Table (or nil)
function ResolveHSVToTable(hsvTableOrName)
    if hsvTableOrName then
        if type(hsvTableOrName) == "string" then
            return (require "content/colour_settings")[hsvTableOrName]  -- can return nil if not found
        elseif type(hsvTableOrName) == "table" and hsvTableOrName.highlights then -- somewhat lame way to check if the table is actually a HSV table
            return hsvTableOrName            
        end
    end

    return nil
end


function Plax:UpdateCurrentHSVInternal(colourtableholder, widget)
   if widget then
        if self.day_phase == DAY_PHASE.DAY then 
            widget:SetHSV(ResolveHSVToTable(colourtableholder.day_colour), colourtableholder.day_colour_ignore_parent)
        elseif self.day_phase == DAY_PHASE.NIGHT then 
            widget:SetHSV(ResolveHSVToTable(colourtableholder.night_colour), colourtableholder.night_colour_ignore_parent)
        end
    end
end

-- elemOrWidget is a plax element or a widget to update the current HSV for:
function Plax:UpdateCurrentHSVForElement(elem)
    local w = self:GetWidgetByElement(elem)

    self:UpdateCurrentHSVInternal(elem, w)
end


-- elemOrWidget is a plax element or a widget to update the current HSV for:
function Plax:UpdateCurrentHSVForWidget(w)
    self:UpdateCurrentHSVInternal(w, w)
end

function Plax:UpdateAllHSV()
    -- Update the HSV settings for all nodes and widgets:
    self:UpdateCurrentHSVForWidget(self.layers_root)
    for elem, widget in pairs(self.element_lookup) do
        self:UpdateCurrentHSVForElement(elem)
    end
end


function Plax:SetColourNames(day, night)
    assert( day == nil or type(day) == "string")
    assert( night == nil or type(night) == "string")

    self.layers_root.day_colour = day
    self.layers_root.night_colour = night
       
    self:UpdateCurrentHSVForWidget(self.layers_root)

    return self
end


function Plax:UpdateAnimation(dt)

    if self.repeating_anims then
        
        for k,v in pairs(self.repeating_anims) do
            if k:IsDone() then
                v.repeat_time = v.repeat_time - dt
                --print (v.repeat_time)
                if v.repeat_time <= 0 then
                    k:Play(v.anim, false)
                    v.repeat_time = v.min_time + math.random()*(v.max_time - v.min_time)
                end
            end
        end
    end


end

function Plax:OnUpdate(dt)
    Plax._base.OnUpdate( self, dt )

    self:UpdateAnimation(self.override_dt or dt) 

    if self.override_dt then
        for k,v in pairs(self.all_particles) do
            v:UpdateSystem(self.override_dt)
        end
    end


    if self.track_mouse then
        self.mouse_shift = engine.inst:GetShiftValue(dt, self.mouse_shift or 0)
        self:UpdatePlaxShift()
    end
    
    if self.plax_sfx_animation and self.plax_sfx_animation.t then
        if self.plax_sfx_animation and self.plax_sfx_animation.delay_time and self.plax_sfx_animation.delay_time > 0 then
            self.plax_sfx_animation.delay_time = self.plax_sfx_animation.delay_time - dt
        else

            self.plax_sfx_animation.t = self.plax_sfx_animation.t + dt
            if self.plax_sfx_animation.t >= self.plax_sfx_animation.total_t then
                self.plax_sfx_animation.t =  nil
                self:SetSaturation(self.plax_sfx_animation.sat_dest)
                self:SetTintColour(self.plax_sfx_animation.col_dest)
                if self.img then
                    self.img:SetBlur(self.plax_sfx_animation.blur_dest)
                    self.img:SetVignette(self.plax_sfx_animation.vig_dest)
                end
            else
                local sat = easing.linear(self.plax_sfx_animation.t, self.plax_sfx_animation.sat_start, self.plax_sfx_animation.sat_dest - self.plax_sfx_animation.sat_start, self.plax_sfx_animation.total_t)
                self:SetSaturation(sat)

                local r1,g1,b1,a1 = HexColour(self.plax_sfx_animation.col_start)            
                local r2,g2,b2,a2 = HexColour(self.plax_sfx_animation.col_dest)            
                
                local r = easing.linear(self.plax_sfx_animation.t, r1, r2 - r1, self.plax_sfx_animation.total_t)
                local g = easing.linear(self.plax_sfx_animation.t, g1, g2 - g1, self.plax_sfx_animation.total_t)
                local b = easing.linear(self.plax_sfx_animation.t, b1, b2 - b1, self.plax_sfx_animation.total_t)
                local a = easing.linear(self.plax_sfx_animation.t, a1, a2 - a1, self.plax_sfx_animation.total_t)
                self:SetTintColour(r, g, b, a)

                if self.img then
                    local blur = easing.linear(self.plax_sfx_animation.t, self.plax_sfx_animation.blur_start, self.plax_sfx_animation.blur_dest - self.plax_sfx_animation.blur_start, self.plax_sfx_animation.total_t)
                    self.img:SetBlur(blur)
        
                    local vig = easing.linear(self.plax_sfx_animation.t, self.plax_sfx_animation.vig_start, self.plax_sfx_animation.vig_dest - self.plax_sfx_animation.vig_start, self.plax_sfx_animation.total_t)
                    self.img:SetVignette(vig)
                end
            end
        end 
    end

    if self.script_coros then
        for i = #self.script_coros, 1, -1 do
            local coro = self.script_coros[i]
            if coro:IsDone() then
                table.remove( self.script_coros, i )
            end
        end
    end

    self:UpdateSlotLabels()
end


function Plax:TrackMouse( val )
    self.track_mouse = val
    return self
end

function Plax:Clear()
    self.layers = nil
    self.labelroot:DestroyAllChildren()
    self.layers_root:DestroyAllChildren()

    self:ClearScripts()

    table.clear(self.layer_roots)
    table.clear(self.rollovers_by_element)
    table.clear(self.slot_roots)
    table.clear(self.layer_root_lookup_by_element)

    table.clear(self.invalid_elements)
    table.clear(self.element_lookup)
    table.clear(self.selectable_elements)
    table.clear(self.agent_to_slot)
    table.clear(self.label_layer_roots)
    if self.repeating_anims then
        table.clear(self.repeating_anims)
    end
    return self
end


function Plax:SetWindy(v)

    self.is_windy = v
    for k,v in pairs(self.agent_to_slot) do
        v.character.anim:SetAutoLoopEnabled(self.is_windy)
    end

    return self
end

function Plax:SetGameMode(mode)
    self.game_mode = mode
    for elem,widget in pairs(self.element_lookup) do
        self:RefreshVisibility( elem, widget )
    end

    if self.game_mode == GAME_MODE.BATTLE and self.img == nil then
        self.layers_root:SetPos( 0, -FIGHT_PLAX_VOFFSET )
    else
        self.layers_root:SetPos( 0, 0 )
    end

    return self
end


function Plax:SetDayPhase(phase)
    self.day_phase = phase
    for elem,widget in pairs(self.element_lookup) do
        self:RefreshVisibility( elem, widget )
    end

    -- Update the HSV settings for all nodes and widgets:
    self:UpdateAllHSV()

    return self
end

function Plax:GetElementForWidget(w)
    for k,v in pairs(self.element_lookup) do
    if v == w then
            return k
        end
    end
end

function Plax:GetWidgetByElement(def)
    return self.element_lookup[def]
end

function Plax:GetElementByID( id )
    return self.element_id_lookup[ id ]
end

function Plax:CollectWidgets( fn )
    local t = {}
    for elem, w in pairs( self.element_lookup ) do
        if fn( elem, w ) then
            table.insert( t, w )
        end
    end
    return t
end

function Plax:GetInvalidElements()
    return self.invalid_elements
end

--free up this agent's slot. Leave the actual widget alone, so that the caller can make it go away gracefully
function Plax:UnhookAgentWidget(agent)
    if self.agent_to_slot[agent] then
        self.agent_to_slot[agent].occupant = nil
        self.agent_to_slot[agent].character = nil
        self.agent_to_slot[agent].agent = nil
        self.agent_to_slot[agent] = nil
    end
end

function Plax:ShowLabels( show )
    self.labelroot:SetShown( show )
end

function Plax:SetPlaxElementVisibility( elements_to_visibility )
    -- this is a map of element_id -> visible status (bool)
    self.elements_to_visibility = elements_to_visibility

    if elements_to_visibility then
        for elem_id, shown in pairs( elements_to_visibility ) do
            local elem = self.element_id_lookup[ elem_id ]
            if elem then
                self:RefreshVisibility( elem )
            end
        end
    end
end

function Plax:RefreshVisibility( elem, w )
    if type(elem) == "string" then
        elem = self.element_id_lookup[ elem ]
        if elem == nil then
            return
        end
    end

    if w == nil then
        w = self:GetWidgetByElement( elem )
    end

    local visible
    if elem.element_id and self.elements_to_visibility then
        visible = self.elements_to_visibility[ elem.element_id ]
    end
    if visible == nil then
        visible = elem.default_visibility
    end
    if visible == nil then
        visible = true
    end
 
    if (self.day_phase == DAY_PHASE.DAY) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.DAY ) then visible = false end
    if (self.day_phase == DAY_PHASE.NIGHT) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.NIGHT ) then visible = false end

    if (self.game_mode == GAME_MODE.CONVERSATION) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.CONVERSATION ) then visible = false end
    if (self.game_mode == GAME_MODE.BATTLE) and not CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.BATTLE ) then visible = false end

    if TheGame:GetLevelOfDetail() == 1 and SPECIAL_BUILD_TYPE >= 1024 then
        if CheckBits( elem.plax_flags or PLAX_DEFAULT_FLAGS, PLAX_FLAGS.EXPENSIVE ) then visible = false end
    end

    if (elem.hidden_in_editor and self.in_editor) or not visible then
        w:Hide()
    else
        w:Show()
    end
end

function Plax:ClearSlot( slot_id )
    local slot = self.slot_roots[ slot_id ]
    slot.occupant = nil
    slot.role = nil
    if slot.agent then
        self.agent_to_slot[ slot.agent ] = nil
        slot.agent = nil
        slot.character:Remove()
        slot.character = nil
    end
    if slot.feature then
        slot.img:Remove()
        slot.img = nil
        slot.feature = nil
    end
    if slot.label then
        slot.label:Remove()
        slot.label = nil
    end
    assert( slot.character == nil and slot.img == nil )
end

function Plax:Slots()
    return pairs( self.slot_roots )
end

function Plax:GetSlotOccupant( slot_id )
    return self.slot_roots[slot_id] and self.slot_roots[slot_id].occupant
end

function Plax:GetSlotLabel( slot_id )
    local slot = self.slot_roots[ slot_id ]
    return slot and slot.label
end

function Plax:SetSlotLabel( slot_id, widget )
    local slot = self.slot_roots[ slot_id ]
    if slot.label then
        slot.label:Remove()
        slot.label = nil
    end

    if widget then
        slot.label = widget
        local root = self.layer_root_lookup_by_element[ slot.data ]
        self.label_layer_roots[ root ]:AddChild( slot.label )
    end
end

function Plax:UpdateSlotLabels()

    for _,slot in pairs(self.slot_roots) do
        if slot.label then
            if slot.character then
                local x, y, z = slot.character.anim.model:SymbolToWorld( "head_swap", 0, 0, 0 )
                x, y = slot.widget.parent:TransformFromWorld( x, y )
                slot.label:SetPos(x + (slot.data.labelx or 0), y + (slot.data.labely or 0) )
            end
        end
    end
end



function Plax:GetBoundingBox()
    if self.img then
        return self.img:GetBoundingBox()
    else
        return Plax._base.GetBoundingBox( self )
    end
end

function Plax:RefreshSlot( v, slot )

    slot.occupant = v

    if is_instance( v, Agent ) then
        slot.agent = v

        local character = slot.character
        character:Refresh( v )
        character:SetPoseDef( slot.data.posedef or "standing")
        character.anim:SetAutoLoopEnabled(self.is_windy)

        if slot.data.prop_file then
            character.anim:AddOverrideFile( slot.data.prop_file )
        end
        if slot.data.prop_file2 then
            character.anim:AddOverrideFile( slot.data.prop_file2 )
        end

    else
        slot.feature = v
        slot.img:SetTexture( engine.asset.Texture( slot.feature:GetTexture(), true ))
    end
end

function Plax:AddAgentToSlotInternal( agent, slot, role )

    assert( slot.occupant == nil and slot.character == nil )

    local sx, sy = slot.widget:GetResolvedScale()
    local scale = math.max(math.abs(sx), math.abs(sy))
    --local use_small = scale <= .3333 or self.in_editor

    slot.occupant = agent
    slot.role = role
    slot.character = Widget.AnimatedCharacter( )
    self:RefreshSlot( agent, slot )

    slot.widget:AddChild( slot.character )
    self.agent_to_slot[agent] = slot

    return slot.character, slot    
end

function Plax:GetAgentSlot(agent)
    return self.agent_to_slot[agent]
end

function Plax:CanSlot( v, slot_entry )
    if is_instance( v, Agent ) then
        -- Determine whether this slot supports this occupant 'v' based on its required plax_flags.
        local slot_flags = (slot_entry.data.plax_flags or PLAX_DEFAULT_FLAGS) | self.plax_flags
        
        if not CheckAnyBits( slot_flags, v.plax_flags or PLAX_DEFAULT_FLAGS ) then
            print ("THIS IS ME", slot_flags, v.plax_flags or PLAX_DEFAULT_FLAGS, slot_entry.data.plax_flags)
            return false
        end
    end
    
    return true
end

function Plax:ShouldMoveAgentToMoreSpecificSlot(agent, role, slot_id)
    local current_slot = self.slot_roots[slot_id]
    if not current_slot then
        return true
    end

    if role and (current_slot.data.character_role == nil or current_slot.data.character_role == "") then

        for _ , slot_entry in pairs(self.slot_roots) do
            if slot_entry.occupant == nil and self:CanSlot( agent, slot_entry ) then
                
                --local open_slot = not slot_entry.data.character_role or slot_entry.data.character_role == ""
                local roles_match = role == slot_entry.data.character_role
                --this is gross,  and should go away when role gets revisited
                if slot_entry.data.character_role == CHARACTER_ROLES.PROPRIETOR and role == CHARACTER_ROLES.ASSISTANT then
                    roles_match = true
                end
                if roles_match then
                    return true --there's something better out there!
                end
            end 
        end
    end
end

function Plax:FillSlot( v, role, allow_open )

    local slot
    if type(role) == "number" and self.slot_roots[ role ] then
        slot = self.slot_roots[ role ]
    else
        local preferred_slots = {}
        local open_slots = {}
        for _ , slot_entry in pairs(self.slot_roots) do
            if slot_entry.occupant == nil and self:CanSlot( v, slot_entry ) then
                
                local open_slot = not slot_entry.data.character_role or slot_entry.data.character_role == ""
                local roles_match = role == slot_entry.data.character_role
                

                --this is gross,  and should go away when role gets revisited
                if slot_entry.data.character_role == CHARACTER_ROLES.PROPRIETOR and role == CHARACTER_ROLES.ASSISTANT then
                    roles_match = true
                end
                
                if roles_match and not open_slot then
                    table.insert(preferred_slots, slot_entry)
                end

                -- These get used as a last resort (role didn't match)
                if open_slot and allow_open then
                    table.insert(open_slots, slot_entry)
                end
            end 
        end

        if #preferred_slots > 0 then
            table.shuffle(preferred_slots)
            table.stable_sort( preferred_slots, function(a,b) return (a.data.priority or 1) < (b.data.priority or 1) end)
            slot = preferred_slots[1]
        elseif #open_slots > 0 then
            table.shuffle(open_slots)
            table.stable_sort( open_slots, function(a,b) return (a.data.priority or 1) < (b.data.priority or 1) end)
            slot = open_slots[1]
        end
    end
    
    if slot then
        if is_instance( v, Agent ) then
            return self:AddAgentToSlotInternal( v, slot, v:GetRoleAtLocation() )
        else
            error( "Unimplemented occupant type" )
        end
    end
end

function Plax:SetOverrideTime(t)
    if t then
        self.override_dt = t - (self.override_time or t)
    else
        self.override_dt = nil
    end
    self.override_time = t
    for k,v in pairs(self.all_anims) do
        v.model:SetTimeOverride(t)
    end

    for k,v in pairs(self.all_particles) do
        v:SetManualMode(t ~= nil)
    end

end


function Plax:PlayRepeatingAnim(widget, anim, min_time, max_time, initial_delay )
    self.repeating_anims = self.repeating_anims or {}
    min_time = min_time or max_time or 0
    max_time = max_time or min_time or 0
    anim = anim or "anim"

    if min_time == max_time and min_time == 0 then
        widget:Play(anim, true, initial_delay and -initial_delay)
        self.repeating_anims[widget] = nil
    else
        min_time, max_time = math.min(min_time, max_time), math.max(min_time, max_time)

        self.repeating_anims[widget] = self.repeating_anims[widget] or {}
        self.repeating_anims[widget].min_time = min_time 
        self.repeating_anims[widget].max_time = max_time 
        self.repeating_anims[widget].anim = anim
        self.repeating_anims[widget].repeat_time = min_time + math.random()*(max_time - min_time)
        
        widget:Play(anim, false, initial_delay and -initial_delay)
    end
end

function Plax:CreateElementWidget( w, v )
    w:SetTintColour( (v.tint_r or 255)/255,
        (v.tint_g or 255)/255,
        (v.tint_b or 255)/255,
        (v.tint_a or 255)/255)

    if v.bloom and v.bloom > 0 and w.SetBloom then
        w:SetBloom(tonumber(v.bloom))
    end

    w:SetPos(v.x or 0, v.y or 0)
    w:SetScale((v.flip and -1 or 1 )*(v.scale or 1), v.scale or 1)
    w:SetRotation(v.rot or 0)

    if w.SetUVSpeed then
        w:SetUVSpeed(v.uspeed or 0, v.vspeed or 0)
    end

    if v.blend_mode then
        w:SetBlendMode(v.blend_mode)
    end

    if v.tooltip then
        w:SetToolTip(v.tooltip)
    end

    self:RefreshVisibility( v, w )
end

function Plax:CreateAnimation( v )
    if v.animfile == nil then
        return
    end

    local w = Widget.Anim()
    local anim = engine.asset.Anim( filepath.normpath( v.animfile ))
    w:AddAnims(anim):SetBuild(anim)
    if v.anim_phase then
        w:SetPhase(v.anim_phase/100)
    end
    if v.anim_speed then
        w:SetSpeed(v.anim_speed/100)
    end

    if v.buildfile then
        w:SetBuild(GetAnim(v.buildfile))
    end
    
    if v.anim then
        self:PlayRepeatingAnim(w, v.anim, v.anim_repeat_delay_min, v.anim_repeat_delay_max, v.anim_initial_delay )
    end

    if v.erode then
		w:SetErodeShader()
    elseif v.blend_mode == BLEND_MODE.MULTIPLY then
        w:SetMultiplyShader()
    end
    self:CreateElementWidget( w, v )

    table.insert(self.all_anims, w)

    return w
end



function Plax:CreateParticles( v )
    if v.particles == nil then
        return
    end

    local w = Widget.ParticleSystem()
    if v.particles then
        w:LoadParams(v.particles)
    end
    self:CreateElementWidget( w, v )

    table.insert(self.all_particles, w)
    return w
end

function Plax:CreatePlaxSlot( v, slot_id )
    local w = Widget()
    local char = {widget=w,data=v, slot_id = slot_id }
    self.slot_roots[slot_id] = char
    self:CreateElementWidget( w, v )
    return w, slot_id + 1
end

function Plax:CreateImage( v, clickable )
    local tex = engine.asset.Texture( filepath.normpath( v.img ), true )
    if tex:IsDefault() then
        table.insert( self.invalid_elements, v )
    end

    local w = clickable and not self.in_editor and Widget.ImageButton(tex) or Widget.Image(tex)
    self:CreateElementWidget( w, v )

    if clickable and self.in_editor then
        w.entry_handle = w:AddChild(Widget.Image(global_images.circle, 16,16):SetTintColour(0xff0000af))
        w.entry_handle:SetPos(v.enter_x or 0, v.enter_y or 0)
    end

    return w
end

function Plax:GetPlaxID() 
    return self.plax_id
end


function Plax:SetPlaxData( plax_id )
    self.plax_id = plax_id
    engine.inst:ProfilerPush( string.format( "SetPlax( %s )", plax_id ) )

    local plaxdata = Content.GetPlaxData( plax_id )
    if not plaxdata then
        LOGWARN( "Unknown plax: '%s'", plax_id )
    else
        local changed = self.layers ~= plaxdata.layers
        self:SetLayers( plaxdata.layers )
        self:SetColourNames( plaxdata.day_colour, plaxdata.night_colour )
        if changed then
            self:RunScripts( plaxdata.scripts )
        end

        self.ambience = plaxdata.ambient_event
        self.music = plaxdata.music_event
        self.entry_event = plaxdata.entry_event
        self.exit_event = plaxdata.exit_event
    end

    engine.inst:ProfilerPop()
    return self
end

local function ConvertPlaxFlags( v )
    local flags = v.plax_flags or PLAX_DEFAULT_FLAGS
    if v.nighthide then
        flags = ClearBits(flags, PLAX_FLAGS.NIGHT)
    end

    if v.dayhide then
        flags = ClearBits(flags, PLAX_FLAGS.DAY)
    end

    v.nighthide = nil
    v.dayhide = nil

    return flags
end

function Plax:ClearScripts()
    if self.script_coros then
        while #self.script_coros > 0 do
            table.remove( self.script_coros ):Stop()
        end
    end
end

function Plax:RunScripts( scripts )
    if self.script_coros then
        self:ClearScripts()
    else
        self.script_coros = {}
    end

    for i, v in ipairs( scripts or table.empty ) do
        local filename = string.format( "scripts/content/plax_scripts/%s", v.filename )

        local ok, result = pcall( require, filename )
        if not ok then
            print( "Plax script error: ", filename, result )
        elseif type(result) ~= "function" then
            print( "Plax script did not return a function: ", filename )
        else
            local coro = self:StartCoroutine( result, self )
            table.insert( self.script_coros, coro )
        end
    end
end

function Plax:SetLayers(layers, editor)
    if layers == self.layers and not editor then
        return self
    end

    self:Clear()

    self.in_editor = editor
    self.layers = layers

    self:SetPlaxShift(0,0)

    local slot_id = 0

    for k,v in ipairs(self.layers) do

        v.plax_flags = ConvertPlaxFlags( v )

        local skip = not editor and v.guide
        if not skip then
            local layer_root = Widget()
            local layer_label_root = Widget()
            layer_root.name = v.name
            self.label_layer_roots[layer_root] = layer_label_root
            table.insert(self.layer_roots, layer_root)

            self.labelroot:AddChild(layer_label_root)
            self.layers_root:AddChild(layer_root)
            self.element_lookup[v] = layer_root
            
            if editor and v.hidden_in_editor then
                layer_root:Hide()
            end

            self:UpdateCurrentHSVForElement(v)

            if v.elements then
                for k,v in pairs(v.elements) do
                    v.plax_flags = ConvertPlaxFlags( v )

                    self.layer_root_lookup_by_element[v] = layer_root

                    if v.element_id then
                        if self.element_id_lookup[ v.element_id ] then
                            print( "WARNING: duplicate ID", v.element_id )
                        end
                        self.element_id_lookup[ v.element_id ] = v
                    end

                    local skip_guide = not editor and v.guide
                    if not skip_guide then
                        local w
                        if v.node_type == NODE_TYPE.PARTICLES then
                            w = self:CreateParticles( v )

                        elseif v.node_type == NODE_TYPE.PATH then
                            if self.in_editor then
                                w = Widget.NavPath(Graph( v.nodes, v.edges ) )
                                w:SetPos(v.x or 0, v.y or 0)
                            end
                            self.path_element = v
                        
                        elseif v.node_type == NODE_TYPE.LOCATION then
                            if self.in_editor then
                                w = Widget.LocationNode(v)
                                w:SetPos(v.x or 0, v.y or 0)
                            end
                            local all_tags = v.location_tags and v.location_tags:split(",")
                            v.location_tag_list = all_tags or {}
                            table.insert(self.location_slots, v)

                        elseif v.node_type == NODE_TYPE.ANIMATION then
                            w = self:CreateAnimation( v )

                        elseif v.node_type == NODE_TYPE.SLOT then
                            w, slot_id = self:CreatePlaxSlot( v, slot_id )

                        elseif v.node_type == NODE_TYPE.IMAGE then
                            w = self:CreateImage( v )
                        end

                        if w then
                            self.element_lookup[v] = w

                            layer_root:AddChild(w)

                            if self.in_editor then
                                table.insert(self.selectable_elements, {element = v, widget = w})
                            end

                            self:UpdateCurrentHSVForElement(v)
                        end
                    end
                end

            end
        end
    end 
    
    return self
end

function Plax:GetPathElement()
    return self.path_element
end

function Plax:GetPlaxShift()
    return self.xshift, self.yshift
end

function Plax:UpdatePlaxShift()

    local xshift, yshift = self.xshift + (self.mouse_shift or 0), self.yshift

    if not TheGame:ScreenSwayEnabled() then
        xshift, yshift = 0, 0 
    end


    for k,root in ipairs(self.layer_roots) do
        local scale = self.layers[k].plax or 0
        local dx, dy = -(xshift or 0) * scale, -(yshift or 0) * scale
        
        root:SetPos( dx, dy )
        local label_root = self.label_layer_roots[root]
        if label_root then
            label_root:SetPos( dx, dy )
        end
    end
end

function Plax:SetPlaxShift(x,y)
    self.xshift = x or 0
    self.yshift = y or 0
    self:UpdatePlaxShift()

    return self
end

function Plax:GetHitsUnderPoint(t,x,y)
    table.clear(t)
    
    for k,v in ipairs(self.selectable_elements) do        
        if not v.element.lock and v.widget.shown then
            if v.widget:CheckHit(x,y) then
                table.insert(t, v)
            end
        end
    end
end

function Plax:CheckMouseHover(x,y,trace)
    
    if trace then
        trace:PushWidget( self )
    end

    local is_blocked = false
    if self.slot_roots then
        for _,slot in pairs(self.slot_roots) do
            if slot and slot.agent then 
                if slot.widget and slot.widget.shown then
                    local blocked, hover = slot.widget:CheckMouseHover(x,y,trace)
                    if blocked and hover then
                        if trace then
                            trace:PopWidget( "true : slot blocked and hover" )
                        end
                        return blocked, hover
                    elseif blocked then
                        is_blocked = true
                        break
                    end
                end
                
                if slot.label and slot.label.shown then
                    local blocked, hover = slot.label:CheckMouseHover(x,y,trace)
                    if blocked and hover then
                        if trace then
                            trace:PopWidget( "true : slot blocked and hover" )
                        end
                        return blocked, hover
                    elseif blocked then
                        is_blocked = true
                        break
                    end
                end
            end
        end
    end

    if trace then
        trace:PopWidget( string.format( "%s: return", tostring(is_blocked)))
    end
    return is_blocked
end
--]]

function Plax:RenderToScene()
    if self.img == nil then
        return
    end

    self.layers_root:InheritTransform(false)
    self.layers_root:SetRenderTerminator(false)

    local w, h = TheGame:FE():GetScreenDims()

    if self.img.blur_amt and self.img.blur_amt > 0 and SPECIAL_BUILD_TYPE < 1024 then
        if TheGame:GetLevelOfDetail() == 1 then -- low level of detail
            w = w * 0.6667
            h = h * 0.6667
        end
    end

    local aspect = w/h
    local author_aspect = RES_X/RES_Y

    local plaxw, plaxh

    if aspect >= author_aspect then
        -- Wide aspect.
        plaxw = RES_Y * math.min(aspect, 21/9)
        plaxh = RES_Y
    else
        -- Tall aspect.
        plaxw = RES_X
        plaxh = RES_X / math.max(aspect, 4/3)
    end

    local dy = (self.game_mode == GAME_MODE.BATTLE) and FIGHT_PLAX_VOFFSET or 0


    self.img:SetSize( plaxw, plaxh, w, h )
    self.img:RenderToScene( self.layers_root, -plaxw/2, -plaxh/2 + dy, plaxw/2, plaxh/2 + dy)

    self.layers_root:InheritTransform(true)
    self.layers_root:SetRenderTerminator(true)
end

function Plax:EditorRenderToScene()
    if self.img == nil then
        return
    end

    self.layers_root:InheritTransform(false)
    self.layers_root:SetRenderTerminator(false)

    local w, h = RES_Y*(21/9), RES_Y
    self.img:InheritTransform( true )
    self.img:SetSize( w, h, w, h )
    self.img:RenderToScene( self.layers_root, -w/2, -h/2, w/2, h/2 )

    self.layers_root:InheritTransform(true)
    self.layers_root:SetRenderTerminator(true)
end

function Plax:RenderDebugUI( ui, panel )
    ui:Columns( 4 )
    ui:Text( "Slot ID / Name" )
    ui:NextColumn()

    ui:Text( "Role" )
    ui:NextColumn()

    ui:Text( "Occupant" )
    ui:NextColumn()

    ui:Text( "Slot Table" )
    ui:NextColumn()

    for slot_id, slot in pairs( self.slot_roots ) do
        if slot.data.name then
            ui:Text( string.format( "%d [%s]", slot_id, slot.data.name ))
        else
            ui:Text( tostring(slot_id) )
        end
        ui:NextColumn()

        ui:Text( tostring(slot.data.character_role))
        ui:NextColumn()

        ui:TextColored( 0xFF77FFFF, tostring(slot.agent or slot.feature or "EMPTY") )
        ui:NextColumn()

        panel:AppendTable( ui, slot )
        ui:NextColumn()
    end
    ui:Columns( 1 )

    ui:Separator()

    panel:AppendKeyValues( ui, self )

end



function Plax:StopBlurEffect()
    if self.blurred then
        self.plax_sfx_animation = self.plax_sfx_animation or {}
        self.plax_sfx_animation.t = 0
        self.plax_sfx_animation.total_t = BG_PLAX_TIME_OUT
        
        self.plax_sfx_animation.sat_start = BG_PLAX_SAT_LEVEL
        self.plax_sfx_animation.sat_dest = 1
        self.plax_sfx_animation.col_start = BG_PLAX_TINT_COL
        self.plax_sfx_animation.col_dest = 0xffffffff
        self.plax_sfx_animation.vig_start = BG_PLAX_VIG
        self.plax_sfx_animation.vig_dest = BG_PLAX_VIG_DEFAULT
        self.plax_sfx_animation.blur_start = BG_PLAX_BLUR_PERCENT
        self.plax_sfx_animation.blur_dest = 0
        self.blurred = false
    end

end


function Plax:DoBlurEffect(delay_time)
    if not self.blurred then
        self.plax_sfx_animation = self.plax_sfx_animation or {}
        self.plax_sfx_animation.delay_time = delay_time or 0
        self.plax_sfx_animation.t = 0
        self.plax_sfx_animation.total_t = BG_PLAX_TIME_IN
        self.plax_sfx_animation.sat_start = 1
        self.plax_sfx_animation.sat_dest = BG_PLAX_SAT_LEVEL
        self.plax_sfx_animation.col_start = 0xffffffff
        self.plax_sfx_animation.col_dest = BG_PLAX_TINT_COL
        self.plax_sfx_animation.vig_start = BG_PLAX_VIG_DEFAULT
        self.plax_sfx_animation.vig_dest = BG_PLAX_VIG

        self.plax_sfx_animation.blur_start = 0
        self.plax_sfx_animation.blur_dest = BG_PLAX_BLUR_PERCENT
        self.blurred = true
    end
end
