require "ui/widgets/label"

local ProgressBar = class("Widget.ProgressBar", Widget)

function ProgressBar:init( w, h )
	ProgressBar._base.init(self)
    self.seg_div_colour = UICOLOURS.PROGRESS_BAR_OUTLINE
	self.outline_width = 0
    w =  (w or 140)
	h = (h or 32)

    self.root = self:AddChild( Widget() )
    self.outline = Widget.SolidBox( self.inner_w, self.inner_h):SetColour(self.seg_div_colour)
    self.container = Widget()
    self.bar_bg = Widget.SolidBox( self.inner_w, self.inner_h, 0,0,0,1 )
    self.bar = Widget.SolidBox( self.inner_w, self.inner_h )

    self.bar_colour_stages = {}
    self.bar_progress = 1.0

    self.root:AddChildren{
        self.outline,
        self.container:AddChildren{
            self.bar_bg,
            self.bar
        }
    }

    self:SetSize(w,h)
end

function ProgressBar:OrientH()
    self.orientation = "horizontal"
    self:SetProgress( self.bar_progress )
    return self
end

function ProgressBar:OrientV()
    self.orientation = "vertical"
    self:SetProgress( self.bar_progress )
    return self
end

function ProgressBar:SetOutline(sz)
    self.outline_width = sz
    self:Layout()
    return self
end

function ProgressBar:Layout()
    self.inner_w = self.w - self.outline_width*2
    self.inner_h = self.h - self.outline_width*2
    self.outline:SetSize(self.inner_w+self.outline_width*2, self.inner_h+self.outline_width*2)
    self.bar_bg:SetSize(self.inner_w, self.inner_h)

    if self.orientation == "horizontal" then
        self.bar:SetSize(self.bar_progress*self.inner_w, self.inner_h)
        self.bar:SetPos(-(1-self.bar_progress)*self.inner_w/2, 0)
    else
        self.bar:SetSize(self.inner_w, self.bar_progress*self.inner_h)
        self.bar:SetPos(0, -(1-self.bar_progress)*self.inner_h/2)
    end
end

function ProgressBar:SetSize(w, h)
    self.w = w
    self.h = h
    self:Layout() 
    return self
end

function ProgressBar:SetWidth( w )
    self.w = w
    self:Layout() 
    return self
end

function ProgressBar:SetBGColour(r,g,b,a)
    self.bar_bg:SetColour(r,g,b,a)
    return self
end

function ProgressBar:SetColourStages( stages )
    for i, stage in ipairs( stages ) do
        assert( stage.colour )
        assert( stage.trigger )
    end

    -- a table of stages for the bar_bg to use
    self.bar_colour_stages = stages
    return self
end

function ProgressBar:SetOutlineColour(r,g,b,a)
    if r and g and b then
        self.seg_div_colour = MakeHexColour( r, g, b, a or 1 )
    elseif r and not g and not b then
        self.seg_div_colour = r
    end
    
    self.outline:SetColour( self.seg_div_colour )

    return self
end

function ProgressBar:SetBarBloom( bloom )
    self.bar:SetBloom( bloom )
    return self
end

function ProgressBar:SetBarColour( r,g,b,a )
    self.bar:SetColour(r,g,b,a)
    return self
end

function ProgressBar:GetBar()
    return self.bar
end

function ProgressBar:GetProgress()
    return self.bar_progress
end

function ProgressBar:SetProgress( progress )
    local bar = self.bar

    self.bar_progress = clamp( progress, 0, 1 )

    if self.orientation == "horizontal" then
        bar:SetSize(self.bar_progress*self.inner_w, self.inner_h)
        bar:SetPos(-(1-self.bar_progress)*self.inner_w/2, 0)
    else
        bar:SetSize(self.inner_w, self.bar_progress*self.inner_h)
        bar:SetPos(0, -(1-self.bar_progress)*self.inner_h/2)
    end

    if self.bar_colour_stages then
        for i, stage in ipairs(self.bar_colour_stages) do
            if self.bar_progress >= stage.trigger then
                self:SetBarColour(stage.colour)
                break
            end
        end
    end

    if self.bar_highlight_values then
        local vals = self.bar_highlight_values
        self:HighlightRange(vals.low, vals.high, vals.col, vals.col2)
    end

    return self
end

function ProgressBar:GetHighlightWidget()
    if not self.bar_highlight then
        self.bar_highlight = Widget.SolidBox():Hide()
        self.bar:AddChild( self.bar_highlight )
    end

    return self.bar_highlight
end

function ProgressBar:GetFillerWidget()
    if not self.bar_filler then
        self.bar_filler = Widget.SolidBox():Hide()
        self.bar:AddChild(self.bar_filler)
    end

    return self.bar_filler
end

function ProgressBar:ClearHighlight()
    self:GetHighlightWidget():Hide()
    self:GetFillerWidget():Hide()
    self.bar_highlight_values = nil
    return self
end

function ProgressBar:HighlightRange(low, high, col, col2)
    if not self.anim_progress then
        low = math.max(0, low)
        high = math.min(1, high)
        local w = (high-low)*self.inner_w
        local h = (high-low)*self.inner_h
        local highlight = self:GetHighlightWidget()
        local filler = self:GetFillerWidget()

        if self.orientation == "horizontal" then
            highlight:SetSize(w, self.inner_h):SetTintColour(col or 0xffffffff):Show()
            highlight:LayoutBounds("left", "center", self.bar):Offset(low * self.inner_w, 0)
        else
            highlight:SetSize(self.inner_w, h):SetTintColour(col or 0xffffffff):Show()
            highlight:LayoutBounds("center", "bottom", self.bar):Offset(0, low*self.inner_h)
        end

        if self.bar_progress > high then
            -- need to cover up the bar a bit
            if self.orientation == "horizontal" then
                local w = self.inner_w * (self.bar_progress - high)
                filler:SetSize(w, self.inner_h):SetTintColour(col2 or 0xffffffff):Show()
                filler:LayoutBounds("after", "center", highlight)
            else
                local h = self.inner_h*math.abs(self.bar_progress-high)
                filler:SetSize(self.inner_w, h):SetTintColour(col2 or 0xffffffff):Show()
                filler:LayoutBounds("center", "above", highlight)
            end

        elseif self.bar_progress < low then
            -- need to connect highlight & bar end
            if self.orientation == "horizontal" then
                local w = self.inner_w*math.abs(low - self.bar_progress)
                filler:SetSize(w, self.inner_h):SetTintColour(col2 or 0xffffffff):Show()
                filler:LayoutBounds("before", "center", highlight)
            else
                local h = self.inner_h*math.abs(low - self.bar_progress)
                filler:SetSize(self.inner_w, h):SetTintColour(col2 or 0xffffffff):Show()
                filler:LayoutBounds("center", "below", highlight)
            end

        elseif self.bar_progress > low and self.bar_progress < high then
            -- I guess the highlight range will cover the end of the bar... don't need a filler
            filler:Hide()
        end

        self.bar_highlight_values = nil
    else
        self.bar_highlight_values =
        {
            low = low,
            high = high,
            col = col,
            col2 = col2,
        }
    end

    return self
end


function ProgressBar:ProgressTo( finish, duration, ease, fn )
    -- only start if not already updating I guess?
    if self.anim_progress then
        self:SetProgress(self.anim_progress.finish)
        self:StopUpdating()
    end

    self:StartUpdating()

    self.anim_progress=
    {
        start = self.bar_progress,
        finish = finish,
        total_t = duration,
        t = 0,
        ease = ease or easing.linear,
        fn = fn,
    }

    return self
end

function ProgressBar:IsAnimating()
    return self.anim_progress ~= nil or ProgressBar._base.IsAnimating( self )
end

function ProgressBar:OnUpdate(dt)
    ProgressBar._base.OnUpdate( self, dt )

    if self.anim_progress then
        local progress = self.anim_progress
        progress.t = progress.t + dt

        if progress.t > progress.total_t then
            self:StopUpdating()
            self:SetProgress( progress.finish )
            self.anim_progress = nil

            local fn = progress.fn
            if fn then fn() end
        else
            local t = progress.ease( progress.t, progress.start, progress.finish - progress.start, progress.total_t )
            self:SetProgress(t)
        end
    end
end

function ProgressBar:GetBoundingBox()
    local w = self.inner_w + self.outline_width*2
    local h = self.inner_h + self.outline_width*2

    return -w/2, -h/2, w/2, h/2
end
