require "ui/widget"

local Screen = class("Widget.Screen", Widget)
Screen._start_table_size = 64

function Screen:init()
    Screen._base.init(self)
    self.hotkeys = {}
    self.is_screen = true
    self.can_focus = true
    self:SetToolTipLayoutFn( function( w, tooltip_widget ) self:LayoutToolTip( tooltip_widget ) end )
end

Screen.CONTROL_MAP = 
{
    {
        hint = function(self, left, right)
            table.insert(left, LOC"UI.CONTROLS.NAV")
        end
    },

    {
        control = Controls.Digital.MENU_PREV, 
        test = function(self)
            return self.focus_ring and #self.focus_ring > 0
        end,
        fn = function(self)
            if self:PrevFocusGroup() then return true end
        end,
        hint = function(self, left, right)
            table.insert(left, loc.format(LOC"UI.CONTROLS.FOCUS_PREV", Controls.Digital.MENU_PREV))
        end
    },
    {
        control = Controls.Digital.MENU_NEXT, 
        test = function(self)
            return self.focus_ring and #self.focus_ring > 0
        end,
        fn = function(self)
            if self:NextFocusGroup() then return true end
        end,
        hint = function(self, left, right)
            table.insert(left, loc.format(LOC"UI.CONTROLS.FOCUS_NEXT", Controls.Digital.MENU_NEXT))
        end
    },
}
    
function Screen:SetFocusRing(ring)
    self.focus_ring = ring
end

function Screen:ClearRobotOptions()
    if self.all_robot_options then
        table.clear(self.all_robot_options)
    end
end

function Screen:AddRobotOption(widget)
    if TheGame:GetLocalSettings().ROBOTICS then
        self.all_robot_options = self.all_robot_options or {}
        table.insert(self.all_robot_options, widget)
    end
end

function Screen:SetRobotTimer( duration )
    self.robotics_timer = duration
end

function Screen:NextFocusGroup()
    
    if self.focus_ring then

        for k,v in ipairs(self.focus_ring) do

            if v.focus then
                local idx = k == #self.focus_ring and 1 or k+1
                
                self.focus_ring[idx]:SetFocus()
                return true
            end
        end
    end
end

function Screen:PrevFocusGroup()
    
    if self.focus_ring then
        for k,v in ipairs(self.focus_ring) do
            if v.focus then
                local idx = k == 1 and #self.focus_ring or k-1
                
                self.focus_ring[idx]:SetFocus()
                return true
            end
        end
    end
end

function Screen:IsPopup()
    return self.is_popup
end

function Screen:IsOverlay()
    return self.is_overlay
end

function Screen:OnScreenResize(w, h)
    Screen._base.OnScreenResize(self, w, h)
end

function Screen:IsOpaque()
    --special case for a popup with a fade-in underlay
    if self.is_popup and self.underlay and ((self.underlay.tint_a or 1) >= 1) and ((self.tint_a or 1)>= 1) then
        return true
    end
    return not self.is_popup and not self.is_overlay and not self.is_transparent
end

function Screen:SinksInput()
    return not self.is_overlay or self.sinks_input
end

function Screen:IsOnStack()
    if self.fe and self.fe:FindScreen( self ) ~= nil then
        return true
    else
        return false
    end
end

function Screen:FadeOut()
    self:GetFE():FadeScreen( function()
        if self:IsOnStack() then
            self:GetFE():PopScreen( self )
        end
    end, self, 0.3 )
end



function Screen:OnUpdate(dt)
    Screen._base.OnUpdate(self, dt)
    if TheGame:GetLocalSettings().ROBOTICS then
        if self.robotics_timer then
            self.robotics_timer = self.robotics_timer - dt
        end
        if (self.robotics_timer or 0) <= 0 then
            self:DoRoboticsChoice()
        end
    end
end

function Screen:DoRoboticsChoice()
    local settings = TheGame:GetRoboticsSettings()
    
    if not self.fe:IsFading() then
        local opt = self:_PickRobotOption() 
        
        if opt then
            table.clear( self.all_robot_options )
            self.robotics_timer = nil
            opt:Click()
            return opt
        end
    end
end

function Screen:_PickRobotOption()
    if self.HarvestRobotOptions then
        self:ClearRobotOptions()
        self:HarvestRobotOptions()
    end

    if not self.all_robot_options or #self.all_robot_options == 0 then
        return
    end

    local candidates = {}
    for k,v in ipairs(self.all_robot_options) do
        if v:IsEnabled() then
            table.insert(candidates, v)
        end
    end

    table.shuffle(candidates)
    table.stable_sort(candidates, function(a,b) 
        return self:_GetOptionWeight(a) < self:_GetOptionWeight(b)
    end)
    local opt = candidates[#candidates]


    print ("ROBOTICS CHOICE:", tostring(self))
    for k,v in ipairs(candidates) do
        print (k == #candidates and "  ****" or "      ", tostring(v), self:_GetOptionWeight(v))
    end

    return opt
end

function Screen:_GetOptionWeight(opt)
    local settings = TheGame:GetRoboticsSettings()
    local score = 0
    if settings.choice_weights and opt.robot_tags then
        for tag, weight in pairs(settings.choice_weights) do
            if table.arrayfind(opt.robot_tags, tag) then
                score = score + weight
            end
        end
    end

    return score
end



function Screen:StartUpdating()
    -- screens are always updating
end

function Screen:StopUpdating()
    -- screens dont stop updating
end

function Screen:OnFocusMove( dir )
    -- override this to handle focus movement that hasn't gone anywhere
end

function Screen:LayoutToolTip( tooltip_widget )
    local mx, my = TheGame:GetInput():GetMousePos()
    tooltip_widget:SetRegistration("left", "top")
    tooltip_widget:SetPos(mx + 5, my + 5)

    self:GetFE():ConstrainToolTipPos( tooltip_widget )
end

function Screen:SetupUnderlay( fade )
    if self.underlay then
        return
    end

    self.underlay = Widget.Image( engine.asset.Texture( "splash/underlay.tex" ))
    self.underlay:SetAnchors( "fill", "fill" )
    self:AddChild( self.underlay, 1 )
    if fade then
        self.underlay:SetTintAlpha( 0 )
        self.underlay:AlphaTo( 1.0, 0.3, easing.outQuad )
    end
end

function Screen:SetSounds( open_sound, close_sound )
    if open_sound then self.open_sound = open_sound end
    if close_sound then self.close_sound = close_sound end
end

function Screen:SetTabLoop(tab_loop)
    for k,v in ipairs(tab_loop) do
        if k == 1 then
            v:SetFocusDir("prev", tab_loop[#tab_loop])
        else
            v:SetFocusDir("prev", tab_loop[k-1])
        end
        
        if k == #tab_loop then
            v:SetFocusDir("next", tab_loop[1])
        else
            v:SetFocusDir("next", tab_loop[k+1])
        end
    end
end

function Screen:RegisterHotkeyButton(control, widget)
    assert( is_instance( widget, Widget ) and (widget.HotkeyUp or widget.HotkeyDown))
    assert( self.hotkeys[control] == nil ) -- Dupe hotkey?
    self.hotkeys[control] = widget
end

function Screen:UnregisterHotkey(control)
    self.hotkeys[control] = nil
end

function Screen:UnregisterHotkeyButton(control, button)
    assert( self.hotkeys[control] == button )
    self.hotkeys[control] = nil
end


function Screen:HandleControlUp(control_set, device)
    if device == "key" then
        for i, control in ipairs( control_set.controls ) do
            local widget = self.hotkeys[control]
            if widget and widget.enabled and widget:IsVisible() and widget.HotkeyUp then
                if widget.removed then
                    LOGWARN("Trying to call hotkey '%s' on removed widget '%s', (did you forget to unregister a hotkey?)", control_set:Dump(), tostring(widget))
                else
                    return widget:HotkeyUp( control )
                end
            end
        end
    end
end

function Screen:HandleControlDown(control_set, device)
    if device == "key" then
        for i, control in ipairs( control_set.controls ) do
            local widget = self.hotkeys[control]
            if widget and widget.enabled and widget:IsVisible() and widget.HotkeyDown then
                if widget.removed then
                    LOGWARN("Trying to call hotkey '%s' on removed widget '%s', (did you forget to unregister a hotkey?)", control_set:Dump(), tostring(widget))
                else
                    return widget:HotkeyDown( control )
                end
            end
        end
    end
end

function Screen:OnOpen()
    if self.open_sound then
        AUDIO:PlayEvent( self.open_sound )
    end

    local mix_event = self.screen_mix_event

    if mix_event then
        self.screen_mix_event_sound = AUDIO:CreateEventInstance(mix_event)
        self.screen_mix_event_sound:Start()
    end

end

function Screen:SetMusicEvent( music_event )
    self.music_event = music_event
end

function Screen:GetMusicEvent()
    return self.music_event
end

function Screen:GetDefaultFocus()
    return self.default_widget
end

function Screen:OnClose()
    if self.close_sound then
        AUDIO:PlayEvent( self.close_sound )
    end

    if self.screen_mix_event_sound then
        self.screen_mix_event_sound:Stop()
        self.screen_mix_event_sound:Release()
        self.screen_mix_event_sound = nil
    end

end

function Screen:SetDefaultFocus()
    if self.fe then
        if self.fe:GetControlMode() == CONTROL_MODE.GAMEPAD then
            local focus_w = self:GetDefaultFocus()
            if focus_w then 
                focus_w:SetFocus()
            end
        end
    end
end

function Screen:WantsReticuleMode()
    return false
end

function Screen:TestReticule(wx, wy)
    local mouse_blocked, hover_widget = self:CheckMouseHover(wx, wy)
    return hover_widget
end

function Screen:OnGainTop()
    self.top = true
    self:SetDefaultFocus()

    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )

    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )

    if self.RefreshStreamerMode then
        self:RefreshStreamerMode()
    end
end

function Screen:OnLoseTop()
    self.top = false
    self._last_focus_widget = self:GetFE():GetFocusWidget()
    self:GetFE():ClearFocusWidget()
end

function Screen:IsOnTop()
    return self.top
end


function Screen:RefreshStreamerMode()
end

-- Returns the controller-specific options (table of strings) to show on the bottom UI bar
