------------------------------------------------------------------------------------------
-- The sidebar showing the various outfits

local SidebarOutfits = class("Widget.SidebarOutfits", Widget.SidebarBlock)

function SidebarOutfits:init( width )
    SidebarOutfits._base.init(self,width)
end

function SidebarOutfits:Refresh( character_data, act_data, current_anim )
    if not character_data then return end

	self.current_anim = current_anim

	SidebarOutfits._base.Refresh( self )

	self.title:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_OUTFITS_TITLE" ):SetGlyphColour( UICOLOURS.ACCENT )
    self.subtitle:Hide()
	self.text:SetText( loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_OUTFITS_TEXT", loc.toupper( character_data:GetLocalizedName() ) ))

	-- Remove old outfit buttons
	self.scrolling_contents:DestroyAllChildren()
    
	-- Add new ones
	local act_id = character_data.id
	local character_id = character_data.data.player_agent
	local selected_outfit_id = TheGame:GetGameProfile():GetSelectedOutfit( act_id )
    for k, outfit_data in ipairs( Content.GetOutfitsForCharacter( character_id ) ) do
    	local is_unlocked = outfit_data.unlocked or TheGame:GetGameProfile():HasUnlock( outfit_data.id )
    	selected_outfit_id = selected_outfit_id or outfit_data.id -- If there wasn't a previously selected outfit, make the first one it
    	local is_selected = outfit_data.id == selected_outfit_id
    	local is_seen = TheGame:GetGameProfile():IsOutfitSeen( act_id, outfit_data.id )
    	local button = self.scrolling_contents:AddChild( Widget.OutfitButton( self.sidebar_width, act_id, outfit_data, not is_unlocked, is_selected, is_seen, function( button, outfit_id ) self:OnOutfitSelect( button, outfit_id ) end ) )
        if is_selected then self.last_focused_button = button end
    end

	-- Layout
    self:Layout()
end

function SidebarOutfits:Layout()

    self.scrolling_contents:StackChildren( 10, 0 )

    SidebarOutfits._base.Layout( self )
end

function SidebarOutfits:OnOutfitSelect( button, outfit_id )
	-- Unselect last button
	for i, widget in ipairs( self.scrolling_contents.children ) do
		if widget.selected then
			-- Tell the selected button to unselect itself, and to run the function when done
			widget:AnimateUnselect( function()
				-- When done, tell the new button to select itself, and go back when done
				button:AnimateSelect( function()
					-- Update outfit on the anim
					self.current_anim:ChangeOutfit( outfit_id )
					-- Go back
					-- self.back_fn()
				end )
			end )
		end
	end
end

function SidebarOutfits:AnimateIn()
	SidebarOutfits._base.AnimateIn( self )

    -- Prepare button fade-in sequence
    local timing = 0.2
    local spacing = 30
    local button_adder = Updater.Parallel()
    for i, widget in ipairs( self.scrolling_contents.children ) do
    	local pos_x, pos_y = widget:GetPos()
        button_adder:Add( Updater.Ease( function(v) widget:SetPos(nil,v) end, pos_y-spacing*i, pos_y, timing, easing.outQuad ) )
    end
    button_adder:Add( Updater.Do( function()
    	for i, widget in ipairs( self.scrolling_contents.children ) do
    		widget.enabled = true
    	end
	end ) )
    self:RunUpdater( button_adder )
end

function SidebarOutfits:GetDefaultFocus()
	return self.last_focused_button or ( self.scrolling_contents.children and #self.scrolling_contents.children > 0 and self.scrolling_contents.children[1] )
end

------------------------------------------------------------------------------------------
-- A button for each outfit level

local OutfitButton = class("Widget.OutfitButton", Widget.Clickable)

local HEIGHT = 90
local ICON_H = 80
local ICON_W = 124
local TEXT_LEFT_MARGIN = 130

function OutfitButton:init( width, act_id, outfit_data, is_locked, selected, seen, on_click_fn )
    OutfitButton._base.init(self)

    self.button_width = width
    self.act_id = act_id
    self.outfit_data = outfit_data
    self.is_locked = is_locked
    self.selected = selected
    self.seen = seen
    self.on_click_fn = on_click_fn

    self.hitbox = self:AddChild( Widget.SolidBox( self.button_width, HEIGHT, 0x00FFFF00 ) )
    self.contents = self:AddChild( Widget() )
    self.selector = self.contents:AddChild( Widget.Panel( engine.asset.Texture("UI/empty_inner_frame_border_selected_flat.tex") ) )
		:SetHiddenBoundingBox( true )
    	:SetSize( self.button_width, HEIGHT )
    	:SetTintColour( UICOLOURS.ACCENT )
    	:SetBloom( 0.1 )
    	:SetShown( self.selected )
    self.glow = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
    	:SetTintColour( UICOLOURS.ACCENT )
    	:SetTintAlpha( 0.2 )
   		:SetHiddenBoundingBox( true )
        :SetSize( ICON_W*2, ICON_W*2 )
	self.icon = self.contents:AddChild( Widget.Image( self.outfit_data.icon ) )
		-- :SetTintColour( self.is_locked and 0x4A6A82FF or UICOLOURS.WHITE )
   		:SetHiddenBoundingBox( true )
        :SetSize( ICON_W, ICON_H )
        :Bloom( 0.05 )
    self.text_content = self.contents:AddChild( Widget() )
		:SetHiddenBoundingBox( true )
    if self.is_locked then
    	self.status_icon = self.text_content:AddChild( Widget.Image( engine.asset.Texture("UI/newgame_sidebar_locked.tex" ) ) )
    		:SetSize( 34, 34 )
    		:Bloom( 0.05 )
    		:SetTintColour( 0x4A6A82FF )
    	self.icon:SetTintColour( 0x000000FF )
    	if not self.icon.tint_add_constant then
			self.icon.tint_add_constant = engine.rendering.ShaderConstantValue("TINT_ADD", 0,0,0,0)
       		self.icon.model:SetShaderConstant(self.icon.tint_add_constant)
    	end
    	self.icon.tint_add_constant:SetValue( HexColour( 0x4A6A8200 ) )
    	self.glow:SetTintColour( 0x4A6A82FF )
    	self.glow:SetTintAlpha( 0.2 )
    end
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, self.is_locked and LOC"UI.NEW_GAME_SCREEN.SIDEBAR_OUTFITS_LOCKED" or self.outfit_data:GetName() ) )
    	:SetGlyphColour( self.is_locked and 0x4A6A82FF or UICOLOURS.ACCENT )
        :LeftAlign()
        :SetAutoSize( self.button_width-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.is_locked and "" or self.outfit_data:GetDesc() ) )
        :LeftAlign()
        :SetAutoSize( self.button_width-TEXT_LEFT_MARGIN )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC*LINE_HEIGHT_TITLE_MULT )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
	self.new_items_label = self.contents:AddChild( Widget.NewItemsLabel( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_NEW" ) )
		:SetShown( not self.seen and not self.is_locked )
	if self.selected then
		-- Save this outfit as "seen"
		TheGame:GetGameProfile():SetSeenOutfit( self.act_id, self.outfit_data.id )
		self.new_items_label:Hide()
	end

	-- Layout things
	self.icon:LayoutBounds( "left", "center", self.hitbox ):Offset( 5, 0 )
	self.glow:LayoutBounds( "center", "center", self.icon ):Offset( 0, 0 )
	if self.is_locked then
		self.title:LayoutBounds( "after", "center", self.status_icon ):Offset( 0, 1 )
		self.text_content:LayoutBounds( "left", "center", self.hitbox ):Offset( TEXT_LEFT_MARGIN-7, 2 )
	else
		self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, 4 )
		self.text_content:LayoutBounds( "left", "center", self.hitbox ):Offset( TEXT_LEFT_MARGIN, 2 )
	end
	self.new_items_label:LayoutBounds( "after", "center", self.title ):Offset( 10, -1 )

    if self.is_locked == false then self:SetOnClickFn( function() self:OnClick() end ) end
    self:UpdateImage()
end

function OutfitButton:OnClick()
	-- Save this outfit as "seen"
	TheGame:GetGameProfile():SetSeenOutfit( self.act_id, self.outfit_data.id )

	-- Save selected outfit
	TheGame:GetGameProfile():SetSelectedOutfit( self.act_id, self.outfit_data.id )
	
	self.on_click_fn( self, self.outfit_data.id )
end

function OutfitButton:AnimateSelect( fn )
	self.selected = true
	self.selector:SetTintAlpha(0):Show()
	self:RunUpdater(
		Updater.Series {
			Updater.Do( function()
				-- Hide the "new" label
				self.new_items_label:Hide()
			end ),
			Updater.Parallel{
				Updater.Ease( function(v) self.selector:SetPos(v,nil) end, -40, 0, 0.1, easing.outQuad ),
				Updater.Ease( function(v) self.selector:SetScale(v,v) end, 1.1, 1, 0.1, easing.outQuad ),
				Updater.Ease( function(v) self.selector:SetTintAlpha(v) end, 0, 1, 0.1, easing.outQuad )
			},
			Updater.Wait( 0.1 ),
			Updater.Do( fn ),
		}
	)
end

function OutfitButton:AnimateUnselect( fn )
	self.selected = false
	self:RunUpdater(
		Updater.Parallel{
			Updater.Ease( function(v) self.selector:SetPos(v,nil) end, 0, -10, 0.05, easing.outQuad ),
			Updater.Ease( function(v) self.selector:SetTintAlpha(v) end, 1, 0, 0.05, easing.outQuad ),
			Updater.Do( fn ),
		}
	)
end

function OutfitButton:UpdateImage()
    if ( not self.is_locked and self.hover ) or self.focus then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.glow:ScaleTo( nil, 1.2, 0.1 )
        self.contents:ScaleTo( nil, 1.05, 0.2 )
        self.contents:MoveTo( 20, 0, 0.2 )
    else
        self.glow:ScaleTo( nil, 1, 0.1 )
        self.contents:ScaleTo( nil, 1, 0.3 )
        self.contents:MoveTo( 0, 0, 0.3 )
    end
end

function OutfitButton:OnGainFocus()
    OutfitButton._base.OnGainFocus(self)
    self:UpdateImage()
end

function OutfitButton:OnLoseFocus()
    OutfitButton._base.OnLoseFocus(self)
    self:UpdateImage()
end

function OutfitButton:OnGainHover()
    OutfitButton._base.OnGainHover(self)
    self:UpdateImage()
end

function OutfitButton:OnLoseHover()
    OutfitButton._base.OnLoseHover(self)
    self:UpdateImage()
end
