require "ui/widgets/perkwidget"
require "ui/screens/perkscreen"
require "ui/screens/perkviewpopup"

------------------------------------------------------------------------------------------
-- The sidebar showing the various selectable perks

local SidebarPerks = class("Widget.SidebarPerks", Widget.SidebarBlock)

function SidebarPerks:init( width )
    SidebarPerks._base.init(self,width)

    -- Amount of perk points we have
    self.balance = self.header_content:AddChild( Widget.Label( "title", FONT_SIZE.SCREEN_TITLE, LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_BALANCE" ) )
        :SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.PERK )
        :RightAlign()
        :SetSize( self.sidebar_width, 60 )
        :Bloom( 0.15 )

    self.title:SetSize( self.sidebar_width - 120, 60 )


    -- Sidebar top buttons
    self.btn_clearall = self.header_content:AddChild( Widget.ShortButton( (self.sidebar_width-10)/2, LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_CLEAR_ALL", nil ) )
        :SetOnClickFn( function() self:ClearAll() end )
        :SetFocusable( false )
    self.btn_clearselected = self.header_content:AddChild( Widget.ShortButton( (self.sidebar_width-10)/2, LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_CLEAR_SELECTED", nil ) )
        :SetOnClickFn( function() self:ClearSlot( self.current_slot_widget and self.current_slot_widget:GetIndex() ) end )
        :SetFocusable( false )
        :Hide()

    -- Sidebar unlockable slots
    self.slot_widget_root = self.scrolling_contents:AddChild(Widget())
    self.slot_widgets = {}
    for k = 1, #PERK_SLOT_UNLOCK_THRESHOLDS do
        local widget = self.scrolling_contents:AddChild( Widget.PerkOption( self.sidebar_width, 130, k ) )
            :SetOnChangeFn( function( widget, def_id, checked ) self:OnSlotSelected( widget, def_id, checked ) end )
        table.insert( self.slot_widgets, widget )
    end

    -- Sidebar bottom buttons
    self.btn_grifts = self.scrolling_contents:AddChild( Widget.ShortButton( (self.sidebar_width-10)/2, LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_GRIFTS", nil ) )
        :SetOnClickFn( function() self:ShowGrifts() end )
        :SetFocusable( false )

    -- Content area on the right
    self.right_content_root = self:AddChild( Widget() )
    self.right_content_scroll = self.right_content_root:AddChild( Widget.ScrollPanel() )
        :SetVirtualTopMargin( 60 )
        :SetVirtualBottomMargin( 250 )
        :SetScrollBarOuterMargin( 0 )
    self.right_content = self.right_content_scroll:AddScrollChild( Widget() )

    -- Get all perks
    self.perk_widgets = {}
    self.tier_widgets = {}
    self.perks_by_tier = {}
    local grafts = Content.GetAllGrafts()
    for _, graft in pairs(grafts) do
        if graft.type == GRAFT_TYPE.PERK then
            self.perks_by_tier[graft.tier] = self.perks_by_tier[graft.tier] or {}
            table.insert( self.perks_by_tier[graft.tier], graft )
        end
    end
    for tier,grafts in ipairs(self.perks_by_tier) do
        table.sort(grafts, function(a,b) 
            if a.unlock_cost ~= b.unlock_cost then
                return a.unlock_cost < b.unlock_cost
            else
                return a.id < b.id
            end
        end)
        local tier_widget = self.right_content:AddChild(Widget.PerkGroup(loc.format(LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER", tier), grafts, function(...) self:OnClickPerk(...) end ))
            :PrepareAnimation()
        table.insert( self.tier_widgets, tier_widget)
        for perk_id, widget in pairs(tier_widget.perk_widgets) do
            self.perk_widgets[ perk_id ] = widget
        end
    end

end

function SidebarPerks:IsAnimVisible()
    return false
end

function SidebarPerks:OnClickPerk(perk)
    local id = perk.id
    local unlocked = TheGame:GetGameProfile() and TheGame:GetGameProfile():HasPerkUnlocked(id)
    if not unlocked then
        TheGame:FE():PushScreen( Screen.PerkViewPopup( id, function(should_refresh) 
            if should_refresh then self:PerkUnlocked( id ) end
         end ) )
    elseif self.current_slot_widget then
        -- There is a slot currently active. Check if we can place this perk on it

        -- How many perks are active from this tier
        local current_tier = perk.tier
        local selected_perks_on_tier = 0
        for k, w in pairs( self.slot_widgets ) do
            if w:GetTier() == current_tier then
                selected_perks_on_tier = selected_perks_on_tier + 1
            end
        end

        if self.current_slot_widget:GetTier() ~= current_tier then
            if selected_perks_on_tier == PERK_TIER_EQUIP_LIMITS[ current_tier ] then
                -- The player tried to select too many perks from this tier. Let them know
                UIHelpers.InfoPopup( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_LIMIT_TITLE", loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_LIMIT_BODY", PERK_TIER_EQUIP_LIMITS[ current_tier ], current_tier ) )
                return self
            end
        end

        -- Check if this perk is selected already
        local selected = false
        for k, w in pairs( self.slot_widgets ) do
            if w:GetPerkId() == id then
                selected = true
            end
        end
        if selected then
            -- The player already had this perk selected. Let them know
            UIHelpers.InfoPopup( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_ALREADY_SELECTED_TITLE", loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_ALREADY_SELECTED_BODY", perk:GetLocalizedName() ) )
            return self
        end

        -- Save this perk to the slot!
        self.current_slot_widget:SetGraftId( id )

        -- Show the Clear slot button, now that this slot has a def
        self.btn_clearselected:SetShown( not TheGame:FE():IsGamepadMode() )

        -- Save this to the profile
        self:SaveToProfile()
    end
    return self
end

function SidebarPerks:OnSlotSelected( widget, def_id, checked )

    if checked then
        -- This slot was just toggled on, so untoggle all other slots first
        for k, v in ipairs(self.slot_widgets) do
            local checked = v==widget
            local notify_changed = false
            v:SetValue( checked, notify_changed )
        end

        -- Show the Clear slot button if this slot has a def
        if widget:GetPerkId() then
            self.btn_clearselected:SetShown( not TheGame:FE():IsGamepadMode() )
        else
            self.btn_clearselected:Hide()
        end

        -- Save selected slot
        self.current_slot_widget = widget
    else
        -- Hide the Clear slot button
        self.btn_clearselected:Hide()

        -- Clear selected slot
        self.current_slot_widget = nil
    end

    return self
end

function SidebarPerks:SelectedSlotHasPerk()
    return self.current_slot_widget and self.current_slot_widget:GetPerkId()
end

function SidebarPerks:SaveToProfile()

    -- Generate a list of the active perks
    local perks = {}
    for k, v in pairs( self.slot_widgets ) do
        if v:GetPerkId() then
            table.insert( perks, v:GetPerkId() )
        end
    end

    -- Save them to profile
    TheGame:GetGameProfile():SetSelectedPerks( self.character_data.id, perks )

    -- Update sidebar count
    self.info:SetText( loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TEXT", #perks, self.num_slots ) )

    return self
end

function SidebarPerks:PerkUnlocked( perk_id )

    -- Update perk points
    self:RefreshPerkBalance()

    -- Animate this perk being unlocked
    local perk_widget = self.perk_widgets[perk_id]
    if perk_widget then
        perk_widget:AnimateUnlock()
    end

    -- Unlock a perk slot if the threshold has been reached
    local total_perks_unlocked = TheGame:GetGameProfile():GetNumPerksUnlocked()
    local num_slots = TheGame:GetGameProfile():GetNumPerkSlots()
    for k, v in ipairs( self.slot_widgets ) do
        local unlocked = num_slots >= k
        if unlocked and v:IsLocked() then
            -- This slot wasn't unlocked before, but it is now
            v:AnimateUnlock()
        elseif v:IsLocked() then
            v:Refresh()
        end
    end

    local total_perks_unlocked = TheGame:GetGameProfile():GetNumPerksUnlocked()
    self.num_slots = TheGame:GetGameProfile():GetNumPerkSlots()

    local selected_perks = TheGame:GetGameProfile():GetSelectedPerks( self.character_data.id )
    -- Update count
    self.info:SetText( loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TEXT", #selected_perks, self.num_slots ) )

    -- Unlock a perk tier if the threshold has been reached
    self:RefreshPerkTiers()

    return self
end

function SidebarPerks:ClearAll()

    for k, v in ipairs(self.slot_widgets) do
        -- Clear each slot
        self:ClearSlot(k)
    end

    return self
end

function SidebarPerks:ClearSlot( idx )
    if idx and self.slot_widgets[idx] then
        -- Get slot widget
        local widget = self.slot_widgets[idx]
        -- Clear slot
        local checked = false
        local notify_changed = false
        widget:SetValue( checked, notify_changed )
            :SetDef( nil )
        -- Clear selected slot
        self.current_slot_widget = nil
        -- Save this to the profile
        self:SaveToProfile()
    end

    -- Hide the Clear slot button
    self.btn_clearselected:Hide()

    return self
end

function SidebarPerks:ShowGrifts()
    TheGame:FE():InsertScreen( Screen.Compendium( 5 ) )
    return self
end

function SidebarPerks:RefreshPerkBalance()
    local points = TheGame:GetGameProfile():GetPerkPoints() or 0
    self.balance:SetText(loc.format(LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_BALANCE", points))
end

function SidebarPerks:RefreshPerkTiers()

    local unlocked_counts, locked_counts = {}, {}
    local grafts = Content.GetAllGrafts()
    for _, graft in pairs(grafts) do
        if graft.type == GRAFT_TYPE.PERK then
            if TheGame:GetGameProfile():HasPerkUnlocked(graft.id) then
                unlocked_counts[graft.tier] = (unlocked_counts[graft.tier] or 0) + 1
            else
                locked_counts[graft.tier] = (locked_counts[graft.tier] or 0) + 1
            end
        end
    end

    local previous_tier
    for tier, widget in ipairs(self.tier_widgets) do
        local is_unlocked 
        local txt 
        local limit = PERK_TIER_EQUIP_LIMITS[tier]
        
        if tier > 1 then
            local total = (locked_counts[tier-1] or 0) + (unlocked_counts[tier-1] or 0) 
            local thresh = PERK_TIER_UNLOCK_THRESH[tier] or math.floor(total * .4)
            is_unlocked = (unlocked_counts[tier-1] or 0) >= thresh
            if not is_unlocked then
                txt = loc.format(limit < #PERK_SLOT_UNLOCK_THRESHOLDS and LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_UNLOCK_LIMIT" or LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_UNLOCK", thresh, tier-1, limit)
            end
        else
            is_unlocked = true
        end

        if is_unlocked then
            txt = loc.format(limit < #PERK_SLOT_UNLOCK_THRESHOLDS and LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_UNLOCKED_LIMIT" or LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TIER_UNLOCKED", unlocked_counts[tier] or 0, (unlocked_counts[tier] or 0) + (locked_counts[tier] or 0), limit)
        end

        widget:SetLocked( not is_unlocked )
            :SetLockedText( txt )
            :Layout()
    end
    
    self:Layout()

    return self
end

function SidebarPerks:Refresh( character_data )
    self.character_data = character_data
	if not character_data then return end

	SidebarPerks._base.Refresh( self, character_data )

	self.title:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TITLE" ):SetGlyphColour( UICOLOURS.PERK )
    self.title:TryFitToRegion( 32, FONT_SIZE.SCREEN_TITLE )
    self.subtitle:Hide()
    self.text:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_SUBTITLE" )

    -- Get unlocked slots count
    local total_perks_unlocked = TheGame:GetGameProfile():GetNumPerksUnlocked()
    self.num_slots = TheGame:GetGameProfile():GetNumPerkSlots()

    -- Get selected perks
    local selected_perks = TheGame:GetGameProfile():GetSelectedPerks( self.character_data.id )

    -- Update count
    self.info:SetText( loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_TEXT", #selected_perks, self.num_slots ) )

    -- Refresh left-side perk slots
    for k, v in ipairs(self.slot_widgets) do
        local unlocked = self.num_slots >= k
        if unlocked then
            v:SetLocked( false )
            -- This slot is unlocked. Is there a selected perk already for it?
            if selected_perks[k] then
                -- There is a previously selected perk for this slot
                v:SetGraftId( selected_perks[k] )
            else
                -- This slot has no perk selected yet
                v:SetDef( nil )
            end
        else
            -- This slot is locked
            v:SetLocked( true )
        end
    end

    self:RefreshPerkTiers()

    self:RefreshPerkBalance()

    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )

	-- Layout
    self:Layout()

    return self
end

function SidebarPerks:Layout()
    SidebarPerks._base.Layout( self )

    -- Layout perk points balance
    self.balance:LayoutBounds( "left", "center", self.title )

    -- Layout header text
    if self.info:IsShown() then self.info:LayoutBounds( "left", "below", self.text:IsShown() and self.text or self.subtitle ):Offset( 0, self.text:IsShown() and 0 or -SPACING.M1 ) end

    -- Layout buttons
    self.btn_clearall:LayoutBounds( "left", "below", self.info ):Offset( 0, -SPACING.M1/2 )
    self.btn_clearselected:LayoutBounds( "after", "center", self.btn_clearall ):Offset( 10, 0 )

    -- Layout sidebar slots
    self.scrolling_contents:StackChildren( SPACING.M1/2 )
        :LayoutBounds( "left", "below", self.header_content ):Offset( 0, -SPACING.M1/2 )
    -- Move the grifts button a bit further down
    self.btn_grifts:LayoutBounds( "left", "below" ):Offset( 0, -SPACING.M1 )

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Get screen scaling mode
    local sm = TheGame:FE():GetScreenMode()

    -- Calculate right-side content size
    local screen_edge_padding = 10
    local padding = 90*self.current_layout_scale
    local scroll_w = self.screen_w - self.sidebar_width*self.current_layout_scale - padding - screen_edge_padding
    local scroll_h = self.screen_h
    self.right_content_scroll:SetSize( scroll_w, scroll_h )

    -- Scale the tiers container
    self.right_content:SetLayoutScale( LAYOUT_SCALE[sm] )

    -- Set a small number of columns
    local columns = 3
    local content_width = 0
    local max_w = scroll_w - 50 -- Add some extra spacing for the left edge of the scroll panel
    -- Now keep increasing the number of columns until they don't fit anymore
    repeat
        columns = columns + 1
        for tier, widget in ipairs( self.tier_widgets ) do
            widget:SetColumns( columns )
        end
        content_width = self.right_content:GetScaledSize()
    until content_width > max_w or columns == 10
    -- Then scale down so it fits
    for tier, widget in ipairs( self.tier_widgets ) do
        widget:SetColumns( columns-1 )
    end

    -- Position tiers
    self.right_content:StackChildren( SPACING.M1*1.25 )
        :LayoutBounds( "left", "top", 20, 0 )

    self.right_content_scroll:RefreshView()
        :LayoutBounds( "left", "top", self.scroll_panel )
        :Offset( self.sidebar_width*self.current_layout_scale + padding, 0 )

    return self
end

function SidebarPerks:OnControlModeChange( cm, device_id )
    SidebarPerks._base.OnControlModeChange( self, cm, device_id )

    if cm == CONTROL_MODE.MOUSE_KEYBOARD then
        self.btn_clearall:Show()
        self.btn_clearselected:SetShown( self.current_slot_widget and self.current_slot_widget:GetPerkId() )
        self.btn_grifts:Show()
    end
    if cm == CONTROL_MODE.TOUCH then
        self.btn_clearall:Show()
        self.btn_clearselected:Show()
        self.btn_grifts:Show()
    end
    if cm == CONTROL_MODE.GAMEPAD then
        self.btn_clearall:Hide()
        self.btn_clearselected:Hide()
        self.btn_grifts:Hide()
    end
    
    self:Layout()
end

function SidebarPerks:OnScreenModeChange( sm )
    SidebarPerks._base.OnScreenModeChange( self, sm )

    self:Layout()
end

function SidebarPerks:OnScreenResize(w,h)
    SidebarPerks._base.OnScreenResize( self,w,h )

    self:Layout()
end

function SidebarPerks:PrepareAnimateIn()
    SidebarPerks._base.PrepareAnimateIn( self )
    for i, widget in ipairs( self.tier_widgets ) do
        widget:PrepareAnimation()
    end
end

function SidebarPerks:AnimateIn()
	SidebarPerks._base.AnimateIn( self )

     -- Go through the tiers and animate each of them in
    local sequence = Updater.Parallel()
    for i, widget in ipairs( self.tier_widgets ) do
        sequence:Add( widget:GetAnimateUpdater( i * 0.2 ) )
    end
    self:RunUpdater( sequence )

    return self
end

function SidebarPerks:GetDefaultFocus()
	return self.slot_widgets[1]
end

-------------------------------------------------------------------------
-- A single perk toggle

local PerkOption = class( "Widget.PerkOption", Widget.Clickable )

local ICON_W = 126
local TEXT_LEFT_MARGIN = 148

PerkOption.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        test = function(self)
            return self and self.enabled and not self:IsAnimatingOut() and self.onclickfn
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.CONTROLS.SELECT", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.GAMEPAD_X,
        test = function(self) 
            return self:GetPerkId()
        end,
        fn = function(self)
            local perks_sidebar = self:GetAncestor( Widget.SidebarPerks )
            if perks_sidebar then perks_sidebar:ClearSlot( self.idx ) end
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.NEW_GAME_SCREEN.CONTROLS_PERKS_CLEAR_SELECTED", Controls.Digital.GAMEPAD_X ))
        end,
    },
}

function PerkOption:init( w, h, idx )
    PerkOption._base.init( self )
    self.features = {}
    self.idx = idx or 0 -- This slot's index
    
    self.hitbox = self:AddChild( Widget.SolidBox( w, h, 0x00FFFF00 ) )
    self.contents = self:AddChild( Widget() )
    self.bg = self.contents:AddChild( Widget.Panel( engine.asset.Texture("UI/mutator_bg_active.tex" ) ) )
        :SetTintColour( UICOLOURS.PERK )
        :SetTintAlpha( 0.2 )
        :SetBloom(0.1)
    self.glow = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetTintColour( UICOLOURS.PERK )
        :SetTintAlpha( 0.2 )
        :SetHiddenBoundingBox( true )
        :SetSize( ICON_W*1.2, ICON_W*1.2 )
    self.empty_slot = self.contents:AddChild( Widget.PerkSlot() )
    self.icon = self.contents:AddChild( Widget.GraftSlot( GRAFT_TYPE.PERK, ICON_W ) )
    self.text_content = self.contents:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.WHITE )
        :SetTintColour( UICOLOURS.PERK )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC ) )
        :LeftAlign()
        :SetAutoSize( w-TEXT_LEFT_MARGIN )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC*LINE_HEIGHT_TITLE_MULT )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )

    self:SetSize( w, h )

    self:SetValue( false )
    self:SetOnClickFn( function()
        if not self.locked then
            self:SetValue( not self.checked )

            if self.checked then
                AUDIO:PlayEvent( "event:/ui/new_game_menu/mutator_on" )
            else
                AUDIO:PlayEvent( "event:/ui/new_game_menu/mutator_off" )
            end

            if self.on_change_fn then self.on_change_fn( self, self.def and self.def.id, self.checked ) end
        end
    end )
end

function PerkOption:GetIndex()
    return self.idx
end

function PerkOption:SetSize( w, h )
    self.width = w or 300
    self.height = h or 90
    self.title:SetAutoSize( w-TEXT_LEFT_MARGIN-SPACING.M1 )
    self.text:SetAutoSize( w-TEXT_LEFT_MARGIN-SPACING.M1 )
    self:Layout()
    return self
end

function PerkOption:Layout()

    self.text:LayoutBounds( "left", "below", self.title )
        :Offset( 0, SPACING.TITLE_DESC )
    local text_w, text_h = self.text_content:GetSize()
    local perk_w, perk_h = self.empty_slot:GetSize()
    text_h = math.max( text_h, perk_h )
    text_h = text_h + SPACING.M1
    self.hitbox:SetSize( self.width, text_h )
    self.bg:SetSize( self.width, text_h )
    self.empty_slot:LayoutBounds( "left", "top", self.bg )
        :Offset( SPACING.M1*0.55, -SPACING.M1*0.5 )
    self.icon:LayoutBounds( "center", "center", self.empty_slot )
    self.glow:LayoutBounds( "center", "center", self.empty_slot )
        :Offset( 0, 0 )
    self.text_content:LayoutBounds( "left", "center", self.bg )
        :Offset( TEXT_LEFT_MARGIN, 1 )

    return self
end

function PerkOption:SetGraftId( id )
    local def = Content.GetGraft( id )
    self:SetDef( def )
    return self
end

function PerkOption:SetDef( def )

    if def then
        -- Valid graft
        self.def = def
        self.graft = GraftInstance( def.id )
    else
        -- Empty slot
        self.def = nil
        self.graft = nil
    end

    self:Refresh()
    return self
end

function PerkOption:SetLocked( locked )
    self.locked = locked
    self:Refresh()
    return self
end

function PerkOption:IsLocked()
    return self.locked
end

function PerkOption:AnimateUnlock()
    self.locked = false

    -- Animate
    self:RunUpdater(
        Updater.Series{
            Updater.Do( function() AUDIO:PlayEvent(SoundEvents.perkslot_unlock) end ),
            Updater.Ease( function(v) self.contents:SetScale(v) end, 1, 1.1, 0.3, easing.inQuad),
            Updater.Do( function() self:Refresh() end ),
            Updater.Ease( function(v) self.contents:SetScale(v) end, 1.2, 1, 0.2, easing.inQuad),
        }
    )
    return self
end

function PerkOption:Refresh()
    if self.locked then
        -- Locked
        self.empty_slot:Refresh( self.idx, false, "" ):Show()
        self.icon:Hide()
        -- Calculate how far we're from unlocking this one
        local total_perks_unlocked = TheGame:GetGameProfile():GetNumPerksUnlocked()
        local unlock_threshold = PERK_SLOT_UNLOCK_THRESHOLDS[self.idx]
        self.title:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_SLOT_TITLE_LOCKED" )
            :SetTintColour( 0x4A6A82ff )
        self.text:SetText( loc.format( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_SLOT_DESC_LOCKED", unlock_threshold-total_perks_unlocked ) )
            :SetGlyphColour( 0x4A6A82ff )
        self:DisableToolTip( true )
    elseif self.graft then
        -- Unlocked and there's a graft here!
        -- self.empty_slot:Hide()
        self.icon:Refresh( self.graft ):Show()
        self.title:SetText( self.def:GetLocalizedName() )
            :SetTintColour( UICOLOURS.PERK )
        self.text:SetText( self.graft:GetDesc( self ) )
            :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        self:SetToolTipClass( Widget.TooltipGraft )
            :SetToolTip( self.graft )
            :ShowToolTipOnFocus( true )
            :DisableToolTip( false )
    else
        -- Unlocked but empty
        self.empty_slot:Refresh( self.idx, true, "" ):Show()
        self.icon:Hide()
        self.title:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_SLOT_TITLE_EMPTY" )
            :SetTintColour( UICOLOURS.PERK )
        self.text:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PERKS_SLOT_DESC_EMPTY" )
            :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        self:DisableToolTip( true )
    end
    self:Layout()
    return self
end

function PerkOption:IsChecked()
    return self.checked
end

function PerkOption:SetValue( checked, notify_changed )
    self.checked = checked
    if self.checked then
        self.bg:SetTexture( engine.asset.Texture("UI/mutator_bg_active.tex" ) )
            :SetTintColour( UICOLOURS.PERK )
        self.title:SetGlyphColour( UICOLOURS.PERK )
        self.icon:SetTintColour( UICOLOURS.WHITE )
        self.glow:SetTintColour( UICOLOURS.PERK )
            :SetTintAlpha( 0.3 )
    else
        self.bg:SetTexture( engine.asset.Texture("UI/mutator_bg_normal.tex" ) )
            :SetTintColour( 0x4A6A82FF )
        self.glow:SetTintColour( 0x4A6A82FF )
            :SetTintAlpha( 0.2 )
    end
    if notify_changed and self.on_change_fn then
        self.on_change_fn( self, self.def.id, self.checked )
    end
    return self
end

function PerkOption:GetPerkId()
    return self.def and self.def.id
end

function PerkOption:GetTier()
    return self.def and self.def.tier
end

function PerkOption:SetOnChangeFn( fn )
    self.on_change_fn = fn
    return self
end

function PerkOption:PostFeature( keyword, name, desc )
    self.features[ keyword ] = { name = name, desc = desc }
end

function PerkOption:PostCard( feature_id, card )
    self.features[ feature_id ] = { card = card }
end

function PerkOption:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.glow:ScaleTo( nil, 1.2, 0.1 )
        if not self.locked then
            self.contents:ScaleTo( nil, 1.05, 0.2 )
            self.contents:MoveTo( 20, 0, 0.2 )
        end
    else
        self.glow:ScaleTo( nil, 1, 0.1 )
        self.contents:ScaleTo( nil, 1, 0.3 )
        self.contents:MoveTo( 0, 0, 0.3 )
    end
end

function PerkOption:OnGainFocus()
    PerkOption._base.OnGainFocus(self)
    self:UpdateImage()
end

function PerkOption:OnLoseFocus()
    PerkOption._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PerkOption:OnGainHover()
    PerkOption._base.OnGainHover(self)
    self:UpdateImage()
end

function PerkOption:OnLoseHover()
    PerkOption._base.OnLoseHover(self)
    self:UpdateImage()
end
