------------------------------------------------------------------------------------------
-- The sidebar showing the prestige levels

local SidebarPrestige = class("Widget.SidebarPrestige", Widget.SidebarBlock)

function SidebarPrestige:init( width )
    SidebarPrestige._base.init(self,width)
end

function SidebarPrestige:Refresh( character_data, act_data )
    if not character_data then return end

	SidebarPrestige._base.Refresh( self, character_data )

	self.act_id = act_data.id
	self.title:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_TITLE" ):SetGlyphColour( UICOLOURS.ADVANCEMENT )
    self.subtitle:Hide()
	self.text:SetText( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_TEXT" )

	-- Get last selected prestige level
	self.selected_prestige = TheGame:GetGameProfile():GetSelectedAdvancement( self.act_id )
	-- Get highest played level
	self.played_prestige = TheGame:GetGameProfile():GetHighestPlayedAdvancement( self.act_id )
	-- Get highest unlocked level
	self.unlocked_prestige = TheGame:GetGameProfile():GetAdvancementLevel( self.act_id )

	-- Remove old prestige buttons
	self.scrolling_contents:DestroyAllChildren()

	-- Add new ones
    self.scrolling_contents:AddChild( Widget.PrestigeButton( self.sidebar_width, 0, engine.asset.Texture( "UI/advancement_level_0.tex" ), LOC"ADVANCEMENT.ADVANCEMENT_LEVEL_0",LOC"ADVANCEMENT.ADVANCEMENT_LEVEL_0_DESC", nil, self.selected_prestige==0, false, false, function( button, level ) self:OnLevelSelect( button, level ) end ) )
    self.last_focused_button = nil

    local tt = { LOC"ADVANCEMENT.ALL_MODIFIERS" }
    for k, v in ipairs( character_data.data.advancement ) do
		-- table.insert( tt, LOC(v.desc) )
		local selected = self.selected_prestige == k
		local new = k > self.played_prestige
		local locked = k > self.unlocked_prestige
    	local button = self.scrolling_contents:AddChild( Widget.PrestigeButton( self.sidebar_width, k, v.icon, LOC(v.name), LOC(v.desc), GetAdvancementTooltip(character_data.id, k), selected, new, locked, function( button, level ) self:OnLevelSelect( button, level ) end ) )
    	if selected then self.last_focused_button = button end
        button.enabled = false
    end

	-- Layout
    self:Layout()

end

function SidebarPrestige:Layout()

    self.scrolling_contents:StackChildren( 10, 0 )

    SidebarPrestige._base.Layout( self )
end

function SidebarPrestige:OnLevelSelect( button, prestige_level )
	-- Save selected prestige
	self.selected_prestige = prestige_level
	TheGame:GetGameProfile():UpdateSelectedAdvancement( self.act_id, self.selected_prestige )

	-- Unselect last button
	for i, widget in ipairs( self.scrolling_contents.children ) do
		if widget.selected then
			-- Tell the selected button to unselect itself, and to run the function when done
			widget:AnimateUnselect( function()
				-- When done, tell the new button to select itself
				button:AnimateSelect( function() end )
			end )
		end
	end
end

function SidebarPrestige:AnimateIn()
	SidebarPrestige._base.AnimateIn( self )

    -- Prepare button fade-in sequence
    local timing = 0.2
    local spacing = 30
    local updater = Updater.Series()
    local button_adder = Updater.Parallel()

    for i, widget in ipairs( self.scrolling_contents.children ) do
    	local pos_x, pos_y = widget:GetPos()
        widget.enabled = false
        button_adder:Add( Updater.Ease( function(v) widget:SetPos(nil,v) end, pos_y-spacing*i, pos_y, timing, easing.outQuad ) )
    end
    
    -- Enable buttons after they've animated in
    updater:Add( button_adder )
    updater:Add( Updater.Do( function()
        local selected
    	for i, widget in ipairs( self.scrolling_contents.children ) do
            if widget.selected then selected = widget end
    		widget.enabled = true
    	end
        -- Scroll to the current one
        if selected then
            self.scroll_panel:EnsureVisible( selected )
        end
	end ) )
    self:RunUpdater( updater )
end


function SidebarPrestige:GetDefaultFocus()
	return self.last_focused_button or ( self.scrolling_contents and self.scrolling_contents.children and #self.scrolling_contents.children > 0 and self.scrolling_contents.children[1] )
end

------------------------------------------------------------------------------------------
-- A button for each prestige level

local PrestigeButton = class("Widget.PrestigeButton", Widget.Clickable)

local HEIGHT = 130
local ICON_W = 60
local TEXT_LEFT_MARGIN = 80

function PrestigeButton:init( width, prestige_level, icon, title, desc, tt, selected, new, locked, on_click_fn )
    PrestigeButton._base.init(self)

    self.button_width = width
    self.prestige_level = prestige_level
    self.selected = selected
    self.new = new
    self.locked = locked

    self.hitbox = self:AddChild( Widget.SolidBox( self.button_width, HEIGHT, 0x00FFFF00 ) )
    self.contents = self:AddChild( Widget() )
    self.selector = self.contents:AddChild( Widget.Panel( engine.asset.Texture("UI/empty_inner_frame_border_selected_flat.tex") ) )
		:SetHiddenBoundingBox( true )
    	:SetSize( self.button_width, HEIGHT )
    	:SetTintColour( UICOLOURS.ADVANCEMENT )
    	:SetBloom( 0.1 )
    	:SetShown( self.selected )
    self.glow = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
    	:SetTintColour( UICOLOURS.ADVANCEMENT )
    	:SetTintAlpha( 0.2 )
   		:SetHiddenBoundingBox( true )
        :SetSize( ICON_W*2, ICON_W*2 )
	self.icon = self.contents:AddChild( Widget.Image( self.locked and engine.asset.Texture("UI/newgame_sidebar_locked.tex") or icon ) )
		:SetTintColour( self.locked and 0x4A6A82FF or UICOLOURS.WHITE )
        :SetSize( ICON_W, ICON_W )
        :Bloom( 0.15 )
    self.text_content = self.contents:AddChild( Widget() )
		:SetHiddenBoundingBox( true )
    self.title = self.text_content:AddChild( Widget.Label( "title", FONT_SIZE.ITEM_SUBTITLE, self.locked and LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_LOCKED_TITLE" or title ) )
    	:SetTintColour( self.locked and 0x4A6A82FF or UICOLOURS.ADVANCEMENT )
        :LeftAlign()
        :SetAutoSize( self.button_width-TEXT_LEFT_MARGIN )
        :SetWordWrap( true )
        :Bloom( 0.1 )
    self.text = self.text_content:AddChild( Widget.Label( "tooltip", FONT_SIZE.ITEM_DESC, self.locked and LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_LOCKED_TEXT" or desc ) )
        :LeftAlign()
        :SetAutoSize( self.button_width-TEXT_LEFT_MARGIN )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC*LINE_HEIGHT_TITLE_MULT )
        :SetWordWrap( true )
        :SetGlyphColour( UICOLOURS.TOOLTIP_TEXT )
        :Bloom( 0.1 )
    self.new_items_label = self.contents:AddChild( Widget.NewItemsLabel( LOC"UI.NEW_GAME_SCREEN.SIDEBAR_PRESTIGE_NEW" ) )
    	:SetShown( self.new and not self.locked )
    if tt and not self.locked then
    	self:SetToolTip( tt )
    	self:ShowToolTipOnFocus( true )
    end

    -- Update size to match text
    self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
    local text_w, text_h = self.text_content:GetSize()
    local new_height = text_h + SPACING.M1*1.5 -- increase this value to *2 if there isn't enough space on the top and bottom
    self.hitbox:SetSize( self.button_width, new_height )
    self.selector:SetSize( self.button_width, new_height )

	-- Layout things
	self.icon:LayoutBounds( "left", "center", self.hitbox ):Offset( 5, 0 )
	self.glow:LayoutBounds( "center", "center", self.icon ):Offset( 0, 0 )
	self.text:LayoutBounds( "left", "below", self.title ):Offset( 0, SPACING.TITLE_DESC )
	self.text_content:LayoutBounds( "left", "center", self.hitbox ):Offset( TEXT_LEFT_MARGIN, 0 )
	self.new_items_label:LayoutBounds( "after", "center", self.title ):Offset( 10, -1 )
    self.icon:LayoutBounds( "left", "center", self.hitbox ):Offset( 5, 0 )

    if not self.locked then self:SetOnClickFn( function() on_click_fn( self, self.prestige_level ) end ) end
    self:UpdateImage()
end

function PrestigeButton:AnimateSelect( fn )
	self.selected = true
	self.selector:SetTintAlpha(0):Show()
	self:RunUpdater(
		Updater.Series {
			Updater.Parallel{
				Updater.Ease( function(v) self.selector:SetPos(v,nil) end, -40, 0, 0.1, easing.outQuad ),
				Updater.Ease( function(v) self.selector:SetScale(v,v) end, 1.1, 1, 0.1, easing.outQuad ),
				Updater.Ease( function(v) self.selector:SetTintAlpha(v) end, 0, 1, 0.1, easing.outQuad )
			},
			Updater.Wait( 0.1 ),
			Updater.Do( fn ),
		}
	)
end

function PrestigeButton:AnimateUnselect( fn )
	self.selected = false
	self:RunUpdater(
		Updater.Parallel{
			Updater.Ease( function(v) self.selector:SetPos(v,nil) end, 0, -10, 0.05, easing.outQuad ),
			Updater.Ease( function(v) self.selector:SetTintAlpha(v) end, 1, 0, 0.05, easing.outQuad ),
			Updater.Do( fn ),
		}
	)
end

function PrestigeButton:UpdateImage()
    if self.enabled and ( self.hover or self.focus ) then
        if self.is_locked then
            AUDIO:PlayEvent( "event:/ui/card_codex/mouseover_locked_card" )
        end
        self.glow:ScaleTo( nil, 1.2, 0.1 )
        self.contents:ScaleTo( nil, 1.05, 0.2 )
        self.contents:MoveTo( 20, 0, 0.2 )
    else
        self.glow:ScaleTo( nil, 1, 0.1 )
        self.contents:ScaleTo( nil, 1, 0.3 )
        self.contents:MoveTo( 0, 0, 0.3 )
    end
end

function PrestigeButton:OnGainFocus()
    PrestigeButton._base.OnGainFocus(self)
    self:UpdateImage()
end

function PrestigeButton:OnLoseFocus()
    PrestigeButton._base.OnLoseFocus(self)
    self:UpdateImage()
end

function PrestigeButton:OnGainHover()
    PrestigeButton._base.OnGainHover(self)
    self:UpdateImage()
end

function PrestigeButton:OnLoseHover()
    PrestigeButton._base.OnLoseHover(self)
    self:UpdateImage()
end
