require "ui/widget"
require "ui/widgets/solidbox"
require "ui/widgets/panelbutton"
require "ui/widgets/dragbutton"
local easing = require "util/easing"
local DEFAULT_SKIN =
{
    bg_colour = 0x5492AAFF,
    button_colour = 0xaa00aaff,
    knob_colour = 0x87A299FF
}

local Slider = class("Widget.Slider", Widget)

local HEIGHT = 20

function Slider:init(length )
    Slider._base.init(self)
    self.hover_check = true
    self.min = 0
    self.max = 10
    self.val = 5
    self.step_size = 1
    self.length = length or 200

    self.bg_border_size = 12
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = self.length + self.bg_border_size*2*self.bg_border_scale, HEIGHT + self.bg_border_size*2*self.bg_border_scale
    self.bg = Widget.Panel( engine.asset.Texture("UI/slider_bg.tex") )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :SetNineSliceCoords( self.bg_border_size, self.bg_border_size, self.bg_border_size, self.bg_border_size )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :Bloom( 0.05 )
    self.handle = Widget.DragButton( nil, nil, UICOLOURS.SUBTITLE, { engine.asset.Texture ("UI/slider_handle.tex"), 0, 10, 16, 45 } )
        :SetDragFn(function(p) return self:OnScroll(p) end):SetHorizontal()
    self.handle_border_size = 12
    self.handle_border_scale = 0.5
    self.handle.bg:SetNineSliceCoords( self.handle_border_size, self.handle_border_size, self.handle_border_size, self.handle_border_size )
        :SetNineSliceBorderScale( self.handle_border_scale )
    local nineslice_w, nineslice_h = 18 + self.handle_border_size*2*self.handle_border_scale, 26 + self.handle_border_size*2*self.handle_border_scale
    self.handle:SetSize( nineslice_w, nineslice_h )
        :SetTintColour( UICOLOURS.SUBTITLE )
        :Bloom( 0.1 )
    self.increment_btn = Widget.ArrowButton():SetOnClickFn(function() self:Increment() end)
    self.decrement_btn = Widget.ArrowButton():SetFlip(true):SetOnClickFn(function() self:Decrement() end)
    self:AddChildren{
        self.bg,
        self.handle,
        self.increment_btn,
        self.decrement_btn,
    }
end

function Slider:SetLabel(val)
    if self.label and not val then
        self.label:Remove()
        self.label = nil

        return
    end

    if val and not self.label then
        self.label = Widget.Label("body", 20):RightAlign()
        self:AddChild(self.label)    
        self.label:SetText(tostring(self.val))
        self:Layout()
    end

    return self
end

function Slider:SetFn(fn)
    self.fn = fn
    return self
end

function Slider:SetWidth(w)
    return self:SetLength(w)
end

function Slider:SetLength(l)
    self.length = l
    self:Layout()
    return self
end

function Slider:SetRange(min, max)
    assert( max >= min )
    self.min = min
    self.max = max
    self.val = clamp(self.val, min, max)
    self:Layout()
    return self
end

function Slider:Layout()
    
    if self.label then
        self.label:SetPos(0, 0):SetRegionSize(self.length, 20)
    end

    local p = 0
    if self.max ~= self.min then
        p = (self.val - self.min)/(self.max - self.min)
    end
    local x = -self.length*.5+8 + p*(self.length-16)

    self.handle:SetPos(x, 0):SetExtents(-self.length/2+8, self.length/2-8)
    local nineslice_w, nineslice_h = self.length + self.bg_border_size*2*self.bg_border_scale, HEIGHT + self.bg_border_size*2*self.bg_border_scale
    self.bg:SetSize( nineslice_w, nineslice_h )
    self.increment_btn:LayoutBounds("after", "center", self.bg):Offset(4, 0)
    self.decrement_btn:LayoutBounds("before", "center", self.bg):Offset(-4, 0)

    return self
end

function Slider:SetStepSize( step_size )
    assert( step_size > 0 )
    self.step_size = step_size
    return self
end

function Slider:Increment()
    self:SetValue(self.val+self.step_size)
end

function Slider:Decrement()
    self:SetValue(self.val-self.step_size)
end

function Slider:SetValue(v, silent, nolayout)
    
    local mults = v  / self.step_size
    v = math.floor( mults + .5 ) * self.step_size

    v = clamp( v, self.min, self.max )

    if v ~= self.val then
        self.val = v
        if not nolayout then
            self:Layout()   
        end
        if self.label then
            self.label:SetText(tostring(self.val))
        end
        if not silent and self.fn then
            self.fn(v)
        end
    end
    return self
end

function Slider:GetValue()
    return self.val
end

function Slider:HandleControlUp(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        return true
    end
end


function Slider:OnScroll(p)
    p = clamp(p,0,1)
    local v = self.min + (self.max-self.min)*p
    self:SetValue(v, false, true)
    return (self.val - self.min) / (self.max- self.min)
end


function Slider:HandleControlDown(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        local x, y = self:TransformFromWorld(TheGame:GetInput():GetMousePos())
        local click_p = clamp ( (x + self.length/2) / self.length, 0, 1)
        local v = self.min + (self.max-self.min)*click_p
        self:SetValue(v)
        --local hx, hy = self.handle:GetPos()
        self.handle:StartDragging()
        return true
    elseif control:Has( Controls.Digital.MENU_LEFT ) then
        self:SetValue(self:GetValue()-1)
        return true
    elseif control:Has( Controls.Digital.MENU_RIGHT ) then
        self:SetValue(self:GetValue()+1)
        return true
    end
end

function Slider:HideButtons()
    self.increment_btn:Hide()
    self.decrement_btn:Hide()
    return self
end