require "ui/widget"
require "ui/widgets/label"
require "ui/widgets/image"
require "ui/widgets/arrowbutton"

local Spinner = class("Widget.Spinner", Widget)
Spinner.can_focus = true

function Spinner:init( fn, width )
	Spinner._base.init(self)
	
	self.bg = self:AddChild(Widget.SolidBox())	
	self.left = self:AddChild(Widget.ArrowButton( function() self:Left() end))
	self.right = self:AddChild(Widget.ArrowButton( function() self:Right() end))

	self.options = {}
	self.idx = 1
	self.width = width or 100
	self.height = 32
	if fn then
		self:SetFn(fn)
	end
	self:Layout()
end

function Spinner:SetFn(fn)
	self.fn = fn
	return self
end

function Spinner:SetColour(...)
	self.bg:SetColour(...)
	return self
end

function Spinner:Layout()
	local scale = 1

	local bw, bh = self.left:GetSize()
	if bh > self.height then
		scale = self.height/bh
	end

	self.left:SetPos(-self.width/2 + bw/2 + 2*scale, 0, 0)
	self.left:SetScale(-scale, scale, scale)
	self.right:SetPos(self.width/2 - bw/2 - 2*scale, 0, 0)
	self.right:SetScale(scale, scale, scale)
	self.bg:SetSize(self.width, self.height)
end

function Spinner:SetWidth(w)
	self.width = w
	self:Layout()
	return self
end

function Spinner:SetHeight(h)
	self.height = h
	self:Layout()
	return self
end

function Spinner:SetSize( w, h )
	self.width, self.height = w, h
	self:Layout()
	return self
end

function Spinner:ClearOptions()
	self.idx = 1
	table.clear(self.options)
	self:UpdateState()
end

function Spinner:AddOption(opt)
	table.insert(self.options, opt)
	-- Update state if this option is either the current index, or adjacent to it (left / right button states)
	if (#self.options - self.idx) <= 1 then
		self:UpdateState()
	end
	return self
end

function Spinner:SetOptions(opt, idx)
	self.idx = idx or 1
	self.options = opt or {}
	self:UpdateState()
	return self
end

function Spinner:GetOptions()
	return self.options
end

function Spinner:SetNumberOptions(min, max, step)
	self:ClearOptions()
	for k = min, max, step or 1 do
		table.insert(self.options, tostring(k))
	end
	self:UpdateState()
	return self
end


function Spinner:Enable()
	Spinner._base.Enable(self)
	self:UpdateState()
	return self
end

function Spinner:Disable()
	Spinner._base.Disable(self)
	self:UpdateState()
	return self
end

function Spinner:GetData()
	return self.options[self.idx]
end

function Spinner:GetIndex()
	return self.idx
end

function Spinner:SetIndexForData(data, nofn)
	for k,v in ipairs(self.options) do
		if data == v then
			self:SetIndex( k, nofn )
			break
		end
	end
	return self
end

function Spinner:SetIndex(idx, nofn)
	if idx >= 1 and idx <= #self.options then
		if idx ~= self.idx then
			self.idx = idx
			self:UpdateState()
			if not nofn then
				self:OnSetIndex()
			end
		end
	end
	return self
end

function Spinner:OnSetIndex()
	if self.fn then
		self.fn(self.idx, self.options[self.idx])
	end
end

function Spinner:UpdateState()
	if self.idx > 1 and self.enabled then
		self.left:Enable()
	else
		self.left:Disable()
	end

	if self.idx < #self.options and self.enabled then
		self.right:Enable()
	else
		self.right:Disable()
	end
end

function Spinner:__tostring()
	return "Spinner(" .. tostring(self.idx) ..")" 
end

function Spinner:Right()
	if self.idx < #self.options then
		self.idx = self.idx + 1
		self:UpdateState()
		self:OnSetIndex()
		return true
	end
    return false
end

function Spinner:Left()
	if self.idx > 1 then
		self.idx = self.idx - 1
		self:UpdateState()
		self:OnSetIndex()
		return true
	end
    return false
end

function Spinner:HandleControlDown( control )
    if control:Has( Controls.Digital.MENU_LEFT ) then
        return self:Left()
    elseif control:Has( Controls.Digital.MENU_RIGHT ) then
        return self:Right()
    end

    return false
end
