
------------------------------------------------------------------------------------------
-- Shows navigation tabs
--
-- Example:
--    self.test_navigation = self:AddChild( Widget.TabNavigation( 300, 70, function(tab) print(tab:GetText()) end ) )
--    self.test_navigation:AddTab( Widget.TextTab("TAB 1") )
--    self.test_navigation:AddTab( Widget.TextTab("TAB 2",UICOLOURS.ACCENT) )
--    self.test_navigation:AddTab( Widget.IconTextTab(engine.asset.Texture( "UI/ic_graftscompendium.tex" ),"TAB 3",UICOLOURS.OBJECTIVE) )
--    self.test_navigation:AddTab( Widget.ProgressTab("TAB 4","PROGRESS","25%",1,4,UICOLOURS.FOCUS) )
--    self.test_navigation:AddTab( Widget.ProgressIconTab(engine.asset.Texture( "UI/ic_peoplecompendium.tex" ),"TAB 5","PROGRESS","33%",3,10,UICOLOURS.QUEST) )
--    self.test_navigation:LayoutBounds( "center", "center", self.content_panel )

local TabNavigation = class( "Widget.TabNavigation", Widget )

function TabNavigation:init( tab_width, tab_height, on_select_fn )
    self._base.init( self )

    self.tab_w = tab_width or 200
    self.tab_h = tab_height or 90
    self.tab_spacing = 20
    self.on_select_fn = on_select_fn
end

function TabNavigation:AddTab( new_tab )
    -- Add new tab
    local tab = self:AddChild( new_tab )
        :SetFn( function() self:OnSelectionChanged( new_tab ) end )

    if self.focusable_override ~= nil then tab:SetFocusable( self.focusable_override ) end

    self:Layout()

    return new_tab
end

function TabNavigation:Layout()
	-- Go through all the tabs
    local previous_tab
    for k, v in ipairs(self.children) do
    	-- Set it to the right size
    	v:SetSize( self.tab_w, self.tab_h )
    	-- Position it relative to the preceding one
    	previous_tab = previous_tab or v
        v:LayoutBounds( "after", "bottom", previous_tab )
        	:Offset( self.tab_spacing, 0 )
        previous_tab = v
    end
end

function TabNavigation:OnSelectionChanged( tab )
    self.current_tab = tab

    -- Unselect all tabs
    for k, v in ipairs(self.children) do
        v:SetToggled( false )
    end

    -- Tell the correct tab to show its true colors
    tab:SetToggled( true )

    -- Notify the parent
    if self.on_select_fn then self.on_select_fn( tab ) end
end

function TabNavigation:SetFn( on_select_fn )
    self.on_select_fn = on_select_fn

    return self
end

function TabNavigation:SetTotalWidth( tabs_width )
    self.total_w = tabs_width or 600
    self.tab_w = (self.total_w-(self.tab_spacing*(#self.children-1)))/#self.children
    self.tab_h = self.tab_h or 90

    self:Layout()
    
    return self
end

function TabNavigation:SetTabSize( tab_width, tab_height )
    self.tab_w = tab_width or 200
    self.tab_h = tab_height or 90

    self:Layout()
    
    return self
end

function TabNavigation:SetSpacing( spacing )
    self.tab_spacing = spacing

    self:Layout()

    return self
end

function TabNavigation:SetBGColour( color )
    for k, v in ipairs(self.children) do
        v:SetBGColour( color )
    end

    return self
end

function TabNavigation:SetCurrent( pos )
    if pos < 1 then pos = #self.children end
    if pos > #self.children then pos = 1 end

    local target_tab
    for k, v in ipairs(self.children) do
        if k == pos then
            target_tab = v
        end
    end
    if target_tab then self:OnSelectionChanged( target_tab ) end

    return self
end

function TabNavigation:GetCurrent()
    return self.current_tab
end

function TabNavigation:GetCurrentIdx()
    for k, v in ipairs(self.children) do
        if self.current_tab == v then
            return k
        end
    end
    return 1
end

function TabNavigation:GetTabs()
    return self.children
end

function TabNavigation:GetTab(idx)
    return self.children and #self.children >= idx and self.children[idx]
end

function TabNavigation:GetDefaultFocus()
    return self.children and self.children[1]
end

function TabNavigation:PrepareAnimation()
    for i, widget in ipairs(self.children) do
        widget:SetTintAlpha( 0 )
    end

    return self
end

function TabNavigation:AnimateTabs()
    local tab_anim_duration_total = 0.3
    local tab_number = math.min( #self.children, 5 )
    local tab_anim_duration = 0.1
    local delay_per_tab = tab_anim_duration_total / tab_number
    local tab_adder = Updater.Parallel()
    for i, widget in ipairs(self.children) do
        tab_adder:Add( 
            Updater.Series{
                Updater.Wait(delay_per_tab*i),
                Updater.Parallel{
                    Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, tab_anim_duration, easing.linear),
                    Updater.Ease( function(v) widget.content:SetScale(v,v) end, 1.2, 1, tab_anim_duration, easing.linear)
                }
            }
        )
    end

    self:RunUpdater( tab_adder )

    return self
end

function TabNavigation:SetFocusable( focusable )
    self.focusable_override = focusable

    -- Update all tabs
    for k, v in ipairs(self.children) do
        v:SetFocusable( self.focusable_override )
    end

    return self
end

------------------------------------------------------------------------------------------
-- Shows a single tab with a background and text

local TextTab = class( "Widget.TextTab", Widget.Clickable )

function TextTab:init( text, color )
    TextTab._base.init( self )

    self.color = color or UICOLOURS.SUBTITLE

    -- Setup sizes
    self.tab_w = 200
    self.tab_h = 90
    self.padding = 10
    self.content_w = self.tab_w - self.padding*2

    -- Get a background going
    self.bg_back = self:AddChild( Widget.Panel( engine.asset.Texture ("UI/compendium_tab_normal.tex") ) )
        :SetNineSliceBorderScale( 0.5 )
        :SetTintColour( self.color )
        :SetTintAlpha( 0.1 )
        :SetSize( self.tab_w, self.tab_h )
        :Bloom( 0.1 )
    self.bg_selection = self:AddChild( Widget.Panel( engine.asset.Texture ("UI/compendium_tab_selected.tex") ) )
        :SetNineSliceBorderScale( 0.5 )
        :SetTintColour( self.color )
        :SetTintAlpha( 0 )
        :SetSize( self.tab_w, self.tab_h )
        :Bloom( 0.1 )

    -- And now content
    self.content = self:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
    self.title = self.content:AddChild( Widget.Label("title", FONT_SIZE.SCREEN_TABS ))
        :LeftAlign()
        :SetGlyphColour( self.color )
        :SetText( text )
        :Bloom(0.1)

    -- Setup click callback stuffs
    self.toggled = false
    self:SetOnClickFn( function( ) self:Clicked() end )

    self:Layout()
end

function TextTab:SetText( text )
	self.title:SetText( text )

	return self
end

function TextTab:GetText( text )
	return self.title:GetText()
end

function TextTab:GetColour()
    return self.color
end

function TextTab:SetFn( on_click_fn )
    self.on_click_fn = on_click_fn

	return self
end

function TextTab:SetPadding( padding )
	self.padding = padding or 10

	return self
end

function TextTab:SetSize( width, height )
    self.tab_w = width or self.tab_w or 200
    self.tab_h = height or self.tab_h or 90
    self.content_w = self.tab_w - self.padding*2
    self.bg_back:SetSize( self.tab_w, self.tab_h )
    self.bg_selection:SetSize( self.tab_w, self.tab_h )
    self:Layout()

	return self
end

function TextTab:Enable()
	TextTab._base.Enable(self)

	self:SetSaturation(1)

    return self
end

function TextTab:Disable()
	TextTab._base.Disable(self)

	self:SetSaturation(0.2)
    self.content:Bloom(0.0)
    self.bg_back:AlphaTo( 0.15, 0.2, easing.outQuad )
    self.bg_selection:AlphaTo( 0, 0.2, easing.outQuad )
    return self
end

function TextTab:SetBGColour( color )
    self.bg_back:SetTintColour( color )
    self.bg_selection:SetTintColour( color )

	return self
end

function TextTab:Layout()
    self.content:LayoutBounds( "left", "center", self.bg_back )
        :Offset( self.padding, ( self.toggled or self.has_hover ) and 4 or 2 )

	return self
end

function TextTab:UpdateImage()
    if self.enabled and self.toggled and ( self.hover or self.focus ) then
        self.bg_back:AlphaTo( 0.15, 0.2, easing.outQuad )
        self.bg_selection:AlphaTo( 1, 0.2, easing.outQuad )
            :Bloom(0.45)
        self.content:Bloom(0.15)
    elseif self.enabled and self.toggled and not ( self.hover or self.focus ) then
        self.bg_back:AlphaTo( 0.15, 0.2, easing.outQuad )
        self.bg_selection:AlphaTo( 1, 0.2, easing.outQuad )
            :Bloom(0.2)
        self.content:Bloom(0.15)
    elseif self.enabled and ( self.hover or self.focus ) then
        self.bg_back:AlphaTo( 0.15, 0.2, easing.outQuad )
        self.bg_selection:AlphaTo( 1, 0.2, easing.outQuad )
            :Bloom(0.15)
        self.content:Bloom(0.15)
    elseif self.enabled then
        self.bg_back:AlphaTo( 0.1, 0.2, easing.outQuad )
        self.bg_selection:AlphaTo( 0, 0.2, easing.outQuad )
            :Bloom(0.1)
        self.content:Bloom(0.1)
    end
end

function TextTab:Clicked()
    if self.on_click_fn then self.on_click_fn( self ) end
    self:UpdateImage()
end

function TextTab:IsToggled()
    return self.toggled
end

function TextTab:SetToggled( toggled )
    self.toggled = toggled
    self:UpdateImage()
    return self
end

function TextTab:OnGainFocus()
    TextTab._base.OnGainFocus(self)
    self:UpdateImage()
end

function TextTab:OnLoseFocus()
    TextTab._base.OnLoseFocus(self)
    self:UpdateImage()
end

function TextTab:OnGainHover()
    TextTab._base.OnGainHover(self)
    self:UpdateImage()
end

function TextTab:OnLoseHover()
    TextTab._base.OnLoseHover(self)
    self:UpdateImage()
end


------------------------------------------------------------------------------------------
-- Shows a single tab with an icon and text

local IconTextTab = class( "Widget.IconTextTab", Widget.TextTab )

function IconTextTab:init( icon, text, color )
    IconTextTab._base.init( self, text, color )

    self.icon_texture = icon or engine.asset.Texture ("UI/ic_checkmark.tex")
    self.icon_size = self.icon_size or 50

    -- And new content
    self.icon = self.content:AddChild( Widget.Image( self.icon_texture ) )
        :SetSize( self.icon_size, self.icon_size )
        :SetTintColour( self.color )
        :Bloom(0.05)

    self:Layout()
end

function IconTextTab:GetIcon()
    return self.icon_texture
end

function IconTextTab:Layout()
    IconTextTab._base.Layout( self )

    if not self.icon then return self end

    self.title:LayoutBounds( "after", "center", self.icon )
    	:Offset( self.padding, 0 )
    self.content:LayoutBounds( "left", "center", self.bg_back )
        :Offset( self.padding, ( self.toggled or self.has_hover ) and 4 or 2 )

    return self
end


------------------------------------------------------------------------------------------
-- Shows a single tab with a relationship icon, and count

local RelationshipCountTab = class( "Widget.RelationshipCountTab", Widget.IconTextTab )

function RelationshipCountTab:init( relationship, text, count )
    self.relationship = relationship
    local icon = global_images.relationship_icons[ relationship ]
    local color = RELATIONSHIP_COLOURS[ relationship ]
    self.icon_size = 24
    RelationshipCountTab._base.init( self, icon, text )
    self.icon:SetTintColour( UICOLOURS.WHITE )

    self.title:OverrideLineHeight( 22 )
        :SetGlyphColour( color )
    self.count_label = self.content:AddChild( Widget.Label("title", FONT_SIZE.ITEM_DESC ))
        :SetGlyphColour( color )
        :SetText( count )
        :Bloom( 0.1 )

    self:Layout()
end

function RelationshipCountTab:Layout()
    RelationshipCountTab._base.Layout( self )

    self.title:LayoutBounds( "after", "top", self.icon )
        :Offset( self.padding, 6 )
    if self.count_label then 
        self.count_label:LayoutBounds( "center", "below", self.icon )
            :Offset( 0, 3 )
    end
    self.content:LayoutBounds( "left", "center", self.bg_back )
        :Offset( self.padding, ( self.toggled or self.has_hover ) and 4 or 2 )

    return self
end


------------------------------------------------------------------------------------------
-- Shows a single tab with a relationship icon, and count

local IconCountTab = class( "Widget.IconCountTab", Widget.IconTextTab )

function IconCountTab:init( icon, text, count, color )
    self.icon_size = 24
    IconCountTab._base.init( self, icon, text, color )

    self.title:OverrideLineHeight( 22 )
        :SetGlyphColour( self.color )
    self.count_label = self.content:AddChild( Widget.Label("title", FONT_SIZE.ITEM_DESC ))
        :SetGlyphColour( self.color )
        :SetText( count )
        :Bloom( 0.1 )

    self:Layout()
end

function IconCountTab:Layout()
    IconCountTab._base.Layout( self )

    self.title:LayoutBounds( "after", "center", self.icon )
        :Offset( self.padding, -10 )
    if self.count_label then 
        self.count_label:LayoutBounds( "center", "below", self.icon )
            :Offset( 0, 3 )
    end
    self.content:LayoutBounds( "left", "center", self.bg_back )
        :Offset( self.padding, ( self.toggled or self.has_hover ) and 1 or -1 )

    return self
end

------------------------------------------------------------------------------------------
-- Shows a single tab with text and a progress bar

local ProgressTab = class( "Widget.ProgressTab", Widget.TextTab )

function ProgressTab:init( text, subtitle_text, value_text, current, max, color )
    ProgressTab._base.init( self, text, color )
    
    -- And new content
    self.subtitle = self.content:AddChild( Widget.Label("body", 22 ))
        :SetGlyphColour( self.color )
        :SetTintAlpha( 0.8 )
        :SetText( subtitle_text or "" )
        :LeftAlign()
        :Bloom(0.05)
        :Hide()
    self.value = self.content:AddChild( Widget.Label("body", 22 ))
        :SetGlyphColour( self.color )
        :SetTintAlpha( 0.8 )
        :SetText( value_text or "" )
        :RightAlign()
        :Bloom(0.05)
        :Hide()
    self.progress_bar = self.content:AddChild( Widget.ProgressBar( self.widget_w, 3 ) )
        :OrientH()
        :SetOutlineColour( 0x00000000 )
        :SetBGColour( 0x00000050 )
        :SetBarColour( self.color )
        :SetBarBloom( 0.05 )
        :Hide()

    self:SetProgress( current, max )

    self:Layout()
end

function ProgressTab:SetProgress( current, max, value_text )
    self.current = current or 0
    self.max = max or 0

    self.value:SetText( value_text or ( self.current .. "/" .. self.max ) )
    self.progress_bar:SetProgress( self.current/self.max )
    self.subtitle:Show()
    self.value:Show()
    self.progress_bar:Show()

    self:Layout()

    return self
end

function TextTab:SetSubtitle( subtitle )
	self.subtitle:SetText( subtitle )

    self:Layout()

	return self
end

function TextTab:SetValue( value )
	self.value:SetText( value )

    self:Layout()

	return self
end

function ProgressTab:SetSize( width, height )
    ProgressTab._base.SetSize( self, width, height )

    if not self.subtitle then return self end

    self.progress_bar:SetSize( self.content_w, 3 )
    self:Layout()

    return self
end

function ProgressTab:Layout()
    ProgressTab._base.Layout( self )

    if not self.subtitle then return self end

    self.subtitle:LayoutBounds( "left", "below", self.title )
        :Offset( 0, 5 )
    self.progress_bar:LayoutBounds( "left", "below", self.subtitle )
        :Offset( 0, -1 )
    self.value:LayoutBounds( "right", "above", self.progress_bar )
        :Offset( 0, 1 )

    self.content:LayoutBounds( "left", "center", self.bg_back )
        :Offset( self.padding, ( self.toggled or self.has_hover ) and 6 or 4 )

    return self
end

------------------------------------------------------------------------------------------
-- Shows a single tab with text, icon and a progress bar

local ProgressIconTab = class( "Widget.ProgressIconTab", Widget.ProgressTab )

function ProgressIconTab:init( icon, text, subtitle_text, value_text, current, max, color )
    ProgressIconTab._base.init( self, text, subtitle_text, value_text, current, max, color )

    self.icon_texture = icon or engine.asset.Texture ("UI/ic_checkmark.tex")
    self.icon_size = 50

    -- And new content
    self.icon = self:AddChild( Widget.Image( self.icon_texture ) )
        :SetSize( self.icon_size, self.icon_size )
        :SetTintColour( self.color )
        :Bloom(0.05)

    self:Layout()
end

function ProgressIconTab:SetSize( width, height )
    ProgressIconTab._base.SetSize( self, width, height )

    if not self.icon then return self end

    self.progress_bar:SetSize( self.content_w - self.icon_size - self.padding, 3 )
    self:Layout()

    return self
end

function ProgressIconTab:GetIcon()
    return self.icon_texture
end

function ProgressIconTab:Layout()
    ProgressIconTab._base.Layout( self )

    if not self.icon then return self end

    self.title:LayoutBounds( "after", "center", self.icon )
    	:Offset( self.padding, 0 )
    self.subtitle:LayoutBounds( "left", "below", self.title )
        :Offset( 0, 5 )
    self.progress_bar:LayoutBounds( "left", "below", self.subtitle )
        :Offset( 0, -1 )
    self.value:LayoutBounds( "right", "above", self.progress_bar )
        :Offset( 0, 1 )

    self.content:LayoutBounds( "right", "center", self.bg_back )
        :Offset( -self.padding, ( self.toggled or self.has_hover ) and 6 or 4 )
    self.icon:LayoutBounds( "before", "center", self.content )
    	:Offset( -self.padding, 0 )

    return self
end