require "ui/widget"

--------------------------------------------------------------------

local ARROW_SIZE = 14
local ARROW_SPACING = 44
local ARROW_SPD = 2
local PULSE_SPEED = 6
local TINTS =
{
    [ TARGET_STATE.VALID ] = 0x25FFFFF0,
    [ TARGET_STATE.INVALID ] = 0xBF2D2DF0,
    [ TARGET_STATE.NONE ] = 0xB2B2B2F0,
    [ "INVISIBLE" ] = 0x00000000,
}

local SOUNDS =
{
    [ TARGET_STATE.VALID ] = "event:/sfx/battle/cards/neg/argument_target_valid",
    [ TARGET_STATE.INVALID ] = "event:/sfx/battle/cards/neg/argument_target_notvalid",
}

local assets = {
    arrow = engine.asset.Texture("UI/card_arrow.tex")
}

--------------------------------------------------------------------

local TargetArrow = class( "Widget.TargetArrow", Widget )

function TargetArrow:init( spline_fn )
    Widget.init( self )
    self.segs = {}
    self.spline_fn = spline_fn
end

function TargetArrow:OnAdded()
    self:StartUpdating()
end

function TargetArrow:AttachSource( source )
    assert( is_instance( source, Widget ))
    self.source = source
end

function TargetArrow:AttachTarget( target, state )
    assert( is_instance( target, Widget ))
    state = state or TARGET_STATE.NONE

    if target ~= self.target or state ~= self.target_state then
        self.target = target
        self.target_pos = nil
        self.target_mouse = nil
        self.target_state = state

        self:SetTintColour( (state == TARGET_STATE.NONE and TheGame:FE():IsGamepadMode()) and TINTS["INVISIBLE"] or TINTS[ state ] )
        if SOUNDS[ state ] then
            AUDIO:PlayEvent( SOUNDS[ state ] )
        end
    end
end

function TargetArrow:AttachToPosition( x, y, state )
    state = state or TARGET_STATE.NONE
    if not self.target_pos or self.target_state ~= state then
        self.target = nil
        self.target_mouse = nil
        self.target_state = state

        self:SetTintColour( (state == TARGET_STATE.NONE and TheGame:FE():IsGamepadMode()) and TINTS["INVISIBLE"] or TINTS[ state ] )
        if SOUNDS[ state ] then
            AUDIO:PlayEvent( SOUNDS[ state ] )
        end
    end
    self.target_pos = { x = x, y = y }
end

function TargetArrow:AttachMouseTarget( state )
    state = state or TARGET_STATE.NONE
    if self.target_mouse ~= true or self.target_state ~= state then
        self.target = nil
        self.target_pos = nil
        self.target_mouse = true
        self.target_state = state

        self:SetTintColour( (state == TARGET_STATE.NONE and TheGame:FE():IsGamepadMode()) and TINTS["INVISIBLE"] or TINTS[ state ] )
        if SOUNDS[ state ] then
            AUDIO:PlayEvent( SOUNDS[ state ] )
        end
    end
end

function TargetArrow:Detach()
    self.target, self.source = nil, nil
    self.target_pos = nil
    self.target_mouse = nil
end

local function GetEvenSpacedPoints(pathx, pathy, outx,outy,outr, offset_p)
    
    local dist_accum = (1-offset_p) * ARROW_SPACING
    for k = 2, #pathx do
        local x1, y1 = pathx[k-1], pathy[k-1]
        local x2, y2 = pathx[k], pathy[k]
        if k == 2 then
            local r = math.atan2(y2-y1, x2-x1)
            table.insert(outx, x1)
            table.insert(outy, y1)
            table.insert(outr, r-90*DEGREES)
        end

        local seg_length = distance(x1,y1,x2,y2)
        if seg_length > 0 then
            local nx, ny = (x2 - x1)/seg_length, (y2 - y1)/seg_length
            while dist_accum + seg_length >= ARROW_SPACING do
                local dist_left = ARROW_SPACING - dist_accum
                dist_accum = 0
                local x, y = x1 + nx*dist_left, y1 + ny*dist_left
                seg_length = math.max(0, seg_length - dist_left)
                
                local r = math.atan2(y2-y1, x2-x1)
                table.insert(outx, x)
                table.insert(outy, y)
                table.insert(outr, r-90*DEGREES)
            end

            dist_accum = dist_accum + seg_length
            
            if k == #pathx then
                local r = math.atan2(y2-y1, x2-x1)
                table.insert(outx, x2)
                table.insert(outy, y2)
                table.insert(outr, r-90*DEGREES)
            end            
        end
    end

    --print (total_length, num_nodes, ideal_path_between_spaces, path_between_spaces, #out_path)
end

function TargetArrow:OnUpdate( dt )
    if self.source and self.source.removed then
        self.source = nil
    end
    if self.target and self.target.removed then
        self.target = nil
    end

    if self.source and (self.target or self.target_pos or (self.target_mouse and not TheGame:GetInput():IsMousePosReset()) ) then
        local x1, y1 = self:TransformFromWidget( self.source, 0, 0 )
        local x2, y2
        if self.target_mouse then
            x2, y2 = self:TransformFromWorld( TheGame:GetInput():GetMousePos() )
        elseif self.target_pos then
            x2, y2 = self.target_pos.x, self.target_pos.y
        else
            x2, y2 = self:TransformFromWidget( self.target, 0, 0 )
        end
        local in_path_x, in_path_y = ObtainWorkTable(), ObtainWorkTable()
        if self.spline_fn then
            self.spline_fn( self, x1, y1, x2, y2, in_path_x, in_path_y )
        else
            self:CalculateSpline( x1, y1, x2, y2, in_path_x, in_path_y)
        end

        local outx,outy,outr= ObtainWorkTable(),ObtainWorkTable(),ObtainWorkTable()

        self.offset_t = self.offset_t or 0
        if self.target_state == TARGET_STATE.VALID then
            self.offset_t = self.offset_t + dt
        end
        
        local spd = 1/ARROW_SPD
        local offset_p = (self.offset_t % spd)/spd
        GetEvenSpacedPoints(in_path_x, in_path_y, outx, outy, outr, offset_p)
        
        local last_sc
        for i =  1, #outx do
            self.segs[i] = self:GetSegment( i )
            local x,y,r = outx[i],outy[i], outr[i]
            self.segs[i]:SetPos( x, y):SetRotation(r):Show()
            local t = (i-1) / #outx
            local sc = 1.5 + .25*math.cos(2*t*PI)
            
            if i == #outx or i == 1 then
                if self.target_state == TARGET_STATE.VALID then
                    local pulse = 2 + math.cos(self.offset_t*PULSE_SPEED)*.25
                    self.segs[i]:SetScale(pulse,pulse)
                else
                    self.segs[i]:SetScale(1,1)
                end
            else
                local my_sc = easing.linear(offset_p, last_sc, sc-last_sc, 1)
                self.segs[i]:SetScale(my_sc,my_sc)
            end
            last_sc = sc
        end


        for i = #outx + 1, #self.segs do
            self.segs[i]:Hide()
        end

        ReleaseWorkTable(outx)
        ReleaseWorkTable(outy)
        ReleaseWorkTable(outr)
        ReleaseWorkTable(in_path_x)
        ReleaseWorkTable(in_path_y)
    else
        for i, seg in ipairs( self.segs ) do
            seg:Hide()
        end
    end
end

function TargetArrow:GetSegment( idx )
    if self.segs[idx] == nil then
        self.segs[idx] = self:AddChild( Widget.Image( assets.arrow, ARROW_SIZE, ARROW_SIZE ):SetBloom(UI_BLOOM*.6) )
    end
    return self.segs[idx]
end

function TargetArrow:CalculateSpline( x1, y1, x2, y2, xs, ys )
    -- Based on http://en.wikibooks.org/wiki/Cg_Programming/Unity/Hermite_Curves
    -- Also see http://en.wikipedia.org/wiki/Cubic_Hermite_spline
    local dist = distance(x1,y1,x2,y2)
    local dx1, dy1 = 0, dist*2
    local dx2, dy2 = dist * math.cos(math.pi/4), dist * -math.sin(math.pi/4)
    if x2 < x1 then
        dx2 = -dx2
    end

    return self:CalculateDirectionalSpline( x1, y1, x2, y2, dx1, dy1, dx2, dy2, xs, ys )
end

function TargetArrow:CalculateDirectionalSpline( x1, y1, x2, y2, dx1, dy1, dx2, dy2, xs, ys )

    local numSegments = 100--math.ceil(math.sqrt(dist/2))

    local t = 0

    local xs = xs or {}
    local ys = ys or {}
    local px, py

    local step = 1/numSegments

    for i = 0, numSegments do 
        t = i * step
 
        local h00 = (2.0*t*t*t - 3.0*t*t + 1)
        local h10 = (t*t*t - 2.0*t*t + t) 
        local h01 = (-2.0*t*t*t + 3.0*t*t)
        local h11 = (t*t*t - t*t)

        px = x1*h00 + dx1*h10 + x2*h01 + dx2*h11
        py = y1*h00 + dy1*h10 + y2*h01 + dy2*h11

        table.insert(xs, px)
        table.insert(ys, py)
        
    end

    return xs, ys
end

