require "ui/widget"
require "ui/widgets/anim"
require "ui/widgets/clickable"
require "ui/widgets/healthbar"
require "ui/widgets/floatlabel"
require "ui/widgets/conditionimage"
require "ui/widgets/tooltip_condition"
local battle_defs = require "battle/battle_defs"

-- Damage floater scaling: from MIN_DMG_SCALE to MAX_DMG_SCALE (achieved at MAX_DMG_VALUE)
local MIN_DMG_SCALE = 0.8
local MAX_DMG_SCALE = 1.5
local MAX_DMG_VALUE = 20

-- Delay between queued floaters
local FLOATER_DELAY = 0.4
-----------------------------------------------------------------------------------------

local UnitFrame = class( "Widget.UnitFrame", Widget )

function UnitFrame:init( fighter, ent )
    UnitFrame._base.init( self )

    self:SetName( fighter:GetName() )

    self.hover_check = true
    self.fighter = fighter
    self.entity = ent
    self.floater_queue = {}

    self.elapsed_time = 0

    -- Add our fancy health bar, which has sub-bars inside and such
    self.health_bar = fighter.fight_data.healthbar_class and fighter.fight_data.healthbar_class(fighter) or Widget.HealthBar( fighter )

    if fighter:GetHealthHidden() then
        self.health_bar:Hide()
    end

    -- Container for all the widgets above the fighter's head
    self.top_widgets = Widget()

    -- Container for all the widgets below the fighter's feet
    self.bottom_widgets = Widget()

    -- This fighter's name
    self.name_label = Widget.Label( "title", 30, fighter:GetName() )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :EnableOutline( true )
        :Bloom( 0.15 )
        :SetTintAlpha( 0 )
        :Hide()

    self.subtitle_label = Widget.Label( "title", 24 )
        :SetGlyphColour( UICOLOURS.SUBTITLE )
        :EnableOutline( true )
        :Bloom( 0.1 )
        :Hide()

    -- Contains all the upcoming effect widgets, including damage ones
    self.upcoming_status_effects = Widget()
    -- Contains all the current effect widgets
    self.current_status_effects = Widget()
    self.condition_widgets = {}
    self.morale_icon = self.current_status_effects:AddChild( Widget.UnitFrameMoraleIcon() )
        :Refresh( self.fighter )

    -- Setup our widget hierarchy
    self.root = self:AddChild( Widget() ):AddChildren{
        self.top_widgets:AddChildren{
            self.upcoming_status_effects,
            self.name_label,
            self.subtitle_label,
        },
        self.bottom_widgets:AddChildren{
            self.current_status_effects,
            self.health_bar,
        }
    }
end

function UnitFrame:OnAdded()
    self:StartUpdating()
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function UnitFrame:OnScreenModeChange( sm )
    UnitFrame._base.OnScreenModeChange( self, sm )

    if sm == SCREEN_MODE.MONITOR then
        self.top_widgets:SetLayoutScale( 1 )
        self.current_status_effects:SetLayoutScale( 1 )
    elseif sm == SCREEN_MODE.SMALL
        or sm == SCREEN_MODE.TV then
        self.top_widgets:SetLayoutScale( 1.2 )
        self.current_status_effects:SetLayoutScale( 1.1 )
    end

    self.health_bar:SetLayoutScale( LAYOUT_SCALE[sm] )
end

function UnitFrame:OnUpdate( dt )
    self.elapsed_time = self.elapsed_time + dt
    if #self.floater_queue > 0 then
        if self.last_floater_time == nil or (self.elapsed_time - self.last_floater_time) > FLOATER_DELAY then
            local floater = table.remove( self.floater_queue, 1 )
            self:PlayFloater(floater.txt, floater.icon, floater.keyframes, floater.fx_name, floater.sfx)
        end
    end

    UnitFrame._base.OnUpdate( self, dt )
end

function UnitFrame:HasFloaters( dt )
    if #self.floater_queue > 0 then
        return true
    end
    if self.last_floater_time and self.elapsed_time - self.last_floater_time <= dt then
        return true
    end
    return false
end

local function PreviewExists( self, preview )
    -- Check if a preview already exists on the self.upcoming_status_data list
    for i, widget in ipairs( self.upcoming_status_effects.children ) do
        if widget.preview_data.type and widget.preview_data.type == preview.type then return widget end
        if ( (widget.preview_data.card and widget.preview_data.card == preview.card )or (widget.preview_data.condition and widget.preview_data.condition == preview.condition) )
        and widget.preview_data.hit_index == preview.hit_index then
        
            return widget end
        --origin and index should be enough to identify a hit for re-use. Actual damage numbers can change (due to conditions, and we actually want to consider the hits to be the same.)
        --and widget.preview_data.min_damage == preview.min_damage
        --and widget.preview_data.max_damage == preview.max_damage then
    end
    return false
end

function UnitFrame:RefreshDamagePreview( previews )

    -- Create new table for all the final preview widgets
    local final_widgets = {}

    -- Go through the new previews
    local damage, piercing_damage = 0, 0
    for i, preview in ipairs( previews ) do
        -- Check if it's a showable preview
        if preview and Widget.UnitFrameIcon.IsShowable( preview ) then
            -- Check if there's already a widget for it
            local preview_widget = PreviewExists( self, preview )
            if preview_widget then
                -- Is this widget animating out?
                if preview_widget:IsFadingOut() then
                    -- Remove this widget straight away to make room for the new one
                    preview_widget:Remove()
                    preview_widget = nil
                else
                    -- Update it, and move it to the final previews table
                    preview_widget:Refresh( preview )
                    table.insert( final_widgets, preview_widget )
                end
            end
            if not preview_widget then
                -- Create a new widget and add it
                local preview_widget = Widget.UnitFrameIcon( preview )
                self.upcoming_status_effects:AddChild( preview_widget )
                table.insert( final_widgets, preview_widget )
            end
            -- Update the total damage
            if preview.piercing then
                piercing_damage = piercing_damage + (preview.min_damage or 0)
            else
                damage = damage + (preview.min_damage or 0)
            end
        end
    end

    -- Go through the conditions
    local condition_damage = 0
    for k,condition in pairs(self.fighter:GetConditions()) do
        if condition.CalculatePreviewDamage then
            condition_damage = condition_damage + condition:CalculatePreviewDamage()
        end
    end

    -- Calculate piercing damage
    piercing_damage = piercing_damage + condition_damage

    -- Check if this damage is enough to make the agent surrender or die
    if (damage + piercing_damage) > 0 then
        local defend = self.fighter:GetConditionStacks( "DEFEND" )
        local health, max_health = self.fighter:GetHealth(), self.fighter:GetMaxHealth()
        local health_left = health - math.max( 0, damage - defend ) - piercing_damage
        if health_left <= 0 and not self.fighter:GetHealthHidden() then
            -- Check if there's already a widget for imminent death
            local preview_widget = PreviewExists( self, { type = "death" } )
            if preview_widget then
                -- Is this widget animating out?
                if preview_widget:IsFadingOut() then
                    -- Remove this widget straight away to make room for the new one
                    preview_widget:Remove()
                    preview_widget = nil
                else
                    -- Update it, and move it to the final previews table
                    table.insert( final_widgets, preview_widget )
                end
            else
                -- Create a new widget and add it
                local preview_data = {
                    type = "death",
                    tt = LOC"UI.UNIT_FRAME.ABOUT_TO_DIE_TOOLTIP"
                }
                local preview_widget = Widget.UnitFrameIcon( preview_data )
                self.upcoming_status_effects:AddChild( preview_widget )
                table.insert( final_widgets, preview_widget )
            end
        elseif self.fighter:HasMorale() and health_left <= self.fighter:GetSurrenderHealth() then
            -- Check if there's already a widget for imminent surrender
            local preview_widget = PreviewExists( self, { type = "surrender" } )
            if preview_widget then
                -- Is this widget animating out?
                if preview_widget:IsFadingOut() then
                    -- Remove this widget straight away to make room for the new one
                    preview_widget:Remove()
                    preview_widget = nil
                else
                    -- Update it, and move it to the final previews table
                    table.insert( final_widgets, preview_widget )
                end
            else
                -- Create a new widget and add it
                local preview_data = {
                    type = "surrender",
                    tt = self.fighter:CanSurrender() and LOC"UI.UNIT_FRAME.ABOUT_TO_SURRENDER_TOOLTIP" or LOC"UI.UNIT_FRAME.ABOUT_TO_FLEE_TOOLTIP"
                }
                local preview_widget = Widget.UnitFrameIcon( preview_data )
                self.upcoming_status_effects:AddChild( preview_widget )
                table.insert( final_widgets, preview_widget )
            end
        end
    end

    -- Go through the remaining existing previews
    for i, widget in ipairs( self.upcoming_status_effects.children ) do
        -- Animate any that isn't one of the final ones
        if not table.contains( final_widgets, widget ) then
            widget:FadeOut()
        end
    end

    -- Update the upcoming damage health bar
    self.health_bar:SetDamagePreview( previews )
end

function UnitFrame:ShowDeltaStat( stat, delta, value, mitigated, new_cap, delta_cap, attack )
    if stat == COMBAT_STAT.HEALTH then
        self.health_bar:AnimHealth( value, new_cap, delta_cap, attack )
        self:ShowDamageFloater( delta, mitigated )
        self.morale_icon:Refresh( self.fighter )
    elseif stat == COMBAT_STAT.MORALE then
        self.health_bar:AnimMorale( value )
        self.morale_icon:Refresh( self.fighter )

        local _, max_health = self.fighter:GetStatBounds( COMBAT_STAT.HEALTH )
        self:ShowMoraleFloater( math.ceil(delta*max_health)  )
    end
end

function UnitFrame:HideMorale()
    self.health_bar:HideMorale()
end

function  UnitFrame:ShowMorale()
    self.health_bar:ShowMorale()
end

function UnitFrame:ShowDeltaMoney( delta )
    local widget = Widget.MoneyIndicator( 36, 36, true )
        :SetMoney( delta )
        :SetScale( self:GetFE():GetBaseWidgetScale() )

    local screen = self:GetScreen()
    screen:AddChild( widget )
    screen:RepositionFloaterWidget( widget, self.entity )

    widget:StartCoroutine( function()
        widget:MoveTo( widget.x, widget.y + 100, 0.5, easing.outCubic ):DelayIfAnimating()
        widget:Delay( 0.5 )
        widget:AlphaTo( 0, 0.5 ):DelayIfAnimating()
        widget:Remove()
    end )
end

function UnitFrame:AnimateConditionBreak( condition, w )
    -- local animname = "anim/UI_negotiation_square_fx.zip"
    -- local anim = Widget.Anim()
    --     :SetHiddenBoundingBox(true)
    --     :SetScale( 0.18 )
    --     :SetBloom( 0.1 )
    -- anim:AddAnims( GetAnim( animname ) )
    -- anim:AddOverrideFile( animname )
    
    -- self:AddChild( anim ):SetPos( self:TransformFromWidget( w, 0, 0 ))
    -- anim:Play( "break", false)

    -- anim:AlphaTo( 0, 0.3 )
    -- anim:MoveTo( anim.x, anim.y + 20 * self:GetFE():GetBaseWidgetScale(), 0.3, easing.inQuad,
    --     function() anim:Remove() end )
    local anim = Widget.ConditionImage( 40 )
    anim:SetCondition( condition.id, 1, condition )
    self:AddChild( anim ):SetPos( self:TransformFromWidget( w, 0, 0 ))

    anim:AlphaTo( 0, 0.5 )
    anim:MoveTo( anim.x, anim.y + 40 * self:GetFE():GetBaseWidgetScale(), 0.5, easing.inQuad,
        function() anim:Remove() end )
end

function UnitFrame:ShowCondition(condition, stacks)
    local widget
    for i, w in ipairs( self.condition_widgets ) do
        if w:GetCondition() == condition.id then
            widget = w
            if not w:DeltaStacks( stacks ) then
                table.remove( self.condition_widgets, i )
                self:AnimateConditionBreak( condition, w )
                w:Remove()
            end
            break
        end
    end

    if widget == nil and not condition.hidden and stacks > 0 then
        widget = Widget.ConditionImage( 40 )
        self.current_status_effects:AddChild( widget )
        table.insert( self.condition_widgets, widget )

        widget:SetToolTipClass( Widget.TooltipCondition )
        widget:SetCondition( condition.id, stacks, condition )
        widget:LayoutBounds( "after", "center" ):Offset( 4, 0 )
    end

    if widget then
        widget:SetToolTip( condition )
    end

    return self
end

function UnitFrame:AddCustomStatusEffectWidget( widget, full_row )
    if full_row then
        -- This widget will take the entire bottom row of conditions, and all the others will be above it
        self.custom_status_effect = self.bottom_widgets:AddChild( widget )
    else
        self.current_status_effects:AddChild( widget )
    end
    return self
end

local FLOAT_HEIGHT = 1.7 -- Distance the label will rise
local CRIT_FLOAT_HEIGHT = 2
local FLOAT_DRIFT = 0.5
local FLOAT_DELAY = 1.0 -- Seconds it takes for the label to rise

local CONDITION_FLOATER_KEYFRAMES =
{
    dy = { 0, 1, easing.outQuart, FLOAT_DELAY, FLOAT_HEIGHT },
    scale = { 0, 0.5, easing.inQuad, 0.2, 1.25, easing.outQuad, 0.5, 0.8 },
    alpha = { 0.0, 1.0, easing.linear, 1.0, 1.0, easing.linear, 1.7, 0.0 },
    bloom = { 0.1, 0.3, easing.linear, 0.7, 0.0 },
}

function UnitFrame:ShowConditionFloater( condition, stacks )
    if condition.hidden then
        return
    end
    
    local txt
    if stacks == nil then
        txt = condition:GetName()
    elseif stacks == 1 then
        txt = loc.format( "+{1}", condition:GetName() )
    elseif stacks > 1 then
        txt = loc.format( "+{1} {2}", condition:GetName(), stacks )
    elseif stacks == -1 then
        txt = loc.format( "-{1}", condition:GetName() )
    elseif stacks < -1 then
        txt = loc.format( "-{1} {2}", condition:GetName(), -stacks )
    end

    local img = condition.icon
    if condition.stacks and type(condition.icon) == "table" then
        img = condition.icon[condition.stacks]
    end

    local sfx = condition.apply_sound

    return self:AddFloaterToQueue( txt, img, CONDITION_FLOATER_KEYFRAMES, nil, sfx )
end

function UnitFrame:ShowConditionBlockedFloater( )
    local txt = LOC"UI.UNIT_FRAME.CONDITION_BLOCKED"
    local img 

    return self:AddFloaterToQueue( txt, img, CONDITION_FLOATER_KEYFRAMES )
end

function UnitFrame:ShowAttackFloater( attack_name, duration, colour )
    local attack_colour = MakeColourString( HexColour( colour or UICOLOURS.WHITE ) )
    local txt = loc.format( LOC"UI.UNIT_FRAME.ATTACK_FLOATER", attack_colour, attack_name )
    return self:AddFloaterToQueue( txt, nil, CONDITION_FLOATER_KEYFRAMES )
end

function UnitFrame:ShowTextFloater( txt, size )
    local widget = Widget.FloatLabel( "title", size or 32, txt )
        :SetBloom( 0.1 )
        :SetScale( self:GetFE():GetBaseWidgetScale() )

    local screen = self:GetScreen()
    screen:AddChild( widget )
    screen:RepositionFloaterWidget( widget, self.entity )

    widget:StartFloater()
end


local function DefendFloaterCoro( w )
    w:SetRotation( 0 )
    local sx = w.sx
    -- w:ScaleTo( sx, sx * 2.0, 0.5, easing.outCubic )
    w:MoveTo( w.x, w.y + 120 * w:GetFE():GetBaseWidgetScale(), 0.3, easing.outCubic )
    w:Delay( 0.2 )
    --w:ScaleTo( sx * 3, sx * 2.5, 0.15, easing.outQuad )    w:Delay( 0.3 )
    w:AlphaTo( 0, 1.2 ):DelayIfAnimating()
    w:Remove()
end


function UnitFrame:ShowDamageFloater( delta, mitigated )
    local txt

    if delta > 0 then
        txt = loc.format( "<#STATUS_HEAL>+{1}</>", math.abs(delta) )
    elseif delta < 0 then
        txt = loc.format( "<#STATUS_HEALTH>-{1}</>", math.abs(delta) )
    end

    if txt then
        local t = math.min( math.abs( delta ), MAX_DMG_VALUE )
        local sx = easing.linear( t, MIN_DMG_SCALE, MAX_DMG_SCALE - MIN_DMG_SCALE, MAX_DMG_VALUE )
        local widget = Widget.FloatLabel( "title", 46, txt )
            :SetBloom( 0.1 )
            :SetScale( sx * self:GetFE():GetBaseWidgetScale() )
        local screen = self:GetScreen()
        screen:AddChild( widget )
        screen:RepositionFloaterWidget( widget, self.entity )

        widget:StartFloater()
    end

    if mitigated and mitigated > 0 then
        local widget = Widget.FloatLabel( "title", 38, tostring(mitigated) )
            :SetScale( self:GetFE():GetBaseWidgetScale() )
            :SetGlyphColour( UICOLOURS.DEFEND )
        local screen = self:GetScreen()
        screen:AddChild( widget )
        screen:RepositionFloaterWidget( widget, self.entity )

        widget:StartFloater( DefendFloaterCoro )
    end
end

function UnitFrame:ShowMoraleFloater( delta )
    local txt

    if delta > 0 then
        txt = loc.format( "<#STATUS_HEALTH>-{1}</> MORALE ", math.floor(math.abs(delta)) )
    elseif delta < 0 then
        txt = loc.format( "<#STATUS_HEAL>+{1}</> MORALE", math.floor(math.abs(delta)) )
    end

    if txt then
        local t = math.min( math.abs( delta ), MAX_DMG_VALUE )

        local widget = Widget.FloatLabel( "title", 46, txt )
            :SetBloom( 0.1 )
            :SetScale( self:GetFE():GetBaseWidgetScale() )
        local screen = self:GetScreen()
        screen:AddChild( widget )
        screen:RepositionFloaterWidget( widget, self.entity )

        widget:StartFloater()
    end
end


function UnitFrame:AddFloaterToQueue( txt, icon, keyframes, fx_name, sfx )
    if not self:GetFE():IsHUDVisible() then
        return
    end

    -- print(loc.format("Add Floater To Queue {1}", txt or 'nil'), debugstack())
    table.insert(self.floater_queue, {txt = txt, icon = icon, keyframes = keyframes, fx_name = fx_name, sfx = sfx})
end

function UnitFrame:PlayFloater( txt, icon, keyframes, fx_name, sfx )
    local entity = self.entity
    if entity.removed then
        return
    end

    local floater = entity.sim:SpawnPrefab( "floater" ).cmp.FloatText
    floater:GetModel():SetFont( "title" )
    floater:SetAutoDespawn( true )
    floater:SetKeyFrames( keyframes )
    floater:SetText( txt )
    floater:GetModel():EnableOutline( LookupTuningValue( "FIGHT_FLOATER_OUTLINE" ))

    local wx, wy, wz = entity.cmp.AnimController.animmodel:SymbolToWorld("head_swap", 0, 0, 0)
    wx = wx + (self.fighter.fight_data.head_symbol_dx or 0)
    wy = wy + (self.fighter.fight_data.head_symbol_dy or 0)
    local xflip = entity.cmp.AnimController:GetXFlip()
    local dx, dy = 1 * (xflip and 1 or -1), 1.5
    local bx1, by1, bx2, by2 = floater:GetRenderBounds()
    local now = TheGame:GetWallTime()  

    if sfx then
        local x, y = entity.sim:GetCamera():WorldToScreen( wx, wy, wz )
        local width = TheGame:FE():GetScreenDims()
        local screen_pos = math.abs(x)/width
        local pan_pos = easing.linear( screen_pos, -1, 2, 1 )
        AUDIO:PlayParamEvent(sfx, "position", pan_pos)
    end

    if self.floater_time == nil or now > self.floater_time then
        floater:StartFloater( txt, { icon }, wx + dx, wy + dy, wz )

        self.floater_time = TheGame:GetWallTime()
        self.floater_count = 1
        self.floater_dy = 0
        self.floater_dz = 0
    else
        local t = self.floater_count * 0.2
        self.floater_dy = self.floater_dy - 0.7
        self.floater_dz = self.floater_dz - 0.05
        self.entity:DoTaskInTime( "floater", t, floater.StartFloater, floater, txt, { icon }, wx + dx, wy + dy + self.floater_dy, wz + self.floater_dz )
        self.floater_count = self.floater_count + 1
    end

    if fx_name then
        local FX_SCALE = 1.2
        local FX_BLOOM = .5
        CacheAnim( fx_name )
        local fx = self.entity.sim.fx_mgr:PlayBillboardFX( fx_name, "fx", wx + dx, wy-.25 + dy, wz+.5, FX_SCALE, FX_SCALE, FX_SCALE, 0, -FIGHT_CAM_ANGLE)
        fx.model:SetBloom( FX_BLOOM )
    end

    self.last_floater_time = self.elapsed_time

    return floater
end

function UnitFrame:AnimProtect( has_protect )
    local out_time = 0.30
    local in_time = out_time/3

    if has_protect then
        local scale_back = function()
            --self.bottom_hitbox:ColourFromCurrentTo(232/255, 181/255, 88/255, 0.9, in_time)
            --self.bottom_hitbox:SetBloom(0.15)
            self.health_bar:ScaleTo(1.75, 1, in_time)
        end

        self.health_bar:ScaleTo(1, 1.75, out_time, easing.inQuint, scale_back)
    else
        local scale_back = function()
            --self.bottom_hitbox:ColourFromCurrentTo(1, 1, 1, 0.67, in_time)
            --self.bottom_hitbox:SetBloom(0)
            self.health_bar:ScaleTo(1.75, 1, in_time)
        end

        self.health_bar:ScaleTo(1, 1.75, out_time, easing.inQuint, scale_back)
    end
end

function UnitFrame:AnimDefend( evasion, attacked  )
    self.health_bar:AnimDefend( evasion, attacked  )
end

function UnitFrame:RefreshCondition(condition)
    local widget
    for i, w in ipairs( self.condition_widgets ) do
        if w:GetCondition() == condition.id then
            widget = w
            local show = widget:RefreshConditionImage(condition)
            if not show then
                table.remove( self.condition_widgets, i )
                self:AnimateConditionBreak( condition, w )
                w:Remove()
                break
            end
        end
    end
end

function UnitFrame:FindConditionWidget( condition )
    local widget
    for i, w in ipairs( self.condition_widgets ) do
        if w:GetCondition() == condition.id then
            widget = w
            break
        end
    end
    return widget
end

function UnitFrame:SetHeadPos( x, y )

    -- Convert the world coordinates to the current widget's coordinates
    local new_x, new_y = self:TransformFromWorld( x, y )

    -- Prepare our "old" position, in case there's none so far
    if self.head_x == nil then self.head_x = new_x end
    if self.head_y == nil then self.head_y = new_y end

    -- Ease our new position into position, so the widgets move softly about
    -- local lerp = 0.01
    -- self.head_x = Lerp( self.head_x, new_x, lerp )
    -- self.head_y = Lerp( self.head_y, new_y, lerp )

    self.head_x = new_x
    self.head_y = new_y

    -- Re-layout everything
    self:DoLayout()
end

function UnitFrame:GetMidpoint()
    return self.head_x/2, self.head_y/2
end

function UnitFrame:DoLayout()
    if not self.head_y then
        return
    end

    -- Layout conditions
    self.current_status_effects:LayoutChildrenInGrid( 6, 4 )

    -- Layout the top widgets first
    self.name_label:LayoutBounds( "center", "above", 0, 0 )
    self.subtitle_label:LayoutBounds( "center", "below", self.name_label )
    self.upcoming_status_effects:StackChildrenRow( -5 )
    self.upcoming_status_effects:LayoutBounds( "center", "above", 0, 30 )
    self.top_widgets:LayoutBounds( "center", "center", self.head_x, self.head_y+80 )

    -- Layout the bottom widgets now
    if self.custom_status_effect then
        self.custom_status_effect:LayoutBounds( "center", "above", self.health_bar ):Offset( 0, 4 )
        self.current_status_effects:LayoutBounds( "center", "above", self.custom_status_effect ):Offset( 0, 4 )
    else
        self.current_status_effects:LayoutBounds( "center", "above", self.health_bar ):Offset( 0, 4 )
    end
    self.bottom_widgets:LayoutBounds( "center", "bottom", 0, -20 )

end

function UnitFrame:Refresh()
    local health = self.fighter:GetStat( COMBAT_STAT.HEALTH )
    local morale = self.fighter:GetStat( COMBAT_STAT.MORALE )
    self.health_bar:SetVals( health, morale or 0 )
    self.morale_icon:Refresh( self.fighter )
end

function UnitFrame:SetFn(fn)
    self.fn = fn
    return self
end

function UnitFrame:HandleControlUp(control)
    if control:Has( Controls.Digital.MENU_ACCEPT ) then
        if self.fn then
            self.fn()
            return true
        end            
    end
end

function UnitFrame:HilightDamagePreview( attacker )
    engine.inst:ProfilerPush("UnitFrame:HilightDamagePreview")

    for i, widget in pairs( self.upcoming_status_effects.children ) do
        if widget.card and widget.card.owner == attacker then
            if not widget.hilited then
                widget:SetHighlighted( true )
            end
        else
            if widget.hilited then
                widget:SetHighlighted( false )
            end
        end
    end

    engine.inst:ProfilerPop()
end

function UnitFrame:UpdateImage()
    if self.highlighted or self.hover or self.focus then
        for i, widget in pairs( self.upcoming_status_effects.children ) do
            widget:Bloom( 0.15 )
        end
        self.top_widgets:ScaleTo( nil, 1.1, 0.2, easing.outQuad )
        self.bottom_widgets:ScaleTo( nil, 1.1, 0.2, easing.outQuad )
        self.name_label:Show()
        self.name_label:AlphaTo( 1, 0.1 )
    else
        for i, widget in pairs( self.upcoming_status_effects.children ) do
            widget:Bloom( 0.1 )
        end
        self.top_widgets:ScaleTo( nil, 1, 0.2, easing.outQuad )
        self.bottom_widgets:ScaleTo( nil, 1, 0.2, easing.outQuad )
        self.name_label:AlphaTo( 0, 0.1, nil, function() self.name_label:Hide() end )
    end
end

function UnitFrame:SetHighlighted( highlighted )
    if self.highlighted ~= highlighted then
        self.highlighted = highlighted

        if highlighted and TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD then
            self.subtitle_label:SetText( LOC "UI.UNIT_FRAME.NAME_LABEL" ):Show()
        else
            self.subtitle_label:Hide()
        end

        self:UpdateImage()
    end
    return self
end

function UnitFrame:OnGainFocus()
    self:UpdateImage()
end

function UnitFrame:OnLoseFocus()
    self:UpdateImage()
end

function UnitFrame:OnGainHover()
    UnitFrame._base.OnGainHover(self)
    self:UpdateImage()
    -- want mouse control to take over, so clear focus
    self:GetFE():ClearFocusWidget()
end

function UnitFrame:OnLoseHover()
    UnitFrame._base.OnLoseHover(self)
    self:UpdateImage()
    -- want mouse control to take over, so clear focus
    self:GetFE():ClearFocusWidget()
end

function UnitFrame:CheckHit(x,y)
    return self.top_widgets:CheckHit(x,y) or self.bottom_widgets:CheckHit(x,y)
end

------------------------------------------------------------------------------------------
-- Displays upcoming damage/effect to a character

local UnitFrameIcon = class("Widget.UnitFrameIcon", Widget)

local ICON_W = 50
local ANIM_FILE = "anim/UI_combat_character_intent_fx.zip"
local ANIM_SCALE = 0.12

function UnitFrameIcon:init( preview_data )
    UnitFrameIcon._base.init(self)

    -- Hover hitbox
    self.hitbox = self:AddChild( Widget.SolidBox( ICON_W, ICON_W, 0xff00ff00 ) )

    -- Scalable contents
    self.contents = self:AddChild( Widget() )
        :SetHiddenBoundingBox( true )

    -- Icon anim
    self.anim = self.contents:AddChild( Widget.Anim() )
        :AddAnims( GetAnim( ANIM_FILE ) )
        :AddOverrideFile( ANIM_FILE )
        :SetScale( ANIM_SCALE )

    -- Preview icon
    self.img = self.contents:AddChild( Widget.Image( engine.asset.Texture("images/white.tex"), ICON_W, ICON_W ) )

    -- Value label
    self.label = self.contents:AddChild( Widget.Label( "title", 24 ) )
        :EnableOutline( true )
        :SetGlyphColour( UICOLOURS.WHITE )

    self:SetToolTipClass( Widget.BattleTooltip )

    self:Refresh( preview_data )
    self:Bloom( 0.1 )
    self.initialized = true
end

function UnitFrameIcon:Refresh( preview_data )
    self.preview_data = preview_data
    self.img:SetTintAlpha(0)
    if self.preview_data.card and self.preview_data.card:IsAttackCard() and self.preview_data.piercing then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "armor_pierce_pre", false )
            self.anim:Push( "armor_pierce_loop", true )
            self.anim_pst = "armor_pierce_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.card and self.preview_data.card:IsAttackCard() and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.DEBUFF ) then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "debuff_pre", false )
            self.anim:Push( "debuff_loop", true )
            self.anim_pst = "debuff_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.card and self.preview_data.card:IsAttackCard() and self.preview_data.card.min_damage and self.preview_data.card.max_damage then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "damage_pre", false )
            self.anim:Push( "damage_loop", true )
            self.anim_pst = "damage_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.card and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.DEBUFF ) then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "debuff_pre", false )
            self.anim:Push( "debuff_loop", true )
            self.anim_pst = "debuff_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.card and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.HEAL ) then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "healing_pre", false )
            self.anim:Push( "healing_loop", true )
            self.anim_pst = "healing_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.card and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.BUFF ) then
        self.card = self.preview_data.card
        if not self.initialized then
            self.anim:Play( "action_pre", false )
            self.anim:Push( "action_loop", true )
            self.anim_pst = "action_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    elseif self.preview_data.type == "death" then
        self.card = nil
        if not self.initialized then
            self.anim:Play( "death_pre", false )
            self.anim:Push( "death_loop", true )
            self.anim_pst = "death_pst"
        end
        self.anim:Show()
        local r1, g1, b1, a1 = HexColour( 0xCC6161C6 )
        local r2, g2, b2, a2 = HexColour( UICOLOURS.WHITE )
        self.anim:PulseColour( r1, g1, b1, a1, r2, g2, b2, a2, 0.9 )
        -- Hide the label text
        self.label:Hide()
        -- Set the tooltip
        self:SetToolTip( self.preview_data.tt )
        return self
    elseif self.preview_data.type == "surrender" then
        self.card = nil
        if not self.initialized then
            self.anim:Play( "surrender_pre", false )
            self.anim:Push( "surrender_loop", true )
            self.anim_pst = "surrender_pst"
        end
        self.anim:Show()
        local r1, g1, b1, a1 = HexColour( 0xCC6161C6 )
        local r2, g2, b2, a2 = HexColour( UICOLOURS.WHITE )
        self.anim:PulseColour( r1, g1, b1, a1, r2, g2, b2, a2, 0.9 )
        -- Hide the label text
        self.label:Hide()
        -- Set the tooltip
        self:SetToolTip( self.preview_data.tt )
        return self
    elseif self.preview_data.condition and self.preview_data.condition.id == "malfunctioning_charge_cells" then
        if not self.initialized then
            self.anim:Play( "damage_cell_pre", false )
            self.anim:Push( "damage_cell_loop", true )
            self.anim_pst = "damage_cell_pst"
        end
        self.anim:Show()
        self.anim:StopPulse()
    
    elseif self.preview_data.condition then
        self.condition = self.preview_data.condition
        if not self.initialized then
            if self.condition.owner:IsPlayer() then
                self.anim:Play( "self_damage_pre", false )
                self.anim:Push( "self_damage_loop", true )
                self.anim_pst = "self_damage_pst"
            else
                self.anim:Play( "damage_pre", false )
                self.anim:Push( "damage_loop", true )
                self.anim_pst = "damage_pst"
            end
        end
        self.anim:Show()
        self.anim:StopPulse()
    end

    -- Check if there's an icon override
    if self.preview_data.card and self.preview_data.card.preview_icon then
        if self.preview_data.card.preview_icon then
            self.anim:Hide()
            self.img:SetTintAlpha(1)
                :SetTexture( self.preview_data.card.preview_icon )
        end
    end

    if self.preview_data.min_damage and self.preview_data.max_damage then
        -- Set the label text
        if self.preview_data.min_damage ~= self.preview_data.max_damage then
            self.label:SetText( loc.format( "{1}-{2}", self.preview_data.min_damage, self.preview_data.max_damage ))
        else
            self.label:SetText( tostring( self.preview_data.max_damage ))
        end
        self.label:LayoutBounds( "center", "center", self.img ):Offset( 0, 2 )
        self.label:Show()

        -- Set the tooltip
        local tt


        if self.preview_data.card and self.preview_data.piercing then
            tt = loc.format( LOC"UI.UNIT_FRAME.PIERCING_DAMAGE_RECEIVED_TOOLTIP", self.preview_data.max_damage, self.preview_data.card.owner:GetName() )
        elseif self.preview_data.condition and self.preview_data.condition.id == "malfunctioning_charge_cells" then
            tt = loc.format( LOC"UI.UNIT_FRAME.DAMAGE_RECEIVED_TOOLTIP_CONDITION", self.preview_data.max_damage, self.preview_data.condition:GetName())
        else
            tt = loc.format( LOC"UI.UNIT_FRAME.DAMAGE_RECEIVED_TOOLTIP", self.preview_data.max_damage, self.preview_data.owner:GetName())
        end
        if self.preview_data.card and self.preview_data.card:IsFlaggedAny( battle_defs.CARD_FLAGS.BUFF | battle_defs.CARD_FLAGS.DEBUFF ) then
            tt = tt .. LOC"UI.UNIT_FRAME.ATTACK_STATUS_TOOLTIP"
        end
        self:SetToolTip( tt )

    elseif self.preview_data.card and not self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.BUFF ) then
        self.label:SetText( "-" )
        self.label:LayoutBounds( "center", "center", self.img ):Offset( 0, 2 )
        self.label:Show()
        self:SetToolTip( loc.format( LOC"UI.UNIT_FRAME.STATUS_TOOLTIP", self.preview_data.card.owner:GetName() ) )   

    elseif self.preview_data.card and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.HEAL ) then
        self.label:Hide()
        self:SetToolTip( loc.format( LOC"UI.UNIT_FRAME.HEAL_TOOLTIP", self.preview_data.card.owner:GetName() ) )   

    elseif self.preview_data.card and self.preview_data.card:IsFlagged( battle_defs.CARD_FLAGS.BUFF ) then
        self.label:Hide()
        self:SetToolTip( loc.format( LOC"UI.UNIT_FRAME.STATUS_TOOLTIP", self.preview_data.card.owner:GetName() ) )   

    else
        -- Hide the label text
        self.label:Hide()
        -- Set the tooltip
        self:SetToolTip( nil )   
    end

    return self
end

-- Makes a preview widget animate out, delete itself when done, and re-layout the previews
-- It should be already off the previews list when this is called on it
function UnitFrameIcon:FadeOut( fn )
    self.is_fading_out = true
    self.anim:SetAnimDoneFn( function()
        self:Remove()
        if fn then fn() end
    end )
    if self.anim_pst then self.anim:Push( self.anim_pst, false ) end
end

function UnitFrameIcon:IsFadingOut()
    return self.is_fading_out and self.is_fading_out == true
end

function UnitFrameIcon.IsShowable( preview )
    return ( preview.card and preview.card:IsAttackCard() and preview.piercing )
        or ( preview.card and preview.card:IsAttackCard() and preview.card:IsFlagged( battle_defs.CARD_FLAGS.DEBUFF ) )
        or ( preview.card and preview.card:IsAttackCard() and preview.card.min_damage and preview.card.max_damage )
        or ( preview.card and preview.card:IsFlagged( battle_defs.CARD_FLAGS.DEBUFF ) )
        or ( preview.card and preview.card:IsFlagged( battle_defs.CARD_FLAGS.BUFF ) )
        or preview.type == "death"
        or preview.type == "surrender"
        or preview.condition
end

function UnitFrameIcon:SetHighlighted( highlighted )
    self.hilited = highlighted
    self.contents:ScaleTo( nil, self.hilited and 1.1 or 1, 0.1 )
    self.contents:Bloom( self.hilited and 0.15 or 0.1 )
    return self
end

------------------------------------------------------------------------------------------
-- Displays a health bar

local UnitFrameHealthBar = class("Widget.UnitFrameHealthBar", Widget)

local LINE_HEIGHT = 3
local HEALTH_HEIGHT = 8
local HEALTH_HEIGHT = 8

local BAR_LENGTH_PER_POINT = 3
local MIN_HEALTHBAR_WIDTH = 50
local MAX_HEALTHBAR_WIDTH = 300

function UnitFrameHealthBar:init()
    UnitFrameHealthBar._base.init(self)

    self.bar_back = self:AddChild( Widget.SolidBox( 10, LINE_HEIGHT+HEALTH_HEIGHT, 0x3C3A3Aff ) )
    self.bar_top = self:AddChild( Widget.SolidBox( 10, LINE_HEIGHT, UICOLOURS.SUBTITLE ) )
    self.bar_morale = self:AddChild( Widget.SolidBox( 10, HEALTH_HEIGHT, UICOLOURS.MORALE ) )
    self.bar_health = self:AddChild( Widget.SolidBox( 10, HEALTH_HEIGHT, UICOLOURS.HEALTH ) )

    self.current_health = self.contents:AddChild( Widget.Label( "title", 40 ) )
        -- :EnableOutline( true )
        :SetGlyphColour( UICOLOURS.HEALTH )
    self.total_health = self.contents:AddChild( Widget.Label( "title", 24 ) )
        -- :EnableOutline( true )
        :SetGlyphColour( UICOLOURS.HEALTH )
end

function UnitFrameHealthBar:SetVals( health, morale )
    local bar_w, bar_h = self.bar_container:GetSize()

end

------------------------------------------------------------------------------------------
-- Displays remaining health until morale panic kicks in

local UnitFrameMoraleIcon = class("Widget.UnitFrameMoraleIcon", Widget)

local ICON_W = 40

function UnitFrameMoraleIcon:init()
    UnitFrameMoraleIcon._base.init(self)

    self.hitbox = self:AddChild( Widget.SolidBox( ICON_W, ICON_W, 0xff00ff00 ) )
    self.icon = self:AddChild( Widget.Image( engine.asset.Texture("battle/conditions/ic_morale.tex") ) )
        :SetSize( ICON_W+4, ICON_W+4 )
        :SetHiddenBoundingBox( true )
    self.label = self:AddChild( Widget.Label( "title", 24 ) )
        -- :EnableOutline( true )
        :SetGlyphColour( UICOLOURS.DARK_BODYTEXT )
        :SetHiddenBoundingBox( true )
end

function UnitFrameMoraleIcon:Refresh( fighter )
    if fighter:HasMorale() and fighter:IsActive() then
        local battle_defs = require "battle/battle_defs"
        local health = fighter:GetHealth()
        local surrender_health = fighter:GetSurrenderHealth()
        if surrender_health < health and surrender_health > 0 then
            local surrender_dmg = math.max( 1, health - surrender_health )
            self.label:SetText( surrender_dmg )
                :LayoutBounds( "center", "center", self.hitbox )
                :Offset( 0, -3 )
            local tt = loc.format( LOC "UI.FIGHT_SCREEN.SURRENDER_TT", fighter:GetName(), surrender_dmg )
            self:SetToolTipClass( Widget.BattleTooltip )
            self:SetToolTip( tt )

            if not self:IsShown() then
                self:Show()
            end
        else
            self:Hide()
        end
    else
        self:Hide()
    end
    return self
end