


--------------------------------------------------------------
-- Displays a track of meta unlockable items for a specific character

local UnlockTrack = class( "Widget.UnlockTrack", Widget )

local VERTICAL_PADDING = 15
local SIDE_PADDING = 20

function UnlockTrack:init( width, height )
    UnlockTrack._base.init( self )

    self.hitbox = self:AddChild( Widget.SolidBox( width, height, 0xFF000000 ) )

    self.background = self:AddChild( Widget.Image( engine.asset.Texture("UI/unlocktrack_bg.tex") ) )
        :SetTintColour( UICOLOURS.OBJECTIVE )
        :Bloom( 0.1 )

    self.track_data = nil
    self.unlock_widgets = {}

    self.unlocks_container = self:AddChild( Widget() )

    self:SetSize( width, height )
    self:Layout()
    self:SetFocusable( false )
end

function UnlockTrack:SetGamestate( gamestate )
	if gamestate and gamestate.options and gamestate.options.toggled_off_unlocks then
		self.gamestate = gamestate
	else
		self.gamestate = nil
	end
    return self
end

function UnlockTrack:SetCharacter( pool )
	self.character = pool
	self.track_data = Content.GetUnlockTrack( self.character )
	self.num_unlocked = TheGame:GetGameProfile():GetNumUnlocked( self.character )
    self.unlocks_container:DestroyAllChildren()

    if #self.track_data > 0 then
		for k, unlock in ipairs( self.track_data ) do
			local unlock_toggled_off = TheGame:GetGameProfile():IsUnlockToggledOff( self.character, unlock.id, self.gamestate )
			local unlock_widget = self.unlocks_container:AddChild( Widget.UnlockItem( self.height, self.height ) )
				:Refresh( unlock, k <= self.num_unlocked, not unlock_toggled_off )
			table.insert( self.unlock_widgets, unlock_widget )
		end
		self:Show()
	else
		self:Hide()
	end
    self:Layout()
    return self
end

function UnlockTrack:SetSize( width, height )
	self.width = width or 800
    self.height = height or 60
    self:Layout()
    return self
end

function UnlockTrack:GetBoundingBox()
	return self.hitbox:GetBoundingBox()
end

function UnlockTrack:Layout()
	self.hitbox:SetSize( self.width, self.height )
	self.background:SetSize( self.width-SIDE_PADDING*2, self.height-VERTICAL_PADDING*2 )
	if self.track_data and #self.track_data > 1 then
		local num_unlocks = #self.track_data
		local unlocks_w = self.width + SIDE_PADDING*2
		local side = self.height
		local spacing = ( unlocks_w-(num_unlocks*self.height) ) / (num_unlocks - 1)
		self.unlocks_container:StackChildrenRow( spacing )
	end
	self.unlocks_container:LayoutBounds( "center", "center", self.background )
	return self
end

function UnlockTrack:UnlockItem( unlock, fn )
	for k,v in pairs( self.unlock_widgets) do
        if v.unlock == unlock then
            v:ShowUnlock( fn )
            return
        end
    end
end

function UnlockTrack:GetDefaultFocus()
	for k, v in pairs( self.unlock_widgets) do
        return v
    end
end


--------------------------------------------------------------
-- Displays a meta unlockable item

local UnlockItem = class( "Widget.UnlockItem", Widget.Clickable )

UnlockItem.CONTROL_MAP = 
{
    {
        control = Controls.Digital.MENU_ACCEPT,
        fn = function(self) 
            self:ShowUnlockPopup()
            return true
        end,
        hint = function(self, left, right)
            table.insert(right, loc.format(LOC"UI.UNLOCK_TRACK.CONTROLS_DETAILS", Controls.Digital.MENU_ACCEPT))
        end
    },
    {
        control = Controls.Digital.MENU_ALT,
        fn = function(self) 
            self:ShowUnlockPopup()
            return true
        end
    },
}

function UnlockItem:init( width, height )
    UnlockItem._base.init( self )

    self.hitbox = self:AddChild( Widget.SolidBox( width, height, 0x0000FF00 ) )
    self.glow = self:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ) ) )
        :SetHiddenBoundingBox( true )
        :SetTintColour( UICOLOURS.OBJECTIVE )
        :SetTintAlpha( 0 )
    self.status_icon = self:AddChild( Widget.Image( engine.asset.Texture("UI/ic_item_locked.tex" ) ) )
    	:Bloom( 0.05 )
    self.icon = self:AddChild( Widget.Image( engine.asset.Texture("images/white.tex" ) ) )
        :SetHiddenBoundingBox( true )
    	:Bloom( 0.05 )
	self:SetFocusable(true)
	self:ShowToolTipOnFocus(true)

    self:SetSize( width, height )
    self:Layout()
end

function UnlockItem:Refresh( unlock, unlocked, toggled )
	self.unlock = unlock
	self.unlocked = unlocked
	self.toggled = toggled
	if self.unlocked == true then
		self.icon:SetTexture( self.unlock.icon )
		local tt = loc.format( self.toggled and LOC"UI.UNLOCK_TRACK.UNLOCKED_ACTIVE_TOOLTIP" or LOC"UI.UNLOCK_TRACK.UNLOCKED_DISABLED_TOOLTIP", self.unlock:GetLocalizedName() )
		if TheGame:FE():GetControlMode() == CONTROL_MODE.MOUSE_KEYBOARD then
			tt = tt .. LOC"UI.UNLOCK_TRACK.CONTROLS_TOOLTIP_SUFFIX"
		end
		self:SetToolTip( tt )
		self.icon:SetSaturation( self.toggled and 1 or 0 )
		self.status_icon:SetTexture( self.toggled and engine.asset.Texture("UI/ic_item_enabled.tex" ) or engine.asset.Texture("UI/ic_item_disabled.tex" ) )
	end
	if self.unlocked == false then
		self.icon:SetTexture( self.unlock.icon_locked )
		self:SetToolTip( loc.format( LOC"UI.UNLOCK_TRACK.LOCKED_TOOLTIP", self.unlock:GetLocalizedName() ) )
		self.status_icon:SetTexture( engine.asset.Texture("UI/ic_item_locked.tex" ) )
	end

	return self
end

function UnlockItem:SetSize( width, height )
	self.width = width or 60
    self.height = height or 60
    self:Layout()
    return self
end

function UnlockItem:Layout()
	self.hitbox:SetSize( self.width, self.height )
	self.glow:SetSize( self.width, self.height )
	self.status_icon:SetSize( self.width*0.2, self.height*0.2 )
		:LayoutBounds( "right", "top", self.hitbox )
		:Offset( -10, -8 )
	self.icon:SetSize( self.width, self.height )
	return self
end

function UnlockItem:ShowUnlock( fn )
	local duration = 0.3
	self:RunUpdater(
		Updater.Series{
	        Updater.Parallel{
	            Updater.Ease( function(v) self.glow:SetTintAlpha(v) end, 0, 1, duration, easing.linear),
	            Updater.Ease( function(v) self.glow:SetScale(v,v) end, 1, 2.5, duration, easing.linear),
	            Updater.Ease( function(v) self.icon:SetScale(v,v) end, 1, 1.2, duration, easing.linear),
	        },
	        Updater.Ease( function(v) self.glow:SetScale(v,v) end, 2.5, 3, duration, easing.linear),
	        Updater.Parallel{
	        	Updater.Do( function()
	        		self.icon:SetTexture( self.unlock.icon )
	        	end ),
	            Updater.Ease( function(v) self.glow:SetTintAlpha(v) end, 1, 0.2, duration, easing.linear),
	            Updater.Ease( function(v) self.glow:SetScale(v,v) end, 3, 2, duration, easing.linear),
	            Updater.Ease( function(v) self.icon:SetScale(v,v) end, 1.2, 1, duration, easing.linear),
	        },
	        Updater.Wait( duration ),
	        Updater.Do( function() fn() end )
	    }
    )
    self:Refresh( self.unlock, true, true )
end
function UnlockItem:OnControlModeChange( cm, device_id )
    UnlockItem._base.OnControlModeChange( self, cm, device_id )

    -- Update tooltips
    self:Refresh( self.unlock, self.unlocked, self.toggled )
end

function UnlockItem:ShowUnlockPopup()
	if self.unlocked then
		TheGame:FE():InsertScreen( Screen.UnlockInspectorScreen( self.unlock, false ) )
	end
    return self
end

function UnlockItem:UpdateImage()
	if self.enabled and ( self.hover or self.focus ) then
		self.icon:ScaleTo( nil, 1.1, 0.1, easing.inOutQuad )
	else
		self.icon:ScaleTo( nil, 1, 0.1, easing.inOutQuad )
	end
    return self
end

function UnlockItem:OnGainFocus()
    UnlockItem._base.OnGainFocus(self)
    self:UpdateImage()
end

function UnlockItem:OnLoseFocus()
    UnlockItem._base.OnLoseFocus(self)
    self:UpdateImage()
end

function UnlockItem:OnGainHover()
    UnlockItem._base.OnGainHover(self)
    self:UpdateImage()
end

function UnlockItem:OnLoseHover()
    UnlockItem._base.OnLoseHover(self)
    self:UpdateImage()
end
