 local VSTeamDisplay = class("Widget.VSTeamDisplay", Widget)

function VSTeamDisplay:init( left, card_type )
    VSTeamDisplay._base.init( self )

    self.left = left
    self.card_type = card_type
    self.bg_colour = self.card_type == QCARD_TYPE.NEGOTIATION and UICOLOURS.NEGOTIATION or UICOLOURS.FIGHT
    self.backup = self:AddChild( Widget() )
end

function VSTeamDisplay:Refresh( members, boss_battle )

    -- Remove all the old portraits
    if self.leader then self.leader:Remove() end
    self.backup:DestroyAllChildren()

    -- Add new ones
    if self.card_type == QCARD_TYPE.BATTLE and members[1].boss and members[1].battle_preview_anim then

        -- This is a boss with a fancy intro. Show that instead of a regular team-member tile
        self.leader = self:AddChild( Widget.VSTeamBossPortrait(members[1], 900, 300, self.bg_colour, self.left ) )
        self.boss_battle = true

        -- Add the boss' team members!
        for k, agent in ipairs( members ) do
            if not agent.boss then
                self.backup:AddChild( Widget.VSTeamMemberPortrait( agent, 140, 140, self.bg_colour, self.left ) )
            end
        end
        
        self.leader:SetAlliesWidget( self.backup )

        -- If this boss has a set audio to play when they animate in, save it
        local anim_data = members[1].battle_preview_anim_fn and members[1].battle_preview_anim_fn( members[1].build )
        local audio = anim_data and anim_data.audio or members[1].battle_preview_audio
        if audio then
            self.team_audio = audio
        else
            -- If they don't, use the default boss audio
            self.team_audio = "event:/ui/prebattle_overlay/prebattle_overlay_whooshin_boss_generic"
        end
    elseif boss_battle and ( members[1].battle_preview_anim or members[1].battle_preview_anim_fn ) then

        -- This is a boss battle. Show our playable character in proper hero pose
        self.leader = self:AddChild( Widget.VSTeamHeroPortrait(members[1], 900, 300, self.bg_colour, self.left ) )
        self.boss_battle = true

        -- Add our team members!
        for k, agent in ipairs( members ) do
            if k > 1 then
                self.backup:AddChild( Widget.VSTeamMemberPortrait( agent, 140, 140, self.bg_colour, self.left ) )
            end
        end
        
        self.leader:SetAlliesWidget( self.backup )
    else
        -- Add our leader
        self.leader = self:AddChild( Widget.VSTeamLeaderPortrait( members[1], 200, 300, self.bg_colour, self.left ) )
        -- Add our team members!
        for k, agent in ipairs( members ) do
            if k > 1 then
                self.backup:AddChild( Widget.VSTeamMemberPortrait( agent, 140, 140, self.bg_colour, self.left ) )
            end
        end
    end

    self:Layout()

    return self
end

function VSTeamDisplay:IsBossBattle()
    return self.boss_battle or false
end

function VSTeamDisplay:GetTeamAudio()
    return self.team_audio or nil
end

function VSTeamDisplay:Layout()

    if not self:IsBossBattle() then
        -- Layout all our team members neatly
        local max_rows = 2
        local spacing = 20
        for k, w in ipairs( self.backup.children ) do
            local new_column = (k % max_rows == 1) or max_rows == 1
            w:LayoutBounds( new_column and ( self.left and "before" or "after" ) or "left", new_column and "top" or "below" )
                :Offset( new_column and ( self.left and -spacing or spacing ) or 0, new_column and 0 or -spacing )
            if new_column then w:LayoutBounds( nil, "top", self.backup.children[1] ) end
        end

        -- Align them behind their leader
        if self.left then
            self.backup:LayoutBounds( "before", "top", self.leader ):Offset( -20, 0 )
        else
            self.backup:LayoutBounds( "after", "top", self.leader ):Offset( 20, 0 )
        end
    end

    return self
end

function VSTeamDisplay:SetResolve( resolve_value, max_value, text_colour, value_colour, tooltip_text )
    self.leader:SetResolve( resolve_value, max_value, text_colour, value_colour, tooltip_text )

    return self
end

function VSTeamDisplay:ShowCombatStrengths()

    self.leader:ShowCombatStrength()
    for k, v in ipairs( self.backup.children ) do
        v:ShowCombatStrength()
    end

    return self
end


function VSTeamDisplay:PrepareAnimation()
    self.leader:SetTintAlpha(0)
    for i, widget in ipairs( self.backup.children ) do
        widget:SetTintAlpha(0)
    end
end

function VSTeamDisplay:AnimateIn()

    -- Prepare for animation by getting current values
    local pacing = 0.1
    local leader_x, leader_y = self.leader:GetPos()

    -- Prepare backup display sequence
    local squad_anim_duration = pacing/2
    local delay_per_squad = pacing/2
    local squad_adder = Updater.Parallel()
    for i, widget in ipairs( self.backup.children ) do
        local w_x, w_y = widget:GetPos()
        squad_adder:Add( 
            Updater.Series{
                Updater.Wait( delay_per_squad*i ),
                Updater.Parallel{
                    Updater.Ease( function(v) widget:SetTintAlpha(v) end, 0, 1, squad_anim_duration, easing.outQuad),
                    Updater.Do( function() widget:AnimateIn() end ),
                }
            }
        )
    end

    -- Play out our beautiful animation!
    self.fadein_anim = self:RunUpdater(
        Updater.Series{
            -- Show the allmighty leader first
            Updater.Parallel{
                Updater.Ease( function(v) self.leader:SetTintAlpha(v) end, 0, 1, pacing, easing.inQuad),
                Updater.Ease2D( function(x,y) self.leader:SetPos(x,y) end, leader_x + ( self.left and 20 or -20), leader_y, leader_x, leader_y, pacing, easing.inQuad),
                Updater.Do( function() self.leader:AnimateIn() end ),
            },
            -- And then the silly minions
            squad_adder,
        }
    )

end

function VSTeamDisplay:AnimateOut()
    
    local pacing = 0.05

    -- Fade out everything!
    self.leader:AlphaTo( 0, pacing )
    for i, widget in ipairs( self.backup.children ) do
        widget:AlphaTo( 0, pacing )
    end

end

------------------------------------------------------------------------------------------
-- Shows a team member

local VSTeamMemberPortrait = class( "Widget.VSTeamMemberPortrait", Widget )

function VSTeamMemberPortrait:init( agent, width, height, colour, facing )
    VSTeamMemberPortrait._base.init( self )

    self.tile_width = width
    self.tile_height = height
    self.padding = 12
    self.bg_border = 50
    self.bg_border_scale = 0.5
    local nineslice_w, nineslice_h = self.tile_width + self.bg_border*2*self.bg_border_scale, self.tile_height + self.bg_border*2*self.bg_border_scale
    self.left = facing
    self.agent = agent
    self.anim_scale = 0.25
    self.category_colour = colour

    -- Character anim
    self.bg = self:AddChild( Widget.Panel( engine.asset.Texture("UI/teammember_bg.tex") ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( self.category_colour )
        -- :SetTintAlpha( 0.2 )
    self.mask = self:AddChild( Widget() )
        :SetStencilContext( true )
        :SetHiddenBoundingBox( true )
    self.mask_image = Widget.Panel( engine.asset.Texture("UI/teammember_mask.tex") )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
    self.anim_root = Widget()
    self.anim = self.anim_root:AddChild( Widget.AnimatedCharacter() )
        :Refresh( self.agent )
        :SetSpeed( 1 )
        :SetAutoLoopEnabled( true )
        :SetOutline( 5, self.category_colour )
    self.anim.anim:SetPhase( math.random() )
    self.gradient = Widget.Image( engine.asset.Texture("UI/teammember_gradient.tex") )
        :SetSize( self.tile_width, self.tile_height )
        :SetTintAlpha( 0.8 )
    self.mask:AddChildren
    {
        self.mask_image:SetMask(true),
        self.anim_root:SetStencilTest(true),
        self.gradient,
    }
    self.overlay = self:AddChild( Widget.Panel( engine.asset.Texture("UI/teammember_overlay.tex") ) )
        :SetNineSliceCoords( self.bg_border, self.bg_border, self.bg_border, self.bg_border )
        :SetNineSliceBorderScale( self.bg_border_scale )
        :SetSize( nineslice_w, nineslice_h )
        :SetTintColour( self.category_colour )
        :Bloom( 0.3 )

    -- Text details
    self.name = self:AddChild( Widget.Label( "title", 24 ) )
        :SetGlyphColour( self.category_colour )
        :SetAutoSize( self.tile_width - self.padding*2 )
        :SetWordWrap( true )
        :SetText( agent:GetName() or agent:GetShortTitle() )
        :SetSaturation( 0.7 )
        :OverrideLineHeight( 19 )
        :Bloom( 0.15 )
    if self.left then self.name:LeftAlign() else self.name:RightAlign() end
    self.def = self:AddChild( Widget.Label( "title", 20 ) )
        :SetGlyphColour( self.category_colour )
        :SetAutoSize( self.tile_width - self.padding*2 )
        :SetWordWrap( true )
        :Bloom( 0.1 )

    self.def:SetText( agent:GetIdentity() )
    if self.left then self.def:LeftAlign() else self.def:RightAlign() end

    -- Agent's relationship icon
    self.relationship_icon = self:AddChild( Widget.OpinionIndicator( 22 ) )
        :ShowBg( false )
        :UpdateOpinion( self.agent )
        :Bloom( 0.1 )


    -- Layout!
    self:DoTaskInTime( "layout_team_member", 0.01, function() self:Layout() end )
end


function VSTeamMemberPortrait:SetResolve( resolve_value, max_value, text_colour, value_colour, tooltip_text )
    -- If this is a negotiation, we'll show the team's resolve under the leader
    if not self.resolve_label then
        self.resolve_value = self:AddChild( Widget.Label("title", 44 ) )
        self.resolve_label = self:AddChild( Widget.Label("body", 20 ) )
    end
    self.resolve_label:SetGlyphColour( text_colour )
        :SetText( LOC"UI.NEGOTIATION_PREVIEW.RESOLVE_DESC" )
        :Bloom( 0.05 )
    self.resolve_value:SetGlyphColour( value_colour or text_colour )
        -- :SetText( loc.format( LOC"UI.NEGOTIATION_PREVIEW.RESOLVE", resolve_value, max_value ) )
        :SetText( loc.format( "{1}", resolve_value ) )
        :Bloom( 0.15 )
    if resolve_value == 0 then
        self.resolve_value:SetGlyphColour( UICOLOURS.PENALTY )
    end
    if tooltip_text then
        self.resolve_label:SetToolTip( tooltip_text )
        self.resolve_value:SetToolTip( tooltip_text )
    end
end

function VSTeamMemberPortrait:ShowCombatStrength()
    if self.agent:IsPlayer() then
        if self.difficulty_stars then
            self.difficulty_stars:Hide()
        end
        return
    end

    if self.agent.boss and not self.agent:IsInPlayerParty() then
        self.difficulty_boss = self.difficulty_boss or self:AddChild( Widget.Label("title", 20, LOC"UI.NEGOTIATION_PREVIEW.HIGH_THREAT" ) )
            :Bloom( 0.05 )
    else
        if self.difficulty_boss then self.difficulty_boss:Hide() end

        if not self.difficulty_stars then
            self.difficulty_stars = self:AddChild( Widget() )
            for k=1, 5 do
                local star = self.difficulty_stars:AddChild( Widget.Image( engine.asset.Texture( "UI/teammember_star_full.tex" ) ) )
                    :SetSize( 13, 13 )
                    :SetTintColour( self.category_colour )
                    :Bloom( 0.1 )
                    :LayoutBounds( "after", "center" )
                    :Offset( 1, 0 )
                if ( self.agent:GetCombatStrength() + 0.5 ) == k then
                    star:SetTexture( engine.asset.Texture( "UI/teammember_star_half.tex" ) )
                elseif self.agent:GetCombatStrength() >= k then
                    star:SetTexture( engine.asset.Texture( "UI/teammember_star_full.tex" ) )
                else
                    star:SetTexture( engine.asset.Texture( "UI/teammember_star_empty.tex" ) )
                end
            end
        end
    end

    self:Layout()
end

function VSTeamMemberPortrait:Layout()

    self.def:LayoutBounds( self.left and "left" or "right", "bottom", self.bg )
        :Offset( self.left and self.padding or -self.padding, self.padding-3 )

    -- Position agent's name over their def
    self.name:LayoutBounds( self.left and "left" or "right", "above", self.def )
        :Offset( 0, -2 )
    -- Check if the name is too high up, meaning the text is too long. If so, hide the def
    local x, y = self.name:GetPos()
    if y > 40 then
        self.name:LayoutBounds( self.left and "left" or "right", "bottom", self.def )
        self.def:Hide()
    else
        self.def:Show()
    end

    -- Position agent's opinion of the player
    self.relationship_icon:LayoutBounds( self.left and "left" or "right", "top", self.bg )
        :Offset( self.left and self.padding-3 or -self.padding+3, -self.padding + 3 )

    -- Position the resolve value
    if self.resolve_value then
        self.resolve_value:LayoutBounds( self.left and "right" or "left", "top", self.bg )
            :Offset( self.left and -self.padding or self.padding, -self.padding+12 )
        self.resolve_label:LayoutBounds( self.left and "right" or "left", "below", self.resolve_value )
            :Offset( 0, 6 )
    end

    -- Position the difficulty stars
    if self.difficulty_stars then
        self.difficulty_stars:LayoutBounds( self.left and "right" or "left", "top", self.bg )
            :Offset( self.left and -self.padding or self.padding, -self.padding + 2 )
    end

    -- Position the boss threat warning
    if self.difficulty_boss then
        self.difficulty_boss:LayoutBounds( self.left and "right" or "left", "top", self.bg )
            :Offset( self.left and -self.padding or self.padding, -self.padding + 5 )
    end

    -- Position anim
    self.anim:SetScale( self.anim_scale , self.anim_scale )
        :LayoutBounds( self.left and "right" or "left", "top", self.bg )
        :SetScale( self.left and self.anim_scale or -self.anim_scale, self.anim_scale )
        :Offset( self.left and -15 or 15, -15 )

    -- Try and position the anim based on its head position
    self:DoTaskInTime( "delay_anim_layout", 0.02, 
        function()
            local head_offset_x, head_offset_y = 100, 0
            local anim_w, anim_h = self.anim_root:GetSize()
            if anim_w > 600 then head_offset_x = head_offset_x + 70 end
            if anim_h > 300 then head_offset_y = head_offset_y + 20 end
            if anim_h > 680 then head_offset_y = head_offset_y + 40 end
            local fight_data = self.agent.fight_data
            local head_x, head_y, head_z = self.anim.anim.model:SymbolToWorld( "head_swap", 0, 0, 0 )
            if fight_data.track_symbol then
                head_x, head_y, head_z = self.anim.anim.model:SymbolToWorld(fight_data.track_symbol, 0, 0, 0)
                head_offset_y = head_offset_y + 60
            end
            head_x, head_y, head_z = self.anim_root.parent:TransformFromWorld( head_x, head_y, head_z )
            self.anim_root:Show()
                :SetPos( ( self.left and -120 or -80 ) - head_x + head_offset_x, 0 - head_y + head_offset_y )
        end )

    return self
end

function VSTeamMemberPortrait:AnimateIn()
end

function VSTeamMemberPortrait:OnAdded()
    -- Refresh control mode display
    self:OnControlModeChange( TheGame:FE():GetControlMode(), TheGame:FE():GetControlDeviceID() )
    self:OnScreenModeChange( TheGame:FE():GetScreenMode() )
end

function VSTeamMemberPortrait:OnControlModeChange( cm, device_id )
    VSTeamMemberPortrait._base.OnControlModeChange( self, cm, device_id )
end

function VSTeamMemberPortrait:OnScreenModeChange( sm )
    VSTeamMemberPortrait._base.OnScreenModeChange( self, sm )

    -- Scale the text
    local scale_factor = 0.75
    self.name:SetFontSize( FONT_SIZE.ITEM_DESC * LAYOUT_SCALE[sm] * scale_factor )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    self.def:SetFontSize( FONT_SIZE.SMALL_TEXT * LAYOUT_SCALE[sm] * scale_factor )
        :OverrideLineHeight( FONT_SIZE.SMALL_TEXT * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    if self.resolve_value then
        self.resolve_value:SetFontSize( FONT_SIZE.ITEM_NAME * LAYOUT_SCALE[sm] * scale_factor )
            :OverrideLineHeight( FONT_SIZE.ITEM_NAME * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    end
    if self.resolve_label then
        self.resolve_label:SetFontSize( FONT_SIZE.SMALL_TEXT * LAYOUT_SCALE[sm] * scale_factor )
            :OverrideLineHeight( FONT_SIZE.SMALL_TEXT * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    end

    self:Layout()
end

------------------------------------------------------------------------------------------
-- Shows a team leader

local VSTeamLeaderPortrait = class( "Widget.VSTeamLeaderPortrait", Widget.VSTeamMemberPortrait )

function VSTeamLeaderPortrait:init( agent, width, height, colour, facing )
    VSTeamLeaderPortrait._base.init( self, agent, width, height, colour, facing )
    self.anim_scale = 0.45
    self.name:SetFontSize( 28 )
end

function VSTeamLeaderPortrait:ShowCombatStrength()
    self._base.ShowCombatStrength( self )

    if self.difficulty_stars then
        for k, star in ipairs( self.difficulty_stars.children ) do
            star:SetSize( 19, 19 )
                :LayoutBounds( "after", nil )
                :Offset( 1, 0 ) 
        end
    end

    if self.difficulty_boss then
        self.difficulty_boss:SetFontSize( 25 )
    end

end

function VSTeamLeaderPortrait:OnScreenModeChange( sm )
    VSTeamLeaderPortrait._base.OnScreenModeChange( self, sm )

    -- Scale the text
    local scale_factor = 0.85
    self.name:SetFontSize( FONT_SIZE.ITEM_DESC * LAYOUT_SCALE[sm] * scale_factor )
        :OverrideLineHeight( FONT_SIZE.ITEM_DESC * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    self.def:SetFontSize( FONT_SIZE.SMALL_TEXT * LAYOUT_SCALE[sm] * scale_factor )
        :OverrideLineHeight( FONT_SIZE.SMALL_TEXT * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    if self.resolve_value then
        self.resolve_value:SetFontSize( FONT_SIZE.ITEM_NAME * LAYOUT_SCALE[sm] * scale_factor )
            :OverrideLineHeight( FONT_SIZE.ITEM_NAME * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    end
    if self.resolve_label then
        self.resolve_label:SetFontSize( FONT_SIZE.SMALL_TEXT * LAYOUT_SCALE[sm] * scale_factor )
            :OverrideLineHeight( FONT_SIZE.SMALL_TEXT * LINE_HEIGHT_TEXT_MULT * LAYOUT_SCALE[sm] * scale_factor )
    end

    self:Layout()
end

------------------------------------------------------------------------------------------
-- Shows a boss

local VSTeamBossPortrait = class( "Widget.VSTeamBossPortrait", Widget.VSTeamMemberPortrait )

local TOP_PADDING = 500
local BOTTOM_PADDING = 20
local LEFT_PADDING = 600

function VSTeamBossPortrait:init( agent, width, height, colour, facing )
    VSTeamBossPortrait._base.init( self, agent, width, height, colour, facing )

    self.anim:Hide()

    self.hitbox = self:AddChild( Widget.SolidBox( width, height + TOP_PADDING + BOTTOM_PADDING, 0xFF000030 ))
        :SetHiddenBoundingBox( true )
        :Hide()
    self.panel_glow_top = self.mask:AddChild( Widget.Image( engine.asset.Texture("images/glow.tex" ), 1800, 800 ) )
        :SetHiddenBoundingBox( true )
        :SetStencilTest( true )
        :SetTintColour( colour )
        :SetTintAlpha( 0.4 )
        :SetPos( self.left and -400 or 400, -100 )

    self.particles_background = self:AddChild( Widget.ParticleSystem() )
        :SetHiddenBoundingBox( true )
        :LoadParams( "fightpreview_boss_background" )
        :SetPos( width/2, -height/2 )
        :Bloom( 0.4 )

    self.anim_scale = 0.73
    self.boss_anim_container = self.mask:AddChild( Widget() )
        :SetHiddenBoundingBox( true )
        :SetStencilTest( true )
        :SetScale( self.anim_scale )
    self.boss_anim = self.boss_anim_container:AddChild( Widget.Anim() )
    self.boss_anim_fx = self.boss_anim_container:AddChild( Widget.Anim() )

    local name = agent.battle_preview_name_fn and agent.battle_preview_name_fn(agent) or self.name:GetText()
    local title = agent.battle_preview_title_fn and agent.battle_preview_title_fn(agent) or self.def:GetText()
    local anim_data = agent.battle_preview_anim_fn and agent.battle_preview_anim_fn( agent.build )
    local anim = anim_data and anim_data.anim or agent.battle_preview_anim
    local offset = anim_data and anim_data.offset or agent.battle_preview_offset
    local glow = anim_data and anim_data.glow or agent.battle_preview_glow
    local audio = anim_data and anim_data.audio or agent.battle_preview_audio

    if anim then
        self.boss_anim:AddAnims( GetAnim( anim ) )
        self.boss_anim:AddOverrideFile( anim )
        self.boss_anim_fx:AddAnims( GetAnim( anim ) )
        self.boss_anim_fx:AddOverrideFile( anim )
        if glow then
            self.boss_anim:SetGlowThresholdAndAmount( glow.threshold, glow.bloom )
            self.boss_anim:SetGlowColour( glow.colour )
        end
        self.boss_anim:SetOutline( 5, UICOLOURS.FIGHT )
        self.boss_anim:Play( "pre", false )
        self.boss_anim:Push( "idle", true )
        if glow and glow.fx_bloom then
            self.boss_anim_fx:Bloom( glow.fx_bloom )
        else
            self.boss_anim_fx:Bloom( glow and glow.bloom or 0.3 )
        end
        self.boss_anim_fx:Play( "pre_fx", false )
        self.boss_anim_fx:Push( "idle_fx", true )
    end

    if offset then
        self.anim_offset_x = offset.x or 0
        self.anim_offset_y = offset.y or 0
        self.anim_offset_scale = offset.scale or 1
    end

    self.particles_foreground = self:AddChild( Widget.ParticleSystem() )
        :SetHiddenBoundingBox( true )
        :LoadParams( "fightpreview_boss_foreground" )
        :SetPos( width/2, -height/2 )
        :Bloom( 0.8 )

    self.boss_overlay_scale = 0.45
    self.boss_overlay = self:AddChild( Widget.Image( engine.asset.Texture("large/teammember_boss_badge.tex") ) )
        :SetHiddenBoundingBox( true )
        :SetScale( self.boss_overlay_scale )
        :SetTintAlpha( 0 )
    self.boss_name = self:AddChild( Widget.VSTeamBossTitle( 100, colour, 0x021A18cc, 3, -2 ) )
        :SetText( name )
        :SetHiddenBoundingBox( true )
    self.boss_def = self:AddChild( Widget.VSTeamBossTitle( 30, colour, 0x021A18cc, 2, -2 ) )
        :SetText( title )
        :SetHiddenBoundingBox( true )

    self.hitbox_edge = self:AddChild( Widget.SolidBox( 4, 4, 0xFF0000ff )) -- Used to calculate the edge of the screen
        :SetAnchors( self.left and "left" or "right", "center" )
        :SetHiddenBoundingBox( true )
        :Hide()
end

function VSTeamBossPortrait:Layout()
    VSTeamBossPortrait._base.Layout( self )

    -- Hide unneeded elements
    self.bg:SetTintAlpha( 0 )
    self.name:Hide()
    self.def:Hide()
    self.relationship_icon:Hide()
    if self.difficulty_boss then self.difficulty_boss:Hide() end
    self.gradient:Hide()
    self.overlay:Hide()

    -- Get the screen size
    local screen_w, screen_h = TheGame:FE():GetScreenScissorDims()
    self.screen_w = math.ceil( screen_w / TheGame:FE():GetBaseWidgetScale() )
    self.screen_h = math.ceil( screen_h / TheGame:FE():GetBaseWidgetScale() )

    -- Get the position of the left edge
    local center_x, center_y = self:GetPos()
    self.hitbox_edge:SetPos( 0, 0 )
    local width, height = self:GetSize()
    local leftedge_x, left_edge_y, rightedge_x, y
    if self.left then
        rightedge_x, y = center_x+width/2, center_y
        leftedge_x, left_edge_y = self.hitbox_edge:GetPos()
        leftedge_x, left_edge_y = self:TransformFromWidget( self.hitbox_edge, leftedge_x, left_edge_y )
    else
        leftedge_x, left_edge_y = center_x-width/2, center_y
        rightedge_x, y = self.hitbox_edge:GetPos()
        rightedge_x, y = self:TransformFromWidget( self.hitbox_edge, rightedge_x, y )
    end
    local mask_w, mask_h = rightedge_x - leftedge_x, self.tile_height + TOP_PADDING + BOTTOM_PADDING
    self.hitbox:SetSize( mask_w, mask_h )
        :LayoutBounds( self.left and "right" or "left", "bottom", self.bg )
        :Offset( 0, -BOTTOM_PADDING )
    local nineslice_w, nineslice_h = mask_w + LEFT_PADDING + self.bg_border*2*self.bg_border_scale, mask_h + self.bg_border*2*self.bg_border_scale
    self.mask_image:SetSize( nineslice_w, nineslice_h )
        :LayoutBounds( self.left and "right" or "left", "bottom", self.bg )
        :Offset( self.left and LEFT_PADDING or -LEFT_PADDING, -BOTTOM_PADDING )
    self.boss_overlay:LayoutBounds( self.left and "left" or "right", "bottom", self.mask_image ):Offset( self.left and -100 or 100, 0 )
    self.boss_name:LayoutBounds( self.left and "left" or "right", "bottom", self.mask_image ):Offset( self.left and 80 or -80, 30 )
    self.boss_def:LayoutBounds( self.left and "left" or "right", "above", self.boss_name ):Offset( self.left and -15 or 15, -24 )

    return self
end

function VSTeamBossPortrait:SetAlliesWidget( allies_container )
    self.allies_container = allies_container
    if self.allies_container then
        self.allies_container:SendToFront()
            :StackChildrenRow( 6 )
            :SetHiddenBoundingBox( true )
            :SetTintAlpha( 0 )
    end
    return self
end

function VSTeamBossPortrait:AnimateIn()
    VSTeamBossPortrait._base.AnimateIn( self )

    self.boss_anim_container:LayoutBounds( self.left and "right" or "left", "center", self.bg )
        :SetScale( self.anim_scale*self.anim_offset_scale )
        :Offset( self.anim_offset_x or 0, self.anim_offset_y or 0 )
    local boss_x, boss_y = self.boss_anim_container:GetPos()
    self.boss_anim_container:SetPos( self.left and boss_x-1800 or boss_x+1800, boss_y )

    self:DoTaskInTime( "layout_boss_anim", 0.05, function()

        -- Animate boss in
        self.boss_anim_container:MoveTo( boss_x, boss_y, 0.7, easing.outQuart )

        self.boss_name:SetTintAlpha( 0 )
        self.boss_def:SetTintAlpha( 0 )

        local has_allies = self.allies_container and not self.allies_container:IsEmpty()

        -- Animate overlay in
        self.boss_overlay_scale = has_allies and 0.7 or 0.45
        self.boss_overlay:SetScale( self.boss_overlay_scale, self.boss_overlay_scale )
            :LayoutBounds( self.left and "left" or "right", "bottom", self.mask_image )
            :Offset( self.left and -100 or 100, 0 )
            :SetScale( self.left and -self.boss_overlay_scale or self.boss_overlay_scale, self.boss_overlay_scale )
            :DoTaskInTime( "boss_overlay_fadein", 0.3, function()
                self.boss_overlay:AlphaTo( 0.4, 0.8, easing.outQuart, function()
                    self.boss_overlay:Pulse( 0.4, 0.8, 0.003 )
                end )
                local boss_overlay_x, boss_overlay_y = self.boss_overlay:GetPos()
                self.boss_overlay:SetPos( boss_overlay_x+(self.left and -400 or 400), boss_overlay_y )
                self.boss_overlay:MoveTo( boss_overlay_x, boss_overlay_y, 0.6, easing.outQuart )
            end )

        self.boss_name:DoTaskInTime( "boss_titles_fadein", 0.4, function()
                -- Animate titles in
                if has_allies then
                    self.allies_container:LayoutBounds( self.left and "left" or "right", "bottom", self.mask_image )
                        :Offset( self.left and 120 or -120, 40 )
                    self.boss_name:LayoutBounds( nil, "above", self.allies_container )
                        :Offset( 0, 20 )
                    self.boss_def:LayoutBounds( nil, "above", self.boss_name )
                        :Offset( 0, -24 )
                end
                local boss_name_x, boss_name_y = self.boss_name:GetPos()
                local boss_def_x, boss_def_y = self.boss_def:GetPos()
                self.boss_name:AlphaTo( 1, 0.6, easing.outQuart )
                self.boss_name:SetPos( boss_name_x+(self.left and -600 or 600), boss_name_y )
                self.boss_name:MoveTo( boss_name_x, boss_name_y, 0.6, easing.outQuart )
                self.boss_def:AlphaTo( 1, 0.6, easing.outQuart )
                self.boss_def:SetPos( boss_def_x+(self.left and -900 or 900), boss_def_y )
                self.boss_def:MoveTo( boss_def_x, boss_def_y, 0.6, easing.outQuart, function()
                    -- Align the allies, if any
                    if has_allies then self.allies_container:AlphaTo( 1, 0.1, easing.outQuart ) end
                end )
            end )
    end )
end

function VSTeamBossPortrait:OnScreenModeChange( sm )
    VSTeamBossPortrait._base.OnScreenModeChange( self, sm )

    -- Scale the text
    self.boss_def:SetLayoutScale( LAYOUT_SCALE[sm] )
    if self.allies_container and not self.allies_container:IsEmpty() then
        self.allies_container:LayoutBounds( self.left and "left" or "right", "bottom", self.mask_image )
            :Offset( self.left and 120 or -120, 40 )
        self.boss_name:LayoutBounds( nil, "above", self.allies_container )
            :Offset( 0, 20 )
        self.boss_def:LayoutBounds( nil, "above", self.boss_name )
            :Offset( 0, -24 )
    else
        self.boss_name:LayoutBounds( "right", "bottom", self.mask_image ):Offset( -80, 30 )
        self.boss_def:LayoutBounds( "right", "above", self.boss_name ):Offset( 15, -24 )
    end

    self:Layout()
end

------------------------------------------------------------------------------------------
-- Shows a hero

local VSTeamHeroPortrait = class( "Widget.VSTeamHeroPortrait", Widget.VSTeamBossPortrait )

function VSTeamHeroPortrait:init( agent, width, height, colour, facing )
    VSTeamHeroPortrait._base.init( self, agent, width, height, colour, facing )

end

------------------------------------------------------------------------------------------
-- Shows a boss title
local VSTeamBossTitle = class( "Widget.VSTeamBossTitle", Widget )

function VSTeamBossTitle:init( font_size, text_colour, shadow_colour, offset_x, offset_y )
    VSTeamBossTitle._base.init( self )

    self.font_size = font_size or 60
    self.text_colour = text_colour or UICOLOURS.SUBTITLE
    self.shadow_colour = shadow_colour or 0x021A18FF
    self.offset_x = offset_x or 4
    self.offset_y = offset_y or -3

    self.shadow = self:AddChild( Widget.Label( "title", self.font_size ) )
        :SetGlyphColour( self.shadow_colour )
    self.title = self:AddChild( Widget.Label( "title", self.font_size ) )
        :SetGlyphColour( self.text_colour )
        :Bloom( 0.1 )
end

function VSTeamBossTitle:SetText( text )
    self.title:SetText( loc.format( "<i>{1}</i>", string.upper (text) ) )
    self.shadow:SetText( loc.format( "<i>{1}</i>", string.upper (text) ) )
        :LayoutBounds( "center", "center", self.title )
        :Offset( self.offset_x, self.offset_y )
    return self
end

function VSTeamBossTitle:SetGlyphColour( colour )
    self.title:SetGlyphColour( colour )
    return self
end