------------------------------------------------------------
-- User settings file: a glorified save-loadable dictionary.

local UserSettings = class( "UserSettings" )

function UserSettings:init( filename, storage, cb )
    assert( filename, "Missing settings filename" )
    assert( storage, "Missing storage instance" )
    self.filename = filename
    self.values = {}
    self.dirty = false
    self.storage = storage
    self:LoadSettings( cb )
end

function UserSettings:GetFilename()
    return self.filename
end

function UserSettings:GetFatalError()
    return self.fatal_error
end

function UserSettings:LoadSettings( cb )
    local err
    local function OnLoad(success, filedata)
        if success and filedata then
            local load_ok, result = pcall(loadstring( filedata ))
            if load_ok and type(result) == "table" then
                self:ApplySettings( result )
            else
                -- Dump to log so we can see what's up.
                local lines = filedata:split("\n")
                print( "---Dump of", self.filename, " follows: --" )
                for i, v in ipairs(lines) do
                    print( v )
                end

                local str = string.format( "\tLoaded invalid UserSettings '%s': load_ok=%s, result=%s", self.filename, tostring(load_ok), tostring(result))
                str = debug.traceback( str, 2 )
                print( str )
                engine.inst:ShowWarning( str )
            end
        else
            err = "\tFailed to load UserSettings:"..self.filename
        end
        if cb then
            cb( self, success )
        end
    end

    local exists = false
    local function OnFileExists( result )
        exists = result
    end
    self.storage:FileExists( self.filename, OnFileExists )

    if exists then
        print( "Loading settings:", self.filename )
        local max_attempts = 5
        for attempts = 1, max_attempts do
            self.storage:LoadFile( self.filename, OnLoad )
            self.dirty  = false
            if err then
                print( "\tErr:", err, "Attempt:", attempts )
                if attempts >= max_attempts then
                    self.fatal_error = err
                else
                    engine.inst:Sleep( 250 )
                end
            else
                err = nil
                break
            end
        end
    else
        print( "Doesn't exist:", self.filename )
    end
end

function UserSettings:ForceSaveSettings( cb )
    self.dirty = true
    self:SaveSettings( cb )
end

function UserSettings:SaveSettings( cb )
    if not self.dirty or self.fatal_error then
        return
    end

    -- Don't ever fail save due to instruction count limit.
     debug.sethook( nil )

    print( "Saving settings:", self.filename )
    if cb == nil then
        cb = function( result )
            print( "\tResult:", result )
            if not result then
                self.errors = (self.errors or 0) + 1
                print( "WARNING: Failed to save profile!", self.errors )
            else
                self.errors = nil
            end
        end
    end

    local out = "return ".. serpent.block(self.values, {comment = false})

    -- NEVER save invalid lua.
    assert(type(loadstring( out )) == "function", "Invalid settings:\n"..out )

    self.storage:SaveFile( self.filename, out, cb)
    self.dirty = false
end

function UserSettings:ApplySettings( t )
    for k, v in pairs( t ) do
        self:ApplySetting( k, v )
    end
end

function UserSettings:VerifySetting( key, default )
    if type( self.values[ key ] ) ~= type(default) then
        print( string.format( "Verifying setting '%s': assigned to default '%s'", tostring(key), tostring(default) ))
        self:ApplySetting( key, default )
    end
end

function UserSettings:ApplySetting( key, value )
    if self.values[ key ] ~= value then
        self.values[ key ] = value
        self.dirty = true
    end
end

function UserSettings:GetSetting( key )
    return self.values[ key ]
end

function UserSettings:ClearSettings()
    self.values = {}
    self.dirty = true
end

function UserSettings:Dump()
    print( "UserSettings:", self.filename )
    print( serpent.block(self.values, {comment = false }))
end