
-- Adapted for Lua from Sedgewick and Wayn's 'Algorithms'
---------------------------------------------------------

-- Three points are a counter-clockwise turn if ccw > 0, clockwise if
-- ccw < 0, and collinear if ccw = 0 because ccw is a determinant that
-- gives twice the signed  area of the triangle formed by p1, p2 and p3.
local function ccw(p1, p2, p3)
    return (p2.x - p1.x)*(p3.y - p1.y) - (p2.y - p1.y)*(p3.x - p1.x)
end

-- points_array is an array of 2 dimensional points { x1, y1, x2, y2, ... }
-- Computation is in-place, thus the return value will be an array
-- of points { x1, y1, ... } that form the convex hull.
local function ComputeConvexHull( points_array )
    -- Would be great to run this algorithm without converting to {x,y} points!
    local pts = {}
    for i = 1, #points_array, 2 do
        table.insert( pts, { x = points_array[i], y = points_array[i+1] })
    end

    local N = #pts

    -- swap points[1] with the point with the lowest y-coordinate
    local min_y, min_i = math.huge
    for i, pt in ipairs( pts ) do
        if pt.y < min_y then
            min_y, min_i = pt.y, i
        end
    end
    pts[1], pts[min_i] = pts[min_i], pts[1]

    -- sort points by polar angle with points[1]
    local pt1 = table.remove( pts, 1 )
    table.sort( pts, function(p, q)
        local pa = math.pi/2 - math.atan( p.x - pt1.x, p.y - pt1.y )
        local pb = math.pi/2 - math.atan( q.x - pt1.x, q.y - pt1.y )
        return pa < pb
    end )
    table.insert( pts, 1, pt1 )

    -- We want points[0] to be a sentinel point that will stop the loop.
    pts[0] = pts[N]

    -- M will denote the number of points on the convex hull.
    local M = 1
    for i = 2, N do
        -- Find next valid point on convex hull.
        while ccw( pts[M-1], pts[M], pts[i]) <= 0 do
            if M > 1 then
                M = M - 1
            -- All points are collinear
            elseif i == N then
                break
            else
                i = i + 1
            end
        end

        -- Update M and swap points[i] to the correct place.
        M = M + 1
        pts[M], pts[i] = pts[i], pts[M]
    end

    -- Would be great to really do this in-place!
    for i = 1, N do
        if i <= M then
            points_array[i * 2 - 1] = pts[i].x
            points_array[i * 2] = pts[i].y
        else
            table.remove( points_array )
            table.remove( points_array )
        end
    end
    assert( table.count( points_array ) == 2 * M )
    return points_array
end

return { ComputeConvexHull = ComputeConvexHull }
