local fun = {}

local mt = {}
mt.__index = mt
setmetatable(fun, mt)

--not correct for tables with mixed key types
local isarray = function(t) 
    return type(t) == "table" and ( not next(t) or t[1] )
end

local clonetable = function(t) 
    local ret = {}
    for k,v in pairs(t) do
        ret[k] = v
    end
    return ret
end

local value_in_array = function(t, val)
    for k,v in ipairs(t) do
        if v == val then
            return true
        end
    end
    return false
end


function mt.__call(class, t)
    local inst = {data = t}
    setmetatable(inst, mt)
    return inst
end

function mt:value()
    local output = self.data
    if self.operations then
        for k,v in ipairs(self.operations) do
            output, self.data = v(self.data)
        end
        self.operations = nil
    end
    return output
end

function mt:__tostring()
    if self.input then
        local ret = self:value()
        return tostr(ret)
    else
        return "fun:"
    end
end



local Wrap = function(wrap_fn)
    return function(t, ...)
        local args = {...}
        if getmetatable(t) == mt then
            t.operations = t.operations or {}
            table.insert(t.operations, function(input) 
                local val = wrap_fn(input, table.unpack(args)) 
                return val, val
            end)
            return t
        else
            return wrap_fn(t, ...)
        end
    end
end

local WrapTerminal = function(wrap_fn)
    return function(t, ...)
        local args = {...}
        if getmetatable(t) == mt then
            t.operations = t.operations or {}
            
            table.insert(t.operations, function(input) 
                local val = wrap_fn(input, table.unpack(args)) 
                return val, input
            end)

            return t:value()
        else
            return wrap_fn(t, ...)
        end
    end
end

mt.map = Wrap(function(input, fn)
    local ret = {}
    if isarray(input) then
        for k,v in ipairs(input) do
            table.insert(ret, fn(v))
        end
    else
        for k,v in pairs(input) do
            ret[k] = fn(v)
        end
    end

    return ret
end)

mt.chunk = Wrap(function( t, chunk_size ) 
    chunk_size = chunk_size or 1
    local num_chunks = math.ceil(#t / chunk_size)
    local ret= {}
    local idx = 1
    for k = 1, num_chunks do
        local chunk = {}
        table.insert(ret, chunk)

        for j = 1, chunk_size do
            if idx <= #t then
                table.insert(chunk, t[idx])
                idx = idx + 1
            else
                return ret
            end
        end
    end

    return ret
end)

mt.zip = Wrap(function(...)
    local args = {...}
    local ret = {}
    local len = 0
    for k,v in ipairs(args) do
        len = math.max(len, #v)
    end
    for k = 1, len do
        
        local zipped = {}
        table.insert(ret, zipped)
        for k2,v in ipairs(args) do
            zipped[k2] = v[k]
        end
    end

    return ret
end)

mt.zipObject = Wrap(function(keys, values)
    local ret = {}
    for k,key in ipairs(keys) do
        ret[key] = values[k]
    end    
    return ret
end)

mt.slice = Wrap(function( t, start_index, end_index ) 
    end_index = end_index or #t+1
    local ret= {}
    assert(start_index > 0, "bad start index")
    
    local idx = start_index
    for k = start_index, math.min( end_index-1, #t)  do
        table.insert(ret, t[k])
    end

    return ret
end)

mt.compact = Wrap(function( input, fn )
    local ret= {}
    
    for k,v in ipairs(input) do
        if (v) then
            table.insert(ret, v)
        end
    end
    return ret
end)

mt.randomPick = WrapTerminal(function( input )
    if #input > 0 then
        return input[math.random(#input)]
    end    
end)

mt.countBy = WrapTerminal(function( input, fn )
    local ret= {}
    for k,v in ipairs(input) do
        local val = fn(v)
        if val then
            ret[val] = 1 + (ret[val] or 0)
        end
    end
    return ret
end)

mt.count = WrapTerminal(function( input, fn )
    local ret = 0
    for k,v in pairs(input) do
        local val = fn == nil or fn(v)
        if val then
            ret = ret + 1
        end
    end
    return ret
end)


mt.groupBy = Wrap(function( input, fn )
    local ret= {}
    for k,v in ipairs(input) do
        local val = fn(v)
        if val then
            ret[val] = ret[val] or {}
            table.insert(ret[val], v)
        end
    end
    return ret
end)

mt.every = WrapTerminal(function( input, fn )
    for k,v in pairs(input) do
        if not fn(v) then
            return false
        end
    end
    return true
end)

mt.any = WrapTerminal(function( input, fn )
    for k,v in pairs(input) do
        if fn(v) then
            return true
        end
    end
    return false
end)


mt.filter = Wrap(function( input, fn )

    assert(type(input) == "table")
    local ret= {}
    
    if isarray(input) then
        for k,v in ipairs(input) do
            if not fn or fn(v, k) then
                table.insert(ret, v)
            end
        end
    else
        for k,v in pairs(input) do
            if not fn or fn(v, k) then
                ret[k] = v
            end
        end
    end

    return ret
end)

mt.filterKeysByValues = Wrap(function(input, fn)
    local ret = {}
    for k,v in pairs(input) do
        if fn(v, k) then
            table.insert(ret, k)
        end
    end
    return ret

end)

mt.forEach = Wrap(function( input, fn )
    assert(type(input) == "table")

    if isarray(input) then
        for k,v in ipairs(input) do
            fn(v)
        end
    else
        for k,v in pairs(input) do
            fn(v)
        end
    end
    return input
end)


mt.first = WrapTerminal(function( input)
    return input[1]
end)
mt.head = mt.first

mt.last = WrapTerminal(function( input)
    return input[#input]
end)

mt.sortBy = Wrap(function(input, key)
    local ret = clonetable(input)
    table.stable_sort(ret, function(a,b) return a[key] < b[key] end)
    return ret
end)

mt.sortByFn = Wrap(function(input, fn)
    local ret = clonetable(input)
    table.stable_sort(ret, fn)
    return ret
end)

mt.shuffle = Wrap(function(input)
    local ret = clonetable(input)
    table.shuffle(ret)
    return ret
end)

mt.values = Wrap(function(input, fn)
    local ret = {}
    for k,v in pairs(input) do
        table.insert(ret, fn and fn(v) or v)
    end
    return ret
end)

mt.keys = Wrap(function(input)
    local ret = {}
    for k,v in pairs(input) do
        table.insert(ret, k)
    end
    return ret
end)

mt.reverse = Wrap(function(input, ...)
    local ret = {}
    for k = #input,1, -1 do
        table.insert(ret, input[k])
    end
    return ret
end)

mt.drop = function(input, num)
    num = num or 1
    return mt.slice(input, num+1)
end

mt.dropRight = Wrap(function( t, num) 
    num = num or 1
    local end_index = #t+1 - num
    local ret= {}
    for k = 1, end_index-1  do
        table.insert(ret, t[k])
    end
    return ret
end)


mt.difference = Wrap(function(input, other)
    local ret = {}
    for k, v in ipairs(input) do
        local include = not (other and value_in_array(other, v))
        if include then
            table.insert(ret, v)
        end
    end
    return ret
end)


mt.union = Wrap(function(...)
    local ret = {}
    local inputs = {...}
    for k,input in ipairs(inputs) do
        for k, v in ipairs(input) do
           if not value_in_array(ret, v) then
                table.insert(ret, v)
            end
        end
    end
    return ret
end)

mt.max = WrapTerminal(function(t, fn)
    local max_item, maxval = nil, nil
    for k, v in pairs(t) do
        local val = fn and fn(v) or v
        if max_item == nil or val > maxval then
            max_item, max = v, val
        end
    end
    return max_item
end)

mt.min = WrapTerminal(function(t, fn)
    local min_item, minval = nil, nil
    for k, v in pairs(t) do
        local val = fn and fn(v) or v
        if min_item == nil or val < minval then
            min_item, min = v, val
        end
    end
    return min_item
end)


mt.concat = Wrap(function(input, ...)
    local ret = clonetable(input)
    
    local args = {...}
    for k,v in ipairs(args) do
        if type(v) == "table" and #v > 0 then
            for kk, vv in ipairs(v) do
                table.insert(ret, vv)
            end
        else
            table.insert(ret, v)
        end
    end
    return ret
end)

mt.reduce = WrapTerminal(function(input, fn, accum)
    assert(type(input) == "table", "Trying to reduce a non-table")
    local ret
    local start_idx = accum and 1 or 2
    local accum = accum or input[1]
    for k = start_idx, #input do
        accum = fn(input[k], accum)
    end
    return accum
end)

function mt.add(augend, addend)
    return augend + addend
end

mt.contains = WrapTerminal(function(input, val)
    for k,v in ipairs(input) do
        if v == val then
            return true
        end
    end
    return false
end)


return fun