local astar = require "util/astar"
local Graph = class("Graph")

function Graph:init(nodes, edges, ox, oy)
    self.nodes = {}
    self.edges = {}

    local temp_nodes
    if nodes then
        temp_nodes = {}
        for k, v in ipairs(nodes) do
            temp_nodes[k] = self:AddNode((v.x or 0) + (ox or 0), (v.y or 0) + (oy or 0), v.userdata)
        end
    end

    if edges then 
        for k, v in ipairs(edges) do
            local node1 = v.node1 and temp_nodes[v.node1]
            local node2 = v.node2 and temp_nodes[v.node2]
            if node1 and node2 then
                self:AddEdge(node1, node2, v.userdata)
            end
        end
    end
end

function Graph:IsNode(elem)
    return table.arrayfind(self.nodes, elem) ~= nil
end

function Graph:IsEdge(elem)
    return table.arrayfind(self.edges, elem) ~= nil
end

function Graph:AddNode(x, y, userdata)
    local node = {x=x, y=y, userdata = userdata, edges={}}
    table.insert( self.nodes, node)
    return node
end


function Graph:RemoveEdge(edge)
    table.arrayremove(self.edges, edge)
    table.arrayremove(edge.node1.edges, edge)
    table.arrayremove(edge.node2.edges, edge)
end

function Graph:RemoveNode(node)
    table.arrayremove(self.nodes, node)
    local to_remove = {}
    for k,edge in pairs(self.edges) do
        if edge.node1 == node or edge.node2 == node then
            table.insert(to_remove, edge)
        end
    end
    for k,edge in pairs(to_remove) do
        self:RemoveEdge(edge)
    end
end

function Graph:AddEdge(node1, node2, userdata)
    local edge = {node1=node1, node2=node2, userdata = userdata}
    table.insert(node1.edges, edge)
    table.insert(node2.edges, edge)

    table.insert(self.edges, edge)
    return edge
end

function Graph:WeldNodes(node, other_node)
    if other_node and node and other_node ~= node then
        for k,edge in pairs(self.edges) do
            if edge.node1 == other_node and edge.node2 ~= node then
                self:AddEdge(edge.node2, node)
            end
            if edge.node2 == other_node and edge.node1 ~= node then
                self:AddEdge(edge.node1, node)
            end
        end
    end

    self:RemoveNode(other_node)
end

function Graph:BisectEdge(edge)
    if edge then
        local new_node = self:AddNode((edge.node1.x + edge.node2.x)/2, (edge.node1.y + edge.node2.y)/2)
        self:AddEdge(edge.node1, new_node)
        self:AddEdge(edge.node2, new_node)
        self:RemoveEdge(edge)
    end
end

function Graph:ExtendNode(node)
    if node then
        local new_node = self:AddNode(node.x+ 20, node.y+ 20, node.userdata and shallowcopy(node.userdata))
        self:AddEdge(node, new_node)
    end
end


function Graph:GetSaveTables()
    local nodes, edges = {}, {}
    local node_to_idx = {}
    for k, node in ipairs(self.nodes) do
        table.insert(nodes, {x=node.x, y=node.y, userdata = node.userdata})
        node_to_idx[node] = k
    end

    for k, edge in ipairs(self.edges) do
        table.insert(edges, {node1=node_to_idx[edge.node1], node2=node_to_idx[edge.node2], userdata = edge.userdata})
    end
    return nodes, edges
end


local function project_point_to_line(x1, y1, x2, y2, px, py)
    local dx, dy = x2 - x1, y2 - y1
    local dstsq = dx*dx+dy*dy
    if dstsq > 0 then
        local dst = math.sqrt(dstsq)
        local nx, ny = dx/dst, dy/dst

        local pdx, pdy = px - x1, py - y1
        local projection = nx*pdx + ny*pdy
        return projection, nx, ny, dst
    end
end

local function find_closest_point_on_line_segment(x1, y1, x2, y2, px, py)
    local projection, nx, ny, dst = project_point_to_line(x1, y1, x2, y2, px, py)
    
    if not projection then
        return x1, y1
    end

    if projection <= 0 then
        return x1, y1
    elseif projection >= dst then
        return x2, y2
    else
        return x1 + projection*nx, y1 + projection*ny, projection
    end
end

function Graph:FindClosestPoint(tx, ty)
    local cx, cy, closest_edge, closest_dist_sq, closest_projection
    
    for k,edge in pairs(self.edges) do
        
        local x, y, projection = find_closest_point_on_line_segment(edge.node1.x, edge.node1.y, edge.node2.x, edge.node2.y, tx, ty)
        local distsq = (x-tx)*(x-tx) + (y-ty)*(y-ty)
        if not closest_dist_sq or distsq < closest_dist_sq then
            cx, cy, closest_dist_sq, closest_edge, closest_projection = x, y, distsq, edge, projection
        end
    end

    return cx, cy, closest_edge, closest_dist_sq, closest_projection
end


function Graph:CanWeldNode(node)
    local WELD_DIST = 8
    for k,other_node in pairs(self.nodes) do
        if other_node ~= node and math.abs(other_node.x - node.x) < WELD_DIST and math.abs(other_node.y - node.y) < WELD_DIST then
            local connected = false
            for k,edge in pairs(self.edges) do
                if (edge.node1 == node and edge.node2 == other_node) or (edge.node2 == node and edge.node1 == other_node) then
                    connected = true
                    break
                end
            end
            if not connected then
                return other_node
            end
        end
    end
end


local function getnextneighbour_startnode(node, index)
    index = index or 1
    if node.neighbours[index] then
        return index+1, node.neighbours[index]
    end
end



local function getnextneighbour(node, index)
    index = index or 1
    local edge = node.edges and node.edges[index]
    if edge then
        return index+1, edge.node1 == node and edge.node2 or edge.node1
    end
end

local function neighbours (node)
    return getnextneighbour, node, nil
end


function Graph:GetClosestNode(x, y)
    local closest, closest_dist
    for k,node in pairs(self.nodes) do
        local distsq = (x-node.x)*(x-node.x) + (y-node.y)*(y-node.y)
        if not closest_dist or distsq < closest_dist then
            closest_dist = distsq
            closest = node
        end
    end
    return node
end

function Graph:FindPath(startx, starty, endx, endy)

    local cx_start, cy_start, closest_edge_start, closest_edge_dist_sq_start, closest_edge_t_start = self:FindClosestPoint(startx, starty)
    local start_node = {x = cx_start, y = cy_start, start_node = true, neighbours = { closest_edge_start.node1, closest_edge_start.node2} }

    local cx_end, cy_end, closest_edge_end, closest_edge_dist_sq_end, closest_edge_t_end = self:FindClosestPoint(endx, endy)
    local end_node = {x = cx_end, y = cy_end, end_node = true, neighbours = { closest_edge_end.node1, closest_edge_end.node2} }

    
    if closest_edge_end == closest_edge_start then
        --hey we're already there!
        return {{x=startx, y=starty}, start_node, end_node, {x=endx, y=endy}}
    end

    local function getnextneighbour_endnode(node, index)
        index = index or 1

        local maxn = node.edges and #node.edges+1 or 1
        
        if index == maxn then
            return index+1, end_node
        end

        local edge = node.edges and node.edges[index]
        if edge then
            return index+1, edge.node1 == node and edge.node2 or edge.node1
        end
    end

    local distfn = function(node1, node2) return distance( node1.x, node1.y, node2.x, node2.y) end
    local heurfn = function(node) return distance(node.x, node.y, endx, endy) end

    local nfn = function(node)

        if node.start_node then
            return getnextneighbour_startnode, node, nil
        end

        if node == closest_edge_end.node1 or node == closest_edge_end.node2 then
            return getnextneighbour_endnode, node, nil
        end

        return getnextneighbour, node, nil
    end

    local searcher = AStarSearcher(distfn, heurfn, nfn)

    searcher:StartSearch(start_node, end_node)
    searcher:RunToCompletion()
    local path = searcher:GetPath()
    table.insert(path, 1, {x=startx, y=starty})
    --table.insert(path, {x=endx, y=endy})
    return path
end

