local Iter = class("Iter")
Iter.can_reload = true

function Iter:init(fn)
    if type(fn) == "function" then
        self.iter_function = fn
    else
        assert(nil)
    end
end

function Iter.keys(t)
    local k
    return Iter(function() 
        k = next(t, k)    
        if k then
            return k
        end
    end)
end

function Iter.values(t)
    local k, v
    return Iter(function() 
        k, v = next(t, k)    
        if v then
            return v
        end
    end)
end

function Iter.array(t)
    local idx = 0
    return Iter(function()
        idx = idx + 1
        local val = t[idx]
        return val
    end)
end

function Iter.array_rev(t)
    local idx = 0
    local idx = #val
    return Iter(function()
        local val = t[idx]
        idx = idx - 1
        return val
    end)
end

function Iter:sort(fn)
    local t = {}
    for v in self.iter_function do
        table.insert(t, v)
    end

    table.sort(t, fn)
    return Iter.values(t)
end

--not exactly performant... would be better if we had reversable iterators
function Iter:reverse()
    local t = {}
    for v in self.iter_function do
        table.insert(t, v)
    end
    table.reverse(t)
    return Iter.values(t)
end

function Iter:shuffle()
    local t = {}
    for v in self.iter_function do
        table.insert(t, v)
    end
    table.shuffle(t)
    return Iter.values(t)
end

function Iter:random_pick()
    local t = {}
    for v in self.iter_function do
        table.insert(t, v)
    end
    return table.arraypick(t)
end

function Iter:sum(t)
    local ret = 0
    for v in self.iter_function do
        ret = ret + v
    end
    return ret
end

function Iter:product(t)
    local ret = 1
    for v in self.iter_function do
        ret = ret * v
    end
    return ret
end

function Iter:map(fn)
    return Iter(function() 
                    local val = self.iter_function()
                    return val and fn(val)
                end)
end

function Iter:inspect(fn)
    return Iter(function() 
                    local val = self.iter_function()
                    if val then
                        fn(val)
                        return val
                    end
                end)
end

function Iter:next()
    return self.iter_function()
end

function Iter:zip(other_iter)
    return Iter(function() 
                    local val = self:next()
                    if val then
                        local val2 = other_iter:next()
                        if val2 then
                            return {val, val2}
                        end
                    end
                end)
end


function Iter:flatten()
    local inner_iterator
    
    return Iter(function() 
        while true do
            if not inner_iterator then
                inner_iterator = self:next()
                if not inner_iterator then
                    return
                end
                inner_iterator = Iter(inner_iterator)
            end
            local val = inner_iterator:next()
            if val then
                return val
            else
                inner_iterator = nil
            end
        end
    end)
end

function Iter:chain(other_iter)
    local current_iter = self
    return Iter(function()
            local val = current_iter:next()
            if val then
                return val
            elseif current_iter == self then
                current_iter = other_iter
                return current_iter:next()
            end
        end)
end

function Iter:filter(fn)
    return Iter(function() 
        while true do
            local nextval = self.iter_function()
            if nextval == nil then
                return
            end
            if fn(nextval) then
                return nextval
            end
        end
    end)
end

function Iter:filtermap(fn)
    return Iter(function() 
        while true do
            local nextval = self.iter_function()
            if nextval == nil then
                return
            end

            local val = fn(nextval)
            if val then
                return val
            end

        end
    end)
end

function Iter:all(fn)
    for v in self.iter_function do
        if not fn(v) then
            return false
        end
    end
    return true
end

function Iter:any(fn)
    for v in self.iter_function do
        if fn(v) then
            return true
        end
    end
    return false
end

function Iter:collect()
    local ret = {}
    for v in self.iter_function do
        table.insert(ret, v)
    end

    return ret
end

function Iter:partition(fn)
    local ret, ret2 = {}, {}
    for v in self.iter_function do
        if fn(v) then
            table.insert(ret, v)
        else
            table.insert(ret2, v)
        end

    end
    return ret, ret2
end

function Iter:fold(init_val, fn)
    local val = init_val
    for v in self.iter_function do
        val = fn(val, v)
    end
    return val
end

function Iter:count_if(fn)
    return self:fold(0, function(val, item) return val + (fn(item) and 1 or 0) end)
end

function Iter:find(fn)
    for v in self.iter_function do
        if fn(v) then
            return v
        end
    end
end

function Iter:position(fn)
    local idx = 1
    for v in self.iter_function do
        if fn(v) then
            return idx
        end
        idx = idx + 1
    end
end

function Iter:count()
    local n = 0
    while self:next() do
        n = n + 1
    end
    return n
end

function Iter:max()
    local maxval 
    while true do
        local val = self:next()
        if val then
            maxval = maxval and math.max(val, maxval) or val
        else
            return maxval
        end
    end
end

function Iter:min()
    local minval 
    while true do
        local val = self:next()
        if val then
            minval = minval and math.min(val, minval) or val
        else
            return minval
        end
    end
end

function Iter:last()
    local last
    while true do
        local val = self:next()
        if val == nil then
            return last
        end
        last = val
    end
end

function Iter:nth(n)
    while true do
        local val = self:next()
        if val == nil then
            return
        end
        n = n - 1
        if n <= 0 then
            return val
        end
    end
end

function Iter:for_each(fn)
    for v in self.iter_function do
        fn(v)
    end
end