----------------------------------------------------------------
-- Localization functions.

local loc = require "util/locmacros"
local utf8_ex = require "lua-utf8"

local LOCALE_ENV =
{
	max_plurality = 2,
}

function loc.getEnv()
	return LOCALE_ENV
end

function loc.setMaxPlurality( n )
	assert( type(n) == "number" )
	LOCALE_ENV.max_plurality = n
end

-- Converts a plurality (integer) to an index.
-- TODO: this can be re-assigned by the locale, through the .pot file.
function loc.convertPlurality( n )
    assert_warning(type(n) == "number", tostring(n))
    if loc.convertPluralityOverride then
        local ok, plurality = pcall( loc.convertPluralityOverride, n )
        if ok and type(plurality) == "number" then
            return plurality
        elseif ok and type(plurality) == "boolean" then
            return plurality == true and 2 or 1
        else
            print( "loc.convertPlurality failed: ", ok, plurality )
        end
    end
    return (n ~= 1 and n ~= -1) and 2 or 1
end


local function parse_operator( str )
	if str and #str > 0 then
		local rest = str:match( "^%.gender:(.*)")
		if rest then
			return ".gender:", rest
		else
			return str:match("^([:*.%%#?])(.*)")
		end
	end
end

-- Localization formatter which should be used exclusively to compose all localized strings.  Each
-- format specifier (a parameter index enclosed in curly braces) can take multiple forms.
-- Note that the location of the specifiers within the format string is immaterial, since they
-- directly refer to the relevant parameter's index.
-- In the examples below, replace 'n' with the 'nth' parameter (not including the format string).
--
-- {n} : replaced with tostring(n)
-- loc.format( "I am a formatted {1}", "string" ) --> "I am a formatted string"

-- {n:word1|word2|...} : replaced with word<n>, so that n indexes into the pipe-delineated list.
-- loc.format( "I choose you, {1:Pikachu|Squirtle}!", 1 ) --> "I choose you, Pikachu!"
--
-- {n*plural1|plural2|...} : replaced with word<i>, where i == convertPlurality(n), so that n is
--      an integral number of something, which is converted into an index into the pipe-delineated list.
--      The plurality function is customizable per locale.
-- loc.format( "I ate {1*an|some} {1*apple|apples}", 1 ) --> "I ate an apple"
--
-- {ng male-word|female-word|neuter-word} : if the table argument's gender property is GENDER.MALE, replace with male-word; if it is GENDER.FEMALE, replace with the female-word; otherwise, replace with neuter-word.
--
-- {n.field} : replaced with n[ field ], so that n is a table of strings indexxed by the fieldname.
--      In the example below, DECKER = { name = "Decker", hisher = "his" }.
-- loc.format( "{1.name} hurt {1.hisher} hand", DECKER ) --> "Decker hurt his hand"
--
-- {n%format} : replaced with string.format( format, n )
-- loc.format( "Pi to 2 places: {1%.2f}", math.pi ) --> "Pi to 2 places: 3.14"
--
-- {n#macro_fn} : replaced with macro_fn( n )
-- loc.format( "Happy {1#time} to you!", gs:GetDateTime() ) --> "Happy 6:05 am to you!"
--
-- {n?true_phrase|false_phrase} : boolean selector
--


function loc.format(str, ...)
     if type(str) == "table" then
        str = table.concat( str )
    end

    local params = {...}

    local ret = string.gsub(str, "%b{}", 
        function(capture) 

            local inner = string.match(capture, "^{(.*{.*)}$")
            if inner then
                capture = string.format("{%s}", loc.format(inner, table.unpack(params) ))
            end


            local ret 
            local param_idx, payload = string.match(capture, "^{%s*([0-9]+)(.*)}$")

            if param_idx then
                param_idx = tonumber(param_idx)
            end
            
            local operator, operator_payload
            if payload then

                payload = loc.format(payload, table.unpack(params))
                operator_payload = string.match(payload, "^.gender%:(.*)")
                if operator_payload then
                    operator = ".gender:"
                else
                    operator, operator_payload = string.match(payload, "([?:#.*%%])(.*)")
                end
            end

            local param = params[param_idx]
            local tokens = {}
            if operator_payload then
                tokens = operator_payload:split("|")
            end
            if operator == "?" then
                -- boolean true|false
                ret = param and tokens[1] or tokens[2]
                ret = ret and tostring(ret) or ""
            elseif operator == "*" then
                -- param is a plurality, which converts to an index into tokens
                if type(param) == "table" then
                    -- Assume table length if param is a table; this could be invalid, but is useful for certain loc macros. eg agent_list.
                    param = #param
                end
                local pluralForm = loc.convertPlurality( param )
                ret = tostring( tokens[pluralForm] or tokens[#tokens] )
            elseif operator == ":" then
                -- param in an index into tokens
                ret = tostring( tokens[param] or tokens[#tokens] )
            elseif operator == '.' then
                -- param is a table, token is the field-name.
				ret = tostring( tokens[1] and type(param) == "table" and param[ tokens[1] ] )
            elseif operator == '.gender:' then
                local gender
                -- loc tables have only localized strings within them, making it impossible to detect which gender they are supposed to reference.
                -- we must use their metatables for comparison.
                local mt = getmetatable( param ) 
                if mt == GENDER_META[ GENDER.MALE ] then
                    gender = GENDER.MALE
                elseif mt == GENDER_META[ GENDER.FEMALE ] then
                    gender = GENDER.FEMALE
                elseif type(param) == "table" and IsEnum( param.gender, GENDER ) then
                    gender = param.gender
                end
                local idx, option
                if gender == GENDER.MALE then
                    idx = 1
                elseif gender == GENDER.FEMALE then
                    idx = 2
                else
                    idx = 3
                end
                ret = tostring( tokens[idx] or tokens[#tokens] )
            elseif operator == '%' then
                if tokens[1] == "d" and type(param) ~= "number" then
                    ret = tostring(param)
                else
                    ret = string.format( "%"..tokens[1], param )
                end
            elseif operator == '#' then
                ret = tostring( (loc[ tokens[1] ] and loc[ tokens[1] ]( param ) or tokens[1]) )
            else
                local str = tostring(param or capture)
                ret = str
                --local inner = string.match(str, "^{(.*{.*)}$")
                --ret = inner and loc.format(inner, table.unpack(params) ) or str
            end
            return ret
        end)
    return ret
end


-- Localization formatter which defers the replacement to the function 'f' which receives
-- 3 parameters: word, fn, field which correspond to the loc formatter fields {word.fn.field}
function loc.custom_format( format, f )
    assert( type(f) == "function" )

    local function repl( formatter )
        local word, rest = formatter:match( "^{([%(%) _%w]+)(.*)}$")
    	if word then
            local operator, rest = parse_operator( rest )
            rest = rest and loc.custom_format(rest, f)
            return f( word, operator, rest)
        end
    end

    local txt = format:gsub( "%b{}", repl )

    return txt
end

function loc.cap( str )
	return loc.toupper( str:sub( 1, 1 ) ) .. str:sub( 2 )
end

function loc.decap( str )
	return loc.tolower( str:sub( 1, 1 ) ) .. str:sub( 2 )
end

function loc.toupper( str )
	-- FIXME: non-unicode garbage
	return utf8_ex.upper( str )
end

function loc.tolower( str )
	-- FIXME: non-unicode garbage
	return utf8_ex.lower( str )
end

function utf8.sub(s,i,j)
    return utf8_ex.sub(s,i,j)
    -- i=utf8.offset(s,i) or 0
    -- if not j then
    --     j = utf8.len(s) or 0
    -- end
    -- local offset = utf8.offset(s,j+1)
    -- j = offset and offset - 1
    
    -- return j and string.sub(s,i,j)
end


return loc