require "util/iterators"
require "util/tablepool"
require "util/ringbuffer"
require "util/graph"

function generic_error( err )
    return debug.traceback( tostring(err), 2 )
end

function is_nan( n )
    return n ~= n
end

-- Takes in a value within an interval, and maps it to another interval
function remap(val, in_low, in_high, out_low, out_high)
   local t = math.min(1, math.max(0, (val - in_low) / (in_high - in_low)))
   return out_low + t*(out_high - out_low)
end

function shallowcompare(a,b)
   if type(a) ~= type(b) then return false end
   
   if type(a) == "table" then
        if table.count( a ) ~= table.count( b ) then
            return false
        end
        for k,v in pairs(a) do
            if v ~= b[k] then
                return false
            end
        end
        return true
    else
        return a == b
    end
end

function deepcompare(a,b)
   if type(a) ~= type(b) then return false end
   
   if type(a) == "table" then
        for k,v in pairs(a) do
            if not deepcompare(v, b[k]) then return false end 
        end

        for k,v in pairs(b) do
            if a[k] == nil then return false end
        end

        return true
   else
        return a == b
    end
end

--doesn't handle circular references
function deepcopy(t)
   if type(t) ~= "table" then return t end

   local t2 = {}
   for k,v in pairs(t) do
        if type(v) == "table" then
            t2[k] = deepcopy(v)
        else
            t2[k] = v
        end
    end
   return t2
end

-- creates an array of all keys in 't'
function copykeys( t, t2 )
    t2 = t2 or {}
    for k, v in pairs( t ) do
        table.insert( t2, k )
    end
    return t2
end

-- creates an array of all values in 't'
function copyvalues( t, t2 )
    t2 = t2 or {}
    for k, v in pairs( t ) do
        table.insert( t2, v )
    end
    return t2
end

-- "extends" a table.  Recursively copies all elements of tbase into an overrides table
-- Usage: t = extend( tbase ) { a = "override", c = { "override" } }
function table.extend( ... )
    local function tcopy( t1, t2 )
        assert( (type(t1) == "table" and type(t2) == "table") or error(string.format("type(t1) == %s, type(t2) == %s", type(t1), type(t2)) ))
        -- Deep-merge all of t1 into t2
        for k,v in pairs( t1 ) do
            if type(v) == "table" then
                if t2[k] == nil then
                    t2[k] = tcopy( v, {} )
                elseif not t2[k]._OVERRIDE then
                    tcopy( v, t2[k] )
                end
            else
                if t2[k] == nil then
                    t2[k] = v
                else
                    assert( type(t2[k]) == type(v), tostring(k) )
                end
            end
        end
            
        return t2
    end
    local tbases = { ... }
    
    return function( t )
        for _, tbase in ipairs( tbases ) do
            t = tcopy( tbase, t or {} )
        end
        return t
    end
end

function table.contains(table, element)
  for _, value in pairs(table) do
    if value == element then
      return true
    end
  end
  return false
end

--handles circular references but allocates a lot!
function deepcopy_circ(t)
    local refs = {}    
    
    local function deepcopy_int(t)
       if type(t) ~= "table" then return t end
       local t2 = {}
       refs[t] = t2

       for k,v in pairs(t) do
            if type(v) == "table" then
                if refs[v] then
                    t2[k] = refs[v]
                else
                    t2[k] = deepcopy_int(v)
                end
            else
                t2[k] = v
            end
        end
       return t2
    end

    return deepcopy_int(t)
end

function shallowcopy(orig, dest)
    local copy
    if type(orig) == 'table' then
        copy = dest or {}
        for k, v in pairs(orig) do
            copy[k] = v
        end

    elseif orig ~= nil then
        -- number, string, boolean, etc
        copy = orig

    else
        copy = dest
    end
    return copy
end

local getinfo = debug.getinfo
local max = math.max
local concat = table.concat

local function filtersource(src)
    if not src then return "[?]" end
    if src:sub(1, 1) == "@" then
        src = src:sub(2)
    end
    return src
end

local function formatinfo(info)
    if not info then return "**error**" end
    local source = filtersource(info.source)
    if info.currentline then
        source = source..":"..info.currentline
    end
    return ("\t%s in (%s) %s (%s) <%d-%d>"):format(source, info.namewhat, info.name or "?", info.what, info.linedefined, info.lastlinedefined)
end

function FormatMemorySize( bytes )
    if bytes > 1024*1024 then
        return string.format("%.2f MB", (bytes)/(1024*1024))
    elseif bytes > 1024 then
        return string.format("%.2f KB", (bytes)/(1024))
    else
        return string.format("%d B", bytes)
    end
end

function debugstackline(start)
    local info = getinfo(start+1)
    local source = filtersource(info.source)
    if info.currentline then
        source = source..":"..info.currentline
    end
    return source
end

function debugstack(start, top, bottom)
        if not bottom then bottom = 10 end
        if not top then top = 12 end
        start = (start or 1) + 1

        local count = max(2, start)
        while getinfo(count) do
            count = count + 1
        end

        count = count - start

        if top + bottom >= count then
            top = count
            bottom = nil
        end

        local s = {"stack traceback:"}
        for i = 1, top, 1 do
            s[#s + 1] = formatinfo(getinfo(start + i - 1))
        end
        if bottom then
            s[#s + 1] = "\t..."
            for i = bottom , 1, -1 do
                s[#s + 1] = formatinfo(getinfo(count - i + 1))
            end
        end

        return concat(s, "\n")
end

function table_diff(from, to)
    local diff = 
    {
        remove = {},
        merge = {},
        add = {},
    }

    for k,v in pairs(from) do
        local t = to[k]
        if t == nil then
            diff.remove[k] = true
        else
            if type(t) == "table" then
                local d = table_diff(v, t)
                if d then
                    diff.merge[k] = d
                end
            else
                if v ~= t then
                    diff.merge[k] = t
                end
            end
        end
    end

    for k,v in pairs(to) do
        local f = from[k]
        if f == nil then
            diff.add[k] = v
        end
    end

    if not next(diff.remove) then
        diff.remove = nil
    end
    if not next(diff.merge) then
        diff.merge = nil
    end
    if not next(diff.add) then
        diff.add = nil
    end

    if next(diff) then
        return diff
    end
end


function clamp(val, min, max)
    if min and val < min then
        val = min
    end

    if max and val > max then
        val = max
    end

    return val
end

function sign(num)
    --returns -1 if num is negative, 1 if num is positive.
    return (num < 0 and -1) or 1
end

function rerange(num, oldmin, oldmax, newmin, newmax)
    --re-ranges a number from one range into a new range. IE:
    --(num = 180, oldmin = 0, oldmax = 360, newmin = -180, newmax = 180) returns 0.
    --(num = PI, oldmin = 0, oldmax = 2PI, newmin = 0, newmax = 360) returns 180.
    return (newmax-newmin) * (num-oldmin)/(oldmax-oldmin) + newmin
end

function table.maxn( t )
    local n = nil
    for k, v in pairs(t) do
        if type(k) == "number" and (n == nil or k > n) then
            n = k
        end
    end
    return n
end

function table.shiftdown( t )
    local n = table.maxn( t )
    local head = t[ 1 ]
    for i = 2, n do
        t[ i - 1 ] = t[ i ]
    end
    t[ n ] = nil
    return head
end

function table.map( t, fn )
    local ret = {}
    for k,v in ipairs(t) do
        local new_val = fn(v)
        if new_val then
            table.insert(ret, new_val)
        end
    end
    return ret
end

function table.arrayfunction( t, fn )
    for k,v in ipairs(t) do
        fn(v)
    end
    return t
end

function table.transform( t, fn )
    local j, n = 1, #t
    for i = 1, #t do
        local v = fn( t[i] )
        if j < i then
            t[i] = nil
        end
        t[j] = v
        if v ~= nil then
            j = j + 1
        end
    end
end

function table.range( a, b )
    local t = {}
    for i = a, b do
        table.insert( t, i )
    end
    return t
end

function table.arrayremove(t, v)
    for k = 1, #t do
        if t[k] == v then
            table.remove(t, k)
            return true
        end
    end
    return false
end

function table.arrayremoveall( t, v )
    local count = 0
    while table.arrayremove(t, v) do
        count = count + 1
    end
    return count
end

function table.eraseback(t, v)
    for k = 1, #t do
        if t[k] == v then
            if k ~= #t then
                t[k], t[#t] = t[#t], t[k]
            end
            table.remove(t)
            return true
        end
    end
    return false
end

function table.arraycontains(t, v)
    for k = 1, #t do
        if t[k] == v then
            return true
        end
    end
    return false
end

function table.arrayfind(t, v)
    for i, tv in ipairs(t) do
        if tv == v then
            return i
        end
    end
    return nil
end

function table.findif( t, fn )
    for k, v in pairs(t) do
        if fn(v) then
            return k, v
        end
    end
end

function table.removevalue(t, value)
    for k, v in pairs(t) do
        if v == value then
            t[k] = nil
        end
    end
end

function table.find(t, v)
    for k, tv in pairs(t) do
        if tv == v then
            return k
        end
    end
    return nil
end

function table.clear(t)
    for k,v in pairs(t) do
        t[k] = nil
    end
end

function table.reverse(t)
    local len = #t
    local half = math.floor(len /2)
    for k = 1, half do
        t[k], t[len - (k-1)] =  t[len - (k-1)], t[k]
    end
end

function table.count( t, value )
    local count = 0
    for k,v in pairs(t) do
        if value == nil or v == value then
            count = count + 1
        end
    end
    return count
end


function table.count_if(t, fn, ...)
    local count = 0
    for k,v in pairs(t) do
        if fn(v, ...) then
            count = count + 1
        end
    end   
    return count 
end

function table.sum( t )
    local total = 0
    for i,v in pairs(t) do
        total = total + v
    end
    return total
end

function table.shuffle(array, start_index, end_index)
    
    start_index = start_index or 1
    end_index = end_index or #array

    local arrayCount = end_index - start_index + 1
    
    for i = end_index, start_index+1, -1 do
        local j = math.random(start_index, i)
        array[i], array[j] = array[j], array[i]
    end
    return array
end

function table.seeded_shuffle(array, rng, start_index, end_index)
    if not rng then
        local time = os.time()
        time = time - (time % (60*60*24))
        rng = engine.Random( time )
    end
    start_index = start_index or 1
    end_index = end_index or #array

    local arrayCount = end_index - start_index + 1
    
    for i = end_index, start_index+1, -1 do
        local j = rng:Random(start_index, i)
        array[i], array[j] = array[j], array[i]
    end
    return array
end

function table.seeded_pick(array, rng)
    if not rng then
        local time = os.time()
        time = time - (time % (60*60*24))
        rng = engine.Random( time )
    end
    if #array > 0 then
        return array[rng:Random(1, #array)]
    end
end

-- See: "Heap's Algorithm for generating permutations"
function table.permute( t, fn, size )
    if size == nil then
        size = #t
    end
    if size == 1 then
        fn( t )
    else
        for i = 1, size do
            table.permute( t, fn, size - 1 )

            if size % 2 == 1 then
                t[1], t[size] = t[size], t[1]
            else
                t[i], t[size] = t[size], t[i]
            end
        end
    end
end


function permutations(elements, out_list, n)
    out_list = out_list or {}
    n = n or #elements
    if n == 0 then
        table.insert(out_list, shallowcopy(elements))
    else
        for i = 1, n do
            elements[i], elements[n] = elements[n], elements[i]
            permutations(elements, out_list, n - 1)
            elements[i], elements[n] = elements[n], elements[i]
        end
    end
    return out_list
end


function Combinations(elements, length)
    local ret = {}
    for i = 1, #elements do
        if length == 1 then
            table.insert(ret, {elements[i]})
        else
            local subelements = {}
            for k = i+1, #elements do
                table.insert(subelements, elements[k])
            end
            local sub_list = Combinations(subelements, length-1)
            for s = 1, #sub_list do
                local next = sub_list[s]
                table.insert(next, 1, elements[i])
                table.insert(ret, next)
            end
        end
    end
    
    return ret
end



function table.add(dst, src)
    for k,v in pairs(src) do
        assert(dst[k] == nil, "Merging duplicate into table")
        dst[k] = v
    end
end

function table.arrayadd(dst, src)
    for i,v in ipairs(src) do
        table.insert(dst, v)
    end
end

function table.insert_unique( t, val )
    for k, v in pairs(t) do
        if v == val then
            return
        end
    end
    table.insert( t, val )
    return true
end

function table.merge( ... )
    local t = {}
    for i = 1, select( "#", ... ) do
        local t2 = select( i, ... )
        if t2 then
            table.arrayadd( t, t2 )
        end
    end
    return t
end

function table.inherit( tbase, t, recurse )
    if t == nil then
        return tbase
    end
    local tnew = shallowcopy( tbase )
    for k, v in pairs(t) do
        if recurse and type(tnew[k]) == "table" then
            tnew[k] = table.inherit( tnew[k], v, recurse )
        else
            tnew[k] = v
        end
    end
    return tnew
end

--jcheng: from http://lua.2524044.n2.nabble.com/A-stable-sort-td7648892.html
local max_chunk_size = 12 

local function insertion_sort( array, first, last, goes_before ) 
  for i = first + 1, last do 
    local k = first 
    local v = array[i] 
    for j = i, first + 1, -1 do 
      if goes_before( v, array[j-1] ) then 
        array[j] = array[j-1] 
      else 
        k = j 
        break 
      end 
    end 
    array[k] = v 
  end 
end 

local function merge( array, workspace, low, middle, high, goes_before ) 
  local i, j, k 
  i = 1 
  -- Copy first half of array to auxiliary array 
  for j = low, middle do 
    workspace[ i ] = array[ j ] 
    i = i + 1 
  end 
  i = 1 
  j = middle + 1 
  k = low 
  while true do 
    if (k >= j) or (j > high) then 
      break 
    end 
    if goes_before( array[ j ], workspace[ i ] )  then 
      array[ k ] = array[ j ] 
      j = j + 1 
    else 
      array[ k ] = workspace[ i ] 
      i = i + 1 
    end 
    k = k + 1 
  end 
  -- Copy back any remaining elements of first half 
  for k = k, j-1 do 
    array[ k ] = workspace[ i ] 
    i = i + 1 
  end 
end 

local function merge_sort( array, workspace, low, high, goes_before ) 
  if high - low < max_chunk_size then 
    insertion_sort( array, low, high, goes_before ) 
  else 
    local middle = math.floor((low + high)/2) 
    merge_sort( array, workspace, low, middle, goes_before ) 
    merge_sort( array, workspace, middle + 1, high, goes_before ) 
    merge( array, workspace, low, middle, high, goes_before ) 
  end 
end 

function table.stable_sort( array, goes_before ) 
  local n = #array 
  if n < 2 then  return array  end 
  goes_before = goes_before or 
    function (a, b)  return a < b  end 
  local workspace = {} 
  --  Allocate some room. 
  workspace[ math.floor( (n+1)/2 ) ] = array[1] 
  merge_sort( array, workspace, 1, n, goes_before ) 
  return array 
end 

-- verify if the parameter is "plain old data".
-- This means either a string, boolean, number, or a simple table with only pods (recursively) as key/values.
-- Plain old data cannot contain more than a single reference to any table.

local _pod_tables = {}
local function is_pod_internal( data )
    if data == nil or type(data) == "string" or type(data) == "boolean" or type(data) == "number" then
        return true

    elseif type(data) == "table" then
        if _pod_tables[ data ] then
            return false

        else
            _pod_tables[ data ] = true 
            for k, v in pairs( data ) do
                if not is_pod_internal(k) or not is_pod_internal(v) then
                    return false
                end
            end
        end
        return true
    end

    return false
end

function table.is_pod( data )
    local is_pod = is_pod_internal( data )
    table.clear( _pod_tables )
    return is_pod
end

table.empty = readonly({})

-- a delegate is an array with t[1] as a function, and t[2...n] as additional optional parameters.
-- Furthermore, the additional parameters ... are appended to the invocation.
-- For the sake of flexibility, delegate might directly be a function.
function call_delegate( delegate, ... )
    if type(delegate) == "function" then
        return delegate( ... )
    elseif type(delegate) == "table" then
        assert( type(delegate[1]) == "function" )
        local args = { table.unpack( delegate, 2 ) }
        local n = select( "#", ... )
        for i = 1, n do
            local p = select( i, ... )
            table.insert( args, p )
        end
        return delegate[1]( table.unpack( args ))
    end
end

function rawstring( t )
    if type(t) == "table" then
        local mt = getmetatable( t )
        if mt then
            -- Seriously, is there any better way to bypass the tostring metamethod?
            setmetatable( t, nil )
            local s = tostring( t )
            setmetatable( t, mt )
            return s
        end
    end

    return tostring(t)
end

local TREFS = {}
function tostr( t, maxRecurse, indent )
    table.clear( TREFS )

    if indent == nil then indent = 0 end
    maxRecurse = maxRecurse or 1

    if type(t) == "table" and maxRecurse ~= 0 then
        local s = tostring(t) .. "\n" .. string.rep(" ", indent) .. "{\n"
        indent = indent + 4
        for k,v in pairs(t) do
            if table.arrayfind  (TREFS, v) == nil then
                TREFS[#TREFS + 1] = v
                s = s .. string.rep(" ", indent) .. tostring(k) .." = " ..tostr(v, maxRecurse - 1, indent) ..",\n"
            else
                s = s .. string.rep(" ", indent) .. tostring(k) .." = " ..tostring(v) .. ",\n"
            end
        end
        indent = indent - 4
        s = s .. string.rep(" ", indent) .. "}"
        return s
    else
        return tostring(t)
    end
end

-- aka. "trim6" http://lua-users.org/wiki/StringTrim
function string.trim(s)
    return s and (s:match'^()%s*$' and '' or s:match'^%s*(.*%S)')
end

function string.crop(s, len)
    if #s > len then
        return s:sub( 1, len ) .. "..."
    else
        return s
    end
end

-- Lua 5.2 compat
if not math.pow then
    math.pow = function( x, y )
        return x^y
    end
end

-- Lua 5.2 compat
if not math.atan2 then
    math.atan2 = math.atan -- atan now serves both purposes
end


local pow = math.pow 
local floor = math.floor
local fmod = math.fmod
local sin = math.sin
local cos = math.cos
local atan2 = math.atan2
local abs = math.abs
local floor = math.floor

math.round = function(num, numDecimalPlaces)
    local mult = 10^(numDecimalPlaces or 0)
    return math.floor(math.floor(num * mult + 0.5) / mult)
end

-- constrain angle delta to [-PI, PI]
function normalizeDelta(angle)
    return atan2(sin(angle), cos(angle))
end

function reduceangle(angle)
    angle = (angle > 2*PI or angle < 0) and fmod(angle, 2*PI) or angle
    if angle < 0 then 
        angle = angle + 2*PI 
    end
    return angle
end

function angledist(src, dest)
    return ((reduceangle(dest) - reduceangle(src)) + PI) % (2*PI) - PI
end

function arclength(angle, radius)
    return angle * radius
end

function arcToAngle(arclength, radius)
    return arclength / radius
end

function rotate2d( dx, dy, theta )
    return dx * math.cos(theta) - dy * math.sin(theta), dx * math.sin(theta) + dy * math.cos(theta)
end

-- Pick a random point within a circle at (x0, y0) with radius.
function randomPoint( radius, x0, y0 )
    local phi = math.random() * 2 * math.pi
    local r = math.random() * radius
    return r * math.cos( phi ) + x0, r * math.sin( phi ) + y0
end

-- Rolls integral values, but permits skewing the distribution if lo/hi are floating point.
-- eg.
--  SkewRandom( 1, 2 ) will return values from [1..2] with a uniform PDF: 50% 1, 50% 2
--  SkewRandom( 1, 2.5 ) will return values from [1..2] but with 50% MORE 2s
function SkewRandom( lo, hi )
    local roll = math.random() * (hi - lo + 1) + lo
    if roll > 0 then
        roll = clamp( math.floor( roll ), math.floor( lo ), math.floor( hi ))
    else
        roll = clamp( math.ceil( roll - 1 ), math.ceil( lo ), math.ceil( hi ))
    end
    return roll
end

-- true if the short direction from angle1 to angle2 is clockwise, false otherwise
function clockwise(angle1, angle2)
    return angledist(angle1, angle2) < 0
end

function truncnum(f, dec)
    local m = pow(10, dec or 2)
    return floor(f * m) / m
end

function normalizeVec3(x,y,z)
    local mag = math.sqrt(x*x+y*y+z*z)
    if mag > 0 then
        return x/mag, y/mag, z/mag
    end
    return 0,0,0
end

function normalizeVec2(x,y)
    local mag = math.sqrt(x*x+y*y)
    if mag > 0 then
        return x/mag, y/mag
    end
    return 0,0
end

function normalizeRads(rads)
    local n = math.floor( rads / (math.pi * 2))
    return rads - math.pi * 2 * n
end

function distsq(x1, y1, x2, y2)
    local dx = x2-x1
    local dy = y2-y1
    return dx*dx + dy*dy
end

function distance(x1, y1, x2, y2)
    local dx = x2-x1
    local dy = y2-y1
    return math.sqrt( dx*dx + dy*dy )
end

function round_zero( n, step )
    step = step or 10
    if n > 0 then
        return math.floor( n / step ) * step
    elseif n < 0 then
        return math.ceil( n / step ) * step
    else
        return n
    end
end

function Dict(args)
    local t = {}
    for k,v in ipairs(args) do
        t[v] = v
    end
    return t
end


local gc_running = true
function StopGC()
    collectgarbage("stop")
    gc_running = false
end

function StartGC()
    collectgarbage("restart")
    gc_running = true
end

function DoFullGC()
    -- Because garbage collecting can release other references on the engine side,
    -- FULL garbage collection entails multiple cycles.
    local i, MAX_CYCLES = 0, 8
    while true do
        local before_bytes = collectgarbage( "count" )
        collectgarbage("collect")
        local after_bytes = collectgarbage( "count" )
        i = i + 1
        if before_bytes - after_bytes < 32 or i >= MAX_CYCLES then
            break
        end
    end
    if not gc_running then
        collectgarbage("stop")
    end
end

function StepGC()
    return collectgarbage("step", 0)
end

function ReportTextureGC()
    local tex_pre = engine.inst:GetTextureMemUsed()
    local pre = collectgarbage("count")
    collectgarbage()
    local pst = collectgarbage("count")
    local tex_pst = engine.inst:GetTextureMemUsed()
    local d_mem = (pst - pre) -- in K
    local d_tmem = (tex_pst - tex_pre)/(1024*1024) -- in M
    print (string.format("FREED UP %2.2fK mem, %2.2fM texture", d_mem, d_tmem))
end

--finds the first index where val would be inserted to preserve ordering in the table t
local default_cmp = function(a,b) return a < b end
function table.binsearch(t, val, cmp)
    cmp = cmp or default_cmp
    local len = #t
    local min_idx = 1
    local max_idx = len--math.floor(len/2)

    while min_idx <= max_idx do
        local idx = math.floor((min_idx + max_idx)*.5)
        
        if cmp(t[idx], val) then
            min_idx = idx + 1
        else
            max_idx = idx - 1
        end
    end
    return min_idx
end

-- binary search 't', returns whether 'val' was found.
function table.binfind(t, val, cmp)
    local idx = table.binsearch( t, val, cmp )
    return t[ idx ] == val
end

function table.binsert(t, val, cmp)
    local idx = table.binsearch(t, val, cmp)
    table.insert(t, idx, val)
end

function table.binsert_unique(t, val, cmp)
    local idx = table.binsearch(t, val, cmp)
    if t[idx] ~= val then
        table.insert(t, idx, val)
    end
end

function table.arraypick( t )
    local num = t and #t or 0
    return num > 0 and t[math.random(num)] or nil
end

function table.multipick( t, amt, duplicates )
    assert( amt )
    amt = math.min(amt, #t)

    local tmp_table = shallowcopy(t)
    local picks = {}

    for i = 1, amt do
        local num = tmp_table and #tmp_table or 0
        local pick = num > 0 and math.random(num) or nil
        table.insert(picks, tmp_table[pick])
        if pick and not duplicates then
            table.remove(tmp_table, pick)
        end
    end

    return picks
end

function table.tablepick( t )
    local num = table.count( t )
    if num > 0 then
        local i = math.random( num )
        for k, v in pairs(t) do
            i = i - 1
            if i <= 0 then
                return k, v
            end
        end
    end
end

function table.tablepicks ( t, amt )
    local num = table.count( t )
    local picks = {}
    if num > 0 and amt <= num then
        for i = 1, amt do
            local i = math.random( num )
            for k, v in pairs(t) do
                i = i - 1
                if i <= 0 then
                    picks[k] = v
                end
            end
        end
    end

    return picks
end

function printtable(t, highlight)
    if highlight ~= nil then
        print (highlight)
    end

    if t == nil then
        print ("The provided table is nil")
    elseif (type(t) ~= "table") then
        print ("The provided value is not a table:")
        print (t)
    else
        print ("Size: " .. #t)
        print ("Content: ")

        for pos,val in pairs(t) do
            print (string.format("Key: [ %s ] \t Value: [ %s ]", tostring(pos), tostring(val)))
        end
    end
end

local random_stack = {}
function push_random( fn )
    assert( type(fn) == "function" )
    table.insert( random_stack, math.random )
    math.random = fn
end

function pop_random()
    local fn = table.remove( random_stack )
    assert( type(fn) == "function" )
    math.random = fn
end

function weightedpick(options, nullweight, rng)
    nullweight = nullweight or 0
    local total = nullweight or 0
    for k,v in pairs(options) do
        total = total + math.max( 0, v )
    end
    if total <= 0 then
        return
    end
    local rand = (rng and rng:Random() or math.random()) * total
    if rand <= nullweight and nullweight > 0 then
        return nil
    end
    rand = rand - nullweight
   
    local option = next(options)
    while option do
        local wt = options[option]
        rand = rand - math.max( 0, wt )
        if rand <= 0 then
            return option, wt/total
        end
        option = next(options, option)
    end
    assert(option, "weighted random is messed up")
end

-- THis weighted table is in the form of an array consisting of: { value1, weight1, value2, weight2, ... valueN, weightN }
function weighted_arraypick(options, nullweight, rng)
    nullweight = nullweight or 0
    local total = nullweight or 0
    for i = 1, #options, 2 do
        local v = options[i+1]
        total = total + math.max( 0, v )
    end
    if total <= 0 then
        return
    end
    local rand = (rng and rng:Random() or math.random()) * total
    if rand <= nullweight and nullweight > 0 then
        return nil
    end
    rand = rand - nullweight
   
    for i = 1, #options, 2 do
        local option, wt = options[i], options[i+1]
        rand = rand - math.max( 0, wt )
        if rand <= 0 then
            return option, wt/total
        end
    end
end

-- Check that 'tags' has all the values contained within '...'
function all_tags( tags, ... )
    local n = select( "#", ... )
    for i = 1, n do
        local tag = select( i, ... )
        if tag ~= nil and not table.arrayfind( tags, tag ) then
            return false
        end
    end
    return true
end

-- Check that 'tags' has ANY of the values contained within '...'
-- NOTE: same as not no_tags()
function any_tags( tags, ... )
    local n = select( "#", ... )
    for i = 1, n do
        local tag = select( i, ... )
        if tag ~= nil and table.arrayfind( tags, tag ) then
            return true
        end
    end
    return false
end

-- Check that 'tags' has NONE of the values contained within '...'
-- NOTE: same as: not any_tags()
function no_tags( tags, ... )
    local n = select( "#", ... )
    for i = 1, n do
        local tag = select( i, ... )
        if tag ~= nil and table.arrayfind( tags, tag ) then
            return false
        end
    end
    return true
end

function math.fequal( a, b, eps )
    return math.abs( a - b ) < (eps or 0.001)
end

function math.random_sign()
    return math.random() <= 0.5 and 1 or -1
end

function math.length2d(x, y)
    return math.sqrt(x*x + y*y)
end

-- Choose a random number in a gaussian distribution.
-- Based on the polar form of the Box-Muller transformation.
function math.randomGauss( min, max )
    local mean, stddev = (min + max)/2, (max - min)/6
	local x1, x2, w
	repeat
		x1 = 2 * math.random() - 1
		x2 = 2 * math.random() - 1
		w = x1 * x1 + x2 * x2
	until w < 1.0

	w = math.sqrt( (-2 * math.log( w ) ) / w )
	local x = (x1 * w)*stddev + mean
    return clamp( x, min, max )
end

-- Generates a random binomial variate.
function math.randomBinomial( p, n )
    local x = 0
    for i = 0, n-1 do
        if math.random() < p then
            x = x + 1
        end
    end
    return x    
end

function printf(fmt, ...)
    print(string.format(fmt, ...))
end


local function sort_cmp(a, b)
    local anum, bnum = type(a) == "number", type(b) == "number"
    if anum and bnum then
        return a < b
    elseif anum then
        return true
    elseif bnum then
        return false
    else
        return tostring(a) < tostring(b)
    end
end

local largeTablePool = TablePool( 200 )

function ObtainWorkTable()
    return largeTablePool:ObtainTable() 
end

function ReleaseWorkTable( t )
    largeTablePool:ReleaseTable( t )
end

function GetLargeTablePool()
    return largeTablePool
end

function string:at( i )
    return self:sub( i, i )
end

function string:pmatch( pattern )
    return select( 2, pcall( string.match, self, pattern ))
end

function string:split(sep)
   local sep, fields = sep or " ", {}
   local pattern = string.format("([^%s]+)", sep)
   self:gsub(pattern, function(c) fields[#fields+1] = c end)
   return fields
end

function ParseTags( str )
    local t = {}
    if str then
        for i, token in ipairs(str:split()) do
            local tag, weight = token:match( "(.+)[:](%d+)$")
            tag = tag or token
            if #tag > 0 then
                t[ tag or token ] = tonumber(weight) or true
            end
        end
    end
    return t
end

local function ParseFormattingColour( attr )
    local jj, kk, subattr, clrattr = attr:find( "^(.*)#(.+)$" )
    if clrattr then
        if tonumber( clrattr, 16 ) then
            while #clrattr < 8 do clrattr = clrattr .. "f" end
            return tonumber( clrattr, 16 )
        elseif UICOLOURS[ clrattr ] then
            return UICOLOURS[ clrattr ]
        end
    end
end

function ParseImageInformation( attr )
    local lowerattr = string.lower(attr)
    local img = string.match(lowerattr, [[img=['’]([^'’]+)]])
    local scale = string.match(lowerattr, [[scale=([%d.]+)]])
    if scale then
        scale = tonumber(scale)
    end
    return img, scale
end

function ApplyFormatting( textnode, str )
    local j, k, sel, attr
    local spans = {}
    repeat
        -- find colourization.
        j, k, sel, attr = str:find( "<([#!bBcCsSiIuUpP/]?)([^>]*)>" )
        if j then
            if sel == '/' then
                -- Close the last span
                local attr = table.remove( spans )
                local sel = table.remove( spans )
                local start_idx = table.remove( spans )
                local end_idx = j - 1
                if textnode then
                    
                    if sel == "b" or sel == "B" then
                        textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_BOLD )
                    elseif sel == "i" or sel == "I" then
                        textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_ITALIC )
                    elseif sel == "u" or sel == "U" then
                        local clr = ParseFormattingColour( attr )
                        textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_UNDERLINE, clr )
                    elseif sel == "s" or sel == "S" then
                        textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_SHADOW )
                    elseif sel == '#' then
                        if tonumber( attr, 16 ) then
                            while #attr < 8 do attr = attr .. "f" end
                            textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_COLOUR, tonumber( attr, 16 ))
                        elseif UICOLOURS[attr] then
                            textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_COLOUR, UICOLOURS[attr] )
                        end
                        if attr == "UPGRADE" or attr == "DOWNGRADE" then
                            textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_SHADOW )
                        end

                    elseif sel == '!' and attr then
                        local clr = ParseFormattingColour( attr )
                        textnode:AddMarkup( start_idx, end_idx, engine.rendering.TextNode.MARKUP_LINK, attr, clr )
                    elseif sel == '' then
                        --just suppressing definitions
                    elseif sel ~= nil then
                        print( "Closing unknown text markup code", tostring(sel), debug.traceback(), str )
                    end
                end
            elseif sel == "p" or sel == "P" then
                -- p (picture) doesn't need a </> to close the span, it just inserts
                -- example options:
                -- <p img=folder/image.tex scale=1.0 advancement=75.0>
                local img, scale = ParseImageInformation( attr )

                local start_idx = j - 1
                if textnode then
                    textnode:AddMarkup( start_idx, start_idx, engine.rendering.TextNode.MARKUP_IMAGE, img, scale )
                end
                
                -- also insert a '/a' character into the string, so that the bitmap font renderer knows to render an image:
                -- (insert it AFTER the <blah> tag, so that it remains in the string after the tag is removed
                str = str:sub( 1, k ) .. '\a' .. str:sub( k + 1 ) 

            elseif sel == "c" or sel=="C" then
                -- replace <c img='r' scale=1.0>
                local img, scale = ParseImageInformation( attr )
                local device_id = TheGame:FE():GetControlDeviceID()
                if device_id and img then
                    local img_path = device_map[device_id] or device_map["XInput Controller"]
                    if img_path then
                        img_path = img_path .. img .. ".tex"

                        local start_idx = j - 1
                        if textnode then
                            textnode:AddMarkup( start_idx, start_idx, engine.rendering.TextNode.MARKUP_IMAGE, img_path, scale )
                        end

                        -- also insert a '/a' character into the string, so that the bitmap font renderer knows to render an image:
                        -- (insert it AFTER the <blah> tag, so that it remains in the string after the tag is removed
                        str = str:sub( 1, k ) .. '\a' .. str:sub( k + 1 )
                    else
                        -- Add a space instead of the controller image if there's no device_id, so the text box has a size
                        str = str:sub( 1, k ) .. ' ' .. str:sub( k + 1 )
                    end
                else
                    -- Add a space instead of the controller image if there's no device_id, so the text box has a size
                    str = str:sub( 1, k ) .. ' ' .. str:sub( k + 1 )
                end
            else
                table.insert( spans, j - 1 )
                table.insert( spans, sel )
                table.insert( spans, attr )
            end

            -- remove the <blah> tag from the string:
            str = str:sub( 1, j - 1 ) .. str:sub( k + 1 )
        end
    until not j
    
    return str
end


function SanitizeString( text, tag )
    repeat
        local ANY_TAG = "<([^>]*)>"
        local i, j = string.find( text, tag or ANY_TAG )
        if i then
            -- Remove matching close element only.
            local ei, ej, m = string.find( text, ANY_TAG, j + 1 )
            if ei and m == "/" then
                text = text:sub( 0, ei - 1 ) .. text:sub( ej + 1 )
            end
            text = text:sub( 0, i - 1 ) .. text:sub( j + 1 )
        end
    until i == nil
    return text
end

--[[function GetTextLinks( str )
    
    if str and type(str) == "string" then
        local j, k, txt
        local elems = {}
        repeat
            j, k, txt = str:find( "<([^/][^>]*)>", k )
            if j then
                table.insert(elems, {type="START", j = j, k = k, txt = txt:trim()})
            end
        until not j

        local ends = {}
        repeat
            j, k, txt = str:find( "</>", k )
            if j then
                table.insert(elems, {type = "END", j = j, k = k})
            end
        until not j

        table.sort(elems, function(a,b) return a.j > b.j end)

        local spans = {}
        local open = {}
        while #elems >0 do
            local elem = table.remove(elems)
            if elem.type == "START" then
                table.insert(open, elem)
            elseif elem.type == "END" then
                local starter = #open > 0 and table.remove(open)
                if starter.type == "START" and starter.txt and starter.txt:sub(1,1) == "!" then
                    table.insert(spans, {type = starter.txt:sub(2), txt = str:sub(starter.k+1, elem.j-1):trim()})
                end

            end
        end
        return spans
    end
end--]]

function SetFormattedText( textnode, str )
    str = tostring(str)
    if textnode then
        textnode:ClearMarkup()
    end
    str = ApplyFormatting( textnode, str )
    if textnode then
        textnode:SetText( str )
    end

    return str
end


function LOC( str )
    return TheGame:Str( str )
end

local _ENUM_META =
{
    __index = function( t, k ) error( "BAD ENUM ACCESS: "..tostring(k) ) end,
    __newindex = function( t, k, v ) error( "BAD ENUM ACCESS "..tostring(k) ) end,
}
function MakeEnum(args)
    local enum = {}
    for k,v in ipairs(args) do
        assert(type(v) == "string", "Enums come from strings")
        enum[v] = v
    end
    setmetatable( enum, _ENUM_META )
    return enum
end

function ExtendEnum( enum, args )
    setmetatable(enum, nil)
    for k,v in ipairs(args) do
        assert(type(v) == "string", "Enums come from strings")
        enum[v] = v
    end
    setmetatable( enum, _ENUM_META )
    return enum
end

function MakeEnumValues(args)
    local enum = {}
    for i = 1, #args do
        assert(type(args[i]) == "string", "Enums come from strings")
        enum[ args[i] ] = i
    end
    setmetatable( enum, _ENUM_META )
    return enum, args
end

function MakeEnumArray( enum )
    local t = {}
    for i, k, v in sorted_pairs( enum ) do
        table.insert( t, k )
    end
    return t
end

function IsEnum( val, enum )
    assert( getmetatable( enum ) == _ENUM_META )
    return val and rawget( enum, val ) == val
end

function MakeBitField(args)
    local bitfield = { NONE = 0, ANY = 0 }
    for k,v in ipairs(args) do
        assert(type(v) == "string", "Bitfields come from strings")
        assert(bitfield[v] == nil)
        local bit = math.floor(2^(k-1)) -- floor so we are dealing with proper Integers
        bitfield[v] = bit
        bitfield.ANY = bitfield.ANY | bit
    end
    setmetatable( bitfield, _ENUM_META )
    return bitfield, args
end

function StringizeBitField( bits, strings, sep )
    local str = {}
    for i, v in ipairs( strings ) do
        local flag = 1 << (i - 1)
        if bits & flag == flag then
            table.insert( str, v )
        end
    end
    if #str == 0 then
        return "NO-BITS"
    else
        return table.concat( str, sep or " " )
    end
end

function ParseBitField( bits, string )
    local t = string:split("|")
    local value = 0
    for i, bit in ipairs(t) do
        value = value | bits[ bit:trim() ]
    end
    return value
end

function IsBitField( val, bitfield )
    assert( getmetatable( bitfield ) == _ENUM_META )
    return (val & bitfield.ANY) ~= 0
end

function CheckAnyBits( bitfield, flags )
    return bitfield & flags ~= 0
end

function CheckBits( bitfield, flags )
    return bitfield & flags == flags
end

function SetBits( bitfield, flags )
    return bitfield | flags
end

function ClearBits( bitfield, flags )
    return bitfield & ~flags
end

function ToggleBits( bitfield, flags )
    return bitfield ~ flags
end

function RandomColourString()
    local r = math.random(0, 255)
    local g = math.random(0, 255)
    local b = math.random(0, 255)
    return string.format("#%02X%02X%02X", r, b, g)
end

function MakeColourString(r, g, b, a)
    return string.format( "%08X", MakeHexColour(r, g, b, a))
end

function ParseColourString( txt )
    local j, k, match = txt:find( "^0x([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])$")
    if match then
        return tonumber(match, 16)
    end
end

function MakeHexColour(r, g, b, a)
    r = clamp( math.floor( r*255 ), 0, 255 )
    g = clamp( math.floor( g*255 ), 0, 255 )
    b = clamp( math.floor( b*255 ), 0, 255 )
    a = clamp( math.floor( (a or 1)*255 ), 0, 255 )
    return (r << 24) | (g << 16) | (b << 8) | a
end

function HexColour(val, alpha)
    return
        ((val & 0xFF000000) >> 24)/255,
        ((val & 0xFF0000) >> 16)/255,
        ((val & 0xFF00) >> 8)/255,
        alpha or ((val & 0xFF) >> 0)/255
end

function THexColour(val)
    return {HexColour(val)}
end

function LerpColour( c1, c2, t )
    return
        c1[1] * (1-t) + c2[1] * t,
        c1[2] * (1-t) + c2[2] * t,
        c1[3] * (1-t) + c2[3] * t,
        c1[4] and c2[4] and (c1[4] * (1-t) + c2[4] * t)
end

function Lerp(x1,x2,t)
    return x1 + x2 * t - x1 * t
end

function TrueFunction()
    return true
end

function FalseFunction()
    return false
end

function NilFunction()
    return nil
end

--maaaaaybe we shouldn't be concatting strings like this in an inner loop...
function GetAnimStateForAngle( angle, prefix, old_flip )
    local flip = false
    local da = math.pi/8
    local anim = prefix or "walk"
    local abs_angle = math.abs(angle)

    if angle < 0 then
        flip = true
    end

    if abs_angle < da then
        anim = anim .. "_back"
        flip = old_flip
    elseif abs_angle < math.pi/2 - da then
        anim = anim .. "_back_3/4"
    elseif abs_angle < math.pi/2 + da then
        anim = anim .. "_side"
    elseif abs_angle < math.pi - da then
        anim = anim .. "_front_3/4"
    elseif angle <= math.pi then
        anim = anim.. "_front"
        flip = old_flip
    end

    return anim, flip
end

function pretty_float(v, precision)
    precision = precision or 2
    local s = string.format("%."..precision.."f", v):gsub("%.*0+$", "")
    return s
end

function FormatTime( total_time )
    local hours = math.floor(total_time / (60*60))
    local minutes = math.floor(total_time / 60 - hours*60)
    local seconds = math.floor(total_time - minutes*60 - hours*60*60)

    if hours > 0 then
        return string.format( "%02d:%02d:%02d", hours, minutes, seconds )
    else
        return string.format( "%02d:%02d", minutes, seconds )
    end
end

function ParseVersion( s )
    local major, minor, patch = s:match( "^(%d+)[.](%d+)[.](%d+)$" )
    return tonumber( major ), tonumber( minor ), tonumber( patch )
end

function IsCompatibleVersion( v1, v2 )
    local major1, minor1, patch1 = ParseVersion( v1 )
    local major2, minor2, patch2 = ParseVersion( v2 )
    return major1 and major2 and major1 >= major2
end

function CreateRecyclers(MAX_SIZE, createfn, recyclefn, clearfn, start_size)
    local t = {}
    
    local create = function(...) 
        if #t == 0 then
            return createfn(...)
        end
        local node = table.remove(t)

        if recyclefn then
            recyclefn(node, ...)
        end
        return node
    end

    local destroy = function(node)
        if #t < MAX_SIZE then
            if clearfn then
                clearfn(node)
            end

            table.insert(t, node)            
        end    
    end

    
    if start_size then
        for k = 1, start_size do
            table.insert(t, createfn())
        end
    end

    return create, destroy, t
end

-- Calculate the shortest distance between two axis-aligned rectangles.
function CalculateRectDistance( from_x1, from_y1, from_x2, from_y2,
                                to_x1, to_y1, to_x2, to_y2 )

    local right = to_x1 > from_x2
    local left = to_x2 < from_x1
    local top = to_y1 > from_y2
    local bottom = to_y2 < from_y1
    local dist

    if top and right then
        dist = distance( from_x2, from_y2, to_x1, to_y1 )
    elseif bottom and right then
        dist = distance( from_x2, from_y1, to_x1, to_y2 )
    elseif bottom and left then
        dist = distance( from_x1, from_y1, to_x2, to_y2 )
    elseif top and left then
        dist = distance( from_x1, from_y2, to_x2, to_x1 )
    elseif right then
        dist = to_x1 - from_x2
    elseif left then
        dist = from_x1 - to_x2
    elseif top then
        dist = to_y1 - from_y2
    elseif bottom then
        dist = from_y1 - to_y2
    else
        -- overlapping.
        dist = 0
    end

    return dist
end

-- Calculate AABB intersection, assuming these rectangles intersect.
function IntersectRect( x1min, y1min, x1max, y1max, x2min, y2min, x2max, y2max )
    local xmin = math.max( x1min, x2min )
    local xmax = math.min( x1max, x2max )
    local ymin = math.max( y1min, y2min )
    local ymax = math.min( y1max, y2max )
    return xmin, ymin, xmax, ymax
end

function CalcInputVectors(dimensions)
    local out_table = {}
    local fn
    fn = function(path, idx)
        path = path or {}
        idx = idx or 1
        local dim = dimensions[idx]
        if dim then
            for d = 1, dim do
                local new_path = shallowcopy(path)
                table.insert(new_path, d)
                fn(new_path, idx+1)
            end
        else
            
            table.insert(out_table, path)
            return
        end
    end
    fn()
    return out_table
end

function IsIdentityHSV(hsvtable)
    return  hsvtable.shadows.hsv[1] == 0 and hsvtable.shadows.hsv[2] == 1 and hsvtable.shadows.hsv[3] == 1 and
            hsvtable.midtones.hsv[1] == 0 and hsvtable.midtones.hsv[2] == 1 and hsvtable.midtones.hsv[3] == 1 and
            hsvtable.highlights.hsv[1] == 0 and hsvtable.highlights.hsv[2] == 1 and hsvtable.highlights.hsv[3] == 1 and
            hsvtable.shadows.colour[1] == 1 and hsvtable.shadows.colour[2] == 1 and hsvtable.shadows.colour[3] == 1 and
            hsvtable.midtones.colour[1] == 1 and hsvtable.midtones.colour[2] == 1 and hsvtable.midtones.colour[3] == 1 and
            hsvtable.highlights.colour[1] == 1 and hsvtable.highlights.colour[2] == 1 and hsvtable.highlights.colour[3] == 1
end


function SendMetricsData( event_name, t )
    assert( type( event_name ) == "string" )
    if t == nil then
        t = {}
    end
    t.INSTALLED_MODS = Content.GetInstalledMods()
    local json_data = json.encode( t )
    engine.inst:SendMetricsData(event_name, json_data )
end


local function tl_loader( module_name )
    local tl = require "util/tl"
    local found_filename, fd, tried = tl.search_module(module_name, false)
    if fd then
        fd:close()
    end

    if found_filename then
        local r, err = tl.process( found_filename )
        local errs = {}
        if #r.type_errors > 0 then
            for i, t in ipairs( r.type_errors ) do
                local str = string.format( "%s (%d,%d): %s", t.filename, t.x, t.y, t.msg )
                table.insert( errs, str )
            end
        end
        if #r.syntax_errors > 0 then
            for i, t in ipairs( r.syntax_errors ) do
                local str = string.format( "%s (%d,%d): %s", t.filename, t.x, t.y, t.msg )
                table.insert( errs, str )
            end
        end
        if #errs > 0 then
            local str = table.concat( errs, "\n" )
            engine.inst:ShowWarning( debug.traceback( str ), 2 )
        end

        local program = r.ast
        local code = tl.pretty_print_ast(program, true)
        local chunk, err = load(code, module_name, "t")
        if chunk then
            return function()
                local ret = chunk()
                package.loaded[module_name] = ret
                return ret
            end
        else
            error("Internal Compiler Error: Teal generator produced invalid Lua. Please report a bug at https://github.com/teal-language/tl")
        end
   end
   return table.concat(tried, "\n\t")
end

function InjectTealLoader()
    table.insert(package.searchers, tl_loader)
end

function table.all(table, fn)
    for k, v in ipairs(table) do
        if not fn(v) then
            return false
        end
    end
    return #table > 0 
end

function table.any(table, fn)
    for k, v in ipairs(table) do
        if fn(v) then
            return true
        end 
    end
    return false
end
