local WorkPosition = class("WorkPosition")

function WorkPosition:init( owner, id )
    assert( is_instance( owner, Location ))
    self.content = owner:GetWorkData( id )
    assert( self.content, id )
    self.location = owner

    self.id = id
end

function WorkPosition:Save()
    return 
    {
        id = self.id,
        location = self.location,
        agent = self.agent,
    }
end

function WorkPosition:Load()
    -- Yup.  Save compatability for when work data had two names (local and content ID)
    if self.id == "sal_spree_contact" then self.id = "faction_contact" end

    self.content = self.location:GetWorkData( self.id )
    if not self.content then
        local id = self.id:match( "[.](.+)$" )
        if id then
            self.id = id
            self.content = self.location:GetWorkData( self.id )
        end
 
        -- This should be checked in PostLoad as appropriate, to remove any invalid work positions.
        if not self.content then
            LOGWARN ("missing work position: %s", self.id )
        end
    end
end

function WorkPosition:GetData()
    return self.content
end

function WorkPosition:GetTitle()
    return self.content.title
end

function WorkPosition:IsBoss()
    return self.content.role == CHARACTER_ROLES.PROPRIETOR
end


function WorkPosition:GetRole()
    return self.content.role
end

function WorkPosition:GetLocation()
    return self.location
end

function WorkPosition:GetAgent()
    return self.agent
end

function WorkPosition:GetSchedule()
    return self.content.hours or 0
end

function WorkPosition:ShouldBeWorking()
    return self:GetSchedule() & TheGame:GetGameState():GetDayPhase() ~= 0
end

function WorkPosition:TryHire(verbose)
    local best_agent
    local character_def
    if self.content.promote_def then
        if type(self.content.promote_def) == "string" then
            character_def = self.content.promote_def
        else
            character_def = table.arraypick(self.content.promote_def)
        end
    end

    if character_def then
        best_agent = TheGame:GetGameState():AddSkinnedAgent( character_def )

        if verbose then
            print( "\tSpawned", self.id, best_agent )
        end
    else
        if verbose then
            print( "\tNo character def defined for spawn fallback" )
        end
    end

    if best_agent then
        if verbose then
            print ("picked ", self.id, best_agent )
        end
        self:Hire(best_agent)
    end
end

function WorkPosition:Hire(agent)
    assert( self.agent == nil )
    self.agent = agent

    local old_work_pos = self.agent:GetBrain():GetWorkPosition() 
    assert( old_work_pos ~= self )
    if old_work_pos then
        old_work_pos:Fire()
    end
    
    local role = self.content.role

    if self.content.cardshop then
        self.agent:GainAspect("cardshop")
        self.agent:GetAspect("cardshop"):SetStockID(self.content.cardshop)
    end

    self.agent:GetBrain():SetWork(self)

    if self.content.on_hire then
        self.content.on_hire( self, agent )
    end
end

function WorkPosition:SetPaused(val)
    self.paused = val
end

function WorkPosition:Fire()
    assert( self.agent ) -- Stop firing if there's nobody to fire.

    if self.content.cardshop then
        self.agent:RemoveAspect("cardshop")
    end

    if self.agent and self.agent:GetBrain() then
        self.agent:GetBrain():SetWork(nil)
    end
    
    self.agent = nil
end

function WorkPosition:DoHiring()
    if not self.agent and not self.paused and self.content.hiring then
        self:TryHire()
    end
end


local function ValidatePromoteDef( defs )
    if defs then
        if type(defs) == "string" then
            if not Content.GetCharacterDef( defs ) then
                assert(false, string.format( "INVALID CHARACTER DEF %s", defs))
            end
        elseif type(defs) == "table" then
            for k,v in ipairs(defs) do
                if not Content.GetCharacterDef( v ) then
                    assert(false, string.format( "INVALID CHARACTER DEF %s", v))
                end
            end
        end
    end
end 


function CreateProductionWorkplace(num_workers, phase_mask, boss_def, worker_def, boss_title, worker_title)
    ValidatePromoteDef( boss_def )
    ValidatePromoteDef( worker_def )

    worker_title = worker_title or "WORKER"
    boss_title = boss_title or "WORKER_BOSS"

    local work = 
    {
        foreman =
        {     
            hiring = true, 
            title = boss_title, 
            role = CHARACTER_ROLES.PROPRIETOR, 
            hours = phase_mask, 
            promote_def = boss_def,
        },
    }

    for k = 1, num_workers do
        local t =
        {
            hiring = true, 
            title = worker_title,
            role = CHARACTER_ROLES.LABOUR,
            hours = phase_mask,
            promote_def = worker_def,
        }
        work["worker"..tostring(k)] = t
    end

    return work
end

function CreateLabourJob( phase_mask, title, role, promote_def, on_hire)
    ValidatePromoteDef(promote_def)
    assert( on_hire == nil or type(on_hire) == "function" )

    return  {
                hiring = true,
                title = title,
                role = role,
                promote_def = promote_def,
                hours = phase_mask,
                on_hire = on_hire,
            }
end



function CreateMerchantJob( phase_mask, title, role, promote_def, shop_def )
    ValidatePromoteDef(promote_def)
    local t = CreateLabourJob( phase_mask, title, role, promote_def)    
    t.cardshop = shop_def
    return t
end

function CreateClosedJob( phase_mask, title, role, shop_def, alias_id )
    return  {
                hiring = false, 
                title = title,
                role = role,
                promote_def = nil,
                hours = phase_mask or PHASE_MASK_ALL,
                cardshop = shop_def,
            }

end
